// *****************************************************************************
//                             Exentia Library
//                        by Stefano Tommesani 2000->2002
//                 Updates by: Joao Paulo Schwarz Schuler
//                             Robert Lee
//                             Chris Rorden
//                             Patrick van Laake
//
//                              version 0.4.0
//                  Web: http://exentia.tommesani.com
//                       E-mail: stefano@tommesani.com
//
// The contents of this file are subject to the Mozilla Public License Version
// 1.0; you may not use this file except in compliance with the License.
// You may obtain a copy of the License at http://www.mozilla.org/MPL/
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
// for the specific language governing rights and limitations under the License.
// *****************************************************************************

//{$DEFINE DELPHI_CODE}

unit Exentia;

interface

{$IFDEF LINUX}
uses SysUtils,Qdialogs;
{$else}
uses SysUtils,dialogs;
{$ENDIF}

type
  TSingleArray = array [0..0] of single;
  PSingleArray = ^TSingleArray;

  TPointerArray = array [0..0] of pointer; //JPSS 14/3/2001
  PPointerArray = ^TPointerArray;          //JPSS 14/3/2001

  TLongWordArray = array [0..0] of LongWord; //JPSS 14/3/2001
  PLongWordArray = ^TLongWordArray;          //JPSS 14/3/2001

type
  TFVector = class(TObject)
  private
    AlignedArray : PSingleArray;
    AllocatedArray : pointer;
    NumElements : integer;
    function GetDataElement(index : integer) : single;
    procedure SetDataElement(index : integer; const value : single);
  public
    property DataArray : PSingleArray read AlignedArray;
    property Data[index : integer] : single read GetDataElement write SetDataElement; default;
    constructor Create(const ALength : integer);
    constructor CreateUsingArray(var A:array of Single);                // JPSS  28/02/2001
    constructor Copy(const V : TFVector); // Self = V
    destructor Destroy;override;
    class function BuildVector(const ALength : integer) : TFVector;
    class function BuildUsingArray(var A:array of Single) : TFVector;   // JPSS  28/02/2001
    class function CopyVector(const V : TFVector) : TFVector;
    class function isSSE : boolean;
    class function is3DNow : boolean;
    procedure UseArray(var A:array of single);                          // JPSS  28/02/2001
    procedure SetLength(const ALength : integer);
    procedure CopyArray(const V : TFVector);                            // FIXED by JPSS 7/4/2001
    procedure ImportArray(const Source : PSingleArray; const Length : integer = 0);
    procedure ExportArray(Destination : PSingleArray; const Length : integer = 0);
    procedure ClearArray; // Self = 0
    procedure SetArray(value : single);  // Self[i] = value
    function Equal(V : TFVector) : boolean;
    // basic operations
    procedure Add(V : TFVector); virtual; abstract;  // Self = Self + V
    procedure Sub(V : TFVector); virtual; abstract;  // Self = Self - V
    procedure Mul(V : TFVector); virtual; abstract;  // Self = Self * V
    procedure Divide(V : TFVector); virtual; abstract;  // Self = Self / V
    procedure Sqrt(V : TFVector); virtual; abstract;  // Self = Sqrt(V)
    procedure Reciprocal(V : TFVector); virtual; abstract;  // Self = 1 / V
    procedure RecSqrt(V : TFVector); virtual; abstract;  // Self = 1 / Sqrt(V)
    procedure Max(V : TFVector); virtual; abstract;  // Self = Max(Self, V)
    procedure Min(V : TFVector); virtual; abstract;  // Self = Min(Self, V)
    procedure LogicalAnd(V : TFVector); virtual; abstract;  // Self = Self and V
    procedure LogicalOr(V : TFVector); virtual; abstract;   // Self = Self or V
    procedure LogicalXor(V : TFVector); virtual; abstract;  // Self = Self xor V
    procedure CmpGreater(V : TFVector); virtual; abstract;       // Self = Self > V
    procedure CmpLower(V : TFVector); virtual; abstract;         // Self = Self < V
    procedure CmpEqual(V : TFVector); virtual; abstract;         // Self = Self == V
    procedure CmpGreaterEqual(V : TFVector); virtual; abstract;  // Self = Self >= V
    procedure CmpLowerEqual(V : TFVector); virtual; abstract;    // Self = Self <= V
    // complex operations
    procedure AddSquare(V : TFVector); virtual; abstract;   // Self = Self + V^2
    function InnerSum : single; virtual; abstract;  // result = Sum(Self[i])
    function InnerSumAbs : single; virtual; abstract;  // result = Sum(abs(Self[i]))
    function DotProduct(V : TFVector) : single; virtual; abstract;  // result = Sum(Self[i] * V[i])
    function MaxValue : single; virtual; abstract;  // result = Max(Self[i])
    function MaxAbsValue : single; virtual; abstract;  // result = Max(abs(Self[i]))
    function IndexMaxValue : integer; virtual; abstract;  // result = index of Max(Self)
    function MinValue : single; virtual; abstract;  // result = Min(Self[i])
    function MinAbsValue : single; virtual; abstract;  // result = Min(abs(Self[i]))
    function IndexMinValue : integer; virtual; abstract;  // result = index of Min(Self)
    function Mean : single; virtual; abstract;  // result = Sum(Self[i]) / NumElements
    procedure Scale(value : single); virtual; abstract;  // Self = value * Self
  end;

type
  TX87Vector = class(TFVector)
  public
    constructor Create(const ALength : integer);
    constructor CreateUsingArray(var A:array of single);
    destructor Destroy;override;
    procedure Add(V : TFVector); override;
    procedure Sub(V : TFVector); override;
    procedure Mul(V : TFVector); override;
    procedure Divide(V : TFVector); override;
    procedure Sqrt(V : TFVector); override;
    procedure Reciprocal(V : TFVector); override;
    procedure RecSqrt(V : TFVector); override;
    procedure Max(V : TFVector); override;
    procedure Min(V : TFVector); override;
    procedure LogicalAnd(V : TFVector); override;  // JPSS FIXED 7/4/2001
    procedure LogicalOr(V : TFVector); override;   // JPSS FIXED 7/4/2001
    procedure LogicalXor(V : TFVector); override;  // JPSS FIXED 7/4/2001
    procedure CmpGreater(V : TFVector); override;  // JPSS FIXED 7/4/2001
    procedure CmpLower(V : TFVector); override;    // JPSS FIXED 7/4/2001
    procedure CmpEqual(V : TFVector); override;    // JPSS FIXED 7/4/2001
    procedure CmpGreaterEqual(V : TFVector); override; // JPSS FIXED 7/4/2001
    procedure CmpLowerEqual(V : TFVector); override;   // JPSS FIXED 7/4/2001
    procedure AddSquare(V : TFVector); override;
    function InnerSum : single; override;
    function InnerSumAbs : single; override;
    function DotProduct(V : TFVector) : single; override;
    function MaxValue : single; override;
    function MaxAbsValue : single; override;
    function IndexMaxValue : integer; override;
    function MinValue : single; override;
    function MinAbsValue : single; override;
    function IndexMinValue : integer; override;
    function Mean : single; override;
    procedure Scale(value : single); override;
  end;

type
  TSSEVector = class(TX87Vector)
  public
    constructor Create(const ALength : integer);
    constructor CreateUsingArray(var A:array of Single);                // JPSS  28/02/2001
    destructor Destroy;override;
    procedure Add(V : TFVector); override;
    procedure Sub(V : TFVector); override;
    procedure Mul(V : TFVector); override;
    procedure Divide(V : TFVector); override;
    procedure Sqrt(V : TFVector); override;
    procedure Reciprocal(V : TFVector); override;
    procedure RecSqrt(V : TFVector); override;
    procedure Max(V : TFVector); override;
    procedure Min(V : TFVector); override;
    procedure LogicalAnd(V : TFVector); override;
    procedure LogicalOr(V : TFVector); override;
    procedure LogicalXor(V : TFVector); override;
    procedure CmpGreater(V : TFVector); override;
    procedure CmpLower(V : TFVector); override;
    procedure CmpEqual(V : TFVector); override;
    procedure CmpGreaterEqual(V : TFVector); override;
    procedure CmpLowerEqual(V : TFVector); override;
    procedure AddSquare(V : TFVector); override;
  end;

type
  T3DNowVector = class(TX87Vector)
  public
    constructor Create(const ALength : integer);
    constructor CreateUsingArray(var A:array of Single);                // JPSS  28/02/2001
    destructor Destroy;override;
    procedure Add(V : TFVector); override;
    procedure Sub(V : TFVector); override;
    procedure Mul(V : TFVector); override;
    procedure Divide(V : TFVector); override;
    // procedure Sqrt(V : TFVector); override;
    procedure Reciprocal(V : TFVector); override;
    // procedure RecSqrt(V : TFVector); override;
    procedure Max(V : TFVector); override;
    procedure Min(V : TFVector); override;
    procedure LogicalAnd(V : TFVector); override;
    procedure LogicalOr(V : TFVector); override;
    procedure LogicalXor(V : TFVector); override;
    procedure CmpGreater(V : TFVector); override;
    procedure CmpLower(V : TFVector); override;
    procedure CmpEqual(V : TFVector); override;
    procedure CmpGreaterEqual(V : TFVector); override;
    procedure CmpLowerEqual(V : TFVector); override;
    procedure AddSquare(V : TFVector); override;
    procedure Scale(value : single); override;
  end;

type
  ETestException = class(Exception);

//procedure TestBinaryBasicOps( var TestVector1, TestVector2,TestVector:TFVector);
//procedure TestBinaryComplexOps( var TestVector1, TestVector2,TestVector:TFVector);


implementation
{$R-}     // JPSS 28/02/2001

const cTestVectorSize = 100;

var
  EnableSSE : boolean;
  Enable3DNow : boolean;

// *************************************************** //
// *                      TFVector                   * //
// *************************************************** //

constructor TFVector.Create(const ALength : integer);
begin
  inherited Create;
  AlignedArray := nil;
  AllocatedArray := nil;
  NumElements := 0;
  SetLength(ALength);
end;

constructor TFVector.CreateUsingArray(var A:array of Single);
begin
  inherited Create;
  AlignedArray := addr(A);
  AllocatedArray := nil;
  NumElements := Length(A);
end;

procedure TFVector.UseArray(var A:array of single); // JPSS  28/02/2001
begin
  if AllocatedArray <> nil
     then FreeMem(AllocatedArray);
  AlignedArray := addr(A);
  AllocatedArray := nil;
  NumElements := Length(A);
end;

constructor TFVector.Copy(const V: TFVector);
begin
  inherited Create;
  AlignedArray := nil;
  AllocatedArray := nil;
  NumElements := 0;
  SetLength(V.NumElements);
  move(V.AlignedArray, AlignedArray, NumElements * SizeOf(Single));
end;

destructor TFVector.Destroy;
begin
  if AllocatedArray <> nil
     then FreeMem(AllocatedArray);
  AllocatedArray := nil;
  inherited Destroy;
end;

class function TFVector.BuildVector(const ALength : integer) : TFVector;
begin
  if EnableSSE
     then result := TSSEVector.Create(ALength)
     else if Enable3DNow
             then result := T3DNowVector.Create(ALength)
             else result := TX87Vector.Create(ALength);
end;

class function TFVector.BuildUsingArray(var A:array of Single) : TFVector;
begin
  if EnableSSE
     then result := TSSEVector.CreateUsingArray(A)
     else if Enable3DNow
             then result := T3DNowVector.CreateUsingArray(A)
             else result := TX87Vector.CreateUsingArray(A);
end;

class function TFVector.CopyVector(const V : TFVector) : TFVector;
begin
  if EnableSSE
     then result := TSSEVector.Copy(V)
     else if Enable3DNow
             then result := T3DNowVector.Copy(V)
             else result := TX87Vector.Copy(V);
end;

class function TFVector.isSSE : boolean;
begin
  result := EnableSSE;
end;

class function TFVector.is3DNow : boolean;
begin
  result := Enable3DNow;
end;

function TFVector.GetDataElement(index : integer) : single;
begin
  result := AlignedArray[index];
end;

procedure TFVector.SetDataElement(index : integer; const value : single);
begin
  AlignedArray[index] := value;
end;

procedure TFVector.SetLength(const ALength : integer);
begin
  if (AllocatedArray <> nil) and (ALength = NumElements)
     then exit;
  if AllocatedArray <> nil
     then FreeMem(AllocatedArray);
  // rounds length on blocks of 4 elements
  if (ALength and 3) > 0
     then NumElements := (ALength and ($FFFFFFFF - 3)) + 4
     else NumElements := ALength;
  try
    GetMem(AllocatedArray, (NumElements * SizeOf(single)) + 16);
    // AlignedArray is aligned on 16 bytes boundary
    AlignedArray := PSingleArray((integer(AllocatedArray) and $FFFFFFF0) + 16);
  except
    on EOutOfMemory do
       begin
       NumElements := 0;
       AllocatedArray := nil;
       AlignedArray := nil;
       raise;  // re-raise exception to calling function
       end;
  end;
end;

procedure TFVector.CopyArray(const V : TFVector);   // JPSS FIXED 7/4/2001
begin
  if NumElements <> V.NumElements
     then SetLength(V.NumElements);
  move(V.AlignedArray^, AlignedArray^, SizeOf(single) * NumElements);
end;

procedure TFVector.ImportArray(const Source : PSingleArray; const Length : integer = 0);
begin
  if (Length > 0) and (Length <> NumElements)
     then begin
          SetLength(Length);
          move(Source, AlignedArray, Length * SizeOf(Single));
          end
     else move(Source, AlignedArray, NumElements * SizeOf(Single));
end;

procedure TFVector.ExportArray(Destination : PSingleArray; const Length : integer = 0);
begin
  if Length > 0
     then move(AlignedArray, Destination, Length * SizeOf(Single))
     else move(AlignedArray, Destination, NumElements * SizeOf(Single));
end;

procedure TFVector.SetArray(value : single);
var counter : integer;
begin
  for counter := 0 to (NumElements - 1) do
      AlignedArray[counter] := value;
end;

procedure TFVector.ClearArray;
begin
  fillchar(AlignedArray, NumElements * SizeOf(Single), 0);
end;

function TFVector.Equal(V : TFVector) : boolean;
var counter : integer;
begin
  result := true;
  if NumElements <> V.NumElements
     then begin
          result := false;
          exit;
          end;
  for counter := 0 to (NumElements - 1) do
      if Self[counter] <> V[counter]
         then begin
              result := false;
              break;
              end;
end;

// *************************************************** //
// *                    TSSEVector                   * //
// *************************************************** //

constructor TSSEVector.Create(const ALength : integer);
begin
  inherited Create(ALength);
end;

constructor TSSEVector.CreateUsingArray(var A:array of single);
begin
  inherited CreateUsingArray(A);
end;

destructor TSSEVector.Destroy;
begin
  inherited Destroy;
end;

procedure TSSEVector.Add(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeAddLoop
@LargeAddLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$58,$C1           /// addps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$58,$D3           /// addps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$58,$E5           /// addps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$58,$F7           /// addps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeAddLoop

@SkipLargeAddLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndAdd
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallAddLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$58,$C1           /// addps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallAddLoop

@EndAdd:
end;

procedure TSSEVector.Sub(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeSubLoop
@LargeSubLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$5C,$C1           /// subps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$5C,$D3           /// subps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$5C,$E5           /// subps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$5C,$F7           /// subps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeSubLoop

@SkipLargeSubLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndSub
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallSubLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$5C,$C1           /// subps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallSubLoop

@EndSub:
end;

procedure TSSEVector.Mul(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeMulLoop
@LargeMulLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$59,$C1           /// mulps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$59,$D3           /// mulps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$59,$E5           /// mulps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$59,$F7           /// mulps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeMulLoop

@SkipLargeMulLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndMul
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallMulLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$59,$C1           /// mulps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallMulLoop

@EndMul:
end;

procedure TSSEVector.Divide(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeDivLoop
@LargeDivLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$5E,$C1           /// divps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$5E,$D3           /// divps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$5E,$E5           /// divps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$5E,$F7           /// divps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeDivLoop

@SkipLargeDivLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndDiv
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallDivLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$5E,$C1           /// divps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallDivLoop

@EndDiv:
end;

procedure TSSEVector.Sqrt(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeSqrtLoop
@LargeSqrtLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$51,$C1           /// sqrtps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$51,$D3           /// sqrtps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$51,$E5           /// sqrtps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$51,$F7           /// sqrtps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeSqrtLoop

@SkipLargeSqrtLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndSqrt
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallSqrtLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$51,$C1           /// sqrtps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallSqrtLoop

@EndSqrt:
end;

procedure TSSEVector.Reciprocal(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeReciprocalLoop
@LargeReciprocalLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$53,$C1           /// rcpps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$53,$D3           /// rcpps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$53,$E5           /// rcpps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$53,$F7           /// rcpps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeReciprocalLoop

@SkipLargeReciprocalLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndReciprocal
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallReciprocalLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$53,$C1           /// rcpps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallReciprocalLoop

@EndReciprocal:
end;

procedure TSSEVector.RecSqrt(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeRSqrtLoop
@LargeRSqrtLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$52,$C1           /// rsqrtps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$52,$D3           /// rsqrtps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$52,$E5           /// rsqrtps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$52,$F7           /// rsqrtps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeRSqrtLoop

@SkipLargeRSqrtLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndRSqrt
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallRSqrtLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$52,$C1           /// rsqrtps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallRSqrtLoop

@EndRSqrt:
end;

procedure TSSEVector.Max(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeMaxLoop
@LargeMaxLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$5F,$C1           /// maxps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$5F,$D3           /// maxps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$5F,$E5           /// maxps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$5F,$F7           /// maxps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeMaxLoop

@SkipLargeMaxLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndMax
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallMaxLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$5F,$C1           /// maxps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallMaxLoop

@EndMax:
end;

procedure TSSEVector.Min(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeMinLoop
@LargeMinLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$5D,$C1           /// minps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$5D,$D3           /// minps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$5D,$E5           /// minps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$5D,$F7           /// minps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeMinLoop

@SkipLargeMinLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndMin
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallMinLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$5D,$C1           /// minps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallMinLoop

@EndMin:
end;

procedure TSSEVector.LogicalAnd(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeLogicalAndLoop
@LargeLogicalAndLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$54,$C1           /// andps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$54,$D3           /// andps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$54,$E5           /// andps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$54,$F7           /// andps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeLogicalAndLoop

@SkipLargeLogicalAndLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndLogicalAnd
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallLogicalAndLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$54,$C1           /// andps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallLogicalAndLoop

@EndLogicalAnd:
end;

procedure TSSEVector.LogicalOr(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeLogicalOrLoop
@LargeLogicalOrLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$56,$C1           /// orps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$56,$D3           /// orps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$56,$E5           /// orps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$56,$F7           /// orps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeLogicalOrLoop

@SkipLargeLogicalOrLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndLogicalOr
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallLogicalOrLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$56,$C1           /// orps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallLogicalOrLoop

@EndLogicalOr:
end;

procedure TSSEVector.LogicalXor(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeLogicalXorLoop
@LargeLogicalXorLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$57,$C1           /// xorps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$57,$D3           /// xorps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$57,$E5           /// xorps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$57,$F7           /// xorps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeLogicalXorLoop

@SkipLargeLogicalXorLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndLogicalXor
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallLogicalXorLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$57,$C1           /// xorps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallLogicalXorLoop

@EndLogicalXor:
end;

procedure TSSEVector.CmpGreater(V : TFVector);
// Self > V becomes V < Self
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpGreaterLoop
@LargeCmpGreaterLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$C2,$C8,$01       /// cmpltps xmm1,xmm0
  db $0F,$29,$08           /// movaps [eax],xmm1

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$C2,$DA,$01       /// cmpltps xmm3,xmm2
  db $0F,$29,$58,$10       /// movaps [eax+16],xmm3

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$C2,$EC,$01       /// cmpltps xmm5,xmm4
  db $0F,$29,$68,$20       /// movaps [eax+32],xmm5

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$C2,$FE,$01       /// cmpltps xmm7,xmm6
  db $0F,$29,$78,$30       /// movaps [eax+48],xmm7

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpGreaterLoop

@SkipLargeCmpGreaterLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpGreater
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpGreaterLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$C2,$C8,$01       /// cmpltps xmm1,xmm0
  db $0F,$29,$08           /// movaps [eax],xmm1
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpGreaterLoop

@EndCmpGreater:
end;

procedure TSSEVector.CmpLower(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpLowerLoop
@LargeCmpLowerLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$C2,$C1,$01       /// cmpltps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$C2,$D3,$01       /// cmpltps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$C2,$E5,$01       /// cmpltps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$C2,$F7,$01       /// cmpltps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpLowerLoop

@SkipLargeCmpLowerLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpLower
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpLowerLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$C2,$C1,$01       /// cmpltps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpLowerLoop

@EndCmpLower:
end;

procedure TSSEVector.CmpEqual(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpEqualLoop
@LargeCmpEqualLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$C2,$C1,$00       /// cmpeqps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$C2,$D3,$00       /// cmpeqps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$C2,$E5,$00       /// cmpeqps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$C2,$F7,$00       /// cmpeqps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpEqualLoop

@SkipLargeCmpEqualLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpEqual
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpEqualLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$C2,$C1,$00       /// cmpeqps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpEqualLoop

@EndCmpEqual:
end;

procedure TSSEVector.CmpGreaterEqual(V : TFVector);
// A >= B becomes B <= A
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpGreaterEqualLoop
@LargeCmpGreaterEqualLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$C2,$C8,$02       /// cmpleps xmm1,xmm0
  db $0F,$29,$08           /// movaps [eax],xmm1

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$C2,$DA,$02       /// cmpleps xmm3,xmm2
  db $0F,$29,$58,$10       /// movaps [eax+16],xmm3

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$C2,$EC,$02       /// cmpleps xmm5,xmm4
  db $0F,$29,$68,$20       /// movaps [eax+32],xmm5

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$C2,$FE,$02       /// cmpleps xmm7,xmm6
  db $0F,$29,$78,$30       /// movaps [eax+48],xmm7

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpGreaterEqualLoop

@SkipLargeCmpGreaterEqualLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpGreaterEqual
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpGreaterEqualLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$C2,$C8,$02       /// cmpleps xmm1,xmm0
  db $0F,$29,$08           /// movaps [eax],xmm1
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpGreaterEqualLoop

@EndCmpGreaterEqual:
end;

procedure TSSEVector.CmpLowerEqual(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpLowerEqualLoop
@LargeCmpLowerEqualLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$C2,$C1,$02       /// cmpleps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$C2,$D3,$02       /// cmpleps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$C2,$E5,$02       /// cmpleps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$C2,$F7,$02       /// cmpleps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpLowerEqualLoop

@SkipLargeCmpLowerEqualLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpLowerEqual
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpLowerEqualLoop:
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$C2,$C1,$02       /// cmpleps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpLowerEqualLoop

@EndCmpLowerEqual:
end;

procedure TSSEVector.AddSquare(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeAddSquareLoop
@LargeAddSquareLoop:
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$18,$4A,$40       /// prefetcht0 [edx+64]
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$18,$48,$40       /// prefetcht0 [eax+64]
  db $0F,$59,$C9           /// mulps xmm1,xmm1
  db $0F,$58,$C1           /// addps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0

  db $0F,$28,$5A,$10       /// movaps xmm3,[edx+16]
  db $0F,$28,$50,$10       /// movaps xmm2,[eax+16]
  db $0F,$59,$DB           /// mulps xmm3,xmm3
  db $0F,$58,$D3           /// addps xmm2,xmm3
  db $0F,$29,$50,$10       /// movaps [eax+16],xmm2

  db $0F,$28,$6A,$20       /// movaps xmm5,[edx+32]
  db $0F,$28,$60,$20       /// movaps xmm4,[eax+32]
  db $0F,$59,$ED           /// mulps xmm5,xmm5
  db $0F,$58,$E5           /// addps xmm4,xmm5
  db $0F,$29,$60,$20       /// movaps [eax+32],xmm4

  db $0F,$28,$7A,$30       /// movaps xmm7,[edx+48]
  db $0F,$28,$70,$30       /// movaps xmm6,[eax+48]
  db $0F,$59,$FF           /// mulps xmm7,xmm7
  db $0F,$58,$F7           /// addps xmm6,xmm7
  db $0F,$29,$70,$30       /// movaps [eax+48],xmm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeAddSquareLoop

@SkipLargeAddSquareLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndAddSquare
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallAddSquareLoop:
  db $0F,$28,$0A           /// movaps xmm1,[edx]
  db $0F,$28,$00           /// movaps xmm0,[eax]
  db $0F,$59,$C9           /// mulps xmm1,xmm1
  db $0F,$58,$C1           /// addps xmm0,xmm1
  db $0F,$29,$00           /// movaps [eax],xmm0
  add eax,16
  add edx,16
  dec ecx
  jnz @SmallAddSquareLoop

@EndAddSquare:
end;

// *************************************************** //
// *                    TX87Vector                   * //
// *************************************************** //

constructor TX87Vector.Create(const ALength : integer);
begin
  inherited Create(ALength);
end;

constructor TX87Vector.CreateUsingArray(var A:array of single);
begin
  inherited CreateUsingArray(A);
end;

destructor TX87Vector.Destroy;
begin
  inherited Destroy;
end;

procedure TX87Vector.Add(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    Data[counter] := Data[counter] + Source[counter];
    Data[counter+1] := Data[counter+1] + Source[counter+1];
    Data[counter+2] := Data[counter+2] + Source[counter+2];
    Data[counter+3] := Data[counter+3] + Source[counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    Data[counter] := Data[counter] + Source[counter];
    inc(Counter);
  end;
end;

procedure TX87Vector.Sub(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    Data[counter] := Data[counter] - Source[counter];
    Data[counter+1] := Data[counter+1] - Source[counter+1];
    Data[counter+2] := Data[counter+2] - Source[counter+2];
    Data[counter+3] := Data[counter+3] - Source[counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    Data[counter] := Data[counter] - Source[counter];
    inc(Counter);
  end;
end;

procedure TX87Vector.Mul(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    Data[counter] := Data[counter] * Source[counter];
    Data[counter+1] := Data[counter+1] * Source[counter+1];
    Data[counter+2] := Data[counter+2] * Source[counter+2];
    Data[counter+3] := Data[counter+3] * Source[counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    Data[counter] := Data[counter] * Source[counter];
    inc(Counter);
  end;
end;

procedure TX87Vector.Divide(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    Data[counter] := Data[counter] / Source[counter];
    inc(Counter);
  end;
end;

procedure TX87Vector.Sqrt(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    Data[counter] := System.Sqrt(Source[counter]);
    inc(Counter);
  end;
end;

procedure TX87Vector.Reciprocal(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    Data[counter] := 1 / Source[counter];
    inc(Counter);
  end;
end;

procedure TX87Vector.RecSqrt(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    Data[counter] := 1 / system.Sqrt(Source[counter]);
    inc(Counter);
  end;
end;

procedure TX87Vector.Max(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    if Source[counter+0] > Data[counter+0] then Data[Counter+0]:= Source[Counter+0];
    if Source[counter+1] > Data[counter+1] then Data[Counter+1]:= Source[Counter+1];
    if Source[counter+1] > Data[counter+1] then Data[Counter+1]:= Source[Counter+1];
    if Source[counter+1] > Data[counter+1] then Data[Counter+1]:= Source[Counter+1];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    if Source[counter] > Data[counter] then
      Data[Counter]:= Source[Counter];
    inc(Counter);
  end;
end;

procedure TX87Vector.Min(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    if Source[counter+0] < Data[counter+0] then Data[Counter+0]:= Source[Counter+0];
    if Source[counter+1] < Data[counter+1] then Data[Counter+1]:= Source[Counter+1];
    if Source[counter+1] < Data[counter+1] then Data[Counter+1]:= Source[Counter+1];
    if Source[counter+1] < Data[counter+1] then Data[Counter+1]:= Source[Counter+1];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    if Source[counter] < Data[counter] then
      Data[Counter]:= Source[Counter];
    inc(Counter);
  end;
end;

procedure TX87Vector.LogicalAnd(V : TFVector);       // JPSS FIXED 7/4/2001
{$IFDEF DELPHI_CODE}
var counter : integer;
    AuxP:pointer;
    DestVec,SrcVec: PLongWordArray;
begin
  AuxP:=AlignedArray; DestVec:=AuxP;
  AuxP:=V.AlignedArray; SrcVec:=AuxP;
  for counter := 0 to (NumElements - 1) do
      DestVec^[counter] := DestVec^[counter] and SrcVec^[counter];
end;
{$ELSE}
asm
   mov ecx,[eax + NumElements]
   mov edx,[V.AlignedArray]
   mov eax,[eax + AlignedArray]
 @AndLoop:
   mov ebx, [edx]
   and [eax], ebx
   add edx,4
   add eax,4
   dec ecx
   jnz @AndLoop
end;
{$ENDIF}

procedure TX87Vector.LogicalOr(V : TFVector);        // JPSS FIXED 7/4/2001
{$IFDEF DELPHI_CODE}
var counter : integer;
    AuxP:pointer;
    DestVec,SrcVec: PLongWordArray;
begin
  AuxP:=AlignedArray; DestVec:=AuxP;
  AuxP:=V.AlignedArray; SrcVec:=AuxP;
  for counter := 0 to (NumElements - 1) do
      DestVec^[counter] := DestVec^[counter] or SrcVec^[counter];
end;
{$ELSE}
asm
   mov ecx,[eax + NumElements]
   mov edx,[V.AlignedArray]
   mov eax,[eax + AlignedArray]
 @AndLoop:
   mov ebx, [edx]
   or [eax], ebx
   add edx,4
   add eax,4
   dec ecx
   jnz @AndLoop
end;
{$ENDIF}

procedure TX87Vector.LogicalXor(V : TFVector);         // JPSS FIXED 7/4/2001
{$IFDEF DELPHI_CODE}
var counter : integer;
    AuxP:pointer;
    DestVec,SrcVec: PLongWordArray;
begin
  AuxP:=AlignedArray; DestVec:=AuxP;
  AuxP:=V.AlignedArray; SrcVec:=AuxP;
  for counter := 0 to (NumElements - 1) do
      DestVec^[counter] := DestVec^[counter] xor SrcVec^[counter];
end;
{$ELSE}
asm
   mov ecx,[eax + NumElements]
   mov edx,[V.AlignedArray]
   mov eax,[eax + AlignedArray]
 @AndLoop:
   mov ebx, [edx]
   xor [eax], ebx
   add edx,4
   add eax,4
   dec ecx
   jnz @AndLoop
end;
{$ENDIF}

procedure TX87Vector.CmpGreater(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if Data[counter] > Source[counter] then
      Pointer(Data[counter]):=pointer($FFFFFFFF)
    else
      Pointer(Data[counter]):=pointer(0);
    inc(Counter);
  end;
end;

procedure TX87Vector.CmpLower(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if Data[counter] < Source[counter] then
      Pointer(Data[counter]):=pointer($FFFFFFFF)
    else
      Pointer(Data[counter]):=pointer(0);
    inc(Counter);
  end;
end;

procedure TX87Vector.CmpEqual(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if Data[counter] = Source[counter] then
      Pointer(Data[counter]):=pointer($FFFFFFFF)
    else
      Pointer(Data[counter]):=pointer(0);
    inc(Counter);
  end;
end;

procedure TX87Vector.CmpGreaterEqual(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if Data[counter] >= Source[counter] then
      Pointer(Data[counter]):=pointer($FFFFFFFF)
    else
      Pointer(Data[counter]):=pointer(0);
    inc(Counter);
  end;
end;

procedure TX87Vector.CmpLowerEqual(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if Data[counter] >= Source[counter] then
      Pointer(Data[counter]):=pointer($FFFFFFFF)
    else
      Pointer(Data[counter]):=pointer(0);
    inc(Counter);
  end;
end;

procedure TX87Vector.AddSquare(V : TFVector);
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    Data[counter+0] := Data[counter+0] + system.sqr(Source[counter+0]);
    Data[counter+1] := Data[counter+1] + system.sqr(Source[counter+1]);
    Data[counter+2] := Data[counter+2] + system.sqr(Source[counter+2]);
    Data[counter+3] := Data[counter+3] + system.sqr(Source[counter+3]);
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    Data[counter] := Data[counter] + system.sqr(Source[counter]);
    inc(Counter);
  end;
end;

function TX87Vector.InnerSum : single;
{$IFDEF DELPHI_CODE}
var
  counter : integer;
  Data:PSingleArray;
begin
  result := 0.0;
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    result := result + Data[counter] + Data[counter+1]+Data[counter+2]+ Data[counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    result := result + Data[counter];
    inc(Counter);
  end;
end;
{$ELSE}
asm
    mov ecx,[eax + NumElements]
    mov eax,[eax + AlignedArray]
    fldz
  @InnerSumLoop:
    fadd DWORD PTR [eax]
    add eax,4
    dec ecx
    jnz @InnerSumLoop
end;
{$ENDIF}

function TX87Vector.InnerSumAbs : single;
{$IFDEF DELPHI_CODE}
var
  counter : integer;
  Data:PSingleArray;
begin
  result := 0.0;
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    result := result + abs(Data[counter]);
    inc(Counter);
  end;
end;
{$ELSE}
asm
    mov ecx,[eax + NumElements]
    mov eax,[eax + AlignedArray]
    fldz
  @InnerSumLoop:
    fld DWORD PTR [eax]
    fabs
    faddp ST(1),ST(0)
    add eax,4
    dec ecx
    jnz @InnerSumLoop
end;
{$ENDIF}

function TX87Vector.DotProduct(V : TFVector) : single;
{$IFDEF DELPHI_CODE}
var
  counter : integer;
  Data,Source:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Source:=@V.AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    result := result + Data[Counter+0]*Source[Counter+0]
                     + Data[Counter+1]*Source[Counter+1]
                     + Data[Counter+2]*Source[Counter+2]
                     + Data[Counter+3]*Source[Counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    result := result + Data[Counter]*Source[Counter];
    inc(Counter);
  end;
end;
{$ELSE}
asm
   mov ecx,[eax + NumElements]
   mov edx,[V.AlignedArray]
   mov eax,[eax + AlignedArray]
   fldz
 @DotProductLoop:
   fld DWORD PTR [eax]
   fmul DWORD PTR [edx]
   add eax,4
   add edx,4
   faddp ST(1),ST(0)
   dec ecx
   jnz @DotProductLoop
end;
{$ENDIF}

function TX87Vector.MaxValue : single;
var
  counter : integer;
  Data:PSingleArray;
begin
  result := AlignedArray[0];
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    if Data[counter+0] > result then result:=Data[counter+0];
    if Data[counter+1] > result then result:=Data[counter+1];
    if Data[counter+2] > result then result:=Data[counter+2];
    if Data[counter+3] > result then result:=Data[counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    if Data[counter] > result then
      result:=Data[counter];
    inc(Counter);
  end;
end;

function TX87Vector.MinValue : single;
var
  counter : integer;
  Data:PSingleArray;
begin
  result := AlignedArray[0];
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    if Data[counter+0] < result then result:=Data[counter+0];
    if Data[counter+1] < result then result:=Data[counter+1];
    if Data[counter+2] < result then result:=Data[counter+2];
    if Data[counter+3] < result then result:=Data[counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    if Data[counter] < result then
      result:=Data[counter];
    inc(Counter);
  end;
end;

function TX87Vector.MaxAbsValue : single;
var
  counter : integer;
  Data:PSingleArray;
begin
  result := AlignedArray[0];
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if abs(Data[counter]) > result then
      result:=abs(Data[counter]);
    inc(Counter);
  end;
end;

function TX87Vector.MinAbsValue : single;
var
  counter : integer;
  Data:PSingleArray;
begin
  result := AlignedArray[0];
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if abs(Data[counter]) < result then
      result:=abs(Data[counter]);
    inc(Counter);
  end;
end;

function TX87Vector.IndexMaxValue : integer;
var
  counter : integer;
  CurrentMaxValue : single;
  Data:PSingleArray;
begin
  result := 0;
  CurrentMaxValue := AlignedArray[0];
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if Data[counter] > CurrentMaxValue then
    begin
      CurrentMaxValue:=Data[counter];
      result:=counter;
    end;
    inc(Counter);
  end;
end;

function TX87Vector.IndexMinValue : integer;
var
  counter : integer;
  CurrentMaxValue : single;
  Data:PSingleArray;
begin
  result := 0;
  CurrentMaxValue := AlignedArray[0];
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<0 do
  begin
    if Data[counter] < CurrentMaxValue then
    begin
      CurrentMaxValue:=Data[counter];
      result:=counter;
    end;
    inc(Counter);
  end;
end;

function TX87Vector.Mean : single;
{$IFDEF DELPHI_CODE}
var
  counter : integer;
  Data:PSingleArray;
begin
  result := 0;
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    result:=result + Data[counter+0] + Data[counter+1] + Data[counter+2] + Data[counter+3];
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    result:=result + Data[counter];
    inc(Counter);
  end;
  result:=result / NumElements;
end;
{$ELSE}
asm
  mov ecx,[eax + NumElements]
  mov edx,[eax + AlignedArray]
  fldz
 @MeanLoop:
  fadd DWORD PTR [edx]
  add edx,4
  dec ecx
  jnz @MeanLoop
  fild DWORD PTR [eax + NumElements]
  fdivp
end;
{$ENDIF}

procedure TX87Vector.Scale(value : single);
{$IFDEF DELPHI_CODE}
var
  counter : integer;
  Data:PSingleArray;
begin
  Data:=@AlignedArray[NumElements];
  Counter:=-NumElements;
  while Counter<-3 do
  begin
    Data[counter+0]:=Data[counter+0] * value;
    Data[counter+1]:=Data[counter+1] * value;
    Data[counter+2]:=Data[counter+2] * value;
    Data[counter+3]:=Data[counter+3] * value;
    inc(Counter,4);
  end;
  while Counter<0 do
  begin
    Data[counter]:=Data[counter] * value;
    inc(Counter);
  end;
end;
{$ELSE}
asm
   mov ecx,[eax + NumElements]
   mov edx,[eax + AlignedArray]
   fld DWORD PTR [ebp + 8]  // load value
 @ScaleLoop:
   fld DWORD PTR [edx]
   fmul ST(0),ST(1)
   fstp DWORD PTR [edx]
   add edx,4
   dec ecx
   jnz @ScaleLoop
end;
{$ENDIF}


// *************************************************** //
// *                  T3DNowVector                   * //
// *************************************************** //

constructor T3DNowVector.Create(const ALength : integer);
begin
  inherited Create(ALength);
end;

constructor T3DNowVector.CreateUsingArray(var A:array of single);
begin
  inherited CreateUsingArray(A);
end;

destructor T3DNowVector.Destroy;
begin
  inherited Destroy;
end;

procedure T3DNowVector.Add(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeAddLoop
@LargeAddLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       ///---------- prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       ///--------- prefetch [edx+64]
  db $0F,$0F,$C1,$9E       /// pfadd mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$9E       /// pfadd mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$9E       /// pfadd mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$9E       /// pfadd mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$9E       /// pfadd mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$9E       /// pfadd mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$9E       /// pfadd mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$9E       /// pfadd mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeAddLoop

@SkipLargeAddLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndAdd
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallAddLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$9E       /// pfadd mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$9E       /// pfadd mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallAddLoop

@EndAdd:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.Sub(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeSubLoop
@LargeSubLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C1,$9A       /// pfsub mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$9A       /// pfsub mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$9A       /// pfsub mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$9A       /// pfsub mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$9A       /// pfsub mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$9A       /// pfsub mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$9A       /// pfsub mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$9A       /// pfsub mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeSubLoop

@SkipLargeSubLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndSub
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallSubLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$9A       /// pfsub mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$9A       /// pfsub mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallSubLoop

@EndSub:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.Mul(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeMulLoop
@LargeMulLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C1,$B4       /// pfmul mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$B4       /// pfmul mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$B4       /// pfmul mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$B4       /// pfmul mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$B4       /// pfmul mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$B4       /// pfmul mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$B4       /// pfmul mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$B4       /// pfmul mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeMulLoop

@SkipLargeMulLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndMul
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallMulLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$B4       /// pfmul mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$B4       /// pfmul mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallMulLoop

@EndMul:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.Divide(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeDivideLoop
@LargeDivideLoop:
  // cmp. AMD Athlon Processor x86 Code Optimization, August 1999, Chapter 10, page 97
  db $0F,$6F,$02           /// movq mm0,[edx]    // load y | x
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0     // approx 1/x | 1/x
  db $0F,$6F,$D0           /// movq mm2,mm0      // y | x
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0 // y | y
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0     // approx 1/y | 1/y
  db $0F,$62,$C8           /// punpckldq mm1,mm0 // approx 1/y | 1/x
  db $0F,$6F,$00           /// movq mm0,[eax]    // z | w
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1  // intermediate 1/y | 1/x
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1  // final 1/y | 1/x
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2     // z/y | w/x
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$42,$08       /// movq mm0,[edx+8]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$08       /// movq mm0,[eax+8]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$08       /// movq [eax+8],mm0

  db $0F,$6F,$42,$10       /// movq mm0,[edx+16]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$10       /// movq mm0,[eax+16]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$10       /// movq [eax+16],mm0

  db $0F,$6F,$42,$18       /// movq mm0,[edx+24]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$18       /// movq mm0,[eax+24]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$18       /// movq [eax+24],mm0

  db $0F,$6F,$42,$20       /// movq mm0,[edx+32]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$42,$28       /// movq mm0,[edx+40]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$28       /// movq mm0,[eax+40]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$28       /// movq [eax+40],mm0

  db $0F,$6F,$42,$30       /// movq mm0,[edx+48]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$30       /// movq mm0,[eax+48]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$30       /// movq [eax+48],mm0

  db $0F,$6F,$42,$38       /// movq mm0,[edx+56]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$38       /// movq mm0,[eax+56]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$38       /// movq [eax+56],mm0

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeDivideLoop

@SkipLargeDivideLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndDivide
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallDivideLoop:
  db $0F,$6F,$02           /// movq mm0,[edx]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$42,$08       /// movq mm0,[edx+8]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$6F,$40,$08       /// movq mm0,[eax+8]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$0F,$C2,$B4       /// pfmul mm0,mm2
  db $0F,$7F,$40,$08       /// movq [eax+8],mm0

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallDivideLoop

@EndDivide:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.Reciprocal(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeReciprocalLoop
@LargeReciprocalLoop:
  // cmp. AMD Athlon Processor x86 Code Optimization, August 1999, Chapter 10, page 97
  db $0F,$6F,$02           /// movq mm0,[edx]    // load y | x
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0     // approx 1/x | 1/x
  db $0F,$6F,$D0           /// movq mm2,mm0      // y | x
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0 // y | y
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0     // approx 1/y | 1/y
  db $0F,$62,$C8           /// punpckldq mm1,mm0 // approx 1/y | 1/x
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1  // intermediate 1/y | 1/x
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1  // final 1/y | 1/x
  db $0F,$7F,$10           /// movq [eax],mm2

  db $0F,$6F,$42,$08       /// movq mm0,[edx+8]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$42,$10       /// movq mm0,[edx+16]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$10       /// movq [eax+16],mm2

  db $0F,$6F,$42,$18       /// movq mm0,[edx+24]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$18       /// movq [eax+24],mm2

  db $0F,$6F,$42,$20       /// movq mm0,[edx+32]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$20       /// movq [eax+32],mm2

  db $0F,$6F,$42,$28       /// movq mm0,[edx+40]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$42,$30       /// movq mm0,[edx+48]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$30       /// movq [eax+48],mm2

  db $0F,$6F,$42,$38       /// movq mm0,[edx+56]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$38       /// movq [eax+56],mm2

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeReciprocalLoop

@SkipLargeReciprocalLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndReciprocal
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallReciprocalLoop:
  db $0F,$6F,$02           /// movq mm0,[edx]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$10           /// movq [eax],mm2

  db $0F,$6F,$42,$08       /// movq mm0,[edx+8]
  db $0F,$0F,$C8,$96       /// pfrcp mm1,mm0
  db $0F,$6F,$D0           /// movq mm2,mm0
  db $0F,$6A,$C0           /// punpckhdq mm0,mm0
  db $0F,$0F,$C0,$96       /// pfrcp mm0,mm0
  db $0F,$62,$C8           /// punpckldq mm1,mm0
  db $0F,$0F,$D1,$A6       /// pfrcpit1 mm2,mm1
  db $0F,$0F,$D1,$B6       /// pfrcpit2 mm2,mm1
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallReciprocalLoop

@EndReciprocal:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.Max(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeMaxLoop
@LargeMaxLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C1,$A4       /// pfmax mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$A4       /// pfmax mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$A4       /// pfmax mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$A4       /// pfmax mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$A4       /// pfmax mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$A4       /// pfmax mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$A4       /// pfmax mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$A4       /// pfmax mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeMaxLoop

@SkipLargeMaxLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndMax
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallMaxLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$A4       /// pfmax mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$A4       /// pfmax mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallMaxLoop

@EndMax:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.Min(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeMinLoop
@LargeMinLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C1,$94       /// pfmin mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$94       /// pfmin mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$94       /// pfmin mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$94       /// pfmin mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$94       /// pfmin mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$94       /// pfmin mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$94       /// pfmin mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$94       /// pfmin mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeMinLoop

@SkipLargeMinLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndMin
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallMinLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$94       /// pfmin mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$94       /// pfmin mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallMinLoop

@EndMin:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.LogicalAnd(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeLogicalAndLoop
@LargeLogicalAndLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$DB,$C1           /// pand mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$DB,$D3           /// pand mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$DB,$E5           /// pand mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$DB,$F7           /// pand mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$DB,$C1           /// pand mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$DB,$D3           /// pand mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$DB,$E5           /// pand mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$DB,$F7           /// pand mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeLogicalAndLoop

@SkipLargeLogicalAndLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndLogicalAnd
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallLogicalAndLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$DB,$C1           /// pand mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$DB,$D3           /// pand mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallLogicalAndLoop

@EndLogicalAnd:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.LogicalOr(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeLogicalOrLoop
@LargeLogicalOrLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$EB,$C1           /// por mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$EB,$D3           /// por mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$EB,$E5           /// por mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$EB,$F7           /// por mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$EB,$C1           /// por mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$EB,$D3           /// por mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$EB,$E5           /// por mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$EB,$F7           /// por mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeLogicalOrLoop

@SkipLargeLogicalOrLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndLogicalOr
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallLogicalOrLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$EB,$C1           /// por mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$EB,$D3           /// por mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallLogicalOrLoop

@EndLogicalOr:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.LogicalXor(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeLogicalXorLoop
@LargeLogicalXorLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$EF,$C1           /// pxor mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$EF,$D3           /// pxor mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$EF,$E5           /// pxor mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$EF,$F7           /// pxor mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$EF,$C1           /// pxor mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$EF,$D3           /// pxor mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$EF,$E5           /// pxor mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$EF,$F7           /// pxor mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeLogicalXorLoop

@SkipLargeLogicalXorLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndLogicalXor
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallLogicalXorLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$EF,$C1           /// pxor mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$EF,$D3           /// pxor mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallLogicalXorLoop

@EndLogicalXor:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.CmpGreater(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpGreaterLoop
@LargeCmpGreaterLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C1,$A0       /// pfcmpgt mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$A0       /// pfcmpgt mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$A0       /// pfcmpgt mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$A0       /// pfcmpgt mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$A0       /// pfcmpgt mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$A0       /// pfcmpgt mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$A0       /// pfcmpgt mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$A0       /// pfcmpgt mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpGreaterLoop

@SkipLargeCmpGreaterLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpGreater
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpGreaterLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$A0       /// pfcmpgt mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$A0       /// pfcmpgt mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpGreaterLoop

@EndCmpGreater:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.CmpLower(V : TFVector);
// A < B decomes B > A
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpLowerLoop
@LargeCmpLowerLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C8,$A0       /// pfcmpgt mm1,mm0
  db $0F,$7F,$08           /// movq [eax],mm1

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$DA,$A0       /// pfcmpgt mm3,mm2
  db $0F,$7F,$58,$08       /// movq [eax+8],mm3

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$EC,$A0       /// pfcmpgt mm5,mm4
  db $0F,$7F,$68,$10       /// movq [eax+16],mm5

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$FE,$A0       /// pfcmpgt mm7,mm6
  db $0F,$7F,$78,$18       /// movq [eax+24],mm7

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C8,$A0       /// pfcmpgt mm1,mm0
  db $0F,$7F,$48,$20       /// movq [eax+32],mm1

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$DA,$A0       /// pfcmpgt mm3,mm2
  db $0F,$7F,$58,$28       /// movq [eax+40],mm3

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$EC,$A0       /// pfcmpgt mm5,mm4
  db $0F,$7F,$68,$30       /// movq [eax+48],mm5

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$FE,$A0       /// pfcmpgt mm7,mm6
  db $0F,$7F,$78,$38       /// movq [eax+56],mm7

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpLowerLoop

@SkipLargeCmpLowerLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpLower
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpLowerLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C8,$A0       /// pfcmpgt mm1,mm0
  db $0F,$7F,$08           /// movq [eax],mm1

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$DA,$A0       /// pfcmpgt mm3,mm2
  db $0F,$7F,$58,$08       /// movq [eax+8],mm3

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpLowerLoop

@EndCmpLower:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.CmpEqual(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpEqualLoop
@LargeCmpEqualLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C1,$B0       /// pfcmpeq mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$B0       /// pfcmpeq mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$B0       /// pfcmpeq mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$B0       /// pfcmpeq mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$B0       /// pfcmpeq mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$B0       /// pfcmpeq mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$B0       /// pfcmpeq mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$B0       /// pfcmpeq mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpEqualLoop

@SkipLargeCmpEqualLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpEqual
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpEqualLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$B0       /// pfcmpeq mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$B0       /// pfcmpeq mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpEqualLoop

@EndCmpEqual:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.CmpGreaterEqual(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpGreaterEqualLoop
@LargeCmpGreaterEqualLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C1,$90       /// pfcmpge mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$90       /// pfcmpge mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$E5,$90       /// pfcmpge mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$F7,$90       /// pfcmpge mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C1,$90       /// pfcmpge mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$D3,$90       /// pfcmpge mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$E5,$90       /// pfcmpge mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$F7,$90       /// pfcmpge mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpGreaterEqualLoop

@SkipLargeCmpGreaterEqualLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpGreaterEqual
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpGreaterEqualLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C1,$90       /// pfcmpge mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$D3,$90       /// pfcmpge mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpGreaterEqualLoop

@EndCmpGreaterEqual:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.CmpLowerEqual(V : TFVector);
// A <= B becomes B >= A
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeCmpLowerEqualLoop
@LargeCmpLowerEqualLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$0F,$C8,$90       /// pfcmpge mm1,mm0
  db $0F,$7F,$08           /// movq [eax],mm1

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$DA,$90       /// pfcmpge mm3,mm2
  db $0F,$7F,$58,$08       /// movq [eax+8],mm3

  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$0F,$EC,$90       /// pfcmpge mm5,mm4
  db $0F,$7F,$68,$10       /// movq [eax+16],mm5

  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$0F,$FE,$90       /// pfcmpge mm7,mm6
  db $0F,$7F,$78,$18       /// movq [eax+24],mm7

  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$0F,$C8,$90       /// pfcmpge mm1,mm0
  db $0F,$7F,$48,$20       /// movq [eax+32],mm1

  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$0F,$DA,$90       /// pfcmpge mm3,mm2
  db $0F,$7F,$58,$28       /// movq [eax+40],mm3

  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$0F,$EC,$90       /// pfcmpge mm5,mm4
  db $0F,$7F,$68,$30       /// movq [eax+48],mm5

  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$0F,$FE,$90       /// pfcmpge mm7,mm6
  db $0F,$7F,$78,$38       /// movq [eax+56],mm7

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeCmpLowerEqualLoop

@SkipLargeCmpLowerEqualLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndCmpLowerEqual
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallCmpLowerEqualLoop:
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0F,$C8,$90       /// pfcmpge mm1,mm0
  db $0F,$7F,$08           /// movq [eax],mm1

  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$0F,$DA,$90       /// pfcmpge mm3,mm2
  db $0F,$7F,$58,$08       /// movq [eax+8],mm3

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallCmpLowerEqualLoop

@EndCmpLowerEqual:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.AddSquare(V : TFVector);
asm
  mov ecx,[eax + NumElements]
  mov eax,[eax + AlignedArray]
  mov edx,[V.AlignedArray]
  push ecx
  shr ecx,4  // number of large iterations = number of elements / 16
  jz @SkipLargeAddSquareLoop
@LargeAddSquareLoop:
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$0D,$42,$40       /// prefetch [edx+64]
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0D,$48,$40       /// prefetchw [eax+64]
  db $0F,$0F,$C9,$B4       /// pfmul mm1,mm1
  db $0F,$0F,$C1,$9E       /// pfadd mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$0F,$DB,$B4       /// pfmul mm3,mm3       DB = 11 011 011
  db $0F,$0F,$D3,$9E       /// pfadd mm2,mm3       D3 = 11 010 011
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  db $0F,$6F,$6A,$10       /// movq mm5,[edx+16]
  db $0F,$6F,$60,$10       /// movq mm4,[eax+16]
  db $0F,$0F,$ED,$B4       /// pfmul mm5,mm5
  db $0F,$0F,$E5,$9E       /// pfadd mm4,mm5
  db $0F,$7F,$60,$10       /// movq [eax+16],mm4

  db $0F,$6F,$7A,$18       /// movq mm7,[edx+24]
  db $0F,$6F,$70,$18       /// movq mm6,[eax+24]
  db $0F,$0F,$FF,$B4       /// pfmul mm7,mm7
  db $0F,$0F,$F7,$9E       /// pfadd mm6,mm7
  db $0F,$7F,$70,$18       /// movq [eax+24],mm6

  db $0F,$6F,$4A,$20       /// movq mm1,[edx+32]
  db $0F,$6F,$40,$20       /// movq mm0,[eax+32]
  db $0F,$0F,$C9,$B4       /// pfmul mm1,mm1
  db $0F,$0F,$C1,$9E       /// pfadd mm0,mm1
  db $0F,$7F,$40,$20       /// movq [eax+32],mm0

  db $0F,$6F,$5A,$28       /// movq mm3,[edx+40]
  db $0F,$6F,$50,$28       /// movq mm2,[eax+40]
  db $0F,$0F,$DB,$B4       /// pfmul mm3,mm3
  db $0F,$0F,$D3,$9E       /// pfadd mm2,mm3
  db $0F,$7F,$50,$28       /// movq [eax+40],mm2

  db $0F,$6F,$6A,$30       /// movq mm5,[edx+48]
  db $0F,$6F,$60,$30       /// movq mm4,[eax+48]
  db $0F,$0F,$ED,$B4       /// pfmul mm5,mm5
  db $0F,$0F,$E5,$9E       /// pfadd mm4,mm5
  db $0F,$7F,$60,$30       /// movq [eax+48],mm4

  db $0F,$6F,$7A,$38       /// movq mm7,[edx+56]
  db $0F,$6F,$70,$38       /// movq mm6,[eax+56]
  db $0F,$0F,$FF,$B4       /// pfmul mm7,mm7
  db $0F,$0F,$F7,$9E       /// pfadd mm6,mm7
  db $0F,$7F,$70,$38       /// movq [eax+56],mm6

  add eax,64
  add edx,64
  dec ecx
  jnz @LargeAddSquareLoop

@SkipLargeAddSquareLoop:
  pop ecx
  and ecx,$0000000F
  jz @EndAddSquare
  shr ecx,2 // number of small iterations = (number of elements modulo 16) / 4
@SmallAddSquareLoop:
  db $0F,$6F,$0A           /// movq mm1,[edx]
  db $0F,$6F,$00           /// movq mm0,[eax]
  db $0F,$0F,$C9,$B4       /// pfmul mm1,mm1
  db $0F,$0F,$C1,$9E       /// pfadd mm0,mm1
  db $0F,$7F,$00           /// movq [eax],mm0

  db $0F,$6F,$5A,$08       /// movq mm3,[edx+8]
  db $0F,$6F,$50,$08       /// movq mm2,[eax+8]
  db $0F,$0F,$DB,$B4       /// pfmul mm3,mm3
  db $0F,$0F,$D3,$9E       /// pfadd mm2,mm3
  db $0F,$7F,$50,$08       /// movq [eax+8],mm2

  add eax,16
  add edx,16
  dec ecx
  jnz @SmallAddSquareLoop

@EndAddSquare:
  db $0F,$77               /// emms
end;

procedure T3DNowVector.Scale(value : single);
asm
   mov ecx,[eax + NumElements]
   mov edx,[eax + AlignedArray]
   shr ecx,1
   db $0F,$6E,$45,$08       /// movd mm0,DWORD PTR [ebp + 8] // load value
   db $0F,$6F,$C8           /// movq mm1,mm0
   db $0F,$73,$F1,$20       /// psllq mm1,32
   db $0F,$EB,$C1           /// por mm0,mm1
 @ScaleLoop:
   db $0F,$6F,$12           /// movq mm2,[edx]
   db $0F,$0F,$D0,$B4       /// pfmul mm2,mm0
   db $0F,$7F,$12           /// movq [edx],mm2
   add edx,8
   dec ecx
   jnz @ScaleLoop
   db $0F,$77               /// emms
end;

// testing functions
procedure SyncVectors(var V1,V2 : TFVector); // modification JPSS 6/4/2001
begin
V1.CopyArray(V2);             // JPSS 6/4/2001
end;
procedure Riteln(lStr: String);
begin
    {$IFDEF console}
    writeln(lStr);
    {$ELSE}
    Showmessage(lStr);
    {$ENDIF}
end;

procedure TestLogicalOps( var TestVector1, TestVector2,TestVector:TFVector); // JPSS 14/3/2001
var    counter : integer;
       AuxP:pointer;
       P1,P2,P:PLongWordArray;
begin
  AuxP:=TestVector1.AlignedArray; P1:=AuxP;
  AuxP:=TestVector2.AlignedArray; P2:=AuxP;
  AuxP:=TestVector.AlignedArray;  P:=AuxP;
  riteln('Testing AndOrXor ops');
  Randomize;
  for counter := 0 to (cTestVectorSize - 1) do
      begin
      TestVector1[counter] := random * 1000;
      TestVector[counter] := TestVector1[counter];
      TestVector2[counter] := random * 1000;
      end;

  // test AND
  for counter := 0 to (cTestVectorSize - 1) do
      P^[counter] := P^[counter] and P2^[counter];

  TestVector1.LogicalAnd(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then riteln('Error: And')
     else riteln('And WORKED!');
  SyncVectors(TestVector, TestVector1);

  // test OR
  for counter := 0 to (cTestVectorSize - 1) do
      P^[counter] := P^[counter] or P2^[counter];

  TestVector1.LogicalOr(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then riteln('Error: Or')
     else riteln('Or WORKED!');
  SyncVectors(TestVector, TestVector1);

  // test XOR
  for counter := 0 to (cTestVectorSize - 1) do
      P^[counter] := P^[counter] xor P2^[counter];

  TestVector1.LogicalXOr(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then riteln('Error: XOr')
     else riteln('XOr WORKED!');
  SyncVectors(TestVector, TestVector1);
end;

procedure TestBinaryBasicOps( var TestVector1, TestVector2,TestVector:TFVector);
var    counter : integer;
begin
  riteln('Testing simple ops');
  Randomize;
  for counter := 0 to (cTestVectorSize - 1) do
      begin
      TestVector1[counter] := random * 1000;
      if TestVector1[counter] = 0
        then TestVector1[counter] := 1.0;
      TestVector[counter] := TestVector1[counter];
      TestVector2[counter] := random * 1000;
      if TestVector2[counter] = 0
        then TestVector2[counter] := 1.0;
      end;

  // test ADD
  for counter := 0 to (cTestVectorSize - 1) do
      TestVector[counter] := TestVector[counter] + TestVector2[counter];
  TestVector1.Add(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then riteln('Error: Add');
  SyncVectors(TestVector, TestVector1);

  // test SUB
  for counter := 0 to (cTestVectorSize - 1) do
      TestVector[counter] := TestVector[counter] - TestVector2[counter];
  TestVector1.Sub(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then riteln('Error: Sub');
  SyncVectors(TestVector, TestVector1);

  // test MUL
  for counter := 0 to (cTestVectorSize - 1) do
      TestVector[counter] := TestVector[counter] * TestVector2[counter];
  TestVector1.Mul(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then riteln('Error: Mul');
  SyncVectors(TestVector, TestVector1);

  // test DIV
  for counter := 0 to (cTestVectorSize - 1) do
      TestVector[counter] := TestVector[counter] / TestVector2[counter];
  TestVector1.Divide(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then riteln('Error: Divide');

end;

procedure TestBinaryComplexOps( var TestVector1, TestVector2,TestVector:TFVector);
var counter : integer;
    res : single;
    otherres : single;
    lStr1,lStr2,lStr3: string;
    procedure ShowError(S:string); // JPSS 6/4/2001
    begin
     riteln(S);
     Str((res-otherres):12:5,lStr1);
     Str((res):12:5,lStr2);
     Str((otherres):12:5,lStr3);
     riteln(' ERROR = '+lStr1+'   Exentia:'+lStr2+' Pascal:'+lStr3);
    end;

begin
  riteln('Testing complex ops');
  Randomize;
  for counter := 0 to (cTestVectorSize - 1) do
      begin
      TestVector1[counter] := (random-0.5) * 1000;   // JPSS modification 6/4/2001
      if TestVector1[counter] = 0
        then TestVector1[counter] := 1.0;
      TestVector[counter] := TestVector1[counter];
      TestVector2[counter] := (random-0.5) * 1000;  // JPSS modification 6/4/2001
      if TestVector2[counter] = 0
        then TestVector2[counter] := 1.0;
      end;

  // test INNER SUM
  res := 0;
  for counter := 0 to (cTestVectorSize - 1) do
      res := res + TestVector1.Data[counter];
  otherres := TestVector1.InnerSum;
  // Note: these results are different (due to rounding error/s?)
  if res <> otherres
     then ShowError('Error: InnerSum:'); // JPSS modification 6/4/2001

  // test INNER ABS SUM
  res := 0;
  for counter := 0 to (cTestVectorSize - 1) do
      res := res + Abs(TestVector1.Data[counter]);
  otherres := TestVector1.InnerSumAbs;
  // Note: these results are different (due to rounding errors?)
  if res <> otherres
     then ShowError('Error: InnerSumAbs:'); // JPSS modification 6/4/2001

  // test ADD + SQUARE
  for counter := 0 to (cTestVectorSize - 1) do
      TestVector[counter] := TestVector[counter] + TestVector2[counter] * TestVector2[counter];
  TestVector1.AddSquare(TestVector2);
  if not(TestVector1.Equal(TestVector))
     then ShowError('Error: Add Square'); // JPSS modification 6/4/2001
  SyncVectors(TestVector, TestVector1);
end;

initialization
  try
    EnableSSE := false;
    asm
      mov eax, 1
      db $0F,$A2               /// cpuid
      test edx,(1 shl 25)
      jnz @SSEFound
      mov EnableSSE,0
      jmp @END_SSE
    @SSEFound:
      mov EnableSSE,1
    @END_SSE:
    end;
  except
    EnableSSE := false;
  end;
  if (EnableSSE)
     then begin
          Enable3DNow := false;
          //NOTE: SSE and 3dNow not necessarily exclusive: this simple removes IDE exceptions from occurring
          end
     else begin
          Enable3DNow := false;
          // check for 3DNow!
          asm
            //test whether extended function 80000001h is supported
            mov eax, 80000000h      //call extended function 80000000h
            db $0F,$A2               /// cpuid                   //reports back highest supported ext. function
            cmp eax, 80000000h      //supports functions > 80000000h?
            jbe @NO_EXTENDED        //no 3DNow! support, either
            //test if function 80000001h indicates 3DNow! support
            mov eax, 80000001h      //call extended function 80000001h
            db $0F,$A2               /// cpuid                   //reports back extended feature flags
            test edx, 80000000h     //bit 31 in extended features
            jz @NO_3DNow            //if set, 3DNow! is supported
            mov Enable3DNow,1
            jmp @END_3DNow
          @NO_EXTENDED:
          @NO_3DNow:
            mov Enable3DNow,0
            jmp @END_3DNow
          @END_3DNow:
          end;
          end;
end.

