/*
classes 
	CProtocolSegment : Definition of an output waveform or expression.
	CProtocol : Collection of segments.
	
	Together these define a 'Protocol' - a description of the output to produce on 
one hardware channel.  These are used in a session to define the generated signal.
*/

class CProtocolSegment : public CItem { 
	private : 
		//int m_iDuration;					// segment duration in msec 
		double m_dV0a, m_dV0b;				// 1st ref value / default start value
		double m_dV1a, m_dV1b;				// 2nd ref value / default end value
		
	public : 
		CString m_cDuration;				// Duration as fixed value or func(i)
		CExpression m_oDuration;	
		
		int GetDuration( double d );
		//void SetDuration( int i );
		bool SetDuration( const char * pchDuration );

		bool m_bV0Range, m_bV1Range;
		double GetV0();
		double GetV0Final();
		void SetV0( double dVa );
		void SetV0Final( double dVb );
		void SetV0( double dVa, bool bRange, double dVb );
		double GetV1();
		double GetV1Final();
		void SetV1( double dVa );
		void SetV1Final( double dVb );
		void SetV1( double dVa, bool bRange, double dVb );

		int m_iDispWidth, m_iIterWidth;		// used by editor
		int m_iRepeatCount;

		int m_iGenStart, m_iGenLength;		// used during generation - within iteration, samples this segment generates

		CString m_cFunction;				// Expression of (i,t,u,v) to generate data i=iteration [0..1]; t=segmenttime[0..1] u=v0; v=v1;
		bool m_bModifiable;					// parameters can be modified in session
		
		CExpression m_oExpr;				// Expression object for generating data in editor or session 
		void Parse();						// Parse expression. If it is empty, use straight line from V0 to V1.
		
		CProtocolSegment( const char * pchDur="500", double dV0a=0.0, double dV1a=10.0, bool bV0=false, bool bV1=false, 
			double dV0b=0.0, double dV1b=0.0, const char * pchFunction = NULL, bool bModifiable=false, int iRepeat=1 );

		void CalcAval1( double * aVal, int iPos, int iLen );
		void CalcAVal23( double * aVal, int iIter, int iIters );
		
		//double Value( double i, double t );
		void Set( const char * pchDur="500", double dV0=0.0, double dV1=10.0, bool bV0Range=false, bool bV1Range=false, 
			double dV0b=0.0, double dV1b=0.0, const char * pchFunction = NULL, bool bModifiable=false, int iRepeat=1 );
		bool LoadFromFile( FILE * h, int iVer );
		bool SaveToFile( FILE * h );
	};
		
class CProtocol { 
	public : 
		int m_iIterations;			// default iterations of Protocol
		bool m_bRandom;
		int * m_aiIterationMap;
		void CreateRandomMap();

		double m_dYMin, m_dYMax;	// min and max of Y Range
		double Range();
		
		CItemArray m_aSeg;
		CProtocolSegment * Segment( int iSeg );

		//-- Generation running params 		
		bool GenPrepare( int iGenRate );
		bool GenSegPrepare(  );
		bool GenerateData( double * pData, int iCount, bool bRepeat, double dScaling, double dMin, double dMax );
		int m_iGenSweep;			// number of full generation sweeps sent
		int m_iGenIteration;		// track iteration during generation
		int m_iGenSegment;			// segment being generated from 
		int m_iGenSample;			// Position of pointer within Iteration
		int m_iGenRate;				// Rate generating at 
		
		CProtocol( );
		~CProtocol( );
		bool AddSegment( const char * pchDur, double dV0a, double dV1a, bool bV0, bool bV1, double dV0b, double dV1b, 
				const char * pchFunction=NULL, bool bModifiable=false, int iRepeat=1, int iPos=-1 );
		bool DelSegment( int iPos );
		bool MoveSegment( int iPos, int iDest );

		int Duration( int iIter=0 );	// msec duration per iteration
		int Duration( double dIter );
		int TotalDurations( );		
		double Iteration(int iIter);// for iteration i, convert to double in range [0.0,1.0]
		double IterationMap(int iIter); // Same as Iteration with random ordering.

		int TotalSamples(int iGenRate);	// total samples for all iterations
		int SegmentFromPos( int iX ) ;
		bool m_bHasModifiable;		// true if any segment is modifiable ( set in load - editor does not require this )
		
		void CalculateDispWidths( int iTtlWid );
		void CalculateIterWidths( int iWidth, double dIter=0.5 );
		
		bool LoadFile( const char * pchFile );
		bool SaveFile( const char * pchFile ) ;
	};

//------------------------------------------------------------
class COptionsGenPulseProtocol {
	public :
		double m_dBaseAmp;
		double m_dAmpMin;
		double m_dAmpMax;
		int m_iTotalDuration;
		int m_iPulseDurMin;
		int m_iPulseDurMax;
		bool m_bPulseExp;
		int m_iSepDurMin;
		int m_iSepDurMax;
		bool m_bSepExp;
		bool m_bRamp;

		COptionsGenPulseProtocol();
		bool RegistryLoad();
		bool RegistrySave();
	};

