const int SESSION_MAXACQCHN = 16;
const int SESSION_MAXGENCHN = 4;

enum ChnType { CT_AI, CT_AO, CT_DI, CT_DO } ;

const double DATA_GLITCH = 10000.0;

const int ANAGENPANEL_HEIGHT = 22;
const int DIGGENPANEL_HEIGHT = 22;
const int DIGACQPANEL_HEIGHT = 32;

const int DIGGENCHANNELID = 1000;		// Call Generate( DIGGENCHANNELID... ) for all generate digitial channels only

const int ERASEWIDTH = 20;				// size of space between current recording and last data being overwritten

const COLORREF crTraceColor = RGB(0,0,255);
const COLORREF crSaturatedColor = RGB(255,192,192);
const COLORREF crPanelSepColor = RGB(160,160,160);

const COLORREF crDigitalOn  = RGB(128,255,128);
const COLORREF crDigitalOff = RGB(255,128,128);
const COLORREF crDigitalTrn= RGB(255,255,128); 

const int NOCAPTURE = -1;		// local mouse capturing is not active
const int CAP_TOP = 1;
const int CAP_LFT = 2;
const int CAP_RGT = 3;
const int CAP_BTM = 4;
const int CAP_MOV = 5;

const int ICONIZE_BOX_SIZE = 10;

const int MEANTEXT_BOXWID = 88;
const int MEANTEXT_BOXHGT = 16;
const int MEANTEXT_WID = 75;
const int MEANTEXT_BASELINE = 12;
const COLORREF MEANTEXT_COLOR = RGB(0x00,0xff,0x00);
const int MEANTEXT_FONTSIZE = 10;

// 
const int TITLE_WIDTH = 14;		// size of title section of panel.
const int UNITS_WIDTH = 14;		// size of units section of panel.
const int YAXIS_WIDTH = 36;		// size of Y axis labels section of panel.
const int YSCROLL_WIDTH = 9;	// size of Y Axis scroll bar section of panel.
const int SCOPE_XPOS = TITLE_WIDTH + UNITS_WIDTH + YAXIS_WIDTH + YSCROLL_WIDTH;

class CDigoutEvent : public CItem { 
	public : 
		int m_iAcqPos;
		int m_iLine;
		int m_iStatus;	// 1=Set On, 0=Set Off, 2=Triggered(auto reset)
		CDigoutEvent( int iAcqPos, int iLine, int iStatus );
	};
	
class CSessionPanel : public CItem { 
	public : 	//-- persistent parameters
		int m_iHgt, m_iWid;		// height of panel, sizeable for analog acq, fixed for others
		int m_iLft, m_iTop;	
		int m_iPage;			// for multipage display, which page this is on.  -1=hidden
		
		double m_dMinDisp;		// minimum displayed value in panel (acquisition only)
		double m_dMaxDisp;		// maximum displayed value in panel (acquisition only)
		double m_dScaling;      // digital scaling factor
		double DispRange();		// Max - Min, at least 0.00001
		int YPos(double dValue);	// position on screen of dValue based on m_dMinDisp and m_dMaxDisp

		ChnType m_iChannelType;	// Type of channel - CT_AI CT_AO CT_DI CT_DO
		int m_iChannelIndex;	// channel number on card ( analog only )

		int m_iMaxLineAcq;			// highest line number used - acquisition
		int m_iMaxLineGen;			// highest line number used - generation
		bool m_abLineSelectAcq[MAXDIGLINES];	// active / inactive - digital acquisition
		bool m_abLineSelectGen[MAXDIGLINES];	// active / inactive - digital generation 
		
		bool m_bIconized;

		bool LoadFromFile( FILE * h );	
		bool SaveToFile( FILE * h );

		HBITMAP m_hBitmap;		// Handle to dc and bitmap to hold scrolling scope display
		HDC m_hdcBitmap;		// device context for drawing on bitmap
		int m_iBitmapOffset;	// offset into bitmap for left edge, wraps around.
		CSessionPanel( int iWid, int iLft, int iTop );
		~CSessionPanel();
		void SetupBitmap( HDC hdc );
		void DeleteBitmap();
		void AIUpdate( int iSampPerPixel, double * adQueue, int iChans, int iChan, int iQueSize, int iQuePos, bool bRecord, HPEN penSep, int iScanMode, int iDrawMode );
		void DIUpdate( int iSampPerPixel, int * aiDigQueue, int iQueSize, int iQuePos, int iLines, BYTE * aiDigAcqMap, int iScanMode, int iDrawMode );
		void AdjustPanel( int iAdjust, int iX, int iY );
		void MovePanel( int iX, int iY, RECT rctPic );
		void PlacePanel( int iLeft, int iTop, int iWid, int iHgt );

		void DrawExtra( HDC hdc, CConfiguration * poCfg );
				
	public : 
		CDlgMiniScroll * pVBar;	// For Ana Acq

		void SetupAnaAcqPanel( const char * pchName, const char * pchUnits, 
					 int iHgt, double dMin, double dMax, double dMinDisp, double dMaxDisp, 
					 bool bVisible, int iChnlIndex, int iPage=0 );
					 
		void SetupAnaGenPanel( const char * pchName, const char * pchUnits, bool bVisible, int iChnlIndex, int iPage=0 );

		void SetupDigAcqPanel( BYTE * abLines, int iPage=0 );
						
		void SetupDigGenPanel( BYTE * abLines, int iPage=0 );

		//----- Following are not persistent.  They are filled in by LoadConfig() 
		// from a combination of CNiCard and CConfiguration::CChannelConfig info
		CString m_cName;		// copy of channel name and units 
		CString m_cUnits;	
		double m_dMin;			// hardware minimum value possible (Analog acquisition)
		double m_dMax;			// hardware maximum value possible (Analog acquisition)

		CDlgControl * m_pUpdate;	// For AO , DO - individual controls
		CDlgControl * m_pPlay;
		CDlgControl * m_pLoop;
		CDlgControl * m_pStop;
	};


	
class CSession { 
	public : 
		CString m_cConfigName;	// Hardware configuration name.
		int m_iPage, m_iPages;	// for multipage displays - current page and page count
		
		//-- analog generation 
		int m_iGenBufferSize;	// size in samples for each channels generate buffer
		int m_iGenRate;			// rate (in hz) of generation
		CString m_acGenFile[SESSION_MAXGENCHN];
		bool m_abGenRepeat[SESSION_MAXGENCHN];
		CProtocol m_aProtocol[SESSION_MAXGENCHN];	// not saved - used during Gen Session
		
		//-- acquisition properties		
		int m_iAcqBufferSize;	// size in samples for each channels acquisition buffer
		int m_iAcqRate;			// rate (in hz) of acquisition
		int m_iAcqSampPerPixel;	// samples per pixel in scope display 
		CString m_cAcqFile;		// data output file name

		//-- Manual Digital generation 
		bool m_bRecordDigout;	// whether to record digital outputs as simulated acquisition channel
		void RecordDigout( int iDigout, int iStatus );	// add record of change to digital output 
		CItemArray m_aDigout;	// Array of manual digital output events 
		CDigoutEvent * DigoutEvent( int iEvent );
		unsigned int m_iDigoutStatus;	// which buttons were made visible
		
		//-- Digital Generation
		CString m_cDigGenFile;
		bool m_bDigGenRepeat;
		bool IsDigoutControl( int iLine );	// Is line named and mode is auto or manual, which indicates button should be shown
		bool IsDigoutControl( );			// are there any digital output controls
		CDigitalProtocol m_oDigProtocol;

		//-- Collection of AI, AO, DI, DO panels 
		CItemArray m_aPanels;	// Array of CSessionPanel objects
		CSessionPanel * Panel( int iPanel );
		
		bool LoadConfig();		// Load hardware configuration and set up panels, etc.
		bool LoadProtocols();	// Open protocol files and count them 
		void AcqOpenFile();		// if not yet opened, open the acq file 
		CConfiguration * m_poCfg; 
		CNICard * m_poCard;
		
		CTask * m_poAnaAcqTask;	// analog acquisition ( or Synchronized IO ) task 
		CTask * m_poAnaGenTask;	// analog generation task if not in Synchronized IO mode 
		CTask * m_poDigAcqTask;	// digital acquisition 
		CTask * m_poDigGenTask; // digital generation 
		int m_iAnaAcqPanels;	// panels to display Analog Acquisitions
		int m_iAnaGenPanels;	// panels to display Analog generations
		int m_iDigAcqLines;	// Lines of Digital Acq to display on single panel
		int m_iDigGenLines;	// Lines of Digital Gen to display on single panel
		BYTE * m_aiDigGenMap;	//  [0]=bit of first line, ..., [m_iDigGenLines-1]=bit of last line
		BYTE * m_aiDigAcqMap;	//  [0]=bit of first line, ..., [m_iDigAcqLines-1]=bit of last line
		
		//-- GainTable : Some amps provide an output channel which provides the current gain setting.
		bool m_bGainTable;		// Apply a gain table setting 
		CString m_cGainTable;	// Amp Gain table name being used
		int m_iGainTableSource;	// acq channel from amp providing index voltage
		int m_iGainTableApply;	// acq channel to apply multiplier to
		CGainTable * m_pGainTable;	// pointer to in use gaintable during acquisition.
		int m_iGainSourceIndex;	// position of gain source channel within acq channels
		int m_iGainApplyIndex;	// position of gain source channel within acq channels
		
		//-- Feedback : Calculate a feedback, such as PID, based on an AI and an AO's protocol to generate the AO output.
		bool m_bFeedback;		// Apply a Feedback loop calculation
		CString m_cFeedback;	// Feedback configuration name being used
		int m_iFeedbackSource;	// Acquisition channel #
		int m_iFeedbackApply;	// Generation channel #
		CFeedback * m_pFeedback;
		bool CreateFeedback();

		// mapping - for a given channel number, which panel index is it 
		int m_aiGenChannelPanel[CONFIG_AOROWS];
		// whether to skip the config dialog
		bool m_bQuick;

		void CreatePanels( int iHgt, int iWid, int iTop, int iLeft );	// For session new - create initial panels	
		void SetPanelPositions( int iTop );
		void SetupPanels( );			// for session load - recreate panels 

		bool NewSession( CQUBIOSettings& oSet );		// Set reasonable default values and clear panels arrays
		bool LoadFile( const char * pchFile, int * piScanMode, bool * pbTimeSep, int iWid );
		bool SaveFile( const char * pchFile, int iScanMode, bool bTimeSep );

		//-- window display 
		HWND m_hView;			// Handle to view window
		void SetupBitmaps(HDC hdc);
		void DrawFgr( HDC hdc, int iScanMode );
		void DrawBgr( HDC hdc, int iPanel );
		
		CSession();
		~CSession();
	
		//----- Active session parameters 		
		bool * m_pbCancel;		// flag to be set when acq should terminate 
		bool m_bRecording;		// true when recording, false when playing
		CQDFFileAcq m_oQDF;		// data output file 
		double * m_adQueue;		// array of last samples read in 
		int * m_aiDigQueue;
		int m_iQueuePos;		// next position to be written
		int m_iQueueSize;		// size of scope queue
		int m_iGenChannel;		// Channel generating when only one is generating ( 1000=Digital )
		NIPOLLER_INIT * m_pAcqInit;		// used during acquisition tasks and acq+gen tasks.
		NIPOLLER_INIT * m_pGenInit;		// used during generation only tasks

		void StartRecording();
		void EndRecording();
		bool Go( int iQueueSize, bool bAcq, bool bGen, bool bRecord );
		bool Synch( int iQueueSize );
		bool Play( int iQueueSize );
		bool Record( int iQueueSize );
		bool Stop();
		bool Generate( int iCh, bool bContinuous );
		bool GenStop( );
		//int m_iGenPos;			// Accumulated position in the Generation buffer
		int m_iAnaGenPos, m_iDigGenPos;
		int m_iAcqPos;			// Accumulated position in the acquisition buffer

		bool GenData( int iSamples, void * pData, void * pDigData, int iSetPos=-1 );
		bool AcqData( int iSamples, void * pAnaData, void * pDigData );

		bool AcqCompleted( );
		bool GenCompleted( );
	};












