/*
Contains : 
basic function to get a file name :
	GetOpenFileName
	GetSaveFileName
	
classes which encapsulate / extend existing controls : 
	CDlgControl
		CDlgEdit
		CDlgCheck
		CDlgCombo
		CDlgScroll
		CDlgList
		CDlgButton
		
classes which create pseudo-controls
	CDlgExtra
		CDlgVertText
		CDlgMiniScroll			
		
class to extend the dialog and manage array of pseudo controls
	CDialog
	CDocumentDialog - for document based dialogue, some generic additions

Note : The following may be required for older versions of windows.
    // Initialize common controls
    INITCOMMONCONTROLSEX iccex;			
    iccex.dwICC = ICC_WIN95_CLASSES;
    iccex.dwSize = sizeof(INITCOMMONCONTROLSEX);
    InitCommonControlsEx(&iccex);
*/

HINSTANCE GetInstance( );
void SetInstance( HINSTANCE hInst );

bool GetOpenFileName( char * pchFile, const char * pchFilter, const char * pchExt, HWND hOwner, char * pchShortName=NULL, CString * pInitialDir = NULL );
bool GetSaveFileName( char * pchFile, const char * pchFilter, const char * pchExt, HWND hOwner, char * pchShortName=NULL, CString * pInitialDir = NULL );

//-----------------------------------------------------------------------------
//---- Dialog Controls - wrapper for child window controls on dialogs 
class CDlgControl : public CItem { 
	protected :
		HWND handle;
	public : 
		HWND m_hTip;
		CDlgControl( HWND h = NULL );
		void SetHandle( HWND h, char * pchTip=NULL );
		HWND GetHandle();
		void SetEnabled( bool bEnable=true );
		void SetVisible( bool bVisible=true );
		void SetText( const char * pch );
		bool GetText( char * pch, int iMaxLen );
		bool GetText( CString * pC );
		int GetTextInt( );
		double GetTextDouble( );
		void GetControlBounds( RECT * prct );
		void GetControlOffsetBounds( RECT * prct );		
		void MoveWindow( int iXOffset, int iYOffset );
		void SizeWindow( int iWidth, int iHeight );
		void SetTextInt( int iVal, char * pchFormat="%d" );
		void SetTextDouble( double dVal, const char * pchFormat=NULL );
		void SetButtonImage( HINSTANCE hInst, int iID, int iDX, int iDY );
		void SetButtonIcon( HINSTANCE hInst, int iID );
		void SetButtonBitmap( HBITMAP hb );
		void CreateTip( char * pchMsg, int iWidth=-1 );
	};
	
//-----------------------------------------------------------------------------
class CDlgButton : public CDlgControl { 
	public : 
		COLORREF m_cr;			// COLOR for owner drawn buttons
	};

//-----------------------------------------------------------------------------
class CDlgEdit : public CDlgControl { 
	public : 
		void SetTextLimit( int iLen );
	};
	
//-----------------------------------------------------------------------------
class CDlgCheck : public CDlgControl { 
	public : 
		void SetCheck( bool bCheck );
		bool GetCheck( );
	};

//-----------------------------------------------------------------------------
class CDlgTab : public CDlgControl { 
	public : 
		void SetTab( int iTab );
		int GetTab( );
		int GetTabCount( );
		void DeleteTabs( );
		void DeleteTab( int iTab );
		bool AddTab( const char * pchName );
	};
	
//-----------------------------------------------------------------------------
class CDlgScroll : public CDlgControl { 
	private : 
		int m_iStepSize;	// size of LINE change	- s.b. negative it TOP > BOTTOM
		int m_iPageSize;	// size of PAGE change  - s.b. negative it TOP > BOTTOM
		int m_iTop;			
		int m_iBottom;
	public : 
		void SetPos( int iPos );
		int GetPos( );
		void SetRange( int iTop, int iBottom, int iStepSize, int iPageSize );
		int ScrollProc( WPARAM wp );
	};

//-----------------------------------------------------------------------------
class CDlgList : public CDlgControl { 
	public : 
		COLORREF m_cr;			// COLOR for owner drawn buttons
		
		void ClearContents();
		void AddString( const char * pch );
		int GetCurSel();			
		int GetCount();				
		int SetCurSel( int iSel );	
		bool ReplaceString( int iSel, const char * pch );
	};

//-----------------------------------------------------------------------------
class CDlgCombo : public CDlgControl { 
	public : 
		void ClearContents();
		void AddString( const char * pch );
		void InsertString(const char * pch, int iPos);
		void DeleteString( int iSel );
		int FindString( const char * pch );
		int FindStringInt( int iVal, const char * pchFormat );
		int FindStringDouble( double dVal, const char * pchFormat );
		int GetCurSel();			// CB_ERR = failed 
		int GetCount();				// CB_ERR = failed 
		int SetCurSel( int iSel );	// CB_ERR = failed 
		int SetCurSel( const char * pch, bool bAdd=false );	// CB_ERR = failed 
		bool GetText( char * pch, int iMaxLen );
		bool GetText( CString * pc );
		bool GetText( int iItem, CString * pc );
	};
	
//-----------------------------------------------------------------------------	
//----- Various types of objects which draw on top of the dialogue or window 
class CDlgExtra : public CItem { 
	public : 
		HWND m_hWndParent;
		CDlgExtra( int iID, HWND hWndParent, int iX0, int iY0, int iX1, int iY1, bool bDraggable );
		RECT m_rct;								// nominal dimensions of item
		int m_iID;

		bool m_bDraggable;						// true if item should respond to drag attempts
		POINT m_ptDragDown;						// starting point of a drag operation
		POINT m_ptDragMove;						// last updated point of a drag operation
		virtual void Draw( HDC hdc );			// cause item to redraw
		virtual bool HitTest( int iX, int iY );	// return true if point is a 'hit' point.
		virtual bool DragDown( int iX, int iY );	// true if point initiates a drag 
		virtual void DragMove( int iX, int iY );	// update coordinates for drag 
		virtual void DragUp( int iX, int iY );	// complete the drag operation
		void Invalidate();
	};

//---- vertical rotated text 
class CDlgVertText : public CDlgExtra { 
	public :	
		COLORREF m_crText;			// Color of text 
		COLORREF m_crFill;			// Color of background
		CString m_cText;
		int m_iSize;
		
		CDlgVertText( int iID, HWND hWndParent, 
			int iX0, int iY0, int iX1, int iY1, 
			const char * pchText, int iSize, 
			COLORREF crText=RGB(0,0,0),
			COLORREF crFill=RGB(255,255,255)
			);
		void Draw( HDC hdc );
	};
	
//---- a vertical scrollbar style range selector.  
// Allows independent movement of the endpoints of a scroll region.  
// Height will be adjusted to be >= 2 tip heights if necc.
class CDlgMiniScroll : public CDlgExtra { 
	public :	
		COLORREF m_crBody;			// color connection between endpoints
		COLORREF m_crFrame;			// color border around bar
		COLORREF m_crTips;			// color endpoints of range
		COLORREF m_crFill;			// color unselected portion of bar
		CDlgMiniScroll( int iID, HWND hWndParent, 
			int iX0, int iY0, int iX1, int iY1, 
			double dMin, double dMax, double dLow, double dHigh, 
			bool bVert=true, 
			COLORREF crFill=RGB(128,128,128),
			COLORREF crTips=RGB(0,0,255),
			COLORREF crBody=RGB(255,255,255),
			COLORREF crFrame=RGB(0,0,0)  );
		void Draw( HDC hdc );
		bool DragDown( int iX, int iY );
		void DragMove( int iX, int iY );
		void DragUp( int iX, int iY );

		double GetLowVal();			// Get or set positions.  If out of range, coerce.  Return final value.
		double SetLowVal(double dVal, bool bPixelSet=true);
		double GetHighVal();		// Get or set positions.  If out of range, coerce.  Return final value.
		double SetHighVal(double dVal, bool bPixelSet=true);
	protected : 
		double m_dMin, m_dMax;		// Values corresponding to ends of scrollbar
		double m_dLow, m_dHigh;		// Values corresponding to user moveable 'tips'

		int m_iLow, m_iHigh;		// Pixel pos corresponding to dLow,dHigh
		int ValueToPixel( double dVal );	// convert value to pixel location on bar
		double PixelToValue( int Val );		// convert pixel location to value 
		
		bool m_bVert;				// Draw vertically ? ( or horizontally ) 
		
		int m_iDragItem;			// Part of bar clicked prior to drag op. : 0:no; 1:top; 2:bottom; 3=body;
		int m_iDragMin,m_iDragMax;	// set at start of drag - limit to drag range.
		
	};

	
//----- 
//Dialog base class - it is intended that this will provide functionality for 
//messages that are sent to the frame window which owns the dialogues.  
class CDialog { 
	public :
		int m_iID;
		HWND m_hDlg;
		HWND m_hTip;
		bool m_bSizing;			// in sizing loop - do not redraw details.
		bool m_bDocumentDialog;	// true if dialog is a CDocumentDialog descendant of CDialog
		
		CDialog();
			
		CItemArray m_aExtra;			// extra items on the dialogue
		CDlgExtra * m_pCapture;			// item which has the mouse captured 
		void DrawExtras( HDC hdc );		// Draw the extras.  Make a virtual ? 
		bool HitTest( int iX, int iY );	// return true if any extra item returns true
		bool DragDown( int iX, int iY );	// true if any control responds
		void DragMove( int iX, int iY );	// pass to control capturing mouse
		void DragUp( int iX, int iY );	// pass to control capturing mouse
		
		CItemArray m_aControls;
		CDlgEdit * CreateEdit( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip=NULL, DWORD dwStyle=0 );
		CDlgControl * CreateLabel( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip=NULL, HFONT fnt=NULL, DWORD dwStyle=0 ); 
		CDlgControl * CreateButton( int iX, int iY, int iWid, int iHgt, int iID, HBITMAP hb, char * pchTip=NULL );
		CDlgCheck * CreateCheckBox( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip=NULL, HFONT fnt=NULL, DWORD dwStyle=0 );
		CDlgScroll * CreateScrollBar( int iX, int iY, int iWid, int iHgt, int iID, char * pchTip=NULL, DWORD dwStyle=SBS_HORZ );
		void CreateTip( int iX0, int iY0, int iX1, int iY1, char * pchMsg );
		CDlgControl * Control( int iPos );
		void DestroyControl( int iID );
		void DestroyExtra( int iID );
				
		virtual void OnKeyUp( DWORD wParam, DWORD lParam ) { }
		virtual void OnKeyDown( DWORD wParam, DWORD lParam ) { }
		virtual void OnChar( DWORD wParam, DWORD lParam ) { }
		virtual bool Close( ) { return true; }	
		virtual bool CanClose( ) { return true; }	
				
		void ShowControls();
		void DestroyControls();
		virtual void OnMinMaxInfo( MINMAXINFO * p ) { } 
		virtual void OnSize( WPARAM wparam, LPARAM * p ) { } 
		virtual void OnExitSizing( LPARAM * p ) {}
		virtual void OnMDIActivate( HWND hwndLosing, HWND hwndGaining ) { } 
		void GetBounds( RECT * prct );

		//-- intended for generic info passing mechanism to be implemented in descendant		
		// Note : It is desirable to allow the Vw classes to be isolated to their 
		// implementation file, and allow minimal access to members via this method, 
		// but if used excessively it defeats this purpose and the classes should be 
		// exposed instead.  
		virtual void * GetData( int iData, void * pInfo ) { return NULL; }

		//virtual void OnEnterSizing( LPARAM * p ) { }		possible future usage
		//virtual void OnExitSizing( LPARAM * p ) { } 
	};

//------------------------------------------------------------------------------------
class CDocumentDialog : public CDialog { 
	public : 
		CString m_cFileName;
		bool m_bUpdated;
		virtual bool Updated() { return m_bUpdated; }
		void SetUpdated();
	
		CDocumentDialog( const char * pchFileFilter, const char * pchFileExtension );

		const char * m_pchFileFilter;
		const char * m_pchFileExtension;
		bool GetSaveAsFileName(  );
		
	};

//------------------------------------------------------------------------------------
LRESULT CALLBACK ChildWndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam );

//------------------------------------------------------------------------------------
// Maintain list of open windows
bool WindowListCanClose( );				// return false on first dialogue to respond false. or true if all are true.
bool WindowListClose( );				// false if 1 or more saves were cancelled (but still closed).
int WindowListFindID( int iID, int iStartPos=0 );
bool WindowListAdd( HWND hFrame, HWND hNew );
bool WindowListRemove( HWND h ); 
bool WindowListRemove( int iH );
bool WindowListProcessMessage( MSG * pmsg );
HWND WindowListHandle( int iPos );


