#include "stdafx.h"
//#include "pdf.h"

const char * pchRptClassName = "RptWin";

const int MAXZOOM=800;
const int MINZOOM=10;

HINSTANCE CRptWin :: m_hInst = (HINSTANCE)0;

//------------------------------------------------------------------------
//-- shorthand 
double R( COLORREF rgb ) { 
	return double(GetRValue(rgb))/256.0;
	}
double G( COLORREF rgb ) { 
	return double(GetGValue(rgb))/256.0;
	}
double B( COLORREF rgb ) { 
	return double(GetBValue(rgb))/256.0;
	}

//------------------------------------------------------------------------
//-- CRptObject 
CRptObject :: CRptObject( int iX0, int iY0, int iX1, int iY1, COLORREF rgb ) { 
	m_iX0=iX0;
	m_iY0=iY0;
	m_iX1=iX1;
	m_iY1=iY1;
	m_rgb=rgb;
	}

void CRptObject :: WritePdf( FILE * hOut ){ }

void CRptObject :: Draw( HDC hdc ){ }

int CRptObject :: PDFSize( ) { 
	return 0;
	}

void CRptObject :: WritePdf_rg( FILE * hOut, COLORREF rgb, bool bStroke ) { 
	fprintf( hOut, "%3.1f %3.1f %3.1f %s ", R(rgb), G(rgb), B(rgb), bStroke ? "RG":"rg" );
	}
	
int CRptObject :: Size_rg( ) { 
	return 15;
	}

//------------------------------------------------------------------------
CRptLine :: CRptLine( int iX0, int iY0, int iX1, int iY1, COLORREF rgb, int iWidth ) 
	: CRptObject( iX0, iY0, iX1, iY1, rgb ) {
	m_rgb=rgb;
	m_iWidth=iWidth;
	m_iType=RPT_LINE;
	}
	
int CRptLine :: PDFSize( ) { 
	return 6 + Size_rg( ) + 27;
	}

void CRptLine :: Draw( HDC hdc ){ 
	HPEN penLine = CreatePen( PS_SOLID, m_iWidth/76, m_rgb );
	HPEN penWas = (HPEN) SelectObject( hdc, penLine );
	MoveToEx( hdc, m_iX0, m_iY0, NULL );
	LineTo( hdc, m_iX1, m_iY1 );
	SelectObject( hdc, penWas );
	DeleteObject( penLine );
	}

void CRptLine :: WritePdf( FILE * hOut ) { 
	fprintf( hOut, "%3.1f w ", m_iWidth<0? 0.5 : double(m_iWidth)/76.0 );
	WritePdf_rg( hOut, m_rgb );
	fprintf( hOut, "%4d %4d m %4d %4d l S \n", m_iX0, m_iY0, m_iX1, m_iY1 );
	}
	
//------------------------------------------------------------------------
CRptBox :: CRptBox( int iX0, int iY0, int iX1, int iY1, bool bStroke, bool bFill, COLORREF rgbEdge, COLORREF rgbFill, int iWidth ) 
	: CRptObject( iX0, iY0, iX1, iY1, rgbEdge )  {
	m_rgb=rgbEdge;
	m_rgbFill=rgbFill;
	m_iType=RPT_BOX;
	m_iWidth=iWidth;	
	m_bFill=bFill;
	m_bStroke=bStroke;	
	}

void CRptBox :: Draw( HDC hdc ){ 
	if( m_bFill ) { 
		RECT rct;
		SetRect( &rct, min(m_iX0,m_iX1), min(m_iY0,m_iY1), max(m_iX0,m_iX1), max(m_iY0,m_iY1) );
		HBRUSH brFill = (HBRUSH) CreateSolidBrush( m_rgbFill );
		FillRect( hdc, &rct, brFill );
		DeleteObject( brFill );
		}
		
	if( m_bStroke ) { 
		HPEN penStroke = CreatePen( PS_SOLID, 0, m_rgb );
		HPEN penWas = (HPEN) SelectObject( hdc, penStroke );
		MoveToEx( hdc, m_iX0, m_iY0, NULL );
		LineTo( hdc, m_iX0, m_iY1 );
		LineTo( hdc, m_iX1, m_iY1 );
		LineTo( hdc, m_iX1, m_iY0 );
		LineTo( hdc, m_iX0, m_iY0 );
		SelectObject( hdc, penWas );
		DeleteObject( penStroke );
		}

	}

void CRptBox :: WritePdf( FILE * hOut ) { 
	fprintf( hOut, "%3.1f w ", m_iWidth<0? 0.5 : double(m_iWidth)/76.0 );
	if( m_bStroke ) 
		WritePdf_rg( hOut, m_rgb );
	if( m_bFill ) 
		WritePdf_rg( hOut, m_rgbFill, false );
	fprintf( hOut, "%4d %4d m %4d %4d l %4d %4d l %4d %4d l %c \n", 
				m_iX0, m_iY0, 
				m_iX0, m_iY1 , 
				m_iX1, m_iY1, 
				m_iX1, m_iY0, 
				m_bFill ? ( m_bStroke ? 'b' : 'f' ) : 's' );
	}

int CRptBox :: PDFSize( ) { 
	return 6 + (m_bFill?Size_rg():0) + (m_bStroke?Size_rg():0) + 51;
	}

//-----------------------------------------------------------------------------
CRptText :: CRptText( int iX0, int iY0, int iSize, const char * pch, COLORREF rgb ) 
	: CRptObject( iX0, iY0, 0, 0, rgb )  {
	m_iSize=iSize;
	if( m_pchText=new char[strlen(pch)+1] ) 
		strcpy( m_pchText, pch );
	else
		m_pchText=0;
	m_iType=RPT_TEXT;
	m_iAngle=0;
	m_bBold=false;
	m_bFixed=false;
	}

CRptText :: ~CRptText() {
	delete [] m_pchText;
	}

void CRptText :: Draw( HDC hdc ){ 
	HFONT fntText = CreateFont(-m_iSize, //MulDiv(m_iSize,-100,76), 
						0 /*Width*/,
						m_iAngle /*escapement*/,
						m_iAngle /*nOrientation*/,
						m_bBold ? FW_BOLD : FW_REGULAR,
						FALSE /* Italic*/, 
						FALSE /* underline */,
						FALSE /* Strikeout */, 
						0,0,0,0, 
						(m_bFixed ? (FF_MODERN | FIXED_PITCH) : (FF_ROMAN | VARIABLE_PITCH)), 
						(m_bFixed ? "Modern" : "Times New Roman"));
	HFONT fntWas = (HFONT) SelectObject( hdc, fntText );
	int iAlignWas = SetTextAlign( hdc, TA_BASELINE | TA_LEFT );
	COLORREF crWas = SetTextColor( hdc, m_rgb );
	TextOut( hdc, m_iX0, m_iY0, m_pchText, strlen( m_pchText ) );
	SetTextColor( hdc, crWas );
	SetTextAlign( hdc, iAlignWas );
	SelectObject( hdc, fntWas );
	DeleteObject( fntText );
	}

void CRptText :: WritePdf( FILE * hOut ) { 
	WritePdf_rg( hOut, m_rgb );
	fprintf( hOut, "BT /F1 %3d Tf %4d %4d Td (%s) Tj ET\n", m_iSize, m_iX0, m_iY0, m_pchText );
	};

int CRptText :: PDFSize( ) { 
	return 6 + Size_rg( ) + 36 + int(strlen(m_pchText));
	}
	
//---------------------------------------------------------------
CRptGroup :: CRptGroup( ) 
	: CRptObject( 0, 0, 600, 800, RGB(0,0,0) ) { 
	}
	
CRptGroup :: CRptGroup( int iX0, int iY0, int iX1, int iY1, COLORREF rgb ) 
	: CRptObject( iX0, iY0, iX1, iY1, rgb )  { 
	}

void CRptGroup :: Draw( HDC hdc ){ 
	for( int i=0; i<m_aObjects.Count(); ++i ) 
		((CRptObject *)m_aObjects[i])->Draw( hdc );
	}

void CRptGroup :: WritePdf( FILE * hOut ) { 
	for( int i=0; i<m_aObjects.Count(); ++i ) 
		((CRptObject *)m_aObjects[i])->WritePdf( hOut );
	}
	
int CRptGroup :: PDFSize( ) { 
	int iSize=0;
	for( int i=0; i<m_aObjects.Count(); ++i ) 
		if( m_aObjects[i]!=NULL ) 
			iSize += ((CRptObject *)m_aObjects[i])->PDFSize();
	return iSize;
	}

void CRptGroup :: AddLine( int iX0, int iY0, int iX1, int iY1, COLORREF rgb, int iWidth ) { 
	m_aObjects.AddItem( new CRptLine( iX0, iY0, iX1, iY1, rgb, iWidth ) ) ;
	}

void CRptGroup :: AddText( int iSize, int iX0, int iY0, const char * pch, COLORREF rgb) { 
	m_aObjects.AddItem( new CRptText( iX0, iY0, iSize, pch, rgb ) );
	}
	
void CRptGroup :: AddBox( int iX0, int iY0, int iX1, int iY1, bool bStroke, bool bFill, COLORREF rgbEdge, COLORREF rgbFill ) { 
	m_aObjects.AddItem( new CRptBox( iX0, iY0, iX1, iY1, bStroke, bFill, rgbEdge, rgbFill ) ) ;
	}

//---------------------------------------------------------------
CRptRGridCol :: CRptRGridCol( int iPos, int iWidth ) {
	m_iPos=iPos;
	m_iWidth=iWidth;
	}

CRptRGrid :: CRptRGrid( int iX0, int iY0, int iY1, int iHeadRows, int iColHeadRows ) 
	: CRptGroup( iX0, iY0, 0, iY1, RGB(0,0,0) )  { 
	m_iHeadRows = iHeadRows;
	m_iColHeadRows = iColHeadRows;
	m_iHeadRowHeight = 10;
	m_iColHeadRowsHeight = 10 ;
	m_iHeadRowFontSize=12;
	m_iColHeadFontSize=10;
	m_iDataFontSize=8;

	m_iWidth=COLOFFSET;
	m_rgbDefault = RGB(0,0,0);
	m_iDataRowHeight = 10;
	}

int CRptRGrid :: DataOffset( int iRow ) { 
	return m_iY0-m_iHeadRows*m_iHeadRowHeight-HEADSEP-m_iColHeadRows*m_iColHeadRowsHeight-iRow*m_iDataRowHeight;
	}

//-- Overall setup 
CRptText * CRptRGrid :: AddHeadText( int iHeadRow, const char * pchText, COLORREF rgb, int iOffset ) { 
	int iY = m_iY0-(iHeadRow+1)*m_iHeadRowHeight;
	CRptText * poRptText = new CRptText( m_iX0+iOffset+HEADOFFSET, iY, m_iColHeadFontSize, pchText, rgb );
	m_aObjects.AddItem( poRptText );
	return poRptText;
	}

void CRptRGrid :: AddBorder() { 
	m_aObjects.AddItem( new CRptBox( m_iX0, m_iY0, m_iX0+m_iWidth+HEADOFFSET, 
		m_iY0-500 /* TODO */, true, false, m_rgbDefault ) );
	
	}

void CRptRGrid :: AddCol( int iWidth ) { 
	m_aColumns.AddItem( new CRptRGridCol( m_iWidth, iWidth ) );
	m_iWidth += iWidth+COLSEP;
	}

CRptText * CRptRGrid :: AddColHeadText( int iCol, int iHeadRow, const char * pchText, COLORREF rgb ) { 
	CRptRGridCol * poCol = (CRptRGridCol *)(m_aColumns[iCol]);
	int iY = m_iY0-m_iHeadRows*m_iHeadRowHeight-HEADSEP-(iHeadRow+1)*m_iColHeadRowsHeight;
	CRptText * poRptText = new CRptText( m_iX0+poCol->m_iPos, iY, m_iColHeadFontSize, pchText, rgb );
	m_aObjects.AddItem( poRptText );
	return poRptText;
	}

//-- Column data contents
//TODO - Right justify data 
CRptText * CRptRGrid :: AddDataItem( int iCol, int iRow, const char * pchText, COLORREF rgb ) { 
	CRptRGridCol * poCol = (CRptRGridCol *)(m_aColumns[iCol]);
	int iY = DataOffset( iRow+1 );
	CRptText * poRptText = new CRptText( m_iX0+poCol->m_iPos, DataOffset( iRow+1 ), m_iDataFontSize, pchText, rgb );
	m_aObjects.AddItem( poRptText );
	return poRptText;
	}

//-- Horizontal sep.  If iCol==-1, should come after all columns are setup
void CRptRGrid :: AddHSep( int iRow, int iCol ) { 
	int iX0, iX1;
	if( iCol<0 || iCol>=m_aColumns.Count() ) {
		iX0=m_iX0+COLOFFSET;
		iX1=m_iX0+m_iWidth;
		}
	else { 
		CRptRGridCol * poCol = (CRptRGridCol *)m_aColumns[iCol];
		iX0=m_iX0+poCol->m_iPos;
		iX1=m_iX0+poCol->m_iPos+poCol->m_iWidth;
		}
	m_aObjects.AddItem( new CRptLine( iX0, DataOffset(iRow)-1, iX1, DataOffset(iRow)-1, m_rgbDefault ) );
	}

//---------------------------------------------------------------
CRptObject * CRptPage :: Object( int iObject ) { 
	return (CRptObject *)m_oGroup.m_aObjects[iObject];
	}

void CRptPage :: WritePdf( FILE * hOut ) { 
	for( int i=0; i<m_oGroup.m_aObjects.Count(); ++i ) 
		if( m_oGroup.m_aObjects[i]!=NULL ) 
			((CRptObject *)m_oGroup.m_aObjects[i])->WritePdf( hOut );
	}

int CRptPage :: PDFSize( ) { 
	int iSize=0;
	for( int i=0; i<m_oGroup.m_aObjects.Count(); ++i ) 
		if( m_oGroup.m_aObjects[i]!=NULL ) 
			iSize += ((CRptObject *)m_oGroup.m_aObjects[i])->PDFSize();
	return iSize;
	}			

void CRptPage :: AddLine( int iX0, int iY0, int iX1, int iY1, COLORREF rgb, int iWidth ) { 
	m_oGroup.m_aObjects.AddItem( new CRptLine( iX0, iY0, iX1, iY1, rgb, iWidth ) ) ;
	}

void CRptPage :: AddText( int iSize, int iX0, int iY0, const char * pch, COLORREF rgb) { 
	m_oGroup.m_aObjects.AddItem( new CRptText( iX0, iY0, iSize, pch, rgb ) );
	}
	
void CRptPage :: AddBox( int iX0, int iY0, int iX1, int iY1, bool bStroke, bool bFill, COLORREF rgbEdge, COLORREF rgbFill ) { 
	m_oGroup.m_aObjects.AddItem( new CRptBox( iX0, iY0, iX1, iY1, bStroke, bFill, rgbEdge, rgbFill ) ) ;
	}

void CRptPage :: Draw( HDC hdc ) { 
	int iObjects = m_oGroup.m_aObjects.Count();
	for( int iObject=0; iObject<iObjects; ++iObject ) 
		Object(iObject)->Draw( hdc );
	}

//-----------------------------------------------------------------------------
CRptPage * CRptWin :: LastPage( ) { 
	int iPages=m_aPages.Count();
	if( iPages==0 ) 
		return NULL;
	return (CRptPage *)m_aPages[iPages-1];
	}

//-- Add a page 
CRptPage * CRptWin :: AddPage( ) { 
	CRptPage * pNew = new CRptPage;
	if( pNew ) 
		m_aPages.AddItem( pNew );
	return pNew;
	}
	
CRptPage * CRptWin :: Page( int iPage ) { 
	return (CRptPage *)m_aPages[iPage];
	}
	
bool CRptWin :: DrawPage( int iPage, HDC hdc ) { 
	if( iPage<0 || iPage>=m_aPages.Count() ) 
		return false;
	Page(iPage)->Draw( hdc );
	return true;
	}
	

bool CRptWin :: WritePdf( const char * pchFile ) { 
	/*
		Object 1 = root.  
		Object 2 = Catalog.  
		Object 3 = ProcSet.  
		Object 4 = Fontset.  
		Object 5 = PageHeader.
		Object 6:6+n-1 = pages
		Object 6+n-1:6+2n-1 = page contents
	*/

	int iPage, iPages=m_aPages.Count();		//TODO - account for NULL pages here and in page indexing
	CRptPage * poPage;

	FILE * hOut = fopen( pchFile, "w" );
	if( hOut==NULL ) 
		return false;
		
	//-- Write Header
	fprintf( hOut, "%%PDF-1.3\n" );	

	int iRootPos=ftell(hOut);
	fprintf( hOut, "1 0 obj << /Type /Catalog /Outlines 2 0 R /Pages 5 0 R >> endobj\n" );

	int iCatalogPos=ftell(hOut);
	fprintf( hOut, "2 0 obj << /Type /Outlines /Count 0 >> endobj\n" );

	int iProcsetPos=ftell(hOut);
	fprintf( hOut, "3 0 obj [/PDF /Text] endobj\n" );

	int iFontsetPos=ftell(hOut);
	fprintf( hOut, "4 0 obj << /Type /Font /Subtype /Type1 /Name /F1 /BaseFont /Helvetica /Encoding /MacRomanEncoding >> endobj \n" );

	//-- Write object 5 : list of pages (Kids) 
	//int iPagesPos=ftell(hOut);				
	fprintf( hOut, "5 0 obj << /Type /Pages /Kids [");	
	for( iPage=0; iPage<iPages; ++iPage ) 
		fprintf( hOut, "%d 0 R ", 6+iPage ) ;
	fprintf( hOut, "] /Count %d /MediaBox [0 0 612 792] >> endobj\n", iPages );
								// todo - experiment with mediabox
								
	//-- last object written
	int iObjs=5;			

	//-- for each page, a page reference object which ID's the page contents object 
	for( iPage=0; iPage<iPages; ++iPage ){	
		((CRptPage *)m_aPages[iPage])->m_iRefPos=ftell(hOut);
		fprintf( hOut, "%d 0 obj << /Type /Page /Parent 5 0 R /Contents %d 0 R /Resources << /ProcSet 3 0 R /Font << /F1 4 0 R >> >> >> endobj \n", 
				iPage+6, 6+iPages+iPage );
		}
		
	//-- for each page - page contents
	for( iPage=0; iPage<iPages; ++iPage ){	
		poPage=(CRptPage *)m_aPages[iPage];
		poPage->m_iContentPos=ftell(hOut);
		fprintf( hOut, "%d 0 obj << /Length %d >> stream\n", 
				iPage+iPages+6, poPage->PDFSize() );
		poPage->WritePdf( hOut );
		fprintf( hOut, "endstream\nendobj\n" );
		}
								
	//-- Write cross-reference table
	int iXRefPos=ftell(hOut);
	fprintf( hOut,	"xref\n"				
					"0 %d\n", 6+2*iPages );	// # of objects, starting at 0
	fprintf( hOut, "0000000000 65535 f \n");	// Linked list of free objects (empty)
	fprintf( hOut, "%010d 00000 n \n", iRootPos);
	fprintf( hOut, "%010d 00000 n \n", iCatalogPos);
	fprintf( hOut, "%010d 00000 n \n", iProcsetPos);
	fprintf( hOut, "%010d 00000 n \n", iFontsetPos);
	for( iPage=0; iPage<iPages; ++iPage )
		fprintf( hOut, "%010d 00000 n \n", ((CRptPage *)m_aPages[iPage])->m_iRefPos );	
	for( iPage=0; iPage<iPages; ++iPage )
		fprintf( hOut, "%010d 00000 n \n", ((CRptPage *)m_aPages[iPage])->m_iContentPos );	
			
	//-- Write Trailer	
	fprintf( hOut, "trailer\n"				
				"<< /Size %d "				// number of objects
				"/Root 1 0 R>>\n"			// index of catalog object = 1 
				"startxref\n"
				"%d\n"						// offset to xref 
				"%%%%EOF", 6+2*iPages, iXRefPos );

	fclose( hOut );
	return true;
	}


//-----------------------------------------------------------------------------
//----- Initialize the report window class 
BOOL CRptWin :: WinInit( HINSTANCE hParent, BOOL lBlack ) { 
	m_hInst=hParent;
	WNDCLASS WindowClass;
	WindowClass.style = CS_HREDRAW | CS_VREDRAW ;	// redraw if sized
	WindowClass.lpfnWndProc = CRptWin::WinProc;			// message proc
	WindowClass.cbClsExtra=0;						// no extra space on object
	WindowClass.cbWndExtra=0;
	WindowClass.hInstance=hParent;
	WindowClass.hIcon=LoadIcon(0,IDI_APPLICATION);
	WindowClass.hCursor=LoadCursor(0,IDC_ARROW);
	WindowClass.hbrBackground=(HBRUSH) GetStockObject(lBlack ? BLACK_BRUSH : WHITE_BRUSH );
	WindowClass.lpszMenuName=NULL;
	WindowClass.lpszClassName=pchRptClassName;
	WindowClass.hIcon=LoadIcon(NULL,IDI_APPLICATION);
	RegisterClass(&WindowClass);	
	return TRUE;
	}

//-----------------------------------------------------------------------------
long WINAPI CRptWin :: WinProc( HWND hWnd, UINT msg, WPARAM wP, LPARAM lP) {
	//---- Get pointer to this report object 
	CRptWin * pRep = (CRptWin *) GetWindowLong( hWnd, GWL_USERDATA );

	switch(msg) { 
		case WM_CLOSE : 
			OutputDebugString(" RPT Close\n");
			break;
			
		//---- Destroy Window 
		case WM_DESTROY : 	//---- Delete this report object, then pass to default handler
			delete pRep;
			OutputDebugString(" RPT Destroy\n");
			break;

		//----- COMMAND (menu choices)
		case WM_COMMAND : 
			switch( LOWORD(wP) ) { 
				//---------- File Menu 
				case IDM_RPT_PRINT : 
					OutputDebugString("RPT Print\n");//TODO
					break;
					
				case IDM_RPT_CLOSE : 
					SendMessage( hWnd, WM_CLOSE, 0, 0L );	//PostQuitMessage(0);
					break;
					
				case IDM_RPT_OPTION+0 : 
				case IDM_RPT_OPTION+1 : 
				case IDM_RPT_OPTION+2 : 
				case IDM_RPT_OPTION+3 : 
				case IDM_RPT_OPTION+4 : 
				case IDM_RPT_OPTION+5 : 
				case IDM_RPT_OPTION+6 : 
				case IDM_RPT_OPTION+7 : 
				case IDM_RPT_OPTION+8 : 
				case IDM_RPT_OPTION+9 : 
				case IDM_RPT_OPTION+10 : 
				case IDM_RPT_OPTION+11 : 
				case IDM_RPT_OPTION+12 : 
				case IDM_RPT_OPTION+13 : 
				case IDM_RPT_OPTION+14 : 
				case IDM_RPT_OPTION+15 : 
				case IDM_RPT_OPTION+16 : 
				case IDM_RPT_OPTION+17 : 
				case IDM_RPT_OPTION+18 : 
				case IDM_RPT_OPTION+19 : { 
					//TODO
					char buf[100];
					sprintf( buf, "RPT Option %d\n", LOWORD(wP)-IDM_RPT_OPTION );
					OutputDebugString( buf );
					}
					break;
					
				//---------- Zoom levels 
				case IDM_RPT_ZOOMNONE : 
					pRep->m_iZoom=100;
					//SetScrollSizes();
					InvalidateRect(hWnd,NULL,TRUE);
					//OutputDebugString("RPT ZOOM 100%\n");
					break;
				case IDM_RPT_ZOOMIN : 
				
					if(pRep->m_iZoom>MINZOOM) {
						pRep->m_iZoom-=(pRep->m_iZoom>100?25:10);
					//	SetScrollSizes();
						InvalidateRect(hWnd,NULL,TRUE);
						}
					//OutputDebugString("RPT ZOOM +\n");
					break;
				case IDM_RPT_ZOOMOUT : 
					if(pRep->m_iZoom<MAXZOOM) {
						pRep->m_iZoom+= (pRep->m_iZoom<100?10:25);
						//SetScrollSizes();
						InvalidateRect(hWnd,NULL,TRUE);
						}
					//OutputDebugString("RPT ZOOM -\n");
					break;
					
				//---------- Page Selection 
				case IDM_RPT_PAGEFIRST : 
					pRep->m_iPage=0;
					InvalidateRect(hWnd,NULL,TRUE);
					break;
				case IDM_RPT_PAGEUP : 
					if( pRep->m_iPage>0 ) {
						--pRep->m_iPage;
						InvalidateRect(hWnd,NULL,TRUE);
						}
					break;
				case IDM_RPT_PAGEDOWN : 
					if( pRep->m_iPage<pRep->m_aPages.Count()-1 ) {
						++pRep->m_iPage;
						InvalidateRect(hWnd,NULL,TRUE);
						}
					break;
				case IDM_RPT_PAGELAST : 
					pRep->m_iPage=max(pRep->m_aPages.Count()-1,0);
					InvalidateRect(hWnd,NULL,TRUE);
					break;

				case IDM_RPT_ABOUT :
					MessageBox( NULL, "Reporting Utility\n\r"
									"Copyright (c) 2001","About...",MB_ICONINFORMATION);
					break;
				default:
					break;
				}
			break;

		//---- character messages 
		case WM_CHAR : 
			switch(wP){
				case '-' :	//---- '-' : ZOOMOUT
					SendMessage(hWnd, WM_COMMAND, IDM_RPT_ZOOMOUT, 0L);
					break;
				case '+' :	//---- '+' : ZOOMIN
					SendMessage(hWnd, WM_COMMAND, IDM_RPT_ZOOMIN, 0L);
					break;
					
				#ifdef _DEBUG	// Output key codes for debugging
				default : {
					char buf[100];
					sprintf( buf, "RPT CHAR %d\n", wP );
					OutputDebugString( buf );
					}
				#endif
				
				}
			return 0;

		//----- Virtual keys 
		case WM_KEYDOWN : 
			switch(wP) { 
				case VK_HOME : 
					SendMessage(hWnd, WM_COMMAND, IDM_RPT_PAGEFIRST, 0L);
					break;
				case VK_END : 
					SendMessage(hWnd, WM_COMMAND, IDM_RPT_PAGELAST, 0L);
					break;
				case VK_PRIOR :	//---- PGUP 
					SendMessage(hWnd, WM_COMMAND, IDM_RPT_PAGEUP, 0L);
					break;
				case VK_NEXT :  //---- PGDN 
					SendMessage(hWnd, WM_COMMAND, IDM_RPT_PAGEDOWN, 0L);
					break;
				
				case VK_LEFT : //----- LEFT <-
					SendMessage(hWnd, WM_HSCROLL, SB_LINEUP, 0L);
					break;
				case VK_RIGHT : //---- RIGHT ->
					SendMessage(hWnd, WM_HSCROLL, SB_LINEDOWN, 0L);
					break;
				case VK_UP : //----- UP
					SendMessage(hWnd, WM_VSCROLL, SB_LINEUP, 0L);
					break;
				case VK_DOWN : //---- DOWN
					SendMessage(hWnd, WM_VSCROLL, SB_LINEDOWN, 0L);
					break;
				
				// Output key codes for debugging
				/*
				default : {
					char buf[100];
					sprintf( buf, "RPT KeyDown %d\n", wP );
					OutputDebugString( buf );
					}
				*/
				}
			return 0;

		//----- Vetical Scrollbar processing
		case WM_VSCROLL : 
			switch (LOWORD(wP)) { 
				case SB_BOTTOM :
					pRep->m_iYOrg = 0;
					break;
				case SB_TOP :
					pRep->m_iYOrg = pRep->m_iYSize;
					break;
				case SB_THUMBPOSITION :
					pRep->m_iYOrg = HIWORD(wP);
					break;
				case SB_THUMBTRACK :
					pRep->m_iYOrg = HIWORD(wP);
					break;
				case SB_PAGEDOWN :
				case SB_LINEDOWN : 
					pRep->m_iYOrg += pRep->m_iYSize/10;
					break;
				case SB_PAGEUP :
				case SB_LINEUP : 
					pRep->m_iYOrg -= pRep->m_iYSize/10;
					break;
				}
				
			pRep->m_iYOrg = __max(0, __min( pRep->m_iYOrg, pRep->m_iYSize ));

			//----- If Vertically scrolled - update view 
			if( pRep->m_iYOrg != GetScrollPos(hWnd, SB_VERT ) ) {
				SetScrollPos( hWnd, SB_VERT, pRep->m_iYOrg, TRUE );
				InvalidateRect( hWnd, NULL, TRUE );
				}
			break;

		//----- Horizontal scrollbar processing
		case WM_HSCROLL : 
			switch (LOWORD(wP)) { 
				case SB_TOP :
					pRep->m_iXOrg = 0;
					break;
				case SB_BOTTOM :
					pRep->m_iXOrg = pRep->m_iXSize-1;
					break;
				case SB_THUMBPOSITION :
					pRep->m_iXOrg = HIWORD(wP);
					break;
				case SB_THUMBTRACK :
					pRep->m_iXOrg = HIWORD(wP);
					break;
				case SB_PAGEUP :
				case SB_LINEUP : 
					pRep->m_iXOrg -= pRep->m_iXSize/10;
					break;
				case SB_PAGEDOWN :
				case SB_LINEDOWN : 
					pRep->m_iXOrg += pRep->m_iXSize/10;
					break;
				}
			pRep->m_iXOrg = __max(0, __min( pRep->m_iXOrg, pRep->m_iXSize ));

			//----- If Horizontally scrolled - update view 
			if( pRep->m_iXOrg != GetScrollPos( hWnd, SB_HORZ ) ) {
				SetScrollPos( hWnd, SB_HORZ, pRep->m_iXOrg, TRUE );
				InvalidateRect( hWnd, NULL, TRUE );
				}
			break;

		case WM_PAINT : { 
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint( hWnd, &ps );

			RECT rCli;
			GetClientRect( hWnd, &rCli );
			SetMapMode( hdc, MM_ISOTROPIC );

			//--- setup drawing
			SetWindowExtEx( hdc, pRep->m_iXSize, pRep->m_iYSize, NULL );
			SetWindowOrgEx( hdc, pRep->m_iXOrg, pRep->m_iYSize-pRep->m_iYOrg, NULL );

			SetViewportExtEx( hdc, rCli.right*100/pRep->m_iZoom, -rCli.bottom*100/pRep->m_iZoom, NULL );
			//SetViewportExtEx( hdc, pRep->m_iXSize*100/pRep->m_iZoom, pRep->m_iYSize*-100/pRep->m_iZoom, NULL );
			SetViewportOrgEx( hdc, 0, 0, NULL);
			
			int iBkModeWas = SetBkMode( hdc, TRANSPARENT );

			//-- Draw contents
			#ifdef _DEBUG
			// alignment grid for reference 
			HPEN penGrid = CreatePen( PS_DOT, 0, RGB(192,192,192) );
			HPEN penWas = (HPEN) SelectObject( hdc, penGrid );
			MoveToEx( hdc, 0, 0, NULL );
			LineTo( hdc, 608, 0 );
			LineTo( hdc, 608, 798 );
			LineTo( hdc, 0, 798 );
			LineTo( hdc, 0, 0 );
			//for( int iX=0; iX<=600; iX+=100 ) { 
			//	MoveToEx( hdc, iX, 0, NULL );
			//	LineTo( hdc, iX, 800 );
			//	}
			//for( int iY=0; iY<=800; iY+=100 ) {
			//	MoveToEx( hdc, 0, iY, NULL );
			//	LineTo( hdc, 600, iY );
			//	}
			SelectObject( hdc, penWas );
			DeleteObject( penGrid );
			#endif
			
			pRep->DrawPage( pRep->m_iPage, hdc );
					
			//-- clean up when done
			SetBkMode( hdc, iBkModeWas );
			EndPaint( hWnd, &ps );
			}
			return 0;

		}

/*
	//---- Zoom implementation


		//----- Resize 
		case WM_SIZE :
			SetSizes();
			// send to toolbar too.
			SendMessage(m_hwndToolBar, msg, wP, lP);
			return 0;



	m_iXMax=50;
	m_iYMax=50;
*/
	//----- all other messages to default window procedure
	return DefWindowProc(hWnd, msg, wP, lP);
	}

//-----------------------------------------------------------------------------
BOOL CRptWin :: WinOpen( const char * pchTitle, bool bPrint, CStringArray * pacOptions ) { 
	if( m_hInst == NULL ) 
		CRptWin::WinInit( GetInstance(), false );
		
	//----- create a new window 
	m_hWnd=CreateWindow( pchRptClassName, pchTitle, 
			WS_OVERLAPPEDWINDOW | WS_VSCROLL | WS_HSCROLL , 
			CW_USEDEFAULT, CW_USEDEFAULT, 
			CW_USEDEFAULT, CW_USEDEFAULT, 
			0, 0, m_hInst, 0);
	SetWindowLong( m_hWnd, GWL_USERDATA, (long) this );

	//----- add the menues
	HMENU hFile = CreateMenu();
	if( bPrint )											//-- Print
		AppendMenu(hFile, MF_STRING, IDM_RPT_PRINT, "Print");
	if( pacOptions!=NULL )									//-- Options { Export PDF , Save as HTML , ... } 
		for( int iOpt=0; iOpt<pacOptions->Count() && iOpt<10; ++iOpt ) 
			AppendMenu( hFile, MF_STRING, IDM_RPT_OPTION+iOpt, pacOptions->GetAt(iOpt).Buffer() );
	AppendMenu(hFile, MF_STRING, IDM_RPT_CLOSE, "e&Xit");	//-- Exit

	HMENU hView = CreateMenu();
	AppendMenu(hView, MF_STRING, IDM_RPT_ZOOMIN, "&+ Zoom in");
	AppendMenu(hView, MF_STRING, IDM_RPT_ZOOMNONE, "Zoom 100%");
	AppendMenu(hView, MF_STRING, IDM_RPT_ZOOMOUT, "&- Zoom out");
	AppendMenu(hView, MF_STRING, IDM_RPT_PAGEFIRST, "First Page");
	AppendMenu(hView, MF_STRING, IDM_RPT_PAGEUP, "Page Up");
	AppendMenu(hView, MF_STRING, IDM_RPT_PAGEDOWN, "Page Down");
	AppendMenu(hView, MF_STRING, IDM_RPT_PAGELAST, "Last Page");

	HMENU hHelp = CreateMenu();
	AppendMenu(hHelp, MF_STRING, IDM_RPT_ABOUT, "&About");

	HMENU hMain = CreateMenu();
	AppendMenu(hMain, MF_POPUP, (UINT) hFile, "&File");
	AppendMenu(hMain, MF_POPUP, (UINT) hView, "&View");
	AppendMenu(hMain, MF_POPUP, (UINT) hHelp, "&Help");

	SetMenu( m_hWnd, hMain );

	//----- Add the toolbar 
	/*
	m_hwndToolBar = CreateToolbarEx(m_hWnd, 
							WS_CHILD|WS_VISIBLE|WS_BORDER|WS_CLIPSIBLINGS,
							1,							// wID Control Identifier 
							8,							// nBitMaps
							AfxGetApp()->m_hInstance,	// App instance -OR- NULL
							IDR_TOOLBAR1,				// Toolbar Res ID -OR- Bitmap handle 
							tbToolButtons,				// Address of tool button array
							10,							// # of buttons
							16, 16, //dxButton dyButton
							16, 16, // dxBitmap, dyBitmap, 
							sizeof(TBBUTTON));
	*/
	/*
	if( lPrint ) {
		SetPageCount();	
		m_lSuppressPrintDialogue=TRUE;
		Print();
		return FALSE;
		}

	m_lSuppressPrintDialogue=FALSE;
	*/

 	//----- Set page format and count 
	/* note : get the print config for page formatting */
	// SetPageCount();	

	m_iPage=0;
	m_iXSize=620;
	m_iXOrg=0;
	m_iYSize=820;
	m_iYOrg=0;
	m_iZoom=100;
	SetScrollRange( m_hWnd, SB_HORZ, 0, m_iXSize, FALSE);
	SetScrollPos( m_hWnd, SB_HORZ, m_iXOrg, TRUE );
	SetScrollRange( m_hWnd, SB_VERT, 0, m_iYSize, FALSE);
	SetScrollPos( m_hWnd, SB_VERT, m_iYOrg, TRUE );

	//----- display the window 
	ShowWindow( m_hWnd, 1 );
	UpdateWindow( m_hWnd );
	//SetSizes();

	return TRUE;
	}

//===============================================================================
void TestPDFRpt() { 
	char buf[100];	
	CRptWin * poRpt = new CRptWin;
	CRptPage * pPage;

	//-- Example creation of table and data 	
	pPage = poRpt->AddPage( );
	CRptRGrid * poGrid = new CRptRGrid( 100, 800, 500, 3, 2);
	poGrid->AddCol( 100 );
	poGrid->AddCol( 222 );
	poGrid->AddCol( 100 );
	poGrid->AddHeadText( 0, "General Heading Line 1" );
	poGrid->AddHeadText( 1, "General Heading Line 2", RGB(255,0,0)  );
	poGrid->AddHeadText( 2, "General Heading Line 3" );
	poGrid->AddColHeadText( 0, 0, "C0H0", RGB(0,0,255) );
	poGrid->AddColHeadText( 1, 0, "C1H0" );
	poGrid->AddColHeadText( 2, 0, "C2H0" );
	poGrid->AddColHeadText( 0, 1, "C0H1" );
	poGrid->AddColHeadText( 1, 1, "C1H1" );
	poGrid->AddColHeadText( 2, 1, "C2H1" );
	poGrid->AddBorder();
	for( int iCol=0; iCol<3; ++iCol ) 
		for( int iRow=0; iRow<10; ++iRow ) {
			sprintf( buf, "row=%d col=%d", iRow, iCol );
			if( iRow==4 || iRow==7 ) 
				poGrid->AddDataItem( iCol, iRow, buf, RGB(128,255,64) );
			else 
				poGrid->AddDataItem( iCol, iRow, buf );
			}
	poGrid->AddHSep( 3, 2 );
	poGrid->AddHSep( 4 );
	pPage->m_oGroup.m_aObjects.AddItem( poGrid );

	//================================
	pPage = poRpt->AddPage();
	for( int iX=0; iX<4; ++iX )
		for( int iY=0; iY<5; ++iY )
			pPage->AddLine( 405, 625, iX*200, iY*200, RGB(iX*25,iY*25,128 ) );

	for( int i=0; i<256; i+=16 ) {
		pPage->AddLine( 300+i*2, 200, 200, 400, RGB(i,0,0), 144 );
		pPage->AddLine( 300+i*2, 900, 600, 300, RGB(MulDiv(rand(),255,RAND_MAX),MulDiv(rand(),255,RAND_MAX),MulDiv(rand(),255,RAND_MAX)) );
		}
	
	
	//================================
	pPage = poRpt->AddPage( );
	int iY;
	for( int i=8; i<20; ++i ) {
		sprintf( buf, "YyJj Text %d", i );
		iY = 10*76-36*(i-8);
		pPage->AddText( i, 200, iY, buf, RGB(i*10,128,128) );
		pPage->AddBox( 200, iY, 400, iY+18, true, false, RGB(255,0,0) );
		}
	
	pPage->AddBox( 20, 10, 30, 20, true, false, RGB(255,0,0), RGB(0,64,0) );
	pPage->AddBox( 40, 10, 50, 20, false, true, RGB(255,0,0), RGB(0,64,0) );
	pPage->AddBox( 60, 10, 70, 20, true, true, RGB(255,0,0), RGB(0,64,0) );
	
	
	//================================
	pPage = poRpt->AddPage();
	for( int iBox=0; iBox<100; ++iBox ) { 
		pPage->AddBox( MulDiv(rand(),360,RAND_MAX), 
						MulDiv(rand(),360,RAND_MAX), 
						MulDiv(rand(),360,RAND_MAX), 
						MulDiv(rand(),360,RAND_MAX), 
						true, 
						false, 
						RGB( MulDiv(rand(),255,RAND_MAX), MulDiv(rand(),255,RAND_MAX), MulDiv(rand(),255,RAND_MAX) ), 
						RGB( 32, 32, 32 ) 
						);
		}
		
	CStringArray ac;
	ac.AddString( "Export PDF" );
	ac.AddString( "Save PDF" );
	poRpt->WinOpen( "Test Report", false, &ac );
	
	//poRpt->WritePdf( "C:\\example1.pdf" ); 
	}





