#include "stdafx.h"

//todo - add a push/pop functionality for m_oCurrent to avoid GoParent() calls in .qdf,

const double SHORT_SCALING = 3200.0;		// default scaling value 

//-----------------------------------------------------------------------------
void CQFNode :: Set( bool bPreload, BYTE btType, int iSize, int iCount, int iDataPos, int iChildPos, int iSiblingPos, const char * pchName ) { 
	*(int *)m_aFlags= bPreload ? 2 : 0 ;
	m_btType=btType;
	m_iSize=iSize;
	m_iCount=iCount;
	m_iDataPos=iDataPos;
	m_iChildPos=iChildPos;
	m_iSiblingPos=iSiblingPos;
	*(double *)m_aReserved=0.0;
	memset(m_aReserved,0,7);
	m_cName.Set(pchName);
	//m_aReserved[7]=min(255,m_cName.Length());
	m_iPos=0;
	}

bool CQFNode :: LoadFromFile( FILE * h, int iPos ){
	char buf[32], bufName[256]={0};
	if( iPos != -1 )
		fseek( h, iPos, SEEK_SET );
	else 
		iPos = ftell( h );
	m_iPos=iPos;
	if( fread( buf, 32, 1, h ) != 1 
		|| (buf[31]>0 && fread( bufName, min(255,buf[31]), 1, h ) != 1) ) 
		return false;
	memcpy( m_aFlags, buf+0, 3 );
	m_btType=buf[3];
	m_iSize=*(int*)(buf+ 4);
	m_iCount=*(int*)(buf+ 8);
	m_iDataPos=*(int*)(buf+12);
	m_iChildPos=*(int*)(buf+16);
	m_iSiblingPos=*(int*)(buf+20);
	memcpy( m_aReserved, buf+25, 7 );
	m_cName.Set( bufName, buf[31] );
	return true;
	}
	
bool CQFNode :: SaveToFile( FILE * h, int iPos, bool bSetDataPos ){
	//-- set m_iPos - where to write 
	if( iPos != -1 )
		fseek( h, m_iPos=iPos, SEEK_SET );
	else 
		m_iPos=ftell(h);

	//-- set data pos 
	if( bSetDataPos ) 
		m_iDataPos=m_iPos+32+m_cName.Length();
	
	return WriteNode( h, -1 ) 
			&& ( m_cName.Length()==0 || fwrite( m_cName.Buffer(), m_cName.Length(), 1, h) == 1);
	}

bool CQFNode :: WriteNode( FILE * h, int iPos ){
	if( iPos != -1 ) 
		fseek( h, iPos, SEEK_SET );

	//-- build write buffer
	BYTE buf[32];
	memcpy( buf+0, m_aFlags, 3 );
	buf[3]=m_btType;
	*(int*)(buf+ 4)=m_iSize;
	*(int*)(buf+ 8)=m_iCount;
	*(int*)(buf+12)=m_iDataPos;
	*(int*)(buf+16)=m_iChildPos;
	*(int*)(buf+20)=m_iSiblingPos;
	memcpy( buf+24, m_aReserved, 7 );
	buf[31]=(BYTE)(m_cName.Length());

	return fwrite( buf, 32, 1, h) == 1 ;
	}

bool CQFNode :: DataBlock( FILE * h, void * pData, int iLen, int iOffset, int * piLen ) { 
	//-- go to data block and read it into provided buffer
	if( fseek( h, m_iDataPos+iOffset, SEEK_SET ) != 0 )
		return false;

	int iRead = fread( (BYTE *)pData, 1, min( iLen, m_iSize*m_iCount-iOffset), h );

	//-- if fewer bytes read than requested, return false and out param the read byte count.
	if( piLen!=NULL ) 
		*piLen=iRead;

	return iRead==iLen;
	}

CString CQFNode :: DataString( FILE * h ) { 
	CString c;
	
	//----- check for string data available
	if( m_btType!=QTP_STRING || m_iCount==0 || m_iDataPos<12 ) 
		return c;
	
	//----- allocate string space; read string data 
	if( c.allocate( m_iCount+1 ) 
		&& fseek( h, m_iDataPos, SEEK_SET) == 0 
		&& fread( c.Buffer(), m_iCount, 1, h) == 1 ) {
		c.Buffer()[m_iCount]=0;
		c.setlength();
		}
	else 
		c.Set("");				
	
	return c;
	}
		
//-----------------------------------------------------------------------------
CQFFile :: CQFFile( ) { 
	m_hFile=NULL;
	m_iDepth=0;
	}

CQFFile :: ~CQFFile() { 
	Close();
	}

bool CQFFile :: Opened() { 
	return m_hFile != NULL;
	} 

bool CQFFile :: Open( const char * pchFileName ) { 
	// Open File 
	BYTE buffer[12];
	if( (m_hFile=fopen(pchFileName,"rb+"))==NULL 
		|| fread( buffer, 12, 1, m_hFile) != 1 
		|| memcmp( buffer, "QUB_(;-)_QFS", 12) != 0 ) 
		return false;
	m_cFile.Set( pchFileName );
	return GoRoot();
	}

int CQFFile :: Create( const char * pchFileName, const char * pchRootNodeName ) { 
	// Create File 
	if( (m_hFile=fopen(pchFileName,"wb+"))==NULL 
		|| fwrite( "QUB_(;-)_QFS", 12, 1, m_hFile ) != 1 ) 
		return -1;
	m_cFile.Set( pchFileName );
	
	// Create Root Node 
	int iPos=12+32+strlen(pchRootNodeName);
	m_oCurrent.Set( true, QTP_EMPTY, 0, 0, 0, 0, 0, pchRootNodeName );
	m_oCurrent.SaveToFile( m_hFile, 12 );
	return iPos;
	}
	
void CQFFile :: Close(){
	if( m_hFile!=NULL )
		fclose(m_hFile);
	m_hFile=NULL;
	}

//----- Current node data load and return 
CString CQFFile :: CurrentString( ) { 
	return m_oCurrent.DataString( m_hFile );
	}

double CQFFile :: CurrentDouble( ) { 
	double d=0.0;
	if( m_oCurrent.m_btType==QTP_DOUBLE ) 
		m_oCurrent.DataBlock( m_hFile, &d, 8, 0 );
	return d;
	}

int CQFFile :: CurrentInt( ) { 
	int i=0;
	if( m_oCurrent.m_btType==QTP_LONG )
		m_oCurrent.DataBlock( m_hFile, &i, 4, 0 );
	else if( m_oCurrent.m_btType==QTP_SHORT ) {
		short ii;			// Note : if this were unsigned, could just read it into the low order bytes.
		m_oCurrent.DataBlock( m_hFile, &ii, 2, 0 );
		i=ii;
		}
	return i;
	}

BYTE * CQFFile :: GetDataBufferToDelete( ) { 
	int iRead, iLen = m_oCurrent.m_iCount*m_oCurrent.m_iSize;
	BYTE * pBuf = new BYTE[iLen];
	if( pBuf!=NULL ) {
		m_oCurrent.DataBlock( m_hFile, pBuf, iLen, 0, &iRead );
		if( iRead != iLen ) {
			delete [] pBuf;
			pBuf=NULL;
			}
		}
	return pBuf;
	}

//----- Navigate file 
bool CQFFile :: GoRoot() { 
	m_iDepth=0;
	return Opened() && m_oCurrent.LoadFromFile( m_hFile, 12 );
	}
	
bool CQFFile :: GoSibling() { 
	return Opened() && m_oCurrent.m_iSiblingPos>=12 && m_oCurrent.LoadFromFile( m_hFile, m_oCurrent.m_iSiblingPos );
	}

bool CQFFile :: GoSibling( const char * pchName ) { 
	while( GoSibling() ) 
		if( m_oCurrent.m_cName==pchName ) 
			return true;
	return false;
	}

bool CQFFile :: GoLastSibling() { 
	if( ! Opened() ) 
		return false;
	while( m_oCurrent.m_iSiblingPos>=12 ) 
		if( ! m_oCurrent.LoadFromFile( m_hFile, m_oCurrent.m_iSiblingPos ) )
			return false;
	return true;
	} 
	
bool CQFFile :: GoChild() { 
	if( ! Opened() || m_oCurrent.m_iChildPos<12 )
		return false;
	if( m_iDepth < MAX_DEPTH ) 
		m_aiParent[m_iDepth++]=m_oCurrent.m_iPos;
	return m_oCurrent.LoadFromFile( m_hFile, m_oCurrent.m_iChildPos);
	}

bool CQFFile :: GoChild( const char * pchName, bool bAdd, BYTE btType, int iSize, int iCount ){
	if( ! Opened() || ( m_oCurrent.m_iChildPos<12 && ! bAdd ) ) 
		return false;
	//-- if child does not exist, add as new child
	if( m_oCurrent.m_iChildPos<12 )
		return AddNode( pchName, true/*child*/, btType, iSize, iCount );
		
	//-- set parent pointer
	if( m_iDepth < MAX_DEPTH ) 
		m_aiParent[m_iDepth++]=m_oCurrent.m_iPos;
	//-- go to first child 
	if( ! GoChild() ) 
		return false;
	//-- check all children for name match 
	while( strcmp(pchName, m_oCurrent.m_cName.Buffer()) != 0 ) { 
		if( ! GoSibling() ) {
			if( bAdd ) 
				return AddNode( pchName, false, btType, iSize, iCount );
			return false;
			}
		}
	return true;
	}

bool CQFFile :: GoParent() { 
	return m_iDepth>0 && m_oCurrent.LoadFromFile( m_hFile, m_aiParent[--m_iDepth]);
	}

bool CQFFile :: DeleteNode() { 
	if( m_iDepth==0 ) 
		return false;

	CQFNode	oDelete = m_oCurrent;
	
	if( ! GoParent() ) 
		return false;
		
	// was first child of parent, update parent
	if( m_oCurrent.m_iChildPos==oDelete.m_iPos ) {	
		m_oCurrent.m_iChildPos=oDelete.m_iSiblingPos;
		m_oCurrent.SaveToFile( m_hFile, m_oCurrent.m_iPos );
		// add recovered space to free space list 
		FreeSpaceAdd( oDelete.m_iPos, 32 );
		if( oDelete.m_iSize*oDelete.m_iCount > 0 ) 
			FreeSpaceAdd( oDelete.m_iDataPos, oDelete.m_iSize*oDelete.m_iCount );
		return true;
		}

	// was link in sibling chain 
	if( ! GoChild() ) 
		return false;
		
	while( m_oCurrent.m_iSiblingPos != oDelete.m_iPos )
		if( ! GoSibling() )
			return false;
	
	m_oCurrent.m_iSiblingPos=oDelete.m_iSiblingPos;
	m_oCurrent.SaveToFile( m_hFile, m_oCurrent.m_iPos );
	// add recovered space to free space list 
	FreeSpaceAdd( oDelete.m_iPos, 32 );
	if( oDelete.m_iSize*oDelete.m_iCount > 0 ) 
		FreeSpaceAdd( oDelete.m_iDataPos, oDelete.m_iSize*oDelete.m_iCount );
	return true;
	} 

//--- add a node as sibling or child.  
// Note : subsequent rewriting of the node should not be inefficient since the file pointer
// is not moved and therefore the write is likely buffered.
//
// Note : if pData is ! null, the DataPos node member is set, though iDataSize could be 0.
// 
bool CQFFile :: AddNode( const char * pchName, bool bChild, BYTE btType, int iSize, int iCount, const void * pData, int iDataSize, bool bPreload ) {
	//-- get address of end of file
	fseek( m_hFile, 0, SEEK_END );
	int iPos=ftell(m_hFile);
	
	//-- update the 'current' node with new child / sibling.  May need to find 'leaf' node.
	if( bChild ) {
		if( m_oCurrent.m_iChildPos>=12 ) { 
			GoChild();
			GoLastSibling();
			m_oCurrent.m_iSiblingPos=iPos;
			}
		else {
			m_aiParent[m_iDepth++]=m_oCurrent.m_iPos;
			m_oCurrent.m_iChildPos=iPos;
			}
		}
	else { 
		GoLastSibling();
		m_oCurrent.m_iSiblingPos=iPos;
		}
	if( ! m_oCurrent.SaveToFile( m_hFile, m_oCurrent.m_iPos ) ) 
		return false;

	//-- add new node 	
	m_oCurrent.Set( bPreload, btType, iSize, iCount, 0, 0, 0, pchName );
	return m_oCurrent.SaveToFile( m_hFile, iPos, pData!=NULL )
		&& ( pData==NULL || iDataSize==0 || fwrite( pData, iDataSize, 1, m_hFile )==1 );
	}
	
bool CQFFile :: AddStringNode( const char * pchName, bool bChild, CString * pcData, bool bIncludeZero ) {
	return AddNode( pchName, bChild, QTP_STRING, 1, pcData->Length()+(bIncludeZero?1:0), pcData->ConstBuffer(), pcData->Length()+(bIncludeZero?1:0) );
	}

bool CQFFile :: AddStringNode( const char * pchName, bool bChild, const char * pchData, bool bIncludeZero ) {
	int iLen = strlen(pchData)+(bIncludeZero?1:0);
	return AddNode( pchName, bChild, QTP_STRING, 1, iLen, pchData, iLen );
	}

bool CQFFile :: ReplaceNodeData( int iPos, int iSize, int iCount, void * pData ) { 
	bool bUpdated=false;					//-- whether node info has changed
	//-- Make sure provided file position is reasonable.
	CQFNode o;
	if( iPos<12 || ! o.LoadFromFile( m_hFile, iPos ) ) 
		return false;
	//--check if data will fit in old space, create new space if not
	if( iSize*iCount > o.m_iCount*o.m_iSize ) { 
		fseek( m_hFile, 0, SEEK_END );		//-- get address of end of file
		o.m_iDataPos=ftell(m_hFile);
		bUpdated=true;
		}
	//-- set data size and count
	if( o.m_iCount!=iCount || o.m_iSize!=iSize ) {
		bUpdated=true;
		o.m_iCount=iCount;
		o.m_iSize=iSize;
		}
	//-- write node info if changed
	if( bUpdated ) 
		o.WriteNode( m_hFile, iPos ); 
	//-- Write New Data block
	if( iSize*iCount>0 ) 
		fwrite( pData, iSize, iCount, m_hFile );
	return true;	
	}

void CQFFile :: FreeSpaceAdd( int iPos, int iLen ) { 
	//TODO 
	}

void CQFFile :: FreeSpaceFind( int iLen ) { 
	//TODO 
	}

void CQFFile :: SaveCurrent() { 
	m_iSaveDepth=m_iDepth;
	memcpy( m_aiSaveParent, m_aiParent, MAX_DEPTH*sizeof(int) );
	m_iSavePos=m_oCurrent.m_iPos;
	}

void CQFFile :: RestoreCurrent() { 
	m_iDepth=m_iSaveDepth;
	memcpy( m_aiParent, m_aiSaveParent, MAX_DEPTH*sizeof(int) );
	m_oCurrent.LoadFromFile( m_hFile, m_iSavePos );
	}


//-----------------------------------------------------------------------------
#ifndef QUBIO
	bool CQDSFFile :: WriteHeader( double dSampling, double dScaling, int iChannels, int iDataSize, int iDataType, bool bSetPos ) { 
		if( bSetPos ) 
			GoRoot();
		return AddNode( "Sampling", true, QTP_DOUBLE, 8, 1, &dSampling, 8 )
			&& AddNode( "Scaling", false, QTP_DOUBLE, 8, 1, &dScaling, 8 )
			&& AddNode( "ADChannelCount", false, QTP_LONG, 4, 1, &iChannels, 4 )
			&& AddNode( "ADDataSize", false, QTP_LONG, 4, 1, &iDataSize, 4 )
			&& AddNode( "ADDataType", false, QTP_LONG, 4, 1, &iDataType, 4 ) ;
		}

	bool CQDSFFile :: WriteComment( const char * pch, bool bSetPos ) { 
		if( bSetPos ) {
			GoRoot();
			GoChild("SessionFile",true);
			}
		if( ! AddStringNode( "Comment", true, "GEM File", true ) ){ 
			//TODO - error message 
			return false;
			}
		GoParent();
		return true;
		}

	bool CQDSFFile :: WriteChannels( CDataConfigure * pConfig ) { 
		//-- Exit if no channels to write
		int iChannels=pConfig->ChannelCount();
		if( iChannels==0 ) 
			return true;

		if( ! AddNode( "DataChannels", true, QTP_EMPTY, 0, 0 ) ) 
			return false;
			
		if( ! AddStringNode( "Config", true, pConfig->ToString().ConstBuffer(), true ) ) 
			return false;
			
		//-- ..DataChannels, ...Channel, ....Name, ....Units, ....Scaling
		for( int iCh=0; iCh<iChannels; ++iCh ) {
			CChannel * pCh = pConfig->Channel(iCh);

			if( ! AddNode( "Channel", false, QTP_EMPTY, 0, 0 )
				|| ! AddStringNode( "Name", true, pCh->m_cName.Buffer(), true )
				|| ! AddStringNode( "Units", false, pCh->m_cUnits.Buffer(), true )
				|| ! AddNode( "Scaling", false, QTP_DOUBLE, 8, 1, &pCh->m_dChScaling, 8 )
				|| ! GoParent()				// Back to Channel
				) {
				//TODO - error message 
				return false;
				}
			}		
			
		GoParent();		// Back to DataChannels
		GoParent();		// Back to original
		return true;
		}

	bool CQDSFFile :: WriteSegments( int iChannels, double * adScaling, CItemArray * paSegments, int * aiSegOffsets, double dFileScale  ) { 
		int iSegments=paSegments->Count();
		if( iSegments==0 || ! GoRoot() ) 
			return false;

		//-- .Segments, ..Segment, ...StartTime, ...Channels
		if( ! AddNode( "Segments", true, QTP_EMPTY, 0, 0 ) ) 
			return false;
		int iPt, iCh, iIndex;
		short * aiData ;
		double dScale;
		
		aiSegOffsets[0]=0;
		for( int iSeg=0; iSeg<iSegments; ++iSeg ) {
			CSegment * pSeg = (CSegment *)paSegments->GetAt(iSeg);
			AddNode( "Segment", iSeg==0, QTP_EMPTY, 0, 0 );
			AddNode( "StartTime", true, QTP_DOUBLE, 8, 1, &pSeg->m_dtStartTime, 8 );

			//-- copy data to buffer
			aiData = new short[pSeg->m_iSamples*pSeg->m_iChannels];
			for( iCh=0; iCh<iChannels; ++iCh ) {
				dScale=adScaling[iCh];

				for( iPt=0; iPt<pSeg->m_iSamples; ++iPt ) {
					iIndex = iPt*iChannels+iCh;
					aiData[iIndex] = short( pSeg->m_pdData[pSeg->m_bInterleaved?iIndex:iPt+iCh*pSeg->m_iSamples] * dScale * dFileScale );
					}
				}
					
			//-- Write buffer to file 
			AddNode( "Channels", false, QTP_SHORT, sizeof(short)*iChannels, pSeg->m_iSamples, 
								aiData, sizeof(short)*iChannels*pSeg->m_iSamples, true/*preload*/ );
			delete [] aiData;
								
			//-- track total offsets
			aiSegOffsets[iSeg+1]=(aiSegOffsets[iSeg]+=pSeg->m_iSamples);	// okay to go 1 too far, allocated for n+1 elements
			GoParent();		// Back to Segment
			}

		GoParent();		// Back to Segments
		GoParent();		// Back to SessionFile
		return true;
		}

	bool CQDSFFile :: WriteLists( CItemArray * paLists, int * aiSegOffsets ) {
		//-- ..Lists, ...List, ....(Item Label), .....Comment
		int iLists=paLists->Count();
		if( iLists==0 ) 
			return true;
		if( ! AddNode( "Lists", true, QTP_EMPTY, 0, 0 ) ) 
			return false;
			
		int aiListData[2];
		for( int iList=0; iList<iLists; ++iList ) {
			CSelList * pList = (CSelList *)paLists->GetAt(iList);
			AddStringNode( "List", iList==0, &pList->m_cName );
			int iItems=pList->m_aItems.Count();
			for( int iItem=0; iItem<iItems; ++iItem ) { 
				CSelListItem * pItem = (CSelListItem *)(pList->m_aItems[iItem]);
				aiListData[0]=aiSegOffsets[pItem->m_iStartSegment]+pItem->m_iStartPoint;
				aiListData[1]=aiSegOffsets[pItem->m_iEndSegment]+pItem->m_iEndPoint;
				AddNode( pItem->m_cLabel.Buffer(), iItem==0, QTP_LONG, 8, 1, aiListData, 8 );
				AddStringNode( "Comment", true, &pItem->m_cNote, true );
				GoParent();	// back to label
				}
			if( iItems>0 ) 
				GoParent();	// back to "list"
			}
		if( iLists>0 ) 
			GoParent();	// back to "lists"
		GoParent();	// back to starting point ("SessionFile").
		return true;
		}


	bool CQDSFFile :: WriteIDL( CItemArray * paChannels, CItemArray * paSegments, int * aiSegOffsets, double dSampling ) {
		//-- find largest required 
		int i, iIdl, iSeg, iFirst;
		int iStates=0, iDwells=0;

		for( iSeg=0; iSeg<paSegments->Count(); ++iSeg ) { 
			CSegment * pSeg = (CSegment *)paSegments->GetAt(iSeg);
			for( iIdl=0; iIdl<pSeg->m_aIdl.Count(); ++iIdl ) { 
				CIdlList * pList = pSeg->IdlList(iIdl);
				iStates = max( iStates, pList->m_aIdlClasses.Count() );
				iDwells = max( iDwells, pList->m_aIdlDwells.Count() );
				}
			}

		double * adAmp = new double[iStates];
		double * adSD = new double[iStates];

		// TODO : "Firsts" and "Durations" are redundant and only appear because they 
		// did in Qub.  Since the format is no longer compatible with Qub, remove them.
		// check that loader does not use them first.
		long * aiFirst = new long[iDwells];
		long * aiLast = new long[iDwells];
		short * aiClass = new short[iDwells];
		short * aiState = new short[iDwells];
		float * afDuration = new float[iDwells];

		//-- Save new Idls array 
		bool bFirst=true;
		AddNode( "Idls", true, QTP_EMPTY, 0, 0 );

		for( iSeg=0; iSeg<paSegments->Count(); ++iSeg ) { 
			CSegment * pSeg = (CSegment *)paSegments->GetAt(iSeg);
			for( iIdl=0; iIdl<pSeg->m_aIdl.Count(); ++iIdl ) { 
				CIdlList * pList = pSeg->IdlList(iIdl);
				AddNode( "Idl", bFirst, QTP_EMPTY, 0, 0 );
				AddNode( "Segment", true, QTP_LONG, 4, 1, &iSeg, 4, true );
				AddNode( "Offset", false, QTP_LONG, 4, 1, &pList->m_iOffset, 4, true );
				AddNode( "Length", false, QTP_LONG, 4, 1, &pList->m_iLength, 4, true );
				AddNode( "RGB", false, QTP_LONG, 4, 1, &pList->m_crColor, 4, true );
				AddNode( "Channel", false, QTP_LONG, 4, 1, &pList->m_iChannel, 4, true );
				
				iStates = pList->m_aIdlClasses.Count();
				for( i=0; i<iStates; ++i ) { 
					adAmp[i] = pList->IdlClass(i)->m_dAmp;
					adSD[i] = pList->IdlClass(i)->m_dStdDev;
					}
				AddNode( "Amps", false, QTP_DOUBLE, 8, iStates, adAmp, 8*iStates, true );
				AddNode( "StdDevs", false, QTP_DOUBLE, 8, iStates, adSD, 8*iStates, true );

				iDwells = pList->m_aIdlDwells.Count();
				iFirst=pList->m_iOffset;
				for( i=0; i<iDwells; ++i ) { 
					CIdlDwell * pDwell = pList->IdlDwell(i);
					aiFirst[i] = iFirst;
					aiLast[i] = iFirst+pDwell->m_iDuration-1;
					aiClass[i] = pDwell->m_iClass;
					aiState[i] = pDwell->m_iState;
					afDuration[i] = float(pDwell->m_iDuration)*float(dSampling);
					iFirst += pDwell->m_iDuration;
					}

				AddNode( "Firsts", false, QTP_LONG, 4, iDwells, aiFirst, 4*iDwells, true );
				AddNode( "Lasts", false, QTP_LONG, 4, iDwells, aiLast, 4*iDwells, true );
				AddNode( "Classes", false, QTP_SHORT, 2, iDwells, aiClass, 2*iDwells, true );
				AddNode( "Durations", false, QTP_FLOAT, 4, iDwells, afDuration, 4*iDwells, true );
				AddNode( "States", false, QTP_SHORT, 2, iDwells, aiState, 2*iDwells, true );

				GoParent();
				bFirst=false;
				}
			}
		if( ! bFirst ) 
			GoParent();				// back to "Idls"
		GoParent();					// back to sessionfile

		//-- cleanup 
		delete [] aiFirst;
		delete [] aiLast;
		delete [] aiClass;
		delete [] aiState;
		delete [] afDuration;

		return true;
		}

#else
	//-----------------------------------------------------------------------------
	CQDFFileAcq :: CQDFFileAcq( int iType ) { 
		SetDataType(iType);
		m_iBufferSize=0;
		m_pBuffer=NULL;
		}

	CQDFFileAcq :: ~CQDFFileAcq( ) {
		AllocateBuffer(0);
		}

	bool CQDFFileAcq :: AddChannel( const char * pchName, const char * pchUnits, double dScaling, double dGain ) { 
		GoRoot();
		GoChild( "SessionFile" );
		GoChild( "DataChannels" );
		AddNode( "Channel", true, QTP_EMPTY, 0, 0 );
		AddNode( "Name", true, QTP_STRING, 1, strlen(pchName), (void *)pchName, strlen(pchName) );

		int iLen=strlen(pchUnits);
		char * pch = new char[iLen+2];
		sprintf( pch, pchUnits );
		pch[iLen]=' ';
		pch[iLen+1]=0;
		AddNode( "Units", false, QTP_STRING, 1, iLen+1, (void *)pch, iLen+1 );
		delete [] pch;
		// add a blank space ( bug in QuB ? ) 
		// AddNode( "Units", false, QTP_STRING, 1, strlen(pchUnits), (void *)pchUnits, strlen(pchUnits) );
		
		AddNode( "Scaling", false, QTP_DOUBLE, 8, 1, &dScaling, 8 );
		AddNode( "Gain", false, QTP_DOUBLE, 8, 1, &dGain, 8 );
		fflush(m_hFile);
		return true;
		}

	bool CQDFFileAcq :: CreateQdf( const char * pchFileName, double dSampling, int iChannels ){ 
		Create( pchFileName, "DataFile" );
		double dScaling=SHORT_SCALING;
		AddNode( "Sampling", true, QTP_DOUBLE, 8, 1, &dSampling, 8);
		AddNode( "Scaling", false, QTP_DOUBLE, 8, 1, &dScaling, 8);
		AddNode( "ADChannelCount", false, QTP_LONG, 4, 1, &iChannels, 4);
		AddNode( "ADDataSize", false, QTP_LONG, 4, 1, &m_iDataSize, 4);
		AddNode( "ADDataType", false, QTP_LONG, 4, 1, &m_iDataType, 4);
		AddNode( "Segments", false, QTP_EMPTY, 0, 0);
		m_bCurrentSaved=false;
		fflush( m_hFile );	
		return true;
		}	

	bool CQDFFileAcq :: AddDataNode( int iChannels ){ 
		BYTE junk;					//Note: ptr with 0 bytes passed to AddNode to get the data pos set in the node.
		if( m_bCurrentSaved ) {
			RestoreCurrent();
			AddNode( "Segment", false, QTP_EMPTY, 0, 0 );
			}
		else { 
			GoRoot();
			GoChild("Segments");
			AddNode( "Segment", true, QTP_EMPTY, 0, 0 );
			SaveCurrent();
			m_bCurrentSaved=true;
			}

		double dTime = Now();
		AddNode( "StartTime", true, QTP_DOUBLE, 8, 1, &dTime, 8 );
		AddNode( "Channels", false, m_iDataType, m_iDataSize*iChannels, 0, &junk, 0, false );
			
		m_iSamplesWritten=0;
		m_iDataPos=m_oCurrent.m_iDataPos;
		m_iSamplesPos=m_oCurrent.m_iPos+COUNT_OFFSET;

		fflush( m_hFile );	
		return true;
		}

	bool CQDFFileAcq :: AddDigoutNode( int iSamples, unsigned int * pData ){ 
		// assume the current node is 'Channels' from last segment save
		AddNode( "Digout", false, QTP_LONG, sizeof(int), iSamples, pData, sizeof(int)*iSamples, false );
		fflush( m_hFile );	
		return true;
		}

	bool CQDFFileAcq :: AllocateBuffer( int iSize ) { 
		if( iSize < m_iBufferSize ) 
			return true;
		if( m_pBuffer!=NULL ) 
			delete[]m_pBuffer;
			
		m_pBuffer=NULL;
		if( iSize == 0 )
			return true;
			
		m_pBuffer=new BYTE[iSize];
		if( m_pBuffer==NULL ) 
			return false;
			
		m_iBufferSize=iSize;
		return true;
		}

	bool CQDFFileAcq :: AddData( double * pData, int iAnaChannels, int * aiData, int iDigChannels, BYTE * aiDigMap,
						int iSamples, bool lUpdateNode ) { 
		int iChannels = iAnaChannels+iDigChannels;
		if( iSamples>0 ) {		//might be 0 for UpdateNode=true calls.
			int iWritten;
			fseek( m_hFile, m_iDataPos+m_iSamplesWritten*m_iDataSize*iChannels, SEEK_SET );
			m_iSamplesWritten += iSamples;
			
			//----- Convert double buffer to short ints
			if( m_iDataType==QTP_SHORT ) {			
				if( ! AllocateBuffer( sizeof(short)*iChannels*iSamples ) ) {
					AddMessage( REPTYPE_ERROR, "Error adding data to qdf file ");
					return false;
					}

				short * pSampleShort;
				double * pSampleDouble;
				for( int iSample=0; iSample<iSamples; ++iSample ) { 
					pSampleShort = ((short *)m_pBuffer)+iSample*iChannels;
					pSampleDouble = pData + iSample*iAnaChannels;
					for( int iAnaCh=0; iAnaCh<iAnaChannels; ++iAnaCh )
						pSampleShort[iAnaCh] = short( SHORT_SCALING * pSampleDouble[iAnaCh] );
					pSampleShort = ((short *)m_pBuffer)+iSample*iChannels+iAnaChannels;
					for( int iDigCh=0; iDigCh<iDigChannels; ++iDigCh )
						pSampleShort[iDigCh] = (aiData[iSample]&(1<<aiDigMap[iDigCh]))?short(SHORT_SCALING):0;
					}

				iWritten = fwrite( m_pBuffer, iChannels*m_iDataSize, iSamples, m_hFile ); 
				}
				
			//----- convert 2 double buffers to 1 
			else if( m_iDataType==QTP_DOUBLE && iDigChannels>0 ) { 
				if( ! AllocateBuffer( sizeof(float)*iChannels*iSamples ) ) {
					AddMessage( REPTYPE_ERROR, "Error adding data to qdf file ");
					return false;
					}
					
				double * pSampleDest;
				double * pSampleSource;
				for( int iSample=0; iSample<iSamples; ++iSample ) { 
					pSampleDest = ((double *)m_pBuffer)+iSample*iChannels;
					pSampleSource = pData + iSample*iAnaChannels;
					for( int iAnaCh=0; iAnaCh<iAnaChannels; ++iAnaCh )
						pSampleDest[iAnaCh] = pSampleSource[iAnaCh];
					pSampleDest = ((double *)m_pBuffer)+iSample*iChannels+iAnaChannels;
					for( int iDigCh=0; iDigCh<iDigChannels; ++iDigCh )
						pSampleDest[iDigCh] = (aiData[iSample]&(1<<aiDigMap[iDigCh]))?1:0;
					}

				iWritten = fwrite( ((double *)m_pBuffer), iChannels*m_iDataSize, iSamples, m_hFile ); 
				}
				
			//----- buffer is compatible, copy to file
			else if( m_iDataType==QTP_DOUBLE ) { 
				iWritten = fwrite( pData, iChannels*sizeof(double), iSamples, m_hFile ); 
				}
				
			//----- convert double buffers to float
			else { 
				if( ! AllocateBuffer( sizeof(float)*iChannels*iSamples ) ) {
					AddMessage( REPTYPE_ERROR, "Error adding data to qdf file ");
					return false;
					}

				float * pSampleFloat;
				double * pSampleDouble;
				for( int iSample=0; iSample<iSamples; ++iSample ) { 
					pSampleFloat = ((float *)m_pBuffer)+iSample*iChannels;
					pSampleDouble = pData + iSample*iAnaChannels;
					for( int iAnaCh=0; iAnaCh<iAnaChannels; ++iAnaCh )
						pSampleFloat[iAnaCh] = float( pSampleDouble[iAnaCh] );
					pSampleFloat = ((float *)m_pBuffer)+iSample*iChannels+iAnaChannels;
					for( int iDigCh=0; iDigCh<iDigChannels; ++iDigCh )
						pSampleFloat[iDigCh] = (aiData[iSample]&(1<<aiDigMap[iDigCh]))?1.0f:0.0f;
					}

				iWritten = fwrite( m_pBuffer, iChannels*m_iDataSize, iSamples, m_hFile ); 
				}

			//-- Success ? 
			if( iWritten != iSamples ) {
				AddMessage( REPTYPE_ERROR, "Error 1 adding data to qdf file." );
				return false;
				}
			}
			
		if( lUpdateNode ) { 
			fseek( m_hFile, m_iSamplesPos, SEEK_SET );
			fwrite( &m_iSamplesWritten, 4, 1, m_hFile );
			fflush( m_hFile );
			m_oCurrent.m_iCount=m_iSamplesWritten;
			}
			
		return true;
		}
		
	bool CQDFFileAcq :: AddNote( int iPos, const char * pchNote, const char * pchComment ) { 
		return m_aDataNotes.AddItem( new CQFDataNote( iPos, pchNote, pchComment ) );
		}

	void CQDFFileAcq :: SetDataType( int iDataType ){ 
		m_iDataType = iDataType;
		m_iDataSize = ( m_iDataType==QTP_SHORT ? 2 : ( m_iDataType==QTP_DOUBLE ? 8 : 4 ) ) ;
		}

	//-----------------------------------------------------------------------------
	bool CQDFFileAcq :: FindOrAddList( const char * pchListName ) { 
		//----- Search for list of specified name or create it 
		if( ! GoRoot() 
			|| ! GoChild( "SessionFile", true, QTP_EMPTY, 0, 0 )
			|| ! GoChild( "Lists", true, QTP_EMPTY, 0, 0 ) ) 
			return false;
		
		//-- if no child node, create as first child 
		if( m_oCurrent.m_iChildPos<12 )
			return AddNode( "List", true, QTP_STRING, 1, strlen(pchListName), pchListName, strlen(pchListName) );

		//-- go to first list 
		if( ! GoChild() ) 
			return false;

		//-- search through siblings for list 
		CString cListName;
		do { 
			//-- If list was found, exit 
			if( strcmp(m_oCurrent.m_cName.Buffer(), "List" )==0 
				&& strcmp( (cListName=m_oCurrent.DataString(m_hFile)).ConstBuffer(), pchListName) == 0 )
				return true;
			
			// no more siblings to check - add a new list 
			if( m_oCurrent.m_iSiblingPos==0 )	
				return AddNode( "List", false, QTP_STRING, 1, strlen(pchListName), pchListName, strlen(pchListName) );

			// keep looking 
			} while( GoSibling() ); 
		return false;
		}	

	bool CQDFFileAcq :: WriteDataNotes( ) { 
		if( m_aDataNotes.Count() == 0 ) 
			return true;	// true ... no notes written successfully 
					
		FindOrAddList( "Datanotes" );
			
		//-- create the Sels data block - array of start and stop positions
		CQFDataNote * pNote;
		//	The 'Sels' array of all start/stops created by qub seems to be not required.

		//-- go to add position
		bool bChild=true;
		if( m_oCurrent.m_iChildPos >= 12 ) {
			bChild=false;
			GoChild();
			while( GoSibling() ) { 
				} 
			}

		int aiSel[2];	
		for( int iNote=0; iNote<m_aDataNotes.Count(); ++iNote ) { 
			pNote=(CQFDataNote *)m_aDataNotes[iNote];
			aiSel[0]=pNote->m_iStart;
			aiSel[1]=pNote->m_iEnd;
			if( ! AddNode( pNote->m_cNote.Buffer(), bChild, QTP_LONG, 8, 1, aiSel, 8 )
				|| ! AddNode( "Comment", true, QTP_STRING, 1, pNote->m_cComment.Length(), pNote->m_cComment.Buffer(), pNote->m_cComment.Length() )
				|| ! GoParent() ) 
				return false;
			bChild=false;
			}		

		m_aDataNotes.DeleteAll();
		return true;
		}
		
	bool CQDFFileAcq :: WriteComment( const char * pchText ) { 
		if( GoRoot()
			&& GoChild( "SessionFile" ) ) 
			return AddNode( "Comment", true, QTP_STRING, 1, strlen(pchText), pchText, strlen(pchText));
		return false;
		}

	CQFDataNote :: CQFDataNote( int iPos, const char * pchNote, const char * pchComment ) { 
		m_iStart=m_iEnd=iPos;
		m_cNote.Set( pchNote );
		m_cComment.Set( pchComment );
		}
#endif
