unit FFT;

interface


//******************************************************************************
implementation

procedure TPig.prNPower(var n, lnn, nn :Integer);
begin
  case n of
      1..2:
        begin
          lnn := 1;
          nn  := 2;
        end;
      3..4:
        begin
          lnn := 2;
          nn  := 4;
        end;
      5..8:
        begin
          lnn := 3;
          nn  := 8;
        end;
      9..16:
        begin
          lnn := 4;
          nn  := 16;
        end;
     17..32:
        begin
          lnn := 5;
          nn  := 32;
        end;
     33..64:
        begin
          lnn := 6;
          nn  := 64;
        end;
     65..128:
        begin
          lnn := 7;
          nn  := 128;
        end;
    129..256:
        begin
          lnn := 8;
          nn  := 256;
        end;
    257..512:
        begin
          lnn := 9;
          nn  := 512;
        end;
    513..1024:
        begin
          lnn := 10;
          nn  := 1024;
        end;
   1025..2048:
        begin
          lnn := 11;
          nn  := 2048;
        end;
     else
        begin
          lnn := 12;
          nn  := 4096;
        end;
  end;
end;

//************************   F I L T E R    ************************************
procedure TPig.prDoFilter;
var
  i, j, k, l          :Integer;
  n, lnn, nn          :Integer;
  f                   :Integer;
  ns, ne              :Array[1..2] of Integer;
  bs, be              :Integer;
  r1, r2, i1, i2      :Double;
begin
  if (prFilter = False) and (Notch[1] = False) and (Notch[2] = False) and (BPass = False) then Exit;

  f := 0;

  n := prFrames.Count;//VarArrayHighBound(prData.Series, 3);

  prNPower(n, lnn, nn);
  //nn = the next power of two
  //lnn = the exponent

  //Low Pass
  if prFilter then begin
    f := round(prFilterFc * prTimeFrame * 2 * (2{to deal with edge effect}*nn - 1));// + 1;
    if f > (2*nn div 2) then f := 2*nn div 2;
  end;

  //Notch
  for i := 1 to 2 do begin
    if Notch[i] then begin
      ns[i] := round(NotchStart[i] * prTimeFrame * 2 * (2*nn - 1));// + 1;
      ne[i] := round(NotchEnd[i] * prTimeFrame * 2 * (2*nn - 1));// + 1;
      if ns[i] > (2*nn div 2) then ns[i] := 2*nn div 2;
      if ne[i] > (2*nn div 2) then ne[i] := 2*nn div 2;
      if ns[i] < 1 then ns[i] := 1;
      if ne[i] < 1 then ne[i] := 1;
    end;
  end;

  //Band Pass
  bs := 1;
  be := 1;
  if BPass then begin
    bs := round(BPassStart * prTimeFrame * 2 * (2*nn - 1));// + 1;
    be := round(BPassEnd * prTimeFrame * 2 * (2*nn - 1));// + 1;
    if bs > (2*nn div 2) then bs := 2*nn div 2;
    if be > (2*nn div 2) then be := 2*nn div 2;
    if bs < 1 then bs := 1;
    if be < 1 then be := 1;
  end;

  //Add TAvg and DAvg

  for i := 0 to prData.SeriaCount - 1 {VarArrayHighBound(prData.Series, 1)} do begin
    for j := 1 to 2 do begin

      //Deconstruct data
      for k := (nn div 2) + 1 to (nn div 2) + n do begin
        prReal[k] := prData.Series[i, j, k - (nn div 2)];
        prImag[k] := 0;
      end;
      for k := (nn div 2) + n + 1 to (nn div 2) + nn do begin
        prReal[k] := prReal[(nn div 2) + n - 1 - (k - ((nn div 2) + n + 1))];
        prImag[k] := 0;
      end;

      for k := 1 to (nn div 2) do begin
        prReal[k] := prReal[(nn div 2) + 1 + ((nn div 2) - k)];
        prImag[k] := 0;
      end;
      for k := (nn div 2) + nn + 1 to 2*nn do begin
        prReal[k] := prReal[(nn div 2) + nn - 1 + (k - ((nn div 2) + nn + 1))];
        prImag[k] := 0;
      end;

      //Direct FFT
      prFFT(lnn + 1, -1);

      r1 := prReal[1];
      r2 := prReal[2*nn];
      i1 := prImag[1];
      i2 := prImag[2*nn];

      //Low pass filter
      if prFilter then begin
        for k := 1 + f to 2*nn - f do begin
          prReal[k] := 0;
          prImag[k] := 0;
        end;
      end;

      //Notch filter                         to be checked !!!
      for l := 1 to 2 do begin
        if Notch[l] then begin
          for k := 1 + ns[l] to 1 + ne[l] do begin
            prReal[k] := 0;
            prImag[k] := 0;
          end;
          for k := 2*nn - ne[l] to 2*nn - ns[l] do begin
            prReal[k] := 0;
            prImag[k] := 0;
          end;
        end;
      end;

      //Band-pass filter
      if BPass then begin
        for k := 1 to bs - 1 do begin
          prReal[k] := 0;
          prImag[k] := 0;
        end;
        for k := be + 1 to 2*nn - be do begin
          prReal[k] := 0;
          prImag[k] := 0;
        end;
        for k := 2*nn - bs + 2 to 2*nn do begin
          prReal[k] := 0;
          prImag[k] := 0;
        end;
      end;

      prReal[1]    := r1;
      prReal[2*nn] := r2;
      prImag[1]    := i1;
      prImag[2*nn] := i2;

      //Indirect FFT
      prFFT(lnn + 1, 1);

      //Reconstruct data
      for k := 1 to n do begin
        prData.Series[i, j, k] := prReal[k + (nn div 2)];
      end;

      if Assigned(prOnProgress) then begin
        prOnProgress(Self, ((i + 1) * 100) div prData.SeriaCount);
      end;

    end;
  end;

  if Assigned(prOnProgress) then begin
    prOnProgress(Self, 0);
  end;

end;

procedure TPig.prFFT(lnN :Integer; sign :Integer);
var
  divN, s, ur, ur1,  ui, wr, wi, tr, ti :Single;
  n, nd2, i, j, k, l, le, le1, ip       :Integer;
label
  1, 2, 3;
begin
  n := prPower2(lnN);
  divN := 1 / sqrt(n*1.0);
  nd2 := n div 2;
  j := 1;
  for i := 1 to n - 1 do begin
    if i >= j then goto 1;
    s := prReal[i]; prReal[i] := prReal[j]; prReal[j] := s;
    s := prImag[i]; prImag[i] := prImag[j]; prImag[j] := s;
    1: k := nd2;
    2: if k >= j then goto 3;
    j := j - k;
    k := k div 2;
    goto 2;
    3: j := j + k;
  end;
  for l := 1 to lnN do begin
    le := prPower2(l);
    le1 := le div 2;
    ur := 1;
    ui := 0;
    wr := cos(pi/le1);
    wi := sign*sin(pi/le1);
    for j:=1 to le1 do begin
      i := j;
      while (i <= N) do begin
        ip := i + le1;
        tr := prReal[ip]*ur - prImag[ip]*ui;
        ti := prReal[ip]*ui + prImag[ip]*ur;
        prReal[ip] := prReal[i] - tr;
        prImag[ip] := prImag[i] - ti;
        prReal[i] := prReal[i] + tr;
        prImag[i] := prImag[i] + ti;
        i := i + le;
      end;
      ur1 := ur*wr - ui*wi;
      ui := ur*wi + ui*wr;
      ur := ur1;
    end;
  end;
  for i := 1 to N do begin
    prReal[i] := prReal[i]*divN;
    prImag[i] := prImag[i]*divN;
  end;
end;

function TPig.prPower2(i :Integer) :Integer;
var
  n :Integer;
begin
  n := 1;
  for i := 1 to i do begin
    n := n*2;
  end;
  Result := n;
end;


//******************************************************************************
end.



