unit SelectorListBox;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

// by Chris c.2003
// The basic TListBox makes it inconvenient to just look at
// ItemChange (which line is selected), ItemDoubleClick,
// and select what you right-clicked prior to popupmenu.
// Not that this one's perfect either, but it's a start.
// Based on the stuff in TQUB_QVO_QVD_Global.


interface


uses


Windows, Classes, Menus, Stdctrls, Messages, Controls;


type


TSelectorListBox = class(TCustomListBox)
protected


  prLastItemIndex             :Integer;


  procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
  procedure WndProc(var Message: TMessage); override;

  procedure prOnKeyPress(Sender: TObject; var Key: Char);


public


  OnItemChange                :TNotifyEvent;
  OnItemDblClick              :TNotifyEvent;


  constructor Create(AOwner :TComponent); override;


published


  PopupMenu                   :TPopupMenu;


end;


implementation


//******************************************************************************
constructor TSelectorListBox.Create(AOwner :TComponent);
begin
  inherited;
  MultiSelect := False;
  OnKeyPress  := prOnKeyPress;
  prLastItemIndex := -1;
end;

procedure TSelectorListBox.CNCommand(var Message: TWMCommand);
var
  PrevItemIndex       :Integer;
  FirstSelected       :Integer;
  i                   :Integer;
begin
  case Message.NotifyCode of
    LBN_SELCHANGE:
      begin
        PrevItemIndex := ItemIndex;

        inherited;

        FirstSelected := -1;
        for i := 0 to Items.Count - 1 do begin
          if Selected[i] then begin
            FirstSelected := i;
            Break;
          end;
        end;

        // the chosen list is the first of multi-selection, or the most recently selected
        if FirstSelected = -1 then begin
          if PrevItemIndex <> -1 then begin
            ItemIndex := PrevItemIndex;
            Selected[ ItemIndex ] := True;
          end;
        end else {if (ItemIndex = -1) or not Selected[ ItemIndex ] then} begin
          ItemIndex := FirstSelected;
        end;

        //Generate OnItemChange Event
        if ItemIndex <> prLastItemIndex then begin
          prLastItemIndex := ItemIndex;
          if Assigned(OnItemChange) then begin
            OnItemChange(Self);
          end;
        end;
      end;
    LBN_DBLCLK:
      begin
        inherited;
        //Generate OnItemDblClick Event
        if Assigned(OnItemDblClick) then begin
          OnItemDblClick(Self);
        end;
      end;
  else
    inherited;
  end;
end;

procedure TSelectorListBox.WndProc(var Message: TMessage);
var
  Point   :TPoint;
begin
  {make r-click := l-click then popup}
  if Message.Msg = WM_RBUTTONDOWN then begin
    Message.Msg := WM_LBUTTONDOWN;
    inherited WndProc(Message);
    Message.Msg := WM_LBUTTONUP;
    inherited WndProc(Message);
    Message.Msg := WM_RBUTTONDOWN;

    if PopupMenu <> nil then begin
      Point.X := Message.lParamLo;
      Point.Y := Message.lParamHi;
      Point   := ClientToScreen( Point );
      PopupMenu.Popup(Point.X,Point.Y);
    end;
  end else begin
    inherited WndProc(Message);
  end;
end;

procedure TSelectorListBox.prOnKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;

  if Key = #13 then begin
    Key := #0;
    OnItemDblClick(Self);
  end;
end;


// *****************************************************************************
end.
