unit qubopt_iface;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


QUB_Node,
AllocTypes;


//Solvopt
function solvopt(optimizer :TObject; var stopflag :Boolean; n :Word; Pars :Pointer; fun :Pointer; grad :Pointer; options :Pointer; func :Pointer; gradc :Pointer; progress :Pointer; warning :Pointer) :Double; stdcall; external 'kinopt.dll';


type


TDFP_MIN_Func  = function(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
TDFP_MIN_Check = function(call_obj :TObject; Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;

// both return 0 for success


function DFP_MIN(call_obj :TObject; NPar :Integer; InitPars :TPADouble; Hessian :TPMDouble;
        Func :TDFP_MIN_Func; Check :TDFP_MIN_Check;
				MaxIter :Integer; ConvLL :Double; ConvGrad :Double; StepMax :Double;
        var ErrOut :Integer) :Double; stdcall; external 'qubopt.dll' name '_dfp_opt_dll@56';

// errOut:  0   success
//          -1  func returned error
//          -2  check returned error
//          -3  exceeded max iter



function SVD(A :TPMDouble; m, n :Integer; W :TPADouble; V :TPMDouble) :Integer; stdcall; external 'qubopt.dll' name '_svdcmp_dll@20';


function qubopt_pyinit: Integer; stdcall; external 'qubopt.dll' name '_qubopt_pyinit@0';


function _BalanceAllLoops(Model :TQUB_QFS_PNodeMem): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_BalanceAllLoops@4';
function BalanceAllLoops(Model :INode): INode;

function _PareConstraints(Model, Removed :TQUB_QFS_PNodeMem): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_PareConstraints@8';
function PareConstraints(Model, Removed :INode): INode;

function _CheckLoopStimuli(Model :TQUB_QFS_PNodeMem): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_CheckLoopStimuli@4';
function CheckLoopStimuli(Model :INode): INode;

function _GetScaledRates(Model :TQUB_QFS_PNodeMem; StateFrom, StateTo :Integer): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name'__GetScaledRates@12';
function _GetScaledExpRates(Model :TQUB_QFS_PNodeMem; StateFrom, StateTo :Integer): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name'__GetScaledExpRates@12';

function GetScaledRates(Model :INode; StateFrom, StateTo :Integer): INode;
function GetScaledExpRates(Model :INode; StateFrom, StateTo :Integer): INode;


implementation


//stuff for SolvOpt

//function WCFun(Opt :TQUB_Thread_Idealize; x :TPADouble) :Double; stdcall; forward;
//procedure WCGrad(Opt :TQUB_Thread_Idealize; x, g :TPADouble); stdcall; forward;
//function WCDFPFunc(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall; forward;



function BalanceAllLoops(Model :INode): INode;
var
  ResultPtr     :TQUB_QFS_PNodeMem;
begin
  ResultPtr     := _BalanceAllLoops(Model.GetNodeMem);
  Result        := TNode.CreateUsing(ResultPtr);
  Node_DECREF(ResultPtr);
end;

function PareConstraints(Model, Removed :INode): INode;
var
  ResultPtr     :TQUB_QFS_PNodeMem;
begin
  ResultPtr     := _PareConstraints(Model.GetNodeMem, Removed.GetNodeMem);
  Result        := TNode.CreateUsing(ResultPtr);
  Node_DECREF(ResultPtr);
end;

function CheckLoopStimuli(Model :INode): INode;
var
  ResultPtr     :TQUB_QFS_PNodeMem;
begin
  ResultPtr     := _CheckLoopStimuli(Model.GetNodeMem);
  Result        := TNode.CreateUsing(ResultPtr);
  Node_DECREF(ResultPtr);
end;


function GetScaledRates(Model :INode; StateFrom, StateTo :Integer): INode;
var
  ResultPtr     :TQUB_QFS_PNodeMem;
begin
  ResultPtr     := _GetScaledRates(Model.GetNodeMem, StateFrom, StateTo);
  Result        := TNode.CreateUsing(ResultPtr);
  Node_DECREF(ResultPtr);
end;

function GetScaledExpRates(Model :INode; StateFrom, StateTo :Integer): INode;
var
  ResultPtr     :TQUB_QFS_PNodeMem;
begin
  ResultPtr     := _GetScaledExpRates(Model.GetNodeMem, StateFrom, StateTo);
  Result        := TNode.CreateUsing(ResultPtr);
  Node_DECREF(ResultPtr);
end;


//******************************************************************************
end.
