unit MessageCheckDialog;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls;


type


TMessageCheckDialog = class(TForm)


  chkCheck          :TCheckBox;
  txtMsg            :TMemo;


end;


function MessageCheckDlg(const Msg: string; const CheckMsg: String; var Check: Boolean; Buttons: TMsgDlgButtons): Integer;


implementation


{$R *.DFM}


const


PAD       :Integer      = 16;
BTNWIDTH  :Integer      = 75;
BTNHEIGHT :Integer      = 25;


procedure AddButton(Dlg :TMessageCheckDialog; var BtnX, BtnY :Integer; const Caption :String; ModalResult :Integer);
var
  Btn          :TButton;
begin
  Btn             := TButton.Create(Dlg);
  Btn.Parent      := Dlg;
  Btn.Height      := BTNHEIGHT;
  Btn.Width       := BTNWIDTH;
  Btn.Caption     := Caption;
  Btn.ModalResult := ModalResult;
  Btn.Top         := BtnY;
  Btn.Left        := BtnX;
  BtnX            := BtnX - BTNWIDTH - PAD;
end;

function MessageCheckDlg(const Msg: string; const CheckMsg: String; var Check: Boolean; Buttons: TMsgDlgButtons): Integer;
var
  Dlg        :TMessageCheckDialog;
  BtnX, BtnY :Integer;
begin
  Dlg                     := TMessageCheckDialog.Create(nil);
  Dlg.txtMsg.Lines.SetText(PChar(Msg));
  Dlg.chkCheck.Caption    := CheckMsg;
  Dlg.chkCheck.Checked    := Check;

  BtnX                    := Dlg.ClientWidth - PAD - BTNWIDTH;
  BtnY                    := Dlg.ClientHeight - PAD - BTNHEIGHT;

  if mbYes in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'Yes', mrYes);
  end;
  if mbNo in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'No', mrNo);
  end;
  if mbOK in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'OK', mrOK);
  end;
  if mbCancel in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'Cancel', mrCancel);
  end;
  if mbAbort in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'Abort', mrAbort);
  end;
  if mbRetry in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'Retry', mrRetry);
  end;
  if mbIgnore in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'Ignore', mrIgnore);
  end;
  if mbAll in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'All', mrAll);
  end;
  if mbNoToAll in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'No To All', mrNoToAll);
  end;
  if mbYesToAll in Buttons then begin
    AddButton(Dlg, BtnX, BtnY, 'Yes To All', mrYesToAll);
  end;

  Result := Dlg.ShowModal;;
  Check  := Dlg.chkCheck.Checked;
  Dlg.Release;
end;


//******************************************************************************
end.
