unit ModelEnergyPlotProperties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, RangedEdit;

type
  TModelEnergyPlotProperties = class(TForm)
    lblMaxStimulusNote: TLabel;
    lblMaxStimulus: TLabel;
    lblZeroState: TLabel;
    lblReferenceRate: TLabel;
    lblRefRateNote: TLabel;
    btnOK: TButton;
    btnCancel: TButton;
    procedure btnOKClick(Sender: TObject);
  private
    { Private declarations }
  public
    edtMaxStimulus: TRangedEdit;
    edtZeroState: TRangedEdit;
    edtReferenceRate: TRangedEdit;

    Procedure Setup( iZeroState, iStates : Integer; lStimulus : Boolean );
    Function GetZeroState() : Integer;
    Function GetReferenceRate() : Integer;
    Function GetMaxStimulus() : double;
  end;

implementation

{$R *.DFM}

//----- Fill in control values
Procedure TModelEnergyPlotProperties.Setup( iZeroState, iStates : Integer; lStimulus : Boolean );
begin
  // create ranged edit controls manually
  // added at start of tab order - Create in reverse order of desired tab order
  edtMaxStimulus := TRangedEdit.Create(Self);
  edtMaxStimulus.Setup( Self, 130, 124, 70, 21, 2, '10', 0.0, 100000.0 );

  edtReferenceRate := TRangedEdit.Create(Self);
  edtReferenceRate.Setup( Self, 130, 64, 70, 21, 2, '1000000', 100.0, 1e12);

  edtZeroState := TRangedEdit.Create(Self);
  edtZeroState.Setup( Self, 130, 24, 70, 21, 2, IntToStr(iZeroState), 1.0, iStates );
  edtZeroState.MaxValue := iStates;

  // set which are visible
  lblMaxStimulus.Visible := lStimulus;
  edtMaxStimulus.Visible := lStimulus;
  lblMaxStimulusNote.Visible := lStimulus;

  //TODO - Read Registry
end;

procedure TModelEnergyPlotProperties.btnOKClick(Sender: TObject);
begin
  // TODO - Write Registry
end;

//----- Return control values
Function TModelEnergyPlotProperties.GetZeroState() : Integer;
begin
  result := StrToInt(edtZeroState.Text)
end;

Function TModelEnergyPlotProperties.GetReferenceRate() : Integer;
begin
  result := StrToInt(edtReferenceRate.Text)
end;

Function TModelEnergyPlotProperties.GetMaxStimulus() : double;
begin
  result := StrToFloat(edtMaxStimulus.Text)
end;

end.
