unit QFC_FittingDlog;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Math,

  QUBFits, QUBFits_GUI;


const

// how many times to pause and process messages (eg the stop button)

SLOWNESS_FOR_THE_STOP_BUTTON = 8;
MSG_START                    = 4321;


type
  TQFC_Fitting_Dlog = class(TForm)
    progIterations: TProgressBar;
    btnStop: TButton;
    procedure FormShow(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
  private
    procedure prOnStart(var Message :TMessage); message MSG_START;
    procedure prOnFitterIter(Sender :TObject);
    procedure prOnFitterDone(Sender :TObject);
  public
    panFits      :TQUB_DialACurve;
  end;

var
  QFC_Fitting_Dlog: TQFC_Fitting_Dlog;


implementation

{$R *.dfm}

procedure TQFC_Fitting_Dlog.FormShow(Sender: TObject);
begin
  panFits.Fits.Events.AddEvent('OnFitterIter', prOnFitterIter);
  panFits.Fits.Events.AddEvent('OnFitterDone', prOnFitterDone);
  progIterations.Max := Round(Abs(panFits.Fits.Fitter.GetParamVal(0)));
  progIterations.Position := 0;
  PostMessage(Handle, MSG_START, 0, 0);
end;

procedure TQFC_Fitting_Dlog.FormHide(Sender: TObject);
begin
  panFits.Fits.Events.DelEvent('OnFitterIter', prOnFitterIter);
  panFits.Fits.Events.DelEvent('OnFitterDone', prOnFitterDone);
end;

procedure TQFC_Fitting_Dlog.prOnStart(var Message :TMessage);
begin
  panFits.Fits.Fit();
end;

procedure TQFC_Fitting_Dlog.btnStopClick(Sender: TObject);
begin
  panFits.Fits.Fitter.Stop();
  panFits.Fits.Events.RaiseNotifyEvent('OnClickStop', panFits.Fits);
end;

procedure TQFC_Fitting_Dlog.prOnFitterIter(Sender :TObject);
begin
  progIterations.Position := panFits.Fits.Fitter.GetIterations();
  // check the stop button and repaint up to 8 times
  if 0 = (progIterations.Position mod (Max(1,progIterations.Max div SLOWNESS_FOR_THE_STOP_BUTTON))) then Application.ProcessMessages;
end;

procedure TQFC_Fitting_Dlog.prOnFitterDone(Sender :TObject);
begin
  ModalResult := mrOK;
end;

end.
