unit QFC_Log;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses

StdCtrls, Classes, SysUtils, Windows, ShellAPI;


type

TQFC_Log = class(TCustomListBox)
protected
  prFullItems  :TStringList;
  prURLs       :TStringList;
  prDepth      :Integer;
  prLastIndex  :Integer;

  procedure prSetDepth(x :Integer);
  procedure prOnClick(Sender :TObject);
  procedure prOnKeyUp(Sender :TObject; var Key :Word; Shift :TShiftState);
  procedure prOnDblClick(Sender :TObject);

public
  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure Add(Txt, URL :String);

  property Depth     :Integer   read prDepth        write prSetDepth;
end;




implementation

constructor TQFC_Log.Create(AOwner :TComponent);
begin
  inherited;

  prFullItems := TStringList.Create;
  prURLS      := TStringList.Create;
  prDepth     := 50;
  prLastIndex := 0;

  OnClick     := prOnClick;
  OnKeyUp     := prOnKeyUp;
  OnDblClick  := prOnDblClick;
  ShowHint    := True;
end;

destructor TQFC_Log.Destroy;
begin
  try
    prFullItems.Free;
    prURLS.Free;
  finally
    inherited;
  end;
end;

procedure TQFC_Log.Add(Txt, URL :String);
var
  endl   :Integer;
  Msg    :String;
begin
  endl   := AnsiPos(#13#10, Txt);
  if endl > 0 then begin
    Msg  := Copy(Txt, 1, endl-1) + '...';
  end else begin
    Msg  := Txt;
  end;

  if Items.Count = prDepth then begin
    Items.Delete(0);
    prFullItems.Delete(0);
    prURLS.Delete(0);
  end;

  Items.Add(Msg);
  prFullItems.Add(Txt);
  prURLS.Add(URL);
  ItemIndex := Items.Count - 1;
  prOnClick(Self);
end;

procedure TQFC_Log.prSetDepth(x :Integer);
begin
  prDepth := x;
  while Items.Count > prDepth do begin
    Items.Delete(0);
    prFullItems.Delete(0);
    prURLS.Delete(0);
  end;
end;

procedure TQFC_Log.prOnClick(Sender :TObject);
begin
  if ItemIndex < 0 then ItemIndex := prLastIndex;
  if ItemIndex < 0 then Exit;
  prLastIndex := ItemIndex;
  Hint        := prFullItems[ItemIndex];
end;

procedure TQFC_Log.prOnKeyUp(Sender :TObject; var Key :Word; Shift :TShiftState);
begin
  inherited;
  prOnClick(Sender);
end;

procedure TQFC_Log.prOnDblClick(Sender :TObject);
begin
  if (ItemIndex >= 0) and (prURLS[ItemIndex] <> '')
  then ShellExecute(Handle, 'open', PChar(prURLS[ItemIndex]), nil, nil, SW_SHOWNORMAL);
end;


//******************************************************************************
end.
