unit QUBFits;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// interface to qubfits.dll

interface

uses

Classes, Math, SysUtils, Windows,

Alloc, AllocTypes, QUB_Node, QUB_Profiles, QUB_Events, QUB_AdaptiveResample;


// from qubfits.dll:


function _InitQUBFits: Integer; stdcall; external 'qubfits.dll' name '_initqubfits@0';


type

IQUB_EvtListener = interface(IUnknown)
 // ok ms and borland forced me to inherit from iunknown
 // but this is official notice i will ignore those 3 methods.
 // please implement them with an empty body, and never call them.
    //function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    //function _AddRef: Integer; stdcall;
    //function _Release: Integer; stdcall;
  procedure QUBEvt(Name :PChar; Sender :Pointer); cdecl;
end;

TQUB_Evts = class
public
  procedure Free; virtual; cdecl; abstract;
  procedure Add(Name :PChar; Rcvr :IQUB_EvtListener); virtual; cdecl; abstract;
  procedure Del(Name :PChar; Rcvr :IQUB_EvtListener); virtual; cdecl; abstract;
  procedure Send(Name :PChar; Sender :Pointer); virtual; cdecl; abstract;
end;

function QUB_Evts_Create: TQUB_Evts; stdcall; external 'qubfits.dll' name '_QUB_Evts_Create@0';

type

TQF_Data = record
  N          :Integer;
  XX, YY, WW :TPADouble;
  NV         :Integer;
  VName      :TPMChar;
  VVV        :TPMDouble;
end;
TPQF_Data                              = ^TQF_Data;
TAQF_Data                              = Array[0..(MaxInt div (SizeOf(TQF_Data)*2))] of TQF_Data;
TPAQF_Data                             = ^TAQF_Data;


TQUB_Curve = class
public
  procedure Free; virtual; cdecl; abstract;
  function  Events: TQUB_Evts; virtual; cdecl; abstract;
	// Curve CompCount
	// Curve Eqn
	// Curve Params
	// Curve ParamVal
        // Curve ParamConst
        // Curve Msg
        // Curve Err
  procedure SetExtraVars(Names :TPMChar; Count :Integer); virtual; cdecl; abstract;
  function  GetName: PChar; virtual; cdecl; abstract;
  function  GetEqn: PChar; virtual; cdecl; abstract;
  procedure SetEqn(x :PChar); virtual; cdecl; abstract;
  function  CanSetEqn: Integer; virtual; cdecl; abstract;
  function  GetComponentCount: Integer; virtual; cdecl; abstract;
  procedure SetComponentCount(x :Integer); virtual; cdecl; abstract;
  function  GetParamCount: Integer; virtual; cdecl; abstract;
  function  GetParamName(ix :Integer): PChar; virtual; cdecl; abstract;
  function  GetParamVal(ix :Integer): Double; virtual; cdecl; abstract;
  procedure SetParamVal(ix :Integer; val :Double; sendEvent :Integer); virtual; cdecl; abstract;
  function  GetParamConst(ix :Integer): Integer; virtual; cdecl; abstract;
  procedure SetParamConst(ix :Integer; val :Integer); virtual; cdecl; abstract;
  procedure InitParams(xx, yy :TPADouble; n :Integer); virtual; cdecl; abstract;
  procedure EvalArrayAll(var Data :TQF_Data; yy :TPADouble); virtual; cdecl; abstract;
  procedure EvalArrayOne(var Data :TQF_Data; yy :TPADouble; Component :Integer); virtual; cdecl; abstract;
  function  HasDeriv: Integer; virtual; cdecl; abstract;
  procedure EvalDerivArray(var Data :TQF_Data; dPar :TPADouble); virtual; cdecl; abstract;
  function  GetMessage: PChar; virtual; cdecl; abstract;
  function  GetState: TQUB_QFS_PNodeMem; virtual; cdecl; abstract;
  procedure UpdateState(state :TQUB_QFS_PNodeMem); virtual; cdecl; abstract;
end;

TQUB_CurveFactory = function: TQUB_Curve; stdcall;

procedure RegisterCurve(name :PChar; factory :TQUB_CurveFactory); stdcall; external 'qubfits.dll' name '_RegisterCurve@8';
function  CreateCurve(name :PChar): TQUB_Curve; stdcall; external 'qubfits.dll' name '_CreateCurve@4';
function  ListCurves: TQUB_QFS_PNodeMem; stdcall; external 'qubfits.dll' name '_ListCurves@0';


type

TQUB_Fitter = class
public
  procedure Free; virtual; cdecl; abstract;
  function  Events: TQUB_Evts; virtual; cdecl; abstract;
	// Fitter ParamVal
	// Fitter Iter
	// Fitter Msg
	// Fitter Err
  function  GetName: PChar; virtual; cdecl; abstract;
  function  GetParamCount: Integer; virtual; cdecl; abstract;
  function  GetParamName(ix :Integer): PChar; virtual; cdecl; abstract;
  function  GetParamVal(ix :Integer): Double; virtual; cdecl; abstract;
  procedure SetParamVal(ix :Integer; val :Double); virtual; cdecl; abstract;
  function  Fit(Curve :TQUB_Curve; Data :TPAQF_Data; NData :Integer): Integer; virtual; cdecl; abstract;
  procedure Stop; virtual; cdecl; abstract; // from another thread or iter msg
  // info during/after iteration:
  function  GetIterations: Integer; virtual; cdecl; abstract;
  function  GetParamErr(ix :Integer): Double; virtual; cdecl; abstract;
  function  GetCurve: TPADouble; virtual; cdecl; abstract;
  function  GetMessage: PChar; virtual; cdecl; abstract; // if event msg or err
  // anytime:
  function  GetState: TQUB_QFS_PNodeMem; virtual; cdecl; abstract;
  procedure UpdateState(state :TQUB_QFS_PNodeMem); virtual; cdecl; abstract;
end;

TQUB_FitterFactory = function: TQUB_Fitter; stdcall;

procedure RegisterFitter(name :PChar; factory :TQUB_FitterFactory); stdcall; external 'qubfits.dll' name '_RegisterFitter@8';
function  CreateFitter(name :PChar): TQUB_Fitter; stdcall; external 'qubfits.dll' name '_CreateFitter@4';
function  ListFitters: TQUB_QFS_PNodeMem; stdcall; external 'qubfits.dll' name '_ListFitters@0';

function ChiSqrUncertainty(Curve :TQUB_Curve; var Data :TQF_Data; FF :TPADouble; iPar :Integer; deltaPct {0<d<1}:Double): Double; stdcall; external 'qubfits.dll' name '_ChiSqrUncertainty@24';


function FilterDataD(Data, FData :TPADouble; ndata :Integer; dt, freq :Double): Integer; stdcall; external 'qubfits.dll' name '_filterdatad@28';



// defined below:


function InitQUBFits(PluginPath :String): Integer;

// just zeros out the pointers:
procedure QF_Data_Init(var Data :TQF_Data);
// you mainly point the QF_Data to arrays and strings of your own
// and free the var arrays when you're done
procedure QF_Data_SetVars(var Data :TQF_Data; Vars :TStringList{name->TPADouble});
procedure QF_Data_FreeVars(var Data :TQF_Data);

// for persistance eg with global fitting,
// allocate internal storage and copy the arrays and strings
// then free storage when you're done
procedure QF_Data_CopyIn(var Data :TQF_Data);
procedure QF_Data_FreeIn(var Data :TQF_Data);

// clean up result with _FreeVars, _FreeIn
function QF_Data_Copy(var Data :TQF_Data): TQF_Data;

function QF_Data_SafeVarName(Nm :String): String;


type

TQUBFits_Plugin_Init = function: integer; stdcall;

TQUB_FitDataset = class
protected
  prVars              :TStringList; // string -> TPADouble
  prOwnStorage         :Boolean;
  prXX, prYY, prWW     :TPADouble;
  prXoff               :Double;
  prXindex, prYindex   :Integer;
  prWindex             :Integer;
  prWinverse           :Boolean;
  prWmean              :Double;
  prSortIx             :TList;

  function prGetVarCount: Integer;
  function prGetVarName(Ix :Integer): String;
  procedure prSetVarName(Ix :Integer; x :String);
  function prGetVals(Ix :Integer) :TPADouble;
  procedure prSetXindex(x :Integer);
  procedure prSetYindex(x :Integer);
  procedure prSetWindex(x :Integer);
  procedure prSetWinverse(x :Boolean);

  procedure prSteal(Other :TQUB_FitDataSet);// copies names and array pointers and takes ownership if possible

public
  Data              :TQF_Data;

  constructor Create; overload;
  constructor Create(aN :Integer); overload;     // for use with addvar
  destructor Destroy; override;

  //procedure Reset(NewCount :Integer);            // for use with addvar
  procedure AddVar(Name :String; Values :TPADouble);
  procedure InsertVar(Name :String; Values :TPADouble; Ix :Integer);

  procedure Assign(Other :TQUB_FitDataSet);// copies names and array pointers
  procedure MakeOwn; // copy external storage to internal
  procedure Realloc(NewCount :Integer);    // (re-)allocs own storage
  procedure ReMeasureW; // recalc wmean after outside changes

  function IndexOf(const VName :String): Integer;
  function XRange(t0, t1 :Double; var x0, x1 :Integer): Boolean;
  procedure Sort;
  function Compare(p, q :Pointer): Integer;

  property N                      :Integer     read Data.N;
  property XX                     :TPADouble   read Data.XX;
  property YY                     :TPADouble   read Data.YY;
  property WW                     :TPADouble   read Data.WW;
  property VarCount               :Integer     read prGetVarCount;
  property VarName[Ix :Integer]   :String      read prGetVarName   write prSetVarName;
  property Vals[Ix :Integer]      :TPADouble   read prGetVals;
  property Xoff                   :Double      read prXoff         write prXoff;
  property Xindex                 :Integer     read prXindex       write prSetXindex;
  property Yindex                 :Integer     read prYindex       write prSetYindex;
  property Windex                 :Integer     read prWindex       write prSetWindex;
  property Winverse               :Boolean     read prWinverse     write prSetWinverse;
  property Wmean                  :Double      read prWmean;
end;

TQUB_FitData_DisplayStyle = (QUBFITS_DISPLAY_DEFAULT,
                             QUBFITS_DISPLAY_POINTS,
                             QUBFITS_DISPLAY_HIST);

TQUB_FitData = class
protected
  prProfile            :INode;

  prDataSet            :TQUB_FitDataSet;
  prUseDataSet         :TQUB_FitDataSet;
  prFY                 :TPADouble;
  prXName, prYName     :String;
  prXmin, prXmax       :Double;
  prAbsX               :Boolean;
  prDisplayStyle       :TQUB_FitData_DisplayStyle;
  prCanFilter          :Boolean;
  prWname              :String;
  prWinverse           :Boolean;

  prSelFromPct         :Double;
  prSelToPct           :Double;
  prFilter             :Boolean;
  prFreq               :Double;
  prResample           :Boolean;
  prResampleStd        :Double;

  //function  prGetXName: String;
  procedure prSetXName(x :String);
  //function  prGetYName: String;
  procedure prSetYName(x :String);
  procedure prSetAbsX(x :Boolean);
  procedure prSetDisplayStyle(x :TQUB_FitData_DisplayStyle);
  procedure prSetSelFrom(x :Double);
  procedure prSetSelTo(x :Double);
  procedure prSetSelFromS(x :Double);
  procedure prSetSelToS(x :Double);
  function  prGetSelFromS: Double;
  function  prGetSelToS: Double;
  procedure prSetFilter(x :Boolean);
  procedure prSetFreq(x :Double);
  procedure prSetResample(x :Boolean);
  procedure prSetResampleStd(x :Double);

  procedure prOnPreSet(Sender :TQUB_ProfileCat; Updates :INode);

  procedure prReSubData;

public
  OnVarsChange          :TNotifyEvent;
  OnChange              :TNotifyEvent;

  constructor Create;
  destructor Destroy; override;

  procedure UseData(aDataSet :TQUB_FitDataset);
  procedure UseWeights(Name :String; Inverse :Boolean);

  property Profile      :INode     read prProfile;

  property DataSet      :TQUB_FitDataset   read prDataSet;
  property UseDataSet   :TQUB_FitDataset   read prUseDataSet;
  property XName        :String            read prXName   write prSetXName;
  property YName        :String            read prYName   write prSetYName;
  property AbsX         :Boolean           read prAbsX       write prSetAbsX;
  property DisplayStyle :TQUB_FitData_DisplayStyle read prDisplayStyle write prSetDisplayStyle;
  property CanFilter    :Boolean           read prCanFilter;
  property WName        :String            read prWName;
  property Winverse     :Boolean           read prWinverse;

  property FY           :TPADouble read prFY;
  // FY uses Use.X, Use.Count
  property Xmin         :Double    read prXmin;
  property Xmax         :Double    read prXmax;

  property SelFromPct   :Double    read prSelFromPct   write prSetSelFrom;
  property SelToPct     :Double    read prSelToPct     write prSetSelTo;
  property SelFromS     :Double    read prGetSelFromS  write prSetSelFromS;
  property SelToS       :Double    read prGetSelToS    write prSetSelToS;
  property Filter       :Boolean   read prFilter       write prSetFilter;
  property Freq         :Double    read prFreq         write prSetFreq;
  property Resample     :Boolean   read prResample     write prSetResample;
  property ResampleStd  :Double    read prResampleStd  write prSetResampleStd;

end;


TQUB_Fits = class(TObject, IQUB_EvtListener)
protected
  prData           :TQUB_FitData;
  prCurve          :TQUB_Curve;
  prFitter         :TQUB_Fitter;
  prLastIter       :Integer;
  prFitting        :Boolean;

  prCurves         :TStringList;
  prFitters        :TStringList;

  prCurveRestore   :INode;
  prCurveActive    :INode;
  prFitterRestore  :INode;
  prFitterActive   :INode;

  procedure prSetCurve(x :TQUB_Curve);
  procedure prSetFitter(x :TQUB_Fitter);

  function prGetCurveCount: Integer;
  function prGetCurveName(Ix :Integer): String;
  function prGetFitterCount: Integer;
  function prGetFitterName(Ix :Integer): String;

  procedure prOnCurvePreSet(Sender :TQUB_ProfileCat; Updates :INode);
  procedure prOnFitterPreSet(Sender :TQUB_ProfileCat; Updates :INode);

  procedure prOnVarsChange(Sender :TObject);
  procedure prOnDataChange(Sender :TObject);
  procedure prReCalc;

public
  Events           :TQUB_Events;
  //OnCurveChange    :TNotifyEvent;
  //OnCurveCompCount :TNotifyEvent;
  //OnCurveEqn       :TNotifyEvent;
  //OnCurveParams    :TNotifyEvent;
  //OnCurveParamVal  :TNotifyEvent;
  //OnCurveParamConst:TNotifyEvent;
  //OnCurveMsg       :TNotifyEvent;
  //OnCurveErr       :TNotifyEvent;

  //OnFitterChange   :TNotifyEvent;
  //OnFitterParamVal :TNotifyEvent;
  //OnFitterIter     :TNotifyEvent;
  //OnFitterMsg      :TNotifyEvent;
  //OnFitterErr      :TNotifyEvent;
  //OnFitterDone     :TNotifyEvent;

  //OnVarsChange     :TNotifyEvent;
  //OnDataChange     :TNotifyEvent;
  //OnErrorsChange   :TNotifyEvent;

  //OnClickStop (from QFC_FittingDlog) after it calls Fitter.Stop()

  Errors             :array of Double; // per curve param
  Residual           :Double; // sum-square difference

  constructor Create;
  destructor Destroy; override;

  // bs delphi interface com garbage:
  function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
  function _AddRef: Integer; stdcall;
  function _Release: Integer; stdcall;
  // the real interface function:
  procedure QUBEvt(Name :PChar; Sender :Pointer); cdecl;

  procedure Fit;

  property Data    :TQUB_FitData          read prData;
  property Curve   :TQUB_Curve            read prCurve      write prSetCurve;
  property Fitter  :TQUB_Fitter           read prFitter     write prSetFitter;

  property CurveCount             :Integer     read prGetCurveCount;
  property CurveName[ix :Integer] :String      read prGetCurveName;
  property FitterCount            :Integer     read prGetFitterCount;
  property FitterName[ix :Integer]:String      read prGetFitterName;

end;


const

  FITDATA_CHISQR_DELTAPCT = 0.01; // 1%



implementation


function InitQUBFits(PluginPath :String): Integer;
var
  j            :Integer;
  Search       :TSearchRec;
  Filename     :String;
  Init         :TQUBFits_Plugin_Init;
  DLLHandle    :THandle;
begin
  Result := _InitQUBFits;
  if Result <> 0 then Exit;

  // dir traversal
  j:=FindFirst(PluginPath+'*.dll',faAnyFile,Search);
  While j=0 do begin
    DLLHandle := 0;
    try
      Filename  := PluginPath+Search.Name;
      DLLHandle := LoadLibrary(PChar(Filename));
      if DLLHandle <> 0 then begin
        @Init := GetProcAddress(DLLHandle, '_initqubfits_plugin@0');
        Result := Result + Init;
      end;
    except
      if DLLHandle <> 0 then FreeLibrary(DLLHandle);
    end;

    j:=FindNext(Search);                         //Get next file now
  end; //while
{You must qualify the FindClose because of a naming conflict with the
windows API}
  sysutils.FindClose(Search);
end;


//******************************************************************************
procedure QF_Data_Init(var Data :TQF_Data);
begin
  with Data do begin
    VName    := nil;
    VVV      := nil;
    XX       := nil;
    YY       := nil;
    WW       := nil;
    N        := 0;
    NV       := 0;
  end;
end;

procedure QF_Data_SetVars(var Data :TQF_Data; Vars :TStringList{name->TPADouble});
var
  i          :Integer;
begin
  with Data do begin
    if Vars.Count > 0 then begin
      ReAllocMem(VName, Vars.Count * SizeOf(Pointer));
      ReAllocMem(VVV, Vars.Count * SizeOf(Pointer));
    end else begin
      VName    := nil;
      VVV      := nil;
    end;
    NV         := Vars.Count;
    for i := 0 to Vars.Count - 1 do begin
      VName[i] := PChar(Vars[i]);
      VVV[i]   := TPADouble(Vars.Objects[i]);
    end;
  end;
end;

procedure QF_Data_FreeVars(var Data :TQF_Data);
begin
  FreeMem(Data.VName); Data.VName := nil;
  FreeMem(Data.VVV);   Data.VVV   := nil;
  Data.NV := 0;
end;

procedure QF_Data_CopyIn(var Data :TQF_Data);
var
  i       :Integer;

  function CopyString(S :PChar): PChar;
  var
    N  :Integer;
  begin
    N     := 0;
    while S[N] <> #0 do Inc(N);
    Result := AllocMem(N+1);
    Move(S[0], Result[0], N+1);
  end;
  function CopyArr(N :Integer; Arr :TPADouble): TPADouble;
  begin
    if N = 0 then Result := nil
    else begin
      Result   := AllocMem(N * SizeOf(Double));
      Move(Arr[0], Result[0], N*SizeOf(Double));
    end;
  end;
begin
  with Data do begin
    XX := CopyArr(N, XX);
    YY := CopyArr(N, YY);
    WW := CopyArr(N, WW);
    for i := 0 to NV - 1 do begin
      VName[i] := CopyString(VName[i]);
      VVV[i]   := CopyArr(N, VVV[i]);
    end;
  end;
end;

procedure QF_Data_FreeIn(var Data :TQF_Data);
var
  i            :Integer;
begin
  with Data do begin
    FreeMem(XX);     XX := nil;
    FreeMem(YY);     YY := nil;
    FreeMem(WW);     WW := nil;
    for i := 0 to NV - 1 do begin
      FreeMem(VVV[i]);   VVV[i]   := nil;
      FreeMem(VName[i]); VName[i] := nil;
    end;
  end;
end;

function QF_Data_Copy(var Data :TQF_Data): TQF_Data;
var
  i          :Integer;
begin
  with Data do begin
    if NV > 0 then begin
      Result.VName := AllocMem(NV * SizeOf(Pointer));
      Result.VVV   := AllocMem(NV * SizeOf(Pointer));
    end else begin
      Result.VName    := nil;
      Result.VVV      := nil;
    end;

    for i := 0 to NV - 1 do begin
      Result.VName[i] := VName[i];
      Result.VVV[i]   := VVV[i];
    end;

    Result.XX := XX;
    Result.YY := YY;
    Result.WW := WW;
    Result.N  := N;
    Result.NV := NV;

    QF_Data_CopyIn(Result);
  end;
end;


function QF_Data_SafeVarName(Nm :String): String;
// python-safe identifiers only; this is not a very complete fix
begin
  Result := Nm;
  Result := StringReplace(Result, ' ', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '=', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '+', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '-', '_', [rfReplaceAll]);
  Result := StringReplace(Result, ')', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '(', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '*', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '&', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '#', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '@', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '!', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '~', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '[', '_', [rfReplaceAll]);
  Result := StringReplace(Result, ']', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '{', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '}', '_', [rfReplaceAll]);
  Result := StringReplace(Result, ':', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '.', '_', [rfReplaceAll]);
  Result := StringReplace(Result, ',', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '/', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '?', '_', [rfReplaceAll]);
  // ...
  if (Result = '') or ((Ord(Result[1]) >= Ord('0')) and (Ord(Result[1]) <= Ord('9')))
  then Result := '_'+Result;
end;

//******************************************************************************
constructor TQUB_FitDataSet.Create;
begin
  prVars := TStringList.Create;
  prXindex := -1;
  prYindex := -1;
  prWindex := -1;
  QF_Data_Init(Data);
end;

constructor TQUB_FitDataSet.Create(aN :Integer);
var
  i        :Integer;
begin
  prVars  := TStringList.Create;
  QF_Data_Init(Data);
  Data.N  := aN;
  prOwnStorage := False;
  prXindex := -1;
  prYindex := -1;
  prWindex := -1;
  prWW     := AllocMem(aN * SizeOf(Double));
  for i := 0 to aN - 1 do begin
    prWW[i] := 1.0;
  end;
  Data.WW  := prWW;
  prWmean  := 1.0;
end;

destructor TQUB_FitDataSet.Destroy;
begin
  try
    Realloc(0);
    prVars.Free;
    QF_Data_FreeVars(Data);
  finally
    inherited;
  end;
end;

procedure TQUB_FitDataSet.AddVar(Name :String; Values :TPADouble);
begin
  prVars.AddObject(Name, TObject(Values));
  if prVars.Count = 1 then begin
    prXindex := 0;
    prXX     := Values;
    Data.XX := prXX;
  end else if prVars.Count = 2 then begin
    prYindex := 1;
    prYY     := Values;
    Data.YY := prYY;
  end else if Name = 'weight' then begin
    Windex   := prVars.Count - 1;
  end;
  QF_Data_SetVars(Data, prVars);
end;

procedure TQUB_FitDataSet.InsertVar(Name :String; Values :TPADouble; Ix :Integer);
begin
  prVars.InsertObject(Ix, Name, TObject(Values));
  if Ix = 0 then begin
    prXindex := 0;
    prXX     := Values;
    Data.XX := prXX;
    if prYindex >= 0 then begin
      prYindex := prYindex + 1;
      prYY     := Vals[prYindex];
      Data.YY := prYY;
    end else if prVars.Count > 1 then begin
      prYindex := 1;
      prYY     := Vals[1];
      Data.YY := prYY;
    end;
    if Windex >= 0 then begin
      Windex   := Windex + 1;
    end;
  end;
  QF_Data_SetVars(Data, prVars);
end;

procedure TQUB_FitDataSet.Assign(Other :TQUB_FitDataSet); // copies names and array pointers
var
  i            :Integer;
begin
  Realloc(0);
  prOwnStorage := False;

  Data.N := Other.N;
  ReAllocMem(prWW, Data.N * SizeOf(Double));
  Data.WW := prWW;
  prVars.Clear;
  for i := 0 to Other.VarCount - 1 do begin
    AddVar(Other.VarName[i], Other.Vals[i]);
  end;
  QF_Data_SetVars(Data, prVars);

  Xindex := Other.Xindex;
  Yindex := Other.Yindex;
  Winverse := Other.Winverse;
  Windex := Other.Windex;
end;

procedure TQUB_FitDataSet.prSteal(Other :TQUB_FitDataSet);
begin
  Assign(Other);
  if Other.prOwnStorage then begin
    prOwnStorage := True;
    Other.prOwnStorage := False;
  end;
end;

procedure TQUB_FitDataSet.MakeOwn;
var
  Arr   :TPADouble;
  i     :Integer;
begin
  if not prOwnStorage then begin
    for i := 0 to VarCount - 1 do begin
      Arr := AllocMem(N*SizeOf(Double));
      Move(Vals[i][0], Arr[0], N*SizeOf(Double));
      prVars.Objects[i] := TObject(Arr);
    end;
    prOwnStorage := True;
  end;
  QF_Data_SetVars(Data, prVars);
  if prXindex >= 0 then begin
    prXX := Vals[prXindex];
    Data.XX := prXX;
  end;
  if prYindex >= 0 then begin
    prYY := Vals[prYindex];
    Data.YY := prYY;
  end;
end;

procedure TQUB_FitDataSet.Realloc(NewCount :Integer);    // (re-)allocs own storage
var
  i            :Integer;
begin
  if prOwnStorage then begin
    for i := 0 to VarCount - 1 do begin
      FreeMem(Vals[i]);     prVars.Objects[i] := nil;
    end;
  end;

  if NewCount > 0 then begin
    for i := 0 to VarCount - 1 do begin
      prVars.Objects[i] := AllocMem(NewCount*SizeOf(Double));
    end;
    prOwnStorage := True;
  end;
  Data.N         := NewCount;
  QF_Data_SetVars(Data, prVars);
  if prXindex >= 0 then begin
    prXX := Vals[prXindex];
    Data.XX := prXX;
  end;
  if prYindex >= 0 then begin
    prYY := Vals[prYindex];
    Data.YY := prYY;
  end;
  ReAllocMem(prWW, Data.N * SizeOf(Double));
  Data.WW   := prWW;
  Windex         := Windex;
end;

function TQUB_FitDataSet.IndexOf(const VName :String): Integer;
begin
  Result := prVars.IndexOf(VName);
end;

function TQUB_FitDataSet.XRange(t0, t1 :Double; var x0, x1 :Integer): Boolean;
begin
  x0 := 0;
  while (x0 < N) and (XX[x0] < t0) do Inc(x0);
  x1 := x0;
  while (x1 < N) and (XX[x1] <= t1) do Inc(x1);
  Dec(x1);
  Result := (x0 <= x1) and (XX[x1] <= t1);
end;

var
  gSortFitDataSet :TQUB_FitDataSet;

function CompareFDSPoints(p, q :Pointer): Integer;
begin
  Result := gSortFitDataSet.Compare(p, q);
end;

procedure TQUB_FitDataSet.Sort;
var
  i, j     :Integer;
  Sorted   :TQUB_FitDataSet;
  A, B     :TPADouble;
begin
  if prXX = nil then Exit;
  prSortIx := TList.Create;
  for i := 0 to N - 1 do begin
    prSortIx.Add(Pointer(i));
  end;
  gSortFitDataSet := Self;
  prSortIx.Sort(CompareFDSPoints);
  Sorted   := TQUB_FitDataSet.Create;
  Sorted.Assign(Self);
  Sorted.Realloc(N);
  for j := 0 to VarCount - 1 do begin
    A      := Vals[j];
    B      := Sorted.Vals[j];
    for i := 0 to N - 1 do begin
      B[i] := A[Integer(prSortIx[i])];
    end;
  end;
  A        := prWW;
  B        := Sorted.WW;
  for i := 0 to N - 1 do begin
    B[i]   := A[Integer(prSortIx[i])];
  end;
  prSteal(Sorted);
  Sorted.Free;
  prSortIx.Free;
end;

function TQUB_FitDataSet.Compare(p, q :Pointer): Integer;
var
  x1, x2        :Double;
begin
  x1  := prXX[Integer(p)];
  x2  := prXX[Integer(q)];
  if x1 < x2 then begin
    Result := -1;
  end else if x1 > x2 then begin
    Result := 1;
  end else begin
    Result := 0;
  end;
end;

function TQUB_FitDataSet.prGetVarCount: Integer;
begin
  Result := prVars.Count;
end;

function TQUB_FitDataSet.prGetVarName(Ix :Integer): String;
begin
  Result := prVars[Ix];
end;

procedure TQUB_FitDataSet.prSetVarName(Ix :Integer; x :String);
begin
  prVars[Ix] := x;
  QF_Data_SetVars(Data, prVars);
end;

function TQUB_FitDataSet.prGetVals(Ix :Integer) :TPADouble;
begin
  Result := TPADouble(prVars.Objects[Ix]);
end;

procedure TQUB_FitDataSet.prSetXindex(x: Integer);
begin
  prXindex := x;
  if x >= 0 then begin
    prXX   := Vals[x];
  end else begin
    prXX   := nil;
  end;
  Data.XX := prXX;
end;

procedure TQUB_FitDataSet.prSetYindex(x :Integer);
begin
  prYindex := x;
  if x >= 0 then begin
    prYY   := Vals[x];
  end else begin
    prYY   := nil;
  end;
  Data.YY := prYY;
end;

procedure TQUB_FitDataSet.prSetWindex(x :Integer);
var
  i        :Integer;
  Wmax     :Double;
begin
  prWindex := x;
  if x >= 0 then begin
    if Winverse then begin
      for i := 0 to N - 1 do begin
        if Vals[x][i] <> 0.0 then begin
          prWW[i] := Abs(1.0 / Vals[x][i]);
        end else begin
          prWW[i] := 0.0;
        end;
      end;
    end else begin
      for i := 0 to N - 1 do begin
        prWW[i] := Abs(Vals[x][i]);
      end;
    end;
  end else begin
    for i := 0 to N - 1 do begin
      prWW[i] := 1.0;
    end;
  end;

  // normalize weights to have max of 1.0 for stability
  if N > 0 then begin
    Wmax := prWW[0];
    for i := 1 to N - 1 do begin
      Wmax := Max(Wmax, prWW[i]);
    end;
    for i := 0 to N - 1 do begin
      prWW[i]   := prWW[i] / Wmax;
    end;
  end;

  ReMeasureW;
end;

procedure TQUB_FitDataSet.ReMeasureW;
var
  i        :Integer;
  Wsum     :Double;
begin
  if N > 0 then begin
    Wsum := 0.0;
    for i := 1 to N - 1 do begin
      Wsum := Wsum + prWW[i];
    end;
    prWmean := Wsum / N;
  end else begin;
    prWmean   := 0.0;
  end;
end;

procedure TQUB_FitDataSet.prSetWinverse(x :Boolean);
begin
  prWinverse := x;
  Windex := Windex;
end;


//******************************************************************************

constructor TQUB_FitData.Create;
begin
  prProfile             := ProfileMgr['FitData'].Active;
  ProfileMgr['FitData'].OnSet := prOnPreSet;

  if prProfile.Find('SelFrom').DataCount = 0 then prProfile['SelFrom'].SetData(0.0/1.0);
  if prProfile.Find('SelTo').DataCount = 0 then prProfile['SelTo'].SetData(100.0/1.0);
  if prProfile.Find('Filter').DataCount = 0 then prProfile['Filter'].SetData(0);
  if prProfile.Find('Freq').DataCount = 0 then prProfile['Freq'].SetData(1000.0/1.0);
  if prProfile.Find('Resample').DataCount = 0 then prProfile['Resample'].SetData(1);
  if prProfile.Find('ResampleStd').DataCount = 0 then prProfile['ResampleStd'].SetData(0.001/1.0);
  if prProfile.Find('Winverse').DataCount = 0 then prProfile['Winverse'].SetData(0);
  prSelFromPct         := prProfile['SelFrom'].Data[0];
  prSelToPct           := prProfile['SelTo'].Data[0];
  prFilter             := Boolean(prProfile['Filter'].Data[0]);
  prFreq               := prProfile['Freq'].Data[0];
  prResample           := Boolean(prProfile['Resample'].Data[0]);
  prResampleStd        := prProfile['ResampleStd'].Data[0];
  prXName              := prProfile.Find('XName').DataAsString;
  prYName              := prProfile.Find('YName').DataAsString;
  prWName              := prProfile.Find('WName').DataAsString;
  prWinverse           := Boolean(prProfile['Winverse'].Data[0]);

  prAbsX               := False;
  prDisplayStyle       := QUBFITS_DISPLAY_DEFAULT;

  prDataSet            := TQUB_FitDataset.Create;
  prUseDataSet         := TQUB_FitDataset.Create;
end;

destructor TQUB_FitData.Destroy;
begin
  try
    prDataSet.Free;
    prUseDataSet.Free;
    FreeMem(prFY);
  finally
    inherited;
  end;
end;

procedure TQUB_FitData.UseData(aDataSet :TQUB_FitDataSet);
begin
  if aDataSet <> nil then begin
    prDataSet.Assign(aDataSet);
  end else begin
    prDataSet.Realloc(0);
  end;
  XName := XName;
  YName := YName;
  UseWeights(WName, Winverse);
  prDataSet.Sort;
  prReSubData();
  if Assigned(OnVarsChange) then OnVarsChange(Self);
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prReSubData;
var
  A, B, i, j :Integer;
  X, W       :Double;
  lo, hi     :Double;
  Start      :Double;
  ResampleCount :Integer;
  Intervals     :TQUB_PAAdResInterval_dbl;
  UseSamples    :TPAInteger;
  SelCount      :Integer;
  Resample      :Boolean;
  FilterDec     :Integer;
  Sampling      :Double;
  ProcData      :TQUB_FitDataSet;
begin
  UseDataSet.Assign(DataSet);
  if DataSet.N = 0 then begin
    Exit;
  end;

  prXmin     := DataSet.XX[0];
  prXmax     := DataSet.XX[DataSet.N - 1];
  Sampling   := -1; // to shut up compiler; only set&used if canfilter

  // filter only uniformly-sampled data
  prCanFilter := DataSet.N >= 10;
  if prCanFilter then begin
    lo       := DataSet.XX[1] - DataSet.XX[0];
    hi       := lo;
    for i := 2 to DataSet.N - 1 do begin
      x      := DataSet.XX[i] - DataSet.XX[i-1];
      lo     := Min(x, lo);
      hi     := Max(x, hi);
    end;
    prCanFilter := Abs(lo - hi) < 1.0e-8;
    Sampling := lo;

    // min freq fits 4 periods in data
    prFreq   := Max(prFreq, 4.0/(prXmax - prXmin));
  end;

  ProcData   := TQUB_FitDataSet.Create;
  ProcData.Assign(DataSet);
  SelCount   := DataSet.N;

  if CanFilter and Filter then begin
    ProcData.Realloc(DataSet.N);
    for j := 0 to ProcData.VarCount - 1 do begin
      if (j = ProcData.Xindex)
      or (0 <> FilterDataD(DataSet.Vals[j], ProcData.Vals[j], DataSet.N, Sampling, Freq))
      then Move(DataSet.Vals[j][0], ProcData.Vals[j][0], DataSet.N*SizeOf(Double));
      // filter each non-x variable, copying raw if there's an error or it's xx
    end;
    if 0 <> FilterDataD(DataSet.WW, ProcData.WW, DataSet.N, Sampling, Freq)
    then Move(DataSet.WW[0], ProcData.WW[0], DataSet.N*SizeOf(Double));
    // decimate by largest D: (2*Freq) < 1/(Sampling*D)
    //   Samp*D < 1/(2*Freq)
    //   D < 1/(2*Freq*Samp); D = floor(1/(2*Freq*Samp))
    FilterDec := Max(1, Floor(1.0/(2*Freq*Sampling)));
    SelCount  := DataSet.N div FilterDec;
    for i := 0 to SelCount - 1 do begin
      for j := 0 to ProcData.VarCount - 1 do begin
        ProcData.Vals[j][i] := ProcData.Vals[j][i*FilterDec];
      end;
      ProcData.WW[i] := ProcData.WW[i*FilterDec];
    end;
    // take the last point if it's not there
    if ((SelCount-1) * FilterDec) < (DataSet.N - 1) then begin
      for j := 0 to ProcData.VarCount - 1 do begin
        ProcData.Vals[j][SelCount] := ProcData.Vals[j][DataSet.N-1];
      end;
      ProcData.WW[SelCount] := ProcData.WW[DataSet.N-1];
      Inc(SelCount);
    end;
  end;

  // measure data
  A          := 0;
  X          := prXmin + (prSelFromPct/100) * (prXmax - prXmin);
  while (A < SelCount) and (ProcData.XX[A] < X) do Inc(A);
  B          := SelCount - 1;
  X          := prXmin + (prSelToPct/100) * (prXmax - prXmin);
  while (B >= 0) and (ProcData.XX[B] > X) do Dec(B);
  SelCount   := B - A + 1;
  if AbsX then begin
    Start    := 0.0;
  end else begin
    Start    := SelFromS;
  end;
  UseDataSet.Xoff := Start;

  if SelCount = 0 then begin
    UseDataSet.Realloc(0);
    Exit;
  end;

  Resample := prResample;
  if Resample then begin
    Intervals  := nil;
    UseSamples := nil;
    try
      AdaptiveResampleClosest_dbl(@(ProcData.YY[A]), SelCount, prResampleStd,
                   {outputs:} ResampleCount, Intervals, UseSamples);
      UseDataSet.Realloc(ResampleCount);
      for i := 0 to ResampleCount - 1 do begin
        for j := 0 to ProcData.VarCount - 1 do begin
          UseDataSet.Vals[j][i] := ProcData.Vals[j][A+UseSamples[i]];
        end;
        UseDataSet.XX[i] := ProcData.XX[A+UseSamples[i]] - Start;
        // fill WW[i] with mean of WW[j] in interval
        W     := 0.0;
        for j := Intervals[i].First to Intervals[i].Last do begin
          W   := W + ProcData.WW[j];
        end;
        UseDataSet.WW[i] := W / (Intervals[i].Last - Intervals[i].First + 1);
      end;
    except
      Resample := False;
    end;
    FreeMem(Intervals);
    FreeMem(UseSamples);
  end;
  if not Resample then begin
    UseDataSet.Realloc(SelCount);
    for i := 0 to ProcData.VarCount - 1 do begin
      if i <> ProcData.Xindex then begin
        Move(ProcData.Vals[i][A], UseDataSet.Vals[i][0], SelCount*SizeOf(Double));
      end;
    end;
    for i := 0 to SelCount - 1 do begin
      UseDataSet.XX[i] := ProcData.XX[A+i] - Start;
    end;
    Move(ProcData.WW[0], UseDataSet.WW[0], SelCount*SizeOf(Double));
  end;
  UseDataSet.ReMeasureW;
  // now re-alloc fit curve to be re-gen by QUB_Fits handler
  ReAllocMem(prFY, UseDataSet.N * SizeOf(Double));
  ProcData.Free;
end;

procedure TQUB_FitData.prOnPreSet(Sender :TQUB_ProfileCat; Updates :INode);
var
  Nm    :String;
  Inv   :Boolean;
begin
  if Updates.Find('SelFrom').DataCount > 0 then begin
    prSelFromPct         := Max(0.0, Double(Updates['SelFrom'].Data[0]));
    prProfile['SelFrom'].SetData(prSelFromPct);
  end;
  if Updates.Find('SelTo').DataCount > 0 then begin
    prSelToPct           := Min(100.0, Double(Updates['SelTo'].Data[0]));
    prProfile['SelTo'].SetData(prSelToPct);
  end;
  if Updates.Find('Filter').DataCount > 0 then begin
    prFilter             := Boolean(Updates['Filter'].Data[0]);
    prProfile['Filter'].SetData(Integer(prFilter));
  end;
  if Updates.Find('Freq').DataCount > 0 then begin
    prFreq               := Updates['Freq'].Data[0];
    prProfile['Freq'].SetData(prFreq);
  end;
  if Updates.Find('Resample').DataCount > 0 then begin
    prResample           := Boolean(Updates['Resample'].Data[0]);
    prProfile['Resample'].SetData(Integer(prResample));
  end;
  if Updates.Find('ResampleStd').DataCount > 0 then begin
    prResampleStd        := Updates['ResampleStd'].Data[0];
    prProfile['ResampleStd'].SetData(prResampleStd);
  end;
  if Updates.Find('XName').DataCount > 0 then begin
    XName                := Updates['XName'].DataAsString;
  end;
  if Updates.Find('YName').DataCount > 0 then begin
    YName                := Updates['YName'].DataAsString;
  end;
  Nm                     := WName;
  if Updates.Find('WName').DataCount > 0 then begin
    Nm                   := Updates['WName'].DataAsString;
  end;
  Inv                    := Winverse;
  if Updates.Find('Winverse').DataCount > 0 then begin
    Inv                  := Boolean(Updates['Winverse'].Data[0]);
  end;
  if (Nm <> WName) or (Inv <> Winverse) then UseWeights(Nm, Inv);

  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetXName(x :String);
var
  i     :Integer;
begin
  i := DataSet.IndexOf(x);
  if (i >= 0) and (i <> DataSet.Xindex) then begin
    DataSet.Xindex := i;
    prXName := x;
    prProfile['XName'].SetData(x);
    DataSet.Sort;
    prReSubData();
    if Assigned(OnChange) then OnChange(Self);
  end else if DataSet.Xindex >= 0 then begin
    prXName := DataSet.VarName[DataSet.Xindex];
  end;
end;

procedure TQUB_FitData.prSetYName(x :String);
var
  i     :Integer;
begin
  i := DataSet.IndexOf(x);
  if (i >= 0) and (i <> DataSet.Yindex) then begin
    DataSet.Yindex := i;
    prYname := x;
    prProfile['YName'].SetData(x);
    prReSubData();
    if Assigned(OnChange) then OnChange(Self);
  end else if DataSet.Yindex >= 0 then begin
    prYName := DataSet.VarName[DataSet.Yindex];
  end;
end;

procedure TQUB_FitData.UseWeights(Name :String; Inverse :Boolean);
var
  i      :Integer;
begin
  if Name <> '' then begin
    i := DataSet.IndexOf(Name);
    if i >= 0 then begin
      DataSet.Windex   := i;
      DataSet.Winverse := Inverse;
      prWName          := Name;
      prWinverse       := Inverse;
      prProfile['WName'].SetData(Name);
      prProfile['Winverse'].SetData(Integer(Inverse));
      prReSubData();
      if Assigned(OnChange) then OnChange(Self);
    end;
  end else begin
    DataSet.Windex   := -1;
    DataSet.Winverse := Inverse;
    prWName          := Name;
    prWinverse       := Inverse;
    prProfile['WName'].SetData(Name);
    prProfile['Winverse'].SetData(Integer(Inverse));
    prReSubData();
    if Assigned(OnChange) then OnChange(Self);
  end;
end;

procedure TQUB_FitData.prSetAbsX(x :Boolean);
begin
  prAbsX := x;
  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetDisplayStyle(x :TQUB_FitData_DisplayStyle);
begin
  prDisplayStyle := x;
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetSelFrom(x :Double);
begin
  if x >= prSelToPct then Exit;
  prSelFromPct       := Max(0.0, x);
  prProfile['SelFrom'].SetData(x);
  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetSelTo(x :Double);
begin
  if x <= prSelFromPct then Exit;
  prSelToPct       := Min(100.0, x);
  prProfile['SelTo'].SetData(x);
  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetSelFromS(x :Double);
begin
  if prXmax <> prXmin
  then SelFromPct  := 100.0 * (x - prXmin) / (prXmax - prXmin);
end;

procedure TQUB_FitData.prSetSelToS(x :Double);
begin
  if prXmax <> prXmin
  then SelToPct    := 100.0 * (x - prXmin) / (prXmax - prXmin);
end;

function  TQUB_FitData.prGetSelFromS: Double;
begin
  if prXmax <> prXmin
  then Result := prXmin + (prSelFromPct / 100.0) * (prXmax - prXmin)
  else Result := prXmin;
end;

function  TQUB_FitData.prGetSelToS: Double;
begin
  if prXmax <> prXmin
  then Result := prXmin + (prSelToPct / 100.0) * (prXmax - prXmin)
  else Result := prXmax;
end;

procedure TQUB_FitData.prSetFilter(x :Boolean);
begin
  prFilter := x;
  prProfile['Filter'].SetData(Integer(x));
  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetFreq(x :Double);
begin
  prFreq := x;
  prProfile['Freq'].SetData(x);
  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetResample(x :Boolean);
begin
  prResample := x;
  prProfile['Resample'].SetData(Integer(x));
  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;

procedure TQUB_FitData.prSetResampleStd(x :Double);
begin
  prResampleStd := x;
  prProfile['ResampleStd'].SetData(x);
  prReSubData();
  if Assigned(OnChange) then OnChange(Self);
end;


// *****************************************************************************


constructor TQUB_Fits.Create;
var
  PCurves             :TQUB_QFS_PNodeMem;
  Curves, Crv         :INode;
  i                   :Integer;
begin
  Events              := TQUB_Events.Create;

  prData              := TQUB_FitData.Create;
  prData.OnVarsChange := prOnVarsChange;
  prData.OnChange     := prOnDataChange;

  prCurveRestore      := ProfileMgr['Curve Restore'].Active;
  prCurves            := TStringList.Create;
  PCurves             := ListCurves();
  Curves              := TNode.CreateUsing(PCurves);
  Node_DECREF(PCurves);
  Crv                 := Curves.Find('Curve');
  while not Crv.IsNull do begin
    prCurves.Add( Crv.DataAsString );
    Crv               := Crv.SiblingSameName;
  end;

  prCurveActive       := ProfileMgr['Curve'].Active;
  ProfileMgr['Curve'].OnSet := prOnCurvePreSet;
  i                   := Max(0, prCurves.IndexOf(prCurveActive['Name'].DataAsString));
  Crv                 := prCurveActive;
  Curve               := CreateCurve(PChar(prCurves[i]));
  Curve.UpdateState(Crv.GetNodeMem());

  prFitterRestore     := ProfileMgr['Fitter Restore'].Active;
  prFitters           := TStringList.Create;
  PCurves             := ListFitters();
  Curves              := TNode.CreateUsing(PCurves);
  Node_DECREF(PCurves);
  Crv                 := Curves.Find('Fitter');
  while not Crv.IsNull do begin
    prFitters.Add( Crv.DataAsString );
    Crv               := Crv.SiblingSameName;
  end;

  prFitterActive      := ProfileMgr['Fitter'].Active;
  ProfileMgr['Fitter'].OnSet := prOnFitterPreSet;
  i                   := Max(0, prFitters.IndexOf(prFitterActive['Name'].DataAsString));
  Crv                 := prFitterActive;
  Fitter              := CreateFitter(PChar(prFitters[i]));
  Fitter.UpdateState(Crv.GetNodeMem());
end;

destructor TQUB_Fits.Destroy;
begin
  try
    Fitter := nil;
    Curve  := nil;
    Data.Free;
    prCurves.Free;
    prFitters.Free;
    Events.Free;
  finally
    inherited;
  end;
end;

function TQUB_Fits.prGetCurveCount: Integer;
begin
  Result := prCurves.Count;
end;

function TQUB_Fits.prGetCurveName(Ix :Integer): String;
begin
  Result := prCurves[Ix];
end;

function TQUB_Fits.prGetFitterCount: Integer;
begin
  Result := prFitters.Count;
end;

function TQUB_Fits.prGetFitterName(Ix :Integer): String;
begin
  Result := prFitters[Ix];
end;

procedure TQUB_Fits.prSetCurve(x :TQUB_Curve);
var
  Crv         :INode;
begin
  if prCurve <> nil then begin
    Crv       := prCurveReStore.Find(prCurve.GetName());
    if not Crv.IsNull then prCurveReStore.RemoveChild(Crv);
    Crv       := TNode.CreateUsing(prCurve.GetState());
    Node_DECREF(Crv.GetNodeMem);
    Crv       := Crv.Clone(True);
    Crv.Name  := Crv['Name'].DataAsString;
    prCurveReStore.InsertChild(TNode.CreateNull(), Crv);

    prCurve.Events.Del('Curve CompCount', Self);
    prCurve.Events.Del('Curve Eqn', Self);
    prCurve.Events.Del('Curve Params', Self);
    prCurve.Events.Del('Curve ParamVal', Self);
    prCurve.Events.Del('Curve ParamConst', Self);
    prCurve.Events.Del('Curve Msg', Self);
    prCurve.Events.Del('Curve Err', Self);

    prCurve.Free;
  end;

  prCurve         := x;

  if prCurve <> nil then begin
    prCurve.SetExtraVars(Data.UseDataSet.Data.VName, Data.UseDataSet.Data.NV);

    prCurveActive   := TNode.CreateUsing(prCurve.GetState());
    Node_DECREF(prCurveActive.GetNodeMem());
    ProfileMgr['Curve'].Active := prCurveActive;

    prCurve.Events.Add('Curve CompCount', Self);
    prCurve.Events.Add('Curve Eqn', Self);
    prCurve.Events.Add('Curve Params', Self);
    prCurve.Events.Add('Curve ParamVal', Self);
    prCurve.Events.Add('Curve ParamConst', Self);
    prCurve.Events.Add('Curve Msg', Self);
    prCurve.Events.Add('Curve Err', Self);

    Crv                   := prCurveReStore.Find(prCurve.GetName());
    if not Crv.IsNull then begin
      prCurve.UpdateState(Crv.GetNodeMem());
    end;
  end;

  prReCalc();
  Events.RaiseNotifyEvent('OnCurveChange', Self);
end;

procedure TQUB_Fits.prSetFitter(x :TQUB_Fitter);
var
  Ftr         :INode;
begin
  if prFitter <> nil then begin
    Ftr       := prFitterReStore.Find(prFitter.GetName());
    if not Ftr.IsNull then prFitterReStore.RemoveChild(Ftr);
    Ftr       := TNode.CreateUsing(prFitter.GetState());
    Node_DECREF(Ftr.GetNodeMem);
    Ftr       := Ftr.Clone(True);
    Ftr.Name  := Ftr['Name'].DataAsString;
    prFitterReStore.InsertChild(TNode.CreateNull(), Ftr);

    prFitter.Events.Del('Fitter ParamVal', Self);
    prFitter.Events.Del('Fitter Iter', Self);
    prFitter.Events.Del('Fitter Msg', Self);
    prFitter.Events.Del('Fitter Err', Self);

    prFitter.Free;
  end;

  prFitter         := x;

  if prFitter <> nil then begin
    prFitterActive   := TNode.CreateUsing(prFitter.GetState());
    Node_DECREF(prFitterActive.GetNodeMem());
    ProfileMgr['Fitter'].Active := prFitterActive;

    prFitter.Events.Add('Fitter ParamVal', Self);
    prFitter.Events.Add('Fitter Iter', Self);
    prFitter.Events.Add('Fitter Msg', Self);
    prFitter.Events.Add('Fitter Err', Self);

    Ftr                   := prFitterReStore.Find(prFitter.GetName());
    if not Ftr.IsNull then begin
      prFitter.UpdateState(Ftr.GetNodeMem());
    end;
  end;

  Events.RaiseNotifyEvent('OnFitterChange', Self);
end;

procedure TQUB_Fits.Fit;
var
  i    :Integer;
  PData :TPQF_Data;
begin
  try
    prLastIter := 0;
    prFitting  := True;
    PData := @(Data.UseDataSet.Data);
    Fitter.Fit(Curve, TPAQF_Data(@PData), 1);
    prFitting  := False;

    for i := 0 to Length(Errors) - 1 do begin
      Errors[i] := Fitter.GetParamErr(i);
    end;
    Events.RaiseNotifyEvent('OnErrorsChange', Self);
  except
    // ?
  end;
  Events.RaiseNotifyEvent('OnFitterDone', Self);
end;

procedure TQUB_Fits.prOnCurvePreSet(Sender :TQUB_ProfileCat; Updates :INode);
var
  PresetName       :String;
begin
  PresetName       := Updates['Name'].DataAsString;
  if (PresetName <> '') and ((prCurve = nil) or (prCurve.GetName() <> PresetName)) then begin
    Curve := CreateCurve(PChar(PresetName));
  end;
  prCurve.UpdateState(Updates.GetNodeMem());
end;

procedure TQUB_Fits.prOnFitterPreSet(Sender :TQUB_ProfileCat; Updates :INode);
var
  PresetName       :String;
begin
  PresetName       := Updates['Name'].DataAsString;
  if (PresetName <> '') and ((prFitter = nil) or (prFitter.GetName() <> PresetName)) then begin
    Fitter := CreateFitter(PChar(PresetName));
  end;
  prFitter.UpdateState(Updates.GetNodeMem());
end;

procedure TQUB_Fits.prReCalc;
var
  i               :Integer;
  x, xs           :Double;
begin
  if Curve = nil then Exit;

  SetLength(Errors, Curve.GetParamCount());

  if Data.UseDataSet.N > 0 then begin
    try
      Curve.EvalArrayAll(Data.UseDataSet.Data, Data.FY);

      for i := 0 to Length(Errors) - 1 do begin
        try
          Errors[i] := ChiSqrUncertainty(Curve, Data.UseDataSet.Data, Data.FY, i, FITDATA_CHISQR_DELTAPCT);
        except
          Errors[i] := -1;
        end;
      end;

      xs    := 0.0;
      for i := 0 to Data.UseDataSet.N - 1 do begin
        x   := Data.FY[i] - Data.UseDataSet.YY[i];
        xs  := xs + (x * x);
      end;
      Residual := xs;
    except
      Residual := -1.0;
    end;
  end;
end;

procedure TQUB_Fits.prOnVarsChange(Sender :TObject);
begin
  Curve.SetExtraVars(Data.UseDataSet.Data.VName, Data.UseDataSet.Data.NV);
  Events.RaiseNotifyEvent('OnVarsChange', Self);
end;

procedure TQUB_Fits.prOnDataChange(Sender :TObject);
begin
  prReCalc();
  Events.RaiseNotifyEvent('OnDataChange', Self);
end;

procedure TQUB_Fits.QUBEvt(Name :PChar; Sender :Pointer); cdecl;
var
  NameS              :String;
begin
  NameS              := Name;
  if NameS = 'Curve CompCount' then begin
    Events.RaiseNotifyEvent('OnCurveCompCount', Self);
  end else if NameS = 'Curve Eqn' then begin
    Events.RaiseNotifyEvent('OnCurveEqn', Self);
  end else if NameS = 'Curve Params' then begin
    prReCalc();
    Events.RaiseNotifyEvent('OnCurveParams', Self);
  end else if NameS = 'Curve ParamVal' then begin
    if not prFitting then prReCalc();
    Events.RaiseNotifyEvent('OnCurveParamVal', Self);
  end else if NameS = 'Curve ParamConst' then begin
    Events.RaiseNotifyEvent('OnCurveParamConst', Self);
  end else if NameS = 'Curve Msg' then begin
    Events.RaiseNotifyEvent('OnCurveMsg', Self);
  end else if NameS = 'Curve Err' then begin
    Events.RaiseNotifyEvent('OnCurveErr', Self);
  end else if NameS = 'Fitter ParamVal' then begin
    Events.RaiseNotifyEvent('OnFitterParamVal', Self);
  end else if NameS = 'Fitter Iter' then begin
    if prLastIter <> Fitter.GetIterations() then begin
      prLastIter := Fitter.GetIterations();
      prReCalc();
      Events.RaiseNotifyEvent('OnFitterIter', Self);
    end;
  end else if NameS = 'Fitter Msg' then begin
    Events.RaiseNotifyEvent('OnFitterMsg', Self);
  end else if NameS = 'Fitter Err' then begin
    Events.RaiseNotifyEvent('OnFitterErr', Self);
  end;
end;

function TQUB_Fits.QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
begin
  Result := 0;
end;

function TQUB_Fits._AddRef: Integer; stdcall;
begin
  Result := -1;
end;

function TQUB_Fits._Release: Integer; stdcall;
begin
  Result := -1;
end;




// *****************************************************************************
end.
