unit QUBFits_Dialog;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Math,
  Dialogs,

  QUB_Node,
  QUB_Profiles,
  QUBFits,
  QUBFits_GUI;

type
  TQUBFits_Dlog = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormHide(Sender: TObject);
  private
    prLayout   :INode;
    procedure prOnResize(Sender :TObject);
    procedure prOnPreSet(Sender :TQUB_ProfileCat; Updates :INode);

  public
    panFitter   :TQUB_DialACurve;
    Fitter      :TQUB_Fits;
    Data        :TQUB_FitData;

    function GetState: INode;
    procedure SetState(State :INode);

  end;


function FitDlog: TQUBFits_Dlog;




implementation

{$R *.dfm}

uses

QUB_GUIInterface;

procedure TQUBFits_Dlog.FormCreate(Sender: TObject);
var
  R        :INode;
begin
  prLayout := ProfileMgr['QUBFits Dialog'].Active;
  ProfileMgr['QUBFits Dialog'].OnSet := prOnPreSet;
  R        := prLayout['Rect'];
  if R.DataCount < 4 then begin
    R.SetupNumData(NODE_DATATYPE_LONGINT, 4, 1);
    R.Data[0] := Left;
    R.Data[1] := Top;
    R.Data[2] := Width;
    R.Data[3] := Height;
  end;

  SetBounds(R.Data[0], R.Data[1], R.Data[2], R.Data[3]);
  OnResize := prOnResize;

  panFitter := TQUB_DialACurve.Create(Self);
  panFitter.Parent := Self;
  panFitter.Align := alClient;
  panFitter.Init;
  Fitter := panFitter.Fits;
  Data   := Fitter.Data;
end;

function TQUBFits_Dlog.GetState: INode;
begin
  Result := panFitter.GetState;
  Result.AppendClone(prLayout, True).Name := 'QUBFits Dialog';
end;

procedure TQUBFits_Dlog.SetState(State :INode);
begin
  ProfileMgr['QUBFits Dialog'].SetProperties(State.Find('QUBFits Dialog'));
  panFitter.SetState(State);
end;

procedure TQUBFits_Dlog.FormHide(Sender: TObject);
begin
  prOnResize(Self);
end;

procedure TQUBFits_Dlog.prOnResize(Sender :TObject);
var
  R         :INode;
begin
  R         := prLayout['Rect'];
  R.Data[0] := Left;
  R.Data[1] := Top;
  R.Data[2] := Width;
  R.Data[3] := Height;
end;

procedure TQUBFits_Dlog.prOnPreSet(Sender :TQUB_ProfileCat; Updates :INode);
var
  R        :INode;
begin
  R        := Updates.Find('Rect');
  if R.DataCount = 4 then begin
    SetBounds(R.Data[0], R.Data[1], R.Data[2], R.Data[3]);
    Move(R.DataPtr^, prLayout['Rect'].DataPtr^, R.RowSize*R.Rows);
  end;
end;


var

  QUBFits_Dlog   :TQUBFits_Dlog;


function FitDlog: TQUBFits_Dlog;
begin
  if QUBFits_Dlog = nil then begin
    QUBFits_Dlog := TQUBFits_Dlog.Create(nil);
  end;
  Result := QUBFits_Dlog;
end;



end.
