unit QUBFits_GUI;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

// "fit all" and "keep it" buttons are present but do nothing
// drawing is naive and hopefully not too slow

// elsewhere:
//    "tag" new data files
//    folder-reading, date/tag refining, info-showing, leak-subtracting, fit-keeping
//    make plugins for other fitters and curves, especially the diffusion ODEs


interface

uses

Windows, Classes, SysUtils, Forms, StdCtrls, ComCtrls, ExtCtrls, Controls,
Graphics, Dialogs, Math, Buttons, ShellAPI,

Alloc, AllocTypes,
QUB_Node,
QUB_Profiles,
QFC_Log,
QUBFits;


const

  CURVESLIDER_STEPS       = 30;
  CURVESLIDER_OCTAVES     = 3;
  CURVESLIDER_WIDTH       = 60;
  CURVESLIDR2_STEPS       = 30;
  CURVESLIDR2_FRACTION    = 48;
  CURVESLIDR2_WIDTH       = 25;
  CURVESLIDER_FORMAT      = '%.5f';
  CURVERESID_FORMAT       = '%.5g';
  FITTERPARAMS_MID        = 80;
  FITTERPARAMS_FORMAT     = '%.6g';
  FITTER_TEXT_HEIGHT      = 50;
  FITDATA_PCT_FORMAT      = '%.2f';
  FITDATA_S_FORMAT        = '%.3f';
  FITDATA_STD_FORMAT      = '%.3g';
  FITDATA_ERR_FORMAT      = '%.3g';
  FITDATA_FREQ_FORMAT     = '%.2f';
  FITDATA_HIRES_EXPAND    = 0.15;
  FITDATA_Y_EXPAND        = 0.1;
  FITDATA_Y_HIST_EXPAND   = 0.02;
  FITDATA_LORES_HEIGHT    = 50;
  FITDATA_CONTROL_HEIGHT  = 45;
  FITDATA_TXT_WIDTH       = 50;
  FITDATA_MNU_WIDTH       = 65;
  FITDATA_POINTS_THRESH   = 32;
  FITDATA_HIST_WIDTH      = 80.0;


type

TQUB_CurveParams = class(TScrollBox)
protected
  prFits            :TQUB_Fits;
  prLabels          :array of TCheckbox;
  prSliders         :array of TTrackBar;
  prSlidr2s         :array of TTrackBar;
  prValues          :array of TEdit;
  prErrors          :array of TEdit;
  prCenters         :array of Double;
  lblPlusMinus      :TLabel;

  prSliding         :Boolean;
  prSliding2        :Boolean;
  prFitting         :Boolean;

  procedure prSetFits(x :TQUB_Fits);

  procedure prOnCheckClick(Sender :TObject);
  procedure prOnKeyPress(Sender :TObject; var Key :Char);
  procedure prOnValueExit(Sender :TObject);
  procedure prOnSliderChange(Sender :TObject);
  procedure prOnSlidr2Change(Sender :TObject);
  procedure prOnSliderExit(Sender :TObject);
  procedure prOnSlidr2Exit(Sender :TObject);
  //procedure prOnSliderMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

  procedure prOnCurveChange(Sender :TObject);
  procedure prOnCurveParams(Sender :TObject);
  procedure prOnCurveParamVal(Sender :TObject);
  procedure prOnCurveParamConst(Sender :TObject);
  procedure prOnDataChange(Sender :TObject);
  procedure prOnFitterIter(Sender :TObject);
  procedure prDoCurveParamVals;

public
  destructor Destroy; override;

  property Fits     :TQUB_Fits     read prFits    write prSetFits;
  property Fitting  :Boolean       read prFitting write prFitting;
end;


TQUB_CurvePanel = class(TPanel)
protected
  prFits          :TQUB_Fits;

  mnuCurve        :TComboBox;
  txtComponents   :TEdit;
  udComponents    :TUpDown;
  txtEqn          :TEdit;
  panMsg          :TPanel;
  panParams       :TQUB_CurveParams;
  txtResidual     :TEdit;
  mnuPresets      :TQUB_ProfileEditBox;
  prUrl           :String;
  prChangingCompCount :Boolean;
  prFitting       :Boolean;

  procedure prSetFits(x :TQUB_Fits);

  procedure prOnChooseCurve(Sender :TObject);
  procedure prOnComponentsChangingEx(Sender: TObject; var AllowChange: Boolean; NewValue: SmallInt; Direction: TUpDownDirection);
  procedure prOnKeyPress(Sender :TObject; var Key :Char);
  procedure prOnEqnEnter(Sender :TObject);
  procedure prOnEqnExit(Sender :TObject);
  procedure prOnEqnClick(Sender :TObject);

  procedure prOnCurveChange(Sender :TObject);
  procedure prOnCurveCompCount(Sender :TObject);
  procedure prOnCurveEqn(Sender :TObject);
  procedure prOnCurveMsg(Sender :TObject);
  procedure prOnWhatever(Sender :TObject);
  procedure prSetFitting(x :Boolean);

public
  destructor Destroy; override;

  property Fits     :TQUB_Fits     read prFits    write prSetFits;
  property Fitting  :Boolean       read prFitting write prSetFitting;
end;


TQUB_FitterParams = class(TScrollBox)
protected
  prFits          :TQUB_Fits;

  prLabels        :array of TLabel;
  prValues        :array of TEdit;

  procedure prSetFits(x :TQUB_Fits);

  procedure prOnKeyPress(Sender :TObject; var Key :Char);
  procedure prOnValueExit(Sender :TObject);

  procedure prOnFitterChange(Sender :TObject);
  procedure prOnFitterParamVal(Sender :TObject);

public
  destructor Destroy; override;

  property Fits     :TQUB_Fits     read prFits    write prSetFits;
end;


TQUB_FitterPanel = class(TPanel)
protected
  prFits          :TQUB_Fits;

  mnuFitter       :TComboBox;
  panParams       :TQUB_FitterParams;
  btnFit          :TButton;
  btnKeepIt       :TButton;
  btnFitAll       :TButton;

  prOnKeepIt         :TNotifyEvent;
  prOnFitAll         :TNotifyEvent;
  prOnFit            :TNotifyEvent;

  procedure prSetFits(x :TQUB_Fits);
  procedure prSetOnFit(x :TNotifyEvent);
  procedure prSetOnKeepIt(x :TNotifyEvent);
  procedure prSetOnFitAll(x :TNotifyEvent);

  procedure prOnChooseFitter(Sender :TObject);
  procedure prOnFitClick(Sender :TObject);
  procedure prOnKeepItClick(Sender :TObject);
  procedure prOnFitAllClick(Sender :TObject);

  procedure prOnFitterChange(Sender :TObject);

protected

  function FitBtnScreenPos: TPoint;

public

  destructor Destroy; override;

  property Fits     :TQUB_Fits     read prFits     write prSetFits;
  property OnFit    :TNotifyEvent  read prOnFit    write prSetOnFit;
  property OnKeepIt :TNotifyEvent  read prOnKeepIt write prSetOnKeepIt;
  property OnFitAll :TNotifyEvent  read prOnFitAll write prSetOnFitAll;
end;

TQUB_FitData_Control = class(TPanel)
protected
  prFits                :TQUB_Fits;
  prData                :TQUB_FitData;

  lblFrom              :TLabel;
  lblTo                :TLabel;
  lblS                 :TLabel;
  txtFromPct           :TEdit;
  txtFromS             :TEdit;
  txtToPct             :TEdit;
  txtToS               :TEdit;
  lblXName             :TLabel;
  lblYName             :TLabel;
  mnuXName             :TComboBox;
  mnuYName             :TComboBox;
  lblWName             :TLabel;
  mnuWName             :TComboBox;
  chkFilter            :TCheckBox;
  txtFreq              :TEdit;
  lblHz                :TLabel;
  lblResample          :TLabel;
  chkResample          :TCheckBox;
  lblResampleStd       :TLabel;
  txtResampleStd       :TEdit;
  lblN                 :TLabel;
  txtN                 :TEdit;
  mnuPresets           :TQUB_ProfileEditBox;

  procedure prOnResize(Sender :TObject);
  procedure prOnKeyPress(Sender :TObject; var Key :Char);
  procedure prOnExitFromPct(Sender :TObject);
  procedure prOnExitFromS(Sender :TObject);
  procedure prOnExitToPct(Sender :TObject);
  procedure prOnExitToS(Sender :TObject);
  procedure prOnXNameChange(Sender :TObject);
  procedure prOnYNameChange(Sender :TObject);
  procedure prOnWNameChange(Sender :TObject);
  procedure prOnClickFilter(Sender :TObject);
  procedure prOnExitFreq(Sender :TObject);
  procedure prOnClickResample(Sender :TObject);
  procedure prOnExitResampleStd(Sender :TObject);

  procedure prSetFits(x :TQUB_Fits);
  procedure prOnDataChange(Sender :TObject);

public
  destructor Destroy; override;

  property Fits        :TQUB_Fits      read prFits       write prSetFits;
end;


TQUB_FitData_Display = class(TCustomControl)
protected
  prFits                :TQUB_Fits;
  prData                :TQUB_FitData;
  prHiRes               :Boolean;
  prFitting             :Boolean;

  prBitmap              :Graphics.TBitmap;
  prBitmapCanvas        :TCanvas;

  prNeedsDraw           :Boolean;

  prXmin, prXmax, prXm  :Double;
  prYMin, prYMax, prYm  :Double;
  prVmin, prVmax, prVm  :array of Double;
  // (x, y) screen -> (Xmin + x*Xm, Ymin + (h-y)*Ym)

  prSelFrom, prSelTo    :Double;

  prMouseX, prMouseY    :Integer;
  prMouseDownX, prMouseDownY :Integer;
  prCanMoveSF           :Boolean;
  prCanMoveSL           :Boolean;
  prMoving              :Boolean;

  procedure prSetFits(x :TQUB_Fits);
  procedure prSetHiRes(x :Boolean);
  procedure prOnDataChange(Sender :TObject);
  procedure prOnCurveChange(Sender :TObject);
  procedure prOnFitterIter(Sender :TObject);

  procedure prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  procedure prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnDblClick(Sender :TObject);

public
  // draw on double-buf canvas, when data changes, underneath the data
  OnBeforeDraw        :procedure (Sender :TQUB_FitData_Display;
                                  Canvas :TCanvas) of object;
  // draw on screen canvas, whenever so much as the mouse moves
  OnAfterDraw         :procedure (Sender :TQUB_FitData_Display;
                                  Canvas :TCanvas) of object;
  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  procedure Paint; override;
  procedure ForceRedraw;

  function XtoScreen(X :Double): Integer;
  function YtoScreen(Y :Double; Extra :Integer = -1): Integer;
  function ScreenToX(i :Integer): Double;
  function ScreenToY(j :Integer; Extra :Integer = -1): Double;

  property Fits       :TQUB_Fits      read prFits       write prSetFits;
  property Data       :TQUB_FitData   read prData;
  property HiRes      :Boolean        read prHiRes      write prSetHiRes;
  property Fitting    :Boolean        read prFitting    write prFitting;
end;



TQUB_DialACurve = class(TPanel)
protected
  prFits               :TQUB_Fits;

  panDisplays          :TPanel;
  panSettings          :TQUB_FitData_Control;
  panWhole             :TQUB_FitData_Display;
  panDetail            :TQUB_FitData_Display;
  panCurveFitLog       :TPanel;
  panCurveFit          :TPanel;
  panCurve             :TQUB_CurvePanel;
  panFitter            :TQUB_FitterPanel;
  lstLog               :TQFC_Log;
  splTB, splLR, splLog :TSplitter;

  prLayout             :INode;

  procedure prOnFit(Sender :TObject);
  procedure prOnLayoutChange(Sender :TObject);
  procedure prOnPreSet(Sender :TQUB_ProfileCat; Updates :INode);
  procedure prSetFitting(X :Boolean);
  procedure prOnFitterMsg(Sender :TObject);

public
  destructor Destroy; override;
  procedure Init;

  function GetState: INode;
  procedure SetState(State :INode);

  property FitterPanel   :TQUB_FitterPanel       read panFitter;
  // unhandled: FitterPanel.OnFitAll, FitterPanel.OnKeepIt
  property Whole         :TQUB_FitData_Display   read panWhole;
  property Detail        :TQUB_FitData_Display   read panDetail;
  property Log           :TQFC_Log               read lstLog;

  property Fits          :TQUB_Fits              read prFits;
  property Fitting       :Boolean    write prSetFitting;
end;


TQUB_CurveFitButton = class(TBitBtn)
private
  prFits   :TQUB_DialACurve;
  prState  :INode;
  prData   :TQUB_FitDataSet;
  prFitY   :INode;

  procedure prSetData(x :TQUB_FitDataSet);
  procedure prOnResize(Sender :TObject);

public
  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure Capture(panFits :TQUB_DialACurve);
  procedure Restore;

  property Data  :TQUB_FitDataSet  read prData  write prSetData;
  property State :INode            read prState;
end;


implementation

uses

{$ifdef QUB}
QUB_Python_Thread,
QUB_Viewport_Python,
{$endif}
PythonEngine,
QFC_FittingDlog;


procedure TQUB_CurveParams.prSetFits(x :TQUB_Fits);
begin
  BorderStyle              := bsNone;

  prFits := x;
  prFits.Events.AddEvent('OnCurveChange', prOnCurveChange);
  prFits.Events.AddEvent('OnCurveParams', prOnCurveParams);
  prFits.Events.AddEvent('OnCurveParamVal', prOnCurveParamVal);
  prFits.Events.AddEvent('OnCurveParamConst', prOnCurveParamConst);
  prFits.Events.AddEvent('OnDataChange', prOnDataChange);
  prFits.Events.AddEvent('OnErrorsChange', prOnDataChange);
  prFits.Events.AddEvent('OnFitterIter', prOnFitterIter);
  prOnCurveChange(prFits);
end;

destructor TQUB_CurveParams.Destroy;
begin
  try
    prFits.Events.DelEvent('OnCurveChange', prOnCurveChange);
    prFits.Events.DelEvent('OnCurveParams', prOnCurveParams);
    prFits.Events.DelEvent('OnCurveParamVal', prOnCurveParamVal);
    prFits.Events.DelEvent('OnCurveParamConst', prOnCurveParamConst);
    prFits.Events.DelEvent('OnDataChange', prOnDataChange);
    prFits.Events.DelEvent('OnErrorsChange', prOnDataChange);
    prFits.Events.DelEvent('OnFitterIter', prOnFitterIter);
  finally
    inherited;
  end;
end;

procedure TQUB_CurveParams.prOnCurveChange(Sender :TObject);
var
  Curve    :TQUB_Curve;
  Ix       :Integer;
  X0       :Integer;
begin
  Curve    := prFits.Curve;

  for Ix := 0 to Length(prValues) - 1 do begin
    prLabels[Ix].Free;
    prSliders[Ix].Free;
    prSlidr2s[Ix].Free;
    prValues[Ix].Free;
    prErrors[Ix].Free;
  end;

  if Curve <> nil then begin
    Ix := Curve.GetParamCount();
    SetLength(prLabels, Ix);
    SetLength(prSliders, Ix);
    SetLength(prSlidr2s, Ix);
    SetLength(prValues, Ix);
    SetLength(prCenters, Ix);
    SetLength(prErrors, Ix);

    if lblPlusMinus = nil then begin
      lblPlusMinus           := TLabel.Create(Self);
      lblPlusMinus.Parent    := Self;
      lblPlusMinus.Caption   := '+/-';
      lblPlusMinus.Top       := ClientHeight - lblPlusMinus.Height - 5;
      lblPlusMinus.Left      := 0;
      lblPlusMinus.Anchors   := [akLeft, akBottom];
    end;
    X0                       := lblPlusMinus.Width + 2;

    for Ix := 0 to Curve.GetParamCount - 1 do begin
      prLabels[Ix]           := TCheckbox.Create(Self);
      prLabels[Ix].Parent    := Self;
      prLabels[Ix].Top       := 0;
      prLabels[Ix].Left      := X0 + (CURVESLIDER_WIDTH * Ix);
      prLabels[Ix].Width     := CURVESLIDER_WIDTH;
      prLabels[Ix].Caption   := Curve.GetParamName(Ix);
      prLabels[Ix].Tag       := Ix;
      prLabels[Ix].Checked   := not Boolean(Curve.GetParamConst(Ix));
      prLabels[Ix].OnClick   := prOnCheckClick;

      prErrors[Ix]             := TEdit.Create(Self);
      prErrors[Ix].Parent      := Self;
      prErrors[Ix].Top         := ClientHeight - prErrors[Ix].Height;
      prErrors[Ix].Left        := X0 + (CURVESLIDER_WIDTH * Ix);
      prErrors[Ix].Width       := CURVESLIDER_WIDTH;
      prErrors[Ix].Anchors     := [akLeft, akBottom];
      prErrors[Ix].ReadOnly    := True;

      prValues[Ix]           := TEdit.Create(Self);
      prValues[Ix].Parent    := Self;
      prValues[Ix].Top       := prErrors[Ix].Top - prValues[Ix].Height;
      prValues[Ix].Left      := X0 + (CURVESLIDER_WIDTH * Ix);
      prValues[Ix].Width     := CURVESLIDER_WIDTH;
      prValues[Ix].Anchors   := [akLeft, akBottom];
      prValues[Ix].Tag       := Ix;
      prValues[Ix].OnKeyPress := prOnKeyPress;
      prValues[Ix].OnExit    := prOnValueExit;

      prCenters[Ix]          := Curve.GetParamVal(Ix);
      prValues[Ix].Text      := Format(CURVESLIDER_FORMAT, [prCenters[Ix]]);

      prSliders[Ix]          := TTrackBar.Create(Self);
      prSliders[Ix].Orientation := trVertical;
      prSliders[Ix].Parent   := Self;
      prSliders[Ix].Top      := prLabels[Ix].Height;
      prSliders[Ix].Left     := X0 + (CURVESLIDER_WIDTH * Ix);
      prSliders[Ix].Width    := CURVESLIDR2_WIDTH;
      prSliders[Ix].Height   := prValues[Ix].Top - prErrors[Ix].Height;
      prSliders[Ix].Anchors  := [akLeft, akTop, akBottom];
      prSliders[Ix].Min      := - CURVESLIDER_OCTAVES * CURVESLIDER_STEPS;
      prSliders[Ix].Max      := CURVESLIDER_OCTAVES * CURVESLIDER_STEPS;
      prSliders[Ix].Tag      := Ix;
      prSliders[Ix].OnChange := prOnSliderChange;
      prSliders[Ix].OnExit   := prOnSliderExit;

      prSlidr2s[Ix]          := TTrackBar.Create(Self);
      prSlidr2s[Ix].Orientation := trVertical;
      prSlidr2s[Ix].Parent   := Self;
      prSlidr2s[Ix].Top      := prSliders[Ix].Top;
      prSlidr2s[Ix].Left     := prSliders[Ix].Left + prSliders[Ix].Width;
      prSlidr2s[Ix].Width    := CURVESLIDR2_WIDTH;
      prSlidr2s[Ix].Height   := prSliders[Ix].Height;
      prSlidr2s[Ix].Anchors  := [akLeft, akTop, akBottom];
      prSlidr2s[Ix].Min      := - CURVESLIDER_OCTAVES * CURVESLIDER_STEPS;
      prSlidr2s[Ix].Max      := CURVESLIDER_OCTAVES * CURVESLIDER_STEPS;
      prSlidr2s[Ix].Tag      := Ix;
      prSlidr2s[Ix].OnChange := prOnSlidr2Change;
      prSlidr2s[Ix].OnExit   := prOnSlidr2Exit;
      prSlidr2s[Ix].TickStyle:= tsNone;
    end;
  end else begin
    SetLength(prLabels, 0);
    SetLength(prSliders, 0);
    SetLength(prSlidr2s, 0);
    SetLength(prValues, 0);
    SetLength(prCenters, 0);
    SetLength(prErrors, 0);
  end;

  prOnDataChange(Sender);
end;

procedure TQUB_CurveParams.prOnCurveParams(Sender :TObject);
begin
  prOnCurveChange(Sender);
end;

procedure TQUB_CurveParams.prOnCurveParamVal(Sender :TObject);
begin
  if not prFitting then prDoCurveParamVals;
end;

procedure TQUB_CurveParams.prOnCurveParamConst(Sender :TObject);
var
  Ix      :Integer;
begin
  if prFits.Curve.GetParamCount() <> Length(prValues) then Exit;
  for Ix := 0 to prFits.Curve.GetParamCount() - 1 do begin
    prLabels[Ix].Checked := not Boolean(prFits.Curve.GetParamConst(Ix));
  end;
end;

procedure TQUB_CurveParams.prOnFitterIter(Sender :TObject);
begin
  prDoCurveParamVals;
end;

procedure TQUB_CurveParams.prDoCurveParamVals;
var
  Ix      :Integer;
begin
  if prFits.Curve.GetParamCount() <> Length(prValues) then Exit;
  for Ix := 0 to prFits.Curve.GetParamCount() - 1 do begin
    prValues[Ix].Text := Format(CURVESLIDER_FORMAT, [prFits.Curve.GetParamVal(Ix)]);
    if not prSliding then begin
      prSliders[Ix].OnChange := nil;
      prSlidr2s[Ix].OnChange := nil;
      prSliders[Ix].Position := 0;
      prSlidr2s[Ix].Position := 0;
      prCenters[Ix] := prFits.Curve.GetParamVal(Ix);
      prSliders[Ix].OnChange := prOnSliderChange;
      prSlidr2s[Ix].OnChange := prOnSlidr2Change;
    end;
  end;
  prOnDataChange(Fits);
end;

procedure TQUB_CurveParams.prOnDataChange(Sender :TObject);
var
  Ix      :Integer;
begin
  if prFits.Curve.GetParamCount() <> Length(prValues) then Exit;
  for Ix := 0 to prFits.Curve.GetParamCount() - 1 do begin
    prErrors[Ix].Text := Format(FITDATA_ERR_FORMAT, [prFits.Errors[Ix]]);
  end;
end;

procedure TQUB_CurveParams.prOnCheckClick(Sender :TObject);
var
  Ix      :Integer;
begin
  Ix      := TCheckbox(Sender).Tag;
  prFits.Curve.SetParamConst(Ix, Integer(not TCheckbox(Sender).Checked));
  TCheckbox(Sender).Checked := not Boolean(prFits.Curve.GetParamConst(Ix));
end;

procedure TQUB_CurveParams.prOnKeyPress(Sender :TObject; var Key :Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    TEdit(Sender).OnExit(Sender);
  end else begin
    TEdit(Sender).Font.Color := clRed;
  end;
end;

procedure TQUB_CurveParams.prOnValueExit(Sender :TObject);
var
  Curve   :TQUB_Curve;
  PyX     :PPyObject;
  x       :Double;
begin
  Curve   := prFits.Curve;
  x       := Curve.GetParamVal(TEdit(Sender).Tag);
  {$ifdef QUB}
  PythonThread.BeginPython;
  {$endif}
  with GetPythonEngine do begin
    try
      PyX := EvalString( 'float('+TEdit(Sender).Text+')' );
      if not PyFloat_Check(PyX) then begin
        Py_DECREF(PyX);
        raise Exception.Create('Please enter a number.');
      end;
      x   := PyFloat_AsDouble(PyX);
      Py_DECREF( PyX );
    except
      on E: Exception do begin
        try
          x := StrToFloat(TEdit(Sender).Text);
        except
          MessageDlg(E.Message, mtError, [mbOK], 0);
        end;
      end;
    end;
  end;
  Curve.SetParamVal( TEdit(Sender).Tag, x, 1{gen event} );
  {$ifdef QUB}
  PythonThread.EndPython;
  {$endif}

  TEdit(Sender).Text := Format(CURVESLIDER_FORMAT, [Curve.GetParamVal(TEdit(Sender).Tag)]);
  TEdit(Sender).Font.Color := clWindowText;
end;

procedure TQUB_CurveParams.prOnSliderChange(Sender :TObject);
var
  Ix        :Integer;
  Sign      :Integer;
begin
  Ix        := TTrackBar(Sender).Tag;
  if prCenters[Ix] < 0 then Sign := -1 else Sign := 1;
  prSliding := True;
  if prCenters[Ix] <> 0.0 then begin
    prFits.Curve.SetParamVal(Ix, prCenters[Ix] * Math.Power(2.0, - Sign * TTrackBar(Sender).Position / CURVESLIDER_STEPS), 1{gen evt});
  end else begin
    prFits.Curve.SetParamVal(Ix, - TTrackBar(Sender).Position / CURVESLIDER_STEPS, 1{gen evt});
  end;
  prSliding := False;
  prSlidr2s[Ix].Position := 0;
end;

procedure TQUB_CurveParams.prOnSliderExit(Sender :TObject);
var
  Slider :TTrackBar;
begin
  Slider := TTrackBar(Sender);
  Slider.OnChange := nil;
  prSlidr2s[Slider.Tag].OnChange := nil;
  Slider.Position := 0;
  prSlidr2s[Slider.Tag].Position := 0;
  prCenters[Slider.Tag] := prFits.Curve.GetParamVal(Slider.Tag);
  Slider.OnChange := prOnSliderChange;
  prSlidr2s[Slider.Tag].OnChange := prOnSlidr2Change;
end;

procedure TQUB_CurveParams.prOnSlidr2Change(Sender :TObject);
var
  Ix        :Integer;
  Sign      :Integer;
begin
  Ix        := TTrackBar(Sender).Tag;
  if prCenters[Ix] < 0 then Sign := -1 else Sign := 1;
  prSliding := True;
  if prCenters[Ix] <> 0.0 then begin
    prFits.Curve.SetParamVal(Ix, prCenters[Ix] * Math.Power(2.0, - Sign * TTrackBar(Sender).Position / (CURVESLIDER_STEPS * CURVESLIDR2_FRACTION)), 1{gen evt});
  end else begin
    prFits.Curve.SetParamVal(Ix, - TTrackBar(Sender).Position / (CURVESLIDER_STEPS * CURVESLIDR2_FRACTION), 1{gen evt});
  end;
  prSliding := False;
end;

procedure TQUB_CurveParams.prOnSlidr2Exit(Sender :TObject);
var
  Slider :TTrackBar;
begin
  Slider := TTrackBar(Sender);
  Slider.OnChange := nil;
  prSliders[Slider.Tag].OnChange := nil;
  Slider.Position := 0;
  prCenters[Slider.Tag] := prFits.Curve.GetParamVal(Slider.Tag);
  Slider.OnChange := prOnSlidr2Change;
  prSliders[Slider.Tag].OnChange := prOnSliderChange;
end;

{
// no  mouseup handler available
procedure TQUB_CurveParams.prOnSliderMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Slider :TTrackBar;
begin
  Slider := TTrackBar(Sender);
  Slider.OnChange := nil;
  Slider.Position := 0;
  prCenters[Slider.Tag] := prFits.Curve.GetParamVal(Slider.Tag);
  Slider.OnChange := prOnSliderChange;
end;
}


//******************************************************************************

procedure TQUB_CurvePanel.prSetFits(x :TQUB_Fits);
var
  Lbl                 :TLabel;
  i                   :Integer;
begin
  prFits              := x;
  prFits.Events.AddEvent('OnCurveChange', prOnCurveChange);
  prFits.Events.AddEvent('OnCurveCompCount', prOnCurveCompCount);
  prFits.Events.AddEvent('OnCurveEqn', prOnCurveEqn);
  prFits.Events.AddEvent('OnCurveMsg', prOnCurveMsg);
  prFits.Events.AddEvent('OnCurveErr', prOnCurveMsg);
  prFits.Events.AddEvent('OnCurveParams', prOnWhatever);
  prFits.Events.AddEvent('OnCurveParamVal', prOnWhatever);
  prFits.Events.AddEvent('OnDataChange', prOnWhatever);

  Lbl                 := TLabel.Create(Self);
  Lbl.Parent          := Self;
  Lbl.Top             := 0;
  Lbl.Left            := 0;
  Lbl.Anchors         := [akTop, akRight];
  Lbl.Caption         := 'Components:';

  txtComponents        := TEdit.Create(Self);
  txtComponents.Parent := Self;
  txtComponents.Top    := Lbl.Top;
  txtComponents.Left   := Lbl.Width + 10;
  txtComponents.Width  := 30;
  txtComponents.Anchors := [akTop, akRight];
  txtComponents.ReadOnly := True;

  udComponents         := TUpDown.Create(Self);
  udComponents.Parent  := Self;
  udComponents.Top     := txtComponents.Top;
  udComponents.Left    := txtComponents.Left + txtComponents.Width;
  udComponents.Height  := txtComponents.Height;
  udComponents.Anchors := [akTop, akRight];
  udComponents.Associate := txtComponents;
  udComponents.OnChangingEx := prOnComponentsChangingEx;
  udComponents.Position := prFits.Curve.GetComponentCount();

  i                    := ClientWidth - (udComponents.Left + udComponents.Width);
  Lbl.Left             := Lbl.Left + i;
  txtComponents.Left   := txtComponents.Left + i;
  udComponents.Left    := udComponents.Left + i;

  Lbl                 := TLabel.Create(Self);
  Lbl.Parent          := Self;
  Lbl.Top             := 0;
  Lbl.Left            := 0;
  Lbl.Caption         := 'Curve:';

  mnuCurve            := TComboBox.Create(Self);
  mnuCurve.Parent     := Self;
  mnuCurve.Top        := 0;
  mnuCurve.Left       := Lbl.Width + 10;
  mnuCurve.Width      := i - mnuCurve.Left;
  mnuCurve.Anchors    := [akTop, akLeft, akRight];
  mnuCurve.Style      := csDropDownList;
  mnuCurve.DropDownCount := 16;
  mnuCurve.OnChange   := prOnChooseCurve;
  mnuCurve.Anchors    := [akTop, akLeft, akRight];

  for i := 0 to prFits.CurveCount - 1 do begin
    mnuCurve.Items.Add(prFits.CurveName[i]);
  end;
  mnuCurve.ItemIndex  := mnuCurve.Items.IndexOf(prFits.Curve.GetName());

  Lbl                 := TLabel.Create(Self);
  Lbl.Parent          := Self;
  Lbl.Top             := txtComponents.Top + txtComponents.Height;
  Lbl.Left            := 0;
  Lbl.Caption         := 'f(x) = ';

  txtEqn              := TEdit.Create(Self);
  txtEqn.Parent       := Self;
  txtEqn.Top          := Lbl.Top;
  txtEqn.Left         := Lbl.Width + 10;
  txtEqn.Width        := ClientWidth - txtEqn.Left;
  txtEqn.Anchors      := [akTop, akLeft, akRight];
  txtEqn.OnKeyPress   := prOnKeyPress;
  txtEqn.OnEnter      := prOnEqnEnter;
  txtEqn.OnExit       := prOnEqnExit;
  txtEqn.OnClick      := prOnEqnClick;
  txtEqn.ReadOnly     := not Boolean(prFits.Curve.CanSetEqn());

  panMsg              := TPanel.Create(Self);
  panMsg.Top          := txtEqn.Top + txtEqn.Height;
  panMsg.Left         := txtEqn.Left;
  panMsg.Width        := txtEqn.Width;
  panMsg.Anchors      := [akTop, akLeft, akRight];
  panMsg.Parent       := Self;
  panMsg.Visible      := False;
  panMsg.Caption      := '';
  panMsg.Color        := clRed;
  panMsg.Font.Color   := clWhite;
  panMsg.Height       := txtEqn.Height;
  //panMsg.Font.Style   := [fsBold];

  mnuPresets          := TQUB_ProfileEditBox.Create(Self);
  mnuPresets.Parent   := Self;
  mnuPresets.Top      := ClientHeight - mnuPresets.Height;
  mnuPresets.Left     := ClientWidth - mnuPresets.Width;
  mnuPresets.Anchors  := [akRight, akBottom];
  mnuPresets.Cat      := 'Curve';

  Lbl                 := TLabel.Create(Self);
  Lbl.Parent          := Self;
  Lbl.Top             := ClientHeight - txtEqn.Height;
  Lbl.Left            := 0;
  Lbl.Anchors         := [akLeft, akBottom];
  Lbl.Caption         := 'SS Residual:';

  txtResidual         := TEdit.Create(Self);
  txtResidual.Parent  := Self;
  txtResidual.Top     := mnuPresets.Top;
  txtResidual.Left    := Lbl.Width + 10;
  txtResidual.Width   := 75;
  txtResidual.Anchors := [akLeft, akBottom];
  txtResidual.ReadOnly := True;
  txtResidual.Text     := Format(CURVERESID_FORMAT, [prFits.Residual]);

  panParams           := TQUB_CurveParams.Create(Self);
  panParams.Parent    := Self;
  panParams.Top       := txtEqn.Top + txtEqn.Height;
  panParams.Left      := 0;
  panParams.Width     := ClientWidth - 10;
  panParams.Height    := mnuPresets.Top - txtEqn.Top - txtEqn.Height;
  panParams.Anchors   := [akLeft, akRight, akTop, akBottom];
  panParams.Fits      := prFits;

  prOnCurveEqn(Self);
end;

procedure TQUB_CurvePanel.prSetFitting(x :Boolean);
begin
  panParams.Fitting := x;
  prFitting := x;
  prOnWhatever(prFits);
end;

destructor TQUB_CurvePanel.Destroy;
begin
  try
    prFits.Events.DelEvent('OnCurveChange', prOnCurveChange);
    prFits.Events.DelEvent('OnCurveCompCount', prOnCurveCompCount);
    prFits.Events.DelEvent('OnCurveEqn', prOnCurveEqn);
    prFits.Events.DelEvent('OnCurveParams', prOnWhatever);
    prFits.Events.DelEvent('OnCurveParamVal', prOnWhatever);
    prFits.Events.DelEvent('OnDataChange', prOnWhatever);
    prFits.Events.DelEvent('OnCurveMsg', prOnCurveMsg);
    prFits.Events.DelEvent('OnCurveErr', prOnCurveMsg);
  finally
    inherited;
  end;
end;

procedure TQUB_CurvePanel.prOnChooseCurve(Sender :TObject);
begin
  if (prFits.Curve <> nil) and (prFits.Curve.GetName() = mnuCurve.Items[mnuCurve.ItemIndex])
  then Exit;

  prFits.Curve := CreateCurve(PChar(mnuCurve.Items[mnuCurve.ItemIndex]));
end;

procedure TQUB_CurvePanel.prOnComponentsChangingEx(Sender: TObject; var AllowChange: Boolean; NewValue: SmallInt; Direction: TUpDownDirection);
begin
  AllowChange := NewValue > 0;
  if AllowChange then begin
    prChangingCompCount := True;
    prFits.Curve.SetComponentCount(NewValue);
    prChangingCompCount := False;
    AllowChange := prFits.Curve.GetComponentCount() = NewValue;
  end;
end;

procedure TQUB_CurvePanel.prOnKeyPress(Sender :TObject; var Key :Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    TEdit(Sender).OnExit(Sender);
  end else begin
    TEdit(Sender).Font.Color := clRed;

    if TEdit(Sender) = txtEqn then begin
      panParams.Enabled := False;
    end;
  end;
end;

procedure TQUB_CurvePanel.prOnEqnEnter(Sender :TObject);
begin
  // we try disabling only on actual edit above
  // panParams.Enabled := False;
end;

procedure TQUB_CurvePanel.prOnEqnExit(Sender :TObject);
begin
  if TEdit(Sender).Font.Color = clRed then begin
    prFits.Curve.SetEqn(PChar(TEdit(Sender).Text));
  end;
  TEdit(Sender).Font.Color := clWindowText;
  panParams.Enabled := True;
  panParams.SetFocus;
end;

procedure TQUB_CurvePanel.prOnEqnClick(Sender :TObject);
begin
  if prUrl <> ''
  then ShellExecute(Handle, 'open', PChar(prUrl),nil,nil, SW_SHOWNORMAL) ;
end;

procedure TQUB_CurvePanel.prOnCurveChange(Sender :TObject);
begin
  mnuCurve.ItemIndex := mnuCurve.Items.IndexOf(prFits.Curve.GetName());
  txtEqn.ReadOnly     := not Boolean(prFits.Curve.CanSetEqn());
  udComponents.Position := prFits.Curve.GetComponentCount();
  prOnCurveEqn(Sender);
end;

procedure TQUB_CurvePanel.prOnCurveCompCount(Sender :TObject);
begin
  if not prChangingCompCount then begin
    udComponents.Position := prFits.Curve.GetComponentCount();
  end;
end;

procedure TQUB_CurvePanel.prOnCurveEqn(Sender :TObject);
var
  Eqn                 :String;
  Pos                 :Integer;
begin
  Eqn                 := prFits.Curve.GetEqn();
  if 0<>AnsiPos('http://', Eqn) then begin
    Pos               := AnsiPos(' ', Eqn);
    if Pos = 0 then Pos := Length(Eqn)+1;
    prUrl             := Copy(Eqn, 1, Pos-1);
    if Pos <= Length(Eqn) then Eqn := Copy(Eqn, Pos+1, Length(Eqn));
    txtEqn.Font.Color := clBlue;
    txtEqn.Font.Style := [fsUnderline];
  end else begin
    txtEqn.Font.Color := clBlack;
    txtEqn.Font.Style := [];
    prUrl             := '';
  end;
  txtEqn.Text         := Eqn;
  prOnCurveMsg(nil);
end;

procedure TQUB_CurvePanel.prOnCurveMsg(Sender :TObject);
begin
  if Sender = nil then begin
    if panMsg.Visible then begin
      panMsg.Visible    := False;
      panParams.Top     := panMsg.Top;
      panParams.Height  := panParams.Height + panMsg.Height;
    end;
  end else begin
    panMsg.Caption    := prFits.Curve.GetMessage();
    panMsg.Visible    := True;
    panParams.Top     := panMsg.Top + panMsg.Height;
    panParams.Height  := Max(0, panParams.Height - panMsg.Height);
  end;
end;

procedure TQUB_CurvePanel.prOnWhatever(Sender :TObject);
begin
  if not prFitting then txtResidual.Text    := Format(CURVERESID_FORMAT, [prFits.Residual]);
  //prOnCurveMsg(nil);
end;


//******************************************************************************

procedure TQUB_FitterParams.prSetFits(x :TQUB_Fits);
begin
  BorderStyle              := bsNone;
  AutoScroll               := True;

  prFits := x;
  prFits.Events.AddEvent('OnFitterChange', prOnFitterChange);
  prFits.Events.AddEvent('OnFitterParamVal', prOnFitterParamVal);
  prOnFitterChange(prFits);
end;

destructor TQUB_FitterParams.Destroy;
begin
  try
    prFits.Events.DelEvent('OnFitterChange', prOnFitterChange);
    prFits.Events.DelEvent('OnFitterParamVal', prOnFitterParamVal);
  finally
    inherited;
  end;
end;

procedure TQUB_FitterParams.prOnFitterChange(Sender :TObject);
var
  Fitter :TQUB_Fitter;
  Ix     :Integer;
begin
  for Ix := 0 to Length(prValues) - 1 do begin
    prLabels[Ix].Free;
    prValues[Ix].Free;
  end;

  Fitter := prFits.Fitter;

  if Fitter <> nil then begin
    Ix := Fitter.GetParamCount();
    SetLength(prLabels, Ix);
    SetLength(prValues, Ix);

    for Ix := 0 to Fitter.GetParamCount() - 1 do begin
      prValues[Ix]           := TEdit.Create(Self);
      prValues[Ix].Parent    := Self;
      prValues[Ix].Top       := Ix * prValues[Ix].Height;
      prValues[Ix].Left      := FITTERPARAMS_MID;
      prValues[Ix].Width     := ClientWidth - FITTERPARAMS_MID;
      prValues[Ix].Anchors   := [akTop, akLeft, akRight];
      prValues[Ix].OnKeyPress:= prOnKeyPress;
      prValues[Ix].OnExit    := prOnValueExit;
      prValues[Ix].Tag       := Ix;
      prValues[Ix].Text      := Format(FITTERPARAMS_FORMAT, [Fitter.GetParamVal(Ix)]);

      prLabels[Ix]           := TLabel.Create(Self);
      prLabels[Ix].Parent    := Self;
      prLabels[Ix].Top       := prValues[Ix].Top;
      prLabels[Ix].Left      := 0;
      prLabels[Ix].Caption   := Fitter.GetParamName(Ix);
    end;
  end else begin
    SetLength(prLabels, 0);
    SetLength(prValues, 0);
  end;
end;

procedure TQUB_FitterParams.prOnFitterParamVal(Sender :TObject);
var
  Ix      :Integer;
begin
  if prFits.Fitter.GetParamCount() <> Length(prValues) then Exit;
  for Ix := 0 to prFits.Fitter.GetParamCount - 1 do begin
    prValues[Ix].Text := Format(FITTERPARAMS_FORMAT, [prFits.Fitter.GetParamVal(Ix)]);
  end;
end;

procedure TQUB_FitterParams.prOnKeyPress(Sender :TObject; var Key :Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    TEdit(Sender).OnExit(Sender);
  end else begin
    TEdit(Sender).Font.Color := clRed;
  end;
end;

procedure TQUB_FitterParams.prOnValueExit(Sender :TObject);
var
  Fitter  :TQUB_Fitter;
  x       :Double;
begin
  Fitter  := prFits.Fitter;
  try
    x     := StrToFloat(TEdit(Sender).Text);
    Fitter.SetParamVal( TEdit(Sender).Tag, x );
  except
    on E: Exception do begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
    end;
  end;
  TEdit(Sender).Text := Format(FITTERPARAMS_FORMAT, [Fitter.GetParamVal(TEdit(Sender).Tag)]);
  TEdit(Sender).Font.Color := clWindowText;
end;


//******************************************************************************

procedure TQUB_FitterPanel.prSetFits(x :TQUB_Fits);
var
  Lbl                  :TLabel;
  i                    :Integer;
begin
  prFits               := x;
  prFits.Events.AddEvent('OnFitterChange', prOnFitterChange);

  Lbl                  := TLabel.Create(Self);
  Lbl.Parent           := Self;
  Lbl.Top              := 0;
  Lbl.Left             := 0;
  Lbl.Caption          := 'Fitter:';

  mnuFitter            := TComboBox.Create(Self);
  mnuFitter.Style      := csDropDownList;
  mnuFitter.DropDownCount := 16;
  mnuFitter.Parent     := Self;
  mnuFitter.Top        := 0;
  mnuFitter.Left       := Lbl.Width + 10;
  mnuFitter.Width      := ClientWidth - mnuFitter.Left;
  mnuFitter.Anchors    := [akTop, akLeft, akRight];
  mnuFitter.OnChange   := prOnChooseFitter;

  for i := 0 to prFits.FitterCount - 1 do begin
    mnuFitter.Items.Add( prFits.FitterName[i] );
  end;
  mnuFitter.ItemIndex  := mnuFitter.Items.IndexOf(prFits.Fitter.GetName());

  btnFit              := TButton.Create(Self);
  btnFit.Parent       := Self;
  btnFit.Top          := ClientHeight - btnFit.Height;
  btnFit.Left         := ClientWidth - btnFit.Width;
  btnFit.Width        := 75;
  btnFit.Anchors      := [akBottom, akRight];
  btnFit.Caption      := 'Fit';
  btnFit.OnClick      := prOnFitClick;
  btnFit.Visible      := Assigned(prOnFit);

  btnKeepIt           := TButton.Create(Self);
  btnKeepIt.Parent    := Self;
  btnKeepIt.Top       := ClientHeight - btnKeepIt.Height;
  btnKeepIt.Left      := btnFit.Left - btnFit.Width - 5;
  btnKeepIt.Width     := btnFit.Width;
  btnKeepIt.Anchors   := [akBottom, akRight];
  btnKeepIt.Caption   := 'Keep It';
  btnKeepIt.OnClick   := prOnKeepItClick;
  btnKeepIt.Visible   := Assigned(prOnKeepIt);

  btnFitAll           := TButton.Create(Self);
  btnFitAll.Parent    := Self;
  btnFitAll.Top       := ClientHeight - btnFitAll.Height;
  btnFitAll.Width     := btnKeepIt.Width;
  btnFitAll.Left      := btnKeepIt.Left - btnKeepIt.Width - 5;
  btnFitAll.Anchors   := [akBottom, akRight];
  btnFitAll.Caption   := 'Fit All';
  btnFitAll.OnClick   := prOnFitAllClick;
  btnFitAll.Visible   := Assigned(prOnFitAll);

  panParams           := TQUB_FitterParams.Create(Self);
  panParams.Parent    := Self;
  panParams.Top       := mnuFitter.Top + mnuFitter.Height;
  panParams.Left      := 0;
  panParams.Width     := ClientWidth;
  panParams.Height    := btnFit.Top - panParams.Top;
  panParams.Anchors   := [akLeft, akRight, akTop, akBottom];
  panParams.Fits      := prFits;
end;

destructor TQUB_FitterPanel.Destroy;
begin
  try
    prFits.Events.DelEvent('OnFitterChange', prOnFitterChange);
  finally
    inherited;
  end;
end;

function TQUB_FitterPanel.FitBtnScreenPos: TPoint;
begin
  Result.x := btnFit.Left;
  Result.y := btnFit.Top;
  Result := ClientToScreen(Result);
end;

procedure TQUB_FitterPanel.prOnFitterChange(Sender :TObject);
begin
  mnuFitter.ItemIndex := mnuFitter.Items.IndexOf(prFits.Fitter.GetName());
end;

procedure TQUB_FitterPanel.prOnChooseFitter(Sender :TObject);
begin
  if (prFits.Fitter.GetName() = mnuFitter.Items[mnuFitter.ItemIndex])
  then Exit;

  prFits.Fitter := CreateFitter(PChar(mnuFitter.Items[mnuFitter.ItemIndex]));
end;

procedure TQUB_FitterPanel.prOnFitClick(Sender :TObject);
begin
  if Assigned(prOnFit) then prOnFit(Self);
end;

procedure TQUB_FitterPanel.prOnKeepItClick(Sender :TObject);
begin
  if Assigned(prOnKeepIt) then prOnKeepIt(Self);
end;

procedure TQUB_FitterPanel.prOnFitAllClick(Sender :TObject);
begin
  if Assigned(prOnFitAll) then prOnFitAll(Self);
end;

procedure TQUB_FitterPanel.prSetOnFit(x :TNotifyEvent);
begin
  if btnFit <> nil then btnFit.Visible := Assigned(x);
  prOnFit := x;
end;

procedure TQUB_FitterPanel.prSetOnKeepIt(x :TNotifyEvent);
begin
  if btnKeepIt <> nil then btnKeepIt.Visible := Assigned(x);
  prOnKeepIt := x;
end;

procedure TQUB_FitterPanel.prSetOnFitAll(x :TNotifyEvent);
begin
  if btnFitAll <> nil then btnFitAll.Visible := Assigned(x);
  prOnFitAll := x;
end;




//******************************************************************************

procedure TQUB_FitData_Control.prSetFits(x :TQUB_Fits);
begin
  OnResize              := prOnResize;

  lblFrom               := TLabel.Create(Self);
  lblFrom.Parent        := Self;
  lblFrom.Caption       := 'sel:';//'Fit window:';
  lblTo                 := TLabel.Create(Self);
  lblTo.Parent          := Self;
  lblTo.Caption         := '% to';
  lblS                  := TLabel.Create(Self);
  lblS.Parent           := Self;
  lblS.Caption          := ''; // 's'; make this configurable; blank for now
  txtFromPct            := TEdit.Create(Self);
  txtFromPct.Parent     := Self;
  txtFromPct.OnKeyPress := prOnKeyPress;
  txtFromPct.OnExit     := prOnExitFromPct;
  txtFromS              := TEdit.Create(Self);
  txtFromS.Parent       := Self;
  txtFromS.OnKeyPress   := prOnKeyPress;
  txtFromS.OnExit       := prOnExitFromS;
  txtToPct              := TEdit.Create(Self);
  txtToPct.Parent       := Self;
  txtToPct.OnKeyPress   := prOnKeyPress;
  txtToPct.OnExit       := prOnExitToPct;
  txtToS                := TEdit.Create(Self);
  txtToS.Parent         := Self;
  txtToS.OnKeyPress     := prOnKeyPress;
  txtToS.OnExit         := prOnExitToS;
  lblXName              := TLabel.Create(Self);
  lblXName.Parent       := Self;
  lblXName.Caption      := 'X: ';
  lblYName              := TLabel.Create(Self);
  lblYName.Parent       := Self;
  lblYName.Caption      := 'Y: ';
  mnuXName              := TComboBox.Create(Self);
  mnuXName.Parent       := Self;
  mnuXName.Style        := csDropDownList;
  mnuXName.OnChange     := prOnXNameChange;
  mnuYName              := TComboBox.Create(Self);
  mnuYName.Parent       := Self;
  mnuYName.Style        := csDropDownList;
  mnuYName.OnChange     := prOnYNameChange;
  lblWName              := TLabel.Create(Self);
  lblWName.Parent       := Self;
  lblWName.Caption      := 'Weights:';
  mnuWName              := TComboBox.Create(Self);
  mnuWName.Parent       := Self;
  mnuWName.Style        := csDropDownList;
  mnuWName.DropDownCount := 16;
  mnuWName.OnChange     := prOnWNameChange;
  mnuWName.Hint         := 'weighted_dy = weight * dy';
  chkFilter             := TCheckBox.Create(Self);
  chkFilter.Parent      := Self;
  chkFilter.Caption     := 'Filter';
  chkFilter.Alignment   := taRightJustify;
  chkFilter.OnClick     := prOnClickFilter;
  txtFreq               := TEdit.Create(Self);
  txtFreq.Parent        := Self;
  txtFreq.OnKeyPress    := prOnKeyPress;
  txtFreq.OnExit        := prOnExitFreq;
  lblHz                 := TLabel.Create(Self);
  lblHz.Parent          := Self;
  lblHz.Caption         := 'Hz';
  chkResample           := TCheckBox.Create(Self);
  chkResample.Parent    := Self;
  chkResample.Caption   := 'Resample';
  chkResample.Alignment := taRightJustify;
  chkResample.OnClick   := prOnClickResample;
  chkResample.Hint      := 'keep one point from each region with small enough std. deviation';
  lblResampleStd        := TLabel.Create(Self);
  lblResampleStd.Parent := Self;
  lblResampleStd.Caption := 'ampl.:';
  lblResampleStd.Hint    := chkResample.Hint;
  txtResampleStd         := TEdit.Create(Self);
  txtResampleStd.Parent  := Self;
  txtResampleStd.OnKeyPress := prOnKeyPress;
  txtResampleStd.OnExit  := prOnExitResampleStd;
  txtResampleStd.Hint    := chkResample.Hint;
  lblN                   := TLabel.Create(Self);
  lblN.Parent            := Self;
  lblN.Caption           := 'N:';
  txtN                   := TEdit.Create(Self);
  txtN.Parent            := Self;
  txtN.ReadOnly          := True;
  txtN.Hint              := 'number of data points to be fit';

  mnuPresets            := TQUB_ProfileEditBox.Create(Self);
  mnuPresets.Parent     := Self;
  mnuPresets.Cat        := 'FitData';

  prFits       := x;
  prData       := prFits.Data;
  prFits.Events.AddEvent('OnDataChange', prOnDataChange);
  prOnResize(Self);
  prOnDataChange(prFits);
end;

destructor TQUB_FitData_Control.Destroy;
begin
  try
    prFits.Events.DelEvent('OnDataChange', prOnDataChange);
  finally
    inherited;
  end;
end;

procedure TQUB_FitData_Control.prOnDataChange(Sender :TObject);
var
  i                     :Integer;
begin
  txtFromPct.Text       := Format(FITDATA_PCT_FORMAT, [prData.SelFromPct]);
  txtToPct.Text         := Format(FITDATA_PCT_FORMAT, [prData.SelToPct]);
  txtFromS.Text         := Format(FITDATA_S_FORMAT, [prData.SelFromS]);
  txtToS.Text           := Format(FITDATA_S_FORMAT, [prData.SelToS]);
  mnuXName.Clear;
  mnuYName.Clear;
  mnuWName.Clear;
  mnuWName.Items.Add('<none>');
  for i := 0 to prData.DataSet.VarCount - 1 do begin
    mnuXName.Items.Add(prData.DataSet.VarName[i]);
    mnuYName.Items.Add(prData.DataSet.VarName[i]);
    mnuWName.Items.Add(prData.DataSet.VarName[i]);
  end;
  for i := 0 to prData.DataSet.VarCount - 1 do begin
    mnuWName.Items.Add('1/'+prData.DataSet.VarName[i]);
  end;
  mnuXName.ItemIndex    := prData.DataSet.Xindex;
  mnuYName.ItemIndex    := prData.DataSet.Yindex;
  if prData.DataSet.Windex < 0 then begin
    mnuWName.ItemIndex  := 0;
  end else begin
    mnuWName.ItemIndex  := 1 + prData.DataSet.Windex + (Integer(prData.DataSet.Winverse) * prData.DataSet.VarCount);
  end;
  chkFilter.Enabled     := prData.CanFilter;
  chkFilter.Checked     := prData.Filter;
  txtFreq.Text          := Format(FITDATA_FREQ_FORMAT, [prData.Freq]);
  chkResample.Checked   := prData.Resample;
  txtResampleStd.Text   := Format(FITDATA_STD_FORMAT, [prData.ResampleStd]);
  txtN.Text             := IntToStr(prData.UseDataSet.N);
end;

procedure TQUB_FitData_Control.prOnResize(Sender :TObject);
begin
  lblFrom.Top          := 0;
  lblFrom.Left         := 0;
  txtFromPct.Top       := 0;
  txtFromPct.Left      := lblFrom.Width;
  txtFromPct.Width     := FITDATA_TXT_WIDTH;
  txtFromS.Top         := txtFromPct.Height;
  txtFromS.Left        := txtFromPct.Left;
  txtFromS.Width       := txtFromPct.Width;
  lblTo.Top            := 0;
  lblTo.Left           := txtFromPct.Left + txtFromPct.Width;
  lblS.Top             := txtFromS.Top;
  lblS.Left            := lblTo.Left;
  txtToPct.Top         := 0;
  txtToPct.Left        := lblTo.Left + lblTo.Width;
  txtToPct.Width       := FITDATA_TXT_WIDTH;
  txtToS.Top           := txtFromS.Top;
  txtToS.Left          := txtToPct.Left;
  txtToS.Width         := txtToPct.Width;
  lblXName.Top         := 0;
  lblXName.Left        := txtToPct.Left + txtToPct.Width + 5;
  mnuXName.Top         := 0;
  mnuXName.Left        := lblXName.Left + lblXName.Width + 5;
  mnuXName.Width       := FITDATA_MNU_WIDTH;
  lblYName.Top         := txtToS.Top;
  lblYName.Left        := lblXName.Left;
  mnuYName.Top         := txtToS.Top;
  mnuYName.Width       := FITDATA_MNU_WIDTH;
  mnuYName.Left        := mnuXName.Left;
  lblWName.Top         := 0;
  lblWName.Left        := mnuXName.Left + mnuXName.Width + 5;
  mnuWName.Top         := mnuYName.Top;
  mnuWName.Left        := lblWName.Left;
  mnuWName.Width       := FITDATA_MNU_WIDTH;
  chkResample.Top      := mnuYName.Top;
  chkResample.Left     := mnuWName.Left + mnuWName.Width + 5;
  chkResample.Width    := 70;
  lblResampleStd.Top   := chkResample.Top;
  lblResampleStd.Left  := chkResample.Left + chkResample.Width;
  txtResampleStd.Top   := chkResample.Top;
  txtResampleStd.Left  := lblResampleStd.Left + lblResampleStd.Width;
  txtResampleStd.Width := FITDATA_TXT_WIDTH;
  chkFilter.Top        := 0;
  chkFilter.Left       := chkResample.Left;
  chkFilter.Width      := 45;
  txtFreq.Top          := chkFilter.Top;
  txtFreq.Left         := txtResampleStd.Left; // chkFilter.Left + chkFilter.Width;
  txtFreq.Width        := FITDATA_TXT_WIDTH;
  lblHz.Top            := txtFreq.Top;
  lblHz.Left           := txtFreq.Left + txtFreq.Width + 3;
(*  lblBaseline.Top      := txtToS.Top;
  lblBaseline.Left     := txtToS.Left + txtToS.Width;
  txtBaseline.Top      := txtToS.Top;
  txtBaseline.Left     := lblBaseline.Left + lblBaseline.Width;
  txtBaseline.Width    := Avail - lblBaseline.Width;  *)
  mnuPresets.Top       := txtToS.Top;
  mnuPresets.Width     := 90;
  mnuPresets.Left      := ClientWidth - mnuPresets.Width;
  lblN.Top             := 0;
  lblN.Left            := mnuPresets.Left;
  txtN.Top             := 0;
  txtN.Left            := lblN.Left + lblN.Width + 3;
  txtN.Width           := ClientWidth - txtN.Left;
end;

procedure TQUB_FitData_Control.prOnKeyPress(Sender :TObject; var Key :Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    TEdit(Sender).OnExit(Sender);
  end else begin
    TEdit(Sender).Font.Color := clRed;
  end;
end;

procedure TQUB_FitData_Control.prOnExitFromPct(Sender :TObject);
var
  x       :Double;
begin
  try
    x       := StrToFloat(TEdit(Sender).Text);
    if (x < 0.0) or (x >= 100.0) then raise Exception.Create('Must be between 0 and 100');
    prData.SelFromPct := x;
  except
    on E: Exception do begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
    end;
  end;
  TEdit(Sender).Font.Color := clWindowText;
  TEdit(Sender).Text := Format(FITDATA_PCT_FORMAT, [prData.SelFromPct]);
end;

procedure TQUB_FitData_Control.prOnExitFromS(Sender :TObject);
var
  x       :Double;
begin
  try
    x       := StrToFloat(TEdit(Sender).Text);
    if (x < prData.Xmin) or (x > prData.Xmax) then raise Exception.Create(Format('Must be between %f and %f', [prData.Xmin, prData.Xmax]));
    prData.SelFromS := x;
  except
    on E: Exception do begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
    end;
  end;
  TEdit(Sender).Text := Format(FITDATA_S_FORMAT, [prData.SelFromS]);
  TEdit(Sender).Font.Color := clWindowText;
end;

procedure TQUB_FitData_Control.prOnExitToPct(Sender :TObject);
var
  x       :Double;
begin
  try
    x       := StrToFloat(TEdit(Sender).Text);
    if (x <= 0.0) or (x > 100.0) then raise Exception.Create('Must be between 0 and 100');
    prData.SelToPct   := x;
  except
    on E: Exception do begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
    end;
  end;
  TEdit(Sender).Text := Format(FITDATA_PCT_FORMAT, [prData.SelToPct]);
  TEdit(Sender).Font.Color := clWindowText;
end;

procedure TQUB_FitData_Control.prOnExitToS(Sender :TObject);
var
  x       :Double;
begin
  try
    x       := StrToFloat(TEdit(Sender).Text);
    if (x < prData.Xmin) or (x > prData.Xmax) then raise Exception.Create(Format('Must be between %f and %f', [prData.Xmin, prData.Xmax]));
    prData.SelToS     := x;
  except
    on E: Exception do begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
    end;
  end;
  TEdit(Sender).Text := Format(FITDATA_S_FORMAT, [prData.SelToS]);
  TEdit(Sender).Font.Color := clWindowText;
end;

procedure TQUB_FitData_Control.prOnXNameChange(Sender :TObject);
begin
  if mnuXName.ItemIndex < 0
  then mnuXName.ItemIndex := prData.DataSet.Xindex
  else prData.XName := mnuXName.Items[mnuXName.ItemIndex];
end;

procedure TQUB_FitData_Control.prOnYNameChange(Sender :TObject);
begin
  if mnuYName.ItemIndex < 0
  then mnuYName.ItemIndex := prData.DataSet.Yindex
  else prData.YName := mnuYName.Items[mnuYName.ItemIndex];
end;

procedure TQUB_FitData_Control.prOnWNameChange(Sender :TObject);
var
  Ix     :Integer;
  Inv    :Boolean;
begin
  if mnuWName.ItemIndex <= 0 then begin
    prData.UseWeights('', prData.Winverse);
  end else begin
    Ix   := (mnuWName.ItemIndex - 1) mod prData.DataSet.VarCount;
    Inv  := Ix <> (mnuWName.ItemIndex - 1);
    prData.UseWeights(prData.DataSet.VarName[Ix], Inv);
  end;
end;

procedure TQUB_FitData_Control.prOnClickFilter(Sender :TObject);
begin
  prData.Filter := chkFilter.Checked;
end;

procedure TQUB_FitData_Control.prOnExitFreq(Sender :TObject);
var
  x       :Double;
begin
  try
    x           := StrToFloat(TEdit(Sender).Text);
    if x <= 0.0 then raise Exception.Create('Filter freq must be > 0');
    prData.Freq := x;
  except
    on E: Exception do begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
    end;
  end;
  TEdit(Sender).Text := Format(FITDATA_FREQ_FORMAT, [prData.Freq]);
  TEdit(Sender).Font.Color := clWindowText;
end;

procedure TQUB_FitData_Control.prOnClickResample(Sender :TObject);
begin
  prData.Resample := chkResample.Checked;
end;

procedure TQUB_FitData_Control.prOnExitResampleStd(Sender :TObject);
var
  x       :Double;
begin
  try
    x           := StrToFloat(TEdit(Sender).Text);
    if x <= 0.0 then raise Exception.Create('Resample std.dev. must be > 0');
    prData.ResampleStd := x;
  except
    on E: Exception do begin
      MessageDlg(E.Message, mtError, [mbOK], 0);
    end;
  end;
  TEdit(Sender).Text := Format(FITDATA_STD_FORMAT, [prData.ResampleStd]);
  TEdit(Sender).Font.Color := clWindowText;
end;


//******************************************************************************

constructor TQUB_FitData_Display.Create(AOwner :TComponent);
begin
  inherited;

  ControlStyle          := ControlStyle + [csOpaque, csReflector];

  OnMouseDown           := prOnMouseDown;
  OnMouseMove           := prOnMouseMove;
  OnMouseUp             := prOnMouseUp;
  OnDblClick            := prOnDblClick;
  prMouseX              := -1;
  prMouseY              := -1;
  prXm                  := 1.0;
  prYm                  := 1.0;

  prBitmap              := Graphics.TBitmap.Create;
  prBitmapCanvas        := prBitmap.Canvas;
  prBitmap.Width        := Width;
  prBitmap.Height       := Height;
end;

procedure TQUB_FitData_Display.prSetFits(x :TQUB_Fits);
begin
  if prFits <> nil then begin
    prFits.Events.DelEvent('OnDataChange', prOnDataChange);
    prFits.Events.DelEvent('OnCurveChange', prOnCurveChange);
    prFits.Events.DelEvent('OnCurveParams', prOnCurveChange);
    prFits.Events.DelEvent('OnCurveParamVal', prOnCurveChange);
    prFits.Events.DelEvent('OnFitterIter', prOnFitterIter);
  end;

  prFits      := x;
  prData      := nil;

  if prFits <> nil then begin
    prData      := prFits.Data;
    prFits.Events.AddEvent('OnDataChange', prOnDataChange);
    prFits.Events.AddEvent('OnCurveChange', prOnCurveChange);
    prFits.Events.AddEvent('OnCurveParams', prOnCurveChange);
    prFits.Events.AddEvent('OnCurveParamVal', prOnCurveChange);
    prFits.Events.AddEvent('OnFitterIter', prOnFitterIter);
    prOnDataChange(prFits);
  end;
end;

destructor TQUB_FitData_Display.Destroy;
begin
  try
    Fits := nil;
    prBitmap.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_FitData_Display.prSetHiRes(x :Boolean);
begin
  prHiRes := x;
  prOnDataChange(Self);
end;

procedure TQUB_FitData_Display.prOnDataChange(Sender :TObject);
var
  i, j            :Integer;

  procedure Expand(var vmin, vmax :Double; Factor :Double);
  var
    Expansion     :Double;
  begin
    if vmin <> vmax then begin
      Expansion   := Factor * (vmax - vmin);
    end else if vmin <> 0 then begin
      Expansion   := vmin;
    end else begin
      Expansion   := 1.0;
    end;
    vmin          := vmin - Expansion;
    vmax          := vmax + Expansion;
  end;

begin
  prSelFrom    := Data.SelFromS;
  prSelTo      := Data.SelToS;

  if (Data.DataSet.XX <> nil) and (Data.DataSet.YY <> nil) and (Data.DataSet.N > 0) then begin
    prYmin     := Data.DataSet.YY[0];
    prYmax     := Data.DataSet.YY[0];
    for i := 1 to Data.DataSet.N - 1 do begin
      if Data.DataSet.YY[i] < prYmin then prYmin := Data.DataSet.YY[i];
      if Data.DataSet.YY[i] > prYmax then prYmax := Data.DataSet.YY[i];
    end;
    if Data.DisplayStyle = QUBFITS_DISPLAY_HIST then begin
      prYmin := Min(0.0, prYmax);
      Expand(prYmin, prYmax, FITDATA_Y_HIST_EXPAND);
    end else begin
      Expand(prYmin, prYmax, FITDATA_Y_EXPAND);
    end;

    SetLength(prVmin, Data.DataSet.VarCount);
    SetLength(prVmax, Data.DataSet.VarCount);
    SetLength(prVm,   Data.DataSet.VarCount);
    for j := 0 to Data.DataSet.VarCount - 1 do begin
      prVmin[j] := Data.DataSet.Vals[j][0];
      prVmax[j] := Data.DataSet.Vals[j][0];
      for i := 1 to Data.DataSet.N - 1 do begin
        if Data.DataSet.Vals[j][i] < prVmin[j] then prVmin[j] := Data.DataSet.Vals[j][i];
        if Data.DataSet.Vals[j][i] > prVmax[j] then prVmax[j] := Data.DataSet.Vals[j][i];
      end;
      Expand(prVmin[j], prVmax[j], FITDATA_Y_EXPAND);
    end;

    if prHiRes then begin
      prXmin   := prSelFrom;
      prXmax   := prSelTo;
      Expand(prXmin, prXmax, FITDATA_HIRES_EXPAND);
    end else begin
      prXmin   := Data.DataSet.XX[0];
      prXmax   := Data.DataSet.XX[Data.DataSet.N - 1];
      if prXmin = prXmax then begin
        if prXmin <> 0.0 then begin
          prXmin := 0.0;
          prXmax := 2.0 * prXmax;
        end else begin
          prXmin := -1.0;
          prYmin := 1.0;
        end;
      end;
    end;
  end;

  ForceRedraw;
end;

procedure TQUB_FitData_Display.prOnCurveChange(Sender :TObject);
begin
  if not Fitting then ForceRedraw;
end;

procedure TQUB_FitData_Display.prOnFitterIter(Sender :TObject);
begin
  ForceRedraw;
end;

procedure TQUB_FitData_Display.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited;
  prBitmap.Width        := AWidth;
  prBitmap.Height       := AHeight;
  ForceRedraw;
end;

procedure TQUB_FitData_Display.ForceRedraw;
begin
  prNeedsDraw := True;
  Invalidate;
end;

procedure TQUB_FitData_Display.Paint;
var
  Rect, Dot        :TRect;
  i, j             :Integer;
  ix               :Integer;
  Coords           :String;
  ExtraCoords      :String;
  Sz               :TSize;
  DrawData         :TQUB_FitDataSet;
  Component        :array of Double;
  DotRad           :Integer;
  x                :Double;
  px, py, y0       :Integer;
  ch               :Integer;
  function YClip(Y :Integer): Integer;
  begin
    Result := Max(0, Min(ch, Y));
  end;
begin
  if (Parent <> nil) and (Data.DataSet.N > 0) then begin
    try
      if not (csDestroying in ComponentState) then begin
        ch := ClientHeight;
        //----- Rebuild geometry and bitmap buffer
        if prNeedsDraw then begin
          prNeedsDraw := False;         // Flag for whether to execute this block

  //    figure out scale:
          prXm        := (prXmax - prXmin) / ClientWidth;
          prYm        := (prYmax - prYmin) / ClientHeight;

          for j := 0 to Data.DataSet.VarCount - 1 do begin
            prVm[j]   := (prVmax[j] - prVmin[j]) / ClientHeight;
          end;

  //    paint non-sel with different bg
          Rect        := ClientRect;
          prBitmapCanvas.Brush.Color := TColor($ffaaaa);
          prBitmapCanvas.FillRect(Rect);

  // pick all/chosen data
          if prHiRes then begin
            DrawData  := Data.DataSet;
            with DrawData do begin
  // draw extra data series
{
              prBitmapCanvas.Pen.Color := clLime;
              for j := 0 to ExtraCount - 1 do begin
                if Count > 0 then prBitmapCanvas.MoveTo(XtoScreen(XX[0]), YtoScreen(ExtraVals[j][0], j));
                for ix := 1 to Count - 1 do begin
                  prBitmapCanvas.LineTo(XtoScreen(XX[ix]), YtoScreen(ExtraVals[j][ix], j));
                end;
              end;
}
  //    draw full data anti-sel
              prBitmapCanvas.Pen.Color   := clYellow;
              prBitmapCanvas.Brush.Color := clYellow;
              if Self.Data.DisplayStyle = QUBFITS_DISPLAY_HIST then begin
                y0     := YtoScreen(0.0);
                for ix := 0 to N - 1 do begin
                  py       := YtoScreen(YY[ix]);
                  x        := XX[ix];
                  if ix = 0 then px := XtoScreen(x - ((XX[1] - x) / 2.0) * (FITDATA_HIST_WIDTH / 100.0))
                  else           px := XtoScreen(x - ((x - XX[ix-1]) / 2.0) * (FITDATA_HIST_WIDTH / 100.0));
                  prBitmapCanvas.MoveTo(px, y0);
                  prBitmapCanvas.LineTo(px, py);
                  if ix = (N-1) then px := XtoScreen(x + ((x - XX[ix-1]) / 2.0) * (FITDATA_HIST_WIDTH / 100.0))
                  else              px := XtoScreen(x + ((XX[ix+1] - x) / 2.0) * (FITDATA_HIST_WIDTH / 100.0));
                  prBitmapCanvas.LineTo(px, py);
                  prBitmapCanvas.LineTo(px, y0);
                end;
              end else if (Self.Data.DisplayStyle = QUBFITS_DISPLAY_POINTS)
                       or (N < FITDATA_POINTS_THRESH) then begin
                for ix := 0 to N - 1 do begin
                  if Wmean <> 0 then begin
                    DotRad     := Max(1, 2+Round(Ln(WW[ix]/Wmean)));
                  end else begin
                    DotRad     := 2;
                  end;
                  Dot.Left     := XtoScreen(XX[ix]) - DotRad;
                  Dot.Right    := Dot.Left + 2*DotRad;
                  Dot.Top      := YtoScreen(YY[ix]) - DotRad;
                  Dot.Bottom   := Dot.Top + 2*DotRad;
                  prBitmapCanvas.Ellipse(Dot);
                end;
              end else begin
                if N > 0 then prBitmapCanvas.MoveTo(XtoScreen(XX[0]), YtoScreen(YY[0]));
                for ix := 1 to N - 1 do begin
                  prBitmapCanvas.LineTo(XtoScreen(XX[ix]), YtoScreen(YY[ix]));
                end;
              end;
            end;

            DrawData  := Data.UseDataSet;
          end else begin
            DrawData  := Data.DataSet;
          end;

  //    sel-bg overwrites grey hi-res full data
          Rect.Left   := XtoScreen(Data.SelFromS);
          Rect.Right  := XtoScreen(Data.SelToS);
          prBitmapCanvas.Brush.Color := clWhite;
          prBitmapCanvas.FillRect(Rect);

          if Assigned(OnBeforeDraw) then OnBeforeDraw(Self, prBitmapCanvas);

          with DrawData do begin
    //    draw extra data series
            if not prHiRes then begin
              prBitmapCanvas.Pen.Color := clLime;
              for j := 0 to VarCount - 1 do begin
                if (j <> DrawData.Xindex) and (j <> DrawData.Yindex) then begin
                  if N > 0 then prBitmapCanvas.MoveTo(XtoScreen(XX[0]+xoff), YtoScreen(Vals[j][0], j));
                  for ix := 1 to N - 1 do begin
                    prBitmapCanvas.LineTo(XtoScreen(XX[ix]+xoff), YtoScreen(Vals[j][ix], j));
                  end;
                end;
              end;
            end;

    //    draw chosen data black
            prBitmapCanvas.Pen.Color := clBlack;
            prBitmapCanvas.Brush.Color := clBlack;
            if Self.Data.DisplayStyle = QUBFITS_DISPLAY_HIST then begin
              y0     := YtoScreen(0.0);
              for ix := 0 to N - 1 do begin
                py       := YtoScreen(YY[ix]);
                x        := XX[ix];
                if ix = 0 then px := XtoScreen(x - ((XX[1] - x) / 2.0) * (FITDATA_HIST_WIDTH / 100.0))
                else           px := XtoScreen(x - ((x - XX[ix-1]) / 2.0) * (FITDATA_HIST_WIDTH / 100.0));
                prBitmapCanvas.MoveTo(px, y0);
                prBitmapCanvas.LineTo(px, py);
                if ix = (N-1) then px := XtoScreen(x + ((x - XX[ix-1]) / 2.0) * (FITDATA_HIST_WIDTH / 100.0))
                else              px := XtoScreen(x + ((XX[ix+1] - x) / 2.0) * (FITDATA_HIST_WIDTH / 100.0));
                prBitmapCanvas.LineTo(px, py);
                prBitmapCanvas.LineTo(px, y0);
              end;
            end else if (Self.Data.DisplayStyle = QUBFITS_DISPLAY_POINTS)
                     or (N < FITDATA_POINTS_THRESH) then begin
              for ix := 0 to N - 1 do begin
                if Wmean <> 0 then begin
                  DotRad     := Max(1, 2+Round(Ln(WW[ix]/Wmean)));
                end else begin
                  DotRad     := 2;
                end;
                Dot.Left     := XtoScreen(XX[ix]+xoff) - DotRad;
                Dot.Right    := Dot.Left + 2*DotRad;
                Dot.Top      := YtoScreen(YY[ix]) - DotRad;
                Dot.Bottom   := Dot.Top + 2*DotRad;
                prBitmapCanvas.Ellipse(Dot);
              end;
            end else begin
              if N > 0 then prBitmapCanvas.MoveTo(XtoScreen(XX[0]+xoff), YtoScreen(YY[0]));
              for ix := 1 to N - 1 do begin
                prBitmapCanvas.LineTo(XtoScreen(XX[ix]+xoff), YtoScreen(YY[ix]));
              end;
            end;
          end;
        end;

        with Data.UseDataSet do begin
  //    draw fits in red
          prBitmapCanvas.Pen.Color := clRed;
          if (N > 0) and (prFits.Curve.GetComponentCount() > 1) then begin
            SetLength(Component, N);
            for j := 0 to prFits.Curve.GetComponentCount() - 1 do begin
              prFits.Curve.EvalArrayOne(Data, @(Component[0]), j);
              prBitmapCanvas.MoveTo(XtoScreen(XX[0]+xoff), YClip(YtoScreen(Component[0])));
              for ix := 1 to N - 1 do begin
                prBitmapCanvas.LineTo(XtoScreen(XX[ix]+xoff), YClip(YtoScreen(Component[ix])));
              end;
            end;
          end;
          prBitmapCanvas.Pen.Width := 2;
          if N > 0 then prBitmapCanvas.MoveTo(XtoScreen(XX[0]+xoff), YClip(YtoScreen(Self.Data.FY[0])));
          for ix := 1 to N - 1 do begin
            prBitmapCanvas.LineTo(XtoScreen(XX[ix]+xoff), YClip(YtoScreen(Self.Data.FY[ix])));
          end;
          prBitmapCanvas.Pen.Width := 1;
        end;

        //----- Paint from bitmap
        try
          Canvas.Lock;
          Canvas.Draw(0, 0, prBitmap);

  // draw moving item (sel bound) in bold pink
          Canvas.Pen.Color := clFuchsia;
          if prCanMoveSF then begin
            i := XtoScreen(prSelFrom);
            Canvas.MoveTo(i, 0);
            Canvas.LineTo(i, Height);
          end else if prCanMoveSL then begin
            i := XtoScreen(prSelTo);
            Canvas.MoveTo(i, 0);
            Canvas.LineTo(i, Height);
          end else if prMoving then begin // lo-res drag-select
            i := Min(prMouseX, prMouseDownX);
            Canvas.MoveTo(i, 0);
            Canvas.LineTo(i, Height);
            j := Max(prMouseX, prMouseDownX);
            Canvas.MoveTo(j, 0);
            Canvas.LineTo(j, Height);
            Canvas.MoveTo(i, Height div 2);
            Canvas.LineTo(j, Height div 2);
          end;
  // draw coords as tooltip down-right of cursor point
          if  (prMouseX > 0) and (prMouseX < (Width-1))
          and (prMouseY > 0) and (prMouseY < (Height-1)) then begin
            ExtraCoords   := '';
{            if (not prHiRes) and (Data.DataSet.VarCount > 2) then begin
              for j := 0 to Data.DataSet.VarCount - 1 do begin
                if (j <> Data.DataSet.Xindex) and (j <> Data.DataSet.Yindex) and (j <> Data.DataSet.Windex) then begin
                  if ExtraCoords = '' then begin
                    ExtraCoords := #13#10'[';
                  end else begin
                    ExtraCoords := ExtraCoords + ', ';
                  end;
                  ExtraCoords := ExtraCoords + Format('%.4f', [ScreenToY(prMouseY, j)]);
                end;
              end;
              if ExtraCoords <> '' then ExtraCoords := ExtraCoords + ']';
            end;
 }           Coords    := Format('(%.4f, %.4f)%s', [ScreenToX(prMouseX), ScreenToY(prMouseY), ExtraCoords]);
            Sz        := Canvas.TextExtent(Coords);
            Rect.Left := prMouseX + 12;
            Rect.Top  := prMouseY + 2;
            Rect.Right := Rect.Left + Sz.cx + 4;
            Rect.Bottom := Rect.Top + Sz.cy + 4;
            j         := Rect.Right - Width;
            if j > 0 then begin
              Rect.Left := Rect.Left - j;
              Rect.Right := Rect.Right - j;
            end;
            j         := Rect.Bottom - Height;
            if j > 0 then begin
              Rect.Top   := Rect.Top - j;
              Rect.Bottom := Rect.Bottom - j;
            end;
            Canvas.Pen.Color := clBlack;
            Canvas.Brush.Color := clYellow;
            Canvas.Rectangle(Rect);
            Canvas.TextRect(Rect, Rect.Left+2, Rect.Top+2, Coords);
          end;

          if Assigned(OnAfterDraw) then OnAfterDraw(Self, Canvas);

        finally
          Canvas.Unlock;
        end;
      end;
    finally
      inherited;
    end;
  end else begin
  // no data white background
    Canvas.Lock;
    Rect        := ClientRect;
    Canvas.Brush.Color := TColor($ffffff);
    Canvas.FillRect(Rect);
    Canvas.Unlock;
  end;
end;

procedure TQUB_FitData_Display.prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Data.DataSet.N = 0 then Exit;
  prMouseDownX := X;
  prMouseDownY := Y;
  prOnMouseMove(Sender, Shift, X, Y);
  prMoving := True;
end;

procedure TQUB_FitData_Display.prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  i           :Integer;
begin
  if Data.DataSet.N = 0 then Exit;

  Invalidate;

  prMouseX    := X;
  prMouseY    := Y;

  if prMoving then begin
    if prCanMoveSF then begin
      prSelFrom  := ScreenToX(X);
    end else if prCanMoveSL then begin
      prSelTo    := ScreenToX(X);
    end;
  end else begin
    prCanMoveSF   := False;
    prCanMoveSL   := False;

    // hit-test sf, sl and change cursor
    i             := XtoScreen(prSelFrom);
    if ((i-3) <= X) and (X <= (i+3)) then begin
      prCanMoveSF := True;
      Cursor      := crHSplit;
      Exit;
    end;

    i             := XtoScreen(prSelTo);
    if ((i-3) <= X) and (X <= (i+3)) then begin
      prCanMoveSL := True;
      Cursor      := crHSplit;
      Exit;
    end;

    Cursor := crDefault;
  end;
end;

procedure TQUB_FitData_Display.prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  sf, st :Double;
begin
  if Data.DataSet.N = 0 then Exit;

  prMouseX    := X;
  prMouseY    := Y;

  if prMoving then begin
    if prCanMoveSF then begin
      Data.SelFromS := ScreenToX(X);
    end else if prCanMoveSL then begin
      Data.SelToS   := ScreenToX(X);
    end else if prMouseX <> prMouseDownX then begin
      sf              := ScreenToX(Min(prMouseX, prMouseDownX));
      st              := ScreenToX(Max(prMouseX, prMouseDownX));
      Data.SelFromPct := 0.0;
      Data.SelToPct   := 100.0;
      Data.SelFromS   := sf;
      Data.SelToS     := st;
    end;
    prMoving := False;

    prSelFrom := Data.SelFromS;
    prSelTo   := Data.SelToS;
  end;
end;

procedure TQUB_FitData_Display.prOnDblClick(Sender :TObject);
begin
  if not prHiRes then begin
    Data.SelFromPct := 0.0;
    Data.SelToPct   := 100.0;
  end;
end;

  // (x, y) screen -> (x0 + x*xm, y0 + (h-y)*ym)

function TQUB_FitData_Display.XtoScreen(X :Double): Integer;
begin
  Result := Round((X - prXmin) / prXm);
end;

function TQUB_FitData_Display.YtoScreen(Y :Double; Extra :Integer = -1): Integer;
var
  Ys       :Double;
const
  MINN     :Double    = Low(SmallInt)+1;
  MAXX     :Double    = High(SmallInt)-1;
begin
  if Extra < 0 then begin
    Ys     := ClientHeight - (Y - prYmin) / prYm;
  end else begin
    Ys     := ClientHeight - (Y - prVmin[Extra]) / prVm[Extra];
  end;
  if Ys < MINN then begin
    Result := Low(Integer);
  end else if Ys > MAXX then begin
    Result := High(Integer);
  end else begin
    Result := Round(Ys);
  end;
end;

function TQUB_FitData_Display.ScreenToX(i :Integer): Double;
begin
  Result := prXmin + i * prXm;
end;

function TQUB_FitData_Display.ScreenToY(j :Integer; Extra :Integer = -1): Double;
begin
  if Extra < 0 then begin
    Result := prYmin + (ClientHeight - j) * prYm;
  end else begin
    Result := prVmin[Extra] + (ClientHeight - j) * prVm[Extra];
  end;
end;


//******************************************************************************
procedure TQUB_DialACurve.Init;
begin
  prFits                := TQUB_Fits.Create;
  prLayout              := ProfileMgr['DialACurve Layout'].Active;
  ProfileMgr['DialACurve Layout'].OnSet := prOnPreSet;
  if prLayout['Data Height'].DataCount = 0
  then prLayout['Data Height'].SetData(0.5/1.0);
  if prLayout['Log Height'].DataCount = 0
  then prLayout['Log Height'].SetData(0.1/1.0);
  if prLayout['Curve Width'].DataCount = 0
  then prLayout['Curve Width'].SetData(0.5/1.0);

  prFits.Events.AddEvent('OnFitterMsg', prOnFitterMsg);
  prFits.Events.AddEvent('OnFitterErr', prOnFitterMsg);

  panCurveFitLog        := TPanel.Create(Self);
  panCurveFitLog.Parent := Self;
  panCurveFitLog.Height := Round((1.0 - prLayout['Data Height'].Data[0]) * ClientHeight);
  panCurveFitLog.Align  := alBottom;
  panCurveFitLog.Constraints.MinHeight := 200;

  lstLog                := TQFC_Log.Create(Self);
  lstLog.Parent         := panCurveFitLog;
  lstLog.Height         := Round(prLayout['Log Height'].Data[0] * ClientHeight);
  lstLog.Align          := alBottom;
  lstLog.Constraints.MinHeight := 20;

  splLog                := TSplitter.Create(Self);
  splLog.Parent         := panCurveFitLog;
  splLog.Top            := panCurveFitLog.ClientHeight - lstLog.Height - 10;
  splLog.Align          := alBottom;
  splLog.AutoSnap       := False;
  splLog.Beveled        := True;
  splLog.OnMoved        := prOnLayoutChange;

  panCurveFit           := TPanel.Create(Self);
  panCurveFit.Parent    := panCurveFitLog;
  panCurveFit.Align     := alClient;

  panFitter             := TQUB_FitterPanel.Create(Self);
  panFitter.Parent      := panCurveFit;
  panFitter.Width       := 240;
  panFitter.Align       := alRight;
  panFitter.OnFit       := prOnFit;
  panFitter.Fits        := prFits;
  panFitter.Constraints.MinWidth := 240;

  splLR                 := TSplitter.Create(Self);
  splLR.Parent          := panCurveFit;
  splLR.Left            := panFitter.Left - 10;
  splLR.Align           := alRight;
  splLR.AutoSnap        := False;
  splLR.Beveled         := True;
  splLR.OnMoved         := prOnLayoutChange;

  panCurve              := TQUB_CurvePanel.Create(Self);
  panCurve.Parent       := panCurveFit;
  panCurve.Align        := alClient;
  panCurve.Width        := Round(prLayout['Curve Width'].Data[0] * panCurveFit.Width);
  panCurve.Fits         := prFits;
  panCurve.Constraints.MinWidth := 200;

  splTB                 := TSplitter.Create(Self);
  splTB.Parent          := Self;
  splTB.Top             := panCurveFitLog.Top - 10;
  splTB.Align           := alBottom;
  splTB.AutoSnap        := False;
  splTB.Beveled         := True;
  splTB.OnMoved         := prOnLayoutChange;

  panDisplays           := TPanel.Create(Self);
  panDisplays.Parent    := Self;
  panDisplays.Align     := alClient;
  panDisplays.Height    := Round(prLayout['Data Height'].Data[0] * ClientHeight);
  panDisplays.Constraints.MinHeight := 150;

  panWhole              := TQUB_FitData_Display.Create(Self);
  panWhole.Parent       := panDisplays;
  panWhole.Top          := 0;
  panWhole.Left         := 0;
  panWhole.Width        := panDisplays.Width;
  panWhole.Height       := FITDATA_LORES_HEIGHT;
  panWhole.Anchors      := [akLeft, akRight, akTop];
  panWhole.Fits         := prFits;

  panSettings           := TQUB_FitData_Control.Create(Self);
  panSettings.Parent    := panDisplays;
  panSettings.Top       := panWhole.Height;
  panSettings.Left      := 0;
  panSettings.Width     := panDisplays.Width;
  panSettings.Height    := FITDATA_CONTROL_HEIGHT;
  panSettings.Anchors   := [akLeft, akRight, akTop];
  panSettings.Fits      := prFits;

  panDetail             := TQUB_FitData_Display.Create(Self);
  panDetail.Parent      := panDisplays;
  panDetail.Top         := panSettings.Top + panSettings.Height;
  panDetail.Left        := 0;
  panDetail.Width       := panDisplays.Width;
  panDetail.Height      := panDisplays.Height - panDetail.Top;
  panDetail.Anchors     := [akLeft, akRight, akTop, akBottom];
  panDetail.Fits        := prFits;
  panDetail.HiRes       := True;
end;

destructor TQUB_DialACurve.Destroy;
begin
  try
    panWhole.Free;
    panSettings.Free;
    panDetail.Free;
    panCurve.Free;
    panFitter.Free;
    // free prFits' users before prFits
    prFits.Free;
  finally
    inherited;
  end;
end;

function TQUB_DialACurve.GetState: INode;
var
  i, N          :Integer;
  Params, Param :INode;
begin
  Result := TNode.Create('DialACurve');
  Result.AppendClone(prLayout, True).Name                     := 'DialACurve Layout';
  Result.AppendClone(ProfileMgr['Curve'].Active, True).Name   := 'Curve';
  Result.AppendClone(ProfileMgr['Fitter'].Active, True).Name  := 'Fitter';
  Result.AppendClone(ProfileMgr['FitData'].Active, True).Name := 'FitData';
  Result['Residual'].SetData(Fits.Residual);
  Result['N'].SetData(Fits.Data.UseDataSet.N);
  with Fits.Curve do begin
    N           := GetParamCount();
    Param       := TNode.CreateNull();
    Params      := Result['Params'];
    for i := 0 to N - 1 do begin
      Param     := Params.InsertChild(Param, GetParamName(i));
      Param.SetupNumData(NODE_DATATYPE_DOUBLE, 2, 1);
      Param.Data[0] := GetParamVal(i);
      Param.Data[1] := Fits.Errors[i];
    end;
  end;
end;

procedure TQUB_DialACurve.SetState(State :INode);
begin
  ProfileMgr['DialACurve Layout'].SetProperties(State.Find('DialACurve Layout'));
  ProfileMgr['Curve'].SetProperties(State.Find('Curve'));
  ProfileMgr['Fitter'].SetProperties(State.Find('Fitter'));
  ProfileMgr['FitData'].SetProperties(State.Find('FitData'));
end;

procedure TQUB_DialACurve.prSetFitting(x :Boolean);
begin
  panCurve.Fitting  := x;
  panWhole.Fitting  := x;
  panDetail.Fitting := x;
end;

procedure TQUB_DialACurve.prOnFit(Sender :TObject);
var
  FitPos, StopPos :TPoint;
begin
  // position dialog so Stop lines up with Fit
  FitPos  := panFitter.FitBtnScreenPos;
  StopPos.x := QFC_Fitting_Dlog.btnStop.Left;
  StopPos.y := QFC_Fitting_Dlog.btnStop.Top;
  StopPos      := QFC_Fitting_Dlog.ClientToScreen(StopPos);
  QFC_Fitting_Dlog.Left := QFC_Fitting_Dlog.Left + FitPos.x - StopPos.x;
  QFC_Fitting_Dlog.Top  := QFC_Fitting_Dlog.Top  + FitPos.y - StopPos.y;

  Fitting := True;
  QFC_Fitting_Dlog.panFits := Self;
  QFC_Fitting_Dlog.ShowModal;
  Fitting := False;
end;

procedure TQUB_DialACurve.prOnLayoutChange(Sender :TObject);
begin
  prLayout['Data Height'].Data[0] := panDisplays.Height / ClientHeight;
  prLayout['Log Height'].Data[0]  := lstLog.Height / ClientHeight;
  prLayout['Curve Width'].Data[0] := panCurve.Width / panCurveFit.Width;
end;

procedure TQUB_DialACurve.prOnPreSet(Sender :TQUB_ProfileCat; Updates :INode);
begin
  if Updates.Find('Data Height').DataCount > 0
  then panDisplays.Height  := Round(Updates['Data Height'].Data[0] * ClientHeight);
  //panCurveFitLog.Height    := ClientHeight - panDisplays.Height;
  if Updates.Find('Log Height').DataCount > 0
  then lstLog.Height  := Round(Updates['Log Height'].Data[0] * ClientHeight);
  if Updates['Curve Width'].DataCount > 0
  then panCurve.Width      := Round(Updates['Curve Width'].Data[0] * panCurveFit.Width);

  prOnLayoutChange(Self);
end;

procedure TQUB_DialACurve.prOnFitterMsg(Sender :TObject);
var
  Msg     :String;
  Space   :Integer;
begin
  Msg     := prFits.Fitter.GetMessage();
  if AnsiPos('http://', Msg) = 1 then begin
    Space := AnsiPos(' ', Msg);
    if Space > 0 then begin
      lstLog.Add(Copy(Msg, Space+1, Length(Msg)),
                 Copy(Msg, 1, Space-1));
    end else begin
      lstLog.Add(Msg, Msg);
    end;
  end else begin
    lstLog.Add(Msg, '');
  end;
end;


//******************************************************************************

constructor TQUB_CurveFitButton.Create(AOwner :TComponent);
begin
  inherited;

  prState := TNode.CreateNull;
  prData  := TQUB_FitDataSet.Create;
  prFitY  := TNode.Create('');
end;

destructor TQUB_CurveFitButton.Destroy;
begin
  try
    prData.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_CurveFitButton.Capture(panFits :TQUB_DialACurve);
begin
  prFits     := panFits;
  prState    := panFits.GetState;
  if prData.N > 0 then begin
    prFitY.SetupNumData(NODE_DATATYPE_DOUBLE, prData.N, 1);
    panFits.Fits.Curve.EvalArrayAll(prData.Data, TPADouble(prFitY.DataPtr));
  end else begin
    prFitY.ClearData;
  end;
  prOnResize(Self);
end;

procedure TQUB_CurveFitButton.Restore;
begin
  prFits.Fits.Data.UseData(prData);
  prFits.SetState(prState);
end;

procedure TQUB_CurveFitButton.prSetData(x :TQUB_FitDataSet);
var
  i          :Integer;
begin
  prData.Assign(x);
  prData.MakeOwn();
  for i := prData.N - 1 downto 0 do begin
    prData.XX[i] := prData.XX[i] - prData.XX[0];
  end;
  // TODO: re-gen fit curve
  prFitY.ClearData();
end;

procedure TQUB_CurveFitButton.prOnResize(Sender :TObject);
var
  bmp  :TBitmap;
  i, N :Integer;
  r    :TRect;

  xm, xb, ym, yb :Double;
  procedure measure;
  var
    xl, xh, yl, yh :Double;
    i              :Integer;
  begin
    if N <= 0 then Exit;
    xl := prData.XX[0];   xh := xl;
    yl := prData.YY[0];   yh := yl;
    for i := 1 to N - 1 do begin
      xl := Min(xl, prData.XX[i]);
      xh := Max(xh, prData.XX[i]);
      yl := Min(yl, prData.YY[i]);
      yh := Max(yh, prData.YY[i]);
      if xh = xl then begin
        xm := Width / 2;
        xb := xh - 1;
      end else begin
        xm := Width / (xh - xl);
        xb := xl;
      end;
      if yh = yl then begin
        ym := Height / 2;
        yb := yh - 1;
      end else begin
        ym := Height / (yh - yl);
        yb := yl;
      end;
    end;
  end;
  function xcoord(t :Double): Integer;
  begin
    Result := Round(xm*(t-xb));
  end;
  function ycoord(v :Double): Integer;
  begin
    Result := Round(Height - ym*(v-yb));
  end;
begin
  N := prData.N;
  measure();
  bmp:=TBitmap.Create;
  try
    bmp.Width := Width;
    bmp.Height := Height;
    with bmp.Canvas do begin
      SetRect(r, 0, 0, Width, Height);
      Brush.Color := clWhite;
      FillRect(r);
      N := Min(prFitY.DataCount, prData.N);
      if N > 0 then begin
        Pen.Color := clBlack;
        MoveTo(xcoord(prData.XX[0]), ycoord(prData.YY[0]));
        for i := 1 to N - 1 do begin
          LineTo(xcoord(prData.XX[i]), ycoord(prData.YY[i]));
        end;
        Pen.Color := clRed;
        MoveTo(xcoord(prData.XX[0]), ycoord(prFitY.Data[0]));
        for i := 1 to N - 1 do begin
          LineTo(xcoord(prData.XX[i]), ycoord(prFitY.Data[i]));
        end;
      end;
    end;
    BMP.Canvas.Pixels[0, BMP.Height-1] := clBtnFace;
    Glyph := bmp;
  finally
    bmp.Free;
  end;
end;


//******************************************************************************
end.
