{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// by Lorin c.2003
// see QUB_A2Q_Global for discussion.

unit QUB_A2Q_Model;


interface


uses


Classes,
AllocTypes, Alloc,
qubopt_iface,

QUB_Global,
QUB_Matrices,

QUB_Q_Model, QUB_Q_Global,
QUB_A2Q_Global,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix, QUB_QFS_Channels,
QUB_ExpCond_Parameters;


type


TQUB_A2Q_Model               = class(TQUB_Q_Model)
protected


  prOptParams                :TQUB_A2Q_Opt_PParams;


public


  constructor Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_A2Q_Opt_PParams);


  function IsPZero(Rate, Index :Integer) :Boolean; override;
  function IsQZero(Rate, Index :Integer) :Boolean; override;


  property OptParams         :TQUB_A2Q_Opt_PParams    read prOptParams;


end;


implementation


//******************************************************************************
constructor TQUB_A2Q_Model.Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_A2Q_Opt_PParams);
begin

  prOptParams := OptParams;

  QMinRate    := prOptParams.ModQMinRate;

  inherited Create(Model, DoInitialize, ExpCondParams);

end;


//****************************************************************************//
//                                                                            //
//                             C O N S T R A I N T S                          //
//                                                                            //
//****************************************************************************//
function TQUB_A2Q_Model.IsPZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := False;
  if not prModel.Rates[Rate].P[Index] then Exit;
  //Not P-dependent --> not zero


  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].PName[Index]];
  if (Par <> nil) and (Par.Value = 0) and (not Par.IsVar) then begin
    Result := True;
  end;

end;

function TQUB_A2Q_Model.IsQZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := True;
  if not prModel.Rates[Rate].Q[Index] then Exit;
  //Not Q-dependent --> zero

  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].QName[Index]];
  if (Par = nil) then Exit;

  if (Par.Value <> 0) or (Par.IsVar) then begin
    Result := False;
  end;

end;


//******************************************************************************
end.
