{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// The Ubiquitous 'About' dialogue.
// Specialization - display of version number is runtime built;
// Initial about screen display is timer controlled.


unit QUB_About;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls, Registry;


type


TQUB_AboutForm = class(TForm)


  Logo             :TImage;
  Timer            :TTimer;
  lblVersion       :TLabel;

  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure FormClick(Sender: TObject);
  procedure FormKeyPress(Sender: TObject; var Key: Char);
  procedure TimerTimer(Sender: TObject);
  procedure lblVersionClick(Sender: TObject);


protected


  prFirstTime      :Boolean;


public


  property FirstTime :Boolean read prFirstTime;


end;


var


QUB_AboutForm: TQUB_AboutForm;


implementation


{$R *.DFM}


//******************************************************************************
procedure TQUB_AboutForm.FormCreate(Sender: TObject);
var
  iJunk, iBufSize :Integer;
  pchBuffer       :PChar;
  FixedFileInfo   :PVSFixedFileInfo;
begin
  Width           := Logo.Width;
  Height          := Logo.Height + (Height - ClientHeight);
  lblVersion.Left := 40;
  lblVersion.Top  := 126;
  prFirstTime     := True;

  //----- Create string of 1.2.3.4 version number
  iBufSize    := GetFileVersionInfoSize(PChar(Application.ExeName), LongWord(iJunk));
  pchBuffer   := nil;
  try
    pchBuffer := StrAlloc( iBufSize + 1);
    GetFileVersionInfo(Pchar(Application.Exename), 0, iBufSize, pchBuffer);
    VerQueryValue(pchBuffer, '\', Pointer(FixedFileInfo), Longword(iJunk));
    lblVersion.Caption := IntToStr( FixedFileInfo.dwFileVersionMS shr 16 ) + '.'
            + IntToStr( FixedFileInfo.dwFileVersionMS mod $10000 ) + '.'
            + IntToStr( FixedFileInfo.dwFileVersionLS shr 16 ) + '.'
            + IntToStr( FixedFileInfo.dwFileVersionLS mod $10000 );
  finally
    StrDispose(pchBuffer);
  end;

end;

procedure TQUB_AboutForm.FormShow(Sender: TObject);
begin
  if (prFirstTime) then begin
    Timer.Enabled := True;
  end;
  prFirstTime := False;

end;

procedure TQUB_AboutForm.FormHide(Sender: TObject);
begin
  Timer.Enabled := False;
end;

procedure TQUB_AboutForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  Close;
end;

procedure TQUB_AboutForm.FormClick(Sender: TObject);
begin
  Close;
end;

procedure TQUB_AboutForm.TimerTimer(Sender: TObject);
begin
  Timer.Enabled := False;
  Hide;
end;

procedure TQUB_AboutForm.lblVersionClick(Sender: TObject);
begin
  Close;
end;

end.
