unit QUB_AcquisitionInput;

{
/* Copyright 1998-2014 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls, StdCtrls, Registry, Math, Buttons, ExtCtrls, Menus, SyncObjs,

Alloc, AllocTypes,
RegistryInterface,

QUB_Global,

QUB_Console, QUB_Rich_Console,

QUB_QVO_QVD, QUB_QVO_QVS, QUB_QVO_QVSpectrum,

QUB_Acquisition_Global,
QUB_Acquisition_Hardware,
QUB_AcquisitionInput_Properties,

QUB_QFS_Segments;


{$include 'QUB_AcquisitionInput.hct'}


const


MIN_TIMEPOINTS                                  = 10;
MAX_TIMEPOINTS                                  = High(Integer);

DEF_BUFFERLENGTH                                = 1000;
MIN_BUFFERLENGTH                                = 10;
MAX_BUFFERLENGTH                                = 5000;

DEF_SAMPLING                                    = 10000;
MIN_SAMPLING                                    = 1e-3;
MAX_SAMPLING                                    = 1e10;


WM_QUB_ACQ_INPUT_PAUSE                          = WM_USER + 100;
WM_QUB_ACQ_INPUT_UPDATERECORDEDPOINTS           = WM_USER + 101;

{
Chris 2006 Aug attempt to stop acq in crashes:
Do anything with potential gui or qub-wide interaction in the gui thread:

The Acq_In holds a "Supply" and "Queue" of buffers from start until stop.
The callback, operating on its own (non-gui) thread,
takes an empty buffer from the Supply, copies in the data, and queues it.
If necessary, it posts a message (_AVAIL).  (the message handler posts its own
_AVAIL on exit if the queue is still not empty.)
The message handler dequeues a buffer and calls QVD.AppendData, then returns
the buffer to the Supply.

TODO: progress bar aesthetics

* the error indicator:}
WM_QUB_ACQ_INPUT_ERR                            = WM_USER + 102;    // wparam is true(nonzero) or false, to show or hide the "Err" panel
// and the buffers-arrived signal
WM_QUB_ACQ_INPUT_AVAIL                          = WM_USER + 103;

type


TQUB_Acquisition_Input_State                    = (ACQ_INPUT_STATE_STOPPED, ACQ_INPUT_STATE_RUNNING, ACQ_INPUT_STATE_RECORDING, ACQ_INPUT_STATE_PAUSED);

TQUB_Acquisition_IDataError           = (IDATA_ERR_NONE, DATA_ERR_UNASS,  DATA_ERR_DIFFCHAN, DATA_ERR_DIFFBITS);
TQUB_Acquisition_IDeviceError         = (DEV_ERR_NONE,  DEV_ERR_UNASS);
TQUB_Acquisition_IChannelsError       = (CHAN_ERR_NONE, CHAN_ERR_NOCHAN, CHAN_ERR_NODEV,    CHAN_ERR_DIFFDEV);

TQUB_OnAcquire = procedure(Sender :TObject; Buf :TQUB_AI_PBuffer; NotLast :Boolean) of object;

TQUB_Acquisition_Input                = class(TFrame)

  Label13                             :TLabel;

  btnRun                              :TSpeedButton;
  btnStop                             :TSpeedButton;
  btnRecord                           :TSpeedButton;
  btnPause                            :TSpeedButton;

  panDuration                         :TPanel;

  chkAcqModeCont                      :TRadioButton;
  chkAcqModeTime                      :TRadioButton;
  chkAcqModePoints                    :TRadioButton;
  txtAcqTimeHours                     :TEdit;
  udAcqTimeHours                      :TUpDown;
  txtAcqTimeMinutes                   :TEdit;
  udAcqTimeMinutes                    :TUpDown;
  txtAcqTimeSeconds                   :TEdit;
  udAcqTimeSeconds                    :TUpDown;
  txtAcqTimePoints                    :TEdit;

  lblRecordedTime                     :TLabel;
  panOverwrite                        :TPanel;

  btnProperties                       :TSpeedButton;

  lblSampling                         :TLabel;
  txtSampling                         :TEdit;
  lblBufferLength                     :TLabel;
  txtBufferLength                     :TEdit;

  lblDevice                           :TLabel;
  lblData                             :TLabel;
  cmbData                             :TComboBox;
  pbCache                             :TProgressBar;


  procedure txtSamplingKeyPress(Sender: TObject; var Key: Char);
  procedure txtSamplingExit(Sender: TObject);

  procedure txtBufferLengthKeyPress(Sender: TObject; var Key: Char);
  procedure txtBufferLengthExit(Sender: TObject);

  procedure chkAcqModeClick(Sender: TObject);
  procedure udAcqTimeHoursClick(Sender: TObject; Button: TUDBtnType);
  procedure udAcqTimeMinutesClick(Sender: TObject; Button: TUDBtnType);
  procedure udAcqTimeSecondsClick(Sender: TObject; Button: TUDBtnType);
  procedure txtAcqTimeHoursKeyPress(Sender: TObject; var Key: Char);
  procedure txtAcqTimeHoursExit(Sender: TObject);
  procedure txtAcqTimeMinutesKeyPress(Sender: TObject; var Key: Char);
  procedure txtAcqTimeMinutesExit(Sender: TObject);
  procedure txtAcqTimeSecondsKeyPress(Sender: TObject; var Key: Char);
  procedure txtAcqTimeSecondsExit(Sender: TObject);
  procedure txtAcqTimePointsKeyPress(Sender: TObject; var Key: Char);
  procedure txtAcqTimePointsExit(Sender: TObject);

  procedure btnRunClick(Sender: TObject);
  procedure btnStopClick(Sender: TObject);
  procedure btnPauseClick(Sender: TObject);
  procedure btnRecordClick(Sender: TObject);

  procedure btnPropertiesClick(Sender: TObject);

  procedure cmbDataDropDown(Sender: TObject);
  procedure cmbDataChange(Sender: TObject);

  procedure SetFocusClick(Sender: TObject);


protected


  prPropertiesForm            :TQUB_AcquisitionInput_Properties_Form;

  prScope                     :TQUB_QVO_QVS;
  prSpectrum                  :TQUB_QVO_QVSpectrum;
  prData                      :TQUB_QVO_QVD;

  prAIChannels                :TQUB_Acquisition_Channels;
  prDevice                    :TQUB_Acquisition_IDevice;
  prHardware                  :TQUB_Acquisition_Hardware;

  prSampling                  :Double;//Hz

  prAcqState                  :TQUB_Acquisition_Input_State;

  prDataError                 :TQUB_Acquisition_IDataError;
  prDeviceError               :TQUB_Acquisition_IDeviceError;
  prChannelsError             :TQUB_Acquisition_IChannelsError;

  prAcqMode                   :TQUB_Acquisition_Mode;
  prAcqTimeHours              :Integer;
  prAcqTimeMinutes            :Integer;
  prAcqTimeSeconds            :Integer;
  prAcqTimePoints             :Integer;

  prPointsRecorded            :Integer;
  prPointsRecording           :Integer;
  prReadingData               :Boolean;
  prUpdatingRecordedPoints    :Boolean;
  prUpdateDisplay             :Boolean;

  prLockCount                 :Integer;

  prBufferLength              :Integer;//ms

  prStatus                    :TLabel;

  prBufferLock                :TCriticalSection;
  prBufferSupply              :TQUB_BufferSupply;
  prBufferQ                   :TQUB_BufferQueue;
  prDefBuffers                :Integer;
  prMaxBuffers                :Integer;
  flagAvail                   :Boolean;
  prPausingSlow               :Boolean;

  prAcquiredSel               :TQUB_FirstLast;
  prAcquiredNotLast           :Boolean;

  //Scripting
  prScripting                 :Boolean;
  prOperations                :TQUB_Acquisition_Operations;
  prCurrOperation             :TQUB_Acquisition_POperation;

  //Events
  prOnDataNew                 :TNotifyEvent;
  prOnDataOpen                :TNotifyEvent;
  prOnDataAssign              :TNotifyEvent;
  prOnAcquire                 :TQUB_OnAcquire; // for SShot


  procedure prUpdateChannels;

  procedure prPropertiesRegistry(Read :Boolean);


  //Acquisition State
  procedure prSetAcquisitionState(s :TQUB_Acquisition_Input_State);

  //Acquisition Mode
  procedure prSetAcquisitionMode(m :TQUB_Acquisition_Mode);

  //Acquisition Duration
  procedure prSetAcquisitionTimeHours(h :Integer);
  procedure prSetAcquisitionTimeMinutes(m :Integer);
  procedure prSetAcquisitionTimeSeconds(s :Integer);
  procedure prSetAcquisitionTimePoints(p :Integer);

  //Data
  procedure prSetData(d :TQUB_QVO_QVD);
  procedure prCheckChannelInfo;

  procedure prSetSampling(s :Double);
  function  prGetSampling :Double;

  procedure prSetBufferLength(b :Integer);
  function  prGetBufferLength :Integer;

  procedure prUpdateRecordedPoints;

  procedure prMsgPause(var Message: TMessage); message WM_QUB_ACQ_INPUT_PAUSE;
  procedure prMsgUpdateRecordedPoints(var Message: TMessage); message WM_QUB_ACQ_INPUT_UPDATERECORDEDPOINTS;

  procedure prMsgErr(var Message: TMessage); message WM_QUB_ACQ_INPUT_ERR;
  procedure prMsgAvail(var Message: TMessage); message WM_QUB_ACQ_INPUT_AVAIL;

  procedure prSetupBuffers;
  procedure prFreeBuffers;

  procedure prAcquisitionPause;
  procedure prAcquisitionRecord;

  function prGetLocked :Boolean;

  procedure prSetStatus;


public


  m_lSynchRun  :Boolean;      // TRUE if btnSynch started acquisition, FALSE if btnRun did.
  m_aBreaks    :TStringList;  // List of segment breakpoints


  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;


  procedure DoOnChangeHardware(Sender :TObject);
  procedure DoOnChangeChannels(Sender :TObject);

  procedure Lock;
  procedure UnLock;

  function AcqStart(Synchronized :Boolean; Scripting :Boolean; Operations :TQUB_Acquisition_Operations) :Boolean;
  function AcqStop :Boolean;
  procedure AcqStopApplyBreaks(i0, i1 : Integer);
  function AcqPause :Boolean;
  function AcqRecord :Boolean;

  procedure AcquisitionCallBack(Buffer :Pointer; BufferSize, BufferCount :Integer; Error :Integer);

  function NewInputAcq(const FileName :String) :Boolean;
  function OpenInputAcq(const FileName :String) :Boolean;
  function AssignInputAcq(AData :TQUB_QVO_QVD) :Boolean;


  //Lock
  property Locked                        :Boolean                         read prGetLocked;

  //Scope
  property Scope                         :TQUB_QVO_QVS                    read prScope                       write prScope;
  //Spectrum
  property Spectrum                      :TQUB_QVO_QVSpectrum             read prSpectrum                    write prSpectrum;

  //Data
  property Data                          :TQUB_QVO_QVD                    read prData                        write prSetData;
  property DataError                     :TQUB_Acquisition_IDataError     read prDataError;

  //Device
  property Device                        :TQUB_Acquisition_IDevice        read prDevice;

  //Hardware
  property Hardware                      :TQUB_Acquisition_Hardware       read prHardware;
  property AIChannels                    :TQUB_Acquisition_Channels       read prAIChannels;

  property Sampling                      :Double                          read prGetSampling                 write prSetSampling;
  property BufferLength                  :Integer                         read prGetBufferLength             write prSetBufferLength;

  property AcqState                      :TQUB_Acquisition_Input_State    read prAcqState                    write prSetAcquisitionState;

  property AcqMode                       :TQUB_Acquisition_Mode           read prAcqMode                     write prSetAcquisitionMode;
  property AcqTimeHours                  :Integer                         read prAcqTimeHours                write prSetAcquisitionTimeHours;
  property AcqTimeMinutes                :Integer                         read prAcqTimeMinutes              write prSetAcquisitionTimeMinutes;
  property AcqTimeSeconds                :Integer                         read prAcqTimeSeconds              write prSetAcquisitionTimeSeconds;
  property AcqTimePoints                 :Integer                         read prAcqTimePoints               write prSetAcquisitionTimePoints;

  //Data events
  property OnDataNew                     :TNotifyEvent                    read prOnDataNew                   write prOnDataNew;
  property OnDataOpen                    :TNotifyEvent                    read prOnDataOpen                  write prOnDataOpen;
  property OnDataAssign                  :TNotifyEvent                    read prOnDataAssign                write prOnDataAssign;

  property OnAcquire                     :TQUB_OnAcquire                  read prOnAcquire                   write prOnAcquire;

  //Status
  property Status                        :TLabel                          read prStatus                      write prStatus;


end;


implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Viewport_AcquisitionInput,
QUB_Viewport_Data,
QUB_Main;


procedure AcquisitionCallBack(Sender :TObject; Buffer :Pointer; BufferSize, BufferCount :Integer; Error :Integer); stdcall;
begin
  TQUB_Acquisition_Input(Sender).AcquisitionCallBack(Buffer, BufferSize, BufferCount, Error);
end;


//******************************************************************************
constructor TQUB_Acquisition_Input.Create(AOwner: TComponent);
begin

  inherited;

  Align            := alClient;

  prAcqState       := ACQ_INPUT_STATE_STOPPED;

  prData           := nil;
  prDevice         := nil;
  prHardware       := nil;

  prAIChannels     := TQUB_Acquisition_Channels.Create;
  prAIChannels.OwnChannels := False;

  prPropertiesRegistry(True{Read});

  //Set Properties Form
  prPropertiesForm := TQUB_AcquisitionInput_Properties_Form.Create(Self);
  prPropertiesForm.AIChannels := prAIChannels;

  prBufferLock     := TCriticalSection.Create;
  
  prUpdateChannels;
  prSetStatus;

  cmbData.ItemIndex := 2;

end;


destructor TQUB_Acquisition_Input.Destroy;
begin

  try

    if AcqState <> ACQ_INPUT_STATE_STOPPED then AcqStop;

    prBufferLock.Acquire;
    prFreeBuffers;
    prBufferLock.Release;
    prBufferLock.Free;

    prPropertiesRegistry(False{Read});

    prAIChannels.Free;

  finally

    inherited;

  end;

end;


//Events ***********************************************************************

//DoOnChangeHardware
procedure TQUB_Acquisition_Input.DoOnChangeHardware(Sender :TObject);
begin
  prUpdateChannels;
  prCheckChannelInfo;
  prSetStatus;
end;

//DoOnChangeChannels
procedure TQUB_Acquisition_Input.DoOnChangeChannels(Sender :TObject);
begin
  prUpdateChannels;
  prCheckChannelInfo;
  prSetStatus;
end;


//Lock *************************************************************************
function TQUB_Acquisition_Input.prGetLocked :Boolean;
begin
  Result := prLockCount > 0;
end;

procedure TQUB_Acquisition_Input.Lock;
begin
  Inc(prLockCount);
end;

procedure TQUB_Acquisition_Input.UnLock;
begin
  Dec(prLockCount);
  if prLockCount < 0 then prLockCount := 0;
end;

//Update Channels *************************************************************
procedure TQUB_Acquisition_Input.prUpdateChannels;
var
  ACh     :TQUB_Acquisition_AIChannel;
  c, i    :Integer;
begin

  //Remove Dead Channels *******************************************************
  for c := prAIChannels.Count - 1 downto 0 do begin
    i := QUBAcqAIChannels.IndexOfObject(prAIChannels.Objects[c]);
    if i < 0 then begin
      prAIChannels.Delete(c);
    end else begin
      prAIChannels.Strings[c] := QUBAcqAIChannels.Strings[i];
    end;
  end;

  //Set Device
  for c := 0 to prAIChannels.Count - 1 do begin
    ACh     := TQUB_Acquisition_AIChannel(prAIChannels.Objects[c]);
    i       := QUBAcqIDevices.IndexOf(ACh.DeviceName);
    if i < 0 then begin
      ACh.Device := nil;
    end else begin
      ACh.Device := TQUB_Acquisition_IDevice(QUBAcqIDevices.Objects[i]);
    end;
  end;

  //Error checking *************************************************************
  prDeviceError   := DEV_ERR_NONE;
  prChannelsError := CHAN_ERR_NONE;

  if prAIChannels.Count > 0 then begin
    prDevice        := TQUB_Acquisition_IDevice(TQUB_Acquisition_AIChannel(prAIChannels.Objects[0]).Device);
    if prDevice     =  nil then prChannelsError := CHAN_ERR_NODEV;
  end else begin
    prDevice        := nil;
    prChannelsError := CHAN_ERR_NOCHAN;
  end;

  if prDevice <> nil then begin
    for c := 0 to prAIChannels.Count - 1 do begin
      Ach := TQUB_Acquisition_AIChannel(prAIChannels.Objects[c]);
      if Ach.Device = nil then begin
        prDevice := nil;
        prChannelsError := CHAN_ERR_NODEV;
        Break;
      end else if Ach.Device <> prDevice then begin
        prDevice := nil;
        prChannelsError := CHAN_ERR_DIFFDEV;
      end;
    end;
  end;

  if prDevice = nil then begin
    prDeviceError := DEV_ERR_UNASS;
    prHardware    := nil;
  end else begin
    prHardware    := prDevice.Hardware;
  end;

  //Data
  prDataError := IDATA_ERR_NONE;
  if prData = nil then begin
    prDataError := DATA_ERR_UNASS;
  end else if prData.DataObject.ADChannelCount <> prAIChannels.Count then begin
    prDataError := DATA_ERR_DIFFCHAN;
    prData := nil;
  end else if (prDevice <> nil) and ((prData.DataObject.ADDataSize) <> ((prDevice.Hardware.InputDeviceParams[prDevice.Index].m_Bits + 7) div 8)) then begin
    prDataError := DATA_ERR_DIFFBITS;
    prData := nil;
  end;

  if (prDevice <> nil) and (prDataError = IDATA_ERR_NONE) then begin
    prData.Scaling := IntPower(2, prDevice.Hardware.InputDeviceParams[prDevice.Index].m_Bits) / prDevice.Hardware.InputDeviceParams[prDevice.Index].m_Range;
  end;

  AcqState        := AcqState;

end;


//******************************************************************************
function TQUB_Acquisition_Input.AcqStart(Synchronized :Boolean; Scripting :Boolean; Operations :TQUB_Acquisition_Operations) :Boolean;
var
  yMin, yMax :Integer;
  i, p       :Integer;
  ErrCode    :Integer;
begin

  Result     := True;
  m_lSynchRun:= Synchronized;
  if m_lSynchRun then begin
    m_aBreaks := TStringList.Create;
  end;

  if Locked then Exit;
  if (prChannelsError <> CHAN_ERR_NONE) or (prDeviceError <> DEV_ERR_NONE) then Exit;
  if (prDataError = IDATA_ERR_NONE) and (Data.LockedAny) then Exit;

  Result     := False;

  //Lock, itself and Data
  Lock;
  if prDataError = IDATA_ERR_NONE then begin
    Data.LockRecord;
  end;

  //Set Device Properties
  prDevice.ASampling        := prSampling;
  prDevice.Caller           := Self;
  prDevice.ACallBack        := @QUB_AcquisitionInput.AcquisitionCallBack;
  prDevice.BufferLength     := BufferLength;
  prDevice.AChannelCount    := prAIChannels.Count;
  for i := 0 to prAIChannels.Count - 1 do begin
    prDevice.AChannel[i]    := TQUB_Acquisition_AIChannel(prAIChannels.Objects[i]).Channel;
    prDevice.AGain[i]       := TQUB_Acquisition_AIChannel(prAIChannels.Objects[i]).Gain;
  end;

  prPointsRecorded   := 0;

  PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_UPDATERECORDEDPOINTS, 0, 0);

  panOverWrite.Visible     := False;

  case prAcqMode of
    ACQ_MODE_CONT://Continuous
      begin
        prPointsRecording := High(Integer);
      end;
    ACQ_MODE_TIME://Time
      begin
        prPointsRecording := Round(prSampling * (prAcqTimeHours * 3600 + prAcqTimeMinutes * 60 + prAcqTimeSeconds));
      end;
    ACQ_MODE_POINTS://Points
      begin
        prPointsRecording := prAcqTimePoints;
      end;
  end;//case AcquisitionMode

  if Scripting then begin
    prScripting := True;
    prOperations := Operations;
    prCurrOperation := nil;
    prPointsRecording := 0;
  end else begin
    prScripting := False;
  end;

  //Scope/Spectrum
  if prUpdateDisplay then begin

    p := Round(IntPower(2, prDevice.Hardware.InputDeviceParams[prDevice.Index].m_Bits));
    if prDevice.Hardware.InputDeviceParams[prDevice.Index].m_Differential then begin
      yMin := -(p div 2);
      yMax := -yMin - 1;
    end else begin
      yMin := 0;
      yMax := p - 1;
    end;

    for i := 0 to prAIChannels.Count - 1 do begin
      Scope.Scaling[i]    := TQUB_Acquisition_AIChannel(prAIChannels.Objects[i]).Gain *
                             TQUB_Acquisition_AIChannel(prAIChannels.Objects[i]).Scaling 
                             {* p / prDevice.Hardware.RangeInput[prDevice.Index]};
      Spectrum.Scaling[i] := TQUB_Acquisition_AIChannel(prAIChannels.Objects[i]).Gain *
                             TQUB_Acquisition_AIChannel(prAIChannels.Objects[i]).Scaling
    end;

    Scope.StartDisplay(prAIChannels.Count, prHardware.InputDeviceParams[prDevice.Index].m_Range, prHardware.InputDeviceParams[prDevice.Index].m_Differential, prHardware.InputDeviceParams[prDevice.Index].m_Bits, prSampling, yMin, yMax);
    Spectrum.StartDisplay(prAIChannels.Count, prHardware.InputDeviceParams[prDevice.Index].m_Range, prHardware.InputDeviceParams[prDevice.Index].m_Differential, prHardware.InputDeviceParams[prDevice.Index].m_Bits, prSampling, yMin, yMax);

  end;

  prSetupBuffers;

  if (not Synchronized) then begin
    ErrCode := prDevice.Start;
    if ErrCode <> 0 then begin

      Scope.StopDisplay;
      Spectrum.StopDisplay;

      UnLock;
      if prDataError = IDATA_ERR_NONE then begin
        Data.UnLockRecord;
      end;

      prFreeBuffers;

      if prScripting then begin
        Console.NewEntry(Format('Cannot start acquisition: code %d', [ErrCode]), COLOR_ERROR);
      end else begin
        MessageHelpDlg(Format('Cannot start acquisition: code %d', [ErrCode]), mtError, [mbOK], 0, 'Error:Cannot_Start_Acquisition');
      end;

      prScripting := False;

      Result := True;

    end else begin
      AcqState := ACQ_INPUT_STATE_RUNNING;
    end;
  end else begin
    AcqState := ACQ_INPUT_STATE_RUNNING;
  end;

end;

//----- Binary divide scheme for applying breaks to minimize time (hopefully)
procedure TQUB_Acquisition_Input.AcqStopApplyBreaks( i0, i1 : Integer );
var
  i, iBreak : Integer;
begin
  if i1=i0 then begin
    iBreak := StrToInt(m_aBreaks[i0]);
    Data.DataObject.BreakSegment(iBreak, iBreak);
  end else if i1>i0 then begin
    i := (i1+i0) Div 2;
    iBreak := StrToInt(m_aBreaks[i]);
    Data.DataObject.BreakSegment(iBreak, iBreak);
    if i > i0 Then
      AcqStopApplyBreaks(i0,i-1);
    AcqStopApplyBreaks(i+1,i1);
  end;
end;

function TQUB_Acquisition_Input.AcqStop :Boolean;
var
  iLast : Integer;
begin
  Result := True;

  if (prDevice = nil) then Exit;

  Result := False;

  if not m_lSynchRun then begin
    if prDevice.Stop then Exit;
  end else begin
    prDevice.Started := False;
    if (m_aBreaks<>nil) Then Begin
      iLast := m_aBreaks.Count-1;
      while (iLast>0) and (StrToInt(m_aBreaks[iLast])>=Data.DataPointCount) do
        Dec(iLast);
      if iLast > 0 Then
        AcqStopApplyBreaks(0, iLast );
      m_aBreaks.Free();
      m_aBreaks:=nil;
    end;
  end;

  prBufferLock.Acquire;

  prAcqState := ACQ_INPUT_STATE_STOPPED;
  prFreeBuffers;

  prBufferLock.Release;

  if prUpdateDisplay then begin
    Scope.StopDisplay;
    Spectrum.StopDisplay;
  end;

  UnLock;
  if prDataError = IDATA_ERR_NONE then begin
    Data.UnLockRecord;
  end;

  prScripting := False;

  if prCurrOperation <> nil then begin
    Dispose(prCurrOperation);
    prCurrOperation := nil;
  end;

  AcqState := ACQ_INPUT_STATE_STOPPED;

end;

function TQUB_Acquisition_Input.AcqPause :Boolean;
begin
  Result := False;
  if AcqState <> ACQ_INPUT_STATE_RECORDING then Exit;
  prAcquisitionPause;
end;

function TQUB_Acquisition_Input.AcqRecord :Boolean;
begin
  Result := False;
  if AcqState = ACQ_INPUT_STATE_RECORDING then Exit;
  if (prAcqState <> ACQ_INPUT_STATE_RUNNING) and (prAcqState <> ACQ_INPUT_STATE_PAUSED) then Exit;
  prAcquisitionRecord;
end;

procedure TQUB_Acquisition_Input.prMsgPause(var Message: TMessage);
begin
  btnPauseClick(btnPause);
end;

procedure TQUB_Acquisition_Input.prAcquisitionPause;
begin
  prAcqState := ACQ_INPUT_STATE_RUNNING;
  AcqState   := ACQ_INPUT_STATE_PAUSED;
end;

procedure TQUB_Acquisition_Input.prAcquisitionRecord;
begin
  AcqState := ACQ_INPUT_STATE_RECORDING;
  prPausingSlow := False;
  pbCache.Visible := True;
end;

procedure TQUB_Acquisition_Input.prMsgErr(var Message: TMessage);
begin
  panOverwrite.Visible := Message.WParam <> 0;
end;

procedure TQUB_Acquisition_Input.prMsgAvail(var Message: TMessage);
var
  Pending    :Integer;
  Buf        :TQUB_AI_PBuffer;
  i          :Integer;
  PauseMsg   :Boolean;
begin
  if prBufferQ = nil then Exit;

  Pending    := prBufferQ.Count;
  pbCache.Position := Round((100.0 * Pending) / prMaxBuffers);

  // If too many buffers pile up, acq in will be paused.
  // You only want to show the message once,
  // but this function is called several times to exhaust the queue.
  PauseMsg   := False;             // "show the message this time"
  if not prPausingSlow then begin  // only if we haven't already
    prPausingSlow := Pending >= prMaxBuffers;  // and the q is huge
    if prPausingSlow then begin
      AcqPause;
      Beep;
      PauseMsg := True;
    end;
  end;                                 // but first let's process some buffers

  i          := 0;
  while (i < MAX_BUF_AT_ONCE) do begin
    Inc(i);
    Buf      := prBufferQ.Get(BUFQ_TIMEOUT);
    if Buf = nil then Break;

    try
      //Scope/Spectrum
      if prUpdateDisplay then begin
        prScope.AddData(Buf.Buffer, Buf.BufferSize, Buf.BufferCount);
        prSpectrum.AddData(Buf.Buffer, Buf.BufferSize, Buf.BufferCount);
      end;

      if Buf.Recording then Data.AppendData(Buf.Buffer, Buf.BufferSize, Buf.BufferCount, Buf.NewSegment, Buf.StartTime, Buf.Refresh and (Pending <= 2), True{Acq});

      if Assigned(OnAcquire) then OnAcquire(Self, Buf, Boolean(Pending-i));

    except
      on ex :Exception do begin
        prPausingSlow := True;
        AcqPause;
        Beep;
        MessageHelpDlg('Acquisition Input was paused because an exception occurred:'#13#10+ex.Message, mtError, [mbOK], 0, 'Warning:Acquisition_Input_Paused');
      end;
    end;

    prBufferSupply.Put(Buf);
    pbCache.Position := Round((100.0 * prBufferQ.Count) / prMaxBuffers);
  end;

  // as a side-effect of flagAvail, buffer processing pauses during this message:

  if PauseMsg then begin
    MessageHelpDlg('Acquisition Input was paused because it was running too slowly.  Try closing other programs or displaying fewer points of data.', mtError, [mbOK], 0, 'Warning:Acquisition_Input_Paused');
  {end else if (not prRunningSlow) and (Pending >= prDefBuffers) then begin
    prRunningSlow := True; // so this message doesn't show again
    PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_AVAIL, 0, 0);
    MessageDlg('Acquisition Input is running slowly.  If this continues it will be stopped.  Try closing other programs or displaying fewer points of data.', mtWarning, [mbOK], 0);
  end else if Pending < prDefBuffers then begin
    if Pending <= 1 then begin
      prRunningSlow := False;
    end; }
  end;

  Pending := prBufferQ.Count;
  if Pending > 0 then begin
    PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_AVAIL, 0, 0);
  end else begin
    flagAvail := False; // really means: "There's no _AVAIL message in the pump,
                        // so if you queue something please post a message."
  end;
end;

procedure TQUB_Acquisition_Input.prSetupBuffers;
var
  BufferCount :Integer;
  Bytes       :Integer;
begin
  prFreeBuffers;
  // determine max buffer size the same way e.g. NIMX driver does:
  BufferCount    := (Round(prSampling*BufferLength/1000)+511) div 512;
  if BufferCount > 2048 then begin
    BufferCount  := 2048 * 512;
  end else begin
    BufferCount  := BufferCount * 512;
  end;
  Bytes          := BufferCount*prDevice.AChannelCount*SizeOf(SmallInt);
  prDefBuffers   := (DEF_SPARE_BYTES div Bytes) + 1;
  prMaxBuffers   := (MAX_SPARE_BYTES div Bytes) + 1;
  prBufferSupply := TQUB_BufferSupply.Create(prDefBuffers, prMaxBuffers, Bytes);
  prBufferQ      := TQUB_BufferQueue.Create;
  flagAvail      := False;
end;

procedure TQUB_Acquisition_Input.prFreeBuffers;
begin
  if prBufferSupply <> nil then begin
    prBufferQ.Clear(prBufferSupply);
    prBufferQ.Free;
    prBufferQ := nil;
    prBufferSupply.Free;
    prBufferSupply := nil;
    pbCache.Visible := False;
  end;
end;

procedure TQUB_Acquisition_Input.AcquisitionCallBack(Buffer :Pointer; BufferSize, BufferCount :Integer; Error :Integer);
const
  APPENDWPARAM = DATA_MSG_DATA_EDIT or DATA_MSG_RECOUNT_DATA or DATA_MSG_BROWSE     or DATA_MSG_READ_CACHE or
                 DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA    or DATA_MSG_SET_SCROLL or DATA_MSG_COND_BROWSE;
var
  //Seg          :TQUB_QFS_Segment;
  StartTime    :Double;
  BStartTime   :Double;
  BuffFirst    :Integer;
  Sel          :TQUB_FirstLast;
  RecPoints    :Integer;
  Buf          :TQUB_AI_PBuffer;
begin

  prBufferLock.Acquire;
  try
   if prAcqState <> ACQ_INPUT_STATE_STOPPED then begin

    {TO DO : read Now at the begining of acquisition and calculate. LM}
    BStartTime := Now - (1/SecsPerDay)*(BufferCount/prSampling);


    try

      if (prReadingData) or (Error <> 0) then begin
        // Note : panOverWrite is a hidden red block with text 'Err' between the
        //        Play/Record buttons and [Channels] buton
        //panOverWrite.Visible := True;
        PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_ERR, 1{True}, 0);

      end else begin

        prReadingData := True;

        try

          if (prDataError = IDATA_ERR_NONE) or (prDataError = DATA_ERR_UNASS) then begin

            if BufferCount > 0 then begin


              //******************************************************************//
              //                                                                  //
              //                        S C R I P T I N G                         //
              //                                                                  //
              //******************************************************************//
              if prScripting then begin

                BuffFirst := 0;

                //Get the current operation
                if prCurrOperation = nil then begin

                  prCurrOperation := prOperations.Pop;
                  if prCurrOperation = nil then begin
                    prOperations.Stopped := True;
                    Exit;
                  end else begin
                    prOperations.Stopped := False;
                  end;

                end;

                prOperations.Stopped := prCurrOperation = nil;

                Sel.First := BuffFirst;

                RecPoints := 0;

                repeat

                  Sel.Last := Sel.First + prCurrOperation.Points - 1;
                  if Sel.Last >= BufferCount then Sel.Last := BufferCount - 1;

                  if  not prCurrOperation.Started then begin
                    //Seg := Data.DataObject.Segments[Data.DataObject.Segments.SegmentCount - 1];
                    //StartTime := Seg.StartTime + Seg.Channels.DataCount / prSampling;
                    //Use the real time, at the beginning of the Operation. LM
                    StartTime := BStartTime + (1/SecsPerDay)*(Sel.First/prSampling);
                  end else begin
                    StartTime := 0;
                  end;

                  Buf               := prBufferSupply.Get();
                  Buf.Recording     := prCurrOperation.Recording;
                  Buf.BufferSize    := BufferSize;
                  Buf.BufferCount   := Sel.Last - Sel.First + 1;
                  Buf.NewSegment    := not prCurrOperation.Started;
                  Buf.StartTime     := StartTime;
                  Buf.Refresh       := False;
                  Move(TPAByte(Buffer)[Sel.First * BufferSize], TPAByte(Buf.Buffer)[0], Buf.BufferSize*Buf.BufferCount);
                  prBufferQ.Put(Buf);
                  if not flagAvail then begin // if there's not a posted _AVAIL message already
                    flagAvail       := True;
                    PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_AVAIL, 0, 0);
                  end;
                  Inc(RecPoints, Sel.Last - Sel.First + 1);

                  prCurrOperation.Started := True;
                  Dec(prCurrOperation.Points, Sel.Last - Sel.First + 1);
                  if prCurrOperation.Points <= 0 then begin
                    Dispose(prCurrOperation);
                    prCurrOperation := prOperations.Pop;
                  end;

                  Sel.First := Sel.Last + 1;

                until ((Sel.First >= BufferCount) or (prCurrOperation = nil));

                if prCurrOperation.Recording and (RecPoints > 0) then begin
                  Inc(prPointsRecorded, RecPoints);
                  Data.DataMessage(APPENDWPARAM, -1);
                  //Post message to update recorded points labels
                  PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_UPDATERECORDEDPOINTS, 0, 0);
                end;

                prOperations.Stopped := prCurrOperation = nil;


              //******************************************************************//
              //                                                                  //
              //                  N O N  -   S C R I P T I N G                    //
              //                                                                  //
              //******************************************************************//
              end else begin

                Buf               := prBufferSupply.Get();
                Buf.Recording := (AcqState = ACQ_INPUT_STATE_RECORDING);

                if Buf.Recording then begin
                  if prPointsRecorded < prPointsRecording then begin
                    if BufferCount > (prPointsRecording - prPointsRecorded) then begin
                      BufferCount := prPointsRecording - prPointsRecorded;
                    end;
                  end else begin
                    Buf.Recording := False;
                    //Post message to Pause
                    PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_PAUSE, 0, 0);
                  end;
                end;

                if prPointsRecorded = 0 then begin
                  //Seg := Data.DataObject.Segments[Data.DataObject.Segments.SegmentCount - 1];
                  //Seg.StartTime + Seg.Channels.DataCount / prSampling;
                  //Use the real time, at the beginning of the Buffer. LM
                  StartTime := Now - (1/SecsPerDay)*(BufferCount/prSampling);
                end else begin
                  StartTime := 0;
                end;

                Buf.BufferSize    := BufferSize;
                Buf.BufferCount   := BufferCount;
                Buf.NewSegment    := prPointsRecorded = 0;
                Buf.StartTime     := StartTime;
                Buf.Refresh       := True;
                Move(TPAByte(Buffer)[0], TPAByte(Buf.Buffer)[0], Buf.BufferSize*Buf.BufferCount);

                if Buf.Recording then begin
                  Inc(prPointsRecorded, BufferCount);
                  //Post message to update recorded points labels
                  PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_UPDATERECORDEDPOINTS, 0, 0);
                end;

                prBufferQ.Put(Buf);
                if not flagAvail then begin // if there's not a posted _AVAIL message already
                  flagAvail       := True;
                  PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_AVAIL, 0, 0);
                end;

              end;

            end;

          end;//DataError = NONE

        finally

          prReadingData := False;

        end;

      end;

      {
      // Error prone - possibly due to being run on thread ?
      // possibly due to taking longer time than one buffer of output
      if m_lSynchRun and (m_aBreaks<>nil) and (m_aBreaks.Count>0) Then Begin
        iBreak := StrToInt(m_aBreaks[0]);
        if ( iBreak <= Data.DataPointCount ) Then
          Data.DataObject.BreakSegment(iBreak, iBreak);
        m_aBreaks.Delete(0);
      end;
      }

    except
      //panOverWrite.Visible := True;
      PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_ERR, 1{True}, 0);
    end;

   end;
  finally
    prBufferLock.Release;
  end;
end;

function TQUB_Acquisition_Input.NewInputAcq(const FileName :String) :Boolean;
begin

  Result := True;

  if AcqState <> ACQ_INPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  //Instruct ViewportData to create a new data file and associate with it
  if not ViewportData.New(FileName, False{Silent}) then begin
    Result := AssignInputAcq(ViewportData.CurrData);
  end else begin
    AssignInputAcq(prData);
  end;

end;

function TQUB_Acquisition_Input.OpenInputAcq(const FileName :String) :Boolean;
begin

  Result := True;

  if AcqState <> ACQ_INPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  //Instruct ViewportData to create a new data file and associate with it
  if not ViewportData.Open(FileName, False{Silent}) then begin
    Result := AssignInputAcq(ViewportData.CurrData);
  end else begin
    AssignInputAcq(prData);
  end;

end;

function TQUB_Acquisition_Input.AssignInputAcq(AData :TQUB_QVO_QVD) :Boolean;
begin

  Result := True;

  if AcqState <> ACQ_INPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  if AData = nil then begin
    Data   := AData;
    Result := False;
  end else if AData.LockedAny then begin
    Exit; //Data was created but is, somehow, locked
  end else begin
    Data   := AData;
    Result := False;
  end;

end;


procedure TQUB_Acquisition_Input.prUpdateRecordedPoints;
var
  h, m, s :Integer;
begin
  if prUpdatingRecordedPoints then Exit;

  prUpdatingRecordedPoints := True;

  try

    try
      s := Round(prPointsRecorded / prSampling);
      //TO DO: John replaced the Trunc()... with div
      //I'm not sure if I had a reason for this.... LM
      //h := Trunc(s / 3600);
      h := s Div 3600;
      s := s - h *3600;
      m := s Div 60;
      s := s - m * 60;
      lblRecordedTime.Caption := Format('%d:%d:%d / %d', [h, m, s, prPointsRecorded]);
    except
    end;

  finally

    prUpdatingRecordedPoints := False;

  end;

end;

procedure TQUB_Acquisition_Input.prMsgUpdateRecordedPoints(var Message: TMessage);
begin
  prUpdateRecordedPoints;
end;

//Acquisition State
procedure TQUB_Acquisition_Input.prSetAcquisitionState(s :TQUB_Acquisition_Input_State);
var
  e :Boolean;
begin

  while prReadingData do begin
  end;

  prReadingData := True;

  try

    case s of

      ACQ_INPUT_STATE_RUNNING:
        begin
          if ((prDevice <> nil)) then begin
            s := ACQ_INPUT_STATE_RUNNING;
          end else begin
            s := ACQ_INPUT_STATE_STOPPED;
          end;
          btnRun.Enabled     := False;
          btnStop.Enabled    := (s = ACQ_INPUT_STATE_RUNNING) and (not prScripting);
          btnPause.Enabled   := False;
          btnRecord.Enabled  := (s = ACQ_INPUT_STATE_RUNNING) and (prDataError = IDATA_ERR_NONE) and (not prScripting);
          btnRecord.Down     := False;
        end;

      ACQ_INPUT_STATE_STOPPED:
        begin
          btnRun.Enabled     := (prDevice <> nil) and (not prScripting);
          btnStop.Enabled    := False;
          btnPause.Enabled   := False;
          btnRecord.Enabled  := False;
          btnRecord.Down     := False;
        end;

      ACQ_INPUT_STATE_PAUSED:
        begin
          btnRun.Enabled     := False;
          btnStop.Enabled    := (not prScripting) and (not m_lSynchRun);
          btnPause.Enabled   := False;
          btnRecord.Enabled  := (prDataError = IDATA_ERR_NONE) and (not prScripting) and (not m_lSynchRun);
          btnRecord.Down     := False;
        end;

      ACQ_INPUT_STATE_RECORDING:
        begin
          prPointsRecorded   := 0;
          PostMessage(Self.Handle, WM_QUB_ACQ_INPUT_UPDATERECORDEDPOINTS, 0, 0);
          panOverWrite.Visible := False;
          btnRun.Enabled     := False;
          btnStop.Enabled    := (not prScripting) and (not m_lSynchRun);
          btnPause.Enabled   := (not prScripting) and (not m_lSynchRun);
          btnRecord.Enabled  := (not prScripting) and (not m_lSynchRun);
          btnRecord.Down     := True;

        end;

    end;//case AcqState

    //Set the Data Recording flag
    if prDataError = IDATA_ERR_NONE then begin
      Data.Recording := (s = ACQ_INPUT_STATE_RECORDING);
    end;

    e                             := (s = ACQ_INPUT_STATE_STOPPED) and (not prScripting);

    chkAcqModeCont.Enabled        := e;
    chkAcqModeTime.Enabled        := e;
    chkAcqModePoints.Enabled      := e;
    txtAcqTimeHours.Enabled       := e;
    udAcqTimeHours.Enabled        := e;
    txtAcqTimeMinutes.Enabled     := e;
    udAcqTimeMinutes.Enabled      := e;
    txtAcqTimeSeconds.Enabled     := e;
    udAcqTimeSeconds.Enabled      := e;
    txtAcqTimePoints.Enabled      := e;

    txtBufferLength.Enabled       := e;
    lblBufferLength.Enabled       := e;
    txtSampling.Enabled           := e;
    lblSampling.Enabled           := e;

    btnProperties.Enabled         := e;

    cmbData.Enabled               := e;

    prAcqState                    := s;

  finally

    prReadingData := False;

  end;

end;


//****************************************************************************//
//                                                                            //
//                                   D A T A                                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Acquisition_Input.prSetData(d :TQUB_QVO_QVD);
begin

  prData           := d;

  prUpdateChannels;

  if prData <> nil then begin
    prSampling     := 1 / prData.Sampling;
  end;

  txtSampling.Text := FloatToStr(prSampling / 1000);

  if prData = nil then begin
    cmbData.ItemIndex := 2;
  end else begin
    cmbData.Items.Append(ExtractFileName(d.FileName));
    cmbData.ItemIndex := cmbData.Items.Count - 1;
  end;

  prCheckChannelInfo;

  AcqState         := AcqState;

  prSetStatus;

end;

procedure TQUB_Acquisition_Input.prCheckChannelInfo;
var
  c          :Integer;
  ACh        :TQUB_Acquisition_AIChannel;
  DataEmpty  :Boolean;
begin
  if prData = nil then Exit;
  DataEmpty := prData.DataObject.Segments.TotalDataCount = 0;

  for c := 0 to prAIChannels.Count - 1 do begin
    ACh     := TQUB_Acquisition_AIChannel(prAIChannels.Objects[c]);
    if DataEmpty
    or (   (  (prData.DataChannelName[c] <> ACh.Name)
           or (prData.DataChannelUnits[c] <> ACh.Units)
           or (prData.DataChannelScaling[c] <> ACh.Scaling)
           or (prData.DataChannelOffset[c] <> 0.0)
           )
       and (mrYes = MessageHelpDlg(Format('Acq. input: Data channel %d [%s/%s/%f/%f] does not match DAQ channel [%s/%s/%f].  Change it?',
           [c, prData.DataChannelName[c], prData.DataChannelUnits[c], prData.DataChannelScaling[c], prData.DataChannelOffset[c], ACh.Name, ACh.Units, ACh.Scaling]),
           mtConfirmation, [mbYes, mbNo], 0, 'Warning:Data_Channel_Does_Not_Match'))
       )
    then begin
      prData.DataChannelName[c]     := ACh.Name;
      prData.DataChannelUnits[c]    := ACh.Units;
      prData.DataChannelScaling[c]  := ACh.Scaling;
      prData.DataChannelOffset[c]   := 0.0;
    end;
  end;
end;

//Sampling
procedure TQUB_Acquisition_Input.prSetSampling(s :Double);
begin
  prSampling       := s;
  txtSampling.Text := FloatToStr(prSampling / 1000);
end;

function  TQUB_Acquisition_Input.prGetSampling :Double;
begin
  Result := prSampling;
end;

//BufferLength
procedure TQUB_Acquisition_Input.prSetBufferLength(b :Integer);
begin
  prBufferLength := b;
  txtBufferLength.Text := IntToStr(BufferLength);
end;

function  TQUB_Acquisition_Input.prGetBufferLength :Integer;
begin
  Result := prBufferLength;
end;

//Acquisition Mode
procedure TQUB_Acquisition_Input.prSetAcquisitionMode(m :TQUB_Acquisition_Mode);
begin
  prAcqMode                := m;
  chkAcqModeCont.Checked   := Integer(prAcqMode) = chkAcqModeCont.Tag;
  chkAcqModeTime.Checked   := Integer(prAcqMode) = chkAcqModeTime.Tag;
  chkAcqModePoints.Checked := Integer(prAcqMode) = chkAcqModePoints.Tag;
end;

//Acquisition Duration
procedure TQUB_Acquisition_Input.prSetAcquisitionTimeHours(h :Integer);
begin
  prAcqTimeHours := h;
  udAcqTimeHours.Position := prAcqTimeHours;
end;

procedure TQUB_Acquisition_Input.prSetAcquisitionTimeMinutes(m :Integer);
begin
  prAcqTimeMinutes := m;
  udAcqTimeMinutes.Position := prAcqTimeMinutes;
end;

procedure TQUB_Acquisition_Input.prSetAcquisitionTimeSeconds(s :Integer);
begin
  prAcqTimeSeconds := s;
  udAcqTimeSeconds.Position := prAcqTimeSeconds;
end;

procedure TQUB_Acquisition_Input.prSetAcquisitionTimePoints(p :Integer);
begin
  prAcqTimePoints := p;
  txtAcqTimePoints.Text := IntToStr(prAcqTimePoints);
end;

//Status
procedure TQUB_Acquisition_Input.prSetStatus;
var
  t :String;
begin
  t := '';
  if prDeviceError <> DEV_ERR_NONE then begin
    if t <> '' then t := t + #13#10;
    t := t + 'Acq Device not assigned -> Options.Acquisition Hardware / Channels.';
  end;
  if prChannelsError <> CHAN_ERR_NONE then begin
    if t <> '' then t := t + #13#10;
    case prChannelsError of
      CHAN_ERR_NOCHAN:  t := t + 'No Acq Channels are assigned -> Channels or Options.Acquisition Channels.';
      CHAN_ERR_NODEV:   t := t + 'One or more Acq Channels have no Device assigned -> Options.Acquisition Hardware / Channels.';
      CHAN_ERR_DIFFDEV: t := t + 'One or more Acq Channels have different Devices (must be identical) -> Options.Acquisition Channels.';
    end;
  end;
  if prDataError <> IDATA_ERR_NONE then begin
    if t <> '' then t := t + #13#10;
    case prDataError of
      DATA_ERR_UNASS:    t := t + 'Data file not assigned for recording -> File.NewDataInputAcq.';
      DATA_ERR_DIFFCHAN: t := t + 'Data file has different Acq Channel Count than Acq Channels -> File.NewDataInputAcq or Channels.';
      DATA_ERR_DIFFBITS: t := t + 'Data file has different Bits Resolution than Acq Device -> File.NewDataInputAcq or Options.Acquisition Hardware.';
    end;
  end;
  if Status <> nil then begin
    Status.Visible  := t <> '';
    Status.Caption  := t;
    Status.ShowHint := True;
    Status.Hint     := t;
  end;

  if lblDevice <> nil then begin
    if prDevice = nil then begin
      lblDevice.Caption := 'Device: <none>';
    end else begin
      lblDevice.Caption := 'Device: ' + prDevice.Name;
    end;
  end;

end;


//****************************************************************************//
//                                                                            //
//                                 E V E N T S                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Acquisition_Input.txtSamplingKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtSamplingExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Input.txtSamplingExit(Sender: TObject);
var
  s :Double;
begin
  try
    s := StrToFloat(TEdit(Sender).Text) * 1000;
    if (s < (MIN_SAMPLING / 1000)) or (s > (MAX_SAMPLING / 1000)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_SAMPLING / 1000) + ' and ' + FloatToStr(MAX_SAMPLING / 1000), mtError, [mbOK], 0);
    Sampling := Sampling;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if s <> Sampling then begin
    Sampling := s;
  end;
end;

procedure TQUB_Acquisition_Input.txtBufferLengthKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtBufferLengthExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Input.txtBufferLengthExit(Sender: TObject);
var
  b :Integer;
begin
  try
    b := StrToInt(TEdit(Sender).Text);
    if (b < MIN_BUFFERLENGTH) or (b > MAX_BUFFERLENGTH) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_BUFFERLENGTH) + ' and ' + IntToStr(MAX_BUFFERLENGTH), mtError, [mbOK], 0);
    BufferLength := BufferLength;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if b <> BufferLength then begin
    BufferLength := b;
  end;
end;

procedure TQUB_Acquisition_Input.chkAcqModeClick(Sender: TObject);
begin
  AcqMode := TQUB_Acquisition_Mode(TRadioButton(Sender).Tag);
end;

procedure TQUB_Acquisition_Input.udAcqTimeHoursClick(Sender: TObject; Button: TUDBtnType);
begin
  AcqTimeHours := TUpDown(Sender).Position;
end;

procedure TQUB_Acquisition_Input.udAcqTimeMinutesClick(Sender: TObject; Button: TUDBtnType);
begin
  AcqTimeMinutes := TUpDown(Sender).Position;
end;

procedure TQUB_Acquisition_Input.udAcqTimeSecondsClick(Sender: TObject; Button: TUDBtnType);
begin
  AcqTimeSeconds := TUpDown(Sender).Position;
end;

procedure TQUB_Acquisition_Input.txtAcqTimePointsKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtAcqTimePointsExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Input.txtAcqTimeHoursKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtAcqTimeHoursExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Input.txtAcqTimeHoursExit(Sender: TObject);
var
  p :Integer;
begin
  try
    p := StrToInt(TEdit(Sender).Text);
    if (p < udAcqTimeHours.Min) or (p > udAcqTimeHours.Max) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(udAcqTimeHours.Min) + ' and ' + IntToStr(udAcqTimeHours.Max), mtError, [mbOK], 0);
    AcqTimeHours := AcqTimeHours;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> AcqTimeHours then begin
    AcqTimeHours := p;
  end;
end;

procedure TQUB_Acquisition_Input.txtAcqTimeMinutesKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtAcqTimeMinutesExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Input.txtAcqTimeMinutesExit(Sender: TObject);
var
  p :Integer;
begin
  try
    p := StrToInt(TEdit(Sender).Text);
    if (p < udAcqTimeMinutes.Min) or (p > udAcqTimeMinutes.Max) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(udAcqTimeMinutes.Min) + ' and ' + IntToStr(udAcqTimeMinutes.Max), mtError, [mbOK], 0);
    AcqTimeMinutes := AcqTimeMinutes;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> AcqTimeMinutes then begin
    AcqTimeMinutes := p;
  end;
end;

procedure TQUB_Acquisition_Input.txtAcqTimeSecondsKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtAcqTimeSecondsExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Input.txtAcqTimeSecondsExit(Sender: TObject);
var
  p :Integer;
begin
  try
    p := StrToInt(TEdit(Sender).Text);
    if (p < udAcqTimeSeconds.Min) or (p > udAcqTimeSeconds.Max) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(udAcqTimeSeconds.Min) + ' and ' + IntToStr(udAcqTimeSeconds.Max), mtError, [mbOK], 0);
    AcqTimeSeconds := AcqTimeSeconds;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> AcqTimeSeconds then begin
    AcqTimeSeconds := p;
  end;
end;

procedure TQUB_Acquisition_Input.txtAcqTimePointsExit(Sender: TObject);
var
  p :Integer;
begin
  try
    p := StrToInt(TEdit(Sender).Text);
    if (p < MIN_TIMEPOINTS){ or (p > MAX_TIMEPOINTS)} then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_TIMEPOINTS) + ' and ' + IntToStr(MAX_TIMEPOINTS), mtError, [mbOK], 0);
    AcqTimePoints := AcqTimePoints;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> AcqTimePoints then begin
    AcqTimePoints := p;
  end;
end;

procedure TQUB_Acquisition_Input.btnRunClick(Sender: TObject);
begin
  AcqStart(False{Synch}, False{Scripting}, nil);
end;

procedure TQUB_Acquisition_Input.btnStopClick(Sender: TObject);
begin
  AcqStop;
end;

procedure TQUB_Acquisition_Input.btnPauseClick(Sender: TObject);
begin
  AcqPause;
end;

procedure TQUB_Acquisition_Input.btnRecordClick(Sender: TObject);
begin
  if AcqState = ACQ_INPUT_STATE_RECORDING then begin
    btnRecord.Down := True;
    Exit;
  end;
  AcqRecord;
end;

procedure TQUB_Acquisition_Input.btnPropertiesClick(Sender: TObject);
begin
  prPropertiesForm.ShowModal;

  prUpdateChannels;
  prCheckChannelInfo;

  prSetStatus;
end;

procedure TQUB_Acquisition_Input.cmbDataDropDown(Sender: TObject);
var
  i                     :Integer;
begin
  while cmbData.Items.Count > 3 do cmbData.Items.Delete( cmbData.Items.Count - 1 );

  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    cmbData.Items.Append(ExtractFileName(ViewportData.ListObjects[i].FileName));
    if prData = ViewportData.ListObjects[i] then begin
      cmbData.ItemIndex := cmbData.Items.Count - 1;
    end;
  end;
end;

procedure TQUB_Acquisition_Input.cmbDataChange(Sender: TObject);
begin
  case cmbData.ItemIndex of
    0:      ViewportAcqInput.NewInputAcq('');
    1:      ViewportAcqInput.OpenInputAcq('');
    2:      ViewportAcqInput.AssignInputAcq(nil);
    else    ViewportAcqInput.AssignInputAcq(TQUB_QVO_QVD(ViewportData.ListObjects[cmbData.ItemIndex - 3]));
  end;
end;


//Registry *********************************************************************
procedure TQUB_Acquisition_Input.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
  Ch  :TQUB_Acquisition_Channel;
  Idx :Integer;
  c   :Integer;
begin
  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Acquisition Input', True) then begin

        if Read then begin

          prUpdateDisplay := ReadRegistryBoolean(Reg, 'Update Display',True);
          Sampling        := ReadRegistryFloat(Reg, 'Sampling',DEF_SAMPLING);
          BufferLength    := ReadRegistryInteger(Reg, 'Buffer Length',DEF_BUFFERLENGTH);
          AcqMode         := TQUB_Acquisition_Mode(ReadRegistryInteger(Reg, 'Acquisition Mode',Integer(ACQ_MODE_CONT)));
          AcqTimeHours    := ReadRegistryInteger(Reg, 'Acquisition Time Hours',0);
          AcqTimeMinutes  := ReadRegistryInteger(Reg, 'Acquisition Time Minutes',0);
          AcqTimeSeconds  := ReadRegistryInteger(Reg, 'Acquisition Time Seconds',10);
          AcqTimePoints   := ReadRegistryInteger(Reg, 'Acquisition Time Points',10000);

        end else begin

          WriteBool('Update Display', prUpdateDisplay);
          WriteFloat('Sampling', prSampling);
          WriteInteger('Buffer Length', prBufferLength);
          WriteInteger('Acquisition Mode', Integer(AcqMode));
          WriteInteger('Acquisition Time Hours', AcqTimeHours);
          WriteInteger('Acquisition Time Minutes', AcqTimeMinutes);
          WriteInteger('Acquisition Time Seconds', AcqTimeSeconds);
          WriteInteger('Acquisition Time Points', AcqTimePoints);

        end;

      end;
      CloseKey;

      if Read then begin
      //Read

        //AIChannels
        c := 0;
        while OpenKey('Software\QUB\Acquisition Input\AIChannels\AIChannel ' + IntToStr(c + 1), False)
              and ValueExists('Name') do begin
          Idx := QUBAcqAIChannels.IndexOf(ReadString('Name'));
          if Idx >= 0 then begin
            Ch := TQUB_Acquisition_Channel(QUBAcqAIChannels.Objects[Idx]);
            prAIChannels.AddObject(Ch.Name, Ch);
          end;//Idx > 0
          CloseKey;
          Inc(c);
        end;

      end else begin
      //Write

        if OpenKey('Software\QUB\Acquisition Input', True) then begin
          DeleteKey('AIChannels');
        end;
        CloseKey;

        //AIChannels
        for c := 0 to prAIChannels.Count - 1 do begin
          Ch := TQUB_Acquisition_Channel(prAIChannels.Objects[c]);
          if OpenKey('Software\QUB\Acquisition Input\AIChannels\AIChannel ' + IntToStr(c + 1), True) then begin
            WriteString('Name', Ch.Name);
          end;
          CloseKey;
        end;

      end;//Read/Write

    end;//with Reg

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
procedure TQUB_Acquisition_Input.SetFocusClick(Sender: TObject);
begin
  SetFocus;
end;


//******************************************************************************
end.
