unit QUB_AcquisitionInput_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Forms, Controls, StdCtrls, ComCtrls, Buttons, Classes,

QUB_Global,
QUB_Acquisition_Hardware;


type


TQUB_AcquisitionInput_Properties_Form = class(TForm)


  lstInstAIChannels                   :TListBox;
  lstAvailAIChannels                  :TListBox;

  btnAddAI                            :TSpeedButton;
  btnDelAI                            :TSpeedButton;
  btnUpAI                             :TSpeedButton;
  btnDnAI                             :TSpeedButton;

  btnClose                            :TButton;
  Label5                              :TLabel;
  Label6                              :TLabel;
    btnHelp: TButton;


  procedure FormShow(Sender: TObject);

  procedure btnAddAIClick(Sender: TObject);
  procedure btnDelAIClick(Sender: TObject);
  procedure btnUpAIClick(Sender: TObject);
  procedure btnDnAIClick(Sender: TObject);

  procedure lstInstAIChannelsDblClick(Sender: TObject);
  procedure lstAvailAIChannelsDblClick(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);


protected


  procedure prAddAIChannel;
  procedure prDelAIChannel;
  procedure prUpAIChannel;
  procedure prDnAIChannel;


public


  AIChannels                         :TQUB_Acquisition_Channels;


end;


implementation


{$R *.DFM}


//AIChannels *******************************************************************
procedure TQUB_AcquisitionInput_Properties_Form.prAddAIChannel;
var
  Ch :TQUB_Acquisition_Channel;
  i  :Integer;
begin

  //Anything selected ?
  i := lstAvailAIChannels.ItemIndex;
  if (i < 0) or (QUBAcqAIChannels.Count = 0) then Exit;

  Ch := TQUB_Acquisition_Channel(QUBAcqAIChannels.Objects[i]);

  //Already installed ?
  if AIChannels.IndexOfObject(Ch) >= 0 then Exit;

  lstInstAIChannels.Items.Add(Ch.Name);
  AIChannels.AddObject(Ch.Name, Ch);

  i := AIChannels.Count - 1;
  if lstInstAIChannels.ItemIndex >= 0 then lstInstAIChannels.Selected[lstInstAIChannels.ItemIndex] := False;
  lstInstAIChannels.ItemIndex   := i;
  lstInstAIChannels.Selected[i] := True;

end;

procedure TQUB_AcquisitionInput_Properties_Form.prDelAIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAIChannels.ItemIndex;
  if (i < 0) or (AIChannels.Count = 0) then Exit;

  AIChannels.Delete(i);
  lstInstAIChannels.Items.Delete(i);

  if i >= AIChannels.Count then i := AIChannels.Count - 1;
  if i >= 0 then begin
    lstInstAIChannels.ItemIndex   := i;
    lstInstAIChannels.Selected[i] := True;
  end;

end;

procedure TQUB_AcquisitionInput_Properties_Form.prUpAIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAIChannels.ItemIndex;
  if (i < 0) or (AIChannels.Count = 0) or (i = 0) then Exit;

  AIChannels.Exchange(i, i - 1);
  lstInstAIChannels.Items.Exchange(i, i - 1);

  lstInstAIChannels.Selected[i]     := False;
  lstInstAIChannels.Selected[i - 1] := True;
  lstInstAIChannels.ItemIndex       := i - 1;

end;

procedure TQUB_AcquisitionInput_Properties_Form.prDnAIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAIChannels.ItemIndex;
  if (i < 0) or (AIChannels.Count = 0) or (i = (AIChannels.Count - 1)) then Exit;

  AIChannels.Exchange(i, i + 1);
  lstInstAIChannels.Items.Exchange(i, i + 1);

  lstInstAIChannels.Selected[i]     := False;
  lstInstAIChannels.Selected[i + 1] := True;
  lstInstAIChannels.ItemIndex       := i + 1;

end;

//******************************************************************************
procedure TQUB_AcquisitionInput_Properties_Form.FormShow(Sender: TObject);
begin

  lstInstAIChannels.Items.Assign(AIChannels);

  lstAvailAIChannels.Items.Assign(QUBAcqAIChannels);

end;

//******************************************************************************
procedure TQUB_AcquisitionInput_Properties_Form.btnAddAIClick(Sender: TObject);
begin
  prAddAIChannel;
end;

procedure TQUB_AcquisitionInput_Properties_Form.btnDelAIClick(Sender: TObject);
begin
  prDelAIChannel;
end;

procedure TQUB_AcquisitionInput_Properties_Form.btnUpAIClick(Sender: TObject);
begin
  prUpAIChannel;
end;

procedure TQUB_AcquisitionInput_Properties_Form.btnDnAIClick(Sender: TObject);
begin
  prDnAIChannel;
end;

procedure TQUB_AcquisitionInput_Properties_Form.lstInstAIChannelsDblClick(Sender: TObject);
begin
  prDelAIChannel;
end;

procedure TQUB_AcquisitionInput_Properties_Form.lstAvailAIChannelsDblClick(Sender: TObject);
begin
  prAddAIChannel;
end;

procedure TQUB_AcquisitionInput_Properties_Form.btnHelpClick(
  Sender: TObject);
begin
  ShowWiki('Acquisition_Input', Handle);
end;


//******************************************************************************
end.
