unit QUB_AcquisitionOutput;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls,
StdCtrls, Registry, Math, Buttons, ExtCtrls, SyncObjs,

AllocTypes, Alloc,
RegistryInterface,

QUB_Global,

QUB_Console, QUB_Rich_Console,

QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QFS_QIF_Segments,
QUB_Acquisition_Hardware,
QUB_Acquisition_Global,
QUB_AcquisitionInput,
QUB_AcquisitionOutput_Properties,
QUB_Viewport_Data, QUB_Viewport_Model;

//KPL_Objects, KPL_Kinetic, KPL_Program,
//KPL_CollectionQUBPulse;


const


MIN_TIMEPOINTS                                  = 10;
MAX_TIMEPOINTS                                  = High(Integer);

DEF_BUFFERLENGTH                                = 1000;
MIN_BUFFERLENGTH                                = 10;
MAX_BUFFERLENGTH                                = 5000;

DEF_SAMPLING                                    = 10000;
MIN_SAMPLING                                    = 1e-3;
MAX_SAMPLING                                    = 1e10;

DEF_DSAMPLING                                   = 100;
MIN_DSAMPLING                                   = 1e-3;
MAX_DSAMPLING                                   = 100000;


WM_QUB_ACQ_OUTPUT_STOP                          = WM_USER + 100;
WM_QUB_ACQ_OUTPUT_UPDATEWRITENPOINTS            = WM_USER + 101;


//QUB_ACQOUT_DATAFILE                             = 0;
//QUB_ACQOUT_PROTOCOL                             = 1;
QUB_ACQOUT_SCRIPTS                              = 0;
QUB_ACQOUT_SEND                                 = 1;


type


TQUB_Acquisition_Output_State         = (ACQ_OUTPUT_STATE_STOPPED, ACQ_OUTPUT_STATE_RUNNING);

TQUB_Acquisition_ODataError           = (DATA_ERR_NONE, DATA_ERR_UNASS,  DATA_ERR_DIFFCHAN, DATA_ERR_DIFFBITS);
TQUB_Acquisition_ODeviceError         = (DEV_ERR_NONE,  DEV_ERR_UNASS);
TQUB_Acquisition_OChannelsError       = (CHAN_ERR_NONE, CHAN_ERR_NOCHAN, CHAN_ERR_NODEV,    CHAN_ERR_DIFFDEV);

TQUB_Acquisition_Output               = class(TFrame)


  panPanels                           :TPanel;
  panSend                             :TPanel;
  btnRun                              :TSpeedButton;
  btnStop                             :TSpeedButton;
  Label1                              :TLabel;
    lblWritenTime: TLabel;
  panScripts                          :TPanel;
  panOverwrite                        :TPanel;
  panDuration                         :TPanel;
  txtAcqTimeHours                     :TEdit;
  udAcqTimeHours                      :TUpDown;
  txtAcqTimeMinutes                   :TEdit;
  udAcqTimeMinutes                    :TUpDown;
  txtAcqTimeSeconds                   :TEdit;
  udAcqTimeSeconds                    :TUpDown;
  chkAcqModeCont                      :TRadioButton;
  chkAcqModeTime                      :TRadioButton;
  chkAcqModePoints                    :TRadioButton;
  txtAcqTimePoints                    :TEdit;
  chkAcqModeOnce                      :TRadioButton;
  lblSampling                         :TLabel;
  txtSampling                         :TEdit;
  lblBufferLength                     :TLabel;
  txtBufferLength                     :TEdit;
  lblData                             :TLabel;
  lblDevice                           :TLabel;
  lblDSampling                        :TLabel;
  txtDSampling                        :TEdit;
  btnProperties                       :TSpeedButton;
  cmbData                             :TComboBox;
  Panel1                              :TPanel;
  btnModeScripts                      :TSpeedButton;
  btnModeAcq                          :TSpeedButton;
  btnSynchRun                         :TSpeedButton;


  procedure txtSamplingKeyPress(Sender: TObject; var Key: Char);
  procedure txtSamplingExit(Sender: TObject);

  procedure txtDSamplingKeyPress(Sender: TObject; var Key: Char);
  procedure txtDSamplingExit(Sender: TObject);

  procedure txtBufferLengthKeyPress(Sender: TObject; var Key: Char);
  procedure txtBufferLengthExit(Sender: TObject);

  procedure chkAcqModeClick(Sender: TObject);
  procedure udAcqTimeHoursClick(Sender: TObject; Button: TUDBtnType);
  procedure udAcqTimeMinutesClick(Sender: TObject; Button: TUDBtnType);
  procedure udAcqTimeSecondsClick(Sender: TObject; Button: TUDBtnType);
  procedure txtAcqTimePointsKeyPress(Sender: TObject; var Key: Char);
  procedure txtAcqTimePointsExit(Sender: TObject);

  procedure btnRunClick(Sender: TObject);
  procedure btnStopClick(Sender: TObject);

  procedure btnPropertiesClick(Sender: TObject);

  procedure btnModeClick(Sender: TObject);

  procedure lstPanelsClick(Sender: TObject);
  procedure cmbDataChange(Sender: TObject);
  procedure cmbDataDropDown(Sender: TObject);

  procedure SetFocusClick(Sender: TObject);
  procedure btnSynchRunClick(Sender: TObject);


protected


  prPropertiesForm            :TQUB_AcquisitionOutput_Properties_Form;

  prAcqInput                  :TQUB_Acquisition_Input;
  prData                      :TQUB_QVO_QVD;

  prAOChannels                :TQUB_Acquisition_Channels;
  prDOChannels                :TQUB_Acquisition_Channels;
  prDevice                    :TQUB_Acquisition_ODevice;
  prHardware                  :TQUB_Acquisition_Hardware;

  prSampling                  :Double;//Hertz
  prDSampling                 :Double;//Hertz

  prAcqState                  :TQUB_Acquisition_Output_State;

  prDataError                 :TQUB_Acquisition_ODataError;
  prDeviceError               :TQUB_Acquisition_ODeviceError;
  prChannelsError             :TQUB_Acquisition_OChannelsError;

  prAcqMode                   :TQUB_Acquisition_Mode;
  prAcqTimeHours              :Integer;
  prAcqTimeMinutes            :Integer;
  prAcqTimeSeconds            :Integer;
  prAcqTimePoints             :Integer;

  prPointsWriten              :Integer;
  prPointsWriting             :Integer;
  prUpdatingWritenPoints      :Boolean;
  prPointsFirst               :Integer;
  prWritingData               :Boolean;
  prWritingCritSection        :TCriticalSection;

  prLockCount                 :Integer;

  prBufferLength              :Integer;//ms

  prStatus                    :TLabel;

  m_lSynchRun                 :Boolean; // TRUE if btnSynch started acquisition, FALSE if btnRun did.
  m_iSynchLoop                :Integer; // For synchronized, Iterations of total output run
  m_iInitialSize              :Integer; // For synchronized, Size of data prior to recording

  //Scripting
  //prScripting                 :Boolean;
  //prOperations                :TQUB_Acquisition_Operations;

  //Events
  prOnDataNew                 :TNotifyEvent;
  prOnDataOpen                :TNotifyEvent;
  prOnDataAssign              :TNotifyEvent;

  prFLCSize                   :Integer;
  prFLCBuffer                 :TQUB_QFS_QIF_PAFirstLastClass;

  prDSegment                  :Integer;
  prDSegmentPos               :Integer;
  {
  //WaveForm
  prWaveFormFile              :String;
  prSource                    :TKPL_Program;
  prKinetic                   :TKPL_Kinetic;

  //KPL Collections ************************************************************
  prKPLConsoleCollection      :TKPL_Object;
  prKPLConsole                :TKPL_Object;//TKPL_Collection_Console;

  prKPLPulseCollection        :TKPL_Object;
  prKPLPulse                  :TKPL_Object;//TKPL_Collection_QUB_Data;

  prKPLDAQCollection          :TKPL_Object;
  prKPLDAQ                    :TKPL_Object;//TKPL_Collection_QUB_DAQ;

  prKPLInterfaceCollection    :TKPL_Object;
  prKPLInterface              :TKPL_Object;//TKPL_Collection_QUB_Interface;

  prKPLODECollection          :TKPL_Object;
  prKPLODE                    :TKPL_Object;//TKPL_Collection_ODE;
  //****************************************************************************
  }

  procedure prUpdateChannels;

  procedure prPropertiesRegistry(Read :Boolean);


  //Acquisition State
  procedure prSetAcquisitionState(s :TQUB_Acquisition_Output_State);

  //Acquisition Mode
  procedure prSetAcquisitionMode(m :TQUB_Acquisition_Mode);

  //Acquisition Duration
  procedure prSetAcquisitionTimeHours(h :Integer);
  procedure prSetAcquisitionTimeMinutes(m :Integer);
  procedure prSetAcquisitionTimeSeconds(s :Integer);
  procedure prSetAcquisitionTimePoints(p :Integer);

  //Data
  procedure prSetData(d :TQUB_QVO_QVD);
  procedure prCheckChannelInfo;

  procedure prSetSampling(s :Double);
  function  prGetSampling :Double;

  procedure prSetDSampling(s :Double);
  function prGetDSampling :Double;

  procedure prSetBufferLength(b :Integer);
  function  prGetBufferLength :Integer;

  procedure prUpdateWritenPoints;

  procedure prMsgUpdateWritenPoints(var Message: TMessage); message WM_QUB_ACQ_OUTPUT_UPDATEWRITENPOINTS;
  procedure prMsgStop(var Message: TMessage); message WM_QUB_ACQ_OUTPUT_STOP;

  function prGetLocked :Boolean;

  procedure prSetStatus;

  procedure prOnStartRun(Sender :TObject);
  procedure prOnEndRun(Sender :TObject; Failed :Boolean);

  procedure prWriteData(Buffer :Pointer; BufferSize, BufferCount :Integer);


public


  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;


  procedure DoOnChangeHardware(Sender :TObject);
  procedure DoOnChangeChannels(Sender :TObject);

  procedure Lock;
  procedure UnLock;


  function AcqStart(Synchronized :Boolean; Scripting :Boolean; Operations :TQUB_Acquisition_Operations) :Boolean;
  function AcqStop :Boolean;

  procedure AcquisitionCallBack(Buffer :Pointer; BufferSize :Integer; var BufferCount :Integer; var Error :Integer);
  procedure AcquisitionDCallBack(var DValue, DCount :Integer; var Error :Integer);

  function NewOutputAcq(const FileName :String) :Boolean;
  function OpenOutputAcq(const FileName :String) :Boolean;
  function AssignOutputAcq(AData :TQUB_QVO_QVD) :Boolean;


  //Lock
  property Locked                        :Boolean                         read prGetLocked;

  //AcqInput
  property AcqInput                      :TQUB_Acquisition_Input          read prAcqInput                    write prAcqInput;

  //Data
  property Data                          :TQUB_QVO_QVD                    read prData                        write prSetData;

  //Device
  property Device                        :TQUB_Acquisition_ODevice        read prDevice;

  //Hardware
  property Hardware                      :TQUB_Acquisition_Hardware       read prHardware;

  property Sampling                      :Double                          read prGetSampling                 write prSetSampling;
  property BufferLength                  :Integer                         read prGetBufferLength             write prSetBufferLength;
  property DSampling                     :Double                          read prGetDSampling                write prSetDSampling;

  property AcqState                      :TQUB_Acquisition_Output_State   read prAcqState                    write prSetAcquisitionState;

  property AcqMode                       :TQUB_Acquisition_Mode           read prAcqMode                     write prSetAcquisitionMode;
  property AcqTimeHours                  :Integer                         read prAcqTimeHours                write prSetAcquisitionTimeHours;
  property AcqTimeMinutes                :Integer                         read prAcqTimeMinutes              write prSetAcquisitionTimeMinutes;
  property AcqTimeSeconds                :Integer                         read prAcqTimeSeconds              write prSetAcquisitionTimeSeconds;
  property AcqTimePoints                 :Integer                         read prAcqTimePoints               write prSetAcquisitionTimePoints;

  //Data events
  property OnDataNew                     :TNotifyEvent                    read prOnDataNew                   write prOnDataNew;
  property OnDataOpen                    :TNotifyEvent                    read prOnDataOpen                  write prOnDataOpen;
  property OnDataAssign                  :TNotifyEvent                    read prOnDataAssign                write prOnDataAssign;

  //Status
  property Status                        :TLabel                          read prStatus                      write prStatus;


end;


implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Node,
//KPL_CollectionConsole, KPL_CollectionQUBAcqProtocol,
//KPL_CollectionQUBInterface, KPL_CollectionDAQ, KPL_CollectionODE,
QUB_Viewport_AcquisitionOutput;


//******************************************************************************
procedure AcquisitionCallBack(Sender :TObject; Buffer :Pointer; BufferSize :Integer; var BufferCount :Integer; var Error :Integer); stdcall;
begin
  TQUB_Acquisition_Output(Sender).AcquisitionCallBack(Buffer, BufferSize, BufferCount, Error);
end;

procedure AcquisitionDCallBack(Sender :TObject; var DValue, DCount :Integer; var Error :Integer); stdcall;
begin
  TQUB_Acquisition_Output(Sender).AcquisitionDCallBack(DValue, DCount, Error);
end;


//******************************************************************************
constructor TQUB_Acquisition_Output.Create(AOwner: TComponent);
begin

  inherited;

  Align            := alClient;

  prAcqState       := ACQ_OUTPUT_STATE_STOPPED;

  prData           := nil;
  prDevice         := nil;
  prHardware       := nil;

  prAOChannels     := TQUB_Acquisition_Channels.Create;
  prAOChannels.OwnChannels := False;
  prDOChannels     := TQUB_Acquisition_Channels.Create;
  prDOChannels.OwnChannels := False;

  prPropertiesRegistry(True{Read});

  //Set Properties Form
  prPropertiesForm := TQUB_AcquisitionOutput_Properties_Form.Create(Self);
  prPropertiesForm.AOChannels := prAOChannels;
  prPropertiesForm.DOChannels := prDOChannels;
  {
  //Collections ****************************************************************
  prKPLConsoleCollection                              := TKPL_Object.Create('Collection', nil);
  prKPLConsole                                        := TKPL_Collection_Console.Create('console', prKPLConsoleCollection);

  prKPLPulseCollection                                := TKPL_Object.Create('Collection', nil);
  prKPLPulse                                          := TKPL_Collection_QUB_Pulse.Create('pulse', prKPLPulseCollection);

  prKPLDAQCollection                                  := TKPL_Object.Create('Collection', nil);
  prKPLDAQ                                            := TKPL_Collection_QUB_DAQ.Create('daq', prKPLDAQCollection);

  prKPLInterfaceCollection                            := TKPL_Object.Create('Collection', nil);
  prKPLInterface                                      := TKPL_Collection_QUB_Interface.Create('qub', prKPLInterfaceCollection);

  prKPLODECollection                                  := TKPL_Object.Create('Collection', nil);
  prKPLODE                                            := TKPL_Collection_ODE.Create('odelib', prKPLODECollection);
  //****************************************************************************

  prKinetic                                           := TKPL_Kinetic.Create('kinetic', nil);
  prKinetic.Register(prKPLConsoleCollection);
  prKinetic.Register(prKPLPulseCollection);
  prKinetic.Register(prKPLDAQCollection);
  prKinetic.Register(prKPLInterfaceCollection);
  prKinetic.Register(prKPLODECollection);

  prSource                                            := TKPL_Program.Create(Self, 'Acquisition Waveform Generator', prKinetic);
  prSource.Options                                    := [KPL_PRG_BUTTONS, KPL_PRG_FILE, KPL_PRG_COMPILE, KPL_PRG_RUN, KPL_PRG_WATCH];
  prSource.CanCompile                                 := False;
  prSource.CanRun                                     := False;
  prSource.Parent                                     := panScripts;
  prSource.Align                                      := alClient;
  prSource.OnStartRun                                 := prOnStartRun;
  prSource.OnEndRun                                   := prOnEndRun;
  }
  panSend{panData}.BringToFront;

  prUpdateChannels;
  prSetStatus;

  cmbData.ItemIndex := 2;

end;


destructor TQUB_Acquisition_Output.Destroy;
begin

  try

    if AcqState <> ACQ_OUTPUT_STATE_STOPPED then AcqStop;

    prPropertiesRegistry(False{Read});

    prAOChannels.Free;
    prDOChannels.Free;
    {
    prKinetic.Free;

    prKPLODECollection.Free;
    prKPLInterfaceCollection.Free;
    prKPLDAQCollection.Free;
    prKPLPulseCollection.Free;
    prKPLConsoleCollection.Free;
    }
    FreeMem(prFLCBuffer);
    prFLCBuffer := nil;

  finally

    inherited;

  end;

end;

//Events ***********************************************************************

//DoOnChangeHardware
procedure TQUB_Acquisition_Output.DoOnChangeHardware(Sender :TObject);
begin
  prUpdateChannels;
  prCheckChannelInfo;
  prSetStatus;
end;

//DoOnChangeChannels
procedure TQUB_Acquisition_Output.DoOnChangeChannels(Sender :TObject);
begin
  prUpdateChannels;
  prCheckChannelInfo;
  prSetStatus;
end;


//Lock *************************************************************************
function TQUB_Acquisition_Output.prGetLocked :Boolean;
begin
  Result := prLockCount > 0;
end;

procedure TQUB_Acquisition_Output.Lock;
begin
  Inc(prLockCount);
end;

procedure TQUB_Acquisition_Output.UnLock;
begin
  Dec(prLockCount);
  if prLockCount < 0 then prLockCount := 0;
end;


//Update Channels *************************************************************
procedure TQUB_Acquisition_Output.prUpdateChannels;
var
  ACh     :TQUB_Acquisition_AOChannel;
  DCh     :TQUB_Acquisition_DOChannel;
  c, i    :Integer;
begin

  //Remove Dead Channels *******************************************************
  //AO
  for c := prAOChannels.Count - 1 downto 0 do begin
    i := QUBAcqAOChannels.IndexOfObject(prAOChannels.Objects[c]);
    if i < 0 then begin
      prAOChannels.Delete(c);
    end else begin
      prAOChannels.Strings[c] := QUBAcqAOChannels.Strings[i];
    end;
  end;
  //DO
  for c := prDOChannels.Count - 1 downto 0 do begin
    i := QUBAcqDOChannels.IndexOfObject(prDOChannels.Objects[c]);
    if i < 0 then begin
      prDOChannels.Delete(c);
    end else begin
      prDOChannels.Strings[c] := QUBAcqDOChannels.Strings[i];
    end;
  end;

  //Set Device
  //AO
  for c := 0 to prAOChannels.Count - 1 do begin
    ACh     := TQUB_Acquisition_AOChannel(prAOChannels.Objects[c]);
    i       := QUBAcqODevices.IndexOf(ACh.DeviceName);
    if i < 0 then begin
      ACh.Device := nil;
    end else begin
      ACh.Device := TQUB_Acquisition_ODevice(QUBAcqODevices.Objects[i]);
    end;
  end;
  //DO
  for c := 0 to prDOChannels.Count - 1 do begin
    DCh     := TQUB_Acquisition_DOChannel(prDOChannels.Objects[c]);
    i       := QUBAcqODevices.IndexOf(DCh.DeviceName);
    if i < 0 then begin
      DCh.Device := nil;
    end else begin
      DCh.Device := TQUB_Acquisition_ODevice(QUBAcqODevices.Objects[i]);
    end;
  end;

  //Error checking *************************************************************
  prDeviceError   := DEV_ERR_NONE;
  prChannelsError := CHAN_ERR_NONE;

  if prAOChannels.Count > 0 then begin
    prDevice        := TQUB_Acquisition_ODevice(TQUB_Acquisition_AOChannel(prAOChannels.Objects[0]).Device);
    if prDevice     =  nil then prChannelsError := CHAN_ERR_NODEV;
  end else begin
    prDevice        := nil;
    prChannelsError := CHAN_ERR_NOCHAN;
  end;

  if prDevice <> nil then begin
    for c := 0 to prAOChannels.Count - 1 do begin
      Ach := TQUB_Acquisition_AOChannel(prAOChannels.Objects[c]);
      if Ach.Device = nil then begin
        prDevice := nil;
        prChannelsError := CHAN_ERR_NODEV;
        Break;
      end else if Ach.Device <> prDevice then begin
        prDevice := nil;
        prChannelsError := CHAN_ERR_DIFFDEV;
      end;
    end;
    for c := 0 to prDOChannels.Count - 1 do begin
      Dch := TQUB_Acquisition_DOChannel(prDOChannels.Objects[c]);
      if Dch.Device = nil then begin
        prDevice := nil;
        prChannelsError := CHAN_ERR_NODEV;
        Break;
      end else if Dch.Device <> prDevice then begin
        prDevice := nil;
        prChannelsError := CHAN_ERR_DIFFDEV;
      end;
    end;
  end;

  if prDevice = nil then begin
    prDeviceError := DEV_ERR_UNASS;
    prHardware    := nil;
  end else begin
    prHardware    := prDevice.Hardware;
  end;

  //Data
  prDataError := DATA_ERR_NONE;
  if prData = nil then begin
    prDataError := DATA_ERR_UNASS;
  end else if prData.DataObject.ADChannelCount <> prAOChannels.Count then begin
    prDataError := DATA_ERR_DIFFCHAN;
    prData := nil;
  end else if (prDevice <> nil) and ((prData.DataObject.ADDataSize) <> ((prDevice.Hardware.OutputDeviceParams[prDevice.Index].m_Bits + 7) div 8)) then begin
    prDataError := DATA_ERR_DIFFBITS;
    prData := nil;
  end;

  if (prDevice <> nil) and (prDataError = DATA_ERR_NONE) then begin
    prData.Scaling := IntPower(2, prDevice.Hardware.OutputDeviceParams[prDevice.Index].m_Bits) / prDevice.Hardware.OutputDeviceParams[prDevice.Index].m_Range;
    //?!?!
  end;

  AcqState        := AcqState;

end;


//******************************************************************************
function TQUB_Acquisition_Output.AcqStart(Synchronized :Boolean; Scripting :Boolean; Operations :TQUB_Acquisition_Operations) :Boolean;
var
  i          :Integer;
  DS, Seg    :INode;
  DwellCount :Integer;
  FF, LL, CC :TPAInteger;
begin

  Result     := True;
  m_lSynchRun := Synchronized;

  if Locked then Exit;
  if (prChannelsError <> CHAN_ERR_NONE) or (prDeviceError <> DEV_ERR_NONE) then Exit;
  if (prDataError = DATA_ERR_NONE) and (Data.LockedAny) then Exit;
  if (prDataError <> DATA_ERR_NONE) then Exit;

  Result     := False;

  //prSource.CanCompile := False;
  //prSource.CanRun     := False;

  //Lock, itself and Data
  Lock;
  Data.LockAny;
  Data.LockPlayback;

  //Set Device Properties
  prDevice.ASampling        := prSampling;
  prDevice.DSampling        := prDSampling;
  prDevice.Caller           := Self;
  prDevice.ACallBack        := @QUB_AcquisitionOutput.AcquisitionCallBack;
  prDevice.DCallBack        := @QUB_AcquisitionOutput.AcquisitionDCallBack;
  prDevice.BufferLength     := BufferLength;
  prDevice.AChannelCount    := prAOChannels.Count;
  for i := 0 to prAOChannels.Count - 1 do begin
    prDevice.AChannel[i]    := TQUB_Acquisition_AOChannel(prAOChannels.Objects[i]).Channel;
  end;
  prDevice.DChannelCount    := prDOChannels.Count;
  for i := 0 to prDOChannels.Count - 1 do begin
    prDevice.DChannel[i]    := TQUB_Acquisition_DOChannel(prDOChannels.Objects[i]).Channel;
  end;

  prPointsWriten            := 0;
  prPointsFirst             := 0;

  prDSegment                := 0;
  prDSegmentPos             := 0;

  PostMessage(Self.Handle, WM_QUB_ACQ_OUTPUT_UPDATEWRITENPOINTS, 0, 0);

  panOverWrite.Visible     := False;

  case prAcqMode of
    ACQ_MODE_CONT://Continuous
      begin
        prPointsWriting := High(Integer);
      end;
    ACQ_MODE_TIME://Time
      begin
        prPointsWriting := Round(prSampling * (prAcqTimeHours * 3600 + prAcqTimeMinutes * 60 + prAcqTimeSeconds));
      end;
    ACQ_MODE_POINTS://Points
      begin
        prPointsWriting := prAcqTimePoints;
      end;
    ACQ_MODE_ONCE://The whole data once
      begin
        prPointsWriting := Data.DataObject.Segments.TotalDataCount;
      end;
  end;//case AcquisitionMode
  {
  if Scripting then begin
    prScripting  := Scripting;
    prOperations := Operations;
  end;
  }
  //FLC Buffer
  if prDOChannels.Count > 0 then begin
    prFLCSize   := 0;
    prFLCBuffer := nil;

    DS    := Data.DataObject.Idealization[0].ToTree;
    Seg   := DS.Find('Segment');
    while not Seg.IsNull do begin
      Inc(prFLCSize, Integer(Seg['DwellCount'].Data[0]));
      Seg := Seg.SiblingSameName;
    end;
    if prFLCSize <> 0 then begin
      try
        ReAllocMem(prFLCBuffer, prFLCSize * SizeOf(TQUB_QFS_QIF_FirstLastClass));
      except
        prFLCSize := 0;
        FreeMem(prFLCBuffer);
        prFLCBuffer := nil;
      end;
    end;
    if prFLCSize <> 0 then begin
      prFLCSize := 0;
      Seg    := DS.Find('Segment');
      while not Seg.IsNull do begin
        DwellCount := Seg['DwellCount'].Data[0];
        FF         := TPAInteger( Seg['Firsts'].DataPtr );
        LL         := TPAInteger( Seg['Lasts'].DataPtr );
        CC         := TPAInteger( Seg['Classes'].DataPtr );

        for i := 0 to DwellCount - 1 do begin
          prFLCBuffer[prFLCSize + i].First := FF[i];
          prFLCBuffer[prFLCSize + i].Last  := LL[i];
          prFLCBuffer[prFLCSize + i].Cls   := CC[i];
        end;

        Inc( prFLCSize, DwellCount );
        Seg  := Seg.SiblingSameName;
      end;
    end;
  end;//DOCount > 0

  //-- For synchronized I/O, preload a list of breakpoints where recorded data will be split.
  if Synchronized and (QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.m_aBreaks <> nil) then begin
    m_iSynchLoop   := 0;
    m_iInitialSize := QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.Data.DataPointCount;
    Data.DataObject.ListSegmentBreaks(m_iInitialSize, QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.m_aBreaks);
  end;

  //-- if not synchronized, attempt to start the generation
  prAcqState := ACQ_OUTPUT_STATE_RUNNING;

  if (not Synchronized) and (prDevice.Start)then begin

    prAcqState          := ACQ_OUTPUT_STATE_STOPPED;

    //prSource.CanCompile := True;
    //prSource.CanRun     := True;

    UnLock;
    Data.UnlockAny;
    Data.UnlockPlayback;

    //if prScripting then begin
    //  Console.NewEntry('Cannot start output acquisition', COLOR_ERROR);
    //end else begin
      MessageHelpDlg('Cannot start output acquisition', mtError, [mbOK], 0, 'Error:Cannot_Start_Acquisition');
    //end;

    //prScripting := False;

    Result := True;

  end else begin

    AcqState := ACQ_OUTPUT_STATE_RUNNING;

  end;

end;

function TQUB_Acquisition_Output.AcqStop :Boolean;
begin
  Result := True;

  if (prDevice = nil) or (prData = nil) then Exit;

  Result := False;

  prAcqState := ACQ_OUTPUT_STATE_STOPPED;
  if not m_lSynchRun Then
    prDevice.Stop
  else begin
    prDevice.Started := False;
  end;

  //prSource.CanCompile := True;
  //prSource.CanRun     := True;

  UnLock;
  Data.UnLockAny;
  Data.UnlockPlayback;

  //prScripting := False;

  AcqState    := ACQ_OUTPUT_STATE_STOPPED;

  FreeMem(prFLCBuffer);
  prFLCBuffer := nil;
  prFLCSize := 0;

end;

procedure TQUB_Acquisition_Output.prMsgStop(var Message: TMessage);
begin
  btnStopClick(btnStop);
end;

procedure TQUB_Acquisition_Output.AcquisitionCallBack(Buffer :Pointer; BufferSize :Integer; var BufferCount :Integer; var Error :Integer);
begin

  if prAcqState = ACQ_OUTPUT_STATE_STOPPED then begin
    BufferCount := 0;
    Exit;
  end;

  if {(prWritingData) or} (Error <> 0) then begin

  //See this message in SetAcqState:
  //this is called AFTER acquisition starts, then it either enters an infinite loop
  //  or causes an IsWriting state in the Callback and leads to freezing. To fix.

    panOverWrite.Visible := True;

  end else begin

    prWritingData := True;

    try

      if BufferCount > 0 then begin

        if (AcqState = ACQ_OUTPUT_STATE_RUNNING) then begin

          if prPointsWriten < prPointsWriting then begin

            //Change to Points
            if BufferCount > (prPointsWriting - prPointsWriten) then begin
              BufferCount := prPointsWriting - prPointsWriten;
            end;

            prWriteData(Buffer, BufferSize, BufferCount);

            Inc(prPointsWriten, BufferCount);
            if BufferCount > 0 then begin
              //Post message to update writen points labels
              PostMessage(Self.Handle, WM_QUB_ACQ_OUTPUT_UPDATEWRITENPOINTS, 0, 0);
            end;

          end else begin

            //Post message to Stop
            PostMessage(Self.Handle, WM_QUB_ACQ_OUTPUT_STOP, 0, 0);

          end;

        end;

      end;

    finally

      prWritingData := False;

    end;

  end;

end;

procedure TQUB_Acquisition_Output.AcquisitionDCallBack(var DValue, DCount :Integer; var Error :Integer);
var
  c, s :Integer;
begin

  try

    DValue := prFLCBuffer[prDSegment].Cls;  // Consider putting this after the repeat loop 

    c := DCount;  // # of samples to advance
    repeat
      // samples left in current segment  (5/05 - added +1 )
      s := 1 + prFLCBuffer[prDSegment].Last - (prFLCBuffer[prDSegment].First + prDSegmentPos);
      if s >= c then begin
        //-- Value in current segment - set pointer to position within segment and exit
        Inc(prDSegmentPos, c);
        Break;
      end else begin
        //-- Value beyond current segment - advance to next segment
        Inc(prDSegment);
        if prDSegment >= prFLCSize then prDSegment := 0;
        prDSegmentPos := 0;
        Dec(c, s);
      end;
    until(False);

  except
  end;
end;

//******************************************************************************
function TQUB_Acquisition_Output.NewOutputAcq(const FileName :String) :Boolean;
begin

  Result := True;

  if AcqState <> ACQ_OUTPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  //Instruct ViewportData to create a new data file and associate with it
  ViewportData.NewPropertiesForm.ADChannelCount := prAOChannels.Count;
  if not ViewportData.New(FileName, False{Silent}) then begin
    Result := AssignOutputAcq(ViewportData.CurrData);
  end else begin
    AssignOutputAcq(prData);
  end;

end;

function TQUB_Acquisition_Output.OpenOutputAcq(const FileName :String) :Boolean;
begin

  Result := True;

  if AcqState <> ACQ_OUTPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  //Instruct ViewportData to create a new data file and associate with it
  if not ViewportData.Open(FileName, False) then begin
    Result := AssignOutputAcq(ViewportData.CurrData);
  end else begin
    AssignOutputAcq(prData);
  end;

end;

function TQUB_Acquisition_Output.AssignOutputAcq(AData :TQUB_QVO_QVD) :Boolean;
begin

  Result := True;

  if AcqState <> ACQ_OUTPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  if AData = nil then begin
    Data := AData;
    Result := False;
  end else if AData.LockedAny then begin
    Exit; //Data was created but is, somehow, locked
  end else begin
    Data := AData;
    Result := False;
  end;

end;

//******************************************************************************
procedure TQUB_Acquisition_Output.prWriteData(Buffer :Pointer; BufferSize, BufferCount :Integer);
var
  Copied :Integer;
  Copy   :TQUB_FirstLast;
  BPoint :Integer;
begin

  Copied := 0;
  BPoint := 0;

  if Data.DataPointCount = 0 then Exit;

  repeat

    //-- set copy range from first to min(first+size,last)
    Copy.First := prPointsFirst;
    Copy.Last  := Copy.First + (BufferCount - Copied) - 1;
    if Copy.Last > Data.DataPoint.Last then begin
      Copy.Last := Data.DataPoint.Last;
    end;

    //-- copy a chunk of data
    Data.DataObject.ReadData(@TPAByte(Buffer)[BPoint], True{Raw}, -1{Channel}, Copy.First, Copy.Last);

    //An attempt to show what's being output.
    //Too slow as it is - should Window-Draw
    //Maybe we can have a reverse Scope?
    //LM.
    //Data.MoveSelectionIntoWindow(QUB_FirstLast(Copy.Last, Copy.Last), Copy.Last);

    Inc(BPoint, (Copy.Last - Copy.First + 1) * BufferSize);
    Inc(Copied, Copy.Last - Copy.First + 1);

    //-- Adjust pointer to next section
    prPointsFirst := Copy.Last + 1;
    if prPointsFirst > Data.DataPoint.Last then begin
      //-- If looping to start, add another set of segment breaks to the list
      if m_lSynchRun and (QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.m_aBreaks<>nil) Then Begin
        inc(m_iSynchLoop);
        Data.DataObject.ListSegmentBreaks( m_iInitialSize+m_iSynchLoop*(Data.DataPoint.Last+1), QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.m_aBreaks );
      end;
      prPointsFirst := Data.DataPoint.First;
    end;

  until (Copied >= BufferCount);

end;

procedure TQUB_Acquisition_Output.prUpdateWritenPoints;
var
  h, m, s :Integer;
begin
  if prUpdatingWritenPoints then Exit;

  prUpdatingWritenPoints := True;

  try

    try
      s := Round(prPointsWriten / prSampling);
      //TO DO: Did I have a reason to use trunc instead of div?! LM.
      //h := Trunc(s / 3600);
      h := s div 3600;
      s := s - h *3600;
      m := s div 60;
      s := s - m * 60;
      lblWritenTime.Caption := Format('%d:%d:%d / %d', [h, m, s, prPointsWriten]);
    except
    end;

  finally

    prUpdatingWritenPoints := False;

  end;

end;

procedure TQUB_Acquisition_Output.prMsgUpdateWritenPoints(var Message: TMessage);
begin
  prUpdateWritenPoints;
end;


//Acquisition State
procedure TQUB_Acquisition_Output.prSetAcquisitionState(s :TQUB_Acquisition_Output_State);
var
  e :Boolean;
begin
  //TO DO: this is called AFTER acquisition starts, then it either enters an infinite loop
  //  or causes an IsWriting state in the Callback and leads to freezing. To fix. LM


  //todo: use critical sections
  while prWritingData do begin
  end;

  prWritingData := True;

  try

    case s of

      ACQ_OUTPUT_STATE_RUNNING:
        begin
          if ((prDevice <> nil) and (prData <> nil)) then begin
            s := ACQ_OUTPUT_STATE_RUNNING;
          end else begin
            s := ACQ_OUTPUT_STATE_STOPPED;
          end;
          btnRun.Enabled     := False;
          btnSynchRun.Enabled := False;
          btnStop.Enabled    := (s = ACQ_OUTPUT_STATE_RUNNING);// and (not prScripting);
        end;

      ACQ_OUTPUT_STATE_STOPPED:
        begin
          btnRun.Enabled     := ((prDevice <> nil) and (prData <> nil));// and (not prScripting);
          btnSynchRun.Enabled := btnRun.Enabled;  // TODO - Also check if input recording is ready 
          btnStop.Enabled    := False;
        end;

    end;//case AcqState

    e                             := (s = ACQ_OUTPUT_STATE_STOPPED);// and (not prScripting);

    chkAcqModeCont.Enabled        := e;
    chkAcqModeTime.Enabled        := e;
    chkAcqModePoints.Enabled      := e;
    chkAcqModeOnce.Enabled        := e;
    txtAcqTimeHours.Enabled       := e;
    udAcqTimeHours.Enabled        := e;
    txtAcqTimeMinutes.Enabled     := e;
    udAcqTimeMinutes.Enabled      := e;
    txtAcqTimeSeconds.Enabled     := e;
    udAcqTimeSeconds.Enabled      := e;
    txtAcqTimePoints.Enabled      := e;

    txtBufferLength.Enabled       := e;
    lblBufferLength.Enabled       := e;
    txtSampling.Enabled           := e;
    lblSampling.Enabled           := e;
    txtDSampling.Enabled          := e;
    lblDSampling.Enabled          := e;

    btnProperties.Enabled         := e;

    cmbData.Enabled               := e;

    //prSource.CanCompile           := (s = ACQ_OUTPUT_STATE_STOPPED) and (prData <> nil);
    //prSource.CanRun               := (s = ACQ_OUTPUT_STATE_STOPPED) and (prData <> nil);

    prAcqState                    := s;

  finally

    prWritingData := False;

  end;

end;


//****************************************************************************//
//                                                                            //
//                                   D A T A                                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Acquisition_Output.prSetData(d :TQUB_QVO_QVD);
begin

  prData           := d;

  prUpdateChannels;

  if prData <> nil then begin
    prSampling     := 1 / prData.Sampling;
  end;

  txtSampling.Text := FloatToStr(prSampling / 1000);

  if prData = nil then begin
    cmbData.ItemIndex := 2;
  end else begin
    cmbData.Items.Append(ExtractFileName(d.FileName));
    cmbData.ItemIndex := cmbData.Items.Count - 1;
  end;

  prCheckChannelInfo;

  AcqState         := AcqState;

  prSetStatus;

end;

procedure TQUB_Acquisition_Output.prCheckChannelInfo;
var
  c          :Integer;
  ACh        :TQUB_Acquisition_AOChannel;
  DataEmpty  :Boolean;
begin
  if prData = nil then Exit;
  DataEmpty := prData.DataObject.Segments.TotalDataCount = 0;

  for c := 0 to prAOChannels.Count - 1 do begin
    ACh     := TQUB_Acquisition_AOChannel(prAOChannels.Objects[c]);
    if DataEmpty
    or (   (  (prData.DataChannelName[c] <> ACh.Name)
           or (prData.DataChannelUnits[c] <> ACh.Units)
           or (prData.DataChannelScaling[c] <> ACh.Scaling)
           )
       and (mrYes = MessageHelpDlg(Format('Acq. output: Data channel %d [%s/%s/%f] does not match DAQ channel [%s/%s/%f].  Change it?',
           [c, prData.DataChannelName[c], prData.DataChannelUnits[c], prData.DataChannelScaling[c], ACh.Name, ACh.Units, ACh.Scaling]),
           mtConfirmation, [mbYes, mbNo], 0, 'Warning:Data_Channel_Does_Not_Match'))
       )
    then begin
      prData.DataChannelName[c]     := ACh.Name;
      prData.DataChannelUnits[c]    := ACh.Units;
      prData.DataChannelScaling[c]  := ACh.Scaling;
    end;
  end;
end;

//Sampling
procedure TQUB_Acquisition_Output.prSetSampling(s :Double);
begin
  prSampling       := s;
  txtSampling.Text := FloatToStr(prSampling / 1000);
end;

function  TQUB_Acquisition_Output.prGetSampling :Double;
begin
  Result := prSampling;
end;

//DSampling
procedure TQUB_Acquisition_Output.prSetDSampling(s :Double);
begin
  prDSampling       := s;
  txtDSampling.Text := FloatToStr(prDSampling / 1000);
end;

function  TQUB_Acquisition_Output.prGetDSampling :Double;
begin
  Result := prDSampling;
end;

//BufferLength
procedure TQUB_Acquisition_Output.prSetBufferLength(b :Integer);
begin
  prBufferLength := b;
  txtBufferLength.Text := IntToStr(BufferLength);
end;

function  TQUB_Acquisition_Output.prGetBufferLength :Integer;
begin
  Result := prBufferLength;
end;

//Acquisition Mode
procedure TQUB_Acquisition_Output.prSetAcquisitionMode(m :TQUB_Acquisition_Mode);
begin
  prAcqMode                := m;
  chkAcqModeCont.Checked   := Integer(prAcqMode) = chkAcqModeCont.Tag;
  chkAcqModeTime.Checked   := Integer(prAcqMode) = chkAcqModeTime.Tag;
  chkAcqModePoints.Checked := Integer(prAcqMode) = chkAcqModePoints.Tag;
  chkAcqModeOnce.Checked   := Integer(prAcqMode) = chkAcqModeOnce.Tag;
end;

//Acquisition Duration
procedure TQUB_Acquisition_Output.prSetAcquisitionTimeHours(h :Integer);
begin
  prAcqTimeHours := h;
  udAcqTimeHours.Position := prAcqTimeHours;
end;

procedure TQUB_Acquisition_Output.prSetAcquisitionTimeMinutes(m :Integer);
begin
  prAcqTimeMinutes := m;
  udAcqTimeMinutes.Position := prAcqTimeMinutes;
end;

procedure TQUB_Acquisition_Output.prSetAcquisitionTimeSeconds(s :Integer);
begin
  prAcqTimeSeconds := s;
  udAcqTimeSeconds.Position := prAcqTimeSeconds;
end;

procedure TQUB_Acquisition_Output.prSetAcquisitionTimePoints(p :Integer);
begin
  prAcqTimePoints := p;
  txtAcqTimePoints.Text := IntToStr(prAcqTimePoints);
end;


procedure TQUB_Acquisition_Output.prSetStatus;
var
  t :String;
begin
  t := '';
  if prDeviceError <> DEV_ERR_NONE then begin
    if t <> '' then t := t + #13#10;
    t := t + 'Acq Device not assigned -> Options.Acquisition Hardware / Channels.';
  end;
  if prChannelsError <> CHAN_ERR_NONE then begin
    if t <> '' then t := t + #13#10;
    case prChannelsError of
      CHAN_ERR_NOCHAN:  t := t + 'No Acq Channels are assigned -> Channels or Options.Acquisition Channels.';
      CHAN_ERR_NODEV:   t := t + 'One or more Acq Channels have no Device assigned -> Options.Acquisition Hardware / Channels.';
      CHAN_ERR_DIFFDEV: t := t + 'One or more Acq Channels have different Devices (must be identical) -> Options.Acquisition Channels.';
    end;
  end;
  if prDataError <> DATA_ERR_NONE then begin
    if t <> '' then t := t + #13#10;
    case prDataError of
      DATA_ERR_UNASS:    t := t + 'Data file not assigned for recording -> File.NewDataOutputAcq.';
      DATA_ERR_DIFFCHAN: t := t + 'Data file has different Acq Channel Count than Acq Channels -> File.NewDataOutputAcq or Channels.';
      DATA_ERR_DIFFBITS: t := t + 'Data file has different Bits Resolution than Acq Device -> File.NewDataOutputAcq or Options.Acquisition Hardware.';
    end;
  end;
  if Status <> nil then begin
    Status.Visible  := t <> '';
    Status.Caption  := t;
    Status.ShowHint := True;
    Status.Hint     := t;
  end;

  if lblDevice <> nil then begin
    if prDevice = nil then begin
      lblDevice.Caption := 'Device: <none>';
    end else begin
      lblDevice.Caption := 'Device: ' + prDevice.Name;
    end;
  end;

end;


//****************************************************************************//
//                                                                            //
//                                 E V E N T S                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Acquisition_Output.lstPanelsClick(Sender: TObject);
begin
  {
  case lstPanels.ItemIndex of
    //QUB_ACQOUT_DATAFILE: panData.BringToFront;
    //QUB_ACQOUT_PROTOCOL: panProtocol.BringToFront;
    QUB_ACQOUT_SCRIPTS:  panScripts.BringToFront;
    QUB_ACQOUT_SEND:     panSend.BringToFront;
  end;
  }
end;

//------------------------------------------------------------------------------
procedure TQUB_Acquisition_Output.txtSamplingKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtSamplingExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Output.txtSamplingExit(Sender: TObject);
var
  s :Double;
begin
  try
    s := StrToFloat(TEdit(Sender).Text) * 1000;
    if (s < (MIN_SAMPLING / 1000)) or (s > (MAX_SAMPLING / 1000)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_SAMPLING / 1000) + ' and ' + FloatToStr(MAX_SAMPLING / 1000), mtError, [mbOK], 0);
    Sampling := Sampling;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if s <> Sampling then begin
    Sampling := s;
  end;
end;

//------------------------------------------------------------------------------
procedure TQUB_Acquisition_Output.txtDSamplingKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtDSamplingExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Output.txtDSamplingExit(Sender: TObject);
var
  s :Double;
begin
  try
    s := StrToFloat(TEdit(Sender).Text) * 1000;
    if (s < MIN_DSAMPLING) or (s > MAX_DSAMPLING) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between (+/-) ' + FloatToStr(MIN_DSAMPLING / 1000) + ' and ' + FloatToStr(MAX_DSAMPLING / 1000), mtError, [mbOK], 0);
    DSampling := DSampling;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if s <> DSampling then begin
    DSampling := s;
  end;
end;

//------------------------------------------------------------------------------
procedure TQUB_Acquisition_Output.txtBufferLengthKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtBufferLengthExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Output.txtBufferLengthExit(Sender: TObject);
var
  b :Integer;
begin
  try
    b := StrToInt(TEdit(Sender).Text);
    if (b < MIN_BUFFERLENGTH) or (b > MAX_BUFFERLENGTH) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_BUFFERLENGTH) + ' and ' + IntToStr(MAX_BUFFERLENGTH), mtError, [mbOK], 0);
    BufferLength := BufferLength;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if b <> BufferLength then begin
    BufferLength := b;
  end;
end;

//------------------------------------------------------------------------------
procedure TQUB_Acquisition_Output.chkAcqModeClick(Sender: TObject);
begin
  AcqMode := TQUB_Acquisition_Mode(TRadioButton(Sender).Tag);
end;

//------------------------------------------------------------------------------
procedure TQUB_Acquisition_Output.udAcqTimeHoursClick(Sender: TObject; Button: TUDBtnType);
begin
  AcqTimeHours := TUpDown(Sender).Position;
end;

procedure TQUB_Acquisition_Output.udAcqTimeMinutesClick(Sender: TObject; Button: TUDBtnType);
begin
  AcqTimeMinutes := TUpDown(Sender).Position;
end;

procedure TQUB_Acquisition_Output.udAcqTimeSecondsClick(Sender: TObject; Button: TUDBtnType);
begin
  AcqTimeSeconds := TUpDown(Sender).Position;
end;

//------------------------------------------------------------------------------
procedure TQUB_Acquisition_Output.txtAcqTimePointsKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtAcqTimePointsExit(Sender);
  end;
end;

procedure TQUB_Acquisition_Output.txtAcqTimePointsExit(Sender: TObject);
var
  p :Integer;
begin
  try
    p := StrToInt(TEdit(Sender).Text);
    if (p < MIN_TIMEPOINTS){ or (p > MAX_TIMEPOINTS)} then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_TIMEPOINTS) + ' and ' + IntToStr(MAX_TIMEPOINTS), mtError, [mbOK], 0);
    AcqTimePoints := AcqTimePoints;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> AcqTimePoints then begin
    AcqTimePoints := p;
  end;
end;

//------------------------------------------------------------------------------
procedure TQUB_Acquisition_Output.btnRunClick(Sender: TObject);
begin
  AcqStart(False{Synch}, False{Scripting}, nil);
end;

procedure TQUB_Acquisition_Output.btnStopClick(Sender: TObject);
begin
  if m_lSynchRun then begin
    prHardware.SynchIOStop(QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.Device, prDevice);
    QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.AcqStop;
  end;
  AcqStop;
end;

procedure TQUB_Acquisition_Output.btnPropertiesClick(Sender: TObject);
begin
  prPropertiesForm.ShowModal;

  prUpdateChannels;
  prCheckChannelInfo;

  prSetStatus;
end;


procedure TQUB_Acquisition_Output.prOnStartRun(Sender :TObject);
begin
  //TKPL_Collection_QUB_Interface(prKPLInterface).OnTaskChange  := TaskView.OnTaskChange;

  //TKPL_Collection_QUB_DAQ(prKPLDAQ).AcqInput                  := prAcqInput;
  //TKPL_Collection_QUB_DAQ(prKPLDAQ).AcqOutput                 := Self;

  if Data <> nil then begin
    if not Data.LockedAny then begin
      Data.LockAny;
      //with prKPLPulse as TKPL_Collection_QUB_Pulse do begin
      //  Sampling          := Data.Sampling;
      //  PulseToData.Data  := Data;
      //end;
    end;
  end;
end;

procedure TQUB_Acquisition_Output.prOnEndRun(Sender :TObject; Failed :Boolean);
begin
  if Data <> nil then begin
    Data.UnLockAny;
  end;
end;

procedure TQUB_Acquisition_Output.cmbDataDropDown(Sender: TObject);
var
  i :Integer;
begin
  while cmbData.Items.Count > 3 do cmbData.Items.Delete( cmbData.Items.Count - 1 );

  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    cmbData.Items.Append( ExtractFileName(ViewportData.ListObjects[i].FileName) );
    if prData = ViewportData.ListObjects[i] then cmbData.ItemIndex := cmbData.Items.Count - 1;
  end;
end;

procedure TQUB_Acquisition_Output.cmbDataChange(Sender: TObject);
begin
  case cmbData.ItemIndex of
    0:      ViewportAcqOutput.NewOutputAcq('');
    1:      ViewportAcqOutput.OpenOutputAcq('');
    2:      ViewportAcqOutput.AssignOutputAcq(nil);
    else    ViewportAcqOutput.AssignOutputAcq(TQUB_QVO_QVD(ViewportData.ListObjects[cmbData.ItemIndex - 3]));
  end;
end;



//Registry *********************************************************************
procedure TQUB_Acquisition_Output.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
  Ch  :TQUB_Acquisition_Channel;
  Idx :Integer;
  c   :Integer;
begin
  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Acquisition Output', True) then begin

        if Read then begin

          Sampling       := ReadRegistryFloat(Reg, 'Sampling',DEF_SAMPLING);
          DSampling      := ReadRegistryFloat(Reg, 'Digital Sampling',DEF_DSAMPLING);
          BufferLength   := ReadRegistryInteger(Reg, 'Buffer Length',DEF_BUFFERLENGTH);
          AcqMode        := TQUB_Acquisition_Mode(ReadRegistryInteger(Reg, 'Acquisition Mode',Integer(ACQ_MODE_CONT)));
          AcqTimeHours   := ReadRegistryInteger(Reg, 'Acquisition Time Hours',0);
          AcqTimeMinutes := ReadRegistryInteger(Reg, 'Acquisition Time Minutes',0);
          AcqTimeSeconds := ReadRegistryInteger(Reg, 'Acquisition Time Seconds',10);
          AcqTimePoints  := ReadRegistryInteger(Reg, 'Acquisition Time Points',10000);
        
        end else begin

          WriteFloat('Sampling', prSampling);
          WriteFloat('Digital Sampling', prDSampling);
          WriteInteger('Buffer Length', prBufferLength);
          WriteInteger('Acquisition Mode', Integer(AcqMode));
          WriteInteger('Acquisition Time Hours', AcqTimeHours);
          WriteInteger('Acquisition Time Minutes', AcqTimeMinutes);
          WriteInteger('Acquisition Time Seconds', AcqTimeSeconds);
          WriteInteger('Acquisition Time Points', AcqTimePoints);
 
        end;

      end;
      CloseKey;

      if Read then begin
      //Read

        //AOChannels
        c := 0;
        while OpenKey('Software\QUB\Acquisition Output\AOChannels\AOChannel ' + IntToStr(c + 1), False)
              and ValueExists('Name') do begin
          Idx := QUBAcqAOChannels.IndexOf(ReadString('Name'));
          if Idx >= 0 then begin
            Ch := TQUB_Acquisition_Channel(QUBAcqAOChannels.Objects[Idx]);
            prAOChannels.AddObject(Ch.Name, Ch);
          end;//Idx > 0
          CloseKey;
          Inc(c);
        end;
  
        //DOChannels
        c := 0;
        while OpenKey('Software\QUB\Acquisition Output\DOChannels\DOChannel ' + IntToStr(c + 1), False)
              and ValueExists('Name') do begin
          Idx := QUBAcqDOChannels.IndexOf(ReadString('Name'));
          if Idx >= 0 then begin
            Ch := TQUB_Acquisition_Channel(QUBAcqDOChannels.Objects[Idx]);
            prDOChannels.AddObject(Ch.Name, Ch);
          end;//Idx > 0
          CloseKey;
          Inc(c);
        end;

      end else begin
      //Write

        if OpenKey('Software\QUB\Acquisition Output', False) then begin
          DeleteKey('AOChannels');
          DeleteKey('DOChannels');
        end;
        CloseKey;

        //AOChannels
        for c := 0 to prAOChannels.Count - 1 do begin
          Ch := TQUB_Acquisition_Channel(prAOChannels.Objects[c]);
          if OpenKey('Software\QUB\Acquisition Output\AOChannels\AOChannel ' + IntToStr(c + 1), True) then begin
            WriteString('Name', Ch.Name);
          end;
          CloseKey;
        end;

        //DOChannels
        for c := 0 to prDOChannels.Count - 1 do begin
          Ch := TQUB_Acquisition_Channel(prDOChannels.Objects[c]);
          if OpenKey('Software\QUB\Acquisition Output\DOChannels\DOChannel ' + IntToStr(c + 1), True) then begin
            WriteString('Name', Ch.Name);
          end;
          CloseKey;
        end;
  
      end;//Read/Write

    end;//with Reg

  finally

    Reg.Free;

  end;
end;

procedure TQUB_Acquisition_Output.btnModeClick(Sender: TObject);
begin
  //btnModeScripts.Visible := TControl(Sender).Tag = 1;
  //btnModeAcq.Visible     := TControl(Sender).Tag = 0;
  //if TControl(Sender).Tag = 0 then begin
  //  panSend.SendToBack;
  //end else begin
  //  panScripts.SendToBack;
  //end
end;

//******************************************************************************
procedure TQUB_Acquisition_Output.SetFocusClick(Sender: TObject);
begin
  SetFocus;
end;


//******************************************************************************
procedure TQUB_Acquisition_Output.btnSynchRunClick(Sender: TObject);
begin
  if (QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.AcqState  <> ACQ_INPUT_STATE_STOPPED) or
     (QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.DataError <> IDATA_ERR_NONE)          then begin

    Application.MessageBox( 'Input must be ready and not running to start synchronous I/O', 'Start synchronized generation / acquisition ', MB_OK); 

  end else begin

    QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.AcqStart(True{Synch}, False{Scripting}, nil{Operations});
    AcqStart(True{Synch}, False{Scripting}, nil{Operations});
    prHardware.SynchIOStart(  QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.Device, prDevice );
    if QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.btnRecord.Enabled then begin
      QUB_GUIInterface.ViewportAcqInput.AcquisitionInput.AcqRecord;
    end;

  end;
end;


//******************************************************************************
end.
