unit QUB_AcquisitionOutput_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, StdCtrls, Controls, Buttons, ComCtrls, Forms, 

QUB_Global,
QUB_Acquisition_Hardware;


{$include 'QUB_AcquisitionOutput_Properties.hct'}


type


TQUB_AcquisitionOutput_Properties_Form = class(TForm)


  pctChannels                          :TPageControl;

  tabAnalogOutput                      :TTabSheet;
  tabDigitalOutput                     :TTabSheet;

  btnAddAO                             :TSpeedButton;
  btnDelAO                             :TSpeedButton;
  btnUpAO                              :TSpeedButton;
  btnDnAO                              :TSpeedButton;

  lstInstAOChannels                    :TListBox;
  lstAvailAOChannels                   :TListBox;

  btnAddDO                             :TSpeedButton;
  btnDelDO                             :TSpeedButton;
  btnUpDO                              :TSpeedButton;
  btnDnDO                              :TSpeedButton;
  lstInstDOChannels                    :TListBox;
  lstAvailDOChannels                   :TListBox;

  Label1                               :TLabel;
  Label2                               :TLabel;
  Label5                               :TLabel;
  Label6                               :TLabel;

  btnClose                             :TButton;
    btnHelp: TButton;


  procedure FormShow(Sender: TObject);

  procedure btnAddAOClick(Sender: TObject);
  procedure btnDelAOClick(Sender: TObject);
  procedure btnUpAOClick(Sender: TObject);
  procedure btnDnAOClick(Sender: TObject);

  procedure lstInstAOChannelsDblClick(Sender: TObject);
  procedure lstAvailAOChannelsDblClick(Sender: TObject);

  procedure btnAddDOClick(Sender: TObject);
  procedure btnDelDOClick(Sender: TObject);
  procedure btnUpDOClick(Sender: TObject);
  procedure btnDnDOClick(Sender: TObject);

  procedure lstInstDOChannelsDblClick(Sender: TObject);
  procedure lstAvailDOChannelsDblClick(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);
  

protected


  procedure prAddAOChannel;
  procedure prDelAOChannel;
  procedure prUpAOChannel;
  procedure prDnAOChannel;

  procedure prAddDOChannel;
  procedure prDelDOChannel;
  procedure prUpDOChannel;
  procedure prDnDOChannel;


public


  AOChannels                         :TQUB_Acquisition_Channels;
  DOChannels                         :TQUB_Acquisition_Channels;


end;


implementation


{$R *.DFM}


//AOChannels *******************************************************************
procedure TQUB_AcquisitionOutput_Properties_Form.prAddAOChannel;
var
  Ch :TQUB_Acquisition_Channel;
  i  :Integer;
begin

  //Anything selected ?
  i := lstAvailAOChannels.ItemIndex;
  if (i < 0) or (QUBAcqAOChannels.Count = 0) then Exit;

  Ch := TQUB_Acquisition_Channel(QUBAcqAOChannels.Objects[i]);

  //Already installed ?
  if AOChannels.IndexOfObject(Ch) >= 0 then Exit;

  lstInstAOChannels.Items.Add(Ch.Name);
  AOChannels.AddObject(Ch.Name, Ch);

  i := AOChannels.Count - 1;
  if lstInstAOChannels.ItemIndex >= 0 then lstInstAOChannels.Selected[lstInstAOChannels.ItemIndex] := False;
  lstInstAOChannels.ItemIndex   := i;
  lstInstAOChannels.Selected[i] := True;

end;

procedure TQUB_AcquisitionOutput_Properties_Form.prDelAOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAOChannels.ItemIndex;
  if (i < 0) or (AOChannels.Count = 0) then Exit;

  AOChannels.Delete(i);
  lstInstAOChannels.Items.Delete(i);

  if i >= AOChannels.Count then i := AOChannels.Count - 1;
  if i >= 0 then begin
    lstInstAOChannels.ItemIndex   := i;
    lstInstAOChannels.Selected[i] := True;
  end;

end;

procedure TQUB_AcquisitionOutput_Properties_Form.prUpAOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAOChannels.ItemIndex;
  if (i < 0) or (AOChannels.Count = 0) or (i = 0) then Exit;

  AOChannels.Exchange(i, i - 1);
  lstInstAOChannels.Items.Exchange(i, i - 1);

  lstInstAOChannels.Selected[i]     := False;
  lstInstAOChannels.Selected[i - 1] := True;
  lstInstAOChannels.ItemIndex       := i - 1;

end;

procedure TQUB_AcquisitionOutput_Properties_Form.prDnAOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAOChannels.ItemIndex;
  if (i < 0) or (AOChannels.Count = 0) or (i = (AOChannels.Count - 1)) then Exit;

  AOChannels.Exchange(i, i + 1);
  lstInstAOChannels.Items.Exchange(i, i + 1);

  lstInstAOChannels.Selected[i]     := False;
  lstInstAOChannels.Selected[i + 1] := True;
  lstInstAOChannels.ItemIndex       := i + 1;

end;

//DOChannels *******************************************************************
procedure TQUB_AcquisitionOutput_Properties_Form.prAddDOChannel;
var
  Ch :TQUB_Acquisition_Channel;
  i  :Integer;
begin

  //Anything selected ?
  i := lstAvailDOChannels.ItemIndex;
  if (i < 0) or (QUBAcqDOChannels.Count = 0) then Exit;

  Ch := TQUB_Acquisition_Channel(QUBAcqDOChannels.Objects[i]);

  //Already installed ?
  if DOChannels.IndexOfObject(Ch) >= 0 then Exit;

  lstInstDOChannels.Items.Add(Ch.Name);
  DOChannels.AddObject(Ch.Name, Ch);

  i := DOChannels.Count - 1;
  if lstInstDOChannels.ItemIndex >= 0 then lstInstDOChannels.Selected[lstInstDOChannels.ItemIndex] := False;
  lstInstDOChannels.ItemIndex   := i;
  lstInstDOChannels.Selected[i] := True;

end;

procedure TQUB_AcquisitionOutput_Properties_Form.prDelDOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDOChannels.ItemIndex;
  if (i < 0) or (DOChannels.Count = 0) then Exit;

  DOChannels.Delete(i);
  lstInstDOChannels.Items.Delete(i);

  if i >= DOChannels.Count then i := DOChannels.Count - 1;
  if i >= 0 then begin
    lstInstDOChannels.ItemIndex   := i;
    lstInstDOChannels.Selected[i] := True;
  end;

end;

procedure TQUB_AcquisitionOutput_Properties_Form.prUpDOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDOChannels.ItemIndex;
  if (i < 0) or (DOChannels.Count = 0) or (i = 0) then Exit;

  DOChannels.Exchange(i, i - 1);
  lstInstDOChannels.Items.Exchange(i, i - 1);

  lstInstDOChannels.Selected[i]     := False;
  lstInstDOChannels.Selected[i - 1] := True;
  lstInstDOChannels.ItemIndex       := i - 1;

end;

procedure TQUB_AcquisitionOutput_Properties_Form.prDnDOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDOChannels.ItemIndex;
  if (i < 0) or (DOChannels.Count = 0) or (i = (DOChannels.Count - 1)) then Exit;

  DOChannels.Exchange(i, i + 1);
  lstInstDOChannels.Items.Exchange(i, i + 1);

  lstInstDOChannels.Selected[i]     := False;
  lstInstDOChannels.Selected[i + 1] := True;
  lstInstDOChannels.ItemIndex       := i + 1;

end;

//******************************************************************************
procedure TQUB_AcquisitionOutput_Properties_Form.FormShow(Sender: TObject);
begin

  lstInstAOChannels.Items.Assign(AOChannels);
  lstAvailAOChannels.Items.Assign(QUBAcqAOChannels);

  lstInstDOChannels.Items.Assign(DOChannels);
  lstAvailDOChannels.Items.Assign(QUBAcqDOChannels);

end;

//AO Channels ******************************************************************
procedure TQUB_AcquisitionOutput_Properties_Form.btnAddAOClick(Sender: TObject);
begin
  prAddAOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.btnDelAOClick(Sender: TObject);
begin
  prDelAOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.btnUpAOClick(Sender: TObject);
begin
  prUpAOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.btnDnAOClick(Sender: TObject);
begin
  prDnAOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.lstInstAOChannelsDblClick(Sender: TObject);
begin
  prDelAOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.lstAvailAOChannelsDblClick(Sender: TObject);
begin
  prAddAOChannel;
end;

//DO Channels ******************************************************************
procedure TQUB_AcquisitionOutput_Properties_Form.btnAddDOClick(Sender: TObject);
begin
  prAddDOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.btnDelDOClick(Sender: TObject);
begin
  prDelDOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.btnUpDOClick(Sender: TObject);
begin
  prUpDOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.btnDnDOClick(Sender: TObject);
begin
  prDnDOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.lstInstDOChannelsDblClick(Sender: TObject);
begin
  prDelDOChannel;
end;

procedure TQUB_AcquisitionOutput_Properties_Form.lstAvailDOChannelsDblClick(Sender: TObject);
begin
  prAddDOChannel;
end;
procedure TQUB_AcquisitionOutput_Properties_Form.btnHelpClick(
  Sender: TObject);
begin
  ShowWiki('Acquisition_Output', Handle);
end;



//******************************************************************************
end.
