unit QUB_AcquisitionProtocol_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
ComCtrls, StdCtrls, Buttons,

QUB_Acquisition_Hardware;


type


TQUB_AcquisitionProtocol_PropertiesForm = class(TForm)


  btnOK                              :TButton;

  pctChannels                        :TPageControl;
  tabAnalogInput                     :TTabSheet;
  tabAnalogOutput                    :TTabSheet;
  tabDigitalInput                    :TTabSheet;
  tabDigitalOutput                   :TTabSheet;

  lstInstAIChannels                  :TListBox;
  lstAvailAIChannels                 :TListBox;

  lstInstAOChannels                  :TListBox;
  lstAvailAOChannels                 :TListBox;

  lstInstDIChannels                  :TListBox;
  lstAvailDIChannels                 :TListBox;
  lstInstDOChannels                  :TListBox;
  lstAvailDOChannels                 :TListBox;

  btnAddAI                           :TSpeedButton;
  btnDelAI                           :TSpeedButton;
  btnAIUp                            :TSpeedButton;
  btnAIDn                            :TSpeedButton;
  btnAddAO                           :TSpeedButton;
  btnDelAO                           :TSpeedButton;
  btnUpAO                            :TSpeedButton;
  btnDnAO                            :TSpeedButton;
  btnAddDI                           :TSpeedButton;
  btnDelDI                           :TSpeedButton;
  btnUpDI                            :TSpeedButton;
  btnDnDI                            :TSpeedButton;
  btnAddDO                           :TSpeedButton;
  btnDelDO                           :TSpeedButton;
  btnUpDO                            :TSpeedButton;
  btnDnDO                            :TSpeedButton;

  Label1                             :TLabel;
  Label2                             :TLabel;
  Label3                             :TLabel;
  Label4                             :TLabel;
  Label5                             :TLabel;
  Label6                             :TLabel;
  Label7                             :TLabel;
  Label8                             :TLabel;


  procedure FormShow(Sender: TObject);

  procedure btnAddClick(Sender: TObject);
  procedure btnDelClick(Sender: TObject);
  procedure btnUpClick(Sender: TObject);
  procedure btnDnClick(Sender: TObject);

  procedure lstInstAIChannelsDblClick(Sender: TObject);
  procedure lstAvailAIChannelsDblClick(Sender: TObject);
  procedure lstInstAOChannelsDblClick(Sender: TObject);
  procedure lstAvailAOChannelsDblClick(Sender: TObject);
  procedure lstInstDIChannelsDblClick(Sender: TObject);
  procedure lstAvailDIChannelsDblClick(Sender: TObject);
  procedure lstInstDOChannelsDblClick(Sender: TObject);
  procedure lstAvailDOChannelsDblClick(Sender: TObject);


protected


  procedure prAddAIChannel;
  procedure prAddAOChannel;
  procedure prAddDIChannel;
  procedure prAddDOChannel;

  procedure prDelAIChannel;
  procedure prDelAOChannel;
  procedure prDelDIChannel;
  procedure prDelDOChannel;

  procedure prUpAIChannel;
  procedure prUpAOChannel;
  procedure prUpDIChannel;
  procedure prUpDOChannel;

  procedure prDnAIChannel;
  procedure prDnAOChannel;
  procedure prDnDIChannel;
  procedure prDnDOChannel;


public


  AIChannels                         :TQUB_Acquisition_Channels;
  AOChannels                         :TQUB_Acquisition_Channels;
  DIChannels                         :TQUB_Acquisition_Channels;
  DOChannels                         :TQUB_Acquisition_Channels;


end;


implementation


{$R *.DFM}


//AIChannels *******************************************************************
procedure TQUB_AcquisitionProtocol_PropertiesForm.prAddAIChannel;
var
  Ch :TQUB_Acquisition_Channel;
  i  :Integer;
begin

  //Anything selected ?
  i := lstAvailAIChannels.ItemIndex;
  if (i < 0) or (QUBAcqAIChannels.Count = 0) then Exit;

  Ch := TQUB_Acquisition_Channel(QUBAcqAIChannels.Objects[i]);

  //Already installed ?
  if AIChannels.IndexOfObject(Ch) >= 0 then Exit;

  lstInstAIChannels.Items.Add(Ch.Name);
  AIChannels.AddObject(Ch.Name, Ch);

  i := AIChannels.Count - 1;
  if lstInstAIChannels.ItemIndex >= 0 then lstInstAIChannels.Selected[lstInstAIChannels.ItemIndex] := False;
  lstInstAIChannels.ItemIndex   := i;
  lstInstAIChannels.Selected[i] := True;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDelAIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAIChannels.ItemIndex;
  if (i < 0) or (AIChannels.Count = 0) then Exit;

  AIChannels.Delete(i);
  lstInstAIChannels.Items.Delete(i);

  if i >= AIChannels.Count then i := AIChannels.Count - 1;
  if i >= 0 then begin
    lstInstAIChannels.ItemIndex   := i;
    lstInstAIChannels.Selected[i] := True;
  end;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prUpAIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAIChannels.ItemIndex;
  if (i < 0) or (AIChannels.Count = 0) or (i = 0) then Exit;

  AIChannels.Exchange(i, i - 1);
  lstInstAIChannels.Items.Exchange(i, i - 1);

  lstInstAIChannels.Selected[i]     := False;
  lstInstAIChannels.Selected[i - 1] := True;
  lstInstAIChannels.ItemIndex       := i - 1;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDnAIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAIChannels.ItemIndex;
  if (i < 0) or (AIChannels.Count = 0) or (i = (AIChannels.Count - 1)) then Exit;

  AIChannels.Exchange(i, i + 1);
  lstInstAIChannels.Items.Exchange(i, i + 1);

  lstInstAIChannels.Selected[i]     := False;
  lstInstAIChannels.Selected[i + 1] := True;
  lstInstAIChannels.ItemIndex       := i + 1;

end;

//AOChannels
procedure TQUB_AcquisitionProtocol_PropertiesForm.prAddAOChannel;
var
  Ch :TQUB_Acquisition_Channel;
  i  :Integer;
begin

  //Anything selected ?
  i := lstAvailAOChannels.ItemIndex;
  if (i < 0) or (QUBAcqAOChannels.Count = 0) then Exit;

  Ch := TQUB_Acquisition_Channel(QUBAcqAOChannels.Objects[i]);

  //Already installed ?
  if AOChannels.IndexOfObject(Ch) >= 0 then Exit;

  lstInstAOChannels.Items.Add(Ch.Name);
  AOChannels.AddObject(Ch.Name, Ch);

  i := AOChannels.Count - 1;
  if lstInstAOChannels.ItemIndex >= 0 then lstInstAOChannels.Selected[lstInstAOChannels.ItemIndex] := False;
  lstInstAOChannels.ItemIndex   := i;
  lstInstAOChannels.Selected[i] := True;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDelAOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAOChannels.ItemIndex;
  if (i < 0) or (AOChannels.Count = 0) then Exit;

  AOChannels.Delete(i);
  lstInstAOChannels.Items.Delete(i);

  if i >= AOChannels.Count then i := AOChannels.Count - 1;
  if i >= 0 then begin
    lstInstAOChannels.ItemIndex   := i;
    lstInstAOChannels.Selected[i] := True;
  end;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prUpAOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAOChannels.ItemIndex;
  if (i < 0) or (AOChannels.Count = 0) or (i = 0) then Exit;

  AOChannels.Exchange(i, i - 1);
  lstInstAOChannels.Items.Exchange(i, i - 1);

  lstInstAOChannels.Selected[i]     := False;
  lstInstAOChannels.Selected[i - 1] := True;
  lstInstAOChannels.ItemIndex       := i - 1;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDnAOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstAOChannels.ItemIndex;
  if (i < 0) or (AOChannels.Count = 0) or (i = (AOChannels.Count - 1)) then Exit;

  AOChannels.Exchange(i, i + 1);
  lstInstAOChannels.Items.Exchange(i, i + 1);

  lstInstAOChannels.Selected[i]     := False;
  lstInstAOChannels.Selected[i + 1] := True;
  lstInstAOChannels.ItemIndex       := i + 1;

end;

//DIChannels
procedure TQUB_AcquisitionProtocol_PropertiesForm.prAddDIChannel;
var
  Ch :TQUB_Acquisition_Channel;
  i  :Integer;
begin

  //Anything selected ?
  i := lstAvailDIChannels.ItemIndex;
  if (i < 0) or (QUBAcqDIChannels.Count = 0) then Exit;

  Ch := TQUB_Acquisition_Channel(QUBAcqDIChannels.Objects[i]);

  //Already installed ?
  if DIChannels.IndexOfObject(Ch) >= 0 then Exit;

  lstInstDIChannels.Items.Add(Ch.Name);
  DIChannels.AddObject(Ch.Name, Ch);

  i := DIChannels.Count - 1;
  if lstInstDIChannels.ItemIndex >= 0 then lstInstDIChannels.Selected[lstInstDIChannels.ItemIndex] := False;
  lstInstDIChannels.ItemIndex   := i;
  lstInstDIChannels.Selected[i] := True;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDelDIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDIChannels.ItemIndex;
  if (i < 0) or (DIChannels.Count = 0) then Exit;

  DIChannels.Delete(i);
  lstInstDIChannels.Items.Delete(i);

  if i >= DIChannels.Count then i := DIChannels.Count - 1;
  if i >= 0 then begin
    lstInstDIChannels.ItemIndex   := i;
    lstInstDIChannels.Selected[i] := True;
  end;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prUpDIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDIChannels.ItemIndex;
  if (i < 0) or (DIChannels.Count = 0) or (i = 0) then Exit;

  DIChannels.Exchange(i, i - 1);
  lstInstDIChannels.Items.Exchange(i, i - 1);

  lstInstDIChannels.Selected[i]     := False;
  lstInstDIChannels.Selected[i - 1] := True;
  lstInstDIChannels.ItemIndex       := i - 1;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDnDIChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDIChannels.ItemIndex;
  if (i < 0) or (DIChannels.Count = 0) or (i = (DIChannels.Count - 1)) then Exit;

  DIChannels.Exchange(i, i + 1);
  lstInstDIChannels.Items.Exchange(i, i + 1);

  lstInstDIChannels.Selected[i]     := False;
  lstInstDIChannels.Selected[i + 1] := True;
  lstInstDIChannels.ItemIndex       := i + 1;

end;

//DOChannels
procedure TQUB_AcquisitionProtocol_PropertiesForm.prAddDOChannel;
var
  Ch :TQUB_Acquisition_Channel;
  i  :Integer;
begin

  //Anything selected ?
  i := lstAvailDOChannels.ItemIndex;
  if (i < 0) or (QUBAcqDOChannels.Count = 0) then Exit;

  Ch := TQUB_Acquisition_Channel(QUBAcqDOChannels.Objects[i]);

  //Already installed ?
  if DOChannels.IndexOfObject(Ch) >= 0 then Exit;

  lstInstDOChannels.Items.Add(Ch.Name);
  DOChannels.AddObject(Ch.Name, Ch);

  i := DOChannels.Count - 1;
  if lstInstDOChannels.ItemIndex >= 0 then lstInstDOChannels.Selected[lstInstDOChannels.ItemIndex] := False;
  lstInstDOChannels.ItemIndex   := i;
  lstInstDOChannels.Selected[i] := True;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDelDOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDOChannels.ItemIndex;
  if (i < 0) or (DOChannels.Count = 0) then Exit;

  DOChannels.Delete(i);
  lstInstDOChannels.Items.Delete(i);

  if i >= DOChannels.Count then i := DOChannels.Count - 1;
  if i >= 0 then begin
    lstInstDOChannels.ItemIndex   := i;
    lstInstDOChannels.Selected[i] := True;
  end;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prUpDOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDOChannels.ItemIndex;
  if (i < 0) or (DOChannels.Count = 0) or (i = 0) then Exit;

  DOChannels.Exchange(i, i - 1);
  lstInstDOChannels.Items.Exchange(i, i - 1);

  lstInstDOChannels.Selected[i]     := False;
  lstInstDOChannels.Selected[i - 1] := True;
  lstInstDOChannels.ItemIndex       := i - 1;

end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.prDnDOChannel;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstInstDOChannels.ItemIndex;
  if (i < 0) or (DOChannels.Count = 0) or (i = (DOChannels.Count - 1)) then Exit;

  DOChannels.Exchange(i, i + 1);
  lstInstDOChannels.Items.Exchange(i, i + 1);

  lstInstDOChannels.Selected[i]     := False;
  lstInstDOChannels.Selected[i + 1] := True;
  lstInstDOChannels.ItemIndex       := i + 1;

end;

//******************************************************************************
procedure TQUB_AcquisitionProtocol_PropertiesForm.FormShow(Sender: TObject);
begin

  lstInstAIChannels.Items.Assign(AIChannels);
  lstInstAOChannels.Items.Assign(AOChannels);
  lstInstDIChannels.Items.Assign(DIChannels);
  lstInstDOChannels.Items.Assign(DOChannels);

  lstAvailAIChannels.Items.Assign(QUBAcqAIChannels);
  lstAvailAOChannels.Items.Assign(QUBAcqAOChannels);
  lstAvailDIChannels.Items.Assign(QUBAcqDIChannels);
  lstAvailDOChannels.Items.Assign(QUBacqDOChannels);

end;

//******************************************************************************
procedure TQUB_AcquisitionProtocol_PropertiesForm.btnAddClick(Sender: TObject);
begin
  case pctChannels.ActivePage.Tag of

    0: prAddAIChannel;
    1: prAddAOChannel;
    2: prAddDIChannel;
    3: prAddDOChannel;

  end;//case
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.btnDelClick(Sender: TObject);
begin
  case pctChannels.ActivePage.Tag of

    0: prDelAIChannel;
    1: prDelAOChannel;
    2: prDelDIChannel;
    3: prDelDOChannel;

  end;//case
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.btnUpClick(Sender: TObject);
begin
  case pctChannels.ActivePage.Tag of

    0: prUpAIChannel;
    1: prUpAOChannel;
    2: prUpDIChannel;
    3: prUpDOChannel;

  end;//case
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.btnDnClick(Sender: TObject);
begin
  case pctChannels.ActivePage.Tag of

    0: prDnAIChannel;
    1: prDnAOChannel;
    2: prDnDIChannel;
    3: prDnDOChannel;

  end;//case
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstInstAIChannelsDblClick(Sender: TObject);
begin
  prDelAIChannel;
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstAvailAIChannelsDblClick(Sender: TObject);
begin
  prAddAIChannel;
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstInstAOChannelsDblClick(Sender: TObject);
begin
  prDelAOChannel;
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstAvailAOChannelsDblClick(Sender: TObject);
begin
  prAddAOChannel;
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstInstDIChannelsDblClick(Sender: TObject);
begin
  prDelDIChannel;
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstAvailDIChannelsDblClick(Sender: TObject);
begin
  prAddDIChannel;
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstInstDOChannelsDblClick(Sender: TObject);
begin
  prDelDOChannel;
end;

procedure TQUB_AcquisitionProtocol_PropertiesForm.lstAvailDOChannelsDblClick(Sender: TObject);
begin
  prAddDOChannel;
end;


//******************************************************************************
end.
