unit QUB_Acquisition_AOChannel_Frame;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls,
Forms, Dialogs, StdCtrls, ExtCtrls, Buttons,
QUB_Global, 
QUB_Acquisition_Hardware;


type


TQUB_Acquisition_AOChannelFrame = class(TFrame)


  panName                       :TPanel;
  panValue                      :TPanel;
  txtValue                      :TEdit;
  panUnits                      :TPanel;
  panWrite                      :TPanel;
  btnWrite                      :TSpeedButton;


  procedure btnWriteClick(Sender: TObject);
  procedure txtValueKeyPress(Sender: TObject; var Key: Char);
  procedure txtValueKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);


protected


  procedure prSetValue(v :Double);
  function prGetValue :Double;


public


  AOChannel                     :TQUB_Acquisition_AOChannel;
  
  OnWrite                       :TNotifyEvent;


  constructor Create(AOwner :TComponent); override;


  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;


  property Value :Double read prGetValue write prSetValue;


end;


implementation


{$R *.DFM}


//******************************************************************************
constructor TQUB_Acquisition_AOChannelFrame.Create(AOwner :TComponent);
begin
  inherited;
  Align  := alTop;
  Height := 20;
end;

procedure TQUB_Acquisition_AOChannelFrame.prSetValue(v :Double);
begin
  txtValue.Text := FloatToStr(v);
end;

function TQUB_Acquisition_AOChannelFrame.prGetValue :Double;
begin
  try
    Result := StrToFloat(txtValue.Text);
  except
    Result := AOChannel.Value;
    Value  := AOChannel.Value;
  end;
end;

procedure TQUB_Acquisition_AOChannelFrame.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited;

  if panName  = nil then Exit;
  if txtValue = nil then Exit;
	if panValue = nil then Exit;

  panName.Width  := Round(ClientWidth * 0.3);
  panValue.Width := Round(ClientWidth * 0.4);
  panUnits.Width := Round(ClientWidth * 0.2);

  txtValue.SetBounds(2, 1, panValue.ClientWidth - 4, panValue.ClientHeight - 2);
  btnWrite.SetBounds(0, 0, panWrite.ClientWidth, panWrite.ClientHeight);

end;

procedure TQUB_Acquisition_AOChannelFrame.btnWriteClick(Sender: TObject);
begin
  AOChannel.Value := Value;
  if Assigned(OnWrite) then OnWrite(Self);
end;

procedure TQUB_Acquisition_AOChannelFrame.txtValueKeyPress(Sender: TObject; var Key: Char);
begin
	if Key = #13 then begin
		Key := #0;
		AOChannel.Value := Value;
		if Assigned(OnWrite) then OnWrite(Self);
	end;
end;


//******************************************************************************
procedure TQUB_Acquisition_AOChannelFrame.txtValueKeyDown(Sender: TObject;
	var Key: Word; Shift: TShiftState);
begin
	if Key = VK_UP then begin
		//Note : this will crash if text is not float compatible
		try
			txtValue.Text := FloatToStr(_round(StrToFloat(txtValue.Text) + AOChannel.Increment,6));
		except
		end;
	end;
	if Key = VK_DOWN then begin
		try
			txtValue.Text := FloatToStr(_round(StrToFloat(txtValue.Text) - AOChannel.Increment,6));
		except
		end;
	end;
end;

end.

