{
/* Copyright 1998-2014 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
{ Acquisition Hardware

  TDeviceParams                 Device / Hardware registry settings - bits, gain,range, differential

  TQUB_Acquisition_Hardware     DLL - entry points and registry interface
  TQUB_Acquisition_Device       Hardware card
    TQUB_Acquisition_IDevice
    TQUB_Acquisition_ODevice
  TQUB_Acquisition_Devices      TStringList of Devices

  TQUB_Acquisition_Channel
    TQUB_Acquisition_AChannel
      TQUB_Acquisition_AIChannel
      TQUB_Acquisition_AOChannel
    TQUB_Acquisition_DChannel
      TQUB_Acquisition_DIChannel
      TQUB_Acquisition_DOChannel

  TQUB_Acquisition_Channels     TStringList of Channels


-------------- Globals
QUBHardwareList : TStringList;                //Global list of all hardware ( DLL's)

QUBAcqIDevices  : TQUB_Acquisition_Devices    //The global lists of all Devices
QUBAcqODevices  : TQUB_Acquisition_Devices

QUBAcqAIChannels: TQUB_Acquisition_Channels   //The global lists of all Channels
QUBAcqAOChannels: TQUB_Acquisition_Channels
QUBAcqDIChannels: TQUB_Acquisition_Channels
QUBAcqDOChannels: TQUB_Acquisition_Channels

}


unit QUB_Acquisition_Hardware;


interface


uses


Windows, SysUtils, Classes, Registry, Contnrs, Controls, Math,
AllocTypes, Alloc,
QUB_Acquisition_Global,
QUB_Console, QUB_Rich_Console;


type


//----- Registry parameters for Hardware
TDeviceParams = class( TObject )
public
  m_ChannelCount    : Integer;
  m_Channels        : Array[ 0..ACQ_MAX_CHANNELS - 1] of Integer;
  m_Gain            : Array[ 0..ACQ_MAX_CHANNELS - 1] of Double;
  m_Range           : Double;
  m_Differential    : Boolean;
  m_Bits            : Integer;
  m_Loaded          : Boolean;  // True when settings found in registry
end;

//----- Interface to a QUB DLL - entry points and registry settings
TQUB_Acquisition_IDevice      = class;
TQUB_Acquisition_ODevice      = class;

TQUB_Acquisition_Hardware     = class(TObject)
protected


  prLoaded                                :Boolean;

  prLibraryName                           :String;
  prLibraryHandle                         :HINST;
  prDllName                               :String;

  //Library Initialize/Terminate
  prLibInitialize                         :TQUB_Acquisition_LibraryInitialize;
  prLibTerminate                          :TQUB_Acquisition_LibraryTerminate;

  prGetLibNameLength                      :TQUB_Acquisition_LibraryNameLength;
  prGetLibName                            :TQUB_Acquisition_LibraryName;


  //Input **********************************************************************
  prIRunning                              :Boolean; // Not really used anywhere

  prIDeviceCount                          :Integer;
  prIDevice                               :Integer;

  prIDataChannelCount                     :Integer;
  prIDataChannel                          :Array[0..ACQ_MAX_CHANNELS - 1] of Integer;
  prIDataGain                             :Array[0..ACQ_MAX_CHANNELS - 1] of Double;

  prISampling                             :Double;

  prICaller                               :TObject;
  prICallBack                             :TQUB_Acquisition_InputCallBack;

  prGetLibIDeviceName                     :TQUB_Acquisition_InputDeviceName;
  prGetLibIChannelCount                   :TQUB_Acquisition_InputChannelCount;

  prReadDBLChannel                        :TQUB_Acquisition_ReadDBLChannel;

  prAIRead                                :TQUB_Acquisition_AIRead;
  prDIRead                                :TQUB_Acquisition_DIRead;

  prIStart                                :TQUB_Acquisition_InputStart;
  prIStop                                 :TQUB_Acquisition_InputStop;


  //Output **********************************************************************
  prORunning                              :Boolean; // Not really used anywhere

  prODeviceCount                          :Integer;
  prODevice                               :Integer;

  prODataChannelCount                     :Integer;
  prODataChannel                          :Array[0..ACQ_MAX_CHANNELS - 1] of Integer;
  prODataGain                             :Array[0..ACQ_MAX_CHANNELS - 1] of Double;

  prODigitalDataChannelCount              :Integer;
  prODigitalDataSampling                  :Double;

  prOSampling                             :Double;

  prOCaller                               :TObject;
  prOCallBack                             :TQUB_Acquisition_OutputCallBack;
  prDigitalOCallBack                      :TQUB_Acquisition_DigitalOutputCallBack;

  prGetLibODeviceName                     :TQUB_Acquisition_OutputDeviceName;
  prGetLibOChannelCount                   :TQUB_Acquisition_OutputChannelCount;
  prGetLibODigitalChannelCount            :TQUB_Acquisition_OutputDigitalChannelCount;

  prWriteDBLChannel                       :TQUB_Acquisition_WriteDBLChannel;

  prAOWrite                               :TQUB_Acquisition_AOWrite;
  prDOWrite                               :TQUB_Acquisition_DOWrite;

  prOStart                                :TQUB_Acquisition_OutputStart;
  prOStop                                 :TQUB_Acquisition_OutputStop;


  //Synchronized Input / Output
  prSynchStart                            :TQUB_Acquisition_SynchStart;
  prSynchStop                             :TQUB_Acquisition_SynchStop;

  //Params
  prIBufferLength                         :Integer;
  prOBufferLength                         :Integer;


  //Input **********************************************************************
  function  prGetIDeviceName(Device :Integer) :String;

  //Output **********************************************************************
  function  prGetODeviceName(Device :Integer) :String;


public


  GetLibIDeviceCount                    :TQUB_Acquisition_InputDeviceCount;
  GetLibODeviceCount                    :TQUB_Acquisition_OutputDeviceCount;
  GetLibIGain                           :TQUB_Acquisition_InputGain;
  GetLibIRange                          :TQUB_Acquisition_InputRange;
  GetLibIDifferential                   :TQUB_Acquisition_InputDifferential;
  GetLibIBits                           :TQUB_Acquisition_InputBits;

  GetLibOGain                           :TQUB_Acquisition_OutputGain;
  GetLibORange                          :TQUB_Acquisition_OutputRange;
  GetLibODifferential                   :TQUB_Acquisition_OutputDifferential;
  GetLibOBits                           :TQUB_Acquisition_OutputBits;

  InputDeviceParams : Array[0..ACQ_MAX_DEVICES-1] of TDeviceParams;
  OutputDeviceParams : Array[0..ACQ_MAX_DEVICES-1] of TDeviceParams;


  constructor Create(const DllName :String);
  destructor Destroy; override;

  //-- Functions Used by KPL_CollectionQUBAcqProtocol
  function IStart :Boolean; //overload;
  //Unused ... function IStart(var Params :TQUB_Acquisition_InputStart_Params) :Boolean; overload;
  //Unused ... function IStopX  :Boolean;
  function OStart :Boolean; //overload;
  //Unused ... function OStart(var Params :TQUB_Acquisition_OutputStart_Params) :Boolean; overload;
  //Unused ... function OStop  :Boolean;
  function SynchStart :Boolean;
  function SynchStop :Boolean;

  //-- Synchronized Input / Output from QUB Acquisition Output
  function SynchIOStart( IDev : TQUB_Acquisition_IDevice; ODev : TQUB_Acquisition_ODevice ) : Boolean ;
  procedure SynchIOStop( IDev : TQUB_Acquisition_IDevice; ODev : TQUB_Acquisition_ODevice ) ;

  property LibraryHandle : HINST read prLibraryHandle;

  property Loaded                                    :Boolean                                   read prLoaded;

  property LibraryName                               :String                                    read prLibraryName;
  property DllName                                   :String                                    read prDllName;


  //Input
  //property IRunning                                  :Boolean                                   read prIRunning;

  property IDeviceCount                              :Integer                                   read prIDeviceCount;
  property IDevice                                   :Integer                                   read prIDevice                    write prIDevice;
  property IDeviceName[Device :Integer]              :String                                    read prGetIDeviceName;

  property ISampling                                 :Double                                    read prISampling                  write prISampling;

  property IDataChannelCount                         :Integer                                   read prIDataChannelCount          write prIDataChannelCount;

  property ICaller                                   :TObject                                   read prICaller                    write prICaller;
  property ICallBack                                 :TQUB_Acquisition_InputCallBack            read prICallBack                  write prICallBack;

  property ReadDBLChannel                            :TQUB_Acquisition_ReadDBLChannel           read prReadDBLChannel;

  property AIRead                                    :TQUB_Acquisition_AIRead                   read prAIRead;
  property DIRead                                    :TQUB_Acquisition_DIRead                   read prDIRead;


  //Output
  //property ORunning                                  :Boolean                                   read prORunning;

  property ODeviceCount                              :Integer                                   read prODeviceCount;
  property ODevice                                   :Integer                                   read prODevice                    write prODevice;
  property ODeviceName[Device :Integer]              :String                                    read prGetODeviceName;

  property OSampling                                 :Double                                    read prOSampling                  write prOSampling;

  property ODataChannelCount                         :Integer                                   read prODataChannelCount          write prODataChannelCount;

  property ODigitalDataChannelCount                  :Integer                                   read prODigitalDataChannelCount   write prODigitalDataChannelCount;
  property ODigitalDataSampling                      :Double                                    read prODigitalDataSampling       write prODigitalDataSampling;

  property OCaller                                   :TObject                                   read prOCaller                    write prOCaller;
  property OCallBack                                 :TQUB_Acquisition_OutputCallBack           read prOCallBack                  write prOCallBack;
  property DigitalOCallBack                          :TQUB_Acquisition_DigitalOutputCallBack    read prDigitalOCallBack           write prDigitalOCallBack;

  property WriteDBLChannel                           :TQUB_Acquisition_WriteDBLChannel          read prWriteDBLChannel;

  property AOWrite                                   :TQUB_Acquisition_AOWrite                  read prAOWrite;
  property DOWrite                                   :TQUB_Acquisition_DOWrite                  read prDOWrite;

  property IBufferLength                             :Integer                                   read prIBufferLength              write prIBufferLength;
  property OBufferLength                             :Integer                                   read prOBufferLength              write prOBufferLength;


  //New style properties *******************************************************
  property HIStart                                   :TQUB_Acquisition_InputStart               read prIStart;
  property HIStop                                    :TQUB_Acquisition_InputStop                read prIStop;

  property HOStart                                   :TQUB_Acquisition_OutputStart              read prOStart;
  property HOStop                                    :TQUB_Acquisition_OutputStop               read prOStop;


end;

//----- Corresponds to card, such as 'NI6024E (1)' or 'Realtek AC97 Audio'
TQUB_Acquisition_Device       = class(TObject)
protected


  prName                      :String;

  prIndex                     :Integer;

  //Mutex objects
  prAMutex                    :THandle;
  prDMutex                    :THandle;

  //Started
  prStarted                   :Boolean;

  //Hardware
  prHardware                  :TQUB_Acquisition_Hardware;

  //Settings for (mostly) Double-Buffered Acquisition
  prASampling                 :Double;
  prDSampling                 :Double;
  prBufferLength              :Integer;
  prAChannelCount             :Integer;
  prAChannel                  :Array[0..ACQ_MAX_CHANNELS - 1] of Integer;
  prDChannelCount             :Integer;
  prDChannel                  :Array[0..ACQ_MAX_CHANNELS - 1] of Integer;
  prAGain                     :Array[0..ACQ_MAX_CHANNELS - 1] of Double;

  prCaller                    :Pointer;
  prACallback                 :Pointer;
  prDCallback                 :Pointer;

  prLockedA                   :Boolean;
  prLockedD                   :Boolean;


  procedure prSetAChannel(Chan :Integer; c :Integer);
  function prGetAChannel(Chan :Integer) :Integer;

  procedure prSetDChannel(Chan :Integer; c :Integer);
  function prGetDChannel(Chan :Integer) :Integer;

  procedure prSetAGain(Chan :Integer; g :Double);
  function prGetAGain(Chan :Integer) :Double;

  //Lock/Unlock - Return True if unsuccessful
  function LockA(TimeOut :Cardinal) :Boolean;
  function UnLockA(TimeOut :Cardinal) :Boolean;
  function LockD(TimeOut :Cardinal) :Boolean;
  function UnLockD(TimeOut :Cardinal) :Boolean;


public


  constructor Create(const Name :String; Index :Integer); virtual;
  destructor Destroy; override;


  property Name                    :String                      read prName;

  property Index                   :Integer                     read prIndex;

  property LockedA                 :Boolean                     read prLockedA;
  property LockedD                 :Boolean                     read prLockedD;

  property Started                 :Boolean                     read prStarted       write prStarted;

  //Mutex object
  property AMutex                  :THandle                     read prAMutex;
  property DMutex                  :THandle                     read prDMutex;

  //Hardware
  property Hardware                :TQUB_Acquisition_Hardware   read prHardware      write prHardware;

  //Settings for (mostly) Double-Buffered Acquisition
  property ASampling               :Double                      read prASampling     write prASampling;
  property DSampling               :Double                      read prDSampling     write prDSampling;

  property BufferLength            :Integer                     read prBufferLength  write prBufferLength;

  property AChannelCount           :Integer                     read prAChannelCount write prAChannelCount;
  property AChannel[Chan :Integer] :Integer                     read prGetAChannel   write prSetAChannel;

  property DChannelCount           :Integer                     read prDChannelCount write prDChannelCount;
  property DChannel[Chan :Integer] :Integer                     read prGetAChannel   write prSetDChannel;

  property AGain[Chan :Integer]    :Double                      read prGetAGain      write prSetAGain;

  property Caller                  :Pointer                     read prCaller        write prCaller;

  property ACallback               :Pointer                     read prACallback      write prACallback;
  property DCallback               :Pointer                     read prDCallback      write prDCallback;


end;

TQUB_Acquisition_IDevice      = class(TQUB_Acquisition_Device)
public


  constructor Create(const Name :String; Index :Integer); override;
  destructor Destroy; override;


  function InitializeParams( var IParams : TQUB_Acquisition_InputStart_Params ) : Integer;
  function Start :Integer; overload;
  function Start(var Params :TQUB_Acquisition_InputStart_Params) :Boolean; overload;
  function Stop :Boolean;


end;

TQUB_Acquisition_ODevice      = class(TQUB_Acquisition_Device)
public


  constructor Create(const Name :String; Index :Integer); override;
  destructor Destroy; override;


  function InitializeParams( var OParams : TQUB_Acquisition_OutputStart_Params ) : Boolean;
  function Start :Boolean; overload;
  function Start(var Params :TQUB_Acquisition_OutputStart_Params) :Boolean; overload;
  function Stop :Boolean;

end;

//----- Devices - Device List
TQUB_Acquisition_Devices      = class(TStringList)
public


  function FindDevice( cName : String ) : TQUB_Acquisition_Device;


end;


//----- Channel classes - Channel base class
TQUB_Acquisition_Channel      = class(TObject)
protected


  prName                      :String;
  prUpdate                    :Boolean;
  prChannel                   :Integer;
  prDevice                    :TQUB_Acquisition_Device;
  prDeviceName                :String;

  prControl                   :TControl;


public


  constructor Create(const Name :String); virtual;


  property Name               :String                   read prName          write prName;
  property Update             :Boolean                  read prUpdate        write prUpdate;
  property Channel            :Integer                  read prChannel       write prChannel;
  property Device             :TQUB_Acquisition_Device  read prDevice        write prDevice;
  property DeviceName         :String                   read prDeviceName    write prDeviceName;

  property Control            :TControl                 read prControl       write prControl;


end;


TQUB_Acquisition_AChannel     = class(TQUB_Acquisition_Channel)
protected


  prValue                     :Double;
  prUnits                     :String;
  prScaling                   :Double;
  prGain                      :Double;


  function prGetBaseScaling() : Double;
  procedure prSetBaseScaling(nBaseScale : Double);


public


  property Value              :Double                   read prValue         write prValue;
  property Units              :String                   read prUnits         write prUnits;
  property Scaling            :Double                   read prScaling       write prScaling;
  property Gain               :Double                   read prGain          write prGain;
  property BaseScaling        :Double                   read prGetBaseScaling       write prSetBaseScaling;


end;

TQUB_Acquisition_AIChannel    = class(TQUB_Acquisition_AChannel)
public


  constructor Create(const Name :String); override;
  destructor Destroy; override;


  function Read :Integer;


end;

TQUB_Acquisition_AOChannel    = class(TQUB_Acquisition_AChannel)
public


  Increment : Double;


  constructor Create(const Name :String); override;
  destructor Destroy; override;


  function Write :Integer;


end;

TQUB_Acquisition_DChannel     = class(TQUB_Acquisition_Channel)
protected


  prValue                     :Boolean;


public


  property Value              :Boolean                  read prValue         write prValue;


end;


TQUB_Acquisition_DIChannel    = class(TQUB_Acquisition_DChannel)
public


  constructor Create(const Name :String); override;
  destructor Destroy; override;


  function Read :Integer;


end;


TQUB_Acquisition_DOChannel    = class(TQUB_Acquisition_DChannel)
public


  constructor Create(const Name :String); override;
  destructor Destroy; override;


  function Write :Integer;


end;


//Channels *********************************************************************
TQUB_Acquisition_Channels     = class(TStringList)
public


  OwnChannels                 :Boolean;


  destructor Destroy; override;


end;

//----- Hardware list - delete objects before destroy
THardwareList = class(TStringList)
public


  destructor Destroy; override;


end;


//-------------------------------------------------------------------------
var


  //Global list of all hardware ( DLL's)
  QUBHardwareList : THardwareList;

  //The global lists of all Devices
  QUBAcqIDevices                :TQUB_Acquisition_Devices  = nil;
  QUBAcqODevices                :TQUB_Acquisition_Devices  = nil;

  //The global lists of all Channels
  QUBAcqAIChannels              :TQUB_Acquisition_Channels = nil;
  QUBAcqAOChannels              :TQUB_Acquisition_Channels = nil;
  QUBAcqDIChannels              :TQUB_Acquisition_Channels = nil;
  QUBAcqDOChannels              :TQUB_Acquisition_Channels = nil;


//-------------------------------------------------------------------------
implementation


uses


RegistryInterface,
QUB_Viewport_Python,
QUB_Python_Thread,
PythonEngine;


// Base Scaling is scaling at gain 1, for back compat save as gained scale
Function TQUB_Acquisition_AChannel.prGetBaseScaling() : Double;
begin
  if prGain = -1 Then
    Result := prScaling  * 2
  else if prGain <= 0 Then
    Result := prScaling
  else
    Result := prScaling  / prGain;
end;

Procedure TQUB_Acquisition_AChannel.prSetBaseScaling(nBaseScale : Double);
begin
  if prGain = -1 Then
    prScaling := nBaseScale / 2
  else if prGain <= 0 Then
    prScaling := nBaseScale
  else
    prScaling := nBaseScale * prGain;
end;


//Channel
constructor TQUB_Acquisition_Channel.Create(const Name :String);
begin
  inherited Create;

  Self.Name := Name;
end;

constructor TQUB_Acquisition_AIChannel.Create(const Name :String);
begin
  inherited;
  QUBAcqAIChannels.AddObject(Self.Name, Self);
end;

destructor TQUB_Acquisition_AIChannel.Destroy;
begin
  try
    QUBAcqAIChannels.Delete(QUBAcqAIChannels.IndexOfObject(Self));
  finally
    inherited;
  end;
end;

function TQUB_Acquisition_AIChannel.Read :Integer;
var
  iValue :Integer;
  s      :Double;
  p      :Integer;
begin

  Result := Integer((prDevice = nil) or (prDevice.Hardware = nil));
  if Result <> 0 then Exit;
  Result := Integer(prDevice.LockA(10));
  if Result <> 0 then Exit;

  try
    try
      iValue := 0;
      p      := Round(IntPower(2, prDevice.Hardware.InputDeviceParams[prDevice.Index].m_Bits));
      Result := prDevice.Hardware.AIRead(prDevice.Index, Channel, Gain, iValue);
      s      := Gain * Scaling * p / prDevice.Hardware.InputDeviceParams[prDevice.Index].m_Range;
      Value  := iValue / s;
    except
      Result := 1;
    end;
  finally
    prDevice.UnlockA(10);
  end;
end;

constructor TQUB_Acquisition_AOChannel.Create(const Name :String);
begin
  inherited;
  QUBAcqAOChannels.AddObject(Self.Name, Self);
end;

destructor TQUB_Acquisition_AOChannel.Destroy;
begin
  try
    QUBAcqAOChannels.Delete(QUBAcqAOChannels.IndexOfObject(Self));
  finally
    inherited;
  end;
end;

function TQUB_Acquisition_AOChannel.Write :Integer;
var
  iValue     :Integer;
  IMin, IMax :Integer;
  p          :Integer;
  s          :Double;
begin

  Result := Integer((prDevice = nil) or (prDevice.Hardware = nil));
  if Result <> 0 then Exit;
  Result := Integer(prDevice.LockA(10));
  if Result <> 0 then Exit;

  try
    try
      p      := Round(IntPower(2, prDevice.Hardware.OutputDeviceParams[prDevice.Index].m_Bits));
      if prDevice.Hardware.OutputDeviceParams[prDevice.Index].m_Differential then begin
        IMin := -(p div 2);
        IMax := -IMin - 1;
      end else begin
        IMin := 0;
        IMax := p - 1;
      end;
      s      := Scaling * p / prDevice.Hardware.OutputDeviceParams[prDevice.Index].m_Range;
      iValue := Round(Value  * s);
      iValue := Max(IMin, Min(IMax, iValue));
      Result := prDevice.Hardware.AOWrite(prDevice.Index, Channel, iValue);
    except
      Result := 1;
    end;
  finally
    prDevice.UnlockA(10);
  end;

end;

constructor TQUB_Acquisition_DIChannel.Create(const Name :String);
begin
  inherited;
  QUBAcqDIChannels.AddObject(Self.Name, Self);
end;

destructor TQUB_Acquisition_DIChannel.Destroy;
begin
  try
    QUBAcqDIChannels.Delete(QUBAcqDIChannels.IndexOfObject(Self));
  finally
    inherited;
  end;
end;

function TQUB_Acquisition_DIChannel.Read :Integer;
var
  iValue :Integer;
begin

  Result := Integer((prDevice = nil) or (prDevice.Hardware = nil));
  if Result <> 0 then Exit;
  Result := Integer(prDevice.LockD(10));
  if Result <> 0 then Exit;

  try
    try
    Result := prDevice.Hardware.DIRead(prDevice.Index, Channel, iValue);
    Value  := Boolean(iValue);
    except
      Result := 1;
    end;
  finally
    prDevice.UnlockD(10);
  end;

end;

destructor TQUB_Acquisition_DOChannel.Destroy;
begin
  try
    QUBAcqDOChannels.Delete(QUBAcqDOChannels.IndexOfObject(Self));
  finally
    inherited;
  end;
end;

constructor TQUB_Acquisition_DOChannel.Create(const Name :String);
begin
  inherited;
  QUBAcqDOChannels.AddObject(Self.Name, Self);
end;

function TQUB_Acquisition_DOChannel.Write :Integer;
begin

  Result := Integer((prDevice = nil) or (prDevice.Hardware = nil));
  if Result <> 0 then Exit;
  Result := Integer(prDevice.LockD(10));
  if Result <> 0 then Exit;

  try
    try
      Result := prDevice.Hardware.DOWrite(prDevice.Index, Channel, Integer(Value));
    except
      Result := 1;
    end;
  finally
    prDevice.UnlockD(10);
  end;

end;


//Channels *********************************************************************
destructor TQUB_Acquisition_Channels.Destroy;
var
  c :Integer;
begin
  try
    if OwnChannels then begin
      for c := Count - 1 downto 0 do begin
        Objects[c].Free;
      end;
    end;
  finally
    inherited;
  end;
end;

//----- Hardware list
destructor THardwareList.Destroy();
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do
      Objects[i].Free;
  finally
    inherited;
  end;
end;

//Device ***********************************************************************
constructor TQUB_Acquisition_Device.Create(const Name :String; Index :Integer);
begin

  inherited Create;

  prName   := Name;
  prIndex  := Index;

  prAMutex := CreateMutex(nil, False, nil);
  prDMutex := CreateMutex(nil, False, nil);

end;

destructor TQUB_Acquisition_Device.Destroy;
begin

  try

    CloseHandle(prAMutex);
    CloseHandle(prDMutex);

  finally

    inherited;

  end;

end;


//Lock, Analog
function TQUB_Acquisition_Device.LockA(TimeOut :Cardinal) :Boolean;
begin
  Result := LockedA;
  if Result then Exit;
  Result := WaitForSingleObject(prAMutex, TimeOut) <> WAIT_OBJECT_0;
  if not Result then begin
    prLockedA := True;
    ReleaseMutex(prAMutex);
  end;
end;

function TQUB_Acquisition_Device.UnLockA(TimeOut :Cardinal) :Boolean;
begin
  Result := False;
  if not LockedA then Exit;
  Result := WaitForSingleObject(prAMutex, TimeOut) <> WAIT_OBJECT_0;
  if not Result then begin
    prLockedA := False;
    ReleaseMutex(prAMutex);
  end;
end;

//Lock, Digital
function TQUB_Acquisition_Device.LockD(TimeOut :Cardinal) :Boolean;
begin
  Result := LockedD;
  if Result then Exit;
  Result := WaitForSingleObject(prDMutex, TimeOut) <> WAIT_OBJECT_0;
  if not Result then begin
    prLockedD := True;
    ReleaseMutex(prDMutex);
  end;
end;

function TQUB_Acquisition_Device.UnLockD(TimeOut :Cardinal) :Boolean;
begin
  Result := False;
  if not LockedD then Exit;
  Result := WaitForSingleObject(prDMutex, TimeOut) <> WAIT_OBJECT_0;
  if not Result then begin
    prLockedD := False;
    ReleaseMutex(prDMutex);
  end;
end;


//******************************************************************************
procedure TQUB_Acquisition_Device.prSetAChannel(Chan :Integer; c :Integer);
begin
  prAChannel[Chan] := c;
end;

function TQUB_Acquisition_Device.prGetAChannel(Chan :Integer) :Integer;
begin
  Result := prAChannel[Chan];
end;

procedure TQUB_Acquisition_Device.prSetDChannel(Chan :Integer; c :Integer);
begin
  prDChannel[Chan] := c;
end;

function TQUB_Acquisition_Device.prGetDChannel(Chan :Integer) :Integer;
begin
  Result := prDChannel[Chan];
end;

procedure TQUB_Acquisition_Device.prSetAGain(Chan :Integer; g :Double);
begin
  prAGain[Chan] := g;
end;

function TQUB_Acquisition_Device.prGetAGain(Chan :Integer) :Double;
begin
  Result := prAGain[Chan];
end;


//IDevice **********************************************************************
constructor TQUB_Acquisition_IDevice.Create(const Name :String; Index :Integer);
begin
  inherited Create(Name, Index);
  QUBAcqIDevices.AddObject(Name, Self);
end;

destructor TQUB_Acquisition_IDevice.Destroy;
begin
  try
    QUBAcqIDevices.Delete(QUBAcqIDevices.IndexOfObject(Self));
  finally
    inherited;
  end;
end;

//----- Fill in structure of device settings which is passed to Acq. DLL
function TQUB_Acquisition_IDevice.InitializeParams( var IParams : TQUB_Acquisition_InputStart_Params ) : Integer;
begin
  Result := 0;
  if (prHardware = nil) then begin
    Result := -1654;
    Exit;
  end else if (LockedA) then begin
    Result := -1655;
    Exit;
  end else if (LockedD) then begin
    Result := -1656;
    Exit;
  end else if (prStarted) then begin
    Result := -1657;
    Exit;
  end;

  //try
    If (LockA(100)) or (LockD(100)) then begin
      Result := -1658;
      Exit;
    end;

    IParams.Size                := SizeOf(IParams);
    IParams.ChannelCount        := prAChannelCount;
    IParams.Channels            := @prAChannel[0];
    IParams.Gain                := @prAGain[0];
    IParams.Bits                := prHardware.InputDeviceParams[Index].m_Bits;
    IParams.Sampling            := prASampling;
    IParams.CallBack            := prACallBack;
    IParams.Caller              := prCaller;
    IParams.BufferLength        := prBufferLength;
    IParams.Buffer              := nil;
    IParams.SingleBuffer        := False;
    IParams.BufferCount         := 0;
    IParams.BufferElemSize      := 0;
  //finally

  //end;
end;

function TQUB_Acquisition_IDevice.Start : Integer;
var
  IParams : TQUB_Acquisition_InputStart_Params;
begin

  try
  
    Result    := InitializeParams( IParams );
    if Result = 0 then begin
      Result    := prHardware.HIStart(Index, IParams);
    end;
    prStarted := not Boolean(Result);

  finally

    if not prStarted then begin
      if LockedA then UnLockA(100);
      if LockedD then UnLockD(100);
    end;

  end;

end;

function TQUB_Acquisition_IDevice.Start(var Params :TQUB_Acquisition_InputStart_Params) :Boolean;
begin

  Result := (prHardware = nil) or (LockedA) or (LockedD) or (prStarted);
  if Result then Exit;

  try

    Result := (LockA(100)) or (LockD(100));
    if Result then Exit;

    prStarted := not Boolean(prHardware.HIStart(Index, Params));
    Result    := not prStarted;

  finally

    if Result then begin
      if LockedA then UnLockA(100);
      if LockedD then UnLockD(100);
    end;

  end;

end;

function TQUB_Acquisition_IDevice.Stop :Boolean;
begin

  Result := (prHardware = nil);
  if Result then Exit;

  if not prStarted then Exit;

  try

    prStarted := Boolean(prHardware.HIStop(Index));
    Result    := prStarted;

  finally

    if not Result then begin
      if LockedA then UnLockA(100);
      if LockedD then UnLockD(100);
    end;

  end;

end;


//ODevice **********************************************************************
constructor TQUB_Acquisition_ODevice.Create(const Name :String; Index :Integer);
begin
  inherited Create(Name, Index);
  QUBAcqODevices.AddObject(Name, Self);
end;

destructor TQUB_Acquisition_ODevice.Destroy;
begin
  try
    QUBAcqODevices.Delete(QUBAcqODevices.IndexOfObject(Self));
  finally
    inherited;
  end;
end;

//----- Fill in structure of device settings which is passed to Acq. DLL
function TQUB_Acquisition_ODevice.InitializeParams( var OParams : TQUB_Acquisition_OutputStart_Params ) : Boolean;
begin
  Result := True;
  if (prHardware = nil) or (LockedA) or (LockedD) or (prStarted) then Exit;

  try
    if (LockA(100)) or (LockD(100)) then Exit;

    OParams.Size                := SizeOf(OParams);
    OParams.ChannelCount        := prAChannelCount;
    OParams.Channels            := @prAChannel[0];
    OParams.Gain                := @prAGain[0];
    OParams.DigitalChannelCount := prDChannelCount;
    OParams.Bits                := prHardware.InputDeviceParams[Index].m_Bits;
    OParams.Sampling            := prASampling;
    OParams.DigitalSampling     := prDSampling;
    OParams.CallBack            := prACallBack;
    OParams.DigitalCallBack     := prDCallBack;
    OParams.Caller              := prCaller;
    OParams.BufferLength        := prBufferLength;
    OParams.Buffer              := nil;
    OParams.SingleBuffer        := False;
    OParams.BufferCount         := 0;
    OParams.BufferElemSize      := 0;
    Result := False;
  finally

  end;
end;

//---- Output start for QUB_AcquisitionOutput
function TQUB_Acquisition_ODevice.Start :Boolean;
var
  i               :Integer;
  OParams : TQUB_Acquisition_OutputStart_Params;
begin

  Result      := True;
  try
    Result := InitializeParams( OParams );
    if Result then Exit;

    i         := prHardware.HOStart(Index, OParams);
    prStarted := i = 0;
    Result    := not prStarted;

    if not prStarted then begin
      PythonThread.BeginPython;
      GetPythonEngine.Run_CommandAsString('acq_err = ' + IntToStr(i), single_input);
      PythonThread.EndPython;
    end;

  finally

    if Result then begin
      if LockedA then UnLockA(100);
      if LockedD then UnLockD(100);
    end;

  end;

end;

//---- Output start for KPL_ODE
function TQUB_Acquisition_ODevice.Start(var Params :TQUB_Acquisition_OutputStart_Params) :Boolean;
begin

  Result := (prHardware = nil) or (LockedA) or (LockedD) or (prStarted);
  if Result then Exit;

  try

    Result := (LockA(100)) or (LockD(100));
    if Result then Exit;

    prStarted := not Boolean(prHardware.HOStart(Index, Params));
    Result    := not prStarted;

  finally

    if Result then begin
      if LockedA then UnLockA(100);
      if LockedD then UnLockD(100);
    end;

  end;

end;

function TQUB_Acquisition_ODevice.Stop :Boolean;
begin

  Result := (prHardware = nil);
  if Result then Exit;

  if not prStarted then Exit;

  try

    prStarted := Boolean(prHardware.HOStop(Index));
    Result    := prStarted;

  finally

    if not Result then begin
      if LockedA then UnLockA(100);
      if LockedD then UnLockD(100);
    end;

  end;

end;

//---- Synchronized Input / Output start
function TQUB_Acquisition_Hardware.SynchIOStart( IDev : TQUB_Acquisition_IDevice; ODev : TQUB_Acquisition_ODevice ) : Boolean ;
  var
    IParams          :TQUB_Acquisition_InputStart_Params;
    OParams          :TQUB_Acquisition_OutputStart_Params;
    iInputStarted    :Integer;
    iOutputStarted   :Integer;
  begin

    Result := True;
    try
      //----- Initialize the Input and Output control structures for the DLL
      if IDev.InitializeParams( IParams ) <> 0 then begin
        Exit;
      end;
      if ODev.InitializeParams( OParams ) then begin
        IDev.UnLockA(100);
        IDev.UnLockD(100);
        Exit;
      end;

      Result := Boolean( prSynchStart( IDev.Index, ODev.Index, IParams, OParams, iInputStarted, iOutputStarted));
      IDev.prStarted := not Result;
      ODev.prStarted := not Result;

    finally
      //-- if error occurred, unlock everything.
      if Result then begin
        if IDev.LockedA then IDev.UnLockA(100);
        if IDev.LockedD then IDev.UnLockD(100);
        if ODev.LockedA then ODev.UnLockA(100);
        if ODev.LockedD then ODev.UnLockD(100);
      end;
    end;
  end;

procedure TQUB_Acquisition_Hardware.SynchIOStop( IDev : TQUB_Acquisition_IDevice; ODev : TQUB_Acquisition_ODevice );
  begin
    prSynchStop;
    if IDev.LockedA then IDev.UnLockA(100);
    if IDev.LockedD then IDev.UnLockD(100);
    if ODev.LockedA then ODev.UnLockA(100);
    if ODev.LockedD then ODev.UnLockD(100);
  end;

//Hardware *********************************************************************
constructor TQUB_Acquisition_Hardware.Create(const DllName :String);
var
  Device : TQUB_Acquisition_Device;
  d : Integer;
  iDev : Integer;
begin

  inherited Create;

  prDllName             := DllName;

  prLibraryHandle       := 0;

  try

    prLibraryHandle := LoadLibrary(PChar(DllName));

    if prLibraryHandle <> 0 then begin

      prLoaded := True;
      for iDev := 0 to ACQ_MAX_DEVICES-1 do begin
        InputDeviceParams[iDev] := TDeviceParams.Create();
        InputDeviceParams[iDev].m_Loaded := False;

        OutputDeviceParams[iDev] := TDeviceParams.Create();
        OutputDeviceParams[iDev].m_Loaded := False;
      end;

      try


        //Find Library functions addresses *************************************

        prLibInitialize := GetProcAddress(prLibraryHandle, 'LibraryInitialize');
        if @prLibInitialize = nil then begin
          prLibInitialize              := GetProcAddress(prLibraryHandle, '_LibraryInitialize@0');
          prLibTerminate               := GetProcAddress(prLibraryHandle, '_LibraryTerminate@0');

          prGetLibNameLength           := GetProcAddress(prLibraryHandle, '_LibraryNameLength@0');
          prGetLibName                 := GetProcAddress(prLibraryHandle, '_LibraryName@4');

          //Input
          GetLibIDeviceCount           := GetProcAddress(prLibraryHandle, '_InputDeviceCount@0');
          prGetLibIDeviceName          := GetProcAddress(prLibraryHandle, '_InputDeviceName@8');
          prGetLibIChannelCount        := GetProcAddress(prLibraryHandle, '_InputChannelCount@4');
          GetLibIGain                  := GetProcAddress(prLibraryHandle, '_InputGain@8');
          GetLibIRange                 := GetProcAddress(prLibraryHandle, '_InputRange@4');
          GetLibIDifferential          := GetProcAddress(prLibraryHandle, '_InputDifferential@4');
          GetLibIBits                  := GetProcAddress(prLibraryHandle, '_InputBits@4');

          prReadDBLChannel             := GetProcAddress(prLibraryHandle, '_ReadDBLChannel@20');

          prAIRead                     := GetProcAddress(prLibraryHandle, '_AIRead@20');
          prDIRead                     := GetProcAddress(prLibraryHandle, '_DIRead@12');

          prIStart                     := GetProcAddress(prLibraryHandle, '_InputStart@8');
          prIStop                      := GetProcAddress(prLibraryHandle, '_InputStop@4');

          //Output
          GetLibODeviceCount           := GetProcAddress(prLibraryHandle, '_OutputDeviceCount@0');
          prGetLibODeviceName          := GetProcAddress(prLibraryHandle, '_OutputDeviceName@8');
          prGetLibOChannelCount        := GetProcAddress(prLibraryHandle, '_OutputChannelCount@4');
          prGetLibODigitalChannelCount := GetProcAddress(prLibraryHandle, '_OutputDigitalChannelCount@4');
          GetLibOGain                  := GetProcAddress(prLibraryHandle, '_OutputGain@8');
          GetLibORange                 := GetProcAddress(prLibraryHandle, '_OutputRange@4');
          GetLibODifferential          := GetProcAddress(prLibraryHandle, '_OutputDifferential@4');
          GetLibOBits                  := GetProcAddress(prLibraryHandle, '_OutputBits@4');

          prWriteDBLChannel            := GetProcAddress(prLibraryHandle, '_WriteDBLChannel@16');

          prAOWrite                    := GetProcAddress(prLibraryHandle, '_AOWrite@12');
          prDOWrite                    := GetProcAddress(prLibraryHandle, '_DOWrite@12');

          prOStart                     := GetProcAddress(prLibraryHandle, '_OutputStart@8');
          prOStop                      := GetProcAddress(prLibraryHandle, '_OutputStop@4');

          //Synch
          prSynchStart                 := GetProcAddress(prLibraryHandle, '_SynchStart@24');
          prSynchStop                  := GetProcAddress(prLibraryHandle, '_SynchStop@0');

        end else Begin

          prLibTerminate               := GetProcAddress(prLibraryHandle, 'LibraryTerminate');

          prGetLibNameLength           := GetProcAddress(prLibraryHandle, 'LibraryNameLength');
          prGetLibName                 := GetProcAddress(prLibraryHandle, 'LibraryName');

          //Input
          GetLibIDeviceCount           := GetProcAddress(prLibraryHandle, 'InputDeviceCount');
          prGetLibIDeviceName          := GetProcAddress(prLibraryHandle, 'InputDeviceName');
          prGetLibIChannelCount        := GetProcAddress(prLibraryHandle, 'InputChannelCount');
          GetLibIGain                  := GetProcAddress(prLibraryHandle, 'InputGain');
          GetLibIRange                 := GetProcAddress(prLibraryHandle, 'InputRange');
          GetLibIDifferential          := GetProcAddress(prLibraryHandle, 'InputDifferential');
          GetLibIBits                  := GetProcAddress(prLibraryHandle, 'InputBits');

          prReadDBLChannel             := GetProcAddress(prLibraryHandle, 'ReadDBLChannel');

          prAIRead                     := GetProcAddress(prLibraryHandle, 'AIRead');
          prDIRead                     := GetProcAddress(prLibraryHandle, 'DIRead');

          prIStart                     := GetProcAddress(prLibraryHandle, 'InputStart');
          prIStop                      := GetProcAddress(prLibraryHandle, 'InputStop');

          //Output
          GetLibODeviceCount           := GetProcAddress(prLibraryHandle, 'OutputDeviceCount');
          prGetLibODeviceName          := GetProcAddress(prLibraryHandle, 'OutputDeviceName');
          prGetLibOChannelCount        := GetProcAddress(prLibraryHandle, 'OutputChannelCount');
          prGetLibODigitalChannelCount := GetProcAddress(prLibraryHandle, 'OutputDigitalChannelCount');
          GetLibOGain                  := GetProcAddress(prLibraryHandle, 'OutputGain');
          GetLibORange                 := GetProcAddress(prLibraryHandle, 'OutputRange');
          GetLibODifferential          := GetProcAddress(prLibraryHandle, 'OutputDifferential');
          GetLibOBits                  := GetProcAddress(prLibraryHandle, 'OutputBits');

          prWriteDBLChannel            := GetProcAddress(prLibraryHandle, 'WriteDBLChannel');

          prAOWrite                    := GetProcAddress(prLibraryHandle, 'AOWrite');
          prDOWrite                    := GetProcAddress(prLibraryHandle, 'DOWrite');

          prOStart                     := GetProcAddress(prLibraryHandle, 'OutputStart');
          prOStop                      := GetProcAddress(prLibraryHandle, 'OutputStop');

          //Synch
          prSynchStart                 := GetProcAddress(prLibraryHandle, 'SynchStart');
          prSynchStop                  := nil;//???
        end;


        //Check if all addresses are valid
        if (@prLibInitialize                      = nil) or
           (@prLibTerminate                       = nil) or

           (@prGetLibNameLength                   = nil) or
           (@prGetLibName                         = nil) or

           //Input
           (@GetLibIDeviceCount                   = nil) or
           (@prGetLibIDeviceName                  = nil) or
           (@prGetLibIChannelCount                = nil) or
           (@GetLibIGain                          = nil) or
           (@GetLibIRange                         = nil) or
           (@GetLibIDifferential                  = nil) or
           (@GetLibIBits                          = nil) or
           (@prReadDBLChannel                     = nil) or
           (@prAIRead                             = nil) or
           (@prDIRead                             = nil) or
           (@prIStart                             = nil) or
           (@prIStop                              = nil) or

           //Output
           (@GetLibODeviceCount                   = nil) or
           (@prGetLibODeviceName                  = nil) or
           (@prGetLibOChannelCount                = nil) or
           (@prGetLibODigitalChannelCount         = nil) or
           (@GetLibOGain                          = nil) or
           (@GetLibORange                         = nil) or
           (@GetLibODifferential                  = nil) or
           (@GetLibOBits                          = nil) or
           (@prWriteDBLChannel                    = nil) or
           (@prAOWrite                            = nil) or
           (@prDOWrite                            = nil) or
           (@prOStart                             = nil) or
           (@prOStop                              = nil) or

           //Synch
           (@prSynchStart                         = nil) then begin

          prLoaded := False;


        end else begin


          if prLibInitialize <> 0 then begin
            prLoaded := False;
          end;

          if prLoaded then begin
            SetLength(prLibraryName, prGetLibNameLength);
            prGetLibName(PChar(prLibraryName));
          end;

          if prLoaded then begin
            
            //Input
            prIDeviceCount  := GetLibIDeviceCount;
            if prIDeviceCount > ACQ_MAX_DEVICES then prIDeviceCount := ACQ_MAX_DEVICES;
            for d := 0 to prIDeviceCount - 1 do begin
              InputDeviceParams[d].m_ChannelCount := Min(ACQ_MAX_CHANNELS, prGetLibIChannelCount(d));
            end;

            if (prIDeviceCount=0) and (Console<>nil) then
              Console.NewEntry('DAQ library ' + DLLName + ' does not support input devices (no hardware connected?).', COLOR_WARNING{COLOR_ERROR});

            //Output
            prODeviceCount  := GetLibODeviceCount;
            if prODeviceCount > ACQ_MAX_DEVICES then prODeviceCount := ACQ_MAX_DEVICES;
            for d := 0 to prODeviceCount - 1 do
              OutputDeviceParams[d].m_ChannelCount := Min(ACQ_MAX_CHANNELS,prGetLibOChannelCount(d));

            if (prODeviceCount=0) and (Console<>nil) then
              Console.NewEntry('DAQ library ' + DLLName + ' does not support output devices (no hardware connected?).', COLOR_WARNING{COLOR_ERROR});

          end;

        end;


        //Devices
        if prLoaded then begin

          //Input
          for d := 0 to prIDeviceCount - 1 do begin
            Device          := TQUB_Acquisition_IDevice.Create(prGetIDeviceName(d), d);
            Device.Hardware := Self;
          end;

          //Output
          for d := 0 to prODeviceCount - 1 do begin
            Device          := TQUB_Acquisition_ODevice.Create(prGetODeviceName(d), d);
            Device.Hardware := Self;
          end;

        end;


      except

        prLoaded := False;

      end;

    end;

  finally

    if (not prLoaded) and (Console <> nil) then begin
      Console.NewEntry('DAQ library ' + DLLName + ' is discarded.', COLOR_ERROR);
    end;
    
  end;

end;

destructor TQUB_Acquisition_Hardware.Destroy;
var
  iDev : Integer;
begin

  try

    try
      if @prLibTerminate <> nil then prLibTerminate;
      FreeLibrary(prLibraryHandle);
    except
    end;

    prLoaded := False;

    for iDev := 0 to ACQ_MAX_DEVICES-1 do begin
      InputDeviceParams[iDev].Free;
      OutputDeviceParams[iDev].Free;
    end;

  finally

    inherited;

  end;

end;


//****************************************************************************//
//                                                                            //
//                           A C Q U I S I T I O N                            //
//                                                                            //
//****************************************************************************//

//Input
function TQUB_Acquisition_Hardware.IStart :Boolean;
var
  i :Integer;
  IParams : TQUB_Acquisition_InputStart_Params;
begin

  for i := 0 to InputDeviceParams[IDevice].m_ChannelCount - 1 do begin
    prIDataGain[i]    := InputDeviceParams[IDevice].m_Gain[i];
    prIDataChannel[i] := InputDeviceParams[IDevice].m_Channels[i];
  end;

  IParams.Size                := SizeOf(IParams);
  IParams.ChannelCount        := InputDeviceParams[IDevice].m_ChannelCount;
  IParams.Channels            := @prIDataChannel[0];
  IParams.Gain                := @prIDataGain[0];
  IParams.Bits                := InputDeviceParams[IDevice].m_Bits;
  IParams.Sampling            := ISampling;
  IParams.CallBack            := ICallBack;
  IParams.Caller              := ICaller;
  IParams.BufferLength        := prIBufferLength;
  IParams.Buffer              := nil;
  IParams.SingleBuffer        := False;
  IParams.BufferCount         := 0;
  IParams.BufferElemSize      := 0;

  prIRunning := not Boolean(prIStart(IDevice, IParams));
  Result := not prIRunning;
end;

(* Unused
function TQUB_Acquisition_Hardware.IStart(var Params :TQUB_Acquisition_InputStart_Params) :Boolean;
begin
  prIRunning := not Boolean(prIStart(IDevice, Params));
  Result := not prIRunning;
end;

function TQUB_Acquisition_Hardware.IStop :Boolean;
begin
  Result := Boolean(prIStop(IDevice));
  if not Result then prIRunning := False;
end;
*)


//Output
function TQUB_Acquisition_Hardware.OStart :Boolean;
var
  i :Integer;
  OParams : TQUB_Acquisition_OutputStart_Params;
begin

  for i := 0 to prODataChannelCount - 1 do begin
    prODataGain[i] := OutputDeviceParams[prODevice].m_Gain[prODataChannel[i]];
  end;

  OParams.Size                := SizeOf(OParams);
  OParams.ChannelCount        := ODataChannelCount;
  OParams.DigitalChannelCount := ODigitalDataChannelCount;
  OParams.Channels            := @prODataChannel[0];
  OParams.Gain                := @prODataGain[0];
  OParams.Bits                := OutputDeviceParams[prODevice].m_Bits;
  OParams.Sampling            := OSampling;
  OParams.DigitalSampling     := ODigitalDataSampling;
  OParams.CallBack            := OCallBack;
  OParams.DigitalCallBack     := DigitalOCallBack;
  OParams.Caller              := OCaller;
  OParams.BufferLength        := prOBufferLength;
  OParams.Buffer              := nil;
  OParams.SingleBuffer        := False;
  OParams.BufferCount         := 0;
  OParams.BufferElemSize      := 0;

  i                     := prOStart(ODevice, OParams);
  prORunning            := i = 0;
  Result                := not prORunning;
  if not prORunning then begin
    PythonThread.BeginPython;
    GetPythonEngine.Run_CommandAsString('acq_err = ' + IntToStr(i), single_input);
    PythonThread.EndPython;
  end;
end;

(* unused
function TQUB_Acquisition_Hardware.OStart(var Params :TQUB_Acquisition_OutputStart_Params) :Boolean;
var
  i           :Integer;
begin
  i           := prOStart(ODevice, Params);
  prORunning  := i = 0;
  Result      := not prORunning;
  if not prORunning then begin
    PythonThread.BeginPython;
    GetPythonEngine.Run_CommandAsString('acq_err = ' + IntToStr(i), single_input);
    PythonThread.EndPython;
  end;
end;

function TQUB_Acquisition_Hardware.OStop :Boolean;
begin
  Result := Boolean(prOStop(ODevice));
  if not Result then prORunning := False;
end;
*)

//Synch
function TQUB_Acquisition_Hardware.SynchStart :Boolean;
var
  IParams  :TQUB_Acquisition_InputStart_Params;
  OParams  :TQUB_Acquisition_OutputStart_Params;
  i_r, o_r :Integer;
  i        :Integer;
begin

  for i := 0 to prIDataChannelCount - 1  do begin
    prIDataGain[i] := InputDeviceParams[prIDevice].m_Gain[prIDataChannel[i]];
  end;

  for i := 0 to prODataChannelCount - 1 do begin
    prODataGain[i] := OutputDeviceParams[prODevice].m_Gain[prODataChannel[i]];
  end;

  IParams.Size                 := SizeOf(IParams);
  IParams.ChannelCount         := IDataChannelCount;
  IParams.Channels             := @prIDataChannel[0];
  IParams.Gain                 := @prIDataGain[0];
  IParams.Bits                 := InputDeviceParams[prIDevice].m_Bits;
  IParams.Sampling             := ISampling;
  IParams.CallBack             := ICallBack;
  IParams.Caller               := ICaller;
  IParams.BufferLength         := prIBufferLength;
  IParams.Buffer               := nil;
  IParams.SingleBuffer         := False;
  IParams.BufferCount          := 0;
  IParams.BufferElemSize       := 0;

  OParams.Size                 := SizeOf(OParams);
  OParams.ChannelCount         := ODataChannelCount;
  OParams.DigitalChannelCount  := ODigitalDataChannelCount;
  OParams.Channels             := @prODataChannel[0];
  OParams.Gain                 := @prODataGain[0];
  OParams.Bits                 := OutputDeviceParams[prODevice].m_Bits;
  OParams.Sampling             := OSampling;
  OParams.DigitalSampling      := ODigitalDataSampling;
  OParams.CallBack             := OCallBack;
  OParams.DigitalCallBack      := DigitalOCallBack;
  OParams.Caller               := OCaller;
  OParams.BufferLength         := prOBufferLength;
  OParams.Buffer               := nil;
  OParams.SingleBuffer         := False;
  OParams.BufferCount          := 0;
  OParams.BufferElemSize       := 0;

  Result      := Boolean(prSynchStart(IDevice, ODevice, IParams, OParams, i_r, o_r));
  prIRunning  := not Boolean(i_r);
  prORunning  := not Boolean(o_r);

end;

function TQUB_Acquisition_Hardware.SynchStop :Boolean;
begin
  Result := Boolean(prSynchStop());
  if not Result then begin
    prORunning := False;
    prIRunning := False;
  end;
end;

//****************************************************************************//
//                                                                            //
//                             P R O P E R T I E S                            //
//                                                                            //
//****************************************************************************//

//Input ************************************************************************

//Device
function TQUB_Acquisition_Hardware.prGetIDeviceName(Device :Integer) :String;
var
  Name : String;
begin
  SetLength(Name, 256);
  prGetLibIDeviceName(Device, PChar(Name));
  SetLength(Name, Pos(#0, Name) - 1);
  Result := Name;
end;

//Output ***********************************************************************

//Device
function TQUB_Acquisition_Hardware.prGetODeviceName(Device :Integer) :String;
var
  Name :String;
begin
  SetLength(Name, 256);
  prGetLibODeviceName(Device, PChar(Name));
  SetLength(Name, Pos(#0, Name) - 1);
  Result := Name;
end;

function TQUB_Acquisition_Devices.FindDevice( cName : String ) : TQUB_Acquisition_Device;
  var
    iDev : Integer;
  begin
    iDev := IndexOf(cName);
    if iDev >= 0 then
    Result := TQUB_Acquisition_Device(Objects[iDev])
    else
    Result := nil;
  end;


//******************************************************************************
initialization


QUBHardwareList := THardwareList.Create();

QUBAcqIDevices   := TQUB_Acquisition_Devices.Create;
QUBAcqODevices   := TQUB_Acquisition_Devices.Create;

QUBAcqAIChannels := TQUB_Acquisition_Channels.Create;
QUBAcqAOChannels := TQUB_Acquisition_Channels.Create;
QUBAcqDIChannels := TQUB_Acquisition_Channels.Create;
QUBAcqDOChannels := TQUB_Acquisition_Channels.Create;

QUBAcqAIChannels.OwnChannels := True;
QUBAcqAOChannels.OwnChannels := True;
QUBAcqDIChannels.OwnChannels := True;
QUBAcqDOChannels.OwnChannels := True;

HardwareReadRegistry();
ChannelsReadRegistry();


finalization


QUBHardwareList.Free();

QUBAcqIDevices.Free;
QUBAcqODevices.Free;

QUBAcqAIChannels.Free;
QUBAcqAOChannels.Free;
QUBAcqDIChannels.Free;
QUBAcqDOChannels.Free;


//******************************************************************************
end.
