unit QUB_ActionSelector_Lists;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls, Menus,

QUB_Action, QUB_ActionSelector;


type


TQUB_ActionSelector_Lists = class(TQUB_ActionSelector)
protected


  procedure prLClickListSegs(OnActionDone :TNotifyEvent);
  procedure prLClickListIdl(OnActionDone :TNotifyEvent);
  procedure prLClickNewList(OnActionDone :TNotifyEvent);
  procedure prRClickNewList(OnActionDone :TNotifyEvent);
  procedure prLClickDelList(OnActionDone :TNotifyEvent);
  procedure prLClickAddSel(OnActionDone :TNotifyEvent);
  procedure prRClickAddSel(OnActionDone :TNotifyEvent);
  procedure prLClickRemSel(OnActionDone :TNotifyEvent);

  procedure prRestoreDefaults; override;


public


  constructor Create(AOwner :TComponent); reintroduce;


end;


implementation


{$R *.DFM}


uses


QUB_GUIInterface;


// *****************************************************************************
constructor TQUB_ActionSelector_Lists.Create(AOwner :TComponent);
begin
  inherited CreateNamed(AOwner, 'Lists');

  prAddAction( TQUB_Action.Create('List Segs', prLClickListSegs, nil, nil,
                                  'Add a list of all segments','AllS'));
  prAddAction( TQUB_Action.Create('List Idl', prLClickListIdl, nil, nil,
                                  'Add a list of all idealized segments','IdlS'));
  prAddAction( TQUB_Action.Create('New List', prLClickNewList, prRClickNewList, nil,
                                  'Add a new selection list','AddL'));
  prAddAction( TQUB_Action.Create('Del List', prLClickDelList, nil, nil,
                                  'Delete the current selection list', 'DelL'));
  prAddAction( TQUB_Action.Create('Add Sel', prLClickAddSel, prRClickAddSel, nil,
                                  'Add the selection to the current list', 'AddS'));
  prAddAction( TQUB_Action.Create('Rem Sel', prLClickRemSel, nil, nil,
                                  'Remove an item from the current list', 'DelS'));
  RestoreDefaults;
end;

procedure TQUB_ActionSelector_Lists.prRestoreDefaults;
var
  Vis                 :TStringList;
begin
  Vis                 := ActionByName['List Segs'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['List Idl'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['New List'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Del List'].VisLayouts;
  Vis.Clear;

  Vis                 := ActionByName['Add Sel'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Rem Sel'].VisLayouts;
  Vis.Clear;

end;

procedure TQUB_ActionSelector_Lists.prLClickListSegs(OnActionDone :TNotifyEvent);
begin
  ViewportData.AddSegmentList;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Lists.prLClickListIdl(OnActionDone :TNotifyEvent);
begin
  ViewportData.AddIdlSegmentList;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Lists.prLClickNewList(OnActionDone :TNotifyEvent);
begin
  ViewportData.AddList('');
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Lists.prRClickNewList(OnActionDone :TNotifyEvent);
begin
  with ViewportData do begin
    if CurrData <> nil then CurrData.AddList( InputBox('Add a new selection list', 'List name:', '') );
  end;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Lists.prLClickDelList(OnActionDone :TNotifyEvent);
begin
  ViewportData.DelList;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Lists.prLClickAddSel(OnActionDone :TNotifyEvent);
begin
  ViewportData.AddSelection;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Lists.prRClickAddSel(OnActionDone :TNotifyEvent);
begin
  with ViewportData do begin
    AddSelection;
    if (CurrData <> nil) and (CurrData.CurrList <> nil) then begin
      with CurrData.CurrList do begin
        if not ShowSelectionProperties(IndexOf(CurrData.DataSel.First, CurrData.DataSel.Last)) then begin
          DelSelection(ItemIndex);
        end;
        CurrData.RefreshLists;
      end;
    end;
  end;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Lists.prLClickRemSel(OnActionDone :TNotifyEvent);
begin
  ViewportData.DelSelection;
  if @OnActionDone <> nil then OnActionDone(nil);
end;


//****************************************************************************//
end.
