unit QUB_ActionSelector_Modeling;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls, Menus, Buttons,

QUB_Global,
QUB_Action, QUB_ActionSelector,
QUB_Task_Simulate,
QUB_Task_Simulate_Properties,
QUB_QVO_QVD,
QUB_Task,
QUB_Task_IdealizeMT,
QUB_Task_Idealize_Properties,
QUB_Task_AMP,
QUB_Task_AMP_Properties,
QUB_Task_IdlBase,
QUB_Task_IdlBase_Properties,
QUB_Task_IdlFit,
QUB_Task_IdlFit_Properties,
QUB_Task_MIL,
QUB_Task_MIL_Properties,
QUB_Task_MSL,
QUB_Task_MSL_Properties,
QUB_Task_MPL,
QUB_Task_MPL_Properties,
QUB_Task_MIP,
QUB_Task_MIP_Properties,
QUB_Task_Drift,
QUB_Task_Drift_Properties,
QUB_Task_Mac,
QUB_Task_Mac_Properties,
QUB_Task_MacLiveFit,
QUB_Task_Exp,
QUB_Task_Exp_Properties,
QUB_Task_MSearch,
QUB_Task_MSearch_Properties,
QUB_Task_MMerge_Properties,
QUB_Task_Stat,
QUB_Task_Stat_Properties,
QUB_Task_MdlRpt,
QUB_Task_MdlRpt_Properties,
QUB_MultiFit,
QUB_Task_Survey,
QUB_Task_Survey_Properties,
QUB_Node;


type


TQUB_ActionSelector_Modeling = class(TQUB_ActionSelector)
protected


  prErrorMsg                     :String;
  prWikiTopic                    :String;

  prTaskMacLiveFit               :TQUB_Task_MacLiveFit;


  function prCanModelData(OpName :String; CanIeqFv :Boolean = False): Boolean;

  procedure prLClickIdl(OnActionDone :TNotifyEvent);
  procedure prRClickIdl(OnActionDone :TNotifyEvent);
  function prIdl(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prLClickAmp(OnActionDone :TNotifyEvent);
  procedure prRClickAmp(OnActionDone :TNotifyEvent);
  function prAMP(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prLClickIdlBase(OnActionDone :TNotifyEvent);
  procedure prRClickIdlBase(OnActionDone :TNotifyEvent);
  function prIdlBase(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prLClickIdlFit(OnActionDone :TNotifyEvent);
  procedure prRClickIdlFit(OnActionDone :TNotifyEvent);
  function prIdlFit(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  function prCanMIL :Boolean;
  procedure prLClickMIL(OnActionDone :TNotifyEvent);
  procedure prRClickMIL(OnActionDone :TNotifyEvent);
  function prMIL(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;  // PPyObject or nil

  function prCanMSL :Boolean;
  procedure prLClickMSL(OnActionDone :TNotifyEvent);
  procedure prRClickMSL(OnActionDone :TNotifyEvent);
  function prMSL(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;  // PPyObject or nil

  procedure prLClickMPL(OnActionDone :TNotifyEvent);
  procedure prRClickMPL(OnActionDone :TNotifyEvent);
  function prMPL(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  function prCanMIP :Boolean;
  procedure prLClickMIP(OnActionDone :TNotifyEvent);
  procedure prRClickMIP(OnActionDone :TNotifyEvent);
  function prMIP(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  function prCanMac :Boolean;
  function prMacStimAssignedChannel: Boolean;
  function prMacUserInsistsNoStimChannel: Boolean;
  procedure prLClickMac(OnActionDone :TNotifyEvent);
  procedure prRClickMac(OnActionDone :TNotifyEvent);
  function prMac(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prMacLiveFit(OnActionDone :TNotifyEvent);
  procedure prMacLiveFitDone(Sender :TObject);

  function prCanExp :Boolean;
  procedure prLClickExp(OnActionDone :TNotifyEvent);
  procedure prRClickExp(OnActionDone :TNotifyEvent);
  function prExp(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  function prCanDrift :Boolean;
  procedure prLClickDrift(OnActionDone :TNotifyEvent);
  procedure prRClickDrift(OnActionDone :TNotifyEvent);
  function prDrift(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prLClickMSearch(OnActionDone :TNotifyEvent);
  procedure prRClickMSearch(OnActionDone :TNotifyEvent);
  function prMSearch(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prRClickMMerge(OnActionDone :TNotifyEvent);

  procedure prLClickStat(OnActionDone :TNotifyEvent);
  procedure prRClickStat(OnActionDone :TNotifyEvent);
  function prStat(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prLClickMdlRpt(OnActionDone :TNotifyEvent);
  procedure prRClickMdlRpt(OnActionDone :TNotifyEvent);
  function prMdlRpt(OnActionDone :TNotifyEvent; Properties :Boolean) :Boolean;

  procedure prLClickIV(OnActionDone :TNotifyEvent);
  procedure prClickMultiFit(OnActionDone :TNotifyEvent);
  procedure prClickSurvey(OnActionDone :TNotifyEvent);

  procedure prRestoreDefaults; override;


public


  TaskIdealizePropertiesForm      :TQUB_Task_Idealize_PropertiesForm;
  TaskAmpPropertiesForm           :TQUB_Task_Amp_PropertiesForm;
  TaskIdlBasePropertiesForm       :TQUB_Task_IdlBase_PropertiesForm;
  TaskIdlFitPropertiesForm        :TQUB_Task_IdlFit_PropertiesForm;
  TaskMILPropertiesForm           :TQUB_Task_MIL_PropertiesForm;
  TaskMSLPropertiesForm           :TQUB_Task_MSL_PropertiesForm;
  TaskMPLPropertiesForm           :TQUB_Task_MPL_PropertiesForm;
  TaskMIPPropertiesForm           :TQUB_Task_MIP_PropertiesForm;
  TaskMacPropertiesForm           :TQUB_Task_Mac_PropertiesForm;
  TaskExpPropertiesForm           :TQUB_Task_Exp_PropertiesForm;
  TaskDriftPropertiesForm         :TQUB_Task_Drift_PropertiesForm;
  TaskMSearchPropertiesForm       :TQUB_Task_MSearch_PropertiesForm;
  TaskStatPropertiesForm          :TQUB_Task_Stat_PropertiesForm;
  TaskMMergePropertiesForm        :TQUB_Task_MMerge_PropertiesForm;
  TaskMdlRptPropertiesForm        :TQUB_Task_MdlRpt_PropertiesForm;
  MultiFitForm                    :TQUB_MultiFitForm;
  TaskSurveyPropertiesForm        :TQUB_Task_Survey_PropertiesForm;


  constructor Create(AOwner :TComponent); reintroduce;


  procedure StopMacLiveFit;

  function MacFit :Boolean;


end;


implementation


{$R *.DFM}


uses


Math,
QUB_Main,
QUB_GUIInterface,
QUB_QFS_File, QUB_QFS_QDF_Global,
QUB_QVO_QVM,
QUB_Python_Data,
QUB_ExpCond_Parameters,
QUB_Viewport_Data, QUB_Viewport_Model, QUB_Viewport_Histogram;


//****************************************************************************//
constructor TQUB_ActionSelector_Modeling.Create(AOwner :TComponent);
begin
  inherited CreateNamed(AOwner, 'Modeling');

  //----- IDL
  TaskIdealizePropertiesForm := TQUB_Task_Idealize_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Idealize', prLClickIdl, prRClickIdl, TaskIdealizePropertiesForm.SetGetProperties,
                                  'Idealize data with SKM or Half-Amp','Idl',clGreen));

  //----- AMPS
  TaskAmpPropertiesForm := TQUB_Task_Amp_PropertiesForm.Create(Application);
  prAddAction(TQUB_Action.Create('Amps', prLClickAMP, prRClickAMP, TaskAmpPropertiesForm.SetGetProperties,
                                  'Estimate conductance levels with Baum-Welch','AMP',clMaroon));

  //----- IDL/BASE
  TaskIdlBasePropertiesForm := TQUB_Task_IdlBase_PropertiesForm.Create(Application);
  prAddAction(TQUB_Action.Create('Idl/Base', prLClickIdlBase, prRClickIdlBase, TaskIdlBasePropertiesForm.SetGetProperties,
                                  'Idealize with baseline correction','IB',clFuchsia));

  //----- MIL
  TaskMilPropertiesForm := TQUB_Task_Mil_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('MIL Rates', prLClickMIL, prRClickMIL, TaskMilPropertiesForm.SetGetProperties,
                                  'Maximum interval likelihood rate estimation','MIL',clPurple));

  //----- MSL
  TaskMSLPropertiesForm := TQUB_Task_MSL_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('MSL Rates', prLClickMSL, prRClickMSL, TaskMSLPropertiesForm.SetGetProperties,
                                  'Maximum sub-interval likelihood rate estimation','MSL',clPurple));

  //----- MPL
  TaskMPLPropertiesForm := TQUB_Task_MPL_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('MPL Rates', prLClickMPL, prRClickMPL, TaskMPLPropertiesForm.SetGetProperties,
                                  'Maximum point likelihood rate estimation', 'MPL', clTeal));

  //----- MIP
  TaskMIPPropertiesForm := TQUB_Task_MIP_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('MIP Rates', prLClickMIP, prRClickMIP, TaskMIPPropertiesForm.SetGetProperties,
                                  'Maximum idealized point likelihood rate estimation','MIP',clBlue));

  //-----Mac
  TaskMacPropertiesForm := TQUB_Task_Mac_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Mac Rates', prLClickMac, prRClickMac, TaskMacPropertiesForm.SetGetProperties,
                                  'Mechanistic fit of macroscopic currents','Mac',clRed));

  prAddAction( TQUB_Action.Create('Mac LiveFit', prMacLiveFit, prMacLiveFit, nil,
                                  'Mac fit curves, updated when model changes','MacLF',clRed));

  //-----Exp fitting
  TaskExpPropertiesForm := TQUB_Task_Exp_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Exp fitting', prLClickExp, prRClickExp, TaskExpPropertiesForm.SetGetProperties,
                                  'Exponential fitting of macroscopic currents','Exp',clRed));
  QUB_GUIInterface.TaskExpPropertiesForm := TaskExpPropertiesForm;

  //-----Staircase (Kinesin)
  TaskDriftPropertiesForm := TQUB_Task_Drift_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Staircase', prLClickDrift, prRClickDrift, TaskDriftPropertiesForm.SetGetProperties,
                                  'Staircase processes rate estimation with MIP', 'Stair'));

  //----- Model Search
  TaskMSearchPropertiesForm := TQUB_Task_MSearch_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Model Srch', prLClickMSearch, prRClickMSearch, TaskMSearchPropertiesForm.SetGetProperties,
                                  'Model Search with MIL','msrch',clPurple));

  //----- Model Merge
  TaskMMergePropertiesForm := TQUB_Task_MMerge_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Model Merge', prRClickMMerge, prRClickMMerge, nil,
                                  '','mmerge'));

  //----- Stat
  TaskStatPropertiesForm := TQUB_Task_Stat_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Stat', prLClickStat, prRClickStat, TaskStatPropertiesForm.SetGetProperties,
                                  '','Stat'));

  //----- Model Report
  TaskMdlRptPropertiesForm := TQUB_Task_MdlRpt_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('MdlReport', prLClickMdlRpt, prRClickMdlRpt, TaskMdlRptPropertiesForm.SetGetProperties,
                                  '', 'MdlRpt'));

  //----- IV Plot
  prAddAction( TQUB_Action.Create('IV Plot', prLClickIV, prLClickIV, nil,
                                  '', 'IV'));

  MultiFitForm := TQUB_MultiFitForm.Create(Application);
  prAddAction( TQUB_Action.Create('MultiFit', prClickMultiFit, prClickMultiFit, nil,
                                  '', 'MFit'));

  TaskSurveyPropertiesForm := TQUB_Task_Survey_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Survey', prClickSurvey, prClickSurvey, TaskSurveyPropertiesForm.SetGetProperties,
                                  'Scan over a range of a model parameter', 'Surv'));


  //----- NEW - chris' attempt to help Arthur ?
  {
  06-2004 todo
  TaskIdlFitPropertiesForm := TQUB_Task_IdlFit_PropertiesForm.Create(Application);
  Action              := TQUB_Action.Create('Idl/Fit', prLClickIdlFit, prRClickIdlFit,
                                            TaskIdlFitPropertiesForm.SetGetProperties);
  Action.Hint         := 'Idealize with baseline curve-fit';
  Action.ShortName    := 'IdlF';
  prAddAction(Action);
  }


  RestoreDefaults;
end;

function TQUB_ActionSelector_Modeling.MacFit :Boolean;
begin
  Result := prMac(nil{OnActionDone}, False);
end;

procedure TQUB_ActionSelector_Modeling.prRestoreDefaults;
var
  Vis                 :TStringList;
begin
  Vis                 := ActionByName['Idealize'].VisLayouts;
  Vis.Clear;
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Amps'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Idl/Base'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
  Vis.Add('Modeling');
  Vis.Add('Results');

  // 06-2004 todo - make this work
  //Vis                 := ActionByName['Idl/Fit'].VisLayouts;
  //Vis.Clear;
  //Vis.Add('Preprocessing');
  //Vis.Add('Modeling');

  Vis                 := ActionByName['MIL Rates'].VisLayouts;
  Vis.Clear;
  Vis.Add('Simulation');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['MSL Rates'].VisLayouts;
  Vis.Clear;
  Vis.Add('Simulation');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['MPL Rates'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Simulation');
  Vis.Add('Results');

  Vis                 := ActionByName['MIP Rates'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Mac Rates'].VisLayouts;
  Vis.Clear;
  Vis.Add('Simulation');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Mac LiveFit'].VisLayouts;
  Vis.Clear;
  Vis.Add('Macroscopic');
  Vis.Add('Results');

  Vis                 := ActionByName['Exp Fitting'].VisLayouts;
  Vis.Clear;
  Vis.Add('Simulation');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Staircase'].VisLayouts;
  Vis.Clear;

  Vis                 := ActionByName['Model Srch'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Model Merge'].VisLayouts;
  Vis.Clear;
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['Stat'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
  Vis.Add('Results');

  Vis                 := ActionByName['MdlReport'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis                 := ActionByName['IV Plot'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Results');

  Vis                 := ActionByName['MultiFit'].VisLayouts;
  Vis.Clear;
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');

  Vis                 := ActionByName['Survey'].VisLayouts;
  Vis.Clear;
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');
end;

function TQUB_ActionSelector_Modeling.prCanModelData(OpName :String; CanIeqFv :Boolean = False): Boolean;
begin
  Result := ViewportData.CurrData <> nil;
  if Result then Result := not ViewportData.CurrData.LockedReadWrite;
  if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
  //TO DO: make a function that returns true if IDL-type. LM
  if Result then Result := (ViewportData.CurrData.DataObject.FileType <> QFS_QDF_TYPE_DWT) and (ViewportData.CurrData.DataObject.FileType <> QFS_QDF_TYPE_EVL) and (ViewportData.CurrData.DataObject.FileType <> QFS_QDF_TYPE_PRT) and (ViewportData.CurrData.DataObject.FileType <> QFS_QDF_TYPE_QSF) and (ViewportData.CurrData.DataObject.FileType <> QFS_QDF_TYPE_TAC);
  if not Result then begin
    prErrorMsg  := 'Cannot ' + OpName + '. The data might be in use or empty or of the wrong type.';
    prWikiTopic := 'Error:No_Data';
  end;
  if not Result then Exit;

  if Result then Result := ViewportData.CurrData.CurrModel <> nil;
  if Result then Result := not ViewportData.CurrData.CurrModel.LockedReadWrite;
  if Result then Result := not TQUB_QVO_QVM(ViewportData.CurrData.CurrModel).ModelObject.States.StateCount <> 0;
  if not Result then begin
    prErrorMsg  := 'Cannot '+ OpName + '. The model might be in use or empty.';
    prWikiTopic := 'Error:No_Model';
  end;
  if not Result then Exit;

  if (not CanIeqFv) and (0<>TQUB_QVO_QVM(ViewportData.CurrData.CurrModel).ModelObject.Properties['IeqFv'].Data[0]) then begin
    Result := False;
    prErrorMsg := 'Cannot ' + OpName + '. The model uses units of pS, but ' + opName + ' works in pA.';
    prWikiTopic := 'Error:Cannot_model_pS';
    Exit;
  end;
end;

procedure TQUB_ActionSelector_Modeling.prLClickIdl(OnActionDone :TNotifyEvent);
begin
  prIdl(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickIdl(OnActionDone :TNotifyEvent);
begin
  prIdl(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prIdl(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data                     :TQUB_QVO_QVD;
  Model                    :TQUB_QVO_QVM;
  TaskIdealize             :TQUB_Task_IdealizeMT;
begin
  
  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_IdealizeMT)) then begin
    Result := True;
    MessageHelpDlg('Only one Idealization task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Data := ViewportData.CurrData;
  if Data <> nil then begin
    Model := TQUB_QVO_QVM(Data.CurrModel);
  end else begin
    Model := nil;
  end;
  if Model <> nil then begin
    TaskIdealizePropertiesForm.ClassCount := Model.ModelObject.ClassCount;
  end;

  Result := not prCanModelData('idealize');

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := ''; // reset
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  //DataSource
  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskIdealizePropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskIdealizePropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  TaskIdealize := TQUB_Task_IdealizeMT.Create(True{Suspended});
  TaskIdealize.OnTerminate := OnActionDone;

  TaskIdealizePropertiesForm.OnGetProperties := TaskIdealize.OnGetProperties;
  TaskIdealizePropertiesForm.OnPutProperties := TaskIdealize.OnPutProperties;

  TaskIdealizePropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskIdealizePropertiesForm.ShowModal;
  end else begin
    TaskIdealizePropertiesForm.FormShow(TaskIdealizePropertiesForm);
    TaskIdealizePropertiesForm.ModalResult := mrOK;
    TaskIdealizePropertiesForm.btnOKClick(TaskIdealizePropertiesForm);
  end;

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and (TaskIdealizePropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskIdealizePropertiesForm.DataSource;
  end;

  if TaskIdealizePropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskIdealize.Terminate;
  end else if TaskIdealizePropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskIdealize.Terminate;
    Application.ProcessMessages;
    if Data <> nil then begin
      Data.DataObject.Idealization[TaskIdealizePropertiesForm.Properties['ChannelIndex'].Data[0]].Clear;
    end;
  end else begin
  //OK
    //TaskIdealize.Servers           := prServers;
    TaskIdealize.OnChange          := TaskView.OnTaskChange;
    TaskIdealize.Data              := Data;
    TaskIdealize.Model             := Model;
    TaskIdealize.FillDataCallback  := DataNoder.FillData;
  end;
  TaskIdealize.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prLClickAMP(OnActionDone :TNotifyEvent);
begin
  prAMP(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickAMP(OnActionDone :TNotifyEvent);
begin
  prAMP(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prAMP(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  TaskAmp       :TQUB_Task_Amp;
begin
  
  Data := ViewportData.CurrData;
  
  if Data <> nil then begin
    Model := TQUB_QVO_QVM(Data.CurrModel);
  end else begin
    Model := nil;
  end;

  Result := not prCanModelData('estimate amps');

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskAmp := TQUB_Task_Amp.Create(True{Suspended});
  TaskAmp.OnTerminate := OnActionDone;

  TaskAmpPropertiesForm.OnGetProperties := TaskAmp.OnGetProperties;
  TaskAmpPropertiesForm.OnPutProperties := TaskAmp.OnPutProperties;

  TaskAmpPropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskAmpPropertiesForm.ShowModal;
  end else begin
    TaskAmpPropertiesForm.FormShow(TaskAmpPropertiesForm);
    TaskAmpPropertiesForm.ModalResult := mrOK;
    TaskAmpPropertiesForm.btnOKClick(TaskAmpPropertiesForm);
  end;

  if TaskAmpPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskAmp.Terminate;
  end else begin
  //OK
    //TaskAmp.Servers           := prServers;
    TaskAmp.OnChange          := TaskView.OnTaskChange;
    TaskAmp.Data              := Data;
    TaskAmp.Model             := Model;
    TaskAmp.FillDataCallback  := DataNoder.FillData;
  end;
  TaskAmp.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prLClickIdlFit(OnActionDone :TNotifyEvent);
begin
  prIdlFit(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickIdlFit(OnActionDone :TNotifyEvent);
begin
  prIdlFit(OnActionDone, True);
end;

// according to Chris, this doesn't do anything useful yet, so comment it out
function TQUB_ActionSelector_Modeling.prIdlFit(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  Task          :TQUB_Task_IdlFit;
begin

  Data := ViewportData.CurrData;
  Model := ViewportModel.CurrModel;

  Result := not prCanModelData('estimate amps');

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  Task := TQUB_Task_IdlFit.Create(True{Suspended});
  Task.OnTerminate := OnActionDone;

  TaskIdlFitPropertiesForm.OnGetProperties := Task.OnGetProperties;
  TaskIdlFitPropertiesForm.OnPutProperties := Task.OnPutProperties;

  TaskIdlFitPropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskIdlFitPropertiesForm.ShowModal;
  end else begin
    TaskIdlFitPropertiesForm.FormShow(TaskIdlFitPropertiesForm);
    TaskIdlFitPropertiesForm.ModalResult := mrOK;
    TaskIdlFitPropertiesForm.btnOKClick(TaskIdlFitPropertiesForm);
  end;

  if TaskIdlFitPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    Task.Terminate;
  end else begin
  //OK
    //Task.Servers           := prServers;
    Task.OnChange          := TaskView.OnTaskChange;
    Task.Data              := Data;
    Task.Model             := Model;
    Task.FillDataCallback  := DataNoder.FillData;
  end;
  Task.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prLClickIdlBase(OnActionDone :TNotifyEvent);
begin
  prIdlBase(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickIdlBase(OnActionDone :TNotifyEvent);
begin
  prIdlBase(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prIdlBase(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  TaskIdlBase   :TQUB_Task_IdlBase;
begin
  
  Data := ViewportData.CurrData;
  if Data <> nil then begin
    Model := TQUB_QVO_QVM(Data.CurrModel);
  end else begin
    Model := nil;
  end;

  Result := not prCanModelData('idealize');

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  //DataSource
  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskIdlBasePropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskIdlBasePropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  TaskIdlBase := TQUB_Task_IdlBase.Create(True{Suspended});
  TaskIdlBase.OnTerminate := OnActionDone;

  TaskIdlBasePropertiesForm.OnGetProperties := TaskIdlBase.OnGetProperties;
  TaskIdlBasePropertiesForm.OnPutProperties := TaskIdlBase.OnPutProperties;

  TaskIdlBasePropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskIdlBasePropertiesForm.ShowModal;
  end else begin
    TaskIdlBasePropertiesForm.FormShow(TaskIdlBasePropertiesForm);
    TaskIdlBasePropertiesForm.ModalResult := mrOK;
    TaskIdlBasePropertiesForm.btnOKClick(TaskIdlBasePropertiesForm);
  end;

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and (TaskIdlBasePropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskIdlBasePropertiesForm.DataSource;
  end;

  if TaskIdlBasePropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskIdlBase.Terminate;
  end else if TaskIdlBasePropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskIdlBase.Terminate;
    Application.ProcessMessages;
    if Data <> nil then begin
      Data.DataObject.Idealization[TaskIdlBasePropertiesForm.Properties['ChannelIndex'].Data[0]].Clear;
    end;
  end else begin
  //OK
    //TaskIdlBase.Servers           := prServers;
    TaskIdlBase.OnChange          := TaskView.OnTaskChange;
    TaskIdlBase.Data              := Data;
    TaskIdlBase.Model             := Model;
    TaskIdlBase.GlobalFillData    := DataNoder.FillData;
  end;
  TaskIdlBase.Resume;
end;

function TQUB_ActionSelector_Modeling.prCanMIL: Boolean;
begin

  Result := (ViewportData.CurrData <> nil) or ({TaskMILPropertiesForm.}QUB_MainForm.DataSource = DATA_SOURCE_FILELIST);
  if Result then begin
    if {prTaskMilPropertiesForm.}QUB_MainForm.DataSource <> DATA_SOURCE_FILELIST then begin
      if Result then Result := not ViewportData.CurrData.LockedReadWrite;
      if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
    end;
  end;
  if not Result then begin
    prErrorMsg := 'Cannot run MIL. The data might be in use or empty.';
    prWikiTopic := 'Error:No_Data';
  end;
  if not Result then Exit;

  if Result then Result := ViewportModel.CurrModel <> nil;
  if Result then Result := not ViewportModel.CurrModel.LockedReadWrite;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.States.StateCount <> 0;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.Rates.RateCount <> 0;
  if not Result then begin
    prErrorMsg := 'Cannot run MIL. The model might be in use or empty.';
    prWikiTopic := 'Error:No_Model';
  end;
  if not Result then Exit;
end;

procedure TQUB_ActionSelector_Modeling.prLClickMIL(OnActionDone :TNotifyEvent);
begin
  prMIL(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickMIL(OnActionDone :TNotifyEvent);
begin
  prMIL(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prMIL(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data    :TQUB_QVO_QVD;
  Model   :TQUB_QVO_QVM;
  //i       :Integer;
  //item    :PPyObject;
  TaskMil :TQUB_Task_MIL;
begin
  
  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_Mil)) then begin
    Result := True;
    MessageHelpDlg('Only one Mil task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Data  := ViewportData.CurrData;
  Model := ViewportModel.CurrModel;

  Result := not prCanMil;

  //if Data = nil then can only use FileList option !!!

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskMil := TQUB_Task_Mil.Create(True{Suspended});
  TaskMil.OnTerminate  := OnActionDone;

  TaskMilPropertiesForm.OnGetProperties := TaskMil.OnGetProperties;
  TaskMilPropertiesForm.OnPutProperties := TaskMil.OnPutProperties;

  TaskMilPropertiesForm.btnOK.Enabled    := not Result;

  if (Properties) or (Result) then begin
    TaskMilPropertiesForm.ShowModal;
  end else begin
    TaskMilPropertiesForm.FormShow(TaskMilPropertiesForm);
    TaskMilPropertiesForm.ModalResult := mrOK;
    TaskMilPropertiesForm.btnOKClick(TaskMilPropertiesForm);
  end;

  if TaskMilPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskMil.Terminate;
  end else if TaskMilPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskMil.Terminate;
  end else begin
  //OK
    //TaskMil.Servers           := prServers;
    Data.ActiveChannel        := TaskMilPropertiesForm.Properties['ChannelIndex'].Data[0];
    TaskMil.OnChange          := TaskView.OnTaskChange;
    TaskMil.Data              := Data;
    TaskMil.Model             := Model;
  end;
  TaskMil.Resume;
end;

function TQUB_ActionSelector_Modeling.prCanMSL: Boolean;
begin
  Result := prCanMIL;
end;

procedure TQUB_ActionSelector_Modeling.prLClickMSL(OnActionDone :TNotifyEvent);
begin
  prMSL(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickMSL(OnActionDone :TNotifyEvent);
begin
  prMSL(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prMSL(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data    :TQUB_QVO_QVD;
  Model   :TQUB_QVO_QVM;
  //i       :Integer;
  //item    :PPyObject;
  TaskMSL :TQUB_Task_MSL;
begin

  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_MSL)) then begin
    Result := True;
    MessageHelpDlg('Only one MSL task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Data  := ViewportData.CurrData;
  Model := ViewportModel.CurrModel;

  Result := not prCanmsl;

  //if Data = nil then can only use FileList option !!!

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskMSL := TQUB_Task_MSL.Create(True{Suspended});
  TaskMSL.OnTerminate  := OnActionDone;

  TaskMSLPropertiesForm.OnGetProperties := TaskMSL.OnGetProperties;
  TaskMSLPropertiesForm.OnPutProperties := TaskMSL.OnPutProperties;

  TaskMSLPropertiesForm.btnOK.Enabled    := not Result;

  if (Properties) or (Result) then begin
    TaskMSLPropertiesForm.ShowModal;
  end else begin
    TaskMSLPropertiesForm.FormShow(TaskMSLPropertiesForm);
    TaskMSLPropertiesForm.ModalResult := mrOK;
    TaskMSLPropertiesForm.btnOKClick(TaskMSLPropertiesForm);
  end;

  if TaskMSLPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskMSL.Terminate;
  end else if TaskMSLPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskMSL.Terminate;
  end else begin
  //OK
    //TaskMSL.Servers           := prServers;
    Data.ActiveChannel        := TaskMSLPropertiesForm.Properties['ChannelIndex'].Data[0];
    TaskMSL.OnChange          := TaskView.OnTaskChange;
    TaskMSL.Data              := Data;
    TaskMSL.Model             := Model;
  end;
  TaskMSL.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prLClickMPL(OnActionDone :TNotifyEvent);
begin
  prMPL(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickMPL(OnActionDone :TNotifyEvent);
begin
  prMPL(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prMPL(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  TaskMPL       :TQUB_Task_MPL;
begin
  
  Data := ViewportData.CurrData;
  if Data <> nil then begin
    Model := TQUB_QVO_QVM(Data.CurrModel);
  end else begin
    Model := nil;
  end;

  Result := not prCanModelData('run MPL');

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskMPL := TQUB_Task_MPL.Create(True{Suspended});
  TaskMPL.OnTerminate := OnActionDone;

  TaskMPLPropertiesForm.OnGetProperties := TaskMPL.OnGetProperties;
  TaskMPLPropertiesForm.OnPutProperties := TaskMPL.OnPutProperties;

  TaskMPLPropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskMPLPropertiesForm.ShowModal;
  end else begin
    TaskMPLPropertiesForm.FormShow(TaskMPLPropertiesForm);
    TaskMPLPropertiesForm.ModalResult := mrOK;
    TaskMPLPropertiesForm.btnOKClick(TaskMPLPropertiesForm);
  end;

  if TaskMPLPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskMPL.Terminate;
  end else begin
  //OK
    //TaskMPL.Servers           := prServers;
    TaskMPL.OnChange          := TaskView.OnTaskChange;
    TaskMPL.Data              := Data;
    TaskMPL.Model             := Model;
    TaskMPL.FillDataCallback  := DataNoder.FillData;
  end;
  TaskMPL.Resume;
end;

function TQUB_ActionSelector_Modeling.prCanMIP: Boolean;
begin
  
  Result := (ViewportData.CurrData <> nil) or ({TaskMIPPropertiesForm.}QUB_MainForm.DataSource = DATA_SOURCE_FILELIST);
  if Result then begin
    if {prTaskMIPPropertiesForm.}QUB_MainForm.DataSource <> DATA_SOURCE_FILELIST then begin
      if Result then Result := not ViewportData.CurrData.LockedReadWrite;
      if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
    end;
  end;
  if not Result then begin
    prErrorMsg := 'Cannot run MIP. The data might be in use or empty.';
    prWikiTopic := 'Error:No_Data';
    Exit;
  end;

  if Result then Result := ViewportModel.CurrModel <> nil;
  if Result then Result := not ViewportModel.CurrModel.LockedReadWrite;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.States.StateCount <> 0;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.Rates.RateCount <> 0;
  if not Result then begin
    prErrorMsg := 'Cannot run MIP. The model might be in use or empty.';
    prWikiTopic := 'Error:No_Model';
    Exit;
  end;

  if Result then Result := ViewportModel.CurrModel.ModelObject.ChannelCount = 1;
  if not Result then begin
    prErrorMsg := 'Cannot run MIP. The model has Channel Count > 1.';
    prWikiTopic := 'Error:ModelChannelCount';
    Exit;
  end;
  if not Result then Exit;

  if (0<>ViewportModel.CurrModel.ModelObject.Properties['IeqFv'].Data[0]) then begin
    Result := False;
    prErrorMsg := 'Cannot run MIP. The model uses units of pS, but MIP works in pA.';
    prWikiTopic := 'Error:Cannot_model_pS';
    Exit;
  end;
end;

procedure TQUB_ActionSelector_Modeling.prLClickMIP(OnActionDone :TNotifyEvent);
begin
  prMIP(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickMIP(OnActionDone :TNotifyEvent);
begin
  prMIP(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prMIP(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  TaskMIP       :TQUB_Task_MIP;
begin

  Data          := ViewportData.CurrData;
  Model         := ViewportModel.CurrModel;

  Result := not prCanMIP;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskMIP := TQUB_Task_MIP.Create(True{Suspended});
  TaskMIP.OnTerminate    := OnActionDone;

  TaskMIPPropertiesForm.OnGetProperties := TaskMIP.OnGetProperties;
  TaskMIPPropertiesForm.OnPutProperties := TaskMIP.OnPutProperties;

  TaskMIPPropertiesForm.btnOK.Enabled    := not Result;

  if (Properties) or (Result) then begin
    TaskMIPPropertiesForm.ShowModal;
  end else begin
    TaskMIPPropertiesForm.FormShow(TaskMIPPropertiesForm);
    TaskMIPPropertiesForm.ModalResult := mrOK;
    TaskMIPPropertiesForm.btnOKClick(TaskMIPPropertiesForm);
  end;

  if TaskMIPPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskMIP.Terminate;
  end else if TaskMIPPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskMIP.Terminate;
  end else begin
  //OK
    //TaskMIP.Servers  := prServers;
    TaskMIP.OnChange := TaskView.OnTaskChange;
    TaskMIP.Data     := Data;
    TaskMIP.Model    := Model;
  end;
  TaskMIP.Resume;
end;

//Mac
function TQUB_ActionSelector_Modeling.prCanMac: Boolean;
begin

  Result := (ViewportData.CurrData <> nil) or ({prTaskMacPropertiesForm.}QUB_MainForm.DataSource = DATA_SOURCE_FILELIST);
  if Result then begin
    if {prTaskMacPropertiesForm.}QUB_MainForm.DataSource <> DATA_SOURCE_FILELIST then begin
      if Result then Result := not ViewportData.CurrData.LockedReadWrite;
      if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
    end;
  end;
  if not Result then begin
    prErrorMsg := 'Cannot run Mac. The data might be in use or empty or of the wrong type.';
    prWikiTopic := 'Error:No_Data';
  end;
  if not Result then Exit;

  if Result then Result := ViewportModel.CurrModel <> nil;
  if Result then Result := not ViewportModel.CurrModel.LockedReadWrite;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.States.StateCount <> 0;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.Rates.RateCount <> 0;
  if not Result then begin
    prErrorMsg := 'Cannot run Mac. The model might be in use or empty.';
    prWikiTopic := 'Error:No_Model';
  end;
  if not Result then Exit;

  // if model has stimulus p or q not assigned to data channel,
  // kick up error message and go to qvd properties: info
  // and maybe select that entry in the table

  // pro: almost certainly what they forgot to do
  // con: debilitatingly obnoxious if the stimulus is an actual constant
  //      e.g. mil-style global fitting
  

end;

function TQUB_ActionSelector_Modeling.prMacStimAssignedChannel: Boolean;
var
  j, k       :Integer;
  Model      :TQUB_QVO_QVM;
  Stimuli    :TStringList;
  Data       :TQUB_QVO_QVD;

  function AssignedInData(Data :TQUB_QVO_QVD): Boolean;
  var
    ExpCond    :TQUB_ExpCond_PParameter;
    i          :Integer;
  begin
    for i := 0 to Stimuli.Count - 1 do begin
      ExpCond  := Data.ExpCond.ParameterByName[Stimuli[i]];
      if (ExpCond = nil) or (ExpCond.Channel < 0) then begin
        Result := False;
        Exit;
      end;
    end;
    Result := True;
  end;

begin
  Stimuli    := TStringList.Create;
  Stimuli.Sorted := True;
  Stimuli.Duplicates := dupIgnore;

  try
    Model       := ViewportModel.CurrModel;
    for j := 0 to Model.ModelObject.Rates.RateCount - 1 do begin
      for k := 1 to 2 do begin
        if Model.ModelObject.Rates[j].P[k] then Stimuli.Add(Model.ModelObject.Rates[j].PName[k]);
        if Model.ModelObject.Rates[j].Q[k] then Stimuli.Add(Model.ModelObject.Rates[j].QName[k]);
      end;
    end;

    if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
      for j := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
        Data := TQUB_QVO_QVD(ViewportData.ListObjects[j]);
        if Data.Selected and not AssignedInData(Data) then begin
          Result := False;
          Exit;
        end;
      end;
    end else begin
      if not AssignedInData(ViewportData.CurrData) then begin
        Result := False;
        Exit;
      end;
    end;

  finally
    Stimuli.Free;
  end;

  Result := True;
end;

function TQUB_ActionSelector_Modeling.prMacUserInsistsNoStimChannel: Boolean;
begin
  Result := mrYes = MessageHelpDlg('A stimulus (P or Q) in the model is not assigned to an A/D channel. Continue with Mac anyway?',
                               mtWarning, [mbYes, mbNo], 0, 'Warning:No_Stimulus');
end;

procedure TQUB_ActionSelector_Modeling.prLClickMac(OnActionDone :TNotifyEvent);
begin
  prMac(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickMac(OnActionDone :TNotifyEvent);
begin
  prMac(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prMac(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  TaskMac       :TQUB_Task_Mac;
begin

  Data          := ViewportData.CurrData;
  Model         := ViewportModel.CurrModel;

  if Data <> nil then begin
    TaskMacPropertiesForm.lstFitData.Items.Assign(Data.SelectionLists.Items);
    TaskMacPropertiesForm.lstExcData.Items.Assign(Data.SelectionLists.Items);
  end else begin
    TaskMacPropertiesForm.lstFitData.Items.Clear;
    TaskMacPropertiesForm.lstExcData.Items.Clear;
  end;

  if Model <> nil then begin
    TaskMacPropertiesForm.OptVRev := Model.ModelObject.VRev;
  end;

  StopMacLiveFit;
  while prTaskMacLiveFit <> nil do Application.ProcessMessages;

  Result := not prCanMac;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskMac := TQUB_Task_Mac.Create(True{Suspended});
  TaskMac.OnTerminate    := OnActionDone;

  TaskMacPropertiesForm.OnGetProperties := TaskMac.OnGetProperties;
  TaskMacPropertiesForm.OnPutProperties := TaskMac.OnPutProperties;

  TaskMacPropertiesForm.btnOK.Enabled := not Result;

  Properties := Properties or not prMacStimAssignedChannel;

  repeat

    if (Properties) or (Result) then begin
      TaskMacPropertiesForm.ShowModal;
    end else begin
      TaskMacPropertiesForm.FormShow(TaskMacPropertiesForm);
      TaskMacPropertiesForm.ModalResult := mrOK;
      TaskMacPropertiesForm.btnOKClick(TaskMacPropertiesForm);
    end;

  until (TaskMacPropertiesForm.ModalResult <> mrOK)
  or    prMacStimAssignedChannel
  or    prMacUserInsistsNoStimChannel;

  if TaskMacPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskMac.Terminate;
  end else begin
  //OK
    //TaskMac.Servers  := prServers;
    TaskMac.OnChange := TaskView.OnTaskChange;
    TaskMac.Data     := Data;
    TaskMac.Model    := Model;
  end;
  TaskMac.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prMacLiveFit(OnActionDone :TNotifyEvent);
begin
  if prTaskMacLiveFit = nil then begin
    prTaskMacLiveFit := TQUB_Task_MacLiveFit.Create(True);
    prTaskMacLiveFit.MacProperties := TaskMacPropertiesForm;
    prTaskMacLiveFit.OnTerminate := prMacLiveFitDone;
    prTaskMacLiveFit.Resume;
    TSpeedButton(prButtons[IndexOf('Mac LiveFit')]).GroupIndex := 1;
    TSpeedButton(prButtons[IndexOf('Mac LiveFit')]).AllowAllUp := True;
    TSpeedButton(prButtons[IndexOf('Mac LiveFit')]).Down := True;
  end else begin
    StopMacLiveFit;
  end;
end;

procedure TQUB_ActionSelector_Modeling.prMacLiveFitDone(Sender :TObject);
begin
    prTaskMacLiveFit := nil;
    TSpeedButton(prButtons[IndexOf('Mac LiveFit')]).Down := False;
  end;


procedure TQUB_ActionSelector_Modeling.StopMacLiveFit;
begin
  if prTaskMacLiveFit <> nil then prTaskMacLiveFit.Stop;
end;


//Exp
function TQUB_ActionSelector_Modeling.prCanExp: Boolean;
begin

  Result := (ViewportData.CurrData <> nil) or ({prTaskExpPropertiesForm.}QUB_MainForm.DataSource = DATA_SOURCE_FILELIST);
  if Result then begin
    if {prTaskExpPropertiesForm.}QUB_MainForm.DataSource <> DATA_SOURCE_FILELIST then begin
      if Result then Result := not ViewportData.CurrData.LockedReadWrite;
      if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
    end;
  end;
  if not Result then begin
    prErrorMsg := 'Cannot run Exp. The data might be in use or empty or of the wrong type.';
    prWikiTopic := 'Error:No_Data';
  end;
  if not Result then Exit;

end;

procedure TQUB_ActionSelector_Modeling.prLClickExp(OnActionDone :TNotifyEvent);
begin
  prExp(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickExp(OnActionDone :TNotifyEvent);
begin
  prExp(OnActionDone, True);
end;   

function TQUB_ActionSelector_Modeling.prExp(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  TaskExp       :TQUB_Task_Exp;
begin

  Data          := ViewportData.CurrData;

  if Data <> nil then begin
    TaskExpPropertiesForm.lstFitData.Items.Assign(Data.SelectionLists.Items);
    //TaskExpPropertiesForm.lstFitDataExclude.Items.Assign(Data.SelectionLists.Items);
    //TaskExpPropertiesForm.FitDataListIdx := 0;
  end else begin
    TaskExpPropertiesForm.lstFitData.Items.Clear;
    //TaskExpPropertiesForm.lstFitDataExclude.Items.Clear;
  end;

  Result := not prCanExp;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskExp := TQUB_Task_Exp.Create(True{Suspended});
  TaskExp.OnTerminate    := OnActionDone;

  TaskExpPropertiesForm.OnGetProperties := TaskExp.OnGetProperties;
  TaskExpPropertiesForm.OnPutProperties := TaskExp.OnPutProperties;

  TaskExpPropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskExpPropertiesForm.ShowModal;
  end else begin
    TaskExpPropertiesForm.FormShow(TaskExpPropertiesForm);
    TaskExpPropertiesForm.ModalResult := mrOK;
    TaskExpPropertiesForm.btnOKClick(TaskExpPropertiesForm);
  end;

  if TaskExpPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskExp.Terminate;
  end else begin
  //OK
    //TaskExp.Servers  := prServers;
    TaskExp.OnChange := TaskView.OnTaskChange;
    TaskExp.Data     := Data;
  end;
  TaskExp.Resume;
end;

//Staircase
function TQUB_ActionSelector_Modeling.prCanDrift: Boolean;
begin
  
  Result := (ViewportData.CurrData <> nil) or ({prTaskDriftPropertiesForm.}QUB_MainForm.DataSource = DATA_SOURCE_FILELIST);
  if Result then begin
    if QUB_MainForm.DataSource <> DATA_SOURCE_FILELIST then begin
      if Result then Result := not ViewportData.CurrData.LockedReadWrite;
      if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
    end;
  end;
  if not Result then begin
    prErrorMsg := 'Cannot run Drift. The data might be in use or empty.';
    prWikiTopic := 'Error:No_Data';
  end;
  if not Result then Exit;

  if Result then Result := ViewportModel.CurrModel <> nil;
  if Result then Result := not ViewportModel.CurrModel.LockedReadWrite;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.States.StateCount <> 0;
  if Result then Result := not ViewportModel.CurrModel.ModelObject.Rates.RateCount <> 0;
  if not Result then begin
    prErrorMsg := 'Cannot run Staircase. The model might be in use or empty.';
    prWikiTopic := 'Error:No_Model';
  end;
  if not Result then Exit;

  if (0<>TQUB_QVO_QVM(ViewportData.CurrData.CurrModel).ModelObject.Properties['IeqFv'].Data[0]) then begin
    Result := False;
    prErrorMsg := 'Cannot run Staircase. The model uses units of pS, but staircase works in pA.';
    prWikiTopic := 'Error:Cannot_model_pS';
    Exit;
  end;
end;

procedure TQUB_ActionSelector_Modeling.prLClickDrift(OnActionDone :TNotifyEvent);
begin
  prDrift(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickDrift(OnActionDone :TNotifyEvent);
begin
  prDrift(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prDrift(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  TaskDrift     :TQUB_Task_Drift;
begin

  Data          := ViewportData.CurrData;
  Model         := ViewportModel.CurrModel;

  Result := not prCanDrift;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskDrift := TQUB_Task_Drift.Create(True{Suspended});
  TaskDrift.OnTerminate    := OnActionDone;

  TaskDriftPropertiesForm.OnGetProperties := TaskDrift.OnGetProperties;
  TaskDriftPropertiesForm.OnPutProperties := TaskDrift.OnPutProperties;

  TaskDriftPropertiesForm.btnOK.Enabled    := not Result;

  if (Properties) or (Result) then begin
    TaskDriftPropertiesForm.ShowModal;
  end else begin
    TaskDriftPropertiesForm.FormShow(TaskDriftPropertiesForm);
    TaskDriftPropertiesForm.ModalResult := mrOK;
    TaskDriftPropertiesForm.btnOKClick(TaskDriftPropertiesForm);
  end;

  if TaskDriftPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskDrift.Terminate;
  end else if TaskDriftPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskDrift.Terminate;
  end else begin
  //OK
    //TaskDrift.Servers           := prServers;
    TaskDrift.OnChange          := TaskView.OnTaskChange;
    TaskDrift.Data              := Data;
    TaskDrift.Model             := Model;
  end;
  TaskDrift.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prLClickMSearch(OnActionDone :TNotifyEvent);
begin
  prMSearch(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickMSearch(OnActionDone :TNotifyEvent);
begin
  prMSearch(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prMSearch(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  Data          :TQUB_QVO_QVD;
  Model         :TQUB_QVO_QVM;
  TaskMSearch   :TQUB_Task_MSearch;
begin
  
  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_MSearch)) then begin
    Result := True;
    MessageHelpDlg('Only one MSearch task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;
  {
   really, there can be more than one as long as all but one are silent
    -- the MIL histograms are the point of contention --
    maybe worth Tony's suggestion to have updating histograms in the results view
  }

  Data  := ViewportData.CurrData;
  Model := ViewportModel.CurrModel;

  Result := not prCanMIL;

  //if Data = nil then can only use FileList option !!!


  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskMSearch := TQUB_Task_MSearch.Create(True{Suspended});
  TaskMSearch.OnTerminate  := OnActionDone;

  TaskMSearchPropertiesForm.OnGetProperties := TaskMSearch.OnGetProperties;
  TaskMSearchPropertiesForm.OnPutProperties := TaskMSearch.OnPutProperties;

  TaskMSearchPropertiesForm.btnOK.Enabled    := not Result;

  if (Properties) or (Result) then begin
    TaskMSearchPropertiesForm.ShowModal;
  end else begin
    TaskMSearchPropertiesForm.FormShow(TaskMSearchPropertiesForm);
    TaskMSearchPropertiesForm.ModalResult := mrOK;
    TaskMSearchPropertiesForm.btnOKClick(TaskMSearchPropertiesForm);
  end;

  if TaskMSearchPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskMSearch.Terminate;
  end else if TaskMSearchPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskMSearch.Terminate;
  end else begin
  //OK
    //TaskMSearch.Servers           := prServers;
    TaskMSearch.OnChange          := TaskView.OnTaskChange;
    TaskMSearch.Data              := Data;
    TaskMSearch.Model             := Model;
  end;
  TaskMSearch.Resume;
end;


procedure TQUB_ActionSelector_Modeling.prLClickStat(OnActionDone :TNotifyEvent);
begin
  prStat(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickStat(OnActionDone :TNotifyEvent);
begin
  prStat(OnActionDone, True);
end;

function TQUB_ActionSelector_Modeling.prStat(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  TaskStat      :TQUB_Task_Stat;
  Data          :TQUB_QVO_QVD;
begin
  Result                   := False;

  Data := ViewportData.CurrData;
  if Data <> nil then begin
  end else begin
    Result                              := True;
    prErrorMsg                          := 'No data';
    prWikiTopic                         := 'Error:No_Data';
  end;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskStat := TQUB_Task_Stat.Create(True{Suspended});
  TaskStat.OnTerminate := OnActionDone;

  TaskStatPropertiesForm.OnGetProperties := TaskStat.OnGetProperties;
  TaskStatPropertiesForm.OnPutProperties := TaskStat.OnPutProperties;

  TaskStatPropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskStatPropertiesForm.ShowModal;
  end else begin
    TaskStatPropertiesForm.FormShow(TaskStatPropertiesForm);
    TaskStatPropertiesForm.ModalResult := mrOK;
    TaskStatPropertiesForm.btnOKClick(TaskStatPropertiesForm);
  end;

  if TaskStatPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskStat.Terminate;
  end else begin
  //OK
    TaskStat.OnChange          := TaskView.OnTaskChange;
  end;

  TaskStat.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prLClickMdlRpt(OnActionDone :TNotifyEvent);
begin
  prMdlRpt(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Modeling.prRClickMdlRpt(OnActionDone :TNotifyEvent);
begin
  prMdlRpt(OnActionDone, True);
end;

//Chris 06-2004 = Model report 
function TQUB_ActionSelector_Modeling.prMdlRpt(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  TaskMdlRpt      :TQUB_Task_MdlRpt;
  Data            :TQUB_QVO_QVD;
begin
  Data := g_CurrData();

  Result := not prCanModelData('make model report', True);

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskMdlRpt := TQUB_Task_MdlRpt.Create(True{Suspended});
  TaskMdlRpt.OnTerminate := OnActionDone;

  TaskMdlRptPropertiesForm.btnOK.Enabled := not Result;

  if (Properties) or (Result) then begin
    TaskMdlRptPropertiesForm.ShowModal;
  end else begin
    TaskMdlRptPropertiesForm.FormShow(TaskStatPropertiesForm);
    TaskMdlRptPropertiesForm.ModalResult := mrOK;
    TaskMdlRptPropertiesForm.btnOKClick(TaskStatPropertiesForm);
  end;

  if TaskMdlRptPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskMdlRpt.Terminate;
  end else begin
  //OK
    TaskMdlRpt.OnChange          := TaskView.OnTaskChange;
    TaskMdlRpt.Data              := Data.GetChosenData;
    TaskMdlRpt.Properties        := TaskMdlRptPropertiesForm.Properties.Clone(True);
    TaskMdlRpt.Properties.AppendChild( TQUB_QVO_QVM(Data.CurrModel).GetAsNode );
  end;

  TaskMdlRpt.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prLClickIV(OnActionDone :TNotifyEvent);
var
  TaskIV        :TQUB_Task_IVPlot;
  Data          :TQUB_QVO_QVD;
begin
  
  TaskIV := TQUB_Task_IVPlot.Create(True{Suspended});

  Data := ViewportData.CurrData;
  if Data = nil then begin
    MessageHelpDlg('No data', mtError, [mbOK], 0, 'Error:No_Data');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end else begin
    TaskIV.Data := Data;
  end;

  TaskIV.OnTerminate := OnActionDone;
  TaskIV.OnChange          := TaskView.OnTaskChange;

  TaskIV.Resume;
end;

procedure TQUB_ActionSelector_Modeling.prRClickMMerge(OnActionDone :TNotifyEvent);
var
  Models     :TList;
  i, j       :Integer;
begin
  TaskMMergePropertiesForm.ShowModal;

  if TaskMMergePropertiesForm.ModalResult = mrOK then begin
    Models   := TList.Create;
    try
      for i := 0 to TaskMMergePropertiesForm.ModelCount - 1 do begin
        for j := 0 to TaskMMergePropertiesForm.ModelReps[i] - 1 do begin
        Models.Add( TaskMMergePropertiesForm.Models[i] );
      end;
      end;
      ViewportModel.Merge(Models);
    except

    end;
    Models.Free;
  end;

  if Assigned(OnActionDone) then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Modeling.prClickMultiFit(OnActionDone :TNotifyEvent);
begin
  if ViewportData.CurrData = nil
  then MessageHelpDlg('No data', mtError, [mbOK], 0, 'Error:No_Data')
  else MultiFitForm.ShowModal;

  if Assigned(OnActionDone) then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Modeling.prClickSurvey(OnActionDone :TNotifyEvent);
var
  Task            :TQUB_Task_Survey;
  Trouble         :Boolean;
begin
  Trouble := not prCanModelData('Survey', True);

  if Trouble then begin
    MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
    prWikiTopic := '';
  end;

  Task := TQUB_Task_Survey.Create(True{Suspended});
  Task.OnTerminate := OnActionDone;

  TaskSurveyPropertiesForm.OnGetProperties := Task.OnGetProperties;
  TaskSurveyPropertiesForm.OnPutProperties := Task.OnPutProperties;
  TaskSurveyPropertiesForm.btnOK.Enabled := not Trouble;
  TaskSurveyPropertiesForm.ShowModal;

  if TaskSurveyPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    Task.Terminate;
  end else begin
  //OK
    Task.OnChange          := TaskView.OnTaskChange;
    Task.MacProperties     := TaskMacPropertiesForm;
    Task.MIPProperties     := TaskMIPPropertiesForm;
    Task.MILProperties     := TaskMILPropertiesForm;
  end;

  Task.Resume;
end;


//****************************************************************************//
end.
