unit QUB_ActionSelector_Preprocessing;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls, Menus,

QUB_Global,
QUB_Action, QUB_ActionSelector,
QUB_QVO_QVD,
QUB_Task_ExtractSelection,
QUB_Task_BaseLineMT,
QUB_Task_IdlStim,
QUB_Task_Stimulus,
QUB_Task_Kalman_Properties,
QUB_Task_ExtractSelection_Properties,
QUB_Task_ApplyBaseLine_Properties,
QUB_Task_EditIdl_Properties,
QUB_Task_IdlStim_Properties,
QUB_Task_Stimulus_Properties,
QUB_Node;


type


TQUB_ActionSelector_Preprocessing = class(TQUB_ActionSelector)
protected


  prErrorMsg                         :String;
  prWikiTopic                        :String;

  prIdlStimChannel                   :Integer;


  function prCanExtract: Boolean;
  procedure prLClickExtract(OnActionDone :TNotifyEvent);
  procedure prRClickExtract(OnActionDone :TNotifyEvent);
  procedure prDoExtract(OnActionDone :TNotifyEvent; Properties :Boolean);

  function prCanBaseline: Boolean;
  procedure prLClickBaseline(OnActionDone :TNotifyEvent);
  procedure prRClickBaseline(OnActionDone :TNotifyEvent);
  procedure prDoBaseline(OnActionDone :TNotifyEvent; Properties :Boolean);

  procedure prLClickChopData(OnActionDone :TNotifyEvent);
  procedure prRClickChopData(OnActionDone :TNotifyEvent);
  procedure prLClickChopEvents(OnActionDone :TNotifyEvent);
  procedure prRClickChopEvents(OnActionDone :TNotifyEvent);
  procedure prLClickEditIdl(OnActionDone :TNotifyEvent);
  procedure prRClickEditIdl(OnActionDone :TNotifyEvent);
  function prDoEditIdl(Properties :Boolean; OnActionDone :TNotifyEvent) :Boolean;
  procedure prLClickGoToPoint(OnActionDone :TNotifyEvent);
  procedure prRClickGoToPoint(OnActionDone :TNotifyEvent);

  function prIdlStimUserInsistsChannelZero: Boolean;
  procedure prLClickIdlStim(OnActionDone :TNotifyEvent);
  procedure prRClickIdlStim(OnActionDone :TNotifyEvent);
  function prDoIdlStim(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
  procedure prLClickStimulus(OnActionDone :TNotifyEvent);
  procedure prRClickStimulus(OnActionDone :TNotifyEvent);
  function prDoStimulus(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
  procedure prLClickKalman(OnActionDone :TNotifyEvent);
  procedure prRClickKalman(OnActionDone :TNotifyEvent);
  function prDoKalman(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;

  procedure prRestoreDefaults; override;


public


  TaskExtractSelectionPropertiesForm :TQUB_Task_ExtractSelection_PropertiesForm;
  TaskApplyBaseLinePropertiesForm    :TQUB_Task_ApplyBaseLine_PropertiesForm;
  TaskIdlStimPropertiesForm          :TQUB_Task_IdlStim_PropertiesForm;
  TaskStimulusPropertiesForm         :TQUB_Task_Stimulus_PropertiesForm;
  TaskKalmanPropertiesForm           :TQUB_Task_Kalman_PropertiesForm;
  TaskEditIdlPropertiesForm          :TQUB_Task_EditIdl_Properties;


  constructor Create(AOwner :TComponent); reintroduce;


end;


implementation


{$R *.DFM}


uses


QUB_Viewport,
QUB_Viewport_Data,
QUB_Viewport_Model,
QUB_GUIInterface,
QUB_Profiles,
QUB_Main,
QUB_Task_EditIdl,
QUB_Task_Kalman,
QUB_Task;


// *****************************************************************************
// self-destructing OnTerminate for Task_Extract
// first displays the extracted data if requested, then calls OnActionDone


type


TQUB_ActionSelector_Pre_AfterExtract = class
protected
  prOnActionDone              :TNotifyEvent;
public
  constructor Create(OnActionDone :TNotifyEvent);
  procedure OnExtractDone(Sender :TObject);
end;

constructor TQUB_ActionSelector_Pre_AfterExtract.Create(OnActionDone :TNotifyEvent);
begin
  prOnActionDone := OnActionDone;
end;

procedure TQUB_ActionSelector_Pre_AfterExtract.OnExtractDone(Sender :TObject);
begin
  if (Sender is TQUB_Task_ExtractSelection)              and
     (not TQUB_Task_ExtractSelection(Sender).Error)      and
     (TQUB_Task_ExtractSelection(Sender).OpenFileAfter)  then begin
    ViewportData.Open(TQUB_Task_ExtractSelection(Sender).FileName, False{Silent});
  end;

  if @prOnActionDone <> nil then prOnActionDone(Sender);
end;


// *****************************************************************************
constructor TQUB_ActionSelector_Preprocessing.Create(AOwner :TComponent);
begin
  inherited CreateNamed(AOwner, 'Preprocessing');

  TaskApplyBaselinePropertiesForm := TQUB_Task_ApplyBaseline_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Baseline', prLClickBaseline, prRClickBaseline, TaskApplyBaselinePropertiesForm.SetGetProperties,
                                  'Correct baseline drift','Base'));

  TaskExtractSelectionPropertiesForm := TQUB_Task_ExtractSelection_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Extract', prLClickExtract, prRClickExtract, TaskExtractSelectionPropertiesForm.SetGetProperties,
                                  'Extract data or idealization', 'Ext'));

  prAddAction( TQUB_Action.Create('Chop Data', prLClickChopData, prRClickChopData, ViewportData.ChopPropertiesForm.PropertiesTree,
                                  'Chop data into segments of fixed size', 'Chop'));

  prAddAction( TQUB_Action.Create('Chop Idl', prLClickChopEvents, prRClickChopEvents, ViewportData.EvtListProcessorPropertiesForm.SetGetProperties,
                                  'Chop idealized data into segments', 'ChpIdl'));

  TaskEditIdlPropertiesForm := TQUB_Task_EditIdl_Properties.Create(Application);
  prAddAction( TQUB_Action.Create('Edit Idl', prLClickEditIdl, prRClickEditIdl, TaskEditIdlPropertiesForm.SetGetProperties,
                                  'Modify idealized data', 'EdIdl'));

  prAddAction( TQUB_Action.Create('Go To', prLClickGoToPoint, prRClickGoToPoint, ViewportData.GoToPointPropertiesForm.PropertiesTree,
                                  'Show a point/time in the data', 'GoTo'));

  TaskIdlStimPropertiesForm := TQUB_Task_IdlStim_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('IdlStim', prLClickIdlStim, prRClickIdlStim, TaskIdlStimPropertiesForm.SetGetProperties,
                                  'Idealize experimental conditions', 'IdlStm'));
  TaskStimulusPropertiesForm := TQUB_Task_Stimulus_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Stimulus', prLClickStimulus, prRClickStimulus, TaskStimulusPropertiesForm.SetGetProperties,
                                  'Prepare stimulus from an A/D channel for Mac and MIP', 'Stim'));

  TaskKalmanPropertiesForm  := TQUB_Task_Kalman_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Bkg Sub', prLClickKalman, prRClickKalman, TaskKalmanPropertiesForm.SetGetProperties,
                                  'Remove baseline drift and periodic waveforms', 'Bkg-'));

  RestoreDefaults;

  prIdlStimChannel    := 0; // so it asks the first time
end;

procedure TQUB_ActionSelector_Preprocessing.prRestoreDefaults;
var
  Vis                 :TStringList;
begin
  Vis                 := ActionByName['Baseline'].VisLayouts;
  Vis.Clear;

  Vis                 := ActionByName['Extract'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Modeling');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Results');

  Vis                 := ActionByName['Chop Data'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Modeling');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Results');

  Vis                 := ActionByName['Chop Idl'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Results');

  Vis                 := ActionByName['Edit Idl'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Results');

  Vis                 := ActionByName['Go To'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Modeling');
  Vis.Add('Simulation');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');
  Vis.Add('Results');

  Vis                 := ActionByName['IdlStim'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');

  Vis                 := ActionByName['Stimulus'].VisLayouts;
  Vis.Clear;
  Vis.Add('Modeling');
  Vis.Add('Preprocessing');
  Vis.Add('Macroscopic');

  Vis                 := ActionByName['Bkg Sub'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
end;

function TQUB_ActionSelector_Preprocessing.prCanExtract: Boolean;
begin
  Result := ViewportData.CurrData <> nil;
  if Result then Result := not ViewportData.CurrData.LockedReadWrite;
  if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
  if not Result then begin
    prErrorMsg  := 'Cannot extract selection. The file might be in use or empty';
    prWikiTopic := 'Error:No_Data';
  end;
end;

procedure TQUB_ActionSelector_Preprocessing.prLClickExtract(OnActionDone :TNotifyEvent);
begin
  prDoExtract(OnActionDone, TaskExtractSelectionPropertiesForm.Properties['DialogOnLClick'].Data[0]<>0);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickExtract(OnActionDone :TNotifyEvent);
begin
  prDoExtract(OnActionDone, True);
end;

procedure TQUB_ActionSelector_Preprocessing.prDoExtract(OnActionDone :TNotifyEvent; Properties :Boolean);
var
  Data          :TQUB_QVO_QVD;
  TaskExt       :TQUB_Task_ExtractSelection;
  AfterExtract  :TQUB_ActionSelector_Pre_AfterExtract;
begin

  Data          := ViewportData.CurrData;

  if not prCanExtract then begin
    MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
    prWikiTopic := '';
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  //DataSource
  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskExtractSelectionPropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskExtractSelectionPropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  TaskExt                   := TQUB_Task_ExtractSelection.Create(True{Suspended});
  if ProfileMgr['Ext'].Active['FileName'].DataCount > 0 then begin
    // python script override meant to be used with DialogOnLClick=0
    TaskExt.FileName            := ProfileMgr['Ext'].Active['FileName'].DataAsString;
  end else if (TaskExtractSelectionPropertiesForm.DataSource = DATA_SOURCE_SELECTIONLIST) and (Data.CurrList <> nil) then begin
    try
      TaskExt.FileName          := ChangeFileExt(Data.FileName, '_' + SafeFileName(Data.SelectionLists.Items.Strings[Data.SelectionLists.Items.IndexOfObject(Data.CurrList)]));
    except
      TaskExt.FileName          := ChangeFileExt(Data.FileName, '_x');
    end;
  end else if (TaskExtractSelectionPropertiesForm.DataSource = DATA_SOURCE_SELECTION) then begin
  //TO DO: LM. Adjust for draw list overlapped
    TaskExt.FileName          := ChangeFileExt(Data.FileName, Format('_%d-%d', [Data.DataSel.First, Data.DataSel.Last]));
  end else begin
    TaskExt.FileName          := ChangeFileExt(Data.FileName, '_x');
  end;

  AfterExtract              := TQUB_ActionSelector_Pre_AfterExtract.Create(OnActionDone);
  TaskExt.OnTerminate       := AfterExtract.OnExtractDone;

  TaskExtractSelectionPropertiesForm.SourceADChannelCount := Data.DataObject.ADChannelCount;
  TaskExtractSelectionPropertiesForm.OnGetProperties := TaskExt.OnGetProperties;
  TaskExtractSelectionPropertiesForm.OnPutProperties := TaskExt.OnPutProperties;

  TaskExtractSelectionPropertiesForm.SamplingFrequency := 1 / Data.Sampling;

  if Data <> nil then begin
    TaskExtractSelectionPropertiesForm.lstOperationList.Items.Assign(Data.SelectionLists.Items);
  end else begin
    TaskExtractSelectionPropertiesForm.lstOperationList.Items.Clear;
  end;

  if Properties then begin
    TaskExtractSelectionPropertiesForm.ShowModal;
  end else begin
    TaskExtractSelectionPropertiesForm.FormShow(TaskExtractSelectionPropertiesForm);
    TaskExtractSelectionPropertiesForm.ModalResult := mrOK;
    TaskExtractSelectionPropertiesForm.btnOKClick(TaskExtractSelectionPropertiesForm);
  end;

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST)
          and (TaskExtractSelectionPropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskExtractSelectionPropertiesForm.DataSource;
  end;

  if TaskExtractSelectionPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskExt.Terminate;
  end else begin
  //OK
    TaskExt.OnChange          := TaskView.OnTaskChange;
    TaskExt.Data              := Data;
  end;
  TaskExt.Resume;
end;

function TQUB_ActionSelector_Preprocessing.prCanBaseline: Boolean;
begin
  Result := ViewportData.CurrData <> nil;
  if Result then Result := not ViewportData.CurrData.LockedReadWrite;
  if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
  if not Result then begin
    prErrorMsg := 'Cannot apply baseline correction. The file might be in use or empty';
    prWikiTopic := 'Error:No_Data';
  end;
end;

procedure TQUB_ActionSelector_Preprocessing.prLClickBaseline(OnActionDone :TNotifyEvent);
begin
  prDoBaseline(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickBaseline(OnActionDone :TNotifyEvent);
begin
  prDoBaseline(OnActionDone, True);
end;

procedure TQUB_ActionSelector_Preprocessing.prDoBaseline(OnActionDone :TNotifyEvent; Properties :Boolean);
var
  CantRun       :Boolean;
  Data          :TQUB_QVO_QVD;
  TaskBase      :TQUB_Task_BaselineMT;
begin

  Data := ViewportData.CurrData;

  CantRun := not prCanBaseline;

  if CantRun then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskBase := TQUB_Task_BaselineMT.Create(True{Suspended});
  TaskBase.OnTerminate := OnActionDone;

  //DataSource
  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskApplyBaselinePropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskApplyBaselinePropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  TaskApplyBaseLinePropertiesForm.OnGetProperties := TaskBase.OnGetProperties;
  TaskApplyBaseLinePropertiesForm.OnPutProperties := TaskBase.OnPutProperties;

  TaskApplyBaseLinePropertiesForm.btnOK.Enabled    := not CantRun;
  TaskApplyBaseLinePropertiesForm.btnReset.Enabled := not CantRun;

  if (Properties) or (CantRun) then begin
    TaskApplyBaseLinePropertiesForm.ShowModal;
  end else begin
    TaskApplyBaseLinePropertiesForm.FormShow(TaskApplyBaseLinePropertiesForm);
    TaskApplyBaseLinePropertiesForm.ModalResult := mrOK;
    TaskApplyBaseLinePropertiesForm.btnOKClick(TaskApplyBaseLinePropertiesForm);
  end;

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST)
          and (TaskApplyBaseLinePropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskApplyBaseLinePropertiesForm.DataSource;
  end;

  if TaskApplyBaseLinePropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskBase.Terminate;
  end else if TaskApplyBaseLinePropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskBase.Terminate;
    Application.ProcessMessages;
    Data.ClearBaseLineNodes(TaskApplyBaseLinePropertiesForm.Properties['ChannelIndex'].Data[0]);
  end else begin
  //OK
    TaskBase.OnChange          := TaskView.OnTaskChange;
    TaskBase.Data              := Data;
  end;
  TaskBase.Resume;
end;

procedure TQUB_ActionSelector_Preprocessing.prLClickChopData(OnActionDone :TNotifyEvent);
begin
  ViewportData.ChopSegments(ViewportData.ChopPropertiesForm.DialogOnLClick);
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickChopData(OnActionDone :TNotifyEvent);
begin
  ViewportData.ChopSegments(True);
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Preprocessing.prLClickChopEvents(OnActionDone :TNotifyEvent);
begin
  ViewportData.EvtListProcess(ViewportData.EvtListProcessorPropertiesForm.Properties['DialogOnLClick'].Data[0]<>0, OnActionDone);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickChopEvents(OnActionDone :TNotifyEvent);
begin
  ViewportData.EvtListProcess(True, OnActionDone);
end;

procedure TQUB_ActionSelector_Preprocessing.prLClickEditIdl(OnActionDone :TNotifyEvent);
begin
  prDoEditIdl(TaskEditIdlPropertiesForm.Properties['DialogOnLClick'].Data[0]<>0, OnActionDone);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickEditIdl(OnActionDone :TNotifyEvent);
begin
  prDoEditIdl(True, OnActionDone);
end;

function TQUB_ActionSelector_Preprocessing.prDoEditIdl(Properties :Boolean; OnActionDone :TNotifyEvent) :Boolean;
var
  Task               :TQUB_Task_EditIdl;
begin
  Result     := False;

  if ViewportData.CurrData = nil then begin
    MessageHelpDlg('No data.', mtError, [mbOK], 0, 'Error:No_data');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Task       := TQUB_Task_EditIdl.Create(True{Suspended});
  Task.OnTerminate := OnActionDone;
  Task.Data  := ViewportData.CurrData;

  TaskEditIdlPropertiesForm.OnGetProperties := Task.OnGetProperties;
  TaskEditIdlPropertiesForm.OnPutProperties := Task.OnPutProperties;
  TaskEditIdlPropertiesForm.ClassCount      := ViewportData.CurrData.DataObject.Idealization[ViewportData.CurrData.ActiveChannel].ClsCount;

  if (Properties) or (Result) then begin
    TaskEditIdlPropertiesForm.ShowModal;
  end else begin
    TaskEditIdlPropertiesForm.FormShow(TaskEditIdlPropertiesForm);
    TaskEditIdlPropertiesForm.ModalResult := mrOK;
    TaskEditIdlPropertiesForm.btnOKClick(TaskEditIdlPropertiesForm);
  end;

  if TaskEditIdlPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    Task.Terminate;
  end else if TaskEditIdlPropertiesForm.ModalResult = mrNo then begin
  //Reset
    Task.Terminate;
  //end else begin
  //OK
  end;
  Task.Resume;

end;

procedure TQUB_ActionSelector_Preprocessing.prLClickGoToPoint(OnActionDone :TNotifyEvent);
begin
  ViewportData.GoToPoint(ViewportData.GoToPointPropertiesForm.DialogOnLClick);
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickGoToPoint(OnActionDone :TNotifyEvent);
begin
  ViewportData.GoToPoint(True);
  if @OnActionDone <> nil then OnActionDone(nil);
end;

function TQUB_ActionSelector_Preprocessing.prIdlStimUserInsistsChannelZero: Boolean;
begin
  Result := mrYes = MessageHelpDlg('Data channel 0 is probably not a stimulus channel.  Idealize it with the stimulus idealizer anyway?',
                               mtWarning, [mbYes, mbNo], 0, 'Warning:Channel_0_not_a_stimulus');
  if not Result then TaskIdlStimPropertiesForm.GetsFocus := TaskIdlStimPropertiesForm.txtXChannelIndex;
end;

procedure TQUB_ActionSelector_Preprocessing.prLClickIdlStim(OnActionDone :TNotifyEvent);
begin
  prDoIdlStim(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickIdlStim(OnActionDone :TNotifyEvent);
begin
  prDoIdlStim(OnActionDone, True);
end;

function TQUB_ActionSelector_Preprocessing.prDoIdlStim(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  TaskIdlStim              :TQUB_Task_IdlStim;
  Data                     :TQUB_QVO_QVD;
begin

  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_IdlStim)) then begin
    Result := True;
    MessageHelpDlg('Only one IdlStim task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Result := False;
  Data := TQUB_QVO_QVD(ViewportData.CurrData);
  if Data = nil then begin
    Result := True;
    prErrorMsg := 'No data available.';
    prWikiTopic := 'Error:No_Data';
  end;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskIdlStim := TQUB_Task_IdlStim.Create(True{Suspended});
  TaskIdlStim.OnTerminate   := OnActionDone;

  TaskIdlStimPropertiesForm.OnGetProperties := TaskIdlStim.OnGetProperties;
  TaskIdlStimPropertiesForm.OnPutProperties := TaskIdlStim.OnPutProperties;

  TaskIdlStimPropertiesForm.btnOK.Enabled   := not Result;

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskIdlStimPropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskIdlStimPropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  if Data <> nil then Data.ActiveChannel := prIdlStimChannel;

  repeat

    if (Properties) or (Result) then begin
      TaskIdlStimPropertiesForm.ShowModal;
    end else begin
      TaskIdlStimPropertiesForm.FormShow(TaskIdlStimPropertiesForm);
      TaskIdlStimPropertiesForm.ModalResult := mrOK;
      TaskIdlStimPropertiesForm.btnOKClick(TaskIdlStimPropertiesForm);
    end;

    Properties := True; // in case we have to repeat. it is not used afterward

  until (TaskIdlStimPropertiesForm.ModalResult <> mrOK)
  or    (TaskIdlStimPropertiesForm.Properties['ChannelIndex'].Data[0] > 0)
  or    prIdlStimUserInsistsChannelZero;

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and (TaskIdlStimPropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskIdlStimPropertiesForm.DataSource;
  end;

  if TaskIdlStimPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskIdlStim.Terminate;
  end else if TaskIdlStimPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskIdlStim.Terminate;
  end else begin
  //OK
    TaskIdlStim.OnChange     := TaskView.OnTaskChange;
    TaskIdlStim.ChannelIndex := Data.ActiveChannel;
    TaskIdlStim.Data         := Data;
  end;
  TaskIdlStim.Resume;
  if Data <> nil then begin
    prIdlStimChannel   := Data.ActiveChannel;
    Data.ActiveChannel := 0;
  end;
end;



procedure TQUB_ActionSelector_Preprocessing.prLClickStimulus(OnActionDone :TNotifyEvent);
begin
  prDoStimulus(OnActionDone, True);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickStimulus(OnActionDone :TNotifyEvent);
begin
  prDoStimulus(OnActionDone, True);
end;

function TQUB_ActionSelector_Preprocessing.prDoStimulus(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  TaskStimulus             :TQUB_Task_Stimulus;
  Data                     :TQUB_QVO_QVD;
begin

  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_Stimulus)) then begin
    Result := True;
    MessageHelpDlg('Only one Stimulus task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Result := False;
  Data := TQUB_QVO_QVD(ViewportData.CurrData);
  if Data = nil then begin
    Result := True;
    prErrorMsg := 'No data available.';
    prWikiTopic := 'Error:No_Data';
  end;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskStimulus := TQUB_Task_Stimulus.Create(True{Suspended});
  TaskStimulus.OnTerminate   := OnActionDone;

  TaskStimulusPropertiesForm.OnGetProperties := TaskStimulus.OnGetProperties;
  TaskStimulusPropertiesForm.OnPutProperties := TaskStimulus.OnPutProperties;

  TaskStimulusPropertiesForm.btnOK.Enabled   := not Result;

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskStimulusPropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskStimulusPropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  if (Properties) or (Result) then begin
    TaskStimulusPropertiesForm.ShowModal;
  end else begin
    TaskStimulusPropertiesForm.FormShow(TaskStimulusPropertiesForm);
    TaskStimulusPropertiesForm.ModalResult := mrOK;
    TaskStimulusPropertiesForm.btnOKClick(TaskStimulusPropertiesForm);
  end;

  Properties := True; // in case we have to repeat. it is not used afterward

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and (TaskStimulusPropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskStimulusPropertiesForm.DataSource;
  end;

  if TaskStimulusPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskStimulus.Terminate;
  end else if TaskStimulusPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskStimulus.Terminate;
  end else begin
  //OK
    TaskStimulus.OnChange     := TaskView.OnTaskChange;
    TaskStimulus.ChannelIndex := TaskStimulusPropertiesForm.Properties['Stimulus Channel'].Data[0];
    TaskStimulus.Data         := Data;
  end;
  TaskStimulus.Resume;
end;


procedure TQUB_ActionSelector_Preprocessing.prLClickKalman(OnActionDone :TNotifyEvent);
begin
  prDoKalman(OnActionDone, True);
end;

procedure TQUB_ActionSelector_Preprocessing.prRClickKalman(OnActionDone :TNotifyEvent);
begin
  prDoKalman(OnActionDone, True);
end;

function TQUB_ActionSelector_Preprocessing.prDoKalman(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  TaskKalman             :TQUB_Task_Kalman;
  Data                     :TQUB_QVO_QVD;
begin

  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_Kalman)) then begin
    Result := True;
    MessageHelpDlg('Only one Kalman task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Result := False;
  Data := TQUB_QVO_QVD(ViewportData.CurrData);
  if Data = nil then begin
    Result := True;
    prErrorMsg := 'No data available.';
    prWikiTopic := 'Error:No_Data';
  end;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskKalman := TQUB_Task_Kalman.Create(True{Suspended});
  TaskKalman.OnTerminate   := OnActionDone;

  TaskKalmanPropertiesForm.OnGetProperties := TaskKalman.OnGetProperties;
  TaskKalmanPropertiesForm.OnPutProperties := TaskKalman.OnPutProperties;

  TaskKalmanPropertiesForm.btnOK.Enabled   := not Result;

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskKalmanPropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskKalmanPropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  if (Properties) or (Result) then begin
    TaskKalmanPropertiesForm.ShowModal;
  end else begin
    TaskKalmanPropertiesForm.FormShow(TaskKalmanPropertiesForm);
    TaskKalmanPropertiesForm.ModalResult := mrOK;
    TaskKalmanPropertiesForm.btnOKClick(TaskKalmanPropertiesForm);
  end;

  Properties := True; // in case we have to repeat. it is not used afterward

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and (TaskKalmanPropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskKalmanPropertiesForm.DataSource;
  end;

  if TaskKalmanPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskKalman.Terminate;
  end else if TaskKalmanPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskKalman.Terminate;
  end else begin
  //OK
    TaskKalman.OnChange     := TaskView.OnTaskChange;
    TaskKalman.Data         := Data;
  end;
  TaskKalman.Resume;
end;


//****************************************************************************//
end.
