unit QUB_ActionSelector_Simulation;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls, Menus, Buttons, Math,

QUB_Global,
QUB_Action, QUB_ActionSelector,
QUB_Task_Simulate,
QUB_Task_Simulate_Properties,
QUB_Task_StimSim_Properties,
QUB_Task_Sim_DoseResponse,
QUB_Task_Sim_DoseResponse_Properties,
QUB_QVO_QVD,
QUB_ExpCond_Parameters,
QUB_Node;


type


TQUB_ActionSelector_Simulation = class(TQUB_ActionSelector)
protected

  prSimData               :TQUB_QVO_QVD;
  prTaskSimLive           :TQUB_Task_SimLive;


  procedure prLClickSim(OnActionDone :TNotifyEvent);
  procedure prRClickSim(OnActionDone :TNotifyEvent);
  procedure prDoSim(OnActionDone :TNotifyEvent; Properties :Boolean);
  // function prPropertiesSim(arg :INode): INode;

  procedure prLClickStim(OnActionDone :TNotifyEvent);
  procedure prRClickStim(OnActionDone :TNotifyEvent);
  procedure prDoStim(OnActionDone :TNotifyEvent; Properties :Boolean);

  //procedure prPrepareDefExpCond;
  procedure prPrepareExpCond;

  procedure prRestoreDefaults; override;

  procedure prSimLive(OnActionDone :TNotifyEvent);
  procedure prSimLiveDone(Sender :TObject);

  procedure prLClickSimDoseResponse(OnActionDone :TNotifyEvent);
  procedure prRClickSimDoseResponse(OnActionDone :TNotifyEvent);
  procedure prDoSimDoseResponse(OnActionDone :TNotifyEvent; Properties :Boolean);


public


  TaskSimulatePropertiesForm      :TQUB_Task_Simulate_PropertiesForm;
  TaskStimSimPropertiesForm       :TQUB_Task_StimSim_PropertiesForm;
  TaskSimDoseResponsePropertiesForm :TQUB_Task_Sim_DoseResponse_PropertiesForm;


  constructor Create(AOwner :TComponent); reintroduce;
  destructor Destroy; override;


  procedure DoOnCloseData(Sender :TObject);


  procedure StopSimLive;


end;


implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Profiles,
QUB_Viewport_Model,
QUB_QVO_QVM, QUB_QFS_QMF, QUB_QFS_Channels,
QUB_QFS_QDF_Global,
QUB_Viewport_Data;


//******************************************************************************
constructor TQUB_ActionSelector_Simulation.Create(AOwner :TComponent);
begin
  inherited CreateNamed(AOwner, 'Simulation');

  TaskSimulatePropertiesForm := TQUB_Task_Simulate_PropertiesForm.Create(Application);
  TaskStimSimPropertiesForm  := TQUB_Task_StimSim_PropertiesForm.Create(Application);
  TaskSimDoseResponsePropertiesForm  := TQUB_Task_Sim_DoseResponse_PropertiesForm.Create(Application);

  prAddAction(TQUB_Action.Create('Simulate', prLClickSim, prRClickSim, TaskSimulatePropertiesForm.SetGetProperties,
                                 'Simulate', 'Sim'));

  prAddAction( TQUB_Action.Create('Live Sim', prSimLive, prSimLive, nil,
                                  'Re-simulate whenever the model changes','LiveS',clRed));

  prAddAction(TQUB_Action.Create('Stim Sim', prLClickStim, prRClickStim, TaskStimSimPropertiesForm.SetGetProperties,
                                 'Simulate with a stimulus waveform', 'Stim'));

  prAddAction(TQUB_Action.Create('Dose-Resp', prLClickSimDoseResponse, prRClickSimDoseResponse, TaskSimDoseResponsePropertiesForm.SetGetProperties,
                                 'Simulate macroscopic response to a pulse train.', 'DRsp'));

  QUB_GUIInterface.TaskSimulatePropertiesForm := TaskSimulatePropertiesForm;

  RestoreDefaults;
end;

destructor TQUB_ActionSelector_Simulation.Destroy;
begin
  try

  finally
    inherited;
  end;
end;

procedure TQUB_ActionSelector_Simulation.prRestoreDefaults;
var
  Vis :TStringList;
begin
  Vis := ActionByName['Simulate'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Simulation');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis := ActionByName['Live Sim'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Simulation');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis := ActionByName['Stim Sim'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Simulation');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');

  Vis := ActionByName['Dose-Resp'].VisLayouts;
  Vis.Clear;
  Vis.Add('Simulation');
  Vis.Add('Macroscopic');
  Vis.Add('Modeling');
  Vis.Add('Results');
end;

procedure TQUB_ActionSelector_Simulation.DoOnCloseData(Sender :TObject);
var
  i           :Integer;
begin
  if Sender = prSimData then begin
    prSimData := nil;
  end;
  for i := 0 to ViewportModel.ListObjects.Items.Count - 1 do begin
    if Sender = TQUB_QVO_QVM(ViewportModel.ListObjects[i]).SimData then begin
      TQUB_QVO_QVM(ViewportModel.ListObjects[i]).SimData := nil;
    end;
  end;
end;

procedure TQUB_ActionSelector_Simulation.prLClickSim(OnActionDone :TNotifyEvent);
begin
  prDoSim(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Simulation.prRClickSim(OnActionDone :TNotifyEvent);
begin
  prDoSim(OnActionDone, True);
end;

{
This method unionizes the Parameters of all QVM's into the ViewportModel DefExpCond
It does so in a rather strange way.
}
procedure EnsureExpCond(Params :TQUB_ExpCond_Parameters; Name :String);
begin
  if Params.ParameterByName[Name] <> nil then Exit;
  Params.AddPar(Name, 1.0, -1, False, -1);
end;

procedure TQUB_ActionSelector_Simulation.prPrepareExpCond;
var
  Model         :TQUB_QVO_QVM;
  i, j, k       :Integer;
  Params        :TQUB_ExpCond_Parameters;
begin
  Params        := ViewportModel.DefExpCond;
  
  for i := 0 to ViewportModel.ListObjects.Items.Count - 1 do begin
    Model       := TQUB_QVO_QVM(ViewportModel.ListObjects[i]);
    for j := 0 to Model.ModelObject.Rates.RateCount - 1 do begin
      for k := 1 to 2 do begin
        if Model.ModelObject.Rates[j].P[k] then EnsureExpCond(Params, Model.ModelObject.Rates[j].PName[k]);
        if Model.ModelObject.Rates[j].Q[k] then EnsureExpCond(Params, Model.ModelObject.Rates[j].QName[k]);
      end;
    end;
  end;

  i             := Params.IndexOf('ChannelCount');
  if i >= 0  then  Params.DelPar(i);
end;

procedure TQUB_ActionSelector_Simulation.prDoSim(OnActionDone :TNotifyEvent; Properties :Boolean);
var
  TaskSimulate  :TQUB_Task_Simulate;
  Stored        :IQUB_Task_Simulate_Stored;
  Model         :TQUB_QVO_QVM;
  SimData       :TQUB_QVO_QVD;
  i             :Integer;
  FName         :String;
begin
  StopSimLive;
  while prTaskSimLive <> nil do Application.ProcessMessages;

  //Task
  TaskSimulate := TQUB_Task_Simulate.Create(True{Suspended});
  TaskSimulate.OnTerminate := OnActionDone;

  //TaskSimulatePropertiesForm.Kinetic := TaskSimulate.Kinetic;

  //Testing .............................................................
  prPrepareExpCond;
  //.....................................................................

  TaskSimulatePropertiesForm.OnGetProperties := TaskSimulate.OnGetProperties;
  TaskSimulatePropertiesForm.OnPutProperties := TaskSimulate.OnPutProperties;

  SimData        := nil;
  Model          := ViewportModel.CurrModel;
  if Model <> nil then begin
    SimData      := TQUB_QVO_QVD(Model.SimData); // replace/append into last-simmed-this-model
  end;
  if SimData = nil then begin
    SimData      := prSimData;                   // or if none, try last-simmed-any-model
  end;

  if (SimData = nil) or (SimData.LockedAny) then begin
    TaskSimulatePropertiesForm.DataMode := 0;//New Data
  end else begin
    if TaskSimulatePropertiesForm.DataMode = 0 then begin
      TaskSimulatePropertiesForm.DataMode := 1;//Replace Data
    end;
  end;
  TaskSimulatePropertiesForm.chkDataModeReplace.Enabled     := SimData <> nil;
  TaskSimulatePropertiesForm.chkDataModeAppend.Enabled      := SimData <> nil;
  TaskSimulatePropertiesForm.chkDataModeAppendCurr.Enabled  := ViewportData.CurrData <> nil;

  TaskSimulatePropertiesForm.lstModels.Items.Assign(ViewportModel.ListObjects.Items);

  TaskSimulatePropertiesForm.Model := Model;

  TaskSimulatePropertiesForm.lstStimData.Items.Assign(ViewportProtocol.ListObjects.Items);
  TaskSimulatePropertiesForm.ProtoCount := ViewportProtocol.ListObjects.Items.Count;
  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    TaskSimulatePropertiesForm.lstStimData.Items.AddObject(ViewportData.ListObjects.Items[i], ViewportData.ListObjects.Items.Objects[i]);
  end;

  if Properties then begin
    TaskSimulatePropertiesForm.ShowModal;
  end else begin
    TaskSimulatePropertiesForm.FormShow(TaskSimulatePropertiesForm);
    TaskSimulatePropertiesForm.ModalResult := mrOK;
    TaskSimulatePropertiesForm.btnOKClick(TaskSimulatePropertiesForm);
    TaskSimulatePropertiesForm.FormHide(TaskSimulatePropertiesForm);
  end;


  try

    if TaskSimulatePropertiesForm.ModalResult = mrCancel then begin
    //Cancel

      TaskSimulate.Terminate;

    end else begin
    //OK

      TaskSimulate.OnChange := TaskView.OnTaskChange;

      for i := 0 to TaskSimulatePropertiesForm.lstModels.Items.Count - 1 do begin
        if TaskSimulatePropertiesForm.lstModels.Selected[i] then begin
          TaskSimulate.Models.AddObject('', TaskSimulatePropertiesForm.lstModels.Items.Objects[i]);
        end;
      end;

      if TaskSimulate.Models.Count = 0 then begin
        MessageHelpDlg('No model available for simulation', mtError, [mbOK], 0, 'Error:No_Model');
        TaskSimulate.Terminate;
        Exit;
      end;

      if TaskSimulatePropertiesForm.DataMode = 3{AppendCurrent} then begin
        SimData := ViewportData.CurrData;
      end else begin
        // the user may have switched models since determination of simdata --
        // use first model's simdata if it exists, otherwise keep previous choice
        // (most likely prSimData)
        if TQUB_QVO_QVM(TaskSimulate.Models.Objects[0]).SimData <> nil then begin
          SimData := TQUB_QVO_QVD(TQUB_QVO_QVM(TaskSimulate.Models.Objects[0]).SimData);
        end;
      end;

      if (SimData = nil) or (TaskSimulatePropertiesForm.DataMode = 0) then begin
        with ViewportData.NewPropertiesForm do begin
          FName := ChangeFileExt(TQUB_QVO_QVM(TaskSimulate.Models.Objects[0]).FileName, FileTypeToExt(FileFormat));
          if FileExists(FName) then FName := '';
        end;
        if ViewportData.New(FName, False{Silent}) then begin
          TaskSimulate.Terminate;
          Exit;
        end;
        SimData := ViewportData.CurrData;
      end;


      if SimData.LockedAny then begin
        MessageHelpDlg('Data file is in use', mtError, [mbOK], 0, 'Error:No_Data');
        TaskSimulate.Terminate;
        Exit;
      end;
      {
      if (TaskSimulatePropertiesForm.Stimulation = SIM_STIM_SCRIPT) and (SimData.DataChannelCount <= 1) then begin
        MessageHelpDlg('No data channel(s) available for stimuli', mtError, [mbOK], 0, 'Error:No_Channels_For_Stimuli');
        TaskSimulate.Terminate;
        Exit;
      end;
      }
      if TaskSimulatePropertiesForm.Stimulation = SIM_STIM_FILE then begin
        if TaskSimulatePropertiesForm.StimData = SimData then begin
          MessageHelpDlg('The response and the stimulus files must be different', mtError, [mbOK], 0, 'Error:Response_and_stimulus_must_be_different');
          TaskSimulate.Terminate;
          Exit;
        end;
        if (TaskSimulatePropertiesForm.StimData = nil) and (TaskSimulatePropertiesForm.StimProto = nil) then begin
          TaskSimulate.Terminate;
          Exit;
          // already warned by sim properties
        end;
        if ((TaskSimulatePropertiesForm.StimData <> nil)
             and ((TaskSimulatePropertiesForm.StimData.DataChannelCount + 1) > SimData.DataChannelCount))
        or ((TaskSimulatePropertiesForm.StimProto <> nil)
             and ((TaskSimulatePropertiesForm.StimProto.Protocol.ChannelCount + 1) > SimData.DataChannelCount)) then begin
          MessageHelpDlg('Not enough data channels available for stimuli', mtError, [mbOK], 0, 'Error:Not_Enough_Channels');
          TaskSimulate.Terminate;
          Exit;
        end;
      end;

      if TaskSimulatePropertiesForm.DataMode = 1{Replace} then begin
      //Replace Data

        if SimData.Clear then begin
          TaskSimulate.Terminate;
          Exit;
        end;

      end;

      if TaskSimulatePropertiesForm.Stimulation = SIM_STIM_FILE then begin
        TaskSimulate.StimData  := TaskSimulatePropertiesForm.StimData;
        TaskSimulate.StimProto := TaskSimulatePropertiesForm.StimProto;
      end;
      TaskSimulate.Data            := SimData;

      TaskSimulate.ModelParameters := ViewportModel.DefExpCond;

      Stored                       := TQUB_Task_Simulate_Stored.Create(TaskSimulate);
      for i := 0 to TaskSimulate.Models.Count - 1 do begin
        TQUB_QVO_QVM(TaskSimulate.Models.Objects[i]).SimData   := SimData;
        TQUB_QVO_QVM(TaskSimulate.Models.Objects[i]).SimStored := Stored;
      end;
      prSimData                    := SimData;

      //Bring data in view/focus
      ViewportData.ListObjects.SelQVO(SimData);

    end;

  finally

    //TaskSimulatePropertiesForm.Kinetic := nil;
    TaskSimulate.Resume;

  end;

end;


procedure TQUB_ActionSelector_Simulation.prSimLive(OnActionDone :TNotifyEvent);
begin
  if prTaskSimLive = nil then begin
    prTaskSimLive := TQUB_Task_SimLive.Create(True);
    prTaskSimLive.OnTerminate := prSimLiveDone;
    prTaskSimLive.Resume;
    TSpeedButton(prButtons[IndexOf('Live Sim')]).GroupIndex := 1;
    TSpeedButton(prButtons[IndexOf('Live Sim')]).AllowAllUp := True;
    TSpeedButton(prButtons[IndexOf('Live Sim')]).Down := True;
  end else begin
    StopSimLive;
  end;
end;

procedure TQUB_ActionSelector_Simulation.prSimLiveDone(Sender :TObject);
begin
  prTaskSimLive := nil;
  TSpeedButton(prButtons[IndexOf('Live Sim')]).Down := False;
end;


procedure TQUB_ActionSelector_Simulation.StopSimLive;
begin
  if prTaskSimLive <> nil then prTaskSimLive.Stop;
end;


procedure TQUB_ActionSelector_Simulation.prLClickStim(OnActionDone :TNotifyEvent);
begin
  prDoStim(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Simulation.prRClickStim(OnActionDone :TNotifyEvent);
begin
  prDoStim(OnActionDone, True);
end;

function ScalingForMaxAmp(Model :TQUB_QVO_QVM; DataBytes :Integer; ChannelCount :Integer=1): Double;
var
  MaxSC         :Double;
  i             :Integer;
begin
  MaxSC         := 0.0;
  for i := 0 to Model.ModelObject.States.StateCount - 1 do begin
    MaxSC       := Max(MaxSC, Abs(Model.ModelObject.States[i].Amp)
                            + 3*Abs(Model.ModelObject.States[i].Std));
  end;
  MaxSC         := MaxSC * 2.0; // pad for noise

  if ChannelCount <= 0 then ChannelCount := Model.ModelObject.ChannelCount;
  // scale so (2^(8*databytes-1)-1) :: MaxSingle (V = pA)
  Result        :=   (Power(2.0, 8*DataBytes-1) - 1.0)
                   / (ChannelCount*MaxSC);
  Result := Result / 2; // head room
  if Result = 0.0 then Result := 1.0; // or at least no zero
end;

function ScalingForMaxCond(Model :TQUB_QVO_QVM; DataBytes :Integer; ChannelCount :Integer=0): Double;
var
  MaxSC         :Double;
  i             :Integer;
begin
  MaxSC         := 0.0;
  for i := 0 to Model.ModelObject.States.StateCount - 1 do begin
    MaxSC       := Max(MaxSC, Abs(Model.ModelObject.States[i].Cond)
                            + 2*Abs(Model.ModelObject.States[i].CondStd)
                            + Model.ModelObject.Amp[0]);
  end;
  MaxSC         := MaxSC * 2.0; // pad for noise

  if ChannelCount <= 0 then ChannelCount := Model.ModelObject.ChannelCount;
  // scale so (2^(8*databytes-1)-1) :: MaxSingle (V = pA)
  Result        :=   (Power(2.0, 8*DataBytes-1) - 1.0)
                   / (ChannelCount*MaxSC);
  Result := Result / 2; // head room
  if Result = 0.0 then Result := 1.0; // or at least no zero
end;

procedure TQUB_ActionSelector_Simulation.prDoStim(OnActionDone :TNotifyEvent; Properties :Boolean);
var
  TaskSimulate  :TQUB_Task_Simulate;
  Stored        :IQUB_Task_Simulate_Stored;
  Model         :TQUB_QVO_QVM;
  SimData       :TQUB_QVO_QVD;
  i             :Integer;
  FName         :String;
  ParamV        :TQUB_ExpCond_PParameter;
  MaxV          :Double;
begin
  StopSimLive;
  while prTaskSimLive <> nil do Application.ProcessMessages;

  //Task
  TaskSimulate := TQUB_Task_Simulate.Create(True{Suspended});
  TaskSimulate.OnTerminate := OnActionDone;

  try
    Model         := ViewportModel.CurrModel;
    if Model = nil then begin
      MessageHelpDlg('No model available for simulation', mtError, [mbOK], 0, 'Error:No_Model');
      TaskSimulate.Terminate;
      Exit;
    end;
    TaskSimulate.Models.AddObject('', Model);
    SimData      := TQUB_QVO_QVD(Model.SimData); // replace/append into last-simmed-this-model

    TaskStimSimPropertiesForm.NewData := Properties or (SimData = nil);
    if SimData <> nil then begin
      TaskStimSimPropertiesForm.Sampling := 1e-3 / SimData.Sampling;
    end else if TaskStimSimPropertiesForm.NewData then begin
      TaskStimSimPropertiesForm.Sampling := ProfileMgr['Data'].Active['Sampling'].Data[0];
    end;
    TaskStimSimPropertiesForm.LastSimData := SimData;

    if TaskStimSimPropertiesForm.NewData then begin
      TaskStimSimPropertiesForm.ShowModal;
    end else begin
      TaskStimSimPropertiesForm.FormShow(TaskStimSimPropertiesForm);
      TaskStimSimPropertiesForm.ModalResult := mrOK;
      TaskStimSimPropertiesForm.btnOKClick(TaskStimSimPropertiesForm);
      TaskStimSimPropertiesForm.FormHide(TaskStimSimPropertiesForm);
    end;

    TaskSimulatePropertiesForm.OnGetProperties := TaskSimulate.OnGetProperties;
    TaskSimulatePropertiesForm.OnPutProperties := TaskSimulate.OnPutProperties;
    TaskSimulatePropertiesForm.FormShow(TaskSimulatePropertiesForm);
    TaskSimulatePropertiesForm.ModalResult := mrOK;
    TaskSimulatePropertiesForm.btnOKClick(nil);
    TaskSimulatePropertiesForm.FormHide(TaskSimulatePropertiesForm);

    if (TaskStimSimPropertiesForm.ModalResult = mrCancel)
    or not TaskStimSimPropertiesForm.Ready then begin
    //Cancel

      TaskSimulate.Terminate;

    end else begin
    //OK

      if TaskStimSimPropertiesForm.NewData or (SimData = nil) or SimData.LockedAny then begin
        with ViewportData.NewPropertiesForm do begin
          FName := ChangeFileExt(TQUB_QVO_QVM(TaskSimulate.Models.Objects[0]).FileName, FileTypeToExt(FileFormat));
          if FileExists(FName) then FName := '';
          FileFormat     := QFS_QDF_TYPE_QDF;
          ADChannelCount := TaskStimSimPropertiesForm.StimCount + 1;
          Sampling       := TaskStimSimPropertiesForm.Sampling;
          ADDataSize     := 2;
        end;
        if ViewportData.New(FName, True{Silent}) then begin
          TaskSimulate.Terminate;
          Exit;
        end;
        SimData := ViewportData.CurrData;
      end else begin
        if SimData.Clear then begin
          TaskSimulate.Terminate;
          Exit;
        end;
      end;

      if Model.IeqFv then begin
        ParamV := ViewportModel.DefExpCond.ParameterByName['Voltage'];
        if ParamV <> nil then begin
          MaxV := ParamV.Value;
          if ParamV.Channel > 0 then begin
            MaxV := 10000; // gonna exceed 10V?
          end;
        end else begin
          MaxV := 1.0;
        end;
        SimData.DataChannelScaling[0] := (ScalingForMaxCond(Model, SimData.DataObject.ADDataSize) / SimData.DataObject.Scaling) / ((MaxV - Model.ModelObject.Vrev) * 1e-3);
      end else begin
        SimData.DataChannelScaling[0] := ScalingForMaxAmp(Model, SimData.DataObject.ADDataSize) / SimData.DataObject.Scaling;
      end;
      SimData.DataChannelOffset[0]  := 0.0;

      TaskSimulate.OnChange := TaskView.OnTaskChange;

      if SimData.LockedAny then begin
        MessageHelpDlg('Data file is in use', mtError, [mbOK], 0, 'Error:No_Data');
        TaskSimulate.Terminate;
        Exit;
      end;

      if SimData = TaskStimSimPropertiesForm.StimData then begin
        MessageHelpDlg('The response and the stimulus files must be different', mtError, [mbOK], 0, 'Error:Response_and_stimulus_must_be_different');
        TaskSimulate.Terminate;
        Exit;
      end;

      TaskSimulate.StimData        := TaskStimSimPropertiesForm.StimData;
      TaskSimulate.StimProto       := TaskStimSimPropertiesForm.StimProto;
      TaskSimulate.Data            := SimData;
      TaskSimulate.ModelParameters := ViewportModel.DefExpCond;

      //Setup Properties:
      //  defaults from the sim dialog
      TaskSimulatePropertiesForm.PropertiesToParams;
      TaskSimulate.Properties := TaskSimulatePropertiesForm.PropertiesRec;
      with TaskSimulate.Properties do begin
      //  unconditional overrides
        Stimulation                     := SIM_STIM_FILE;
        Randomize                       := True;
        Staircase                       := False;
        BaselineFluct                   := False;
        BaselineDrift                   := False;
        BaselineJumps                   := False;
        BaselinePeriodic                := False;
        BaselineLGM                     := False;
        MakeMovie                       := False;
        IDLOutput                       := Integer(OUT_CLASS);
      //  conditional overrides
        IdealizedData                   := Model.ModelObject.ChannelCount = 1;
        if Model.ModelObject.ChannelCount > 200 then begin
          Response                      := SIM_RESPONSE_DETERMINISTIC;
        end else begin
          Response                      := SIM_RESPONSE_STOCHASTIC;
        end;
      end;

      Stored                       := TQUB_Task_Simulate_Stored.Create(TaskSimulate);
      for i := 0 to TaskSimulate.Models.Count - 1 do begin
        TQUB_QVO_QVM(TaskSimulate.Models.Objects[i]).SimData   := SimData;
        TQUB_QVO_QVM(TaskSimulate.Models.Objects[i]).SimStored := Stored;
      end;
      prSimData                    := SimData;

      //Bring data in view/focus
      ViewportData.ListObjects.SelQVO(SimData);

    end;  // OK

  finally

    TaskSimulate.Resume;

  end;

end;


procedure TQUB_ActionSelector_Simulation.prLClickSimDoseResponse(OnActionDone :TNotifyEvent);
begin
  prDoSimDoseResponse(OnActionDone, True);
end;

procedure TQUB_ActionSelector_Simulation.prRClickSimDoseResponse(OnActionDone :TNotifyEvent);
begin
  prDoSimDoseResponse(OnActionDone, True);
end;

procedure TQUB_ActionSelector_Simulation.prDoSimDoseResponse(OnActionDone :TNotifyEvent; Properties :Boolean);
var
  Task          :TQUB_Task_Sim_DoseResponse;
  FName         :String;
  MaxmV         :Double;
  Levels        :INode;
  i             :Integer;
begin
  if (ViewportModel.CurrModel = nil) or ViewportModel.CurrModel.LockedAny then begin
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  StopSimLive;
  while prTaskSimLive <> nil do Application.ProcessMessages;

  //Task
  Task             := TQUB_Task_Sim_DoseResponse.Create(True{Suspended});
  Task.OnTerminate := OnActionDone;

  TaskSimDoseResponsePropertiesForm.OnGetProperties := Task.OnGetProperties;
  TaskSimDoseResponsePropertiesForm.OnPutProperties := Task.OnPutProperties;

  if Properties then begin
    TaskSimDoseResponsePropertiesForm.ShowModal;
  end else begin
    TaskSimDoseResponsePropertiesForm.FormShow(TaskSimDoseResponsePropertiesForm);
    TaskSimDoseResponsePropertiesForm.ModalResult := mrOK;
    TaskSimDoseResponsePropertiesForm.btnOKClick(TaskSimDoseResponsePropertiesForm);
    TaskSimDoseResponsePropertiesForm.FormHide(TaskSimDoseResponsePropertiesForm);
  end;

  try

    if TaskSimDoseResponsePropertiesForm.ModalResult = mrCancel then begin
    //Cancel

      Task.Terminate;

    end else begin
    //OK

      Task.OnChange := TaskView.OnTaskChange;

      with ViewportData.NewPropertiesForm do begin
        FName := GetBaseStampedFileName('DoseResponse', '.qdf');
        if FileExists(FName) then FName := '';
        FileFormat     := QFS_QDF_TYPE_QDF;
        if 0 <> TaskSimDoseResponsePropertiesForm.Properties['Variable2'].Data[0] then begin
          ADChannelCount := 3;
        end else begin
          ADChannelCount := 2;
        end;
        Sampling       := TaskSimDoseResponsePropertiesForm.Properties['Sampling'].Data[0];
        ADDataSize     := 2;
        if ViewportModel.CurrModel.IeqFv then begin
          // MaxmV := max(abs(v[mv] - Task.Model.ModelObject.Vrev)   for v in Properties['PulseLevels'].data)
          MaxmV := 0.0;
          Levels := TaskSimDoseResponsePropertiesForm.Properties['PulseLevels'];
          for i := 0 to Levels.DataCount - 1 do begin
            if (Abs(Levels.Data[i] - ViewportModel.CurrModel.ModelObject.Vrev) > MaxmV) then begin
              MaxmV := Abs(Levels.Data[i] - ViewportModel.CurrModel.ModelObject.Vrev);
            end;
          end;
          Scaling   := (ScalingForMaxCond(ViewportModel.CurrModel, Task.Data.DataObject.ADDataSize) / Task.Data.DataObject.Scaling) / (MaxmV * 1e-3);
        end else begin
          Scaling    := ScalingForMaxAmp(ViewportModel.CurrModel, ADDataSize,
                                         TaskSimDoseResponsePropertiesForm.Properties['ChannelCount'].Data[0]);
        end;
      end;
      if ViewportData.New(FName, True{Silent}) then begin
        Task.Terminate;
        Exit;
      end;
      Task.Data        := ViewportData.CurrData;
      Task.Model       := ViewportModel.CurrModel;
      Task.Data.DataChannelOffset[0]  := 0.0;
    end;

  finally

    Task.Resume;

  end;

end;




//****************************************************************************//
end.
