unit QUB_ActionSelector_Sound;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses

Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls,
QUB_Global,
QUB_Action, QUB_ActionSelector,
QUB_Task_DataWav,
QUB_Task_DataWav_Properties,
QUB_QVO_QVD,
QUB_Node, Menus;


type TQUB_ActionSelector_Sound = class(TQUB_ActionSelector)

protected

  prErrorMsg                   :String;
  prWikiTopic                  :String;

  procedure prLClickPlay(OnActionDone :TNotifyEvent);
  procedure prRClickPlay(OnActionDone :TNotifyEvent);
  procedure prLClickStop(OnActionDone :TNotifyEvent);

  function prCanDataWav :Boolean;
  procedure prLClickIdlWav(OnActionDone :TNotifyEvent);
  procedure prRClickIdlWav(OnActionDone :TNotifyEvent);
  function prDoIdlWav(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
  procedure prLClickDataWav(OnActionDone :TNotifyEvent);
  procedure prRClickDataWav(OnActionDone :TNotifyEvent);
  function prDoDataWav(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;

  procedure prRestoreDefaults; override;

public

  TaskDataWavPropertiesForm    :TQUB_Task_DataWav_PropertiesForm;

  constructor Create(AOwner :TComponent); reintroduce;

end;


implementation


{$R *.DFM}


uses


QUB_Viewport,
QUB_Viewport_Data,
QUB_GUIInterface,
QUB_Main,
QUB_Task,
QUB_Python_Data;


// *****************************************************************************
// self-destructing OnTerminate for Task_IdlWav
// first starts it playing, then calls OnActionDone

type TQUB_ActionSelector_Sound_AfterDataWav = class
protected
  prOnActionDone              :TNotifyEvent;
public
  constructor Create(OnActionDone :TNotifyEvent);
  procedure OnDataWavDone(Sender :TObject);
end;

constructor TQUB_ActionSelector_Sound_AfterDataWav.Create(OnActionDone :TNotifyEvent);
begin
  prOnActionDone := OnActionDone;
end;

procedure TQUB_ActionSelector_Sound_AfterDataWav.OnDataWavDone(Sender :TObject);
begin
  if Sender is TQUB_Task_DataWav then begin
    if not TQUB_Task_DataWav(Sender).Error then begin
      if TQUB_Task_DataWav(Sender).Data <> nil then begin
        TQUB_QVO_QVD(TQUB_Task_DataWav(Sender).Data).PlaySelection(-1);
      end;
    end;
  end;

  if @prOnActionDone <> nil then prOnActionDone(Sender);
end;


// *****************************************************************************
constructor TQUB_ActionSelector_Sound.Create(AOwner :TComponent);
begin
  inherited CreateNamed(AOwner, 'Sound');

  prAddAction( TQUB_Action.Create('Play', prLClickPlay, prRClickPlay, nil, 'Play data selection'));
                                                    
  prAddAction( TQUB_Action.Create('Stop', prLClickStop, nil, nil, 'Stop playing data selection'));

  TaskDataWavPropertiesForm := TQUB_Task_DataWav_PropertiesForm.Create(Application);
  prAddAction( TQUB_Action.Create('Idl -> WAV', prLClickIdlWav, prRClickIdlWav, TaskDataWavPropertiesForm.SetGetProperties,
                                  'Convert idealized data to a sound file', 'iWAV'));

  prAddAction( TQUB_Action.Create('Data -> WAV', prLClickDataWav, prRClickDataWav, TaskDataWavPropertiesForm.SetGetProperties,
                                  'Convert sampled data to a sound file','dWAV'));

  RestoreDefaults;
end;

procedure TQUB_ActionSelector_Sound.prRestoreDefaults;
var
  Vis                 :TStringList;
begin
  Vis                 := ActionByName['Play'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Preprocessing');

  Vis                 := ActionByName['Stop'].VisLayouts;
  Vis.Clear;
  Vis.Add('Acquisition');
  Vis.Add('Preprocessing');

  Vis                 := ActionByName['Idl -> WAV'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
  Vis.Add('Modeling');

  Vis                 := ActionByName['Data -> WAV'].VisLayouts;
  Vis.Clear;
  Vis.Add('Preprocessing');
  Vis.Add('Modeling');
end;

procedure TQUB_ActionSelector_Sound.prLClickPlay(OnActionDone :TNotifyEvent);
begin
  ViewportData.PlaySelection;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

//RClick is used to play the current channel. 
//todo properties ?! LM.
procedure TQUB_ActionSelector_Sound.prRClickPlay(OnActionDone :TNotifyEvent);
begin
  if ViewportData.CurrData <> nil then ViewportData.CurrData.PlaySelection(ViewportData.CurrData.ActiveChannel);
  if @OnActionDone <> nil then OnActionDone(nil);
end;

procedure TQUB_ActionSelector_Sound.prLClickStop(OnActionDone :TNotifyEvent);
begin
  ViewportData.StopSelection;
  if @OnActionDone <> nil then OnActionDone(nil);
end;

function TQUB_ActionSelector_Sound.prCanDataWav: Boolean;
begin
  Result := ViewportData.CurrData <> nil;
  if Result then begin
    if Result then Result := not ViewportData.CurrData.LockedReadWrite;
    if Result then Result := ViewportData.CurrData.DataPointCount <> 0;
  end;
  if not Result then begin
    prErrorMsg := 'Cannot make sound from data. The data might be in use or empty.';
    prWikiTopic := 'Error:No_Data';
  end;
  if not Result then Exit;
end;

procedure TQUB_ActionSelector_Sound.prLClickIdlWav(OnActionDone :TNotifyEvent);
begin
  prDoIdlWav(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Sound.prRClickIdlWav(OnActionDone :TNotifyEvent);
begin
  prDoIdlWav(OnActionDone, True);
end;

function TQUB_ActionSelector_Sound.prDoIdlWav(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  TaskIdlWav               :TQUB_Task_IdealWav;
  AfterDataWav             :TQUB_ActionSelector_Sound_AfterDataWav;
begin

  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_IdealWav)) then begin
    Result := True;
    MessageHelpDlg('Only one Idl->Wav task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Result := not prCanDataWav;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskIdlWav := TQUB_Task_IdealWav.Create(True{Suspended});

  AfterDataWav             := TQUB_ActionSelector_Sound_AfterDataWav.Create(OnActionDone);
  TaskIdlWav.OnTerminate   := AfterDataWav.OnDataWavDone;

  TaskDataWavPropertiesForm.OnGetProperties := TaskIdlWav.OnGetProperties;
  TaskDataWavPropertiesForm.OnPutProperties := TaskIdlWav.OnPutProperties;

  TaskDataWavPropertiesForm.btnOK.Enabled   := not Result;

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskDataWavPropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskDataWavPropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  if (Properties) or (Result) then begin
    TaskDataWavPropertiesForm.ShowModal;
  end else begin
    TaskDataWavPropertiesForm.FormShow(TaskDataWavPropertiesForm);
    TaskDataWavPropertiesForm.ModalResult := mrOK;
    TaskDataWavPropertiesForm.btnOKClick(TaskDataWavPropertiesForm);
  end;

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and (TaskDataWavPropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskDataWavPropertiesForm.DataSource;
  end;

  if TaskDataWavPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskIdlWav.Terminate;
  end else if TaskDataWavPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskIdlWav.Terminate;
  end else begin
  //OK
    TaskIdlWav.OnChange := TaskView.OnTaskChange;
  end;
  TaskIdlWav.Resume;
end;

procedure TQUB_ActionSelector_Sound.prLClickDataWav(OnActionDone :TNotifyEvent);
begin
  prDoDataWav(OnActionDone, False);
end;

procedure TQUB_ActionSelector_Sound.prRClickDataWav(OnActionDone :TNotifyEvent);
begin
  prDoDataWav(OnActionDone, True);
end;

function TQUB_ActionSelector_Sound.prDoDataWav(OnActionDone :TNotifyEvent; Properties :Boolean): Boolean;
var
  TaskDataWav              :TQUB_Task_SampleDataWav;
  AfterDataWav             :TQUB_ActionSelector_Sound_AfterDataWav;
begin
  if (TaskView <> nil) and (TaskView.HasTaskOfType(TQUB_Task_SampleDataWav)) then begin
    Result := True;
    MessageHelpDlg('Only one Data->Wav task can run at a time', mtError, [mbOK], 0, 'Error:Only_One_Task');
    if Assigned(OnActionDone) then OnActionDone(nil);
    Exit;
  end;

  Result := not prCanDataWav;

  if Result then begin
    if not Properties then begin
      MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
      if Assigned(OnActionDone) then OnActionDone(nil);
      Exit;
    end else begin
      MessageHelpDlg(prErrorMsg, mtWarning, [mbOK], 0, prWikiTopic);
      prWikiTopic := '';
    end;
  end;

  TaskDataWav := TQUB_Task_SampleDataWav.Create(True{Suspended});

  AfterDataWav              := TQUB_ActionSelector_Sound_AfterDataWav.Create(OnActionDone);
  TaskDataWav.OnTerminate   := AfterDataWav.OnDataWavDone;

  TaskDataWavPropertiesForm.OnGetProperties := TaskDataWav.OnGetProperties;
  TaskDataWavPropertiesForm.OnPutProperties := TaskDataWav.OnPutProperties;

  TaskDataWavPropertiesForm.btnOK.Enabled    := not Result;

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    TaskDataWavPropertiesForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else begin
    TaskDataWavPropertiesForm.DataSource := QUB_MainForm.DataSource;
  end;

  if (Properties) or (Result) then begin
    TaskDataWavPropertiesForm.ShowModal;
  end else begin
    TaskDataWavPropertiesForm.FormShow(TaskDataWavPropertiesForm);
    TaskDataWavPropertiesForm.ModalResult := mrOK;
    TaskDataWavPropertiesForm.btnOKClick(TaskDataWavPropertiesForm);
  end;

  if not ((QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and (TaskDataWavPropertiesForm.DataSource = DATA_SOURCE_WHOLEFILE)) then begin
    QUB_MainForm.DataSource := TaskDataWavPropertiesForm.DataSource;
  end;

  if TaskDataWavPropertiesForm.ModalResult = mrCancel then begin
  //Cancel
    TaskDataWav.Terminate;
  end else if TaskDataWavPropertiesForm.ModalResult = mrNo then begin
  //Reset
    TaskDataWav.Terminate;
  end else begin
  //OK
    TaskDataWav.OnChange          := TaskView.OnTaskChange;
  end;
  TaskDataWav.Resume;
end;

//****************************************************************************//
end.
