{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
{ 03-15-2004 JB
  Grid style form for selecting which buttons are available in each layout.

}

unit QUB_Action_Properties;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Math,
  StdCtrls, Grids;

type
  TfrmActionProperties = class(TForm)
    grdSelections: TStringGrid;
    Label1: TLabel;
    btnSave: TButton;
    btnCancel: TButton;
    btnRestoreDefaults: TButton;
    btnApply: TButton;
    btnHelp: TButton;
    procedure OnShow( Sender: TObject);
    procedure ClickGrid( Sender: TObject);
    procedure MouseDownGrid( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure MouseUpGrid( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure OnHide( Sender: TObject);
    procedure DrawCellGrid( Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
    procedure ClickSave( Sender: TObject);
    procedure ClickApply( Sender: TObject);
    procedure ClickRestoreDefaults( Sender: TObject);
    procedure btnHelpClick(Sender: TObject);
    procedure grdSelectionsKeyPress(Sender: TObject; var Key: Char);
  private
    { Private declarations }
    m_iCurrentLayout : Integer;
    procedure SaveChanges();
    procedure LoadSettings();
  public
    { Public declarations }
  end;

implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Main,
QUB_Action,
QUB_ActionSelector,
QUB_Global;


var


  //When mouse wheel is used in grid, causes click message for every cell passed.
  //track left mouse button down to prevent wheel from faking clicks.
  g_lMouseDown : Boolean;

{------------------------------------------------------------------------------}
//-- Main refresh - Build grid and load list of options from QUB_ActionSelector objects.
procedure TfrmActionProperties.LoadSettings();
var
  iGrp,iRow,iCol, iItem, iRows : Integer;
  Actions : TQUB_ActionSelector;
  Action : TQUB_Action;
  cLayout : String;
begin
  g_lMouseDown := False;

  //----- set # of cols, more than enough rows
  grdSelections.ColCount := InterfaceList.Count + 3;
  grdSelections.ColWidths[0] := 75;
  grdSelections.ColWidths[1] := 90;
  grdSelections.ColWidths[2] := 250;
  iRows := 100;
  grdSelections.RowCount := iRows;

  //----- Row 0 column headers ( 3 fixed, then layout names )
  m_iCurrentLayout := -1;
  grdSelections.Cells[0,0] := 'Group';
  grdSelections.Cells[1,0] := 'Name';
  grdSelections.Cells[2,0] := 'Description';
  for iCol := 0 to InterfaceList.Count-1 do begin
    cLayout := InterfaceList[iCol].Name;
    if cLayout = 'Simulation' then cLayout := 'Sim';
    if cLayout = 'Acquisition' then cLayout := 'Acq';
    if cLayout = 'Preprocessing' then cLayout := 'Pre';
    if cLayout = 'Macroscopic' then cLayout := 'Mac';
    if cLayout = 'Modeling' then cLayout := 'Mod';
    if cLayout <> InterfaceList[iCol].Name then grdSelections.ColWidths[iCol+3] := 25;
    grdSelections.Cells[iCol+3,0] := cLayout;

    //-- remember which layout is current
    if InterfaceList[iCol].ID = QUB_MainForm.InterfaceIndex then m_iCurrentLayout := iCol;
  end;

  //----- For each list, for each item - set grid text
  iRow:=1;
  for iGrp := 0 to ActionsList.Count-1 do begin
    Actions := TQUB_ActionSelector(ActionsList[iGrp]);
    //-- Col 0 - Group Name
    grdSelections.Cells[0,iRow] := Actions.Name;

    for iItem := 0 to Actions.ActionCount -1 do begin
      Action := Actions.Actions[iItem];
      //-- Cols 1,2 = Name, Description
      grdSelections.Cells[1,iRow] := Action.Name;
      grdSelections.Cells[2,iRow] := Action.Description;

      //-- Cols 3+ : matrix of True / False ( button available in layout ? )
      for iCol := 0 to InterfaceList.Count-1 do begin
        cLayout := InterfaceList[iCol].Name;
        grdSelections.Cells[iCol+3, iRow] := Choice(Actions.ActionVisible[iItem, cLayout],'.','');
      end;

      //-- Next row, increase grid if necc.
      Inc(iRow);
      if iRow >= iRows Then begin
        iRows := iRows + 25;
        grdSelections.RowCount := iRows;
      end;
    end;
  end;
  //-- Reduce grid to size actually used.
  grdSelections.RowCount := iRow;

end;

//----- Update Actions.ActionVisible based on Grid settings.
procedure TfrmActionProperties.SaveChanges();
var
  iGrp,iRow,iCol, iItem : Integer;
  Actions : TQUB_ActionSelector;
  cLayout : String;
begin
  //----- For each list, for each item - set grid text
  iRow:=1;
  for iGrp := 0 to ActionsList.Count-1 do begin         // for each button group
    Actions := TQUB_ActionSelector(ActionsList[iGrp]);
    for iItem := 0 to Actions.ActionCount -1 do begin   // for each action button
      for iCol := 0 to InterfaceList.Count-1 do begin   // for each layout
        cLayout := InterfaceList[iCol].Name;
        if (grdSelections.Cells[iCol+3, iRow]='.') <> Actions.ActionVisible[iItem, cLayout]
            then Actions.ActionVisible[iItem, cLayout] := not Actions.ActionVisible[iItem, cLayout];
      end;
      Inc(iRow);
    end;
  end;
end;

//----- Toggle cells between '' and '.' when double Clicked.
procedure TfrmActionProperties.ClickGrid(Sender: TObject);
var
  c : String;
  iRow , iCol : Integer ;
begin
    iCol := grdSelections.Col;
    iRow := grdSelections.Row;
    if g_lMouseDown and (iCol >= 3) then begin
      c := grdSelections.Cells[iCol, iRow ];
      if c = '' Then c := '.' else c := '' ;
      grdSelections.Cells[iCol, iRow] := c ;
    end;
end;

procedure TfrmActionProperties.grdSelectionsKeyPress(Sender: TObject;
  var Key: Char);
var
  c : String;
  iRow , iCol : Integer ;
begin
  if Key = ' ' then begin
    // do like click
    iCol := grdSelections.Col;
    iRow := grdSelections.Row;
    if iCol >= 3 then begin
      c := grdSelections.Cells[iCol, iRow ];
      if c = '' Then c := '.' else c := '' ;
      grdSelections.Cells[iCol, iRow] := c ;
    end;
    Key := #0;
  end;
end;

//----- track left mouse button down to prevent wheel from faking clicks
procedure TfrmActionProperties.MouseDownGrid(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft Then g_lMouseDown := True;
end;

procedure TfrmActionProperties.MouseUpGrid(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft Then g_lMouseDown := False;
end;

//----- Intercept after grid cell is drawn and redraw as graphic.
procedure TfrmActionProperties.DrawCellGrid(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  iCtr : Integer;
begin
  // For layout columns ( col 3+ ) draw the contents manually.
  if (ARow>0) and (ACol>=3) Then begin
    if grdSelections.Cells[ACol, ARow]='.' then begin
      // Draw a Gray box - dark for current layout, light for non-current.
      InflateRect(Rect,-1,-1);
      iCtr := (Rect.Right+Rect.Left) Div 2;
      if m_iCurrentLayout = ACol-3 Then begin
        grdSelections.Canvas.Brush.Color:=clDkGray;
        grdSelections.Canvas.Pen.Color:=clDkGray;
      end else begin
        grdSelections.Canvas.Brush.Color:=clLtGray;
        grdSelections.Canvas.Pen.Color:=clLtGray;
      end;
      grdSelections.Canvas.Rectangle(Rect);
      // Draw  a Black X (i.e. two diagonal lines)
      grdSelections.Canvas.Pen.Color:=clBlack;
      grdSelections.Canvas.Pen.Width:=2;
      grdSelections.Canvas.MoveTo(Max(Rect.Left,iCtr-6), Rect.Top);
      grdSelections.Canvas.LineTo(Min(Rect.Right-1,iCtr+6), Rect.Bottom-1); //-1 for 2 pixel pen
      grdSelections.Canvas.MoveTo(Max(Rect.Left,iCtr-6), Rect.Bottom-1);  //-1 for 2 pixel pen
      grdSelections.Canvas.LineTo(Min(Rect.Right-1,iCtr+6), Rect.Top)
    end else begin
      // Draw a white rectangle to cover any contents
      InflateRect(Rect,-1,-1);
      grdSelections.Canvas.Pen.Color:=clWhite;
      grdSelections.Canvas.Brush.Color:=clWhite;
      grdSelections.Canvas.Rectangle(Rect);
    end;
  end;
end;

//---------- Events
procedure TfrmActionProperties.OnHide(Sender: TObject);
begin
  g_lMouseDown := False;
end;

procedure TfrmActionProperties.OnShow(Sender: TObject);
begin
  LoadSettings();
end;

procedure TfrmActionProperties.ClickSave(Sender: TObject);
begin
  SaveChanges();
end;

procedure TfrmActionProperties.ClickApply(Sender: TObject);
begin
  SaveChanges();
end;

procedure TfrmActionProperties.ClickRestoreDefaults(Sender: TObject);
var
  iList : Integer;
begin
  for iList := 0 to ActionsList.Count - 1 do
    ActionsList[iList].RestoreDefaults;
  LoadSettings();
end;

procedure TfrmActionProperties.btnHelpClick(Sender: TObject);
begin
  ShowWiki('Button Settings', Handle);
end;

end.
