unit QUB_AdaptiveResample;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

//TO DO: LM
//VERY IMPORTANT!!!
//A big problem here is that the variance information is lost. 
//Solution: replace Avg with a randomly chosen point, 
//or randomly generated point with Avg and Std,
//or provide info about how many points were averaged, 
//so Mac knows how to adjust the expected variance. 

//OK this is done -=- Chris 12-22-2005
// look at AdaptiveResample2() which is now used by Mac

{
Chris 12-2005

For Mac

Breaks data into segments with std < MaxStd
returning a list of segment midpoints (sample indices)
and overwriting all data points with segment averages.
Recursively breaks data in half until the condition is reached.

}

interface

uses

{$ifdef QUB}
QUB_QFS_QIF_Segments,
{$endif}
SysUtils,
Alloc, AllocTypes;

function AdaptiveResample(Data :TPAInteger; DataCount :Integer;
                          MaxStd :Double; var ResampleCount :Integer)
            : TPAInteger;

type

{$ifdef QUB}
{$else}
TQUB_QFS_QIF_NClass                            = record
  Cls                                          :Integer;
  N                                            :Integer;
end;
TQUB_QFS_QIF_PNClass                           = ^TQUB_QFS_QIF_NClass;
TQUB_QFS_QIF_ANClass                           = Array[0..(MaxInt div (SizeOf(TQUB_QFS_QIF_NClass)*2))] of TQUB_QFS_QIF_NClass;
TQUB_QFS_QIF_PANClass                          = ^TQUB_QFS_QIF_ANClass;
{$endif}

TQUB_AdResInterval = record
  First, Last       :Integer;
  Mean, Std         :Single;
end;
TQUB_PAdResInterval         = ^TQUB_AdResInterval;

TQUB_AAdResInterval         = Array[0..(MaxInt div (SizeOf(TQUB_AdResInterval)*2))] of TQUB_AdResInterval;
TQUB_PAAdResInterval        = ^TQUB_AAdResInterval;


function AdaptiveResample2(Data :TPAInteger; DataCount :Integer;
                           ECIdl :TQUB_QFS_QIF_PANClass; Count :Integer;
                           MaxStd :Double;
                           var ResampleCount :Integer;
                           var Intervals     :TQUB_PAAdResInterval;
                           var UseSamples    :TPAInteger)
            : Boolean;

{
The following modified 9-2006 for Double data, for the volume chamber / new-style curve fitting

Also unlike the original, we don't overwrite the data with the adaptive averages
}

type

TQUB_AdResInterval_dbl = record
  First, Last       :Integer;
  Mean, Std         :Double;
end;
TQUB_PAdResInterval_dbl         = ^TQUB_AdResInterval_dbl;

TQUB_AAdResInterval_dbl         = Array[0..(MaxInt div (SizeOf(TQUB_AdResInterval_dbl)*2))] of TQUB_AdResInterval_dbl;
TQUB_PAAdResInterval_dbl        = ^TQUB_AAdResInterval_dbl;


function AdaptiveResample_dbl(Data :TPADouble; DataCount :Integer;
                          MaxStd :Double;
                          var ResampleCount :Integer;
                          var Intervals     :TQUB_PAAdResInterval_dbl;
                          var UseSamples    :TPAInteger)
            : Boolean;

// Intervals has the complete output info.
// you must FreeMem(Intervals)

// UseSamples has the index (First+Last)/2 for each interval.
// you must FreeMem(UseSamples)
// perhaps not as useful if the data is not uniformly sampled.

// below is a variant which picks the closest data point for UseSamples
// rather than the midpoint

function AdaptiveResampleClosest_dbl(Data :TPADouble; DataCount :Integer;
                          MaxStd :Double;
                          var ResampleCount :Integer;
                          var Intervals     :TQUB_PAAdResInterval_dbl;
                          var UseSamples    :TPAInteger)
            : Boolean;

//******************************************************************************

implementation

type

TAdResTree = class
public
  First, Last   :Integer;
  Avg, Std      :Double;
  Left, Right   :TAdResTree;

  constructor Create(First, Last :Integer);
  destructor Destroy; override;

  procedure Split;
  procedure SplitBefore(RightFirst :Integer);
end;

TAdResTree_DFS = class
private
  LeavesOnly   :Boolean;
  ToVisit      :Array of TAdResTree;

  procedure Push(Node :TAdResTree);
  function Pop: TAdResTree;
public
  Here         :TAdResTree;

  constructor Create(Root :TAdResTree; LeavesOnly :Boolean);
  function Next: Boolean; // returns false if no more
end;

//******************************************************************************

constructor TAdResTree.Create(First, Last :Integer);
begin
  Self.First := First;
  Self.Last  := Last;
end;

destructor TAdResTree.Destroy;
begin
  if Left <> nil then Left.Free;
  if Right <> nil then Right.Free;
end;

procedure TAdResTree.Split;
var
  Count      :Integer;
begin
  Assert((Left = nil) and (Right = nil), 'TAdResTree.Split: not a leaf node');
  Count      := Last - First + 1;
  SplitBefore(First + (Count div 2));
end;

procedure TAdResTree.SplitBefore(RightFirst :Integer);
begin
  Left       := TAdResTree.Create(First, RightFirst-1);
  Right      := TAdResTree.Create(RightFirst, Last);
end;


constructor TAdResTree_DFS.Create(Root :TAdResTree; LeavesOnly :Boolean);
begin
  Self.LeavesOnly   := LeavesOnly;
  Here              := nil;
  Push(Root);
end;

function TAdResTree_DFS.Next: Boolean;
begin
  Result := 0 <> Length(ToVisit);
  if Result then begin
    Here            := Pop;
    if Here.Left <> nil then begin
      Push(Here.Right);
      Push(Here.Left);
      if LeavesOnly then Result := Next;
    end;
  end;
end;

procedure TAdResTree_DFS.Push(Node :TAdResTree);
var
  i     :Integer;
begin
  i     := Length(ToVisit);
  SetLength(ToVisit, i+1);
  ToVisit[i] := Node;
end;

function TAdResTree_DFS.Pop: TAdResTree;
var
  i      :Integer;
begin
  i      := Length(ToVisit) - 1;
  Result := ToVisit[i];
  SetLength(ToVisit, i);
end;


//******************************************************************************

procedure CalcAvgStd(Node :TAdResTree; Data :TPAInteger);
var
  i, N         :Integer;
  Avg, Std, x  :Double;
begin
  N         := Node.Last - Node.First + 1;
  if N = 1 then begin
    Node.Avg := Data[Node.First];
    Node.Std := 0.0;
    Exit;
  end;
  Avg       := 0;
  for i := Node.First to Node.Last do begin
    Avg     := Avg + Data[i];
  end;
  Avg       := Avg / N;
  Node.Avg  := Avg;
  Std       := 0;
  for i := Node.First to Node.Last do begin
    x       := Data[i] - Avg;
    Std     := Std + x * x;
  end;
  Std       := Sqrt(Std / (N-1));
  Node.Std  := Std;
end;

procedure SplitTree(Tree :TAdResTree; Data :TPAInteger; MaxStd :Double);
begin
  if Tree.Left = nil then begin
    CalcAvgStd(Tree, Data);
    if Tree.Std > MaxStd then begin
      Tree.Split;
    end;
  end;
  if Tree.Left <> nil then begin
    SplitTree(Tree.Left, Data, MaxStd);
    SplitTree(Tree.Right, Data, MaxStd);
  end;
end;

//******************************************************************************

function AdaptiveResample(Data :TPAInteger; DataCount :Integer;
                          MaxStd :Double; var ResampleCount :Integer)
            : TPAInteger;
var
  Tree    :TAdResTree;
  Trav    :TAdResTree_DFS;
  i, x, j :Integer;
begin
  Tree    := TAdResTree.Create(0, DataCount-1);
  SplitTree(Tree, Data, MaxStd);

  i := 0;
  Trav    := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do Inc(i);
  ResampleCount := i;
  Trav.Free;

  Result  := AllocMem(ResampleCount * SizeOf(Integer));
  i       := 0;
  Trav    := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do begin
    Result[i]     := (Trav.Here.First + Trav.Here.Last) div 2;
    x             := Round(Trav.Here.Avg);
    for j := Trav.Here.First to Trav.Here.Last do begin
      Data[j]     := x;
    end;
    Inc(i);
  end;
  Trav.Free;
  Tree.Free;
end;


procedure ECIdlToTree(ECIdl :TQUB_QFS_QIF_PANClass; Count :Integer; Tree :TAdResTree);
var
  Branch :TAdResTree;
  i      :Integer;
begin
  Branch := Tree;
  for i := 1 to Count - 1 do begin
    Branch.SplitBefore(Branch.First + ECIdl[i-1].N);
    Branch := Branch.Right;
  end;
end;


function AdaptiveResample2(Data :TPAInteger; DataCount :Integer;
                           ECIdl :TQUB_QFS_QIF_PANClass; Count :Integer;
                           MaxStd :Double;
                           var ResampleCount :Integer;
                           var Intervals     :TQUB_PAAdResInterval;
                           var UseSamples    :TPAInteger)
            : Boolean;
var
  Tree    :TAdResTree;
  Trav    :TAdResTree_DFS;
  i, x, j :Integer;
begin
  Result  := False;

  Tree    := TAdResTree.Create(0, DataCount-1);
  ECIdlToTree(ECIdl, Count, Tree);
  SplitTree(Tree, Data, MaxStd);

  i := 0;
  Trav    := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do Inc(i);
  ResampleCount := i;
  Trav.Free;

  Intervals  := AllocMem(ResampleCount * SizeOf(TQUB_AdResInterval));
  UseSamples := AllocMem(ResampleCount * SizeOf(Integer));
  i          := 0;
  Trav       := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do begin
    Intervals[i].First := Trav.Here.First;
    Intervals[i].Last  := Trav.Here.Last;
    Intervals[i].Mean  := Trav.Here.Avg;
    Intervals[i].Std   := Trav.Here.Std;
    UseSamples[i]      := (Intervals[i].First + Intervals[i].Last) div 2;

    x             := Round(Intervals[i].Mean);
    for j := Intervals[i].First to Intervals[i].Last do begin
      Data[j]     := x;
    end;
    Inc(i);
  end;
  Trav.Free;
  Tree.Free;
end;



//******************************************************************************

procedure CalcAvgStd_dbl(Node :TAdResTree; Data :TPADouble);
var
  i, N         :Integer;
  Avg, Std, x  :Double;
begin
  N         := Node.Last - Node.First + 1;
  if N = 1 then begin
    Node.Avg := Data[Node.First];
    Node.Std := 0.0;
    Exit;
  end;
  Avg       := 0;
  for i := Node.First to Node.Last do begin
    Avg     := Avg + Data[i];
  end;
  Avg       := Avg / N;
  Node.Avg  := Avg;
  Std       := 0;
  for i := Node.First to Node.Last do begin
    x       := Data[i] - Avg;
    Std     := Std + x * x;
  end;
  Std       := Sqrt(Std / (N-1));
  Node.Std  := Std;
end;

procedure SplitTree_dbl(Tree :TAdResTree; Data :TPADouble; MaxStd :Double);
begin
  if Tree.Left = nil then begin
    CalcAvgStd_dbl(Tree, Data);
    if Tree.Std > MaxStd then begin
      Tree.Split;
    end;
  end;
  if Tree.Left <> nil then begin
    SplitTree_dbl(Tree.Left, Data, MaxStd);
    SplitTree_dbl(Tree.Right, Data, MaxStd);
  end;
end;

//******************************************************************************

function AdaptiveResample_dbl(Data :TPADouble; DataCount :Integer;
                          MaxStd :Double;
                          var ResampleCount :Integer;
                          var Intervals     :TQUB_PAAdResInterval_dbl;
                          var UseSamples    :TPAInteger)
            : Boolean;
var
  Tree    :TAdResTree;
  Trav    :TAdResTree_DFS;
  i       :Integer;
begin
  Result  := False;

  Tree    := TAdResTree.Create(0, DataCount-1);
  SplitTree_dbl(Tree, Data, MaxStd);

  i := 0;
  Trav    := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do Inc(i);
  ResampleCount := i;
  Trav.Free;

  Intervals  := AllocMem(ResampleCount * SizeOf(TQUB_AdResInterval_dbl));
  UseSamples := AllocMem(ResampleCount * SizeOf(Double));
  i          := 0;
  Trav       := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do begin
    Intervals[i].First := Trav.Here.First;
    Intervals[i].Last  := Trav.Here.Last;
    Intervals[i].Mean  := Trav.Here.Avg;
    Intervals[i].Std   := Trav.Here.Std;
    UseSamples[i]      := (Intervals[i].First + Intervals[i].Last) div 2;

    Inc(i);
  end;
  Trav.Free;
  Tree.Free;
end;

//******************************************************************************

function AdaptiveResampleClosest_dbl(Data :TPADouble; DataCount :Integer;
                          MaxStd :Double;
                          var ResampleCount :Integer;
                          var Intervals     :TQUB_PAAdResInterval_dbl;
                          var UseSamples    :TPAInteger)
            : Boolean;
var
  Tree    :TAdResTree;
  Trav    :TAdResTree_DFS;
  i, j    :Integer;
  imin    :Integer;
  dmin, m, d :Double;
begin
  Result  := False;

  Tree    := TAdResTree.Create(0, DataCount-1);
  SplitTree_dbl(Tree, Data, MaxStd);

  i := 0;
  Trav    := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do Inc(i);
  ResampleCount := i;
  Trav.Free;

  Intervals  := AllocMem(ResampleCount * SizeOf(TQUB_AdResInterval_dbl));
  UseSamples := AllocMem(ResampleCount * SizeOf(Double));
  i          := 0;
  Trav       := TAdResTree_DFS.Create(Tree, True);
  while Trav.Next do begin
    Intervals[i].First := Trav.Here.First;
    Intervals[i].Last  := Trav.Here.Last;
    Intervals[i].Mean  := Trav.Here.Avg;
    Intervals[i].Std   := Trav.Here.Std;

    m    := Intervals[i].Mean;
    imin := Intervals[i].First;
    dmin := Abs(m - Data[Intervals[i].First]);
    for j := (Intervals[i].First+1) to Intervals[i].Last do begin
      d  := Abs(m - Data[j]);
      if d < dmin then begin
        dmin := d;
        imin := j;
      end;
    end;
    UseSamples[i]      := imin;

    Inc(i);
  end;
  Trav.Free;
  Tree.Free;
end;


//******************************************************************************
end.
