{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2003

// Imports "segment of baseline nodes" datatype from qubopt,
// then wraps it with multi-segment, thread-safe, qubtree-convertible class QUB_Baseline.

// This stuff immediately precedes QUB_Idealization.  A lot of it served as lessons about what not to do there.

// The essential problem of the baseline is that nodes are sparse, and often inserted between other nodes.
// To apply baseline, you need to find the last node <= some data point,
// then iterate through until the first node >= some data point.
// Additionally, the data's segment bounds may change due to editing (delete, cut, paste, append).

unit QUB_Baseline;


interface



uses

Alloc, AllocTypes, sysutils, Classes, SyncObjs, Math, Windows,
QUB_Global,
QUB_Node,
PythonEngine;



type

TQUB_BaselineSeg_Rec = record
  First         :Integer;
  Last          :Integer;
  // Nodes       :std::map<int, QUB_BaselineSeg_Node*>
end;
TQUB_BaselineSeg_PRec = ^TQUB_BaselineSeg_Rec;

TQUB_BaselineSeg_Node_Rec = record
  Point         :Integer;
  Value         :Integer;
  // Iter       :std::map<int, struct QUB_BaselineSeg_Node*>::iterator;
end;
TQUB_BaselineSeg_Node_PRec = ^TQUB_BaselineSeg_Node_Rec;
TQUB_BaselineSeg_Node_PPRec = ^TQUB_BaselineSeg_Node_PRec;
TQUB_BaselineSeg_Node      = TQUB_BaselineSeg_Node_PRec;

function  QUB_BaselineSeg_Create             (First, Last :Integer): TQUB_BaselineSeg_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_Create@8';
procedure QUB_BaselineSeg_Free( Seg :TQUB_BaselineSeg_PRec ); stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_Free@4';
function  QUB_BaselineSeg_GetNodeCount( Seg :TQUB_BaselineSeg_PRec ): Integer; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetNodeCount@4';
procedure QUB_BaselineSeg_SetFirst( Seg :TQUB_BaselineSeg_PRec; First :Integer ); stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_SetFirst@8';
procedure QUB_BaselineSeg_SetLast( Seg :TQUB_BaselineSeg_PRec; Last :Integer ); stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_SetLast@8';
function  QUB_BaselineSeg_ExciseFrom( Seg :TQUB_BaselineSeg_PRec; First :Integer ): TQUB_BaselineSeg_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_ExciseFrom@8';
function  QUB_BaselineSeg_GetFirstNode( Seg :TQUB_BaselineSeg_PRec ): TQUB_BaselineSeg_Node_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetFirstNode@4';
function  QUB_BaselineSeg_GetLastNode( Seg :TQUB_BaselineSeg_PRec ): TQUB_BaselineSeg_Node_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetLastNode@4';
function  QUB_BaselineSeg_GetNodeBefore( Seg :TQUB_BaselineSeg_PRec; Node :TQUB_BaselineSeg_Node_PRec ): TQUB_BaselineSeg_Node_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetNodeBefore@8';
function  QUB_BaselineSeg_GetNodeAfter( Seg :TQUB_BaselineSeg_PRec; Node :TQUB_BaselineSeg_Node_PRec ): TQUB_BaselineSeg_Node_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetNodeAfter@8';

function  QUB_BaselineSeg_InterpolateValueAt ( Seg :TQUB_BaselineSeg_PRec; Point :Integer ): Integer; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_InterpolateValueAt@8';
// rn non-) f.g. out-of-range / past-the-end
function  QUB_BaselineSeg_GetRangeAt( Seg :TQUB_BaselineSeg_PRec; Point :Integer; First, Last :TQUB_BaselineSeg_Node ): Integer; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetRangeAt@16';
function  QUB_BaselineSeg_GetRangeBefore( Seg :TQUB_BaselineSeg_PRec; First, Last :TQUB_BaselineSeg_Node_PPRec ): Integer; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetRangeBefore@12';
function  QUB_BaselineSeg_GetRangeAfter( Seg :TQUB_BaselineSeg_PRec; First, Last :TQUB_BaselineSeg_Node_PPRec ): Integer; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_GetRangeAfter@12';
// t the pranill be cleared first
function  QUB_BaselineSeg_AddNode( Seg :TQUB_BaselineSeg_PRec; Point, Value :Integer): TQUB_BaselineSeg_Node_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_AddNode@12';
function  QUB_BaselineSeg_AddNodes( Seg :TQUB_BaselineSeg_PRec; Count :Integer; Points, Values :TPAInteger): TQUB_BaselineSeg_Node_PRec; stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_AddNodes@16';
procedure QUB_BaselineSeg_ClearNodes( Seg :TQUB_BaselineSeg_PRec; First, Last, RepairFirstLast :Integer ); stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_ClearNodes@16';

procedure QUB_BaselineSeg_Offset( Seg :TQUB_BaselineSeg_PRec; Points :Integer ); stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_Offset@8';
procedure QUB_BaselineSeg_Insert( Seg :TQUB_BaselineSeg_PRec; Before, Points :Integer ); stdcall; external 'qubopt.dll' name '_QUB_BaselineSeg_Insert@12';


type


TQUB_BaselineSeg_Node_Py = class(TPyObject)
protected
  prNode                  :TQUB_BaselineSeg_Node;

public

  function  GetAttr(key : PChar) : PPyObject; override;

  property Node        :TQUB_BaselineSeg_Node      read prNode    write prNode;
end;


TQUB_BaselineSeg = class
protected
  prSeg        :TQUB_BaselineSeg_PRec;

  constructor CreateUsing(Seg :TQUB_BaselineSeg_PRec);

  function  prGetFirst: Integer;
  function  prGetLast: Integer;
  function  prGetNodeCount: Integer;

public
  OnFree            :TNotifyEvent;

  destructor Destroy; override;    // prSeg (aka Rec) is freed by the TQUB_Baseline

  function AsTree: INode;
  function GetNodes(First, Last :Integer): INode;

  function ValueAt(Point :Integer): Integer;

  function FirstNode: TQUB_BaselineSeg_Node;
  function LastNode: TQUB_BaselineSeg_Node;

  function NextNode(Node :TQUB_BaselineSeg_Node): TQUB_BaselineSeg_Node; // or nil
  function PrevNode(Node :TQUB_BaselineSeg_Node): TQUB_BaselineSeg_Node;

  function RangeAt(Point :Integer; var First, Last :TQUB_BaselineSeg_Node): Boolean;
  function PrevRange(var First, Last :TQUB_BaselineSeg_Node): Boolean;
  function NextRange(var First, Last :TQUB_BaselineSeg_Node): Boolean;

  //procedure CountNodes(First, Last :Integer; Partials :Boolean; Counts, FirstNodes :TList);  // of Integer and Node respectively
  //procedure GetNodes(First, Last :Integer; Fragments :Boolean; DataSet :INode);

  property Rec              :TQUB_BaselineSeg_PRec  read prSeg;

  property First            :Integer          read prGetFirst;
  property Last             :Integer          read prGetLast;
  property NodeCount       :Integer          read prGetNodeCount;
end;

// py object holding an BaselineSeg, freeing it on Free
TQUB_BaselineSeg_Py = class(TPyObject)
protected
  prSeg                  :TQUB_BaselineSeg;

public
  destructor Destroy; override;

  function  GetAttr(key : PChar) : PPyObject; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;
  function NextNode( args :PPyObject ): PPyObject; cdecl;
  function PrevNode( args :PPyObject ): PPyObject; cdecl;
  function RangeAt( args :PPyObject ): PPyObject; cdecl;
  function NextRange( args :PPyObject ): PPyObject; cdecl;
  function PrevRange( args :PPyObject ): PPyObject; cdecl;
  function ValueAt( args :PPyObject ): PPyObject; cdecl;

  property Seg           :TQUB_BaselineSeg      read prSeg         write prSeg;
end;


TQUB_BaselineSegEditor = class(TQUB_BaselineSeg)
protected

  constructor Create(First, Last :Integer);
  // constructor CreateFromTree(Segment :INode);

public

  procedure SetFirst(x :Integer);
  procedure SetLast(x :Integer);

  procedure AddNode(Point, Value :Integer);
  procedure AddNodes(Count :Integer; Points, Values :TPAInteger); overload;
  procedure AddNodes(Segment :INode); overload;

  procedure ClearNodes(F, L :Integer);

  procedure Offset(Points :Integer);
  procedure OffsetValues(ADUnits :Integer);
  procedure Insert(Before, Points :Integer);

end;


// py object holding an BaselineSegEditor, freeing it on Free
TQUB_BaselineSegEditor_Py = class(TQUB_BaselineSeg_Py)
protected
  function SegEditor: TQUB_BaselineSegEditor;

public
  function  GetAttr(key : PChar) : PPyObject; override;
  function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;
  function ClearNodes( args :PPyObject ): PPyObject; cdecl;
  function AddNode( args :PPyObject ): PPyObject; cdecl;
  function AddNodes( args :PPyObject ): PPyObject; cdecl;
  function Offset( args :PPyObject ): PPyObject; cdecl;
end;


TQUB_Baseline_Py = class;


TQUB_Baseline = class
protected
  prLock          :TCriticalSection;
  prSegRecs       :TList;
  prRdrCounts     :TList;

  prPyHandle      :TQUB_Baseline_Py; // python surrogate of Self

  procedure prOnFreeReadSeg(Sender :TObject);
  procedure prOnFreeWriteSeg(Sender :TObject);

  function prGetSegCount: Integer;
  function prGetSegRec(Ix :Integer): TQUB_BaselineSeg_PRec;

  property SegRec[Ix :Integer]     :TQUB_BaselineSeg_PRec       read prGetSegRec;

public
  constructor Create;
  destructor Destroy; override;

  function ToTree: INode;
  function GetNodes(First, Last :Integer): INode;

  function GetReadSeg(Ix :Integer): TQUB_BaselineSeg; // free it to show you're done    (exception if it's write-locked for 5 seconds)
  function GetReadSegs(First, Last :Integer; Segs :TList): Boolean;
  function GetWriteSeg(Ix :Integer): TQUB_BaselineSegEditor; // may be nil (someone's reading) // free it as soon as possible to show you're done
  function GetWriteSegs(First, Last :Integer; Segs :TList): Boolean;

  procedure AddSeg(First, Last :Integer);
  function DeleteData(First, Last :Integer): Boolean; // truncate, break, erase in range, renumber following seg.(first,last)
  function InsertPoints(Before, Points :Integer): Boolean;
  function InsertSegment(First, Last :Integer): Boolean;
  function BreakSegment(First :Integer): Boolean; // breaks segment containing first into two
  function JoinSegments(First :Integer): Boolean;
  function OffsetValues(ADUnits :Integer): Boolean;

  function ClearNodes(F, L :Integer): Boolean;

  function AddNode(Point, Value :Integer): Boolean;
  function AddNodes(Count :Integer; Points, Values :TPAInteger): Boolean; overload;
  function AddNodes(BaselineSegments :INode): Boolean; overload;

  property SegCount       :Integer               read prGetSegCount;

  property PyHandle       :TQUB_Baseline_Py  read prPyHandle;
end;


// py surrogate for idealization, as with the viewports
TQUB_Baseline_Py = class(TPyObject)
protected
  prBaseline              :TQUB_Baseline;

public
  function  GetAttr(key : PChar) : PPyObject; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;
  function GetReadSeg( args : PPyObject ) : PPyObject; cdecl;
  function GetWriteSeg( args : PPyObject ) : PPyObject; cdecl;
  function AddSeg( args : PPyObject ) : PPyObject; cdecl;
  function DeleteData( args : PPyObject ) : PPyObject; cdecl;
  function InsertPoints( args :PPyObject ) : PPyObject; cdecl;
  function BreakSegment( args :PPyObject ) : PPyObject; cdecl;
  function OffsetValues( args :PPyObject ) : PPyObject; cdecl;
  function ClearNodes( args : PPyObject ) : PPyObject; cdecl;
  function AddNode( args : PPyObject ) : PPyObject; cdecl;
  function AddNodes( args : PPyObject ) : PPyObject; cdecl;

  property Baseline :TQUB_Baseline  read prBaseline  write prBaseline;
end;



var
  QUB_BaselineSeg_Node_PyType    :TPythonType;
  QUB_BaselineSeg_PyType         :TPythonType;
  QUB_BaselineSegEditor_PyType   :TPythonType;
  QUB_Baseline_PyType            :TPythonType;

  
procedure QUB_Baseline_Py_Init;




implementation


uses

QUB_Python_Thread,
QUB_Viewport_Python,
QUB_Node_Python;


constructor TQUB_BaselineSeg.CreateUsing(Seg :TQUB_BaselineSeg_PRec);
begin
  prSeg         := Seg;
end;

destructor TQUB_BaselineSeg.Destroy;
begin
  try
    if Assigned(OnFree) then OnFree(Self);
  finally
    inherited;
  end;
end;

function  TQUB_BaselineSeg.prGetFirst: Integer;
begin
  Result := prSeg.First;
end;

function  TQUB_BaselineSeg.prGetLast: Integer;
begin
  Result := prSeg.Last;
end;

function  TQUB_BaselineSeg.prGetNodeCount: Integer;
begin
  Result := QUB_BaselineSeg_GetNodeCount(prSeg);
end;

function TQUB_BaselineSeg.ValueAt(Point :Integer): Integer;
begin
  Result := QUB_BaselineSeg_InterpolateValueAt(prSeg, Point);
end;

function TQUB_BaselineSeg.FirstNode: TQUB_BaselineSeg_Node;
begin
  Result := QUB_BaselineSeg_GetFirstNode(prSeg);
end;

function TQUB_BaselineSeg.LastNode: TQUB_BaselineSeg_Node;
begin
  Result := QUB_BaselineSeg_GetLastNode(prSeg);
end;

function TQUB_BaselineSeg.NextNode(Node :TQUB_BaselineSeg_Node): TQUB_BaselineSeg_Node; // or nil
begin
  Result := QUB_BaselineSeg_GetNodeAfter(prSeg, Node);
end;

function TQUB_BaselineSeg.PrevNode(Node :TQUB_BaselineSeg_Node): TQUB_BaselineSeg_Node;
begin
  Result := QUB_BaselineSeg_GetNodeBefore(prSeg, Node);
end;

function TQUB_BaselineSeg.RangeAt(Point :Integer; var First, Last :TQUB_BaselineSeg_Node): Boolean;
begin
  Result := 0 <> QUB_BaselineSeg_GetRangeAt(prSeg, Point, @First, @Last);
end;

function TQUB_BaselineSeg.PrevRange(var First, Last :TQUB_BaselineSeg_Node): Boolean;
begin
  Result := 0 <> QUB_BaselineSeg_GetRangeBefore(prSeg, @First, @Last);
end;

function TQUB_BaselineSeg.NextRange(var First, Last :TQUB_BaselineSeg_Node): Boolean;
begin
  Result := 0 <> QUB_BaselineSeg_GetRangeAfter(prSeg, @First, @Last);
end;

function TQUB_BaselineSeg.AsTree: INode;
var
  Node       :TQUB_BaselineSeg_Node;
  Child      :INode;
  Points     :TPAInteger;
  Values     :TPAInteger;
  i, N       :Integer;
begin
  N          := NodeCount;
  Points     := nil;
  Values     := nil;

  Result     := TNode.Create('BaselineSegment');
  Result.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
  Result.Data[0]  := prSeg.First;
  Result.Data[1]  := prSeg.Last;
  Result['NodeCount'].SetData( N );

  Child      := Result['Points'];
  if N > 0 then begin
    Child.SetupNumData( NODE_DATATYPE_LONGINT, N, 1 );
    Points   := TPAInteger( Child.DataPtr );
  end;
  Child      := Result['Values'];
  if N > 0 then begin
    Child.SetupNumData( NODE_DATATYPE_LONGINT, N, 1 );
    Values   := TPAInteger( Child.DataPtr );
  end;

  Node       := FirstNode;
  for i := 0 to N - 1 do begin
    Points[i] := Node.Point;
    Values[i] := Node.Value;
    Node     := NextNode(Node);
  end;
end;

function TQUB_BaselineSeg.GetNodes(First, Last :Integer): INode;
var
  Node          :TQUB_BaselineSeg_Node;
  FirstInRange  :TQUB_BaselineSeg_Node;
  Child         :INode;
  Points        :TPAInteger;
  Values        :TPAInteger;
  i, N          :Integer;
begin
  Node          := FirstNode;
  FirstInRange  := nil;
  N             := 0;
  while (Node <> nil) and (Node.Point <= Last) do begin
    if Node.Point >= First then begin
      Inc(N);
      if FirstInRange = nil then FirstInRange := Node;
    end;
    Node     := NextNode(Node);
  end;

  Points     := nil;
  Values     := nil;

  Result     := TNode.Create('BaselineSegment');
  Result.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
  Result.Data[0]  := First;
  Result.Data[1]  := Last;
  Result['NodeCount'].SetData( N );

  Child      := Result['Points'];
  if N > 0 then begin
    Child.SetupNumData( NODE_DATATYPE_LONGINT, N, 1 );
    Points   := TPAInteger( Child.DataPtr );
  end;
  Child      := Result['Values'];
  if N > 0 then begin
    Child.SetupNumData( NODE_DATATYPE_LONGINT, N, 1 );
    Values   := TPAInteger( Child.DataPtr );
  end;

  Node       := FirstInRange;
  for i := 0 to N - 1 do begin
    Points[i] := Node.Point;
    Values[i] := Node.Value;
    Node     := NextNode(Node);
  end;
end;


//procedure CountNodes(First, Last :Integer; Partials :Boolean; Counts, FirstNodes :TList);  // of Integer and Node respectively
//procedure GetNodes(First, Last :Integer; Fragments :Boolean; DataSet :INode);



// *****************************************************************************

constructor TQUB_BaselineSegEditor.Create(First, Last :Integer);
begin
  prSeg := QUB_BaselineSeg_Create(First, Last);
end;

// constructor CreateFromTree(Segment :INode);

procedure TQUB_BaselineSegEditor.SetFirst(x :Integer);
begin
  QUB_BaselineSeg_SetFirst(prSeg, x);
end;

procedure TQUB_BaselineSegEditor.SetLast(x :Integer);
begin
  QUB_BaselineSeg_SetLast(prSeg, x);
end;

procedure TQUB_BaselineSegEditor.AddNode(Point :Integer; Value :Integer);
begin
  QUB_BaselineSeg_AddNode(prSeg, Point, Value);
end;

procedure TQUB_BaselineSegEditor.AddNodes(Count :Integer; Points :TPAInteger; Values :TPAInteger); //overload;
begin
  QUB_BaselineSeg_AddNodes(prSeg, Count, Points, Values);
end;

procedure TQUB_BaselineSegEditor.AddNodes(Segment :INode);
begin
  try
    ClearNodes( Segment.Data[0], Segment.Data[1] );
    AddNodes( Segment['NodeCount'].Data[0],
              TPAInteger( Segment['Points'].DataPtr ),
              TPAInteger( Segment['Values'].DataPtr ) );
  except
  end;
end;

procedure TQUB_BaselineSegEditor.ClearNodes(F, L :Integer);
begin
  QUB_BaselineSeg_ClearNodes(prSeg, F, L, 1{re-gen unbalanced marginal node});
end;

procedure TQUB_BaselineSegEditor.Offset(Points :Integer);
begin
  QUB_BaselineSeg_Offset(prSeg, Points);
end;

procedure TQUB_BaselineSegEditor.OffsetValues(ADUnits :Integer);
var
  Node            :TQUB_BaselineSeg_Node;
begin
  Node            := FirstNode;
  while Node <> nil do begin
    Node.Value    := Node.Value + ADUnits;

    Node          := NextNode(Node);
  end;
end;

procedure TQUB_BaselineSegEditor.Insert(Before, Points :Integer);
begin
  QUB_BaselineSeg_Insert(prSeg, Before, Points);
end;



// prLock          :TCriticalSection;
// prSegRecs       :TList;
// prRdrCounts     :TList;



// *****************************************************************************

constructor TQUB_Baseline.Create;
var
  PyInst          :PPyObject;
begin
  prLock          := TCriticalSection.Create;
  prSegRecs       := TList.Create;
  prRdrCounts     := TList.Create;

  PythonThread.BeginPython;
  PyInst                  := QUB_Baseline_PyType.CreateInstance;
  prPyHandle              := TQUB_Baseline_Py( PythonToDelphi(PyInst) );
  prPyHandle.Baseline     := Self;
  PythonThread.EndPython;
end;

destructor TQUB_Baseline.Destroy;
var
  i               :Integer;
begin
  try
    for i := 0 to SegCount - 1 do begin
      while Integer( prRdrCounts[i] ) <> 0 do Sleep(50);
      QUB_BaselineSeg_Free( TQUB_BaselineSeg_PRec( prSegRecs[i] ) );
    end;
    prSegRecs.Free;
    prRdrCounts.Free;
    prLock.Free;

    PythonThread.BeginPython;
    prPyHandle.Baseline    := nil;
    GetPythonEngine.Py_DECREF( prPyHandle.GetSelf );
    PythonThread.EndPython;
  finally
    inherited;
  end;
end;

procedure TQUB_Baseline.prOnFreeReadSeg(Sender :TObject);
var
  i          :Integer;
begin
  prLock.Acquire;
  i          := prSegRecs.IndexOf( TQUB_BaselineSeg(Sender).Rec );
  prRdrCounts[i] := Pointer( Integer(prRdrCounts[i]) - 1 );
  prLock.Release;
end;

procedure TQUB_Baseline.prOnFreeWriteSeg(Sender :TObject);
var
  i          :Integer;
begin
  prLock.Acquire;
  i          := prSegRecs.IndexOf( TQUB_BaselineSegEditor(Sender).Rec );
  prRdrCounts[i] := Pointer( 0 );
  prLock.Release;
end;

function TQUB_Baseline.prGetSegCount: Integer;
begin
  Result := prSegRecs.Count;
end;

function TQUB_Baseline.prGetSegRec(Ix :Integer): TQUB_BaselineSeg_PRec;
begin
  Result := TQUB_BaselineSeg_PRec( prSegRecs[Ix] );
end;

function TQUB_Baseline.ToTree: INode;
var
  LastSeg  :INode;
  i        :Integer;
  Seg      :TQUB_BaselineSeg;
begin
  try

    Result   := TNode.Create('BaselineSegments');
    LastSeg  := TNode.CreateNull;

    for i := 0 to SegCount - 1 do begin
      Seg     := GetReadSeg(i);
      LastSeg := Result.InsertChild( LastSeg, Seg.AsTree );
      Seg.Free;
    end;

  except
  end;

end;

function TQUB_Baseline.GetNodes(First, Last :Integer): INode;
var
  LastSeg  :INode;
  i        :Integer;
  Seg      :TQUB_BaselineSeg;
  f, l     :Integer;
begin
  Result   := TNode.Create('BaselineSegments');
  LastSeg  := TNode.CreateNull;

  for i := 0 to SegCount - 1 do begin
    f       := Max(First, SegRec[i].First);
    l       := Min(Last,  SegRec[i].Last);
    if f <= l then begin
      Seg     := GetReadSeg(i);
      LastSeg := Result.InsertChild( LastSeg, Seg.GetNodes(f, l) );
      Seg.Free;
    end;
  end;
end;

function TQUB_Baseline.GetReadSeg(Ix :Integer): TQUB_BaselineSeg; // free it to show you're done    (exception if it's write-locked for 5 seconds)
var
  Rdrs       :Integer;
  SleepTime  :Integer;
begin
  Result     := nil;
  SleepTime  := 0;
  while Result = nil do begin
    prLock.Acquire;
    Rdrs     := Integer( prRdrCounts[Ix] );
    if Rdrs >= 0 then begin
      prRdrCounts[Ix] := Pointer( Rdrs + 1 );
      Result          := TQUB_BaselineSeg.CreateUsing(SegRec[Ix]);
      Result.OnFree   := prOnFreeReadSeg;
    end;
    prLock.Release;

    if Result = nil then begin
      if SleepTime >= 5000 then begin
        raise Exception.Create(Format('Can''t read baseline segment %d: another part of the program is editing it.', [Ix+1]) );
      end;

      Sleep(100);
      SleepTime := SleepTime + 100;
    end;
  end;
end;

function TQUB_Baseline.GetReadSegs(First, Last :Integer; Segs :TList): Boolean;
var
  i              :Integer;
  OrigCount      :Integer;
begin
  OrigCount      := Segs.Count;
  Result         := False;

  try
    for i := 0 to SegCount - 1 do begin
      if (SegRec[i].First <= Last) or (Last < First) then begin
        if (SegRec[i].Last >= First) or (SegRec[i].Last = (SegRec[i].First-1)) then begin
          Segs.Add( GetReadSeg(i) );
        end;
      end else begin
        Break;
      end;
    end;
  except
    for i := OrigCount to Segs.Count - 1 do begin
      TQUB_BaselineSeg( Segs[i] ).Free;
    end;
    Segs.Count   := OrigCount;
    Result       := True;
  end;
end;

function TQUB_Baseline.GetWriteSeg(Ix :Integer): TQUB_BaselineSegEditor; // may be nil (someone's reading) // free it as soon as possible to show you're done
begin
  Result := nil;
  prLock.Acquire;
  if Integer( prRdrCounts[Ix] ) = 0 then begin
    prRdrCounts[Ix] := Pointer( -1 );
    Result          := TQUB_BaselineSegEditor.CreateUsing(SegRec[Ix]);
    Result.OnFree   := prOnFreeWriteSeg;
  end;
  prLock.Release;
end;

function TQUB_Baseline.GetWriteSegs(First, Last :Integer; Segs :TList): Boolean;
var
  i              :Integer;
  WaitsRemaining :Integer;
  OrigCount      :Integer;
  Seg            :TQUB_BaselineSegEditor;
begin
  WaitsRemaining := 10;
  OrigCount      := Segs.Count;
  Result         := False;

  try
    for i := 0 to SegCount - 1 do begin
      if (SegRec[i].First <= Last) or (Last < First) then begin
        if (SegRec[i].Last >= First) or (SegRec[i].Last = (SegRec[i].First-1)) then begin
          Seg    := nil;
          while Seg = nil do begin
            Seg  := GetWriteSeg(i);
            if Seg <> nil then begin
              Segs.Add(seg);
            end else begin
              if WaitsRemaining = 0 then begin
                raise Exception.Create('');
              end else begin
                Sleep(100);
                Dec(WaitsRemaining);
              end;
            end;
          end;
        end;
      end else begin
        Break;
      end;
    end;
  except
    for i := OrigCount to Segs.Count - 1 do begin
      TQUB_BaselineSegEditor( Segs[i] ).Free;
    end;
    Segs.Count   := OrigCount;
    Result       := True;
  end;
end;

procedure TQUB_Baseline.AddSeg(First, Last :Integer);
var
  SegEditor       :TQUB_BaselineSegEditor;
begin
  prLock.Acquire;
  try
    SegEditor   := TQUB_BaselineSegEditor.Create(First, Last);
    prSegRecs.Add( SegEditor.Rec );
    prRdrCounts.Add( Pointer(0) );
    SegEditor.Free;
  except
  end;
  prLock.Release;
end;

function TQUB_Baseline.DeleteData(First, Last :Integer): Boolean; // truncate, break, erase in range, renumber following seg.(first,last)
var
  Segs        :TList;
  Seg         :TQUB_BaselineSegEditor;
  i           :Integer;
  SegIx       :Integer;
  Offset      :Integer;
  NewSeg      :TQUB_BaselineSeg_PRec;
  OldSegCount :Integer;

  procedure NextSeg;
  begin
    if Seg <> nil then Seg.Free;
    Inc(i);
    Inc(SegIx);
    if i < Segs.Count then begin
      Seg     := TQUB_BaselineSegEditor( Segs[i] );
    end else begin
      Seg     := nil;
    end;
  end;

  procedure RemSeg;
  begin
    if Seg = nil then Exit;

    if prSegRecs.Count = 1 then begin // to match QDF/QVD, leave an empty seg
      Seg.SetLast( Seg.First - 1 );
      Seg.Free;
      Inc(i);
    end else begin
      Seg.Free;
      prLock.Acquire;
      prSegRecs.Delete(SegIx);
      prRdrCounts.Delete(SegIx);
      Segs.Delete(i);
      prLock.Release;
    end;

    if i < Segs.Count then begin
      Seg    := TQUB_BaselineSegEditor( Segs[i] );
    end else begin
      Seg    := nil;
    end;
  end;

begin
  OldSegCount := SegCount;
  Segs        := TList.Create;
  Result      := GetWriteSegs(First, SegRec[SegCount-1].Last, Segs);
  if (Segs.Count > 0) and not Result then begin
    i         := -1;
    Seg       := nil;
    NextSeg;
    SegIx     := prSegRecs.IndexOf(Seg.Rec);

    if Seg.First < First then begin
      if Last < Seg.Last then begin
        prLock.Acquire;
        NewSeg    := QUB_BaselineSeg_ExciseFrom(Seg.Rec, Last+1);
        prSegRecs.Insert( SegIx + 1, NewSeg );
        prRdrCounts.Insert( SegIx + 1, Pointer(0) );
        prLock.Release;
        Segs.Insert( 1, GetWriteSeg( SegIx + 1 ) );

        Seg.SetLast( First - 1 );
        NextSeg;
      end else begin
        Seg.SetLast( First - 1 );
        NextSeg;
      end;
    end;

    while (Seg <> nil) and (Seg.Last <= Last) do begin
      RemSeg;
    end;

    if Seg <> nil then begin

      if Seg.First <= Last then begin
        Seg.SetFirst( Last + 1 );
      end;

      Offset    := - (Last - First + 1);
      while Seg <> nil do begin
        Seg.Offset( Offset );
        NextSeg;
      end;

    end;

  end;
  Segs.Free;

  // for consistency with QDF, undo any seg break
  if OldSegCount < SegCount then JoinSegments(First);
end;

function TQUB_Baseline.InsertPoints(Before, Points :Integer): Boolean;
var
  Segs         :TList;
  i            :Integer;
begin
  if Points <= 0 then begin
    Result     := True;
    Exit;
  end;

  Segs         := TList.Create;
  Result       := GetWriteSegs(Before, SegRec[SegCount-1].Last, Segs);
  if Segs.Count = 0 then begin
    Result     := True;
    Segs.Free;
    Exit;
  end;

  TQUB_BaselineSegEditor( Segs[0] ).Insert(Before, Points);
  TQUB_BaselineSegEditor( Segs[0] ).Free;

  for i := 1 to Segs.Count - 1 do begin
    TQUB_BaselineSegEditor( Segs[i] ).Offset(Points);
    TQUB_BaselineSegEditor( Segs[i] ).Free;
  end;

  Segs.Free;
end;

function TQUB_Baseline.InsertSegment(First, Last :Integer): Boolean;
var
  Segs        :TList;
  Seg         :TQUB_BaselineSegEditor;
  NewSeg      :TQUB_BaselineSeg_PRec;
  i           :Integer;
  InsertAt    :Integer;
begin
  if First > SegRec[ SegCount-1 ].Last then begin
    AddSeg(First, Last);
    Result    := False;
    Exit;
  end;
  
  Segs        := TList.Create;
  try

    Result      := GetWriteSegs(First, SegRec[SegCount-1].Last, Segs);
    if (Segs.Count > 0) and not Result then begin
      prLock.Acquire;

      Seg       := TQUB_BaselineSegEditor(Segs[0]);
      InsertAt  := prSegRecs.IndexOf(Seg.Rec);
      if Seg.First < First then begin
        Inc(InsertAt);
        NewSeg  := QUB_BaselineSeg_ExciseFrom(Seg.Rec, First);
        prSegRecs.Insert( InsertAt, NewSeg );
        prRdrCounts.Insert( InsertAt, Pointer(0) );

        Seg.Free;
        Segs[0] := GetWriteSeg(InsertAt);
      end;

      NewSeg    := QUB_BaselineSeg_Create(First, Last);
      prSegRecs.Insert( InsertAt, NewSeg );
      prRdrCounts.Insert( InsertAt, Pointer(0) );

      for i := 0 to Segs.Count - 1 do begin
        TQUB_BaselineSegEditor(Segs[i]).Offset(Last - First + 1);
      end;

      prLock.Release;
    end;
  finally
    for i := 0 to Segs.Count - 1 do begin
      TQUB_BaselineSegEditor(Segs[i]).Free;
    end;
    Segs.Free;
  end;
end;

function TQUB_Baseline.BreakSegment(First :Integer): Boolean;
var
  Segs        :TList;
  Seg         :TQUB_BaselineSegEditor;
  NewSeg      :TQUB_BaselineSeg_PRec;
  i           :Integer;
begin
  Segs        := TList.Create;
  Result      := GetWriteSegs(First, First, Segs);
  if (Segs.Count > 0) and not Result then begin
    Seg       := TQUB_BaselineSegEditor(Segs[0]);
    if Seg.First < First then begin
      prLock.Acquire;
      NewSeg  := QUB_BaselineSeg_ExciseFrom(Seg.Rec, First);
      i       := prSegRecs.IndexOf(Seg.Rec) + 1;
      prSegRecs.Insert( i, NewSeg );
      prRdrCounts.Insert( i, Pointer(0) );
      prLock.Release;
    end;
  end;

  for i := 0 to Segs.Count - 1 do begin
    TQUB_BaselineSegEditor(Segs[i]).Free;
  end;
  Segs.Free;
end;

function TQUB_Baseline.JoinSegments(First :Integer): Boolean;
var
  Segs        :TList;
  SegA, SegB  :TQUB_BaselineSegEditor;
  SegBix      :Integer;
  i           :Integer;
  Node        :TQUB_BaselineSeg_Node;
begin
  Segs        := TList.Create;
  Result      := GetWriteSegs(First-1, First, Segs);
  try
    if (Segs.Count = 2) and not Result then begin
      SegA      := TQUB_BaselineSegEditor(Segs[0]);
      SegB      := TQUB_BaselineSegEditor(Segs[1]);
      SegA.SetLast( SegB.Last );

      Node      := SegB.FirstNode;
      while Node <> nil do begin
        SegA.AddNode(Node.Point, Node.Value);
        Node    := SegB.NextNode(Node);
      end;

      SegBix    := prSegRecs.IndexOf(SegB.Rec);
      SegB.Free;
      prLock.Acquire;
      prSegRecs.Delete(SegBix);
      prRdrCounts.Delete(SegBix);
      Segs.Delete(1);
      prLock.Release;
    end;
  finally
    for i := 0 to Segs.Count - 1 do begin
      TQUB_BaselineSegEditor(Segs[i]).Free;
    end;
    Segs.Free;
  end;
end;

function TQUB_Baseline.OffsetValues(ADUnits :Integer): Boolean;
var
  i           :Integer;
  Segs        :TList;
begin
  Result      := False;
  if SegCount = 0 then Exit;

  Segs        := TList.Create;
  Result      := GetWriteSegs(0, SegRec[ SegCount-1 ].Last, Segs);
  try
    if not Result then begin
      for i := 0 to Segs.Count - 1 do begin
        TQUB_BaselineSegEditor(Segs[i]).OffsetValues(ADUnits);
      end;
    end;
  except
    Result    := True;
  end;

  for i := 0 to Segs.Count - 1 do begin
    TQUB_BaselineSegEditor(Segs[i]).Free;
  end;
  Segs.Free;
end;

function TQUB_Baseline.ClearNodes(F, L :Integer): Boolean;
var
  Segs   :TList;
  Seg    :TQUB_BaselineSegEditor;
  i      :Integer;
begin
  Segs   := TList.Create;
  Result := GetWriteSegs(F, L, Segs);
  if not Result then begin
    for i := 0 to Segs.Count - 1 do begin
      Seg := TQUB_BaselineSegEditor(Segs[i]);
      Seg.ClearNodes( Max(F, Seg.First), Min(L, Seg.Last) );
      Seg.Free;
    end;
  end;

  Segs.Free;
end;

function TQUB_Baseline.AddNode(Point, Value :Integer): Boolean;
var
  i          :Integer;
  Seg        :TQUB_BaselineSegEditor;
begin
  Result := True;
  if (SegCount = 0) or (Point < SegRec[0].First) or (Point > SegRec[SegCount-1].Last) then Exit;

  for i := 0 to SegCount - 1 do begin
    if Point <= SegRec[i].Last then begin
      Seg    := GetWriteSeg(i);
      if Seg <> nil then begin
        try
          Seg.AddNode(Point, Value);
          Result := False;
        finally
          Seg.Free;
        end;
      end;
      Break;
    end;
  end;
end;

function TQUB_Baseline.AddNodes(Count :Integer; Points, Values :TPAInteger): Boolean; //overload;
var
  Segs            :TList;
  Seg             :TQUB_BaselineSegEditor;
  i               :Integer;
begin
  Segs            := TList.Create;
  Result          := GetWriteSegs( Points[0], Points[Count-1], Segs );
  if not Result then begin
    for i := 0 to Segs.Count - 1 do begin
      Seg         := TQUB_BaselineSegEditor( Segs[i] );
      Seg.AddNodes( Count, Points, Values );
      Seg.Free;
    end;
  end;

  Segs.Free;
end;

function TQUB_Baseline.AddNodes(BaselineSegments :INode): Boolean;
var
  Segment         :INode;
  Segs            :TList;
  Seg             :TQUB_BaselineSegEditor;
  i               :Integer;
  F, L            :Integer;
begin
  Result          := False;

  // find bounds
  Segment         := BaselineSegments.Find('BaselineSegment');
  if Segment.IsNull then Exit;
  F               := Integer(Segment.Data[0]);
  while not Segment.SiblingSameName.IsNull do begin
    Segment       := Segment.SiblingSameName;
  end;
  L               := Integer(Segment.Data[1]);

  Segs            := TList.Create;
  Result          := GetWriteSegs( F, L, Segs );
  if not Result then begin

    i             := 0;
    Segment       := BaselineSegments.Find('BaselineSegment');
    while (i < Segs.Count) and not Segment.IsNull do begin
      Seg         := TQUB_BaselineSegEditor( Segs[i] );

      if Seg.Last < Integer(Segment.Data[0]) then begin
        Inc(i);
      end else begin
        Seg.AddNodes( Segment );

        if Seg.Last <= Integer(Segment.Data[1]) then begin
          Inc(i);
        end;
        if Integer(Segment.Data[1]) <= Seg.Last then begin
          Segment := Segment.SiblingSameName;
        end;
      end;
    end;

  end;

  for i := 0 to Segs.Count - 1 do begin
    TQUB_BaselineSegEditor(Segs[i]).Free;
  end;
  Segs.Free;
end;

//function GetNodes(First, Last :Integer; Fragments :Boolean; BaselineSegments :INode): Boolean; overload;

// these two GetNodes will trust you on the segmentation  /slash/ return true and leave you with an array full of unknown
//function GetSegmentation(First, Last :Integer; Fragments :Boolean; Firsts, Lasts, Counts :TList): Boolean;
//function GetNodes(First, Last, Count :Integer; FLC :TQUB_QFS_QIF_PAFirstLastClass): Boolean; overload;
//function GetNodes(First, Last, Count :Integer; DC :TQUB_QFS_QIF_PANodeClass): Boolean; overload;


// *****************************************************************************


function  TQUB_BaselineSeg_Node_Py.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do begin
    if key = 'point' then begin
      Result := PyInt_FromLong( Node.Point );
    end else if key = 'value' then begin
      Result := PyInt_FromLong( Node.Value );
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;


// *****************************************************************************


destructor TQUB_BaselineSeg_Py.Destroy;
begin
  prSeg.Free;
end;

function  TQUB_BaselineSeg_Py.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do begin
    if key = 'first' then begin
      Result := PyInt_FromLong( Seg.First );
    end else if key = 'last' then begin
      Result := PyInt_FromLong( Seg.Last );
    end else if key = 'firstNode' then begin
      Result := QUB_BaselineSeg_Node_PyType.CreateInstance;
      TQUB_BaselineSeg_Node_Py( PythonToDelphi(Result) ).Node := Seg.FirstNode;
    end else if key = 'lastNode' then begin
      Result := QUB_BaselineSeg_Node_PyType.CreateInstance;
      TQUB_BaselineSeg_Node_Py( PythonToDelphi(Result) ).Node := Seg.LastNode;
    end else if key = 'nodeCount' then begin
      Result := PyInt_FromLong( Seg.NodeCount );
    end else if key = 'asTree' then begin
      Result := PyNode_FromNode( Seg.AsTree );
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

class procedure TQUB_BaselineSeg_Py.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do
    begin
      AddMethod( 'nextNode',   @TQUB_BaselineSeg_Py.NextNode, '' );
      AddMethod( 'prevNode',   @TQUB_BaselineSeg_Py.PrevNode, '' );
      AddMethod( 'rangeAt',    @TQUB_BaselineSeg_Py.RangeAt, '' );
      AddMethod( 'nextRange',  @TQUB_BaselineSeg_Py.NextRange, '' );
      AddMethod( 'prevRange',  @TQUB_BaselineSeg_Py.PrevRange, '' );
      AddMethod( 'valueAt',    @TQUB_BaselineSeg_Py.ValueAt, '' );
    end;
end;

function TQUB_BaselineSeg_Py.NextNode( args :PPyObject ): PPyObject;
var
  Node   :TQUB_BaselineSeg_Node;
  NodePy :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'O:nextNode(Node)', [@NodePy] ) <> 0 then begin
      if NodePy = Py_None then begin
        Node    := nil;
      end else begin
        Node    := TQUB_BaselineSeg_Node_Py( PythonToDelphi(NodePy) ).Node;
      end;

      Node  := Seg.NextNode(Node);

      if Node = nil then begin
        Result := ReturnNone;
      end else begin
        Result := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(Result) ).Node := Node;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSeg_Py.PrevNode( args :PPyObject ): PPyObject;
var
  Node   :TQUB_BaselineSeg_Node;
  NodePy :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'O:prevNode(Node)', [@NodePy] ) <> 0 then begin
      if NodePy = Py_None then begin
        Node    := nil;
      end else begin
        Node    := TQUB_BaselineSeg_Node_Py( PythonToDelphi(NodePy) ).Node;
      end;

      Node  := Seg.PrevNode(Node);

      if Node = nil then begin
        Result := ReturnNone;
      end else begin
        Result := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(Result) ).Node := Node;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSeg_Py.RangeAt( args :PPyObject ): PPyObject;
var
  i      :Integer;
  First  :TQUB_BaselineSeg_Node;
  Last   :TQUB_BaselineSeg_Node;
  Node   :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:rangeAt(point)', [@i] ) <> 0 then begin
      if Seg.RangeAt(i, First, Last) then begin
        Result := ReturnNone;
      end else begin
        Result := PyTuple_New(2);
        Node   := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(Node) ).Node := First;
        PyTuple_SetItem(Result, 0, Node);
        Node   := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(Node) ).Node := Last;
        PyTuple_SetItem(Result, 1, Node);
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSeg_Py.NextRange( args :PPyObject ): PPyObject;
var
  First   :TQUB_BaselineSeg_Node;
  Last    :TQUB_BaselineSeg_Node;
  FirstPy :PPyObject;
  LastPy  :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, '(OO):nextRange( (from, to) )', [@FirstPy, @LastPy] ) <> 0 then begin
      First    := TQUB_BaselineSeg_Node_Py( PythonToDelphi(FirstPy) ).Node;
      Last     := TQUB_BaselineSeg_Node_Py( PythonToDelphi(LastPy) ).Node;
      if Seg.NextRange(First, Last) then begin
        Result := ReturnNone;
      end else begin
        FirstPy   := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(FirstPy) ).Node := First;
        LastPy   := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(LastPy) ).Node := Last;

        Result := PyTuple_New(2);
        PyTuple_SetItem(Result, 0, FirstPy);
        PyTuple_SetItem(Result, 1, LastPy);
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSeg_Py.PrevRange( args :PPyObject ): PPyObject;
var
  First   :TQUB_BaselineSeg_Node;
  Last    :TQUB_BaselineSeg_Node;
  FirstPy :PPyObject;
  LastPy  :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, '(OO):nextRange( (from, to) )', [@FirstPy, @LastPy] ) <> 0 then begin
      First    := TQUB_BaselineSeg_Node_Py( PythonToDelphi(FirstPy) ).Node;
      Last     := TQUB_BaselineSeg_Node_Py( PythonToDelphi(LastPy) ).Node;
      if Seg.PrevRange(First, Last) then begin
        Result := ReturnNone;
      end else begin
        FirstPy   := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(FirstPy) ).Node := First;
        LastPy   := QUB_BaselineSeg_Node_PyType.CreateInstance;
        TQUB_BaselineSeg_Node_Py( PythonToDelphi(LastPy) ).Node := Last;

        Result := PyTuple_New(2);
        PyTuple_SetItem(Result, 0, FirstPy);
        PyTuple_SetItem(Result, 1, LastPy);
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSeg_Py.ValueAt( args :PPyObject ): PPyObject;
var
  i      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:valueAt(point)', [@i] ) <> 0 then begin
      Result := PyInt_FromLong( Seg.ValueAt(i) );
    end else begin
      Result := nil;
    end;
  end;
end;


// *****************************************************************************

function TQUB_BaselineSegEditor_Py.SegEditor: TQUB_BaselineSegEditor;
begin
  Result := TQUB_BaselineSegEditor( Seg );
end;

function  TQUB_BaselineSegEditor_Py.GetAttr(key : PChar) : PPyObject;
begin
//  with GetPythonEngine do begin
//    if key = 'first' then begin
//      Result := PyInt_FromLong( Seg.First );
//    end else begin
      Result := inherited GetAttr(key);
//    end;
//  end;
end;

function  TQUB_BaselineSegEditor_Py.SetAttr(key : PChar; value : PPyObject) : Integer;
var
  p             :Integer;
begin
  Result := 0;
  with GetPythonEngine do begin
    if key = 'first' then begin
      if PyArg_Parse( value, 'i:first', [@p] ) = 0 then begin
        Result := -1;
      end else begin
        SegEditor.SetFirst(p);
      end;
    end else if key = 'last' then begin
      if PyArg_Parse( value, 'i:last', [@p] ) = 0 then begin
        Result := -1;
      end else begin
        SegEditor.SetLast(p);
      end;
    end else begin
      Result := inherited SetAttr(key, value);
    end;
  end;
end;

class procedure TQUB_BaselineSegEditor_Py.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do
    begin
      AddMethod( 'clearNodes',  @TQUB_BaselineSegEditor_Py.ClearNodes, '' );
      AddMethod( 'addNode',     @TQUB_BaselineSegEditor_Py.AddNode, '' );
      AddMethod( 'addNodes',    @TQUB_BaselineSegEditor_Py.AddNodes, '' );
      AddMethod( 'offset',      @TQUB_BaselineSegEditor_Py.Offset, '' );
    end;
end;

function TQUB_BaselineSegEditor_Py.ClearNodes( args :PPyObject ): PPyObject;
var
  f, l      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'ii:clearNodes(first, last)', [@f, @l] ) <> 0 then begin
      SegEditor.ClearNodes(f, l);
      Result := ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSegEditor_Py.AddNode( args :PPyObject ): PPyObject;
var
  p, v    :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'ii:addNode(point, value)', [@p, @v] ) <> 0 then begin
      SegEditor.AddNode(p, v);
      Result := ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSegEditor_Py.AddNodes( args :PPyObject ): PPyObject;
var
  SegmentPy    :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'O:addNodes(BaselineSegment)', [@SegmentPy] ) <> 0 then begin
      if PyNode_IsNode( SegmentPy ) then begin
        SegEditor.AddNodes( PyNode_ToNode( SegmentPy ) );
        Result := ReturnNone;
      end else begin
        PyErr_SetString( PyExc_TypeError^, PChar('argument must be a qubtree.Node') );
        Result := nil;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_BaselineSegEditor_Py.Offset( args :PPyObject ): PPyObject;
var
  p         :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:offset(points)', [@p] ) <> 0 then begin
      SegEditor.Offset(p);
      Result := ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;


// *****************************************************************************

function  TQUB_Baseline_Py.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if key = 'segCount' then begin
      Result := PyInt_FromLong( prBaseline.SegCount );
    end else if key = 'asTree' then begin
      Result := PyNode_FromNode( prBaseline.ToTree );
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

class procedure TQUB_Baseline_Py.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do
    begin
      AddMethod( 'getReadSeg',       @TQUB_Baseline_Py.GetReadSeg, '' );
      AddMethod( 'getWriteSeg',      @TQUB_Baseline_Py.GetWriteSeg, '' );
      AddMethod( 'addSeg',           @TQUB_Baseline_Py.AddSeg, '' );
      AddMethod( 'deleteData',       @TQUB_Baseline_Py.DeleteData, '' );
      AddMethod( 'insertPoints',     @TQUB_Baseline_Py.InsertPoints, '' );
      AddMethod( 'breakSegment',     @TQUB_Baseline_Py.BreakSegment, '' );
      AddMethod( 'offsetValues',     @TQUB_Baseline_Py.OffsetValues, '' );
      AddMethod( 'clearNodes',       @TQUB_Baseline_Py.ClearNodes, '' );
      AddMethod( 'addNode',          @TQUB_Baseline_Py.AddNode, '' );
      AddMethod( 'addNodes',         @TQUB_Baseline_Py.AddNodes, '' );
    end;
end;

function TQUB_Baseline_Py.GetReadSeg( args : PPyObject ) : PPyObject;
var
  i      :Integer;
  Seg    :TQUB_BaselineSeg;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:getReadSeg(i)', [@i] ) <> 0 then begin
      if prBaseline = nil then begin
        Result := ReturnNone;
      end else if (i >= 0) and (i < prBaseline.SegCount) then begin
        try
          Seg := prBaseline.GetReadSeg(i);
          Result := QUB_BaselineSeg_PyType.CreateInstance;
          TQUB_BaselineSeg_Py( PythonToDelphi(Result) ).Seg := Seg;
        except
          PyErr_SetString( PyExc_Exception^, PChar('segment is busy being edited') );
          Result := nil;
        end;
      end else begin
        PyErr_SetString( PyExc_IndexError^, PChar('index out of range') );
        Result := nil;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.GetWriteSeg( args : PPyObject ) : PPyObject;
var
  i      :Integer;
  Editor :TQUB_BaselineSegEditor;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:getWriteSeg(i)', [@i] ) <> 0 then begin
      if (i >= 0) and (i < prBaseline.SegCount) then begin
        if prBaseline = nil then begin
          Result := ReturnNone;
        end else begin
          Editor := prBaseline.GetWriteSeg(i);
          if Editor <> nil then begin
            Result := QUB_BaselineSegEditor_PyType.CreateInstance;
            TQUB_BaselineSegEditor_Py( PythonToDelphi(Result) ).Seg := Editor;
          end else begin
            Result := ReturnNone;
          end;
        end;
      end else begin
        PyErr_SetString( PyExc_IndexError^, PChar('index out of range') );
        Result := nil;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.AddSeg( args : PPyObject ) : PPyObject;
var
  f, l      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'ii:addSeg(first, last)', [@f, @l] ) <> 0 then begin
      prBaseline.AddSeg(f, l);
      Result := ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.DeleteData( args : PPyObject ) : PPyObject;
var
  f, l      :Integer;
  Trouble   :Boolean;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'ii:deleteData(first, last)', [@f, @l] ) <> 0 then begin
      Trouble := prBaseline.DeleteData(f, l);
      if Trouble then begin
        Result := PyString_FromString( PChar('failed (busy)') );
      end else begin
        Result := ReturnNone;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.InsertPoints( args :PPyObject ) : PPyObject;
var
  before, points   :Integer;
  Trouble          :Boolean;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'ii:insertPoints(before, points)', [@before, @points] ) <> 0 then begin
      Trouble := prBaseline.InsertPoints(before, points);
      if Trouble then begin
        Result := PyString_FromString( PChar('failed (busy)') );
      end else begin
        Result := ReturnNone;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.BreakSegment( args :PPyObject ) : PPyObject;
var
  first            :Integer;
  Trouble          :Boolean;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'i:breakSegment(first)', [@first] ) <> 0 then begin
      Trouble := prBaseline.BreakSegment(first);
      if Trouble then begin
        Result := PyString_FromString( PChar('failed (busy)') );
      end else begin
        Result := ReturnNone;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.OffsetValues( args :PPyObject ) : PPyObject;
var
  ADUnits          :Integer;
  Trouble          :Boolean;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'i:offsetValues(adUnits)', [@ADUnits] ) <> 0 then begin
      Trouble := prBaseline.OffsetValues(ADUnits);
      if Trouble then begin
        Result := PyString_FromString( PChar('failed (busy)') );
      end else begin
        Result := ReturnNone;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.ClearNodes( args : PPyObject ) : PPyObject;
var
  f, l      :Integer;
  Trouble   :Boolean;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'ii:clearNodes(first, last)', [@f, @l] ) <> 0 then begin
      Trouble := prBaseline.ClearNodes(f, l);
      if Trouble then begin
        Result := PyString_FromString( PChar('failed (busy)') );
      end else begin
        Result := ReturnNone;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.AddNode( args : PPyObject ) : PPyObject;
var
  p, v       :Integer;
  Trouble    :Boolean;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'ii:addNode(point, value)', [@p, @v] ) <> 0 then begin
      Trouble := prBaseline.AddNode(p, v);
      if Trouble then begin
        Result := PyString_FromString( PChar('failed (busy)') );
      end else begin
        Result := ReturnNone;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Baseline_Py.AddNodes( args : PPyObject ) : PPyObject;
var
  DataSetPy    :PPyObject;
  Trouble      :Boolean;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if prBaseline = nil then begin
      Result := ReturnNone;
    end else if PyArg_ParseTuple( args, 'O:addNodes(BaselineSegments)', [@DataSetPy] ) <> 0 then begin
      if PyNode_IsNode( DataSetPy ) then begin
        Trouble := prBaseline.AddNodes( PyNode_ToNode( DataSetPy ) );
        if Trouble then begin
          Result := PyString_FromString( PChar('failed (busy)') );
        end else begin
          Result := ReturnNone;
        end;
      end else begin
        PyErr_SetString( PyExc_TypeError^, PChar('argument must be a qubtree.Node') );
        Result := nil;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;


procedure QUB_Baseline_Py_Init;
begin
  QUB_BaselineSeg_Node_PyType := TPythonType.Create(nil);
  QUB_BaselineSeg_Node_PyType.Engine := GetPythonEngine;
  QUB_BaselineSeg_Node_PyType.Module := PythonViewport.pmQUBTypes;
  QUB_BaselineSeg_Node_PyType.TypeName := 'BaselineSeg_Node';
  QUB_BaselineSeg_Node_PyType.Services.Basic := [bsGetAttr];
  QUB_BaselineSeg_Node_PyType.PyObjectClass := TQUB_BaselineSeg_Node_Py;
  QUB_BaselineSeg_Node_PyType.Initialize;

  QUB_BaselineSeg_PyType := TPythonType.Create(nil);
  QUB_BaselineSeg_PyType.Engine := GetPythonEngine;
  QUB_BaselineSeg_PyType.Module := PythonViewport.pmQUBTypes;
  QUB_BaselineSeg_PyType.TypeName := 'BaselineSeg';
  QUB_BaselineSeg_PyType.Services.Basic := [bsGetAttr];
  QUB_BaselineSeg_PyType.PyObjectClass := TQUB_BaselineSeg_Py;
  QUB_BaselineSeg_PyType.Initialize;

  QUB_BaselineSegEditor_PyType := TPythonType.Create(nil);
  QUB_BaselineSegEditor_PyType.Engine := GetPythonEngine;
  QUB_BaselineSegEditor_PyType.Module := PythonViewport.pmQUBTypes;
  QUB_BaselineSegEditor_PyType.TypeName := 'BaselineSegEditor';
  QUB_BaselineSegEditor_PyType.Services.Basic := [bsGetAttr, bsSetAttr];
  QUB_BaselineSegEditor_PyType.PyObjectClass := TQUB_BaselineSegEditor_Py;
  QUB_BaselineSegEditor_PyType.Initialize;

  QUB_Baseline_PyType := TPythonType.Create(nil);
  QUB_Baseline_PyType.Engine := GetPythonEngine;
  QUB_Baseline_PyType.Module := PythonViewport.pmQUBTypes;
  QUB_Baseline_PyType.TypeName := 'Baseline';
  QUB_Baseline_PyType.Services.Basic := [bsGetAttr];
  QUB_Baseline_PyType.PyObjectClass := TQUB_Baseline_Py;
  QUB_Baseline_PyType.Initialize;

end;



// *****************************************************************************
end.
