{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Breaking with convention, this has nothing to do with QUB_Baseline.
// This is actually the companion to QUB_Task_ApplyBaseline, which I have not heard of anyone using.


unit QUB_Baseline_Global;


interface


uses


AllocTypes,

QUB_QVO_QVD_Global,
QUB_Global,
QUB_Stats;


type


TQUB_Baseline_Params_CapTrans = record

  StimChannel                 :Integer;
  DeltaStim                   :Integer;
  FitDur                      :Double;
  CorrectDur                  :Double;

end;

TQUB_Baseline_Params_Adaptive = record

  Sense                       :Integer;
  WindowWidth                 :Integer;
  WindowHeight                :Double;
  MinDifference               :Integer;

end;

TQUB_Baseline_Params_Linear   = record

  RegressionSegment           :Integer;
  ExtensionSegment            :Integer;
  VarianceConfidence          :Double;
  CoincidenceConfidence       :Double;
  OutliersTolerance           :Double;
  MaxBaselineShift            :Double;
  MaxSlope                    :Double;
  RegressionAdvance           :Integer;
  MinBaseline                 :Integer;
  MinSegment                  :Integer;
  ThirdOrderInterpolation     :Boolean;

end;

TQUB_Baseline_Params_Init     = record

  N                           :Integer;
  Avg                         :Double;
  Std                         :Double;
  xAvg                        :Double;
  yStdR                       :Double;
  Slope                       :Double;
  Intercept                   :Double;
  SlopeStd                    :Double;
  yVar                        :Double;

end;


TQUB_Baseline_Params          = record

  Method                      :Integer;

  CapTrans                    :TQUB_Baseline_Params_CapTrans;
  Adaptive                    :TQUB_Baseline_Params_Adaptive;
  Linear                      :TQUB_Baseline_Params_Linear;

  Init                        :TQUB_Baseline_Params_Init;

  StartPoint                  :Integer;
  DCOffset                    :Integer;
  CorrectDCOffset             :Boolean;

  Sampling                    :Double;
  Scaling                     :Double;

end;
TQUB_Baseline_PParams         = ^TQUB_Baseline_Params;

TQUB_Baseline_Segment         = record
  Sel                         :TQUB_FirstLast;
  xAvg                        :Double;
  yAvg                        :Double;
  yStdR                       :Double;
  Slope                       :Double;
  Intercept                   :Double;
  Accept                      :Boolean;
end;
TQUB_Baseline_PSegment        = ^TQUB_Baseline_Segment;


const


METHOD_ADAPTIVE               = 0;
METHOD_LINEAR                 = 1;
METHOD_CAPTRANS               = 2;
DEF_METHOD                    = METHOD_ADAPTIVE;

MIN_WINDOW_WIDTH              = 10;
MAX_WINDOW_WIDTH              = 100000;
DEF_WINDOW_WIDTH              = 100;

MIN_WINDOW_HEIGHT            :Double = 0.1;
MAX_WINDOW_HEIGHT            :Double = 10;
DEF_WINDOW_HEIGHT            :Double = 2;

MIN_MIN_DIFFERENCE            = 1;
MAX_MIN_DIFFERENCE            = 32767;
DEF_MIN_DIFFERENCE            = 10;

MIN_REGRESSIONSEGMENT         = 10;
MAX_REGRESSIONSEGMENT         = ONE_MEGABYTE;
DEF_REGRESSIONSEGMENT         = 1000;

MIN_VARIANCECONFIDENCE       :Double = 0.1;
MAX_VARIANCECONFIDENCE       :Double = 1;
DEF_VARIANCECONFIDENCE       :Double = 0.95;

MIN_COINCIDENCECONFIDENCE    :Double = 0.1;
MAX_COINCIDENCECONFIDENCE    :Double = 1;
DEF_COINCIDENCECONFIDENCE    :Double = 0.995;

MIN_OUTLIERSTOLERANCE        :Double = 0.0;
MAX_OUTLIERSTOLERANCE        :Double = 100.0;
DEF_OUTLIERSTOLERANCE        :Double = 3;

MIN_REGRESSIONADVANCE         = 0;
MAX_REGRESSIONADVANCE         = 100;
DEF_REGRESSIONADVANCE         = 0;

MIN_MINBASELINE               = 4;
MAX_MINBASELINE               = ONE_MEGABYTE;
DEF_MINBASELINE               = 200;

MIN_MAXBASELINESHIFT         :Double = 0.0;
MAX_MAXBASELINESHIFT         :Double = 10.0;
DEF_MAXBASELINESHIFT         :Double = 0.1;

MIN_MAXSLOPE                 :Double = 0.0;
MAX_MAXSLOPE                 :Double = 1000.0;
DEF_MAXSLOPE                 :Double = 1.0;

MIN_MINSEGMENT                = 4;
MAX_MINSEGMENT                = ONE_MEGABYTE;
DEF_MINSEGMENT                = 50;

MIN_EXTENSIONSEGMENT          = 4;
MAX_EXTENSIONSEGMENT          = 100000;
DEF_EXTENSIONSEGMENT          = 20;

DEF_CORRECTDCOFFSET           = 0;

DEF_THIRDORDERINTERPOLATION   = 0;

MIN_DELTASTIM                 = 1;
MAX_DELTASTIM                 = 65536;
DEF_DELTASTIM                 = 10;

DEF_STIMCHANNEL               = 1;

MIN_FITDUR                   :Double = 1e-10;
DEF_FITDUR                   :Double = 0.01;

MIN_CORRECTDUR               :Double = 1e-10;
DEF_CORRECTDUR               :Double = 0.05;


procedure CalculateWindow(Buffer :TPAInteger; Sel :TQUB_FirstLast; var xAvg, yAvg, yStd, yStdR, Slope, Intercept, SlopeStd :Double; UseVariance :Boolean; var Vyx :Double); forward;
procedure CalculateWindowAvgStd(Buffer :TPAInteger; Sel :TQUB_FirstLast; var Avg, Std :Double); forward;
procedure CalculateAdaptiveWindowAvg(Buffer :TPAInteger; Sel :TQUB_FirstLast; Min, Max :Double; var Avg, Sum :Double; var N :Integer); forward;
procedure CalculateVariance(Buffer :TPAInteger; Sel :TQUB_FirstLast; var xAvg, yAvg, yStdR, Slope, Intercept :Double); forward;
function HasOutliers(Buffer :TPAInteger; Sel :TQUB_FirstLast; xAvg, yAvg, Slope, StdR :Double; var p :Integer; var dy :Double) :Boolean; forward;
function SegsCoincide(Buffer_P, Buffer_N :TPAInteger; Sel_P, Sel_N :TQUB_FirstLast; Confidence :Double; PopStd :Double) :Boolean; forward;
function MSRatioFTest(MSRatio :Double; DF1, DF2 :Integer; Alpha{Confidence} :Double) :Boolean; forward;


implementation


//******************************************************************************
procedure CalculateWindow(Buffer :TPAInteger; Sel :TQUB_FirstLast; var xAvg, yAvg, yStd, yStdR, Slope, Intercept, SlopeStd :Double; UseVariance :Boolean; var Vyx :Double);
const
  AllowedExcess = 0;
var
  N             :Integer;
  i             :Integer;
  Syy           :Double;
  Sxx           :Double;
  Sxy           :Double;
  Syf           :Double;
  Min, Max      :Double;
  AdjustSlope   :Boolean;
  Vsl           :Double;
  Iter          :Integer;
  //t             :Double;
begin

  N             := Sel.Last - Sel.First + 1;

  //xAvg, yAvg .................................................................
  xAvg          := 0;
  yAvg          := 0;
  for i := 0 to N - 1 do begin
    xAvg  := xAvg  + i;
    yAvg  := yAvg  + Buffer[i];
  end;
  if N <> 0 then begin
    xAvg  := xAvg / N;
    yAvg  := yAvg / N;
  end;

  //Sxx, Syy, Sxy ..............................................................
  Syy := 0;
  Sxx := 0;
  Sxy := 0;
  for i := 0 to N - 1 do begin
    Syy := Syy + (Buffer[i] - yAvg) * (Buffer[i] - yAvg);
    Sxx := Sxx + (i         - xAvg) * (i         - xAvg);
    Sxy := Sxy + (i         - xAvg) * (Buffer[i] - yAvg);
  end;

  //Slope, Intercept ...........................................................
  Slope     := Sxy / Sxx;
  Intercept := yAvg - Slope * xAvg;

  //Syf ........................................................................
  Syf := 0;
  for i := 0 to N - 1 do begin
    Syf := Syf + (Buffer[i] - (Slope * i + Intercept)) * (Buffer[i] - (Slope * i + Intercept));
  end;

  AdjustSlope := False;

  if not UseVariance then begin

    Vyx := Syf / (N - 1);

  end else begin


    //If Slope can be zero, then make it zero

    //If Slope cannot be zero, adjust it such that the new Slope is within accepted limits


    Vsl := Syy / (N - 1);

    if (N - 1) <= 700 then begin

      Min := Sqrt(Chi2(0.95, N - 1) / (N - 1));

    end else begin

      Min := 0.95;

    end;


    //Test the Hypothesis that Slope = 0

    //New way
    {
    SlopeStd := Sqrt((Syf / Sxx) * (1 / (N - 2)));
    t := Student(0.9, N - 2);
    if (Slope / SlopeStd) < t then begin

      Slope := 0;

    end else begin

      AdjustSlope := True;

    end;
    }

    //Old way
    if (Vsl < (Vyx / Min)) {and (Vsl > Vyx / Max)} then begin

      Slope := 0;

    end else begin

      AdjustSlope := True;

    end;


    if AdjustSlope then begin

      if (Syf / (N - 1)) < (Vyx / Min) then begin

        Min := 0;
        Max := Slope;
        Vsl := Syf / (N - 1);
        Iter := 0;
        while (Vsl < (Vyx - 0.1 * Vyx)) or (Vsl > (Vyx + 0.1 * Vyx)) do begin

          Slope := (Min + Max) / 2;
          Intercept := yAvg - Slope * xAvg;

          Vsl := 0;
          for i := 0 to N - 1 do begin
            Vsl := Vsl + (Buffer[i] - (Slope * i + Intercept)) * (Buffer[i] - (Slope * i + Intercept));
          end;
          Vsl := Vsl / (N - 1);

          if Vsl < Vyx then begin
            Max := Slope;
          end else begin
            Min := Slope;
          end;

          Inc(Iter);
          if Iter > 100 then Break;

        end;

      end;

    end;
    //*)
  end;

  Intercept := yAvg - Slope * xAvg;


  //Std, StdR, SlopeStd ........................................................
  yStd      := 0;
  yStdR     := 0;
  if N > 1 then begin
    yStd    := Sqrt(Syy / (N - 1));
    yStdR   := Sqrt(Syf / (N - 1));
  end;

  //SlopeStd ...................................................................
  SlopeStd := 0;
  if N > 2 then begin
    SlopeStd  := Sqrt((N / (N - 2)) * (Syf / Sxx));//Sqrt((Syf / Sxx) / (N - 2));
  end;

  //Adjust Intercept ?
  //Intercept := Intercept + Slope*Sel.First;

end;

procedure CalculateWindowAvgStd(Buffer :TPAInteger; Sel :TQUB_FirstLast; var Avg, Std :Double);
var
  yAvg  :Double;
  yStd  :Double;
  N     :Integer;
  i     :Integer;
begin

  N     := Sel.Last - Sel.First + 1;

  yAvg  := 0;
  for i := 0 to N - 1 do begin
    yAvg := yAvg + Buffer[i];
  end;
  Avg := yAvg / N;

  yStd := 0;
  for i := 0 to N - 1 do begin
    yStd := yStd + (Buffer[i] - Avg) * (Buffer[i] - Avg);
  end;
  if N < 2 then Std := 0 else Std := Sqrt(yStd / (N - 1));

end;

procedure CalculateAdaptiveWindowAvg(Buffer :TPAInteger; Sel :TQUB_FirstLast; Min, Max :Double; var Avg, Sum :Double; var N :Integer); 
var
  i     :Integer;
begin
  N     := 0;
  Sum   := 0;
  for i := 0 to (Sel.Last - Sel.First) do begin
    if (Min <= Buffer[i]) and (Max >= Buffer[i]) then begin
      Inc(N);
      Sum := Sum + Buffer[i];
    end;
  end;
  if N <> 0 then Avg := Sum / N else Avg := 0;
end;

procedure CalculateVariance(Buffer :TPAInteger; Sel :TQUB_FirstLast; var xAvg, yAvg, yStdR, Slope, Intercept :Double);
var
  Syy    :Double;
  Sxx    :Double;
  Sxy    :Double;
  Syf    :Double;
  b1     :Double;
  b0     :Double;
  N      :Integer;
  i      :Integer;
begin

  N := Sel.Last - Sel.First + 1;

  //xAvg, yAvg .................................................................
  xAvg          := 0;
  yAvg          := 0;
  for i := 0 to N - 1 do begin
    xAvg  := xAvg  + (i + Sel.First);
    yAvg  := yAvg  + Buffer[i];
  end;
  if N <> 0 then begin
    xAvg  := xAvg / N;
    yAvg  := yAvg / N;
  end;

  //Sxx, Syy, Sxy ..............................................................
  Syy := 0;
  Sxx := 0;
  Sxy := 0;
  for i := 0 to N - 1 do begin
    Syy := Syy + (Buffer[i]       - yAvg) * (Buffer[i]       - yAvg);
    Sxx := Sxx + ((i + Sel.First) - xAvg) * ((i + Sel.First) - xAvg);
    Sxy := Sxy + ((i + Sel.First) - xAvg) * (Buffer[i]       - yAvg);
  end;

  //Slope, Intercept ...........................................................
  b1 := Sxy / Sxx;
  b0 := yAvg - b1 * xAvg;

  //Syf ........................................................................
  Syf := 0;
  for i := 0 to N - 1 do begin
    Syf := Syf + (Buffer[i] - (b1 * (i + Sel.First) + b0)) * (Buffer[i] - (b1 * (i + Sel.First) + b0));
  end;

  //If (unlikely but...) b1, b0 is worse than b0 model, use b0 model
  if Syf > Syy then begin
    b1  := 0;
    b0  := yAvg;
    Syf := Syy;
  end;
  //............................................................................

  yStdR     := Sqrt(Syf / (N - 1));
  Intercept := b0;
  Slope     := b1;

end;

function HasOutliers(Buffer :TPAInteger; Sel :TQUB_FirstLast; xAvg, yAvg, Slope, StdR :Double; var p :Integer; var dy :Double) :Boolean; 
var
  N      :Integer;
  i      :Integer;
  yPred  :Double;
  Max    :Double;
  pMax   :Integer;
begin
  Result := False;
  N      := Sel.Last - Sel.First + 1;
  Max    := 0;
  pMax   := 0;
  for i := 0 to N - 1 do begin
    yPred := yAvg + Slope * (i + Sel.First - xAvg);
    dy := Abs(Buffer[i] - yPred);
    p := i + Sel.First;
    if dy > StdR then begin
      Result := True;
      Exit;
    end else begin
      if dy > Max then begin
        Max  := dy;
        pMax := p;
      end;
    end;
  end;
  dy := Max;
  p  := pMax;
end;

function SegsCoincide(Buffer_P, Buffer_N :TPAInteger; Sel_P, Sel_N :TQUB_FirstLast; Confidence :Double; PopStd :Double) :Boolean;
var
  N_P, N_N        :Integer;
  Syy_P, Syy_N    :Double;
  Sxx_P, Sxx_N    :Double;
  Sxy_P, Sxy_N    :Double;
  Syf_P, Syf_N    :Double;
  xAvg_P, xAvg_N  :Double;
  yAvg_P, yAvg_N  :Double;
  yAvg            :Double;
  b1_P, b1_N      :Double;
  b0_P, b0_N      :Double;
  b0, b1, b2, b3  :Double;
  SSY_X           :Double;
  SSY_X_Z_XZ      :Double;
  SSE_X           :Double;
  SSE_X_Z_XZ      :Double;
  SS_X            :Double;
  SS_X_Z_XZ       :Double;
  Sb_PN           :Double;
  Tb1             :Double;
  Tb0             :Double;
  dy_N            :Double;
  F               :Double;
  i               :Integer;
begin

  N_P := Sel_P.Last - Sel_P.First + 1;
  N_N := Sel_N.Last - Sel_N.First + 1;

  //xAvg, yAvg .................................................................
  xAvg_P          := 0;
  yAvg_P          := 0;
  xAvg_N          := 0;
  yAvg_N          := 0;
  for i := 0 to N_P - 1 do begin
    xAvg_P  := xAvg_P  + (i + Sel_P.First);
    yAvg_P  := yAvg_P  + Buffer_P[i];
  end;
  for i := 0 to N_N - 1 do begin
    xAvg_N  := xAvg_N  + (i + Sel_N.First);
    yAvg_N  := yAvg_N  + Buffer_N[i];
  end;
  yAvg := (yAvg_P + yAvg_N) / (N_N + N_P);
  if N_P <> 0 then begin
    xAvg_P  := xAvg_P / N_P;
    yAvg_P  := yAvg_P / N_P;
  end;
  if N_N <> 0 then begin
    xAvg_N  := xAvg_N / N_N;
    yAvg_N  := yAvg_N / N_N;
  end;

  //Sxx, Syy, Sxy ..............................................................
  Syy_P := 0;
  Sxx_P := 0;
  Sxy_P := 0;
  Syy_N := 0;
  Sxx_N := 0;
  Sxy_N := 0;
  for i := 0 to N_P - 1 do begin
    Syy_P := Syy_P + (Buffer_P[i]       - yAvg_P) * (Buffer_P[i]       - yAvg_P);
    Sxx_P := Sxx_P + ((i + Sel_P.First) - xAvg_P) * ((i + Sel_P.First) - xAvg_P);
    Sxy_P := Sxy_P + ((i + Sel_P.First) - xAvg_P) * (Buffer_P[i]       - yAvg_P);
  end;
  for i := 0 to N_N - 1 do begin
    Syy_N := Syy_N + (Buffer_N[i]       - yAvg_N) * (Buffer_N[i]       - yAvg_N);
    Sxx_N := Sxx_N + ((i + Sel_N.First) - xAvg_N) * ((i + Sel_N.First) - xAvg_N);
    Sxy_N := Sxy_N + ((i + Sel_N.First) - xAvg_N) * (Buffer_N[i]       - yAvg_N);
  end;

  //Slope, Intercept ...........................................................
  b1_P := Sxy_P / Sxx_P;
  b0_P := yAvg_P - b1_P * xAvg_P;
  b1_N := Sxy_N / Sxx_N;
  b0_N := yAvg_N - b1_N * xAvg_N;

  //Syf ........................................................................
  Syf_P := 0;
  Syf_N := 0;
  for i := 0 to N_P - 1 do begin
    Syf_P := Syf_P + (Buffer_P[i] - (b1_P * (i + Sel_P.First) + b0_P)) * (Buffer_P[i] - (b1_P * (i + Sel_P.First) + b0_P));
  end;
  for i := 0 to N_N - 1 do begin
    Syf_N := Syf_N + (Buffer_N[i] - (b1_N * (i + Sel_N.First) + b0_N)) * (Buffer_N[i] - (b1_N * (i + Sel_N.First) + b0_N));
  end;

  //If (unlikely but...) b1, b0 is worse than b0 model, use b0 model
  if Syf_P > Syy_P then begin
    b1_P  := 0;
    b0_P  := yAvg_P;
    Syf_P := Syy_P;
  end;
  if Syf_N > Syy_N then begin
    b1_N  := 0;
    b0_N  := yAvg_N;
    Syf_N := Syy_N;
  end;
  //............................................................................

  b0 := b0_P;
  b1 := b1_P;
  b2 := b0_N - b0_P;
  b3 := b1_N - b1_P;

  SSE_X := 0;
  for i := 0 to N_N - 1 do begin
    SSE_X := SSE_X + (Buffer_N[i] - (b0 + b1*(i + Sel_N.First))) * (Buffer_N[i] - (b0 + b1*(i + Sel_N.First)));
  end;

  SSE_X_Z_XZ := Syf_P + Syf_N;

  F := ((SSE_X - Syf_N) / 2) / (((N_P - 1)*PopStd*PopStd + Syf_N) / (N_P + N_N - 4));

  //Some correction, although empirical and approximative
  F := Sqrt(F);

  Result := MSRatioFTest(F, 2, N_P + N_N - 4, Confidence);

end;

function MSRatioFTest(MSRatio :Double; DF1, DF2 :Integer; Alpha{Confidence} :Double) :Boolean;
var
  c, c1, c2      :Integer;
  a1, a2         :Double;
  F_1, F_2       :Double;
  F              :Double;
begin

  Result := False;
  if (DF1 < 1) or (DF1 > 2) then Exit;

  if DF2 < 1   then DF2 := 1;
  if DF2 > 700 then DF2 := 700;

  //Find the nearest Alpha in the table, in order to extrapolate, if necessary
  c1 := 1;
  c2 := 1;
  a1 := FTable[DF1, 0, c1];
  a2 := FTable[DF1, 0, c2];
  for c := 21 downto 1 do begin
    //Fucking truncation problem to be solved !!!
    if Abs(FTable[DF1, 0, c] - Alpha) < 1e-10 then begin
      c2 := c;
      c1 := c + 1;
      if c1 > 21 then c1 := c;
      a1 := FTable[DF1, 0, c1];
      a2 := FTable[DF1, 0, c2];
      Break;
    end;
  end;

  F_1 := FTable[DF1, DF2, c1];
  F_2 := FTable[DF1, DF2, c2];
  if a2 = a1 then begin
    F := F_1;
  end else begin
    F  := F_1 + (((F_2 - F_1) * (Alpha - a1)) / (a2 - a1));
  end;

  Result := F > MSRatio;

end;


//******************************************************************************
end.
