{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2002
// a palette of colors, one of which is visibly chosen.

unit QUB_ColorPanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus,
  ExtCtrls;


type TQUB_ColorPanel_Orient = (QUB_COLORPANEL_HORIZ, QUB_COLORPANEL_VERT);


type TQUB_ColorPanel = class(TPanel)

protected

  prOrientation                :TQUB_ColorPanel_Orient;
  prColorCount                 :Integer;
  prColorWidth                 :Integer; // to preserve across resizes
  prPanels                     :array of TPanel;
  prPopupMenu                  :TPopupMenu;
  prChosen                     :Integer;

  procedure prOnClick(Sender :TObject);
  procedure prOnDblClick(Sender :TObject);
  procedure prOnResize(Sender :TObject);

  procedure prSetOrientation(Orient :TQUB_ColorPanel_Orient);
  procedure prSetColorWidth(CWidth :Integer);
  procedure prSetPopupMenu(Popup :TPopupMenu);
  procedure prSetChosen(Choice :Integer);

public

  OnColorClick                 :procedure (Sender :TObject; Color :Integer) of object;
  OnColorDblClick              :procedure (Sender :TObject; Color :Integer) of object;

  constructor CreateCounted(AOwner :TComponent; ColorCount :Integer);

  property Orientation         :TQUB_ColorPanel_Orient      read prOrientation   write prSetOrientation;
  property ColorCount          :Integer                     read prColorCount;
  property ColorWidth          :Integer                     read prColorWidth    write prSetColorWidth;
  property ColorPopupMenu      :TPopupMenu                  read prPopupMenu     write prSetPopupMenu;
  property Chosen              :Integer                     read prChosen        write prSetChosen;

end;


procedure Register;


implementation

uses

QUB_QFS_Channels;


procedure TQUB_ColorPanel.prOnClick(Sender :TObject);
begin
  Chosen     := TPanel(Sender).Tag;
  if Assigned(OnColorClick) then OnColorClick(Self, Chosen);
end;

procedure TQUB_ColorPanel.prOnDblClick(Sender :TObject);
begin
  Chosen     := TPanel(Sender).Tag;
  if Assigned(OnColorDblClick) then OnColorDblClick(Self, Chosen);
end;

procedure TQUB_ColorPanel.prOnResize(Sender :TObject);
var
  i                       :Integer;
  ColHeight               :Integer;
begin
  if Orientation = QUB_COLORPANEL_HORIZ then begin

    Height                     := ColorWidth;
    ColHeight                  := Width div prColorCount;
    for i := 0 to prColorCount - 1 do begin
      prPanels[i].Left         := (i * Width) div prColorCount;
      prPanels[i].Width        := ColHeight;
      prPanels[i].Top          := 0;
      prPanels[i].Height       := Height;
    end;

  end else begin

    Width                      := ColorWidth;
    ColHeight                  := Height div prColorCount;
    for i := 0 to prColorCount - 1 do begin
      prPanels[i].Top          := (i * Height) div prColorCount;
      prPanels[i].Height       := ColHeight;
      prPanels[i].Left         := 0;
      prPanels[i].Width        := Width;
    end;

  end;
end;

procedure TQUB_ColorPanel.prSetOrientation(Orient :TQUB_ColorPanel_Orient);
begin
  if Orient <> prOrientation then begin
    prOrientation := Orient;
    prOnResize(Self);
  end;
end;

procedure TQUB_ColorPanel.prSetColorWidth(CWidth :Integer);
begin
  prColorWidth := CWidth;
  prOnResize(Self);
end;

procedure TQUB_ColorPanel.prSetPopupMenu(Popup :TPopupMenu);
var
  i           :Integer;
begin
  prPopupMenu := Popup;

  for i := 0 to prColorCount - 1 do begin
     prPanels[i].PopupMenu := Popup;
  end;
end;

procedure TQUB_ColorPanel.prSetChosen(Choice :Integer);
begin
  prPanels[Chosen].BevelOuter      := bvRaised;
  prChosen                         := Choice;
  prPanels[Chosen].BevelOuter      := bvLowered;
end;

constructor TQUB_ColorPanel.CreateCounted(AOwner :TComponent; ColorCount :Integer);
var
  i                          :Integer;
begin
  inherited Create(AOwner);

  prColorCount               := ColorCount;
  prChosen                   := 0;

  SetLength( prPanels, prColorCount );
  for i := 0 to prColorCount - 1 do begin
    prPanels[i]              := TPanel.Create(Self);
    prPanels[i].Parent       := Self;
    prPanels[i].Color        := QUB_GetClassColor(i);
    prPanels[i].BevelWidth   := 3;
    prPanels[i].BevelOuter   := bvRaised;
    prPanels[i].Tag          := i;
    prPanels[i].OnClick      := prOnClick;
    prPanels[i].OnDblClick   := prOnDblClick;
  end;

  OnResize                   := prOnResize;

  Chosen                     := Chosen;
  prColorWidth               := 20;
  Orientation                := QUB_COLORPANEL_HORIZ;
end;

// *****************************************************************************

procedure Register;
begin
  RegisterComponents('QUB', [TQUB_ColorPanel]);
end;

end.
