unit QUB_Cuda;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses

Windows, SysUtils, Classes, Registry, Contnrs, Controls, Math,
AllocTypes, Alloc;


type

TQUB_Cuda_init                   = procedure; cdecl;
TQUB_Cuda_done                   = procedure; cdecl;
TQUB_Cuda_dAdq_context_create    = function(Nstim, Ns, Nq :Integer): Pointer; cdecl;
TQUB_Cuda_dAdq_context_destroy   = procedure(Ctx :Pointer); cdecl;
TQUB_Cuda_dAdq_context_set_dQdq  = procedure(Ctx, dQdq :Pointer); cdecl;
TQUB_Cuda_dAdq_context_set_Aek   = procedure(Ctx, Ak :Pointer); cdecl;
TQUB_Cuda_dAdq_context_set_gklt  = procedure(Ctx, gklt :Pointer); cdecl;
TQUB_Cuda_dAdq_context_get_dAdq  = procedure(Ctx, dAdq :Pointer; ec :Integer); cdecl;

TQUB_Cuda_Lib = class
public
  Avail         :Boolean;
  LibHandle     :HINST;

  cuda_init     :TQUB_Cuda_init;
  cuda_done     :TQUB_Cuda_done;
  dAdq_create   :TQUB_Cuda_dAdq_context_create;
  dAdq_destroy  :TQUB_Cuda_dAdq_context_destroy;
  dAdq_set_dQdq :TQUB_Cuda_dAdq_context_set_dQdq;
  dAdq_set_Aek  :TQUB_Cuda_dAdq_context_set_Aek;
  dAdq_set_gklt :TQUB_Cuda_dAdq_context_set_gklt;
  dAdq_get_dAdq :TQUB_Cuda_dAdq_context_get_dAdq;

  constructor Create;
  destructor Destroy; override;
end;

TQUB_Cuda_dAdq = class
public
  Lib           :TQUB_Cuda_Lib;
  Context       :Pointer;

  constructor Create(Lib :TQUB_Cuda_Lib; Nstim, Ns, Nq :Integer);
  destructor Destroy; override;

  procedure Set_dQdq(dQdq :Pointer); // [qq][i][j] :Double
  procedure Set_Aek(Aek :Pointer);    // [ec][k][i][j]  :Double
  procedure Set_gklt(gklt :TPMDouble);
  procedure Get_dAdq(dAdq :Pointer; ec :Integer); // [qq][i][j] :Double
end;


implementation


constructor TQUB_Cuda_Lib.Create;
begin
  Avail      := False;
  LibHandle  := LoadLibrary(PChar('qubcuda.dll'));
  if LibHandle = 0 then Exit;

  cuda_init       := GetProcAddress(LibHandle, 'qubcuda_init');
  cuda_done       := GetProcAddress(LibHandle, 'qubcuda_done');
  dAdq_create     := GetProcAddress(LibHandle, 'dAdq_context_create');
  dAdq_destroy    := GetProcAddress(LibHandle, 'dAdq_context_destroy');
  dAdq_set_dQdq   := GetProcAddress(LibHandle, 'dAdq_context_set_dQdq');
  dAdq_set_Aek    := GetProcAddress(LibHandle, 'dAdq_context_set_Aek');
  dAdq_set_Gklt   := GetProcAddress(LibHandle, 'dAdq_context_set_gklt');
  dAdq_get_dAdq   := GetProcAddress(LibHandle, 'dAdq_context_get_dAdq');
  Avail      := Assigned(dAdq_get_dAdq);
  if Avail then cuda_init();
end;

destructor TQUB_Cuda_Lib.Destroy;
begin
  try
    if LibHandle <> 0 then begin
      cuda_done();
      FreeLibrary(LibHandle);
    end;
  finally
    inherited;
  end;
end;

constructor TQUB_Cuda_dAdq.Create(Lib :TQUB_Cuda_Lib; Nstim, Ns, Nq :Integer);
begin
  Self.Lib   := Lib;
  Context    := Lib.dAdq_create(Nstim, Ns, Nq);
end;

destructor TQUB_Cuda_dAdq.Destroy;
begin
  try
    if Context <> nil then begin
      Lib.dAdq_destroy(Context);
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_Cuda_dAdq.Set_dQdq(dQdq :Pointer); // [qq][i][j] :Double
begin
  if Context <> nil then Lib.dAdq_set_dQdq(Context, dQdq);
end;

procedure TQUB_Cuda_dAdq.Set_Aek(Aek :Pointer);    // [ec][k][i][j]  :Double
begin
  if Context <> nil then Lib.dAdq_set_Aek(Context, Aek);
end;

procedure TQUB_Cuda_dAdq.Set_gklt(gklt :TPMDouble);
begin
  if Context <> nil then Lib.dAdq_set_gklt(Context, gklt);
end;

procedure TQUB_Cuda_dAdq.Get_dAdq(dAdq :Pointer; ec :Integer); // [qq][i][j] :Double
begin
  if Context <> nil then Lib.dAdq_get_dAdq(Context, dAdq, ec);
end;



end.
