unit QUB_DFilter;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


function DFII32_New: Pointer;                    cdecl; external 'qubopt.dll' name 'DFII32_New';
procedure DFII32_Free(F :Pointer);               cdecl; external 'qubopt.dll' name 'DFII32_Free';
function DFII32_Init(FileName :PChar; F :Pointer; IC :Single): Integer;
                                                 cdecl; external 'qubopt.dll' name 'DFII32_Init';
function DFII32(X :Single; F :Pointer): Single;  cdecl; external 'qubopt.dll' name 'DFII32';
function DFIIT32(X :Single; F :Pointer): Single; cdecl; external 'qubopt.dll' name 'DFIIT32';


function DFII64_New: Pointer;                    cdecl; external 'qubopt.dll' name 'DFII64_New';
procedure DFII64_Free(F :Pointer);               cdecl; external 'qubopt.dll' name 'DFII64_Free';
function DFII64_Init(FileName :PChar; F :Pointer; IC :Double): Integer;
                                                 cdecl; external 'qubopt.dll' name 'DFII64_Init';
function DFII64(X :Double; F :Pointer): Double;  cdecl; external 'qubopt.dll' name 'DFII64';
function DFIIT64(X :Double; F :Pointer): Double; cdecl; external 'qubopt.dll' name 'DFIIT64';


type


TQUB_DFilter_Single = class
protected
  prObj    :Pointer;

public
  constructor Create;
  destructor Destroy; override;

  function Init(FileName :String; InitialVal :Single = 0.0): Boolean;
  function Next(X :Single): Single;
  function NextT(X :Single): Single;
end;


TQUB_DFilter_Double = class
protected
  prObj    :Pointer;

public
  constructor Create;
  destructor Destroy; override;

  function Init(FileName :String; InitialVal :Double = 0.0): Boolean;
  function Next(X :Double): Double;
  function NextT(X :Double): Double;
end;


implementation


constructor TQUB_DFilter_Single.Create;
begin
  prObj := DFII32_New;
end;

destructor TQUB_DFilter_Single.Destroy;
begin
  try
    DFII32_Free(prObj);
  finally
    inherited;
  end;
end;

function TQUB_DFilter_Single.Init(FileName :String; InitialVal :Single = 0.0): Boolean;
begin
  Result := 0<>DFII32_Init(PChar(FileName), prObj, InitialVal);
end;

function TQUB_DFilter_Single.Next(X :Single): Single;
begin
  Result := DFII32(X, prObj);
end;

function TQUB_DFilter_Single.NextT(X :Single): Single;
begin
  Result := DFIIT32(X, prObj);
end;


constructor TQUB_DFilter_Double.Create;
begin
  prObj := DFII64_New;
end;

destructor TQUB_DFilter_Double.Destroy;
begin
  try
    DFII64_Free(prObj);
  finally
    inherited;
  end;
end;

function TQUB_DFilter_Double.Init(FileName :String; InitialVal :Double = 0.0): Boolean;
begin
  Result := 0 <> DFII64_Init(PChar(FileName), prObj, InitialVal);
end;

function TQUB_DFilter_Double.Next(X :Double): Double;
begin
  Result := DFII64(X, prObj);
end;

function TQUB_DFilter_Double.NextT(X :Double): Double;
begin
  Result := DFIIT64(X, prObj);
end;



end.
