{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2003
// the file list visible in many task properties
// the checkboxes reflect QVO_QVD.Listed
// the var columns reflect stuff from QVO_QVD.ExpCond

unit QUB_DataList;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
Grids, QUB_Table, StdCtrls, ExtCtrls, Math,

QUB_QVO, QUB_QVO_QVD, QUB_QVO_QVM,
QUB_ListBox,
QUB_ExpCond_Editor;


const


COL_USED                          = 0;
COL_NAME                          = 1;
COL_CHANCOUNT                     = 2;
COL_VAR                           = 3;


type


TQUB_DataList                     = class(TFrame)


  panDataSource                   :TPanel;
  Label3                          :TLabel;
  chkDataSourceFileList           :TRadioButton;
  chkDataSourceSelection          :TRadioButton;
  chkDataSourceSelectionList      :TRadioButton;
  chkDataSourceWholeFile          :TRadioButton;
  panGrid                         :TPanel;


  procedure chkDataSourceClick(Sender: TObject);


protected


  grdFiles                        :TQUB_Table;
  grdExpCond                      :TQUB_ExpCond_Editor;

  prListObjects     :TQUB_ListBox;

  prVisData         :TList; // of qvd
  prVars            :TStringList;
  prVarDefs         :Array of Double;
  prVarsWCond       :TStringList;
  prVarDefsWCond    :Array of Double;
  prShowingFiles    :Boolean;

  prDataSourceAvail :Boolean;
  prFileListAvail   :Boolean;
  prForceList       :Boolean;


  procedure grdFilesSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
  procedure grdFilesDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
  procedure grdFilesMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

  function prGetDataCount: Integer;
  function prGetData(i :Integer): TQUB_QVO_QVD;
  function prGetVisCount: Integer;
  function prGetVisData(i :Integer): TQUB_QVO_QVD;

  procedure prSetDataSourceAvail(x :Boolean);
  procedure prSetFileListAvail(x :Boolean);
  procedure prSetForceList(x :Boolean);

  procedure prOnListedChanged(Sender :TObject);

  procedure prShowFile(ix :Integer);

  function prVarOfCol(Col :Integer; var IsCond :Boolean): String;


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure ShowFiles;
  procedure HideFiles;

  procedure ShowVar(const VarName, Units :String; DefVal :Double; WithCond :Boolean);
  procedure ShowModelVars(Model :TQUB_QVO_QVM; WithCond :Boolean);
  procedure HideVars;

  procedure HideAll;


  property DataCount           :Integer        read prGetDataCount;
  property Data[i :Integer]    :TQUB_QVO_QVD   read prGetData;
  property VisCount            :Integer        read prGetVisCount;
  property VisData[i :Integer] :TQUB_QVO_QVD   read prGetVisData;

  property DataSourceAvail     :Boolean        read prDataSourceAvail  write prSetDataSourceAvail;
  property FileListAvail       :Boolean        read prFileListAvail    write prSetFileListAvail;
  property ForceList           :Boolean        read prForceList        write prSetForceList;


end;



implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Viewport_Data,
QUB_ExpCond_Parameters,
QUB_Task, QUB_Main;


//TODO: check this whole file. LM.


//******************************************************************************
constructor TQUB_DataList.Create(AOwner :TComponent);
begin

  inherited;

  grdFiles                 := TQUB_Table.Create(Self);
  grdFiles.Parent          := panGrid;
  grdFiles.Align           := alClient;
  grdFiles.ColCount        := 2;
  grdFiles.DefaultColWidth := 75;
  grdFiles.DefaultRowHeight := 16;
  grdFiles.RowCount        := 2;
  grdFiles.Options         := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRangeSelect, goRowSizing, goColSizing, goThumbTracking];
  grdFiles.TabOrder        := 1;
  grdFiles.OnDrawCell      := grdFilesDrawCell;
  grdFiles.OnMouseDown     := grdFilesMouseDown;
  grdFiles.OnSetEditText   := grdFilesSetEditText;
  grdFiles.ColWidths[0]    := 34;
  grdFiles.ColWidths[1]    := 75;

  grdExpCond               := TQUB_ExpCond_Editor.Create(Self);
  grdExpCond.Parent        := panGrid;
  grdExpCond.Align         := alClient;

  prListObjects       := ViewportData.ListObjects;

  prVisData           := TList.Create;
  prVars              := TStringList.Create;
  prVarsWCond         := TStringList.Create;

  grdFiles.Cells[COL_USED, 0]      := 'Use';
  grdFiles.Cells[COL_NAME, 0]      := 'FileName';
  grdFiles.Cells[COL_CHANCOUNT, 0] := 'ChannelCount';

  prDataSourceAvail   := True;
  prFileListAvail     := True;
  prForceList         := False;
end;

destructor TQUB_DataList.Destroy;
begin
  try
    prVisData.Free;
    prVars.Free;
    prVarsWCond.Free;
  finally
    inherited;
  end;
end;


procedure TQUB_DataList.ShowFiles;
var
  i :Integer;
begin
  if (QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and not FileListAvail then begin
    QUB_MainForm.DataSource      := DATA_SOURCE_WHOLEFILE;
  end;

  chkDataSourceSelection.Checked := chkDataSourceSelection.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceSelectionList.Checked := chkDataSourceSelectionList.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceWholeFile.Checked := chkDataSourceWholeFile.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceFileList.Checked := chkDataSourceFileList.Tag = Integer(QUB_MainForm.DataSource);

  HideFiles;

  for i := 0 to DataCount - 1 do begin
    prShowFile(i);
  end;

  if prForceList or (QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) then begin
    grdFiles.BringToFront;
  end else if ViewportData.CurrData <> nil then begin
    ViewportData.CurrData.ExpCond.DelPar( ViewportData.CurrData.ExpCond.IndexOf('ChannelCount') );
    //prShowFile( prListObjects.ItemIndex );
    grdExpCond.BringToFront;
    grdExpCond.ExpCond := ViewportData.CurrData.ExpCond;
  end;

  prShowingFiles := True;
end;

procedure TQUB_DataList.prShowFile(ix :Integer);
var
  i,Base,cc:Integer;
  Dat      :TQUB_QVO_QVD;
  ARow     :Integer;
  Par      :TQUB_ExpCond_PParameter;
begin
  if (ix < 0) or (ix >= prListObjects.Items.Count) then Exit;
  Dat      := Data[ix];
  prVisData.Add( Dat );

  ARow     := prVisData.Count;
  grdFiles.RowCount := Max(2, 1 + ARow);
  grdFiles.Cells[ COL_NAME, ARow ] := ExtractFileName( Dat.FileName );

  cc       := 1;
  if ViewportModel.CurrModel <> nil then cc := ViewportModel.CurrModel.ModelObject.ChannelCount;
  Par      := Dat.ExpCond.ParameterByName['ChannelCount'];
  if Par = nil then begin
    Dat.ExpCond.AddPar('ChannelCount', cc, -1, False);
    Par    := Dat.ExpCond.ParameterByName['ChannelCount'];
  end;
  grdFiles.Cells[COL_CHANCOUNT, ARow] := IntToStr(Round(Par.Value));

  for i := 0 to prVars.Count - 1 do begin
    with Dat.ExpCond do begin
      Par    := ParameterByName[prVars[i]];
      if Par = nil then begin
        grdFiles.Cells[COL_VAR + i, ARow] := FloatToStr(prVarDefs[i]);
        AddPar(prVars[i], prVarDefs[i], -1, False);
      end else begin
        grdFiles.Cells[COL_VAR + i, ARow] := FloatToStr(Par.Value);
      end;
    end;
  end;

  Base := COL_VAR + prVars.Count;
  for i := 0 to prVarsWCond.Count - 1 do begin
    with Dat.ExpCond do begin
      Par    := ParameterByName[prVarsWCond[i]];
      if Par = nil then begin
        AddPar(prVarsWCond[i], prVarDefsWCond[i], -1, False);
        grdFiles.Cells[Base + 2*i, ARow] := FloatToStr(prVarDefsWCond[i]);
        grdFiles.Cells[Base + 2*i + 1, ARow] := FloatToStr(prVarDefsWCond[i]);
      end else begin
        grdFiles.Cells[Base + 2*i, ARow] := FloatToStr(Par.Cond_Value);
        grdFiles.Cells[Base + 2*i + 1, ARow] := FloatToStr(Par.Value);
      end;
    end;
  end;
end;

procedure TQUB_DataList.HideFiles;
begin
  grdFiles.RowCount := 2;
  prVisData.Clear;
  grdExpCond.ExpCond := nil;
  prShowingFiles    := False;
end;

procedure TQUB_DataList.ShowVar(const VarName, Units :String; DefVal :Double; WithCond :Boolean);
var
  ACol, i      :Integer;
  Par          :TQUB_ExpCond_PParameter;
  UnitsText    :String;
begin
  if (prVars.IndexOf(VarName) < 0) and (prVarsWCond.IndexOf(VarName) < 0) then begin
    if Units <> '' then UnitsText := ' (' + Units + ')' else UnitsText := '';

    if WithCond then begin

      prVarsWCond.Add(VarName);
      grdFiles.ColCount := COL_VAR + prVars.Count + 2*prVarsWCond.Count;
      ACol              := grdFiles.ColCount - 2;
      grdFiles.Cells[ACol, 0] := 'Cond.' + VarName + UnitsText;
      grdFiles.Cells[ACol+1, 0] := VarName + UnitsText;

      SetLength( prVarDefsWCond, prVarsWCond.Count );
      prVarDefsWCond[ prVarsWCond.Count - 1 ] := DefVal;

      for i := 0 to VisCount - 1 do begin
        with VisData[i].ExpCond do begin
          Par := ParameterByName[VarName];
          if Par = nil then begin
            AddPar(VarName, DefVal, -1, False);
            Par := ParameterByName[VarName];
          end;
          grdFiles.Cells[ACol, i+1] := FloatToStr(Par.Cond_Value);
          grdFiles.Cells[ACol+1, i+1] := FloatToStr(Par.Value);
        end;
      end;

    end else begin

      prVars.Add(VarName);
      ACol              := COL_VAR + prVars.Count - 1;
      grdFiles.InsertColumn(ACol);
      grdFiles.Cells[ACol, 0] := VarName + UnitsText;

      SetLength( prVarDefs, prVars.Count );
      prVarDefs[ prVars.Count - 1 ] := DefVal;

      for i := 0 to VisCount - 1 do begin
        with VisData[i].ExpCond do begin
          Par    := ParameterByName[VarName];
          if Par = nil then begin
            grdFiles.Cells[ACol, i+1] := FloatToStr(DefVal);
            AddPar(VarName, DefVal, -1, False);
          end else begin
            grdFiles.Cells[ACol, i+1] := FloatToStr(Par.Value);
          end;
        end;
      end;

    end;
  end;

  grdExpCond.ShowCond := WithCond;
end;

procedure TQUB_DataList.ShowModelVars(Model :TQUB_QVO_QVM; WithCond :Boolean);
var
 i        :Integer;
begin
  with Model.ModelObject.Rates do begin
    for i := 0 to RateCount - 1 do begin
      if Rate[i].P[1] then ShowVar(Rate[i].PName[1], '', 1.0, WithCond);
      if Rate[i].P[2] then ShowVar(Rate[i].PName[2], '', 1.0, WithCond);
      if Rate[i].Q[1] then ShowVar(Rate[i].QName[1], '', 1.0, WithCond);
      if Rate[i].Q[2] then ShowVar(Rate[i].QName[2], '', 1.0, WithCond);
    end;
  end;

  grdExpCond.ShowCond := WithCond;
end;

procedure TQUB_DataList.HideVars;
begin
  grdFiles.ColCount := COL_VAR;
  prVars.Clear;
  prVarsWCond.Clear;
  SetLength(prVarDefs, 0);
  SetLength(prVarDefsWCond, 0);
end;

procedure TQUB_DataList.HideAll;
begin
  HideFiles;
  HideVars;
end;

function TQUB_DataList.prVarOfCol(Col :Integer; var IsCond :Boolean): String;
begin
  if Col = COL_CHANCOUNT then begin
    Result := 'ChannelCount';
    Exit;
  end else if Col < COL_VAR then begin
    Result := '';
    Exit;
  end;

  Col := Col - COL_VAR;
  if Col < prVars.Count then begin
    IsCond := False;
    Result := prVars[Col];
    Exit;
  end;

  Col := Col - prVars.Count;
  IsCond := (Col mod 2) = 0;
  Col := Col div 2;
  if Col < prVarsWCond.Count then begin
    Result := prVarsWCond[Col];
  end else begin
    Result := '';
  end;
end;

function TQUB_DataList.prGetDataCount: Integer;
begin
  Result := prListObjects.Items.Count;
end;

function TQUB_DataList.prGetData(i :Integer): TQUB_QVO_QVD;
begin
  Result := TQUB_QVO_QVD( prListObjects[i] );
end;

function TQUB_DataList.prGetVisCount: Integer;
begin
  Result := prVisData.Count;
end;

function TQUB_DataList.prGetVisData(i :Integer): TQUB_QVO_QVD;
begin
  Result := TQUB_QVO_QVD( prVisData[i] );
end;

procedure TQUB_DataList.prSetDataSourceAvail(x :Boolean);
begin
  prDataSourceAvail     := x;
  panDataSource.Visible := x;
end;

procedure TQUB_DataList.prSetFileListAvail(x :Boolean);
begin
  prFileListAvail                := x;
  chkDataSourceFileList.Visible  := x;
  if (QUB_MainForm.DataSource = DATA_SOURCE_FILELIST) and not x then begin
    QUB_MainForm.DataSource      := DATA_SOURCE_WHOLEFILE;
    chkDataSourceWholeFile.Checked := True;
  end;
end;

procedure TQUB_DataList.prSetForceList(x :Boolean);
begin
  prForceList        := x;
  if x and prShowingFiles then ShowFiles;
end;

procedure TQUB_DataList.prOnListedChanged(Sender :TObject);
var
  i        :Integer;
begin
  for i := 0 to VisCount - 1 do begin
    if VisData[i] = Sender then begin
      VisData[i].Listed := not VisData[i].Listed;
      grdFiles.Col := COL_NAME;
      grdFiles.Cells[COL_USED, i+1] := '';
    end;
  end;
end;

procedure TQUB_DataList.grdFilesSetEditText(Sender: TObject; ACol,
  ARow: Integer; const Value: String);
var
  VarName      :String;
  IsCond       :Boolean;
begin
  if (Value = '') or (Value = '.') or (Value = '-') or (Value = '-.') then Exit;

  try
    VarName    := prVarOfCol( ACol, IsCond );
    if IsCond then begin
      VisData[ ARow - 1 ].ExpCond.ChangeParCond( VisData[ ARow - 1 ].ExpCond.IndexOf(VarName), StrToFloat(Value) );
    end else begin
      VisData[ ARow - 1 ].ExpCond.ChangePar( VisData[ ARow - 1 ].ExpCond.IndexOf(VarName), StrToFloat(Value) );
    end;
  except
  end;
end;

procedure TQUB_DataList.grdFilesDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  Data       :TQUB_QVO_QVD;
  r          :TRect;
begin
  if (ACol = COL_USED) and (ARow > 0) and (ARow <= prListObjects.Items.Count) then begin
    Data     := VisData[ ARow - 1 ];
    r        := Rect;
    InflateRect(r, 0, -((r.Bottom - r.Top + 1) - 15) div 2);
    if (Data <> nil) and Data.Listed then begin
      DrawFrameControl(grdFiles.Canvas.Handle, r, DFC_BUTTON, DFCS_BUTTONCHECK or DFCS_CHECKED);
    end else begin
      DrawFrameControl(grdFiles.Canvas.Handle, r, DFC_BUTTON, DFCS_BUTTONCHECK);
    end;
  end;
end;

procedure TQUB_DataList.grdFilesMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ACol, ARow :Integer;
  Data       :TQUB_QVO_QVD;
begin
  grdFiles.MouseToCell(X, Y, ACol, ARow);
  if (ACol = COL_USED) and (ARow > 0) and (ARow <= prListObjects.Items.Count) then begin
    Data     := VisData[ ARow - 1 ];
    if Data <> nil then begin
      Data.Listed := not Data.Listed;
      grdFiles.Col := COL_NAME;
      grdFiles.Cells[COL_USED, ARow] := '';
    end;
  end;
end;


procedure TQUB_DataList.chkDataSourceClick(Sender: TObject);
var
  PrevDataSource          :TQUB_Task_Data_Source;
begin
  PrevDataSource          := QUB_MainForm.DataSource;
  QUB_MainForm.DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);

  if prShowingFiles
  and (((PrevDataSource = DATA_SOURCE_FILELIST) and (QUB_MainForm.DataSource <> DATA_SOURCE_FILELIST))
  or   ((PrevDataSource <> DATA_SOURCE_FILELIST) and (QUB_MainForm.DataSource = DATA_SOURCE_FILELIST)))
  then begin
    ShowFiles;
  end;
end;

// *****************************************************************************
end.
