unit QUB_DockManager;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, Graphics, Controls,

QUB_Viewport;


type


TQUB_DockManager = class(TDockTree)
protected


  procedure PaintDockFrame(Canvas: TCanvas; Control: TControl; const ARect: TRect); override;


end;


implementation


//******************************************************************************
procedure TQUB_DockManager.PaintDockFrame(Canvas: TCanvas; Control: TControl; const ARect: TRect);
const
  GrabberSize = 12;
  procedure DrawCloseButton(Left, Top: Integer);
  begin
    DrawFrameControl(Canvas.Handle, Rect(Left, Top, Left + GrabberSize - 2, Top + GrabberSize - 2), DFC_CAPTION, DFCS_CAPTIONCLOSE);
  end;
  procedure DrawCaption(Left, Top, Right, Bottom: Integer);
  var
    Active  :Boolean;
    Caption :String;
  begin
    with Canvas do begin
      if Control is TQUB_Viewport then begin
        Active  := TQUB_Viewport(Control).Active;
        Caption := TQUB_Viewport(Control).Caption;
      end else begin
        Active := False;
        Caption := '';
      end;
      if Active then begin
        Brush.Color := clActiveCaption;
      end else begin
        Brush.Color := clInActiveCaption;
      end;
      Font.Color := clCaptionText;
      TextRect(Rect(Left, Top, Right, Bottom), Left, Top - 1, Caption);
    end;
  end;
begin
  with ARect do begin
    if DockSite.Align in [alTop, alBottom] then begin
      DrawCloseButton(Left + 1, Top + 1);
    end else if DockSite.Align in [alClient] then begin
      DrawCaption(Left, Top, Right, Top + GrabberSize);
      DrawCloseButton(Right - GrabberSize + 1, Top + 1);
    end;
  end;
end;


//******************************************************************************
end.
