// this is probably duplicated with tpmath1 (also by J Debord).
// dunno if it's specialized or modified at all

// This module is not used ... {JB}

{ **********************************************************************
*                           Unit EIGEN.PAS                           *
*                             Version 1.6                            *
*                    (c) J. Debord, January 2001                     *
**********************************************************************
        Procedures for computing eigenvalues and eigenvectors
**********************************************************************
References:
1) Borland's Numerical Methods Toolbox : Jacobi
2) 'Numerical Recipes' by Press et al. : EigenVals, RootPol
********************************************************************** }

unit QUB_Eigen;


interface


uses


FMath, QUB_Matrices;


function Jacobi(A :PMatrix; Lbound, Ubound, MaxIter :Integer; Tol :Float; V :PMatrix; Lambda :PVector) :Integer;
{ ----------------------------------------------------------------------
Eigenvalues and eigenvectors of a symmetric matrix by the iterative
method of Jacobi
----------------------------------------------------------------------
Input parameters  : A       = matrix
                    Lbound  = index of first matrix element
                    Ubound  = index of last matrix element
                    MaxIter = maximum number of iterations
                    Tol     = required precision
----------------------------------------------------------------------
Output parameters : V      = matrix of eigenvectors (stored by lines)
                    Lambda = eigenvalues in decreasing order
----------------------------------------------------------------------
Possible results  : MAT_OK
                    MAT_NON_CONV
----------------------------------------------------------------------
NB : 1. The eigenvectors are normalized, with their first component > 0
     2. This procedure destroys the original matrix A
---------------------------------------------------------------------- }

procedure Balance(A :PMatrix; Lbound, Ubound :Integer);
procedure ElmHes(A :PMatrix; Lbound, Ubound : Integer);


function EigenVals(A :PMatrix; Lbound, Ubound :Integer; Lambda_Re, Lambda_Im :PVector) :Integer;
{ ----------------------------------------------------------------------
Eigenvalues of a general square matrix
----------------------------------------------------------------------
Input parameters  : A      = matrix
                    Lbound = index of first matrix element
                    Ubound = index of last matrix element
----------------------------------------------------------------------
Output parameters : Lambda_Re = real part of eigenvalues
                    Lambda_Im = imaginary part of eigenvalues
----------------------------------------------------------------------
Possible results  : MAT_OK
                    MAT_NON_CONV
----------------------------------------------------------------------
NB : This procedure destroys the original matrix A
---------------------------------------------------------------------- }

function EigenVect(A :PMatrix; Lbound, Ubound :Integer; Lambda :Float; V :PVector) :Integer;
{ ----------------------------------------------------------------------
Computes the eigenvector associated to a real eigenvalue
----------------------------------------------------------------------
Input parameters  : A       = matrix
                    Lbound  = index of first matrix element
                    Ubound  = index of last matrix element
                    Lambda  = eigenvalue
----------------------------------------------------------------------
Output parameters : V       = eigenvector
----------------------------------------------------------------------
Possible results  : MAT_OK
                    MAT_SINGUL
                    MAT_NON_CONV
----------------------------------------------------------------------
NB : 1. The eigenvector is normalized, with its first component > 0
     2. The function returns only one eigenvector, even if the
        eigenvalue has a multiplicity greater than 1.
---------------------------------------------------------------------- }

procedure DivLargest(V :PVector; Lbound, Ubound :Integer; var Largest :Float);
{ ----------------------------------------------------------------------
Normalizes an eigenvector V by dividing by the element with the
largest absolute value
---------------------------------------------------------------------- }

function RootPol(Coef :PVector; Deg :Integer; X_Re, X_Im :PVector) :Integer;
{ ----------------------------------------------------------------------
Real and complex roots of a real polynomial by the method of the
companion matrix
----------------------------------------------------------------------
Input parameters  : Coef = coefficients of polynomial
                    Deg  = degree of polynomial
----------------------------------------------------------------------
Output parameters : X_Re = real parts of root (in increasing order)
                    X_Im = imaginary parts of root
----------------------------------------------------------------------
Possible results  : MAT_OK
                    MAT_NON_CONV
---------------------------------------------------------------------- }

implementation


function Jacobi(A :PMatrix; Lbound, Ubound, MaxIter :Integer; Tol :Float; V :PMatrix; Lambda :PVector) :Integer;
var
  SinTheta, CosTheta, TanTheta, Tan2Theta :Float;
  CosSqr, SinSqr, SinCos, SumSqrDiag      :Float;
  AII, AJJ, AIJ, AIK, AJK, VIK, VJK, D    :Float;
  I, J, K, Iter                           :Integer;
  Done                                    :Boolean;
begin
  Iter := 0;
  for I := Lbound to Ubound do
    for J := Lbound to Ubound do
      if I = J then
        V^[I]^[J] := 1.0
      else
        V^[I]^[J] := 0.0;

  repeat
    Iter := Succ(Iter);
    SumSqrDiag := 0.0;
    for I := Lbound to Ubound do
      SumSqrDiag := SumSqrDiag + Sqr(A^[I]^[I]);
    Done := True;

    for I := Lbound to Pred(Ubound) do
      for J := Succ(I) to Ubound do
        if Abs(A^[I]^[J]) > Tol * SumSqrDiag then
          begin
            Done := False;

            { Calculate rotation }
            D := A^[I]^[I] - A^[J]^[J];
            if Abs(D) > MACHEP then
              begin
                Tan2Theta := D / (2.0 * A^[I]^[J]);
                TanTheta := - Tan2Theta + Sgn(Tan2Theta) *
                Sqrt(1.0 + Sqr(Tan2Theta));
                CosTheta := 1.0 / Sqrt(1.0 + Sqr(TanTheta));
                SinTheta := CosTheta * TanTheta;
              end
            else
              begin
                CosTheta := SQRT2DIV2;                   { Sqrt(2)/2 }
                SinTheta := Sgn(A^[I]^[J]) * SQRT2DIV2;
              end;

            { Rotate matrix }
            CosSqr := Sqr(CosTheta);
            SinSqr := Sqr(SinTheta);
            SinCos := SinTheta * CosTheta;
            AII := A^[I]^[I] * CosSqr + 2.0 * A^[I]^[J] * SinCos
            + A^[J]^[J] * SinSqr;
            AJJ := A^[I]^[I] * SinSqr - 2.0 * A^[I]^[J] * SinCos
            + A^[J]^[J] * CosSqr;
            AIJ := (A^[J]^[J] - A^[I]^[I]) * SinCos
            + A^[I]^[J] * (CosSqr - SinSqr);
            for K := Lbound to Ubound do
              if not(K in [I, J]) then
                begin
                  AIK := A^[I]^[K] * CosTheta + A^[J]^[K] * SinTheta;
                  AJK := - A^[I]^[K] * SinTheta + A^[J]^[K] * CosTheta;
                  A^[I]^[K] := AIK;
                  A^[K]^[I] := AIK;
                  A^[J]^[K] := AJK;
                  A^[K]^[J] := AJK;
                end;
            A^[I]^[I] := AII;
            A^[J]^[J] := AJJ;
            A^[I]^[J] := AIJ;
            A^[J]^[I] := AIJ;

            { Rotate eigenvectors }
            for K := Lbound to Ubound do
              begin
                VIK := CosTheta * V^[I]^[K] + SinTheta * V^[J]^[K];
                VJK := - SinTheta * V^[I]^[K] + CosTheta * V^[J]^[K];
                V^[I]^[K] := VIK;
                V^[J]^[K] := VJK;
              end;
          end;
  until Done or (Iter > MaxIter);

  { The diagonal terms of the transformed matrix are the eigenvalues }
  for I := Lbound to Ubound do
    Lambda^[I] := A^[I]^[I];

  if Iter > MaxIter then
    begin
      Jacobi := MAT_NON_CONV;
      Exit;
    end;

  { Sort eigenvalues and eigenvectors }
  for I := Lbound to Pred(Ubound) do
    begin
      K := I;
      D := Lambda^[I];
      for J := Succ(I) to Ubound do
        if Lambda^[J] > D then
          begin
            K := J;
            D := Lambda^[J];
          end;
      FSwap(Lambda^[I], Lambda^[K]);
      SwapRows(I, K, V, Lbound, Ubound);
    end;

  { Make sure that the first component of each eigenvector is > 0 }
  for I := Lbound to Ubound do
    if V^[I]^[Lbound] < 0.0 then
      for J := Lbound to Ubound do
        V^[I]^[J] := - V^[I]^[J];

  Jacobi := MAT_OK;
end;


//******************************************************************************
procedure Balance(A :PMatrix; Lbound, Ubound :Integer);
{ Balances the matrix, ie reduces norm without affecting eigenvalues }
const
  RADIX = 2;  { Base used for machine computations }
var
  C, F, G, R, S, Sqrdx :Float;
  I, J, Last           :Integer;
begin

  Sqrdx := Sqr(RADIX);

  repeat

    Last := 1;
    for I := Lbound to Ubound do begin

      C := 0.0;
      R := 0.0;

      for J := Lbound to Ubound do begin
        if J <> I then begin
          C := C + Abs(A^[J]^[I]);
          R := R + Abs(A^[I]^[J]);
        end;
      end;

      if (C <> 0.0) and (R <> 0.0) then begin
        G := R / RADIX;
        F := 1.0;
        S := C + R;

        while C < G do begin
          F := F * RADIX;
          C := C * Sqrdx;
        end;

        G := R * RADIX;
        while C > G do begin
          F := F / RADIX;
          C := C / Sqrdx;
        end;

        if ((C + R) / F) < (0.95 * S) then begin
          Last := 0;
          G := 1.0 / F;
          for J := Lbound to Ubound do
            A^[I]^[J] := A^[I]^[J] * G;
          for J := Lbound to Ubound do
            A^[J]^[I] := A^[J]^[I] * F;
        end;

      end;
    end;

  until (Last <> 0);

end;


//******************************************************************************
procedure ElmHes(A :PMatrix; Lbound, Ubound : Integer);
{ Reduces the matrix to upper Hessenberg form by elimination }
var
  X, Y     :Float;
  I, J, M  :Integer;
begin
  for M := Succ(Lbound) to Pred(Ubound) do begin
    X := 0.0;
    I := M;

    for J := M to Ubound do begin
      if Abs(A^[J]^[M - 1]) > Abs(X) then begin
        X := A^[J]^[M - 1];
        I := J;
      end;
    end;

    if I <> M then begin
      for J := Pred(M) to Ubound do
        FSwap(A^[I]^[J], A^[M]^[J]);
      for J := Lbound to Ubound do
        FSwap(A^[J]^[I], A^[J]^[M]);
    end;

    if X <> 0.0 then begin
      for I := Succ(M) to Ubound do begin
        Y := A^[I]^[M - 1];
        if Y <> 0.0 then begin
          Y := Y / X;
          A^[I]^[M - 1] := Y;
          for J := M to Ubound do
            A^[I]^[J] := A^[I]^[J] - Y * A^[M]^[J];
          for J := Lbound to Ubound do
            A^[J]^[M] := A^[J]^[M] + Y * A^[J]^[I];
        end;
      end;
    end;

  end;

  for I := (Lbound + 2) to Ubound do
    for J := Lbound to (I - 2) do
      A^[I]^[J] := 0.0;

end;

function Hqr(A :PMatrix; Lbound, Ubound :Integer; Lambda_Re, Lambda_Im :PVector) :Integer;
{ Finds the eigenvalues of an upper Hessenberg matrix }
label
  2;
var
  Anorm, P, Q, R, S, T, U, V, W, X, Y, Z :Float;
  I, Its, J, K, L, M, N                  :Integer;

  function Sign(A, B :Float) :Float;
  begin
    if B < 0.0 then Sign := - Abs(A) else Sign := Abs(A);
  end;

begin

  P := 0;
  Q := 0;
  R := 0;

  Anorm := Abs(A^[1]^[1]);//?! should be Lbound ?

  for I := Succ(Lbound) to Ubound do begin
    for J := I - 1 to Ubound do begin
      Anorm := Anorm + Abs(A^[I]^[J]);
    end;
  end;

  N := Ubound;
  T := 0.0;

  while N >= Lbound do begin

    Its := 0;

2:

    for L := N downto Succ(Lbound) do begin

      S := Abs(A^[L - 1]^[L - 1]) + Abs(A^[L]^[L]);
      if S = 0.0 then S := Anorm;
      if Abs(A^[L]^[L - 1]) <= MACHEP * S then Break;

    end;

    L := Lbound;
    X := A^[N]^[N];

    if L = N then begin

      Lambda_Re^[N] := X + T;
      Lambda_Im^[N] := 0.0;
      N := N - 1;

    end else begin

      Y := A^[N - 1]^[N - 1];
      W := A^[N]^[N - 1] * A^[N - 1]^[N];

      if L = N - 1 then begin

        P := 0.5 * (Y - X);
        Q := Sqr(P) + W;
        Z := Sqrt(Abs(Q));
        X := X + T;

        if Q >= 0.0 then begin

          Z := P + Sign(Z, P);
          Lambda_Re^[N] := X + Z;
          Lambda_Re^[N - 1] := Lambda_Re^[N];
          if Z <> 0.0 then Lambda_Re^[N] := X - W / Z;
          Lambda_Im^[N] := 0.0;
          Lambda_Im^[N - 1] := 0.0

        end else begin

          Lambda_Re^[N] := X + P;
          Lambda_Re^[N - 1] := Lambda_Re^[N];
          Lambda_Im^[N] := Z;
          Lambda_Im^[N - 1] := - Z

        end;

        N := N - 2;

      end else begin

        if Its = 30 then begin
          Hqr := MAT_NON_CONV;
          Exit;
        end;

        if (Its = 10) or (Its = 20) then begin
          T := T + X;

          for I := Lbound to N do
            A^[I]^[I] := A^[I]^[I] - X;
          S := Abs(A^[N]^[N - 1]) + Abs(A^[N - 1]^[N - 2]);
          X := 0.75 * S;
          Y := X;
          W := - 0.4375 * Sqr(S)
        end;

        Its := Its + 1;

        for M := N - 2 downto L do begin

          Z := A^[M]^[M];
          R := X - Z;
          S := Y - Z;
          P := (R * S - W) / A^[M + 1]^[M] + A^[M]^[M + 1];
          Q := A^[M + 1]^[M + 1] - Z - R - S;
          R := A^[M + 2]^[M + 1];
          S := Abs(P) + Abs(Q) + Abs(R);
          P := P / S;
          Q := Q / S;
          R := R / S;

          if M = L then Break;

          U := Abs(A^[M]^[M - 1]) * (Abs(Q) + Abs(R));
          V := Abs(P) * (Abs(A^[M - 1]^[M - 1]) + Abs(Z) + Abs(A^[M + 1]^[M + 1]));

          if U <= MACHEP * V then Break;

        end;

        M := L;

        for I := M + 2 to N do begin

          A^[I]^[I - 2] := 0.0;
          if I <> (M + 2) then A^[I]^[I - 3] := 0.0

        end;

        for K := M to N - 1 do begin

          if K <> M then begin

            P := A^[K]^[K - 1];
            Q := A^[K + 1]^[K - 1];
            R := 0.0;
            if K <> (N - 1) then
              R := A^[K + 2]^[K - 1];
            X := Abs(P) + Abs(Q) + Abs(R);
            if X <> 0.0 then begin
              P := P / X;
              Q := Q / X;
              R := R / X
            end;
          end;

          S := Sign(Sqrt(Sqr(P) + Sqr(Q) + Sqr(R)), P);

          if S <> 0.0 then begin

            if K = M then begin
              if L <> M then A^[K]^[K - 1] := - A^[K]^[K - 1];
            end else begin
              A^[K]^[K - 1] := - S * X
            end;

            P := P + S;
            X := P / S;
            Y := Q / S;
            Z := R / S;
            Q := Q / P;
            R := R / P;

            for J := K to N do begin

              P := A^[K]^[J] + Q * A^[K + 1]^[J];
              if K <> (N - 1) then begin
                P := P + R * A^[K + 2]^[J];
                A^[K + 2]^[J] := A^[K + 2]^[J] - P * Z
              end;
              A^[K + 1]^[J] := A^[K + 1]^[J] - P * Y;
              A^[K]^[J] := A^[K]^[J] - P * X
            end;

              for I := L to IMin(N, K + 3) do
                begin
                  P := X * A^[I]^[K] + Y * A^[I]^[K + 1];
                  if K <> (N - 1) then
                    begin
                      P := P + Z * A^[I]^[K + 2];
                      A^[I]^[K + 2] := A^[I]^[K + 2] - P * R
                    end;
                  A^[I]^[K + 1] := A^[I]^[K + 1] - P * Q;
                  A^[I]^[K] := A^[I]^[K] - P
                end

          end;
        end;


        goto 2;

      end;

    end;

  end;

  Hqr := MAT_OK;
end;


//******************************************************************************
function EigenVals(A :PMatrix; Lbound, Ubound :Integer; Lambda_Re, Lambda_Im :PVector) :Integer;
begin
  Balance(A, Lbound, Ubound);
  ElmHes(A, Lbound, Ubound);
  EigenVals := Hqr(A, Lbound, Ubound, Lambda_Re, Lambda_Im);
end;


//******************************************************************************
procedure DivLargest(V :PVector; Lbound, Ubound :Integer; var Largest :Float);
var
  I : Integer;
begin
  Largest := V^[Lbound];
  for I := Succ(Lbound) to Ubound do
    if Abs(V^[I]) > Abs(Largest) then
      Largest := V^[I];
  for I := Lbound to Ubound do
    V^[I] := V^[I] / Largest;
end;


//******************************************************************************
function EigenVect(A :PMatrix; Lbound, Ubound :Integer; Lambda :Float; V :PVector) :Integer;
const
  TOL = MACHEP * 10.0;  { Tolerance for singular values }
var
  I, J, I1, J1, Rank, ErrCode : Integer;
  Sum, Norm                   : Float;
  B, S, X                     : PVector;
  A1, W                       : PMatrix;
  Zero_B                      : Boolean;
begin

  DimMatrix(A1, Ubound, Ubound);
  DimMatrix(W, Ubound, Ubound);
  DimVector(S, Ubound);

  { Form (A - Lambda * I) and perform its singular value decomposition }
  CopyMatrix(A1, A, Lbound, Lbound, Ubound, Ubound);
  for I := Lbound to Ubound do
    A1^[I]^[I] := A^[I]^[I] - Lambda;

  ErrCode := SV_Decomp(A1, Lbound, Ubound, Ubound, S, W);

  if ErrCode <> MAT_OK then
    begin
      EigenVect := ErrCode;
      Exit;
    end;

  SV_SetZero(S, Lbound, Ubound, TOL);

  { Find rank of (A - lambda * I) }
  Rank := 0;
  for I := Lbound to Ubound do
    if S^[I] <> 0.0 then
      Inc(Rank);

  { The components of the eigenvector which correspond to
    the null singular value of (A - lambda * I) are set to 1 }
  for I := Lbound to Ubound do
    if S^[I] = 0.0 then
      V^[I] := 1.0;

  { The other components are found by solving a subsystem A1 * X = B
    in which only the matrix elements corresponding to the nonzero
    singular values are used }

  DelMatrix(W, Ubound, Ubound);
  DelMatrix(A1, Ubound, Ubound);
  DimMatrix(A1, Rank, Rank);
  DimVector(B, Rank);
  DimVector(X, Rank);

  I1 := 0;
  for I := Lbound to Ubound do
    if S^[I] <> 0 then
      begin
        Inc(I1);
        J1 := 0;
        for J := Lbound to Ubound do
          if S^[J] = 0 then
            B^[I1] := B^[I1] - A^[I]^[J]
          else
            begin
              Inc(J1);
              if I = J then
                A1^[I1]^[J1] := A^[I]^[J] - Lambda
              else
                A1^[I1]^[J1] := A^[I]^[J];
            end;
      end;

  { Check if the constant vector B is zero }
  I1 := 0;
  repeat
    Inc(I1);
  until (I1 > Rank) or (Abs(B^[I1]) > MACHEP);
  Zero_B := (I1 > Rank);

  { If B is not zero, solve the subsystem by LU decomposition }
  if not Zero_B then
    begin
      ErrCode := LU_Decomp(A1, 1, Rank);
      if ErrCode <> MAT_OK then
        begin
          EigenVect := ErrCode;
          Exit;
        end;
      LU_Solve(A1, B, 1, Rank, X);
    end;

  { Update eigenvector }
  I1 := 0;
  for I := Lbound to Ubound do
    if S^[I] <> 0 then
      begin
        Inc(I1);
        V^[I] := X^[I1];
      end;

  { Normalize eigenvector and make sure that the first component is > 0 }
  Sum := 0.0;
  for I := Lbound to Ubound do
    Sum := Sum + Sqr(V^[I]);
  Norm := Sqrt(Sum);

  //My modification - Lorin Milescu
  if Norm = 0 then Norm := 1e-30;

  for I := Lbound to Ubound do
    V^[I] := V^[I] / Norm;
  if V^[Lbound] < 0.0 then
    for I := Lbound to Ubound do
      V^[I] := - V^[I];

  DelVector(B, Rank);
  DelVector(X, Rank);
  DelVector(S, Ubound);
  DelMatrix(A1, Rank, Rank);

  EigenVect := ErrCode;
end;


//******************************************************************************
function RootPol(Coef :PVector; Deg :Integer; X_Re, X_Im :PVector) :Integer;
var
  A          :PMatrix;        { Companion matrix }
  N          :Integer;        { Size of matrix }
  I, J, K    :Integer;  { Loop variables }
  ErrCode    :Integer;  { Error code }
  Temp       :Float;
begin
  N := Pred(Deg);
  DimMatrix(A, N, N);

  { Set up the companion matrix (to save space, begin at index 0) }
  for J := 0 to N do
    A^[0]^[J] := - Coef^[Deg - J - 1] / Coef^[Deg];
  for J := 0 to Pred(N) do
    A^[J + 1]^[J] := 1.0;

  { The roots of the polynomial are the eigenvalues of the companion matrix }
  Balance(A, 0, N);
  ErrCode := Hqr(A, 0, N, X_Re, X_Im);

  if ErrCode = MAT_OK then
    begin
      { Sort roots in increasing order of real parts }
      for I := 0 to N - 1 do
        begin
          K := I;
          Temp := X_Re^[I];
          for J := Succ(I) to N do
            if X_Re^[J] < Temp then
              begin
                K := J;
                Temp := X_Re^[J];
              end;
          FSwap(X_Re^[I], X_Re^[K]);
          FSwap(X_Im^[I], X_Im^[K]);
        end;

      { Transfer roots from 0..(Deg - 1) to 1..Deg }
      for J := N downto 0 do
        begin
          X_Re^[J + 1] := X_Re^[J];
          X_Im^[J + 1] := X_Im^[J];
        end;
    end;

  DelMatrix(A, N, N);
  RootPol := ErrCode;
end;


//******************************************************************************
end.
