{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// by Lorin
// Intended for situations with more than one receiver for a single TNotifyEvent.

unit QUB_Events;


interface


uses


{$ifdef QUB}
Alloc, AllocTypes,
{$endif}
SysUtils, Classes;


type


{$ifdef QUB}
{$else}
TAPointer                           = Array[0..(MaxInt div (SizeOf(Pointer)*2))] of Pointer;
TPAPointer                          = ^TAPointer;
{$endif}

TPNotifyEvent = ^TNotifyEvent;


//forward class declarations
TQUB_EventClass = class;
TQUB_Events     = class;


TQUB_EventClass = class(TList)
protected


  function prGetEvent(Index :Integer) :TNotifyEvent;


public


  procedure Clear; override;

  procedure AddEvent(Evt :TNotifyEvent);
  procedure DelEvent(Evt :TNotifyEvent);

  function IndexOfEvent(Evt :TNotifyEvent): Integer;


  property Event[Index :Integer]           :TNotifyEvent  read prGetEvent; default;


end;


TQUB_Events     = class(TStringList)
protected


  function prGetEventClass(Index :Integer) :TQUB_EventClass;
  function prGetEventClassName(Index :String) :TQUB_EventClass;


public

  destructor Destroy; override;

  procedure Clear; override;

  procedure RaiseNotifyEvent(const EventName :String; Sender :TObject);
  procedure AddEvent(const EventClass :String; Evt :TNotifyEvent);
  procedure DelEvent(const EventClass :String; Evt :TNotifyEvent);

  procedure AddEventClass(const EventClass :String);
  procedure DelEventClass(const EventClass :String);


  property EventClass[Index :Integer]      :TQUB_EventClass   read prGetEventClass;
  property EventClassName[Index :String]   :TQUB_EventClass   read prGetEventClassName; default;


end;


implementation


//******************************************************************************
procedure TQUB_EventClass.Clear;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      Dispose(TPNotifyEvent(List[i]));
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_EventClass.AddEvent(Evt :TNotifyEvent);
var
  p :TPNotifyEvent;
begin
  New(p);
  p^ := Evt;
  Add(p);
end;

procedure TQUB_EventClass.DelEvent(Evt :TNotifyEvent);
var
  i :Integer;
begin
  i := IndexOfEvent(Evt);
  if i >= 0 then begin
    Dispose(TPNotifyEvent(List[i]));
    Delete(i);
  end;
  {
  for i := 0 to Count - 1 do begin
    if @Event[i] = @Evt then begin           // this compares only the class and method, but not the instance // replaced by .IndexOfEvent
      Dispose(TPNotifyEvent(List[i]));
      Delete(i);
      Break;
    end;
  end;
  }
end;

function TQUB_EventClass.prGetEvent(Index :Integer) :TNotifyEvent;
begin
  if  (Index >= 0) and (Index < Count) then Result := TNotifyEvent(List[Index]^) else Result := nil;
end;

function TQUB_EventClass.IndexOfEvent(Evt :TNotifyEvent): Integer;
var
  p       :TPAPointer;
  findp   :TPAPointer;
  i       :Integer;
begin
  Result := -1;

  findp   := TPAPointer( @@Evt );

  for i := 0 to Count - 1 do begin
    p    := TPAPointer( List[i] );

    if (p[0] = findp[0]) and (p[1] = findp[1]) then begin
      Result := i;
      Break;
    end;
  end;
end;

//******************************************************************************
destructor TQUB_Events.Destroy;
begin
  try
    Clear;
  finally
    inherited;
  end;
end;

procedure TQUB_Events.Clear;
var
  i :Integer;
begin
  try
    for i := 0 to Count - 1 do begin
      EventClass[i].Free;
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_Events.RaiseNotifyEvent(const EventName :String; Sender :TObject);
var
  c :TQUB_EventClass;
  i :Integer;
begin
  c := Self[EventName];
  if c <> nil then begin
    for i := c.Count - 1 downto 0 do begin
      c.Event[i](Sender);
    end;
  end;
end;

procedure TQUB_Events.AddEvent(const EventClass :String; Evt :TNotifyEvent);
var
  c :TQUB_EventClass;
begin
  c := Self[EventClass];
  if c = nil then begin
    AddEventClass(EventClass);
    c := Self[EventClass];
  end;
  if c = nil then Exit;
  c.AddEvent(Evt);
end;

procedure TQUB_Events.DelEvent(const EventClass :String; Evt :TNotifyEvent);
var
  c :TQUB_EventClass;
begin
  c := Self[EventClass];
  if c = nil then Exit;
  c.DelEvent(Evt);
end;

procedure TQUB_Events.AddEventClass(const EventClass :String);
begin
  if Self[EventClass] <> nil then Exit;
  AddObject(EventClass, TQUB_EventClass.Create);
end;

procedure TQUB_Events.DelEventClass(const EventClass :String);
var
  i :Integer;
begin
  i := IndexOf(EventClass);
  if i = -1 then Exit;
  TQUB_EventClass(Objects[i]).Free;
  Delete(i);
end;

function TQUB_Events.prGetEventClass(Index :Integer) :TQUB_EventClass;
begin
  if (Index >= 0) and (Index < Count) then Result := TQUB_EventClass(Objects[Index]) else Result := nil;
end;

function TQUB_Events.prGetEventClassName(Index :String) :TQUB_EventClass;
var
  i :Integer;
begin
  i := IndexOf(Index);
  if i <> -1 then Result := TQUB_EventClass(Objects[i]) else Result := nil;
end;


//******************************************************************************
end.

