{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_ExpCond_Editor;

interface

uses

Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
Grids, QUB_Table, StdCtrls, ExtCtrls, Math,

QUB_Node,
QUB_Profiles,
QUB_ExpCond_Parameters, QUB_Events;


type

TQUB_ExpCond_Editor = class(TFrame)

  panButtons             :TPanel;
  btnAdd                 :TButton;
  btnDel                 :TButton;

  procedure btnAddClick(Sender :TObject);
  procedure btnDelClick(Sender :TObject);

protected

  grdExpCond             :TQUB_Table;
  mnuProfiles            :TQUB_ProfileEditBox;

  prExpCond              :TQUB_ExpCond_Parameters;
  prProfileCat           :TQUB_ProfileCat;
  prProfile              :INode;

  prShowValueAlways      :Boolean;
  prShowCond             :Boolean;
  prShowChannel          :Boolean;
  prColCond              :Integer;
  prColChannel           :Integer;

  procedure prSetExpCond(x :TQUB_ExpCond_Parameters);
  procedure prSetShowValueAlways(x :Boolean);
  procedure prSetShowCond(x :Boolean);
  procedure prSetShowChannel(x :Boolean);

  procedure prOnSetProfile(Sender :TQUB_ProfileCat; Updates :INode);

  procedure prReLayout;

  procedure grdExpCondSetEditText(Sender: TObject; ACol, ARow: Integer; const Valu: String);


public


  constructor Create(AOwner :TComponent); override;


  // don't let expcond change elsewhere while this is assigned
  property ExpCond             :TQUB_ExpCond_Parameters read prExpCond           write prSetExpCond;
  
  // show value even if a channel is assigned
  property ShowValueAlways     :Boolean                 read prShowValueAlways   write prSetShowValueAlways;
  property ShowCond            :Boolean                 read prShowCond          write prSetShowCond;
  property ShowChannel         :Boolean                 read prShowChannel       write prSetShowChannel;


end;


implementation


uses


QUB_GUIInterface;


{$R *.DFM}


constructor TQUB_ExpCond_Editor.Create(AOwner :TComponent);
begin
  inherited;

  grdExpCond        := TQUB_Table.Create(Self);
  grdExpCond.Align  := alClient;
  grdExpCond.DefaultColWidth   := 75;
  grdExpCond.DefaultRowHeight  := 16;
  grdExpCond.Parent := Self;

  grdExpCond.Options := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRowSizing, goColSizing, goThumbTracking];
  grdExpCond.OnSetEditText := grdExpCondSetEditText;

  mnuProfiles                             := TQUB_ProfileEditBox.Create(Self);
  mnuProfiles.Parent                      := Self;
  mnuProfiles.SetBounds(Width - 75, // btnAdd.Width,
                        btnDel.Top + (btnDel.Top - btnAdd.Top),
                        75, mnuProfiles.Height);
  mnuProfiles.Anchors                     := [akRight, akTop];
  mnuProfiles.Cat   := 'ExpCond';

  prExpCond         := nil;
  prProfileCat      := ProfileMgr['ExpCond'];
  prProfile         := QUB_Node_NULL;

  prShowValueAlways := False;
  prShowCond        := True;
  prShowChannel     := True;

  prReLayout;
end;

procedure TQUB_ExpCond_Editor.prReLayout;
var
  i           :Integer;
  Par         :TQUB_ExpCond_PParameter;
  Active      :Boolean;
begin
  // i as ColCount
  i           := 2; // param name, value
  if ShowCond    then begin
    prColCond   := i;
    Inc(i);
  end;
  if ShowChannel then begin
    prColChannel := i;
    Inc(i);
  end;
  grdExpCond.ColCount    := i;

  btnAdd.Enabled := ExpCond <> nil;
  btnDel.Enabled := ExpCond <> nil;

  if ExpCond = nil then begin
    grdExpCond.RowCount  := 1;
    prProfile            := QUB_Node_NULL;
    Exit;
  end;

  Active                 := prProfileCat.Active.Equals(prProfile);
  prProfile              := ExpCond.ToTree;
  if Active then prProfileCat.Active := prProfile;

  grdExpCond.RowCount    := ExpCond.Count + 1;
  grdExpCond.FixedRows   := 1;
  grdExpCond.FixedCols   := 0;

  grdExpCond.Cells[1, 0] := 'Value';
  for i := 0 to ExpCond.Count - 1 do begin
    Par       := ExpCond[i];
    grdExpCond.Cells[0, i+1] := Par.Name;
    if ShowValueAlways or (Par.Channel = -1) then begin
      grdExpCond.Cells[1, i+1] := FloatToStr(Par.Value);
    end else begin
      grdExpCond.Cells[1, i+1] := '';
    end;
  end;

  if ShowCond then begin
    grdExpCond.Cells[prColCond, 0] := 'Conditioning';
    for i := 0 to ExpCond.Count - 1 do begin
      grdExpCond.Cells[prColCond, i+1]  := FloatToStr(ExpCond[i].Cond_Value);
    end;
  end;

  if ShowChannel then begin
    grdExpCond.Cells[prColChannel, 0] := 'Channel';
    for i := 0 to ExpCond.Count - 1 do begin
      if ExpCond[i].Channel >= 0 then begin
        grdExpCond.Cells[prColChannel, i+1]  := IntToStr(ExpCond[i].Channel);
      end else begin
        grdExpCond.Cells[prColChannel, i+1]  := '';
      end;
    end;
  end;
end;

procedure TQUB_ExpCond_Editor.btnAddClick(Sender :TObject);
begin
  if prExpCond = nil then Exit;
  prExpCond.AddPar('Param', 1.0, -1, False, -1);
  prReLayout;
end;

procedure TQUB_ExpCond_Editor.btnDelClick(Sender :TObject);
begin
  if prExpCond = nil then Exit;
  if grdExpCond.Row > 0 then begin
    prExpCond.DelPar(grdExpCond.Row - 1);
    prReLayout;
  end;
end;

procedure TQUB_ExpCond_Editor.grdExpCondSetEditText(Sender: TObject; ACol, ARow: Integer; const Valu: String);
var
  ParNode  :INode;
  Value    :String;
  c        :Integer;
begin
  if ShowChannel and (ACol = prColChannel) and (Valu = '') then begin
    Value := '-1';
  end else begin
    Value := Valu;
  end;

  if (Value = '') or (Value = '.') or (Value = '-') or (Value = '-.') then Exit;
  if ExpCond = nil then Exit;
  if (ARow < 0) or (ARow > ExpCond.Count) then Exit;

  try
    ParNode := prProfile.Find(ExpCond[ARow-1].Name);
    if ACol = 0 then begin
      ExpCond.ChangeParName(ARow-1, Value);
      ParNode.Name := Value;
    end else if ACol = 1 then begin
      if (ExpCond[ARow-1].Channel >= 0) and not ShowValueAlways then begin
        grdExpCond.Cells[ACol, ARow] := '';
      end else begin
        ExpCond.ChangePar(ARow-1, StrToFloat(Value));
        ParNode.SetData(StrToFloat(Value));
      end;
    end else if ShowCond and (ACol = prColCond) then begin
      ExpCond.ChangeParCond(ARow-1, StrToFloat(Value));
      ParNode['Cond_Value'].SetData(StrToFloat(Value));
    end else if ShowChannel and (ACol = prColChannel) then begin
      c := StrToInt(Value);
      ExpCond.ChangeParChannel(ARow-1, c);
      ParNode['ChannelIndex'].SetData(c);
      if not ShowValueAlways then begin
        if c >= 0 then begin
          grdExpCond.Cells[1, ARow] := '';
        end else begin
          grdExpCond.Cells[1, ARow] := FloatToStr(ExpCond[ARow-1].Value);
        end;
      end;
    end;
  except
  end;
end;

procedure TQUB_ExpCond_Editor.prOnSetProfile(Sender :TQUB_ProfileCat; Updates :INode);
begin
  ExpCond.ReadTree(Updates);
  prReLayout;
end;

procedure TQUB_ExpCond_Editor.prSetExpCond(x :TQUB_ExpCond_Parameters);
var
  Foo          :TQUB_Profile_OnSet;
begin
  prExpCond := x;
  if x <> nil then begin
    prProfileCat.Active := prProfile;
    prProfileCat.OnSet  := prOnSetProfile;
  end else begin
    Foo          := prOnSetProfile;
    if (@prProfileCat.OnSet = @Foo)
    then prProfileCat.OnSet := nil;
  end;
  prReLayout;
end;

procedure TQUB_ExpCond_Editor.prSetShowValueAlways(x :Boolean);
begin
  prShowValueAlways := x;
  if prExpCond <> nil then prReLayout;
end;

procedure TQUB_ExpCond_Editor.prSetShowCond(x :Boolean);
begin
  prShowCond := x;
  if prExpCond <> nil then prReLayout;
end;

procedure TQUB_ExpCond_Editor.prSetShowChannel(x :Boolean);
begin
  prShowChannel := x;
  if prExpCond <> nil then prReLayout;
end;



end.
 