unit QUB_FitCurves;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris 4-28-2004
// imports Fit curve repository from qubopt for use by QVO_QVD
// wraps it for events

interface


uses

Alloc, AllocTypes,

QUB_Events;


type

TQUB_FitCurve = class
public
  function First: Integer; virtual; cdecl; abstract;
  function Last:  Integer; virtual; cdecl; abstract;
  function Data:  TPAInteger; virtual; cdecl; abstract;
end;

TQUB_FitCurveIter = class
public
  procedure Free; virtual; cdecl; abstract;
  function Done: Integer; virtual; cdecl; abstract;
  function Next: Integer; virtual; cdecl; abstract;
  function Curve: TQUB_FitCurve; virtual; cdecl; abstract;
  procedure MarkChanged; virtual; cdecl; abstract;
end;

TQUB_FitCurves_ = class
public
  procedure Free; virtual; cdecl; abstract;
  procedure Clear; virtual; cdecl; abstract;
  function Add(First, Last :Integer): TQUB_FitCurveIter; virtual; cdecl; abstract;
  function Find(First :Integer): TQUB_FitCurveIter; virtual; cdecl; abstract;
  function Empty: Integer; virtual; cdecl; abstract;

  procedure Lock; virtual; cdecl; abstract;
  procedure Unlock; virtual; cdecl; abstract;
  procedure UseTempFile; virtual; cdecl; abstract;
end;

function TQUB_FitCurves_Create: TQUB_FitCurves_; stdcall; external 'qubopt.dll' name '_QUB_FitCurves_Create@0';

type

TQUB_FitCurves = class
protected
  prFitCurves      :TQUB_FitCurves_;
public
  Events           :TQUB_Events;       //OnChanged

  //Testing
  FirstFindSynch   :Integer;
  FitFoundSynch    :TQUB_FitCurveIter;
  //

  constructor Create;
  destructor Destroy; override;

  procedure Clear;
  function Add(First, Last :Integer): TQUB_FitCurveIter;
  function Find(First: Integer): TQUB_FitCurveIter;
  procedure FindSynch;
  function IsEmpty: Boolean;

  procedure Lock;
  procedure Unlock;
  procedure UseTempFile;

  procedure RaiseChanged; // no automatic event trigger
end;


implementation



constructor TQUB_FitCurves.Create;
begin
  prFitCurves := TQUB_FitCurves_Create;
  Events      := TQUB_Events.Create;
end;

destructor TQUB_FitCurves.Destroy;
begin
  try
    prFitCurves.Free;
    Events.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_FitCurves.Clear;
begin
  prFitCurves.Clear;
end;

function TQUB_FitCurves.Add(First, Last :Integer): TQUB_FitCurveIter;
begin
  Result := prFitCurves.Add(First, Last);
end;

function TQUB_FitCurves.Find(First: Integer): TQUB_FitCurveIter;
begin
  Result := prFitCurves.Find(First);
end;

procedure TQUB_FitCurves.FindSynch;
begin
  FitFoundSynch := prFitCurves.Find(FirstFindSynch);
end;

function TQUB_FitCurves.IsEmpty: Boolean;
begin
  Result := 0 <> prFitCurves.Empty;
end;

procedure TQUB_FitCurves.Lock;
begin
  prFitCurves.Lock;
end;

procedure TQUB_FitCurves.Unlock;
begin
  prFitCurves.Unlock;
end;

procedure TQUB_FitCurves.UseTempFile;
begin
  prFitCurves.UseTempFile;
end;

procedure TQUB_FitCurves.RaiseChanged; // no automatic event trigger
begin
  Events.RaiseNotifyEvent('OnChanged', Self);
end;


// *****************************************************************************
end.
 