{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// various utilities with no particular theme


unit QUB_Global;


interface


uses


Windows, SysUtils, ShellAPI, Dialogs, Forms,
Math, 
AllocTypes, Alloc;


const


Y_STR                   = 'Y';
N_STR                   = 'N';

ONE_MEGABYTE            = 1048576;

QUB_IDATA_WINSIZE       = 1000000;

//MAX_CHANNEL             = 15;

DATA_MSG_RECOUNT_DATA   =    1;
DATA_MSG_BROWSE         =    2;
DATA_MSG_READ_CACHE     =    4;
DATA_MSG_DRAW_DATA      =    8;
DATA_MSG_PROC_DATA      =   16;
DATA_MSG_SET_SCROLL     =   32;
DATA_MSG_COND_BROWSE    =   64;
DATA_MSG_BROWSE_TO_P    =  128;
DATA_MSG_BROWSE_TO_T    =  256;
DATA_MSG_BROWSE_CMD     =  512;
DATA_MSG_DATA_EDIT      = 1024;
DATA_MSG_DATA_APPEND    = 2048;
DATA_MSG_SCROLLINVIEW   = 4096;
DATA_MSG_USE_ACC        = 8192;

DATA_MSG_NOT_BENIGN_FOR_ACC = not (DATA_MSG_BROWSE or
                                   DATA_MSG_DRAW_DATA or
                                   DATA_MSG_PROC_DATA);

HELPCTX_WIKI            = 5241;

type


TQUB_FirstLast          = record
  First                 :Integer;
  Last                  :Integer;
end;
TQUB_PFirstLast         = ^TQUB_FirstLast;

TQUB_AFirstLast         = Array[0..(MaxInt div (SizeOf(TQUB_FirstLast)*2))] of TQUB_FirstLast;
TQUB_PAFirstLast        = ^TQUB_AFirstLast;


function Choice(Value :Boolean; const A, B :String) :String; overload; forward;
function Choice(Value :Boolean; const A, B :Integer) :Integer; overload; forward;
function Choice(Value :Boolean; const A, B :Double) :Double; overload; forward;
//function Choice(Value :Boolean; const A, B :Variant) :Double; overload; forward;


function QUB_FirstLast(First, Last :Integer) :TQUB_FirstLast; forward;


function GetBaseStampedFileName(const BaseName, BaseExt :String): String; overload;
function GetBaseStampedFileName(const BaseName, BaseExt :String; Date, Time :TDateTime): String; overload;
function GetNewUniqueFileName(const BaseName, BaseExt :String; var NoNameCount :Integer) :String; forward;

// shows a page from the wiki in the default web browser.
// Topic should have _ in place of <space>
procedure ShowWiki(Topic :String; Handle :HWND = 0);

function MessageHelpDlg(const Msg: string; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: Longint; const WikiTopic :String): Word;
function CheckWikiHelp(Command: Word; Data: Longint; var CallHelp: Boolean): Boolean;

function Gaussian(Mean, Std :Double; Value :Double) :Double; forward;
function SafeLog(Value :Double) :Double; forward;
function MatrixLogarithm(A, Q :TPMDouble; dt :Double; Size :Integer; MaxIter :Integer; Conv :Double) :Boolean; forward;

// Round to iDec decimals, iDec >= 0 {JB 02-2004} John, the _ is used for ASM functions. Call it differently. LM.
function _round(n : Double; iDec:Integer): Double;

// test for nonnumeric, empty strings, return converted data or defaults 5-2004 {JB}
function IsValidInteger( c :String ) : Boolean;
function ValidStrToInt( c :String; iDefault : Integer ) : Integer;
function IsValidFloat( c : String ) : Boolean;
function ValidStrToFloat( c :String; nDefault : Double ) : Double;

procedure StrToArrayOfInt(const Text :String; ArrayOfInt :TPAInteger; var Count :Integer);

function SafeFileName(const FName :String) :String;

// IntegerArray
type TIntegerArray = class(TObject)
private
  

  m_a          :Array of Integer;
  m_iAllocated :Integer;
  m_iCount     :Integer;


public

  
  constructor Create;
  

  function Count :Integer;
  procedure Add(i :Integer);
  function Get(i :Integer; iDefault :Integer = 0) :Integer;
  function GetLast( iDefault :Integer = 0) :Integer;


end;


implementation


uses

QUB_Viewport_Python, QUB_Python_Thread, PythonEngine;


//========================================================== Integer Array Class
constructor TIntegerArray.Create( );
begin
  SetLength(m_a,10);
  m_iAllocated :=10;
  m_iCount := 0;
end;

Function TIntegerArray.Count() : Integer;
begin
  result := m_iCount;
end;

Procedure TIntegerArray.Add( i:Integer );
begin
  if m_iCount >= m_iAllocated then begin
    SetLength(m_a, m_iCount+10 );
    m_iAllocated:=m_iCount+10;
  end;
  m_a[m_iCount]:=i;
  inc(m_iCount);
end;

Function TIntegerArray.Get( i:Integer; iDefault : Integer = 0 ) : Integer;
begin
  if (i<0) or (i>=m_iCount) Then
    result := iDefault
  else
    result := m_a[i];
end;

Function TIntegerArray.GetLast( iDefault : Integer = 0) : Integer;
begin
  if (m_iCount<=0) Then
    result := iDefault
  else
    result := m_a[m_iCount-1];
end;

//******************************************************************************
function Choice(Value :Boolean; const A, B :String) :String;
begin
  if Value then Result := A else Result := B;
end;
function Choice(Value :Boolean; const A, B :Integer) :Integer;
begin
  if Value then Result := A else Result := B;
end;
function Choice(Value :Boolean; const A, B :Double) :Double;
begin
  if Value then Result := A else Result := B;
end;
{
function Choice(Value :Boolean; const A, B :Variant) :Double;
begin
  if Value then Result := A else Result := B;
end;
}

//******************************************************************************
function Gaussian(Mean, Std :Double; Value :Double) :Double; 
var
  tmp :Double;
begin
  Result := 0;
  try
    if Std = 0 then begin
      if Abs(Value - Mean) < 1e-100 then begin
        Result := 0;
      end else begin
        Result := 1e100//a substitute for infinity
      end;
    end else begin
      tmp := (Value - Mean) / Std;
      tmp := exp(-tmp * tmp / 2.0);
      tmp := tmp / (2.5066283 * Std);
      if tmp < 1e-100 then begin
        Result := 1e-100;
      end else begin
        Result := tmp;
      end;
    end;
  except
  end;
end;

function SafeLog(Value :Double) :Double;
begin
  Result := 0;
  try
    if Value <= 0 then begin
      Result := -1e100//a substitute for -infinity
    end else begin
      Result := ln(Value);
    end;
  except
  end;
end;


{
  Returns the logarithm of a matrix, defined as the (truncated) series:

    Sum(-[(I-A)^n]/n)
}
function MatrixLogarithm(A, Q :TPMDouble; dt :Double; Size :Integer; MaxIter :Integer; Conv :Double) :Boolean;
var
  IA, AN, QQ, QP, p   :TPMDouble;
  Converged           :Boolean;
  i, j, k, n          :Integer;
begin
  Result := False;

  IA := nil;
  AN := nil;
	QQ := nil;
  QP := nil;

  try


    try


      IA := DAlloc2(Size, Size);
      AN := DAlloc2(Size, Size);
      QQ := DAlloc2(Size, Size);
      QP := DAlloc2(Size, Size);


      //IA = I - A
      //AN = I
      //QQ = 0
      //QP = 0
      for i := 0 to Size - 1 do begin
        for j := 0 to Size - 1 do begin
          if i <> j then begin
          //off-diagonal
            IA[i, j] :=   - A[i, j];
            AN[i, j] := 0;
          end else begin
          //diagonal
            IA[i, j] := 1 - A[i, j];
            AN[i, j] := 1;
          end;
          QQ[i, j] := 0;
					QP[i, j] := 0;
        end;
      end;

      n := 1;

      repeat


        Converged := True;


        //AN := AN.IA;
        for i := 0 to Size - 1 do begin
          for j := 0 to Size - 1 do begin
            QQ[i, j] := 0;
            for k := 0 to Size - 1 do begin
              QQ[i, j] := QQ[i, j] + AN[i, k]*IA[k, j];
            end;
          end;
        end;
        for i := 0 to Size - 1 do for j := 0 to Size - 1 do AN[i, j] := QQ[i, j];

        //QQ := QP - AN/n;
        for i := 0 to Size - 1 do begin
          for j := 0 to Size - 1 do begin
            QQ[i, j] := QP[i, j] - AN[i, j];
          end;
        end;


        for i := 0 to Size - 1 do begin
          for j := 0 to Size - 1 do begin
            if Abs((QQ[i, j] + 1e-100 - QP[i, j])/(QQ[i, j] + 1e-100)) > Conv then begin
              Converged := False;
              Break;
            end;
          end;
          if not Converged then Break;
        end;

        p  := QQ;
        QQ := QP;
        QP := p;

        Inc(n);


      until((Converged) or (n > MaxIter));


      //Q := QQ / dt;
      for i := 0 to Size - 1 do begin
        for j := 0 to Size - 1 do begin
          Q[i, j] := QP[i, j] / dt;
        end;
      end


    except

      Result := True;

    end;


  finally

    Free2(TPMChar(IA));
    Free2(TPMChar(AN));
    Free2(TPMChar(QQ));
    Free2(TPMChar(QP));

  end;

end;


procedure ShowWiki(Topic :String; Handle :HWND);
begin
  ShellExecute(Handle, 'open', PChar('http://www.qub.buffalo.edu/wiki/index.php/'+Topic),nil,nil, SW_SHOWNORMAL) ;
end;

var
  gWikiTopic :String;

function MessageHelpDlg(const Msg: string; DlgType: TMsgDlgType; Buttons: TMsgDlgButtons; HelpCtx: Longint; const WikiTopic :String): Word;
begin
  if WikiTopic <> '' then begin
    gWikiTopic := WikiTopic;
    Result   := MessageDlg(Msg, DlgType, Buttons+[mbHelp], HELPCTX_WIKI);
    gWikiTopic := '';
  end else if HelpCtx <> 0 then begin
    Result   := MessageDlg(Msg, DlgType, Buttons+[mbHelp], HelpCtx);
  end else begin
    Result := MessageDlg(Msg, DlgType, Buttons, HelpCtx);
  end;
end;

function CheckWikiHelp(Command: Word; Data: Longint; var CallHelp: Boolean): Boolean;
begin
  if gWikiTopic <> '' then begin
    CallHelp := False;
    ShowWiki(gWikiTopic);
    Result := True;
  end else begin
    Result := False;
  end;
end;

function GetBaseStampedFileName(const BaseName, BaseExt :String): String;
begin
  Result := GetBaseStampedFileName(BaseName, BaseExt, Date, Time);
end;

function GetBaseStampedFileName(const BaseName, BaseExt :String; Date, Time :TDateTime): String;
var
  Y,M,D      :Word;
  H,Min,S,Ms :Word;
begin
  DecodeDate(Date, Y, M, D);
  DecodeTime(Time, H, Min, S, Ms);
  Result := Format('%s-%.4d%.2d%.2d-%.2d%.2d%.2d%s', [BaseName, Y, M, D, H, Min, S, BaseExt]);
end;

function GetNewUniqueFileName(const BaseName, BaseExt :String; var NoNameCount :Integer) :String;
begin
  Inc(NoNameCount);
  Result := BaseName + IntToStr(NoNameCount) + BaseExt;
  while FileExists(Result) do begin
    Inc(NoNameCount);
    Result := BaseName + IntToStr(NoNameCount) + BaseExt;
    if NoNameCount = 100000 then Break;
  end;
end;

function SafeFileName(const FName :String) :String;
begin
  Result := FName;
  Result := StringReplace(Result, '"', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '*', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '/', '_', [rfReplaceAll]);
  Result := StringReplace(Result, ':', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '<', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '>', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '?', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '\', '_', [rfReplaceAll]);
  Result := StringReplace(Result, '|', '_', [rfReplaceAll]);
end;

{function UniqueFileName(const FileName, FileExt, Title, Filter :String) :String;
var
  Dlg            :TSaveDialog;
begin
  Result         := '';

  Dlg            := TSaveDialog.Create(nil);
  Dlg.Options    := [ofPathMustExist];
  Dlg.DefaultExt := FileExt;
  Dlg.Filter     := Filter;
  Dlg.Title      := Title;

  try

    Result := FileName;
    while(FileExists(ChangeFileExt(Result, FileExt))) do begin
      case MessageDlg('File ' + ChangeFileExt(Result, FileExt) + ' already exists. Do you want to ovewrite it ?', mtConfirmation, [mbYes, mbNo, mbCancel], 0) of
        idYes:
          Break;
        idNo:
          begin
            Dlg.FileName := Result;
            if Dlg.Execute then begin
              Result := ChangeFileExt(Dlg.FileName, '');
            end;
          end;
        idCancel:
          begin
            Result := '';
            Exit;
          end;
      end;
    end;

  finally

    Dlg.Free;

  end;

end;
}


function QUB_FirstLast(First, Last :Integer) :TQUB_FirstLast;
begin
  Result.First := First;
  Result.Last  := Last;
end;

function _round(n : Double; iDec:Integer): Double;
begin
  Result := Round(n  * IntPower( 10.0, iDec ) ) / IntPower( 10.0, iDec );
end;
  
//----- Returns true if StrToInt does not throw exception
function IsValidInteger(c : String) :Boolean;
var
  i :Integer;
begin
  Result := True;
  try
    if Length(c) = 0 then Result := False else i := StrToInt(c);
    //Too fool the compiler: (LM)
    if i <> 0 then i := 0;
  except
    on EConvertError do Result := False;
  end;
end;

//----- Converts string to integer or uses iDefault if exception is thrown
function ValidStrToInt(c :String; iDefault :Integer) :Integer;
begin
  if IsValidInteger(c) then Result := StrToInt(c) else Result := iDefault;
end;

//----- Returns true if StrToFloat does not throw exception
function IsValidFloat(c :String) :Boolean;
var
  n : Double;
begin
  Result := True;
  try
    if Length(c) = 0 then Result := False else n := StrToFloat(c);
    //To fool the compiler: (LM)
    if n <> 0 then n := 0;
  except
    on EConvertError do Result := False;
  end;
end;

//----- Converts string to Double or uses nDefault if exception is thrown
function ValidStrToFloat(c :String; nDefault :Double) :Double;
begin
  if IsValidFloat(c) then Result := StrToFloat(c) else Result := nDefault;
end;

procedure StrToArrayOfInt(const Text :String; ArrayOfInt :TPAInteger; var Count :Integer);
var
  PyResult :PPyObject;
begin
  PythonThread.BeginPython;
  try
    with GetPythonEngine do begin
      PyResult := EvalString(Format('StrToArrayOfInt("""%s""", %d, %d)',
                                    [Text, DWORD(ArrayOfInt), Count]));
      if PyResult = nil then begin
        PyErr_Print;
        raise Exception.Create('');
      end;
      Count    := PyInt_AsLong(PyResult);
      Py_DECREF(PyResult);
    end;
  except
    Count := 0;
  end;
  PythonThread.EndPython;
end;


//******************************************************************************
end.
