{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the form you see when you right-click the first histogram in Viewport_Histogram and choose Properties


unit QUB_Histogram_Amp_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,
RegistryInterface,

QUB_Profiles,
QUB_Properties,
QUB_Task_Properties;

{$include 'QUB_Histogram_Amp_Properties.hct'}


const


DEF_BINCOUNT   = 100;
MIN_BINCOUNT   = 1;
MAX_BINCOUNT   = 10000;

DEF_AUTOSCALE  = 1;

DEF_MIN        :Double = 0;
MIN_MIN        :Double = -1e6;
MAX_MIN        :Double = +1e6;

DEF_MAX        :Double = 1;
MIN_MAX        :Double = -1e6;
MAX_MAX        :Double = +1e6;

DEF_YTITLE     = 'Count';
DEF_XTITLE     = 'Amplitude';

DEF_COPYWIDTH  = 1000;
MIN_COPYWIDTH  = 10;
MAX_COPYWIDTH  = 10000;

DEF_COPYHEIGHT = 200;
MIN_COPYHEIGHT = 10;
MAX_COPYHEIGHT = 10000;


type


TQUB_Histogram_Amp_PropertiesForm = class(TQUB_Task_PropertiesForm)


  Label1                          :TLabel;
  Label2                          :TLabel;
  Label3                          :TLabel;
  Label4                          :TLabel;
  Label5                          :TLabel;
  Label6                          :TLabel;
  Label7                          :TLabel;


  btnReset                        :TButton;

  txtBinCount                     :TEdit;
  udBinCount                      :TUpDown;

  chkAutoScale                    :TCheckBox;
  txtMin                          :TEdit;
  txtMax                          :TEdit;
  txtYTitle                       :TEdit;
  txtXTitle                       :TEdit;
  txtCopyWidth                    :TEdit;
  txtCopyHeight                   :TEdit;
    panBinColor: TPanel;
    Label8: TLabel;
    Label9: TLabel;
    panFitColor: TPanel;


  procedure FormCreate(Sender: TObject);
  procedure FormHide(Sender: TObject);


protected


  prBinCount                      :Integer;
  prAutoScale                     :Boolean;
  prMin                           :Double;
  prMax                           :Double;
  prYTitle                        :String;
  prXTitle                        :String;
  prCopyWidth                     :Integer;
  prCopyHeight                    :Integer;
  prBinColor                      :TColor;
  prFitColor                      :TColor;


public


  procedure PropertiesToParams;


  property BinCount               :Integer     read prBinCount;
  property AutoScale              :Boolean     read prAutoScale;             
  property Min                    :Double      read prMin;                   
  property Max                    :Double      read prMax;                   
  property YTitle                 :String      read prYTitle;                
  property XTitle                 :String      read prXTitle;                
  property CopyWidth              :Integer     read prCopyWidth;             
  property CopyHeight             :Integer     read prCopyHeight;            
  property BinColor               :TColor      read prBinColor;              
  property FitColor               :TColor      read prFitColor;              


end;


var


QUB_Histogram_Amp_PropertiesForm: TQUB_Histogram_Amp_PropertiesForm;



implementation


{$R *.DFM}


uses


QUB_Viewport_Data;


//******************************************************************************
procedure TQUB_Histogram_Amp_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;


  udBinCount.Min      := MIN_BINCOUNT;
  udBinCount.Max      := MAX_BINCOUNT;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Bin Count"]',
                      'Bin Count', REGTYPE_INT, DEF_BINCOUNT,
                      udBinCount,
                      txtBinCount,
                      AcceptIntGreaterThan(1)));
  AddColorPanelProperty(TQUB_Property_ColorPanel.Create('["Bin Color"]',
                          'Bin Color', REGTYPE_INT, Integer(clRed),
                          panBinColor));
  AddColorPanelProperty(TQUB_Property_ColorPanel.Create('["Fit Color"]',
                          'Fit Color', REGTYPE_INT, Integer(clNavy),
                          panFitColor));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Autoscale"]',
                        'Autoscale', REGTYPE_BOOL, DEF_AUTOSCALE,
                        chkAutoScale));
  AddEditProperty(TQUB_Property_Edit.Create('["Min"]',
                    'Min', REGTYPE_FLOAT, DEF_MIN,
                    txtMin,
                    AcceptFloatBetween(MIN_MIN, MAX_MIN),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Max"]',
                    'Max', REGTYPE_FLOAT, DEF_MAX,
                    txtMAX,
                    AcceptFloatBetween(MIN_MAX, MAX_MAX),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["YTitle"]',
                    'YTitle', REGTYPE_STRING, DEF_YTITLE,
                    txtYTitle,
                    '',
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["XTitle"]',
                    'XTitle', REGTYPE_STRING, DEF_XTITLE,
                    txtXTitle,
                    '',
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Copy Width"]',
                    'Copy Width', REGTYPE_INT, DEF_COPYWIDTH,
                    txtCopyWidth,
                    AcceptIntBetween(MIN_COPYWIDTH, MAX_COPYWIDTH),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Copy Height"]',
                    'Copy Height', REGTYPE_INT, DEF_COPYHEIGHT,
                    txtCopyHeight,
                    AcceptIntBetween(MIN_COPYHEIGHT, MAX_COPYHEIGHT),
                    '') );

  RegKeyName            := 'Software\QUB\Histogram Amp';
  ProfileCat            := 'HistogramAmp';

  prBeforeGetProperties := PropertiesToParams;

  {$include 'QUB_Histogram_Amp_Properties.htx'}

  WikiTopic             := 'Histogram#Amplitude';

end;

procedure TQUB_Histogram_Amp_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;

procedure TQUB_Histogram_Amp_PropertiesForm.PropertiesToParams;
begin
  prBinCount     := Properties['Bin Count'].Data[0];
  prBinColor     := TColor(Properties['Bin Color'].Data[0]);
  prFitColor     := TColor(Properties['Fit Color'].Data[0]);
  prAutoscale    := 0<>Properties['Autoscale'].Data[0];
  prMin          := Properties['Min'].Data[0];
  prMax          := Properties['Max'].Data[0];
  prYTitle       := Properties['YTitle'].DataAsString;
  prXTitle       := Properties['XTitle'].DataAsString;
  prCopyWidth    := Properties['Copy Width'].Data[0];
  prCopyHeight   := Properties['Copy Height'].Data[0];
end;


//******************************************************************************
end.
