{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// properties dialog for third histogram in Viewport_Histogram


unit QUB_Histogram_Dur_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,
RegistryInterface,

QUB_Profiles,
QUB_Properties,
QUB_Task_Properties;

{$include 'QUB_Histogram_Dur_Properties.hct'}

const


DEF_BINCOUNT  = 100;
MIN_BINCOUNT  = 1;
MAX_BINCOUNT  = 10000;

DEF_DURCLASS  = 0;
MIN_DURCLASS  = 0;
MAX_DURCLASS  = 10000;

DEF_DURLOG    = 1;

DEF_AUTOSCALE = 1;

DEF_MIN      :Double = +1e-7;
MIN_MIN      :Double = +1e-100;
MAX_MIN      :Double = +1e+100;

DEF_MAX      :Double = 1;
MIN_MAX      :Double = +1e-100;
MAX_MAX      :Double = +1e+100;


type


TQUB_Histogram_Dur_PropertiesForm = class(TQUB_Task_PropertiesForm)


  Label1                          :TLabel;
  Label2                          :TLabel;

  btnReset                        :TButton;

  txtBinCount                     :TEdit;
  udBinCount                      :TUpDown;
  txtDurClass                     :TEdit;
  udDurClass                      :TUpDown;
  chkDurLog                       :TCheckBox;
  chkAutoScale                    :TCheckBox;
  txtMin                          :TEdit;
  Label3                          :TLabel;
  txtMax                          :TEdit;
  Label4                          :TLabel;
  Label5                          :TLabel;
  Label6                          :TLabel;


  procedure FormCreate(Sender: TObject);
  procedure FormHide(Sender: TObject);


protected


  prBinCount                      :Integer;
  prDurClass                      :Integer;
  prDurLog                        :Boolean;
  prAutoScale                     :Boolean;
  prMin                           :Double;
  prMax                           :Double;


public


  procedure PropertiesToParams;


  property BinCount               :Integer     read prBinCount;    
  property DurClass               :Integer     read prDurClass;    
  property DurLog                 :Boolean     read prDurLog;      
  property AutoScale              :Boolean     read prAutoScale;   
  property Min                    :Double      read prMin;
  property Max                    :Double      read prMax;         


end;


var


QUB_Histogram_Dur_PropertiesForm: TQUB_Histogram_Dur_PropertiesForm;


implementation


{$R *.DFM}


uses


QUB_Viewport_Data;


//******************************************************************************
procedure TQUB_Histogram_Dur_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  udBinCount.Min      := MIN_BINCOUNT;
  udBinCount.Max      := MAX_BINCOUNT;
  udBinCount.Position := DEF_BINCOUNT;

  udDurClass.Min      := MIN_DURCLASS;
  udDurClass.Max      := MAX_DURCLASS;
  udDurClass.Position := DEF_DURCLASS;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Bin Count"]',
                      'Bin Count', REGTYPE_INT, DEF_BINCOUNT,
                      udBinCount,
                      txtBinCount,
                      AcceptIntGreaterThan(1)));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["Dur Class"]',
                      'Dur Class', REGTYPE_INT, DEF_DURCLASS,
                      udDurClass,
                      txtDurClass,
                      AcceptIntGreaterThan(-1)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Dur Log"]',
                        'Dur Log', REGTYPE_BOOL, DEF_DURLOG,
                        chkDurLog));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Autoscale"]',
                        'Autoscale', REGTYPE_BOOL, DEF_AUTOSCALE,
                        chkAutoScale));
  AddEditProperty(TQUB_Property_Edit.Create('["Min"]',
                    'Min', REGTYPE_FLOAT, DEF_MIN,
                    txtMin,
                    AcceptFloatBetween(MIN_MIN, MAX_MIN),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Max"]',
                    'Max', REGTYPE_FLOAT, DEF_MAX,
                    txtMAX,
                    AcceptFloatBetween(MIN_MAX, MAX_MAX),
                    '') );

  RegKeyName            := 'Software\QUB\Histogram Dur';
  ProfileCat            := 'HistogramDur';

  prBeforeGetProperties := PropertiesToParams;

  WikiTopic             := 'Histogram#Duration';
end;

procedure TQUB_Histogram_Dur_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;


//******************************************************************************

procedure TQUB_Histogram_Dur_PropertiesForm.PropertiesToParams;
begin
  prBinCount     := Properties['Bin Count'].Data[0];
  prDurClass     := Properties['Dur Class'].Data[0];
  prDurLog       := 0<>Properties['Dur Log'].Data[0];
  prAutoscale    := 0<>Properties['Autoscale'].Data[0];
  prMin          := Properties['Min'].Data[0];
  prMax          := Properties['Max'].Data[0];
end;


//******************************************************************************
end.
