{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// properties dialog for second histogram in Viewport_Histogram


unit QUB_Histogram_FFT_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,
RegistryInterface,

QUB_Task_Histogram,
QUB_Spectrum,
QUB_Profiles,
QUB_Properties,
QUB_Task_Properties,
QUB_Node;


{$include 'QUB_Histogram_FFT_Properties.hct'}


const


DEF_BINCOUNT  = 100;
MIN_BINCOUNT  = 1;
MAX_BINCOUNT  = 10000;

DEF_FFTLOG    = 1;
DEF_FFTBARSORLINE = 1;

DEF_AUTOSCALE = 1;

DEF_MIN      :Double = 1;
MIN_MIN      :Double = 1e-7;
MAX_MIN      :Double = 1e7;

DEF_MAX      :Double = 1000;
MIN_MAX      :Double = 1e-7;
MAX_MAX      :Double = 1e7;

DEF_MAXWINDOW = 14;
MIN_MAXWINDOW = 5;
MAX_MAXWINDOW = 30;

DEF_WINDOWFUNC = FFTHIST_WINDOW_HANNING;

DEF_OVERLAP   = 1;
MIN_OVERLAP   = 0;
MAX_OVERLAP   = 6;

DEF_BINCOLOR  = clNavy;


type


TQUB_Histogram_FFT_PropertiesForm = class(TQUB_Task_PropertiesForm)


  Label1                          :TLabel;

  btnReset                        :TButton;

  txtBinCount                     :TEdit;
  udBinCount                      :TUpDown;
  chkFFTLog                       :TCheckBox;
  chkAutoScale                    :TCheckBox;
  txtMin                          :TEdit;
  Label2                          :TLabel;
  txtMax                          :TEdit;
  Label3                          :TLabel;
  Label4                          :TLabel;
  Label5                          :TLabel;
  Label6                          :TLabel;
  txtMaxWindow                    :TEdit;
  udMaxWindow                     :TUpDown;
  txtOverlap                      :TEdit;
  udOverlap                       :TUpDown;
  cmbWindow                       :TComboBox;
  chkFFTBarsOrLine                :TCheckBox;
  Label8                          :TLabel;
  panBinColor                     :TPanel;


  procedure FormCreate(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure udMaxWindowClick(Sender: TObject; Button: TUDBtnType);
  procedure udOverlapClick(Sender: TObject; Button: TUDBtnType);
  procedure cmbWindowChange(Sender: TObject);


protected


  prBinCount                      :Integer;
  prFFTLog                        :Boolean;
  prFFTBarsOrLine                 :Boolean;
  prAutoScale                     :Boolean;
  prMin                           :Double;
  prMax                           :Double;
  prMaxWindow                     :Integer;
  prWindowFunc                    :Integer;
  prOverlap                       :Integer;
  prBinColor                      :TColor;

  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;

public

  procedure PropertiesToParams;


  property BinCount               :Integer     read prBinCount;            
  property FFTLog                 :Boolean     read prFFTLog;              
  property FFTBarsOrLine          :Boolean     read prFFTBarsOrLine;       
  property AutoScale              :Boolean     read prAutoScale;
  property Min                    :Double      read prMin;                 
  property Max                    :Double      read prMax;                 
  property MaxWindow              :Integer     read prMaxWindow;           
  property WindowFunc             :Integer     read prWindowFunc;          
  property Overlap                :Integer     read prOverlap;             
  property BinColor               :TColor      read prBinColor;            


end;


implementation


{$R *.DFM}


uses


QUB_Viewport_Data;


//******************************************************************************
procedure TQUB_Histogram_FFT_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  udBinCount.Min      := MIN_BINCOUNT;
  udBinCount.Max      := MAX_BINCOUNT;
  udBinCount.Position := DEF_BINCOUNT;

  udMaxWindow.Min     := MIN_MAXWINDOW;
  udMaxWindow.Max     := MAX_MAXWINDOW;
  udMaxWindow.Position := DEF_MAXWINDOW;

  udOverlap.Min       := MIN_OVERLAP;
  udOverlap.Max       := MAX_OVERLAP;
  udOverlap.Position  := DEF_OVERLAP;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Bin Count"]',
                      'Bin Count', REGTYPE_INT, DEF_BINCOUNT,
                      udBinCount,
                      txtBinCount,
                      AcceptIntGreaterThan(1)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["FFT Log"]',
                        'FFT Log', REGTYPE_BOOL, DEF_FFTLOG,
                        chkFFTLog));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["FFT BarsOrLine"]',
                        'FFT BarsOrLine', REGTYPE_BOOL, DEF_FFTBARSORLINE,
                        chkFFTBarsOrLine));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Autoscale"]',
                        'Autoscale', REGTYPE_BOOL, DEF_AUTOSCALE,
                        chkAutoScale));
  AddEditProperty(TQUB_Property_Edit.Create('["Min"]',
                    'Min', REGTYPE_FLOAT, DEF_MIN,
                    txtMin,
                    AcceptFloatBetween(MIN_MIN, MAX_MIN),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Max"]',
                    'Max', REGTYPE_FLOAT, DEF_MAX,
                    txtMAX,
                    AcceptFloatBetween(MIN_MAX, MAX_MAX),
                    '') );
  AddColorPanelProperty(TQUB_Property_ColorPanel.Create('["Bin Color"]',
                          'Bin Color', REGTYPE_INT, Integer(clRed),
                          panBinColor));

  RegKeyName            := 'Software\QUB\Histogram FFT';
  ProfileCat            := 'HistogramFFT';

  prBeforeGetProperties := PropertiesToParams;

  //{$include 'QUB_Histogram_FFT_Properties.htx'}

  WikiTopic             := 'Histogram#Spectrum';
end;

procedure TQUB_Histogram_FFT_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;

procedure TQUB_Histogram_FFT_PropertiesForm.udMaxWindowClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  Properties['MaxWindow'].Data[0]   := udMaxWindow.Position;
  txtMaxWindow.Text := IntToStr(Power2(udMaxWindow.Position));
end;

procedure TQUB_Histogram_FFT_PropertiesForm.udOverlapClick(Sender: TObject;Button: TUDBtnType);
begin
  inherited;
  Properties['Overlap'].Data[0]   := udOverlap.Position;
  txtOverlap.Text := IntToStr(Power2(udOverlap.Position));
end;

procedure TQUB_Histogram_FFT_PropertiesForm.cmbWindowChange(Sender: TObject);
begin
  inherited;
  Properties['WindowFunc'].Data[0] := cmbWindow.ItemIndex;
end;


//******************************************************************************
procedure TQUB_Histogram_FFT_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Reg   :TRegistry;
  Prop  :INode;
  iVal  :Integer;
begin
  inherited;

  Reg   := nil; // will be created as needed by ReadXRegistry*
  try
    Prop  := Props['MaxWindow'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, 'MaxWindow', iVal) then begin
        Prop.SetData(iVal);
      end else begin
        Prop.SetData(DEF_MAXWINDOW);
      end;
    end;
    Prop  := Props['Overlap'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, 'Overlap', iVal) then begin
        Prop.SetData(iVal);
      end else begin
        Prop.SetData(DEF_OVERLAP);
      end;
    end;
    Prop  := Props['WindowFunc'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, 'WindowFunc', iVal) then begin
        Prop.SetData(iVal);
      end else begin
        Prop.SetData(DEF_WINDOWFUNC);
      end;
    end;
  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_Histogram_FFT_PropertiesForm.prUpdateComponents;
begin
  inherited;

  udMaxWindow.Position := Properties['MaxWindow'].Data[0];
  txtMaxWindow.Text    := IntToStr(Power2(udMaxWindow.Position));
  udOverlap.Position   := Properties['Overlap'].Data[0];
  txtOverlap.Text      := IntToStr(Power2(udOverlap.Position));
  cmbWindow.ItemIndex  := Properties['WindowFunc'].Data[0];
end;

procedure TQUB_Histogram_FFT_PropertiesForm.PropertiesToParams;
begin
  prBinCount           := Properties['Bin Count'].Data[0];
  prFFTLog             := 0<>Properties['FFT Log'].Data[0];
  prFFTBarsOrLine      := 0<>Properties['FFT BarsOrLine'].Data[0];
  prAutoScale          := 0<>Properties['Autoscale'].Data[0];
  prMin                := Properties['Min'].Data[0];
  prMax                := Properties['Max'].Data[0];
  prMaxWindow          := Properties['MaxWindow'].Data[0];
  prWindowFunc         := Properties['WindowFunc'].Data[0];
  prOverlap            := Properties['Overlap'].Data[0];
  prBinColor           := TColor(Properties['Bin Color'].Data[0]);
end;


//******************************************************************************
end.
