unit QUB_Histogram_Print_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// properties form for copying and printing any histogram

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls,
Printers, Math, Registry, Clipbrd,
TeeProcs, TeEngine, Chart, Series,

QUB_Profiles,
QUB_Properties, QUB_Task_Properties, 
QUB_Node;


const


DEF_LANDSCAPE        = 0;
DEF_BITMAP           = 0;
MIN_WIDTH            = 20;
MAX_WIDTH            = 10000;
DEF_WIDTH            = 600;
MIN_HEIGHT           = 20;
MAX_HEIGHT           = 10000;
DEF_HEIGHT           = 400;

DEF_STYLE          :Integer      = 1; // as such
DEF_MIN_X          :Double       = 0.0;
DEF_MIN_Y          :Double       = 0.0;
DEF_MAX_X          :Double       = 10.0;
DEF_MAX_Y          :Double       = 10.0;
DEF_MAJOR_X        :Double       = 1.0;
DEF_MAJOR_Y        :Double       = 1.0;
DEF_MINOR_X        :Double       = 0.5;
DEF_MINOR_Y        :Double       = 0.5;
DEF_LABELS_X       :Integer      = 1;
DEF_LABELS_Y       :Integer      = 1;
DEF_PRECISION_X    :Integer      = 2;
DEF_PRECISION_Y    :Integer      = 2;
DEF_LINEWIDTH      :Integer      = 2;
DEF_FONTSIZE       :Integer      = 7;
DEF_BARWIDTH       :Double       = 100.0;


type


TQUB_Histogram_Print_PropertiesForm = class(TQUB_Task_PropertiesForm)


  panOptions                      :TPanel;
   panStuff                        :TPanel;
  lblOrientation                  :TLabel;
  Panel3                          :TPanel;
  chkPortrait                     :TRadioButton;
  chkLandscape                    :TRadioButton;
  lblImageFormat                  :TLabel;
  Panel1                          :TPanel;
  chkPicture                      :TRadioButton;
  chkBitmap                       :TRadioButton;
  txtWidth                        :TEdit;
  txtHeight                       :TEdit;
  lblWidth                        :TLabel;
  lblHeight                       :TLabel;
  Label2                          :TLabel;
  txtMaxY                         :TEdit;
  txtMajorX                       :TEdit;
  txtMajorY                       :TEdit;
  txtMinorX                       :TEdit;
  txtMinorY                       :TEdit;
  txtPrecisionY                   :TEdit;
    txtLineWidth: TEdit;
    Label13: TLabel;
    txtPrecisionX: TEdit;
    Panel2: TPanel;
    chkAsDisplayed: TRadioButton;
    chkAsSuch: TRadioButton;
    Label1: TLabel;
    Label3: TLabel;
    btnGrabX: TButton;
    Label4: TLabel;
    btnGrabY: TButton;
    Label5: TLabel;
    txtMinX: TEdit;
    txtMinY: TEdit;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    chkLabelsX: TCheckBox;
    chkLabelsY: TCheckBox;
    Label10: TLabel;
    Label14: TLabel;
    Label11: TLabel;
    txtFontSize: TEdit;
    Label12: TLabel;
    txtBarWidth: TEdit;
    txtMaxX: TEdit;


  procedure FormCreate(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure btnOKClick(Sender: TObject);
    procedure btnGrabXClick(Sender: TObject);
    procedure btnGrabYClick(Sender: TObject);


protected

  prHist                          :TChart;
  prTree                          :INode;

  prDlg                           :TPrintDialog;

  prPrinting                      :Boolean;

  prDrawFrame                     :Boolean;
  prShowTitles                    :Boolean;
  prLandscape                     :Boolean;
  prDrawWidth                     :Integer;
  prDrawHeight                    :Integer;
  prBitmap                        :Boolean;

  prStyle                         :Integer;
  prMinX, prMinY                  :Double;
  prMaxX, prMaxY                  :Double;
  prMajorX, prMajorY              :Double;
  prMinorX, prMinorY              :Double;
  prLabelsX, prLabelsY            :Boolean;
  prPrecisionX, prPrecisionY      :Integer;
  prLineWidth                     :Integer;
  prFontSize                      :Integer;
  prBarWidth                      :Double;

  prCopyImage                     :Boolean;

  procedure prSetCopyImage(c :Boolean);

  procedure prPrint;
  procedure prCopy;

  procedure prDraw(Canvas :TCanvas; PenWidth :Integer; var PageRect :TRect);


public


  procedure PropertiesToParams;

  procedure UseProfile(Name :String);
  function Print(Histogram :TChart; Tree :INode;{Dlg :TPrintDialog;}CopyImage :Boolean; ShowProperties :Boolean) :Boolean;

  property Landscape              :Boolean           read prLandscape;
  property Bitmap                 :Boolean           read prBitmap;
  property DrawWidth              :Integer           read prDrawWidth;
  property DrawHeight             :Integer           read prDrawHeight;

  property CopyImage              :Boolean           read prCopyImage           write prSetCopyImage;

end;


implementation


uses


QUB_Global,
QUB_QVO, QUB_QVO_QVD_Global,
QUB_QVO_QVM;


{$R *.DFM}


//******************************************************************************
function TQUB_Histogram_Print_PropertiesForm.Print(Histogram :TChart; Tree :INode; CopyImage :Boolean; ShowProperties :Boolean) :Boolean;
var
  Dlg :TPrintDialog;
begin
  Result := False;
  if prPrinting then begin
    Result := True;
    Exit;
  end;

  prPrinting := True;
  Dlg        := nil;
  try
    prHist         := Histogram;
    prTree         := Tree;
    //prDlg          := Dlg;
    Self.CopyImage := CopyImage;

    if not CopyImage {Printing} then begin
      Dlg := TPrintDialog.Create(Self);

      Result := not Dlg.Execute;
      if Result then Exit;

      Printer.Copies := Dlg.Copies;
      Printer.Title  := 'QuB Histogram';
    end;

    if ShowProperties then ShowModal else btnOkClick(Self);
  finally
    prPrinting := False;
    Dlg.Free;
  end;

end;

procedure TQUB_Histogram_Print_PropertiesForm.prPrint;
var
  OldColor                :TColor;
  OldParent               :Boolean;
  PageRect                :TRect;
  MetaRect                :TRect;
  Meta                    :TMetafile;
begin
  if Landscape then Printer.Orientation := poLandscape else Printer.Orientation := poPortrait;

  try
    Printer.BeginDoc;

    try
      PageRect.Left   := Printer.PageWidth  div 20;
      PageRect.Top    := Printer.PageHeight div 20;
      PageRect.Right  := Printer.PageWidth  - PageRect.Left;
      PageRect.Bottom := Printer.PageHeight - PageRect.Top;

      if prStyle = 0 {as displayed} then begin
        OldColor  := prHist.Color;
        OldParent := prHist.ParentColor;
        Meta      := nil;
        try
          prHist.ParentColor := False;
          prHist.Color       := clWhite;

          SetRect(MetaRect, 0, 0, prDrawWidth, (Printer.PageHeight * prDrawWidth) div Printer.PageWidth);
          Meta := prHist.TeeCreateMetafile(True, MetaRect);
          Printer.Canvas.StretchDraw(PageRect, Meta);
        finally
          prHist.Color       := OldColor;
          prHist.ParentColor := OldParent;
          Meta.Free;
        end;
      end else begin
        prDraw(Printer.Canvas, (prLineWidth * Printer.PageWidth) div prDrawWidth, PageRect);
      end;
    finally

      Printer.EndDoc;

    end;

  except
  end;
end;

procedure TQUB_Histogram_Print_PropertiesForm.prCopy;
var
  OldColor                :TColor;
  OldParent               :Boolean;
  PageRect                :TRect;
  Canvas                  :TCanvas;
  Metafile                :TMetafile;
  Bitmap                  :TBitmap;
begin
  Metafile                := nil;
  Bitmap                  := nil;

  if prStyle = 0 {as displayed} then begin
    try
      OldColor  := prHist.Color;
      OldParent := prHist.ParentColor;
      try
        prHist.ParentColor := False;
        prHist.Color       := clNone;//clWhite;
        prHist.Align       := alNone;
        prHist.Width       := prDrawWidth;
        prHist.Height      := prDrawHeight;

        if Self.Bitmap then begin
          prHist.CopyToClipboardBitmap;
        end else begin
          prHist.CopyToClipboardMetafile(True);
        end;
      finally
        prHist.Align       := alClient;
        prHist.Color       := OldColor;
        prHist.ParentColor := OldParent;
      end;
    except
    end;
    Exit;
  end;

  try
    if Self.Bitmap then begin
      Bitmap             := TBitmap.Create;
      Bitmap.Width       := prDrawWidth;
      Bitmap.Height      := prDrawHeight;
      Bitmap.PixelFormat := pf8bit;
      Canvas             := Bitmap.Canvas;
    end else begin
      Metafile           := TMetafile.Create;
      Metafile.Width     := prDrawWidth;
      Metafile.Height    := prDrawHeight;
      Canvas             := TMetafileCanvas.Create(Metafile, 0);
    end;

    try
      try
        PageRect.Left   := 0;
        PageRect.Top    := 0;
        PageRect.Right  := prDrawWidth;
        PageRect.Bottom := prDrawHeight;

        prDraw(Canvas, prLineWidth, PageRect);

      finally
        if not Self.Bitmap then begin
          Canvas.Free;
          Clipboard.Assign(Metafile);
        end else begin
          Clipboard.Assign(Bitmap);
        end;
      end;

    except
    end;

  finally
    Metafile.Free;
    Bitmap.Free;
  end;
end;

procedure TQUB_Histogram_Print_PropertiesForm.prDraw(Canvas :TCanvas; PenWidth :Integer; var PageRect :TRect);
const
  MAJOR_LEN      :Integer         = 5;
  MINOR_LEN      :Integer         = 2;
var
  TextFormatX, TextFormatY        :String;
  TextSize                        :TSize;
  MajorLen, MinorLen              :Integer;
  // Chart*: excluding axis labels (just axis lines and hist)
  ChartWidth, ChartHeight         :Integer;
  ChartX, ChartY                  :Integer;
  // Plot*: excluding axis labels and lines (just hist area)
  PlotWidth, PlotHeight           :Integer;
  PlotX, PlotY                    :Integer;
  // Factor*: scale (x-x0) to Plot coords
  FactorX, FactorY                :Double;
  i, N                            :Integer;
  Bins, Bars, Line                :INode;
  x, y                            :Double;
  px, py                          :Integer;
  Text                            :String;
  ClipRgn                         :hRgn;
  OldClip                         :TRect;

  function XtoPlot(xx :Double): Integer;
  begin
    Result := Round((xx - prMinX) * FactorX) + PlotX;
  end;
  function YtoPlot(yy :Double): Integer;
  begin
    Result := - Round((yy - prMinY) * FactorY) + PlotY;
  end;
begin
  TextFormatX          := Format('%%.%df', [prPrecisionX]);
  TextFormatY          := Format('%%.%df', [prPrecisionY]);
  Canvas.Font.Size    := prFontSize;
  TextSize            := Canvas.TextExtent(Format(TextFormatY, [prMajorY*Floor(prMaxY/prMajorY)]));
  if not prLabelsX then TextSize.cy := 0;
  if not prLabelsY then TextSize.cx := 0;
  ChartX              := PageRect.Left + TextSize.cx;
  ChartY              := PageRect.Bottom - TextSize.cy;
  ChartWidth          := (PageRect.Right - PageRect.Left) - TextSize.cx;
  ChartHeight         := (PageRect.Bottom - PageRect.Top) - TextSize.cy;
  MajorLen            := MAJOR_LEN * PenWidth;
  MinorLen            := MINOR_LEN * PenWidth;
  PlotWidth           := ChartWidth - MajorLen;
  PlotHeight          := ChartHeight - MajorLen;
  PlotX               := ChartX + MajorLen;
  PlotY               := ChartY - MajorLen;
  FactorX             := PlotWidth / (prMaxX - prMinX);
  FactorY             := PlotHeight / (prMaxY - prMinY);

  with Canvas do begin
    Pen.Width         := PenWidth;
    // Axes
    px                := XtoPlot(prMinX);
    py                := YtoPlot(prMinY);
    MoveTo(PageRect.Right, py);
    LineTo(px, py);
    LineTo(px, PageRect.Top);
    // Majors and maybe labels
    x := prMajorX * Floor(prMinX / prMajorX);
    while x <= prMaxX do begin
      if x >= prMinX then begin
        px := XtoPlot(x);
        MoveTo(px, PlotY+MajorLen);
        LineTo(px, PlotY+PenWidth);
        if prLabelsX then begin
          Text := Format(TextFormatX, [x]);
          TextOut(Min(PageRect.Right-TextWidth(Text), px-(TextWidth(Text) div 2)), ChartY, Text);
        end;
      end;
      x := x + prMajorX;
    end;
    y := prMajorY * Floor(prMinY / prMajorY);
    while y <= prMaxY do begin
      if y >= prMinY then begin
        py := YtoPlot(y);
        MoveTo(PlotX-MajorLen, py);
        LineTo(PlotX-PenWidth, py);
        if prLabelsY then begin
          Text := Format(TextFormatY, [y]);
          TextOut(PageRect.Left, Max(0, py-(TextHeight(Text) div 2)), Text);
        end;
      end;
      y := y + prMajorY;
    end;
    // Minors
    x := prMinorX * Floor(prMinX / prMinorX);
    while x <= prMaxX do begin
      if x >= prMinX then begin
        px := XtoPlot(x);
        MoveTo(px, PlotY+MinorLen);
        LineTo(px, PlotY+PenWidth);
      end;
      x := x + prMinorX;
    end;
    y := prMinorY * Floor(prMinY / prMinorY);
    while y <= prMaxY do begin
      if y >= prMinY then begin
        py := YtoPlot(y);
        MoveTo(PlotX-MinorLen, py);
        LineTo(PlotX-PenWidth, py);
      end;
      y := y + prMinorY;
    end;

    // plot
    OldClip         := ClipRect;
    ClipRgn         := CreateRectRgn(PlotX, PageRect.Top, PlotX+PlotWidth, PlotY); // left, top, right, bottom
    SelectClipRgn(Canvas.Handle, ClipRgn);
    DeleteObject(ClipRgn);

    Bins       := prTree['Bins'];
    Bars       := prTree['Bars'];
    N          := Min(Bins.DataCount, Bars.DataCount);

    // bars
    for i := 0 to N - 1 do begin
      py       := YtoPlot(Bars.Data[i]);
      x        := Bins.Data[i];
      if i = 0 then px := XtoPlot(x - ((Bins.Data[1] - x) / 2.0) * (prBarWidth / 100.0))
      else          px := XtoPlot(x - ((x - Bins.Data[i-1]) / 2.0) * (prBarWidth / 100.0));
      MoveTo(px, PlotY);
      LineTo(px, py);
      if i = (N-1) then px := XtoPlot(x + ((x - Bins.Data[i-1]) / 2.0) * (prBarWidth / 100.0))
      else              px := XtoPlot(x + ((Bins.Data[i+1] - x) / 2.0) * (prBarWidth / 100.0));
      LineTo(px, py);
      LineTo(px, PlotY);
    end;

    // lines
    Line       := prTree.Find('Line');
    while not Line.IsNull do begin
      MoveTo(XtoPlot(Bins.Data[0]), YtoPlot(Line.Data[0]));
      for i := 1 to N - 1 do begin
        LineTo(XtoPlot(Bins.Data[i]), YtoPlot(Line.Data[i]));
      end;

      Line     := Line.SiblingSameName;
    end;
    Line       := prTree.Find('Component');
    while not Line.IsNull do begin
      MoveTo(XtoPlot(Bins.Data[0]), YtoPlot(Line.Data[0]));
      for i := 1 to N - 1 do begin
        LineTo(XtoPlot(Bins.Data[i]), YtoPlot(Line.Data[i]));
      end;

      Line     := Line.SiblingSameName;
    end;

    ClipRgn         := CreateRectRgn(OldClip.Left, OldClip.Top, OldClip.Right, OldClip.Bottom); // left, top, right, bottom
    SelectClipRgn(Canvas.Handle, ClipRgn);
    DeleteObject(ClipRgn);
  end;
end;


//******************************************************************************
procedure TQUB_Histogram_Print_PropertiesForm.prSetCopyImage(c :Boolean);
begin
  prCopyImage            := c;
  chkBitmap.Enabled      := c;
  chkPicture.Enabled     := c;
  lblImageFormat.Enabled := c;
  lblWidth.Enabled       := c;
  lblHeight.Enabled      := c;
  txtWidth.Enabled       := c;
  txtHeight.Enabled      := c;
  // anything else not relevant for printing?
  if c then btnOK.Caption := '&Copy' else btnOK.Caption := '&Print';
end;

//******************************************************************************
procedure TQUB_Histogram_Print_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Orientation"]',
                          '', REGTYPE_NONE, DEF_LANDSCAPE)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkPortrait,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkLandscape,
                                       1,
                                       1))
                       );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Format"]',
                          '', REGTYPE_NONE, DEF_BITMAP)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkPicture,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkBitmap,
                                       1,
                                       1))
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["Draw Width"]',
                    'Draw Width', REGTYPE_INT, DEF_WIDTH,
                    txtWidth,
                    AcceptIntBetween(MIN_WIDTH, MAX_WIDTH),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Draw Height"]',
                    'Draw Height', REGTYPE_INT, DEF_HEIGHT,
                    txtHeight,
                    AcceptIntBetween(MIN_HEIGHT, MAX_HEIGHT),
                    '') );

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Style"]',
                          '', REGTYPE_NONE, DEF_STYLE)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkAsDisplayed,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkAsSuch,
                                       1,
                                       1)
                                         .Enables(panStuff)
                                      )
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["MinX"]',
                    '', REGTYPE_NONE, DEF_MIN_X,
                    txtMinX,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MinY"]',
                    '', REGTYPE_NONE, DEF_MIN_Y,
                    txtMinY,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxX"]',
                    '', REGTYPE_NONE, DEF_MAX_X,
                    txtMaxX,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxY"]',
                    '', REGTYPE_NONE, DEF_MAX_Y,
                    txtMaxY,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MajorX"]',
                    '', REGTYPE_NONE, DEF_MAJOR_X,
                    txtMajorX,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MajorY"]',
                    '', REGTYPE_NONE, DEF_MAJOR_Y,
                    txtMajorY,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MinorX"]',
                    '', REGTYPE_NONE, DEF_MINOR_X,
                    txtMinorX,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MinorY"]',
                    '', REGTYPE_NONE, DEF_MINOR_Y,
                    txtMinorY,
                    AcceptFloat,
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["LabelsX"]',
                        '', REGTYPE_NONE, DEF_LABELS_X,
                        chkLabelsX));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["LabelsY"]',
                        '', REGTYPE_NONE, DEF_LABELS_Y,
                        chkLabelsY));
  AddEditProperty(TQUB_Property_Edit.Create('["PrecisionX"]',
                    '', REGTYPE_NONE, DEF_PRECISION_X,
                    txtPrecisionX,
                    AcceptIntGreaterThanOrEqualTo(0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PrecisionY"]',
                    '', REGTYPE_NONE, DEF_PRECISION_Y,
                    txtPrecisionY,
                    AcceptIntGreaterThanOrEqualTo(0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["LineWidth"]',
                    '', REGTYPE_NONE, DEF_LINEWIDTH,
                    txtLineWidth,
                    AcceptIntGreaterThan(0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["FontSize"]',
                    '', REGTYPE_NONE, DEF_FONTSIZE,
                    txtFontSize,
                    AcceptIntGreaterThan(0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["BarWidth"]',
                    '', REGTYPE_NONE, DEF_BARWIDTH,
                    txtBarWidth,
                    AcceptFloatBetween(0.0, 100.0),
                    '') );

  prBeforeGetProperties := PropertiesToParams;

  WikiTopic             := 'Copy_Histogram';

end;

procedure TQUB_Histogram_Print_PropertiesForm.UseProfile(Name :String);
begin
  ProfileCat            := 'CopyHist '+Name;
end;

procedure TQUB_Histogram_Print_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;

//******************************************************************************
procedure TQUB_Histogram_Print_PropertiesForm.btnOKClick(Sender: TObject);
begin
  inherited;
  if CopyImage then begin
    prCopy;
  end else begin
    prPrint;
  end;
end;

procedure TQUB_Histogram_Print_PropertiesForm.btnGrabXClick(Sender: TObject);
var
  Props        :INode;
begin
  inherited;
  Props        := TNode.Create('');
  Props['MinX'].SetData( prHist.BottomAxis.Minimum );
  Props['MaxX'].SetData( prHist.BottomAxis.Maximum );
  prProfile.SetProperties(Props);
end;

procedure TQUB_Histogram_Print_PropertiesForm.btnGrabYClick(Sender: TObject);
var
  Props        :INode;
begin
  inherited;
  Props        := TNode.Create('');
  Props['MinY'].SetData( prHist.LeftAxis.Minimum );
  Props['MaxY'].SetData( prHist.LeftAxis.Maximum );
  prProfile.SetProperties(Props);
end;

//******************************************************************************
procedure TQUB_Histogram_Print_PropertiesForm.PropertiesToParams;
begin
  prLandscape        := 0<>Properties['Orientation'].Data[0];
  prDrawWidth        := Properties['Draw Width'].Data[0];
  prDrawHeight       := Properties['Draw Height'].Data[0];
  prBitmap           := 0<>Properties['Format'].Data[0];
  prStyle            := Properties['Style'].Data[0];
  prMinX             := Properties['MinX'].Data[0];
  prMinY             := Properties['MinY'].Data[0];
  prMaxX             := Properties['MaxX'].Data[0];
  prMaxY             := Properties['MaxY'].Data[0];
  prMajorX           := Properties['MajorX'].Data[0];
  prMajorY           := Properties['MajorY'].Data[0];
  prMinorX           := Properties['MinorX'].Data[0];
  prMinorY           := Properties['MinorY'].Data[0];
  prLabelsX          := 0<>Properties['LabelsX'].Data[0];
  prLabelsY          := 0<>Properties['LabelsY'].Data[0];
  prPrecisionX       := Properties['PrecisionX'].Data[0];
  prPrecisionY       := Properties['PrecisionY'].Data[0];
  prLineWidth        := Properties['LineWidth'].Data[0];
  prFontSize         := Properties['FontSize'].Data[0];
  prBarWidth         := Properties['BarWidth'].Data[0];
end;


//******************************************************************************
end.
