{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2003

// Imports QUB_IdlSegments from qubopt, and wraps it in a thin layer of Delphi convenience.
// The C++ object already handles thread-safety, segmentation, qubtree-conversion, and even read/writing DWT and EVL files.


unit QUB_Idealization;


interface


uses


SysUtils, Classes, SyncObjs, Math, Windows,

PythonEngine,

Alloc, AllocTypes,

QUB_Global,
QUB_Node,
QUB_QFS_QIF_Segments,
QUB_Console, QUB_Rich_Console;


type


TQUB_IdlSegmentsIter = class
public
  procedure Free; virtual; cdecl; abstract;

  function Valid: Integer; virtual; cdecl; abstract;
  procedure Read(F, L, C :TPInteger; A :TPDouble); virtual; cdecl; abstract;
  procedure Next; virtual; cdecl; abstract;
  procedure Prev; virtual; cdecl; abstract;

  function Copy: TQUB_IdlSegmentsIter; virtual; cdecl; abstract;
end;


TQUB_IdlSegments_Impl = class
public
  procedure Free; virtual; cdecl; abstract;
  procedure Lock; virtual; cdecl; abstract;
  procedure Unlock; virtual; cdecl; abstract;

  function  GetSampling: Double; virtual; cdecl; abstract;
  procedure SetSampling(x :Double); virtual; cdecl; abstract;
  function  GetClsCount: Integer; virtual; cdecl; abstract;
  function  GetSegCount: Integer; virtual; cdecl; abstract;
  function  GetSegmentation: TQUB_QFS_PNodeMem; virtual; cdecl; abstract;

  function  ToTree: TQUB_QFS_PNodeMem; virtual; cdecl; abstract;
  function  ReadDWT(FileName :PChar): Integer; virtual; cdecl; abstract;
  function  GetDwells(First, Last, Fragments, WithDurs :Integer): TQUB_QFS_PNodeMem; virtual; cdecl; abstract;
  procedure GetDwellsInto(First, Last, Fragments, WithDurs :Integer;
                      DataSet :TQUB_QFS_PNodeMem; InsertAsFirst :Integer); virtual; cdecl; abstract;
  procedure GetDwellsIntoSegs(Seg :TQUB_QFS_PNodeMem; WithDurs :Integer); virtual; cdecl; abstract;
  procedure GetDwellsWDrop(First, Last, WithDurs, DropFirstLast, DropIfCls, DropCls :Integer; DataSet :TQUB_QFS_PNodeMem; InsertAsFirst :Integer); virtual; cdecl; abstract;
  function  SetDwells(DataSet :TQUB_QFS_PNodeMem): Integer; overload; virtual; cdecl; abstract;
  function  SetDwellsSeg(Seg :TQUB_QFS_PNodeMem): Integer; overload; virtual; cdecl; abstract;
  function  SetDwells(Count :Integer; Buf :TQUB_QFS_QIF_PAFirstLastClass): Integer; overload; virtual; cdecl; abstract;
  function  SetDwells(Count :Integer; FF, LL, CC :TPAInteger): Integer; overload; virtual; cdecl; abstract;
  function  Clear: Integer; virtual; cdecl; abstract;
  function  AddSeg(First, Last :Integer): Integer; virtual; cdecl; abstract;
  function  RemovePoints(First, Last :Integer): Integer; virtual; cdecl; abstract;
  function  InsertPoints(First, Last :Integer): Integer; virtual; cdecl; abstract;
  function  BreakSeg(At :Integer): Integer; virtual; cdecl; abstract;
  function  JoinSegs(At :Integer): Integer; virtual; cdecl; abstract;
  function  Erase(First, Last :Integer): Integer; virtual; cdecl; abstract;
  function  Join(First, Last :Integer): Integer; virtual; cdecl; abstract;

  function  BeginIter: TQUB_IdlSegmentsIter; virtual; cdecl; abstract;
  function  EndIter:   TQUB_IdlSegmentsIter; virtual; cdecl; abstract;
  function  FindIter(Point :Integer):  TQUB_IdlSegmentsIter; virtual; cdecl; abstract;

  function  UseTempFile(DoUse :Integer): Integer; virtual; cdecl; abstract;

  function  SaveASCII(FileName :PChar): Integer; virtual; cdecl; abstract;
end;

function TQUB_IdlSegments_Create(Sampling :Double): TQUB_IdlSegments_Impl; stdcall; external 'qubopt.dll' name '_QUB_IdlSegments_Create@8';

function QUB_Idl_ReadDWT_(FileName :PChar): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_QUB_Idl_ReadDWT@4';
function QUB_Idl_ReadDWT(FileName :String): INode;

function QUB_Idl_ReadEVL_(FileName :PChar): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_QUB_Idl_ReadEVL@4';
function QUB_Idl_ReadEVL(FileName :String): INode;

function QUB_Idl_ReadTAC_(FileName :PChar): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_QUB_Idl_ReadTAC@4';
function QUB_Idl_ReadTAC(FileName :String): INode;

function QUB_Idl_ReadPRT(FileName :String; Acrit :String): INode;

type

TQUB_IdlSegments = class
protected

  function  prGetSampling: Double;
  procedure prSetSampling(x :Double);
  function  prGetSegCount: Integer;
  function  prGetClsCount: Integer;

  procedure prEnsureFirstsLasts(DataSet :INode);
  procedure prEnsureFirstsLastsSeg(Seg :INode);

public
  Impl      :TQUB_IdlSegments_Impl;

  OnEdit    :TNotifyEvent;

  constructor Create(Sampling :Double);
  destructor  Destroy; override;

  procedure Lock;
  procedure Unlock;

  function  ToTree: INode;
  function  ReadDWT(FileName :String): Integer;
  function  GetSegmentation: INode;
  function  GetDwells(First, Last, Fragments, WithDurs :Integer): INode; overload;
  procedure GetDwellsInto(First, Last, Fragments, WithDurs :Integer; DataSet :INode; InsertAsFirst :Integer); overload;
  procedure GetDwellsIntoSegs(Seg :INode; WithDurs :Integer);
  procedure GetDwellsWDrop(First, Last, WithDurs, DropFirstLast, DropIfCls, DropCls :Integer; DataSet :INode; InsertAsFirst :Integer);
  function  SetDwells(DataSet :INode): Integer; overload;
  function  SetDwellsSeg(Seg :INode): Integer; overload;
  function  SetDwells(Count :Integer; Buf :TQUB_QFS_QIF_PAFirstLastClass): Integer; overload;
  function  SetDwells(Count :Integer; FF, LL, CC :TPAInteger): Integer; overload;
  function  SetDwell(F, L, C :Integer): Integer;
  function  Clear: Integer;
  function  AddSeg(First, Last :Integer): Integer;
  function  RemovePoints(First, Last :Integer): Integer;
  function  InsertPoints(First, Last :Integer): Integer;
  function  BreakSeg(At :Integer): Integer;
  function  JoinSegs(At :Integer): Integer;
  function  Erase(First, Last :Integer): Integer;
  function  Join(First, Last :Integer): Integer;

  function  BeginIter: TQUB_IdlSegmentsIter;
  function  EndIter:   TQUB_IdlSegmentsIter;
  function  FindIter(Point :Integer):  TQUB_IdlSegmentsIter;

  function  UseTempFile(DoUse :Integer): Integer;

  function SaveASCII(const FileName :String): Boolean;

  property Sampling      :Double     read prGetSampling       write prSetSampling;
  property ClsCount      :Integer    read prGetClsCount;
  property SegCount      :Integer    read prGetSegCount;
end;


implementation


uses

QUB_Python_Thread,
QUB_Viewport_Python,
QUB_Node_Python;


constructor TQUB_IdlSegments.Create(Sampling :Double);
begin
  Impl  := TQUB_IdlSegments_Create(Sampling);
end;

destructor  TQUB_IdlSegments.Destroy;
begin
  try
    Impl.Free;
  except
  end;

  inherited;
end;

function  TQUB_IdlSegments.prGetSampling: Double;
begin
  Result := Impl.GetSampling;
end;

procedure TQUB_IdlSegments.prSetSampling(x :Double);
begin
  Impl.SetSampling(x);
end;

function  TQUB_IdlSegments.prGetClsCount: Integer;
begin
  Result := Impl.GetClsCount;
end;

function  TQUB_IdlSegments.prGetSegCount: Integer;
begin
  Result := Impl.GetSegCount;
end;

procedure TQUB_IdlSegments.prEnsureFirstsLasts(DataSet :INode);
var
  Seg :INode;
begin
  Seg := DataSet.Find('Segment');
  //TO DO: LM. There is a problem with this, in that there is a sibling, when there shouldn't be
  while not Seg.IsNull do begin
    prEnsureFirstsLastsSeg(Seg);
    Seg := Seg.SiblingSameName;
    //Same here
    if Seg['Durations'].DataPtr = nil then begin
      DataSet.RemoveChild(Seg);
      Break;
    end;
  end;
end;

procedure TQUB_IdlSegments.prEnsureFirstsLastsSeg(Seg :INode);
var
  SampMS          :Double;
  DwellCount      :Integer;
  Firsts, Lasts   :TPAInteger;
  Durations       :TPASingle;
  At              :Integer;
  i               :Integer;
begin
  try
    DwellCount      := Seg['DwellCount'].Data[0];
  except
    DwellCount      := 0;
  end;
  if LongInt(Seg.Find('Firsts').DataCount) < DwellCount then begin
    SampMS        := Sampling * 1.0e3;
    Durations     := TPASingle( Seg['Durations'].DataPtr );
    //TO DO: LM. I found the above problem, and as a quick fix I added this check here
    if Durations <> nil then begin
      try
        At          := Seg.Data[0];
      except
        At          := 0;
      end;

      Seg['Firsts'].SetupNumData( NODE_DATATYPE_LONGINT, DwellCount, 1 );
      Firsts        := TPAInteger( Seg['Firsts'].DataPtr );
      Seg['Lasts'].SetupNumData( NODE_DATATYPE_LONGINT, DwellCount, 1 );
      Lasts         := TPAInteger( Seg['Lasts'].DataPtr );

      for i := 0 to DwellCount - 1 do begin
        Firsts[i]   := At;
        At          := At + Round( Durations[i] / SampMS );
        Lasts[i]    := At - 1;
      end;
    end;//Durations <> nil
  end;
end;

procedure TQUB_IdlSegments.Lock;
begin
  Impl.Lock;
end;

procedure TQUB_IdlSegments.Unlock;
begin
  Impl.Unlock;
end;

function  TQUB_IdlSegments.ToTree: INode;
begin
  Result := TNode.CreateUsing( Impl.ToTree );
  Node_DECREF( Result.GetNodeMem );
end;

function  TQUB_IdlSegments.ReadDWT(FileName :String): Integer;
begin
  Result := Impl.ReadDWT(PChar(FileName));
end;

function  TQUB_IdlSegments.GetSegmentation: INode;
begin
  Result := TNode.CreateUsing( Impl.GetSegmentation );
  Node_DECREF( Result.GetNodeMem );
end;

function  TQUB_IdlSegments.GetDwells(First, Last, Fragments, WithDurs :Integer): INode;
begin
  Result := TNode.CreateUsing( Impl.GetDwells(First, Last, Fragments, WithDurs) );
  Node_DECREF( Result.GetNodeMem );
end;

procedure TQUB_IdlSegments.GetDwellsInto(First, Last, Fragments, WithDurs :Integer; DataSet :INode; InsertAsFirst :Integer);
begin
  Impl.GetDwellsInto(First, Last, Fragments, WithDurs, DataSet.GetNodeMem, InsertAsFirst);
end;

procedure TQUB_IdlSegments.GetDwellsIntoSegs(Seg :INode; WithDurs :Integer);
begin
  Impl.GetDwellsIntoSegs(Seg.GetNodeMem, WithDurs);
end;

procedure TQUB_IdlSegments.GetDwellsWDrop(First, Last, WithDurs, DropFirstLast, DropIfCls, DropCls :Integer; DataSet :INode; InsertAsFirst :Integer);
begin
  Impl.GetDwellsWDrop(First, Last, WithDurs, DropFirstLast, DropIfCls, DropCls, DataSet.GetNodeMem, InsertAsFirst);
end;

function  TQUB_IdlSegments.SetDwells(DataSet :INode): Integer;
begin
  if DataSet.Find('Segment').Find('DwellCount').DataCount = 0 then begin
    Result := 1;
    Exit;
  end;

  prEnsureFirstsLasts(DataSet);
  Result := Impl.SetDwells(DataSet.GetNodeMem);
  if Assigned(OnEdit) then OnEdit(Self);
end;

function  TQUB_IdlSegments.SetDwellsSeg(Seg :INode): Integer;
begin
  if Seg.Find('DwellCount').DataCount = 0 then begin
    Result := 1;
    Exit;
  end;

  prEnsureFirstsLastsSeg(Seg);
  Result := Impl.SetDwellsSeg(Seg.GetNodeMem);
  if Assigned(OnEdit) then OnEdit(Self);
end;

function  TQUB_IdlSegments.SetDwells(Count :Integer; Buf :TQUB_QFS_QIF_PAFirstLastClass): Integer;
begin
  Result := Impl.SetDwells(Count, Buf);
  if Assigned(OnEdit) then OnEdit(Self);
end;

function  TQUB_IdlSegments.SetDwells(Count :Integer; FF, LL, CC :TPAInteger): Integer;
begin
  Result := Impl.SetDwells(Count, FF, LL, CC);
  if Assigned(OnEdit) then OnEdit(Self);
end;

function  TQUB_IdlSegments.SetDwell(F, L, C :Integer): Integer;
begin
  Result := SetDwells(1, @F, @L, @C);
end;

function  TQUB_IdlSegments.Clear: Integer;
begin
  Result := Impl.Clear;
  if Assigned(OnEdit) then OnEdit(Self);
end;

function  TQUB_IdlSegments.AddSeg(First, Last :Integer): Integer;
begin
  Result := Impl.AddSeg(First, Last);
end;

function  TQUB_IdlSegments.RemovePoints(First, Last :Integer): Integer;
var
  OldSegCount  :Integer;
begin
  OldSegCount  := SegCount;
  Result       := Impl.RemovePoints(First, Last);
  if OldSegCount < SegCount then Impl.JoinSegs(First);
end;

function  TQUB_IdlSegments.InsertPoints(First, Last :Integer): Integer;
begin
  Result := Impl.InsertPoints(First, Last);
end;

function  TQUB_IdlSegments.BreakSeg(At :Integer): Integer;
begin
  Result := Impl.BreakSeg(At);
end;

function  TQUB_IdlSegments.JoinSegs(At :Integer): Integer;
begin
  Result := Impl.JoinSegs(At);
end;

function  TQUB_IdlSegments.Erase(First, Last :Integer): Integer;
begin
  Result := Impl.Erase(First, Last);
  if Assigned(OnEdit) then OnEdit(Self);
end;

function  TQUB_IdlSegments.Join(First, Last :Integer): Integer;
begin
  Result := Impl.Join(First, Last);
  if Assigned(OnEdit) then OnEdit(Self);
end;

function  TQUB_IdlSegments.BeginIter: TQUB_IdlSegmentsIter;
begin
  Result := Impl.BeginIter;
end;

function  TQUB_IdlSegments.EndIter:   TQUB_IdlSegmentsIter;
begin
  Result := Impl.EndIter;
end;

function  TQUB_IdlSegments.FindIter(Point :Integer):  TQUB_IdlSegmentsIter;
begin
  Result := Impl.FindIter(Point);
end;

function  TQUB_IdlSegments.UseTempFile(DoUse :Integer): Integer;
begin
  Result := Impl.UseTempFile(DoUse);
end;

function TQUB_IdlSegments.SaveASCII(const FileName :String): Boolean;
begin
  Result := 0 <> Impl.SaveASCII(PChar(FileName));
end;

function QUB_Idl_ReadDWT(FileName :String): INode;
begin
  Result := TNode.CreateUsing( QUB_Idl_ReadDWT_(PChar(FileName)) );
  Node_DECREF( Result.GetNodeMem );
end;

function QUB_Idl_ReadEVL(FileName :String): INode;
begin
  Result := TNode.CreateUsing( QUB_Idl_ReadEVL_(PChar(FileName)) );
  Node_DECREF( Result.GetNodeMem );
end;

function QUB_Idl_ReadTAC(FileName :String): INode;
begin
  Result := TNode.CreateUsing( QUB_Idl_ReadTAC_(PChar(FileName)) );
  Node_DECREF( Result.GetNodeMem );
end;

function QUB_Idl_ReadPRT(FileName :String; Acrit :String): INode;
var
  Func           :PPyObject;
  Args           :PPyObject;
  PyResult       :PPyObject;
begin
  Func           := nil;
  Args           := nil;
  PyResult       := nil;
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    try
      Run_CommandAsString('import ReadPRT', single_input);
      Func := EvalString('ReadPRT.ReadPRT');
      if Func = nil then begin
        PyErr_Print;
        raise Exception.Create('Error loading ReadPRT.py -- see PythonScripts window for details.');
      end;

      Args := PyTuple_New(2);
      PyTuple_SetItem(Args, 0, PyString_FromString(PChar(FileName)));
      PyTuple_SetItem(Args, 1, PyString_FromString(PChar(Acrit)));
      PyResult := PyEval_CallObject(Func, Args);
      if PyResult = nil then begin
        PyErr_Print;
        raise Exception.Create('Exception in ReadPRT -- see PythonScripts window for details.');
      end else if PyNode_IsNode( PyResult ) then begin
        Result := PyNode_ToNode( PyResult );
      end else begin
        Result := TNode.CreateNull;
      end;
    finally
      Py_XDECREF(Func);
      Py_XDECREF(Args);
      Py_XDECREF(PyResult);
      PythonThread.EndPython;
    end;
  end;
end;


// *****************************************************************************
end.
