{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// data structures and constants for QUB_Task_IdealizeMT and QUB_Thread_Idealize (SKM and HalfAmp)


unit QUB_Idealize_Global;


interface


uses


AllocTypes,

QUB_QVO_QVD_Global,
QUB_Global,
QUB_Thread,
QUB_QFS_Channels,

QUB_QFS, QUB_Node;


type


TQUB_Thread_Idealize_Chunk = record

  Chunk                    :TQUB_Thread_Chunk;

  Sampling                 :Single;//Seconds
  Scaling                  :Single;
  Offset                   :Single;

  DataIs16Bit              :Integer;
  Data                     :TPASingle;//De/Allocated in DataCallback

  SegCount                 :Integer;
  SegFirsts                :TPAInteger;
  SegLasts                 :TPAInteger;
  SegClasses               :TPAInteger;

  ClassCount               :Integer;

  Amps                     :TPASingle;
  Stds                     :TPASingle;
  Occupancy                :TPASingle;
  Lifetime                 :TPASingle;
  NEvents                  :TPAInteger;

  StateCount               :Integer;
  AMatrix                  :TPMSingle;

  LL                       :Single;
  Iterations               :Integer;

  HistBinCount             :Integer;
  Histogram                :TPMSingle;
  //Histogram[class, bin]  (row, col)
  //row 0 is class 0 pdf
  //row SegAmpCount is Sum pdf
  //row SegAmpCount + 1 is Bin x
  //row SegAmpCount + 2 is Bin y

  BaseTime                 :Double;

  Report                   :PChar;
  ReportSize               :Integer;

  ResultNode               :TQUB_QFS_PNodeMem;

end;
TQUB_Thread_Idealize_PChunk = ^TQUB_Thread_Idealize_Chunk;


TQUB_Idealize_Params_HalfAmp = record

  ClassCount                 :Integer;
  Classes                    :TPAInteger;
  HalfAmps                   :TPAInteger;
  Amps                       :TPAInteger;
  IdxAmps                    :TPAInteger;
  AmpVals                    :TPADouble;
  StdVals                    :TPADouble;
  ComputeStats               :Boolean;

end;


TQUB_SKM_Output              = (OUT_STATE, OUT_CLASS, OUT_GROUP);


TQUB_Idealize_Params_SKM     = record

  HistBinCount               :Integer; //Should it be in both places ?!
  FixQ                       :Integer;
  FixA                       :TPAInteger;
  FixS                       :TPAInteger;
  LLConv                     :Double;
  MaxIter                    :Integer;
  Output                     :Integer;//State, Class, Group
  Viterbi                    :Integer;
 
end;

TQUB_Idealize_Params         = record

  Method                     :Integer;

  LibName                    :PChar;
  LibNameSize                :Integer;

  Model                      :PChar;
  ModelSize                  :Integer;
  ModelChannelCount          :Integer;
  ModelObject                :TObject;

  HalfAmp                    :TQUB_Idealize_Params_HalfAmp;
  SKM                        :TQUB_Idealize_Params_SKM;

  Remote                     :Boolean;
  Host                       :String;
  Port                       :Integer;

  WriteOutput                :Boolean;

  DropFirstLast              :Boolean;
  DropIfClass                :Boolean;
  DropClass                  :Integer;

  Sampling                   :Double;
  Scaling                    :Double;
  Offset                     :Double;

  ChopIfLonger               :Boolean;
  ChopIfLongerPoints         :Integer;

  LargeMemory                :Boolean;

  BaselineTrack              :Boolean;
  BaselineTrackStd           :Single;

  TestMethod                 :Integer;
  BaselineReEstVar           :Boolean;
  Baseline2State             :Boolean;
  BaselineOutput             :Boolean;

  ApplyDeadTime              :Boolean;
  IdlDeadTime                :Boolean;

  ClearIdl                   :Boolean;

end;
TQUB_Idealize_PParams         = ^TQUB_Idealize_Params;


function IdlMethodToString(x :Integer): String;
function StringToIdlMethod(const x :String): Integer;


const


METHOD_HALFAMP                = 0;
METHOD_SKM                    = 1;
METHOD_TEST_SKM               = 2;
METHOD_TEST_BAUMWELCH         = 3;
METHOD_TEST_VITERBIGROUP      = 4;
DEF_METHOD                    = METHOD_SKM;

DEF_SKMDLLNAME                = 'qubskm.dll';

DEF_HASTATS                   = 0;

DEF_SKMHISTBINCOUNT           = 50;
MIN_SKMHISTBINCOUNT           = 2;
MAX_SKMHISTBINCOUNT           = 500;

DEF_SKMOUTPUT                 = Integer(OUT_CLASS);
DEF_SKMVITERBI                = 0;

DEF_SKMFIXQ                   = 0;
DEF_SKMFIXA                   = 0;
DEF_SKMFIXS                   = 0;

MIN_SKMLLCONV                :Double = 1e-10;
MAX_SKMLLCONV                :Double = 1;
DEF_SKMLLCONV                :Double = 1e-2;

MIN_SKMMAXITER                = 1;
MAX_SKMMAXITER                = 1000;
DEF_SKMMAXITER                = 10;

DEF_WRITEOUTPUT               = 1;

DEF_DROPFIRSTLAST             = 0;
DEF_DROPIFCLASS               = 1;
DEF_DROPCLASS                 = 0;

DEF_DEADTIME                 :Double = 1.0e-3;
MIN_DEADTIME                 :Double = 1.0e-10;
MAX_DEADTIME                 :Double = 1.0e10;
DEF_APPLYDEADTIME             = 1;
DEF_IDLDEADTIME               = 0;

DEF_CLEARIDL                  = 0;



implementation


//******************************************************************************
function IdlMethodToString(x :Integer): String;
begin
  if x = METHOD_HALFAMP then begin
    Result := 'HalfAmp';
  end else{if x = METHOD_SKM then}begin
    Result := 'SKM';
  end;
end;

function StringToIdlMethod(const x :String): Integer;
begin
  if x = 'HalfAmp' then begin
    Result := METHOD_HALFAMP;
  end else{if x = 'SKM' then}begin
    Result := METHOD_SKM;
  end;
end;


//******************************************************************************
end.
