{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// data structures and constants for QUB_Task_IdlBase


unit QUB_IdlBase_Global;


interface


uses


AllocTypes,

QUB_QVO_QVD_Global,
QUB_Global,
QUB_Thread,
QUB_QFS_Channels,
QUB_QFS,
QUB_Node;


type


TQUB_IdlBase_Chunk               = record

  Chunk                          :TQUB_Thread_Chunk;

  Sampling                       :Double;
  Scaling                        :Double;
  Offset                         :Double;

  Data                           :TPAInteger;

  SegCount                       :Integer;
  SegFirsts                      :TPAInteger;
  SegLasts                       :TPAInteger;
  SegClasses                     :TPAInteger;

  ClassCount                     :Integer;
  StateCount                     :Integer;
  Amps                           :TPASingle;
  Stds                           :TPASingle;
  AMatrix                        :TPMDouble;

  NodeCount                      :Integer;
  Nodes                          :TPAInteger;
  NodesDBL                       :TPAInteger;

  LL                             :Double;
  Iterations                     :Integer;

  StatResult                     :INode;

end;
TQUB_IdlBase_PChunk              = ^TQUB_IdlBase_Chunk;


{$Z4}
TQUB_IdlBase_Reestimation        = (IDLBASE_REEST_NONE, IDLBASE_REEST_SKM, IDLBASE_REEST_BW);
{$Z1}

{$Z4}
TQUB_IdlBase_MLSS                = (IDLBASE_MLSS_VITERBI, IDLBASE_MLSS_GAMMA, IDLBASE_MLSS_FORWARDVITERBI);
{$Z1}

{$Z4}
TQUB_IdlBase_Output              = (IDLBASE_OUTPUT_CLASS, IDLBASE_OUTPUT_GROUP);
{$Z1}

{$Z4}
TQUB_IdlBase_JumpAmpReestMethod  = (JUMPAMP_CONSECUTIVE, JUMPAMP_COMBINATIONS, JUMPAMP_CONSTRAINED);
{$Z1}

TQUB_IdlBase_Params              = record

  Reestimation                   :TQUB_IdlBase_Reestimation;
  MLSS                           :TQUB_IdlBase_MLSS;
  Output                         :TQUB_IdlBase_Output;

  A2Q                            :Boolean;
  A2QEveryIter                   :Boolean;
  A2QEval                        :Boolean;

  LLConv                         :Double;
  MaxIter                        :Integer;

  FixKinetics                    :Boolean;
  FixAmp                         :Boolean;
  FixStd                         :Boolean;

  DropFirstLastIfClosed          :Boolean;
  DropFirstLastIfOpen            :Boolean;

  ApplyDeadTime                  :Boolean;
  DeadTime                       :Double;

  BaselineTrack                  :Boolean;
  BaselineStd                    :Double;
  BaselineCorrect                :Boolean;
  BaselineDecFactor              :Integer;
  BaselineMinY                   :Integer;
  BaselineBackKalman             :Boolean;
  BaselineReestPar               :Boolean;
  BaselineFBKCount               :Integer;
  BaselineFKCount                :Integer;

  Staircase                      :Boolean;
  StaircaseMaxJump               :Integer;
  StaircaseRightOnly             :Boolean;
  StaircaseGlobalReest           :Boolean;
  StaircaseReestBW               :Boolean;
  StaircaseLLAddJump             :Boolean;
  StaircaseBaseJumpVar           :Double;
  StaircaseUpdateRefState        :Boolean;
  StaircaseUpdateRefStateEvery   :Integer;
                    
  StaircaseKineticsReest         :Boolean;
  StaircaseKineticsUseInit       :Boolean;
  StaircaseKinetics              :Array[0..1] of Double;

  StaircaseJumpAmpReestimate     :Boolean;
  StaircaseJumpAmpReestMethod    :TQUB_IdlBase_JumpAmpReestMethod;
  StaircaseJumpAmpUseInit        :Boolean;
  StaircaseJumpAmpStepModel      :Integer;
  StaircaseJumpAmps              :Array[0..1] of Double;
  StaircaseJumpAmpWeight         :Double;
  StaircaseJumpAmpWeighed        :Boolean;
  StaircaseJumpAmpPlsMin         :Integer;

  StaircaseJumpStdReest          :Boolean;
  StaircaseJumpStdUseInit        :Boolean;
  StaircaseJumpStd               :Double;
  StaircaseJumpStdMinGamma       :Double;
  StaircaseJumpStdWeighed        :Boolean;

  StaircaseMeasStdReest          :Boolean;
  StaircaseMeasStdUseInit        :Boolean;
  StaircaseMeasStd               :Double;

  StaircaseBaseAmpUseInit        :Boolean;
  StaircaseBaseAmp               :Double;

  StaircaseMergeDwells           :Boolean;
  StaircaseMergeDwellsPercent    :Double;
  StaircaseSplitDwells           :Boolean;
  StaircaseSplitDwellsPercent    :Double;

  ReplaceIDL                     :Boolean;
  QuietOutput                    :Boolean;
  QuietOutputEveryIter           :Boolean

end;
TQUB_IdlBase_PParams             = ^TQUB_IdlBase_Params;


function ReestimationToString(x :TQUB_IdlBase_Reestimation): String;
function StringToReestimation(const x :String): TQUB_IdlBase_Reestimation;

function MLSSToString(x :TQUB_IdlBase_MLSS): String;
function StringToMLSS(const x :String): TQUB_IdlBase_MLSS;

function OutputToString(x :TQUB_IdlBase_Output): String;
function StringToOutput(const x :String): TQUB_IdlBase_Output;


const


DEF_REESTIMATION                 = IDLBASE_REEST_BW;

DEF_A2Q                          = 0;
DEF_A2QEVERYITER                 = 0;
DEF_A2QEVAL                      = 0;

DEF_MLSS                         = IDLBASE_MLSS_GAMMA;

DEF_OUTPUT                       = IDLBASE_OUTPUT_CLASS;

MIN_LLCONV                       :Double = 1e-10;
MAX_LLCONV                       :Double = 1e10;
DEF_LLCONV                       :Double = 1e-3;

MIN_MAXITER                      = 1;
MAX_MAXITER                      = 1000;
DEF_MAXITER                      = 100;

DEF_FIXKINETICS                  = 0;
DEF_FIXAMP                       = 0;
DEF_FIXSTD                       = 0;

DEF_DROPFIRSTLASTIFCLOSED        = 0;
DEF_DROPFIRSTLASTIFOPEN          = 0;

DEF_APPLYDEADTIME                = 0;
DEF_DEADTIME                     :Double = 0e0;
MIN_DEADTIME                     :Double = 0e0;
MAX_DEADTIME                     :Double = 1e6;

//Baseline
DEF_BASELINETRACK                = 0;
MIN_BASELINESTD                  :Double = 1e-10;
MAX_BASELINESTD                  :Double = 1e+10;
DEF_BASELINESTD                  :Double = 1e-4;
DEF_BASELINECORRECT              = 0;
MIN_BASELINEDECFACTOR            = 1;
MAX_BASELINEDECFACTOR            = 100;
DEF_BASELINEDECFACTOR            = 1;
MIN_BASELINEMINY                 = 1;
MAX_BASELINEMINY                 = High(SmallInt);
DEF_BASELINEMINY                 = 1;
DEF_BASELINEBACKKALMAN           = 0;
MIN_BASELINEFKCOUNT              = 0;
MAX_BASELINEFKCOUNT              = 50;
DEF_BASELINEFKCOUNT              = 10;
MIN_BASELINEFBKCOUNT             = 0;
MAX_BASELINEFBKCOUNT             = 50;
DEF_BASELINEFBKCOUNT             = 0;
DEF_BASELINEREESTPAR             = 0;


//Staircase
DEF_STAIRCASE                    = 0;
MIN_STAIRCASE_MAXJUMP            = 6;
MAX_STAIRCASE_MAXJUMP            = 31;
DEF_STAIRCASE_MAXJUMP            = 20;
DEF_STAIRCASE_RIGHTONLY          = 1;
DEF_STAIRCASE_REESTBW            = 1;
DEF_STAIRCASE_GLOBALREEST        = 0;
MIN_STAIRCASE_BASEJUMPVAR        :Double = 0.0;
MAX_STAIRCASE_BASEJUMPVAR        :Double = 1e100;
DEF_STAIRCASE_BASEJUMPVAR        :Double = 0.2;
DEF_STAIRCASE_UPDATEREFSTATE     = 0;
DEF_STAIRCASE_UPDATEREFSTATEEVERY= 5;
DEF_STAIRCASE_LLADDJUMP          = 0;

DEF_STAIRCASE_KINETICSREEST      = 1;
DEF_STAIRCASE_KINETICSUSEINIT    = 1;
MIN_STAIRCASE_KINETICS           :Double = 0.0;
MAX_STAIRCASE_KINETICS           :Double = 1e+100;
DEF_STAIRCASE_KINETICS           :Double = 0.25;

STAIRCASE_STEPMODEL_UNIFORM      = 0;
STAIRCASE_STEPMODEL_ALTERNATE    = 1;

DEF_STAIRCASE_JUMPAMPREESTIMATE  = 1;
DEF_STAIRCASE_JUMPAMPREESTMETHOD = JUMPAMP_CONSECUTIVE;
DEF_STAIRCASE_JUMPAMPSTEPMODEL   = STAIRCASE_STEPMODEL_UNIFORM;
DEF_STAIRCASE_JUMPAMPUSEINIT     = 1;
MIN_STAIRCASE_JUMPAMP            :Double = 1e-10;
MAX_STAIRCASE_JUMPAMP            :Double = 1e+10;
DEF_STAIRCASE_JUMPAMP            :Double = 1.0;
DEF_STAIRCASE_JUMPAMPWEIGHT      :Double = 1.0;
MIN_STAIRCASE_JUMPAMPWEIGHT      :Double = 1e-100;
MAX_STAIRCASE_JUMPAMPWEIGHT      :Double = 1e+100;
DEF_STAIRCASE_JUMPAMPWEIGHED     = 0;
DEF_STAIRCASE_JUMPAMPPLSMIN      = 0;
MIN_STAIRCASE_JUMPAMPPLSMIN      = 0;
MAX_STAIRCASE_JUMPAMPPLSMIN      = 100;

DEF_STAIRCASE_JUMPSTDREEST       = 1;
DEF_STAIRCASE_JUMPSTDUSEINIT     = 1;
DEF_STAIRCASE_JUMPSTD            :Double = 0.1;
MIN_STAIRCASE_JUMPSTD            :Double = 1e-10;
MAX_STAIRCASE_JUMPSTD            :Double = 1e+10;
DEF_STAIRCASE_JUMPSTDMINGAMMA    :Double = 0.0;
MIN_STAIRCASE_JUMPSTDMINGAMMA    :Double = 0.0;
MAX_STAIRCASE_JUMPSTDMINGAMMA    :Double = 1e+10;
DEF_STAIRCASE_JUMPSTDWEIGHED     = 0;

DEF_STAIRCASE_MEASSTDREEST       = 1;
DEF_STAIRCASE_MEASSTDUSEINIT     = 1;
DEF_STAIRCASE_MEASSTD            :Double = 0.1;
MIN_STAIRCASE_MEASSTD            :Double = 1e-10;
MAX_STAIRCASE_MEASSTD            :Double = 1e+10;

DEF_STAIRCASE_BASEAMPUSEINIT     = 1;
DEF_STAIRCASE_BASEAMP            :Double = 0.0;
MIN_STAIRCASE_BASEAMP            :Double = 1e-10;
MAX_STAIRCASE_BASEAMP            :Double = 1e+10;

DEF_STAIRCASE_MERGEDWELLS        = 0;
DEF_STAIRCASE_MERGEDWELLSPERCENT :Double = 0.5;
DEF_STAIRCASE_SPLITDWELLS        = 0;
DEF_STAIRCASE_SPLITDWELLSPERCENT :Double = 0.5;


DEF_REPLACEIDL                   = 1;
DEF_QUIETOUTPUT                  = 0;
DEF_QUIETOUTPUTEVERYITER         = 0;


implementation


function ReestimationToString(x :TQUB_IdlBase_Reestimation): String;
begin
  if x = IDLBASE_REEST_NONE then begin
    Result                       := 'None';
  end else if x = IDLBASE_REEST_SKM then begin
    Result                       := 'Seg K-Means';
  end else{if x = IDLBASE_REEST_BW then}begin
    Result                       := 'Baum-Welch';
  end;
end;

function StringToReestimation(const x :String): TQUB_IdlBase_Reestimation;
begin
  if x = 'None' then begin
    Result               := IDLBASE_REEST_NONE;
  end else if x = 'Seg K-Means' then begin
    Result               := IDLBASE_REEST_SKM;
  end else{if x = 'Baum-Welch' then}begin
    Result               := IDLBASE_REEST_BW;
  end;
end;

function MLSSToString(x :TQUB_IdlBase_MLSS): String;
begin
  if x = IDLBASE_MLSS_VITERBI then begin
    Result                       := 'None';
  end else if x = IDLBASE_MLSS_GAMMA then begin
    Result                       := 'Seg K-Means';
  end else{if MLSS = IDLBASE_MLSS_FORWARDVITERBI then}begin
    Result                       := 'Baum-Welch';
  end;
end;

function StringToMLSS(const x :String): TQUB_IdlBase_MLSS;
begin
  if x = 'None' then begin
    Result                       := IDLBASE_MLSS_VITERBI;
  end else if x = 'Seg K-Means' then begin
    Result                       := IDLBASE_MLSS_GAMMA;
  end else{if x = 'Baum-Welch' then}begin
    Result                       := IDLBASE_MLSS_FORWARDVITERBI;
  end;
end;

function OutputToString(x :TQUB_IdlBase_Output): String;
begin
  if x = IDLBASE_OUTPUT_CLASS then begin
    Result                       := 'Class';
  end else if x = IDLBASE_OUTPUT_GROUP then begin
    Result                       := 'Group';
  end;
end;

function StringToOutput(const x :String): TQUB_IdlBase_Output;
begin
  if x = 'Class' then begin
    Result                       := IDLBASE_OUTPUT_CLASS;
  end else begin // if x = 'Group' then begin
    Result                       := IDLBASE_OUTPUT_GROUP;
  end;
end;


//******************************************************************************
end.
