{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// used by QUB_Task_MIP and _Mac to "idealize" experimental conditions.
// If one or more of the variables in the model (e.g. Ligand) are assigned to an A/D channel,
// this breaks up segments into pieces in which all conditions are constant.

// The part that's in use is confusing and possibly slightly broken.  Lorin's seen it lose the last point of a segment.
// It actually combines the idealized data with the idealized conditions
/// to form an idealization of (cls, ec, npoint) in which nothing at all changes during one event.

/// There's a replacement in the works but it'll probably change a lot before it's worth describing.

unit QUB_IdlCond;


interface


uses


SysUtils, Classes, Math,
Alloc, AllocTypes,

QUB_QVO_QVD,
QUB_QFS_QDF,
QUB_QFS_QDF_Global,
QUB_QFS_QIF_Segments,
QUB_ExpCond_Parameters,
QUB_Global,
QUB_Node;


type


TQUB_IData_Stream = class
protected


  prHasMore          :Boolean;


public


  function Next: Integer; virtual; abstract;


  property HasMore   :Boolean             read prHasMore;


end;


TQUB_QFSChan_Stream = class(TQUB_IData_Stream)
protected


  prData             :TQUB_QFS_QDF;
  prChannel          :Integer;
  prBounds           :TQUB_FirstLast;
  prNext             :Integer;

  prLoaded           :TPAInteger;
  prLoadBounds       :TQUB_FirstLast;
  prLoadNextIx       :Integer;        // index of next in loaded
  m_iLatency         :Integer; // Number of points to skip at start and duplicate at end.


  procedure prLoadNext;


public


  constructor Create(Data :TQUB_QFS_QDF; Channel :Integer; Bounds :TQUB_FirstLast; iLatency : Integer = 0);
  destructor Destroy; override;


  function Next: Integer; override;


end;


TQUB_QFSChan_ConcatChunkStream = class(TQUB_IData_Stream)
protected


  prData             :TQUB_QFS_QDF;
  prChannel          :Integer;
  prBounds           :TList;
  prCurBounds        :Integer;
  prChunkStream      :TQUB_QFSChan_Stream;
  m_iLatency         :Integer;


  procedure prNextChunk;


public


  constructor Create(Data :TQUB_QFS_QDF; Channel :Integer; Bounds :TList; iLatency : Integer = 0);
  destructor Destroy; override;


  function Next: Integer; override;


end;


TQUB_EvtGen = class
protected
  

  prHasMore         :Boolean;


public
  

  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; virtual; abstract;      // Evt is not valid if false


  property HasMore  :Boolean           read prHasMore;


end;


TQUB_DeltaIdlGen = class(TQUB_EvtGen)
// Cls -- avg value of points in dwell
// N -- number of points within Delta of avg.
protected


  prData           :TQUB_IData_Stream;
  prDelta          :Double;

  prAvg            :Double;
  prN              :Integer;
  //prSum : Double;
  //function Average() : Double;


public
  

  constructor Create(Data :TQUB_IData_Stream; Delta :Integer);
  destructor Destroy; override;

  
  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; override;


end;


TQUB_GlobalDeltaIdlGen = class(TQUB_EvtGen)
// unify events with cls within delta into one class with cls = avg
protected


  prEvents         :TList;
  prAt             :Integer;


  procedure prGenIdl(DeltaIdl :TQUB_DeltaIdlGen; Delta :Integer);


public


  constructor Create(DeltaIdl :TQUB_DeltaIdlGen; Delta :Integer);
  destructor Destroy; override;


  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; override;


end;


TQUB_ConstIdlGen = class(TQUB_EvtGen)
// serve up one event, then done
protected


  prCls            :Integer;
  prN              :Integer;


public


  constructor Create(Cls, N :Integer);
  destructor Destroy; override;


  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; override;


end;


{
TQUB_DCIdlGen = class(TQUB_EvtGen)
// serves up data idealization, from Dur,Cls form
protected
  prDwells         :TQUB_QFS_QIF_PADwellClass;
  prCount          :Integer;
  prSampling       :Double;

  prAt             :Integer;

public
  constructor Create(Dwells :TQUB_QFS_QIF_PADwellClass; Count :Integer; Sampling :Double);
  destructor Destroy; override;

  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; override;
end;
}

TQUB_SegNodeIdlGen = class(TQUB_EvtGen)
// serves up data idealization, from Segment{ DwellCount, Firsts, Lasts, Classes } form for MIP
protected


  prSeg            :INode;
  prFF, prLL, prCC :TPAInteger;
  prCount          :Integer;

  prAt             :Integer;


public


  constructor Create(Seg :INode);
  destructor Destroy; override;


  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; override;


end;


TQUB_NCIdlGen = class(TQUB_EvtGen)
// serves up data idealization, from N,Cls form for MIP
protected


  prDwells         :TQUB_QFS_QIF_PANClass;
  prCount          :Integer;

  prAt             :Integer;


public


  constructor Create(Dwells :TQUB_QFS_QIF_PANClass; Count :Integer);
  destructor Destroy; override;


  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; override;


end;


// For MIP
TQUB_MPlexIdlGen = class(TQUB_EvtGen)
protected


  prIdlIn           :TList;    // of IdlGen
  prIdlCount        :Integer;  // prIdlIn.Count
  prClasses         :TList;    // of TPAInteger -- prClasses[out.n] has the contributing class of each input
  prCur             :array of TQUB_QFS_QIF_NClass;


  function prGetClass(Ix :Integer): TPAInteger;
  function prGetClassCount: Integer;

  function prClassOfCur: Integer;  // create, append one if it's not in the list


public


  constructor Create;
  destructor Destroy; override;


  procedure Add(Idl :TQUB_EvtGen); // add all before generating events or face the wrath

  function Next(var Evt :TQUB_QFS_QIF_NClass): Boolean; override;


  property Classes[Ix :Integer] :TPAInteger        read prGetClass;
  property ClassCount           :Integer           read prGetClassCount;
  property ComponentCount       :Integer           read prIdlCount;


end;


TQUB_ChunkIdlGen = class
protected


  prIdl              :TQUB_EvtGen;
  prOverflow         :TQUB_QFS_QIF_NClass;


public


  constructor Create(Idl :TQUB_EvtGen);
  destructor Destroy; override;


  function NextN(N :Integer; var Dwells :TQUB_QFS_QIF_PANClass; var DwellCount :Integer): Integer; // returns actual N (e.g. if the stream ends short)


end;


TQUB_ExpCondIdlGen = class
protected


  prParams              :TQUB_ExpCond_Parameters;
  prBounds              :TList;


public


  ClassCount            :Integer;
  Params                :Array of TQUB_ExpCond_Parameters;

  ChunkCount            :Integer;
  Idl                   :Array of TQUB_QFS_QIF_PANClass;
  IdlCount              :Array of Integer;

  FreeOutputs           :Boolean;   // def true; whether destructor frees elements of public arrays


  constructor Create(Params :TQUB_ExpCond_Parameters; Data :TQUB_QVO_QVD; Bounds :TList; Delta :Integer; iLatency : Integer = 0);
  destructor Destroy; override;


  property Bounds       :TList  read prBounds;

end;


function MakeExpCondClass(Params :TQUB_ExpCond_Parameters; Vals :TPAInteger; Data :TQUB_QVO_QVD): TQUB_ExpCond_Parameters;

// Chris tries again

{type
// this is exported from qubopt.dll
TQUB_UniqueArrays_Integer = class
public
  procedure Free; virtual; cdecl; abstract;
  function Insert(Arr :TPAInteger): Integer; virtual; cdecl; abstract;
  // copies it, returns 0-based index in order of first seen
  function Count: Integer; virtual; cdecl; abstract;
  function Get(i :Integer): TPAInteger; virtual; cdecl; abstract;
end;

function TQUB_UniqueArrays_Integer_Create(Len :Integer): TQUB_UniqueArrays_Integer; stdcall; external 'qubopt.dll' name '_QUB_UniqueArrays_Integer_Create@4';

// this is defined below   new one - tested from python but not used yet
function IdealizeExpCond(DataSet :INode; DTol :Double): INode;
// Inserts Segment breaks when an ExpCond changes by more than DTol.
// An ExpCond is changing if DataSet['ExpCond'][x]['ChannelIndex'] >= 0
//                        and <> DataSet['ActiveChannel']
//                        and is the first one assigned to that channel.
// DataSet as in GetChosenData().
// Result is likewise, except:
//    Result['ExpConds'] has one or more children ExpCond
//    each Segment has child ExpCondIndex (index of relevant ExpCond in ExpConds)
//    no ExpCond elsewhere in Result

}

//procedure InitIdlCondPy;


// -----------------------------------------------------------------------------
implementation


uses


QUB_GUIInterface,
QUB_ListBox,
QUB_Viewport_Data,
QUB_Node_Python;


{
function IdealizeExpCondPy(DataSet :INode): INode;
var
  Data   :TQUB_QVO_QVD;
begin
  Result := IdealizeExpCond(DataSet, DataSet['DTol'].Data[0]);
  Data   := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DataSet['FileName'].DataAsString]);
  Data.DataObject.Idealization[ Data.ActiveChannel ].GetDwellsIntoSegs(Result, 1);
end;

procedure InitIdlCondPy;
begin
  PyNode_PublishNodeFunc( TNodeFunc.Create( IdealizeExpCondPy ), 'IdlCond' );
end;

function IdealizeExpCond(DataSet :INode; DTol :Double): INode;
// Inserts Segment breaks when an ExpCond changes by more than DTol.
// An ExpCond is changing if DataSet['ExpCond'][x]['ChannelIndex'] >= 0
//                        and <> DataSet['ActiveChannel']
//                        and is the first one assigned to that channel.
// DataSet as in GetChosenData().
// Result is likewise, except:
//    Result['ExpConds'] has one or more children ExpCond
//    each Segment has child ExpCondIndex (index of relevant ExpCond in ExpConds)
//    no ExpCond elsewhere in Result
var
  ResultDS         :INode;
  ExpCondChannels  :Array of Integer;  // ExpCond[i]['ChannelIndex'] (changing only)
  ITol, HalfTol    :Array of Integer;  // Round(DTol * Scaling[ ExpCondChannels[i] ])
  ExpCondNodes     :Array of INode;    // ExpCond[i]
  ExpCondTemplate  :INode;
  ExpCondClasses   :TQUB_UniqueArrays_Integer;
  EC, ECChan       :INode;
  Sampling         :Double;
  ChannelCount     :Integer;
  ActiveChannel    :Integer;
  Scaling          :Array of Double;   // all channels
  Segment          :INode;
  c, i             :Integer;
  x                :Double;
  Data             :TQUB_QVO_QVD;
  Buf              :TQUB_QVO_QVD_DataBuffer;

  // written by sub-procs
  ExpCondVals      :Array of Integer;
  OutSeg           :INode;

  procedure AppendSeg(First, Last, Cls :Integer; Start :Double);
  begin
    OutSeg         := ResultDS.InsertChild(OutSeg, 'Segment');
    OutSeg.SetupNumData(NODE_DATATYPE_LONGINT, 2, 1);
    OutSeg.Data[0] := First;
    OutSeg.Data[1] := Last;
    OutSeg['ExpCondIndex'].SetData( Cls );
    OutSeg['start'].SetData( Start );
  end;

  procedure IdealizeSeg(Segment :INode);
  var
    FreshSeg       :Boolean;
    First, Last    :Integer;
    StartTime      :Double;
    c, i, v, p, N  :Integer;
    BufCount       :Integer;
  begin
    FreshSeg       := True;
    First          := Segment.Data[0];
    Last           := Segment.Data[1];
    StartTime      := Segment['start'].Data[0];
    N              := Length(ExpCondChannels);
    BufCount       := Buf.BufferCount;
    i              := BufCount;

    for p := First to Last do begin
      if i = BufCount then begin
        Buf.LoadFrom(p);
        i          := 0;
      end;

      if FreshSeg then begin
        for c := 0 to N - 1 do begin
          ExpCondVals[c] := Buf.Buffer[ExpCondChannels[c]][i];
          if ExpCondVals[c] >= 0 then begin
            ExpCondVals[c] := (ExpCondVals[c] div ITol[c])*ITol[c] + HalfTol[c];
          end else begin
            ExpCondVals[c] := (ExpCondVals[c] div ITol[c])*ITol[c] - HalfTol[c];
          end;
        end;
        FreshSeg   := False;
      end else begin
        for c := 0 to N - 1 do begin
          v        := Buf.Buffer[ExpCondChannels[c]][i];
          if v >= 0 then begin
            v      := (v div ITol[c])*ITol[c] + HalfTol[c];
          end else begin
            v      := (v div ITol[c])*ITol[c] - HalfTol[c];
          end;

          if (not FreshSeg) and (v <> ExpCondVals[c]) then begin
            FreshSeg := True;
            AppendSeg(First, p-1, ExpCondClasses.Insert(@(ExpCondVals[0])), StartTime);
            StartTime := StartTime + Sampling * (p - First);
            First   := p;
          end;
          if FreshSeg then ExpCondVals[c] := v;
        end;
      end;

      Inc(i);
    end;

    if not FreshSeg then begin
      AppendSeg(First, Last, ExpCondClasses.Insert(@(ExpCondVals[0])), StartTime);
    end;
  end;

  function ExpCondClassesToTree: INode;
  var
    EC            :INode;
    i, j, N       :Integer;
    Vals          :TPAInteger;
  begin
    Result        := TNode.Create('ExpConds');
    EC            := QUB_Node_NULL;
    N             := Length(ExpCondChannels);

    for i := 0 to ExpCondClasses.Count() - 1 do begin
      Vals        := ExpCondClasses.Get(i);
      for j := 0 to N - 1 do begin
        ExpCondNodes[j].Data[0] := Vals[j] / Scaling[ ExpCondChannels[j] ];
      end;
      EC          := Result.InsertClone(EC, ExpCondTemplate, True);
    end;
  end;

label SKIP_EXPCOND;
begin
  Result           := TNode.Create('DataSet');
  ResultDS         := Result;
  Result.AppendClone( DataSet['FileName'], True );
  Result.AppendClone( DataSet['sampling'], True );
  Result.AppendClone( DataSet['Scaling'], True );
  Result.AppendClone( DataSet['ADChannelCount'], True );
  Result.AppendClone( DataSet['ActiveChannel'], True );
  Result.AppendClone( DataSet['Channels'], True );

  Sampling         := DataSet['sampling'].Data[0];
  ChannelCount     := DataSet['ADChannelCount'].Data[0];
  ActiveChannel    := DataSet['ActiveChannel'].Data[0];

  SetLength(Scaling, ChannelCount);
  x                := DataSet['Scaling'].Data[0];
  ECChan           := Result['Channels'].Find('Channel');
  for c := 0 to ChannelCount - 1 do begin
    Scaling[c]     := x * ECChan['Scaling'].Data[0];
    ECChan         := ECChan.SiblingSameName;
  end;

  // find which ExpConds go with which channels
  ExpCondTemplate  := DataSet['ExpCond'].Clone(True);
  EC               := ExpCondTemplate.Child;
  while not EC.IsNull do begin
    ECChan         := EC.Find('ChannelIndex');
    if ECChan.DataCount > 0 then begin
      c            := ECChan.Data[0];
      if (c >= 0) and (c <> ActiveChannel) then begin
        for i := 0 to Length(ExpCondChannels) - 1 do begin
          if ExpCondChannels[i] = c then begin
            goto SKIP_EXPCOND;
          end;
        end;

        // ExpCond is good
        i          := Length(ExpCondChannels) + 1;
        SetLength(ExpCondChannels, i);
        SetLength(ExpCondNodes, i);
        SetLength(ExpCondVals, i);
        SetLength(ITol, i);
        SetLength(HalfTol, i);
        Dec(i);
        ExpCondChannels[i] := c;
        ExpCondNodes[i]    := EC;
        ITol[i]            := Round(Scaling[c] * DTol);
        HalfTol[i]         := Round(Scaling[c] * (DTol / 2));
      end;
    end;
  SKIP_EXPCOND:
    EC             := EC.Sibling;
  end;

  Buf              := nil;
  ExpCondClasses   := nil;
  try
    Data           := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DataSet['FileName'].DataAsString] );
    if Data = nil then Exit;

    ExpCondClasses := TQUB_UniqueArrays_Integer_Create(Length(ExpCondChannels));
    Buf            := TQUB_QVO_QVD_DataBuffer.Create(Data, ONE_MEGABYTE div ChannelCount, False);

    Segment          := DataSet.Find('Segment');
    OutSeg           := QUB_Node_NULL;
    while not Segment.IsNull do begin
      IdealizeSeg(Segment);
      Segment        := Segment.SiblingSameName;
    end;
    Result.InsertChild(QUB_Node_NULL, ExpCondClassesToTree());
  finally
    ExpCondClasses.Free;
    Buf.Free;
  end;
end;
}

// -----------------------------------------------------------------------------
constructor TQUB_QFSChan_Stream.Create(Data :TQUB_QFS_QDF; Channel :Integer; Bounds :TQUB_FirstLast; iLatency : Integer = 0);
begin
  prData             := Data;
  prChannel          := Channel;
  prBounds           := Bounds;
  prNext             := Bounds.First;
  prHasMore          := True;

  m_iLatency := iLatency;

  try
    prLoaded           := IAlloc1(QUB_IDATA_WINSIZE);
    prLoadBounds.First := -1;
    prLoadBounds.Last  := -1;

    prLoadNext;
  except
    prLoaded           := nil;
    prNext             := Bounds.Last + 1;
    prHasMore          := False;
  end;

end;

destructor TQUB_QFSChan_Stream.Destroy;
begin
  try
    FreeMem(prLoaded);
  except
  end;
end;

function TQUB_QFSChan_Stream.Next: Integer;
begin
  Result       := prLoaded[prLoadNextIx];

  prNext       := prNext + 1;
  prLoadNextIx := prLoadNextIx + 1;
  if prNext > prLoadBounds.Last then begin
    prLoadNext;
  end;
end;

procedure TQUB_QFSChan_Stream.prLoadNext;
var
  i, iLat :Integer;
begin
  if prNext > prBounds.Last then begin
    prHasMore          := False;
    Exit;
  end;

  if prLoadBounds.First = -1 then begin
    prLoadBounds.First := prBounds.First;
  end else begin
    prLoadBounds.First := prLoadBounds.Last + 1;
    prHasMore          := False;
  end;

  prLoadBounds.Last    := Min(prBounds.Last, prLoadBounds.First + QUB_IDATA_WINSIZE - 1);
  prLoadNextIx         := 0;

  try
    //prData.ReadData(prLoaded, False{Raw16 not 32}, prChannel, prLoadBounds.First, prLoadBounds.Last);

    //---- 7-2004 JTB : Shift the idealized data by m_iLatency pts.
    iLat := max(0,m_iLatency-(prBounds.First-prLoadBounds.First));
    prLoaded[0]:=0; // safeguard - if latency > load size (1M-unlikely), fill with 0's.
    prData.ReadData(prLoaded, False, prChannel, max(prLoadBounds.First-m_iLatency,prBounds.First), prLoadBounds.Last-m_iLatency);
    //Careful! The last iLat points are undefined. LM

    if (iLat > 0) then begin
      //Shift the stimulus by iLat. LM
      //for i := prLoadBounds.Last-prLoadBounds.First-1 downto iLat do begin
      // -1 on Last-First removed 3/2005 - was causing MAC errors. {JB}
      for i := prLoadBounds.Last-prLoadBounds.First downto iLat do begin
        prLoaded[i] := prLoaded[i-iLat];
      end;
      for i := iLat-1 downto 1 do begin
        prLoaded[i] := prLoaded[0];
      end;
    end;

  except
    prNext             := prBounds.Last + 1;
    prHasMore          := False;
  end;
end;


// -----------------------------------------------------------------------------
constructor TQUB_QFSChan_ConcatChunkStream.Create(Data :TQUB_QFS_QDF; Channel :Integer; Bounds :TList; iLatency : Integer=0);
begin
  prData             := Data;
  prChannel          := Channel;
  prBounds           := Bounds;
  prCurBounds        := -1;
  prChunkStream      := nil;
  m_iLatency         := iLatency;
  prNextChunk;
end;

destructor TQUB_QFSChan_ConcatChunkStream.Destroy;
begin
  if prChunkStream <> nil then prChunkStream.Free;
end;

procedure TQUB_QFSChan_ConcatChunkStream.prNextChunk;
begin
  if prChunkStream <> nil then prChunkStream.Free;
  prChunkStream      := nil;

  prCurBounds        := prCurBounds + 1;
  prHasMore          := prCurBounds < prBounds.Count;
  if prHasMore then begin
    prChunkStream    := TQUB_QFSChan_Stream.Create(prData, prChannel, TQUB_PFirstLast(prBounds[prCurBounds])^, m_iLatency);
    //prHasMore        := prChunkStream.HasMore;
  end;
end;

function TQUB_QFSChan_ConcatChunkStream.Next: Integer;
begin
  if prChunkStream = nil then begin
    Result := 0;
  end else begin
    if prChunkStream.HasMore then begin
      Result := prChunkStream.Next;

      if not prChunkStream.HasMore then prNextChunk;
    end else begin      // to keep from stalling on an empty/inaccessible chunk
      prNextChunk;
      Result := Next;
    end;
  end;
end;


// -----------------------------------------------------------------------------
constructor TQUB_DeltaIdlGen.Create(Data :TQUB_IData_Stream; Delta :Integer);
begin
  prData           := Data;
  prDelta          := Delta;

  prHasMore        := prData.HasMore;
  prAvg            := Data.Next;
  //prSum            := Data.Next;
  prN              := 1;
end;

destructor TQUB_DeltaIdlGen.Destroy;
begin
end;

function TQUB_DeltaIdlGen.Next(var Evt :TQUB_QFS_QIF_NClass): Boolean;
var
  NextVal          :Double;
begin
  if not HasMore then begin
    Result         := False;
    Exit;
  end;

  Result           := True;

  NextVal          := prData.Next;
  if not prData.HasMore then begin
    prHasMore      := False;
    Evt.Cls        := Round(prAvg);
    //Evt.Cls        := Round(Average());
    Evt.N          := prN;
    Exit;
  end;

  while Abs(prAvg - NextVal) <= prDelta do begin
    prAvg          := (prAvg * prN + NextVal) / (prN + 1);
    //prSum := prSum + NextVal;
    prN            := prN + 1;

    NextVal        := prData.Next;
    if not prData.HasMore then begin
      prHasMore    := False;
      Break;
    end;
  end;

  Evt.Cls          := Round(prAvg);
  //Evt.Cls          := Round(Average());
  Evt.N            := prN;
  prAvg            := NextVal;
  //prSum := NextVal;
  prN              := 1;
end;

{
function TQUB_DeltaIdlGen.Average() : Double;
begin
  Result := prSum / prN;
end;
}

// -----------------------------------------------------------------------------
constructor TQUB_GlobalDeltaIdlGen.Create(DeltaIdl :TQUB_DeltaIdlGen; Delta :Integer);
begin
  prEvents         := TList.Create;
  prAt             := 0;

  prGenIdl(DeltaIdl, Delta);
  prHasMore        := prEvents.Count > 0;
end;

destructor TQUB_GlobalDeltaIdlGen.Destroy;
var
  i                :Integer;
begin
  for i := 0 to prEvents.Count - 1 do begin
    Dispose( prEvents[i] );
  end;
  prEvents.Free;
end;

procedure TQUB_GlobalDeltaIdlGen.prGenIdl(DeltaIdl :TQUB_DeltaIdlGen; Delta :Integer);
var
  Evt              :TQUB_QFS_QIF_PNClass;
  i, j             :Integer;
  Evi, Evj         :TQUB_QFS_QIF_PNClass;
  Clsum            :Double; // sum(cls * n)
  Nsum             :Integer;
  N                :Integer;
  Visited          :array of Boolean;
  IxInCls          :array of Integer;
  NinCls           :Integer;
begin
  while DeltaIdl.HasMore do begin
    New(Evt);
    DeltaIdl.Next(Evt^);
    prEvents.Add(Evt);
  end;

  N               := prEvents.Count;

  // pass 1: foreach i, j: if cls[i] within delta of cls[j] then cls[i] := cls[j] := weighted avg;
  SetLength(Visited, N);
  for i := 0 to N - 1 do begin
    Visited[i]     := False;
  end;

  SetLength(IxInCls, N);
  for i := 0 to N - 2 do begin
    if Visited[i] then Continue;

    Evi            := TQUB_QFS_QIF_PNClass(prEvents[i]);
    Clsum          := Evi.Cls * Evi.N;
    Nsum           := Evi.N;
    NinCls         := 0;

    for j := i + 1 to N - 1 do begin
      if Visited[j] then Continue;

      Evj          := TQUB_QFS_QIF_PNClass(prEvents[j]);

      if Abs(Evi.Cls - Evj.Cls) <= Delta then begin
        Clsum      := Clsum + Evj.Cls * Evj.N;
        Nsum       := Nsum + Evj.N;
        IxInCls[ NinCls ] := j;
        NinCls     := NinCls + 1;
        Visited[j] := True;
      end;
    end;

    Evi.Cls        := Round(Clsum / Nsum);

    {
    for j := 0 to NinCls - 1 do begin
      TQUB_QFS_QIF_PNClass(prEvents[j]).Cls := Evi.Cls;//Bad.
    end;
    }
    for j := 0 to N - 1 do begin
      Evj          := TQUB_QFS_QIF_PNClass(prEvents[j]);
      if Abs(Evi.Cls - Evj.Cls) <= Delta then begin
        Evj.Cls := Evi.Cls;
      end;
    end;

  end;

  // pass 2: combine adjacent events of same cls
  i                := N - 1;
  while i > 0 do begin
    Evi            := TQUB_QFS_QIF_PNClass(prEvents[i]);

    j              := i - 1;
    Evj            := TQUB_QFS_QIF_PNClass(prEvents[j]);

    while (j >= 0) and (Evj.Cls = Evi.Cls) do begin
      Evi.N        := Evi.N + Evj.N;
      Dispose(Evj);
      prEvents.Delete(j);
      i            := i - 1;
      j            := j - 1;
      if j >= 0 then Evj := TQUB_QFS_QIF_PNClass(prEvents[j]);
    end;

    i              := i - 1;
  end;
end;

function TQUB_GlobalDeltaIdlGen.Next(var Evt :TQUB_QFS_QIF_NClass): Boolean;
begin
  if not HasMore then begin
    Result         := False;
    Exit;
  end;

  Result           := True;
  Evt              := TQUB_QFS_QIF_PNClass( prEvents[prAt] )^;
  prAt             := prAt + 1;
  prHasMore        := prAt < prEvents.Count;
end;

// -----------------------------------------------------------------------------
constructor TQUB_ConstIdlGen.Create(Cls, N :Integer);
begin
  prCls            := Cls;
  prN              := N;
  prHasMore        := True;
end;

destructor TQUB_ConstIdlGen.Destroy;
begin
end;

function TQUB_ConstIdlGen.Next(var Evt :TQUB_QFS_QIF_NClass): Boolean;
begin
  Evt.Cls          := prCls;
  Evt.N            := prN;
  Result           := HasMore;
  prHasMore        := False;
end;

// -----------------------------------------------------------------------------
{
constructor TQUB_DCIdlGen.Create(Dwells :TQUB_QFS_QIF_PADwellClass; Count :Integer; Sampling :Double);
begin
  prDwells         := Dwells;
  prCount          := Count;
  prSampling       := Sampling;
  prHasMore        := Count > 0;
  prAt             := 0;
end;

destructor TQUB_DCIdlGen.Destroy;
begin
end;

function TQUB_DCIdlGen.Next(var Evt :TQUB_QFS_QIF_NClass): Boolean;
begin
  if not HasMore then begin
    Result         := False;
    Exit;
  end;

  Result           := True;
  Evt.Cls          := prDwells[prAt].Cls;
  Evt.N            := Round( prDwells[prAt].Dwell / prSampling );
  prAt             := prAt + 1;
  prHasMore        := prAt < prCount;
end;
}

// -----------------------------------------------------------------------------
constructor TQUB_SegNodeIdlGen.Create(Seg :INode);
begin
  prSeg            := Seg;
  prCount          := Seg['DwellCount'].Data[0];
  prFF             := TPAInteger( Seg['Firsts'].DataPtr );
  prLL             := TPAInteger( Seg['Lasts'].DataPtr );
  prCC             := TPAInteger( Seg['Classes'].DataPtr );
  prHasMore        := prCount > 0;
  prAt             := 0;
end;

destructor TQUB_SegNodeIdlGen.Destroy;
begin
end;

function TQUB_SegNodeIdlGen.Next(var Evt :TQUB_QFS_QIF_NClass): Boolean;
begin
  if not HasMore then begin
    Result         := False;
    Exit;
  end;

  Result           := True;
  Evt.Cls          := prCC[ prAt ];
  Evt.N            := prLL[ prAt ] - prFF[ prAt ] + 1;
  prAt             := prAt + 1;
  prHasMore        := prAt < prCount;
end;

// -----------------------------------------------------------------------------
constructor TQUB_NCIdlGen.Create(Dwells :TQUB_QFS_QIF_PANClass; Count :Integer);
begin
  prDwells         := Dwells;
  prCount          := Count;
  prHasMore        := Count > 0;
  prAt             := 0;
end;

destructor TQUB_NCIdlGen.Destroy;
begin
end;

function TQUB_NCIdlGen.Next(var Evt :TQUB_QFS_QIF_NClass): Boolean;
begin
  if not HasMore then begin
    Result         := False;
    Exit;
  end;

  Result           := True;
  Evt.Cls          := prDwells[prAt].Cls;
  Evt.N            := prDwells[prAt].N;
  prAt             := prAt + 1;
  prHasMore        := prAt < prCount;
end;

// -----------------------------------------------------------------------------
constructor TQUB_MPlexIdlGen.Create;
begin
  prIdlIn           := TList.Create;
  prIdlCount        := 0;
  prClasses         := TList.Create;
  prHasMore         := True;
end;

destructor TQUB_MPlexIdlGen.Destroy;
var
  i                 :Integer;
begin
  for i := 0 to ClassCount - 1 do begin
    FreeMem( prClasses[i] );
  end;
  prClasses.Free;
  prIdlIn.Free;
end;

procedure TQUB_MPlexIdlGen.Add(Idl :TQUB_EvtGen); // add all before generating events or face the wrath
begin
  prIdlIn.Add( Idl );
  prIdlCount := prIdlCount + 1;

  SetLength( prCur, prIdlCount );
  if not Idl.Next( prCur[ prIdlCount - 1 ] ) then begin
    prHasMore := False;
  end;
end;

function TQUB_MPlexIdlGen.prGetClass(Ix :Integer): TPAInteger;
begin
  Result := TPAInteger( prClasses[Ix] );
end;

function TQUB_MPlexIdlGen.prGetClassCount: Integer;
begin
  Result := prClasses.Count;
end;

function TQUB_MPlexIdlGen.prClassOfCur: Integer;
var
  c, i              :Integer;
  cls               :TPAInteger;
  same              :Boolean;
begin
  Result            := -1; // temp -- not found yet

  for c := 0 to ClassCount - 1 do begin
    cls             := Classes[c];
    same            := True;
    for i := 0 to ComponentCount - 1 do begin
      if cls[i] <> prCur[i].Cls then begin
        same        := False;
        Break; // check next class
      end;
    end;
    if same then begin
      Result        := c;
      Break;
    end;
  end;

  if Result = -1 then begin
    cls             := IAlloc1( ComponentCount );
    for i := 0 to ComponentCount - 1 do begin
      cls[i]        := prCur[i].Cls;
    end;
    prClasses.Add( cls );
    Result := ClassCount - 1;
  end;
end;

function TQUB_MPlexIdlGen.Next(var Evt :TQUB_QFS_QIF_NClass): Boolean;
var
  MinN              :Integer;
  i                 :Integer;
begin
  if not HasMore then begin
    Result          := False;
    Exit;
  end;

  Result            := True;

  MinN              := prCur[0].N;
  for i := 1 to ComponentCount - 1 do begin
    if prCur[i].N < MinN then MinN := prCur[i].N;
  end;

  Evt.N             := MinN;
  Evt.Cls           := prClassOfCur;

  for i := 0 to ComponentCount - 1 do begin
    prCur[i].N      := prCur[i].N - MinN;
    if prCur[i].N = 0 then begin
      if not TQUB_EvtGen(prIdlIn[i]).Next( prCur[i] ) then begin
        prHasMore   := False;
        Break;
      end;
    end;
  end;
end;

// -----------------------------------------------------------------------------
constructor TQUB_ChunkIdlGen.Create(Idl :TQUB_EvtGen);
begin
  prIdl              := Idl;
  prOverflow.N       := 0;
end;

destructor TQUB_ChunkIdlGen.Destroy;
begin
end;

function TQUB_ChunkIdlGen.NextN(N :Integer; var Dwells :TQUB_QFS_QIF_PANClass; var DwellCount :Integer): Integer; // returns actual N (e.g. if the stream ends short)
var
  Cap                :Integer;
begin
  Cap                := 256;
  Dwells             := AllocMem(Cap * SizeOf(TQUB_QFS_QIF_NClass));
  DwellCount         := 0;
  Result             := 0;

  if prOverflow.N > 0 then begin
    Dwells[0].N      := prOverflow.N;
    Dwells[0].Cls    := prOverflow.Cls;
    DwellCount       := 1;
    Result           := prOverflow.N;
    prOverflow.N     := 0;
  end;

  while prIdl.HasMore and (Result < N) do begin
    if DwellCount = Cap then begin
      Cap            := Cap * 2;
      ReallocMem(Dwells, Cap * SizeOf(TQUB_QFS_QIF_NClass));
    end;

    prIdl.Next(Dwells[DwellCount]);
    Result           := Result + Dwells[DwellCount].N;
    DwellCount       := DwellCount + 1;
  end;

  ReallocMem(Dwells, DwellCount * SizeOf(TQUB_QFS_QIF_NClass));

  if Result > N then begin
    prOverflow.N     := Result - N;
    prOverflow.Cls   := Dwells[ DwellCount - 1 ].Cls;
    Dwells[ DwellCount - 1 ].N := Dwells[ DwellCount - 1 ].N - prOverflow.N;
  end;
end;


// -----------------------------------------------------------------------------
function MakeExpCondClass(Params :TQUB_ExpCond_Parameters; Vals :TPAInteger; Data :TQUB_QVO_QVD): TQUB_ExpCond_Parameters;
var
  i                   :Integer;
  Par                 :TQUB_ExpCond_PParameter;
  ChannelScaling      :Double;
  ChannelOffset       :Double;
begin
  Result              := TQUB_ExpCond_Parameters.Create;

  for i := 0 to Params.Count - 1 do begin
    Par               := Params[i];
    ChannelScaling    := 1.0;
    ChannelOffset     := 0.0;
    if Par.Channel >= 0 then begin
      ChannelScaling := Data.DataChannelScaling[Par.Channel];
      ChannelOffset  := Data.DataChannelOffset[Par.Channel];
    end;
    Result.AddPar(Par.Name, Vals[i] / (Data.DataObject.Scaling * ChannelScaling) + ChannelOffset, Par.Index, Par.IsVar, Par.Channel);
  end;
end;

// -----------------------------------------------------------------------------
constructor TQUB_ExpCondIdlGen.Create(Params :TQUB_ExpCond_Parameters; Data :TQUB_QVO_QVD; Bounds :TList; Delta :Integer; iLatency : Integer = 0 );
var
  i                   :Integer;
  SumN                :Integer;
  FL                  :TQUB_PFirstLast;
  Mplex               :TQUB_MPlexIdlGen;
  ChunkGen            :TQUB_ChunkIdlGen;
  Channel             :Integer;
  DataStream          :TQUB_IData_Stream;
  IdlStream           :TQUB_EvtGen;
  DataStreams         :TList;
  IdlStreams          :TList;
begin
  prParams            := Params;
  prBounds            := Bounds;
  FreeOutputs         := True;

  if Params.Count > 0 then begin

    SumN                := 0;
    for i := 0 to Bounds.Count - 1 do begin
      FL                := TQUB_PFirstLast( Bounds[i] );
      SumN              := SumN + FL.Last - FL.First + 1;
    end;

    Mplex               := TQUB_MPlexIdlGen.Create;
    DataStreams         := TList.Create;
    IdlStreams          := TList.Create;

    for i := 0 to Params.Count - 1 do begin
      Channel           := Params[i].Channel;

      //Should it be processed if not in use?!
      if (Channel >= 0) and (Channel < Data.DataObject.ADChannelCount) then begin
        DataStream      := TQUB_QFSChan_ConcatChunkStream.Create(Data.DataObject, Channel, Bounds, iLatency);
        DataStreams.Add(DataStream);

        IdlStream       := TQUB_DeltaIdlGen.Create(DataStream, Delta);
        IdlStreams.Add(IdlStream);

        IdlStream       := TQUB_GlobalDeltaIdlGen.Create(TQUB_DeltaIdlGen(IdlStream), Delta);
        IdlStreams.Add(IdlStream);
        Mplex.Add(IdlStream);
      end else begin
        IdlStream       := TQUB_ConstIdlGen.Create(Round(Params[i].Value * Data.DataObject.Scaling), SumN);
        IdlStreams.Add(IdlStream);
        Mplex.Add(IdlStream);
      end;
    end;

    ChunkCount          := Bounds.Count;
    SetLength(Idl, ChunkCount);
    SetLength(IdlCount, ChunkCount);

    ChunkGen            := TQUB_ChunkIdlGen.Create(Mplex);
    for i := 0 to ChunkCount - 1 do begin
      FL                := TQUB_PFirstLast( Bounds[i] );
      ChunkGen.NextN(FL.Last - FL.First + 1, Idl[i], IdlCount[i]); // make sure IdlCount[i] = rtn val?
    end;

    // now that mplex has generated all, mplex.classes are valid
    ClassCount          := Mplex.ClassCount;
    SetLength(Self.Params, ClassCount);
    for i := 0 to ClassCount - 1 do begin
      Self.Params[i]    := MakeExpCondClass(Params, Mplex.Classes[i], Data);
    end;

    ChunkGen.Free;
    Mplex.Free;

    for i := 0 to IdlStreams.Count - 1 do begin
      TQUB_EvtGen(IdlStreams[i]).Free;
    end;
    IdlStreams.Free;

    for i := 0 to DataStreams.Count - 1 do begin
      TQUB_IData_Stream(DataStreams[i]).Free;
    end;
    DataStreams.Free;

  end else begin

    ChunkCount          := Bounds.Count;
    SetLength(Idl, ChunkCount);
    SetLength(IdlCount, ChunkCount);

    for i := 0 to ChunkCount - 1 do begin
      FL                := TQUB_PFirstLast( Bounds[i] );
      ReallocMem(Idl[i], 1 * SizeOf(TQUB_QFS_QIF_NClass));
      Idl[i][0].N       := FL.Last - FL.First + 1;
      Idl[i][0].Cls     := 0;
      IdlCount[i]       := 1;
    end;

    ClassCount          := 1;
    SetLength(Self.Params, ClassCount);
    Self.Params[0]      := TQUB_ExpCond_Parameters.Create;

  end;
end;

destructor TQUB_ExpCondIdlGen.Destroy;
var
  i                   :Integer;
begin
  if FreeOutputs then begin
    for i := 0 to ClassCount - 1 do begin
      Params[i].Free;
    end;

    for i := 0 to ChunkCount - 1 do begin
      FreeMem(Idl[i]);
    end;
  end;
end;


// -----------------------------------------------------------------------------
end.
