{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the visual list of files in all QUB_Viewport_Files (e.g. data, model)
// it is known as Viewport.ListObjects


unit QUB_ListBox;


interface


uses

Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls,
QUB_QVO;


type


TQUB_ListBox = class(TListBox)
protected


  prQVO           :TQUB_QVO;

  prSaveItems     :TStringList;
  prSaveTopIndex  :Integer;
  prSaveItemIndex :Integer;


  function prGetQVO (Index :Integer) :TQUB_QVO;
  function prGetQVOByName( nm :String ): TQUB_QVO;
  function prGetQVOByPath( path :String ): TQUB_QVO;

  procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;

  procedure CreateWnd; override;
  procedure DestroyWnd; override;


public


  destructor Destroy; override;


  function AddQVO(o :TQUB_QVO; const Name :String) :Boolean;
  function DelQVO(o :TQUB_QVO) :Boolean;

  function SelQVO(o :TQUB_QVO): Boolean;
  function SelQVOAt(i :Integer): Boolean;

  property QVO[Index :Integer] :TQUB_QVO read prGetQVO; default;
  property QVOByName[nm :String] :TQUB_QVO read prGetQVOByName;
  property QVOByPath[path :String] :TQUB_QVO read prGetQVOByPath;


end;


procedure Register;


implementation


//******************************************************************************
//This shit is only necessary to cope wity Docking !!!
//Which somehow calls Create and DestroyWnd in the wrong order for TCustomListBox
//Fuck !!! (The same had to be done for Selections !!!)
destructor TQUB_ListBox.Destroy;
begin
  try
    prSaveItems.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_ListBox.CreateWnd;
begin
  inherited;
  if prSaveItems <> nil then begin
    Items.Assign(prSaveItems);
    TopIndex  := prSaveTopIndex;
    ItemIndex := prSaveItemIndex;
  end;
end;

procedure TQUB_ListBox.DestroyWnd;
begin
  if Items.Count > 0 then begin
    prSaveItems := TStringList.Create;
    prSaveItems.Assign(Items);
    prSaveTopIndex  := TopIndex;
    prSaveItemIndex := ItemIndex;
  end;
  inherited;
end;

function TQUB_ListBox.AddQVO(o :TQUB_QVO; const Name :String) :Boolean;
begin
  Result := False;
  {
  if (Items.Count <> 0) and (ItemIndex <> -1) then begin
    QVO[ItemIndex].Deactivate;
    if SelCount <= 1 then begin
      Selected[ItemIndex] := False;
    end;
  end;
  }
  Items.AddObject(Name, o);
  //ItemIndex := Items.IndexOfObject(o);}
  SelQVOAt(Items.IndexOfObject(o));
  {Selected[ItemIndex] := True;
  if (Owner is TControl) and (TControl(Owner).Visible) and (TControl(Owner).Enabled) then QVO[ItemIndex].Activate;}
end;

function TQUB_ListBox.SelQVO(o :TQUB_QVO): Boolean;
begin
  Result := SelQVOAt(Items.IndexOfObject(o));
end;

function TQUB_ListBox.SelQVOAt(i :Integer): Boolean;
begin
  Result := (i < 0) or (i >= Items.Count);
  if Result then Exit;

  if (Items.Count <> 0) and (ItemIndex <> -1) then begin
    QVO[ItemIndex].Deactivate;
    if SelCount <= 1 then begin
      Selected[ItemIndex] := False;
    end;
  end;

  ItemIndex := i;
  Selected[ItemIndex] := True;
  {if (Owner is TControl) and (TControl(Owner).Visible) and (TControl(Owner).Enabled) then }QVO[ItemIndex].Activate;
end;

function TQUB_ListBox.DelQVO(o :TQUB_QVO) :Boolean;
var
  i :Integer;
begin
  Result := False;
  i := Items.IndexOfObject(o);
  if i = -1 then Exit;
  o.DeActivate;
  Items.Delete(i);

  if i >= Items.Count then i := Items.Count - 1;
  if i < 0 then Exit;
  ItemIndex := i;
  Selected[ItemIndex] := True;
  if (Owner is TControl) and (TControl(Owner).Visible) and (TControl(Owner).Enabled) then QVO[ItemIndex].Activate;
end;

procedure TQUB_ListBox.CNCommand(var Message: TWMCommand);
begin
  case Message.NotifyCode of
    LBN_SELCHANGE:
      begin
        if (Items.Count <> 0) and (ItemIndex <> -1) then begin
          if (prQVO <> nil) and (prQVO <> QVO[ItemIndex]) then begin
            prQVO.DeActivate;
          end;
          if (Owner is TControl) and (TControl(Owner).Visible) and (TControl(Owner).Enabled) then QVO[ItemIndex].Activate;
          prQVO := nil;
        end;
      end;
    LBN_SETFOCUS:
      begin
        if (Items.Count > 1) and (ItemIndex <> -1) then begin
          prQVO := QVO[ItemIndex];
        end else begin
          prQVO := nil;
        end;
      end;
  end;
  inherited;
end;

function TQUB_ListBox.prGetQVO(Index :Integer) :TQUB_QVO;
begin
  if (Index >= 0) and (Index < Items.Count) then Result := TQUB_QVO(Items.Objects[Index]) else Result := nil;
end;

function TQUB_ListBox.prGetQVOByName(nm :String): TQUB_QVO; // or nil
var
  i      :Integer;
begin
  Result := nil;
  for i := 0 to Items.Count - 1 do begin
    if Items[i] = nm then begin
      Result := TQUB_QVO(Items.Objects[i]);
    end;
  end;
end;

function TQUB_ListBox.prGetQVOByPath(path :String): TQUB_QVO; // or nil
var
  i      :Integer;
begin
  Result := nil;
  for i := 0 to Items.Count - 1 do begin
    if QVO[i].FileName = path then begin
      Result := TQUB_QVO(Items.Objects[i]);
    end;
  end;
end;


//******************************************************************************
procedure Register;
begin
  RegisterComponents('QUB', [TQUB_ListBox]);
end;


//******************************************************************************
end.
