{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// this file is almost unused.
// MIL doesn't use it.
// MIP (in QUB_MIP_Model or QUB_Q_Model) uses the last bit,
// but the intention to bring the full multichannel model to MIP was never realized.

unit QUB_MIL;


interface


uses


AllocTypes,
QUB_Thread,
QUB_Mil_Global,
QUB_Multicast,
QUB_Node;


type


TQUB_MIL_MODEL_DESC = record
  Model            :PChar;
  NChannel         :Integer;
  SearchLimit      :Single;
end;


TQUB_MIL_PARS      = record
  HistBinCount     :Integer;
  HistPerIter      :Integer;
  convLL           :Single;
  convGrad         :Single;
  maxIter          :Integer;
  Restarts         :Integer;
  MaxStep          :Single;
  Method           :TQUB_MIL_Method;
  Mode             :TQUB_MIL_Mode;
  SegMode          :TQUB_MIL_Seg_Mode;
  JoinSegments     :Integer;
end;

TQUB_MIL_RATES     = record
  States           :Array[1..2] of Integer;
  k0               :Array[1..2] of Single;
  k1               :Array[1..2] of Single;
  k_std0           :Array[1..2] of Single;
  k_std1           :Array[1..2] of Single;
end;
TQUB_MIL_ARATES    = Array[0..(MaxInt div (SizeOf(TQUB_MIL_RATES)*2))] of TQUB_MIL_RATES;
TQUB_MIL_PARATES   = ^TQUB_MIL_ARATES;


TQUB_MIL_RESULTS   = record
  Index            :Integer;  //-1 == ALL
  Iteration        :Integer;
  IsLastIter       :Integer;
  IsLastSeg        :Integer;
  LL               :Single;
  Rates            :TQUB_MIL_PARATES;

  nFreePar         :Integer;  // == size(grads)
  grads            :TPASingle;//
  segLL            :TPASingle;// per-seg(only for index==ALL and isLastIter==TRUE)

  ClassCount       :TPAInteger;  //takes into account NChannel -> nmulticlass
                                 //it is per SegClass

  //There are K SegClasses, each SegClass has one ore more segments (of identical params (voltage, conc etc)).
  //For each SegClass there is one histogram
  //Each Chunk should have SegClass index

  SegClassCount    :Integer;//returned by MIL
  HistSegMap       :TPAInteger; //maps the Histogram's index into Chunk's SegClass
  Histogram        :TPAPointer;{each element is a TPMSingle}
  //Histogram[SegClass][class, bin]  (segclass, row, col)
  //  row i              is class i hist
  //  row ClassCount + i is class i pdf
  //  row 2 * ClassCount is bin x

  // time constants
  ComponentCount   :TPMInteger; // [segclass, class]
  TimeConst        :TPAPointer; // each element is a TPMSingle
  AreaFract        :TPAPointer; // [SegClass][class, metastate] // beware of invalid numbers for some metastates that don't exist

  HistValid        :Integer; // false/0 if there was an exception making pdfs -- ignore pdf and time constants
end;
TQUB_MIL_PRESULTS  = ^TQUB_MIL_RESULTS;


//In here I have to allocate/deallocate the Dwells buffer
// chunk->dwells ? dealloc : alloc
TQUB_MIL_DATA_CALLBACK      = function(caller_obj :TObject; Mil_Chunk :TQUB_Thread_Mil_PChunk) :Integer; stdcall;
TQUB_MIL_IDLDATA_CALLBACK   = function(caller_obj :TObject; Mil_Chunk :TQUB_Thread_Mil_PChunk) :Integer; stdcall;
TQUB_MIL_REPORT_CALLBACK    = function(caller_obj :TObject; ReportKind :TQUB_Thread_Event_Message_Kind; message :PChar) :Integer; stdcall;
//Chris allocates everything in Results, and Results. I am responsible with making a copy, at least at the last iteraiton
TQUB_MIL_RESULTS_CALLBACK   = function(caller_obj :TObject; Results :TQUB_MIL_PRESULTS) :Integer; stdcall;


TQUB_MIL_INTERFACE          = record

  Model                     :TQUB_MIL_MODEL_DESC;
  Pars                      :TQUB_MIL_PARS;
  RemotePars                :TQUB_Remote_Params;

  DataCallback              :TQUB_MIL_DATA_CALLBACK;
  IdlDataCallback           :TQUB_MIL_IDLDATA_CALLBACK;
  ReportCallback            :TQUB_MIL_REPORT_CALLBACK;
  ResultsCallBack           :TQUB_MIL_RESULTS_CALLBACK;

  Stopped                   :Integer;

end;


TQUB_MIL_FUNCTION           = function(caller_obj :TObject; var MILInterface :TQUB_MIL_INTERFACE; MilChunks :TPAPointer{an array of pointers, each to a MilChunk}; MilChunkCount :Integer) :Integer; stdcall;
TQUB_MIL_FUNCTION_W_TREE    = function(caller_obj :TObject; var MILInterface :TQUB_MIL_INTERFACE; MilChunks :TPAPointer{an array of pointers, each to a MilChunk}; MilChunkCount :Integer; tree :TQUB_QFS_PNodeMem) :Integer; stdcall;


TQUB_MIL_MODEL              = record
  ClassCount                :Integer;
  ClassOfState              :TPAInteger;

  MetastateCount            :Integer;
  MetaclassCount            :Integer;
  MetapathCount             :Integer;
  MaxNMetastate             :Integer;
  MaxNMetapath              :Integer;

  Metaclass                 :TPAInteger;
  Metagroup                 :TPAInteger;
  Metapr                    :TPADouble;
  Metastate                 :TPMInteger;
  Metapath                  :TPMInteger;
  Subpath                   :TPMInteger;
  Metaindex                 :TPMInteger;

  // c++ objects that create and manage the above info
  mil_model                 :Pointer;
  mil_metamodel             :Pointer;
end;
TQUB_MIL_PMODEL  = ^TQUB_MIL_MODEL;

function MIL_Model_New( Model :TQUB_QFS_PNodeMem; ChannelCount :Integer ): TQUB_MIL_PMODEL; stdcall; external 'qubopt.dll' name '_new_mil_model@8';
procedure MIL_Model_Free( Model :TQUB_MIL_PMODEL ); stdcall; external 'qubopt.dll' name '_del_mil_model@4';


implementation


end.
