{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// constants and such for QUB_Task_MIP

unit QUB_MIP_Global;


interface


uses


AllocTypes, Alloc,

QUB_Thread,
QUB_QVO_QVD,
QUB_QFS_QDF,
QUB_QFS_QIF_Segments,
QUB_Q_Model,
QUB_Global,
QUB_Node;


const


//Conditions
DEF_VOLTAGE           :Double       = 0;
MIN_VOLTAGE           :Double       = -1e6;
MAX_VOLTAGE           :Double       = +1e6;

DEF_CONC              :Double       = 1;
MIN_CONC              :Double       = 0;
MAX_CONC              :Double       = 1e6;

//Optimization
MIN_OPT_LLCONV        :Double       = 1e-10;
MAX_OPT_LLCONV        :Double       = 1e10;
DEF_OPT_LLCONV        :Double       = 1e-3;

MIN_OPT_GRADCONV      :Double       = 1e-10;
MAX_OPT_GRADCONV      :Double       = 1e10;
DEF_OPT_GRADCONV      :Double       = 1e-3;

MIN_OPT_MAXITER              = 1;
MAX_OPT_MAXITER              = 10000;
DEF_OPT_MAXITER              = 20;

MIN_OPT_MAXSTEP       :Double       = 1e-10;
MAX_OPT_MAXSTEP       :Double       = 1e3;
DEF_OPT_MAXSTEP       :Double       = 0.1;

MIN_OPT_MAXPOWER             = 0;
MAX_OPT_MAXPOWER             = 16;
DEF_OPT_MAXPOWER             = 7;//128

MIN_OPT_STIMDELTA            = 1;
MAX_OPT_STIMDELTA            = High(Integer);
DEF_OPT_STIMDELTA            = 10;

DEF_OPT_STIMLATENCY          = 0;

DEF_OPT_RESTART              = 0;

DEF_OPT_ALLOWNEGRATES        = 0;

DEF_OPT_STARTATEQ            = 1;

MIN_MOD_MAXJUMP              = 1;
MAX_MOD_MAXJUMP              = 10;
DEF_MOD_MAXJUMP              = 2;

DEF_MOD_RIGHTONLY            = 0;
     
DEF_MOD_MINRATE       :Double       = 0.0;
MIN_MOD_MINRATE       :Double       = 0.0;
MAX_MOD_MINRATE       :Double       = 1e10;

MIN_OPT_MAXBATCH             = 1;
MAX_OPT_MAXBATCH             = 1000000;
DEF_OPT_MAXBATCH             = 100;

DEF_OPT_IDENTICALCHUNKS      = 0;

DEF_OPT_MUXFILES             = 0;


type


{$Z4}
TQUB_MIP_Opt_Mode            = (MODE_OPTIMIZE, MODE_CHECK);
{$Z1}

{$Z4}
TQUB_OptimizeSegments        = (OPTSEGS_TOGETHER, OPTSEGS_GROUP, OPTSEGS_GROUPLIST);
{$Z1}


const


DEF_OPT_MODE                 = MODE_OPTIMIZE;

DEF_OPT_SEGMENTS             = OPTSEGS_TOGETHER;
DEF_OPT_SEGMENTS_GROUPCOUNT  = 1;

DEF_EST_INITPROBS            = INITPROBS_FIXED;


function OptModeToString(x :TQUB_MIP_Opt_Mode): String;
function StringToOptMode(x :String): TQUB_MIP_Opt_Mode;

function SegModeToString(x :TQUB_OptimizeSegments): String;
function StringToSegMode(x :String): TQUB_OptimizeSegments;


type


TQUB_MIP_Opt_Params          = record

  EstInitProbs               :TQUB_Q_InitProbs;

  OptLLConv                  :Double;
  OptGradConv                :Double;
  OptMaxIter                 :Integer;
  OptRestart                 :Integer;
  OptSearchLimit             :Double;
  OptMaxStep                 :Double;
  OptMaxPower                :Integer;
  OptStimDelta               :Integer;
  OptIdenticalChunks         :Integer;
  OptMode                    :TQUB_MIP_Opt_Mode;
  OptSegments                :TQUB_OptimizeSegments;
  OptSegmentsGroupCount      :Integer;
  OptSegmentsMaxBatch        :Integer;
  OptMUXFiles                :Integer;

  OptAllowNegRates           :Boolean;

  ModMaxJump                 :Integer;//Used by Drift
  ModRightOnly               :Boolean;//
  ModQMinRate                :Double;

end;
TQUB_MIP_Opt_PParams         = ^TQUB_MIP_Opt_Params;

TQUB_MIP_DataPoint           = record
  Cls                        :ShortInt;
  Pow                        :Byte;
  ExpCondCls                 :Byte;
end;

TQUB_MIP_DataArray           = Array[0..(MaxInt div (SizeOf(TQUB_MIP_DataPoint)*2))] of TQUB_MIP_DataPoint;
TQUB_MIP_Data                = ^TQUB_MIP_DataArray;


TQUB_MIP_Chunk         = record

  Chunk                      :TQUB_Thread_Chunk;                                //Sel is in Data Points

  IdlIdx                     :Integer;
  IdlSel                     :TQUB_FirstLast;                                   //Is in Dwells, and referes to original DC

  IdlSeg                     :TQUB_QFS_PNodeMem;

  Sampling                   :Double;                                           //Seconds
  Scaling                    :Double;
  //ChannelCount               :Integer;

  Data                       :TQUB_MIP_Data;
  DataCount                  :Integer;

  ExpCondIdl                 :TQUB_QFS_QIF_PANClass;
  ExpCondIdlCount            :Integer;

  SkipPoints                 :TQUB_FirstLast;

  AlphaS                     :TPMExtended;
  BetaS                      :TPMExtended;
  LScale                     :TPAExtended;

  QVD                        :TQUB_QVO_QVD;
  QDF                        :TQUB_QFS_QDF;
  MIPModel                   :TQUB_Q_Model;
  FileIndex                  :Integer;

  LL                         :Double;                                           //Output: the LL of this segment, after the last iteration
  ResultSeg                  :TQUB_QFS_PNodeMem; // borrowed reference -- don't decref

end;
TQUB_MIP_PChunk              = ^TQUB_MIP_Chunk;


function MIP_PowerOf2(Pow :Integer) :Integer;
function MIP_MaxPower(Count, Max :Integer; var PowOf2 :Integer) :Integer;


implementation


//******************************************************************************
function OptModeToString(x :TQUB_MIP_Opt_Mode): String;
begin
  if x = MODE_OPTIMIZE then begin
    Result := 'optimize';
  end else begin // if OptMode = MODE_CHECK then begin
    Result := 'evaluate';
  end;
end;

function StringToOptMode(x :String): TQUB_MIP_Opt_Mode;
begin
  if x = 'optimize' then begin
    Result := MODE_OPTIMIZE;
  end else begin // if x = 'evaluate' then begin
    Result := MODE_CHECK;
  end;
end;

function SegModeToString(x :TQUB_OptimizeSegments): String;
begin
  if          x = OPTSEGS_TOGETHER  then begin
    Result := 'together';
  end else if x = OPTSEGS_GROUP     then begin
    Result := 'in groups of';
  end else if x = OPTSEGS_GROUPLIST then begin
    Result := 'in group list of';
  end;
end;

function StringToSegMode(x :String): TQUB_OptimizeSegments;
begin
  if          x = 'together'         then begin
    Result := OPTSEGS_TOGETHER;
  end else if x = 'in groups of'     then begin
    Result := OPTSEGS_GROUP;
  end else{if x = 'in group list of' then}begin
    Result := OPTSEGS_GROUPLIST;
  end;
end;


//******************************************************************************
function MIP_PowerOf2(Pow :Integer) :Integer;
var
  p :Integer;
begin
  p      := 0;
  Result := 1;
  while p < Pow do begin
    Inc(p);
    Result := Result shl 1;
  end;
end;

function MIP_MaxPower(Count, Max :Integer; var PowOf2 :Integer) :Integer;
var
  r      :Integer;
begin
  Result := 0;
  PowOf2 := 1;
  r      := 2;
  while ((r <= Count) and (r <= Max)) do begin
    Inc(Result);
    PowOf2 := r;
    r      := r shl 1;
  end;
end;


//******************************************************************************
end.
