{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// specialization of QUB_Q_Model for use by QUB_Task_MIP
// computes matrices and derivatives and stuff

unit QUB_MIP_Model;


interface


uses


Classes,

AllocTypes, Alloc,
qubopt_iface,

QUB_Global,
QUB_Matrices,

QUB_Q_Model, QUB_Q_Global,
QUB_MIP_Global,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix, QUB_QFS_Channels,
QUB_ExpCond_Parameters;


type


TQUB_MIP_Model               = class(TQUB_Q_Model)
protected


  prOptParams                :TQUB_MIP_Opt_PParams;


  procedure prAllocModelStuff; override;
  procedure prInitModelStuff; override;
  procedure prFreeModelStuff; override;

  function prGet_ChannelCount :Integer; override;


public


  //MIP - Specific:

  //Conductance matrices .......................................................
  B                          :TPAPointer; //The conductance matrices

  //A.B powers and their derivatives
  AB                         :TPAPointer;                                       ////
  dABdq                      :TPAPointer;                                       ////

  //Temporary quantities .......................................................
  Vec1E                      :TPAExtended;


  TotalDwellCount            :Integer;
  TotalDataPoints            :Integer;
  Multiplications            :Integer;


  constructor Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_MIP_Opt_PParams);


  function IsPZero(Rate, Index :Integer) :Boolean; override;
  function IsQZero(Rate, Index :Integer) :Boolean; override;

  function Do_fParsToModel :Integer; override;

  function Do_P0 :Integer; override;
  function Do_dP0_dq :Integer; override;
  function Do_dP0_dCond_q :Integer; override;

  function Do_A(t :Double) :Integer; override;
  function Do_A(t :Double; ExpCond :Integer) :Integer; override;
  function Do_dA_dq(t :Double) :Integer; override;
  function Do_dA_dq(t :Double; ExpCond :Integer) :Integer; override;

  function Do_APowers :Integer; overload;
  function Do_APowers(ExpCond :Integer) :Integer; overload;
  function Do_dAPowers_dq :Integer; overload;
  function Do_dAPowers_dq(ExpCond :Integer) :Integer; overload;


  property OptParams         :TQUB_MIP_Opt_PParams    read prOptParams;


end;


implementation


//******************************************************************************
constructor TQUB_MIP_Model.Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_MIP_Opt_PParams);
begin

  prOptParams    := OptParams;

  QMinRate       := prOptParams.ModQMinRate;
  AllowNegRates  := prOptParams.OptAllowNegRates;

  inherited Create(Model, DoInitialize, ExpCondParams);

end;



//****************************************************************************//
//                                                                            //
//                             C O N S T R A I N T S                          //
//                                                                            //
//****************************************************************************//
function TQUB_MIP_Model.IsPZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := False;
  if not prModel.Rates[Rate].P[Index] then Exit;
  //Not P-dependent --> not zero

  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].PName[Index]];
  if (Par <> nil) and (Par.Value = 0) and (not Par.IsVar) then begin
  // Lorin: I have never understood IsPZero, but if Par.Channel >= 0 then its value may change throughout the file.  
  //if (Par <> nil) and (Par.Channel < 0) and (Par.Value = 0) and (not Par.IsVar) then begin
  //TO DO: LM. Check this, or explain it better
    Result := True;
  end;

end;

function TQUB_MIP_Model.IsQZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := True;
  if not prModel.Rates[Rate].Q[Index] then Exit;
  //Not Q-dependent --> zero

  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].QName[Index]];
  if (Par = nil) then Exit;

  //LM. I commented out this, because Voltage = 0 has no special meaning
  //if (Par.Value <> 0) or (Par.IsVar) then begin
    Result := False;
  //end;

end;



//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_MIP_Model.Do_fParsToModel :Integer;
begin

  Result := inherited Do_fParsToModel;

  //Do_dP0_dq;

  //Do_Peq_dPeq_dq;

end;



//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_MIP_Model.prAllocModelStuff;
var
  i, j, k, c, p, ec :Integer;
  //Sum             :Double;
begin


  inherited;


  //MIP - specific:

  AB                           := PAlloc1(ExpCondCount);
  dABdq                        := PAlloc1(ExpCondCount);

  for ec := 0 to ExpCondCount - 1 do begin

    //A.B powers and their derivatives, for each class ...........................
    AB[ec]                     := PAlloc2(ClassCount, prOptParams.OptMaxPower + 1);
    for c := 0 to ClassCount - 1 do begin
      for p := 0 to prOptParams.OptMaxPower do begin
        TPMPointer(AB[ec])[c, p] := DAlloc2(StateCount, StateCount);
      end;
    end;

    dABdq[ec]                  := PAlloc1(RateCount);
    for k := 0 to RateCount - 1 do begin
      TPAPointer(dABdq[ec])[k]             := PAlloc2(ClassCount, prOptParams.OptMaxPower + 1);
      for c := 0 to ClassCount - 1 do begin
        for p := 0 to prOptParams.OptMaxPower do begin
          TPMPointer(TPAPointer(dABdq[ec])[k])[c, p] := DAlloc2(StateCount, StateCount);
        end;
      end;
    end;

  end;

  //Conductance matrices .......................................................
  B := PAlloc1(ClassCount);
  for c := 0 to ClassCount - 1 do begin
    B[c] := DAlloc2(StateCount, StateCount);
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        TPMDouble(B[c])[i, j] := 0;
      end;
    end;
    for j := 0 to StateCount - 1 do begin
      if prModel.States[j].Cls = Classes[c] then begin
        TPMDouble(B[c])[j, j] := 1;
      end else begin
        TPMDouble(B[c])[j, j] := 0;//1e-8;//To prevent zero state vector
      end;
    end;
  end;

  //Temporary quantities .......................................................
  Vec1E                      := EAlloc1(StateCount);

end;

procedure TQUB_MIP_Model.prInitModelStuff;
begin

  inherited;

  MakeInitProbs(prModel, P0, StateCount);

end;


procedure TQUB_MIP_Model.prFreeModelStuff;
var
  k, c, p, ec :Integer;
begin

  inherited;

  //MIP - specific:

  //A.B powers and their derivatives, for each class ...........................
  if AB <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      for c := 0 to ClassCount - 1 do begin
        for p := 0 to prOptParams.OptMaxPower do begin
          Free2(TPMChar(TPMPointer(AB[ec])[c, p]));
        end;
      end;
      Free2(TPMChar(AB[ec]));
    end;
    FreeMem(AB);
    AB := nil;
  end;

  if dABdq <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      for k := 0 to RateCount - 1 do begin
        for c := 0 to ClassCount - 1 do begin
          for p := 0 to prOptParams.OptMaxPower do begin
            Free2(TPMChar(TPMPointer(TPAPointer(dABdq[ec])[k])[c, p]));
          end;
        end;
        Free2(TPMChar(TPAPointer(dABdq[ec])[k]));
      end;
      FreeMem(dABdq[ec]);
    end;
    FreeMem(dABdq);
    dABdq := nil;
  end;

  //Conductance matrices .......................................................
  if B <> nil then begin
    for c := 0 to ClassCount - 1 do begin
      Free2(TPMChar(B[c]));
    end;
  end;
  FreeMem(B);
  B := nil;

  //Temporary quantities .......................................................
  FreeMem(Vec1E);
  Vec1E := nil;

end;



//****************************************************************************//
//                                                                            //
//                         M I P   -    s p e c i f i c                       //
//                                                                            //
//****************************************************************************//
function TQUB_MIP_Model.Do_P0 :Integer;
var
  i :Integer;
begin

  Result := 0;

  if          prOptParams.EstInitProbs = INITPROBS_FIXED  then begin

    //The fixed probabilities from the model
    Result := inherited Do_P0;

  end else if prOptParams.EstInitProbs = INITPROBS_EQ     then begin

    //Calculate the equilibrium probabilities
    Result := Do_Peq;
    for i := 0 to StateCount - 1 do begin
      P0[i] := Peq[i];
    end;

  end else if prOptParams.EstInitProbs = INITPROBS_CONDEQ then begin

    //Calculate the equilibrium probabilities for the Conditioning experimental conditions
    Result := Do_Cond_Peq;
    for i := 0 to StateCount - 1 do begin
      P0[i] := Cond_Peq[i];
    end;

  end;

end;

function TQUB_MIP_Model.Do_dP0_dq :Integer;
var
  qq, k, ec :Integer;
begin

  Result := 0;

  if          prOptParams.EstInitProbs = INITPROBS_FIXED  then begin
  //Fixed P0 - derivatives are zero

    Result := inherited Do_dP0_dq;

  end else if prOptParams.EstInitProbs = INITPROBS_EQ     then begin
  //Equilibrium P0 - calculate

    Result := Do_dPeq_dq;
    for ec := 0 to ExpCondCount - 1 do begin
      for qq := 0 to RateCount - 1 do begin
        for k := 0 to StateCount - 1 do begin
          TPMDouble(dP0dq[ec])[qq, k] := TPMDouble(dPeqdq[ec])[qq, k];
        end;
      end;
    end;

  end else if prOptParams.EstInitProbs = INITPROBS_CONDEQ then begin
  //Conditioning P0 - derivatives are zero

    Result := inherited Do_dP0_dq;

  end;

end;

function TQUB_MIP_Model.Do_dP0_dCond_q :Integer;
var
  qq, k :Integer;
begin

  Result := 0;

  if          prOptParams.EstInitProbs = INITPROBS_FIXED  then begin
  //Fixed P0 - derivatives are zero

    Result := inherited Do_dP0_dCond_q;

  end else if prOptParams.EstInitProbs = INITPROBS_EQ     then begin
  //Equilibrium P0 - derivatives are zero

    Result := inherited Do_dP0_dCond_q;

  end else if prOptParams.EstInitProbs = INITPROBS_CONDEQ then begin
  //Conditioning P0

    Result := Do_dCond_Peq_dCond_q;
    for qq := 0 to RateCount - 1 do begin
      for k := 0 to StateCount - 1 do begin
        dP0dCond_q[qq, k] := dCond_PeqdCond_q[qq, k];
      end;
    end;

  end;

end;



//****************************************************************************//
//                                                                            //
//                              A    M a t r i x                              //
//                                                                            //
//****************************************************************************//
function TQUB_MIP_Model.Do_A(t :Double) :Integer;
begin

  Result := inherited Do_A(t);

  if Result = 0 then if Do_APowers <> 0 then Result := 1;

end;

function TQUB_MIP_Model.Do_A(t :Double; ExpCond :Integer) :Integer;
begin

  Result := inherited Do_A(t, ExpCond);

  if Result = 0 then if Do_APowers(ExpCond) <> 0 then Result := 1;

end;

function TQUB_MIP_Model.Do_dA_dq(t :Double) :Integer;
begin

  Result := inherited Do_dA_dq(t);

  if Result = 0 then if Do_dAPowers_dq <> 0 then Result := 1;

end;

function TQUB_MIP_Model.Do_dA_dq(t :Double; ExpCond :Integer) :Integer;
begin

  Result := inherited Do_dA_dq(t, ExpCond);

  if Result = 0 then if Do_dAPowers_dq(ExpCond) <> 0 then Result := 1;

end;

//A.B Powers
function TQUB_MIP_Model.Do_APowers :Integer;
var
  c, p, ec :Integer;
begin

  Result := 0;

  for ec := 0 to ExpCondCount - 1 do begin
    //AB[c, p] (StateCount, StateCount);
    for c := 0 to ClassCount - 1 do begin
      p := 0;
      MulMatMat(TPMDouble(A[ec]), B[c], TPMPointer(AB[ec])[c, p], StateCount);
      for p := 1 to prOptParams.OptMaxPower do begin
        MulMatMat(TPMPointer(AB[ec])[c, p - 1], TPMPointer(AB[ec])[c, p - 1], TPMPointer(AB[ec])[c, p], StateCount);
      end;//MaxPower
    end;//Classcount

  end;

end;

function TQUB_MIP_Model.Do_APowers(ExpCond :Integer) :Integer;
var
  c, p :Integer;
begin

  Result := 0;

  //AB[c, p] (StateCount, StateCount);
  for c := 0 to ClassCount - 1 do begin
    p := 0;
    MulMatMat(TPMDouble(A[ExpCond]), B[c], TPMPointer(AB[ExpCond])[c, p], StateCount);
    for p := 1 to prOptParams.OptMaxPower do begin
      MulMatMat(TPMPointer(AB[ExpCond])[c, p - 1], TPMPointer(AB[ExpCond])[c, p - 1], TPMPointer(AB[ExpCond])[c, p], StateCount);
    end;//MaxPower
  end;//Classcount

end;

function TQUB_MIP_Model.Do_dAPowers_dq :Integer;
var
  k, c, p, ec :Integer;
begin

  Result := 0;

  for ec := 0 to ExpCondCount - 1 do begin
    //dABdq(RateCount)
    for k := 0 to RateCount - 1 do begin
      //dABdq[k](ClassCount, prOptParams.OptMaxPower + 1);
      for c := 0 to ClassCount - 1 do begin
        p := 0;
        MulMatMat(TPAPointer(dAdq[ec])[k], B[c], TPMPointer(TPAPointer(dABdq[ec])[k])[c, p], StateCount);
        for p := 1 to prOptParams.OptMaxPower do begin
          MulMatMat(TPMPointer(TPAPointer(dABdq[ec])[k])[c, p - 1], TPMPointer(AB[ec])[c, p - 1], Mat1, StateCount);
          MulMatMat(TPMPointer(AB[ec])[c, p - 1], TPMPointer(TPAPointer(dABdq[ec])[k])[c, p - 1], Mat2, StateCount);
          SumMatMat(Mat1, Mat2, TPMPointer(TPAPointer(dABdq[ec])[k])[c, p], StateCount);
        end;//MaxPower + 1
      end;//ClassCount
    end;//RateCount
  end;//ExpCondCount

end;

function TQUB_MIP_Model.Do_dAPowers_dq(ExpCond :Integer) :Integer;
var
  k, c, p :Integer;
begin

  Result := 0;

  //dABdq(RateCount)
  for k := 0 to RateCount - 1 do begin
    //dABdq[k](ClassCount, prOptParams.OptMaxPower + 1);
    for c := 0 to ClassCount - 1 do begin
      p := 0;
      MulMatMat(TPAPointer(dAdq[ExpCond])[k], B[c], TPMPointer(TPAPointer(dABdq[ExpCond])[k])[c, p], StateCount);
      for p := 1 to prOptParams.OptMaxPower do begin
        MulMatMat(TPMPointer(TPAPointer(dABdq[ExpCond])[k])[c, p - 1], TPMPointer(AB[ExpCond])[c, p - 1], Mat1, StateCount);
        MulMatMat(TPMPointer(AB[ExpCond])[c, p - 1], TPMPointer(TPAPointer(dABdq[ExpCond])[k])[c, p - 1], Mat2, StateCount);
        SumMatMat(Mat1, Mat2, TPMPointer(TPAPointer(dABdq[ExpCond])[k])[c, p], StateCount);
      end;//MaxPower + 1
    end;//ClassCount
  end;//RateCount

end;

function TQUB_MIP_Model.prGet_ChannelCount :Integer;
var
  Par    :TQUB_ExpCond_PParameter;
begin

  Result := inherited prGet_ChannelCount;

  Par    := ExpCondParams[0].ParameterByName['ChannelCount'];
  if Par <> nil then begin
    if Par.Value > 0 then begin
      try
        Result := Round(Par.Value);
      except
      end;
    end;
  end;

end;


//******************************************************************************
end.
