{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// constants and such for QUB_Task_Mac

unit QUB_Mac_Global;


interface


uses


Classes,

AllocTypes, Alloc,

QUB_Thread,             //QUB_Task,
QUB_QVO_QVD,
QUB_QFS_QDF,            //QUB_ExpCond_Parameters,
QUB_QFS_QIF_Segments,
QUB_Q_Model,
QUB_Node,
QUB_AdaptiveResample;


type


//----- Enumerated types for radio button control selections
{$Z4}
TQUB_Mac_Opt_Mode            = (MODE_OPTIMIZE, MODE_CHECK, MODE_SIM);
TQUB_Mac_Opt_FitDataMode     = (FITDATA_WHOLE, FITDATA_LIST);
TQUB_Mac_Opt_Segments        = (OPTSEGS_TOGETHER, OPTSEGS_GROUP, OPTSEGS_GROUPLIST);
{$Z1}


function MacFitDataModeStringOf(x :TQUB_Mac_Opt_FitDataMode): String;
function MacFitDataModeOfString(x :String): TQUB_Mac_Opt_FitDataMode;
function MacOptModeStringOf(x :TQUB_Mac_Opt_Mode): String;
function MacOptModeOfString(x :String): TQUB_Mac_Opt_Mode;
function MacSegModeStringOf(x :TQUB_Mac_Opt_Segments): String;
function MacSegModeOfString(x :String): TQUB_Mac_Opt_Segments;


const


//TODO: The conductance/voltage mechanism.
//TODO: Express delta stimulus in mV units as well.


//----- The following constants are used in the properties dialogue for various control defaults
DEF_EST_RATES                = 1;
DEF_EST_CHANNELCOUNT         = 0;
DEF_EST_CHANNELCOUNTPERFILE  = 0;
DEF_EST_AMPOPEN              = 0;
DEF_EST_VAROPEN              = 0;
DEF_EST_INITPROBS            = INITPROBS_FIXED;
DEF_EST_USEVARIANCE          = 0;
//DEF_EST_PENALTY              = False;
//DEF_EST_PENALTY_HIGHRATE     = 500.0;
//MIN_EST_PENALTY_HIGHRATE     = 0.0;
//MAX_EST_PENALTY_HIGHRATE     = 1.0e10;
DEF_EST_USEMINRATE           = 0;
DEF_EST_MINRATE              :Double = 1e-3;
MIN_EST_MINRATE              :Double = 0.0;
MAX_EST_MINRATE              :Double = 1.0e10;
DEF_EST_USEMAXRATE           = 0;
DEF_EST_MAXRATE              :Double = 1e6;
MIN_EST_MAXRATE              :Double = 0.0;
MAX_EST_MAXRATE              :Double = 1.0e10;

MIN_OPT_LLCONV               :Double = 1e-10;
MAX_OPT_LLCONV               :Double = 1e10;
DEF_OPT_LLCONV               :Double = 1e-3;

MIN_OPT_GRADCONV             :Double = 1e-10;
MAX_OPT_GRADCONV             :Double = 1e10;
DEF_OPT_GRADCONV             :Double = 1e-3;

MIN_OPT_MAXITER              = 1;
MAX_OPT_MAXITER              = 1000;
DEF_OPT_MAXITER              = 20;

DEF_OPT_STIMLATENCY          = 0;
MIN_OPT_STIMLATENCY          = -100;
MAX_OPT_STIMLATENCY          =  100;
DEF_OPT_STIMDELTA            = 10;
MIN_OPT_STIMDELTA            = 1;
//MAX_OPT_STIMDELTA            = High(Integer);

DEF_OPT_RESTART              = 0;

DEF_OPT_NUMGRADS             = 0;

MIN_OPT_NUMGRADSTEP          :Double = 1e-10;
MAX_OPT_NUMGRADSTEP          :Double = 1e10;
DEF_OPT_NUMGRADSTEP          :Double = 1e-3;

DEF_OPT_NUMGRADSFAST         = 0;

DEF_OPT_PLOTLL               = 0;
DEF_OPT_PLOTLLPAGE           = 0;
DEF_OPT_PLOTLLTRACE          = 0;

MIN_OPT_MAXSTEP              :Double = 1e-10;
MAX_OPT_MAXSTEP              :Double = 1e3;
DEF_OPT_MAXSTEP              :Double = 0.1;//LM. 0.005 is not a good default, in my experience. Who uses it?  Fred.

DEF_OPT_SKIP                 = 0;
DEF_OPT_SKIPTIME            :Double = 0;
MIN_OPT_SKIPTIME            :Double = 0;
MAX_OPT_SKIPTIME            :Double = 1e10;

DEF_OPT_SHOWFIT              = 0;

SHOWFIT_AVG                  = 0;
SHOWFIT_STD                  = 1;
SHOWFIT_RES                  = 2;

DEF_OPT_SHOWFIT_WHAT         = Integer(SHOWFIT_AVG);

MIN_OPT_MAXBATCH             = 1;
MAX_OPT_MAXBATCH             = 1000000;
DEF_OPT_MAXBATCH             = 100;

DEF_OPT_IDENTICALCHUNKS      = 0;

DEF_OPT_MUXFILES             = 0;

DEF_OPT_PROPAGATEPARAMS      = 0;

DEF_OPT_USEVREV              = 0;
DEF_OPT_VREV                 :Double = 0.0;
MIN_OPT_VREV                 :Double = -10000.0;
MAX_OPT_VREV                 :Double = +10000.0;

DEF_OPT_GLEAK                :Double = 0.0;
MIN_OPT_GLEAK                :Double = 0.0;
MAX_OPT_GLEAK                :Double = 1e10;

DEF_OPT_MODE                 = MODE_OPTIMIZE;

DEF_OPT_SEGMENTS             = OPTSEGS_TOGETHER;
DEF_OPT_SEGMENTS_GROUPCOUNT  = 1;

DEF_OPT_FITDATAMODE          = FITDATA_WHOLE;
DEF_OPT_EXCDATA              = 0;

//DEF_MOD_RATE                 = 0.0;
//MIN_MOD_RATE                 = 0.0;
//MAX_MOD_RATE                 = 1e10;
DEF_MINRATE                  :Double = 1e-3;
MIN_MINRATE                  :Double = 1e-10;
MAX_MINRATE                  :Double = 1e10;
DEF_MAXRATE                  :Double = 1e5;
MIN_MAXRATE                  :Double = 1e-10;
MAX_MAXRATE                  :Double = 1e10;

DEF_SKIPSLOWCHANGE           = 0;
DEF_SKIPSLOWSTD              :Double = 1.0;

DEF_OPT_SCRIPTPENALTY        = 0;
DEF_OPT_SCRIPTPENALTYNAME    = '<none>';
DEF_OPT_SCRIPTITERATION      = 0;
DEF_OPT_SCRIPTITERATIONNAME  = '<none>';
DEF_OPT_SCRIPTSETPARS        = 0;
DEF_OPT_SCRIPTSETPARSNAME    = '<none>';
DEF_MINABSSTIMULUS           :Double = 0.0;


type


//----- Collection of control settings for Task_Mac_Properties
// does not include settings for pre process data, quiet output.
TQUB_Mac_Opt_Params          = record

  EstRates                   :Integer;
  EstInitProbs               :TQUB_Q_InitProbs;
  EstChannelCount            :Integer;
  EstChannelCountPerFile     :Integer;
  EstAmpOpen                 :Integer;
  EstVarOpen                 :Integer;
  EstUseVariance             :Integer;
  //EstPenalty                 :Integer;
  //EstPenaltyHighRate         :Double;
  EstMinRate                 :Double;
  EstMaxRate                 :Double;
  EstUseMinRate              :Boolean;
  EstUseMaxRate              :Boolean;

  OptLLConv                  :Double;
  OptGradConv                :Double;
  OptMaxIter                 :Integer;
  OptRestart                 :Integer;
  OptNumGrads                :Integer;
  OptNumGradStep             :Double;
  OptNumGradsFast            :Integer;
  OptSearchLimit             :Double;
  OptMaxStep                 :Double;
  OptPlotLL                  :Integer;
  OptPlotLLPage              :Integer;
  OptPlotLLTrace             :Integer;
  OptStimLatency             :Integer;
  OptStimDelta               :Integer;
  OptSkip                    :Integer;
  OptSkipTime                :Double;
  OptShowFit                 :Integer;
  OptShowFitWhat             :Integer;
  OptIdenticalChunks         :Integer;
  OptMode                    :TQUB_Mac_Opt_Mode;
  OptSegments                :TQUB_Mac_Opt_Segments;
  OptSegmentsGroupCount      :Integer;
  OptSegmentsMaxBatch        :Integer;
  OptFitDataMode             :TQUB_Mac_Opt_FitDataMode;
  OptExcData                 :Integer;
  OptMUXFiles                :Integer;
  OptPropagateParams         :Integer;
  OptUseVRev                 :Integer;
  OptVRevFromModel           :Boolean;
  OptVRev                    :Double;
  OptGLeakFromTable          :Boolean;
  OptGLeak                   :Double;
  OptVLeakFromTable          :Boolean;
  OptVLeak                   :Double;

  OptAllowNegRates           :Boolean;

  OptScriptPenalty           :Integer;
  OptScriptPenaltyName       :String;
  OptScriptIteration         :Integer;
  OptScriptIterationName     :String;
  OptScriptSetPars           :Integer;
  OptScriptSetParsName       :String;

  ModQMinRate                :Double;

  ChannelCountCount          :Integer;
  ChannelCountIdx            :Integer;

end;
TQUB_Mac_Opt_PParams         = ^TQUB_Mac_Opt_Params;


TQUB_Mac_Chunk               = record

  Chunk                      :TQUB_Thread_Chunk;                                //Sel is in Data Points

  Sampling                   :Double;                                           //Seconds
  Scaling                    :Double;
  Offset                     :Double;

  Data                       :TPAInteger;

  HasTime                    :Boolean;
  TimeData                   :TPASingle;
  TimeScaling                :Double;
  TimeOffset                 :Double;
  TimeChIndex                :Integer;

  FitPoint                   :TPABoolean;

  ExpCondIdl                 :TQUB_QFS_QIF_PANClass;
  ExpCondIdlCount            :Integer;

  ResampleCount              :Integer;
  ResampleIntervals          :TQUB_PAAdResInterval;
  ResampleSamples            :TPAInteger;

  QVD                        :TQUB_QVO_QVD;
  QDF                        :TQUB_QFS_QDF;
  MacModel                   :TQUB_Q_Model;
  MacModels                  :TList;
  FileIndex                  :Integer;

  LL                         :Double;                                           //Output: the LL of this segment, after the last iteration

  FitCurve                   :TPAInteger;
  ResultSeg                  :TQUB_QFS_PNodeMem; // borrowed reference -- don't decref

  FitList                    :TPAPoint;                                         //Now each chunk has its own lists, to allow for Fit/Exclude lists accross files
  FitListCount               :Integer;
  FitListIdx                 :Integer;
  ExcList                    :TPAPoint;
  ExcListCount               :Integer;
  ExcListIdx                 :Integer;
end;
TQUB_Mac_PChunk              = ^TQUB_Mac_Chunk;


implementation


//******************************************************************************
function MacFitDataModeStringOf(x :TQUB_Mac_Opt_FitDataMode): String;
begin
  if          x = FITDATA_WHOLE    then begin
    Result := 'Whole';
  end else if x = FITDATA_LIST     then begin
    Result := 'List';
  end;
end;

function MacFitDataModeOfString(x :String): TQUB_Mac_Opt_FitDataMode;
begin
  if          x = 'Whole'          then begin
    Result := FITDATA_WHOLE;
  end else{if x = 'List'           then}begin
    Result := FITDATA_LIST;
  end;
end;

function MacOptModeStringOf(x :TQUB_Mac_Opt_Mode): String;
begin
  if          x = MODE_OPTIMIZE then begin
    Result := 'optimize';
  end else if x = MODE_CHECK    then begin
    Result := 'evaluate';
  end else if x = MODE_SIM      then begin
    Result := 'simulate';
  end;
end;

function MacOptModeOfString(x :String): TQUB_Mac_Opt_Mode;
begin
  if          x = 'optimize' then begin
    Result := MODE_OPTIMIZE;
  end else if x = 'evaluate' then begin
    Result := MODE_CHECK;
  end else begin // if x = 'simulate' then begin
    Result := MODE_SIM;
  end;
end;

function MacSegModeStringOf(x :TQUB_Mac_Opt_Segments): String;
begin
  if          x = OPTSEGS_TOGETHER  then begin
    Result := 'together';
  end else if x = OPTSEGS_GROUP     then begin
    Result := 'in groups of';
  end else if x = OPTSEGS_GROUPLIST then begin
    Result := 'in group list of';
  end;
end;

function MacSegModeOfString(x :String): TQUB_Mac_Opt_Segments;
begin
  if          x = 'together'         then begin
    Result := OPTSEGS_TOGETHER;
  end else if x = 'in groups of'     then begin
    Result := OPTSEGS_GROUP;
  end else{if x = 'in group list of' then}begin
    Result := OPTSEGS_GROUPLIST;
  end;
end;
 

//******************************************************************************
end.
