{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// specialization of QUB_Q_Model for QUB_Task_Mac
// computes matrices and derivatives and such

unit QUB_Mac_Model;


interface


uses


Classes,

AllocTypes, Alloc,

QUB_Cuda,
QUB_Q_Model, QUB_Q_Global,
QUB_Mac_Global,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix,
QUB_ExpCond_Parameters;


type


TQUB_Mac_Model               = class(TQUB_Q_Model)
protected


  prOptParams                :TQUB_Mac_Opt_PParams;
  prCuda_dAdq                :TQUB_Cuda_dAdq;
  prGklt                     :TPMDouble;

  procedure prAllocModelStuff; override;
  procedure prInitModelStuff; override;
  procedure prFreeModelStuff; override;

  procedure prInitOptStuff; override;

  function prGet_cParCount :Integer; override;
  function prGet_ChannelCount :Integer; override;
  function prGet_AverageCount :Integer; override;
  function prGet_FitWeight :Double; override;
  function prGet_VRev :Double;
  function prGet_VLeak :Double;
  function prGet_GLeak :Double;
  function prGet_CCFixed(ExpCondParams :TQUB_ExpCond_ParametersList) :Boolean;


public


  //Mac - Specific:

  //Conductance Info ...........................................................
  xAmp                       :TPADouble;  //Excess state amplitudes
  xVar                       :TPADouble;  //Excess state variances
  bAmp                       :Double;     //Baseline amplitude
  bVar                       :Double;     //Baseline variance

  MinAmp                     :Double;     //The smallest amplitude

  VarPt                      :TPMDouble;  //State covariance matrix at time t
  dVarPtdq                   :TPMPointer; //The derivative of state cov. matrix at t w.r.t. each q
  dVarPtdCond_q              :TPAPointer; //The derivative of state cov. matrix at t w.r.t. each Cond q

  dLLdN                      :Double;     //The derivative of LL w.r.t. ChannelCount
  dLLdA                      :Double;     //The derivative of LL w.r.t. AmpOpen
  dLLdV                      :Double;     //The derivative of LL w.r.t. VarOpen

  ChannelCountIdx            :Integer;    //Used with EstChannelCountPerFile; if -1 then it is fixed
  InitChannelCount           :Integer;
  ChannelCountCount          :Integer;
  AmpOpenCount               :Integer;
  VarOpenCount               :Integer;

  xAmpOpen                   :Double;     //the excess amplitude of the open class, to be estimated
  xVarOpen                   :Double;     //the excess variance of the open class, to be estimated

  CCFixed                    :Boolean;

  FitP                       :TPADouble;  //Double so I can plug it in the same formulae as xAmp
  HasFitP                    :Boolean;


  constructor Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_Mac_Opt_PParams; CudaLib :TQUB_Cuda_Lib);
  destructor Destroy; override;


  function ResetLLandGR :Boolean; override;

  function Do_InitcPars :Integer; override;

  function IsPZero(Rate, Index :Integer) :Boolean; override;
  function IsQZero(Rate, Index :Integer) :Boolean; override;

  function Do_fParsToModel :Integer; override;
  function Do_cParsToN :Integer;
  function Do_cParsToA :Integer;
  function Do_cParsToV :Integer;

  function Do_P0 :Integer; override;
  function Do_dP0_dq :Integer; override;
  function Do_dP0_dCond_q :Integer; override;
  procedure RestartPt;

  function Do_VarPt :Integer;
  function Do_dVarPt_dq :Integer;
  function Do_dVarPt_dCond_q :Integer;

  function Do_Mean_t(var mt :Double) :Integer; overload;
  function Do_Mean_t(var mt :Double; dV, ILeak :Double) :Integer; overload;
  function Do_dMean_t_dq(var dmt :Double; qq :Integer; ec :Integer) :Integer;
  function Do_dMean_t_dCond_q(var dmt :Double; qq :Integer) :Integer;
  function Do_dMean_t_dN(var dmt :Double) :Integer;
  function Do_dMean_t_dA(var dmt :Double) :Integer;
  function Do_dMean_t_dV(var dmt :Double) :Integer;

  function Do_Var_t(var Vt :Double) :Integer; overload;
  function Do_Var_t(var Vt :Double; dV :Double) :Integer; overload;
  function Do_dVar_t_dq(var dVt :Double; qq :Integer; ec :Integer) :Integer;
  function Do_dVar_t_dCond_q(var dVt :Double; qq :Integer) :Integer;
  function Do_dVar_t_dN(var dVt :Double) :Integer;
  function Do_dVar_t_dA(var dVt :Double) :Integer;
  function Do_dVar_t_dV(var dVt :Double) :Integer;

  function Do_dLLdfPar :Integer; override;
  function Do_Varf2Varc :Integer; override;

  // cuda -- also Do_fParsToModel
  function Do_dAt_dq(t :Double; ExpCond :Integer) :Integer; override;


  property OptParams         :TQUB_Mac_Opt_PParams    read prOptParams;

  property CCount            :Integer                 read prGet_ChannelCount;
  property VRev              :Double                  read prGet_VRev;
  property VLeak             :Double                  read prGet_VLeak;
  property GLeak             :Double                  read prGet_GLeak;


  //testing - John
  //procedure AMatrixAndCurrent(t :Double; ECClass :Integer);
  //procedure CopyPtToP0;


end;


implementation


//******************************************************************************
constructor TQUB_Mac_Model.Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ExpCondParams :TQUB_ExpCond_ParametersList; OptParams :TQUB_Mac_Opt_PParams; CudaLib :TQUB_Cuda_Lib);
begin

  prOptParams := OptParams;

  QMinRate    := prOptParams.ModQMinRate;

  CCFixed     := prGet_CCFixed(ExpCondParams);

  if prOptParams.EstChannelCount = 0 then begin
  //Don't estimate ChannelCount

    ChannelCountCount := 0;
    ChannelCountIdx   := 0;

  end else begin
  //Estimate ChannelCount

    if prOptParams.EstChannelCountPerFile = 0 then begin
    //Global

      ChannelCountCount := 1;
      ChannelCountIdx   := 0;

    end else begin
    //Per file

      ChannelCountCount := prOptParams.ChannelCountCount;

      if CCFixed then begin
      //Fixed for this model
        ChannelCountIdx   := -1;
      end else begin
        ChannelCountIdx   := prOptParams.ChannelCountIdx;
      end;

    end;

  end;

  if prOptParams.EstAmpOpen = 0 then begin
    AmpOpenCount := 0;
  end else begin
    AmpOpenCount := 1;
  end;

  inherited Create(Model, DoInitialize, ExpCondParams);

  if CudaLib.Avail and (0<>prOptParams.EstUseVariance) and (StateCount >= 4) then begin
    prCuda_dAdq := TQUB_Cuda_dAdq.Create(CudaLib, ExpCondCount, StateCount, RateCount);
    if prCuda_dAdq.Context = nil then begin
      prCuda_dAdq.Free;
      prCuda_dAdq := nil;
      prGklt      := nil;
    end else begin
      prGklt      := DAlloc2(StateCount, StateCount);
      prCuda_dAdq.Set_dQdq(dQdq);
    end;
  end else begin
    prCuda_dAdq := nil;
    prGklt      := nil;
  end;

end;

destructor TQUB_Mac_Model.Destroy;
begin
  try
    if prCuda_dAdq <> nil then prCuda_dAdq.Free;
    if prGklt <> nil      then Free2(TPMChar(prGklt));
  finally
    inherited;
  end;
end;


//****************************************************************************//
//                                                                            //
//                             C O N S T R A I N T S                          //
//                                                                            //
//****************************************************************************//
function TQUB_Mac_Model.IsPZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := False;
  if not prModel.Rates[Rate].P[Index] then Exit;
  //Not P-dependent --> not zero


  //Chris, see comment in TaskMac.

  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].PName[Index]];
  if Par = nil then begin
  //Par does not exist, unknown, thus not zero (?!)
    Exit;
  end else if Par.Channel <> -1 then begin
  //Par changes within data. Maybe one value is not zero ... Why take chances ?
    Exit;
  end else if (Par.Value = 0) and (not Par.IsVar{across data}) then begin
  //Since nobody sets IsVar, skip it:
  //Somebody should. See my comment to Chris. LM
    Result := True;
  end;

end;

function TQUB_Mac_Model.IsQZero(Rate, Index :Integer) :Boolean;
var
  Par       :TQUB_ExpCond_PParameter;
begin

  Result    := True;
  if not prModel.Rates[Rate].Q[Index] then Exit;
  //Not Q-dependent --> zero

  Par := ExpCondParams[0].ParameterByName[prModel.Rates[Rate].QName[Index]];
  if (Par = nil) then Exit;

  if (Par.Value <> 0) or (Par.IsVar) or (Par.Channel <> -1) then begin
    Result := False;
  end;

end;


//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_Mac_Model.Do_fParsToModel :Integer;
begin

  Result := inherited Do_fParsToModel;

  Do_cParsToN;
  Do_cParsToA;
  Do_cParsToV;

  if prCuda_dAdq <> nil then begin
    prCuda_dAdq.Set_Aek(Ak);
  end;
end;

function TQUB_Mac_Model.Do_dAt_dq(t :Double; ExpCond :Integer) :Integer;
var
  k, l        :Integer;
begin
  if prCuda_dAdq <> nil then begin
    for k := 0 to StateCount - 1 do begin
      for l := 0 to StateCount - 1 do begin
        Do_G_kl_t(k, l, ExpCond, t, prGklt[k,l]);
      end;
    end;
    prCuda_dAdq.Set_Gklt(prGklt);
    prCuda_dAdq.Get_dAdq(dAtdq[ExpCond], ExpCond);
    
    Result := 0;
  end else begin
    Result := inherited Do_dAt_dq(t, ExpCond);
  end;
end;

function TQUB_Mac_Model.Do_cParsToN :Integer;
begin

  Result := 0;

  if prOptParams.EstChannelCount = 0 then Exit;

  if ChannelCountIdx >= 0 then begin
    ChannelCount := Exp(cPars[2*RateCount + ExtraKinParCount + ChannelCountIdx]);
  end;

end;

function TQUB_Mac_Model.Do_cParsToA :Integer;
var
  i :Integer;
begin

  Result := 0;

  //AmpOpen
  if prOptParams.EstAmpOpen <> 0 then begin
    xAmpOpen := cPars[2*RateCount + ExtraKinParCount + ChannelCountCount];
    for i := 0 to StateCount - 1 do begin
      if Cls[i] = 1 then begin
        xAmp[i] := xAmpOpen;//new
      end;
    end;
  end;

end;

function TQUB_Mac_Model.Do_cParsToV :Integer;
var
  i :Integer;
begin

  Result := 0;

  //StdOpen
  if prOptParams.EstVarOpen <> 0 then begin
    xVarOpen := Exp(cPars[2*RateCount + ExtraKinParCount + ChannelCountCount + AmpOpenCount]);
    for i := 0 to StateCount - 1 do begin
      if Cls[i] = 1 then begin
        xVar[i] := xVarOpen;//new
      end;
    end;
  end;

end;

//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Mac_Model.prAllocModelStuff;
var
  ec, k :Integer;
begin


  inherited;


  //Mac - specific:

  //Conductance Info ...........................................................
  xAmp                       := DAlloc1(StateCount);
  xVar                       := DAlloc1(StateCount);

  //Variance ...................................................................
  VarPt                      := DAlloc2(StateCount, StateCount);

  dVarPtdq                   := PAlloc2(ExpCondCount, RateCount);
  dVarPtdCond_q              := PAlloc1(RateCount);
  for ec := 0 to ExpCondCount - 1 do begin
    for k := 0 to RateCount - 1 do begin
      dVarPtdq[ec, k]        := DAlloc2(StateCount, StateCount);
    end;
  end;
  for k := 0 to RateCount - 1 do begin
    dVarPtdCond_q[k]         := DAlloc2(StateCount, StateCount);
  end;

  //Fit States
  FitP                       := DAlloc1(StateCount);
  HasFitP                    := False;

end;

procedure TQUB_Mac_Model.prInitModelStuff;
var
  s :Integer;
begin

  inherited;

  if 0 <> prOptParams.OptUseVRev then begin
    MakeCondVar(prModel, bAmp, bVar, xAmp, xVar, StateCount);
  end else begin
    MakeAmpVar(prModel, bAmp, bVar, xAmp, xVar, StateCount);
  end;
  
  //If xAmp is per file (but not estimated):
  //Beep;

  //AmpOpen
  xAmpOpen := 0;
  for s := 0 to StateCount - 1 do begin
    if Cls[s] = 1 then begin
      xAmpOpen := xAmp[s];
      Break;
    end;
  end;

  //StdOpen
  xVarOpen := 0;
  for s := 0 to StateCount - 1 do begin
    if Cls[s] = 1 then begin
      xVarOpen := xVar[s];
      Break;
    end;
  end;

  MinAmp := xAmp[0];
  for s := 1 to StateCount - 1 do begin
    if ((MinAmp = 0) and (xAmp[s] <> 0)) or ((MinAmp <> 0) and (xAmp[s] <> 0) and (Abs(xAmp[s]) < Abs(MinAmp))) then begin
      MinAmp := xAmp[s];
    end;
  end;

end;


procedure TQUB_Mac_Model.prFreeModelStuff;
var
  ec, k :Integer;
begin

  inherited;

  //Mac - specific:

  //Conductance Info ...........................................................
  FreeMem(xAmp);
  xAmp := nil;
  FreeMem(xVar);
  xVar := nil;

  //Variance ...................................................................
  Free2(TPMChar(VarPt));
  VarPt := nil;

  if dVarPtdq <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      for k := 0 to RateCount - 1 do begin
        Free2(TPMChar(dVarPtdq[ec, k]));
        dVarPtdq[ec, k] := nil;
      end;
    end;
  end;
  Free2(TPMChar(dVarPtdq));
  dVarPtdq := nil;
  if dVarPtdCond_q <> nil then begin
    for k := 0 to RateCount - 1 do begin
      Free2(TPMChar(dVarPtdCond_q[k]));
      dVarPtdCond_q[k] := nil;
    end;
  end;
  FreeMem(dVarPtdCond_q);
  dVarPtdCond_q := nil;

  //Fit States
  FreeMem(FitP);
  FitP := nil;

end;



//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
function TQUB_Mac_Model.Do_InitcPars :Integer;
begin

  Result := inherited Do_InitcPars;

  //ChannelCount
  if prOptParams.EstChannelCount <> 0 then begin

    if ChannelCountIdx >= 0 then begin

      //cPars[2*RateCount + ExtraKinParCount + ChannelCountIdx] := Ln(ChannelCount);
      //cPars[2*RateCount + ExtraKinParCount + ChannelCountIdx] := Ln(prModel.ChannelCount);
      cPars[2*RateCount + ExtraKinParCount + ChannelCountIdx] := Ln(InitChannelCount);

    end;

  end;

  //AmpOpen
  if prOptParams.EstAmpOpen <> 0 then begin
    cPars[2*RateCount + ExtraKinParCount + ChannelCountCount] := xAmpOpen;
  end;

  //StdOpen
  if prOptParams.EstVarOpen <> 0 then begin
    cPars[2*RateCount + ExtraKinParCount + ChannelCountCount + AmpOpenCount] := Ln(xVarOpen);
  end;

end;

procedure TQUB_Mac_Model.prInitOptStuff;
begin

  inherited prInitOptStuff;

end;



//****************************************************************************//
//                                                                            //
//                         M A C   -    s p e c i f i c                       //
//                                                                            //
//****************************************************************************//
function TQUB_Mac_Model.Do_P0 :Integer;
var
  i :Integer;
begin

  Result := 0;

  if          prOptParams.EstInitProbs = INITPROBS_FIXED  then begin

    //The fixed probabilities from the model
    Result := inherited Do_P0;

  end else if prOptParams.EstInitProbs = INITPROBS_EQ     then begin

    //Calculate the equilibrium probabilities
    Result := Do_Peq;
    for i := 0 to StateCount - 1 do begin
      P0[i] := Peq[i];
    end;

  end else if prOptParams.EstInitProbs = INITPROBS_CONDEQ then begin

    //Calculate the equilibrium probabilities for the Conditioning experimental conditions
    Result := Do_Cond_Peq;
    for i := 0 to StateCount - 1 do begin
      P0[i] := Cond_Peq[i];
    end;

  end;

end;

function TQUB_Mac_Model.Do_dP0_dq :Integer;
var
  qq, k, ec :Integer;
begin

  Result := 0;

  if          prOptParams.EstInitProbs = INITPROBS_FIXED  then begin
  //Fixed P0 - derivatives are zero

    Result := inherited Do_dP0_dq;

  end else if prOptParams.EstInitProbs = INITPROBS_EQ     then begin
  //Equilibrium P0 - calculate

    Result := Do_dPeq_dq;
    for ec := 0 to ExpCondCount - 1 do begin
      for qq := 0 to RateCount - 1 do begin
        for k := 0 to StateCount - 1 do begin
          TPMDouble(dP0dq[ec])[qq, k] := TPMDouble(dPeqdq[ec])[qq, k];
        end;
      end;
    end;

  end else if prOptParams.EstInitProbs = INITPROBS_CONDEQ then begin
  //Conditioning P0 - derivatives are zero

    Result := inherited Do_dP0_dq;

  end;

end;

function TQUB_Mac_Model.Do_dP0_dCond_q :Integer;
var
  qq, k :Integer;
begin

  Result := 0;

  if          prOptParams.EstInitProbs = INITPROBS_FIXED  then begin
  //Fixed P0 - derivatives are zero

    Result := inherited Do_dP0_dCond_q;

  end else if prOptParams.EstInitProbs = INITPROBS_EQ     then begin
  //Equilibrium P0 - derivatives are zero

    Result := inherited Do_dP0_dCond_q;

  end else if prOptParams.EstInitProbs = INITPROBS_CONDEQ then begin
  //Conditioning P0

    Result := Do_dCond_Peq_dCond_q;
    for qq := 0 to RateCount - 1 do begin
      for k := 0 to StateCount - 1 do begin
        dP0dCond_q[qq, k] := dCond_PeqdCond_q[qq, k];
      end;
    end;

  end;

end;


procedure TQUB_Mac_Model.RestartPt;
var
  i :Integer;
begin
  for i := 0 to StateCount - 1 do begin
    Pt[i] := P0[i];
  end;
end;


//Current mean and variance
function TQUB_Mac_Model.Do_Mean_t(var mt :Double) :Integer;
var
  m :Double;
begin

  Result := 0;

  if HasFitP then begin

    //mt = Pt.FitP
    MulVecVecS(Pt, FitP, mt, StateCount);

  end else begin

    //mt = bAmp + ChannelCount*(Pt.xAmp)
    MulVecVecS(Pt, xAmp, m, StateCount);
    mt := bAmp + ChannelCount * m;

  end;

end;

function TQUB_Mac_Model.Do_Mean_t(var mt :Double; dV, ILeak :Double) :Integer;
var
  m :Double;
begin

  Result := 0;

  if HasFitP then begin

    //mt = Pt.FitP
    MulVecVecS(Pt, FitP, mt, StateCount);

  end else begin

    //mt = bAmp + ChannelCount*(Pt.xAmp)
    MulVecVecS(Pt, xAmp, m, StateCount);
    mt := bAmp + ILeak + dV * ChannelCount * m;

  end;

end;

function TQUB_Mac_Model.Do_dMean_t_dq(var dmt :Double; qq :Integer; ec :Integer) :Integer;
begin

  Result := 0;

  if HasFitP then begin

    //mt = Pt.FitP
    //dmt = dPt.FitP
    MulVecVecS(TPMDouble(dPtdq[ec])[qq], FitP, dmt, StateCount);

  end else begin

    //mt = bAmp + ChannelCount*(Pt.xAmp)
    //dmt = ChannelCount*(dPtdq.xAmp)

    MulVecVecS(TPMDouble(dPtdq[ec])[qq], xAmp, dmt, StateCount);
    dmt := ChannelCount * dmt;

  end;

end;

function TQUB_Mac_Model.Do_dMean_t_dCond_q(var dmt :Double; qq :Integer) :Integer;
begin

  Result := 0;

  if HasFitP then begin

    //mt = Pt.FitP
    //dmt = dPt.FitP
    MulVecVecS(dPtdCond_q[qq], FitP, dmt, StateCount);

  end else begin

    //mt = bAmp + ChannelCount*(Pt.xAmp)
    //dmt = ChannelCount*(dPtdCond_q.xAmp)

    MulVecVecS(dPtdCond_q[qq], xAmp, dmt, StateCount);
    dmt := ChannelCount * dmt;

  end;

end;

function TQUB_Mac_Model.Do_dMean_t_dN(var dmt :Double) :Integer;
begin

  Result := 0;

  if HasFitP then begin

    //mt = Pt.FitP
    //dmt = 0
    dmt := 0;;

  end else begin

    //mt = bAmp + ChannelCount*(Pt.xAmp)
    //dmt = Pt.xAmp

    MulVecVecS(Pt, xAmp, dmt, StateCount);

  end;

end;

function TQUB_Mac_Model.Do_dMean_t_dA(var dmt :Double) :Integer;
var
  i :Integer;
begin

  Result := 0;

  if HasFitP then begin

    //mt = Pt.FitP
    //dmt = 0
    dmt := 0;;

  end else begin

    //mt = bAmp + ChannelCount*(Pt.xAmp)
    //dmt = ChannelCount*(Pt.dxAmp_dA)

    //Vec1 := dxAmp_dA
    for i := 0 to StateCount - 1 do begin
      if Cls[i] = 1 then begin
        Vec1[i] := 1;
      end else begin
        Vec1[i] := 0;
      end;
    end;

    MulVecVecS(Pt, Vec1, dmt, StateCount);

    dmt := ChannelCount * dmt;

  end;

end;

function TQUB_Mac_Model.Do_dMean_t_dV(var dmt :Double) :Integer;
begin

  Result := 0;

  dmt := 0;

end;


function TQUB_Mac_Model.Do_Var_t(var Vt :Double) :Integer;
var
  V1, V2 :Double;
begin

  Result := 0;

  if HasFitP then begin

    //I'm not sure yet what Vt should be
    Vt := 0;

  end else begin

    //Vt = [bVar + ChannelCount*(xAmp.VarPt.xAmp + Pt.xVar)] / AverageCount

    //Vec1 = xAmp*VarPt
    MulVecMatVec(xAmp, VarPt, Vec1, StateCount);

    //V1 = Vec1*xAmp
    MulVecVecS(Vec1, xAmp, V1, StateCount);

    //V2 = Pt*xVar
    MulVecVecS(Pt, xVar, V2, StateCount);

    Vt := (bVar + ChannelCount * (V1 + V2)) / AverageCount;

  end;

end;

function TQUB_Mac_Model.Do_Var_t(var Vt :Double; dV :Double) :Integer;
var
  V1, V2 :Double;
begin

  Result := 0;

  if HasFitP then begin

    //I'm not sure yet what Vt should be
    Vt := 0;

  end else begin

    //Vt = [bVar + ChannelCount*(xAmp.VarPt.xAmp + Pt.xVar)] / AverageCount

    //Vec1 = xAmp*VarPt
    MulVecMatVec(xAmp, VarPt, Vec1, StateCount);

    //V1 = Vec1*xAmp
    MulVecVecS(Vec1, xAmp, V1, StateCount);

    //V2 = Pt*xVar
    MulVecVecS(Pt, xVar, V2, StateCount);

    Vt := (bVar + dV * ChannelCount * (V1 + V2)) / AverageCount;

  end;

end;

function TQUB_Mac_Model.Do_dVar_t_dq(var dVt :Double; qq :Integer; ec :Integer) :Integer;
var
  V1, V2 :Double;
begin

  Result := 0;

  if HasFitP then begin

    //I'm not sure yet what dVt should be
    dVt := 0;

  end else begin

    //dVtdQ = [ChannelCount * (xAmp.dVarPtdq.xAmp + dPtdq.xVar)] / AverageCount

    //V1 = xAmp.dVarPtdq.xAmp
    MulVecMatVec(xAmp, dVarPtdq[ec, qq], Vec1, StateCount);
    MulVecVecS(Vec1, xAmp, V1, StateCount);

    //V2 = dPtdQ.xVar
    MulVecVecS(TPMDouble(dPtdq[ec])[qq], xVar, V2, StateCount);

    dVt := (ChannelCount * (V1 + V2)) / AverageCount;

  end;

end;

function TQUB_Mac_Model.Do_dVar_t_dCond_q(var dVt :Double; qq :Integer) :Integer;
var
  V1, V2 :Double;
begin

  Result := 0;

  if HasFitP then begin

    //I'm not sure yet what dVt should be
    dVt := 0;

  end else begin

    //dVtdQ = [ChannelCount * (xAmp.dVarPtdCond_q.xAmp + dPtdCond_q.xVar)] / AverageCount

    //V1 = xAmp.dVarPtdCond_q.xAmp
    MulVecMatVec(xAmp, dVarPtdCond_q[qq], Vec1, StateCount);
    MulVecVecS(Vec1, xAmp, V1, StateCount);

    //V2 = dPtdCond_q.xVar
    MulVecVecS(dPtdCond_q[qq], xVar, V2, StateCount);

    dVt := (ChannelCount * (V1 + V2)) / AverageCount;

  end;

end;

function TQUB_Mac_Model.Do_dVar_t_dN(var dVt :Double) :Integer;
var
  V1, V2 :Double;
begin

  Result := 0;

  if HasFitP then begin

    //I'm not sure yet what dVt should be
    dVt := 0;

  end else begin

    //dVtdN = [xAmp.VarPt.xAmp + Pt.xVar] / AverageCount

    //V1 = xAmp.VarPt.xAmp
    MulVecMatVec(xAmp, VarPt, Vec1, StateCount);
    MulVecVecS(Vec1, xAmp, V1, StateCount);

    //V2 = Pt.xVar
    MulVecVecS(Pt, xVar, V2, StateCount);

    dVt := (V1 + V2) / AverageCount;

  end;

end;

function TQUB_Mac_Model.Do_dVar_t_dA(var dVt :Double) :Integer;
var
  V1, V2 :Double;
  i      :Integer;
begin

  Result := 0;

  if HasFitP then begin

    //I'm not sure yet what dVt should be
    dVt := 0;

  end else begin

    //dVtdA = [ChannelCount*(dxAmp_dA.VarPt.xAmp + xAmp.VarPt.dxAmp_dA)] / AverageCount

    //Vec1 := dxAmp_dA
    for i := 0 to StateCount - 1 do begin
      if Cls[i] = 1 then begin
        Vec1[i] := 1;
      end else begin
        Vec1[i] := 0;
      end;
    end;

    //V1 = dxAmp_dA.VarPt.xAmp
    MulVecMatVec(Vec1, VarPt, Vec2, StateCount);
    MulVecVecS(Vec2, xAmp, V1, StateCount);

    //V2 = xAmp.VarPt.dxAmp_dA
    MulVecMatVec(xAmp, VarPt, Vec2, StateCount);
    MulVecVecS(Vec2, Vec1, V2, StateCount);

    dVt := (ChannelCount * (V1 + V2)) / AverageCount;

  end;

end;

function TQUB_Mac_Model.Do_dVar_t_dV(var dVt :Double) :Integer;
var
  V1     :Double;
  i      :Integer;
begin

  Result := 0;

  if HasFitP then begin

    //I'm not sure yet what dVt should be
    dVt := 0;

  end else begin

    //dVtdV = [ChannelCount*(Pt.dxVar_dV)] / AverageCount

    //Vec1 := dxAmp_dA
    for i := 0 to StateCount - 1 do begin
      if Cls[i] = 1 then begin
        Vec1[i] := 1;
      end else begin
        Vec1[i] := 0;
      end;
    end;

    //V1 = Pt.dxVar_dV
    MulVecVecS(Pt, Vec1, V1, StateCount);

    dVt := (ChannelCount * V1) / AverageCount;

  end;

end;


function TQUB_Mac_Model.Do_VarPt :Integer;
var
  i, j :Integer;
begin

  Result := 0;

  for i := 0 to StateCount - 1 do begin
    for j := 0 to StateCount - 1 do begin
      if i = j then begin
      //Diagonal
        VarPt[i, j] :=   Pt[i] * (1 - Pt[i]);
      end else begin
      //Off-diagonal
        VarPt[i, j] := - Pt[i] * Pt[j];
      end;
    end;
  end;

end;

function TQUB_Mac_Model.Do_dVarPt_dq :Integer;
var
  m, n, qq, ec :Integer;
begin

  Result := 0;

  for ec := 0 to ExpCondCount -1 do begin

    for qq := 0 to RateCount - 1 do begin

      //Do the diagonals
      for m := 0 to StateCount - 1 do begin
        TPMDouble(dVarPtdq[ec, qq])[m, m] := (1 - 2*Pt[m]) * TPMDouble(dPtdq[ec])[qq, m];
      end;//m

      //Now the off-diagonals and symetricals
      for m := 0 to StateCount - 1 do begin
        for n := m + 1 to StateCount - 1 do begin
          TPMDouble(dVarPtdq[ec, qq])[m, n] := - (TPMDouble(dPtdq[ec])[qq, m] * Pt[n]) - (Pt[m] * TPMDouble(dPtdq[ec])[qq, n]);
          TPMDouble(dVarPtdq[ec, qq])[n, m] := TPMDouble(dVarPtdq[ec, qq])[m, n];
        end;//n
      end;//m

    end;//qq

  end;//ec

end;

function TQUB_Mac_Model.Do_dVarPt_dCond_q :Integer;
var
  m, n, qq :Integer;
begin

  Result := 0;

  for qq := 0 to RateCount - 1 do begin

    //Do the diagonals
    for m := 0 to StateCount - 1 do begin
      TPMDouble(dVarPtdCond_q[qq])[m, m] := (1 - 2*Pt[m]) * dPtdCond_q[qq, m];
    end;//m

    //Now the off-diagonals and symetricals
    for m := 0 to StateCount - 1 do begin
      for n := m + 1 to StateCount - 1 do begin
        TPMDouble(dVarPtdCond_q[qq])[m, n] := - (dPtdCond_q[qq, m] * Pt[n]) - (Pt[m] * dPtdCond_q[qq, n]);
        TPMDouble(dVarPtdCond_q[qq])[n, m] := TPMDouble(dVarPtdCond_q[qq])[m, n];
      end;//n
    end;//m

  end;

end;



//****************************************************************************//
//                                                                            //
//                 L o g  -  L i k e l i h o o d     G r a d s                //
//                                                                            //
//****************************************************************************//
function TQUB_Mac_Model.Do_dLLdfPar :Integer;
var
  dNdcPar :Double;
  dAdcPar :Double;
  dVdcPar :Double;
  k       :Integer;

begin


  //The inherited, for Rates
  Result := inherited Do_dLLdfPar;


  //ChannelCount
  if prOptParams.EstChannelCount = 1 then begin

    if ChannelCountIdx >= 0 then begin

      //ChannelCount:
      dNdcPar := Exp(cPars[2*RateCount + ExtraKinParCount + ChannelCountIdx]);

      //dLLdfPar
      for k := 0 to fParCount - 1 do begin

        dLLdfPar[k] := dLLdfPar[k] + dLLdN * dNdcPar * {dcPardfPar}f2cA[2*RateCount + ExtraKinParCount + ChannelCountIdx, k];

      end;

      dLLdcPar[2*RateCount + ExtraKinParCount + ChannelCountIdx] := dLLdN;

    end;

  end;


  //AmpOpen
  if prOptParams.EstAmpOpen = 1 then begin

    //AmpOpen:
    dAdcPar := 1;

    //dLLdfPar
    for k := 0 to fParCount - 1 do begin

      dLLdfPar[k] := dLLdfPar[k] + dLLdA * dAdcPar * {dcPardfPar}f2cA[2*RateCount + ExtraKinParCount + ChannelCountCount, k];

    end;

    dLLdcPar[2*RateCount + ExtraKinParCount + ChannelCountCount] := dLLdA;

  end;


  //StdOpen
  if prOptParams.EstVarOpen = 1 then begin

    //StdOpen:
    dVdcPar := Exp(cPars[2*RateCount + ExtraKinParCount + ChannelCountCount + AmpOpenCount]);

    //dLLdfPar
    for k := 0 to fParCount - 1 do begin

      dLLdfPar[k] := dLLdfPar[k] + dLLdV * dVdcPar * {dcPardfPar}f2cA[2*RateCount + ExtraKinParCount + ChannelCountCount + AmpOpenCount, k];

    end;

    dLLdcPar[2*RateCount + ExtraKinParCount + ChannelCountCount + AmpOpenCount] := dLLdV;

  end;

end;



//****************************************************************************//
//                                                                            //
//                   P a r a m e t e r     V a r i a n c e                    //
//                                                                            //
//****************************************************************************//
function TQUB_Mac_Model.Do_Varf2Varc :Integer;
var
  i, j :Integer;
begin

  //The inherited, for Rates
  Result := inherited Do_Varf2Varc;

  if prOptParams.EstChannelCount = 1 then begin
  //The new one, for ChannelCount

    //I'm not sure if this is right
    if ChannelCountIdx >= 0 then begin

      //var(N) ~= Sum i (Sqr(dN/dzi) * var(zi))
      //dN/dz = dN/dc0 * dc0/dz

      for i := 2*RateCount + ExtraKinParCount to 2*RateCount + ExtraKinParCount + ChannelCountCount - 1 do begin  //do all NC's ?!

        cParsVar[i] := 0;

        for j := 0 to fParCount - 1 do begin

          //Parametrization is ln(N):
          cParsVar[i] := cParsVar[i] + Sqr(Exp(cPars[i])*f2cA[i, j]) * Abs(fParsVar[j, j]);

          //If parametrization is N:
          //cParsVar[i] := cParsVar[i] + Sqr(              f2cA[i, j]) * Abs(fParsVar[j, j]);

        end;

      end;

    end;

  end;

  if prOptParams.EstAmpOpen = 1 then begin
  //The newest one, for AmpOpen

    //var(A) ~= Sum i (Sqr(dA/dzi) * var(zi))
    //dA/dz = dA/dc0 * dc0/dz

    for i := (2*RateCount + ExtraKinParCount + ChannelCountCount) to (2*RateCount + ExtraKinParCount + ChannelCountCount) do begin

      cParsVar[i] := 0;

      for j := 0 to fParCount - 1 do begin

        //Parametrization is A:
        cParsVar[i] := cParsVar[i] + Sqr(f2cA[i, j]) * Abs(fParsVar[j, j]);

      end;

    end;

  end;

  if prOptParams.EstVarOpen = 1 then begin
  //The very newest one, for StdOpen

    //var(S) ~= Sum i (Sqr(dS/dzi) * var(zi))
    //dS/dz = dS/dc0 * dc0/dz

    for i := (2*RateCount + ExtraKinParCount + ChannelCountCount + AmpOpenCount) to (2*RateCount + ExtraKinParCount + ChannelCountCount + AmpOpenCount) do begin

      cParsVar[i] := 0;

      for j := 0 to fParCount - 1 do begin

        //Parametrization is ln(V):
        cParsVar[i] := cParsVar[i] + Sqr(Exp(cPars[i])*f2cA[i, j]) * Abs(fParsVar[j, j]);

      end;

    end;

  end;

end;


function TQUB_Mac_Model.prGet_cParCount;
begin

  Result := inherited prGet_cParCount;

  if prOptParams.EstChannelCount <> 0 then begin
    Result := Result + ChannelCountCount;
  end;

  if prOptParams.EstAmpOpen <> 0 then begin
    Result := Result + 1{currently, only class 1};
  end;

  if prOptParams.EstVarOpen <> 0 then begin
    Result := Result + 1{currently, only class 1};
  end;

end;

function TQUB_Mac_Model.prGet_ChannelCount :Integer;
var
  Par :TQUB_ExpCond_PParameter;
begin

  //Default value
  Result := inherited prGet_ChannelCount;

  if prOptParams.EstChannelCountPerFile = 1 then begin

    Par    := ExpCondParams[0].ParameterByName['ChannelCount'];
    if Par <> nil then begin
      if Par.Value > 0 then begin
        try
          Result := Round(Par.Value);
        except
        end;
      end;
    end;

  end;

  //What is ChannelCount here?
  InitChannelCount := Result;

end;

function TQUB_Mac_Model.prGet_CCFixed(ExpCondParams :TQUB_ExpCond_ParametersList) :Boolean;
var
  Par :TQUB_ExpCond_PParameter;
begin

  //Default value
  Result := False;

  if prOptParams.EstChannelCountPerFile = 1 then begin

    Par    := ExpCondParams[0].ParameterByName['CCFixed'];
    if Par <> nil then begin
      if Par.Value > 0 then begin
        try
          Result := Boolean(Round(Par.Value));
        except
        end;
      end;
    end;

  end;

end;

function TQUB_Mac_Model.prGet_AverageCount :Integer;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Result := inherited prGet_AverageCount;
  Par    := ExpCondParams[0].ParameterByName['AverageCount'];
  if Par <> nil then begin
    if Par.Value > 0 then begin
      try
        Result := Round(Par.Value);
      except
      end;
    end;
  end;

end;

function TQUB_Mac_Model.prGet_FitWeight :Double;
var
  Par :TQUB_ExpCond_PParameter;
begin

  Result := inherited prGet_FitWeight;
  Par    := ExpCondParams[0].ParameterByName['FitWeight'];
  if Par <> nil then begin
    if Par.Value > 0 then begin
      try
        Result := Par.Value;
      except
      end;
    end;
  end;

end;

function TQUB_Mac_Model.prGet_VRev :Double;
begin

  //Default value
  Result := prOptParams.OptVRev;

  if prOptParams.OptVRevFromModel then begin
    Result := prModel.VRev;
  end;

end;

function TQUB_Mac_Model.prGet_VLeak :Double;
var
  Par :TQUB_ExpCond_PParameter;
begin

  //Default value
  Result := prOptParams.OptVLeak;
  if not prOptParams.OptVLeakFromTable then Exit;

  Par    := ExpCondParams[0].ParameterByName['VLeak'];
  if Par <> nil then begin
    //if Par.Value > 0 then begin
      try
        Result := Round(Par.Value);
      except
        Result := prOptParams.OptVLeak;
      end;
    //end;
  end;

end;

function TQUB_Mac_Model.prGet_GLeak :Double;
var
  Par :TQUB_ExpCond_PParameter;
begin

  //Default value
  Result := prOptParams.OptGLeak;
  if not prOptParams.OptGLeakFromTable then Exit;

  Par    := ExpCondParams[0].ParameterByName['GLeak'];
  if Par <> nil then begin
    if Par.Value > 0 then begin
      try
        Result := Round(Par.Value);
      except
        Result := prOptParams.OptGLeak;
      end;
    end;
  end;

end;

// Result is true if one or more models have confluent eigenvalues
function TQUB_Mac_Model.ResetLLandGR :Boolean;
var
  ec, k : Integer;
begin

  Result := inherited ResetLLandGR;

  //dLLdq[ec]
  for ec := 0 to ExpCondCount - 1 do begin
    if Confluent[ec] <> 0 then Result := True; // Confluent
    for k := 0 to RateCount - 1 do begin
      dLLdq[ec, k] := 0;
    end;//k
  end;//ec

  //dLLdCond_q
  for k := 0 to RateCount - 1 do begin
    dLLdCond_q[k]  := 0;
  end;//k

  //dLLdN
  dLLdN            := 0;//dLLdN should not be an array, because it's local (per Model).
  //dLLdA
  dLLdA            := 0;
  //dLLdV
  dLLdV            := 0;

end;




//******************************************************************************
end.
