{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// constants and such for QUB_Task_MIL.  I'm pretty sure some of this can be cleaned out

unit QUB_Mil_Global;


interface


uses


AllocTypes,

QUB_QVO_QVD_Global,
QUB_Global,
QUB_Thread,
QUB_QFS_QIF_Segments;


type


{$Z4}
TQUB_MIL_Mode              = (MIL_MODE_OPT, MIL_MODE_EVAL, MIL_MODE_HIST);
TQUB_MIL_Seg_Mode          = (MIL_MODE_SEG_EACH, MIL_MODE_SEG_ALL, MIL_MODE_SEG_JOINED);
TQUB_MIL_Method            = (OPT_METHOD_MIL, OPT_METHOD_MPL, OPT_METHOD_AMP);


function MILModeStringOf(x :TQUB_MIL_Mode): String;
function MILModeOfString(const x :String): TQUB_MIL_Mode;
function MILSegModeStringOf(x :TQUB_MIL_Seg_Mode): String;
function MILSegModeOfString(const x :String): TQUB_MIL_Seg_Mode;


type


TQUB_Thread_Mil_Chunk      = record

  Chunk                    :TQUB_Thread_Chunk;                                  //Sel is in Data Points

  Sampling                 :Single;//Seconds
  Scaling                  :Single;

  Data                     :TPAInteger;                                         //only used by MPL

  SegClass                 :Integer;                                            //the histogram class

  IdlIdx                   :Integer;
  IdlSel                   :TQUB_FirstLast;                                     //Is in Dwells

  ClassDwells              :TQUB_QFS_QIF_PADwellClass;                          //Class :Integer; Dwell(seconds) :Single

  Voltage                  :Single;                                             //the conditions per chunk
  Conc                     :Single;
  DeadTime                 :Single;
  NChannel                 :Integer;

  LL                       :Single;                                             //Output: the LL of this segment, after the last iteration

end;
TQUB_Thread_Mil_PChunk     = ^TQUB_Thread_Mil_Chunk;


TQUB_Mil_Params            = record

  LibName                  :PChar;
  LibNameSize              :Integer;

  Model                    :PChar;
  ModelSize                :Integer;
  ModelChannelCount        :Integer;

  HistBinCount             :Integer;
  HistPerIter              :Boolean;

  LLConv                   :Double;
  GradConv                 :Double;

  MaxIter                  :Integer;
  SearchLimit              :Double;
  Restarts                 :Integer;
  MaxStep                  :Double;

  QuietOutput              :Boolean;
  WordReport               :Boolean;

  Method                   :TQUB_MIL_Method;

  Mode                     :TQUB_MIL_Mode;
  SegMode                  :TQUB_MIL_Seg_Mode;
  JoinSegments             :Boolean;
end;
TQUB_Mil_PParams           = ^TQUB_Mil_Params;


const


DEF_MILMODE                   = MIL_MODE_OPT;
DEF_SEGMODE                   = MIL_MODE_SEG_ALL;
DEF_MILJOINSEGMENTS           = 0;
DEF_MILSHOWITER               = 1;
DEF_GROUPVITERBI              = 0;

DEF_MILHISTBINCOUNT        = 50;
MIN_MILHISTBINCOUNT        = 2;
MAX_MILHISTBINCOUNT        = 500;

DEF_MILHISTPERITER         = 1;

DEF_MILLLCONV             :Double = 1e-4;
MIN_MILLLCONV             :Double = 1e-10;
MAX_MILLLCONV             :Double = 1.0;

DEF_MILGRADCONV           :Double = 1e-2;
MIN_MILGRADCONV           :Double = 1e-10;
MAX_MILGRADCONV           :Double = 1.0;

DEF_MILMAXITER             = 100;
MIN_MILMAXITER             = 1;
MAX_MILMAXITER             = 10000;

DEF_MILSEARCHLIMIT        :Double = 1000.0;
MIN_MILSEARCHLIMIT        :Double = 10.0;
MAX_MILSEARCHLIMIT        :Double = 1e6;

DEF_MILRESTARTS            = 0;
MIN_MILRESTARTS            = 0;
MAX_MILRESTARTS            = 100;

DEF_MILMAXSTEP            :Double = 1.0;
MIN_MILMAXSTEP            :Double = 1e-6;
MAX_MILMAXSTEP            :Double = 1e6;

MIL_MAX_NCHANNEL           = 48;


implementation


//******************************************************************************
function MILModeStringOf(x :TQUB_MIL_Mode): String;
begin
  if x = MIL_MODE_OPT then begin
    Result := 'optimize';
  end else if x = MIL_MODE_EVAL then begin
    Result := 'evaluate';
  end else begin
    Result := 'histograms';
  end;
end;

function MILModeOfString(const x :String): TQUB_MIL_Mode;
begin
  if x = 'optimize' then begin
    Result := MIL_MODE_OPT;
  end else if x = 'evaluate' then begin
    Result := MIL_MODE_EVAL;
  end else{if x = 'histograms' then}begin
    Result := MIL_MODE_HIST;
  end;
end;

function MILSegModeStringOf(x :TQUB_MIL_Seg_Mode): String;
begin
  if x = MIL_MODE_SEG_EACH then begin
    Result := 'separately';
  end else begin
    Result := 'together';
  end;
end;

function MILSegModeOfString(const x :String): TQUB_MIL_Seg_Mode;
begin
  if x = 'separately' then begin
    Result := MIL_MODE_SEG_EACH;
  end else{if x = 'together' then}begin
    Result := MIL_MODE_SEG_ALL;
  end;
end;


//******************************************************************************
end.
