unit QUB_MultiFit;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls,

  Alloc, AllocTypes,
  QUB_Global,
  QUB_Node, QUB_Profiles,
  QUB_QFS_QDF, QUB_QVO_QVD, QUB_QVO_QVD_Global, QUB_QVO_QVD_Overlap,
  QUBFits, QUBFits_GUI, QUBFits_Dialog;

type

TQUB_MultiFitForm = class(TForm)
  panLists: TPanel;
  panFitsParent: TPanel;
  lstResults: TListBox;
  Label1: TLabel;
  Label2: TLabel;
  panSelections: TPanel;
  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure lstResultsDblClick(Sender: TObject);

private
  panFits   :TQUB_DialACurve;
  prFits    :TQUB_Fits;
  prData    :TQUB_FitData;
  lstSels   :TQUB_QVO_QVD_SelectionList;
  lstSelsActive :TQUB_QVO_QVD_ActiveSelList;
  prSrcSegs :Array of INode;
  prResult  :INode;
  prResultSegs :Array of INode;
  prFitsParentStored :TWinControl;
  prSelIndex :Integer;
  prDataNode :INode;
  prStopped  :Boolean;

  procedure prOnItemChanged(Sender :TObject);
  procedure prOnFitAll(Sender :TObject);
  procedure prOnStopFitAll(Sender :TObject);
  procedure prOnKeepIt(Sender :TObject);

public


end;


implementation

uses

QUB_Main,
QUB_Task,
QUB_FitCurves,
QUB_Viewport_Results,
QUB_GUIInterface;


{$R *.DFM}


procedure TQUB_MultiFitForm.FormCreate(Sender: TObject);
var
  R        :INode;
begin
  R         := ProfileMgr['MultiFit'].Active['Rect'];
  if R.DataCount < 4 then begin
    R.SetupNumData(NODE_DATATYPE_LONGINT, 4, 1);
    R.Data[0] := Left;
    R.Data[1] := Top;
    R.Data[2] := Width;
    R.Data[3] := Height;
  end;

  SetBounds(R.Data[0], R.Data[1], R.Data[2], R.Data[3]);
end;

procedure TQUB_MultiFitForm.FormShow(Sender: TObject);
var
  i         :Integer;
  DS, Seg   :INode;
begin
  if panFits = nil then begin
    panFits := FitDlog.panFitter;
    prFits  := panFits.Fits;
    prData  := prFits.Data;

    lstSels := TQUB_QVO_QVD_SelectionList.Create(Self);
    lstSels.Visible := False;
    lstSels.Parent := Self;
    lstSelsActive := TQUB_QVO_QVD_ActiveSelList.Create(Self);
    lstSelsActive.Parent := panSelections;
    lstSelsActive.Align := alClient;
  end;

  prFitsParentStored := panFits.Parent;
  panFits.Parent     := panFitsParent;
  panFits.FitterPanel.OnKeepIt := prOnKeepIt;
  panFits.FitterPanel.OnFitAll := prOnFitAll;

  // populate lstSels
  if ViewportData.CurrData <> nil then begin
    DS  := ViewportData.CurrData.GetChosenData;
    Seg := DS.Find('Segment');
    i   := 0;
    while not Seg.IsNull do begin
      SetLength(prSrcSegs, i+1);
      prSrcSegs[i] := Seg;
      lstSels.AddSelection(Seg.Data[0], Seg.Data[1], '');
      Seg := Seg.SiblingSameName;
      Inc(i);
    end;
  end;

  lstSelsActive.Data := ViewportData.CurrData;
  lstSelsActive.List := lstSels;
  lstSels.Events.AddEvent('ItemChanged', prOnItemChanged);

  prResult := TNode.Create('MultiFit Results');
  prResult['ResultType'].SetData('MultiFit');
  prResult['Hold'].SetData(1);

  panFits.SetState(ProfileMgr['MultiFitter'].Active);
  prData.DisplayStyle := QUBFITS_DISPLAY_DEFAULT;

  if lstSels.Items.Count > 0 then lstSels.Choose(0);

  if ViewportData.CurrData <> nil then with ViewportData.CurrData do begin
    prData.YName := QF_Data_SafeVarName(DataChannelName[ActiveChannel]);
    prResult['DataSet']['FileName'].SetData(FileName);
  end;
end;

procedure TQUB_MultiFitForm.FormHide(Sender: TObject);
var
  R         :INode;
begin
  R         := ProfileMgr['MultiFit'].Active['Rect'];
  R.Data[0] := Left;
  R.Data[1] := Top;
  R.Data[2] := Width;
  R.Data[3] := Height;

  ProfileMgr['MultiFitter'].Active := panFits.GetState;
  prData.UseData(nil);

  lstResults.Items.Clear;
  lstSelsActive.List := nil;
  lstSels.Events.DelEvent('ItemChanged', prOnItemChanged);
  lstSels.Items.Clear;
  panFits.FitterPanel.OnKeepIt := nil;
  panFits.FitterPanel.OnFitAll := nil;
  panFits.Parent               := prFitsParentStored;

  if Length(prResultSegs) > 0 then ViewportResults.AddResults(prResult, False);

  prResult := TNode.CreateNull;
  SetLength(prResultSegs, 0);
  prDataNode                   := TNode.CreateNull;
end;

procedure TQUB_MultiFitForm.lstResultsDblClick(Sender: TObject);
var
  Res     :INode;
begin
  if lstResults.ItemIndex >= 0 then begin
    Res   := prResultSegs[lstResults.ItemIndex];
    lstSels.Choose(Res['SelIndex'].Data[0]);
    panFits.SetState(Res['DialACurve']);
  end;
end;

procedure TQUB_MultiFitForm.prOnItemChanged(Sender :TObject);
var
  i, j         :Integer;
  x, s         :Double;
  DS           :TQUB_FitDataSet;
  Seg          :INode;
  Vr           :INode;
  Sel          :TQUB_FirstLast;
  DNode        :INode;
begin
  if lstSels.ItemIndex < 0 then begin
    lstSels.Choose(prSelIndex);
    Exit;
  end;

  prSelIndex  := lstSels.ItemIndex;
  Sel         := lstSels[lstSels.ItemIndex].Sel;
  x           := ViewportData.CurrData.WhichTimeForPoint(Sel.First, True);
  s           := ViewportData.CurrData.Sampling;

  // qubtree data rq/storage structure
  DNode       := TNode.Create('DataSet');
  DNode['ProcessData'].SetData('AsDisplayed');
  Seg         := DNode['Segment'];
  Seg.SetupNumData(NODE_DATATYPE_LONGINT, 2, 1);
  Seg.Data[0] := Sel.First;
  Seg.Data[1] := Sel.Last;

  // qubfits data descriptor
  DS          := TQUB_FitDataSet.Create(Sel.Last - Sel.First + 1);
  Vr          := Seg['X'];
  Vr.SetupNumData(NODE_DATATYPE_DOUBLE, DS.N, 1);
  for j := 0 to DS.N - 1 do begin
    Vr.Data[j] := x + j*s;
  end;
  Ds.AddVar(Vr.Name, TPADouble(Vr.DataPtr));

  for i := 0 to ViewportData.CurrData.DataObject.ADChannelCount - 1 do begin
    DNode['ActiveChannel'].SetData(i);
    ViewportData.CurrData.FillDataSegment(Seg);
    Vr := Seg['Channel'];
    Vr.Name := QF_Data_SafeVarName(ViewportData.CurrData.DataChannelName[i]);
    DS.AddVar(Vr.Name, TPADouble(Vr.DataPtr));
  end;

  prData.UseData(DS);
  DS.Free;
  prDataNode := DNode; // overwrite may ref-count-free old prDataNode, wait until after UseData(new)
end;

procedure TQUB_MultiFitForm.prOnFitAll(Sender :TObject);
var
  State   :INode;
  i       :Integer;
begin
  prStopped := False;
  prFits.Events.AddEvent('OnClickStop', prOnStopFitAll);
  State   := panFits.GetState;
  for i := 0 to lstSels.Items.Count - 1 do begin
    if lstSels[i].Active then begin
      lstSels.Choose(i);
      if i > 0 then panFits.SetState(State); // same init params each sel
      panFits.FitterPanel.OnFit(panFits.FitterPanel);
      if prStopped then Break;
      prOnKeepIt(panFits.FitterPanel);
    end;
  end;
  prFits.Events.DelEvent('OnClickStop', prOnStopFitAll);
end;

procedure TQUB_MultiFitForm.prOnStopFitAll(Sender :TObject);
begin
  prStopped := True;
end;

procedure TQUB_MultiFitForm.prOnKeepIt(Sender :TObject);
var
  Replacing :Integer;
  Res, Prior:INode;
  i         :Integer;
  x         :Double;
  Param     :INode;
  Lbl       :String;
  FitIter   :TQUB_FitCurveIter;
  FitData   :TPAInteger;
  Sel       :TQUB_FirstLast;
  Scaling   :Double;
  Offset    :Double;
begin
  // Find stale result to replace it
  Replacing := -1;
  for i := 0 to Length(prResultSegs) - 1 do begin
    if prResultSegs[i]['SelIndex'].Data[0] = prSelIndex then begin
      Replacing := i;
      Res       := prResultSegs[i];
      Break;
    end;
  end;
  if Replacing >= 0 then begin
    Prior   := prResult['DataSet'].Child;
    while (not Prior.IsNull) and (not Prior.Sibling.Equals(Res)) do begin
      Prior := Prior.Sibling;
    end;
    Res       := prResult['DataSet'].InsertClone(Prior, prSrcSegs[prSelIndex], True);
    prResult['DataSet'].RemoveChild(Res.Sibling, Res);
  end else begin
    Res      := prResult['DataSet'].AppendClone(prSrcSegs[prSelIndex], True);
  end;

  Res['SelIndex'].SetData(prSelIndex);
  Res.AppendClone(panFits.GetState, True);
  Res['Residual'].SetData(Res['DialACurve']['Residual'].Data[0]);
  Res['N'].SetData(Res['DialACurve']['N'].Data[0]);
  Param := Res['DialACurve']['Params'].Child;
  while not Param.IsNull do begin
    if Param.DataCount >= 2 then begin
      Res[Param.Name].SetData(Param.Data[0]);
      Res[Param.Name+' +/-'].SetData(Param.Data[1]);
    end;
    Param := Param.Sibling;
  end;
  for i := 0 to panFits.Fits.Data.UseDataSet.VarCount - 1 do begin
    x := panFits.Fits.Data.UseDataSet.Vals[i][0];
    if i = panFits.Fits.Data.UseDataSet.Xindex
    then x := x + panFits.Fits.Data.UseDataSet.Xoff;
    Res[panFits.Fits.Data.UseDataSet.VarName[i]+'_start'].SetData(x);
  end;

  Lbl := Format('%d-%d: %s', [Integer(Res.Data[0]), Integer(Res.Data[1]), prFits.Curve.GetName()]);
  if Replacing >= 0 then begin
    prResultSegs[Replacing]     := Res;
    lstResults.Items[Replacing] := Lbl;
  end else begin
    i := Length(prResultSegs);
    SetLength(prResultSegs, i+1);
    prResultSegs[i] := Res;
    lstResults.Items.Add(Lbl);
  end;

  Sel           := lstSels[lstSels.ItemIndex].Sel;
  Sel.First     := Sel.First + Round((panFits.Fits.Data.UseDataSet.Xoff - panFits.Fits.Data.DataSet.XX[0]) / ViewportData.CurrData.DataObject.Sampling);
  Sel.Last      := Sel.First + panFits.Fits.Data.UseDataSet.N - 1;
  Scaling       := ViewportData.CurrData.DataObject.Scaling * ViewportData.CurrData.DataChannelScaling[ViewportData.CurrData.ActiveChannel];
  Offset        := - ViewportData.CurrData.DataChannelOffset[ViewportData.CurrData.ActiveChannel];
  FitIter       := nil;
  try
    FitIter       := ViewportData.CurrData.FitCurves.Find(Sel.First);
    if (0 <> FitIter.Done) or (FitIter.Curve.First <> Sel.First) or (FitIter.Curve.Last <> Sel.Last) then begin
      FitIter.Free;
      FitIter   := ViewportData.CurrData.FitCurves.Add(Sel.First, Sel.Last);
    end;
    FitData     := FitIter.Curve.Data;
    for i := 0 to Sel.Last - Sel.First + 1 - 1 do begin
      FitData[i] := Round((panFits.Fits.Data.FY[i] + Offset) * Scaling);
    end;
    FitIter.MarkChanged;
  finally
    FitIter.Free;
  end;

end;



//******************************************************************************
end.
