{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
{

    QUB Tree API header

}
unit QUB_Node;


interface


uses


SysUtils, Classes, Windows, Forms,
{$IFDEF DELPHI6_OR_HIGHER}
    Variants,
{$ENDIF}
AllocTypes,
QUB_Global;


// ************* first the stuff that's shared with QFS ************************
type


TQUB_QFS_Node_DataType = (NODE_DATATYPE_EMPTY,    NODE_DATATYPE_UNKNOWN,
                          NODE_DATATYPE_POINTER,
                          NODE_DATATYPE_STRING,   //DataSize = 1; for WideString is 2, but we don't check it -
                          NODE_DATATYPE_BYTE,     NODE_DATATYPE_SHORTINT,
													NODE_DATATYPE_WORD,     NODE_DATATYPE_SMALLINT,
                          NODE_DATATYPE_LONGWORD, NODE_DATATYPE_LONGINT,
                          NODE_DATATYPE_UINT64,   NODE_DATATYPE_INT64,
                          NODE_DATATYPE_SINGLE,   NODE_DATATYPE_DOUBLE,   NODE_DATATYPE_EXTENDED
												 );

TQUB_QFS_Node_Flags    = Word;

TQUB_QFS_NodeStream    = record

  Flags                :TQUB_QFS_Node_Flags;
  Flags2               :Byte;

  DataType             :TQUB_QFS_Node_DataType;
  DataSize             :LongInt;
  DataCount            :LongInt;
  DataStreamPos        :LongInt; //  0 should mean NONE

  ChildStreamPos       :LongInt; //       -,,-
  SiblingStreamPos     :LongInt; //       -,,-

  Reserved             :Array[0..6] of Byte;

  NameLen              :Byte;

end;
TQUB_QFS_PNodeStream   = ^TQUB_QFS_NodeStream;

TQUB_QFS_PNodeMem      = ^TQUB_QFS_NodeMem;
TQUB_QFS_NodeMem       = record

  NodeStream           :TQUB_QFS_NodeStream;

  RefCount             :Integer;

	Name                 :PChar;

  ParentNode           :TQUB_QFS_PNodeMem;
  ChildNode            :TQUB_QFS_PNodeMem;
  SiblingNode          :TQUB_QFS_PNodeMem;

  DataLoadSel          :TQUB_FirstLast;
  Data                 :Pointer;

  Stream               :TStream;
  StreamPos            :LongInt;

  Mutex                :Pointer;
	ReaderCount          :Integer;

end;


const


//Flags ************************************************************************
//On Stream
NODE_FLAG_UNSIGNED                              = 1;		// UNUSED {JB}
NODE_FLAG_PRELOAD                               = 2;
NODE_FLAG_DATA_IN_NODE                          = 4;
NODE_FLAG_MATRIX                                = 8;		// UNUSED {JB}
NODE_FLAG_NONWRITABLE                           = 16;

//In Memory only
NODE_FLAG_CHANGED                               = $1000;
NODE_FLAG_CHILDCHANGED                          = $2000;


//Flag Masks *******************************************************************
NODE_FLAG_ALL_MASK                              = High(TQUB_QFS_Node_Flags);
NODE_FLAG_UNSIGNED_MASK                         = NODE_FLAG_ALL_MASK - NODE_FLAG_UNSIGNED;			// UNUSED {JB}
NODE_FLAG_PRELOAD_MASK                          = NODE_FLAG_ALL_MASK - NODE_FLAG_PRELOAD;
NODE_FLAG_DATA_IN_NODE_MASK                     = NODE_FLAG_ALL_MASK - NODE_FLAG_DATA_IN_NODE;
NODE_FLAG_MATRIX_MASK                           = NODE_FLAG_ALL_MASK - NODE_FLAG_MATRIX;				// UNUSED {JB}
NODE_FLAG_CHANGED_MASK                          = NODE_FLAG_ALL_MASK - NODE_FLAG_CHANGED;				// UNUSED {JB}
NODE_FLAG_CHILDCHANGED_MASK                     = NODE_FLAG_ALL_MASK - NODE_FLAG_CHILDCHANGED;	// UNUSED {JB}
NODE_FLAG_NONWRITABLE_MASK                      = NODE_FLAG_ALL_MASK - NODE_FLAG_NONWRITABLE;


//Default flags ****************************************************************
NODE_FLAGS_DEFAULT                              = NODE_FLAG_PRELOAD or NODE_FLAG_CHANGED;
NODE_FLAGS_MEMONLY                              = NODE_FLAG_CHANGED or NODE_FLAG_CHILDCHANGED;	// UNUSED {JB}


//node type     ****************************************************************
QUB_NODETYPE_TNODE                              = 1;


// ************* next the stuff imported from qubtree.dll **********************

// const


QUB_NODE_MAX_LOAD_FOR_COPY                           = 1000000;


function Node_INCREF( n: TQUB_QFS_PNodeMem ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_INCREF';
function Node_DECREF( n: TQUB_QFS_PNodeMem ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_DECREF'; // returns # refs remaining cdecl;

function Node_Lock( n: TQUB_QFS_PNodeMem; timeoutMS: Integer ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_Lock'; // negative to wait indefinitely
procedure Node_Unlock( n: TQUB_QFS_PNodeMem ); cdecl; external 'qubtree.dll' name 'QTR_Unlock';

function Node_Create( name: PChar  ) : TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_Create';
function Node_Clone( other: TQUB_QFS_PNodeMem; deep :Integer  ) : TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_Clone';
function Node_Open( path: PChar; readOnly :Integer  ) : TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_Open';
procedure Node_Close( n :TQUB_QFS_PNodeMem ); cdecl; external 'qubtree.dll' name 'QTR_Close';
function Node_Save( n :TQUB_QFS_PNodeMem ): Integer; cdecl; external 'qubtree.dll' name 'QTR_Save';
function Node_SaveAs( n :TQUB_QFS_PNodeMem; path :PChar ): Integer; cdecl; external 'qubtree.dll' name 'QTR_SaveAs';
function Node_SaveAsTemp( n :TQUB_QFS_PNodeMem ): Integer; cdecl; external 'qubtree.dll' name 'QTR_SaveAsTemp';

function Node_FromString( str: PChar ): TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_FromString';

function Node_CreateChild( parent, presib :TQUB_QFS_PNodeMem; name :PChar ): TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_CreateChild';
function Node_InsertClone( parent, presib, other :TQUB_QFS_PNodeMem; deep :Integer ): TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_InsertClone';
procedure Node_InsertChild( parent, presib, child :TQUB_QFS_PNodeMem ); cdecl; external 'qubtree.dll' name 'QTR_InsertChild';
procedure Node_RemoveChild( parent, presib, child :TQUB_QFS_PNodeMem ); cdecl; external 'qubtree.dll' name 'QTR_RemoveChild';

function Node_Flag( n: TQUB_QFS_PNodeMem; mask: Integer ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_Flag';
procedure Node_SetFlag( n: TQUB_QFS_PNodeMem; mask: Integer; val: Integer ); cdecl; external 'qubtree.dll' name 'QTR_SetFlag';

procedure Node_Changed( n: TQUB_QFS_PNodeMem ); cdecl; external 'qubtree.dll' name 'QTR_Changed';
procedure Node_ChildChanged( n: TQUB_QFS_PNodeMem ); cdecl; external 'qubtree.dll' name 'QTR_ChildChanged';

function Node_Name( n: TQUB_QFS_PNodeMem ) : TPShortInt; cdecl; external 'qubtree.dll' name 'QTR_Name';
procedure Node_SetName( n: TQUB_QFS_PNodeMem; name: PChar  ); cdecl; external 'qubtree.dll' name 'QTR_SetName';

function Node_IsInFile( n: TQUB_QFS_PNodeMem ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_IsInFile';
function Node_FilePath( n: TQUB_QFS_PNodeMem ): PChar; cdecl; external 'qubtree.dll' name 'QTR_FilePath';

function Node_SetupData( n: TQUB_QFS_PNodeMem; dtype: Integer; size: LongWord; count: LongWord ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_SetupData';
function Node_SetupNumData( n: TQUB_QFS_PNodeMem; dtype: Integer; rows: LongWord; cols: LongWord ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_SetupNumData';
function Node_SetupStringData( n: TQUB_QFS_PNodeMem; count: LongWord ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_SetupStringData';
function Node_ResizeData( n: TQUB_QFS_PNodeMem; newRowCount: LongWord ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_ResizeData';
function Node_ClearData( n: TQUB_QFS_PNodeMem ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_ClearData';

function Node_NumCol( n :TQUB_QFS_PNodeMem ) : LongWord; cdecl; external 'qubtree.dll' name 'QTR_NumCol';

function Node_LoadRows( n: TQUB_QFS_PNodeMem; first: Integer; last: Integer; doRead: Integer ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_LoadRows';
function Node_UnloadRows( n: TQUB_QFS_PNodeMem; doWrite: Integer ) : Integer; cdecl; external 'qubtree.dll' name 'QTR_UnloadRows';

function Node_ReadRows( n: TQUB_QFS_PNodeMem; buf: Pointer; first: Integer; last: Integer ) : LongWord; cdecl; external 'qubtree.dll' name 'QTR_GetRows';
function Node_WriteRows( n: TQUB_QFS_PNodeMem; buf: Pointer; first: Integer; last: Integer ) : LongWord; cdecl; external 'qubtree.dll' name 'QTR_SetRows';

function Node_DataToCString( n: TQUB_QFS_PNodeMem ): PChar; cdecl; external 'qubtree.dll' name 'QTR_DataToCString';
function Node_ToCString( n: TQUB_QFS_PNodeMem ): PChar; cdecl; external 'qubtree.dll' name 'QTR_ToCString';
procedure Node_FreeCString( cstr: PChar ); cdecl; external 'qubtree.dll' name 'QTR_FreeDataCString';

function QTR_ReadTextFile( path: PChar ) : TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_FromTextFile';
procedure QTR_SaveTextFile( root: TQUB_QFS_PNodeMem; path: PChar ); cdecl; external 'qubtree.dll' name 'QTR_SaveTextFile';

function QTR_ReadTBL( path: PChar ) : TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_FromTBL';

function QTR_LookupName( name: PChar ) : PChar; cdecl; external 'qubtree.dll' name 'QTR_LookupName'; // get the canonical interned char * for quick comparison instead of strcmp


// conversion to/from a contiguous buffer in ondisk format

function QTR_ToBufferSize(impl :TQUB_QFS_PNodeMem): LongWord; cdecl; external 'qubtree.dll' name 'QTR_ToBufferSize';
function QTR_ToBuffer(impl :TQUB_QFS_PNodeMem; buffer :Pointer): LongWord; cdecl; external 'qubtree.dll' name 'QTR_ToBuffer';
function QTR_FromBuffer(buffer :Pointer): TQUB_QFS_PNodeMem; cdecl; external 'qubtree.dll' name 'QTR_FromBuffer';


// ********************  now the objects ***************************************

type


// interface ref-counting only works when you always store nodes as INode rather than TNode.

INode = interface (IUnknown)

   function GetName: String;
   procedure SetName(const nm : String);

   function GetChild(nameOrIx: Variant): INode; // dispatches to Child(...)
   function Child: INode; overload;
   function Child(ix	: Integer): INode; overload;
   function Child(const nm : String): INode; overload; // may create and append

   function IsPreload: Boolean;
   procedure SetPreload(pl : Boolean);
   function IsChanged: Boolean;
   procedure SetChanged(ch : Boolean);
   function IsChildChanged: Boolean;
   function IsDataInNode: Boolean;

   function GetDataType: TQUB_QFS_Node_DataType;
   function GetRowSize: LongWord;
   function GetDataCount: LongWord;
   function GetDataRows: LongWord;
   function GetDataCols: LongWord;
   function GetLoadBounds: TQUB_FirstLast;
   function GetLoadRowBounds: TQUB_FirstLast;

   function GetDataElem(ix :Integer): Variant;
   procedure SetDataElem(ix :Integer; v: Variant);

   function GetMatrixElem(i, j :Integer): Variant;
   procedure SetMatrixElem(i, j :Integer; v: Variant);

   function AsString: String;
   procedure SaveAsText(const path: String );

   function Save: Boolean;
   function SaveAs( const path :String ): Boolean;
   function SaveAsTemp: Boolean;
   function Close: Boolean;
   function IsInFile: Boolean;
   function GetPath: String;

   function Clone(deep : Boolean) : INode;
   function Copy: Boolean;

   function Equals(node : INode) : Boolean;
   function IsNull: Boolean;

   function Parent: INode;
   function Sibling: INode; overload;
   function Sibling(ix : Integer): INode; overload;
   function Sibling(const name: String): INode; overload;
   function SiblingSameName: INode;

   function Find(const nm : String): INode; // may return a null node

   function AppendChild(const newchildname : String): INode; overload;
   function AppendChild(newchild : INode): INode; overload;
   function AppendClone(Template :INode; deep :Boolean): INode;

   function InsertChild(after :INode; const newchildname : String): INode; overload;
   function InsertChild(after :INode; newchild : INode): INode; overload;
   function InsertClone(after :INode; orig : INode; deep :Boolean): INode;

   procedure RemoveChild(remch : INode); overload;
   procedure RemoveChild(remch : INode; Prev :INode); overload;
   procedure RemoveLast(const ChildName :String);

   function Flags: LongWord;

   function Lock(timeoutMS : Integer) : Boolean;
   procedure Unlock;

   function GetDataPtr: Pointer;
   function GetDataAsString: String;

   function SizeOfType(dtype :TQUB_QFS_Node_DataType): LongWord;

   function SetupData(dtype :TQUB_QFS_Node_DataType; dsize, dcount: LongWord): Boolean;
   function SetupNumData(dtype :TQUB_QFS_Node_DataType; rows, cols :LongWord): Boolean;
   function SetData(val: Variant): Boolean;
   function EmulateData(Other :INode): Boolean;

   function ResizeData(newrowcount : LongWord): Boolean;
   function ClearData: Boolean;

   function LoadRows(var bounds	: TQUB_FirstLast; doRead: Boolean): Boolean;
   function UnloadRows(doWrite : Boolean): Boolean;

   // next 2 return # rows read or written
   function ReadRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;
   function WriteRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;

   function GetLineComment: String;
   procedure SetLineComment(const com : String);

   function GetNodeMem: TQUB_QFS_PNodeMem;
   function GetNodeType: Integer;

   property Name                    :String                  read GetName             write SetName;
   property Path                    :String                  read GetPath;
   property LineComment             :String                  read GetLineComment      write SetLineComment;
   property Preload                 :Boolean                 read IsPreload           write SetPreload;
   property Changed                 :Boolean                 read IsChanged           write SetChanged;
   property ChildChanged            :Boolean                 read IsChildChanged;
   property DataInNode              :Boolean                 read IsDataInNode;
   property Nodes[nmOrIx: Variant]  :INode                   read GetChild; default;     // raises EVariantError for unsupported type or unloaded index
   property DataType                :TQUB_QFS_Node_DataType  read GetDataType;
   property RowSize                 :LongWord                read GetRowSize;
   property DataCount               :LongWord                read GetDataCount;
   property Rows                    :LongWord                read GetDataRows;
   property Cols                    :LongWord                read GetDataCols;
   property LoadBounds              :TQUB_FirstLast          read GetLoadBounds;
   property LoadRowBounds           :TQUB_FirstLast          read GetLoadRowBounds;
   property DataPtr                 :Pointer                 read GetDataPtr;
   property Data[ix :Integer]       :Variant                 read GetDataElem         write SetDataElem;
   property MatrixData[i,j :Integer]:Variant                 read GetMatrixElem       write SetMatrixElem;
   property DataAsString            :String                  read GetDataAsString;

   // nodes of the same type are compatible for things like .AppendChild
   // TNode is of type QUB_NODETYPE_TNODE
   // TQUB_QFS_Node:   QUB_NODETYPE_QFS
   property NodeType                :Integer                 read GetNodeType;

end;


TNode = class (TInterfacedObject, INode)
protected


   impl :TQUB_QFS_PNodeMem;


public


   constructor Create(const name: String );
   constructor CreateParented(const name: String; parent: INode );
   constructor CreateComment(const txt: String );
   constructor CreateNull;
   constructor CreateFromText(const path: String );
   constructor CreateFromString(const txt: String);
   constructor CreateUsing( pnode: TQUB_QFS_PNodeMem );
   constructor Paste;

   constructor Open(const path :String);
   constructor OpenReadOnly(const path :String);

   destructor Destroy; override;


   function GetName: String;
   procedure SetName(const nm : String);

   function GetChild(nameOrIx: Variant): INode; // dispatches to Child(...)
   function Child: INode; overload;
   function Child(ix	: Integer): INode; overload;
   function Child(const nm : String): INode; overload; // may create and append

   function IsPreload: Boolean;
   procedure SetPreload(pl : Boolean);
   function IsChanged: Boolean;
   procedure SetChanged(ch : Boolean);
   function IsChildChanged: Boolean;
   function IsDataInNode: Boolean;

   function GetDataType: TQUB_QFS_Node_DataType;
   function GetRowSize: LongWord;
   function GetDataCount: LongWord;
   function GetDataRows: LongWord;
   function GetDataCols: LongWord;
   function GetLoadBounds: TQUB_FirstLast;
   function GetLoadRowBounds: TQUB_FirstLast;

   function GetDataElem(ix :Integer): Variant;
   procedure SetDataElem(ix :Integer; v: Variant);

   function GetMatrixElem(i, j :Integer): Variant;
   procedure SetMatrixElem(i, j :Integer; v: Variant);

   function AsString: String;
   procedure SaveAsText(const path: String );

   function Save: Boolean;
   function SaveAs( const path :String ): Boolean;
   function SaveAsTemp: Boolean;
   function Close: Boolean;
   function IsInFile: Boolean;
   function GetPath: String;

   function Clone(deep : Boolean) : INode;
   function Copy: Boolean;

   function Equals(node : INode) : Boolean;
   function IsNull: Boolean;

   function Parent: INode;
   function Sibling: INode; overload;
   function Sibling(ix : Integer): INode; overload;
   function Sibling(const name: String): INode; overload;
   function SiblingSameName: INode;

   function Find(const nm : String): INode; // may return a null node

   function AppendChild(const newchildname : String): INode; overload;
   function AppendChild(newchild : INode): INode; overload;
   function AppendClone(Template :INode; deep :Boolean): INode;

   function InsertChild(after :INode; const newchildname : String): INode; overload;
   function InsertChild(after :INode; newchild : INode): INode; overload;
   function InsertClone(after :INode; orig : INode; deep :Boolean): INode;

   procedure RemoveChild(remch : INode); overload;
   procedure RemoveChild(remch : INode; Prev :INode); overload;
   procedure RemoveLast(const ChildName :String);

   function Flags: LongWord;

   function Lock(timeoutMS : Integer) : Boolean;
   procedure Unlock;

   function GetDataPtr: Pointer;
   function GetDataAsString: String;

   function SizeOfType(dtype :TQUB_QFS_Node_DataType): LongWord;

   function SetupData(dtype :TQUB_QFS_Node_DataType; dsize, dcount: LongWord): Boolean;
   function SetupNumData(dtype :TQUB_QFS_Node_DataType; rows, cols :LongWord): Boolean;
   function SetData(val: Variant): Boolean;
   function EmulateData(Other :INode): Boolean;

   function ResizeData(newrowcount : LongWord): Boolean;
   function ClearData: Boolean;

   function LoadRows(var bounds	: TQUB_FirstLast; doRead: Boolean): Boolean;
   function UnloadRows(doWrite : Boolean): Boolean;

   // next 2 return # bytes read or written
   function ReadRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;
   function WriteRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;

   function GetLineComment: String;
   procedure SetLineComment(const com : String);

   function GetNodeMem: TQUB_QFS_PNodeMem;
   function GetNodeType: Integer;

end;

function ReadTBLAsNode( filename :String ): INode;

function QUB_Node_DataTypeOfVariant(x :Variant): TQUB_QFS_Node_DataType;

function ReadIntProperty( Node :INode; Val :Integer ): Integer;
function ReadBoolProperty( Node :INode; Val :Boolean ): Boolean;
function ReadDoubleProperty( Node :INode; Val :Double ): Double;

function GetOrSetNodeInteger( Item : INode; iDefault : Integer ): Integer;
function GetOrSetNodeBoolean( Item : INode; Default : Boolean ): Boolean;
function GetOrSetNodeDouble( Item : INode; nDefault : Double ): Double;
function GetOrSetNodeString( Item : INode; cDefault : String ): String;

function CanPasteNode: Boolean; overload;
function CanPasteNode(var Name :String): Boolean; overload;


// ********************  next delphi wrappers of those buffer conversions ******

function Node_ToBufferSize(Node :INode): LongWord;
function Node_ToBuffer(Node :INode; buffer :Pointer): LongWord;
function Node_FromBuffer(buffer :Pointer): INode;


var
  QUB_Node_Null          :INode;              // warning i do not know if this var is thread-safe
  QUB_QFS_PNodeMem_Null  :TQUB_QFS_PNodeMem;


  QTR_ClipboardFormatName :String = 'qubtree';
  QTR_ClipNameFormatName  :String = 'qubtreeName';
  QTR_ClipboardFormat     :Word   = 0;
  QTR_ClipNameFormat      :Word   = 0;




procedure TreeNonFinitesToZero(Tree :INode);


implementation

uses IEEE754;

constructor TNode.Create( const name: String );
begin
   impl := Node_Create( PChar(name) );
end;

constructor TNode.CreateParented(const name: String; parent: INode );
begin
  impl := Node_Create( PChar(name) );
  parent.AppendChild( self );
end;

constructor TNode.CreateComment( const txt: String );
begin
   impl := Node_Create( PChar('QTR_COMMENT') );
   SetData( txt );
end;

constructor TNode.CreateNull;
begin
   impl := nil;
end;

constructor TNode.CreateFromText( const path: String );
begin
	impl := QTR_ReadTextFile( PChar(path) );
end;

constructor TNode.CreateFromString(const txt: String);
begin
  impl := Node_FromString( PChar(txt) );
end;

constructor TNode.CreateUsing( pnode: TQUB_QFS_PNodeMem );
begin
   impl := pnode;
   if impl <> nil then Node_INCREF( impl );
end;

constructor TNode.Paste;
var
  HMem     :Integer;
  PMem     :PChar;
begin
  impl := nil;
  HMem := 0;
  PMem := nil;

  if CanPasteNode and OpenClipboard(Application.Handle) then begin
    try
      //Retrieve Data Handle
      HMem := GetClipboardData(QTR_ClipboardFormat);
      if HMem <> 0 then begin
        PMem := GlobalLock(HMem);
        impl := QTR_FromBuffer(PMem);
      end;
    finally
      if PMem <> nil then GlobalUnlock(HMem);
      CloseClipboard;
    end;
  end;
end;

constructor TNode.Open(const path :String);
begin
  impl := Node_Open( PChar(path), 0 );
end;

constructor TNode.OpenReadOnly(const path :String);
begin
  impl := Node_Open( PChar(path), 1 );
end;

destructor TNode.Destroy;
begin
   if impl <> nil then Node_DECREF( impl );
   inherited;
end;

function TNode.GetName: String;
begin
   if Lock(-1) then begin
     try
       if (impl <> nil) and (impl.name <> nil) then Result := impl.name;
     finally
       Unlock;
     end;
   end;
end;

procedure TNode.SetName(const nm : String);
begin
   if impl <> nil then Node_SetName( impl, PChar(nm) );
end;

function TNode.GetChild(nameOrIx: Variant): INode; // for property [], dispatches to Child(...)
begin
  case VarType(nameOrIx) and varTypeMask of
    varSmallint, varInteger:
      Result := Child( Integer(nameOrIx) );
    varString:
      Result := Child( String(nameOrIx) );
  else
    Result := TNode.CreateNull;
  end;
end;

function TNode.Child: INode;
begin
   if impl <> nil then Result := TNode.CreateUsing( impl.ChildNode )
   else Result := TNode.CreateNull;
end;

function TNode.Child(ix	: Integer): INode;
begin
   if impl <> nil then begin
      if ix = 0 then begin
	 Result := Child;
      end else begin
	 Result := Child.Sibling( ix );
      end;
   end else begin
      Result := TNode.CreateNull;
   end;
end;

function TNode.Child(const nm : String): INode;
var
   last      : TQUB_QFS_PNodeMem;
begin
  Result := Find(nm);
  if (impl <> nil) and Result.IsNull then begin
    last      := Child.GetNodeMem;
    if last <> nil then begin
      while last.SiblingNode <> nil do begin
        last  := last.SiblingNode;
      end;
    end;

    Result := TNode.CreateUsing( Node_CreateChild( impl, last, PChar(nm) ) );
  end;
end;


function TNode.IsPreload: Boolean;
begin
  if impl <> nil then Result := Node_Flag( impl, NODE_FLAG_PRELOAD ) <> 0
  else                Result := False;
end;

procedure TNode.SetPreload(pl : Boolean);
begin
  if impl <> nil then Node_SetFlag( impl, NODE_FLAG_PRELOAD, Integer(pl) );
end;

function TNode.IsChanged: Boolean;
begin
  if impl <> nil then Result := Node_Flag( impl, NODE_FLAG_CHANGED ) <> 0
  else                Result := False;
end;

procedure TNode.SetChanged(ch : Boolean);
begin
  if impl <> nil then begin
    if ch then begin
      Node_Changed( impl );
    end else begin
      Node_SetFlag( impl, NODE_FLAG_CHANGED, 0 );
    end;
  end;
end;

function TNode.IsChildChanged: Boolean;
begin
  if impl <> nil then Result := Node_Flag( impl, NODE_FLAG_CHILDCHANGED ) <> 0
  else                Result := False;
end;

function TNode.IsDataInNode: Boolean;
begin
  if impl <> nil then Result := Node_Flag( impl, NODE_FLAG_DATA_IN_NODE ) <> 0
  else                Result := False;
end;

function TNode.GetDataType: TQUB_QFS_Node_DataType;
begin
   if impl <> nil then Result := impl.NodeStream.DataType
   else Result := NODE_DATATYPE_EMPTY;
end;

function TNode.GetRowSize: LongWord;
begin
   if impl <> nil then Result := impl.NodeStream.DataSize
   else Result := 0;
end;

function TNode.GetDataCount: LongWord;
begin
   if impl <> nil then begin
     if (impl.NodeStream.DataType = NODE_DATATYPE_STRING)
     and (PChar(GetDataPtr)[0] = #0) then begin
       Result := 0;
     end else begin
       Result := LongWord(impl.NodeStream.DataCount) * GetDataCols
     end;
   end else begin
     Result := 0;
   end;
end;

function TNode.GetDataRows: LongWord;
begin
   if impl <> nil then begin
     if (impl.NodeStream.DataType = NODE_DATATYPE_STRING)
     and (PChar(GetDataPtr)[0] = #0) then begin
       Result := 0;
     end else begin
       Result := LongWord(impl.NodeStream.DataCount)
     end;
   end else begin
     Result := 0;
   end;
end;

function TNode.GetDataCols: LongWord;
begin
   if impl <> nil then Result := Node_NumCol( impl )
   else Result := 0;
end;

function TNode.GetLoadBounds: TQUB_FirstLast;
var
  Cols           :LongWord;
begin
  Result         := GetLoadRowBounds;
  if Result.First <> -1 then begin
    Cols         := GetDataCols;
    Result.First := Result.First * LongInt(Cols);
    Result.Last  := (Result.Last + 1) * LongInt(Cols) - 1;
  end;
end;

function TNode.GetLoadRowBounds: TQUB_FirstLast;
begin
   if impl <> nil then Result := impl.DataLoadSel
   else begin
     Result.First := -1;
     Result.Last := -1;
   end;
end;

function TNode.GetDataElem(ix :Integer): Variant;
var
  loaded                      :TQUB_FirstLast;
  dataptr                     :Pointer;
begin
  loaded := GetLoadBounds;
  dataptr := GetDataPtr;
  if (loaded.First >= 0) and (loaded.First <= ix) and (ix <= loaded.Last) then begin
    ix := ix - loaded.First;
    case GetDataType of
      NODE_DATATYPE_UNKNOWN:   raise EVariantError.Create('can''t get unknown data as a variant');
      NODE_DATATYPE_POINTER:   Result := Integer(TPAPointer(dataptr)[ix]);   // yeah yeah
      NODE_DATATYPE_STRING:    Result := Integer(PChar(dataptr)[ix]);
      NODE_DATATYPE_BYTE:      Result := TPAByte(dataptr)[ix];
      NODE_DATATYPE_SHORTINT:  Result := TPAShort(dataptr)[ix];
      NODE_DATATYPE_WORD:      Result := TPAWord(dataptr)[ix];
      NODE_DATATYPE_SMALLINT:  Result := TPASmall(dataptr)[ix];
      NODE_DATATYPE_LONGWORD:  Result := Integer(TPALongWord(dataptr)[ix]);
      NODE_DATATYPE_LONGINT:   Result := TPAInteger(dataptr)[ix];
      NODE_DATATYPE_UINT64:    Result := Integer( TPAUInt64(dataptr)[ix] );
      NODE_DATATYPE_INT64:     Result := Integer( TPAInt64(dataptr)[ix] );
      NODE_DATATYPE_SINGLE:    Result := TPASingle(dataptr)[ix];
      NODE_DATATYPE_DOUBLE:    Result := TPADouble(dataptr)[ix];
      NODE_DATATYPE_EXTENDED:  Result := TPAExtended(dataptr)[ix];
    end;
  end else begin
    raise EVariantError.Create(Format('node data index out of loaded bounds: %d', [ix]));
  end;
end;

procedure TNode.SetDataElem(ix :Integer; v: Variant);
var
  loaded        :TQUB_FirstLast;
  dataptr       :Pointer;
begin
  loaded        := GetLoadBounds;
  dataptr       := GetDataPtr;
  if (loaded.First >= 0) and (loaded.First <= ix) and (ix <= loaded.Last) then begin
    ix := ix - loaded.First;
    case GetDataType of
      NODE_DATATYPE_UNKNOWN:   raise EVariantError.Create('can''t assign a variant to unknown data');
      NODE_DATATYPE_POINTER:   TPAPointer(dataptr)[ix] := Pointer(Integer(v));
      NODE_DATATYPE_STRING:    PChar(dataptr)[ix] := Char(Integer(v));
      NODE_DATATYPE_BYTE:      TPAByte(dataptr)[ix] := Byte(v);
      NODE_DATATYPE_SHORTINT:  TPAShort(dataptr)[ix] := ShortInt(v);
      NODE_DATATYPE_WORD:      TPAWord(dataptr)[ix] := Word(v);
      NODE_DATATYPE_SMALLINT:  TPASmall(dataptr)[ix] := SmallInt(v);
      NODE_DATATYPE_LONGWORD:  TPALongWord(dataptr)[ix] := LongWord(v);
      NODE_DATATYPE_LONGINT:   TPAInteger(dataptr)[ix] := Integer(v);
      NODE_DATATYPE_UINT64:    TPAUInt64(dataptr)[ix] := Integer(v);
      NODE_DATATYPE_INT64:     TPAInt64(dataptr)[ix] := Integer(v);
      NODE_DATATYPE_SINGLE:    TPASingle(dataptr)[ix] := Single(v);
      NODE_DATATYPE_DOUBLE:    TPADouble(dataptr)[ix] := Double(v);
      NODE_DATATYPE_EXTENDED:  TPAExtended(dataptr)[ix] := Extended(v);
    end;
    SetChanged(True);
  end else begin
    raise EVariantError.Create(Format('node data index out of loaded bounds: %d', [ix]));
  end;
end;

function TNode.GetMatrixElem(i, j :Integer): Variant;
begin
  Result := GetDataElem( i * LongInt(GetDataCols) + j );
end;

procedure TNode.SetMatrixElem(i, j :Integer; v: Variant);
begin
  SetDataElem( i * LongInt(GetDataCols) + j, v );
end;

function TNode.AsString: String;
var
  cstr :PChar;
begin
  if impl <> nil then begin
    cstr := Node_ToCString( impl );
    Result := cstr;
    Node_FreeCString( cstr );
  end else begin
    Result := '';
  end;
end;

procedure TNode.SaveAsText( const path: String );
begin
  if impl <> nil then QTR_SaveTextFile( impl, PChar(path) );
end;

function TNode.Save: Boolean;
begin
  if impl <> nil then Result := 0 <= Node_Save( impl )
  else                Result := False;
end;

function TNode.SaveAs( const path :String ): Boolean;
begin
  if impl <> nil then Result := 0 <= Node_SaveAs( impl, PChar(path) )
  else                Result := False;

  if Result then Result := Save;
end;

function TNode.SaveAsTemp: Boolean;
begin
  if impl <> nil then Result := 0 <= Node_SaveAsTemp( impl )
  else                Result := False;

  if Result then Result := Save;
end;

function TNode.Close: Boolean;
begin
  if impl <> nil then begin
    Node_Close( impl );
    Result := True;
  end else begin
    Result := False;
  end;
end;

function TNode.IsInFile: Boolean;
begin
  if impl <> nil then Result := Node_IsInFile( impl ) <> 0
  else                Result := False;
end;

function TNode.GetPath: String;
var
  Path        :PChar;
begin
  if impl <> nil then Path := Node_FilePath( impl )
  else                Path := nil;

  if path <> nil then Result := Path;
end;

function TNode.Clone(deep : Boolean) : INode;
var
   newimpl   : TQUB_QFS_PNodeMem;
begin
   newimpl := nil;
   if impl <> nil then begin
     newimpl := Node_Clone( impl, Integer(deep) );
   end;
   Result := TNode.CreateUsing( newimpl );
   if newimpl <> nil then Node_DECREF( newimpl );
end;

function TNode.Copy: Boolean;
var
  HMem      :Integer;
  PMem      :PChar;
  Size      :Integer;
  Name      :String;
begin
  HMem      := 0;

  Result := False;
  if Self.IsNull then Exit;
  if QTR_ClipboardFormat = 0 then Exit;
  Size  := QTR_ToBufferSize(impl);
  if Size = 0 then Exit;

  if OpenClipboard(Application.Handle) then begin
    try
      EmptyClipboard;
      HMem  := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Size);
      if HMem = 0 then Exit;
      PMem  := GlobalLock(HMem);
      try
        QTR_ToBuffer(impl, PMem);
      except
      end;
      GlobalUnlock(HMem);
      if 0 <> SetClipboardData(QTR_ClipboardFormat, HMem) then begin
        HMem   := 0;
        Result := True;

        // now put up the name
        Name   := Self.GetName;
        Size   := Length(Name)+1;
        HMem   := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Size);
        if HMem <> 0 then begin
          PMem := GlobalLock(HMem);
          Move(PChar(Name)[0], PMem[0], Size);
          GlobalUnlock(HMem);
          if 0 <> SetClipboardData(QTR_ClipNameFormat, HMem) then begin
            HMem := 0;
          end;
        end;
      end;
    finally
      if HMem <> 0 then GlobalFree(HMem);
      CloseClipboard;
    end;
  end;
end;

function TNode.Equals(node : INode) : Boolean;
begin
   Result := impl = node.GetNodeMem;
end;

function TNode.IsNull: Boolean;
begin
   Result := impl = nil;
end;

function TNode.Parent: INode;
begin
   if impl <> nil then Result := TNode.CreateUsing( impl.ParentNode )
   else Result := TNode.CreateNull;
end;

function TNode.Sibling: INode;
begin
   if impl <> nil then Result := TNode.CreateUsing( impl.SiblingNode )
   else Result := TNode.CreateNull;
end;

function TNode.Sibling(ix : Integer): INode;
begin
   if impl <> nil then begin
      if ix = 0 then begin
	 Result := Self;
      end else begin
	 Result := Sibling.Sibling( ix - 1 );
      end;
   end else begin
      Result := TNode.CreateNull;
   end;
end;

function TNode.Sibling(const name : String): INode;
var
   sb : INode;
begin
   sb := Sibling;

   if (impl <> nil) and not sb.IsNull then begin
      if sb.GetName = name then begin
	 Result := sb;
      end else begin
	 Result := sb.Sibling( name );
      end;
   end else begin
      Result := TNode.CreateNull;
   end;
end;


function TNode.Find(const nm : String): INode; // may return a null node
var
   PName    :Pointer;
   PNode    :TQUB_QFS_PNodeMem;
begin
   Result   := TNode.CreateNull;
   PName    := Pointer( QTR_LookupName( PChar(nm) ) );
   PNode    := Child.GetNodeMem;

   while PNode <> nil do begin
     if Pointer(PNode.Name) = PName then begin
       Result := TNode.CreateUsing(PNode);
       Exit;
     end;

     PNode  := PNode.SiblingNode;
   end;
end;

function TNode.SiblingSameName: INode;
var
   PName    :Pointer;
   PNode    :TQUB_QFS_PNodeMem;
begin
   Result   := TNode.CreateNull;
   if impl = nil then Exit;

   PName    := Pointer( impl.Name );
   PNode    := impl.SiblingNode;

   while PNode <> nil do begin
     if Pointer(PNode.Name) = PName then begin
       Result := TNode.CreateUsing(PNode);
       Exit;
     end;

     PNode  := PNode.SiblingNode;
   end;
end;

(*
function TNode.Find(const nm : String): INode; // may return a null node
var
   ch : INode;
begin
   ch := Child;

   if (impl <> nil) and not ch.IsNull then begin
      if ch.GetName = nm then begin
	 Result := ch;
      end else begin
	 Result := ch.Sibling( nm );
      end;
   end else begin
      Result := TNode.CreateNull;
   end;
end;
*)
function TNode.AppendChild(const newchildname : String): INode;
var
   newimpl   : TQUB_QFS_PNodeMem;
   last      : TQUB_QFS_PNodeMem;
begin
   newimpl := nil;

   if impl <> nil then begin
     last      := Child.GetNodeMem;
     if last <> nil then begin
       while last.SiblingNode <> nil do begin
         last  := last.SiblingNode;
       end;
     end;

     newimpl := Node_CreateChild( impl, last, PChar(newchildname) );
   end;
   Result := TNode.CreateUsing( newimpl );
end;

function TNode.AppendChild(newchild : INode): INode;
var
   last      : TQUB_QFS_PNodeMem;
begin
   Result := newchild;

   if (impl <> nil) and not newchild.IsNull then begin
     last      := Child.GetNodeMem;
     if last <> nil then begin
       while last.SiblingNode <> nil do begin
         last  := last.SiblingNode;
       end;
     end;

     Node_InsertChild( impl, last, newchild.GetNodeMem );
   end;
end;

function TNode.AppendClone(Template :INode; deep :Boolean): INode;
var
   newimpl   : TQUB_QFS_PNodeMem;
   last      : TQUB_QFS_PNodeMem;
begin
   newimpl := nil;
   if (impl <> nil) and not Template.IsNull then begin
     last      := Child.GetNodeMem;
     if last <> nil then begin
       while last.SiblingNode <> nil do begin
         last  := last.SiblingNode;
       end;
     end;

     newimpl := Node_InsertClone( impl, last, Template.GetNodeMem, Integer(deep) );
   end;
   Result := TNode.CreateUsing( newimpl );
end;

function TNode.InsertChild(after :INode; const newchildname : String): INode;
var
   newimpl   : TQUB_QFS_PNodeMem;
begin
   newimpl   := nil;
   if impl <> nil then begin
     newimpl := Node_CreateChild( impl, after.GetNodeMem, PChar(newchildname) );
   end;
   Result    := TNode.CreateUsing( newimpl );
end;

function TNode.InsertChild(after :INode; newchild : INode): INode;
begin
   if (impl <> nil) and not newchild.IsNull then begin
     Node_InsertChild( impl, after.GetNodeMem, newchild.GetNodeMem );
   end;
   Result    := newchild;
end;

function TNode.InsertClone(after :INode; orig : INode; deep :Boolean): INode;
var
   newimpl   : TQUB_QFS_PNodeMem;
begin
   if orig.IsNull then begin
     Result := orig;
     Exit;
   end;

   newimpl   := nil;
   if impl <> nil then begin
     newimpl := Node_InsertClone( impl, after.GetNodeMem, orig.GetNodeMem, Integer(deep) );
   end;
   Result    := TNode.CreateUsing( newimpl );
end;

procedure TNode.RemoveChild(remch : INode);
var
  chimpl       :TQUB_QFS_PNodeMem;
  Prev         :TQUB_QFS_PNodeMem;
begin
  chimpl       := remch.GetNodeMem;
  if (impl = nil) or (chimpl = nil) then Exit;

  if chimpl = impl.ChildNode then begin
    Prev       := nil;
  end else begin
    Prev       := Child.GetNodeMem;
    while (Prev <> nil) and (Prev.SiblingNode <> chimpl) do begin
      Prev     := Prev.SiblingNode;
    end;
    if Prev = nil then Exit; // this is not a child!
  end;
  Node_RemoveChild( impl, Prev, chimpl );
end;

procedure TNode.RemoveChild(remch : INode; Prev :INode);
begin
  if (impl <> nil) and (not remch.IsNull) and (remch.Equals( Prev.Sibling )) then begin
    Node_RemoveChild( impl, Prev.GetNodeMem, remch.GetNodeMem );
  end;
end;

procedure TNode.RemoveLast(const ChildName :String);
var
  Ch, Nxt :INode;
begin
  Ch := Find(ChildName);
  if Ch.IsNull then Exit;

  Nxt := Ch.Sibling(ChildName);
  while not Nxt.IsNull do begin
    Ch := Nxt;
    Nxt := Ch.Sibling(ChildName);
  end;

  RemoveChild(Ch);
end;


function TNode.Flags: LongWord;
begin
  if impl <> nil then Result := TPLongWord(@(impl.NodeStream))^ and $FFFFFF
  else                Result := 0;
end;

function TNode.Lock(timeoutMS : Integer) : Boolean;
begin
   if impl <> nil then Result := Boolean( Node_Lock( impl, timeoutMS ) )
   else Result := True;
end;

procedure TNode.Unlock;
begin
   if impl <> nil then Node_Unlock( impl );
end;

function TNode.GetDataPtr: Pointer;
begin
   if impl <> nil then Result := impl.Data
   else Result := nil;
end;

function TNode.GetDataAsString: String;
var
  cstr :PChar;
begin
  if impl <> nil then begin
    cstr := Node_DataToCString( impl );
    Result := cstr;
    Node_FreeCString( cstr );
  end;
end;

function TNode.SizeOfType(dtype :TQUB_QFS_Node_DataType): LongWord;
begin
  case dtype of
    NODE_DATATYPE_EMPTY:        Result := 0;
    NODE_DATATYPE_UNKNOWN:      Result := SizeOf(Pointer);
    NODE_DATATYPE_POINTER:      Result := SizeOf(Pointer);
    NODE_DATATYPE_STRING:       Result := SizeOf(Char);
    NODE_DATATYPE_BYTE:         Result := SizeOf(Byte);
    NODE_DATATYPE_SHORTINT:     Result := SizeOf(ShortInt);
    NODE_DATATYPE_WORD:         Result := SizeOf(Word);
    NODE_DATATYPE_SMALLINT:     Result := SizeOf(SmallInt);
    NODE_DATATYPE_LONGWORD:     Result := SizeOf(LongWord);
    NODE_DATATYPE_LONGINT:      Result := SizeOf(Integer);
    NODE_DATATYPE_UINT64:       Result := SizeOf(UInt64);
    NODE_DATATYPE_INT64:        Result := SizeOf(Int64);
    NODE_DATATYPE_SINGLE:       Result := SizeOf(Single);
    NODE_DATATYPE_DOUBLE:       Result := SizeOf(Double);
    NODE_DATATYPE_EXTENDED:     Result := SizeOf(Extended);
  else
    Result := 0;
  end;
end;

function TNode.SetupData(dtype :TQUB_QFS_Node_DataType; dsize, dcount: LongWord): Boolean;
begin
  if impl <> nil then Result := 0 <= Node_SetupData( impl, Integer(dtype), dsize, dcount )
  else                Result := False;
end;

function TNode.SetupNumData(dtype :TQUB_QFS_Node_DataType; rows, cols :LongWord): Boolean;
begin
  if impl <> nil then Result := 0 <= Node_SetupNumData( impl, Integer(dtype), rows, cols )
  else                Result := False;
end;

function TNode.SetData(val: Variant): Boolean;
var
  vtype                :Integer;
  dtype                :TQUB_QFS_NODE_DataType;
  high, low            :Integer;
  count                :LongWord;
  sval                 :String;
  allbounds            :TQUB_FirstLast;
  ptr                  :Pointer;
begin

  Result := False;
  ptr := nil;

  if impl <> nil then begin
    allBounds.First := 0;
    allBounds.Last := 0;

    vtype := VarType(val);
    if ((vtype and varString) <> 0) or ((vtype and varOLEStr) <> 0) then begin
    //String

      sval := val;
      count := Length(sval);

      // while qubtree allocates space for a null terminator, to avoid access violations with c libraries,
      // it does not count the null as part of the string's length.

      if count > 0 then begin
        SetupData( NODE_DATATYPE_STRING, 1, count );
        allBounds.Last := count - 1;
        LoadRows( allBounds, False{noread} );
        Move(PChar(sval)^, GetDataPtr^, count);
      end else begin

        // but I guess this won't break anything, to have an 'empty string' convention.

        //ClearData;
        //There is a problem with ClearData, in that it changes the type to EMPTY. Lorin
        SetupData(NODE_DATATYPE_STRING, 1, 1{to account for the zero-terminator});
        allBounds.Last := 0{to account for the zero-terminator};
        LoadRows(allBounds, False{noread});
        //Just a zero - when read with DataAsString, the PChar conversion takes care of it and returns an empty string
        PChar(GetDataPtr)[0] := Char(0);
      end;
      Result := True;

    end else begin

      dtype := QUB_Node_DataTypeOfVariant(val);

      if (vtype and varArray) <> 0 then begin
      //Array
        high := VarArrayHighBound(val, 1);
        low := VarArrayLowBound(val, 1);
        count := high - low + 1;
        SetupNumData( dtype, count, 1 );
        allBounds.Last := Count - 1;
        LoadRows( allBounds, False{noread} );

        try
          ptr := VarArrayLock(val);
          if ptr <> nil then Move(ptr^, GetDataPtr^, count * SizeOfType(dtype));
        finally
          if ptr <> nil then VarArrayUnLock(val);
        end;

      end else begin

        SetupNumData( dtype, 1, 1 );
        LoadRows( allBounds, False{noread} );
        SetDataElem( 0, val );

      end;

      Result := True;

    end;
  end;
end;

function TNode.EmulateData(Other :INode): Boolean;
var
  All      :TQUB_FirstLast;
begin
  if Other.DataCount = 0 then begin
    Result := ClearData;
  end else if (Other.DataType = NODE_DATATYPE_STRING) then begin
    Result := SetData( Other.DataAsString );
  end else begin
    if  (GetDataType <> Other.DataType)
    or (GetDataCount <> Other.DataCount) then begin
      SetupData(Other.DataType, Other.RowSize, Other.DataCount);
    end;
    All    := QUB_FirstLast(0, GetDataRows-1);
    Result := GetDataRows = Other.ReadRows(GetDataPtr, All);
  end;
end;

function TNode.ResizeData(newrowcount : LongWord): Boolean;
begin
// some nasty bugs "node data out of bounds" happen
// when you try to resize 0-length data.
// (what type should the resized data have?)
// uncomment this code for testing:
//  if (impl <> nil) and (impl.NodeStream.DataCount = 0) and (newrowcount <> 0)
//  then raise Exception.Create('Attempt to resize empty node data!');
  
  if impl <> nil then Result := 0 <= Node_ResizeData( impl, newrowcount )
  else                Result := False;
end;

function TNode.ClearData: Boolean;
begin
  if impl <> nil then Result := 0 <= Node_ClearData( impl )
  else                Result := False;
end;

function TNode.LoadRows(var bounds : TQUB_FirstLast; doRead: Boolean): Boolean;
begin
  if impl <> nil then Result := 0 <= Node_LoadRows( impl, bounds.First, bounds.Last, Integer(doRead) )
  else                Result := False;
end;

function TNode.UnloadRows(doWrite : Boolean): Boolean;
begin
  if impl <> nil then Result := 0 <= Node_UnloadRows( impl, Integer(doWrite) )
  else                Result := False;
end;

// next 2 return # rows read or written
function TNode.ReadRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;
begin
   if impl <> nil then Result := Node_ReadRows( impl, buf, bounds.First, bounds.Last )
   else                Result := 0;
end;

function TNode.WriteRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;
begin
   if impl <> nil then Result := Node_WriteRows( impl, buf, bounds.First, bounds.Last )
   else                Result := 0;
end;

function TNode.GetLineComment: String;
var
   com : INode;
begin
   com := Find('QTR_LINE_COMMENT');
   if not com.IsNull then Result := com.DataAsString
   else Result := '';
end;

procedure TNode.SetLineComment(const com : String);
begin
  Child('QTR_LINE_COMMENT').SetData( com );
end;

function TNode.GetNodeMem: TQUB_QFS_PNodeMem;
begin
   Result := impl;
end;

function TNode.GetNodeType: Integer;
begin
  Result := QUB_NODETYPE_TNODE;
end;

function ReadTBLAsNode( filename :String ): INode;
var
  impl   :TQUB_QFS_PNodeMem;
begin
  impl   := QTR_ReadTBL( PChar(filename) );
  Result := TNode.CreateUsing( impl );
  if impl <> nil then Node_DECREF( impl );
end;

// *****************************************************************************

function QUB_Node_DataTypeOfVariant(x :Variant): TQUB_QFS_Node_DataType;
begin
  case TVarData(x).VType of
    varString,
    varOLEStr:    Result := NODE_DATATYPE_STRING;
    varByte:      Result := NODE_DATATYPE_BYTE;
    varSmallint:  Result := NODE_DATATYPE_SMALLINT;
    varInteger:   Result := NODE_DATATYPE_LONGINT;
    varSingle:    Result := NODE_DATATYPE_SINGLE;
    varDouble:    Result := NODE_DATATYPE_DOUBLE;
  else
    raise EVariantError.Create('unknown variant type');
  end;
end;

// *****************************************************************************

function ReadIntProperty( Node :INode; Val :Integer ): Integer;
begin
  if Node.DataCount > 0 then begin
    try
      Val := Integer(Node.Data[0]);
    except
    end;
  end;

  Result := Val;
end;

function ReadBoolProperty( Node :INode; Val :Boolean ): Boolean;
begin
  if Node.DataCount > 0 then begin
    try
      Val := 0 <> Integer(Node.Data[0]);
    except
    end;
  end;

  Result := Val;
end;

function ReadDoubleProperty( Node :INode; Val :Double ): Double;
begin
  if Node.DataCount > 0 then begin
    try
      Val := Double(Node.Data[0]);
    except
    end;
  end;
  Result := Val;
end;

function GetOrSetNodeInteger( Item : INode; iDefault : Integer ): Integer;
begin
  if Item.DataCount > 0 then
    Result := Integer( Item.Data[0] )
  else begin
    Item.SetData( iDefault );
    Result := iDefault;
  end;
end;

function GetOrSetNodeBoolean( Item : INode; Default : Boolean ): Boolean;
begin
  if Item.DataCount > 0 then
    if Integer(Item.Data[0])=1 then Result := True else Result := False
  else begin
    Item.SetData( Choice(Default,Integer(1),Integer(0)) );
    Result := Default;
  end;
end;

function GetOrSetNodeDouble( Item : INode; nDefault : Double ): Double;
begin
  if Item.DataCount > 0 then
    Result := Double( Item.Data[0] )
  else begin
    Item.SetData( nDefault );
    Result := nDefault;
  end;
end;

function GetOrSetNodeString( Item : INode; cDefault : String ): String;
begin
  if Item.DataCount > 0 then
    Result := Item.DataAsString
  else begin
    Item.SetData( cDefault );
    Result := cDefault;
  end;
end;



// *****************************************************************************

function Node_ToBufferSize(Node :INode): LongWord;
begin
  Result := QTR_ToBufferSize(Node.GetNodeMem);
end;

function Node_ToBuffer(Node :INode; buffer :Pointer): LongWord;
begin
  Result := QTR_ToBuffer(Node.GetNodeMem, buffer);
end;

function Node_FromBuffer(buffer :Pointer): INode;
var
  impl   :TQUB_QFS_PNodeMem;
begin
  impl   := QTR_FromBuffer(buffer);
  Result := TNode.CreateUsing(impl);
  if impl <> nil then Node_DECREF(impl);
end;



function CanPasteNode: Boolean;
begin
  Result := OpenClipboard(Application.Handle);
  if Result then begin
    Result := IsClipboardFormatAvailable(QTR_ClipboardFormat);
    CloseClipboard;
  end;
end;

function CanPasteNode(var Name :String): Boolean;
var
  HMem   :Integer;
  PMem   :PChar;
  Size   :Integer;
begin
  Result := OpenClipboard(Application.Handle);
  if Result then begin
    Result := IsClipboardFormatAvailable(QTR_ClipboardFormat);

    if IsClipboardFormatAvailable(QTR_ClipNameFormat) then begin
      HMem := GetClipboardData(QTR_ClipNameFormat);
      if HMem <> 0 then begin
        Size := GlobalSize(HMem)-1;
        PMem := GlobalLock(HMem);
        SetLength(Name, Size);
        Move(PMem[0], Name[1], Size);
        GlobalUnlock(HMem);
      end;
    end else begin
      Name := 'QTR_GET_IT_THE_HARD_WAY';
    end;

    CloseClipboard;
  end;
  if Name = 'QTR_GET_IT_THE_HARD_WAY' then begin
    Name := TNode.Paste.GetName;
  end;
end;

// *****************************************************************************

procedure TreeNonFinitesToZero(Tree :INode);
var
  SData    :TPASingle;
  DData    :TPADouble;
  i        :Integer;
  Child    :INode;
begin
  if Tree.DataType = NODE_DATATYPE_SINGLE then begin
    SData  := TPASingle(Tree.DataPtr);
    for i := 0 to Tree.DataCount - 1 do begin
      if IsNAN(SData[i]) or IsInfinity(SData[i]) then begin
        SData[i] := 0.0;
      end;
    end;
  end else if Tree.DataType = NODE_DATATYPE_DOUBLE then begin
    DData  := TPADouble(Tree.DataPtr);
    for i := 0 to Tree.DataCount - 1 do begin
      if IsNAN(DData[i]) or IsInfinity(DData[i]) then begin
        DData[i] := 0.0;
      end;
    end;
  end;

  Child := Tree.Child;
  while not Child.IsNull do begin
    TreeNonFinitesToZero(Child);
    Child := Child.Sibling;
  end;
end;

// *****************************************************************************

initialization


  QUB_Node_Null         := TNode.CreateNull;
  QUB_QFS_PNodeMem_Null := nil;

  QTR_ClipboardFormat   := RegisterClipboardFormat(PChar(QTR_ClipboardFormatName));
  QTR_ClipNameFormat    := RegisterClipboardFormat(PChar(QTR_ClipNameFormatName));


// *****************************************************************************

end.

