{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2003
// a component for visual browsing and editing of a qubtree

unit QUB_NodeEditor;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
StdCtrls, ExtCtrls, ComCtrls, Menus, Grids, Math,

QUB_Table,

QUB_Console, QUB_Rich_Console,
QUB_Node,
QUB_Global;


const


QUB_NODEEDITOR_MAXROWS = 10240;


type


TQUB_NodeEditEvent = procedure (Sender :TObject; Node :INode; R, C :Integer) of object;


TQUB_NodeEditorEntry = class
public
  Node          :INode;
  Constr        :INode;
end;


TQUB_NodeEditor = class(TFrame)


  treeTree: TTreeView;
  splLeft: TSplitter;
  panEditor: TPanel;
  panEditString: TPanel;
  panEditNum: TPanel;
  panEditNone: TPanel;
  txtData: TMemo;
  popupTree: TPopupMenu;
  itemInsertChild: TMenuItem;
  itemInsertSibling: TMenuItem;
  itemRename: TMenuItem;
  itemCut: TMenuItem;
  itemCopy: TMenuItem;
  itemPasteChild: TMenuItem;
  itemPasteSibling: TMenuItem;
  itemDelete: TMenuItem;
  lblComment: TLabel;
  lblDataType: TLabel;
  cmbDataType: TComboBox;
  txtComment: TEdit;
  Button1: TButton;
  panMeta: TPanel;
  panData: TPanel;
  panRowsCols: TPanel;
  panRowsLoaded: TPanel;
  lblRows: TLabel;
  txtRows: TEdit;
  lblCols: TLabel;
  txtCols: TEdit;
  Label5: TLabel;
  txtRowsLoadedFrom: TEdit;
  Label6: TLabel;
  txtRowsLoadedTo: TEdit;
  btnUnload: TButton;
  panGridData: TPanel;
  panSlider: TPanel;
    Panel1: TPanel;
    lblSlider: TLabel;
    trkSlider: TTrackBar;


  procedure txtCommentExit(Sender: TObject);
  procedure txtCommentKeyPress(Sender: TObject; var Key: Char);
  procedure txtDataChange(Sender: TObject);
  procedure txtDataExit(Sender: TObject);
  procedure txtDataMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure txtRowsExit(Sender: TObject);
  procedure txtColsExit(Sender: TObject);
  procedure txtRowsKeyPress(Sender: TObject; var Key: Char);
  procedure txtColsKeyPress(Sender: TObject; var Key: Char);
  procedure txtRowsLoadedFromExit(Sender: TObject);
  procedure txtRowsLoadedFromKeyPress(Sender: TObject; var Key: Char);
  procedure txtRowsLoadedToExit(Sender: TObject);
  procedure txtRowsLoadedToKeyPress(Sender: TObject; var Key: Char);
  procedure treeTreeChanging(Sender: TObject; Node: TTreeNode; var AllowChange: Boolean);
  procedure treeTreeContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
  procedure treeTreeDeletion(Sender: TObject; Node: TTreeNode);
  procedure cmbDataTypeChange(Sender: TObject);
  procedure itemInsertChildClick(Sender: TObject);
  procedure itemInsertSiblingClick(Sender: TObject);
  procedure itemRenameClick(Sender: TObject);
  procedure itemCutClick(Sender: TObject);
  procedure itemCopyClick(Sender: TObject);
  procedure itemPasteChildClick(Sender: TObject);
  procedure itemPasteSiblingClick(Sender: TObject);
  procedure itemDeleteClick(Sender: TObject);
  procedure Button1Click(Sender: TObject);
  procedure btnUnloadClick(Sender: TObject);
  procedure trkSliderChange(Sender: TObject);
  procedure trkSliderKeyPress(Sender: TObject; var Key: Char);


protected


  grdData              :TQUB_Table;

  prRoot               :INode;
  prConstraints        :INode;

  prChangingRoot       :Boolean;
  prChangingSel        :Boolean;

  prEditTree           :Boolean;
  prEditData           :Boolean;
  prSizeData           :Boolean;
  prLoadData           :Boolean;
  prDataOnly           :Boolean;
  prEventsEveryKey     :Boolean;

  prDataConstrPopup    :TPopupMenu;
  prConstrR            :Integer;
  prConstrC            :Integer;


  procedure grdDataSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
  procedure grdDataSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);

  procedure prSetRoot(x :INode);
  procedure prSetConstraints(x :INode);

  procedure prSetEditTree(x :Boolean);
  procedure prSetEditData(x :Boolean);
  procedure prSetSizeData(x :Boolean);
  procedure prSetLoadData(x :Boolean);
  procedure prSetDataOnly(x :Boolean);
  procedure prSetEventsEveryKey(x :Boolean);

  function prGetSelected: INode;
  procedure prSetSelected(x :INode);
  procedure prSetSelectedConstraints(Constr :INode);

  function prFindNode(Root :TTreeNode; x :INode): TTreeNode;

  procedure prUpdateData(RowFrom, RowTo :Integer); overload;
  procedure prUpdateData; overload;

  procedure prOnConstrClick(Sender :TObject);


public


  OnSelectedChanged    :TQUB_NodeEditEvent;
  OnNameChanged        :TQUB_NodeEditEvent;
  OnCommentChanged     :TQUB_NodeEditEvent;
  OnDataTypeChanged    :TQUB_NodeEditEvent;
  OnDataBoundsChanged  :TQUB_NodeEditEvent;
  OnDataLoadChanged    :TQUB_NodeEditEvent;
  OnDataChanged        :TQUB_NodeEditEvent;
  OnSliderChanged      :TQUB_NodeEditEvent;
  OnSliderDefault      :TQUB_NodeEditEvent;


  constructor Create(AOwner :TComponent); override;


  property Root        :INode          read prRoot         write prSetRoot;
  property Constraints :INode          read prConstraints  write prSetConstraints;

  property Selected    :INode          read prGetSelected  write prSetSelected;


published


  property EditTree    :Boolean        read prEditTree     write prSetEditTree;
  property EditData    :Boolean        read prEditData     write prSetEditData;
  property SizeData    :Boolean        read prSizeData     write prSetSizeData;
  property LoadData    :Boolean        read prLoadData     write prSetLoadData;
  property DataOnly    :Boolean        read prDataOnly     write prSetDataOnly;
  property EventsEveryKey :Boolean     read prEventsEveryKey write prSetEventsEveryKey;


end;


  //   NodeEditor.Root := NewTree will attempt to keep the current node selected,
  //     only if current node is still in NewTree.


implementation


{$R *.DFM}


//******************************************************************************
constructor TQUB_NodeEditor.Create(AOwner :TComponent);
begin
  inherited;

  grdData                   := TQUB_Table.Create(Self);
  grdData.Parent            := panGridData;
  grdData.Align             := alClient;
  grdData.DefaultColWidth   := 40;
  grdData.DefaultRowHeight  := 16;
  grdData.FixedCols         := 0;
  grdData.FixedRows         := 0;
  grdData.ColCount          := 1;
  grdData.RowCount          := 1;
  grdData.Options           := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRangeSelect, goColSizing, goEditing, goTabs, goThumbTracking];
  grdData.TabOrder          := 0;
  grdData.OnSelectCell      := grdDataSelectCell;
  grdData.OnSetEditText     := grdDataSetEditText;

  prDataConstrPopup         := TPopupMenu.Create(Self);
  prDataConstrPopup.AutoHotkeys   := maManual;

  Root                      := QUB_Node_Null;

  EditTree                  := True;
  EditData                  := True;
  SizeData                  := True;
  LoadData                  := True;
  EventsEveryKey            := True;

  prChangingRoot            := False;
  prChangingSel             := False;
end;

procedure TQUB_NodeEditor.prUpdateData(RowFrom, RowTo :Integer);
var
  OldLoad, NewLoad       :TQUB_FirstLast;
  c, i, j, x, w          :Integer;
  iGrid                  :Integer;
  Node                   :INode;
  IsFloat                :Boolean;
begin
  try

    Node                   := Selected;
    cmbDataType.ItemIndex  := Integer(Node.DataType);
    try
      txtData.OnChange     := nil;
      txtData.Text         := '';
      case Node.DataType of
        NODE_DATATYPE_EMPTY,
        NODE_DATATYPE_UNKNOWN:
        begin
          panEditNone.BringToFront;
          Exit;
        end;
        NODE_DATATYPE_STRING:
        begin
          txtData.Text       := Node.DataAsString;
          panEditString.BringToFront;
          Exit;
        end;
      else
        panEditNum.BringToFront;
      end;
    finally
      txtData.OnChange     := txtDataChange;
    end;

    NewLoad                := Node.LoadRowBounds;

    txtRows.Text           := IntToStr(Node.Rows);
    txtCols.Text           := IntToStr(Node.Cols);
    txtRowsLoadedFrom.Text := IntToStr(NewLoad.First);
    txtRowsLoadedTo.Text   := IntToStr(NewLoad.Last);

    NewLoad.Last           := Min(NewLoad.First + QUB_NODEEDITOR_MAXROWS - 1, NewLoad.Last);

    if NewLoad.First < 0 then begin
      grdData.RowCount     := 0;
      grdData.ColCount     := 0;
      Exit;
    end;

    if grdData.RowCount < 2 then begin
      grdData.RowCount     := 1;
      grdData.ColCount     := 1;

      OldLoad.First        := NewLoad.First;
      OldLoad.Last         := NewLoad.First - 1;
    end else begin
      OldLoad.First        := StrToInt(grdData.Cells[0,1]);
      OldLoad.Last         := OldLoad.First + grdData.RowCount - 2;
    end;

    if OldLoad.First < NewLoad.First then begin
      if (NewLoad.First - OldLoad.First) < 16 then begin
        for i := 0 to NewLoad.First - OldLoad.First - 1 do begin
          grdData.DeleteRow(1);
        end;
      end else begin
        grdData.RowCount   := 0;
        grdData.ColCount   := 0;
        OldLoad.First      := NewLoad.First;
        OldLoad.Last       := NewLoad.First - 1;
      end;
    end else if NewLoad.First < OldLoad.First then begin
      if (OldLoad.First - NewLoad.First) < 16 then begin
        for i := 0 to OldLoad.First - NewLoad.First - 1 do begin
          grdData.InsertRow(1);
        end;
      end else begin
        grdData.RowCount   := 0;
        grdData.ColCount   := 0;
        OldLoad.First      := NewLoad.First;
        OldLoad.Last       := NewLoad.First - 1;
      end;
    end;

    grdData.RowCount := grdData.RowCount + NewLoad.Last - OldLoad.Last;

    x                      := grdData.ColCount - 1;
    grdData.ColCount       := Node.Cols + 1;
    if x < LongInt(Node.Cols) then begin
      for i := x to Node.Cols - 1 do begin
        grdData.Cells[i+1, 0] := IntToStr(i);
      end;
    end;

    grdData.Cells[0,0]     := '';

    RowFrom                := Max(RowFrom, NewLoad.First);
    RowTo                  := Min(RowTo,   NewLoad.Last);

    IsFloat                := False;
    case Node.DataType of
      NODE_DATATYPE_SINGLE,
      NODE_DATATYPE_DOUBLE,
      NODE_DATATYPE_EXTENDED:    IsFloat       := True;
    end;

    for i := RowFrom to RowTo do begin
      iGrid                := i - NewLoad.First + 1;
      grdData.Cells[0, iGrid] := IntToStr(i);

      if IsFloat then begin
        for j := 0 to Node.Cols - 1 do begin
          grdData.Cells[j+1, iGrid] := FloatToStr(Double(Node.MatrixData[i, j]));
        end;
      end else begin
        for j := 0 to Node.Cols - 1 do begin
          grdData.Cells[j+1, iGrid] := IntToStr(Integer(Node.MatrixData[i, j]));
        end;
      end;
    end;

    //grdData.ColWidths[0] := 20;//40;
    grdData.FixedRows    := 1;
    grdData.FixedCols    := 1;

  finally

    c  := grdData.ColCount - 1;
    if c > 0 then begin
    w  := (grdData.ClientWidth - grdData.ColWidths[0] - c) div c;
    for i := 1 to c do begin
      grdData.ColWidths[i] := w;
    end;
    end;

  end;

end;

procedure TQUB_NodeEditor.prUpdateData;
begin
  prUpdateData(Max(0,Selected.LoadRowBounds.First), Selected.LoadRowBounds.Last);
end;

procedure SetTreeViewFromQUBTree(treeTree :TTreeView; q :INode);
var
  v           :TTreeNode;
  Entry       :TQUB_NodeEditorEntry;

  procedure AddSubtree(qsub :INode; vsubprev :TTreeNode; AsChild :Boolean);
  var
    vsub         :TTreeNode;
    qsubsub      :INode;
  begin
    if qsub.Name = 'QTR_LINE_COMMENT' then begin
      qsub         := qsub.Sibling;
      if qsub.IsNull then Exit;
    end;

    Entry      := TQUB_NodeEditorEntry.Create;
    Entry.Node := qsub;
    Entry.Constr := QUB_Node_Null;

    if AsChild then begin
      vsub       := treeTree.Items.AddChildObject(vsubprev, qsub.Name, Entry);
    end else begin
      vsub       := treeTree.Items.AddObject(vsubprev, qsub.Name, Entry);
    end;

    qsubsub      := qsub.Child;
    if not qsubsub.IsNull then AddSubtree(qsubsub, vsub, True);
    qsubsub      := qsub.Sibling;
    if not qsubsub.IsNull then AddSubtree(qsubsub, vsub, False);
  end;
begin
  treeTree.Items.BeginUpdate;
  try
    treeTree.Items.Clear;
    if not q.IsNull then begin
      Entry      := TQUB_NodeEditorEntry.Create;
      Entry.Node := q;
      Entry.Constr := QUB_Node_Null;
      v          := treeTree.Items.AddObject(nil, q.Name, Entry);
      if not q.Child.IsNull then AddSubtree(q.Child, v, True);
      v.Expanded := True;
    end;
  finally
    treeTree.Items.EndUpdate;
  end;
end;

procedure TQUB_NodeEditor.prSetRoot(x :INode);
var
  ReSelected             :INode;
begin
  ReSelected             := Selected;
  Selected               := QUB_Node_Null;
  prRoot                 := x;
  prConstraints          := QUB_Node_Null;
  prChangingRoot         := True;
  try
    SetTreeViewFromQUBTree(treeTree, x);
  finally
    prChangingRoot         := False;
  end;
  Selected               := ReSelected;
end;

procedure TQUB_NodeEditor.prSetConstraints(x :INode);
  procedure SetConstraints(q :INode; v :TTreeNode);
  var
    Entry        :TQUB_NodeEditorEntry;
    i            :Integer;
    qsub         :INode;
  begin
    Entry        := TQUB_NodeEditorEntry( v.Data );
    Entry.Constr := q;

    i            := 0;
    qsub         := q.Child;
    while i < v.Count do begin
      SetConstraints(qsub, v[i]);
      Inc(i);
      qsub       := qsub.Sibling;
    end;
  end;
begin
  SetConstraints(x, treeTree.Items[0]);
end;

procedure TQUB_NodeEditor.prSetEditTree(x :Boolean);
begin
  prEditTree                := x;
end;

procedure TQUB_NodeEditor.prSetEditData(x :Boolean);
begin
  prEditData             := x;
  cmbDataType.Enabled    := x;
  grdData.Enabled        := x;
  prSetSizeData( prSizeData );
end;

procedure TQUB_NodeEditor.prSetSizeData(x :Boolean);
begin
  prSizeData             := x;
  txtRows.Enabled        := x and prEditData;
  txtCols.Enabled        := x and prEditData;
end;

procedure TQUB_NodeEditor.prSetLoadData(x :Boolean);
begin
  prLoadData                 := x;
  txtRowsLoadedFrom.Enabled  := x;
  txtRowsLoadedTo.Enabled    := x;
end;

procedure TQUB_NodeEditor.prSetDataOnly(x :Boolean);
begin
  prDataOnly                 := not x;
  panMeta.Visible            := not x;
  panRowsCols.Visible        := not x;
  panRowsLoaded.Visible      := not x;
end;

procedure TQUB_NodeEditor.prSetEventsEveryKey(x :Boolean);
begin
  prEventsEveryKey           := x;
end;

function TQUB_NodeEditor.prGetSelected: INode;
begin
  if treeTree.Selected <> nil then begin
    Result := TQUB_NodeEditorEntry(treeTree.Selected.Data).Node;
  end else begin
    Result := TNode.CreateNull;
  end;
end;

function TQUB_NodeEditor.prFindNode(Root :TTreeNode; x :INode): TTreeNode;
var
  i                     :Integer;
begin
  if TQUB_NodeEditorEntry(Root.Data).Node.Equals( x ) then begin
    Result              := Root;
    Exit;
  end;

  for i := 0 to Root.Count - 1 do begin
    Result              := prFindNode( Root[i], x );
    if Result <> nil then Exit;
  end;

  Result                := nil;
end;

procedure TQUB_NodeEditor.prOnConstrClick(Sender :TObject);
begin
  case Selected.DataType of
    NODE_DATATYPE_STRING:
      begin
        Selected.SetData( TMenuItem(Sender).Caption );
        prUpdateData;
        if Assigned(OnDataChanged) then OnDataChanged(Self, Selected, 0, 0);
      end;
    NODE_DATATYPE_SINGLE,
    NODE_DATATYPE_DOUBLE,
    NODE_DATATYPE_EXTENDED:
      begin
        Selected.MatrixData[ prConstrR, prConstrC ] := StrToFloat( TMenuItem(Sender).Caption );
        prUpdateData( prConstrR, prConstrR );
        if Assigned(OnDataChanged) then OnDataChanged(Self, Selected, prConstrR, prConstrC);
      end;
  else begin
    Selected.MatrixData[ prConstrR, prConstrC ] := StrToInt( TMenuItem(Sender).Caption );
    prUpdateData( prConstrR, prConstrR );
    if Assigned(OnDataChanged) then OnDataChanged(Self, Selected, prConstrR, prConstrC);
  end end;
end;

function StringToLines(s :PChar; Lines :TStrings): Integer;
var
  endl                  :PChar;
  len                   :Integer;
  Line                  :String;
begin
  Result                := 0;

  while s^ <> #0 do begin
    // endl: one past the end
    endl                    := StrScan(s, #10);
    if endl = nil then endl := StrScan(s, #0);
    if (endl-1)^ = #13 then Dec(endl);

    len                     := endl - s;
    SetLength(Line, len);
    StrLCopy(PChar(Line), s, len);
    PChar(Line)[len]        := #0;
    Lines.Add(Copy(Line, 0, len));
    Inc(Result);

    if endl^ = #0 then begin
      s                     := endl;
    end else begin
      s                     := endl + 1;
      while (s^ = #13) or (s^ = #10) do Inc(s);
    end;
  end;
end;

procedure TQUB_NodeEditor.prSetSelectedConstraints(Constr :INode);
var
  ConstrItem            :TMenuItem;
  i                     :Integer;
  Lines                 :TStringList;
begin
  prDataConstrPopup.Items.Clear;
  txtData.ReadOnly      := False;

  if Constr.DataCount = 0 then Exit;

  case Constr.DataType of
    NODE_DATATYPE_EMPTY,
    NODE_DATATYPE_UNKNOWN:
      ;
    NODE_DATATYPE_STRING:
    begin
      Lines             := TStringList.Create;


        //TO DO: Chris, unless you mean it, you can use simply:
        //  Lines.Text := Constr.DataAsString; LM.
      StringToLines(PChar(Constr.DataAsString), Lines);


      for i := 0 to Lines.Count - 1 do begin
        ConstrItem         := TMenuItem.Create(prDataConstrPopup);
        ConstrItem.Caption := Lines[i];
        ConstrItem.OnClick := prOnConstrClick;
        prDataConstrPopup.Items.Add(ConstrItem);
      end;
      txtData.ReadOnly   := True;
      Lines.Free;
    end;
    NODE_DATATYPE_SINGLE,
    NODE_DATATYPE_DOUBLE,
    NODE_DATATYPE_EXTENDED:
    begin
      for i := 0 to Constr.DataCount - 1 do begin
        ConstrItem         := TMenuItem.Create(prDataConstrPopup);
        ConstrItem.Caption := FloatToStr(Double(Constr.Data[i]));
        ConstrItem.OnClick := prOnConstrClick;
        prDataConstrPopup.Items.Add(ConstrItem);
      end;
    end;
  else
    for i := 0 to Constr.DataCount - 1 do begin
      ConstrItem         := TMenuItem.Create(prDataConstrPopup);
      ConstrItem.Caption := IntToStr(Integer(Constr.Data[i]));
      ConstrItem.OnClick := prOnConstrClick;
      prDataConstrPopup.Items.Add(ConstrItem);
    end;
  end;//case DataType
end;

procedure TQUB_NodeEditor.prSetSelected(x :INode);
var
  TreeNode              :TTreeNode;
  y                     :INode;
  xInRoot               :Boolean;
begin
  if x = nil then x := QUB_Node_Null;
  
  y                     := Selected;
  if not y.IsNull then begin
    // OnExit is so unreliable
    if txtComment.Text <> y.LineComment then y.LineComment := txtComment.Text;
    if (y.DataType = NODE_DATATYPE_STRING) and (y.DataAsString <> txtData.Text) then begin
      y.SetData(txtData.Text);
    end;
    grdData.RowCount  := 0;
    grdData.ColCount  := 0;
  end;

  prChangingSel         := True;

  txtComment.Text       := '';
  cmbDataType.ItemIndex := 0;

  xInRoot               := False;
  y                     := x.Parent;
  while not y.IsNull do begin
    if y.Equals(Root) then begin
      xInRoot           := True;
      Break;
    end else begin
      y                 := y.Parent;
    end;
  end;

  if (not xInRoot) or (treeTree.Items.Count = 0) or x.IsNull then begin

    treeTree.Selected     := nil;
    txtComment.Enabled    := False;
    cmbDataType.Enabled   := False;
    panEditNone.BringToFront;

    prSetSelectedConstraints( QUB_Node_Null );

  end else begin

    TreeNode            := prFindNode(treeTree.Items[0], x);

    treeTree.Selected     := TreeNode;
    txtComment.Text       := x.LineComment;
    txtComment.Enabled    := True;
    cmbDataType.Enabled   := prEditData and TQUB_NodeEditorEntry(TreeNode.Data).Constr.IsNull;
    cmbDataType.ItemIndex := Integer(x.DataType);
    prUpdateData;

    prSetSelectedConstraints(TQUB_NodeEditorEntry(TreeNode.Data).Constr);

  end;

  prChangingSel      := False;

  itemInsertChild.Visible   := prEditTree and not x.IsNull;
  itemInsertSibling.Visible := prEditTree and not (x.IsNull or x.Equals(Root));
  itemRename.Visible        := prEditTree and not x.IsNull;
  itemDelete.Visible        := prEditTree and not x.IsNull;

  if Assigned(OnSelectedChanged) then begin
    OnSelectedChanged(Self, x, 0, 0);
  end;
end;


// *****************************************************************************
procedure TQUB_NodeEditor.treeTreeChanging(Sender: TObject; Node: TTreeNode; var AllowChange: Boolean);
begin
  if prChangingSel then Exit;
  Selected := TQUB_NodeEditorEntry(Node.Data).Node;
end;

procedure TQUB_NodeEditor.treeTreeContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
var
  Pointed :TTreeNode;
begin
  Pointed          := treeTree.GetNodeAt(MousePos.X, MousePos.Y);
  if Pointed <> nil then treeTree.Selected := Pointed;
  if Selected.IsNull then Exit;
  MousePos   := ClientToScreen(MousePos);
  popupTree.Popup(MousePos.X, MousePos.Y);
  Handled := True;
end;

procedure TQUB_NodeEditor.treeTreeDeletion(Sender: TObject; Node: TTreeNode);
var
  Entry :TQUB_NodeEditorEntry;
begin
  if prChangingRoot then Exit;

  try
    Entry      := TQUB_NodeEditorEntry(Node.Data);
    if not (csDestroying in treeTree.ComponentState) then begin
      Entry.Node.Parent.RemoveChild(Entry.Node);
    end;
    Entry.Free;
  except
  end;
end;

procedure TQUB_NodeEditor.itemInsertChildClick(Sender: TObject);
var
  Parent        :INode;
  Name          :String;
  Entry         :TQUB_NodeEditorEntry;
begin
  Parent        := Selected;
  if Parent.IsNull then Exit;
  Name          := InputBox('Insert Child', 'Child name:', '');
  if Name <> '' then begin
    Entry        := TQUB_NodeEditorEntry.Create;
    Entry.Node   := Parent.InsertChild(QUB_Node_Null, Name);
    Entry.Constr := QUB_Node_Null;
    treeTree.Items.AddChildObjectFirst(treeTree.Selected, Name, Entry);
  end;
  treeTree.Selected.Expanded := True;
end;

procedure TQUB_NodeEditor.itemInsertSiblingClick(Sender: TObject);
var
  PreSib        :INode;
  Name          :String;
  Parent        :TTreeNode;
  Index         :Integer;
  Entry         :TQUB_NodeEditorEntry;
begin
  PreSib         := Selected;
  if PreSib.IsNull or (PreSib.Equals(Root)) then Exit;
  Name           := InputBox('Insert Sibling', 'Sibling name:', '');
  if Name <> '' then begin
    Parent       := treeTree.Selected.Parent;
    Index        := treeTree.Selected.Index + 1;
    Entry        := TQUB_NodeEditorEntry.Create;
    Entry.Node   := PreSib.Parent.InsertChild(PreSib, Name);
    Entry.Constr := QUB_Node_Null;
    if Index = Parent.Count then begin
      treeTree.Items.AddObject(treeTree.Selected, Name, Entry);
    end else begin
      treeTree.Items.InsertObject(Parent[Index], Name, Entry);
    end;
  end;
end;

procedure TQUB_NodeEditor.itemRenameClick(Sender: TObject);
var
  NewName           :String;
begin
  if not Selected.IsNull then begin
    NewName := InputBox('Rename Node', 'New name:', Selected.Name);
    if NewName <> '' then begin
      Selected.Name          := NewName;
      treeTree.Selected.Text := NewName;
      if Assigned(OnNameChanged) then OnNameChanged(Self, Selected, 0, 0);
    end;
  end;
end;

procedure TQUB_NodeEditor.itemCutClick(Sender: TObject);
begin
  //
end;

procedure TQUB_NodeEditor.itemCopyClick(Sender: TObject);
begin
  //
end;

procedure TQUB_NodeEditor.itemPasteChildClick(Sender: TObject);
begin
  //
end;

procedure TQUB_NodeEditor.itemPasteSiblingClick(Sender: TObject);
begin
  //
end;

procedure TQUB_NodeEditor.itemDeleteClick(Sender: TObject);
begin
  treeTree.Items.Delete( treeTree.Selected );
end;


procedure TQUB_NodeEditor.txtCommentKeyPress(Sender: TObject;
  var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtCommentExit(Sender);
  end;
end;

procedure TQUB_NodeEditor.txtCommentExit(Sender: TObject);
begin
  if Selected.LineComment <> txtComment.Text then begin
    Selected.LineComment := txtComment.Text;
    if Assigned(OnCommentChanged) then OnCommentChanged(Self, Selected, 0, 0);
  end;
end;

procedure TQUB_NodeEditor.cmbDataTypeChange(Sender: TObject);
var
  NewType             :TQUB_QFS_Node_DataType;
  Node, Clone         :INode;
  i                   :Integer;
begin
  if prChangingSel then Exit;
  NewType             := TQUB_QFS_Node_DataType(cmbDataType.ItemIndex);
  Node                := Selected;

  grdData.RowCount  := 0;
  grdData.ColCount  := 0;

  case NewType of
    NODE_DATATYPE_EMPTY:
    begin
      Node.ClearData;
      panEditNone.BringToFront;
    end;
    NODE_DATATYPE_UNKNOWN:
    begin
      case Node.DataType of
        NODE_DATATYPE_EMPTY:
        begin
          Node.SetupData(NODE_DATATYPE_UNKNOWN, 4, 1);
        end;
        else begin
          Clone     := Node.Clone(False);
          Node.SetupData(NODE_DATATYPE_UNKNOWN, Node.RowSize, Node.Rows);
          Move(Clone.DataPtr^, Node.DataPtr^, Node.RowSize * Node.Rows);
        end;
      end;
    end;
    NODE_DATATYPE_STRING:
    begin
      if (Node.DataType = NODE_DATATYPE_EMPTY) or (Node.DataType = NODE_DATATYPE_UNKNOWN) then begin
        Node.SetData(' ');
      end else begin
        Clone        := Node.Clone(False);
        Node.SetupData(NODE_DATATYPE_STRING, 1, Node.Rows * Node.Cols);
        for i := 0 to Clone.DataCount - 1 do begin
          Node.Data[i] := Clone.Data[i];
        end;
      end;
    end;
    else begin
      case Node.DataType of
        NODE_DATATYPE_EMPTY,
        NODE_DATATYPE_UNKNOWN:
        begin
          Node.SetupNumData(NewType, 1, 1);
        end;
        else begin
          Clone        := Node.Clone(False);
          Node.SetupNumData(NewType, Clone.Rows, Clone.Cols);
          for i := 0 to Clone.DataCount - 1 do begin
            Node.Data[i] := Clone.Data[i];
          end;
        end;
      end;
    end;
  end;

  prUpdateData;
  if Assigned(OnDataTypeChanged) then OnDataTypeChanged(Self, Node, Integer(NewType), Integer(NewType));
end;

procedure TQUB_NodeEditor.txtDataChange(Sender: TObject);
begin
  Selected.SetData(txtData.Text);
  if Assigned(OnDataChanged) and (prEventsEveryKey) then OnDataChanged(Self, Selected, 0, 0);
end;

procedure TQUB_NodeEditor.txtDataMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Pt                :TPoint;
begin
  if prDataConstrPopup.Items.Count > 0 then begin
    Pt.X            := X;
    Pt.Y            := Y;
    Pt              := grdData.ClientToScreen( Pt );
    prDataConstrPopup.Popup(Pt.X,Pt.Y);
  end;
end;

procedure TQUB_NodeEditor.txtDataExit(Sender: TObject);
begin
  Selected.SetData(txtData.Text);
  if Assigned(OnDataChanged) then OnDataChanged(Self, Selected, 0, 0);
end;

procedure TQUB_NodeEditor.txtRowsKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtRowsExit(Sender);
  end;
end;

procedure TQUB_NodeEditor.txtRowsExit(Sender: TObject);
var
  Node                 :INode;
  OldRows, NewRows     :Integer;
begin
  try
    Node               := Selected;
    OldRows            := Node.Rows;
    NewRows            := StrToInt( txtRows.Text );
    if NewRows = 0 then begin
      // simulate datatype := none
      cmbDataType.ItemIndex := Integer(NODE_DATATYPE_EMPTY);
      cmbDataTypeChange(cmbDataType);
    end else if NewRows < 1 then begin
      raise Exception.Create('');
    end else if NewRows = OldRows then begin
      Exit;
    end else begin
      Node.ResizeData(NewRows);
      prUpdateData( Min(OldRows, NewRows), Max(OldRows, NewRows)-1 );
      if Assigned(OnDataBoundsChanged) then OnDataBoundsChanged(Self, Node, Node.Rows, Node.Cols);
    end;
  except
    txtRows.Text    := IntToStr( Selected.Rows );
  end;
end;

procedure TQUB_NodeEditor.txtColsKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtColsExit(Sender);
  end;
end;

procedure TQUB_NodeEditor.txtColsExit(Sender: TObject);
var
  Node, Clone       :INode;
  i, j              :Integer;
  NewCols           :Integer;
begin
  try
    NewCols         := StrToInt( txtCols.Text );
    if NewCols < 1 then raise Exception.Create('');
    if NewCols = LongInt(Selected.Cols) then Exit;

    Node             := Selected;
    Clone            := Node.Clone(False);
    Node.SetupNumData( Clone.DataType, Clone.Rows, NewCols );
    for i := 0 to Clone.Rows - 1 do begin
      for j := 0 to Min(Clone.Cols, NewCols) - 1 do begin
        Node.MatrixData[i,j] := Clone.MatrixData[i,j];
      end;
    end;

    prUpdateData;
    if Assigned(OnDataBoundsChanged) then OnDataBoundsChanged(Self, Node, Node.Rows, Node.Cols);

  except
    txtCols.Text    := IntToStr( Selected.Cols );
  end;
end;

procedure TQUB_NodeEditor.grdDataSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
var
  Rect              :TRect;
  Pt                :TPoint;
begin
  CanSelect         := (ARow > 0) and (ACol > 0);
  if CanSelect and (prDataConstrPopup.Items.Count > 0) then begin
    CanSelect       := False;
    prConstrR       := StrToInt(grdData.Cells[0,ARow]);
    prConstrC       := ACol - 1;

    Rect := grdData.CellRect(ACol, ARow);
    Pt   := grdData.ClientToScreen(Rect.TopLeft);
    prDataConstrPopup.Popup(Pt.X, Pt.Y);
  end;
end;

procedure TQUB_NodeEditor.grdDataSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
var
  Node          :INode;
  R, C          :Integer;
  CanSelect     :Boolean;
begin

  if (not EventsEveryKey) and grdData.EditorMode then Exit;

  if (ARow = 0) or (ACol = 0) then Exit;
  if (Value = '') or (Value = '.') or (Value = '-') or (Value = '-.') then Exit;

  Node          := Selected;
  R             := Node.LoadRowBounds.First + ARow - 1;
  C             := ACol - 1;

  if prDataConstrPopup.Items.Count > 0 then begin     // oops you can't edit this
    grdData.Cells[ACol, ARow] := FloatToStr(Double(Node.MatrixData[R,C]));
    grdDataSelectCell(Sender, ACol, ARow, CanSelect);
    Exit;
  end;

  try
    case Node.DataType of
      NODE_DATATYPE_SINGLE,
      NODE_DATATYPE_DOUBLE,
      NODE_DATATYPE_EXTENDED:
        begin
          Node.MatrixData[R,C] := StrToFloat(Value);
        end;
      else
        begin
          Node.MatrixData[R,C] := StrToInt(Value);
        end;
    end;//case
    if Assigned(OnDataChanged) then OnDataChanged(Self, Selected, R, C);
  except
  end;

end;


procedure TQUB_NodeEditor.txtRowsLoadedFromKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtRowsLoadedFromExit(Sender);
  end;
end;

procedure TQUB_NodeEditor.txtRowsLoadedFromExit(Sender: TObject);
var
  NewSel                  :TQUB_FirstLast;
begin
  try
    NewSel.First          := StrToInt( txtRowsLoadedFrom.Text );
    if (NewSel.First = -1) and (Selected.LoadRowBounds.First = -1) then Exit;

    if NewSel.First < 0 then NewSel.First := 0;
    if NewSel.First >= LongInt(Selected.Rows) then NewSel.First := LongInt(Selected.Rows) - 1;
    if NewSel.Last < NewSel.First then NewSel.Last := NewSel.First;

    Selected.LoadRows(NewSel, True);
    prUpdateData;
    if Assigned(OnDataLoadChanged) then OnDataLoadChanged(Self, Selected, NewSel.First, NewSel.Last);
  except
    txtRowsLoadedFrom.Text    := IntToStr( Selected.LoadRowBounds.First );
  end;
end;

procedure TQUB_NodeEditor.txtRowsLoadedToKeyPress(Sender: TObject;
  var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtRowsLoadedToExit(Sender);
  end;
end;

procedure TQUB_NodeEditor.txtRowsLoadedToExit(Sender: TObject);
var
  NewSel                  :TQUB_FirstLast;
begin
  try
    NewSel.Last           := StrToInt( txtRowsLoadedTo.Text );
    if (NewSel.Last = -1) and (Selected.LoadRowBounds.Last = -1) then Exit;

    if NewSel.Last < 0 then NewSel.Last := 0;
    if NewSel.Last >= LongInt(Selected.Rows) then NewSel.Last := LongInt(Selected.Rows) - 1;
    if NewSel.First > NewSel.Last then NewSel.First := NewSel.Last;

    Selected.LoadRows(NewSel, True);
    prUpdateData;
    if Assigned(OnDataLoadChanged) then OnDataLoadChanged(Self, Selected, NewSel.First, NewSel.Last);
  except
    txtRowsLoadedTo.Text       := IntToStr( Selected.LoadRowBounds.Last );
  end;
end;

procedure TQUB_NodeEditor.btnUnloadClick(Sender: TObject);
begin
  Selected.UnloadRows(True);
  prUpdateData;
  if Assigned(OnDataLoadChanged) then OnDataLoadChanged(Self, Selected, -1, -1);
end;

// >>> QUB.Report(QUB.editTree(QUB.getModel()))
// >>> QUB.editTree(QUB.ResultsView.currResults.node)

procedure TQUB_NodeEditor.Button1Click(Sender: TObject);
begin
  Console.NewEntry(Root.AsString, COLOR_TEXT);
end;

procedure TQUB_NodeEditor.trkSliderChange(Sender: TObject);
begin
  if Assigned(OnSliderChanged) then OnSliderChanged(Sender, Selected, 0, 0);
end;

procedure TQUB_NodeEditor.trkSliderKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = Char(VK_ESCAPE) then begin
    if Assigned(OnSliderDefault) then OnSliderDefault(Sender, Selected, 0, 0);
  end;
end;


//******************************************************************************
end.

