{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// qubtree and python interface on the model

unit QUB_Node_Model;

interface


uses QUB_QFS, QUB_Node,

QUB_QFS_QMF,
PythonEngine, QUB_Python_Thread


;

function GetModelAsNode: INode;
procedure SetModelFromNode( ModelObject :TQUB_QFS_QMF; ModelNode :INode );
procedure SetRatesFromNode( ModelObject :TQUB_QFS_QMF; RatesNode :INode ); // just k0 and k1 for now


type

TGetModel = class(TPyObject)
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;

TSetModel = class(TPyObject)
protected
  prModelNode          :INode;
  procedure prRun;
public
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;

TSetRates = class(TPyObject)
protected
  prRatesNode          :INode;
  procedure prRun;
public
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;


procedure InitModelScripts;


implementation


uses


Math, IEEE754,

QUB_GUIInterface,
QUB_Viewport_Model, QUB_QFS_Channels,
QUB_Viewport_Python, QUB_Node_Python,
QUB_QFS_Node;


var


GetModelType, SetModelType, SetRatesType     :TPythonType;


function GetModelAsNode: INode;
var
  ModelObject            :TQUB_QFS_QMF;
begin
  try
    ModelObject := ViewportModel.CurrModel.ModelObject;
    Result      := ModelObject.CloneAsNode(True, TNode.CreateNull);
  except
    Result      := TNode.CreateNull;
  end;
end;

function GetModelAsQFSNode: INode;
begin
  try
    Result      := QUB_QFS_INode(ViewportModel.CurrModel.ModelObject);
  except
    Result      := TNode.CreateNull;
  end;
end;

procedure SetModelFromNode( ModelObject :TQUB_QFS_QMF; ModelNode :INode );
var
  i, j, c                :Integer;
  node, node2            :INode;
  x, y                   :Double;
  Constraint             :TQUB_QFS_Constraint;
  ConstraintAV           :TQUB_QFS_ConstraintAmpVar;
begin
  try
    ModelObject.Delete;

    // if States not null:
    //   foreach State:
    //     add                                          you know control structures are good
    //     for prop in [x, y, class, pr]:               when you use 'em in pseudocode
    //       if have prop: set prop                     (for x in lst; block by indent and colon)
    node := ModelNode.Find('States');
    if not node.IsNull then begin
      node := node.Child;
      i := 0;
      while not node.IsNull do begin
        x := 0.0;     node2 := node.Find('x');     if not node2.IsNull then x := node2.Data[0];
        y := 0.0;     node2 := node.Find('y');     if not node2.IsNull then y := node2.Data[0];
        ModelObject.AddState( x, y );

        node2 := node.Find('Class');
        if not node2.IsNull then begin
          c := Integer(node2.Data[0]);
          if c = ModelObject.ClassCount then begin
            ModelObject.Amp[c] := 0.0;
          end;
          ModelObject.States[i].Cls := c;
        end;

        node2 := node.Find('Pr');
        if not node2.IsNull then
          ModelObject.States[i].Pr := node2.Data[0];

        node2 := node.Find('Group');
        if not node2.IsNull then
          ModelObject.States[i].Gr := node2.Data[0];

        node := node.Sibling;
        Inc(i);
      end;

      ModelObject.States.UpdateStates;
    end;

    // if Rates not null:
    //   foreach Rate:
    //     add, set props [States, k0, k1, dk0, dk1, P, Q, PNames, QNames] as above
    node := ModelNode.Find('Rates');
    if not node.IsNull then begin
      node := node.Child;
      while not node.IsNull do begin
        node2 := node['States'];
        ModelObject.AddRate( ModelObject.States[ Integer(node2.Data[0]) ],
                             ModelObject.States[ Integer(node2.Data[1]) ] );
        node := node.Sibling;
      end;

      SetRatesFromNode( ModelObject, ModelNode['Rates'] );
    end;

    // if Constraints not null:
    //   add each
    node := ModelNode.Find('Constraints');
    if not node.IsNull then begin
      node := node.Child;
      while not node.IsNull do begin
        Constraint := TQUB_QFS_Constraint.Create(node.Name, nil);
        ModelObject.Constraints.AddConstraint(Constraint, ModelObject.Constraints.ConstraintCount);
        for j := 0 to node.DataCount - 1 do begin
          Constraint.Index[j] := Integer( node.Data[j] );
        end;
        if (node.Find('HasValue').DataCount > 0) and (node.Find('HasValue').Data[0]<>0) and (node.Find('Value').DataCount > 0) then begin
          Constraint.HasValue := True;
          Constraint.Value    := node['Value'].Data[0];
        end;

        node := node.Sibling;
      end;
      SetRatesFromNode( ModelObject, ModelNode['Rates'] );
    end;

    // if ConstraintsAmpVar not null:
    //   add each
    node := ModelNode.Find('ConstraintsAmpVar');
    if not node.IsNull then begin
      node := node.Child;
      while not node.IsNull do begin
        ConstraintAV := TQUB_QFS_ConstraintAmpVar.Create(node.Name, nil);
        ModelObject.ConstraintsAmpVar.AddConstraint(ConstraintAV, ModelObject.ConstraintsAmpVar.ConstraintCount);
        for j := 0 to node.DataCount - 1 do begin
          ConstraintAV.Index[j] := Integer( node.Data[j] );
        end;

        node := node.Sibling;
      end;
    end;

    // if ChannelCount: set
    node := ModelNode.Find('ChannelCount');
    if not node.IsNull then begin
      ModelObject.ChannelCount := Integer(node.Data[0]);
    end;

    // for x in [Amp, Std, Nar]:
    //   if x not null:
    //     len = min( len(x), maxclass )
    //     for i in range(0,len): set x[i]
    node := ModelNode.Find('Amps');
    if not node.IsNull then begin
      for i := 0 to node.DataCount - 1 do begin
        if not (IsNan(node.Data[i]) or IsInfinity(node.Data[i])) then begin
          ModelObject.Amp[i] := node.Data[i];
        end;
      end;
    end;

    node := ModelNode.Find('Stds');
    if not node.IsNull then begin
      for i := 0 to node.DataCount - 1 do begin
        if not (IsNan(node.Data[i]) or IsInfinity(node.Data[i])) then begin
          ModelObject.Std[i] := node.Data[i];
        end;
      end;
    end;

    node := ModelNode.Find(CHANNELS_CONDS_NAME);
    if not node.IsNull then begin
      for i := 0 to node.DataCount - 1 do begin
        if not (IsNan(node.Data[i]) or IsInfinity(node.Data[i])) then begin
          ModelObject.Cond[i] := node.Data[i];
        end;
      end;
    end;

    node := ModelNode.Find(CHANNELS_CONDSTDS_NAME);
    if not node.IsNull then begin
      for i := 0 to node.DataCount - 1 do begin
        if not (IsNan(node.Data[i]) or IsInfinity(node.Data[i])) then begin
          ModelObject.CondStd[i] := node.Data[i];
        end;
      end;
    end;

    node := ModelNode.Find(CHANNELS_VREVS_NAME);
    if not node.IsNull then begin
      for i := 0 to node.DataCount - 1 do begin
        if not (IsNan(node.Data[i]) or IsInfinity(node.Data[i])) then begin
          ModelObject.ClsVrev[i] := node.Data[i];
        end;
      end;
    end;

    node := ModelNode.Find('NAr');
    if not node.IsNull then begin
      for i := 0 to node.DataCount - 1 do begin
        ModelObject.NAr[i] := node.Data[i];
      end;
    end;

    // if Ars not null:
    //   foreach Ar:
    //     set from 0 to min( len(Ar), maxnar )
    node := ModelNode.Find('Ars');
    if not node.IsNull then begin
      node := node.Child;
      i    := 0;

      while not node.IsNull do begin
        for j := Min( node.DataCount, QUB_QFS_QMF_MAX_COEFF ) - 1 downto 0 do begin
          ModelObject.Cff[i,j] := node.Data[j];
        end;

        node := node.Sibling;
        Inc(i);
      end;
    end;

  except
  end;

end;

procedure SetRatesFromNode( ModelObject :TQUB_QFS_QMF; RatesNode :INode );
var
  Rates                  :TQUB_QFS_Rates;
  Rate                   :TQUB_QFS_Rate;
  i                      :Integer;
  RateNode               :INode;
  CondName               :INode;
begin
  try
    Rates       := ModelObject.Rates;

    RateNode    := RatesNode.Find('Rate');
    for i := 0 to Rates.RateCount - 1 do begin
      Rate       := Rates[i];

      if RateNode['k0'].DataCount >= 2 then begin
        Rate.K0[1] := Double( RateNode['k0'].Data[0] );
        Rate.K0[2] := Double( RateNode['k0'].Data[1] );
      end;
      if RateNode['k1'].DataCount >= 2 then begin
        Rate.K1[1] := Double( RateNode['k1'].Data[0] );
        Rate.K1[2] := Double( RateNode['k1'].Data[1] );
      end;

      if RateNode.Find('P').DataCount >= 2 then try
        Rate.P[1]  := 0 <> Integer( RateNode['P'].Data[0] );
        Rate.P[2]  := 0 <> Integer( RateNode['P'].Data[1] );
      except
      end;

      if RateNode.Find('Q').DataCount >= 2 then try
        Rate.Q[1]  := 0 <> Integer( RateNode['Q'].Data[0] );
        Rate.Q[2]  := 0 <> Integer( RateNode['Q'].Data[1] );

        CondName   := RateNode['PNames'].Child;
        Rate.PName[1] := CondName.DataAsString;
        CondName   := CondName.Sibling;
        Rate.PName[2] := CondName.DataAsString;

        CondName   := RateNode['QNames'].Child;
        Rate.QName[1] := CondName.DataAsString;
        CondName   := CondName.Sibling;
        Rate.QName[2] := CondName.DataAsString;
      except
      end;

      RateNode   := RateNode.SiblingSameName;
    end;

  except
  end;
end;

function TGetModel.Call( args, kw : PPyObject) : PPyObject;
var
  ModelNode :INode;
begin
  ModelNode := GetModelAsNode;

  if not ModelNode.Find('Properties').IsNull then begin
    ModelNode.RemoveChild(ModelNode['Properties']);
  end;
  ModelNode.AppendClone(ViewportModel.CurrModel.ModelObject.Properties, True);

  Result := PyNode_FromNode( ModelNode );
end;

function TSetModel.Call( args, kw : PPyObject) : PPyObject;
var
  arg :PPyObject;
begin
  with GetPythonEngine do begin
    if ( PyArg_ParseTuple( args, 'O:SetModel(ModelNode)', [@arg] ) = 0 ) then begin
      Result := nil;
    end else begin
      if PyNode_IsNode( arg ) then begin
        prModelNode := PyNode_ToNode( arg );
        PythonThread.SyncGUIProc( prRun );

        Result := ReturnNone;
      end else begin
        PyErr_SetString( PyExc_TypeError^, PChar('SetModel(ModelNode)') );
        Result := nil;
      end;
    end;
  end;
end;

procedure TSetModel.prRun;
begin
  SetModelFromNode(ViewportModel.CurrModel.ModelObject, prModelNode);
  ViewportModel.CurrModel.DoOnSetProperties(ProfileMgr['QVM'], prModelNode.Find('Properties'));
end;

function TSetRates.Call( args, kw : PPyObject) : PPyObject;
var
  arg                    :PPyObject;
begin
  with GetPythonEngine do begin
    if ( PyArg_ParseTuple( args, 'O:SetRates(RatesNode)', [@arg] ) = 0 ) then begin
      Result := nil;
    end else begin
      if PyNode_IsNode( arg ) then begin
        prRatesNode := PyNode_ToNode( arg );
        PythonThread.SyncGUIProc( prRun );
        Result := ReturnNone;
      end else begin
        PyErr_SetString( PyExc_TypeError^, PChar('SetRates(RatesNode)') );
        Result := nil;
      end;
    end;
  end;
end;

procedure TSetRates.prRun;
begin
  SetRatesFromNode(ViewportModel.CurrModel.ModelObject, prRatesNode);
end;


procedure InitModelScripts;
var
  inst    :PPyObject;
begin
  GetModelType := TPythonType.Create(nil);
  GetModelType.Engine := GetPythonEngine;
  GetModelType.Module := PythonViewport.pmQUBTypes;
  GetModelType.TypeName := 'GetModelFunc';
  GetModelType.Services.Basic := [bsCall];
  GetModelType.PyObjectClass := TGetModel;
  GetModelType.Initialize;

  inst := GetModelType.CreateInstance;
  PyNode_PublishPyObj( inst, 'getModel' );
  GetPythonEngine.Py_DECREF( inst );

  SetModelType := TPythonType.Create(nil);
  SetModelType.Engine := GetPythonEngine;
  SetModelType.Module := PythonViewport.pmQUBTypes;
  SetModelType.TypeName := 'SetModelFunc';
  SetModelType.Services.Basic := [bsCall];
  SetModelType.PyObjectClass := TSetModel;
  SetModelType.Initialize;

  inst := SetModelType.CreateInstance;
  PyNode_PublishPyObj( inst, 'setModel' );
  GetPythonEngine.Py_DECREF( inst );

  SetRatesType := TPythonType.Create(nil);
  SetRatesType.Engine := GetPythonEngine;
  SetRatesType.Module := PythonViewport.pmQUBTypes;
  SetRatesType.TypeName := 'SetRatesFunc';
  SetRatesType.Services.Basic := [bsCall];
  SetRatesType.PyObjectClass := TSetRates;
  SetRatesType.Initialize;

  inst := SetRatesType.CreateInstance;
  PyNode_PublishPyObj( inst, 'setRates' );
  GetPythonEngine.Py_DECREF( inst );

end;


end.
