{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// by Chris
// grab bag of sketchy
// 1. wrappers and voodoo for sharing function/method pointers across c++/delphi/python if they have signature:
//       function (arg[, arg2] : INode): INode
// 2. utilities for converting data between delphi and python
// 3. utilities for publishing callbacks and objects in the QUB module
// 4. utilities for sharing qubtrees between delphi and python

   
unit QUB_Node_Python;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
StdCtrls, ExtCtrls, ComCtrls, Menus, Grids,

AllocTypes,

QUB_Node, PythonEngine, QUB_NodeEditForm, QUB_NodeEditor;


type

// interfaces for common delphi use

INodeFunc = interface(IUnknown)

  function NodeCall( arg: INode ): INode;

end;

INodeNodeFunc = interface(IUnknown)

  function NodeNodeCall( arg1: INode; arg2: INode ): INode;

end;

// wrapper interface and implementation suitable for passing to C++ dlls

INodeCFunc = interface(IUnknown)

  function NodeCall( arg: TQUB_QFS_PNodeMem ) : TQUB_QFS_PNodeMem; stdcall;

end;

INodeNodeCFunc = interface(IUnknown)

  function NodeNodeCall( arg1, arg2: TQUB_QFS_PNodeMem ) : TQUB_QFS_PNodeMem; stdcall;

end;

TNodeCFunc = class(TInterfacedObject, INodeCFunc)
protected
  func          :INodeFunc;
public
  constructor Wrap( fn: INodeFunc );
  function NodeCall( arg: TQUB_QFS_PNodeMem ) : TQUB_QFS_PNodeMem; stdcall;
end;

TNodeNodeCFunc = class(TInterfacedObject, INodeNodeCFunc)
protected
  func          :INodeNodeFunc;
public
  constructor Wrap( fn: INodeNodeFunc );
  function NodeNodeCall( arg1, arg2: TQUB_QFS_PNodeMem ): TQUB_QFS_PNodeMem; stdcall;
end;

// python-exported object which tests python-native callbacks by calling them 5 times
{
TCall5 = class(TPyObject)
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;
}
// wrapper (intended for delphi-native classes) suitable for python users
// ?? release GIL during call ??

TNodeFuncPyWrap = class(TPyObject)
protected
  func          :INodeFunc;
public
  function  Call( args, kw : PPyObject) : PPyObject; override;
  property Fn     :INodeFunc        write func;
end;

PNodeFuncPyWrap = ^TNodeFuncPyWrap;

TNodeNodeFuncPyWrap = class(TPyObject)
protected
  func          :INodeNodeFunc;
public
  function  Repr : PPyObject; override;
  function  Call( args, kw : PPyObject) : PPyObject; override;
  property Fn     :INodeNodeFunc        write func;
end;

PNodeNodeFuncPyWrap = ^TNodeNodeFuncPyWrap;

TNodeMethPtr = function( arg: INode ): INode of object;
TNodeMeth = class(TInterfacedObject, INodeFunc)
protected
  meth      :TNodeMethPtr;
public
  constructor Create( method :TNodeMethPtr );

  function NodeCall( arg: INode ): INode;
end;


TQUB_PyFuncPtr = function( args, kw :PPyObject ): PPyObject;
TQUB_PyFunc = class(TPyObject)
protected
  func         :TQUB_PyFuncPtr;
public
  function Call( args, kw :PPyObject ): PPyObject; override;
  property Fn   :TQUB_PyFuncPtr         read func        write func;
end;


{

myclass = class
  function dosomething(node: INode): INode;

var fn : INodeFunc;
fn := TNodeMeth.Create( myclass_instance.dosomething );
PyNode_PublishNodeFunc( fn, 'dosomthing' );   QUB.dosomething

del QUB.dosomething



tmyimpl = class(TInterfacedObject, INodeFunc)
  function NodeCall(node: INode): INode;

var fn :INodeFunc;
fn := tmyimpl.create;
PyNode_PublishNodeFunc( fn, 'dosomethingelse' );   QUB.dosomethingelse

}


TNodeNodeMethPtr = function( arg1: INode; arg2: Inode ): INode of object;
TNodeNodeMeth = class(TInterfacedObject, INodeNodeFunc)
protected
  meth      :TNodeNodeMethPtr;
public
  constructor Create( method :TNodeNodeMethPtr );

  function NodeNodeCall( arg1: INode; arg2: INode ): INode;
end;

TNodeFuncPtr = function( arg: INode ): INode;
TNodeFunc = class(TInterfacedObject, INodeFunc)
protected
  func       :TNodeFuncPtr;
public
  constructor Create( fn :TNodeFuncPtr );

  function NodeCall( arg: INode ): INode;
end;

TNodeNodeFuncPtr = function( arg1: INode; arg2: Inode ): INode;
TNodeNodeFunc = class(TInterfacedObject, INodeNodeFunc)
protected
  func      :TNodeNodeFuncPtr;
public
  constructor Create( fn :TNodeNodeFuncPtr );

  function NodeNodeCall( arg1: INode; arg2: INode ): INode;
end;

TNodePyFunc = class(TInterfacedObject, INodeFunc)
protected
  func        :PPyObject;
public
  constructor Create( fn: PPyObject );
  destructor Destroy; override;

  function NodeCall( arg: INode ): INode;
end;

TNodeNodePyFunc = class(TInterfacedObject, INodeNodeFunc)
protected
  func      :PPyObject;
public
  constructor Create( fn :PPyObject );
  destructor Destroy; override;

  function NodeNodeCall( arg1: INode; arg2: INode ): INode;
end;


// u must call these while holding the python GIL
function PyObject_DelphiStr(x :PPyObject): String;
procedure PyStrings_ToStrings(PyStrings :PPyObject; Strings :TStrings);
function PyStrings_FromStrings(Strings :TStrings): PPyObject;
function PyObject_GetAttrStringAsString(PyObject :PPyObject; Attr :String): String;
function PyObject_GetAttrStringAsNode(PyObject :PPyObject; Attr :String): INode;
function PyErr_ShortMessage: String;
procedure Py_ParseIntList(x :String; lst :TList);
procedure Py_ParseFloatList(x :String; lst :TList);


procedure PyNode_InitCallbacks;

procedure PyNode_Publish( node: INode; name: String );
procedure PyNode_PublishPyObj( obj: PPyObject; name: String );
procedure PyNode_PublishPyFunc( func :TQUB_PyFuncPtr; name :String );
procedure PyNode_PublishNodeFunc( func: INodeFunc; name: String );
function PyNode_WrapNodeFunc( func :INodeFunc ): PPyObject;
procedure PyNode_PublishNodeNodeFunc( func: INodeNodeFunc; name: String );
function PyNode_WrapNodeNodeFunc( func :INodeNodeFunc ): PPyObject;


procedure Init_QUBTree; cdecl; external '_qubtree.pyd' name 'init_qubtree';

function PyNodelib_Using( impl: TQUB_QFS_PNodeMem ): PPyObject; cdecl; external '_qubtree.pyd' name 'qubtree_node_using';
function PyNodelib_NodeMemOf( node: PPyObject ): TQUB_QFS_PNodeMem; cdecl; external '_qubtree.pyd' name 'qubtree_node_impl';
function PyNodelib_IsNode( obj: PPyObject ): Integer; cdecl; external '_qubtree.pyd' name 'qubtree_node_check';

function PyNode_FromNode( node: INode ): PPyObject;
function PyNode_ToNode( node: PPyObject ): INode;
function PyNode_IsNode( obj: PPyObject ): Boolean;


type

TQUB_NodeEditClosure = class // for invoking an editor from other threads
public
  EditForm          :TQUB_NodeEditForm;
  Node              :INode;
  Constr            :INode;
  Done              :Boolean;
  procedure DoEdit;
  procedure prOnEditDone(Sender :TObject);
end;

function PyNode_ShowEdit(x :INode): INode;
function PyNode_ShowEditConstr(x, c :INode): INode;


implementation


uses


QUB_Python_Thread, QUB_Viewport_Python,
QUB_Main;


var 


NodeFuncPyType, NodeNodeFuncPyType, PyFuncPyType    :TPythonType;
//Call5Type :TPythonType;

{
function DoSomething( arg: INode ): INode;
begin
  Console.NewEntry( 'foo!', COLOR_ERROR );
  Result := TNode.Create('foo');
end;
}
procedure PyNode_InitCallbacks;
//var
//  call5   :PPyObject;
begin
  NodeFuncPyType := TPythonType.Create(nil); // owner?
  NodeFuncPyType.Engine := GetPythonEngine;
  NodeFuncPyType.Module := PythonViewport.pmQUBTypes;
  NodeFuncPyType.TypeName := 'NodeFunc';
  NodeFuncPyType.Prefix := 'Create';
  NodeFuncPyType.Services.Basic := [bsCall, bsRepr];
  NodeFuncPyType.PyObjectClass := TNodeFuncPyWrap;
  NodeFuncPyType.Initialize;

  NodeNodeFuncPyType := TPythonType.Create(nil); // owner?
  NodeNodeFuncPyType.Engine := GetPythonEngine;
  NodeNodeFuncPyType.Module := PythonViewport.pmQUBTypes;
  NodeNodeFuncPyType.TypeName := 'NodeNodeFunc';
  NodeNodeFuncPyType.Prefix := 'Create';
  NodeNodeFuncPyType.Services.Basic := [bsCall, bsRepr];
  NodeNodeFuncPyType.PyObjectClass := TNodeNodeFuncPyWrap;
  NodeNodeFuncPyType.Initialize;

  PyFuncPyType := TPythonType.Create(nil); // owner?
  PyFuncPyType.Engine := GetPythonEngine;
  PyFuncPyType.Module := PythonViewport.pmQUBTypes;
  PyFuncPyType.TypeName := 'QUB_PyFunc';
  PyFuncPyType.Prefix := 'Create';
  PyFuncPyType.Services.Basic := [bsCall];
  PyFuncPyType.PyObjectClass := TQUB_PyFunc;
  PyFuncPyType.Initialize;
  {
  PyNode_PublishNodeFunc( TNodeFunc.Create( DoSomething ), 'testfn' );

  Call5Type := TPythonType.Create(nil);
  Call5Type.Engine := GetPythonEngine;
  Call5Type.Module := PythonViewport.pmQUBTypes;
  Call5Type.TypeName := 'Call5TimesObj';
  Call5Type.Services.Basic := [bsCall];
  Call5Type.PyObjectClass := TCall5;
  Call5Type.Initialize;

  call5 := Call5Type.CreateInstance;
  PyNode_PublishPyObj( call5, 'call5times' );
  GetPythonEngine.Py_DECREF( call5 );
  }

  PyNode_PublishNodeFunc( TNodeFunc.Create(PyNode_ShowEdit), 'editTree' );
  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create(PyNode_ShowEditConstr), 'editCTree' );

end;
{
function TCall5.Call( args, kw : PPyObject) : PPyObject;
var
  pyfunc   :PPyObject;
  func     :INodeFunc;
  node     :INode;
begin
  with GetPythonEngine do begin
    if ( PyArg_ParseTuple( args, 'O:call5times', [@pyfunc] ) = 0 ) then begin
      Result := nil
    end else begin
      func := TNodePyFunc.Create( pyfunc );
      node := TNode.Create('abc');
      func.NodeCall( node );
      func.NodeCall( node );
      func.NodeCall( node );
      func.NodeCall( node );
      func.NodeCall( node );
      Result := ReturnNone;
    end;
  end;
end;
}
constructor TNodeCFunc.Wrap( fn: INodeFunc );
begin
  func := fn;
end;

function TNodeCFunc.NodeCall( arg: TQUB_QFS_PNodeMem ) : TQUB_QFS_PNodeMem;
var
  node   :INode;
begin
  node := TNode.CreateUsing( arg );
  node := func.NodeCall( node );
  Result := node.GetNodeMem;
  if Result <> nil then Node_INCREF( Result );
end;

constructor TNodeNodeCFunc.Wrap( fn: INodeNodeFunc );
begin
  func := fn;
end;

function TNodeNodeCFunc.NodeNodeCall( arg1, arg2: TQUB_QFS_PNodeMem ) : TQUB_QFS_PNodeMem;
var
  node1, node2   :INode;
begin
  node1 := TNode.CreateUsing( arg1 );
  node2 := TNode.CreateUsing( arg2 );
  node1 := func.NodeNodeCall( node1, node2 );
  Result := node1.GetNodeMem;
  if Result <> nil then Node_INCREF( Result );
end;

function  TNodeFuncPyWrap.Call( args, kw : PPyObject) : PPyObject;
var
  arg                    :PPyObject;
  argNode                :INode;
  ResultNode             :INode;
  //xThreadState           :PPyThreadState;
begin
  with GetPythonEngine do begin
    arg      := nil;
    argNode  := nil;
    if ( PyArg_ParseTuple( args, '|O:f(node)', [@arg] ) = 0 ) then begin
      Result := nil;
    end else begin
      try
        if arg = nil then begin
          argNode := TNode.CreateNull;
        end else if PyNode_IsNode( arg ) then begin
          argNode := PyNode_ToNode( arg );
        end;

        if argNode <> nil then begin
          // release GIL
          PythonThread.EndPython;
          //xThreadState        := PyThreadState_Get;
          //PyThreadState_Swap(nil);
          //PyEval_ReleaseLock;

          ResultNode := func.NodeCall( argNode );

          // reclaim GIL
          PythonThread.BeginPython;
          //PyEval_AcquireLock;
          //PyThreadState_Swap( xThreadState );

          Result  := PyNode_FromNode( ResultNode );
        end else begin
          PyErr_SetString( PyExc_TypeError^, PChar('f(node)') );
          Result := nil;
        end;
      except
        Result := PyNode_FromNode( TNode.CreateNull );
      end;
    end;
  end;
end;

function  TNodeNodeFuncPyWrap.Repr : PPyObject;
begin
  with GetPythonEngine do
    Result := PyString_FromString( PChar('f(node, node) -> node') );
end;

function  TNodeNodeFuncPyWrap.Call( args, kw : PPyObject) : PPyObject;
var
  arg1, arg2             :PPyObject;
  arg1Node, arg2Node     :INode;
  ResultNode             :INode;
  //xThreadState           :PPyThreadState;
begin
  with GetPythonEngine do begin
    arg1     := nil;
    arg2     := nil;
    arg1Node := nil;
    arg2Node := nil;

    if ( PyArg_ParseTuple( args, '|OO:f(node, node)', [@arg1, @arg2] ) = 0 ) then begin
      Result := nil;
    end else begin
      try
        if arg1 = nil then begin
          arg1Node := TNode.CreateNull;
        end else if PyNode_IsNode( arg1 ) then begin
          arg1Node := PyNode_ToNode( arg1 );
        end;

        if arg2 = nil then begin
          arg2Node := TNode.CreateNull;
        end else if PyNode_IsNode( arg2 ) then begin
          arg2Node := PyNode_ToNode( arg2 );
        end;

        if (arg1Node <> nil) and (arg2Node <> nil) then begin
          // release GIL
          PythonThread.EndPython;
          //xThreadState        := PyThreadState_Get;
          //PyThreadState_Swap(nil);
          //PyEval_ReleaseLock;

          ResultNode := func.NodeNodeCall( arg1Node, arg2Node );

          // reclaim GIL
          PythonThread.BeginPython;
          //PyEval_AcquireLock;
          //PyThreadState_Swap( xThreadState );

          Result     := PyNode_FromNode( ResultNode );
        end else begin
          PyErr_SetString( PyExc_TypeError^, PChar('f(node,node)') );
          Result := nil;
        end;
      except
        Result := PyNode_FromNode( TNode.CreateNull );
      end;
    end;
  end;
end;

constructor TNodeMeth.Create( method :TNodeMethPtr );
begin
  meth := method;
end;

function TNodeMeth.NodeCall( arg: INode ): INode;
begin
  try
    Result := meth( arg );
  except
    Result := TNode.CreateNull;
  end;
end;

constructor TNodeNodeMeth.Create( method :TNodeNodeMethPtr );
begin
  meth := method;
end;

function TNodeNodeMeth.NodeNodeCall( arg1: INode; arg2: INode ): INode;
begin
  try
    Result := meth( arg1, arg2 );
  except
    Result := TNode.CreateNull;
  end;
end;

constructor TNodeFunc.Create( fn :TNodeFuncPtr );
begin
  func := fn;
end;

function TNodeFunc.NodeCall( arg: INode ): INode;
begin
  try
    Result := func( arg );
  except
    Result := TNode.CreateNull;
  end;
end;

constructor TNodeNodeFunc.Create( fn :TNodeNodeFuncPtr );
begin
  func := fn;
end;

function TNodeNodeFunc.NodeNodeCall( arg1: INode; arg2: INode ): INode;
begin
  try
    Result := func( arg1, arg2 );
  except
    Result := TNode.CreateNull;
  end;
end;

constructor TNodePyFunc.Create( fn: PPyObject );
begin
  func := fn;
  GetPythonEngine.Py_INCREF( fn );
end;

destructor TNodePyFunc.Destroy;
begin
  GetPythonEngine.Py_DECREF( func );
end;

function TNodePyFunc.NodeCall( arg: INode ): INode;
var
  arglist                      :PPyObject;
  resObj                       :PPyObject;
begin
  with GetPythonEngine do begin
    try
      arglist := Py_BuildValue('(O)', [PyNode_FromNode( arg )]);
      resObj := PyEval_CallObject(func, arglist);
      Py_DECREF(arglist);
      if resObj = nil then begin
        Result := TNode.CreateNull;
        PyErr_Print;
      end else if PyNode_IsNode( resObj ) then begin
        Result := PyNode_ToNode( resObj );
        Py_DECREF(resObj);
      end else begin
        Result := TNode.CreateNull;
        Py_DECREF(resObj);
      end;
    except
      Result := TNode.CreateNull;
    end;
  end;
end;

constructor TNodeNodePyFunc.Create( fn: PPyObject );
begin
  func := fn;
  GetPythonEngine.Py_INCREF( fn );
end;

destructor TNodeNodePyFunc.Destroy;
begin
  GetPythonEngine.Py_DECREF( func );
end;

function TNodeNodePyFunc.NodeNodeCall( arg1: INode; arg2: INode ): INode;
var
  arglist                      :PPyObject;
  resObj                       :PPyObject;
begin
  with GetPythonEngine do begin
    try
      arglist := Py_BuildValue('(OO)', [PyNode_FromNode( arg1 ), PyNode_FromNode( arg2 )]);
      resObj := PyEval_CallObject(func, arglist);
      Py_DECREF(arglist);
      if resObj = nil then begin
        Result := TNode.CreateNull;
        PyErr_Print;
      end else if PyNode_IsNode( resObj ) then begin
        Result := PyNode_ToNode( resObj );
        Py_DECREF(resObj);
      end else begin
        Result := TNode.CreateNull;
        Py_DECREF(resObj);
      end;
    except
      Result := TNode.CreateNull;
    end;
  end;
end;


function TQUB_PyFunc.Call( args, kw :PPyObject ): PPyObject;
begin
  Result := func( args, kw );
end;


procedure PyNode_Publish( node: INode; name: String );
var
  nodeObj                 :PPyObject;
begin
  nodeObj := PyNode_FromNode( node );
  PythonViewport.pmQUB.setVar( name, nodeObj );
  GetPythonEngine.Py_DECREF( nodeObj );
end;

procedure PyNode_PublishPyObj( obj: PPyObject; name: String );
begin
  PythonViewport.pmQUB.setVar( name, obj );
end;

procedure PyNode_PublishPyFunc( func :TQUB_PyFuncPtr; name: String );
var
  funcObj                 :PPyObject;
begin
  funcObj := PyFuncPyType.CreateInstance;
  TQUB_PyFunc( PythonToDelphi( funcObj ) ).Fn := func;
  PythonViewport.pmQUB.setVar( name, funcObj );
  GetPythonEngine.Py_DECREF( funcObj );
end;

procedure PyNode_PublishNodeFunc( func: INodeFunc; name: String );
var
  wrap    :PPyObject;
begin
  wrap := PyNode_WrapNodeFunc( func );
  PythonViewport.pmQUB.setVar( name, wrap );
  GetPythonEngine.Py_DECREF( wrap );
end;

function PyNode_WrapNodeFunc( func :INodeFunc ): PPyObject;
var
  wrapObj     :TNodeFuncPyWrap;
begin
  Result      := NodeFuncPyType.CreateInstance;
  wrapObj     := TNodeFuncPyWrap( PythonToDelphi( Result ) );
  wrapObj.Fn  := func;
end;

procedure PyNode_PublishNodeNodeFunc( func: INodeNodeFunc; name: String );
var
  wrap    :PPyObject;
  wrapObj :TNodeNodeFuncPyWrap;
begin
  wrap := NodeNodeFuncPyType.CreateInstance;
  wrapObj := TNodeNodeFuncPyWrap( PythonToDelphi( wrap ) );
  wrapObj.Fn := func;
  PythonViewport.pmQUB.setVar( name, PPyObject(wrap) );
  GetPythonEngine.Py_DECREF( wrap );
end;

function PyNode_WrapNodeNodeFunc( func :INodeNodeFunc ): PPyObject;
var
  wrapObj     :TNodeNodeFuncPyWrap;
begin
  Result      := NodeNodeFuncPyType.CreateInstance;
  wrapObj     := TNodeNodeFuncPyWrap( PythonToDelphi( Result ) );
  wrapObj.Fn  := func;
end;

// convenience wrappers for _qubtree.pyd


function PyNode_FromNode( node: INode ): PPyObject;
begin
  Result := PyNodelib_Using( node.GetNodeMem );
end;

function PyNode_ToNode( node: PPyObject ): INode;
begin
  Result := TNode.CreateUsing( PyNodelib_NodeMemOf( node ) );
end;

function PyNode_IsNode( obj: PPyObject ): Boolean;
begin
  Result := Boolean( PyNodelib_IsNode( obj ) );
end;


// *****************************************************************************

function PyNode_ShowEdit(x :INode): INode;
var
  Closure               :TQUB_NodeEditClosure;
begin
  Closure               := TQUB_NodeEditClosure.Create;
  Closure.Node          := x;
  Closure.Constr        := QUB_Node_Null;
  SyncGUIProc( Closure.DoEdit );

  while not Closure.Done do Sleep(1000);
  Result                := Closure.Node;
  Closure.Free;
end;

function PyNode_ShowEditConstr(x, c :INode): INode;
var
  Closure               :TQUB_NodeEditClosure;
begin
  Closure               := TQUB_NodeEditClosure.Create;
  Closure.Node          := x;
  Closure.Constr        := c;
  SyncGUIProc( Closure.DoEdit );

  while not Closure.Done do Sleep(1000);
  Result                := Closure.Node;
  Closure.Free;
end;

procedure TQUB_NodeEditClosure.DoEdit;
begin
  EditForm                 := TQUB_NodeEditForm.Create(Application);
  EditForm.OnDone          := prOnEditDone;
  EditForm.EditFrame.Root  := Node;
  if not Constr.IsNull then begin
    EditForm.EditFrame.Constraints := Constr;
    EditForm.EditFrame.EditTree    := False;
    EditForm.EditFrame.LoadData    := False;
    EditForm.EditFrame.SizeData    := False;
  end;
  EditForm.Show;
  Done                     := False;
end;

procedure TQUB_NodeEditClosure.prOnEditDone(Sender :TObject);
begin
  EditForm.Release;
  Done := True;
end;


// *****************************************************************************

function PyObject_DelphiStr(x :PPyObject): String;
var
  PyJunk            :PPyObject;
begin
  if x = nil then begin
    Result          := '';
    Exit;
  end;

  with GetPythonEngine do begin
    PyJunk          := PyObject_Str(x);
    Result          := PyString_AsString(PyJunk);
    Py_DECREF( PyJunk );
  end;
end;

procedure PyStrings_ToStrings(PyStrings :PPyObject; Strings :TStrings);
var
  PyStr       :PPyObject;
  i           :Integer;
begin
  with GetPythonEngine do begin
    if 0 = PySequence_Check(PyStrings) then Exit;

    Strings.Clear;
    for i := 0 to PySequence_Length(PyStrings) - 1 do begin
      PyStr   := PySequence_GetItem(PyStrings, i);
      Strings.Add( PyObject_DelphiStr(PyStr) );
      Py_DECREF( PyStr );
    end;
  end;
end;

function PyStrings_FromStrings(Strings :TStrings): PPyObject;
var
  i           :Integer;
begin
  with GetPythonEngine do begin
    Result    := PyList_New(Strings.Count);
    for i := 0 to Strings.Count - 1 do begin
      PyList_SetItem(Result, i, PyString_FromString(PChar(Strings[i])));
    end;
  end;
end;

function PyObject_GetAttrStringAsString(PyObject :PPyObject; Attr :String): String;
var
  PyJunk         :PPyObject;
begin
  with GetPythonEngine do begin
    PyJunk       := PyObject_GetAttrString(PyObject, PChar(Attr));
    if PyJunk = nil then begin
      Result     := '';
      PyErr_Clear();
    end else begin
      Result     := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;
  end;
end;

function PyObject_GetAttrStringAsNode(PyObject :PPyObject; Attr :String): INode;
var
  PyJunk         :PPyObject;
begin
  with GetPythonEngine do begin
    PyJunk       := PyObject_GetAttrString(PyObject, PChar(Attr));
    if PyJunk = nil then begin
      Result     := QUB_Node_NULL;
      PyErr_Clear();
    end else begin
      Result     := PyNode_ToNode(PyJunk);
      Py_DECREF(PyJunk);
    end;
  end;
end;

function PyErr_ShortMessage: String;
var
  ErrType, ErrVal, ErrTrace :PPyObject;
begin
  Result     := '';
  ErrType    := nil;
  ErrVal     := nil;
  ErrTrace   := nil;

  with GetPythonEngine do begin
    try
      PyErr_Fetch(@ErrType, @ErrVal, @ErrTrace);
      if (ErrType = nil) or (ErrVal = nil) then Exit;
      Result := PyObject_DelphiStr(ErrType) + ': ' + PyObject_DelphiStr(ErrVal);
    finally
      Py_XDECREF(ErrType);
      Py_XDECREF(ErrVal);
      Py_XDECREF(ErrTrace);
    end;
  end;
end;

procedure Py_ParseIntList(x :String; lst :TList);
var
  PyList    :PPyObject;
  i         :Integer;
begin
  with GetPythonEngine do begin
    try
      PyList  := EvalString(Format('QUBB.ParseIntList("%s")', [x]));
      lst.Clear;
      for i := 0 to PyList_Size(PyList) - 1 do begin
        Lst.Add( Pointer(PyInt_AsLong(PyList_GetItem(PyList, i))) );
      end;
      Py_DECREF(PyList);
    except
      raise Exception.Create(Format('ParseIntList: invalid list: %s', [x]));
    end;
  end;
end;

procedure Py_ParseFloatList(x :String; lst :TList);
var
  PyList    :PPyObject;
  i         :Integer;
  s         :Single;
begin
  with GetPythonEngine do begin
    try
      PyList  := EvalString(Format('QUBB.ParseTypedList("%s", float)', [x]));
      lst.Clear;
      for i := 0 to PyList_Size(PyList) - 1 do begin
        s := PyFloat_AsDouble(PyList_GetItem(PyList, i));
        Lst.Add( Pointer(s) );
      end;
      Py_DECREF(PyList);
    except
      raise Exception.Create(Format('ParseFloatList: Invalid list: %s', [x]));
    end;
  end;
end;


// *****************************************************************************
end.
