{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Profiles_Python;



interface


uses

Math,
Alloc, AllocTypes,
PythonEngine,

QUB_Node,
QUB_Profiles;


type

TQUB_ProfileMgr_Python = class(TPyObject)
protected
  prMgr     :TQUB_ProfileMgr;

public

  // Mapping services
  function  MpSubscript( obj : PPyObject) : PPyObject; override;

  property Mgr   :TQUB_ProfileMgr read prMgr write prMgr;

end;

TQUB_ProfileCat_Python = class(TPyObject)
protected
  prCat     :TQUB_ProfileCat;

public

  function  GetAttr(key : PChar) : PPyObject; override;
  function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  class procedure RegisterMethods( PythonType : TPythonType ); override;
  function SetProperties(Args :PPyObject): PPyObject; cdecl;
{
  function ListProfiles(Args :PPyObject): PPyObject; cdecl;
  function SaveProfile(Args :PPyObject): PPyObject; cdecl;
  function LoadProfile(Args :PPyObject): PPyObject; cdecl;
  function ExportProfile(Args :PPyObject): PPyObject; cdecl;
  function ImportProfile(Args :PPyObject): PPyObject; cdecl;
}
  property Cat     :TQUB_ProfileCat   read prCat    write prCat;
end;



// call while holding GIL
procedure InitProfiles_Py;



implementation

uses

QUB_Viewport_Python, QUB_Python_Thread,
QUB_Node_Python,
QUB_Main, QUB_GUIInterface;


var

  ProfileCatPyType            :TPythonType;


//******************************************************************************

function  TQUB_ProfileMgr_Python.MpSubscript( obj : PPyObject) : PPyObject;
var
  Cat       :TQUB_ProfileCat;
  PyCat     :TQUB_ProfileCat_Python;
begin
  with GetPythonEngine do begin
    if PyString_Check( obj ) then begin
      if Mgr = nil then begin
        Result := ReturnNone;
      end else begin
        Cat := Mgr[ PyString_AsString( obj ) ];
        Result      := ProfileCatPyType.CreateInstance;
        PyCat       := TQUB_ProfileCat_Python( PythonToDelphi(Result) );
        PyCat.Cat   := Cat;
      end;
    end else begin
      PyErr_SetString( PyExc_TypeError^, PChar('get profile cat by string') );
      Result := nil;
    end;
  end;
end;


//******************************************************************************

function  TQUB_ProfileCat_Python.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do begin
    if key = 'name' then begin
      Result := PyString_FromString( PChar(Cat.Name) );
    end else if key = 'active' then begin
      Result := PyNode_FromNode( Cat.Active );
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

function  TQUB_ProfileCat_Python.SetAttr(key : PChar; value : PPyObject) : Integer;
var
  PyActive             :PPyObject;
begin
  Result := 0;
  with GetPythonEngine do begin
    if key = 'active' then begin
      if PyArg_Parse( value, 'O:active', [@PyActive] ) = 0 then begin
        Result := -1;
      end else begin
        if not PyNode_IsNode(PyActive) then begin
          PyErr_SetString( PyExc_TypeError^, PChar('expected qubtree.Node') );
          Result := -1;
        end else begin
          Cat.Active := PyNode_ToNode(PyActive);
        end;
      end;
    end else begin
      Result := inherited SetAttr(key, value);
    end;
  end;
end;

class procedure TQUB_ProfileCat_Python.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do begin
    AddMethod( 'setProperties',  @TQUB_ProfileCat_Python.SetProperties,  'ProfileCat.setProperties(qubtree.Node)' );
  end;
end;

type TQUB_Run_SetProperties = class(TQUB_Runnable)
public
  Cat         :TQUB_ProfileCat;
  Node        :INode;

  constructor Create(ACat :TQUB_ProfileCat; ANode :INode);
  procedure Run; override;
end;

constructor TQUB_Run_SetProperties.Create(ACat :TQUB_ProfileCat; ANode :INode);
begin
  Cat         := ACat;
  Node        := ANode;
end;

procedure TQUB_Run_SetProperties.Run;
begin
  Cat.SetProperties(Node);
end;

function TQUB_ProfileCat_Python.SetProperties(Args :PPyObject): PPyObject; cdecl;
var
  PyNode      :PPyObject;
  Run         :TQUB_Run_SetProperties;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'O:setProperties( qubtree.Node )', [@PyNode] ) <> 0 then begin
      if PyNode_IsNode(PyNode) then begin
        Run   := TQUB_Run_SetProperties.Create(Cat, PyNode_ToNode(PyNode));
        try
          PythonThread.EndPython;
          SyncGUIRunnable(Run);
          PythonThread.BeginPython;
        except
          //??
        end;
        Run.Free;
        Result := ReturnNone;
      end else begin
        PyErr_SetString( PyExc_TypeError^, PChar('expected qubtree.Node') );
        Result := nil;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

{
  function ListProfiles(Args :PPyObject): PPyObject; cdecl;
  function SaveProfile(Args :PPyObject): PPyObject; cdecl;
  function LoadProfile(Args :PPyObject): PPyObject; cdecl;
  function ExportProfile(Args :PPyObject): PPyObject; cdecl;
  function ImportProfile(Args :PPyObject): PPyObject; cdecl;
}


//******************************************************************************
procedure InitProfiles_Py;
var
  PyType                    :TPythonType;
  PyInst                    :PPyObject;
  Inst                      :TQUB_ProfileMgr_Python;
begin
  ProfileCatPyType                    := TPythonType.Create(nil);
  ProfileCatPyType.Engine             := GetPythonEngine;
  ProfileCatPyType.Module             := PythonViewport.pmQUBTypes;
  ProfileCatPyType.TypeName           := 'ProfileCatType';
  ProfileCatPyType.Services.Basic     := [bsGetAttr, bsSetAttr];
  ProfileCatPyType.PyObjectClass      := TQUB_ProfileCat_Python;
  ProfileCatPyType.Initialize;

  PyType                    := TPythonType.Create(nil);
  PyType.Engine             := GetPythonEngine;
  PyType.Module             := PythonViewport.pmQUBTypes;
  PyType.TypeName           := 'ProfileMgrType';
  PyType.Services.Mapping   := [msSubscript];
  PyType.PyObjectClass      := TQUB_ProfileMgr_Python;
  PyType.Initialize;

  PyInst                    := PyType.CreateInstance;
  Inst                      := TQUB_ProfileMgr_Python( PythonToDelphi(PyInst) );
  Inst.Mgr                  := ProfileMgr;
  PyNode_PublishPyObj(PyInst, 'Profiles');
  GetPythonEngine.Py_DECREF(PyInst);
end;


//******************************************************************************
end.
