{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2002
// Weird nonstandard component that looks like a button, but when you click it you get a popup menu of python scripts.
// The scripts are functions that have been registered by Commands.Register(category, name, func)
// Alternatively, you can append a category of scripts to your own popup menu or combo box.


unit QUB_PyCmd_MenuButton;

interface

uses


SyncObjs, Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls, Buttons, Menus;

type

TQUB_PyCmd_MenuButton = class(TObject)

protected

  prCategory    :String;

  procedure prOnPopup(Sender :TObject);
  procedure prOnChoice(Sender :TObject);

  procedure prSetParent(par :TWinControl);
  procedure prSetCaption(lbl :String);

public

  Button              :TSpeedButton;
  Menu                :TPopupMenu;

  constructor Create( cat :String );
  destructor Destroy; override;

  procedure DoOnMouseUp(Sender: TObject; MouseButton: TMouseButton; Shift: TShiftState; X, Y: Integer);

  property Parent     :TWinControl                            write prSetParent;
  property Caption    :String                                 write prSetCaption;
  property Category   :String            read prCategory      write prCategory;

end;

function PyCmd_FillMenu(Menu :TMenuItem; CmdType :String; Handler :TNotifyEvent): Integer; overload;
function PyCmd_FillMenu(Menu :TComboBox; CmdType :String): Integer; overload;


implementation


uses
  QUB_Viewport_Python, PythonEngine, QUB_Python_Thread;


procedure TQUB_PyCmd_MenuButton.DoOnMouseUp(Sender: TObject;
          MouseButton: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Point   :TPoint;
begin
  inherited;
  Point.X := X;
  Point.Y := Y;
  Point   := Button.ClientToScreen( Point );
  Menu.Popup(Point.X,Point.Y);
end;

procedure TQUB_PyCmd_MenuButton.prOnPopup(Sender :TObject);
begin
  inherited;
  Menu.Items.Clear;

  PyCmd_FillMenu(Menu.Items, prCategory, prOnChoice);
end;

procedure TQUB_PyCmd_MenuButton.prOnChoice(Sender :TObject);
begin
  PythonThread.InterpretNonSilently(
       Format('Commands.CallByName( "%s", "%s" )', [prCategory, TMenuItem(Sender).Caption]) );
end;

procedure TQUB_PyCmd_MenuButton.prSetParent(par :TWinControl);
begin
  Button.Parent := par;
end;

procedure TQUB_PyCmd_MenuButton.prSetCaption(lbl :String);
begin
  Button.Caption     := lbl;
end;

constructor TQUB_PyCmd_MenuButton.Create( cat :String );
begin
  Button             := TSpeedButton.Create(nil);
  Button.OnMouseUp   := DoOnMouseUp;

  Menu               := TPopupMenu.Create(nil);
  Menu.AutoHotkeys   := maManual;
  Menu.OnPopup       := prOnPopup;

  Category           := cat;
  Caption            := cat;
end;

destructor TQUB_PyCmd_MenuButton.Destroy;
begin
  Button.Free;
  Menu.Free;
end;
//******************************************************************************

function PyCmd_FillMenu(Menu :TMenuItem; CmdType :String; Handler :TNotifyEvent): Integer;
var
  busy       :Boolean;
  lst        :PPyObject;
  PyItem     :PPyObject;
  len, i     :Integer;
  Item       :TMenuItem;
begin
  busy := PythonThread.Interpret(Format('Commands.Names("%s")', [CmdType]), lst, QP_EVALUATE, False, True, 1000);
  len  := 0;

  try
    if not busy then with GetPythonEngine do begin
      Menu.AutoHotkeys := maManual; // here or for each item?  better do both
      len    := PySequence_Length( lst );
      for i := 0 to len - 1 do begin
        PyItem := PySequence_GetItem(lst, i);
        Item   := TMenuItem.Create( nil );
        Item.AutoHotkeys := maManual; // don't insert no '&', so the caption can be used to choose the action
        Item.Caption := PyObjectAsString( PyItem );
        Item.OnClick := Handler;
        Menu.Add(Item);
        Py_DECREF( PyItem );
      end;
    end;
  except
  end;

  Result := len;
end;

function PyCmd_FillMenu(Menu :TComboBox; CmdType :String): Integer;
var
  busy       :Boolean;
  lst        :PPyObject;
  PyItem     :PPyObject;
  len, i     :Integer;
begin
  busy := PythonThread.Interpret(Format('Commands.Names("%s")', [CmdType]), lst, QP_EVALUATE, False, True, 1000);
  len  := 0;

  try
    if not busy then with GetPythonEngine do begin
      len    := PySequence_Length( lst );
      for i := 0 to len - 1 do begin
        PyItem := PySequence_GetItem(lst, i);
        Menu.Items.Add(PyObjectAsString( PyItem ));
        Py_DECREF( PyItem );
      end;
    end;
  except
  end;

  Result := len;
end;

end.
 
