{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// python: QUB.DataView, QUB.getChosenData, QUB.getChosenIdealization, QUB.fillData
// delphi: DataNoder with GetChosenData etc.


unit QUB_Python_Data;

interface

uses

Classes, Windows, SysUtils,
QUB_Viewport_Data,
PythonEngine,
QUB_Python_Thread,
QUB_Node;


procedure QUB_Python_DataView_Init(ViewportData :TQUB_Viewport_Data);


type


TQUB_Python_DataView = class(TPyObject)
protected

  prSegToMove        :INode;
  prCopyImageConfig  :PPyObject;
  prDataCount        :Integer;
  prIndex            :Integer;
  prResult           :PPyObject;
  prIndexToSet        :Integer;       //- used to store file index to be used by prSetCurrIndex - GUI thread
  prIResult           :Integer;

  procedure prMoveSelectionIntoWindow;
  procedure prCopyImage;
  function prGetDataCount: Integer;
  procedure prGetDataCountSync;
  procedure prGetCurrData;
  procedure prGetSelectedData;
  procedure prGetIndexData;
  procedure prGetCurrIndex;
  procedure prSetCurrIndex;           //- Change file - execute on GUI thread

  property DataCount         :Integer         read prGetDataCount;   // suitable to be called from python thread

public
  function  GetAttr(key : PChar) : PPyObject; override;
  function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  function  SqLength : Integer; override;
  function  SqItem( idx : Integer ) : PPyObject; override;

  function CopyImage( args : PPyObject ) : PPyObject; cdecl;
  function MoveSelectionIntoWindow( Seg :INode ): INode; // aka QUB.ShowSeg( DS['Segment'] )

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;

end;

// since Delphi distinguishes between function and 'function of object',
// these three functions must be in a class for flexibility
// (functions can be repackaged as methods, but not vice versa, so always ask for a method)

TQUB_DataNoder = class
public

  constructor Create;

  function GetChosenData(Parent :INode): INode; virtual;
  function GetChosenIdealization(Parent :INode): INode; virtual;
  function GetChosenIdlSignals(Parent :INode): INode; virtual;
  function FillData(SegNode :INode): INode; virtual;
  function FillDataCh(SegNode :INode): INode; virtual;
  function NullCallback(arg :INode): INode;

end;

TQUB_DataNoderScript = class(TQUB_DataNoder)
public

  function GetChosenData(Parent :INode): INode; override;
  function GetChosenIdealization(Parent :INode): INode; override;
  //function GetChosenIdlSignals(Parent :INode): INode; override;
  function FillData(SegNode :INode): INode; override;
  function FillDataCh(SegNode :INode): INode; override;

end;


var


DataNoder      :TQUB_DataNoder;
DataNoderScript :TQUB_DataNoderScript;


function GetDataSourcePy( args, kw :PPyObject ): PPyObject;
function SetDataSourcePy( args, kw :PPyObject ): PPyObject;



implementation


uses

Alloc, AllocTypes,
QUB_Viewport_Python,
QUB_Node_Python,
QUB_QVO_QVD,
QUB_QVO_QVD_Global,
QUB_QVO_QVD_Print_Properties,
QUB_Viewport_Model,
QUB_Global,
QUB_Main,
QUB_Task,
QUB_Idealization,
QUB_QFS_Channels,
QUB_GUIInterface,
QUB_Profiles;


var
  DataViewPyType                    :TPythonType;


procedure QUB_Python_DataView_Init(ViewportData :TQUB_Viewport_Data);
var
  PyInst                  :PPyObject;
  DataViewObj             :TQUB_Python_DataView;
begin
  InitDataPyTypes;

  DataViewPyType := TPythonType.Create(nil);
  DataViewPyType.Engine := GetPythonEngine;
  DataViewPyType.Module := PythonViewport.pmQUBTypes;
  DataViewPyType.TypeName := 'DataViewType';
  DataViewPyType.Services.Basic := [bsGetAttr, bsSetAttr];
  DataViewPyType.Services.Sequence  := [ssItem, ssLength];
  DataViewPyType.PyObjectClass := TQUB_Python_DataView;
  DataViewPyType.Initialize;

  PyInst := DataViewPyType.CreateInstance;
  DataViewObj := TQUB_Python_DataView( PythonToDelphi(PyInst) );
  PyNode_PublishPyObj( PyInst, 'DataView' );
  GetPythonEngine.Py_DECREF( PyInst );

  PyNode_PublishNodeFunc( TNodeMeth.Create( DataViewObj.MoveSelectionIntoWindow ), 'ShowSeg' );

  PyNode_PublishPyFunc( GetDataSourcePy, 'GetDataSource' );
  PyNode_PublishPyFunc( SetDataSourcePy, 'SetDataSource' );

  DataNoder := TQUB_DataNoder.Create;
  DataNoderScript := TQUB_DataNoderScript.Create;
end;



function TQUB_Python_DataView.prGetDataCount: Integer;
begin
  PythonThread.SyncGUIProc( prGetDataCountSync );
  Result := prDataCount;
end;

procedure TQUB_Python_DataView.prGetDataCountSync;
begin
  prDataCount := ViewportData.ListObjects.Items.Count;
end;

procedure TQUB_Python_DataView.prGetCurrData;
begin
  if ViewportData.CurrData <> nil then begin
    PythonThread.BeginPython;
    prResult := ViewportData.CurrData.PyHandle.GetSelf;
    GetPythonEngine.Py_INCREF( prResult );
    PythonThread.EndPython;
  end else begin
    prResult := nil;
  end;
end;

procedure TQUB_Python_DataView.prGetSelectedData;
var
  PyData                :PPyObject;
  i                     :Integer;
begin
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    prResult := PyList_New( 0 );
    for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
      if ViewportData.ListObjects.Selected[i] then begin
        PyData := TQUB_QVO_QVD(ViewportData.ListObjects.QVO[i]).PyHandle.GetSelf;
        PyList_Append( prResult, PyData );
      end;
    end;
  end;
  PythonThread.EndPython;
end;

procedure TQUB_Python_DataView.prGetIndexData;
begin
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    if (prIndex >= 0) and (prIndex < DataCount) then begin
      prResult := TQUB_QVO_QVD(ViewportData.ListObjects[prIndex]).PyHandle.GetSelf;
      Py_INCREF( prResult );
    end else begin
      prResult := nil;
    end;
  end;
  PythonThread.EndPython;
end;

function  TQUB_Python_DataView.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do begin
    if key = 'currData' then begin
      PythonThread.SyncGUIProc( prGetCurrData );
      if prResult <> nil then begin
        Result := prResult;
      end else begin
        PyErr_SetString( PyExc_IndexError^, PChar('no data') );
        Result := nil;
      end;
    end else if key = 'currIndex' then begin
      PythonThread.SyncGUIProc(prGetCurrIndex);
      Result := prResult;
    end else if key = 'selectedData' then begin
      PythonThread.SyncGUIProc( prGetSelectedData );
      Result := prResult;
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

function  TQUB_Python_DataView.SetAttr(key : PChar; value : PPyObject) : Integer;
var
  i                       :Integer;
begin
  Result := 0;
  with GetPythonEngine do begin
    if key = 'currIndex' then begin
      if PyArg_Parse( value, 'i:DataView.currIndex', [@i] ) = 0 then begin
        Result := -1;
      end else begin
        prIndexToSet := i;
        PythonThread.SyncGUIProc(prSetCurrIndex);
      end;
    end else begin
      Result := inherited SetAttr(key, value);
    end;
  end;
end;

procedure TQUB_Python_DataView.prSetCurrIndex;
begin
  PythonThread.BeginPython;
  try
    ViewportData.ListObjects.SelQVOAt(prIndexToSet);
  finally
    PythonThread.EndPython;
  end;
end;

procedure TQUB_Python_DataView.prGetCurrIndex;
begin
  PythonThread.BeginPython;
  try
    with GetPythonEngine do begin
      prResult := PyInt_FromLong( ViewportData.ListObjects.ItemIndex );
    end;
  finally
    PythonThread.EndPython;
  end;
end;

function  TQUB_Python_DataView.SqLength : Integer;
begin
  Result := DataCount;
end;

function  TQUB_Python_DataView.SqItem( idx : Integer ) : PPyObject;
begin
  prIndex := idx;
  PythonThread.SyncGUIProc( prGetIndexData );
  Result := prResult;
  if Result = nil then with GetPythonEngine do begin
    PyErr_SetString( PyExc_IndexError^, PChar('index out of range') );
  end;
end;

function TQUB_Python_DataView.CopyImage( args : PPyObject ) : PPyObject;
begin
  Adjust(@Self);

  prCopyImageConfig    := nil;
  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, '|O:DataView.CopyImage( configDict )', [@prCopyImageConfig] ) <> 0 then begin
      PythonThread.SyncGUIProc( prCopyImage );
      Result := GetPythonEngine.ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;

procedure TQUB_Python_DataView.prCopyImage;
var
  PyVal                  :PPyObject;
  PyInt                  :PPyObject;
  SavedPars              :INode;
  NewPars                :INode;
begin
  PythonThread.BeginPython;
  try
    if ViewportData.CurrData <> nil then with GetPythonEngine do begin
      SavedPars     := ProfileMgr['CopyData'].Active.Clone(True);
      NewPars       := TNode.Create('Properties');

      if (prCopyImageConfig <> nil) and (PyMapping_Check(prCopyImageConfig) <> 0) then begin
        if PyMapping_HasKeyString(prCopyImageConfig, 'DrawData') <> 0 then begin
          PyVal                 := PyMapping_GetItemString(prCopyImageConfig, 'DrawData');
          NewPars['Draw Data'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'DrawIDLData') <> 0 then begin
          PyVal                 := PyMapping_GetItemString(prCopyImageConfig, 'DrawIDLData');
          NewPars['Draw IDL Data'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'DrawBase') <> 0 then begin
          PyVal                 := PyMapping_GetItemString(prCopyImageConfig, 'DrawBase');
          NewPars['Draw Base'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'ScaleBars') <> 0 then begin
          PyVal                 := PyMapping_GetItemString(prCopyImageConfig, 'ScaleBars');
          NewPars['Scale Bars'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'ScaleBarsTop') <> 0 then begin
          PyVal                 := PyMapping_GetItemString(prCopyImageConfig, 'ScaleBarsTop');
          NewPars['Scale Bars Location'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'DrawFrame') <> 0 then begin
          PyVal               := PyMapping_GetItemString(prCopyImageConfig, 'DrawFrame');
          NewPars['Draw Frame'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'ShowTitles') <> 0 then begin
          PyVal               := PyMapping_GetItemString(prCopyImageConfig, 'ShowTitles');
          NewPars['Show Titles'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'Landscape') <> 0 then begin
          PyVal               := PyMapping_GetItemString(prCopyImageConfig, 'Landscape');
          NewPars['Orientation'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'Bitmap') <> 0 then begin
          PyVal               := PyMapping_GetItemString(prCopyImageConfig, 'Bitmap');
          NewPars['Format'].SetData(PyObject_IsTrue(PyVal));
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'Width') <> 0 then begin
          PyVal               := PyMapping_GetItemString(prCopyImageConfig, 'Width');
          if PyNumber_Check( PyVal ) <> 0 then begin
            PyInt             := PyNumber_Int( PyVal );
            NewPars['Draw Width'].SetData(PyInt_AsLong(PyInt));
            Py_DECREF( PyInt );
          end;
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'Height') <> 0 then begin
          PyVal               := PyMapping_GetItemString(prCopyImageConfig, 'Height');
          if PyNumber_Check( PyVal ) <> 0 then begin
            PyInt               := PyNumber_Int( PyVal );
            NewPars['Draw Height'].SetData(PyInt_AsLong(PyInt));
            Py_DECREF( PyInt );
          end;
          Py_DECREF( PyVal );
        end;
        if PyMapping_HasKeyString(prCopyImageConfig, 'HiRes') <> 0 then begin
          PyVal               := PyMapping_GetItemString(prCopyImageConfig, 'HiRes');
          ViewportData.CurrData.ActOnHiRes := PyObject_IsTrue( PyVal ) <> 0;
          Py_DECREF( PyVal );
        end else begin
          ViewportData.CurrData.ActOnHiRes := False;
        end;
      end;

      PythonThread.EndPython;

      ProfileMgr['CopyData'].SetProperties(NewPars);
      ViewportData.CurrData.CopyImage(False);
      ProfileMgr['CopyData'].SetProperties(SavedPars);

      PythonThread.BeginPython;
    end;
  finally
    PythonThread.EndPython;
  end;
end;

procedure TQUB_Python_DataView.prMoveSelectionIntoWindow;
var
  i                :Integer;
  NameNode         :INode;
  Name             :String;
  Data             :TQUB_QVO_QVD;
  Sel              :TQUB_FirstLast;
begin
  Data             := nil;

  // if filename specified, locate it in the list of data.  else use curr.
  NameNode         := prSegToMove.Parent;
  if not NameNode.IsNull then begin
    NameNode       := NameNode.Find('FileName');
    if not NameNode.IsNull then begin
      Name         := NameNode.DataAsString;

      for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
        if ViewportData.ListObjects.QVO[i].FileName = Name then begin
          Data     := TQUB_QVO_QVD(ViewportData.ListObjects.QVO[i]);
          Break;
        end;
      end;
    end;
  end;

  if NameNode.IsNull then Data := ViewportData.CurrData;
  if Data = nil then Exit;

  try
    Sel.First      := Integer( prSegToMove.Data[0] );
    Sel.Last       := Integer( prSegToMove.Data[1] );
    Data.MoveSelectionIntoWindow( Sel, Sel.First );
  except
  end;
end;

function TQUB_Python_DataView.MoveSelectionIntoWindow( Seg :INode ): INode; // aka QUB.ShowSeg( DS['Segment'] )
begin
  prSegToMove := Seg;
  PythonThread.SyncGUIProc( prMoveSelectionIntoWindow );
  Result := Seg;
  prSegToMove := nil;
end;

class procedure TQUB_Python_DataView.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do
    begin
      AddMethod( 'CopyImage', @TQUB_Python_DataView.CopyImage, 'DataView.CopyImage( configDict )' );
    end;
end;

// *****************************************************************************
type

TQUB_Runnable_Node = class(TQUB_Runnable)
public
  Node         :INode;

  constructor Create(ANode :INode);
end;

TQUB_Run_GetChosenData = class(TQUB_Runnable_Node)
public
  procedure Run; override;
end;

TQUB_Run_GetChosenIdl = class(TQUB_Runnable_Node)
public
  procedure Run; override;
end;

TQUB_Run_GetChosenIdlSignals = class(TQUB_Runnable_Node)
public
  procedure Run; override;
end;

TQUB_Run_FillData = class(TQUB_Runnable_Node)
public
  procedure Run; override;
end;

TQUB_Run_FillDataCh = class(TQUB_Runnable_Node)
public
  procedure Run; override;
end;


constructor TQUB_Runnable_Node.Create(ANode :INode);
begin
  Node := ANode;
end;


// *****************************************************************************
constructor TQUB_DataNoder.Create;
begin
  PyNode_PublishNodeFunc( TNodeMeth.Create( GetChosenData ),         'getChosenData' );
  PyNode_PublishNodeFunc( TNodeMeth.Create( GetChosenIdealization ), 'getChosenIdealization' );
  PyNode_PublishNodeFunc( TNodeMeth.Create( FillData ),              'fillData' );
  PyNode_PublishNodeFunc( TNodeMeth.Create( FillDataCh ),              'fillDataCh' );
end;

function TQUB_DataNoder.GetChosenData(Parent :INode): INode;
var
  Run   :TQUB_Run_GetChosenData;
begin
  Run   := TQUB_Run_GetChosenData.Create(Parent);
  try
    SyncGUIRunnable(Run);
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

procedure TQUB_Run_GetChosenData.Run;
var
  Parent                   :INode;
  Data                     :TQUB_QVO_QVD;
  DataSet                  :INode;
  w                        :Integer;
begin
  Parent := Node;

  Node := TNode.Create('DataSets');

  if not Parent.IsNull then Parent.AppendChild( Node );

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
  //File List

    for w := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
      Data          := TQUB_QVO_QVD( ViewportData.ListObjects[w] );
      if Data.Listed then begin
        Node.AppendChild( Data.GetChosenData );
      end;
    end;

    Node['DataDescription'].SetData('file list');

  end else begin
  //Current File

    Data         := ViewportData.CurrData;
    if Data = nil then Exit;

    DataSet      := Node.AppendChild( Data.GetChosenData );
    Node.AppendChild( DataSet.Find('DataDescription') );

  end;
end;

function TQUB_DataNoder.GetChosenIdealization(Parent :INode): INode;
var
  Run   :TQUB_Run_GetChosenIdl;
begin
  Run   := TQUB_Run_GetChosenIdl.Create(Parent);
  try
    SyncGUIRunnable(Run);
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

procedure TQUB_Run_GetChosenIdl.Run;
var
  Parent           :INode;
  w                :Integer;

  DataSet          :INode;

  Data             :TQUB_QVO_QVD;    // == ViewportData.CurrData
begin
  Parent           := Node;

  Node := TNode.Create('DataSets');
  if not Parent.IsNull then Parent.AppendChild( Node );

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
  //File List

    for w := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
      Data          := TQUB_QVO_QVD(ViewportData.ListObjects[w]);
      if Data.Listed then begin
        DataSet     := Node.AppendChild( Data.GetChosenIdl );
      end;
    end;

    Node['DataDescription'].SetData('file list');

  end else begin
  //Current File

    Data         := ViewportData.CurrData;
    if Data = nil then Exit;

    DataSet      := Node.AppendChild( Data.GetChosenIdl );
    Node.AppendChild( DataSet.Find('DataDescription') );

  end;//Data Source
end;

function TQUB_DataNoder.GetChosenIdlSignals(Parent :INode): INode;
var
  Run   :TQUB_Run_GetChosenIdlSignals;
begin
  Run   := TQUB_Run_GetChosenIdlSignals.Create(Parent);
  try
    SyncGUIRunnable(Run);
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

procedure TQUB_Run_GetChosenIdlSignals.Run;
var
  Parent           :INode;
  w                :Integer;

  DataSet          :INode;

  Data             :TQUB_QVO_QVD;    // == ViewportData.CurrData
begin
  Parent           := Node;

  Node := TNode.Create('DataSets');
  if not Parent.IsNull then Parent.AppendChild( Node );

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
  //File List

    for w := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
      Data          := TQUB_QVO_QVD(ViewportData.ListObjects[w]);
      if Data.Listed then begin
        DataSet     := Node.AppendChild( Data.GetChosenIdlSignals );
      end;
    end;

    Node['DataDescription'].SetData('file list');

  end else begin
  //Current File

    Data         := ViewportData.CurrData;
    if Data = nil then Exit;

    DataSet      := Node.AppendChild( Data.GetChosenIdlSignals );
    Node.AppendChild( DataSet.Find('DataDescription') );

  end;//Data Source
end;

function TQUB_DataNoder.FillData(SegNode :INode): INode;
var
  Run   :TQUB_Run_FillData;
begin
  Run   := TQUB_Run_FillData.Create(SegNode);
  try
    SyncGUIRunnable(Run);
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

procedure TQUB_Run_FillData.Run;
var
  SegNode         :INode;
  i               :Integer;
  FileName        :String;
  Data            :TQUB_QVO_QVD;
begin
  SegNode         := Node;
  if SegNode.Find('Channel').DataCount <> 0 then Exit;

  Data            := nil;
  FileName        := SegNode.Parent['FileName'].DataAsString;
  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    if ViewportData.ListObjects[i].FileName = FileName then Data := TQUB_QVO_QVD( ViewportData.ListObjects[i] );
  end;
  if Data = nil then Exit;

  Data.FillDataSegment(SegNode);
end;

function TQUB_DataNoder.FillDataCh(SegNode :INode): INode;
var
  Run   :TQUB_Run_FillDataCh;
begin
  Run   := TQUB_Run_FillDataCh.Create(SegNode);
  try
    SyncGUIRunnable(Run);
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

procedure TQUB_Run_FillDataCh.Run;
var
  SegNode         :INode;
  i               :Integer;
  FileName        :String;
  Data            :TQUB_QVO_QVD;
begin
  SegNode         := Node;
  if SegNode.Find('Channel').DataCount <> 0 then Exit;

  Data            := nil;
  FileName        := SegNode.Parent['FileName'].DataAsString;
  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    if ViewportData.ListObjects[i].FileName = FileName then Data := TQUB_QVO_QVD( ViewportData.ListObjects[i] );
  end;
  if Data = nil then Exit;

  Data.FillDataSegmentCh(SegNode);
end;


function TQUB_DataNoder.NullCallback(arg :INode): INode;
begin
  Result := arg;
end;


//******************************************************************************
function TQUB_DataNoderScript.GetChosenData(Parent :INode): INode;
var
  Run   :TQUB_Run_GetChosenData;
begin
  Run   := TQUB_Run_GetChosenData.Create(Parent);
  try
    Run.Run;
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

function TQUB_DataNoderScript.GetChosenIdealization(Parent :INode): INode;
var
  Run   :TQUB_Run_GetChosenIdl;
begin
  Run   := TQUB_Run_GetChosenIdl.Create(Parent);
  try
    Run.Run;
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

function TQUB_DataNoderScript.FillData(SegNode :INode): INode;
var
  Run   :TQUB_Run_FillData;
begin
  Run   := TQUB_Run_FillData.Create(SegNode);
  try
    Run.Run;
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

function TQUB_DataNoderScript.FillDataCh(SegNode :INode): INode;
var
  Run   :TQUB_Run_FillData;
begin
  Run   := TQUB_Run_FillData.Create(SegNode);
  try
    Run.Run;
  finally
    Result := Run.Node;
    Run.Free;
  end;
end;

//******************************************************************************
function GetDataSourcePy( args, kw :PPyObject ): PPyObject;
begin
  with GetPythonEngine do begin
    case QUB_MainForm.DataSource of
      DATA_SOURCE_WHOLEFILE:
        Result := PyString_FromString( PChar('file') );
      DATA_SOURCE_SELECTIONLIST:
        Result := PyString_FromString( PChar('list') );
      DATA_SOURCE_SELECTION:
        Result := PyString_FromString( PChar('sel') );
      else // DATA_SOURCE_FILELIST:
        Result := PyString_FromString( PChar('file list') );
      end;
  end;
end;

type TQUB_SetDataSource = class
public
  DataSource         :PChar;
  Err                :Boolean;

  constructor Create(DataSourceStr :PChar);

  procedure DoSetDataSource;
end;

constructor TQUB_SetDataSource.Create(DataSourceStr :PChar);
begin
  DataSource := DataSourceStr;
  Err        := False;
  SyncGUIProc(DoSetDataSource);
end;

procedure TQUB_SetDataSource.DoSetDataSource;
begin
  if DataSource = 'file' then begin
    QUB_MainForm.DataSource := DATA_SOURCE_WHOLEFILE;
  end else if DataSource = 'list' then begin
    QUB_MainForm.DataSource := DATA_SOURCE_SELECTIONLIST;
  end else if DataSource = 'sel' then begin
    QUB_MainForm.DataSource := DATA_SOURCE_SELECTION;
  end else if DataSource = 'file list' then begin
    QUB_MainForm.DataSource := DATA_SOURCE_FILELIST;
  end else begin
    Err := True;
  end;
end;

function SetDataSourcePy( args, kw :PPyObject ): PPyObject;
var
  DataSourceStr     :PChar;
  Setter            :TQUB_SetDataSource;
begin
  with GetPythonEngine do begin
    if PyArg_ParseTuple(args, 's:SetDataSource', [@DataSourceStr]) = 0 then begin
      PyErr_SetString( PyExc_TypeError^, PChar('QUB.SetDataSource(''sel'' or ''list'' or ''file'' or ''file list''') );
      Result := nil;
      Exit;
    end else begin
      Setter := TQUB_SetDataSource.Create(DataSourceStr);
      if Setter.Err then begin
        PyErr_SetString( PyExc_TypeError^, PChar('QUB.SetDataSource(''sel'' or ''list'' or ''file'' or ''file list''') );
        Result := nil;
        Setter.Free;
        Exit;
      end;
      Setter.Free;
    end;
    Result := ReturnNone;
  end;
end;


// *****************************************************************************
end.

