{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Python_Eigen;

// testing only: to route all qublib.eigen calls through scipy
// to enable: uncomment in PythonThread.prInitialize and QUBB.py

interface

uses

Alloc, AllocTypes;

type

BypassEigenFunc = function(N :Integer; A, Wr, Wi, V, Vinv :TPADouble): Integer; stdcall;


// call once, holding GIL
procedure Eigen_Scipy_Init;


// call NOT holding GIL:
function Eigen_Scipy(N :Integer; A, Wr, Wi, V, Vinv :TPADouble): Integer; stdcall;


implementation

uses

QUB_Viewport_Python, QUB_Python_Thread, PythonEngine;

var

g_pyEigen :PPyObject;

procedure QUBLIB_SetBypassEigen(Func :BypassEigenFunc); stdcall; external 'qublib.dll' name '_set_bypass_eigen@4';


procedure Eigen_Scipy_Init;
begin
  try
    with GetPythonEngine do begin
      g_pyEigen := EvalString('Eigen_Scipy');
    end;
    QUBLIB_SetBypassEigen(Eigen_Scipy);
  except
  end;
end;

function Eigen_Scipy(N :Integer; A, Wr, Wi, V, Vinv :TPADouble): Integer; stdcall;
var
  pyArgs     :PPyObject;
  pyResult   :PPyObject;
begin
  PythonThread.BeginPython;

  try
    with GetPythonEngine do begin
      pyArgs   := Py_BuildValue('(iiiiii)', [N, Integer(A), Integer(Wr), Integer(Wi), Integer(V), Integer(Vinv)]);
      pyResult := PyObject_CallObject(g_pyEigen, pyArgs);
      if pyResult = nil then begin
        PyErr_Print;
        Result     := -2;
      end else begin
        Result     := PyInt_AsLong(pyResult);
        Py_DECREF(PyResult);
      end;
      Py_DECREF(pyArgs);
    end;
  except
    Result := -3;
  end;

  PythonThread.EndPython;
end;


end.
