{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2004
// python-invokable custom modal dialog
// expects fields etc from Dialogs.py
// supersedes QUB_Python_Prompt


unit QUB_Python_ModalDialog;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
StdCtrls, ExtCtrls, Math,

PythonEngine;


type

//----- Base class for Python/Delphi controls
TQUB_Python_ModalDialog_Item = class
public

  Item               :PPyObject;
  Component          :TControl;
  Name               :String;

  constructor Create(AItem :PPyObject; AParent :TWinControl); virtual;
  destructor Destroy; override;

  function Value: PPyObject; virtual;

end;

//----- Label Control
TQUB_Python_ModalDialog_Label = class(TQUB_Python_ModalDialog_Item)
public

  constructor Create(AItem :PPyObject; AParent :TWinControl); override;

end;

//----- TextBox Control
TQUB_Python_ModalDialog_TextBox = class(TQUB_Python_ModalDialog_Item)
protected
  prAcceptor         :PPyObject;
  prFormat           :PPyObject;
  prCurrValue        :PPyObject;
  prLbl              :TLabel;
  prTxt              :TEdit;

  procedure prOnExit(Sender :TObject);
  procedure prOnKeyPress(Sender: TObject; var Key: Char);
  procedure prOnResize(Sender :TObject);
  procedure prSetText;

public

  constructor Create(AItem :PPyObject; AParent :TWinControl); override;
  destructor Destroy; override;

  function Value: PPyObject; override;

end;

//----- Checkbox control
TQUB_Python_ModalDialog_CheckBox = class(TQUB_Python_ModalDialog_Item)
public

  constructor Create(AItem :PPyObject; AParent :TWinControl); override;

  function Value: PPyObject; override;

end;

//----- Radio group control
TQUB_Python_ModalDialog_RadioGroup = class(TQUB_Python_ModalDialog_Item)
public

  constructor Create(AItem :PPyObject; AParent :TWinControl); override;

  function Value: PPyObject; override;

end;

//----- Modal Dialogue - collection of controls
TQUB_Python_ModalDialog = class(TForm)
  
  procedure FormDestroy(Sender: TObject);

protected

  prItems                        :TList;
  prResults                      :PPyObject;

  procedure prButtonClick(Sender: TObject);
  procedure prBuildResults;

public

  procedure Initialize(ACaption, Fields, Buttons :PPyObject);

  property Results               :PPyObject        read prResults;

end;

//----- Make Dialog available to Python
procedure InitPythonModalDialog;


//------------------------------------------------------------------------------
implementation


uses


QUB_Viewport_Python, QUB_Python_Thread, QUB_Node_Python;


{$R *.DFM}


const


XPAD                      = 10;
YPAD                      = 6;
RADIO_ITEM_HEIGHT         = 16;
RADIO_GROUP_HEIGHT        = 25;


// *****************************************************************************

type

TQUB_ModalDialog_PyFunc = class(TPyObject)
protected
  prCaption         :PPyObject;
  prButtons         :PPyObject;
  prFields          :PPyObject;
  prResults         :PPyObject;

  procedure prDoDialog;

public

  function  Call( args, kw : PPyObject) : PPyObject; override;

end;

function TQUB_ModalDialog_PyFunc.Call( args, kw :PPyObject ) : PPyObject;
var
  argc, argi         :Integer;
begin
  with GetPythonEngine do begin
    argc := PyTuple_Size(args);
    argi := 0;
    if argi < argc then begin
      prCaption := PyTuple_GetItem(args, argi);
      Inc(argi);
    end else if kw <> nil then begin
      prCaption := PyDict_GetItemString(kw, PChar('caption'));
    end else begin
      prCaption := nil;
    end;
    if argi < argc then begin
      prButtons := PyTuple_GetItem(args, argi);
      Inc(argi);
    end else if kw <> nil then begin
      prButtons := PyDict_GetItemString(kw, PChar('buttons'));
    end else begin
      prButtons := nil;
    end;
    if argi < argc then begin
      prFields := PyTuple_GetItem(args, argi);
    end else if kw <> nil then begin
      prFields := PyDict_GetItemString(kw, PChar('fields'));
    end else begin
      prFields := nil;
    end;

    PythonThread.SyncGUIProc( prDoDialog );

    Result := prResults;
  end;
end;

procedure TQUB_ModalDialog_PyFunc.prDoDialog;
var
  Form                     :TQUB_Python_ModalDialog;
begin
  Form                     := TQUB_Python_ModalDialog.Create(nil);

  Form.Initialize(prCaption, prFields, prButtons);
  Form.ShowModal;
  prResults                := Form.Results;
  PythonThread.BeginPython;
  GetPythonEngine.Py_INCREF(prResults);
  PythonThread.EndPython;

  Form.Release;
end;


var
  QUB_ModalDialog_PyFuncType      :TPythonType;


procedure InitPythonModalDialog;
var
  PromptFunc        :PPyObject;
begin
  QUB_ModalDialog_PyFuncType := TPythonType.Create(nil);
  QUB_ModalDialog_PyFuncType.Engine := GetPythonEngine;
  QUB_ModalDialog_PyFuncType.Module := PythonViewport.pmQUBTypes;
  QUB_ModalDialog_PyFuncType.TypeName := 'DialogFunc';
  QUB_ModalDialog_PyFuncType.Services.Basic := [bsCall];
  QUB_ModalDialog_PyFuncType.PyObjectClass := TQUB_ModalDialog_PyFunc;
  QUB_ModalDialog_PyFuncType.Initialize;

  PromptFunc := QUB_ModalDialog_PyFuncType.CreateInstance;
  PyNode_PublishPyObj( PromptFunc, 'Dialog' );
  GetPythonEngine.Py_DECREF( PromptFunc );
end;


// *****************************************************************************

constructor TQUB_Python_ModalDialog_Item.Create(AItem :PPyObject; AParent :TWinControl);
var
  PyJunk               :PPyObject;
begin
  with GetPythonEngine do begin
    Py_INCREF( AItem );
    Item               := AItem;
    PyJunk             := PyObject_GetAttrString(Item, PChar('name'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Self.Name        := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;
  end;
end;

destructor TQUB_Python_ModalDialog_Item.Destroy;
begin
  try
    GetPythonEngine.Py_DECREF( Item );
    if Component <> nil then Component.Free;
  finally
    inherited;
  end;
end;

function TQUB_Python_ModalDialog_Item.Value: PPyObject;
begin
  Result := GetPythonEngine.PyObject_GetAttrString(Item, PChar('value'));
  if Result = nil then GetPythonEngine.PyErr_Clear;
end;


constructor TQUB_Python_ModalDialog_Label.Create(AItem :PPyObject; AParent :TWinControl);
var
  PyJunk               :PPyObject;
  Lbl                  :TLabel;
begin
  inherited;

  Lbl                  := TLabel.Create(nil);
  Lbl.Parent           := AParent;
  Lbl.Caption          := '';
  Component            := Lbl;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(Item, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Lbl.Caption       := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;
  end;
end;


constructor TQUB_Python_ModalDialog_TextBox.Create(AItem :PPyObject; AParent :TWinControl);
var
  PyJunk               :PPyObject;
  Panel                :TPanel;
begin
  inherited;

  Panel                := TPanel.Create(nil);
  Panel.Parent         := AParent;
  Panel.Caption        := '';
  Panel.BevelOuter     := bvNone;
  Panel.OnResize       := prOnResize;
  Component            := Panel;

  prLbl                := TLabel.Create(Panel);
  prLbl.Parent         := Panel;
  prLbl.Caption        := '';

  prTxt                := TEdit.Create(Panel);
  prTxt.Parent         := Panel;
  prTxt.Text           := '';
  prTxt.OnExit         := prOnExit;
  prTxt.OnKeyPress     := prOnKeyPress;

  Panel.Height         := Max(prTxt.Height, prLbl.Height);

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(Item, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      prLbl.Caption     := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;

    PyJunk             := PyObject_GetAttrString(Item, PChar('acceptor'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      if 0 <> PyCallable_Check(PyJunk) then begin
        prAcceptor     := PyJunk;
      end else begin
        Py_DECREF( PyJunk );
      end;
    end;

    PyJunk             := PyObject_GetAttrString(Item, PChar('format'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      if 0 <> PyCallable_Check(PyJunk) then begin
        prFormat       := PyJunk;
      end else begin
        Py_DECREF( PyJunk );
      end;
    end;

    PyJunk             := PyObject_GetAttrString(Item, PChar('value'));
    if PyJunk = nil then begin
      PyErr_Clear();
      prCurrValue      := ReturnNone;
    end else begin
      prCurrValue      := PyJunk;
    end;

    prSetText;
  end;
end;

destructor TQUB_Python_ModalDialog_TextBox.Destroy;
begin
  try
    with GetPythonEngine do begin
      Py_XDECREF( prAcceptor );
      Py_XDECREF( prFormat );
      Py_XDECREF( prCurrValue );
    end;
  finally
    inherited;
  end;
end;

function TQUB_Python_ModalDialog_TextBox.Value: PPyObject;
begin
  GetPythonEngine.Py_INCREF( prCurrValue );
  Result := prCurrValue;
end;

procedure TQUB_Python_ModalDialog_TextBox.prOnExit(Sender :TObject);
var
  Args         :PPyObject;
  NewValue     :PPyObject;
begin
  PythonThread.BeginPython;

  with GetPythonEngine do begin
    if prAcceptor = nil then begin
      NewValue   := PyString_FromString( PChar( prTxt.Text ) );
    end else begin
      Args       := Py_BuildValue('(s)', [PChar( prTxt.Text )]);
      NewValue   := PyObject_CallObject( prAcceptor, Args );
      if NewValue = nil then begin
        MessageDlg('Invalid value; ' + PyErr_ShortMessage, mtError, [mbOK], 0);
        prSetText;
        NewValue := prCurrValue;
        Py_INCREF( NewValue );
      end;
      Py_DECREF(Args);
    end;

    Py_DECREF( prCurrValue );
    prCurrValue := NewValue;
  end;

  PythonThread.EndPython;
end;

procedure TQUB_Python_ModalDialog_TextBox.prOnKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    prOnExit(Sender);
  end;
end;

procedure TQUB_Python_ModalDialog_TextBox.prOnResize(Sender :TObject);
begin
  prLbl.Top            := 0;
  prLbl.Left           := 0;
  prLbl.Width          := TPanel(Component).ClientWidth div 2;
  prLbl.Height         := TPanel(Component).ClientHeight;
  prTxt.Top            := 0;
  prTxt.Left           := prLbl.Width;
  prTxt.Width          := prLbl.Width;
  prTxt.Height         := prLbl.Height;
end;

procedure TQUB_Python_ModalDialog_TextBox.prSetText;
var
  Args, Result   :PPyObject;
begin
  with GetPythonEngine do begin
    Args                 := nil;
    Result               := nil;
    try
      if prFormat <> nil then begin
        Args             := Py_BuildValue('(O)', [prCurrValue]);
        Result           := PyObject_CallObject( prFormat, Args );
        if Result = nil then begin
          PyErr_Print;
          PyErr_Clear;
        end else begin
          prTxt.Text       := PyObject_DelphiStr(Result);
          Exit;
        end;
      end;
    finally
      Py_XDECREF(Args);
      Py_XDECREF(Result);
    end;

    prTxt.Text       := PyObject_DelphiStr(prCurrValue);
  end;
end;




constructor TQUB_Python_ModalDialog_CheckBox.Create(AItem :PPyObject; AParent :TWinControl);
var
  Chk                  :TCheckBox;
  PyJunk               :PPyObject;
begin
  inherited;

  Chk                  := TCheckBox.Create(nil);
  Chk.Parent           := AParent;
  Chk.Caption          := '';
  Chk.Checked          := False;
  Chk.Alignment        := taRightJustify;
  Component            := Chk;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(Item, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Chk.Caption      := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;

    PyJunk             := GetPythonEngine.PyObject_GetAttrString(Item, PChar('value'));
    Chk.Checked        := 0 <> PyObject_IsTrue(PyJunk);
    Py_DECREF(PyJunk);
  end;
end;

function TQUB_Python_ModalDialog_CheckBox.Value: PPyObject;
begin
  with GetPythonEngine do begin
    if TCheckBox(Component).Checked then begin
      Result := PyInt_FromLong( 1 );
    end else begin
      Result := PyInt_FromLong( 0 );
    end;
  end;
end;


constructor TQUB_Python_ModalDialog_RadioGroup.Create(AItem :PPyObject; AParent :TWinControl);
var
  Rad               :TRadioGroup;
  PyJunk, PyItems   :PPyObject;
  InitVal           :String;
  i                 :Integer;
begin
  inherited;

  Rad               := TRadioGroup.Create(nil);
  Rad.Parent        := AParent;
  Rad.Caption       := '';
  Component         := Rad;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(Item, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Rad.Caption   := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;

    PyItems         := PyObject_GetAttrString(Item, PChar('choices'));
    if PyItems = nil then begin
      PyErr_Clear();
    end else if 0 <> PySequence_Check(PyItems) then begin
      for i := 0 to PySequence_Length(PyItems) - 1 do begin
        PyJunk      := PySequence_GetItem(PyItems, i);
        Rad.Items.Add( PyObject_DelphiStr(PyJunk) );
        Py_DECREF( PyJunk );
      end;
      Py_DECREF( PyItems );
    end;

    InitVal         := '';
    if Rad.Items.Count > 0 then InitVal := Rad.Items[0];

    PyJunk          := PyObject_GetAttrString(Item, PChar('value'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      InitVal       := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;

    i := Rad.Items.IndexOf( InitVal );
    if i < 0 then i := 0;
    if Rad.Items.Count > 0 then Rad.ItemIndex := i;

    Rad.Height := RADIO_GROUP_HEIGHT + Rad.Items.Count * RADIO_ITEM_HEIGHT;
  end;
end;

function TQUB_Python_ModalDialog_RadioGroup.Value: PPyObject;
var
  Rad            :TRadioGroup;
begin
  Rad            := TRadioGroup(Component);
  if Rad.ItemIndex < 0 then begin
    Result       := inherited Value;
  end else begin
    Result := GetPythonEngine.PyString_FromString( PChar( Rad.Items[ Rad.ItemIndex ] ) );
  end;
end;


// *****************************************************************************


procedure TQUB_Python_ModalDialog.Initialize(ACaption, Fields, Buttons :PPyObject);
var
  Item             :TQUB_Python_ModalDialog_Item;
  i                :Integer;
  x, y             :Integer;
  PyItem           :PPyObject;
  PyJunk           :PPyObject;
  ItemType         :String;
  Button           :TButton;
begin
  PythonThread.BeginPython;

  prItems          := TList.Create;
  Button           := nil;
  
  with GetPythonEngine do begin
    if ACaption <> nil then begin
      Caption      := PyObject_DelphiStr( ACaption );
    end;

    y              := YPAD;
    if (Fields <> nil) and (0 <> PySequence_Check(Fields)) then begin
      for i := 0 to PySequence_Length(Fields) - 1 do begin
        PyItem         := PySequence_GetItem(Fields, i);
        ItemType       := 'bullshit';
        Item           := nil;

        PyJunk         := PyObject_GetAttrString(PyItem, PChar('type'));
        if PyJunk = nil then begin
          PyErr_Clear();
        end else begin
          ItemType     := PyObject_DelphiStr(PyJunk);
          Py_DECREF( PyJunk );
        end;

        if ItemType = 'hidden' then begin
          Item         := TQUB_Python_ModalDialog_Item.Create(PyItem, Self);
        end else if ItemType = 'label' then begin
          Item         := TQUB_Python_ModalDialog_Label.Create(PyItem, Self);
        end else if ItemType = 'textbox' then begin
          Item         := TQUB_Python_ModalDialog_TextBox.Create(PyItem, Self);
        end else if ItemType = 'checkbox' then begin
          Item         := TQUB_Python_ModalDialog_CheckBox.Create(PyItem, Self);
        end else if ItemType = 'radiogroup' then begin
          Item         := TQUB_Python_ModalDialog_RadioGroup.Create(PyItem, Self);
        end;

        if Item <> nil then begin
          prItems.Add(Item);

          if Item.Component <> nil then begin
            Item.Component.Left   := XPAD;
            Item.Component.Width  := ClientWidth - 2 * XPAD;
            Item.Component.Top    := y;
            y                     := y + Item.Component.Height + YPAD;
          end;
        end;
        
        Py_DECREF( PyItem );
      end;
    end;

    y                  := y + YPAD;
    x                  := XPAD;
    if (Buttons <> nil) and (0 <> PySequence_Check(Buttons)) then begin
      for i := 0 to PySequence_Length(Buttons) - 1 do begin
        PyJunk         := PySequence_GetItem(Buttons, i);
        Button         := TButton.Create(Self);
        Button.Caption := PyObject_DelphiStr( PyJunk );
        Py_DECREF( PyJunk );

        Button.OnClick := prButtonClick;
        Button.Parent  := Self;
        Button.Top     := y;
        Button.Left    := x;
        x              := x + Button.Width + XPAD;
      end;
    end;

    if x = XPAD {no buttons} then begin
      Button           := TButton.Create(Self);
      Button.Caption   := 'OK';
      Button.OnClick   := prButtonClick;
      Button.Parent    := Self;
      Button.Top       := y;
      Button.Left      := x;
    end;
  end;

  Height               := y + Button.Height + YPAD + 20;

  PythonThread.EndPython;
end;

procedure TQUB_Python_ModalDialog.FormDestroy(Sender: TObject);
var
  i           :Integer;
begin
  PythonThread.BeginPython;

  GetPythonEngine.Py_XDECREF( prResults );

  for i := 0 to prItems.Count - 1 do begin
    TQUB_Python_ModalDialog_Item( prItems[i] ).Free;
  end;

  prItems.Free;

  PythonThread.EndPython;
end;

procedure TQUB_Python_ModalDialog.prButtonClick(Sender: TObject);
var
  PyJunk        :PPyObject;
begin
  PythonThread.BeginPython;

  prBuildResults;
  with GetPythonEngine do begin
    PyJunk      := PyString_FromString( PChar( TButton(Sender).Caption ) );
    PyMapping_SetItemString(prResults, PChar('button'), PyJunk);
    Py_DECREF( PyJunk );
  end;

  PythonThread.EndPython;

  Close;
end;

procedure TQUB_Python_ModalDialog.prBuildResults;
var
  i                :Integer;
  Item             :TQUB_Python_ModalDialog_Item;
  PyJunk           :PPyObject;
begin
  with GetPythonEngine do begin
    Py_XDECREF( prResults );
    prResults  := PyDict_New();

    for i := 0 to prItems.Count - 1 do begin
      Item         := TQUB_Python_ModalDialog_Item(prItems[i]);
      PyJunk       := Item.Value;
      if PyJunk <> nil then begin
        PyDict_SetItemString(prResults, PChar(Item.Name), PyJunk);
        Py_DECREF(PyJunk);
      end;
    end;
  end;
end;



// *****************************************************************************
end.
