{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// python: QUB.ModelView
// maybe this should be at the bottom of QUB_Viewport_Model?


unit QUB_Python_Model;

interface

uses

QUB_Viewport_Model,
PythonEngine,
QUB_Python_Thread;



procedure QUB_Python_ModelView_Init(ViewportModel :TQUB_Viewport_Model);


type

TQUB_Python_ModelView = class(TPyObject)
protected
  prViewportModel     :TQUB_Viewport_Model;

  prIndexToSet        :Integer;       //- used to store model index to be used by prSetCurrIndex - GUI thread
  prResult            :PPyObject;
  prIResult           :Integer;
  prName              :String;

  procedure prGetCurrIndex;
  procedure prSetCurrIndex;           //- Change model - execute on GUI thread
  procedure prGetIndexModel;
  procedure prGetCurrModel;
  procedure prGetModelCount;
  procedure prMakeNewRate;
  procedure prNewModel;
  procedure prChooseModel;

public
  //- Python attributes - generic + GetAttr('currModel'),('currIndex'), SetAttr('currIndex')
  function  GetAttr(key : PChar) : PPyObject; override;
  function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  function  SqLength: Integer; override;
  function  SqItem( idx : Integer ) : PPyObject; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;

  function NewModel(args :PPyObject): PPyObject; cdecl;
  function ChooseModel(args :PPyObject): PPyObject; cdecl;
  // should add save, saveAs, close

  function MakeNewRate(args :PPyObject): PPyObject; cdecl;

  property ViewportModel :TQUB_Viewport_Model  read prViewportModel   write prViewportModel;
end;


implementation


uses

QUB_Main,
QUB_Viewport_Python,
QUB_Node_Python,
QUB_QVO_QVM,
QUB_QVO_QVM_Print_Properties;


var
  ModelViewPyType                    :TPythonType;


procedure QUB_Python_ModelView_Init(ViewportModel :TQUB_Viewport_Model);
var
  PyInst                  :PPyObject;
  ModelViewObj            :TQUB_Python_ModelView;
begin
  ModelViewPyType := TPythonType.Create(nil);
  ModelViewPyType.Engine := GetPythonEngine;
  ModelViewPyType.Module := PythonViewport.pmQUBTypes;
  ModelViewPyType.TypeName := 'ModelViewType';
  ModelViewPyType.Services.Basic := [bsGetAttr, bsSetAttr];
  ModelViewPyType.Services.Sequence := [ssLength, ssItem];
  ModelViewPyType.PyObjectClass := TQUB_Python_ModelView;
  ModelViewPyType.Initialize;

  PyInst := ModelViewPyType.CreateInstance;
  ModelViewObj := TQUB_Python_ModelView( PythonToDelphi(PyInst) );
  ModelViewObj.ViewportModel := ViewportModel;
  PyNode_PublishPyObj( PyInst, 'ModelView' );
  GetPythonEngine.Py_DECREF( PyInst );

  ModelPyType := TPythonType.Create(nil);
  ModelPyType.Engine := GetPythonEngine;
  ModelPyType.Module := PythonViewport.pmQUBTypes;
  ModelPyType.TypeName := 'ModelType';
  ModelPyType.PyObjectClass := TQUB_Python_Model;
  ModelPyType.Initialize;
end;

function  TQUB_Python_ModelView.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do begin
    if key = 'currModel' then begin
      PythonThread.SyncGUIProc(prGetCurrModel);
      if prResult <> nil then begin
        Result := prResult;
      end else begin
        Result := ReturnNone;
      end;
    end else if key = 'currIndex' then begin
      PythonThread.SyncGUIProc(prGetCurrIndex);
      Result := prResult;
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

function  TQUB_Python_ModelView.SetAttr(key : PChar; value : PPyObject) : Integer;
var
  i                       :Integer;
begin
  Result := 0;
  with GetPythonEngine do begin
    if key = 'currIndex' then begin
      if PyArg_Parse( value, 'i:ModelView.currIndex', [@i] ) = 0 then begin
        Result := -1;
      end else begin
        prIndexToSet := i;
        PythonThread.SyncGUIProc(prSetCurrIndex);
      end;
    end else begin
      Result := inherited SetAttr(key, value);
    end;
  end;
end;

procedure TQUB_Python_ModelView.prGetIndexModel;
begin
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    if (prIndexToSet >= 0) and (prIndexToSet < prViewportModel.ModelCount) then begin
      prResult := TQUB_QVO_QVM(prViewportModel.ListObjects[prIndexToSet]).PyHandle.GetSelf;
      Py_INCREF( prResult );
    end else begin
      prResult := nil;
    end;
  end;
  PythonThread.EndPython;
end;

procedure TQUB_Python_ModelView.prGetCurrModel;
begin
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    if prViewportModel.CurrModel <> nil then begin
      prResult := prViewportModel.CurrModel.PyHandle.GetSelf;
      Py_INCREF( prResult );
    end else begin
      prResult := nil;
    end;
    PythonThread.EndPython;
  end;
end;


procedure TQUB_Python_ModelView.prSetCurrIndex;
begin
  PythonThread.BeginPython;
  try
    prViewportModel.ListObjects.SelQVOAt(prIndexToSet);
    {
    with prViewportModel do begin
      ListObjects.ItemIndex := prIndexToSet;
      for i := 0 to ModelCount - 1 do begin
        ListObjects.Selected[i] := i = prIndexToSet;
      end;
    end;}
  finally
    PythonThread.EndPython;
  end;
end;

procedure TQUB_Python_ModelView.prGetCurrIndex;
begin
  PythonThread.BeginPython;
  try
    with GetPythonEngine do begin
      prResult := PyInt_FromLong( prViewportModel.ListObjects.ItemIndex );
    end;
  finally
    PythonThread.EndPython;
  end;
end;

function  TQUB_Python_ModelView.SqLength: Integer;
begin
  with GetPythonEngine do begin
    PythonThread.SyncGUIProc(prGetModelCount);
    Result := prIResult;
  end;
end;

procedure TQUB_Python_ModelView.prGetModelCount;
begin
  prIResult := prViewportModel.ModelCount;
end;

function  TQUB_Python_ModelView.SqItem( idx : Integer ) : PPyObject;
begin
  with GetPythonEngine do begin
    prIndexToSet := idx;
    PythonThread.SyncGUIProc(prGetIndexModel);
    if prResult <> nil then begin
      Result := prResult;
    end else begin
      PyErr_SetString( PyExc_IndexError^, PChar('index out of range') );
      Result := nil;
    end;
  end;
end;

class procedure TQUB_Python_ModelView.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do begin
    AddMethod( 'makeNewRate',     @TQUB_Python_ModelView.MakeNewRate,  'ModelView.makeNewRate()' );
    AddMethod( 'New',             @TQUB_Python_ModelView.NewModel,     'ModelView.New( name )' );
    AddMethod( 'Choose',          @TQUB_Python_ModelView.ChooseModel,  'ModelView.Choose( index or object )' );
  end;
end;

function TQUB_Python_ModelView.MakeNewRate(args :PPyObject): PPyObject; cdecl;
begin
  Adjust(@Self);

  PythonThread.SyncGUIProc( prMakeNewRate );
  if prResult <> nil then begin
    Result := prResult;
  end else begin
    Result := GetPythonEngine.ReturnNone;
  end;
end;

procedure TQUB_Python_ModelView.prMakeNewRate;
begin
  PythonThread.BeginPython;
  if prViewportModel.CurrModel <> nil then begin
    prResult := PyNode_FromNode( prViewportModel.CurrModel.MakeNewRateNode );
  end else begin
    prResult := nil;
  end;
  PythonThread.EndPython;
end;


function TQUB_Python_ModelView.NewModel(args :PPyObject): PPyObject; cdecl;
var
  FName       :PChar;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    FName      := nil;
    if 0 = PyArg_ParseTuple(args, '|s', [@FName]) then begin
      PyErr_SetString( PyExc_TypeError^, PChar('name must be a string') );
      Result   := nil;
      Exit;
    end;
    if FName <> nil then begin
      prName   := FName;
    end else begin
      prName   := '';
    end;

    PythonThread.SyncGUIProc( prNewModel );

    Result := PyBool_FromLong( prIResult );
  end;
end;

procedure TQUB_Python_ModelView.prNewModel;
begin
  prIResult := Integer( prViewportModel.New(prName, True{Silent}) );
end;

function TQUB_Python_ModelView.ChooseModel(args :PPyObject): PPyObject; cdecl;
var
  Arg         :PPyObject;
  i           :Integer;
  ix          :Integer;
  M           :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if 0 = PyArg_ParseTuple(args, 'O', [@Arg]) then begin
      Result   := nil;
      Exit;
    end;
    if PyInt_Check(Arg) then begin
      prIndexToSet := PyInt_AsLong(Arg);
    end else begin
      ix := -1;
      for i := 0 to SqLength - 1 do begin
        M   := SqItem(i);
        if Arg = M then begin
          ix := i;
        end;
        Py_DECREF(M);
        if ix >= 0 then begin
          prIndexToSet := ix;
          Break;
        end;
      end;
    end;
    if (prIndexToSet >= 0) and (prIndexToSet < SqLength) then begin
      PythonThread.SyncGuiProc(prChooseModel);
    end;
    Result := ReturnNone; // ok should raise indexerror if out of range
  end;
end;

procedure TQUB_Python_ModelView.prChooseModel;
begin
  prViewportModel.ListObjects.SelQVOAt(prIndexToSet);
end;


// *****************************************************************************
end.



