{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// first attempt at python custom dialog, before I knew anything.


unit QUB_Python_Prompt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

//{$include 'QUB_Python_Prompt.hct'}

type
  TQUB_Python_Prompt = class(TForm)
    lbMessage: TLabel;

    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);

  protected

    prBottomComponent              :TControl;
    prLastButton                   :TControl;

    prTextQueries                  :TStringList; // -> TEdit
    prRadioGroups                  :TStringList; // -> TRadioGroup
    prButtons                      :TStringList; // -> TButton

    prChosenButton                 :String;

    procedure prSetMessage(m :String); // lbMessage.Caption := p;

    procedure prAddLine(lineHeight :Integer; var linePos :Integer);

  public

    procedure ButtonClick(Sender: TObject);

    procedure AddTextQuery(prompt :String; value :String);
    procedure AddRadioQuery(prompt :String; values :TStrings; chosen :String);
    procedure AddButton(caption :String);

    function GetValue(prompt :String) :String;
    function GetButton: String;

    property TopMessage     :String                           write prSetMessage;


  end;


procedure InitPythonPrompt;


implementation


uses

{$IFDEF DELPHI6_OR_HIGHER}
Variants,
{$ENDIF}
PythonEngine, QUB_Viewport_Python, QUB_Python_Thread, QUB_Node_Python,
PythonAtom, AtomPythonEngine;


{$R *.DFM}



const
  XPAD                      = 10;
  YPAD                      = 6;
  RADIO_ITEM_HEIGHT         = 16;
  RADIO_GROUP_HEIGHT        = 25;


// *****************************************************************************

type

TQUB_Prompt_PyFunc = class(TPyObject)
protected
  prMsg             :String;
  prButtons         :PPyObject;
  prFields          :PPyObject;
  prResults         :PPyObject;

  procedure prDoPrompt;

public
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;

function TQUB_Prompt_PyFunc.Call( args, kw :PPyObject ) : PPyObject;
var
  argc               :Integer;
  msg                :PPyObject;
  msgStr             :String;
  btns               :PPyObject;
  ownBtns            :Boolean;
  fields             :PPyObject;
  ownFields          :Boolean;
begin
  with GetPythonEngine do begin
    argc := PyTuple_Size(args);
    if argc > 0 then begin
      msg := PyTuple_GetItem(args, 0);
    end else if kw <> nil then begin
      msg := PyDict_GetItemString(kw, PChar('message'));
    end else msg := nil;
    if argc > 1 then begin
      btns := PyTuple_GetItem(args, 1);
    end else if kw <> nil then begin
      btns := PyDict_GetItemString(kw, PChar('buttons'));
    end else btns := nil;
    if argc > 2 then begin
      fields := PyTuple_GetItem(args, 2);
    end else if kw <> nil then begin
      fields := PyDict_GetItemString(kw, PChar('fields'));
    end else fields := nil;

    if msg <> nil then msgStr := PyString_AsString(msg) else msgStr := '';
    if btns = nil then begin
      ownBtns := True;
      btns := PyTuple_New(1);
      PyTuple_SetItem(btns, 0, PyString_FromString(PChar('OK')));
    end else ownBtns := False;
    if fields = nil then begin
      ownFields := True;
      fields := PyList_New(0);
    end else ownFields := False;

    prMsg           := msgStr;
    prButtons       := btns;
    prFields        := fields;
    prResults       := nil;

    PythonThread.SyncGUIProc( prDoPrompt );

    Result := prResults;

    if ownBtns   then Py_DECREF(btns);
    if ownFields then Py_DECREF(fields);
  end;
end;

procedure TQUB_Prompt_PyFunc.prDoPrompt;
var
  form                        :TQUB_Python_Prompt;
  field                       :OleVariant;
  fieldCount                  :Integer;
  minChoice, maxChoice        :Integer;
  choices                     :TStringList;
  i,j                         :Integer;
  ss                          :TStringList;
  btnStr                      :PPyObject;
  pname, pval                 :PPyObject;
begin
  PythonThread.BeginPython;

  with GetPythonEngine do begin
    form := TQUB_Python_Prompt.Create(PythonViewport);
    form.TopMessage := prMsg;

    if PyList_Check(prFields)
    then fieldCount := PyList_Size(prFields)
    else if PyTuple_Check(prFields)
    then fieldCount := PyTuple_Size(prFields)
    else fieldCount := 0;

    for i := 0 to fieldCount - 1 do begin
      try
        if PyList_Check(prFields)
        then field := getAtom(PyList_GetItem(prFields, i))
        else field := getAtom(PyTuple_GetItem(prFields, i));

        minChoice := VarArrayLowBound(field.choices,1);
        maxChoice := VarArrayHighBound(field.choices,1);
        if (maxChoice - minChoice + 1) > 0 then begin
          choices := TStringList.Create;
          for j := minChoice to maxChoice do begin
            choices.Add(field.choices[j]);
          end;
          form.AddRadioQuery(field.text, choices, field.value);
          choices.Free;
        end else begin
          form.AddTextQuery(field.text, field.value);
        end;
      except
      end;
    end;

    ss := TStringList.Create;
    if PyTuple_Check(prButtons) then begin
      PyTupleToStrings(prButtons, ss);
    end else if PyList_Check(prButtons) then begin
      PyListToStrings(prButtons, ss);
    end;

    for i := 0 to ss.Count - 1 do begin
      form.AddButton(ss[i]);
    end;
    ss.Free;

    VarClear(field);
    PythonThread.EndPython;
    form.ShowModal;
    PythonThread.BeginPython;

    prResults := PyDict_New();

    btnStr := PyString_FromString(PChar(form.GetButton()));
    PyDict_SetItemString( prResults, PChar('button'), btnStr );
    Py_DECREF(btnStr);

    for i := 0 to fieldCount - 1 do begin
      try
        if PyList_Check(prFields)
        then field := getAtom(PyList_GetItem(prFields, i))
        else field := getAtom(PyTuple_GetItem(prFields, i));
        pname      := VariantAsPyObject(field.name);
        pval       := VariantAsPyObject(form.getValue(field.text));
        PyDict_SetItem( prResults, pname, pval );
        Py_XDECREF( pname );
        Py_XDECREF( pval );
      except
      end;
    end;

    form.Release;
  end;

  VarClear(field);
  PythonThread.EndPython;
  
end;


var
  QUB_Prompt_PyFuncType      :TPythonType;


procedure InitPythonPrompt;
var
  PromptFunc        :PPyObject;
begin
  QUB_Prompt_PyFuncType := TPythonType.Create(nil);
  QUB_Prompt_PyFuncType.Engine := GetPythonEngine;
  QUB_Prompt_PyFuncType.Module := PythonViewport.pmQUBTypes;
  QUB_Prompt_PyFuncType.TypeName := 'PromptFunc';
  QUB_Prompt_PyFuncType.Services.Basic := [bsCall];
  QUB_Prompt_PyFuncType.PyObjectClass := TQUB_Prompt_PyFunc;
  QUB_Prompt_PyFuncType.Initialize;

  PromptFunc := QUB_Prompt_PyFuncType.CreateInstance;
  PyNode_PublishPyObj( PromptFunc, 'Prompt' );
  GetPythonEngine.Py_DECREF( PromptFunc );
end;


// *****************************************************************************

procedure TQUB_Python_Prompt.FormCreate(Sender: TObject);
begin
  prBottomComponent := lbMessage;
  prLastButton      := nil;

  prTextQueries := TStringList.Create;
  prRadioGroups := TStringList.Create;
  prButtons := TStringList.Create;
  //{$include 'QUB_Python_Prompt.htx'}
end;

procedure TQUB_Python_Prompt.FormShow(Sender: TObject);
begin
  if prLastButton <> nil then prBottomComponent := prLastButton;
  ClientHeight := prBottomComponent.Top + prBottomComponent.Height + YPAD;
end;

procedure TQUB_Python_Prompt.FormDestroy(Sender: TObject);
begin
  prTextQueries.Free;
  prRadioGroups.Free;
  prButtons.Free;
end;

procedure TQUB_Python_Prompt.prSetMessage(m :String);
begin
  lbMessage.Caption := m;
end;

procedure TQUB_Python_Prompt.prAddLine(lineHeight :Integer; var linePos :Integer);
begin
  Self.Height := Self.Height + lineHeight + YPAD;
  linePos := prBottomComponent.Top + prBottomComponent.Height + YPAD;
end;

procedure TQUB_Python_Prompt.ButtonClick(Sender: TObject);
begin
  prChosenButton := TButton(Sender).Caption;
  Close;
end;

procedure TQUB_Python_Prompt.AddTextQuery(prompt :String; value :String);
var
  NameC             :TLabel;
  ValueC            :TEdit;
  Top               :Integer;
  HalfWidth         :Integer;

begin
  HalfWidth := ClientWidth div 2 - XPAD;

  NameC := TLabel.Create(Self);
  NameC.Caption := prompt;
  NameC.Parent := Self;
  NameC.Left := XPAD;
  NameC.Width := HalfWidth;

  ValueC := TEdit.Create(Self);
  ValueC.Text := value;
  ValueC.Parent := Self;
  ValueC.Left := ClientWidth div 2;
  ValueC.Width := HalfWidth;

  prAddLine(ValueC.Height, Top);
  NameC.Top := Top;
  ValueC.Top := Top;

  prTextQueries.AddObject(prompt, ValueC);

  prBottomComponent := ValueC;
end;

procedure TQUB_Python_Prompt.AddRadioQuery(prompt :String; values :TStrings; chosen :String);
var
  Top            :Integer;
  GroupC         :TRadioGroup;
  i              :Integer;
begin
  GroupC := TRadioGroup.Create(Self);
  GroupC.Caption := prompt;
  GroupC.Parent := Self;
  prRadioGroups.AddObject(prompt, GroupC);

  for i := 0 to values.Count - 1 do begin
    GroupC.Items.Add( values[i] );
    if values[i] = chosen then GroupC.ItemIndex := i;
  end;

  GroupC.Width := Self.ClientWidth - XPAD;
  GroupC.Height := RADIO_GROUP_HEIGHT + values.Count * RADIO_ITEM_HEIGHT;
  if GroupC.ItemIndex < 0 then GroupC.ItemIndex := 0;

  prAddLine(GroupC.Height, Top);
  GroupC.Top := Top;
  GroupC.Left := XPAD;

  prBottomComponent := GroupC;
end;
procedure TQUB_Python_Prompt.AddButton(caption :String);
var
  b       :TButton;
  Top     :Integer;
begin
  b := TButton.Create(Self);
  b.Caption := caption;
  b.Parent := Self;
  b.OnClick := Self.ButtonClick;

  if prLastButton <> nil then begin
    b.Top := prLastButton.Top;
    b.Left := prLastButton.Left + prLastButton.Width + XPAD;
  end else begin
    prAddLine(b.Height, Top);
    b.Top := Top;
    b.Left := XPAD;
  end;

  prButtons.AddObject(caption, b);
  prLastButton := b;
  prBottomComponent := b;
end;

function TQUB_Python_Prompt.GetValue(prompt :String) :String;
var
  group            :TRadioGroup;
begin
  Result := '';
  try
    if prTextQueries.IndexOf(prompt) >= 0 then begin
      Result := TEdit(prTextQueries.Objects[ prTextQueries.IndexOf(prompt) ]).Text;
    end else if prRadioGroups.IndexOf(prompt) >= 0 then begin
      group := TRadioGroup(prRadioGroups.Objects[ prRadioGroups.IndexOf(prompt) ]);
      Result := group.Items[ group.ItemIndex ];
    end;
  except
  end;
end;

function TQUB_Python_Prompt.GetButton: String;
begin
  Result := prChosenButton;
end;

end.
