{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2003
// python: custom floating window with message and maybe buttons


unit QUB_Python_Wizard;

interface

uses

Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
StdCtrls, ExtCtrls, SyncObjs,

PythonEngine;


type

TQUB_Python_WizardForm = class(TForm)
protected
  txtMessage             :TMemo;
  panButtons             :TPanel;
  prButtons              :TList;

  procedure prOnButtonClick(Sender :TObject);
  procedure prSetMessage(Msg :String);

public
  OnButtonClick          :procedure (Caption :String) of object;

  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure ClearButtons;
  procedure AddButton(Caption :String);

  property Msg           :String        write prSetMessage;
end;


TQUB_Python_Wizard = class(TPyObject)
protected
  prForm                 :TQUB_Python_WizardForm;
  prSignal               :TEvent;
  prButtonClicked        :String;
  prString               :String;
  prButtonNames          :PPyObject;

  procedure prOnButtonClick(Caption :String);

  // gui sync methods
  procedure prDestroySync;
  procedure prInit;
  procedure prShow;
  procedure prHide;
  procedure prSetCaption;
  procedure prSetMessage;
  procedure prSetButtons;

public
  constructor Create( APythonType : TPythonType ); override;
  destructor Destroy; override;

  function  GetAttr(key : PChar) : PPyObject; override;
  function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;
  function Show( args :PPyObject ): PPyObject; cdecl;
  function Hide( args :PPyObject ): PPyObject; cdecl;
  function WaitForButton( args :PPyObject ): PPyObject; cdecl;
end;


var
  QUB_Python_WizardType         :TPythonType;

procedure QUB_Python_Wizard_Init;


implementation


uses

QUB_Main,
QUB_Viewport_Python, QUB_Python_Thread, QUB_Node_Python;


const
QUB_PYTHON_WIZARD_WIDTH    = 300;
QUB_PYTHON_WIZARD_HEIGHT   = 150;


{$R *.DFM}

constructor TQUB_Python_WizardForm.Create(AOwner :TComponent);
begin
  inherited;

  Width                  := QUB_PYTHON_WIZARD_WIDTH;
  Height                 := QUB_PYTHON_WIZARD_HEIGHT;
  BorderStyle            := bsDialog;
  FormStyle              := fsStayOnTop;
  Position               := poMainFormCenter;
  BorderIcons            := [];

  prButtons              := TList.Create;
  panButtons             := TPanel.Create(Self);
  panButtons.Align       := alBottom;
  panButtons.Height      := 22;
  panButtons.Parent      := Self;

  txtMessage             := TMemo.Create(Self);
  txtMessage.Align       := alClient;
  txtMessage.Alignment   := taCenter;
  txtMessage.Parent      := Self;
  txtMessage.ReadOnly    := True;
  txtMessage.Font.Style  := [fsBold];
  txtMessage.Font.Size   := 10;
end;

destructor TQUB_Python_WizardForm.Destroy;
begin
  prButtons.Free;
  inherited;
end;

procedure TQUB_Python_WizardForm.ClearButtons;
var
  i                      :Integer;
begin
  for i := 0 to prButtons.Count - 1 do begin
    TButton( prButtons[i] ).Free;
  end;
  prButtons.Clear;
end;

procedure TQUB_Python_WizardForm.AddButton(Caption :String);
var
  Btn                    :TButton;
  i, w                   :Integer;
begin
  Btn                    := TButton.Create(Self);
  Btn.Parent             := panButtons;
  Btn.Caption            := Caption;
  Btn.OnClick            := prOnButtonClick;
  prButtons.Add(Btn);

  w                      := panButtons.ClientWidth div prButtons.Count;
  for i := prButtons.Count - 1 downto 0 do begin
    TButton(prButtons[i]).SetBounds(i*w, 0, w, panButtons.Height);
  end;
end;

procedure TQUB_Python_WizardForm.prOnButtonClick(Sender :TObject);
begin
  if Assigned(OnButtonClick) then OnButtonClick( TButton(Sender).Caption );
end;

procedure TQUB_Python_WizardForm.prSetMessage(Msg :String);
begin
  txtMessage.Text := Msg;
end;

// *****************************************************************************
class procedure TQUB_Python_Wizard.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do begin
    AddMethod( 'show',          @TQUB_Python_Wizard.Show,          'show()' );
    AddMethod( 'hide',          @TQUB_Python_Wizard.Hide,          'hide()' );
    AddMethod( 'waitForButton', @TQUB_Python_Wizard.WaitForButton, 'waitForButton()' );
  end;
end;

constructor TQUB_Python_Wizard.Create( APythonType : TPythonType );
begin
  inherited;
  PythonThread.SyncGUIProc( prInit );
end;

procedure TQUB_Python_Wizard.prInit;
begin
  prForm    := TQUB_Python_WizardForm.Create(Application);
  prForm.OnButtonClick := prOnButtonClick;
  prSignal  := TEvent.Create(nil{security}, False{auto-reset}, False{init:off}, ''{unique});
end;

destructor TQUB_Python_Wizard.Destroy;
begin
  PythonThread.SyncGUIProc( prDestroySync );
  inherited;
end;

procedure TQUB_Python_Wizard.prDestroySync;
begin
  try
    prForm.Release;
    prSignal.Free;
  except
  end;
end;

function  TQUB_Python_Wizard.GetAttr(key : PChar) : PPyObject;
begin
  Result := inherited GetAttr(key);
end;

function  TQUB_Python_Wizard.SetAttr(key : PChar; value : PPyObject) : Integer;
var
  s           :PChar;
  i           :Integer;
begin
  Result := 0;
  with GetPythonEngine do begin
    if key = 'caption' then begin
      if PyArg_Parse( value, 's:caption', [@s] ) = 0 then begin
        Result := -1;
        Exit;
      end else begin
        prString := s;
        PythonThread.SyncGUIProc( prSetCaption );
      end;
    end else if key = 'message' then begin
      if PyArg_Parse( value, 's:message', [@s] ) = 0 then begin
        Result := -1;
        Exit;
      end else begin
        prString := s;
        PythonThread.SyncGUIProc( prSetMessage );
      end;
    end else if key = 'buttons' then begin
      if not PyList_Check( value ) then begin
        PyErr_SetString( PyExc_TypeError^, 'expected a list of button names' );
        Result := -1;
        Exit;
      end else begin
        for i := 0 to PyList_Size( value ) - 1 do begin
          if not PyString_Check( PyList_GetItem( value, i ) ) then begin
            PyErr_SetString( PyExc_TypeError^, 'expected a list of button names' );
            Result := -1;
            Exit;
          end;
        end;

        prButtonNames := value;
        PythonThread.SyncGUIProc( prSetButtons );
      end;
    end else begin
      Result := inherited SetAttr(key, value);
    end;
  end;
end;

procedure TQUB_Python_Wizard.prSetCaption;
begin
  prForm.Caption := prString;
end;

procedure TQUB_Python_Wizard.prSetMessage;
begin
  prForm.Msg     := prString;
end;

procedure TQUB_Python_Wizard.prSetButtons;
var
  i              :Integer;
begin
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    prForm.ClearButtons;
    for i := 0 to PyList_Size( prButtonNames ) - 1 do begin
      prForm.AddButton( PyString_AsString( PyList_GetItem( prButtonNames, i ) ) );
    end;
  end;
  PythonThread.EndPython;
end;

function TQUB_Python_Wizard.Show( args :PPyObject ): PPyObject;
begin
  Adjust(@Self);

  PythonThread.SyncGUIProc( prShow );
  Result := GetPythonEngine.ReturnNone;
end;

procedure TQUB_Python_Wizard.prShow;
begin
  prForm.Visible := True;
  prForm.SetFocus;
end;

function TQUB_Python_Wizard.Hide( args :PPyObject ): PPyObject;
begin
  Adjust(@Self);

  PythonThread.SyncGUIProc( prHide );
  Result := GetPythonEngine.ReturnNone;
end;

procedure TQUB_Python_Wizard.prHide;
begin
  prForm.Visible := False;
end;

function TQUB_Python_Wizard.WaitForButton( args :PPyObject ): PPyObject;
begin
  Adjust(@Self);

  PythonThread.EndPython;

  SyncGUIProc( prShow );

  prSignal.ResetEvent;
  while wrTimeout = prSignal.WaitFor( 5000 ) do SyncGUIProc( prShow ); // goddamn floating windows style hides my wizard at the first opportunity; it must be held in check

  PythonThread.BeginPython;

  Result := GetPythonEngine.PyString_FromString( PChar(prButtonClicked) );
end;

procedure TQUB_Python_Wizard.prOnButtonClick(Caption :String);
begin
  prButtonClicked := Caption;
  prSignal.SetEvent;
end;

function QUB_Python_Wizard_Create(args, kw :PPyObject): PPyObject;
begin
  Result := QUB_Python_WizardType.CreateInstance;
end;

procedure QUB_Python_Wizard_Init;
begin
  QUB_Python_WizardType := TPythonType.Create(nil);
  QUB_Python_WizardType.Engine := GetPythonEngine;
  QUB_Python_WizardType.TypeName := 'Wizard';
  QUB_Python_WizardType.Services.Basic := [bsGetAttr, bsSetAttr];
  QUB_Python_WizardType.PyObjectClass := TQUB_Python_Wizard;
  QUB_Python_WizardType.Module := PythonViewport.pmQUBTypes;
  QUB_Python_WizardType.Initialize;
  PyNode_PublishPyFunc( QUB_Python_Wizard_Create, 'CreateWizard' );
end;

// *****************************************************************************
end.
