{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// mainly components for the model


unit QUB_QFS_Channels;


interface


uses


Windows, SysUtils, Classes, Messages, Graphics, Controls, StdCtrls, ExtCtrls, Math,

AllocTypes, Alloc,

RotRect, RotLabel,

QUB_QFS, QUB_QFS_Edit, QUB_QVO_QVD_Global,
QUB_ExpCond_Parameters,
QUB_Node,
QUB_Globals;
//KPL_Expression;


const


//Names for Stream I/O
CHANNELS_MODEL_FILE_NAME                       = 'ModelFile';
CHANNELS_DATA_FILE_NAME                        = 'DataFile';

CHANNELS_CHANNEL_COUNT_NAME                    = 'ChannelCount';

CHANNELS_STATES_NAME                           = 'States';
CHANNELS_RATES_NAME                            = 'Rates';
CHANNELS_CONSTRAINTS_NAME                      = 'Constraints';
CHANNELS_CONSTRAINTS_AMPVAR_NAME               = 'ConstraintsAmpVar';
CHANNELS_EXTRAKINETICPARS_NAME                 = 'ExtraKineticPars';

CHANNELS_STATE_NAME                            = 'State';
CHANNELS_RATE_NAME                             = 'Rate';
CHANNELS_CONSTRAINT_NAME                       = 'Constraint';
CHANNELS_CONSTRAINT_AMPVAR_NAME                = 'ConstraintAmpVar';

CHANNELS_STATE_X_NAME                          = 'x';
CHANNELS_STATE_Y_NAME                          = 'y';
CHANNELS_STATE_CLASS_NAME                      = 'Class';
CHANNELS_STATE_PR_NAME                         = 'Pr';
CHANNELS_STATE_GR_NAME                         = 'Gr';

CHANNELS_RATE_STATES_NAME                      = 'States';
CHANNELS_RATE_K0_NAME                          = 'k0';
CHANNELS_RATE_K1_NAME                          = 'k1';
CHANNELS_RATE_dK0_NAME                         = 'dk0';
CHANNELS_RATE_dK1_NAME                         = 'dk1';
CHANNELS_RATE_P_NAME                           = 'P';
CHANNELS_RATE_Q_NAME                           = 'Q';
CHANNELS_RATE_PNAMES_NAME                      = 'PNames';
CHANNELS_RATE_PNAME_NAME                       = 'PName';
CHANNELS_RATE_DEF_PNAME                        = 'Ligand';
CHANNELS_RATE_QNAMES_NAME                      = 'QNames';
CHANNELS_RATE_QNAME_NAME                       = 'QName';
CHANNELS_RATE_DEF_QNAME                        = 'Voltage';
CHANNELS_RATE_PVALUE_NAME                      = 'PValue';
CHANNELS_RATE_QVALUE_NAME                      = 'QValue';
CHANNELS_RATE_FORMATS_NAME                     = 'RateFormats';
CHANNELS_RATE_FORMAT_NAME                      = 'RateFormat';
CHANNELS_RATE_DEF_FORMAT                       = '';

CHANNELS_CONSTRAINT_HASVALUE_NAME              = 'HasValue';
CHANNELS_CONSTRAINT_VALUE_NAME                 = 'Value';
CHANNELS_CONSTRAINT_COEFFS_NAME                = 'Coeffs';

CHANNELS_CONSTRAINT_FIX_RATE_NAME              = 'FixRate';
CHANNELS_CONSTRAINT_SCALE_RATE_NAME            = 'ScaleRate';
CHANNELS_CONSTRAINT_FIX_EXP_NAME               = 'FixExp';
CHANNELS_CONSTRAINT_SCALE_EXP_NAME             = 'ScaleExp';
CHANNELS_CONSTRAINT_LOOP_BAL_NAME              = 'LoopBal';
CHANNELS_CONSTRAINT_LOOP_IMBAL_NAME            = 'LoopImbal';
CHANNELS_CONSTRAINT_LOOP_BAL_EXP_NAME          = 'LoopBalExp';
CHANNELS_CONSTRAINT_LOOP_IMBAL_EXP_NAME        = 'LoopImbalExp';
CHANNELS_CONSTRAINT_GEN_NAME                   = 'Generalized';

CHANNELS_CONSTRAINT_FIX_AMP_NAME               = 'FixAmp';
CHANNELS_CONSTRAINT_FIX_EXC_AMP_NAME           = 'FixExcAmp';
CHANNELS_CONSTRAINT_FIX_DIFF_EXC_AMP_NAME      = 'FixDiffExcAmp';
CHANNELS_CONSTRAINT_SCALE_EXC_AMP_NAME         = 'ScaleExcAmp';
CHANNELS_CONSTRAINT_SCALE_DIFF_EXC_AMP_NAME    = 'ScaleDiffExcAmp';

CHANNELS_CONSTRAINT_FIX_VAR_NAME               = 'FixVar';
CHANNELS_CONSTRAINT_FIX_EXC_VAR_NAME           = 'FixExcVar';
CHANNELS_CONSTRAINT_FIX_DIFF_EXC_VAR_NAME      = 'FixDiffExcVar';
CHANNELS_CONSTRAINT_SCALE_EXC_VAR_NAME         = 'ScaleExcVar';
CHANNELS_CONSTRAINT_SCALE_DIFF_EXC_VAR_NAME    = 'ScaleDiffExcVar';

CHANNELS_CONSTRAINT_UNKNOWN_NAME               = 'Unknown';

CHANNELS_EXTRAKINETICPAR_NAME                  = 'ExtraKineticPar';
CHANNELS_EXTRAKINETICPAR_TYPE_NAME             = 'ExtraKineticParType';

CHANNELS_AMPS_NAME                             = 'Amps';
CHANNELS_STDS_NAME                             = 'Stds';
CHANNELS_CONDS_NAME                            = 'Conds';
CHANNELS_CONDSTDS_NAME                         = 'CondStds';
CHANNELS_VREVS_NAME                            = 'VRevs';
CHANNELS_NAR_NAME                              = 'NAr';
CHANNELS_ARS_NAME                              = 'Ars';
CHANNELS_AR_NAME                               = 'Ar';
CHANNELS_VRev_NAME                             = 'VRev';


PASC_FMT                         :String = '#######0.0###';

MIN_K0                           :Double = 0;
MAX_K0                           :Double = 1e30;

MIN_K1                           :Double = -1e30;
MAX_K1                           :Double = 1e30;


type


TQUB_QFS_QMF_ExtraKineticParType               = (EKP_TYPE_LIN, EKP_TYPE_LOG, EKP_TYPE_EXP);

TQUB_QFS_QMF_ConstraintType                    = (CSTR_FIX_RATE, CSTR_SCALE_RATE, CSTR_FIX_EXP, CSTR_SCALE_EXP,
                                                  CSTR_LOOP_BALANCE, CSTR_LOOP_IMBALANCE, CSTR_LOOP_BALANCE_EXP, CSTR_LOOP_IMBALANCE_EXP, CSTR_GEN, CSTR_UNKNOWN);
TQUB_QFS_QMF_ConstraintAmpVarType              = (CSTR_FIX_AMP,  CSTR_FIX_EXC_AMP, CSTR_FIX_DIFF_EXC_AMP, CSTR_SCALE_EXC_AMP, CSTR_SCALE_DIFF_EXC_AMP,
                                                  CSTR_FIX_VAR,  CSTR_FIX_EXC_VAR, CSTR_FIX_DIFF_EXC_VAR, CSTR_SCALE_EXC_VAR, CSTR_SCALE_DIFF_EXC_VAR, CSTR_UNKNOWN_AMPVAR);

const


CHANNELS_CONSTRAINT_TYPE_NAME                  :Array[TQUB_QFS_QMF_ConstraintType] of String = (

CHANNELS_CONSTRAINT_FIX_RATE_NAME,
CHANNELS_CONSTRAINT_SCALE_RATE_NAME,
CHANNELS_CONSTRAINT_FIX_EXP_NAME,
CHANNELS_CONSTRAINT_SCALE_EXP_NAME,
CHANNELS_CONSTRAINT_LOOP_BAL_NAME,
CHANNELS_CONSTRAINT_LOOP_IMBAL_NAME,
CHANNELS_CONSTRAINT_LOOP_BAL_EXP_NAME,
CHANNELS_CONSTRAINT_LOOP_IMBAL_EXP_NAME,
CHANNELS_CONSTRAINT_GEN_NAME,
CHANNELS_CONSTRAINT_UNKNOWN_NAME
);

CHANNELS_CONSTRAINT_AMPVAR_TYPE_NAME           :Array[TQUB_QFS_QMF_ConstraintAmpVarType] of String = (

CHANNELS_CONSTRAINT_FIX_AMP_NAME,
CHANNELS_CONSTRAINT_FIX_EXC_AMP_NAME,
CHANNELS_CONSTRAINT_FIX_DIFF_EXC_AMP_NAME,
CHANNELS_CONSTRAINT_SCALE_EXC_AMP_NAME,
CHANNELS_CONSTRAINT_SCALE_DIFF_EXC_AMP_NAME,
CHANNELS_CONSTRAINT_FIX_VAR_NAME,
CHANNELS_CONSTRAINT_FIX_EXC_VAR_NAME,
CHANNELS_CONSTRAINT_FIX_DIFF_EXC_VAR_NAME,
CHANNELS_CONSTRAINT_SCALE_EXC_VAR_NAME,
CHANNELS_CONSTRAINT_SCALE_DIFF_EXC_VAR_NAME,
CHANNELS_CONSTRAINT_UNKNOWN_NAME
);


QUB_QFS_QMF_MIN_CHANNEL_COUNT                  = 1;
QUB_QFS_QMF_MAX_CHANNEL_COUNT                  = High(Integer);
QUB_QFS_QMF_DEF_CHANNEL_COUNT                  = 1;


QUB_QFS_QMF_DEF_NCLASS                                                          = 10;
QUB_QFS_QMF_MAX_STATE                          :Integer                         = 500;
QUB_QFS_QMF_MAX_RATE                           :Integer                         = 5000;

QUB_QFS_QMF_MAX_COEFF                                                           = 10;

QUB_QFS_QMF_DEF_VREV                           :Double                          =   0;
QUB_QFS_QMF_MIN_VREV                           :Double                          = -10000;
QUB_QFS_QMF_MAX_VREV                           :Double                          =  10000;
                      
QUB_QFS_QMF_DEF_AMP                            :Array[0..QUB_QFS_QMF_DEF_NCLASS - 1] of Double = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9);
QUB_QFS_QMF_DEF_STD                            :Array[0..QUB_QFS_QMF_DEF_NCLASS - 1] of Double = (0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55);
QUB_QFS_QMF_DEF_COND                           :Array[0..QUB_QFS_QMF_DEF_NCLASS - 1] of Double = (0, 1, 2, 3, 4, 5, 6, 7, 8, 9);
QUB_QFS_QMF_DEF_CONDSTD                        :Array[0..QUB_QFS_QMF_DEF_NCLASS - 1] of Double = (0.0, 0.1, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55);
QUB_QFS_QMF_DEF_CLSVREV                        :Array[0..QUB_QFS_QMF_DEF_NCLASS - 1] of Double = (UNSET_VALUE, UNSET_VALUE, UNSET_VALUE, UNSET_VALUE, UNSET_VALUE, UNSET_VALUE, UNSET_VALUE, UNSET_VALUE, UNSET_VALUE, UNSET_VALUE);
QUB_QFS_QMF_DEF_NAR                            :Integer                         = 0;
QUB_QFS_QMF_DEF_CFF                            :Double                          = 0;


QUB_QFS_QMF_DEF_K0                             :Double                          = 1000;
QUB_QFS_QMF_DEF_K1                             :Double                          = 0;
QUB_QFS_QMF_DEF_P                              :Boolean                         = False;
QUB_QFS_QMF_DEF_Q                              :Boolean                         = False;
QUB_QFS_QMF_DEF_PName                          :String                          = 'Ligand';
QUB_QFS_QMF_DEF_QName                          :String                          = 'Voltage';
QUB_QFS_QMF_DEF_RATEFORMAT                     :String                          = '';


QUB_QFS_QMF_CLASS_COLORS                       :Array[0..QUB_QFS_QMF_DEF_NCLASS    ] of TColor = (clBlack,   clRed,  clNavy, clGreen, clYellow, clFuchsia,  clLime,  clBlue,  clAqua, clPurple, clWhite);
QUB_QFS_QMF_BACK_COLORS                        :Array[0..QUB_QFS_QMF_DEF_NCLASS - 1] of TColor = (clWhite, clBlack, clWhite, clWhite,  clBlack,   clBlack, clBlack, clWhite, clBlack,  clWhite);

QUB_QFS_QMF_RATE_WIDTH                         :Integer                         = 76;
QUB_QFS_QMF_RATE_HEIGHT                        :Integer                         = 16;
QUB_QFS_QMF_RATE_WIDTH2                        :Integer                         = 38;
QUB_QFS_QMF_RATE_HEIGHT2                       :Integer                         = 8;
QUB_QFS_QMF_RATE_FONTHEIGHT                    :Integer                         = 15;

QUB_QFS_QMF_MAX_PARAMETER                      = 15;

QUB_QFS_QMF_STATE_FMT                          :String                          = '#######0.0###';

QUB_QFS_QMF_RATE_K0                            :Integer                         = 0;
QUB_QFS_QMF_RATE_K1                            :Integer                         = 1;


function QUB_GetClassColor(i :Integer): TColor;
function QUB_GetClassBackColor(i :Integer): TColor;

procedure DrawState(C :TCanvas; rctBox :TRect; iClass :Integer; cCaption :String);


type


//Forward class declarations ***************************************************


TQUB_QFS_States                                = class;
TQUB_QFS_State                                 = class;

TQUB_QFS_Rates                                 = class;
TQUB_QFS_Rate                                  = class;

TQUB_QFS_Constraints                           = class;
TQUB_QFS_Constraint                            = class;

TQUB_QFS_ConstraintsAmpVar                     = class;
TQUB_QFS_ConstraintAmpVar                      = class;

TQUB_QFS_ExtraKineticPars                      = class;
TQUB_QFS_ExtraKineticPar                       = class;

TQUB_QFS_StatePanels                           = class;
TQUB_QFS_RatePanels                            = class;

TQUB_QFS_Line                                  = class;


TQUB_QFS_StatePanel                            = class(TPanel)
protected


  m_iClass                                     :Integer;			//Class set by user. ( Formerly stored in prOnesPanel.Caption )
  m_cCaption                                   :String;				//UpdateStates calls SetCaptionI or SetCaptionC to change this.

  prState                                      :TQUB_QFS_State;


  procedure prSetClass(x :Integer);


public


  constructor Create(AOwner :TComponent); override;


  procedure Paint; override;
  procedure SetCaptionI(i : Integer );
  procedure SetCaptionC(c : String );

  procedure Activate;
  procedure DeActivate;


  property State         :TQUB_QFS_State     read prState       write prState;
  property Cls           :Integer                               write prSetClass;
  property StateCaption  :String             read m_cCaption;


end;


TQUB_QFS_RatePanel                             = class(TRotLabel)
public

  Rate              :TQUB_QFS_Rate;
  Dir               :Integer;
  K01               :Integer;
  Hint              :String;  // displayed by QVM.MouseMove

  constructor Create; override;


  procedure Activate;
  procedure DeActivate;

  ///procedure PaintTo(NewCanvas :TCanvas);

end;


//States ***********************************************************************
TQUB_QFS_States                                = class(TQUB_QFS_Record)
protected


  prRates                                      :TQUB_QFS_Rates;
  prStateCount                                 :Integer;
  prCurrState                                  :TQUB_QFS_State;
  prPanels                                     :TQUB_QFS_StatePanels;

  prOnChangeActiveState                        :TNotifyEvent;


  function prGetState(s :Integer) :TQUB_QFS_State;
  procedure prSetCurrState(State :TQUB_QFS_State);


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  destructor Destroy; override;


  procedure Reset; override;

  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;
  function NewState(x, y :Double) :TQUB_QFS_State;
  procedure DelState(State :TQUB_QFS_State);
  function WhichState(Panel :TQUB_QFS_StatePanel) :TQUB_QFS_State;
  function IndexOf(State :TQUB_QFS_State) :Integer;
  procedure UpdateStates(lClassLabels :Boolean = False);


  property Rates                   :TQUB_QFS_Rates   read prRates               write prRates;

  property State[Index :Integer]   :TQUB_QFS_State   read prGetState;                                     default;
  property StateCount              :Integer          read prRecordCount;

  property CurrState               :TQUB_QFS_State   read prCurrState           write prSetCurrState;

  property OnChangeActiveState     :TNotifyEvent     read prOnChangeActiveState write prOnChangeActiveState;


end;


//State ************************************************************************
TQUB_QFS_State                                 = class(TQUB_QFS_Record)
protected


  prPanel                                      :TQUB_QFS_StatePanel;

  prStateSize                                  :Integer;

  prAmp                                        :Double;
  prStd                                        :Double;
  prCond                                       :Double;
  prCondStd                                    :Double;
  prVrev                                       :Double;
  
  prCff                                        :Array[0..QUB_QFS_QMF_MAX_COEFF - 1] of Double;

  prOnEdit                                     :TNotifyEvent;


  //Records for Stream-stored properties ***************************************
  prRecX                                       :TQUB_QFS_Record;
  prRecY                                       :TQUB_QFS_Record;
  prRecClass                                   :TQUB_QFS_Record;
  prRecPr                                      :TQUB_QFS_Record;
  prRecGr                                      :TQUB_QFS_Record;
  //****************************************************************************


  procedure prSetStateSize(s :Integer);

  procedure prSetX(x :Double);
  function prGetX :Double;
  procedure prSetY(y :Double);
  function prGetY :Double;
  procedure prSetPr(p :Double);
  function prGetPr :Double;
  procedure prSetGr(g :Integer);
  function prGetGr :Integer;
  procedure prSetClass(c :Integer);
  function prGetClass :Integer;

  procedure prSetAmp(a :Double);
  procedure prSetStd(x :Double);
  procedure prSetCond(c :Double);
  procedure prSetCondStd(x :Double);
  procedure prSetVrev(x :Double);
  function  prGetCff(Index :Integer) :Double;
  procedure prSetCff(Index :Integer; c :Double);


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  constructor CreateParam(const Name :String; Parent :TQUB_QFS_Record; x, y :Double; Panel :TQUB_QFS_StatePanel);


  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  procedure Activate;
  procedure DeActivate;

  procedure ChangeClass;

  procedure SetHint;

  property Panel                               :TQUB_QFS_StatePanel    read prPanel;
  property StateSize                           :Integer                                             write prSetStateSize;


  //Streamed *******************************************************************
  property X                                   :Double                 read prGetX                  write prSetX;
  property Y                                   :Double                 read prGetY                  write prSetY;
  property Pr                                  :Double                 read prGetPr                 write prSetPr;
  property Gr                                  :Integer                read prGetGr                 write prSetGr;
  property Cls                                 :Integer                read prGetClass              write prSetClass;
  //****************************************************************************

  property Amp                                 :Double                 read prAmp                   write prSetAmp;
  property Std                                 :Double                 read prStd                   write prSetStd;
  property Cond                                :Double                 read prCond                  write prSetCond;
  property CondStd                             :Double                 read prCondStd               write prSetCondStd;
  property Vrev                                :Double                 read prVrev                  write prSetVrev;
  property Cff[Index :Integer]                 :Double                 read prGetCff                write prSetCff;

  property OnEdit                              :TNotifyEvent           read prOnEdit                write prOnEdit;


end;



//Rates ************************************************************************
TQUB_QFS_Rates                                 = class(TQUB_QFS_Record)
protected


  prStates                                     :TQUB_QFS_States;

  prRateCount                                  :Integer;

  prPanels                                     :TQUB_QFS_RatePanels;

  prCurrRateIdx                                :Integer;
  prCurrHalfIdx                                :Integer;
  prCurrK01                                    :Integer;

  prDiagonalRates                              :Boolean;
  prShowK1                                     :Boolean;
  prPreFormat, prExpFormat                     :INode;


  function prGetRate(s :Integer) :TQUB_QFS_Rate;
  function prGetRateIndex(r :TQUB_QFS_Rate) :Integer;
  function prGetRateOfStates(State1, State2 :TQUB_QFS_State) :TQUB_QFS_Rate;
  function prGetRateValueOfStates(State1, State2 :TQUB_QFS_State) :Double;

  procedure prSetPreFormat(x :INode);
  procedure prSetExpFormat(x :INode);
  procedure prSetDiagonalRates(x :Boolean);
  procedure prSetShowK1(x :Boolean);

  procedure prUpdateRates;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  destructor Destroy; override;


  procedure Reset; override;

  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  function NewRate(State1, State2 :TQUB_QFS_State) :TQUB_QFS_Rate;
  procedure DelRate(Rate :TQUB_QFS_Rate);
  function WhichPanel(Pos :TPoint): TQUB_QFS_RatePanel;

  procedure SetRateK0OfStates(State1, State2 :TQUB_QFS_State; k0 :Double);

  procedure GetRateIdxHalfIdxOfStates(StateIdx1, StateIdx2 :Integer; var RateIdx, HalfIdx :Integer);
  procedure GetStatesOfRate(RateIdx, HalfIdx :Integer; var StateIdx1, StateIdx2 :Integer);

  function CopyFrom(Source :TQUB_QFS_Rates) :Boolean;
  function UpdateFrom(Source :TQUB_QFS_Rates; Change :Boolean) :Boolean;

  procedure UpdateCaptions;

  procedure DoMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
  function  DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; var Handled :Boolean): TQUB_QFS_RatePanel;
  procedure DoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
  procedure DoMouseWheel(Sender :TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled :Boolean);
  procedure DoDblClick(Sender: TObject; X, Y: Integer; var Handled :Boolean);

  property States                                            :TQUB_QFS_States   read prStates               write prStates;

  property CurrRateIdx                                       :Integer           read prCurrRateIdx          write prCurrRateIdx;
  property CurrHalfIdx                                       :Integer           read prCurrHalfIdx          write prCurrHalfIdx;
  property CurrK01                                           :Integer           read prCurrK01              write prCurrK01;

  property Rate[Index :Integer]                              :TQUB_QFS_Rate     read prGetRate; default;
  property RateIndex[Rate :TQUB_QFS_Rate]                    :Integer           read prGetRateIndex;
  property RateOfStates[State1, State2 :TQUB_QFS_State]      :TQUB_QFS_Rate     read prGetRateOfStates;
  property RateValueOfStates[State1, State2 :TQUB_QFS_State] :Double            read prGetRateValueOfStates;
  property RateCount                                         :Integer           read prRecordCount;

  property DiagonalRates                                     :Boolean           read prDiagonalRates        write prSetDiagonalRates;
  property ShowK1                                            :Boolean           read prShowK1               write prSetShowK1;
  property PreFormat                                         :INode             read prPreFormat            write prSetPreFormat;
  property ExpFormat                                         :INode             read prExpFormat            write prSetExpFormat;

end;


//Rate *************************************************************************
TQUB_QFS_Rate                                  = class(TQUB_QFS_Record)
protected


  prStates                                     :Array[1..2] of TQUB_QFS_State;
  prPanels                                     :Array[1..4] of TQUB_QFS_RatePanel;


  //Records for Stream properties **********************************************
  prRecStates                                  :TQUB_QFS_Record;
  prRecK0                                      :TQUB_QFS_Record;
  prRecK1                                      :TQUB_QFS_Record;
  prRecdK0                                     :TQUB_QFS_Record;
  prRecdK1                                     :TQUB_QFS_Record;
  prRecP                                       :TQUB_QFS_Record;
  prRecQ                                       :TQUB_QFS_Record;
  prRecPNames                                  :TQUB_QFS_Record;
  prRecQNames                                  :TQUB_QFS_Record;
  prRecPValue                                  :TQUB_QFS_Record;
  prRecQValue                                  :TQUB_QFS_Record;
  prRecRateFormats                             :TQUB_QFS_Record;
  //****************************************************************************

  //prRateExp                                    :Array[1..2] of TKPL_Expression;//Alternative rate expression,
  //                                                                             //used if non-nil

  prArrowPos                                   :Array[1..2, 1..2] of TPoint;
  prLinesPos                                   :Array[1..2] of TPoint;
  prAngle                                      :Double;

  prDiagonalRates                              :Boolean;
  prShowK1                                     :Boolean;
  prPreFormat, prExpFormat                     :INode;

  prOnEdit                                     :TNotifyEvent;


  function prGetState(Index :Integer) :TQUB_QFS_State;
  procedure prSetState(Index :Integer; State :TQUB_QFS_State);

  function prGetStateIndex(Index :Integer) :Integer;

  function prGetPanel(Index :Integer) :TQUB_QFS_RatePanel;

  function prGetValueFormat(Index :Integer) :String;
  function prGetValue(Index :Integer) :Double;

  function prGetPreExpFormat(Index :Integer; v :Double) :String;
  function prGetExpFormat(v :Double) :String;

  procedure prSetHint;

  procedure prSetShowK1(x :Boolean);
  procedure prUpdateCaptions(Index :Integer);

  function prGetK0Format(Index :Integer) :String;
  function prGetK0(Index :Integer) :Double;
  procedure prSetK0(Index :Integer; k :Double);
  function prGetdK0(Index :Integer) :Double;
  procedure prSetdK0(Index :Integer; dk :Double);

  function prGetK1Format(Index :Integer) :String;
  function prGetK1(Index :Integer) :Double;
  procedure prSetK1(Index :Integer; k :Double);
  function prGetdK1(Index :Integer) :Double;
  procedure prSetdK1(Index :Integer; dk :Double);

  function prGetP(Index :Integer) :Boolean;
  procedure prSetP(Index :Integer; p :Boolean);
  function prGetQ(Index :Integer) :Boolean;
  procedure prSetQ(Index :Integer; q :Boolean);

  function prGetPName(Index :Integer) :String;
  procedure prSetPName(Index :Integer; p :String);
  function prGetQName(Index :Integer) :String;
  procedure prSetQName(Index :Integer; q :String);

  procedure prSetPValue(Index :Integer; p :TPDouble);
  function prGetPValue(Index :Integer) :TPDouble;
  procedure prSetQValue(Index :Integer; q :TPDouble);
  function prGetQValue(Index :Integer) :TPDouble;

  function prGetRateFormat(Index :Integer): String;
  procedure prSetRateFormat(Index :Integer; x :String);

  function prGetK0Adr(Index :Integer) :TPDouble;
  function prGetK1Adr(Index :Integer) :TPDouble;

  //function prGetRateExp(Index :Integer) :TKPL_Expression;
  //procedure prSetRateExp(Index :Integer; Exp :TKPL_Expression);

  function prGetLinesPos(Index :Integer) :TPoint;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  constructor CreateParam(const Name :String; Parent :TQUB_QFS_Record; State1, State2 :TQUB_QFS_State; Panel1, Panel2, Panel3, Panel4 :TQUB_QFS_RatePanel);


  procedure SetRate(Index :Integer; oRate :TRate);

  function Write(WriteStream :TStream) :Boolean; override;

  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  procedure Activate;
  procedure Deactivate;

  procedure Update;
  procedure Paint(Canvas :TCanvas; dx, dy :Integer);
  procedure PaintActive(Canvas :TCanvas; dx, dy :Integer; ActiveIdx :Integer);

  function Unconstrained(Index :Integer; Constraint :Integer) :Boolean;

  procedure ClearPQ; // no graphical updates

  function UpdateFrom(Source :TQUB_QFS_Rate; Change :Boolean) :Boolean;

  procedure UpdateCaptions;

  //procedure OnDeleteRateExp(Sender :TObject);


  property States [Index :Integer]              :TQUB_QFS_State     read prGetState        write prSetState;
  property StateIndex[Index :Integer]           :Integer            read prGetStateIndex;

  property Panels [Index :Integer]              :TQUB_QFS_RatePanel read prGetPanel;

  property Value  [Index :Integer]              :Double             read prGetValue;
  property ValueFormat[Index :Integer]          :String             read prGetValueFormat;


  //Streamed *******************************************************************
  property K0     [Index :Integer]              :Double             read prGetK0           write prSetK0;
  property K1     [Index :Integer]              :Double             read prGetK1           write prSetK1;
  property dK0    [Index :Integer]              :Double             read prGetdK0          write prSetdK0;
  property dK1    [Index :Integer]              :Double             read prGetdK1          write prSetdK1;
  property P      [Index :Integer]              :Boolean            read prGetP            write prSetP;
  property Q      [Index :Integer]              :Boolean            read prGetQ            write prSetQ;
  property PName  [Index :Integer]              :String             read prGetPName        write prSetPName;
  property QName  [Index :Integer]              :String             read prGetQName        write prSetQName;
  property PValue [Index :Integer]              :TPDouble           read prGetPValue       write prSetPValue;
  property QValue [Index :Integer]              :TPDouble           read prGetQValue       write prSetQValue;
  property RateFormat[Index :Integer]           :String             read prGetRateFormat   write prSetRateFormat;
  //****************************************************************************

  property K0Adr  [Index :Integer]              :TPDouble           read prGetK0Adr;
  property K1Adr  [Index :Integer]              :TPDouble           read prGetK1Adr;

  //property RateExp[Index :Integer]              :TKPL_Expression    read prGetRateExp      write prSetRateExp;

  property DiagonalRates                        :Boolean            read prDiagonalRates   write prDiagonalRates;
  property ShowK1                               :Boolean            read prShowK1          write prSetShowK1;
  property PreFormat                                         :INode             read prPreFormat            write prPreFormat;
  property ExpFormat                                         :INode             read prExpFormat            write prExpFormat;

  property LinesPos[Index :Integer]             :TPoint             read prGetLinesPos;
  property Angle                                :Double             read prAngle;

  property OnEdit                               :TNotifyEvent       read prOnEdit          write prOnEdit;


end;



//Constraints ******************************************************************
TQUB_QFS_Constraints                           = class(TQUB_QFS_Record)
protected


  prConstraintCount                            :Integer;

  prOnEdit                                     :TNotifyEvent;


  function prGetConstraint(s :Integer) :TQUB_QFS_Constraint;


public


  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  procedure AddConstraint(Constraint :TQUB_QFS_Constraint; ConstraintIndex :Integer);
  procedure DelConstraint(ConstraintIndex :Integer);
  procedure DelState(State, StateCount, RateCount, ExtraKinParCount :Integer);
  procedure AddState;
  procedure DelRate(Rate :TQUB_QFS_Rate; RateIndex, StateCount, RateCount, ExtraKinParCount :Integer);
  procedure AddRate(StateCount, RateCount, ExtraKinParCount :Integer);
  procedure DelExtraKineticPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount :Integer);
  procedure AddExtraKineticPar(StateCount, RateCount, ExtraKinParCount :Integer);

  procedure Copy(c :TQUB_QFS_Constraints);


  property Constraint[Index :Integer]          :TQUB_QFS_Constraint    read prGetConstraint; default;
  property ConstraintCount                     :Integer                read prRecordCount;

  property OnEdit                              :TNotifyEvent           read prOnEdit                write prOnEdit;


end;


//Constraint
//The name of the constraint indicates its type.
TQUB_QFS_Constraint                            = class(TQUB_QFS_Record)
protected


  //Records for Stream-stored properties ***************************************
  prRecHasValue                                :TQUB_QFS_Record;//Boolean
  prRecValue                                   :TQUB_QFS_Record;//Double
  prRecCoeffs                                  :TQUB_QFS_Record;//Array of Double


  function prGetCType :TQUB_QFS_QMF_ConstraintType;
  function prGetIndex(Index :Integer) :Integer;
  procedure prSetIndex(Index, Value :Integer);
  function prGetCoeffCount :Integer;
  function prGetCoeffPointer: TPADouble;
  function prGetCoeff(Index :Integer) :Double;
  procedure prSetCoeff(Index :Integer; Value :Double);
  procedure prSetHasValue(v :Boolean);
  function prGetHasValue :Boolean;
  procedure prSetValue(v :Double);
  function prGetValue :Double;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;


  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  function CheckHasState(State, StateCount, RateCount, ExtraKinParCount :Integer) :Boolean;
  function CheckHasRate(Rate, StateCount, RateCount, ExtraKinParCount :Integer) :Boolean;
  function CheckHasExtraKinPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount :Integer) :Boolean;
  procedure AdjustOnDelRate(Rate, StateCount, RateCount, ExtraKinParCount :Integer);
  procedure AdjustOnAddRate(StateCount, RateCount, ExtraKinParCount :Integer);
  procedure AdjustOnDelExtraKinPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount :Integer);
  procedure AdjustOnAddExtraKinPar(StateCount, RateCount, ExtraKinParCount :Integer);


  property IndexCount            :LongInt                      read prNodeMem.NodeStream.DataCount;
  property Index[Index :Integer] :Integer                      read prGetIndex                       write prSetIndex;
  property CoeffCount            :LongInt                      read prGetCoeffCount;
  property CoeffPointer          :TPADouble                    read prGetCoeffPointer;
  property Coeff[Index :Integer] :Double                       read prGetCoeff                       write prSetCoeff;

  property CType                 :TQUB_QFS_QMF_ConstraintType  read prGetCType;
  property HasValue              :Boolean                      read prGetHasValue                    write prSetHasValue;
  property Value                 :Double                       read prGetValue                       write prSetValue;


end;


//Extra Kinetic Parameters *****************************************************
TQUB_QFS_ExtraKineticPars        = class(TQUB_QFS_Record)
protected


  prExtraKineticParCount         :Integer;

  prOnEdit                       :TNotifyEvent;


  function prGetExtraKineticPar(p :Integer) :TQUB_QFS_ExtraKineticPar;
  function prGetExtraKineticParByName(n :String) :TQUB_QFS_ExtraKineticPar;


public


  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  procedure AddExtraKineticPar(Par :TQUB_QFS_ExtraKineticPar; ParIndex :Integer);
  procedure DelExtraKineticPar(p :Integer);
  //function IndexOf(Par :TQUB_QFS_ExtraKineticPar) :Integer;

  procedure Copy(p :TQUB_QFS_ExtraKineticPars);


  property ExtraKineticPar[Index :Integer]   :TQUB_QFS_ExtraKineticPar read prGetExtraKineticPar; default;
  property ExtraKineticParByName[Index :String] :TQUB_QFS_ExtraKineticPar read prGetExtraKineticParByName;
  property ExtraKineticParCount              :Integer                  read prRecordCount;

  property OnEdit                            :TNotifyEvent             read prOnEdit              write prOnEdit;


end;


//ExtraKineticPar **************************************************************
TQUB_QFS_ExtraKineticPar                              = class(TQUB_QFS_Record)
protected


  //Records for Stream-stored properties ***************************************
  prRecEKPType                                  :TQUB_QFS_Record;


  procedure prSetValue(Value :Double);
  function prGetValue :Double;
  procedure prSetEKPType(t :TQUB_QFS_QMF_ExtraKineticParType);
  function prGetEKPType :TQUB_QFS_QMF_ExtraKineticParType;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record);


  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;


  property Value :Double read prGetValue write prSetValue;
  property EKPType :TQUB_QFS_QMF_ExtraKineticParType read prGetEKPType write prSetEKPType;


end;


//Amplitude Constraints ********************************************************
TQUB_QFS_ConstraintsAmpVar                     = class(TQUB_QFS_Record)
protected


  prConstraintCount                            :Integer;

  prOnEdit                                     :TNotifyEvent;


  function prGetConstraint(s :Integer) :TQUB_QFS_ConstraintAmpVar;


public


  procedure AddConstraint(Constraint :TQUB_QFS_ConstraintAmpVar; ConstraintIndex :Integer);
  procedure DelConstraint(ConstraintIndex :Integer);
  procedure DelClass(Cls :Integer);

  procedure Copy(c :TQUB_QFS_ConstraintsAmpVar);


  property Constraint[Index :Integer]          :TQUB_QFS_ConstraintAmpVar read prGetConstraint; default;
  property ConstraintCount                     :Integer                   read prRecordCount;

  property OnEdit                              :TNotifyEvent              read prOnEdit                write prOnEdit;


end;

//Amplitude Constraint
//The name of the constraint indicates its type.
TQUB_QFS_ConstraintAmpVar                      = class(TQUB_QFS_Record)
protected


  function prGetCType :TQUB_QFS_QMF_ConstraintAmpVarType;
  function prGetIndex(Index :Integer) :Integer;
  procedure prSetIndex(Index, Value :Integer);


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;


  property IndexCount            :LongInt                           read prNodeMem.NodeStream.DataCount;
  property Index[Index :Integer] :Integer                           read prGetIndex                       write prSetIndex;

  property CType                 :TQUB_QFS_QMF_ConstraintAmpVarType read prGetCType;


end;


//******************************************************************************
TQUB_QFS_StatePanels = class(TList)
protected


  function prGetPanel(Index :Integer) :TQUB_QFS_StatePanel;


public


  destructor Destroy; override;


  procedure Clear; override;

  function NewPanel :TQUB_QFS_StatePanel;
  procedure DelPanel(p :TQUB_QFS_StatePanel);


  property Panel[Index :Integer] :TQUB_QFS_StatePanel read prGetPanel;


end;


TQUB_QFS_RatePanels = class(TList)
protected


  function prGetPanel(Index :Integer) :TQUB_QFS_RatePanel;


public


  destructor Destroy; override;


  procedure Clear; override;

  function NewPanel :TQUB_QFS_RatePanel;
  procedure DelPanel(p :TQUB_QFS_RatePanel);

  property Panel[Index :Integer] :TQUB_QFS_RatePanel read prGetPanel;


end;


//******************************************************************************
TQUB_QFS_Line = class(TObject)
protected


  prPoints                                     :Array[1..2] of TPoint;


public


  Left                                         :Integer;
  Top                                          :Integer;
  Width                                        :Integer;
  Height                                       :Integer;


  class procedure Draw(Canvas :TCanvas; x1, y1, x2, y2 :Integer);
  procedure SetBounds(Left, Top, Width, Height: Integer);


end;


TRatePanelPosList = class
protected
  prPanels        :Array of TQUB_QFS_RatePanel;
  prRects         :Array of TRotRect;
  prSize          :Integer;

  function prGetPanel(i :Integer): TQUB_QFS_RatePanel;
  procedure prSetPanel(i :Integer; x :TQUB_QFS_RatePanel);
  function prGetRect(i :Integer): TRotRect;
  procedure prSetSize(x :Integer);

public
  constructor Create;
  destructor Destroy; override;

  property Panel[i :Integer]  :TQUB_QFS_RatePanel   read prGetPanel  write prSetPanel;
  property Rect[i :Integer]   :TRotRect             read prGetRect;
  property Size               :Integer              read prSize      write prSetSize;
end;



implementation


//****************************************************************************//
//                                                                            //
//                                 S T A T E S                                //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_States.Create(const Name :String; Parent :TQUB_QFS_Record);
begin

  inherited Create(Name, Parent);

  prPanels  := TQUB_QFS_StatePanels.Create;

end;

destructor TQUB_QFS_States.Destroy;
begin

  inherited Destroy;

  prPanels.Free;

end;

procedure TQUB_QFS_States.Reset;
begin
  inherited;
  prCurrState := nil;
  if prPanels <> nil then prPanels.Clear;
end;

function TQUB_QFS_States.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = CHANNELS_STATE_NAME      then begin
      Result := NewState(-1, -1);
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

//******************************************************************************
function TQUB_QFS_States.NewState(x, y :Double) :TQUB_QFS_State;
begin
  Result := TQUB_QFS_State.CreateParam(CHANNELS_STATE_NAME, Self, x, y, prPanels.NewPanel);
  Result.Panel.State := Result;
  CurrState := Result;
  UpdateStates;
end;

procedure TQUB_QFS_States.DelState(State :TQUB_QFS_State);
var
  p :TQUB_QFS_StatePanel;
begin
  p := State.Panel;
  if not DelRecord(State) then begin

    prPanels.DelPanel(p);

    if prCurrState = State then prCurrState := nil;

    CurrState := Self[StateCount - 1];
    UpdateStates;

  end;
end;

function TQUB_QFS_States.prGetState(s :Integer) :TQUB_QFS_State;
begin
  Result := TQUB_QFS_State(Records[s]);
end;

function TQUB_QFS_States.WhichState(Panel :TQUB_QFS_StatePanel) :TQUB_QFS_State;
var
  i :Integer;
begin
  Result := nil;
  for i := 0 to StateCount - 1 do begin
    if State[i].Panel = Panel then begin
      Result := State[i];
      Exit;
    end;
  end;
end;

function TQUB_QFS_States.IndexOf(State :TQUB_QFS_State) :Integer;
var
  i :Integer;
begin
  Result := -1;
  for i := 0 to StateCount - 1 do begin
    if Self.State[i] = State then begin
      Result := i;
      Exit;
    end;
  end;
end;

procedure TQUB_QFS_States.prSetCurrState(State :TQUB_QFS_State);
var
  i :Integer;
begin
  if prCurrState <> nil then begin
    prCurrState.Deactivate;
    if Rates <> nil then begin
      for i := 0 to Rates.RateCount - 1 do begin
        if (Rates[i].States[1] = prCurrState) or (Rates[i].States[2] = prCurrState) then begin
          Rates[i].DeActivate;
        end;
      end;
    end;
  end;
  prCurrState := State;
  if State <> nil then begin
    prCurrState.Activate;
    if Rates <> nil then begin
      for i := 0 to Rates.RateCount - 1 do begin
        if (Rates[i].States[1] = prCurrState) or (Rates[i].States[2] = prCurrState) then begin
          Rates[i].Activate;
        end;
      end;
    end;
  end;
  if Assigned(OnChangeActiveState) then OnChangeActiveState(CurrState);
end;

procedure TQUB_QFS_States.UpdateStates(lClassLabels :Boolean = False);
const
  acPrefix :array[0..26] of Char = ('C','O','P','Q','R','S','T','U','V','W','X','Y','Z','N','M','L','K','J','I','H','G','F','E','D','B','A','.');
var
  i        :Integer;

  //----- Option to label states as C1,C2,...; O1,O2,...; P1,P2,...;       2004-02 {JB}
  iIndex   :Integer;
  aiCount  :array[0..26] of Integer;
begin

  //----- Initialize count of States/Class to 0
  for i := 0 to 26 do begin
    aiCount[i] := 0;
  end;

  //----- Label Each State
  for i := 0 to StateCount - 1 do begin
    if lClassLabels then begin
    //-- Label states in C1,C2,...,O1,O2,... mode
    //-- Note : if the option to view states as names is added, will need to prevent QUB_QVO_QVM_Print_Properties.prCopy from setting/resetting
      iIndex := Min(State[i].Cls,26);
      aiCount[iIndex] := aiCount[iIndex] + 1;
      State[i].Panel.SetCaptionC(acPrefix[iIndex]+IntToStr(aiCount[iIndex]));
    end else begin
      //-- Label States in 1,2,3... mode
      State[i].Panel.SetCaptionI(i);
      State[i].Panel.Cls := State[i].Cls;
    end;
  end;
end;


//****************************************************************************//
//                                                                            //
//                                  S T A T E                                 //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_State.Create(const Name :String; Parent :TQUB_QFS_Record);
begin

  inherited Create(Name, Parent);

  //Child records
  prRecX      := TQUB_QFS_Record.CreateAsValue(CHANNELS_STATE_X_NAME,     Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  1{Count});
  prRecY      := TQUB_QFS_Record.CreateAsValue(CHANNELS_STATE_Y_NAME,     Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  1{Count});
  prRecClass  := TQUB_QFS_Record.CreateAsValue(CHANNELS_STATE_CLASS_NAME, Self, True{Permanent}, 0, NODE_DATATYPE_LONGINT, 1{Count});
  prRecPr     := TQUB_QFS_Record.CreateAsValue(CHANNELS_STATE_PR_NAME,    Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  1{Count});
  prRecGr     := TQUB_QFS_Record.CreateAsValue(CHANNELS_STATE_GR_NAME,    Self, True{Permanent}, 0, NODE_DATATYPE_LONGINT, 1{Count});

end;


constructor TQUB_QFS_State.CreateParam(const Name :String; Parent :TQUB_QFS_Record; x, y :Double; Panel :TQUB_QFS_StatePanel);
begin

  Create(Name, Parent);

  prRecX.DataValue     := x;
  prRecY.DataValue     := y;
  prRecPr.DataValue    := 0;
  prRecGr.DataValue    := 0;
  prRecClass.DataValue := 0;

  prPanel := Panel;

end;

function TQUB_QFS_State.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = CHANNELS_STATE_X_NAME      then begin
      Result := prRecX;
    end else if NodeName = CHANNELS_STATE_Y_NAME      then begin
      Result := prRecY;
    end else if NodeName = CHANNELS_STATE_CLASS_NAME  then begin
      Result := prRecClass;
    end else if NodeName = CHANNELS_STATE_PR_NAME     then begin
      Result := prRecPr;
    end else if NodeName = CHANNELS_STATE_GR_NAME     then begin
      Result := prRecGr;
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;


//Activate/DeActivate
procedure TQUB_QFS_State.Activate;
begin
  Panel.Activate;
end;

procedure TQUB_QFS_State.DeActivate;
begin
  Panel.Deactivate;
end;

//Change Class
procedure TQUB_QFS_State.ChangeClass;
begin
  Cls := (Cls + 1); //  mod QUB_QFS_QMF_MAX_CLASS;
end;

//StateSize
procedure TQUB_QFS_State.prSetStateSize(s :Integer);
begin
  prPanel.SetBounds(prPanel.Left + prStateSize div 2 - s div 2, prPanel.Top + prStateSize div 2 - s div 2, s, s);
  prStateSize := s;
end;

//X, Y
procedure TQUB_QFS_State.prSetX(x :Double);
begin
  prRecX.DataValue := x;
end;

function TQUB_QFS_State.prGetX :Double;
begin
  Result := prRecX.DataValue;
end;

procedure TQUB_QFS_State.prSetY(y :Double);
begin
  prRecY.DataValue := y;
end;

function TQUB_QFS_State.prGetY :Double;
begin
  Result := prRecY.DataValue;
end;

//Pr
function TQUB_QFS_State.prGetPr :Double;
begin
  Result := prRecPr.DataValue;
end;

procedure TQUB_QFS_State.prSetPr(p :Double);
begin
  if prRecPr.DataValue <> p then begin
    prRecPr.DataValue := StrToFloat(FormatFloat('0.000000e+000', p));
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

//Group
function TQUB_QFS_State.prGetGr :Integer;
begin
  Result := prRecGr.DataValue;
end;

procedure TQUB_QFS_State.prSetGr(g :Integer);
begin
  if prRecGr.DataValue <> g then begin
    prRecGr.DataValue := g;
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

//Class
procedure TQUB_QFS_State.prSetClass(c :Integer);
begin
  if prRecClass.DataValue <> c then begin
    prRecClass.DataValue := c;
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

function TQUB_QFS_State.prGetClass :Integer;
begin
  Result := prRecClass.DataValue;
end;


//******************************************************************************
procedure TQUB_QFS_State.prSetAmp(a :Double);
begin
  if prAmp <> a then begin
    prAmp := StrToFloat(FormatFloat('0.000000e+000', a));
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

procedure TQUB_QFS_State.prSetStd(x :Double);
begin
  if prStd <> x then begin
    prStd := StrToFloat(FormatFloat('0.000000e+000', x));
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

procedure TQUB_QFS_State.prSetCond(c :Double);
begin
  if prCond <> c then begin
    prCond := StrToFloat(FormatFloat('0.000000e+000', c));
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

procedure TQUB_QFS_State.prSetCondStd(x :Double);
begin
  if prCondStd <> x then begin
    prCondStd := StrToFloat(FormatFloat('0.000000e+000', x));
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

procedure TQUB_QFS_State.prSetVrev(x :Double);
begin
  if prVrev <> x then begin
    prVrev := StrToFloat(FormatFloat('0.000000e+000', x));
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

function TQUB_QFS_State.prGetCff(Index :Integer) :Double;
begin
  Result := prCff[Index];
end;

procedure TQUB_QFS_State.prSetCff(Index :Integer; c :Double);
begin
  if prCff[Index] <> c then begin
    prCff[Index] := StrToFloat(FormatFloat('0.000000e+000', c));
    if Assigned(OnEdit) then OnEdit(Self);
  end;
  SetHint;
end;

procedure TQUB_QFS_State.SetHint;
begin
  prPanel.Hint := Format('pr0 = %2.6g'#13#10'amp = %2.6g'#13#10'std = %2.6g'#13#10'cond = %2.6g'#13#10'condstd = %2.6g'#13#10'grp = %d', [Pr, Amp, Std, Cond, CondStd, Gr]);
end;


//****************************************************************************//
//                                                                            //
//                                  R A T E S                                 //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_Rates.Create(const Name :String; Parent :TQUB_QFS_Record);
begin

  PreFormat := QUB_NODE_Null;
  ExpFormat := QUB_NODE_Null;

  inherited Create(Name, Parent);

  prCurrRateIdx := -1;
  prCurrHalfIdx := 1;
  prCurrK01     := QUB_QFS_QMF_RATE_K0;
  prPanels      := TQUB_QFS_RatePanels.Create;

end;

destructor TQUB_QFS_Rates.Destroy;
begin

  inherited Destroy;

  prPanels.Free;

end;

procedure TQUB_QFS_Rates.Reset;
begin
  inherited;
  prCurrRateIdx := -1;
  prCurrHalfIdx := 1;
  prCurrK01     := QUB_QFS_QMF_RATE_K0;
  if prPanels <> nil then prPanels.Clear;
end;

function TQUB_QFS_Rates.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = CHANNELS_RATE_NAME      then begin
      Result := NewRate(nil, nil);
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

procedure TQUB_QFS_Rates.prSetPreFormat(x :INode);
var
  i           :Integer;
begin
  prPreFormat := x;
  for i := 0 to RateCount - 1 do begin
    Rate[i].PreFormat := x;
    Rate[i].UpdateCaptions;
  end;
end;

procedure TQUB_QFS_Rates.prSetExpFormat(x :INode);
var
  i           :Integer;
begin
  prExpFormat := x;
  for i := 0 to RateCount - 1 do begin
    Rate[i].ExpFormat := x;
    Rate[i].UpdateCaptions;
  end;
end;

//******************************************************************************
function TQUB_QFS_Rates.NewRate(State1, State2 :TQUB_QFS_State) :TQUB_QFS_Rate;
begin
  Result        := TQUB_QFS_Rate.CreateParam(CHANNELS_RATE_NAME, Self, State1, State2, prPanels.NewPanel, prPanels.NewPanel, prPanels.NewPanel, prPanels.NewPanel);
  Result.PreFormat     := PreFormat;
  Result.ExpFormat     := ExpFormat;
  Result.DiagonalRates := DiagonalRates;
  Result.ShowK1        := ShowK1;
  prCurrRateIdx := RecordCount - 1;
  prCurrHalfIdx := 1;
  prCurrK01     := QUB_QFS_QMF_RATE_K0;
  prUpdateRates;
end;

procedure TQUB_QFS_Rates.DelRate(Rate :TQUB_QFS_Rate);
var
  p1, p2, p3, p4 :TQUB_QFS_RatePanel;
begin
  p1 := Rate.Panels[1];
  p2 := Rate.Panels[2];
  p3 := Rate.Panels[3];
  p4 := Rate.Panels[4];
  if not DelRecord(Rate) then begin
    prPanels.DelPanel(p1);
    prPanels.DelPanel(p2);
    prPanels.DelPanel(p3);
    prPanels.DelPanel(p4);
    if prCurrRateIdx > (RecordCount - 1) then prCurrRateIdx := RecordCount - 1;
    prUpdateRates;
  end;
end;

procedure TQUB_QFS_Rates.prSetDiagonalRates(x :Boolean);
var
  i    :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    Rate[i].DiagonalRates := x;
  end;
  prDiagonalRates := x;
  prUpdateRates;
end;

procedure TQUB_QFS_Rates.prSetShowK1(x :Boolean);
var
  i    :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    Rate[i].ShowK1 := x;
  end;
  prShowK1 := x;
  prUpdateRates;
end;

function TQUB_QFS_Rates.WhichPanel(Pos :TPoint): TQUB_QFS_RatePanel;
var
  i, j :Integer;
begin
  Result := nil;
  for i := 0 to RateCount - 1 do begin
    for j := 1 to 4 do begin
      if Rate[i].Panels[j].Visible and Rate[i].Panels[j].Rect.PtInRect(Pos) then begin
        Result := Rate[i].Panels[j];
      end;
    end;
  end;
end;

procedure TQUB_QFS_Rates.GetRateIdxHalfIdxOfStates(StateIdx1, StateIdx2 :Integer; var RateIdx, HalfIdx :Integer);
var
  s1, s2 :Integer;
  i      :Integer;
begin
  RateIdx := -1;
  HalfIdx := 0;
  for i := 0 to RateCount - 1 do begin
    s1 := States.IndexOf(Rate[i].States[1]);
    s2 := States.IndexOf(Rate[i].States[2]);

    if          (s1 = StateIdx1) and (s2 = StateIdx2) then begin
      RateIdx := i;
      HalfIdx := 1;
      Break;
    end else if (s1 = StateIdx2) and (s2 = StateIdx1) then begin
      RateIdx := i;
      HalfIdx := 2;
      Break;
    end;
  end;
end;

procedure TQUB_QFS_Rates.GetStatesOfRate(RateIdx, HalfIdx :Integer; var StateIdx1, StateIdx2 :Integer);
var
  Rate       :TQUB_QFS_Rate;
begin
  Rate       := prGetRate(RateIdx);
  if HalfIdx = 1 then begin
    StateIdx1 := Rate.StateIndex[1];
    StateIdx2 := Rate.StateIndex[2];
  end else begin
    StateIdx1 := Rate.StateIndex[2];
    StateIdx2 := Rate.StateIndex[1];
  end;
end;

function TQUB_QFS_Rates.prGetRate(s :Integer) :TQUB_QFS_Rate;
begin
  Result := TQUB_QFS_Rate(Records[s]);
end;

function TQUB_QFS_Rates.prGetRateIndex(r :TQUB_QFS_Rate) :Integer;
var
  i :Integer;
begin
  Result := -1;
  for i := 0 to RateCount - 1 do begin
    if Rate[i] = r then begin
      Result := i;
      Break;
    end;
  end;
end;

function TQUB_QFS_Rates.prGetRateOfStates(State1, State2 :TQUB_QFS_State) :TQUB_QFS_Rate;
var
  i :Integer;
begin
  Result := nil;
  for i := 0 to RateCount - 1 do begin
    if ((Rate[i].States[1] = State1) and (Rate[i].States[2] = State2)) or ((Rate[i].States[1] = State2) and (Rate[i].States[2] = State1)) then begin
      Result := Rate[i];
      Break;
    end;
  end;
end;

function TQUB_QFS_Rates.prGetRateValueOfStates(State1, State2 :TQUB_QFS_State) :Double;
var
  i :Integer;
begin
  Result := 0;
  for i := 0 to RateCount - 1 do begin
    if          ((Rate[i].States[1] = State1) and (Rate[i].States[2] = State2)) then begin
      Result := Rate[i].Value[1];
      Break;
    end else if ((Rate[i].States[1] = State2) and (Rate[i].States[2] = State1)) then begin
      Result := Rate[i].Value[2];
      Break;
    end;
  end;
end;

procedure TQUB_QFS_Rates.SetRateK0OfStates(State1, State2 :TQUB_QFS_State; k0 :Double);
var
  i :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    if          ((Rate[i].States[1] = State1) and (Rate[i].States[2] = State2)) then begin
      Rate[i].K0[1] := k0;
      Break;
    end else if ((Rate[i].States[1] = State2) and (Rate[i].States[2] = State1)) then begin
      Rate[i].K0[2] := k0;
      Break;
    end;
  end;
end;

procedure TQUB_QFS_Rates.DoMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
var
  i, j   :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    for j := 1 to 4 do begin
      Rate[i].Panels[j].DoMouseDown(Sender, Button, Shift, X, Y, Handled);
      if Handled then Exit;
    end;
  end;
end;

function TQUB_QFS_Rates.DoMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer; var Handled :Boolean): TQUB_QFS_RatePanel;
var
  i, j   :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    for j := 1 to 4 do begin
      Rate[i].Panels[j].DoMouseMove(Sender, Shift, X, Y, Handled);
      if Handled then begin
        Result := Rate[i].Panels[j];
        Exit;
      end;
    end;
  end;
  Result := nil;
end;

procedure TQUB_QFS_Rates.DoMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer; var Handled :Boolean);
var
  i, j   :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    for j := 1 to 4 do begin
      Rate[i].Panels[j].DoMouseUp(Sender, Button, Shift, X, Y, Handled);
      if Handled then Exit;
    end;
  end;
end;

procedure TQUB_QFS_Rates.DoMouseWheel(Sender :TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled :Boolean);
var
  i, j   :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    for j := 1 to 4 do begin
      Rate[i].Panels[j].DoMouseWheel(Sender, Shift, WheelDelta, MousePos, Handled);
      if Handled then Exit;
    end;
  end;
end;

procedure TQUB_QFS_Rates.DoDblClick(Sender :TObject; X, Y :Integer; var Handled :Boolean);
var
  i, j   :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    for j := 1 to 4 do begin
      Rate[i].Panels[j].DoDblClick(Sender, X, Y, Handled);
      if Handled then Exit;
    end;
  end;
end;

//TODO: finish! LM
function TQUB_QFS_Rates.CopyFrom(Source :TQUB_QFS_Rates) :Boolean;
begin
  Result := False;
  while RateCount > Source.RateCount do begin
    DelRate(Self[RateCount - 1]);
  end;
  while RateCount < Source.RateCount do begin
    NewRate(nil, nil);
  end;
end;

function TQUB_QFS_Rates.UpdateFrom(Source :TQUB_QFS_Rates; Change :Boolean) :Boolean;
var
  i :Integer;
begin
  Result := False;
  for i := 0 to RateCount - 1 do begin
    if Self[i].UpdateFrom(Source[i], Change) then begin
      Result := True;
    end;
  end;
end;

procedure TQUB_QFS_Rates.UpdateCaptions;
var
  i :Integer;
begin
  for i := 0 to RateCount - 1 do begin
    Self[i].UpdateCaptions;
  end;
end;

procedure TQUB_QFS_Rates.prUpdateRates;
begin
end;


//****************************************************************************//
//                                                                            //
//                                   R A T E                                  //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_Rate.Create(const Name :String; Parent :TQUB_QFS_Record);
begin

  inherited Create(Name, Parent);

  PreFormat := QUB_NODE_Null;
  ExpFormat := QUB_NODE_Null;

  //Child records
  prRecStates  := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_STATES_NAME, Self, True{Permanent}, 0, NODE_DATATYPE_LONGINT,  2{Count});
  prRecStates.DataArrayValue[0] := -1;
  prRecStates.DataArrayValue[1] := -1;

  prRecK0      := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_K0_NAME,     Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  2{Count});
  prRecK0.DataArrayValue[0] := 0;
  prRecK0.DataArrayValue[1] := 0;

  prRecK1      := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_K1_NAME,     Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  2{Count});
  prRecK1.DataArrayValue[0] := 0;
  prRecK1.DataArrayValue[1] := 0;

  prRecdK0     := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_dK0_NAME,    Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  2{Count});
  prRecdK0.DataArrayValue[0] := 0;
  prRecdK0.DataArrayValue[1] := 0;

  prRecdK1     := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_dK1_NAME,    Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  2{Count});
  prRecK1.DataArrayValue[0] := 0;
  prRecK1.DataArrayValue[1] := 0;

  prRecP       := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_P_NAME,      Self, True{Permanent}, 0, NODE_DATATYPE_LONGINT, 2{Count});
  prRecP.DataArrayValue[0] := Integer(False);
  prRecP.DataArrayValue[1] := Integer(False);

  prRecQ       := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_Q_NAME,      Self, True{Permanent}, 0, NODE_DATATYPE_LONGINT, 2{Count});
  prRecQ.DataArrayValue[0] := Integer(False);
  prRecQ.DataArrayValue[1] := Integer(False);

  prRecPNames  := TQUB_QFS_Record.Create(CHANNELS_RATE_PNAMES_NAME, Self);
  prRecPNames.Permanent := True;
  TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_PNAME_NAME,  prRecPNames, False{Permanent}, CHANNELS_RATE_DEF_PNAME, NODE_DATATYPE_STRING,  0{Count});
  TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_PNAME_NAME,  prRecPNames, False{Permanent}, CHANNELS_RATE_DEF_PNAME, NODE_DATATYPE_STRING,  0{Count});

  prRecQNames  := TQUB_QFS_Record.Create(CHANNELS_RATE_QNAMES_NAME, Self);
  prRecQNames.Permanent := True;
  TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_QNAME_NAME,  prRecQNames, False{Permanent}, CHANNELS_RATE_DEF_QNAME, NODE_DATATYPE_STRING,  0{Count});
  TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_QNAME_NAME,  prRecQNames, False{Permanent}, CHANNELS_RATE_DEF_QNAME, NODE_DATATYPE_STRING,  0{Count});

  prRecPValue  := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_PVALUE_NAME, Self, True{Permanent}, 0, NODE_DATATYPE_POINTER, 2{Count});
  prRecPValue.DataArrayValue[0] := Integer(nil);
  prRecPValue.DataArrayValue[1] := Integer(nil);

  prRecQValue  := TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_QVALUE_NAME, Self, True{Permanent}, 0, NODE_DATATYPE_POINTER, 2{Count});
  prRecQValue.DataArrayValue[0] := Integer(nil);
  prRecQValue.DataArrayValue[1] := Integer(nil);

  prRecRateFormats := TQUB_QFS_Record.Create(CHANNELS_RATE_FORMATS_NAME, Self);
  prRecRateFormats.Permanent := True;
  TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_FORMAT_NAME, prRecRateFormats, False{Permanent}, CHANNELS_RATE_DEF_FORMAT, NODE_DATATYPE_STRING, 0{Count});
  TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_FORMAT_NAME, prRecRateFormats, False{Permanent}, CHANNELS_RATE_DEF_FORMAT, NODE_DATATYPE_STRING, 0{Count});
end;

constructor TQUB_QFS_Rate.CreateParam(const Name :String; Parent :TQUB_QFS_Record; State1, State2 :TQUB_QFS_State; Panel1, Panel2, Panel3, Panel4 :TQUB_QFS_RatePanel);
var
  i :Integer;
begin

  Create(Name, Parent);

  PreFormat := QUB_NODE_Null;
  ExpFormat := QUB_NODE_Null;

  States[1] := State1;
  States[2] := State2;

  prPanels[1] := Panel1;
  prPanels[1].Rate := Self;
  prPanels[1].Dir  := 1;
  prPanels[1].K01  := QUB_QFS_QMF_RATE_K0;
  prPanels[2] := Panel2;
  prPanels[2].Rate := Self;
  prPanels[2].Dir  := 2;
  prPanels[2].K01  := QUB_QFS_QMF_RATE_K0;
  prPanels[3] := Panel3;
  prPanels[3].Rate := Self;
  prPanels[3].Dir  := 1;
  prPanels[3].K01  := QUB_QFS_QMF_RATE_K1;
  prPanels[4] := Panel4;
  prPanels[4].Rate := Self;
  prPanels[4].Dir  := 2;
  prPanels[4].K01  := QUB_QFS_QMF_RATE_K1;

  {//Are they necessary ?
  prPanels[1].Width  := QUB_QFS_QMF_RATE_WIDTH;
  prPanels[2].Width  := QUB_QFS_QMF_RATE_WIDTH;
  prPanels[1].Height := QUB_QFS_QMF_RATE_HEIGHT;
  prPanels[2].Height := QUB_QFS_QMF_RATE_HEIGHT;
  }
  for i := 1 to 2 do begin
    K0[i]     := QUB_QFS_QMF_DEF_K0;
    K1[i]     := QUB_QFS_QMF_DEF_K1;
    P[i]      := QUB_QFS_QMF_DEF_P;
    Q[i]      := QUB_QFS_QMF_DEF_Q;
    PName[i]  := QUB_QFS_QMF_DEF_PNAME;
    QName[i]  := QUB_QFS_QMF_DEF_QNAME;
    PValue[i] := nil;
    QValue[i] := nil;
    RateFormat[i] := QUB_QFS_QMF_DEF_RATEFORMAT;
  end;

end;


//Does not change the Modified flag - to be used for backup
function TQUB_QFS_Rate.UpdateFrom(Source :TQUB_QFS_Rate; Change :Boolean) :Boolean;
var
  i :Integer;
begin
  Result := False;
  if Change then begin
    for i := 1 to 2 do begin
      K0[i]                          := Source.K0[i];
      K1[i]                          := Source.K1[i];
      P[i]                           := Source.P[i];
      Q[i]                           := Source.Q[i];
      PName[i]                       := Source.PName[i];
      QName[i]                       := Source.QName[i];
      RateFormat[i]                  := Source.RateFormat[i];
    end;
  end else begin
  //Does not generate events
    for i := 1 to 2 do begin
      prRecK0.DataArrayValue[i-1]    := Source.K0[i];
      prRecK1.DataArrayValue[i-1]    := Source.K1[i];
      prRecP.DataArrayValue[i-1]     := Source.P[i];
      prRecQ.DataArrayValue[i-1]     := Source.Q[i];
      prRecPNames[i-1].DataValue     := Source.PName[i];
      prRecQNames[i-1].DataValue     := Source.QName[i];
      prRecRateFormats[i-1].DataValue:= Source.RateFormat[i];
    end;
  end;
end;


function TQUB_QFS_Rate.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = CHANNELS_RATE_STATES_NAME  then begin
      Result := prRecStates;
    end else if NodeName = CHANNELS_RATE_K0_NAME      then begin
      Result := prRecK0;
    end else if NodeName = CHANNELS_RATE_K1_NAME      then begin
      Result := prRecK1;
    end else if NodeName = CHANNELS_RATE_dK0_NAME     then begin
      Result := prRecdK0;
    end else if NodeName = CHANNELS_RATE_dK1_NAME     then begin
      Result := prRecdK1;
    end else if NodeName = CHANNELS_RATE_P_NAME       then begin
      Result := prRecP;
    end else if NodeName = CHANNELS_RATE_Q_NAME       then begin
      Result := prRecQ;
    end else if NodeName = CHANNELS_RATE_PNAMES_NAME  then begin
      Result := prRecPNames;
    end else if NodeName = CHANNELS_RATE_QNAMES_NAME  then begin
      Result := prRecQNames;
    end else if NodeName = CHANNELS_RATE_PVALUE_NAME  then begin
      Result := prRecPValue;
    end else if NodeName = CHANNELS_RATE_QVALUE_NAME  then begin
      Result := prRecQValue;
    end else if NodeName = CHANNELS_RATE_FORMATS_NAME then begin
      Result := prRecRateFormats;
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

function TQUB_QFS_Rate.Write(WriteStream :TStream) :Boolean;
begin
  try
    if States[1] <> nil then begin
      prRecStates.DataArrayValue[0] := States[1].Parent.RecordIndex(States[1]);
    end else begin
      prRecStates.DataArrayValue[0] := -1;
    end;
    if States[2] <> nil then begin
      prRecStates.DataArrayValue[1] := States[2].Parent.RecordIndex(States[2]);
    end else begin
      prRecStates.DataArrayValue[1] := -1;
    end;
  finally
    Result := inherited Write(WriteStream);
  end;
end;

procedure TQUB_QFS_Rate.Activate;
begin
  prPanels[1].Activate;
  prPanels[2].Activate;
  prPanels[3].Activate;
  prPanels[4].Activate;
end;

procedure TQUB_QFS_Rate.Deactivate;
begin
  prPanels[1].DeActivate;
  prPanels[2].DeActivate;
  prPanels[3].DeActivate;
  prPanels[4].DeActivate;
end;

procedure TQUB_QFS_Rate.Update;
var
  s          :Array[1..2] of TPoint;
  r          :Array[1..2] of TPoint;
  m          :TPoint;
  sina, cosa :Single;
  dx, dy     :Single;
  Flipped    :Boolean;
  l          :Integer;
  w          :Integer;
  rw         :Array[1..4] of Integer;
  rw2        :Array[1..4] of Integer;
  TextsPos   :Array[1..2] of TPoint;
begin

  if (States[1] = nil) or (States[2] = nil) then Exit;

  prPanels[1].Anchors := [akBottom];
  prPanels[2].Anchors := [akBottom]; // different for k0 and k1 display
  prPanels[3].Visible := False;
  prPanels[4].Visible := False;

  w  := States[1].Panel.Width;

  rw[1] := Panels[1].Rect.Width;
  rw[2] := Panels[2].Rect.Width;
  rw[3] := Panels[3].Rect.Width;
  rw[4] := Panels[4].Rect.Width;
  rw2[1] := rw[1] div 2;
  rw2[2] := rw[2] div 2;
  rw2[3] := rw[3] div 2;
  rw2[4] := rw[4] div 2;

  //Coordinates of the line between the centers of the two state panels
  prLinesPos[1].x := States[1].Panel.Left + w div 2;
  prLinesPos[1].y := States[1].Panel.Top  + w div 2;
  prLinesPos[2].x := States[2].Panel.Left + w div 2;
  prLinesPos[2].y := States[2].Panel.Top  + w div 2;

  //Cordinates for the middle of the above line
  m.x := (prLinesPos[1].x + prLinesPos[2].x) div 2;
  m.y := (prLinesPos[1].y + prLinesPos[2].y) div 2;

  //Calculate the hypotenuse - the 2D distance between the centers of the two state panels
  l := Round(Hypot(prLinesPos[2].x - prLinesPos[1].x, prLinesPos[2].y - prLinesPos[1].y));
  if l = 0 then l := 1;

  //s has the coordinates of the line, un-rotated (parallel with x axis)
  s[1] := prLinesPos[1];
  s[2].x := s[1].x + l;
  s[2].y := s[1].y;

  //r[1] and r[2] are the coordinates of the two text boxes
  r[1].x := s[1].x + l div 2;
  r[1].y := s[1].y - (rw2[1]{QUB_QFS_QMF_RATE_WIDTH2}); //to be replaced
  r[2].x := r[1].x;
  r[2].y := s[1].y + (rw2[2]{QUB_QFS_QMF_RATE_WIDTH2}); //to be replaced

  dx := prLinesPos[2].x - prLinesPos[1].x;
  if dx = 0 then dx := 1e-10;
  dy := prLinesPos[2].y - prLinesPos[1].y;
  if dy = 0 then dy := 1e-10;

  //the sinus and cosinus of the angle between the line connecting the two states and the x axis
  sina := dy / l;
  cosa := dx / l;

  prAngle := - ArcTan2(dy, dx);

  if prDiagonalRates then begin
    Flipped := Abs(prAngle) > (PI/2);

    if prShowK1 and Q[1] then begin
      prPanels[1].Anchors := [akBottom, akRight];
      prPanels[3].Anchors := [akBottom, akLeft];
      if Flipped then begin
        prPanels[1].SetBounds(r[1].x, s[1].y-QUB_QFS_QMF_RATE_HEIGHT+1, rw[1], QUB_QFS_QMF_RATE_HEIGHT);
        prPanels[3].SetBounds(r[1].x - rw[3], s[1].y-QUB_QFS_QMF_RATE_HEIGHT+1, rw[3], QUB_QFS_QMF_RATE_HEIGHT);
      end else begin
        prPanels[1].SetBounds(r[1].x - rw[1], s[1].y-QUB_QFS_QMF_RATE_HEIGHT+1, rw[1], QUB_QFS_QMF_RATE_HEIGHT);
        prPanels[3].SetBounds(r[1].x, s[1].y-QUB_QFS_QMF_RATE_HEIGHT+1, rw[3], QUB_QFS_QMF_RATE_HEIGHT);
      end;
      prPanels[1].Rotate(s[1].x, s[1].y, prAngle);
      prPanels[3].Rotate(s[1].x, s[1].y, prAngle);
      prPanels[3].Visible := True;
    end else begin
      prPanels[1].SetBounds(r[1].x - rw2[1], s[1].y-QUB_QFS_QMF_RATE_HEIGHT+1, rw[1], QUB_QFS_QMF_RATE_HEIGHT);
      prPanels[1].Rotate(s[1].x, s[1].y, prAngle);
    end;

    if prShowK1 and Q[2] then begin
      prPanels[2].Anchors := [akBottom, akRight];
      prPanels[4].Anchors := [akBottom, akLeft];
      if Flipped then begin
        prPanels[2].SetBounds(r[2].x, s[2].y+3, rw[2], QUB_QFS_QMF_RATE_HEIGHT);
        prPanels[4].SetBounds(r[2].x - rw[4], s[1].y+3, rw[4], QUB_QFS_QMF_RATE_HEIGHT);
      end else begin
        prPanels[2].SetBounds(r[2].x - rw[2], s[2].y+3, rw[2], QUB_QFS_QMF_RATE_HEIGHT);
        prPanels[4].SetBounds(r[2].x, s[1].y+3, rw[4], QUB_QFS_QMF_RATE_HEIGHT);
      end;
      prPanels[2].Rotate(s[1].x, s[1].y, prAngle);
      prPanels[4].Rotate(s[1].x, s[1].y, prAngle);
      prPanels[4].Visible := True;
    end else begin
      prPanels[2].SetBounds(r[2].x - rw2[2], s[1].y+3, rw[2], QUB_QFS_QMF_RATE_HEIGHT);
      prPanels[2].Rotate(s[1].x, s[1].y, prAngle);
    end;

  end else begin

    //Calculate the rotated coordinates of the two rate panels
    TextsPos[1].x := Round(s[1].x + ((r[1].x - s[1].x)*cosa - (r[1].y - s[1].y)*sina));
    TextsPos[1].y := Round(s[1].y + ((r[1].x - s[1].x)*sina + (r[1].y - s[1].y)*cosa));
    TextsPos[2].x := Round(s[1].x + ((r[2].x - s[1].x)*cosa - (r[2].y - s[1].y)*sina));
    TextsPos[2].y := Round(s[1].y + ((r[2].x - s[1].x)*sina + (r[2].y - s[1].y)*cosa));

    //Check if they do overlap, and adjust, if necessary
    if abs(TextsPos[1].y - TextsPos[2].y) > (QUB_QFS_QMF_RATE_HEIGHT + 2) then begin
      if TextsPos[2].y < TextsPos[1].y then begin
        TextsPos[1].y := m.y + ((QUB_QFS_QMF_RATE_HEIGHT2) + 1);
        TextsPos[2].y := m.y - ((QUB_QFS_QMF_RATE_HEIGHT2) + 1);
      end else begin
        TextsPos[1].y := m.y - ((QUB_QFS_QMF_RATE_HEIGHT2) + 1);
        TextsPos[2].y := m.y + ((QUB_QFS_QMF_RATE_HEIGHT2) + 1);
      end;
    end;

    //Save the two rate panels coordinates
    TextsPos[1].x := TextsPos[1].x - (rw2[1]{QUB_QFS_QMF_RATE_WIDTH2});
    TextsPos[1].y := TextsPos[1].y - (QUB_QFS_QMF_RATE_HEIGHT2);
    TextsPos[2].x := TextsPos[2].x - (rw2[2]{QUB_QFS_QMF_RATE_WIDTH2});
    TextsPos[2].y := TextsPos[2].y - (QUB_QFS_QMF_RATE_HEIGHT2);

    //Set the rate panels position
    prPanels[1].SetBounds(TextsPos[1].x, TextsPos[1].y, rw[1]{QUB_QFS_QMF_RATE_WIDTH}, QUB_QFS_QMF_RATE_HEIGHT);
    prPanels[2].SetBounds(TextsPos[2].x, TextsPos[2].y, rw[2]{QUB_QFS_QMF_RATE_WIDTH}, QUB_QFS_QMF_RATE_HEIGHT);
  end;

  //Calculate the coordinates of the arrows

  //First panel - the rightmost arrow
  //the arrow tail
  r[1].x := s[2].x - w div 2 - w div 4;
  r[1].y := s[1].y           - w div 4;
  prArrowPos[1][1].x := Round(s[1].x + ((r[1].x - s[1].x)*cosa - (r[1].y - s[1].y)*sina));
  prArrowPos[1][1].y := Round(s[1].y + ((r[1].x - s[1].x)*sina + (r[1].y - s[1].y)*cosa));
  //the arrow head
  r[1].x := s[2].x - w div 2;
  r[1].y := s[1].y;
  prArrowPos[1][2].x := Round(s[1].x + ((r[1].x - s[1].x)*cosa - (r[1].y - s[1].y)*sina));
  prArrowPos[1][2].y := Round(s[1].y + ((r[1].x - s[1].x)*sina + (r[1].y - s[1].y)*cosa));

  //Second panel - the leftmost arrow
  //the arrow tail
  r[2].x := s[1].x + w div 2 + w div 4;
  r[2].y := s[1].y           + w div 4;
  prArrowPos[2][1].x := Round(s[1].x + ((r[2].x - s[1].x)*cosa - (r[2].y - s[1].y)*sina));
  prArrowPos[2][1].y := Round(s[1].y + ((r[2].x - s[1].x)*sina + (r[2].y - s[1].y)*cosa));
  //the arrow head
  r[2].x := s[1].x + w div 2;
  r[2].y := s[1].y;
  prArrowPos[2][2].x := Round(s[1].x + ((r[2].x - s[1].x)*cosa - (r[2].y - s[1].y)*sina));
  prArrowPos[2][2].y := Round(s[1].y + ((r[2].x - s[1].x)*sina + (r[2].y - s[1].y)*cosa));

end;

procedure TQUB_QFS_Rate.Paint(Canvas :TCanvas; dx, dy :Integer);
begin
  //Connecting Line
  Canvas.MoveTo(prArrowPos[2][2].x - dx, prArrowPos[2][2].y - dy);
  Canvas.LineTo(prArrowPos[1][2].x - dx, prArrowPos[1][2].y - dy);
  //Arrows
  Canvas.MoveTo(prArrowPos[1][1].x - dx, prArrowPos[1][1].y - dy);
  Canvas.LineTo(prArrowPos[1][2].x - dx, prArrowPos[1][2].y - dy);
  Canvas.MoveTo(prArrowPos[2][1].x - dx, prArrowPos[2][1].y - dy);
  Canvas.LineTo(prArrowPos[2][2].x - dx, prArrowPos[2][2].y - dy);
end;

procedure TQUB_QFS_Rate.PaintActive(Canvas :TCanvas; dx, dy :Integer; ActiveIdx :Integer);
begin
  //Connecting Line
  Canvas.MoveTo(prArrowPos[2][2].x - dx, prArrowPos[2][2].y - dy);
  Canvas.LineTo(prArrowPos[1][2].x - dx, prArrowPos[1][2].y - dy);
  //Arrow
  Canvas.MoveTo(prArrowPos[ActiveIdx][1].x - dx, prArrowPos[ActiveIdx][1].y - dy);
  Canvas.LineTo(prArrowPos[ActiveIdx][2].x - dx, prArrowPos[ActiveIdx][2].y - dy);
end;

function TQUB_QFS_Rate.Unconstrained(Index :Integer; Constraint :Integer) :Boolean;
//var
//  i, j :Integer;
begin
  Result := True;
  (*
  for i := 0 to prParent.Constraints.Count - 1 do begin
    if i <> Constraint then begin
      //Loop constraint, both rates are involved
      if prParent.Constraints[i].iType = CONSTRAINT_LOOP_BALANCE then begin
        for j := 0 to prParent.Constraints[i].sIndex.Count - 2 do begin
          if ((States[1] = prParent.States[prParent.Constraints[i].sIndex[j]]) and (States[2] = prParent.States[prParent.Constraints[i].sIndex[j + 1]])) or
	     ((States[2] = prParent.States[prParent.Constraints[i].sIndex[j]]) and (States[1] = prParent.States[prParent.Constraints[i].sIndex[j + 1]])) then begin
            Result := False;
            Exit;
          end;
        end;
        if ((States[1] = prParent.States[prParent.Constraints[i].sIndex[prParent.Constraints[i].sIndex.Count - 1]]) and (States[2] = prParent.States[prParent.Constraints[i].sIndex[0]])) or
	   ((States[2] = prParent.States[prParent.Constraints[i].sIndex[prParent.Constraints[i].sIndex.Count - 1]]) and (States[1] = prParent.States[prParent.Constraints[i].sIndex[0]])) then begin
          Result := False;
          Exit;
        end;
      end else begin
      //Non-Loop constraint, only one rate is involved
        for j := 0 to prParent.Constraints[i].sIndex.Count - 2 do begin
          if (Index = 1) and ((States[1] = prParent.States[prParent.Constraints[i].sIndex[j]]) and (States[2] = prParent.States[prParent.Constraints[i].sIndex[j + 1]])) or
             (Index = 2) and ((States[2] = prParent.States[prParent.Constraints[i].sIndex[j]]) and (States[1] = prParent.States[prParent.Constraints[i].sIndex[j + 1]])) then begin
            Result := False;
	    Exit;
          end;
        end;
      end;
    end;
  end;
  *)
end;

procedure TQUB_QFS_Rate.ClearPQ;
var
  i  :Integer;
begin
  for i := 1 to 2 do begin
    prRecPValue.DataArrayValue[i - 1] := 0;
    prRecQValue.DataArrayValue[i - 1] := 0;
  end;
end;

function TQUB_QFS_Rate.prGetState(Index :Integer) :TQUB_QFS_State;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := prStates[Index] else Result := nil;
end;

procedure TQUB_QFS_Rate.prSetState(Index :Integer; State :TQUB_QFS_State);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    if State <> nil then prRecStates.DataArrayValue[Index - 1] := State.Parent.RecordIndex(State);
    prStates[Index] := State;
  end;
end;

function TQUB_QFS_Rate.prGetStateIndex(Index :Integer) :Integer;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := prRecStates.DataArrayValue[Index - 1] else Result := -1;
end;

function TQUB_QFS_Rate.prGetPanel(Index :Integer) :TQUB_QFS_RatePanel;
begin
  if ((Index >= 1) and (Index <= 4)) then Result := TQUB_QFS_RatePanel(prPanels[Index]) else Result := nil;
end;

function TQUB_QFS_Rate.prGetValueFormat(Index :Integer) :String;
var
  pq :String;
begin
  if Showk1 then begin
    Result := ' ' + prGetPreExpFormat(Index, K0[Index]) + ' ';
    if Q[Index] then begin
      Result := Result + '(' + prGetExpFormat(K1[Index]) + ') ';
    end;
  end else begin
    Result := ' ' + prGetPreExpFormat(Index, prGetValue(Index)) + ' ';
    pq := '';
    if P[Index] then begin
      pq := pq + 'p';
    end;
    if Q[Index] then begin
      pq := pq + 'q';
    end;
    if pq <> ' ' then begin
      Result := Result + pq + ' ';
    end;
  end;
end;

function TQUB_QFS_Rate.prGetValue(Index :Integer) :Double;
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    try
      //if prRateExp[Index] = nil then begin
        Result := K0[Index];
        if (P[Index]) and (PValue[Index] <> nil) then begin
          Result := Result * PValue[Index]^;
        end;
        if (Q[Index]) and (QValue[Index] <> nil) then begin
          Result := Result * exp(K1[Index] * QValue[Index]^);
        end;
     { end else begin
        Result := prRateExp[Index].Value;
      end;  }
    except
      Result := 0;
    end;
  end else begin
    Result := 0;
  end;
end;

procedure TQUB_QFS_Rate.prSetHint;
var
  i         :Integer;
  delG      :Double;
begin
  for i := 1 to 2 do begin
    if (K0[1] <> 0.0) and (K0[2] <> 0.0) then begin
      prPanels[i].Hint := 'k0        = ' + Format('%8.4g'#13#10, [K0[i]]) +
                          'k1        = ' + Format('%8.4g'#13#10, [K1[i]]) +
                          'k         = ' + Format('%8.4g'#13#10, [prGetValue(i)]);

      try
        delG   := prGetValue( 1 + (i mod 2) );
        if delG <> 0 then begin
          delG := prGetValue(i) / delG;
          if delG > 0.0 then begin
            prPanels[i].Hint := prPanels[i].Hint +
                              'delG (kT) = ' + Format('%8.4g'#13#10, [Ln(delG)]);
          end;
        end;
      except
      end;

    end else begin
      prPanels[i].Hint := 'k0   = ' + Format('%8.4g'#13#10, [K0[i]]) +
                          'k1   = ' + Format('%8.4g'#13#10, [K1[i]]) +
                          'k         = ' + Format('%8.4g'#13#10, [prGetValue(i)]);
    end;

    if P[i] then begin
      prPanels[i].Hint := prPanels[i].Hint + 'P (' + PName[i] + ') = True' + #13#10;
    end;
    if Q[i] then begin
      prPanels[i].Hint := prPanels[i].Hint + 'Q (' + QName[i] + ') = True' + #13#10;
    end;
    if P[i] or Q[i] then begin
      prPanels[i].Hint := prPanels[i].Hint + 'To see net "k" at a particular time in the stimulus,'#13#10
                                           + 'click in the Data window to position the cursor.';
    end;
  end;

  prPanels[3].Hint := prPanels[1].Hint;
  prPanels[4].Hint := prPanels[2].Hint;

end;

//----- Same as 6 separate sets with less activity.
procedure TQUB_QFS_Rate.SetRate(Index :Integer; oRate :TRate);
begin
  if not ((Index >= 1) and (Index <= 2)) then Exit;

  //Why would it not have 2? LM.
  while prRecPNames.RecordCount < 2 do begin
    TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_PNAME_NAME,  prRecPNames, False{Permanent}, CHANNELS_RATE_DEF_PNAME, NODE_DATATYPE_STRING,  0{Count});
  end;
  while prRecQNames.RecordCount < 2 do begin
    TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_QNAME_NAME,  prRecQNames, False{Permanent}, CHANNELS_RATE_DEF_QNAME, NODE_DATATYPE_STRING,  0{Count});
  end;
  while prRecRateFormats.RecordCount < 2 do begin
    TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_FORMAT_NAME, prRecRateFormats, False{Permanent}, CHANNELS_RATE_DEF_FORMAT, NODE_DATATYPE_STRING, 0{Count});
  end;

  if (prRecK0.DataArrayValue[Index - 1] <> Abs(oRate.m_K0))      or
     (prRecK1.DataArrayValue[Index - 1] <> oRate.m_K1)           or
     (Boolean(prRecP.DataArrayValue[Index - 1]) <> oRate.m_P)    or
     (Boolean(prRecQ.DataArrayValue[Index - 1]) <> oRate.m_Q)    or
     (prRecPNames.Records[Index - 1].DataValue <> oRate.m_PName) or
     (prRecQNames.Records[Index - 1].DataValue <> oRate.m_QName) or
     (prRecRateFormats.Records[Index - 1].DataValue <> oRate.m_RateFormat) then begin

    prRecK0.DataArrayValue[Index - 1]         := Abs(oRate.m_K0);
    prRecK1.DataArrayValue[Index - 1]         := oRate.m_K1;
    prRecP.DataArrayValue[Index - 1]          := Integer(oRate.m_P);
    prRecQ.DataArrayValue[Index - 1]          := Integer(oRate.m_Q);
    prRecPNames.Records[Index - 1].DataValue  := oRate.m_PName;
    prRecQNames.Records[Index - 1].DataValue  := oRate.m_QName;
    prRecRateFormats.Records[Index - 1].DataValue := oRate.m_RateFormat;
    
    prUpdateCaptions(Index);

    if Assigned(OnEdit) then OnEdit(Self);
    prSetHint;

  end;
    
end;

procedure TQUB_QFS_Rate.prSetShowK1(x :Boolean);
begin
  prShowK1 := x;
  prUpdateCaptions(1);
  prUpdateCaptions(2);
end;

procedure TQUB_QFS_Rate.prUpdateCaptions(Index :Integer);
begin
  if DiagonalRates and ShowK1 and Q[Index] then begin
    prPanels[Index].Caption   := prGetK0Format(Index);
  end else begin
    prPanels[Index].Caption  := prGetValueFormat(Index);
  end;
  prPanels[Index+2].Caption := prGetK1Format(Index);
end;

procedure TQUB_QFS_Rate.UpdateCaptions;
var
  i :Integer;
begin
  for i := 1 to 2 do begin
    prUpdateCaptions(i);
  end;
end;

function TQUB_QFS_Rate.prGetPreExpFormat(Index :Integer; v :Double) :String;
begin
  if RateFormat[Index] <> '' then begin
    try
      Result := Format(RateFormat[Index], [v]);
      Exit;
    except
    end;
  end;
  if PreFormat.DataCount > 0 then begin
    try
      Result := Format(PreFormat.DataAsString, [v]);
      Exit;
    except
    end;
  end;
  if (v = 0) then begin
    Result := '0.0';
  end else if (Abs(v) < 1e-3) or (Abs(v) > 1e6)  then begin
    Result := FormatFloat('0.000e+00', v);
  end else if Abs(v) < 1 then begin
    Result := FormatFloat('0.0####', v);
  end else if Abs(v) >= 1 then begin
    Result := FormatFloat('######0.0#', v);
  end;
end;

function TQUB_QFS_Rate.prGetExpFormat(v :Double) :String;
begin
  if ExpFormat.DataCount > 0 then begin
    try
      Result := Format(ExpFormat.DataAsString, [v]);
      Exit;
    except
    end;
  end;

  if (Abs(v) = 0) then begin
    Result := '0.0';
  end else if (Abs(v) < 1e-3) or (Abs(v) > 1e3)  then begin
    Result := FormatFloat('0.000e+00', v);
  end else if Abs(v) < 1 then begin
    Result := FormatFloat('0.0##', v);
  end else if Abs(v) >= 1 then begin
    Result := FormatFloat('##0.0##', v);
  end;
end;


//K0
function TQUB_QFS_Rate.prGetK0Format(Index :Integer) :String;
var
  pq :String;
begin
  Result := ' ' + prGetPreExpFormat(Index, prGetK0(Index)) + ' ';
  pq := ' ';
  if P[Index] then begin
    pq := pq + 'p';
  end;
  if Q[Index] then begin
    pq := pq + 'q';
  end;
  if pq <> ' ' then begin
    Result := Result + pq;
  end;
end;

function TQUB_QFS_Rate.prGetK0(Index :Integer) :Double;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := prRecK0.DataArrayValue[Index - 1] else Result := 0;
end;

procedure TQUB_QFS_Rate.prSetK0(Index :Integer; k :Double);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    if prRecK0.DataArrayValue[Index - 1] <> abs(k) then begin
      prRecK0.DataArrayValue[Index - 1]  := abs(k);
      prUpdateCaptions(Index);
      if Assigned(OnEdit) then OnEdit(Self);
    end;
    prSetHint;
  end;
end;

//K1
function TQUB_QFS_Rate.prGetK1Format(Index :Integer) :String;
begin
  Result := prGetExpFormat(prGetK1(Index));
end;

function TQUB_QFS_Rate.prGetK1(Index :Integer) :Double;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := prRecK1.DataArrayValue[Index - 1] else Result := 0;
end;

procedure TQUB_QFS_Rate.prSetK1(Index :Integer; k :Double);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    if prRecK1.DataArrayValue[Index - 1] <> k then begin
      prRecK1.DataArrayValue[Index - 1]  := k;
      prUpdateCaptions(Index);
      if Assigned(OnEdit) then OnEdit(Self);
    end;
    prSetHint;
  end;
end;

//dK0
function TQUB_QFS_Rate.prGetdK0(Index :Integer) :Double;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := prRecdK0.DataArrayValue[Index - 1] else Result := 0;
end;

procedure TQUB_QFS_Rate.prSetdK0(Index :Integer; dk :Double);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    prRecdK0.DataArrayValue[Index - 1] := abs(dk);
  end;
end;

//dK1
function TQUB_QFS_Rate.prGetdK1(Index :Integer) :Double;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := prRecdK1.DataArrayValue[Index - 1] else Result := 0;
end;

procedure TQUB_QFS_Rate.prSetdK1(Index :Integer; dk :Double);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    prRecdK1.DataArrayValue[Index - 1] := abs(dk);
  end;
end;

//P
function TQUB_QFS_Rate.prGetP(Index :Integer) :Boolean;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := Boolean(prRecP.DataArrayValue[Index - 1]) else Result := False;
end;

procedure TQUB_QFS_Rate.prSetP(Index :Integer; p :Boolean);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    if Boolean(prRecP.DataArrayValue[Index - 1]) <> p then begin
      prRecP.DataArrayValue[Index - 1] := Integer(p);
      prUpdateCaptions(Index);
      if Assigned(OnEdit) then OnEdit(Self);
    end;
    prSetHint;
  end;
end;

//Q
function TQUB_QFS_Rate.prGetQ(Index :Integer) :Boolean;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := Boolean(prRecQ.DataArrayValue[Index - 1]) else Result := False;
end;

procedure TQUB_QFS_Rate.prSetQ(Index :Integer; q :Boolean);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    if Boolean(prRecQ.DataArrayValue[Index - 1]) <> q then begin
      prRecQ.DataArrayValue[Index - 1] := Integer(q);
      prUpdateCaptions(Index);
      if Assigned(OnEdit) then OnEdit(Self);
    end;
    prSetHint;
  end;
end;

//PName
function TQUB_QFS_Rate.prGetPName(Index :Integer) :String;
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    while prRecPNames.RecordCount < 2 do begin
      TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_PNAME_NAME,  prRecPNames, False{Permanent}, CHANNELS_RATE_DEF_PNAME, NODE_DATATYPE_STRING,  0{Count});
    end;
    Result := prRecPNames.Records[Index - 1].DataValue;
  end else begin
    Result := '';
  end;
end;

procedure TQUB_QFS_Rate.prSetPName(Index :Integer; p :String);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    while prRecPNames.RecordCount < 2 do begin
      TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_PNAME_NAME,  prRecPNames, False{Permanent}, CHANNELS_RATE_DEF_PNAME, NODE_DATATYPE_STRING,  0{Count});
    end;
    if prRecPNames.Records[Index - 1].DataValue <> p then begin
      prRecPNames.Records[Index - 1].DataValue  := p;
      prUpdateCaptions(Index);
      if Assigned(OnEdit) then OnEdit(Self);
    end;
    prSetHint;
  end;
end;

//QName
function TQUB_QFS_Rate.prGetQName(Index :Integer) :String;
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    while prRecQNames.RecordCount < 2 do begin
      TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_QNAME_NAME,  prRecQNames, False{Permanent}, CHANNELS_RATE_DEF_QNAME, NODE_DATATYPE_STRING,  0{Count});
    end;
    Result := prRecQNames.Records[Index - 1].DataValue;
  end else begin
    Result := '';
  end;
end;

procedure TQUB_QFS_Rate.prSetQName(Index :Integer; q :String);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    while prRecQNames.RecordCount < 2 do begin
      TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_QNAME_NAME,  prRecQNames, False{Permanent}, CHANNELS_RATE_DEF_QNAME, NODE_DATATYPE_STRING,  0{Count});
    end;
    if prRecQNames.Records[Index - 1].DataValue <> q then begin
      prRecQNames.Records[Index - 1].DataValue  := q;
      prUpdateCaptions(Index);
      if Assigned(OnEdit) then OnEdit(Self);
    end;
    prSetHint;
  end;
end;


//PValue
procedure TQUB_QFS_Rate.prSetPValue(Index :Integer; p :TPDouble);
begin
  prRecPValue.DataArrayValue[Index - 1] := Integer(p);
  prUpdateCaptions(Index);
  prSetHint;
end;

function TQUB_QFS_Rate.prGetPValue(Index :Integer) :TPDouble;
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    Result := TPDouble(Integer(prRecPValue.DataArrayValue[Index - 1]));
  end else begin
    Result := nil;
  end;
end;

procedure TQUB_QFS_Rate.prSetQValue(Index :Integer; q :TPDouble);
begin
  prRecQValue.DataArrayValue[Index - 1] := Integer(q);
  prUpdateCaptions(Index);
  prSetHint;
end;

function TQUB_QFS_Rate.prGetQValue(Index :Integer) :TPDouble;
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    Result := TPDouble(Integer(prRecQValue.DataArrayValue[Index - 1]));
  end else begin
    Result := nil;
  end;
end;


//RateFormat
function TQUB_QFS_Rate.prGetRateFormat(Index :Integer) :String;
var
  x        :Variant;
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    while prRecRateFormats.RecordCount < 2 do begin
      TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_FORMAT_NAME,  prRecRateFormats, False{Permanent}, CHANNELS_RATE_DEF_FORMAT, NODE_DATATYPE_STRING,  0{Count});
    end;
    x := prRecRateFormats.Records[Index - 1].DataValue;
    if (varType(x) and varTypeMask) = varString then begin
      Result := x;
    end else begin // "DataValue" turns empty strings into '0'!
      Result := '';
    end;
  end else begin
    Result := '';
  end;
end;

procedure TQUB_QFS_Rate.prSetRateFormat(Index :Integer; x :String);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    while prRecRateFormats.RecordCount < 2 do begin
      TQUB_QFS_Record.CreateAsValue(CHANNELS_RATE_FORMAT_NAME,  prRecRateFormats, False{Permanent}, CHANNELS_RATE_DEF_FORMAT, NODE_DATATYPE_STRING,  0{Count});
    end;
    if prRecRateFormats.Records[Index - 1].DataValue <> x then begin
      prRecRateFormats.Records[Index - 1].DataValue  := x;
      prUpdateCaptions(Index);
      if Assigned(OnEdit) then OnEdit(Self);
    end;
    prSetHint;
  end;
end;


//K0, K1 addresses
function TQUB_QFS_Rate.prGetK0Adr(Index :Integer) :TPDouble;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := @(TPADouble(prRecK0.Data)[Index - 1]){@(prK0[Index])} else Result := nil;
end;

function TQUB_QFS_Rate.prGetK1Adr(Index :Integer) :TPDouble;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := @(TPADouble(prRecK1.Data)[Index - 1]){@(prK1[Index])} else Result := nil;
end;


//Rate expression
(*
function TQUB_QFS_Rate.prGetRateExp(Index :Integer) :TKPL_Expression;
begin
  if ((Index >= 1) and (Index <= 2)) then Result := prRateExp[Index] else Result := nil;
end;

procedure TQUB_QFS_Rate.prSetRateExp(Index :Integer; Exp :TKPL_Expression);
begin
  if ((Index >= 1) and (Index <= 2)) then begin
    if prRateExp[Index] <> nil then begin
      {if prRateExp[Index].OnDelete = OnDeleteRateExp then} prRateExp[Index].OnDelete := nil;
      prRateExp[Index] := nil;
    end;
    prRateExp[Index] := Exp;
    if prRateExp[Index] <> nil then begin
      prRateExp[Index].OnDelete := OnDeleteRateExp;
    end;
  end;
end;


//This event should be called by RateExp when destroyed
procedure TQUB_QFS_Rate.OnDeleteRateExp(Sender :TObject);
begin
  if prRateExp[1] = Sender then prRateExp[1] := nil;
  if prRateExp[2] = Sender then prRateExp[2] := nil;
end;
*)

function TQUB_QFS_Rate.prGetLinesPos(Index :Integer) :TPoint;
begin
  if Index = 1 then Result := prLinesPos[1] else if Index = 2 then Result := prLinesPos[2];
end;


//****************************************************************************//
//                                                                            //
//                          C O N S T R A I N T S                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_Constraints.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    //if NodeName          = CHANNELS_CONSTRAINT_NAME      then begin
    if (NodeName = CHANNELS_CONSTRAINT_FIX_RATE_NAME)       or
       (NodeName = CHANNELS_CONSTRAINT_SCALE_RATE_NAME)     or
       (NodeName = CHANNELS_CONSTRAINT_FIX_EXP_NAME)        or
       (NodeName = CHANNELS_CONSTRAINT_SCALE_EXP_NAME)      or
       (NodeName = CHANNELS_CONSTRAINT_LOOP_BAL_NAME)       or
       (NodeName = CHANNELS_CONSTRAINT_LOOP_IMBAL_NAME)     or
       (NodeName = CHANNELS_CONSTRAINT_LOOP_BAL_EXP_NAME)   or
       (NodeName = CHANNELS_CONSTRAINT_LOOP_IMBAL_EXP_NAME) or
       (NodeName = CHANNELS_CONSTRAINT_GEN_NAME)            then begin
      Result := TQUB_QFS_Constraint.Create(NodeName, Self);
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

procedure TQUB_QFS_Constraints.AddConstraint(Constraint :TQUB_QFS_Constraint; ConstraintIndex :Integer);
begin
  if not AddRecord(Constraint, ConstraintIndex) then if Assigned(OnEdit) then OnEdit(Self);
end;

procedure TQUB_QFS_Constraints.DelConstraint(ConstraintIndex :Integer);
begin
  if not DelRecord(Constraint[ConstraintIndex]) then if Assigned(OnEdit) then OnEdit(Self);
end;

function TQUB_QFS_Constraints.prGetConstraint(s :Integer) :TQUB_QFS_Constraint;
begin
  Result := TQUB_QFS_Constraint(Records[s]);
end;

procedure TQUB_QFS_Constraints.DelState(State, StateCount, RateCount, ExtraKinParCount :Integer);
var
  i, j, s :Integer;
begin
  for i := ConstraintCount - 1 downto 0 do begin
    if Constraint[i].cType = CSTR_GEN then begin
      if Constraint[i].CheckHasState(State, StateCount, RateCount, ExtraKinParCount) then begin
        DelConstraint(i);
      end;
    end else begin
      for j := 0 to Constraint[i].IndexCount - 1 do begin
        s   := Constraint[i].Index[j];
        if s = State then begin
          DelConstraint(i);
          Break;
        end else if s > State then begin
          Constraint[i].Index[j] := s - 1;
        end;
      end;
    end;
  end;
end;

procedure TQUB_QFS_Constraints.AddState;
begin
end;

procedure TQUB_QFS_Constraints.DelRate(Rate :TQUB_QFS_Rate; RateIndex, StateCount, RateCount, ExtraKinParCount :Integer);
var
  i, j :Integer;
begin
  for i := ConstraintCount - 1 downto 0 do begin
    case Constraint[i].cType of
      CSTR_FIX_RATE,
      CSTR_FIX_EXP:
        begin
          if ((Constraint[i].Index[0] = Rate.StateIndex[1]) and (Constraint[i].Index[1] = Rate.StateIndex[2])) or
             ((Constraint[i].Index[0] = Rate.StateIndex[2]) and (Constraint[i].Index[1] = Rate.StateIndex[1]))
          then begin
            DelConstraint(i);
          end;
        end;
      CSTR_SCALE_RATE,
      CSTR_SCALE_EXP:
        begin
          if ((Constraint[i].Index[0] = Rate.StateIndex[1]) and (Constraint[i].Index[1] = Rate.StateIndex[2])) or
             ((Constraint[i].Index[0] = Rate.StateIndex[2]) and (Constraint[i].Index[1] = Rate.StateIndex[1])) or
             ((Constraint[i].Index[2] = Rate.StateIndex[1]) and (Constraint[i].Index[3] = Rate.StateIndex[2])) or
             ((Constraint[i].Index[2] = Rate.StateIndex[2]) and (Constraint[i].Index[3] = Rate.StateIndex[1]))
          then begin
            DelConstraint(i);
          end;
        end;
      CSTR_LOOP_BALANCE,
      CSTR_LOOP_IMBALANCE,
      CSTR_LOOP_BALANCE_EXP,
      CSTR_LOOP_IMBALANCE_EXP:
        begin
          if ((Constraint[i].Index[0] = Rate.StateIndex[1]) and (Constraint[i].Index[Constraint[i].IndexCount - 1] = Rate.StateIndex[2])) or
             ((Constraint[i].Index[0] = Rate.StateIndex[2]) and (Constraint[i].Index[Constraint[i].IndexCount - 1] = Rate.StateIndex[1]))
            then begin
              DelConstraint(i);
          end else begin
            for j := 0 to Constraint[i].IndexCount - 2 do begin
              if ((Constraint[i].Index[j] = Rate.StateIndex[1]) and (Constraint[i].Index[j + 1] = Rate.StateIndex[2])) or
                 ((Constraint[i].Index[j] = Rate.StateIndex[2]) and (Constraint[i].Index[j + 1] = Rate.StateIndex[1]))
              then begin
                DelConstraint(i);
                Break;
              end;
            end;
          end;
        end;
      CSTR_GEN:
        begin
          if Constraint[i].CheckHasRate(RateIndex, StateCount, RateCount, ExtraKinParCount) then begin
            DelConstraint(i);
          end else begin
            Constraint[i].AdjustOnDelRate(RateIndex, StateCount, RateCount, ExtraKinParCount);
          end;
        end;
    end;//case;
  end;
end;

procedure TQUB_QFS_Constraints.AddRate(StateCount, RateCount, ExtraKinParCount :Integer);
var
  i :Integer;
begin
  for i := ConstraintCount - 1 downto 0 do begin
    if Constraint[i].cType = CSTR_GEN then begin
      Constraint[i].AdjustOnAddRate(StateCount, RateCount, ExtraKinParCount);
    end;//case;
  end;
end;

procedure TQUB_QFS_Constraints.DelExtraKineticPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount :Integer);
var
  i :Integer;
begin
  for i := ConstraintCount - 1 downto 0 do begin
    if Constraint[i].cType = CSTR_GEN then begin
      if Constraint[i].CheckHasExtraKinPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount) then begin
      DelConstraint(i);
      end else begin
        Constraint[i].AdjustOnDelExtraKinPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount);
      end;
    end;//case;
  end;
end;

procedure TQUB_QFS_Constraints.AddExtraKineticPar(StateCount, RateCount, ExtraKinParCount :Integer);
var
  i :Integer;
begin
  for i := ConstraintCount - 1 downto 0 do begin
    if Constraint[i].cType = CSTR_GEN then begin
      Constraint[i].AdjustOnAddExtraKinPar(StateCount, RateCount, ExtraKinParCount);
    end;//case;
  end;
end;

//******************************************************************************
procedure TQUB_QFS_Constraints.Copy(c :TQUB_QFS_Constraints);
var
  i, j :Integer;
  cc   :TQUB_QFS_Constraint;
begin
  Reset;
  for i := 0 to c.ConstraintCount - 1 do begin
    cc := TQUB_QFS_Constraint.Create(c.Constraint[i].Name, nil);
    cc.HasValue := c.Constraint[i].HasValue;
    cc.Value    := c.Constraint[i].Value;
    AddConstraint(cc, ConstraintCount);
    for j := 0 to c[i].IndexCount - 1 do begin
      cc.Index[j] := c[i].Index[j];
    end;
    for j := 0 to c[i].CoeffCount - 1 do begin
      cc.Coeff[j] := c[i].Coeff[j];
    end;
  end;
end;



//****************************************************************************//
//                                                                            //
//                             C O N S T R A I N T                            //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_Constraint.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited Create(Name, Parent);

  Flags      := NODE_FLAG_PRELOAD;
  DataCount  := 0;
  DataType   := NODE_DATATYPE_LONGINT;

  //Child records
  prRecHasValue := TQUB_QFS_Record.CreateAsValue(CHANNELS_CONSTRAINT_HASVALUE_NAME, Self, True{Permanent}, 0, NODE_DATATYPE_BYTE,   1{Count});
  prRecValue    := TQUB_QFS_Record.CreateAsValue(CHANNELS_CONSTRAINT_VALUE_NAME,    Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE, 1{Count});

  prRecCoeffs   := TQUB_QFS_Record.Create(CHANNELS_CONSTRAINT_COEFFS_NAME, Self);
  prRecCoeffs.Permanent := True;
  prRecCoeffs.Flags     := NODE_FLAG_PRELOAD;
  prRecCoeffs.DataCount := 0;
  prRecCoeffs.DataType  := NODE_DATATYPE_DOUBLE;

  prRecHasValue.DataValue := 0;
  prRecValue.DataValue    := 1;

end;

function TQUB_QFS_Constraint.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = CHANNELS_CONSTRAINT_HASVALUE_NAME then begin
      Result := prRecHasValue;
    end else if NodeName = CHANNELS_CONSTRAINT_VALUE_NAME    then begin
      Result := prRecValue;
    end else if NodeName = CHANNELS_CONSTRAINT_COEFFS_NAME   then begin
      Result := prRecCoeffs;
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

function TQUB_QFS_Constraint.CheckHasState(State, StateCount, RateCount, ExtraKinParCount :Integer) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_Constraint.CheckHasRate(Rate, StateCount, RateCount, ExtraKinParCount :Integer) :Boolean;
begin
  Result := False;
  if (Coeff[Rate*4 + 0] <> 0) or
     (Coeff[Rate*4 + 1] <> 0) or
     (Coeff[Rate*4 + 2] <> 0) or
     (Coeff[Rate*4 + 3] <> 0) then begin
    Result := True;
  end;
end;

function TQUB_QFS_Constraint.CheckHasExtraKinPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount :Integer) :Boolean;
begin
  Result := False;
  if (Coeff[RateCount*4 + ExtraKinPar] <> 0) then begin
    Result := True;
  end;
end;

procedure TQUB_QFS_Constraint.AdjustOnDelRate(Rate, StateCount, RateCount, ExtraKinParCount :Integer);
var
  i :Integer;
begin
  for i := Rate*4 to RateCount*4 + ExtraKinParCount - 1 - 4 do begin
    Coeff[i] := Coeff[i + 4];
  end;
  prRecCoeffs.DataCount := prRecCoeffs.DataCount - 4;
end;

procedure TQUB_QFS_Constraint.AdjustOnAddRate(StateCount, RateCount, ExtraKinParCount :Integer);
var
  i :Integer;
begin
  prRecCoeffs.DataCount := prRecCoeffs.DataCount + 4;
  //Move the ExtraKineticParams one rate up
  for i := RateCount*4 + 4 + ExtraKinParCount - 1 downto RateCount*4 + 4 + 0 do begin
    Coeff[i] := Coeff[i - 4];
  end;
  //Clear the added rate coeffs
  Coeff[RateCount*4 + 0] := 0;
  Coeff[RateCount*4 + 1] := 0;
  Coeff[RateCount*4 + 2] := 0;
  Coeff[RateCount*4 + 3] := 0;
end;

procedure TQUB_QFS_Constraint.AdjustOnDelExtraKinPar(ExtraKinPar, StateCount, RateCount, ExtraKinParCount :Integer);
var
  i :Integer;
begin
  for i := RateCount*4 + ExtraKinPar to RateCount*4 + ExtraKinParCount - 1 - 1 do begin
    Coeff[i] := Coeff[i + 1];
  end;
  prRecCoeffs.DataCount := prRecCoeffs.DataCount - 1;
end;

procedure TQUB_QFS_Constraint.AdjustOnAddExtraKinPar(StateCount, RateCount, ExtraKinParCount :Integer);
begin
  prRecCoeffs.DataCount := prRecCoeffs.DataCount + 1;
  Coeff[RateCount*4 + ExtraKinParCount] := 0;
end;

function TQUB_QFS_Constraint.prGetCType :TQUB_QFS_QMF_ConstraintType;
var
  n :String;
begin
  n := Name;
  if          n = CHANNELS_CONSTRAINT_FIX_RATE_NAME        then begin
    Result := CSTR_FIX_RATE;
  end else if n = CHANNELS_CONSTRAINT_SCALE_RATE_NAME      then begin
    Result := CSTR_SCALE_RATE;
  end else if n = CHANNELS_CONSTRAINT_FIX_EXP_NAME         then begin
    Result := CSTR_FIX_EXP;
  end else if n = CHANNELS_CONSTRAINT_SCALE_EXP_NAME       then begin
    Result := CSTR_SCALE_EXP;
  end else if n = CHANNELS_CONSTRAINT_LOOP_BAL_NAME        then begin
    Result := CSTR_LOOP_BALANCE;
  end else if n = CHANNELS_CONSTRAINT_LOOP_IMBAL_NAME      then begin
    Result := CSTR_LOOP_IMBALANCE;
  end else if n = CHANNELS_CONSTRAINT_LOOP_BAL_EXP_NAME    then begin
    Result := CSTR_LOOP_BALANCE_EXP;
  end else if n = CHANNELS_CONSTRAINT_LOOP_IMBAL_EXP_NAME  then begin
    Result := CSTR_LOOP_IMBALANCE_EXP;
  end else if n = CHANNELS_CONSTRAINT_GEN_NAME             then begin
    Result := CSTR_GEN;
  end else begin
    Result := CSTR_UNKNOWN;
  end;
end;

function TQUB_QFS_Constraint.prGetIndex(Index :Integer) :Integer;
begin
  Result := Integer(DataArrayValue[Index]);
end;

procedure TQUB_QFS_Constraint.prSetIndex(Index, Value :Integer);
begin
  DataArrayValue[Index] := Value;
end;

function TQUB_QFS_Constraint.prGetCoeffCount :Integer;
begin
  Result := prRecCoeffs.DataCount;
end;

function TQUB_QFS_Constraint.prGetCoeffPointer: TPADouble;
begin
  Result := TPADouble(prRecCoeffs.Data);
end;

function TQUB_QFS_Constraint.prGetCoeff(Index :Integer) :Double;
begin
  Result := Double(prRecCoeffs.DataArrayValue[Index]);
end;

procedure TQUB_QFS_Constraint.prSetCoeff(Index :Integer; Value :Double);
begin
  prRecCoeffs.DataArrayValue[Index] := Value;
end;

procedure TQUB_QFS_Constraint.prSetHasValue(v :Boolean);
begin
  prRecHasValue.DataValue := Integer(v);
end;

function TQUB_QFS_Constraint.prGetHasValue :Boolean;
begin
  Result := Boolean(prRecHasValue.DataValue);
end;

procedure TQUB_QFS_Constraint.prSetValue(v :Double);
begin
  prRecValue.DataValue := v;
end;

function TQUB_QFS_Constraint.prGetValue :Double;
begin
  Result := prRecValue.DataValue;
end;



//****************************************************************************//
//                                                                            //
//                 A M P L I T U D E     C O N S T R A I N T S                //
//                                                                            //
//****************************************************************************//
procedure TQUB_QFS_ConstraintsAmpVar.AddConstraint(Constraint :TQUB_QFS_ConstraintAmpVar; ConstraintIndex :Integer);
begin
  if not AddRecord(Constraint, ConstraintIndex) then if Assigned(OnEdit) then OnEdit(Self);
end;

procedure TQUB_QFS_ConstraintsAmpVar.DelConstraint(ConstraintIndex :Integer);
begin
  if not DelRecord(Constraint[ConstraintIndex]) then if Assigned(OnEdit) then OnEdit(Self);
end;

function TQUB_QFS_ConstraintsAmpVar.prGetConstraint(s :Integer) :TQUB_QFS_ConstraintAmpVar;
begin
  Result := TQUB_QFS_ConstraintAmpVar(Records[s]);
end;

procedure TQUB_QFS_ConstraintsAmpVar.DelClass(Cls :Integer);//Cls is zero based !!!
var
  i, j, c :Integer;
begin
  for i := ConstraintCount - 1 downto 0 do begin
    for j := 0 to Constraint[i].IndexCount - 1 do begin
      c   := Constraint[i].Index[j];
      if c = Cls then begin
        DelConstraint(i);
        Break;
      end else if c > Cls then begin
        Constraint[i].Index[j] := c - 1;
      end;
    end;
  end;
end;

procedure TQUB_QFS_ConstraintsAmpVar.Copy(c :TQUB_QFS_ConstraintsAmpVar);
var
  i, j :Integer;
  cc   :TQUB_QFS_ConstraintAmpVar;
begin
  Reset;
  for i := 0 to c.ConstraintCount - 1 do begin
    cc := TQUB_QFS_ConstraintAmpVar.Create(c.Constraint[i].Name, nil);
    AddConstraint(cc, ConstraintCount);
    for j := 0 to c[i].IndexCount - 1 do begin
      cc.Index[j] := c[i].Index[j];
    end;
  end;
end;



//****************************************************************************//
//                                                                            //
//                  A M P L I T U D E     C O N S T R A I N T                 //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_ConstraintAmpVar.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited Create(Name, Parent);

  Flags      := NODE_FLAG_PRELOAD;
  DataCount  := 0;
  DataType   := NODE_DATATYPE_LONGINT;

end;

function TQUB_QFS_ConstraintAmpVar.prGetCType :TQUB_QFS_QMF_ConstraintAmpVarType;
var
  n :String;
begin
  n := Name;
  if          n = CHANNELS_CONSTRAINT_FIX_AMP_NAME             then begin
    Result := CSTR_FIX_AMP;
  end else if n = CHANNELS_CONSTRAINT_FIX_EXC_AMP_NAME         then begin
    Result := CSTR_FIX_EXC_AMP;
  end else if n = CHANNELS_CONSTRAINT_FIX_DIFF_EXC_AMP_NAME    then begin
    Result := CSTR_FIX_DIFF_EXC_AMP;
  end else if n = CHANNELS_CONSTRAINT_SCALE_EXC_AMP_NAME       then begin
    Result := CSTR_SCALE_EXC_AMP;
  end else if n = CHANNELS_CONSTRAINT_SCALE_DIFF_EXC_AMP_NAME  then begin
    Result := CSTR_SCALE_DIFF_EXC_AMP;
  end else if n = CHANNELS_CONSTRAINT_FIX_VAR_NAME             then begin
    Result := CSTR_FIX_VAR;
  end else if n = CHANNELS_CONSTRAINT_FIX_EXC_VAR_NAME         then begin
    Result := CSTR_FIX_EXC_VAR;
  end else if n = CHANNELS_CONSTRAINT_FIX_DIFF_EXC_VAR_NAME    then begin
    Result := CSTR_FIX_DIFF_EXC_VAR;
  end else if n = CHANNELS_CONSTRAINT_SCALE_EXC_VAR_NAME       then begin
    Result := CSTR_SCALE_EXC_VAR;
  end else if n = CHANNELS_CONSTRAINT_SCALE_DIFF_EXC_VAR_NAME  then begin
    Result := CSTR_SCALE_DIFF_EXC_VAR;
  end else begin
    Result := CSTR_UNKNOWN_AMPVAR;
  end;
end;

function TQUB_QFS_ConstraintAmpVar.prGetIndex(Index :Integer) :Integer;
begin
  Result := Integer(DataArrayValue[Index]);
end;

procedure TQUB_QFS_ConstraintAmpVar.prSetIndex(Index, Value :Integer);
begin
  DataArrayValue[Index] := Value;
end;


//******************************************************************************
function TQUB_QFS_ExtraKineticPars.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    //if NodeName          = CHANNELS_EXTRAKINETICPAR_NAME      then begin
      Result := TQUB_QFS_ExtraKineticPar.Create(NodeName, Self);
    //end else begin
    //  Result := inherited NewChild(ReadStream);
    //end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

procedure TQUB_QFS_ExtraKineticPars.AddExtraKineticPar(Par :TQUB_QFS_ExtraKineticPar; ParIndex :Integer);
begin
  if not AddRecord(Par, ParIndex) then if Assigned(OnEdit) then OnEdit(Self);
end;

procedure TQUB_QFS_ExtraKineticPars.DelExtraKineticPar(p :Integer);
begin
  if not DelRecord(ExtraKineticPar[p]) then if Assigned(OnEdit) then OnEdit(Self);
end;

{
function TQUB_QFS_ExtraKineticPars.IndexOf(Par :TQUB_QFS_ExtraKineticPar) :Integer;
begin
end;
}

function TQUB_QFS_ExtraKineticPars.prGetExtraKineticPar(p :Integer) :TQUB_QFS_ExtraKineticPar;
begin
  Result := TQUB_QFS_ExtraKineticPar(Records[p]);
end;

function TQUB_QFS_ExtraKineticPars.prGetExtraKineticParByName(n :String) :TQUB_QFS_ExtraKineticPar;
var
  i :Integer;
begin
  Result := nil;
  for i := 0 to ExtraKineticParCount - 1 do begin
    if TQUB_QFS_ExtraKineticPar(Records[i]).Name = n then begin
      Result := TQUB_QFS_ExtraKineticPar(Records[i]);
      Exit;
    end;
  end;
end;

procedure TQUB_QFS_ExtraKineticPars.Copy(p :TQUB_QFS_ExtraKineticPars);
var
  ekc :TQUB_QFS_ExtraKineticPar;
  i   :Integer;
begin
  Reset;
  for i := 0 to p.ExtraKineticParCount - 1 do begin
    ekc := TQUB_QFS_ExtraKineticPar.Create(p.ExtraKineticPar[i].Name, nil);
    ekc.Flags    := p.ExtraKineticPar[i].Flags;
    ekc.DataType := p.ExtraKineticPar[i].DataType;
    ekc.Value    := p.ExtraKineticPar[i].Value;
    AddExtraKineticPar(ekc, ExtraKineticParCount);
  end;
end;


//******************************************************************************
constructor TQUB_QFS_ExtraKineticPar.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited Create(Name, Parent);

  Flags      := NODE_FLAG_PRELOAD;
  DataType   := NODE_DATATYPE_DOUBLE;
  Value      := 0;

  //Child records
  prRecEKPType           := TQUB_QFS_Record.CreateAsValue(CHANNELS_EXTRAKINETICPAR_TYPE_NAME, Self, True{Permanent}, 0, NODE_DATATYPE_LONGINT, 1{Count});
  prRecEKPType.DataValue := EKP_TYPE_LOG;

end;

function TQUB_QFS_ExtraKineticPar.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = CHANNELS_EXTRAKINETICPAR_TYPE_NAME then begin
      Result := prRecEKPType;
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

procedure TQUB_QFS_ExtraKineticPar.prSetValue(Value :Double);
begin
  DataValue := Value;
end;

function TQUB_QFS_ExtraKineticPar.prGetValue :Double;
begin
  Result := DataValue;
end;

procedure TQUB_QFS_ExtraKineticPar.prSetEKPType(t :TQUB_QFS_QMF_ExtraKineticParType);
begin
  prRecEKPType.DataValue := Integer(t);
end;

function TQUB_QFS_ExtraKineticPar.prGetEKPType :TQUB_QFS_QMF_ExtraKineticParType;
begin
  Result := prRecEKPType.DataValue;
end;


//****************************************************************************//
//                                                                            //
//                                 P A N E L S                                //
//                                                                            //
//****************************************************************************//
destructor TQUB_QFS_StatePanels.Destroy;
begin
  try
    Clear;
  finally
    inherited;
  end;
end;

procedure TQUB_QFS_StatePanels.Clear;
var
  i :Integer;
begin
  try
    for i := 0 to Count - 1 do begin
      TQUB_QFS_StatePanel(List[i]).Free;
    end;
  finally
    inherited;
  end;
end;

function TQUB_QFS_StatePanels.prGetPanel(Index :Integer) :TQUB_QFS_StatePanel;
begin
  if ((Index >= 0) and (Index <= Count - 1)) then Result := TQUB_QFS_StatePanel(List[Index]) else Result := nil;
end;

function TQUB_QFS_StatePanels.NewPanel :TQUB_QFS_StatePanel;
begin
  if Count > 0 then begin
    Result := Panel[0];
    Remove(Result);
  end else begin
    Result := TQUB_QFS_StatePanel.Create(nil);
  end;
end;

procedure TQUB_QFS_StatePanels.DelPanel(p :TQUB_QFS_StatePanel);
begin
  Add(p);
  p.Parent  := nil;
end;

//******************************************************************************
destructor TQUB_QFS_RatePanels.Destroy;
begin
  try
    Clear;
  finally
    inherited;
  end;
end;

procedure TQUB_QFS_RatePanels.Clear;
var
  i :Integer;
begin
  try
    for i := 0 to Count - 1 do begin
      TQUB_QFS_RatePanel(List[i]).Free;
    end;
  finally
    inherited;
  end;
end;

function TQUB_QFS_RatePanels.prGetPanel(Index :Integer) :TQUB_QFS_RatePanel;
begin
  if ((Index >= 0) and (Index <= Count - 1)) then Result := TQUB_QFS_RatePanel(List[Index]) else Result := nil;
end;

function TQUB_QFS_RatePanels.NewPanel :TQUB_QFS_RatePanel;
begin
  if Count > 0 then begin
    Result             := Panel[0];
    Remove(Result);
  end else begin
    Result             := TQUB_QFS_RatePanel.Create;
    //Result.Visible     := False;
    //Result.ShowHint    := True;
    //Result.Transparent := True;
    //Result.Autosize    := True;//}False;
    //Result.Alignment   := taCenter;
    //Result.Layout      := tlTop;//tlCenter;
  end;
end;

procedure TQUB_QFS_RatePanels.DelPanel(p :TQUB_QFS_RatePanel);
begin
  Add(p);
  //p.Visible := False;
  //p.Parent  := nil;
end;


//******************************************************************************
constructor TQUB_QFS_StatePanel.Create(AOwner :TComponent);
begin
  inherited;
  Visible        := False;
  BevelWidth     := 2;
  ShowHint       := True;
  m_iClass	 := 1;
end;

//----- Draw one state box.
//This was move out of class to be accessed by rate properties form. 11/2004 {JB}
procedure DrawState(C :TCanvas; rctBox :TRect; iClass :Integer; cCaption :String);
var
  rctText  :TRect;
  iX, iY   :Integer;
begin
  
  rctText := rctBox;
  //----- Fill entire box
  C.Pen.Color := QUB_GetClassBackColor( iClass );
  C.Brush.Color := QUB_GetClassColor( iClass );
  C.FillRect(rctBox);

  //----- Fill top half
  if iClass >= 10 then begin
    rctBox.Bottom := (rctBox.Bottom * 382 + rctBox.Top * 618 ) Div 1000 ;
		rctText.Top := rctBox.Bottom ;
		C.Brush.Color := QUB_GetClassColor( iClass div 10);
		C.FillRect(rctBox);
		C.MoveTo(rctBox.Left, rctBox.Bottom);
		C.LineTo(rctBox.Right, rctBox.Bottom);
	end;

	//----- Fill top left quarter
	if iClass >= 100 Then Begin
		rctBox.Right := (rctBox.Right * 382 + rctBox.Left * 618 ) Div 1000 ;
		C.Brush.Color := QUB_GetClassColor(iClass div 100);
		C.FillRect(rctBox);
		C.MoveTo(rctBox.Right, rctBox.Top);
		C.LineTo(rctBox.Right, rctBox.Bottom);
	end;

	//---- Draw text label
	C.Font.Color := QUB_GetClassBackColor( iClass );
	C.Brush.Style := bsClear;
	iY := (rctText.Top+rctText.Bottom-C.TextHeight( cCaption )) DIV 2;
	iX := (rctText.Left+rctText.Right-C.TextWidth( cCaption )) DIV 2;

	//----- Under Construction... Subscripting
//	if Copy(cCaption,1,1) = 'O' or Copy(cCaption,1,1) = 'C' Then Begin
//			cSub := Copy(cCaption,1,1);
//			C.TextRect( rctText, iX, iY, cSub );
//			C.TextRect( rctText, iX+C.TextWidth(cSub), iY+2, Copy(cCaption,2,Length(cCaption)-1) );
//  end else begin
	C.TextRect( rctText, iX, iY, cCaption );
//		end;
end;

procedure TQUB_QFS_StatePanel.Paint;
var
  rctText, rctBox :TRect;
  iX, iY, iBevel  :Integer;
  //cSub            :String;
begin
  inherited;
  // Note : 618 / 382 used for golden rectangle ratio
  // Change 02-2004 to directly draw boxes instead of 4 panels display.  {JB}
  //----- Client rectangle less bevel area
  if BevelOuter = bvNone then iBevel := 0 else iBevel := BevelWidth;
  rctBox          := GetClientRect;
  rctBox.Top      := rctBox.Top + iBevel;
  rctBox.Bottom   := rctBox.Bottom - iBevel;
  rctBox.Left     := rctBox.Left + iBevel;
  rctBox.Right    := rctBox.Right - iBevel;

  rctText := rctBox;

  //----- Fill entire box
  Canvas.Pen.Color := QUB_GetClassBackColor( m_iClass );
  Canvas.Brush.Color := QUB_GetClassColor( m_iClass );
  Canvas.FillRect(rctBox);

  //----- Fill top half
  if m_iClass >= 10 then begin
    rctBox.Bottom := (rctBox.Bottom * 382 + rctBox.Top * 618 ) div 1000 ;
    rctText.Top := rctBox.Bottom ;
    Canvas.Brush.Color := QUB_GetClassColor( m_iClass div 10);
    Canvas.FillRect(rctBox);
    Canvas.MoveTo(rctBox.Left, rctBox.Bottom);
    Canvas.LineTo(rctBox.Right, rctBox.Bottom);
  end;

  //----- Fill top left quarter
  if m_iClass >= 100 then begin
    rctBox.Right := (rctBox.Right * 382 + rctBox.Left * 618 ) div 1000 ;
    Canvas.Brush.Color := QUB_GetClassColor(m_iClass div 100);
    Canvas.FillRect(rctBox);
    Canvas.MoveTo(rctBox.Right, rctBox.Top);
    Canvas.LineTo(rctBox.Right, rctBox.Bottom);
  end;

  //---- Draw text label
  Canvas.Font.Color := QUB_GetClassBackColor( m_iClass );
  Canvas.Brush.Style := bsClear;
  iY := (rctText.Top+rctText.Bottom-Canvas.TextHeight( m_cCaption )) DIV 2;
  iX := (rctText.Left+rctText.Right-Canvas.TextWidth( m_cCaption )) DIV 2;

  //----- Under Construction... Subscripting
  //if Copy(m_cCaption,1,1) = 'O' or Copy(m_cCaption,1,1) = 'C' then begin
  //  cSub := Copy(m_cCaption,1,1);
  //  Canvas.TextRect( rctText, iX, iY, cSub );
  //  Canvas.TextRect( rctText, iX+Canvas.TextWidth(cSub), iY+2, Copy(m_cCaption,2,Length(m_cCaption)-1) );
  //end else begin
    Canvas.TextRect( rctText, iX, iY, m_cCaption );
  //end;

end;

//Check: John's new Paint
(*
procedure TQUB_QFS_StatePanel.Paint;
var
  rctBox :TRect;
  iBevel :Integer;
  //cSub :String;
begin
  inherited;
  // Note : 618 / 382 used for golden rectangle ratio
  // Change 02-2004 to directly draw boxes instead of 4 panels display.  {JB}
  //----- Client rectangle less bevel area
  if BevelOuter = bvNone then iBevel := 0 else iBevel := BevelWidth;
  rctBox          := GetClientRect;
  rctBox.Top      := rctBox.Top + iBevel;
  rctBox.Bottom   := rctBox.Bottom - iBevel;
  rctBox.Left     := rctBox.Left + iBevel;
  rctBox.Right    := rctBox.Right - iBevel;

  DrawState( Canvas, rctBox, m_iClass, m_cCaption );

end;
*)

//todo: Use overloaded function. LM
//----- Set state caption to specified integer
procedure TQUB_QFS_StatePanel.SetCaptionI(i : Integer);
begin
  m_cCaption := IntToStr(i+1);
end;

//----- Set State caption to specified string, generally of form O1, O2, ... C1, C2, ..
procedure TQUB_QFS_StatePanel.SetCaptionC(c : String);
begin
  m_cCaption := c;
end;


procedure TQUB_QFS_StatePanel.prSetClass(x :Integer);
begin
  m_iClass := x;
  Invalidate;
end;

procedure TQUB_QFS_StatePanel.Activate;
begin
  BevelOuter := bvLowered;
  BringToFront;
end;

procedure TQUB_QFS_StatePanel.DeActivate;
begin
  //BevelOuter := bvNone;
  BevelOuter := bvRaised;
end;

constructor TQUB_QFS_RatePanel.Create;
begin
  inherited;
  FontHeight  := QUB_QFS_QMF_RATE_FONTHEIGHT;
  FontWeight  := 700;
  ///Visible   := False;
  ///ShowHint  := True;
  ///Alignment := taCenter;
  ///AutoSize  := True;
end;

procedure TQUB_QFS_RatePanel.Activate;
begin
  //FontWeight := 700; // .Style := [fsBold];
end;

procedure TQUB_QFS_RatePanel.DeActivate;
begin
  //FontWeight := 400; // .Style := [];
end;
(*
procedure TQUB_QFS_RatePanel.PaintTo(NewCanvas :TCanvas);
const
  Alignments :array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  WordWraps  :array[Boolean]    of Word = (0, DT_WORDBREAK);
var
  OldDC             :HDC;
  Rect, CalcRect    :TRect;
  DrawStyle         :Longint;
begin

  with Canvas do begin

    OldDC := Canvas.Handle;

    try

      Canvas.Lock;
      Canvas.Handle := NewCanvas.Handle;
      Brush.Style := bsClear;
      Rect := BoundsRect;
      DrawStyle := DT_EXPANDTABS or WordWraps[WordWrap] or Alignments[Alignment];
      if Layout <> tlTop then begin
        CalcRect := Rect;
        DoDrawText(CalcRect, DrawStyle or DT_CALCRECT);
        if Layout = tlBottom then begin
          OffsetRect(Rect, 0, Height - CalcRect.Bottom + Top);
        end else begin
          OffsetRect(Rect, 0, (Height - CalcRect.Bottom + Top) div 2);
        end;
      end;
      DoDrawText(Rect, DrawStyle);

    finally

      Canvas.Handle := OldDC;
      Canvas.UnLock;

    end;

  end;
end;
 *)

//****************************************************************************//
//                                                                            //
//                                   L I N E                                  //
//                                                                            //
//****************************************************************************//
class procedure TQUB_QFS_Line.Draw(Canvas :TCanvas; x1, y1, x2, y2 :Integer);
begin
  Canvas.MoveTo(x1, y1);
  Canvas.LineTo(x2, y2);
end;

procedure TQUB_QFS_Line.SetBounds(Left, Top, Width, Height: Integer);
begin
  Self.Left := Left;
  Self.Top := Top;
  Self.Width := Width;
  Self.Height := Height;
end;


//******************************************************************************


constructor TRatePanelPosList.Create;
begin
  prSize := 0;
end;

destructor TRatePanelPosList.Destroy;
begin
  Size := 0;
  inherited;
end;

function TRatePanelPosList.prGetPanel(i :Integer): TQUB_QFS_RatePanel;
begin
  Result := prPanels[i];
end;

procedure TRatePanelPosList.prSetPanel(i :Integer; x :TQUB_QFS_RatePanel);
begin
  prPanels[i] := x;
end;

function TRatePanelPosList.prGetRect(i :Integer): TRotRect;
begin
  Result := prRects[i];
end;

procedure TRatePanelPosList.prSetSize(x :Integer);
var
  i         :Integer;
begin
  for i := x to prSize - 1 do begin
    prRects[i].Free;
  end;
  SetLength(prRects, x);
  SetLength(prPanels, x);
  for i := prSize to x - 1 do begin
    prRects[i] := TRotRect.Create(0, 0, 1, 1);
  end;
  prSize := x;
end;


//******************************************************************************
function QUB_GetClassColor(i :Integer): TColor;
begin
  Result := QUB_QFS_QMF_CLASS_COLORS[ i mod QUB_QFS_QMF_DEF_NCLASS ];
end;

function QUB_GetClassBackColor(i :Integer): TColor;
begin
  Result := QUB_QFS_QMF_BACK_COLORS[ i mod QUB_QFS_QMF_DEF_NCLASS ];
end;


//******************************************************************************
end.
