{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// specialized data segments for QUB_QFS_QDF when it's representing a DWT (or EVL) file opened from disk.
// they recreate data from the idealization rather than reading samples from disk.
// this is necessary to compute e.g. a spectrum histogram from a dwt file.


unit QUB_QFS_DWT_Segments;


interface


uses


SysUtils, Classes,

Alloc, AllocTypes,

QUB_QFS, QUB_QFS_Segments,
QUB_Idealization;


type


//Forward class declarations ***************************************************


TQUB_QFS_DWT_Segment                           = class;
TQUB_QFS_DWT_Channels                          = class;


//Class definitions


TQUB_QFS_DWT_Segment                           = class(TQUB_QFS_Segment)
protected


  prChannels                                   :TQUB_QFS_DWT_Channels;


  function prGetChannels :TQUB_QFS_Channels; override;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;


end;


TQUB_QFS_DWT_Channels                          = class(TQUB_QFS_Channels)
protected

  prScaling                                    :Double;

  prIdl                                        :TQUB_IdlSegments;
  prOffset                                     :Integer;

public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;


  function LoadData(First, Last :Integer) :Boolean; override;
  function LoadDataInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean; override;

  function LoadSelections(First, Last :Integer) :Boolean; override;
  function LoadSelectionsInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean; override;

  property Scaling    :Double                 read prScaling    write prScaling;

  property Idl        :TQUB_IdlSegments       read prIdl        write prIdl;
  property Offset     :Integer                read prOffset     write prOffset;

end;


implementation


//******************************************************************************
constructor TQUB_QFS_DWT_Segment.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited Create(Name, Parent);

  inherited Channels.Free;

  prChannels  := TQUB_QFS_DWT_Channels.Create(QUB_QFS_QDF_CHANNELS_NAME, Self);

end;

function TQUB_QFS_DWT_Segment.prGetChannels :TQUB_QFS_Channels;
begin
  Result := prChannels;
end;


//******************************************************************************
constructor TQUB_QFS_DWT_Channels.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited;

  CanEdit := False;

end;

function TQUB_QFS_DWT_Channels.LoadData(First, Last :Integer) :Boolean;
begin
  Result := True;
end;

function TQUB_QFS_DWT_Channels.LoadDataInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean;
var
  IntBuffer     :TPASmallInt;
  IdlIter       :TQUB_IdlSegmentsIter;
  F, L, C       :Integer;
  A             :Double;
  Amp           :SmallInt;
  i             :Integer;
begin
  Result := False;

  IntBuffer     := TPASmallInt(Buffer);

  IdlIter := nil;

  try

    IdlIter    := Idl.FindIter(First + Offset);
    while 0 <> IdlIter.Valid do begin
      IdlIter.Read(@F, @L, @C, @A);
      IdlIter.Next;

      F := F - Offset;
      L := L - Offset;

      if F > Last  then Exit;
      if F < First then F := First;
      if L > Last  then L := Last;

      if C < 0 then begin
        Amp := Round(-1.0 * Scaling);
      end else begin
        Amp := Round(A * Scaling);
      end;

      for i := F - First to L - First do begin
        IntBuffer[i] := Amp;
      end;
    end;

  finally

    if IdlIter <> nil then IdlIter.Free;

  end;
end;

function TQUB_QFS_DWT_Channels.LoadSelections(First, Last :Integer) :Boolean;
begin
  Result := True;
end;

function TQUB_QFS_DWT_Channels.LoadSelectionsInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean;
begin
  Result := LoadDataInBuffer(Buffer, First, Last); // ??
  // FillChar(Buffer^, DataSize * (Last - First + 1), 0);
end;


//******************************************************************************
end.
