{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// apparently a file opened by QFS for editing has two files associated.
// This is part of the magic of reading from one but writing changes to the other.


unit QUB_QFS_Edit;


interface


uses


SysUtils, Classes, Dialogs,

AllocTypes, Alloc,

QUB_Global,
QUB_QFS,
QUB_Node;


type


//Forward class declarations ***************************************************


TQUB_QFS_Edit                                  = class;

TQUB_QFS_EditOperationType                     = (OPER_TYPE_BASE,

                                                  OPER_TYPE_ADD_SEGMENT,   OPER_TYPE_DEL_SEGMENT,

                                                  OPER_TYPE_ADD_SELECTION, OPER_TYPE_DEL_SELECTION,

                                                  OPER_TYPE_INS_ISEG,      OPER_TYPE_INS_INTO_ISEG,
                                                  OPER_TYPE_DEL_ISEG,      OPER_TYPE_DEL_ISEL,

                                                  OPER_TYPE_BREAK
                                                  );

TQUB_QFS_EditOperations                        = class;
TQUB_QFS_EditSelections                        = class;

TQUB_QFS_EditOperation                         = class;


TQUB_QFS_PEditSelection                        = ^TQUB_QFS_EditSelection;
TQUB_QFS_EditSelection                         = record
  DataCount                                    :Integer;
  FromEditStream                               :Boolean;
  StreamPos                                    :Integer;
end;


//OperationTypes ***************************************************************
TQUB_QFS_PEditOperationBase                    = ^TQUB_QFS_EditOperationBase;
TQUB_QFS_EditOperationBase                     = record
  OperType                                     :TQUB_QFS_EditOperationType;
end;

//seg/sel of data points *******************************************************
TQUB_QFS_PEditOperationDelSegment              = ^TQUB_QFS_EditOperationDelSegment;
TQUB_QFS_EditOperationDelSegment               = record
  OperType                                     :TQUB_QFS_EditOperationType;
  Segment                                      :TQUB_QFS_Record;
  SegmentIndex                                 :Integer;
end;

TQUB_QFS_PEditOperationAddSegment              = ^TQUB_QFS_EditOperationAddSegment;
TQUB_QFS_EditOperationAddSegment               = record
  OperType                                     :TQUB_QFS_EditOperationType;
  Segment                                      :TQUB_QFS_Record;
  SegmentIndex                                 :Integer;
end;

TQUB_QFS_PEditOperationDelSelection            = ^TQUB_QFS_EditOperationDelSelection;
TQUB_QFS_EditOperationDelSelection             = record
  OperType                                     :TQUB_QFS_EditOperationType;
  Selection                                    :TQUB_QFS_PEditSelection;
  SelectionIndex                               :Integer;
  EditRecord                                   :TQUB_QFS_Edit;
end;

TQUB_QFS_PEditOperationAddSelection            = ^TQUB_QFS_EditOperationAddSelection;
TQUB_QFS_EditOperationAddSelection             = record
  OperType                                     :TQUB_QFS_EditOperationType;
  Selection                                    :TQUB_QFS_PEditSelection;
  SelectionIndex                               :Integer;
  EditRecord                                   :TQUB_QFS_Edit;
end;


//segmentation/meta ************************************************************

// generated by QDF in parallel to the above operations.
// they follow the ideals of QUB_Baseline and QUB_Idealization --
// an insertion or deletion instantly heals the point indices after it,
// so go backwards or carefully offset subsequent delete/insert

// QVD responds to these in DoOnFileEdit, modifying its base/idl and putting the
// undo info into the Data field when available.\
// QVD also takes care of undo
TQUB_QFS_PEditOperationBreakSeg                = ^TQUB_QFS_EditOperationBreakSeg;
TQUB_QFS_EditOperationBreakSeg                 = record
  OperType                                     :TQUB_QFS_EditOperationType;
  Point                                        :Integer;
end;

TQUB_QFS_PEditOperationInsertSeg               = ^TQUB_QFS_EditOperationInsertSeg;
TQUB_QFS_EditOperationInsertSeg                = record
  OperType                                     :TQUB_QFS_EditOperationType;
  First                                        :Integer;
  Last                                         :Integer;
end;

TQUB_QFS_PEditOperationInsertIntoSeg           = ^TQUB_QFS_EditOperationInsertIntoSeg;
TQUB_QFS_EditOperationInsertIntoSeg            = record
  OperType                                     :TQUB_QFS_EditOperationType;
  First                                        :Integer;
  Last                                         :Integer;
end;

TQUB_QFS_PEditOperationDeleteSeg               = ^TQUB_QFS_EditOperationDeleteSeg;
TQUB_QFS_EditOperationDeleteSeg                = record
  OperType                                     :TQUB_QFS_EditOperationType;
  First                                        :Integer;
  Last                                         :Integer;
  Data                                         :TQUB_QFS_PNodeMem;
end;

TQUB_QFS_PEditOperationDeleteSel               = ^TQUB_QFS_EditOperationDeleteSel;
TQUB_QFS_EditOperationDeleteSel                = record
  OperType                                     :TQUB_QFS_EditOperationType;
  First                                        :Integer;
  Last                                         :Integer;
  Data                                         :TQUB_QFS_PNodeMem;
end;



//******************************************************************************





//Class definitions

//EditOperations
TQUB_QFS_EditOperations                        = class(TList)
protected


  function prGetOperation(o :Integer) :TQUB_QFS_EditOperation;


public


  procedure Clear; override;


  property Operation[Index :Integer]           :TQUB_QFS_EditOperation read prGetOperation; default;


end;

//EditSelections
TQUB_QFS_EditSelections                        = class(TList)
protected


  function prGetSelection(s :Integer) :TQUB_QFS_PEditSelection;


public


  procedure Clear; override;


  property Selection[Index :Integer]           :TQUB_QFS_PEditSelection read prGetSelection; default;


end;

TQUB_QFS_EditOperation                         = class(TList)
public


  procedure Clear; override;


end;


//Record_Edit
TQUB_QFS_Edit                                    = class(TQUB_QFS_Record)
protected


  prCanEdit                                      :Boolean;
  prEditStream                                   :TStream;

  prEditSelections                               :TQUB_QFS_EditSelections;


  procedure prUpdateDataCount; virtual;

  procedure prCheckHasEditSelection;

  function LoadSelections(First, Last :Integer) :Boolean; virtual;
  function LoadSelectionsInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean; virtual;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  destructor Destroy; override;


  procedure UpdateSelections;
  procedure InsertSelection(Selection :TQUB_QFS_PEditSelection; SelectionIndex :Integer);
  procedure RemoveSelection(SelectionIndex :Integer);

  function DeleteSelection(First, Last :Integer; Operation :TQUB_QFS_EditOperation) :Boolean;
  function BreakSelection(Point :Integer; Operation :TQUB_QFS_EditOperation) :Boolean;

  function WhichSelectionForPoint(p :Integer; IncIfLast :Boolean) :Integer;

  function LoadData(First, Last :Integer) :Boolean; override;
  function LoadDataInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean; override;
  function ReplaceData(Buffer :Pointer; First, Last :Integer) :Boolean; virtual;


  property CanEdit     :Boolean                  read prCanEdit    write prCanEdit;
  property EditStream  :TStream                  read prEditStream write prEditStream;

  property Selections  :TQUB_QFS_EditSelections  read prEditSelections;


end;


implementation


uses

QUB_QVO_QVD_Global;


//****************************************************************************//
//                                                                            //
//                           TQUB_QFS_EditOperations                          //
//                                                                            //
//****************************************************************************//
procedure TQUB_QFS_EditOperations.Clear;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      try
        TQUB_QFS_EditOperation(Items[i]).Free;
      except
      end;
    end;
  finally
    inherited Clear;
  end;
end;

function TQUB_QFS_EditOperations.prGetOperation(o :Integer) :TQUB_QFS_EditOperation;
begin
  if (o >= 0) and (o < Count) then Result := TQUB_QFS_EditOperation(Items[o]) else Result := nil;
end;

//****************************************************************************//
//                                                                            //
//                           TQUB_QFS_EditSelections                          //
//                                                                            //
//****************************************************************************//
procedure TQUB_QFS_EditSelections.Clear;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      try
        FreeMem(TQUB_QFS_PEditSelection(Items[i]));
      except
      end;
    end;
  finally
    inherited Clear;
  end;  
end;

function TQUB_QFS_EditSelections.prGetSelection(s :Integer) :TQUB_QFS_PEditSelection;
begin
  if (s >= 0) and (s < Count) then Result := TQUB_QFS_PEditSelection(Items[s]) else Result := nil;
end;


//****************************************************************************//
//                                                                            //
//                           TQUB_QFS_EditOperation                           //
//                                                                            //
//****************************************************************************//
procedure TQUB_QFS_EditOperation.Clear;
var
  OperItem :TQUB_QFS_PEditOperationBase;
  i :Integer;
begin
  for i := Count - 1 downto 0 do begin
    OperItem := TQUB_QFS_PEditOperationBase(Items[i]);
    case OperItem.OperType of
      OPER_TYPE_DEL_SEGMENT:
        TQUB_QFS_PEditOperationDelSegment(OperItem).Segment.Free;
      OPER_TYPE_ADD_SEGMENT:
        ;//Don't delete Segment, since it is owned and freed by its owner

      OPER_TYPE_DEL_SELECTION:
        FreeMem(TQUB_QFS_PEditOperationDelSelection(OperItem).Selection);
      OPER_TYPE_ADD_SELECTION:
        ;//Don't delete Selection, since it is owned and freed by the EditRecord's Selections

      OPER_TYPE_INS_ISEG:
        ;
      OPER_TYPE_INS_INTO_ISEG:
        ;
      OPER_TYPE_DEL_ISEG:
        Node_DECREF( TQUB_QFS_PEditOperationDeleteSeg(OperItem).Data );
      OPER_TYPE_DEL_ISEL:
        Node_DECREF( TQUB_QFS_PEditOperationDeleteSel(OperItem).Data );

      OPER_TYPE_BREAK:
        ;

      // . . .

    end;
    FreeMem(Items[i]);
  end;
  inherited Clear;
end;



//****************************************************************************//
//                                                                            //
//                                  E D I T                                   //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_Edit.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited;
  prEditSelections := TQUB_QFS_EditSelections.Create;
end;

destructor TQUB_QFS_Edit.Destroy;
begin
  try
    prEditSelections.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_QFS_Edit.UpdateSelections;
begin
  prUpdateDataCount;
end;

procedure TQUB_QFS_Edit.InsertSelection(Selection :TQUB_QFS_PEditSelection; SelectionIndex :Integer);
begin
  //It has Data already, but not in Selections
  if (DataCount <> 0) and (prEditSelections.Count = 0) then begin
    prCheckHasEditSelection;
  end;
  prEditSelections.Insert(SelectionIndex, Selection);
  prUpdateDataCount;
end;

procedure TQUB_QFS_Edit.RemoveSelection(SelectionIndex :Integer);
begin
  prEditSelections.Delete(SelectionIndex);
  prUpdateDataCount;
end;

function TQUB_QFS_Edit.BreakSelection(Point :Integer; Operation :TQUB_QFS_EditOperation) :Boolean;
var
  Selection  :TQUB_QFS_PEditSelection;
  OperDelSel :TQUB_QFS_PEditOperationDelSelection;
  OperAddSel :TQUB_QFS_PEditOperationAddSelection;
  f, l       :Integer;
  s          :Integer;
begin
  Result := False;
  prCheckHasEditSelection;

  //What if the Selection is only one Point ???

  try

    f := 0;
    s := 0;
    repeat

      //calculate the last point
      l := f + prEditSelections[s].DataCount - 1;

      //Selection s'th has (some) data within First to Last range
      if (Point >= f) and (Point <= l) then begin

        //Doesn't make much sense to break it ...
        if (Point = f) or (Point = l) then Break;

        //Delete the old selection ...
        OperDelSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationDelSelection));
        OperDelSel.OperType       := OPER_TYPE_DEL_SELECTION;
        OperDelSel.Selection      := prEditSelections[s];
        OperDelSel.SelectionIndex := s;
        OperDelSel.EditRecord     := Self;
        Operation.Add(OperDelSel);

        prEditSelections.Delete(s);

        //... and create two new ones

        //First
        Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
        Selection.DataCount       := Point - f + 1;
        Selection.StreamPos       := OperDelSel.Selection.StreamPos;
        Selection.FromEditStream  := OperDelSel.Selection.FromEditStream;

        prEditSelections.Insert(s, Selection);

        OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
        OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
        OperAddSel.Selection      := prEditSelections[s];
        OperAddSel.SelectionIndex := s;
        OperAddSel.EditRecord     := Self;
        Operation.Add(OperAddSel);

        Inc(s);

        //Second
        Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
        Selection.DataCount       := l - Point - 1 + 1;
        Selection.StreamPos       := OperDelSel.Selection.StreamPos + DataSize * (Point + 1 - f);
        Selection.FromEditStream  := OperDelSel.Selection.FromEditStream;

        prEditSelections.Insert(s, Selection);

        OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
        OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
        OperAddSel.Selection      := prEditSelections[s];
        OperAddSel.SelectionIndex := s;
        OperAddSel.EditRecord     := Self;
        Operation.Add(OperAddSel);

        Break;

      end else begin

        Inc(s);

      end;

      f := l + 1;

    until ((s >= prEditSelections.Count) or (f > Point));

  finally

    prUpdateDataCount;

  end;
end;

function TQUB_QFS_Edit.DeleteSelection(First, Last :Integer; Operation :TQUB_QFS_EditOperation) :Boolean;
var
  Selection  :TQUB_QFS_PEditSelection;
  OperDelSel :TQUB_QFS_PEditOperationDelSelection;
  OperAddSel :TQUB_QFS_PEditOperationAddSelection;
  f, l       :Integer;
  cf, cl     :Integer;
  s          :Integer;
begin
  Result := False;
  prCheckHasEditSelection;

  try

    f := 0;
    s := 0;
    repeat

      //calculate the last point
      l := f + prEditSelections[s].DataCount - 1;

      //Selection s'th has (some) data within First to Last range
      if not ((Last < f) or (First > l)) then begin

        cf := f;
        if First > cf then cf := First;
        cl := l;
        if Last < cl  then cl := Last;

        if (cf = f) and (cl = l) then begin
        //Delete the whole selection

          OperDelSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationDelSelection));
          OperDelSel.OperType       := OPER_TYPE_DEL_SELECTION;
          OperDelSel.Selection      := prEditSelections[s];
          OperDelSel.SelectionIndex := s;
          OperDelSel.EditRecord     := Self;
          Operation.Add(OperDelSel);
          prEditSelections.Delete(s);

        end else begin
        //Delete just a part of the selection

          //Delete the old selection ...
          OperDelSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationDelSelection));
          OperDelSel.OperType       := OPER_TYPE_DEL_SELECTION;
          OperDelSel.Selection      := prEditSelections[s];
          OperDelSel.SelectionIndex := s;
          OperDelSel.EditRecord     := Self;
          Operation.Add(OperDelSel);

          prEditSelections.Delete(s);

          if (cf > f) and (cl < l) then begin
          //Delete in the middle

            //... and create two new ones

            //First
            Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
            Selection.DataCount       := cf - 1 - f + 1;
            Selection.StreamPos       := OperDelSel.Selection.StreamPos;
            Selection.FromEditStream  := OperDelSel.Selection.FromEditStream;

            prEditSelections.Insert(s, Selection);

            OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
            OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
            OperAddSel.Selection      := prEditSelections[s];
            OperAddSel.SelectionIndex := s;
            OperAddSel.EditRecord     := Self;
            Operation.Add(OperAddSel);

            Inc(s);

            //Second
            Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
            Selection.DataCount       := l - cl - 1 + 1;
            Selection.StreamPos       := OperDelSel.Selection.StreamPos + DataSize * (cl + 1 - f);
            Selection.FromEditStream  := OperDelSel.Selection.FromEditStream;

            prEditSelections.Insert(s, Selection);

            OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
            OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
            OperAddSel.Selection      := prEditSelections[s];
            OperAddSel.SelectionIndex := s;
            OperAddSel.EditRecord     := Self;
            Operation.Add(OperAddSel);

            Inc(s);

          end else begin
          //Delete an end

            //... and create a new one
            Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
            Selection.DataCount       := OperDelSel.Selection.DataCount - (cl - cf + 1);
            if cf = f then begin
            //Delete the start
              Selection.StreamPos     := OperDelSel.Selection.StreamPos + DataSize * (cl + 1 - f);
            end else if cl = l then begin
            //Delete the end
              Selection.StreamPos     := OperDelSel.Selection.StreamPos;
            end;
            Selection.FromEditStream  := OperDelSel.Selection.FromEditStream;

            prEditSelections.Insert(s, Selection);

            OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
            OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
            OperAddSel.Selection      := prEditSelections[s];
            OperAddSel.SelectionIndex := s;
            OperAddSel.EditRecord     := Self;
            Operation.Add(OperAddSel);

            Inc(s);

          end;

        end;

      end else begin

        Inc(s);

      end;

      f := l + 1;

    until ((s >= prEditSelections.Count) or (f > Last));

  finally

    prUpdateDataCount;

  end;

end;

function TQUB_QFS_Edit.WhichSelectionForPoint(p :Integer; IncIfLast :Boolean) :Integer;
var
  s, f, l :Integer;
begin
  prCheckHasEditSelection;

  Result := prEditSelections.Count;

  if p = -1 then Exit;

  f := 0;
  for s := 0 to prEditSelections.Count - 1 do begin
    l := f + prEditSelections[s].DataCount - 1;
    if (p >= f) and (p <= l) then begin
      Result := s;
      if (p = l) and (IncIfLast) then begin
        Inc(Result);
      end;
      Break;
    end;
    f := l + 1;
  end;
end;

//Should be called whithin any Edit method
procedure TQUB_QFS_Edit.prCheckHasEditSelection;
var
  Sel :TQUB_QFS_PEditSelection;
begin
  if prEditSelections.Count > 0 then Exit;

  Sel                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
  Sel.DataCount       := DataCount;
  Sel.FromEditStream  := False;//?!?!
  Sel.StreamPos       := DataStreamPos;

  prEditSelections.Add(Sel);

end;

procedure TQUB_QFS_Edit.prUpdateDataCount;
var
  s :Integer;
begin
  DataCount := 0;
  for s := 0 to prEditSelections.Count - 1 do begin
    DataCount := DataCount + prEditSelections[s].DataCount;
  end;
end;


//****************************************************************************//

function TQUB_QFS_Edit.LoadData(First, Last :Integer) :Boolean;
begin
  if prEditSelections.Count <> 0 then begin
  //Read Selections

    Result := LoadSelections(First, Last);

  end else begin
  //Read inherited

    Result := inherited LoadData(First, Last);

  end;
end;

function TQUB_QFS_Edit.LoadDataInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean;
begin
  if prEditSelections.Count <> 0 then begin
  //Read Selections

    Result := LoadSelectionsInBuffer(Buffer, First, Last);

  end else begin
  //Read inherited

    Result := inherited LoadDataInBuffer(Buffer, First, Last);

  end;
end;

function TQUB_QFS_Edit.LoadSelections(First, Last :Integer) :Boolean;
begin

  Result := False;

  try

    try

      ReallocMem(prNodeMem.Data, (Last - First + 1) * DataSize);
      DataLoadSel := QUB_FirstLast(First, Last);

    except

      Result := True;

      try
        UnloadData;
      finally
      end;

      Exit;

    end;

    Result := LoadSelectionsInBuffer(Data, First, Last);

  finally

    if not Result then begin
      DataLoadSel := QUB_FirstLast(First, Last);
    end else begin
      DataLoadSel := QUB_FirstLast(-1, -1);
    end;

  end;

end;

function TQUB_QFS_Edit.LoadSelectionsInBuffer(Buffer :Pointer; First, Last :Integer) :Boolean;
var
  BufferB   :TPAByte;
  Stream    :TStream;
  f, l      :Integer;
  cf, cl    :Integer;
  s         :Integer;
  OldPos    :Integer;
begin

  Result := False;

  try

    BufferB := Buffer;

    //Load Selections
    f := 0;
    s := 0;
    repeat

      //calculate the last point
      l := f + prEditSelections[s].DataCount - 1;

      //Selection s'th has (some) data within First to Last range
      if not ((Last < f) or (First > l)) then begin

        cf := f;
        if First > cf then cf := First;
        cl := l;
        if Last < cl  then cl := Last;

        if prEditSelections[s].FromEditStream then begin
          Stream := EditStream;
        end else begin
          Stream := Self.Stream;
        end;

        //Read the data within the Selection's range
        try
          OldPos := Stream.Position;
          try
            Stream.Position := prEditSelections[s].StreamPos + DataSize * (cf - f);
            Stream.ReadBuffer((@BufferB[DataSize * (cf - First)])^, DataSize * (cl - cf + 1));
          finally
            Stream.Position := OldPos;
          end;
        except
          Result := True;
        end;

        if Result then Exit;

      end;

      Inc(s);

      f := l + 1;

    until ((s = prEditSelections.Count) or (f > Last));

  finally

  end;

end;


function TQUB_QFS_Edit.ReplaceData(Buffer :Pointer; First, Last :Integer) :Boolean;
var
  OldPos    :Integer;
  Stream    :TStream;
  f, l      :Integer;
  cf, cl    :Integer;
  s         :Integer;
begin
  Result := False;

  try


    if prEditSelections.Count <> 0 then begin
    //Replace Selections


      //Load Selections
      f := 0;
      s := 0;
      repeat

        //calculate the last point
        l := f + prEditSelections[s].DataCount - 1;

        //Selection s'th has (some) data within First to Last range
        if not ((Last < f) or (First > l)) then begin

          cf := f;
          if First > cf then cf := First;
          cl := l;
          if Last < cl  then cl := Last;

          if prEditSelections[s].FromEditStream then begin
            Stream := EditStream;
          end else begin
            Stream := Self.Stream;
          end;

          OldPos := Stream.Position;
          try
            Stream.Position := prEditSelections[s].StreamPos + DataSize * (cf - f);
            Stream.WriteBuffer((TPAByte(Buffer)[DataSize * (cf - First)]), DataSize * (cl - cf + 1));
          finally
            Stream.Position := OldPos;
          end;

        end;

        Inc(s);

        f := l + 1;

      until ((s = prEditSelections.Count) or (f > Last));


    end else begin
    //Replace Data


      Stream := Self.Stream;
      OldPos := Stream.Position;
      try
        Stream.Position := DataStreamPos + First * DataSize;
        Stream.WriteBuffer(Buffer^, (Last - First + 1) * DataSize);
      finally
        Stream.Position := OldPos;
      end;


    end;


  except

    Result := True;

  end;

end;


//******************************************************************************
end.
