{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2003
// unfinished attempt to let you interact with a QFS tree as though it were qubtree.
// abandoned due to confusion over how QFS handles node data.


unit QUB_QFS_Node;

interface

uses

AllocTypes,
QUB_Global,
sysutils,
QUB_QFS, QUB_Node;


type

TQUB_QFS_Node = class(TInterfacedObject, INode)
protected

  prQFS                      :TQUB_QFS_Record;


public

  constructor Create(QFS :TQUB_QFS_Record);
  // destructor Destroy; override;

   function GetName: String;
   procedure SetName(const nm : String);

   function GetChild(nameOrIx: Variant): INode; // dispatches to Child(...)
   function Child: INode; overload;
   function Child(ix	: Integer): INode; overload;
   function Child(const nm : String): INode; overload; // may create and append

   function IsPreload: Boolean;
   procedure SetPreload(pl : Boolean);
   function IsChanged: Boolean;
   procedure SetChanged(ch : Boolean);
   function IsChildChanged: Boolean;
   function IsDataInNode: Boolean;

   function GetDataType: TQUB_QFS_Node_DataType;
   function GetRowSize: LongWord;
   function GetDataCount: LongWord;
   function GetDataRows: LongWord;
   function GetDataCols: LongWord;
   function GetLoadBounds: TQUB_FirstLast;
   function GetLoadRowBounds: TQUB_FirstLast;

   function GetDataElem(ix :Integer): Variant;
   procedure SetDataElem(ix :Integer; v: Variant);

   function GetMatrixElem(i, j :Integer): Variant;
   procedure SetMatrixElem(i, j :Integer; v: Variant);

   function AsString: String;
   procedure SaveAsText(const path: String );

   function Save: Boolean;
   function SaveAs( const path :String ): Boolean;
   function SaveAsTemp: Boolean;
   function Close: Boolean;
   function IsInFile: Boolean;
   function GetPath: String;

   function Clone(deep : Boolean) : INode;
   function Copy: Boolean;

   function Equals(node : INode) : Boolean;
   function IsNull: Boolean;

   function Parent: INode;
   function Sibling: INode; overload;
   function Sibling(ix : Integer): INode; overload;
   function Sibling(const name: String): INode; overload;
   function SiblingSameName: INode;

   function Find(const nm : String): INode; // may return a null node

   function AppendChild(const newchildname : String): INode; overload;
   function AppendChild(newchild : INode): INode; overload;
   function AppendClone(Template :INode; deep :Boolean): INode;

   function InsertChild(after :INode; const newchildname : String): INode; overload;
   function InsertChild(after :INode; newchild : INode): INode; overload;
   function InsertClone(after :INode; orig : INode; deep :Boolean): INode;

   procedure RemoveChild(remch : INode); overload;
   procedure RemoveChild(remch : INode; Prev :INode); overload;
   procedure RemoveLast(const ChildName :String);

   function Flags: LongWord;

   function Lock(timeoutMS : Integer) : Boolean;
   procedure Unlock;

   function GetDataPtr: Pointer;
   function GetDataAsString: String;

   function SizeOfType(dtype :TQUB_QFS_Node_DataType): LongWord;

   function SetupData(dtype :TQUB_QFS_Node_DataType; dsize, dcount: LongWord): Boolean;
   function SetupNumData(dtype :TQUB_QFS_Node_DataType; rows, cols :LongWord): Boolean;
   function SetData(val: Variant): Boolean;
   function EmulateData(Other :INode): Boolean;
   
   function ResizeData(newrowcount : LongWord): Boolean;
   function ClearData: Boolean;

   function LoadRows(var bounds	: TQUB_FirstLast; doRead: Boolean): Boolean;
   function UnloadRows(doWrite : Boolean): Boolean;

   // next 2 return # rows read or written
   function ReadRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;
   function WriteRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;

   function GetLineComment: String;
   procedure SetLineComment(const com : String);

   function GetNodeMem: TQUB_QFS_PNodeMem;
   function GetNodeType: Integer;


end;


const
  QUB_NODETYPE_QFS                            = 2;


function QUB_QFS_INode(QFS :TQUB_QFS_Record): INode;


implementation


{$IFDEF DELPHI6_OR_HIGHER}
uses

Variants;
{$ENDIF}


function QUB_QFS_INode(QFS :TQUB_QFS_Record): INode;
begin
  if QFS <> nil then begin
    Result := TQUB_QFS_Node.Create( QFS );
  end else begin
    Result := TNode.CreateNull;
  end;
end;

constructor TQUB_QFS_Node.Create(QFS :TQUB_QFS_Record);
begin
  prQFS                := QFS;
  if prQFS = nil then raise Exception.Create('Attempt to create a QFS_Node from nil.');
end;

function TQUB_QFS_Node.GetName: String;
begin
  Result               := prQFS.Name;
end;

procedure TQUB_QFS_Node.SetName(const nm : String);
begin
  raise Exception.Create( Format('Can''t rename QFS node %s to %s.', [GetName, nm]) );
end;

function TQUB_QFS_Node.GetChild(nameOrIx: Variant): INode;
begin
  case VarType(nameOrIx) and varTypeMask of
    varSmallint, varInteger:
      Result := Child( Integer(nameOrIx) );
    varString:
      Result := Child( String(nameOrIx) );
  else
    Result := TNode.CreateNull;
  end;
end;

function TQUB_QFS_Node.Child: INode;
begin
  Result := Child(0);
end;

function TQUB_QFS_Node.Child(ix	: Integer): INode;
begin
  if prQFS.RecordCount > ix then begin
    Result := TQUB_QFS_Node.Create( prQFS.Records[ix] );
  end else begin
    Result := TNode.CreateNull;
  end;
end;

function TQUB_QFS_Node.Child(const nm : String): INode;
var
  ChildQFS            :TQUB_QFS_Record;
begin
  ChildQFS            := prQFS.RecordsByName[ nm ];
  if ChildQFS = nil then begin
    ChildQFS          := TQUB_QFS_Record.Create(nm, prQFS);
  end;
  Result              := QUB_QFS_INode( ChildQFS );
end;

function TQUB_QFS_Node.IsPreload: Boolean;
begin
  Result              := prQFS.FlagPreload;
end;

procedure TQUB_QFS_Node.SetPreload(pl : Boolean);
begin
  prQFS.FlagPreload   := pl;
end;

function TQUB_QFS_Node.IsChanged: Boolean;
begin
  Result              := prQFS.Modified;
end;

procedure TQUB_QFS_Node.SetChanged(ch : Boolean);
begin
  // ??
end;

function TQUB_QFS_Node.IsChildChanged: Boolean;
begin
  Result              := False;
  // ??
end;

function TQUB_QFS_Node.IsDataInNode: Boolean;
begin
  Result              := prQFS.FlagDataInNode;
end;

function TQUB_QFS_Node.GetDataType: TQUB_QFS_Node_DataType;
begin
  Result              := prQFS.DataType;
end;

function TQUB_QFS_Node.GetRowSize: LongWord;
begin
  Result              := prQFS.DataSize;
end;

function TQUB_QFS_Node.GetDataCount: LongWord;
begin
  Result              := LongWord(prQFS.DataCount) * GetDataCols;
end;

function TQUB_QFS_Node.GetDataRows: LongWord;
begin
  Result              := prQFS.DataCount;
end;

function TQUB_QFS_Node.GetDataCols: LongWord;
begin
  if Integer( prQFS.DataType ) > 1 then begin
    Result := LongWord(prQFS.DataSize) div SizeOfType( prQFS.DataType );
  end else begin
    Result := 1;
  end;
end;

function TQUB_QFS_Node.GetLoadBounds: TQUB_FirstLast;
var
  Cols                :LongWord;
begin
  Result              := prQFS.DataLoadSel;
  Cols                := GetDataCols;
  Result.First        := Result.First * LongInt(Cols);
  Result.Last         := (Result.Last + 1) * LongInt(Cols) - 1;
end;

function TQUB_QFS_Node.GetLoadRowBounds: TQUB_FirstLast;
begin
  Result              := prQFS.DataLoadSel;
end;

function TQUB_QFS_Node.GetDataElem(ix :Integer): Variant;
begin
  Result              := prQFS.DataArrayValue[ix]; // [ix div GetDataRows];
end;

procedure TQUB_QFS_Node.SetDataElem(ix :Integer; v: Variant);
begin
  prQFS.DataArrayValue[ix] := v; // [ix div GetDataRows] := v;
end;

function TQUB_QFS_Node.GetMatrixElem(i, j :Integer): Variant;
begin
  Result              := prQFS.DataArrayValue[i*LongInt(GetDataCols)+j]; // [i div GetDataRows];
end;

procedure TQUB_QFS_Node.SetMatrixElem(i, j :Integer; v: Variant);
begin
  prQFS.DataArrayValue[i*LongInt(GetDataCols)+j] := v; // [i div GetDataRows] := v;
end;

function TQUB_QFS_Node.AsString: String;
var
  Temp                :INode;
begin
  Temp                := prQFS.CloneAsNode(True, TNode.CreateNull);
  Result              := Temp.AsString;
end;

procedure TQUB_QFS_Node.SaveAsText(const path: String );
var
  Temp                :INode;
begin
  Temp                := prQFS.CloneAsNode(True, TNode.CreateNull);
  Temp.SaveAsText(path);
end;

function TQUB_QFS_Node.Save: Boolean;
begin
  raise Exception.Create('Can''t save a TQUB_QFS_Node');
end;

function TQUB_QFS_Node.SaveAs( const path :String ): Boolean;
begin
  raise Exception.Create('Can''t save_as a TQUB_QFS_Node');
end;

function TQUB_QFS_Node.SaveAsTemp: Boolean;
begin
  raise Exception.Create('Can''t save_as a TQUB_QFS_Node');
end;

function TQUB_QFS_Node.Close: Boolean;
begin
  Result := True;
end;

function TQUB_QFS_Node.IsInFile: Boolean;
begin
  Result := False;
end;

function TQUB_QFS_Node.GetPath: String;
begin
  Result := '';
end;

function TQUB_QFS_Node.Clone(deep : Boolean) : INode;
begin
  Result              := prQFS.CloneAsNode(Deep, TNode.CreateNull);
end;

function TQUB_QFS_Node.Copy: Boolean;
begin
  Result              := Clone(True).Copy;
end;

function TQUB_QFS_Node.Equals(node : INode) : Boolean;
begin
  if node.NodeType <> GetNodeType then begin
    Result            := False;
  end else begin
    Result            := TQUB_QFS_Node(node).prQFS = prQFS;
  end;
end;

function TQUB_QFS_Node.IsNull: Boolean;
begin
  Result              := False; // due to exception in the constructor
end;

function TQUB_QFS_Node.Parent: INode;
begin
  Result              := QUB_QFS_INode( prQFS.Parent );
end;

function TQUB_QFS_Node.Sibling: INode;
begin
  Result              := Sibling(0);
end;

function TQUB_QFS_Node.Sibling(ix : Integer): INode;
var
  ParentQFS           :TQUB_QFS_Record;
begin
  ParentQFS           := prQFS.Parent;
  if not (ParentQFS = nil) then begin
    Result              := QUB_QFS_INode( ParentQFS.Records[ 1 + ix + ParentQFS.RecordIndex( prQFS ) ] );
  end else begin
    Result              := TNode.CreateNull;
  end;
end;

function TQUB_QFS_Node.Sibling(const name: String): INode;
var
  ParentQFS           :TQUB_QFS_Record;
  SiblingI            :TQUB_QFS_Record;
  i                   :Integer;
begin
  Result              := nil;
  ParentQFS           := prQFS.Parent;
  for i := ParentQFS.RecordIndex( prQFS ) + 1 to ParentQFS.RecordCount - 1 do begin
    SiblingI          := ParentQFS.Records[i];
    if SiblingI.Name = name then begin
      Result          := QUB_QFS_INode( SiblingI );
      Break;
    end;
  end;

  if Result = nil then Result := TNode.CreateNull;
end;

function TQUB_QFS_Node.SiblingSameName: INode;
begin
  Result := Sibling(GetName);
end;

function TQUB_QFS_Node.Find(const nm : String): INode;
begin
  Result              := QUB_QFS_INode( prQFS.RecordsByName[ nm ] );
end;

function TQUB_QFS_Node.AppendChild(const newchildname : String): INode;
begin
  Result := TQUB_QFS_Node.Create( TQUB_QFS_Record.Create( newchildname, prQFS ) );
end;

function TQUB_QFS_Node.AppendChild(newchild : INode): INode;
begin
  raise Exception.Create('Can''t append INode to QFS_Node.');
end;

function TQUB_QFS_Node.AppendClone(Template :INode; deep :Boolean): INode;
begin
  prQFS.AddRecordFromNode(Template);
  Result := TQUB_QFS_Node.Create( prQFS.Records[ prQFS.RecordCount - 1 ] );
end;

function TQUB_QFS_Node.InsertChild(after :INode; const newchildname : String): INode;
begin
  raise Exception.Create('sorry not available yet');
end;

function TQUB_QFS_Node.InsertChild(after :INode; newchild : INode): INode;
begin
  raise Exception.Create('Can''t insert INode into QFS_Node.');
end;

function TQUB_QFS_Node.InsertClone(after :INode; orig : INode; deep :Boolean): INode;
begin
  raise Exception.Create('sorry not available yet');
end;

procedure TQUB_QFS_Node.RemoveChild(remch : INode);
var
  ch                   :INode;
  i                    :Integer;
begin
  if remch.NodeType = GetNodeType then begin
    ch                   := Child;
    i                    := 0;
    while not ch.IsNull do begin
      if ch.Equals(remch) then begin
        prQFS.Records[i].RemoveFromParent;
        Break;
      end;

      ch                 := ch.Sibling;
      i                  := i + 1;
    end;
  end;
end;

procedure TQUB_QFS_Node.RemoveChild(remch : INode; Prev :INode);
begin
  RemoveChild( remch );
end;

procedure TQUB_QFS_Node.RemoveLast(const ChildName :String);
begin
  raise Exception.Create('I didn''t think anyone would use this method.  oops.');
end;

function TQUB_QFS_Node.Flags: LongWord;
begin
  Result := TPLongWord(@(prQFS.Flags))^ and $FFFFFF
end;

function TQUB_QFS_Node.Lock(timeoutMS : Integer) : Boolean;
begin
  Result              := True;
end;

procedure TQUB_QFS_Node.Unlock;
begin
end;

function TQUB_QFS_Node.GetDataPtr: Pointer;
begin
  Result               := prQFS.Data;
end;

function TQUB_QFS_Node.GetDataAsString: String;
var
  Temp                 :INode;
begin
  Temp                 := TNode.Create('');
  Temp.SetupData( GetDataType, GetRowSize, GetDataCount );
  prQFS.LoadDataInBuffer( Temp.DataPtr, 0, Temp.DataCount - 1 );
  Result               := Temp.DataAsString;
end;

function TQUB_QFS_Node.SizeOfType(dtype :TQUB_QFS_Node_DataType): LongWord;
begin
  case dtype of
    NODE_DATATYPE_EMPTY:        Result := 0;
    NODE_DATATYPE_UNKNOWN:      Result := SizeOf(Pointer);
    NODE_DATATYPE_POINTER:      Result := SizeOf(Pointer);
    NODE_DATATYPE_STRING:       Result := SizeOf(Char);
    NODE_DATATYPE_BYTE:         Result := SizeOf(Byte);
    NODE_DATATYPE_SHORTINT:     Result := SizeOf(ShortInt);
    NODE_DATATYPE_WORD:         Result := SizeOf(Word);
    NODE_DATATYPE_SMALLINT:     Result := SizeOf(SmallInt);
    NODE_DATATYPE_LONGWORD:     Result := SizeOf(LongWord);
    NODE_DATATYPE_LONGINT:      Result := SizeOf(Integer);
    NODE_DATATYPE_UINT64:       Result := SizeOf(UInt64);
    NODE_DATATYPE_INT64:        Result := SizeOf(Int64);
    NODE_DATATYPE_SINGLE:       Result := SizeOf(Single);
    NODE_DATATYPE_DOUBLE:       Result := SizeOf(Double);
    NODE_DATATYPE_EXTENDED:     Result := SizeOf(Extended);
  else
    Result := 0;
  end;
end;

function TQUB_QFS_Node.SetupData(dtype :TQUB_QFS_Node_DataType; dsize, dcount: LongWord): Boolean;
begin
  Result := False;
end;

function TQUB_QFS_Node.SetupNumData(dtype :TQUB_QFS_Node_DataType; rows, cols :LongWord): Boolean;
begin
  Result := False;
end;

function TQUB_QFS_Node.SetData(val: Variant): Boolean;
var
  vtype                :Integer;
  dtype                :TQUB_QFS_NODE_DataType;
  high, low            :Integer;
  count                :LongWord;
  allbounds            :TQUB_FirstLast;
  ptr                  :Pointer;
begin
  vtype := VarType(val);
  ptr   := nil;

  if (vtype and varString) <> 0 then begin
    prQFS.DataValue := val;
    Result          := True;
  end else begin
    case vtype and varTypeMask of
      varByte:      dtype := NODE_DATATYPE_BYTE;
      varSmallint:  dtype := NODE_DATATYPE_SMALLINT;
      varInteger:   dtype := NODE_DATATYPE_LONGINT;
      varSingle:    dtype := NODE_DATATYPE_SINGLE;
      varDouble:    dtype := NODE_DATATYPE_DOUBLE;
    else
      raise EVariantError.Create('unknown variant type');
    end;

    if (vtype and varArray) <> 0 then begin
    //Array
      high := VarArrayHighBound(val, 1);
      low := VarArrayLowBound(val, 1);
      count := high - low + 1;
      SetupNumData( dtype, count, 1 );
      allBounds.Last := GetDataRows;
      LoadRows( allBounds, False{noread} );

      try
        ptr := VarArrayLock(val);
        if ptr <> nil then Move(ptr^, GetDataPtr^, count * SizeOfType(dtype));
      finally
        if ptr <> nil then VarArrayUnLock(val);
      end;

    end else begin

      SetupNumData( dtype, 1, 1 );
      LoadRows( allBounds, False{noread} );
      SetDataElem( 0, val );

    end;

    Result := True;

  end;
end;

function TQUB_QFS_Node.EmulateData(Other :INode): Boolean;
begin
  raise Exception.Create('not ready for use');
end;

function TQUB_QFS_Node.ResizeData(newrowcount : LongWord): Boolean;
begin
  Result := False;
end;

function TQUB_QFS_Node.ClearData: Boolean;
begin
  Result := not prQFS.ClearData;
end;

function TQUB_QFS_Node.LoadRows(var bounds : TQUB_FirstLast; doRead: Boolean): Boolean;
begin
  Result := not prQFS.LoadData( bounds.First, bounds.Last );
end;

function TQUB_QFS_Node.UnloadRows(doWrite : Boolean): Boolean;
begin
  Result := not prQFS.UnloadData;
end;

// next 2 return # rows read or written
function TQUB_QFS_Node.ReadRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;
begin
  if prQFS.LoadDataInBuffer(buf, bounds.First, bounds.Last) then begin
    Result := 0;
  end else begin
    Result := bounds.Last - bounds.First + 1;
  end;
end;

function TQUB_QFS_Node.WriteRows(buf :Pointer; var bounds: TQUB_FirstLast): LongWord;
begin
  Result := 0;
  // ??
end;

function TQUB_QFS_Node.GetLineComment: String;
var
   com : INode;
begin
   com := Find('QTR_LINE_COMMENT');
   if not com.IsNull then Result := com.DataAsString
   else Result := '';
end;

procedure TQUB_QFS_Node.SetLineComment(const com : String);
begin
  Child('QTR_LINE_COMMENT').SetData(com);
end;

function TQUB_QFS_Node.GetNodeMem: TQUB_QFS_PNodeMem;
begin
  raise Exception.Create('Attempt to get NodeMem of QFS_Node.');
end;

function TQUB_QFS_Node.GetNodeType: Integer;
begin
  Result := QUB_NODETYPE_QFS;
end;

// *****************************************************************************
end.
