{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// enums and stuff for QUB_QFS_QDF


unit QUB_QFS_QDF_Global;


interface


uses


AllocTypes,

QUB_Global,
QUB_QFS, QUB_Node;


type


//This is the place where to register File types (although changes in other places are required)
TQUB_QFS_QDF_FileType              = (QFS_QDF_TYPE_QDF,
                                      QFS_QDF_TYPE_LDT,
                                      QFS_QDF_TYPE_DAT,
                                      QFS_QDF_TYPE_ASC,
                                      QFS_QDF_TYPE_ABF,
                                      QFS_QDF_TYPE_WAV,
                                      QFS_QDF_TYPE_DWT,
                                      QFS_QDF_TYPE_EVL,
                                      QFS_QDF_TYPE_PRT,
                                      QFS_QDF_TYPE_QSF,
                                      QFS_QDF_TYPE_PUL,
                                      QFS_QDF_TYPE_ACQ,
                                      QFS_QDF_TYPE_IBW,
                                      QFS_QDF_TYPE_TAC,
                                      QFS_QDF_TYPE_UNK);


const


//File Extension
QUB_QFS_QDF_FILE_EXTS             :Array[TQUB_QFS_QDF_FileType] of String =
                                     ('.qdf',
                                      '.ldt',
                                      '.dat',
                                      '.txt',//'.asc',
                                      '.abf',
                                      '.wav',
                                      '.dwt',
                                      '.evl',
                                      '.prt',
                                      '.qsf',
                                      '.pul',
                                      '.acquire',
                                      '.ibw',
                                      '.tacidl',
                                      '');


function FileTypeFromFileExt(FileExt :String) :TQUB_QFS_QDF_FileType;


const


QUB_QFS_QDF_MAX_ADCHANNELCOUNT     = 256;//16

QUB_QFS_QDF_MAX_ADCHANNELCOUNTS    :Array[TQUB_QFS_QDF_FileType] of Integer =
                                     (QUB_QFS_QDF_MAX_ADCHANNELCOUNT, {QDF}
                                      1,                              {LDT}
                                      QUB_QFS_QDF_MAX_ADCHANNELCOUNT, {DAT}
                                      QUB_QFS_QDF_MAX_ADCHANNELCOUNT, {TXT}
                                      QUB_QFS_QDF_MAX_ADCHANNELCOUNT, {ABF}
                                      2,                              {WAV}
                                      1,                              {DWT}
                                      1,                              {EVL}
                                      1,                              {PRT}
                                      1,                              {QSF}
                                      QUB_QFS_QDF_MAX_ADCHANNELCOUNT, {PUL} {1?!}
                                      QUB_QFS_QDF_MAX_ADCHANNELCOUNT, {ACQ}
                                      4,                              {IBW}
                                      1,                              {TAC}
                                      0);                             {UNK}



//Streamed properties **********************************************************
//Sampling/Scaling
QUB_QFS_QDF_SAMPLING_NAME                      = 'Sampling';
QUB_QFS_QDF_SCALING_NAME                       = 'Scaling';
//ADChannels
QUB_QFS_QDF_ADCHANNELCOUNT_NAME                = 'ADChannelCount';
QUB_QFS_QDF_ADDATASIZE_NAME                    = 'ADDataSize';
QUB_QFS_QDF_ADDATATYPE_NAME                    = 'ADDataType';
//******************************************************************************



//LDT
QUB_QFS_QDF_LDT_HEADER_COUNT_NAME              = 'LDT Header Count';
QUB_QFS_QDF_LDT_HEADER_SAMPLING_NAME           = 'LDT Header Sampling';
QUB_QFS_QDF_LDT_HEADER_SCALING_NAME            = 'LDT Header Scaling';

QUB_QFS_QDF_LDT_SEGMENT_HEADER_STARTTIME_NAME  = 'LDT Segment Header StartTime';
QUB_QFS_QDF_LDT_SEGMENT_HEADER_DATAPOINTS_NAME = 'LDT Segment Header DataPoints';


//******************************************************************************
type


TQUB_QFS_QDF_ListName                          = Array[0..100] of Char;

TQUB_QFS_QDF_PCopyInfo                         = ^TQUB_QFS_QDF_CopyInfo;
TQUB_QFS_QDF_CopyInfo                          = record
  ChannelCount                                 :Integer;
  DataType                                     :TQUB_QFS_Node_DataType;
  DataSize                                     :Integer;
  DataCount                                    :Integer;
  SegCount                                     :Integer;
  ListName                                     :TQUB_QFS_QDF_ListName;                          
  Data                                         :TPAByte;
end;


TQUB_QFS_QDF_OnAddSeries                       = procedure(Sender :TObject; const SeriesName :String; Private :Boolean) of object;
TQUB_QFS_QDF_OnAddSweep                        = procedure(Sender :TObject; const SeriesName, SweepName :String; const Sel :TQUB_FirstLast; ForceAdd :Boolean) of object;


implementation


//******************************************************************************
function FileTypeFromFileExt(FileExt :String) :TQUB_QFS_QDF_FileType;
begin
  if      FileExt = '.qdf' then Result := QFS_QDF_TYPE_QDF
  else if FileExt = '.ldt' then Result := QFS_QDF_TYPE_LDT
  else if FileExt = '.dat' then Result := QFS_QDF_TYPE_DAT
  else if FileExt = '.txt' then Result := QFS_QDF_TYPE_ASC
  else if FileExt = '.abf' then Result := QFS_QDF_TYPE_ABF
  else if FileExt = '.wav' then Result := QFS_QDF_TYPE_WAV
  else if FileExt = '.dwt' then Result := QFS_QDF_TYPE_DWT
  else if FileExt = '.evl' then Result := QFS_QDF_TYPE_EVL
  else if FileExt = '.prt' then Result := QFS_QDF_TYPE_PRT
  else if FileExt = '.qsf' then Result := QFS_QDF_TYPE_QSF
  else if FileExt = '.pul' then Result := QFS_QDF_TYPE_PUL
  else if FileExt = '.acquire' then Result := QFS_QDF_TYPE_ACQ
  else if FileExt = '.ibw' then Result := QFS_QDF_TYPE_IBW
  else if FileExt = '.tacidl' then Result := QFS_QDF_TYPE_TAC
  else                          Result := QFS_QDF_TYPE_UNK;
end;


//******************************************************************************
end.
