{

                            Qub Data (De-) Multiplexor
                                     unit


core magic for converting between (array or matrix of integer) and (interleaved array of DataType)


(C) 2000 Lorin Milescu
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


//07-14-2000   Creation date


unit QUB_QFS_QDF_MPX;


interface


uses


QUB_QFS, QUB_Node;


function MPXData(Buffer,                   // V, always Integer, either tpa (T*) or tpm (T**)
                 BufferMPX :Pointer;       // V, row(chan)-major array of datasize, datatype
                 ChannelCount,             // of buffermpx
                 Channel :Integer;         // -1: buf is matrix, use all channels
                 DataSize :Integer; DataType :TQUB_QFS_Node_DataType;
                 DataFirst, DataLast :Integer;
                 MPX :Boolean; nMult :Double = 1.0) :Boolean;

function MPXDataX(Buffer,
                  BufferMPX :Pointer;
                  ChannelCount,
                  Channel :Integer;
                  DataSize, DataSizeMPX :Integer;
                  DataType, DataTypeMPX :TQUB_QFS_Node_DataType;
                  DataFirst, DataLast :Integer;
                  MPX :Boolean; Scaling :Double) :Boolean;

function MPXDataFloat(Buffer,              // V, either tpa (T*) or tpm (T**)
                 BufferMPX :Pointer;       // V, row(chan)-major array of datasize, datatype
                 ChannelCount,             // of buffermpx
                 Channel :Integer;         // -1: buf is matrix, use all channels
                 DataSize :Integer; DataType :TQUB_QFS_Node_DataType; 
                 DataFirst, DataLast :Integer; 
                 MPX :Boolean; nMult :Double = 1.0) :Boolean;


implementation


uses


AllocTypes, Alloc;


//******************************************************************************
function MPXData(Buffer, BufferMPX :Pointer; ChannelCount, Channel :Integer; DataSize :Integer; DataType :TQUB_QFS_Node_DataType; DataFirst, DataLast :Integer; MPX :Boolean; nMult :Double) :Boolean;
var
  BufferM  :TPMInteger;
  BufferA  :TPAInteger;
  RecAddr  :TPByte;
  c, i     :Integer;
begin

  Result := False;

  BufferM := Buffer;
  BufferA := Buffer;

  //i-th record's address = pointer to an array of bytes
  RecAddr := BufferMPX;

  try

  case DataType of

    NODE_DATATYPE_BYTE://Byte **************************************************

      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPByte(RecAddr)^ := Byte(BufferM[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPByte(RecAddr)^ := Byte(BufferM[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPByte(RecAddr)^ := Byte(BufferA[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPByte(RecAddr)^ := Byte(BufferA[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferM[c, i] := TPByte(RecAddr)^;
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferM[0, i] := TPByte(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferA[i] := TPByte(RecAddr)^;
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferA[i] := TPByte(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

    NODE_DATATYPE_WORD://Word **************************************************

      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPWord(RecAddr)^ := Word(BufferM[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPWord(RecAddr)^ := Word(BufferM[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPWord(RecAddr)^ := Word(BufferA[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPWord(RecAddr)^ := Word(BufferA[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferM[c, i] := TPWord(RecAddr)^;
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferM[0, i] := TPWord(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferA[i] := TPWord(RecAddr)^;
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferA[i] := TPWord(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

    NODE_DATATYPE_LONGWORD://LongWord ******************************************

      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPLongWord(RecAddr)^ := LongWord(BufferM[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPLongWord(RecAddr)^ := LongWord(BufferM[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPLongWord(RecAddr)^ := LongWord(BufferA[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPLongWord(RecAddr)^ := LongWord(BufferA[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferM[c, i] := TPLongWord(RecAddr)^;
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferM[0, i] := TPLongWord(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferA[i] := TPLongWord(RecAddr)^;
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferA[i] := TPLongWord(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

    NODE_DATATYPE_SHORTINT://ShortInt ******************************************

      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPShortInt(RecAddr)^ := ShortInt(BufferM[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPShortInt(RecAddr)^ := ShortInt(BufferM[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPShortInt(RecAddr)^ := ShortInt(BufferA[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPShortInt(RecAddr)^ := ShortInt(BufferA[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferM[c, i] := TPShortInt(RecAddr)^;
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferM[0, i] := TPShortInt(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferA[i] := TPShortInt(RecAddr)^;
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferA[i] := TPShortInt(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

    NODE_DATATYPE_SMALLINT://SmallInt ******************************************

      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPSmallInt(RecAddr)^ := SmallInt(BufferM[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPSmallInt(RecAddr)^ := SmallInt(BufferM[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPSmallInt(RecAddr)^ := SmallInt(BufferA[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPSmallInt(RecAddr)^ := SmallInt(BufferA[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferM[c, i] := TPSmallInt(RecAddr)^;
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferM[0, i] := TPSmallInt(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferA[i] := TPSmallInt(RecAddr)^;
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferA[i] := TPSmallInt(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

    NODE_DATATYPE_LONGINT://LongInt ********************************************

      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPLongInt(RecAddr)^ := LongInt(BufferM[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPLongInt(RecAddr)^ := LongInt(BufferM[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPLongInt(RecAddr)^ := LongInt(BufferA[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPLongInt(RecAddr)^ := LongInt(BufferA[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferM[c, i] := TPLongInt(RecAddr)^;
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferM[0, i] := TPLongInt(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferA[i] := TPLongInt(RecAddr)^;
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferA[i] := TPLongInt(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

    NODE_DATATYPE_SINGLE: //----- SINGLE precision reals - added 10/2004 - JB
    //TO DO: this here is not the proper use for it. LM
    //I added a separate routine for floats, used by Extract.Dat.Float
    
      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPLongInt(RecAddr)^ := LongInt(BufferM[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPLongInt(RecAddr)^ := LongInt(BufferM[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPLongInt(RecAddr)^ := LongInt(BufferA[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPLongInt(RecAddr)^ := LongInt(BufferA[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferM[c, i] := Round(nMult*(TPSingle(RecAddr)^));
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferM[0, i] := Round(nMult*(TPSingle(RecAddr)^));
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferA[i] := Round(nMult*(TPSingle(RecAddr)^));
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferA[i] := Round(nMult*(TPSingle(RecAddr)^));
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

  end;//case DataType

  except
    Result := True;
  end;

end;

function MPXDataX(Buffer, BufferMPX :Pointer; ChannelCount, Channel :Integer; DataSize, DataSizeMPX :Integer; DataType, DataTypeMPX :TQUB_QFS_Node_DataType; DataFirst, DataLast :Integer; MPX :Boolean; Scaling :Double) :Boolean;
var
  BufferMSmallInt  :TPMSmallInt;
  BufferMLongInt   :TPMLongInt;
  BufferMSingle    :TPMSingle;
  BufferASmallInt  :TPASmallInt;
  BufferALongInt   :TPALongInt;
  BufferASingle    :TPASingle;
  RecAddr          :TPByte;
  //c, i             :Integer;

  //Integer = Scaling * Float
  //Float = (1 / Scaling) * Integer

  //MPX All Many
  procedure MPX_All_Many_SMALLINT_SMALLINT;
  var c, i :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPSmallInt(RecAddr)^ := (BufferMSmallInt[c, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_Many_SMALLINT_LONGINT;
  var c, i :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPLongInt(RecAddr)^ := (BufferMSmallInt[c, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_Many_SMALLINT_SINGLE;
  var c, i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPSingle(RecAddr)^ := s*(BufferMSmallInt[c, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_All_Many_LONGINT_SMALLINT;
  var c, i :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPSmallInt(RecAddr)^ := (BufferMLongInt[c, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_Many_LONGINT_LONGINT;
  var c, i :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPLongInt(RecAddr)^ := (BufferMLongInt[c, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_Many_LONGINT_SINGLE;
  var c, i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPSingle(RecAddr)^ := s*(BufferMLongInt[c, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_All_Many_SINGLE_SMALLINT;
  var c, i :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPSmallInt(RecAddr)^ := Round(Scaling*(BufferMSingle[c, i])); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_Many_SINGLE_LONGINT;
  var c, i :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPLongInt(RecAddr)^ := Round(Scaling*(BufferMSingle[c, i])); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_Many_SINGLE_SINGLE;
  var c, i :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      TPSingle(RecAddr)^ := (BufferMSingle[c, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;

  //MPX All One
  procedure MPX_All_One_SMALLINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSmallInt(RecAddr)^ := (BufferMSmallInt[0, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_One_SMALLINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPLongInt(RecAddr)^ := (BufferMSmallInt[0, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_One_SMALLINT_SINGLE;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      TPSingle(RecAddr)^ := s*(BufferMSmallInt[0, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_All_One_LONGINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSmallInt(RecAddr)^ := (BufferMLongInt[0, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_One_LONGINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPLongInt(RecAddr)^ := (BufferMLongInt[0, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_One_LONGINT_SINGLE;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      TPSingle(RecAddr)^ := s*(BufferMLongInt[0, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_All_One_SINGLE_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSmallInt(RecAddr)^ := Round(Scaling*(BufferMSingle[0, i])); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_One_SINGLE_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPLongInt(RecAddr)^ := Round(Scaling*(BufferMSingle[0, i])); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_All_One_SINGLE_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSingle(RecAddr)^ := (BufferMSingle[0, i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;

  //MPX One Many
  procedure MPX_One_Many_SMALLINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPSmallInt(RecAddr)^ := (BufferASmallInt[i]); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure MPX_One_Many_SMALLINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPLongInt(RecAddr)^ := (BufferASmallInt[i]); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure MPX_One_Many_SMALLINT_SINGLE;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPSingle(RecAddr)^ := s*(BufferASmallInt[i]); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_One_Many_LONGINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPSmallInt(RecAddr)^ := (BufferALongInt[i]); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure MPX_One_Many_LONGINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPLongInt(RecAddr)^ := (BufferALongInt[i]); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure MPX_One_Many_LONGINT_SINGLE;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPSingle(RecAddr)^ := s*(BufferALongInt[i]); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_One_Many_SINGLE_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPSmallInt(RecAddr)^ := Round(Scaling*(BufferASingle[i])); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure MPX_One_Many_SINGLE_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPLongInt(RecAddr)^ := Round(Scaling*(BufferASingle[i])); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure MPX_One_Many_SINGLE_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      TPSingle(RecAddr)^ := (BufferASingle[i]); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;

  //MPX One One
  procedure MPX_One_One_SMALLINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSmallInt(RecAddr)^ := (BufferASmallInt[i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_One_One_SMALLINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPLongInt(RecAddr)^ := (BufferASmallInt[i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_One_One_SMALLINT_SINGLE;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      TPSingle(RecAddr)^ := s*(BufferASmallInt[i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_One_One_LONGINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSmallInt(RecAddr)^ := (BufferALongInt[i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_One_One_LONGINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPLongInt(RecAddr)^ := (BufferALongInt[i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_One_One_LONGINT_SINGLE;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      TPSingle(RecAddr)^ := s*(BufferALongInt[i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure MPX_One_One_SINGLE_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSmallInt(RecAddr)^ := Round(Scaling*(BufferASingle[i])); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_One_One_SINGLE_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPLongInt(RecAddr)^ := Round(Scaling*(BufferASingle[i])); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure MPX_One_One_SINGLE_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      TPSingle(RecAddr)^ := (BufferASingle[i]); Inc(RecAddr, DataSizeMPX);
    end;
  end;

  //De-MPX

  //All Many
  procedure DeMPX_All_Many_SMALLINT_SMALLINT;
  var i, c :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMSmallInt[c, i] := TPSmallInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_Many_SMALLINT_LONGINT;
  var i, c :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMSmallInt[c, i] := TPLongInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_Many_SMALLINT_SINGLE;
  var i, c :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMSmallInt[c, i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_All_Many_LONGINT_SMALLINT;
  var i, c :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMLongInt[c, i] := TPSmallInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_Many_LONGINT_LONGINT;
  var i, c :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMLongInt[c, i] := TPLongInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_Many_LONGINT_SINGLE;
  var i, c :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMLongInt[c, i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_All_Many_SINGLE_SMALLINT;
  var i, c :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMSingle[c, i] := s*(TPSmallInt(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_Many_SINGLE_LONGINT;
  var i, c :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMSingle[c, i] := s*(TPLongInt(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_Many_SINGLE_SINGLE;
  var i, c :Integer; begin
    for i := DataFirst to DataLast do for c := 0 to ChannelCount - 1 do begin
      BufferMSingle[c, i] := (TPSingle(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;

  //All one
  procedure DeMPX_All_One_SMALLINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferMSmallInt[0, i] := TPSmallInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_One_SMALLINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferMSmallInt[0, i] := TPLongInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_One_SMALLINT_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferMSmallInt[0, i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_All_One_LONGINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferMLongInt[0, i] := TPSmallInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_One_LONGINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferMLongInt[0, i] := TPLongInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_One_LONGINT_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferMLongInt[0, i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_All_One_SINGLE_SMALLINT;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      BufferMSingle[0, i] := s*(TPSmallInt(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_One_SINGLE_LONGINT;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      BufferMSingle[0, i] := s*(TPLongInt(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_All_One_SINGLE_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferMSingle[0, i] := TPSingle(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;

  //One Many
  procedure DeMPX_One_Many_SMALLINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferASmallInt[i] := TPSmallInt(RecAddr)^; Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_Many_SMALLINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferASmallInt[i] := TPLongInt(RecAddr)^; Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_Many_SMALLINT_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferASmallInt[i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_One_Many_LONGINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferALongInt[i] := TPSmallInt(RecAddr)^; Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_Many_LONGINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferALongInt[i] := TPLongInt(RecAddr)^; Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_Many_LONGINT_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferALongInt[i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_One_Many_SINGLE_SMALLINT;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferASingle[i] := s*(TPSmallInt(RecAddr)^); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_Many_SINGLE_LONGINT;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferASingle[i] := s*(TPLongInt(RecAddr)^); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_Many_SINGLE_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin Inc(RecAddr, Channel * DataSizeMPX);
      BufferASingle[i] := (TPSingle(RecAddr)^); Inc(RecAddr, (ChannelCount - Channel) * DataSizeMPX);
    end;
  end;

  //One One
  procedure DeMPX_One_One_SMALLINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferASmallInt[i] := TPSmallInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_One_SMALLINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferASmallInt[i] := TPLongInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_One_SMALLINT_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferASmallInt[i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_One_One_LONGINT_SMALLINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferALongInt[i] := TPSmallInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_One_LONGINT_LONGINT;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferALongInt[i] := TPLongInt(RecAddr)^; Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_One_LONGINT_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferALongInt[i] := Round(Scaling*(TPSingle(RecAddr)^)); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  //****************************************************************************
  procedure DeMPX_One_One_SINGLE_SMALLINT;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      BufferASingle[i] := s*(TPSmallInt(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_One_SINGLE_LONGINT;
  var i :Integer; s :Double; begin s := 1 / Scaling;
    for i := DataFirst to DataLast do begin
      BufferASingle[i] := s*(TPLongInt(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;
  procedure DeMPX_One_One_SINGLE_SINGLE;
  var i :Integer; begin
    for i := DataFirst to DataLast do begin
      BufferASingle[i] := (TPSingle(RecAddr)^); Inc(RecAddr, DataSizeMPX);
    end;
  end;
     
begin

  case DataType of
    NODE_DATATYPE_SMALLINT,
    NODE_DATATYPE_LONGINT,
    NODE_DATATYPE_SINGLE:
      Result := False;
    else
      Result := True;
  end;
  if Result = True then Exit;
  case DataTypeMPX of
    NODE_DATATYPE_SMALLINT,
    NODE_DATATYPE_LONGINT,
    NODE_DATATYPE_SINGLE:
      Result := False;
    else
      Result := True;
  end;
  if Result = True then Exit;


  BufferMSmallInt  := Buffer;
  BufferMLongInt   := Buffer;
  BufferMSingle    := Buffer;
  BufferASmallInt  := Buffer;
  BufferALongInt   := Buffer;
  BufferASingle    := Buffer;

  //i-th record's address = pointer to an array of bytes
  RecAddr          := BufferMPX;

  try

    if MPX then begin
    //MPX

      if Channel = -1 then begin
      //All Channels

        if ChannelCount > 1 then begin
        //More than one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_All_Many_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_All_Many_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_All_Many_SMALLINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_All_Many_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_All_Many_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_All_Many_LONGINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_All_Many_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_All_Many_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_All_Many_SINGLE_SINGLE;
              end;//case DataTypeMPX
          end;//case DataType

        end else begin
        //Only one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_All_One_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_All_One_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_All_One_SMALLINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_All_One_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_All_One_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_All_One_LONGINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_All_One_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_All_One_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_All_One_SINGLE_SINGLE;
              end;//case DataTypeMPX
          end;//case DataType

        end;

      end else begin
      //One Channel

        if ChannelCount > 1 then begin
        //More than one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_One_Many_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_One_Many_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_One_Many_SMALLINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_One_Many_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_One_Many_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_One_Many_LONGINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_One_Many_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_One_Many_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_One_Many_SINGLE_SINGLE;
              end;//case DataTypeMPX
          end;//case DataType

        end else begin
        //Only one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_One_One_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_One_One_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_One_One_SMALLINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_One_One_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_One_One_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_One_One_LONGINT_SINGLE;
              end;//case DataTypeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  MPX_One_One_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  MPX_One_One_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  MPX_One_One_SINGLE_SINGLE;
              end;//case DataTypeMPX
          end;//case DataType

        end;

      end;//of Channel All / One

    end else begin
    //De-MPX

      if Channel = -1 then begin
      //All Channels

        if ChannelCount > 1 then begin
        //More than one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_All_Many_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_All_Many_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_All_Many_SMALLINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_All_Many_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_All_Many_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_All_Many_LONGINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_All_Many_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_All_Many_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_All_Many_SINGLE_SINGLE;
              end;//case DataTypeDeMPX
          end;//case DataType

        end else begin
        //Only one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_All_One_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_All_One_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_All_One_SMALLINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_All_One_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_All_One_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_All_One_LONGINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_All_One_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_All_One_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_All_One_SINGLE_SINGLE;
              end;//case DataTypeDeMPX
          end;//case DataType

        end;

      end else begin
      //One Channel

        if ChannelCount > 1 then begin
        //More than one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_One_Many_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_One_Many_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_One_Many_SMALLINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_One_Many_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_One_Many_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_One_Many_LONGINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_One_Many_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_One_Many_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_One_Many_SINGLE_SINGLE;
              end;//case DataTypeDeMPX
          end;//case DataType

        end else begin
        //Only one channel

          case DataType of
            NODE_DATATYPE_SMALLINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_One_One_SMALLINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_One_One_SMALLINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_One_One_SMALLINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_LONGINT:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_One_One_LONGINT_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_One_One_LONGINT_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_One_One_LONGINT_SINGLE;
              end;//case DataTypeDeMPX
            NODE_DATATYPE_SINGLE:
              case DataTypeMPX of
                NODE_DATATYPE_SMALLINT:
                  DeMPX_One_One_SINGLE_SMALLINT;
                NODE_DATATYPE_LONGINT:
                  DeMPX_One_One_SINGLE_LONGINT;
                NODE_DATATYPE_SINGLE:
                  DeMPX_One_One_SINGLE_SINGLE;
              end;//case DataTypeDeMPX
          end;//case DataType

        end;

      end;//of Channel All / One

    end;//MPX - De-MPX

  except
    Result := True;
  end;

end;

function MPXDataFloat(Buffer, BufferMPX :Pointer; ChannelCount, Channel :Integer; DataSize :Integer; DataType :TQUB_QFS_Node_DataType; DataFirst, DataLast :Integer; MPX :Boolean; nMult :Double) :Boolean;
var
  BufferMF :TPMSingle;
  BufferAF :TPASingle;
  RecAddr  :TPByte;
  c, i     :Integer;
begin

  Result   := False;

  BufferMF := Buffer;
  BufferAF := Buffer;

  //i-th record's address = pointer to an array of bytes
  RecAddr := BufferMPX;

  try

  case DataType of

    NODE_DATATYPE_SINGLE:

      if MPX then begin
      //MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                TPSingle(RecAddr)^ := Single(BufferMF[c, i]);
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPSingle(RecAddr)^ := Single(BufferMF[0, i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              TPSingle(RecAddr)^ := Single(BufferAF[i]);
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              TPSingle(RecAddr)^ := Single(BufferAF[i]);
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end else begin
      //De-MPX

        if Channel = -1 then begin
        //All Channels

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              for c := 0 to ChannelCount - 1 do begin
                BufferMF[c, i] := TPSingle(RecAddr)^;
                Inc(RecAddr, DataSize);
              end;
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferMF[0, i] := TPSingle(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end else begin
        //One Channel

          if ChannelCount > 1 then begin
            for i := DataFirst to DataLast do begin
              Inc(RecAddr, Channel * DataSize);
              BufferAF[i] := TPSingle(RecAddr)^;
              Inc(RecAddr, (ChannelCount - Channel) * DataSize);
            end;
          end else begin
            for i := DataFirst to DataLast do begin
              BufferAF[i] := TPSingle(RecAddr)^;
              Inc(RecAddr, DataSize);
            end;
          end;

        end;//of Channel All / One

      end;

  end;//case DataType

  except
    Result := True;
  end;

end;


//******************************************************************************
end.
