{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_QFS_QDF_ScalingPanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, ExtCtrls, Forms, Dialogs,
  StdCtrls;

type


TQUB_QFS_QDF_ScalingDetails = class(TCustomControl)
public
  txtMax              :TEdit;
  txtRes              :TEdit;
  btnMin              :TButton;
  btnClose            :TButton;

  constructor Create(AOwner :TComponent); override;
end;

TQUB_QFS_QDF_ScalingPanel = class(TFrame)
  Label1              :TLabel;
  Edit                :TEdit;
  procedure EditKeyPress(Sender: TObject; var Key: Char);
  procedure EditExit(Sender: TObject);
  procedure EditEnter(Sender: TObject);
private
  prMaxInt            :Integer;
  prScaling           :Double;

  frmDetails          :TQUB_QFS_QDF_ScalingDetails;

  procedure prSetMaxInt(x :Integer);
  procedure prSetScaling(x :Double);
  function prGetDetails: TControl;
  procedure MaxExit(Sender :TObject);
  procedure ResExit(Sender :TObject);
  procedure MinPress(Sender :TObject);
  procedure ClosePress(Sender :TObject);

public
  OnScalingChange     :TNotifyEvent;
  OnEditing           :TNotifyEvent;
  OnMinimize          :TNotifyEvent;
  OnClose             :TNotifyEvent;

  constructor Create(AOwner :TComponent); override;

  property MaxInt     :Integer      read prMaxInt         write prSetMaxInt;
  property Scaling    :Double       read prScaling        write prSetScaling;
  property Details    :TControl     read prGetDetails;
end;


implementation

{$R *.DFM}

constructor TQUB_QFS_QDF_ScalingDetails.Create(AOwner :TComponent);
var
  P                   :TPanel;
  L                   :TLabel;
begin
  inherited;
  Width               := 119;
  Height              := 54;
  P                   := TPanel.Create(Self);
  P.SetBounds(0, 0, 119, 54);
  P.Anchors           := [akLeft, akRight, akTop, akBottom];
  P.Parent            := Self;
  P.BevelOuter        := bvRaised;
  P.BevelInner        := bvLowered;
  P.Color             := TColor($CCCCFF);
  btnMin              := TButton.Create(Self);
  btnMin.Font.Name    := 'Arial Black';
  btnMin.Font.Size    := 4;
  btnMin.Caption      := '__';
  btnMin.SetBounds(78, 2, 20, 10);
  btnMin.Anchors      := [akTop, akRight];
  btnMin.Parent       := P;
  btnMin.Hint         := 'Hide the scaling calculator (double-click scaling to bring it back)';
  btnMin.ShowHint     := True;
  btnClose            := TButton.Create(Self);
  btnClose.Font.Name  := 'Arial Black';
  btnClose.Font.Size  := 9;
  btnClose.Font.Color := clRed;
  btnClose.Caption    := 'x';
  btnClose.SetBounds(98, 2, 20, 10);
  btnClose.Anchors    := [akTop, akRight];
  btnClose.Parent     := P;
  btnClose.Hint       := 'Close the scaling calculator';
  btnClose.ShowHint   := True;
  L                   := TLabel.Create(Self);
  L.Caption           := 'Max val:';
  L.Parent            := P;
  L.Left              := 2;
  L.Top               := 15;
  L                   := TLabel.Create(Self);
  L.Caption           := 'Res.:';
  L.Parent            := P;
  L.Left              := 2;
  L.Top               := 35;
  txtMax              := TEdit.Create(Self);
  txtMax.Parent       := P;
  txtMax.SetBounds(52, 12, 65, 20);
  txtMax.Anchors      := [akLeft, akTop, akRight];
  txtRes              := TEdit.Create(Self);
  txtRes.Parent       := P;
  txtRes.SetBounds(52, 32, 65, 20);
  txtRes.Anchors      := [akLeft, akTop, akRight];
end;

constructor TQUB_QFS_QDF_ScalingPanel.Create(AOwner :TComponent);
begin
  inherited;

  prMaxInt            := High(SmallInt);
  prScaling           := 1;
  frmDetails          := TQUB_QFS_QDF_ScalingDetails.Create(Self);
  frmDetails.txtMax.OnKeyPress := EditKeyPress;
  frmDetails.txtRes.OnKeyPress := EditKeyPress;
  frmDetails.txtMax.OnExit     := MaxExit;
  frmDetails.txtRes.OnExit     := ResExit;
  frmDetails.btnMin.OnClick    := MinPress;
  frmDetails.btnClose.OnClick  := ClosePress;
end;

procedure TQUB_QFS_QDF_ScalingPanel.prSetMaxInt(x :Integer);
begin
  prMaxInt     := x;
  Scaling      := Scaling;
end;

procedure TQUB_QFS_QDF_ScalingPanel.prSetScaling(x :Double);
begin
  prScaling    := x;
  Edit.Text    := FloatToStr(x);
  frmDetails.txtMax.Text := Format('%.6g', [prMaxInt / x]);
  frmDetails.txtRes.Text := Format('%.6g', [1.0 / x]);
  if Assigned(OnScalingChange) then OnScalingChange(Self);
end;

function TQUB_QFS_QDF_ScalingPanel.prGetDetails: TControl;
begin
  Result := frmDetails;
end;

procedure TQUB_QFS_QDF_ScalingPanel.MinPress(Sender :TObject);
begin
  if Assigned(OnMinimize) then OnMinimize(Self);
end;

procedure TQUB_QFS_QDF_ScalingPanel.ClosePress(Sender :TObject);
begin
  if Assigned(OnClose) then OnClose(Self);
end;

procedure TQUB_QFS_QDF_ScalingPanel.EditEnter(Sender: TObject);
begin
  if Assigned(OnEditing) then OnEditing(Self);
end;

procedure TQUB_QFS_QDF_ScalingPanel.EditKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    if Assigned(TEdit(Sender).OnExit) then TEdit(Sender).OnExit(Sender);
  end else begin
    TEdit(Sender).Font.Color := clRed;
  end;
end;

procedure TQUB_QFS_QDF_ScalingPanel.EditExit(Sender: TObject);
var
  s       :Double;
begin
  TEdit(Sender).Font.Color := clWindowText;
  try
    s := StrToFloat(TEdit(Sender).Text);
    if s <= 0.0 then raise Exception.Create('');
    Scaling := s;
  except
    MessageDlg('Invalid value; must be positive.', mtError, [mbOK], 0);
    Scaling := Scaling;
    TEdit(Sender).SetFocus;
    Exit;
  end;
end;

procedure TQUB_QFS_QDF_ScalingPanel.MaxExit(Sender :TObject);
var
  s       :Double;
begin
  TEdit(Sender).Font.Color := clWindowText;
  try
    s := StrToFloat(TEdit(Sender).Text);
    if s <= 0.0 then raise Exception.Create('');
    Scaling := prMaxInt / s;
  except
    MessageDlg('Invalid value; must be positive.', mtError, [mbOK], 0);
    Scaling := Scaling;
    TEdit(Sender).SetFocus;
    Exit;
  end;
end;

procedure TQUB_QFS_QDF_ScalingPanel.ResExit(Sender :TObject);
var
  s       :Double;
begin
  TEdit(Sender).Font.Color := clWindowText;
  try
    s := StrToFloat(TEdit(Sender).Text);
    if s <= 0.0 then raise Exception.Create('');
    Scaling := 1.0 / s;
  except
    MessageDlg('Invalid value; must be positive.', mtError, [mbOK], 0);
    Scaling := Scaling;
    TEdit(Sender).SetFocus;
    Exit;
  end;
end;


// *****************************************************************************
end.
