{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// used to be the utilities for QUB_QFS_QIF, which was the predecessor to QUB_Idealization.
// now it's just the idealization-related data types that are still in use.


unit QUB_QFS_QIF_Segments;


interface


uses


SysUtils, Classes,

Alloc,

QUB_Global,
QUB_QFS, QUB_QFS_Edit;


const


QUB_QFS_QDF_QIF_IDLDC_NAME                     = 'DwellsClass';
QUB_QFS_QDF_QIF_IDLFLC_NAME                    = 'FirstLastClass';
QUB_QFS_QDF_QIF_AMPSTD_NAME                    = 'AmpStd';

QUB_QFS_QDF_QIF_SEGMENTS_NAME                  = 'Segments Info';
QUB_QFS_QDF_QIF_SEGMENT_NAME                   = 'Segment Info';


type


TQUB_QFS_QIF_DwellClass                        = record
  Cls                                          :Integer;
  Dwell                                        :Single;
end;
TQUB_QFS_QIF_ADwellClass                       = Array[0..(MaxInt div (SizeOf(TQUB_QFS_QIF_DwellClass)*2))] of TQUB_QFS_QIF_DwellClass;
TQUB_QFS_QIF_PADwellClass                      = ^TQUB_QFS_QIF_ADwellClass;

TQUB_QFS_QIF_FirstLastClass                    = record
  Cls                                          :Integer;
  First                                        :Integer;
  Last                                         :Integer;
end;
TQUB_QFS_QIF_PFirstLastClass                   = ^TQUB_QFS_QIF_FirstLastClass;
TQUB_QFS_QIF_AFirstLastClass                   = Array[0..(MaxInt div (SizeOf(TQUB_QFS_QIF_FirstLastClass)*2))] of TQUB_QFS_QIF_FirstLastClass;
TQUB_QFS_QIF_PAFirstLastClass                  = ^TQUB_QFS_QIF_AFirstLastClass;

TQUB_QFS_QIF_NClass                            = record
  Cls                                          :Integer;
  N                                            :Integer;
end;
TQUB_QFS_QIF_PNClass                           = ^TQUB_QFS_QIF_NClass;
TQUB_QFS_QIF_ANClass                           = Array[0..(MaxInt div (SizeOf(TQUB_QFS_QIF_NClass)*2))] of TQUB_QFS_QIF_NClass;
TQUB_QFS_QIF_PANClass                          = ^TQUB_QFS_QIF_ANClass;

TQUB_QFS_QIF_AmpStdRec                         = record
  Amp                                          :Single;
  Std                                          :Single;
end;
TQUB_QFS_QIF_AAmpStdRec                        = Array[0..(MaxInt div (SizeOf(TQUB_QFS_QIF_AmpStdRec)*2))] of TQUB_QFS_QIF_AmpStdRec;
TQUB_QFS_QIF_PAAmpStdRec                       = ^TQUB_QFS_QIF_AAmpStdRec ;


//Forward class declarations ***************************************************

{

TQUB_QFS_QIF_Segments                          = class;
TQUB_QFS_QIF_Segment                           = class;
TQUB_QFS_QIF_IdlDC                             = class;
TQUB_QFS_QIF_IdlFLC                            = class;
TQUB_QFS_QIF_AmpStd                            = class;


//Class definitions

TQUB_QFS_QIF_Segments                          = class(TQUB_QFS_Record)
protected


  function prGetSegment(s :Integer) :TQUB_QFS_QIF_Segment;


public


  procedure InsertSegment(Segment :TQUB_QFS_QIF_Segment; SegmentIndex :Integer);
  procedure RemoveSegment(SegmentIndex :Integer);

  function TotDwells :Integer;
  function MaxClass :Integer;


  property Segment[Index :Integer]             :TQUB_QFS_QIF_Segment read prGetSegment; default;
  property SegmentCount                        :Integer              read prRecordCount;


end;


TQUB_QFS_QIF_Segment                           = class(TQUB_QFS_Record)
protected


  prSel                                        :TQUB_FirstLast;
  prStartTime                                  :Double;

  prIdlDC                                      :TQUB_QFS_QIF_IdlDC;
  prIdlFLC                                     :TQUB_QFS_QIF_IdlFLC;
  prAmpStd                                     :TQUB_QFS_QIF_AmpStd;

  prIdlReport                                  :String;
  prTBL2                                       :String;

  prDwellCount                                 :Integer;
  prMaxClass                                   :Integer;


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;


  property Sel                                 :TQUB_FirstLast      read prSel        write prSel;
  property StartTime                           :Double              read prStartTime  write prStartTime;

  property IdlDC                               :TQUB_QFS_QIF_IdlDC  read prIdlDC;
  property IdlFLC                              :TQUB_QFS_QIF_IdlFLC read prIdlFLC;
  property AmpStd                              :TQUB_QFS_QIF_AmpStd read prAmpStd;

  property IdlReport                           :String              read prIdlReport  write prIdlReport;
  property TBL2                                :String              read prTBL2       write prTBL2;

  property DwellCount                          :Integer             read prDwellCount write prDwellCount;
  property MaxClass                            :Integer             read prMaxClass   write prMaxClass;


end;


TQUB_QFS_QIF_IdlDC                             = class(TQUB_QFS_Edit)
end;

TQUB_QFS_QIF_IdlFLC                            = class(TQUB_QFS_Edit)
end;

TQUB_QFS_QIF_AmpStd                            = class(TQUB_QFS_Edit)
end;
}

implementation

{
//******************************************************************************
procedure TQUB_QFS_QIF_Segments.InsertSegment(Segment :TQUB_QFS_QIF_Segment; SegmentIndex :Integer);
begin
  AddRecord(Segment, SegmentIndex);
end;

procedure TQUB_QFS_QIF_Segments.RemoveSegment(SegmentIndex :Integer);
begin
  DelRecord(Segment[SegmentIndex]);
end;

function TQUB_QFS_QIF_Segments.prGetSegment(s :Integer) :TQUB_QFS_QIF_Segment;
begin
  Result := TQUB_QFS_QIF_Segment(Records[s]);
end;

function TQUB_QFS_QIF_Segments.TotDwells :Integer;
var
  s :Integer;
begin
  Result := 0;
  for s := 0 to SegmentCount - 1 do begin
    Inc(Result, Segment[s].DwellCount);
  end;
end;

function TQUB_QFS_QIF_Segments.MaxClass :Integer;
var
  s :Integer;
begin
  Result := 0;
  for s := 0 to SegmentCount - 1 do begin
    if Segment[s].MaxClass > Result then Result := Segment[s].MaxClass;
  end;
end;


//******************************************************************************
constructor TQUB_QFS_QIF_Segment.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited Create(Name, Parent);

  prIdlDC     := TQUB_QFS_QIF_IdlDC.Create(QUB_QFS_QDF_QIF_IDLDC_NAME, Self);
  prIdlFLC    := TQUB_QFS_QIF_IdlFLC.Create(QUB_QFS_QDF_QIF_IDLFLC_NAME, Self);
  prAmpStd    := TQUB_QFS_QIF_AmpStd.Create(QUB_QFS_QDF_QIF_AMPSTD_NAME, Self);

  prStartTime := 0;

end;
}

//******************************************************************************
end.
