{

                                 Qub Model File
                                     unit



                        Interface unit for QUB Model files


(C) 2000 Lorin Milescu
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


//01-31-2000   Creation date


unit QUB_QFS_QMF;


//******************************************************************************
interface


uses


Windows, SysUtils, Classes, Forms, Controls, ExtCtrls, Graphics, Dialogs, Math,

AllocTypes, Alloc,

QUB_Global, QUB_Globals,

//QUB_Eigen,
QUB_Deviates, QUB_Matrices,

QUB_Console, QUB_Rich_Console,

QUB_QFS, QUB_QFS_File, QUB_QFS_Edit, QUB_QFS_Channels,
QUB_Node,
QUB_Parser,
QUB_ExpCond_Parameters,
QUB_Profiles;


//******************************************************************************
const


QUB_QFS_QMF_MAX_COPYSIZE                       = ONE_MEGABYTE;


//******************************************************************************
type


TQUB_QFS_QMF_FileType = (QFS_QMF_TYPE_QMF, QFS_QMF_TYPE_MDL, QFS_QMF_TYPE_MOD, QFS_QMF_TYPE_UNK);


TQUB_QFS_QMF_PCopyInfo                         = ^TQUB_QFS_QMF_CopyInfo;
TQUB_QFS_QMF_CopyInfo                          = record
  ModelSize                                    :Integer;
  //Model                                        :TPAByte;
end;


TQUB_QFS_QMF_PGroup                            = ^TQUB_QFS_QMF_Group;
TQUB_QFS_QMF_Group                             = record
  Cls                                          :Integer;
  StateCount                                   :Integer;
  States                                       :TPAInteger;
  Lambdas                                      :TPADouble;
  Taus                                         :TPADouble;
  Areas                                        :TPADouble;
end;


//Class definitions


//QMF
TQUB_QFS_QMF                                   = class(TQUB_QFS_File)
protected

  //Unused?? 
  //TODO: I'm not sure what was intended for. Have to check. LM
  prCanvas                                     :TCanvas;

  prFileType                                   :TQUB_QFS_QMF_FileType;

  prStates                                     :TQUB_QFS_States;
  prRates                                      :TQUB_QFS_Rates;
  prConstraints                                :TQUB_QFS_Constraints;
  prConstraintsAmpVar                          :TQUB_QFS_ConstraintsAmpVar;
  prExtraKineticPars                           :TQUB_QFS_ExtraKineticPars;

  //prSimChannelCount                            :Integer;


  //Streamed properties **************************************************
  prRecChannelCount                            :TQUB_QFS_Record;
  prRecAmps                                    :TQUB_QFS_Record;
  prRecStds                                    :TQUB_QFS_Record;
  prRecConds                                   :TQUB_QFS_Record;
  prRecCondStds                                :TQUB_QFS_Record;
  prRecVrevs                                   :TQUB_QFS_Record;
  prRecNar                                     :TQUB_QFS_Record;
  prRecArs                                     :TQUB_QFS_Record;
  prRecVRev                                    :TQUB_QFS_Record;
  //****************************************************************************


  prOnNewState                                 :TNotifyEvent;
  prOnOpenState                                :TNotifyEvent;
  prOnNewRate                                  :TNotifyEvent;
  prOnOpenRate                                 :TNotifyEvent;
  prOnChangeActiveState                        :TNotifyEvent;
  prOnNeedDraw                                 :TNotifyEvent;


  prParameters                                 :TQUB_ExpCond_Parameters;


  prQ                                          :TPMDouble;
  prADt                                        :Double;
  prA                                          :TPMDouble;
  prP                                          :TPMDouble;
  prPt                                         :TPADouble;
  prP0                                         :TPADouble;
  prPEq                                        :TPADouble;
  prLambdas                                    :TPADouble;
  prAk                                         :TPAPointer;
  prMathStateCount                             :Integer;

  prFreePars                                   :TPADouble;
  prFreeParsCount                              :Integer;

  //Replace all the above with a QModel?
  //prQModel                                     :TObject;//TQUB_Q_Model;

  prProperties                                 :INode;

  //prGroupCount                                 :Integer;
  //prGroups                                     :TPAPointer;


  //Low level Read / Write
  function prReadQMF :Boolean;

  function prReadMOD :Boolean;
  function prReadMODParse(const MDL :TStringList) :Boolean;

  function prReadMDL :Boolean;
  function prReadMDLParse(const MDL :String) :Boolean;

  function prWriteMDLHeader(Stream :TStream) :Boolean;
  function prWriteMDLStates(Stream :TStream) :Boolean;
  function prWriteMDLRates(Stream :TStream) :Boolean;
  function prWriteMDLConstraints(Stream :TStream) :Boolean;
  function prWriteMDLCoefficients(Stream :TStream) :Boolean;

  //Low level File *************************************************************
  function prNew :Boolean; override;
  function prOpen(const FileName :String) :Boolean; override;
  function prSave(const FileName :String) :Boolean; override;
  function prSaveToString(var MemFile :String; FileType :TQUB_QFS_QMF_FileType) :Boolean;

  function prGetFileTypeFromFileName(const FileName :String) :TQUB_QFS_QMF_FileType;
  function prGetFileExt :String;

  //Low level Edit *************************************************************
  function prCutModel :Boolean;
  function prCopyModel :Boolean;
  function prPasteModel :Boolean;
  function prDeleteModel :Boolean;
  function prAddState(x, y :Double) :Boolean;
  function prDelState(State :TQUB_QFS_State) :Boolean;
  function prPosState(State :TQUB_QFS_State; xp, yp :Double; x, y :Integer) :Boolean;
  function prAddRate(State1, State2 :TQUB_QFS_State) :Boolean;
  function prDelRate(Rate :TQUB_QFS_Rate) :Boolean;

  //ChannelCount property
  procedure prSetChannelCount(c :Integer);
  function prGetChannelCount :Integer;

  //VRev property
  procedure prSetVRev(v :Double);
  function prGetVRev :Double;

  function prGetClassCount: Integer;

  //State properties
  procedure prEnsureClassExists(Index :Integer);
  procedure prSetAmp(Index :Integer; a :Double);
  function  prGetAmp(Index :Integer) :Double;
  procedure prSetStd(Index :Integer; s :Double);
  function  prGetStd(Index :Integer) :Double;
  procedure prSetCond(Index :Integer; c :Double);
  function  prGetCond(Index :Integer) :Double;
  procedure prSetCondStd(Index :Integer; x :Double);
  function  prGetCondStd(Index :Integer): Double;
  procedure prSetClsVrev(Index :Integer; x :Double);
  function  prGetClsVrev(Index :Integer): Double;
  function  prGetEffVrev(Index :Integer): Double;
  procedure prSetNar(Index :Integer; n :Integer);
  function  prGetNar(Index :Integer) :Integer;
  procedure prSetCff(Index1, Index2 :Integer; c :Double);
  function  prGetCff(Index1, Index2 :Integer) :Double;

  procedure prOnEdit(Sender :TObject);

  procedure DoOnChangeActiveState(Sender :TObject);

  procedure prClearMath;
  procedure prReAllocMath;

  function prMakeFreePars :Boolean;

  procedure prInitProperties;

  //procedure prClearGroups;
  //function prGetGroup(g :Integer) :TQUB_QFS_QMF_PGroup;


public


  OnInitProperties        :TNotifyEvent;
  OnSetProperties         :TQUB_Profile_OnSet;


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  destructor Destroy; override;


  procedure Reset; override;


  function New(const FileName :String) :Boolean; override;
  function Open(const FileName :String) :Boolean; override;
  function SaveToString(var MemFile :String; FileType :TQUB_QFS_QMF_FileType) :Boolean;

  function Read(ReadStream :TStream) :Boolean; override;

  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  function CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;

  //Edit
  function CutModel :Boolean;
  function CopyModel :Boolean;
  function PasteModel :Boolean;
  function DeleteModel :Boolean;
  function AddState(x, y :Double) :Boolean;
  function DelState(State :TQUB_QFS_State) :Boolean;
  function PosState(State :TQUB_QFS_State; xp, yp :Double; x, y :Integer) :Boolean;
  function AddRate(State1, State2 :TQUB_QFS_State) :Boolean;
  function DelRate(Rate :TQUB_QFS_Rate) :Boolean;

  function CopyFrom(Source :TQUB_QFS_QMF) :Boolean;

  procedure SwapStates(iOldState : Integer; iNewState : Integer);

  procedure UnModify;

  procedure Delete;

  function Connected(State1, State2 :TQUB_QFS_State) :Boolean;
  function RateFixed(State1, State2 :Integer) :Boolean;

  procedure ChangeStateClass(State :TQUB_QFS_State; c :Integer);

  //Simulation
  function GenerateState :Integer; overload;
  function GenerateDwell(State :Integer) :Double;
  function GenerateState(State :Integer) :Integer; overload;
  procedure GenerateP(FromP, ToP :TPADouble; StateCount, ChannelCount :Integer);

  //Parameters
  procedure DoOnParametersChanged(Sender :TObject);
  procedure UpdateParameters;
  function UsesParameter(Par :TQUB_ExpCond_PParameter) :Boolean;
  function SameParameters(Params1, Params2 :TQUB_ExpCond_Parameters) :Boolean;

  //Calculate Q, A, P, P0, PEq, Ak and Lambdas
  function CalculateMath(NeedPars :Boolean = True): Boolean;
  function CalculateProbEq(DoAssign :Boolean) :Boolean;
  function AdvanceProb(t :Double) :Boolean;

  function MakeStartState(StateIdx :Integer) :Boolean;

  function EnforceConstraints :Boolean;
  function RandomizeRates(SE :Double) :Boolean;
  function ConstrainedSetRate(StateFrom, StateTo :Integer; IsK0 :Boolean; Value :Double): Boolean;

  //function FindGroups :Boolean;

  //Conductance
  function Conductance :Double;


  //Properties
  procedure OnGetPropertiesNew(Sender :TObject); override;
  procedure OnPutPropertiesNew(Sender :TObject); override;


  procedure DoOnSetProperties(Sender :TQUB_ProfileCat; Updates :INode);


  property Canvas                              :TCanvas                    read prCanvas              write prCanvas;
  property States                              :TQUB_QFS_States            read prStates;
  property Rates                               :TQUB_QFS_Rates             read prRates;
  property Constraints                         :TQUB_QFS_Constraints       read prConstraints;
  property ConstraintsAmpVar                   :TQUB_QFS_ConstraintsAmpVar read prConstraintsAmpVar;
  property ExtraKineticPars                    :TQUB_QFS_ExtraKineticPars  read prExtraKineticPars;

  //Properties
  property FileType                            :TQUB_QFS_QMF_FileType      read prFileType;
  property FileExt                             :String                     read prGetFileExt;

  //Can's ......................................................................
  property CanEdit                             :Boolean                    read prCanEdit             write prSetCanEdit;

  property Amp[Index :Integer]                 :Double                     read prGetAmp              write prSetAmp;
  property Std[Index :Integer]                 :Double                     read prGetStd              write prSetStd;
  property Cond[Index :Integer]                :Double                     read prGetCond             write prSetCond;
  property CondStd[Index :Integer]             :Double                     read prGetCondStd          write prSetCondStd;
  property ClsVrev[Index :Integer]             :Double                     read prGetClsVrev          write prSetClsVrev;
  property EffVrev[Index :Integer]             :Double                     read prGetEffVrev; // effective Vrev: ClsVrev[i] if set, else Vrev.
  property Nar[Index :Integer]                 :Integer                    read prGetNar              write prSetNar;
  property Cff[Index1, Index2 :Integer]        :Double                     read prGetCff              write prSetCff;

  property VRev                                :Double                     read prGetVRev             write prSetVRev;

  property ChannelCount                        :Integer                    read prGetChannelCount     write prSetChannelCount;
  property ClassCount                          :Integer                    read prGetClassCount;

  property OnNewState                          :TNotifyEvent               read prOnNewState          write prOnNewState;
  property OnOpenState                         :TNotifyEvent               read prOnOpenState         write prOnOpenState;
  property OnNewRate                           :TNotifyEvent               read prOnNewRate           write prOnNewRate;
  property OnOpenRate                          :TNotifyEvent               read prOnOpenRate          write prOnOpenRate;
  property OnChangeActiveState                 :TNotifyEvent               read prOnChangeActiveState write prOnChangeActiveState;
  property OnNeedDraw                          :TNotifyEvent               read prOnNeedDraw          write prOnNeedDraw;

  property Parameters                          :TQUB_ExpCond_Parameters    read prParameters          write prParameters;

  property Q                                   :TPMDouble                  read prQ;
  property ADt                                 :Double                     read prADt                 write prADt;
  property A                                   :TPMDouble                  read prA;
  property P                                   :TPMDouble                  read prP;
  property Pt                                  :TPADouble                  read prPt;
  property P0                                  :TPADouble                  read prP0;
  property PEq                                 :TPADouble                  read prPEq;
  property Lambdas                             :TPADouble                  read prLambdas;
  property Ak                                  :TPAPointer                 read prAk;
  property MathStateCount                      :Integer                    read prMathStateCount;
  property FreePars                            :TPADouble                  read prFreePars;
  property FreeParsCount                       :Integer                    read prFreeParsCount;

  property Properties                          :INode                      read prProperties;

  //property GroupCount                          :Integer                    read prGroupCount;
  //property Groups[Index :Integer]              :TQUB_QFS_QMF_PGroup        read prGetGroup;


end;
//..............................................................................


//******************************************************************************
implementation


uses


FMath,
qubopt_iface,
QUB_QFS_QMF_QMatrix,
QUB_Viewport_Model_New_Properties,
QUB_Q_Model,
QUB_GUIInterface,
QUB_Viewport_Model,
QUB_Node_Model;


//****************************************************************************//
//                                                                            //
//                              TQUB_QFS_QMF                                  //
//                                                                            //
//****************************************************************************//



//****************************************************************************//
//                                                                            //
//              c o n s t r u c t o r   /   d e s t r u c t o r               //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_QMF.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited;

  prParent := Parent;

  prProperties                  := QUB_Node_NULL;

  prStates                      := TQUB_QFS_States.Create(CHANNELS_STATES_NAME, Self);
  prStates.Permanent            := True;
  prRates                       := TQUB_QFS_Rates.Create(CHANNELS_RATES_NAME, Self);
  prRates.Permanent             := True;
  prConstraints                 := TQUB_QFS_Constraints.Create(CHANNELS_CONSTRAINTS_NAME, Self);
  prConstraints.Permanent       := True;
  prConstraintsAmpVar           := TQUB_QFS_ConstraintsAmpVar.Create(CHANNELS_CONSTRAINTS_AMPVAR_NAME, Self);
  prConstraintsAmpVar.Permanent := True;
  prExtraKineticPars            := TQUB_QFS_ExtraKineticPars.Create(CHANNELS_EXTRAKINETICPARS_NAME, Self);
  prExtraKineticPars.Permanent  := True;

  prStates.Rates := Rates;
  prRates.States := States;

  prStates.OnChangeActiveState := DoOnChangeActiveState;

  prConstraints.OnEdit       := prOnEdit;
  prExtraKineticPars.OnEdit  := prOnEdit;
  prConstraintsAmpVar.OnEdit := prOnEdit;

  prRecChannelCount  := TQUB_QFS_Record.CreateAsValue(CHANNELS_CHANNEL_COUNT_NAME, Self, True{Permanent}, 1, NODE_DATATYPE_LONGINT, 1);
  prRecAmps          := TQUB_QFS_Record.CreateAsValue(CHANNELS_AMPS_NAME,          Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  QUB_QFS_QMF_DEF_NCLASS);
  prRecStds          := TQUB_QFS_Record.CreateAsValue(CHANNELS_STDS_NAME,          Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  QUB_QFS_QMF_DEF_NCLASS);
  prRecConds         := TQUB_QFS_Record.CreateAsValue(CHANNELS_CONDS_NAME,         Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  QUB_QFS_QMF_DEF_NCLASS);
  prRecCondStds      := TQUB_QFS_Record.CreateAsValue(CHANNELS_CONDSTDS_NAME,      Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  QUB_QFS_QMF_DEF_NCLASS);
  prRecVrevs         := TQUB_QFS_Record.CreateAsValue(CHANNELS_VREVS_NAME,         Self, True{Permanant}, 0, NODE_DATATYPE_DOUBLE,  QUB_QFS_QMF_DEF_NCLASS);
  prRecNar           := TQUB_QFS_Record.CreateAsValue(CHANNELS_NAR_NAME,           Self, True{Permanent}, 0, NODE_DATATYPE_LONGINT, QUB_QFS_QMF_DEF_NCLASS);
  prRecArs           := TQUB_QFS_Record.Create(CHANNELS_ARS_NAME,  Self);
  prRecArs.Permanent := True;
  prRecVRev          := TQUB_QFS_Record.CreateAsValue(CHANNELS_VREV_NAME,          Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE,  1);

  prRecChannelCount.DataValue := 1;
  //prSimChannelCount := 1;

end;

destructor TQUB_QFS_QMF.Destroy;
begin
  try

    Close;

    prClearMath;
    //prClearGroups;

  finally

    inherited;

  end;
end;

procedure TQUB_QFS_QMF.Reset;
begin

  Delete;

  inherited Reset;
  
end;

procedure TQUB_QFS_QMF.UnModify;
begin
  prModified := False;
  if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
end;

procedure TQUB_QFS_QMF.DoOnChangeActiveState(Sender :TObject);
begin
  if not prEditing then begin
    if Assigned(OnChangeActiveState) then OnChangeActiveState(Sender);
  end;
end;


//****************************************************************************//
//                                                                            //
//                            F I L E    M E T H O D S                        //
//                                                                            //
//****************************************************************************//


//****************************************************************************//
//                                                                            //
//                                    N e w                                   //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QMF.New(const FileName :String) :Boolean;
begin
  Result := inherited New(FileName);

  UpdateParameters;

  prProperties := TNode.Create('Properties');
  prInitProperties;

  //May have to change this !!!
  if not Result then begin
    Stream.Free;
    Stream := nil;
  end;
end;

function TQUB_QFS_QMF.prNew :Boolean;
begin

  Result := False;

end;


//****************************************************************************//
//                                                                            //
//                                   O p e n                                  //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QMF.Open(const FileName :String) :Boolean;
begin

  Result := inherited Open(FileName);

  UpdateParameters;

  prProperties := CloneAsNode(True, TNode.CreateNull)['Properties'];
  prInitProperties;

  {
  //TODO: check this
  if not Result then begin
    Stream.Free;
    Stream := nil;
  end;
  }
end;

function TQUB_QFS_QMF.prOpen(const FileName :String) :Boolean;
var
  Ext :String;
begin

  Result := False;

  Ext := LowerCase(ExtractFileExt(FileName));

  if Ext = '.qmf' then begin
  //QMF
    prFileType := QFS_QMF_TYPE_QMF;

  end else if Ext = '.mdl' then begin
  //MDL
    prFileType := QFS_QMF_TYPE_MDL;

  end else if Ext = '.mod' then begin
  //MOD 
    prFileType := QFS_QMF_TYPE_MOD;

  end else begin
  //UNK
    prFileType := QFS_QMF_TYPE_UNK;
    Result := True;
    Console.NewEntry('File type unknown: ' + FileName, COLOR_ERROR);
    Exit;
  end;

end;


//****************************************************************************//
//                                                                            //
//                                   S a v e                                  //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QMF.prSave(const FileName :String) :Boolean;
var
  FStream   :TFileStream;
  FType     :TQUB_QFS_QMF_FileType;
  FName     :String;
  SName     :String;
  Entry     :Integer;
  AsQubtree :INode;
begin

  SName := FileName;
  FType := prGetFileTypeFromFileName(SName);
  FName := prGetFileNameForSave(ChangeFileExt(SName, ''), '.~s', False{IncFile}, 1000{MaxInc});
  Result := FName = '';
  if Result then Exit;

  FStream := nil;

  Entry := Console.NewEntry('Saving model file ' + SName + ' ...', COLOR_TEXT);

  try

    try

      prEditCritSection.Acquire;

      try
        FStream := TFileStream.Create(FName, fmCreate or fmOpenWrite or fmShareExclusive);
      except
        Result := True;
        Console.NewEntry('Could not create file ' + SName, COLOR_ERROR);
        Exit;
      end;

      if FType = QFS_QMF_TYPE_QMF then begin
      //Native type
        Result := Write(FStream);
      end else begin

        case FType of
          QFS_QMF_TYPE_MDL:
            begin
              Result := prWriteMDLHeader(FStream);
              if Result then Exit;
              Result := prWriteMDLStates(FStream);
              if Result then Exit;
              Result := prWriteMDLRates(FStream);
              if Result then Exit;
              Result := prWriteMDLConstraints(FStream);
              if Result then Exit;
              Result := prWriteMDLCoefficients(FStream);
            end;
        end;//case FType

      end;//FType

    finally

      prEditCritSection.Release;

    end;


  finally


    Beep;

    if Result then begin
      Console.SetEntry(Entry, 'Saving model file ' + SName + ' ... Error!', COLOR_ERROR);
    end else begin
      Console.SetEntry(Entry, 'Saving model file ' + SName + ' ... Done.', COLOR_GRAY);
    end;

    try FStream.Free; except end;

    if Result then begin

      DeleteFile(FName);

    end else begin

      prSaved := True;

      Close;

      try

        if Self.FileName = SName then begin
          DeleteFile(Self.FileName);
        end;

        if FileExists(SName) then begin
          Result := not DeleteFile(SName);
          if Result then begin
            SName := prGetFileNameForSave(ChangeFileExt(SName, ''), ExtractFileExt(SName), True{IncFile}, 1000{MaxInc});
            Result := SName = '';
            if Result then begin
              Console.NewEntry('Could not save as requested. Destination file ' + FileName  + ' could not be overwritten', COLOR_ERROR);
            end else begin
              Console.NewEntry('Could not save as requested. Instead saved as ' + SName, COLOR_WARNING);
            end;
          end;
        end;

        if not Result then begin
          AsQubtree := TNode.Open(FName);
          AsQubtree.RemoveChild(AsQubtree.Find('Properties'));
          AsQubtree.AppendClone(Properties, True);
          AsQubtree.Save;
          AsQubtree.Close;
        end;

        if not Result then begin
          RenameFile(FName, SName);
        end;

      except

        Result := True;
        New(SName);

      end;

      if not Result then Result := Open(SName);

    end;

  end;

end;


procedure TQUB_QFS_QMF.prInitProperties;
begin
  prRates.PreFormat := prProperties['PreFormat'];
  prRates.ExpFormat := prProperties['ExpFormat'];

  // if any were managed by qmf, set them up here

  if Assigned(OnInitProperties) then OnInitProperties(Self);
end;

procedure TQUB_QFS_QMF.DoOnSetProperties(Sender :TQUB_ProfileCat; Updates :INode);
begin
  // if any were managed by qmf, update them here

  if Assigned(OnSetProperties) then OnSetProperties(Sender, Updates);

  prRates.UpdateCaptions;
end;

function TQUB_QFS_QMF.prGetFileExt :String;
begin
  case prFileType of
    QFS_QMF_TYPE_QMF:
      Result := '.qmf';
    QFS_QMF_TYPE_MDL:
      Result := '.mdl';
    QFS_QMF_TYPE_UNK:
      Result := '';
  end;//case
end;

function TQUB_QFS_QMF.prGetFileTypeFromFileName(const FileName :String) :TQUB_QFS_QMF_FileType;
var
  Ext :String;
begin
  Ext := LowerCase(ExtractFileExt(FileName));
  if Ext = '' then begin
  //QMF
    Result := prFileType;
  end else if Ext = '.qmf' then begin
  //QMF
    Result := QFS_QMF_TYPE_QMF;
  end else if Ext = '.mdl' then begin
  //MDL
    Result := QFS_QMF_TYPE_MDL;
  end else if Ext = '.mod' then begin
  //MOD
    Result := QFS_QMF_TYPE_MOD;
  end else begin
  //UNK
    Result := QFS_QMF_TYPE_UNK;
  end;
end;

function TQUB_QFS_QMF.CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;
var
  FType :TQUB_QFS_QMF_FileType;
begin

  FType := prGetFileTypeFromFileName(FileName);

  Result := False;
  case FType of

    QFS_QMF_TYPE_QMF:
      Result := True;
    QFS_QMF_TYPE_MDL:
      Result := True;
    QFS_QMF_TYPE_MOD:
      Result := False;
    QFS_QMF_TYPE_UNK:
      Result := False;

  end;//case

  //John, this should be kept. When saving to mdl,
  //exp.cond. dependence is NOT saved.
  // TODO - determine whether features will be lost before warning.
  if (Result) and (Confirmation) then begin
  //Can be saved but the user wants Confirmation regarding lost features
    case FType of
      QFS_QMF_TYPE_MDL:
        Result := not (MessageHelpDlg('Some features may be lost after saving'#13#10'Are you sure you want to save in this format? (choose *.qmf)', mtConfirmation, [mbYes, mbNo], 0, 'Warning:Some_Model_Features_Lost') = idNo);
    end;//case
  end;

end;


//****************************************************************************//
//                                                                            //
//                        S a v e    T o    S t r i n g                       //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QMF.SaveToString(var MemFile :String; FileType :TQUB_QFS_QMF_FileType) :Boolean;
begin
  Result := prSaveToString(MemFile, FileType);
end;

function TQUB_QFS_QMF.prSaveToString(var MemFile :String; FileType :TQUB_QFS_QMF_FileType) :Boolean;
var
  MStream :TMemoryStream;
begin

  Result := False;

  try
    MStream := TMemoryStream.Create;
  except
    Result := True;
    Exit;
  end;

  try

    case FileType of
      QFS_QMF_TYPE_QMF:
        Result := Write(MStream);
      QFS_QMF_TYPE_MDL:
        begin
          Result := prWriteMDLHeader(MStream);
          if Result then Exit;
          Result := prWriteMDLStates(MStream);
          if Result then Exit;
          Result := prWriteMDLRates(MStream);
          if Result then Exit;
          Result := prWriteMDLConstraints(MStream);
          if Result then Exit;
          Result := prWriteMDLCoefficients(MStream);
        end;
    end;//case

  finally

    SetLength(MemFile, MStream.Size);
    Move(MStream.Memory^, PChar(MemFile)^, MStream.Size);
    MStream.Free;

  end;

end;

function TQUB_QFS_QMF.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = CHANNELS_CHANNEL_COUNT_NAME      then begin
      Result := prRecChannelCount;
    end else if NodeName = CHANNELS_STATES_NAME             then begin
      Result := prStates;
    end else if NodeName = CHANNELS_RATES_NAME              then begin
      Result := prRates;
    end else if NodeName = CHANNELS_CONSTRAINTS_NAME        then begin
      Result := prConstraints;
    end else if NodeName = CHANNELS_CONSTRAINTS_AMPVAR_NAME then begin
      Result := prConstraintsAmpVar;
    end else if NodeName = CHANNELS_EXTRAKINETICPARS_NAME   then begin
      Result := prExtraKineticPars;
    end else if NodeName = CHANNELS_AMPS_NAME               then begin
      Result := prRecAmps;
    end else if NodeNAme = CHANNELS_STDS_NAME               then begin
      Result := prRecStds;
    end else if NodeName = CHANNELS_CONDS_NAME              then begin
      Result := prRecConds;
    end else if NodeName = CHANNELS_CONDSTDS_NAME           then begin
      Result := prRecCondStds;
    end else if NodeName = CHANNELS_VREVS_NAME              then begin
      Result := prRecVrevs;
    end else if NodeName = CHANNELS_NAR_NAME                then begin
      Result := prRecNAr;
    end else if NodeName = CHANNELS_ARS_NAME                then begin
      Result := prRecArs;
    end else if NodeName = CHANNELS_VREV_NAME               then begin
      Result := prRecVRev;
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;


//****************************************************************************//
//                                                                            //
//                       R e a d   -   l o w   l e v e l                      //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QMF.Read(ReadStream :TStream) :Boolean;
var
  Entry :Integer;
begin

  Result := False;

  //Entry := Console.NewEntry('Reading model file ' + FileName , COLOR_TEXT);
  Entry := Console.NewEntry('Reading model file ' + FileName + ' ... Done.', COLOR_GRAY);


  try


    Delete;

    //Read according to the file type
    case prFileType of

      //Native QMF file format
      QFS_QMF_TYPE_QMF:
        begin
          Result := inherited Read(ReadStream);
          if not Result then Result := prReadQMF;
        end;

      //Old MDL binary format
      QFS_QMF_TYPE_MDL:
        Result := prReadMDL;

      // Synaptosoft format
      QFS_QMF_TYPE_MOD:
        Result := prReadMOD;

    end;//case

    if Result then Delete;//Exit;


  finally


    //To do: maybe find a way to control scrolling? LM
    if Result then begin
      Console.SetEntry(Entry, 'Reading model file ' + FileName + ' ... Error!', COLOR_ERROR);
    // Removed 8/05 - to avoid scrolling to right after saving model.  Request from Fred.
    //end else begin
    //  Console.SetEntry(Entry, 'Reading model file ' + FileName + ' ... Done.', COLOR_GRAY);
    end;


  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    Q M F                             //
//                                                                            //
//****************************************************************************//

//The structure is already read from stream, it remains to make the links;
function TQUB_QFS_QMF.prReadQMF :Boolean;
var
  Rate    :TQUB_QFS_Rate;
  State   :TQUB_QFS_State;
  s, r, i :Integer;
begin

  Result := False;

  for s := 0 to prStates.StateCount - 1 do begin

    State      := prStates[s];
    State.Amp  := Amp[State.Cls];
    State.Std  := Std[State.Cls];
    State.Cond := Cond[State.Cls];
    State.CondStd := CondStd[State.Cls];
    State.Vrev    := ClsVrev[State.Cls];
    for i := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
      State.Cff[i] := Cff[State.Cls, i];
    end;
    State.OnEdit := prOnEdit;
    if Assigned(prOnOpenState) then begin
      prOnOpenState(State);
    end;
    State.Panel.Visible := True;

  end;

  prStates.UpdateStates;

  for r := prRates.RateCount - 1 downto 0 do begin

    Rate := prRates[r];

    if (Rate.StateIndex[1] = -1) or (Rate.StateIndex[2] = -1) then begin
      prDelRate(Rate);
    end else begin
      Rate.States[1] := prStates[Rate.StateIndex[1]];
      Rate.States[2] := prStates[Rate.StateIndex[2]];
      //Rate.Panels[1].Caption := Rate.ValueFormat[1];
      //Rate.Panels[2].Caption := Rate.ValueFormat[2];
      Rate.Update;
      Rate.OnEdit := prOnEdit;
      if Assigned(prOnOpenRate) then begin
        prOnOpenRate(Rate);
      end;
      Rate.Panels[1].Visible := True;
      Rate.Panels[2].Visible := True;
    end;

  end;//of r

  prStates.CurrState := prStates.CurrState;

end;

//------------------------- Read Synaptosoft .MOD file
function TQUB_QFS_QMF.prReadMOD : Boolean;
var
  List       :TStringList;
begin

  List       := TStringList.Create;
  prEditing  := True;

  try
    List.LoadFromFile(prFileName);
    Result := prReadMODParse(List);
  finally
    List.Free;
    prEditing := False;
  end;
end;

//John, wouldn't it be easier to use Parser?
// Lorin - I don't think so, but prove me wrong.
// I didn't try it, but it looked like there was more overhead than savings.
// Note also that since writing this I received documentation on their model format and
// will need to enhance this.
function TQUB_QFS_QMF.prReadMODParse(const MDL : TStringList) :Boolean;
var
  iLine : Integer;                  // Index to scan string list
  iMode : Integer;                  // Currently scanning .. {0=none 1=states, 9-rates }
  iRow : Integer;                   // Row # within group

  c : String;

  aiRow2State : Array of Integer;  // x-ref Model state to .MOD state #
  oState : TQUB_QFS_State;
  oRate : TQUB_QFS_Rate;

  //----- States
  iStates, iClass : Integer;

  //----- Rates
  iRate, iState0, iState1 : Integer;
  nRate : Double;
  iDepend : Integer;

  //----- extract next token from string c as double or integer
  Function GetNextDbl() : Double;
  var
    iPos : Integer;
  Begin
    c := TrimLeft(c);
    iPos := Pos(' ',c);
    if( iPos = 0 ) Then Begin
      Result := StrToFloat(c);
      c := '';
    end else begin
      Result := StrToFloat(Copy(c,1,iPos-1));
      c := Copy(c,iPos+1,Length(c));
    end;
  End;

  Function GetNextInt : Integer;
  Begin
    Result := Round( GetNextDbl() );
  End;


Begin
  iMode := 0;
  iStates := 0;
  iRow := 0;
  SetLength(aiRow2State,20);
  Result := False;

  for iLine := 0 to MDL.Count-1 do Begin
    c := TrimLeft(MDL.Strings[iLine]);

    //----- Start of new section
    if Copy(c,0,1) = '[' Then Begin
      iMode := 0;
      iRow := 0;
    end else
        inc(iRow);

    //----- State names
    {if c = '[State1-20: labels]' Then
      iMode := ?
    else if iMode = ? Then Begin

    end;}

    //----- Active States
    if c = '[State1-20: onoff]' Then
      iMode := 1
    else if (iMode=1) and (iRow<20) Then Begin
      // Format is : Active, Class, Conductance... ?
      // 1 1 55.000000 0.000000
      if GetNextInt() > 0 Then Begin  // Active?
        inc(iStates);
        aiRow2State[iRow] := iStates;
        prAddState( 10 + 20 * ( (iRow-1) mod 5 ){x},
                    12 + 25 * Round( (-1.0+iRow) / 5.0 -0.49 ){y} );
        oState := prStates[prStates.StateCount - 1];

        iClass := GetNextInt();   // Class
        Amp[iClass] := GetNextDbl();
        Std[iClass] := 0.0;
        //if iClass > MaxClass then MaxClass := iClass;

        oState.Cls := iClass;
        oState.Pr  := 0.0;
      end;
    end;

    //----- Rate constants
    if c = '[Rate constants for transitions between all states]' Then
      iMode := 9
    else if iMode = 9 then begin
      // Format is state,state,rate,???...
      // 1 2 0.000000 1 0 0 0 0 0 0 0.000000 0.000000

      iState0 := aiRow2State[Math.max(1,Math.min(20,GetNextInt()))];    //-- State 0
      iState1 := aiRow2State[Math.max(1,Math.min(20,GetNextInt()))];    //-- State 1
      nRate := GetNextDbl();      //-- Rate
      if nRate > 0.0 Then Begin
        iDepend := GetNextInt();    //-- Ligand dependent ...
        //GetNextInt();GetNextInt();GetNextInt();GetNextInt();GetNextInt();GetNextInt();GetNextDbl();
        //GetNextDbl();

        for iRate := 0 to prRates.RateCount do begin
          if iRate = prRates.RateCount then begin
            prAddRate( prStates[iState0-1], prStates[iState1-1]);
            oRate := prRates[prRates.RateCount - 1];
            oRate.K0[1] := nRate;
            oRate.P[1] := Boolean(iDepend);
            oRate.K1[1] := 0.0;
            oRate.Q[1] := False;
            oRate.K0[2] := 0.0;
            oRate.P[2] := false;
            oRate.K1[2] := 0.0;
            oRate.Q[2] := false;
          end else begin
            oRate := prRates[iRate];
            if( (oRate.States[2]=prStates[iState0-1])
            and (oRate.States[1]=prStates[iState1-1]) ) then begin
              oRate.K0[2] := nRate;
              oRate.P[2] := Boolean(iDepend);
              oRate.K1[2] := 0.0;
              oRate.Q[2] := False;
              break;
            end;
          end;
        end;
      end;
    end;

    //----- Number of Channels
    if Copy(c,1,14)='[QMatrixMonteCarlo' then begin
      iMode := 10;
    end else if (iMode=10) and (iRow=5)  then begin
      ChannelCount := Math.Min(GetNextInt(),1); //Max(1, Min(QUB_QFS_QMF_MAX_CHANNEL_COUNT, GetNextInt()));
    end;

  end;
end;


//****************************************************************************//
//                                                                            //
//                               R e a d    M D L                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QMF.prReadMDL :Boolean;
var
  List       :TStringList;
begin

  List       := TStringList.Create;
  prEditing  := True;

  try

    try

      List.LoadFromFile(prFileName);

    except

      Result := True;

    end;

    Result := prReadMDLParse(List.Text);

  finally

    List.Free;

    prEditing := False;

  end;

end;

function TQUB_QFS_QMF.prReadMDLParse(const MDL :String) :Boolean;
var
  State      :TQUB_QFS_State;
  Rate       :TQUB_QFS_Rate;
  Constraint :TQUB_QFS_Constraint;
  Parser     :TQUB_Parser;
  amp, std   :Double;
  pr         :Double;
  cf         :Double;
  s1, s2     :Integer;
  k1, k2     :Double;
  b1, b2     :Double;
  d1, d2     :Integer;
  x, y       :Double;
  n, c       :Integer;
  i, j       :Integer;
  MaxClass   :Integer;
begin

  Result     := False;
  Parser     := nil;

  MaxClass   := 0;

  try

    Parser := TQUB_Parser.Create;
    Parser.ReturnTab := False;
    Parser.ReturnNewLine := False;

    Parser.Start;

    try

      //Number of Channels *****************************************************
      Parser.Next(MDL);//prChannels
      ChannelCount := Parser.Token.TokenNumber;
      if (ChannelCount < 1) or (ChannelCount > QUB_QFS_QMF_MAX_CHANNEL_COUNT) then ChannelCount := 1;
      Parser.Next(MDL);//Number
      Parser.Next(MDL);//of
      Parser.Next(MDL);//Channels

      //Conductance Classes ****************************************************
      Parser.Next(MDL);//n
      n := Parser.Token.TokenNumber;
      Parser.Next(MDL);//Number
      Parser.Next(MDL);//of
      Parser.Next(MDL);//Classes
      Parser.Next(MDL);//Amp
      Parser.Next(MDL);//Sd
      Parser.Next(MDL);//Color
      Parser.Next(MDL);//(
      Parser.Next(MDL);//conductance
      Parser.Next(MDL);//info
      Parser.Next(MDL);//)
      for i := 0 to n - 1 do begin
        amp := 1;
        Parser.Next(MDL);//amp
        if Parser.Token.Token = '-' then begin
          Parser.Next(MDL);//amp
          amp := -1;
        end;
        amp := amp * Parser.Token.TokenNumber;
        Parser.Next(MDL);//std
        std := Parser.Token.TokenNumber;
        Parser.Next(MDL);//c
        c := Parser.Token.TokenNumber;
        Self.Amp[c] := amp;
        Self.Std[c] := std;
        if c > MaxClass then MaxClass := c;
      end;

      //States *****************************************************************
      Parser.Next(MDL);//n;
      n := Parser.Token.TokenNumber;
      Parser.Next(MDL);//Number
      Parser.Next(MDL);//of
      Parser.Next(MDL);//States
      Parser.Next(MDL);//Pr
      Parser.Next(MDL);//Color
      Parser.Next(MDL);//x
      Parser.Next(MDL);//y
      Parser.Next(MDL);//(
      Parser.Next(MDL);//state
      Parser.Next(MDL);//info
      Parser.Next(MDL);//)
      for i := 0 to n - 1 do begin
        Parser.Next(MDL);//pr
        pr := Parser.Token.TokenNumber;
        Parser.Next(MDL);//c
        c := Parser.Token.TokenNumber;
        Parser.Next(MDL);//x
        x := Parser.Token.TokenNumber;
        Parser.Next(MDL);//y
        y := Parser.Token.TokenNumber;
        Result := prAddState(x, y);
        State := prStates[prStates.StateCount - 1];
        State.Cls := c;
        State.Pr  := pr;
      end;

      //Rates ******************************************************************
      Parser.Next(MDL);//n;
      n := Parser.Token.TokenNumber;
      Parser.Next(MDL);//Number
      Parser.Next(MDL);//of
      Parser.Next(MDL);//Connections

      for i := 0 to n - 1 do begin
        Parser.Next(MDL);//s1
        s1 := Parser.Token.TokenNumber;
        Parser.Next(MDL);//s2
        s2 := Parser.Token.TokenNumber;
        Parser.Next(MDL);//k1
        k1 := Parser.Token.TokenNumber;
        b1 := 1;
        Parser.Next(MDL);//b1
        if Parser.Token.Token = '-' then begin
          Parser.Next(MDL);//b1
          b1 := -1;
        end;
        b1 := b1 * Parser.Token.TokenNumber;
        Parser.Next(MDL);//d1
        d1 := Parser.Token.TokenNumber;
        Parser.Next(MDL);//k2
        k2 := Parser.Token.TokenNumber;
        b2 := 1;
        Parser.Next(MDL);//b2
        if Parser.Token.Token = '-' then begin
          Parser.Next(MDL);//b2
          b2 := -1;
        end;
        b2 := b2 * Parser.Token.TokenNumber;
        Parser.Next(MDL);//d2
        d2 := Parser.Token.TokenNumber;
        Result := prAddRate(prStates[s1 - 1], prStates[s2 - 1]);
        Rate := prRates[prRates.RateCount - 1];

        Rate.K0[1] := k1;
        Rate.K0[2] := k2;
        Rate.K1[1] := b1;
        Rate.K1[2] := b2;
        Rate.P[1] := Boolean(d1);
        Rate.P[2] := Boolean(d2);
        Rate.Q[1] := False;
        Rate.Q[2] := False;
        Rate.PName[1] := CHANNELS_RATE_DEF_PNAME;
        Rate.PName[2] := CHANNELS_RATE_DEF_PNAME;
        Rate.QName[1] := CHANNELS_RATE_DEF_QNAME;
        Rate.QName[2] := CHANNELS_RATE_DEF_QNAME;
        Rate.RateFormat[1] := CHANNELS_RATE_DEF_FORMAT;
        Rate.RateFormat[2] := CHANNELS_RATE_DEF_FORMAT;

      end;

      //Constraints ************************************************************
      Parser.Next(MDL);//n
      n := Parser.Token.TokenNumber;
      Parser.Next(MDL);//Number
      Parser.Next(MDL);//of
      Parser.Next(MDL);//Constraint
      Parser.Next(MDL);//Parameters
      for i := 1 to 25 do begin
        Parser.Next(MDL);//...
      end;
      for i := 0 to n - 1 do begin
        Parser.Next(MDL);//Constraint.iType
        Constraint := TQUB_QFS_Constraint.Create(CHANNELS_CONSTRAINT_TYPE_NAME[TQUB_QFS_QMF_ConstraintType(Parser.Token.TokenNumber)], nil);
        Parser.ReturnNewLine := True;
        repeat
          Parser.Next(MDL);// c / NewLine / NULL ?!
          if Parser.Token.TokenType = TOKEN_NUMBER then begin
            c := Parser.Token.TokenNumber;
            Constraint.Index[Constraint.IndexCount] := c;
          end;
        until((Parser.Token.TokenType = TOKEN_NEWLINE) or (Parser.Token.TokenType = TOKEN_NULL));
        Parser.ReturnNewLine := False;
        prConstraints.AddConstraint(Constraint, prConstraints.ConstraintCount);
      end;


      //Coefficients ***********************************************************

      //Reset
      for i := 0 to MaxClass do begin
        for j := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
          Cff[i, j] := 0;
        end;
      end;

      //Read
      Nar[0] := 0;
      Parser.Next(MDL);//prNar
      Nar[0] := Parser.Token.TokenNumber;
      Parser.Next(MDL);//Number
      Parser.Next(MDL);//of
      Parser.Next(MDL);//Coefficients
      if (Nar[0] < 0) or (Nar[0] > QUB_QFS_QMF_MAX_COEFF) then begin
        Nar[0] := 0;
      end;

      for i := 0 to MaxClass do begin

        for j := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin

          Parser.Next(MDL);//cf / NULL ?!
          cf := 1;
          if Parser.Token.Token = '-' then begin
            Parser.Next(MDL);
            cf := -1;
          end;
          if Parser.Token.TokenType = TOKEN_NUMBER then begin
            cf := cf * Parser.Token.TokenNumber;
            Cff[i, j] := cf;
          end else if Parser.Token.TokenType = TOKEN_NULL then begin
            Break;
          end;

        end;

        if Parser.Token.TokenType = TOKEN_NULL then begin
          Break;
        end;

      end;

      //Update States' properties
      for i := 0 to States.StateCount - 1 do begin
        States[i].Amp  := Self.Amp[States[i].Cls];
        States[i].Std  := Self.Std[States[i].Cls];
        States[i].Cond := Self.Cond[States[i].Cls];
        States[i].CondStd := Self.CondStd[States[i].Cls];
        States[i].Vrev    := Self.ClsVrev[States[i].Cls];
        for j := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
          States[i].Cff[j] := Self.Cff[States[i].Cls, j];
        end;
      end;

      prStates.UpdateStates;

    except

      Result := True;

    end;

  finally

    Parser.Free;

  end;

end;


//****************************************************************************//
//                                                                            //
//                      W r i t e   -   l o w   l e v e l                     //
//                                                                            //
//****************************************************************************//
//Header
function TQUB_QFS_QMF.prWriteMDLHeader(Stream :TStream) :Boolean;
var
  Buffer :String;
  c      :Array of Boolean;
  mc     :Integer;
  i      :Integer;
begin
  Result := False;

  //Number of Channels
  Buffer := Format('%2d' + #9 + 'Number of Channels' + #13#10, [ChannelCount]);
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  //Conductance Classes ****************************************************
  SetLength(c, ClassCount);
  mc := 0;
  for i := 0 to ClassCount - 1 do begin
    c[i] := False;
  end;

  for i := 0 to prStates.StateCount - 1 do begin
    c[prStates[i].Cls] := True;
  end;
  for i := 0 to ClassCount - 1 do begin
    if c[i] then Inc(mc);
  end;

  Buffer := Format('%2d' + #9 + 'Number of Classes' + #13#10, [mc]);
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  Buffer := 'Amp' + #9 + 'Sd' + #9 + 'Color' + #9 + '(conductance info)' + #13#10;
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  for i := 0 to ClassCount - 1 do begin
    if c[i] then begin
      Buffer := Format('%.6g %.6g %2d' + #13#10, [Amp[i], Std[i], i]);
      Stream.Write(PChar(Buffer)^, Length(Buffer));
    end;
  end;

end;

//States
function TQUB_QFS_QMF.prWriteMDLStates(Stream :TStream) :Boolean;
var
  Buffer :String;
  i      :Integer;
begin
  Result := False;

  //States *****************************************************************
  Buffer := Format('%2d' + #9 + 'Number of States' + #13#10, [prStates.StateCount]);
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  Buffer := 'Pr' + #9 + 'Color' + #9 + 'x' + #9 + 'y' + #9 + '(state info)' + #13#10;
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  for i := 0 to prStates.StateCount - 1 do begin
    Buffer := Format('%.6g %2d %.4g %.4g' + #13#10, [prStates[i].Pr, prStates[i].Cls, prStates[i].x, prStates[i].y]);
    Stream.Write(PChar(Buffer)^, Length(Buffer));
  end;

end;

//Rates
function TQUB_QFS_QMF.prWriteMDLRates(Stream :TStream) :Boolean;
var
  Buffer    :String;
  s1, s2, i :Integer;
begin
  Result := False;

  //Rates ******************************************************************
  Buffer := Format('%2d' + #9 + 'Number of Connections' + #13#10, [prRates.RateCount]);
  Stream.Write(PChar(Buffer)^, Length(Buffer));
  for i := 0 to prRates.RateCount - 1 do begin
    s1 := prStates.IndexOf(prRates[i].States[1]);
    s2 := prStates.IndexOf(prRates[i].States[2]);
    Buffer := Format('%2d %2d %.6g %.6g %2d %.6g %.6g %2d' + #13#10, [s1 + 1, s2 + 1, prRates[i].K0[1], prRates[i].K1[1], Integer(prRates[i].P[1]), prRates[i].K0[2], prRates[i].K1[2], Integer(prRates[i].P[2])]);
    Stream.Write(PChar(Buffer)^, Length(Buffer));
  end;

end;


//Constraints
function TQUB_QFS_QMF.prWriteMDLConstraints(Stream :TStream) :Boolean;
var
  Buffer :String;
  con    :String;
  i, j   :Integer;
begin
  Result := False;

  //Constraints ************************************************************
  Buffer := Format('%2d' + #9 + 'Number of Constraint Parameters' + #13#10, [prConstraints.ConstraintCount]);
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  Buffer := 'itype->0:fix a' + #9 + '1:scale a' + #9 + '2:loop' + #9 + '3:fix b' + #9 + '4:scale b' + #13#10;
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  Buffer := 'itype' + #9 + 'state index' + #13#10;
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  for i := 0 to prConstraints.ConstraintCount - 1 do begin
    Buffer := Format('%2d', [Integer(prConstraints[i].CType)]);
    Stream.Write(PChar(Buffer)^, Length(Buffer));
    con := '';
    for j := 0 to prConstraints[i].IndexCount - 1 do begin
      con := con + #9 + IntToStr(prConstraints[i].Index[j]);
    end;
    Buffer := con + #13#10;
    Stream.Write(PChar(Buffer)^, Length(Buffer));
  end;

end;

//Coefficients
function TQUB_QFS_QMF.prWriteMDLCoefficients(Stream :TStream) :Boolean;
var
  Buffer :String;
  i, j   :Integer;
begin
  Result := False;

  //Coefficients ***********************************************************
  Buffer := Format('%2d' + #9 + 'Number of Coefficients' + #13#10, [NAr[0]]);
  Stream.Write(PChar(Buffer)^, Length(Buffer));

  for i := 0 to ClassCount - 1 do begin
    for j := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
      Buffer := Format('%.6g' + #13#10, [Cff[i, j]]);
      Stream.Write(PChar(Buffer)^, Length(Buffer));
    end;
  end;

end;



//****************************************************************************//
//                                                                            //
//                        E D I T     M E T H O D S                           //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QMF.CutModel :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prCutModel;
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QMF.CopyModel :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prCopyModel;
      except
        Result := True;
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QMF.PasteModel :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prPasteModel;
        UpdateParameters;
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QMF.DeleteModel :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prDeleteModel;
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

//******************************************************************************
function TQUB_QFS_QMF.AddState(x, y :Double) :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prAddState(x, y);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        prReAllocMath;
        //prReAllocGroups;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QMF.DelState(State :TQUB_QFS_State) :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prDelState(State);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        prReAllocMath;
        //prReAllocGroups;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QMF.PosState(State :TQUB_QFS_State; xp, yp :Double; x, y :Integer) :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prPosState(State, xp, yp, x, y);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QMF.AddRate(State1, State2 :TQUB_QFS_State) :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prAddRate(State1, State2);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QMF.DelRate(Rate :TQUB_QFS_Rate) :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prDelRate(Rate);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

// 12-05 JB - swap states is used in the renumbering of model states
procedure TQUB_QFS_QMF.SwapStates( iOldState : Integer; iNewState : Integer );
var
  i, j : Integer;
begin
    for i := Constraints.ConstraintCount - 1 downto 0 do Begin
        for j := 0 to Constraints.Constraint[i].IndexCount - 1 do
            if (Constraints.Constraint[i].Index[j] = iOldState) Then
                Constraints.Constraint[i].Index[j]:=iNewState
            else if ( Constraints.Constraint[i].Index[j] = iNewState) Then
                Constraints.Constraint[i].Index[j]:=iOldState;
    end;
    States.Swap(iOldState,iNewState);
    States.UpdateStates;
end;


//******************************************************************************
function TQUB_QFS_QMF.Connected(State1, State2 :TQUB_QFS_State) :Boolean;
var
  i :Integer;
begin
  Result := False;
  for i := 0 to prRates.RateCount - 1 do begin
    if (((prRates[i].States[1] = State1) and (prRates[i].States[2] = State2)) or ((prRates[i].States[1] = State2) and (prRates[i].States[2] = State1))) then begin
      Result := True;
      Exit;
    end;
  end;
end;

function TQUB_QFS_QMF.RateFixed(State1, State2 :Integer) :Boolean;
var
  i :Integer;
begin
  Result := False;
  if prConstraints.ConstraintCount = 0 then Exit;

  for i := 0 to prConstraints.ConstraintCount - 1 do begin
    with prConstraints[i] do begin
      if CType = CSTR_FIX_RATE then begin
        if (Index[0] = State1) and (Index[1] = State2) then begin
          Result := True;
          Exit;
        end;
      end;
    end;
  end;

end;

procedure TQUB_QFS_QMF.Delete;
var
  State :TQUB_QFS_State;
  Rate  :TQUB_QFS_Rate;
  i     :Integer;
begin
  if prExtraKineticPars <> nil then begin
    for i := prExtraKineticPars.ExtraKineticParCount - 1 downto 0 do begin
      prExtraKineticPars.DelExtraKineticPar(i);
    end;
  end;
  if prConstraints <> nil then begin
    for i := prConstraints.ConstraintCount - 1 downto 0 do begin
      prConstraints.DelConstraint(i);
    end;
  end;
  if prConstraintsAmpVar <> nil then begin
    for i := prConstraintsAmpVar.ConstraintCount - 1 downto 0 do begin
      prConstraintsAmpVar.DelConstraint(i);
    end;
  end;
  if prRates <> nil then begin
    for i := prRates.RateCount - 1 downto 0 do begin
      Rate := prRates[i];
      prRates.DelRate(Rate);
      Rate.Free;
    end;
  end;
  if prStates <> nil then begin
    for i := prStates.StateCount - 1 downto 0 do begin
      State := prStates[i];
      prStates.DelState(State);
      State.Free;
    end;
  end;
end;

{
function TQUB_QFS_QMF.CopyFrom(Source :TQUB_QFS_QMF) :Boolean;
var
  Mdl :String;
begin

  Result := Source.prSaveToString(Mdl, QFS_QMF_TYPE_MDL);
  if Result then Exit;

  try
    prEditCritSection.Acquire;
    prEditing := True;
    try
     prDeleteModel;
     Result := prReadMDLParse(MDL);
    except
      Result := True;
    end;
    if not Result then begin
      prModified := True;
      if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
    end;
  finally
    prEditing := False;
    prEditCritSection.Release;
  end;

end;
}
{
This new version uses the QMF format, and thus doesn't miss any feature, such as P/QNames
LM.
}
function TQUB_QFS_QMF.CopyFrom(Source :TQUB_QFS_QMF) :Boolean;
var
  SourceStream :TMemoryStream;
begin

  SourceStream := TMemoryStream.Create;

  Result := Source.Write(SourceStream);
  if Result then Exit;

  try

    prEditCritSection.Acquire;
    prEditing := True;

    try

     prDeleteModel;

     if Stream <> nil then begin
       Stream.Free;
       Stream := nil;
     end;

     SourceStream.Position := 0;
     Result := inherited Read(SourceStream);
     if not Result then Result := prReadQMF;
     //Close the Stream, as if it were a new file
     Stream := nil;

     //Set Parameters
     if Parameters <> nil then begin
       Parameters.Events.DelEvent('OnChange', DoOnParametersChanged);
     end;
     Parameters := Source.Parameters;
     if Parameters <> nil then begin
       Parameters.Events.AddEvent('OnChange', DoOnParametersChanged);
     end;
     UpdateParameters;

    except

      Result := True;

    end;

    if not Result then begin

      prModified := True;

    end else begin

      prDeleteModel;

    end;

    if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);

  finally

    SourceStream.Free;

    prEditing := False;
    prEditCritSection.Release;

  end;

end;


//Protected Edit methods *******************************************************
function TQUB_QFS_QMF.prCutModel :Boolean;
begin
  Result := prCopyModel;
  if not Result then begin
    Result := prDeleteModel;
  end;
end;

function TQUB_QFS_QMF.prCopyModel :Boolean;
begin
  Result    := not CloneAsNode(True, TNode.CreateNull).Copy;
end;

function TQUB_QFS_QMF.prPasteModel :Boolean;
var
  Name        :String;
  WasEditing  :Boolean;
begin

  Result := not prOpened;
  if Result then Exit;

  Result      := not (CanPasteNode(Name) and (Name = 'ModelFile'));
  if Result then Exit;

  WasEditing  := prEditing; // ???
  prEditing   := False;
  SetModelFromNode(Self, TNode.Paste);
  prEditing   := WasEditing;

end;

function TQUB_QFS_QMF.prDeleteModel :Boolean;
var
  s :Integer;
begin

  Result := (not prOpened) or (prStates = nil) or (prRates = nil);
  if Result then Exit;

  try

    try
      for s := prStates.StateCount - 1 downto 0 do begin
        prDelState(prStates[s]);
      end;
    except
      Result := True;
    end;

  finally

  end;
end;

//******************************************************************************
function TQUB_QFS_QMF.prAddState(x, y :Double) :Boolean;
var
  State :TQUB_QFS_State;
begin

  Result := False;
  if prStates.StateCount = QUB_QFS_QMF_MAX_STATE then begin
    Result := True;
    Exit;
  end;

  prConstraints.AddState;

  State := prStates.NewState(x, y);
  State.OnEdit := prOnEdit;
  if Assigned(prOnNewState) then begin
    prOnNewState(State);
  end;
  State.Panel.Visible := True;

  prStates.CurrState := State;

  //Add to Edit Operations
end;

function TQUB_QFS_QMF.prDelState(State :TQUB_QFS_State) :Boolean;
var
  i, si, si1, si2 :Integer;
begin
  Result := False;

  si     := prStates.RecordIndex(State);

  for i := prRates.RateCount - 1 downto 0 do begin
    if ((prRates[i].States[1] = State) or (prRates[i].States[2] = State)) then begin
      prDelRate(prRates[i]);
    end;
  end;


  //This should delete all generalized constraints
  prConstraints.DelState(si, prStates.StateCount, prRates.RateCount, prExtraKineticPars.ExtraKineticParCount);//Index

  prStates.DelState(State);
  State.Free;

  prStates.CurrState := prStates.CurrState;

  // re-link rates connecting states whose numbers were shifted down, so they record the new number
  for i := prRates.RateCount - 1 downto 0 do begin
    si1 := prStates.RecordIndex(prRates[i].States[1]);
    si2 := prStates.RecordIndex(prRates[i].States[2]);
    if si1 >= si then prRates[i].States[1] := prStates[si1];
    if si2 >= si then prRates[i].States[2] := prStates[si2];
  end;

  //Add to Edit Operations
end;

function TQUB_QFS_QMF.prPosState(State :TQUB_QFS_State; xp, yp :Double; x, y :Integer) :Boolean;
var
  i    :Integer;
  p    :Boolean;
begin

  Result := False;

  p := False;
  for i := prRates.RateCount - 1 downto 0 do begin
    if ((prRates[i].States[1] = State) or (prRates[i].States[2] = State)) then begin
      p := True;
    end;
  end;

  State.x := xp;
  State.y := yp;
  State.Panel.SetBounds(x - (State.Panel.Width div 2), y - (State.Panel.Width div 2), State.Panel.Width, State.Panel.Height);

  if p then begin

    for i := 0 to prRates.RateCount - 1 do begin
      if ((prRates[i].States[1] = State) or (prRates[i].States[2] = State)) then begin
        prRates[i].Update;
      end;
    end;

  end;

  //Add to Edit Operations
end;

function TQUB_QFS_QMF.prAddRate(State1, State2 :TQUB_QFS_State) :Boolean;
var
  Rate   :TQUB_QFS_Rate;
begin
  Result := False;
  if prRates.RateCount = QUB_QFS_QMF_MAX_RATE then begin
    Result := True;
    Exit;
  end;

  if Connected(State1, State2) then begin
    Result := True;
    Exit;
  end;

  prConstraints.AddRate(prStates.StateCount, prRates.RateCount, prExtraKineticPars.ExtraKineticParCount);

  Rate := prRates.NewRate(State1, State2);

  Rate.OnEdit := prOnEdit;
  if Assigned(prOnNewRate) then begin
    prOnNewRate(Rate);
  end;

  Rate.Panels[1].Visible := True;
  Rate.Panels[2].Visible := True;

  prStates.CurrState := prStates.CurrState;

  //Add to Edit Operations
end;

function TQUB_QFS_QMF.prDelRate(Rate :TQUB_QFS_Rate) :Boolean;
begin
  Result := False;

  //QuB crashed whenever a state or rate involved in constraints was deleted. LM
  prConstraints.DelRate(Rate, prRates.RateIndex[Rate], prStates.StateCount, prRates.RateCount, prExtraKineticPars.ExtraKineticParCount);

  prRates.DelRate(Rate);
  Rate.Free;

  prStates.CurrState := prStates.CurrState;

  //Add to Edit Operations
end;



//Simulation functions
function TQUB_QFS_QMF.GenerateState :Integer;
const
  MAXS  = 20;
var
  ps    :Array[0..MAXS-1] of Double;
  p     :TPADouble;
  s, u  :Double;
  i, sc :Integer;
begin

  Result := 0;

  sc := prStates.StateCount;
  if sc = 0 then begin
    Result := -1;
    Exit;
  end else if sc = 1 then begin
    Exit;
  end;

  //Avoids unnecessary memory fragmentation (a drop in an ocean...)
  if sc > MAXS then begin
    p := DAlloc1(sc);
  end else begin
    p := @ps[0];
  end;

  try

    s := 0;
    for i := 0 to sc - 1 do begin
      p[i] := prStates[i].Pr;
      s := s + p[i];
    end;
    if s = 0 then begin
      s := 1 / sc;
      for i := 0 to sc - 1 do begin
        p[i] := s;
      end;
    end else begin
      for i := 0 to sc - 1 do begin
        p[i] := p[i] / s;
      end;
    end;

    //Randomize; Beware !!! Ramdomize takes the seed from system data/time.
    //If it is called N times very fast, chances are that the date/time won't change
    //so the seed is the same !!!
    u := Random;
    Result := 0;
    s := p[0];
    while (s < u) and (Result < sc) do begin
      Inc(Result);
      s := s + p[Result];
    end;

  finally

    if sc > MAXS then FreeMem(p);

  end;

end;

function TQUB_QFS_QMF.GenerateDwell(State :Integer) :Double;
begin

  Result := ExpDev / -Q[State, State];

end;

function TQUB_QFS_QMF.GenerateState(State :Integer) :Integer;
begin

  Result := DisDev(MathStateCount - 1, P[State]);
  if (Result >= State) and (Result < (MathStateCount - 1)) then begin
    Inc(Result);
  end;

end;

procedure TQUB_QFS_QMF.GenerateP(FromP, ToP :TPADouble; StateCount, ChannelCount :Integer);
var
  s, c :Integer;
begin

  //Clear P0 first
  for s := 0 to StateCount - 1 do begin
    ToP[s] := 0;
  end;

  for c := 0 to ChannelCount - 1 do begin
    s := DisDev(StateCount, FromP);
    ToP[s] := ToP[s] + 1/ChannelCount;
  end;

end;


//Conductance
function TQUB_QFS_QMF.Conductance :Double;
var
  i :Integer;
begin

  Result := 0;
  for i := 0 to prMathStateCount - 1 do begin
    Result := Result + Pt[i]*States[i].Cond;
  end;

end;


//Parameters
procedure TQUB_QFS_QMF.DoOnParametersChanged(Sender :TObject);
begin
  UpdateParameters;
end;

procedure TQUB_QFS_QMF.UpdateParameters;
var
  Rate :TQUB_QFS_Rate;
  Par  :TQUB_ExpCond_PParameter;
  r, i :Integer;
begin

  if prParameters = nil then Exit;

  for r := 0 to prRates.RateCount - 1 do begin

    Rate := prRates[r];
    Rate.ClearPQ;

    for i := 1 to 2 do begin

      //P
      Par := prParameters.ParameterByName[Rate.PName[i]];
      if Par = nil then begin
        prParameters.AddPar(Rate.PName[i], 1.0, -1, False);
        Par := prParameters.ParameterByName[Rate.PName[i]];
      end;
      if Par <> nil then begin
        Rate.PValue[i] := @Par.Value;
      end else begin
        Rate.PValue[i] := nil;
      end;

      //Q
      Par := prParameters.ParameterByName[Rate.QName[i]];
      if Par = nil then begin
        prParameters.AddPar(Rate.QName[i], 1.0, -1, False);
        Par := prParameters.ParameterByName[Rate.QName[i]];
      end;
      if Par <> nil then begin
        Rate.QValue[i] := @Par.Value;
      end else begin
        Rate.QValue[i] := nil;
      end;

    end;

  end;

  // there is no point to a global channel count
  // this will actively root out any that remain
  // while still updating CurrData.ExpCond['ChannelCount']
  // (that is necessary for global fitting, unless we remove the ability to global-fit with a different # of channels per file)
  // Chris 11-05 no more channel count in expcond
  // (except as put there by datalist for global fits and ignored by qmf)

  if prParameters = ViewportModel.DefExpCond then begin
  Par := prParameters.ParameterByName['ChannelCount'];
    if Par <> nil then begin
      prParameters.DelPar(prParameters.IndexOf('ChannelCount'));
    end;
  end;
  (*
  if Par = nil then begin
    if prParameters <> ViewportModel.DefExpCond then begin
      prParameters.AddPar('ChannelCount', ChannelCount, -1, False);
    end;
  end else begin
    if prParameters <> ViewportModel.DefExpCond then begin
      if ChannelCount <> Round(Par.Value) then begin
    ChannelCount := Round(Par.Value);
  end;
    end else begin
      prParameters.DelPar(prParameters.IndexOf('ChannelCount'));
    end;
  end;
  *)
  if Assigned(OnNeedDraw) then OnNeedDraw(Self);

end;

function TQUB_QFS_QMF.UsesParameter(Par :TQUB_ExpCond_PParameter) :Boolean;
var
  Rate :TQUB_QFS_Rate;
  r, i :Integer;
begin
  Result := False;
  for r := 0 to prRates.RateCount - 1 do begin
    Rate := prRates[r];
    for i := 1 to 2 do begin
      if (Rate.P[i]) and (Rate.PName[i] = Par.Name){ = @(Par.Value))} then Result := True;
      if (Rate.Q[i]) and (Rate.QName[i] = Par.Name){ = @(Par.Value))} then Result := True;
      if Result then Exit;
    end;
  end;
end;


function TQUB_QFS_QMF.SameParameters(Params1, Params2 :TQUB_ExpCond_Parameters) :Boolean;
var
  Rate       :TQUB_QFS_Rate;
  Par1, Par2 :TQUB_ExpCond_PParameter;
  k, l       :Integer;
begin

  Result := True;

  for k := 0 to Rates.RateCount - 1 do begin

    Rate := Rates[k];

    for l := 0 to 2 do begin

      //P-dependent ?
      if Rate.P[l] then begin

        Par1 := Params1.ParameterByName[Rate.PName[l]];
        Par2 := Params2.ParameterByName[Rate.PName[l]];

        if (Par1 <> nil) = (Par2 <> nil) then begin
        //Either both nil or both non-nil
          if Par1 <> nil then begin
          //Non-nil
            if Par1.Value <> Par2.Value then begin
              Result := False;
              Break;
            end;
          end;
        end else begin
          Result := False;
        end;

      end;//P ?

      //Q-dependent ?
      if Rate.Q[l] then begin

        Par1 := Params1.ParameterByName[Rate.QName[l]];
        Par2 := Params2.ParameterByName[Rate.QName[l]];

        if (Par1 <> nil) = (Par2 <> nil) then begin
        //Either both nil or both non-nil
          if Par1 <> nil then begin
          //Non-nil
            if Par1.Value <> Par2.Value then begin
              Result := False;
              Break;
            end;
          end;
        end else begin
          Result := False;
        end;

      end;//Q ?

    end;//l

    if not Result then Break;

  end;

end;
                                                                           


{
This function identifies groups of aggregated states.
A group is defined as a collection of states of the same class,
such that, starting from any state within the group, the channel can visit any
other state in the group only by making transitions within the group.
Example:
C-C-O-C   = there are two C groups and one O group
The groups can be used, for example, to separate clusters
}
(*
function TQUB_QFS_QMF.prGetGroup(g :Integer) :TQUB_QFS_QMF_PGroup;
begin
  if (g >= 0) and (g < GroupCount) then Result := prGroups[g] else Result := nil;
end;

procedure TQUB_QFS_QMF.prClearGroups;
var
  g :Integer;
begin
  for g := 0 to GroupCount - 1 do begin
    FreeMem(TQUB_QFS_QMF_PGroup(prGroups[g]).States);
    FreeMem(TQUB_QFS_QMF_PGroup(prGroups[g]).Lambdas);
    FreeMem(TQUB_QFS_QMF_PGroup(prGroups[g]).Taus);
    FreeMem(TQUB_QFS_QMF_PGroup(prGroups[g]).Areas);
    Dispose(TQUB_QFS_QMF_PGroup(prGroups[g]));
  end;
  FreeMem(prGroups);
  prGroups := nil;
  prGroupCount := 0;
end;

function TQUB_QFS_QMF.FindGroups :Boolean;
var
  SCount            :Integer;
  RCount            :Integer;

  QAllowed          :TPMInteger;
  Group             :TQUB_QFS_QMF_PGroup;
  Grouped           :TPAInteger;
  Connected         :TPAInteger;
  Q, A              :TPMDouble;
  i, j, r, s, c, g  :Integer;
  s1, s2            :Integer;
begin

  Result := False;

  prClearGroups;

  SCount := prStates.StateCount;
  RCount := prRates.RateCount;
  if (SCount = 0) or ((SCount > 1) and (RCount = 0)) then begin
    Result := True;
    Exit;
  end;

  QAllowed  := nil;
  Grouped   := nil;
  Connected := nil;


  try

    //Construct the Boolean matrix of alowed transitions
    QAllowed := IAlloc2(SCount, SCount);
    for i := 0 to SCount - 1 do begin
      for j := 0 to SCount - 1 do begin
        QAllowed[j, i] := 0;
      end;
    end;

    for r := 0 to RCount - 1 do begin
      s1 := prStates.IndexOf(prRates[r].States[1]);
      s2 := prStates.IndexOf(prRates[r].States[2]);
      QAllowed[s1, s2] := Integer(prRates[r].Value[1] <> 0);
      QAllowed[s2, s1] := Integer(prRates[r].Value[2] <> 0);
    end;

    Grouped   := IAlloc1(SCount);//whether a state was grouped
    Connected := IAlloc1(SCount);//whether a state was connected to the current group
    for i := 0 to SCount - 1 do begin
      Grouped[i] := 0;
    end;

    repeat

      //No state found yet to start a new group
      s := -1;
      for i := 0 to SCount - 1 do begin
        if Grouped[i] = 0 then begin
          s := i;
          Break;
        end;
      end;

      if s = -1 then Break;
      //State found

      //Augment the Groups
      Inc(prGroupCount);
      ReAllocMem(prGroups, SizeOf(Pointer) * prGroupCount);

      //Create a new group
      System.New(Group);
      Group.Cls := prStates[s].Cls;
      Group.StateCount := 1;
      Group.States     := nil;
      Group.Lambdas    := nil;
      Group.Taus       := nil;
      Group.Areas      := nil;

      prGroups[prGroupCount - 1] := Group;

      //Set the Connected indicator
      for i := 0 to SCount - 1 do begin
        if Boolean(Grouped[i]) then begin
        //Already grouped, ignore
          Connected[i] := -1;
        end else if i = s then begin
        //the starting state, connect (to itself)
          Connected[i] := 1;
        end else begin
        //not connected
          Connected[i] := 0;
        end;
      end;

      //Check for new connections
      repeat
        c := -1;
        for i := 0 to SCount - 1 do begin
          //If this state is already connected, find its possible groupists
          if Connected[i] = 1 then begin
            for j := 0 to SCount - 1 do begin
              //This state is not connected yet, and was not previously grouped
              if Connected[j] = 0 then begin
                if prStates[j].Cls = Group.Cls then begin
                  if (Boolean(QAllowed[i, j])) or{and} (Boolean(QAllowed[j, i])) then begin
                    Connected[j] := 1;
                    c := 0;
                    Inc(Group.StateCount);
                  end;
                end;
              end;
            end;
          end;
        end;
      until(c = -1);

      Group.States  := IAlloc1(Group.StateCount);
      Group.Lambdas := DAlloc1(Group.StateCount);
      Group.Taus    := DAlloc1(Group.StateCount);
      Group.Areas   := DAlloc1(Group.StateCount);
      j := 0;
      for i := 0 to SCount - 1 do begin
        if Connected[i] = 1 then begin
          Group.States[j]  := i;
          Group.Lambdas[j] := 0;//Replace it with calculated value
          Group.Taus[j]    := 0;//Replace it with calculated value
          Group.Areas[j]   := 0;//Replace it with calculated value
          Inc(j);
          Grouped[i] := 1;
        end;
      end;


    until (s = -1);

    for i := 0 to prGroupCount - 1 do begin
      for j := 0 to TQUB_QFS_QMF_PGroup(prGroups[i]).StateCount - 1 do begin
        prStates[TQUB_QFS_QMF_PGroup(prGroups[i]).States[j]].Gr := i;
      end;
    end;


    Q := nil;
    A := nil;

    try

      //For each group, calculate its Taus and Areas
      for g := 0 to prGroupCount - 1 do begin

        Group := Groups[g];

        Free2(TPMChar(Q));
        Free2(TPMChar(A));
        Q := nil;
        A := nil;
        Q := DAlloc2(Group.StateCount, Group.StateCount);
        A := DAlloc2(Group.StateCount, Group.StateCount);

        for i := 0 to Group.StateCount - 1 do begin
          for j := 0 to Group.StateCount - 1 do begin
            Q[i, j] := prQ[Group.States[i], Group.States[j]];
          end;
        end;
        MakeAMatrixPEq(Q, A, nil, nil{Ak}, Group.Lambdas, Group.Taus, Group.Areas, prADt, Group.StateCount);

      end;

    finally

      Free2(TPMChar(Q));
      Free2(TPMChar(A));

    end;


  finally


    Free2(TPMChar(QAllowed));
    FreeMem(Grouped);
    FreeMem(Connected);


  end;


end;
*)


procedure TQUB_QFS_QMF.prClearMath;
var
  k  :Integer;
begin
  try
    Free2(TPMChar(prQ));
    Free2(TPMChar(prA));
    Free2(TPMChar(prP));
    FreeMem(prPt);
    FreeMem(prP0);
    FreeMem(prPEq);
    if prAk <> nil then begin
      for k := 0 to prMathStateCount - 1 do begin
        Free2(TPMChar(prAk[k]));
      end;
    end;
    FreeMem(prAk);
    FreeMem(prLambdas);
    FreeMem(prFreePars);
    prQ        := nil;
    prA        := nil;
    prP        := nil;
    prPt       := nil;
    prP0       := nil;
    prPEq      := nil;
    prLambdas  := nil;
    prAk       := nil;
    prFreePars := nil;
  finally
    prMathStateCount := 0;
    prFreeParsCount  := 0;
  end;
end;

procedure TQUB_QFS_QMF.prReAllocMath;
var
  k :Integer;
begin

  if prMathStateCount <> States.StateCount then begin
    prClearMath;
  end;

  prMathStateCount := States.StateCount;
  if prMathStateCount < 1{2} then Exit;

  //Q Matrix
  prQ := DAlloc2(prMathStateCount, prMathStateCount);

  //P Matrix of state change probabilities
  prP := DAlloc2(prMathStateCount + 1, prMathStateCount);

  //A Matrix and P equilibrium
  prA       := DAlloc2(prMathStateCount, prMathStateCount);
  prPt      := DAlloc1(prMathStateCount);
  prP0      := DAlloc1(prMathStateCount);
  prPEq     := DAlloc1(prMathStateCount);

  //Eigen-v-v
  prLambdas := DAlloc1(prMathStateCount);
  prAk      := PAlloc1(prMathStateCount);
  for k := 0 to prMathStateCount - 1 do begin
    prAk[k] := DAlloc2(prMathStateCount, prMathStateCount);
  end;

  //FreePars
  prFreeParsCount := prMathStateCount * prMathStateCount;
  prFreePars:= DAlloc1(prFreeParsCount);

end;

function TQUB_QFS_QMF.CalculateMath(NeedPars :Boolean = True): Boolean;
var
  k :Integer;
begin

  prReAllocMath;

  Result := prMathStateCount < 1;
  if Result then Exit;

  MakeQMatrix(Self, prQ);
  MakePMatrix(prQ, prP, prMathStateCount);
  MakeAMatrixPEq(prQ, prA, prPEq, prAk, prLambdas, nil, nil, prADt, prMathStateCount);
  MakeInitProbs(Self, prP0, prMathStateCount);
  for k := 0 to prMathStateCount - 1 do begin
    prPt[k] := prP0[k];
  end;

  if NeedPars then prMakeFreePars;

  //Find the groups of aggregated states. Called here for testing, mainly
  //FindGroups;

end;

function TQUB_QFS_QMF.CalculateProbEq(DoAssign :Boolean) :Boolean;
var
  i :Integer;
begin

  Result := CalculateMath;
  if DoAssign then begin
    if not Result then begin
      for i := 0 to States.StateCount - 1 do begin
        States[i].Pr := prPEq[i];
      end;
    end;
  end;

end;

function TQUB_QFS_QMF.AdvanceProb(t :Double) :Boolean;
var
  QModel     :TQUB_Q_Model;
  ParamsList :TQUB_ExpCond_ParametersList;
  i          :Integer;
begin

  Result     := False;

  ParamsList := TQUB_ExpCond_ParametersList.Create;
  ParamsList.Add(Parameters);
  QModel     := nil;

  try

    try

      QModel := TQUB_Q_Model.Create(Self, True, ParamsList);
      QModel.Do_fParsToModel;
      QModel.Do_P0;
      QModel.Do_At(t, 0);
      QModel.Do_Pt(t, 0);

      for i := 0 to States.StateCount - 1 do begin
        States[i].Pr := QModel.Pt[i];
      end;

    except
      Result := True;
    end;

  finally

    QModel.Free;
    ParamsList.Delete(0);//doesn't own them, so make sure they are not freed
    ParamsList.Free;

  end;

end;

function TQUB_QFS_QMF.MakeStartState(StateIdx :Integer) :Boolean;
var
  i :Integer;
begin

  Result := False;

  for i := 0 to States.StateCount - 1 do begin
    States[i].Pr := 0;
  end;
  States[StateIdx].Pr := 1;

end;

function TQUB_QFS_QMF.EnforceConstraints :Boolean;
var
  QModel     :TQUB_Q_Model;
  ParamsList :TQUB_ExpCond_ParametersList;
begin

  Result     := False;

  if States.StateCount = 0 then Exit;

  ParamsList := TQUB_ExpCond_ParametersList.Create;
  ParamsList.Add(Parameters);
  QModel     := nil;

  try

    try
      QModel := TQUB_Q_Model.Create(Self, True, ParamsList);
      QModel.Do_fParsTocPars;
      QModel.Do_cParsToQ;
      QModel.Do_cParsToEKP;
      MakeRatesFromParams(Self, QModel.cPars, 2*QModel.RateCount, False{AllowNegRates});
      MakeExtraKinPars(Self, QModel.ExtraKinPars, QModel.ExtraKinParCount);
    except
      Result := True;
    end;

  finally

    QModel.Free;
    ParamsList.Delete(0);//doesn't own them, so make sure they are not freed
    ParamsList.Free;

  end;

end;

function TQUB_QFS_QMF.prMakeFreePars :Boolean;
var
  QModel     :TQUB_Q_Model;
  ParamsList :TQUB_ExpCond_ParametersList;
  i          :Integer;
begin

  Result     := False;

  if States.StateCount = 0 then Exit;

  ParamsList := TQUB_ExpCond_ParametersList.Create;
  ParamsList.Add(Parameters);
  QModel     := nil;

  try

    try
      QModel := TQUB_Q_Model.Create(Self, True, ParamsList);
      prFreeParsCount := QModel.fParCount;
      for i := 0 to prFreeParsCount - 1 do begin
        prFreePars[i] := QModel.fPars[i];
      end;
    except
      Result := True;
    end;

  finally

    QModel.Free;
    ParamsList.Delete(0);//doesn't own them, so make sure they are not freed
    ParamsList.Free;

  end;

end;

{
This function calculates the free parameters corresponding to the current rates,
and perturbs them by SE (a gaussian with Avg = freepar, Std = freepar*SE)
LM.
}
function TQUB_QFS_QMF.RandomizeRates(SE :Double) :Boolean;
var
  QModel     :TQUB_Q_Model;
  ParamsList :TQUB_ExpCond_ParametersList;
  i          :Integer;
begin

  Result     := False;

  ParamsList := TQUB_ExpCond_ParametersList.Create;
  ParamsList.Add(Parameters);
  QModel     := nil;

  try

    try
      QModel := TQUB_Q_Model.Create(Self, True, ParamsList);

      for i := 0 to QModel.fParCount - 1 do begin
        QModel.fPars[i] := RandG(QModel.fPars[i], QModel.fPars[i]*SE);
      end;

      QModel.Do_fParsTocPars;
      QModel.Do_cParsToQ;
      QModel.Do_cParsToEKP;
      MakeRatesFromParams(Self, QModel.cPars, 2*QModel.RateCount, False{AllowNegRates});
      MakeExtraKinPars(Self, QModel.ExtraKinPars, QModel.ExtraKinParCount);
    except
      Result := True;
    end;

  finally

    QModel.Free;
    ParamsList.Delete(0);//doesn't own them, so make sure they are not freed
    ParamsList.Free;

  end;

end;


//TO DO: check how it works. LM
function TQUB_QFS_QMF.ConstrainedSetRate(StateFrom, StateTo :Integer; IsK0 :Boolean; Value :Double): Boolean;
var
  RateIdx, HalfIdx      :Integer;
  Rate                  :TQUB_QFS_Rate;
  Ratio                 :Double;
  WeFixRate             :Boolean;
  i                     :Integer;
  Cnstr                 :TQUB_QFS_Constraint;
  ScaledRates           :INode;
  ScaledRate            :INode;

  procedure ScaleK0(s1, s2 :Integer);
  var
    ix, half            :Integer;
    ScaleRate           :TQUB_QFS_Rate;
  begin
    Rates.GetRateIdxHalfIdxOfStates(s1, s2, ix, half);
    ScaleRate           := Rates[ix];
    ScaleRate.K0[half]  := ScaleRate.K0[half] * Ratio;
  end;

  procedure ScaleK1(s1, s2 :Integer);
  var
    ix, half            :Integer;
    ScaleRate           :TQUB_QFS_Rate;
  begin
    Rates.GetRateIdxHalfIdxOfStates(s1, s2, ix, half);
    ScaleRate           := Rates[ix];
    ScaleRate.K1[half]  := ScaleRate.K1[half] * Ratio;
  end;
begin
  Result                := False;

  Rates.GetRateIdxHalfIdxOfStates(StateFrom, StateTo, RateIdx, HalfIdx);
  Rate                  := Rates[RateIdx];
  // set rate
  if IsK0 then begin
    Ratio               := Value / Rate.K0[HalfIdx];
    Rate.K0[HalfIdx]    := Value;
  end else begin
    Ratio               := Value / Rate.K1[HalfIdx];
    Rate.K1[HalfIdx]    := Value;
  end;

  // update any rates which are scaled to this one
  if IsK0 then begin
    ScaledRates         := GetScaledRates(CloneAsNode(True, TNode.CreateNull)['Constraints'], StateFrom, StateTo);
  end else begin
    ScaledRates         := GetScaledExpRates(CloneAsNode(True, TNode.CreateNull)['Constraints'], StateFrom, StateTo);
  end;

  ScaledRate            := ScaledRates.Find('Rate');
  while not ScaledRate.IsNull do begin
    if IsK0 then begin
      ScaleK0(ScaledRate.Data[0], ScaledRate.Data[1]);
    end else begin
      ScaleK1(ScaledRate.Data[0], ScaledRate.Data[1]);
    end;
    ScaledRate          := ScaledRate.SiblingSameName;
  end;

  // check if rate already fixed, otherwise we must fix it
  // so it is not changed by the loop balancing in EnforceConstraints
  WeFixRate             := True;
  for i := 0 to Constraints.ConstraintCount - 1 do begin
    Cnstr               := Constraints[i];
    if IsK0 then begin
      if (Cnstr.Ctype = CSTR_FIX_RATE) and (Cnstr.Index[0] = StateFrom) and (Cnstr.Index[1] = StateTo) then begin
        WeFixRate       := False;
      end;
    end else begin
      if (Cnstr.Ctype = CSTR_FIX_EXP) and (Cnstr.Index[0] = StateFrom) and (Cnstr.Index[1] = StateTo) then begin
        WeFixRate       := False;
      end;
    end;
  end;

  // if rate not fixed: fix it
  if WeFixRate then begin
    if IsK0 then begin
      Cnstr := TQUB_QFS_Constraint.Create(CHANNELS_CONSTRAINT_TYPE_NAME[CSTR_FIX_RATE], nil);
    end else begin
      Cnstr := TQUB_QFS_Constraint.Create(CHANNELS_CONSTRAINT_TYPE_NAME[CSTR_FIX_EXP], nil);
    end;
    Cnstr.Index[0] := StateFrom;
    Cnstr.Index[1] := StateTo;
    Constraints.AddConstraint(Cnstr, Constraints.ConstraintCount);
  end;

  EnforceConstraints();

  // if we fixed rate: unfix it
  if WeFixRate then begin
    Constraints.DelConstraint(Constraints.ConstraintCount - 1);
  end;
end;

//State/Rate/Constraint EDIT
procedure TQUB_QFS_QMF.prOnEdit(Sender :TObject);
begin
  prModified := True;
  if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
  if not prEditing then begin
    if Sender is TQUB_QFS_Rate then begin
      if Assigned(OnNeedDraw) then begin
        TQUB_QFS_Rate(Sender).Update;
        OnNeedDraw(Self);
      end;
    end;
  end;
end;


//Change the class of a state
procedure TQUB_QFS_QMF.ChangeStateClass(State :TQUB_QFS_State; c :Integer);
var
  i :Integer;
begin

  // next line: ugly hack to make room for the new class before anything can mess with it
  //if (State.Cls + 1) = prModelObject.ClassCount then begin
  //  prModelObject.Amp[State.Cls + 1] := 0.0;
  //end;
  if c >= ClassCount then begin
    Amp[c] := 0.0;
  end;

  State.Cls  := c;
  State.Amp  := Amp[State.Cls];
  State.Std  := Std[State.Cls];
  State.Cond := Cond[State.Cls];
  State.CondStd := CondStd[State.Cls];
  State.Vrev    := ClsVrev[State.Cls];
  for i := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
    State.Cff[i] := Cff[State.Cls, i];
  end;

  States.UpdateStates;

end;



//Can's ************************************************************************



//New File Properties
procedure TQUB_QFS_QMF.OnGetPropertiesNew(Sender :TObject);
begin
  prFileName  := TQUB_Viewport_Model_New_PropertiesForm(Sender).FileName;
  prFileType  := TQUB_Viewport_Model_New_PropertiesForm(Sender).FileFormat;
end;

procedure TQUB_QFS_QMF.OnPutPropertiesNew(Sender :TObject);
begin
end;


{LM.
The problem: if one model gets ChannelCount changed, then all other models (and data files) will.
This is not right, so I commented out the code below.
TODO Was this something chris changed or something old Lorin just discovered ?
  It's something Chris did, fairly long ago. The idea seemed good at the time, 
  but when I started running optimizations with multiple data and model files, 
  unexpected and misterious side effects started to happen.
  02-03-05. LM.
John, if you know of a situation when it should be this way, let me know, and we'll treat it as an exception.

OK Lorin this should make us all happy: QMF ignores the expcond channel count,
which is only there for global fits,
and indeed actively removed otherwise
(mainly still touched by qub_*_Model, qub_datalist, qubopt)
--11/05 chris
}

procedure TQUB_QFS_QMF.prSetChannelCount(c :Integer);
//var
//  Par    :TQUB_ExpCond_PParameter;
begin
  (*
  if (prParameters <> nil) and (prParameters <> ViewportModel.DefExpCond) then begin
    //testing
    Par := prParameters.ParameterByName['ChannelCount'];
    if Par <> nil then begin
      if c <> Round(Par.Value) then begin
        prParameters.ChangePar('ChannelCount', c);
      end;
    end else begin
      prParameters.AddPar('ChannelCount', c, -1, False);
    end;
  end;
  *)
  if c <> ChannelCount then begin
    prRecChannelCount.DataValue := c;

    prModified := True;
    if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
  end;
end;

function TQUB_QFS_QMF.prGetChannelCount :Integer;
//var
//  Par    :TQUB_ExpCond_PParameter;
begin
  Result := -1;

  if prParameters <> nil then begin
    //testing - should it get it from there?
    {
    Par := prParameters.ParameterByName['ChannelCount'];
    if Par <> nil then begin
      Result := Round(Par.Value);
    end;
    }
  end;

  if Result = -1 then Result := prRecChannelCount.DataValue;
end;

function TQUB_QFS_QMF.prGetClassCount: Integer;
begin
  Result := prRecAmps.DataCount;
end;

procedure TQUB_QFS_QMF.prSetVRev(v :Double);
begin
  if v <> VRev then begin
    prRecVRev.DataValue := v;
    prModified := True;
    if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
  end;
end;

function TQUB_QFS_QMF.prGetVRev :Double;
begin
  Result := prRecVRev.DataValue;
end;

//State properties
procedure TQUB_QFS_QMF.prEnsureClassExists(Index :Integer);
var
  PrevMax             :Integer;
  i                   :Integer;
begin
  // do ars outside if/loop since there may be fewer than Nars
  while Index >= prRecArs.RecordCount do begin
    TQUB_QFS_Record.CreateAsValue(CHANNELS_AR_NAME, prRecArs, False{Perm}, 0, NODE_DATATYPE_DOUBLE, QUB_QFS_QMF_MAX_COEFF);
    for i := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
      prRecArs.Records[ prRecArs.RecordCount - 1 ].DataArrayValue[i] := 0.0;
    end;
  end;

  if Index >= ClassCount then begin
    PrevMax           := ClassCount - 1;

    for i := Index downto PrevMax + 1 do begin
      prRecAmps.DataArrayValue[i]  := i;//?
      prRecStds.DataArrayValue[i]  := prRecStds.DataArrayValue[PrevMax];
      prRecConds.DataArrayValue[i] := prRecConds.DataArrayValue[PrevMax];
      prRecCondStds.DataArrayValue[i] := prRecCondStds.DataArrayValue[PrevMax];
      prRecVrevs.DataArrayValue[i]    := prRecVrevs.DataArrayValue[PrevMax];
      prRecNar.DataArrayValue[i]   := 0;
    end;
  end;
end;

procedure TQUB_QFS_QMF.prSetAmp(Index :Integer; a :Double);
var
  i :Integer;
begin
  prEnsureClassExists(Index);
  prRecAmps.DataArrayValue[Index] := a;
  for i := 0 to prStates.StateCount - 1 do begin
    prStates.State[i].Amp := Amp[prStates.State[i].Cls];
  end;
end;

function  TQUB_QFS_QMF.prGetAmp(Index :Integer) :Double;
begin
  prEnsureClassExists(Index);
  Result := prRecAmps.DataArrayValue[Index];
end;

procedure TQUB_QFS_QMF.prSetStd(Index :Integer; s :Double);
var
  i :Integer;
begin
  prEnsureClassExists(Index);
  prRecStds.DataArrayValue[Index] := s;
  for i := 0 to prStates.StateCount - 1 do begin
    prStates.State[i].Std := Std[prStates.State[i].Cls];
  end;
end;

function  TQUB_QFS_QMF.prGetStd(Index :Integer) :Double;
begin
  prEnsureClassExists(Index);
  Result := prRecStds.DataArrayValue[Index];
end;

procedure TQUB_QFS_QMF.prSetCond(Index :Integer; c :Double);
var
  i :Integer;
begin
  prEnsureClassExists(Index);
  prRecConds.DataArrayValue[Index] := c;
  for i := 0 to prStates.StateCount - 1 do begin
    prStates.State[i].Cond := Cond[prStates.State[i].Cls];
  end;
end;

function  TQUB_QFS_QMF.prGetCond(Index :Integer) :Double;
begin
  prEnsureClassExists(Index);
  Result := prRecConds.DataArrayValue[Index];
end;

procedure TQUB_QFS_QMF.prSetCondStd(Index :Integer; x :Double);
var
  i :Integer;
begin
  prEnsureClassExists(Index);
  prRecCondStds.DataArrayValue[Index] := x;
  for i := 0 to prStates.StateCount - 1 do begin
    prStates.State[i].CondStd := CondStd[prStates.State[i].Cls];
  end;
end;

function  TQUB_QFS_QMF.prGetCondStd(Index :Integer) :Double;
begin
  prEnsureClassExists(Index);
  Result := prRecCondStds.DataArrayValue[Index];
end;

procedure TQUB_QFS_QMF.prSetClsVrev(Index :Integer; x :Double);
var
  i :Integer;
begin
  prEnsureClassExists(Index);
  prRecVrevs.DataArrayValue[Index] := x;
  for i := 0 to prStates.StateCount - 1 do begin
    prStates.State[i].Vrev := ClsVrev[prStates.State[i].Cls];
  end;
end;

function  TQUB_QFS_QMF.prGetClsVrev(Index :Integer) :Double;
begin
  prEnsureClassExists(Index);
  Result := prRecVrevs.DataArrayValue[Index];
end;

function  TQUB_QFS_QMF.prGetEffVrev(Index :Integer): Double;
begin
  Result := ClsVrev[Index];
  if Abs(Result - UNSET_VALUE) < 1e-8 then Result := Vrev;
end;

procedure TQUB_QFS_QMF.prSetNar(Index :Integer; n :Integer);
begin
  prEnsureClassExists(Index);
  if n <> prRecNar.DataArrayValue[Index] then begin
    prModified := True;
    prRecNar.DataArrayValue[Index] := n;
    if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, nil^, nil^);
  end;
end;

function  TQUB_QFS_QMF.prGetNar(Index :Integer) :Integer;
begin
  prEnsureClassExists(Index);
  Result := prRecNar.DataArrayValue[Index];
end;

procedure TQUB_QFS_QMF.prSetCff(Index1{Class}, Index2{ArIdx} :Integer; c :Double);
var
  i :Integer;
begin
  prEnsureClassExists(Index1);
  prRecArs[Index1].DataArrayValue[Index2] := c;
  for i := 0 to prStates.StateCount - 1 do begin
    if prStates.State[i].Cls = Index1 then begin
      prStates.State[i].Cff[Index2] := Cff[Index1, Index2];
    end;
  end;
end;

function  TQUB_QFS_QMF.prGetCff(Index1, Index2 :Integer) :Double;
begin
  prEnsureClassExists(Index1);
  Result := prRecArs[Index1].DataArrayValue[Index2];
end;


//******************************************************************************
end.

