{

                             Qub Q Matrix & Co File
                                     unit


(C) 2002 Lorin Milescu
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


//01-09-2002   Creation date


unit QUB_QFS_QMF_QMatrix;


//******************************************************************************
interface


uses


SysUtils,

AllocTypes, Alloc,

FMath, QUB_Q_Global, QUB_Eigen, QUB_Matrices,
QUB_QFS_QMF;


//******************************************************************************
function MakeQMatrix(Model :TQUB_QFS_QMF; Q :TPMDouble) :Boolean; forward;
function MakePMatrix(Q, P :TPMDouble; StateCount :Integer) :Boolean; forward;
function MakeAMatrixPEq(Q, A :TPMDouble; PEq :TPADouble; Ak :TPAPointer; Lambdas, Taus, Areas :TPADouble; Sampling :Double; StateCount :Integer) :Boolean; forward; overload;
function MakeAMatrixPEq(Q, A :TPMDouble; PEq :TPADouble; Ak :TPAPointer; Lambdas, Taus, Areas :TPADouble; Sampling :Double; StateCount :Integer;
                        qq, v, vv, vi :TPMDouble;
                        aa            :TPAPointer;
                        wr, wi        :TPADouble) :Boolean; forward; overload;
function MakeAMatrix(A :TPMDouble; Ak :TPAPointer; Lambda :TPADouble; t :Double; StateCount :Integer) :Boolean; forward;
function MakeRatesFromQMatrix(Model :TQUB_QFS_QMF; Q :TPMDouble) :Boolean; forward;
function MakeRatesFromParams(Model :TQUB_QFS_QMF; Params :TPADouble; ParamCount :Integer; AllowNegRates :Boolean) :Boolean; forward;
function MakeExtraKinPars(Model :TQUB_QFS_QMF; ExtraKinPars :TPADouble; ExtraKinParCount :Integer) :Boolean; forward;
function MakeInitProbs(Model :TQUB_QFS_QMF; P0 :TPADouble; StateCount :Integer) :Integer; forward;
function MakePeq (Q :TPMDouble; Peq :TPADouble; StateCount :Integer) :Boolean; forward;
function MakeAmpVar(Model :TQUB_QFS_QMF; var m0, V0 :Double; Amps, Vars :TPADouble; StateCount :Integer) :Integer; forward;
function MakeCondVar(Model :TQUB_QFS_QMF; var m0, V0 :Double; Amps, Vars :TPADouble; StateCount :Integer) :Integer; forward;



procedure QUBLIB_Eigen(a :TPMDouble; n :Integer; wr, wi :TPADouble; v :TPMDouble); stdcall; external 'qublib.dll' name '_eigen@20';




//testing
//procedure QuBEigen_Eigen(a :TPMDouble; n :Integer; wr, wi :TPADouble; v :TPMDouble); stdcall; external 'qubeigen.dll' name 'eigen';
//*******

//******************************************************************************
implementation


uses


QUB_Q_Model,
QUB_Rich_Console, QUB_Console;


//******************************************************************************
function MakeQMatrix(Model :TQUB_QFS_QMF; Q :TPMDouble) :Boolean;
var
  sc     :Integer;
  i, j   :Integer;
begin

  sc := Model.States.StateCount;
  Result := sc <= 0;
  if Result then Exit;

  try

    for i := 0 to sc - 1 do begin
      for j := 0 to sc - 1 do begin
        if j <> i then begin
          Q[i, j] := Model.Rates.RateValueOfStates[Model.States[i], Model.States[j]];
        end;
      end;
    end;

    //Correct for zero rates ?
    for i := 0 to sc - 1 do begin
      for j := 0 to sc - 1 do begin
        if j <> i then begin
          if (Q[i, j] =  0) and (Q[j, i] <> 0) then begin
            Q[i, j] := MACHEP;
          end;
          if (Q[i, j] <> 0) and (Q[j, i] =  0) then begin
            Q[j, i] := MACHEP;
          end;
        end;
      end;
    end;

    for i := 0 to sc - 1 do begin
      Q[i, i] := 0;
      for j := 0 to sc - 1 do begin
        if j <> i then begin
          Q[i, i] := Q[i, i] - Q[i, j];
        end;
      end;
    end;

  except

    Result := True;

  end;
end;

function MakeRatesFromQMatrix(Model :TQUB_QFS_QMF; Q :TPMDouble) :Boolean;
var
  sc     :Integer;
  i, j   :Integer;
begin

  sc := Model.States.StateCount;
  Result := sc <= 0;
  if Result then Exit;

  try

    for i := 0 to sc - 1 do begin
      for j := 0 to sc - 1 do begin
        if j <> i then begin
          Model.Rates.SetRateK0OfStates(Model.States[i], Model.States[j], Q[i, j]);
        end;
      end;
    end;

    //Correct for zero rates
    for i := 0 to sc - 1 do begin
      for j := 0 to sc - 1 do begin
        if j <> i then begin
          if (Q[i, j] =  0) and (Q[j, i] <> 0) then Q[i, j] := MACHEP;
          if (Q[i, j] <> 0) and (Q[j, i] =  0) then Q[j, i] := MACHEP;
        end;
      end;
    end;

  except

    Result := True;

  end;
end;

function MakeRatesFromParams(Model :TQUB_QFS_QMF; Params :TPADouble; ParamCount :Integer; AllowNegRates :Boolean) :Boolean;
var
  k :Integer;
begin

  Result := False;

  try

    for k := 0 to (ParamCount div 4) - 1 do begin

      if AllowNegRates then begin
        Model.Rates[k].K0[1] := Params[4*k + 0];
        Model.Rates[k].K0[2] := Params[4*k + 2];
      end else begin
        Model.Rates[k].K0[1] := Exp(Params[4*k + 0]);
        Model.Rates[k].K0[2] := Exp(Params[4*k + 2]);
      end;
      Model.Rates[k].K1[1] :=   Params[4*k + 1];
      Model.Rates[k].K1[2] :=   Params[4*k + 3];

    end;//k

  except

    Result := True;

  end;
end;

function MakeExtraKinPars(Model :TQUB_QFS_QMF; ExtraKinPars :TPADouble; ExtraKinParCount :Integer) :Boolean;
var
  k :Integer;
begin

  Result := False;

  for k := 0 to ExtraKinParCount - 1 do begin

    Model.ExtraKineticPars[k].Value := ExtraKinPars[k];

  end;//k

end;

function MakePMatrix(Q, P :TPMDouble; StateCount :Integer) :Boolean;
var
  i, j, jj :Integer;
begin
  Result := False;

  try

    for i := 0 to StateCount - 1 do begin
      jj := 0;
      for j := 0 to StateCount - 1 do begin
        if j <> i then begin
          if Q[i, i] = 0 then begin
            P[i, jj] := 0;
          end else begin
            P[i, jj] := -(Q[i, j] / Q[i, i]);
          end;
          Inc(jj);
        end;
      end;
    end;

  except

    Result := True;

  end;
end;

function MakeAMatrixPEq(Q, A :TPMDouble; PEq :TPADouble; Ak :TPAPointer; Lambdas, Taus, Areas :TPADouble; Sampling :Double; StateCount :Integer) :Boolean;
var
  i, j, k     :Integer;
  qq          :TPMDouble;
  v           :TPMDouble;
  vv          :TPMDouble;
  vi          :TPMDouble;
  aa          :TPAPointer;
  wr, wi      :TPADouble;
  et          :Double;
  t           :Double;
  Sorted      :Boolean;
  //m           :String;
begin

  qq  := nil;
  v   := nil;
  vv  := nil;
  vi  := nil;
  aa  := nil;
  wr  := nil;
  wi  := nil;

  try

    try


      qq  := DAlloc2(StateCount, StateCount);
      v   := DAlloc2(StateCount, StateCount);
      vv  := DAlloc2(StateCount, StateCount);
      vi  := DAlloc2(StateCount, StateCount);
      aa  := PAlloc1(StateCount);
      for i := 0 to StateCount - 1 do aa[i] := DAlloc2(StateCount, StateCount);
      wr  := DAlloc1(StateCount);
      wi  := DAlloc1(StateCount);


      //qq := -Q
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          qq[i, j] := - Q[i, j];
        end;
      end;


      //Find Eigenvalues and Eigenvectors
      QUBLIB_Eigen(qq, StateCount, wr, wi, v);
      //QuBEigen_Eigen(qq, StateCount, wr, wi, v);





      (*
      m := 'Eigen (DLL):'#13#10'';
      m := m + 'Values:'#13#10;
      for i := 0 to StateCount - 1 do begin
        //for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(wr[i]) + #13#10;
        //end;
      end;
      m := m + 'Vectors:'#13#10;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(v[i, j]) + #9;
        end;
        m := m + #13#10;
      end;
      Console.NewEntry(m, COLOR_WARNING);
      *)


      (*
      //qq := -Q
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          qq[i, j] := - Q[i, j];
          wr[i] := 0;
        end;
      end;
      //Find Eigenvalues and Eigenvectors
      QuBEigen_Eigen(qq, StateCount, wr, wi, v);
      m := 'Eigen (Test):'#13#10'';
      m := m + 'Values:'#13#10;
      for i := 0 to StateCount - 1 do begin
        //for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(wr[i]) + #13#10;
        //end;
      end;
      m := m + 'Vectors:'#13#10;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(v[i, j]) + #9;
        end;
        m := m + #13#10;
      end;
      Console.NewEntry(m, COLOR_ERROR);
      *)





      //Transpose v col/row into row/col = put Eigen vectors in rows, so they can be easily sorted
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          vv[i, j] := v[j, i];
        end;
      end;

      //Sort Eigen values and Eigen vectors in ascending order:
      repeat
        Sorted := True;
        for i := 0 to StateCount - 2 do begin
          if wr[i] > wr[i + 1] then begin
            Sorted := False;
            t := wr[i]; wr[i] := wr[i + 1]; wr[i + 1] := t;
            t := wi[i]; wi[i] := wi[i + 1]; wi[i + 1] := t;
            for j := 0 to StateCount - 1 do begin
              t := vv[i, j]; vv[i, j] := vv[i + 1, j];  vv[i + 1, j] := t;
            end;
          end;
        end;
      until(Sorted);


      //Transpose v back into col/row
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          v[i, j] := vv[j, i];
        end;
      end;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          vv[i, j] := v[i, j];
        end;
      end;



      {
      m := 'Eigen (DLL):'#13#10'';
      m := m + 'Values:'#13#10;
      for i := 0 to StateCount - 1 do begin
        //for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(wr[i]) + #13#10;
        //end;
      end;
      m := m + 'Vectors:'#13#10;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(vv[i, j]) + #9;
        end;
        m := m + #13#10;
      end;
      Console.NewEntry(m, COLOR_WARNING);
      }


      (*
      m := 'Sorted(v):'#13#10;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(vv[i, j]) + #9;
        end;
        m := m + #13#10;
      end;
      Console.NewEntry(m, COLOR_WARNING);
      *)



      Result := InvMat(PMatrix(vv), 0, StateCount - 1, PMatrix(vi)) <> MAT_OK;
      //Now vi is v-1





      (*
      m := 'Inv(v):'#13#10;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          m := m + FloatToStr(vi[i, j]) + #9;
        end;
        m := m + #13#10;
      end;
      Console.NewEntry(m, COLOR_WARNING);
      *)







      //Clear A Matrix
      if A <> nil then begin
        for i := 0 to StateCount - 1 do begin
          for j := 0 to StateCount - 1 do begin
            A[i, j] := 0;
          end;
        end;
      end;

      //Find spectral matrices
      for k := 0 to StateCount - 1 do begin
        //postmultiply each column k of v with the corresponding row k of vi
        for i := 0 to StateCount - 1 do begin
          for j := 0 to StateCount - 1 do begin
            TPMDouble(aa[k])[i, j] := v[i, k] * vi[k, j];
          end;
        end;
      end;

      //Find spectral matrices
      for k := 0 to StateCount - 1 do begin

        if Ak <> nil then begin
          for i := 0 to StateCount - 1 do begin
            for j := 0 to StateCount - 1 do begin
              TPMDouble(Ak[k])[i, j] := TPMDouble(aa[k])[i, j];
            end;
          end;
        end;

        //postmultiply each column k of v with the corresponding row k of vi
        if A <> nil then begin
          et := exp(- wr[k] * Sampling);
          for i := 0 to StateCount - 1 do begin
            for j := 0 to StateCount - 1 do begin
              A[i, j] := A[i, j] + TPMDouble(aa[k])[i, j] * et;
            end;
          end;
          if PEq <> nil then begin
            if k = 0 then begin
              for j := 0 to StateCount - 1 do begin
                PEq[j] := A[0, j];
              end;
            end;
          end;
        end;

      end;


      //Calculate Time Constants and Areas
      if (Taus <> nil) and (Areas <> nil) then begin

        //Taus are the inverse of EigenValues, except the first
        if wr[0] < 1e-10 then begin
          wr[0] := 0;
        end;
        if wr[0] = 0 then begin
          Taus[0] := MAXNUM;
        end else begin
          Taus[0] := 1 / wr[0];
        end;
        for i := 1 to StateCount - 1 do begin
          Taus[i] := 1 / wr[i];
        end;

        //The Areas cannot be calculated without the starting probability vector
        for i := 0 to StateCount - 1 do begin
          Areas[i] := -Taus[i];
        end;

      end;

      //Write Lambdas
      if (Lambdas <> nil) then begin
        Lambdas[0] := 0;
        for i := 1 to StateCount - 1 do begin
          Lambdas[i] := -wr[i];
        end;
      end;

      if StateCount = 1 then begin
        if A <> nil then begin
          A[0, 0] := 1;
        end;
        if PEq <> nil then begin
          PEq[0]  := 1;
        end;
        if Taus <> nil then begin
          Taus[0] := 0;
        end;
        if Areas <> nil then begin
          Areas[0] := 1;
        end;
      end;

    except

      Result := True;

    end;

  finally

    Free2(TPMChar(qq));
    Free2(TPMChar(v));
    Free2(TPMChar(vv));
    Free2(TPMChar(vi));
    if aa <> nil then for i := 0 to StateCount - 1 do Free2(TPMChar(aa[i]));//This was a memory leak! LM
    FreeMem(aa);
    FreeMem(wr);
    FreeMem(wi);

  end;

end;

function MakeAMatrixPEq(Q, A :TPMDouble; PEq :TPADouble; Ak :TPAPointer; Lambdas, Taus, Areas :TPADouble; Sampling :Double; StateCount :Integer;
                        qq, v, vv, vi :TPMDouble;
                        aa            :TPAPointer;
                        wr, wi        :TPADouble) :Boolean;
var
  i, j, k     :Integer;
  et          :Double;
  t           :Double;
  Sorted      :Boolean;
begin

  try

    try

      //qq := -Q
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          qq[i, j] := - Q[i, j];
        end;
      end;


      //Find Eigenvalues and Eigenvectors
      QUBLIB_Eigen(qq, StateCount, wr, wi, v);
      //QuBEigen_Eigen(qq, StateCount, wr, wi, v);


      //Transpose v col/row into row/col = put Eigen vectors in rows, so they can be easily sorted
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          vv[i, j] := v[j, i];
        end;
      end;

      //Sort Eigen values and Eigen vectors in ascending order:
      repeat
        Sorted := True;
        for i := 0 to StateCount - 2 do begin
          if wr[i] > wr[i + 1] then begin
            Sorted := False;
            t := wr[i]; wr[i] := wr[i + 1]; wr[i + 1] := t;
            t := wi[i]; wi[i] := wi[i + 1]; wi[i + 1] := t;
            for j := 0 to StateCount - 1 do begin
              t := vv[i, j]; vv[i, j] := vv[i + 1, j];  vv[i + 1, j] := t;
            end;
          end;
        end;
      until(Sorted);


      //Transpose v back into col/row
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          v[i, j] := vv[j, i];
        end;
      end;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          vv[i, j] := v[i, j];
        end;
      end;


      Result := InvMat(PMatrix(vv), 0, StateCount - 1, PMatrix(vi)) <> MAT_OK;
      //Now vi is v-1


      //Clear A Matrix
      if A <> nil then begin
        for i := 0 to StateCount - 1 do begin
          for j := 0 to StateCount - 1 do begin
            A[i, j] := 0;
          end;
        end;
      end;

      //Find spectral matrices
      for k := 0 to StateCount - 1 do begin
        //postmultiply each column k of v with the corresponding row k of vi
        for i := 0 to StateCount - 1 do begin
          for j := 0 to StateCount - 1 do begin
            TPMDouble(aa[k])[i, j] := v[i, k] * vi[k, j];
          end;
        end;
      end;

      //Find spectral matrices
      for k := 0 to StateCount - 1 do begin

        if Ak <> nil then begin
          for i := 0 to StateCount - 1 do begin
            for j := 0 to StateCount - 1 do begin
              TPMDouble(Ak[k])[i, j] := TPMDouble(aa[k])[i, j];
            end;
          end;
        end;

        //postmultiply each column k of v with the corresponding row k of vi
        if A <> nil then begin
          et := exp(- wr[k] * Sampling);
          for i := 0 to StateCount - 1 do begin
            for j := 0 to StateCount - 1 do begin
              A[i, j] := A[i, j] + TPMDouble(aa[k])[i, j] * et;
            end;
          end;
          if PEq <> nil then begin
            if k = 0 then begin
              for j := 0 to StateCount - 1 do begin
                PEq[j] := A[0, j];
              end;
            end;
          end;
        end;

      end;


      //Calculate Time Constants and Areas
      if (Taus <> nil) and (Areas <> nil) then begin

        //Taus are the inverse of EigenValues, except the first
        if wr[0] < 1e-10 then begin
          wr[0] := 0;
        end;
        if wr[0] = 0 then begin
          Taus[0] := MAXNUM;
        end else begin
          Taus[0] := 1 / wr[0];
        end;
        for i := 1 to StateCount - 1 do begin
          Taus[i] := 1 / wr[i];
        end;

        //The Areas cannot be calculated without the starting probability vector
        for i := 0 to StateCount - 1 do begin
          Areas[i] := -Taus[i];
        end;

      end;

      //Write Lambdas
      if (Lambdas <> nil) then begin
        Lambdas[0] := 0;
        for i := 1 to StateCount - 1 do begin
          Lambdas[i] := -wr[i];
        end;
      end;

      if StateCount = 1 then begin
        if A <> nil then begin
          A[0, 0] := 1;
        end;
        if PEq <> nil then begin
          PEq[0]  := 1;
        end;
        if Taus <> nil then begin
          Taus[0] := 0;
        end;
        if Areas <> nil then begin
          Areas[0] := 1;
        end;
      end;

    except

      Result := True;

    end;

  finally


  end;

end;

function MakeAMatrix(A :TPMDouble; Ak :TPAPointer; Lambda :TPADouble; t :Double; StateCount :Integer) :Boolean;
var
  i, j, k  :Integer;
begin

  Result := False;

  for i := 0 to StateCount - 1 do begin
    for j := 0 to StateCount - 1 do begin
      A[i, j] := 0;
      for k := 0 to StateCount - 1 do begin
        A[i, j] := A[i, j] + TPMDouble(Ak[k])[i, j] * Exp(Lambda[k] * t);
      end;//k
    end;//j
  end;//i

end;

function MakePeq(Q :TPMDouble; Peq :TPADouble; StateCount :Integer) :Boolean;
var
  R, RT, M :TPMDouble;
  U        :TPADouble;
  i, j, k  :Integer;
begin

  R        := nil;
  RT       := nil;
  U        := nil;
  M        := nil;

  //Peq = 1.(R.RT)-1

  try

    U  := DAlloc1(StateCount);
    R  := DAlloc2(StateCount,     StateCount + 1);
    RT := DAlloc2(StateCount + 1, StateCount);
    M  := DAlloc2(StateCount, StateCount);

    //U
    for i := 0 to StateCount - 1 do begin
      U[i] := 1;
    end;

    //R, RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        R [i, j] := Q[i, j];
        RT[j, i] := Q[i, j];
      end;
    end;

    for k := 0 to StateCount - 1 do begin
      R [k, StateCount] := 1;
      RT[StateCount, k] := 1;
    end;

    //R.RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        M[i, j] := 0;
        for k := 0 to StateCount do begin
          M[i, j] := M[i, j] + R[i, k]*RT[k, j];
        end;
      end;
    end;

    //RRTI = (R.RT)-1
    Result := Boolean(InvMat(PMatrix(M), 0, StateCount - 1, PMatrix(R{re-use it})));

    MulVecMatVec(U, R, Peq, StateCount);

  finally

    FreeMem(U);
    Free2(TPMChar(R));
    Free2(TPMChar(RT));
    Free2(TPMChar(M));

  end;

end;

function MakeInitProbs(Model :TQUB_QFS_QMF; P0 :TPADouble; StateCount :Integer) :Integer;
var
  Sum :Double;
  i   :Integer;
begin

  Result := 0;

  Sum := 0;
  for i := 0 to StateCount - 1 do begin
    P0[i] := Model.States[i].Pr;
    Sum := Sum + P0[i];
  end;
  if Sum = 0 then begin
    for i := 0 to StateCount - 1 do begin
      P0[i] := 1 / StateCount;
    end;
  end else begin
    for i := 0 to StateCount - 1 do begin
      P0[i] := P0[i] / Sum;
    end;
  end;

end;

function MakeAmpVar(Model :TQUB_QFS_QMF; var m0, V0 :Double; Amps, Vars :TPADouble; StateCount :Integer) :Integer;
var
  i :Integer;
begin

  Result := 0;

  m0 := Model.Amp[0];
  V0 := Sqr(Model.Std[0]);

  for i := 0 to StateCount - 1 do begin
    Amps[i]      := Model.Amp[Model.States[i].Cls] - Model.Amp[0];
    if Model.Std[Model.States[i].Cls] >= Model.Std[0] then begin
      Vars[i]    := Sqr(Model.Std[Model.States[i].Cls]) - V0;
    end else begin
      Vars[i]    := 0;
    end;
  end;

end;

function MakeCondVar(Model :TQUB_QFS_QMF; var m0, V0 :Double; Amps, Vars :TPADouble; StateCount :Integer) :Integer;
var
  i :Integer;
begin

  Result := 0;

  m0 := Model.Amp[0];
  V0 := Sqr(Model.Std[0]);

  for i := 0 to StateCount - 1 do begin
    Amps[i]      := Model.Cond[Model.States[i].Cls];
    Vars[i]      := Sqr(Model.CondStd[Model.States[i].Cls]);
  end;

end;


//******************************************************************************
end.
