{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2004
/// QUB FileSystem :: QUB Protocol File
// nonvisual layer: opening and manipulating files, instantiating shapes from python
//
// commented out RepeatMode and RepeatTimes in TQUB_QFS_QPF - looks like the
//  same vars in _Stage are preferable.  6/2005 {JB}
//
// the unrealized goal was to be able to repeat a multi-stage protocol a few times,
// averaging corresponding iterations of each stage to get more reliable data
// (apparently axon has this feature)   -- Chris



unit QUB_QFS_QPF;


interface


uses

Classes, Forms, Dialogs, Sysutils, Math, Graphics, Windows,

Alloc, AllocTypes,

PythonEngine,
QUB_Node;


type

TQUB_QFS_QPF_RepeatMode     = (QUB_QFS_QPF_REPEAT_NONE,
                               QUB_QFS_QPF_REPEAT_TIMES);

TQUB_QFS_QPF_PauseMode      = (QUB_QFS_QPF_PAUSE_NONE,
                               QUB_QFS_QPF_PAUSE_SECONDS,
                               QUB_QFS_QPF_PAUSE_USER);

TQUB_QFS_QPF_Shape          = class;

TQUB_QFS_QPF_ShapeFactory   = class
protected
  prName                    :String;
  prFactory                 :PPyObject;
  prIcon                    :Graphics.TBitmap;

public
  constructor Create(AFactory :PPyObject);
  destructor Destroy; override;

  function Instantiate(Properties :INode): TQUB_QFS_QPF_Shape; // properties can be null

  property Name             :String           read prName;
  property Factory          :PPyObject        read prFactory;
  property Icon             :Graphics.TBitmap read prIcon        write prIcon;  // set by ShapeLib
end;


TQUB_QFS_QPF_Shape          = class
protected
  prShape                   :PPyObject;
  prTree                    :INode;
  prFactory                 :TQUB_QFS_QPF_ShapeFactory;

  function prGetProperty(Nm :String): String;
  procedure prSetProperty(Nm, Val :String);

public

  OnChanged                 :TNotifyEvent;

  constructor Create(AShape :PPyObject; AFactory :TQUB_QFS_QPF_ShapeFactory);
  destructor Destroy; override;

  function SampleCount(TStart, Sampling :Double; Rep_Index :Integer): Integer; overload;
  function SampleCount(TStart, Sampling :Double; Duration :Double): Integer; overload; // for partial counts
  function Duration(Rep_Index :Integer): Double;

  function Generate(First, Last :Integer; TStart, Sampling :Double; Rep_Index :Integer): INode;
  procedure GenerateInto(First, Last :Integer; TStart, Sampling :Double; Rep_Index :Integer; Node :INode; IStart :Integer);
  procedure GetBounds(Sampling :Double; Reps :Integer; var Low, High :Double);
  
  property Shape            :PPyObject                 read prShape;
  property Tree             :INode                     read prTree;
  property Factory          :TQUB_QFS_QPF_ShapeFactory read prFactory;
  property Properties[Nm :String] :String              read prGetProperty       write prSetProperty;
end;


TQUB_QFS_QPF_ShapeLib            = class
protected

  prShapes            :TStringList;

  prAddShapePyType    :TPythonType;
  prGenerateIntoPyType :TPythonType;

  procedure prInitPyFunc;
  procedure prAddQueuedShapes;

  function prGetShapeCount: Integer;
  function prGetShape(NameOrIx :Variant): TQUB_QFS_QPF_ShapeFactory;

public
  OnShapeAdded        :TNotifyEvent;

  constructor Create;
  destructor Destroy; override;

  procedure Add(ShapeFactory :PPyObject);

  property ShapeCount           :Integer             read prGetShapeCount;
  property Shapes[Ix :Variant]  :TQUB_QFS_QPF_ShapeFactory  read prGetShape;       default;

end;

TQUB_QFS_QPF_AddShape       = class(TPyObject)
protected
  prFactory                 :PPyObject;

  procedure prAddShape;
public
  ShapeLib                  :TQUB_QFS_QPF_ShapeLib;

  function  Call( args, kw : PPyObject) : PPyObject; override;
end;

TQUB_ProtocolShape_GenerateInto   = class(TPyObject)
protected
  function prGenerate(Eval :PPyObject; Node :INode; t_off, sampling :Double; first, remain, rep_index, i_start :Integer): PPyObject;
public
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;


TQUB_QFS_QPF_Channel        = class
protected

  prTree                    :INode;
  prShapes                  :TList;

  prShapeLib                :TQUB_QFS_QPF_ShapeLib;

  function prGetShapeCount: Integer;
  function prGetShape(Ix :Integer): TQUB_QFS_QPF_Shape;

public

  OnShapeAdded              :procedure (Sender :TObject; Index :Integer) of object;
  OnShapeRemoved            :procedure (Sender :TObject; OldIndex :Integer; Shape :TQUB_QFS_QPF_Shape) of object;

  constructor Create(ShapeLib :TQUB_QFS_QPF_ShapeLib; Tree :INode);
  destructor Destroy; override;

  procedure InsertShape(Shape :INode; Index :Integer); overload;
  procedure InsertShape(ShapeFactory :TQUB_QFS_QPF_ShapeFactory; Properties :INode; Index :Integer); overload;
  procedure RemoveShape(Index :Integer);

  function Duration(Rep_Index :Integer): Double;
  procedure GetBounds(Sampling :Double; Reps :Integer; var Low, High :Double);

  property Tree                :INode                  read prTree;
  property ShapeCount          :Integer                read prGetShapeCount;
  property Shapes[Ix :Integer] :TQUB_QFS_QPF_Shape     read prGetShape;         default;

end;


TQUB_QFS_QPF_Stage          = class
protected

  prTree                    :INode;
  prChannels                :TList;

  prShapeLib                :TQUB_QFS_QPF_ShapeLib;

  prRepeatMode              :TQUB_QFS_QPF_RepeatMode;
  prRepeatTimes             :Integer;
  prPauseMode               :TQUB_QFS_QPF_PauseMode;
  prPauseSeconds            :Double;
  prRandomize : Boolean;


  function prGetChannelCount: Integer;
  procedure prSetChannelCount(x :Integer);
  function prGetChannel(Ix :Integer): TQUB_QFS_QPF_Channel;
  procedure prSetRepeatMode(x :TQUB_QFS_QPF_RepeatMode);
  procedure prSetRepeatTimes(x :Integer);
  procedure prSetPauseMode(x :TQUB_QFS_QPF_PauseMode);
  procedure prSetPauseSeconds(x :Double);
  procedure prSetRandomize( x : Boolean );

public

  OnRepeatChanged           :TNotifyEvent;
  OnPauseChanged            :TNotifyEvent;
  OnChannelDeleting         :TNotifyEvent; // sender is channel

  constructor Create(ShapeLib :TQUB_QFS_QPF_ShapeLib; ChanCount :Integer; Tree :INode);
  destructor Destroy; override;

  function Duration(Rep_Index :Integer): Double;
  procedure GetBoundsOfChannel(c :Integer; Sampling :Double; var Low, High :Double);

  property Tree                  :INode                   read prTree;
  property ChannelCount          :Integer                 read prGetChannelCount    write prSetChannelCount;
  property Channels[Ix :Integer] :TQUB_QFS_QPF_Channel    read prGetChannel;    default;
  property RepeatMode            :TQUB_QFS_QPF_RepeatMode read prRepeatMode         write prSetRepeatMode;
  property RepeatTimes           :Integer                 read prRepeatTimes        write prSetRepeatTimes;
  property PauseMode             :TQUB_QFS_QPF_PauseMode  read prPauseMode          write prSetPauseMode;
  property PauseSeconds          :Double                  read prPauseSeconds       write prSetPauseSeconds;
  property Randomize             :Boolean                 read prRandomize          write prSetRandomize;

end;


TQUB_QFS_QPF                = class // not QFS_Edit or anything -- uses QUB_Node
protected

  prShapeLib                :TQUB_QFS_QPF_ShapeLib;

  prFileName                :String;
  prTree                    :INode;
  prStages                  :TList;

  prChannelCount            :Integer;

//  prRepeatMode              :TQUB_QFS_QPF_RepeatMode;
//  prRepeatTimes             :Integer;

  function prGetModified: Boolean;
  procedure prSetTree(x :INode);
  function prGetStageCount: Integer;
  function prGetStage(Ix :Integer): TQUB_QFS_QPF_Stage;
  procedure prSetChannelCount(x :Integer);
//  procedure prSetRepeatMode(x :TQUB_QFS_QPF_RepeatMode);
//  procedure prSetRepeatTimes(x :Integer);

public

  OnChannelCountChanged     :TNotifyEvent;
  OnRepeatChanged           :TNotifyEvent;

  OnStageAdded              :procedure (Sender :TObject; Index :Integer) of object;
  OnStageRemoved            :procedure (Sender :TObject; OldIndex :Integer; Stage :TQUB_QFS_QPF_Stage) of object;

  constructor Create(ShapeLib :TQUB_QFS_QPF_ShapeLib);
  destructor Destroy; override;

  // New: it's "new" and ready on construction
  function Open(FileName :String): Boolean;
  function SaveAs(FileName :String): Boolean;

  procedure InsertStage(Stage :INode; Index :Integer);
  procedure RemoveStage(Index :Integer);

  procedure GetBoundsOfChannel(c :Integer; Sampling :Double; var Low, High :Double);

  property FileName              :String                  read prFileName           write prFileName;
  property Modified              :Boolean                 read prGetModified;
  property Tree                  :INode                   read prTree               write prSetTree;
  property StageCount            :Integer                 read prGetStageCount;
  property Stages[Ix :Integer]   :TQUB_QFS_QPF_Stage      read prGetStage;      default;
  property ChannelCount          :Integer                 read prChannelCount       write prSetChannelCount;
//  property RepeatMode            :TQUB_QFS_QPF_RepeatMode read prRepeatMode         write prSetRepeatMode;
//  property RepeatTimes           :Integer                 read prRepeatTimes        write prSetRepeatTimes;

end;


function ProtocolShape_Step_GenerateInto(Node, Params :INode): INode;
function ProtocolShape_Sine_GenerateInto(Node, Params :INode): INode;
function ProtocolShape_Ramp_GenerateInto(Node, Params :INode): INode;
function ProtocolShape_Hold_GenerateInto(Node, Params :INode): INode;



implementation


uses


{$IFDEF DELPHI6_OR_HIGHER}
Variants,
{$ENDIF}
QUB_Python_Thread, QUB_Viewport_Python, QUB_Node_Python;


// *****************************************************************************

constructor TQUB_QFS_QPF_ShapeFactory.Create(AFactory :PPyObject);
begin
  PythonThread.BeginPython;
  prFactory               := AFactory;
  if prFactory <> nil then begin
    GetPythonEngine.Py_INCREF( prFactory );
    prName                  := PyObject_GetAttrStringAsString(AFactory, 'name');
  end else begin
    prName                  := '<exception>'
  end;
  PythonThread.EndPython;
end;

destructor TQUB_QFS_QPF_ShapeFactory.Destroy;
begin
  try
    PythonThread.BeginPython;
    GetPythonEngine.Py_XDECREF( prFactory );
    PythonThread.EndPython;

    if prIcon <> nil then prIcon.Free;
  finally
    inherited;
  end;
end;

function TQUB_QFS_QPF_ShapeFactory.Instantiate(Properties :INode): TQUB_QFS_QPF_Shape;
var
  Args         :PPyObject;
  Shape        :PPyObject;
begin
  Result       := nil;
  if prFactory = nil then Exit;

  Args         := nil;
  Shape        := nil;
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    try
      if Properties.IsNull then begin
        Args       := PyTuple_New(0);
      end else begin
        Args       := Py_BuildValue('(N)', [PyNode_FromNode(Properties)]);
      end;
      Shape := PyObject_CallObject(Factory, Args);
      if Shape = nil then Exit;

      PythonThread.EndPython;
      Result := TQUB_QFS_QPF_Shape.Create( Shape, Self );
      PythonThread.BeginPython;
    finally
      if PyErr_Occurred <> nil then PyErr_Print;
      Py_XDECREF( Args );
      Py_XDECREF( Shape );
      PythonThread.EndPython;
    end;
  end;
end;


// *****************************************************************************

constructor TQUB_QFS_QPF_Shape.Create(AShape :PPyObject; AFactory :TQUB_QFS_QPF_ShapeFactory);
begin
  prShape            := AShape;
  prFactory          := AFactory;
  if prShape <> nil then begin
    PythonThread.BeginPython;
    GetPythonEngine.Py_INCREF( prShape );
    prTree             := PyObject_GetAttrStringAsNode(prShape, 'properties');
    PythonThread.EndPython;
  end;
end;

destructor TQUB_QFS_QPF_Shape.Destroy;
begin
  try
    PythonThread.BeginPython;
    GetPythonEngine.Py_XDECREF( prShape );
    PythonThread.EndPython;
  finally
    inherited;
  end;
end;

function TQUB_QFS_QPF_Shape.SampleCount(TStart, Sampling :Double; Rep_Index :Integer): Integer;
var
  PyArgs, PyMeth, PyResult, PyJunk   :PPyObject;
begin
  Result := 0;
  if Shape = nil then Exit;

  PyArgs             := nil;
  PyMeth             := nil;
  PyResult           := nil;
  PyJunk             := nil;
  with GetPythonEngine do begin
    PythonThread.BeginPython;
    try
    // Result := Shape.count(TStart, Sampling);
      PyArgs           := Py_BuildValue('(ddi)', [TStart, Sampling, Rep_Index]);
      PyMeth           := PyObject_GetAttrString(prShape, PChar('count'));
      if PyMeth = nil then Exit;
      PyResult         := PyObject_CallObject(PyMeth, PyArgs);
      if PyResult = nil then Exit;
      PyJunk           := PyNumber_Int(PyResult);
      if PyJunk = nil then Exit;
      Result           := PyLong_AsLong(PyJunk);
    finally
      if PyErr_Occurred <> nil then PyErr_Print;
      Py_XDECREF( PyResult );
      Py_XDECREF( PyMeth );
      Py_XDECREF( PyArgs );
      Py_XDECREF( PyJunk );
      PythonThread.EndPython;
    end;
  end;
end;

function TQUB_QFS_QPF_Shape.SampleCount(TStart, Sampling :Double; Duration :Double): Integer;
var
  PyArgs, PyMeth, PyResult, PyJunk   :PPyObject;
begin
  Result := 0;
  if Shape = nil then Exit;

  PyArgs             := nil;
  PyMeth             := nil;
  PyResult           := nil;
  PyJunk             := nil;
  with GetPythonEngine do begin
    PythonThread.BeginPython;
    try
    // Result := Shape.count(TStart, Sampling);
      PyArgs           := Py_BuildValue('(ddd)', [TStart, Sampling, Duration]);
      PyMeth           := EvalString('AcquisitionProtocol.SampleCount');
      if PyMeth = nil then Exit;
      PyResult         := PyObject_CallObject(PyMeth, PyArgs);
      if PyResult = nil then Exit;
      PyJunk           := PyNumber_Int(PyResult);
      if PyJunk = nil then Exit;
      Result           := PyLong_AsLong(PyJunk);
      if Result < 0 then Result := 0;
    finally
      if PyErr_Occurred <> nil then PyErr_Print;
      Py_XDECREF( PyResult );
      Py_XDECREF( PyMeth );
      Py_XDECREF( PyArgs );
      Py_XDECREF( PyJunk );
      PythonThread.EndPython;
    end;
  end;
end;

function TQUB_QFS_QPF_Shape.Duration(Rep_Index :Integer): Double;
var
  PyArgs, PyMeth, PyResult, PyJunk   :PPyObject;
begin
  Result := 1.0;  // easy to click?
  if Shape = nil then Exit;

  PyArgs             := nil;
  PyMeth             := nil;
  PyResult           := nil;
  PyJunk             := nil;
  with GetPythonEngine do begin
    PythonThread.BeginPython;
    try
    // Result := Shape.duration();
      PyArgs           := Py_BuildValue('(i)', [Rep_Index]);
      PyMeth           := PyObject_GetAttrString(prShape, PChar('duration'));
      if PyMeth = nil then Exit;
      PyResult         := PyObject_CallObject(PyMeth, PyArgs);
      if PyResult = nil then Exit;
      PyJunk           := PyNumber_Float(PyResult);
      if PyJunk = nil then Exit;
      Result           := PyFloat_AsDouble(PyJunk);
      if Result < 0.0 then Result := 0.0;
    finally
      if PyErr_Occurred <> nil then PyErr_Print;
      Py_XDECREF( PyResult );
      Py_XDECREF( PyMeth );
      Py_XDECREF( PyArgs );
      Py_XDECREF( PyJunk );
      PythonThread.EndPython;
    end;
  end;
end;

procedure TQUB_QFS_QPF_Shape.GetBounds(Sampling :Double; Reps :Integer; var Low, High :Double);
var
  Samples      :Integer;
  Rep          :Integer;
  Gen          :INode;
  Data         :TPADouble;
  i            :Integer;
begin
  for Rep := 0 to Reps - 1 do begin
    Samples      := SampleCount(0, Sampling, Duration(Rep));
    Gen          := Generate(0, Samples-1, 0, Sampling, Rep);
    Data         := TPADouble(Gen.DataPtr);
    Samples      := Gen.DataCount;
    for i := 0 to Samples - 1 do begin
      if Data[i] < Low then Low := Data[i];
      if Data[i] > High then High := Data[i];
    end;
  end;
end;

function TQUB_QFS_QPF_Shape.Generate(First, Last :Integer; TStart, Sampling :Double; Rep_Index :Integer): INode;
var
  PyArgs, PyMeth, PyResult   :PPyObject;
  DataPtr                    :TPADouble;
  i                          :Integer;
begin
  Result    := QUB_Node_NULL;
  if Shape = nil then Exit;

  PyArgs    := nil;
  PyMeth    := nil;
  PyResult  := nil;
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    try
    // Result := Shape.generate(First, Last, TStart, Sampling, Rep_Index);
      PyArgs           := Py_BuildValue('(iiddi)', [First, Last, TStart, Sampling, Rep_Index]);
      PyMeth           := PyObject_GetAttrString(prShape, PChar('generate'));
      if PyMeth = nil then Exit;
      PyResult         := PyObject_CallObject(PyMeth, PyArgs);
      if (PyResult = nil) or not PyNode_IsNode(PyResult) then Exit;
      Result           := PyNode_ToNode(PyResult);
    finally
      if PyErr_Occurred <> nil then begin
        Result := TNode.Create('');
        Result['Err'].SetData( PyErr_ShortMessage );
        Result.SetupNumData(NODE_DATATYPE_DOUBLE, Last-First+1, 1);
        DataPtr := TPADouble(Result.DataPtr);
        for i := 0 to Last - First do begin
          DataPtr[i] := 0.0;
        end;
      end;
      Py_XDECREF( PyResult );
      Py_XDECREF( PyMeth );
      Py_XDECREF( PyArgs );
      PythonThread.EndPython;
    end;
  end;
end;

procedure TQUB_QFS_QPF_Shape.GenerateInto(First, Last :Integer; TStart, Sampling :Double; Rep_Index :Integer; Node :INode; IStart :Integer);
var
  PyArgs, PyMeth, PyResult   :PPyObject;
  DataPtr                    :TPADouble;
  i                          :Integer;
begin
  if Shape = nil then Exit;

  PyArgs    := nil;
  PyMeth    := nil;
  PyResult  := nil;
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    try
    // Shape.generateInto(First, Last, TStart, Sampling, Rep_Index, Node, IStart);
      PyArgs           := Py_BuildValue('(iiddiNi)', [First, Last, TStart, Sampling, Rep_Index,
                                                     PyNode_FromNode(Node), IStart]);
      PyMeth           := PyObject_GetAttrString(prShape, PChar('generateInto'));
      if PyMeth = nil then Exit;
      PyResult         := PyObject_CallObject(PyMeth, PyArgs);
    finally
      if PyErr_Occurred <> nil then begin
        Node['Err'].SetData( PyErr_ShortMessage );
        DataPtr        := TPADouble(Node.DataPtr);
        for i := IStart to IStart + Last - First do begin
          DataPtr[i]   := 0.0;
        end;
      end;
      Py_XDECREF( PyResult );
      Py_XDECREF( PyMeth );
      Py_XDECREF( PyArgs );
      PythonThread.EndPython;
    end;
  end;
end;

function TQUB_QFS_QPF_Shape.prGetProperty(Nm :String): String;
begin
  Result := Tree.Find(Nm).DataAsString;
end;

procedure TQUB_QFS_QPF_Shape.prSetProperty(Nm, Val :String);
var
  PyArgs, PyMeth, PyResult   :PPyObject;
begin
  if Shape = nil then Exit;

  PyArgs    := nil;
  PyMeth    := nil;
  PyResult  := nil;
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    try
    // Shape.set(Nm, Val);
      PyArgs           := Py_BuildValue('(ss)', [PChar(Nm), PChar(Val)]);
      PyMeth           := PyObject_GetAttrString(prShape, PChar('set'));
      if PyMeth = nil then Exit;
      PyResult         := PyObject_CallObject(PyMeth, PyArgs);
    finally
      if PyErr_Occurred <> nil then PyErr_Print;
      Py_XDECREF( PyResult );
      Py_XDECREF( PyMeth );
      Py_XDECREF( PyArgs );
      PythonThread.EndPython;
    end;
  end;

  if Assigned(OnChanged)  then OnChanged (Self);
end;


// *****************************************************************************

constructor TQUB_QFS_QPF_ShapeLib.Create;
var
  FlatLine       :PPyObject;
begin
  prShapes       := TStringList.Create;

  // by enqueueing this method/synchronized, we force VCL thread to wait for all python initialization
  PythonThread.SyncPyProc(prInitPyFunc, False, True, -1);

  with GetPythonEngine do begin
    PythonThread.BeginPython;
    try
      FlatLine       := EvalString('AcquisitionProtocol.Shape()');
    except
      FlatLine       := nil;
    end;
    PythonThread.EndPython;
    Add(FlatLine);
    PythonThread.BeginPython;
    Py_XDECREF(FlatLine);
    PythonThread.EndPython;
  end;

  prAddQueuedShapes;
end;

destructor TQUB_QFS_QPF_ShapeLib.Destroy;
var
  i           :Integer;
begin
  try
    for i := 0 to ShapeCount - 1 do begin
      Shapes[i].Free;
    end;
    prShapes.Free;
  except
  end;

  inherited;
end;

procedure TQUB_QFS_QPF_ShapeLib.Add(ShapeFactory :PPyObject);
var
  Factory      :TQUB_QFS_QPF_ShapeFactory;
  IconName     :String;
  i            :Integer;
begin
  Factory      := TQUB_QFS_QPF_ShapeFactory.Create(ShapeFactory);

  i            := prShapes.IndexOf(Factory.Name);
  if i >= 0 then begin
    prShapes.Objects[i].Free;
    prShapes.Objects[i] := Factory;
  end else begin
    prShapes.AddObject(Factory.Name, Factory);
  end;

  IconName := ExtractFileDir(Application.ExeName) + '\ProtocolShapes\' + Factory.Name + '.bmp';
  if FileExists(IconName) then try
    Factory.Icon := Graphics.TBitmap.Create;
    Factory.Icon.LoadFromFile(IconName);
  except
    Factory.Icon.Free;
    Factory.Icon := nil;
  end;

  if (i < 0) and Assigned(OnShapeAdded) then OnShapeAdded(Self);
end;

procedure TQUB_QFS_QPF_ShapeLib.prInitPyFunc;
var
  Fn                    :INodeNodeFunc;
  AddShapePyFunc        :PPyObject;
  GenerateIntoPyFunc    :PPyObject;
begin
  Fn := TNodeNodeFunc.Create( ProtocolShape_Step_GenerateInto );
  PyNode_PublishNodeNodeFunc( Fn, 'ProtocolShape_Step_GenerateInto' );
  Fn := TNodeNodeFunc.Create( ProtocolShape_Sine_GenerateInto );
  PyNode_PublishNodeNodeFunc( Fn, 'ProtocolShape_Sine_GenerateInto' );
  Fn := TNodeNodeFunc.Create( ProtocolShape_Ramp_GenerateInto );
  PyNode_PublishNodeNodeFunc( Fn, 'ProtocolShape_Ramp_GenerateInto' );
  Fn := TNodeNodeFunc.Create( ProtocolShape_Hold_GenerateInto );
  PyNode_PublishNodeNodeFunc( Fn, 'ProtocolShape_Hold_GenerateInto' );

  prAddShapePyType := TPythonType.Create(nil);
  prAddShapePyType.Engine := GetPythonEngine;
  prAddShapePyType.Module := PythonViewport.pmQUBTypes;
  prAddShapePyType.TypeName := 'AddProtocolShapeFunc';
  prAddShapePyType.Services.Basic := [bsCall];
  prAddShapePyType.PyObjectClass := TQUB_QFS_QPF_AddShape;
  prAddShapePyType.Initialize;

  AddShapePyFunc := prAddShapePyType.CreateInstance;
  PyNode_PublishPyObj( AddShapePyFunc, 'AddProtocolShape' );
  GetPythonEngine.Py_DECREF( AddShapePyFunc );

  prGenerateIntoPyType := TPythonType.Create(nil);
  prGenerateIntoPyType.Engine := GetPythonEngine;
  prGenerateIntoPyType.Module := PythonViewport.pmQUBTypes;
  prGenerateIntoPyType.TypeName := 'ProtocolShape_GenerateIntoFunc';
  prGenerateIntoPyType.Services.Basic := [bsCall];
  prGenerateIntoPyType.PyObjectClass := TQUB_ProtocolShape_GenerateInto;
  prGenerateIntoPyType.Initialize;

  GenerateIntoPyFunc := prGenerateIntoPyType.CreateInstance;
  PyNode_PublishPyObj( GenerateIntoPyFunc, 'ProtocolShape_GenerateInto' );
  GetPythonEngine.Py_DECREF( GenerateIntoPyFunc );
end;

procedure TQUB_QFS_QPF_ShapeLib.prAddQueuedShapes;
var
  ShapeList         :PPyObject;
  Factory           :PPyObject;
  i                 :Integer;
begin
  ShapeList         := nil;

  PythonThread.BeginPython;
  with GetPythonEngine do begin
    try
      ShapeList     := EvalString('AcquisitionProtocol.ShapeQueue');
      for i := 0 to PyList_Size(ShapeList) - 1 do begin
        Factory     := PyList_GetItem(ShapeList, i);    // borrowed ref
        PythonThread.EndPython;
        Add(Factory);
        PythonThread.BeginPython;
      end;
      PyList_SetSlice(ShapeList, 0, PyList_Size(ShapeList), nil); // clear
    finally
      if PyErr_Occurred <> nil then PyErr_Print;
      Py_XDECREF( ShapeList );
      PythonThread.EndPython;
    end;
  end;
end;

function TQUB_QFS_QPF_ShapeLib.prGetShapeCount;
begin
  Result := prShapes.Count;
end;

function TQUB_QFS_QPF_ShapeLib.prGetShape(NameOrIx :Variant): TQUB_QFS_QPF_ShapeFactory;
var
  Name     :String;
  i        :Integer;
begin
  Result := TQUB_QFS_QPF_ShapeFactory(prShapes.Objects[0]);
  if (VarType(NameOrIx) and varTypeMask) = varString then begin
    Name   := String(NameOrIx);
    for i := 1 to ShapeCount - 1 do begin
      if TQUB_QFS_QPF_ShapeFactory(prShapes.Objects[i]).Name = Name then begin
        Result := TQUB_QFS_QPF_ShapeFactory(prShapes.Objects[i]);
        Exit;
      end;
    end;
  end else begin
    try
      i      := Integer(NameOrIx);
      if (i >= 0) and (i < ShapeCount) then begin
        Result := TQUB_QFS_QPF_ShapeFactory(prShapes.Objects[i]);
      end;
    except
    end;
  end;
end;

// *****************************************************************************

function  TQUB_QFS_QPF_AddShape.Call( args, kw : PPyObject) : PPyObject;
begin
  with GetPythonEngine do begin
    if 0 = PyArg_ParseTuple( args, 'O:AddProtocolShape', [@prFactory] ) then begin
      Result := nil;
    end else begin
      PythonThread.SyncGUIProc( prAddShape );
      Result := ReturnNone;
    end;
  end;
end;

procedure TQUB_QFS_QPF_AddShape.prAddShape;
begin
  ShapeLib.Add( prFactory );
end;

// *****************************************************************************


function TQUB_ProtocolShape_GenerateInto.Call(Args, kw :PPyObject): PPyObject;
var
  Eval        :PPyObject;
  NodePy      :PPyObject;
  ParamsPy    :PPyObject;
  Params      :INode;
begin
  with GetPythonEngine do begin
    if ( PyArg_ParseTuple( args, 'OOO:ProtocolShape_GenerateInto(evaluate, node, params)', [@Eval, @NodePy, @ParamsPy] ) = 0 ) then begin
      Result := nil;
    end else begin
      // should really check arg types, but this will only be called from one place (in AcquisitionProtocol.py)
      Params := PyNode_ToNode(ParamsPy);
      Result := prGenerate( Eval,
                            PyNode_ToNode(NodePy),
                            Params['t_off'].Data[0],
                            Params['sampling'].Data[0],
                            Params['first'].Data[0],
                            Params['remain'].Data[0],
                            Params['rep_index'].Data[0],
                            Params['i_start'].Data[0] );
    end;
  end;
end;

function TQUB_ProtocolShape_GenerateInto.prGenerate(Eval :PPyObject; Node :INode; t_off, sampling :Double; first, remain, rep_index, i_start :Integer): PPyObject;
var
  i           :Integer;
  t           :Double;
  EvalArgs    :PPyObject;
  EvalResult  :PPyObject;
  Data        :TPADouble;
begin
  with GetPythonEngine do begin
    EvalArgs    := nil;
    try
      EvalArgs  := PyTuple_New(2);
      PyTuple_SetItem(EvalArgs, 1, PyInt_FromLong(rep_index));

      Data      := TPADouble(Node.DataPtr);

      for i := 0 to remain - 1 do begin
        t       := t_off + (first + i) * sampling;
        PyTuple_SetItem(EvalArgs, 0, PyFloat_FromDouble(t));      // assuming SetItem decrefs the existing item
        EvalResult := PyObject_CallObject(Eval, EvalArgs);
        if EvalResult = nil then begin
          Result := nil;
          Exit;
        end;
        Data[i_start + i] := PyFloat_AsDouble(EvalResult);
        Py_DECREF(EvalResult);
      end;
    finally
      Py_XDECREF(EvalArgs);
    end;

    Result := ReturnNone;
  end;
end;


//******************************************************************************

function ProtocolShape_Step_GenerateInto(Node, Params :INode): INode;
var
  t_off       :Double;
  sampling    :Double;
  first       :Integer;
  remain      :Integer;
  i_start     :Integer;
  preDur      :Double;
  preAmp      :Double;
  pre2Dur     :Double;
  amp         :Double;
  postAmp     :Double;
  i           :Integer;
  t           :Double;
  Data        :TPADouble;
begin
  Result      := Node;
  try
    t_off     := Params['t_off'].Data[0];
    sampling  := Params['sampling'].Data[0];
    first     := Params['first'].Data[0];
    remain    := Params['remain'].Data[0];
    i_start   := Params['i_start'].Data[0];
    preDur    := Params['preDur'].Data[0];
    preAmp    := Params['preAmp'].Data[0];
    pre2Dur   := Params['pre2Dur'].Data[0];
    amp       := Params['amp'].Data[0];
    postAmp   := Params['postAmp'].Data[0];

    Data      := TPADouble(Node.DataPtr);
    for i := 0 to remain - 1 do begin
      t       := t_off + (first + i) * sampling;
      if t < preDur then begin
        Data[i_start + i] := preAmp;
      end else if t < pre2Dur then begin
        Data[i_start + i] := amp;
      end else begin
        Data[i_start + i] := postAmp;
      end;
    end;
  except
    // ??
  end;
end;


//******************************************************************************

function ProtocolShape_Sine_GenerateInto(Node, Params :INode): INode;
var
  t_off       :Double;
  sampling    :Double;
  first       :Integer;
  remain      :Integer;
  i_start     :Integer;
  off         :Double;
  amp         :Double;
  freq2pi     :Double;
  phase       :Double;
  i           :Integer;
  t           :Double;
  Data        :TPADouble;
begin
  Result      := Node;
  try
    t_off     := Params['t_off'].Data[0];
    sampling  := Params['sampling'].Data[0];
    first     := Params['first'].Data[0];
    remain    := Params['remain'].Data[0];
    i_start   := Params['i_start'].Data[0];
    off       := Params['off'].Data[0];
    amp       := Params['amp'].Data[0];
    freq2pi   := 2*PI*Params['freq'].Data[0];
    phase     := Params['phase'].Data[0];

    Data      := TPADouble(Node.DataPtr);
    for i := 0 to remain - 1 do begin
      t       := t_off + (first + i) * sampling;
      Data[i_start + i] := off + amp * sin(t * freq2pi + phase);
    end;
  except
    // ??
  end;
end;


//******************************************************************************

function ProtocolShape_Ramp_GenerateInto(Node, Params :INode): INode;
var
  t_off       :Double;
  sampling    :Double;
  first       :Integer;
  remain      :Integer;
  i_start     :Integer;
  fromAmp     :Double;
  toAmp       :Double;
  dur         :Double;
  slope       :Double;
  i           :Integer;
  t           :Double;
  Data        :TPADouble;
begin
  Result      := Node;
  try
    t_off     := Params['t_off'].Data[0];
    sampling  := Params['sampling'].Data[0];
    first     := Params['first'].Data[0];
    remain    := Params['remain'].Data[0];
    i_start   := Params['i_start'].Data[0];
    fromAmp   := Params['fromAmp'].Data[0];
    toAmp     := Params['toAmp'].Data[0];
    dur       := Params['dur'].Data[0];
    slope     := (toAmp - fromAmp) / dur;

    Data      := TPADouble(Node.DataPtr);
    for i := 0 to remain - 1 do begin
      t       := t_off + (first + i) * sampling;
      Data[i_start + i] := fromAmp + t * slope;
    end;
  except
    // ??
  end;
end;


//******************************************************************************

function ProtocolShape_Hold_GenerateInto(Node, Params :INode): INode;
var
  remain      :Integer;
  i_start     :Integer;
  amp         :Double;
  i           :Integer;
  Data        :TPADouble;
begin
  Result      := Node;
  try
    // first     := Params['first'].Data[0];
    remain    := Params['remain'].Data[0];
    i_start   := Params['i_start'].Data[0];
    amp       := Params['amp'].Data[0];

    Data      := TPADouble(Node.DataPtr);
    for i := 0 to remain - 1 do begin
      Data[i_start + i] := amp;
    end;
  except
    // ??
  end;
end;


// *****************************************************************************

constructor TQUB_QFS_QPF_Channel.Create(ShapeLib :TQUB_QFS_QPF_ShapeLib; Tree :INode);
var
  ShapeTree         :INode;
  ShapeTemplate     :TQUB_QFS_QPF_ShapeFactory;
  Shape             :TQUB_QFS_QPF_Shape;
begin
  prShapeLib        := ShapeLib;
  prTree            := Tree;
  prShapes          := TList.Create;

  ShapeTree         := Tree.Find('Shape');
  while not ShapeTree.IsNull do begin
    ShapeTemplate   := ShapeLib[ShapeTree.DataAsString];
    Shape           := ShapeTemplate.Instantiate(ShapeTree);
    prShapes.Add( Shape );
    ShapeTree        := ShapeTree.SiblingSameName;
  end;
end;

destructor TQUB_QFS_QPF_Channel.Destroy;
var
  i          :Integer;
begin
  try
    for i := 0 to ShapeCount - 1 do begin
      Shapes[i].Free;
    end;
    prShapes.Free;
  except
  end;

  inherited;
end;

procedure TQUB_QFS_QPF_Channel.InsertShape(Shape :INode; Index :Integer);
begin
  InsertShape( prShapeLib[Shape.DataAsString], Shape, Index );
end;

procedure TQUB_QFS_QPF_Channel.InsertShape(ShapeFactory :TQUB_QFS_QPF_ShapeFactory; Properties :INode; Index :Integer);
var
  PrevShape       :INode;
  ShapeObj        :TQUB_QFS_QPF_Shape;
begin
  if Index < 0 then Index := 0;
  if Index > ShapeCount then Index := ShapeCount;

  if Index = 0 then begin
    PrevShape := QUB_Node_Null;
  end else begin
    PrevShape := Shapes[Index-1].Tree;
  end;

  ShapeObj    := ShapeFactory.Instantiate(Properties);
  prShapes.Insert( Index, ShapeObj );
  Tree.InsertChild(PrevShape, ShapeObj.Tree);

  if Assigned(OnShapeAdded) then OnShapeAdded(Self, Index);
end;

procedure TQUB_QFS_QPF_Channel.RemoveShape(Index :Integer);
var
  Removed           :TQUB_QFS_QPF_Shape;
begin
  if (Index < 0) or (Index >= ShapeCount) then Exit;
  Removed := Shapes[Index];
  prShapes.Delete(Index);
  Tree.RemoveChild(Removed.Tree);
  if Assigned(OnShapeRemoved) then OnShapeRemoved(Self, Index, Removed);
  Removed.Free;
end;

function TQUB_QFS_QPF_Channel.Duration(Rep_Index :Integer): Double;
var
  i            :Integer;
begin
  Result       := 0.0;
  for i := 0 to ShapeCount - 1 do begin
    Result     := Result + Shapes[i].Duration(Rep_Index);
  end;
end;

procedure TQUB_QFS_QPF_Channel.GetBounds(Sampling :Double; Reps :Integer; var Low, High :Double);
var
  i        :Integer;
begin
  for i := 0 to ShapeCount - 1 do begin
    Shapes[i].GetBounds(Sampling, Reps, Low, High);
  end;
end;

function TQUB_QFS_QPF_Channel.prGetShapeCount: Integer;
begin
  Result := prShapes.Count;
end;

function TQUB_QFS_QPF_Channel.prGetShape(Ix :Integer): TQUB_QFS_QPF_Shape;
begin
  if (Ix >= 0) and (Ix < prShapes.Count) then begin
    Result := TQUB_QFS_QPF_Shape(prShapes[Ix]);
  end else begin
    Result := nil;
  end;
end;


// *****************************************************************************

constructor TQUB_QFS_QPF_Stage.Create(ShapeLib :TQUB_QFS_QPF_ShapeLib; ChanCount :Integer; Tree :INode);
var
  i                       :Integer;
  ChanTree                :INode;
  Chan                    :TQUB_QFS_QPF_Channel;
  Prop                    :INode;
begin
  prShapeLib              := ShapeLib;
  prTree                  := Tree;
  prChannels              := TList.Create;

  ChanTree                := Tree.Find('Channel');
  for i := 0 to ChanCount - 1 do begin
    if ChanTree.IsNull then begin
      ChanTree            := Tree.AppendChild('Channel');
    end;
    Chan                  := TQUB_QFS_QPF_Channel.Create(ShapeLib, ChanTree);
    prChannels.Add(Chan);

    ChanTree              := ChanTree.SiblingSameName;
  end;

  Prop                    := Tree['RepeatMode'];
  if Prop.DataCount > 0 then begin
    prRepeatMode          := TQUB_QFS_QPF_RepeatMode( Integer(Prop.Data[0]) );
  end else begin
    prRepeatMode          := QUB_QFS_QPF_REPEAT_NONE;
    Prop.SetData( Integer(prRepeatMode) );
  end;

  Prop                    := Tree['RepeatTimes'];
  if Prop.DataCount > 0 then begin
    prRepeatTimes          := Prop.Data[0];
  end else begin
    prRepeatTimes          := 1;
    Prop.SetData( prRepeatTimes );
  end;

  Prop                    := Tree['PauseMode'];
  if Prop.DataCount > 0 then begin
    prPauseMode           := TQUB_QFS_QPF_PauseMode( Integer(Prop.Data[0]) );
  end else begin
    prPauseMode           := QUB_QFS_QPF_PAUSE_NONE;
    Prop.SetData( Integer(prPauseMode) );
  end;

  Prop                    := Tree['PauseSeconds'];
  if Prop.DataCount > 0 then begin
    prPauseSeconds        := Prop.Data[0];
  end else begin
    prPauseSeconds        := 1.0;
    Prop.SetData( prPauseSeconds );
  end;

  Prop := Tree['Randomize'];
  if Prop.DataCount > 0 then begin
    if Prop.Data[0]=0 Then Randomize := False else Randomize := True;
  end else begin
    Randomize := False;
    Prop.SetData( 0 );
  end;

end;

destructor TQUB_QFS_QPF_Stage.Destroy;
var
  i                :Integer;
begin
  try
    for i := 0 to ChannelCount - 1 do begin
      Channels[i].Free;
    end;
    prChannels.Free;
  except
  end;

  inherited;
end;

function TQUB_QFS_QPF_Stage.Duration(Rep_Index :Integer): Double;
var
  i          :Integer;
begin
  Result     := 0.0;
  for i := 0 to ChannelCount - 1 do begin
    Result   := Max(Result, Channels[i].Duration(Rep_Index));
  end;
end;

procedure TQUB_QFS_QPF_Stage.GetBoundsOfChannel(c :Integer; Sampling :Double; var Low, High :Double);
begin
  Channels[c].GetBounds(Sampling, RepeatTimes, Low, High);
end;

function TQUB_QFS_QPF_Stage.prGetChannelCount: Integer;
begin
  Result := prChannels.Count;
end;

procedure TQUB_QFS_QPF_Stage.prSetChannelCount(x :Integer);
var
  i            :Integer;
  ChanTree     :INode;
  Chan         :TQUB_QFS_QPF_Channel;
begin
  for i := ChannelCount - 1 downto x do begin
    if Assigned(OnChannelDeleting) then OnChannelDeleting(Channels[i]);
    Tree.RemoveChild(Channels[i].Tree);
    Channels[i].Free;
    prChannels.Delete(i);
  end;

  for i := ChannelCount to x - 1 do begin
    ChanTree            := Tree.AppendChild('Channel');

    Chan := TQUB_QFS_QPF_Channel.Create(prShapeLib, ChanTree);
    prChannels.Add( Chan );
  end;
end;

function TQUB_QFS_QPF_Stage.prGetChannel(Ix :Integer): TQUB_QFS_QPF_Channel;
begin
  if (Ix >= 0) and (Ix < ChannelCount) then begin
    Result := TQUB_QFS_QPF_Channel(prChannels[Ix]);
  end else begin
    Result := nil;
  end;
end;

procedure TQUB_QFS_QPF_Stage.prSetRepeatMode(x :TQUB_QFS_QPF_RepeatMode);
begin
  if x = prRepeatMode then Exit;
  prRepeatMode := x;
  Tree['RepeatMode'].Data[0] := Integer(x);
  if Assigned(OnRepeatChanged) then OnRepeatChanged(Self);
end;

procedure TQUB_QFS_QPF_Stage.prSetRepeatTimes(x :Integer);
begin
  if x = prRepeatTimes then Exit;
  prRepeatTimes := x;
  Tree['RepeatTimes'].Data[0] := x;
  if Assigned(OnRepeatChanged) then OnRepeatChanged(Self);
end;

procedure TQUB_QFS_QPF_Stage.prSetPauseMode(x :TQUB_QFS_QPF_PauseMode);
begin
  if x = prPauseMode then Exit;
  prPauseMode := x;
  Tree['PauseMode'].Data[0] := Integer(x);
  if Assigned(OnPauseChanged) then OnPauseChanged(Self);
end;

procedure TQUB_QFS_QPF_Stage.prSetPauseSeconds(x :Double);
begin
  if x = prPauseSeconds then Exit;
  prPauseSeconds := x;
  Tree['PauseSeconds'].Data[0] := x;
  if Assigned(OnPauseChanged) then OnPauseChanged(Self);
end;

procedure TQUB_QFS_QPF_Stage.prSetRandomize( x : Boolean );
begin
  if x = prRandomize then
    Exit;
  prRandomize := x;
  if x then Tree['Randomize'].Data[0] := 1 else Tree['Randomize'].Data[0]:=0;
  if Assigned(OnRepeatChanged) then OnRepeatChanged(Self);
end;

// *****************************************************************************

constructor TQUB_QFS_QPF.Create(ShapeLib :TQUB_QFS_QPF_ShapeLib);
var
  Stage           :TQUB_QFS_QPF_Stage;
begin
  prShapeLib      := ShapeLib;
  prFileName      := 'noname.qpf';
  prTree          := TNode.Create('ProtocolFile');
  prStages        := TList.Create;
  prChannelCount  := 1;
  //prRepeatMode    := QUB_QFS_QPF_REPEAT_NONE;
  //prRepeatTimes   := 1;
  prTree['ChannelCount'].SetData( prChannelCount );
  //prTree['RepeatMode'].SetData( Integer(prRepeatMode) );
  //prTree['RepeatTimes'].SetData( prRepeatTimes );
  Stage           := TQUB_QFS_QPF_Stage.Create(ShapeLib, prChannelCount, prTree['Stages']['Stage']);
  prStages.Add( Stage );
end;

destructor TQUB_QFS_QPF.Destroy;
var
  i               :Integer;
begin
  try
    for i := 0 to StageCount - 1 do begin
      //if Assigned(OnStageRemoved) then OnStageRemoved(Self, i, Stages[i]):
      Stages[i].Free;
    end;
    prStages.Free;
  except
  end;

  inherited;
end;

function TQUB_QFS_QPF.Open(FileName :String): Boolean;
var
  NewTree    :INode;
begin
  Result     := True;
  try
    NewTree  := TNode.Open(FileName);
    if not NewTree.IsNull then begin
      Tree        := NewTree;
      prFileName  := FileName;
      Result      := False;
    end;
  except
  end;
end;

function TQUB_QFS_QPF.SaveAs(FileName :String): Boolean;
begin
  Result := not Tree.SaveAs(FileName);
  if not Result then prFileName := FileName;
  Tree.Close;
end;

procedure TQUB_QFS_QPF.InsertStage(Stage :INode; Index :Integer);
var
  PrevStage        :INode;
  i                :Integer;
  Chan, Last       :INode;
begin
  if Index < 0 then Index := 0;
  if Index > StageCount then Index := StageCount;
  if Stage.IsNull then Stage := TNode.Create('Stage');
  if Index = 0 then begin
    PrevStage := QUB_Node_Null;
  end else begin
    PrevStage := Stages[Index-1].Tree;
  end;
  Chan             := Stage.Find('Channel');
  i                := 0;
  while i < ChannelCount do begin
    if Chan.IsNull then Chan := Stage.AppendChild('Channel');
    Chan           := Chan.SiblingSameName;
    Inc(i);
  end;
  while not Chan.IsNull do begin
    Last           := Chan;
    Chan           := Chan.SiblingSameName;
    Stage.RemoveChild(Last);
  end;
  Tree['Stages'].InsertChild(PrevStage, Stage);
  prStages.Insert( Index, TQUB_QFS_QPF_Stage.Create(prShapeLib, prChannelCount, Stage) );

  if Assigned(OnStageAdded) then OnStageAdded(Self, Index);
end;

procedure TQUB_QFS_QPF.RemoveStage(Index :Integer);
var
  Removed           :TQUB_QFS_QPF_Stage;
begin
  if (Index < 0) or (Index >= StageCount) then Exit;
  Removed := Stages[Index];
  prStages.Delete(Index);
  Tree['Stages'].RemoveChild(Removed.Tree);
  if Assigned(OnStageRemoved) then OnStageRemoved(Self, Index, Removed);
  Removed.Free;
end;

procedure TQUB_QFS_QPF.GetBoundsOfChannel(c :Integer; Sampling :Double; var Low, High :Double);
var
  s         :Integer;
begin
  Low       := 1.0e20;
  High      := -Low;
  for s := 0 to StageCount - 1 do begin
    Stages[s].GetBoundsOfChannel(c, Sampling, Low, High);
  end;
  if Low > High then begin
    Low     := 0;
    High    := 0;
  end;
end;

function TQUB_QFS_QPF.prGetModified: Boolean;
begin
  Result := Tree.Changed;
end;

procedure TQUB_QFS_QPF.prSetTree(x :INode);
var
  i         :Integer;
  Stage     :TQUB_QFS_QPF_Stage;
  StageTree :INode;
  Prop      :INode;
begin
  if x.IsNull then Exit;
  
  for i := StageCount - 1 downto 0 do begin
    Stage := Stages[i];
    prStages.Delete(i);
    if Assigned(OnStageRemoved) then OnStageRemoved(Self, i, Stage);
    Stage.Free;
  end;

  prTree    := x;

  Prop                    := Tree['ChannelCount'];
  if Prop.DataCount > 0 then begin
    ChannelCount          := Prop.Data[0];
  end else begin
    ChannelCount          := 1;
    Prop.SetData( prChannelCount );
  end;

  {
  Prop                    := Tree['RepeatMode'];
  if Prop.DataCount > 0 then begin
    RepeatMode            := TQUB_QFS_QPF_RepeatMode( Integer(Prop.Data[0]) );
  end else begin
    RepeatMode            := QUB_QFS_QPF_REPEAT_NONE;
    Prop.SetData( Integer(prRepeatMode) );
  end;

  Prop                    := Tree['RepeatTimes'];
  if Prop.DataCount > 0 then begin
    RepeatTimes            := Prop.Data[0];
  end else begin
    RepeatTimes            := 1;
    Prop.SetData( prRepeatTimes );
  end;
  }

  StageTree := prTree['Stages']['Stage']; // square brackets make sure there's at least one
  while not StageTree.IsNull do begin
    Stage   := TQUB_QFS_QPF_Stage.Create(prShapeLib, prChannelCount, StageTree);
    prStages.Add(Stage);
    if Assigned(OnStageAdded) then OnStageAdded(Self, prStages.Count-1);
    StageTree := StageTree.SiblingSameName;
  end;
end;

function TQUB_QFS_QPF.prGetStageCount: Integer;
begin
  Result := prStages.Count;
end;

function TQUB_QFS_QPF.prGetStage(Ix :Integer): TQUB_QFS_QPF_Stage;
begin
  if (Ix >= 0) and (Ix < StageCount) then begin
    Result := TQUB_QFS_QPF_Stage(prStages[Ix]);
  end else begin
    Result := nil;
  end;
end;

procedure TQUB_QFS_QPF.prSetChannelCount(x :Integer);
var
  i             :Integer;
begin
  if x = prChannelCount then Exit;
  for i := 0 to StageCount - 1 do begin
    Stages[i].ChannelCount := x;
  end;
  prChannelCount := x;
  Tree['ChannelCount'].Data[0] := x;
  if Assigned(OnChannelCountChanged) then OnChannelCountChanged(Self);
end;

{
procedure TQUB_QFS_QPF.prSetRepeatMode(x :TQUB_QFS_QPF_RepeatMode);
begin
  if x = prRepeatMode then Exit;
  prRepeatMode  := x;
  Tree['RepeatMode'].Data[0] := Integer(x);
  if Assigned(OnRepeatChanged) then OnRepeatChanged(Self);
end;

procedure TQUB_QFS_QPF.prSetRepeatTimes(x :Integer);
begin
  if x = prRepeatTimes then Exit;
  prRepeatTimes  := x;
  Tree['RepeatTimes'].Data[0] := x;
  if Assigned(OnRepeatChanged) then OnRepeatChanged(Self);
end;
}



// *****************************************************************************
end.
