{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// apparently does some work for QUB_QFS_QDF


unit QUB_QFS_Segments;


interface


uses


SysUtils, Classes,
Alloc,
QUB_QFS, QUB_QFS_Edit, QUB_Node;


const


QUB_QFS_QDF_CHANNELS_NAME                      = 'Channels';

QUB_QFS_QDF_SEGMENTS_NAME                      = 'Segments';
QUB_QFS_QDF_SEGMENT_NAME                       = 'Segment';

QUB_QFS_QDF_SEGMENTSTARTTIME_NAME              = 'StartTime';


type


//Forward class declarations ***************************************************


TQUB_QFS_Segments                              = class;
TQUB_QFS_Segment                               = class;
TQUB_QFS_Channels                              = class;


//Class definitions

//Record_Edit
TQUB_QFS_Segments                              = class(TQUB_QFS_Record)
protected


  prTotalDataCount                             :Integer;


  function prGetSegment(s :Integer) :TQUB_QFS_Segment;

  procedure prUpdateDataCount;


public


  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  function AddRecord(r :TQUB_QFS_Record) :Boolean; override;
  function AddRecord(r :TQUB_QFS_Record; Index :Integer) :Boolean; override;
  function DelRecord(r :TQUB_QFS_Record) :Boolean; override;


  property Segment[Index :Integer]             :TQUB_QFS_Segment read prGetSegment; default;
  property SegmentCount                        :Integer          read prRecordCount;

  property TotalDataCount                      :Integer          read prTotalDataCount;


end;


TQUB_QFS_Segment                               = class(TQUB_QFS_Record)
protected


  prChannels                                   :TQUB_QFS_Channels;
  prRecStartTime                               :TQUB_QFS_Record;


  function prGetChannels :TQUB_QFS_Channels; virtual;
  function prGetStartTime :Double;
  procedure prSetStartTime(t :Double);


public


  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;


  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;


  property Channels                            :TQUB_QFS_Channels read prGetChannels;
  property StartTime                           :Double            read prGetStartTime  write prSetStartTime;


end;


TQUB_QFS_Channels                              = class(TQUB_QFS_Edit)
protected


  procedure prUpdateDataCount; override;


public


  function Read(ReadStream :TStream) :Boolean; override;


end;


implementation


//******************************************************************************
//Segments should only have children of type Segment !!!

function TQUB_QFS_Segments.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = QUB_QFS_QDF_SEGMENT_NAME    then begin
      Result := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, Self);
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

function TQUB_QFS_Segments.AddRecord(r :TQUB_QFS_Record) :Boolean;
begin
  Result := inherited AddRecord(r);
  if not Result then begin
    if r is TQUB_QFS_Segment then begin
      prUpdateDataCount;
    end;
  end;
end;

function TQUB_QFS_Segments.AddRecord(r :TQUB_QFS_Record; Index :Integer) :Boolean;
begin
  Result := inherited AddRecord(r, Index);
  if not Result then begin
    if r is TQUB_QFS_Segment then begin
      prUpdateDataCount;
    end;
  end;
end;

function TQUB_QFS_Segments.DelRecord(r :TQUB_QFS_Record) :Boolean;
begin
  Result := inherited DelRecord(r);
  if not Result then begin
    if r is TQUB_QFS_Segment then begin
      prUpdateDataCount;
    end;
  end;
end;

function TQUB_QFS_Segments.prGetSegment(s :Integer) :TQUB_QFS_Segment;
begin
  Result := TQUB_QFS_Segment(Records[s]);
end;

procedure TQUB_QFS_Segments.prUpdateDataCount;
var
  s :Integer;
begin
  prTotalDataCount := 0;
  for s := 0 to SegmentCount - 1 do begin
    if Segment[s].Channels <> nil then begin
      Inc(prTotalDataCount, Segment[s].Channels.DataCount);
    end;
  end;
end;

//******************************************************************************
constructor TQUB_QFS_Segment.Create(const Name :String; Parent :TQUB_QFS_Record);
begin
  inherited Create(Name, Parent);

  prRecStartTime        := TQUB_QFS_Record.CreateAsValue(QUB_QFS_QDF_SEGMENTSTARTTIME_NAME, Self, True{Permanent}, 0, NODE_DATATYPE_DOUBLE, 1{Count});

  prChannels            := TQUB_QFS_Channels.Create(QUB_QFS_QDF_CHANNELS_NAME, Self);
  prChannels.Permanent  := True;

end;

function TQUB_QFS_Segment.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = QUB_QFS_QDF_CHANNELS_NAME            then begin
      Result := prChannels;
    end else if NodeName = QUB_QFS_QDF_SEGMENTSTARTTIME_NAME    then begin
      Result := prRecStartTime;
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

function TQUB_QFS_Segment.prGetChannels :TQUB_QFS_Channels;
begin
  Result := prChannels;
end;

function TQUB_QFS_Segment.prGetStartTime :Double;
begin
  Result := prRecStartTime.DataValue;
end;

procedure TQUB_QFS_Segment.prSetStartTime(t :Double);
begin
  prRecStartTime.DataValue := t;
end;


//******************************************************************************
function TQUB_QFS_Channels.Read(ReadStream :TStream) :Boolean;
begin

  Result := inherited Read(ReadStream);

  prCheckHasEditSelection;
  prUpdateDataCount;

end;


procedure TQUB_QFS_Channels.prUpdateDataCount;
begin
  inherited prUpdateDataCount;

  if Parent.Parent is TQUB_QFS_Segments then TQUB_QFS_Segments(Parent.Parent).prUpdateDataCount;

end;


//******************************************************************************
end.
