{

                              Qub Visual Objects
                                     unit



(C) 2000 Lorin Milescu
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


//01-31-2000   Creation date


unit QUB_QVO;


interface


uses


Windows, SysUtils, Classes, SyncObjs, Controls, Graphics, Menus, Dialogs, Forms,
Alloc,

QUB_Global,
QUB_Node,
QUB_QFS_File,
QUB_Events;


type


//Forward class declarations ***************************************************
TQUB_QVO                                       = class;


//Object
TQUB_QVO                                       = class(TCustomControl)
protected


  prCanvas                                     :TCanvas;
  prBitmapCanvas                               :TCanvas;
  prBitmap                                     :TBitmap;

  prDrawWidth                                  :Integer;
  prDrawHeight                                 :Integer;

  prDrawFrame                                  :Boolean;
  prDrawFocusRect                              :Boolean;

  prTipsOn                                     :Boolean;

  //Margin
  prMargin                                     :TRect;

  //Colors
  prColor                                      :TColor;
  prColorBack                                  :TColor;
  prColorFrame                                 :TColor;
  prColorText                                  :TColor;
  
  //Flags and other settings
  prDrawing                                    :Boolean;
  prChanging                                   :Boolean;
  prNeedsChange                                :Boolean;
  prMustRedraw                                 :Boolean;
  prNoDraw                                     :Boolean;
  prDrawForPrinting                            :Boolean;
  prDrawBW                                     :Boolean;

  prMouseDowning                               :Boolean;  // (for subclasses) True when mouse drag selection is occuring
  prMouseRunning                               :Boolean;
  prHasDblClicked                              :Boolean;
  prMouseHitCount                              :Integer;

  prPrintPenWidth                              :Integer;
  prPrintFontSize                              :Integer;


  prSelected                                   :Boolean;
  prActive                                     :Boolean;

  prLockCloseCount                             :Integer;
  prLockWriteCount                             :Integer;
  prLockReadWriteCount                         :Integer;

  //Popup menus
  prPopUpMenu                                  :TPopUpMenu;
  prEditCopyImageItem                          :TMenuItem;
  prPrintItem                                  :TMenuItem;
  prPropertiesItem                             :TMenuItem;

  //List
  prAssociatedObject                           :TObject;

  prEvents                                     :TQUB_Events;

  //Draw Critical Section
  prDrawCritSection                            :TCriticalSection;


  //Events
  procedure DoOnActivate;
  procedure DoOnDeActivate;
  procedure DoOnEdit;
  procedure DoEnter; override;
  procedure DoExit; override;

  //File Events - called by the QDF object
  procedure DoOnFileOpen(Sender :TObject); virtual;
  procedure DoOnFileClose(Sender :TObject); virtual;
  procedure DoOnFileEdit(Sender :TObject; EditOper :TQUB_EditOperation; var Param; var Param2); virtual;

  //Mouse
  procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;

  //Lock
  function prGetLockedClose :Boolean;
  function prGetLockedAny :Boolean;
  function prGetLockedWrite :Boolean;
  function prGetLockedReadWrite :Boolean;

  //Draw
  procedure prErase; virtual;

  //Print
  function prPrint(Dlg :TPrintDialog) :Boolean; virtual; abstract;

  //Set Margin
  procedure prSetMargin(Index: Integer; m: Integer); virtual;

  //Set Colors
  procedure prSetColor(c :TColor); virtual;
  procedure prSetColorBack(c :TColor); virtual;
  procedure prSetColorFrame(c :TColor); virtual;
  procedure prSetColorText(c :TColor); virtual;

  //Draw
  procedure prSetDrawFrame(f :Boolean);

  procedure prSetSelected(s :Boolean); virtual;

  //Tips
  procedure prSetTipsOn(t :Boolean); virtual;

  //Registry, properties
  procedure prSetInitialProperties; virtual;
  procedure prPropertiesRegistry(Read :Boolean); virtual;

  //Properties
  procedure prOnProperties(Sender :TObject); virtual;
  procedure prOnGetProperties(Sender :TObject); virtual;
  procedure prOnPutProperties(Sender :TObject); virtual;
  procedure prOnResetProperties(Sender :TObject); virtual;
  procedure prOnSetAsDefaultProperties(Sender :TObject); virtual;

  //Popup Menu
  procedure prCreatePopUpMenu; virtual;

  //Popup Menu events
  procedure prOnCopyImage(Sender :TObject); virtual;
  procedure prOnPrint(Sender :TObject); virtual;
  procedure prOnPopUp(Sender :TObject); virtual;

  //Flag methods
  function prGetFileName :String; virtual;
  function prGetFileExt :String; virtual;
  function prGetModified :Boolean; virtual;
  function prGetSaved :Boolean; virtual;

  //Print
  function prGetMinPrintPage :Integer; virtual;
  function prGetMaxPrintPage :Integer; virtual;

  //File methods
  function New(const FileName :String) :Boolean; virtual; abstract;
  function Open(const FileName :String) :Boolean; overload; virtual; abstract;
  function Open(const FileName :String; OpenStream :TStream) :Boolean; overload; virtual; abstract;
  function Save(const FileName :String) :Boolean; virtual; abstract;
  function Close :Boolean; virtual; abstract;
  function OpenFromString(const MemFile :String) :Boolean; virtual; abstract;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; virtual; abstract;

  function Print :Boolean; virtual;

  function CopyImage(ShowProperties :Boolean) :Boolean; virtual;

  function Properties :Boolean; virtual;
  function NodeToProperties(Node :INode) :Boolean; virtual;
  function PropertiesToNode(Node :INode) :Boolean; virtual;

  procedure DrawFocusRect;

  //-----support for descendants
  //Note : add a class with common QVD / QVT elements descendant from QVO ?
  function TimeText(d : Double):String;


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

  procedure Paint; override;

  procedure Activate; virtual;
  procedure Deactivate; virtual;

  function KeyShift   :Boolean;
  function KeyControl :Boolean;

  procedure LockClose; virtual;
  procedure UnLockClose; virtual;
  procedure LockAny; virtual;
  procedure UnLockAny; virtual;
  procedure LockWrite; virtual;
  procedure UnLockWrite; virtual;
  procedure LockReadWrite; virtual;
  procedure UnLockReadWrite; virtual;

  function Draw :Boolean; virtual;
  function DrawTo(Canvas :TCanvas; Left, Right, Width, Height :Integer) :Boolean; virtual;

  function AddPopUpMenuItem(sc : TShortCut; cCaption :String; evtOnClick :TNotifyEvent; mnuPopUp :TPopUpMenu; var iPos :Integer; mnuParent :TMenuItem = nil) :TMenuItem;


  //Active
  property Active                              :Boolean                read prActive;

  //Bitmap
  property Bitmap                              :TBitmap                read prBitmap;
  property BitmapCanvas                        :TCanvas                read prBitmapCanvas              write prBitmapCanvas;

  property Events                              :TQUB_Events            read prEvents;

  property AssociatedObject                    :TObject                read prAssociatedObject          write prAssociatedObject;

  property DrawWidth                           :Integer                read prDrawWidth                 write prDrawWidth;
  property DrawHeight                          :Integer                read prDrawHeight                write prDrawHeight;
  property DrawForPrinting                     :Boolean                read prDrawForPrinting           write prDrawForPrinting;
  property DrawBW                              :Boolean                read prDrawBW                    write prDrawBW;
  property PrintPenWidth                       :Integer                read prPrintPenWidth             write prPrintPenWidth;
  property PrintFontSize                       :Integer                read prPrintFontSize             write prPrintFontSize;
  property NoDraw                              :Boolean                read prNoDraw                    write prNoDraw;

  //Tips
  property TipsOn                              :Boolean                read prTipsOn                    write prSetTipsOn;


  //Flags
  property FileName                            :String                 read prGetFileName;
  property FileExt                             :String                 read prGetFileExt;
  property Modified                            :Boolean                read prGetModified;
  property Saved                               :Boolean                read prGetSaved;
  property Selected                            :Boolean                read prSelected                  write prSetSelected;

  property LockedClose                         :Boolean                read prGetLockedClose;
  property LockedAny                           :Boolean                read prGetLockedAny;
  property LockedWrite                         :Boolean                read prGetLockedWrite;
  property LockedReadWrite                     :Boolean                read prGetLockedReadWrite;

  //Print
  property MinPrintPage                        :Integer                read prGetMinPrintPage;
  property MaxPrintPage                        :Integer                read prGetMaxPrintPage;


published


  property Align;
  property PopupMenu;
  property ShowHint;
  property TabOrder;
  property TabStop;
  property Visible;

  //Margin
  property Margin                              :TRect                  read prMargin;
  property MarginLeft                          :Integer    index 0     read prMargin.Left               write prSetMargin;
  property MarginTop                           :Integer    index 1     read prMargin.Top                write prSetMargin;
  property MarginRight                         :Integer    index 2     read prMargin.Right              write prSetMargin;
  property MarginBottom                        :Integer    index 3     read prMargin.Bottom             write prSetMargin;

  //Colors
  property Color                               :TColor                 read prColor                     write prSetColor;
  property ColorBack                           :TColor                 read prColorBack                 write prSetColorBack;
  property ColorFrame                          :TColor                 read prColorFrame                write prSetColorFrame;
  property ColorText                           :TColor                 read prColorText                 write prSetColorText;

  //Draw
  property DrawFrame                           :Boolean                read prDrawFrame                 write prSetDrawFrame;


end;
//..............................................................................


//******************************************************************************
implementation


//****************************************************************************//
//                                                                            //
//                              TQUB_QVO                                      //
//                                                                            //
//****************************************************************************//
constructor TQUB_QVO.Create(AOwner :TComponent);
begin
  inherited Create(AOwner);

  prDrawCritSection   := TCriticalSection.Create;

  prEvents            := TQUB_Events.Create;

  ControlStyle        := ControlStyle + [csReplicatable, csOpaque, csCaptureMouse] - [csNoStdEvents];

  SetBounds(Left, Top, 105, 105);

  prCanvas            := Canvas;
  prBitmap            := TBitmap.Create;
  prBitmap.Width      := Width;
  prBitmap.Height     := Height;
  prBitmapCanvas      := prBitmap.Canvas;

  prColor             := clBtnFace;

  Align               := alClient;

end;

destructor TQUB_QVO.Destroy;
begin
  try

    prEvents.Free;
    prEvents := nil;
    prBitmap.Free;

    prDrawCritSection.Free;

  finally

    inherited;

  end;
end;

//Activate/DeActivate
procedure TQUB_QVO.Activate;
begin
  BringToFront;
  prActive := True;
  DrawFocusRect;
  if (CanFocus) and (Visible) and (Enabled) then if not Focused then try SetFocus except end;
  DoOnActivate;
end;

procedure TQUB_QVO.Deactivate;
begin
  prActive := False;
  DrawFocusRect;
  prMouseHitCount := 0;
  DoOnDeActivate;
end;

//Events
procedure TQUB_QVO.DoOnActivate;
begin
  if Events <> nil then Events.RaiseNotifyEvent('OnActivate', Self);
end;

procedure TQUB_QVO.DoOnDeActivate;
begin
  if Events <> nil then Events.RaiseNotifyEvent('OnDeActivate', Self);
end;

procedure TQUB_QVO.DoOnEdit;
begin
  if Events <> nil then Events.RaiseNotifyEvent('OnEdit', Self);
end;

procedure TQUB_QVO.DoEnter;
begin
  inherited;
  if Events <> nil then Events.RaiseNotifyEvent('OnEnter', Self);
end;

procedure TQUB_QVO.DoExit;
begin
  inherited;
  prMouseHitCount := 0;
  if Events <> nil then Events.RaiseNotifyEvent('OnExit', Self);
end;

//File Events - called by the associated QDF object, and redefined in descendent classes
procedure TQUB_QVO.DoOnFileOpen(Sender :TObject);
begin
end;

procedure TQUB_QVO.DoOnFileClose(Sender :TObject);
begin
end;

procedure TQUB_QVO.DoOnFileEdit(Sender :TObject; EditOper :TQUB_EditOperation; var Param; var Param2);
begin
end;

//Mouse
procedure TQUB_QVO.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;

  if Button = mbLeft then Inc(prMouseHitCount);
  if prMouseHitCount > 2 then prMouseHitCount := 2;

  if CanFocus then if not Focused then try SetFocus except end;

end;

procedure TQUB_QVO.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;

  if not Focused then try SetFocus except end;
end;


//Paint
procedure TQUB_QVO.Paint;
begin
  try
    if Parent <> nil then begin
      if not IsWindowVisible(Handle) then Exit; 
      try
        if not (csDestroying in ComponentState) then begin
          if not prNoDraw then begin
            try
              prCanvas.Lock;
              prBitmapCanvas.Lock;
              prCanvas.Draw(0, 0, prBitmap);
              DrawFocusRect;
            finally
              prBitmapCanvas.UnLock;
              prCanvas.Unlock;
            end;
          end;
        end;
      finally
        inherited;
      end;
    end;
  except
  end;
end;

procedure TQUB_QVO.DrawFocusRect;
begin
  if not prDrawFocusRect then Exit;
  if prCanvas = nil then Exit;
  try
    prCanvas.Lock;
    if Active then begin
      prCanvas.Pen.Color := ColorFrame;
    end else begin
      prCanvas.Pen.Color := Color;
    end;
    prCanvas.Pen.Style := psDot;
    prCanvas.Pen.Mode  := pmCopy;
    prCanvas.Pen.Width := 1;
    prCanvas.Brush.Style := bsClear;
    prCanvas.Rectangle(MarginLeft - 2, MarginTop - 2, prDrawWidth - MarginRight + 1, prDrawHeight - MarginBottom + 1);
  finally
    prCanvas.Unlock;
  end;
end;

function TQUB_QVO.Print :Boolean;
var
  Dlg :TPrintDialog;
begin

  Dlg := TPrintDialog.Create(Self);
  Dlg.MinPage := MinPrintPage;
  Dlg.MaxPage := MaxPrintPage;
  Dlg.Options := [poPageNums, poSelection];
  Dlg.PrintRange := prSelection;

  try
    Result := not Dlg.Execute;
    if not Result then Result := prPrint(Dlg);
  finally
    Dlg.Free;
  end;

end;

function TQUB_QVO.CopyImage(ShowProperties :Boolean) :Boolean;
begin
  Result := False;
end;

//Properties
function TQUB_QVO.Properties :Boolean;
begin
  Result := False;
end;

function TQUB_QVO.NodeToProperties(Node :INode) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO.PropertiesToNode(Node :INode) :Boolean; 
begin
  Result := False;
end;

//Draw
function TQUB_QVO.Draw :Boolean;
begin
  Result := False;
  prMustRedraw := True;
  Paint;
end;

function TQUB_QVO.DrawTo(Canvas :TCanvas; Left, Right, Width, Height :Integer) :Boolean;
begin
  Result := False;
end;

//Lock
procedure TQUB_QVO.LockClose;
begin
  Inc(prLockCloseCount);
end;

procedure TQUB_QVO.UnLockClose;
begin
  Dec(prLockCloseCount);
  if prLockCloseCount < 0 then prLockCloseCount := 0;
end;

procedure TQUB_QVO.LockAny;
begin
  LockWrite;
  LockReadWrite;
end;

procedure TQUB_QVO.UnLockAny;
begin
  UnLockWrite;
  UnLockReadWrite;
end;

procedure TQUB_QVO.LockWrite;
begin
  Inc(prLockWriteCount);
end;

procedure TQUB_QVO.UnLockWrite;
begin
  Dec(prLockWriteCount);
  if prLockWriteCount < 0 then prLockWriteCount := 0;
end;

procedure TQUB_QVO.LockReadWrite;
begin
  Inc(prLockReadWriteCount);
end;

procedure TQUB_QVO.UnLockReadWrite;
begin
  Dec(prLockReadWriteCount);
  if prLockReadWriteCount < 0 then prLockReadWriteCount := 0;
end;

//******************************************************************************
procedure TQUB_QVO.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  try
    if prBitmap <> nil then begin
      prBitmap.Width  := AWidth;
      prBitmap.Height := AHeight;
    end;
    prDrawWidth       := AWidth;
    prDrawHeight      := AHeight;

    prMustRedraw := True;

  finally
    inherited;
  end;
end;


//Locked
function TQUB_QVO.prGetLockedClose :Boolean;
begin
  Result := prLockCloseCount > 0;
end;

function TQUB_QVO.prGetLockedAny :Boolean;
begin
  Result := (LockedWrite) or (LockedReadWrite);
end;

function TQUB_QVO.prGetLockedWrite :Boolean;
begin
  Result := prLockWriteCount > 0;
end;

function TQUB_QVO.prGetLockedReadWrite :Boolean;
begin
  Result := prLockReadWriteCount > 0;
end;

//Key status
function TQUB_QVO.KeyShift :Boolean;
begin
  Result := GetAsyncKeyState(VK_SHIFT) < 0;
end;

function TQUB_QVO.KeyControl :Boolean;
begin
  Result := GetAsyncKeyState(VK_CONTROL) < 0;
end;

procedure TQUB_QVO.prErase;
begin
  if not prDrawForPrinting then begin
    prBitmapCanvas.Pen.Width := 0;
    prBitmapCanvas.Pen.Style := psClear;
    prBitmapCanvas.Brush.Style := bsSolid;
    prBitmapCanvas.Brush.Color := prColor;
    prBitmapCanvas.Rectangle(0, 0, prDrawWidth + 1, prDrawHeight + 1);
  end;
end;

procedure TQUB_QVO.prCreatePopUpMenu;
begin

  //Create PopUp Menu **********************************************************
  prPopUpMenu                     := TPopUpMenu.Create(Self);

  //PRINT ----------------------------------------------------------------------
  prPrintItem                     := TMenuItem.Create(Self);
  prPrintItem.Caption             := 'Print';
  prPrintItem.OnClick             := prOnPrint;
  prPopUpMenu.Items.Add(prPrintItem);
  //COPY IMAGE -----------------------------------------------------------------
  prEditCopyImageItem             := TMenuItem.Create(Self);
  prEditCopyImageItem.Caption     := 'Copy Image';
  prEditCopyImageItem.OnClick     := prOnCopyImage;
  prPopUpMenu.Items.Add(prEditCopyImageItem);
  //----------------------------------------------------------------------------
  prPopUpMenu.Items.InsertNewLineAfter(prEditCopyImageItem);
  //PROPERTIES -----------------------------------------------------------------
  prPropertiesItem                := TMenuItem.Create(Self);
  prPropertiesItem.ShortCut       := ShortCut(Word('P'), [ssCtrl]);
  prPropertiesItem.Caption        := '&Properties';
  prPropertiesItem.OnClick        := prOnProperties;
  prPopUpMenu.Items.Add(prPropertiesItem);
  //----------------------------------------------------------------------------

  prPopUpMenu.OnPopUp             := prOnPopUp;
  PopUpMenu                       := prPopUpMenu;

end;


//Popup Menu Events
procedure TQUB_QVO.prOnCopyImage(Sender :TObject);
begin
end;

procedure TQUB_QVO.prOnPrint(Sender :TObject);
begin
end;

procedure TQUB_QVO.prOnPopUp(Sender :TObject);
begin
end;

//Flag methods
function TQUB_QVO.prGetFileName :String;
begin
  Result := '';
end;

function TQUB_QVO.prGetFileExt :String;
begin
  Result := '';
end;

function TQUB_QVO.prGetModified :Boolean;
begin
  Result := False;
end;

function TQUB_QVO.prGetSaved :Boolean;
begin
  Result := False;
end;

//Print properties
function TQUB_QVO.prGetMinPrintPage :Integer;
begin
  Result := 1;
end;

function TQUB_QVO.prGetMaxPrintPage :Integer;
begin
  Result := 1;
end;


//Properties *******************************************************************
procedure TQUB_QVO.prOnProperties(Sender :TObject);
begin
end;

procedure TQUB_QVO.prOnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_QVO.prOnPutProperties(Sender :TObject);
begin
end;

procedure TQUB_QVO.prOnResetProperties(Sender :TObject);
begin
end;

procedure TQUB_QVO.prOnSetAsDefaultProperties(Sender :TObject);
begin
end;

//******************************************************************************
procedure TQUB_QVO.prSetInitialProperties;
begin
end;

procedure TQUB_QVO.prPropertiesRegistry(Read :Boolean);
begin
end;

//******************************************************************************
procedure TQUB_QVO.prSetMargin(Index :Integer; m :Integer);
begin
  with prMargin do begin
    case Index of
      0: Left   := m;
      1: Top    := m;
      2: Right  := m;
      3: Bottom := m;
    end;
  end;
  prMustRedraw := True;
end;

//******************************************************************************
procedure TQUB_QVO.prSetColor(c :TColor);
begin
  prColor := c;
  prMustRedraw := True;
end;

procedure TQUB_QVO.prSetColorBack(c :TColor);
begin
  prColorBack := c;
  prMustRedraw := True;
end;

procedure TQUB_QVO.prSetColorFrame(c :TColor);
begin
  prColorFrame := c;
  prMustRedraw := True;
end;

procedure TQUB_QVO.prSetColorText(c :TColor);
begin
  prColorText  := c;
  prMustRedraw := True;
end;

//Draw
procedure TQUB_QVO.prSetDrawFrame(f :Boolean);
begin
  prDrawFrame := f;
  prMustRedraw := True;
end;

//******************************************************************************
procedure TQUB_QVO.prSetSelected(s :Boolean);
begin
  prSelected := s;
end;

//******************************************************************************
procedure TQUB_QVO.prSetTipsOn(t :Boolean);
begin
  prTipsOn := t;
  ShowHint := t;
end;

//----- Convert time value to string of minimal units, format [ms|us|s]= #.###
function TQUB_QVO.TimeText(d : Double):String;
begin
  if d < 0.001      then begin
    Result := Format('[us] = %10.6g', [d*1e6]);
  end else if d < 1 then begin
    Result := Format('[ms] = %10.6g', [d*1e3]);
  end else begin
    Result := Format('[s] = %10.6g', [d]);
  end;
end;

//----- Create a menu item and add it to the popup menu
function TQUB_QVO.AddPopUpMenuItem(sc :TShortCut; cCaption :String; evtOnClick :TNotifyEvent; mnuPopUp :TPopUpMenu; var iPos :Integer; mnuParent :TMenuItem = nil) :TMenuItem;
var
  item :TMenuItem;
begin
  item := TMenuItem.Create(Self);
  if sc <> 0 then item.ShortCut := sc;
  if Length(cCaption) > 0 then item.Caption := cCaption;
  if @evtOnClick <> nil then item.OnClick := evtOnClick;
  if mnuPopUp <> nil then mnuPopUp.Items.Insert(iPos, item);
  if mnuParent <> nil then mnuParent.Add(item);
  if @iPos <> nil then Inc(iPos);
  Result := item;
end;


//******************************************************************************
end.

