{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// chop data properties


unit QUB_QVO_QVD_Chop_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Registry,
QUB_Classic_Properties,
QUB_Node,
RegistryInterface, ExtCtrls;


{$include 'QUB_QVO_QVD_Chop_Properties.hct'}


const


DEF_IGNORESEGMENTS           = False;
DEF_ONCE                     = False;
DEF_DISCARDPARTIAL           = False;
DEF_ADDTOLIST                = False;

MIN_POINTS                   = 0;
MAX_POINTS                   = High(Integer) - 1;
DEF_POINTS                   = 0;

MIN_GETPOINTS                = 1;
DEF_GETPOINTS                = 10000;

DEF_OUTPUTLIST               = 'Chopped segments';

DEF_USEGETLIST               = False;


type


TQUB_QVO_QVD_Chop_PropertiesForm = class(TQUB_Classic_PropertiesForm)


  Label1                         :TLabel;
  Label2                         :TLabel;
  Label3                         :TLabel;
  Label4                         :TLabel;
  Label5                         :TLabel;
  Label6                         :TLabel;
  Label7                         :TLabel;

  txtSkipBeforePoints            :TEdit;
  txtSkipBeforeTime              :TEdit;
  txtGetPoints                   :TEdit;
  txtGetTime                     :TEdit;
  txtSkipAfterPoints             :TEdit;
  txtSkipAfterTime               :TEdit;

  txtOutputList                  :TEdit;

  chkIgnoreSegments              :TCheckBox;
  chkOnce                        :TCheckBox;
  chkDiscardPartial              :TCheckBox;
  chkAddToList                   :TCheckBox;
  lstGetList                     :TComboBox;
  chkUseGetList                  :TCheckBox;
  Button1                        :TButton;
    panDataSource: TPanel;
    Label8: TLabel;
    chkDataSourceSel: TRadioButton;
    chkDataSourceList: TRadioButton;
    chkDataSourceFile: TRadioButton;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure btnOKClick(Sender: TObject);

  procedure txtSkipBeforePointsKeyPress(Sender: TObject; var Key: Char);
  procedure txtSkipBeforePointsExit(Sender: TObject);
  procedure txtSkipBeforeTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtSkipBeforeTimeExit(Sender: TObject);
  procedure txtGetPointsKeyPress(Sender: TObject; var Key: Char);
  procedure txtGetPointsExit(Sender: TObject);
  procedure txtGetTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtGetTimeExit(Sender: TObject);
  procedure txtSkipAfterPointsKeyPress(Sender: TObject; var Key: Char);
  procedure txtSkipAfterPointsExit(Sender: TObject);
  procedure txtSkipAfterTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtSkipAfterTimeExit(Sender: TObject);
  procedure chkIgnoreSegmentsClick(Sender: TObject);
  procedure chkOnceClick(Sender: TObject);
  procedure chkDiscardPartialClick(Sender: TObject);
  procedure chkAddToListClick(Sender: TObject);
  procedure chkUseGetListClick(Sender: TObject);
  procedure lstGetListClick(Sender: TObject);
  procedure btnGrabClick(Sender: TObject);
    procedure chkDataSourceListClick(Sender: TObject);


protected


  prSamplingTime                 :Extended;

  prSkipBeforePoints             :Integer;
  prSkipAfterPoints              :Integer;
  prGetPoints                    :Integer;

  prIgnoreSegments               :Boolean;
  prOnce                         :Boolean;
  prDiscardPartial               :Boolean;
  prAddToList                    :Boolean;
  
  prOutputList                   :String;

  prUseGetList                   :Boolean;
  prGetListName                  :String;

  prDialogOnLClick               :Boolean;


  procedure prSetSkipBeforePoints(p :Integer);
  procedure prSetSkipAfterPoints(p :Integer);
  procedure prSetGetPoints(p :Integer);
  procedure prSetIgnoreSegments(s :Boolean);
  procedure prSetOnce(o :Boolean);
  procedure prSetDiscardPartial(x :Boolean);
  procedure prSetOutputList(o :String);
  procedure prSetAddToList(a :Boolean);
  procedure prSetUseGetList(u :Boolean);
  procedure prSetGetListName(n :String);

  procedure prPropertiesRegistry(Read :Boolean);


public


  m_iGrabStart, m_iGrabLength : Integer;        // preloaded setting for SkipBeforePoints and GetPoints to use if [Grab] is clicked.


  function PropertiesTree(cfg :INode): INode;


  property SamplingTime          :Extended       read prSamplingTime      write prSamplingTime;

  property SkipBeforePoints      :Integer        read prSkipBeforePoints  write prSetSkipBeforePoints;
  property SkipAfterPoints       :Integer        read prSkipAfterPoints   write prSetSkipAfterPoints;
  property GetPoints             :Integer        read prGetPoints         write prSetGetPoints;
  property IgnoreSegments        :Boolean        read prIgnoreSegments    write prSetIgnoreSegments;
  property Once                  :Boolean        read prOnce              write prSetOnce;
  property DiscardPartial        :Boolean        read prDiscardPartial    write prSetDiscardPartial;
  property AddToList             :Boolean        read prAddToList         write prSetAddToList;

  property OutputList            :String         read prOutputList        write prSetOutputList;

  property UseGetList            :Boolean        read prUseGetList        write prSetUseGetList;
  property GetListName           :String         read prGetListName       write prSetGetListName;

  property DialogOnLClick        :Boolean        read prDialogOnLClick    write prDialogOnLClick;


end;


//******************************************************************************
implementation


uses


QUB_QVO_QVD,
QUB_GUIInterface,
QUB_Task,
QUB_Main;


{$R *.DFM}


//******************************************************************************

//----- Set skip before points and convert to time units also
procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetSkipBeforePoints(p :Integer);
begin
  prSkipBeforePoints := p;
  txtSkipBeforePoints.Text := IntToStr(p);
  txtSkipBeforeTime.Text   := FloatToStr(p * SamplingTime * 1e3);
end;

//----- Set skip after points and convert to time units also
procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetSkipAfterPoints(p :Integer);
begin
  prSkipAfterPoints := p;
  txtSkipAfterPoints.Text := IntToStr(p);
  txtSkipAfterTime.Text   := FloatToStr(p * SamplingTime * 1e3);
end;

//----- Set point count and convert to time units also
procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetGetPoints(p :Integer);
begin
  prGetPoints := p;
  txtGetPoints.Text := IntToStr(p);
  txtGetTime.Text   := FloatToStr(p * SamplingTime * 1e3);
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetIgnoreSegments(s :Boolean);
begin
  prIgnoreSegments := s;
  chkIgnoreSegments.Checked := prIgnoreSegments;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetOnce(o :Boolean);
begin
  prOnce := o;
  chkOnce.Checked := prOnce;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetDiscardPartial(x :Boolean);
begin
  prDiscardPartial := x;
  chkDiscardPartial.Checked := prDiscardPartial;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetAddToList(a :Boolean);
begin
  prAddToList := a;
  if chkAddToList.Checked <> prAddToList then chkAddToList.Checked := prAddToList;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetOutputList(o :String);
begin
  prOutputList := o;
  txtOutputList.Text := prOutputList;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetUseGetList(u :Boolean);
begin
  prUseGetList := u;
  if chkUseGetList.Checked <> prUseGetList then chkUseGetList.Checked := prUseGetList;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.prSetGetListName(n :String);
begin
  prGetListName := n;
  //???
end;

//******************************************************************************
procedure TQUB_QVO_QVD_Chop_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(True{Read});
  {$include 'QUB_QVO_QVD_Chop_Properties.htx'}

  prDialogOnLClick             := True;

  WikiTopic                    := 'Preprocessing:Chop_Data';

end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.FormDestroy(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(False{Write});

end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.FormShow(Sender: TObject);
var
  Data :TQUB_QVO_QVD;
begin
  inherited;

  lstGetList.Items.Clear;
  Data := ViewportData.CurrData;
  if Data <> nil then lstGetList.Items.Assign(Data.SelectionLists.Items);
  lstGetList.ItemIndex := -1;

  IgnoreSegments   := IgnoreSegments;
  Once             := Once;
  DiscardPartial   := DiscardPartial;
  AddToList        := AddToList;
  SkipBeforePoints := SkipBeforePoints;
  SkipAfterPoints  := SkipAfterPoints;
  GetPoints        := GetPoints;
  OutputList       := OutputList;
  UseGetList       := UseGetList;

  chkDataSourceSel.Checked := chkDataSourceSel.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceList.Checked := chkDataSourceList.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceFile.Checked := (chkDataSourceFile.Tag = Integer(QUB_MainForm.DataSource)) or (DATA_SOURCE_FILELIST = QUB_MainForm.DataSource);
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.btnOKClick(Sender: TObject);
begin
  OutputList := Trim(txtOutputList.Text);
  if OutputList = '' then OutputList := DEF_OUTPUTLIST;
  inherited;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.btnGrabClick(Sender: TObject);
begin
  inherited;
  SkipBeforePoints := m_iGrabStart;
  GetPoints := m_iGrabLength;
end;

//******************************************************************************
procedure TQUB_QVO_QVD_Chop_PropertiesForm.chkIgnoreSegmentsClick(Sender: TObject);
begin
  inherited;
  IgnoreSegments := TCheckBox(Sender).Checked;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.chkOnceClick(Sender: TObject);
begin
  inherited;
  Once := TCheckBox(Sender).Checked;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.chkDiscardPartialClick(Sender: TObject);
begin
  inherited;
  DiscardPartial := TCheckBox(Sender).Checked;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.chkAddToListClick(Sender: TObject);
begin
  inherited;
  AddToList := TCheckBox(Sender).Checked;
end;

//Skip Before
procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipBeforePointsKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSkipBeforePointsExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipBeforePointsExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := StrToInt(TEdit(Sender).Text);
    if ((p < MIN_POINTS) or (p > MAX_POINTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_POINTS) + ' and ' + IntToStr(MAX_POINTS), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(SkipBeforePoints);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> SkipBeforePoints then begin
    SkipBeforePoints := p;
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipBeforeTimeKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSkipBeforeTimeExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipBeforeTimeExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := Round((StrToFloat(TEdit(Sender).Text) * 1e-3) / SamplingTime);
    if ((p < MIN_POINTS) or (p > MAX_POINTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_POINTS * 1e3 * SamplingTime) + ' and ' + FloatToStr(MAX_POINTS * 1e3 * SamplingTime), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(SkipBeforePoints * 1e3 * SamplingTime);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> SkipBeforePoints then begin
    SkipBeforePoints := p;
  end;
end;

//Skip After
procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipAfterPointsKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSkipAfterPointsExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipAfterPointsExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := StrToInt(TEdit(Sender).Text);
    if ((p < MIN_POINTS) or (p > MAX_POINTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_POINTS) + ' and ' + IntToStr(MAX_POINTS), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(SkipAfterPoints);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> SkipAfterPoints then begin
    SkipAfterPoints := p;
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipAfterTimeKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSkipAfterTimeExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtSkipAfterTimeExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := Round((StrToFloat(TEdit(Sender).Text) * 1e-3) / SamplingTime);
    if ((p < MIN_POINTS) or (p > MAX_POINTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_POINTS * 1e3 * SamplingTime) + ' and ' + FloatToStr(MAX_POINTS * 1e3 * SamplingTime), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(SkipAfterPoints * 1e3 * SamplingTime);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> SkipAfterPoints then begin
    SkipAfterPoints := p;
  end;
end;

//Get Points
procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtGetPointsKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtGetPointsExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtGetPointsExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := StrToInt(TEdit(Sender).Text);
    if ((p < MIN_GETPOINTS{sic!}) or (p > MAX_POINTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_GETPOINTS) + ' and ' + IntToStr(MAX_POINTS), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(GetPoints);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> GetPoints then begin
    GetPoints := p;
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtGetTimeKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtGetTimeExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.txtGetTimeExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := Round((StrToFloat(TEdit(Sender).Text) * 1e-3) / SamplingTime);
    if ((p < MIN_GETPOINTS) or (p > MAX_POINTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_GETPOINTS * 1e3 * SamplingTime) + ' and ' + FloatToStr(MAX_POINTS * 1e3 * SamplingTime), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(GetPoints * 1e3 * SamplingTime);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> GetPoints then begin
    GetPoints := p;
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.chkUseGetListClick(Sender: TObject);
begin
  inherited;
  UseGetList := TCheckBox(Sender).Checked;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.lstGetListClick(Sender: TObject);
begin
  inherited;
  if lstGetList.ItemIndex <> -1 then begin
    GetListName := lstGetList.Items[lstGetList.ItemIndex];
  end else begin
    GetListName := '';
  end;
end;

procedure TQUB_QVO_QVD_Chop_PropertiesForm.chkDataSourceListClick(Sender: TObject);
begin
  QUB_MainForm.DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;


//Registry
procedure TQUB_QVO_QVD_Chop_PropertiesForm.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
begin
  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Chop Data', True) then begin
        if Read then begin
          prSkipBeforePoints := ReadRegistryInteger(Reg, 'Skip Before Points', DEF_POINTS);
          prSkipAfterPoints  := ReadRegistryInteger(Reg, 'Skip After Points', DEF_POINTS);
          prGetPoints        := ReadRegistryInteger(Reg, 'Get Points', DEF_GETPOINTS);
          prIgnoreSegments   := ReadRegistryBoolean(Reg, 'Ignore Segments', DEF_IGNORESEGMENTS);
          prOnce             := ReadRegistryBoolean(Reg, 'Once', DEF_ONCE);
          prDiscardPartial   := ReadRegistryBoolean(Reg, 'Discard Partial', DEF_DISCARDPARTIAL);
          prAddToList        := ReadRegistryBoolean(Reg, 'Add To List', DEF_ADDTOLIST);
          prOutputList       := ReadRegistryString(Reg,  'Output List', DEF_OUTPUTLIST);
          prUseGetList       := ReadRegistryBoolean(Reg, 'Use Get List', DEF_USEGETLIST);
        end else begin
          WriteInteger('Skip Before Points', SkipBeforePoints);
          WriteInteger('Skip After Points', SkipAfterPoints);
          WriteInteger('Get Points', GetPoints);
          WriteBool('Ignore Segments', IgnoreSegments);
          WriteBool('Once', Once);
          WriteBool('Discard Partial', DiscardPartial);
          WriteBool('Add To List', AddToList);
          WriteString('Output List', OutputList);
          WriteBool('Use Get List', UseGetList);
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;

function TQUB_QVO_QVD_Chop_PropertiesForm.PropertiesTree(cfg :INode): INode;
begin
  cfg                            := cfg.Clone(True); // don't modify what's passed in
  Result                         := cfg;
  { enable this if you're going to read/write times as well
  SamplingTime := GetOrSetNodeDouble( cfg['Sampling'], SamplingTime );
  }

  SkipBeforePoints  := GetOrSetNodeInteger( cfg['SkipBeforePoints'], SkipBeforePoints );
  GetPoints         := GetOrSetNodeInteger( cfg['GetPoints'], GetPoints );
  SkipAfterPoints   := GetOrSetNodeInteger( cfg['SkipAfterPoints'], SkipAfterPoints);
  IgnoreSegments    := GetOrSetNodeBoolean( cfg['IgnoreSegments'], IgnoreSegments);
  Once              := GetOrSetNodeBoolean( cfg['OncePerSegment'], Once);
  DiscardPartial    := GetOrSetNodeBoolean( cfg['DiscardPartial'], DiscardPartial);
  AddToList         := GetOrSetNodeBoolean( cfg['AddToList'], AddToList);
  OutputList        := GetOrSetNodeString( cfg['OutputList'], OutputList);
  DialogOnLClick    := GetOrSetNodeBoolean( cfg['DialogOnLClick'], DialogOnLClick);

end;


//******************************************************************************
end.
