{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// chop idl properties


unit QUB_QVO_QVD_EventProcessor_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Registry,
QUB_Task_Properties, ComCtrls, ExtCtrls;

{$include 'QUB_QVO_QVD_EventProcessor_Properties.hct'}

const


MIN_SEPARATIONTIME           = 1e-10;
MAX_SEPARATIONTIME           = 1e+10;
DEF_SEPARATIONTIME           = 10;

MIN_MIN_EVENTS               = 0;
MAX_MIN_EVENTS               = 1000000000000000;
DEF_MIN_EVENTS               = 3;

MIN_MIN_DURATION             = 0;
MAX_MIN_DURATION             = 1e+10;
DEF_MIN_DURATION             = 10;

DEF_OUTPUTLIST               = 'Processed events';


type


TQUB_QVO_QVD_EventProcessor_PropertiesForm = class(TQUB_Task_PropertiesForm)


  Label1                                   :TLabel;
  Label2                                   :TLabel;
  Label3                                   :TLabel;
  Label4                                   :TLabel;

  txtSeparationTime                        :TEdit;
  txtOutputList                            :TEdit;
  txtMinEvents                             :TEdit;
  txtMinDuration                           :TEdit;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure btnOKClick(Sender: TObject);

  procedure txtSeparationTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtSeparationTimeExit(Sender: TObject);
  procedure txtMinEventsExit(Sender: TObject);
  procedure txtMinEventsKeyPress(Sender: TObject; var Key: Char);
  procedure txtMinDurationExit(Sender: TObject);
  procedure txtMinDurationKeyPress(Sender: TObject; var Key: Char);


protected


  prSamplingTime                           :Extended;

  prSeparationTime                         :Double;
  prMinEvents                              :Integer;
  prMinDuration                            :Double;

  prOutputList                             :String;


  procedure prSetSeparationTime(t :Double);
  procedure prSetMinEvents(m :Integer);
  procedure prSetMinDuration(t :Double);
  procedure prSetOutputList(o :String);

  procedure prPropertiesRegistry(Read :Boolean);


public


  property SamplingTime          :Extended       read prSamplingTime      write prSamplingTime;

  property SeparationTime        :Double         read prSeparationTime    write prSetSeparationTime;
  property MinEvents             :Integer        read prMinEvents         write prSetMinEvents;
  property MinDuration           :Double         read prMinDuration       write prSetMinDuration;

  property OutputList            :String         read prOutputList        write prSetOutputList;


end;


implementation


{$R *.DFM}


//******************************************************************************
procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.prSetSeparationTime(t :Double);
begin
  prSeparationTime := t;
  txtSeparationTime.Text := Format('%.6g', [t * 1000]);
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.prSetMinEvents(m :Integer);
begin
  prMinEvents := m;
  txtMinEvents.Text := IntToStr(m);
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.prSetMinDuration(t :Double);
begin
  prMinDuration := t;
  txtMinDuration.Text := Format('%.6g', [t * 1000]);
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.prSetOutputList(o :String);
begin
  prOutputList := o;
  txtOutputList.Text := prOutputList;
end;

//******************************************************************************
procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(True{Read});
  {$include 'QUB_QVO_QVD_EventProcessor_Properties.htx'}
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.FormDestroy(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(False{Write});

end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;
  SeparationTime   := SeparationTime;
  MinEvents        := MinEvents;
  MinDuration      := MinDuration;
  OutputList       := OutputList;
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.btnOKClick(Sender: TObject);
begin
  OutputList := Trim(txtOutputList.Text);
  if OutputList = '' then OutputList := DEF_OUTPUTLIST;
  inherited;
end;

//Separation Time
procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.txtSeparationTimeKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSeparationTimeExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.txtSeparationTimeExit(Sender: TObject);
var
  t :Double;
begin
  inherited;
  try
    t := StrToFloat(TEdit(Sender).Text) / 1000;
    if ((t < MIN_SEPARATIONTIME) or (t > MAX_SEPARATIONTIME)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_SEPARATIONTIME) + ' and ' + FloatToStr(MAX_SEPARATIONTIME), mtError, [mbOK], 0);
    TEdit(Sender).Text := Format('%.6g', [SeparationTime * 1000]);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if t <> SeparationTime then begin
    SeparationTime := t;
  end;
end;

//******************************************************************************
procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.txtMinEventsKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtMinEventsExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.txtMinEventsExit(Sender: TObject);
var
  m :Integer;
begin
  inherited;
  try
    m := StrToInt(TEdit(Sender).Text);
    if ((m < MIN_MIN_EVENTS) or (m > MAX_MIN_EVENTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_MIN_EVENTS) + ' and ' + IntToStr(MAX_MIN_EVENTS), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(MinEvents);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if m <> MinEvents then begin
    MinEvents := m;
  end;
end;

//******************************************************************************
procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.txtMinDurationKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtMinDurationExit(Sender);
  end;
end;

procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.txtMinDurationExit(Sender: TObject);
var
  t :Double;
begin
  inherited;
  try
    t := StrToFloat(TEdit(Sender).Text) / 1000;
    if ((t < MIN_MIN_DURATION) or (t > MAX_MIN_DURATION)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_MIN_DURATION) + ' and ' + FloatToStr(MAX_MIN_DURATION), mtError, [mbOK], 0);
    TEdit(Sender).Text := Format('%.6g', [MinDuration * 1000]);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if t <> MinDuration then begin
    MinDuration := t;
  end;
end;

//Registry *********************************************************************
procedure TQUB_QVO_QVD_EventProcessor_PropertiesForm.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
begin
  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software', True) then begin
        if OpenKey('QUB', True) then begin
          if OpenKey('Event List Processor Data', True) then begin

            if Read then begin
              if ValueExists('Separation Time') then begin
                prSeparationTime := ReadFloat('Separation Time');
              end else begin
                prSeparationTime := DEF_SEPARATIONTIME;
              end;
            end else begin
              WriteFloat('Separation Time', SeparationTime);
            end;
            if Read then begin
              if ValueExists('Min Events') then begin
                prMinEvents := ReadInteger('Min Events');
              end else begin
                prMinEvents := DEF_MIN_EVENTS;
              end;
            end else begin
              WriteInteger('Min Events', MinEvents);
            end;
            if Read then begin
              if ValueExists('Min Duration') then begin
                prMinDuration := ReadFloat('Min Duration');
              end else begin
                prMinDuration := DEF_MIN_DURATION;
              end;
            end else begin
              WriteFloat('Min Duration', MinDuration);
            end;
            if Read then begin
              if ValueExists('Output List') then begin
                prOutputList := ReadString('Output List');
              end else begin
                prOutputList := DEF_OUTPUTLIST;
              end;
            end else begin
              WriteString('Output List', OutputList);
            end;

          end;
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
end.
