{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// selection lists, results list, and misc for QUB Visual Data


unit QUB_QVO_QVD_Global;


interface


uses


Windows, SysUtils, Classes, Messages, Graphics, Controls, StdCtrls, ExtCtrls, Forms, Menus, Clipbrd, Dialogs,

AllocTypes,
PythonEngine,

QUB_QFS_QDF_Global,
QUB_QVO, QUB_QVO_QVR,
QUB_Node,
QUB_Viewport_Python, QUB_Python_Thread,
QUB_Global, QUB_Events,
QUB_ObjectCache;


//******************************************************************************
type


TQUB_QVO_QVD_DataChannelDisplay = Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of TPAInteger;
TQUB_QVO_QVD_PDataChannelDisplay = ^TQUB_QVO_QVD_DataChannelDisplay;

TQUB_OnEditPointEvent           = procedure (Sender: TObject; Point :Integer; Change :Integer) of object;

//Draw Events --> hook-up and draw over
TQUB_OnDrawChannel              = procedure(Sender :TObject; Channel :Integer) of object;


TQUB_QVO_QVD_DrawMode           = (DRAWMODE_SLOW, DRAWMODE_NORM, DRAWMODE_FAST);




TQUB_QVO_QVD_Command            = (QUB_QVO_QVD_CMD_HOME,      QUB_QVO_QVD_CMD_END,        QUB_QVO_QVD_CMD_SEG_HOME,     QUB_QVO_QVD_CMD_SEG_END,
                                   QUB_QVO_QVD_CMD_PAGE_UP,   QUB_QVO_QVD_CMD_PAGE_DN,    QUB_QVO_QVD_CMD_HALF_PAGE_UP, QUB_QVO_QVD_CMD_HALF_PAGE_DN,
                                   QUB_QVO_QVD_CMD_TRACE_UP,  QUB_QVO_QVD_CMD_TRACE_DN,
                                   QUB_QVO_QVD_CMD_SEL_UP,    QUB_QVO_QVD_CMD_SEL_DN,
                                   QUB_QVO_QVD_CMD_SEL_UPSEG, QUB_QVO_QVD_CMD_SEL_DNSEG,
                                   QUB_QVO_QVD_CMD_GOTO_SEG,  QUB_QVO_QVD_CMD_GOTO_TRACE, QUB_QVO_QVD_CMD_GOTO_POINT, QUB_QVO_QVD_CMD_GOTO_PAGE,
                                   QUB_QVO_QVD_CMD_CHANGE_SEL,
                                   QUB_QVO_QVD_CMD_CHANGE_CURR
                                  );

//Selection List

TQUB_QVO_QVD_Selection          = class
protected


  prSel                         :TQUB_FirstLast;
  prLabel                       :String;
  prComment                     :String;
  prActive                      :Boolean;


  procedure prSetLabel(x :String);
  procedure prSetComment(x :String);
  procedure prSetActive(x :Boolean);


public


  OnLabelChanged                :TNotifyEvent;
  OnActiveChanged               :TNotifyEvent;


  constructor Create(First, Last :Integer; Lbl :String);


  property Sel                  :TQUB_FirstLast         read prSel         write prSel;
  property Lbl                  :String                 read prLabel       write prSetLabel;
  property Comment              :String                 read prComment     write prSetComment;
  property Active               :Boolean                read prActive      write prSetActive;


end;


TQUB_QVO_QVD_SelectionListIter  = class;

TQUB_QVO_QVD_SelectionList      = class(TListBox)
protected


  prActiveRange                :TQUB_FirstLast;

  prLockCount                  :Integer;
  prEvents                     :TQUB_Events;
  prEventIndex                 :Integer;

  prResultObject               :TQUB_QVO;

  prPrivate                    :Boolean;


  function prGetSelection(Index :Integer) :TQUB_QVO_QVD_Selection;
  function prGetCurrSelection :TQUB_QVO_QVD_Selection;
  procedure prSetActiveRange(x :TQUB_FirstLast);

  function prAddSelection(First, Last :Integer; SelName :String; InBunch :Boolean): TQUB_QVO_QVD_Selection;
  procedure prDelSelection(Sel :Integer);
  procedure prRenumber;

  function prGetLocked :Boolean;

  function prGetAsTree: INode;
  function prGetAsTreeSelected: INode;
  procedure prSetAsTree(x :INode);
  procedure prSetAsTreeAppend(x :INode);

  procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;

  procedure prOnKeyPress(Sender: TObject; var Key: Char);

  procedure prOnLabelChanged(Sender :TObject);
  procedure prOnActiveChanged(Sender :TObject);


public


  OnItemChange                :TNotifyEvent;
  OnItemDblClick              :TNotifyEvent;


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure Clear; {$IFDEF DELPHI6_OR_HIGHER} override; {$ENDIF}
  procedure Activate;
  procedure DeActivate;
  function AddSelection(First, Last :Integer; SelName :String; InBunch :Boolean =False): TQUB_QVO_QVD_Selection;
  procedure DelSelection(Sel :Integer);
  procedure DelSelected(KeyControl :Boolean);
  procedure Choose(Ix :Integer);
  function ShowSelectionProperties(Ix :Integer) :Boolean;

  function Find(First, Last :Integer; ActiveOnly :Boolean): TQUB_QVO_QVD_SelectionListIter;
  function FindAll(ActiveOnly :Boolean): TQUB_QVO_QVD_SelectionListIter;

  function IndexOf(First, Last :Integer): Integer;
  function IndexAtOrAfter(First :Integer): Integer;

  function SaveToFile(const FileName :String): Boolean;
  function LoadFromFile(const FileName :String): Boolean;

  procedure Lock;
  procedure UnLock;


  property Events                    :TQUB_Events            read prEvents;
  property EventIndex                :Integer                read prEventIndex; // index of affected selection, during appropriate event handlers e.g. ActiveChanged
  property Locked                    :Boolean                read prGetLocked;

  property Selection[Index :Integer] :TQUB_QVO_QVD_Selection read prGetSelection; default;
  property CurrSelection             :TQUB_QVO_QVD_Selection read prGetCurrSelection;

  property ActiveRange               :TQUB_FirstLast         read prActiveRange    write prSetActiveRange;

  property ResultObject              :TQUB_QVO               read prResultObject   write prResultObject;

  property AsTree                    :INode                  read prGetAsTree      write prSetAsTree;
  property AsTreeAppend              :INode                                        write prSetAsTreeAppend;
  property AsTreeSelected            :INode                  read prGetAsTreeSelected;

  property Private                   :Boolean                read prPrivate        write prPrivate;


published


  property OnMouseDown;


end;

// SelectionList Events:
//   ItemChanged
//   ItemLabelChanged
//   ItemActiveChanged
//   ActiveRangeChanged
//   ItemAdded
//   ItemDeleting
//   ItemsChanged       -- one or more added or deleted -- in addition to the above 2 events
//   ItemDblClicked


TQUB_QVO_QVD_SelectionListIter                 = class
protected
  prList                             :TQUB_QVO_QVD_SelectionList;
  prBounds                           :TQUB_FirstLast;
  prIndexRange                       :TQUB_FirstLast;

  prIndex                            :Integer;
  prSel                              :TQUB_FirstLast;
  prDone                             :Boolean;
  prActiveOnly                       :Boolean; // limit to active sels (it always honors ActiveRange)

  function prGetFullSel: TQUB_FirstLast;
  function prGetLbl: String;
  function prGetComment: String;

public
  constructor Create(List :TQUB_QVO_QVD_SelectionList; First, Last :Integer; ActiveOnly :Boolean); // -1, -1 for all

  function Next: Boolean; // returns Done

  property Done                      :Boolean             read prDone;
  property Ix                        :Integer             read prIndex;
  property Sel                       :TQUB_FirstLast      read prSel;
  property FullSel                   :TQUB_FirstLast      read prGetFullSel;
  property Lbl                       :String              read prGetLbl;
  property Comment                   :String              read prGetComment;
end;


TQUB_QVO_QVD_SelectionLists                    = class(TListBox)
protected


  prOnItemChange              :TNotifyEvent;
  prOnItemDblClick            :TNotifyEvent;

  prSaveItems                 :TStringList;
  prSaveTopIndex              :Integer;
  prSaveItemIndex             :Integer;


  function prGetList(Index :Integer) :TQUB_QVO_QVD_SelectionList;
  function prGetListByName(Index :String) :TQUB_QVO_QVD_SelectionList;

  procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;

  procedure prOnKeyPress(Sender: TObject; var Key: Char);

  procedure CreateWnd; override;
  procedure DestroyWnd; override;


public


  OnListAdded                 :procedure (Sender :TObject; Index :Integer) of object;
  OnListDeleting              :procedure (Sender :TObject; Index :Integer) of object;
  OnListRenamed               :procedure (Sender :TObject; Index :Integer) of object;


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure Clear; {$IFDEF DELPHI6_OR_HIGHER} override; {$ENDIF}
  procedure Activate;
  procedure DeActivate;
  procedure AddList(List :TQUB_QVO_QVD_SelectionList; const ListName :String);
  procedure DelList(List :Integer);
  procedure DelSelLists;
  procedure RenameList(List :Integer; ListName :String);


  property List[Index :Integer]      :TQUB_QVO_QVD_SelectionList read prGetList; default;
  property ListByName[Index :String] :TQUB_QVO_QVD_SelectionList read prGetListByName;

  property OnItemChange              :TNotifyEvent           read prOnItemChange   write prOnItemChange;
  property OnItemDblClick            :TNotifyEvent           read prOnItemDblClick write prOnItemDblClick;


end;


TQUB_QVO_QVD_ResultsRec                      = class
public
  Node                        :INode;
  constructor Create(ResultsNode :INode);
end;

TQUB_QVO_QVD_ResultsList                     = class(TListBox)
protected

  prPopup                     :TPopupMenu;
  prItemRename                :TMenuItem;
  prItemHold                  :TMenuItem;
  prItemDelete                :TMenuItem;

  prOnItemChange              :TNotifyEvent;
  prOnItemDblClick            :TNotifyEvent;

  prQVRs                      :TQUB_ObjectCache;

  function prGetCurrResults: TQUB_QVO_QVR;
  function prGetResults(Index :Integer) :TQUB_QVO_QVR;
  function prGetResultsNode(Index :Integer) :INode;

  procedure prRemoveOldResults;

  function prCreateQVR(Key :Pointer): Pointer;
  procedure prDestroyQVR(Key, Obj :Pointer);

  procedure prAddBuiltResults(Sender :TObject);

  procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;

  procedure prOnKeyPress(Sender: TObject; var Key: Char);
  procedure prOnDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);

  procedure prOnPopup(Sender :TObject);
  procedure prOnClickRename(Sender :TObject);
  procedure prOnClickHold(Sender :TObject);
  procedure prOnClickDelete(Sender :TObject);


public


  DataImageList               :TImageList;


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure Activate;
  procedure DeActivate;
  procedure AddResults(Results :INode);
  procedure AddResultsBehind(Results :INode); // don't select -- also assumes Date and Index are present
  procedure DelResults(Ix :Integer);
  procedure DelSelected;
  procedure ChooseResults(Ix :Integer);
  procedure Clear; {$IFDEF DELPHI6_OR_HIGHER} override; {$ENDIF}

  property CurrResults                  :TQUB_QVO_QVR read prGetCurrResults;
  property Results[Index :Integer]      :TQUB_QVO_QVR read prGetResults; default;
  property ResultsNode[Index :Integer]  :INode        read prGetResultsNode;

  property OnItemChange              :TNotifyEvent           read prOnItemChange   write prOnItemChange;
  property OnItemDblClick            :TNotifyEvent           read prOnItemDblClick write prOnItemDblClick;

end;


//******************************************************************************
const


PRINT_PEN_WIDTH                             :Integer                            = 3;

QUB_QVO_QVD_PROCESS_NONE                    :Byte                               = 0;
QUB_QVO_QVD_PROCESS_SET_FILTER_ON           :Byte                               = 1;
QUB_QVO_QVD_PROCESS_SET_FILTER_OFF          :Byte                               = 254;
QUB_QVO_QVD_PROCESS_SET_BASELINE_ON         :Byte                               = 2;
QUB_QVO_QVD_PROCESS_SET_BASELINE_OFF        :Byte                               = 253;
QUB_QVO_QVD_PROCESS_GET_FILTER              :Byte                               = 1;
QUB_QVO_QVD_PROCESS_GET_BASELINE            :Byte                               = 2;

QUB_QVO_QVD_TraceVMargin                    :Integer                            = 4;
QUB_QVO_QVD_TraceHMargin                    :Integer                            = 4;

QUB_QVO_QVD_DEF_COLOR                       :TColor                             = clBtnFace;
QUB_QVO_QVD_DEF_COLORBACK                   :TColor                             = TColor($fafafa); // clNavy;
QUB_QVO_QVD_DEF_COLORDATA                   :Array[0..7] of TColor              = ({clWhite}clBlack, TColor($3399ff), clLime, clMaroon, clOlive, clTeal, clBlue, clFuchsia);
QUB_QVO_QVD_DEF_COLORSEL                    :TColor                             = clBlue; // clRed;
QUB_QVO_QVD_DEF_COLORBREAK                  :TColor                             = clBlue;
QUB_QVO_QVD_DEF_COLORFRAME                  :TColor                             = clWhite;
QUB_QVO_QVD_DEF_COLORBASELINE               :TColor                             = clPurple; // clLime;
QUB_QVO_QVD_DEF_COLORBASELINENODES          :TColor                             = clFuchsia;
QUB_QVO_QVD_DEF_COLORIDEALIZED              :TColor                             = clRed;
QUB_QVO_QVD_DEF_COLORIDEALIZEDSEL           :TColor                             = clBlue;
QUB_QVO_QVD_DEF_COLORCURR                   :TColor                             = clRed;
QUB_QVO_QVD_DEF_COLORSELLIST                :TColor                             = TColor($e2d2bc); // $d5bc9b);
QUB_QVO_QVD_DEF_COLORPARAMETERS             :TColor                             = clLime;
QUB_QVO_QVD_DEF_COLORSEGLABEL               :TColor                             = clGray;
QUB_QVO_QVD_DEF_COLORTEXT                   :TColor                             = clBtnText;

QUB_QVO_QVD_DEF_TRACESPERPAGE               :Integer                            = 10;
QUB_QVO_QVD_MIN_TRACESPERPAGE               :Integer                            = 1;
QUB_QVO_QVD_MAX_TRACESPERPAGE               :Integer                            = 1000;

QUB_QVO_QVD_DEF_POINTSPERTRACE              :Integer                            = 100000;
QUB_QVO_QVD_MIN_POINTSPERTRACE              :Integer                            = 2;
QUB_QVO_QVD_MAX_POINTSPERTRACE              :Integer                            = 1000000;

QUB_QVO_QVD_DEF_CACHESIZE                   :Integer                            =   4 * ONE_MEGABYTE;
QUB_QVO_QVD_MIN_CACHESIZE                   :Integer                            =   1 * ONE_MEGABYTE;
QUB_QVO_QVD_MAX_CACHESIZE                   :Integer                            = 256 * ONE_MEGABYTE;


//Scaling
QUB_QVO_QVD_DEF_SCALING                     :Double                             =   320.0;
QUB_QVO_QVD_MIN_SCALING                     :Double                             =     1.0e-10;
QUB_QVO_QVD_MAX_SCALING                     :Double                             =     1.0e16;

//Sampling, kHz
QUB_QVO_QVD_DEF_SAMPLING                    :Double                             =   20.0;
QUB_QVO_QVD_MIN_SAMPLING                    :Double                             =   1e-6;
QUB_QVO_QVD_MAX_SAMPLING                    :Double                             =   1e+6;

// PRT import
QUB_QVO_QVD_DEF_PRT_SAMPLING                :Double                             = 0.001;
QUB_QVO_QVD_DEF_PRT_USE_ACRIT               :Integer                            = 0;
QUB_QVO_QVD_DEF_PRT_ACRIT                   :String                             = '-0.5';

QUB_QVO_QVD_DEF_TRACECHANGEONBREAK          :Boolean                            = True;
QUB_QVO_QVD_DEF_OVERLAPCHANNELS             :Boolean                            = False;
QUB_QVO_QVD_DEF_OVERLAPTRACES               :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWPOINTS                  :Boolean                            = True;
QUB_QVO_QVD_DEF_DRAWINVERTED                :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWACTIVETHICK             :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWRASTER                  :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWLABELS                  :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWBASELINE                :Boolean                            = True;
QUB_QVO_QVD_DEF_DRAWBASELINENODES           :Boolean                            = True;
QUB_QVO_QVD_DEF_DRAWSEGLABELS               :Boolean                            = True;
QUB_QVO_QVD_DEF_DRAWIDEALIZED               :Boolean                            = True;
QUB_QVO_QVD_DEF_DRAWIDEALIZEDABOVE          :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWIDEALIZEDCOLOR          :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWLISTOVERLAPPED          :Boolean                            = False;
QUB_QVO_QVD_DEF_DRAWSELLIST                 :Boolean                            = True;
QUB_QVO_QVD_DEF_DRAWPARAMETERS              :Boolean                            = True;

QUB_QVO_QVD_DEF_AUTOSCALE                   :Boolean                            = True;
QUB_QVO_QVD_DEF_DATAYMIN                    :Integer                            = -10000;
QUB_QVO_QVD_DEF_DATAYMAX                    :Integer                            = +10000;
QUB_QVO_QVD_DEF_DATAYFACTOR                 :Double                             = 1.0;
QUB_QVO_QVD_MIN_DATAY                       :Integer                            = Low(Integer);
QUB_QVO_QVD_MAX_DATAY                       :Integer                            = High(Integer);
QUB_QVO_QVD_MIN_DATAYREAL                   :Double                             = -1e100;
QUB_QVO_QVD_MAX_DATAYREAL                   :Double                             = 1e100;

QUB_QVO_QVD_DEF_AUTOSCROLL                  :Boolean                            = True;

QUB_QVO_QVD_DEF_CHANNELVISIBLE              :Boolean                            = True;

QUB_QVO_QVD_DEF_FILTERON                    :Boolean                            = False;
QUB_QVO_QVD_DEF_FILTERMAON                  :Boolean                            = False;

QUB_QVO_QVD_DEF_FILTERFREQ                  :Double                             = 5000;
QUB_QVO_QVD_MIN_FILTERFREQ                  :Double                             = 0.001;
QUB_QVO_QVD_MAX_FILTERFREQ                  :Double                             = 1000000;

QUB_QVO_QVD_DEF_FILTERMACOUNT               :Integer                            = 11;
QUB_QVO_QVD_MIN_FILTERMACOUNT               :Integer                            = 3;
QUB_QVO_QVD_MAX_FILTERMACOUNT               :Integer                            = 511;

QUB_QVO_QVD_DEF_BASELINEON                  :Boolean                            = False;
QUB_QVO_QVD_DEF_WARNONBASELINE              :Boolean                            = True; // except channel 0 set programatically

QUB_QVO_QVD_MIN_AXIS_TICKCOUNT              :Integer                            = 1;
QUB_QVO_QVD_MAX_AXIS_TICKCOUNT              :Integer                            = 100;
QUB_QVO_QVD_DEF_AXIS_TICKCOUNT              :Integer                            = 5;

QUB_QVO_QVD_DEF_DRAWITOG                    :Boolean                            = False;

QUB_QVO_QVD_DEF_DRAWMODE                    :TQUB_QVO_QVD_DrawMode              = DRAWMODE_NORM;

QUB_QVO_QVD_DEF_MARGINLEFT                  :Integer                            = 10;
QUB_QVO_QVD_DEF_MARGINTOP                   :Integer                            = 10;
QUB_QVO_QVD_DEF_MARGINRIGHT                 :Integer                            = 10;
QUB_QVO_QVD_DEF_MARGINBOTTOM                :Integer                            = 10;

QUB_QVO_QVD_MIN_MARGIN                      :Integer                            = 0;
QUB_QVO_QVD_MAX_MARGIN                      :Integer                            = 50;
QUB_QVO_QVD_DEF_MARGIN                      :Integer                            = 10;

QUB_QVO_QVD_DEF_SHOWMINMAX                  :Boolean                            = True;

QUB_QVO_QVD_DEF_DRAWRMS                     :Boolean                            = False;

QUB_QVO_QVD_MOVE_UP                                                             = -1;
QUB_QVO_QVD_MOVE_NO                                                             =  0;
QUB_QVO_QVD_MOVE_DN                                                             = +1;

QUB_QVO_QVD_DEF_TIPSON                      :Boolean                            = True;

QUB_QVO_QVD_QVR_CACHESIZE                   :Integer                            = 4;




QUB_QVO_QVD_RESULTS_MAXCOUNT                :Integer                            = 6;



implementation


uses


Math, Scanf,
QUB_Console, QUB_Rich_Console,
QUB_GUIInterface,
QUB_Viewport_Results,
QUB_QVO_QVD_SelectionProperties,
QUB_Main;


//****************************************************************************//
//                                                                            //
//                          Selection   L I S T S                             //
//                                                                            //
//****************************************************************************//
constructor TQUB_QVO_QVD_SelectionLists.Create(AOwner :TComponent);
begin
  inherited;
  ShowHint    := True;
  Align       := alClient;
  MultiSelect := True;
  OnKeyPress  := prOnKeyPress;
end;

destructor TQUB_QVO_QVD_SelectionLists.Destroy;
begin
  try
    prSaveItems.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.CreateWnd;
begin
  inherited;
  if prSaveItems <> nil then begin
    Items.Assign(prSaveItems);
    TopIndex  := prSaveTopIndex;
    ItemIndex := prSaveItemIndex;
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.DestroyWnd;
begin
  if Items.Count > 0 then begin
    prSaveItems := TStringList.Create;
    prSaveItems.Assign(Items);
    prSaveTopIndex  := TopIndex;
    prSaveItemIndex := ItemIndex;
  end;
  inherited;
end;

procedure TQUB_QVO_QVD_SelectionLists.Clear;
var
  i :Integer;
begin
  if Parent <> nil then begin
    Items.BeginUpdate;
    try
      for i := Items.Count - 1 downto 0 do begin
        DelList(i);
      end;
      if Items.Count = 0 then ItemIndex := -1;
    finally
      Items.EndUpdate;
    end;
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.CNCommand(var Message: TWMCommand);
var
  PrevItemIndex       :Integer;
  FirstSelected       :Integer;
  i                   :Integer;
begin
  case Message.NotifyCode of
    LBN_SELCHANGE:
      begin
        PrevItemIndex := ItemIndex;
        if ItemIndex <> -1 then begin
          List[ItemIndex].DeActivate; //?!
        end;

        inherited;

        FirstSelected := -1;
        for i := 0 to Items.Count - 1 do begin
          if Selected[i] then begin
            FirstSelected := i;
            Break;
          end;
        end;

        // the chosen list is the first of multi-selection, or the most recently selected
        if FirstSelected = -1 then begin
          if PrevItemIndex <> -1 then begin
            ItemIndex := PrevItemIndex;
            Selected[ItemIndex] := True;
          end;
        end else {if (ItemIndex = -1) or not Selected[ ItemIndex ] then} begin
          ItemIndex := FirstSelected;
        end;

        if ItemIndex <> -1 then begin
          List[ItemIndex].Activate;
        end;
        //Generate OnItemChange Event
        if Assigned(prOnItemChange) then begin
          prOnItemChange(Self);
        end;
      end;
    LBN_DBLCLK:
      begin
        inherited;
        //Generate OnItemDblClick Event
        if Assigned(prOnItemDblClick) then begin
          prOnItemDblClick(Self);
        end;
      end;
  else
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.prOnKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    prOnItemDblClick(Self);
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.Activate;
begin
  if Items.Count <> 0 then begin
    if ItemIndex <> -1 then begin
       List[ItemIndex].Activate;
    end;
  end;
  if Parent <> nil then begin
    BringToFront;
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.DeActivate;
begin
  if Items.Count <> 0 then begin
    if ItemIndex <> -1 then begin
      List[ItemIndex].DeActivate;
    end;
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.AddList(List :TQUB_QVO_QVD_SelectionList; const ListName :String);
begin
  Items.AddObject(ListName, List);
  {
  ItemIndex := Items.Count - 1;
  for i := 0 to Items.Count - 2 do begin
    Selected[i] := False;
  end;
  Selected[ Items.Count - 1 ] := True;
  }
  if Assigned(OnListAdded) then OnListAdded(Self, Items.Count - 1);
  {
  //Generate OnItemChange Event
  if Assigned(prOnItemChange) then begin
    prOnItemChange(Self);
  end;
  }
end;

procedure TQUB_QVO_QVD_SelectionLists.DelList(List :Integer);
var
  OldName  :String;
  //OldIndex :Integer;
  i        :Integer;
begin
  OldName  := '';
  if ItemIndex >= 0 then OldName := Items.Strings[ ItemIndex ];

  if Self.List[List].Locked then Exit;

  if Assigned(OnListDeleting) then OnListDeleting(Self, List);
  Self.List[List].Free;
  Items.Delete(List);

  if (Items.Count > 0) and (OldName <> '') then begin
    for i := 0 to Items.Count - 1 do begin
      if Items.Strings[i] = OldName then begin
        ItemIndex := i;
        Break;
      end;
    end;
  end else begin
    ItemIndex := -1;
  end;
  {
  OldIndex := ItemIndex;
  if Self.List[List].Locked then Exit;
  Self.List[List].Free;
  Items.Delete(List);
  if Items.Count > 0 then begin
    if OldIndex > (Items.Count - 1) then OldIndex := (Items.Count - 1);
    if ItemIndex <> OldIndex then ItemIndex := OldIndex;
  end;
  if Items.Count = 0 then ItemIndex := -1;
  }
  for i := 0 to Items.Count - 1 do begin
    Selected[i] := False;
  end;
  if (ItemIndex >= 0) and (ItemIndex < Items.Count) then begin
    Selected[ ItemIndex ] := True;
  end;

  //Generate OnItemChange Event
  if Assigned(prOnItemChange) then begin
    prOnItemChange(Self);
  end;
end;

//Delete all selected lists
procedure TQUB_QVO_QVD_SelectionLists.DelSelLists;
var
  OldName  :String;
  i        :Integer;
begin
  OldName  := '';
  if ItemIndex >= 0 then OldName := Items.Strings[ItemIndex];

  for i := Items.Count - 1 downto 0 do begin
    if Selected[i] then begin
      if List[i].Locked then Continue;
      if Assigned(OnListDeleting) then OnListDeleting(Self, i);
      List[i].Free;
      Items.Delete(i);
    end;
  end;

  i := -1;
  if (Items.Count > 0) and (OldName <> '') then begin
    i := Items.IndexOf(OldName);
  end;
  ItemIndex := i;
  if ItemIndex >= 0 then Selected[ItemIndex] := True;

  //Generate OnItemChange Event
  if Assigned(prOnItemChange) then begin
    prOnItemChange(Self);
  end;
end;

procedure TQUB_QVO_QVD_SelectionLists.RenameList(List :Integer; ListName :String);
begin
  Items.Strings[List] := ListName;
  if Assigned(OnListRenamed) then OnListRenamed(Self, List);
end;

function TQUB_QVO_QVD_SelectionLists.prGetList(Index :Integer) :TQUB_QVO_QVD_SelectionList;
begin
  if (Index >= 0) and (Index < Items.Count) then Result := TQUB_QVO_QVD_SelectionList(Items.Objects[Index]) else Result := nil;
end;

function TQUB_QVO_QVD_SelectionLists.prGetListByName(Index :String) :TQUB_QVO_QVD_SelectionList;
var
  Idx :Integer;
begin
  Idx := Items.IndexOf(Index);
  if Idx = -1 then Result := nil else Result := List[Idx];
end;



//****************************************************************************//
//                                                                            //
//                          Selection   L I S T                               //
//                                                                            //
//****************************************************************************//
constructor TQUB_QVO_QVD_SelectionList.Create(AOwner :TComponent);
begin
  inherited;
  ShowHint       := True;
  Align          := alClient;
  MultiSelect    := True;
  OnKeyPress     := prOnKeyPress;
  prActiveRange  := QUB_FirstLast(-1, -1);
  prEvents       := TQUB_Events.Create;
end;

destructor TQUB_QVO_QVD_SelectionList.Destroy;
begin
  try
    prEvents.Free;
    prEvents := nil;
    Clear;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_SelectionList.Activate;
begin
  BringToFront;
end;

procedure TQUB_QVO_QVD_SelectionList.DeActivate;
begin
end;

function TQUB_QVO_QVD_SelectionList.prGetLocked :Boolean;
begin
  Result := prLockCount > 0;
end;

procedure TQUB_QVO_QVD_SelectionList.Lock;
begin
  Inc(prLockCount);
end;

procedure TQUB_QVO_QVD_SelectionList.UnLock;
begin
  Dec(prLockCount);
  if prLockCount < 0 then prLockCount := 0;
end;

procedure TQUB_QVO_QVD_SelectionList.CNCommand(var Message: TWMCommand);
begin
  case Message.NotifyCode of
    LBN_SELCHANGE:
      begin
        inherited;
        Events.RaiseNotifyEvent('ItemChanged', Self);
      end;
    LBN_DBLCLK:
      begin
        inherited;
        Events.RaiseNotifyEvent('ItemDblClicked', Self);
      end;
  else
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_SelectionList.prOnKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;

  if Key = #13 then begin
    Key := #0;
    Events.RaiseNotifyEvent('ItemDblClicked', Self);
  end;
end;

//----- Add selection item to list, sorted by start of selection
function TQUB_QVO_QVD_SelectionList.prAddSelection(First, Last :Integer; SelName :String; InBunch :Boolean): TQUB_QVO_QVD_Selection;
var
  s :TQUB_QVO_QVD_Selection;
  p :Integer;
begin
  Result := nil;
  if Locked then Exit;

  //----- Set P = point in list to insert this selection
  //p := -1;
  if Items.Count<1 then begin
  //Empty list - place at 0.
    p := 0;
  end else if First > Selection[Items.Count - 1].Sel.First then begin
  //Check if new seg is > last seg so far
    p := Items.Count;
  end else begin
    p := IndexAtOrAfter(First);
  end;

  //----- Create new selection list item to be added : s
  s := TQUB_QVO_QVD_Selection.Create(First, Last, SelName);
  s.OnLabelChanged  := prOnLabelChanged;
  s.OnActiveChanged := prOnActiveChanged;
  if SelName <> '' then SelName := SelName + ': ';
  SelName := SelName + Format('%d - %d', [First, Last]);
  Result := s;

  //----- add to list and make current
  Items.InsertObject(p, SelName, s);
  ItemIndex := p;

  if not InBunch then prRenumber;

  prEventIndex := p;
  Events.RaiseNotifyEvent('ItemAdded', Self);
end;

function TQUB_QVO_QVD_SelectionList.AddSelection(First, Last :Integer; SelName :String; InBunch :Boolean =False): TQUB_QVO_QVD_Selection;
begin
  Result := prAddSelection(First, Last, SelName, InBunch);
  Events.RaiseNotifyEvent('ItemsChanged', Self);
  Events.RaiseNotifyEvent('ItemChanged', Self);
end;

procedure TQUB_QVO_QVD_SelectionList.prDelSelection(Sel :Integer);
var
  OldIndex :Integer;
begin
  if Locked then Exit;

  OldIndex := ItemIndex;
  prEventIndex := Sel;
  if Events <> nil then Events.RaiseNotifyEvent('ItemDeleting', Self);
  TQUB_QVO_QVD_Selection(Items.Objects[Sel]).Free;
  Items.Delete(Sel);
  if Items.Count > 0 then begin
    if OldIndex > (Items.Count - 1) then OldIndex := (Items.Count - 1);
    if ItemIndex <> OldIndex then ItemIndex := OldIndex;
  end;
  if Items.Count = 0 then ItemIndex := -1;

  if Sel < Items.Count then prRenumber;
end;

procedure TQUB_QVO_QVD_SelectionList.prRenumber;
var
  i :Integer;
  s :Boolean;
begin
  for i := 0 to Items.Count - 1 do begin
    if Selection[i].Lbl = '' then begin
      s := Selected[i];
      Items[i] := Format('%d: %d - %d', [i+1, Selection[i].Sel.First, Selection[i].Sel.Last]);
      Selected[i] := s;
    end;
  end;
end;

procedure TQUB_QVO_QVD_SelectionList.DelSelection(Sel :Integer);
begin
  prDelSelection(Sel);
  if Events <> nil then begin
    Events.RaiseNotifyEvent('ItemsChanged', Self);
    Events.RaiseNotifyEvent('ItemChanged', Self);
  end;
end;

procedure TQUB_QVO_QVD_SelectionList.DelSelected(KeyControl :Boolean);
var
  OldIndex :Integer;
  s        :Integer;
begin
  if Locked then Exit;

  OldIndex := ItemIndex;

  for s := Items.Count - 1 downto 0 do begin
    if ((Selected[s]) and (not KeyControl)) or
       ((not Selected[s]) and (KeyControl)) then begin
      prEventIndex := s;
      Events.RaiseNotifyEvent('ItemDeleting', Self);
      TQUB_QVO_QVD_Selection(Items.Objects[s]).Free;
      Items.Delete(s);
    end;
  end;

  if Items.Count > 0 then begin
    if OldIndex > (Items.Count - 1) then OldIndex := (Items.Count - 1);
    if ItemIndex <> OldIndex then ItemIndex := OldIndex;
  end;
  if Items.Count = 0 then ItemIndex := -1;

  prRenumber;

  Events.RaiseNotifyEvent('ItemsChanged', Self);
  Events.RaiseNotifyEvent('ItemChanged', Self);
end;

procedure TQUB_QVO_QVD_SelectionList.Clear;
begin
  if Locked then Exit;

  if Parent <> nil then begin
    Items.BeginUpdate;
    try
      while Items.Count > 0 do begin
        DelSelection(Items.Count - 1);
      end;
      if Items.Count = 0 then ItemIndex := -1;
    finally
      Items.EndUpdate;
    end;
    if Events <> nil then begin
      Events.RaiseNotifyEvent('ItemsChanged', Self);
      Events.RaiseNotifyEvent('ItemChanged', Self);
    end;
  end;
end;

procedure TQUB_QVO_QVD_SelectionList.Choose(Ix :Integer);
var
  i            :Integer;
begin
  for i := 0 to Items.Count - 1 do begin
    Selected[i] := False;
  end;

  if Ix >= 0 then Selected[Ix] := True;

  // ??
  ItemIndex := Ix;
  Events.RaiseNotifyEvent('ItemChanged', Self);
end;

function TQUB_QVO_QVD_SelectionList.ShowSelectionProperties(Ix :Integer) :Boolean;
var
  Sel :TQUB_QVO_QVD_Selection;
begin
  Result := False;

  if Ix < 0 then Exit;

  Sel := Selection[Ix];
  with SelectionPropertiesForm do begin
    //lblSelection.Caption := Format('Selection: %d - %d', [Sel.Sel.First, Sel.Sel.Last]);
    txtSelFirst.Text     := IntToStr(Sel.Sel.First);
    txtSelLast.Text      := IntToStr(Sel.Sel.Last);
    lstLabel.ItemIndex   := 0;
    lstLabel.Text        := Sel.Lbl;
    txtComment.Text      := Sel.Comment;
    if ShowModal = mrOK then begin
      try
        Sel.Sel          := QUB_FirstLast(StrToInt(txtSelFirst.Text), StrToInt(txtSelLast.Text));
      except
      end;
      Sel.Lbl            := lstLabel.Text;
      Sel.Comment        := txtComment.Text;
      Result             := True;
    end else begin
      Result             := False
    end;
  end;
end;

function TQUB_QVO_QVD_SelectionList.prGetSelection(Index :Integer) :TQUB_QVO_QVD_Selection;
begin
  if (Index >= 0) and (Index < Items.Count) then Result := TQUB_QVO_QVD_Selection(Items.Objects[Index]) else Result := nil;
end;

function TQUB_QVO_QVD_SelectionList.prGetCurrSelection :TQUB_QVO_QVD_Selection;
begin
  if (Items.Count > 0) and (ItemIndex <> -1) then Result := TQUB_QVO_QVD_Selection(Items.Objects[ItemIndex]) else Result := nil;
end;

procedure TQUB_QVO_QVD_SelectionList.prSetActiveRange(x :TQUB_FirstLast);
begin
  prActiveRange := x;
  Events.RaiseNotifyEvent('ActiveRangeChanged', Self);
end;

procedure TQUB_QVO_QVD_SelectionList.prOnLabelChanged(Sender :TObject);
var
  i         :Integer;
  Lbl       :String;
  Sel       :TQUB_QVO_QVD_Selection;
begin
  Sel := TQUB_QVO_QVD_Selection(Sender);

  for i := 0 to Items.Count - 1 do begin
    if Selection[i] = Sel then begin
      Lbl   := Sel.Lbl;
      if Lbl = '' then Lbl := IntToStr(i+1);
      if Lbl <> '' then Lbl := Lbl + ': ';
      Lbl := Lbl + Format('%d - %d', [Sel.Sel.First, Sel.Sel.Last]);
      Items[i] := Lbl;
      prEventIndex := i;
      Events.RaiseNotifyEvent('LabelChanged', Self);
      Break;
    end;
  end;
end;

procedure TQUB_QVO_QVD_SelectionList.prOnActiveChanged(Sender :TObject);
var
  i         :Integer;
begin
  for i := 0 to Items.Count - 1 do begin
    if Selection[i] = Sender then begin
      prEventIndex := i;
      Events.RaiseNotifyEvent('ActiveChanged', Self);
      Break;
    end;
  end;
end;

function TQUB_QVO_QVD_SelectionList.Find(First, Last :Integer; ActiveOnly :Boolean): TQUB_QVO_QVD_SelectionListIter;
begin
  Result := TQUB_QVO_QVD_SelectionListIter.Create(Self, First, Last, ActiveOnly);
end;

function TQUB_QVO_QVD_SelectionList.FindAll(ActiveOnly :Boolean): TQUB_QVO_QVD_SelectionListIter;
begin
  Result := TQUB_QVO_QVD_SelectionListIter.Create(Self, -1, -1, ActiveOnly);
end;

// Selections are sorted by Sel.First; Find insertion point in list for (First).
function TQUB_QVO_QVD_SelectionList.IndexAtOrAfter(First :Integer): Integer;
var
  i, N, Step   :Integer;
  NotLT        :Boolean;
begin
  N            := Items.Count;
  Result       := N;

  if Result > 0 then begin
    i          := 0;
    Step       := N;
    while (i >= 0) and (i < N) do begin
      NotLT    := Selection[i].Sel.First >= First;
      if NotLT then Result := i;

      if Step = 1 then Break;
      Step     := (Step div 2) + (Step mod 2);
      if NotLT then Dec(i, Step) else Inc(i, Step);
    end;
  end;
end;

function TQUB_QVO_QVD_SelectionList.IndexOf(First, Last :Integer): Integer;
var
  i           :Integer;
begin
  Result      := -1;

  for i := IndexAtOrAfter(First) to Items.Count - 1 do begin
    if Selection[i].Sel.First <> First then Break;
    if Selection[i].Sel.Last = Last then begin
      Result  := i;
      Break;
    end;
  end;
end;

function TQUB_QVO_QVD_SelectionList.SaveToFile(const FileName :String): Boolean;
var
  f                     :TextFile;
  FileOpened            :Boolean;
  i                     :Integer;
  Sel                   :TQUB_QVO_QVD_Selection;
begin
  Result                := True;
  FileOpened            := False;

  try

    try
      AssignFile(f, FileName);
      ReWrite(f);
      FileOpened := True;
    except
      Console.NewEntry( 'Cannot open output file.'#13#10'Medium may be read-only or a file with the same name is in use', COLOR_TEXT);
      Exit;
    end;

    for i := 0 to Items.Count - 1 do begin
      Sel              := Selection[i];
      Write(f, Format('%d - %d'#13#10, [Sel.Sel.First, Sel.Sel.Last]));
    end;

    Result              := False;

  finally

    if FileOpened then begin
      try
        CloseFile(f);
      except
      end;
    end;

  end;
end;

function TQUB_QVO_QVD_SelectionList.LoadFromFile(const FileName :String): Boolean;
var
  Strings          :TStringList;
  i                :Integer;
  F, L             :Integer;
  iPos : Integer;
begin
  Strings          := TStringList.Create;
  Result           := True;
  try
    Strings.LoadFromFile(FileName);
    if Strings.Count = 0 then Exit;

    for i := 0 to Strings.Count - 1 do begin
      { SCANF_C has some unusual asm code which Delphi 2005 does not
      accept.  Since this is the only place it is used, it is easier
      to remove the module and write some extra code here than to fix
      the module.  [JB] 2/2005

      Sscanf(PChar(Strings[i]), '%d - %d', [@F, @L]);
      }
      F            := -1;
      L            := -1;
      iPos := Pos(' - ', Strings[i]);
      if iPos > 0 then begin
        F := ValidStrToInt(Copy(Strings[i], 1,        iPos - 1),           -1);
        L := ValidStrToInt(Copy(Strings[i], iPos + 3, Length(Strings[i])), -1);
      end;

      if (F < 0) or (L < 0) then Exit;
      AddSelection(F, L, '', i<(Strings.Count-1));
    end;

    Result         := False;
  finally
    Strings.Free;
  end;
end;

function TQUB_QVO_QVD_SelectionList.prGetAsTree: INode;
var
  i          :Integer;
  Sel        :TQUB_QVO_QVD_Selection;
  Node       :INode;
  Sels       :INode;
begin
  Result     := TNode.Create('List');
  // Result.Data := Name ??

  if Items.Count > 0 then begin
    Sels     := Result['Sels'];
    Sels.SetupNumData(NODE_DATATYPE_LONGINT, Items.Count, 2);
    Node     := Sels;
    for i := 0 to Items.Count - 1 do begin
      Sel    := Selection[i];

      //Sels.MatrixData[i,0] := Sel.Sel.First;
      //Sels.MatrixData[i,1] := Sel.Sel.Last;
      Move(Sel.Sel.First, (PChar(Sels.DataPtr) + 2*SizeOf(LongInt)*i)^, 2*SizeOf(LongInt));

      Node   := Result.InsertChild(Node, Sel.Lbl);
      Node.SetupNumData(NODE_DATATYPE_LONGINT, 1, 2);
      Move(Sel.Sel.First, Node.DataPtr^, 2*SizeOf(LongInt));

      if Sel.Comment <> '' then Node['Comment'].SetData(Sel.Comment);
      Node['Selected'].SetData(Integer(Selected[i]));

    end;
  end;
end;

function TQUB_QVO_QVD_SelectionList.prGetAsTreeSelected: INode;
var
  i, j       :Integer;
  Sel        :TQUB_QVO_QVD_Selection;
  Node       :INode;
  Sels       :INode;
begin
  Result     := TNode.Create('List');

  if Items.Count > 0 then begin
    Sels     := Result['Sels'];
    Sels.SetupNumData(NODE_DATATYPE_LONGINT, Max(SelCount, 0), 2);
    Node     := Sels;
    j        := 0;
    for i := 0 to Items.Count - 1 do begin
      if not Selected[i] then Continue;
      Sel    := Selection[i];

      //Sels.MatrixData[j,0] := Sel.Sel.First;
      //Sels.MatrixData[j,1] := Sel.Sel.Last;
      Move(Sel.Sel.First, (PChar(Sels.DataPtr) + 2*SizeOf(LongInt)*j)^, 2*SizeOf(LongInt));
      Inc(j);

      Node   := Result.InsertChild(Node, Sel.Lbl);
      Node.SetupNumData(NODE_DATATYPE_LONGINT, 1, 2);
      Move(Sel.Sel.First, Node.DataPtr^, 2*SizeOf(LongInt));

      if Sel.Comment <> '' then Node['Comment'].SetData(Sel.Comment);

    end;
  end;
end;

//My new format - not backward compatible, so it cannot be used. LM
(*
function TQUB_QVO_QVD_SelectionList.prGetAsTree: INode;
var
  SelsNode    :INode;
  SelNode     :INode;
  RangeNode   :INode;
  LabelNode   :INode;
  CommentNode :INode;
  Sel         :TQUB_QVO_QVD_Selection;
  i           :Integer;
begin
  Result     := TNode.Create('List');

  if Items.Count > 0 then begin

    SelsNode     := Result['Sels'];

    for i := 0 to Items.Count - 1 do begin

      Sel         := Selection[i];

      SelNode     := SelsNode.AppendChild('Sel');
      RangeNode   := SelNode['Range'];
      RangeNode.SetupNumData(NODE_DATATYPE_LONGINT, 1, 2);
      RangeNode.MatrixData[0, 0] := Sel.Sel.First;
      RangeNode.MatrixData[0, 1] := Sel.Sel.Last;

      if Sel.Lbl <> '' then begin
        LabelNode   := SelNode['Label'];
        LabelNode.SetData(Sel.Lbl);
      end;

      if Sel.Comment <> '' then begin
        CommentNode := SelNode['Comment'];
        CommentNode.SetData(Sel.Comment);
      end;

    end;
  end;
end;

function TQUB_QVO_QVD_SelectionList.prGetAsTreeSelected: INode;
var
  SelsNode    :INode;
  SelNode     :INode;
  RangeNode   :INode;
  LabelNode   :INode;
  CommentNode :INode;
  Sel         :TQUB_QVO_QVD_Selection;
  i           :Integer;
begin
  Result     := TNode.Create('List');

  if Items.Count > 0 then begin

    SelsNode     := Result['Sels'];

    for i := 0 to Items.Count - 1 do begin

      if not Selected[i] then Continue;

      Sel         := Selection[i];

      SelNode     := SelsNode.AppendChild('Sel');
      RangeNode   := SelNode['Range'];
      RangeNode.SetupNumData(NODE_DATATYPE_LONGINT, 1, 2);
      RangeNode.MatrixData[0, 0] := Sel.Sel.First;
      RangeNode.MatrixData[0, 1] := Sel.Sel.Last;

      if Sel.Lbl <> '' then begin
        LabelNode   := SelNode['Label'];
        LabelNode.SetData(Sel.Lbl);
      end;

      if Sel.Comment <> '' then begin
        CommentNode := SelNode['Comment'];
        CommentNode.SetData(Sel.Comment);
      end;

    end;
  end;
end;
*)

//Chris's old stuff
procedure TQUB_QVO_QVD_SelectionList.prSetAsTree(x :INode);
begin
  Clear;
  prSetAsTreeAppend(x);
end;

procedure TQUB_QVO_QVD_SelectionList.prSetAsTreeAppend(x :INode);
var
  Node, Sels      :INode;
  Next            :INode;
  SelNode         :INode;
  r               :Integer;
begin

  Sels            := QUB_Node_Null;
  Node            := x.Child;
  r               := 0;//
  while not Node.IsNull do begin
    Next          := Node.Sibling;
    if Node.Name = 'Sels' then begin
    //First Child is 'Sels'
      Sels        := Node;
    end else if  (Node.DataType = NODE_DATATYPE_LONGINT)
             and (Node.DataCount = 2) then begin
      AddSelection(Node.Data[0], Node.Data[1], Node.Name, not Next.IsNull).Comment := Node.Find('Comment').DataAsString;

      SelNode := Node.Find('Selected');//
      if not SelNode.IsNull then Selected[r] := SelNode.Data[0];//
      Inc(r);//

    end;

    Node          := Next;
  end;

  //Items.Count produces an exception when the list has no parent. LM
  if (Items.Count = 0) and (Sels.Rows > 0) then begin
    for r := 0 to Sels.Rows - 1 do begin
      AddSelection(Sels.MatrixData[r,0], Sels.MatrixData[r,1], '', r<(LongInt(Sels.Rows)-1));
    end;
  end;
end;
(*
//My new format - not backward compatible, so cannot be used. LM
procedure TQUB_QVO_QVD_SelectionList.prSetAsTree(x :INode);
var
  SelsNode    :INode;
  SelNode     :INode;
  RangeNode   :INode;
  LabelNode   :INode;
  CommentNode :INode;
  Sel         :TQUB_QVO_QVD_Selection;
begin
  Clear;

  SelsNode := x.Find('Sels');
  if SelsNode.IsNull then Exit;

  SelNode := SelsNode.Find('Sel');
  while not SelNode.IsNull do begin
    RangeNode   := SelNode.Find('Range');
    LabelNode   := SelNode.Find('Label');
    CommentNode := SelNode.Find('Comment');

    if not RangeNode.IsNull then begin
      Sel         := AddSelection(RangeNode.MatrixData[0, 0], RangeNode.MatrixData[0, 1], '');
      //Sel         := AddSelection(TPAInteger(RangeNode.DataPtr)[0], TPAInteger(RangeNode.DataPtr)[1], '');
      if not LabelNode.IsNull then begin
        Sel.Lbl     := LabelNode.DataAsString;
      end;
      if not CommentNode.IsNull then begin
        Sel.Comment := CommentNode.DataAsString;
      end;
    end;

    SelNode          := SelNode.Sibling;
  end;

end;

procedure TQUB_QVO_QVD_SelectionList.prSetAsTreeAppend(x :INode);
var
  SelsNode    :INode;
  SelNode     :INode;
  RangeNode   :INode;
  LabelNode   :INode;
  CommentNode :INode;
  Sel         :TQUB_QVO_QVD_Selection;
begin

  SelsNode := x.Find('Sels');
  if SelsNode.IsNull then Exit;

  SelNode := SelsNode.Find('Sel');
  while not SelNode.IsNull do begin
    RangeNode   := SelNode.Find('Range');
    LabelNode   := SelNode.Find('Label');
    CommentNode := SelNode.Find('Comment');

    if not RangeNode.IsNull then begin
      Sel         := AddSelection(RangeNode.MatrixData[0, 0], RangeNode.MatrixData[0, 1], '');
      //Sel         := AddSelection(TPAInteger(RangeNode.DataPtr)[0], TPAInteger(RangeNode.DataPtr)[1], '');
      if not LabelNode.IsNull then begin
        Sel.Lbl     := LabelNode.DataAsString;
      end;
      if not CommentNode.IsNull then begin
        Sel.Comment := CommentNode.DataAsString;
      end;
    end;

    SelNode          := SelNode.Sibling;
  end;

end;
*)

// *****************************************************************************
constructor TQUB_QVO_QVD_Selection.Create(First, Last :Integer; Lbl :String);
begin
  prSel.First := First;
  prSel.Last  := Last;
  prLabel     := Lbl;
  prActive    := True;
end;

procedure TQUB_QVO_QVD_Selection.prSetLabel(x :String);
begin
  prLabel := x;
  if Assigned(OnLabelChanged) then OnLabelChanged(Self);
end;

procedure TQUB_QVO_QVD_Selection.prSetComment(x :String);
begin
  prComment := x;
end;

procedure TQUB_QVO_QVD_Selection.prSetActive(x :Boolean);
begin
  if prActive = x then Exit;
  prActive := x;
  if Assigned(OnActiveChanged) then OnActiveChanged(Self);
end;


// *****************************************************************************

constructor TQUB_QVO_QVD_SelectionListIter.Create(List :TQUB_QVO_QVD_SelectionList; First, Last :Integer; ActiveOnly :Boolean); // -1, -1 for all
var
  i                    :Integer;
  Start                :Integer;
  Sel                  :TQUB_QVO_QVD_Selection;
  InRange              :Boolean;
begin
  prList               := List;
  prBounds.First       := First;
  prBounds.Last        := Last;
  prDone               := False;
  prActiveOnly         := ActiveOnly;

  Start := List.IndexAtOrAfter(First);
  while Start > 0 do begin
    if List[Start-1].Sel.Last < First then Break;
    Dec(Start);
  end;

  if prBounds.Last < 0 then begin
    prIndexRange       := QUB_FirstLast(0, List.Items.Count-1);
  end else begin
    InRange              := False;
    prIndexRange         := QUB_FirstLast(-1, -1);
    for i := Start to List.Items.Count - 1 do begin
      Sel                := List[i];

      if Sel.Sel.First > Last then begin
        Break;
      end else if InRange then begin
        prIndexRange.Last := i;
      end else begin
        if Sel.Sel.Last >= First then begin
          prIndexRange.First := i;
          prIndexRange.Last  := i;
          InRange            := True;
        end;
      end;
    end;
  end;
  
  prIndex              := Max(-1, prIndexRange.First - 1);
  prSel                := QUB_FirstLast(-1, -1);
  Next;
end;

function TQUB_QVO_QVD_SelectionListIter.Next: Boolean; // returns Done
var
  Sl                   :TQUB_FirstLast;
  Ix                   :Integer;
  Skip                 :Boolean;
begin
  if prIndex > prIndexRange.Last then begin
    prDone             := True;
    Result             := prDone;
    Exit;
  end;
  Ix                   := prIndex;
  Skip                 := True;
  repeat
    Inc(Ix);
    if Ix > prIndexRange.Last then Break;
    
    Sl                 := prList[Ix].Sel;
    Skip               := prActiveOnly and not prList[Ix].Active;
    if not Skip then begin
      if prBounds.First >= 0 then begin
        Sl.First       := Max(Sl.First, prBounds.First);
        Sl.Last        := Min(Sl.Last, prBounds.Last);
      end;
      if prList.ActiveRange.First >= 0 then begin
        Sl.Last        := Min(Sl.First + prList.ActiveRange.Last, Sl.Last);
        Sl.First       := Min(Sl.First + prList.ActiveRange.First, Sl.Last + 1);
      end;
      Skip             := Sl.First > Sl.Last;
    end;
  until not Skip;

  if Ix <= prIndexRange.Last then begin
    prSel              := Sl;
    prIndex            := Ix;
  end else begin
    prDone             := True;
  end;

  Result := prDone;
end;

function TQUB_QVO_QVD_SelectionListIter.prGetFullSel: TQUB_FirstLast;
begin
  Result := prList[prIndex].Sel;
end;

function TQUB_QVO_QVD_SelectionListIter.prGetLbl: String;
begin
  Result := prList[prIndex].Lbl;
  if Result = '' then Result := IntToStr(prIndex+1);
end;

function TQUB_QVO_QVD_SelectionListIter.prGetComment: String;
begin
  Result := prList[prIndex].Comment;
end;


// *****************************************************************************

constructor TQUB_QVO_QVD_ResultsRec.Create(ResultsNode :INode);
begin
  Node := ResultsNode;
end;

constructor TQUB_QVO_QVD_ResultsList.Create(AOwner :TComponent);
begin
  inherited;
  ShowHint    := True;
  Align       := alClient;
  MultiSelect := True;
  OnKeyPress  := prOnKeyPress;
  Style       := lbOwnerDrawFixed;
  OnDrawItem  := prOnDrawItem;
  prQVRs      := TQUB_ObjectCache.Create( prCreateQVR, prDestroyQVR, QUB_QVO_QVD_QVR_CACHESIZE);

  prPopup              := TPopupMenu.Create(Self);
  prPopup.OnPopup      := prOnPopup;
  Self.PopupMenu       := prPopup;
  prItemRename         := TMenuItem.Create(Self);
  prItemRename.Caption := 'Rename Results';
  prItemRename.OnClick := prOnClickRename;
  prPopup.Items.Add(prItemRename);
  prItemHold           := TMenuItem.Create(Self);
  prItemHold.Caption   := 'Hold Results';
  prItemHold.OnClick   := prOnClickHold;
  prPopup.Items.Add(prItemHold);
  prItemDelete         := TMenuItem.Create(Self);
  prItemDelete.Caption := 'Delete Results';
  prItemDelete.OnClick := prOnClickDelete;
  prPopup.Items.Add(prItemDelete);
end;

destructor TQUB_QVO_QVD_ResultsList.Destroy;
var
  i           :Integer;
begin
  try
    for i := 0 to Items.Count - 1 do begin
      TQUB_QVO_QVD_ResultsRec( Items.Objects[i] ).Free;
    end;
    prQVRs.Free;

  finally
    inherited;
  end;
end;

function TQUB_QVO_QVD_ResultsList.prCreateQVR(Key :Pointer): Pointer;
var
  Task        :TQUB_Task_BuildResults;
begin
  Task                     := TQUB_Task_BuildResults.Create(True);
  Task.FileName            := '';
  Task.ResultsNode         := TQUB_QVO_QVD_ResultsRec(Key).Node;
  Task.OnDone              := prAddBuiltResults;
  Task.FreeOnTerminate     := False;
  Task.Resume;

  while not Task.Done do Application.ProcessMessages;
  Task.WaitFor;
  Result                   := Task.QVR;
  Task.Free;
end;

procedure TQUB_QVO_QVD_ResultsList.prAddBuiltResults(Sender :TObject);
var
  Task                     :TQUB_Task_BuildResults;
  Results                  :TQUB_QVO_QVR;
begin
  Task                     := TQUB_Task_BuildResults(Sender);

  Results                  := TQUB_QVO_QVR.Create(nil, Task.Results);
  Results.Results.FileName := Task.FileName;
  ViewportResults.AddToDisplay(Results);
  Results.BuildInterface;

  Task.QVR                := Results;
end;


procedure TQUB_QVO_QVD_ResultsList.prDestroyQVR(Key, Obj :Pointer);
begin
  TQUB_QVO_QVR(Obj).Free;
end;

function TQUB_QVO_QVD_ResultsList.prGetCurrResults: TQUB_QVO_QVR;
begin
  if ItemIndex < 0 then Result := nil
	else Result := Results[ItemIndex]; // TQUB_QVO_QVR( prQVRs[ Items.Objects[ItemIndex] ] );
end;

function TQUB_QVO_QVD_ResultsList.prGetResults(Index :Integer) :TQUB_QVO_QVR;
begin
  Result := TQUB_QVO_QVR( prQVRs[ Items.Objects[Index] ] );
end;

function TQUB_QVO_QVD_ResultsList.prGetResultsNode(Index :Integer) :INode;
begin
  if (Index >= 0) and (Index < Items.Count) then begin
    Result := TQUB_QVO_QVD_ResultsRec( Items.Objects[Index] ).Node;
  end else begin
    Result := TNode.CreateNull;
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.CNCommand(var Message: TWMCommand);
var
  PrevItemIndex       :Integer;
  FirstSelected       :Integer;
  i                   :Integer;
begin
  case Message.NotifyCode of
    LBN_SELCHANGE:
      begin
        PrevItemIndex := ItemIndex;

        inherited;

        FirstSelected := -1;
        for i := 0 to Items.Count - 1 do begin
          if Selected[i] then begin
            FirstSelected := i;
            Break;
          end;
        end;

        // the chosen list is the first of multi-selection, or the most recently selected
        if FirstSelected = -1 then begin
          if PrevItemIndex <> -1 then begin
            ItemIndex := PrevItemIndex;
            Selected[ ItemIndex ] := True;
          end;
        end else {if (ItemIndex = -1) or not Selected[ ItemIndex ] then} begin
          ItemIndex := FirstSelected;
        end;

        //Generate OnItemChange Event
        if Assigned(prOnItemChange) then begin
          prOnItemChange(Self);
        end;
      end;
    LBN_DBLCLK:
      begin
        inherited;
        //Generate OnItemDblClick Event
        if Assigned(prOnItemDblClick) then begin
          prOnItemDblClick(Self);
        end;
      end;
  else
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.prOnKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;

  if Key = #13 then begin
    Key := #0;
    prOnItemDblClick(Self);
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.Activate;
begin
  if Items.Count <> 0 then begin
    if ItemIndex <> -1 then begin
       CurrResults.Activate;
    end;
  end;
  if CanFocus then begin
    try
      BringToFront;
      SetFocus;
    except
    end;
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.DeActivate;
begin
end;

procedure TQUB_QVO_QVD_ResultsList.prRemoveOldResults;
var
  i            :Integer;
begin
  i            := 0;
  while (i < Items.Count) and (Items.Count > QUB_QVO_QVD_RESULTS_MAXCOUNT) do begin
    if ResultsNode[i]['Hold'].Data[0] = 0 then begin
      DelResults(i);
    end else begin
      Inc(i);
    end;
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.AddResults(Results :INode);
var
  i            :Integer;
  DS, Seg      :INode;
  Nm           :String;
begin
  if Results.Find('Date').IsNull then
    Results['Date'].SetData( Format('%s %s',[DateTimeToStr( Date ),TimeToStr( Time )] ) );

  // set segment indices
  i := 0;
  DS := Results.Find('DataSet');  while not DS.IsNull do begin
    Seg := DS.Find('Segment');    while not Seg.IsNull do begin
      Seg['Index'].SetData(i);
      Seg := Seg.SiblingSameName;
      i   := i + 1;
    end;
    DS  := DS.SiblingSameName;
  end;

  // by default, a fresh result is ephemeral
  if Results.Find('Hold').DataCount = 0 then Results['Hold'].SetData( 0 );

  prRemoveOldResults;
  
  Nm := Results.Find('ResultName').DataAsString;
  if Nm = '' then Nm := Format('%s %s', [Results['ResultType'].DataAsString, Results.Find('DataDescription').DataAsString]);
  Items.AddObject( Nm, TQUB_QVO_QVD_ResultsRec.Create(Results) );
  ItemIndex := Items.Count - 1;

  for i := 0 to Items.Count - 1 do begin
    Selected[i] := (i = ItemIndex);
  end;

  //Generate OnItemChange Event
  if Assigned(prOnItemChange) then begin
    prOnItemChange(Self);
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.AddResultsBehind(Results :INode);
var
  Nm      :String;
begin
  Nm := Results.Find('ResultName').DataAsString;
  if Nm = '' then Nm := Format('%s %s', [Results['ResultType'].DataAsString, Results.Find('DataDescription').DataAsString]);
  Items.AddObject( Nm, TQUB_QVO_QVD_ResultsRec.Create(Results) );
end;

procedure TQUB_QVO_QVD_ResultsList.DelResults(Ix :Integer);
begin
  if (Ix >= 0) and (Ix < Items.Count) then begin
    prQVRs.FreeObject( Items.Objects[Ix] );
    TQUB_QVO_QVD_ResultsRec( Items.Objects[Ix] ).Free;
    Items.Delete(Ix);
  end;

  //Generate OnItemChange Event
  if Assigned(prOnItemChange) then begin
    prOnItemChange(Self);
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.DelSelected;
var
  i        :Integer;
begin
  for i := Items.Count - 1 downto 0 do begin
    if Selected[i] then begin
      prQVRs.FreeObject( Items.Objects[i] );
      TQUB_QVO_QVD_ResultsRec( Items.Objects[i] ).Free;
      Items.Delete(i);
    end;
  end;

  //Generate OnItemChange Event
  if Assigned(prOnItemChange) then begin
    prOnItemChange(Self);
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.ChooseResults(Ix :Integer);
var
  i           :Integer;
begin
  if (Ix >= 0) and (Ix < Items.Count) then begin
    for i := 0 to Items.Count - 1 do begin
      Selected[i] := (Ix = i);
    end;

    if Ix <> ItemIndex then begin
      ItemIndex := Ix;

      //Generate OnItemChange Event
      if Assigned(prOnItemChange) then begin
        prOnItemChange(Self);
      end;
    end;
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.Clear;
var
  Ix           :Integer;
begin
  if Parent <> nil then begin
    Items.BeginUpdate;
    try
      for Ix := Items.Count - 1 downto 0 do begin
        prQVRs.FreeObject( Items.Objects[Ix] );
        TQUB_QVO_QVD_ResultsRec( Items.Objects[Ix] ).Free;
        Items.Delete(Ix);
      end;
      ItemIndex := -1;
    finally
      Items.EndUpdate;
    end;
  end;
end;

procedure TQUB_QVO_QVD_ResultsList.prOnPopup(Sender :TObject);
begin
  prItemDelete.Enabled := ItemIndex >= 0;
  prItemHold.Enabled   := ItemIndex >= 0;
  prItemHold.Checked   := prItemHold.Enabled and (ResultsNode[ItemIndex]['Hold'].Data[0] <> 0);
  prItemRename.Enabled := ItemIndex >= 0;
end;

procedure TQUB_QVO_QVD_ResultsList.prOnClickDelete(Sender :TObject);
begin
  DelSelected;
end;

procedure TQUB_QVO_QVD_ResultsList.prOnClickHold(Sender :TObject);
begin
  if prItemHold.Checked then begin
    ResultsNode[ItemIndex]['Hold'].Data[0] := 0;
  end else begin
    ResultsNode[ItemIndex]['Hold'].Data[0] := 1;
  end;
  Invalidate;
end;

procedure TQUB_QVO_QVD_ResultsList.prOnClickRename(Sender :TObject);
var
  Nm      :String;
begin
  Nm      := InputBox('Rename Results', '', Items.Strings[ItemIndex]);
  Items.Strings[ItemIndex] := Nm;
  ResultsNode[ItemIndex]['ResultName'].SetData(Nm);
end;

procedure TQUB_QVO_QVD_ResultsList.prOnDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  Offset :Integer;
  Hold   :INode;
begin

  inherited;

  with Canvas do begin

    FillRect(Rect);

    Offset := 2;

    Hold := ResultsNode[Index]['Hold'];
    if (DataImageList <> nil) and not Hold.IsNull then begin
      if Hold.Data[0] <> 0 then begin
        //DataImageList.GetBitmap(0, Bitmap);
        DataImageList.Draw(Canvas, Rect.Left + Offset, Rect.Top, 2, True);
        Inc(Offset, 14);
      end;
      TextOut(Rect.Left + Offset, Rect.Top, Items[Index]);
    end;

  end;

end;


//******************************************************************************
end.
