{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// GoTo properties


unit QUB_QVO_QVD_GoToPoint_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Registry, ExtCtrls,

RegistryInterface,

QUB_Node,
QUB_Global,
QUB_Classic_Properties;


{$include 'QUB_QVO_QVD_GoToPoint_Properties.hct'}


type


TQUB_GoToPoint_Mode           = (GOTOPOINT_POINT, GOTOPOINT_TIME, GOTOPOINT_MARKER, GOTOPOINT_POINTS, GOTOPOINT_DURATION);


const


DEF_GOTOPOINT_START           = GOTOPOINT_POINT;
DEF_GOTOPOINT_END             = GOTOPOINT_POINTS;

DEF_GOTOPOINT_RELATIVETOSEL   = False;
DEF_GOTOPOINT_REALTIME        = False;

MIN_GOTOPOINT_POINT           = 0;
MAX_GOTOPOINT_POINT           = High(Integer) - 1;
DEF_GOTOPOINT_POINT           = 0;

MIN_GOTOPOINT_TIME            = 0;
MAX_GOTOPOINT_TIME            = 1e10;
DEF_GOTOPOINT_TIME            = 0;

MIN_GOTOPOINT_MARKER          = 0;
MAX_GOTOPOINT_MARKER          = High(Integer) - 1;
DEF_GOTOPOINT_MARKER          = 0;

MIN_GOTOPOINT_POINTS          = 1;
MAX_GOTOPOINT_POINTS          = High(Integer) - 1;
DEF_GOTOPOINT_POINTS          = 1;

MIN_GOTOPOINT_DURATION        = 0;
MAX_GOTOPOINT_DURATION        = 1e10;
DEF_GOTOPOINT_DURATION        = 0;

DEF_GOTOPOINT_LISTNAME        = 'GotoList';


type


TQUB_QVO_QVD_GoToPoint_PropertiesForm = class(TQUB_Classic_PropertiesForm)

  chkRealTime                    :TCheckBox;
  chkRelativeToSel               :TCheckBox;

  panStart                       :TPanel;
  lblStart                       :TLabel;
  txtStartPoint                  :TEdit;
  txtStartTime                   :TEdit;
  chkStartPoint                  :TRadioButton;
  chkStartTime                   :TRadioButton;
  chkStartMarker                 :TRadioButton;
  lstStartMarkers                :TComboBox;
  btnStartCrt                    :TButton;

  panEnd                         :TPanel;
  lblEnd                         :TLabel;
  lblEnd1                        :TLabel;
  chkEndPoint                    :TRadioButton;
  chkEndTime                     :TRadioButton;
  chkEndMarker                   :TRadioButton;
  chkEndPoints                   :TRadioButton;
  chkEndDuration                 :TRadioButton;
  txtEndPoint                    :TEdit;
  txtEndTime                     :TEdit;
  txtEndPoints                   :TEdit;
  txtEndDuration                 :TEdit;
  lstEndMarkers                  :TComboBox;
  btnEndCrt                      :TButton;

  lstAddToList                   :TComboBox;
  btnAddToList                   :TButton;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure chkStartClick(Sender: TObject);
  procedure txtStartPointKeyPress(Sender: TObject; var Key: Char);
  procedure txtStartPointExit(Sender: TObject);
  procedure txtStartTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtStartTimeExit(Sender: TObject);

  procedure chkEndClick(Sender: TObject);
  procedure txtEndPointExit(Sender: TObject);
  procedure txtEndPointKeyPress(Sender: TObject; var Key: Char);
  procedure txtEndTimeExit(Sender: TObject);
  procedure txtEndTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtEndPointsKeyPress(Sender: TObject; var Key: Char);
  procedure txtEndPointsExit(Sender: TObject);
  procedure txtEndDurationKeyPress(Sender: TObject; var Key: Char);
  procedure txtEndDurationExit(Sender: TObject);

  procedure lstStartMarkersClick(Sender: TObject);
  procedure lstEndMarkersClick(Sender: TObject);

  procedure btnStartCrtClick(Sender: TObject);
  procedure btnEndCrtClick(Sender: TObject);

  procedure chkRelativeToSelClick(Sender: TObject);
  procedure chkRealTimeClick(Sender: TObject);

  procedure btnAddToListClick(Sender: TObject);
  procedure lstAddToListClick(Sender: TObject);
  procedure lstAddToListKeyPress(Sender: TObject; var Key: Char);
  procedure lstAddToListExit(Sender: TObject);
  procedure txtStartPointMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure txtStartTimeMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure txtEndPointMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure txtEndTimeMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure txtEndPointsMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure txtEndDurationMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure btnStartCrtMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure btnEndCrtMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);


protected


  prGoToPointStart         :TQUB_GoToPoint_Mode;
  prGoToPointStartPoint    :Integer;
  prGoToPointStartTime     :Double;
  prGoToPointStartMarker   :Integer;

  prGoToPointEnd           :TQUB_GoToPoint_Mode;
  prGoToPointEndPoint      :Integer;
  prGoToPointEndTime       :Double;
  prGoToPointEndMarker     :Integer;
  prGoToPointEndPoints     :Integer;
  prGoToPointEndDuration   :Double;

  prGoToPointRelativeToSel :Boolean;
  prGoToPointRealTime      :Boolean;
  prGoToPointAddToList     :Boolean;
  prGoToPointListName      :String;

  prDialogOnLClick         :Boolean;

  prDataSel                :TQUB_FirstLast;


  procedure prSetGoToPointStart(s :TQUB_GoToPoint_Mode);
  procedure prSetGoToPointStartPoint(p :Integer);
  procedure prSetGoToPointStartTime(t :Double);
  procedure prSetGoToPointStartMarker(m :Integer);

  procedure prSetGoToPointEnd(e :TQUB_GoToPoint_Mode);
  procedure prSetGoToPointEndPoint(p :Integer);
  procedure prSetGoToPointEndTime(t :Double);
  procedure prSetGoToPointEndMarker(m :Integer);
  procedure prSetGoToPointEndPoints(p :Integer);
  procedure prSetGoToPointEndDuration(d :Double);

  procedure prSetGoToPointRelativeToSel(r :Boolean);
  procedure prSetGoToPointRealTime(r :Boolean);
  procedure prSetGoToPointAddToList(a :Boolean);
  procedure prSetGoToPointListName(n :String);

  procedure prMakeMarkerLists;

  procedure prPropertiesRegistry(Read :Boolean);


public


  function PropertiesTree(cfg :INode): INode;


  property GoToPointStart         :TQUB_GoToPoint_Mode    read prGoToPointStart          write prSetGoToPointStart;
  property GoToPointStartPoint    :Integer                read prGoToPointStartPoint     write prSetGoToPointStartPoint;
  property GoToPointStartTime     :Double                 read prGoToPointStartTime      write prSetGoToPointStartTime;
  property GoToPointStartMarker   :Integer                read prGoToPointStartMarker    write prSetGoToPointStartMarker;

  property GoToPointEnd           :TQUB_GoToPoint_Mode    read prGoToPointEnd            write prSetGoToPointEnd;
  property GoToPointEndPoint      :Integer                read prGoToPointEndPoint       write prSetGoToPointEndPoint;
  property GoToPointEndTime       :Double                 read prGoToPointEndTime        write prSetGoToPointEndTime;
  property GoToPointEndMarker     :Integer                read prGoToPointEndMarker      write prSetGoToPointEndMarker;
  property GoToPointEndPoints     :Integer                read prGoToPointEndPoints      write prSetGoToPointEndPoints;
  property GoToPointEndDuration   :Double                 read prGoToPointEndDuration    write prSetGoToPointEndDuration;

  property GoToPointRelativeToSel :Boolean                read prGoToPointRelativeToSel  write prSetGoToPointRelativeToSel;
  property GoToPointRealTime      :Boolean                read prGoToPointRealTime       write prSetGoToPointRealTime;
  property GoToPointAddToList     :Boolean                read prGoToPointAddToList      write prSetGoToPointAddToList;
  property GoToPointListName      :String                 read prGoToPointListName       write prSetGoToPointListName;

  property DataSel                :TQUB_FirstLast         read prDataSel                 write prDataSel;

  property DialogOnLClick         :Boolean                read prDialogOnLClick          write prDialogOnLClick;


end;


implementation


uses


QUB_QVO_QVD,
QUB_GUIInterface;


{$R *.DFM}


//******************************************************************************
//GoToPoint Start
procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointStart(s :TQUB_GoToPoint_Mode);
begin
  prGoToPointStart          := s;
  chkStartPoint.Checked     := s = GOTOPOINT_POINT;
  chkStartTime.Checked      := s = GOTOPOINT_TIME;
  chkStartMarker.Checked    := s = GOTOPOINT_MARKER;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointStartPoint(p :Integer);
begin
  prGoToPointStartPoint     := p;
  txtStartPoint.Text        := IntToStr(p);
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointStartTime(t :Double);
begin
  prGoToPointStartTime      := t;
  txtStartTime.Text         := FloatToStr(t * 1e3);
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointStartMarker(m :Integer);
begin
  prGoToPointStartMarker    := m;
  lstStartMarkers.ItemIndex := m;
end;

//GoToPoint End
procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointEnd(e :TQUB_GoToPoint_Mode);
begin
  prGoToPointEnd            := e;
  chkEndPoint.Checked       := e = GOTOPOINT_POINT;
  chkEndTime.Checked        := e = GOTOPOINT_TIME;
  chkEndMarker.Checked      := e = GOTOPOINT_MARKER;
  chkEndPoints.Checked      := e = GOTOPOINT_POINTS;
  chkEndDuration.Checked    := e = GOTOPOINT_DURATION;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointEndPoint(p :Integer);
begin
  prGoToPointEndPoint       := p;
  txtEndPoint.Text          := IntToStr(p);
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointEndTime(t :Double);
begin
  prGoToPointEndTime        := t;
  txtEndTime.Text           := FloatToStr(t * 1e3);
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointEndMarker(m :Integer);
begin
  prGoToPointEndMarker      := m;
  lstEndMarkers.ItemIndex   := m;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointEndPoints(p :Integer);
begin
  prGoToPointEndPoints      := p;
  txtEndPoints.Text         := IntToStr(p);
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointEndDuration(d :Double);
begin
  prGoToPointEndDuration    := d;
  txtEndDuration.Text       := FloatToStr(d * 1e3);
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointRelativeToSel(r :Boolean);
begin
  prGoToPointRelativeToSel  := r;
  chkRelativeToSel.Checked  := r;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointRealTime(r :Boolean);
begin
  prGoToPointRealTime       := r;
  chkRealTime.Checked       := r;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointAddToList(a :Boolean);
begin
  prGoToPointAddToList      := a;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prSetGoToPointListName(n :String);
var
  t :String;
begin
  t := Trim(n);
  if t = '' then begin
    n := prGoToPointListName;
  end else begin
    n := t;
  end;
  if lstAddToList.Items.IndexOf(n) = -1 then lstAddToList.Items.Add(n);
  lstAddToList.ItemIndex := lstAddToList.Items.IndexOf(n);
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prMakeMarkerLists;
var
  Data :TQUB_QVO_QVD;
  i    :Integer;
begin
  lstStartMarkers.Items.Clear;
  lstEndMarkers.Items.Clear;

  Data := ViewportData.CurrData;
  if Data = nil then Exit;
  if Data.DataTimeMarkersCount = 0 then Exit;

  for i := 0 to Data.DataTimeMarkersCount - 1 do begin
    lstStartMarkers.Items.Add(Format('%d (%d)', [i, Round(Data.DataTimeMarkers[i])]));
  end;
  lstEndMarkers.Items.Assign(lstStartMarkers.Items);

end;


//******************************************************************************
procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(True{Read});

  {$include 'QUB_QVO_QVD_GoToPoint_Properties.htx'}

  prDialogOnLClick             := True;

  WikiTopic                    := 'Preprocessing:Go_To_Point';

end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.FormDestroy(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(False{Write});

end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  prMakeMarkerLists;

  GoToPointStart             := GoToPointStart;
  GoToPointStartPoint        := GoToPointStartPoint;
  GoToPointStartTime         := GoToPointStartTime;
  GoToPointStartMarker       := GoToPointStartMarker;
  GoToPointEnd               := GoToPointEnd;
  GoToPointEndPoint          := GoToPointEndPoint;
  GoToPointEndTime           := GoToPointEndTime;
  GoToPointEndMarker         := GoToPointEndMarker;
  GoToPointEndPoints         := GoToPointEndPoints;
  GoToPointEndDuration       := GoToPointEndDuration;
  GoToPointRelativeToSel     := GoToPointRelativeToSel;
  GoToPointRealTime          := GoToPointRealTime;
  GoToPointAddToList         := False;
  GoToPointListName          := GoToPointListName;
end;

//******************************************************************************
procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.chkStartClick(Sender: TObject);
begin
  inherited;
  if          Sender = chkStartPoint  then begin
    if GoToPointStart <> GOTOPOINT_POINT  then GoToPointStart := GOTOPOINT_POINT;
  end else if Sender = chkStartTime   then begin
    if GoToPointStart <> GOTOPOINT_TIME   then GoToPointStart := GOTOPOINT_TIME;
  end else if Sender = chkStartMarker then begin
    if GoToPointStart <> GOTOPOINT_MARKER then GoToPointStart := GOTOPOINT_MARKER;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.chkEndClick(Sender: TObject);
begin
  inherited;
  if          Sender = chkEndPoint    then begin
    if GoToPointEnd <> GOTOPOINT_POINT    then GoToPointEnd   := GOTOPOINT_POINT;
  end else if Sender = chkEndTime     then begin
    if GoToPointEnd <> GOTOPOINT_TIME     then GoToPointEnd   := GOTOPOINT_TIME;
  end else if Sender = chkEndMarker   then begin
    if GoToPointEnd <> GOTOPOINT_MARKER   then GoToPointEnd   := GOTOPOINT_MARKER;
  end else if Sender = chkEndPoints   then begin
    if GoToPointEnd <> GOTOPOINT_POINTS   then GoToPointEnd   := GOTOPOINT_POINTS;
  end else if Sender = chkEndDuration then begin
    if GoToPointEnd <> GOTOPOINT_DURATION then GoToPointEnd   := GOTOPOINT_DURATION;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.lstStartMarkersClick(Sender: TObject);
begin
  inherited;
  if GoToPointStartMarker <> lstStartMarkers.ItemIndex then GoToPointStartMarker := lstStartMarkers.ItemIndex;
  if GoToPointStart <> GOTOPOINT_MARKER then GoToPointStart := GOTOPOINT_MARKER;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.lstEndMarkersClick(Sender: TObject);
begin
  inherited;
  if GoToPointEndMarker   <> lstEndMarkers.ItemIndex   then GoToPointEndMarker   := lstEndMarkers.ItemIndex;
  if GoToPointEnd <> GOTOPOINT_MARKER then GoToPointEnd := GOTOPOINT_MARKER;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.chkRelativeToSelClick(Sender: TObject);
begin
  inherited;
  if GoToPointRelativeToSel <> TCheckBox(Sender).Checked then GoToPointRelativeToSel := TCheckBox(Sender).Checked;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.chkRealTimeClick(Sender: TObject);
begin
  inherited;
  if GoToPointRealTime <> TCheckBox(Sender).Checked then GoToPointRealTime := TCheckBox(Sender).Checked;
end;

//Start
procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtStartPointKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtStartPointExit(Sender);
  end;
  // if changes to point parameters, force radio button to point mode
  chkStartPoint.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtStartPointMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkStartPoint.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtStartPointExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := StrToInt(TEdit(Sender).Text);
    if ((p < MIN_GOTOPOINT_POINT) or (p > MAX_GOTOPOINT_POINT)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_GOTOPOINT_POINT) + ' and ' + IntToStr(MAX_GOTOPOINT_POINT), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(GoToPointStartPoint);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> GoToPointStartPoint then begin
    GoToPointStartPoint := p;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtStartTimeKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtStartTimeExit(Sender);
  end;
  // if changes to time parameters, force radio button to time mode
  chkStartTime.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtStartTimeMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkStartTime.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtStartTimeExit(Sender: TObject);
var
  t :Double;
begin
  inherited;
  try
    t := StrToFloat(TEdit(Sender).Text) * 1e-3;
    if ((t < MIN_GOTOPOINT_TIME) or (t > MAX_GOTOPOINT_TIME)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_GOTOPOINT_TIME * 1e3) + ' and ' + FloatToStr(MAX_GOTOPOINT_TIME * 1e3) + ' ms', mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(GoToPointStartTime * 1e3);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if t <> GoToPointStartTime then begin
    GoToPointStartTime := t;
  end;
end;

//End
procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndPointKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtEndPointExit(Sender);
  end;
  // if changes to point parameters, force radio button to point mode
  chkEndPoint.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndPointMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkEndPoint.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndPointExit(Sender: TObject);
var
  p :Integer;
begin
  inherited;
  try
    p := StrToInt(TEdit(Sender).Text);
    if ((p < MIN_GOTOPOINT_POINT) or (p > MAX_GOTOPOINT_POINT)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_GOTOPOINT_POINT) + ' and ' + IntToStr(MAX_GOTOPOINT_POINT), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(GoToPointEndPoint);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> GoToPointEndPoint then begin
    GoToPointEndPoint := p;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndTimeKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtEndTimeExit(Sender);
  end;
  // if changes to time parameters, force radio button to time mode
  chkEndTime.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndTimeMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkEndTime.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndTimeExit(Sender: TObject);
var
  t :Double;
begin
  inherited;
  try
    t := StrToFloat(TEdit(Sender).Text) * 1e-3;
    if ((t < MIN_GOTOPOINT_TIME) or (t > MAX_GOTOPOINT_TIME)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_GOTOPOINT_TIME * 1e3) + ' and ' + FloatToStr(MAX_GOTOPOINT_TIME * 1e3) + ' ms', mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(GoToPointEndTime * 1e3);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if t <> GoToPointEndTime then begin
    GoToPointEndTime := t;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndPointsKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtEndPointsExit(Sender);
  end;
  // if changes to point parameters, force radio button to point mode
  chkEndPoints.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndPointsMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkEndPoints.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndPointsExit(Sender: TObject);
var
  s :Integer;
begin
  inherited;
  try
    s := StrToInt(TEdit(Sender).Text);
    if ((s < MIN_GOTOPOINT_POINTS) or (s > MAX_GOTOPOINT_POINTS)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_GOTOPOINT_POINTS) + ' and ' + IntToStr(MAX_GOTOPOINT_POINTS), mtError, [mbOK], 0);
    GoToPointEndPoints := GoToPointEndPoints;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if s <> GoToPointEndPoints then begin
    GoToPointEndPoints := s;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndDurationKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtEndDurationExit(Sender);
  end;
  // if changes to time parameters, force radio button to time mode
  chkEndDuration.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndDurationMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkEndDuration.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.txtEndDurationExit(Sender: TObject);
var
  d :Double;
begin
  inherited;
  try
    d := StrToFloat(TEdit(Sender).Text) * 1e-3;
    if ((Abs(d) < MIN_GOTOPOINT_DURATION) or (Abs(d) > MAX_GOTOPOINT_DURATION)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between +/-' + FloatToStr(MIN_GOTOPOINT_DURATION * 1e3) + ' and +/-' + FloatToStr(MAX_GOTOPOINT_DURATION * 1e3) + ' ms', mtError, [mbOK], 0);
    GoToPointEndDuration := GoToPointEndDuration;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if d <> GoToPointEndDuration then begin
    GoToPointEndDuration := d;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.btnStartCrtClick(Sender: TObject);
begin
  inherited;
  GoToPointStartPoint := prDataSel.First;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.btnStartCrtMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkStartPoint.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.btnEndCrtClick(Sender: TObject);
begin
  inherited;
  GoToPointEndPoint := prDataSel.First;
end;


procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.btnEndCrtMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  chkEndPoint.Checked := True;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.btnAddToListClick(Sender: TObject);
begin
  if Assigned(prOnGetProperties) then begin
    try
      prGoToPointAddToList := True;
      prOnGetProperties(Self);
    finally
      prGoToPointAddToList := False;
    end;
  end;
  inherited;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.lstAddToListClick(Sender: TObject);
begin
  inherited;
  GoToPointListName := lstAddToList.Text;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.lstAddToListKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    GoToPointListName := lstAddToList.Text;
  end;
end;

procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.lstAddToListExit(Sender: TObject);
begin
  inherited;
  GoToPointListName := lstAddToList.Text;
end;


//Registry
procedure TQUB_QVO_QVD_GoToPoint_PropertiesForm.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
begin
  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Go To Point', True) then begin
        if Read then begin

          prGoToPointStart         := TQUB_GoToPoint_Mode(ReadRegistryInteger(Reg, 'GoToPoint Start', Integer(DEF_GOTOPOINT_START)));
          prGoToPointStartPoint    := ReadRegistryInteger(Reg, 'GoToPoint Start Point', DEF_GOTOPOINT_POINT);
          prGoToPointStartTime     := ReadRegistryFloat(Reg, 'GoToPoint Start Time', DEF_GOTOPOINT_TIME);
          prGoToPointStartMarker   := ReadRegistryInteger(Reg, 'GoToPoint Start Marker', DEF_GOTOPOINT_MARKER);

          prGoToPointEnd           := TQUB_GoToPoint_Mode(ReadRegistryInteger(Reg, 'GoToPoint End', Integer(DEF_GOTOPOINT_END)));
          prGoToPointEndPoint      := ReadRegistryInteger(Reg, 'GoToPoint End Point', DEF_GOTOPOINT_POINT);
          prGoToPointEndTime       := ReadRegistryFloat(Reg, 'GoToPoint End Time', DEF_GOTOPOINT_TIME);
          prGoToPointEndMarker     := ReadRegistryInteger(Reg, 'GoToPoint End Marker', DEF_GOTOPOINT_MARKER);
          prGoToPointEndPoints     := ReadRegistryInteger(Reg, 'GoToPoint End Points', DEF_GOTOPOINT_POINTS);
          prGoToPointEndDuration   := ReadRegistryFloat(Reg, 'GoToPoint End Duration', DEF_GOTOPOINT_DURATION);

          prGoToPointRelativeToSel := ReadRegistryBoolean(Reg, 'GoToPoint Relative To Sel', DEF_GOTOPOINT_RELATIVETOSEL);
          prGoToPointRealTime      := ReadRegistryBoolean(Reg, 'GoToPoint Real Time', DEF_GOTOPOINT_REALTIME);
          prGoToPointListName      := ReadRegistryString(Reg, 'GoToPoint List Name', DEF_GOTOPOINT_LISTNAME);

        end else begin

          WriteInteger('GoToPoint Start', Integer(GoToPointStart));
          WriteInteger('GoToPoint Start Point', GoToPointStartPoint);
          WriteFloat('GoToPoint Start Time', GoToPointStartTime);
          WriteInteger('GoToPoint Start Marker', GoToPointStartMarker);

          WriteInteger('GoToPoint End', Integer(GoToPointEnd));
          WriteInteger('GoToPoint End Point', GoToPointEndPoint);
          WriteFloat('GoToPoint End Time', GoToPointEndTime);
          WriteInteger('GoToPoint End Marker', GoToPointEndMarker);
          WriteInteger('GoToPoint End Points', GoToPointEndPoints);
          WriteFloat('GoToPoint End Duration', GoToPointEndDuration);

          WriteBool('GoToPoint Relative To Sel', GoToPointRelativeToSel);
          WriteBool('GoToPoint Real Time', GoToPointRealTime);
          WriteString('GoToPoint List Name', GoToPointListName);

        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;

function TQUB_QVO_QVD_GoToPoint_PropertiesForm.PropertiesTree(cfg :INode): INode;
var
  Item                           :INode;
  oneOrZero                      :Integer;
begin
  cfg                            := cfg.Clone(True); // don't modify what's passed in
  Result                         := cfg;

  //TO DO:

  {
  Item                           := cfg['PointNotTime'];
  if Item.DataCount > 0 then begin
    GoToPointStart               := Integer( Item.Data[0] ) <> 0;
  end else begin
    if GoToPointPointOrTime then oneOrZero := 1 else oneOrZero := 0;
    Item.SetData( oneOrZero );
  end;

  Item                           := cfg['Point'];
  if Item.DataCount > 0 then begin
    GoToPointPoint               := Integer( Item.Data[0] );
  end else begin
    Item.SetData( GoToPointPoint );
  end;

  Item                           := cfg['Time'];
  if Item.DataCount > 0 then begin
    GoToPointTime                := Double( Item.Data[0] );
  end else begin
    Item.SetData( GoToPointTime );
  end;

  Item                           := cfg['Selected'];
  if Item.DataCount > 0 then begin
    GoToPointSelected            := Integer( Item.Data[0] );
  end else begin
    Item.SetData( GoToPointSelected );
  end;

  Item                           := cfg['Duration'];
  if Item.DataCount > 0 then begin
    GoToPointDuration            := Double( Item.Data[0] );
  end else begin
    Item.SetData( GoToPointDuration );
  end;
  }

  Item                           := cfg['RealTime'];
  if Item.DataCount > 0 then begin
    GoToPointRealTime            := Integer( Item.Data[0] ) <> 0;
  end else begin
    if GoToPointRealTime then oneOrZero := 1 else oneOrZero := 0;
    Item.SetData( oneOrZero );
  end;

  Item                           := cfg['DialogOnLClick'];
  if Item.DataCount > 0 then begin
    DialogOnLClick               := Integer( Item.Data[0] ) <> 0;
  end else begin
    if DialogOnLClick then oneOrZero := 1 else oneOrZero := 0;
    Item.SetData( oneOrZero );
  end;

end;


//******************************************************************************
end.
