{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2003
// components for QUB Viewport DataOverlap


unit QUB_QVO_QVD_Overlap;


interface


uses


Classes, Forms, Menus, Controls, Extctrls, Stdctrls, Grids, Dialogs, Registry, Sysutils, Graphics, Windows, Math, Clipbrd,

PythonEngine,
SelectorListBox,

Alloc, AllocTypes,

QUB_Global,
QUB_Table,
QUB_Node,
QUBFits,
QUBFits_Dialog,
QUB_QFS, QUB_QFS_QDF,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_Viewport_Data, QUB_ListBox;


type


ArrayOfSingle                = Array of Single;
ArrayOfArrayOfSingle         = Array of ArrayOfSingle;
ArrayOfArrayOfArrayOfSingle  = Array of ArrayOfArrayOfSingle;

TQUB_QVO_QVD_Overlap         = class;
TQUB_QVO_QVD_Overlap_Channel = class;
TQUB_QVO_QVD_OverlapActive   = class;
TQUB_QVO_QVD_ListResampler   = class;


TQUB_QVO_QVD_ActiveSelList   = class(TQUB_Table)
protected


  prList                     :TQUB_QVO_QVD_SelectionList;
  prData                     :TQUB_QVO_QVD;

  prPopup                    :TPopupMenu;
  itemSelectAll              :TMenuItem;
  itemSelectNone             :TMenuItem;


  procedure prSetList(x :TQUB_QVO_QVD_SelectionList);
  procedure prRelabelItem(i :Integer);

  procedure prOnSelectAll(Sender :TObject);

  procedure prOnItemAdded(Sender :TObject);
  procedure prOnItemDeleting(Sender :TObject);
  procedure prOnItemChanged(Sender :TObject);
  procedure prOnLabelChanged(Sender :TObject);
  procedure prOnActiveChanged(Sender :TObject);

  procedure prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnKeyUp(Sender :TObject; var Key : Word; Shift: TShiftState);
  procedure prOnSelectCell(Sender: TObject; ACol, ARow: Longint; var CanSelect: Boolean);
  procedure prOnDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;


  property List         :TQUB_QVO_QVD_SelectionList       read prList      write prSetList;
  property Data         :TQUB_QVO_QVD                     read prData      write prData;


end;


TQUB_QVO_QVD_DataListMirror = class(TSelectorListBox)
protected


  procedure prOnItemChange(Sender :TObject);
  procedure prOnItemDblClick(Sender :TObject);


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure OnOpenData(Sender :TObject);
  procedure OnClosingData(Sender :TObject);
  procedure OnChangeData(Sender :TObject);


end;


TQUB_QVO_QVD_ListZoomState = class
protected
  prList                  :TQUB_QVO_QVD_SelectionList;
  prActive                :array of Boolean;
  prActiveRange           :TQUB_FirstLast;

public
  constructor Create(List :TQUB_QVO_QVD_SelectionList);

  procedure Apply;
end;

TQUB_QVO_QVD_ListZoomStack = class
protected
  prList                  :TQUB_QVO_QVD_SelectionList;
  prStack                 :TList;

  procedure prSetList(x :TQUB_QVO_QVD_SelectionList);
  function prGetEmpty: Boolean;

  procedure prOnItemAdded(Sender :TObject);
  procedure prOnItemDeleting(Sender :TObject);

public
  constructor Create;
  destructor Destroy; override;

  procedure Push;
  procedure Pop;
  procedure Clear;

  property List           :TQUB_QVO_QVD_SelectionList   read prList   write prSetList;
  property Empty          :Boolean                      read prGetEmpty;
end;


TQUB_QVO_QVD_Overlap = class(TCustomControl)
protected
  prLblStatus             :TLabel;
  prLblStatusActive       :TLabel;

  prData                  :TQUB_QVO_QVD;
  prList                  :TQUB_QVO_QVD_SelectionList;
  prResampler             :TQUB_QVO_QVD_ListResampler;
  prZoomStack             :TQUB_QVO_QVD_ListZoomStack;

  prPopup                 :TPopupMenu;
  itemSelectAll           :TMenuItem;

  prChannels              :TList;

  procedure prSetLblStatus(x :TLabel);
  procedure prSetLblStatusActive(x :TLabel);
  procedure prSetData(x :TQUB_QVO_QVD);
  procedure prSetList(x :TQUB_QVO_QVD_SelectionList);
  function prGetChannelCount: Integer;
  function prGetChannel(Ix :Integer): TQUB_QVO_QVD_Overlap_Channel;

  procedure prOnPopup(Sender :TObject);
  procedure prOnClickSelectAll(Sender :TObject);

  procedure prOnNeedsResample(Sender :TObject; Channel, Sel :Integer);
  procedure prOnYRangeChanged(Sender :TObject; Channel :Integer);

  procedure prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  procedure prOnBeforeZoomChannel(Sender :TObject);

public

  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  procedure Paint; override;

  property LblStatus      :TLabel                      read prLblStatus      write prSetLblStatus;
  property LblStatusActive :TLabel                     read prLblStatusActive write prSetLblStatusActive;
  property Data           :TQUB_QVO_QVD                read prData           write prSetData;
  property List           :TQUB_QVO_QVD_SelectionList  read prList           write prSetList;
  property ZoomStack      :TQUB_QVO_QVD_ListZoomStack  read prZoomStack      write prZoomStack;
  property ChannelCount   :Integer                     read prGetChannelCount;
  property Channels[Ix :Integer] :TQUB_QVO_QVD_Overlap_Channel    read prGetChannel;

end;


TQUB_QVO_QVD_OverlapMode = (QUB_QVO_QVD_OVERLAP_OVERLAP,
                            QUB_QVO_QVD_OVERLAP_AVERAGE,
                            QUB_QVO_QVD_OVERLAP_FUNCTION,
                            QUB_QVO_QVD_OVERLAP_SUM);

TQUB_QVO_QVD_Overlap_Channel = class(TCustomControl)
protected

  prLblStatus             :TLabel;
  prLblStatusActive       :TLabel;
  prChannelIndex          :Integer;
  prList                  :TQUB_QVO_QVD_SelectionList;
  prResampler             :TQUB_QVO_QVD_ListResampler;

  prBitmap                :Graphics.TBitmap;
  prBitmapCanvas          :TCanvas;

  prNeedsRedraw           :array of Boolean;
  prAllNeedRedraw         :Boolean;
  prDrawnItemIndex        :Integer;
  prYfactor               :Single; // ypix = Yfactor * (ymax - yval)
  prYmin, prYmax          :Single;
  prAvg, prStd            :array of Single;   // prAvg doubles as function samples

  prMarqueeing            :Boolean;
  prMarqueeStart          :TPoint;
  prMarquee               :TRect;
  prMarqueeInvalid        :Boolean;
  prCanMarquee            :Boolean;
  prShowMarquee           :Boolean;

  prMode                  :TQUB_QVO_QVD_OverlapMode;
  prJustChangedMode       :Boolean;
  prFunc                  :PPyObject;
  prFitData               :INode;
  prFitNeedsRedraw        :Boolean;
  prVisDataSet            :TQUB_FitDataSet;


  procedure prSetChannelIndex(x :Integer);
  procedure prSetList(x :TQUB_QVO_QVD_SelectionList);
  procedure prSetShowMarquee(x :Boolean);                                       
  procedure prSetMode(x :TQUB_QVO_QVD_OverlapMode);
  procedure prSetFunc(x :PPyObject);

  procedure prOnItemAdded(Sender :TObject);
  procedure prOnItemDeleting(Sender :TObject);
  procedure prOnItemChanged(Sender :TObject);

  // mouse events         : drag marquee: starts if far enough from mousedown;
  //                      :               set rect then invalidate
  //                      : mouseUp: if marqueeing: set list.activeRange and each sel.active
  //                      : on move: set LblStatus.Caption := 'Mouse: (%f, %f)'
  //                        + (maybe) '\tTrace %d: (%f, %f)'
  //                      : on click: List.Choose( closest/cycling )
  procedure prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

  procedure prOnActiveChanged(Sender :TObject);
  procedure prOnActiveRangeChanged(Sender :TObject);

  procedure prOnCurvesChanged(Sender :TObject);
  procedure prUpdateFitData;

  procedure prMarqueeToActive;
  procedure prActiveToMarquee;

  procedure prRedrawTrace(Ix :Integer);
  procedure prRedrawTraces;

  procedure prRecalcAvgStd;
  procedure prRecalcSum;
  procedure prRecalcFunc;
  procedure prRedrawAvgStd;

public

  OnBeforeZoom         :TNotifyEvent;

  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  procedure Paint; override;

  procedure DoOnNeedsResample(Sender :TObject; Sel :Integer);
  procedure DoOnYRangeChanged(Sender :TObject);

  procedure PixToData(x, y :Integer; var Sec, Val :Double);
  procedure DataToPix(Sec, Val :Double; var x, y :Integer);

  procedure ShowFit;
  procedure ClearFit;

  property LblStatus      :TLabel                      read prLblStatus      write prLblStatus;
  property LblStatusActive :TLabel                     read prLblStatusActive write prLblStatusActive;
  property ChannelIndex   :Integer                     read prChannelIndex   write prSetChannelIndex;
  property List           :TQUB_QVO_QVD_SelectionList  read prList           write prSetList;
  property Resampler      :TQUB_QVO_QVD_ListResampler  read prResampler      write prResampler;

  property CanMarquee     :Boolean                     read prCanMarquee     write prCanMarquee;
  property ShowMarquee    :Boolean                     read prShowMarquee    write prSetShowMarquee;

  property Mode           :TQUB_QVO_QVD_OverlapMode    read prMode           write prSetMode;

  // please set this one while Mode is not _FUNCTION, while holding the python g.i.lock
  property Func           :PPyObject                   read prFunc           write prSetFunc;

  property Bitmap         :Graphics.TBitmap            read prBitmap;
end;


TQUB_QVO_QVD_OverlapActive = class(TCustomControl)
protected
  prLblStatus             :TLabel;
  prLblStatusActive       :TLabel;

  prData                  :TQUB_QVO_QVD;
  prList                  :TQUB_QVO_QVD_SelectionList;
  prResampler             :TQUB_QVO_QVD_ListResampler;
  prFitResampler          :TQUB_QVO_QVD_ListResampler;
  prZoomStack             :TQUB_QVO_QVD_ListZoomStack;

  prChannel               :TQUB_QVO_QVD_Overlap_Channel;
  prFitChannel            :TQUB_QVO_QVD_Overlap_Channel;

  prFuncText              :String;

  prPopup                 :TPopupMenu;
  itemZoomOut             :TMenuItem;
  itemOverlap             :TMenuItem;
  itemAverage             :TMenuItem;
  itemSum                 :TMenuItem;
  itemFunction            :TMenuItem;
  itemAppend              :TMenuItem;
  itemCurveFit            :TMenuItem;
  itemClearFit            :TMenuItem;
  itemNewList             :TMenuItem;
  itemCopyImage           :TMenuItem;

  procedure prSetLblStatus(x :TLabel);
  procedure prSetLblStatusActive(x :TLabel);
  procedure prSetData(x :TQUB_QVO_QVD);
  procedure prSetList(x :TQUB_QVO_QVD_SelectionList);

  procedure prOnActiveChannelChange(Sender :TObject);
  procedure prOnFitCurvesChanged(Sender :TObject);

  procedure prOnNeedsResample(Sender :TObject; Channel, Sel :Integer);
  procedure prOnYRangeChanged(Sender :TObject; Channel :Integer);

  procedure prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

  procedure prOnPopup(Sender :TObject);
  procedure prOnClickZoomOut(Sender :TObject);
  procedure prOnClickOverlap(Sender :TObject);
  procedure prOnClickAverage(Sender :TObject);
  procedure prOnClickSum(Sender :TObject);
  procedure prOnClickFunction(Sender :TObject);
  procedure prOnClickAppend(Sender :TObject);
  procedure prOnClickCurveFit(Sender :TObject);
  procedure prOnClickClearFit(Sender :TObject);
  procedure prOnClickNewList(Sender :TObject);
  procedure prOnClickCopyImage(Sender :TObject);

  procedure prOnBeforeZoomChannel(Sender :TObject);

public

  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  procedure Paint; override;

  property LblStatus      :TLabel                      read prLblStatus      write prSetLblStatus;
  property LblStatusActive :TLabel                     read prLblStatusActive write prSetLblStatusActive;
  property Data           :TQUB_QVO_QVD                read prData           write prSetData;
  property List           :TQUB_QVO_QVD_SelectionList  read prList           write prSetList;
  property ZoomStack      :TQUB_QVO_QVD_ListZoomStack  read prZoomStack      write prZoomStack;
  property Channel        :TQUB_QVO_QVD_Overlap_Channel    read prChannel;
  property FitChannel     :TQUB_QVO_QVD_Overlap_Channel    read prFitChannel;

end;


TQUB_QVO_QVD_ListResampler = class
protected
  prData                  :TQUB_QVO_QVD;
  prSampleCount           :Integer;
  prChannelIndex          :Integer;
  prActiveOnly            :Boolean;
  prFitCurves             :Boolean;
  prList                  :TQUB_QVO_QVD_SelectionList;

  procedure prResampleAll;
  procedure prResampleSome;
  function prResampleOne(ch, s :Integer): Boolean; // returns true if Ymin, Ymax changed outside ChannelY* (or if they were the channel minmax)

  procedure prSetData(x :TQUB_QVO_QVD);
  procedure prSetSampleCount(x :Integer);
  procedure prSetChannelIndex(x :Integer);
  procedure prSetActiveOnly(x :Boolean);
  procedure prSetFitCurves(x :Boolean);
  procedure prSetList(x :TQUB_QVO_QVD_SelectionList);

  procedure prOnDataEdit(Sender :TObject);
  procedure prOnDataProcessChange(Sender :TObject);
  procedure prOnFitCurvesChanged(Sender :TObject);

  // list events
  procedure prOnItemAdded(Sender :TObject);
  procedure prOnItemDeleting(Sender :TObject);
  procedure prOnActiveChanged(Sender :TObject);
  procedure prOnActiveRangeChanged(Sender :TObject);


public

  // always accurate
  SelCount                :Integer;
  DataCount               :array of Integer; // [sel]
  MaxDataCount            :Integer;
  ChannelCount            :Integer;

  Invalid                 :array of array of Boolean; // [ch][sel]
  AllInvalid              :Boolean;
  AllValid                :Boolean;

  // accurate unless invalid (needs resample)
  Samples                 :ArrayOfArrayOfArrayOfSingle; // [ch][sel][s]
  SelSampleCount          :array of Integer;
  Xfactor                 :Double; // point = factor * sample + offset
  Xoffset                 :Integer;
  Ymin                    :ArrayOfArrayOfSingle; // [ch][sel]
  Ymax                    :ArrayOfArrayOfSingle;
  ChannelYmin             :ArrayOfSingle;
  ChannelYmax             :ArrayOfSingle;

  OnNeedsResample         :procedure (Sender :TObject; Channel, Sel :Integer) of object;
  // some or all sels are invalid (as a result of an event from the list or data, or a fresh assignment of list or data)

  OnYRangeChanged         :procedure (Sender :TObject; Channel :Integer) of object;
  // a channel's ymin/max has changed (during DoResample)

  constructor Create;
  destructor Destroy; override;

  function DoResample: Boolean; // True if any resampling done

  property Data           :TQUB_QVO_QVD               read prData           write prSetData;
  property SampleCount    :Integer                    read prSampleCount    write prSetSampleCount;
  property ChannelIndex   :Integer                    read prChannelIndex   write prSetChannelIndex;
  property ActiveOnly     :Boolean                    read prActiveOnly     write prSetActiveOnly;
  property FitCurves      :Boolean                    read prFitCurves      write prFitCurves;
  property List           :TQUB_QVO_QVD_SelectionList read prList           write prSetList;

end;


implementation


uses


PythonAtom,
QUB_Viewport_Python,
QUB_Python_Thread,
QUB_FitCurves,
QUB_Task_AppendOverlap,
QUB_GUIInterface,
QUB_Profiles;


constructor TQUB_QVO_QVD_ActiveSelList.Create(AOwner :TComponent);
begin
  inherited;

  DefaultRowHeight              := 16;
  FixedRows                     := 0;
  FixedCols                     := 0;
  RowCount                      := 0;
  ColCount                      := 0;
  Options                       := [goFixedVertLine, goFixedHorzLine,
                                    goVertLine,      goHorzLine,
                                    goRowSizing,     goColSizing,
                                    goRowSelect,     goThumbTracking];
  ScrollBars                    := ssVertical;

  OnMouseUp                     := prOnMouseUp;
  OnKeyUp                       := prOnKeyUp;
  OnDrawCell                    := prOnDrawCell;

  prPopup                       := TPopupMenu.Create(Self);
  PopupMenu                     := prPopup;

  itemSelectAll                 := TMenuItem.Create(Self);
  itemSelectAll.Caption         := 'Select all';
  itemSelectAll.OnClick         := prOnSelectAll;
  prPopup.Items.Add(itemSelectAll);

  itemSelectNone                := TMenuItem.Create(Self);
  itemSelectNone.Caption        := 'Select none';
  itemSelectNone.OnClick        := prOnSelectAll;
  prPopup.Items.Add(itemSelectNone);
end;

destructor TQUB_QVO_QVD_ActiveSelList.Destroy;
begin
  try
    List := nil;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited;

  if (Parent <> nil) and (ColCount >= 2) then begin
    ColWidths[1] := ClientWidth - ColWidths[0];
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prSetList(x :TQUB_QVO_QVD_SelectionList);
var
  i            :Integer;
begin
  if prList <> nil then begin
    FixedCols := 0;
    RowCount  := 0;
    ColCount  := 0;
    prList.Events.DelEvent('ItemAdded', prOnItemAdded);
    prList.Events.DelEvent('ItemDeleting', prOnItemDeleting);
    prList.Events.DelEvent('ItemChanged', prOnItemChanged);
    prList.Events.DelEvent('LabelChanged', prOnLabelChanged);
    prList.Events.DelEvent('ActiveChanged', prOnActiveChanged);

    prList.ActiveRange := QUB_FirstLast(-1, -1);
    for i := 0 to prList.Items.Count - 1 do begin
      prList[i].Active := True;
    end;
  end;

  prList := x;

  if prList <> nil then begin
    RowCount  := List.Items.Count;
    if RowCount > 0 then begin
      ColWidths[0] := 20;
      ColCount  := 2;
      FixedCols := 1;
      if Parent <> nil then ColWidths[1] := ClientWidth - ColWidths[0];
    end;
    for i := 0 to List.Items.Count - 1 do begin
      prRelabelItem(i);
    end;
    prList.Events.AddEvent('ItemAdded', prOnItemAdded);
    prList.Events.AddEvent('ItemDeleting', prOnItemDeleting);
    prList.Events.AddEvent('ItemChanged', prOnItemChanged);
    prList.Events.AddEvent('LabelChanged', prOnLabelChanged);
    prList.Events.AddEvent('ActiveChanged', prOnActiveChanged);
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prRelabelItem(i :Integer);
var
  Sel     :TQUB_QVO_QVD_Selection;
begin
  Sel     := List[i];
  if Sel.Lbl = '' then begin
    Cells[1, i] := Format('%d - %d', [Sel.Sel.First, Sel.Sel.Last]);
  end else begin
    Cells[1, i] := Format('%s: %d - %d', [Sel.Lbl, Sel.Sel.First, Sel.Sel.Last]);
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnSelectAll(Sender :TObject);
var
  Active      :Boolean;
  i           :Integer;
begin
  if List = nil then Exit;
  Active := Sender = itemSelectAll;

  for i := 0 to List.Items.Count - 1 do begin
    List[i].Active := Active;
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnItemAdded(Sender :TObject);
begin
  if List.EventIndex = (List.Items.Count - 1) then begin
    RowCount := List.Items.Count;
  end else begin
    InsertRow(List.EventIndex);
  end;
  if RowCount = 1 then begin
    ColCount  := 2;
    FixedCols := 1;
    ColWidths[0] := 20;
    if Parent <> nil then ColWidths[1] := ClientWidth - ColWidths[0];
  end;

  prRelabelItem(List.EventIndex);
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnItemDeleting(Sender :TObject);
begin
  if List.EventIndex = (List.Items.Count - 1) then begin
    if List.Items.Count = 1 then begin
      FixedCols := 0;
      ColCount  := 0;
    end;
    RowCount := List.Items.Count - 1;
  end else begin
    DeleteRow(List.EventIndex);
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnItemChanged(Sender :TObject);
begin
  if (List.ItemIndex >= 0) and (List.ItemIndex < List.Items.Count) then begin
    Row := List.ItemIndex;
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnLabelChanged(Sender :TObject);
begin
  prRelabelItem(List.EventIndex);
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnActiveChanged(Sender :TObject);
begin
  Cells[0, List.EventIndex] := '';
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ACol, ARow  :Integer;
  CanSelect   :Boolean;
begin
  MouseToCell( X, Y, ACol, ARow );
  prOnSelectCell(Sender, ACol, ARow, CanSelect);
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnKeyUp(Sender :TObject; var Key : Word; Shift: TShiftState);
var
  CanSelect  :Boolean;
  k          :Word;
begin
  k          := Key;
  inherited;
  if k = Ord(' ') then begin
    prOnSelectCell(Sender, 0, Row, CanSelect);
  end else begin
    prOnSelectCell(Sender, 1, Row, CanSelect);
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnSelectCell(Sender: TObject; ACol, ARow: Longint; var CanSelect: Boolean);
var
  Sel :TQUB_FirstLast;
begin
  if (List = nil) or (ARow < 0) or (ARow >= List.Items.Count) then Exit;

  if ACol = 0 then begin
    List[ARow].Active := not List[ARow].Active;
  end else begin
    Sel             := List[ARow].Sel;
    if List.ActiveRange.First >= 0 then begin
      Sel.Last        := Min(Sel.Last, Sel.First + List.ActiveRange.Last);
      Sel.First       := Min(Sel.Last, Sel.First + List.ActiveRange.First);
    end;
    if Sel.First <= Sel.Last then begin
      Data.MoveSelectionIntoWindow( Sel, Sel.First );
    end;
    List.Choose(ARow);
  end;
end;

procedure TQUB_QVO_QVD_ActiveSelList.prOnDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  r          :TRect;
begin
  if List = nil then Exit;

  if (ACol = 0) and (ARow < prList.Items.Count) then begin
    r        := Rect;
    InflateRect(r, 0, -((r.Bottom - r.Top + 1) - 15) div 2);
    if List[ARow].Active then begin
      DrawFrameControl(Canvas.Handle, r, DFC_BUTTON, DFCS_BUTTONCHECK or DFCS_CHECKED);
    end else begin
      DrawFrameControl(Canvas.Handle, r, DFC_BUTTON, DFCS_BUTTONCHECK);
    end;
  end;
end;


// *****************************************************************************
constructor TQUB_QVO_QVD_DataListMirror.Create;
begin

  inherited;

  OnItemChange   := prOnItemChange;
  OnItemDblClick := prOnItemDblClick;

end;

destructor TQUB_QVO_QVD_DataListMirror.Destroy;
begin

  try
    OnItemChange   := prOnItemChange;
    OnItemDblClick := prOnItemDblClick;
  finally
    inherited;
  end;

end;

procedure TQUB_QVO_QVD_DataListMirror.prOnItemChange(Sender :TObject);
begin
  if ItemIndex >= 0 then ViewportData.ListObjects.SelQVOAt(ItemIndex);
end;

procedure TQUB_QVO_QVD_DataListMirror.prOnItemDblClick(Sender :TObject);
begin
  ViewportData.Properties;
end;

procedure TQUB_QVO_QVD_DataListMirror.OnOpenData(Sender :TObject);
begin
  Items.Add(ViewportData.ListObjects.Items[Items.Count]);
end;

procedure TQUB_QVO_QVD_DataListMirror.OnClosingData(Sender :TObject);
var
  i            :Integer;
begin
  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    if ViewportData.ListObjects[i] = Sender then begin
      Items.Delete(i);
      Exit;
    end;
  end;
end;

procedure TQUB_QVO_QVD_DataListMirror.OnChangeData(Sender :TObject);
var
  NewItemIndex     :Integer;
begin
  NewItemIndex     := ViewportData.ListObjects.ItemIndex;
  if (NewItemIndex >= 0) and (NewItemIndex <> ItemIndex) then ItemIndex := NewItemIndex;
end;


// *****************************************************************************
constructor TQUB_QVO_QVD_ListZoomState.Create(List :TQUB_QVO_QVD_SelectionList);
var
  i                 :Integer;
begin
  prList            := List;
  prActiveRange     := List.ActiveRange;

  SetLength(prActive, List.Items.Count);
  for i := 0 to List.Items.Count - 1 do begin
    prActive[i]     := List[i].Active;
  end;
end;

procedure TQUB_QVO_QVD_ListZoomState.Apply;
var
  i                   :Integer;
begin
  for i := 0 to Min(Length(prActive), prList.Items.Count) - 1 do begin
    prList[i].Active  := prActive[i];
  end;

  prList.ActiveRange  := prActiveRange;
end;


constructor TQUB_QVO_QVD_ListZoomStack.Create;
begin
  prList            := nil;
  prStack           := TList.Create;
end;

destructor TQUB_QVO_QVD_ListZoomStack.Destroy;
begin
  try
    Clear;
    prList.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_ListZoomStack.Push;
begin
  if List <> nil then prStack.Add( TQUB_QVO_QVD_ListZoomState.Create(List) );
end;

procedure TQUB_QVO_QVD_ListZoomStack.Pop;
begin
  if (List = nil) or (prStack.Count = 0) then Exit;

  TQUB_QVO_QVD_ListZoomState(prStack[prStack.Count - 1]).Apply;
  TQUB_QVO_QVD_ListZoomState(prStack[prStack.Count - 1]).Free;
  prStack.Delete( prStack.Count - 1 );
end;

procedure TQUB_QVO_QVD_ListZoomStack.Clear;
var
  i            :Integer;
begin
  for i := 0 to prStack.Count - 1 do begin
    TQUB_QVO_QVD_ListZoomState(prStack[i]).Free;
  end;
  prStack.Clear;
end;

procedure TQUB_QVO_QVD_ListZoomStack.prSetList(x :TQUB_QVO_QVD_SelectionList);
begin
  if prList <> nil then begin
    prList.Events.DelEvent('ItemAdded', prOnItemAdded);
    prList.Events.DelEvent('ItemDeleting', prOnItemDeleting);
  end;

  prList := x;
  Clear;

  if prList <> nil then begin
    prList.Events.AddEvent('ItemAdded', prOnItemAdded);
    prList.Events.AddEvent('ItemDeleting', prOnItemDeleting);
  end;
end;

function TQUB_QVO_QVD_ListZoomStack.prGetEmpty: Boolean;
begin
  Result := prStack.Count = 0;
end;

procedure TQUB_QVO_QVD_ListZoomStack.prOnItemAdded(Sender :TObject);
begin
  Clear;
end;

procedure TQUB_QVO_QVD_ListZoomStack.prOnItemDeleting(Sender :TObject);
begin
  Clear;
end;


// *****************************************************************************

constructor TQUB_QVO_QVD_Overlap.Create(AOwner :TComponent);
begin
  inherited;

  prChannels              := TList.Create;

  prResampler             := TQUB_QVO_QVD_ListResampler.Create;
  prResampler.ChannelIndex := -1; // all
  prResampler.OnNeedsResample := prOnNeedsResample;
  prResampler.OnYRangeChanged := prOnYRangeChanged;

  prPopup                 := TPopupMenu.Create(Self);
  prPopup.OnPopup           := prOnPopup;

  itemSelectAll           := TMenuItem.Create(Self);
  itemSelectAll.Caption     := 'Select all';
  itemSelectAll.OnClick     := prOnClickSelectAll;
  prPopup.Items.Add(itemSelectAll);

  OnMouseMove             := prOnMouseMove;
end;

destructor TQUB_QVO_QVD_Overlap.Destroy;
begin
  try
    Data                  := nil;
    prChannels.Free;
    prResampler.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_Overlap.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
const
  MARGIN_W       = 6;
  MARGIN_H       = 6;
  MIN_TRACEWIDTH = 30;
  MIN_TRACEHEIGHT = 10;
var
  TraceWidth     :Integer;
  TraceHeight    :Integer;
  EffHeight      :Integer;
  EffWidth       :Integer;
  i              :Integer;
begin
  inherited;

  if Parent <> nil then begin
    EffHeight := ClientHeight;
    EffWidth  := ClientWidth;
  end else begin
    EffHeight := Height;
    EffWidth  := Width;
  end;

  TraceWidth     := Max(MIN_TRACEWIDTH, EffWidth - 2 * MARGIN_W);

  if ChannelCount > 0 then begin
    TraceHeight    := Max(MIN_TRACEHEIGHT, (EffHeight - (ChannelCount+1)*MARGIN_H) div ChannelCount);
  end else begin
    TraceHeight    := Max(MIN_TRACEHEIGHT, (EffHeight - (1+1)*MARGIN_H));
  end;

  for i := 0 to ChannelCount - 1 do begin
    Channels[i].SetBounds(MARGIN_W, (i+1)*MARGIN_H + i*TraceHeight, TraceWidth, TraceHeight);
  end;

  prResampler.SampleCount := TraceWidth;
end;

procedure TQUB_QVO_QVD_Overlap.Paint;
begin
  inherited;
end;

procedure TQUB_QVO_QVD_Overlap.prSetLblStatus(x :TLabel);
var
  i               :Integer;
begin
  for i := 0 to ChannelCount - 1 do begin
    Channels[i].LblStatus := x;
  end;
  prLblStatus     := x;
end;

procedure TQUB_QVO_QVD_Overlap.prSetLblStatusActive(x :TLabel);
var
  i               :Integer;
begin
  for i := 0 to ChannelCount - 1 do begin
    Channels[i].LblStatusActive := x;
  end;
  prLblStatusActive     := x;
end;

procedure TQUB_QVO_QVD_Overlap.prSetData(x :TQUB_QVO_QVD);
var
  i               :Integer;
  NewChannelCount :Integer;
  Channel         :TQUB_QVO_QVD_Overlap_Channel;
begin
  if x = prData then Exit;

  NewChannelCount := 0;
  if x <> nil then NewChannelCount := x.DataObject.ADChannelCount;

  if prData <> nil then begin
    List := nil;

    for i := ChannelCount - 1 downto NewChannelCount do begin
      Channel := Channels[i];
      prChannels.Delete(i);
      Channel.Free;
    end;
  end;

  prData := x;

  if prData <> nil then begin
    for i := 0 to NewChannelCount - 1 do begin
      if i >= ChannelCount then begin
        Channel              := TQUB_QVO_QVD_Overlap_Channel.Create(Self);
        Channel.Parent       := Self;
        Channel.LblStatus    := lblStatus;
        Channel.LblStatusActive := lblStatusActive;
        Channel.ChannelIndex := i;
        Channel.PopupMenu    := prPopup;
        Channel.OnBeforeZoom := prOnBeforeZoomChannel;
        prChannels.Add( Channel );
      end else begin
        Channel              := Channels[i];
      end;

      Channel.Resampler      := prResampler;
    end;
  end;

  prResampler.Data           := prData;

  SetBounds(Left, Top, Width, Height);
end;

procedure TQUB_QVO_QVD_Overlap.prSetList(x :TQUB_QVO_QVD_SelectionList);
var
  i            :Integer;
begin
  if x = prList then Exit;
  prList       := x;

  for i := 0 to ChannelCount - 1 do begin
    Channels[i].List := x;
  end;
  prResampler.List := x;
end;

function TQUB_QVO_QVD_Overlap.prGetChannelCount: Integer;
begin
  Result := prChannels.Count;
end;

function TQUB_QVO_QVD_Overlap.prGetChannel(Ix :Integer): TQUB_QVO_QVD_Overlap_Channel;
begin
  Result := TQUB_QVO_QVD_Overlap_Channel(prChannels[Ix]);
end;

procedure TQUB_QVO_QVD_Overlap.prOnPopup(Sender :TObject);
begin
  itemSelectAll.Visible := prResampler.SelCount > 0;
end;

procedure TQUB_QVO_QVD_Overlap.prOnClickSelectAll(Sender :TObject);
var
  i                :Integer;
begin
  ZoomStack.Clear;

  List.ActiveRange := QUB_FirstLast(-1, -1);
  for i := 0 to List.Items.Count - 1 do begin
    List[i].Active := True;
  end;
end;

procedure TQUB_QVO_QVD_Overlap.prOnBeforeZoomChannel(Sender :TObject);
begin
  ZoomStack.Clear;
end;

procedure TQUB_QVO_QVD_Overlap.prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if LblStatus <> nil then LblStatus.Caption := '';
  if LblStatusActive <> nil then LblStatusActive.Caption := '';
end;

procedure TQUB_QVO_QVD_Overlap.prOnNeedsResample(Sender :TObject; Channel, Sel :Integer);
var
  c           :Integer;
begin
  if Channel >= 0 then begin
    Channels[Channel].DoOnNeedsResample(Sender, Sel);
  end else begin
    for c := 0 to ChannelCount - 1 do begin
      Channels[c].DoOnNeedsResample(Sender, Sel);
    end;
  end;
end;

procedure TQUB_QVO_QVD_Overlap.prOnYRangeChanged(Sender :TObject; Channel :Integer);
var
  c           :Integer;
begin
  if Channel >= 0 then begin
    Channels[Channel].DoOnYRangeChanged(Sender);
  end else begin
    for c := 0 to ChannelCount - 1 do begin
      Channels[c].DoOnYRangeChanged(Sender);
    end;
  end;
end;

// *****************************************************************************
constructor TQUB_QVO_QVD_Overlap_Channel.Create(AOwner :TComponent);
begin
  inherited;

  ControlStyle          := ControlStyle + [csOpaque];

  prBitmap              := Graphics.TBitmap.Create;
  prBitmapCanvas        := prBitmap.Canvas;
  prBitmap.Width        := Width;
  prBitmap.Height       := Height;

  prChannelIndex        := 0;
  prCanMarquee          := True;
  prShowMarquee         := True;
  prYfactor             := 1;

  OnMouseDown           := prOnMouseDown;
  OnMouseUp             := prOnMouseUp;
  OnMouseMove           := prOnMouseMove;
end;

destructor TQUB_QVO_QVD_Overlap_Channel.Destroy;
begin
  try
    List                := nil;
    Resampler           := nil;
    prBitmap.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if Height <> AHeight then begin
    prAllNeedRedraw     := True;
    prMarqueeInvalid    := ShowMarquee;
    Invalidate;
  end;
  if Width <> AWidth then begin
    prMarqueeInvalid    := ShowMarquee;
    Invalidate;
  end;

  inherited;

  prBitmap.Width        := AWidth;
  prBitmap.Height       := AHeight;
end;

procedure SortRect(var A, B :TRect);
begin
  B          := A;
  if A.Left > A.Right then begin
    B.Left   := A.Right;
    B.Right  := A.Left;
  end;
  if A.Top > A.Bottom then begin
    B.Top    := A.Bottom;
    B.Bottom := A.Top;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.Paint;
var
  i              :Integer;
  SortedMarquee  :TRect;
  DidResample    :Boolean;
  Curve          :INode;
  yratio, ymax   :Double;
  Data           :TQUB_QVO_QVD;
begin
  if (Parent <> nil) then begin
    try
      if not (csDestroying in ComponentState) then begin
        try
          DidResample := Resampler.DoResample;

          if Mode = QUB_QVO_QVD_OVERLAP_OVERLAP then begin

            if prAllNeedRedraw then begin
              prRedrawTraces;
            end else begin
              for i := 0 to Length(prNeedsRedraw) - 1 do begin
                if prNeedsRedraw[i] and (i <> List.ItemIndex) then prRedrawTrace(i);
              end;
              if (List <> nil) and (List.ItemIndex >= 0) and (List.ItemIndex < List.Items.Count) and prNeedsRedraw[List.ItemIndex] then prRedrawTrace(List.ItemIndex);
            end;

          end else if Mode = QUB_QVO_QVD_OVERLAP_AVERAGE then begin

            if DidResample or prJustChangedMode then begin
              prRecalcAvgStd;
              prAllNeedRedraw := True;
            end;

            if prAllNeedRedraw then begin
              prRedrawAvgStd;
            end;

          end else if Mode = QUB_QVO_QVD_OVERLAP_SUM then begin

            if DidResample or prJustChangedMode then begin
              prRecalcSum;
              prAllNeedRedraw := True;
            end;

            if prAllNeedRedraw then begin
              prRedrawAvgStd;
            end;

          end else if Mode = QUB_QVO_QVD_OVERLAP_FUNCTION then begin

            if DidResample or prJustChangedMode then begin
              prRecalcFunc;
              prAllNeedRedraw := True;
            end;

            if prAllNeedRedraw then begin
              prRedrawAvgStd;
            end;

          end;

          Canvas.Lock;
          Canvas.Draw(0, 0, prBitmap);

          if prFitNeedsRedraw and (prFitData <> nil) and (Length(prAvg) > 0) then begin
            Canvas.Pen.Color := clRed;
            yratio := prYfactor;
            ymax   := prYmax;
            Curve := prFitData['F'];
            while not Curve.IsNull do begin
              Canvas.MoveTo(0, Round(yratio * (ymax - Double(Curve.Data[0]))));
              for i := 1 to Curve.DataCount - 1 do begin
                Canvas.LineTo(i, Round(yratio * (ymax - Double(Curve.Data[i]))));
              end;

              Curve := Curve.Sibling;
            end;
          end;

          if prMarqueeInvalid then prActiveToMarquee;
          if prMarqueeing or prShowMarquee then begin
            SortRect(prMarquee, SortedMarquee);
            SortedMarquee.Right  := SortedMarquee.Right + 1;
            SortedMarquee.Bottom := SortedMarquee.Bottom + 1;
            Data := ViewportData.CurrData;
            if Data <> nil then begin
              Canvas.Brush.Color := Data.ColorSel;
            end else begin
              Canvas.Brush.Color := clBlue;
            end;
            Canvas.FrameRect(SortedMarquee);
            SortedMarquee.Right  := SortedMarquee.Right - 1;
            SortedMarquee.Bottom := SortedMarquee.Bottom - 1;
          end;

          prJustChangedMode := False;
        finally
          Canvas.Unlock;
        end;
      end;
    finally
      inherited;
    end;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.DoOnNeedsResample(Sender :TObject; Sel :Integer);
begin
  if (Sel >= 0) then begin
    if Resampler.SelCount = Length(prNeedsRedraw) then prNeedsRedraw[Sel] := True;
  end else begin
    prAllNeedRedraw := True;
  end;

  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.PixToData(x, y :Integer; var Sec, Val :Double);
begin
  Sec := Resampler.Data.Sampling * (Resampler.Xoffset + Floor(x * Resampler.Xfactor));
  Val := prYmax - y / prYfactor;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.DataToPix(Sec, Val :Double; var x, y :Integer);
begin
  x   := Round( (Sec / Resampler.Data.Sampling - Resampler.Xoffset) / Resampler.Xfactor );
  y   := Round( prYfactor * (prYmax - Val) );
end;

procedure TQUB_QVO_QVD_Overlap_Channel.DoOnYRangeChanged(Sender :TObject);
begin
  prAllNeedRedraw := True;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.ShowFit;
begin
  if (Length(prAvg) = 0) or (Mode = QUB_QVO_QVD_OVERLAP_OVERLAP) then Exit;

  prUpdateFitData;
  FitDlog.SetState(ProfileMgr['Fitter-Overlap'].Active);
  FitDlog.Fitter.Events.AddEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.ShowModal;
  FitDlog.Fitter.Events.DelEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  ProfileMgr['Fitter-Overlap'].Active := FitDlog.GetState;
  FitDlog.Data.UseData(nil);
end;

procedure TQUB_QVO_QVD_Overlap_Channel.ClearFit;
begin
  if prFitData <> nil then begin
    prFitData := nil;
    Invalidate;
  end;
  if prVisDataSet <> nil then begin
    prVisDataSet.Free;
    prVisDataSet := nil;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prUpdateFitData;
var
  i, N                     :Integer;
  X, Y, F                  :INode;
  DataSet                  :TQUB_FitDataSet;
begin
  N                        := Length(prAvg);
  prFitData                := TNode.Create('DataSet');
  X                        := prFitData['X'];
  X.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
  Y                        := prFitData['Y'];
  Y.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
  F                        := prFitData['F'];
  F.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );

  for i := 0 to N - 1 do begin
    X.Data[i]              := Resampler.Data.Sampling * Floor(i * Resampler.Xfactor);
    Y.Data[i]              := prAvg[i];
  end;

  FitDlog.Data.AbsX := False;
  FitDlog.Data.DisplayStyle := QUBFITS_DISPLAY_DEFAULT;
  DataSet  := TQUB_FitDataSet.Create(N);
  DataSet.AddVar('X', TPADouble(X.DataPtr));
  DataSet.AddVar('Y', TPADouble(Y.DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prVisDataSet <> nil then prVisDataSet.Free;
  prVisDataSet               := TQUB_FitDataSet.Create;
  prVisDataSet.Assign(DataSet);
  prVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  DataSet.Free;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnCurvesChanged(Sender :TObject);
var
  X        :INode;
  i, N     :Integer;
  s        :Double;
begin
  X        := prFitData['X'];
  N        := prVisDataSet.N;
  s        := FitDlog.Data.UseDataSet.Xoff;
  for i := 0 to N - 1 do begin
    prVisDataSet.XX[i]  := X.Data[i] - s;
  end;
  FitDlog.Fitter.Curve.EvalArrayAll(prVisDataSet.Data,
                                TPADouble(prFitData['F'].DataPtr));
  i := 0;
  while i < N do begin
    if X.Data[i] < S then break;
    prFitData['F'].Data[i]     := 0.0;
    Inc(i);
  end;
  prFitNeedsRedraw             := True;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prSetChannelIndex(x :Integer);
begin
  if x = prChannelIndex then Exit;

  prChannelIndex          := x;
  prAllNeedRedraw := True;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prSetList(x :TQUB_QVO_QVD_SelectionList);
begin
  if x = prList then Exit;

  if prList <> nil then begin
    ClearFit;

    prList.Events.DelEvent('ItemAdded', prOnItemAdded);
    prList.Events.DelEvent('ItemDeleting', prOnItemDeleting);
    prList.Events.DelEvent('ItemChanged', prOnItemChanged);
    prList.Events.DelEvent('ActiveChanged', prOnActiveChanged);
    prList.Events.DelEvent('ActiveRangeChanged', prOnActiveRangeChanged);

    SetLength(prNeedsRedraw, 0);
  end;

  prList                  := x;

  if prList <> nil then begin
    prList.Events.AddEvent('ItemAdded', prOnItemAdded);
    prList.Events.AddEvent('ItemDeleting', prOnItemDeleting);
    prList.Events.AddEvent('ItemChanged', prOnItemChanged);
    prList.Events.AddEvent('ActiveChanged', prOnActiveChanged);
    prList.Events.AddEvent('ActiveRangeChanged', prOnActiveRangeChanged);

    SetLength(prNeedsRedraw, prList.Items.Count);
  end;

  prDrawnItemIndex  := -1;

  prAllNeedRedraw   := True;
  prMarqueeInvalid  := ShowMarquee;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prSetShowMarquee(x :Boolean);
begin
  prShowMarquee     := x;
  prMarqueeInvalid  := x;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prSetMode(x :TQUB_QVO_QVD_OverlapMode);
begin
  if prMode = x then Exit;
  if (x = QUB_QVO_QVD_OVERLAP_FUNCTION) and (prFunc = nil) then Exit;
  
  prMode := x;
  prAllNeedRedraw   := True;
  // CanMarquee        := Mode = QUB_QVO_QVD_OVERLAP_OVERLAP;
  prMarqueeInvalid  := ShowMarquee;
  prJustChangedMode := True;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prSetFunc(x :PPyObject);
begin
  if prFunc = x then Exit;

  if prFunc <> nil then begin
    GetPythonEngine.Py_DECREF( prFunc );
  end;

  prFunc := x;

  if prFunc <> nil then begin
    GetPythonEngine.Py_INCREF( prFunc );
  end else begin
    if Mode = QUB_QVO_QVD_OVERLAP_FUNCTION then Mode := QUB_QVO_QVD_OVERLAP_OVERLAP;
  end;

  if Mode = QUB_QVO_QVD_OVERLAP_FUNCTION then begin
    prAllNeedRedraw   := True;
    prJustChangedMode := True;
    Invalidate;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnItemAdded(Sender :TObject);
var
  i, p       :Integer;
begin
  p := List.EventIndex;

  SetLength(prNeedsRedraw, List.Items.Count);
  for i := p + 1 to List.Items.Count - 1 do begin
    prNeedsRedraw[i] := prNeedsRedraw[i-1];
  end;

  if prDrawnItemIndex >= p then Inc(prDrawnItemIndex);

  prNeedsRedraw[p]   := True;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnItemDeleting(Sender :TObject);
var
  i, p       :Integer;
begin
  p          := List.EventIndex;
  for i := p to List.Items.Count - 2 do begin
    prNeedsRedraw[i] := prNeedsRedraw[i+1];
  end;
  SetLength(prNeedsRedraw, List.Items.Count - 1);

  if prDrawnItemIndex = p then begin
    prDrawnItemIndex := -1;
  end else if prDrawnItemIndex > p then begin
    Dec(prDrawnItemIndex);
  end;

  prAllNeedRedraw := True;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnItemChanged(Sender :TObject);
begin
  if List.ItemIndex >= 0 then begin
    if prDrawnItemIndex >= 0 then begin
      prNeedsRedraw[ prDrawnItemIndex ] := True;
    end;
    prNeedsRedraw[ List.ItemIndex ] := True;
    Invalidate;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if CanMarquee then begin
    prMarqueeStart.X    := X;
    prMarqueeStart.Y    := Y;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
const
  PIX_TOL             = 4.0;
var
  Sec, Val            :Double;
  i, prox_i           :Integer;
  prox, prox_min      :Double;
  Sel                 :TQUB_FirstLast;
begin
  if Resampler.SelCount = 0 then Exit;
  X  := Max(0, X);
  X  := Min(Width-1, X);
  Y  := Max(0, Y);
  Y  := Min(Height-1, Y);

  PixToData(X, Y, Sec, Val);

  if prMarqueeing then begin

    prMarqueeing      := False;
    prMarqueeToActive;          // redo list.activerange, list[i].active
    // then in active-change-listeners: set need-recalc-marquee and invalidate
    // also in setbounds:                "   "                   "   "
    // then in paint: if n-r-c, set it to activerange X yminmax in active range/traces
    Invalidate;

  end else if Mode = QUB_QVO_QVD_OVERLAP_OVERLAP then begin

    prox_i              := -1;
    prox_min            := -1; // whatever
    for i := 0 to Resampler.SelCount - 1 do begin
      if Resampler.SelSampleCount[i] > X then begin
        prox            := Abs(Val - Resampler.Samples[ChannelIndex][i][X]);
        if (prox_i < 0) or (prox < prox_min) then begin
          prox_i        := i;
          prox_min      := prox;
        end;
      end;
    end;

    if prox_i >= 0 then begin
      prox_min := prox_min * prYfactor;
      if prox_min < PIX_TOL then begin
        Sel             := List[prox_i].Sel;
        if List.ActiveRange.First >= 0 then begin
          Sel.Last        := Min(Sel.Last, Sel.First + List.ActiveRange.Last);
          Sel.First       := Min(Sel.Last, Sel.First + List.ActiveRange.First);
        end;
        if Sel.First <= Sel.Last then begin
          Resampler.Data.MoveSelectionIntoWindow( Sel, Sel.First );
        end;
        List.Choose(prox_i);
      end;
    end;

  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
const
  CLICK_TOL           = 4.0;
var
  Sec, Val            :Double;
begin
  if Resampler.SelCount = 0 then Exit;
  X  := Max(0, X);
  X  := Min(Width-1, X);
  Y  := Max(0, Y);
  Y  := Min(Height-1, Y);

  PixToData(X, Y, Sec, Val);
  if LblStatus <> nil then begin
    LblStatus.Caption := Format('Mouse: (%-10.6g, %-10.6g)', [Sec, Val]);
  end;

  if LblStatusActive <> nil then begin
    if (Mode = QUB_QVO_QVD_OVERLAP_AVERAGE) and (Length(prAvg) = Resampler.SampleCount) then begin
      LblStatusActive.Caption := Format('Average: (%-10.6g, %-10.6g)', [Sec, prAvg[X]]);
    end else if (Mode = QUB_QVO_QVD_OVERLAP_SUM) and (Length(prAvg) = Resampler.SampleCount) then begin
      LblStatusActive.Caption := Format('Sum: (%-10.6g, %-10.6g)', [Sec, prAvg[X]]);
    end else if (Mode = QUB_QVO_QVD_OVERLAP_FUNCTION) and (Length(prAvg) = Resampler.SampleCount) then begin
      LblStatusActive.Caption := Format('Function: (%-10.6g, %-10.6g)', [Sec, prAvg[X]]);
    end else if (List.ItemIndex >= 0) and (X >= 0) and (Resampler.SelSampleCount[List.ItemIndex] > X) then begin
      LblStatusActive.Caption := Format('Trace %d: (%-10.6g, %-10.6g)', [List.ItemIndex+1, Sec, Resampler.Samples[ChannelIndex][List.ItemIndex][X]]);
    end;
  end;
  
  if CanMarquee and (ssLeft in Shift) and (not prMarqueeing) and (Max(Abs(prMarqueeStart.X - X), Abs(prMarqueeStart.Y - Y)) > CLICK_TOL) then begin
    prMarqueeing      := True;
    prMarquee.TopLeft := prMarqueeStart;
  end;

  if prMarqueeing then begin
    prMarquee.Right   := X;
    prMarquee.Bottom  := Y;
    Invalidate;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnActiveChanged(Sender :TObject);
begin
  prNeedsRedraw[ List.EventIndex ] := True;
  prMarqueeInvalid := ShowMarquee;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prOnActiveRangeChanged(Sender :TObject);
begin
  prMarqueeInvalid := ShowMarquee;
  Invalidate;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prMarqueeToActive;
var
  DataSel       :TQUB_FirstLast;
  Ymin, Ymax    :Double;
  i, s          :Integer;
  Active        :Boolean;
  Val           :Single;
  DVal          :Double;
  SortedMarquee :TRect;
begin
  if Resampler.SelCount = 0 then Exit;

  if Assigned(OnBeforeZoom) then OnBeforeZoom(Self);

  SortRect(prMarquee, SortedMarquee);
  DataSel.First := Floor(SortedMarquee.Left * Resampler.Xfactor) + Resampler.Xoffset;
  DataSel.Last  := Max( Floor(SortedMarquee.Right * Resampler.Xfactor) + Resampler.Xoffset,
                        Floor((SortedMarquee.Right + 1) * Resampler.Xfactor) + Resampler.Xoffset - 1 );
  List.ActiveRange := DataSel;

  PixToData(SortedMarquee.Left, SortedMarquee.Top, DVal, Ymax);
  PixToData(SortedMarquee.Right, SortedMarquee.Bottom, DVal, Ymin);

  if Mode = QUB_QVO_QVD_OVERLAP_OVERLAP then begin
    for i := 0 to Resampler.SelCount - 1 do begin
      if (not Resampler.ActiveOnly) or List[i].Active then begin
        Active      := False;
        for s := SortedMarquee.Left to Min(SortedMarquee.Right, Resampler.SelSampleCount[i] - 1) do begin
          Val       := Resampler.Samples[ChannelIndex][i][s];
          if (Val >= Ymin) and (Val <= Ymax) then begin
            Active  := True;
            Break;
          end;
        end;
        List[i].Active := Active;
      end;
    end;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prActiveToMarquee;
var
  Ymin, Ymax    :Single;
  i, s          :Integer;
  Val           :Single;
begin
  if List = nil then begin
    SetRect(prMarquee, 0, 0, Width-1, Height-1);
    Exit;
  end;

  if List.ActiveRange.First < 0 then begin
    prMarquee.Left  := 0;
    prMarquee.Right := Width - 1;
  end else begin
    prMarquee.Left  := Round((List.ActiveRange.First - Resampler.Xoffset) / Resampler.Xfactor);
    prMarquee.Right := Round((List.ActiveRange.Last + 1 - Resampler.Xoffset) / Resampler.Xfactor) - 1;
  end;

  Ymin          := Resampler.ChannelYmax[ChannelIndex];
  Ymax          := Resampler.ChannelYmin[ChannelIndex];
  for i := 0 to Resampler.SelCount - 1 do begin
    if List[i].Active then begin
      for s := prMarquee.Left to Min(prMarquee.Right, Resampler.SelSampleCount[i] - 1) do begin
        Val     := Resampler.Samples[ChannelIndex][i][s];
        if Val < Ymin then Ymin := Val;
        if Val > Ymax then Ymax := Val;
      end;
    end;
  end;

  if Resampler.SelCount = 0 then begin
    prMarquee.Top    := 0;
    prMarquee.Bottom := Height - 1;
  end else begin
    DataToPix(0.0, Ymax, i, prMarquee.Top);
    DataToPix(0.0, Ymin, i, prMarquee.Bottom);
  end;
  
  prMarqueeInvalid := False;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prRedrawTrace(Ix :Integer);
var
  p                       :Integer;
  ymax, yratio            :Single;
  Data                    :TQUB_QVO_QVD;
begin

  Data := ViewportData.CurrData;

  if Ix >= List.Items.Count then Exit;
  if Resampler.SelSampleCount[Ix] = 0 then Exit;
  if Resampler.ActiveOnly and not List[Ix].Active then Exit;

  if Ix = List.ItemIndex then begin
    prBitmapCanvas.Pen.Color := clRed;
    prDrawnItemIndex         := Ix;
  end else if List[Ix].Active then begin
    if Data <> nil then begin
      prBitmapCanvas.Pen.Color := Data.ColorData[ChannelIndex];
    end else begin
      prBitmapCanvas.Pen.Color := clBlack;
    end;
  end else begin
    prBitmapCanvas.Pen.Color := clLtGray;
  end;

  ymax                    := prYmax;
  yratio                  := prYfactor;

  prBitmapCanvas.MoveTo(0, Round(yratio * (ymax - Resampler.Samples[ChannelIndex][Ix][0])));
  for p := 1 to Resampler.SelSampleCount[Ix] - 1 do begin
    prBitmapCanvas.LineTo(p, Round(yratio * (ymax - Resampler.Samples[ChannelIndex][Ix][p])));
  end;

  prNeedsRedraw[Ix]       := False;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prRedrawTraces;
var
  i                       :Integer;
  Rect                    :TRect;
  ymin, ymax, ymargin     :Single;
  Data                    :TQUB_QVO_QVD;
begin

  Data := ViewportData.CurrData;

  Rect.Top                := 0;
  Rect.Left               := 0;
  Rect.Right              := Width;
  Rect.Bottom             := Height;
  if Data <> nil then begin
    prBitmapCanvas.Brush.Color := Data.ColorBack;
  end else begin
    prBitmapCanvas.Brush.Color := clWhite;
  end;
  prBitmapCanvas.FillRect(Rect);

  if List <> nil then begin
    ymin                    := Resampler.ChannelYmin[ChannelIndex];
    ymax                    := Resampler.ChannelYmax[ChannelIndex];
    if ymin = ymax then begin
      ymin    := ymin - 0.1;
      ymax    := ymax + 0.1;
    end else begin
      ymargin := 0.1 * (ymax - ymin);
      ymin    := ymin - ymargin;
      ymax    := ymax + ymargin;
    end;
    prYmin    := ymin;
    prYmax    := ymax;
    prYfactor := Height / (ymax - ymin);

    for i := 0 to List.Items.Count - 1 do begin
      if i <> List.ItemIndex then prRedrawTrace(i);
    end;
    if List.ItemIndex >= 0 then prRedrawTrace(List.ItemIndex);
  end;

  prAllNeedRedraw         := False;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prRecalcAvgStd;
var
  i, s        :Integer;
  x, ym, yx   :Single;
  ymargin     :Single;
  nn          :array of Integer;
begin
  if List = nil then Exit;

  SetLength(prAvg, Resampler.SampleCount);
  SetLength(prStd, Resampler.SampleCount);
  SetLength(nn,    Resampler.SampleCount);
  for s := 0 to Resampler.SampleCount - 1 do begin
    prAvg[s]       := 0.0;
    prStd[s]       := 0.0;
    nn[s]          := 0;
  end;

  for i := 0 to Resampler.SelCount - 1 do begin
    if ((not Resampler.ActiveOnly) or List[i].Active) and (Resampler.SelSampleCount[i] > 0) then begin
      for s := 0 to Resampler.SelSampleCount[i] - 1 do begin
        prAvg[s]   := prAvg[s] + Resampler.Samples[ChannelIndex][i][s];
        nn[s]      := nn[s] + 1;
      end;
    end;
  end;

  for s := 0 to Resampler.SampleCount - 1 do begin
    if nn[s] > 0 then prAvg[s] := prAvg[s] / nn[s];
  end;

  for i := 0 to Resampler.SelCount - 1 do begin
    if ((not Resampler.ActiveOnly) or List[i].Active) and (Resampler.SelSampleCount[i] > 0) then begin
      for s := 0 to Resampler.SelSampleCount[i] - 1 do begin
        x          := prAvg[s] - Resampler.Samples[ChannelIndex][i][s];
        prStd[s]   := prStd[s] + x * x;
      end;
    end;
  end;

  ym  := 1.0e20;
  yx  := -1.0e20;
  for s := 0 to Resampler.SampleCount - 1 do begin
    if nn[s] > 0 then begin
      prStd[s] := Sqrt(prStd[s] / nn[s]);
      ym := Min(ym, prAvg[s] - prStd[s]);
      yx := Max(yx, prAvg[s] + prStd[s]);
    end;
  end;

  if ym = yx then begin
    prYmin    := ym - 0.1;
    prYmax    := yx + 0.1;
  end else begin
    ymargin := 0.1 * (yx - ym);
    prYmin    := ym - ymargin;
    prYmax    := yx + ymargin;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prRecalcSum;
var
  i, s        :Integer;
  ym, yx      :Single;
  ymargin     :Single;
begin
  if List = nil then Exit;

  SetLength(prAvg, Resampler.SampleCount);
  SetLength(prStd, Resampler.SampleCount);
  for s := 0 to Resampler.SampleCount - 1 do begin
    prAvg[s]       := 0.0;
    prStd[s]       := 0.0;
  end;

  for i := 0 to Resampler.SelCount - 1 do begin
    if ((not Resampler.ActiveOnly) or List[i].Active) and (Resampler.SelSampleCount[i] > 0) then begin
      for s := 0 to Resampler.SelSampleCount[i] - 1 do begin
        prAvg[s]   := prAvg[s] + Resampler.Samples[ChannelIndex][i][s];
      end;
    end;
  end;

  ym  := 1.0e20;
  yx  := -1.0e20;
  for s := 0 to Resampler.SampleCount - 1 do begin
     ym := Min(ym, prAvg[s]);
     yx := Max(yx, prAvg[s]);
  end;

  if ym = yx then begin
    prYmin    := ym - 0.1;
    prYmax    := yx + 0.1;
  end else begin
    ymargin := 0.1 * (yx - ym);
    prYmin    := ym - ymargin;
    prYmax    := yx + ymargin;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prRecalcFunc;
var
  i, s        :Integer;
  x, ym, yx   :Single;
  ymargin     :Single;
  Args        :PPyObject;
  PyResult    :PPyObject;
begin
  if List = nil then Exit;

  ym  := 1.0e20;
  yx  := -1.0e20;

  PythonThread.BeginPython;
  with GetPythonEngine do begin
    // prep tuple of args
    Args      := PyTuple_New(Resampler.SelCount);

    SetLength(prAvg, Resampler.SampleCount);
    SetLength(prStd, Resampler.SampleCount);
    for s := 0 to Resampler.SampleCount - 1 do begin
      //fill tuple[i] with samples[channelindex][i][s] (or 0.0 as needed)
      for i := 0 to Resampler.SelCount - 1 do begin
        //if s > 0 then Py_DECREF( PyTuple_GetItem(Args, i) );
        if s < Resampler.SelSampleCount[i] then begin
          x := Resampler.Samples[ChannelIndex][i][s];
        end else begin
          x := 0;
        end;
        PyTuple_SetItem(Args, i, PyFloat_FromDouble(x));
      end;

      PyResult       := PyObject_CallObject(prFunc, Args);
      prAvg[s]       := PyFloat_AsDouble(PyResult);
      Py_DECREF( PyResult );

      if prAvg[s] < ym then ym := prAvg[s];
      if prAvg[s] > yx then yx := prAvg[s];

      prStd[s]       := 0.0;
    end;

    Py_DECREF( Args );
  end;
  PythonThread.EndPython;

  if ym = yx then begin
    prYmin    := ym - 0.1;
    prYmax    := yx + 0.1;
  end else begin
    ymargin := 0.1 * (yx - ym);
    prYmin    := ym - ymargin;
    prYmax    := yx + ymargin;
  end;
end;

procedure TQUB_QVO_QVD_Overlap_Channel.prRedrawAvgStd;
var
  i                       :Integer;
  Rect                    :TRect;
  ymax, yratio            :Single;
begin
  Rect.Top                := 0;
  Rect.Left               := 0;
  Rect.Right              := Width;
  Rect.Bottom             := Height;
  prBitmapCanvas.Brush.Color := clWhite;
  prBitmapCanvas.FillRect(Rect);

  if List = nil then Exit;

  prYfactor := Height / (prYmax - prYmin);
  ymax                    := prYmax;
  yratio                  := prYfactor;

  prBitmapCanvas.Pen.Color := clLtGray;
  for i := 0 to Resampler.SampleCount - 1 do begin
    if prStd[i] > 0.0 then begin
      prBitmapCanvas.MoveTo(i, Round(yratio * (ymax - (prAvg[i] - prStd[i]))));
      prBitmapCanvas.LineTo(i, Round(yratio * (ymax - (prAvg[i] + prStd[i]))));
    end;
  end;
  prBitmapCanvas.Pen.Color := clBlack;
  if Resampler.SampleCount > 0 then prBitmapCanvas.MoveTo(0, Round(yratio * (ymax - prAvg[0])));
  for i := 1 to Resampler.SampleCount - 1 do begin
    prBitmapCanvas.LineTo(i, Round(yratio * (ymax - prAvg[i])));
  end;

  prAllNeedRedraw         := False;
end;


// *****************************************************************************

constructor TQUB_QVO_QVD_OverlapActive.Create(AOwner :TComponent);
type
  TMenuFunc = procedure(Sender : TObject) of Object;

  Function SetupMenuItem( cCaption : String; prOnClick : TMenuFunc ) : TMenuItem;
  begin
    Result := TMenuItem.Create(Self);
    Result.Caption := cCaption;
    Result.OnClick := prOnClick;
    prPopup.Items.Add(Result);
  end;

begin
  inherited;

  prResampler             := TQUB_QVO_QVD_ListResampler.Create;
  prResampler.ActiveOnly   := True;
  prResampler.ChannelIndex := 0; // should be ActiveChannel
  prResampler.OnNeedsResample := prOnNeedsResample;
  prResampler.OnYRangeChanged := prOnYRangeChanged;

  prFitResampler             := TQUB_QVO_QVD_ListResampler.Create;
  prFitResampler.ActiveOnly   := True;
  prFitResampler.FitCurves    := True;
  prFitResampler.ChannelIndex := 0; // should be ActiveChannel
  prFitResampler.OnNeedsResample := prOnNeedsResample;
  prFitResampler.OnYRangeChanged := prOnYRangeChanged;

  prPopup                 := TPopupMenu.Create(Self);
  prPopup.AutoLineReduction := maAutomatic;
  prPopup.OnPopup           := prOnPopup;

  itemZoomOut := SetupMenuItem( 'Zoom out', prOnClickZoomOut );
  prPopup.Items.NewBottomLine;
  itemOverlap := SetupMenuItem( 'Overlap', prOnClickOverlap);
  itemAverage := SetupMenuItem( 'Average', prOnClickAverage);
  itemSum     := SetupMenuItem( 'Sum', prOnClickSum);
  itemFunction := SetupMenuItem( 'Function...', prOnClickFunction);
  prPopup.Items.NewBottomLine;
  itemAppend := SetupMenuItem( 'Append to data', prOnClickAppend);
  itemCurveFit := SetupMenuItem( 'Curve fitting...', prOnClickCurveFit);
  itemClearFit := SetupMenuItem( 'Clear fit curve', prOnClickClearFit);
  itemCopyImage := SetupMenuItem( 'Copy Image', prOnClickCopyImage);
  prPopup.Items.NewBottomLine;
  itemNewList := SetupMenuItem( 'New list of selected', prOnClickNewList);

  prChannel               := TQUB_QVO_QVD_Overlap_Channel.Create(Self);
  prChannel.Parent        := Self;
  prChannel.ChannelIndex  := 0;
  prChannel.ShowMarquee   := False;
  prChannel.Resampler     := prResampler;
  prChannel.PopupMenu     := prPopup;
  prChannel.OnBeforeZoom  := prOnBeforeZoomChannel;

  prFitChannel               := TQUB_QVO_QVD_Overlap_Channel.Create(Self);
  prFitChannel.Parent        := Self;
  prFitChannel.ChannelIndex  := 0;
  prFitChannel.ShowMarquee   := False;
  prFitChannel.Resampler     := prFitResampler;
  prFitChannel.PopupMenu     := prPopup;
  prFitChannel.OnBeforeZoom  := prOnBeforeZoomChannel;

  OnMouseMove             := prOnMouseMove;
end;

destructor TQUB_QVO_QVD_OverlapActive.Destroy;
begin
  try
    Data                  := nil;
    prChannel.Free;
    prResampler.Free;
    prFitChannel.Free;
    prFitResampler.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVD_OverlapActive.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
const
  MARGIN_W       = 6;
  MARGIN_H       = 6;
  MIN_TRACEWIDTH = 30;
  MIN_TRACEHEIGHT = 10;
var
  TraceWidth     :Integer;
  TraceHeight    :Integer;
  EffHeight      :Integer;
  EffWidth       :Integer;
  HasFit         :Boolean;
  CurveIter      :TQUB_FitCurveIter;
begin
  inherited;

  if Parent <> nil then begin
    EffHeight := ClientHeight;
    EffWidth  := ClientWidth;
  end else begin
    EffHeight := Height;
    EffWidth  := Width;
  end;

  TraceWidth     := Max(MIN_TRACEWIDTH, EffWidth - 2 * MARGIN_W);
  CurveIter      := nil;

  if (prData <> nil) and (prList <> nil) and (prList.Items.Count > 0) then begin
    try
      CurveIter      := prData.FitCurves.Find(prList[0].Sel.First + prList.ActiveRange.First);
      HasFit         := 0 = CurveIter.Done;
    finally
      CurveIter.Free;
    end;
  end else begin
    HasFit         := False;
  end;

  if (prData <> nil) and prData.DrawIdealized and HasFit then begin
    TraceHeight    := Max(MIN_TRACEHEIGHT, (EffHeight div 2 - 2*MARGIN_H));
    prFitChannel.SetBounds(MARGIN_W, MARGIN_H, TraceWidth, TraceHeight);
    prChannel.SetBounds(MARGIN_W, MARGIN_H+TraceHeight, TraceWidth, TraceHeight);
  end else begin
    TraceHeight    := Max(MIN_TRACEHEIGHT, (EffHeight - 2*MARGIN_H));
    prFitChannel.SetBounds(MARGIN_W, MARGIN_H, TraceWidth, 0);
    prChannel.SetBounds(MARGIN_W, MARGIN_H, TraceWidth, TraceHeight);
  end;

  prResampler.SampleCount := TraceWidth;
  prFitResampler.SampleCount := TraceWidth;
end;

procedure TQUB_QVO_QVD_OverlapActive.Paint;
begin
  inherited;
end;

procedure TQUB_QVO_QVD_OverlapActive.prSetLblStatus(x :TLabel);
begin
  prChannel.LblStatus := x;
  prFitChannel.LblStatus := x;
  prLblStatus     := x;
end;

procedure TQUB_QVO_QVD_OverlapActive.prSetLblStatusActive(x :TLabel);
begin
  prChannel.LblStatusActive := x;
  prFitChannel.LblStatusActive := x;
  prLblStatusActive         := x;
end;

procedure TQUB_QVO_QVD_OverlapActive.prSetData(x :TQUB_QVO_QVD);
begin
  if x = prData then Exit;

  if prData <> nil then begin
    List := nil;
    Data.Events.DelEvent('OnActiveChannelChange', prOnActiveChannelChange);
    Data.Events.DelEvent('OnDrawIdealizedChange', prOnFitCurvesChanged);
    Data.FitCurves.Events.DelEvent('OnChanged',   prOnFitCurvesChanged);
  end;

  prData := x;

  if prData <> nil then begin
    Data.Events.AddEvent('OnActiveChannelChange', prOnActiveChannelChange);
    Data.Events.AddEvent('OnDrawIdealizedChange', prOnFitCurvesChanged);
    Data.FitCurves.Events.AddEvent('OnChanged',   prOnFitCurvesChanged);
  end;

  prResampler.Data           := prData;
  prFitResampler.Data        := prData;
end;

procedure TQUB_QVO_QVD_OverlapActive.prSetList(x :TQUB_QVO_QVD_SelectionList);
begin
  if x = prList then Exit;

  if (x = nil) and (prChannel.Mode = QUB_QVO_QVD_OVERLAP_FUNCTION) then begin
    PythonThread.BeginPython;
    prChannel.Mode := QUB_QVO_QVD_OVERLAP_OVERLAP;
    prChannel.Func := nil;
    PythonThread.EndPython;
  end;

  prList       := x;
  prChannel.List := x;
  prFitChannel.List := x;
  prResampler.List := x;
  prFitResampler.List := x;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnActiveChannelChange(Sender :TObject);
begin
  prResampler.ChannelIndex := Data.ActiveChannel;
  prChannel.ChannelIndex   := Data.ActiveChannel;
  prFitResampler.ChannelIndex := Data.ActiveChannel;
  prFitChannel.ChannelIndex   := Data.ActiveChannel;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnFitCurvesChanged(Sender :TObject);
begin
  if prList <> nil then SetBounds(Left, Top, Width, Height);
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if LblStatus <> nil then LblStatus.Caption := '';
  if LblStatusActive <> nil then LblStatusActive.Caption := '';
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnNeedsResample(Sender :TObject; Channel, Sel :Integer);
begin
  prChannel.DoOnNeedsResample(Sender, Sel);
  prFitChannel.DoOnNeedsResample(Sender, Sel);
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnYRangeChanged(Sender :TObject; Channel :Integer);
begin
  prChannel.DoOnYRangeChanged(Sender);
  prFitChannel.DoOnYRangeChanged(Sender);
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnPopup(Sender :TObject);
begin
  itemOverlap.Checked := prChannel.Mode = QUB_QVO_QVD_OVERLAP_OVERLAP;
  itemAverage.Checked := prChannel.Mode = QUB_QVO_QVD_OVERLAP_AVERAGE;
  itemSum.Checked     := prChannel.Mode = QUB_QVO_QVD_OVERLAP_SUM;
  itemFunction.Checked := prChannel.Mode = QUB_QVO_QVD_OVERLAP_FUNCTION;

  itemOverlap.Visible := List <> nil;
  itemAverage.Visible := List <> nil;
  itemSum.Visible     := List <> nil;
  itemFunction.Visible := List <> nil;

  itemAppend.Visible   := (prChannel.Mode <> QUB_QVO_QVD_OVERLAP_OVERLAP) and (prResampler.SelCount > 0);
  itemCurveFit.Visible := (prChannel.Mode <> QUB_QVO_QVD_OVERLAP_OVERLAP) and (prResampler.SelCount > 0);
  itemClearFit.Visible := itemCurveFit.Visible;

  itemNewList.Visible  := prResampler.SelCount > 0;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickZoomOut(Sender :TObject);
var
  i                :Integer;
begin
  if ZoomStack.Empty then begin      // Select all
    List.ActiveRange := QUB_FirstLast(-1, -1);
    for i := 0 to List.Items.Count - 1 do begin
      List[i].Active := True;
    end;
  end else begin
    ZoomStack.Pop;
  end;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickOverlap(Sender :TObject);
begin
  prChannel.Mode := QUB_QVO_QVD_OVERLAP_OVERLAP;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickAverage(Sender :TObject);
begin
  prChannel.Mode := QUB_QVO_QVD_OVERLAP_AVERAGE;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickSum(Sender :TObject);
begin
  prChannel.Mode := QUB_QVO_QVD_OVERLAP_SUM;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickFunction(Sender :TObject);
var
  Header         :String;
  Lambda         :PPyObject;
  Args           :PPyObject;
  PyResult       :PPyObject;
  i              :Integer;
begin
  if prResampler.SelCount = 0 then Exit;

  PythonThread.BeginPython;
  with GetPythonEngine do begin
    Header         := 'lambda ';
    Args           := PyTuple_New(prResampler.SelCount);
    for i := 0 to prResampler.SelCount - 2 do begin
      Header       := Header + 'trace' + IntToStr(i+1) + ', ';
      PyTuple_SetItem(Args, i, PyFloat_FromDouble(1.0));
    end;
    Header         := Header + 'trace' + IntToStr(prResampler.SelCount) + ': ';
    PyTuple_SetItem(Args, prResampler.SelCount-1, PyFloat_FromDouble(1.0));

    Lambda         := nil;
    repeat
      if not InputQuery('Custom function', 'f(trace1, ..., trace'+IntToStr(prResampler.SelCount)+') = ', prFuncText) then begin
        Lambda     := nil;
        Break;
      end;
      if prFuncText = '' then prFuncText := '0.0';

      Lambda       := nil;
      try
        Lambda     := EvalString(Header + prFuncText);
      except
      end;

      PyResult   := PyObject_CallObject(Lambda, Args);
      if PyResult = nil then begin
        PyErr_Clear;
        Py_DECREF(Lambda);
        Lambda   := nil;
      end else begin
        Py_DECREF(PyResult);
      end;
    until Lambda <> nil;

    if Lambda <> nil then begin
      prChannel.Func := Lambda;
      Py_DECREF(Lambda);
    end;
    
    Py_DECREF(Args);
  end;
  PythonThread.EndPython;

  if Lambda <> nil then prChannel.Mode := QUB_QVO_QVD_OVERLAP_FUNCTION;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickAppend(Sender :TObject);
var
  Task           :TQUB_Task_AppendOverlap;
begin
  if prChannel.Mode = QUB_QVO_QVD_OVERLAP_AVERAGE then begin
    Task         := TQUB_Task_AppendOverlapAvg.Create(True{Suspended}); 
  end else if prChannel.Mode = QUB_QVO_QVD_OVERLAP_SUM then begin
    Task         := TQUB_Task_AppendOverlapSum.Create(True{Suspended});
  end else if prChannel.Mode = QUB_QVO_QVD_OVERLAP_FUNCTION then begin
    Task         := TQUB_Task_AppendOverlapFunc.Create(True{Suspended});
    TQUB_Task_AppendOverlapFunc(Task).Func    := prChannel.Func;
  end else begin
    Exit;
  end;

  Task.Data         := Data;
  Task.ChannelIndex := prResampler.ChannelIndex;
  Task.List         := List;
  Task.Resume;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickCurveFit(Sender :TObject);
begin
  prChannel.ShowFit;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickClearFit(Sender :TObject);
begin
  prChannel.ClearFit;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickCopyImage(Sender :TObject);
var
  Bitmap      :Graphics.TBitmap;
begin
  // 10/05 : rudimentary copy image - copies the last prBitmap contents
  // Clipboard.Assign(prChannel.Bitmap);
  // now with fit curves and all, let's draw a frame like qvd,
  // then paint both channels into it.

  // 10/05 : just as rudimentary, but includes the fit channel
  // also demonstrates where we could draw in fancier stuff
  Bitmap              := Graphics.TBitmap.Create;
  Bitmap.Width        := prChannel.Width;
  Bitmap.Height       := prFitChannel.Height + prChannel.Height;
  Bitmap.Canvas.Draw(0, 0, prFitChannel.Bitmap);
  Bitmap.Canvas.Draw(0, prFitChannel.Height, prChannel.Bitmap);
  Clipboard.Assign(Bitmap);
  Bitmap.Free;//?

end;

procedure TQUB_QVO_QVD_OverlapActive.prOnClickNewList(Sender :TObject);
var
  Name         :String;
  xList        :TQUB_QVO_QVD_SelectionList;
  Iter         :TQUB_QVO_QVD_SelectionListIter;
begin
  Name         := Data.CurrListName + ' x';
  if not InputQuery('New list of active', 'List name:', Name) then Exit;

  xList         := Data.GetList(Name);
  xList.Clear;
  Iter          := List.FindAll(True{ActiveOnly});

  if not Iter.Done then repeat
    xList.AddSelection(Iter.Sel.First, Iter.Sel.Last, '');
  until Iter.Next;

  Iter.Free;
end;

procedure TQUB_QVO_QVD_OverlapActive.prOnBeforeZoomChannel(Sender :TObject);
begin
  ZoomStack.Push;
end;


// *****************************************************************************

constructor TQUB_QVO_QVD_ListResampler.Create;
begin
  prChannelIndex := -1;
end;

destructor TQUB_QVO_QVD_ListResampler.Destroy;
begin
  try
    Data := nil;
  finally
    inherited;
  end;
end;

function TQUB_QVO_QVD_ListResampler.DoResample: Boolean;
begin
  Result := False;
  if AllValid then Exit;

  if ActiveOnly and (List <> nil) and (List.ActiveRange.First >= 0) then begin
    Xfactor              := (List.ActiveRange.Last - List.ActiveRange.First + 1) * 1.0 / SampleCount;
    Xoffset              := List.ActiveRange.First;
  end else begin
    Xfactor              := MaxDataCount * 1.0 / SampleCount;
    if Xfactor = 0 then Xfactor := 1.0;
    Xoffset              := 0;
  end;

  if AllInvalid and (ChannelIndex < 0) and (SelCount > 0) and (not FitCurves) then begin
    prResampleAll;
  end else begin
    prResampleSome;
  end;

  AllInvalid              := False;
  AllValid                := True;
  Result                  := True;
end;

procedure TQUB_QVO_QVD_ListResampler.prResampleAll;
var
  c, i, s, d              :Integer;
  ym, yx                  :Single;
  YRangeChanged           :Boolean;
  val                     :Single;
  Sel                     :TQUB_FirstLast;
  SelDataCount            :Integer;
  Buffer                  :TQUB_QVO_QVD_DataBuffer;
  BufferCount             :Integer;
  ActiveRangeOnly         :Boolean;
begin
  Buffer                  := nil;

  try

    ActiveRangeOnly       := ActiveOnly and (List <> nil) and (List.ActiveRange.First >= 0);

    BufferCount           := MaxDataCount;
    if ActiveRangeOnly then BufferCount := Min(BufferCount, List.ActiveRange.Last - List.ActiveRange.First + 1);
    BufferCount           := Min(BufferCount, 1048576 div ChannelCount);
    Buffer                := TQUB_QVO_QVD_DataBuffer.Create(prData, BufferCount, True);

    for i := 0 to SelCount - 1 do begin
      Sel                   := List[i].Sel;
      SelDataCount          := DataCount[i];
      if ActiveRangeOnly then begin
        Sel.Last            := Min(Sel.Last, Sel.First + List.ActiveRange.Last);
        Sel.First           := Sel.First + List.ActiveRange.First;
        SelDataCount        := Max(0, Sel.Last - Sel.First + 1);
      end;
      SelSampleCount[i]     := Min( SampleCount, Ceil(SelDataCount / Xfactor) );

      for c := 0 to ChannelCount - 1 do begin
        SetLength(Samples[c][i], SelSampleCount[i]);
        if SelSampleCount[i] > 0 then begin
          val               := Buffer.ValueAt(c, Sel.First);
          Samples[c][i][0]  := val;
          Ymin[c][i]        := val;
          Ymax[c][i]        := val;
        end;
        Invalid[c][i]       := False;
      end;

      for s := 1 to SelSampleCount[i] - 1 do begin
        d                   := Sel.First + Floor(Xfactor * s);
        for c := 0 to ChannelCount - 1 do begin
          val               := Buffer.ValueAt(c, d);
          Samples[c][i][s]  := val;
          if val < Ymin[c][i] then Ymin[c][i] := val;
          if val > Ymax[c][i] then Ymax[c][i] := val;
        end;
      end;
    end;

    for c := 0 to ChannelCount - 1 do begin
      ym                    := 1.0e20;
      yx                    := -1.0e20;
      for i := 0 to SelCount - 1 do begin
        if (not ActiveOnly) or List[i].Active then begin
          if Ymin[c][i] < ym then ym := Ymin[c][i];
          if Ymax[c][i] > yx then yx := Ymax[c][i];
        end;
      end;
      YRangeChanged         := (ym <> ChannelYmin[c]) or (yx <> ChannelYmax[c]);
      ChannelYmin[c]        := ym;
      ChannelYmax[c]        := yx;
      if YRangeChanged and Assigned(OnYRangeChanged) then OnYRangeChanged(Self, c);
    end;

  finally

    Buffer.Free;

  end;
end;

procedure TQUB_QVO_QVD_ListResampler.prResampleSome;
var
  c, i                    :Integer;
  YRangeChanged           :array of Boolean;
  NewChannelYmin          :array of Single;
  NewChannelYmax          :array of Single;
begin
  SetLength(YRangeChanged, ChannelCount);
  SetLength(NewChannelYmin, ChannelCount);
  SetLength(NewChannelYmax, ChannelCount);

  for i := 0 to SelCount - 1 do begin
    for c := 0 to ChannelCount - 1 do begin
      if (ChannelIndex < 0) or (c = ChannelIndex) then begin
        if i = 0 then YRangeChanged[c]  := False;
        if AllInvalid or Invalid[c][i] then begin
          if (not ActiveOnly) or List[i].Active then begin
            YRangeChanged[c] := prResampleOne(c, i) or YRangeChanged[c];
          end else begin
            prResampleOne(c, i);
            YRangeChanged[c] := True; // since it was maybe just inactivated
          end;
        end;
      end;
    end;
  end;

  for c := 0 to ChannelCount - 1 do begin
    if YRangeChanged[c] then begin
      NewChannelYmin[c] := 1.0e20;
      NewChannelYmax[c] := -1.0e20;
      for i := 0 to SelCount - 1 do begin
        if (not ActiveOnly) or List[i].Active then begin
          NewChannelYmin[c]  := Min(NewChannelYmin[c], Ymin[c][i]);
          NewChannelYmax[c]  := Max(NewChannelYmax[c], Ymax[c][i]);
        end;
      end;

      if (NewChannelYmin[c] <> ChannelYmin[c])
      or (NewChannelYmax[c] <> ChannelYmax[c]) then begin
        ChannelYmin[c] := NewChannelYmin[c];
        ChannelYmax[c] := NewChannelYmax[c];
        
        if Assigned(OnYRangeChanged) then OnYRangeChanged(Self, c);
      end;
    end;
  end;
end;

function TQUB_QVO_QVD_ListResampler.prResampleOne(ch, s :Integer): Boolean; // returns true if Ymin, Ymax changed
var
  ym, yx                  :Single;
  val                     :Single;
  si                      :Integer;
  Sel                     :TQUB_FirstLast;
  SelDataCount            :Integer;
  Buffer                  :TQUB_QVO_QVD_DataBufferOne;
  BufferCount             :Integer;
  ActiveRangeOnly         :Boolean;
begin
  Result                  := False;
  Buffer                  := nil;

  try

    ActiveRangeOnly := ActiveOnly and (List <> nil) and (List.ActiveRange.First >= 0);

    BufferCount           := MaxDataCount;
    if ActiveRangeOnly then BufferCount := Min(BufferCount, List.ActiveRange.Last - List.ActiveRange.First + 1);
    BufferCount           := Min(BufferCount, 1048576);
    Buffer                := TQUB_QVO_QVD_DataBufferOne.Create(prData, ch, BufferCount, True{process}, FitCurves);
    Sel                   := List[s].Sel;
    SelDataCount          := DataCount[s];
    if ActiveRangeOnly then begin
      Sel.Last            := Min(Sel.Last, Sel.First + List.ActiveRange.Last);
      Sel.First           := Sel.First + List.ActiveRange.First;
      SelDataCount        := Max(0, Sel.Last - Sel.First + 1);
    end;

    SelSampleCount[s]     := Min(SampleCount, Ceil(SelDataCount / Xfactor));

    SetLength(Samples[ch][s], SelSampleCount[s]);
    if SelSampleCount[s] > 0 then begin
      Samples[ch][s][0]     := Buffer.ValueAt(Sel.First);
      ym                    := Samples[ch][s][0];
      yx                    := Samples[ch][s][0];
      for si := 1 to SelSampleCount[s] - 1 do begin
        val                 := Buffer.ValueAt(Sel.First + Floor(Xfactor * si));
        Samples[ch][s][si]:= val;
        if val < ym then ym := val;
        if val > yx then yx := val;
      end;

      if ym <> Ymin[ch][s] then begin
        Result            := True;
        Ymin[ch][s]         := ym;
      end;
      if yx <> Ymax[ch][s] then begin
        Result            := True;
        Ymax[ch][s]       := yx;
      end;
    end;

    Invalid[ch][s]      := False;

  finally

    Buffer.Free;
    
  end;
end;

procedure TQUB_QVO_QVD_ListResampler.prSetData(x :TQUB_QVO_QVD);
begin
  if x = prData then Exit;

  List             := nil;

  if prData <> nil then begin
    Data.Events.DelEvent('OnEdit',                prOnDataEdit);
    Data.Events.DelEvent('OnFilterChange',        prOnDataProcessChange);
    Data.Events.DelEvent('OnBaselineApplyChange', prOnDataProcessChange);
    Data.FitCurves.Events.DelEvent('OnChanged',   prOnFitCurvesChanged);
    ChannelCount   := 0;
    SetLength(Samples, 0);
    SetLength(Ymin, 0);
    SetLength(Ymax, 0);
    SetLength(ChannelYmin, 0);
    SetLength(ChannelYmax, 0);
    SetLength(Invalid, 0);
  end;

  prData           := x;

  if prData <> nil then begin
    Data.Events.AddEvent('OnEdit',                prOnDataEdit);
    Data.Events.AddEvent('OnFilterChange',        prOnDataProcessChange);
    Data.Events.AddEvent('OnBaselineApplyChange', prOnDataProcessChange);
    Data.FitCurves.Events.AddEvent('OnChanged',   prOnFitCurvesChanged);
    ChannelCount   := prData.DataObject.ADChannelCount;
    SetLength(Samples, ChannelCount);
    SetLength(Ymin, ChannelCount);
    SetLength(Ymax, ChannelCount);
    SetLength(ChannelYmin, ChannelCount);
    SetLength(ChannelYmax, ChannelCount);
    SetLength(Invalid, ChannelCount);

    if prChannelIndex >= ChannelCount then prChannelIndex := -1;
  end;

  AllInvalid := True;
  AllValid   := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prSetSampleCount(x :Integer);
begin
  if x = prSampleCount then Exit;

  prSampleCount := x;
  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prSetChannelIndex(x :Integer);
var
  c           :Integer;
begin
  if x = prChannelIndex then Exit;

  prChannelIndex := x;

  if prList <> nil then begin
    for c := 0 to ChannelCount - 1 do begin
      if (ChannelIndex < 0) or (ChannelIndex = c) then begin
        SetLength(Samples[c], SelCount);
        SetLength(Ymin[c], SelCount);
        SetLength(Ymax[c], SelCount);
        SetLength(Invalid[c], SelCount);
      end else begin
        SetLength(Samples[c], 0);
        SetLength(Ymin[c], 0);
        SetLength(Ymax[c], 0);
        SetLength(Invalid[c], 0);
      end;
    end;
  end;

  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prSetActiveOnly(x :Boolean);
begin
  if x = prActiveOnly then Exit;
  prActiveOnly  := x;
  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prSetFitCurves(x :Boolean);
begin
  if x = prFitCurves then Exit;
  prFitCurves   := x;
  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prSetList(x :TQUB_QVO_QVD_SelectionList);
var
  c, i        :Integer;
begin
  if x = prList then Exit;

  if prList <> nil then begin
    prList.Events.DelEvent('ItemAdded', prOnItemAdded);
    prList.Events.DelEvent('ItemDeleting', prOnItemDeleting);
    prList.Events.DelEvent('ActiveChanged', prOnActiveChanged);
    prList.Events.DelEvent('ActiveRangeChanged', prOnActiveRangeChanged);

    SelCount := 0;
    SetLength(DataCount, SelCount);
    SetLength(SelSampleCount, SelCount);
    MaxDataCount := 0;
    for c := 0 to ChannelCount - 1 do begin
      SetLength(Samples[c], 0);
      SetLength(Ymin[c], 0);
      SetLength(Ymax[c], 0);
      SetLength(Invalid[c], 0);
    end;
  end;

  prList := x;

  if prList <> nil then begin
    prList.Events.AddEvent('ItemAdded', prOnItemAdded);
    prList.Events.AddEvent('ItemDeleting', prOnItemDeleting);
    prList.Events.AddEvent('ActiveChanged', prOnActiveChanged);
    prList.Events.AddEvent('ActiveRangeChanged', prOnActiveRangeChanged);

    SelCount := prList.Items.Count;
    SetLength(DataCount, SelCount);
    SetLength(SelSampleCount, SelCount);
    MaxDataCount := 0;
    for i := 0 to SelCount - 1 do begin
      DataCount[i] := List[i].Sel.Last - List[i].Sel.First + 1;
      MaxDataCount := Max(MaxDataCount, DataCount[i]);
    end;
    for c := 0 to ChannelCount - 1 do begin
      if (ChannelIndex < 0) or (ChannelIndex = c) then begin
        SetLength(Samples[c], SelCount);
        SetLength(Ymin[c], SelCount);
        SetLength(Ymax[c], SelCount);
        SetLength(Invalid[c], SelCount);
      end;
    end;
  end;

  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prOnDataEdit(Sender :TObject);
begin
  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prOnDataProcessChange(Sender :TObject);
begin
  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prOnFitCurvesChanged(Sender :TObject);
begin
  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prOnItemAdded(Sender :TObject);
var
  i, p, n, c :Integer;
  NewSamples :ArrayOfArrayOfSingle;
begin
  p := List.EventIndex;
  n := List.Items.Count;
  SelCount := n;

  SetLength(DataCount, n);
  SetLength(SelSampleCount, n);
  for i := p + 1 to n - 1 do begin
    DataCount[i]      := DataCount[i-1];
    SelSampleCount[i] := SelSampleCount[i-1];
  end;
  DataCount[p]        := List[p].Sel.Last - List[p].Sel.First + 1;

  SetLength(NewSamples, ChannelCount, SampleCount);

  for c := 0 to ChannelCount - 1 do begin
    if (ChannelIndex < 0) or (ChannelIndex = c) then begin
      SetLength(Samples[c], n);
      SetLength(Ymin[c], n);
      SetLength(Ymax[c], n);
      SetLength(Invalid[c], n);

      for i := p + 1 to n - 1 do begin
        Samples[c][i] := Samples[c][i-1];
        Ymin[c][i]    := Ymin[c][i-1];
        Ymax[c][i]    := Ymax[c][i-1];
        Invalid[c][i] := Invalid[c][i-1];
      end;

      Samples[c][p]   := NewSamples[c];
      Invalid[c][p]   := True;
      AllValid        := False;
    end;
  end;

  if DataCount[p] > MaxDataCount then begin
    MaxDataCount := DataCount[p];
    AllInvalid   := True;
    AllValid     := False;
  end;

  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prOnItemDeleting(Sender :TObject);
var
  i, p, n, c :Integer;
begin
  p := List.EventIndex;
  n := List.Items.Count;
  SelCount := n - 1;

  for i := p to n - 2 do begin
    DataCount[i]       := DataCount[i+1];
    SelSampleCount[i]  := SelSampleCount[i+1];
  end;
  SetLength(DataCount, n-1);
  SetLength(SelSampleCount, n-1);

  MaxDataCount   := 0;
  for i := 0 to n - 2 do begin
    if DataCount[i] > MaxDataCount then MaxDataCount := DataCount[i];
  end;

  for c := 0 to ChannelCount - 1 do begin
    if (ChannelIndex < 0) or (ChannelIndex = c) then begin
      for i := p to n - 2 do begin
        Samples[c][i] := Samples[c][i+1];
        Ymin[c][i]    := Ymin[c][i+1];
        Ymax[c][i]    := Ymax[c][i+1];
        Invalid[c][i] := Invalid[c][i+1];
      end;
      SetLength(Samples[c], n-1);
      SetLength(Ymin[c], n-1);
      SetLength(Ymax[c], n-1);
      SetLength(Invalid[c], n-1);
    end;
  end;

  AllInvalid    := True;
  AllValid      := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prOnActiveChanged(Sender :TObject);
var
  p, c       :Integer;
  Activating :Boolean;
begin
  if not ActiveOnly then Exit;

  p          := List.EventIndex;
  Activating := List[p].Active;
  for c := 0 to ChannelCount - 1 do begin
    if (ChannelIndex < 0) or (ChannelIndex = c) then begin
      if Activating then begin
        Invalid[c][p] := True;
      end else begin
        AllInvalid    := True;
      end;
      AllValid        := False;

      // force recalc of y range
      Ymin[c][p]    := 0.0;
      Ymax[c][p]    := 0.0;
    end;
  end;

  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;

procedure TQUB_QVO_QVD_ListResampler.prOnActiveRangeChanged(Sender :TObject);
begin
  if not ActiveOnly then Exit;

  AllInvalid := True;
  AllValid   := False;
  if Assigned(OnNeedsResample) then OnNeedsResample(Self, -1, -1);
end;


// *****************************************************************************
end.
