{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// functions to play data through the sound card


unit QUB_QVO_QVD_Play;


interface


uses


Windows, MMSystem, Dialogs, Math,
AllocTypes, Alloc,
QUB_Global, QUB_QVO_QVD_Global, QUB_QVO_QVD, QUB_QFS_QDF_MPX;


var


Play_Data               :TQUB_QVO_QVD;
Play_Selection          :TQUB_FirstLast;
Play_ADChannel          :Integer;
Play_ADChannelCount     :Integer;
Play_Playing            :Boolean;
Play_WaveOut            :Integer;
Play_Header             :Array[1..2] of TWaveHdr;
Play_Buffer             :TPMInteger;
Play_BufferW            :Array[1..2] of Pointer;
Play_BufferCount        :Integer;
Play_BufferSize         :Integer;
Play_BufferIndex        :Integer;
Play_BufferElemSize     :Integer;
Play_BufferElemUnSigned :Boolean;
Play_SelPlaying         :TQUB_FirstLast;


function  Play_Play(Data :TQUB_QVO_QVD; ADChannel :Integer) :Boolean;
function  Play_Stop(Data :TQUB_QVO_QVD) :Boolean;
function  Play_LoadBuffer(b :Integer; Play_SelPlaying :TQUB_FirstLast) :Boolean;
procedure Play_CallBack(hdrvr :HDRVR; uMsg: UINT; dwUser: DWORD; dw1, dw2: DWORD) stdcall;


implementation


//****************************************************************************//
//                                                                            //
//                          P L A Y     S O U N D                             //
//                                                                            //
//****************************************************************************//
function Play_Play(Data :TQUB_QVO_QVD; ADChannel :Integer) :Boolean;
var
  WaveFormat :TWaveFormatEx;
  i          :Integer;
begin

  Result := Play_Stop(Data);
  if Result then Exit;

  if Data.Data = nil then Exit;
  if Data.LockedReadWrite    then Exit;
  if Data.DataPointCount = 0 then Exit;

  Play_ADChannel := ADChannel;
  if (Play_ADChannel = -1{All}) then begin
    Play_ADChannelCount := Data.DataObject.ADChannelCount;
  end else begin
    Play_ADChannelCount := 1;
  end;

  if Play_ADChannelCount > 2 then begin
    MessageHelpDlg('Cannot play more than two A/D channels', mtError, [mbOK], 0, 'Error:Cannot_Play');
    Result := True;
    Exit;
  end;


  //Add an onclose/onchange event ...


  try

    //Play Selection
    Play_Data := Data;
    Play_Selection := Play_Data.DataSel;
    if Play_Selection.First = Play_Selection.Last then begin
      Play_Selection.First := Play_Data.DataCurr;
      Play_Selection.Last  := Play_Data.DataPoint.Last;
    end;

    Play_BufferSize  := Round(1 / Play_Data.Sampling);
    Play_BufferCount := Ceil((Play_Selection.Last - Play_Selection.First + 1) / Play_BufferSize);

    //Wave Format
    with WaveFormat do begin
      wFormatTag       := WAVE_FORMAT_PCM;
      nChannels        := Play_ADChannelCount;
      nSamplesPerSec   := Round(1 / Play_Data.Sampling);
      wBitsPerSample   := Play_Data.DataObject.WAVBitsPerSample;
      nBlockAlign      := Play_Data.DataObject.WAVBlockAlign;
      nAvgBytesPerSec  := Play_Data.DataObject.WAVAvgBytesPerSec;
      cbSize           := 0;
    end;

    //Buffer
    try
      Play_Buffer  := IAlloc2(Play_ADChannelCount, Play_BufferSize);
    except
      Result := True;
    end;

    if Result then Exit;

    //BufferW
    for i := 1 to 2 do begin

      if WaveFormat.wBitsPerSample <= 8 then begin
        Play_BufferElemSize     := 1;
        Play_BufferElemUnSigned := True;
      end else begin
        Play_BufferElemSize     := 2;
        Play_BufferElemUnSigned := False;
      end;
      try
        Play_BufferW[i] := Alloc1(Play_BufferSize, Play_BufferElemSize * Play_ADChannelCount);
      except
        Result := True;
      end;

      if Result then Exit;

      if Play_BufferCount = 1 then Break;

    end;

    //Open Device
    Play_WaveOut := 0;
    Result := Boolean(waveOutOpen(@Play_WaveOut, WAVE_MAPPER, @WaveFormat, Cardinal(@Play_CallBack), 0, CALLBACK_FUNCTION));

    if Result then Exit;

    Play_SelPlaying.Last  := Play_Selection.First - 1;

    //Header
    for i := 1 to 2 do begin

      Play_SelPlaying.First := Play_SelPlaying.Last + 1;
      Play_SelPlaying.Last  := Play_SelPlaying.First + Play_BufferSize - 1;
      if Play_SelPlaying.Last > Play_Selection.Last then Play_SelPlaying.Last := Play_Selection.Last;

      with Play_Header[i] do begin
        lpData         := LPSTR(Play_BufferW[i]);
        dwBufferLength := Play_ADChannelCount * Play_BufferElemSize * (Play_SelPlaying.Last - Play_SelPlaying.First + 1);
        dwFlags        := 0;
        dwLoops        := 1;
      end;

      Result := Play_LoadBuffer(i, Play_SelPlaying);

      if Result then Exit;

      Result := Boolean(waveOutPrepareHeader(Play_WaveOut, @Play_Header[i], SizeOf(Play_Header[i])));

      if Play_BufferCount = 1 then Break;

      if Result then Exit;

    end;

    //Header Flags
    Play_Header[1].dwFlags := Play_Header[1].dwFlags or WHDR_BEGINLOOP;
    if Play_BufferCount = 1 then begin
      Play_Header[1].dwFlags := Play_Header[1].dwFlags or WHDR_ENDLOOP;
    end;
    if Play_BufferCount = 2 then begin
      Play_Header[2].dwFlags := Play_Header[2].dwFlags or WHDR_ENDLOOP;
    end;

    Play_BufferIndex := 1;

    Result := Boolean(waveOutWrite(Play_WaveOut, @Play_Header[1], SizeOf(Play_Header[1])));
    if Result then Exit;
    if Play_BufferCount > 1 then begin
      Result := Boolean(waveOutWrite(Play_WaveOut, @Play_Header[2], SizeOf(Play_Header[2])));
    end;
    if Result then Exit;

    Play_Playing := True;
    //Play_Data.Lock;

  finally

    if Result then begin

      Play_Playing := False;

      //Unprepare Headers
      if (Play_Header[1].dwFlags and WHDR_PREPARED) <> 0 then begin
        waveOutUnPrepareHeader(Play_WaveOut, @Play_Header[1], SizeOf(Play_Header[1]));
      end;
      if (Play_BufferCount > 1) and ((Play_Header[2].dwFlags and WHDR_PREPARED) <> 0) then begin
        waveOutUnPrepareHeader(Play_WaveOut, @Play_Header[2], SizeOf(Play_Header[2]));
      end;

      //Free Memory
      Free2(TPMChar(Play_Buffer));
      FreeMem(Play_BufferW[1]);
      FreeMem(Play_BufferW[2]);
      Play_Buffer     := nil;
      Play_BufferW[1] := nil;
      Play_BufferW[2] := nil;
      //Play_Data.UnLock;
      Play_Data := nil;

    end;

  end;

end;

function Play_Stop(Data :TQUB_QVO_QVD) :Boolean;
var
  i :Integer;
begin
  Result := False;

  if not Play_Playing then Exit;

  Play_Playing := False;
  waveOutReset(Play_WaveOut);
  for i := 1 to 2 do begin
    waveOutUnPrepareHeader(Play_WaveOut, @Play_Header[i], SizeOf(Play_Header[i]));
  end;
  waveOutClose(Play_WaveOut);

end;

procedure Play_CallBack(hdrvr :HDRVR; uMsg: UINT; dwUser: DWORD; dw1, dw2: DWORD) stdcall;
var
  Index :Integer;
  r     :Boolean;
begin

  //Buffer Index
  if Odd(Play_BufferIndex) then Index := 1 else Index := 2;

  case uMsg of

    WOM_DONE:
      begin

        //Unprepare the Header
        if Play_Playing then waveOutUnPrepareHeader(Play_WaveOut, @Play_Header[Index], SizeOf(Play_Header[Index]));

        Inc(Play_BufferIndex);

        //If more Buffers to play
        if (Play_SelPlaying.Last < Play_Selection.Last) and (Play_BufferIndex <= Play_BufferCount) and (Play_Playing) then begin

          Play_SelPlaying.First := Play_SelPlaying.Last  + 1;
          Play_SelPlaying.Last  := Play_SelPlaying.First + Play_BufferSize - 1;
          if Play_SelPlaying.Last > Play_Selection.Last then Play_SelPlaying.Last := Play_Selection.Last;

          Play_Header[Index].dwBufferLength := Play_ADChannelCount * Play_BufferElemSize * (Play_SelPlaying.Last - Play_SelPlaying.First + 1);

          r := Play_LoadBuffer(Index, Play_SelPlaying);

          if not r then begin

            Play_Header[Index].dwFlags := 0;
            waveOutPrepareHeader(Play_WaveOut, @Play_Header[Index], SizeOf(Play_Header[Index]));

            if Play_BufferIndex = Play_BufferCount then begin
            //This would be the last Buffer, therefore add EndLoop flag
              Play_Header[Index].dwFlags := Play_Header[Index].dwFlags or WHDR_ENDLOOP;
            end;

            waveOutWrite(Play_WaveOut, @Play_Header[Index], SizeOf(Play_Header[Index]));

            //Update Pointer in the data. 
            //Not finished. 12-2005. LM.
            //Play_Data.DataCurr := Play_SelPlaying.First;
            //Play_Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_BROWSE_CMD, LParam(QUB_QVO_QVD_CMD_CHANGE_SEL));

          end;

        end else begin

          //Play_Playing := False;
          //waveOutClose(Play_WaveOut);

        end;

      end;

    WOM_OPEN:
      begin
      end;

    WOM_CLOSE:
      begin

        Play_Playing := False;
        Free2(TPMChar(Play_Buffer));
        FreeMem(Play_BufferW[1]);
        FreeMem(Play_BufferW[2]);
        Play_Buffer     := nil;
        Play_BufferW[1] := nil;
        Play_BufferW[2] := nil;
        //Play_Data.UnLock;

        //Update Pointer in the data
        //Play_Data.DataCurr := Play_Selection.First;
        //Play_Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_BROWSE_CMD, LParam(QUB_QVO_QVD_CMD_CHANGE_SEL));

        Play_Data := nil;

      end;

  end;//case
end;

function Play_LoadBuffer(b :Integer; Play_SelPlaying :TQUB_FirstLast) :Boolean;
begin
  Result := Play_Data.DataPoint.Last < Play_SelPlaying.Last;
  if not Result then begin
    if Play_ADChannel = - 1 then begin
      Result := Play_Data.LoadDataProcessed(Play_Buffer, Play_SelPlaying);
    end else begin
      Result := Play_Data.LoadDataChannel(Play_Buffer[0], Play_ADChannel, Play_SelPlaying, True{Process}, Play_Data.FilterOn[Play_ADChannel], Play_Data.BaselineOn[Play_ADChannel]);
    end;
  end;
  if not Result then Result := MPXData(Play_Buffer, Play_BufferW[b], Play_ADChannelCount, -1 {Channel; All}, Play_BufferElemSize, Play_Data.DataObject.ADDataType, Play_SelPlaying.First - Play_SelPlaying.First, Play_SelPlaying.Last - Play_SelPlaying.First, True {MPX});
end;


//******************************************************************************
end.
