{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// properties form for printing and copying data


unit QUB_QVO_QVD_Print_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, ComCtrls, Forms, 
Dialogs, StdCtrls, ExtCtrls, Printers, Math, Registry, Clipbrd,

QUB_Global,
QUB_QVO, QUB_QVO_QVD_Global,
QUB_QVO_QVM,
QUB_Profiles,
QUB_Properties,
QUB_Task_Properties;


{$include 'QUB_QVO_QVD_Print_Properties.hct'}

const


DEF_DRAWDATA         = 1;
DEF_DRAWIDLDATA      = 1;
DEF_DRAWBASE         = 0;
DEF_DRAWMODE         = DRAWMODE_NORM;
DEF_INVERTDATA       = 0;
DEF_DRAWFRAME        = 1;
DEF_SHOWTITLES       = 0;
DEF_SHOWFILENAME     = 1;
DEF_SHOWFULLPATH     = 0;
DEF_SCALEBARS        = 1;
DEF_SCALEBARSTOP     = 1;
DEF_SCALEBARTIME     = 1;
DEF_SCALEBARAMP      = 0;
DEF_SCALEBARFONTSIZE = 8;
MIN_SCALEBARFONTSIZE = 4;
MAX_SCALEBARFONTSIZE = 120;
DEF_AMPFROMMODEL     = 0;
DEF_LANDSCAPE        = 1;
DEF_BITMAP           = 1;
MIN_WIDTH            = 20;
MAX_WIDTH            = 10000;
DEF_WIDTH            = 800;
MIN_HEIGHT           = 20;
MAX_HEIGHT           = 10000;
DEF_HEIGHT           = 600;
DEF_BLACKANDWHITE    = 1;


type


TQUB_QVO_QVD_Print_PropertiesForm = class(TQUB_Task_PropertiesForm)


  panOptions                      :TPanel;
  chkInvertData                   :TCheckBox;
  chkDrawFrame                    :TCheckBox;
  chkShowFileName                 :TCheckBox;
  chkShowFullPath                 :TCheckBox;
  chkShowTitles                   :TCheckBox;
  chkScaleBars                    :TCheckBox;
  Panel2                          :TPanel;
  chkScaleBarsTop                 :TRadioButton;
  chkScaleBarsBottom              :TRadioButton;
  lblOrientation                  :TLabel;
  Panel3                          :TPanel;
  chkPortrait                     :TRadioButton;
  chkLandscape                    :TRadioButton;
  lblImageFormat                  :TLabel;
  Panel1                          :TPanel;
  chkPicture                      :TRadioButton;
  chkBitmap                       :TRadioButton;
  txtWidth                        :TEdit;
  txtHeight                       :TEdit;
  lblWidth                        :TLabel;
  lblHeight                       :TLabel;
  Label2                          :TLabel;
  chkDrawData                     :TCheckBox;
  chkDrawIDLData                  :TCheckBox;
  chkDrawBase                     :TCheckBox;
  chkScaleBarTime                 :TCheckBox;
  chkScaleBarAmp                  :TCheckBox;
  chkAmpFromModel                 :TCheckBox;
  txtScaleBarFontSize             :TEdit;
  udScaleBarFontSize              :TUpDown;
  lblScaleBarFontSize             :TLabel;
  chkBlackAndWhite                :TCheckBox;
  Label11                         :TLabel;
  panDrawMode                     :TPanel;
  optDrawMode2                    :TRadioButton;
  optDrawMode1                    :TRadioButton;
  optDrawMode0                    :TRadioButton;


  procedure FormCreate(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure btnOKClick(Sender: TObject);


protected


  prData                          :TQUB_QVO;
  prModel                         :TQUB_QVO_QVM;

  prDlg                           :TPrintDialog;

  prPrinting                      :Boolean;

  prDrawData                      :Boolean;
  prDrawIDLData                   :Boolean;
  prDrawBase                      :Boolean;
  prDrawMode                      :TQUB_QVO_QVD_DrawMode;
  prInvertData                    :Boolean;
  prDrawFrame                     :Boolean;
  prShowFileName                  :Boolean;
  prShowFullPath                  :Boolean;
  prShowTitles                    :Boolean;
  prScaleBars                     :Boolean;
  prScaleBarsTop                  :Boolean;
  prScaleBarTime                  :Boolean;
  prScaleBarAmp                   :Boolean;
  prScaleBarFontSize              :Integer;
  prAmpFromModel                  :Boolean;
  prLandscape                     :Boolean;
  prDrawWidth                     :Integer;
  prDrawHeight                    :Integer;
  prBitmap                        :Boolean;
  prBlackAndWhite                 :Boolean;

  prCopyImage                     :Boolean;


  procedure prSetCopyImage(c :Boolean);

  procedure prPrint;
  procedure prCopy;


public


  procedure PropertiesToParams;

  function Print(Data :TQUB_QVO; Dlg :TPrintDialog; CopyImage :Boolean; ShowProperties :Boolean) :Boolean;


  property DrawData               :Boolean                     read prDrawData;
  property DrawIDLData            :Boolean                     read prDrawIDLData;
  property DrawBase               :Boolean                     read prDrawBase;
  property DrawMode               :TQUB_QVO_QVD_DrawMode       read prDrawMode;
  property InvertData             :Boolean                     read prInvertData;
  property DrawFrame              :Boolean                     read prDrawFrame;
  property ShowFileName           :Boolean                     read prShowFilename;
  property ShowFullPath           :Boolean                     read prShowFullPath;
  property ShowTitles             :Boolean                     read prShowTitles;
  property ScaleBars              :Boolean                     read prScaleBars;
  property ScaleBarsTop           :Boolean                     read prScaleBarsTop;
  property ScaleBarTime           :Boolean                     read prScaleBarTime;
  property ScaleBarAmp            :Boolean                     read prScaleBarAmp;
  property ScaleBarFontSize       :Integer                     read prScaleBarFontSize;
  property AmpFromModel           :Boolean                     read prAmpFromModel;
  property Landscape              :Boolean                     read prLandscape;
  property Bitmap                 :Boolean                     read prBitmap;
  property DrawWidth              :Integer                     read prDrawWidth;
  property DrawHeight             :Integer                     read prDrawHeight;
  property BlackAndWhite          :Boolean                     read prBlackAndWhite;

  property CopyImage              :Boolean                     read prCopyImage           write prSetCopyImage;

  property Data                   :TQUB_QVO                    read prData                write prData;
  property Model                  :TQUB_QVO_QVM                read prModel               write prModel;


end;


implementation


uses


QUB_QVO_QVD, QUB_QVO_QVT;


{$R *.DFM}


//******************************************************************************
function TQUB_QVO_QVD_Print_PropertiesForm.Print(Data :TQUB_QVO; Dlg :TPrintDialog; CopyImage :Boolean; ShowProperties :Boolean) :Boolean;
begin
  Result := False;
  if prPrinting then begin
    Result := True;
    Exit;
  end;

  prPrinting := True;
  try
    Self.Data      := Data;
    prDlg          := Dlg;
    Self.CopyImage := CopyImage;

    if ShowProperties then begin
      ShowModal;
    end else begin
      btnOKClick(Self);
    end;
  finally
    prPrinting := False;
  end;

end;

procedure TQUB_QVO_QVD_Print_PropertiesForm.prPrint;
var
  PageCount              :Integer;
  PageFirst              :Integer;
  PageLast               :Integer;
  PageCurr               :Integer;
  OldCanvas              :TCanvas;
  OldWidth               :Integer;
  OldHeight              :Integer;
  OldMargin              :TRect;
  OldDataDraw            :TQUB_FirstLast;
  OldDrawFrame           :Boolean;
  Text                   :String;
  PageRect               :TRect;
  TopTitleHeight         :Integer;
  BottomTitleHeight      :Integer;
  ScaleBarHeight         :Integer;
  t                      :Integer;
  h, w                   :Integer;
  sp                     :Integer;
  sc                     :Extended;
  x0, y0                 :Integer;
  DecVal                 :TFloatRec;
  sci                    :Integer;
  max                    :Integer;
  ChannelCount           :Integer;
  ChannelIndex           :Integer;
  VisChannelCount        :Integer;
  VisChannelIndex        :Integer;
begin

  Data.NoDraw             := True;

  OldCanvas               := Data.BitmapCanvas;
  OldWidth                := Data.DrawWidth;
  OldHeight               := Data.DrawHeight;
  OldMargin               := Data.Margin;
  if Data is TQUB_QVO_QVD then begin
    OldDataDraw           := TQUB_QVO_QVD(Data).DataDraw;
  end;
  OldDrawFrame            := Data.DrawFrame;

  try

    Data.DrawForPrinting  := True;
    Data.DrawBW           := BlackAndWhite;

    Data.DrawFrame        := DrawFrame;

    if Data is TQUB_QVO_QVD then begin
      TQUB_QVO_QVD(Data).PrintDrawData    := DrawData;
      TQUB_QVO_QVD(Data).PrintDrawIdlData := DrawIDLData;
      TQUB_QVO_QVD(Data).PrintDrawBase    := DrawBase;
      TQUB_QVO_QVD(Data).PrintDrawMode    := DrawMode;
    end else if Data is TQUB_QVO_QVT then begin
      TQUB_QVO_QVT(Data).PrintDrawData    := DrawData;
      TQUB_QVO_QVT(Data).PrintDrawIdlData := DrawIDLData;
      TQUB_QVO_QVT(Data).PrintDrawBase    := DrawBase;
      TQUB_QVO_QVT(Data).PrintDrawMode    := DrawMode;
    end;

    Data.PrintPenWidth    := Round(Printer.PageWidth / 2000);

    if Landscape then Printer.Orientation := poLandscape else Printer.Orientation := poPortrait;

    Data.BitmapCanvas     := Printer.Canvas;

    Data.DrawWidth        := Printer.PageWidth;
    Data.DrawHeight       := Printer.PageHeight;

    PageCurr              := 0;
    PageFirst             := -1;
    PageLast              := -1;

    try

      Printer.Title       := 'QuB Data';
      Printer.BeginDoc;

      try

        case prDlg.PrintRange of
          prAllPages:
            begin
              PageFirst := 0;
              PageLast  := Data.MaxPrintPage;
            end;
          prSelection:
            begin
              PageFirst := -1;
              PageLast  := -1;
            end;
          prPageNums:
            begin
              PageFirst := prDlg.FromPage - 1;
              PageLast  := prDlg.ToPage   - 1;
            end;
        end;//case PrintRange

        PageCount := PageLast - PageFirst + 1;
        if PageFirst > -1 then begin
          if Data is TQUB_QVO_QVD then begin
            TQUB_QVO_QVD(Data).Browse(QUB_QVO_QVD_CMD_GOTO_PAGE, PageFirst, max, False, False, True, False, False{Append}, True{SetZoom});
          end;
        end;

        if Data is TQUB_QVO_QVT then begin
          PageCount := 1;
        end;

        PageRect.Left   := Printer.PageWidth  div 20;
        PageRect.Top    := Printer.PageHeight div 20;
        PageRect.Right  := Printer.PageWidth  - PageRect.Left;
        PageRect.Bottom := Printer.PageHeight - PageRect.Top;

        with Data.BitmapCanvas do begin

          TopTitleHeight    := 0;
          BottomTitleHeight := 0;
          if ShowTitles then begin
            Font.Size  := 12;
            Font.Style := [fsBold];
            TopTitleHeight := TopTitleHeight + TextHeight('A');
            Font.Size  := 10;
            Font.Style := [];
            TopTitleHeight := TopTitleHeight + 3 * TextHeight('A');
            Font.Size  := 6;
            Font.Style := [fsBold];
            BottomTitleHeight := BottomTitleHeight + TextHeight('A');
          end;

          ScaleBarHeight := 0;
          if ScaleBars then begin
            Font.Size  := 8;
            Font.Style := [fsBold];
            ScaleBarHeight := ScaleBarHeight + 5 * TextHeight('A');
          end;

          Data.MarginLeft   := PageRect.Left;
          Data.MarginRight  := Printer.PageWidth - PageRect.Right;
          if ((ScaleBarsTop) and (TopTitleHeight >= ScaleBarHeight)) or (not ScaleBarsTop) then begin
            Data.MarginTop := PageRect.Top + TopTitleHeight;
          end else begin
            Data.MarginTop := PageRect.Top + ScaleBarHeight;
          end;
          if ((not ScaleBarsTop) and (BottomTitleHeight >= ScaleBarHeight)) or (ScaleBarsTop) then begin
            Data.MarginBottom := (Printer.PageHeight - PageRect.Bottom) + BottomTitleHeight;
          end else begin
            Data.MarginBottom := (Printer.PageHeight - PageRect.Bottom) + ScaleBarHeight;
          end;

          repeat

            if ShowTitles then begin

              Brush.Style := bsClear;
              TextFlags := 0;
              Font.Name  := 'Arial';
              Font.Color := clBlack;

              t := PageRect.Top;

              //FileName
              Font.Size  := 12;
              Font.Style := [fsBold];
              if Data is TQUB_QVO_QVD then begin
                Text := TQUB_QVO_QVD(Data).Filename;
              end else begin
                Text := TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).Filename;
              end;
              if not ShowFullPath then Text := ExtractFileName(Text);
              h := TextHeight(Text);
              w := TextWidth(Text);
              TextOut(PageRect.Left + (PageRect.Right - PageRect.Left) div 2 - w div 2, t, Text);
              t := t + h;

              //Selection
              Font.Size  := 10;
              Font.Style := [];
              if Data is TQUB_QVO_QVD then begin
                Text := Format('Drawing %d - %d, %d points/trace', [TQUB_QVO_QVD(Data).DataDraw.First, TQUB_QVO_QVD(Data).DataDraw.Last, TQUB_QVO_QVD(Data).PointsPerTrace]);
              end else begin
                Text := Format('Drawing %d - %d',                  [TQUB_QVO_QVT(Data).DataDraw.First + TQUB_QVO_QVT(Data).DataOffset, TQUB_QVO_QVT(Data).DataDraw.Last  + TQUB_QVO_QVT(Data).DataOffset]);
              end;
              h := TextHeight(Text);
              TextOut(PageRect.Left, t, Text);
              t := t + h;

              //Sampling
              if Data is TQUB_QVO_QVD then begin
                Text := Format('Sampling frequency: %.4g kHz', [(1 / TQUB_QVO_QVD(Data).Sampling) / 1000]);
              end else begin
                Text := Format('Sampling frequency: %.4g kHz', [(1 / TQUB_QVO_QVT(Data).Sampling) / 1000]);
              end;
              TextOut(PageRect.Left, t, Text);
              t := t + h;

              //Filter
              if Data is TQUB_QVO_QVD then begin
                if TQUB_QVO_QVD(Data).FilterOn[TQUB_QVO_QVD(Data).ActiveChannel] then begin
                  Text := Format('Low pass filter: %.4g kHz', [TQUB_QVO_QVD(Data).FilterFreq[TQUB_QVO_QVD(Data).ActiveChannel] / 1000]);
                end else begin
                  Text := 'Low pass filter: none';
                end;
              end else begin
                if TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).FilterOn[TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).ActiveChannel] then begin
                  Text := Format('Low pass filter: %.4g kHz', [TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).FilterFreq[TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).ActiveChannel] / 1000]);
                end else begin
                  Text := 'Low pass filter: none';
                end;
              end;
              TextOut(PageRect.Left, t, Text);

              //Date
              t := Printer.PageHeight - Data.Margin.Bottom;
              Font.Size  := 6;
              Font.Style := [fsBold];
              Text := 'QuB (C) 2000-07 University at Buffalo. ' + DateTimeToStr(Now) + '   Page ' + IntToStr(PageCurr + 1) + '/' + IntToStr(PageCount);
              TextOut(PageRect.Left, t, Text);

            end;

            //Draw
            Data.NoDraw := False;
            try
              Data.Draw;
            finally
              Data.NoDraw := True;
            end;

            if ScaleBars then begin
            //Draw Scale Bars

              Brush.Style := bsClear;
              Pen.Width := Data.PrintPenWidth;
              Pen.Color := clBlack;
              TextFlags := 0;
              Font.Name  := 'Arial';
              Font.Color := clBlack;
              Font.Size  := 8;
              Font.Style := [fsBold];

              h := TextHeight('A');

              if ScaleBarsTop then begin
                y0 := Data.Margin.Top - 2*h;
              end else begin
                y0 := PageRect.Bottom - 2*h;
              end;

              VisChannelCount := 0;
              if Data is TQUB_QVO_QVD then begin
                ChannelCount := TQUB_QVO_QVD(Data).DataChannelCount;
                for ChannelIndex := 0 to ChannelCount - 1 do begin
                  if TQUB_QVO_QVD(Data).ChannelVisible[ChannelIndex] then VisChannelCount := VisChannelCount + 1;
                end;
              end else begin
                ChannelCount := TQUB_QVO_QVT(Data).DataChannelCount;
                for ChannelIndex := 0 to ChannelCount - 1 do begin
                  if TQUB_QVO_QVT(Data).ChannelVisible[ChannelIndex] then VisChannelCount := VisChannelCount + 1;
                end;
              end;

              VisChannelIndex := 0;
              for ChannelIndex := 0 to ChannelCount - 1 do begin

                if Data is TQUB_QVO_QVD then begin
                  if not TQUB_QVO_QVD(Data).ChannelVisible[ChannelIndex] then Continue;
                end else begin
                  if not TQUB_QVO_QVT(Data).ChannelVisible[ChannelIndex] then Continue;
                end;

                x0 := Printer.PageWidth - Data.Margin.Right - (4 * h) - (9 * h) * (VisChannelCount - VisChannelIndex - 1);

                //y bar **************************************************************

                sp := 2 * h;//the initial number of pixels;
                if (ChannelIndex = 0) and (prAmpFromModel) and (Model <> nil) then begin
                  sc := Model.ModelObject.Amp[1] - Model.ModelObject.Amp[0];
                end else if Data is TQUB_QVO_QVD then begin
                  sc := ((TQUB_QVO_QVD(Data).ScaleBarADUnits[ChannelIndex] * sp) / TQUB_QVO_QVD(Data).ScaleBarPixel.y) / (TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex]);
                end else begin
                  sc := ((TQUB_QVO_QVT(Data).ScaleBarADUnits[ChannelIndex] * sp) / TQUB_QVO_QVT(Data).ScaleBarPixel.y) / (TQUB_QVO_QVT(Data).Scaling * TQUB_QVO_QVT(Data).DataChannelScaling[ChannelIndex]);
                end;
                FloatToDecimal(DecVal, sc, fvExtended, 1, 9999);
                sci := StrToInt(DecVal.Digits);
                if (sci > 1) and (sci < 4) then sci := 2 else if (sci >= 4) and (sci < 8) then sci := 5 else if sci >= 8 then sci := 10;
                sc := sci * Power(10, DecVal.Exponent - 1);
                if DecVal.Negative then sc := sc * -1;
                if Data is TQUB_QVO_QVD then begin
                  sp := Round((sc * TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex] * TQUB_QVO_QVD(Data).ScaleBarPixel.y) / TQUB_QVO_QVD(Data).ScaleBarADUnits[ChannelIndex]);
                end else begin
                  sp := Round((sc * TQUB_QVO_QVT(Data).Scaling * TQUB_QVO_QVT(Data).DataChannelScaling[ChannelIndex] * TQUB_QVO_QVT(Data).ScaleBarPixel.y) / TQUB_QVO_QVT(Data).ScaleBarADUnits[ChannelIndex]);
                end;

                MoveTo(x0, y0);
                LineTo(x0, y0 - sp);

                if Data is TQUB_QVO_QVD then begin
                  Text := Format('%.4g %s', [sc, TQUB_QVO_QVD(Data).DataChannelUnits[ChannelIndex]]);
                end else begin
                  Text := Format('%.4g %s', [sc, TQUB_QVO_QVT(Data).DataChannelUnits[ChannelIndex]]);
                end;
                w := TextWidth(Text);

                TextOut(x0 - w - TextWidth('I'), y0 - h, Text);

                //x bar
                sp := 2 * h;//the initial number of pixels;
                if Data is TQUB_QVO_QVD then begin
                  sc := ((TQUB_QVO_QVD(Data).ScaleBarSamples * sp) / TQUB_QVO_QVD(Data).ScaleBarPixel.x) * TQUB_QVO_QVD(Data).Sampling;
                end else begin
                  sc := ((TQUB_QVO_QVT(Data).ScaleBarSamples * sp) / TQUB_QVO_QVT(Data).ScaleBarPixel.x) * TQUB_QVO_QVT(Data).Sampling;
                end;
                FloatToDecimal(DecVal, sc, fvExtended, 1, 9999);
                sci := StrToInt(DecVal.Digits);
                if (sci > 1) and (sci < 4) then sci := 2 else if (sci >= 4) and (sci < 8) then sci := 5 else if sci >= 8 then sci := 10;
                sc := sci * Power(10, DecVal.Exponent - 1);
                if DecVal.Negative then sc := sc * -1;
                if Data is TQUB_QVO_QVD then begin
                  sp := Round(((sc / TQUB_QVO_QVD(Data).Sampling) * TQUB_QVO_QVD(Data).ScaleBarPixel.x) / TQUB_QVO_QVD(Data).ScaleBarSamples);
                end else begin
                  sp := Round(((sc / TQUB_QVO_QVT(Data).Sampling) * TQUB_QVO_QVT(Data).ScaleBarPixel.x) / TQUB_QVO_QVT(Data).ScaleBarSamples);
                end;
                MoveTo(x0,      y0);
                LineTo(x0 + sp, y0);

                Text := '%.4g sec';
                if (sc < 1e-3) then begin                                         //< 1ms
                  sc := sc * 1e6;
                  Text := '%.4g us';
                end else if (sc >= 1e-3) and (sc < 1) then begin                  //>= 1ms
                  sc := sc * 1e3;
                  Text := '%.4g ms';
                end else if (sc >= 1) then begin                                  //>= 1ms
                  Text := '%.4g s';
                end;
                Text := Format(Text, [sc]);

                w := TextWidth(Text);
                TextOut(x0 + sp - w, y0, Text);

                VisChannelIndex := VisChannelIndex + 1;

              end;

            end;  // ScaleBars

            Inc(PageCurr);
            if (PageCurr < PageCount) then begin

              if Data is TQUB_QVO_QVD then begin
                TQUB_QVO_QVD(Data).Browse(QUB_QVO_QVD_CMD_PAGE_DN, PageFirst, max, False, False, True, False, False{Append}, True{SetZoom});
              end;
              Printer.NewPage;

            end else begin

              Break;

            end;

          until(PageCurr >= PageCount);//pages

        end;//with Printer.Canvas


      finally

        Printer.EndDoc;

      end;

    except
    end;

  finally

    Data.DrawFrame       := OldDrawFrame;
    Data.DrawForPrinting := False;
    Data.BitmapCanvas    := OldCanvas;
    Data.DrawWidth       := OldWidth;
    Data.DrawHeight      := OldHeight;
    Data.MarginLeft      := OldMargin.Left;
    Data.MarginTop       := OldMargin.Top;
    Data.MarginRight     := OldMargin.Right;
    Data.MarginBottom    := OldMargin.Bottom;
    Data.NoDraw          := False;

    if Data is TQUB_QVO_QVD then begin
      TQUB_QVO_QVD(Data).Browse(QUB_QVO_QVD_CMD_GOTO_POINT, OldDataDraw.First, max, False, False, True, False, False{Append}, True{SetZoom});
    end else begin
      TQUB_QVO_QVT(Data).Draw;
    end;

  end;

end;


procedure TQUB_QVO_QVD_Print_PropertiesForm.prCopy;
var
  Metafile               :TMetafile;
  Bitmap                 :TBitmap;
  PageWidth              :Integer;
  PageHeight             :Integer;
  OldCanvas              :TCanvas;
  OldWidth               :Integer;
  OldHeight              :Integer;
  OldMargin              :TRect;
  OldDrawFrame           :Boolean;
  Text                   :String;
  PageRect               :TRect;
  TopTitleHeight         :Integer;
  BottomTitleHeight      :Integer;
  ScaleBarHeight         :Integer;
  t                      :Integer;
  h, w                   :Integer;
  sp                     :Integer;
  sc                     :Extended;
  x0, y0                 :Integer;
  DecVal                 :TFloatRec;
  sci                    :Integer;
  //max                    :Integer;
  Ff                     :Double;
  ChannelCount           :Integer;
  ChannelIndex           :Integer;
  VisChannelCount        :Integer;
  VisChannelIndex        :Integer;
begin

  Data.NoDraw            := True;

  OldCanvas              := Data.BitmapCanvas;
  OldWidth               := Data.DrawWidth;
  OldHeight              := Data.DrawHeight;
  OldMargin              := Data.Margin;
  OldDrawFrame           := Data.DrawFrame;

  Metafile               := nil;
  Bitmap                 := nil;

  try

    if Landscape then begin
      PageWidth            := DrawWidth;
      PageHeight           := DrawHeight;
    end else begin
      PageWidth            := DrawHeight;
      PageHeight           := DrawWidth;
    end;
    Data.DrawWidth       := PageWidth;
    Data.DrawHeight      := PageHeight;

    Ff := 1;

    if Self.Bitmap then begin
      Bitmap             := TBitmap.Create;
      Bitmap.Width       := Data.DrawWidth;
      Bitmap.Height      := Data.DrawHeight;
      Bitmap.PixelFormat := pf8bit;
      Data.BitmapCanvas  := Bitmap.Canvas;
    end else begin
      Metafile           := TMetafile.Create;
      Metafile.Width     := Data.DrawWidth;
      Metafile.Height    := Data.DrawHeight;
      Data.BitmapCanvas  := TMetafileCanvas.Create(Metafile, Data.BitmapCanvas.Handle);
    end;

    Data.DrawForPrinting := True;
    Data.DrawBW          := BlackAndWhite;

    Data.DrawFrame       := DrawFrame;

    if Data is TQUB_QVO_QVD then begin
      TQUB_QVO_QVD(Data).PrintDrawData    := DrawData;
      TQUB_QVO_QVD(Data).PrintDrawIdlData := DrawIDLData;
      TQUB_QVO_QVD(Data).PrintDrawBase    := DrawBase;
      TQUB_QVO_QVD(Data).PrintDrawMode    := DrawMode;
    end else if Data is TQUB_QVO_QVT then begin
      TQUB_QVO_QVT(Data).PrintDrawData    := DrawData;
      TQUB_QVO_QVT(Data).PrintDrawIdlData := DrawIDLData;
      TQUB_QVO_QVT(Data).PrintDrawBase    := DrawBase;
      TQUB_QVO_QVT(Data).PrintDrawMode    := DrawMode;
    end;

    Data.PrintPenWidth   := 0;
    Data.PrintFontSize   := prScaleBarFontSize;

    try

      try

        PageRect.Left   := 1;
        PageRect.Top    := 1;
        PageRect.Right  := PageWidth  - 1;
        PageRect.Bottom := PageHeight - 1;

        with Data.BitmapCanvas do begin

          TopTitleHeight    := 0;
          BottomTitleHeight := 0;
          if ShowFilename then begin
            Font.Size  := Round(Ff {* 1.50} * prScaleBarFontSize{12});
            Font.Style := [];//[fsBold];
            TopTitleHeight := TopTitleHeight + TextHeight('A');
          end;
          if ShowTitles then begin
            Font.Size  := Round(Ff {* 1.25} * prScaleBarFontSize);
            Font.Style := [];
            TopTitleHeight := TopTitleHeight + 3 * TextHeight('A');
            Font.Size  := Round(Ff * 0.75 * prScaleBarFontSize);
            Font.Style := [fsBold];
            BottomTitleHeight := BottomTitleHeight + TextHeight('A');
          end;

          ScaleBarHeight := 0;
          //if ScaleBars then begin
            Font.Size  := Round(Ff * prScaleBarFontSize);
            Font.Style := [fsBold];
            ScaleBarHeight := ScaleBarHeight + 5 * TextHeight('A');    // approx
          //end;

          Data.MarginLeft   := PageRect.Left;
          Data.MarginRight  := PageWidth - PageRect.Right;
          if ((ScaleBarsTop) and (TopTitleHeight >= ScaleBarHeight)) or (not ScaleBarsTop) then begin
            Data.MarginTop := PageRect.Top + TopTitleHeight;
          end else begin
            Data.MarginTop := PageRect.Top + ScaleBarHeight;
          end;
          if ((not ScaleBarsTop) and (BottomTitleHeight >= ScaleBarHeight)) or (ScaleBarsTop) then begin
            Data.MarginBottom := (PageHeight - PageRect.Bottom) + BottomTitleHeight;
          end else begin
            Data.MarginBottom := (PageHeight - PageRect.Bottom) + ScaleBarHeight;
          end;

          t := PageRect.Top;

          if ShowFilename then begin
          //FileName

            Brush.Style := bsClear;
            TextFlags   := 0;
            Font.Name   := 'Arial';
            Font.Color  := clBlack;
            Font.Size   := Round(Ff {* 1.50} * prScaleBarFontSize{12});
            Font.Style  := [];//[fsBold];
            if Data is TQUB_QVO_QVD then begin
              Text      := TQUB_QVO_QVD(Data).Filename;
            end else begin
              Text      := TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).Filename;
            end;
            if not ShowFullPath then Text := ExtractFileName(Text);
            h           := TextHeight(Text);
            w           := TextWidth(Text);
            TextOut(PageRect.Left + (PageRect.Right - PageRect.Left) div 2 - w div 2, t, Text);
            t := t + h;

          end;

          if ShowTitles then begin
          //Selection, Sampling, Filter

            Brush.Style := bsClear;
            TextFlags := 0;
            Font.Name  := 'Arial';
            Font.Color := clBlack;

            //Selection
            Font.Size  := Round(Ff {* 1.25} * prScaleBarFontSize{10});
            Font.Style := [];
            if Data is TQUB_QVO_QVD then begin
              Text := Format('Drawing %d - %d, %d points/trace', [TQUB_QVO_QVD(Data).DataDraw.First, TQUB_QVO_QVD(Data).DataDraw.Last, TQUB_QVO_QVD(Data).PointsPerTrace]);
            end else begin
              Text := Format('Drawing %d - %d',                  [TQUB_QVO_QVT(Data).DataDraw.First, TQUB_QVO_QVT(Data).DataDraw.Last]);
            end;
            h := TextHeight(Text);
            TextOut(PageRect.Left, t, Text);
            t := t + h;

            //Sampling
            if Data is TQUB_QVO_QVD then begin
              Text := Format('Sampling frequency: %.4g kHz', [(1 / TQUB_QVO_QVD(Data).Sampling) / 1000]);
            end else begin
              Text := Format('Sampling frequency: %.4g kHz', [(1 / TQUB_QVO_QVT(Data).Sampling) / 1000]);
            end;
            TextOut(PageRect.Left, t, Text);
            t := t + h;

            //Filter
            if Data is TQUB_QVO_QVD then begin
              if TQUB_QVO_QVD(Data).FilterOn[TQUB_QVO_QVD(Data).ActiveChannel] then begin
                Text := Format('Low pass filter: %.4g kHz', [TQUB_QVO_QVD(Data).FilterFreq[TQUB_QVO_QVD(Data).ActiveChannel] / 1000]);
              end else begin
                Text := 'Low pass filter: none';
              end;
            end else begin
              if TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).FilterOn[TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).ActiveChannel] then begin
                Text := Format('Low pass filter: %.4g kHz', [TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).FilterFreq[TQUB_QVO_QVD(TQUB_QVO_QVT(Data).DataSource).ActiveChannel] / 1000]);
              end else begin
                Text := 'Low pass filter: none';
              end;
            end;
            TextOut(PageRect.Left, t, Text);

            //Date
            t := PageHeight - Data.Margin.Bottom;
            Font.Size  := Round(Ff * 0.75 * prScaleBarFontSize{6});
            Font.Style := [fsBold];
            Text := 'QuB (C) 2000-07 University at Buffalo. ' + DateTimeToStr(Now);
            TextOut(PageRect.Left, t, Text);

          end;//ShowTitles

          //Draw
          Data.NoDraw := False;
          try
            Data.Draw;
          finally
            Data.NoDraw := True;
          end;

          //if ScaleBars then begin
          //Draw Scale Bars

            Brush.Style := bsClear;
            Pen.Width := Data.PrintPenWidth;
            Pen.Color := clBlack;
            TextFlags := 0;
            Font.Name  := 'Arial';
            Font.Color := clBlack;
            Font.Size  := Round(Ff * 1.00 * prScaleBarFontSize{8});
            Font.Style := [fsBold];

            h := TextHeight('A');

            if ScaleBarsTop then begin
              y0 := Data.Margin.Top - 2*h;
            end else begin
              y0 := PageRect.Bottom - 2*h;
            end;

            VisChannelCount := 0;
            if Data is TQUB_QVO_QVD then begin
              ChannelCount := TQUB_QVO_QVD(Data).DataChannelCount;
              for ChannelIndex := 0 to ChannelCount - 1 do begin
                if TQUB_QVO_QVD(Data).ChannelVisible[ChannelIndex] then VisChannelCount := VisChannelCount + 1;
              end;
            end else begin
              ChannelCount := TQUB_QVO_QVT(Data).DataChannelCount;
              for ChannelIndex := 0 to ChannelCount - 1 do begin
                if TQUB_QVO_QVT(Data).ChannelVisible[ChannelIndex] then VisChannelCount := VisChannelCount + 1;
              end;
            end;

            VisChannelIndex := 0;
            for ChannelIndex := 0 to ChannelCount - 1 do begin

              if Data is TQUB_QVO_QVD then begin
                if not TQUB_QVO_QVD(Data).ChannelVisible[ChannelIndex] then Continue;
              end else begin
                if not TQUB_QVO_QVT(Data).ChannelVisible[ChannelIndex] then Continue;
              end;

              x0 := PageWidth - Data.Margin.Right - (4 * h) - (6 * h) * (VisChannelCount - VisChannelIndex - 1);

              //y bar **************************************************************

              sp := 2 * h;//the initial number of pixels;
              if (ChannelIndex = 0) and (prAmpFromModel) and (Model <> nil) then begin
                sc := Model.ModelObject.Amp[1] - Model.ModelObject.Amp[0];
              end else if Data is TQUB_QVO_QVD then begin
                sc := ((TQUB_QVO_QVD(Data).ScaleBarADUnits[ChannelIndex] * sp) / TQUB_QVO_QVD(Data).ScaleBarPixel.y) / (TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex]);
              end else begin
                sc := ((TQUB_QVO_QVT(Data).ScaleBarADUnits[ChannelIndex] * sp) / TQUB_QVO_QVT(Data).ScaleBarPixel.y) / (TQUB_QVO_QVT(Data).Scaling * TQUB_QVO_QVT(Data).DataChannelScaling[ChannelIndex]);
              end;
              FloatToDecimal(DecVal, sc, fvExtended, 1, 9999);
              sci := StrToInt(DecVal.Digits);
              if (sci > 1) and (sci < 4) then sci := 2 else if (sci >= 4) and (sci < 8) then sci := 5 else if sci >= 8 then sci := 10;
              sc := sci * Power(10, DecVal.Exponent - 1);
              if DecVal.Negative then sc := sc * -1;
              if Data is TQUB_QVO_QVD then begin
                sp := Round((sc * TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex] * TQUB_QVO_QVD(Data).ScaleBarPixel.y) / TQUB_QVO_QVD(Data).ScaleBarADUnits[ChannelIndex]);
              end else begin
                sp := Round((sc * TQUB_QVO_QVT(Data).Scaling * TQUB_QVO_QVT(Data).DataChannelScaling[ChannelIndex] * TQUB_QVO_QVT(Data).ScaleBarPixel.y) / TQUB_QVO_QVT(Data).ScaleBarADUnits[ChannelIndex]);
              end;

              if ScaleBars then MoveTo(x0, y0);
              if ScaleBars then LineTo(x0, y0 - sp);

              if Data is TQUB_QVO_QVD then begin
                Text := Format('%.4g %s', [sc, TQUB_QVO_QVD(Data).DataChannelUnits[ChannelIndex]]);
              end else begin
                Text := Format('%.4g %s', [sc, TQUB_QVO_QVT(Data).DataChannelUnits[ChannelIndex]]);
              end;
              w := TextWidth(Text);

              if ScaleBars then TextOut(x0 - w - TextWidth('I'), y0 - h, Text);

              //x bar
              sp := 2 * h;//the initial number of pixels;
              if Data is TQUB_QVO_QVD then begin
                sc := ((TQUB_QVO_QVD(Data).ScaleBarSamples * sp) / TQUB_QVO_QVD(Data).ScaleBarPixel.x) * TQUB_QVO_QVD(Data).Sampling;
              end else begin
                sc := ((TQUB_QVO_QVT(Data).ScaleBarSamples * sp) / TQUB_QVO_QVT(Data).ScaleBarPixel.x) * TQUB_QVO_QVT(Data).Sampling;
              end;
              FloatToDecimal(DecVal, sc, fvExtended, 1, 9999);
              sci := StrToInt(DecVal.Digits);
              if (sci > 1) and (sci < 4) then sci := 2 else if (sci >= 4) and (sci < 8) then sci := 5 else if sci >= 8 then sci := 10;
              sc := sci * Power(10, DecVal.Exponent - 1);
              if DecVal.Negative then sc := sc * -1;
              if Data is TQUB_QVO_QVD then begin
                sp := Round(((sc / TQUB_QVO_QVD(Data).Sampling) * TQUB_QVO_QVD(Data).ScaleBarPixel.x) / TQUB_QVO_QVD(Data).ScaleBarSamples);
              end else begin
                sp := Round(((sc / TQUB_QVO_QVT(Data).Sampling) * TQUB_QVO_QVT(Data).ScaleBarPixel.x) / TQUB_QVO_QVT(Data).ScaleBarSamples);
              end;
              if ScaleBars then MoveTo(x0,      y0);
              if ScaleBars then LineTo(x0 + sp, y0);

              Text := '%.4g sec';
              if (sc < 1e-3) then begin                                           //< 1ms
                sc := sc * 1e6;
                Text := '%.4g us';
              end else if (sc >= 1e-3) and (sc < 1) then begin                    //>= 1ms
                sc := sc * 1e3;
                Text := '%.4g ms';
              end else if (sc >= 1) then begin                                    //>= 1ms
                Text := '%.4g s';
              end;
              Text := Format(Text, [sc]);

              w := TextWidth(Text);
              if ScaleBars then TextOut(x0 + sp - w, y0, Text);

              VisChannelIndex := VisChannelIndex + 1;

            end;

          //end;//Show ScaleBars

        end;//with BitmapCanvas

      finally

        if not Self.Bitmap then begin
          Data.BitmapCanvas.Free;
          Clipboard.Assign(Metafile);
        end else begin
          Clipboard.Assign(Bitmap);
        end;

      end;

    except
    end;

  finally

    Metafile.Free;
    Bitmap.Free;

    Data.DrawFrame       := OldDrawFrame;
    Data.DrawForPrinting := False;
    Data.DrawBW          := False;
    Data.BitmapCanvas    := OldCanvas;
    Data.DrawWidth       := OldWidth;
    Data.DrawHeight      := OldHeight;
    Data.MarginLeft      := OldMargin.Left;
    Data.MarginTop       := OldMargin.Top;
    Data.MarginRight     := OldMargin.Right;
    Data.MarginBottom    := OldMargin.Bottom;
    Data.NoDraw          := False;

    if Data is TQUB_QVO_QVD then begin
      TQUB_QVO_QVD(Data).Draw;
    end else begin
      TQUB_QVO_QVT(Data).Draw;
    end;

  end;

end;

procedure TQUB_QVO_QVD_Print_PropertiesForm.prSetCopyImage(c :Boolean);
begin
  prCopyImage            := c;
  chkBitmap.Enabled      := c;
  chkPicture.Enabled     := c;
  lblImageFormat.Enabled := c;
  lblWidth.Enabled       := c;
  lblHeight.Enabled      := c;
  txtWidth.Enabled       := c;
  txtHeight.Enabled      := c;
  if c then btnOK.Caption := '&Copy' else btnOK.Caption := '&Print';
end;

//******************************************************************************
procedure TQUB_QVO_QVD_Print_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Draw Data"]',
                        'Draw Data', REGTYPE_BOOL, DEF_DRAWDATA,
                        chkDrawData));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Draw IDL Data"]',
                        'Draw IDL Data', REGTYPE_BOOL, DEF_DRAWIDLDATA,
                        chkDrawIdlData));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Draw Base"]',
                        'Draw Base', REGTYPE_BOOL, DEF_DRAWBASE,
                        chkDrawBase));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Invert Data"]',
                        'Invert Data', REGTYPE_BOOL, DEF_INVERTDATA,
                        chkInvertData));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Draw Frame"]',
                        'Draw Frame', REGTYPE_BOOL, DEF_DRAWFRAME,
                        chkDrawFrame));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Show Filename"]',
                        'Show Filename', REGTYPE_BOOL, DEF_SHOWFILENAME,
                        chkShowFilename));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Show Full Path"]',
                        'Show Filename', REGTYPE_BOOL, DEF_SHOWFULLPATH,
                        chkShowFullPath));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Show Titles"]',
                        'Show Titles', REGTYPE_BOOL, DEF_SHOWTITLES,
                        chkShowTitles));
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Draw Mode"]')
                        .AddButton(TQUB_Property_RadioButton.Create(optDrawMode0, 0))
                        .AddButton(TQUB_Property_RadioButton.Create(optDrawMode1, 1))
                        .AddButton(TQUB_Property_RadioButton.Create(optDrawMode2, 2))
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Scale Bars"]',
                        'Scale Bars', REGTYPE_BOOL, DEF_SCALEBARS,
                        chkScaleBars)
                      .Enables(chkScaleBarsTop)
                      .Enables(chkScaleBarsBottom)
                      .Enables(chkScaleBarTime)
                      .Enables(chkScaleBarAmp)
                      //.Enables(txtScaleBarFontSize)
                      //.Enables(udScaleBarFontSize)
                     );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Scale Bars Location"]',
                          '', REGTYPE_NONE, DEF_SCALEBARSTOP)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkScaleBarsBottom,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkScaleBarsTop,
                                       1,
                                       1))
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Scale Bar Time"]',
                        'Scale Bar Time', REGTYPE_BOOL, DEF_SCALEBARTIME,
                        chkScaleBarTime));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Scale Bar Amp"]',
                        'Scale Bar Amp', REGTYPE_BOOL, DEF_SCALEBARAMP,
                        chkScaleBarAmp));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["Scale Bar Font Size"]',
                      'Scale Bar Font Size', REGTYPE_INT, DEF_SCALEBARFONTSIZE,
                      udScaleBarFontSize,
                      txtScaleBarFontSize,
                      AcceptIntBetween(MIN_SCALEBARFONTSIZE, MAX_SCALEBARFONTSIZE)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Amp From Model"]',
                        'Amp From Model', REGTYPE_BOOL, DEF_AMPFROMMODEL,
                        chkAmpFromModel));
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Orientation"]',
                          '', REGTYPE_INT, DEF_LANDSCAPE)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkPortrait,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkLandscape,
                                       1,
                                       1))
                       );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Format"]',
                          '', REGTYPE_INT, DEF_BITMAP)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkPicture,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkBitmap,
                                       1,
                                       1))
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["Draw Width"]',
                    'Draw Width', REGTYPE_INT, DEF_WIDTH,
                    txtWidth,
                    AcceptIntBetween(MIN_WIDTH, MAX_WIDTH),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Draw Height"]',
                    'Draw Height', REGTYPE_INT, DEF_HEIGHT,
                    txtHeight,
                    AcceptIntBetween(MIN_HEIGHT, MAX_HEIGHT),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["BlackAndWhite"]',
                        'BlackAndWhite', REGTYPE_BOOL, DEF_BLACKANDWHITE,
                        chkBlackAndWhite));

  RegKeyName            := 'Software\QUB\Print Data';
  ProfileCat            := 'CopyData';
  prBeforeGetProperties := PropertiesToParams;

end;

procedure TQUB_QVO_QVD_Print_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;

//******************************************************************************
procedure TQUB_QVO_QVD_Print_PropertiesForm.btnOKClick(Sender: TObject);
begin
  inherited;
  if CopyImage then begin
    prCopy;
  end else begin
    prPrint;
  end;
end;

//******************************************************************************
procedure TQUB_QVO_QVD_Print_PropertiesForm.PropertiesToParams;
begin
  prDrawFrame        := 0<>Properties['Draw Frame'].Data[0];
  prDrawData         := 0<>Properties['Draw Data'].Data[0];
  prDrawIDLData      := 0<>Properties['Draw IDL Data'].Data[0];
  prDrawBase         := 0<>Properties['Draw Base'].Data[0];
  prDrawMode         := TQUB_QVO_QVD_DrawMode(Properties['Draw Mode'].Data[0]);
  prInvertData       := 0<>Properties['Invert Data'].Data[0];
  prDrawFrame        := 0<>Properties['Draw Frame'].Data[0];
  prShowFilename     := 0<>Properties['Show Filename'].Data[0];
  prShowFullPath     := 0<>Properties['Show Full Path'].Data[0];
  prShowTitles       := 0<>Properties['Show Titles'].Data[0];
  prScaleBars        := 0<>Properties['Scale Bars'].Data[0];
  prScaleBarsTop     := 0<>Properties['Scale Bars Location'].Data[0];
  prScaleBarTime     := 0<>Properties['Scale Bar Time'].Data[0];
  prScaleBarAmp      := 0<>Properties['Scale Bar Amp'].Data[0];
  prScaleBarFontSize := Properties['Scale Bar Font Size'].Data[0];
  prAmpFromModel     := 0<>Properties['Amp From Model'].Data[0];
  prLandscape        := 0<>Properties['Orientation'].Data[0];
  prDrawWidth        := Properties['Draw Width'].Data[0];
  prDrawHeight       := Properties['Draw Height'].Data[0];
  prBitmap           := 0<>Properties['Format'].Data[0];
  prBlackAndWhite    := 0<>Properties['BlackAndWhite'].Data[0];
end;


//******************************************************************************
end.
