{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// edits properties of a selection in a sel.list (right-click the entry in the sel.list)


unit QUB_QVO_QVD_SelectionProperties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Buttons, Registry,

QUB_Global,
QUB_QVO_QVD_Global;


const


LABEL_MAX_COUNT   = 20;


type


TQUB_QVO_QVD_SelectionPropertiesForm = class(TForm)


  lblSelection                       :TLabel;
  Label2                             :TLabel;
  Label3                             :TLabel;
  txtComment                         :TMemo;
  btnOK                              :TButton;
  btnCancel                          :TButton;
  lstLabel                           :TComboBox;
  Label1                             :TLabel;
  txtSelFirst                        :TEdit;
  txtSelLast                         :TEdit;
    btnShiftLeft: TSpeedButton;
    btnShiftRight: TSpeedButton;
    txtShiftSel: TEdit;
  btnHelp                            : TButton;


  procedure FormDestroy(Sender: TObject);
  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormKeyPress(Sender: TObject; var Key: Char);
  procedure lstLabelClick(Sender: TObject);
  procedure lstLabelKeyPress(Sender: TObject; var Key: Char);
  procedure lstLabelExit(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);


protected


  procedure prPropertiesRegistry(Read :Boolean);


end;


implementation


{$R *.DFM}


//******************************************************************************
procedure TQUB_QVO_QVD_SelectionPropertiesForm.FormCreate(Sender: TObject);
begin
  //read properties from registry
  prPropertiesRegistry(True{Read});
  if lstLabel.Items.Count > 0 then begin
    lstLabel.ItemIndex := 0;
  end else begin
    lstLabel.Text := 'Note';
  end;
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.FormDestroy(Sender: TObject);
begin
  //save properties in registry
  prPropertiesRegistry(False{Read});
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.FormShow(Sender: TObject);
begin
  lstLabel.SetFocus;
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Word(Key) = VK_ESCAPE then begin
    Key := #0;
    ModalResult := mrCancel;
  end;
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.lstLabelClick(Sender: TObject);
begin
  lstLabelExit(Sender);
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.lstLabelKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    lstLabelExit(Sender);
    if Trim(lstLabel.Text) <> '' then begin
      btnOK.SetFocus;
    end else begin
      btnCancel.SetFocus;
    end;
  end;
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.lstLabelExit(Sender: TObject);
var
  Idx :Integer;
  Txt :String;
begin
  Txt := Trim(lstLabel.Text);
  if Txt = '' then Exit;//?!?!
  Idx := lstLabel.Items.IndexOf(Txt);
  if Idx < 0 then begin
    lstLabel.Items.Insert(0, Txt);
    Idx := 0;
  end else if lstLabel.Items.Strings[Idx] <> Txt then begin
  //for differences in case
    lstLabel.Items.Strings[Idx] := Txt;
  end;
  if Idx > 0 then begin
    lstLabel.Items.Move(Idx, 0);
    lstLabel.ItemIndex := 0;
  end;
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.FormResize(Sender: TObject);
begin
  btnOK.Top := ClientHeight - txtSelFirst.Top - btnOK.Height;
  btnCancel.Top := btnOK.Top;
  txtComment.Height := btnOK.Top - txtComment.Top - 10;
  txtComment.Width  := ClientWidth - txtComment.Left - 10;
end;

procedure TQUB_QVO_QVD_SelectionPropertiesForm.btnHelpClick(
  Sender: TObject);
begin
  ShowWiki('Selection_Properties', Handle);
end;


//Registry *********************************************************************
procedure TQUB_QVO_QVD_SelectionPropertiesForm.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
  i   :Integer;
begin
  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Data Selection Properties', True) then begin

        for i := 0 to LABEL_MAX_COUNT - 1 do begin
          if Read then begin
            if ValueExists('Label ' + IntToStr(i)) then begin
              lstLabel.Items.Add(ReadString('Label ' + IntToStr(i)));
            end else begin
              Break;
            end;
          end else begin
            if i > (lstLabel.Items.Count - 1) then begin
              DeleteKey('Label ' + IntToStr(i));
            end else begin
              WriteString('Label ' + IntToStr(i), lstLabel.Items[i]);
            end;
          end;
        end;

        if Read then begin
          txtComment.Text := ReadString('Notes');
        end else begin
          WriteString('Notes', txtComment.Text);
        end;

      end;
      CloseKey;

    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
end.
