{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
//The editor of a generalized rate constraint


unit QUB_QVO_QVM_GenConstraint_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Grids, ExtCtrls,

QUB_Global,
QUB_QVO_Properties,
QUB_QFS_Channels, QUB_QFS_QMF,
QUB_Table;


type


TQUB_QVO_QVM_GenConstraint_PropertiesForm = class(TQUB_QVO_PropertiesForm)


  panGrid: TPanel;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);



protected


  prModel                        :TQUB_QFS_QMF;
  prConstraint                   :TQUB_QFS_Constraint;
  

public


  grdConstraint                  :TQUB_Table;


  property Model                 :TQUB_QFS_QMF        read prModel      write prModel;
  property Constraint            :TQUB_QFS_Constraint read prConstraint write prConstraint;


end;


implementation


{$R *.DFM}


//******************************************************************************
procedure TQUB_QVO_QVM_GenConstraint_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;
  grdConstraint                  := TQUB_Table.Create(Self);
  grdConstraint.Parent           := panGrid;
  grdConstraint.Align            := alClient;
  grdConstraint.DefaultColWidth  := 60;
  grdConstraint.DefaultRowHeight := 18;
  grdConstraint.FixedCols        := 1;
  grdConstraint.FixedRows        := 1;
  grdConstraint.ColCount         := 2;
  grdConstraint.RowCount         := 2;
  grdConstraint.Options          := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRangeSelect, goEditing, goTabs, goThumbTracking];
  grdConstraint.TabOrder         := 0;
end;

procedure TQUB_QVO_QVM_GenConstraint_PropertiesForm.FormShow(Sender: TObject);
var
  r :Integer;
begin
  inherited;

  with grdConstraint, Model do begin

    ColCount := 1{fixed} + Rates.RateCount * 4 + ExtraKineticPars.ExtraKineticParCount + 1{Value};

    Cells[0, 1] := 'Factor';

    for r := 0 to Rates.RateCount - 1 do begin

      Cells[r*4 + 1, 0] := Format('ln(k0[%d,%d])', [Rates[r].StateIndex[1]+1, Rates[r].StateIndex[2]+1]);
      Cells[r*4 + 2, 0] := Format('k1[%d,%d]',     [Rates[r].StateIndex[1]+1, Rates[r].StateIndex[2]+1]);
      Cells[r*4 + 3, 0] := Format('ln(k0[%d,%d])', [Rates[r].StateIndex[2]+1, Rates[r].StateIndex[1]+1]);
      Cells[r*4 + 4, 0] := Format('k1[%d,%d]',     [Rates[r].StateIndex[2]+1, Rates[r].StateIndex[1]+1]);

      if prConstraint <> nil then begin
        Cells[r*4 + 1, 1] := Choice(prConstraint.Coeff[r*4 + 0] = 0, '', FloatToStr(prConstraint.Coeff[r*4 + 0]));
        Cells[r*4 + 2, 1] := Choice(prConstraint.Coeff[r*4 + 1] = 0, '', FloatToStr(prConstraint.Coeff[r*4 + 1]));
        Cells[r*4 + 3, 1] := Choice(prConstraint.Coeff[r*4 + 2] = 0, '', FloatToStr(prConstraint.Coeff[r*4 + 2]));
        Cells[r*4 + 4, 1] := Choice(prConstraint.Coeff[r*4 + 3] = 0, '', FloatToStr(prConstraint.Coeff[r*4 + 3]));
      end else begin
        Cells[r*4 + 1, 1] := '';
        Cells[r*4 + 2, 1] := '';
        Cells[r*4 + 3, 1] := '';
        Cells[r*4 + 4, 1] := '';
      end;

    end;

    for r := 0 to ExtraKineticPars.ExtraKineticParCount - 1 do begin

      Cells[Rates.RateCount*4 + r + 1, 0] := 'ln(' + ExtraKineticPars[r].Name + ')';
      if prConstraint <> nil then begin
        Cells[Rates.RateCount*4 + r + 1, 1] := Choice(prConstraint.Coeff[Rates.RateCount*4 + r] = 0, '', FloatToStr(prConstraint.Coeff[Rates.RateCount*4 + r]));
      end else begin
        Cells[Rates.RateCount*4 + r + 1, 1] := '';
      end;

    end;

    //Value
    Cells[Rates.RateCount*4 + ExtraKineticPars.ExtraKineticParCount + 1, 0] := 'Value';
    if prConstraint <> nil then begin
      Cells[Rates.RateCount*4 + ExtraKineticPars.ExtraKineticParCount + 1, 1] := FloatToStr(prConstraint.Value);
    end else begin
      Cells[Rates.RateCount*4 + ExtraKineticPars.ExtraKineticParCount + 1, 1] := '0';
    end;

  end;
  ActiveControl := grdConstraint;
end;


//******************************************************************************
end.
