{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_QVO_QVM_Print_Properties;

// properties form for copying and printing the model

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls,
Printers, Math, Registry, Clipbrd,

QUB_QFS_Channels,
QUB_Global,
QUB_QVO, QUB_QVO_QVD_Global,
QUB_Profiles,
QUB_Properties, QUB_Task_Properties, 
QUB_Node;


const


DEF_DRAWFRAME        = 1;
DEF_SHOWTITLES       = 1;
DEF_NAMEDSTATES      = 0;
DEF_DRAWSTATEPANEL   = 1;
DEF_DRAWSTATEPANELCOLOR = 1;
DEF_SHOWK1           = 0;
DEF_DIAGONALRATES    = 1;
DEF_LANDSCAPE        = 1;
DEF_BITMAP           = 1;
MIN_WIDTH            = 20;
MAX_WIDTH            = 10000;
DEF_WIDTH            = 800;
MIN_HEIGHT           = 20;
MAX_HEIGHT           = 10000;
DEF_HEIGHT           = 600;
DEF_STATEFONTSIZE    = 12;
DEF_RATEFONTSIZE     = 8;
MIN_FONTSIZE         = 4;
MAX_FONTSIZE         = 20;
DEF_STATEPANELSIZE   = 20;
MIN_STATEPANELSIZE   = 10;
MAX_STATEPANELSIZE   = 100;


type


TQUB_QVO_QVM_Print_PropertiesForm = class(TQUB_Task_PropertiesForm)


  panOptions                      :TPanel;
  chkDrawFrame                    :TCheckBox;
  chkShowTitles                   :TCheckBox;
  lblOrientation                  :TLabel;
  Panel3                          :TPanel;
  chkPortrait                     :TRadioButton;
  chkLandscape                    :TRadioButton;
  lblImageFormat                  :TLabel;
  Panel1                          :TPanel;
  chkPicture                      :TRadioButton;
  chkBitmap                       :TRadioButton;
  txtWidth                        :TEdit;
  txtHeight                       :TEdit;
  lblWidth                        :TLabel;
  lblHeight                       :TLabel;
  Label2                          :TLabel;
  chkNamedStates                  :TCheckBox;
  lblFontSize                     :TLabel;
  txtStateFontSize                :TEdit;
  udStateFontSize                 :TUpDown;
  chkDiagonalRates                :TCheckBox;
  chkShowK1                       :TCheckBox;
  Label1                          :TLabel;
  txtRateFontSize                 :TEdit;
  udRateFontSize                  :TUpDown;
  Label3                          :TLabel;
  txtStatePanelSize               :TEdit;
  udStatePanelSize                :TUpDown;
  chkDrawStatePanel               :TCheckBox;
  chkDrawStatePanelColor          :TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure btnOKClick(Sender: TObject);


protected


  prModel                         :TQUB_QVO;

  prDlg                           :TPrintDialog;

  prPrinting                      :Boolean;

  prDrawFrame                     :Boolean;
  prShowTitles                    :Boolean;
  prLandscape                     :Boolean;
  prDrawWidth                     :Integer;
  prDrawHeight                    :Integer;
  prBitmap                        :Boolean;
  prNamedStates                   :Boolean;
  prDrawStatePanel                :Boolean;
  prDrawStatePanelColor           :Boolean;
  prStatePanelSize                :Integer;
  prStateFontSize                 :Integer;
  prRateFontSize                  :Integer;
  prShowk1                        :Boolean;
  prDiagonalRates                 :Boolean;
  prCopyImage                     :Boolean;


  procedure prSetCopyImage(c :Boolean);

  procedure prPrint;
  procedure prCopy;


public


  procedure PropertiesToParams;

  function Print(Model :TQUB_QVO; Dlg :TPrintDialog; CopyImage :Boolean; ShowProperties :Boolean) :Boolean;


  property DrawFrame              :Boolean           read prDrawFrame;
  property ShowTitles             :Boolean           read prShowTitles;
  property Landscape              :Boolean           read prLandscape;
  property Bitmap                 :Boolean           read prBitmap;
  property DrawWidth              :Integer           read prDrawWidth;
  property DrawHeight             :Integer           read prDrawHeight;

  property NamedStates            :Boolean           read prNamedStates;
  property DrawStatePanel         :Boolean           read prDrawStatePanel;
  property DrawStatePanelColor    :Boolean           read prDrawStatePanelColor;
  property StatePanelSize         :Integer           read prStatePanelSize;
  property StateFontSize          :Integer           read prStateFontSize;
  property RateFontSize           :Integer           read prRateFontSize;
  property Showk1                 :Boolean           read prShowk1;
  property DiagonalRates          :Boolean           read prDiagonalRates;

  property CopyImage              :Boolean           read prCopyImage           write prSetCopyImage;

  property Model                  :TQUB_QVO          read prModel               write prModel;


end;


implementation


uses


QUB_QVO_QVM;


{$R *.DFM}


//******************************************************************************
function TQUB_QVO_QVM_Print_PropertiesForm.Print(Model :TQUB_QVO; Dlg :TPrintDialog; CopyImage :Boolean; ShowProperties :Boolean) :Boolean;
begin
  Result := False;
  if prPrinting then begin
    Result := True;
    Exit;
  end;

  prPrinting := True;
  try
    Self.Model     := Model;
    prDlg          := Dlg;
    Self.CopyImage := CopyImage;

    if ShowProperties then ShowModal else btnOkClick(Self);
  finally
    prPrinting := False;
  end;

end;

procedure TQUB_QVO_QVM_Print_PropertiesForm.prPrint;
var
  OldCanvas               :TCanvas;
  OldWidth                :Integer;
  OldHeight               :Integer;
  OldMargin               :TRect;
  OldDrawFrame            :Boolean;
  OldStatePanelSize       :Integer;
  OldShowk1               :Boolean;
  OldDiagonalRates        :Boolean;
  Text                    :String;
  PageRect                :TRect;
  TopTitleHeight          :Integer;
  BottomTitleHeight       :Integer;
  t                       :Integer;
begin

  Model.NoDraw            := True;

  OldCanvas               := Model.BitmapCanvas;
  OldWidth                := Model.DrawWidth;
  OldHeight               := Model.DrawHeight;
  OldMargin               := Model.Margin;
  OldDrawFrame            := Model.DrawFrame;

  OldStatePanelSize       := TQUB_QVO_QVM(Model).StateSize;
  OldShowk1               := TQUB_QVO_QVM(Model).Showk1;
  OldDiagonalRates        := TQUB_QVO_QVM(Model).DiagonalRates;

  try

    Model.DrawForPrinting := True;

    Model.DrawFrame       := DrawFrame;

    Model.PrintPenWidth   := Round(Printer.PageWidth / 2000);
    TQUB_QVO_QVM(Model).PrintNamedStates    := NamedStates;
    TQUB_QVO_QVM(Model).PrintDrawStatePanel := DrawStatePanel;
    TQUB_QVO_QVM(Model).PrintDrawStatePanelColor := DrawStatePanelColor;
    TQUB_QVO_QVM(Model).PrintStateFontSize  := StateFontSize;
    TQUB_QVO_QVM(Model).PrintRateFontSize   := RateFontSize;
    TQUB_QVO_QVM(Model).StateSize           := StatePanelSize;
    TQUB_QVO_QVM(Model).Showk1              := Showk1;
    TQUB_QVO_QVM(Model).DiagonalRates       := DiagonalRates;

    if Landscape then Printer.Orientation := poLandscape else Printer.Orientation := poPortrait;

    Model.BitmapCanvas    := Printer.Canvas;

    Model.DrawWidth       := Printer.PageWidth;
    Model.DrawHeight      := Printer.PageHeight;

    try

      Printer.Title       := 'QuB Model';
      Printer.BeginDoc;

      try

        PageRect.Left   := Printer.PageWidth  div 20;
        PageRect.Top    := Printer.PageHeight div 20;
        PageRect.Right  := Printer.PageWidth  - PageRect.Left;
        PageRect.Bottom := Printer.PageHeight - PageRect.Top;

        with Model.BitmapCanvas do begin

          TopTitleHeight    := 0;
          BottomTitleHeight := 0;
          if ShowTitles then begin
            Font.Size  := 12;
            Font.Style := [fsBold];
            TopTitleHeight := TopTitleHeight + TextHeight('A');
            Font.Size  := 6;
            Font.Style := [fsBold];
            BottomTitleHeight := BottomTitleHeight + TextHeight('A');
          end;

          Model.MarginLeft   := PageRect.Left;
          Model.MarginRight  := Printer.PageWidth - PageRect.Right;
          Model.MarginTop    := PageRect.Top + TopTitleHeight;
          Model.MarginBottom := (Printer.PageHeight - PageRect.Bottom) + BottomTitleHeight;

          if ShowTitles then begin

	    Brush.Style := bsClear;
            TextFlags := 0;
            Font.Name  := 'Arial';
            Font.Color := clBlack;

            //FileName
            t := PageRect.Top;
            Font.Size  := 12;
            Font.Style := [fsBold];
            Text := TQUB_QVO_QVM(Model).Filename;
            TextOut(PageRect.Left, t, Text);

            //Date
            t := Printer.PageHeight - Model.Margin.Bottom;
            Font.Size  := 6;
            Font.Style := [fsBold];
            Text := 'QuB (C) 2000-01 University at Buffalo. ' + DateTimeToStr(Now);
            TextOut(PageRect.Left, t, Text);

          end;

          //Draw
          TQUB_QVO_QVM(Model).ResizeModel(Printer.PageWidth, Printer.PageHeight);
          Model.NoDraw := False;
          try
            Model.Draw;
          finally
            Model.NoDraw := True;
          end;

        end;//with Printer.Canvas


      finally

        Printer.EndDoc;

      end;

    except
    end;

  finally

    TQUB_QVO_QVM(Model).StateSize     := OldStatePanelSize;
    TQUB_QVO_QVM(Model).Showk1        := OldShowk1;
    TQUB_QVO_QVM(Model).DiagonalRates := OldDiagonalRates;
    Model.DrawFrame       := OldDrawFrame;
    Model.DrawForPrinting := False;
    Model.BitmapCanvas    := OldCanvas;
    Model.DrawWidth       := OldWidth;
    Model.DrawHeight      := OldHeight;
    Model.MarginLeft      := OldMargin.Left;
    Model.MarginTop       := OldMargin.Top;
    Model.MarginRight     := OldMargin.Right;
    Model.MarginBottom    := OldMargin.Bottom;
    Model.NoDraw          := False;

    Model.NoDraw := True;
    try
      TQUB_QVO_QVM(Model).ResizeModel(Model.Width, Model.Height);
    finally
      Model.NoDraw := False;
    end;
    Model.Draw;

  end;

end;


procedure TQUB_QVO_QVM_Print_PropertiesForm.prCopy;
var
  Metafile                :TMetafile;
  Bitmap                  :TBitmap;
  PageWidth               :Integer;
  PageHeight              :Integer;
  OldCanvas               :TCanvas;
  OldWidth                :Integer;
  OldHeight               :Integer;
  OldMargin               :TRect;
  OldDrawFrame            :Boolean;
  OldStatePanelSize       :Integer;
  OldShowk1               :Boolean;
  OldDiagonalRates        :Boolean;
  Text                    :String;
  PageRect                :TRect;
  TopTitleHeight          :Integer;
  BottomTitleHeight       :Integer;
  t                       :Integer;
  Ff                      :Double;
  iTitleWidth, iTitleSize :Integer;
begin

  Model.NoDraw            := True;

  OldCanvas               := Model.BitmapCanvas;
  OldWidth                := Model.DrawWidth;
  OldHeight               := Model.DrawHeight;
  OldMargin               := Model.Margin;
  OldDrawFrame            := Model.DrawFrame;

  OldStatePanelSize       := TQUB_QVO_QVM(Model).StateSize;
  OldShowk1               := TQUB_QVO_QVM(Model).Showk1;
  OldDiagonalRates        := TQUB_QVO_QVM(Model).DiagonalRates;

  Metafile                := nil;
  Bitmap                  := nil;

  try

    if Landscape then begin
      PageWidth          := DrawWidth;
      PageHeight         := DrawHeight;
    end else begin
      PageWidth          := DrawHeight;
      PageHeight         := DrawWidth;
    end;
    Model.DrawWidth      := PageWidth;
    Model.DrawHeight     := PageHeight;

    Ff := 1;

    if Self.Bitmap then begin
      Bitmap             := TBitmap.Create;
      Bitmap.Width       := Model.DrawWidth;
      Bitmap.Height      := Model.DrawHeight;
      Bitmap.PixelFormat := pf8bit;
      Model.BitmapCanvas := Bitmap.Canvas;
    end else begin
      Metafile           := TMetafile.Create;
      Metafile.Width     := Model.DrawWidth;
      Metafile.Height    := Model.DrawHeight;
      Model.BitmapCanvas := TMetafileCanvas.Create(Metafile, Model.BitmapCanvas.Handle);
    end;

    Model.DrawForPrinting := True;

    Model.DrawFrame       := DrawFrame;

    Model.PrintPenWidth   := 0;
    TQUB_QVO_QVM(Model).PrintNamedStates    := NamedStates;
    TQUB_QVO_QVM(Model).PrintDrawStatePanel := DrawStatePanel;
    TQUB_QVO_QVM(Model).PrintDrawStatePanelColor := DrawStatePanelColor;
    TQUB_QVO_QVM(Model).PrintStateFontSize  := StateFontSize;
    TQUB_QVO_QVM(Model).PrintRateFontSize   := RateFontSize;
    TQUB_QVO_QVM(Model).StateSize           := StatePanelSize;
    TQUB_QVO_QVM(Model).Showk1              := Showk1;
    TQUB_QVO_QVM(Model).DiagonalRates       := DiagonalRates;


    try

      try

        PageRect.Left   := 1;
        PageRect.Top    := 1;
        PageRect.Right  := PageWidth  - 1;
        PageRect.Bottom := PageHeight - 1;

        with Model.BitmapCanvas do begin

          TopTitleHeight    := 0;
          BottomTitleHeight := 0;
          iTitleSize := 12;
          if ShowTitles then begin
            // set font size to largest between 4,12 which does not exceed width
            Font.Name  := 'Arial';
            Font.Size  := Round(Ff * iTitleSize);
            Font.Style := [fsBold];
            iTitleWidth := Model.BitmapCanvas.TextWidth(TQUB_QVO_QVM(Model).Filename);
            while (iTitleWidth>PageWidth) and (iTitleSize>4) do Begin
              iTitleSize := iTitleSize - 1;
              Font.Size := Round( Ff * iTitleSize );
              iTitleWidth := Model.BitmapCanvas.TextWidth(TQUB_QVO_QVM(Model).Filename);
            end;
            TopTitleHeight := TopTitleHeight + TextHeight('A');
            Font.Size  := Round(Ff * 6);
            Font.Style := [fsBold];
            BottomTitleHeight := BottomTitleHeight + TextHeight('A');
          end;

          Model.MarginLeft   := PageRect.Left;
          Model.MarginRight  := PageWidth - PageRect.Right;
          Model.MarginTop    := PageRect.Top + TopTitleHeight - 1;
          Model.MarginBottom := (PageHeight - PageRect.Bottom) + BottomTitleHeight - 1;

          if ShowTitles then begin

            Brush.Style := bsClear;
            TextFlags := 0;
            Font.Name  := 'Arial';
            Font.Color := clBlack;

            //FileName
            t := PageRect.Top;
            Font.Size  := Round(Ff * iTitleSize);
            Font.Style := [fsBold];
            Text := TQUB_QVO_QVM(Model).Filename;
            TextOut(PageRect.Left, t, Text);

            //Date
            t := PageHeight - BottomTitleHeight;
            Font.Size  := Round(Ff * 6);
            Font.Style := [fsBold];
            Text := 'QuB (C) 2000-01 University at Buffalo. ' + DateTimeToStr(Now);
            TextOut(PageRect.Left, t, Text);

          end;

          //Draw
          TQUB_QVO_QVM(Model).ResizeModel(PageWidth, PageHeight);
          Model.NoDraw := False;

          try
            Model.Draw;
          finally
            Model.NoDraw := True;
          end;

        end;//with BitmapCanvas

      finally

        if not Self.Bitmap then begin
          Model.BitmapCanvas.Free;
          Clipboard.Assign(Metafile);
        end else begin
          Clipboard.Assign(Bitmap);
        end;

      end;

    except
    end;


  finally


    Metafile.Free;
    Bitmap.Free;

    TQUB_QVO_QVM(Model).StateSize     := OldStatePanelSize;
    TQUB_QVO_QVM(Model).Showk1        := OldShowk1;
    TQUB_QVO_QVM(Model).DiagonalRates := OldDiagonalRates;
    Model.DrawFrame       := OldDrawFrame;
    Model.DrawForPrinting := False;
    Model.BitmapCanvas    := OldCanvas;
    Model.DrawWidth       := OldWidth;
    Model.DrawHeight      := OldHeight;
    Model.MarginLeft      := OldMargin.Left;
    Model.MarginTop       := OldMargin.Top;
    Model.MarginRight     := OldMargin.Right;
    Model.MarginBottom    := OldMargin.Bottom;
    Model.NoDraw          := False;

    Model.NoDraw := True;
    try
      TQUB_QVO_QVM(Model).ResizeModel(Model.Width, Model.Height);
    finally
      Model.NoDraw := False;
    end;
    Model.Draw;


  end;

end;

//******************************************************************************
procedure TQUB_QVO_QVM_Print_PropertiesForm.prSetCopyImage(c :Boolean);
begin
  prCopyImage            := c;
  chkBitmap.Enabled      := c;
  chkPicture.Enabled     := c;
  lblImageFormat.Enabled := c;
  lblWidth.Enabled       := c;
  lblHeight.Enabled      := c;
  txtWidth.Enabled       := c;
  txtHeight.Enabled      := c;
  if c then btnOK.Caption := '&Copy' else btnOK.Caption := '&Print';
  //if c then chkNamedStates.Visible:=True;
end;

//******************************************************************************
procedure TQUB_QVO_QVM_Print_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Draw Frame"]',
                        'Draw Frame', REGTYPE_BOOL, DEF_DRAWFRAME,
                        chkDrawFrame));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Show Titles"]',
                        'Show Titles', REGTYPE_BOOL, DEF_SHOWTITLES,
                        chkShowTitles));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Named States"]',
                        '', REGTYPE_NONE, DEF_NAMEDSTATES,
                        chkNamedStates));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Draw State Panel"]',
                        '', REGTYPE_BOOL, DEF_DRAWSTATEPANEL,
                        chkDrawStatePanel));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Draw State Panel Color"]',
                        '', REGTYPE_NONE, DEF_DRAWSTATEPANELCOLOR,
                        chkDrawStatePanelColor));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["State Panel Size"]',
                      'State Panel Size', REGTYPE_INT, DEF_STATEPANELSIZE,
                      udStatePanelSize,
                      txtStatePanelSize,
                      AcceptIntBetween(MIN_STATEPANELSIZE, MAX_STATEPANELSIZE)));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["State Font Size"]',
                      'State Font Size', REGTYPE_INT, DEF_STATEFONTSIZE,
                      udStateFontSize,
                      txtStateFontSize,
                      AcceptIntBetween(MIN_FONTSIZE, MAX_FONTSIZE)));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["Rate Font Size"]',
                      'Rate Font Size', REGTYPE_INT, DEF_RATEFONTSIZE,
                      udRateFontSize,
                      txtRateFontSize,
                      AcceptIntBetween(MIN_FONTSIZE, MAX_FONTSIZE)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Show k1"]',
                        '', REGTYPE_BOOL, DEF_SHOWK1,
                        chkShowk1));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Diagonal Rates"]',
                        '', REGTYPE_BOOL, DEF_DIAGONALRATES,
                        chkDiagonalRates));
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Orientation"]',
                          '', REGTYPE_INT, DEF_LANDSCAPE)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkPortrait,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkLandscape,
                                       1,
                                       1))
                       );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Format"]',
                          '', REGTYPE_INT, DEF_BITMAP)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkPicture,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkBitmap,
                                       1,
                                       1))
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["Draw Width"]',
                    'Draw Width', REGTYPE_INT, DEF_WIDTH,
                    txtWidth,
                    AcceptIntBetween(MIN_WIDTH, MAX_WIDTH),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Draw Height"]',
                    'Draw Height', REGTYPE_INT, DEF_HEIGHT,
                    txtHeight,
                    AcceptIntBetween(MIN_HEIGHT, MAX_HEIGHT),
                    '') );

  RegKeyName            := 'Software\QUB\Print Model';
  ProfileCat            := 'CopyModel';
  prBeforeGetProperties := PropertiesToParams;

  WikiTopic             := 'Copy_Model';

end;

procedure TQUB_QVO_QVM_Print_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;

//******************************************************************************
procedure TQUB_QVO_QVM_Print_PropertiesForm.btnOKClick(Sender: TObject);
begin
  inherited;
  if CopyImage then begin
    //----- If 'Named States' is checked then UpdateStates prior to copy, undo subsequently. 2004-02 {JB}
    //if chkNamedStates.Checked then TQUB_QVO_QVM(prModel).ModelObject.States.UpdateStates(TRUE);
    if TQUB_QVO_QVM(prModel).ModelObject.States.CurrState <> nil then TQUB_QVO_QVM(prModel).ModelObject.States.CurrState.DeActivate;
    prCopy;
    if TQUB_QVO_QVM(prModel).ModelObject.States.CurrState <> nil then TQUB_QVO_QVM(prModel).ModelObject.States.CurrState.Activate;
    //if chkNamedStates.Checked then TQUB_QVO_QVM(prModel).ModelObject.States.UpdateStates(FALSE);
  end else begin
    prPrint;
  end;
end;

//******************************************************************************
procedure TQUB_QVO_QVM_Print_PropertiesForm.PropertiesToParams;
begin
  prDrawFrame        := 0<>Properties['Draw Frame'].Data[0];
  prShowTitles       := 0<>Properties['Show Titles'].Data[0];
  prLandscape        := 0<>Properties['Orientation'].Data[0];
  prDrawWidth        := Properties['Draw Width'].Data[0];
  prDrawHeight       := Properties['Draw Height'].Data[0];
  prBitmap           := 0<>Properties['Format'].Data[0];
  prNamedStates      := 0<>Properties['Named States'].Data[0];
  prDrawStatePanel   := 0<>Properties['Draw State Panel'].Data[0];
  prDrawStatePanelColor := 0<>Properties['Draw State Panel Color'].Data[0];
  prStatePanelSize   := Properties['State Panel Size'].Data[0];
  prStateFontSize    := Properties['State Font Size'].Data[0];
  prRateFontSize     := Properties['Rate Font Size'].Data[0];
  prShowk1           := 0<>Properties['Show k1'].Data[0];
  prDiagonalRates    := 0<>Properties['Diagonal Rates'].Data[0];
end;


//******************************************************************************
end.
