{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// what you see when you right-click a state (choose properties) in the model

{-------------------------------------------------------------------------------
  According to Lorin :
  The AR's are not used for the moment.
  They should be used in Sim.
  I left them out because of some unclear theoretical issues.
  They are used in MPL, optionally.
  They're also part of AMP (the big obstacle to AMP of multi-channels).
  It's very restricted, specialized analysis that nobody quite does.
  (see Feng Qin's 2000 paper in Biophys.J)
}

unit QUB_QVO_QVM_State_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ComCtrls, ExtCtrls,
Alloc, AllocTypes, QUB_Global, QUB_Globals,
QUB_QVO_Properties, QUB_QFS_Channels;

{$include 'QUB_QVO_QVM_State_Properties.hct'}


const


PASC_FMT                          :String = '#######0.0#####';

MIN_AMP                           :Double    = -1e9;
DEF_AMP                           :Double    = 0.0;
MAX_AMP                           :Double    = 1e9;

MIN_STD                           :Double    = 0;
DEF_STD                           :Double    = 0.1;
MAX_STD                           :Double    = 1e9;

MIN_COND                          :Double    = 1e-9;
DEF_COND                          :Double    = 10.0;
MAX_COND                          :Double    = 1e9;

MIN_CONDSTD                       :Double    = 0;
DEF_CONDSTD                       :Double    = 1.0;
MAX_CONDSTD                       :Double    = 1e9;

DEF_VREV                          :Double    = 0;
MIN_VREV                          :Double    = -10000;
MAX_VREV                          :Double    = 10000;

MIN_CFF                           :Double    = -1;
MAX_CFF                           :Double    = +1;
DEF_CFF                           :Double    = 0;

MAX_GROUP                         :Integer   = 100;

STATE_PROPERTIES_HEIGHT           :Array[0..1] of Integer = (190, 290);
STATE_PROPERTIES_BTNLBL           :Array[0..1] of String = ('More...', 'Less...');

type


TQUB_QVO_QVM_State_PropertiesForm = class(TQUB_QVO_PropertiesForm)
    lblAmp: TLabel;
    lblStd: TLabel;
  Label3                          :TLabel;
  Label4                          :TLabel;
  Label5                          :TLabel;
  Label6                          :TLabel;
  Label7                          :TLabel;
    lblCond: TLabel;

  lblCoeff                        :TLabel;

  txtPr                           :TEdit;
  txtAmp                          :TEdit;
  txtStd                          :TEdit;
  txtCond                         :TEdit;
  txtGroup                        :TEdit;
  txtNar                          :TEdit;

  udNar                           :TUpDown;

  panCoeff                        :TPanel;
  txtCoeff0                       :TEdit;
  txtCoeff1                       :TEdit;
  txtCoeff2                       :TEdit;
  txtCoeff3                       :TEdit;
  txtCoeff4                       :TEdit;
  txtCoeff5                       :TEdit;
  txtCoeff6                       :TEdit;
  txtCoeff7                       :TEdit;
  txtCoeff8                       :TEdit;
  txtCoeff9                       :TEdit;

  btnGrab                         :TButton;

  panCls                          :TPanel;
  txtCls                          :TEdit;
  udCls                           :TUpDown;
    btnHelp: TButton;
    panIeqFv: TPanel;
    chkpA: TRadioButton;
    chkpS: TRadioButton;
    btnMoreLess: TButton;
    txtVrev: TEdit;
    lblVrev: TLabel;
    txtCondStd: TEdit;
    lblCondStd: TLabel;
    panBaseline: TPanel;
    Label1: TLabel;
    Label2: TLabel;
    txtBaseline: TEdit;
    txtBaselineStd: TEdit;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);

  procedure txtPrKeyPress(Sender: TObject; var Key: Char);
  procedure txtPrExit(Sender: TObject);
  procedure txtAmpKeyPress(Sender: TObject; var Key: Char);
  procedure txtAmpExit(Sender: TObject);
  procedure txtStdExit(Sender: TObject);
  procedure txtStdKeyPress(Sender: TObject; var Key: Char);
  procedure txtCondKeyPress(Sender: TObject; var Key: Char);
  procedure txtCondExit(Sender: TObject);
  procedure txtGroupKeyPress(Sender: TObject; var Key: Char);
  procedure txtGroupExit(Sender: TObject);
  procedure udNarClick(Sender: TObject; Button: TUDBtnType);
  procedure txtCoeffKeyPress(Sender: TObject; var Key: Char);
  procedure txtCoeffExit(Sender: TObject);
  procedure btnGrabClick(Sender: TObject);
  procedure udClsClick(Sender: TObject; Button: TUDBtnType);
  procedure btnHelpClick(Sender: TObject);
  procedure chkpSClick(Sender: TObject);
    procedure btnMoreLessClick(Sender: TObject);
    procedure txtVrevExit(Sender: TObject);
    procedure txtVrevKeyPress(Sender: TObject; var Key: Char);
    procedure txtCondStdExit(Sender: TObject);
    procedure txtCondStdKeyPress(Sender: TObject; var Key: Char);
    procedure txtBaselineKeyPress(Sender: TObject; var Key: Char);
    procedure txtBaselineExit(Sender: TObject);
    procedure txtBaselineStdKeyPress(Sender: TObject; var Key: Char);
    procedure txtBaselineStdExit(Sender: TObject);


protected


  prDemoStatePanel                 :TQUB_QFS_StatePanel;

  prClassCount                     :Integer;

  prCls                            :Integer;
  prPr                             :Double;
  prAmp                            :Array of Double;
  prStd                            :Array of Double;
  prCond                           :Array of Double;
  prCondStd                        :Array of Double;
  prVrev                           :Array of Double;
  prGroup                          :Integer;
  prNar                            :Array of Integer;
  prCff                            :TList; // of TPADouble // Array[0..QUB_QFS_QMF_MAX_CLASS - 1, 0..QUB_QFS_QMF_MAX_COEFF - 1] of Double;

  prIeqFv                          :Boolean;

  prGrabAmp                        :Double;
  prGrabStd                        :Double;

  prError                          :Boolean;

  prCanGrab                        :Boolean;


  procedure prSetClassCount(x :Integer);

  procedure prSetCls(c :Integer);
  procedure prSetPr(p :Double);
  function prGetAmp(Index :Integer) :Double;
  procedure prSetAmp(Index :Integer; a :Double);
  function prGetStd(Index :Integer) :Double;
  procedure prSetStd(Index :Integer; x :Double);
  function prGetCond(Index :Integer) :Double;
  procedure prSetCond(Index :Integer; c :Double);
  function prGetCondStd(Index :Integer) :Double;
  procedure prSetCondStd(Index :Integer; x :Double);
  function prGetVrev(Index :Integer) :Double;
  procedure prSetVrev(Index :Integer; x :Double);
  procedure prSetGroup(g :Integer);
  function prGetNar(Index :Integer) :Integer;
  procedure prSetNar(Index :Integer; n :Integer);
  function prGetCff(Index1, Index2 :Integer) :Double;
  procedure prSetCff(Index1, Index2 :Integer; c :Double);

  procedure prSetIeqFv(x :Boolean);

  procedure prSetCanGrab(g :Boolean);


public

  // classcount: # classes accessed -- set explicitly to be efficient if you expect more
  property ClassCount                    :Integer     read prClassCount     write prSetClassCount;

  property Cls                           :Integer     read prCls            write prSetCls;
  property Pr                            :Double      read prPr             write prSetPr;
  property Amp[Index :Integer]           :Double      read prGetAmp         write prSetAmp;
  property Std[Index :Integer]           :Double      read prGetStd         write prSetStd;
  property Cond[Index :Integer]          :Double      read prGetCond        write prSetCond;
  property CondStd[Index :Integer]       :Double      read prGetCondStd     write prSetCondStd;
  property Vrev[Index :Integer]          :Double      read prGetVrev        write prSetVrev;
  property Group                         :Integer     read prGroup          write prSetGroup;
  property Nar[Index :Integer]           :Integer     read prGetNar         write prSetNar;
  property Cff[Index1, Index2 :Integer]  :Double      read prGetCff         write prSetCff;

  property IeqFv                         :Boolean     read prIeqFv          write prSetIeqFv;

  property GrabAmp                       :Double      read prGrabAmp        write prGrabAmp;
  property GrabStd                       :Double      read prGrabStd        write prGrabStd;
  
  property CanGrab                       :Boolean     read prCanGrab        write prSetCanGrab;


end;


implementation


{$R *.DFM}


//******************************************************************************
procedure TQUB_QVO_QVM_State_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  {$include 'QUB_QVO_QVM_State_Properties.htx'}

  prClassCount                          := 0;
  prCff                                 := TList.Create;

  prDemoStatePanel                      := TQUB_QFS_StatePanel.Create(Self);
  prDemoStatePanel.Parent               := panCls;
  prDemoStatePanel.Align                := alClient;
  prDemoStatePanel.Cls                  := Cls;
  prDemoStatePanel.Visible              := True;

  Self.Height                           := STATE_PROPERTIES_HEIGHT[btnMoreLess.Tag];

end;

procedure TQUB_QVO_QVM_State_PropertiesForm.FormDestroy(Sender: TObject);
begin
  try
    ClassCount                          := 0;
    prCff.Free;
  finally
    inherited;
  end;
end;

//******************************************************************************
procedure TQUB_QVO_QVM_State_PropertiesForm.prSetClassCount(x :Integer);
var
  i, j         :Integer;
  OneCff       :TPADouble;
begin
  if x < 0 then x := 0;

  if x < ClassCount then begin
    for i := ClassCount - 1 downto x do begin
      FreeMem( prCff[i] );
      prCff.Delete(i);
    end;
    // ah let the dyn.arrays stay big
  end else if x > ClassCount then begin
    SetLength( prAmp, x );
    SetLength( prStd, x );
    SetLength( prCond, x );
    SetLength( prCondStd, x );
    SetLength( prVrev, x );
    SetLength( prNar, x );
    for i := ClassCount to x - 1 do begin
      prAmp[i]    := i;
      prStd[i]    := 0.1;
      prCond[i]   := i;
      prCondStd[i] := DEF_CONDSTD;
      prVrev[i]    := DEF_VREV;
      prNar[i]    := 0;

      OneCff   := DAlloc1( QUB_QFS_QMF_MAX_COEFF );
      for j := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
        OneCff[j] := 0.0;
      end;
      prCff.Add( OneCff );
    end;
  end;

  prClassCount := x;

  if (Cls >= prClassCount) and (prClassCount > 0) then Cls := prClassCount - 1;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetCls(c :Integer);
var
  i :Integer;
begin
  if c >= ClassCount then ClassCount := c + 1;
  prCls                              := c;
  // udCls.Position                     := c + 1;
  udCls.Max                          := c + 2;
  udCls.Position                     := c + 1;
  prDemoStatePanel.Cls               := c;

  Amp[Cls]  := Amp[Cls];
  Std[Cls]  := Std[Cls];
  Cond[Cls] := Cond[Cls];
  CondStd[Cls] := CondStd[Cls];
  Vrev[Cls] := Vrev[Cls];
  Nar[Cls]  := Nar[Cls];
  for i := 0 to QUB_QFS_QMF_MAX_COEFF - 1 do begin
    Cff[Cls, i] := Cff[Cls, i];
  end;
  prPropertyChanged := True;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetPr(p :Double);
begin
  txtPr.Text := FloatToStr(p);
  prPr := p;
  prPropertyChanged := True;
end;

function TQUB_QVO_QVM_State_PropertiesForm.prGetAmp(Index :Integer) :Double;
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  Result := prAmp[Index];
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetAmp(Index :Integer; a :Double);
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  txtAmp.Text := FormatFloat(PASC_FMT, a);
  prAmp[Index] := a;
  if Index = 0 then txtBaseline.Text := txtAmp.Text;
  prPropertyChanged := True;
end;

function TQUB_QVO_QVM_State_PropertiesForm.prGetStd(Index :Integer) :Double;
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  Result := prStd[Index];
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetStd(Index :Integer; x :Double);
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  txtStd.Text := FormatFloat(PASC_FMT, x);
  prStd[Index] := x;
  if Index = 0 then txtBaselineStd.Text := txtStd.Text;
  prPropertyChanged := True;
end;

function TQUB_QVO_QVM_State_PropertiesForm.prGetCond(Index :Integer) :Double;
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  Result := prCond[Index];
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetCond(Index :Integer; c :Double);
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  txtCond.Text  := FormatFloat(PASC_FMT, c);
  prCond[Index] := c;
  prPropertyChanged := True;
end;

function TQUB_QVO_QVM_State_PropertiesForm.prGetCondStd(Index :Integer) :Double;
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  Result := prCondStd[Index];
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetCondStd(Index :Integer; x :Double);
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  txtCondStd.Text  := FormatFloat(PASC_FMT, x);
  prCondStd[Index] := x;
  prPropertyChanged := True;
end;

function TQUB_QVO_QVM_State_PropertiesForm.prGetVrev(Index :Integer) :Double;
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  Result := prVrev[Index];
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetVrev(Index :Integer; x :Double);
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  if Abs(x - UNSET_VALUE) < 1e-8 then begin
    txtVrev.Text  := '';
  end else begin
    txtVrev.Text  := FormatFloat(PASC_FMT, x);
  end;
  prVrev[Index] := x;
  prPropertyChanged := True;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetGroup(g :Integer);
begin
  txtGroup.Text := IntToStr(g);
  prGroup := g;
  prPropertyChanged := True;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetNar(Index :Integer; n :Integer);
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  prNar[Index] := n;
  udNar.Position := prNar[Index];
  txtCoeff0.Enabled  := prNar[Index] > 0;
  txtCoeff1.Enabled  := prNar[Index] > 1;
  txtCoeff2.Enabled  := prNar[Index] > 2;
  txtCoeff3.Enabled  := prNar[Index] > 3;
  txtCoeff4.Enabled  := prNar[Index] > 4;
  txtCoeff5.Enabled  := prNar[Index] > 5;
  txtCoeff6.Enabled  := prNar[Index] > 6;
  txtCoeff7.Enabled  := prNar[Index] > 7;
  txtCoeff8.Enabled  := prNar[Index] > 8;
  txtCoeff9.Enabled  := prNar[Index] > 9;
  prPropertyChanged := True;
end;

function TQUB_QVO_QVM_State_PropertiesForm.prGetNar(Index :Integer) :Integer;
begin
  if Index >= ClassCount then ClassCount := Index + 1;

  Result := prNar[Index];
end;

function TQUB_QVO_QVM_State_PropertiesForm.prGetCff(Index1, Index2 :Integer) :Double;
begin
  if Index1 >= ClassCount then ClassCount := Index1 + 1;

  Result := TPADouble(prCff[Index1])[Index2];
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetCff(Index1, Index2 :Integer; c :Double);
begin
  if Index1 >= ClassCount then ClassCount := Index1 + 1;

  case Index2 of
    0:  txtCoeff0.Text  := FormatFloat(PASC_FMT, c);
    1:  txtCoeff1.Text  := FormatFloat(PASC_FMT, c);
    2:  txtCoeff2.Text  := FormatFloat(PASC_FMT, c);
    3:  txtCoeff3.Text  := FormatFloat(PASC_FMT, c);
    4:  txtCoeff4.Text  := FormatFloat(PASC_FMT, c);
    5:  txtCoeff5.Text  := FormatFloat(PASC_FMT, c);
    6:  txtCoeff6.Text  := FormatFloat(PASC_FMT, c);
    7:  txtCoeff7.Text  := FormatFloat(PASC_FMT, c);
    8:  txtCoeff8.Text  := FormatFloat(PASC_FMT, c);
    9:  txtCoeff9.Text  := FormatFloat(PASC_FMT, c);
  end;
  TPADouble(prCff[Index1])[Index2] := c;
  prPropertyChanged := True;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetIeqFv(x :Boolean);
begin
  prIeqFv              := x;

  chkpA.Checked        := not x;
  lblAmp.Visible       := not x;
  txtAmp.Visible       := not x;
  lblStd.Visible       := not x;
  txtStd.Visible       := not x;

  chkpS.Checked        := x;
  lblCond.Visible      := x;
  txtCond.Visible      := x;
  lblCondStd.Visible   := x;
  txtCondStd.Visible   := x;
  panBaseline.Visible  := x;

  // when Mac is ready for Vrev to be per-class, show this field.
  // It can have a number or be blank, in which case its value is QUB_Globals.UNSET_VALUE
  // and Mac should use the model-wide Vrev.

  lblVrev.Visible      := False; // x;
  txtVrev.Visible      := False; // x;

  prPropertyChanged := True;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.prSetCanGrab(g :Boolean);
begin
  prCanGrab := g;
  btnGrab.Enabled := prCanGrab;
end;


//******************************************************************************
procedure TQUB_QVO_QVM_State_PropertiesForm.txtPrKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtPrExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtPrExit(Sender: TObject);
var
  P :Double;
begin
  inherited;
  prError := False;
  try
    P := StrToFloat(TEdit(Sender).Text);
    if (P < 0) or (P > 1) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(0) + ' and ' + FloatToStr(1), mtError, [mbOK], 0);
    Pr := Pr;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if P <> Pr then begin
    Pr := P;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtAmpKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtAmpExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtAmpExit(Sender: TObject);
var
  A :Double;
begin
  inherited;
  prError := False;
  try
    A := StrToFloat(TEdit(Sender).Text);
    if (A < MIN_AMP) or (A > MAX_AMP) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_AMP) + ' and ' + FloatToStr(MAX_AMP), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(Amp[prCls]);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if A <> Amp[prCls] then begin
    Amp[prCls] := A;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtStdKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtStdExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtStdExit(Sender: TObject);
var
  S :Double;
begin
  inherited;
  prError := False;
  try
    S := StrToFloat(TEdit(Sender).Text);
    if (S < MIN_STD) or (S > MAX_STD) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_STD) + ' and ' + FloatToStr(MAX_STD), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(Std[prCls]);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if S <> Std[prCls] then begin
    Std[prCls] := S;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtCondKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtCondExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtCondExit(Sender: TObject);
var
  G :Double;
begin
  inherited;
  prError := False;
  try
    G := StrToFloat(TEdit(Sender).Text);
    if (G < MIN_COND) or (G > MAX_COND) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_COND) + ' and ' + FloatToStr(MAX_COND), mtError, [mbOK], 0);
    Cond[prCls] := Cond[prCls];
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if G <> Cond[prCls] then begin
    Cond[prCls] := G;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtCondStdExit(Sender: TObject);
var
  G :Double;
begin
  inherited;
  prError := False;
  try
    G := StrToFloat(TEdit(Sender).Text);
    if (G < MIN_CONDSTD) or (G > MAX_CONDSTD) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_CONDSTD) + ' and ' + FloatToStr(MAX_CONDSTD), mtError, [mbOK], 0);
    CondStd[prCls] := CondStd[prCls];
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if G <> CondStd[prCls] then begin
    CondStd[prCls] := G;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtCondStdKeyPress(Sender: TObject;
  var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtCondStdExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtVrevExit(Sender: TObject);
var
  G :Double;
begin
  inherited;
  prError := False;
  try
    if TEdit(Sender).Text = '' then begin
      G := UNSET_VALUE;
    end else begin
      G := StrToFloat(TEdit(Sender).Text);
      if (G < MIN_VREV) or (G > MAX_VREV) then raise Exception.Create('');
    end;
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_VREV) + ' and ' + FloatToStr(MAX_VREV), mtError, [mbOK], 0);
    Vrev[prCls] := Vrev[prCls];
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if G <> Vrev[prCls] then begin
    Vrev[prCls] := G;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtVrevKeyPress(
  Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtVrevExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtBaselineKeyPress(
  Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtBaselineExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtBaselineExit(Sender: TObject);
var
  A :Double;
begin
  inherited;
  prError := False;
  try
    A := StrToFloat(TEdit(Sender).Text);
    if (A < MIN_AMP) or (A > MAX_AMP) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_AMP) + ' and ' + FloatToStr(MAX_AMP), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(Amp[0]);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if A <> Amp[0] then begin
    Amp[0] := A;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtBaselineStdKeyPress(
  Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtBaselineStdExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtBaselineStdExit(Sender: TObject);
var
  S :Double;
begin
  inherited;
  prError := False;
  try
    S := StrToFloat(TEdit(Sender).Text);
    if (S < MIN_STD) or (S > MAX_STD) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_STD) + ' and ' + FloatToStr(MAX_STD), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(Std[0]);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if S <> Std[0] then begin
    Std[0] := S;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.udNarClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  Nar[Cls] := TUpDown(Sender).Position;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtCoeffKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtCoeffExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtCoeffExit(Sender: TObject);
var
  C :Double;
begin
  inherited;
  prError := False;
  try
    C := StrToFloat(TEdit(Sender).Text);
    if (C < MIN_CFF) or (C > MAX_CFF) then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_CFF) + ' and ' + FloatToStr(MAX_CFF), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(Cff[Cls, TEdit(Sender).Tag]);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if C <> Cff[Cls, TEdit(Sender).Tag] then begin
    Cff[Cls, TEdit(Sender).Tag] := C;
  end;
end;

//Group
procedure TQUB_QVO_QVM_State_PropertiesForm.txtGroupKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtGroupExit(Sender);
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.txtGroupExit(Sender: TObject);
var
  G :Integer;
begin
  inherited;
  prError := False;
  try
    G := StrToInt(TEdit(Sender).Text);
    if Abs(G) > MAX_GROUP then raise Exception.Create('');
  except
    prError := True;
    MessageDlg('Invalid value; must be between ' + IntToStr(-MAX_GROUP) + ' and ' + IntToStr(MAX_GROUP), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(Group);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if G <> Group then begin
    Group := G;
  end;
end;

//Grab
procedure TQUB_QVO_QVM_State_PropertiesForm.btnGrabClick(Sender: TObject);
begin
  inherited;
  if GrabAmp <> Amp[Cls] then begin
    Amp[Cls] := GrabAmp;
  end;
  if GrabStd <> Std[Cls] then begin
    Std[Cls] := GrabStd;
  end;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.udClsClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  Cls := udCls.Position - 1;
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.btnHelpClick(Sender: TObject);
begin
  //inherited;
  ShowWiki('State_Properties', Handle);
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.chkpSClick(Sender: TObject);
begin
  inherited;
  IeqFv := Boolean(TRadioButton(Sender).Tag);
end;

procedure TQUB_QVO_QVM_State_PropertiesForm.btnMoreLessClick(Sender: TObject);
begin
  inherited;
  btnMoreLess.Tag := Integer(not Boolean(btnMoreLess.Tag));
  btnMoreLess.Caption := STATE_PROPERTIES_BTNLBL[btnMoreLess.Tag];
  Self.Height         := STATE_PROPERTIES_HEIGHT[btnMoreLess.Tag];
end;


//******************************************************************************
end.
