{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// QUB Visual Protocol (displayed in Viewport: Acquisition Protocol)


unit QUB_QVO_QVP;

interface

uses
  Classes, Forms, Menus, Controls, Extctrls, Stdctrls, Dialogs, Registry, Sysutils, Graphics, Windows, Math,
  SelectorListBox,
  QUB_QVO,
  QUB_QFS_QPF,
  QUB_QVO_QVP_Stage_Properties,
  QUB_ZoomBar,
  QUB_Node,
  QUB_Table;

type
  TQUB_QVO_QVP_Channel = class;

  TQUB_QVO_QVP_ShapeEditor = class(TQUB_Table)
    protected
      prShapes      :array of TQUB_QFS_QPF_Shape;

      function prGetShapeCount: Integer;
      function prGetShape(Ix :Integer): TQUB_QFS_QPF_Shape;

      procedure prOnSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
      procedure prOnResize(Sender :TObject);

    public
      constructor Create(AOwner :TComponent); override;

      procedure Add(Shape :TQUB_QFS_QPF_Shape);
      procedure Remove(Shape :TQUB_QFS_QPF_Shape);
      procedure Clear;
      function Contains(Shape :TQUB_QFS_QPF_Shape): Boolean;

      property ShapeCount : Integer read prGetShapeCount;
      property Shapes[Ix :Integer] : TQUB_QFS_QPF_Shape read prGetShape;
    end;

  TQUB_QVO_QVP         = class(TQUB_QVO)
    protected
      prProtocol            :TQUB_QFS_QPF;
      prSaved               :Boolean;

      prShapeEditor         :TQUB_QVO_QVP_ShapeEditor;
      splPars               :TSplitter;
      panStage              :TPanel;
      panStageProperties    :TQUB_QVO_QVP_Stage_PropertiesFrame;
      panChannels           :TPanel;
      prChannels            :TList;
      lstStages             :TSelectorListBox;
      prStagesPanel         :TPanel;
      prStagesPopup         :TPopupMenu;
      itemStageIns          :TMenuItem;
      itemStageAdd          :TMenuItem;
      itemStageCut          :TMenuItem;
      itemStageCopy         :TMenuItem;
      itemStagePaste        :TMenuItem;
      itemStagePasteAfter   :TMenuItem;
      itemStageDel          :TMenuItem;
      prShapePopup          :TPopupMenu;
      itemShapeCut          :TMenuItem;
      itemShapeCopy         :TMenuItem;
      itemShapePaste        :TMenuItem;
      itemShapePasteAfter   :TMenuItem;
      itemShapeDel          :TMenuItem;
      panZoom               :TQUB_ZoomBar;

      prCurrStage           :TQUB_QFS_QPF_Stage;

      prCurrChannel         :TQUB_QVO_QVP_Channel;
      prChangingCurrShape   :Boolean;

      m_nProtocolStartTime : Double;

      procedure prSetCurrStage(x :TQUB_QFS_QPF_Stage);
      function prGetChannel(Ix :Integer): TQUB_QVO_QVP_Channel;

      procedure prOnChannelCountChanged(Sender :TObject);
      procedure prOnStageAdded(Sender :TObject; Index :Integer);
      procedure prOnStageRemoved(Sender :TObject; OldIndex :Integer; Stage :TQUB_QFS_QPF_Stage);
      procedure prOnRepeatChanged(Sender :TObject);
      procedure prOnPauseChanged(Sender :TObject);
      procedure prOnShapesChanged(Sender :TObject);
      procedure prOnCurrShapeChanged(Sender :TQUB_QVO_QVP_Channel; Shape :TQUB_QFS_QPF_Shape);
      procedure prOnChannelDeleting(Sender :TObject);

      procedure panChannelsResize(Sender :TObject);
      procedure lstStagesItemChanged(Sender :TObject);

      procedure prOnZoomChange(Sender :TObject);
      procedure prOnStartChange(Sender :TObject);

      procedure prOnStagesPopup(Sender :TObject);
      procedure itemStageInsClick(Sender :TObject);
      procedure itemStageAddClick(Sender :TObject);
      procedure itemStageCutClick(Sender :TObject);
      procedure itemStageCopyClick(Sender :TObject);
      procedure itemStagePasteClick(Sender :TObject);
      procedure itemStagePasteAfterClick(Sender :TObject);
      procedure itemStageDelClick(Sender :TObject);
      procedure prOnShapePopup(Sender :TObject);
      procedure itemShapeCutClick(Sender :TObject);
      procedure itemShapeCopyClick(Sender :TObject);
      procedure itemShapePasteClick(Sender :TObject);
      procedure itemShapePasteAfterClick(Sender :TObject);
      procedure itemShapeDelClick(Sender :TObject);

      //Events
      procedure DoOnFileOpen(Sender :TObject); override;
      procedure DoOnFileClose(Sender :TObject); override;
      // procedure DoOnFileEdit(Sender :TObject; EditOper :TQUB_EditOperation; var Param, var Param2); override;

      procedure prSetTipsOn(t :Boolean); override;

      //Print
      function prPrint(Dlg :TPrintDialog) :Boolean; override;

      procedure prSetInitialProperties; override;

      function prGetFileName :String; override;
      function prGetModified :Boolean; override;
      function prGetSaved :Boolean; override;

      //Popup Menu
      procedure prCreatePopUpMenu; override;

    public
      m_nSecondsPerPixel : Double;
      OnChannelCountChanged       :TNotifyEvent;

      constructor Create(AOwner :TComponent; ShapeLib :TQUB_QFS_QPF_ShapeLib); reintroduce; virtual;
      destructor Destroy; override;

      procedure Initialize; // call this after contstructing then parenting

      procedure Paint; override;

      //File .......................................................................
      // these four should never be used, but they're abstract in QVO
      function New(const FileName :String) :Boolean; override;
      function Open(const FileName :String) :Boolean; override;
      function Open(const FileName :String; OpenStream :TStream) :Boolean; override;
      function Close :Boolean; override;
      function OpenFromString(const MemFile :String) :Boolean; override;

      function Save(const FileName :String) :Boolean; override;
      function CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;
      function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

      function Properties :Boolean; override;

      procedure Activate; override;
      // procedure DeActivate; override;

      //Draw .......................................................................
      function Draw :Boolean; override;
      procedure ForceRedraw;

      procedure AddShape(ShapeFactory :TQUB_QFS_QPF_ShapeFactory);
      procedure DoneDragging;

      procedure ReMeasure;

      //Lock
      procedure LockWrite; override;
      procedure UnLockWrite; override;
      procedure LockReadWrite; override;
      procedure UnLockReadWrite; override;


      property Protocol           :TQUB_QFS_QPF       read prProtocol;

      property StagesPanel        :TPanel                                    write prStagesPanel;

      property CurrStage          :TQUB_QFS_QPF_Stage read prCurrStage       write prSetCurrStage;

      property Channels[Ix :Integer] :TQUB_QVO_QVP_Channel read prGetChannel;
      property CurrChannel           :TQUB_QVO_QVP_Channel read prCurrChannel;
    end;


  TQUB_QVO_QVP_Channel = class(TCustomControl)
    protected
      prProtocol            :TQUB_QVO_QVP;
      prChannel             :TQUB_QFS_QPF_Channel;
      prShapeEditor         :TQUB_QVO_QVP_ShapeEditor;
      prBitmap              :Graphics.TBitmap;
      prBitmapCanvas        :TCanvas;
      prRightOfShape        :Array of Integer;
      prDragIx              :Integer;
      prDragSourceIx        :Integer;
      prCurrShape           :TQUB_QFS_QPF_Shape;
      prNeedsDraw           :Boolean;

      procedure prSetChannel(x :TQUB_QFS_QPF_Channel);
      procedure prSetCurrShape(x :TQUB_QFS_QPF_Shape);
      function prGetCurrShapeIndex: Integer;
      function prGetSelected(i :Integer): Boolean;

      procedure prOnDragOver(Sender, Source :TObject; X, Y :Integer;
                             State :TDragState; var Accept :Boolean);
      procedure prOnDragDrop(Sender, Source :TObject; X, Y :Integer);

      procedure prOnEndDrag(Sender, Target: TObject; X, Y: Integer);

      procedure prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      procedure prOnKeyPress(Sender: TObject; var Key: Char);

      procedure prOnShapeAdded(Sender :TObject; Index :Integer);
      procedure prOnShapeRemoved(Sender :TObject; OldIndex :Integer; Shape :TQUB_QFS_QPF_Shape);
      procedure prOnShapeChanged(Sender :TObject);

    public
      OnCurrShapeChanged     :procedure (Sender :TQUB_QVO_QVP_Channel; Shape :TQUB_QFS_QPF_Shape) of object;
      OnShapesChanged        :TNotifyEvent; // e.g. added, removed, changed -> all channels forceRedraw

      constructor Create(AOwner :TComponent); override;
      destructor Destroy; override;

      procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
      procedure Paint; override;

      procedure ForceRedraw;

      procedure AddShape(ShapeFactory :TQUB_QFS_QPF_ShapeFactory);
      procedure DeleteCurrShape;

      procedure DoneDragging;

      property Protocol             :TQUB_QVO_QVP          read prProtocol         write prProtocol;
      property Channel              :TQUB_QFS_QPF_Channel  read prChannel          write prSetChannel;
      property ShapeEditor          :TQUB_QVO_QVP_ShapeEditor read prShapeEditor   write prShapeEditor;

      property CurrShape            :TQUB_QFS_QPF_Shape    read prCurrShape        write prSetCurrShape;
      property CurrShapeIndex       :Integer               read prGetCurrShapeIndex;
      property Selected[i :Integer] :Boolean               read prGetSelected;
  end;

//==============================================================================
implementation

uses
  QUB_Console, QUB_Rich_Console, QUB_Viewport_Protocol;

//==============================================================================
constructor TQUB_QVO_QVP_ShapeEditor.Create(AOwner :TComponent);
begin
  inherited;
  Clear;

  OnSetEditText := prOnSetEditText;
  OnResize      := prOnResize;
end;

procedure TQUB_QVO_QVP_ShapeEditor.Add(Shape :TQUB_QFS_QPF_Shape);
var
  r            :Integer;
  Param        :INode;
begin
  if ShapeCount = 0 then begin
    RowCount      := 1;
    ColCount      := 2;
    ColWidths[0]  := 100;
    if (Width - ColWidths[0] - ColWidths[1]) > 20 then
      ColWidths[1]  := Width - ColWidths[0] - 10;
    Cells[0, 0]   := 'Shape:';
    Cells[1, 0]   := Shape.Tree.DataAsString;

    Param         := Shape.Tree.Child;
    r             := 1;
    while not Param.IsNull do begin
      if (not Param.IsNull) and (Param.Name <> 'rep_index') then begin
        RowCount  := r+1;
        Cells[0, r] := Param.Name;
        Cells[1, r] := Param.DataAsString;

        Inc(r);
      end;

      Param       := Param.Sibling;
    end;

    FixedRows := 1;
    FixedCols := 1;
  end else begin
    if Shape.Tree.DataAsString <> Cells[1, 0] then Cells[1, 0] := '';
    for r := RowCount - 1 downto 1 do begin
      Param       := Shape.Tree.Find(Cells[0, r]);
      if Param.IsNull then begin
        DeleteRow(r);
      end else if Cells[1, r] <> '' then begin
        if Param.DataAsString <> Cells[1, r] then begin
          Cells[1, r] := '';
        end;
      end;
    end;
  end;

  SetLength(prShapes, ShapeCount+1);
  prShapes[ShapeCount-1] := Shape;
end;

procedure TQUB_QVO_QVP_ShapeEditor.Remove(Shape :TQUB_QFS_QPF_Shape);
var
  Kept             :array of TQUB_QFS_QPF_Shape;
  i                :Integer;
  RemIx            :Integer;
begin
  if ShapeCount > 1 then begin
    SetLength(Kept, ShapeCount);
    RemIx          := -1;
    for i := 0 to ShapeCount - 1 do begin
      if Shapes[i] = Shape then RemIx := i;
      Kept[i] := Shapes[i];
    end;

    if RemIx >= 0 then begin
      Clear;
      for i := 0 to Length(Kept) - 1 do begin
        if i <> RemIx then Add(Kept[i]);
      end;
    end;
  end else if (ShapeCount = 1) and (Shapes[0] = Shape) then begin
    Clear;
  end;
end;

procedure TQUB_QVO_QVP_ShapeEditor.Clear;
begin
  SetLength(prShapes, 0);
  RowCount := 0;
  ColCount := 0;
end;

function TQUB_QVO_QVP_ShapeEditor.Contains(Shape :TQUB_QFS_QPF_Shape): Boolean;
var
  i            :Integer;
begin
  for i := 0 to ShapeCount - 1 do begin
    if Shapes[i] = Shape then begin
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

function TQUB_QVO_QVP_ShapeEditor.prGetShapeCount: Integer;
begin
  Result := Length(prShapes);
end;

function TQUB_QVO_QVP_ShapeEditor.prGetShape(Ix :Integer): TQUB_QFS_QPF_Shape;
begin
  Result := prShapes[Ix];
end;

procedure TQUB_QVO_QVP_ShapeEditor.prOnSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
var
  Param         :String;
  i             :Integer;
begin

  // if (not EventsEveryKey) and grdData.EditorMode then Exit;

  if (ARow = 0) or (ACol = 0) then Exit;
  Param := Cells[0, ARow];
  if Param = '' then Exit;

  // if (Value = '') or (Value = '.') or (Value = '-') or (Value = '-.') then Exit;

  try
    for i := 0 to ShapeCount - 1 do begin
      Shapes[i].Properties[Param] := Value;
    end;
  except
  end;
end;

procedure TQUB_QVO_QVP_ShapeEditor.prOnResize(Sender :TObject);
begin
  if ColCount = 2 then ColWidths[1] := Width - ColWidths[0];
end;

//==============================================================================
constructor TQUB_QVO_QVP.Create(AOwner :TComponent; ShapeLib :TQUB_QFS_QPF_ShapeLib);
begin
  inherited Create(AOwner);
  prProtocol            := TQUB_QFS_QPF.Create(ShapeLib);
end;

procedure TQUB_QVO_QVP.Initialize;
begin

  try


    //PopUp Menu
    prCreatePopUpMenu;

    prDrawFocusRect                 := False;

    //To give some initial values, before any calculations are made:
    prSetInitialProperties;

    //Flags
    prMustRedraw                    := True;

    m_nSecondsPerPixel := 1.0;
    m_nProtocolStartTime := 0.0;

    prShapeEditor         := TQUB_QVO_QVP_ShapeEditor.Create(Self);
    prShapeEditor.Align   := alLeft;
    prShapeEditor.Width   := 220;
    prShapeEditor.Parent  := Self;

    splPars               := TSplitter.Create(Self);
    splPars.Align         := alLeft;
    splPars.Left          := prShapeEditor.Width + 1;
    splPars.Parent        := Self;

    panStage              := TPanel.Create(Self);
    panStage.Left         := splPars.Left + 10;
    panStage.Align        := alClient;
    panStage.Parent       := Self;

    panStageProperties        := TQUB_QVO_QVP_Stage_PropertiesFrame.Create(Self);
    panStageProperties.Align  := alBottom;
    panStageProperties.Parent := panStage;

    panZoom               := TQUB_ZoomBar.Create(Self);
    panZoom.Align         := alBottom;
    panZoom.Parent        := panStage;
    panZoom.OnZoomChange  := prOnZoomChange;
    panZoom.OnStartChange := prOnStartChange;

    panChannels           := TPanel.Create(Self);
    panChannels.Align     := alClient;
    panChannels.Parent    := panStage;
    panChannels.OnResize  := panChannelsResize;

    prChannels            := TList.Create;
    prOnChannelCountChanged(Protocol);
    panChannelsResize(Self);

    lstStages             := TSelectorListBox.Create(Self);
    lstStages.Align       := alClient;
    lstStages.Parent      := prStagesPanel;
    lstStages.OnItemChange := lstStagesItemChanged;
    lstStages.PopupMenu   := prStagesPopup;
    lstStages.Items.Add('Stage 1');
    lstStages.ItemIndex   := 0;

    Protocol.OnChannelCountChanged := prOnChannelCountChanged;
    Protocol.OnStageAdded          := prOnStageAdded;
    Protocol.OnStageRemoved        := prOnStageRemoved;

    CurrStage             := Protocol.Stages[0];

  except
  end;
end;

destructor TQUB_QVO_QVP.Destroy;
begin
  try
    CurrStage := nil;
    prProtocol.Free;
    prChannels.Free;
  finally
    inherited;
  end;
end;


procedure TQUB_QVO_QVP.prCreatePopUpMenu;
begin

  inherited;
  PopupMenu                       := nil;

  prStagesPopup                   := TPopupMenu.Create(Self);
  prStagesPopup.OnPopup           := prOnStagesPopup;

  itemStageIns                    := TMenuItem.Create(Self);
  itemStageIns.Caption            := 'Insert stage';
  itemStageIns.OnClick            := itemStageInsClick;
  prStagesPopup.Items.Add( itemStageIns );

  itemStageAdd                    := TMenuItem.Create(Self);
  itemStageAdd.Caption            := 'Add stage';
  itemStageAdd.OnClick            := itemStageAddClick;
  prStagesPopup.Items.Add( itemStageAdd );

  itemStageCut                    := TMenuItem.Create(Self);
  itemStageCut.Caption            := 'Cut stage';
  itemStageCut.OnClick            := itemStageCutClick;
  prStagesPopup.Items.Add( itemStageCut );

  itemStageCopy                   := TMenuItem.Create(Self);
  itemStageCopy.Caption           := 'Copy stage';
  itemStageCopy.OnClick           := itemStageCopyClick;
  prStagesPopup.Items.Add( itemStageCopy );

  itemStagePaste                  := TMenuItem.Create(Self);
  itemStagePaste.Caption          := 'Paste stage before';
  itemStagePaste.OnClick          := itemStagePasteClick;
  prStagesPopup.Items.Add( itemStagePaste );

  itemStagePasteAfter             := TMenuItem.Create(Self);
  itemStagePasteAfter.Caption     := 'Paste stage after';
  itemStagePasteAfter.OnClick     := itemStagePasteAfterClick;
  prStagesPopup.Items.Add( itemStagePasteAfter );

  itemStageDel                    := TMenuItem.Create(Self);
  itemStageDel.Caption            := 'Delete stage';
  itemStageDel.OnClick            := itemStageDelClick;
  prStagesPopup.Items.Add( itemStageDel );
  prStagesPopup.Items.InsertNewlineAfter( itemStageIns );

  prShapePopup                    := TPopupMenu.Create(Self);
  prShapePopup.OnPopup            := prOnShapePopup;

  itemShapeCut                    := TMenuItem.Create(Self);
  itemShapeCut.Caption            := 'Cut';
  itemShapeCut.OnClick            := itemShapeCutClick;
  prShapePopup.Items.Add( itemShapeCut );

  itemShapeCopy                   := TMenuItem.Create(Self);
  itemShapeCopy.Caption           := 'Copy';
  itemShapeCopy.OnClick           := itemShapeCopyClick;
  prShapePopup.Items.Add( itemShapeCopy );

  itemShapePaste                  := TMenuItem.Create(Self);
  itemShapePaste.Caption          := 'Paste';
  itemShapePaste.OnClick          := itemShapePasteClick;
  prShapePopup.Items.Add( itemShapePaste );

  itemShapePasteAfter             := TMenuItem.Create(Self);
  itemShapePasteAfter.Caption     := 'Paste after';
  itemShapePasteAfter.OnClick     := itemShapePasteAfterClick;
  prShapePopup.Items.Add( itemShapePasteAfter );

  itemShapeDel                    := TMenuItem.Create(Self);
  itemShapeDel.Caption            := 'Delete';
  itemShapeDel.OnClick            := itemShapeDelClick;
  prShapePopup.Items.Add( itemShapeDel );

end;

procedure TQUB_QVO_QVP.prOnStagesPopup(Sender :TObject);
var
  Name                            :String;
  CanPaste                        :Boolean;
begin
  CanPaste     := CanPasteNode(Name) and (Name = 'Stage');
  itemStageCut.Visible            := (lstStages.ItemIndex > -1) and (lstStages.Items.Count > 1);
  itemStageCopy.Visible           := lstStages.ItemIndex > -1;
  itemStagePaste.Visible          := CanPaste;
  itemStagePasteAfter.Visible     := CanPaste and (lstStages.ItemIndex > -1);
  itemStageDel.Visible            := (lstStages.ItemIndex > -1) and (lstStages.Items.Count > 1);;
end;

procedure TQUB_QVO_QVP.itemStageInsClick(Sender :TObject);
begin
  Protocol.InsertStage( QUB_Node_Null, lstStages.ItemIndex );
end;

procedure TQUB_QVO_QVP.itemStageAddClick(Sender :TObject);
begin
  Protocol.InsertStage( QUB_Node_Null, lstStages.Items.Count );
end;

procedure TQUB_QVO_QVP.itemStageCutClick(Sender :TObject);
begin
  if Protocol.Stages[ lstStages.ItemIndex ].Tree.Copy
  then Protocol.RemoveStage( lstStages.ItemIndex );
end;

procedure TQUB_QVO_QVP.itemStageCopyClick(Sender :TObject);
begin
  Protocol.Stages[ lstStages.ItemIndex ].Tree.Copy;
end;

procedure TQUB_QVO_QVP.itemStagePasteClick(Sender :TObject);
begin
  Protocol.InsertStage( TNode.Paste, lstStages.ItemIndex );
end;

procedure TQUB_QVO_QVP.itemStagePasteAfterClick(Sender :TObject);
begin
  Protocol.InsertStage( TNode.Paste, lstStages.ItemIndex + 1 );
end;

procedure TQUB_QVO_QVP.itemStageDelClick(Sender :TObject);
begin
  Protocol.RemoveStage( lstStages.ItemIndex );
end;

procedure TQUB_QVO_QVP.prOnShapePopup(Sender :TObject);
var
  HaveShape                       :Boolean;
  Name                            :String;
  CanPaste                        :Boolean;
begin
  HaveShape                       := (CurrChannel <> nil) and (CurrChannel.CurrShape <> nil);
  CanPaste                        := CanPasteNode(Name) and (Name = 'Shape');
  itemShapeCut.Visible            := HaveShape;
  itemShapeCopy.Visible           := HaveShape;
  itemShapePaste.Visible          := CanPaste;
  itemShapePasteAfter.Visible     := HaveShape and CanPaste;
  itemShapeDel.Visible            := HaveShape;
end;

procedure TQUB_QVO_QVP.itemShapeCutClick(Sender :TObject);
begin
  if CurrChannel.CurrShape.Tree.Copy
  then CurrChannel.DeleteCurrShape;
end;

procedure TQUB_QVO_QVP.itemShapeCopyClick(Sender :TObject);
begin
  CurrChannel.CurrShape.Tree.Copy;
end;

procedure TQUB_QVO_QVP.itemShapePasteClick(Sender :TObject);
var
  Ix    :Integer;
begin
  Ix                := CurrChannel.CurrShapeIndex;
  if Ix < 0 then Ix := CurrChannel.Channel.ShapeCount - 1;
  CurrChannel.Channel.InsertShape( TNode.Paste, Ix );
end;

procedure TQUB_QVO_QVP.itemShapePasteAfterClick(Sender :TObject);
var
  Ix    :Integer;
begin
  Ix                := CurrChannel.CurrShapeIndex;
  if Ix < 0 then Ix := CurrChannel.Channel.ShapeCount - 1;
  CurrChannel.Channel.InsertShape( TNode.Paste, Ix + 1 );
end;

procedure TQUB_QVO_QVP.itemShapeDelClick(Sender :TObject);
begin
  CurrChannel.DeleteCurrShape;
end;

procedure TQUB_QVO_QVP.prSetCurrStage(x :TQUB_QFS_QPF_Stage);
var
  i         :Integer;
begin
  if prCurrStage = x then Exit;

  if prCurrStage <> nil then begin
    prCurrStage.OnRepeatChanged := nil;
    prCurrStage.OnPauseChanged  := nil;
    prCurrStage.OnChannelDeleting := nil;
    prShapeEditor.Clear;
  end;

  prCurrStage := x;
  panStageProperties.Stage := x;

  if prCurrStage <> nil then begin
    prCurrStage.OnRepeatChanged := prOnRepeatChanged;
    prCurrStage.OnPauseChanged  := prOnPauseChanged;
    prCurrStage.OnChannelDeleting := prOnChannelDeleting;

    for i := 0 to Protocol.StageCount - 1 do begin
      if Protocol.Stages[i] = x then begin
        lstStages.ItemIndex := i;
        Break;
      end;
    end;

    for i := 0 to Protocol.ChannelCount - 1 do begin
      Channels[i].Channel := prCurrStage.Channels[i];
    end;
  end else begin
    for i := 0 to Protocol.ChannelCount - 1 do begin
      Channels[i].Channel := nil;
    end;
  end;

  ReMeasure;
end;

function TQUB_QVO_QVP.prGetChannel(Ix :Integer): TQUB_QVO_QVP_Channel;
begin
  if (Ix >= 0) or (Ix < Protocol.ChannelCount) then begin
    Result := TQUB_QVO_QVP_Channel(prChannels[Ix]);
  end else begin
    Result := nil;
  end;
end;

procedure TQUB_QVO_QVP.prOnChannelCountChanged(Sender :TObject);
var
  i                     :Integer;
  Chan                  :TQUB_QVO_QVP_Channel;
begin
  for i := prChannels.Count - 1 downto Protocol.ChannelCount do begin
    Channels[i].Free;
    prChannels.Delete(i);
  end;

  for i := prChannels.Count to Protocol.ChannelCount - 1 do begin
    Chan                    := TQUB_QVO_QVP_Channel.Create(Self);
    Chan.Parent             := panChannels;
    Chan.BevelOuter         := bvRaised;
    Chan.ShapeEditor        := prShapeEditor;
    Chan.Protocol           := Self;
    Chan.OnCurrShapeChanged := prOnCurrShapeChanged;
    Chan.OnShapesChanged    := prOnShapesChanged;
    Chan.PopupMenu          := prShapePopup;
    prChannels.Add( Chan );
    if prCurrStage <> nil then Chan.Channel := prCurrStage.Channels[i];
  end;

  ReMeasure;

  if Assigned(OnChannelCountChanged) then OnChannelCountChanged(Self);
  panChannelsResize(panChannels);
end;

procedure TQUB_QVO_QVP.prOnStageAdded(Sender :TObject; Index :Integer);
begin
  lstStages.Items.Add( Format('Stage %d', [lstStages.Items.Count+1]) );
  CurrStage := Protocol.Stages[Index];
end;

procedure TQUB_QVO_QVP.prOnStageRemoved(Sender :TObject; OldIndex :Integer; Stage :TQUB_QFS_QPF_Stage);
begin
  lstStages.Items.Delete( lstStages.Items.Count - 1 );
  if Stage = prCurrStage then begin
    if OldIndex = 0 then begin
      CurrStage := Protocol.Stages[0];
    end else begin
      CurrStage := Protocol.Stages[OldIndex - 1];
    end;
  end;
end;

procedure TQUB_QVO_QVP.prOnRepeatChanged(Sender :TObject);
begin
  panStageProperties.DoOnRepeatChanged(Sender);
  ReMeasure;
end;

procedure TQUB_QVO_QVP.prOnPauseChanged(Sender :TObject);
begin
  panStageProperties.DoOnPauseChanged(Sender);
end;

procedure TQUB_QVO_QVP.prOnShapesChanged(Sender :TObject);
begin
  ReMeasure;
end;

procedure TQUB_QVO_QVP.prOnChannelDeleting(Sender :TObject);
var
  i            :Integer;
begin
  for i := 0 to Protocol.ChannelCount - 1 do begin
    if Channels[i].Channel = TQUB_QFS_QPF_Channel(Sender) then begin
      Channels[i].Channel := nil;
      Break;
    end;
  end;
end;

procedure TQUB_QVO_QVP.prOnZoomChange(Sender :TObject);
begin
  ReMeasure;
end;

procedure TQUB_QVO_QVP.prOnStartChange(Sender :TObject);
begin
  ReMeasure;
end;

procedure TQUB_QVO_QVP.ReMeasure;
var
  nDuration : Double;
  iChannel : Integer;
begin
  if prCurrStage = nil then Exit;

  nDuration := 0.0;
  for iChannel := 0 to prChannels.Count - 1 do begin
    if Channels[iChannel].Channel <> nil then begin
      nDuration := Max(nDuration, Channels[iChannel].Channel.Duration(CurrStage.RepeatTimes-1));
    end;
  end;

  if nDuration = 0.0 then begin
    m_nSecondsPerPixel := 1.0;
    m_nProtocolStartTime := 0.0;
  end else begin
    m_nSecondsPerPixel := (nDuration / panChannels.Width) / (panZoom.Zoom / 100.0);
    m_nProtocolStartTime := nDuration * (panZoom.Start / 100.0);
  end;

  ForceRedraw;
end;

procedure TQUB_QVO_QVP.prOnCurrShapeChanged(Sender :TQUB_QVO_QVP_Channel; Shape :TQUB_QFS_QPF_Shape);
var
  iChannel : Integer;
begin
  if prChangingCurrShape then Exit;

  prChangingCurrShape := True;
  prCurrChannel := Sender;
  for iChannel := 0 to Protocol.ChannelCount - 1 do begin
    if Channels[iChannel] <> Sender then Channels[iChannel].CurrShape := nil;
  end;
  prChangingCurrShape := False;
end;

procedure TQUB_QVO_QVP.panChannelsResize(Sender :TObject);
var
  iChannel : Integer;
  iHeight : Integer;
begin
  ReMeasure;

  iHeight := Round(panChannels.ClientHeight * 1.0 / Protocol.ChannelCount);
  for iChannel := 0 to prChannels.Count - 1 do begin
    Channels[iChannel].SetBounds(0, iChannel*iHeight, ClientWidth, iHeight);
  end;
end;

procedure TQUB_QVO_QVP.AddShape(ShapeFactory :TQUB_QFS_QPF_ShapeFactory);
var
  i                      :Integer;
  c                      :Integer;
  s                      :Double;
  MinSeconds             :Double;
  MaxSeconds             :Double;
begin
  // find channel with min length
  c                      := 0;
  MinSeconds             := Channels[0].Channel.Duration(CurrStage.RepeatTimes - 1);
  MaxSeconds             := MinSeconds;
  for i := 1 to prChannels.Count - 1 do begin
    s                    := Channels[i].Channel.Duration(CurrStage.RepeatTimes - 1);
    if s < MinSeconds then begin
      c                  := i;
      MinSeconds         := s;
    end else if s > MaxSeconds then begin
      MaxSeconds         := s;
    end;
  end;

  // if all the same, use CurrChannel or 0
  if MinSeconds = MaxSeconds then begin
    if CurrChannel <> nil then begin
      CurrChannel.AddShape(ShapeFactory);
    end else begin
      Channels[0].AddShape(ShapeFactory);
    end;
  end else begin
    Channels[c].AddShape(ShapeFactory);
  end;
end;

procedure TQUB_QVO_QVP.DoneDragging;
var
  i : Integer;
begin
  for i := 0 to prChannels.Count - 1 do begin
    Channels[i].DoneDragging;
  end;
end;

procedure TQUB_QVO_QVP.ForceRedraw;
var
  i                      :Integer;
begin
  for i := 0 to prChannels.Count - 1 do begin
    Channels[i].ForceRedraw;
  end;
end;

procedure TQUB_QVO_QVP.lstStagesItemChanged(Sender :TObject);
begin
   CurrStage := Protocol.Stages[ lstStages.ItemIndex ];
end;

//-------------------------------- FILE
function TQUB_QVO_QVP.New(const FileName :String) :Boolean;
begin
  //MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVP.Open(const FileName :String) :Boolean;
begin
  Result := Protocol.Open(FileName);
  if not Result then prSaved := True;
end;

function TQUB_QVO_QVP.Open(const FileName :String; OpenStream :TStream) :Boolean;
begin
  //MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVP.Close :Boolean;
begin
  //MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVP.OpenFromString(const MemFile :String) :Boolean;
var
  NewTree : INode;
begin
  Result := True;
  try
    NewTree := TNode.CreateFromString(MemFile);
    if not NewTree.IsNull then begin
      Protocol.Tree := NewTree;
      Result := False;
    end;
  except
  end;
end;

function TQUB_QVO_QVP.Save(const FileName :String) :Boolean;
begin
  try
    Result := Protocol.SaveAs(FileName);
    if not Result then prSaved := True;
  except
    Result := False;
  end;
end;

function TQUB_QVO_QVP.CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;
begin
  Result := True;
end;

function TQUB_QVO_QVP.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  MemFile := Protocol.Tree.AsString;
  Result := False;
end;

procedure TQUB_QVO_QVP.prSetTipsOn(t :Boolean);
begin
  inherited;
end;

function TQUB_QVO_QVP.prPrint(Dlg :TPrintDialog) :Boolean;
begin
  Result := False; // prPrintPropertiesForm.Print(Self, Dlg, False{CopyImage}, True);
end;

procedure TQUB_QVO_QVP.Activate;
begin
  inherited;
  lstStages.BringToFront;
end;


//------------------------------------- DRAW
procedure TQUB_QVO_QVP.Paint;
begin
end;

//------------------------------------- LOCK
procedure TQUB_QVO_QVP.LockWrite;
begin
  inherited;
  Enabled := False;
end;

procedure TQUB_QVO_QVP.UnLockWrite;
begin
  inherited;
  Enabled := not LockedAny;
end;

procedure TQUB_QVO_QVP.LockReadWrite;
begin
  inherited;
  Enabled := False;
end;

procedure TQUB_QVO_QVP.UnLockReadWrite;
begin
  inherited;
  Enabled := not LockedAny;
end;


//-------------------------------------
function TQUB_QVO_QVP.prGetFileName :String;
begin
  Result := Protocol.FileName;
end;

function TQUB_QVO_QVP.prGetModified :Boolean;
begin
  Result := Protocol.Modified;
end;

function TQUB_QVO_QVP.prGetSaved :Boolean;
begin
  Result := prSaved;
end;

function TQUB_QVO_QVP.Draw :Boolean;
begin
  Result := inherited Draw;
  Update;
end;

//------------------------------------- MODEL  EVENTS
procedure TQUB_QVO_QVP.DoOnFileOpen(Sender :TObject);
begin
  Draw;
end;

procedure TQUB_QVO_QVP.DoOnFileClose(Sender :TObject);
begin
  Draw;
end;

//-------------------------------------Properties
function TQUB_QVO_QVP.Properties :Boolean;
begin
  Result := False;
  Console.NewEntry(Protocol.Tree.AsString, COLOR_TEXT);
end;

//---------Popup Menu Events
procedure TQUB_QVO_QVP.prSetInitialProperties;
const
  QUB_QVO_QVP_DEF_MARGIN : Integer = 5;
begin
  //Margins
  prMargin.Left       := QUB_QVO_QVP_DEF_MARGIN;//LEFT;
  prMargin.Top        := QUB_QVO_QVP_DEF_MARGIN;//TOP;
  prMargin.Right      := QUB_QVO_QVP_DEF_MARGIN;//RIGHT;
  prMargin.Bottom     := QUB_QVO_QVP_DEF_MARGIN;//BOTTOM;

  //Colors
  prColor             := 0;
  prColorBack         := 0;
  prColorFrame        := 0;
end;

//==============================================================================
constructor TQUB_QVO_QVP_Channel.Create(AOwner :TComponent);
begin
  inherited;

  ControlStyle          := ControlStyle + [csOpaque, csReflector];

  OnDragOver            := prOnDragOver;
  OnDragDrop            := prOnDragDrop;
  OnEndDrag             := prOnEndDrag;
  prDragIx              := -1;
  OnMouseDown           := prOnMouseDown;
  OnKeyPress            := prOnKeyPress;

  prBitmap              := Graphics.TBitmap.Create;
  prBitmapCanvas        := prBitmap.Canvas;
  prBitmap.Width        := Width;
  prBitmap.Height       := Height;
end;

destructor TQUB_QVO_QVP_Channel.Destroy;
begin
  try
    Channel             := nil;
    prBitmap.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVP_Channel.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited;
  prBitmap.Width        := AWidth;
  prBitmap.Height       := AHeight;
  ForceRedraw;
end;

procedure TQUB_QVO_QVP_Channel.Paint;
const
  SHAPE_VPAD_PCT = 0.05;
  NOVALUE : Double = -100000001;

var
  Shape                 : TQUB_QFS_QPF_Shape;
  i                     : Integer;
  iDisplayRep           : Integer;      // Which repetition is displayed (last)
  Rect                  : TRect;
  Samples               : INode;
  Err                   : INode;
  iSampleCount          : Integer;    // Shape length in Samples
  iSample               : Integer;
  nShapeStartTime       : Double;
  nShapeDuration        : Double;     // Shape length in seconds

  nYScale, nYMin, nYMax : Double;     // Y Range
  lYInit : boolean;                   // Whether nYMax and nYMin hve been initialized yet.

  iStartSample : Integer;
  iCurrentSample : Integer;
  nWindowStartTime : Double;
  iVisibleWidth : Integer;

  iData, iBit : Integer;              // For drawing bit data

begin
  if (Parent <> nil) and (prChannel <> nil) then begin
    try
      if not (csDestroying in ComponentState) then begin
        Rect.Bottom     := Height;

        //----- Rebuild bitmap buffer
        if prNeedsDraw then begin
          prNeedsDraw := False;         // Flag for whether to execute this block
          lYInit := False;              //
          nYMax := 1.0;
          nYMin := -1.0;

          //----- Prepare Sample Array
          iVisibleWidth := Width-prShapeEditor.Width;
          if iVisibleWidth <= 0 then Exit;
          
          Samples := TNode.Create('');
          Samples.SetupNumData(NODE_DATATYPE_DOUBLE, iVisibleWidth, 1);
          SetLength( prRightOfShape, Channel.ShapeCount );

          //----------------------------- Generate each shape
          iDisplayRep := Protocol.CurrStage.RepeatTimes - 1;
          nWindowStartTime := prProtocol.m_nProtocolStartTime;
          nShapeStartTime := 0.0;
          iCurrentSample := 0;
          //----- For drawing shape background...
          prBitmapCanvas.Pen.Width := 1;
          prBitmapCanvas.Font.Color := clWhite;
          Rect.Left := 0;
          Rect.Top := -1;
          //----- For each shape in channel
          for i := 0 to Channel.ShapeCount - 1 do begin
            //----- Shape object and time duration
            Shape := Channel.Shapes[i];
            nShapeDuration := Shape.Duration( iDisplayRep );
            prRightOfShape[i] := Round((nShapeStartTime+nShapeDuration - nWindowStartTime) / prProtocol.m_nSecondsPerPixel) - 1;

            //----- Draw maroon or navy background (active/inactive) and gray frame
            Rect.Right := prRightOfShape[i]+1;
            if Selected[i] then
              prBitmapCanvas.Brush.Color := clMaroon
            else
              prBitmapCanvas.Brush.Color := clNavy;
            prBitmapCanvas.FillRect(Rect);
            prBitmapCanvas.Brush.Color := clLtGray;
            prBitmapCanvas.FrameRect(Rect);

            //----- If at all visible, add to array of data points and draw background
            if nShapeStartTime+nShapeDuration >= nWindowStartTime then begin
              iSampleCount := Shape.SampleCount( nShapeStartTime, prProtocol.m_nSecondsPerPixel, iDisplayRep );

              if nShapeStartTime >= nWindowStartTime then
                iStartSample := 0   // Left edge visible
              else
                iStartSample := iSampleCount - Shape.SampleCount(nWindowStartTime, prProtocol.m_nSecondsPerPixel, nShapeDuration - nWindowStartTime + nShapeStartTime );

              iSampleCount := Min( iSampleCount-iStartSample, LongInt(Samples.DataCount) - iCurrentSample );
              Shape.GenerateInto(iStartSample, iStartSample + iSampleCount - 1, Max(nShapeStartTime,nWindowStartTime), prProtocol.m_nSecondsPerPixel, iDisplayRep, Samples, iCurrentSample);

              if Shape.Tree.DataAsString <> 'Bits' Then Begin
                //----- Calc min . max y values
                if not lYInit Then Begin
                  lYInit := true;
                  nYMax := Samples.Data[iCurrentSample];
                  nYMin := Samples.Data[iCurrentSample];
                end;
                for iSample := iCurrentSample to iCurrentSample + iSampleCount - 1 do begin
                  if Samples.Data[iSample] > nYMax Then nYMax := Samples.Data[iSample];
                  if Samples.Data[iSample] < nYMin Then nYMin := Samples.Data[iSample];
                end;
              end else Begin
                //----- BITS data can be drawn to bitmap since it is not Y Scale dependent.
                for iSample := iCurrentSample to iCurrentSample + iSampleCount - 1 do begin
                  iData := Round(Samples.Data[iSample]);
                  for iBit := 0 to 7 do Begin
                    if (iData and 1) > 0  Then
                      prBitmapCanvas.Pen.Color := clWhite
                    else
                      prBitmapCanvas.Pen.Color := clDkGray;
                    iData := iData shr 1;
                    prBitmapCanvas.MoveTo(iSample, (iBit*3+1) * Height DIV 26);
                    prBitmapCanvas.LineTo(iSample, (iBit*3+3) * Height DIV 26);
                  end;
                  Samples.Data[iSample]:=NOVALUE;
                end;
              end;

              iCurrentSample := iCurrentSample + iSampleCount;
            end;

            //----- check for errors, draw as text on bitmap if any.
            Err := Samples.Find('Err');
            if not Err.IsNull then begin
              prBitmapCanvas.TextRect(Rect, Rect.Left, Rect.Top, Err.DataAsString);
              Samples.RemoveChild(Err);
            end;
            Rect.Left := Rect.Right;

            //----- Increment time ptrs whether drawn or not
            nShapeStartTime := nShapeStartTime + nShapeDuration;
          end;

          //----- If additional room after last shape, fill in gray
          if Rect.Left < iVisibleWidth then begin
            Rect.Right := iVisibleWidth + 1;
            prBitmapCanvas.Brush.Color := clLtGray;
            prBitmapCanvas.FillRect(Rect);
          end;

          //----- Calc min / max Y + 5% padding
          if nYMax=nYMin Then Begin
            nYMax := nYMax+1.0;
            nYMin := nYMin-1.0;
          end;
          nYMax := (1+SHAPE_VPAD_PCT) * nYMax - SHAPE_VPAD_PCT * nYMin;
          nYScale := Height / ((nYMax-nYMin)*1.1/1.05);

          //----- Draw data trace
          if iCurrentSample > 0 then begin
            prBitmapCanvas.Pen.Color := clWhite;
            prBitmapCanvas.Pen.Width := 2;

            if Samples.Data[0] <> NOVALUE Then Begin
              i := (nYMax - Samples.Data[0]) * nYScale;
              prBitmapCanvas.MoveTo(0, i);
            end;
            for iSample := 1 to iCurrentSample - 1 do begin
              if Samples.Data[iSample]<>NOVALUE Then Begin
                i := (nYMax - Samples.Data[iSample]) * nYScale;
                if Samples.Data[iSample-1]=NOVALUE Then
                  prBitmapCanvas.MoveTo(iSample, i);
                prBitmapCanvas.LineTo(iSample, i);
              end;
            end;

            //prBitmapCanvas.LineTo(iVisibleWidth, i);
          end;

        end;

        //----- Paint from bitmap
        try
          Canvas.Lock;
          Canvas.Draw(0, 0, prBitmap);

          if (prDragIx >= 0) and (prDragIx <= Channel.ShapeCount) then begin
            Rect.Top := 0;
            if prDragIx = 0 then begin
              Rect.Left := 0;
            end else begin
              Rect.Left := prRightOfShape[prDragIx - 1];
            end;

            Rect.Right  := Rect.Left + 2;
            Canvas.Brush.Color := clFuchsia;
            Canvas.FillRect(Rect);
          end;

        finally
          Canvas.Unlock;
        end;
      end;
    finally
      inherited;
    end;
  end;
end;

procedure TQUB_QVO_QVP_Channel.ForceRedraw;
begin
  prNeedsDraw := True;
  Invalidate;
end;

procedure TQUB_QVO_QVP_Channel.AddShape(ShapeFactory :TQUB_QFS_QPF_ShapeFactory);
begin
  prChannel.InsertShape( ShapeFactory, QUB_Node_NULL, Channel.ShapeCount );
end;

procedure TQUB_QVO_QVP_Channel.DeleteCurrShape;
var
  i, j               :Integer;
begin
  for i := ShapeEditor.ShapeCount - 1 downto 0 do begin
    for j := Channel.ShapeCount - 1 downto 0 do begin
      if ShapeEditor.Shapes[i] = Channel.Shapes[j] then begin
        prChannel.RemoveShape(j);
        Break;
      end;
    end;
  end;
end;

procedure TQUB_QVO_QVP_Channel.DoneDragging;
begin
  prDragIx := -1;
  Invalidate;
end;

procedure TQUB_QVO_QVP_Channel.prSetChannel(x :TQUB_QFS_QPF_Channel);
var
  i                     :Integer;
begin
  if prChannel <> nil then begin
    for i := 0 to prChannel.ShapeCount - 1 do begin
      prChannel.Shapes[i].OnChanged := nil;
    end;

    prChannel.OnShapeAdded   := nil;
    prChannel.OnShapeRemoved := nil;
  end;

  prChannel := x;

  if prChannel <> nil then begin
    for i := 0 to prChannel.ShapeCount - 1 do begin
      prChannel.Shapes[i].OnChanged := prOnShapeChanged;
    end;

    prChannel.OnShapeAdded   := prOnShapeAdded;
    prChannel.OnShapeRemoved := prOnShapeRemoved;
  end;

  if Assigned(OnShapesChanged) then OnShapesChanged(Self);
end;

procedure TQUB_QVO_QVP_Channel.prSetCurrShape(x :TQUB_QFS_QPF_Shape);
begin
  if prCurrShape = x then Exit;
  prCurrShape := x;
  ForceRedraw;

  if Assigned(OnCurrShapeChanged) then OnCurrShapeChanged(Self, CurrShape);
end;

function TQUB_QVO_QVP_Channel.prGetCurrShapeIndex: Integer;
var
  i    :Integer;
begin
  Result := -1;
  for i := 0 to Channel.ShapeCount - 1 do begin
    if Channel.Shapes[i] = prCurrShape then begin
      Result := i;
      Exit;
    end;
  end;
end;

function TQUB_QVO_QVP_Channel.prGetSelected(i :Integer): Boolean;
begin
  Result := prShapeEditor.Contains(Channel.Shapes[i]);
end;

procedure TQUB_QVO_QVP_Channel.prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i, j        :Integer;
  Point       :TPoint;
begin
  for i := 0 to Channel.ShapeCount - 1 do begin
    if X <= prRightOfShape[i] then begin
      if (CurrShape <> nil) and ((ssShift in Shift) or (ssCtrl in Shift)) then begin

        if not Selected[i] then begin
          ShapeEditor.Add(Channel.Shapes[i]);
          CurrShape := Channel.Shapes[i];
        end else begin
          ShapeEditor.Remove(Channel.Shapes[i]);
          if ShapeEditor.ShapeCount = 0 then begin
            CurrShape := nil;
          end else begin
            for j := 0 to Channel.ShapeCount - 1 do begin
              if Selected[j] then begin
                CurrShape := Channel.Shapes[j];
                Break;
              end;
            end;
          end;
        end;

        ForceRedraw;

      end else if (Button = mbLeft) and (CurrShape = Channel.Shapes[i]) then begin

      end else if (Button = mbLeft) or not Selected[i] then begin

        ShapeEditor.Clear;
        ShapeEditor.Add(Channel.Shapes[i]);

        CurrShape := Channel.Shapes[i];

        ForceRedraw;

      end;

      if Selected[i] and (ShapeEditor.ShapeCount = 1) and (Button = mbLeft) then begin
        prDragSourceIx := i;
        BeginDrag(False, 5);
      end else if (ShapeEditor.ShapeCount > 0) and (Button = mbRight) and (PopupMenu <> nil) then begin
        Point.X := X;
        Point.Y := Y;
        Point   := ClientToScreen( Point );
        PopupMenu.Popup(Point.X, Point.Y);
      end;

      SetFocus;
      Exit;
    end;
  end;

  CurrShape := nil;
  ShapeEditor.Clear;
end;

// Note : This executes for backspace but not special chars, including delete,
// home, end, ... etc.  Need to use OnKeyDown, onKeyUp to handle delete Key.
procedure TQUB_QVO_QVP_Channel.prOnKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = Char($8){Backspace}) or (Key = Char($7f){Delete}) then begin
    DeleteCurrShape;
    Key := Char($0);
  end;
end;

procedure TQUB_QVO_QVP_Channel.prOnDragOver(Sender, Source :TObject; X, Y :Integer;
                       State :TDragState; var Accept :Boolean);
var
  i          :Integer;
  Left       :Integer;
begin
  Accept := (Source is TQUB_ProtocolShapeButton) or (Source is TQUB_QVO_QVP_Channel);

  i := 0;
  while (i < Channel.ShapeCount) and (prRightOfShape[i] < X) do Inc(i);
  if i < Channel.ShapeCount then begin
    Left     := 0;
    if i > 0 then Left := prRightOfShape[i-1];
    Inc(i, Round( (X - Left) * 1.0 / (prRightOfShape[i] - Left) ));
  end;

  prDragIx := i;
  Invalidate;
end;

procedure TQUB_QVO_QVP_Channel.prOnDragDrop(Sender, Source :TObject; X, Y :Integer);
var
  i           :Integer;
  ShapeTree   :INode;
begin
  i           := prDragIx;
  prDragIx    := -1;
  if Source is TQUB_ProtocolShapeButton then begin
    ShapeTree := TNode.Create('Shape');
    ShapeTree.SetData( TQUB_ProtocolShapeButton(Source).Shape.Name );
    prChannel.InsertShape( ShapeTree, i );
  end else if Source is TQUB_QVO_QVP_Channel then begin
    if (Source = Self) and (i <= prDragSourceIx) then Inc(prDragSourceIx);
    prChannel.InsertShape( TQUB_QVO_QVP_Channel(Source).CurrShape.Tree.Clone(True), i );
  end;
end;

procedure TQUB_QVO_QVP_Channel.prOnEndDrag(Sender, Target: TObject; X, Y: Integer);
begin
  if Target <> nil then prChannel.RemoveShape( prDragSourceIx );
end;

procedure TQUB_QVO_QVP_Channel.prOnShapeAdded(Sender :TObject; Index :Integer);
var
  Shape                :TQUB_QFS_QPF_Shape;
begin
  CurrShape            := nil; // unselect all
  ShapeEditor.Clear;
  Shape                := Channel.Shapes[Index];
  ShapeEditor.Add(Shape);
  CurrShape            := Shape;
  Shape.OnChanged      := prOnShapeChanged;
  if Assigned(OnShapesChanged) then OnShapesChanged(Self);
end;

procedure TQUB_QVO_QVP_Channel.prOnShapeRemoved(Sender :TObject; OldIndex :Integer; Shape :TQUB_QFS_QPF_Shape);
begin
  ShapeEditor.Remove(Shape);
  if Channel.Shapes[OldIndex] = CurrShape then CurrShape := nil;
  if Assigned(OnShapesChanged) then OnShapesChanged(Self);
end;

procedure TQUB_QVO_QVP_Channel.prOnShapeChanged(Sender :TObject);
begin
  if Assigned(OnShapesChanged) then OnShapesChanged(Self);
end;


// *****************************************************************************
end.
