{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the bar underneath the shapes in the Acquisition Protocol

//  added option to Randomize the order of reps within a stage. 6/2005 {JB}


unit QUB_QVO_QVP_Stage_Properties;

interface

uses 
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, StdCtrls, ExtCtrls,

  QUB_QFS_QPF;

type
  TQUB_QVO_QVP_Stage_PropertiesFrame = class(TFrame)
    Panel1: TPanel;
    chkRepeatNone: TRadioButton;
    chkRepeatTimes: TRadioButton;
    txtRepeatTimes: TEdit;
    udRepeatTimes: TUpDown;
    Label4: TLabel;
    Label2: TLabel;
    Panel2: TPanel;
    Label3: TLabel;
    Label5: TLabel;
    chkPauseNone: TRadioButton;
    chkPauseSeconds: TRadioButton;
    txtPauseSeconds: TEdit;
    chkPauseUser: TRadioButton;
    chkRandomize: TCheckBox;
    procedure chkRepeatNoneClick(Sender: TObject);
    procedure chkPauseNoneClick(Sender: TObject);
    procedure udRepeatTimesClick(Sender: TObject; Button: TUDBtnType);
    procedure txtPauseSecondsKeyPress(Sender: TObject; var Key: Char);
    procedure txtPauseSecondsExit(Sender: TObject);
    procedure txtRepeatTimesKeyPress(Sender: TObject; var Key: Char);
    procedure txtRepeatTimesExit(Sender: TObject);
    procedure chkRandomizeClick(Sender: TObject);
  protected

    prStage                  :TQUB_QFS_QPF_Stage;
    prChangingUD             :Boolean;

    procedure prSetStage(x :TQUB_QFS_QPF_Stage);

  public

    procedure DoOnRepeatChanged(Sender :TObject);
    procedure DoOnPauseChanged(Sender :TObject);

    property Stage      :TQUB_QFS_QPF_Stage    read prStage    write prSetStage;
  end;

implementation

{$R *.DFM}

// *****************************************************************************

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.prSetStage(x :TQUB_QFS_QPF_Stage);
begin
  prStage := x;

  if Stage <> nil then begin
    DoOnRepeatChanged(Stage);
    DoOnPauseChanged(Stage);
  end;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.DoOnRepeatChanged(Sender :TObject);
begin
  if prChangingUD then Exit;

  chkRepeatNone.Checked    := Stage.RepeatMode = QUB_QFS_QPF_REPEAT_NONE;
  chkRepeatTimes.Checked   := Stage.RepeatMode = QUB_QFS_QPF_REPEAT_TIMES;
  udRepeatTimes.Position   := Stage.RepeatTimes;
  udRepeatTimes.Enabled    := chkRepeatTimes.Checked;
  txtRepeatTimes.Enabled   := chkRepeatTimes.Checked;

  chkRandomize.Checked := Stage.Randomize;
  chkRandomize.Enabled := chkRepeatTimes.Checked;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.DoOnPauseChanged(Sender :TObject);
begin
  chkPauseNone.Checked     := Stage.PauseMode = QUB_QFS_QPF_PAUSE_NONE;
  chkPauseSeconds.Checked  := Stage.PauseMode = QUB_QFS_QPF_PAUSE_SECONDS;
  txtPauseSeconds.Text     := FloatToStr( Stage.PauseSeconds );
  txtPauseSeconds.Enabled  := chkPauseSeconds.Checked;
end;

// *****************************************************************************

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.chkRepeatNoneClick(
  Sender: TObject);
begin
  Stage.RepeatMode := TQUB_QFS_QPF_RepeatMode(TRadioButton(Sender).Tag);
  txtRepeatTimes.Enabled := Stage.RepeatMode = QUB_QFS_QPF_REPEAT_TIMES;
  udRepeatTimes.Enabled := Stage.RepeatMode = QUB_QFS_QPF_REPEAT_TIMES;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.chkPauseNoneClick(
  Sender: TObject);
begin
  Stage.PauseMode := TQUB_QFS_QPF_PauseMode(TRadioButton(Sender).Tag);
  txtPauseSeconds.Enabled := Stage.PauseMode = QUB_QFS_QPF_PAUSE_SECONDS;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.udRepeatTimesClick(
  Sender: TObject; Button: TUDBtnType);
begin
  prChangingUD := True;
  Stage.RepeatTimes := udRepeatTimes.Position;
  prChangingUD := False;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.txtPauseSecondsKeyPress(
  Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtPauseSecondsExit(Sender);
  end;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.txtPauseSecondsExit(Sender: TObject);
var
  x    :Double;
begin
  inherited;
  try
    x := StrToFloat(TEdit(Sender).Text);
    if x <= 0.0 then raise Exception.Create('');
    Stage.PauseSeconds := x;
  except
    MessageDlg('Invalid value; must be positive.', mtError, [mbOK], 0);
    txtPauseSeconds.Text := FloatToStr(Stage.PauseSeconds);
    TEdit(Sender).SetFocus;
    Exit;
  end;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.txtRepeatTimesKeyPress(
  Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtRepeatTimesExit(Sender);
  end;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.txtRepeatTimesExit(Sender: TObject);
var
  x    :Integer;
begin
  inherited;
  try
    x := StrToInt(TEdit(Sender).Text);
    if x <= 0 then raise Exception.Create('');
    Stage.RepeatTimes := x;
  except
    MessageDlg('Invalid value; must be positive.', mtError, [mbOK], 0);
    txtRepeatTimes.Text := IntToStr(Stage.RepeatTimes);
    TEdit(Sender).SetFocus;
    Exit;
  end;
end;

procedure TQUB_QVO_QVP_Stage_PropertiesFrame.chkRandomizeClick( Sender: TObject);
begin
  Stage.Randomize := chkRandomize.Checked;
end;


// *****************************************************************************
end.
