{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
{

                               Qub Visual Results
                                     unit



}
unit QUB_QVO_QVR;


//******************************************************************************
interface


uses


Windows, SysUtils, Classes, Graphics, Menus, Controls, Messages, Printers, Registry, StdCtrls, ExtCtrls, ComCtrls, Dialogs, Buttons, Forms, FileCtrl,

Alloc,

QUB_Console, QUB_Rich_Console,          

QUB_Node,
QUB_PyCmd_MenuButton,
QUB_Viewport_Histogram,
QUB_QVO,
QUB_Table,
QUB_Global,
AllocTypes,

Chart, Series, TeEngine, TeeProcs,
QUB_ColorPanel,
PythonEngine,
QUB_Python_Thread,

QUB_Results,
QUB_Results_Select,
QUB_Results_Criteria,
QUB_Results_Models,
QUB_Results_Histograms;



const

WM_QUB_RESULTS_ADD = WM_USER + 105;

//******************************************************************************
type


//Forward class declarations ***************************************************
TQUB_QVO_QVR                                   = class;
TQUB_Python_Results                            = class;

TQUB_Results_SumPanel                          = class;
TQUB_Results_SegPanel                          = class;
TQUB_Results_ModelPanel                        = class;


//Object
TQUB_QVO_QVR                                   = class(TQUB_QVO)
protected

  prResults                                    :TQUB_Results;
  prPyHandle                                   :TQUB_Python_Results; // python surrogate of Self

  panMenus                                     :TPanel;
  btnReports                                   :TSpeedButton;
  popupReports                                 :TPopupMenu;

  btnShowInWindow                              :TSpeedButton;

  prTabs                                       :TPageControl;

  tabSum                                       :TTabSheet;
  panSum                                       :TQUB_Results_SumPanel;

  tabSelect                                    :TTabSheet;
  panSelect                                    :TQUB_Results_SelectPanel;

  tabCrit                                      :TTabSheet;
  panCrit                                      :TQUB_Results_CriteriaPanel;

  tabSegs                                      :TTabSheet;
  panSegs                                      :TQUB_Results_SegPanel;

  tabModels                                    :TTabSheet;
  //panModels                                    :TQUB_Results_ModelPanel;
  grdModels                                    :TQUB_Results_ModelTable;
  prModelName                                  :String; // name of QVM

  tabHist                                      :TTabSheet;
  panHist                                      :TQUB_Results_HistGroupDisplay;

  //tabText                                      :TTabSheet;
  //memText                                      :TMemo;

  prDataFiles                                  :TStringList; // name -> QVD or nil
  prSegsPerFile                                :array of Integer;

  procedure prCreatePyInstance;
  procedure prCleanupPyInstance;

  procedure prOnReportsPopup(Sender :TObject);
  procedure prOnReportsNew(Sender :TObject);
  procedure prOnReportsChoice(Sender :TObject);
  procedure prOnReportsEdit(Sender :TObject);
  procedure prOnReportsMouseUp(Sender: TObject; MouseButton: TMouseButton; Shift: TShiftState; X, Y: Integer);

  procedure prOnClickShowInWindow(Sender :TObject);

  procedure prOnSelectionChanged(Sender :TObject);
  procedure prOnShowSegClick(Sender :TObject; Ix :Integer);
  procedure prOnColorChange(Sender :TObject);
  procedure prOnCriteriaAction(Sender :TObject);

  function prGetModelObject: TQUB_QVO;
  procedure prOnModelChosen(Sender :TObject);

  procedure prSetupDataLinks;
  procedure prUpdateDataLinks; // make prDataFiles valid
  procedure prUpdateSelLists;

  //Events
  procedure DoOnFileOpen(Sender :TObject); override;
  procedure DoOnFileClose(Sender :TObject); override;
  // procedure DoOnFileEdit(Sender :TObject; EditOper :TQUB_EditOperation; var Param, var Param2); override;

  procedure prSetTipsOn(t :Boolean); override;

  //Print
  function prPrint(Dlg :TPrintDialog) :Boolean; override;

  procedure prSetInitialProperties; override;

  //Registry
  procedure prPropertiesRegistry(Read :Boolean); override;
  //****************************************************************************

  function prGetFileName :String; override;
  function prGetFileExt :String; override;
  function prGetModified :Boolean; override;
  function prGetSaved :Boolean; override;

  //Properties
  procedure prOnGetProperties(Sender :TObject); override;
  procedure prOnPutProperties(Sender :TObject); override;
  procedure prOnResetProperties(Sender :TObject); override;
  procedure prOnSetAsDefaultProperties(Sender :TObject); override;

  //Popup Menu
  procedure prCreatePopUpMenu; override;

  {//PopUp Menu Events
  procedure prOnCopyImage(Sender :TObject); override;
  procedure prOnPrint(Sender :TObject); override;
  procedure prOnProperties(Sender :TObject); override;
  procedure prOnPopUp(Sender :TObject); override;
            }

public

  constructor Create(AOwner :TComponent; AResults :TQUB_Results); reintroduce; virtual;
  destructor Destroy; override;

  procedure BuildInterface; // do constructor things that need parent set first    e.g. tab sheets

  procedure ChooseSelection( SelData :TQUB_QVO; Sel :TQUB_FirstLast ); // highlight the seg with these bounds

  procedure ShowSelect;
  procedure ShowModels;
  procedure ShowHistograms;

  procedure Paint; override;
  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

  //File .......................................................................
  // these four should never be used, but they're abstract in QVO
  function New(const FileName :String) :Boolean; override;
  function Open(const FileName :String) :Boolean; override;
  function Open(const FileName :String; OpenStream :TStream) :Boolean; override;
  function Close :Boolean; override;
  function OpenFromString(const MemFile :String) :Boolean; override;

  function Save(const FileName :String) :Boolean; override;
  function CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;
  function Rename(const Name :String) :Boolean;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

  function Properties :Boolean; override;

  procedure Activate; override;
  // procedure DeActivate; override;

  //Draw .......................................................................
  function Draw :Boolean; override;

  //Lock
  procedure LockWrite; override;
  procedure UnLockWrite; override;
  procedure LockReadWrite; override;
  procedure UnLockReadWrite; override;

  property Results                             :TQUB_Results             read prResults;
  property PyHandle                            :TQUB_Python_Results      read prPyHandle;

  property Select                              :TQUB_Results_SelectPanel read panSelect;
  property Criteria                            :TQUB_Results_CriteriaPanel read panCrit;
  property Histograms                          :TQUB_Results_HistGroupDisplay read panHist;

  //property PropertiesForm                      :TQUB_QVO_QVM_PropertiesForm        read prPropertiesForm            write prPropertiesForm;


end;

TQUB_Python_Results = class(TPyObject)
protected
  prResults     :TQUB_QVO_QVR;

  prSegIndex    :Integer;
  prSegColor    :Integer;
  prChartIndex  :Integer;

  procedure prDoUpdate;
  procedure prSetSegColor;
  procedure prCopyChartImage;

public
  function  GetAttr(key : PChar) : PPyObject; override;
  //function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  // Mapping services
  function  MpLength : Integer; override;
  function  MpSubscript( obj : PPyObject) : PPyObject; override;
  //function  MpAssSubscript( obj1, obj2 : PPyObject) : Integer; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;

  function Keys( args :PPyObject ): PPyObject; cdecl;
  function UpdateSelection( args :PPyObject): PPyObject; cdecl;
  function DoUpdatesPy( args :PPyObject ): PPyObject; cdecl;
  function Segment( args :PPyObject ): PPyObject; cdecl;
  function SegmentColor( args :PPyObject ): PPyObject; cdecl;
  function SetSegmentColor( args :PPyObject ): PPyObject; cdecl;
  function SegmentsOfColor( args :PPyObject ): PPyObject; cdecl;
  function CopyChartImage( args :PPyObject ): PPyObject; cdecl;

  property Results   :TQUB_QVO_QVR  read prResults   write prResults;
end;


TQUB_Results_SumPanel                          = class(TPanel)
protected
  prResults                     :TQUB_Results;
  gridVars                      :TQUB_Table;

  prPopup                       :TPopupMenu;

  procedure prOnCopy(Sender :TObject);

public
  constructor Create(AOwner :TComponent; Results :TQUB_Results); reintroduce; virtual;
  destructor Destroy; override;

end;

TQUB_Results_SegPanel                          = class(TPanel)
protected
  prResults                     :TQUB_Results;
  gridSegs                      :TQUB_Table;

  prPopup                       :TPopupMenu;

  procedure prOnCopy(Sender :TObject);

public
  constructor Create(AOwner :TComponent; Results :TQUB_Results); reintroduce; virtual;
  destructor Destroy; override;

end;

TQUB_Results_ModelPanel                        = class(TPanel)
protected
  prResults                     :TQUB_Results;
  lstModels                     :TListBox;
  prChosen                      :Integer;

  procedure prOnListClick(Sender :TObject);

public
  OnModelChosen                 :TNotifyEvent;

  constructor Create(AOwner :TComponent; Results :TQUB_Results); reintroduce; virtual;
  procedure Initialize;
  destructor Destroy; override;

  property Chosen               :Integer              read prChosen;
end;


var
  ResultsPyType                        :TPythonType;

//..............................................................................




// procedure Register;



//******************************************************************************
implementation


uses


QUB_QVO_QVR_Window,
QUB_Viewport_Python,
QUB_Node_Python,
QUB_Node_Model,
QUB_QFS_Channels,
QUB_GUIInterface,
QUB_Viewport_Data,
QUB_QVO_QVD,
QUB_QVO_QVD_Global,
QUB_Viewport_Model,
QUB_Viewport_Results,
QUB_QVO_QVM,
QUB_ListBox,
QUB_Reports_Editor,
QUB_Reports_Manager,
QUB_Main,
SystemPath,
Math;



//******************************************************************************
//                              TQUB_QVO_QVR
//******************************************************************************
constructor TQUB_QVO_QVR.Create(AOwner :TComponent; AResults :TQUB_Results);
begin
  inherited Create(AOwner);

  prResults               := AResults;
  prResults.Incref;

  prDataFiles             := TStringList.Create;

  //while PythonThread.SyncPyProc( prCreatePyInstance, False, 50 ) do Application.HandleMessage;
   PythonThread.SneakPyProc( prCreatePyInstance );
end;

procedure TQUB_QVO_QVR.prCreatePyInstance;
var
  PyInst                  :PPyObject;
begin
  PyInst                  := ResultsPyType.CreateInstance;
  prPyHandle              := TQUB_Python_Results( PythonToDelphi(PyInst) );
  prPyHandle.Results      := Self;
end;

procedure TQUB_QVO_QVR.prCleanupPyInstance;
begin
  prPyHandle.Results := nil;
  GetPythonEngine.Py_DECREF( prPyHandle.GetSelf );
end;

procedure TQUB_QVO_QVR.BuildInterface;
var
  i                       :Integer;
begin
  btnReports              := TSpeedButton.Create(Self);
  btnReports.Caption      := 'Reports';
  btnReports.OnMouseUp    := prOnReportsMouseUp;

  popupReports            := TPopupMenu.Create(Self);
  popupReports.OnPopup    := prOnReportsPopup;
  popupReports.AutoHotkeys := maManual;

  panMenus                := TPanel.Create(Self);
  panMenus.Anchors        := [akLeft, akTop, akRight];
  panMenus.Left           := 0;
  panMenus.Top            := 0;
  panMenus.Width          := Width;
  panMenus.Height         := btnReports.Height + 4;
  panMenus.Parent         := Self;

  btnReports.Width        := 80;
  btnReports.Top          := 2;
  btnReports.Left         := 5;
  btnReports.Parent       := panMenus;

  btnShowInWindow         := TSpeedButton.Create(Self);
  btnShowInWindow.Anchors := [akRight, akTop];
  btnShowInWindow.Caption := 'Pop up window';
  btnShowInWindow.Parent  := panMenus;
  btnShowInWindow.Top     := 2;
  btnShowInWindow.Width   := 100;
  btnShowInWindow.Left    := panMenus.Width - btnShowInWindow.Width - 5;
  btnShowInWindow.OnClick := prOnClickShowInWindow;


  prTabs                  := TPageControl.Create(Self);
  prTabs.Anchors          := [akLeft, akTop, akRight, akBottom];
  prTabs.Left             := 0;
  prTabs.Top              := panMenus.Height;
  prTabs.Width            := Width;
  prTabs.Height           := Height - panMenus.Height;
  prTabs.Parent           := Self;
  prTabs.ParentShowHint   := True;

  try
    tabSum                  := TTabSheet.Create(Self);
    tabSum.Caption          := 'Summary';
    tabSum.ParentShowHint   := True;
    tabSum.PageControl      := prTabs;

    panSum                  := TQUB_Results_SumPanel.Create(Self, Results);
    panSum.Parent           := tabSum;
    panSum.Align            := alClient;
  except
  end;

  panSelect                 := TQUB_Results_SelectPanel.Create(Self, Results);
  panCrit                   := TQUB_Results_CriteriaPanel.Create(Self);

  if Results.SegCount > 1 then begin
    try
      tabSelect               := TTabSheet.Create(Self);
      tabSelect.Caption       := 'Select';
      tabSelect.ParentShowHint := True;
      tabSelect.PageControl   := prTabs;

      panSelect.Parent         := tabSelect;
      panSelect.Align          := alClient;
      panSelect.BuildInterface;
      panSelect.OnShowSegClick := prOnShowSegClick;
      panSelect.OnColorChange  := prOnColorChange;

      tabCrit                 := TTabSheet.Create(Self);
      tabCrit.Caption         := 'Criteria';
      tabCrit.PageControl     := prTabs;

      panCrit.Parent          := tabCrit;
      panCrit.Initialize(Results);
      panCrit.Align           := alClient;
      panCrit.OnAction        := prOnCriteriaAction;
      panCrit.OnColorChange   := prOnColorChange;
    except
    end;
  end;

  panSegs                     := TQUB_Results_SegPanel.Create(Self, Results);

  if Results.SegCount > 0 then begin
    try
      tabSegs                 := TTabSheet.Create(Self);
      tabSegs.Caption         := 'Segments';
      tabSegs.PageControl     := prTabs;

      panSegs.Parent          := tabSegs;
      panSegs.Align           := alClient;
    except
    end;
  end;

  grdModels                   := TQUB_Results_ModelTable.Create(Self, Results);

  if Results.Models.Count > 0 then begin
    try
      tabModels               := TTabSheet.Create(Self);
      tabModels.Caption       := 'Models';
      tabModels.PageControl   := prTabs;

      grdModels.Parent        := tabModels;
      grdModels.Align         := alClient;
      grdModels.OnModelChosen := prOnModelChosen;

      if Results.Models[0].HasValue[' LL'] then begin
        Results.Models.SortBy(' LL');
      end else begin
        Results.Models.SortBy( Results.Models.VarName[0] );
      end;
    except
    end;
  end;

  panHist                   := TQUB_Results_HistGroupDisplay.Create(Self);

  try
    tabHist                 := TTabSheet.Create(Self);
    tabHist.Caption         := 'Histograms';
    tabHist.PageControl     := prTabs;

    panHist.Parent          := tabHist;
    panHist.TipsOn          := TipsOn;
    panHist.Align           := alClient;
    panHist.OnModelChosen   := prOnModelChosen;
    for i := 0 to Results.HistGroupCount - 1 do begin
      panHist.AddGroup( Results.HistGroup[i] );
    end;
  except
  end;

  if (Results.SegCount = 0) and (Results.Models.Count > 0) and (tabModels <> nil) then begin
    prTabs.ActivePage     := tabModels;
  end else if (Results.SegCount = 0) or (panHist.GroupCount > 0) and (tabHist <> nil) then begin
    prTabs.ActivePage     := tabHist;
  end else if (Results.SegCount > 1) and (tabSelect <> nil) then begin
    prTabs.ActivePage     := tabSelect;
  end;

  try
    //PopUp Menu
    prCreatePopUpMenu;

    prDrawFocusRect                 := False;

    //To give some initial values, before any calculations are made:
    prSetInitialProperties;

    //Read properties from registry:
    prPropertiesRegistry(True{Read});

   //Flags
    prMustRedraw                    := True;

    if tabSelect <> nil then prUpdateSelLists;

    Results.Events.AddEvent('SelectionChanged', prOnSelectionChanged);
  except
  end;
end;

destructor TQUB_QVO_QVR.Destroy;
var
  Data    :TQUB_QVO_QVD;
  i       :Integer;
begin
  try
    if (Self = ViewportResults.CurrResults)
    and (Parent = ViewportResults.ConteinerLeft) then begin
      ViewportResults.CurrResults := nil;
    end;
    
    prUpdateDataLinks;
    for i := 0 to prDataFiles.Count - 1 do begin
      Data := TQUB_QVO_QVD( prDataFiles.Objects[i] );
      if Data <> nil then Data.DelList( Results.ResultType + ' Results' );
    end;

    prDataFiles.Free;

    // free all the components that rely on results
    if panSelect <> nil then panSelect.Free;
    if panSegs <> nil then panSegs.Free;
    if grdModels <> nil then grdModels.Free;
    panHist.Free;

    prResults.Events.DelEvent('SelectionChanged', prOnSelectionChanged);
    prResults.Decref;

    //while PythonThread.SyncPyProc( prCleanupPyInstance, False, 50 ) do Application.HandleMessage;
    PythonThread.SneakPyProc( prCleanupPyInstance );

  finally
    inherited;
  end;
end;


procedure TQUB_QVO_QVR.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  if prDrawForPrinting then Exit;
  try
    if (AWidth <> Width) or (AHeight <> Height) then begin
      //ResizeModel(AWidth, AHeight);
      prMustRedraw := True;
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_QVO_QVR.ChooseSelection( SelData :TQUB_QVO; Sel :TQUB_FirstLast );
var
  i       :Integer;
begin
  for i := 0 to prResults.SegCount - 1 do begin
    if (prResults.Segment[i].Data[0] = Sel.First) and (prResults.Segment[i].Data[1] = Sel.Last)
    and (prResults.Segment[i].Parent['FileName'].DataAsString = SelData.FileName) then begin
      panSelect.ChosenSeg := i;
      Exit;
    end;
  end;
end;

procedure TQUB_QVO_QVR.ShowSelect;
begin
  prTabs.ActivePage       := tabSelect;
end;

procedure TQUB_QVO_QVR.ShowModels;
begin
  prTabs.ActivePage       := tabModels;
end;

procedure TQUB_QVO_QVR.ShowHistograms;
begin
  prTabs.ActivePage       := tabHist;
end;

procedure TQUB_QVO_QVR.prOnReportsPopup(Sender :TObject);
var
  Item           :TMenuItem;
  Dir            :String;
  sr             :TSearchRec;
begin
  popupReports.Items.Clear;

  Item           := TMenuItem.Create(Self);
  Item.Caption   := 'New Report...';
  Item.OnClick   := prOnReportsNew;
  popupReports.Items.Add(Item);

  popupReports.Items.NewBottomLine;

  Dir := ExtractFilePath(Application.ExeName) + 'Reports\' + Results.Node['ResultType'].DataAsString + '\';
  if 0 = FindFirst(Dir + '*.txt', 0, sr) then begin
    repeat
      Item          := TMenuItem.Create(Self);
      Item.Caption  := Copy(sr.Name, 0, Length(sr.Name) - 4);
      Item.OnClick  := prOnReportsChoice;
      popupReports.Items.Add(Item);
    until 0 <> FindNext(sr);
  end;
  Sysutils.FindClose(sr);

  popupReports.Items.NewBottomLine;

  Item           := TMenuItem.Create(Self);
  Item.Caption   := 'Edit Reports...';
  Item.OnClick   := prOnReportsEdit;
  popupReports.Items.Add(Item);
end;

procedure TQUB_QVO_QVR.prOnReportsNew(Sender :TObject);
var
  ReportsForm         :TQUB_Reports_EditForm;
begin
  ReportsForm         := TQUB_Reports_EditForm.Create(nil);
  ReportsForm.Results := Self;
  ReportsForm.ReportName := 'untitled';
  ReportsForm.ShowModal;
  ReportsForm.Release;
end;

procedure TQUB_QVO_QVR.prOnReportsChoice(Sender :TObject);
var
  Task         :TQUB_Reports_Task_Run;
  Options      :TStringList;
begin
  Options      := TStringList.Create;
  try
    Options.LoadFromFile(ExtractFilePath(Application.ExeName) + 'Reports\' + Results.Node['ResultType'].DataAsString + '\' + TMenuItem(Sender).Caption + '.txt');
    Task       := TQUB_Reports_Task_Run.Create(Self, Options);
    PythonThread.SyncPyTask(Task, True{Async -- auto-freed}, False{not Silent}, -1{no timeout});
  finally
    Options.Free;
  end;
end;

procedure TQUB_QVO_QVR.prOnReportsEdit(Sender :TObject);
var
  ReportsForm         :TQUB_Reports_ManageForm;
begin
  ReportsForm         := TQUB_Reports_ManageForm.Create(nil);
  ReportsForm.Results := Self;
  ReportsForm.ShowModal;
  ReportsForm.Release;
end;

procedure TQUB_QVO_QVR.prOnReportsMouseUp(Sender: TObject; MouseButton: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Point   :TPoint;
begin
  inherited;
  Point.X := X;
  Point.Y := Y;
  Point   := btnReports.ClientToScreen( Point );
  popupReports.Popup(Point.X,Point.Y);
end;

procedure TQUB_QVO_QVR.prOnClickShowInWindow(Sender :TObject);
var
  Win         :TQUB_QVO_QVR_Window;
begin
  Win         := TQUB_QVO_QVR_Window.Create(Application, Results);
  if prTabs.ActivePage = tabSelect then begin
    Win.QVR.ShowSelect;
  end else if prTabs.ActivePage = tabModels then begin
    Win.QVR.ShowModels;
  end else if prTabs.ActivePage = tabHist then begin
    Win.QVR.ShowHistograms;
  end;
end;

procedure TQUB_QVO_QVR.prOnSelectionChanged(Sender :TObject);
begin
  prUpdateSelLists;
end;

procedure TQUB_QVO_QVR.prOnShowSegClick(Sender :TObject; Ix :Integer);
var
  i, j    :Integer;
  Data    :TQUB_QVO_QVD;
  Sel     :TQUB_FirstLast;
begin
  prUpdateDataLinks;

  if Length(prSegsPerFile) <= 0 then Exit;

  i       := 0; // data ix
  j       := prSegsPerFile[0]; // ix of first seg next file
  while j <= Ix do begin
    i     := i + 1;
    j     := j + prSegsPerFile[i];
  end;

  Data    := TQUB_QVO_QVD( prDataFiles.Objects[i] );
  if Data <> nil then begin
    Sel.First := Integer(Results.Segment[ Ix ].Data[0]);
    Sel.Last  := Integer(Results.Segment[ Ix ].Data[1]);
    ViewportData.ListObjects.SelQVO(Data);
    Data.MoveSelectionIntoWindow( Sel, Sel.First );

    while (Sel.First <> (Data.Zoom.DataDraw.First+Data.Zoom.DataOffset)) or (Sel.Last <> (Data.Zoom.DataDraw.Last+Data.Zoom.DataOffset))
    do Application.ProcessMessages;

    if not Results.Segment[Ix].Find('DialACurve').IsNull then begin
      Data.Zoom.ShowFit(Results.Segment[Ix]['DialACurve'], False);
    end;

    //Name             := Results.ResultType + ' Results';
    //List             := Data.GetList(Name);
    //List.Choose( List.IndexOf(Sel.First, Sel.Last) );
  end;
end;

procedure TQUB_QVO_QVR.prOnColorChange(Sender: TObject);
begin
  if Sender = panSelect then begin
    panCrit.ChosenColor := panSelect.ChosenColor;
  end else begin
    panSelect.ChosenColor := panCrit.ChosenColor;
  end;

  prUpdateSelLists;
end;

procedure TQUB_QVO_QVR.prOnCriteriaAction(Sender: TObject);
begin
  prTabs.ActivePage := tabSelect;
end;

function TQUB_QVO_QVR.prGetModelObject: TQUB_QVO;
var
  ModelPath  :String;
begin
  Result := ViewportModel.ListObjects.QVOByName[ prModelName ];
  if Result = nil then begin
    ModelPath := GetSystemPath(SystemPath.AppData)+'QUB\Models\';
    if not DirectoryExists( ModelPath ) then CreateDir( ModelPath );
    ModelPath := ModelPath + ExtractFileName(prModelName);
    if FileExists( ModelPath ) then DeleteFile( ModelPath );
    if not FileExists( ModelPath ) then ViewportModel.New( ModelPath, True );
    Result := g_CurrModel();
    if (Result <> nil) and (ExtractFileName(Result.FileName) <> ExtractFileName(prModelName)) then begin
      Result := nil;
    end;
  end;

  if Result <> nil then ViewportModel.ListObjects.SelQVO( Result );
end;

procedure TQUB_QVO_QVR.prOnModelChosen(Sender :TObject);
var
  ix      :Integer;
  Model   :TQUB_QVO_QVM;
  Node    :INode;
begin
  Node    := QUB_Node_Null;

  if Sender = grdModels then begin
    ix      := grdModels.Chosen;
    if ix >= 0 then begin
      Node  := prResults.Models[ix].Node;
    end;
  end else begin
    Node := panHist.CurrGroup.Node.Parent;
    while (not Node.IsNull) and (Node.Name <> 'ModelTrial') do begin
      Node := Node.Parent;
    end;
    Node := Node.Find('ModelFile');
  end;

  if not Node.IsNull then begin
    prModelName := GetBaseStampedFileName(
                     SafeFileName(prResults.ResultType+'-'+Node.DataAsString),
                     '.qmf', prResults.DateTime, prResults.DateTime );
    Model := TQUB_QVO_QVM( prGetModelObject );
    if (Model <> nil) and (not Model.LockedAny) then begin
      SetModelFromNode( Model.ModelObject, Node );
      // Model.Modified := False;
    end;
  end;

  if Sender = grdModels then grdModels.SetFocus;
end;

procedure TQUB_QVO_QVR.prSetupDataLinks;
var
  Count       :Integer;
  dsi         :Integer;
  ds, seg     :INode;
  overallSegi :Integer;
begin
  Count       := 0; // dataset count
  ds          := Results.Node.Find('DataSet');
  while not ds.isNull do begin
    prDataFiles.Add( ds['FileName'].GetDataAsString );
    Count     := Count + 1;
    ds        := ds.SiblingSameName;
  end;

  SetLength( prSegsPerFile, Count );

  overallSegi := 0;
  dsi         := 0;
  ds          := Results.Node.Find('DataSet');
  while not ds.isNull do begin
    seg       := ds.Find('Segment');
    Count     := 0; // segment count this file
    while not seg.isNull do begin
      Count   := Count + 1;

      if overallSegi < panHist.GroupCount then begin
        try
          panHist.Groups[overallSegi].DataSel := QUB_FirstLast(Integer(seg.Data[0]), Integer(seg.Data[1]));
        except
        end;
      end;

      seg     := seg.SiblingSameName;
      overallSegi := overallSegi + 1;
    end;

    prSegsPerFile[dsi] := Count;

    dsi       := dsi + 1;
    ds        := ds.SiblingSameName;
  end;
end;

procedure TQUB_QVO_QVR.prUpdateDataLinks;
var
  DataList         :TQUB_ListBox;
  i, segoff, j     :Integer;
  Data             :TQUB_QVO_QVD;
begin
  if prDataFiles.Count = 0 then prSetupDataLinks;

  DataList := ViewportData.ListObjects;

  for i := 0 to prDataFiles.Count - 1 do begin
    prDataFiles.Objects[i] := DataList.QVOByName[ ExtractFileName( prDataFiles[i] ) ];
  end;

  if panHist.GroupCount = Results.SegCount then begin
    segoff := 0;
    for i := 0 to prDataFiles.Count - 1 do begin
      Data := TQUB_QVO_QVD( prDataFiles.Objects[i] );
      for j := 0 to prSegsPerFile[i] - 1 do begin
        panHist.Groups[ segoff + j ].Data := Data;
      end;
      segoff := segoff + prSegsPerFile[i];
    end;
  end;
end;

procedure TQUB_QVO_QVR.prUpdateSelLists;
begin
end;
{
var
  i, segoff, j     :Integer;
  Data             :TQUB_QVO_QVD;
  List             :TQUB_QVO_QVD_SelectionList;
  Name             :String;
  FrontList        :String;
begin
  prUpdateDataLinks;

  Name             := Results.ResultType + ' Results';

  segoff := 0;
  for i := 0 to prDataFiles.Count - 1 do begin
    Data := TQUB_QVO_QVD( prDataFiles.Objects[i] );
    if (Data <> nil) and not Data.LockedReadWrite then begin
      Data.LockWrite;

      FrontList    := Data.CurrListName; // keep current list in front
      List         := Data.GetList(Name);
      List.ResultObject := Self;
      List.Clear;

      for j := 0 to prSegsPerFile[i] - 1 do begin
        if Results.SegColor[ segoff + j ] = Select.ChosenColor then begin
          List.AddSelection( Integer(Results.Segment[ segoff + j ].Data[0]),
                             Integer(Results.Segment[ segoff + j ].Data[1]),
                             IntToStr(j+1) );
        end;
      end;

      if FrontList <> '' then Data.ChooseList(FrontList);

      Data.UnlockWrite;
      Data.RefreshLists;
    end;
    segoff := segoff + prSegsPerFile[i];
  end;
end;
}

procedure TQUB_QVO_QVR.prCreatePopUpMenu;
begin

  inherited;
  PopupMenu                       := nil;

end;


//****************************************************************************//
//                                                                            //
//                                  F I L E                                   //
//                                                                            //
//****************************************************************************//
function TQUB_QVO_QVR.New(const FileName :String) :Boolean;
begin
  MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVR.Open(const FileName :String) :Boolean;
begin
  MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVR.Open(const FileName :String; OpenStream :TStream) :Boolean;
begin
  MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVR.Close :Boolean;
begin
  MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVR.OpenFromString(const MemFile :String) :Boolean;
begin
  MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
end;

function TQUB_QVO_QVR.Save(const FileName :String) :Boolean;
begin
  try
    if (ExtractFileExt(FileName) <> '.qrt') and (FileName = Results.FileName) then begin
      Result := not Results.Save;
    end else begin
      Result := not Results.SaveAs(FileName);
    end;
  except
    Result := False;
  end;
end;

function TQUB_QVO_QVR.CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;
begin
  Result := True;
end;

function TQUB_QVO_QVR.Rename(const Name :String) :Boolean;
begin
  Results.FileName        := Name;
  Result     := False;
end;

function TQUB_QVO_QVR.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  MemFile := Results.Node.AsString;
  Result := False ;
end;

procedure TQUB_QVO_QVR.prSetTipsOn(t :Boolean);
begin
  inherited;
  if tabHist <> nil then panHist.TipsOn := t;
end;

function TQUB_QVO_QVR.prPrint(Dlg :TPrintDialog) :Boolean;
begin
  Result := False; // prPrintPropertiesForm.Print(Self, Dlg, False{CopyImage}, True);
end;


procedure TQUB_QVO_QVR.Activate;
begin
  if Parent = ViewportResults.ConteinerLeft then begin
    ViewportResults.CurrResults := Self; // nil; // inherited Activate calls SetFocus which Activates the CurrResults
    inherited;
    //ViewportResults.CurrResults := Self;
  end else begin
    inherited;
  end;
end;


//****************************************************************************//
//                                                                            //
//                                  D R A W                                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVR.Paint;
begin
        {
  if prDrawing then Exit;

  try

    if ((csDesigning in ComponentState) or prMustRedraw) then begin
      prMustRedraw := False;
      if not prNoDraw then begin
      //This flag is used when setting many properties at once, to prevent drawing each time
        try
          //prDrawCritSection.Acquire;
          //prColor := clRed;
          //prErase;
          //prDrawModel;
        finally
          prDrawCritSection.Release;
        end;
      end;
    end;
  finally

    //inherited;

  end;}
end;

//******************************************************************************
function TQUB_QVO_QVR.prGetFileName :String;
begin
  Result := Results.FileName;
end;

function TQUB_QVO_QVR.prGetFileExt :String;
begin
  Result := ExtractFileExt( FileName );
end;

function TQUB_QVO_QVR.prGetModified :Boolean;
begin
  Result := Results.Modified;
end;

function TQUB_QVO_QVR.prGetSaved :Boolean;
begin
  Result := Results.FileIsOpen;
end;

function TQUB_QVO_QVR.Draw :Boolean;
begin
  Result := inherited Draw;
  Update;
end;

//Lock
procedure TQUB_QVO_QVR.LockWrite;
begin
  inherited;
  //prModelObject.LockWrite;
end;

procedure TQUB_QVO_QVR.UnLockWrite;
begin
  inherited;
  //prModelObject.UnLockWrite;
end;

procedure TQUB_QVO_QVR.LockReadWrite;
begin
  inherited;
  //prModelObject.LockReadWrite;
end;

procedure TQUB_QVO_QVR.UnLockReadWrite;
begin
  inherited;
  //prModelObject.UnLockReadWrite;
end;

//****************************************************************************//
//                                                                            //
//                          M O D E L    E V E N T S                          //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVR.DoOnFileOpen(Sender :TObject);
begin
  Draw;
  //prModelSetHint;
end;

procedure TQUB_QVO_QVR.DoOnFileClose(Sender :TObject);
begin
  Draw;
end;

//****************************************************************************//
//                                                                            //
//                             P R O P E R T I E S                            //
//                                                                            //
//****************************************************************************//
function TQUB_QVO_QVR.Properties :Boolean;
begin
  {prPropertiesForm.OnGetProperties          := prOnGetProperties;
  prPropertiesForm.OnPutProperties          := prOnPutProperties;
  prPropertiesForm.OnResetProperties        := prOnResetProperties;
  prPropertiesForm.OnSetAsDefaultProperties := prOnSetAsDefaultProperties;
  prPropertiesForm.ShowModal;
  Result := prPropertiesForm.ModalResult <> mrOK;    }
  Result := False;
end;
{
procedure TQUB_QVO_QVR.prOnProperties(Sender :TObject);
begin
  {if prPopUpMenu.PopUpComponent = Self then begin
    Properties;
  end else begin
    if (not prModelObject.LockedAny) and (prModelObject.CanEdit) then begin
      State := prModelObject.States.WhichState(TQUB_QFS_StatePanel(prPopUpMenu.PopUpComponent));
      if State <> nil then begin
        prStateProperties(State);
        Exit;
      end;
      Rate := prModelObject.Rates.WhichRate(TQUB_QFS_RatePanel(prPopUpMenu.PopUpComponent));
      if Rate <> nil then begin
        prRateIndexForProperties := Rate.WhichState(TQUB_QFS_RatePanel(prPopUpMenu.PopUpComponent));
        prRateProperties(Rate);
        Exit;
      end;
    end;
  end;
end;
}
procedure TQUB_QVO_QVR.prOnGetProperties(Sender :TObject);
begin
  {with Sender as TQUB_QVO_QVM_PropertiesForm do begin
    try
      Self.Color                           := Color;
      Self.ColorBack                       := ColorBack;
      Self.ColorFrame                      := ColorFrame;
      Self.ColorLine                       := ColorLine;
      Self.ColorRate                       := ColorRate;
      Self.StateSize                       := StateSize;
      Self.LineWidth                       := LineWidth;
      Self.AlignToGrid                     := AlignToGrid;
      Self.MarginH                         := MarginH;
      Self.MarginV                         := MarginV;
      Self.ModelObject.ChannelCount        := ModelChannelCount;
      Self.ModelObject.Constraints.Copy(Constraints);
      prModelSetHint;
    finally
      ResizeModel(prDrawWidth, prDrawHeight);
      Draw;
    end;
  end;//with  }
end;

procedure TQUB_QVO_QVR.prOnPutProperties(Sender :TObject);
begin
  {with Sender as TQUB_QVO_QVM_PropertiesForm do begin
    Color                           := Self.Color;
    ColorBack                       := Self.ColorBack;
    ColorFrame                      := Self.ColorFrame;
    ColorLine                       := Self.ColorLine;
    ColorRate                       := Self.ColorRate;
    StateSize                       := Self.StateSize;
    LineWidth                       := Self.LineWidth;
    AlignToGrid                     := Self.AlignToGrid;
    MarginH                         := Self.MarginH;
    MarginV                         := Self.MarginV;
    Model                           := Self.ModelObject;
    ModelChannelCount               := Self.ModelObject.ChannelCount;
    Constraints.Copy(Self.ModelObject.Constraints);
  end;//with        }
end;

procedure TQUB_QVO_QVR.prOnResetProperties(Sender :TObject);
begin
      {
  with Sender as TQUB_QVO_QVM_PropertiesForm do begin
    try
      prPropertiesRegistry(True // Read);
    finally
      ResizeModel(prDrawWidth, prDrawHeight);
      Draw;
    end;
  end;  }
end;

procedure TQUB_QVO_QVR.prOnSetAsDefaultProperties(Sender :TObject);
begin
  prPropertiesRegistry(False{Write});
end;


//Popup Menu Events ************************************************************

//Edit
//make commands, like in data...


//******************************************************************************
procedure TQUB_QVO_QVR.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
begin

  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Visual Results', True) then begin

        //All the interactive properties must be here **********************

        {//Margins
        if Read then begin
          if ValueExists('Margin Left') then begin
            try MarginLeft := ReadInteger('Margin Left'); except end;
          end else begin
            MarginLeft := QUB_QVO_QVM_DEF_MARGINLEFT;
          end;
        end}
      
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;

procedure TQUB_QVO_QVR.prSetInitialProperties;
const
  QUB_QVO_QVM_DEF_MARGINLEFT                  :Integer                            = 10;
  QUB_QVO_QVM_DEF_MARGINTOP                   :Integer                            = 10;
  QUB_QVO_QVM_DEF_MARGINRIGHT                 :Integer                            = 10;
  QUB_QVO_QVM_DEF_MARGINBOTTOM                :Integer                            = 10;
  QUB_QVO_QVM_DEF_MARGIN                      :Integer                            = 5;
  QUB_QVO_QVM_MIN_MARGIN                      :Integer                            = 0;
  QUB_QVO_QVM_MAX_MARGIN                      :Integer                            = 50;
begin
 //Margins
  prMargin.Left       := QUB_QVO_QVM_DEF_MARGIN;//LEFT;
  prMargin.Top        := QUB_QVO_QVM_DEF_MARGIN;//TOP;
  prMargin.Right      := QUB_QVO_QVM_DEF_MARGIN;//RIGHT;
  prMargin.Bottom     := QUB_QVO_QVM_DEF_MARGIN;//BOTTOM;

  //Colors
  prColor             := 0;
  prColorBack         := 0;
  prColorFrame        := 0;

end;


//******************************************************************************
function TQUB_Python_Results.GetAttr(key : PChar) : PPyObject;
var
  i               :Integer;
  colorCount      :array of Integer;
  color           :Integer;
  pyval           :PPyObject;
  VisVars         :TStringList;
begin
  if Results = nil then begin
    Result := inherited GetAttr(key);
    Exit;
  end;

  with GetPythonEngine do begin
    if key = 'node' then begin
      Result := PyNode_FromNode( Results.Results.Node );
    end else if key = 'fileName' then begin
      Result := PyString_FromString( PChar(Results.FileName) );
    end else if key = 'visibleVars' then begin
      VisVars            := TStringList.Create;
      VisVars.Sorted     := True;
      VisVars.Duplicates := dupIgnore;
      for i := 0 to Results.Select.ChartCount - 1 do begin
        VisVars.Add( Results.Select.Charts[i].SelVar.Name );
      end;
      for i := 0 to Results.Select.Chart2Count - 1 do begin
        VisVars.Add( Results.Select.Charts2[i].XSelVar.Name );
        VisVars.Add( Results.Select.Charts2[i].YSelVar.Name );
      end;
      Result := PyList_New( VisVars.Count );
      for i := 0 to VisVars.Count - 1 do begin
        PyList_SetItem( Result, i, PyString_FromString( PChar(VisVars[i]) ) );
      end;
      VisVars.Free;
    end else if key = 'segmentCount' then begin
      Result := PyInt_FromLong( Results.Results.SegCount );
    end else if key = 'numSD' then begin
      Result := PyFloat_FromDouble( Results.Select.NumSD );
    end else if key = 'chosenColor' then begin
      Result := PyInt_FromLong( Results.Select.ChosenColor );
    end else if key = 'usedColors' then begin
      SetLength(colorCount, Results.Results.ColorCount);
      for i := 0 to Results.Results.ColorCount-1 do colorCount[i] := 0;
      for i := 0 to Results.Results.SegCount - 1 do begin
        color := Results.Results.SegColor[i];
        colorCount[color] := colorCount[color] + 1;
      end;

      Result := PyList_New( 0 );
      for i := 0 to Results.Results.ColorCount-1 do begin
        if colorCount[i] > 0 then begin
          pyval := PyInt_FromLong( i );
          PyList_Append( Result, pyval );
          Py_DECREF( pyval );
        end;
      end;
    end else if key = 'colorCounts' then begin
      SetLength(colorCount, Results.Results.ColorCount);

      for i := 0 to Results.Results.ColorCount-1 do colorCount[i] := 0;
      for i := 0 to Results.Results.SegCount - 1 do begin
        color := Results.Results.SegColor[i];
        colorCount[color] := colorCount[color] + 1;
      end;

      Result := PyList_New( 0 );
      for i := 0 to Results.Results.ColorCount-1 do begin
        pyval := PyInt_FromLong( colorCount[i] );
        PyList_Append( Result, pyval );
        Py_DECREF( pyval );
      end;
    end else if key = 'criteria' then begin
      Result := PyNode_FromNode( Results.Criteria.AsNode );
    end else if key = 'chartCount' then begin
      if Results.Select = nil then begin
        Result := PyInt_FromLong(0);
      end else begin
        Result := PyInt_FromLong( Results.Select.ChartCount + Results.Select.Chart2Count );
      end;
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

function TQUB_Python_Results.MpLength : Integer;
begin
  if Results = nil then Result := 0 else Result := Results.Results.SelVarCount;
end;

function TQUB_Python_Results.MpSubscript( obj : PPyObject) : PPyObject;
var
  SelVar       :TQUB_Results_Var;
begin
  with GetPythonEngine do begin
    if PyString_Check( obj ) then begin
      if Results = nil then begin
        Result := ReturnNone;
      end else begin
        SelVar := Results.Results.SelVarByName[ PyString_AsString( obj ) ];
        if SelVar = nil then begin
          Result := ReturnNone;
        end else begin
          Result := SelVar.GetSelf;
          Py_INCREF( Result );
        end;
      end;
    end else begin
      PyErr_SetString( PyExc_TypeError^, PChar('get results by string') );
      Result := nil;
    end;
  end;
end;

class procedure TQUB_Python_Results.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do begin
    AddMethod( 'keys',            @TQUB_Python_Results.Keys,             'Results.keys()' );
    AddMethod( 'updateSelection', @TQUB_Python_Results.UpdateSelection,  'Results.updateSelection()' );
    AddMethod( 'doUpdates',       @TQUB_Python_Results.DoUpdatesPy,      'Results.doUpdates( t/f )' );
    AddMethod( 'segment',         @TQUB_Python_Results.Segment,          'Results.segment(i)'       );
    AddMethod( 'segmentColor',    @TQUB_Python_Results.SegmentColor,     'Results.segmentColor(i)'  );
    AddMethod( 'setSegmentColor', @TQUB_Python_Results.SetSegmentColor,  'Results.setSegmentColor(i, c)' );
    AddMethod( 'segmentsOfColor', @TQUB_Python_Results.SegmentsOfColor,  'Results.segmentsOfColor(c)' );
    AddMethod( 'copyChartImage',  @TQUB_Python_Results.CopyChartImage,   'Results.copyChartImage(i)' );
  end;
end;

function TQUB_Python_Results.Keys( args :PPyObject ): PPyObject;
var
  i      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if Results = nil then begin
      Result := PyList_New( 0 );
      Exit;
    end;

    Result := PyList_New( Results.Results.SelVarCount );
    for i := 0 to Results.Results.SelVarCount - 1 do begin
      PyList_SetItem( Result, i, PyString_FromString( PChar(Results.Results.SelVar[i].Name) ) );
    end;
  end;
end;

function TQUB_Python_Results.UpdateSelection( args :PPyObject): PPyObject;
begin
  Adjust(@Self);

  if Results <> nil then begin
    PythonThread.SyncGUIProc( prDoUpdate );
  end;
  Result := GetPythonEngine.ReturnNone;
end;

function TQUB_Python_Results.DoUpdatesPy( args :PPyObject ): PPyObject;
var
  obj    :PPyObject;
begin
  Adjust(@Self);

  if Results = nil then begin
    Result := GetPythonEngine.ReturnNone;
    Exit;
  end;

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'O:Results.doUpdates( t/f )', [@obj] ) <> 0 then begin
      Results.Results.DoUpdates := PyObject_IsTrue( obj ) <> 0;
      if Results.Results.DoUpdates then begin
        PythonThread.SyncGUIProc( prDoUpdate );
      end;
      Result := ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Python_Results.Segment( args :PPyObject ): PPyObject;
var
  i      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:Results.segment( i )', [@i] ) <> 0 then begin
      if Results <> nil then begin
        Result := PyNode_FromNode( Results.Results.Segment[i] );
      end else begin
        Result := PyNode_FromNode( TNode.CreateNull );
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Python_Results.SegmentColor( args :PPyObject ): PPyObject;
var
  i      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:Results.segmentColor( i )', [@i] ) <> 0 then begin
      if Results <> nil then begin
        Result := PyInt_FromLong( Results.Results.SegColor[i] );
      end else begin
        Result := PyInt_FromLong( 0 );
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Python_Results.SetSegmentColor( args :PPyObject ): PPyObject;
var
  i, c      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'ii:Results.setSegmentColor(i, c)', [@i, @c] ) <> 0 then begin
      if Results <> nil then begin
        if Results.Results.DoUpdates then begin
          prSegIndex := i;
          prSegColor := c;
          PythonThread.SyncGUIProc( prSetSegColor );
        end else begin
          Results.Results.SegColor[i] := c;
        end;
      end;
      Result := ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Python_Results.SegmentsOfColor( args :PPyObject ): PPyObject;
var
  i, c      :Integer;
  pyval     :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:Results.segmentsOfColor(c)', [@c] ) <> 0 then begin
      Result := PyList_New(0);
      if Results <> nil then begin
        for i := 0 to Results.Results.SegCount - 1 do begin
          if Results.Results.SegColor[i] = c then begin
            pyval := PyInt_FromLong( i );
            PyList_Append( Result, pyval );
            Py_DECREF( pyval );
          end;
        end;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

function TQUB_Python_Results.CopyChartImage( args :PPyObject ): PPyObject;
var
  i      :Integer;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'i:Results.copyChartImage(i)', [@i] ) <> 0 then begin
      Result := ReturnNone;
      prChartIndex := i;
      if Results.Select <> nil then PythonThread.SyncGUIProc(prCopyChartImage);
    end else begin
      Result := nil;
    end;
  end;
end;

procedure TQUB_Python_Results.prCopyChartImage;
begin
  if prChartIndex < Results.Select.ChartCount then begin
    Results.Select.Charts[prChartIndex].CopyToClipboardMetafile(False);
  end else begin
    prChartIndex := prChartIndex - Results.Select.ChartCount;
    if prChartIndex < Results.Select.Chart2Count then begin
      Results.Select.Charts2[prChartIndex].CopyToClipboardMetafile(False);
    end;
  end;
end;

procedure TQUB_Python_Results.prDoUpdate;
begin
  if Results <> nil then Results.Results.UpdateSelection;
end;

procedure TQUB_Python_Results.prSetSegColor;
begin
  if Results = nil then Exit;
  
  Results.Results.SegColor[ prSegIndex ] := prSegColor;
end;

//******************************************************************************

constructor TQUB_Results_SumPanel.Create(AOwner :TComponent; Results :TQUB_Results);
var
  i, j            :Integer;
  Node            :INode;
  VarCount        :Integer;
  MaxCols         :Integer;
  itemCopy        :TMenuItem;
begin
  inherited Create(AOwner);

  prResults          := Results;
  gridVars           := TQUB_Table.Create(Self);
  gridVars.Parent    := Self;
  gridVars.Align     := alClient;

  MaxCols            := 0;
  VarCount           := 0;
  Node               := Results.Node.Child;
  while not Node.IsNull do begin
    if (Node.DataCount > 0) and (Node.DataCount < 200) and (Node.Child.Sibling.IsNull) and (Node.Name <> 'Hold') then begin
      Inc(VarCount);
      if (Node.DataType <> NODE_DATATYPE_STRING) and (LongInt(Node.DataCount) > MaxCols) then begin
        MaxCols := LongInt(Node.DataCount);
      end;
    end;

    Node             := Node.Sibling;
  end;

  gridVars.RowCount  := VarCount;// + 1;
  gridVars.ColCount  := MaxCols + 1;
  gridVars.FixedRows := 0;//Min(gridVars.RowCount - 1, 1);
  gridVars.FixedCols := Min(gridVars.ColCount - 1, 1);

  i                  := 0;
  Node               := Results.Node.Child;
  while not Node.IsNull do begin
    if (Node.DataCount > 0) and (Node.DataCount < 200) and (Node.Child.Sibling.IsNull) and (Node.Name <> 'Hold') then begin
      gridVars.Cells[0, i]     := Node.Name;
      if Node.DataType = NODE_DATATYPE_STRING then begin
        gridVars.Cells[1, i]   := Node.DataAsString;
      end else begin
        for j := 0 to Node.DataCount - 1 do begin
          gridVars.Cells[j+1, i] := Format( '%.7g', [Double(Node.Data[j])] );
        end;
      end;
      Inc(i);
    end;

    Node             := Node.Sibling;
  end;

  Node := Results.Node.Find('DataSet').Find('ProcessData');
  if (not Node.Find('Filter').IsNull) and (0 <> Node['Filter'].Data[0]) then begin
    Inc(VarCount);
    gridVars.RowCount := VarCount;// + 1;
    gridVars.Cells[0, VarCount-1]  := 'Filter (Hz)';
    gridVars.Cells[1, VarCount-1]  := FloatToStr(Node['FilterFreq'].Data[0]);
  end;

  prPopup             := TPopupMenu.Create(Self);
  gridVars.PopupMenu  := prPopup;
  itemCopy            := TMenuItem.Create(prPopup);
  itemCopy.Caption    := 'Copy';
  itemCopy.OnClick    := prOnCopy;
  prPopup.Items.Add( itemCopy );
end;

destructor TQUB_Results_SumPanel.Destroy;
begin
  inherited;
end;

procedure TQUB_Results_SumPanel.prOnCopy(Sender :TObject);
begin
  gridVars.Copy;
end;

constructor TQUB_Results_SegPanel.Create(AOwner :TComponent; Results :TQUB_Results);
var
  i, j            :Integer;
  SelVar          :TQUB_Results_Var;
  itemCopy        :TMenuItem;
begin
  inherited Create(AOwner);

  prResults          := Results;
  gridSegs           := TQUB_Table.Create(Self);
  gridSegs.Parent    := Self;
  gridSegs.Align     := alClient;

  gridSegs.RowCount  := Results.SegCount + 1;
  gridSegs.FixedRows := 0;
  gridSegs.ColCount  := Results.SelVarCount + 2;
  gridSegs.FixedCols := 0;

  gridSegs.Cells[0,0] := 'Segment';
  gridSegs.Cells[1,0] := 'Bounds';

  for j := 0 to Results.SegCount - 1 do begin
    gridSegs.Cells[0,j+1] := IntToStr(j+1);
    gridSegs.Cells[1,j+1] := Format('%d - %d', [Integer(Results.Segment[j].Data[0]), Integer(Results.Segment[j].Data[1])]);
  end;

  for i := 0 to Results.SelVarCount - 1 do begin
    SelVar := Results.SelVar[i];
    gridSegs.Cells[2+i,0] := SelVar.Name;
    if SelVar.YLog then begin
      for j := 0 to Results.SegCount - 1 do begin
        gridSegs.Cells[2+i,j+1] := Format( '%.7g', [Exp(SelVar[j])] );
      end;
    end else begin
      for j := 0 to Results.SegCount - 1 do begin
        gridSegs.Cells[2+i,j+1] := Format( '%.7g', [SelVar[j]] );
      end;
    end;
  end;

  prPopup             := TPopupMenu.Create(Self);
  gridSegs.PopupMenu  := prPopup;
  itemCopy            := TMenuItem.Create(prPopup);
  itemCopy.Caption    := 'Copy';
  itemCopy.OnClick    := prOnCopy;
  prPopup.Items.Add( itemCopy );

end;

destructor TQUB_Results_SegPanel.Destroy;
begin
  inherited;
end;

procedure TQUB_Results_SegPanel.prOnCopy(Sender :TObject);
begin
  gridSegs.Copy;
end;

constructor TQUB_Results_ModelPanel.Create(AOwner :TComponent; Results :TQUB_Results);
begin
  inherited Create(AOwner);

  prResults               := Results;
  prChosen                := -1;

  lstModels               := TListBox.Create(Self);
  lstModels.Parent        := Self;
  lstModels.Align         := alClient;
  lstModels.OnClick       := prOnListClick;
end;

procedure TQUB_Results_ModelPanel.Initialize;
var
  i                       :Integer;
begin
  for i := 0 to prResults.Models.Count - 1 do begin
    lstModels.Items.Add( prResults.Models[i].Node.DataAsString );
  end;
end;

destructor TQUB_Results_ModelPanel.Destroy;
begin
  inherited;
end;

procedure TQUB_Results_ModelPanel.prOnListClick(Sender :TObject);
begin
  if lstModels.Selected[ lstModels.ItemIndex ] then begin
    prChosen := lstModels.ItemIndex;
  end else begin
    prChosen := -1;
  end;

  if (prChosen <> -1) and Assigned(OnModelChosen) then OnModelChosen(Self);
end;

//******************************************************************************

procedure Register;
begin
  RegisterComponents('QUB', [TQUB_QVO_QVR]);
end;


//******************************************************************************
end.


