{

                               Qub Visual Scope
                                     unit

                            Oscilloscope Display




(C) 2000 Lorin Milescu
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


//06-29-2000   Creation date

unit QUB_QVO_QVS;


//******************************************************************************
interface


uses


Windows, SysUtils, Classes, SyncObjs, Messages, Graphics, Controls,
StdCtrls, ExtCtrls, Registry, Menus, Dialogs, Math,

AllocTypes, Alloc,
RegistryInterface,

QUB_Global, QUB_QVO, QUB_QVO_QVD_Global, QUB_QVO_QVS_Properties,
QUB_QFS_QDF_Global;


const


DEF_SHOWMINMAX            = False;
DEF_DRAWALLPOINTS         = False;
DEF_DRAWRMS               = False;

DEF_XRESOLUTION           = 1;
MIN_XRESOLUTION           = 10;
MAX_XRESOLUTION           = 0.001;

DEF_CHARTMODE             = False;

DEF_OFFSET                = 0.0;
MIN_OFFSET                = -10.0;
MAX_OFFSET                = +10.0;

DEF_GAIN                  = 1;//[V/div]
MIN_GAIN                  = 0.001;
MAX_GAIN                  = 1;

MAX_DATACOUNT             = ONE_MEGABYTE;

DEF_DISPLAY               = True;

DEF_FILTERON              = False;

DEF_FILTERPTS             = 5;
MIN_FILTERPTS             = 3;
MAX_FILTERPTS             = 21;


WM_QUB_SCOPE_DRAW_MESSAGE = WM_USER + 100;


type


//Forward class declarations ***************************************************
TQUB_QVO_QVS                                   = class;


TQUB_QVO_QVS                                   = class(TQUB_QVO)
protected


  //Timer
  prTimer                                      :TTimer;

  //Data
  prData                                       :TPAByte;
  prDataChannelCount                           :Integer;
  prDataSize                                   :Integer;
  prDataCount                                  :Integer;
  prDataDraw                                   :TQUB_FirstLast;
  prDataAddFirst                               :Integer;
  prDataDrawCount                              :Integer;

  prPolyDataCount                              :Integer;
  prPolyDataIdx                                :Integer;
  prPolyData                                   :TPAPoint;
  prPolyDataF                                  :TPAInteger;

  //Page settings
  prDataYMin                                   :Integer;
  prDataYMax                                   :Integer;

  prDataLastYAvg                               :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Double;
  prDataLastYStd                               :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Double;

  //Colors
  prColorData                                  :Array[0..7] of TColor;

  prPopUpMenu                                  :TPopUpMenu;

  prRunning                                    :Boolean;
  prCopying                                    :Boolean;

  prRange                                      :Double;
  prBitsPower                                  :Integer;
  prSampling                                   :Double;
  prRecordTime                                 :Integer;
  prDifferential                               :Boolean;

  prShowMinMax                                 :Boolean;
  prDrawAllPoints                              :Boolean;
  prDrawRMS                                    :Boolean;

  prClipped                                    :Boolean;
  prClippedCount                               :Integer;

  prXResolution                                :Double;

  //Not finished. LM
  prChartMode                                  :Boolean;

  prOffset                                     :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Double;
  prGain                                       :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Double;
  prDisplay                                    :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Boolean;
  prScaling                                    :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Double;

  prFilterOn                                   :Boolean;
  prFilterPts                                  :Integer;

  prPropertiesForm                             :TQUB_QVO_QVS_PropertiesForm;

  prGlobalMin                                  :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Integer;
  prGlobalMax                                  :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Integer;
  prBits                                       :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1, 0..15] of LongWord;

  prCopyCritSection                            :TCriticalSection;
  prAddCritSection                             :TCriticalSection;

  prOffsetting                                 :Boolean;
  prOffsettingM                                :Boolean;
  prOffsettingCh                               :Integer;
  prLastMouseY                                 :Integer;

  prChannelIndex                               :Integer;
  prOnChannelChange                            :TNotifyEvent;

  //Clipping Regions etc
  prScopeRgn                                   :HRGN;//The whole frame, except the inside
  prGridRgn                                    :HRGN;//The grid, inside frame
  prTraceRgn                                   :HRGN;//The data, inside grid
  prTraceRect                                  :TRect;
  prTraceWidth                                 :Integer;
  prTraceHeight                                :Integer;

  //Not finished. LM
  prActive                                     :Boolean;


  //Print
  function prPrint(Dlg :TPrintDialog) :Boolean; override;

  procedure prResetData;

  //Registry, properties
  procedure prPropertiesRegistry(Read :Boolean); override;

  //Properties
  procedure prOnProperties(Sender :TObject); override;
  procedure prOnGetProperties(Sender :TObject); override;
  procedure prOnPutProperties(Sender :TObject); override;
  procedure prOnResetProperties(Sender :TObject); override;
  procedure prOnSetAsDefaultProperties(Sender :TObject); override;

  //Draw
  procedure prErase; override;
  procedure prDrawData;
  procedure prDrawChart;

  //Property setters
  procedure prSetColorData(Index :Integer; c :TColor);
  function  prGetColorData(Index :Integer) :TColor;

  procedure prSetShowMinMax(s :Boolean);
  procedure prSetDrawRMS(r :Boolean);
  procedure prSetDrawAllPoints(a :Boolean);

  procedure prSetXResolution(x :Double);

  procedure prSetChartMode(c :Boolean);

  procedure prSetOffset(Index :Integer; o :Double);
  function prGetOffset(Index :Integer) :Double;
  procedure prSetGain(Index :Integer; g :Double);
  function prGetGain(Index :Integer) :Double;
  procedure prSetDisplay(Index :Integer; d :Boolean);
  function prGetDisplay(Index :Integer) :Boolean;
  procedure prSetScaling(Index :Integer; s :Double);
  function prGetScaling(Index :Integer) :Double;
  procedure prSetChannelIndex(c :Integer);

  procedure prSetPopUpMenu(p :TPopUpMenu);
  procedure prCreatePopUpMenu; override;

  function prUnpackData(Channel :Integer; x :Integer) :Integer;

  procedure prOnTimer(Sender :TObject);

  procedure prDrawChannels;
  procedure prDrawBits;
  procedure prDrawOffsets;
  procedure prDrawClipped;
  procedure prDrawGrid;

  procedure prDoClipRegns(Create :Boolean);

  procedure DoOnDrawMessage(var Message: TMessage); message WM_QUB_SCOPE_DRAW_MESSAGE;

  procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

  function New(const FileName :String) :Boolean; override;
  function Open(const FileName :String) :Boolean; override;
  function Open(const FileName :String; OpenStream :TStream) :Boolean; override;
  function Save(const FileName :String) :Boolean; override;
  function Close :Boolean; override;
  function OpenFromString(const MemFile :String) :Boolean; override;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

  procedure StartDisplay(ChannelCount :Integer; Range :Double; Differential :Boolean; Bits :Integer; Sampling :Double; yMin, yMax :Integer);
  procedure StopDisplay;

  procedure AddData(Buffer :Pointer; BufferSize, BufferCount :Integer);
  procedure FindTrace;

  procedure Paint; override;


  property PopUpMenu                           :TPopUpMenu                                                write prSetPopUpMenu;

  property ColorData[Index :Integer]           :TColor                          read prGetColorData       write prSetColorData;

  property ShowMinMax                          :Boolean                         read prShowMinMax         write prSetShowMinMax;
  property DrawAllPoints                       :Boolean                         read prDrawAllPoints      write prSetDrawAllPoints;
  property DrawRMS                             :Boolean                         read prDrawRMS            write prSetDrawRMS;

  property XResolution                         :Double                          read prXResolution        write prSetXResolution;

  property ChartMode                           :Boolean                         read prChartMode          write prSetChartMode;

  property Offset [Index :Integer]             :Double                          read prGetOffset          write prSetOffset;
  property Gain   [Index :Integer]             :Double                          read prGetGain            write prSetGain;
  property Display[Index :Integer]             :Boolean                         read prGetDisplay         write prSetDisplay;
  property Scaling[Index :Integer]             :Double                          read prGetScaling         write prSetScaling;

  property FilterOn                            :Boolean                         read prFilterOn           write prFilterOn;
  property FilterPts                           :Integer                         read prFilterPts          write prFilterPts;

  property PropertiesForm                      :TQUB_QVO_QVS_PropertiesForm     read prPropertiesForm     write prPropertiesForm;

  property OnChannelChange                     :TNotifyEvent                    read prOnChannelChange    write prOnChannelChange;
  property ChannelIndex                        :Integer                         read prChannelIndex       write prSetChannelIndex;
  property ChannelCount                        :Integer                         read prDataChannelCount;

  property Active                              :Boolean                         read prActive             write prActive;


published


  property OnMouseDown;


end;
//..............................................................................



procedure Register;



//******************************************************************************
implementation


//******************************************************************************
//                              TQUB_QVO_QVS
//******************************************************************************
constructor TQUB_QVO_QVS.Create(AOwner :TComponent);
var
  c :Integer;
begin
  inherited Create(AOwner);

  prChannelIndex    := 0;
  for c := Low(prScaling) to High(prScaling) do begin
    prScaling[c]    := 1;
  end;

  prCopyCritSection := TCriticalSection.Create;
  prAddCritSection  := TCriticalSection.Create;

  prPropertiesRegistry(True{Read});

  prMargin.Top      := 15;
  prMargin.Bottom   := 20;

  prCreatePopUpMenu;

  prResetData;

  ShowHint          := True;

  prTimer           := TTimer.Create(Self);
  prTimer.Enabled   := False;
  prTimer.OnTimer   := prOnTimer;

  prDoClipRegns(True{Create});

  //Flags
  prMustRedraw      := True;
end;

destructor TQUB_QVO_QVS.Destroy;
begin
  try

    prTimer.Enabled := False;

    prPropertiesRegistry(False{Read});

    prResetData;

    prAddCritSection.Free;
    prCopyCritSection.Free;

    prDoClipRegns(False{Create});

  finally

    inherited;

  end;
end;

procedure TQUB_QVO_QVS.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin

  inherited;

  prDoClipRegns(False);
  prDoClipRegns(True);

end;

procedure TQUB_QVO_QVS.prDoClipRegns(Create :Boolean);
begin

  //Trace settings .............................................................
  prTraceWidth          := prDrawWidth  - prMargin.Left - prMargin.Right  - 1;
  prTraceHeight         := prDrawHeight - prMargin.Top  - prMargin.Bottom - 1;

  prTraceRect.Left      := prMargin.Left;
  prTraceRect.Right     := prDrawWidth  - prMargin.Right  - 1;
  prTraceRect.Top       := prMargin.Top;
  prTraceRect.Bottom    := prDrawHeight - prMargin.Bottom - 1;
                                                
  if Create then begin
    //Whole frame
    prScopeRgn := CreateRectRgn(0, 0, prDrawWidth + 1, prDrawHeight + 1);
    prGridRgn  := CreateRectRgn(prMargin.Left - 1, prMargin.Top - 1, prDrawWidth - prMargin.Right, prDrawHeight - prMargin.Bottom);
    prTraceRgn := CreateRectRgnIndirect(prTraceRect);
    CombineRgn(prScopeRgn, prScopeRgn, prGridRgn, RGN_XOR);
  end else begin
    DeleteObject(prScopeRgn);
    DeleteObject(prGridRgn);
    DeleteObject(prTraceRgn);
  end;
end;


function TQUB_QVO_QVS.New(const FileName :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVS.Open(const FileName :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVS.Open(const FileName :String; OpenStream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVS.Save(const FileName :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVS.Close :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVS.OpenFromString(const MemFile :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVS.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVS.prPrint(Dlg :TPrintDialog) :Boolean;
begin
  Result := False;
end;


//Mouse
procedure TQUB_QVO_QVS.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  TraceDY             :Integer;
  yMin, yMax          :Integer;
  c                   :Integer;
  ox, oy              :Integer;
  g, s                :Double;
begin

  inherited;

  if prData = nil then Exit;

  TraceDy := prDataYMin - prDataYMax;
  if TraceDy = 0 then TraceDy := 1;

  prLastMouseY := y;
  //It offsets the default channel

  //ChannelLoop
  for c := 0 to prDataChannelCount - 1 do begin

    if not prDisplay[c] then Continue;

    g := (prGain[c] * 10) / prRange;
    s := prScaling[c] * prBitsPower / prRange;

    ox := prTraceRect.Right + 2;
    oy := ((Round(prOffset[c] * s / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;

    if ((ox - 2) < x) and ((ox + 9) > x) and ((oy - 4) < y) and ((oy + 4) > y) then begin

      prOffsettingM       := True;
      prOffsettingCh      := c;
      ChannelIndex        := c;
      Exit;
    end;

  end;

  prOffsetting := True;

  //ChannelLoop
  for c := 0 to prDataChannelCount - 1 do begin

    if not prDisplay[c] then Continue;

    g := (prGain[c] * 10) / prRange;
    s := prScaling[c] * prBitsPower / prRange;

    yMin := (((Round(prOffset[c] * s / g)) * prTraceHeight) div TraceDy) + ((Round((prDataLastYAvg[c] - 2*prDataLastYStd[c]) / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;
    yMax := (((Round(prOffset[c] * s / g)) * prTraceHeight) div TraceDy) + ((Round((prDataLastYAvg[c] + 2*prDataLastYStd[c]) / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;

    if yMin = yMax then begin
      Dec(yMin);
      Inc(yMax);
    end;

    if ((y >= yMin) and (y <= yMax)) or ((y <= yMin) and (y >= yMax)) then begin
      prOffsettingCh      := c;
      ChannelIndex        := c;
      Break;
    end;

  end;//ChannelIndex

end;

procedure TQUB_QVO_QVS.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  TraceDY             :Integer;
  g, s                :Double;
begin
  inherited;

  if prOffsettingM then begin

    if (y < prTraceRect.Top) or (y > prTraceRect.Bottom) then Exit;

    TraceDy    := prDataYMin - prDataYMax;
    if TraceDy = 0 then TraceDy := 1;

    g := (prGain[prOffsettingCh] * 10) / prRange;
    s := prScaling[prOffsettingCh] * prBitsPower / prRange;

    prOffset[prOffsettingCh] := (((y - prTraceRect.Top) * TraceDy / prTraceHeight) + prDataYMax) / s * g;

    prMustRedraw := True;

    if Assigned(prOnChannelChange) then prOnChannelChange(Self);

  end else if prOffsetting then begin

    TraceDy    := prDataYMin - prDataYMax;
    if TraceDy = 0 then TraceDy := 1;

    g := (prGain[prOffsettingCh] * 10) / prRange;
    s := prScaling[prOffsettingCh] * prBitsPower / prRange;

    prOffset[prOffsettingCh] := prOffset[prOffsettingCh] +
                                ((((y            - prTraceRect.Top) * TraceDy / prTraceHeight) + prDataYMax) / s * g) -
                                ((((prLastMouseY - prTraceRect.Top) * TraceDy / prTraceHeight) + prDataYMax) / s * g);

    prLastMouseY := y;

    prMustRedraw := True;

    if Assigned(prOnChannelChange) then prOnChannelChange(Self);

  end;

end;

procedure TQUB_QVO_QVS.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;

  prOffsetting  := False;
  prOffsettingM := False;

end;


//****************************************************************************//
//                                                                            //
//                              D I S P L A Y                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVS.StartDisplay(ChannelCount :Integer; Range :Double; Differential :Boolean; Bits :Integer; Sampling :Double; yMin, yMax :Integer);
begin

  prTimer.Enabled    := False;

  prResetData;

  prDataChannelCount := ChannelCount;
  prDataYmin         := yMin;
  prDataYmax         := yMax;

  prRange            := Range;
  prBitsPower        := Round(IntPower(2, Bits));

  prDifferential     := Differential;

  prDataSize         := ((Bits - 1) div 8) + 1;

  prSampling         := Sampling;

  prDataDrawCount    := Round(XResolution * Sampling);
  prDataCount        := ONE_MEGABYTE;//prDataDrawCount
  if prDataDrawCount > (2 * prDataCount - 1) then prDataDrawCount := 2 * prDataCount - 1;
  prDataDraw.First   := 0;
  prDataDraw.Last    := prDataDraw.First + prDataDrawCount - 1;
  prDataAddFirst     := prDataDraw.First;

  prClipped          := False;
  prClippedCount     := 0;

  //Clear the bits
  FillChar(prBits, (QUB_QFS_QDF_MAX_ADCHANNELCOUNT) * 16 * SizeOf(LongWord), 0);

  try
    prData           := AllocMem(prDataChannelCount * prDataCount * prDataSize);
  except
  end;

  prTimer.Interval   := Round(1000 * XResolution);//Min(100, Round(1000 * XResolution));
  prRunning          := True;
  prTimer.Enabled    := True;

  if prChannelIndex >= prDataChannelCount then prChannelIndex := 0;

  if Assigned(prOnChannelChange) then prOnChannelChange(Self);

end;

procedure TQUB_QVO_QVS.prOnTimer(Sender :TObject);
begin
  //PostMessage(Self.Handle, WM_QUB_SCOPE_DRAW_MESSAGE, 0, 0);
  if prDrawing then Exit;
  try
    Draw;
  except
  end;
end;

procedure TQUB_QVO_QVS.DoOnDrawMessage(var Message: TMessage);
begin
  if prDrawing then Exit;
  try
    Draw;
  except
  end;
end;

procedure TQUB_QVO_QVS.StopDisplay;
begin
  prTimer.Enabled := False;
  prRunning       := False;
  prResetData;
end;

function TQUB_QVO_QVS.prUnpackData(Channel :Integer; x :Integer) :Integer;
var
  RecAddr :Pointer;
begin

  Result := 0;
  try

    RecAddr := @prData[prDataChannelCount * prDataSize * x + Channel * prDataSize];
    case prDataSize * 8 of
       8:
         begin
           if prDifferential then begin
             Result := TPShortInt(RecAddr)^;
           end else begin
             Result := TPByte(RecAddr)^;
           end;
         end;
      16:
         begin
           if prDifferential then begin
             Result := TPSmallint(RecAddr)^;
           end else begin
             Result := TPWord(RecAddr)^;
           end;
         end;
      32:
         begin
           if prDifferential then begin
             Result := TPLongInt(RecAddr)^;
           end else begin
             Result := TPLongWord(RecAddr)^;
           end;
         end;

    else

      Result := 0;

    end;//case DataSize

  except
  end;
end;


//****************************************************************************//
//                                                                            //
//                          F I N D     T R A C E                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVS.FindTrace;
var
  FindRange :TQUB_FirstLast;
  p         :Integer;
  Avg, s    :Double;
begin

  try

    if prData = nil then Exit;

    FindRange := prDataDraw;

    Avg := 0;

    for p := FindRange.First to FindRange.Last do begin
      Avg := Avg + prUnpackData(prChannelIndex, p);
    end;

    s   := prScaling[ChannelIndex] * prBitsPower / prRange;

    Avg := Avg / (FindRange.Last - FindRange.First + 1) / s;

    Offset[ChannelIndex] := -Avg;

    prMustRedraw := True;

    if Assigned(prOnChannelChange) then prOnChannelChange(Self);

  except
  end;
  
end;


//****************************************************************************//
//                                                                            //
//                            A D D     D A T A                               //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVS.AddData(Buffer :Pointer; BufferSize, BufferCount :Integer);
var
  Copied   :Integer;
  Copy     :TQUB_FirstLast;
  AddFirst :Integer;
begin

  if prCopying then Exit;

  try

    prCopyCritSection.Acquire;

    prCopying := True;

    AddFirst  := prDataAddFirst;
    Copied    := 0;

    try

      repeat

        Copy.First := AddFirst;
        Copy.Last  := Copy.First + (BufferCount - Copied) - 1;
        if Copy.Last > (prDataCount - 1) then begin
          Copy.Last := (prDataCount - 1);
        end;

        Move(TPAByte(Buffer)[BufferSize * Copied], prData[BufferSize * Copy.First], BufferSize * (Copy.Last - Copy.First + 1));

        Inc(Copied, Copy.Last - Copy.First + 1);

        AddFirst := Copy.Last + 1;
        if AddFirst > (prDataCount - 1) then begin
          AddFirst := 0;
        end;

      until (Copied >= BufferCount);

    except
    end;

    try
      prAddCritSection.Acquire;
      prDataAddFirst := AddFirst;
    finally
      prAddCritSection.Release;
    end;

  finally

    prMustRedraw := True;
    prCopying    := False;
    prCopyCritSection.Release;

  end;

end;


//****************************************************************************//
//                                                                            //
//                           D R A W   Methods                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVS.Paint;
begin
  if prDrawing then Exit;

  try

    if ((csDesigning in ComponentState) or prMustRedraw) then begin
      prMustRedraw := False;
      if not prNoDraw then begin
      //This flag is used when setting many properties at once, to prevent drawing each time
        try
          prDrawCritSection.Acquire;
          Canvas.Lock;
          prBitmapCanvas.Lock;
          try
            prErase;
            if prData <> nil then begin
              if prDataDraw.First <> prDataDraw.Last then begin
                if prChartMode then begin
                  prDrawChart;
                end else begin
                  prDrawData;
                end;
              end;
            end;
          except
          end;
        finally
          prBitmapCanvas.UnLock;
          Canvas.UnLock;
          prDrawCritSection.Release;
        end;
      end;
    end;

  finally

    inherited;

  end;
end;

procedure TQUB_QVO_QVS.prDrawChart;
begin
  prDrawData;
end;

procedure TQUB_QVO_QVS.prDrawData;
var

  //FilterMem           :Array[0..MAX_FILTERPTS div 2 + 1] of Integer;
  FilterSum           :Integer;
  FilterPts2          :Integer;

  TraceDX             :Integer;
  TraceDY             :Integer;

  ChannelIndex        :Integer;
  PointIndex          :Integer;

  PrvX                :Integer;

  Min                 :Integer;
  Max                 :Integer;

  x                   :Integer;
  y                   :Integer;
  yo                  :Integer;
  p                   :Integer;

  //Scale               :Integer;

  g, s                :Double;

  t                   :String;

  h, w                :Integer;

  RMSSumA             :Double;
  RMSSumS             :Double;
  RMSAvg              :Double;
  RMSStd              :Double;

  ChIdx               :Integer;

  procedure AddBits;
  var
    yy :Word;
  begin
    yy := Abs(y);
    //----- Bits display
    // causes representation % of each bit to the total for all bits.
    // displays brighter red for higher contribution.
    // perhaps this was intended as "if yy and 128 > 0 Then inc(prBits[ChannelIndex,7])" ?
    // as is, does not seem like relevant information... {JB}
    Inc(prBits[ChannelIndex,  0], yy and     1);
    Inc(prBits[ChannelIndex,  1], yy and     2);
    Inc(prBits[ChannelIndex,  2], yy and     4);
    Inc(prBits[ChannelIndex,  3], yy and     8);
    Inc(prBits[ChannelIndex,  4], yy and    16);
    Inc(prBits[ChannelIndex,  5], yy and    32);
    Inc(prBits[ChannelIndex,  6], yy and    64);
    Inc(prBits[ChannelIndex,  7], yy and   128);
    Inc(prBits[ChannelIndex,  8], yy and   256);
    Inc(prBits[ChannelIndex,  9], yy and   512);
    Inc(prBits[ChannelIndex, 10], yy and  1024);
    Inc(prBits[ChannelIndex, 11], yy and  2048);
    Inc(prBits[ChannelIndex, 12], yy and  4096);
    Inc(prBits[ChannelIndex, 13], yy and  8192);
    Inc(prBits[ChannelIndex, 14], yy and 16384);
    Inc(prBits[ChannelIndex, 15], yy and 32768);
  end;

begin

  if prDrawing then Exit;

  prDrawing := True;
  try

    TraceDx             := prDataDrawCount - 1;
    if TraceDx = 0 then Exit;

    if prDrawAllPoints then begin
      if prPolyDataCount < (prDataDrawCount + 100{ForSafety}) then begin
        prPolyDataCount := prDataDrawCount + 100{ForSafety};
        ReAllocMem(prPolyData,  SizeOf(TPoint)  * prPolyDataCount);
        ReAllocMem(prPolyDataF, SizeOf(Integer) * prPolyDataCount);
      end;
    end else begin
      if prPolyDataCount < (prTraceRect.Right - prTraceRect.Left + 1 + 100{ForSafety}) then begin
        prPolyDataCount := (prTraceRect.Right - prTraceRect.Left + 1 + 100{ForSafety});
        ReAllocMem(prPolyData,  SizeOf(TPoint)  * prPolyDataCount);
        ReAllocMem(prPolyDataF, SizeOf(Integer) * prPolyDataCount);
      end;
    end;


    //Select clip region
    SelectClipRgn(prBitmapCanvas.Handle, prTraceRgn);

    //Canvas settings ............................................................
    prBitmapCanvas.Pen.Style  := psSolid;
    prBitmapCanvas.Pen.Width  := 0;
    prBitmapCanvas.Font.Color := prColorBack;


    TraceDy    := prDataYMin - prDataYMax;
    if TraceDy = 0 then TraceDy := 1;

    ChIdx := -1;

    //ChannelLoop
    for ChannelIndex := 0 to prDataChannelCount - 1 do begin

      if not prDisplay[ChannelIndex] then Continue;

      Inc(ChIdx);
      
      RMSSumA := 0;
      RMSSumS := 0;

      //Gain
      g := (prGain[ChannelIndex] * 10) / prRange;
      s := prScaling[ChannelIndex] * prBitsPower / prRange;

      //Offset: V --> pixels
      yo := ((Round(prOffset[ChannelIndex] * s / g)) * prTraceHeight) div TraceDy;

      prBitmapCanvas.Pen.Color := ColorData[ChannelIndex];

      //Make the data array
      prPolyDataIdx := 0;

      //MoveTo TraceXFirst
      x := prTraceRect.Left;
      y := prUnpackData(ChannelIndex, prDataDraw.First);

      //if prDrawRMS then begin
        RMSSumA := RMSSumA + y;
        RMSAvg  := RMSSumA / 1;
        RMSSumS := RMSSumS + (y - RMSAvg)*(y - RMSAvg);
      //end;

      AddBits;

      //Clipped
      if y <= prDataYMin then begin
        prClipped := True;
        Inc(prClippedCount);
      end;
      if y >= prDataYMax then begin
        prClipped := True;
        Inc(prClippedCount);
      end;

      //Transform Y
      y   := yo + ((Round(y / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;
      Min := y;
      Max := y;

      PrvX := x;

      //Add a data point
      prPolyData[prPolyDataIdx].x := x;
      prPolyData[prPolyDataIdx].y := y;
      Inc(prPolyDataIdx);

      for p := 1 to prDataDrawCount - 1 do begin

        PointIndex := prDataDraw.First + p;
        x := (p * prTraceWidth) div (TraceDx) + prTraceRect.Left;
        if PointIndex >= prDataCount then begin
          PointIndex := PointIndex - prDataCount;
        end;

        y := prUnpackData(ChannelIndex, PointIndex);

        //if prDrawRMS then begin
          //Running RMS
          RMSSumA := RMSSumA + y;
          RMSAvg  := RMSSumA / (p + 1);
          RMSSumS := RMSSumS + (y - RMSAvg)*(y - RMSAvg);
        //end;

        AddBits;

        //Clipped
        if y <= prDataYMin then begin
          prClipped := True;
          Inc(prClippedCount);
        end;
        if y >= prDataYMax then begin
          prClipped := True;
          Inc(prClippedCount);
        end;
        if y < prGlobalMin[ChannelIndex] then prGlobalMin[ChannelIndex] := y;
        if y > prGlobalMax[ChannelIndex] then prGlobalMax[ChannelIndex] := y;

        if prShowMinMax then begin

          y := yo + ((Round(y / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;

          if y < Min then Min := y;
          if y > Max then Max := y;
        end;

        if (prDrawAllPoints) or (x <> PrvX) then begin

          if not prShowMinMax then begin
            y  := yo + ((Round(y / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;
          end;
          PrvX := x;

          //Add a data point
          prPolyData[prPolyDataIdx].x := x;
          prPolyData[prPolyDataIdx].y := y;
          Inc(prPolyDataIdx);

        end;

      end;//PointIndex

      RMSStd  := sqrt(RMSSumS / prDataDrawCount);
      prDataLastYAvg[ChannelIndex] := RMSAvg;
      prDataLastYStd[ChannelIndex] := RMSStd;

      //Filter data
      if (prFilterOn) and (FilterPts < prPolyDataIdx) then begin

        FilterPts2 := FilterPts div 2 + 1;

        //Middle
        FilterSum := 0;
        for p := 0 to FilterPts - 2 do begin
          Inc(FilterSum, prPolyData[p].y);
        end;
        for p := FilterPts2 - 1 to prPolyDataIdx - FilterPts2 do begin
          Inc(FilterSum, prPolyData[p + FilterPts2 - 1].y);
          prPolyDataF[p] := FilterSum div FilterPts;
          Dec(FilterSum, prPolyData[p - FilterPts2 + 1].y);
        end;
        //Start
        FilterSum := 0;
        for p := 0 to FilterPts2 - 1 do begin
          Inc(FilterSum, prPolyData[p].y);
        end;
        for p := 0 to FilterPts2 - 1 do begin
          prPolyDataF[p] := FilterSum div (FilterPts2 + p);
          Inc(FilterSum, prPolyData[p + FilterPts2].y);
        end;
        //End
        FilterSum := 0;
        for p := prPolyDataIdx - 1 downto prPolyDataIdx - FilterPts2 do begin
          Inc(FilterSum, prPolyData[p].y);
        end;
        for p := prPolyDataIdx - 1 downto prPolyDataIdx - FilterPts2 do begin
          prPolyDataF[p] := FilterSum div (FilterPts2 + prPolyDataIdx - p - 1);
          Inc(FilterSum, prPolyData[p - FilterPts2].y);
        end;
        for p := 0 to prPolyDataIdx - 1 do begin
          prPolyData[p].y := prPolyDataF[p];
        end;

      end;

      //Draw Data
      Polyline(prBitmapCanvas.Handle, prPolyData[0], prPolyDataIdx);

      if prShowMinMax then begin
        prBitmapCanvas.MoveTo(prTraceRect.Left, Min);
        prBitmapCanvas.LineTo(prTraceRect.Right, Min);
        prBitmapCanvas.MoveTo(prTraceRect.Left, Max);
        prBitmapCanvas.LineTo(prTraceRect.Right, Max);

        y   := yo + ((Round(prGlobalMin[ChannelIndex] / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;
        prBitmapCanvas.MoveTo(prTraceRect.Left, y);
        prBitmapCanvas.LineTo(prTraceRect.Right, y);
        y   := yo + ((Round(prGlobalMax[ChannelIndex] / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;
        prBitmapCanvas.MoveTo(prTraceRect.Left, y);
        prBitmapCanvas.LineTo(prTraceRect.Right, y);
      end;

      //......................................................................

      if prDrawRMS then begin
        prBitmapCanvas.Brush.Style := bsSolid;
        prBitmapCanvas.Brush.Color := {pr}ColorData[ChannelIndex];
        t := Format('%-8.6g', [RMSStd]);
        w := prBitmapCanvas.TextWidth(t);
        h := prBitmapCanvas.TextHeight(t);
        prBitmapCanvas.TextOut(ClientWidth - prMargin.Right - w, prMargin.Top + (h * ChIdx), t);
      end;

    end;//ChannelIndex

    SelectClipRgn(prBitmapCanvas.Handle, 0);

    if prClipped then prDrawClipped;

    prDrawBits;


  finally


    prDataDraw.First := prDataDraw.Last + 1;
    if prDataDraw.First >= prDataCount then begin
      prDataDraw.First := prDataDraw.First - prDataCount;
    end;
    prDataDraw.Last := prDataDraw.First + prDataDrawCount - 1;
    if prDataDraw.Last >= prDataCount then begin
      prDataDraw.Last := prDataDraw.Last - prDataCount;
    end;

    try
      prAddCritSection.Acquire;
      prDataAddFirst := prDataDraw.First;
    finally
      prAddCritSection.Release;
    end;

    prDrawing := False;

  end;

end;

procedure TQUB_QVO_QVS.prErase;
begin

  try

    SelectClipRgn(prBitmapCanvas.Handle, prScopeRgn);

    inherited;

    prDrawBits;
    prDrawOffsets;
    prDrawChannels;

    if prClipped then prDrawClipped;

    SelectClipRgn(prBitmapCanvas.Handle, prGridRgn);

    //if not prChartMode then begin
      prDrawGrid;
    //end;

  finally

    SelectClipRgn(prBitmapCanvas.Handle, 0);

  end;

end;

procedure TQUB_QVO_QVS.prDrawGrid;
var
  dx, dy     :Single;
  i          :Integer;
begin

  prBitmapCanvas.Pen.Mode    := pmCopy;
  prBitmapCanvas.Pen.Width   := 0;
  prBitmapCanvas.Pen.Style   := psSolid;
  prBitmapCanvas.Pen.Color   := prColorFrame;
  prBitmapCanvas.Brush.Style := bsSolid;
  prBitmapCanvas.Brush.Color := prColorBack;

  prBitmapCanvas.Rectangle(prMargin.Left - 1, prMargin.Top - 1, Width - prMargin.Right, Height - prMargin.Bottom);

  dx := (ClientWidth  - prMargin.Left - prMargin.Right  - 2) / 10;
  dy := (ClientHeight - prMargin.Top  - prMargin.Bottom - 2) / 10;

  prBitmapCanvas.Pen.Style := psSolid;

  for i := 2 to 10 do begin

    if i = 6 then begin
      prBitmapCanvas.Pen.Style := psDot;
    end;

    prBitmapCanvas.MoveTo(prMargin.Left,                     prMargin.Top + Round((i - 1)*dy));
    prBitmapCanvas.LineTo(ClientWidth - prMargin.Right - 1,  prMargin.Top + Round((i - 1)*dy));
    prBitmapCanvas.MoveTo(prMargin.Left + Round((i - 1)*dx), prMargin.Top);
    prBitmapCanvas.LineTo(prMargin.Left + Round((i - 1)*dx), ClientHeight - prMargin.Bottom - 1);

    if i = 6 then begin
      prBitmapCanvas.Pen.Style := psSolid;
    end;

  end;

end;

procedure TQUB_QVO_QVS.prDrawChannels;
var
  l, w, h, i, c :Integer;
  Text          :String;
  //bgr : TColor;
begin
  c := 0;
  prBitmapCanvas.Pen.Style   := psSolid;
  prBitmapCanvas.Pen.Mode    := pmNot;
  prBitmapCanvas.Brush.Style := bsSolid;
  prBitmapCanvas.Font.Name   := 'Courier New';
  prBitmapCanvas.Font.Size   := 7;
  l := prMargin.Left;
  h := prBitmapCanvas.TextHeight('No');
  for i := 0 to prDataChannelCount - 1 do begin
    if prDisplay[i] then begin
      Text := 'ch' + IntToStr(i + 1);
      w := prBitmapCanvas.TextWidth(Text);
      prBitmapCanvas.Brush.Color := {pr}ColorData[i];
      prBitmapCanvas.Font.Color  := prColorBack;
      //John, it's not always effective. I use a light green for data, and I get white forground
      //The scope background should always work - presumably the user chose the best contrast. LM
      // change label foreground color to black or white, dependent on background color 03-2004 {JB}
      {
      bgr := not prColorData[i];
      if ((bgr and $FF0000) shr 16)+((bgr and $FF00) shr 8)+(bgr and $FF) < 3 * 128 then begin
        prBitmapCanvas.Font.Color := clWhite;
      end else begin
        prBitmapCanvas.Font.Color := clBlack;
      end;
      }
      prBitmapCanvas.TextOut(l, (prMargin.Top div 2) - (h div 2), Text);
      Inc(c);
      Inc(l, w);
    end;
  end;
  if c = 0 then begin
    Text := 'No channels displayed';
    w := prBitmapCanvas.TextWidth(Text);
    prBitmapCanvas.Brush.Color := clRed;
    prBitmapCanvas.TextOut(ClientWidth - prMargin.Right - w, (prMargin.Top div 2) - (h div 2), Text);
  end;
  prBitmapCanvas.Pen.Mode := pmCopy;
end;

procedure TQUB_QVO_QVS.prDrawClipped;
var
  w, h       :Integer;
  t          :String;
begin
  prBitmapCanvas.Pen.Style   := psSolid;
  prBitmapCanvas.Pen.Mode    := pmNot;
  prBitmapCanvas.Brush.Style := bsSolid;
  prBitmapCanvas.Brush.Color := clRed;
  prBitmapCanvas.Font.Name   := 'Courier New';
  prBitmapCanvas.Font.Size   := 7;
  t := 'Clipped ' + IntToStr(prClippedCount);
  w := prBitmapCanvas.TextWidth(t);
  h := prBitmapCanvas.TextHeight(t);
  prBitmapCanvas.TextOut(ClientWidth - prMargin.Right - w, ClientHeight - (prMargin.Bottom div 2) - (h div 2), t);
  prBitmapCanvas.Pen.Mode    := pmCopy;
end;

procedure TQUB_QVO_QVS.prDrawBits;
var
  Total, Percent   :Single;
  l, t, i, j, w, h :Integer;
begin
  if prDataChannelCount = 0 then Exit;

  prBitmapCanvas.Pen.Style   := psClear;
  prBitmapCanvas.Brush.Style := bsSolid;

  w := 10;
  h := (prMargin.Bottom - 2) div prDataChannelCount;
  t := ClientHeight - prMargin.Bottom + 1;
  for i := 0 to prDataChannelCount - 1 do begin
    Total := 0;
    for j := 0 to 15 do begin
      Total := Total + prBits[i, j];
    end;
    l := prMargin.Left;
    for j := 0 to 15 do begin
      if Total = 0 then Percent := 0 else Percent := prBits[i, j] / Total;
      prBitmapCanvas.Brush.Color := Round(Percent * 16 * 255) and $FF;
      prBitmapCanvas.Rectangle(l, t, l + w, t + h);
      Inc(l, w);
      prBits[i, j] := 0;
    end;
    Inc(t, h);
  end;
end;

//----- Draw small pentagon offset marker to right of scope display for each channel
procedure TQUB_QVO_QVS.prDrawOffsets;
var
  TraceDY             :Integer;
  ChannelIndex        :Integer;
  x                   :Integer;
  y                   :Integer;
  g, s                :Double;
begin

  //Canvas settings ............................................................
  prBitmapCanvas.Pen.Style   := psSolid;
  prBitmapCanvas.Pen.Width   := 0;
  prBitmapCanvas.Brush.Style := bsSolid;

  TraceDy    := prDataYMin - prDataYMax;
  if TraceDy = 0 then TraceDy := 1;

  //ChannelLoop
  for ChannelIndex := 0 to prDataChannelCount - 1 do begin

    if not prDisplay[ChannelIndex] then Continue;

    prBitmapCanvas.Pen.Color   := ColorData[ChannelIndex];
    prBitmapCanvas.Brush.Color := ColorData[ChannelIndex];

    g := (prGain[ChannelIndex] * 10) / prRange;
    s := prScaling[ChannelIndex] * prBitsPower / prRange;

    x := prTraceRect.Right + 2;
    y := ((Round(prOffset[ChannelIndex] * s / g) - prDataYMax) * prTraceHeight) div TraceDy + prTraceRect.Top;

    prBitmapCanvas.Polygon([Point(x - 1, y),
                            Point(x + 2, y - 3),
                            Point(x + 8, y - 3),
                            Point(x + 8, y + 3),
                            Point(x + 2, y + 3)]);

  end;//ChannelIndex

end;

//****************************************************************************//
//                                                                            //
//                           D A T A    R E S E T                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVS.prResetData;
var
  i :Integer;
begin
  FreeMem(prData);
  prData := nil;
  FreeMem(prPolyData);
  prPolyData := nil;
  FreeMem(prPolyDataF);
  prPolyDataF := nil;
  prPolyDataCount := 0;
  for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
    prGlobalMin[i] := 0;
    prGlobalMax[i] := 0;
  end;
end;


//****************************************************************************//
//                                                                            //
//                           P R O P E R T I E S                              //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVS.prSetColorData(Index :Integer; c :TColor);
begin
  prColorData[((Index) mod 8)] := c;
  prMustRedraw := True;
end;

function TQUB_QVO_QVS.prGetColorData(Index :Integer) :TColor;
begin
  Result := prColorData[((Index) mod 8)];
end;

procedure TQUB_QVO_QVS.prSetShowMinMax(s :Boolean);
begin
  prShowMinMax := s;
  prMustRedraw := True;
end;

procedure TQUB_QVO_QVS.prSetDrawRMS(r :Boolean);
begin
  prDrawRMS := r;
  prMustRedraw := True;
end;

procedure TQUB_QVO_QVS.prSetDrawAllPoints(a :Boolean);
begin
  prDrawAllPoints := a;
  prMustRedraw := True;
end;

procedure TQUB_QVO_QVS.prSetXResolution(x :Double);
var
  t :Boolean;
begin
  t := prTimer.Enabled;
  prTimer.Enabled := False;

  while prDrawing do;
  while prCopying do;

  try

    prDrawCritSection.Acquire;
    prCopyCritSection.Acquire;
    prCopying := True;
    prDrawing := True;
    prXResolution    := x;
    prDataDrawCount  := Round(XResolution * prSampling);
    if prDataDrawCount > (2 * prDataCount - 1) then prDataDrawCount := 2 * prDataCount - 1;
    prDataDraw.First := 0;
    prDataDraw.Last  := prDataDraw.First + prDataDrawCount - 1;
    if prDataDraw.Last >= prDataCount then begin
      prDataDraw.Last := prDataCount - prDataDraw.Last;
    end;
    try
      prAddCritSection.Acquire;
      prDataAddFirst := prDataDraw.First;
    finally
      prAddCritSection.Release;
    end;

  finally

    prTimer.Interval := Round(1000 * XResolution);//Min(100, Round(1000 * XResolution));
    prTimer.Enabled  := t;

    prMustRedraw     := True;
    prDrawing := False;
    prCopying := False;

    prCopyCritSection.Release;
    prDrawCritSection.Release;

  end;
end;

//ChartMode
procedure TQUB_QVO_QVS.prSetChartMode(c :Boolean);
var
  t :Boolean;
begin
  t := prTimer.Enabled;
  prTimer.Enabled := False;

  while prDrawing do;
  while prCopying do;

  try

    prDrawCritSection.Acquire;
    prCopyCritSection.Acquire;
    prCopying := True;
    prDrawing := True;
    prChartMode := c;
    prDataDraw.First := 0;
    prDataDraw.Last  := prDataDraw.First + prDataDrawCount - 1;
    if prDataDraw.Last >= prDataCount then begin
      prDataDraw.Last := prDataCount - prDataDraw.Last;
    end;
    try
      prAddCritSection.Acquire;
      prDataAddFirst := prDataDraw.First;
    finally
      prAddCritSection.Release;
    end;

  finally

    prTimer.Enabled  := t;

    prMustRedraw     := True;
    prDrawing := False;
    prCopying := False;

    prCopyCritSection.Release;
    prDrawCritSection.Release;

  end;
end;

procedure TQUB_QVO_QVS.prSetOffset(Index :Integer; o :Double);
begin
  prOffset[Index] := o;
end;

function TQUB_QVO_QVS.prGetOffset(Index :Integer) :Double;
begin
  Result := prOffset[Index];
end;

procedure TQUB_QVO_QVS.prSetGain(Index :Integer; g :Double);
var
  //TraceDy    :Double;
  s          :Double;
  OldO, NewO :Double;
  OldF, NewF :Double;
begin

  //Change the offset in order to keep the trace in the same position on screen

  s          := 0.0;
  OldF       := 0.0;
  OldO       := 0.0;

  if prRunning then begin
  //Find the current position on screen

    //TraceDy := prDataYMin - prDataYMax;
    //if TraceDy = 0 then TraceDy := 1;

    s := prScaling[Index] * prBitsPower / prRange;

    OldO := prOffset[Index];
    OldF := (prGain[Index] * 10) / prRange;

    //Pixel position of DataYAvg + Offset
    {
    y := (prOffset[Index] * s / f * prTraceHeight / TraceDy) +
               (((prDataLastYAvg[Index] / f) - prDataYMax) * prTraceHeight / TraceDy) +
               prTraceRect.Top;
    }
  end;

  //Change the Gain to the new value
  prGain[Index] := g;

  if prRunning then begin
  //Change Offset in order to keep the same position on screen:

    NewF := (prGain[Index] * 10) / prRange;

    //(NewO * s / NewF) + ((prDataLastYAvg[Index] / NewF) - prDataYMax) :=
    //(OldO * s / OldF) + ((prDataLastYAvg[Index] / OldF) - prDataYMax);

    NewO := - ((prDataLastYAvg[Index] / NewF) - prDataYMax) + (OldO * s / OldF) + ((prDataLastYAvg[Index] / OldF) - prDataYMax);
    NewO := NewO * NewF / s;

    prOffset[Index] := NewO;


    //f := (prGain[Index] * 10) / prRange;
    //f := s / f * prTraceHeight / TraceDy;

    //Pixel position of DataYAvg + Offset
    //prOffset[Index] := (y - prTraceRect.Top - (((prDataLastYAvg[Index] / f) - prDataYMax) * prTraceHeight / TraceDy)) / f;

  end;

end;

function TQUB_QVO_QVS.prGetGain(Index :Integer) :Double;
begin
  Result := prGain[Index];
end;

procedure TQUB_QVO_QVS.prSetDisplay(Index :Integer; d :Boolean);
begin
  prDisplay[Index] := d;
end;

function TQUB_QVO_QVS.prGetDisplay(Index :Integer) :Boolean;
begin
  Result := prDisplay[Index];
end;

procedure TQUB_QVO_QVS.prSetScaling(Index :Integer; s :Double);
begin
  prScaling[Index] := s;
end;

function TQUB_QVO_QVS.prGetScaling(Index :Integer) :Double;
begin
  Result := prScaling[Index];
end;

procedure TQUB_QVO_QVS.prSetChannelIndex(c :Integer);
begin
  if c <> prChannelIndex then begin
    if c < ChannelCount then begin
      prChannelIndex := c;
    end;
    if Assigned(prOnChannelChange) then prOnChannelChange(Self);
  end;
end;

//******************************************************************************
procedure TQUB_QVO_QVS.prCreatePopUpMenu;
begin

  inherited;

  prPrintItem.Visible := False;
  prEditCopyImageItem.Visible := False;

end;

procedure TQUB_QVO_QVS.prSetPopUpMenu(p :TPopUpMenu);
begin
  inherited PopUpMenu := p;
end;



//******************************************************************************
procedure TQUB_QVO_QVS.prOnProperties(Sender :TObject);
begin
  prPropertiesForm.OnGetProperties          := prOnGetProperties;
  prPropertiesForm.OnPutProperties          := prOnPutProperties;
  prPropertiesForm.OnResetProperties        := prOnResetProperties;
  prPropertiesForm.OnSetAsDefaultProperties := prOnSetAsDefaultProperties;
  prPropertiesForm.ShowModal;
end;

procedure TQUB_QVO_QVS.prOnGetProperties(Sender :TObject);
var
  i :Integer;
begin
  with Sender as TQUB_QVO_QVS_PropertiesForm do begin
    Self.Color                                   := Color;
    Self.ColorBack                               := ColorBack;
    for i := 0 to 7 do Self.ColorData[i]         := ColorData[i];
    Self.ColorFrame                              := ColorFrame;
    Self.ShowMinMax                              := ShowMinMax;
    Self.DrawRMS                                 := DrawRMS;
    Self.DrawAllPoints                           := DrawAllPoints;
    Self.XResolution                             := XResolution;
    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do Self.Offset[i]  := Offset[i];
    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do Self.Gain[i]    := Gain[i];
    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do Self.Display[i] := Display[i];
    Self.FilterOn                                := FilterOn;
    Self.FilterPts                               := FilterPts;
    prMustRedraw := True;
    Paint;
    prPropertiesRegistry(False{Write});
  end;//with
end;

procedure TQUB_QVO_QVS.prOnPutProperties(Sender :TObject);
var
  i :Integer;
begin
  with Sender as TQUB_QVO_QVS_PropertiesForm do begin
    Color                                   := Self.Color;
    ColorBack                               := Self.ColorBack;
    for i := 0 to 7 do ColorData[i]         := Self.ColorData[i];
    ColorFrame                              := Self.ColorFrame;
    ShowMinMax                              := Self.ShowMinMax;
    DrawRMS                                 := Self.DrawRMS;
    DrawAllPoints                           := Self.DrawAllPoints;
    XResolution                             := Self.XResolution;
    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do Offset[i]  := Self.Offset[i];
    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do Gain[i]    := Self.Gain[i];
    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do Display[i] := Self.Display[i];
    FilterOn                                := Self.FilterOn;
    FilterPts                               := Self.FilterPts;
  end;//with
end;

procedure TQUB_QVO_QVS.prOnResetProperties(Sender :TObject);
begin
  with Sender as TQUB_QVO_QVS_PropertiesForm do begin
    prPropertiesRegistry(True{Read});
  end;//with
end;

procedure TQUB_QVO_QVS.prOnSetAsDefaultProperties(Sender :TObject);
begin
  prPropertiesRegistry(False{Write});
end;


//******************************************************************************
procedure TQUB_QVO_QVS.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
  i   :Integer;
begin

  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Oscilloscope', True) then begin

        //All the interactive properties must be here **********************

        if Read then begin
        
          //Margins
          MarginLeft        := ReadRegistryInteger(Reg, 'Margin Left', QUB_QVO_QVD_DEF_MARGINLEFT);
          MarginTop         := ReadRegistryInteger(Reg, 'Margin Top', QUB_QVO_QVD_DEF_MARGINTOP);
          MarginRight       := ReadRegistryInteger(Reg, 'Margin Right', QUB_QVO_QVD_DEF_MARGINRIGHT);
          MarginBottom      := ReadRegistryInteger(Reg, 'Margin Bottom', QUB_QVO_QVD_DEF_MARGINBOTTOM);

          //Colors
          Color             := ReadRegistryInteger(Reg, 'Color', QUB_QVO_QVD_DEF_COLOR);
          ColorBack         := ReadRegistryInteger(Reg, 'Color Back', QUB_QVO_QVD_DEF_COLORBACK);
          for i := 0 to 7 do begin
            ColorData[i]    := ReadRegistryInteger(Reg, 'Color Data ' + IntToStr(i), QUB_QVO_QVD_DEF_COLORDATA[i]);
          end;
          ColorFrame        := ReadRegistryInteger(Reg, 'Color Border', QUB_QVO_QVD_DEF_COLORFRAME);
          
          //Other
          ShowMinMax        := ReadRegistryBoolean(Reg, 'Show Min Max', DEF_SHOWMINMAX);
          DrawRMS           := ReadRegistryBoolean(Reg, 'Draw RMS', DEF_DRAWRMS);
          DrawAllPoints     := ReadRegistryBoolean(Reg, 'Draw All Points', DEF_DRAWALLPOINTS);
          prXResolution     := StrToFloat(ReadRegistryString(Reg, 'X Resolution', FloatToStr(DEF_XRESOLUTION)));

          for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
            prOffset[i]     := ReadRegistryFloat(Reg, 'OffsetV' + IntToStr(i), DEF_OFFSET);
            prGain[i]       := StrToFloat(ReadRegistryString(Reg, 'Gain' + IntToStr(i), FloatToStr(DEF_GAIN)));
            prDisplay[i]    := ReadRegistryBoolean(Reg, 'Display' + IntToStr(i), DEF_DISPLAY);
          end;

          // These two moved outside of loop - no set side effect, so not a per channel set.
          prFilterOn        := ReadRegistryBoolean(Reg, 'Filter On',DEF_FILTERON);
          prFilterPts       := ReadRegistryInteger(Reg, 'Filter Pts',DEF_FILTERPTS);

        end else begin

          //Margins
          WriteInteger('Margin Left', MarginLeft);
          WriteInteger('Margin Top', MarginTop);
          WriteInteger('Margin Right', MarginRight);
          WriteInteger('Margin Bottom', MarginBottom);
          
          //Colors
          WriteInteger('Color', Color);
          WriteInteger('Color Back', ColorBack);
          for i := 0 to 7 do begin
            WriteInteger('Color Data ' + IntToStr(i), ColorData[i]);
          end;
          WriteInteger('Color Border', ColorFrame);
          
          //Other
          WriteBool('Show Min Max', ShowMinMax);
          WriteBool('Draw RMS', DrawRMS);
          WriteBool('Draw All Points', DrawAllPoints);
          WriteBool('Chart Mode', ChartMode);
          WriteString('X Resolution', FloatToStr(XResolution));
          for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
            //TODO - QUB_QFS_QDF_MAX_ADCHANNELCOUNT is 256 - don't write if not changed (do it in Spectrum too)
            WriteFloat('OffsetV' + IntToStr(i), Offset[i]);
            WriteString('Gain' + IntToStr(i), FloatToStr(Gain[i]));
            WriteBool('Display' + IntToStr(i), Display[i]);
          end;
          WriteBool('Filter On', FilterOn);
          WriteInteger('Filter Pts', FilterPts);

        end;

      end;
       CloseKey;

    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
procedure Register;
begin
  RegisterComponents('QUB', [TQUB_QVO_QVS]);
end;


//******************************************************************************
end.

