{

                             Qub Visual Spectrum
                                     unit

                               Spectrum Display




(C) 2004 Lorin Milescu
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


//12-27-2004   Creation date

unit QUB_QVO_QVSpectrum;


//******************************************************************************
interface


uses


Windows, SysUtils, Classes, SyncObjs, Messages, Graphics, Controls,
StdCtrls, ExtCtrls, Registry, Menus, Dialogs, Math,

AllocTypes, Alloc,
RegistryInterface,

QUB_Global, QUB_QVO, QUB_QVO_QVD_Global, QUB_QVO_QVSpectrum_Properties,
QUB_QFS_QDF_Global,
QUB_Spectrum;


const


MAX_DATACOUNT             = ONE_MEGABYTE;

DEF_DECADECOUNT           = 10;
DEF_DECADEOFFSET          = -5;
DEF_TIME                  = 12;

DEF_BINCOUNT              = 128;
MIN_BINCOUNT              = 10;
MAX_BINCOUNT              = 65536;

DEF_ACC                   = False;
DEF_ACCUPDATE             = True;
DEF_ACCPERCENT            = 0.1;
MIN_ACCPERCENT            = 0.0;
MAX_ACCPERCENT            = 1.0;

WM_QUB_SCOPE_DRAW_MESSAGE = WM_USER + 100;


type


//Forward class declarations ***************************************************
TQUB_QVO_QVSpectrum                            = class;


TQUB_QVO_QVSpectrum                            = class(TQUB_QVO)
protected

  prAddedOnce                                  :Boolean;
  
  //Timer
  prTimer                                      :TTimer;

  //Data
  prData                                       :TPAByte;
  prDataChannelCount                           :Integer;
  prDataSize                                   :Integer;
  prDataCount                                  :Integer;
  prDataDraw                                   :TQUB_FirstLast;
  prDataAddFirst                               :Integer;
  prDataDrawCount                              :Integer;

  prPolyDataCount                              :Integer;
  prPolyDataIdx                                :Integer;
  prPolyData                                   :TPAPoint;

  //Spectrum
  prTime                                       :Integer;
  prBinCount                                   :Integer;
  prDecadeCount                                :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Integer;
  prDecadeOffset                               :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Integer;
  prAcc                                        :Boolean;
  prAccUpdate                                  :Boolean;
  prAccPercent                                 :Double;
  prRe                                         :TPADouble;
  prIm                                         :TPADouble;
  prSpectrum                                   :TPADouble;
  prSpectrumSize                               :Integer;
  prBinX                                       :TPADouble;
  prBinY                                       :TPADouble;
  prBinN                                       :TPAInteger;
  prBinSize                                    :Integer;
  prPassCount                                  :Integer;

  //Page settings
  prDataYMin                                   :Integer;
  prDataYMax                                   :Integer;

  //Colors
  prColorData                                  :Array[0..7] of TColor;

  prPopUpMenu                                  :TPopUpMenu;

  prCopying                                    :Boolean;

  prRange                                      :Double;
  prBitsPower                                  :Integer;
  prSampling                                   :Double;
  prRecordTime                                 :Integer;
  prDifferential                               :Boolean;

  prXResolution                                :Double;

  prGain                                       :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Double;
  prScaling                                    :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of Double;

  prPropertiesForm                             :TQUB_QVO_QVSpectrum_PropertiesForm;

  prCopyCritSection                            :TCriticalSection;
  prAddCritSection                             :TCriticalSection;

  prOffsetting                                 :Boolean;
  prOffsettingCh                               :Integer;

  prChannelIndex                               :Integer;
  prOnChannelChange                            :TNotifyEvent;

  //Clipping Regions etc
  prScopeRgn                                   :HRGN;//The whole frame, except the inside
  prGridRgn                                    :HRGN;//The grid, inside frame
  prTraceRgn                                   :HRGN;//The data, inside grid
  prTraceRect                                  :TRect;
  prTraceWidth                                 :Integer;
  prTraceHeight                                :Integer;

  prActive                                     :Boolean;


  //Print
  function prPrint(Dlg :TPrintDialog) :Boolean; override;

  procedure prResetData;

  //Registry, properties
  procedure prPropertiesRegistry(Read :Boolean); override;

  //Properties
  procedure prOnProperties(Sender :TObject); override;
  procedure prOnGetProperties(Sender :TObject); override;
  procedure prOnPutProperties(Sender :TObject); override;
  procedure prOnResetProperties(Sender :TObject); override;
  procedure prOnSetAsDefaultProperties(Sender :TObject); override;

  //Draw
  procedure prErase; override;
  procedure prDrawData;

  //Property setters
  procedure prSetColorData(Index :Integer; c :TColor);
  function  prGetColorData(Index :Integer) :TColor;

  procedure prSetXResolution(x :Double);

  procedure prSetGain(Index :Integer; g :Double);
  function prGetGain(Index :Integer) :Double;
  procedure prSetScaling(Index :Integer; s :Double);
  function prGetScaling(Index :Integer) :Double;
  procedure prSetChannelIndex(c :Integer);

  //Spectrum
  procedure prSetDecadeCount(Index :Integer; c :Integer);
  function prGetDecadeCount(Index :Integer) :Integer;
  procedure prSetDecadeOffset(Index :Integer; o :Integer);
  function prGetDecadeOffset(Index :Integer) :Integer;
  procedure prSetTime(t :Integer);
  function prGetTime :Integer;
  procedure prSetBinCount(c :Integer);
  function prGetBinCount :Integer;
  procedure prSetAcc(a :Boolean);
  function prGetAcc :Boolean;
  procedure prSetAccUpdate(u :Boolean);
  function prGetAccUpdate :Boolean;
  procedure prSetAccPercent(p :Double);
  function prGetAccPercent :Double;

  procedure prSetPopUpMenu(p :TPopUpMenu);
  procedure prCreatePopUpMenu; override;

  function prUnpackData(Channel :Integer; x :Integer) :Integer;

  procedure prOnTimer(Sender :TObject);

  procedure prDrawDecades;
  procedure prDrawGrid;

  procedure prDoClipRegns(Create :Boolean);

  procedure DoOnDrawMessage(var Message: TMessage); message WM_QUB_SCOPE_DRAW_MESSAGE;

  procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;

  //FFT
  procedure prDoSpectrum;
  procedure prBinSpectrum;
  procedure prAllocSpectrum(Count :Integer);
  procedure prAllocBins(Count :Integer);
  procedure prResetBins;


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;

  function New(const FileName :String) :Boolean; override;
  function Open(const FileName :String) :Boolean; override;
  function Open(const FileName :String; OpenStream :TStream) :Boolean; override;
  function Save(const FileName :String) :Boolean; override;
  function Close :Boolean; override;
  function OpenFromString(const MemFile :String) :Boolean; override;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

  procedure StartDisplay(ChannelCount :Integer; Range :Double; Differential :Boolean; Bits :Integer; Sampling :Double; yMin, yMax :Integer);
  procedure StopDisplay;

  procedure AddData(Buffer :Pointer; BufferSize, BufferCount :Integer);
  procedure FindTrace;

  procedure Paint; override;


  property PopUpMenu                           :TPopUpMenu                                                        write prSetPopUpMenu;

  property ColorData[Index :Integer]           :TColor                                  read prGetColorData       write prSetColorData;

  property XResolution                         :Double                                  read prXResolution        write prSetXResolution;

  property Scaling[Index :Integer]             :Double                                  read prGetScaling         write prSetScaling;

  //Spectrum
  property Time                                :Integer                                 read prGetTime            write prSetTime;
  property BinCount                            :Integer                                 read prGetBinCount        write prSetBinCount;
  property DecadeCount[Index :Integer]         :Integer                                 read prGetDecadeCount     write prSetDecadeCount;
  property DecadeOffset[Index :Integer]        :Integer                                 read prGetDecadeOffset    write prSetDecadeOffset;
  property Acc                                 :Boolean                                 read prGetAcc             write prSetAcc;
  property AccUpdate                           :Boolean                                 read prGetAccUpdate       write prSetAccUpdate;
  property AccPercent                          :Double                                  read prGetAccPercent      write prSetAccPercent;

  property PropertiesForm                      :TQUB_QVO_QVSpectrum_PropertiesForm      read prPropertiesForm     write prPropertiesForm;

  property OnChannelChange                     :TNotifyEvent                            read prOnChannelChange    write prOnChannelChange;
  property ChannelIndex                        :Integer                                 read prChannelIndex       write prSetChannelIndex;
  property ChannelCount                        :Integer                                 read prDataChannelCount;

  property Active                              :Boolean                                 read prActive             write prActive;


published


  property OnMouseDown;


end;
//..............................................................................



procedure Register;



//******************************************************************************
implementation



//******************************************************************************
constructor TQUB_QVO_QVSpectrum.Create(AOwner :TComponent);
var
  c :Integer;
begin
  inherited Create(AOwner);

  prChannelIndex    := 0;
  for c := Low(prScaling) to High(prScaling) do begin
    prScaling[c]    := 1;
  end;

  prCopyCritSection := TCriticalSection.Create;
  prAddCritSection  := TCriticalSection.Create;

  prPropertiesRegistry(True{Read});

  prMargin.Left     := 30;
  prMargin.Right    := 5;
  prMargin.Top      := 5;
  prMargin.Bottom   := 20;

  prCreatePopUpMenu;

  prResetData;

  ShowHint          := True;

  prTimer           := TTimer.Create(Self);
  prTimer.Enabled   := False;
  prTimer.OnTimer   := prOnTimer;

  prDoClipRegns(True{Create});

  //Flags
  prMustRedraw      := True;
end;

destructor TQUB_QVO_QVSpectrum.Destroy;
begin
  try

    prTimer.Enabled := False;

    prPropertiesRegistry(False{Read});

    prResetData;

    prAddCritSection.Free;
    prCopyCritSection.Free;

    prDoClipRegns(False{Create});

    prAllocSpectrum(0);
    prAllocBins(0);

  finally

    inherited;

  end;
end;

procedure TQUB_QVO_QVSpectrum.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin

  inherited;

  prDoClipRegns(False);
  prDoClipRegns(True);

end;

procedure TQUB_QVO_QVSpectrum.prDoClipRegns(Create :Boolean);
begin

  //Trace settings .............................................................
  prTraceWidth          := prDrawWidth  - prMargin.Left - prMargin.Right  - 1;
  prTraceHeight         := prDrawHeight - prMargin.Top  - prMargin.Bottom - 1;

  prTraceRect.Left      := prMargin.Left;
  prTraceRect.Right     := prDrawWidth  - prMargin.Right  - 1;
  prTraceRect.Top       := prMargin.Top;
  prTraceRect.Bottom    := prDrawHeight - prMargin.Bottom - 1;
                                                
  if Create then begin
    //Whole frame
    prScopeRgn := CreateRectRgn(0, 0, prDrawWidth + 1, prDrawHeight + 1);
    prGridRgn  := CreateRectRgn(prMargin.Left - 1, prMargin.Top - 1, prDrawWidth - prMargin.Right, prDrawHeight - prMargin.Bottom);
    prTraceRgn := CreateRectRgnIndirect(prTraceRect);
    CombineRgn(prScopeRgn, prScopeRgn, prGridRgn, RGN_XOR);
  end else begin
    DeleteObject(prScopeRgn);
    DeleteObject(prGridRgn);
    DeleteObject(prTraceRgn);
  end;
end;


function TQUB_QVO_QVSpectrum.New(const FileName :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVSpectrum.Open(const FileName :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVSpectrum.Open(const FileName :String; OpenStream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVSpectrum.Save(const FileName :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVSpectrum.Close :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVSpectrum.OpenFromString(const MemFile :String) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVSpectrum.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  Result := False;
end;

function TQUB_QVO_QVSpectrum.prPrint(Dlg :TPrintDialog) :Boolean;
begin
  Result := False;
end;


//FFT
procedure TQUB_QVO_QVSpectrum.prDoSpectrum;
var
  StopFlag   :Boolean;
  PointIdx   :Integer;
  Scaling    :Double;
  Avg        :Double;
  i          :Integer;
begin

  StopFlag   := False;

  prAllocSpectrum(Round(IntPower(2, prTime)));

  Scaling := prScaling[ChannelIndex] * prBitsPower / prRange;

  //Data --> Re
  for i := 0 to prDataDrawCount - 1 do begin
    PointIdx := prDataDraw.First + i;
    if PointIdx >= prDataCount then begin
      PointIdx := PointIdx - prDataCount;
    end;
    prRe[i + 1] := prUnpackData(ChannelIndex, PointIdx) / Scaling;
  end;//p

  //Clear Im
  for i := 0 to prSpectrumSize - 1 do begin
    prIm[i + 1] := 0;
  end;

  //Calculate and subtract average
  Avg := 0;
  for i := 1 to prSpectrumSize do begin
    Avg := Avg + prRe[i];
  end;
  Avg := Avg / prSpectrumSize;
  for i := 1 to prSpectrumSize do begin
    prRe[i] := prRe[i] - Avg;
  end;

  //Do FFT
  FFT(prRe, prIm, prSpectrumSize, prTime, -1{Forward}, StopFlag);

  //Calculate spectrum
  for i := 0 to prSpectrumSize - 1 do begin
    prSpectrum[i] := prRe[i+1] * prRe[i+1] + prIm[i+1] * prIm[i+1];
  end;

  //Bin spectrum
  prAllocBins(BinCount);
  prBinSpectrum;

end;

procedure TQUB_QVO_QVSpectrum.prBinSpectrum;
var
  ri                 :Integer;
  i                  :Integer;
  r                  :Double;
  f                  :Integer;
  fr1, fr2, frm, frd :Double;
begin

  fr1 := 2 / (prSpectrumSize * prSampling);
  fr2 := (prSpectrumSize div 2) / (prSpectrumSize * prSampling);

  if fr1 = 0 then begin
    fr1 := 1e-20;
  end;
  if fr2 = 0 then begin
    fr2 := 1e-20;
  end;

  //Logarithmic boundaries
  fr1 := log10(fr1);
  fr2 := log10(fr2);
  frd := (fr2 - fr1) / prBinCount;

  //First frame
  f := 2;
  for i := 0 to BinCount - 1 do begin
    r  := 0;
    ri := 0;
    frm := fr1 + i * frd + frd;
    repeat
      if (log10(f / (prSpectrumSize * prSampling)) > frm) then begin
        Break;
      end;
      r := r + 2 * prSpectrum[f-1];
      Inc(ri);
      Inc(f);
      if (f > (prSpectrumSize div 2)) then begin
        Break;
      end;
    until(False);

    prBinX[i] := frm - (frd / 2);

    if ri > 0 then begin

      r := r / ri;
      if r = 0 then begin
        r := 1e-20;
      end;

      //Accumulate?
      if prAcc then begin
        if prAccUpdate then begin
          if prPassCount * (1 - prAccPercent) >= 1 then begin
            prBinY[i] := (1 - prAccPercent)*prBinY[i] + prAccPercent*log10(r);
          end else begin
            prBinY[i] := (prPassCount*prBinY[i] + log10(r)) / (prPassCount + 1);
          end;
        end else begin
          prBinY[i] := (prPassCount*prBinY[i] + log10(r)) / (prPassCount + 1);
        end;
        prBinN[i] := 1;
      end else begin
        prBinY[i] := log10(r);
        prBinN[i] := 1;
      end;

    end else begin

      //Accumulate
      if prAcc then begin

      end else begin
        prBinY[i] := 0.0;
        prBinN[i] := 0;
      end;

    end;
  end;

  Inc(prPassCount);


end;



//Mouse
procedure TQUB_QVO_QVSpectrum.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
{
var
  TraceDY             :Integer;
  ChannelIndex        :Integer;
  ox, oy              :Integer;
  g, s                :Double;
}
begin

  inherited;

  {
  if prData = nil then Exit;

  TraceDy := prDataYMin - prDataYMax;
  if TraceDy = 0 then TraceDy := 1;

  //ChannelLoop
  for ChannelIndex := 0 to prDataChannelCount - 1 do begin

    if not prDisplay[ChannelIndex] then Continue;

    g := (prGain[ChannelIndex] * 10) / prRange;
    s := prScaling[ChannelIndex] * prBitsPower / prRange;

    ox := prTraceRect.Right + 2;
    oy := ((Round(prOffset[ChannelIndex] * s / g) - prDataYMax) * prTraceHeight) div (TraceDy) + prTraceRect.Top;

    if ((ox - 2) < x) and ((ox + 9) > x) and ((oy - 4) < y) and ((oy + 4) > y) then begin

      prOffsetting        := True;
      prOffsettingCh      := ChannelIndex;

      prChannelIndex      := ChannelIndex;

      if Assigned(prOnChannelChange) then prOnChannelChange(Self);

      Break;

    end;

  end;//ChannelIndex
  }

end;

procedure TQUB_QVO_QVSpectrum.MouseMove(Shift: TShiftState; X, Y: Integer);
{
var
  TraceDY             :Integer;
  g, s                :Double;
}
begin
  inherited;

  {
  if prOffsetting then begin

    if (y < prTraceRect.Top) or (y > prTraceRect.Bottom) then Exit;

    TraceDy    := prDataYMin - prDataYMax;
    if TraceDy = 0 then TraceDy := 1;

    g := (prGain[prOffsettingCh] * 10) / prRange;
    s := prScaling[prOffsettingCh] * prBitsPower / prRange;

    prOffset[prOffsettingCh] := (((y - prTraceRect.Top) * TraceDy / prTraceHeight) + prDataYMax) / s * g;

    prMustRedraw := True;

    if Assigned(prOnChannelChange) then prOnChannelChange(Self);

  end;
  }
end;

procedure TQUB_QVO_QVSpectrum.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;

  //prOffsetting := False;

end;


//****************************************************************************//
//                                                                            //
//                              D I S P L A Y                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVSpectrum.StartDisplay(ChannelCount :Integer; Range :Double; Differential :Boolean; Bits :Integer; Sampling :Double; yMin, yMax :Integer);
begin

  prTimer.Enabled    := False;

  prResetData;
  prResetBins;

  prDataChannelCount := ChannelCount;
  prDataYmin         := yMin;
  prDataYmax         := yMax;

  prRange            := Range;
  prBitsPower        := Round(IntPower(2, Bits));

  prDifferential     := Differential;

  prDataSize         := ((Bits - 1) div 8) + 1;

  prSampling         := Sampling;

  prDataDrawCount    := Round(XResolution * Sampling);
  prDataCount        := ONE_MEGABYTE;//prDataDrawCount
  if prDataDrawCount > (2 * prDataCount - 1) then prDataDrawCount := 2 * prDataCount - 1;
  prDataDraw.First   := 0;
  prDataDraw.Last    := prDataDraw.First + prDataDrawCount - 1;
  prDataAddFirst     := prDataDraw.First;

  try
    prData           := AllocMem(prDataChannelCount * prDataCount * prDataSize);
  except
  end;

  prAddedOnce        := False;

  prTimer.Interval   := Round(1000 * XResolution);
  prTimer.Enabled    := True;

  if prChannelIndex >= prDataChannelCount then prChannelIndex := 0;

  if Assigned(prOnChannelChange) then prOnChannelChange(Self);

end;

procedure TQUB_QVO_QVSpectrum.prOnTimer(Sender :TObject);
begin
  if not prAddedOnce then Exit;
  if prDrawing then Exit;
  try
    Draw;
  except
  end;
end;

procedure TQUB_QVO_QVSpectrum.DoOnDrawMessage(var Message: TMessage);
begin
  if prDrawing then Exit;
  try
    Draw;
  except
  end;
end;

procedure TQUB_QVO_QVSpectrum.StopDisplay;
begin
  prTimer.Enabled := False;
  //prResetData;
  //prAllocSpectrum(0);
  //prAllocBins(0);
  //prResetBins;
end;

function TQUB_QVO_QVSpectrum.prUnpackData(Channel :Integer; x :Integer) :Integer;
var
  RecAddr :Pointer;
begin

  Result := 0;
  try

    RecAddr := @prData[prDataChannelCount * prDataSize * x + Channel * prDataSize];
    case prDataSize * 8 of
       8:
         begin
           if prDifferential then begin
             Result := TPShortInt(RecAddr)^;
           end else begin
             Result := TPByte(RecAddr)^;
           end;
         end;
      16:
         begin
           if prDifferential then begin
             Result := TPSmallint(RecAddr)^;
           end else begin
             Result := TPWord(RecAddr)^;
           end;
         end;
      32:
         begin
           if prDifferential then begin
             Result := TPLongInt(RecAddr)^;
           end else begin
             Result := TPLongWord(RecAddr)^;
           end;
         end;

    else

      Result := 0;

    end;//case DataSize

  except
  end;
end;


//****************************************************************************//
//                                                                            //
//                          F I N D     T R A C E                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVSpectrum.FindTrace;
{
var
  FindRange :TQUB_FirstLast;
  p         :Integer;
  Avg, s    :Double;
}
begin

  {
  try

    if prData = nil then Exit;

    FindRange := prDataDraw;

    Avg := 0;

    for p := FindRange.First to FindRange.Last do begin
      Avg := Avg + prUnpackData(prChannelIndex, p);
    end;

    s   := prScaling[ChannelIndex] * prBitsPower / prRange;

    Avg := Avg / (FindRange.Last - FindRange.First + 1) / s;

    Offset[ChannelIndex] := -Avg;

    prMustRedraw := True;

    if Assigned(prOnChannelChange) then prOnChannelChange(Self);

  except
  end;
  }

end;


//****************************************************************************//
//                                                                            //
//                            A D D     D A T A                               //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVSpectrum.AddData(Buffer :Pointer; BufferSize, BufferCount :Integer);
var
  Copied   :Integer;
  Copy     :TQUB_FirstLast;
  AddFirst :Integer;
begin

  if prCopying then Exit;

  try

    prCopyCritSection.Acquire;

    prCopying := True;

    AddFirst  := prDataAddFirst;
    Copied    := 0;

    try

      repeat

        Copy.First := AddFirst;
        Copy.Last  := Copy.First + (BufferCount - Copied) - 1;
        if Copy.Last > (prDataCount - 1) then begin
          Copy.Last := (prDataCount - 1);
        end;

        Move(TPAByte(Buffer)[BufferSize * Copied], prData[BufferSize * Copy.First], BufferSize * (Copy.Last - Copy.First + 1));

        Inc(Copied, Copy.Last - Copy.First + 1);

        AddFirst := Copy.Last + 1;
        if AddFirst > (prDataCount - 1) then begin
          AddFirst := 0;
        end;

      until (Copied >= BufferCount);

    except
    end;

    try
      prAddCritSection.Acquire;
      prDataAddFirst := AddFirst;
    finally
      prAddCritSection.Release;
    end;

  finally

    prMustRedraw := True;
    prCopying    := False;
    prCopyCritSection.Release;

  end;

  prAddedOnce := True;

end;


//****************************************************************************//
//                                                                            //
//                           D R A W   Methods                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVSpectrum.Paint;
begin
  if prDrawing then Exit;

  try

    if ((csDesigning in ComponentState) or prMustRedraw) then begin
      prMustRedraw := False;
      if not prNoDraw then begin
      //This flag is used when setting many properties at once, to prevent drawing each time
        try
          prDrawCritSection.Acquire;
          Canvas.Lock;
          prBitmapCanvas.Lock;
          try
            prErase;
            if prData <> nil then begin
              if prDataDraw.First <> prDataDraw.Last then begin
                prDrawData;
              end;
            end;
          except
          end;
        finally
          prBitmapCanvas.UnLock;
          Canvas.UnLock;
          prDrawCritSection.Release;
        end;
      end;
    end;

  finally

    inherited;

  end;
end;

procedure TQUB_QVO_QVSpectrum.prDrawData;
var

  TraceDX             :Double;
  TraceDY             :Integer;

  x                   :Integer;
  y                   :Integer;
  p, pp               :Integer;

begin

  if prDrawing then Exit;

  prDrawing := True;
  try

    //FFT .......
    prDoSpectrum;
    //...........

    TraceDx := prBinX[prBinSize - 1] - prBinX[0];
    if TraceDx = 0 then Exit;

    TraceDy := prDecadeCount[ChannelIndex];
    if TraceDy = 0 then Exit;


    if prPolyDataCount < prBinCount then begin
      prPolyDataCount := prBinCount;
      ReAllocMem(prPolyData, SizeOf(TPoint) * prPolyDataCount);
    end;

    //Select clip region
    SelectClipRgn(prBitmapCanvas.Handle, prTraceRgn);

    //Canvas settings ............................................................
    prBitmapCanvas.Pen.Style  := psSolid;
    prBitmapCanvas.Pen.Width  := 0;
    prBitmapCanvas.Font.Color := prColorBack;

    prBitmapCanvas.Pen.Color := ColorData[ChannelIndex];

    //Make the data array
    prPolyDataIdx := 0;

    //MoveTo TraceXFirst
    pp := 0;
    for p := 0 to prBinCount - 1 do begin
      if prBinN[p] > 0 then begin
        pp := p;
        Break;
      end;
    end;//PointIndex

    p := pp;
    x := Round((prBinX[p] - prBinX[0])                    * prTraceWidth  / TraceDx) + prTraceRect.Left;
    y := Round((prBinY[p] - prDecadeOffset[ChannelIndex]) * (-prTraceHeight) / TraceDy) + prTraceRect.Bottom;

    //Add a data point
    prPolyData[prPolyDataIdx].x := x;
    prPolyData[prPolyDataIdx].y := y;
    Inc(prPolyDataIdx);

    for p := pp + 1 to prBinCount - 1 do begin

      if prBinN[p] > 0 then begin

        x := Round((prBinX[p] - prBinX[0])                    * prTraceWidth  / TraceDx) + prTraceRect.Left;
        y := Round((prBinY[p] - prDecadeOffset[ChannelIndex]) * (-prTraceHeight) / TraceDy) + prTraceRect.Bottom;

        //Add a data point
        prPolyData[prPolyDataIdx].x := x;
        prPolyData[prPolyDataIdx].y := y;
        Inc(prPolyDataIdx);
      end;

    end;//PointIndex


    //Draw Data
    Polyline(prBitmapCanvas.Handle, prPolyData[0], prPolyDataIdx);

    
    SelectClipRgn(prBitmapCanvas.Handle, 0);


  finally


    prDataDraw.First := prDataDraw.Last + 1;
    if prDataDraw.First >= prDataCount then begin
      prDataDraw.First := prDataDraw.First - prDataCount;
    end;
    prDataDraw.Last := prDataDraw.First + prDataDrawCount - 1;
    if prDataDraw.Last >= prDataCount then begin
      prDataDraw.Last := prDataDraw.Last - prDataCount;
    end;

    try
      prAddCritSection.Acquire;
      prDataAddFirst := prDataDraw.First;
    finally
      prAddCritSection.Release;
    end;

    prDrawing := False;

  end;

end;

procedure TQUB_QVO_QVSpectrum.prErase;
begin

  try

    SelectClipRgn(prBitmapCanvas.Handle, prScopeRgn);

    inherited;

    prDrawDecades;

    SelectClipRgn(prBitmapCanvas.Handle, prGridRgn);

    prDrawGrid;

  finally

    SelectClipRgn(prBitmapCanvas.Handle, 0);

  end;

end;

procedure TQUB_QVO_QVSpectrum.prDrawGrid;
var
  dx, dy       :Double;
  f1, f2, f, r :Double;
  i, j         :Integer;
begin

  prBitmapCanvas.Pen.Mode    := pmCopy;
  prBitmapCanvas.Pen.Width   := 0;
  prBitmapCanvas.Pen.Style   := psSolid;
  prBitmapCanvas.Pen.Color   := prColorFrame;
  prBitmapCanvas.Brush.Style := bsSolid;
  prBitmapCanvas.Brush.Color := prColorBack;

  prBitmapCanvas.Rectangle(prMargin.Left - 1, prMargin.Top - 1, Width - prMargin.Right, Height - prMargin.Bottom);

  if (prSpectrumSize = 0) or (prSampling = 0) then Exit;

  prBitmapCanvas.Pen.Style   := psDot;

  dy := (ClientHeight - prMargin.Top  - prMargin.Bottom - 2) / DecadeCount[ChannelIndex];
  for i := 0 to DecadeCount[ChannelIndex] do begin

    prBitmapCanvas.MoveTo(prMargin.Left,                     prMargin.Top + Round((i)*dy));
    prBitmapCanvas.LineTo(ClientWidth - prMargin.Right - 1,  prMargin.Top + Round((i)*dy));

  end;

  f1 := 2 / (prSpectrumSize / prSampling);
  f2 := (prSpectrumSize div 2) / (prSpectrumSize / prSampling);

  if f1 = f2 then Exit;
  if (f1 = 0) or (f2 = 0) then Exit;

  f1 := log10(f1);
  f2 := log10(f2);

  if f1 = f2 then Exit;

  dx := (ClientWidth - prMargin.Left  - prMargin.Right - 2) / (f2 - f1);
  try
    //for i := Floor(f1) to Ceil(f2) do begin
    //TO DO: LM. Check below, proposed by Chris and John
    for i := Round(f1) to Round(f2) do begin

      prBitmapCanvas.MoveTo(prMargin.Left + Round((i - f1)*dx), prMargin.Top);
      prBitmapCanvas.LineTo(prMargin.Left + Round((i - f1)*dx), ClientHeight - prMargin.Bottom - 1);

      try
        f := IntPower(10, i);
      except
        f := 0;
      end;

      for j := 2 to 9 do begin
        r := log10(f*j);
        prBitmapCanvas.MoveTo(prMargin.Left + Round((r - f1)*dx), prMargin.Top);
        prBitmapCanvas.LineTo(prMargin.Left + Round((r - f1)*dx), ClientHeight - prMargin.Bottom - 1);
      end;

    end;
  except
  end;

end;

procedure TQUB_QVO_QVSpectrum.prDrawDecades;
var
  Text       :String;
  dx, dy     :Double;
  f1, f2, f  :Double;
  h, w       :Integer;
  i          :Integer;
begin

  if (prSpectrumSize = 0) or (prSampling = 0) then Exit;

  prBitmapCanvas.Brush.Style := bsClear;

  prBitmapCanvas.Font.Name   := 'Courier New';
  prBitmapCanvas.Font.Size   := 8;
  prBitmapCanvas.Font.Color  := prColorText;

  dy := -(ClientHeight - prMargin.Top - prMargin.Bottom - 2) / (DecadeCount[ChannelIndex]);
  for i := DecadeOffset[ChannelIndex] to DecadeOffset[ChannelIndex] + DecadeCount[ChannelIndex] do begin

    Text := IntToStr(i);
    h    := prBitmapCanvas.TextHeight(Text);
    w    := prBitmapCanvas.TextWidth(Text);
    prBitmapCanvas.TextOut(prMargin.Left - 2 - w, prTraceRect.Bottom + Round((i - DecadeOffset[ChannelIndex])*dy) - (h div 2), Text);

  end;

  f1 := 2 / (prSpectrumSize / prSampling);
  f2 := (prSpectrumSize div 2) / (prSpectrumSize / prSampling);

  if f1 = f2 then Exit;
  if (f1 = 0) or (f2 = 0) then Exit;

  f1 := log10(f1);
  f2 := log10(f2);

  if f1 = f2 then Exit;

  dx := (ClientWidth - prMargin.Left  - prMargin.Right - 2) / (f2 - f1);
  try//was crashing here - apparntly here, at least. LM

    //for i := Ceil(f1) to Floor(f2) do begin
    //TO DO: LM. Check below, proposed by Chris and John
    for i := Round(f1) to Round(f2) do begin

      try
        f    := IntPower(10, i);
      except
        f := 0;
      end;

      if f >= 1000 then begin
        Text := FloatToStr(f / 1000) + 'kHz';
      end else begin
        Text := FloatToStr(f) + 'Hz';
      end;
      w    := prBitmapCanvas.TextWidth(Text);
      prBitmapCanvas.TextOut(prMargin.Left + Round((i - f1)*dx) - (w div 2), prTraceRect.Bottom + 2, Text);

    end;
  except
  end;
  
end;



//****************************************************************************//
//                                                                            //
//                           D A T A    R E S E T                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVSpectrum.prResetData;
begin
  FreeMem(prData);
  prData := nil;
  FreeMem(prPolyData);
  prPolyData := nil;
  prPolyDataCount := 0;
end;


//****************************************************************************//
//                                                                            //
//                           P R O P E R T I E S                              //
//                                                                            //
//****************************************************************************//
procedure TQUB_QVO_QVSpectrum.prSetColorData(Index :Integer; c :TColor);
begin
  prColorData[((Index) mod 8)] := c;
  prMustRedraw := True;
end;

function TQUB_QVO_QVSpectrum.prGetColorData(Index :Integer) :TColor;
begin
  Result := prColorData[((Index) mod 8)];
end;

procedure TQUB_QVO_QVSpectrum.prSetXResolution(x :Double);
begin
  prXResolution    := x;
end;

procedure TQUB_QVO_QVSpectrum.prSetGain(Index :Integer; g :Double);
begin
  prGain[Index] := g;
end;

function TQUB_QVO_QVSpectrum.prGetGain(Index :Integer) :Double;
begin
  Result := prGain[Index];
end;

procedure TQUB_QVO_QVSpectrum.prSetScaling(Index :Integer; s :Double);
begin
  prScaling[Index] := s;
end;

function TQUB_QVO_QVSpectrum.prGetScaling(Index :Integer) :Double;
begin
  Result := prScaling[Index];
end;

procedure TQUB_QVO_QVSpectrum.prSetChannelIndex(c :Integer);
begin
  if c <> prChannelIndex then begin
    if c < ChannelCount then begin
      prChannelIndex := c;
    end;
    if Assigned(prOnChannelChange) then prOnChannelChange(Self);
  end;
end;

//Spectrum
procedure TQUB_QVO_QVSpectrum.prSetDecadeCount(Index :Integer; c :Integer);
begin
  prDecadeCount[Index] := c;
end;

function TQUB_QVO_QVSpectrum.prGetDecadeCount(Index :Integer) :Integer;
begin
  Result := prDecadeCount[Index];
end;

procedure TQUB_QVO_QVSpectrum.prSetDecadeOffset(Index :Integer; o :Integer);
begin
  prDecadeOffset[Index] := o;
end;

function TQUB_QVO_QVSpectrum.prGetDecadeOffset(Index :Integer) :Integer;
begin
  Result := prDecadeOffset[Index];
end;

procedure TQUB_QVO_QVSpectrum.prSetTime(t :Integer);
var
  e :Boolean;
  c :Integer;
begin
  e := prTimer.Enabled;
  prTimer.Enabled := False;

  while prDrawing do;
  while prCopying do;

  try

    prDrawCritSection.Acquire;
    prCopyCritSection.Acquire;
    prCopying := True;
    prDrawing := True;

    prTime        := t;
    c             := Round(IntPower(2, t));
    prAllocSpectrum(c);
    prResetBins;
    prXResolution := c / prSampling;

    prDataDrawCount  := Round(XResolution * prSampling);
    if prDataDrawCount > (2 * prDataCount - 1) then prDataDrawCount := 2 * prDataCount - 1;
    prDataDraw.First := 0;
    prDataDraw.Last  := prDataDraw.First + prDataDrawCount - 1;
    if prDataDraw.Last >= prDataCount then begin
      prDataDraw.Last := prDataCount - prDataDraw.Last;
    end;
    try
      prAddCritSection.Acquire;
      prDataAddFirst := prDataDraw.First;
    finally
      prAddCritSection.Release;
    end;

  finally

    prTimer.Interval := Round(1000 * XResolution);
    prTimer.Enabled  := e;

    prMustRedraw     := True;
    prDrawing := False;
    prCopying := False;

    prCopyCritSection.Release;
    prDrawCritSection.Release;

  end;

end;

function TQUB_QVO_QVSpectrum.prGetTime :Integer;
begin
  Result := prTime;
end;

procedure TQUB_QVO_QVSpectrum.prSetBinCount(c :Integer);
begin

  while prDrawing do;

  try

    prDrawCritSection.Acquire;
    prDrawing := True;

    prBinCount := c;
    prAllocBins(c);
    prResetBins;

  finally

    prMustRedraw     := True;
    prDrawing        := False;

    prDrawCritSection.Release;

  end;

end;

function TQUB_QVO_QVSpectrum.prGetBinCount :Integer;
begin
  Result := prBinCount;
end;

procedure TQUB_QVO_QVSpectrum.prSetAcc(a :Boolean);
begin

  while prDrawing do;

  try

    prDrawCritSection.Acquire;
    prDrawing := True;

    prAcc := a;
    prResetBins;

  finally

    prMustRedraw     := True;
    prDrawing        := False;

    prDrawCritSection.Release;

  end;

end;

function TQUB_QVO_QVSpectrum.prGetAcc :Boolean;
begin
  Result := prAcc;
end;

procedure TQUB_QVO_QVSpectrum.prSetAccUpdate(u :Boolean);
begin

  while prDrawing do;

  try

    prDrawCritSection.Acquire;
    prDrawing := True;

    prAccUpdate := u;
    prResetBins;

  finally

    prMustRedraw     := True;
    prDrawing        := False;

    prDrawCritSection.Release;

  end;

end;

function TQUB_QVO_QVSpectrum.prGetAccUpdate :Boolean;
begin
  Result := prAccUpdate;
end;

procedure TQUB_QVO_QVSpectrum.prSetAccPercent(p :Double);
begin

  while prDrawing do;

  try

    prDrawCritSection.Acquire;
    prDrawing := True;

    prAccPercent := p;
    //prResetBins;

  finally

    prMustRedraw     := True;
    prDrawing        := False;

    prDrawCritSection.Release;

  end;
end;

function TQUB_QVO_QVSpectrum.prGetAccPercent :Double;
begin
  Result := prAccPercent;
end;

//******************************************************************************
procedure TQUB_QVO_QVSpectrum.prAllocSpectrum(Count :Integer);
begin
  if Count <> prSpectrumSize then begin
    FreeMem(prRe);
    prRe := nil;
    FreeMem(prIm);
    prIm := nil;
    FreeMem(prSpectrum);
    prSpectrum := nil;

    if Count > 0 then begin
      prRe        := DAlloc1(Count + 1);
      prIm        := DAlloc1(Count + 1);
      prSpectrum  := DAlloc1(Count + 1);
    end;
    prSpectrumSize := Count;
  end;
end;

procedure TQUB_QVO_QVSpectrum.prAllocBins(Count :Integer);
begin
  if Count <> prBinSize then begin
    FreeMem(prBinX);
    prBinX := nil;
    FreeMem(prBinY);
    prBinY := nil;
    FreeMem(prBinN);
    prBinN := nil;

    if Count > 0 then begin
      prBinX  := DAlloc1(Count + 1);
      prBinY  := DAlloc1(Count + 1);
      prBinN  := IAlloc1(Count + 1);
    end;
    prBinSize := Count;
  end;
end;

procedure TQUB_QVO_QVSpectrum.prResetBins;
var
  i :Integer;
begin

  for i := 0 to prBinSize - 1 do begin
    prBinY[i]  := 0.0;
    prBinN[i]  := 0;
  end;

  prPassCount  := 0;

end;

//******************************************************************************
procedure TQUB_QVO_QVSpectrum.prCreatePopUpMenu;
begin

  inherited;

  prPrintItem.Visible := False;
  prEditCopyImageItem.Visible := False;

end;

procedure TQUB_QVO_QVSpectrum.prSetPopUpMenu(p :TPopUpMenu);
begin
  inherited PopUpMenu := p;
end;



//******************************************************************************
procedure TQUB_QVO_QVSpectrum.prOnProperties(Sender :TObject);
begin
  prPropertiesForm.OnGetProperties          := prOnGetProperties;
  prPropertiesForm.OnPutProperties          := prOnPutProperties;
  prPropertiesForm.OnResetProperties        := prOnResetProperties;
  prPropertiesForm.OnSetAsDefaultProperties := prOnSetAsDefaultProperties;
  prPropertiesForm.ShowModal;
end;

procedure TQUB_QVO_QVSpectrum.prOnGetProperties(Sender :TObject);
var
  i :Integer;
begin
  with Sender as TQUB_QVO_QVSpectrum_PropertiesForm do begin
    Self.Color                                   := Color;
    Self.ColorBack                               := ColorBack;
    for i := 0 to 7 do Self.ColorData[i]         := ColorData[i];
    Self.ColorFrame                              := ColorFrame;
    Self.ColorText                               := ColorText;
    prMustRedraw := True;
    Paint;
    prPropertiesRegistry(False{Write});
  end;//with
end;

procedure TQUB_QVO_QVSpectrum.prOnPutProperties(Sender :TObject);
var
  i :Integer;
begin
  with Sender as TQUB_QVO_QVSpectrum_PropertiesForm do begin
    Color                                   := Self.Color;
    ColorBack                               := Self.ColorBack;
    for i := 0 to 7 do ColorData[i]         := Self.ColorData[i];
    ColorFrame                              := Self.ColorFrame;
    ColorText                               := Self.ColorText;
  end;//with
end;

procedure TQUB_QVO_QVSpectrum.prOnResetProperties(Sender :TObject);
begin
  with Sender as TQUB_QVO_QVSpectrum_PropertiesForm do begin
    prPropertiesRegistry(True{Read});
  end;//with
end;

procedure TQUB_QVO_QVSpectrum.prOnSetAsDefaultProperties(Sender :TObject);
begin
  prPropertiesRegistry(False{Write});
end;


//******************************************************************************
procedure TQUB_QVO_QVSpectrum.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
  i   :Integer;
begin

  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Spectrum', True) then begin

        //All the interactive properties must be here **********************

        if Read then begin

          //Margins
          MarginLeft          := ReadRegistryInteger(Reg, 'Margin Left', QUB_QVO_QVD_DEF_MARGINLEFT);
          MarginTop           := ReadRegistryInteger(Reg, 'Margin Top', QUB_QVO_QVD_DEF_MARGINTOP);
          MarginRight         := ReadRegistryInteger(Reg, 'Margin Right', QUB_QVO_QVD_DEF_MARGINRIGHT);
          MarginBottom        := ReadRegistryInteger(Reg, 'Margin Bottom', QUB_QVO_QVD_DEF_MARGINBOTTOM);

          //Colors
          Color               := ReadRegistryInteger(Reg, 'Color', clBlack);
          ColorBack           := ReadRegistryInteger(Reg, 'Color Back', 4194304);
          for i := 0 to 7 do begin
            ColorData[i]      := ReadRegistryInteger(Reg, 'Color Data ' + IntToStr(i), clLime);
          end;
          ColorFrame          := ReadRegistryInteger(Reg, 'Color Border', 5460819);
          ColorText           := ReadRegistryInteger(Reg, 'Color Text', clWhite);

          //Other
          prTime              := ReadRegistryInteger(Reg, 'Time', DEF_TIME);
          prBinCount          := ReadRegistryInteger(Reg, 'BinCount', DEF_BINCOUNT);
          prAcc               := ReadRegistryBoolean(Reg, 'Acc', DEF_ACC);
          prAccUpdate         := ReadRegistryBoolean(Reg, 'AccUpdate', DEF_ACCUPDATE);
          prAccPercent        := ReadRegistryFloat(Reg, 'AccPercent', DEF_ACCPERCENT);

          for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
            prDecadeCount[i]  := ReadRegistryInteger(Reg, 'DecadeCount' + IntToStr(i), DEF_DECADECOUNT);
            prDecadeOffset[i] := ReadRegistryInteger(Reg, 'DecadeOffset' + IntToStr(i), DEF_DECADEOFFSET);
          end;

        end else begin

          //Margins
          WriteInteger('Margin Left', MarginLeft);
          WriteInteger('Margin Top', MarginTop);
          WriteInteger('Margin Right', MarginRight);
          WriteInteger('Margin Bottom', MarginBottom);

          //Colors
          WriteInteger('Color', Color);
          WriteInteger('Color Back', ColorBack);
          for i := 0 to 7 do begin
            WriteInteger('Color Data ' + IntToStr(i), ColorData[i]);
          end;
          WriteInteger('Color Border', ColorFrame);
          WriteInteger('Color Text', ColorText);

          //Other
          WriteInteger('Time', Time);
          WriteInteger('BinCount', BinCount);
          WriteBool('Acc', Acc);
          WriteBool('AccUpdate', AccUpdate);
          WriteFloat('AccPercent', AccPercent);

          for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
            WriteInteger('DecadeCount' + IntToStr(i), DecadeCount[i]);
            WriteInteger('DecadeOffset' + IntToStr(i), DecadeOffset[i]);
          end;

        end;

      end;
      CloseKey;

    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
procedure Register;
begin
  RegisterComponents('QUB', [TQUB_QVO_QVSpectrum]);
end;


//******************************************************************************
end.

