{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// base class for model and data properties forms
// reformed to use QUB_Profiles
// you set ProfileCatName in FormCreate
// it uses QUB_Profiles mechanisms to edit a copy of the Active and communicate back changes.

// old-style OnGet/OnPutProperties mechanism retained for now, but maybe not for long...
// the trouble is with e.g. kinetic constraints


unit QUB_QVO_XProperties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls,

QUB_Global, QUB_Node, QUB_Profiles, QUB_Properties;


type


TQUB_QVO_XPropertiesForm = class(TQUB_PropertiesForm)


  btnOK                                   :TButton;
  btnApply                                :TButton;
  btnCancel                               :TButton;
    btnHelp: TButton;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure btnOKClick(Sender: TObject);
  procedure btnApplyClick(Sender: TObject);
  procedure btnCancelClick(Sender: TObject);
  procedure btnHelpClick(Sender: TObject);


protected

  prProfileCatName                        :String;
  prProfileCat                            :TQUB_ProfileCat;
  mnuProfiles                             :TQUB_ProfileEditBox;
  prSetasideProperties                    :INode;
  prSetasideOnSet                         :TQUB_Profile_OnSet;

  prPropertyChanged                       :Boolean;

  prOnGetProperties                       :TNotifyEvent;
  prOnPutProperties                       :TNotifyEvent;
  prWikiTopic                             :String;

  procedure prSetProfileCatName(x :String);
  procedure prOnSetProperties(Sender :TQUB_ProfileCat; Updates :INode);
  function prReadyToClose: Boolean; virtual;

  property ProfileCatName                 :String                read prProfileCatName            write prSetProfileCatName;

public

  property OnGetProperties                :TNotifyEvent          read prOnGetProperties           write prOnGetProperties;
  property OnPutProperties                :TNotifyEvent          read prOnPutProperties           write prOnPutProperties;
  property WikiTopic                      :String                read prWikiTopic                 write prWikiTopic;


end;


implementation

{$R *.DFM}


uses

QUB_GUIInterface;



//******************************************************************************
procedure TQUB_QVO_XPropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  mnuProfiles                             := TQUB_ProfileEditBox.Create(Self);
  mnuProfiles.Parent                      := Self;
  mnuProfiles.SetBounds(ClientWidth - 100 - 10,
                        ClientHeight - mnuProfiles.Height - 10,
                        100, mnuProfiles.Height);
  mnuProfiles.Anchors                     := [akRight, akBottom];
end;

procedure TQUB_QVO_XPropertiesForm.FormShow(Sender: TObject);
begin
  prProperties           := prProfileCat.Active.Clone(True);
  prSetasideProperties   := prProfileCat.Active;
  prSetasideOnSet        := prProfileCat.OnSet;
  prProfileCat.Active    := prProperties;
  prProfileCat.OnSet     := prOnSetProperties;
  prUpdateComponents;

  if Assigned(prOnPutProperties) then begin
    prOnPutProperties(Self);
  end;
  prPropertyChanged := False;
end;

procedure TQUB_QVO_XPropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  prProfileCat.SaveActive;
end;

procedure TQUB_QVO_XPropertiesForm.prSetProfileCatName(x :String);
begin
  prProfileCatName       := x;
  prProfileCat           := ProfileMgr[x];
  mnuProfiles.Cat        := x;
end;

procedure TQUB_QVO_XPropertiesForm.prOnSetProperties(Sender :TQUB_ProfileCat; Updates :INode);
begin
  SetProperties(Updates);
  prUpdateComponents;
end;

procedure TQUB_QVO_XPropertiesForm.btnOKClick(Sender: TObject);
begin
  if not prReadyToClose then ModalResult := mrNone;

  prProfileCat.Active     := prSetasideProperties;
  prProfileCat.OnSet      := prSetasideOnSet;
  prProfileCat.SetProperties(prProperties);

  if prPropertyChanged then begin
    if Assigned(prOnGetProperties) then begin
      prOnGetProperties(Self);
    end;
    prPropertyChanged := False;
  end;
end;

function TQUB_QVO_XPropertiesForm.prReadyToClose: Boolean;
begin
  Result := True;
end;

procedure TQUB_QVO_XPropertiesForm.btnApplyClick(Sender: TObject);
begin
  prProfileCat.Active     := prSetasideProperties;
  prProfileCat.OnSet      := prSetasideOnSet;
  prProfileCat.SetProperties(prProperties);

  if Assigned(prOnGetProperties) then begin
    prOnGetProperties(Self);
  end;
  prPropertyChanged := False;

  prProfileCat.Active    := prProperties;
  prProfileCat.OnSet     := prOnSetProperties;
end;

procedure TQUB_QVO_XPropertiesForm.btnCancelClick(Sender: TObject);
begin
  inherited;

  prProfileCat.Active     := prSetasideProperties;
  prProfileCat.OnSet      := prSetasideOnSet;
end;


//******************************************************************************
procedure TQUB_QVO_XPropertiesForm.btnHelpClick(Sender: TObject);
begin
  //inherited;
  if WikiTopic <> '' then begin
    ShowWiki(WikiTopic, Handle);
  end else if HelpContext <> 0 then begin
    Application.HelpCommand(HELP_CONTEXT, HelpContext);
  end;
end;

end.
