{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// utility functions for QUB_Q_Model and its clients e.g. QUB_Task_MIP


unit QUB_Q_Global;


interface


uses


AllocTypes;


//Simple matrix algebra
function MulVecVecS(V1, V2 :TPADouble; var S :Double; Size :Integer): Integer;
function MulVecEVecESE(V1, V2 :TPAExtended; var S :Extended; Size :Integer): Integer;
function MulVecMatVec(V1 :TPADouble; M1 :TPMDouble; V2 :TPADouble; Size :Integer): Integer;
function MulVecMatVecE(V1 :TPADouble; M1 :TPMDouble; V2 :TPAExtended; Size :Integer): Integer;
function MulVecEMatVecE(V1 :TPAExtended; M1 :TPMDouble; V2 :TPAExtended; Size :Integer): Integer;
function MulMatVecVec(M1 :TPMDouble; V1 :TPADouble; V2 :TPADouble; Size :Integer): Integer;
function MulMatVecEVecE(M1 :TPMDouble; V1 :TPAExtended; V2 :TPAExtended; Size :Integer): Integer;
function MulMatMat(M1, M2, M3 :TPMDouble; Size :Integer): Integer;
function SumMatMat(M1, M2, M3 :TPMDouble; Size :Integer): Integer;
function SumVecVec(V1, V2, V3 :TPADouble; Size :Integer): Integer;
function CopyMat(M1, M2 :TPMDouble; Size :Integer): Integer;
function CopyVec(V1, V2 :TPADouble; Size :Integer): Integer;


implementation


//Some optimizations by John and Chris
//Simple Matrix algebra ********************************************************
function MulVecVecS(V1, V2 :TPADouble; var S :Double; Size :Integer): Integer;
var
  i :Integer;
  x :Double;
begin
  Result := 0;

  x := 0; // S := 0;
  for i := 0 to Size - 1 do begin
    x := {S := S} x + V1[i] * V2[i];
  end;
  S := x;

end;

function MulVecEVecESE(V1, V2 :TPAExtended; var S :Extended; Size :Integer): Integer;
var
  i :Integer;
  x :Extended;
begin
  Result := 0;

  x := 0; // S := 0;
  for i := 0 to Size - 1 do begin
    x := {S := S} x + V1[i] * V2[i];
  end;
  S := x;

end;

function MulVecMatVec(V1 :TPADouble; M1 :TPMDouble; V2 :TPADouble; Size :Integer): Integer;
var
  k, j :Integer;
  x    :Double;
begin
  Result := 0;

  for k := 0 to Size - 1 do begin
    x := 0; // V2[k] := 0;
    for j := 0 to Size - 1 do begin
      x := {V2[k] := V2[k]} x + V1[j]*M1[j, k];
    end;
    V2[k] := x;
  end;

end;

function MulVecMatVecE(V1 :TPADouble; M1 :TPMDouble; V2 :TPAExtended; Size :Integer): Integer;
var
  k, j :Integer;
  x    :Extended;
begin
  Result := 0;

  for k := 0 to Size - 1 do begin
    x := 0; // V2[k] := 0;
    for j := 0 to Size - 1 do begin
      x := {V2[k] := V2[k]} x + V1[j]*M1[j, k];
    end;
    V2[k] := x;
  end;

end;

function MulVecEMatVecE(V1 :TPAExtended; M1 :TPMDouble; V2 :TPAExtended; Size :Integer): Integer;
var
  k, j :Integer;
  x    :Extended;
begin
  Result := 0;

  for k := 0 to Size - 1 do begin
    x := 0; // V2[k] := 0;
    for j := 0 to Size - 1 do begin
      x := {V2[k] := V2[k]} x + V1[j]*M1[j, k];
    end;
    V2[k] := x;
  end;

end;

function MulMatVecVec(M1 :TPMDouble; V1 :TPADouble; V2 :TPADouble; Size :Integer): Integer;
var
  k, j :Integer;
  x    :Double;
begin
  Result := 0;

  for k := 0 to Size - 1 do begin
    x := 0; // V2[k] := 0;
    for j := 0 to Size - 1 do begin
      x := {V2[k] := V2[k]} x + M1[k, j]*V1[j];
    end;
    V2[k] := x;
  end;

end;

function MulMatVecEVecE(M1 :TPMDouble; V1 :TPAExtended; V2 :TPAExtended; Size :Integer): Integer;
var
  k, j :Integer;
  x    :Extended;
begin
  Result := 0;

  for k := 0 to Size - 1 do begin
    x := 0; // V2[k] := 0;
    for j := 0 to Size - 1 do begin
      x := {V2[k] := V2[k]} x + M1[k, j]*V1[j];
    end;
    V2[k] := x;
  end;

end;

function MulMatMat(M1, M2, M3 :TPMDouble; Size :Integer): Integer;
var
  i, j, k :Integer;
  x       :Double;
begin
  Result := 0;

  for i := 0 to Size - 1 do begin
    for j := 0 to Size - 1 do begin
      x := 0.0; // M3[i, j] := 0;
      for k := 0 to Size - 1 do begin
        x := {M3[i, j] := M3[i, j]} x + M1[i, k] * M2[k, j];
      end;
      M3[i, j] := x;
    end;
  end;

end;

//Some tests to speed up matrix computation
{
function MulMatMatFast(M1, M2, M3 :TPMDouble; Size :Integer): Integer;
var
  i, j, k, l :Integer;
  MaxK       :Integer;
  MaxK4      :Integer;
  Size4      :Integer;
  x          :Double;
begin
  Result  := 0;

  MaxK    := Size div 4;
  Size4   := MaxK*4;
  MaxK4   := Size - Size4;
  for i := 0 to Size - 1 do begin
    for j := 0 to Size - 1 do begin
      x := 0.0; // M3[i, j] := 0;
      for k := 0 to MaxK - 1 do begin
        l := k*4;
        x := x + M1[i, l]   * M2[l, j] +
                 M1[i, l+1] * M2[l+1, j] +
                 M1[i, l+2] * M2[l+2, j] +
                 M1[i, l+3] * M2[l+3, j];
      end;
      case MaxK4 of
        1:  x := x + M1[i, Size4] * M2[Size4, j];
        2:  x := x + M1[i, Size4] * M2[Size4, j] + M1[i, Size4+1] * M2[Size4+1, j];
        3:  x := x + M1[i, Size4] * M2[Size4, j] + M1[i, Size4+1] * M2[Size4+1, j] + M1[i, Size4+2] * M2[Size4+2, j];
      end;
      M3[i, j] := x;
    end;
  end;

end;
}

function SumMatMat(M1, M2, M3 :TPMDouble; Size :Integer): Integer;
var
  i, j :Integer;
begin
  Result := 0;

  for i := 0 to Size - 1 do begin
    for j := 0 to Size -1 do begin
      M3[i, j] := M1[i, j] + M2[i, j];
    end;
  end;

end;

function SumVecVec(V1, V2, V3 :TPADouble; Size :Integer): Integer;
var
  i :Integer;
begin
  Result := 0;

  for i := 0 to Size - 1 do begin
    V3[i] := V1[i] + V2[i];
  end;

end;

function CopyMat(M1, M2 :TPMDouble; Size :Integer): Integer;
var
  i, j :Integer;
begin
  Result := 0;

  for i := 0 to Size - 1 do begin
    for j := 0 to Size -1 do begin
      M2[i, j] := M1[i, j];
    end;
  end;

end;

function CopyVec(V1, V2 :TPADouble; Size :Integer): Integer;
var
  i :Integer;
begin
  Result := 0;

  for i := 0 to Size - 1 do begin
    V2[i] := V1[i];
  end;

end;


//******************************************************************************
end.
