{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the common core of matrices and derivatives and stuff for algorithms such as MIP


unit QUB_Q_Model;


interface


uses


SysUtils, Classes,

AllocTypes, Alloc,

qubopt_iface,

QUB_Matrices,

QUB_Q_Global,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix, QUB_QFS_Channels,
QUB_ExpCond_Parameters,
QUB_MIL;


type


TQUB_Q_InitProbs             = (INITPROBS_FIXED, INITPROBS_EQ, INITPROBS_CONDEQ);


function Q_InitProbsStringOf(x :TQUB_Q_InitProbs): String;
function Q_InitProbsOfString(x :String): TQUB_Q_InitProbs;


type


TQUB_Q_Model                 = class
protected


  prModel                    :TQUB_QFS_QMF;
  prConstraints              :TQUB_QFS_Constraints;
  prExpCondParams            :TQUB_ExpCond_ParametersList;//Which is a list of TQUB_ExpCond_Parameters, one for each exp. condition
  prMilModel                 :TQUB_MIL_PMODEL;


  procedure prAllocExpCondStuff; virtual;
  procedure prInitExpCondStuff; virtual;
  procedure prFreeExpCondStuff; virtual;

  procedure prAllocModelStuff; virtual;
  procedure prInitModelStuff; virtual;
  procedure prFreeModelStuff; virtual;

  procedure prAllocOptStuff; virtual;
  procedure prInitOptStuff; virtual;
  procedure prFreeOptStuff; virtual;

  function prRateIdxOf(idx1, idx2, k01 :Integer) :Integer;

  function prGet_cParCount :Integer; virtual;
  function prGet_ChannelCount :Integer; virtual;
  function prGet_AverageCount :Integer; virtual;
  function prGet_FitWeight :Double; virtual;

  function prGetExpCondParams(ec :Integer): TQUB_ExpCond_Parameters;


public


  ChannelCount               :Double;     //Model channel count
  AverageCount               :Integer;    //How many traces were averaged into Data
  FitWeight                  :Double;     //Multiply the LL of this model by FitWeight - useful for multiple files
  StateCount                 :Integer;    //State count
  RateCount                  :Integer;    //Rate count (2 * ConnectionCount)
  ExtraKinParCount           :Integer;    //The number of extra kinetic parameters
  ClassCount                 :Integer;    //Class count
  ConstCount                 :Integer;    //Constraints count


  QMinRate                   :Double;     //Minimum value a rate is allowed to take
  AllowNegRates              :Boolean;

  RateIdx                    :TPAPoint;   //Indexes of the connected rates for each q



  //Q matrix and its spectral decomposition ....................................
  Q                          :TPAPointer; //Q matrix
  Ak                         :TPAPointer; //Spectral matrices of Q
  Lambda                     :TPMDouble;  //Eigenvalues of Q

  //Q matrix derivatives
  dQdq                       :TPAPointer; //The derivatives of Q w.r.t. each q

  //Conditioning Q
  Cond_Q                     :TPMDouble;



  //A matrices .................................................................
  A                          :TPAPointer; //A matrix for dt                           ////
  At                         :TPAPointer; //A matrix at time t

  //A matrices derivatives
  dAdq                       :TPAPointer; //The derivative of A w.r.t. each q, for each ExpCond
  dAtdq                      :TPAPointer; //The derivative of At w.r.t. each q, for each ExpCond



  //State probabilities ........................................................
  P0                         :TPADouble;  //Initial state probabilities
  Pt                         :TPADouble;  //At time t
  Peq                        :TPADouble;  //Equilibrium
  Cond_Peq                   :TPADouble;  //Conditioning equilibrium probabilities

  //State probabilities derivatives (for each experimental condition)
  dP0dq                      :TPAPointer; //The derivative of P0 w.r.t. each q, for each ExpCond
  dP0dCond_q                 :TPMDouble;  //The derivative of P0 w.r.t. each Cond_q
  dPtdq                      :TPAPointer; //The derivative of Pt w.r.t. each q, for each ExpCond
  dPtdCond_q                 :TPMDouble;  //The derivative of Pt w.r.t. each Cond_q
  dPeqdq                     :TPAPointer; //The derivative of Peq w.r.t. each q, for each ExpCond
  dCond_PeqdCond_q           :TPMDouble;  //The derivative of dCond_Peq w.r.t. each Cond_q



  //Class info
  Cls                        :TPAInteger; //Class index of each state
  Classes                    :TPAInteger;



  //Parameters .................................................................

  //Free parameters ............................................................
  fPars                      :TPADouble;  //Free parameters, passed to the optimizer
  fParsVar                   :TPMDouble;  //Free parameters covariance
  fParCount                  :Integer;    //Free parameter count

  //Extra Kinetic Parameters ...................................................
  ExtraKinPars               :TPADouble;
  ExtraKinParTypes           :TPAInteger;

  //Constrained parameters .....................................................
  cPars                      :TPADouble;  //Constrained parameters (K0's and K1's)
  cParsVar                   :TPADouble;  //Constrained parameters variance
  cParCount                  :Integer;    //Constrained parameter count

  //Mapping ....................................................................
  f2cA                       :TPMDouble;  //Mapping between free and constrained parameters
  f2cB                       :TPADouble;  //  cPars = f2cA.fPars + f2cB

  c2qS                       :TPAPoint;   //Mapping between the constrained parameters and the Q matrix
                                          //  --> the two states i, j connected by cPars[k]
  //Constraints ................................................................
  ConstA                     :TPMDouble;  //ConstA x cPars = ConstB
  ConstB                     :TPADouble;


  //Experimental conditions ....................................................
  ExpCondCount               :Integer;    //Number of sets of conditions                   ////
  ExpCondP                   :TPMDouble;  //"Concentrations"; they multiply the K0's       ////
  ExpCondQ                   :TPMDouble;  //"Voltages"; they multiply the K1's             ////

  //Conditioning ...............................................................
  Cond_ExpCondP              :TPADouble;  //"Concentrations"; they multiply the K0's
  Cond_ExpCondQ              :TPADouble;  //"Voltages"; they multiply the K1's


  //Optimizer ..................................................................
  Error                      :Integer;
  Iterations                 :Integer;
  FuncCalls                  :Integer;
  GradCalls                  :Integer;
  LL                         :Double;
  SS                         :Double;     //Sum of squares
  PointCount                 :Integer;



  //Log-Likelihood and its Derivatives
  dLLdq                      :TPMDouble;  //Derivative of LL w.r.t. each q                     ////
  dLLdCond_q                 :TPADouble;  //Derivative of LL w.r.t. each conditioning q
  dLLdfPar                   :TPADouble;  //Derivative of LL w.r.t each free parameter
  dqdcPar                    :TPADouble;  //Derivative of each q w.r.t. to its K0 and K1       //// (display only)
  dCond_qdcPar               :TPADouble;  //Derivative of each Cond q w.r.t. to its K0 and K1
  dLLdcPar                   :TPADouble;  //Derivative of LL w.r.t. each constrained parameter



  //Temporary quantities .......................................................
  Vec1                       :TPADouble;  //used for calculations
  Vec2                       :TPADouble;
  Mat1                       :TPMDouble;
  Mat2                       :TPMDouble;


  //Confluent Eigenvalues
  Confluent                  :TPAInteger;
  SmallRateWarning           :Boolean;
  BigRateWarning             :Boolean;

  //Errors and warnings about constraints (valid after init)
  ConstraintMsg              :String;

  constructor Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ECParams :TQUB_ExpCond_ParametersList);
  destructor Destroy; override;


  procedure Initialize;

  function ResetLLandGR :Boolean; virtual;

  function ErrorA :Boolean;

  function IsPZero(Rate, Index :Integer) :Boolean; virtual;
  function IsQZero(Rate, Index :Integer) :Boolean; virtual;

  // function Do_ExpCond :Integer; virtual;
  function Do_AddConstraints :Integer; virtual;
  function Do_CheckConstraints: Boolean; virtual;

  function Do_fParsToModel :Integer; virtual;
  function Do_fParsTocPars :Integer; virtual;
  function Do_cParsToQ :Integer; virtual;
  function Do_cParsToEKP :Integer; virtual;

  function Do_InitfPars :Integer;
  function Do_InitcPars :Integer; virtual;
  function Do_InitExtraKinPars :Integer; virtual;

  function Do_ConstAB :Integer;
  function Do_f2cAB :Integer;
  function Do_c2s :Integer;


  //Q matrix
  function Do_dQ_dq :Integer;

  //A matrix
  function Do_A(dt :Double) :Integer; overload; virtual;
  function Do_A(dt :Double; ExpCond :Integer) :Integer; overload; virtual;
  function Do_At(t :Double) :Integer; overload; virtual;
  function Do_At(t :Double; ExpCond :Integer) :Integer; overload; virtual;

  function Do_dA_dq(dt :Double) :Integer; overload; virtual;
  function Do_dA_dq(dt :Double; ExpCond :Integer) :Integer; overload; virtual;
  function Do_dAt_dq(t :Double) :Integer; overload; virtual;
  function Do_dAt_dq(t :Double; ExpCond :Integer) :Integer; overload; virtual;

  //P vector
  function Do_P0 :Integer; virtual;
  function Do_Pt(t :Double; ExpCond :Integer) :Integer; virtual;
  function Do_Peq :Integer; virtual;
  function Do_Cond_Peq :Integer; virtual;

  function Do_dP0_dq :Integer; virtual;
  function Do_dP0_dCond_q :Integer; virtual;
  function Do_dPt_dq(ExpCond :Integer) :Integer; virtual;
  function Do_dPt_dCond_q(ExpCond :Integer) :Integer; virtual;
  function Do_dPeq_dq :Integer; virtual;
  function Do_dCond_Peq_dCond_q :Integer; virtual;

  procedure PtToP0;
  procedure dPtdqTodP0dq;
  procedure dPtdCond_qTodP0dCond_q;


  //Class Info
  function Do_ClassCount :Integer;

  function Do_G_kl_t(k, l, ec :Integer; t :Double; var gklt :Double) :Integer;

  function Do_dLLdfPar :Integer; virtual;
  function Do_Varf2Varc :Integer; virtual;


  property ExpCondParams[ec :Integer]        :TQUB_ExpCond_Parameters  read prGetExpCondParams;
  property Model                             :TQUB_QFS_QMF             read prModel;
  

end;


implementation


uses


Math,

QUB_Node,
QUB_Console, QUB_Rich_Console;


//******************************************************************************
function Q_InitProbsStringOf(x :TQUB_Q_InitProbs): String;
begin
  if          x = INITPROBS_FIXED  then begin
    Result := 'Fixed';
  end else if x = INITPROBS_EQ     then begin
    Result := 'Equilibrium';
  end else if x = INITPROBS_CONDEQ then begin
    Result := 'Conditioning Eq';
  end;
end;

function Q_InitProbsOfString(x :String): TQUB_Q_InitProbs;
begin
  if          x = 'Fixed'           then begin
    Result := INITPROBS_FIXED;
  end else if x = 'Equilibrium'     then begin
    Result := INITPROBS_EQ;
  end else{if x = 'Conditioning Eq' then}begin
    Result := INITPROBS_CONDEQ;
  end;
end;


//******************************************************************************
constructor TQUB_Q_Model.Create(Model :TQUB_QFS_QMF; DoInitialize :Boolean; ECParams :TQUB_ExpCond_ParametersList);
var
  ModelNode                  :INode;
begin

  inherited Create;


  prModel                    := Model;
  prExpCondParams            := ECParams;


  //Alloc local constraints
  prConstraints              := TQUB_QFS_Constraints.Create('Constraints', nil);
  prConstraints.Copy(prModel.Constraints);

  ChannelCount               := prGet_ChannelCount;
  AverageCount               := prGet_AverageCount;
  FitWeight                  := prGet_FitWeight;
  StateCount                 := prModel.States.StateCount;
  RateCount                  := 2*prModel.Rates.RateCount;
  ExtraKinParCount           := prModel.ExtraKineticPars.ExtraKineticParCount;
  cParCount                  := prGet_cParCount;                                // needed by ExpCond setup

  ModelNode                  := Model.CloneAsNode(True, TNode.CreateNull);
  prMilModel                 := MIL_Model_New( ModelNode.GetNodeMem, 1 );       // prGet_ChannelCount ); ChannelCount for MIP but 1 for Mac -- 1 for all now since MIP only uses single-channel info so far.

  //Zeroth part, expcond-related quantities
  prAllocExpCondStuff;

  //First part, model-related quantities
  prAllocModelStuff;

  //Second part, optimization-related quantities
  prAllocOptStuff;


  //Initialize
  if DoInitialize then begin
    Initialize;
  end;

end;

destructor TQUB_Q_Model.Destroy;
begin
  try

    prFreeModelStuff;
    prFreeOptStuff;
    prFreeExpCondStuff;

    prConstraints.Free;

    MIL_Model_Free(prMilModel);

  finally
    inherited;
  end;

end;


procedure TQUB_Q_Model.Initialize;
begin

  //Initialize the experimental conditions-related quantities
  prInitExpCondStuff;

  //Initialize the model-related quantities
  prInitModelStuff;

  //Initialize optimization-related quantities
  prInitOptStuff;

end;

//Result is true if one or more models have confluent eigenvalues
function TQUB_Q_Model.ResetLLandGR :Boolean;
begin

  Result     := False;

  //LL, SS
  LL         := 0;
  SS         := 0;
  PointCount := 0;

end;

//Result is true if the A matrix is incorrectly calculated.
//For now, just detect negative elements
function TQUB_Q_Model.ErrorA :Boolean;
var
  ec, i, j   :Integer;
begin

  Result     := False;

  for ec := 0 to ExpCondCount - 1 do begin
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        if TPMDouble(A[ec])[i, j] < 0 then begin
          Result := True;
          Exit;
        end;
      end;
    end;
  end;

end;


//****************************************************************************//
//                                                                            //
//                             C O N S T R A I N T S                          //
//                                                                            //
//****************************************************************************//
{
  This function serves two purposes:
    1. Fixes K0 and K1 if P is zero, or fixes K1 if Q is zero.
       Obviously, K0 and K1 cannot be meaningfully estimated when multiplied with a zero
    2. For loop balance constraints, for the balance to hold under all
       experimental conditions, another constraint must be added: the clockwise
       sum of K1's must equal the counterclockwise sum.
    3. Removes overdetermined constraints.

    06-01-16 - I am removing the all-voltage-balance, and making it explicit, until I figure out why it doesn't work sometimes.
               It seems it crashes when it is overdetermined. LM
}
function TQUB_Q_Model.Do_AddConstraints :Integer;
var
  LoopConstraint :TQUB_QFS_Constraint;
  Constraint     :TQUB_QFS_Constraint;
  CCount         :Integer;
  HasQ           :Boolean;
  PZ, QZ         :Boolean;
  i, j, k, l     :Integer;
  c, cc          :Integer;
begin

  Result := 0;

  //If a K0 is multiplied with a zero P, then both K0 and K1 are constrained - fixed
  //If a K1 is multiplied with a zero Q, then K1 is constrained - fixed
  for k := 0 to (RateCount div 2) - 1 do begin

    for l := 1 to 2 do begin

      PZ := False;
      QZ := False;

      if IsPZero(k, l) then begin
      //P zero --> fix K0 and K1
        PZ := True;
        QZ := True;
      end;

      if (not PZ) and IsQZero(k, l) then begin
      //Q zero --> fix K1
        QZ := True;
      end;

      if PZ then begin
      //Add constraint for K0
        Constraint := TQUB_QFS_Constraint.Create(CHANNELS_CONSTRAINT_TYPE_NAME[CSTR_FIX_RATE], nil);
        prConstraints.AddConstraint(Constraint, prConstraints.ConstraintCount);
        if l = 1 then begin
          Constraint.Index[0] := prModel.States.IndexOf(prModel.Rates[k].States[1]);
          Constraint.Index[1] := prModel.States.IndexOf(prModel.Rates[k].States[2]);
        end else begin
          Constraint.Index[0] := prModel.States.IndexOf(prModel.Rates[k].States[2]);
          Constraint.Index[1] := prModel.States.IndexOf(prModel.Rates[k].States[1]);
        end;
      end;

      if QZ then begin
        //Add constraint for K1
        Constraint := TQUB_QFS_Constraint.Create(CHANNELS_CONSTRAINT_TYPE_NAME[CSTR_FIX_EXP], nil);
        prConstraints.AddConstraint(Constraint, prConstraints.ConstraintCount);
        if l = 1 then begin
          Constraint.Index[0] := prModel.States.IndexOf(prModel.Rates[k].States[1]);
          Constraint.Index[1] := prModel.States.IndexOf(prModel.Rates[k].States[2]);
        end else begin
          Constraint.Index[0] := prModel.States.IndexOf(prModel.Rates[k].States[2]);
          Constraint.Index[1] := prModel.States.IndexOf(prModel.Rates[k].States[1]);
        end;
      end;

    end;

  end;

  //If Loop (Im)Balance, and at least one Q-dependent rate, add constraint CSTR_LOOP_BALANCE_EXP (TO DO: Maybe it should be explicitely added? LM
  CCount := prConstraints.ConstraintCount;
  for c := 0 to CCount - 1 do begin
    LoopConstraint := prConstraints[c];
    //Is it Loop?
    if (LoopConstraint.cType = CSTR_LOOP_BALANCE) or (LoopConstraint.cType = CSTR_LOOP_IMBALANCE) then begin
      //Does it have at least one Q-dependent rate?
      HasQ := False;
      //Index[0] <> Index[IndexCount - 1] !!!
      for cc := 0 to LoopConstraint.IndexCount - 1 do begin

        //i and j are the indexes of the states
        i := LoopConstraint.Index[cc];
        if cc < (LoopConstraint.IndexCount - 1) then begin
          j := LoopConstraint.Index[cc + 1];
        end else begin
          j := LoopConstraint.Index[0];
        end;

        //k, l are the RateIdx, HalfIdx:
        prModel.Rates.GetRateIdxHalfIdxOfStates(i, j, k, l);
        PZ   := False;
        QZ   := False;
        if IsPZero(k, l) then begin
          PZ := True;
          QZ := True;
        end;
        if (not PZ) and IsQZero(k, l) then begin
          QZ := True;
        end;
        HasQ := not QZ;
        if HasQ then Break;

        prModel.Rates.GetRateIdxHalfIdxOfStates(j, i, k, l);
        PZ   := False;
        QZ   := False;
        if IsPZero(k, l) then begin
          PZ := True;
          QZ := True;
        end;
        if (not PZ) and IsQZero(k, l) then begin
          QZ := True;
        end;
        HasQ := not QZ;
        if HasQ then Break;

      end;

      //For a while, I thought this constraint introduced some problems. Seems to be fine now. ... . Now again I'm not sure. LM.
      //Testing - see above note
      if HasQ then begin
        Constraint := TQUB_QFS_Constraint.Create(CHANNELS_CONSTRAINT_TYPE_NAME[CSTR_LOOP_BALANCE_EXP], nil);
        prConstraints.AddConstraint(Constraint, prConstraints.ConstraintCount);
        for k := 0 to LoopConstraint.IndexCount - 1 do begin
          Constraint.Index[k] := LoopConstraint.Index[k];
        end;
      end;

    end;
  end;

end;

function TQUB_Q_Model.Do_CheckConstraints: Boolean;
var
  MdlTree        :INode;
  CnsTree        :INode;
  CnsNode        :INode;
  CnsRem         :INode;
  Constraint     :TQUB_QFS_Constraint;
  i, j, k           :Integer;
  LoopMsg        :INode;
begin
  Result         := True; // no problems yet
  ConstraintMsg  := '';

  // remove unused ScaleExp; they cause problems
  for k := prConstraints.ConstraintCount - 1 downto 0 do begin
    with prConstraints[k] do begin
      if cType = CSTR_SCALE_EXP then begin
        i := prRateIdxOf(Index[0], Index[1], 1);
        j := prRateIdxOf(Index[2], Index[3], 1);
        if (i < 0) or (j < 0) then begin
          prConstraints.DelConstraint(k);
          Continue;
        end;
        if (cPars[i] = 0) or (cPars[j] = 0) then begin
          prConstraints.DelConstraint(k);
          Continue;
        end;
      end;
    end;
  end;

  // remove extra constraints (overdetermined) using svd in qubopt
  // model as tree
  MdlTree := prModel.CloneAsNode(True, TNode.CreateNull);
  // constraints as tree, replacing those in model
  CnsTree := prConstraints.CloneAsNode(True, TNode.CreateNull);
  MdlTree.RemoveChild(MdlTree['Constraints']);
  MdlTree.AppendChild(CnsTree);
  // PareConstraints
  CnsRem  := TNode.Create('');
  MdlTree := PareConstraints(MdlTree, CnsRem);
  CnsTree := MdlTree['Constraints'];
  // rebuild prconstraints
  i := 0;
  CnsNode := CnsTree.Child;
  while not CnsNode.IsNull do begin
    // delete any from our list that don't match the returned list
    if (CnsNode.Name <> CHANNELS_CONSTRAINT_TYPE_NAME[prConstraints[i].Ctype])
    or (CnsNode.DataCount <> prConstraints[i].IndexCount) then begin
      prConstraints.DelConstraint(i);
      Continue;
    end;

    for j := 0 to CnsNode.DataCount - 1 do begin
      if CnsNode.Data[j] <> prConstraints[i].Index[j] then begin
        prConstraints.DelConstraint(i);
        Continue;
      end;
    end;

    if prConstraints[i].HasValue then begin
      if CnsNode.Find('HasValue').DataCount = 0 then begin
        prConstraints.DelConstraint(i);
        Continue;
      end else if CnsNode['HasValue'].Data[0] = 0 then begin
        prConstraints.DelConstraint(i);
        Continue;
      end else if CnsNode.Find('Value').DataCount = 0 then begin
        prConstraints.DelConstraint(i);
        Continue;
      end else if CnsNode['Value'].Data[0] <> prConstraints[i].Value then begin
        prConstraints.DelConstraint(i);
        Continue;
      end;
    end;

    if prConstraints[i].CoeffCount > 0 then begin
      if CnsNode.Find('Coeffs').DataCount <> prConstraints[i].CoeffCount then begin
        prConstraints.DelConstraint(i);
        Continue;
      end;
      for j := 0 to prConstraints[i].CoeffCount - 1 do begin
        if prConstraints[i].Coeff[j] <> CnsNode['Coeffs'].Data[j] then begin
          prConstraints.DelConstraint(i);
          Continue;
        end;
      end;
    end;

    CnsNode := CnsNode.Sibling;
    Inc(i);
  end;
{ changed: keep originals rather than recreating, since there are more and more sub-node details
  for i := prConstraints.ConstraintCount - 1 downto 0 do begin
    prConstraints.DelConstraint(i);
  end;
  CnsNode := CnsTree.Child;
  while not CnsNode.IsNull do begin
    if CnsNode.Name = 'Generalized' then begin

    end else begin
      Constraint := TQUB_QFS_Constraint.Create(CnsNode.Name, nil);
      prConstraints.AddConstraint(Constraint, prConstraints.ConstraintCount);
      for k := 0 to CnsNode.DataCount - 1 do begin
        Constraint.Index[k] := CnsNode.Data[k];
      end;
    end;
    CnsNode := CnsNode.Sibling;
  end;
 }
  CnsNode := CnsRem.Child;
  while not CnsNode.IsNull do begin
    ConstraintMsg := ConstraintMsg + 'Ignoring surplus constraint: '+CnsNode.Name+': '+CnsNode.DataAsString+#13#10;
    CnsNode := CnsNode.Sibling;
  end;

  fParCount := cParCount - prConstraints.ConstraintCount;//Assume fParCount > 0
  if fParCount <= 0 then begin
    ConstraintMsg := ConstraintMsg + 'Too many constraints; zero or negative number of free parameters.'#13#10;
    Result        := False;
    Exit;
  end;

  // now CheckLoopStimuli
  LoopMsg := CheckLoopStimuli(MdlTree);
  if LoopMsg.DataCount > 0 then begin
    ConstraintMsg := ConstraintMsg + LoopMsg.DataAsString;
    Result        := False;
    Exit;
  end;
end;

function TQUB_Q_Model.IsPZero(Rate, Index :Integer) :Boolean;
begin

  Result := False;

  if (prModel.Rates[Rate].P     [Index])        and
     (prModel.Rates[Rate].PValue[Index] <> nil) and
     (prModel.Rates[Rate].PValue[Index]^ = 0)   then begin
  //it is P-dependent, P is defined and it is zero
    Result := True;
  end;

end;

function TQUB_Q_Model.IsQZero(Rate, Index :Integer) :Boolean;
begin

  Result := False;

  if (prModel.Rates[Rate].Q     [Index])        and
     (prModel.Rates[Rate].QValue[Index] <> nil) and
     (prModel.Rates[Rate].QValue[Index]^ = 0)   then begin
  //it is Q-dependent, Q is defined and it is zero
    //TO DO: LM. if QValue is zero it means nothing
    //Result := True;
  end else if not prModel.Rates[Rate].Q[Index]  then begin
  //it is not Q-dependent
    Result := True;
  end;

end;



//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_Q_Model.Do_fParsToModel :Integer;
var
  i, ec :Integer;
begin
  Result := 0;

  Do_fParsTocPars;

  SmallRateWarning := False;
  BigRateWarning := False;

  Do_cParsToQ;
  Do_cParsToEKP;

  for ec := 0 to ExpCondCount - 1 do begin

    //Q spectral decomposition
    MakeAMatrixPEq(TPMDouble(Q[ec]), nil{A}, nil{PEq}, TPAPointer(Ak[ec]), Lambda[ec], nil{Tau}, nil{Area}, 0{Sampling}, StateCount);

    Confluent[ec] := 0; //False;
    for i := 0 to StateCount - 2 do begin
      if Lambda[ec][i] = Lambda[ec][i + 1] then begin
        Confluent[ec] := 1; // True;
      end;
    end;
  end;

end;

function TQUB_Q_Model.Do_fParsTocPars :Integer;
var
  i, j :Integer;
begin

  Result := 0;

  //cPars = f2cA.fPars + f2cB

  for i := 0 to cParCount - 1 do begin
    cPars[i] := f2cB[i];
    for j := 0 to fParCount - 1 do begin
      cPars[i] := cPars[i] + f2cA[i, j] * fPars[j];
    end;
    if Abs(cPars[i]) < 1e-10 then cPars[i] := 0.0;

    if i < 2*RateCount then begin
    //There are rates up to this, above are ExtraKinPars, N etc

      if not AllowNegRates then begin
        //testing - QMinRate .......................
        if not Odd(i) then begin
          //cPars[i] := ln(Math.Max(exp(cPars[i]), QMinRate));
          cPars[i] := ln(exp(cPars[i]) + QMinRate);
        end;
        //.........................................
      end;
    end;

  end;

end;

function TQUB_Q_Model.Do_cParsToQ :Integer;
var
  i, j, k, ec :Integer;
const
  MAX_RATE = 1e6;
  MIN_RATE = 1e-6;
begin
  /// separate out cond, in orig form;
  /// do q for each ExpCond

  Result := 0;

  for i := 0 to StateCount - 1 do begin
    for j := 0 to StateCount - 1 do begin
      Cond_Q[i, j] := 0;
    end;
  end;

  for k := 0 to (RateCount div 2) - 1 do  begin
    i              := c2qS[4*k].x;
    j              := c2qS[4*k].y;

    if AllowNegRates then begin
      Cond_Q[i, j]   := Cond_ExpCondP[4*k + 0]*cPars[4*k + 0]*Exp(Cond_ExpCondQ[4*k + 1]*cPars[4*k + 1]);
      Cond_Q[j, i]   := Cond_ExpCondP[4*k + 2]*cPars[4*k + 2]*Exp(Cond_ExpCondQ[4*k + 3]*cPars[4*k + 3]);
    end else begin
      Cond_Q[i, j]   := Cond_ExpCondP[4*k + 0]*Exp(cPars[4*k + 0] + Cond_ExpCondQ[4*k + 1]*cPars[4*k + 1]);
      Cond_Q[j, i]   := Cond_ExpCondP[4*k + 2]*Exp(cPars[4*k + 2] + Cond_ExpCondQ[4*k + 3]*cPars[4*k + 3]);
      if Cond_Q[i, j] < MIN_RATE then SmallRateWarning := True;
      if Cond_Q[j, i] < MIN_RATE then SmallRateWarning := True;
      if Cond_Q[i, j] > MAX_RATE then BigRateWarning := True;
      if Cond_Q[j, i] > MAX_RATE then BigRateWarning := True;
    end;

  end;//k

  for i := 0 to StateCount - 1 do begin
    for j := 0 to StateCount - 1 do begin
      if j <> i then Cond_Q[i, i] := Cond_Q[i, i] - Cond_Q[i, j];
    end;
  end;

  for ec := 0 to ExpCondCount - 1 do begin

    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        TPMDouble(Q[ec])[i, j]  := 0;
      end;
    end;

    for k := 0 to (RateCount div 2) - 1 do  begin
      i                         := c2qS[4*k].x;
      j                         := c2qS[4*k].y;

      if AllowNegRates then begin
        TPMDouble(Q[ec])[i, j]  := ExpCondP[ec][4*k + 0]*cPars[4*k + 0]*Exp(ExpCondQ [ec][4*k + 1]*cPars[4*k + 1]);
        TPMDouble(Q[ec])[j, i]  := ExpCondP[ec][4*k + 2]*cPars[4*k + 2]*Exp(ExpCondQ [ec][4*k + 3]*cPars[4*k + 3]);
      end else begin
        TPMDouble(Q[ec])[i, j]  := ExpCondP[ec][4*k + 0]*Exp(cPars[4*k + 0] + ExpCondQ [ec][4*k + 1]*cPars[4*k + 1]);
        TPMDouble(Q[ec])[j, i]  := ExpCondP[ec][4*k + 2]*Exp(cPars[4*k + 2] + ExpCondQ [ec][4*k + 3]*cPars[4*k + 3]);
        //                         ^C                        ^ln(k0)          ^V                     ^k1
        if TPMDouble(Q[ec])[i, j] < MIN_RATE then SmallRateWarning := True;
        if TPMDouble(Q[ec])[j, i] < MIN_RATE then SmallRateWarning := True;
        if TPMDouble(Q[ec])[i, j] > MAX_RATE then BigRateWarning := True;
        if TPMDouble(Q[ec])[j, i] > MAX_RATE then BigRateWarning := True;
      end;
    end;//k

    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        if j <> i then TPMDouble(Q [ec])[i, i] := TPMDouble(Q [ec])[i, i] - TPMDouble(Q [ec])[i, j];
      end;
    end;

  end;

end;

function TQUB_Q_Model.Do_cParsToEKP :Integer;
var
  k :Integer;
begin

  Result := 0;

  for k := 0 to ExtraKinParCount - 1 do begin
    case TQUB_QFS_QMF_ExtraKineticParType(ExtraKinParTypes[k]) of
      EKP_TYPE_LIN:
        ExtraKinPars[k] := cPars[2*RateCount + k];
      EKP_TYPE_LOG:
        ExtraKinPars[k] := Exp(cPars[2*RateCount + k]);
      EKP_TYPE_EXP:
        ExtraKinPars[k] := Ln(cPars[2*RateCount + k]);
      else
        ExtraKinPars[k] := Exp(cPars[2*RateCount + k]);
    end;
  end;//k

end;



//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Q_Model.prAllocModelStuff;
var                                                                      ////  add Confluent
  i, j, k, ec :Integer;
begin

  //Rate index .................................................................
  RateIdx                    := PointAlloc1(RateCount);
  for k := 0 to RateCount div 2 - 1 do begin
    i := prModel.States.IndexOf(prModel.Rates[k].States[1]);
    j := prModel.States.IndexOf(prModel.Rates[k].States[2]);
    RateIdx[k*2 + 0].x := i;
    RateIdx[k*2 + 0].y := j;
    RateIdx[k*2 + 1].x := j;
    RateIdx[k*2 + 1].y := i;
  end;

  //Q matrix and its spectral decomposition ....................................
  Q                          := PAlloc1(ExpCondCount);
  for ec := 0 to ExpCondCount - 1 do begin
    Q[ec]                    := DAlloc2(StateCount, StateCount);
  end;

  Ak                         := PAlloc1(ExpCondCount);
  for ec := 0 to ExpCondCount - 1 do begin
    Ak[ec]                   := PAlloc1(StateCount);
    for k := 0 to StateCount - 1 do begin
      TPAPointer(Ak[ec])[k]  := DAlloc2(StateCount, StateCount);
    end;
  end;

  Lambda                     := DAlloc2(ExpCondCount, StateCount);

  dQdq                       := PAlloc1(RateCount);
  for k := 0 to RateCount - 1 do begin
    dQdq[k]                  := DAlloc2(StateCount, StateCount);
  end;
  Do_dQ_dq;

  Cond_Q                     := DAlloc2(StateCount, StateCount);

  //A matrix ...................................................................
  A                          := PAlloc1(ExpCondCount);
  for ec := 0 to ExpCondCount - 1 do begin
    A[ec]                    := DAlloc2(StateCount, StateCount);
  end;

  dAdq                       := PAlloc1(ExpCondCount);
  for ec := 0 to ExpCondCount - 1 do begin
    dAdq[ec]                 := PAlloc1(RateCount);
    for k := 0 to RateCount - 1 do begin
      TPAPointer(dAdq[ec])[k] := DAlloc2(StateCount, StateCount);
    end;
  end;

  At                         := PAlloc1(ExpCondCount);
  for ec := 0 to ExpCondCount - 1 do begin
    At[ec]                   := DAlloc2(StateCount, StateCount);
  end;

  dAtdq                      := PAlloc1(ExpCondCount);
  for ec := 0 to ExpCondCount - 1 do begin
    dAtdq[ec]                 := PAlloc1(RateCount);
    for k := 0 to RateCount - 1 do begin
      TPAPointer(dAtdq[ec])[k] := DAlloc2(StateCount, StateCount);
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          TPMDouble(TPAPointer(dAtdq[ec])[k])[i, j] := 0;
        end;
      end;
    end;
  end;

  Confluent                  := IAlloc1(ExpCondCount);

  //State probabilities ........................................................
  P0                         := DAlloc1(StateCount);
  Pt                         := DAlloc1(StateCount);
  Peq                        := DAlloc1(StateCount);
  Cond_Peq                   := DAlloc1(StateCount);

  dP0dq                      := PAlloc1(ExpCondCount);
  dP0dCond_q                 := DAlloc2(RateCount, StateCount);
  dPtdq                      := PAlloc1(ExpCondCount);
  dPtdCond_q                 := DAlloc2(RateCount, StateCount);
  dPeqdq                     := PAlloc1(ExpCondCount);
  dCond_PeqdCond_q           := DAlloc2(RateCount, StateCount);
  for ec := 0 to ExpCondCount - 1 do begin
    dP0dq[ec]                := DAlloc2(RateCount, StateCount);
    dPtdq[ec]                := DAlloc2(RateCount, StateCount);
    dPeqdq[ec]               := DAlloc2(RateCount, StateCount);
  end;

  //Class info .................................................................
  Do_ClassCount;


  //Temporary quantities
  Vec1                       := DAlloc1(StateCount);
  Vec2                       := DAlloc1(StateCount);
  Mat1                       := DAlloc2(StateCount, StateCount);
  Mat2                       := DAlloc2(StateCount, StateCount);

end;

procedure TQUB_Q_Model.prInitModelStuff;
begin
end;


procedure TQUB_Q_Model.prFreeModelStuff;
var                                                 //// add Confluent
  k :Integer;
  ec :Integer;
begin

  //Rate index .................................................................
  FreeMem(RateIdx);
  RateIdx := nil;


  //Q matrix and its spectral decomposition ....................................
  if Q <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      Free2(TPMChar( Q[ec] ));
    end;
    FreeMem(Q);
    Q := nil;
  end;

  if Ak <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      for k := 0 to StateCount - 1 do begin
        Free2(TPMChar(TPAPointer(Ak[ec])[k]));
        //Ak[ec][k] := nil;
      end;
      FreeMem(Ak[ec]);
    end;
    FreeMem(Ak);
    Ak := nil;
  end;

  Free2(TPMChar(Lambda));
  Lambda := nil;

  Free2(TPMChar(Cond_Q));
  Cond_Q := nil;

  if dQdq <> nil then begin
    for k := 0 to RateCount - 1 do begin
      Free2(TPMChar(dQdq[k]));
      dQdq[k] := nil;
    end;
  end;
  dQdq := nil;


  //A matrix ...................................................................
  if A <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      Free2(TPMChar(A[ec]));
    end;
    FreeMem(A);
    A := nil;
  end;

  if dAdq <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      for k := 0 to RateCount - 1 do begin
        Free2(TPMChar(TPAPointer(dAdq[ec])[k]));
      end;
      FreeMem(dAdq[ec]);
    end;
    FreeMem(dAdq);
    dAdq := nil;
  end;

  if At <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      Free2(TPMChar(At[ec]));
    end;
    FreeMem(At);
    At := nil;
  end;

  if dAtdq <> nil then begin
    for ec := 0 to ExpCondCount - 1 do begin
      for k := 0 to RateCount - 1 do begin
        Free2(TPMChar(TPAPointer(dAtdq[ec])[k]));
      end;
      FreeMem(dAtdq[ec]);
    end;
    FreeMem(dAtdq);
    dAtdq := nil;
  end;


  //State probabilities ........................................................
  FreeMem(P0);
  P0 := nil;
  FreeMem(Pt);
  Pt := nil;
  FreeMem(Peq);
  Peq := nil;
  FreeMem(Cond_Peq);
  Cond_Peq := nil;


  //State probabilities derivatives ............................................
  if dP0dq            <> nil then for ec := 0 to ExpCondCount - 1 do Free2(TPMChar(dP0dq[ec]));
  if dPtdq            <> nil then for ec := 0 to ExpCondCount - 1 do Free2(TPMChar(dPtdq[ec]));
  if dPeqdq           <> nil then for ec := 0 to ExpCondCount - 1 do Free2(TPMChar(dPeqdq[ec]));
  FreeMem(dP0dq);
  dP0dq := nil;
  FreeMem(dPtdq);
  dPtdq := nil;
  Free2(TPMChar(dPtdCond_q));
  dPtdCond_q := nil;
  Free2(TPMChar(dP0dCond_q));
  dP0dCond_q := nil;
  FreeMem(dPeqdq);
  dPeqdq := nil;
  Free2(TPMChar(dCond_PeqdCond_q));
  dCond_PeqdCond_q := nil;


  FreeMem(Confluent);


  //Class info .................................................................
  FreeMem(Classes);
  Classes := nil;
  FreeMem(Cls);
  Cls := nil;


  //Temporary quantities .......................................................
  FreeMem(Vec1);
  Vec1 := nil;
  FreeMem(Vec2);
  Vec2 := nil;
  Free2(TPMChar(Mat1));
  Mat1 := nil;
  Free2(TPMChar(Mat2));
  Mat2 := nil;

end;



//****************************************************************************//
//                                                                            //
//               E X P E R I M E N T A L     C O N D I T I O N S              //
//                                                                            //
//****************************************************************************//
procedure TQUB_Q_Model.prAllocExpCondStuff;
begin
  if (prExpCondParams = nil) or (prExpCondParams.Count = 0) then begin
    ExpCondCount             := 1;
  end else begin
    ExpCondCount             := prExpCondParams.Count;
  end;

  ExpCondP                   := DAlloc2(ExpCondCount, cParCount);
  ExpCondQ                   := DAlloc2(ExpCondCount, cParCount);
  Cond_ExpCondP              := DAlloc1(cParCount);
  Cond_ExpCondQ              := DAlloc1(cParCount);
end;

procedure TQUB_Q_Model.prInitExpCondStuff;
var
  Par      :TQUB_ExpCond_PParameter;
  k, l     :Integer;
  ec       :Integer;
  ExpCond  :TQUB_ExpCond_Parameters;
begin

  //Result := 0;

  try


    //************************************************************************//
    //                                                                        //
    //                                  T E S T                               //
    //                                                                        //
    //************************************************************************//
    if prExpCondParams = nil then begin
    //No parameters

      //Make ExpCondP and ExpCondQ .................................................
      for k := 0 to (RateCount div 2) - 1 do begin

        //P
        for l := 0 to 3 do ExpCondP[0][4*k + l] := 1;

        //Q
        for l := 0 to 3 do ExpCondQ[0][4*k + l] := 0;

      end;

    end else begin
    //Get the parameters from ExpCondParams

      for ec := 0 to ExpCondCount - 1 do begin
        ExpCond := ExpCondParams[ec];

        //Make ExpCondP and ExpCondQ .................................................
        for k := 0 to (RateCount div 2) - 1 do begin

          for l := 1 to 2 do begin

            //P
            ExpCondP[ec][4*k + (l - 1)*2 + 0] := 1;
            ExpCondP[ec][4*k + (l - 1)*2 + 1] := 1;
            if prModel.Rates[k].P[l] then begin
              Par := ExpCond.ParameterByName[prModel.Rates[k].PName[l]];
              if Par <> nil then begin
                ExpCondP[ec][4*k + (l - 1)*2 + 0] := Par.Value;
              end;
            end;

            //Q
            ExpCondQ[ec][4*k + (l - 1)*2 + 0] := 0;
            ExpCondQ[ec][4*k + (l - 1)*2 + 1] := 0;
            if prModel.Rates[k].Q[l] then begin
              Par := ExpCond.ParameterByName[prModel.Rates[k].QName[l]];
              if Par <> nil then begin
                ExpCondQ[ec][4*k + (l - 1)*2 + 1] := Par.Value;
              end;
            end;

          end;//l

        end;//k

      end; //ec

    end;//Model or ExpCondParams


    //************************************************************************//
    //                                                                        //
    //                         C O N D I T I O N I N G                        //
    //                                                                        //
    //************************************************************************//
    ExpCond := ExpCondParams[0];

    for k := 0 to (RateCount div 2) - 1 do begin

      for l := 1 to 2 do begin

        //P
        Cond_ExpCondP[4*k + (l - 1)*2 + 0] := 1;
        Cond_ExpCondP[4*k + (l - 1)*2 + 1] := 1;
        if prModel.Rates[k].P[l] then begin
          Par := ExpCond.ParameterByName['Cond_' + prModel.Rates[k].PName[l]];
          if Par <> nil then begin
            Cond_ExpCondP[4*k + (l - 1)*2 + 0] := Par.Value;
          end else begin
            Par := ExpCond.ParameterByName[prModel.Rates[k].PName[l]];
            if Par <> nil then begin
              Cond_ExpCondP[4*k + (l - 1)*2 + 0] := Par.Value;
            end;
          end;
        end;

        //Q
        Cond_ExpCondQ[4*k + (l - 1)*2 + 0] := 0;
        Cond_ExpCondQ[4*k + (l - 1)*2 + 1] := 0;
        if prModel.Rates[k].Q[l] then begin
          Par := ExpCond.ParameterByName['Cond_' + prModel.Rates[k].QName[l]];
          if Par <> nil then begin
            Cond_ExpCondQ[4*k + (l - 1)*2 + 1] := Par.Value;
          end else begin
            Par := ExpCond.ParameterByName[prModel.Rates[k].QName[l]];
            if Par <> nil then begin
              Cond_ExpCondQ[4*k + (l - 1)*2 + 1] := Par.Value;
            end;
          end;
        end;

      end;//l

    end;//k

  except

    //Result := 1;

  end;
end;

procedure TQUB_Q_Model.prFreeExpCondStuff;
begin

  Free2(TPMChar(ExpCondP));
  ExpCondP := nil;
  Free2(TPMChar(ExpCondQ));
  ExpCondQ := nil;
  FreeMem(Cond_ExpCondP);
  Cond_ExpCondP := nil;
  FreeMem(Cond_ExpCondQ);
  Cond_ExpCondQ := nil;

end;

function TQUB_Q_Model.prGetExpCondParams(ec :Integer): TQUB_ExpCond_Parameters;
begin
  if (ec >= 0) and (ec < prExpCondParams.Count) then begin
    Result := prExpCondParams[ec];
  end else begin
    Result := nil;
  end;
end;

//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Q_Model.prAllocOptStuff;
var
  i, j :Integer;
begin

  //Constrained parameters .....................................................
  cParCount                  := prGet_cParCount;
  cPars                      := DAlloc1(cParCount);
  cParsVar                   := DAlloc1(cParCount);

  //Extra Kinetic Parameters
  if ExtraKinParCount > 0 then begin
    ExtraKinPars             := DAlloc1(ExtraKinParCount);
    ExtraKinParTypes         := IAlloc1(ExtraKinParCount);
  end;

  //Mapping ....................................................................
  c2qS                       := PointAlloc1(cParCount);

  //Constraints ................................................................
  ConstA                     := DAlloc2(cParCount, cParCount);//at maximum; TODO: checkit. LM
  ConstB                     := DAlloc1(cParCount);
  for i := 0 to cParCount - 1 do begin
    ConstB[i] := 0;
    for j := 0 to cParCount - 1 do begin
      ConstA[i, j] := 0;
    end;
  end;


  //Mapping ....................................................................
  f2cA                       := DAlloc2(cParCount, cParCount);//at maximum
  f2cB                       := DAlloc1(cParCount);
  for i := 0 to cParCount - 1 do begin
    f2cB[i] := 0;
    for j := 0 to cParCount - 1 do begin
      f2cA[i, j] := 0;
    end;
  end;


  //Free parameters ............................................................
  fPars                      := DAlloc1(cParCount);
  fParsVar                   := DAlloc2(cParCount, cParCount);
  for i := 0 to cParCount - 1 do begin             //cParCount because at this point we don't know fParCount
    fPars[i] := 0;//new. OK?
    for j := 0 to cParCount - 1 do begin
      fParsVar[i, j] := 0;
    end;
  end;
  for j := 0 to cParCount - 1 do begin
    fParsVar[j, j] := 1;
  end;


  //Derivatives
  dLLdq                      := DAlloc2(ExpCondCount, RateCount);
  dqdcPar                    := DAlloc1(cParCount);

  dLLdCond_q                 := DAlloc1(RateCount);
  for i := 0 to RateCount - 1 do begin
    dLLdCond_q[i]            := 0;
  end;
  dLLdfPar                   := DAlloc1(cParCount);           //at most
  dCond_qdcPar               := DAlloc1(cParCount);
  dLLdcPar                   := DAlloc1(cParCount);

end;

procedure TQUB_Q_Model.prInitOptStuff;
var
  HasUserConstraints   :Boolean;   // save pointless checking for duplicates (much faster sim of big models)
begin

  //Make c2qS ..................................................................
  Do_c2s;


  //Make initial Extra Kinetic Parameters ......................................
  Do_InitExtraKinPars;


  //Make initial cPars .........................................................
  Do_InitcPars;


  //Constraints ................................................................
  HasUserConstraints := prConstraints.ConstraintCount > 0;
  Do_AddConstraints;
  // tentative constraint count may include duplicates
  ConstCount := prConstraints.ConstraintCount;
  fParCount := cParCount - ConstCount;//Assume fParCount > 0
  if HasUserConstraints and not Do_CheckConstraints then begin
    raise Exception.Create(ConstraintMsg);
  end;
  // now there may be fewer constraints so update the counts:
  ConstCount := prConstraints.ConstraintCount;


  //Free Parameters ............................................................
  fParCount := cParCount - ConstCount;//Assume fParCount > 0
  Assert(fParCount >= 0, 'Negative number of free parameters');

  //Make ConstA and ConstB .....................................................
  Do_ConstAB;


  //Make f2cA and f2cB by SVD ..................................................
  Do_f2cAB;


  //Init fPars .................................................................
  Do_InitfPars;

end;

procedure TQUB_Q_Model.prFreeOptStuff;
begin

  //Free parameters ............................................................
  FreeMem(fPars);
  fPars := nil;
  Free2(TPMChar(fParsVar));
  fParsVar := nil;

  //Extra Kinetic Parameters ...................................................
  FreeMem(ExtraKinPars);
  ExtraKinPars := nil;
  FreeMem(ExtraKinParTypes);
  ExtraKinParTypes := nil;

  //Constrained parameters .....................................................
  FreeMem(cPars);
  cPars := nil;
  FreeMem(cParsVar);
  cParsVar := nil;

  //Constraints ................................................................
  Free2(TPMChar(ConstA));
  ConstA := nil;
  FreeMem(ConstB);

  //Mapping ....................................................................
  Free2(TPMChar(f2cA));
  f2cA := nil;
  FreeMem(f2cB);
  f2cB := nil;
  FreeMem(c2qS);
  c2qS := nil;

  //Derivatives ................................................................
  Free2(TPMChar(dLLdq));
  dLLdq := nil;
  FreeMem(dqdcPar);
  dqdcPar := nil;
  FreeMem(dLLdfPar);
  dLLdfPar := nil;
  FreeMem(dLLdCond_q);
  dLLdCond_q := nil;
  FreeMem(dCond_qdcPar);
  dCond_qdcPar := nil;
  FreeMem(dLLdcPar);
  dLLdcPar := nil;

end;

function TQUB_Q_Model.prRateIdxOf(idx1, idx2, k01 :Integer) :Integer;
var
  k :Integer;
begin
  Result := -1;
  for k := 0 to cParCount - 1 do begin
    if (c2qS[k].x = idx1) and (c2qS[k].y = idx2) then begin
      if k01 = 0 then Result := k else Result := k + 1;
      Break;
    end;
  end;
end;



//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
function TQUB_Q_Model.Do_InitcPars :Integer;
var
  k :Integer;
begin

  Result := 0;

  //Rates
  for k := 0 to (RateCount div 2) - 1 do begin

    //i-->j
    //cPars[4*k + 0]  := ln(prModel.Rates[k].K0[1]);
    //cPars[4*k + 0]  := ln(Math.Max(prModel.Rates[k].K0[1], QMinRate));
    if AllowNegRates then begin
      cPars[4*k + 0]  := prModel.Rates[k].K0[1];
    end else begin
      cPars[4*k + 0]  := ln(prModel.Rates[k].K0[1] + QMinRate);
    end;
    cPars[4*k + 1]  := prModel.Rates[k].K1[1];

    //j-->i
    //cPars[4*k + 2]  := ln(prModel.Rates[k].K0[2]);
    //cPars[4*k + 2]  := ln(Math.Max(prModel.Rates[k].K0[2], QMinRate));
    if AllowNegRates then begin
      cPars[4*k + 2]  := prModel.Rates[k].K0[2];
    end else begin
      cPars[4*k + 2]  := ln(prModel.Rates[k].K0[2] + QMinRate);
    end;
    cPars[4*k + 3]  := prModel.Rates[k].K1[2];

  end;

  //Extra Kinetic Parameters
  for k := 0 to ExtraKinParCount - 1 do begin

    case prModel.ExtraKineticPars[k].EKPType of
      EKP_TYPE_LIN:
        cPars[2*RateCount + k] := prModel.ExtraKineticPars[k].Value;
      EKP_TYPE_LOG:
        cPars[2*RateCount + k]  := ln(prModel.ExtraKineticPars[k].Value);
      EKP_TYPE_EXP:
        cPars[2*RateCount + k] := exp(prModel.ExtraKineticPars[k].Value);
      else
        cPars[2*RateCount + k] := ln(prModel.ExtraKineticPars[k].Value);
    end;

  end;
  //

end;

function TQUB_Q_Model.Do_InitExtraKinPars :Integer;
var
  k :Integer;
begin

  Result := 0;

  //Extra Kinetic Parameters
  for k := 0 to ExtraKinParCount - 1 do begin

    ExtraKinPars[k]  := prModel.ExtraKineticPars[k].Value;
    ExtraKinParTypes[k]  := Integer(prModel.ExtraKineticPars[k].EKPType);

  end;

end;

//Make ConstA and ConstB
function TQUB_Q_Model.Do_ConstAB :Integer;
var
  c, i, j, k, ii, jj :Integer;
  PZ, QZ             :Boolean;
  r, h               :Integer;
begin

  //It currently doesn't work with negative rates!

  Result := 0;

  for c := 0 to prConstraints.ConstraintCount - 1 do begin

    with prConstraints[c] do begin

      case CType of

        CSTR_FIX_RATE://Actually, fix K0
          begin
            i := prRateIdxOf(Index[0], Index[1], 0);
            if i >= 0 then begin
              if HasValue then begin
                ConstB[c]    := ln(Value);
              end else begin
                ConstB[c]    := cPars[i];
              end;
              ConstA[c, i] := 1;
            end;
          end;

        CSTR_FIX_EXP:
          begin
            i := prRateIdxOf(Index[0], Index[1], 1);
            if i >= 0 then begin
              if HasValue then begin
                ConstB[c]    := Value;
              end else begin
                ConstB[c]    := cPars[i];
              end;
              ConstA[c, i] := 1;
            end;
          end;

        CSTR_SCALE_RATE:
          begin
            i := prRateIdxOf(Index[0], Index[1], 0);
            j := prRateIdxOf(Index[2], Index[3], 0);
            if (i >= 0) and (j >= 0) then begin
              if HasValue then begin
                ConstB[c]    := ln(Value);
              end else begin
                ConstB[c]    := cPars[i] - cPars[j];
              end;
              ConstA[c, i] :=  1;
              ConstA[c, j] := -1;
            end;
          end;

        CSTR_SCALE_EXP:
          begin
            i := prRateIdxOf(Index[0], Index[1], 1);
            j := prRateIdxOf(Index[2], Index[3], 1);
            if (i >= 0) and (j >= 0) then begin
              ConstB[c]    :=  0;
              ConstA[c, i] :=  1;
              if HasValue then begin
                ConstA[c, j] := -Value;
              end else begin
                ConstA[c, j] := -cPars[i]/cPars[j];
              end;
            end;
          end;

        CSTR_LOOP_BALANCE,
        CSTR_LOOP_IMBALANCE:

          begin
            //Index[0] <> Index[IndexCount - 1] !!!
            for k := 0 to IndexCount - 1 do begin
              ii := Index[k];
              if k < (IndexCount - 1) then begin
                jj := Index[k + 1];
              end else begin
                jj := Index[0];
              end;

              //k0
              i := prRateIdxOf(ii, jj, 0);//-->
              j := prRateIdxOf(jj, ii, 0);//<--

              if (i >= 0) and (j >= 0) then begin
                ConstA[c, i] :=  1;
                ConstA[c, j] := -1;

                ConstB[c]    := ConstB[c] - ln(max(1e-10,ExpCondP[0][i])) + ln(max(1e-10,ExpCondP[0][j]));

                //k1
                i := prRateIdxOf(ii, jj, 1);//-->
                j := prRateIdxOf(jj, ii, 1);//<--

                //Testing
                //Using the actual voltage may result in different constraint matrices between different models, which doesn't quite make sense.
                //Voltage = 0 is just as good:
                //ConstA[c, i] :=  ExpCondQ[0][i];
                //ConstA[c, j] := -ExpCondQ[0][j];
                ConstA[c, i] :=  0;
                ConstA[c, j] := -0;
              end;
            end;

            //If IMBALANCE, add the imbalance factor to ConstB
            if CType = CSTR_LOOP_IMBALANCE then begin
              if HasValue then begin
              //Explicitly user-specified factor
                ConstB[c] := ConstB[c] + ln(Value);
              end else begin
              //Take it from the clock/counterclockwise product ratio
                { TODO -oLorin :  }
              end;
            end;

            {
              The loop constraint set as above does not guarantee balance at all
              concentrations and at all voltages. See "How to satisfy cycle.pdf"
              for solution. Must be accompanied by CSTR_LOOP_BALANCE_EXP below.
            }

          end;

        CSTR_LOOP_BALANCE_EXP:
        //CSTR_LOOP_IMBALANCE_EXP: TO DO: this is not implemented yet. Does anyone use imbalance? Good luck! LM

          begin

            ConstB[c]    := 0;

            //Index[0] <> Index[IndexCount - 1] !!!
            for k := 0 to IndexCount - 1 do begin
              ii := Index[k];
              if k < (IndexCount - 1) then begin
                jj := Index[k + 1];
              end else begin
                jj := Index[0];
              end;

              //k1
              i := prRateIdxOf(ii, jj, 1);//-->
              j := prRateIdxOf(jj, ii, 1);//<--

              if (i >= 0) and (j >= 0) then begin
                //k, l are the RateIdx, HalfIdx:
                prModel.Rates.GetRateIdxHalfIdxOfStates(ii, jj, r, h);
                QZ := False;
                PZ := False;
                if IsPZero(r, h) then begin
                  PZ := True;
                  QZ := True;
                end;
                if (not PZ) and IsQZero(r, h) then begin
                  QZ := True;
                end;
                if not QZ then begin
                  ConstA[c, i] :=  1;
                end;
  
                //k, l are the RateIdx, HalfIdx:
                prModel.Rates.GetRateIdxHalfIdxOfStates(jj, ii, r, h);
                QZ := False;
                PZ := False;
                if IsPZero(r, h) then begin
                  PZ := True;
                  QZ := True;
                end;
                if (not PZ) and IsQZero(r, h) then begin
                  QZ := True;
                end;
                if not QZ then begin
                  ConstA[c, j] := -1;
                end;
              end;
            end;

          end;

        //testing
        CSTR_GEN:  //generalized constraint, may refer to Extra Kinetic Parameters
          begin
            ConstB[c] := Value;
            for i := 0 to 2*RateCount + ExtraKinParCount - 1 do begin
              ConstA[c, i] := Coeff[i];
            end;
          end;


      end;//CType

    end;//with prConstraints[c]

  end;//for c

end;

//Make f2cA and f2cB by SVD
function TQUB_Q_Model.Do_f2cAB :Integer;
var
  U, V      :TPMDouble;
  W, H      :TPADouble;
  i, j      :Integer;
  //t         :String;
const
  MAD_SMALL = 1e-9;
begin

  Result := 0;

  if ConstCount > 0 then begin

    U := DAlloc2(cParCount + 1, cParCount + 1);
    W := DAlloc1(cParCount + 1);
    V := DAlloc2(cParCount + 1, cParCount + 1);
    H := DAlloc1(cParCount);

    try

      for i := 0 to ConstCount - 1 do begin
        for j := 0 to cParCount - 1 do begin
          U[i + 1, j + 1] := ConstA[i, j];
        end;
      end;
      try
        SVD(U, ConstCount, cParCount, W, V);
      except
      end;
      for i := 0 to ConstCount - 1 do begin
        for j := 0 to cParCount - 1 do begin
          U[i, j] := U[i + 1, j + 1];
        end;
      end;
      for i := 0 to cParCount - 1 do begin
        if Abs(W[i + 1]) < MAD_SMALL then begin          // chris attempt to tame it 7-20-5
          W[i] := 0;
        end else begin
          W[i] := W[i + 1];
        end;
        for j := 0 to cParCount - 1 do begin
          V[i, j] := V[i + 1, j + 1];
        end;
      end;

      //H := W-1.U-1.ConstB
      for i := 0 to ConstCount - 1 do begin
        H[i] := 0;
        for j := 0 to ConstCount - 1 do begin
          H[i] := H[i] + U[j, i] * ConstB[j];
        end;
        if W[i] = 0 then begin                      // chris attempt part 2
          H[i] := 0;
        end else begin
          H[i] := H[i] / W[i];
        end;
      end;

      //f2cB
      for i := 0 to cParCount - 1 do begin
        f2cB[i] := 0;
        for j := 0 to ConstCount - 1 do begin
          f2cB[i] := f2cB[i] + V[i, j] * H[j];
        end;
      end;

      //f2cA
      for i := 0 to cParCount - 1 do begin
        for j := ConstCount to cParCount - 1 do begin
          if Abs(V[i, j]) < MAD_SMALL then begin          // look for MAD_SMALL in qubopt.dll to find corresponding c code
            f2cA[i, j - ConstCount] := 0;
          end else begin
            f2cA[i, j - ConstCount] := V[i, j];
          end;
        end;
      end;


    finally

      Free2(TPMChar(U));
      Free2(TPMChar(V));
      FreeMem(W);
      FreeMem(H);

    end;


  end else begin
  //one-to-one mapping between fPars and cPars

    for i := 0 to cParCount - 1 do begin
      for j := 0 to fParCount - 1 do f2cA[i, j] := 0;
      f2cA[i, i] := 1;
      f2cB[i]    := 0;
    end;

  end;

end;

function TQUB_Q_Model.Do_InitfPars :Integer;
var
  i, j   :Integer;
begin

  Result := 0;

  //fPars from cPars: fPars = f2cA-1.cPars
  for i := 0 to fParCount - 1 do begin
    fPars[i] := 0;
    for j := 0 to cParCount - 1 do begin
      fPars[i] := fPars[i] + f2cA[j, i] * cPars[j];
    end;
  end;

  //It seems numerical Receipes asks for unity matrix
  //for j := 0 to fParCount - 1 do begin
  //  fParsVar[j, j] := Max(Abs(fPars[j]), 1);
  //end;

end;

function TQUB_Q_Model.Do_c2s :Integer;
var
  i, j, k :Integer;
begin

  Result := 0;

  for k := 0 to (RateCount div 2) - 1 do begin
    i := prModel.States.IndexOf(prModel.Rates[k].States[1]);
    j := prModel.States.IndexOf(prModel.Rates[k].States[2]);
    c2qS[4*k + 0].x  := i;
    c2qS[4*k + 0].y  := j;
    c2qS[4*k + 1].x  := i;
    c2qS[4*k + 1].y  := j;
    c2qS[4*k + 2].x  := j;
    c2qS[4*k + 2].y  := i;
    c2qS[4*k + 3].x  := j;
    c2qS[4*k + 3].y  := i;
  end;

end;



//****************************************************************************//
//                                                                            //
//                              Q    M a t r i x                              //
//                                                                            //
//****************************************************************************//
function TQUB_Q_Model.Do_dQ_dq :Integer;
var
  i, j, qq :Integer;
begin

  Result := 0;

  for qq := 0 to RateCount - 1 do begin
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        TPMDouble(dQdq[qq])[i, j] := 0;
      end;
    end;
    TPMDouble(dQdq[qq])[RateIdx[qq].x, RateIdx[qq].x] := -1;
    TPMDouble(dQdq[qq])[RateIdx[qq].x, RateIdx[qq].y] :=  1;
  end;
end;


//****************************************************************************//
//                                                                            //
//                              A    M a t r i x                              //
//                                                                            //
//****************************************************************************//

//Calculates the A matrix for each exp. cond. ==> ExpCondCount A matrices
function TQUB_Q_Model.Do_A(dt :Double) :Integer;
var
  i, j, k, ec :Integer;
begin

  Result := 0;

  for ec := 0 to ExpCondCount - 1 do begin

    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        TPMDouble(A[ec])[i, j] := 0;
        for k := 0 to StateCount - 1 do begin
          TPMDouble(A[ec])[i, j] := TPMDouble(A[ec])[i, j] + TPMDouble(TPAPointer(Ak[ec])[k])[i, j] * Exp(Lambda[ec][k] * dt);
        end;//k
      end;//j
    end;//i

  end;//ec

end;

//Calculates the A matrix for exp. cond. ExpCond ==> one A matrix, the rest are undefined
function TQUB_Q_Model.Do_A(dt :Double; ExpCond :Integer) :Integer;
var
  i, j, k :Integer;
begin

  Result  := 0;

  for i := 0 to StateCount - 1 do begin
    for j := 0 to StateCount - 1 do begin
      TPMDouble(A[ExpCond])[i, j] := 0;
      for k := 0 to StateCount - 1 do begin
        TPMDouble(A[ExpCond])[i, j] := TPMDouble(A[ExpCond])[i, j] + TPMDouble(TPAPointer(Ak[ExpCond])[k])[i, j] * Exp(Lambda[ExpCond][k] * dt);
      end;//k
    end;//j
  end;//i

end;



//Calculates the At matrix for each exp. cond. ==> ExpCondCount At matrices
function TQUB_Q_Model.Do_At(t :Double) :Integer;
var
  i, j, k, ec :Integer;
begin

  Result := 0;

  for ec := 0 to ExpCondCount - 1 do begin

    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        TPMDouble(At[ec])[i, j] := 0;
        for k := 0 to StateCount - 1 do begin
          TPMDouble(At[ec])[i, j] := TPMDouble(At[ec])[i, j] + TPMDouble(TPAPointer(Ak[ec])[k])[i, j] * Exp(Lambda[ec][k] * t);
        end;//k
      end;//j
    end;//i

  end;//ec

end;

//Calculates the At matrix for exp. cond. ExpCond ==> one At matrix, the rest are undefined
function TQUB_Q_Model.Do_At(t :Double; ExpCond :Integer) :Integer;
var
  M       :TPMDouble;
  Mk      :TPAPointer;
  L       :TPADouble;
  a       :Double;
  i, j, k :Integer;
begin

  Result  := 0;
  (*
  for i := 0 to StateCount - 1 do begin
    for j := 0 to StateCount - 1 do begin
      TPMDouble(At[ExpCond])[i, j] := 0;
      for k := 0 to StateCount - 1 do begin
        TPMDouble(At[ExpCond])[i, j] := TPMDouble(At[ExpCond])[i, j] + TPMDouble(TPAPointer(Ak[ExpCond])[k])[i, j] * Exp(Lambda[ExpCond][k] * t);
      end;//k
    end;//j
  end;//i
  *)

  M       := TPMDouble(At[ExpCond]);
  MK      := TPAPointer(Ak[ExpCond]);
  L       := TPADouble(Lambda[ExpCond]);
  for i := 0 to StateCount - 1 do begin
    for j := 0 to StateCount - 1 do begin
      a := 0;
      for k := 0 to StateCount - 1 do begin
        a := a + TPMDouble(Mk[k])[i, j] * Exp(L[k] * t);
      end;//k
      M[i, j] := a;
    end;//j
  end;//i

end;

function TQUB_Q_Model.Do_dA_dq(dt :Double) :Integer;
var
  i, j, k, l, qq, ec :Integer;
  gklt               :Double;
begin

  Result := 0;

  {
    dA[i]/dq[j] should be a square matrix, but since only the diagonal
                terms are non-zero, it is stored as a vector
  }

  for ec := 0 to ExpCondCount - 1 do begin

    for qq := 0 to RateCount - 1 do begin

      //Clear it first
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          TPMDouble(TPAPointer(dAdq[ec])[qq])[i, j] := 0;
        end;
      end;

      for k := 0 to StateCount - 1 do begin
        for l := 0 to StateCount - 1 do begin

          //M1 = Ak*dQdq
          MulMatMat(TPAPointer(Ak[ec])[k], dQdq[qq], Mat1, StateCount);//this can be optimized
          //M2 = M1*Al
          MulMatMat(Mat1, TPAPointer(Ak[ec])[l], Mat2, StateCount);

          Do_G_kl_t(k, l, ec, dt, gklt);
          //M2 = M2*g_kl(t)
          for i := 0 to StateCount - 1 do begin
            for j := 0 to StateCount - 1 do begin
              Mat2[i, j] := Mat2[i, j] * gklt;
            end;
          end;
          //dAdq = dAdq + M2
          SumMatMat(TPAPointer(dAdq[ec])[qq], Mat2, TPAPointer(dAdq[ec])[qq], StateCount);

        end;//l
      end;//k

    end;//qq

  end;//ec

end;

function TQUB_Q_Model.Do_dA_dq(dt :Double; ExpCond :Integer) :Integer;
var
  i, j, k, l, qq, ec :Integer;
  gklt               :Double;
begin

  Result := 0;

  {
    Same as above, but it calculates only dA[ExpCond] / dq[ExpCond], and zeroes the rest
  }

  for ec := 0 to ExpCondCount - 1 do begin

    for qq := 0 to RateCount - 1 do begin

      //Clear it first
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          TPMDouble(TPAPointer(dAdq[ec])[qq])[i, j] := 0;
        end;
      end;

      if ExpCond = ec then begin

        for k := 0 to StateCount - 1 do begin
          for l := 0 to StateCount - 1 do begin

            //M1 = Ak*dQdq
            MulMatMat(TPAPointer(Ak[ec])[k], dQdq[qq], Mat1, StateCount);//this can be optimized
            //M2 = M1*Al
            MulMatMat(Mat1, TPAPointer(Ak[ec])[l], Mat2, StateCount);

            Do_G_kl_t(k, l, ec, dt, gklt);
            //M2 = M2*g_kl(t)
            for i := 0 to StateCount - 1 do begin
              for j := 0 to StateCount - 1 do begin
                Mat2[i, j] := Mat2[i, j] * gklt;
              end;
            end;
            //dAdq = dAdq + M2
            SumMatMat(TPAPointer(dAdq[ec])[qq], Mat2, TPAPointer(dAdq[ec])[qq], StateCount);

          end;//l
        end;//k

      end;//ExpCond = ec

    end;//qq

  end;//ec

end;

function TQUB_Q_Model.Do_dAt_dq(t :Double) :Integer;
var
  i, j, k, l, qq, ec :Integer;
  gklt               :Double;
begin

  Result := 0;

  {
    dAt[i]/dq[j] should be a square matrix, but since only the diagonal
                terms are non-zero, it is stored as a vector
  }


  for ec := 0 to ExpCondCount - 1 do begin

    for qq := 0 to RateCount - 1 do begin

      //Clear it first
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          TPMDouble(TPAPointer(dAtdq[ec])[qq])[i, j] := 0;
        end;
      end;

      for k := 0 to StateCount - 1 do begin
        for l := 0 to StateCount - 1 do begin

          //M1 = Ak*dQdq
          MulMatMat(TPAPointer(Ak[ec])[k], dQdq[qq], Mat1, StateCount);//this can be optimized
          //M2 = M1*Al
          MulMatMat(Mat1, TPAPointer(Ak[ec])[l], Mat2, StateCount);

          Do_G_kl_t(k, l, ec, t, gklt);
          //M2 = M2*g_kl(t)
          for i := 0 to StateCount - 1 do begin
            for j := 0 to StateCount - 1 do begin
              Mat2[i, j] := Mat2[i, j] * gklt;
            end;
          end;
          //dAtdq = dAtdq + M2
          SumMatMat(TPAPointer(dAtdq[ec])[qq], Mat2, TPMDouble(TPAPointer(dAtdq[ec])[qq]), StateCount);

        end;//l
      end;//k

    end;//qq

  end;//ec

end;

function TQUB_Q_Model.Do_dAt_dq(t :Double; ExpCond :Integer) :Integer;
var
  i, j, k, l, qq, ec :Integer;
  //M                  :TPMDouble;
  gklt               :Double;
begin

  Result := 0;

  {
    Same as above, but it calculates only dAt[ExpCond] / dq[ExpCond], and zeroes the rest
  }

  for ec := 0 to ExpCondCount - 1 do begin

    for qq := 0 to RateCount - 1 do begin

      { moved it below, and only done when ExpCond = ec, the rest are always cleared
      //Clear it first
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          TPMDouble(TPAPointer(dAtdq[ec])[qq])[i, j] := 0;
        end;
      end;
      }

      if ExpCond = ec then begin

        //Clear it first
        for i := 0 to StateCount - 1 do begin
          for j := 0 to StateCount - 1 do begin
            TPMDouble(TPAPointer(dAtdq[ec])[qq])[i, j] := 0;
          end;
        end;

        for k := 0 to StateCount - 1 do begin
          for l := 0 to StateCount - 1 do begin

            //M1 = Ak*dQdq
            MulMatMat(TPAPointer(Ak[ec])[k], dQdq[qq], Mat1, StateCount);//this can be optimized
            //M2 = M1*Al
            MulMatMat(Mat1, TPAPointer(Ak[ec])[l], Mat2, StateCount);

            Do_G_kl_t(k, l, ec, t, gklt);
            //M2 = M2*g_kl(t)
            for i := 0 to StateCount - 1 do begin
              for j := 0 to StateCount - 1 do begin
                Mat2[i, j] := Mat2[i, j] * gklt;
              end;
            end;
            //dAtdq = dAtdq + M2
            SumMatMat(TPAPointer(dAtdq[ec])[qq], Mat2, TPMDouble(TPAPointer(dAtdq[ec])[qq]), StateCount);

          end;//l
        end;//k

      end;//ExpCond = ec

    end;//qq

  end;//ec
  
  (*
  ec := ExpCond;

  for qq := 0 to RateCount - 1 do begin

    M := TPMDouble(TPAPointer(dAtdq[ec])[qq]);

    //Clear it first
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        M[i, j] := 0;
      end;
    end;

    for k := 0 to StateCount - 1 do begin
      for l := 0 to StateCount - 1 do begin

        //M1 = Ak*dQdq
        MulMatMat(TPAPointer(Ak[ec])[k], dQdq[qq], Mat1, StateCount);//this can be optimized
        //M2 = M1*Al
        MulMatMat(Mat1, TPAPointer(Ak[ec])[l], Mat2, StateCount);

        Do_G_kl_t(k, l, ec, t, gklt);
        //M2 = M2*g_kl(t)
        for i := 0 to StateCount - 1 do begin
          for j := 0 to StateCount - 1 do begin
            Mat2[i, j] := Mat2[i, j] * gklt;
          end;
        end;
        //dAtdq = dAtdq + M2
        SumMatMat(M, Mat2, M, StateCount);

      end;//l
    end;//k

  end;//qq
  *)

end;



//****************************************************************************//
//                                                                            //
//                              P    V e c t o r                              //
//                                                                            //
//****************************************************************************//
function TQUB_Q_Model.Do_P0 :Integer;
begin

  Result := 0;

  MakeInitProbs(prModel, P0, StateCount);

end;

//Calculates the state vector at time t (relative to where P0 was calculated), using the At matrix for ExpCond
//At[ExpCond] should be calculated prior to this
function TQUB_Q_Model.Do_Pt(t :Double; ExpCond :Integer) :Integer;
begin

  Result := 0;

  MulVecMatVec(P0, At[ExpCond], Pt, StateCount);

end;

function TQUB_Q_Model.Do_Peq :Integer;
var
  R, RT   :TPMDouble;
  U       :TPADouble;
  i, j, k :Integer;
begin

  R       := nil;
  RT      := nil;
  U       := nil;


  //Peq = 1.(R.RT)-1

  try

    U       := DAlloc1(StateCount);
    R       := DAlloc2(StateCount,     StateCount + 1);
    RT      := DAlloc2(StateCount + 1, StateCount);

    //U
    for i := 0 to StateCount - 1 do begin
      U[i] := 1;
    end;

    //R, RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        R [i, j] := TPMDouble(Q[0{FirstExpCond}])[i, j];
        RT[j, i] := TPMDouble(Q[0{FirstExpCond}])[i, j];
      end;
    end;

    for k := 0 to StateCount - 1 do begin
      R [k, StateCount] := 1;
      RT[StateCount, k] := 1;
    end;

    //R.RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        Mat1[i, j] := 0;
        for k := 0 to StateCount do begin
          Mat1[i, j] := Mat1[i, j] + R[i, k]*RT[k, j];
        end;
      end;
    end;

    //RRTI = (R.RT)-1
    Result := InvMat(PMatrix(Mat1), 0, StateCount - 1, PMatrix(Mat2));

    MulVecMatVec(U, Mat2, Peq, StateCount);

  finally

    FreeMem(U);
    Free2(TPMChar(R));
    Free2(TPMChar(RT));

  end;

end;

function TQUB_Q_Model.Do_Cond_Peq :Integer;
var
  R, RT   :TPMDouble;
  U       :TPADouble;
  i, j, k :Integer;
begin

  R       := nil;
  RT      := nil;
  U       := nil;


  //Peq = 1.(R.RT)-1

  try

    U       := DAlloc1(StateCount);
    R       := DAlloc2(StateCount,     StateCount + 1);
    RT      := DAlloc2(StateCount + 1, StateCount);

    //U
    for i := 0 to StateCount - 1 do begin
      U[i] := 1;
    end;

    //R, RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        R [i, j] := Cond_Q[i, j]{CondExpCond};
        RT[j, i] := Cond_Q[i, j]{CondExpCond};
      end;
    end;

    for k := 0 to StateCount - 1 do begin
      R [k, StateCount] := 1;
      RT[StateCount, k] := 1;
    end;

    //R.RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        Mat1[i, j] := 0;
        for k := 0 to StateCount do begin
          Mat1[i, j] := Mat1[i, j] + R[i, k]*RT[k, j];
        end;
      end;
    end;

    //RRTI = (R.RT)-1
    Result := InvMat(PMatrix(Mat1), 0, StateCount - 1, PMatrix(Mat2));

    MulVecMatVec(U, Mat2, Cond_Peq, StateCount);

  finally

    FreeMem(U);
    Free2(TPMChar(R));
    Free2(TPMChar(RT));

  end;

end;


function TQUB_Q_Model.Do_dP0_dq :Integer;
var
  qq, k, ec :Integer;
begin

  Result := 0;

  for ec := 0 to ExpCondCount - 1 do begin
    for qq := 0 to RateCount - 1 do begin
      for k := 0 to StateCount - 1 do begin
        TPMDouble(dP0dq[ec])[qq, k] := 0;
      end;
    end;
  end;

end;

function TQUB_Q_Model.Do_dP0_dCond_q :Integer;
var
  qq, k :Integer;
begin

  Result := 0;

  for qq := 0 to RateCount - 1 do begin
    for k := 0 to StateCount - 1 do begin
      dP0dCond_q[qq, k] := 0;
    end;
  end;

end;

//Calculates the derivative of the state vector at time t, with respect to q for each exp. cond.
//ExpCond indicates the exp. cond. used to calculate Pt
function TQUB_Q_Model.Do_dPt_dq(ExpCond :Integer) :Integer;
var
  i, k, qq, ec  :Integer;
  a             :Double;
  M, M1, M2, M3 :TPMDouble;
begin

  Result := 0;

  {
    Pt         = P0.At[ExpCond]

    dPt/dq[ec] = (dP0/dq[ec]).At[ExpCond] + P0.(dAt[ExpCond]/dq[ec])

    dP0/dq[ec] is assumed to be already calculated

    dAt[ExpCond]/dq[ec]  = 0 if ExpCond <> ec
                        <> 0 if ExpCond  = ec
  }

  (*
  for ec := 0 to ExpCondCount - 1 do begin

    for qq := 0 to RateCount - 1 do begin

      for k := 0 to StateCount - 1 do begin

        TPMDouble(dPtdq[ec])[qq, k] := 0;

        //w.r.t. P0
        for i := 0 to StateCount - 1 do begin
          TPMDouble(dPtdq[ec])[qq, k]   := TPMDouble(dPtdq[ec])[qq, k] + TPMDouble(dP0dq[ec])[qq, i] * TPMDouble(At[ExpCond])[i, k];
        end;//i

        //w.r.t. A
        if ExpCond  = ec then begin
        //Only dAt[ExpCond] / dq[ExpCond] is non-zero
          for i := 0 to StateCount - 1 do begin
            TPMDouble(dPtdq[ec])[qq, k] := TPMDouble(dPtdq[ec])[qq, k] + P0[i]                       * TPMDouble(TPAPointer(dAtdq[ec])[qq])[i, k];
          end;//i                                                                                                           ^--must correspond to At[ExpCond]
        end;//ExpCond <> ec

      end;//k

    end;//qq

  end;//ec
  *)

  M1 := TPMDouble(At[ExpCond]);

  for ec := 0 to ExpCondCount - 1 do begin

    M  := TPMDouble(dPtdq[ec]);
    M2 := TPMDouble(dP0dq[ec]);

    for qq := 0 to RateCount - 1 do begin

      M3 := TPMDouble(TPAPointer(dAtdq[ec])[qq]);

      for k := 0 to StateCount - 1 do begin

        a := 0;

        //w.r.t. P0
        for i := 0 to StateCount - 1 do begin
          a := a + M2[qq, i] * M1[i, k];
        end;//i

        //w.r.t. A
        if ExpCond  = ec then begin
        //Only dAt[ExpCond] / dq[ExpCond] is non-zero
          for i := 0 to StateCount - 1 do begin
            a := a + P0[i] * M3[i, k];
          end;//i            ^--must correspond to At[ExpCond]
        end;//ExpCond <> ec

        M[qq, k] := a;

      end;//k

    end;//qq

  end;//ec

end;

//Calculates the derivative of the state vector at time t, with respect to q for conditioning exp. cond.
//ExpCond indicates the exp. cond. used to calculate Pt
function TQUB_Q_Model.Do_dPt_dCond_q(ExpCond :Integer) :Integer;
var
  i, k, qq :Integer;
  a        :Double;
  M        :TPMDouble;
begin

  Result := 0;

  {
    Pt         = P0.At[ExpCond]

    dPt/dCondq = (dP0/dCondq).At[ExpCond] + P0.(dAt[ExpCond]/dCondq)

    dP0/dCondq is already calculated

    dAt[ExpCond]/dCondq  = 0
  }

  (*
  for qq := 0 to RateCount - 1 do begin

    for k := 0 to StateCount - 1 do begin

      dPtdCond_q[qq, k] := 0;

      //w.r.t. P0
      for i := 0 to StateCount - 1 do begin
        dPtdCond_q[qq, k] := dPtdCond_q[qq, k] + dP0dCond_q[qq, i] * TPMDouble(At[ExpCond])[i, k];
      end;//i

      //w.r.t. A --> = 0 (not dependent)

    end;//k

  end;//qq
  *)

  M := TPMDouble(At[ExpCond]);
  for qq := 0 to RateCount - 1 do begin

    for k := 0 to StateCount - 1 do begin

      a := 0;

      //w.r.t. P0
      for i := 0 to StateCount - 1 do begin
        a := a + dP0dCond_q[qq, i] * M[i, k];
      end;//i

      //w.r.t. A --> = 0 (not dependent)

      dPtdCond_q[qq, k] := a;

    end;//k

  end;//qq

end;

function TQUB_Q_Model.Do_dPeq_dq :Integer;
var
  R, RT, RRT, RRTI, dRRTdq :TPMDouble;
  U                        :TPADouble;
  i, j, qq, k, ec          :Integer;
begin

  R       := nil;
  RT      := nil;
  RRT     := nil;
  RRTI    := nil;
  dRRTdq  := nil;
  U       := nil;


  //dP0/dq = -1.(R.RT)-1.(d(R.RT)/dq).(R.RT)-1

  try

    R       := DAlloc2(StateCount,     StateCount + 1);
    RT      := DAlloc2(StateCount + 1, StateCount);
    RRT     := DAlloc2(StateCount,     StateCount);
    RRTI    := DAlloc2(StateCount,     StateCount);
    dRRTdq  := DAlloc2(StateCount,     StateCount);
    U       := DAlloc1(StateCount);

    for i := 0 to StateCount - 1 do begin
      U[i] := -1;
    end;

    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        R [i, j] := TPMDouble(Q[0])[i, j];
        RT[j, i] := TPMDouble(Q[0])[i, j];
      end;
    end;

    //R, RT
    for k := 0 to StateCount - 1 do begin
      R [k, StateCount] := 1;
      RT[StateCount, k] := 1;
    end;

    //R.RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        RRT[i, j] := 0;
        for k := 0 to StateCount do begin
          RRT[i, j] := RRT[i, j] + R[i, k]*RT[k, j];
        end;
      end;
    end;

    //RRTI = (R.RT)-1
    Result := InvMat(PMatrix(RRT), 0, StateCount - 1, PMatrix(RRTI));

    //For each optimizable rate do begin
    for qq := 0 to RateCount - 1 do begin

      //d(R.RT)/dq = dRdq.RT + R.dRTdq

      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          dRRTdq[i, j] := 0;
          for k := 0 to StateCount do begin

            //dRdq.RT
            if k < StateCount then begin
              dRRTdq[i, j] := dRRTdq[i, j] + TPMDouble(dQdq[qq])[i, k]*RT[k, j];
            end else begin
              dRRTdq[i, j] := dRRTdq[i, j] +                         0*RT[k, j];
            end;

            //R.dRTdq
            if k < StateCount then begin
              dRRTdq[i, j] := dRRTdq[i, j] + R[i, k]*TPMDouble(dQdq[qq])[j, k];
            end else begin
              dRRTdq[i, j] := dRRTdq[i, j] + R[i, k]*0;
            end;

          end;
        end;
      end;

      //Mat1 = RRTI.dRRTdq
      MulMatMat(RRTI, dRRTdq, Mat1, StateCount);

      //dRRTdq = Mat1.RRTI
      MulMatMat(Mat1, RRTI, dRRTdq, StateCount);

      MulVecMatVec(U, dRRTdq, TPMDouble(dPeqdq[0])[qq], StateCount);


    end;//qq


  finally


    for ec := 1 to ExpCondCount - 1 do begin
      for qq := 0 to RateCount - 1 do begin
        for i := 0 to StateCount - 1 do begin
          TPMDouble(dPeqdq[ec])[qq][i] := 0;
        end;
      end;
    end;//qq


    FreeMem(U);
    Free2(TPMChar(R));
    Free2(TPMChar(RT));
    Free2(TPMChar(RRT));
    Free2(TPMChar(RRTI));
    Free2(TPMChar(dRRTdq));


  end;

end;

function TQUB_Q_Model.Do_dCond_Peq_dCond_q :Integer;
var
  R, RT, RRT, RRTI, dRRTdq :TPMDouble;
  U                        :TPADouble;
  i, j, qq, k              :Integer;
begin

  R       := nil;
  RT      := nil;
  RRT     := nil;
  RRTI    := nil;
  dRRTdq  := nil;
  U       := nil;


  //dPeq/dq = -1.(R.RT)-1.(d(R.RT)/dq).(R.RT)-1

  try

    R       := DAlloc2(StateCount,     StateCount + 1);
    RT      := DAlloc2(StateCount + 1, StateCount);
    RRT     := DAlloc2(StateCount,     StateCount);
    RRTI    := DAlloc2(StateCount,     StateCount);
    dRRTdq  := DAlloc2(StateCount,     StateCount);
    U       := DAlloc1(StateCount);

    for i := 0 to StateCount - 1 do begin
      U[i] := -1;
    end;

    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        R [i, j] := Cond_Q[i, j];
        RT[j, i] := Cond_Q[i, j];
      end;
    end;

    //R, RT
    for k := 0 to StateCount - 1 do begin
      R [k, StateCount] := 1;
      RT[StateCount, k] := 1;
    end;

    //R.RT
    for i := 0 to StateCount - 1 do begin
      for j := 0 to StateCount - 1 do begin
        RRT[i, j] := 0;
        for k := 0 to StateCount do begin
          RRT[i, j] := RRT[i, j] + R[i, k]*RT[k, j];
        end;
      end;
    end;

    //RRTI = (R.RT)-1
    Result := InvMat(PMatrix(RRT), 0, StateCount - 1, PMatrix(RRTI));

    //For each optimizable rate do begin
    for qq := 0 to RateCount - 1 do begin

      //d(R.RT)/dq = dRdq.RT + R.dRTdq

      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          dRRTdq[i, j] := 0;
          for k := 0 to StateCount do begin

            //dRdq.RT
            if k < StateCount then begin
              dRRTdq[i, j] := dRRTdq[i, j] + TPMDouble(dQdq[qq])[i, k]*RT[k, j];
            end else begin
              dRRTdq[i, j] := dRRTdq[i, j] +                         0*RT[k, j];
            end;

            //R.dRTdq
            if k < StateCount then begin
              dRRTdq[i, j] := dRRTdq[i, j] + R[i, k]*TPMDouble(dQdq[qq])[j, k];
            end else begin
              dRRTdq[i, j] := dRRTdq[i, j] + R[i, k]*0;
            end;

          end;
        end;
      end;

      //Mat1 = RRTI.dRRTdq
      MulMatMat(RRTI, dRRTdq, Mat1, StateCount);

      //dRRTdq = Mat1.RRTI
      MulMatMat(Mat1, RRTI, dRRTdq, StateCount);

      MulVecMatVec(U, dRRTdq, dCond_PeqdCond_q[qq], StateCount);

    end;//qq


  finally


    FreeMem(U);
    Free2(TPMChar(R));
    Free2(TPMChar(RT));
    Free2(TPMChar(RRT));
    Free2(TPMChar(RRTI));
    Free2(TPMChar(dRRTdq));


  end;

end;

procedure TQUB_Q_Model.PtToP0;
var
  k :Integer;
begin
  for k := 0 to StateCount - 1 do begin
    P0[k] := Pt[k];
  end;
end;

procedure TQUB_Q_Model.dPtdqTodP0dq;
var
  k, qq, ec :Integer;
begin

  //dPtdq[ec]
  for ec := 0 to ExpCondCount - 1 do begin
    for qq := 0 to RateCount - 1 do begin
      for k := 0 to StateCount - 1 do begin
        TPMDouble(dP0dq[ec])[qq][k] := TPMDouble(dPtdq[ec])[qq][k];
      end;//i
    end;//qq
  end;//ec

end;

procedure TQUB_Q_Model.dPtdCond_qTodP0dCond_q;
var
  s, qq :Integer;
begin

  for qq := 0 to RateCount - 1 do begin
    for s := 0 to StateCount - 1 do begin
      dP0dCond_q[qq][s] := dPtdCond_q[qq][s];
    end;//i
  end;//qq

end;


//****************************************************************************//
//                                                                            //
//                            C L A S S     I n f o                           //
//                                                                            //
//****************************************************************************//

{
function TQUB_Q_Model.Do_ClassCount :Integer;
var
  s           :Integer;
begin

  Result      := 0;

  ClassCount  := prMilModel.ClassCount;
  Classes     := IAlloc1(StateCount);
  Cls         := IAlloc1(StateCount);

  for s := 0 to StateCount - 1 do begin
    Cls[s] := prMilModel.ClassOfState[s];
    Classes[ Cls[s] ] := prModel.States[s].Cls;
  end;

end;
}

{

    this algorithm had trouble:
    in answer to the problem of sparse classes (e.g. cls 0 and 3 used, but 1 and 2 empty),
    it re-ordered the classes in order of first appearance.

    the mil model correctly condenses the classes into a contiguous list.
    rather than re-implement in delphi, we get the ordering from qubopt
    as a precursor to sharing the multi-channel model.
}
//{
function TQUB_Q_Model.Do_ClassCount :Integer;
var
  s, c        :Integer;
  Found       :Boolean;
begin

  Result      := 0;

  ClassCount  := 0;
  Classes     := IAlloc1(StateCount);
  Cls         := IAlloc1(StateCount);

  for s := 0 to StateCount - 1 do begin
    Cls[s] := prModel.States[s].Cls;
    Found := False;
    for c := 0 to ClassCount - 1 do begin
      if Classes[c] = prModel.States[s].Cls then begin
        Found := True;
        Break;
      end;
    end;
    if not Found then begin
      Classes[ClassCount] := prModel.States[s].Cls;
      Inc(ClassCount);
    end;
  end;

end;
//}



//****************************************************************************//
//                                                                            //
//                        L O G  -  L I K E L I H O O D                       //
//                                                                            //
//****************************************************************************//
function TQUB_Q_Model.Do_dLLdfPar :Integer;
var
  i, j, k, l, ec :Integer;
begin

  Result := 0;

  //Initialize:
  for k := 0 to fParCount - 1 do begin
    dLLdfPar[k] := 0.0;
  end;

  for k := 0 to RateCount - 1 do begin
    dLLdcPar[2*k + 0] := 0.0;
    dLLdcPar[2*k + 1] := 0.0;
  end;

  for k := 0 to ExtraKinParCount - 1 do begin
    dLLdcPar[k + 2*RateCount] := 0.0;
  end;

  //Conditioning part **********************************************************
  //We have dLLdCond_q, and need dCond_qdcPars
  for k := 0 to RateCount - 1 do begin

    i := c2qS[2*k + 0].x;
    j := c2qS[2*k + 0].y;

    if AllowNegRates then begin
      //cPars[even] = k0
      //k0 = cPars[even]
      //q = k0*C*exp(k1*V)
      //q = cPars[even]*C*exp(k1*V)
      //dq/dcPars[even] = C*exp(k1*V)
      dCond_qdcPar[2*k + 0] := Cond_ExpCondP[2*k + 0]*exp(Cond_Q[i, j] * Cond_ExpCondQ[2*k + 0]);
    end else begin
      //cPars[even] = ln(k0)
      //k0 = exp(cPars[even])
      //q = k0*C*exp(k1*V)
      //q = exp(cPars[even])*C*exp(k1*V)
      //dq/dcPars[even] = q
      dCond_qdcPar[2*k + 0] := Cond_Q[i, j];
    end;
    dCond_qdcPar[2*k + 1] := Cond_Q[i, j] * Cond_ExpCondQ[2*k + 1];

  end;

  //dLLdfPar
  for k := 0 to fParCount - 1 do begin

    for l := 0 to RateCount - 1 do begin

       dLLdfPar[k] := dLLdfPar[k] +
                        dLLdCond_q[l] * (dCond_qdcPar[2*l + 0]*f2cA[2*l + 0, k] + dCond_qdcPar[2*l + 1]*f2cA[2*l + 1, k]);

    end;

  end;

  //dLLdcPar
  for k := 0 to RateCount - 1 do begin

    if AllowNegRates then begin
      dLLdcPar[2*k + 0] := dLLdcPar[2*k + 0] +
                           dLLdCond_q[k] * dCond_qdcPar[2*k + 0] * 1;
    end else begin
      dLLdcPar[2*k + 0] := dLLdcPar[2*k + 0] +
                           dLLdCond_q[k] * dCond_qdcPar[2*k + 0] * (1/Exp(cPars[2*k + 0]));
    end;
    dLLdcPar[2*k + 1] := dLLdcPar[2*k + 1] +
                         dLLdCond_q[k] * dCond_qdcPar[2*k + 1] * 1;

  end;


  //Testing part ***************************************************************
  for ec := 0 to ExpCondCount - 1 do begin

    //We have dLLdq, and need dqdcPars
    for k := 0 to RateCount - 1 do begin

      i := c2qS[2*k + 0].x;
      j := c2qS[2*k + 0].y;


      if AllowNegRates then begin
        //cPars[even] = k0
        //k0 = cPars[even]
        //q = k0*C*exp(k1*V)
        //q = cPars[even]*C*exp(k1*V)
        //dq/dcPars[even] = C*exp(k1*V)

        dqdcPar[2*k + 0] := ExpCondP[ec][2*k + 0]*exp(TPMDouble(Q[ec])[i, j] * ExpCondQ[ec][2*k + 0]);
        //TO DO: is this correct?!                    ^^^^^^^^^^^^^^^^^^^^^^
        //shouldn't it be k1?!

      end else begin
        //cPars[even] = ln(k0)
        //k0 = exp(cPars[even])
        //q = k0*C*exp(k1*V)
        //q = exp(cPars[even])*C*exp(k1*V)
        //dq/dcPars[even] = q

        //If:
        //cPars[even] = ln(k0 + QMinRate)
        //k0 = exp(cPars[even]) - QMinRate
        //q = k0*C*exp(k1*V)
        //q = exp(cPars[even])*C*exp(k1*V) - QMinRate*C*exp(k1*V)
        //dq/dcPars[even] = q
        dqdcPar[2*k + 0] := TPMDouble(Q[ec])[i, j];
      end;

      dqdcPar[2*k + 1] := TPMDouble(Q[ec])[i, j] * ExpCondQ[ec][2*k + 1];

    end;


    //Used by the optimizer:

    //dLLdfPar
    for k := 0 to fParCount - 1 do begin

      {dLLdfPar[k] =
         Sum qq [
                 dLLdq[qq]      * ( dqdcPar[2*qq + 0]      * dcParsdfPar[2*qq + 0, k] + dqdcPar[2*qq + 1]      * dcParsdfPar[2*qq + 1, k] ) +
[moved above]    dLLdCond_q[qq] * ( dCond_qdcPar[2*qq + 0] * dcParsdfPar[2*qq + 0, k] + dCond_qdcPar[2*qq + 1] * dcParsdfPar[2*qq + 1, k] )
                ]
      }

      for l := 0 to RateCount - 1 do begin

         dLLdfPar[k] := dLLdfPar[k] + dLLdq [ec][l] * (dqdcPar[2*l + 0]*f2cA[2*l + 0, k] + dqdcPar[2*l + 1]*f2cA[2*l + 1, k]);

      end;

    end;

    //For display only:

    //dLLdcPar
    for k := 0 to RateCount - 1 do begin

                 //dLLdK{0|1} = dLLdq * dqdcPar * dcPardK{0|1}
//[moved above]  //dcPardK0 = dln(K0)/dK0 = 1 / K0 = 1 / Exp(cPar)
                 //dcPardK1 = 1

      if AllowNegRates then begin
        dLLdcPar[2*k + 0] := dLLdcPar[2*k + 0] +
                             dLLdq [ec][k] * dqdcPar[2*k + 0] * 1;
      end else begin
        dLLdcPar[2*k + 0] := dLLdcPar[2*k + 0] +
                             dLLdq [ec][k] * dqdcPar[2*k + 0] * (1/Exp(cPars[2*k + 0]));
      end;
      dLLdcPar[2*k + 1] := dLLdcPar[2*k + 1] +
                           dLLdq [ec][k] * dqdcPar [2*k + 1] * 1;

    end;

  end;

end;

function TQUB_Q_Model.Do_Varf2Varc :Integer;
var
  k, l :Integer;
begin

  Result := 0;

  //var(k0|k1) ~= Sum i (Sqr(dk{0|1}/dzi) * var(zi))
  //dk0/dz = dk0/dc0 * dc0/dz
  //dk1/dz =    1    * dc1/dz
  for k := 0 to RateCount - 1 do begin

    cParsVar[2*k + 0] := 0;
    cParsVar[2*k + 1] := 0;

    for l := 0 to fParCount - 1 do begin

      if AllowNegRates then begin
        cParsVar[2*k + 0] := cParsVar[2*k + 0] + Sqr(                    f2cA[2*k + 0, l]) * Abs(fParsVar[l, l]);
      end else begin
        cParsVar[2*k + 0] := cParsVar[2*k + 0] + Sqr(Exp(cPars[2*k + 0])*f2cA[2*k + 0, l]) * Abs(fParsVar[l, l]);
      end;
      cParsVar[2*k + 1]   := cParsVar[2*k + 1] + Sqr(                    f2cA[2*k + 1, l]) * Abs(fParsVar[l, l]);

    end;

  end;

  //var(ekp) ~= Sum i (Sqr(dekp/dzi) * var(zi))
  //dekp/dz = dekp/dc * dc/dz
  //  EKP_TYPE_LIN:
  //    dekp/dz = 1 * dc/dz
  //  EKP_TYPE_LOG:
  //    dekp/dz = dekp/dc * dc/dz         ekp = exp(c)  ==> dexp(c)/dc = exp(c)
  //  EKP_TYPE_EXP:
  //    dekp/dz = dekp/dc * dc/dz         ekp = ln(c)   ==> dln(c)/dc  = 1/c
  for k := 0 to ExtraKinParCount - 1 do begin

    cParsVar[2*RateCount + k] := 0;

    for l := 0 to fParCount - 1 do begin

      case TQUB_QFS_QMF_ExtraKineticParType(ExtraKinParTypes[k]) of
        EKP_TYPE_LIN:
          cParsVar[2*RateCount + k] := cParsVar[2*RateCount + k] + Sqr(                            f2cA[2*RateCount + k, l]) * Abs(fParsVar[l, l]);
        EKP_TYPE_LOG:
          cParsVar[2*RateCount + k] := cParsVar[2*RateCount + k] + Sqr(Exp(cPars[2*RateCount + k])*f2cA[2*RateCount + k, l]) * Abs(fParsVar[l, l]);
        EKP_TYPE_EXP:
          cParsVar[2*RateCount + k] := cParsVar[2*RateCount + k] + Sqr( 1/(cPars[2*RateCount + k])*f2cA[2*RateCount + k, l]) * Abs(fParsVar[l, l]);
        else
          cParsVar[2*RateCount + k] := cParsVar[2*RateCount + k] + Sqr(Exp(cPars[2*RateCount + k])*f2cA[2*RateCount + k, l]) * Abs(fParsVar[l, l]);
      end;

    end;

  end;

end;

function TQUB_Q_Model.Do_G_kl_t(k, l, ec :Integer; t :Double; var gklt :Double) :Integer;
begin

  Result := 0;

  if Lambda[ec, k] = Lambda[ec, l] then begin
    gklt := t * Exp(Lambda[ec, k] * t);
  end else begin
    gklt := (Exp(Lambda[ec, k] * t) - Exp(Lambda[ec, l] * t)) / (Lambda[ec, k] - Lambda[ec, l]);
  end;

end;

function TQUB_Q_Model.prGet_cParCount :Integer;
begin
  Result := 2*RateCount + ExtraKinParCount;
end;

function TQUB_Q_Model.prGet_ChannelCount :Integer;
begin
  Result := prModel.ChannelCount;
end;

function TQUB_Q_Model.prGet_AverageCount :Integer;
begin
  Result := 1;
end;

function TQUB_Q_Model.prGet_FitWeight :Double;
begin
  Result := 1;
end;


//******************************************************************************
end.

