{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2004
// edit and run a report (in Viewport: Results)


unit QUB_Reports_Editor;


interface


uses


Forms, Controls, StdCtrls, ComCtrls, Buttons, Classes, Sysutils, Filectrl,

QUB_Global,
PythonEngine,
QUB_Python_Thread,
QUB_QVO_QVR;



type


TQUB_Reports_EditForm = class(TForm)
  lstReport                           :TListBox;
  lstOptions                          :TListBox;
  btnAddOption                        :TSpeedButton;
  btnDelOption                        :TSpeedButton;
  btnUpOption                         :TSpeedButton;
  btnDnOption                         :TSpeedButton;

  btnClose                            :TButton;
  Label5                              :TLabel;
  Label6                              :TLabel;
  txtName                             :TEdit;
  btnSave                             :TButton;
  btnRun                              :TButton;
    btnHelp: TButton;


  procedure FormShow(Sender: TObject);

  procedure btnAddOptionClick(Sender: TObject);
  procedure btnDelOptionClick(Sender: TObject);
  procedure btnUpOptionClick(Sender: TObject);
  procedure btnDnOptionClick(Sender: TObject);

  procedure lstReportDblClick(Sender: TObject);
  procedure lstOptionsDblClick(Sender: TObject);
  procedure btnSaveClick(Sender: TObject);
  procedure btnRunClick(Sender: TObject);
    procedure txtNameChange(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);


protected

  prName              :String;
  prResults           :TQUB_QVO_QVR;
  prResultType        :String;

  function prGetName: String;
  procedure prSetName(x :String);
  procedure prSetOptions(x :TStrings);
  procedure prSetResults(x :TQUB_QVO_QVR);

  procedure prAddOption;
  procedure prDelOption;
  procedure prUpOption;
  procedure prDnOption;

public

  property ReportName   :String         read prGetName      write prSetName;
  property Options      :TStrings                           write prSetOptions;
  property Results      :TQUB_QVO_QVR   read prResults      write prSetResults;

end;


// we need to wrap up some data and hand it off to python for running in the python thread
TQUB_Reports_Task_Run = class(TQUB_Python_SyncTask)
protected

  prResults             :PPyObject;
  prOptions             :PPyObject;

public

  constructor Create(Results :TQUB_QVO_QVR; Options :TStrings);
  destructor Destroy; override;

  procedure Execute; override;

end;



implementation


{$R *.DFM}


uses

QUB_Viewport_Python,
QUB_Node_Python;



function TQUB_Reports_EditForm.prGetName: String;
begin
  Result := txtName.Text;
end;

procedure TQUB_Reports_EditForm.prSetName(x :String);
begin
  txtName.Text := x;
end;

procedure TQUB_Reports_EditForm.prSetOptions(x :TStrings);
begin
  lstReport.Items.Assign(x);
end;

procedure TQUB_Reports_EditForm.prSetResults(x :TQUB_QVO_QVR);
var
  PyOptions     :PPyObject;
begin
  prResults     := x;
  prResultType  := x.Results.Node['ResultType'].DataAsString;

  // populate lstOptions with Reports.ItemsForType[ prResultType ]
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    try
      PyOptions := EvalString( Format('Reports.ListItems("%s")', [prResultType]) );
      PyStrings_ToStrings(PyOptions, lstOptions.Items);
      Py_DECREF( PyOptions );
    except
    end;
  end;
  PythonThread.EndPython;
end;

// Options *******************************************************************
procedure TQUB_Reports_EditForm.prAddOption;
var
  i      :Integer;
  Option :String;
begin

  //Anything selected ?
  i := lstOptions.ItemIndex;
  if (i < 0) or (lstOptions.Items.Count = 0) then Exit;

  Option := lstOptions.Items[i];
  lstReport.Items.Add(Option);

  i := lstReport.Items.Count - 1;
  if lstReport.ItemIndex >= 0 then lstReport.Selected[lstReport.ItemIndex] := False;
  lstReport.ItemIndex   := i;
  lstReport.Selected[i] := True;

  btnSave.Enabled  := True;

end;

procedure TQUB_Reports_EditForm.prDelOption;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstReport.ItemIndex;
  if (i < 0) or (lstReport.Items.Count = 0) then Exit;

  lstReport.Items.Delete(i);

  if i >= lstReport.Items.Count then i := lstReport.Items.Count - 1;
  if i >= 0 then begin
    lstReport.ItemIndex   := i;
    lstReport.Selected[i] := True;
  end;

  btnSave.Enabled  := True;

end;

procedure TQUB_Reports_EditForm.prUpOption;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstReport.ItemIndex;
  if (i < 0) or (lstReport.Items.Count = 0) or (i = 0) then Exit;

  lstReport.Items.Exchange(i, i - 1);

  lstReport.Selected[i]     := False;
  lstReport.Selected[i - 1] := True;
  lstReport.ItemIndex       := i - 1;

  btnSave.Enabled  := True;

end;

procedure TQUB_Reports_EditForm.prDnOption;
var
  i  :Integer;
begin

  //Anything selected ?
  i := lstReport.ItemIndex;
  if (i < 0) or (lstReport.Items.Count = 0) or (i = (lstReport.Items.Count - 1)) then Exit;

  lstReport.Items.Exchange(i, i + 1);

  lstReport.Selected[i]     := False;
  lstReport.Selected[i + 1] := True;
  lstReport.ItemIndex       := i + 1;

  btnSave.Enabled  := True;

end;

//******************************************************************************
procedure TQUB_Reports_EditForm.FormShow(Sender: TObject);
begin

  btnSave.Enabled  := True;

end;

//******************************************************************************
procedure TQUB_Reports_EditForm.btnAddOptionClick(Sender: TObject);
begin
  prAddOption;
end;

procedure TQUB_Reports_EditForm.btnDelOptionClick(Sender: TObject);
begin
  prDelOption;
end;

procedure TQUB_Reports_EditForm.btnUpOptionClick(Sender: TObject);
begin
  prUpOption;
end;

procedure TQUB_Reports_EditForm.btnDnOptionClick(Sender: TObject);
begin
  prDnOption;
end;

procedure TQUB_Reports_EditForm.lstReportDblClick(Sender: TObject);
begin
  prDelOption;
end;

procedure TQUB_Reports_EditForm.lstOptionsDblClick(Sender: TObject);
begin
  prAddOption;
end;

procedure TQUB_Reports_EditForm.txtNameChange(Sender: TObject);
begin
  if txtName.Modified then begin
    btnSave.Enabled  := True;
    txtName.Modified := False;
  end;
end;

procedure TQUB_Reports_EditForm.btnSaveClick(Sender: TObject);
var
  Dir      :String;
begin
  Dir := ExtractFilePath(Application.ExeName) + 'Reports\' + prResultType;
  if ForceDirectories( Dir ) then begin
    try
      lstReport.Items.SaveToFile( Dir + '\' + ReportName + '.txt' );
      btnSave.Enabled := False;
    except
    end;
  end;

  // so this should really tell them if it didn't save and why:
  //    e.g. you can't save stuff in that location
  // but you really shouldn't be out of luck if you're not privileged ?!
end;

procedure TQUB_Reports_EditForm.btnRunClick(Sender: TObject);
var
  Task            :TQUB_Reports_Task_Run;
begin
  Task            := TQUB_Reports_Task_Run.Create(prResults, lstReport.Items);
  PythonThread.SyncPyTask(Task, True{Async -- auto-freed}, False{not Silent}, -1{no timeout});
end;

constructor TQUB_Reports_Task_Run.Create(Results :TQUB_QVO_QVR; Options :TStrings);
begin
  PythonThread.BeginPython;
  prResults     := Results.PyHandle.GetSelf;
  GetPythonEngine.Py_INCREF(prResults);
  prOptions     := PyStrings_FromStrings(Options);
  PythonThread.EndPython;
end;

destructor TQUB_Reports_Task_Run.Destroy;
begin
  PythonThread.BeginPython;
  GetPythonEngine.Py_DECREF(prResults);
  GetPythonEngine.Py_DECREF(prOptions);
  PythonThread.EndPython;
end;

procedure TQUB_Reports_Task_Run.Execute;
var
  RunList        :PPyObject;
  Args           :PPyObject;
begin
  Args           := nil;
  RunList        := nil;

  with GetPythonEngine do begin
    try
      Args       := Py_BuildValue('(OO)', [prOptions, prResults]);
      RunList    := EvalString('Reports.RunList');
      PyObject_CallObject(RunList, Args);
    except
    end;

    Py_XDECREF( Args );
    Py_XDECREF( RunList );
  end;
end;

procedure TQUB_Reports_EditForm.btnHelpClick(Sender: TObject);
begin
  ShowWiki('Results#Reports', Handle);
end;


//******************************************************************************
end.
