{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2004
// lists available reports (Viewport: Results) and lets you new/edit/delete


unit QUB_Reports_Manager;


interface


uses


Forms, Controls, StdCtrls, ComCtrls, Buttons, Classes, Sysutils, Filectrl,

QUB_Global,
QUB_QVO_QVR;



type


TQUB_Reports_ManageForm = class(TForm)
  lstReports                          :TListBox;
  btnClose                            :TButton;
  btnEdit                             :TButton;
  btnDelete                           :TButton;
  btnNew                              :TButton;
    btnHelp: TButton;

  procedure lstReportsDblClick(Sender: TObject);
  procedure btnEditClick(Sender: TObject);
  procedure btnDeleteClick(Sender: TObject);
  procedure btnNewClick(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);


protected

  prResults           :TQUB_QVO_QVR;
  prResultType        :String;

  procedure prSetResults(x :TQUB_QVO_QVR);

public

  property Results      :TQUB_QVO_QVR   read prResults      write prSetResults;

end;



implementation


{$R *.DFM}


uses

QUB_Reports_Editor;



procedure TQUB_Reports_ManageForm.prSetResults(x :TQUB_QVO_QVR);
var
  Dir            :String;
  sr             :TSearchRec;
begin
  prResults      := x;
  prResultType   := x.Results.Node['ResultType'].DataAsString;

  // clear and repopulate lstReports from Reports\ResultType\*.txt
  lstReports.Items.Clear;
  Dir := ExtractFilePath(Application.ExeName) + 'Reports\' + prResultType + '\';
  if 0 = FindFirst(Dir + '*.txt', 0, sr) then begin
    repeat
      lstReports.Items.Add( Copy(sr.Name, 0, Length(sr.Name) - 4) );
    until 0 <> FindNext(sr);
  end;
  Sysutils.FindClose(sr);
end;

//******************************************************************************
procedure TQUB_Reports_ManageForm.lstReportsDblClick(Sender: TObject);
begin
  btnEditClick(btnEdit);
end;

procedure TQUB_Reports_ManageForm.btnEditClick(Sender: TObject);
var
  ReportsForm         :TQUB_Reports_EditForm;
  Options             :TStringList;
begin
  if (lstReports.ItemIndex < 0) or (lstReports.Items.Count = 0) then Exit;

  ReportsForm         := TQUB_Reports_EditForm.Create(nil);
  Options             := TStringList.Create;
  try
    ReportsForm.Results := Results;
    ReportsForm.ReportName := lstReports.Items[ lstReports.ItemIndex ];

    Options.LoadFromFile(ExtractFilePath(Application.ExeName) + 'Reports\' + prResultType + '\' + lstReports.Items[ lstReports.ItemIndex ] + '.txt');
    ReportsForm.Options := Options;
    
    ReportsForm.ShowModal;
  finally
    Options.Free;
    ReportsForm.Release;
    Self.Results      := Results; // re-read list
  end;
end;

procedure TQUB_Reports_ManageForm.btnDeleteClick(Sender: TObject);
begin
  if (lstReports.ItemIndex < 0) or (lstReports.Items.Count = 0) then Exit;
  DeleteFile(ExtractFilePath(Application.ExeName) + 'Reports\' + prResultType + '\' + lstReports.Items[ lstReports.ItemIndex ] + '.txt');
  Self.Results    := Results; // re-read list
end;

procedure TQUB_Reports_ManageForm.btnNewClick(Sender: TObject);
var
  ReportsForm         :TQUB_Reports_EditForm;
begin
  ReportsForm         := TQUB_Reports_EditForm.Create(nil);
  ReportsForm.Results := Results;
  ReportsForm.ReportName := 'untitled';
  ReportsForm.ShowModal;
  ReportsForm.Release;
  
  Self.Results        := Results; // re-read list
end;

procedure TQUB_Reports_ManageForm.btnHelpClick(Sender: TObject);
begin
  ShowWiki('Results#Reports', Handle);
end;


//******************************************************************************
end.
