{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// specializations of TChart for the charts in Results: Select


unit QUB_Results_Chart;


interface


uses


Windows, SysUtils, Classes, Graphics, Menus, Controls, Messages, Printers, Registry, StdCtrls, ExtCtrls, ComCtrls, Dialogs, Buttons, Forms,
Chart, Series, TeEngine, TeeProcs,
Math,

AllocTypes, Alloc,

QUB_Global,

QUB_Node,
QUB_Events,

QUB_Results;


type

TQUB_Results_Trend = class
public
  X                     :INode;
  Y                     :INode;
  Color                 :Integer;
  Visible               :Boolean;
end;

TQUB_Results_Chart = class(TChart)
protected

  btnHide               :TSpeedButton;

  prPoints              :TList;
  prPointSeries         :TPointSeries;

  prLines               :TList;

  prTrends              :TList; // of TQUB_Results_Trend;
  prTrendSeries         :TList; // of TLineSeries;

  prChosen              :Integer;

  procedure prOnSeriesClick(Sender: TCustomChart; Series: TChartSeries; ValueIndex: Longint;
                            Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnZoom(Sender :TObject);

  procedure prBeforeDrawValues(Sender :TObject); virtual;
  procedure prAfterDrawValues(Sender :TObject);

  procedure prSetChosen(x :Integer);

  function prGetTrend(Ix :Integer): TQUB_Results_Trend;
  function prGetTrendCount: Integer;

public

  Lbl                   :String;

  OnClickPoint          :procedure(Sender :TQUB_Results_Chart; Ix :Integer) of object;
  OnSelectPoints        :procedure(Sender :TQUB_Results_Chart; Points :TList) of object;
  PointPopupMenu        :TPopupMenu;
  PopupPoint            :Integer; // valid OnPopup

  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure AddPoint(Point :TQUB_Results_Point);
  procedure RefreshPoints; virtual;
  procedure RefreshColors;

  procedure AddLine(Line :TQUB_Results_Line);
  procedure RefreshLines;

  procedure AddTrend(Trend :TQUB_Results_Trend);
  procedure ClearTrends;
  procedure RefreshTrends;

  property HideButton    :TSpeedButton        read btnHide;
  property ChosenPoint   :Integer             read prChosen        write prSetChosen;

  property Trends[Ix :Integer] :TQUB_Results_Trend  read prGetTrend;
  property TrendCount          :Integer             read prGetTrendCount;

end;


TQUB_Results_2Chart = class(TChart)
protected

  btnHide                :TSpeedButton;

  prXPoints              :TList;
  prXLines               :TList;

  prYPoints              :TList;
  prYLines               :TList;

  prPointSeries          :TPointSeries;
  prChosen               :Integer;

  prTrends              :TList; // of TQUB_Results_Trend;
  prTrendSeries         :TList; // of TLineSeries;

  procedure prOnSeriesClick(Sender: TCustomChart; Series: TChartSeries; ValueIndex: Longint;
                            Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnZoom(Sender :TObject);

  procedure prBeforeDrawValues(Sender :TObject);
  procedure prAfterDrawValues(Sender :TObject);

  procedure prSetChosen(x :Integer);

  function prGetTrend(Ix :Integer): TQUB_Results_Trend;
  function prGetTrendCount: Integer;

public

  Lbl                   :String;

  OnClickPoint          :procedure(Sender :TQUB_Results_2Chart; Ix :Integer) of object;
  OnSelectPoints        :procedure(Sender :TQUB_Results_2Chart; Points :TList) of object;
  PointPopupMenu        :TPopupMenu;
  PopupPoint            :Integer; // valid OnPopup

  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure AddPoint(XPoint, YPoint :TQUB_Results_Point);
  procedure RefreshPoints;
  procedure RefreshColors;

  procedure AddXLine(Line :TQUB_Results_Line);
  procedure AddYLine(Line :TQUB_Results_Line);
  procedure RefreshLines;

  procedure AddTrend(Trend :TQUB_Results_Trend);
  procedure ClearTrends;
  procedure RefreshTrends;

  procedure SwapAxes; virtual;

  property HideButton    :TSpeedButton       read btnHide;
  property ChosenPoint   :Integer            read prChosen        write prSetChosen;

  property Trends[Ix :Integer] :TQUB_Results_Trend  read prGetTrend;
  property TrendCount          :Integer             read prGetTrendCount;

end;


implementation

uses

QUB_QFS_Channels;


constructor TQUB_Results_Chart.Create(AOwner :TComponent);
begin
  inherited;

  btnHide                     := TSpeedButton.Create(Self);
  btnHide.Parent              := Self;
  btnHide.Caption             := 'X';
  btnHide.Width               := 15;
  btnHide.Height              := 15;
  btnHide.Top                 := 0;
  btnHide.Left                := Width - btnHide.Width;
  btnHide.Anchors             := [akTop, akRight];

  prPoints                           := TList.Create;
  prPointSeries                      := TPointSeries.Create(Self);
  prPointSeries.XValues.Order        := loNone;
  prPointSeries.ParentChart          := Self;
  prPointSeries.Pointer.Style        := psCircle;
  prPointSeries.Pointer.HorizSize    := 5;
  prPointSeries.Pointer.VertSize     := 5;
  prPointSeries.ColorEachPoint       := True;

  prLines                     := TList.Create;

  prChosen                    := -1;

  prTrends                    := TList.Create;
  prTrendSeries               := TList.Create;

  OnZoom                      := prOnZoom;
  OnClickSeries               := prOnSeriesClick;

  prPointSeries.BeforeDrawValues     := prBeforeDrawValues;
  prPointSeries.AfterDrawValues      := prAfterDrawValues;

end;

destructor TQUB_Results_Chart.Destroy;
var
  i             :Integer;
begin
  try
    prPoints.Free;
    prLines.Free;
    for i := 0 to prTrends.Count - 1 do begin
      TQUB_Results_Trend(prTrends[i]).Free;
    end;
    prTrends.Free;
    prTrendSeries.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Results_Chart.AddPoint(Point :TQUB_Results_Point);
begin
  prPoints.Add( Point );
  prPointSeries.AddXY( Point.XValue, Point.YValue, '', QUB_QFS_QMF_CLASS_COLORS[ Min(QUB_SELECT_COLOR_COUNT, Point.Color) ]);
end;

procedure TQUB_Results_Chart.RefreshPoints;
var
  i           :Integer;
  Point       :TQUB_Results_Point;
begin
  for i := 0 to prPoints.Count - 1 do begin
    Point := TQUB_Results_Point( prPoints[i] );
    prPointSeries.XValues[i]    := Point.XValue;
    prPointSeries.YValues[i]    := Point.YValue;
  end;

  if prPointSeries.MinYValue = prPointSeries.MaxYValue then begin
    prPointSeries.YValues[ prPoints.Count - 1 ] := prPointSeries.MinYValue + prPointSeries.MinYValue * 1.0e-6;
  end;

  UndoZoom;
end;

procedure TQUB_Results_Chart.RefreshColors;
var
  i       :Integer;
  Point   :TQUB_Results_Point;
begin
  for i := 0 to prPoints.Count - 1 do begin
    Point := TQUB_Results_Point( prPoints[i] );
    prPointSeries.ValueColor[i] := QUB_QFS_QMF_CLASS_COLORS[ Min(QUB_SELECT_COLOR_COUNT, Point.Color) ];
  end;

  Repaint;
end;

procedure TQUB_Results_Chart.AddLine(Line :TQUB_Results_Line);
begin
  prLines.Add( Line );
end;

procedure TQUB_Results_Chart.RefreshLines;
begin
  Repaint;
end;

function TQUB_Results_Chart.prGetTrend(Ix :Integer): TQUB_Results_Trend;
begin
  if (Ix >= 0) and (Ix < TrendCount) then begin
    Result := TQUB_Results_Trend( prTrends[Ix] );
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results_Chart.prGetTrendCount: Integer;
begin
  Result := prTrends.Count;
end;

procedure TQUB_Results_Chart.AddTrend(Trend :TQUB_Results_Trend);
var
  Series                    :TLineSeries;
  i                         :Integer;
begin
  prTrends.Add(Trend);

  Series                    := TLineSeries.Create(Self);
  prTrendSeries.Add(Series);

  Series.XValues.Order      := loNone;
  Series.Marks.Visible      := False;
  Series.LinePen.Width      := 3;//Draw a thick line

  with Trend do begin
    for i := 0 to X.DataCount - 1 do begin
      Series.AddXY(Double(X.Data[i]), Double(Y.Data[i]), '', QUB_QFS_QMF_CLASS_COLORS[Min(QUB_SELECT_COLOR_COUNT, Color)]);
    end;
  end;

  prPointSeries.ParentChart := nil;
  Series.ParentChart        := Self;
  prPointSeries.ParentChart := Self;   // make sure the points are in front

  if prTrends.Count = 1 then begin // first trend added
    prPointSeries.BeforeDrawValues := nil;
    Series.BeforeDrawValues        := prBeforeDrawValues;
  end;
end;

procedure TQUB_Results_Chart.ClearTrends;
var
  i                      :Integer;
begin
  for i := 0 to TrendCount - 1 do begin
    Trends[i].Free;
    TLineSeries(prTrendSeries[i]).Free;
  end;
  prTrends.Clear;
  prTrendSeries.Clear;

  prPointSeries.BeforeDrawValues := prBeforeDrawValues;
  UndoZoom;
end;

// need to modify to make sure the # of points in the trend hasn't changed?

procedure TQUB_Results_Chart.RefreshTrends;
var
  i, j                    :Integer;
  Series                  :TLineSeries;
begin
  for i := 0 to TrendCount - 1 do begin
    Series                := TLineSeries( prTrendSeries[i] );
    with Trends[i] do begin
      Series.Active       := Visible;
      for j := 0 to Series.Count - 1 do begin
        Series.XValues[j]   := Double(X.Data[j]);
        Series.YValues[j]   := Double(Y.Data[j]);
      end;
    end;
  end;
  UndoZoom;
end;

procedure TQUB_Results_Chart.prOnSeriesClick(Sender: TCustomChart; Series: TChartSeries; ValueIndex: Longint;
                          Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Point   :TPoint;
begin
  if not (Series is TPointSeries) then Exit;

  if ssRight in Shift then begin
    if PointPopupMenu <> nil then begin
      PopupPoint := ValueIndex;

      Point.X := X;
      Point.Y := Y;
      Point   := Sender.ClientToScreen( Point );

      PointPopupMenu.PopupComponent := Self;
      PointPopupMenu.Popup( Point.X, Point.Y );
    end;
  end else begin
    if Assigned(OnClickPoint) then OnClickPoint(Self, ValueIndex);
  end;
end;

procedure TQUB_Results_Chart.prOnZoom(Sender :TObject);
var
  i           :Integer;
  Points      :TList;
  XMin, XMax  :Double;
  Min, Max    :Double;
begin
  if not Assigned( OnSelectPoints ) then Exit;

  XMin    := BottomAxis.Minimum;
  XMax    := BottomAxis.Maximum;
  Min     := LeftAxis.Minimum;
  Max     := LeftAxis.Maximum;

  Points := TList.Create;

  for i := prPointSeries.FirstValueIndex to prPointSeries.LastValueIndex do begin
    if (prPointSeries.XValue[i] >= XMin) and (prPointSeries.XValue[i] <= XMax)
    and (prPointSeries.YValue[i] >= Min) and (prPointSeries.YValue[i] <= Max) then begin
      Points.Add( Pointer(i) );
    end;
  end;

  OnSelectPoints(Self, Points);
  Points.Free;

  UndoZoom;
end;

procedure TQUB_Results_Chart.prSetChosen(x :Integer);
begin
  prChosen := x;
  Repaint;
end;

procedure TQUB_Results_Chart.prBeforeDrawValues(Sender :TObject);
var
  i                :Integer;
  Line             :TQUB_Results_Line;
  X1, X2, Y        :Integer;
begin
  X1 := ChartRect.Left+1; // prPointSeries.CalcXPosValue( prPointSeries.MinXValue );
  X2 := ChartRect.Right-1; // prPointSeries.CalcXPosValue( prPointSeries.MaxXValue );

  for i := 0 to prLines.Count - 1 do begin
    Line           := TQUB_Results_Line( prLines[i] );
    if not Line.Visible then Continue;

    if Line.Color > 0 then begin
      Canvas.Pen.Color             := QUB_QFS_QMF_CLASS_COLORS[ Min(QUB_SELECT_COLOR_COUNT, Line.Color) ];
    end else begin
      Canvas.Pen.Color             := clGray;
    end;

    Y                              := prPointSeries.CalcYPosValue( Line.Value );

    if (Y > ChartRect.Top) and (Y < ChartRect.Bottom) then begin
      Canvas.Pen.Width               := Line.Width;
      Canvas.Line( X1, Y, X2, Y );
    end;
  end;
end;

procedure TQUB_Results_Chart.prAfterDrawValues(Sender :TObject);
var
  X, Y    :Integer;
begin
  if (prChosen >= 0) and (prChosen < prPointSeries.Count) then begin
    X     := prPointSeries.CalcXPos( prChosen );
    Y     := prPointSeries.CalcYPos( prChosen );

    Canvas.Pen.Color   := clWhite;
    Canvas.Brush.Color := clWhite;
    Canvas.Ellipse( X-2, Y-2, X+2, Y+2 );
  end;
end;

// *****************************************************************************
constructor TQUB_Results_2Chart.Create(AOwner :TComponent);
begin
  inherited;

  btnHide                            := TSpeedButton.Create(Self);
  btnHide.Parent                     := Self;
  btnHide.Caption                    := 'X';
  btnHide.Width                      := 15;
  btnHide.Height                     := 15;
  btnHide.Top                        := 0;
  btnHide.Left                       := Width - btnHide.Width;
  btnHide.Anchors                    := [akTop, akRight];

  prXPoints                          := TList.Create;
  prXLines                           := TList.Create;
  prYPoints                          := TList.Create;
  prYLines                           := TList.Create;

  prPointSeries                      := TPointSeries.Create(Self);
  prPointSeries.ParentChart          := Self;
  prPointSeries.Pointer.Style        := psCircle;
  prPointSeries.Pointer.HorizSize    := 5;
  prPointSeries.Pointer.VertSize     := 5;
  prPointSeries.ColorEachPoint       := True;

  prTrends                    := TList.Create;
  prTrendSeries               := TList.Create;

  prChosen                           := -1;

  OnZoom                             := prOnZoom;
  OnClickSeries                      := prOnSeriesClick;

  prPointSeries.BeforeDrawValues     := prBeforeDrawValues;
  prPointSeries.AfterDrawValues      := prAfterDrawValues;

end;

destructor TQUB_Results_2Chart.Destroy;
var
  i                :Integer;
begin
  try
    prXPoints.Free;
    prYPoints.Free;
    prXLines.Free;
    prYLines.Free;

    for i := 0 to prTrends.Count - 1 do begin
      TQUB_Results_Trend(prTrends[i]).Free;
    end;
    prTrends.Free;
    prTrendSeries.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Results_2Chart.AddPoint(XPoint, YPoint :TQUB_Results_Point);
begin
  prXPoints.Add( XPoint );
  prYPoints.Add( YPoint );
  prPointSeries.AddXY( XPoint.YValue, YPoint.YValue, '', QUB_QFS_QMF_CLASS_COLORS[ Min(QUB_SELECT_COLOR_COUNT, XPoint.Color) ]);
end;

procedure TQUB_Results_2Chart.RefreshPoints;
var
  i            :Integer;
  XPoint       :TQUB_Results_Point;
  YPoint       :TQUB_Results_Point;
begin
  for i := 0 to Min(prXPoints.Count, prYPoints.Count) - 1 do begin
    XPoint := TQUB_Results_Point( prXPoints[i] );
    YPoint := TQUB_Results_Point( prYPoints[i] );
    prPointSeries.XValues[i]    := XPoint.YValue;
    prPointSeries.YValues[i]    := YPoint.YValue;
  end;

  if prPointSeries.MinYValue = prPointSeries.MaxYValue then begin
    prPointSeries.YValues[ prYPoints.Count - 1 ] := prPointSeries.MinYValue + prPointSeries.MinYValue * 1.0e-6;
  end;

  if prPointSeries.MinXValue = prPointSeries.MaxXValue then begin
    prPointSeries.XValues[ prXPoints.Count - 1 ] := prPointSeries.MinXValue + prPointSeries.MinXValue * 1.0e-6;
  end;

  UndoZoom;
end;

procedure TQUB_Results_2Chart.RefreshColors;
var
  i       :Integer;
  Point   :TQUB_Results_Point;
begin
  for i := 0 to prXPoints.Count - 1 do begin
    Point := TQUB_Results_Point( prXPoints[i] );
    prPointSeries.ValueColor[i] := QUB_QFS_QMF_CLASS_COLORS[ Min(QUB_SELECT_COLOR_COUNT, Point.Color) ];
  end;

  Repaint;
end;

procedure TQUB_Results_2Chart.AddXLine(Line :TQUB_Results_Line);
begin
  prXLines.Add( Line );
end;

procedure TQUB_Results_2Chart.AddYLine(Line :TQUB_Results_Line);
begin
  prYLines.Add( Line );
end;

procedure TQUB_Results_2Chart.RefreshLines;
begin
  Repaint;
end;

function TQUB_Results_2Chart.prGetTrend(Ix :Integer): TQUB_Results_Trend;
begin
  if (Ix >= 0) and (Ix < TrendCount) then begin
    Result := TQUB_Results_Trend( prTrends[Ix] );
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results_2Chart.prGetTrendCount: Integer;
begin
  Result := prTrends.Count;
end;

procedure TQUB_Results_2Chart.AddTrend(Trend :TQUB_Results_Trend);
var
  Series                    :TLineSeries;
  i                         :Integer;
begin
  prTrends.Add(Trend);

  Series                    := TLineSeries.Create(Self);
  prTrendSeries.Add(Series);

  Series.XValues.Order      := loNone;
  Series.Marks.Visible      := False;
  Series.LinePen.Width      := 3;//Draw a thick line

  with Trend do begin
    for i := 0 to X.DataCount - 1 do begin
      Series.AddXY(Double(X.Data[i]), Double(Y.Data[i]), '', QUB_QFS_QMF_CLASS_COLORS[Min(QUB_SELECT_COLOR_COUNT, Color)]);
    end;
  end;

  prPointSeries.ParentChart := nil;
  Series.ParentChart        := Self;
  prPointSeries.ParentChart := Self;   // make sure the points are in front
end;

procedure TQUB_Results_2Chart.ClearTrends;
var
  i                      :Integer;
begin
  for i := 0 to TrendCount - 1 do begin
    Trends[i].Free;
    TLineSeries(prTrendSeries[i]).Free;
  end;
  prTrends.Clear;
  prTrendSeries.Clear;
  UndoZoom;
end;

// need to modify to make sure the # of points in the trend hasn't changed?

procedure TQUB_Results_2Chart.RefreshTrends;
var
  i, j                    :Integer;
  Series                  :TLineSeries;
begin
  for i := 0 to TrendCount - 1 do begin
    Series                := TLineSeries( prTrendSeries[i] );
    with Trends[i] do begin
      Series.Active       := Visible;
      for j := 0 to Series.Count - 1 do begin
        Series.XValues[j]   := Double(X.Data[j]);
        Series.YValues[j]   := Double(Y.Data[j]);
      end;
    end;
  end;
  UndoZoom;
end;

procedure TQUB_Results_2Chart.SwapAxes;
var
  TmpList                  :TList;
begin
  TmpList                  := prXPoints;
  prXPoints                := prYPoints;
  prYPoints                := TmpList;

  TmpList                  := prXLines;
  prXLines                 := prYLines;
  prYLines                 := TmpList;

  RefreshPoints;
  RefreshLines;
end;

procedure TQUB_Results_2Chart.prOnSeriesClick(Sender: TCustomChart; Series: TChartSeries; ValueIndex: Longint;
                          Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Point   :TPoint;
begin
  if not (Series is TPointSeries) then Exit;

  if ssRight in Shift then begin
    if PointPopupMenu <> nil then begin
      PopupPoint := ValueIndex;

      Point.X := X;
      Point.Y := Y;
      Point   := Sender.ClientToScreen( Point );

      PointPopupMenu.PopupComponent := Self;
      PointPopupMenu.Popup( Point.X, Point.Y );
    end;
  end else begin
    if Assigned(OnClickPoint) then OnClickPoint(Self, ValueIndex);
  end;
end;

procedure TQUB_Results_2Chart.prOnZoom(Sender :TObject);
var
  i           :Integer;
  Points      :TList;
  XMin, XMax  :Double;
  Min, Max    :Double;
begin
  if not Assigned( OnSelectPoints ) then Exit;

  XMin    := BottomAxis.Minimum;
  XMax    := BottomAxis.Maximum;
  Min     := LeftAxis.Minimum;
  Max     := LeftAxis.Maximum;

  Points := TList.Create;

  for i := prPointSeries.FirstValueIndex to prPointSeries.LastValueIndex do begin
    if (prPointSeries.XValue[i] >= XMin) and (prPointSeries.XValue[i] <= XMax)
    and (prPointSeries.YValue[i] >= Min) and (prPointSeries.YValue[i] <= Max) then begin
      Points.Add( Pointer(i) );
    end;
  end;

  OnSelectPoints(Self, Points);
  Points.Free;

  UndoZoom;
end;

procedure TQUB_Results_2Chart.prSetChosen(x :Integer);
begin
  prChosen := x;
  Repaint;
end;

procedure TQUB_Results_2Chart.prBeforeDrawValues(Sender :TObject);
var
  i                :Integer;
  Line             :TQUB_Results_Line;
  X1, X2, Y        :Integer;
  Y1, Y2, X        :Integer;
begin
  X1 := ChartRect.Left+1; // prPointSeries.CalcXPosValue( prPointSeries.MinXValue );
  X2 := ChartRect.Right-1; // prPointSeries.CalcXPosValue( prPointSeries.MaxXValue );
  Y1 := ChartRect.Top+1; // prPointSeries.CalcYPosValue( prPointSeries.MinYValue );
  Y2 := ChartRect.Bottom-1; // prPointSeries.CalcYPosValue( prPointSeries.MaxYValue );

  for i := 0 to prXLines.Count - 1 do begin
    Line           := TQUB_Results_Line( prXLines[i] );
    if not Line.Visible then Continue;

    if Line.Color > 0 then begin
      Canvas.Pen.Color             := QUB_QFS_QMF_CLASS_COLORS[ Min(QUB_SELECT_COLOR_COUNT, Line.Color) ];
    end else begin
      Canvas.Pen.Color             := clGray;
    end;

    X                              := prPointSeries.CalcXPosValue( Line.Value );

    if (X > ChartRect.Left) and (X < ChartRect.Right) then begin
      Canvas.Pen.Width               := Line.Width;
      Canvas.Line( X, Y1, X, Y2 );
    end;
  end;

  for i := 0 to prYLines.Count - 1 do begin
    Line           := TQUB_Results_Line( prYLines[i] );
    if not Line.Visible then Continue;

    if Line.Color > 0 then begin
      Canvas.Pen.Color             := QUB_QFS_QMF_CLASS_COLORS[ Min(QUB_SELECT_COLOR_COUNT, Line.Color) ];
    end else begin
      Canvas.Pen.Color             := clGray;
    end;

    Y                              := prPointSeries.CalcYPosValue( Line.Value );

    if (Y > ChartRect.Top) and (Y < ChartRect.Bottom) then begin
      Canvas.Pen.Width               := Line.Width;
      Canvas.Line( X1, Y, X2, Y );
    end;
  end;
end;

procedure TQUB_Results_2Chart.prAfterDrawValues(Sender :TObject);
var
  X, Y    :Integer;
begin
  if (prChosen >= 0) and (prChosen < prPointSeries.Count) then begin
    X     := prPointSeries.CalcXPos( prChosen );
    Y     := prPointSeries.CalcYPos( prChosen );

    Canvas.Pen.Color   := clWhite;
    Canvas.Brush.Color := clWhite;
    Canvas.Ellipse( X-2, Y-2, X+2, Y+2 );
  end;
end;

// *****************************************************************************
end.

