{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the Criteria tab of QUB Visual Results


unit QUB_Results_Criteria;


interface


uses


Windows, SysUtils, Classes, Graphics, Menus, Controls, Messages, Printers, Registry, StdCtrls, ExtCtrls, ComCtrls, Dialogs, Buttons, Forms,

AllocTypes, Alloc,

QUB_Global,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python,
QUB_Events,

QUB_Results,
QUB_PyCmd_MenuButton,
QUB_ColorPanel;


const

QUB_SELCRIT_LINES = 10;


type

TQUB_SelCritLine = class(TPanel)

protected

  mnuVars                          :TComboBox;
  mnuOps                           :TComboBox;
  txtExpr                          :TEdit;
  panColors                        :TQUB_ColorPanel;
  prValid                          :Boolean;         // whether a var is selected

  procedure prSetVars( Results :TQUB_Results );
  procedure prOnVarClick(Sender :TObject);
  procedure prOnResize(Sender :TObject);

public

  OnCritChange                     :procedure (Sender :TObject; Valid :Boolean) of object;

  constructor Create(AOwner :TComponent); override;

  function AsNode: INode;
  procedure SetFromNode(Crit :INode);

  property Results                 :TQUB_Results                        write prSetVars;
  property Valid                   :Boolean            read prValid;

end;


TQUB_Results_CriteriaPanel = class(TPanel)
protected

  lblCrit                                      :TLabel;
  panCritColors                                :TQUB_ColorPanel;
  panCrits                                     :TScrollBox;
  panCritCtrls                                 :TPanel;
  prCritLines                                  :TList;
  mnuDoCrit                                    :TQUB_PyCmd_MenuButton;

  procedure prOnResize(Sender :TObject);
  procedure prOnColorClick(Sender: TObject; Color :Integer);
  procedure prOnDoCritMouseUp(Sender: TObject; MouseButton: TMouseButton; Shift: TShiftState; X, Y: Integer);

  function prGetChosenColor: Integer;
  procedure prSetChosenColor(x :Integer);

public

  OnAction                                     :TNotifyEvent;
  OnColorChange                                :TNotifyEvent;

  procedure Initialize(Results :TQUB_Results);
  destructor Destroy; override;

  function AsNode: INode;

  property ChosenColor                         :Integer          read prGetChosenColor    write prSetChosenColor;

end;


implementation


constructor TQUB_SelCritLine.Create(AOwner :TComponent);
begin
  inherited;

  mnuVars         := TComboBox.Create(Self);
  mnuVars.Parent  := Self;
  mnuVars.Style   := csDropDownList;
  mnuVars.OnClick := prOnVarClick;
  mnuVars.DropDownCount     := 20;
  //mnuVars.Anchors := [akTop, akBottom, akLeft];

  Height          := mnuVars.Height + 2;

  mnuOps          := TComboBox.Create(Self);
  mnuOps.Parent   := Self;
  mnuOps.Style    := csDropDownList;
  mnuOps.Visible  := False;
  //mnuOps.Anchors  := [akTop, akBottom, akLeft];

  txtExpr         := TEdit.Create(Self);
  txtExpr.Parent  := Self;
  txtExpr.Visible := False;
  //txtExpr.Anchors := [akTop, akBottom, akLeft];

  panColors              := TQUB_ColorPanel.CreateCounted(Self, QUB_SELECT_COLOR_COUNT);
  panColors.Parent       := Self;
  panColors.Orientation  := QUB_COLORPANEL_HORIZ;
  panColors.ColorWidth   := Height;
  panColors.Visible      := False;
  //panColors.Anchors      := [akTop, akBottom, akLeft];

  prValid         := False;

  OnResize        := prOnResize;
end;

procedure TQUB_SelCritLine.prSetVars( Results :TQUB_Results );
var
  i           :Integer;
begin
  mnuVars.Clear;
  mnuVars.Items.Add('---');
  mnuVars.Items.Add('Color');

  for i := 0 to Results.SelVarCount - 1 do begin
    mnuVars.Items.Add( Results.SelVar[i].Name );
  end;

  mnuVars.ItemIndex := 0;
end;

procedure TQUB_SelCritLine.prOnVarClick(Sender :TObject);
var
  Name            :String;
begin
  Name            := mnuVars.Text;

  if Name = '---' then begin

    prValid            := False;
    mnuOps.Visible     := False;
    txtExpr.Visible    := False;
    panColors.Visible  := False;

  end else if Name = 'Color' then begin

    prValid            := True;
    mnuOps.Visible     := True;
    txtExpr.Visible    := False;
    panColors.Visible  := True;

    mnuOps.Clear;
    mnuOps.Items.Add('is');
    mnuOps.Items.Add('is not');
    mnuOps.ItemIndex := 0;

  end else begin

    prValid           := True;
    mnuOps.Visible    := True;
    txtExpr.Visible   := True;
    panColors.Visible := False;

    mnuOps.Clear;
    mnuOps.Items.Add('<');
    mnuOps.Items.Add('<=');
    mnuOps.Items.Add('>');
    mnuOps.Items.Add('>=');
    mnuOps.ItemIndex := 0;

  end;

  if Assigned(OnCritChange) then OnCritChange(Self, prValid);
end;

procedure TQUB_SelCritLine.prOnResize(Sender :TObject);
begin
  mnuVars.Left    := 0;
  mnuVars.Width   := Width div 3;

  Height          := mnuVars.Height + 2;

  mnuOps.Left     := mnuVars.Width;
  mnuOps.Width    := mnuVars.Width;

  txtExpr.Left    := mnuOps.Left + mnuOps.Width;
  txtExpr.Width   := mnuOps.Width;

  panColors.Left         := mnuOps.Left + mnuOps.Width;
  panColors.Width        := Width - panColors.Left;
  panColors.ColorWidth   := Height;
end;

function TQUB_SelCritLine.AsNode: INode;
var
  N      :INode;
begin
  Result := TNode.Create('Criterion');
  if prValid then begin
    N    := TNode.Create('');
    N.SetData( mnuVars.Text );
    Result.AppendChild( N );

    N    := TNode.Create('');
    N.SetData( mnuOps.Text );
    Result.AppendChild( N );

    N    := TNode.Create('');
    if mnuVars.Text = 'Color' then begin
      N.SetData( panColors.Chosen );
    end else begin
      N.SetData( txtExpr.Text );
    end;
    Result.AppendChild( N );
  end;
end;

procedure TQUB_SelCritLine.SetFromNode(Crit :INode);
var
  s       :String;
  n       :INode;
  i       :Integer;
begin
  if Crit.Child.IsNull then begin
    mnuVars.ItemIndex := 0;
    prOnVarClick(mnuVars);
  end else begin
    n := Crit.Child;
    s := n.DataAsString;
    i := mnuVars.Items.IndexOf(s);
    if i >= 0 then begin
      mnuVars.ItemIndex := i;
      prOnVarClick(mnuVars);

      n := n.Sibling;
      i := mnuOps.Items.IndexOf( n.DataAsString );
      if i >= 0 then begin
        mnuOps.ItemIndex := i;
      end;

      n := n.Sibling;
      if s = 'Color' then begin
        try
          panColors.Chosen := Integer( n.Data[0] );
        except
        end;
      end else begin
        txtExpr.Text := n.DataAsString;
      end;
    end;
  end;
end;


// *****************************************************************************

procedure TQUB_Results_CriteriaPanel.Initialize(Results :TQUB_Results);
var
  i                       :Integer;
  CritLine                :TQUB_SelCritLine;
begin
  OnResize                := prOnResize;

  prCritLines             := TList.Create;

  lblCrit                 := TLabel.Create(Self);
  lblCrit.Parent          := Self;
  lblCrit.Caption         := 'Change the color of segments matching:';
  lblCrit.Left            := 0;
  lblCrit.Width           := Width;
  lblCrit.Top             := 0;
  lblCrit.Anchors         := [akLeft, akTop, akRight];

  panCritCtrls            := TPanel.Create(Self);
  panCritCtrls.Parent     := Self;
  panCritCtrls.Left       := 0;
  panCritCtrls.Width      := Width;

  mnuDoCrit               := TQUB_PyCmd_MenuButton.Create('CriteriaAction');
  mnuDoCrit.Caption       := 'Change Color';
  mnuDoCrit.Parent        := panCritCtrls;
  mnuDoCrit.Button.Top    := 2;
  mnuDoCrit.Button.Width  := 90;
  mnuDoCrit.Button.Left   := panCritCtrls.Width - mnuDoCrit.Button.Width - 2;
  mnuDoCrit.Button.OnMouseUp := prOnDoCritMouseUp;

  panCritColors               := TQUB_ColorPanel.CreateCounted(Self, QUB_SELECT_COLOR_COUNT);
  panCritColors.Parent        := panCritCtrls;
  panCritColors.OnColorClick  := prOnColorClick;
  panCritColors.Chosen        := 1;
  panCritColors.ColorWidth    := mnuDoCrit.Button.Height;
  panCritColors.Orientation   := QUB_COLORPANEL_HORIZ;
  panCritColors.Top           := 0;
  //panCritColors.Height        := mnuDoCrit.Button.Height;
  panCritColors.Left          := 0;
  panCritColors.Width         := mnuDoCrit.Button.Left;
  panCritColors.Anchors       := [akLeft, akTop];

  panCritCtrls.Height      := mnuDoCrit.Button.Height + 4;
  panCritCtrls.Top         := Height - panCritCtrls.Height;
  panCritCtrls.Anchors     := [akLeft, akBottom, akRight];

  panCrits                := TScrollBox.Create(Self);
  panCrits.Parent         := Self;
  panCrits.AutoScroll     := True;
  panCrits.Left           := panCritCtrls.Left;
  panCrits.Top            := lblCrit.Height;
  panCrits.Width          := panCritCtrls.Width;
  panCrits.Height         := Height - lblCrit.Height - panCritCtrls.Height;
  panCrits.Anchors        := [akTop, akLeft, akRight, akBottom];
  // panCrits.Scroll

  for i := 0 to QUB_SELCRIT_LINES - 1 do begin
    CritLine         := TQUB_SelCritLine.Create(Self);
    CritLine.Parent  := panCrits;
    CritLine.Results := Results;
    CritLine.Left    := 0;
    CritLine.Width   := panCrits.Width;
    CritLine.Top     := i * CritLine.Height;
    CritLine.Align   := alTop;
    prCritLines.Add( CritLine );
  end;

  prOnResize(Self);
end;

destructor TQUB_Results_CriteriaPanel.Destroy;
begin
  prCritLines.Free;
  inherited;
end;

function TQUB_Results_CriteriaPanel.AsNode: INode;
var
  i                      :Integer;
  CritLine               :TQUB_SelCritLine;
begin
  Result        := TNode.Create('Criteria');
  for i := 0 to QUB_SELCRIT_LINES - 1 do begin
    CritLine    := TQUB_SelCritLine(prCritLines[i]);
    if CritLine.Valid then begin
      Result.AppendChild( CritLine.AsNode );
    end;
  end;
end;

procedure TQUB_Results_CriteriaPanel.prOnResize(Sender :TObject);
begin
  //lblCrit.Width           := tabCrit.Width;
  //panCritCtrls.Width      := tabCrit.Width - panCritCtrls.Left;
  mnuDoCrit.Button.Top    := 2;
  mnuDoCrit.Button.Width  := 90;
  mnuDoCrit.Button.Left   := panCritCtrls.Width - mnuDoCrit.Button.Width - 2;
  panCritColors.Width     := mnuDoCrit.Button.Left;
  //panCritCtrls.Height      := mnuDoCrit.Button.Height + 4;
  //panCritCtrls.Top         := tabCrit.Height - panCritCtrls.Height;
  //panCrits.SetBounds( panCritCtrls.Left, lblCrit.Height, panCritCtrls.Width, tabCrit.Height - lblCrit.Height - panCritCtrls.Height );
end;

procedure TQUB_Results_CriteriaPanel.prOnColorClick(Sender: TObject; Color :Integer);
begin
  if Assigned(OnColorChange) then OnColorChange(Self);
end;

procedure TQUB_Results_CriteriaPanel.prOnDoCritMouseUp(Sender: TObject; MouseButton: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if MouseButton = mbRight then begin
    mnuDoCrit.DoOnMouseUp(Sender, MouseButton, Shift, X, Y);
  end else begin
    PythonThread.InterpretNonSilently('Commands.CallByName( "CriteriaAction", "Select" )');
    if Assigned(OnAction) then OnAction(Self);
  end;
end;

function TQUB_Results_CriteriaPanel.prGetChosenColor: Integer;
begin
  Result := panCritColors.Chosen;
end;

procedure TQUB_Results_CriteriaPanel.prSetChosenColor(x :Integer);
begin
  panCritColors.Chosen := x;
end;

// *****************************************************************************
end.

