{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// maintains list of all HistogramGroup nodes in a Result tree
// displays one group (in the Histograms tab of QUB Visual Results)


unit QUB_Results_Histograms;


interface


uses


Windows, Messages, SysUtils, Classes, Math, IEEE754, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, Series, ComCtrls, Menus, Buttons, TeeProcs, TeEngine, Chart, Registry, Clipbrd,

AllocTypes, Alloc,

QUB_Results,
QUB_Node,
QUBFits,
QUBFits_Dialog,
QUB_QVO,
QUB_Global;


const


QUB_SELHIST_HINTCYCLE = 10;


type


TQUB_Results_HistGroup = class
protected

  
  prNode                  :INode;
  prData                  :TQUB_QVO;
  prDataSel               :TQUB_FirstLast;


  procedure prSetNode(NewNode :INode); // or set to the same to force a re-read
  function prGetCount: Integer;


public


  OnNodeChanged           :TNotifyEvent;


  property Node           :INode              read prNode           write prSetNode;
  property Data           :TQUB_QVO           read prData           write prData;
  property DataSel        :TQUB_FirstLast     read prDataSel        write prDataSel;
  property HistogramCount :Integer            read prGetCount;


end;

TQUB_Results_HistGroupDisplay = class;

TQUB_Results_HistDisplay = class(TChart)
protected


  prNode                  :INode;
  prData                  :TQUB_QVO;
  prDataSel               :TQUB_FirstLast;

  prVLines                :array of Double;

  prHint                  :String; // the stored part, to be combined with coords
  prHintCycle             :Integer;
  prX, prY                :Double;

  prShowComponents        :Boolean;

  prGroupDisplay          :TQUB_Results_HistGroupDisplay;

  prPopup                 :TPopupMenu;
  prCopyXItem             :TMenuItem;
  prCopyTenXItem          :TMenuItem;
  prCopyYItem             :TMenuItem;
  prCopyItem              :TMenuItem;
  prPrintItem             :TMenuItem;
  prCurveFitItem          :TMenuItem;
  prHideCurvesItem        :TMenuItem;
  prShowComponentsItem    :TMenuItem;
  prSendToItem            :TMenuItem;
  prExtractItem           :TMenuItem;

  prChosenPythonItem      :String;

  prFitData               :INode;
  prFitterState           :INode;
  prVisDataSet            :TQUB_FitDataSet;

  
  OnAfterCopy             :TNotifyEvent; // resizing for copy messes up page layout
  
  
  procedure prSetNode(NewNode :INode);
  procedure prAddOrigLines;

  procedure prOnPopup(Sender :TObject);
  procedure prOnCopyXClick(Sender :TObject);
  procedure prOnCopyTenXClick(Sender :TObject);
  procedure prOnCopyYClick(Sender :TObject);
  procedure prOnCopyClick(Sender :TObject);
  procedure prOnPrintClick(Sender :TObject);
  procedure prOnCurveFitClick(Sender :TObject);
  procedure prOnHideCurvesClick(Sender :TObject);
  procedure prOnShowComponentsClick(Sender :TObject);
  procedure prOnExtractClick(Sender :TObject);
  procedure prSaveHistogram(FileName :String);
  procedure prOnQtiPlotClick(Sender :TObject);
  procedure prOnPythonItemClick(Sender :TObject);
  procedure prInvokePythonItem;

  procedure prOnDblClick(Sender: TObject);
  procedure prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

  procedure prOnAfterDraw(Sender :TObject);

  procedure prOnCurvesChanged(Sender :TObject);

  procedure prSetShowComponents(x :Boolean);

public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure CopyImage(Printing :Boolean = False; Silent :Boolean = True);
  function HasFit: Boolean;


  property Node           :INode              read prNode           write prSetNode;
  property Data           :TQUB_QVO           read prData           write prData;
  property DataSel        :TQUB_FirstLast     read prDataSel        write prDataSel;

  property GroupDisplay   :TQUB_Results_HistGroupDisplay read prGroupDisplay write prGroupDisplay;
  property ShowComponents :Boolean            read prShowComponents write prSetShowComponents;


end;


TQUB_Results_HistGroupDisplay = class(TPanel)
protected


  prGroups                :TList;
  prCurrIndex             :Integer;

  prHistograms            :TList;
  prHistogramsActive      :Integer;

  prScroll                :TScrollBar;
  prGroupName             :TLabel;

  prTipsOn                :Boolean;
  prShowHistComponents    :Boolean;


  procedure prOnResize(Sender :TObject);
  procedure prOnScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
  
  procedure prOnAfterCopy(Sender :TObject);

  procedure prOnGroupChanged(Sender :TObject);

  procedure prSetCurrIndex(x :Integer);
  procedure prSetupEnoughDisplays(DCount :Integer);

  function prGetGroup(i :Integer): TQUB_Results_HistGroup;
  function prGetGroupCount: Integer;
  function prGetCurrGroup: TQUB_Results_HistGroup;
  function prGetHist(i :Integer): TQUB_Results_HistDisplay;

  procedure prSetShowHistComponents(x :Boolean);
  function prGetTitle: String;


public


  OnModelChosen                   :TNotifyEvent;


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure AddGroup(Node :INode); overload;
  procedure AddGroup(Node :INode; Data :TQUB_QVO; DataSel :TQUB_FirstLast); overload;
  procedure ClearGroups;


  property GroupCount             :Integer                   read prGetGroupCount;
  property Groups[i :Integer]     :TQUB_Results_HistGroup    read prGetGroup;
  property CurrGroup              :TQUB_Results_HistGroup    read prGetCurrGroup;
  property CurrIndex              :Integer                   read prCurrIndex           write prSetCurrIndex;

  property HistogramCount         :Integer                   read prHistogramsActive;
  property Histograms[i :Integer] :TQUB_Results_HistDisplay  read prGetHist;

  property TipsOn                 :Boolean                   read prTipsOn              write prTipsOn;
  property ShowHistComponents     :Boolean                   read prShowHistComponents  write prSetShowHistComponents;

  property Title                  :String                    read prGetTitle;


end;


implementation


uses


ShellAPI,
QUB_GUIInterface, QUB_Viewport_Results,
QUB_PyCmd_MenuButton,
QUB_Console, QUB_Rich_Console,
PythonEngine, QUB_Viewport_Python, QUB_Python_Thread, QUB_Node_Python,
QUB_QFS_Channels,
QUB_QVO_QVD,
QUB_Profiles,
QUB_QtiPlot;


//*****************************************************************************
procedure TQUB_Results_HistGroup.prSetNode(NewNode :INode);
begin
  prNode                  := NewNode;
  if Assigned(OnNodeChanged) then OnNodeChanged(Self);
end;

function TQUB_Results_HistGroup.prGetCount: Integer;
var
  Hist                    :INode;
begin
  Result                  := 0;
  Hist                    := prNode.Find('Histogram');
  while not Hist.IsNull do begin
    Result                := Result + 1;
    Hist                  := Hist.SiblingSameName;
  end;
end;

//*****************************************************************************
constructor TQUB_Results_HistDisplay.Create(AOwner :TComponent);
begin
  inherited;

  ParentColor             := True;
  //Color                   := clWhite;
  View3D                  := False;
  View3DWalls             := False;
  Legend.Visible          := False;

  OnDblClick              := prOnDblClick;

  prPopup                 := TPopupMenu.Create(Self);
  prPopup.OnPopup         := prOnPopup;

  prCopyXItem             := TMenuItem.Create(Self);
  prCopyXItem.OnClick     := prOnCopyXClick;
  prPopup.Items.Add(prCopyXItem);

  prCopyTenXItem          := TMenuItem.Create(Self);
  prCopyTenXItem.OnClick  := prOnCopyTenXClick;
  prPopup.Items.Add(prCopyTenXItem);

  prCopyYItem             := TMenuItem.Create(Self);
  prCopyYItem.OnClick     := prOnCopyYClick;
  prPopup.Items.Add(prCopyYItem);

  prCopyItem              := TMenuItem.Create(Self);
  prCopyItem.Caption      := 'Copy Image...';
  prCopyItem.OnClick      := prOnCopyClick;
  prPopup.Items.Add(prCopyItem);

  prPrintItem             := TMenuItem.Create(Self);
  prPrintItem.Caption     := 'Print Image...';
  prPrintItem.OnClick     := prOnPrintClick;
  prPopup.Items.Add(prPrintItem);

  prCurveFitItem          := TMenuItem.Create(Self);
  prCurveFitItem.Caption  := 'Curve Fitting...';
  prCurveFitItem.OnClick  := prOnCurveFitClick;
  prPopup.Items.Add(prCurveFitItem);

  prHideCurvesItem         := TMenuItem.Create(Self);
  prHideCurvesItem.Caption := 'Hide Curves';
  prHideCurvesItem.OnClick := prOnHideCurvesClick;
  prPopup.Items.Add(prHideCurvesItem);

  prShowComponentsItem     := TMenuItem.Create(Self);
  prShowComponentsItem.Caption := 'Components Visible';
  prShowComponentsItem.OnClick := prOnShowComponentsClick;
  prPopup.Items.Add(prShowComponentsItem);

  prSendToItem            := TMenuItem.Create(Self);
  prSendToItem.Caption    := 'Send To';
  prPopup.Items.Add(prSendToItem);

  prExtractItem           := TMenuItem.Create(Self);
  prExtractItem.Caption   := 'Notepad';
  prExtractItem.OnClick   := prOnExtractClick;
  prSendToItem.Add(prExtractItem);

  prNode                   := TNode.CreateNull;

end;

destructor TQUB_Results_HistDisplay.Destroy;
begin
  try
    if HasFit then prOnHideCurvesClick(prHideCurvesItem);
  finally
    inherited;
  end;
end;

function TQUB_Results_HistDisplay.HasFit: Boolean;
begin
  Result                   := prFitData <> nil;
end;

procedure TQUB_Results_HistDisplay.prSetNode(NewNode :INode);
var
  BinCount                 :Integer;
  Bars                     :TBarSeries;
  Bins                     :INode;
  Series                   :INode;
  ANode                    :INode;
  Color                    :TColor;
  VLine                    :INode;
  b                        :Integer;
  HasValue                 :Double;
begin
  prNode                   := NewNode;

  if HasFit then prOnHideCurvesClick(prHideCurvesItem);
  RemoveAllSeries;
  SetLength(prVLines, 0);

  if prNode.IsNull then begin

    Title.Text.Text        := '';
    Hint                   := '';
    ShowHint               := False;
    PopupMenu              := nil;
    OnMouseMove            := nil;
    OnAfterDraw            := nil;

  end else begin

    PopupMenu              := prPopup;

    try
      BinCount                  := Integer(prNode['BinCount'].Data[0]);
      Bins                      := prNode['Bins'];
      Series                    := prNode['Bars'];

      Bars                      := TBarSeries.Create(Self);
      Bars.BarStyle             := bsCilinder;
      Bars.MultiBar             := mbSide;
      Bars.UseYOrigin           := False;
      Bars.Marks.Visible        := False;
      Bars.BarWidthPercent      := 100;

      ANode                     := Series.Find('Color');
      if ANode.IsNull then Color := clWhite else Color := QUB_QFS_QMF_CLASS_COLORS[Integer(ANode.Data[0])];

      HasValue        := 0.0;
      for b := 0 to BinCount - 1 do begin
        if not (IsNan(Bins.Data[b])   or IsInfinity(Bins.Data[b]) or
                IsNan(Series.Data[b]) or IsInfinity(Series.Data[b])) then begin
          Bars.AddXY(Bins.Data[b], Series.Data[b], '', Color);
          HasValue := Max(Abs(Series.Data[b]), HasValue);
        end;
      end;
      //if HasValue > 0.0 then begin
      AddSeries( Bars );
      //end else begin
        //Bars.Free;
      //end;

      b := 0;
      VLine                     := prNode.Find('VLine');
      while not VLine.IsNull do begin
        if VLine.DataCount > 0 then begin
          SetLength( prVLines, b+1 );
          prVLines[b] := Double(VLine.Data[0]);
          Inc(b);
        end;
        VLine                   := VLine.SiblingSameName;
      end;

      prAddOrigLines;

    except
    end;

    ANode := prNode.Find('Title');
    if not ANode.IsNull then begin
      Title.Text.Text              := ANode.GetDataAsString;
    end else begin
      Title.Text.Text              := '';
    end;

    ANode := prNode.Find('XLabel');
    if not ANode.IsNull then begin
      BottomAxis.Title.Caption     := ANode.GetDataAsString;
    end else begin
      BottomAxis.Title.Caption     := '';
    end;

    ANode := prNode.Find('YLabel');
    if not ANode.IsNull then begin
      LeftAxis.Title.Caption       := ANode.GetDataAsString;
    end else begin
      LeftAxis.Title.Caption       := '';
    end;

    ANode := prNode.Find('Hint');
    if not ANode.IsNull then begin
      prHint               := ANode.GetDataAsString;
      if GroupDisplay <> nil then ShowHint := GroupDisplay.TipsOn;
    end else begin
      Hint                 := '';
      ShowHint             := False;
    end;

    OnMouseMove            := prOnMouseMove;
    OnAfterDraw            := prOnAfterDraw;

  end;

end;

procedure TQUB_Results_HistDisplay.prAddOrigLines;
var
  BinCount                 :Integer;
  Line                     :TLineSeries;
  Bins                     :INode;
  Series                   :INode;
  ANode                    :INode;
  Color                    :TColor;
  b                        :Integer;
  HasValue                 :Double;
begin
  try
    BinCount               := Integer(prNode['BinCount'].Data[0]);
    Bins                   := prNode['Bins'];

    if ShowComponents then begin
      Series               := prNode.Find('Component');
      while not Series.IsNull do begin
        Line               := TLineSeries.Create(Self);
        Line.Marks.Visible := False;
        Line.Title         := 'pdf';
        Line.LinePen.Width := 2;
        Color              := clGray;

        HasValue           := 0.0;
        for b := 0 to BinCount - 1 do begin
          if not (IsNan(Bins.Data[b])   or IsInfinity(Bins.Data[b]) or
                IsNan(Series.Data[b]) or IsInfinity(Series.Data[b])) then begin
            Line.AddXY(Bins.Data[b], Series.Data[b], '', Color);
            HasValue := Max(Abs(Series.Data[b]), HasValue);
          end;
        end;
        if HasValue > 0.0 then begin
          AddSeries( Line );
        end else begin
          Line.Free;
        end;

        Series             := Series.SiblingSameName;
      end;
    end;

    Series                 := prNode.Find('Line');
    while not Series.IsNull do begin
      Line                      := TLineSeries.Create(Self);
      Line.Marks.Visible        := False;
      Line.Title                := 'pdf';
      Line.LinePen.Width        := 3;//Draw a thick line

      ANode := Series.Find('Color');
      if ANode.IsNull then Color := clBlack else Color := QUB_QFS_QMF_CLASS_COLORS[Integer(ANode.Data[0])];

      HasValue := 0.0;
      for b := 0 to BinCount - 1 do begin
        if not (IsNan(Bins.Data[b])   or IsInfinity(Bins.Data[b]) or
                IsNan(Series.Data[b]) or IsInfinity(Series.Data[b])) then begin
          Line.AddXY(Bins.Data[b], Series.Data[b], '', Color);
          HasValue := Max(Abs(Series.Data[b]), HasValue);
        end;
      end;
      if HasValue > 0.0 then begin
        AddSeries( Line );
      end else begin
        Line.Free;
      end;

      Series := Series.SiblingSameName;
    end;
  except
  end;
end;

procedure TQUB_Results_HistDisplay.prOnPopup(Sender :TObject);
var
  i             :Integer;
  Items         :TStringList;
  Item          :TMenuItem;
begin
  prCopyXItem.Caption    := Format('Copy X (%.4g) to clipboard', [prX]);
  prCopyTenXItem.Caption := Format('Copy 10^X (%.4g) to clipboard', [Math.Power(10, prX)]);
  prCopyYItem.Caption    := Format('Copy Y (%.4g) to clipboard', [prY]);

  prHideCurvesItem.Visible := HasFit;

  for i := prSendToItem.Count - 1 downto 1 do begin
    prSendToItem[i].Free;
  end;

  PyCmd_FillMenu( prSendToItem, 'Histogram', prOnPythonItemClick );
  Items         := TStringList.Create;
  Items.Sorted  := True;
  TQUB_QtiPlot_Connection.ListScripts('Histogram', Items);
  for i := 0 to Items.Count - 1 do begin
    Item           := TMenuItem.Create(Self);
    Item.Caption   := ChangeFileExt(Items[i], '');
    Item.OnClick   := prOnQtiPlotClick;
    prSendToItem.Add(Item);
  end;
  Items.Free;
end;

procedure TQUB_Results_HistDisplay.prOnCopyXClick(Sender :TObject);
var
  Text      :String;
begin
  Text      := Format('%.4g', [prX]);
  Clipboard.SetTextBuf(PChar(Text));
end;

procedure TQUB_Results_HistDisplay.prOnCopyTenXClick(Sender :TObject);
var
  Text      :String;
begin
  Text      := Format('%.4g', [Math.Power(10, prX)]);
  Clipboard.SetTextBuf(PChar(Text));
end;

procedure TQUB_Results_HistDisplay.prOnCopyYClick(Sender :TObject);
var
  Text      :String;
begin
  Text      := Format('%.4g', [prY]);
  Clipboard.SetTextBuf(PChar(Text));
end;

procedure TQUB_Results_HistDisplay.prOnExtractClick(Sender :TObject);
var
  Dlg     :TSaveDialog;
  FName   :String;
begin
  Dlg := TSaveDialog.Create(Self);
  try
    Dlg.FileName := GroupDisplay.Title + ' ' + Trim(Title.Text.Text) + '.txt';
    Dlg.Filter := 'Text files (*.txt)|*.txt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofOverwritePrompt];
    Dlg.Title := 'Save histogram for Notepad';
    if Dlg.Execute then begin
      FName := ChangeFileExt(Dlg.FileName, '.txt');
      prSaveHistogram(FName);
      ShellExecute(Application.Handle, 'open', 'notepad', PChar(FName), PChar(ExtractFilePath(FName)), SW_SHOWNORMAL);
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Results_HistDisplay.prSaveHistogram(FileName :String);
var
  Text    :String;
  f       :File;
  i, j    :Integer;
  Entry   :Integer;
begin

  AssignFile(f, FileName);
  try
    Rewrite(f, 1);
  except
    MessageDlg('Cannot open file ' + FileName + ' for writing', mtError, [mbOK], 0);
    Exit;
  end;

  Entry := Console.NewEntry('Extracting histogram file ' + FileName + ' ...', COLOR_TEXT);

  try

    Text := BottomAxis.Title.Caption + #9 + LeftAxis.Title.Caption + #13#10;
    for j := 0 to Series[0].Count - 1 do begin
      Text := Text + FormatFloat('0.000000e+', Series[0].XValue[j]);
      for i := 0 to SeriesCount - 1 do begin
        Text := Text + FormatFloat(#9'0.000000e+', Series[i].YValue[j]);
      end;
      Text := Text + #13#10;
    end;

    try
      BlockWrite(f, PChar(Text)^, Length(Text));
    except
      MessageDlg('Could not write data to file ' + FileName, mtError, [mbOK], 0);
    end;

  finally

    try
      CloseFile(f);
    except
    end;

    Console.SetEntry(Entry, 'Extracting histogram file ' + FileName + ' ... Done.', COLOR_GRAY);
    //Beep;

  end;

end;

procedure TQUB_Results_HistDisplay.prOnQtiPlotClick(Sender :TObject);
var
  QtiPlot           :TQUB_QtiPlot_Connection;
  ScriptPath        :String;
  Buf               :TPMSingle;
  i, j              :Integer;
  Colors            :String;
  Ser               :INode;
begin
  Buf               := nil;
  ScriptPath        := ChangeFileExt(StringReplace(TMenuItem(Sender).Caption, '&', '', [rfReplaceAll]), '.py');
  QtiPlot           := TQUB_QtiPlot_Connection.Create;
  try
    Buf             := SAlloc2(Series[0].Count, SeriesCount+1);
    for i := 0 to Series[0].Count - 1 do begin
      for j := 0 to SeriesCount do begin
        if j = 0 then begin
          Buf[i][j] := Series[0].XValue[i];
        end else begin
          Buf[i][j] := Series[j-1].YValue[i];
        end;
      end;
    end;

    Colors          := '(';
    Ser             := prNode.Find('Bars').Find('Color');
    if Ser.DataCount > 0 then begin
      Colors        := Colors + IntToStr(Ser.Data[0]);
    end else begin
      Colors        := Colors + '0';
    end;
    Ser             := prNode.Find('Line');
    if Ser.IsNull then Colors := Colors + ','; // singleton tuple, python requires trailing comma
    while not Ser.IsNull do begin
      if Ser.Find('Color').DataCount > 0 then begin
        Colors      := Colors + ', ' + IntToStr(Ser['Color'].Data[0]);
      end else begin
        Colors      := Colors + ', 0';
      end;
      Ser           := Ser.SiblingSameName;
    end;
    Colors          := Colors + ')';

    QtiPlot.BringToFront;
    QtiPlot.ExecScript('Histogram', ScriptPath);
    QtiPlot.SetArray(Buf[0], Series[0].Count, SeriesCount+1, 'float32', 'a');
    QtiPlot.Exec(Format('add_hist("%s", "%s", "%s", a[:,0], a[:,1], %s, *[a[:,i] for i in xrange(2, a.shape[1])])',
                        [GroupDisplay.Title + ' ' + Trim(Title.Text.Text),
                         BottomAxis.Title.Caption,
                         LeftAxis.Title.Caption, Colors]));

  finally
    QtiPlot.Free;
    Free2(TPMChar(Buf));
  end;
end;

procedure TQUB_Results_HistDisplay.prOnCopyClick(Sender :TObject);
begin
  CopyImage(False{Print}, False{Silent});
end;

procedure TQUB_Results_HistDisplay.prOnPrintClick(Sender :TObject);
begin
  CopyImage(True{Print}, False{Silent});
end;

procedure TQUB_Results_HistDisplay.CopyImage(Printing :Boolean = False; Silent :Boolean = True);
var
  Root        :INode;
  ResultType  :String;
  //OldColor  :TColor;
  //OldParent :Boolean;
begin
  ResultType  := 'def';         // chase up the result tree to find type
  Root        := Node.Parent;   // (different result-types have their own hist prefs)
  while not Root.IsNull do begin
    if not Root.Find('ResultType').IsNull then begin
      ResultType := Root['ResultType'].DataAsString;
    end;
    Root      := Root.Parent;
  end;
  with ViewportHistogram.PrintPropertiesForm do begin
    UseProfile(ResultType);
    Print(Self, Node, not Printing{Copy}, True{Properties});
    if Assigned(OnAfterCopy) then OnAfterCopy(Self);
  end;
{
  try
    OldColor            := Color;
    OldParent           := ParentColor;
    try
      ParentColor       := False;
      Color             := clWhite;
      CopyToClipboardMetafile(True);
    finally
      Color             := OldColor;
      ParentColor       := OldParent;
    end;
  except
  end; }
end;

procedure TQUB_Results_HistDisplay.prOnCurveFitClick(Sender :TObject);
var
  N                        :Integer;
  Line                     :INode;
  DataSet                  :TQUB_FitDataSet;

  function CopyToDouble(inName, outName :String): TPADouble;
  var
    Src        :INode;
    Dest       :INode;
    i          :Integer;
  begin
    Src        := prFitData[inName];
    Dest       := prFitData[outName];
    Dest.SetupNumData(NODE_DATATYPE_DOUBLE, Src.Rows, Src.Cols);
    Result     := TPADouble(Dest.DataPtr);
    for i := 0 to Src.DataCount - 1 do begin
      Result[i] := Src.Data[i];
    end;
  end;

begin
  N                        := Series[0].Count;

  if not HasFit then begin
    prFitData                := prNode.Clone(True);
    CopyToDouble('Bins', 'DBins');
    CopyToDouble('Bars', 'DBars');

    Line                     := prFitData.Find('Line');
    while not Line.IsNull do begin
      prFitData.RemoveChild(Line);
      Series[ SeriesCount - 1 ].Free;
      Line                   := prFitData.Find('Line');
    end;
    Line                     := prFitData.Find('Component');
    while not Line.IsNull do begin
      prFitData.RemoveChild(Line);
      Series[ SeriesCount - 1 ].Free;
      Line                   := prFitData.Find('Component');
    end;
  end;

  if prFitterState = nil then begin
    FitDlog.SetState(ProfileMgr['Fitter-HistogramR'].Active); // init from last Select fit
  end else begin
    FitDlog.SetState(prFitterState);  // init from still-active fit this chart
  end;

  FitDlog.Data.AbsX         := True;
  FitDlog.Data.DisplayStyle := QUBFITS_DISPLAY_HIST;
  DataSet                   := TQUB_FitDataSet.Create(N);
  DataSet.AddVar('X', TPADouble(prFitData['DBins'].DataPtr));
  DataSet.AddVar('Y', TPADouble(prFitData['DBars'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prVisDataSet = nil then begin
    prVisDataSet               := TQUB_FitDataSet.Create;
    prVisDataSet.Assign(DataSet);
    // prVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  DataSet.Free;
  FitDlog.Fitter.Events.AddEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.ShowModal;
  FitDlog.Fitter.Events.DelEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  prFitterState := FitDlog.GetState;     // save state in case they re-open fitting
  ProfileMgr['Fitter-HistogramR'].Active := prFitterState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
end;

procedure TQUB_Results_HistDisplay.prOnHideCurvesClick(Sender :TObject);
begin
  if HasFit then begin
    while SeriesCount > 1 do begin
      Series[ SeriesCount - 1 ].Free;
    end;
    prFitData := nil;
    prVisDataSet.Free;
    prVisDataSet := nil;

    prAddOrigLines;
  end;
end;

procedure TQUB_Results_HistDisplay.prOnCurvesChanged(Sender :TObject);
var
  FitSeries           :TLineSeries;
  i, N, c             :Integer;
  Color               :TColor;
  Line                :INode;
  NComp               :Integer;
begin
  N                   := Series[0].Count;
  NComp               := FitDlog.Fitter.Curve.GetComponentCount();

  while SeriesCount > (NComp + 2) do begin
    Series[ SeriesCount - 1 ].Free;
    prFitData.RemoveLast('Component');
  end;

  while SeriesCount < (NComp + 2) do begin
    FitSeries                := TLineSeries.Create(nil);
    FitSeries.Marks.Visible  := False;

    if SeriesCount = 1 then begin
      FitSeries.LinePen.Width  := 3;//Draw a thick line
      Line                     := TNode.Create('Line');
      Line.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
      prFitData.AppendChild(Line);
      Color := clRed;
      Line['Color'].SetData( 1 );
    end else begin
    FitSeries.LinePen.Width  := 2;//Draw a thin line
      Line                     := TNode.Create('Component');
      Line.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
      prFitData.AppendChild(Line);
      Color := clBlack;
      Line['Color'].SetData( 0 );
    end;

    for i := 0 to N - 1 do begin
      FitSeries.AddXY(Series[0].XValue[i], 0.0, '', Color);
    end;

    AddSeries( FitSeries );
  end;

  Line := prFitData['Line'];
  FitDlog.Fitter.Curve.EvalArrayAll(prVisDataSet.Data,
                                TPADouble(Line.DataPtr));
  FitSeries                := TLineSeries( Series[1] );
  for i := 0 to N - 1 do begin
    FitSeries.YValue[i]    := Double(Line.Data[i]);
  end;

  Line := prFitData.Find('Component');
  for c := 0 to NComp - 1 do begin
    FitDlog.Fitter.Curve.EvalArrayOne(FitDlog.Data.DataSet.Data,
                                TPADouble(Line.DataPtr),
                                c);
    FitSeries                := TLineSeries( Series[1+c+1] );
    for i := 0 to N - 1 do begin
      FitSeries.YValue[i]    := Double(Line.Data[i]);
    end;
    Line := Line.SiblingSameName;
  end;
end;

procedure TQUB_Results_HistDisplay.prOnShowComponentsClick(Sender :TObject);
begin
  GroupDisplay.ShowHistComponents := not prShowComponents;
end;

procedure TQUB_Results_HistDisplay.prSetShowComponents(x :Boolean);
begin
  prShowComponents := x;
  prShowComponentsItem.Checked := x;

  if (not prNode.IsNull) and (not HasFit) then begin
    while SeriesCount > 1 do begin
      Series[ SeriesCount - 1 ].Free;
    end;
    prAddOrigLines;
  end;
end;

procedure TQUB_Results_HistDisplay.prOnPythonItemClick(Sender :TObject);
begin
  with Sender as TMenuItem do prChosenPythonItem := Caption;

  if PythonThread.SyncPyProc( prInvokePythonItem, False, True, 2000 ) then begin
    MessageHelpDlg('Python is busy.  Try again later.', mtError, [mbOK], 0, 'Error:Python_Is_Busy');
  end;
end;

procedure TQUB_Results_HistDisplay.prInvokePythonItem;
var
  InvokeCmd  :PPyObject;
  histarg    :PPyObject;
  arglist    :PPyObject;
  cmdresult  :PPyObject;
begin
  with GetPythonEngine do begin
    InvokeCmd   := nil;
    histarg     := nil;
    arglist     := nil;
    cmdresult   := nil;
    try
      InvokeCmd := EvalString('Commands.CallByName');
      if HasFit then begin
        histarg := PyNode_FromNode( prFitData );
      end else begin
      histarg   := PyNode_FromNode( prNode );
      end;
      arglist   := Py_BuildValue('(ssOi)', [PChar('Histogram'), PChar(prChosenPythonItem), histarg, Integer(ShowComponents)]);
      cmdresult := PyEval_CallObject( InvokeCmd, arglist );
      if cmdresult = nil then PyErr_Print;
    finally
      Py_XDECREF( cmdresult );
      Py_XDECREF( arglist );
      Py_XDECREF( histarg );
      Py_XDECREF( InvokeCmd );
    end;
  end;
end;

procedure TQUB_Results_HistDisplay.prOnDblClick(Sender: TObject);
begin
  if Data <> nil then begin
    if not TQUB_QVO_QVD(Data).LockedReadWrite then begin
      TQUB_QVO_QVD(Data).MoveSelectionIntoWindow(DataSel, DataSel.First);
    end;
  end;
end;

procedure TQUB_Results_HistDisplay.prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  prX := Series[0].XScreenToValue(X);
  prY := Series[0].YScreenToValue(Y);

  if prHintCycle = QUB_SELHIST_HINTCYCLE then begin
    ShowHint             := False;
    prHintCycle          := 0;
  end else begin
    if GroupDisplay.TipsOn then begin
      Hint               := Format('X: %f'#13#10'Y: %f'#13#10 + prHint, [prX, prY]);
      ShowHint             := True;
    end;
    prHintCycle          := prHintCycle + 1;
  end;
end;

procedure TQUB_Results_HistDisplay.prOnAfterDraw(Sender :TObject);
var
  i                   :Integer;
  X, Y0, Y1           :Integer;
begin
  Canvas.Pen.Color    := clRed;
  Canvas.Pen.Width    := 2;
  Y0                  := ChartRect.Top;
  Y1                  := ChartRect.Bottom;
  for i := 0 to Length(prVLines) - 1 do begin
    X                 := Series[0].CalcXPosValue( prVLines[i] );
    if (X >= ChartRect.Left) and (X <= ChartRect.Right) then begin
      Canvas.MoveTo( X, Y0 );
      Canvas.LineTo( X, Y1 );
    end;
  end;
end;

// *****************************************************************************
constructor TQUB_Results_HistGroupDisplay.Create(AOwner :TComponent);
var
  Display                 :TQUB_Results_HistDisplay;
begin
  inherited;

  prGroups                := TList.Create;
  prCurrIndex             := -1;
  prHistograms            := TList.Create;
  prHistogramsActive      := 0;

  prGroupName             := TLabel.Create(Self);
  prGroupName.Parent      := Self;
  prGroupName.Caption     := '';
  prGroupName.Align       := alTop;
  prGroupName.Alignment   := taCenter;

  prScroll                := TScrollBar.Create(Self);
  prScroll.Parent         := Self;
  prScroll.Align          := alTop;
  prScroll.Max            := 0;
  prScroll.OnScroll       := prOnScroll;

  Display                 := TQUB_Results_HistDisplay.Create(Self);
  Display.Parent          := Self;
  Display.Align           := alClient;
  Display.GroupDisplay    := Self;
  prHistograms.Add(Display);

  prShowHistComponents    := ViewportResults.ShowHistComponents;
  Display.ShowComponents  := ShowHistComponents;

  BevelOuter              := bvNone;
  OnResize                := prOnResize;
end;

destructor TQUB_Results_HistGroupDisplay.Destroy;
var
  i                       :Integer;
begin
  try
    for i := 0 to GroupCount - 1 do begin
      Groups[i].Free;
    end;
    prGroups.Free;
    prHistograms.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Results_HistGroupDisplay.AddGroup(Node :INode);
var
  Group                   :TQUB_Results_HistGroup;
begin
  Group                   := TQUB_Results_HistGroup.Create;
  Group.Node              := Node;
  prGroups.Add(Group);
  prScroll.Max            := GroupCount - 1;

  if CurrIndex = -1 then CurrIndex := 0;
end;

procedure TQUB_Results_HistGroupDisplay.AddGroup(Node :INode; Data :TQUB_QVO; DataSel :TQUB_FirstLast);
begin
  AddGroup(Node);
  Groups[ GroupCount - 1 ].Data     := Data;
  Groups[ GroupCount - 1 ].DataSel  := DataSel;
end;

procedure TQUB_Results_HistGroupDisplay.ClearGroups;
var
  i                       :Integer;
begin
  CurrIndex               := -1;

  for i := 0 to GroupCount - 1 do begin
    Groups[i].Free;
  end;
  prGroups.Clear;

  prScroll.Max            := 0;
  prScroll.Position       := 0;
end;

procedure TQUB_Results_HistGroupDisplay.prOnResize(Sender :TObject);
var
  i                       :Integer;
  ChartHeight             :Integer;
begin
  for i := prHistograms.Count - 1 downto prHistogramsActive do begin
    Histograms[i].Height  := 1;
  end;

  ChartHeight             := 1;
  if prHistogramsActive > 0 then begin
    ChartHeight           := (Height - prScroll.Height - prGroupName.Height) div prHistogramsActive;
  end;

  for i := prHistogramsActive - 1 downto 0 do begin
    Histograms[i].Height  := ChartHeight;
  end;
end;

procedure TQUB_Results_HistGroupDisplay.prOnScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
begin
  CurrIndex               := ScrollPos;
end;

procedure TQUB_Results_HistGroupDisplay.prOnGroupChanged(Sender :TObject);
var
  ReloadIndex              :Integer;
begin
  if TQUB_Results_HistGroup(Sender) = CurrGroup then begin
    ReloadIndex            := prCurrIndex;
    prCurrIndex            := -1; // force reload
    CurrIndex              := ReloadIndex;
  end;
end;

procedure TQUB_Results_HistGroupDisplay.prSetCurrIndex(x :Integer);
var
  i                       :Integer;
  Group                   :TQUB_Results_HistGroup;
  Hist                    :INode;
  ANode                   :INode;
begin
  if CurrGroup <> nil then CurrGroup.OnNodeChanged := nil;

  if (x < 0) or (x >= GroupCount) then x := -1;
  prCurrIndex             := x;

  if prCurrIndex = -1 then begin

    for i := 0 to prHistogramsActive - 1 do begin
      Histograms[i].Node  := TNode.CreateNull;
    end;
    prHistogramsActive    := 0;
    prGroupName.Caption   := '';

  end else begin

    Group                 := CurrGroup;
    Group.OnNodeChanged   := prOnGroupChanged;
    prHistogramsActive    := Group.HistogramCount;
    prSetupEnoughDisplays( prHistogramsActive );

    i                     := 0;
    Hist                  := Group.Node.Find('Histogram');
    while not Hist.IsNull do begin
      Histograms[i].Node  := Hist;
      Histograms[i].Data  := Group.Data;
      Histograms[i].DataSel := Group.DataSel;
      
      i                   := i + 1;
      Hist                := Hist.SiblingSameName;
    end;
    while i < prHistograms.Count do begin
      Histograms[i].Node  := TNode.CreateNull;
      i                   := i + 1;
    end;

    ANode                 := Group.Node.Find('Title');
    if not ANode.IsNull then begin
      prGroupName.Caption := ANode.DataAsString;
    end else begin
      prGroupName.Caption := '';
    end;

    // if this is part of a model trial, signal QVR to choose it
    ANode                 := Group.Node.Parent;
    while not ANode.IsNull do begin
      if ANode.Name = 'ModelTrial' then begin
        if Assigned(OnModelChosen) then OnModelChosen(Self);
        Break;
      end else begin
        ANode             := ANode.Parent;
      end;
    end;

  end;

  prOnResize(Self);
end;

procedure TQUB_Results_HistGroupDisplay.prSetupEnoughDisplays(DCount :Integer);
var
  i                       :Integer;
  Display                 :TQUB_Results_HistDisplay;
  DisplaysToAdd           :TList; // due to alBottom we gotta reverse 'em
begin
  DisplaysToAdd         := TList.Create;
  for i := prHistograms.Count to DCount - 1 do begin
    Display                 := TQUB_Results_HistDisplay.Create(Self);
    Display.Parent          := Self;
    Display.Align           := alBottom;
    Display.GroupDisplay    := Self;
    Display.ShowComponents  := ShowHistComponents;
    Display.OnAfterCopy     := prOnAfterCopy;
    DisplaysToAdd.Add(Display);
  end;
  for i := DisplaysToAdd.Count - 1 downto 0 do begin
    prHistograms.Add( DisplaysToAdd[i] );
  end;
  DisplaysToAdd.Free;
end;
  
procedure TQUB_Results_HistGroupDisplay.prOnAfterCopy(Sender :TObject);
var
  i :Integer;
begin
  for i := 0 to prHistograms.Count - 1 do begin
    TQUB_Results_HistDisplay(prHistograms[i]).Parent := nil;
  end;
  for i := 0 to prHistograms.Count - 1 do begin
    TQUB_Results_HistDisplay(prHistograms[i]).Top := 5000;
    TQUB_Results_HistDisplay(prHistograms[i]).Align := alBottom;
    TQUB_Results_HistDisplay(prHistograms[i]).Parent := Self;
  end;
  prOnResize(Self);
end;

function TQUB_Results_HistGroupDisplay.prGetGroup(i :Integer): TQUB_Results_HistGroup;
begin
  if (i >= 0) and (i < GroupCount) then begin
    Result    := TQUB_Results_HistGroup( prGroups[i] );
  end else begin
    Result    := nil;
  end;
end;

function TQUB_Results_HistGroupDisplay.prGetGroupCount: Integer;
begin
  Result      := prGroups.Count;
end;

function TQUB_Results_HistGroupDisplay.prGetCurrGroup: TQUB_Results_HistGroup;
begin
  Result      := Groups[ CurrIndex ];
end;

function TQUB_Results_HistGroupDisplay.prGetHist(i :Integer): TQUB_Results_HistDisplay;
begin
  if (i >= 0) and (i < prHistograms.Count) then begin
    Result    := TQUB_Results_HistDisplay( prHistograms[i] );
  end else begin
    Result    := nil;
  end;
end;

procedure TQUB_Results_HistGroupDisplay.prSetShowHistComponents(x :Boolean);
var
  i :Integer;
begin
  if x <> prShowHistComponents then begin
    prShowHistComponents := x;

    for i := 0 to prHistograms.Count - 1 do begin
      TQUB_Results_HistDisplay( prHistograms[i] ).ShowComponents := x;
    end;

    ViewportResults.ShowHistComponents := x;
  end;
end;

function TQUB_Results_HistGroupDisplay.prGetTitle: String;
begin
  Result := prGroupName.Caption;
end;


// *****************************************************************************
end.
