{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the table of models in the Models tab of QUB Visual Results


unit QUB_Results_Models;


interface


uses


Windows, SysUtils, Classes, Graphics, Menus, Controls, Messages, Printers, Registry, StdCtrls, ExtCtrls, ComCtrls, Dialogs, Buttons, Forms,
Chart, Series, TeEngine, TeeProcs,


AllocTypes, Alloc,

QUB_Global,

QUB_Node, QUB_Node_Python, PythonEngine,
QUB_PyCmd_MenuButton,
QUB_ColorPanel,
QUB_Events,

QUB_Results,
QUB_Table,
Grids;


type

TQUB_Results_ModelTable = class(TQUB_Table)
protected

  prResults                   :TQUB_Results;
  prChosen                    :Integer;

  procedure prOnModelsSorted(Sender :TObject);
  procedure prOnSelectCell(Sender: TObject; ACol, ARow: Longint; var CanSelect: Boolean);
  procedure prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

public

  OnModelChosen               :TNotifyEvent;

  constructor Create(AOwner :TComponent; Results :TQUB_Results); reintroduce; virtual;
  destructor Destroy; override;

  property Chosen             :Integer        read prChosen;

end;


implementation


uses


QUB_QFS_Channels,
QUB_Python_Thread,
QUB_Viewport_Python;


constructor TQUB_Results_ModelTable.Create(AOwner :TComponent; Results :TQUB_Results);
begin
  inherited Create(AOwner);

  ColWidths[0]                  := 100;
  FixedRows                     := 1;
  FixedCols                     := 1;
  Options                       := [goFixedVertLine, goFixedHorzLine,
                                    goVertLine,      goHorzLine,
                                    goRowSizing,     goColSizing,
                                    goRowSelect,     goThumbTracking];

  prResults                     := Results;
  Results.Models.OnModelsSorted := prOnModelsSorted;

  OnMouseUp                     := prOnMouseUp;
end;

destructor TQUB_Results_ModelTable.Destroy;
begin
  inherited;
end;

procedure TQUB_Results_ModelTable.prOnSelectCell(Sender: TObject; ACol, ARow: Longint; var CanSelect: Boolean);
begin
  prChosen                      := ARow - 1;
  if prChosen >= 0 then begin
    if Assigned(OnModelChosen) then OnModelChosen(Self);
  end else if ACol > 0 then begin
    prResults.Models.SortBy( prResults.Models.VarName[ACol-1] );
    CanSelect := False;
  end;
end;

procedure TQUB_Results_ModelTable.prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ACol, ARow                    :Integer;
  CanSelect                     :Boolean;
begin
  MouseToCell( X, Y, ACol, ARow );
  prOnSelectCell(Sender, ACol, ARow, CanSelect);
end;

procedure TQUB_Results_ModelTable.prOnModelsSorted(Sender :TObject);
var
  m, v                          :Integer;
begin
  DeleteAll;
  RowCount    := prResults.Models.Count + 1;
  ColCount    := prResults.Models.VarCount + 1;

  with prResults.Models do begin
    for v := 0 to VarCount - 1 do begin
      Cells[v+1, 0] := VarName[v];
    end;

    for m := 0 to Count - 1 do begin
      with Model[m] do begin
        Cells[0, m+1] := Node.DataAsString;
        for v := 0 to VarCount - 1 do begin
          Cells[v+1, m+1] := Format('%.4g', [Value[v]]);
        end;
      end;
    end;
  end;
end;


// *****************************************************************************
end.

