{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the Select tab of QUB Visual Results


unit QUB_Results_Select;


interface                                                           


uses


Windows, SysUtils, Classes, Graphics, Menus, Controls, Messages, Printers, Registry, StdCtrls, ExtCtrls, ComCtrls, Dialogs, Buttons, Forms,
Chart, Series, TeEngine, TeeProcs,
Math,


AllocTypes, Alloc,

QUB_Global,

QUB_Node, QUB_Node_Python, PythonEngine,
QUB_PyCmd_MenuButton,
QUB_ColorPanel,
QUB_Events,

QUB_Results,
QUB_Results_Chart,
QUBFits,
QUBFits_Dialog;


const

WM_QUB_SELCHART_HIDE  = WM_USER + 100;
WM_QUB_SELCHART_2HIDE = WM_USER + 101;
SELECT_RECENT2_MAX    = 10;




type

TQUB_Results_SelectChart = class(TQUB_Results_Chart)
protected

  prVar                 :TQUB_Results_Var;

  prFitterX             :INode;
  prFitData             :INode;
  prFitterState         :INode;
  prVisDataSet          :TQUB_FitDataSet;

  procedure prOnValuesChanged(Sender :TObject);
  procedure prOnStatsChanged(Sender :TObject);
  procedure prOnAxesChanged(Sender :TObject);
  procedure prOnCurvesChanged(Sender :TObject);
  procedure prSetFitLbl;

  procedure prBeforeDrawValues(Sender :TObject); override; // draw bars in hist mode

public

  constructor Create(AOwner :TComponent; SelVar :TQUB_Results_Var); reintroduce; virtual;
  destructor Destroy; override;

  procedure RefreshPoints; override;

  function GetHistAsNode: INode;

  procedure ShowFit;
  procedure AutoFit(Curve, Eqn :String; Nparam :Integer; PVals :TPADouble; PConst :TPAInteger);
  procedure HideFit;

  property SelVar       :TQUB_Results_Var     read prVar;

end;


TQUB_Results_Select2Chart = class(TQUB_Results_2Chart)
protected

  prXVar                 :TQUB_Results_Var;
  prYVar                 :TQUB_Results_Var;

  prFitterX             :INode;
  prFitData             :INode;
  prFitterState         :INode;
  prVisDataSet          :TQUB_FitDataSet;

  procedure prOnValuesChanged(Sender :TObject);
  procedure prOnStatsChanged(Sender :TObject);
  procedure prOnAxesChanged(Sender :TObject);
  procedure prOnCurvesChanged(Sender :TObject);
  procedure prSetFitLbl;

public

  constructor Create(AOwner :TComponent; XSelVar, YSelVar :TQUB_Results_Var); reintroduce; virtual;
  destructor Destroy; override;

  procedure SwapAxes; override;

  procedure SetupFit(State :INode);
  procedure ShowFit;
  procedure AutoFit(Curve, Eqn :String; Nparam :Integer; PVals :TPADouble; PConst :TPAInteger);
  procedure HideFit;

  property XSelVar       :TQUB_Results_Var     read prXVar;
  property YSelVar       :TQUB_Results_Var     read prYVar;

end;


TQUB_Results_SelectPanel = class(TPanel)
protected

  prResults                 :TQUB_Results;
  prIdealAspect             :Double;
  prNumSDs                  :Double;
  prChangingSDs             :Boolean;
  prChosenColor             :Integer;
  prChosenSeg               :Integer;

  mnuVars                   :TComboBox;
  mnuSelect                 :TQUB_PyCmd_MenuButton;

  lblHistBins               :TLabel;
  tbHistBins                :TTrackBar;

  tbSDs                     :TTrackBar;
  edSDs                     :TEdit;

  panColors                 :TQUB_ColorPanel;
  prColorPopup              :TPopupMenu;

  panCharts                 :TPanel;
  prCharts                  :TStringList;
  pr2Charts                 :TList;

  prChartPopup              :TPopupMenu;
  itemXIndex                :TMenuItem;
  itemXStart                :TMenuItem;
  itemXHist                 :TMenuItem;
  itemYLog                  :TMenuItem;
  itemCurveFit              :TMenuItem;
  itemCopyImage             :TMenuItem;
  itemSendTo                :TMenuItem;
  itemExtractNotepad        :TMenuItem;
  itemExtractExcel          :TMenuItem;

  pr2ChartPopup             :TPopupMenu;
  item2XLog                 :TMenuItem;
  item2YLog                 :TMenuItem;
  item2SwapAxes             :TMenuItem;
  item2CurveFit             :TMenuItem;
  item2CopyImage            :TMenuItem;
  item2SendTo                :TMenuItem;
  item2ExtractNotepad       :TMenuItem;
  item2ExtractExcel         :TMenuItem;

  prPointPopup              :TPopupMenu;
  itemShowSeg               :TMenuItem;

  prChartToExtract          :TQUB_Results_SelectChart;
  pr2ChartToExtract         :TQUB_Results_Select2Chart;

  procedure prSetIdealAspect(x :Double);
  procedure prSetChosenColor(x :Integer);
  procedure prSetChosenSeg(x :Integer);

  function prGetChartCount: Integer;
  function prGetChart(Ix :Integer): TQUB_Results_SelectChart;

  function prGet2ChartCount: Integer;
  function prGet2Chart(Ix :Integer): TQUB_Results_Select2Chart;

  procedure prOnResize(Sender :TObject);

  procedure prOnColorClick(Sender: TObject; Color :Integer);
  procedure prOnColorPopup(Sender: TObject);
  procedure prOnColorPopupChoice(Sender: TObject);

  procedure prOnChangeHistBins(Sender :TObject);
  procedure prOnHaveHist(Sender :TObject);
  procedure prOnHaveNoHist(Sender :TObject);

  procedure prOnChangeSD(Sender :TObject);
  procedure prOnChangeSDSlider(Sender :TObject);
  procedure prUpdateSD; // pass NumSD along to the select vars

  procedure prOnClickVar(Sender :TObject);
  procedure prAddSelChart( Name :String );
  procedure prRemSelChart( Name :String );
  procedure prPromptAdd2Chart;
  procedure prAddSel2Chart( XName, YName :String; DoPrefs :Boolean; Force :Boolean = False );
  procedure prRemSel2Chart( Ix :Integer );

  procedure prOnChartPopup(Sender :TObject);
  procedure prOnClickYLog(Sender :TObject);
  procedure prOnClickXStyle(Sender :TObject);
  procedure prOnClickCurveFit(Sender :TObject);
  procedure prOnClickCopyImage(Sender :TObject);
  procedure prOnClickExtractNotepad(Sender :TObject);
  procedure prOnClickExtractQtiPlot(Sender :TObject);
  procedure prOnClickExtractExcel(Sender :TObject);

  procedure prOn2ChartPopup(Sender :TObject);
  procedure prOnClick2XLog(Sender :TObject);
  procedure prOnClick2YLog(Sender :TObject);
  procedure prOnClick2SwapAxes(Sender :TObject);
  procedure prOnClick2CurveFit(Sender :TObject);
  procedure prOnClick2CopyImage(Sender :TObject);
  procedure prOnClick2ExtractNotepad(Sender :TObject);
  procedure prOnClick2ExtractQtiPlot(Sender :TObject);
  procedure prOnClick2ExtractExcel(Sender :TObject);

  procedure prOnClickHide(Sender :TObject);
  procedure prOnHideMsg(var Message: TMessage); message WM_QUB_SELCHART_HIDE;

  procedure prOnClick2Hide(Sender :TObject);
  procedure prOnHide2Msg(var Message: TMessage); message WM_QUB_SELCHART_2HIDE;


  procedure prOnClickPoint(Sender :TQUB_Results_Chart; Ix :Integer);
  procedure prOnSelectPoints(Sender :TQUB_Results_Chart; Points :TList);

  procedure prOnClick2Point(Sender :TQUB_Results_2Chart; Ix :Integer);
  procedure prOnSelect2Points(Sender :TQUB_Results_2Chart; Points :TList);

  procedure prOnPointPopup(Sender :TObject);
  procedure prOnClickShowSeg(Sender :TObject);

  function  prSaveHistogram(Chart :TChart; const FileName :String) :Boolean;
  procedure prExtractQtiPlot(Script :String; Chart :TChart; PlotTitle :String; HistData :INode; Lbl :String);
  procedure prDoExtractExcel;
  procedure prDo2ExtractExcel;

public

  OnShowSegClick            :procedure(Sender :TObject; Ix :Integer) of object;
  OnColorChange             :TNotifyEvent;

  constructor Create(AOwner :TComponent; Results :TQUB_Results); reintroduce; virtual;
  destructor Destroy; override;

  procedure BuildInterface;

  property IdealAspect      :Double           read prIdealAspect       write prSetIdealAspect;
  property NumSD            :Double           read prNumSDs;
  property ChosenColor      :Integer          read prChosenColor       write prSetChosenColor;
  property ChosenSeg        :Integer                                   write prSetChosenSeg;

  property ChartCount       :Integer          read prGetChartCount;
  property Charts[Ix :Integer] :TQUB_Results_SelectChart  read prGetChart;

  property Chart2Count       :Integer          read prGet2ChartCount;
  property Charts2[Ix :Integer] :TQUB_Results_Select2Chart  read prGet2Chart;

end;


implementation


uses

ShellAPI,
QUB_QtiPlot,
QUB_Console, QUB_Rich_Console,
QUB_Profiles,
QUB_QFS_Channels,
QUB_Python_Thread,
QUB_Viewport_Python,
QUB_Results_VersusDialog;


constructor TQUB_Results_SelectChart.Create(AOwner :TComponent; SelVar :TQUB_Results_Var);
var
  i                       :Integer;
begin
  inherited Create(AOwner);

  prVar                   := SelVar;

  prVar.Events.AddEvent('OnValuesChanged', prOnValuesChanged);
  prVar.Events.AddEvent('OnStatsChanged', prOnStatsChanged);
  prVar.Events.AddEvent('OnAxesChanged', prOnAxesChanged);

  Color                   := clWhite;
  View3D                  := False;
  View3DWalls             := False;
  Legend.Visible          := False;
  ParentShowHint          := True;
  Title.Text.Text         := SelVar.Name;

  for i := 0 to prVar.Count - 1 do begin
    AddPoint( prVar.Points[i] );
  end;

  for i := 0 to QUB_SELECT_COLOR_COUNT - 1 do begin
    AddLine( prVar.LowerSDLine[i] );
    AddLine( prVar.MeanLine[i] );
    AddLine( prVar.UpperSDLine[i] );
  end;

  RefreshPoints; // yeah they already have correct values but if all y-values are the same then one of them needs to be _in_correct or else it crashes hard
  prOnAxesChanged(Self);
  prOnStatsChanged(Self);
  // RefreshColors; // motherfucking tchart re-orders the points as you add them (now called by prOnStatsChanged)
end;

destructor TQUB_Results_SelectChart.Destroy;
begin
  try
    prVar.Events.DelEvent('OnValuesChanged', prOnValuesChanged);
    prVar.Events.DelEvent('OnStatsChanged', prOnStatsChanged);
    prVar.Events.DelEvent('OnAxesChanged', prOnAxesChanged);
  finally
    inherited;
  end;
end;

procedure TQUB_Results_SelectChart.RefreshPoints;
var
  i           :Integer;
  Point       :TQUB_Results_Point;
begin
  if prVar.XStyle = XSTYLE_HIST then begin
    for i := 0 to prPoints.Count - 1 do begin
      Point := TQUB_Results_Point( prPoints[i] );
      prPointSeries.XValues[i]    := Point.YValue;
      prPointSeries.YValues[i]    := Point.XValue;
    end;

    if prPointSeries.MinXValue = prPointSeries.MaxXValue then begin
      prPointSeries.XValues[ prPoints.Count - 1 ] := prPointSeries.MinXValue + prPointSeries.MinXValue * 1.0e-6;
    end;

    UndoZoom;
  end else begin
    inherited;
  end;
end;

function TQUB_Results_SelectChart.GetHistAsNode: INode;
var
  i                 :Integer;
  Line              :INode;
begin
  if prVar.XStyle = XSTYLE_HIST then begin
    Result := prVar.GetHistAsNode;
    for i := 0 to TrendCount - 1 do begin
      Line         := Result.AppendClone( Trends[i].Y, False );
      Line.Name    := 'Line';
      Line['Color'].SetData( Trends[i].Color );
    end;
  end else begin
    Result := TNode.CreateNull;
  end;
end;

procedure TQUB_Results_SelectChart.prOnValuesChanged(Sender :TObject);
begin
  HideFit;
  RefreshPoints;
end;

procedure TQUB_Results_SelectChart.prOnStatsChanged(Sender :TObject);
var
  i                            :Integer;
begin
  RefreshColors;
  RefreshLines;

  Hint               := Format('%12s%12s%12s%12s', ['Color','Mean','Std','Median']);
  for i := 0 to SelVar.Results.ColorCount - 1 do begin
    if SelVar.Results.SegsOfColor[i] <> 0 then begin
      Hint           := Hint + #13#10 + Format('%12d%12.4g%12.4g%12.4g', [i, SelVar.Mean(i), SelVar.SD(i), SelVar.Median(i)]);
    end;
  end;
  if 0 < Length(Lbl) then begin
    Hint := Hint + #13#10 + Lbl;
  end;
end;

procedure TQUB_Results_SelectChart.prOnAxesChanged(Sender :TObject);
begin
  ClearTrends;
  if prVar.XStyle = XSTYLE_INDEX then begin
    BottomAxis.Title.Caption := '';
    LeftAxis.Title.Caption   := prVar.Units;
  end else if prVar.XStyle = XSTYLE_START then begin
    BottomAxis.Title.Caption := 'start (msec)';
    LeftAxis.Title.Caption   := prVar.Units;
  end else if prVar.XStyle = XSTYLE_HIST then begin
    BottomAxis.Title.Caption := prVar.Units;
    LeftAxis.Title.Caption   := 'Count';
  end;
end;

procedure TQUB_Results_SelectChart.prBeforeDrawValues(Sender :TObject);
var
  i                :Integer;
  XNode, YNode     :INode;
  HalfWidth        :Double;
  ThisX            :Double;
  ThisXi           :Integer;
  PrevX            :Integer;
  ThisY            :Integer;
  BottomY          :Integer;

  Line             :TQUB_Results_Line;
  Y1, Y2, X        :Integer;
begin
  if prVar.XStyle = XSTYLE_HIST then begin
    // draw the histogram bars
    XNode          := prVar.HistData['X'];
    YNode          := prVar.HistData['Y'];

    HalfWidth      := (Double(XNode.Data[1]) - Double(XNode.Data[0])) / 2.0;
    BottomY        := ChartRect.Bottom - 1;
    PrevX          := Max( prPointSeries.CalcXPosValue( Double(XNode.Data[0]) - HalfWidth ),
                           ChartRect.Left + 1 );
    Canvas.Pen.Color   := clBlack;
    Canvas.Pen.Width   := 1;
    Canvas.Brush.Color := clGray;
    Canvas.Brush.Style := bsSolid;

    for i := 0 to SelHistBinCount - 1 do begin
      ThisX        := Double(XNode.Data[i]) + HalfWidth;
      ThisXi       := prPointSeries.CalcXPosValue( ThisX );
      ThisY        := Min( BottomY,
                           prPointSeries.CalcYPosValue(Double(YNode.Data[i])) );
      ThisY        := Max( ChartRect.Top + 1, ThisY );

      Canvas.Rectangle( PrevX, BottomY, ThisXi, ThisY );

      PrevX         := ThisXi;
    end;

    // and draw lines vertical instead of horizontal
    Y1 := ChartRect.Top+1;
    Y2 := ChartRect.Bottom-1;

    for i := 0 to prLines.Count - 1 do begin
      Line           := TQUB_Results_Line( prLines[i] );
      if not Line.Visible then Continue;

      if Line.Color > 0 then begin
        Canvas.Pen.Color             := QUB_GetClassColor( Line.Color );
      end else begin
        Canvas.Pen.Color             := clGray;
      end;

      X                              := prPointSeries.CalcXPosValue( Line.Value );

      if (X > ChartRect.Left) and (X < ChartRect.Right) then begin
        Canvas.Pen.Width               := Line.Width;
        Canvas.Line( X, Y1, X, Y2 );
      end;
    end;
  end else begin
    inherited;
  end;
end;

procedure TQUB_Results_SelectChart.prOnCurvesChanged(Sender :TObject);
var
  Curve                    :INode;
  Trend                    :TQUB_Results_Trend;
  ic                       :Integer;
begin
  if (TrendCount > 0) and (TrendCount <> (FitDlog.Fitter.Curve.GetComponentCount()+1))
  then ClearTrends;

  Curve                    := prFitData.Find('C');
  for ic := 0 to FitDlog.Fitter.Curve.GetComponentCount() - 1 do begin
    if Curve.IsNull then begin
      Curve := prFitData.AppendChild('C');
    end;
    if Curve.DataCount <> prFitterX.DataCount then begin
      Curve.SetupNumData(NODE_DATATYPE_DOUBLE, prFitterX.DataCount, 1);
    end;
    FitDlog.Fitter.Curve.EvalArrayOne(prVisDataSet.Data,
                                TPADouble(Curve.DataPtr),
                                ic);
    if ic >= TrendCount then begin
      Trend              := TQUB_Results_Trend.Create;
      Trend.X            := prFitterX;
      Trend.Y            := Curve;
      Trend.Color        := 0;
      Trend.Visible      := True;
      AddTrend(Trend);
    end;
    Curve                  := Curve.SiblingSameName;
  end;

  Curve              := prFitData['F'];
  FitDlog.Fitter.Curve.EvalArrayAll(FitDlog.Data.DataSet.Data,
                                TPADouble(Curve.DataPtr));
  if TrendCount < (1 + FitDlog.Fitter.Curve.GetComponentCount()) then begin
    Trend              := TQUB_Results_Trend.Create;
    Trend.X            := prFitterX;
    Trend.Y            := Curve;
    Trend.Color        := 1;
    Trend.Visible      := True;
    AddTrend(Trend);
  end;

  RefreshTrends;
end;

procedure TQUB_Results_SelectChart.ShowFit;
var
  X, Y, F                  :INode;
  i                        :Integer;
  DataSet                  :TQUB_FitDataSet;
begin
  if prFitData = nil then begin
    prFitData                := TNode.Create('DataSet');

    if prVar.XStyle = XSTYLE_HIST then begin
      FitDlog.Data.DisplayStyle  := QUBFITS_DISPLAY_HIST;
      prFitterX := prFitData.AppendClone(prVar.HistData['X'], True);
      prFitData.AppendClone(prVar.HistData['Y'], True);
    end else begin
      FitDlog.Data.DisplayStyle  := QUBFITS_DISPLAY_POINTS;
      X                        := prFitData['X'];
      X.SetupNumData( NODE_DATATYPE_DOUBLE, prVar.Count, 1 );
      Y                        := prFitData['Y'];
      Y.SetupNumData( NODE_DATATYPE_DOUBLE, prVar.Count, 1 );

      for i := 0 to prVar.Count - 1 do begin
        X.Data[i]              := prVar.XValue[i];
        Y.Data[i]              := prVar.YValue[i];
      end;

      prFitterX                := X;
    end;

    F                        := prFitData['F'];
    F.SetupNumData( NODE_DATATYPE_DOUBLE, prFitterX.DataCount, 1 );
  end;

  if prFitterState = nil then begin
    FitDlog.SetState(ProfileMgr['Fitter-Select'].Active); // init from last Select fit
  end else begin
    FitDlog.SetState(prFitterState);  // init from still-active fit this chart
  end;

  FitDlog.Data.AbsX          := True;
  DataSet                    := TQUB_FitDataSet.Create(prFitterX.DataCount);
  DataSet.AddVar('X', TPADouble(prFitterX.DataPtr));
  DataSet.AddVar(prVar.Name, TPADouble(prFitData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prVisDataSet = nil then begin
    prVisDataSet               := TQUB_FitDataSet.Create;
    prVisDataSet.Assign(DataSet);
    // prVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  DataSet.Free;
  FitDlog.Fitter.Events.AddEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.ShowModal;
  prSetFitLbl;
  FitDlog.Fitter.Events.DelEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  prFitterState := FitDlog.GetState;     // save state in case they re-open fitting
  ProfileMgr['Fitter-Select'].Active := prFitterState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
end;

procedure TQUB_Results_SelectChart.AutoFit(Curve, Eqn :String; Nparam :Integer; PVals :TPADouble; PConst :TPAInteger);
var
  X, Y, F                  :INode;
  i                        :Integer;
  DataSet                  :TQUB_FitDataSet;
begin
  if prFitData = nil then begin
    prFitData                := TNode.Create('DataSet');

    if prVar.XStyle = XSTYLE_HIST then begin
      FitDlog.Data.DisplayStyle  := QUBFITS_DISPLAY_HIST;
      prFitterX := prFitData.AppendClone(prVar.HistData['X'], True);
      prFitData.AppendClone(prVar.HistData['Y'], True);
    end else begin
      FitDlog.Data.DisplayStyle  := QUBFITS_DISPLAY_POINTS;
      X                        := prFitData['X'];
      X.SetupNumData( NODE_DATATYPE_DOUBLE, prVar.Count, 1 );
      Y                        := prFitData['Y'];
      Y.SetupNumData( NODE_DATATYPE_DOUBLE, prVar.Count, 1 );

      for i := 0 to prVar.Count - 1 do begin
        X.Data[i]              := prVar.XValue[i];
        Y.Data[i]              := prVar.YValue[i];
      end;

      prFitterX                := X;
    end;

    F                        := prFitData['F'];
    F.SetupNumData( NODE_DATATYPE_DOUBLE, prFitterX.DataCount, 1 );
  end;

  if prFitterState = nil then begin
    FitDlog.SetState(ProfileMgr['Fitter-Select'].Active); // init from last Select fit
  end else begin
    FitDlog.SetState(prFitterState);  // init from still-active fit this chart
  end;

  FitDlog.Data.AbsX          := True;
  DataSet                    := TQUB_FitDataSet.Create(prFitterX.DataCount);
  DataSet.AddVar('X', TPADouble(prFitterX.DataPtr));
  DataSet.AddVar(prVar.Name, TPADouble(prFitData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prVisDataSet = nil then begin
    prVisDataSet               := TQUB_FitDataSet.Create;
    prVisDataSet.Assign(DataSet);
    // prVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  DataSet.Free;

  FitDlog.Fitter.Curve       := CreateCurve(PChar(Curve));
  if 0 < Length(Eqn) then begin
    FitDlog.Fitter.Curve.SetEqn(PChar(Eqn));
  end;
  for i := 0 to Nparam - 1 do begin
    if PVals <> nil then FitDlog.Fitter.Curve.SetParamVal(i, PVals[i], 1);
    if PConst <> nil then FitDlog.Fitter.Curve.SetParamConst(i, PConst[i]);
  end;
  FitDlog.Fitter.Fit;
  prSetFitLbl;

  prOnCurvesChanged(FitDlog.Fitter);
  prFitterState := FitDlog.GetState;     // save state in case they re-open fitting
  ProfileMgr['Fitter-Select'].Active := prFitterState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
end;

procedure TQUB_Results_SelectChart.HideFit;
begin
  if prFitData <> nil then begin
    ClearTrends;
    prFitData := nil;
    prVisDataSet.Free;
    prVisDataSet := nil;
  end;
end;

procedure TQUB_Results_SelectChart.prSetFitLbl;
var
  S                        :String;
  i                        :Integer;
begin
  S := '';
  for i := 0 to FitDlog.Fitter.Curve.GetParamCount - 1 do begin
    if 0 = FitDlog.Fitter.Curve.GetParamConst(i) then begin
      if 0 < Length(s) then s := s + '; ';
      s := s + FitDlog.Fitter.Curve.GetParamName(i)
             + Format(' = %.4g +/- %.4g',
                     [FitDlog.Fitter.Curve.GetParamVal(i),
                      FitDlog.Fitter.Errors[i]]);
    end;
  end;
  Lbl := S;
  prOnStatsChanged(nil);
end;


// *****************************************************************************

constructor TQUB_Results_Select2Chart.Create(AOwner :TComponent; XSelVar, YSelVar :TQUB_Results_Var);
var
  i                       :Integer;
begin
  inherited Create(AOwner);

  prXVar                  := XSelVar;
  prYVar                  := YSelVar;

  prXVar.Events.AddEvent('OnValuesChanged', prOnValuesChanged);
  prXVar.Events.AddEvent('OnStatsChanged', prOnStatsChanged);
  prXVar.Events.AddEvent('OnAxesChanged', prOnAxesChanged);

  prYVar.Events.AddEvent('OnValuesChanged', prOnValuesChanged);
  prYVar.Events.AddEvent('OnStatsChanged', prOnStatsChanged);
  prYVar.Events.AddEvent('OnAxesChanged', prOnAxesChanged);

  Color                   := clWhite;
  View3D                  := False;
  View3DWalls             := False;
  Legend.Visible          := False;
  ParentShowHint          := True;
  Title.Text.Text         := XSelVar.Name + ' vs. ' + YSelVar.Name;

  for i := 0 to Min(prXVar.Count, prYVar.Count) - 1 do begin
    AddPoint( prXVar.Points[i], prYVar.Points[i] );
  end;

  for i := 0 to QUB_SELECT_COLOR_COUNT - 1 do begin
    AddXLine( prXVar.LowerSDLine[i] );
    AddXLine( prXVar.MeanLine[i] );
    AddXLine( prXVar.UpperSDLine[i] );
    AddYLine( prYVar.LowerSDLine[i] );
    AddYLine( prYVar.MeanLine[i] );
    AddYLine( prYVar.UpperSDLine[i] );
  end;

  RefreshPoints; // yeah they already have correct values but if all y-values are the same then one of them needs to be _in_correct or else it crashes hard
  prOnAxesChanged(Self);
  prOnStatsChanged(Self);
  // RefreshColors; // motherfucking tchart re-orders the points as you add them (now called by prOnStatsChanged)
end;

destructor TQUB_Results_Select2Chart.Destroy;
begin
  try
    prXVar.Events.DelEvent('OnValuesChanged', prOnValuesChanged);
    prXVar.Events.DelEvent('OnStatsChanged', prOnStatsChanged);
    prXVar.Events.DelEvent('OnAxesChanged', prOnAxesChanged);
    prYVar.Events.DelEvent('OnValuesChanged', prOnValuesChanged);
    prYVar.Events.DelEvent('OnStatsChanged', prOnStatsChanged);
    prYVar.Events.DelEvent('OnAxesChanged', prOnAxesChanged);
  finally
    inherited;
  end;
end;

procedure TQUB_Results_Select2Chart.prOnCurvesChanged(Sender :TObject);
var
  Curve                    :INode;
  Trend                    :TQUB_Results_Trend;
  ic                       :Integer;
begin
  if (TrendCount > 0) and (TrendCount <> (FitDlog.Fitter.Curve.GetComponentCount()+1))
  then ClearTrends;

  Curve                    := prFitData.Find('C');
  for ic := 0 to FitDlog.Fitter.Curve.GetComponentCount() - 1 do begin
    if Curve.IsNull then begin
      Curve := prFitData.AppendChild('C');
    end;
    if Curve.DataCount <> prFitterX.DataCount then begin
      Curve.SetupNumData(NODE_DATATYPE_DOUBLE, prFitterX.DataCount, 1);
    end;
    FitDlog.Fitter.Curve.EvalArrayOne(FitDlog.Data.DataSet.Data,
                                TPADouble(Curve.DataPtr),
                                ic);
    if ic >= TrendCount then begin
      Trend              := TQUB_Results_Trend.Create;
      Trend.X            := prFitterX;
      Trend.Y            := Curve;
      Trend.Color        := 0;
      Trend.Visible      := True;
      AddTrend(Trend);
    end;
    Curve                  := Curve.SiblingSameName;
  end;

  Curve              := prFitData['F'];
  FitDlog.Fitter.Curve.EvalArrayAll(prVisDataSet.Data,
                                TPADouble(Curve.DataPtr));
  if TrendCount < (1 + FitDlog.Fitter.Curve.GetComponentCount()) then begin
    Trend              := TQUB_Results_Trend.Create;
    Trend.X            := prFitterX;
    Trend.Y            := Curve;
    Trend.Color        := 1;
    Trend.Visible      := True;
    AddTrend(Trend);
  end;

  RefreshTrends;
end;

var
  gSortX       :INode;

function CompareX(Item1, Item2 :Pointer): Integer;
var
  Diff       :Double;
begin
  Diff       := gSortX.Data[ Integer(Item1) ]
              - gSortX.Data[ Integer(Item2) ];
  if Diff > 0.0 then begin
    Result := 1;
  end else if Diff = 0.0 then begin
    Result := 0;
  end else begin
    Result := -1;
  end;
end;

procedure MassiveSwap( A :INode; Order :TList );
var
  tmp                      :INode;
  i                        :Integer;
begin
  tmp                      := A.Clone(False);
  for i := 0 to Order.Count - 1 do begin
    A.Data[ i ] := tmp.Data[ Integer(Order[i]) ];
  end;
end;

procedure TQUB_Results_Select2Chart.SetupFit(State :INode);
var
  X, Y, F                  :INode;
  i                        :Integer;
  Order                    :TList;
  DataSet                  :TQUB_FitDataSet;
begin
  if prFitData = nil then begin
    prFitData                := TNode.Create('DataSet');
    X                        := prFitData['X'];
    X.SetupNumData( NODE_DATATYPE_DOUBLE, prXVar.Count, 1 );
    prFitterX                := X;
    Y                        := prFitData['Y'];
    Y.SetupNumData( NODE_DATATYPE_DOUBLE, prXVar.Count, 1 );
    F                        := prFitData['F'];
    F.SetupNumData( NODE_DATATYPE_DOUBLE, prFitterX.DataCount, 1 );


    Order                    := TList.Create;

    for i := 0 to prXVar.Count - 1 do begin
      X.Data[i]              := prXVar.YValue[i];
      Y.Data[i]              := prYVar.YValue[i];
      Order.Add( Pointer(i) );
    end;
    prFitterX                := X;
    gSortX                   := X;

    Order.Sort(CompareX);

    MassiveSwap( X, Order );
    MassiveSwap( Y, Order );

    Order.Free;
    gSortX                   := nil;
  end;

  if State = nil then begin
    prFitterState := ProfileMgr['Fitter-Select2'].Active;
  end else begin
    prFitterState            := State.Clone(True);
    if prFitterState.Find('Curve').IsNull then begin
      prFitterState.AppendClone(ProfileMgr['Fitter-Select2'].Active['Curve'], True);
    end;
    if prFitterState.Find('Fitter').IsNull then begin
      prFitterState.AppendClone(ProfileMgr['Fitter-Select2'].Active['Fitter'], True);
    end;
    if prFitterState.Find('FitData').IsNull then begin
      prFitterState.AppendClone(ProfileMgr['Fitter-Select2'].Active['FitData'], True);
    end;
  end;
  FitDlog.SetState(prFitterState);

  FitDlog.Data.AbsX          := True;
  FitDlog.Data.DisplayStyle  := QUBFITS_DISPLAY_POINTS;
  DataSet                    := TQUB_FitDataSet.Create(prFitterX.DataCount);
  DataSet.AddVar(prXVar.Name, TPADouble(prFitterX.DataPtr));
  DataSet.AddVar(prYVar.Name, TPADouble(prFitData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prVisDataSet = nil then begin
    prVisDataSet               := TQUB_FitDataSet.Create;
    prVisDataSet.Assign(DataSet);
    // prVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  DataSet.Free;
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.Data.UseData(nil);
end;

procedure TQUB_Results_Select2Chart.ShowFit;
var
  DataSet                  :TQUB_FitDataSet;
begin
  SetupFit(prFitterState);

  FitDlog.Data.AbsX          := True;
  FitDlog.Data.DisplayStyle  := QUBFITS_DISPLAY_POINTS;
  DataSet                    := TQUB_FitDataSet.Create(prFitterX.DataCount);
  DataSet.AddVar(prXVar.Name, TPADouble(prFitterX.DataPtr));
  DataSet.AddVar(prYVar.Name, TPADouble(prFitData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prVisDataSet = nil then begin
    prVisDataSet               := TQUB_FitDataSet.Create;
    prVisDataSet.Assign(DataSet);
    // prVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  DataSet.Free;
  FitDlog.Fitter.Events.AddEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.ShowModal;
  prSetFitLbl;
  FitDlog.Fitter.Events.DelEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  prFitterState := FitDlog.GetState;     // save state in case they re-open fitting
  ProfileMgr['Fitter-Select2'].Active := prFitterState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
end;

procedure TQUB_Results_Select2Chart.AutoFit(Curve, Eqn :String; Nparam :Integer; PVals :TPADouble; PConst :TPAInteger);
var
  DataSet                  :TQUB_FitDataSet;
  i                        :Integer;
begin
  SetupFit(prFitterState);

  FitDlog.Data.AbsX          := True;
  FitDlog.Data.DisplayStyle  := QUBFITS_DISPLAY_POINTS;
  DataSet                    := TQUB_FitDataSet.Create(prFitterX.DataCount);
  DataSet.AddVar(prXVar.Name, TPADouble(prFitterX.DataPtr));
  DataSet.AddVar(prYVar.Name, TPADouble(prFitData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prVisDataSet = nil then begin
    prVisDataSet               := TQUB_FitDataSet.Create;
    prVisDataSet.Assign(DataSet);
    // prVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  DataSet.Free;

  FitDlog.Fitter.Curve       := CreateCurve(PChar(Curve));
  if 0 < Length(Eqn) then begin
    FitDlog.Fitter.Curve.SetEqn(PChar(Eqn));
  end;
  if PVals <> nil then begin
    for i := 0 to FitDlog.Fitter.Curve.GetParamCount - 1 do begin
      if i < Nparam then begin
        FitDlog.Fitter.Curve.SetParamVal(i, PVals[i], 1);
      end else begin
        FitDlog.Fitter.Curve.SetParamVal(i, 0, 1);
      end;
    end;
  end;
  if PConst <> nil then begin
    for i := 0 to FitDlog.Fitter.Curve.GetParamCount - 1 do begin
      if i < Nparam then begin
        FitDlog.Fitter.Curve.SetParamConst(i, PConst[i]);
      end else begin
        FitDlog.Fitter.Curve.SetParamConst(i, 1);
      end;
    end;
  end;
  
  FitDlog.Fitter.Fit;
  prSetFitLbl;

  prOnCurvesChanged(FitDlog.Fitter);
  prFitterState := FitDlog.GetState;     // save state in case they re-open fitting
  ProfileMgr['Fitter-Select2'].Active := prFitterState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
end;

procedure TQUB_Results_Select2Chart.HideFit;
begin
  if prFitData <> nil then begin
    ClearTrends;
    prFitData := nil;
    prVisDataSet.Free;
    prVisDataSet := nil;
  end;
end;

procedure TQUB_Results_Select2Chart.prSetFitLbl;
var
  S                        :String;
  i                        :Integer;
begin
  S := '';
  for i := 0 to FitDlog.Fitter.Curve.GetParamCount - 1 do begin
    if 0 = FitDlog.Fitter.Curve.GetParamConst(i) then begin
      if 0 < Length(s) then s := s + '; ';
      s := s + FitDlog.Fitter.Curve.GetParamName(i)
             + Format(' = %.4g +/- %.4g',
                     [FitDlog.Fitter.Curve.GetParamVal(i),
                      FitDlog.Fitter.Errors[i]]);
    end;
  end;
  Lbl := S;
  prOnStatsChanged(nil);
end;

procedure TQUB_Results_Select2Chart.SwapAxes;
var
  TmpSelVar            :TQUB_Results_Var;
begin
  ClearTrends;
  inherited;

  TmpSelVar            := prXVar;
  prXVar               := prYVar;
  prYVar               := TmpSelVar;

  HideFit;
  prOnStatsChanged(Self);
  prOnAxesChanged(Self);
end;

procedure TQUB_Results_Select2Chart.prOnValuesChanged(Sender :TObject);
begin
  ClearTrends;
  RefreshPoints;
end;

procedure TQUB_Results_Select2Chart.prOnStatsChanged(Sender :TObject);
var
  i                            :Integer;
begin
  RefreshColors;
  RefreshLines;

  Hint               := Format('%12s%12s%12s%12s%12s', ['Color','Mean X','Std X','Mean Y','Std Y']);
  for i := 0 to XSelVar.Results.ColorCount - 1 do begin
    if XSelVar.Results.SegsOfColor[i] <> 0 then begin
      Hint           := Hint + #13#10 + Format('%12d%12.4g%12.4g%12.4g%12.4g',
                        [i, XSelVar.Mean(i), XSelVar.SD(i), YSelVar.Mean(i), YSelVar.SD(i)]);
    end;
  end;
  if 0 < Length(Lbl) then begin
    Hint := Hint + #13#10 + Lbl;
  end;
end;

procedure TQUB_Results_Select2Chart.prOnAxesChanged(Sender :TObject);
var
  Units                      :String;
begin
  Units                      := prXVar.Units;
  if Units = '' then
    BottomAxis.Title.Caption := prXVar.Name
  else
    BottomAxis.Title.Caption := Format('%s [%s]', [prXVar.Name, Units]);

  Units                      := prYVar.Units;
  if Units = '' then
    LeftAxis.Title.Caption   := prYVar.Name
  else
    LeftAxis.Title.Caption   := Format('%s [%s]', [prYVar.Name, Units]);
end;

// *****************************************************************************

constructor TQUB_Results_SelectPanel.Create(AOwner :TComponent; Results :TQUB_Results);
begin
  inherited Create(AOwner);

  prResults                 := Results;
  prIdealAspect             := 1.618; // golden mean
  prNumSDs                  := 1.0;
  prChosenColor             := 1;
  prChosenSeg               := -1;

  prCharts                  := TStringList.Create;
  pr2Charts                 := TList.Create;
end;

procedure TQUB_Results_SelectPanel.BuildInterface;
var
  pan                       :TPanel;
  lbl                       :TLabel;
  i                         :Integer;
  Visible2                  :INode;
  XVarName, YVarName        :String;
begin
  pan                       := TPanel.Create(Self);
  pan.Parent                := Self;
  pan.BevelInner            := bvNone;
  pan.BevelOuter            := bvNone;
  pan.Top                   := 0;
  pan.Left                  := 0;
  pan.Width                 := Width;
  // pan.Height
  pan.Anchors               := [akLeft, akRight, akTop];

  mnuVars                   := TComboBox.Create(Self);
  mnuVars.Parent            := pan;
  mnuVars.Style             := csDropDownList;
  mnuVars.DropDownCount     := 20;
  mnuVars.OnClick           := prOnClickVar;
  mnuVars.Left              := 2;
  mnuVars.Top               := 2;
  pan.Height                := mnuVars.Height + 4;

  mnuVars.Items.Add('Show Var...');
  mnuVars.ItemIndex := 0;
  mnuVars.Items.Add('All');
  mnuVars.Items.Add('None');
  mnuVars.Items.Add('Custom...');

  for i := 0 to prResults.SelVarCount - 1 do begin
    mnuVars.Items.Add( prResults.SelVar[i].Name );
  end;

  mnuSelect                 := TQUB_PyCmd_MenuButton.Create('Select');
  mnuSelect.Parent          := pan;
  mnuSelect.Button.Top      := 2;
  mnuSelect.Button.Left     := mnuVars.Width + 10;
  mnuSelect.Button.Width    := 60;

  tbSDs                     := TTrackBar.Create(Self);
  //tbSDs.Orientation       := tbHorizontal;
  tbSDs.TickStyle           := tsNone;
  tbSDs.Min                 := 1;
  tbSDs.Max                 := 30;
  tbSDs.Position            := 10;
  tbSDs.OnChange            := prOnChangeSDSlider;
  tbSDs.Parent              := pan;
  tbSDs.Width               := 60;
  tbSDs.Left                := pan.Width - tbSDs.Width;
  tbSDs.Anchors             := [akTop, akRight];

  prChangingSDs             := False;
  edSDs                     := TEdit.Create(Self);
  edSDs.Text                := Format('%f', [NumSD]);
  edSDs.Parent              := pan;
  edSDs.Width               := 40;
  edSDs.Left                := tbSDs.Left - edSDs.Width;
  edSDs.Anchors             := [akRight, akTop];
  edSDs.OnChange            := prOnChangeSD;

  lbl                       := TLabel.Create(Self);
  lbl.Caption               := 'SDs:';
  lbl.Top                   := 3;
  lbl.Left                  := edSDs.Left - lbl.Width - 5;
  lbl.Anchors               := [akRight, akTop];
  lbl.Parent                := pan;

  tbHistBins                := TTrackBar.Create(Self);
  tbHistBins.TickStyle      := tsNone;
  tbHistBins.Min            := SELECTVAR_MIN_BINCOUNT;
  tbHistBins.Max            := SELECTVAR_MAX_BINCOUNT;
  tbHistBins.Position       := SelHistBinCount;
  tbHistBins.OnChange       := prOnChangeHistBins;
  tbHistBins.Parent         := pan;
  tbHistBins.Width          := 60;
  tbHistBins.Left           := lbl.Left - tbHistBins.Width - 10;
  tbHistBins.Anchors        := [akRight, akTop];
  tbHistBins.Visible        := False;

  lblHistBins               := TLabel.Create(Self);
  lblHistBins.Caption       := 'Bins:';
  lblHistBins.Top           := 3;
  lblHistBins.Left          := tbHistBins.Left - lblHistBins.Width - 5;
  lblHistBins.Anchors       := [akRight, akTop];
  lblHistBins.Parent        := pan;
  lblHistBins.Visible       := False;

  panColors                 := TQUB_ColorPanel.CreateCounted(Self, QUB_SELECT_COLOR_COUNT);
  panColors.Parent          := Self;
  panColors.Orientation     := QUB_COLORPANEL_VERT;
  panColors.OnColorClick    := prOnColorClick;
  panColors.Chosen          := prChosenColor;
  panColors.Top             := pan.Height;
  panColors.Height          := Height - pan.Height;
  panColors.Left            := 0;
  panColors.Anchors         := [akLeft, akTop, akBottom];

  prColorPopup              := TPopupMenu.Create(Self);
  prColorPopup.OnPopup      := prOnColorPopup;
  panColors.ColorPopupMenu  := prColorPopup;

  panCharts                 := TPanel.Create(Self);
  panCharts.Parent          := Self;
  panCharts.Left            := panColors.Width;
  panCharts.Width           := Width - panColors.Width;
  panCharts.Top             := pan.Height;
  panCharts.Height          := Height - pan.Height;
  panCharts.Anchors         := [akLeft, akTop, akRight, akBottom];
  panCharts.ParentShowHint  := True;

  prCharts.Sorted           := True;
  prCharts.Duplicates       := dupError;

  prChartPopup              := TPopupMenu.Create(Self);
  prChartPopup.OnPopup      := prOnChartPopup;

  itemXIndex                := TMenuItem.Create(Self);
  itemXIndex.Caption        := 'By Index';
  itemXIndex.Tag            := Integer( XSTYLE_INDEX );
  itemXIndex.OnClick        := prOnClickXStyle;
  prChartPopup.Items.Add( itemXIndex );

  itemXStart                := TMenuItem.Create(Self);
  itemXStart.Caption        := 'By Start Time';
  itemXStart.Tag            := Integer( XSTYLE_START );
  itemXStart.OnClick        := prOnClickXStyle;
  prChartPopup.Items.Add( itemXStart );

  itemXHist                 := TMenuItem.Create(Self);
  itemXHist.Caption         := 'As Histogram';
  itemXHist.Tag             := Integer( XSTYLE_HIST );
  itemXHist.OnClick         := prOnClickXStyle;
  prChartPopup.Items.Add( itemXHist );

  prChartPopup.Items.InsertNewLineAfter( itemXHist );

  itemYLog                  := TMenuItem.Create(Self);
  itemYLog.Caption          := 'Log10 Y Axis';
  itemYLog.OnClick          := prOnClickYLog;
  prChartPopup.Items.Add( itemYLog );

  prChartPopup.Items.InsertNewLineAfter( itemYLog );

  itemCurveFit              := TMenuItem.Create(Self);
  itemCurveFit.Caption      := 'Curve fitting...';
  itemCurveFit.OnClick      := prOnClickCurveFit;
  prChartPopup.Items.Add( itemCurveFit );

  prChartPopup.Items.InsertNewLineAfter( itemCurveFit );

  itemCopyImage             := TMenuItem.Create(Self);
  itemCopyImage.Caption     := 'Copy Image';
  itemCopyImage.OnClick     := prOnClickCopyImage;
  prChartPopup.Items.Add( itemCopyImage );

  itemSendTo                := TMenuItem.Create(Self);
  itemSendTo.Caption        := 'Send to';
  prChartPopup.Items.Add( itemSendTo );

  itemExtractNotepad          := TMenuItem.Create(Self);
  itemExtractNotepad.Caption  := 'Notepad';
  itemExtractNotepad.OnClick  := prOnClickExtractNotepad;
  itemSendTo.Add( itemExtractNotepad );

  itemExtractExcel          := TMenuItem.Create(Self);
  itemExtractExcel.Caption  := 'Excel';
  itemExtractExcel.OnClick  := prOnClickExtractExcel;
  itemSendTo.Add( itemExtractExcel );

  pr2ChartPopup             := TPopupMenu.Create(Self);
  pr2ChartPopup.OnPopup     := prOn2ChartPopup;

  item2XLog                 := TMenuItem.Create(Self);
  item2XLog.Caption         := 'Log10 X Axis';
  item2XLog.OnClick         := prOnClick2XLog;
  pr2ChartPopup.Items.Add( item2XLog );

  item2YLog                 := TMenuItem.Create(Self);
  item2YLog.Caption         := 'Log10 Y Axis';
  item2YLog.OnClick         := prOnClick2YLog;
  pr2ChartPopup.Items.Add( item2YLog );

  item2SwapAxes             := TMenuItem.Create(Self);
  item2SwapAxes.Caption     := 'Swap Axes';
  item2SwapAxes.OnClick     := prOnClick2SwapAxes;
  pr2ChartPopup.Items.Add( item2SwapAxes );

  pr2ChartPopup.Items.InsertNewLineAfter( item2SwapAxes );

  item2CurveFit              := TMenuItem.Create(Self);
  item2CurveFit.Caption      := 'Curve fitting...';
  item2CurveFit.OnClick      := prOnClick2CurveFit;
  pr2ChartPopup.Items.Add( item2CurveFit );

  pr2ChartPopup.Items.InsertNewLineAfter( item2CurveFit );

  item2CopyImage            := TMenuItem.Create(Self);
  item2CopyImage.Caption    := 'Copy Image';
  item2CopyImage.OnClick    := prOnClick2CopyImage;
  pr2ChartPopup.Items.Add( item2CopyImage );

  item2SendTo                := TMenuItem.Create(Self);
  item2SendTo.Caption        := 'Send to';
  pr2ChartPopup.Items.Add( item2SendTo );

  item2ExtractNotepad          := TMenuItem.Create(Self);
  item2ExtractNotepad.Caption  := 'Notepad';
  item2ExtractNotepad.OnClick  := prOnClick2ExtractNotepad;
  item2SendTo.Add( item2ExtractNotepad );

  item2ExtractExcel          := TMenuItem.Create(Self);
  item2ExtractExcel.Caption  := 'Excel';
  item2ExtractExcel.OnClick  := prOnClick2ExtractExcel;
  item2SendTo.Add( item2ExtractExcel );

  prPointPopup              := TPopupMenu.Create(Self);
  prPointPopup.OnPopup      := prOnPointPopup;

  itemShowSeg               := TMenuItem.Create(Self);
  itemShowSeg.Caption       := 'Show Segment';
  itemShowSeg.OnClick       := prOnClickShowSeg;
  prPointPopup.Items.Add( itemShowSeg );

  for i := 0 to prResults.SelVarCount - 1 do begin
    if prResults.SelVar[i].Visible then prAddSelChart( prResults.SelVar[i].Name );
  end;

  Visible2                  := prResults.VsDefaultPrefs.Child;
  while not Visible2.IsNull do begin
    XVarName                := Visible2.Name;
    YVarName                := Visible2.DataAsString;

    if (prResults.SelVarByName[XVarName] <> nil) and
       (prResults.SelVarByName[YVarName] <> nil) then begin

      prAddSel2Chart( XVarName, YVarName, True, True );
      if not Visible2.Find('AutoFit').IsNull then begin
        Charts2[Chart2Count - 1].AutoFit(Visible2['AutoFit']['Curve'].DataAsString,
                                         Visible2['AutoFit']['Eqn'].DataAsString,
                                         Visible2['AutoFit']['Params'].DataCount,
                                         TPADouble(Visible2['AutoFit']['Params'].DataPtr),
                                         TPAInteger(Visible2['AutoFit']['Const'].DataPtr));
      end;
    end;

    Visible2                := Visible2.Sibling;
  end;

  Visible2                  := prResults.VsVisPrefs.Child;
  while not Visible2.IsNull do begin
    XVarName                := Visible2.Name;
    YVarName                := Visible2.DataAsString;

    if (prResults.SelVarByName[XVarName] <> nil) and
       (prResults.SelVarByName[YVarName] <> nil) then begin

      prAddSel2Chart( XVarName, YVarName, False );
    end;

    Visible2                := Visible2.Sibling;
  end;

  OnResize                  := prOnResize;

  SelHistBinEvents.AddEvent('HaveHist', prOnHaveHist);
  SelHistBinEvents.AddEvent('HaveNoHist', prOnHaveNoHist);
end;

destructor TQUB_Results_SelectPanel.Destroy;
begin
  try
    SelHistBinEvents.DelEvent('HaveHist', prOnHaveHist);
    SelHistBinEvents.DelEvent('HaveNoHist', prOnHaveNoHist);
    
    prCharts.Free;   // elements auto deleted by owner?
    pr2Charts.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Results_SelectPanel.prSetIdealAspect(x :Double);
begin
  prIdealAspect := x;
  prOnResize(Self);
end;

procedure TQUB_Results_SelectPanel.prSetChosenColor(x :Integer);
begin
  prChosenColor    := x;
  panColors.Chosen := x;
end;

procedure TQUB_Results_SelectPanel.prSetChosenSeg(x :Integer);
var
  i                :Integer;
begin
  prChosenSeg := x;
  for i := 0 to ChartCount - 1 do begin
    Charts[i].ChosenPoint := x;
  end;
  for i := 0 to Chart2Count - 1 do begin
    Charts2[i].ChosenPoint := x;
  end;
end;

function TQUB_Results_SelectPanel.prGetChartCount: Integer;
begin
  Result := prCharts.Count;
end;

function TQUB_Results_SelectPanel.prGetChart(Ix :Integer): TQUB_Results_SelectChart;
begin
  if (Ix >= 0) and (Ix < prCharts.Count) then begin
    Result := TQUB_Results_SelectChart( prCharts.Objects[Ix] );
  end else begin
    Result := nil;
  end;
end;

function TQUB_Results_SelectPanel.prGet2ChartCount: Integer;
begin
  Result := pr2Charts.Count;
end;

function TQUB_Results_SelectPanel.prGet2Chart(Ix :Integer): TQUB_Results_Select2Chart;
begin
  if (Ix >= 0) and (Ix < pr2Charts.Count) then begin
    Result := TQUB_Results_Select2Chart( pr2Charts[Ix] );
  end else begin
    Result := nil;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnResize(Sender :TObject);
var
  ChartCount    :Integer;
  R, C          :Integer; // #rows, #cols
  W, H          :Integer; // of a single chart
  i, j, k       :Integer; // cur chart ix, row, col
begin
  ChartCount    := prCharts.Count + pr2Charts.Count;
  if ChartCount = 0 then Exit;

  R             := 1;
  C             := 1;
  W             := panCharts.Width;
  H             := panCharts.Height;

  while R * C < ChartCount do begin
    if prIdealAspect > (W / H) then R := R + 1
                               else C := C + 1;
    W := panCharts.Width  div C;
    H := panCharts.Height div R;
  end;

  while (R > 1) and ((R - 1) * C >= ChartCount) do begin
    R := R - 1;
  end;

  while (C > 1) and ((C - 1) * R >= ChartCount) do begin
    C := C - 1;
  end;

  W := panCharts.Width  div C;
  H := panCharts.Height div R;

  j             := 0;
  k             := 0;
  for i := 0 to prCharts.Count - 1 do begin
    Charts[i].SetBounds( k * W, j * H, W, H );

    k := k + 1;
    if k >= C then begin
      k := 0;
      j := j + 1;
    end;
  end;
  for i := 0 to pr2Charts.Count - 1 do begin
    Charts2[i].SetBounds( k * W, j * H, W, H );

    k := k + 1;
    if k >= C then begin
      k := 0;
      j := j + 1;
    end;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnColorClick(Sender: TObject; Color :Integer);
begin
  prChosenColor    := Color;
  if Assigned(OnColorChange) then OnColorChange(Self);
end;

procedure TQUB_Results_SelectPanel.prOnColorPopup(Sender: TObject);
begin
  prChosenColor        := TPanel( prColorPopup.PopupComponent ).Tag;
  panColors.Chosen     := prChosenColor;

  prColorPopup.Items.Clear;
  PyCmd_FillMenu(prColorPopup.Items, 'SelectColor', prOnColorPopupChoice);

  if Assigned(OnColorChange) then OnColorChange(Self);
end;

procedure TQUB_Results_SelectPanel.prOnColorPopupChoice(Sender: TObject);
begin
  PythonThread.InterpretNonSilently(
       Format('Commands.CallByName( "%s", "%s" )', ['SelectColor', TMenuItem(Sender).Caption]) );
end;

procedure TQUB_Results_SelectPanel.prOnChangeHistBins(Sender :TObject);
begin
  QUB_Results_SetHistBins(tbHistBins.Position);
end;

procedure TQUB_Results_SelectPanel.prOnHaveHist(Sender :TObject);
begin
  tbHistBins.Visible  := True;
  lblHistBins.Visible := True;
end;

procedure TQUB_Results_SelectPanel.prOnHaveNoHist(Sender :TObject);
begin
  tbHistBins.Visible  := False;
  lblHistBins.Visible := False;
end;

procedure TQUB_Results_SelectPanel.prOnChangeSD(Sender :TObject);
begin
  if prChangingSDs then Exit;

  prChangingSDs     := True;
  try
    prNumSDs        := StrToFloat(edSDs.Text);
    tbSDs.Position  := Round(prNumSDs * 10);
    prUpdateSD;
  except
  end;
  prChangingSDs     := False;
end;

procedure TQUB_Results_SelectPanel.prOnChangeSDSlider(Sender :TObject);
var
  newsd             :Double;
begin
  if prChangingSDs then Exit;

  prChangingSDs     := True;

  newsd             := tbSDs.Position * 0.1;
  edSDs.Text        := Format('%f', [NumSD]);

  if newsd <> prNumSDs then begin
    prNumSDs        := newsd;
    prUpdateSD;
  end;

  prChangingSDs     := False;
end;

procedure TQUB_Results_SelectPanel.prUpdateSD;
var
  i             :Integer;
begin
  for i := 0 to prResults.SelVarCount - 1 do begin
    prResults.SelVar[i].NumSDForLines := prNumSDs;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClickVar(Sender :TObject);
var
  i       :Integer;
begin
  with Sender as TComboBox do begin
    if ItemIndex = 1 then begin
      for i := 4 to mnuVars.Items.Count - 1 do begin
        prAddSelChart( mnuVars.Items[i] );
      end;
    end else if ItemIndex = 2 then begin
      while ChartCount > 0 do begin
        prRemSelChart( prCharts.Strings[ prCharts.Count - 1 ] );
      end;
      while Chart2Count > 0 do begin
        prRemSel2Chart( Chart2Count - 1 );
      end;
    end else if ItemIndex = 3 then begin
      prPromptAdd2Chart;
    end else if ItemIndex > 3 then begin
      prAddSelChart( Text );
    end;

    if ItemIndex <> 0 then ItemIndex := 0;
  end;
end;

procedure TQUB_Results_SelectPanel.prAddSelChart( Name :String );
var
  SelVar              :TQUB_Results_Var;
  Chart               :TQUB_Results_SelectChart;
begin
  if prCharts.IndexOf( Name ) = -1 then begin
    SelVar                    := prResults.SelVarByName[ Name ];
    SelVar.Visible            := True;
    Chart                     := TQUB_Results_SelectChart.Create(Self, SelVar);
    Chart.Parent              := panCharts;
    Chart.OnClickPoint        := prOnClickPoint;
    Chart.OnSelectPoints      := prOnSelectPoints;
    Chart.PopupMenu           := prChartPopup;
    Chart.PointPopupMenu      := prPointPopup;
    Chart.HideButton.OnClick  := prOnClickHide;
    // Chart.ChosenPoint         := prChosenSeg;

    prCharts.Add( Name );
    prCharts.Objects[ prCharts.IndexOf(Name) ] := Chart;

    prOnResize(Self);
  end;
end;

procedure TQUB_Results_SelectPanel.prRemSelChart( Name :String );
var
  Ix                  :Integer;
  Chart               :TQUB_Results_SelectChart;
begin
  Ix                  := prCharts.IndexOf( Name );
  if Ix <> -1 then begin
    prResults.SelVarByName[ Name ].Visible := False;
    Chart                                  := Charts[Ix];
    Chart.Parent                           := nil;
    Chart.Free;
    prCharts.Delete(Ix);
    prOnResize(Self);
  end;
end;

procedure TQUB_Results_SelectPanel.prPromptAdd2Chart;
var
  Chooser                   :TQUB_Results_VersusDialog;
begin
  Chooser                   := TQUB_Results_VersusDialog.Create(nil);
  Chooser.Results           := prResults;
  if Chooser.ShowModal = mrOK then begin
    prAddSel2Chart( Chooser.XVar, Chooser.YVar, True );
  end;
  Chooser.Free;
end;

procedure TQUB_Results_SelectPanel.prAddSel2Chart( XName, YName :String; DoPrefs :Boolean; Force :Boolean = False );
var
  XSelVar, YSelVar          :TQUB_Results_Var;
  Chart                     :TQUB_Results_Select2Chart;
  Visible2                  :INode;
  Recent2                   :INode;
  RecentCount               :Integer;
  ThisRecentEntry           :INode;
  i                         :Integer;
begin
  XSelVar                   := prResults.SelVarByName[ XName ];
  YSelVar                   := prResults.SelVarByName[ YName ];
  if (XSelVar = nil) or (YSelVar = nil) then Exit;

  if not Force then begin
    // don't add it if it's already visible
    for i := 0 to Chart2Count - 1 do begin
      if (Charts2[i].XSelVar = XSelVar) and (Charts2[i].YSelVar = YSelVar) then begin
        Exit;
      end;
    end;
  end;

  Chart                     := TQUB_Results_Select2Chart.Create(Self, XSelVar, YSelVar);
  Chart.Parent              := panCharts;
  Chart.OnClickPoint        := prOnClick2Point;
  Chart.OnSelectPoints      := prOnSelect2Points;
  Chart.HideButton.OnClick  := prOnClick2Hide;
  Chart.PopupMenu           := pr2ChartPopup;
  Chart.PointPopupMenu      := prPointPopup;

  pr2Charts.Add( Chart );

  prOnResize(Self);

  if DoPrefs then begin
    // update visibility prefs:
    Visible2                  := TNode.Create(XName);
    Visible2.SetData( YName );
    prResults.VsVisPrefs.AppendChild( Visible2 );

    // update recent prefs: if it's there, remove it first so the list is ordered in increasing recentness
    Recent2                   := prResults.VsRecentPrefs.Child;
    ThisRecentEntry           := TNode.CreateNull;
    RecentCount               := 0;
    while not Recent2.IsNull do begin
      if ThisRecentEntry.IsNull and (Recent2.Name = XName) and (Recent2.DataAsString = YName) then begin
        ThisRecentEntry       := Recent2;
      end else begin
        RecentCount           := RecentCount + 1;
      end;

      Recent2                 := Recent2.Sibling;
    end;
    if RecentCount >= SELECT_RECENT2_MAX then prResults.VsRecentPrefs.RemoveChild( prResults.VsRecentPrefs.Child );
    if ThisRecentEntry.IsNull then begin
      ThisRecentEntry         := Visible2.Clone(True);
    end else begin
      prResults.VsRecentPrefs.RemoveChild( ThisRecentEntry );
    end;

    prResults.VsRecentPrefs.AppendChild( ThisRecentEntry );
  end;
end;

procedure TQUB_Results_SelectPanel.prRemSel2Chart( Ix :Integer );
var
  Chart               :TQUB_Results_Select2Chart;
  Visible2            :INode;
  XName, YName        :String;
begin
  Chart               := Charts2[Ix];
  if Chart <> nil then begin
    Visible2                  := prResults.VsVisPrefs.Child;
    XName                     := Chart.XSelVar.Name;
    YName                     := Chart.YSelVar.Name;
    while not Visible2.IsNull do begin
      if (Visible2.Name = XName) and (Visible2.DataAsString = YName) then begin
        prResults.VsVisPrefs.RemoveChild( Visible2 );
        Break;
      end;
      Visible2                := Visible2.Sibling;
    end;

    Chart.Parent                           := nil;
    Chart.Free;
    pr2Charts.Delete(Ix);
    prOnResize(Self);
  end;
end;

procedure TQUB_Results_SelectPanel.prOnChartPopup(Sender :TObject);
var
  SelVar                    :TQUB_Results_Var;
  i                         :Integer;
  Items                     :TStringList;
  Item                      :TMenuItem;
begin
  SelVar                    := TQUB_Results_SelectChart( prChartPopup.PopupComponent ).SelVar;
  itemXIndex.Checked        := SelVar.XStyle = XSTYLE_INDEX;
  itemXStart.Checked        := SelVar.XStyle = XSTYLE_START;
  itemXHist.Checked         := SelVar.XStyle = XSTYLE_HIST;
  itemYLog.Checked          := SelVar.YLog;
  for i := itemSendTo.Count - 1 downto 2 do begin
    itemSendTo[i].Free;
  end;
  Items                     := TStringList.Create;
  Items.Sorted              := True;
  TQUB_QtiPlot_Connection.ListScripts('Scatter', Items);
  for i := 0 to Items.Count - 1 do begin
    Item                    := TMenuItem.Create(Self);
    Item.Caption            := ChangeFileExt(Items[i], '');
    Item.OnClick            := prOnClickExtractQtiPlot;
    itemSendTo.Add( Item );
  end;
  Items.Free;
end;

procedure TQUB_Results_SelectPanel.prOnClickYLog(Sender :TObject);
begin
  with TQUB_Results_SelectChart( prChartPopup.PopupComponent ) do begin
    SelVar.YLog           := not SelVar.YLog;
    //RefreshPoints;
    //RefreshLines;
  end;

end;

procedure TQUB_Results_SelectPanel.prOnClickXStyle(Sender :TObject);
begin
  with TQUB_Results_SelectChart( prChartPopup.PopupComponent ) do begin
    SelVar.XStyle         := TQUB_Results_XStyle( TMenuItem(Sender).Tag );
    //RefreshPoints;
    //RefreshLines;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClickCurveFit(Sender :TObject);
begin
  with TQUB_Results_SelectChart( prChartPopup.PopupComponent ) do begin
    ShowFit;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClickCopyImage(Sender :TObject);
begin
  TChart(prChartPopup.PopupComponent).CopyToClipboardMetafile(False);
end;

procedure TQUB_Results_SelectPanel.prOnClickExtractNotepad(Sender :TObject);
var
  Chart     :TQUB_Results_SelectChart;
  Dlg       :TSaveDialog;
  FName     :String;
begin
  Chart     := TQUB_Results_SelectChart( prChartPopup.PopupComponent );
  if Chart.SeriesCount < 1 then begin
    MessageHelpDlg('No data available for extraction', mtError, [mbOK], 0, 'Error:No_Histogram_Data_Available');
    Exit;
  end;
  Dlg := TSaveDialog.Create(Self);
  try
    Dlg.FileName := Chart.SelVar.Name+'.txt';
    Dlg.Filter := 'Text files (*.txt)|*.txt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofOverwritePrompt];
    Dlg.Title := 'Save plot for notepad';
    if Dlg.Execute then begin
      FName := ChangeFileExt(Dlg.FileName, '.txt');
      prSaveHistogram(Chart, ChangeFileExt(Dlg.FileName, '.txt'));
      ShellExecute(Application.Handle, 'open', 'notepad', PChar(FName), PChar(ExtractFilePath(FName)), SW_SHOWNORMAL);
    end;
  finally
    Dlg.Free;
  end;
end;

function TQUB_Results_SelectPanel.prSaveHistogram(Chart :TChart; const FileName :String) :Boolean;
var
  Series      :TChartSeries;
  Text        :String;
  f           :File;
  i, j        :Integer;
  Entry       :Integer;
begin

  Result := False;

  Series := Chart.Series[0];

  AssignFile(f, FileName);
  try
    Rewrite(f, 1);
  except
    Result := True;
    MessageDlg('Cannot open file ' + FileName + ' for writing', mtError, [mbOK], 0);
    Exit;
  end;

  Entry := Console.NewEntry('Extracting histogram file ' + FileName + ' ...', COLOR_TEXT);

  try

    Text := '';

    for i := 0 to Chart.Series[0].Count - 1 do begin
      Text := Text + FloatToStr(Series.XValues[i]);
      for j := 0 to Chart.SeriesCount - 1 do begin
        Text := Text + #9 + FloatToStr(Chart.Series[j].YValues[i]);
      end;
      Text := Text + #13#10;
    end;
    i := Series.Count - 1;
    Text := Text + FloatToStr(Chart.Series[0].XValues[i]);
    for j := 0 to Chart.SeriesCount - 1 do begin
      Text := Text + #9 + FloatToStr(Chart.Series[j].YValues[i]);
    end;

    try
      BlockWrite(f, PChar(Text)^, Length(Text));
    except
      Result := True;
      MessageDlg('Could not write data to file ' + FileName, mtError, [mbOK], 0);
    end;

  finally

    try
      CloseFile(f);
    except
    end;

    Console.SetEntry(Entry, 'Extracting histogram file ' + FileName + ' ... Done.', COLOR_GRAY);
    //Beep;

  end;

end;


procedure TQUB_Results_SelectPanel.prOnClickExtractQtiPlot(Sender :TObject);
var
  Script    :String;
  Chart     :TQUB_Results_SelectChart;
  Hist      :INode;
begin
  Script    := ChangeFileExt(StringReplace(TMenuItem(Sender).Caption, '&', '', [rfReplaceAll]), '.py');
  Chart     := TQUB_Results_SelectChart( prChartPopup.PopupComponent );
  if Chart.SelVar.XStyle = XSTYLE_HIST then begin
    Hist    := Chart.SelVar.HistData;
  end else begin
    Hist    := TNode.CreateNull;
  end;
  prExtractQtiPlot(Script, Chart, Chart.SelVar.Name, Hist, Chart.Lbl);
end;

procedure TQUB_Results_SelectPanel.prExtractQtiPlot(Script :String; Chart :TChart; PlotTitle :String; HistData :INode; Lbl :String);
var
  QtiPlot           :TQUB_QtiPlot_Connection;
  Buf               :TPMSingle;
  Colors            :Array of Integer;
  i, j              :Integer;
begin
  with Chart do begin
    Buf               := nil;
    QtiPlot           := TQUB_QtiPlot_Connection.Create;
    try
      Buf             := SAlloc2(SeriesCount+2, Series[0].Count);
      for i := 0 to Series[0].Count - 1 do begin
        for j := 0 to SeriesCount + 1 do begin
          if j = 0 then begin
            Buf[j][i] := Series[SeriesCount-1].XValue[i]; // scatter x
          end else if j = 1 then begin
            Buf[j][i] := Series[0].XValue[i]; // fit x
          end else begin
            Buf[j][i] := Series[j-2].YValue[i];
          end;
        end;
      end;

      SetLength(Colors, Series[0].Count);
      for i := 0 to Series[0].Count - 1 do begin
        Colors[i] := prResults.SegColor[i];
      end;

      QtiPlot.BringToFront;
      QtiPlot.ExecScript('Scatter', Script);
      QtiPlot.SetArray(Buf[0], SeriesCount+2, Series[0].Count, 'float32', 'a');
      QtiPlot.SetArray(@Colors[0], Series[0].Count, 1, 'int32', 'colors');
      QtiPlot.Exec(Format('t,g,l = add_scatter("%s", "%s", "%s", a[0,:], a[-1,:], colors, None, *[a[i,:] for i in xrange(1, a.shape[0]-1)])',
                          [PlotTitle,
                           BottomAxis.Title.Caption,
                           LeftAxis.Title.Caption]));

      if not HistData.IsNull then begin
        QtiPlot.SetArray(HistData['X'].DataPtr, HistData['X'].DataCount, 1, 'float64', 'hist_x');
        QtiPlot.SetArray(HistData['Y'].DataPtr, HistData['Y'].DataCount, 1, 'float64', 'hist_y');
        QtiPlot.Exec('add_hist_to_scatter(t, g, l, hist_x, hist_y)');
      end;

      if 0 < Length(Lbl) then begin
        QtiPlot.Exec('add_lbl_to_scatter(t, g, l, "'+Lbl+'")');
      end;

    finally
      Free2(TPMChar(Buf));
      QtiPlot.Free;
    end;
  end;
end;


procedure TQUB_Results_SelectPanel.prOnClickExtractExcel(Sender :TObject);
begin
  prChartToExtract := TQUB_Results_SelectChart( prChartPopup.PopupComponent );
  PythonThread.SyncPyProc( prDoExtractExcel, True, True, -1 );
end;

procedure TQUB_Results_SelectPanel.prDoExtractExcel;
var
  ExtractFunc        :PPyObject;
  ArgList            :PPyObject;
  PyResult           :PPyObject;
  OneOrZero          :Integer;
begin
  with GetPythonEngine do begin
    ExtractFunc      := nil;
    ArgList          := nil;
    PyResult         := nil;
    try
      if prChartToExtract.SelVar.XStyle = XSTYLE_HIST then begin
        Run_CommandAsString('import HistogramCommands', single_input);
        ExtractFunc    := EvalString('HistogramCommands.ExtractToExcel');
        ArgList        := Py_BuildValue('(O)', [PyNode_FromNode(prChartToExtract.GetHistAsNode)]);
        PyResult       := PyEval_CallObject( ExtractFunc, ArgList );
        if PyResult = nil then PyErr_Print;
      end else begin
        OneOrZero      := 0;
        if prChartToExtract.SelVar.XStyle = XSTYLE_START then OneOrZero := 1;  // this should be unneeded soon?

        Run_CommandAsString('import SelectCommands', single_input);
        ExtractFunc    := EvalString('SelectCommands.ExtractChartToExcel');
        ArgList        := Py_BuildValue('(si)', [PChar(prChartToExtract.SelVar.Name),
                                                 OneOrZero]);
        PyResult       := PyEval_CallObject( ExtractFunc,  ArgList );
        if PyResult = nil then PyErr_Print;
      end;
    finally
      if ExtractFunc <> nil then Py_DECREF( ExtractFunc );
      if ArgList <> nil     then Py_DECREF( ArgList );
      if PyResult <> nil    then Py_DECREF( PyResult );
    end;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClickHide(Sender :TObject);
begin
  PostMessage( Self.Handle, WM_QUB_SELCHART_HIDE, WPARAM(Sender), LPARAM(0) );
end;

procedure TQUB_Results_SelectPanel.prOnHideMsg(var Message: TMessage);
var
  Chart  :TQUB_Results_SelectChart;
  Button :TButton;
begin
  Button := TButton(Message.WParam);
  Chart  := TQUB_Results_SelectChart( Button.Parent );
  prRemSelChart( Chart.SelVar.Name );
  SetFocus;
end;

procedure TQUB_Results_SelectPanel.prOn2ChartPopup(Sender :TObject);
var
  Chart                     :TQUB_Results_Select2Chart;
  i                         :Integer;
  Items                     :TStringList;
  Item                      :TMenuItem;
begin
  Chart                     := TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent );
  item2XLog.Checked         := Chart.XSelVar.YLog;
  item2YLog.Checked         := Chart.YSelVar.YLog;
  for i := item2SendTo.Count - 1 downto 2 do begin
    item2SendTo[i].Free;
  end;
  Items                     := TStringList.Create;
  Items.Sorted              := True;
  TQUB_QtiPlot_Connection.ListScripts('Scatter', Items);
  for i := 0 to Items.Count - 1 do begin
    Item                    := TMenuItem.Create(Self);
    Item.Caption            := ChangeFileExt(Items[i], '');
    Item.OnClick            := prOnClick2ExtractQtiPlot;
    item2SendTo.Add( Item );
  end;
  Items.Free;
end;

procedure TQUB_Results_SelectPanel.prOnClick2XLog(Sender :TObject);
begin
  with TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent ) do begin
    XSelVar.YLog           := not XSelVar.YLog;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClick2YLog(Sender :TObject);
begin
  with TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent ) do begin
    YSelVar.YLog           := not YSelVar.YLog;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClick2SwapAxes(Sender :TObject);
begin
  with TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent ) do begin
    SwapAxes;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClick2CurveFit(Sender :TObject);
begin
  with TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent ) do begin
    ShowFit;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClick2CopyImage(Sender :TObject);
begin
  TChart(pr2ChartPopup.PopupComponent).CopyToClipboardMetafile(False);
end;

procedure TQUB_Results_SelectPanel.prOnClick2ExtractNotepad(Sender :TObject);
var
  Chart     :TQUB_Results_Select2Chart;
  Dlg       :TSaveDialog;
  FName     :String;
begin
  Chart     := TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent );
  if Chart.SeriesCount < 1 then begin
    MessageHelpDlg('No data available for extraction', mtError, [mbOK], 0, 'Error:No_Histogram_Data_Available');
    Exit;
  end;
  Dlg := TSaveDialog.Create(Self);
  try
    Dlg.FileName := Chart.XSelVar.Name+'_v_'+Chart.YSelVar.Name+'.txt';
    Dlg.Filter := 'Text files (*.txt)|*.txt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofOverwritePrompt];
    Dlg.Title := 'Save plot for notepad';
    if Dlg.Execute then begin
      FName := ChangeFileExt(Dlg.FileName, '.txt');
      prSaveHistogram(Chart, ChangeFileExt(Dlg.FileName, '.txt'));
      ShellExecute(Application.Handle, 'open', 'notepad', PChar(FName), PChar(ExtractFilePath(FName)), SW_SHOWNORMAL);
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClick2ExtractQtiPlot(Sender :TObject);
var
  Chart     :TQUB_Results_Select2Chart;
  Script    :String;
begin
  Script    := ChangeFileExt(StringReplace(TMenuItem(Sender).Caption, '&', '', [rfReplaceAll]), '.py');
  Chart     := TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent );
  prExtractQtiPlot(Script, Chart, Chart.XSelVar.Name+' v '+Chart.YSelVar.Name,
                   TNode.CreateNull, Chart.Lbl);
end;

procedure TQUB_Results_SelectPanel.prOnClick2ExtractExcel(Sender :TObject);
begin
  pr2ChartToExtract := TQUB_Results_Select2Chart( pr2ChartPopup.PopupComponent );
  PythonThread.SyncPyProc( prDo2ExtractExcel, True, True, -1 );
end;

procedure TQUB_Results_SelectPanel.prDo2ExtractExcel;
var
  ExtractFunc        :PPyObject;
  ArgList            :PPyObject;
  PyResult           :PPyObject;
begin

  with GetPythonEngine do begin
    ExtractFunc      := nil;
    ArgList          := nil;
    PyResult         := nil;
    try
      Run_CommandAsString('import SelectCommands', single_input);
      ExtractFunc    := EvalString('SelectCommands.Extract2ChartToExcel');
      ArgList        := Py_BuildValue('(ss)', [PChar(pr2ChartToExtract.XSelVar.Name),
                                               PChar(pr2ChartToExtract.YSelVar.Name)]);
      PyResult       := PyEval_CallObject( ExtractFunc,  ArgList );
      if PyResult = nil then PyErr_Print;
    finally
      if ExtractFunc <> nil then Py_DECREF( ExtractFunc );
      if ArgList <> nil     then Py_DECREF( ArgList );
      if PyResult <> nil    then Py_DECREF( PyResult ) else PyErr_Clear;
    end;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnClick2Hide(Sender :TObject);
begin
  PostMessage( Self.Handle, WM_QUB_SELCHART_2HIDE, WPARAM(Sender), LPARAM(0) );
end;

procedure TQUB_Results_SelectPanel.prOnHide2Msg(var Message: TMessage);
var
  Chart  :TQUB_Results_Select2Chart;
  Button :TButton;
begin
  Button := TButton(Message.WParam);
  Chart  := TQUB_Results_Select2Chart( Button.Parent );
  prRemSel2Chart( pr2Charts.IndexOf(Chart) );
  SetFocus;
end;

procedure TQUB_Results_SelectPanel.prOnClickPoint(Sender :TQUB_Results_Chart; Ix :Integer);
begin
  if prResults.SegColor[Ix] = prChosenColor then begin
    prResults.SegColor[Ix]  := 0;
  end else begin
    prResults.SegColor[Ix]  := prChosenColor;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnSelectPoints(Sender :TQUB_Results_Chart; Points :TList);
var
  i                         :Integer;
begin
  prResults.DoUpdates       := False;
  for i := 0 to Points.Count - 1 do begin
    prResults.SegColor[ Integer(Points[i]) ] := prChosenColor;
  end;
  prResults.DoUpdates       := True;
  prResults.UpdateSelection;
end;

procedure TQUB_Results_SelectPanel.prOnClick2Point(Sender :TQUB_Results_2Chart; Ix :Integer);
begin
  if prResults.SegColor[Ix] = prChosenColor then begin
    prResults.SegColor[Ix]  := 0;
  end else begin
    prResults.SegColor[Ix]  := prChosenColor;
  end;
end;

procedure TQUB_Results_SelectPanel.prOnSelect2Points(Sender :TQUB_Results_2Chart; Points :TList);
var
  i                         :Integer;
begin
  prResults.DoUpdates       := False;
  for i := 0 to Points.Count - 1 do begin
    prResults.SegColor[ Integer(Points[i]) ] := prChosenColor;
  end;
  prResults.DoUpdates       := True;
  prResults.UpdateSelection;
end;

procedure TQUB_Results_SelectPanel.prOnPointPopup(Sender :TObject);
var
  SegIx       :Integer;
begin
  if prPointPopup.PopupComponent is TQUB_Results_SelectChart then begin
    SegIx := TQUB_Results_SelectChart(prPointPopup.PopupComponent).PopupPoint;
  end else begin
    SegIx := TQUB_Results_Select2Chart(prPointPopup.PopupComponent).PopupPoint;
  end;
  itemShowSeg.Caption := 'Show Segment ' + IntToStr(SegIx+1);
end;

procedure TQUB_Results_SelectPanel.prOnClickShowSeg(Sender :TObject);
begin
  if Assigned(OnShowSegClick) then begin
    if prPointPopup.PopupComponent is TQUB_Results_SelectChart then begin
      OnShowSegClick(Self, TQUB_Results_SelectChart(prPointPopup.PopupComponent).PopupPoint);
    end else begin
      OnShowSegClick(Self, TQUB_Results_Select2Chart(prPointPopup.PopupComponent).PopupPoint);
    end;
  end;
end;


// *****************************************************************************
end.

