{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the dialog in Results: Select when you choose Show Var: Custom...


unit QUB_Results_VersusDialog;


interface


uses

Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls,
Buttons, ExtCtrls,

QUB_Results;

type
  TQUB_Results_VersusDialog = class(TForm)
    OKBtn                 :TButton;
    CancelBtn             :TButton;
    cmbXVar               :TComboBox;
    cmbYVar               :TComboBox;
    Label1                :TLabel;
    Label2                :TLabel;
    Label3: TLabel;
    cmbRecent: TComboBox;
    procedure cmbRecentClick(Sender: TObject);

  protected

    prResults             :TQUB_Results;
    prRecentX             :TStringList;
    prRecentY             :TStringList;

    procedure prSetResults(Results :TQUB_Results);
    function prGetXVar: String;
    function prGetYVar: String;

  public

    constructor Create(AOwner :TComponent); override;
    destructor Destroy; override;

    property Results      :TQUB_Results        read prResults       write prSetResults;
    property XVar         :String              read prGetXVar;
    property YVar         :String              read prGetYVar;

  end;


implementation

{$R *.DFM}


uses


QUB_Results_Select,
QUB_Node;




constructor TQUB_Results_VersusDialog.Create(AOwner :TComponent);
begin
  inherited;

  prRecentX                 := TStringList.Create;
  prRecentY                 := TStringList.Create;
end;

destructor TQUB_Results_VersusDialog.Destroy;
begin
  try
    prRecentX.Free;
    prRecentY.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Results_VersusDialog.prSetResults(Results :TQUB_Results);
var
  i                         :Integer;
  Recent                    :INode;
  XName, YName              :String;
begin
  prResults                 := Results;

  cmbXVar.Clear;
  cmbYVar.Clear;
  cmbXVar.Style             := csDropDownList;
  cmbYVar.Style             := csDropDownList;
  cmbXVar.DropDownCount     := 20;
  cmbYVar.DropDownCount     := 20;

  for i := 0 to Results.SelVarCount - 1 do begin
    cmbXVar.Items.Add( Results.SelVar[i].Name );
    cmbYVar.Items.Add( Results.SelVar[i].Name );
    if Results.SelVar[i].Name = 'Index' then cmbXVar.ItemIndex := i;
  end;

  if cmbXVar.ItemIndex < 0 then cmbXVar.ItemIndex         := 0;
  if cmbYVar.ItemIndex < 0 then cmbYVar.ItemIndex         := 0;

  prRecentX.Clear;
  prRecentY.Clear;
  cmbRecent.Clear;
  cmbRecent.Style           := csDropDownList;
  cmbRecent.DropDownCount   := SELECT_RECENT2_MAX;

  Recent                    := Results.VsRecentPrefs.Child;
  while not Recent.IsNull do begin
    XName                   := Recent.Name;
    YName                   := Recent.DataAsString;

    if (Results.SelVarByName[XName] <> nil) and (Results.SelVarByName[YName] <> nil) then begin
      prRecentX.Add( XName );
      prRecentY.Add( YName );
      cmbRecent.Items.Add( XName + ' vs. ' + YName );
    end;

    Recent                  := Recent.Sibling;
  end;
end;

procedure TQUB_Results_VersusDialog.cmbRecentClick(Sender: TObject);
begin
  cmbXVar.ItemIndex         := cmbXVar.Items.IndexOf( prRecentX[ cmbRecent.ItemIndex ] );
  cmbYVar.ItemIndex         := cmbYVar.Items.IndexOf( prRecentY[ cmbRecent.ItemIndex ] );
end;

function TQUB_Results_VersusDialog.prGetXVar: String;
begin
  Result                    := cmbXVar.Text;
end;

function TQUB_Results_VersusDialog.prGetYVar: String;
begin
  Result                    := cmbYVar.Text;
end;



//******************************************************************************
end.
