{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the main component in QUB_Viewport_Report


unit QUB_Rich_Console;


interface


uses


Windows, Messages, SysUtils, Classes, SyncObjs, Graphics, Controls, Forms, StdCtrls, ComCtrls, RichEdit;


type


TQUB_Rich_Console_ColorCode  = (COLOR_BACK, COLOR_TEXT, COLOR_GRAY, COLOR_ERROR, COLOR_WARNING);


TQUB_Rich_Console = class(TRichEdit)
protected


  prCriticalSection        :TCriticalSection;

  //prSaveText               :TRichEdit;

  //Colors
  prColors                 :Array[TQUB_Rich_Console_ColorCode] of TColor;

  prBlackAndWhite          :Boolean;

  prMaxLines               :Integer;


  function prGetColors(Index :TQUB_Rich_Console_ColorCode) :TColor;
  procedure prSetColors(Index :TQUB_Rich_Console_ColorCode; Color :TColor);

  procedure CNKeyDown(var Message: TWMKeyDown); message CN_KEYDOWN;

  //procedure CreateWnd; override;
  //procedure DestroyWnd; override;


public


  constructor Create(AOwner: TComponent); override;
  destructor Destroy; override;

  procedure Redraw;

  procedure Initialize;  // to be called once it has a parent

  function NewEntry(const Text :String; ColorCode :TQUB_Rich_Console_ColorCode) :Integer;
  function GetEntry(Entry :Integer) :String;
  function SetEntry(Entry :Integer; const Text :String; ColorCode :TQUB_Rich_Console_ColorCode) :Boolean;


  property Colors[Index :TQUB_Rich_Console_ColorCode] :TColor     read prGetColors      write prSetColors;
  property BlackAndWhite                              :Boolean    read prBlackAndWhite  write prBlackAndWhite;
  property MaxLines                                   :Integer    read prMaxLines       write prMaxLines;


end;


procedure Register;


//******************************************************************************
implementation


//******************************************************************************
constructor TQUB_Rich_Console.Create(AOwner: TComponent);
begin
  inherited;
  prCriticalSection := TCriticalSection.Create;

  prColors[COLOR_BACK]    := clNavy;
  prColors[COLOR_TEXT]    := clWhite;
  prColors[COLOR_GRAY]    := clGray;
  prColors[COLOR_ERROR]   := clRed;
  prColors[COLOR_WARNING] := clYellow;

end;

destructor TQUB_Rich_Console.Destroy;
begin
  try
    prCriticalSection.Free;
    //prSaveText.Free;
  finally
    inherited;
  end;
end;

//TO DO: LM. I tried to fix the issue with the console losing control of the Rich formatting, a la SelectionLists. Don't know yet how to do it.
{
procedure TQUB_Rich_Console.CreateWnd;
begin
  inherited;
  if prSaveText <> nil then begin
    Assign(prSaveText.Lines);
    prSaveText.Free;
    prSaveText := nil;
  end;
end;

procedure TQUB_Rich_Console.DestroyWnd;
begin
  if Text <> '' then begin
    prSaveText := TRichEdit.Create(Application);
    prSaveText.Visible := False;
    prSaveText.Parent := Application.MainForm;
    prSaveText.PlainText := False;
    prSaveText.Assign(Self.Lines);
  end;
  inherited;
end;
}

procedure TQUB_Rich_Console.Initialize;
begin
  BlackAndWhite           := False;
  MaxLines                := 200;
  HideSelection           := False; // added 6-10-2004 - window will not autoscroll w/o this.
end;

procedure TQUB_Rich_Console.CNKeyDown(var Message: TWMKeyDown);
var
  Eaten :Boolean;
  l     :Integer;
begin

  Eaten := False;

  case Message.CharCode of
    Word('Y')://Delete current line
      begin
        if GetAsyncKeyState(VK_CONTROL) < 0 then begin
          l := SendMessage(Handle, EM_EXLINEFROMCHAR, 0, -1);
          Lines.Delete(l);
          Eaten := True;
        end;
      end;
  end;

  if not Eaten then inherited;

end;


function TQUB_Rich_Console.NewEntry(const Text :String; ColorCode :TQUB_Rich_Console_ColorCode) :Integer;
var
  l {i}  :Integer;
  ss, sl :Integer;
begin

  try

    prCriticalSection.Acquire;

    SendMessage(Handle, WM_SETREDRAW, 0, 0);

    //Delete Lines is necessary
    if Lines.Count > MaxLines then begin
      repeat
        Lines.Delete(0);
      until(Lines.Count <= MaxLines);
    end;

    {l := 0;
    for i := 0 to Lines.Count - 1 do begin
      Inc(l, Length(Lines.Strings[i]) + 2);
    end; }
    l := Length(Self.Text);

    Lines.Add(Text);
    ss := SelStart;
    sl := SelLength;

    try
      SelStart  := l;
      SelLength := Length(Text) + 2; // Length(Lines[Lines.Count - 1]);
      if BlackAndWhite then begin
        SelAttributes.Color := clBlack;
      end else begin
        SelAttributes.Color := prColors[ColorCode];
      end;
      SelAttributes.Name := Font.Name;
    finally
      SelStart  := ss;
      SelLength := sl;
    end;

    SendMessage(Self.Handle, EM_SCROLLCARET, 0, 0);
    Result := Lines.Count - 1;

  finally

    SendMessage(Handle, WM_SETREDRAW, 1, 0);
    Refresh;
    prCriticalSection.Leave;

  end;
end;

function TQUB_Rich_Console.GetEntry(Entry :Integer) :String;
begin

  try

    prCriticalSection.Acquire;

    if (Entry >= 0) and (Entry < Lines.Count) then begin
      Result := Lines.Strings[Entry];
    end else begin
      Result := '';
    end;

  finally

    prCriticalSection.Leave;

  end;
end;

function TQUB_Rich_Console.SetEntry(Entry :Integer; const Text :String; ColorCode :TQUB_Rich_Console_ColorCode) :Boolean;
var
  l, i   :Integer;
  ss, sl :Integer;
begin

  try

    prCriticalSection.Acquire;

    SendMessage(Handle, WM_SETREDRAW, 0, 0);

    l := 0;
    for i := 0 to Entry - 1 do begin
      Inc(l, Length(Lines.Strings[i]) + 2);
    end;
    ///l := Length(Self.Text);
    ss := SelStart;
    sl := SelLength;
    if (Entry >= 0) and (Entry < Lines.Count) then begin
      Result := False;
      try
        SelStart  := l;
        SelLength := Length(Lines.Strings[Entry]);
        if BlackAndWhite then begin
          SelAttributes.Color := clBlack;
        end else begin
          SelAttributes.Color := prColors[ColorCode];
        end;
        SelAttributes.Name := Font.Name;
        SelText := Text;
      finally
        SelStart  := ss;
        SelLength := sl;
      end;
    end else begin
      Result := True;
    end;

  finally

    SendMessage(Handle, WM_SETREDRAW, 1, 0);
    Refresh;
    prCriticalSection.Leave;

  end;
end;


//******************************************************************************
function TQUB_Rich_Console.prGetColors(Index :TQUB_Rich_Console_ColorCode) :TColor;
begin
  Result := prColors[Index];
end;

procedure TQUB_Rich_Console.Redraw;
var
  ss, sl :Integer;
begin
  ss := 0;
  sl := 0;
  try
    prCriticalSection.Acquire;
    SendMessage(Handle, WM_SETREDRAW, 0, 0);
    ss := SelStart;
    sl := SelLength;
    SelStart  := 0;
    SelLength := Length(Text);
    if BlackAndWhite then begin
      SelAttributes.Color := clBlack;
      Self.Color := clWhite;
    end else begin
      Self.Color := prColors[COLOR_BACK];
    end;
  finally
    SelStart  := ss;
    SelLength := sl;
    SendMessage(Handle, WM_SETREDRAW, 1, 0);
    Refresh;
    prCriticalSection.Leave;
  end;
end;

procedure TQUB_Rich_Console.prSetColors(Index :TQUB_Rich_Console_ColorCode; Color :TColor);
begin
  prColors[Index] := Color;
end;

//******************************************************************************
procedure Register;
begin
  RegisterComponents('QUB', [TQUB_Rich_Console]);
end;


//******************************************************************************
end.
