{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Sin_Model;


interface


uses


Classes,
AllocTypes, Alloc,
qubopt_iface,
QUB_Matrices,

QUB_Fit_Model,
QUB_Q_Global, QUB_Exp_Global;


type


TQUB_Sin_Model                 = class(TQUB_Fit_Model)
protected


  procedure prAllocModelStuff; override;
  procedure prInitModelStuff; override;

  function prGet_cParCount :Integer; override;


public


  //Ieq ........................................................................
  Ieq                        :Double;
  dSSdIeq                    :Double;

  //Amp ........................................................................
  Amp                        :Double;
  dSSdAmp                    :Double;

  //Freq .......................................................................
  Freq                       :Double;
  dSSdFreq                   :Double;

  //Phase ......................................................................
  Phase                      :Double;
  dSSdPhase                  :Double;

  FixedIeq                   :Boolean;
  FixedAmp                   :Boolean;
  FixedFreq                  :Boolean;
  FixedPhase                 :Boolean;

  IdxIeq                     :Integer;
  IdxAmp                     :Integer;
  IdxFreq                    :Integer;
  IdxPhase                   :Integer;


  constructor Create(var OptParams :TQUB_Exp_Opt_Params); override;


  function ResetSSandGR :Boolean; override;

  function Do_Mean_t(var mt :Double; t :Double) :Integer; override;
  function Do_dMean_t_dIeq(var dmt :Double; t :Double) :Integer;
  function Do_dMean_t_dAmp(var dmt :Double; t :Double) :Integer;
  function Do_dMean_t_dFreq(var dmt :Double; t :Double) :Integer;
  function Do_dMean_t_dPhase(var dmt :Double; t :Double) :Integer;

  function Do_fParsToModel :Integer; override;

  function Do_InitcPars :Integer; override;

  function Do_dSSdfPar :Integer; override;
  function Do_Varf2Varc :Integer; override;


end;


implementation


uses


QUB_Node;


//******************************************************************************
constructor TQUB_Sin_Model.Create(var OptParams :TQUB_Exp_Opt_Params);
begin

  Self.FixedIeq   := OptParams.FixedIeq;
  Self.FixedAmp   := OptParams.FixedAmp;
  Self.FixedFreq  := OptParams.FixedFreq;
  Self.FixedPhase := OptParams.FixedPhase;

  if FixedIeq   then IdxIeq   := -1      else IdxIeq   := 0;
  if FixedAmp   then IdxAmp   := IdxIeq  else IdxAmp   := IdxIeq  + 1;
  if FixedFreq  then IdxFreq  := IdxAmp  else IdxFreq  := IdxAmp  + 1;
  if FixedPhase then IdxPhase := IdxFreq else IdxPhase := IdxFreq + 1;

  inherited Create(OptParams);

end;

function TQUB_Sin_Model.ResetSSandGR :Boolean;
begin

  Result := inherited ResetSSandGR;

  //dSSdIeq
  dSSdIeq   := 0;
  //dSSdAmp
  dSSdAmp   := 0;
  //dSSdFreq
  dSSdFreq  := 0;
  //dSSdPhase
  dSSdPhase := 0;

end;


//****************************************************************************//
//                                                                            //
//                              P A R A M E T E R S                           //
//                                                                            //
//****************************************************************************//
function TQUB_Sin_Model.Do_fParsToModel :Integer;
begin

  Result := inherited Do_fParsToModel;

  //cParsToModel
  if not FixedIeq   then Ieq   :=     cPars[IdxIeq];
  if not FixedAmp   then Amp   :=     cPars[IdxAmp];
  if not FixedFreq  then Freq  := Exp(cPars[IdxFreq]);
  if not FixedPhase then Phase :=     cPars[IdxPhase];

end;


//****************************************************************************//
//                                                                            //
//                                  M O D E L                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Sin_Model.prAllocModelStuff;
begin

  inherited;

  //Temporary quantities
  Vec1                       := DAlloc1(3);
  Vec2                       := DAlloc1(3);
  Mat1                       := DAlloc2(3, 3);
  Mat2                       := DAlloc2(3, 3);

end;

procedure TQUB_Sin_Model.prInitModelStuff;
begin

  inherited;

  Ieq   := prOptParams.SineIeq;
  Amp   := prOptParams.SineAmp;
  Freq  := prOptParams.SineFreq;
  Phase := prOptParams.SinePhase;
end;



//****************************************************************************//
//                                                                            //
//                              O P T I M I Z E R                             //
//                                                                            //
//****************************************************************************//
function TQUB_Sin_Model.Do_InitcPars :Integer;
begin

  Result := inherited Do_InitcPars;

  if not FixedIeq   then cPars[IdxIeq]   := Ieq;
  if not FixedAmp   then cPars[IdxAmp]   := Amp;
  if not FixedFreq  then cPars[IdxFreq]  := ln(Freq);
  if not FixedPhase then cPars[IdxPhase] := Phase;

end;


//****************************************************************************//
//                                                                            //
//                        L O G  -  L I K E L I H O O D                       //
//                                                                            //
//****************************************************************************//

//Current mean
function TQUB_Sin_Model.Do_Mean_t(var mt :Double; t :Double) :Integer;
begin

  Result := 0;

  //mt = Ieq + Amp.sin(2.PI.Freq.t + Phase)
  mt := Ieq + Amp*sin(2*PI*Freq*t + Phase);

end;

function TQUB_Sin_Model.Do_dMean_t_dIeq(var dmt :Double; t :Double) :Integer;
begin
  Result := 0;

  //mt = Ieq + Amp.sin(2.PI.Freq.t + Phase)
  //dmt = 1
  dmt := 1;

end;

function TQUB_Sin_Model.Do_dMean_t_dAmp(var dmt :Double; t :Double) :Integer;
begin
  Result := 0;

  //mt = Ieq + Amp.sin(2.PI.Freq.t + Phase)
  //dmt = sin(2.PI.Freq.t + Phase)
  dmt := sin(2*PI*Freq*t + Phase);

end;

function TQUB_Sin_Model.Do_dMean_t_dFreq(var dmt :Double; t :Double) :Integer;
begin
  Result := 0;

  //mt = Ieq + Amp.sin(2.PI.Freq.t + Phase)
  //dmt = Amp.2.PI.t.cos(2.PI.Freq.t + Phase)
  dmt := Amp*2*PI*t*cos(2*PI*Freq*t + Phase);

end;

function TQUB_Sin_Model.Do_dMean_t_dPhase(var dmt :Double; t :Double) :Integer;
begin
  Result := 0;

  //mt = Ieq + Amp.sin(2.PI.Freq.t + Phase)
  //dmt = Amp.cos(2.PI.Freq.t + Phase)
  dmt := Amp*cos(2*PI*Freq*t + Phase);

end;

function TQUB_Sin_Model.Do_dSSdfPar :Integer;
begin

  Result := inherited Do_dSSdfPar;

  //dSSdfPar
  if not FixedIeq   then dSSdfPar[IdxIeq]   := dSSdIeq;
  if not FixedAmp   then dSSdfPar[IdxAmp]   := dSSdAmp;
  if not FixedFreq  then dSSdfPar[IdxFreq]  := dSSdFreq*exp(fPars[IdxFreq]);//Freq=exp(fPar)
  if not FixedPhase then dSSdfPar[IdxPhase] := dSSdPhase;

end;

function TQUB_Sin_Model.Do_Varf2Varc :Integer;
begin

  Result := inherited Do_Varf2Varc;

  //var(Ieq|Amp|Freq|Phase) ~= Sqr(d{Ieq|Amp|Freq|Phase}/dz) * var(z)
  //dIeq  /dz =   1      * 1
  //dAmp  /dz =   1      * 1
  //dFreq /dz = dFreq/dc * 1 = Exp(c) * 1
  //dPhase/dz =   1      * 1

  //Ieq
  if not FixedAmp   then cParsVar[IdxIeq]   :=                             Abs(fParsVar[IdxIeq,   IdxIeq]);
  //Amp
  if not FixedAmp   then cParsVar[IdxAmp]   :=                             Abs(fParsVar[IdxAmp,   IdxAmp]);
  //Freq
  if not FixedFreq  then cParsVar[IdxFreq]  := Sqr(Exp(cPars[IdxFreq]))  * Abs(fParsVar[IdxFreq,  IdxFreq]);
  //Phase
  if not FixedPhase then cParsVar[IdxPhase] :=                             Abs(fParsVar[IdxPhase, IdxPhase]);
end;

function TQUB_Sin_Model.prGet_cParCount :Integer;
begin
  Result := 0;
  if not FixedIeq   then Inc(Result);
  if not FixedAmp   then Inc(Result);
  if not FixedFreq  then Inc(Result);
  if not FixedPhase then Inc(Result);
end;


//******************************************************************************
end.
