{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// specialized string grid; not sure what's special

unit QUB_Table;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Clipbrd, Grids,

QUB_Parser, QUB_Global;


type


TQUB_Table = class(TStringGrid)
protected


  prSaved                   :Boolean;
  prModified                :Boolean;
  prPrevValue               :String;
  prInEdit                  :Boolean;
  prSelecting               :Boolean;
  prOldCell                 :TPoint;
  prOldSel                  :TGridRect;


  procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  procedure KeyPress(var Key: Char); override;
  procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
  procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;

  function GetEditText(ACol, ARow: Longint): string; override;
  procedure SetEditText(ACol, ARow: Longint; const Value: string); override;

  procedure ColumnMoved(FromIndex, ToIndex: Longint); override;
  procedure RowMoved(FromIndex, ToIndex: Longint); override;


public


  constructor Create(AOwner: TComponent); override;


  function GetText(r1, c1, r2, c2 :Integer) :String;
  procedure SetText(Text :String; r1, c1, r2, c2 :Integer; CanExpand :Boolean);

  procedure CountRows;

  function Cut :Boolean;
  function Copy :Boolean;
  function Paste :Boolean;
  function Delete :Boolean;
  function DeleteAll :Boolean;
  function SelectAll :Boolean;
  procedure DeleteColumn(c :Integer); override;
  procedure DeleteRow(r :Integer); override;
  procedure InsertColumn(c :Integer);
  procedure InsertRow(r :Integer);


  property InEdit       :Boolean     read prInEdit;

  property Modified     :Boolean     read prModified  write prModified;

  
end;


procedure Register;                                                  


//******************************************************************************
implementation


//******************************************************************************
constructor TQUB_Table.Create(AOwner: TComponent);
begin
  inherited;
  Options := [{goFixedVertLine,}goFixedHorzLine,goVertLine,goHorzLine,goRangeSelect,goRowSizing,goColSizing{,goRowMoving,goColMoving},goThumbTracking]
end;

procedure TQUB_Table.KeyDown(var Key: Word; Shift: TShiftState);
const
  KEY_INSERT = 45;
  KEY_DELETE = 46;
begin
  inherited;
  if not prInEdit then begin
    if (Key = KEY_DELETE) and (ssShift in Shift) then begin
      Cut;
    end else if (Key = KEY_INSERT) and (ssShift in Shift) then begin
      Paste;
    end else if (Key = KEY_INSERT) and (ssCtrl in Shift) then begin
      Copy;
    end else if Key = KEY_DELETE then begin
      Delete;
    end else if ((Key = Ord('c')) or (Key = Ord('C'))) and (ssCtrl in Shift) then begin
      Copy;
    end;
  end;
end;

procedure TQUB_Table.KeyPress(var Key: Char);
var
  Message :TWMKey;
begin
  inherited;
  if (not prInEdit) {and (Key in [Char(32)..Char(255)])} then begin
    prInEdit         := True;
    Options          := Options + [goEditing];
    Message.Msg      := WM_CHAR;
    Message.CharCode := Word(Key);
    //Prevent double input
    Key              := Char(0);
    Message.KeyData  := 1;
    Dispatch(Message);
  end;
end;

procedure TQUB_Table.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Cell :TGridCoord;
  Sel  :TGridRect;
begin
  inherited;
  Options := Options - [goEditing];

  Cell := MouseCoord(x, y);

  prOldCell.x   := Cell.x;
  prOldCell.y   := Cell.x;

  prOldSel.Left   := Cell.x;
  prOldSel.Right  := Cell.x;
  prOldSel.Top    := Cell.y;
  prOldSel.Bottom := Cell.y;

  if (FixedRows > 0) and (Cell.y = 0) and (Cell.x <> 0) then begin
  //Select Col
    Sel.Top := 1;
    Sel.Bottom := RowCount - 1;
    Sel.Left := Cell.x;
    Sel.Right := Cell.x;
    Selection := Sel;
  end else if (FixedCols > 0) and (Cell.y <> 0) and (Cell.x = 0) then begin
  //Select Row
    Sel.Top := Cell.y;
    Sel.Bottom := Cell.y;
    Sel.Left := 1;
    Sel.Right := ColCount - 1;
    Selection := Sel;
  end else if (FixedRows > 0) and (FixedCols > 0) and (Cell.y = 0) and (Cell.x = 0) then begin
  //Select All
    Sel.Top := 1;
    Sel.Bottom := RowCount - 1;
    Sel.Left := 1;
    Sel.Right := ColCount - 1;
    Selection := Sel;
  end;
  prSelecting := True;
end;

procedure TQUB_Table.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Cell :TGridCoord;
  Sel  :TGridRect;
begin
  inherited;
  if not prSelecting then Exit;

  Cell := MouseCoord(x, y);
  if (Cell.x = prOldCell.x) and (Cell.y = prOldCell.y) then Exit;

  prOldCell.x := Cell.x;
  prOldCell.y := Cell.y;

  if (FixedRows > 0) and (Cell.y = 0) and (Cell.x <> 0) then begin
  //Change Col selection
    Sel := Selection;
    prOldSel.Right := Cell.x;
    if prOldSel.Right < prOldSel.Left then begin
      Sel.Left  := prOldSel.Right;
      Sel.Right := prOldSel.Left;
    end else begin
      Sel.Left  := prOldSel.Left;
      Sel.Right := prOldSel.Right;
    end;
    prOldSel.Right := Cell.x;
    Selection := Sel;
  end else if (FixedCols > 0) and (Cell.y <> 0) and (Cell.x = 0) then begin
  //Change Row selection
    Sel := Selection;
    prOldSel.Bottom := Cell.y;
    if prOldSel.Bottom < prOldSel.Top then begin
      Sel.Top    := prOldSel.Bottom;
      Sel.Bottom := prOldSel.Top;
    end else begin
      Sel.Top    := prOldSel.Top;
      Sel.Bottom := prOldSel.Bottom;
    end;
    prOldSel.Bottom := Cell.y;
    Selection := Sel;
  end;
end;

procedure TQUB_Table.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  prInEdit := False;
  prSelecting := False;
end;

function TQUB_Table.GetEditText(ACol, ARow: Longint): string;
begin
  Result := inherited GetEditText(ACol, ARow);
  prPrevValue := Result;
end;

procedure TQUB_Table.SetEditText(ACol, ARow: Longint; const Value: string);
begin
  inherited;
  if Value <> prPrevValue then begin
    prModified := True;
  end;
end;

procedure TQUB_Table.ColumnMoved(FromIndex, ToIndex: Longint);
begin
  inherited;
  prModified := True;
end;

procedure TQUB_Table.RowMoved(FromIndex, ToIndex: Longint);
begin
  inherited;
  prModified := True;
end;


//Edit methods *****************************************************************

//Cut
function TQUB_Table.Cut :Boolean;
begin
  Result := Copy;
  Delete;
end;

//Copy
function TQUB_Table.Copy :Boolean;
var
  Text :String;
begin
  Result := False;
  if (Selection.Top = Selection.Bottom) and (Selection.Left = Selection.Right) then begin
  //Current cell
    Text := Cells[Selection.Top, Selection.Left];
  end else begin
  //Selection
    Text := GetText(Selection.Top, Selection.Left, Selection.Bottom, Selection.Right);
  end;
  Clipboard.SetTextBuf(PChar(Text));
end;

//Paste
function TQUB_Table.Paste :Boolean;
var
  Text           :String;
  CanExpand      :Boolean;
  Data           :THandle;
  r1, r2, c1, c2 :Integer;
begin
  Result := False;

  //Get the text from clipboard
  Clipboard.Open;
  Data := GetClipboardData(CF_TEXT);
  try
    if Data <> 0 then begin
      Text := PChar(GlobalLock(Data));
    end else begin
      Text := '';
    end;
  finally
    if Data <> 0 then begin
      GlobalUnlock(Data);
      Clipboard.Close;
    end;
  end;

  //Any text ?
  if Length(Text) = 0 then Exit;

  //Find where to copy text
  c1 := Selection.Left;
  c2 := Selection.Right;
  r1 := Selection.Top;
  r2 := Selection.Bottom;
  CanExpand := False;
  if (c1 = c2) and (r1 = r2) then begin
    c2 := ColCount - 1;
    r2 := RowCount - 1;
    CanExpand := True;
  end;

  SetText(Text, r1, c1, r2, c2, CanExpand);

end;

//Delete
function TQUB_Table.Delete :Boolean;
var
  i, j :Integer;
begin
  Result := False;
  for i := Selection.Top to Selection.Bottom do begin
    for j := Selection.Left to Selection.Right do begin
      Cells[j, i] := '';
    end;
  end;
  prModified := True;
end;

//DeleteAll
function TQUB_Table.DeleteAll :Boolean;
var
  i, j :Integer;
begin
  Result := False;
  for i := 1 to RowCount - 1 do begin
    for j := 1 to ColCount - 1 do begin
      Cells[j, i] := '';
    end;
  end;
  prModified := True;
end;

//SelectAll
function TQUB_Table.SelectAll :Boolean;
var
  Sel :TGridRect;
begin
  Result := False;
  Sel.Left := 1;
  Sel.Top := 1;
  Sel.Right := RowCount - 1;
  Sel.Bottom := ColCount - 1;
  Selection := Sel;
end;

//Delete/Insert
procedure TQUB_Table.DeleteColumn(c :Integer);
begin
  inherited;
  prModified := True;
end;

procedure TQUB_Table.DeleteRow(r :Integer);
begin
  inherited;
  prModified := True;
end;

procedure TQUB_Table.InsertColumn(c :Integer);
var
  ro, co :Integer;
  tr, lc :Integer;
  i      :Integer;
begin
  ro := Row;
  co := Col;
  lc := LeftCol;
  tr := TopRow;
  ColCount := ColCount + 1;
  for i := ColCount - 1 downto c + 1 do begin
    Cols[i] := Cols[i - 1];
  end;
  for i := 0 to RowCount - 1 do begin
    Cells[c, i] := '';
  end;
  LeftCol := lc;
  TopRow := tr;
  Col := co;
  Row := ro;
  prModified := True;
end;

procedure TQUB_Table.InsertRow(r :Integer);
var
  ro, co :Integer;
  tr, lc :Integer;
  i      :Integer;
begin
  ro := Row;
  co := Col;
  lc := LeftCol;
  tr := TopRow;
  RowCount := RowCount + 1;
  for i := RowCount - 1 downto r + 1 do begin
    Rows[i] := Rows[i - 1]
  end;
  for i := 0 to ColCount - 1 do begin
    Cells[i, r] := '';
  end;
  LeftCol := lc;
  TopRow := tr;
  Col := co;
  Row := ro;
  CountRows;
  prModified := True;
end;


procedure TQUB_Table.CountRows;
var
  i :Integer;
begin
  for i := 1 to RowCount - 1 do begin
    Cells[0, i] := IntToStr(i);
  end;
end;


//Protected methods
function TQUB_Table.GetText(r1, c1, r2, c2 :Integer) :String;
const
  BUF_INCR        = ONE_MEGABYTE div 8;
var
  AsciiBufLen     :Integer;
  AsciiBufOffset  :Integer;

  Text            :String;
  Tabs            :String;
  Line            :String;
  i, j, l         :Integer;
begin
  // chris 11/5/2 wants to copy the headers as well:
  if (FixedRows = 1) and (FixedCols = 1) and (r1 = 1) and (c1 = 1) then begin
    r1 := 0;
    c1 := 0;
  end;

  Screen.Cursor := crHourglass;

  AsciiBufOffset     := 1;//0;Strings are 1-based

  Result             := '';

  try

    AsciiBufLen := 16 * (r2 - r1 + 1) * (c2 - c1 + 1);
    SetLength(Result, AsciiBufLen);

    Text := '';
    for i := r1 to r2 do begin
      Tabs := '';
      Line := '';
      for j := c1 to c2 do begin
        if Trim(Cells[j, i]) = '' then begin
          Tabs := Tabs + #9;
        end else begin
          Line := Line + Tabs + Cells[j, i];
          Tabs := #9;
        end;
      end;
      if Line <> '' then begin
        if i <> r2 then begin
          Line := Line + #13#10;
        end;
        l := Length(Line);
        if (AsciiBufOffset + l) > AsciiBufLen then begin
          AsciiBufLen  := AsciiBufLen + BUF_INCR;
          SetLength(Result, AsciiBufLen);
        end;
        Move(PChar(Line)^, (@Result[AsciiBufOffset])^, l);
        AsciiBufOffset := AsciiBufOffset + l;
      end;
    end;

    SetLength(Result, AsciiBufOffset);

  finally

    Screen.Cursor := crDefault;

  end;

end;

procedure TQUB_Table.SetText(Text :String; r1, c1, r2, c2 :Integer; CanExpand :Boolean);
var
  Parser         :TQUB_Parser;
  InLimits       :Boolean;
  CanExit        :Boolean;
  First          :Integer;
  OldColCount    :Integer;
  OldRowCount    :Integer;
  oc, ocr        :Integer;
  i, j           :Integer;
begin

  Parser := nil;

  try
    //Create the parser
    Parser := TQUB_Parser.Create;
    Parser.ReturnTab := True;
    Parser.ReturnNewLine := True;

    Parser.Start;
    i := r1;
    First := 1;
    CanExit := False;
    repeat
      if (i > r2) and (not CanExpand) then begin
        Break;
      end;
      j := c1;
      InLimits := True;
      repeat
        Parser.Next(Text);
        if Parser.Token.TokenType = TOKEN_NULL then begin
          if InLimits then begin
            if First <> Parser.Token.TokenPos then begin
              Cells[j, i] := System.Copy(Text, First, Parser.Token.TokenPos - First);
            end;
          end;
          CanExit := True;
          Break;
        end else if Parser.Token.TokenType = TOKEN_TAB then begin
          if InLimits then begin
            Cells[j, i] := System.Copy(Text, First, Parser.Token.TokenPos - First);
          end;
          First := Parser.Token.TokenPos + Parser.Token.TokenLen;
          Inc(j);
        end else if Parser.Token.TokenType = TOKEN_NEWLINE then begin
          if InLimits then begin
            Cells[j, i] := System.Copy(Text, First, Parser.Token.TokenPos - First);
          end;
          First := Parser.Token.TokenPos + Parser.Token.TokenLen;
          Inc(i);
          Break;
        end;

        //CheckPoint for rows
        if i > r2 then begin
          if CanExpand then begin
            if i + 1 > RowCount then begin
              OldRowCount := RowCount;
              RowCount := RowCount + 10;
              for ocr := OldRowCount + 1 to RowCount do begin
                for oc := 0 to ColCount - 1 do begin
                  Cells[oc, ocr] := '';
                end;
              end;
              CountRows;
            end;
          end else begin
            CanExit := True;
            Break;
          end;
        end;

        //CheckPoint for cols
        if not CanExit then begin
          if j > c2 then begin
            if CanExpand then begin
              if j + 1 > ColCount then begin
                OldColCount := ColCount;
                ColCount := ColCount + 10;
                for oc := OldColCount + 1 to ColCount do begin
                  for ocr := 0 to RowCount - 1 do begin
                    Cells[oc, ocr] := '';
                  end;
                end;
              end;
            end else begin
              InLimits := False;
            end;
          end;
        end;

      until(False);

    until (CanExit);

  finally

    Parser.Free;
    prModified := True;

  end;

end;


//******************************************************************************
procedure Register;
begin
  RegisterComponents('QUB', [TQUB_Table]);
end;

//******************************************************************************
end.
