{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// base class for delphi actions that run in their own thread


unit QUB_Task;


interface


uses


SysUtils, Classes, Dialogs,

QUB_Global,  
QUB_Console, QUB_Rich_Console,
QUB_Thread,
QUB_Thread_Socket;


type


TQUB_Task_Data_Source        = (DATA_SOURCE_WHOLEFILE, DATA_SOURCE_SELECTIONLIST,  DATA_SOURCE_SELECTION, DATA_SOURCE_FILELIST);
TQUB_Task_Data_Processed     = (DATA_PROCESSED_NONE,   DATA_PROCESSED_ASDISPLAYED, DATA_PROCESSED_ASSUCH);


const


DEF_DATA_SOURCE              = DATA_SOURCE_SELECTION;
DEF_DATA_PROCESSED           = DATA_PROCESSED_ASDISPLAYED;

DEF_DATA_SOURCEGLOBAL        = True; // maybe obsolete?


function DataProcessedToString(x :TQUB_Task_Data_Processed): String;
function StringToDataProcessed(x :String): TQUB_Task_Data_Processed;


type


//TQUB_Task_OnGetSocketEvent = procedure(Sender :TObject; var Socket :TQUB_Thread_Socket) of object;


TQUB_Task                  = class(TThread)
protected


  prName                   :String;

  prStatus                 :String;
  prPercent                :Integer;
  prETA                    :String;

  prStTime                 :Comp;

  prStopped                :Boolean;
  prStoppedCount           :Integer;
  prSkipped                :Boolean;

  prCanStop                :Boolean;
  prCanPause               :Boolean;
  prCanSkip                :Boolean;

  prError                  :Boolean;
  prErrorMsg               :String;

  prWikiTopic              :String;

  prWarning                :Boolean;
  prWarningMsg             :String;

  prSilentOutput           :Boolean;

  prOnChange               :TNotifyEvent;

  prScripting              :Boolean;

  //Console output
  prConsoleMsg             :String;
  prConsoleEntry           :Integer;
  prConsoleCode            :TQUB_Thread_Event_Message_Kind;

  prResultText             :PString;


  //Console
  procedure prWriteNewConsole;
  procedure prWriteSetConsole;
  procedure prWriteGrayConsole;

  procedure prDoOnChange;
  procedure prUpdateStatus(NewPercent :Integer; StartTime :Comp);


public


  constructor Create(Suspended :Boolean); virtual;
  destructor Destroy; override;


  procedure Stop; virtual;
  procedure Skip; virtual;

  procedure ShowMessage;

  procedure OnGetProperties(Sender :TObject); virtual; abstract;
  procedure OnPutProperties(Sender :TObject); virtual; abstract;


  property Error                   :Boolean         read prError;
  property ErrorMsg                :String          read prErrorMsg;
  property WikiTopic               :String          read prWikiTopic;

  property Terminated;

  property OnChange                :TNotifyEvent    read prOnChange write prOnChange;
  property Name                    :String          read prName;

  property Status                  :String          read prStatus;
  property Percent                 :Integer         read prPercent;
  property ETA                     :String          read prETA;

  property CanStop                 :Boolean         read prCanStop;
  property CanPause                :Boolean         read prCanPause;
  property CanSkip                 :Boolean         read prCanSkip;

  property SilentOutput      :Boolean                      read prSilentOutput      write prSilentOutput;

  property Scripting               :Boolean         read prScripting         write prScripting;
  property ResultText              :PString         read prResultText        write prResultText;


end;


implementation


//******************************************************************************

function DataProcessedToString(x :TQUB_Task_Data_Processed): String;
begin
  if      x = DATA_PROCESSED_NONE        then Result := 'None'
  else if x = DATA_PROCESSED_ASDISPLAYED then Result := 'AsDisplayed'
  else{if x = DATA_PROCESSED_ASSUCH      then}Result := 'AsSuch';
end;

function StringToDataProcessed(x :String): TQUB_Task_Data_Processed;
begin
  if      x = 'AsSuch'      then Result := DATA_PROCESSED_ASSUCH
  else if x = 'AsDisplayed' then Result := DATA_PROCESSED_ASDISPLAYED
  else{if x = 'None'        then}Result := DATA_PROCESSED_NONE;
end;


//******************************************************************************
constructor TQUB_Task.Create(Suspended :Boolean);
begin
  inherited;
  FreeOnTerminate := True;
  prCanStop       := True;
  prCanPause      := False;
  prCanSkip       := False;
end;

destructor TQUB_Task.Destroy;
begin
  try
    Terminate;
    Synchronize(prDoOnChange);
  finally
    inherited;
  end;
end;

procedure TQUB_Task.Stop;
begin
  prStopped := True;
  Inc(prStoppedCount);
end;

procedure TQUB_Task.Skip;
begin
  prSkipped := True;
end;

procedure TQUB_Task.prDoOnChange;
begin
  if Assigned(prOnChange) then begin
    prOnChange(Self);
  end;
end;

procedure TQUB_Task.prUpdateStatus(NewPercent :Integer; StartTime :Comp);
var
  DiffTime  :Comp;
  Hour      :Integer;
  Min       :Integer;
  Sec       :Integer;
//  MSec      :Word;
begin

  if NewPercent <> 0 then begin
    try
      DiffTime  := (TimeStampToMSecs(DateTimeToTimeStamp(Now)) - StartTime);
      // # msec since start
      DiffTime  := (DiffTime / NewPercent * 100) - DiffTime;
      // # msec est. remaining
      //DecodeTime(TimeStampToDateTime(MSecsToTimeStamp(DiffTime)), Hour, Min, Sec, MSec);
      // Exception class EConvertError with message "'0.4427953' is not a valid timestamp"
      // Workaround:
      Hour      := Round(CompToDouble(DiffTime)) div (1000*60*60);
      DiffTime  := DiffTime - Hour*1000*60*60;
      Min       := Round(CompToDouble(DiffTime)) div (1000*60);
      DiffTime  := DiffTime - Min*1000*60;
      Sec       := Round(CompToDouble(DiffTime)) div 1000;
      DiffTime  := DiffTime - Sec*1000;
      if DiffTime >= 500 then Sec := Sec + 1;

      prETA       := IntToStr(Hour) + ':';
      if Min < 10 then prETA := prETA + '0';
      prETA       := prETA + IntToStr(Min) + ':';
      if Sec < 10 then prETA := prETA + '0';
      prETA       := prETA + IntToStr(Sec);
    except
    end;
  end;

  prPercent   := NewPercent;
  Synchronize(prDoOnChange);
end;

procedure TQUB_Task.ShowMessage;
begin
  if prError then begin
    MessageHelpDlg(prErrorMsg, mtError, [mbOK], 0, prWikiTopic);
  end else if prWarning then begin
    MessageHelpDlg(prWarningMsg, mtWarning, [mbOK], 0, prWikiTopic);
  end else begin
    MessageHelpDlg(prStatus, mtInformation, [mbOK], 0, prWikiTopic);
  end;
  prWikiTopic := '';
end;


//Console
procedure TQUB_Task.prWriteNewConsole;
begin
  if prSilentOutput then Exit;
  case prConsoleCode of
    MESSAGE_MESSAGE: prConsoleEntry := Console.NewEntry(prConsoleMsg, COLOR_TEXT);
    MESSAGE_WARNING: prConsoleEntry := Console.NewEntry(prConsoleMsg, COLOR_WARNING);
    MESSAGE_ERROR:   prConsoleEntry := Console.NewEntry(prConsoleMsg, COLOR_ERROR);
  end;
end;

procedure TQUB_Task.prWriteSetConsole;
begin
  if prSilentOutput then Exit;
  case prConsoleCode of
    MESSAGE_MESSAGE: Console.SetEntry(prConsoleEntry, prConsoleMsg, COLOR_TEXT);
    MESSAGE_WARNING: Console.SetEntry(prConsoleEntry, prConsoleMsg, COLOR_WARNING);
    MESSAGE_ERROR:   Console.SetEntry(prConsoleEntry, prConsoleMsg, COLOR_ERROR);
  end;
end;

procedure TQUB_Task.prWriteGrayConsole;
begin
  if prSilentOutput then Exit;
  Console.SetEntry(prConsoleEntry, prConsoleMsg, COLOR_GRAY);
end;


//******************************************************************************
end.
