{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// View -> Tasks


unit QUB_TaskView;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ComCtrls, ExtCtrls, Grids, Buttons, Registry,
RegistryInterface,

QUB_Console, QUB_Rich_Console,
QUB_Task;


//{$include 'QUB_TaskView.hct'}


type


TQUB_TaskViewForm = class(TForm)


  panTaskView     :TPanel;
  grdTaskView     :TStringGrid;
  btnOK           :TBitBtn;
  btnTaskStop     :TSpeedButton;
  btnTaskSuspend  :TSpeedButton;
  btnTaskResume   :TSpeedButton;
  btnTaskSkip     :TSpeedButton;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure FormResize(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure FormShow(Sender: TObject);


  procedure btnOKClick(Sender: TObject);
  procedure btnTaskStopClick(Sender: TObject);
  procedure btnTaskSuspendResumeClick(Sender: TObject);
  procedure btnTaskSkipClick(Sender: TObject);
  procedure grdTaskViewSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);


protected


  prVisible            :Boolean;
  prCanResize          :Boolean;

  prJobCount           :Integer;
  prGlobalPct          :Integer;

  prOnTaskCountChange  :TNotifyEvent;
  prOnGlobalPctChange  :TNotifyEvent;


  procedure prUpdateGlobalPct;
  function prGetJobCount: Integer;

  function prWhichRowForTask(Task :TQUB_Task) :Integer;

  procedure prAddTask(Task :TQUB_Task; TaskIndex :Integer);
  procedure prDelTask(Task :Integer);

  procedure prSetTaskButtons;

  procedure prPropertiesRegistry(Read :Boolean);


public


  procedure StopAll;

  function HasTask :Boolean;
  function HasTaskOfType(TaskClass :TClass) :Boolean;

  procedure OnTaskChange(Sender :TObject);


  property GlobalPct           :Integer      read prGlobalPct;
  property JobCount            :Integer      read prGetJobCount;

  property OnTaskCountChange   :TNotifyEvent read prOnTaskCountChange  write prOnTaskCountChange;
  property OnGlobalPctChange   :TNotifyEvent read prOnGlobalPctChange  write prOnGlobalPctChange;


end;


implementation


{$R *.DFM}


const


COL_JOB           = 0;
COL_TASK          = 1;
COL_STATUS        = 2;
COL_PERCENT       = 3;
COL_ETA           = 4;


//******************************************************************************
procedure TQUB_TaskViewForm.FormCreate(Sender: TObject);
begin
  //grdTaskView.DefaultColWidth := (grdTaskView.ClientWidth - grdTaskView.ColCount) div grdTaskView.ColCount;
  grdTaskView.Cells[COL_JOB, 0]      := 'Job';
  grdTaskView.Cells[COL_TASK, 0]     := 'Task';
  grdTaskView.Cells[COL_STATUS, 0]   := 'Status';
  grdTaskView.Cells[COL_PERCENT, 0]  := 'Percent';
  grdTaskView.Cells[COL_ETA, 0]      := 'ETA';

  prGlobalPct                        := 100;

  prPropertiesRegistry(True{Read});

  if prVisible then Show;
  //{$include 'QUB_TaskView.htx'}

end;

procedure TQUB_TaskViewForm.FormDestroy(Sender: TObject);
begin
  prPropertiesRegistry(False{Read});
end;

procedure TQUB_TaskViewForm.FormShow(Sender: TObject);
begin
  prVisible := True;
  prPropertiesRegistry(False{Read});
  prCanResize := True;

  prSetTaskButtons;
end;

procedure TQUB_TaskViewForm.FormHide(Sender: TObject);
begin
  prPropertiesRegistry(False{Read});
  prCanResize := False;
end;

procedure TQUB_TaskViewForm.FormResize(Sender: TObject);
begin
  if prCanResize then begin
    prPropertiesRegistry(False{Read});
  end;

  try panTaskView.Width  := ClientWidth  - 2 * panTaskView.Left; except end;
  try panTaskView.Height := ClientHeight - 3 * panTaskView.Top   - btnOK.Height; except end;

  try btnOK.Top          := ClientHeight -     panTaskView.Top   - btnOK.Height; except end;
  try btnTaskStop.Top    := btnOK.Top; except end;
  try btnTaskSuspend.Top := btnOK.Top; except end;
  try btnTaskResume.Top  := btnOK.Top; except end;
  try btnTaskSkip.Top    := btnOK.Top; except end;

end;

procedure TQUB_TaskViewForm.grdTaskViewSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
begin
  prSetTaskButtons;
end;

//******************************************************************************
procedure TQUB_TaskViewForm.btnOKClick(Sender: TObject);
begin
  prVisible := False;
  Hide;
end;

procedure TQUB_TaskViewForm.btnTaskStopClick(Sender: TObject);
var
  TaskIdx :Integer;
  Task    :TQUB_Task;
begin
  TaskIdx := grdTaskView.Selection.Top;
  if (TaskIdx > 0) and (TaskIdx < grdTaskView.RowCount) then begin
    Task := TQUB_Task(grdTaskView.Objects[COL_TASK, TaskIdx]);
    if Task <> nil then begin
      if Task.CanStop then begin
        if (Task.CanPause) and (Task.Suspended) then begin
          Task.Resume;
          Sleep(100);
        end;
        Task.Stop;
      end;
    end;
  end;
end;

procedure TQUB_TaskViewForm.btnTaskSuspendResumeClick(Sender: TObject);
var
  TaskIdx :Integer;
  Task    :TQUB_Task;
begin
  TaskIdx := grdTaskView.Selection.Top;
  if (TaskIdx > 0) and (TaskIdx < grdTaskView.RowCount) then begin
    Task := TQUB_Task(grdTaskView.Objects[COL_TASK, TaskIdx]);
    if Task <> nil then begin
      if Task.CanPause then begin
        if Task.Suspended then begin
          Task.Resume;
        end else begin
          Task.Suspend;
        end;
        Sleep(100);
        OnTaskChange(Task);
      end;
    end;
  end;
end;

procedure TQUB_TaskViewForm.btnTaskSkipClick(Sender: TObject);
var
  TaskIdx :Integer;
  Task    :TQUB_Task;
begin
  TaskIdx := grdTaskView.Selection.Top;
  if (TaskIdx > 0) and (TaskIdx < grdTaskView.RowCount) then begin
    Task := TQUB_Task(grdTaskView.Objects[COL_TASK, TaskIdx]);
    if Task <> nil then begin
      if Task.CanSkip then begin
        if (Task.CanPause) and (Task.Suspended) then begin
          Task.Resume;
          Sleep(100);
        end;
        Task.Skip;
      end;
    end;
  end;
end;


//******************************************************************************
procedure TQUB_TaskViewForm.StopAll;
var
  TaskIdx :Integer;
  Task    :TQUB_Task;
begin
  for TaskIdx := 1 to grdTaskView.RowCount - 1 do begin
    Task := TQUB_Task(grdTaskView.Objects[COL_TASK, TaskIdx]);
    if Task <> nil then begin
      if Task.CanStop then begin
        Task.Stop;
      end;
    end;
  end;
end;

function TQUB_TaskViewForm.HasTask :Boolean;
begin
  Result := grdTaskView.Objects[COL_TASK, 1] <> nil;
end;

function TQUB_TaskViewForm.HasTaskOfType(TaskClass :TClass) :Boolean;
var
  i :Integer;
begin
  Result := False;
  for i := 1 to grdTaskView.RowCount - 1 do begin
    if grdTaskView.Objects[COL_TASK, i] <> nil then begin
      if grdTaskView.Objects[COL_TASK, i] is TaskClass then begin
        Result := True;
        Exit;
      end;
    end;
  end;
end;

procedure TQUB_TaskViewForm.prSetTaskButtons;
var
  TaskIdx :Integer;
  Task    :TQUB_Task;
begin
  try
    TaskIdx := grdTaskView.Selection.Top;
    Task    := nil;
    if (TaskIdx > 0) and (TaskIdx < grdTaskView.RowCount) then begin
      Task := TQUB_Task(grdTaskView.Objects[COL_TASK, TaskIdx]);
    end;
    btnTaskStop.Enabled    := (Task <> nil) and (Task.CanStop);
    btnTaskSuspend.Enabled := (Task <> nil) and (Task.CanPause);
    btnTaskResume.Enabled  := (Task <> nil) and (Task.CanPause);
    btnTaskSkip.Enabled    := (Task <> nil) and (Task.CanSkip);

    if Task = nil then begin
      btnTaskSuspend.Visible := True;
    end else if Task.Suspended then begin
      btnTaskSuspend.Visible := False;
    end else begin
      btnTaskSuspend.Visible := True;
    end;
    btnTaskResume.Visible  := not btnTaskSuspend.Visible;
  except
  end;
end;


//******************************************************************************
procedure TQUB_TaskViewForm.OnTaskChange(Sender :TObject);
var
  TaskIdx :Integer;
  Task    :TQUB_Task;
begin
  Task    := TQUB_Task(Sender);
  TaskIdx := prWhichRowForTask(Task);

  if Task.Terminated then begin
    prDelTask(TaskIdx);
    if not Task.SilentOutput then begin
      Console.NewEntry('Task ''' + Task.Name + ''' has terminated.', COLOR_GRAY);
    end;
  end else begin
    if Task.Suspended then begin
      grdTaskView.Cells[COL_STATUS, TaskIdx] := 'Suspended';
    end else begin
      grdTaskView.Cells[COL_STATUS, TaskIdx] := Task.Status;
    end;
    grdTaskView.Cells[COL_PERCENT, TaskIdx]  := IntToStr(Task.Percent);
    grdTaskView.Cells[COL_ETA, TaskIdx]      := Task.ETA;
  end;
  try grdTaskView.Refresh; except end;

  prUpdateGlobalPct;

  prSetTaskButtons;
end;

// now if eta was just a number of seconds this would be ez:
//   global pct = 100 * [1 - sum(ETA) / sum( ETA * 100 / (100 - pct) )]
// for now we just average pct over tasks
procedure TQUB_TaskViewForm.prUpdateGlobalPct;
var
  i              :Integer;
  pct            :Integer;
  njobs          :Integer;
begin
  njobs := JobCount;
  if njobs = 0 then begin
    pct := 100;
  end else begin
    pct   := 0;
    for i := 1 to njobs do begin
      pct := pct + StrToInt( grdTaskView.Cells[COL_PERCENT, i] );
    end;
    pct   := pct div njobs;
  end;

  prGlobalPct := pct;

  if Assigned(OnGlobalPctChange) then begin
    OnGlobalPctChange(Self);
  end;
end;

//******************************************************************************
procedure TQUB_TaskViewForm.prAddTask(Task :TQUB_Task; TaskIndex :Integer);
begin
  grdTaskView.Cells  [COL_JOB, TaskIndex]  := IntToStr(prJobCount);
  grdTaskView.Objects[COL_TASK, TaskIndex] := Task;
  grdTaskView.Cells  [COL_TASK, TaskIndex] := Task.Name;

  grdTaskView.Row := TaskIndex;
  grdTaskView.Col := COL_TASK;

  Inc(prJobCount);

  if Assigned(OnTaskCountChange) then begin
    OnTaskCountChange(Self);
  end

end;

procedure TQUB_TaskViewForm.prDelTask(Task :Integer);
var
  i, j :Integer;
begin
  if grdTaskView.RowCount > 2 then begin
    for i := Task + 1 to grdTaskView.RowCount - 1 do begin
      grdTaskView.Objects[COL_TASK, i - 1] := grdTaskView.Objects[COL_TASK, i];
      for j := 0 to grdTaskView.ColCount - 1 do begin
        grdTaskView.Cells  [j, i - 1] := grdTaskView.Cells[j, i];
      end;
    end;
    grdTaskView.RowCount := grdTaskView.RowCount - 1;
  end else begin
    grdTaskView.Objects[COL_TASK, 1] := nil;
    for j := 0 to grdTaskView.ColCount - 1 do begin
      grdTaskView.Cells[j, 1] := '';
    end;
  end;

  if Assigned(OnTaskCountChange) then begin
    OnTaskCountChange(Self);
  end

end;

function TQUB_TaskViewForm.prWhichRowForTask(Task :TQUB_Task) :Integer;
var
  i :Integer;
begin
  Result := -1;
  for i := 1 to grdTaskView.RowCount - 1 do begin
    if grdTaskView.Objects[COL_TASK, i] = Task then begin
      Result := i;
      Break;
    end;
  end;
  if Result = -1 then begin
    if grdTaskView.Objects[COL_TASK, 1] <> nil then begin
      grdTaskView.RowCount := grdTaskView.RowCount + 1;
      Result := grdTaskView.RowCount - 1;
    end else begin
      Result := 1;
    end;
    prAddTask(Task, Result);
  end;
end;

function TQUB_TaskViewForm.prGetJobCount: Integer;
begin
  if grdTaskView.Objects[COL_TASK, 1] = nil then begin
    Result := 0;
  end else begin
    Result := grdTaskView.RowCount - 1;
  end;
end;

//Properties
procedure TQUB_TaskViewForm.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
begin

  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Task View', True) then begin

        if Read then begin

          prVisible                          := ReadRegistryBoolean(Reg, 'Visible',False);
          Left                               := ReadRegistryInteger(Reg, 'Left', 132);
          Top                                := ReadRegistryInteger(Reg, 'Top', 132);
          Width                              := ReadRegistryInteger(Reg, 'Width', 768);
          Height                             := ReadRegistryInteger(Reg, 'Height', 537);
          grdTaskView.ColWidths[COL_JOB]     := ReadRegistryInteger(Reg, 'Col Job',30);
          grdTaskView.ColWidths[COL_TASK]    := ReadRegistryInteger(Reg, 'Col Task',210);
          grdTaskView.ColWidths[COL_STATUS]  := ReadRegistryInteger(Reg, 'Col Status',80);
          grdTaskView.ColWidths[COL_PERCENT] := ReadRegistryInteger(Reg, 'Col Percent',70);
          grdTaskView.ColWidths[COL_ETA]     := ReadRegistryInteger(Reg, 'Col ETA',80);

        end else begin

          WriteBool('Visible', prVisible);
          WriteInteger('Left', Left);
          WriteInteger('Top', Top);
          WriteInteger('Width', Width);
          WriteInteger('Height', Height);
          WriteInteger('Col Job', grdTaskView.ColWidths[COL_JOB]);
          WriteInteger('Col Task', grdTaskView.ColWidths[COL_TASK]);
          WriteInteger('Col Status', grdTaskView.ColWidths[COL_STATUS]);
          WriteInteger('Col Percent', grdTaskView.ColWidths[COL_PERCENT]);
          WriteInteger('Col ETA', grdTaskView.ColWidths[COL_ETA]);

        end;

      end;
      CloseKey;

    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
end.
