{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Task_AMP;


interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math, Controls, Dialogs,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QVO_QVM,

QUB_QFS_QMF,

QUB_Viewport_Data,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_AMP_Properties,

QUB_MIL_Global,
QUB_MIL,

Chart, Series,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, QUB_Viewport_Results,
QUB_GUIInterface;


const


AMP_REASONABLE_NPOINTS                              = 500000;



type


TQUB_Task_AMP  = class(TQUB_Task_Data)
protected

  prOptName                :String;

  //Source Objects
  prModel                  :TQUB_QVO_QVM;

  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  prConfig                 :INode;
  prDataDescription        :INode;
  prDatasets               :Array of INode;
  prResultNode             :INode;
  prReportNode             :INode;
  prRealityCheckSaysGo     :Boolean;

  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prRealityCheck; // synchronize; sets prRealityCheckSaysGo := true to proceed
  procedure prPublishResults;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetModel(Model :TQUB_QVO_QVM);

  function prReportCallback( arg :INode ): INode;
  procedure prDoReport;
  function prPercentCallback( arg :INode ): INode;

  procedure prReportResults;
  procedure prShowIdealization;


public

  FillDataCallback           :TNodeMethPtr;

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Stop; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model             :TQUB_QVO_QVM              read prModel             write prSetModel;


end;

function SimpleAmp(DataSet, Model :INode; OpensUp, Idealize :Boolean): INode;
// runs silently in current thread, returns usual result tree
// pass QUB_Node_Null to use default: chosen data, curr model, Task_Amp_Properties.PropertiesTree

function AMPFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; fillDataCB, rptCB, pctCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_amptreeiface@20';


implementation


uses
  IEEE754,
  QUB_Node_Model, QUB_Python_Data, QUB_QFS_QDF, QUB_Main,
  QUB_ActionSelector_Modeling;


function SimpleAmp(DataSet :INode; Model :INode; OpensUp, Idealize :Boolean): INode;
var
  DataSets        :INode;
  DataSetPtrs     :TPAPointer;
  Config          :INode;
  fillDataCB      :INodeCFunc;
  ResultNodePtr   :TQUB_QFS_PNodeMem;
  i               :Integer;
begin
  DataSetPtrs     := nil;
  Result          := QUB_Node_Null;

  try

    if DataSet.IsNull then begin
      DataSets      := DataNoder.GetChosenData(QUB_Node_Null);
      DataSet       := DataSets.Find('DataSet');
      i             := 0;
      while not DataSet.IsNull do begin
        Inc(i);
        DataSet     := DataSet.SiblingSameName;
      end;
      GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * (i + 1) );
      DataSetPtrs[ i ] := nil;

      DataSet       := DataSets.Find('DataSet');
      i             := 0;
      while not DataSet.IsNull do begin
        DataSet['ProcessData'].SetData('AsDisplayed');
        DataSetPtrs[i] := Pointer(DataSet.GetNodeMem);
        DataSet     := DataSet.SiblingSameName;
      end;
    end else begin
      DataSet['ProcessData'].SetData('AsDisplayed');
      GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * 2 );
      DataSetPtrs[1] := nil;
      DataSetPtrs[0] := DataSet.GetNodeMem;
    end;

    Config := TQUB_ActionSelector_Modeling(ActionsList.ByName['Modeling']).TaskAmpPropertiesForm.Properties.Clone(True);
    Config['ProcessData'].SetData('AsDisplayed');
    Config['Histograms']['BinCount'].SetData( 3 );
    Config['AutoInit'].SetData( 1 );
    Config['AutoInitUp'].SetData( Integer(OpensUp) );
    Config['Idealize'].SetData( Integer(Idealize) );
    Config['QuietOutput'].SetData( 1 );
    Config['ChannelIndex'].SetData( ViewportData.CurrData.ActiveChannel );

    if Model.IsNull then Model := GetModelAsNode;
    Model['ChannelCount'].SetData( 1 ); // since it doesn't work with >1 anyway
    Config.AppendChild( Model );

    fillDataCB    := TNodeCFunc.Wrap( TNodeMeth.Create( DataNoder.FillData ) );

    ResultNodePtr := AMPFromDll( Config.GetNodeMem, DataSetPtrs, fillDataCB, nil, nil );
    Result        := TNode.CreateUsing( ResultNodePtr );
    Node_DECREF( ResultNodePtr );

  except
  end;

  FreeMem(DataSetPtrs);

end;

//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _A M P                         //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_AMP.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'Amp estimator';

end;

destructor TQUB_Task_AMP.Destroy;
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_AMP.Stop;
begin

  prConfig['StopFlag'].Data[0] := 1;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_AMP.prDoTask;
var
  i               :Integer;

  DataSetPtrs     :TPAPointer;
  fillDataCB      :INodeCFunc;
  rptCB, pctCB    :INodeCFunc;
  ResultNode      :INode;
  ResultNodePtr   :TQUB_QFS_PNodeMem;
begin
  Synchronize( prRealityCheck );
  if not prRealityCheckSaysGo then Exit;

  GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * (Length(prDatasets) + 1) );
  for i := 0 to Length(prDataSets) - 1 do begin
    DataSetPtrs[i] := Pointer(prDataSets[i].GetNodeMem);
  end;
  DataSetPtrs[ Length(prDataSets) ] := nil;

  fillDataCB  := TNodeCFunc.Wrap( TNodeMeth.Create( FillDataCallback ) );
  rptCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prReportCallback ) );
  pctCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prPercentCallback ) );

  prConfig['StopFlag'].SetData( 0 );

  try
    prConsoleMsg  := 'Amp estimator running ... ';
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);

    ResultNodePtr := AMPFromDll( prConfig.GetNodeMem, DataSetPtrs, fillDataCB, rptCB, pctCB );
    ResultNode    := TNode.CreateUsing( ResultNodePtr );
    Node_DECREF( ResultNodePtr );

    prResultNode := ResultNode;
    Synchronize( prReportResults );
    if 0<>prConfig['Idealize'].Data[0] then prShowIdealization;
    Synchronize( prPublishResults );

    for i := 0 to prChunks.Count - 1 do begin
      if TQUB_Thread_PChunk(prChunks[i]).Status = CHUNK_STATUS_EMPTY then begin
        TQUB_Thread_PChunk(prChunks[i]).Status := CHUNK_STATUS_DONE;
      end;
    end;
  except
    prError    := True;
    prErrorMsg := 'Unknown error in AMP library';
    prWikiTopic := 'Error:Task';
  end;

  FreeMem( DataSetPtrs );

end;

procedure TQUB_Task_AMP.prRealityCheck;
var
  i, r         :Integer;
  npoint       :Integer;
  Seg          :INode;
begin
  npoint       := 0;
  for i := 0 to Length(prDataSets) - 1 do begin
    Seg        := prDataSets[i].Find('Segment');
    while not Seg.IsNull do begin
      npoint   := npoint + (Seg.Data[1] - Seg.Data[0] + 1);
      Seg      := Seg.SiblingSameName;
    end;
  end;

  if npoint <= AMP_REASONABLE_NPOINTS then begin
    prRealityCheckSaysGo := True;
    Exit;
  end;

  r := MessageHelpDlg( Format('AMP: %d data points could take too much time and memory.  Continue anyway?', [npoint]),
                   mtConfirmation, [mbYes, mbNo], 0, 'Warning:Too_Much_Time_And_Memory');

  prRealityCheckSaysGo := (r = mrYes);
end;

procedure TQUB_Task_AMP.prPublishResults;
var
  ds, seg         :INode;
begin
  prResultNode.Name  := 'AMP ' + ChangeFileExt(ExtractFileName(Data.FileName), '');

  prResultNode['ResultType'].SetData( 'AMP' );
  if (prDataDescription <> nil) and not prDataDescription.IsNull then prResultNode.AppendChild( prDataDescription );

  ds              := prResultNode.Find('DataSet');
  while not ds.isNull do begin
    seg           := ds.Find('Segment');
    while not seg.isNull do begin
      seg.RemoveChild( seg['Classes'] );
      seg.RemoveChild( seg['Durations'] );
      seg.RemoveChild( seg['Firsts'] );
      seg.RemoveChild( seg['Lasts'] );
      seg.RemoveChild( seg['States'] );
      seg.RemoveChild( seg['StateDurations'] );

      seg         := seg.SiblingSameName;
    end;
    ds            := ds.SiblingSameName;
  end;

  if prSelectionList <> nil then begin          // if the selection list is 'AMP Results',
    prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
    prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
  end;

  ViewportResults.AddResults(prResultNode, False);
end;

procedure TQUB_Task_AMP.prDoExecuteStart;
var
  AutoInitUp               :Integer;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;

    try
      if TQUB_QVO_QVD(Data).DrawInverted then begin
        AutoInitUp := Integer(prConfig['AutoInitUp'].Data[0]);
        if AutoInitUp <> 0 then AutoInitUp := 0 else AutoInitUp := 1;
        prConfig['AutoInitUp'].SetData( AutoInitUp );
      end;
    except
    end;
  end;

  inherited;

end;

procedure TQUB_Task_AMP.prDoExecuteEnd;
begin

  inherited;

  if prSelectionList <> nil then begin
    prSelectionList.UnLock;
  end;

  if Model <> nil then begin
    Model.UnlockAny;
  end;

end;


function TQUB_Task_AMP.prDoExecuteInitialize :Boolean;
var
  MdlNode       :INode;
begin
  Result        := False;

  MdlNode       := GetModelAsNode;
  MdlNode.SetData( 'Initial' );

  prConfig.AppendChild( MdlNode );
end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_AMP.prChunkData :Boolean;
var
  DataSets         :INode;
  DS               :INode;
  RemDS            :INode;
  Process          :INode;
begin
  Result           := False;

  Process          := prConfig.Find('ProcessData');

  DataSets         := DataNoder.GetChosenData(TNode.CreateNull);
  prDataDescription := DataSets.Find('DataDescription');

  DS               := DataSets.Find('DataSet');
  while not DS.IsNull do begin
    if DS.Find('Segment').IsNull then begin
      RemDS          := DS;
      DS             := DS.SiblingSameName;
      DataSets.RemoveChild( RemDS );
    end else begin
      SetLength( prDataSets, Length(prDataSets) + 1 );
      prDataSets[ Length(prDataSets) - 1 ] := DS;

      if not Process.IsNull then begin
        DS.AppendChild( Process.Clone(True) );
      end;

      DS             := DS.SiblingSameName;
    end;
  end;
end;


procedure TQUB_Task_AMP.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MilChunk :TQUB_Thread_Mil_PChunk;
begin

  MilChunk := TQUB_Thread_Mil_PChunk(Chunk);

  if MilChunk.ClassDwells <> nil then begin
    FreeMem(MilChunk.ClassDwells);
    MilChunk.ClassDwells := nil;
  end;

  FreeMem(MilChunk);

end;

function TQUB_Task_AMP.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Mil_Chunk));

  Result.Sel := Sel;

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_AMP.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_AMP_PropertiesForm do begin
    Self.ChannelIndex           := Properties['ChannelIndex'].Data[0];

    prConfig := Properties.Clone(True);

  end;//with
end;

procedure TQUB_Task_AMP.OnPutProperties(Sender :TObject);
begin
end;


procedure TQUB_Task_AMP.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then prModel.LockAny;
end;


//****************************************************************************//
//                                                                            //
//                          C A L L B A C K S                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Task_AMP.prReportCallback( arg :INode ): INode;
begin
  prReportNode := arg;
  Synchronize( prDoReport );
  Result := arg;
end;

procedure TQUB_Task_AMP.prDoReport;
begin
  // just print the node data's text rep'n for now
  prConsoleMsg := prReportNode.GetDataAsString;
  prConsoleCode := MESSAGE_ERROR;
  prWriteNewConsole;
end;

function TQUB_Task_AMP.prPercentCallback( arg :INode ): INode;
begin
  prUpdateStatus( Integer( arg.Data[0] ), prStTime );
  Result := arg;
end;


//****************************************************************************//
//                                                                            //
//                            R E S U L T S                                   //
//                                                                            //
//****************************************************************************//

procedure TQUB_Task_AMP.prReportResults;
var
  AmpNode, SDNode         :INode;
  i                       :Integer;
  Text                    :String;
begin
  AmpNode                 := prResultNode.Find('amp');
  SDNode                  := prResultNode.Find('sd');
  Text                    := Format('%8s%12s%12s', ['Class','Amp','Std']);

  if (Model <> nil) and not AmpNode.IsNull then begin
    with Model.ModelObject do begin
      for i := 0 to AmpNode.DataCount - 1 do begin
        if not (IsNan(AmpNode.Data[i]) or IsInfinity(AmpNode.Data[i])) then begin
          Amp[i] := AmpNode.Data[i];
        end;
        if not (IsNan(SDNode.Data[i])  or IsInfinity(SDNode.Data[i])) then begin
          Std[i] := SDNode.Data[i];
        end;

        if 0=prConfig['QuietOutput'].Data[0] then begin
          Text := Text + Format(#13#10'%8d%12.4g%12.4g', [i+1, Amp[i], Std[i]]);
        end;
      end;
    end;
  end;

  if 0=prConfig['QuietOutput'].Data[0] then begin
    prConsoleMsg := Text;
    prConsoleCode := MESSAGE_MESSAGE;
    prWriteNewConsole;
  end;
end;

procedure TQUB_Task_AMP.prShowIdealization;
var
  DataSet      :INode;
begin

  //Append Info
  prStatus := 'Applying Idealization';
  Synchronize(prDoOnChange);

  DataSet := prResultNode['DataSet'];
  TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].SetDwells(DataSet);
  TQUB_QVO_QVD(Data).DrawIdealized := True;

end;


//******************************************************************************
end.

