{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Task_AMP_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_QFS_Channels,
QUB_Task,
QUB_Task_Data,
QUB_Properties,
QUB_Task_Properties,
QUB_Task_Data_Properties,
QUB_Mil_Global,
QUB_Node,
QUB_DataList;


{$include 'QUB_Task_AMP_Properties.hct'}


const


DEF_AMPQUIETOUTPUT            = 1;

DEF_AMPHISTBINCOUNT           = 50;
MIN_AMPHISTBINCOUNT           = 2;
MAX_AMPHISTBINCOUNT           = 500;

DEF_AMPAUTOINIT               = 1;

DEF_AMPOPENSUP                = 1;
AMP_OPENS_UP                  = 0;
AMP_OPENS_DOWN                = 1;

DEF_VITERBI                   = 0;


type


TQUB_Task_AMP_PropertiesForm = class(TQUB_Task_Data_PropertiesForm)
  Label2                          :TLabel;
  Label4                          :TLabel;
  Label7                          :TLabel;
  panMethod                       :TPanel;
  chkViterbi                      :TCheckBox;
  chkQuiet                        :TCheckBox;
  chkAutoInit                     :TCheckBox;
  cmbOpens                        :TComboBox;
  txtMILHistBinCount              :TEdit;
  udMILHistBinCount               :TUpDown;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure cmbOpensChange(Sender: TObject);


protected


  prDataList                      :TQUB_DataList;


  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public


end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface, QUB_Main;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_AMP_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := 20;
  prDataList.Top        := 50;
  prDataList.Width      := ClientWidth - 40;
  prDataList.Height     := 100;
  prDataList.FileListAvail := False;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Histograms"]["BinCount"]',
                      'MIL Hist Bin Count', REGTYPE_INT, DEF_AMPHISTBINCOUNT,
                      udMILHistBinCount,
                      txtMILHistBinCount,
                      AcceptIntBetween(MIN_AMPHISTBINCOUNT, MAX_AMPHISTBINCOUNT)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Idealize"]',
                        'Viterbi', REGTYPE_BOOL, DEF_VITERBI,
                        chkViterbi));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["QuietOutput"]',
                        'QuietOutput', REGTYPE_BOOL, DEF_AMPQUIETOUTPUT,
                        chkQuiet));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["AutoInit"]',
                        'AutoInit', REGTYPE_BOOL, DEF_AMPAUTOINIT,
                        chkAutoInit));

  RegKeyName            := 'Software\QUB\Task AMP data';
  ProfileCat            := 'AMP';

  WikiTopic             := 'Modeling:Amps#Properties';

end;

procedure TQUB_Task_AMP_PropertiesForm.FormShow(Sender: TObject);
var
  Model                  :TQUB_QVO_QVM;
begin
  inherited;

  prDataList.ShowFiles;

  prDataList.ShowVar('ChannelCount', '', 1, False);

  Model                  := g_CurrModel();
  if Model <> nil then begin
    prDataList.ShowModelVars( Model, False );
  end;

end;

procedure TQUB_Task_AMP_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  prDataList.HideAll;
end;

procedure TQUB_Task_AMP_PropertiesForm.cmbOpensChange(Sender: TObject);
begin
  Properties['AutoInitUp'].Data[0] := (cmbOpens.Items.IndexOf( cmbOpens.Text )+1) mod 2;
end;


//******************************************************************************

procedure TQUB_Task_AMP_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Reg   :TRegistry;
  Prop  :INode;
  iVal  :Integer;
begin
  inherited;

  Reg   := nil; // will be created as needed by ReadXRegistry*
  try
    Prop  := Props['AutoInitUp'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, 'AutoInit opens up', iVal) then begin
        Prop.SetData(iVal);
      end else begin
        Prop.SetData(DEF_AMPOPENSUP);
      end;
    end;
  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_Task_AMP_PropertiesForm.prUpdateComponents;
begin
  inherited;
  cmbOpens.ItemIndex       := (Properties['AutoInitUp'].Data[0] + 1) mod 2;
end;


//******************************************************************************
end.
