{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// right-click the detail view in Viewport: Data Overlap and choose append; this task ensues


unit QUB_Task_AppendOverlap;


interface


uses


Windows, SyncObjs, Sysutils,

Alloc, AllocTypes,

QUB_Task, QUB_QVO, QUB_QVO_QVD, QUB_QVO_QVD_Global, QUB_Global,
PythonEngine;


type


TQUB_Task_AppendOverlap = class(TQUB_Task)
protected


  prData               :TQUB_QVO_QVD;
  prChannelIndex       :Integer;
  prSels               :array of TQUB_FirstLast;

  prActiveOnly         :Boolean; // set true in constructor to ignore inactive sels

  prBuffer             :TPAInteger;
  prBufferCount        :Integer;


  procedure prSetList(x :TQUB_QVO_QVD_SelectionList);

  procedure prAppendData; // to be called synchronized by subclass.Execute


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  property Data         :TQUB_QVO_QVD        read prData         write prData;
  property ChannelIndex :Integer             read prChannelIndex write prChannelIndex;
  property List         :TQUB_QVO_QVD_SelectionList              write prSetList;


end;


TQUB_Task_AppendOverlapAvg = class(TQUB_Task_AppendOverlap)
public
  constructor Create(Suspended :Boolean); override;

  procedure Execute; override;

end;


TQUB_Task_AppendOverlapSum = class(TQUB_Task_AppendOverlap)
public
  constructor Create(Suspended :Boolean); override;

  procedure Execute; override;

end;


TQUB_Task_AppendOverlapFunc = class(TQUB_Task_AppendOverlap)
protected

  prFunc            :PPyObject;

  procedure prSetFunc(x :PPyObject);

public
  destructor Destroy; override;

  procedure Execute; override;

  property Func     :PPyObject           read prFunc           write prSetFunc;
  
end;


implementation


uses

Math,
QUB_GUIInterface,
QUB_QFS_QDF_MPX,
QUB_Python_Thread, QUB_Viewport_Python;


constructor TQUB_Task_AppendOverlap.Create(Suspended :Boolean);
begin
  inherited;
  OnChange := TaskView.OnTaskChange;
  prName   := 'Append Function'
end;

destructor TQUB_Task_AppendOverlap.Destroy;
begin
  inherited;
end;

procedure TQUB_Task_AppendOverlap.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_AppendOverlap.OnPutProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_AppendOverlap.prSetList(x :TQUB_QVO_QVD_SelectionList);
var
  ListIter         :TQUB_QVO_QVD_SelectionListIter;
  i                :Integer;
begin
  SetLength(prSels, 0);
  if (Data = nil) or (x = nil) then Exit;

  ListIter := x.Find(0, Data.DataPoint.Last, prActiveOnly);
  i        := 0;
  if not ListIter.Done then repeat
    SetLength(prSels, i+1);
    prSels[i] := ListIter.Sel;
    Inc(i);
  until ListIter.Next;
  ListIter.Free;
end;

procedure TQUB_Task_AppendOverlap.prAppendData;
var
  BufferMPX   :Pointer;
  StartTime   :Double;
begin
  BufferMPX   := nil;
  try
    BufferMPX   := AllocMem(Data.DataObject.ADDataSize * Data.DataObject.ADChannelCount * prBufferCount);

    if Data.DataObject.Segments.SegmentCount > 0 then begin
      StartTime := Data.DataObject.Segments[ Data.DataObject.Segments.SegmentCount - 1 ].StartTime + Data.DataObject.Segments[ Data.DataObject.Segments.SegmentCount - 1 ].Channels.DataCount * Data.Sampling;
    end else begin
      StartTime := 0.0;
    end;

    MPXData(prBuffer, BufferMPX, Data.DataObject.ADChannelCount, ChannelIndex, Data.DataObject.ADDataSize, Data.DataObject.ADDataType, 0, prBufferCount-1, True{MPX});
    Data.AppendData(BufferMPX, Data.DataObject.ADChannelCount * Data.DataObject.ADDataSize, prBufferCount, True{NewSegment}, StartTime, True{CanRefresh - Doit}, False{Acq});
  finally
    FreeMem(BufferMPX);
  end;
end;


// *****************************************************************************

constructor TQUB_Task_AppendOverlapAvg.Create(Suspended :Boolean);
begin
  inherited;
  prActiveOnly := True;
end;

procedure TQUB_Task_AppendOverlapAvg.Execute;
var
  BufferCount     :Integer;
  Buffer          :TQUB_QVO_QVD_DataBufferOne;
  i, s, N         :Integer;
  OutBuffer       :TPADouble;
  Counts          :TPAInteger;
  Scale           :Double;
  Off             :Double;
begin
  if (Data = nil) or Data.LockedReadWrite then begin
    prError           := True;
    prErrorMsg        := 'No data or data in use';
    prWikiTopic       := 'Error:No_Data';
    Data              := nil;
  end;
  Data.LockReadWrite;

  Buffer          := nil;
  OutBuffer       := nil;
  Counts          := nil;

  prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  prStatus := 'Running';
  Synchronize(prDoOnChange);

  try
    if not prError then begin
    
      BufferCount   := 0;
      for s := 0 to Length(prSels) - 1 do begin
        BufferCount := Max(BufferCount, (prSels[s].Last - prSels[s].First + 1));
      end;
      prBufferCount := BufferCount;

      Buffer        := TQUB_QVO_QVD_DataBufferOne.Create(Data, ChannelIndex, BufferCount, True{Process}, False{use fit curve data}, False{add const baseline offset}, True{baseline nodes}, False{filter});
      OutBuffer     := DAlloc1(BufferCount);
      prBuffer      := TPAInteger(OutBuffer);
      Counts        := IAlloc1(BufferCount);

      Scale         := Data.Scaling * Data.DataChannelScaling[ChannelIndex];
      Off           := Data.DataChannelOffset[ChannelIndex];

      for i := 0 to BufferCount - 1 do begin
        OutBuffer[i] := 0.0;
        Counts[i]    := 0;
      end;

      for s := 0 to Length(prSels) - 1 do begin
        N              := (prSels[s].Last - prSels[s].First + 1);
        for i := 0 to N - 1 do begin
          OutBuffer[i] := OutBuffer[i] + Buffer.ValueAt(prSels[s].First + i);
          Counts[i]    := Counts[i] + 1;
        end;

        prUpdateStatus(Round(100.0 * (s + 1) / Length(prSels)), prStTime);
        if prStopped then Exit;
      end;

      for i := 0 to BufferCount - 1 do begin
        if Counts[i] > 0 then begin
          prBuffer[i]  := Round(Scale * (OutBuffer[i]-Off) / Counts[i]);
        end;
      end;

      prAppendData;

    end;

  finally

    prUpdateStatus( 100, prStTime );

    if not prError then begin

      if prStopped then begin
        prStatus := 'Stopping';
      end else begin
        prStatus := 'Terminating';
      end;
      Synchronize(prDoOnChange);
      
    end else if prError then begin

      prStatus := 'Error';
      Synchronize(prDoOnChange);

      Synchronize(ShowMessage);

    end;

    Data.UnlockReadWrite;
    Buffer.Free;
    FreeMem(OutBuffer);
    FreeMem(Counts);

  end;
end;

// *****************************************************************************

constructor TQUB_Task_AppendOverlapSum.Create(Suspended :Boolean);
begin
  inherited;
  prActiveOnly := True;
end;

procedure TQUB_Task_AppendOverlapSum.Execute;
var
  BufferCount     :Integer;
  Buffer          :TQUB_QVO_QVD_DataBufferOne;
  i, s, N         :Integer;
  OutBuffer       :TPADouble;
  Scale           :Double;
  Off             :Double;
begin
  if (Data = nil) or Data.LockedReadWrite then begin
    prError           := True;
    prErrorMsg        := 'No data or data in use';
    prWikiTopic       := 'Error:No_Data';
    Data              := nil;
  end;
  Data.LockReadWrite;

  Buffer          := nil;
  OutBuffer       := nil;

  prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  prStatus := 'Running';
  Synchronize(prDoOnChange);

  try
    if not prError then begin
    
      BufferCount   := 0;
      for s := 0 to Length(prSels) - 1 do begin
        BufferCount := Max(BufferCount, (prSels[s].Last - prSels[s].First + 1));
      end;
      prBufferCount := BufferCount;

      Buffer        := TQUB_QVO_QVD_DataBufferOne.Create(Data, ChannelIndex, BufferCount, True{Process}, False{use fit curve data}, False{add const baseline offset}, True{baseline nodes}, False{filter});
      OutBuffer     := DAlloc1(BufferCount);
      prBuffer      := TPAInteger(OutBuffer);

      Scale         := Data.Scaling * Data.DataChannelScaling[ChannelIndex];
      Off           := Data.DataChannelOffset[ChannelIndex];

      for i := 0 to BufferCount - 1 do begin
        OutBuffer[i] := 0.0;
      end;

      for s := 0 to Length(prSels) - 1 do begin
        N              := (prSels[s].Last - prSels[s].First + 1);
        for i := 0 to N - 1 do begin
          OutBuffer[i] := OutBuffer[i] + Buffer.ValueAt(prSels[s].First + i);
        end;

        prUpdateStatus(Round(100.0 * (s + 1) / Length(prSels)), prStTime);
        if prStopped then Exit;
      end;

      for i := 0 to BufferCount - 1 do begin
        prBuffer[i]  := Round(Scale * (OutBuffer[i] - Off));
      end;

      prAppendData;

    end;

  finally

    prUpdateStatus( 100, prStTime );

    if not prError then begin

      if prStopped then begin
        prStatus := 'Stopping';
      end else begin
        prStatus := 'Terminating';
      end;
      Synchronize(prDoOnChange);
      
    end else if prError then begin

      prStatus := 'Error';
      Synchronize(prDoOnChange);

      Synchronize(ShowMessage);

    end;

    Data.UnlockReadWrite;
    Buffer.Free;
    FreeMem(OutBuffer);

  end;
end;

// *****************************************************************************

destructor TQUB_Task_AppendOverlapFunc.Destroy;
begin
  try
    Func := nil;
  finally
    inherited;
  end;
end;

procedure TQUB_Task_AppendOverlapFunc.Execute;
var
  BufferCount     :Integer;
  Buffer          :TQUB_QVO_QVD_DataBufferOne;
  i, s, N         :Integer;
  OutBuffer       :TPAInteger;
  Scale           :Double;
  Off             :Double;
  x               :Single;
  Args            :PPyObject;
  PyResult        :PPyObject;
begin
  if (Data = nil) or Data.LockedReadWrite then begin
    prError           := True;
    prErrorMsg        := 'No data or data in use';
    prWikiTopic       := 'Error:No_Data';
    Data              := nil;
  end;
  Data.LockReadWrite;

  Buffer          := nil;
  OutBuffer       := nil;

  prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  prStatus := 'Running';
  Synchronize(prDoOnChange);

  with GetPythonEngine do begin
    try
      if not prError then begin

        N             := Length(prSels);

        BufferCount   := 0;
        for s := 0 to N - 1 do begin
          BufferCount := Max(BufferCount, (prSels[s].Last - prSels[s].First + 1));
        end;
        prBufferCount := BufferCount;

        Buffer        := TQUB_QVO_QVD_DataBufferOne.Create(Data, ChannelIndex, 10{1 sample at a time/hopping about, but the filter freaks out with a 1-sample buffer}, True{Process}, False{use fit curve average}, False{add const baseline offset}, True{baseline nodes}, False{filter});
        OutBuffer     := IAlloc1(BufferCount);
        prBuffer      := OutBuffer;
        Scale         := Data.Scaling * Data.DataChannelScaling[ChannelIndex];
        Off           := Data.DataChannelOffset[ChannelIndex];

        try

          PythonThread.BeginPython;

          Args      := PyTuple_New(N);

          for i := 0 to BufferCount - 1 do begin
            for s := 0 to N - 1 do begin
              if (prSels[s].First + i) < prSels[s].Last then begin
                x := Buffer.ValueAt(prSels[s].First + i);
              end else begin
                x := 0;
              end;
              PyTuple_SetItem(Args, s, PyFloat_FromDouble(x));
            end;

            PyResult       := PyObject_CallObject(prFunc, Args);
            x              := PyFloat_AsDouble(PyResult);
            Py_DECREF( PyResult );

            OutBuffer[i] := Round(Scale * (x - Off));

            if (i mod 100) = 0 then
              prUpdateStatus(Round(100.0 * (i + 1) / BufferCount), prStTime);
            if prStopped then Exit;
          end;

          Py_DECREF( Args );

        finally

          PythonThread.EndPython;

        end;

        prAppendData;

      end;

    finally

      prUpdateStatus( 100, prStTime );

      if not prError then begin

        if prStopped then begin
          prStatus := 'Stopping';
        end else begin
          prStatus := 'Terminating';
        end;
        Synchronize(prDoOnChange);
      
      end else if prError then begin

        prStatus := 'Error';
        Synchronize(prDoOnChange);

        Synchronize(ShowMessage);

      end;

      Data.UnlockReadWrite;
      Buffer.Free;
      FreeMem(OutBuffer);

    end;
  end;
end;

procedure TQUB_Task_AppendOverlapFunc.prSetFunc(x :PPyObject);
begin
  PythonThread.BeginPython;
  if prFunc <> nil then begin
    GetPythonEngine.Py_DECREF( prFunc );
  end;

  prFunc := x;

  if prFunc <> nil then begin
    GetPythonEngine.Py_INCREF( prFunc );
  end;
  PythonThread.EndPython;
end;



// *****************************************************************************
end.
