{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Task_ApplyBaseLine_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ComCtrls, Registry, ExtCtrls,

RegistryInterface,

QUB_QVO_QVD_Global,

QUB_Global,

QUB_Task,
QUB_Task_Data,
QUB_Properties,
QUB_Task_Properties,
QUB_Task_Data_Properties,
QUB_Node,
QUB_Baseline_Global;


{$include 'QUB_Task_ApplyBaseLine_Properties.hct'}


const


//Data Source
DEF_DATA_SOURCE               = DATA_SOURCE_SELECTION;


type


TQUB_Task_ApplyBaseline_PropertiesForm = class(TQUB_Task_Data_PropertiesForm)


  chkCorrectDCOffset                      :TCheckBox;

  btnReset                                :TButton;

  panMethod                               :TPanel;
  chkMethodLinear                         :TRadioButton;
  chkMethodAdaptive                       :TRadioButton;

  panMethodLinear                         :TPanel;
  Label2                                  :TLabel;
  Label3                                  :TLabel;
  Label4                                  :TLabel;
  Label5                                  :TLabel;
  Label6                                  :TLabel;
  Label8                                  :TLabel;
  Label9                                  :TLabel;
  Label10                                 :TLabel;
  Label11                                 :TLabel;
  Label12                                 :TLabel;
  txtRegressionSegment                    :TEdit;
  txtExtensionSegment                     :TEdit;
  txtMinBaseline                          :TEdit;
  txtMinSegment                           :TEdit;
  txtRegressionAdvance                    :TEdit;
  txtVarianceConfidence                   :TEdit;
  txtCoincidenceConfidence                :TEdit;
  txtOutliersTolerance                    :TEdit;
  txtMaxBaselineShift                     :TEdit;
  txtMaxSlope                             :TEdit;
  panMethodAdaptive                       :TPanel;
  Label7                                  :TLabel;
  Label13                                 :TLabel;
  Label14                                 :TLabel;
  txtWindowWidth                          :TEdit;
  txtWindowHeight                         :TEdit;
  txtMinDifference                        :TEdit;
  chkThirdOrderInterpolation              :TCheckBox;
  panDataSource                           :TPanel;
  Label16                                 :TLabel;
  chkDataSourceWholeFile                  :TRadioButton;
  chkDataSourceSelectionList              :TRadioButton;
  chkDataSourceSelection                  :TRadioButton;
  panMethodCapTrans                       :TPanel;
  Label19                                 :TLabel;
  Label20                                 :TLabel;
  Label21                                 :TLabel;
  Label22                                 :TLabel;
  txtStimChannel                          :TEdit;
  udStimChannel                           :TUpDown;
  txtDeltaStim                            :TEdit;
  txtFitDur                               :TEdit;
  txtCorrectDur                           :TEdit;
  chkMethodCapTrans                       :TRadioButton;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure chkMethodClick(Sender: TObject);

  procedure chkDataSourceClick(Sender: TObject);


protected


  //Data Channel
  prChannelIndex                          :Integer;
  prMaxChannel                            :Integer;

  //Data Source
  prDataSource                            :TQUB_Task_Data_Source;

  prParams                                :TQUB_Baseline_Params;


  //Data Source
  procedure prSetDataSource(s :TQUB_Task_Data_Source);

  //Parameters
  procedure prSetMethod(m :Integer);
  function prGetMethodString: String;
  procedure prSetMethodString(x :String);

  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public

  procedure PropertiesToParams;

  //Data Source
  property DataSource                     :TQUB_Task_Data_Source     read prDataSource                            write prSetDataSource;

  property Params                         :TQUB_Baseline_Params      read prParams;                                

  property Method                         :Integer                   read prParams.Method                         write prSetMethod;                         
  property MethodString                   :String                    read prGetMethodString                       write prSetMethodString;

  property WindowWidth                    :Integer                   read prParams.Adaptive.WindowWidth;           
  property WindowHeight                   :Double                    read prParams.Adaptive.WindowHeight;          
  property MinDifference                  :Integer                   read prParams.Adaptive.MinDifference;         

  property RegressionSegment              :Integer                   read prParams.Linear.RegressionSegment;       
  property ExtensionSegment               :Integer                   read prParams.Linear.ExtensionSegment;        
  property VarianceConfidence             :Double                    read prParams.Linear.VarianceConfidence;      
  property CoincidenceConfidence          :Double                    read prParams.Linear.CoincidenceConfidence;   
  property OutliersTolerance              :Double                    read prParams.Linear.OutliersTolerance;       
  property RegressionAdvance              :Integer                   read prParams.Linear.RegressionAdvance;       
  property MinBaseline                    :Integer                   read prParams.Linear.MinBaseline;             
  property MinSegment                     :Integer                   read prParams.Linear.MinSegment;              
  property MaxBaselineShift               :Double                    read prParams.Linear.MaxBaselineShift;        
  property MaxSlope                       :Double                    read prParams.Linear.MaxSlope;                
  property ThirdOrderInterpolation        :Boolean                   read prParams.Linear.ThirdOrderInterpolation; 

  property StimChannel                    :Integer                   read prParams.CapTrans.StimChannel;           
  property DeltaStim                      :Integer                   read prParams.CapTrans.DeltaStim;             
  property FitDur                         :Double                    read prParams.CapTrans.FitDur;                
  property CorrectDur                     :Double                    read prParams.CapTrans.CorrectDur;            

  property CorrectDCOffset                :Boolean                   read prParams.CorrectDCOffset;                


end;


var


QUB_Task_ApplyBaseline_PropertiesForm: TQUB_Task_ApplyBaseline_PropertiesForm;


implementation


{$R *.DFM}


//******************************************************************************
//Data Source
procedure TQUB_Task_ApplyBaseline_PropertiesForm.prSetDataSource(s :TQUB_Task_Data_Source);
begin
  prDataSource                       := s;
  chkDataSourceWholeFile.Checked     := DataSource = DATA_SOURCE_WHOLEFILE;
  chkDataSourceSelectionList.Checked := DataSource = DATA_SOURCE_SELECTIONLIST;
  chkDataSourceSelection.Checked     := DataSource = DATA_SOURCE_SELECTION;
end;

procedure TQUB_Task_ApplyBaseline_PropertiesForm.chkDataSourceClick(Sender: TObject);
begin
  inherited;
  DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;

//Method ***********************************************************************
procedure TQUB_Task_ApplyBaseline_PropertiesForm.prSetMethod(m :Integer);
begin

  Properties['Method'].SetData( MethodString );
  prParams.Method                := m;

  chkMethodAdaptive.Checked      := Method = METHOD_ADAPTIVE;
  chkMethodLinear.Checked        := Method = METHOD_LINEAR;
  chkMethodCapTrans.Checked      := Method = METHOD_CAPTRANS;

  panMethodAdaptive.Visible      := Method = METHOD_ADAPTIVE;
  panMethodLinear.Visible        := Method = METHOD_LINEAR;
  panMethodCapTrans.Visible      := Method = METHOD_CAPTRANS;

end;

function TQUB_Task_ApplyBaseline_PropertiesForm.prGetMethodString: String;
begin
  if Method = METHOD_ADAPTIVE then begin
    Result                       := 'Adaptive';
  end else if Method = METHOD_LINEAR then begin
    Result                       := 'Linear';
  end else begin // if Method = METHOD_CAPTRANS then begin
    Result                       := 'CapacityTransients';
  end;
end;

procedure TQUB_Task_ApplyBaseline_PropertiesForm.prSetMethodString(x :String);
begin
  if x = 'Adaptive' then begin
    Method                       := METHOD_ADAPTIVE;
  end else if x = 'Linear' then begin
    Method                       := METHOD_LINEAR;
  end else{if x = 'CapacityTransients' then}begin
    Method                       := METHOD_CAPTRANS;
  end;
end;

procedure TQUB_Task_ApplyBaseline_PropertiesForm.chkMethodClick(Sender: TObject);
begin
  inherited;
  Method := TRadioButton(Sender).Tag;
end;


//******************************************************************************
procedure TQUB_Task_ApplyBaseline_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["CorrectDCOffset"]',
                        'Correct DC Offset', REGTYPE_BOOL, DEF_CORRECTDCOFFSET,
                        chkCorrectDCOffset));

  //Adaptive Window
  AddEditProperty(TQUB_Property_Edit.Create('["Adaptive"]["WindowWidth"]',
                    'Window Width', REGTYPE_INT, DEF_WINDOW_WIDTH,
                    txtWindowWidth,
                    AcceptIntBetween(MIN_WINDOW_WIDTH, MAX_WINDOW_WIDTH),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Adaptive"]["WindowHeight"]',
                    'Window Height', REGTYPE_FLOAT, DEF_WINDOW_HEIGHT,
                    txtWindowHeight,
                    AcceptFloatBetween(MIN_WINDOW_HEIGHT, MAX_WINDOW_HEIGHT),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Adaptive"]["MinDifference"]',
                    'Min Difference', REGTYPE_INT, DEF_MIN_DIFFERENCE,
                    txtMinDifference,
                    AcceptIntBetween(MIN_MIN_DIFFERENCE, MAX_MIN_DIFFERENCE),
                   '') );

  //Linear Segments
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["RegressionSegment"]',
                    'Regression Segment', REGTYPE_INT, DEF_REGRESSIONSEGMENT,
                    txtRegressionSegment,
                    AcceptIntBetween(MIN_REGRESSIONSEGMENT, MAX_REGRESSIONSEGMENT),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["ExtensionSegment"]',
                    'Extension Segment', REGTYPE_INT, DEF_EXTENSIONSEGMENT,
                    txtExtensionSegment,
                    AcceptIntBetween(MIN_EXTENSIONSEGMENT, MAX_EXTENSIONSEGMENT),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["MinBaseline"]',
                    'Min Baseline', REGTYPE_INT, DEF_MINBASELINE,
                    txtMinBaseline,
                    AcceptIntBetween(MIN_MINBASELINE, MAX_MINBASELINE),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["MinSegment"]',
                    'Min Segment', REGTYPE_INT, DEF_MINSEGMENT,
                    txtMinSegment,
                    AcceptIntBetween(MIN_MINSEGMENT, MAX_MINSEGMENT),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["RegressionAdvance"]',
                    'Regression Advance', REGTYPE_INT, DEF_REGRESSIONADVANCE,
                    txtRegressionAdvance,
                    AcceptIntBetween(MIN_REGRESSIONADVANCE, MAX_REGRESSIONADVANCE),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["VarianceConfidence"]',
                    'Variance Confidence', REGTYPE_FLOAT, DEF_VARIANCECONFIDENCE,
                    txtVarianceConfidence,
                    AcceptFloatBetween(MIN_VARIANCECONFIDENCE, MAX_VARIANCECONFIDENCE),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["CoincidenceConfidence"]',
                    'Coincidence Confidence', REGTYPE_FLOAT, DEF_COINCIDENCECONFIDENCE,
                    txtCoincidenceConfidence,
                    AcceptFloatBetween(MIN_COINCIDENCECONFIDENCE, MAX_COINCIDENCECONFIDENCE),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["OutliersTolerance"]',
                    'Outliers Tolerance', REGTYPE_FLOAT, DEF_OUTLIERSTOLERANCE,
                    txtOutliersTolerance,
                    AcceptFloatBetween(MIN_OUTLIERSTOLERANCE, MAX_OUTLIERSTOLERANCE),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["MaxBaselineShift"]',
                    'Max Baseline Shift', REGTYPE_FLOAT, DEF_MAXBASELINESHIFT,
                    txtMaxBaselineShift,
                    AcceptFloatBetween(MIN_MAXBASELINESHIFT, MAX_MAXBASELINESHIFT),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Linear"]["MaxSlope"]',
                    'Max Slope', REGTYPE_FLOAT, DEF_MAXSLOPE,
                    txtMaxSlope,
                    AcceptFloatBetween(MIN_MAXSLOPE, MAX_MAXSLOPE),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Linear"]["ThirdOrderInterpolation"]',
                        'Third Order Interpolation', REGTYPE_BOOL, DEF_THIRDORDERINTERPOLATION,
                        chkThirdOrderInterpolation));

  RegKeyName            := 'Software\QUB\Task Apply Baseline';
  ProfileCat            := 'Base';
  prBeforeGetProperties := PropertiesToParams;

  WikiTopic             := 'Preprocessing:Baseline';

  txtRegressionSegment.Hint :=
  '   While the window is shorter than this size,'#13#10 +
  'after each extension is added, a regression line'#13#10 +
  'is calculated. Three acceptance criteria are used:'#13#10 +
  '  - Std;'#13#10 +
  '  - Slope;'#13#10 +
  '  - Outliers';

  txtExtensionSegment.Hint :=
  '   A segment this size is appended to the window.'#13#10 +
  'Two acceptance criteria are used for the increment:'#13#10 +
  '  - Avg;'#13#10 +
  '  - Outliers';

  txtMinBaseline.Hint :=
  '   A segment of this size is automatically considered'#13#10 +
  'to be baseline';

  txtMinSegment.Hint :=
  '   Segments shorter than this are automatically rejected';

  {$include 'QUB_Task_ApplyBaseLine_Properties.htx'}
end;

procedure TQUB_Task_ApplyBaseline_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  DataSource             := DataSource;
  Method                 := Method;
end;

procedure TQUB_Task_ApplyBaseline_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;


//******************************************************************************

procedure TQUB_Task_ApplyBaseline_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Reg   :TRegistry;
  Prop  :INode;
  iVal  :Integer;
begin
  inherited;

  Reg   := nil; // will be created as needed by ReadXRegistry*
  try
    Prop  := Props['Method'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, 'Method', iVal) then begin
        Method      := iVal;
      end else begin
        Method      := DEF_METHOD;
      end;
    end;
  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_Task_ApplyBaseline_PropertiesForm.prUpdateComponents;
begin
  inherited;
  MethodString       := Properties['Method'].DataAsString;
end;


procedure TQUB_Task_ApplyBaseline_PropertiesForm.PropertiesToParams;
begin
  prParams.CorrectDCOffset                := 0<>Properties['CorrectDCOffset'].Data[0];

  prParams.Adaptive.WindowWidth           := Properties['Adaptive']['WindowWidth'].Data[0];
  prParams.Adaptive.WindowHeight          := Properties['Adaptive']['WindowHeight'].Data[0];
  prParams.Adaptive.MinDifference         := Properties['Adaptive']['MinDifference'].Data[0];

  prParams.Linear.RegressionSegment       := Properties['Linear']['RegressionSegment'].Data[0];
  prParams.Linear.ExtensionSegment        := Properties['Linear']['ExtensionSegment'].Data[0];
  prParams.Linear.VarianceConfidence      := Properties['Linear']['VarianceConfidence'].Data[0];
  prParams.Linear.CoincidenceConfidence   := Properties['Linear']['CoincidenceConfidence'].Data[0];
  prParams.Linear.OutliersTolerance       := Properties['Linear']['OutliersTolerance'].Data[0];
  prParams.Linear.MaxBaselineShift        := Properties['Linear']['MaxBaselineShift'].Data[0];
  prParams.Linear.MaxSlope                := Properties['Linear']['MaxSlope'].Data[0];
  prParams.Linear.RegressionAdvance       := Properties['Linear']['RegressionAdvance'].Data[0];
  prParams.Linear.MinBaseline             := Properties['Linear']['MinBaseline'].Data[0];
  prParams.Linear.MinSegment              := Properties['Linear']['MinSegment'].Data[0];
  prParams.Linear.ThirdOrderInterpolation := 0<>Properties['Linear']['ThirdOrderInterpolation'].Data[0];
end;


//******************************************************************************
end.
