{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// aka Staircase


unit QUB_Task_Drift;


interface


uses


SysUtils, Classes,

Alloc, AllocTypes,
qubopt_iface,

QUB_Global,
QUB_Drift_Model, QUB_MIP_Global, QUB_Q_Global,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_Thread,
QUB_Task, QUB_Task_Data,
QUB_Idealization, QUB_Node,
QUB_QFS_QIF_Segments,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix, QUB_QFS_Channels,
QUB_QVO_QVM,
QUB_ExpCond_Parameters,
QUB_Task_Drift_Properties;

//KPL_Parser;


type


TQUB_Task_Drift            = class(TQUB_Task_Data)
protected


  //Source Objects
  prModel                  :TQUB_QVO_QVM;
  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  //Optimizer Parameters
  prOptParams              :TQUB_MIP_Opt_Params;

  //DriftModels
  prDriftModels            :TList;

  //ExpConds
  prExpCondParams          :TQUB_ExpCond_ParametersListList;

  //Previous LL
  prPLL                    :Double;
  
  //Best Grads
  prfParsGrads             :TPADouble;

  //OptSegments
  prOptSegsChunks          :TList;
  prOptSegsOutput          :String;
  prOptSegsOldMdl          :TQUB_QFS_QMF;

  prQuietOutput            :Boolean;

  prGroupList              :String;
  prBatchCount             :Integer;
  prBatchIdx               :Integer;

  prQuietEntry             :Integer;


  //Execute
  procedure prDoTask; override;

  procedure prDoExecute; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetModel(Model :TQUB_QVO_QVM);

  //Update Model
  procedure prUpdateModel;
  procedure prResetModel;
  procedure prShowData;

  function DataCallBack(MIPChunk :TQUB_MIP_PChunk) :Boolean;

  procedure prDrift;
  function prLLGrad(CompGrad :Integer) :Integer;

  procedure prPrintOptParams;
  procedure prPrintParams(Final :Boolean);
  procedure prPrintGrads;


public


  constructor Create(Suspended :Boolean); override;


  function DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
  function DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model             :TQUB_QVO_QVM                 read prModel             write prSetModel;

  //Parameters
  property OptParams         :TQUB_MIP_Opt_Params          read prOptParams         write prOptParams;

  property QuietOutput       :Boolean                      read prQuietOutput       write prQuietOutput;


end;


implementation


uses

Math,
QUB_Q_Model,
QUB_Main,
QUB_Node_Python, QUB_Python_Data,
QUB_GUIInterface, QUB_Viewport_Data;


//DFP Callbacks ****************************************************************
function DFPFunc(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
begin
  Result := TQUB_Task_Drift(call_obj).DFPFunc(Pars, Grad, LL, nf, ndf, CompGrad);
end;

function DFPCheck(call_obj :TObject; Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
begin
  Result := TQUB_Task_Drift(call_obj).DFPCheck(Iters, nf, ndf, LL, Pars, Grads);
end;



//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _ M I P                        //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_Drift.Create(Suspended :Boolean);
begin

  inherited;

  prName := 'Drift optimizer';

  prCanPause := True;
  prCanSkip  := True;

end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Drift.prDoTask;
begin

  inherited;

  prDrift;

end;

procedure TQUB_Task_Drift.prDrift;
var
  Done             :Boolean;
  Cnt              :Integer;
  m, n             :Integer;
  BatchFirst       :Integer;
  BatchIter        :Integer;
  BatchCountCount  :Integer;
  BatchCountList   :TPAInteger;
begin

  prOptSegsChunks := TList.Create;

  BatchCountCount := 1;
  BatchCountList  := nil;


  if prOptParams.OptSegments = OPTSEGS_GROUPLIST then begin
    BatchCountCount := 100;
    BatchCountList  := IAlloc1(BatchCountCount);
    StrToArrayOfInt(prGroupList, BatchCountList, BatchCountCount);
    if BatchCountCount = 0 then begin
      prOptParams.OptSegments := OPTSEGS_GROUP;
      BatchCountCount := 1;
    end;
  end;



  try


   for BatchIter := 0 to BatchCountCount - 1 do begin


      prSkipped := False;

      if prStopped then Break;

      //Save them all
      for m := 0 to prChunks.Count - 1 do begin
        prOptSegsChunks.Add(prChunks[m]);
      end;

      //Initialize pointers in list
      BatchFirst := 0;
      if          prOptParams.OptSegments = OPTSEGS_TOGETHER{ALL} then begin
        prBatchCount := prChunks.Count;
      end else if prOptParams.OptSegments = OPTSEGS_GROUP then begin
        prBatchCount := prOptParams.OptSegmentsGroupCount;
      end else begin
        prBatchCount := Abs(BatchCountList[BatchIter]);
      end;


      Synchronize(prPrintOptParams);


      prBatchIdx := 0;


      while (prBatchIdx < prOptParams.OptSegmentsMaxBatch) and (BatchFirst < prOptSegsChunks.Count) and (not prStopped) do begin
      //Batches


        prSkipped := False;

        Inc(prBatchIdx);


        if (prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST) then begin
          Synchronize(prResetModel);
        end;

        if (BatchFirst + prBatchCount - 1) >= prOptSegsChunks.Count then begin
          prBatchCount := prOptSegsChunks.Count - BatchFirst;
        end;


        //Next Batch
        prChunks.Clear;


        for m := BatchFirst to BatchFirst + prBatchCount - 1 do begin
          prChunks.Add(prOptSegsChunks[m]);
        end;


        //Multiplex files
        if prOptParams.OptMUXFiles = 1 then begin
          for m := BatchFirst + prBatchCount to prOptSegsChunks.Count - 1 do begin
            if TQUB_MIP_PChunk(prOptSegsChunks[BatchFirst]).QVD <> TQUB_MIP_PChunk(prOptSegsChunks[m]).QVD then begin
              for n := m + BatchFirst to m + BatchFirst + prBatchCount - 1 do begin
                if n >= prOptSegsChunks.Count then Break;
                prChunks.Add(prOptSegsChunks[n]);
              end;
              Break;
            end;
          end;
        end;


        Synchronize(prShowData);


        //**********************************************************************
        Cnt  := 0;

        repeat
        //restart if Grads > some threshold...


          if Cnt > 0 then begin
            prConsoleMsg  := 'Restarting...';
            prConsoleCode := MESSAGE_WARNING;
            if not prQuietOutput then Synchronize(prWriteNewConsole);
            prpLL := 0;
          end;

          Done := True;

          prfParsGrads := DAlloc1(TQUB_Drift_Model(prDriftModels[0]).fParCount);
          for m := 0 to TQUB_Drift_Model(prDriftModels[0]).fParCount - 1 do prfParsGrads[m] := 0;


          try


            for m := 0 to prDriftModels.Count - 1 do begin

              TQUB_Drift_Model(prDriftModels[m]).Do_fParsToModel;

            end;


            if prOptParams.OptMode = MODE_OPTIMIZE then begin
            //Optimize

              if prQuietOutput then begin
                prConsoleMsg  := Format('Batch index %d: starting ...', [prBatchIdx]);
                prConsoleCode := MESSAGE_WARNING;
                Synchronize(prWriteNewConsole);
                prQuietEntry := prConsoleEntry;
              end;

              try
                DFP_MIN(Self, TQUB_Drift_Model(prDriftModels[0]).fParCount, TQUB_Drift_Model(prDriftModels[0]).fPars, TQUB_Drift_Model(prDriftModels[0]).fParsVar,
                        QUB_Task_Drift.DFPFunc, QUB_Task_Drift.DFPCheck,
                        OptParams.OptMaxIter, OptParams.OptLLConv, OptParams.OptGradConv, OptParams.OptMaxStep,
                        TQUB_Drift_Model(prDriftModels[0]).Error);
              except
              end;

              TQUB_Drift_Model(prDriftModels[0]).Do_Varf2Varc;


            end else begin
            //Evaluate Log-Likelihood


              prLLGrad(1{CompGrad});

              TQUB_Drift_Model(prDriftModels[0]).Do_Varf2Varc;


              prPrintGrads;


            end;


            prPrintParams(True{Final});


          finally


            //Check Grads:
            for m := 0 to TQUB_Drift_Model(prDriftModels[0]).fParCount - 1 do begin
              if Abs(prfParsGrads[m]) >= 1 then begin
                Done := False;
                Break;
              end;
            end;

            FreeMem(prfParsGrads);


          end;


          Inc(Cnt);


          TQUB_Drift_Model(prDriftModels[0]).Iterations := 0;

          if (OptParams.OptRestart = 0) and (Cnt > 0) then Break;

        until ((Done) or (Cnt >= 2) or (prStopped) or (prError));
        //************************************************************************


        prConsoleMsg  := prOptSegsOutput;
        prConsoleCode := MESSAGE_MESSAGE;
        if prQuietOutput then begin
          prConsoleEntry := prQuietEntry;
          Synchronize(prWriteSetConsole);
        end else begin
          Synchronize(prWriteNewConsole);
        end;


        //Next batch
        Inc(BatchFirst, prBatchCount);


      end;


      if ((prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST)) and (prBatchCount > 1) then Synchronize(prResetModel);

      prChunks.Clear;

      for m := 0 to prOptSegsChunks.Count - 1 do begin
        prChunks.Add(prOptSegsChunks[m]);
      end;

      prOptSegsChunks.Clear;


    end;//BatchIter


  finally

    prOptSegsChunks.Free;
    prOptSegsOldMdl.Free;
    FreeMem(BatchCountList);

  end;

end;

procedure TQUB_Task_Drift.prPrintOptParams;
var
  t, h :String;
  k    :Integer;
begin

  with prOptParams do begin

    t :=     #13#10'Drift parameters for:';
    t := t + #13#10#9'data  '    + Choice(prData  <> nil, ExtractFileName(prData.FileName),  '<none>');
    t := t + #13#10#9'model '    + Choice(prModel <> nil, ExtractFileName(prModel.FileName), '<none>');
    t := t + #13#10#9'segcount ' + IntToStr(prBatchCount) + '.';
    t := t + #13#10;


    //Print column headers for quiet output:
    if QuietOutput then begin
      //TODO: finish it. LM
      h := '';
      with TQUB_Drift_Model(prDriftModels[0]) do begin
        for k := 0 to RateCount - 1 do begin
          h := h + Format('%.10s', [Format('k%d-%d', [RateIdx[k].x + 1, RateIdx[k].y + 1]) + '          ']) + #9 +
                   Format('%.10s', ['std' + '          ']) + #9;
        end;

        h := h + Format('%.14s', ['LL'    + '              ']) + #9;

      end;//with

      h := #13#10 + h;

    end;

  end;

  prConsoleMsg  := t + h;
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
end;  

procedure TQUB_Task_Drift.prPrintParams(Final :Boolean);
var
  LogL :Double;
  t    :String;
  k, m :Integer;
begin

  LogL := 0;
  for m := 0 to prDriftModels.Count - 1 do begin
    LogL := LogL + TQUB_Drift_Model(prDriftModels[m]).LL;
  end;

  prOptSegsOutput :=     '';

  if Final then begin
    try
      t :=            #13#10'================================================================';
      t := t + Format(#13#10'Drift final results:  LL = %-14.10g Iterations = %d,'#13#10 +
                            'Dwells = %d, Points = %d, Multps. = %d, TrunCount = %d',
                     [-LogL, TQUB_Drift_Model(prDriftModels[0]).Iterations,
                      TQUB_Drift_Model(prDriftModels[0]).TotalDwellCount, TQUB_Drift_Model(prDriftModels[0]).TotalDataPoints, TQUB_Drift_Model(prDriftModels[0]).Multiplications,
                      TQUB_Drift_Model(prDriftModels[0]).TruncatedJumpCount]);
    except
    end;
  end;

  t := t +   #13#10#9'Rates:  '#13#10#9's1  s2   k0         k0.std     k1         k1.std' +
                             #13#10#9'--------------------------------------------------';

  for k := 0 to TQUB_Drift_Model(prDriftModels[0]).RateCount - 1 do begin
    try
      t := t + Format(#13#10#9'%d   %d    %-10.6g %-10.6g %-10.6g %-10.6g',
                     [TQUB_Drift_Model(prDriftModels[0]).RateIdx[k].x+1, TQUB_Drift_Model(prDriftModels[0]).RateIdx[k].y+1,
                     //!!! cParsVar
                      Choice(OptParams.OptAllowNegRates,
                            TQUB_Drift_Model(prDriftModels[0]).cPars[2*k + 0],
                        Exp(TQUB_Drift_Model(prDriftModels[0]).cPars[2*k + 0])
                      ),
                                                                          Sqrt(TQUB_Drift_Model(prDriftModels[0]).cParsVar[2*k + 0]),
                      TQUB_Drift_Model(prDriftModels[0]).cPars[2*k + 1],  Sqrt(TQUB_Drift_Model(prDriftModels[0]).cParsVar[2*k + 1])
                   ]);
      prOptSegsOutput := prOptSegsOutput + Trim(Format('%10.6g', [Choice(OptParams.OptAllowNegRates,
                                                                        TQUB_Drift_Model(prDriftModels[0]).cPars[2*k + 0],
                                                                    Exp(TQUB_Drift_Model(prDriftModels[0]).cPars[2*k + 0])
                                                                  )
                                                                                                                            ])) + #9 +
                                           Trim(Format('%10.6g', [Sqrt(TQUB_Drift_Model(prDriftModels[0]).cParsVar[2*k + 0])])) + #9;
    except
    end;
  end;
  prOptSegsOutput := prOptSegsOutput + Trim(Format('%14.10g', [-LogL]));


  t := t + #13#10#9'--------------------------------------------------';

  //prBatchOutput := #13#10 + prBatchOutput;

  if Final then begin
    if OptParams.OptMode = MODE_OPTIMIZE then begin
      t := t + #13#10'Free pars grads: ';
      for k := 0 to TQUB_Drift_Model(prDriftModels[0]).fParCount - 1 do begin
        try
          t := t + Format('%-12.8g ', [-prfParsGrads[k]]);
        except
        end;
      end;
    end;
    t := t + #13#10 + '================================================================';
  end else begin
    t := t + #13#10;
  end;

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);
end;

procedure TQUB_Task_Drift.prPrintGrads;
var
  LogL  :Double;
  dLogL :Double;
  t     :String;
  k, m  :Integer;
begin

  LogL := 0;
  for m := 0 to prDriftModels.Count - 1 do begin
    LogL := LogL + TQUB_Drift_Model(prDriftModels[m]).LL;
  end;

  prConsoleMsg  := Format('Batch %d-%d, LL: %-14.10g, Gr: ', [prBatchIdx, TQUB_Drift_Model(prDriftModels[0]).Iterations, -LogL]);

  t := Format('Iteration %d, LL = %-14.10g, Func calls = %d, Grad calls = %d',
              [TQUB_Drift_Model(prDriftModels[0]).Iterations, -LogL, TQUB_Drift_Model(prDriftModels[0]).FuncCalls, TQUB_Drift_Model(prDriftModels[0]).GradCalls]);

  t := t + #13#10#9'Free pars grads: ';
  for k := 0 to TQUB_Drift_Model(prDriftModels[0]).fParCount - 1 do begin
    dLogL := 0;
    for m := 0 to prDriftModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Drift_Model(prDriftModels[m]).dLLdfPar[k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
      prConsoleMsg  := prConsoleMsg + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  //t must be assigned here
  if prQuietOutput then begin
    prConsoleCode := MESSAGE_WARNING;
    Synchronize(prWriteSetConsole);
  end;

  t := t + #13#10#9'Cstr pars grads: ';
  for k := 0 to TQUB_Drift_Model(prDriftModels[0]).cParCount - 1 do begin
    dLogL := 0;
    for m := 0 to prDriftModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Drift_Model(prDriftModels[m]).dLLdcPar[k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  t := t + #13#10#9'Rates grads:     ';
  for k := 0 to TQUB_Drift_Model(prDriftModels[0]).RateCount - 1 do begin
    dLogL := 0;
    for m := 0 to prDriftModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Drift_Model(prDriftModels[m]).dLLdq[0,k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;  

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);

end;


//******************************************************************************
function TQUB_Task_Drift.DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
var
  i, m :Integer;
begin

  Result := 0;

  if TQUB_Drift_Model(prDriftModels[0]).Iterations > 0 then begin

    for m := 0 to prDriftModels.Count - 1 do begin

      //Copy the new pars into the Drift models
      for i := 0 to TQUB_Drift_Model(prDriftModels[m]).fParCount - 1 do begin
        TQUB_Drift_Model(prDriftModels[m]).fPars[i] := Pars[i];
      end;

      TQUB_Drift_Model(prDriftModels[m]).Do_fParsToModel;

    end;

  end;


  try
    prLLGrad(CompGrad);
  except
  end;

  
  LL := 0;
  for m := 0 to prDriftModels.Count - 1 do begin
    LL := LL + TQUB_Drift_Model(prDriftModels[m]).LL;
  end;


  if CompGrad <> 0 then begin
    for i := 0 to TQUB_Drift_Model(prDriftModels[0]).fParCount - 1 do begin
      Grad[i] := 0;
      for m := 0 to prDriftModels.Count - 1 do begin
        Grad[i] := Grad[i] + TQUB_Drift_Model(prDriftModels[m]).dLLdfPar[i];
      end;
    end;
    Move(Grad[0], prfParsGrads[0], TQUB_Drift_Model(prDriftModels[0]).fParCount * SizeOf(Double));
  end;


  Inc(TQUB_Drift_Model(prDriftModels[0]).Iterations);
  Inc(nf);
  if CompGrad <> 0 then Inc(ndf);
  TQUB_Drift_Model(prDriftModels[0]).FuncCalls := nf;
  TQUB_Drift_Model(prDriftModels[0]).GradCalls := ndf;

end;

function TQUB_Task_Drift.DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
var
  LLConverged :Boolean;
  GRConverged :Boolean;
  k, m        :Integer;
begin

  Result := 0;

  TQUB_Drift_Model(prDriftModels[0]).Iterations := Iters;

  if Iters >= OptParams.OptMaxIter then Result := 1;

  LLConverged := Abs(LL - prpLL) < OptParams.OptLLConv;
  prpLL := LL;

  GRConverged := True;
  for k := 0 to TQUB_Drift_Model(prDriftModels[0]).fParCount - 1 do begin
    if Abs(Grads[k]) > OptParams.OptGradConv then GRConverged := False;
  end;


  Synchronize(prUpdateModel);


  prPrintGrads;
  prPrintParams(False{Final});


  if (LLConverged) and (GRConverged) then Result := 1;
  if prStopped then Result := 1;
  if prSkipped then Result := 1;

  //if Result = 1 then begin
    //Change the parameters to the last good ones (???)
    for m := 0 to prDriftModels.Count - 1 do begin
      //Copy the new pars into the models
      for k := 0 to TQUB_Drift_Model(prDriftModels[m]).fParCount - 1 do begin
        TQUB_Drift_Model(prDriftModels[m]).fPars[k] := Pars[k];
      end;
      TQUB_Drift_Model(prDriftModels[m]).Do_fParsToModel;
    end;
    //Only the fParsGrads
    Move(Grads[0], prfParsGrads[0], TQUB_Drift_Model(prDriftModels[0]).fParCount * SizeOf(Double));
  //end;

end;



//****************************************************************************//
//                                                                            //
//                             L L   +   G R A D                              //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Drift.prLLGrad(CompGrad :Integer) :Integer;
const
  VERY_SMALL     = 1e-500;
var
  MIPChunk       :TQUB_MIP_PChunk;
  DriftModel     :TQUB_Drift_Model;
  OldModel       :TQUB_Drift_Model;
  Data           :TQUB_MIP_Data;
  DataCount      :Integer;

  AlphaS         :TPMExtended;
  BetaS          :TPMExtended;
  LScale         :TPAExtended;

  ThisP0         :TPADouble;
  ThisdP0dq      :TPMDouble;

  cLL, cGR       :Extended;
  c, k, p, s, m  :Integer;
  dt             :Double;

  ZeroDetected   :Boolean;
  SomeConfluent  :Boolean;


function ChunkData(c :Integer) :Integer;
begin
  Result   := 0;
  MIPChunk := prChunks[c];
  if MIPChunk.Data = nil then begin
    Result := Integer(DataCallback(MIPChunk));
  end;
end;

begin



  Result         := 0;

  ZeroDetected   := False;
  SomeConfluent := False;
  for m := 0 to prDriftModels.Count - 1 do begin
    if TQUB_Drift_Model(prDriftModels[m]).Confluent[0] <> 0 then SomeConfluent := True;
  end;



  //Reset LL and derivatives w.r.t. q's ****************************************
  for m := 0 to prDriftModels.Count - 1 do begin
    TQUB_Drift_Model(prDriftModels[m]).LL         := 0;
    for k := 0 to TQUB_Drift_Model(prDriftModels[m]).RateCount - 1 do begin
      TQUB_Drift_Model(prDriftModels[m]).dLLdq[0,k] := 0;
    end;
  end;
  //****************************************************************************



  //Load Data, if not already **************************************************
  for c := 0 to prChunks.Count - 1 do begin
    Result := ChunkData(c);
    if Result <> 0 then Exit;;
  end;


  OldModel  := nil;
  ThisP0    := nil;
  ThisdP0dq := nil;


  //For each chunk *************************************************************
  for c := 0 to prChunks.Count - 1 do begin


    MIPChunk  := prChunks[c];
    Data      := MIPChunk.Data;
    DataCount := MIPChunk.DataCount;
    AlphaS    := MIPChunk.AlphaS;
    BetaS     := MIPChunk.BetaS;
    LScale    := MIPChunk.LScale;
    dt        := MIPChunk.Sampling;


    //Choose the right DriftModel
    DriftModel := TQUB_Drift_Model(MIPChunk.MIPModel);


    if DriftModel <> OldModel then begin

      //Calculate A matrix and its derivatives w.r.t. each qq, and the products and their derivatives
      DriftModel.Do_A(dt);
      DriftModel.Do_dA_dq(dt, 0{ECClass});

      //Starting state
      DriftModel.Do_P0;
      DriftModel.Do_dP0_dq;
      ThisP0    := DriftModel.sP0;
      ThisdP0dq := DriftModel.dsP0dq;

    end;

    OldModel := DriftModel;


    //Initialize AlphaS, LScale ................................................
    p := 0;

    MulVecMatVecE(ThisP0, DriftModel.B[Data[p].Cls], AlphaS[p], DriftModel.UsedStateCount);
    LScale[p] := 0;
    for s := 0 to DriftModel.UsedStateCount - 1 do begin
      LScale[p]    := LScale[p] + AlphaS[p, s];
    end;
    if LScale[p] < VERY_SMALL then begin
      LScale[0] := VERY_SMALL;
      ZeroDetected := True;
    end;
    for s := 0 to DriftModel.UsedStateCount - 1 do begin
      AlphaS[p, s] := AlphaS[p, s] / LScale[p];
    end;


    //For each DataPoint calculate AlphaS, LScale ..............................
    for p := 1 to DataCount - 1 do begin

      MulVecEMatVecE(AlphaS[p - 1], DriftModel.ABSM[Data[p].Cls, Data[p].Pow], AlphaS[p], DriftModel.UsedStateCount);

      LScale[p]      := 0;
      for s := 0 to DriftModel.UsedStateCount - 1 do begin
        LScale[p]    := LScale[p] + AlphaS[p, s];
      end;
      if LScale[p] < VERY_SMALL then begin
        LScale[p]    := VERY_SMALL;
        ZeroDetected := True;
      end;
      for s := 0 to DriftModel.UsedStateCount - 1 do begin
        AlphaS[p, s] := AlphaS[p, s] / LScale[p];
      end;

    end;


    //Initialize BetaS .........................................................
    p := DataCount - 1;
    for s := 0 to DriftModel.UsedStateCount - 1 do begin
      BetaS[p, s] := 1 / LScale[p];
    end;


    //For each DataPoint calculate BetaS .......................................
    for p := DataCount - 2 downto 0 do begin

      MulMatVecEVecE(DriftModel.ABSM[Data[p + 1].Cls, Data[p + 1].Pow], BetaS[p + 1], BetaS[p], DriftModel.UsedStateCount);

      for s := 0 to DriftModel.UsedStateCount - 1 do begin
        BetaS[p, s] := BetaS[p, s] / LScale[p];
      end;

    end;



    //Log-Likelihood ...........................................................

    //For each DataPoint
    for p := 0 to DataCount - 1 do begin
      cLL := Ln(LScale[p]);
      DriftModel.LL := DriftModel.LL - cLL;//Use minus sign in order to minimize LL
    end;
    //..........................................................................



    //Rates Gradients ..........................................................

    //P0
    for k := 0 to DriftModel.RateCount - 1 do begin

      //dP0dq.B[0].BetaS[0]

      MulVecMatVecE(ThisdP0dq[k], DriftModel.B[Data[0].Cls], DriftModel.Vec1E, DriftModel.UsedStateCount);

      MulVecEVecESE(DriftModel.Vec1E, BetaS[0], cGR, DriftModel.UsedStateCount);

      DriftModel.dLLdq[0,k] := DriftModel.dLLdq[0,k] - cGR;//Use minus sign in order to minimize LL

    end;//of k

    //For each DataPoint
    for p := 0 to DataCount - 2 do begin

      for k := 0 to DriftModel.RateCount - 1 do begin

        //cGR = cGR + Sum t [AlphaS[t].dAtdq[t+1].Cls[t+1].BetaS[t+1]]

        MulVecEMatVecE(AlphaS[p], TPMPointer(DriftModel.dABSMdq[k])[Data[p + 1].Cls, Data[p + 1].Pow], DriftModel.Vec1E, DriftModel.UsedStateCount);

        MulVecEVecESE(DriftModel.Vec1E, BetaS[p + 1], cGR, DriftModel.UsedStateCount);

        DriftModel.dLLdq[0,k] := DriftModel.dLLdq[0,k] - cGR;//Use minus sign in order to minimize LL

      end;//of k

    end;//of DataPoint
    //..........................................................................



  end;//of Chunk
  //****************************************************************************



  //Calculate free pars gradients
  for m := 0 to prDriftModels.Count - 1 do begin
    TQUB_Drift_Model(prDriftModels[m]).Do_dLLdfPar;
  end;

  if (ZeroDetected) or (SomeConfluent)  then begin
    prConsoleCode := MESSAGE_ERROR;
    prConsoleMsg  := Choice(ZeroDetected, 'Zero state vector ' + Choice(SomeConfluent, 'and confluent eigen values ', '') + 'detected', 'Confluent eigen values detected');
    if prQuietOutput then begin
      prConsoleEntry := prQuietEntry;
      Synchronize(prWriteSetConsole);
    end else begin
      Synchronize(prWriteNewConsole);
    end;
  end;

end;


//******************************************************************************
procedure TQUB_Task_Drift.prUpdateModel;
begin

  MakeRatesFromParams(Model.ModelObject, TQUB_Drift_Model(prDriftModels[0]).cPars, 2*TQUB_Drift_Model(prDriftModels[0]).RateCount, OptParams.OptAllowNegRates);

end;

procedure TQUB_Task_Drift.prResetModel;
var
  m :Integer;
begin
  for m := 0 to prOptSegsOldMdl.Rates.RateCount - 1 do begin
    prModel.ModelObject.Rates[m].k0[1] := prOptSegsOldMdl.Rates[m].k0[1];
    prModel.ModelObject.Rates[m].k0[2] := prOptSegsOldMdl.Rates[m].k0[2];
    prModel.ModelObject.Rates[m].k1[1] := prOptSegsOldMdl.Rates[m].k1[1];
    prModel.ModelObject.Rates[m].k1[2] := prOptSegsOldMdl.Rates[m].k1[2];
  end;

  for m := 0 to prDriftModels.Count - 1 do begin
    TQUB_Drift_Model(prDriftModels[m]).Do_InitcPars;
    TQUB_Drift_Model(prDriftModels[m]).Do_InitfPars;
  end;
end;

procedure TQUB_Task_Drift.prShowData;
begin
  //todo :Do it for FileList. LM
  if Data <> nil then begin
    TQUB_QVO_QVD(Data).MoveSelectionIntoWindow(QUB_FirstLast(TQUB_Thread_PChunk(prChunks[0]).Sel.First, TQUB_Thread_PChunk(prChunks[prChunks.Count - 1]).Sel.Last), TQUB_Thread_PChunk(prChunks[0]).Sel.First);
  end;
end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Drift.prDoExecute;
begin
  prConsoleMsg  := 'Drift optimizer running ... ';
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
  inherited;
end;

procedure TQUB_Task_Drift.prDoExecuteStart;
begin

  inherited;

  //todo :Do it for FileList. LM
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  //DriftModels
  prDriftModels   := TList.Create;

  //ExpConds
  prExpCondParams := TQUB_ExpCond_ParametersListList.Create;

end;


procedure TQUB_Task_Drift.prDoExecuteEnd;
var
  i :Integer;
begin
  if Model <> nil then Model.UnlockAny;

  if prSelectionList <> nil then prSelectionList.UnLock;

  //Free Params & DriftModels
  if prDriftModels <> nil then begin
    for i := 0 to prDriftModels.Count - 1 do begin
      TQUB_Drift_Model(prDriftModels[i]).Free;
    end;
    prDriftModels.Free;
    prDriftModels := nil;
  end;

  prExpCondParams.Free;
  prExpCondParams := nil;

  inherited;
end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//

function TQUB_Task_Drift.prChunkData :Boolean;
var
  DataSets         :INode;
  DS, Seg          :INode;

  QVD              :TQUB_QVO_QVD;

  ParamClasses     :TQUB_ExpCond_ParametersList;
  Params           :TQUB_ExpCond_Parameters;
  ParamsAreVar     :TQUB_ExpCond_Parameters;

  Chunk            :TQUB_Thread_PChunk;
  MIPChunk         :TQUB_MIP_PChunk;
  DriftModel       :TQUB_Drift_Model;

  Dat              :TQUB_FirstLast;

  FileIndex        :Integer;

  c, p             :Integer;
begin

  Result           := False;

  prSetChunkParameters;

  ParamsAreVar     := nil;


  try

    DataSets       := DataNoder.GetChosenIdealization( TNode.CreateNull );
    DS             := DataSets.Find('DataSet');
    FileIndex      := 0;
    while not DS.IsNull do begin
      Seg          := DS.Find('Segment');

      if not Seg.IsNull then begin
        //ExpConds
        ParamClasses := TQUB_ExpCond_ParametersList.Create;
        prExpCondParams.Add(ParamClasses);
        Params := TQUB_ExpCond_Parameters.Create;
        ParamClasses.Add(Params);
        Params.ReadTree( DS.Find('ExpCond') );
        for p := 0 to Params.Count - 1 do begin
          Params.AddPar('Cond_' + Params[p].Name, Params[p].Cond_Value,
                        Params[p].Index, Params[p].IsVar, -1);
        end;

        //DriftModel
        DriftModel  := TQUB_Drift_Model.Create(Model.ModelObject, True{DoInitialize}, ParamClasses, @prOptParams);
        prDriftModels.Add(DriftModel);

        QVD        := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DS['FileName'].DataAsString]);

        while not Seg.IsNull do begin

          Dat                   := QUB_FirstLast(Seg.Data[0], Seg.Data[1]);
          Chunk                 := prNewChunk(Dat);

          MIPChunk              := TQUB_MIP_PChunk(Chunk);

          prChunks.Add(MIPChunk);

          MIPChunk.IdlSeg       := Seg.GetNodeMem;
          Node_INCREF( MIPChunk.IdlSeg );

          MIPChunk.SkipPoints   := QUB_FirstLast(0, 0); // ??? Skp;

          MIPChunk.IdlSel       := QUB_FirstLast(0, Seg['DwellCount'].Data[0]); // ??? Idl^;

          MIPChunk.QVD          := QVD;

          MIPChunk.Sampling     := DS['sampling'].Data[0] * 1.0e-3;
          MIPChunk.Scaling      := QVD.DataObject.Scaling;

          MIPChunk.Data         := nil;

          MIPChunk.MIPModel     := DriftModel;

          MIPChunk.FileIndex    := FileIndex;

          Seg        := Seg.SiblingSameName;
        end;

        Inc(FileIndex);
      end;

      DS           := DS.SiblingSameName;
    end;


 (*
    if DataSource = DATA_SOURCE_FILELIST then begin
    //File List

      prConsoleMsg  := 'Opening files ... ';
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);
      OldEntry      := prConsoleEntry;


      //Par Is Var ? ...........................................................
      ParamsAreVar  := TQUB_ExpCond_Parameters.Create;
      ParFirstRow   := -1;
      for w := 1 to prMIPExpCondView.grdInputList.RowCount - 1 do begin
        if prMIPExpCondView.grdInputList.Objects[COL_FILE, w] <> nil then begin
          ParFirstRow := w;
          Break;
        end;
      end;

      for p := 0 to prMIPExpCondView.Parameters.Count - 1 do begin
        ParIsVar := False;
        if ParFirstRow > -1 then begin
          Val1       := StrToFloat(prMIPExpCondView.grdInputList.Cells[prMIPExpCondView.Parameters[p].Index, ParFirstRow]);
          for w := ParFirstRow + 1 to prMIPExpCondView.grdInputList.RowCount - 1 do begin
            if prMIPExpCondView.grdInputList.Objects[COL_FILE, w] <> nil then begin
              Val2     := StrToFloat(prMIPExpCondView.grdInputList.Cells[prMIPExpCondView.Parameters[p].Index, w]);
              ParIsVar := Val1 <> Val2;
              if ParIsVar then Break;
            end;
          end;
        end;
        ParamsAreVar.AddPar(prMIPExpCondView.Parameters[p].Name, prMIPExpCondView.Parameters[p].Value, prMIPExpCondView.Parameters[p].Index, ParIsVar);
      end;
      //........................................................................


      FileIndex     := 0;

      //For each file
      for w := 1 to prMIPExpCondView.grdInputList.RowCount - 1 do begin

        prConsoleEntry := OldEntry;


        FileName := Trim(prMIPExpCondView.grdInputList.Cells[COL_FILE, w]);
        Idl := TQUB_IdlSegments(prMIPExpCondView.grdInputList.Objects[COL_FILE, w]);


        if (FileName <> '') and (Idl <> nil) then begin


          prConsoleMsg   := 'Preparing file ' + FileName;
          prConsoleCode  := MESSAGE_MESSAGE;
          Synchronize(prWriteSetConsole);


          DS             := Idl.ToTree;
          if not DS.Find('Segment').IsNull then begin

            //ExpConds
            ParamClasses := TQUB_ExpCond_ParametersList.Create;
            prExpCondParams.Add(ParamClasses);
            Params := TQUB_ExpCond_Parameters.Create;
            ParamClasses.Add(Params);
            for p := 0 to prMIPExpCondView.Parameters.Count - 1 do begin
              Params.AddPar(prMIPExpCondView.Parameters[p].Name,
                            StrToFloat(prMIPExpCondView.grdInputList.Cells[prMIPExpCondView.Parameters[p].Index, w]),
                            prMIPExpCondView.Parameters[p].Index,
                            ParamsAreVar.ParameterByName[prMIPExpCondView.Parameters[p].Name].IsVar);
            end;


            //DriftModel
            DriftModel  := TQUB_Drift_Model.Create(Model.ModelObject, ParamClasses, @prOptParams);
            prDriftModels.Add(DriftModel);



            s          := 0;
            Seg        := DS.Find('Segment');
            while not Seg.IsNull do begin
              Dat                   := QUB_FirstLast( Integer(Seg.Data[0]), Integer(Seg.Data[1]) );
              Chunk                 := prNewChunk(Dat);
              MIPChunk              := TQUB_MIP_PChunk(Chunk);

              prChunks.Add(MIPChunk);

              MIPChunk.SkipPoints   := QUB_FirstLast(0, 0);

              MIPChunk.IdlSeg       := Seg.GetNodeMem;
              Node_INCREF( MIPChunk.IdlSeg );

              MIPChunk.IdlIdx       := s;
              MIPChunk.IdlSel.First := 0;
              MIPChunk.IdlSel.Last  := Seg['DwellCount'].Data[0] - 1;

              MIPChunk.Sampling     := DS['sampling'].Data[0] * 1.0e-3;
              MIPChunk.Scaling      := 1;

              MIPChunk.Data         := nil;

              MIPChunk.MIPModel     := DriftModel;

              MIPChunk.FileIndex    := FileIndex;

              Inc(s);
              Seg := Seg.SiblingSameName;

            end;

            Inc(FileIndex);

          end;


        end else begin

          prConsoleMsg  := 'Could not use file ' + FileName;
          prConsoleCode := MESSAGE_ERROR;
          Synchronize(prWriteNewConsole);

        end;

      end;

      prConsoleEntry := OldEntry;
      prConsoleMsg  := 'Opening files ... Done';
      Synchronize(prWriteGrayConsole);


    //**************************************************************************
    end else begin
    //Current File


      DS         := TQUB_QVO_QVD(Data).GetChosenIdl;
      if DS.Find('Segment').IsNull then begin
        Result     := True;
        prError    := True;
        prErrorMsg := 'No Idealized Data';
        Exit;
      end;


      //ExpConds
      ParamClasses := TQUB_ExpCond_ParametersList.Create;
      prExpCondParams.Add(ParamClasses);
      Params := TQUB_ExpCond_Parameters.Create;
      ParamClasses.Add(Params);
      for p := 0 to prModel.ModelObject.Parameters.Count - 1 do begin
        Params.AddPar(prModel.ModelObject.Parameters[p].Name, prModel.ModelObject.Parameters[p].Value, prModel.ModelObject.Parameters[p].Index, False);
      end;


      //DriftModel
      DriftModel  := TQUB_Drift_Model.Create(Model.ModelObject, ParamClasses, @prOptParams);
      prDriftModels.Add(DriftModel);


      l             := 0;
      Seg           := DS.Find('Segment');
      while not Seg.IsNull do begin

        Dat                   := QUB_FirstLast( Integer(Seg.Data[0]), Integer(Seg.Data[1]) );
        Chunk                 := prNewChunk(Dat);

        MIPChunk              := TQUB_MIP_PChunk(Chunk);

        prChunks.Add(MIPChunk);

        MIPChunk.IdlSeg       := Seg.GetNodeMem;
        Node_INCREF( MIPChunk.IdlSeg );

        MIPChunk.SkipPoints   := QUB_FirstLast(0, 0); // ??? Skp;

        MIPChunk.IdlIdx       := l; // ??? TQUB_PFirstLast(OIdlx.List[l]).First;
        MIPChunk.IdlSel       := QUB_FirstLast(0, Seg['DwellCount'].Data[0]); // ??? Idl^;

        MIPChunk.Sampling     := DS['sampling'].Data[0] * 1.0e-3;
        MIPChunk.Scaling      := TQUB_QVO_QVD(Data).DataObject.Scaling;

        MIPChunk.Data         := nil;

        MIPChunk.MIPModel     := DriftModel;

        MIPChunk.FileIndex    := 0;

        Inc(l);
        Seg                   := Seg.SiblingSameName;

      end;

    end;//Data Source
*)


    for c := 0 to prChunks.Count - 1 do begin

      Chunk               := prChunks[c];
      Chunk.Status        := CHUNK_STATUS_EMPTY;
      Chunk.Trials        := 0;
      Chunk.Index         := c;

    end;

    if prChunks.Count = 0 then begin

      Result     := True;
      prError    := True;

      if DataSource = DATA_SOURCE_FILELIST then begin
        prErrorMsg := 'The file list is empty (View menu -> MIP Exp Cond List)';
        prWikiTopic := 'Error:No_Data';
      end else begin
        prErrorMsg := 'No Idealized Data';
        prWikiTopic := 'Error:No_Data';
      end;

      Exit;

    end;


  finally


    ParamsAreVar.Free;


  end;

end;

procedure TQUB_Task_Drift.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MIPChunk :TQUB_MIP_PChunk;
begin

  MIPChunk := TQUB_MIP_PChunk(Chunk);

  if MIPChunk.IdlSeg <> nil then begin
    Node_DECREF(MIPChunk.IdlSeg);
    MIPChunk.IdlSeg := nil;
  end;

  if MIPChunk.Data <> nil then begin
    FreeMem(MIPChunk.Data);
    MIPChunk.Data := nil;
  end;
  if MIPChunk.AlphaS <> nil then begin
    Free2(TPMChar(MIPChunk.AlphaS));
    MIPChunk.AlphaS := nil;
  end;
  if MIPChunk.BetaS <> nil then begin
    Free2(TPMChar(MIPChunk.BetaS));
    MIPChunk.BetaS := nil;
  end;
  if MIPChunk.LScale <> nil then begin
    FreeMem(MIPChunk.LScale);
    MIPChunk.LScale := nil;
  end;

  FreeMem(MIPChunk);

end;

function TQUB_Task_Drift.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result           := AllocMem(SizeOf(TQUB_MIP_Chunk));

  Result.Sel       := Sel;

end;

//******************************************************************************
procedure TQUB_Task_Drift.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then begin
    if (prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST) then begin
      prOptSegsOldMdl := TQUB_QFS_QMF.Create('', nil);
      prOptSegsOldMdl.CopyFrom(prModel.ModelObject);
    end;
    prModel.LockAny;
  end;
end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Drift.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_Drift_PropertiesForm do begin

    Self.ChannelIndex           := ChannelIndex;

    //Data Source
    Self.DataSource             := DataSource;//TODO: make data source local, and get it here.

    //Client/Server
    Self.ThreadCount            := ThreadCount;
    Self.Remote                 := Remote;

    Self.QuietOutput            := QuietOutput;

    Self.OptParams              := Params;

    prGroupList                 := txtOptSegmentsGroupList.Text;

  end;//with
end;

procedure TQUB_Task_Drift.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_Drift_PropertiesForm do begin
  ////  ChannelIndex                := Self.ChannelIndex;
  //end;//with
end;



//****************************************************************************//
//                                                                            //
//                      D A T A     C A L L   -   B A C K                     //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Drift.DataCallBack(MIPChunk :TQUB_MIP_PChunk) :Boolean;
var
  Seg          :INode;
  Durs         :TPASingle;
  Clss         :TPAInteger;
  InCount      :Integer;
  SampMS       :Double;

  DataCountMax :Integer;
  Count        :Integer;
  CntPow       :Integer;
  MaxPow       :Integer;
  PowOf2       :Integer;
  PrevCls      :Integer;
  i            :Integer;
begin

  Result       := False;

  try

    if MIPChunk.Data = nil then begin
    //Allocate

      //DC       := nil;

      MaxPow   := MIP_PowerOf2(prOptParams.OptMaxPower);

      // try

        Seg      := TNode.CreateUsing( MIPChunk.IdlSeg );
        InCount  := Seg['DwellCount'].Data[0];
        Clss     := TPAInteger( Seg['Classes'].DataPtr );
        Durs     := TPASingle( Seg['Durations'].DataPtr );


        Inc(TQUB_Drift_Model(prDriftModels[0]).TotalDwellCount, InCount); // MIPChunk.IdlSel.Last - MIPChunk.IdlSel.First + 1);


        //TO DO. LM. This segmentation is causing some problems. I increased it to 100000.
        //DataCountMax       := 10000;
        DataCountMax       := 100000;
        MIPChunk.DataCount := 0;
        ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * DataCountMax);


        SampMS             := MIPChunk.Sampling * 1.0e3;
        PrevCls            := Clss[0]; // DC[0].Cls;


        for i := 0 to InCount - 1 do begin // (MIPChunk.IdlSel.Last - MIPChunk.IdlSel.First) do begin

          Count := Round(Durs[i] / SampMS);
          Inc(TQUB_Drift_Model(prDriftModels[0]).TotalDataPoints, Count);


          //First Dwell: the very first point goes into a one-point-long dwell
          if i = 0 then begin

            CntPow := 0;
            PowOf2 := 1;
            MIPChunk.Data[MIPChunk.DataCount].Cls := 0;//PrevCls;
            Inc(MIPChunk.Data[MIPChunk.DataCount].Cls, TQUB_Drift_Model(prDriftModels[0]).BlackUnit{ + TQUB_Drift_Model(prDriftModels[0]).BlackUnitOffset});
            MIPChunk.Data[MIPChunk.DataCount].Pow := CntPow;
            Inc(MIPChunk.DataCount);
            Dec(Count, PowOf2);
            Inc(TQUB_Drift_Model(prDriftModels[0]).Multiplications);

          end;

          if i <> 0 then begin
          //Other than the first dwell

            CntPow := 0;
            PowOf2 := 1;

            MIPChunk.Data[MIPChunk.DataCount].Cls := (Clss[i] - PrevCls);

            if prOptParams.ModRightOnly then begin

              if MIPChunk.Data[MIPChunk.DataCount].Cls > prOptParams.ModMaxJump then begin
                MIPChunk.Data[MIPChunk.DataCount].Cls := prOptParams.ModMaxJump;
                Inc(TQUB_Drift_Model(prDriftModels[0]).TruncatedJumpCount);
              end;

              if ((MIPChunk.Data[MIPChunk.DataCount].Cls + 1) > TQUB_Drift_Model(prDriftModels[0]).UnitCount) then begin
                Result        := True;
                prError       := True;
                prErrorMsg    := 'Data has higher order jumps than units in the model';
                prWikiTopic   := 'Error:Higher_Order_Jumps';
                prConsoleMsg  := prErrorMsg;
                prConsoleCode := MESSAGE_ERROR;
                Synchronize(prWriteNewConsole);
              end else if (MIPChunk.Data[MIPChunk.DataCount].Cls < 0) then begin
                MIPChunk.Data[MIPChunk.DataCount].Cls := 0;
                prConsoleMsg  := 'Data is reversible, in chunk ' + IntToStr(TQUB_Thread_PChunk(MIPChunk).Index);
                prConsoleCode := MESSAGE_WARNING;
                Synchronize(prWriteNewConsole);
              end;

            end else begin

              if MIPChunk.Data[MIPChunk.DataCount].Cls > prOptParams.ModMaxJump then begin
                MIPChunk.Data[MIPChunk.DataCount].Cls := prOptParams.ModMaxJump;
                Inc(TQUB_Drift_Model(prDriftModels[0]).TruncatedJumpCount);
              end else if MIPChunk.Data[MIPChunk.DataCount].Cls < -prOptParams.ModMaxJump then begin
                MIPChunk.Data[MIPChunk.DataCount].Cls := -prOptParams.ModMaxJump;
                Inc(TQUB_Drift_Model(prDriftModels[0]).TruncatedJumpCount);
              end;

              if (Abs(MIPChunk.Data[MIPChunk.DataCount].Cls) + 1) > TQUB_Drift_Model(prDriftModels[0]).UnitCount then begin
                Result        := True;
                prError       := True;
                prErrorMsg    := 'Data has higher order jumps than units in the model';
                prWikiTopic   := 'Error:Higher_Order_Jumps';
                prConsoleMsg  := prErrorMsg;
                prConsoleCode := MESSAGE_ERROR;
                Synchronize(prWriteNewConsole);
              end;

            end;

            Inc(MIPChunk.Data[MIPChunk.DataCount].Cls, TQUB_Drift_Model(prDriftModels[0]).BlackUnit{ + TQUB_Drift_Model(prDriftModels[0]).BlackUnitOffset});


            MIPChunk.Data[MIPChunk.DataCount].Pow := CntPow;
            Inc(MIPChunk.DataCount);
            Dec(Count, PowOf2);
            Inc(TQUB_Drift_Model(prDriftModels[0]).Multiplications);

          end;

          if Count > 0 then begin

            repeat

              CntPow := MIP_MaxPower(Count, MaxPow, PowOf2);

              Inc(TQUB_Drift_Model(prDriftModels[0]).Multiplications);

              //Need to reallocate ?
              if (1 + MIPChunk.DataCount) > DataCountMax then begin
                Inc(DataCountMax, 10000);
                ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * DataCountMax);
              end;

              MIPChunk.Data[MIPChunk.DataCount].Cls := TQUB_Drift_Model(prDriftModels[0]).BlackUnit{ + TQUB_Drift_Model(prDriftModels[0]).BlackUnitOffset};
              MIPChunk.Data[MIPChunk.DataCount].Pow := CntPow;

              Inc(MIPChunk.DataCount);

              Dec(Count, PowOf2);

            until(Count <= 0);

          end;

          PrevCls := Clss[i];

        end;

        if MIPChunk.DataCount <> DataCountMax then begin
          ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * MIPChunk.DataCount);
        end;

        MIPChunk.AlphaS := EAlloc2(MIPChunk.DataCount, TQUB_Drift_Model(MIPChunk.MIPModel).UsedStateCount);
        MIPChunk.BetaS  := EAlloc2(MIPChunk.DataCount, TQUB_Drift_Model(MIPChunk.MIPModel).UsedStateCount);
        MIPChunk.LScale := EAlloc1(MIPChunk.DataCount);

      //finally

      //  FreeMem(DC);

      //end;


    end else begin
    //DeAllocate

      FreeMem(MIPChunk.Data);
      MIPChunk.Data := nil;
      MIPChunk.DataCount := 0;
      Free2(TPMChar(MIPChunk.AlphaS));
      MIPChunk.AlphaS := nil;
      Free2(TPMChar(MIPChunk.BetaS));
      MIPChunk.BetaS := nil;
      FreeMem(MIPChunk.LScale);
      MIPChunk.LScale := nil;

    end;


  except

    Result := True;

  end;

end;


//******************************************************************************
end.
