{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// aka staircase


unit QUB_Task_Drift_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_Task,
QUB_Properties,
QUB_Task_Properties,
QUB_Task_Data_Properties,
QUB_MIP_Global,
QUB_Node,
QUB_DataList;


const


DEF_QUIETOUTPUT               = 0;


type


TQUB_Task_Drift_PropertiesForm = class(TQUB_Task_PropertiesForm)


  Label7                          :TLabel;
  Label9                          :TLabel;
  Label10                         :TLabel;
  Label11                         :TLabel;
  Label19                         :TLabel;

  btnReset                        :TButton;
  panOpt                          :TPanel;
  txtOptLLConv                    :TEdit;
  txtOptGradConv                  :TEdit;
  txtOptMaxIter                   :TEdit;
  txtOptMaxStep                   :TEdit;
  Label5                          :TLabel;
  txtOptMaxPower                  :TEdit;
    Panel7: TPanel;
    Label6: TLabel;
    txtModMaxJump: TEdit;
    udModMaxJump: TUpDown;
    chkModRightOnly: TCheckBox;
    Label12: TLabel;
    Panel8: TPanel;
    Label15: TLabel;
    Label16: TLabel;
    chkOptSegmentsTogether: TRadioButton;
    chkOptSegmentsGroup: TRadioButton;
    txtOptSegmentsGroupCount: TEdit;
    udOptSegmentsGroupCount: TUpDown;
    chkOptSegmentsIdChunks: TCheckBox;
    chkOptSegmentsGroupList: TRadioButton;
    txtOptSegmentsGroupList: TEdit;
    txtOptSegmentsMaxBatch: TEdit;
    chkOptSegmentsMUXFiles: TCheckBox;
    Panel1: TPanel;
    Label4: TLabel;
    chkOptModeOpt: TRadioButton;
    chkOptModeCheck: TRadioButton;
    Label2: TLabel;
    txtModMinRate: TEdit;
    chkQuietOutput: TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);


protected


  //Data Source
  prDataList                      :TQUB_DataList;

  //Params
  prQuietOutput                   :Boolean;
  prParams                        :TQUB_MIP_Opt_Params;


public


  procedure PropertiesToParams;


  //Optimization
  property OptLLConv                      :Double                    read prParams.OptLLConv;          
  property OptGradConv                    :Double                    read prParams.OptGradConv;        
  property OptMaxStep                     :Double                    read prParams.OptMaxStep;         
  property OptMaxIter                     :Integer                   read prParams.OptMaxIter;         
  property OptMaxPower                    :Integer                   read prParams.OptMaxPower;        
  property OptMode                        :TQUB_MIP_Opt_Mode         read prParams.OptMode;            

  //Model
  property ModMaxJump                     :Integer                   read prParams.ModMaxJump;         
  property ModRightOnly                   :Boolean                   read prParams.ModRightOnly;       
  property ModQMinRate                    :Double                    read prParams.ModQMinRate;

  //Batch
  property OptSegments                    :TQUB_OptimizeSegments     read prParams.OptSegments;            
  property OptSegmentsGroupCount          :Integer                   read prParams.OptSegmentsGroupCount;  
  property OptSegmentsMaxBatch            :Integer                   read prParams.OptSegmentsMaxBatch;    
  property OptSegmentsIdChunks            :Integer                   read prParams.OptIdenticalChunks;     
  property OptSegmentsMUXFiles            :Integer                   read prParams.OptMUXFiles;            

  //Quiet output
  property QuietOutput                    :Boolean                   read prQuietOutput;

  property Params                         :TQUB_MIP_Opt_Params       read prParams;


end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, 
QUB_GUIInterface;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_Drift_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;



  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvLL"]',
                    'Optimizer LLConv', REGTYPE_FLOAT, DEF_OPT_LLCONV,
                    txtOptLLConv,
                    AcceptFloatBetween(MIN_OPT_LLCONV, MAX_OPT_LLCONV),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvGrad"]',
                    'Optimizer GradConv', REGTYPE_FLOAT, DEF_OPT_GRADCONV,
                    txtOptGradConv,
                    AcceptFloatBetween(MIN_OPT_GRADCONV, MAX_OPT_GRADCONV),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxIterations"]',
                    'Optimizer MaxIter', REGTYPE_INT, DEF_OPT_MAXITER,
                    txtOptMaxIter,
                    AcceptIntBetween(MIN_OPT_MAXITER, MAX_OPT_MAXITER),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxStep"]',
                    'Optimizer MaxStep', REGTYPE_FLOAT, DEF_OPT_MAXSTEP,
                    txtOptMaxStep,
                    AcceptFloatBetween(MIN_OPT_MAXSTEP, MAX_OPT_MAXSTEP),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxPower"]',
                    'Optimizer MaxPower', REGTYPE_INT, DEF_OPT_MAXPOWER,
                    txtOptMaxPower,
                    AcceptIntBetween(MIN_OPT_MAXPOWER, MAX_OPT_MAXPOWER),
                   '') );

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Mode"]',
                          'Optimizer Mode', REGTYPE_INT, OptModeToString(DEF_OPT_MODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptModeOpt,
                                       OptModeToString(MODE_OPTIMIZE),
                                       Integer(MODE_OPTIMIZE)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptModeCheck,
                                       OptModeToString(MODE_CHECK),
                                       Integer(MODE_CHECK)))
                       );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["MaxJump"]',
                      'Model MaxJump', REGTYPE_INT, DEF_MOD_MAXJUMP,
                      udModMaxJump,
                      txtModMaxJump,
                      AcceptIntBetween(MIN_MOD_MAXJUMP, MAX_MOD_MAXJUMP)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["RightOnly"]',
                        'Model RightOnly', REGTYPE_INT, DEF_MOD_RIGHTONLY,
                        chkModRightOnly));
  AddEditProperty(TQUB_Property_Edit.Create('["MinRate"]',
                    'Model QMinRate', REGTYPE_FLOAT, DEF_MOD_MINRATE,
                    txtModMinRate,
                    AcceptFloatBetween(MIN_MOD_MINRATE, MAX_MOD_MINRATE),
                    '') );

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["use segments"]',
                          'Optimizer Segments', REGTYPE_INT, SegModeToString(DEF_OPT_SEGMENTS))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsTogether,
                                       SegModeToString(OPTSEGS_TOGETHER),
                                       Integer(OPTSEGS_TOGETHER)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsGroup,
                                       SegModeToString(OPTSEGS_GROUP),
                                       Integer(OPTSEGS_GROUP)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsGroupList,
                                       SegModeToString(OPTSEGS_GROUPLIST),
                                       Integer(OPTSEGS_GROUPLIST)))
                       );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["SegGroupCount"]',
                      'Optimizer Segments GroupCount', REGTYPE_INT, DEF_OPT_SEGMENTS_GROUPCOUNT,
                      udOptSegmentsGroupCount,
                      txtOptSegmentsGroupCount,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["OptSegGroupList"]',
                    'Optimizer Segments GroupList', REGTYPE_STRING, IntToStr(DEF_OPT_SEGMENTS_GROUPCOUNT),
                    txtOptSegmentsGroupList,
                    '',
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SegMaxBatch"]',
                    'Optimizer Segments MaxBatch', REGTYPE_INT, DEF_OPT_MAXBATCH,
                    txtOptSegmentsMaxBatch,
                    AcceptIntBetween(MIN_OPT_MAXBATCH, MAX_OPT_MAXBATCH),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["IdenticalChunks"]',
                        'Optimizer IdenticalChunks', REGTYPE_INT, DEF_OPT_IDENTICALCHUNKS,
                        chkOptSegmentsIdChunks));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["MUXFiles"]',
                        'Optimizer MUXFiles', REGTYPE_INT, DEF_OPT_MUXFILES,
                        chkOptSegmentsMUXFiles));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["QuietOutput"]',
                        'QuietOutput', REGTYPE_BOOL, DEF_QUIETOUTPUT,
                        chkQuietOutput));

  RegKeyName            := 'Software\QUB\Task Staircase optimizer';
  ProfileCat            := 'Stair';
  prBeforeGetProperties := PropertiesToParams;

  WikiTopic             := 'Modeling:Staircase#Properties';

  udModMaxJump.Min := MIN_MOD_MAXJUMP;
  udModMaxJump.Max := MAX_MOD_MAXJUMP;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := panOpt.Left;
  prDataList.Top        := 50;
  prDataList.Width      := ClientWidth - 2*prDataList.Left;
  prDataList.Height     := 100;
  // prDataList.FileListAvail := False;

//  {$include 'QUB_Task_Drift_Properties.htx'}
end;

procedure TQUB_Task_Drift_PropertiesForm.FormShow(Sender: TObject);
var
  Model                  :TQUB_QVO_QVM;
begin
  inherited;

  prDataList.ShowFiles;

  Model                  := ViewportModel.CurrModel;
  if Model <> nil then begin
    prDataList.ShowModelVars( Model, True );
  end;

end;

procedure TQUB_Task_Drift_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
  prDataList.HideAll;
end;


procedure TQUB_Task_Drift_PropertiesForm.PropertiesToParams;
begin
  prParams.OptLLConv             := Properties['DFP']['ConvLL'].Data[0];
  prParams.OptGradConv           := Properties['DFP']['ConvGrad'].Data[0];
  prParams.OptMaxIter            := Properties['DFP']['MaxIterations'].Data[0];
  prParams.OptMaxStep            := Properties['DFP']['MaxStep'].Data[0];
  prParams.OptMaxPower           := Properties['MaxPower'].Data[0];
  prParams.OptMode               := StringToOptMode(Properties['Mode'].DataAsString);
          //Model
  prParams.ModMaxJump            := Properties['MaxJump'].Data[0];
  prParams.ModRightOnly          := 0<>Properties['RightOnly'].Data[0];
  prParams.ModQMinRate           := Properties['MinRate'].Data[0];
  prParams.OptSegments           := StringToSegMode(Properties['use segments'].DataAsString);
  prParams.OptSegmentsGroupCount := Properties['SegGroupCount'].Data[0];
  txtOptSegmentsGroupList.Text   := Properties['OptSegGroupList'].DataAsString;
  prParams.OptSegmentsMaxBatch   := Properties['SegMaxBatch'].Data[0];
  prParams.OptIdenticalChunks    := Properties['IdenticalChunks'].Data[0];
  prParams.OptMUXFiles           := Properties['MUXFiles'].Data[0];
          //Quiet output
  prQuietOutput                  := 0<>Properties['QuietOutput'].Data[0];
end;


//******************************************************************************
end.
