{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// aka Chop Idl


unit QUB_Task_EditIdl;


interface


uses


SysUtils, Classes, Dialogs, Math,
Alloc, AllocTypes,
QUB_Console, QUB_Rich_Console,
QUB_Task,
QUB_QVO,
QUB_QVO_QVD,
QUB_Global,
QUB_Task_EditIdl_Properties,
QUB_Node,
QUB_Node_Python,
QUB_Main;


type


TQUB_EditIdl_Terminator = function (DataSet :INode): Boolean of object;
TQUB_EditIdl_Keeper     = function (Segment, StatSeg :INode): Boolean of object;


TQUB_Task_EditIdl                  = class(TQUB_Task)
protected

  prData                                   :TQUB_QVO_QVD;
  prChannel                                :Integer;
  prProperties                             :INode;

  procedure prApplyDeadTime;
  procedure prChangeClass;
  procedure prJoinClass;
  procedure prFillBlanks;
  procedure prSetAmp;


public

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  procedure Execute; override;

  property Data                  :TQUB_QVO_QVD        read prData              write prData;


end;


implementation


uses

QUB_Thread,
QUB_Thread_Idealize,
QUB_Idealization,
QUB_QFS_QIF_Segments,
QUB_QFS_Channels,
QUB_GUIInterface,
//QUB_Interface_Modeling,
QUB_Viewport_Results,
QUB_Python_Data,
QUB_QVO_QVD_Global;


constructor TQUB_Task_EditIdl.Create(Suspended :Boolean);
begin
  inherited;

  OnChange           := TaskView.OnTaskChange;

  prName             := 'EditIdl';
end;

destructor TQUB_Task_EditIdl.Destroy;
begin
  inherited;
end;

procedure TQUB_Task_EditIdl.Execute;
begin
  if Terminated or prStopped then Exit;

  if Data = nil then begin
    prErrorMsg := 'No Data';
    prWikiTopic := 'Error:No_Data';
    prError    := True;
    Exit;
  end;

  if Data.LockedReadWrite then begin
    prErrorMsg := 'Data is in use by another task';
    prWikiTopic := 'Error:No_Data';
    prError    := True;
    Exit;
  end;

  Data.LockWrite;
  prChannel := Data.ActiveChannel;

  Synchronize(prDoOnChange);

  prConsoleCode := MESSAGE_MESSAGE;
  prConsoleMsg  := 'Task ''EditIdl'' started';
  Synchronize(prWriteNewConsole);

  try

    if 0 <> prProperties['ApplyDeadTime'].Data[0] then prApplyDeadTime;
    if 0 <> prProperties['ChangeClass'].Data[0]   then prChangeClass;
    if 0 <> prProperties['JoinClass'].Data[0]     then prJoinClass;
    if 0 <> prProperties['FillBlanks'].Data[0]    then prFillBlanks;
    if 0 <> prProperties['SetAmp'].Data[0]        then prSetAmp;

  finally
    Beep;

    Data.UnlockWrite;
  end;
end;


procedure TQUB_Task_EditIdl.prApplyDeadTime;
var
  DS, Seg        :INode;
  DeadTime       :Double;
begin
  DeadTime       := prProperties['DeadTime'].Data[0];
  DS             := Data.GetChosenIdl(True);
  Seg            := DS.Find('Segment');
  while not Seg.IsNull do begin
    ApplyDeadTime(Seg, DeadTime);
    Seg          := Seg.SiblingSameName;
  end;
  Data.DataObject.Idealization[prChannel].SetDwells(DS);
end;


procedure TQUB_Task_EditIdl.prChangeClass;
var
  DS, Seg             :INode;
  idwt, fdwt, ldwt    :TPAInteger;
  i, n                :Integer;
  ClassFrom, ClassTo  :Integer;
  Idl                 :TQUB_IdlSegments;

  Sels                :Array of TQUB_FirstLast;
  SelCount, SelCap    :Integer;
  procedure AddSel(First, Last :Integer);
  begin
    if SelCount = SelCap then begin
      SelCap          := SelCap * 2;
      SetLength(Sels, SelCap);
    end;
    Sels[SelCount].First := First;
    Sels[SelCount].Last  := Last;
    Inc(SelCount);
  end;

begin
  SelCount            := 0;
  SelCap              := 1000000;
  SetLength(Sels, SelCap);

  ClassFrom           := prProperties['ClassFrom'].Data[0] - 1;
  ClassTo             := prProperties['ClassTo'].Data[0] - 1;

  DS                  := Data.GetChosenIdl(True);
  Seg                 := DS.Find('Segment');
  while not Seg.IsNull do begin
    idwt              := Seg.Find('Classes').DataPtr;
    fdwt              := Seg.Find('Firsts').DataPtr;
    ldwt              := Seg.Find('Lasts').DataPtr;
    n                 := Seg['DwellCount'].Data[0];
    for i := 0 to n - 1 do begin
      if idwt[i] = ClassFrom then AddSel(fdwt[i], ldwt[i]);
    end;
    Seg               := Seg.SiblingSameName;
  end;

  Idl                 := Data.DataObject.Idealization[prChannel];
  for i := 1 to SelCount - 1 do begin
    // bypass Idl.OnEdit except final (first) dwell
    Idl.Impl.SetDwells(1, @(Sels[i].First), @(Sels[i].Last), @(ClassTo));
  end;
  if SelCount > 0 then begin
    Idl.SetDwell(Sels[0].First, Sels[0].Last, ClassTo);
  end;
end;

type
  TQUB_JoinEvt = record
    First, Last, Cls :Integer;
  end;

procedure TQUB_Task_EditIdl.prJoinClass;
var
  DS, Seg             :INode;
  idwt, fdwt, ldwt    :TPAInteger;
  i, n                :Integer;
  ClassFrom           :Integer;
  Idl                 :TQUB_IdlSegments;
  Prior               :Integer;

  Sels                :Array of TQUB_JoinEvt;
  SelCount, SelCap    :Integer;
  procedure AddSel(First, Last, Cls :Integer);
  begin
    if SelCount = SelCap then begin
      SelCap          := SelCap * 2;
      SetLength(Sels, SelCap);
    end;
    Sels[SelCount].First := First;
    Sels[SelCount].Last  := Last;
    Sels[SelCount].Cls   := Cls;
    Inc(SelCount);
  end;

begin
  SelCount            := 0;
  SelCap              := 1000000;
  SetLength(Sels, SelCap);

  ClassFrom           := prProperties['JoinClassFrom'].Data[0] - 1;
  Prior               := -1;
  DS                  := Data.GetChosenIdl(True);
  Seg                 := DS.Find('Segment');
  while not Seg.IsNull do begin
    idwt              := Seg.Find('Classes').DataPtr;
    fdwt              := Seg.Find('Firsts').DataPtr;
    ldwt              := Seg.Find('Lasts').DataPtr;
    n                 := Seg['DwellCount'].Data[0];
    for i := 0 to n - 1 do begin
      if idwt[i] = ClassFrom then begin
        AddSel(fdwt[i], ldwt[i], Prior);
      end else begin
        Prior := idwt[i];
      end;
    end;
    Seg               := Seg.SiblingSameName;
  end;

  Idl                 := Data.DataObject.Idealization[prChannel];
  for i := 1 to SelCount - 1 do begin
    // bypass Idl.OnEdit except final (first) dwell
    Idl.Impl.SetDwells(1, @(Sels[i].First), @(Sels[i].Last), @(Sels[i].Cls));
  end;
  if SelCount > 0 then begin
    Idl.SetDwell(Sels[0].First, Sels[0].Last, Sels[0].Cls);
  end;
end;


procedure TQUB_Task_EditIdl.prFillBlanks;
var
  DS, Seg             :INode;
  SegLast             :Integer;
  i, n                :Integer;
  f, l, c             :Integer;
  a                   :Double;
  FillClass           :Integer;
  Idl                 :TQUB_IdlSegments;
  IdlIter             :TQUB_IdlSegmentsIter;

  Sels                :Array of TQUB_FirstLast;
  SelCount, SelCap    :Integer;
  procedure AddSel(First, Last :Integer);
  begin
    if SelCount = SelCap then begin
      SelCap          := SelCap * 2;
      SetLength(Sels, SelCap);
    end;
    Sels[SelCount].First := First;
    Sels[SelCount].Last  := Last;
    Inc(SelCount);
  end;

begin
  SelCount            := 0;
  SelCap              := 1000000;
  SetLength(Sels, SelCap);

  FillClass           := prProperties['FillClass'].Data[0] - 1;
  Idl                 := Data.DataObject.Idealization[prChannel];
  IdlIter             := nil;

  DS                  := Data.GetChosenData;
  Seg                 := DS.Find('Segment');
  while not Seg.IsNull do begin
    SegLast           := Seg.Data[1];
    try
      IdlIter         := Idl.FindIter(Seg.Data[0]);
      while 0<>IdlIter.Valid do begin
        IdlIter.Read(@f, @l, @c, @a);
        if f > SegLast then Break;
        if l > SegLast then l := SegLast;
        if c < 0 then AddSel(f, l);
        IdlIter.Next;
      end;
    finally
      IdlIter.Free;
      IdlIter := nil;
    end;
    Seg               := Seg.SiblingSameName;
  end;

  for i := 1 to SelCount - 1 do begin
    // bypass Idl.OnEdit except final (first) dwell
    Idl.Impl.SetDwells(1, @(Sels[i].First), @(Sels[i].Last), @(FillClass));
  end;
  if SelCount > 0 then begin
    Idl.SetDwell(Sels[0].First, Sels[0].Last, FillClass);
  end;
end;


procedure TQUB_Task_EditIdl.prSetAmp;
// QUB_IdlSegments keeps one set of amps per segment
// No way to modify just the amps, so we edit the whole idl:
var
  DS, Seg          :INode;
  Cls              :Integer;
  Amp              :Double;
begin
  Cls            := prProperties['SetAmpClass'].Data[0] - 1;
  Amp            := prProperties['SetAmpValue'].Data[0];
  DS             := Data.GetChosenIdl(True);
  Seg            := DS.Find('Segment');
  while not Seg.IsNull do begin
    if Seg['amp'].DataCount > Cls then Seg['amp'].Data[Cls] := Amp;
    Seg          := Seg.SiblingSameName;
  end;
  Data.DataObject.Idealization[prChannel].SetDwells(DS);
end;


// ****************************************************************************

procedure TQUB_Task_EditIdl.OnGetProperties(Sender :TObject);
begin
  prProperties := TQUB_Task_EditIdl_Properties(Sender).Properties.Clone(True);
end;


procedure TQUB_Task_EditIdl.OnPutProperties(Sender :TObject);
begin
end;

// *****************************************************************************
end.


