{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Apply dead time, change class of events, set amp of class

unit QUB_Task_EditIdl_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Registry, ComCtrls,

RegistryInterface,

QUB_Properties,
QUB_Task_Properties,
QUB_Node,
QUB_QFS_Channels, ExtCtrls;


type


TQUB_Task_EditIdl_Properties = class(TQUB_Task_PropertiesForm)
  chkApplyDeadTime: TCheckBox;
  panDataSource: TPanel;
  Label8: TLabel;
  chkDataSourceSel: TRadioButton;
  chkDataSourceList: TRadioButton;
  chkDataSourceFile: TRadioButton;
    txtDeadTime: TEdit;
    Label1: TLabel;
    chkChangeClass: TCheckBox;
    txtClassFrom: TEdit;
    udClassFrom: TUpDown;
    Label2: TLabel;
    txtClassTo: TEdit;
    udClassTo: TUpDown;
    chkFillBlanks: TCheckBox;
    txtFillClass: TEdit;
    udFillClass: TUpDown;
    chkSetAmp: TCheckBox;
    txtSetAmpClass: TEdit;
    udSetAmpClass: TUpDown;
    txtSetAmp: TEdit;
    Label3: TLabel;
    chkJoinClass: TCheckBox;
    txtJoinClass: TEdit;
    udJoinClass: TUpDown;
    Label4: TLabel;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure chkDataSourceListClick(Sender: TObject);


protected

  prClassCount                             :Integer;

  procedure prSetClassCount(x :Integer);
  procedure prSetInitialProperties(Props :INode); override;

public

  property ClassCount            :Integer              read prClassCount          write prSetClassCount;

end;


implementation


uses


QUB_GUIInterface,
QUB_Task,
QUB_Main;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_EditIdl_Properties.FormCreate(Sender: TObject);
begin
  inherited;

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ApplyDeadTime"]',
                        '', REGTYPE_NONE, 0,
                        chkApplyDeadTime)
                       .Enables(txtDeadTime) );
  AddEditProperty(TQUB_Property_Edit.Create('["DeadTime"]',
                    '', REGTYPE_NONE, 1.0/1.0, txtDeadTime,
                    AcceptFloatGreaterThan(0.0),
                    '%.3g') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ChangeClass"]',
                        '', REGTYPE_NONE, 0,
                        chkChangeClass)
                       .Enables(txtClassFrom)
                       .Enables(udClassFrom)
                       .Enables(txtClassTo)
                       .Enables(udClassTo) );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["ClassFrom"]',
                      '', REGTYPE_NONE, 3,
                      udClassFrom,
                      txtClassFrom,
                      AcceptIntGreaterThan(0)));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["ClassTo"]',
                      '', REGTYPE_NONE, 2,
                      udClassTo,
                      txtClassTo,
                      AcceptIntGreaterThan(0)));

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["JoinClass"]',
                        '', REGTYPE_NONE, 0,
                        chkJoinClass)
                       .Enables(txtJoinClass)
                       .Enables(udJoinClass) );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["JoinClassFrom"]',
                      '', REGTYPE_NONE, 3,
                      udJoinClass,
                      txtJoinClass,
                      AcceptIntGreaterThan(0)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["FillBlanks"]',
                        '', REGTYPE_NONE, 0,
                        chkFillBlanks)
                       .Enables(txtFillClass)
                       .Enables(udFillClass) );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["FillClass"]',
                      '', REGTYPE_NONE, 1,
                      udFillClass,
                      txtFillClass,
                      AcceptIntGreaterThan(0)));

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["SetAmp"]',
                        '', REGTYPE_NONE, 0,
                        chkSetAmp)
                       .Enables(txtSetAmpClass)
                       .Enables(udSetAmpClass)
                       .Enables(txtSetAmp) );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["SetAmpClass"]',
                      '', REGTYPE_NONE, 1,
                      udSetAmpClass,
                      txtSetAmpClass,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["SetAmpValue"]',
                    '', REGTYPE_NONE, 1.0/1.0, txtSetAmp,
                    AcceptFloat,
                    '%.6g') );

  RegKeyName               := 'Software\QUB\Edit Idl';
  ProfileCat               := 'EdIdl';

  WikiTopic                := 'Preprocessing:Edit_Idl';

  ClassCount                   := QUB_QFS_QMF_DEF_NCLASS;

end;

procedure TQUB_Task_EditIdl_Properties.FormShow(Sender: TObject);
begin
  inherited;

  chkDataSourceSel.Checked := chkDataSourceSel.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceList.Checked := chkDataSourceList.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceFile.Checked := (chkDataSourceFile.Tag = Integer(QUB_MainForm.DataSource)) or (DATA_SOURCE_FILELIST = QUB_MainForm.DataSource);
end;


//******************************************************************************
procedure TQUB_Task_EditIdl_Properties.chkDataSourceListClick(Sender: TObject);
begin
  inherited;
  QUB_MainForm.DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;


//******************************************************************************
procedure TQUB_Task_EditIdl_Properties.prSetClassCount(x :Integer);
  procedure Limit(Prop :INode; UpDown :TUpDown);
  begin
    if Prop.Data[0] > x then Prop.Data[0] := x;
    UpDown.Max   := x;
  end;
begin
  if x <= 0   then x := 1;
  prClassCount       := x;
  Limit(Properties['ClassFrom'], udClassFrom);
  Limit(Properties['ClassTo'], udClassTo);
  Limit(Properties['JoinClassFrom'], udJoinClass);
  Limit(Properties['FillClass'], udFillClass);
  Limit(Properties['SetAmpClass'], udSetAmpClass);

end;


procedure TQUB_Task_EditIdl_Properties.prSetInitialProperties(Props :INode);
begin
  inherited;

  // a script could unset this:
  Properties['DialogOnLClick'].SetData(1);
end;


//******************************************************************************
end.
