{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// aka Chop Idl


unit QUB_Task_EvP;


interface


uses


SysUtils, Classes, Dialogs, Math,
Alloc, AllocTypes,
QUB_Console, QUB_Rich_Console,
QUB_Task,
QUB_QVO,
QUB_QVO_QVD,
QUB_Global,
QUB_Task_EvP_Properties,
QUB_Node,
QUB_Node_Python,
QUB_Main;


type


TQUB_EvP_Terminator = function (DataSet :INode): Boolean of object;
TQUB_EvP_Keeper     = function (Segment, StatSeg :INode): Boolean of object;


TQUB_Task_EvP                  = class(TQUB_Task)
protected

  prData                                   :TQUB_QVO_QVD;

  prTermOnDwellLen                         :Boolean;
  prDwLenClass                             :Integer;
  prDwLenComp                              :TQUB_EvP_DwLenComp;
  prDwLen                                  :Double;

  prTermOnDwellCount                       :Boolean;
  prDwCount                                :Integer;

  prDiscardOnClass                         :Boolean;
  prDiscardClass                           :Integer;

  prDiscardOnLen                           :Boolean;
  prDiscardLen                             :Double;

  prDiscardOnCount                         :Boolean;
  prDiscardCount                           :Integer;

  prComputeStats                           :Boolean;
  prApplyDeadTime                          :Boolean;
  prBalanceEvtCount                        :Boolean;
  prOutputList                             :String;

  prStatConfig                             :INode;
  prAcceptedSegCount                       :Integer;
  prLastBurstThisSeg                       :INode;

  prResults                                :INode;
  prResultsDataSet                         :INode;
  prSrcResults, prSrcResDS                 :INode;

  function prCheckTerminate(DataSet :INode): Boolean;
  function prTryKeep(DataSet, Segment, SegIn :INode; Sampling :Double; SegIx, BurstCount :Integer): Boolean;
  procedure prDoBalanceEvtCount(Seg :INode; Sampling :Double);
  function prDoStat(DataSet, Segment :INode): INode;
  procedure prPublishResults;

  function prCheckTermOnDwellLen(DataSet :INode): Boolean;
  function prCheckTermOnDwellCount(DataSet :INode): Boolean;

  function prKeepOnClass(Segment, StatSeg :INode): Boolean;
  function prKeepOnLen(Segment, StatSeg :INode): Boolean;
  function prKeepOnCount(Segment, StatSeg :INode): Boolean;

public

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  procedure Execute; override;

  property Data                  :TQUB_QVO_QVD        read prData              write prData;

  property TermOnDwellLen        :Boolean             read prTermOnDwellLen    write prTermOnDwellLen;
  property DwLenClass            :Integer             read prDwLenClass        write prDwLenClass;
  property DwLenComp             :TQUB_EvP_DwLenComp  read prDwLenComp         write prDwLenComp;
  property DwLen                 :Double              read prDwLen             write prDwLen;

  property TermOnDwellCount      :Boolean             read prTermOnDwellCount  write prTermOnDwellCount;
  property DwCount               :Integer             read prDwCount           write prDwCount;

  property DiscardOnClass        :Boolean             read prDiscardOnClass    write prDiscardOnClass;
  property DiscardClass          :Integer             read prDiscardClass      write prDiscardClass;

  property DiscardOnLen          :Boolean             read prDiscardOnLen      write prDiscardOnLen;
  property DiscardLen            :Double              read prDiscardLen        write prDiscardLen;

  property DiscardOnCount        :Boolean             read prDiscardOnCount    write prDiscardOnCount;
  property DiscardCount          :Integer             read prDiscardCount      write prDiscardCount;

  property ComputeStats          :Boolean             read prComputeStats      write prComputeStats;
  property ApplyDeadTime         :Boolean             read prApplyDeadTime     write prApplyDeadTime;
  property BalanceEvtCount       :Boolean             read prBalanceEvtCount   write prBalanceEvtCount;
  property OutputList            :String              read prOutputList        write prOutputList;

end;


implementation


uses

QUB_Thread,
QUB_Task_Stat,
QUB_Idealization,
QUB_QFS_QIF_Segments,
QUB_QFS_Channels,
QUB_GUIInterface,
//QUB_Interface_Modeling,
QUB_Viewport_Results,
QUB_Python_Data,
QUB_QVO_QVD_Global;


constructor TQUB_Task_EvP.Create(Suspended :Boolean);
begin
  inherited;

  prStatConfig       := TNode.Create('Properties');
  prResults          := TNode.Create('EvP Results');
  prSrcResults       := TNode.Create('EvP Src Results');
  prResultsDataSet   := prResults['DataSet'];

  OnChange           := TaskView.OnTaskChange;

  prName             := 'ChopIdl';
  prAcceptedSegCount := 0;
end;

destructor TQUB_Task_EvP.Destroy;
begin
  inherited;
end;

procedure TQUB_Task_EvP.Execute;
var
  DS                :INode;
  Seg               :INode;
  Durs              :INode;
  Clss              :INode;
  DCount            :INode;
  SegLen            :INode;
  StartMS           :Double;
  Sampling          :Double;

  DSIn, SegIn       :INode;
  ClssIn            :TPAInteger;
  FirstsIn, LastsIn :TPAInteger;
  DwellCountIn      :Integer;

  SegIx             :Integer;
  LastSegIx         :Integer;
  BurstCount        :Integer;
  BurstDur          :Array of Double;
  BurstStart        :Array of Double;
  MeanDur           :Double;
  SrcIx             :Integer;
  SegSrc            :INode;
  bi                :Integer;

  Last              :Integer;

  SumEvent          :Integer;
  EventSoFar        :Integer;

  procedure ResetSeg(First :Integer; MaxDwells :Integer);
  begin
    Seg             := DS['Segment'];
    if Seg.DataCount <> 2 then begin
      Seg.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
    end;

    Seg.Data[0] := First;
    Seg.Data[1] := First - 1;
    
    Seg['start'].SetData( StartMS );

    DCount := Seg['DwellCount'];
    if DCount.DataCount = 0 then begin
      DCount.SetData( 0 );
    end else begin
      DCount.Data[0] := 0;
    end;

    Clss := Seg['Classes'];
    if LongInt(Clss.DataCount) < MaxDwells then begin
      Clss.SetupNumData( NODE_DATATYPE_LONGINT, MaxDwells, 1 );
    end;

    Durs := Seg['Durations'];
    if LongInt(Durs.DataCount) < MaxDwells then begin
      Durs.SetupNumData( NODE_DATATYPE_SINGLE, MaxDwells, 1 );
    end;

    SegLen := Seg['length'];
    if SegLen.DataCount = 0 then begin
      SegLen.SetData( 0.0 / 1.0 );   // division converts constant from currency to double
    end else begin
      SegLen.Data[0] := 0.0;
    end;
  end;
  procedure AddEvent( F, L, C :Integer );
  var
    at                  :Integer;
    dur                 :Double;
  begin
    Seg.Data[1]         := L;

    at                  := DCount.Data[0];
    DCount.Data[0]      := at + 1;
    dur                 := (L - F + 1) * Data.Sampling * 1.0e3;
    StartMS             := StartMS + dur;

    Clss.Data[at]       := C;
    Durs.Data[at]       := dur;
    SegLen.Data[0]      := (Seg.Data[1] - Seg.Data[0] + 1) * Sampling;

    EventSoFar          := EventSoFar + 1;
    if (EventSoFar mod 100) = 0 then begin
      prPercent         := EventSoFar * 100 div SumEvent;
      Synchronize(prDoOnChange);
    end;
  end;


begin
  if Terminated or prStopped then Exit;

  if Data = nil then begin
    prErrorMsg := 'No Data';
    prWikiTopic := 'Error:No_Data';
    prError    := True;
    Exit;
  end;

  DSIn         := Data.GetChosenIdl(False{Fragments}); // Data.DataObject.Idealization[Data.ActiveChannel].ToTree;
  SegIn        := DSIn.Find('Segment');
  Sampling     := Data.Sampling * 1.0e3;
  LastSegIx    := -1;

  if Data.LockedReadWrite then begin
    prErrorMsg := 'Data is in use by another task';
    prWikiTopic := 'Error:No_Data';
    prError    := True;
    Exit;
  end;

  Data.LockWrite;

  if SegIn.IsNull then begin
    prErrorMsg := 'Data (probably) not idealized';
    prWikiTopic := 'Error:No_Data';
    prError    := True;
    Data.UnlockWrite;
    Exit;
  end;

  Synchronize(prDoOnChange);

  prConsoleCode := MESSAGE_MESSAGE;
  prConsoleMsg  := 'Task ''ChopIdl'' started';
  Synchronize(prWriteNewConsole);

  try
    prResultsDataSet['FileName'].SetData( Data.FileName );
    prResultsDataSet['sampling'].SetData( Sampling );

    DS := prResultsDataSet.Clone(True);
    DS['ProcessData'].SetData( 'AsDisplayed' );
    DS.AppendChild( DSIn.Find('ActiveChannel').Clone(True) );
    DS.AppendChild( DSIn.Find('ADChannelCount').Clone(True) );
    DS.AppendChild( DSIn.Find('Channels').Clone(True) );

    prSrcResDS   := prSrcResults.AppendClone(prResultsDataSet, True);

    SumEvent     := 0;
    EventSoFar   := 0;

    SegIn        := DSIn.Find('Segment');
    while not SegIn.IsNull do begin
      SumEvent   := SumEvent + SegIn['DwellCount'].Data[0];
      SegIn      := SegIn.SiblingSameName;
    end;

    //Apply all the criteria, for seg
    SegIn        := DSIn.Find('Segment');
    SrcIx        := 0;
    while not SegIn.IsNull do begin

      if Terminated or prStopped then Exit;

      SegIx        := Data.WhichSegmentForPoint(SegIn.Data[0]);
      if SegIx <> LastSegIx then begin
        prLastBurstThisSeg := TNode.CreateNull;
        LastSegIx  := SegIx;
      end;

      DwellCountIn := SegIn['DwellCount'].Data[0];
      ClssIn       := TPAInteger( SegIn['Classes'].DataPtr );
      FirstsIn     := TPAInteger( SegIn['Firsts'].DataPtr );
      LastsIn      := TPAInteger( SegIn['Lasts'].DataPtr );

      BurstCount   := 0;
      SetLength(BurstDur, 0);
      SetLength(BurstStart, 0);
      MeanDur      := 0.0;

      StartMS      := Data.WhichTimeForPoint( SegIn.Data[0], True ) * 1.0e3;
      ResetSeg( FirstsIn[0], DwellCountIn );

      SegSrc       := prSrcResDS.AppendChild('Segment');
      SegSrc.SetupNumData(NODE_DATATYPE_LONGINT, 2, 1);
      SegSrc.Data[0] := SegIn.Data[0];
      SegSrc.Data[1] := SegIn.Data[1];
      SegSrc['start'].SetData( StartMS );
      SegSrc['length'].SetData( (SegSrc.Data[1] - SegSrc.Data[0]) * Sampling );

      for Last := 0 to DwellCountIn - 1 do begin
        if prStopped then Exit;

        AddEvent( FirstsIn[Last], LastsIn[Last], ClssIn[Last] );
        if prCheckTerminate(DS) then begin
          if prTryKeep(DS, Seg, SegIn, Sampling, SrcIx, BurstCount) then begin
            SetLength(BurstDur, BurstCount+1);
            SetLength(BurstStart, BurstCount+1);
            BurstDur[BurstCount] := (Seg.Data[1] - Seg.Data[0]) * Sampling;
            MeanDur := MeanDur + (Seg.Data[1] - Seg.Data[0]) * Sampling;
            BurstStart[BurstCount] := (Seg.Data[0] - SegIn.Data[0]) * Sampling;
            Inc(BurstCount);
          end;
          if Last < (DwellCountIn - 1) then begin
            //StartMS := StartMS + (FirstsIn[Last+1] - SegIn.Data[0]) * Sampling;
            ResetSeg( FirstsIn[Last+1], (DwellCountIn - 1) - Last );
          end;
        end else if Last = (DwellCountIn - 1) then begin // Idl.Last then begin
          prTryKeep(DS, Seg, SegIn, Sampling, SrcIx, BurstCount);
        end;
      end;

      SegSrc['burst_count'].SetData(BurstCount);
      if BurstCount > 0 then begin
        SegSrc['burst_dur_mean'].SetData(MeanDur / BurstCount);
        for bi := 0 to BurstCount - 1 do begin
          SegSrc[Format('burst_dur %d', [bi+1])].SetData(BurstDur[bi]);
          SegSrc[Format('burst_start %d', [bi+1])].SetData(BurstStart[bi]);
        end;
      end;
      
      Inc(SrcIx);
      SegIn        := SegIn.SiblingSameName;
    end;

    Synchronize( prPublishResults );
  finally
    Beep;

    Data.UnlockWrite;
  end;
end;

function TQUB_Task_EvP.prCheckTerminate(DataSet :INode): Boolean;
begin
  Result := False;

  if (TermOnDwellLen   and prCheckTermOnDwellLen(DataSet))
  or (TermOnDwellCount and prCheckTermOnDwellCount(DataSet)) then begin
    Result := True;
  end;
end;

function TQUB_Task_EvP.prTryKeep(DataSet, Segment, SegIn :INode; Sampling :Double; SegIx, BurstCount :Integer): Boolean;
var
  StatSeg         :INode;
begin
  Result := False;
  if Segment['DwellCount'].Data[0] <= 0 then Exit;

  if BalanceEvtCount then prDoBalanceEvtCount(Segment, Sampling);

  if (DiscardOnClass and not prKeepOnClass(Segment, StatSeg))
  or (DiscardOnLen   and not prKeepOnLen(Segment, StatSeg))
  or (DiscardOnCount and not prKeepOnCount(Segment, StatSeg)) then begin
    Exit;
  end;

  StatSeg := prDoStat(DataSet, Segment);
  prLastBurstThisSeg := prResultsDataSet.AppendChild( StatSeg );
  StatSeg['src_segment'].SetData(SegIx+1);
  StatSeg['src_time'].SetData((Segment.Data[0] - SegIn.Data[0]) * Sampling);
  StatSeg['src_ord'].SetData(BurstCount+1);
  prAcceptedSegCount := prAcceptedSegCount + 1;
  Result := True
end;

procedure TQUB_Task_EvP.prDoBalanceEvtCount(Seg :INode; Sampling :Double);
var
  Classes    :INode;
  Durs       :INode;
  DwellCount :Integer;
  i          :Integer;
  MaxCls     :Integer;
begin
  DwellCount := Seg['DwellCount'].Data[0];
  if DwellCount <= 1 then Exit;
  Classes    := Seg['Classes'];
  if Classes.Data[0] <> Classes.Data[DwellCount-1] then Exit;
  MaxCls     := 0;
  for i := 0 to DwellCount - 1 do begin
    MaxCls   := Max(MaxCls, Classes.Data[i]);
  end;
  if MaxCls > 1 then Exit; // too hard to balance with substates
  Durs       := Seg['Durations'];
  Seg['DwellCount'].Data[0] := DwellCount - 1;
  Seg.Data[1] := Seg.Data[1] - Integer(Round(Double(Durs.Data[DwellCount-1]) / Sampling));
  Seg['length'].SetData( Sampling * (Seg.Data[1] - Seg.Data[0]) );
  Durs.ResizeData(DwellCount-1);
  Classes.ResizeData(DwellCount-1);
end;

function TQUB_Task_EvP.prDoStat(DataSet, Segment :INode): INode;
var
  MaxClass, i, N              :Integer;
  Clss                        :INode;
  StatResult                  :INode;
begin
  if not ComputeStats then begin
    Result                      := Segment.Clone(False);
  end else begin
    // make seg: amp, sd to give stat the number of classes
    Clss                        := Segment['Classes'];
    N                           := Integer(Segment['DwellCount'].Data[0]);
    MaxClass                    := -1;
    for i := 0 to N - 1 do begin
      if Clss.Data[i] > MaxClass then MaxClass := Clss.Data[i];
    end;
    Segment['amp'].SetupNumData( NODE_DATATYPE_DOUBLE, MaxClass+1, 1 );
    Segment['sd'].SetupNumData( NODE_DATATYPE_DOUBLE, MaxClass+1, 1 );

    prStatConfig['ApplyDeadTime'].SetData(Integer(prApplyDeadTime));
    StatResult  := StatDataSet( DataSet, prStatConfig, DataNoder.FillData, DataNoder.NullCallback );
    Result     := StatResult.Find('DataSet').Find('Segment');

    Result['HistogramGroup']['Title'].SetData(Format('Segment %d', [prAcceptedSegCount+1]));

    if not prLastBurstThisSeg.IsNull then begin
      Result['interburst'].SetData(Result['start'].Data[0] - (  prLastBurstThisSeg['start'].Data[0]
                                                              + prLastBurstThisSeg['length'].Data[0]));
    end;
  end;
end;

procedure TQUB_Task_EvP.prPublishResults;
var
  List             :TQUB_QVO_QVD_SelectionList;
  Seg, Next        :INode;
begin
  prResults.Name   := 'ChopIdl ' + ChangeFileExt(ExtractFileName(Data.FileName), '');

  prResults['ResultType'].SetData( 'ChopIdl' );
  prSrcResults['ResultType'].SetData('ChopIdl Src');

  ViewportResults.AddResults( prSrcResults, False );
  ViewportResults.AddResults( prResults, False );

  List             := Data.GetList(OutputList);
  List.Clear;
  Seg              := prResults.Find('DataSet').Find('Segment');
  while not Seg.IsNull do begin
    Next           := Seg.SiblingSameName;
    List.AddSelection(Seg.Data[0], Seg.Data[1], '', not Next.IsNull);
    Seg            := Next;
  end;
  Data.ChooseList(OutputList);
end;


// *****************************************************************************

procedure RemoveLastEvent(Seg :INode; Sampling :Double);
var
  dc                   :Integer;
  lastDur              :Double;
begin
  dc                        := seg['DwellCount'].Data[0];
  lastDur                   := seg['Durations'].Data[ dc - 1 ];

  seg['DwellCount'].Data[0] := dc - 1;
  seg['length'].Data[0]     := seg['length'].Data[0] - lastDur;
  seg.Data[1]               := Integer(Seg.Data[1]) - Integer(Round(1.0e-3 * lastDur / Sampling));
end;

function TQUB_Task_EvP.prCheckTermOnDwellLen(DataSet :INode): Boolean;
var
  i                    :Integer;
  seg                  :INode;
begin
  try
    seg    := DataSet['Segment'];
    i      := Integer(seg['DwellCount'].Data[0]) - 1;

    if DwLenComp = DWLEN_LONGER then begin
      Result := (Integer(seg['Classes'].Data[i]) = (prDwLenClass - 1)) and
                (Double(seg['Durations'].Data[i]) > prDwLen);
    end else begin
      Result := (Integer(seg['Classes'].Data[i]) = (prDwLenClass - 1)) and
                (Double(seg['Durations'].Data[i]) < prDwLen);
    end;

    if Result then begin  // remove that final too-long/short dwell
      RemoveLastEvent(seg, Data.Sampling);
    end;
  except
    Result := False;
  end;
end;

function TQUB_Task_EvP.prCheckTermOnDwellCount(DataSet :INode): Boolean;
begin
  try
    Result := Integer(DataSet['Segment']['DwellCount'].Data[0]) >= prDwCount;
  except
    Result := False;
  end;
end;

// *****************************************************************************

function TQUB_Task_EvP.prKeepOnClass(Segment, StatSeg :INode): Boolean;
var
  Classes          :INode;
  i                :Integer;
begin
  Result  := True;
  try
    Classes := Segment['Classes'];
    for i := 0 to Integer(Segment['DwellCount'].Data[0]) - 1 do begin
      if Integer(Classes.Data[i]) >= (prDiscardClass - 1) then begin
        Result := False;
        Break;
      end;
    end;
  except
  end;
end;

function TQUB_Task_EvP.prKeepOnLen(Segment, StatSeg :INode): Boolean;
begin
  try
    Result := Double(Segment['length'].Data[0]) >= prDiscardLen;
  except
    Result := True;
  end;
end;

function TQUB_Task_EvP.prKeepOnCount(Segment, StatSeg :INode): Boolean;
begin
  try
    Result := Integer(Segment['DwellCount'].Data[0]) >= prDiscardCount;
  except
    Result := True;
  end;
end;

// *****************************************************************************

procedure TQUB_Task_EvP.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_EvP_Properties do begin

    Self.TermOnDwellLen          := 0<>Properties['BurstTerminator']['DwellLength'].Data[0];
    Self.DwLenClass              := Properties['BurstTerminator']['DwellLength']['Class'].Data[0];
    Self.DwLenComp               := StringToDwLenComp(Properties['BurstTerminator']['DwellLength']['Comparison'].DataAsString);
    Self.DwLen                   := Properties['BurstTerminator']['DwellLength']['Length'].Data[0];

    Self.TermOnDwellCount        := 0<>Properties['BurstTerminator']['DwellCount'].Data[0];
    Self.DwCount                 := Properties['BurstTerminator']['DwellCount']['Count'].Data[0];

    Self.DiscardOnClass          := 0<>Properties['DiscardBursts']['DwellInClass'].Data[0];
    Self.DiscardClass            := Properties['DiscardBursts']['DwellInClass']['Class'].Data[0];

    Self.DiscardOnLen            := 0<>Properties['DiscardBursts']['Length'].Data[0];
    Self.DiscardLen              := Properties['DiscardBursts']['Length']['LessThan'].Data[0];

    Self.DiscardOnCount          := 0<>Properties['DiscardBursts']['DwellCount'].Data[0];
    Self.DiscardCount            := Properties['DiscardBursts']['DwellCount']['LessThan'].Data[0];

    Self.ComputeStats            := 0<>Properties['ComputeStats'].Data[0];
    Self.ApplyDeadTime           := 0<>Properties['ApplyDeadTime'].Data[0];
    Self.BalanceEvtCount         := 0<>Properties['BalanceEvtCount'].Data[0];
    Self.OutputList              := Properties['OutputList'].DataAsString;

  end;
end;


procedure TQUB_Task_EvP.OnPutProperties(Sender :TObject);
begin
end;

// *****************************************************************************
end.


