{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// aka Chop Idl

// demonstrates how you can use the QUB_Properties automation system
// but steal back the ui events to handle dis/abling related components
// (and handle setting the tree value yourself)

unit QUB_Task_EvP_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, Registry, ComCtrls,

RegistryInterface,

QUB_Properties,
QUB_Task_Properties,
QUB_Node,
QUB_QFS_Channels, ExtCtrls;


type


TQUB_EvP_DwLenComp = (DWLEN_LONGER, DWLEN_SHORTER);


const


DEF_TERM_DWLEN               = 0;
DEF_DWLEN_CLASS              = 1;
MIN_DWLEN_CLASS              = 1;
DEF_DWLEN_COMP               = DWLEN_LONGER;
DEF_DWLEN                   :Double = 1000;
MIN_DWLEN                   :Double = 1e-10;
MAX_DWLEN                   :Double = 1e+10;

DEF_TERM_DWCOUNT             = 0;
DEF_DWCOUNT                  = 24;
MIN_DWCOUNT                  = 1;
MAX_DWCOUNT                  = 1000000000000000;

DEF_DISCARDON_CLASS          = 0;
DEF_DISCARD_CLASS            = 3;
MIN_DISCARD_CLASS            = 1;

DEF_DISCARDON_LEN            = 0;
DEF_DISCARD_LEN             :Double = 1000;
MIN_DISCARD_LEN             :Double = 1e-10;
MAX_DISCARD_LEN             :Double = 1e+10;

DEF_DISCARDON_COUNT          = 1;
DEF_DISCARD_COUNT            = 10;
MIN_DISCARD_COUNT            = 1;
MAX_DISCARD_COUNT            = 1000000000000000;

DEF_COMPUTE_STATS            = 1;
DEF_APPLYDEADTIME            = 1;
DEF_BALANCEEVTCOUNT          = 1;
DEF_OUTPUTLIST               = 'Chopped Events';


function DwLenCompToString(x :TQUB_EvP_DwLenComp): String;
function StringToDwLenComp(x :String): TQUB_EvP_DwLenComp;


type


TQUB_Task_EvP_Properties = class(TQUB_Task_PropertiesForm)
  GroupBox1                                :TGroupBox;
  chkDwLen                                 :TRadioButton;
  edDwLenClass                             :TEdit;
  udDwLenClass                             :TUpDown;
  edDwLen                                  :TEdit;
  lblDwLen1                                :TLabel;
  chkDwCount                               :TRadioButton;
  edDwCount                                :TEdit;
  lblDwCount1                              :TLabel;
  GroupBox2                                :TGroupBox;
  chkDiscardClass                          :TCheckBox;
  edDiscardClass                           :TEdit;
  udDiscardClass                           :TUpDown;
  lblDiscardClass1                         :TLabel;
  chkDiscardLen                            :TCheckBox;
  edDiscardLen                             :TEdit;
  lblDiscardLen1                           :TLabel;
  chkDiscardCount                          :TCheckBox;
  edDiscardCount                           :TEdit;
  lblDiscardCount                          :TLabel;
  cmbDwLenComp                             :TComboBox;
  chkComputeStats                          :TCheckBox;
    Label1: TLabel;
    txtOutputList: TEdit;
    chkApplyDeadTime: TCheckBox;
    chkBalanceEvtCount: TCheckBox;
    panDataSource: TPanel;
    Label8: TLabel;
    chkDataSourceSel: TRadioButton;
    chkDataSourceList: TRadioButton;
    chkDataSourceFile: TRadioButton;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure chkDwLenClick(Sender: TObject);
  procedure chkDwCountClick(Sender: TObject);
  procedure cmbDwLenCompChange(Sender: TObject);
    procedure chkDataSourceListClick(Sender: TObject);


protected


  prClassCount                             :Integer;

  prDwLenComp                              :TQUB_EvP_DwLenComp;


  procedure prSetClassCount(x :Integer);

  procedure prSetDwLenComp(x :TQUB_EvP_DwLenComp);
  function prGetDwLenCompString: String;
  procedure prSetDwLenCompString(x :String);

  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;

public

  property ClassCount            :Integer              read prClassCount          write prSetClassCount;

  property DwLenComp             :TQUB_EvP_DwLenComp   read prDwLenComp           write prSetDwLenComp;
  property DwLenCompString       :String               read prGetDwLenCompString  write prSetDwLenCompString;

end;


implementation


uses


QUB_GUIInterface,
QUB_Task,
QUB_Main;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_EvP_Properties.FormCreate(Sender: TObject);
begin
  inherited;

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["BurstTerminator"]',
                          '', REGTYPE_NONE, 0)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkDwLen,
                                       0,
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkDwCount,
                                       1,
                                       1))
                       );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["BurstTerminator"]["DwellLength"]["Class"]',
                      'DwLenClass', REGTYPE_INT, DEF_DWLEN_CLASS,
                      udDwLenClass,
                      edDwLenClass,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["BurstTerminator"]["DwellLength"]["Length"]',
                    'Separation Time', REGTYPE_FLOAT, DEF_DWLEN,
                    edDwLen,
                    AcceptFloatBetween(MIN_DWLEN, MAX_DWLEN),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["BurstTerminator"]["DwellCount"]["Count"]',
                    'DwCount', REGTYPE_INT, DEF_DWCOUNT,
                    edDwCount,
                    AcceptIntGreaterThan(0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DiscardBursts"]["DwellInClass"]',
                        'DiscardOnClass', REGTYPE_BOOL, DEF_DISCARDON_CLASS,
                        chkDiscardClass)
                       .Enables(edDiscardClass)
                       .Enables(udDiscardClass)
                       .Enables(lblDiscardClass1) );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["DiscardBursts"]["DwellInClass"]["Class"]',
                      'DiscardClass', REGTYPE_INT, DEF_DISCARD_CLASS,
                      udDiscardClass,
                      edDiscardClass,
                      AcceptIntGreaterThan(0)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DiscardBursts"]["Length"]',
                        'DiscardOnLen', REGTYPE_BOOL, DEF_DISCARDON_LEN,
                        chkDiscardLen)
                       .Enables(edDiscardLen)
                       .Enables(lblDiscardLen1) );
  AddEditProperty(TQUB_Property_Edit.Create('["DiscardBursts"]["Length"]["LessThan"]',
                    'Min Duration', REGTYPE_FLOAT, DEF_DISCARD_LEN,
                     edDiscardLen,
                     AcceptFloatBetween(MIN_DISCARD_LEN, MAX_DISCARD_LEN),
                     '%.6g') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DiscardBursts"]["DwellCount"]',
                        'DiscardOnCount', REGTYPE_BOOL, DEF_DISCARDON_COUNT,
                        chkDiscardCount)
                       .Enables(edDiscardCount)
                       .Enables(lblDiscardCount) );
  AddEditProperty(TQUB_Property_Edit.Create('["DiscardBursts"]["DwellCount"]["LessThan"]',
                    'Min Events', REGTYPE_INT, DEF_DISCARD_COUNT,
                     edDiscardCount,
                     AcceptIntGreaterThan(0),
                     '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ComputeStats"]',
                        'ComputeStats', REGTYPE_BOOL, DEF_COMPUTE_STATS,
                        chkComputeStats)
                       .Enables(chkApplyDeadTime) );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ApplyDeadTime"]',
                        '', REGTYPE_NONE, DEF_APPLYDEADTIME,
                        chkApplyDeadTime));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["BalanceEvtCount"]',
                        '', REGTYPE_NONE, DEF_BALANCEEVTCOUNT,
                        chkBalanceEvtCount));
  AddEditProperty(TQUB_Property_Edit.Create('["OutputList"]',
                     '', REGTYPE_NONE, DEF_OUTPUTLIST,
                     txtOutputList,
                     '', '') );

  RegKeyName               := 'Software\QUB\Event List Processor Data';
  ProfileCat               := 'ChpIdl';

  WikiTopic                := 'Preprocessing:Chop_Idl';

  // steal back ui events for properties we'll set in the tree and do extra enabling for
  chkDwLen.OnClick         := chkDwLenClick;
  chkDwCount.OnClick       := chkDwCountClick;

  // manually set a couple properties that don't fit the new scheme
  Properties['BurstTerminator']['DwellLength'].SetData((Properties['BurstTerminator'].Data[0] + 1) mod 2);
  Properties['BurstTerminator']['DwellCount'].SetData(Properties['BurstTerminator'].Data[0]);

  // and force them to fix the enabling
  chkDwLenClick(chkDwLen);
  chkDwCountClick(chkDwCount);

  ClassCount                   := QUB_QFS_QMF_DEF_NCLASS;

end;

procedure TQUB_Task_EvP_Properties.FormShow(Sender: TObject);
begin
  inherited;

  chkDataSourceSel.Checked := chkDataSourceSel.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceList.Checked := chkDataSourceList.Tag = Integer(QUB_MainForm.DataSource);
  chkDataSourceFile.Checked := (chkDataSourceFile.Tag = Integer(QUB_MainForm.DataSource)) or (DATA_SOURCE_FILELIST = QUB_MainForm.DataSource);
end;


//******************************************************************************
procedure TQUB_Task_EvP_Properties.chkDwLenClick(Sender: TObject);
var
  x      :Boolean;
begin
  inherited;
  x   := chkDwLen.Checked;
  Properties['BurstTerminator']['DwellLength'].Data[0] := Integer(x);
  Properties['BurstTerminator']['DwellCount'].Data[0]  := (Integer(x) + 1) mod 2;
  Properties['BurstTerminator'].Data[0] := (Integer(x) + 1) mod 2;
  edDwLenClass.Enabled   := x;
  udDwLenClass.Enabled   := x;
  cmbDwLenComp.Enabled   := x;
  edDwLen.Enabled        := x;
  lblDwLen1.Enabled      := x;

  edDwCount.Enabled      := not x;
  lblDwCount1.Enabled    := not x;
end;

procedure TQUB_Task_EvP_Properties.chkDwCountClick(Sender: TObject);
var
  x      :Boolean;
begin
  inherited;
  x   := chkDwCount.Checked;
  Properties['BurstTerminator']['DwellCount'].Data[0]  := Integer(x);
  Properties['BurstTerminator']['DwellLength'].Data[0] := (Integer(x) + 1) mod 2;
  Properties['BurstTerminator'].Data[0] := Integer(x);
  edDwLenClass.Enabled   := not x;
  udDwLenClass.Enabled   := not x;
  cmbDwLenComp.Enabled   := not x;
  edDwLen.Enabled        := not x;
  lblDwLen1.Enabled      := not x;

  edDwCount.Enabled      := x;
  lblDwCount1.Enabled    := x;
end;

procedure TQUB_Task_EvP_Properties.chkDataSourceListClick(Sender: TObject);
begin
  inherited;
  QUB_MainForm.DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;


//******************************************************************************
procedure TQUB_Task_EvP_Properties.prSetClassCount(x :Integer);
var
  Prop         :INode;
begin
  if x <= 0   then x := 1;
  prClassCount       := x;
  Prop := Properties['BurstTerminator']['DwellLength']['Class'];
  if Prop.Data[0] > x then Prop.Data[0] := x;
  udDwLenClass.Max   := x;
  Prop := Properties['DiscardBursts']['DwellInClass']['Class'];
  if Prop.Data[0] > x then Prop.Data[0] := x;
  udDiscardClass.Max := x;

  prUpdateComponents;
end;

procedure TQUB_Task_EvP_Properties.prSetDwLenComp(x :TQUB_EvP_DwLenComp);
begin
  prDwLenComp            := x;
  cmbDwLenComp.ItemIndex := Integer(x);
end;

function TQUB_Task_EvP_Properties.prGetDwLenCompString: String;
begin
  Result := DwLenCompToString(prDwLenComp);
end;

procedure TQUB_Task_EvP_Properties.prSetDwLenCompString(x :String);
begin
  DwLenComp := StringToDwLenComp(x);
end;

procedure TQUB_Task_EvP_Properties.cmbDwLenCompChange(Sender: TObject);
begin
  inherited;
  DwLenComp  := TQUB_EvP_DwLenComp( cmbDwLenComp.ItemIndex );
end;

function DwLenCompToString(x :TQUB_EvP_DwLenComp): String;
begin
  if x = DWLEN_LONGER then begin
    Result               := 'longer than';
  end else begin // if DwLenComp = DWLEN_SHORTER
    Result               := 'shorter than';
  end;
end;

function StringToDwLenComp(x :String): TQUB_EvP_DwLenComp;
begin
  if x = 'longer than' then begin
    Result            := DWLEN_LONGER;
  end else{if x = 'shorter than' then}begin
    Result            := DWLEN_SHORTER;
  end;
end;


//******************************************************************************

procedure TQUB_Task_EvP_Properties.prSetInitialProperties(Props :INode);
var
  Reg   :TRegistry;
  Prop  :INode;
  iVal  :Integer;
begin
  inherited;

  Properties['DialogOnLClick'].SetData(1);
  // a script could unset this

  Reg   := nil; // will be created as needed by ReadXRegistry*
  try
    Prop  := Props['BurstTerminator']['DwellLength']['Comparison'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, 'DwLenComp', iVal) then begin
        DwLenComp := TQUB_EvP_DwLenComp(iVal);
      end else begin
        DwLenComp := DEF_DWLEN_COMP;
      end;
      Prop.SetData(DwLenCompString);
    end;
  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_Task_EvP_Properties.prUpdateComponents;
var
  x            :Integer;
begin
  inherited;

  dwLenCompString    := Properties['BurstTerminator']['DwellLength']['Comparison'].DataAsString;

  x                  := Properties['BurstTerminator'].Data[0];
  chkDwLen.Checked   := x = 0;
  chkDwCount.Checked := x = 1;
  chkDwLenClick(chkDwLen);
  chkDwCountClick(chkDwCount);
end;


//******************************************************************************
end.
