{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
unit QUB_Task_Exp_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls,
ExtCtrls, ComCtrls, Registry, Grids, Math,

RegistryInterface,

QUB_Task,
QUB_Task_Data,
QUB_Properties,
QUB_Task_Properties,
QUB_Task_Data_Properties,
QUB_Q_Model,
QUB_Exp_Global,
QUB_Node,
QUB_DataList, QUB_Table;


const


DEF_QUIETOUTPUT               = 0;

DEF_UPDATEPARAMS              = 0;


type


TQUB_Task_Exp_PropertiesForm  = class(TQUB_Task_Data_PropertiesForm)

  btnReset                        :TButton;
  Panel2                          :TPanel;
  Label11                         :TLabel;
  Label7                          :TLabel;
  Label9                          :TLabel;
  Label10                         :TLabel;
  Label19                         :TLabel;
  txtOptSSConv                    :TEdit;
  txtOptGradConv                  :TEdit;
  txtOptMaxIter                   :TEdit;
  txtOptMaxStep                   :TEdit;
  chkQuietOutput                  :TCheckBox;
  Panel6                          :TPanel;
  chkOptShowFit                   :TCheckBox;
  chkOptSkip                      :TCheckBox;
  txtOptSkipTime                  :TEdit;
  Label13                         :TLabel;
  chkOptFitDataWhole              :TRadioButton;
  chkOptFitDataList               :TRadioButton;
  Panel7                          :TPanel;
  chkOptModeOpt                   :TRadioButton;
  chkOptModeCheck                 :TRadioButton;
  Label12                         :TLabel;
  Panel8                          :TPanel;
  Label15                         :TLabel;
  chkOptSegmentsTogether          :TRadioButton;
  chkOptSegmentsGroup             :TRadioButton;
  txtOptSegmentsGroupCount        :TEdit;
  udOptSegmentsGroupCount         :TUpDown;
  lstFitData                      :TComboBox;
  Panel3                          :TPanel;
  chkOptShowFitAvg                :TRadioButton;
  chkOptSegmentsGroupList         :TRadioButton;
  txtOptSegmentsGroupList         :TEdit;
  Label16                         :TLabel;
  txtOptSegmentsMaxBatch          :TEdit;
  chkOptRestart                   :TCheckBox;
  chkOptMUXFiles                  :TCheckBox;
  chkOptShowFitRes                :TRadioButton;
  pageMethod                      :TPageControl;
  tabExp                          :TTabSheet;
  Label14                         :TLabel;
  Label4                          :TLabel;
  txtStateCount                   :TEdit;
  txtIeq                          :TEdit;
  udStateCount                    :TUpDown;
  tabSine                         :TTabSheet;
  txtSineAmp                      :TEdit;
  txtSineFreq                     :TEdit;
  txtSinePhase                    :TEdit;
  Label20                         :TLabel;
  chkFixedAmp                     :TCheckBox;
  chkFixedFreq                    :TCheckBox;
  chkFixedPhase                   :TCheckBox;
  chkFixedIeq                     :TCheckBox;
  txtSineIeq                      :TEdit;
  tabLeakCap                      :TTabSheet;
  Label3                          :TLabel;
  Label17                         :TLabel;
  txtLKCGL                        :TEdit;
  txtLeakCapN                     :TEdit;
  udLKCN                          :TUpDown;
  Label18                         :TLabel;
  txtLKCVL                        :TEdit;
  chkUpdateParams                 :TCheckBox;
  chkOptIdenticalChunks           :TCheckBox;
  Label21                         :TLabel;
  txtLKCV0Points                  :TEdit;
  Label22                         :TLabel;
  txtLKCSkipPoints                :TEdit;
  Label23                         :TLabel;
  txtLKCVChannel                  :TEdit;
  chkOptPropagateParams           :TCheckBox;
  Button1                         :TButton;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure lstFitDataClick(Sender: TObject);
  procedure udStateCountClick(Sender: TObject; Button: TUDBtnType);
  procedure udLKCNClick(Sender: TObject; Button: TUDBtnType);


protected


  grdModel                        :TQUB_Table;
  grdLKCAmpTau                    :TQUB_Table;

  //Data Source
  prDataList                      :TQUB_DataList;

  //QuietOutput
  prQuietOutput                   :Boolean;

  //UpdateParams
  prUpdateParams                  :Boolean;
  
  //Params
  prParams                        :TQUB_Exp_Opt_Params;

  //FitDataList
  prFitDataListName               :String;


  //Optimizer
  procedure prSetFitDataListIdx(i :Integer);
  function prGetFitDataListIdx :Integer;
  procedure prSetFitDataList;

  procedure prSetStateCount(c :Integer);
  procedure prSetTau(Index :Integer; t :Double);
  function prGetTau(Index :Integer) :Double;
  procedure prSetAmp(Index :Integer; a :Double);
  function prGetAmp(Index :Integer) :Double;

  procedure prSetLKCN(n :Integer);
  procedure prSetLKCTau(Index :Integer; t :Double);
  function prGetLKCTau(Index :Integer) :Double;
  procedure prSetLKCAmp(Index :Integer; a :Double);
  function prGetLKCAmp(Index :Integer) :Double;

  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public


  procedure PropertiesToParams;


  //Quiet output
  property QuietOutput                    :Boolean                   read prQuietOutput;

  //UpdateParams
  property UpdateParams                   :Boolean                   read prUpdateParams;

  //Optimization
  property StateCount                     :Integer                   read prParams.StateCount             write prSetStateCount;
  property Ieq                            :Double                    read prParams.Ieq;
  property Tau[Index :Integer]            :Double                    read prGetTau                        write prSetTau;
  property Amp[Index :Integer]            :Double                    read prGetAmp                        write prSetAmp;
  property SineIeq                        :Double                    read prParams.SineIeq;
  property SineAmp                        :Double                    read prParams.SineAmp;
  property SineFreq                       :Double                    read prParams.SineFreq;
  property SinePhase                      :Double                    read prParams.SinePhase;
  property FixedIeq                       :Boolean                   read prParams.FixedIeq;
  property FixedAmp                       :Boolean                   read prParams.FixedAmp;
  property FixedFreq                      :Boolean                   read prParams.FixedFreq;
  property FixedPhase                     :Boolean                   read prParams.FixedPhase;             
  property LKCN                           :Integer                   read prParams.LKCN                   write prSetLKCN;
  property LKCGL                          :Double                    read prParams.LKCGL;
  property LKCVL                          :Double                    read prParams.LKCVL;                  
  property LKCTau[Index :Integer]         :Double                    read prGetLKCTau                     write prSetLKCTau;
  property LKCAmp[Index :Integer]         :Double                    read prGetLKCAmp                     write prSetLKCAmp;
  property LKCVChannel                    :Integer                   read prParams.LKCVChannel;
  property LKCV0Points                    :Integer                   read prParams.LKCV0Points;            
  property LKCSkipPoints                  :Integer                   read prParams.LKCSkipPoints;          
  property OptSSConv                      :Double                    read prParams.OptSSConv;              
  property OptGradConv                    :Double                    read prParams.OptGradConv;            
  property OptMaxStep                     :Double                    read prParams.OptMaxStep;             
  property OptMaxIter                     :Integer                   read prParams.OptMaxIter;             
  property OptRestart                     :Integer                   read prParams.OptRestart;             
  property OptFitDataMode                 :TQUB_Mac_Opt_FitDataMode  read prParams.OptFitDataMode;         
  property OptShowFit                     :Integer                   read prParams.OptShowFit;             
  property OptShowFitWhat                 :Integer                   read prParams.OptShowFitWhat;         
  property OptSkip                        :Integer                   read prParams.OptSkip;                
  property OptSkipTime                    :Double                    read prParams.OptSkipTime;            
  property OptMode                        :TQUB_Mac_Opt_Mode         read prParams.OptMode;                
  property OptSegments                    :TQUB_OptimizeSegments     read prParams.OptSegments;            
  property OptSegmentsGroupCount          :Integer                   read prParams.OptSegmentsGroupCount;  
  property OptSegmentsMaxBatch            :Integer                   read prParams.OptSegmentsMaxBatch;    
  property OptIdenticalChunks             :Integer                   read prParams.OptIdenticalChunks;     
  property OptMUXFiles                    :Integer                   read prParams.OptMUXFiles;            
  property OptPropagateParams             :Integer                   read prParams.OptPropagateParams;

  property FitDataListIdx                 :Integer                   read prGetFitDataListIdx              write prSetFitDataListIdx;

  property Params                         :TQUB_Exp_Opt_Params       read prParams;


end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface, QUB_Main;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_Exp_PropertiesForm.prSetFitDataListIdx(i :Integer);
begin
  lstFitData.ItemIndex := i;
  prFitDataListName    := '';
  if lstFitData.ItemIndex <> -1 then begin
    prFitDataListName := lstFitData.Items[lstFitData.ItemIndex];
  end;
end;

function TQUB_Task_Exp_PropertiesForm.prGetFitDataListIdx :Integer;
begin
  if lstFitData.Items.Count > 0 then Result := lstFitData.ItemIndex else Result := -1;
end;

procedure TQUB_Task_Exp_PropertiesForm.prSetFitDataList;
begin
  FitDataListIdx := lstFitData.Items.IndexOf(prFitDataListName);
end;

//StateCount
procedure TQUB_Task_Exp_PropertiesForm.prSetStateCount(c :Integer);
var
  i :Integer;
begin
  prParams.StateCount   := c;
  Properties['StateCount'].Data[0] := c;
  Properties['Tau'].ResizeData(c - 1);
  Properties['Amp'].ResizeData(c - 1);
  if udStateCount.Position <> c then udStateCount.Position := c;
  grdModel.RowCount     := prParams.StateCount;
  for i := 0 to prParams.StateCount - 2 do begin
    grdModel.Cells[0, i + 1] := IntToStr(i + 1);
    Tau[i]           := Tau[i];
    Amp[i]           := Amp[i];
  end;
end;
//Tau
procedure TQUB_Task_Exp_PropertiesForm.prSetTau(Index :Integer; t :Double);
begin
  prParams.Tau[Index] := t;
  Properties['Tau'].Data[Index] := t;
  if Index < (grdModel.RowCount - 1) then grdModel.Cells[1, Index + 1] := FloatToStr(prParams.Tau[Index]*1e3);
end;
function TQUB_Task_Exp_PropertiesForm.prGetTau(Index :Integer) :Double;
begin
  if (Index < (grdModel.RowCount - 1)) and (grdModel.Cells[1, Index + 1] <> '') then begin
    try
      prParams.Tau[Index]          := StrToFloat(grdModel.Cells[1, Index + 1])*1e-3;
    except
      grdModel.Cells[1, Index + 1] := FloatToStr(prParams.Tau[Index]*1e3);
    end;
  end;
  Result := prParams.Tau[Index];
end;
//Amp
procedure TQUB_Task_Exp_PropertiesForm.prSetAmp(Index :Integer; a :Double);
begin
  prParams.Amp[Index] := a;
  Properties['Amp'].Data[Index] := a;
  if Index < (grdModel.RowCount - 1) then grdModel.Cells[2, Index + 1] := FloatToStr(a);
end;
function TQUB_Task_Exp_PropertiesForm.prGetAmp(Index :Integer) :Double;
begin
  if (Index < (grdModel.RowCount - 1)) and (grdModel.Cells[2, Index + 1] <> '') then begin
    try
      prParams.Amp[Index]          := StrToFloat(grdModel.Cells[2, Index + 1]);
    except
      grdModel.Cells[2, Index + 1] := FloatToStr(prParams.Amp[Index]);
    end;
  end;
  Result := prParams.Amp[Index];
end;

//LKCN
procedure TQUB_Task_Exp_PropertiesForm.prSetLKCN(n :Integer);
var
  i :Integer;
begin
  prParams.LKCN   := n;
  Properties['LKCN'].Data[0] := n;
  Properties['LKCTau'].ResizeData(n);
  Properties['LKCAmp'].ResizeData(n);
  if udLKCN.Position <> n then udLKCN.Position := n;
  grdLKCAmpTau.RowCount     := prParams.LKCN + 1;
  for i := 0 to prParams.LKCN - 1 do begin
    grdLKCAmpTau.Cells[0, i + 1] := IntToStr(i + 1);
    LKCTau[i] := LKCTau[i];
    LKCAmp[i] := LKCAmp[i];
  end;
end;

procedure TQUB_Task_Exp_PropertiesForm.prSetLKCTau(Index :Integer; t :Double);
begin
  prParams.LKCTau[Index] := t;
  Properties['LKCTau'].Data[Index] := t;
  if Index < (grdLKCAmpTau.RowCount - 1) then grdLKCAmpTau.Cells[1, Index + 1] := FloatToStr(prParams.LKCTau[Index]*1e3);
end;
function TQUB_Task_Exp_PropertiesForm.prGetLKCTau(Index :Integer) :Double;
begin
  if (Index < (grdLKCAmpTau.RowCount - 1)) and (grdLKCAmpTau.Cells[1, Index + 1] <> '') then begin
    try
      prParams.LKCTau[Index]     := StrToFloat(grdLKCAmpTau.Cells[1, Index + 1])*1e-3;
    except
      grdLKCAmpTau.Cells[1, Index + 1] := FloatToStr(prParams.LKCTau[Index]*1e3);
    end;
  end;
  Result := prParams.LKCTau[Index];
end;

//Amp
procedure TQUB_Task_Exp_PropertiesForm.prSetLKCAmp(Index :Integer; a :Double);
begin
  prParams.LKCAmp[Index] := a;
  Properties['LKCAmp'].Data[Index] := a;
  if Index < (grdLKCAmpTau.RowCount - 1) then grdLKCAmpTau.Cells[2, Index + 1] := FloatToStr(a);
end;
function TQUB_Task_Exp_PropertiesForm.prGetLKCAmp(Index :Integer) :Double;
begin
  if (Index < (grdLKCAmpTau.RowCount - 1)) and (grdLKCAmpTau.Cells[2, Index + 1] <> '') then begin
    try
      prParams.LKCAmp[Index]     := StrToFloat(grdLKCAmpTau.Cells[2, Index + 1]);
    except
      grdLKCAmpTau.Cells[2, Index + 1] := FloatToStr(prParams.LKCAmp[Index]);
    end;
  end;
  Result := prParams.LKCAmp[Index];
end;

//******************************************************************************
procedure TQUB_Task_Exp_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  grdModel                              := TQUB_Table.Create(Self);
  grdModel.Parent                       := tabExp;
  grdModel.Left                         := 4;
  grdModel.Top                          := 50;
  grdModel.Width                        := 257;
  grdModel.Height                       := 131;
  grdModel.ColCount                     := 3;
  grdModel.DefaultRowHeight             := 16;
  grdModel.Options                      := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRangeSelect, goRowSizing, goColSizing, goRowMoving, goColMoving, goThumbTracking];
  grdModel.TabOrder                     := 0;


  grdLKCAmpTau                          := TQUB_Table.Create(Self);
  grdLKCAmpTau.Parent                   := tabLeakCap;
  grdLKCAmpTau.Left                     := 4;
  grdLKCAmpTau.Top                      := 74;
  grdLKCAmpTau.Width                    := 257;
  grdLKCAmpTau.Height                   := 107;
  grdLKCAmpTau.ColCount                 := 3;
  grdLKCAmpTau.DefaultRowHeight         := 16;
  grdLKCAmpTau.Options                  := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRangeSelect, goRowSizing, goColSizing, goRowMoving, goColMoving, goThumbTracking];
  grdLKCAmpTau.TabOrder                 := 3;

  prDataList                            := TQUB_DataList.Create(Self);
  prDataList.Parent                     := Self;
  prDataList.Left                       := 20;
  prDataList.Top                        := 50;
  prDataList.Width                      := 600;
  prDataList.Height                     := 100;
  // prDataList.FileListAvail             := False;

  udStateCount.Min                      := 2;
  udStateCount.Max                      := MAX_STATE_COUNT;

  grdModel.RowCount                     := 2;
  grdModel.Cells[0, 0]                  := 'Idx';
  grdModel.Cells[1, 0]                  := 'Tau [ms]';
  grdModel.Cells[2, 0]                  := 'Amp';

  udLKCN.Min                            := 1;
  udLKCN.Max                            := MAX_STATE_COUNT;

  grdLKCAmpTau.RowCount                 := 2;
  grdLKCAmpTau.Cells[0, 0]              := 'Idx';
  grdLKCAmpTau.Cells[1, 0]              := 'Tau [ms]';
  grdLKCAmpTau.Cells[2, 0]              := 'Amp [pA]';

  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvSS"]',
                    'Optimizer SSConv', REGTYPE_FLOAT, DEF_OPT_SSCONV,
                    txtOptSSConv,
                    AcceptFloatBetween(MIN_OPT_SSCONV, MAX_OPT_SSCONV),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvGrad"]',
                    'Optimizer GradConv', REGTYPE_FLOAT, DEF_OPT_GRADCONV,
                    txtOptGradConv,
                    AcceptFloatBetween(MIN_OPT_GRADCONV, MAX_OPT_GRADCONV),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxIterations"]',
                    'Optimizer MaxIter', REGTYPE_INT, DEF_OPT_MAXITER,
                    txtOptMaxIter,
                    AcceptIntBetween(MIN_OPT_MAXITER, MAX_OPT_MAXITER),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DFP"]["Restart"]',
                        'Optimizer Restart', REGTYPE_INT, DEF_OPT_RESTART,
                        chkOptRestart));
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxStep"]',
                    'Optimizer MaxStep', REGTYPE_FLOAT, DEF_OPT_MAXSTEP,
                    txtOptMaxStep,
                    AcceptFloatBetween(MIN_OPT_MAXSTEP, MAX_OPT_MAXSTEP),
                   '') );

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["OptFitDataMode"]',
                          'Optimizer FitDataMode', REGTYPE_INT, MacFitDataModeStringOf(DEF_OPT_FITDATAMODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptFitDataWhole,
                                       MacFitDataModeStringOf(FITDATA_WHOLE),
                                       Integer(FITDATA_WHOLE)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptFitDataList,
                                       MacFitDataModeStringOf(FITDATA_LIST),
                                       Integer(FITDATA_LIST)))
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ShowFit"]',
                        'Optimizer ShowFit', REGTYPE_INT, DEF_OPT_SHOWFIT,
                        chkOptShowFit));
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["ShowFitWhat"]',
                          'Optimizer ShowFit What', REGTYPE_INT, DEF_OPT_SHOWFIT_WHAT)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptShowFitAvg,
                                       SHOWFIT_AVG,
                                       SHOWFIT_AVG))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptShowFitRes,
                                       SHOWFIT_RES,
                                       SHOWFIT_RES))
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["IdenticalChunks"]',
                        'Optimizer IdenticalChunks', REGTYPE_INT, DEF_OPT_IDENTICALCHUNKS,
                        chkOptIdenticalChunks));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["MUXFiles"]',
                        'Optimizer MUXFiles', REGTYPE_INT, DEF_OPT_MUXFILES,
                        chkOptMUXFiles));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["PropagateParams"]',
                        'Optimizer PropagateParams', REGTYPE_INT, DEF_OPT_PROPAGATEPARAMS,
                        chkOptPropagateParams));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["OptSkip"]',
                        'Optimizer Skip', REGTYPE_INT, DEF_OPT_SKIP,
                        chkOptSkip));
  AddEditProperty(TQUB_Property_Edit.Create('["OptSkipTime"]',
                    'Optimizer SkipTime', REGTYPE_FLOAT, DEF_OPT_SKIPTIME,
                    txtOptSkipTime,
                    AcceptFloatBetween(MIN_OPT_SKIPTIME, MAX_OPT_SKIPTIME),
                    '') );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Mode"]',
                          'Optimizer Mode', REGTYPE_INT, MacOptModeStringOf(DEF_OPT_MODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptModeOpt,
                                       MacOptModeStringOf(MODE_OPTIMIZE),
                                       Integer(MODE_OPTIMIZE)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptModeCheck,
                                       MacOptModeStringOf(MODE_CHECK),
                                       Integer(MODE_CHECK)))
                       );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["use segments"]',
                          'Optimizer Segments', REGTYPE_INT, MacSegModeStringOf(DEF_OPT_SEGMENTS))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsTogether,
                                       MacSegModeStringOf(OPTSEGS_TOGETHER),
                                       Integer(OPTSEGS_TOGETHER)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsGroup,
                                       MacSegModeStringOf(OPTSEGS_GROUP),
                                       Integer(OPTSEGS_GROUP)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsGroupList,
                                       MacSegModeStringOf(OPTSEGS_GROUPLIST),
                                       Integer(OPTSEGS_GROUPLIST)))
                       );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["SegGroupCount"]',
                      'Optimizer Segments GroupCount', REGTYPE_INT, DEF_OPT_SEGMENTS_GROUPCOUNT,
                      udOptSegmentsGroupCount,
                      txtOptSegmentsGroupCount,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["OptSegGroupList"]',
                    'Optimizer Segments GroupList', REGTYPE_STRING, IntToStr(DEF_OPT_SEGMENTS_GROUPCOUNT),
                    txtOptSegmentsGroupList,
                    '',
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SegMaxBatch"]',
                    'Optimizer Segments MaxBatch', REGTYPE_INT, DEF_OPT_MAXBATCH,
                    txtOptSegmentsMaxBatch,
                    AcceptIntBetween(MIN_OPT_MAXBATCH, MAX_OPT_MAXBATCH),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["UpdateParams"]',
                        'UpdateParams', REGTYPE_BOOL, DEF_UPDATEPARAMS,
                        chkUpdateParams));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["QuietOutput"]',
                        'QuietOutput', REGTYPE_BOOL, DEF_QUIETOUTPUT,
                        chkQuietOutput));

  AddEditProperty(TQUB_Property_Edit.Create('["Ieq"]',
                    'Model Ieq', REGTYPE_FLOAT, DEF_IEQ,
                    txtIeq,
                    AcceptFloatBetween(MIN_IEQ, MAX_IEQ),
                    '') );

  AddEditProperty(TQUB_Property_Edit.Create('["LKCGL"]',
                    'LKCGL', REGTYPE_FLOAT, 1,
                    txtLKCGL,
                    AcceptFloatBetween(MIN_LKC_GL, MAX_LKC_GL),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["LKCVL"]',
                    'LKCVL', REGTYPE_FLOAT, -50,
                    txtLKCVL,
                    AcceptFloatBetween(MIN_LKC_VL, MAX_LKC_VL),
                    '') );

  AddEditProperty(TQUB_Property_Edit.Create('["LKCVChannel"]',
                    'LKCVChannel', REGTYPE_INT, DEF_LKC_VCHANNEL,
                    txtLKCVChannel,
                    AcceptIntBetween(MIN_LKC_VCHANNEL, MAX_LKC_VCHANNEL),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["LKCV0Points"]',
                    'LKCV0Points', REGTYPE_INT, DEF_LKC_V0POINTS,
                    txtLKCV0Points,
                    AcceptIntBetween(MIN_LKC_V0POINTS, MAX_LKC_V0POINTS),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["LKCSkipPoints"]',
                    'LKCSkipPoints', REGTYPE_INT, DEF_LKC_SKIPPOINTS,
                    txtLKCSkipPoints,
                    AcceptIntBetween(MIN_LKC_SKIPPOINTS, MAX_LKC_SKIPPOINTS),
                    '') );

  AddEditProperty(TQUB_Property_Edit.Create('["Sine Ieq"]',
                    'Sine Ieq', REGTYPE_FLOAT, DEF_SINE_IEQ,
                    txtSineIeq,
                    AcceptFloatBetween(MIN_SINE_IEQ, MAX_SINE_IEQ),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Sine Amp"]',
                    'Sine Amp', REGTYPE_FLOAT, DEF_SINE_AMP,
                    txtSineAmp,
                    AcceptFloatBetween(MIN_SINE_AMP, MAX_SINE_AMP),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Sine Freq"]',
                    'Sine Freq', REGTYPE_FLOAT, DEF_SINE_FREQ,
                    txtSineFreq,
                    AcceptFloatBetween(MIN_SINE_FREQ, MAX_SINE_FREQ),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Sine Phase"]',
                    'Sine Phase', REGTYPE_FLOAT, DEF_SINE_PHASE,
                    txtSinePhase,
                    AcceptFloatBetween(MIN_SINE_PHASE, MAX_SINE_PHASE),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Fixed Ieq"]',
                        'Fixed Ieq', REGTYPE_BOOL, DEF_FIXED_IEQ,
                        chkFixedIeq));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Fixed Amp"]',
                        'Fixed Amp', REGTYPE_BOOL, DEF_FIXED_AMP,
                        chkFixedAmp));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Fixed Freq"]',
                        'Fixed Freq', REGTYPE_BOOL, DEF_FIXED_FREQ,
                        chkFixedFreq));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Fixed Phase"]',
                        'Fixed Phase', REGTYPE_BOOL, DEF_FIXED_PHASE,
                        chkFixedPhase));

  RegKeyName            := 'Software\QUB\Task Exp optimizer';
  ProfileCat            := 'Exp';
  prBeforeGetProperties := PropertiesToParams;

  mnuProfiles.SetBounds(330, mnuProfiles.Top, 155, mnuProfiles.Height);

end;

procedure TQUB_Task_Exp_PropertiesForm.FormShow(Sender: TObject);
var
  Model                  :TQUB_QVO_QVM;
begin
  inherited;

  prSetFitDataList;

  prDataList.ShowFiles;

  Model                  := g_CurrModel();
  if Model <> nil then begin
    prDataList.ShowModelVars( Model, True );
  end;

end;

procedure TQUB_Task_Exp_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;

  prDataList.HideAll;
  
  //Properties['Model GridWidth 0'].Data[0] := grdModel.ColWidths[0];
  Properties['Model GridWidth 1'].Data[0] := grdModel.ColWidths[1];
  Properties['Model GridWidth 2'].Data[0] := grdModel.ColWidths[2];
  //Properties['LKC GridWidth 0'].Data[0]   := grdLKCAmpTau.ColWidths[0];
  Properties['LKC GridWidth 1'].Data[0]   := grdLKCAmpTau.ColWidths[1];
  Properties['LKC GridWidth 2'].Data[0]   := grdLKCAmpTau.ColWidths[2];

end;

//******************************************************************************
procedure TQUB_Task_Exp_PropertiesForm.lstFitDataClick(Sender: TObject);
begin
  inherited;
  FitDataListIdx := TComboBox(Sender).ItemIndex;
end;

procedure TQUB_Task_Exp_PropertiesForm.udStateCountClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  StateCount := TUpDown(Sender).Position;
end;

procedure TQUB_Task_Exp_PropertiesForm.udLKCNClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  LKCN := TUpDown(Sender).Position;
end;

//******************************************************************************

procedure TQUB_Task_Exp_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Reg     :TRegistry;
  PropTau :INode;
  PropAmp :INode;
  dVal    :Double;
  i       :Integer;
  C, N    :Integer;
  s       :String;
  procedure SetNodeFromRegOrDefI(Node :INode; RegName :String; Def :Integer);
  var
    iVal  :Integer;
  begin
    if Node.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, RegName, iVal) then begin
        Node.SetData(iVal);
      end else begin
        Node.SetData(Def);
      end;
    end;
  end;
  function CondOp(x :Boolean; a :Integer; b :Integer): Integer;
  begin
    if x then Result := a else Result := b;
  end;

begin
  inherited;

  Reg   := nil; // will be created as needed by ReadXRegistry*
  try
    for i := 0 to 2 do begin
      s    := 'Model GridWidth '+IntToStr(i);
      SetNodeFromRegOrDefI(Props[s], s, CondOp(i=0, 30, 60));
      s    := 'LKC GridWidth '+IntToStr(i);
      SetNodeFromRegOrDefI(Props[s], s, CondOp(i=0, 30, 60));
    end;

    grdModel.ColWidths[0]     := 20;//Properties['Model GridWidth 0'].Data[0];
    grdModel.ColWidths[1]     := Properties['Model GridWidth 1'].Data[0];
    grdModel.ColWidths[2]     := Properties['Model GridWidth 2'].Data[0];
    grdLKCAmpTau.ColWidths[0] := 20;//Properties['LKC GridWidth 0'].Data[0];
    grdLKCAmpTau.ColWidths[1] := Properties['LKC GridWidth 1'].Data[0];
    grdLKCAmpTau.ColWidths[2] := Properties['LKC GridWidth 2'].Data[0];


    //Exp
    SetNodeFromRegOrDefI(Props['StateCount'], 'Model StateCount', 2);
    N := Props['StateCount'].Data[0] - 1;
    PropTau  := Props['Tau'];
    PropAmp  := Props['Amp'];
    C := PropTau.DataCount;
    if C < N then begin
      //TO DO: This crashes - temporary fix - I think it erases old values (bad). LM
      //PropTau.ResizeData(N);
      PropTau.SetupNumData(NODE_DATATYPE_DOUBLE, 1, N);
      for i := C to N - 1 do begin
        if ReadXRegistryFloat(Reg, RegKeyName, 'Model Tau ' + IntToStr(i), dVal) then begin
          PropTau.Data[i] := dVal;
        end else begin
          PropTau.Data[i] := DEF_TAU;
        end;
      end;
    end;
    C := PropAmp.DataCount;
    if C < N then begin
      //TO DO: This crashes - temporary fix - I think it erases old values (bad). LM
      //PropAmp.ResizeData(N);
      PropAmp.SetupNumData(NODE_DATATYPE_DOUBLE, 1, N);
      for i := C to N - 1 do begin
        if ReadXRegistryFloat(Reg, RegKeyName, 'Model Amp ' + IntToStr(i), dVal) then begin
          PropAmp.Data[i] := dVal;
        end else begin
          PropAmp.Data[i] := DEF_AMP;
        end;
      end;
    end;
    for i := 0 to N - 1 do begin
      Tau[i] := PropTau.Data[i];
      Amp[i] := PropAmp.Data[i];
    end;


    //LKC
    SetNodeFromRegOrDefI(Props['LKCN'], 'LKCN', 1);
    N := Props['LKCN'].Data[0];
    PropTau  := Props['LKCTau'];
    PropAmp  := Props['LKCAmp'];
    C := PropTau.DataCount;
    if C < N then begin
      //TO DO: This crashes - temporary fix - I think it erases old values (bad). LM
      //PropTau.ResizeData(N);
      PropTau.SetupNumData(NODE_DATATYPE_DOUBLE, 1, N);
      for i := C to N - 1 do begin
        if ReadXRegistryFloat(Reg, RegKeyName, 'LKCTau ' + IntToStr(i), dVal) then begin
          PropTau.Data[i] := dVal;
        end else begin
          PropTau.Data[i] := DEF_TAU;
        end;
      end;
    end;
    C := PropAmp.DataCount;
    if C < N then begin
      //TO DO: This crashes - temporary fix - I think it erases old values (bad). LM
      //PropAmp.ResizeData(N);
      PropAmp.SetupNumData(NODE_DATATYPE_DOUBLE, 1, N);
      for i := C to N - 1 do begin
        if ReadXRegistryFloat(Reg, RegKeyName, 'LKCAmp ' + IntToStr(i), dVal) then begin
          PropAmp.Data[i] := dVal;
        end else begin
          PropAmp.Data[i] := DEF_AMP;
        end;
      end;
    end;
    for i := 0 to N - 1 do begin
      LKCTau[i] := PropTau.Data[i];
      LKCAmp[i] := PropAmp.Data[i];
    end;

    //Method
    SetNodeFromRegOrDefI(Props['Method'], 'Method', Integer(METHOD_EXP));

  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_Task_Exp_PropertiesForm.prUpdateComponents;
var
  i        :Integer;
  PropTau  :INode;
  PropAmp  :INode;
begin
  inherited;

  StateCount         := Properties['StateCount'].Data[0];
  PropTau            := Properties['Tau'];
  PropAmp            := Properties['Amp'];
  for i := 0 to StateCount - 2 do begin
    Tau[i]           := PropTau.Data[i];
    Amp[i]           := PropAmp.Data[i];
  end;

  LKCN               := Properties['LKCN'].Data[0];
  PropTau            := Properties['LKCTau'];
  PropAmp            := Properties['LKCAmp'];
  for i := 0 to LKCN - 1 do begin
    LKCTau[i]        := PropTau.Data[i];
    LKCAmp[i]        := PropAmp.Data[i];
  end;

  //TO DO: LM. Add Sine properties
  //. . .

  i := Properties['Method'].Data[0];
  pageMethod.ActivePageIndex := i;
end;

procedure TQUB_Task_Exp_PropertiesForm.PropertiesToParams;
var
  i        :Integer;
  PropTau  :INode;
  PropAmp  :INode;
begin
  prParams.OptSSConv                 := Properties['DFP']['ConvSS'].Data[0];
  prParams.OptGradConv               := Properties['DFP']['ConvGrad'].Data[0];
  prParams.OptMaxStep                := Properties['DFP']['MaxStep'].Data[0];
  prParams.OptMaxIter                := Properties['DFP']['MaxIterations'].Data[0];
  prParams.OptRestart                := Properties['DFP']['Restart'].Data[0];

  prParams.OptMode                   := MacOptModeOfString( Properties['Mode'].DataAsString );
  prParams.OptSegments               := MacSegModeOfString( Properties['use segments'].DataAsString );

  prParams.OptFitDataMode            := MacFitDataModeOfString( Properties['OptFitDataMode'].DataAsString );
  prParams.OptShowFit                := Properties['ShowFit'].Data[0];
  prParams.OptShowFitWhat            := Properties['ShowFitWhat'].Data[0];
  prParams.OptSkip                   := Properties['OptSkip'].Data[0];
  prParams.OptSkipTime               := Properties['OptSkipTime'].Data[0];
  prParams.OptSegmentsGroupCount     := Properties['SegGroupCount'].Data[0];
  prParams.OptSegmentsMaxBatch       := Properties['SegMaxBatch'].Data[0];
  prParams.OptIdenticalChunks        := Properties['IdenticalChunks'].Data[0];
  prParams.OptMUXFiles               := Properties['MUXFiles'].Data[0];
  prParams.OptPropagateParams        := Properties['PropagateParams'].Data[0];

  prUpdateParams                     := 0<>Properties['UpdateParams'].Data[0];
  prQuietOutput                      := 0<>Properties['QuietOutput'].Data[0];

  prParams.Ieq                       := Properties['Ieq'].Data[0];

  prParams.LKCGL                     := Properties['LKCGL'].Data[0];
  prParams.LKCVL                     := Properties['LKCVL'].Data[0];
  prParams.LKCVChannel               := Properties['LKCVChannel'].Data[0];
  prParams.LKCV0Points               := Properties['LKCV0Points'].Data[0];
  prParams.LKCSkipPoints             := Properties['LKCSkipPoints'].Data[0];

  prParams.SineIeq                   := Properties['Sine Ieq'].Data[0];
  prParams.SineAmp                   := Properties['Sine Amp'].Data[0];
  prParams.SineFreq                  := Properties['Sine Freq'].Data[0];
  prParams.SinePhase                 := Properties['Sine Phase'].Data[0];
  prParams.FixedIeq                  := 0<>Properties['Fixed Ieq'].Data[0];
  prParams.FixedAmp                  := 0<>Properties['Fixed Amp'].Data[0];
  prParams.FixedFreq                 := 0<>Properties['Fixed Freq'].Data[0];
  prParams.FixedPhase                := 0<>Properties['Fixed Phase'].Data[0];

  StateCount                         := Properties['StateCount'].Data[0];
  PropTau                            := Properties['Tau'];
  PropAmp                            := Properties['Amp'];
  for i := 0 to StateCount - 2 do begin
    Tau[i]                           := PropTau.Data[i];
    Amp[i]                           := PropAmp.Data[i];
  end;

  LKCN                               := Properties['LKCN'].Data[0];
  PropTau                            := Properties['LKCTau'];
  PropAmp                            := Properties['LKCAmp'];
  for i := 0 to LKCN - 1 do begin
    LKCTau[i]                        := PropTau.Data[i];
    LKCAmp[i]                        := PropAmp.Data[i];
  end;

  Properties['Method'].Data[0]       := pageMethod.ActivePageIndex;

end;


//******************************************************************************
end.
