unit QUB_Task_ExtractSelection_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

AllocTypes,
RegistryInterface,

QUB_DataList_Form,
QUB_Global,
QUB_QFS_File, QUB_QFS_QDF, QUB_QFS_QDF_Global,
QUB_Node,
QUB_Task,
QUB_Task_Data,
QUB_Properties,
QUB_Task_Properties;


{$include 'QUB_Task_ExtractSelection_Properties.hct'}


type


TQUB_Task_ExtractSelection_Operation = (OPER_NONE, OPER_ADD, OPER_SUB, OPER_AVG, OPER_STD, OPER_SCALE);
TQUB_Task_ExtractSelection_Channels  = (CHANNELS_ALL, CHANNELS_ACTIVE, CHANNELS_LIST);


function OperationToString(x :TQUB_Task_ExtractSelection_Operation): String;
function StringToOperation(x :String): TQUB_Task_ExtractSelection_Operation;


const


DEF_OPENFILEAFTER             = 1;

//Data Source
DEF_DATA_SOURCE               = DATA_SOURCE_SELECTIONLIST;

//Data Processed
DEF_DATA_PROCESSED            = DATA_PROCESSED_ASSUCH;

DEF_PROCESSFILTER             = 0;
DEF_PROCESSFILTERFREQ         :Double = 5000.0;  //Hz
MIN_PROCESSFILTERFREQ         :Double = 100.0;   //Hz
MAX_PROCESSFILTERFREQ         :Double = 100000.0;//Hz

DEF_PROCESSBASELINE           = 0;

DEF_PROCESSRESAMPLE           = 0;
DEF_PROCESSRESAMPLEFACTOR     :Extended = 2;
MIN_PROCESSRESAMPLEFACTOR     :Extended = 1;
MAX_PROCESSRESAMPLEFACTOR     :Extended = 100;
DEF_PROCESSRESAMPLEFREQ       :Extended = 1000; // whatever i think it's ignored

//File Format
DEF_FILEFORMAT                = QFS_QDF_TYPE_QDF;//QFS_QDF_TYPE_LDT;
DEF_ASCIITIME                 = 1;
DEF_ASCIIFIT                  = 0;
DEF_ASCIISEGCOL               = 0;
DEF_FLOAT4BYTE                = 0;

//Other Options
DEF_CHANNELS                  = CHANNELS_ALL;
DEF_PROC_ONLY_ACTIVE_CHANNEL  = 0;
DEF_JOINSEGMENTS              = 0;
DEF_DROPFIRSTLAST             = 0;
DEF_DROPIFCLASS               = 1;
DEF_DROPCLASS                 = 0;

//KinPars
DEF_KINPARS                   = 0;
DEF_KINPARSALL                = 0;
DEF_KINPARSLIST               = '';

//Advanced Options
DEF_ADVANCED                  = 0;

//Function
DEF_FUNCTION_DRIFT            = 0;
DEF_FUNCTION_DRIFTA           :Double = 1;
DEF_FUNCTION_DRIFTB           :Double = 0;

DEF_FUNCTION_SINE             = 0;
DEF_FUNCTION_SINEFRQ          :Double = 60;
DEF_FUNCTION_SINEAMP          :Double = 1;
DEF_FUNCTION_SINEPH           :Double = 0;

DEF_FUNCTION_DIGITIZE         = 0;
DEF_FUNCTION_DIGITIZE_THRESHOLD = 0;
DEF_FUNCTION_DIGITIZE_TOMIN   :Double = 0;
DEF_FUNCTION_DIGITIZE_TOMAX   :Double = 1;

DEF_FUNCTION_DERIVATE         = 0;

DEF_FUNCTION_LKC              = 0;

DEF_FUNCTION_ITOG             = 0;
DEF_FUNCTION_ITOG_VREV        :Double  = 0;
DEF_FUNCTION_ITOG_VCH         :Integer = 0;

DEF_FUNCTION_RSCORR           = 0;
DEF_FUNCTION_RSVALUE          :Double  = 0.0;
DEF_FUNCTION_RSPER            :Double  = 75.0;
DEF_FUNCTION_RSADDLEAK        = 0;
DEF_FUNCTION_RSLEAKCH         :Integer = 2;

DEF_FUNCTION_PNADD            = 0;
DEF_FUNCTION_PNSUB            = 0;
DEF_FUNCTION_PNOFFSET         :Double = -80;
DEF_FUNCTION_PNDELAY          :Double = 100;
DEF_FUNCTION_PNSCALE          :Double = 0.25;
DEF_FUNCTION_PNCOUNT          :Integer = 4;
DEF_FUNCTION_PNVM             :Double = -80;
DEF_FUNCTION_PNINTERVAL       :Double = 100;

DEF_FUNCTION_FIT              = 0;

DEF_OPERATION                 = OPER_NONE;
DEF_OPERATION_SCALE           = 0;
DEF_OPERATION_SCALEONEOVER    = 0;
MAX_OPERATION_SCALEFACTOR     :Double = 1e+3;
MIN_OPERATION_SCALEFACTOR     :Double = 1e-3;
DEF_OPERATION_SCALEFACTOR     :Double = 1;

DEF_OPERATION_LIST            = 0;
DEF_OPERATION_FILES           = 0;


type


TQUB_Task_ExtractSelection_PropertiesForm = class(TQUB_Task_PropertiesForm)


  Label1                                  :TLabel;
  Label2                                  :TLabel;
  Label3                                  :TLabel;
  Label4                                  :TLabel;
  Label5                                  :TLabel;
    lblProcessResampleFactor: TLabel;
    lblFunctionDriftA: TLabel;
    lblFunctionDriftB: TLabel;
    lblFunctionSineFrq: TLabel;
    lblFunctionSineAmp: TLabel;
    lblFunctionSinePh: TLabel;
    lblProcessResampleFreq: TLabel;

  lblFileName                             :TLabel;

  panFormat                               :TPanel;

  txtFileName                             :TEdit;
  chkOpenFileAfter                        :TCheckBox;

  optFileFormat0                          :TRadioButton;
  optFileFormat1                          :TRadioButton;
  optFileFormat2                          :TRadioButton;
  optFileFormat3                          :TRadioButton;
  optFileFormat5                          :TRadioButton;
  optFileFormat6                          :TRadioButton;

  chkASCIITime                            :TCheckBox;

  btnBrowseFileName                       :TButton;

  panOperation                            :TPanel;
  chkOperationNone                        :TRadioButton;
  chkOperationAdd                         :TRadioButton;
  chkOperationSub                         :TRadioButton;
  chkOperationAvg                         :TRadioButton;
  chkOperationStd                         :TRadioButton;
  chkOperationList                        :TCheckBox;
  lstOperationList                        :TComboBox;
  chkOperationFiles                       :TCheckBox;

  panFunction                             :TPanel;
  chkFunctionDrift                        :TCheckBox;
  txtFunctionDriftA                       :TEdit;
  txtFunctionDriftB                       :TEdit;
  chkFunctionSine                         :TCheckBox;
  txtFunctionSineFrq                      :TEdit;
  txtFunctionSineAmp                      :TEdit;
  txtFunctionSinePh                       :TEdit;
  chkFunctionDerivate                     :TCheckBox;

  btnAdvanced                             :TButton;
  panDataSource                           :TPanel;
  chkDataSourceWholeFile                  :TRadioButton;
  chkDataSourceSelectionList              :TRadioButton;
  chkDataSourceSelection                  :TRadioButton;
  panDataProcessed                        :TPanel;
  chkDataProcessedAsSuch                  :TRadioButton;
  chkDataProcessedAsDisplayed             :TRadioButton;
  chkProcessFilter                        :TCheckBox;
  txtProcessFilterFreq                    :TEdit;
  chkProcessBaseline                      :TCheckBox;
  chkDataProcessedNone                    :TRadioButton;
  chkProcessResample                      :TCheckBox;
  txtProcessResampleFactor                :TEdit;
  txtProcessResampleFreq                  :TEdit;
  chkFunctionDigitize                     :TCheckBox;
    lblFunctionDigitizeThreshold: TLabel;
  txtFunctionDigitizeThreshold            :TEdit;
    lblFunctionDigitizeToMin: TLabel;
  txtFunctionDigitizeToMin                :TEdit;
    lblFunctionDigitizeToMax: TLabel;
  txtFunctionDigitizeToMax                :TEdit;
  chkProcOnlyActiveChannel                :TCheckBox;
  chkFloat4Byte                           :TCheckBox;
    Panel3: TPanel;
    chkJoinSegments: TCheckBox;
    Label29: TLabel;
    chkAllChannels: TRadioButton;
    chkActiveChannel: TRadioButton;
    txtChannelList: TEdit;
    chkListedChannels: TRadioButton;
    Panel1: TPanel;
    lblFunctionPnOffset: TLabel;
    lblFunctionPnCount: TLabel;
    Label21: TLabel;
    chkFunctionPnAdd: TCheckBox;
    txtFunctionPnOffset: TEdit;
    chkFunctionPnSub: TCheckBox;
    lblFunctionPnDelay: TLabel;
    txtFunctionPnDelay: TEdit;
    txtFunctionPnCount: TEdit;
    lblFunctionPnScale: TLabel;
    txtFunctionPnScale: TEdit;
    lblFunctionPnVm: TLabel;
    txtFunctionPnVm: TEdit;
    lblFunctionPnInterval: TLabel;
    txtFunctionPnInterval: TEdit;
    Panel2: TPanel;
    Label28: TLabel;
    chkFunctionLKC: TCheckBox;
    chkFunctionFit: TCheckBox;
    txtFunctionItoGVCh: TEdit;
    txtFunctionRsLeakCh: TEdit;
    chkFunctionRsAddLeak: TCheckBox;
    lblFunctionItoGVCh: TLabel;
    txtFunctionItoGVRev: TEdit;
    lblFunctionItoGVRev: TLabel;
    lblFunctionRsCorrPer: TLabel;
    txtFunctionRsCorrPer: TEdit;
    txtFunctionRsCorr: TEdit;
    chkFunctionRsCorr: TCheckBox;
    chkFunctionItoG: TCheckBox;
    chkDropFirstLast: TCheckBox;
    chkDropIfClass: TCheckBox;
    txtDropClass: TEdit;
    udDropClass: TUpDown;
    chkASCIISegCol: TCheckBox;
    chkKinPars: TCheckBox;
    chkKinParsAll: TCheckBox;
    txtKinParsList: TEdit;
    chkOperationScale: TCheckBox;
    chkOperationScaleOneOver: TCheckBox;
    txtOperationScaleFactor: TEdit;
    chkASCIIFit: TCheckBox;
    chkFileList: TCheckBox;
    btnFileList: TButton;
    lblFileListStatus: TLabel;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure btnBrowseFileNameClick(Sender: TObject);
  procedure btnOKClick(Sender: TObject);

  //Data Source
  procedure chkDataSourceClick(Sender: TObject);

  //File Format
  procedure optFileFormatClick(Sender: TObject);
  procedure txtFileNameKeyPress(Sender: TObject; var Key: Char);
  procedure txtFileNameExit(Sender :TObject);

  procedure lstOperationListClick(Sender: TObject);

  //Advanced Options
  procedure btnAdvancedClick(Sender: TObject);

  procedure txtProcessResampleFreqExit(Sender: TObject);
  procedure txtProcessResampleFactorExit(Sender: TObject);

  procedure prSetEnabledOnClick(Sender: TObject);

  procedure chkFunctionPnAddClick(Sender: TObject);
  procedure chkFunctionPnSubClick(Sender: TObject);
    procedure btnFileListClick(Sender: TObject);


protected


  prSourceADChannelCount                  :Integer;

  prSamplingFrequency                     :Extended;

  //Data Source
  prDataSource                            :TQUB_Task_Data_Source;
  //File Format
  prFileFormat                            :TQUB_QFS_QDF_FileType;

  prFileName                              :String;

  //Advanced options
  prAdvanced                              :Boolean;

  prOperationListName                     :String;

  formDataList                            :TQUB_DataList_Form;


  procedure prSetFileName(f :String);

  //Data Source
  procedure prSetDataSource(s :TQUB_Task_Data_Source);

  //File format
  procedure prSetFileFormat(f :TQUB_QFS_QDF_FileType);
  function prGetFileFormatString: String;
  procedure prSetFileFormatString(x :String);

  //Advanced options
  procedure prSetAdvanced(a :Boolean);

  procedure prSetEnabled;
  procedure prCountFileList;

  procedure prSetOperationListName(n :String);

  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public


  property SourceADChannelCount           :Integer                              read prSourceADChannelCount       write prSourceADChannelCount;

  property FileName                       :String                               read prFileName                   write prSetFileName;

  property SamplingFrequency              :Extended                                                               write prSamplingFrequency;

  //Data Source
  property DataSource                     :TQUB_Task_Data_Source                read prDataSource                 write prSetDataSource;

  //File format
  property FileFormat                     :TQUB_QFS_QDF_FileType                read prFileFormat                 write prSetFileFormat;
  property FileFormatString               :String                               read prGetFileFormatString        write prSetFileFormatString;

  //Advanced options
  property Advanced                       :Boolean                              read prAdvanced                   write prSetAdvanced;

  property OperationListName              :String                               read prOperationListName          write prSetOperationListName;


end;


implementation


{$R *.DFM}


uses


PythonEngine,
QUB_Python_Thread,
QUB_Viewport_Python,
QUB_Node_Python,
QUB_QVO_QVD,
QUB_GUIInterface;


//******************************************************************************
function OperationToString(x :TQUB_Task_ExtractSelection_Operation): String;
begin
  if          x = OPER_NONE  then begin
    Result := 'None';
  end else if x = OPER_ADD   then begin
    Result := 'Add';
  end else if x = OPER_SUB   then begin
    Result := 'Subtract';
  end else if x = OPER_AVG   then begin
    Result := 'Average';
  end else if x = OPER_STD   then begin
    Result := 'Std';
  end else begin // if Operation = OPER_SCALE then begin
    Result := 'Scale';
  end;
end;

function StringToOperation(x :String): TQUB_Task_ExtractSelection_Operation;
begin
  if          x = 'None'     then begin
    Result := OPER_NONE;
  end else if x = 'Add'      then begin
    Result := OPER_ADD;
  end else if x = 'Subtract' then begin
    Result := OPER_SUB;
  end else if x = 'Average'  then begin
    Result := OPER_AVG;
  end else if x = 'Std'      then begin
    Result := OPER_STD;
  end else {if x = 'Scale'    then} begin
    Result := OPER_SCALE;
  end;
end;


//******************************************************************************
procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetFileName(f :String);
begin
  prFileName                         := ChangeFileExt(f, '');
  txtFileName.Text                   := prFileName;
end;

//File format
procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetFileFormat(f :TQUB_QFS_QDF_FileType);
begin

  prFileFormat := f;
  Properties['FileFormat'].SetData(FileFormatString);

  case prFileFormat of
    QFS_QDF_TYPE_QDF: optFileFormat0.Checked := True;
    QFS_QDF_TYPE_LDT: optFileFormat1.Checked := True;
    QFS_QDF_TYPE_DAT: optFileFormat2.Checked := True;
    QFS_QDF_TYPE_ASC: optFileFormat3.Checked := True;
    QFS_QDF_TYPE_WAV: optFileFormat5.Checked := True;
    QFS_QDF_TYPE_DWT: optFileFormat6.Checked := True;
  end;

  prSetEnabled;

end;

function TQUB_Task_ExtractSelection_PropertiesForm.prGetFileFormatString: String;
begin
  if FileFormat = QFS_QDF_TYPE_QDF then begin
    Result                           := 'QDF';
  end else if FileFormat = QFS_QDF_TYPE_LDT then begin
    Result                           := 'LDT';
  end else if FileFormat = QFS_QDF_TYPE_DAT then begin
    Result                           := 'DAT';
  end else if FileFormat = QFS_QDF_TYPE_ASC then begin
    Result                           := 'ASCII';
  end else if FileFormat = QFS_QDF_TYPE_WAV then begin
    Result                           := 'WAV';
  end else begin // if FileFormat = QFS_QDF_TYPE_DWT then begin
    Result                           := 'DWT';
  end;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetFileFormatString(x :String);
begin
  if x = 'QDF' then begin
    FileFormat                       := QFS_QDF_TYPE_QDF;
  end else if x = 'LDT' then begin
    FileFormat                       := QFS_QDF_TYPE_LDT;
  end else if x = 'DAT' then begin
    FileFormat                       := QFS_QDF_TYPE_DAT;
  end else if x = 'ASC' then begin
    FileFormat                       := QFS_QDF_TYPE_ASC;
  end else if x = 'WAV' then begin
    FileFormat                       := QFS_QDF_TYPE_WAV;
  end else if x = 'DWT' then begin
    FileFormat                       := QFS_QDF_TYPE_DWT;
  end;
end;


//Data Source
procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetDataSource(s :TQUB_Task_Data_Source);
begin
  prDataSource                       := s;
  chkDataSourceWholeFile.Checked     := DataSource = DATA_SOURCE_WHOLEFILE;
  chkDataSourceSelectionList.Checked := DataSource = DATA_SOURCE_SELECTIONLIST;
  chkDataSourceSelection.Checked     := DataSource = DATA_SOURCE_SELECTION;
end;



//Advanced options
procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetAdvanced(a :Boolean);
begin
  //Make it always advanced
  prAdvanced                          := a;
  Properties['Advanced'].Data[0]      := Integer(prAdvanced);

  {
  panOperation.Visible                := prAdvanced;
  panFunction.Visible                 := prAdvanced;

  if Advanced then begin
    btnAdvanced.Caption := '&Simple';
    ClientWidth := panFunction.Left + panFunction.Width + btnOK.Left;
  end else begin
    btnAdvanced.Caption := '&Advanced';
    ClientWidth := btnBrowseFileName.Left + btnBrowseFileName.Width + btnOK.Left;
  end;
  }

end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetEnabled;
begin

  //Allow including the time column only for ASCII files
  chkASCIITime.Enabled                := FileFormat = QFS_QDF_TYPE_ASC;
  chkASCIISegCol.Enabled              := FileFormat = QFS_QDF_TYPE_ASC;

  //Not for these two formats
  chkASCIIFit.Enabled                 := (FileFormat <> QFS_QDF_TYPE_WAV) and (FileFormat <> QFS_QDF_TYPE_DWT);

  //Only for DAT - must add other formats
  chkFloat4Byte.Enabled               := FileFormat = QFS_QDF_TYPE_DAT;

  //Process Data
  //Not for DWT
  chkDataProcessedNone.Enabled        := FileFormat <> QFS_QDF_TYPE_DWT;
  chkDataProcessedAsDisplayed.Enabled := chkDataProcessedNone.Enabled;
  chkDataProcessedAsSuch.Enabled      := chkDataProcessedNone.Enabled;
  chkProcessFilter.Enabled            := chkDataProcessedAsSuch.Enabled and chkDataProcessedAsSuch.Checked;
  txtProcessFilterFreq.Enabled        := chkDataProcessedAsSuch.Enabled and chkProcessFilter.Checked;
  chkProcessBaseline.Enabled          := chkDataProcessedAsSuch.Enabled and chkDataProcessedAsSuch.Checked;
  chkProcessResample.Enabled          := chkDataProcessedAsSuch.Enabled and chkDataProcessedAsSuch.Checked;
  lblProcessResampleFactor.Enabled    := chkDataProcessedAsSuch.Enabled and chkDataProcessedAsSuch.Checked and chkProcessResample.Checked;
  txtProcessResampleFactor.Enabled    := chkDataProcessedAsSuch.Enabled and chkDataProcessedAsSuch.Checked and chkProcessResample.Checked;
  lblProcessResampleFreq.Enabled      := chkDataProcessedAsSuch.Enabled and chkDataProcessedAsSuch.Checked and chkProcessResample.Checked;
  txtProcessResampleFreq.Enabled      := chkDataProcessedAsSuch.Enabled and chkDataProcessedAsSuch.Checked and chkProcessResample.Checked;
  chkDataProcessedAsSuch.Enabled      := chkDataProcessedNone.Enabled;
  chkProcOnlyActiveChannel.Enabled    := FileFormat <> QFS_QDF_TYPE_DWT;

  //Channels
  //Not for DWT
  chkAllChannels.Enabled              := FileFormat <> QFS_QDF_TYPE_DWT;
  chkActiveChannel.Enabled            := chkAllChannels.Enabled;
  chkListedChannels.Enabled           := chkAllChannels.Enabled;
  txtChannelList.Enabled              := chkAllChannels.Enabled and chkListedChannels.Checked;

  //DAT is the only non-segmented format
  chkJoinSegments.Enabled             := FileFormat <> QFS_QDF_TYPE_DAT;

  chkKinPars.Enabled                  := FileFormat <> QFS_QDF_TYPE_DWT;
  chkKinParsAll.Enabled               := chkKinPars.Enabled and chkKinPars.Checked;
  txtKinParsList.Enabled              := chkKinPars.Enabled and chkKinPars.Checked and (not chkKinParsAll.Checked);

  //Only for DWT
  chkDropFirstLast.Enabled            := FileFormat = QFS_QDF_TYPE_DWT;
  chkDropIfClass.Enabled              := chkDropFirstLast.Enabled and chkDropFirstLast.Checked;
  txtDropClass.Enabled                := chkDropFirstLast.Enabled and chkDropFirstLast.Checked;
  udDropClass.Enabled                 := chkDropFirstLast.Enabled and chkDropFirstLast.Checked;

  //Not for DWT
  chkOperationNone.Enabled            := FileFormat <> QFS_QDF_TYPE_DWT;
  chkOperationAdd.Enabled             := chkOperationNone.Enabled;
  chkOperationSub.Enabled             := chkOperationNone.Enabled;
  chkOperationAvg.Enabled             := chkOperationNone.Enabled;
  chkOperationStd.Enabled             := chkOperationNone.Enabled;
  chkOperationScale.Enabled           := chkOperationNone.Enabled;
  chkOperationScaleOneOver.Enabled    := chkOperationScale.Enabled and chkOperationScale.Checked;
  txtOperationScaleFactor.Enabled     := chkOperationScale.Enabled and chkOperationScale.Checked;
  chkOperationList.Enabled            := chkOperationNone.Enabled;
  lstOperationList.Enabled            := chkOperationNone.Enabled;
  chkOperationFiles.Enabled           := chkOperationNone.Enabled;

  //No functions for DWT
  chkFunctionDrift.Enabled            := FileFormat <> QFS_QDF_TYPE_DWT;
  txtFunctionDriftA.Enabled           := chkFunctionDrift.Enabled and chkFunctionDrift.Checked;
  lblFunctionDriftA.Enabled           := chkFunctionDrift.Enabled and chkFunctionDrift.Checked;
  txtFunctionDriftB.Enabled           := chkFunctionDrift.Enabled and chkFunctionDrift.Checked;
  lblFunctionDriftB.Enabled           := chkFunctionDrift.Enabled and chkFunctionDrift.Checked;
  chkFunctionSine.Enabled             := FileFormat <> QFS_QDF_TYPE_DWT;
  txtFunctionSineFrq.Enabled          := chkFunctionSine.Enabled and chkFunctionSine.Checked;
  lblFunctionSineFrq.Enabled          := chkFunctionSine.Enabled and chkFunctionSine.Checked;
  txtFunctionSineAmp.Enabled          := chkFunctionSine.Enabled and chkFunctionSine.Checked;
  lblFunctionSineAmp.Enabled          := chkFunctionSine.Enabled and chkFunctionSine.Checked;
  txtFunctionSinePh.Enabled           := chkFunctionSine.Enabled and chkFunctionSine.Checked;
  lblFunctionSinePh.Enabled           := chkFunctionSine.Enabled and chkFunctionSine.Checked;
  chkFunctionDigitize.Enabled         := FileFormat <> QFS_QDF_TYPE_DWT;
  txtFunctionDigitizeThreshold.Enabled:= chkFunctionDigitize.Enabled and chkFunctionDigitize.Checked;
  lblFunctionDigitizeThreshold.Enabled:= chkFunctionDigitize.Enabled and chkFunctionDigitize.Checked;
  txtFunctionDigitizeToMin.Enabled    := chkFunctionDigitize.Enabled and chkFunctionDigitize.Checked;
  lblFunctionDigitizeToMin.Enabled    := chkFunctionDigitize.Enabled and chkFunctionDigitize.Checked;
  txtFunctionDigitizeToMax.Enabled    := chkFunctionDigitize.Enabled and chkFunctionDigitize.Checked;
  lblFunctionDigitizeToMax.Enabled    := chkFunctionDigitize.Enabled and chkFunctionDigitize.Checked;
  chkFunctionDerivate.Enabled         := FileFormat <> QFS_QDF_TYPE_DWT;
  chkFunctionLKC.Enabled              := FileFormat <> QFS_QDF_TYPE_DWT;
  chkFunctionItoG.Enabled             := FileFormat <> QFS_QDF_TYPE_DWT;
  txtFunctionItoGVRev.Enabled         := chkFunctionItoG.Enabled and chkFunctionItoG.Checked;
  lblFunctionItoGVRev.Enabled         := chkFunctionItoG.Enabled and chkFunctionItoG.Checked;
  txtFunctionItoGVCh.Enabled          := chkFunctionItoG.Enabled and chkFunctionItoG.Checked;
  lblFunctionItoGVCh.Enabled          := chkFunctionItoG.Enabled and chkFunctionItoG.Checked;
  chkFunctionRsCorr.Enabled           := FileFormat <> QFS_QDF_TYPE_DWT;
  txtFunctionRsCorr.Enabled           := chkFunctionRsCorr.Enabled and chkFunctionRsCorr.Checked;
  txtFunctionRsCorrPer.Enabled        := chkFunctionRsCorr.Enabled and chkFunctionRsCorr.Checked;
  lblFunctionRsCorrPer.Enabled        := chkFunctionRsCorr.Enabled and chkFunctionRsCorr.Checked;
  chkFunctionRsAddLeak.Enabled        := chkFunctionRsCorr.Enabled and chkFunctionRsCorr.Checked;
  txtFunctionRsLeakCh.Enabled         := chkFunctionRsCorr.Enabled and chkFunctionRsCorr.Checked;
  chkFunctionFit.Enabled              := FileFormat <> QFS_QDF_TYPE_DWT;
  chkFunctionPnAdd.Enabled            := FileFormat <> QFS_QDF_TYPE_DWT;
  chkFunctionPnSub.Enabled            := FileFormat <> QFS_QDF_TYPE_DWT;
  txtFunctionPnOffset.Enabled         := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  lblFunctionPnOffset.Enabled         := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  txtFunctionPnCount.Enabled          := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  lblFunctionPnCount.Enabled          := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  txtFunctionPnDelay.Enabled          := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  lblFunctionPnDelay.Enabled          := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  txtFunctionPnScale.Enabled          := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  lblFunctionPnScale.Enabled          := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  txtFunctionPnVm.Enabled             := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  lblFunctionPnVm.Enabled             := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  txtFunctionPnInterval.Enabled       := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);
  lblFunctionPnInterval.Enabled       := chkFunctionPnSub.Enabled and (chkFunctionPnAdd.Checked or chkFunctionPnSub.Checked);


end;


procedure TQUB_Task_ExtractSelection_PropertiesForm.prCountFileList;
var
  i          :Integer;
  n          :Integer;
begin
  n  := 0;
  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    if TQUB_QVO_QVD(ViewportData.ListObjects[i]).Listed then Inc(n);
  end;
  if n = 0 then begin
    lblFileListStatus.Caption := '(no files chosen)'
  end else if n = 1 then begin
    lblFileListStatus.Caption := '(1 file chosen)';
  end else begin
    lblFileListStatus.Caption := Format('(%d files chosen)', [n]);
  end;
end;

//******************************************************************************
procedure TQUB_Task_ExtractSelection_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  formDataList             := TQUB_DataList_Form.Create(Self);

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ExtractFileList"]',
                         '', REGTYPE_NONE, 0, chkFileList)
                      );

  // copied from task_data_properties;
  // added resample and proc-only-active-channel
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["ProcessData"]',
                          'Data Processed', REGTYPE_INT, DataProcessedToString(DEF_DATA_PROCESSED))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkDataProcessedNone,
                                       DataProcessedToString(DATA_PROCESSED_NONE),
                                       Integer(DATA_PROCESSED_NONE)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkDataProcessedAsDisplayed,
                                       DataProcessedToString(DATA_PROCESSED_ASDISPLAYED),
                                       Integer(DATA_PROCESSED_ASDISPLAYED)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkDataProcessedAsSuch,
                                       DataProcessedToString(DATA_PROCESSED_ASSUCH),
                                       Integer(DATA_PROCESSED_ASSUCH))
                               .Enables(chkProcessFilter)
                               .Enables(chkProcessBaseline)
                               .Enables(chkProcessResample)
                             )
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ProcessData"]["Filter"]',
                        'Process Filter', REGTYPE_BOOL, DEF_PROCESSFILTER,
                        chkProcessFilter)
                       .Enables(txtProcessFilterFreq) );
  AddEditProperty(TQUB_Property_Edit.Create('["ProcessData"]["FilterFreq"]',
                    'Process Filter Freq', REGTYPE_FLOAT, DEF_PROCESSFILTERFREQ,
                    txtProcessFilterFreq,
                    'lambda x: acceptFloatBetween('+FloatToStr(MIN_PROCESSFILTERFREQ/1000)+', '+FloatToStr(MAX_PROCESSFILTERFREQ/1000)+')(x) * 1000.0',
                    'lambda node, ix: str(node.data[ix] / 1000.0)') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ProcessData"]["Baseline"]',
                        'Process Baseline', REGTYPE_BOOL, DEF_PROCESSBASELINE,
                        chkProcessBaseline));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ProcessData"]["Resample"]',
                        'Process Resample', REGTYPE_BOOL, DEF_PROCESSRESAMPLE,
                        chkProcessResample)
                       .Enables(txtProcessResampleFactor)
                       .Enables(txtProcessResampleFreq) );
  AddEditProperty(TQUB_Property_Edit.Create('["ProcessData"]["ResampleFactor"]',
                    'Process Resample Factor', REGTYPE_STRINGFLOAT, DEF_PROCESSRESAMPLEFACTOR,
                    txtProcessResampleFactor,
                    AcceptFloatBetween(MIN_PROCESSRESAMPLEFACTOR, MAX_PROCESSRESAMPLEFACTOR),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["ProcessData"]["ResampleFreq"]',
                    'Process Resample Freq', REGTYPE_NONE, DEF_PROCESSRESAMPLEFREQ,
                    txtProcessResampleFreq,
                    'lambda x: acceptFloatGreaterThan(0.0)(x) * 1000.0',
                    'lambda node, ix: str(node.data[ix] / 1000.0)') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ProcOnlyActiveChannel"]',
                        'Process Only Active Channel', REGTYPE_BOOL, DEF_PROC_ONLY_ACTIVE_CHANNEL,
                        chkProcOnlyActiveChannel));

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["OpenFileAfter"]',
                        'Open File After', REGTYPE_BOOL, DEF_OPENFILEAFTER,
                        chkOpenFileAfter));

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ASCIIHasTime"]',
                        'ASCII Time', REGTYPE_BOOL, DEF_ASCIITIME,
                        chkASCIITime));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ASCIIHasFit"]',
                        'ASCII Fit', REGTYPE_BOOL, DEF_ASCIIFIT,
                        chkASCIIFit));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ASCIISegCol"]',
                        '', REGTYPE_NONE, DEF_ASCIISEGCOL,
                        chkASCIISegCol));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Float4Byte"]',
                        'Float 4 Byte', REGTYPE_BOOL, DEF_FLOAT4BYTE,
                        chkFloat4Byte));

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Channels"]',
                          'Channels', REGTYPE_INT, Integer(DEF_CHANNELS))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkAllChannels,
                                       Integer(CHANNELS_ALL),
                                       Integer(CHANNELS_ALL)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkActiveChannel,
                                       Integer(CHANNELS_ACTIVE),
                                       Integer(CHANNELS_ACTIVE)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkListedChannels,
                                       Integer(CHANNELS_LIST),
                                       Integer(CHANNELS_LIST))
                               .Enables(txtChannelList)
                             )
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["ChannelList"]',
                          '', REGTYPE_NONE, 0,
                          txtChannelList,
                          AcceptIntList(AcceptIntGreaterThan(-1)),
                          'lambda node, ix: SpacedNodeData(node)') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["JoinSegments"]',
                        'Collate Segments', REGTYPE_BOOL, DEF_JOINSEGMENTS,
                        chkJoinSegments));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DropFirstLast"]',
                        'Drop First Last', REGTYPE_BOOL, DEF_DROPFIRSTLAST,
                        chkDropFirstLast)
                       .Enables(chkDropIfClass)
                       .Enables(txtDropClass)
                       .Enables(udDropClass)
                      );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DropIfClass"]',
                        'Drop If Class', REGTYPE_BOOL, DEF_DROPIFCLASS,
                        chkDropIfClass));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["DropClass"]',
                      'Drop Class', REGTYPE_INT, DEF_DROPCLASS,
                      udDropClass,
                      txtDropClass,
                      AcceptIntGreaterThan(-1)));

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["KinPars"]',
                        'Kin Pars', REGTYPE_BOOL, DEF_KINPARS,
                        chkKinPars));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["KinParsAll"]',
                        'Kin Pars All', REGTYPE_BOOL, DEF_KINPARSALL,
                        chkKinParsAll));
  AddEditProperty(TQUB_Property_Edit.Create('["KinParsList"]',
                          '', REGTYPE_NONE, '',
                          txtKinParsList,
                          AcceptString,
                          ''));

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Operation"]',
                          'Operation', REGTYPE_INT, OperationToString(DEF_OPERATION))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOperationNone,
                                       OperationToString(OPER_NONE),
                                       Integer(OPER_NONE)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOperationAdd,
                                       OperationToString(OPER_ADD),
                                       Integer(OPER_ADD)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOperationSub,
                                       OperationToString(OPER_SUB),
                                       Integer(OPER_SUB)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOperationAvg,
                                       OperationToString(OPER_AVG),
                                       Integer(OPER_AVG)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOperationStd,
                                       OperationToString(OPER_STD),
                                       Integer(OPER_STD))));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Operation"]["Scale"]',
                        '', REGTYPE_NONE, DEF_OPERATION_SCALE,
                        chkOperationScale));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Operation"]["ScaleOneOver"]',
                        '', REGTYPE_NONE, DEF_OPERATION_SCALEONEOVER,
                        chkOperationScaleOneOver));
  AddEditProperty(TQUB_Property_Edit.Create('["Operation"]["ScaleFactor"]',
                    'Operation Scale Factor', REGTYPE_STRINGFLOAT, DEF_OPERATION_SCALEFACTOR,
                    txtOperationScaleFactor,
                    AcceptFloatAbsBetween(MIN_OPERATION_SCALEFACTOR, MAX_OPERATION_SCALEFACTOR),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Operation"]["List"]',
                        '', REGTYPE_NONE, DEF_OPERATION_LIST,
                        chkOperationList));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Operation"]["Files"]',
                        '', REGTYPE_NONE, DEF_OPERATION_FILES,
                        chkOperationFiles));

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["Drift"]',
                        'Function Drift', REGTYPE_BOOL, DEF_FUNCTION_DRIFT,
                        chkFunctionDrift));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["DriftA"]',
                    'Function Drift A', REGTYPE_STRINGFLOAT, DEF_FUNCTION_DRIFTA,
                    txtFunctionDriftA,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["DriftB"]',
                    'Function Drift B', REGTYPE_STRINGFLOAT, DEF_FUNCTION_DRIFTB,
                    txtFunctionDriftB,
                    AcceptFloat,
                    '') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["Sine"]',
                        'Function Sine', REGTYPE_BOOL, DEF_FUNCTION_SINE,
                        chkFunctionSine));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["SineFreq"]',
                    'Function Sine Frq', REGTYPE_STRINGFLOAT, DEF_FUNCTION_SINEFRQ,
                    txtFunctionSineFrq,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["SineAmp"]',
                    'Function Sine Amp', REGTYPE_STRINGFLOAT, DEF_FUNCTION_SINEAMP,
                    txtFunctionSineAmp,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["SinePhase"]',
                    'Function Sine Ph', REGTYPE_STRINGFLOAT, DEF_FUNCTION_SINEPH,
                    txtFunctionSinePh,
                    AcceptFloat,
                    '') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["Digitize"]',
                        'Function Digitize', REGTYPE_BOOL, DEF_FUNCTION_DIGITIZE,
                        chkFunctionDigitize));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["DigitizeThreshold"]',
                    'Digitize Threshold', REGTYPE_STRINGFLOAT, DEF_FUNCTION_DIGITIZE_THRESHOLD,
                    txtFunctionDigitizeThreshold,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["DigitizeToMin"]',
                    'Digitize ToMin', REGTYPE_STRINGFLOAT, DEF_FUNCTION_DIGITIZE_TOMIN,
                    txtFunctionDigitizeToMin,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["DigitizeToMax"]',
                    'Digitize ToMax', REGTYPE_STRINGFLOAT, DEF_FUNCTION_DIGITIZE_TOMAX,
                    txtFunctionDigitizeToMax,
                    AcceptFloat,
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["Derivate"]',
                        'Function Derivate', REGTYPE_BOOL, DEF_FUNCTION_DERIVATE,
                        chkFunctionDerivate)
                     );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["LKC"]',
                        'Function LKC', REGTYPE_BOOL, DEF_FUNCTION_LKC,
                        chkFunctionLKC));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["ItoG"]',
                        'Function ItoG', REGTYPE_BOOL, DEF_FUNCTION_ItoG,
                        chkFunctionItoG));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["ItoGVRev"]',
                    'ItoG VRev', REGTYPE_STRINGFLOAT, DEF_FUNCTION_ITOG_VREV,
                    txtFunctionItoGVRev,
                    AcceptFloat,
                    ''));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["ItoGVCh"]',
                    'ItoG VCh', REGTYPE_INT, DEF_FUNCTION_ITOG_VCH,
                    txtFunctionItoGVCh,
                    AcceptInt,
                    ''));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["RsCorr"]',
                        'Function RsCorr', REGTYPE_BOOL, DEF_FUNCTION_RSCORR,
                        chkFunctionRsCorr));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["RsValue"]',
                    'RsValue', REGTYPE_STRINGFLOAT, DEF_FUNCTION_RSVALUE,
                    txtFunctionRsCorr,
                    AcceptFloatGreaterThanOrEqualTo(0),
                    ''));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["RsPer"]',
                    'RsPer', REGTYPE_STRINGFLOAT, DEF_FUNCTION_RSPER,
                    txtFunctionRsCorrPer,
                    AcceptFloatBetween(0, 100),
                    ''));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["RsAddLeak"]',
                        'RsAddLeak', REGTYPE_BOOL, DEF_FUNCTION_RSADDLEAK,
                        chkFunctionRsAddLeak)
                        );
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["RsLeakCh"]',
                    'RsLeakCh', REGTYPE_INT, DEF_FUNCTION_RSLEAKCH,
                    txtFunctionRsLeakCh,
                    AcceptInt,
                    ''));

  //P/n
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["PnAdd"]',
                        'Function PnAdd', REGTYPE_BOOL, DEF_FUNCTION_PNADD,
                        chkFunctionPnAdd));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["PnSub"]',
                        'Function PnSub', REGTYPE_BOOL, DEF_FUNCTION_PNSUB,
                        chkFunctionPnSub));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["PnOffset"]',
                    'PnOffset', REGTYPE_STRINGFLOAT, DEF_FUNCTION_PNOFFSET,
                    txtFunctionPnOffset,
                    AcceptFloat,
                    ''));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["PnDelay"]',
                    'PnDelay', REGTYPE_STRINGFLOAT, DEF_FUNCTION_PNDELAY,
                    txtFunctionPnDelay,
                    AcceptFloat,
                    ''));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["PnScale"]',
                    'PnScale', REGTYPE_STRINGFLOAT, DEF_FUNCTION_PNSCALE,
                    txtFunctionPnScale,
                    AcceptFloat,
                    ''));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["PnCount"]',
                    'PnCount', REGTYPE_INT, DEF_FUNCTION_PNCOUNT,
                    txtFunctionPnCount,
                    AcceptIntGreaterThan(0),
                    ''));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["PnVm"]',
                    'PnVm', REGTYPE_STRINGFLOAT, DEF_FUNCTION_PNVM,
                    txtFunctionPnVm,
                    AcceptFloat,
                    ''));
  AddEditProperty(TQUB_Property_Edit.Create('["Function"]["PnInterval"]',
                    'PnInterval', REGTYPE_STRINGFLOAT, DEF_FUNCTION_PNINTERVAL,
                    txtFunctionPnInterval,
                    AcceptFloatGreaterThanOrEqualTo(0),
                    ''));

  //Fit
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Function"]["Fit"]',
                        'Function Fit', REGTYPE_BOOL, DEF_FUNCTION_FIT,
                        chkFunctionFit));

  RegKeyName               := 'Software\QUB\Task Extract Selection';
  ProfileCat               := 'Ext';
  WikiTopic                := 'Preprocessing:Extract';

  txtProcessResampleFactor.OnExit := txtProcessResampleFactorExit;
  txtProcessResampleFreq.OnExit   := txtProcessResampleFreqExit;

  {$include 'QUB_Task_ExtractSelection_Properties.htx'}

  mnuProfiles.Anchors          := [akTop, akLeft];
  mnuProfiles.SetBounds(btnHelp.Left + btnHelp.Width + 16, btnHelp.Top, panOperation.Width - (btnHelp.Left + btnHelp.Width + 16 - panOperation.Left), mnuProfiles.Height);

end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetOperationListName(n :String);
begin
  prOperationListName          := n;
  //set index
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  FileFormat                  := FileFormat;

  DataSource                  := DataSource;

  Advanced                    := Advanced;

  OperationListName           := OperationListName;

  //Memoryless
  Properties['Operation'].SetData(OperationToString(OPER_NONE));
  Properties['Operation']['Scale'].Data[0]    := 0;
  Properties['ASCIIHasFit'].Data[0]           := 0;
  Properties['KinPars'].Data[0]               := 0;
  Properties['Function']['Drift'].Data[0]     := 0;
  Properties['Function']['Sine'].Data[0]      := 0;
  Properties['Function']['Digitize'].Data[0]  := 0;
  Properties['Function']['Derivate'].Data[0]  := 0;
  Properties['Function']['LKC'].Data[0]       := 0;
  Properties['Function']['ItoG'].Data[0]      := 0;
  Properties['Function']['RsCorr'].Data[0]    := 0;
  Properties['Function']['PnAdd'].Data[0]     := 0;
  Properties['Function']['PnSub'].Data[0]     := 0;
  Properties['Function']['Fit'].Data[0]       := 0;
  Properties['ExtractFileList'].Data[0]       := 0;

  // update resample freq given sampling rate
  txtProcessResampleFactorExit(txtProcessResampleFactor);

  prUpdateComponents;
  prCountFileList;

end;


//******************************************************************************

procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Reg   :TRegistry;
  Prop  :INode;
  bVal  :Boolean;
  iVal  :Integer;
begin
  inherited;

  Properties['DialogOnLClick'].SetData(1);
  // a script could unset this

  Reg   := nil; // will be created as needed by ReadXRegistry*
  try
    Prop  := Props['FileFormat'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryInteger(Reg, RegKeyName, 'File Format', iVal) then begin
        FileFormat := TQUB_QFS_QDF_FileType(iVal);
      end else begin
        FileFormat := DEF_FILEFORMAT;
      end;
    end;
    Prop  := Props['Advanced'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryBoolean(Reg, RegKeyName, 'Advanced', bVal) then begin
        Prop.SetData( Integer(bVal) );
      end else begin
        Prop.SetData( DEF_ADVANCED );
      end;
    end;
  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.prUpdateComponents;
begin
  inherited;

  FileFormatString   := Properties['FileFormat'].DataAsString;
  Advanced           := 0<>Properties['Advanced'].Data[0];
end;


//******************************************************************************
procedure TQUB_Task_ExtractSelection_PropertiesForm.btnBrowseFileNameClick(Sender: TObject);
var
  Dlg :TSaveDialog;
begin
  inherited;
  Dlg := TSaveDialog.Create(Application);
  try
    Dlg.Filter := 'All QUB data files (*.qdf;*.dat;*.ldt;*.txt)|*.qdf;*.dat;*.ldt;*.txt|QUB data files (*.qdf)|*.qdf|QUB old binary files (*.dat)|*.dat|' +
                  'QUB old ldt files (*.ldt)|*.ldt|ASCII data files (*.txt)|*.txt|WAVE sound files (*.wav)|*.wav|Event list files (*.dwt)|*.dwt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofEnableSizing];
    Dlg.FileName := FileName;
    if Dlg.Execute then begin
      FileName := Dlg.FileName;
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.txtFileNameKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtFileNameExit(Sender);
  end;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.txtFileNameExit(Sender: TObject);
var
  f :String;
begin
  inherited;
  f := TEdit(Sender).Text;
  if f <> FileName then begin
    FileName := f;
  end;
end;

//Data Source
procedure TQUB_Task_ExtractSelection_PropertiesForm.chkDataSourceClick(Sender: TObject);
begin
  inherited;
  DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.txtProcessResampleFactorExit(Sender: TObject);
var
  x :Double;
  r :Double;
begin
  inherited;

  prOnExit(Sender);

  r := Properties['ProcessData']['ResampleFactor'].Data[0];
  x := prSamplingFrequency / r;
  Properties['ProcessData']['ResampleFreq'].Data[0] := x;
  txtProcessResampleFreq.Text := FloatToStr(x / 1000);
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.txtProcessResampleFreqExit(Sender: TObject);
var
  x :Double;
  f :Double;
begin
  inherited;

  prOnExit(Sender);

  f := Properties['ProcessData']['ResampleFreq'].Data[0];
  x := prSamplingFrequency / f;
  Properties['ProcessData']['ResampleFactor'].Data[0] := x;
  txtProcessResampleFactor.Text := FloatToStr(x);
end;

//File Format
procedure TQUB_Task_ExtractSelection_PropertiesForm.optFileFormatClick(Sender: TObject);
begin
  inherited;
  FileFormat := TQUB_QFS_QDF_FileType(TRadioButton(Sender).Tag);
end;

//Advanced Options
procedure TQUB_Task_ExtractSelection_PropertiesForm.btnAdvancedClick(Sender: TObject);
begin
  inherited;
  Advanced := not Advanced;
end;


procedure TQUB_Task_ExtractSelection_PropertiesForm.lstOperationListClick(Sender: TObject);
begin
  inherited;
  if lstOperationList.ItemIndex >= 0 then begin
    OperationListName := lstOperationList.Items[lstOperationList.ItemIndex];
  end else begin
    OperationListName := '';
  end;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.chkFunctionPnAddClick(Sender: TObject);
begin
  inherited;
  if chkFunctionPnAdd.Checked and (chkFunctionPnSub.Checked <> not chkFunctionPnAdd.Checked) then begin
    chkFunctionPnSub.Checked := not chkFunctionPnAdd.Checked;
  end;
  prSetEnabled;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.chkFunctionPnSubClick(Sender: TObject);
begin
  inherited;
  if chkFunctionPnSub.Checked and (chkFunctionPnAdd.Checked <> not chkFunctionPnSub.Checked) then begin
    chkFunctionPnAdd.Checked := not chkFunctionPnSub.Checked;
  end;
  prSetEnabled;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.prSetEnabledOnClick(Sender: TObject);
begin
  inherited;
  prSetEnabled;
end;

procedure TQUB_Task_ExtractSelection_PropertiesForm.btnFileListClick(Sender: TObject);
begin
  inherited;

  formDataList.ShowModal;
  prCountFileList;
end;


//******************************************************************************
procedure TQUB_Task_ExtractSelection_PropertiesForm.btnOKClick(Sender: TObject);
var
  Dlg :TSaveDialog;
  r   :Integer;
  f   :String;
  Channels :TQUB_Task_ExtractSelection_Channels;
begin
  case FileFormat of
    QFS_QDF_TYPE_QDF:
      f := ChangeFileExt(FileName, '.qdf');
    QFS_QDF_TYPE_LDT:
      f := ChangeFileExt(FileName, '.ldt');
    QFS_QDF_TYPE_DAT:
      f := ChangeFileExt(FileName, '.dat');
    QFS_QDF_TYPE_ASC:
      f := ChangeFileExt(FileName, '.txt');
    QFS_QDF_TYPE_WAV:
      f := ChangeFileExt(FileName, '.wav');
    QFS_QDF_TYPE_DWT:
      f := ChangeFileExt(FileName, '.dwt');
  end;

  Channels := TQUB_Task_ExtractSelection_Channels(Properties['Channels'].Data[0]);

  if FileFormat <> QFS_QDF_TYPE_DWT then begin
    if ((Channels = CHANNELS_ALL) and (prSourceADChannelCount > QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[FileFormat]))
    or ((Channels = CHANNELS_LIST) and (LongInt(Properties['ChannelList'].DataCount) > QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[FileFormat]))
    then begin
    //if (prSourceADChannelCount > QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[FileFormat]) and (not OnlyActiveChannel) then begin
      MessageHelpDlg(Format('The selected output file format does not allow ADChannelCount > %d'#13#10'Please select an appropriate format', [QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[FileFormat]]), mtError, [mbOK], 0, 'Error:Too_Many_Channels');
      ModalResult := mrNone;
      Exit;
    end;
  end;

  Dlg := nil;

  try

    Dlg := TSaveDialog.Create(Application);

    while FileExists(f) do begin

      Dlg.Filter := 'All QUB data files (*.qdf;*.dat;*.ldt;*.txt)|*.qdf;*.dat;*.ldt;*.txt|QUB data files (*.qdf)|*.qdf|QUB old binary files (*.dat)|*.dat|' +
                    'QUB old ldt files (*.ldt)|*.ldt|ASCII data files (*.txt)|*.txt|WAVE sound files (*.wav)|*.wav|Event list files (*.dwt)|*.dwt|All files (*.*)|*.*';
      Dlg.Options := [ofPathMustExist, ofEnableSizing];
      Dlg.FileName := f;
      r := MessageHelpDlg('File ''' + f + ''' already exists. Do you want to overwrite it ?', mtConfirmation, [mbYes, mbNo, mbCancel], 0, 'Warning:File_Already_Exists');
      case r of
        idCancel:
          begin
            ModalResult := mrNone;
            Exit;
          end;
        idYes:
          begin
            inherited;
            Exit;
          end;
        idNo:
          begin
            if Dlg.Execute then begin
              f := Dlg.FileName;
              case FileFormat of
                QFS_QDF_TYPE_QDF:
                  f := ChangeFileExt(f, '.qdf');
                QFS_QDF_TYPE_LDT:
                  f := ChangeFileExt(f, '.ldt');
                QFS_QDF_TYPE_DAT:
                  f := ChangeFileExt(f, '.dat');
                QFS_QDF_TYPE_ASC:
                  f := ChangeFileExt(f, '.txt');
                QFS_QDF_TYPE_WAV:
                  f := ChangeFileExt(f, '.wav');
                QFS_QDF_TYPE_DWT:
                  f := ChangeFileExt(f, '.dwt');
              end;
            end else begin
              ModalResult := mrNone;
              Exit;
            end;
          end;
      end;//case
    end;//while

    FileName := f;

    inherited;

  finally
    Dlg.Free;
  end;

end;


//******************************************************************************
end.
