{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// curve fitting dialog used by data, histograms, results:select
// calls on python (CurveFitting.py) to do most real work
// python calls on fit functions defined at bottom of this file


unit QUB_Task_Fit_Dlog;


interface


uses


Windows, SysUtils, Classes, Graphics, Forms, Controls, StdCtrls, Dialogs,
Buttons, ExtCtrls, Grids, QUB_Table, ComCtrls,
PythonEngine,
Alloc, AllocTypes,
FMath, matrices, regress,
QUB_Task,
QUB_Node;


type


TQUB_Task_Fit = class;
TQUB_Task_FitOptions = class;

TQUB_Task_Fit_Dlog = class(TForm)

  OKBtn                           : TButton;
  CancelBtn                       : TButton;
  Bevel1                          : TBevel;
  HelpBtn                         : TButton;
  Label1                          : TLabel;
  cmbCurveType                    : TComboBox;
  btnOptions                      : TButton;
  Label2                          : TLabel;
  txtComponents                   : TEdit;
  Label3                          : TLabel;
  txtBaseline                     : TEdit;
  Label4                          : TLabel;
  txtMaxIter                      : TEdit;
  btnFit                          : TButton;
  btnUpdate                       : TButton;
  txtResidual                     : TEdit;
  Label5                          : TLabel;
  Label6                          : TLabel;
  txtPrecision                    : TEdit;
  txtEqn                          : TEdit;

  procedure HelpBtnClick(Sender: TObject);
  procedure cmbCurveTypeChange(Sender: TObject);
  procedure btnOptionsClick(Sender: TObject);
  procedure txtBaselineKeyPress(Sender: TObject; var Key: Char);
  procedure txtBaselineExit(Sender: TObject);
  procedure txtMaxIterKeyPress(Sender: TObject; var Key: Char);
  procedure txtMaxIterExit(Sender: TObject);
  procedure btnFitClick(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure btnUpdateClick(Sender: TObject);
  procedure txtPrecisionKeyPress(Sender: TObject; var Key: Char);
  procedure txtPrecisionExit(Sender: TObject);
  procedure txtComponentsKeyPress(Sender: TObject; var Key: Char);
  procedure txtComponentsExit(Sender: TObject);

protected

  grdParams                       :TQUB_Table;

  prFitter                        :PPyObject;
  prFitterVar                     :OleVariant;
  prCurveType                     :String;
  prComponentCount                :Integer;
  prBaseline                      :Double;
  prMaxIter                       :Integer;
  prPrecision                     :Double;
  prData                          :INode;
  prProcData                      :INode;
  prCurves                        :INode;
  prParams                        :INode;
  prParamCount                    :Integer;
  prEqn                           :String;

  prTryCurveType                  :String;
  prTryComponentCount             :Integer;
  prCurveTypes                    :TStringList;

  procedure prDestroyFitterPy;

  function prGetParams: INode;
  procedure prGetParamsPy;

  procedure prSetCurveType(x :String);
  procedure prSetCurveTypePy;
  procedure prSetComponentCount(x :Integer);
  procedure prSetComponentCountPy;
  procedure prSetBaseline(x :Double);
  procedure prSetMaxIter(x :Integer);
  procedure prSetData(x :INode);
  procedure prSetDataPy;
  procedure prSetResidual(x :Double);
  procedure prSetPrecision(x :Double);

  procedure prProcessData;
  procedure prReadParams; // into grid
  procedure prRebuildCurvesPy;
  procedure prRebuildCurveTypes;
  procedure prRebuildCurveTypesPy;

  procedure prOnFitDone(Sender :TObject);

  procedure prDoOnCurvesChanged;

public

  OnCurvesChanged                 :TNotifyEvent;

  constructor Create(AOwner :TComponent); override;
  procedure Finalize; // call just before .Release

  property Fitter                 :PPyObject           read prFitter;
  property CurveType              :String              read prCurveType           write prSetCurveType;
  property ComponentCount         :Integer             read prComponentCount      write prSetComponentCount;
  property Baseline               :Double              read prBaseline            write prSetBaseline;
  property MaxIter                :Integer             read prMaxIter             write prSetMaxIter;
  property Params                 :INode               read prGetParams;
  property Data                   :INode               read prData                write prSetData;
  property ProcData               :INode               read prProcData;
  property Curves                 :INode               read prCurves;
  property Precision              :Double              read prPrecision           write prSetPrecision;
  property Residual               :Double                                         write prSetResidual;

end;


TQUB_Task_Fit = class(TQUB_Task)
protected

  prFitter                        :OleVariant;
  prMaxIter                       :Integer;
  prPrecision                     :Double;
  prIter                          :Integer;

  procedure prSetFitter(Fitter :PPyObject);

  procedure prDoFitPy;
  procedure prDoOnFitDone;

public

  OnFitDone                       :TNotifyEvent;

  procedure Execute; override;

  procedure OnGetProperties(Sender :TObject); override; // useless;
  procedure OnPutProperties(Sender :TObject); override; // useless;

  property Fitter                 :PPyObject                              write prSetFitter;
  property MaxIter                :Integer         read prMaxIter         write prMaxIter;
  property Precision              :Double          read prPrecision       write prPrecision;

end;


TQUB_Task_FitOptions = class(TQUB_Task)
protected

  prFitter                        :OleVariant;

  procedure prSetFitter(Fitter :PPyObject);

  procedure prDoOptionsPy;
  procedure prDoOnOptionsDone;

public

  OnOptionsDone                   :TNotifyEvent;

  procedure Execute; override;

  procedure OnGetProperties(Sender :TObject); override; // useless;
  procedure OnPutProperties(Sender :TObject); override; // useless;

  property Fitter                 :PPyObject       write prSetFitter;

end;


TQUB_PyFunc_NLRegFit = class(TPyObject)
public
  function  Call( args, kw : PPyObject) : PPyObject; override;
end;



function FitGauss_Eval(State, Comp :INode): INode;
function FitGauss_Fit(State, Data :INode): INode;


function OneGaussian(X : Float; B : PVector; comp : Integer) : Float;
function RegGaussian(X : Float; B : PVector) : Float;
// the sum of gaussians as a regression function suitable for tpmath1::regress
// B[0..2] = yscale, mean, sdev  --  repeating per component
procedure DerivGaussian(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);
// compute the partial derivatives of the above gaussian into D


function FitLogExpo_Eval(State, Comp :INode): INode;
function FitLogExpo_Fit(State, Data :INode): INode;

function OneLogExpo(X : Float; B : PVector; comp : Integer) : Float;
function RegLogExpo(X : Float; B : PVector) : Float;
procedure DerivLogExpo(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);


function FitExpo_Eval(State, Comp :INode): INode;
function FitExpo_Fit(State, Data :INode): INode;

function OneExpo(X : Float; B : PVector; comp : Integer) : Float;
function RegExpo(X : Float; B : PVector) : Float;
procedure DerivExpo(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);

function FitNegExpo_Eval(State, Comp :INode): INode;
function FitNegExpo_Fit(State, Data :INode): INode;

function OneNegExpo(X : Float; B : PVector; comp : Integer) : Float;
function RegNegExpo(X : Float; B : PVector) : Float;
procedure DerivNegExpo(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);



procedure InitCurveFits;


// !!! make sure nodes have mutexes


implementation


{$R *.DFM}


uses


QUB_Viewport_Python,
QUB_Python_Thread,
PythonAtom,
QUB_Node_Python,
Math;


var


// allows regression callbacks to know the number of components without being member functions
Histogram_CurveFit_Components           :Integer;


const


DEF_MAXITER                     = 20;
MIN_MAXITER                     = 1;
MAX_MAXITER                     = 1000;
DEF_PRECISION                   = 0.01;
MIN_PRECISION                   = 1.0e-20;
MAX_PRECISION                   = 1.0e20;

HISTOGRAM_CURVEFIT_COMPONENTS_MAX             = 8;
HISTOGRAM_CURVEFIT_ALGORITHM_MARQUARDT        = 0;
HISTOGRAM_CURVEFIT_ALGORITHM_SIMPLEX          = 1;
HISTOGRAM_CURVEFIT_ALGORITHM_BFGS             = 2;
HISTOGRAM_CURVEFIT_ALGORITHM_SA               = 3;


constructor TQUB_Task_Fit_Dlog.Create(AOwner :TComponent);
begin
  inherited;

  grdParams             := TQUB_Table.Create(Self);
  grdParams.Parent      := Self;
  grdParams.Left        := 8;
  grdParams.Top         := 166;
  grdParams.Width       := 340;
  grdParams.Height      := 196;
  grdParams.DefaultColWidth := 66;
  grdParams.FixedCols   := 0;
  grdParams.RowCount    := 2;
  grdParams.Options     := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRangeSelect, goRowSizing, goColSizing, goEditing, goThumbTracking];
  grdParams.TabOrder    := 9;

  prFitter                        := nil;
  prCurveType                     := '';
  prComponentCount                := 1;
  prBaseline                      := 0.0;
  prMaxIter                       := DEF_MAXITER;
  prPrecision                     := DEF_PRECISION;
  prData                          := TNode.CreateNull;
  prProcData                      := TNode.Create('DataSet');
  prCurves                        := TNode.Create('Curves');
  prParams                        := TNode.CreateNull;
  prParamCount                    := 0;

  prCurveTypes                    := TStringList.Create;
  prRebuildCurveTypes;
end;

procedure TQUB_Task_Fit_Dlog.FormShow(Sender: TObject);
begin
  inherited;

  Baseline                        := Baseline;
  MaxIter                         := MaxIter;
  Precision                       := Precision;

  //PythonThread.SyncPyProc( prRebuildCurveTypesPy, False, -1 );
  //CurveType := CurveType;
end;

procedure TQUB_Task_Fit_Dlog.Finalize;
begin
  try
    if fitter <> nil then begin
      PythonThread.SyncPyProc( prDestroyFitterPy, False, False, -1 );
    end;
    prCurveTypes.Free;
  except
  end;
end;

procedure TQUB_Task_Fit_Dlog.prDestroyFitterPy;
begin
  GetPythonEngine.Py_DECREF( prFitter );
  prFitter := nil;
  VarClear(prFitterVar);
end;

function TQUB_Task_Fit_Dlog.prGetParams: INode;
begin
  PythonThread.SyncPyProc( prGetParamsPy, False, False, -1 );
  prReadParams;
  Result := prParams;
end;

procedure TQUB_Task_Fit_Dlog.prGetParamsPy;
var
  ParamsPy         :PPyObject;
begin
  if prFitter = nil then Exit;

  with GetPythonEngine do begin
    ParamsPy := PyObject_GetAttrString( fitter, PChar('paramsNode') );
    if ParamsPy = nil then begin
      PyErr_Print;
    end else begin
      prParams := PyNode_ToNode( ParamsPy );
      Py_DECREF( ParamsPy );
    end;
  end;
end;

procedure TQUB_Task_Fit_Dlog.prSetCurveType(x :String);
begin
  prTryCurveType           := x;
  PythonThread.SyncPyProc( prSetCurveTypePy, False, False, -1 );
  if fitter <> nil then begin
    prCurveType            := x;
    txtComponents.Text     := IntToStr(prComponentCount);
    txtEqn.Text            := prEqn;
    if not Data.IsNull then Data := Data;
    prGetParams;         // also reads them in
    cmbCurveType.ItemIndex := cmbCurveType.Items.IndexOf(x);
    prDoOnCurvesChanged;
  end else begin
    prCurveType            := '';
    cmbCurveType.Text      := '';
    txtEqn.Text            := '';
  end;
end;

procedure TQUB_Task_Fit_Dlog.prSetCurveTypePy;
var
  FitterMaker              :PPyObject;
begin
  with GetPythonEngine do begin
    if prFitter <> nil then Py_DECREF( prFitter );
    prFitter                 := nil;

    try
      FitterMaker            := EvalString(Format('CurveFitting.fitters["%s"]',[prTryCurveType]));

      prFitter             := PyObject_CallObject( FitterMaker, nil );
      if prFitter = nil then begin
        PyErr_Print;
      end else begin
        PyNode_PublishPyObj( prFitter, 'fitter' );
        prFitterVar        := getAtom(prFitter);
        prRebuildCurvesPy;

        try
          prComponentCount := prFitterVar.getNumComponents;
          prEqn            := prFitterVar.getEqn;
        except
          prComponentCount := 1;
          prEqn            := '';
        end;
      end;
      Py_DECREF( FitterMaker );
    except
    end;
  end;
end;

procedure TQUB_Task_Fit_Dlog.prSetComponentCount(x :Integer);
begin
  if (x = prComponentCount) or (prFitter = nil) then Exit;

  prTryComponentCount := x;
  PythonThread.SyncPyProc( prSetComponentCountPy, False, False, -1 );

  txtComponents.Text := IntToStr(prComponentCount);

  prReadParams;
  PythonThread.SyncPyProc( prRebuildCurvesPy, False, False, -1 );
  prDoOnCurvesChanged;
end;

procedure TQUB_Task_Fit_Dlog.prSetComponentCountPy;
begin
  if prFitter = nil then Exit;

  try
    prFitterVar.setNumComponents( prTryComponentCount );
    prComponentCount := prFitterVar.getNumComponents();
  except
  end;
end;

procedure TQUB_Task_Fit_Dlog.prSetBaseline(x :Double);
begin
  prBaseline       := x;
  txtBaseline.Text := FloatToStr(x);
  prProcessData;
end;

procedure TQUB_Task_Fit_Dlog.prSetMaxIter(x :Integer);
begin
  prMaxIter       := x;
  txtMaxIter.Text := IntToStr(x);
end;

procedure TQUB_Task_Fit_Dlog.prSetData(x :INode);
begin
  prData          := x;
  prProcessData;
  PythonThread.SyncPyProc( prSetDataPy, False, False, -1 );
  prDoOnCurvesChanged;
  prReadParams;
end;

procedure TQUB_Task_Fit_Dlog.prSetDataPy;
var
  ProcDataPy           :PPyObject;
begin
  if prFitter = nil then Exit;

  try
    ProcDataPy := PyNode_FromNode( prProcData );
    prFitterVar.useData( getAtom(ProcDataPy) );
    GetPythonEngine.Py_DECREF( ProcDataPy );

    prRebuildCurvesPy;
  except
  end;
end;

procedure TQUB_Task_Fit_Dlog.prSetPrecision(x :Double);
begin
  prPrecision       := x;
  txtPrecision.Text := FloatToStr(x);
end;

procedure TQUB_Task_Fit_Dlog.prSetResidual(x :Double);
begin
  txtResidual.Text := Format('%.7g', [x]);
end;

procedure TQUB_Task_Fit_Dlog.prDoOnCurvesChanged;
var
  Curve               :INode;
  i                   :Integer;
  Resid               :Double;
  OneResid            :Double;
  Y                   :INode;
begin
  txtEqn.Text         := prEqn;

  Curve               := prCurves.Child;
  Y                   := prData.Find('Y');
  if Curve.IsNull or Y.IsNull then Exit;

  Resid               := 0.0;
  for i := 0 to Math.Min(Curve.DataCount, Y.DataCount) - 1 do begin
    OneResid          := Curve.Data[i] -  Y.Data[i];
    Resid             := Resid + OneResid * OneResid;
  end;
  Residual            := Resid;

  if Assigned(OnCurvesChanged) then OnCurvesChanged(Self);
end;

procedure TQUB_Task_Fit_Dlog.prProcessData;
var
  Y, ProcY         :INode;
  i                :Integer;
begin
  Y                := prProcData.Find('X');
  if not Y.IsNull then prProcData.RemoveChild(Y);
  prProcData.AppendClone( prData['X'], True );

  Y                := prData['Y'];
  ProcY            := prProcData['Y'];
  if Y.DataCount <> ProcY.DataCount then begin
    ProcY.SetupNumData( NODE_DATATYPE_DOUBLE, Y.DataCount, 1 );
  end;

  for i := 0 to Y.DataCount - 1 do begin
    ProcY.Data[i]  := Y.Data[i] - prBaseline;
  end;
end;

procedure TQUB_Task_Fit_Dlog.prReadParams;
var
  r, c                 :Integer;
  Comp, Param          :INode;
begin
  if prParams.IsNull then Exit;

  grdParams.RowCount   := 2;
  grdParams.ColCount   := 1;

  r                    := 1;
  Comp                 := prParams.Child;
  while not Comp.IsNull do begin
    c                  := 0;
    Param              := Comp.Child;
    while not Param.IsNull do begin
      try
        grdParams.RowCount                 := r + 1;
        grdParams.ColCount                 := c + 1;
        grdParams.Cells[c,r]               := Format('%.6f', [Double(Param.Data[0])]);
        if r = 1 then grdParams.Cells[c,0] := Param.Name;
        c                                  := c + 1;
      except
      end;

      Param            := Param.Sibling;
    end;

    r                  := r + 1;
    Comp               := Comp.Sibling;
  end;

  prParamCount         := grdParams.ColCount;
end;

procedure TQUB_Task_Fit_Dlog.prRebuildCurvesPy;
var
  CurvesPy            :PPyObject;
  Curve               :INode;
  i                   :Integer;
begin
  if prFitter = nil then Exit;

  CurvesPy            := nil;
  try
    CurvesPy          := PyNode_FromNode( prCurves );
    getAtom(prFitter).getCurves( getAtom(CurvesPy) );
  finally
    GetPythonEngine.Py_XDECREF( CurvesPy );
  end;

  Curve := prCurves.Child;
  while not Curve.IsNull do begin
    for i := 0 to Curve.DataCount - 1 do begin
      Curve.Data[i] := Curve.Data[i] + Baseline;
    end;

    Curve := Curve.Sibling;
  end;

  prEqn            := prFitterVar.getEqn;
end;

procedure TQUB_Task_Fit_Dlog.prRebuildCurveTypes;
var
  i              :Integer;
begin
  cmbCurveType.Clear;
  PythonThread.SyncPyProc( prRebuildCurveTypesPy, False, False, -1 );

  for i := 0 to prCurveTypes.Count - 1 do begin
    cmbCurveType.Items.Add( prCurveTypes[i] );
  end;
end;

procedure TQUB_Task_Fit_Dlog.prRebuildCurveTypesPy;
var
  Keys, Key          :PPyObject;
  i, N               :Integer;
begin
  prCurveTypes.Clear;

  with GetPythonEngine do begin
    Keys             := nil;
    try
      Keys           := EvalString('CurveFitting.fitters.keys()');
      N              := PySequence_Length( Keys );

      for i := 0 to N - 1 do begin
        Key          := PySequence_GetItem( Keys, i );
        if PyString_Check(Key) then begin
          prCurveTypes.Add( PyString_AsString(Key) );
        end;
        Py_DECREF(Key);
      end;
    finally
      Py_XDECREF( Keys );
    end;
  end;
end;

procedure TQUB_Task_Fit_Dlog.prOnFitDone(Sender :TObject);
begin
  PythonThread.SyncPyProc( prRebuildCurvesPy, False, False, -1 );
  prDoOnCurvesChanged;
  prReadParams;

  btnFit.Caption               := 'Fit';
  btnFit.Enabled               := True;

  OKBtn.Enabled                := True;
  btnUpdate.Enabled            := True;
  HelpBtn.Enabled              := True;
  grdParams.Enabled            := True;
  txtMaxIter.Enabled           := True;
  txtBaseline.Enabled          := True;
  txtComponents.Enabled        := True;
  txtPrecision.Enabled         := True;
  cmbCurveType.Enabled         := True;
  btnOptions.Enabled           := True;

  Cursor                       := crDefault;
end;

// *****************************************************************************
procedure TQUB_Task_Fit_Dlog.HelpBtnClick(Sender: TObject);
begin
  Application.HelpContext(HelpContext);
end;

procedure TQUB_Task_Fit_Dlog.cmbCurveTypeChange(Sender: TObject);
begin
  inherited;

  CurveType          := cmbCurveType.Text;
  cmbCurveType.Text  := CurveType;
end;

procedure TQUB_Task_Fit_Dlog.btnOptionsClick(Sender: TObject);
var
  Task                         :TQUB_Task_FitOptions;
begin
  inherited;

  if prFitter = nil then Exit;

  PythonThread.BeginPython;
  Task                         := TQUB_Task_FitOptions.Create(True);
  Task.Fitter                  := Fitter;
  Task.OnOptionsDone           := prOnFitDone;
  PythonThread.EndPython;

  Task.Resume;
end;

procedure TQUB_Task_Fit_Dlog.txtComponentsKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtComponentsExit(Sender);
  end;
end;

procedure TQUB_Task_Fit_Dlog.txtComponentsExit(Sender: TObject);
var
  i         :Integer;
begin
  try
    i := StrToInt(TEdit(Sender).Text);
    if i <= 0 then raise Exception.Create('');
  except
    MessageDlg('Invalid value', mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(ComponentCount);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  ComponentCount := i;
end;

procedure TQUB_Task_Fit_Dlog.txtBaselineKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtBaselineExit(Sender);
  end;
end;

procedure TQUB_Task_Fit_Dlog.txtBaselineExit(Sender: TObject);
var
  c :Double;
begin
  inherited;
  try
    c := StrToFloat(TEdit(Sender).Text);
  except
    MessageDlg('Invalid value', mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(Baseline);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if c <> Baseline then begin
    Baseline := c;
  end;
end;

procedure TQUB_Task_Fit_Dlog.txtMaxIterKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtMaxIterExit(Sender);
  end;
end;

procedure TQUB_Task_Fit_Dlog.txtMaxIterExit(Sender: TObject);
var
  m :Integer;
begin
  inherited;
  try
    m := StrToInt(TEdit(Sender).Text);
    if ((m < MIN_MAXITER) or (m > MAX_MAXITER)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_MAXITER) + ' and ' + IntToStr(MAX_MAXITER), mtError, [mbOK], 0);
    MaxIter := MaxIter;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if m <> MaxIter then begin
    MaxIter := m;
  end;
end;

procedure TQUB_Task_Fit_Dlog.txtPrecisionKeyPress(Sender: TObject;
  var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtPrecisionExit(Sender);
  end;
end;

procedure TQUB_Task_Fit_Dlog.txtPrecisionExit(Sender: TObject);
var
  c :Double;
begin
  inherited;
  try
    c := StrToFloat(TEdit(Sender).Text);
    if ((c < MIN_PRECISION) or (c > MAX_PRECISION)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_PRECISION) + ' and ' + FloatToStr(MAX_PRECISION), mtError, [mbOK], 0);
    Precision := Precision;
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if c <> Precision then begin
    Precision := c;
  end;
end;

procedure TQUB_Task_Fit_Dlog.btnFitClick(Sender: TObject);
var
  Task                         :TQUB_Task_Fit;
begin
  inherited;

  if prFitter = nil then Exit;

  btnUpdateClick(Sender); // make sure have new values of params and baseline

  // btnFit.Caption               := 'Stop';
  btnFit.Enabled               := False;

  OKBtn.Enabled                := False;
  btnUpdate.Enabled            := False;
  HelpBtn.Enabled              := False;
  grdParams.Enabled            := False;
  txtMaxIter.Enabled           := False;
  txtBaseline.Enabled          := False;
  txtComponents.Enabled        := False;
  txtPrecision.Enabled         := False;
  cmbCurveType.Enabled         := False;
  btnOptions.Enabled           := False;

  Cursor                       := crHourglass;

  Task                         := TQUB_Task_Fit.Create(True);
  PythonThread.BeginPython;
  Task.Fitter                  := Fitter;
  Task.MaxIter                 := MaxIter;
  Task.Precision               := Precision;
  Task.OnFitDone               := prOnFitDone;
  PythonThread.EndPython;
  
  Task.Resume;
end;

// procedure TQUB_Task_Fit_Dlog.grdParamsSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
procedure TQUB_Task_Fit_Dlog.btnUpdateClick(Sender: TObject);
var
  Params, Param    :INode;
  r, c             :Integer;
begin
  inherited;

  Params           := prParams.Child;
  r                := 1;
  while not Params.IsNull do begin
    Param          := Params.Child;
    c              := 0;
    while not Param.IsNull do begin
      try
        Param.Data[0] := StrToFloat( grdParams.Cells[c,r] );
      except
        grdParams.Cells[c,r] := FloatToStr(Double(Param.Data[0]));
      end;

      Param        := Param.Sibling;
      c            := c + 1;
    end;

    Params         := Params.Sibling;
    r              := r + 1;
  end;

  try
    PythonThread.SyncPyProc( prRebuildCurvesPy, False, False, -1 );
    prDoOnCurvesChanged;
  except
  end;
end;


// *****************************************************************************
procedure TQUB_Task_Fit.Execute;
begin
  Synchronize( prDoOnChange );
  prIter := 0;
  PythonThread.SyncPyProc( prDoFitPy, False, False, -1 );
  Synchronize( prDoOnFitDone );
end;

procedure TQUB_Task_Fit.prSetFitter(Fitter :PPyObject);
begin
  prFitter := getAtom(Fitter);
end;

procedure TQUB_Task_Fit.prDoFitPy;
begin
  try
    prFitter.fit( MaxIter, Precision );
  finally
  end;
end;

procedure TQUB_Task_Fit.prDoOnFitDone;
begin
  if Assigned(OnFitDone) then OnFitDone(Self);
end;

procedure TQUB_Task_FitOptions.Execute;
begin
  Synchronize( prDoOnChange );
  PythonThread.SyncPyProc( prDoOptionsPy, False, False, -1 );
  Synchronize( prDoOnOptionsDone );
end;

procedure TQUB_Task_FitOptions.prSetFitter(Fitter :PPyObject);
begin
  prFitter := getAtom(Fitter);
end;

procedure TQUB_Task_FitOptions.prDoOptionsPy;
begin
  try
    prFitter.promptOptions;
  except
  end;
end;

procedure TQUB_Task_FitOptions.prDoOnOptionsDone;
begin
  if Assigned(OnOptionsDone) then OnOptionsDone(Self);
end;

procedure TQUB_Task_Fit.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_Fit.OnPutProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_FitOptions.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_FitOptions.OnPutProperties(Sender :TObject);
begin
end;


// *****************************************************************************

function FitGauss_Eval(State, Comp :INode): INode;
var
  X, Mean, Std, Amp, xdev       :Double;
begin
  Result := State['Compute'];
  X      := Double(Result.Data[0]);
  Mean   := Double(Comp['Mean'].Data[0]);
  Std    := Double(Comp['Std'].Data[0]);
  Amp    := Double(Comp['Amp'].Data[0]);

  xdev  := X - Mean;
  Result.Data[1] := Amp * exp( - xdev * xdev / (2 * Std * Std) );
end;

function FitGauss_Fit(State, Data :INode): INode;
var
  MaxIter               :Integer;
  Components            :Integer;
  Component             :INode;
  XNode, YNode          :INode;

  N                     :Integer;
  X, Y                  :PVector;
  B, B_min, B_max       :PVector; // reg.func parameters and bounds
  V                     :PMatrix; // inverse matrix of normal eqns (output)
  i                     :Integer;
  comp                  :Integer; // start index of component in B
  //FitStats              :TRegTest;
  err                   :Integer;
begin
  Result := State;

  MaxIter        := Integer( State['MaxIter'].Data[0] );

  Components     := 0;
  Component      := State['Params'].Child;
  while not Component.IsNull do begin
    Components   := Components + 1;
    Component    := Component.Sibling;
  end;

  X     := nil;
  Y     := nil;
  B     := nil;
  B_min := nil;
  B_max := nil;
  V     := nil;

  XNode := Data['X'];
  YNode := Data['Y'];
  N     := Math.Min(XNode.DataCount, YNode.DataCount);
  DimVector(X, N);
  DimVector(Y, N);
  for i := 0 to N - 1 do begin
    X[i] := XNode.Data[i];
    Y[i] := YNode.Data[i];
  end;

  try
    DimVector(B,     Components*3);
    DimVector(B_min, Components*3);
    DimVector(B_max, Components*3);
    DimMatrix(V, Components*3, Components*3); // ?

    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 3;

      B^[comp+0] := Component['Amp'].Data[0];
      B_min^[comp+0] := 1.0e-20;
      B_max^[comp+0] := 1.0e20;

      B^[comp+1] := Component['Mean'].Data[0];
      B_min^[comp+1] := X^[0]; // ?
      B_max^[comp+1] := X^[N-1]; // ?

      B^[comp+2] := Component['Std'].Data[0];
      B_min^[comp+2] := 1.0e-20;
      B_max^[comp+2] := X^[N-1] - X^[0]; // ?

      Component := Component.Sibling;
    end;

    SetFirstPoint(0); // you'd think they'd at least mention this kind of weirdness in the docs
    Histogram_CurveFit_Components := Components;
    err := NLFit(RegGaussian, DerivGaussian, X, Y, N-1, 0, 3*Components-1, MaxIter, {Precision}0.01, B, B_min, B_max, V);
    State['Err'].SetData( err );

    {
    AreaSum := 0.0; // == (area of sum of components) / sqrt(2*PI)
    for i := 0 to Components - 1 do begin
      comp := i * 3;
      AreaSum := AreaSum + B^[comp+0] * B^[comp+2];
    end;
    }
    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 3;
      Component['Amp'].Data[0]  := B[comp+0];
      Component['Mean'].Data[0] := B[comp+1];
      Component['Std'].Data[0]  := B[comp+2];

      Component := Component.Sibling;
    end;

    //RegTest(Y, Yout, N-1, 0, 3*Components-1, V, FitStats);
    //State['Residual'].SetData( FitStats.Vr );
  except
  end;

  if X     <> nil then DelVector(X    , N);
  if Y     <> nil then DelVector(Y    , N);
  if B     <> nil then DelVector(B    , 3*Components);
  if B_min <> nil then DelVector(B_min, 3*Components);
  if B_max <> nil then DelVector(B_max, 3*Components);
  if V     <> nil then DelMatrix(V, 3*Components, 3*Components);
end;


//******************************************************************************
function OneGaussian(X : Float; B : PVector; comp : Integer) : Float;
// B: yscale, mean, sdev   --  repeating per component
var
  xdev          :Float;
begin
  Result := 0.0;
  try
    comp := comp * 3; // to become the start index of parameters in B
    xdev  := X - B^[comp + 1];
    Result := B^[comp+0] * exp( - xdev * xdev / (2 * B^[comp+2] * B^[comp+2]) );
  except
  end;
end;

function RegGaussian(X : Float; B : PVector) : Float;
// B: yscale, mean, sdev   --  repeating per component
var
  xdev          :Float;
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
begin
  Result := 0;
  for i := 0 to Histogram_CurveFit_Components - 1 do begin
    try
      comp := i * 3;
      xdev  := X - B^[comp + 1];
      Result := Result + B^[comp+0] * exp( - xdev * xdev / (2 * B^[comp+2] * B^[comp+2]) );
    except
    end;
  end;
end;

procedure DerivGaussian(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);
var
  yval           :Float;
  xdev           :Float;
  variance       :Float;
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
begin
  for i := 0 to Histogram_CurveFit_Components - 1 do begin
    try
      comp := i * 3;
      xdev := X - B^[comp+1];
      variance := B^[comp+2] * B^[comp+2];    
      if variance = 0.0 then variance := 1.0e-10;
      yval := B^[comp+0] * exp( - xdev * xdev / (2 * variance) );

      if B^[comp+0] > 0.0 then begin
        D^[comp+0] := yval / B^[comp+0];                             // df/dscale
      end else begin
        D^[comp+0] := yval / 1.0e-10;
      end;
      D^[comp+1] := yval * xdev / variance;                        // df/dmean
      D^[comp+2] := yval * xdev * xdev / (variance * B^[comp+2]);  // df/dsdev
    except
    end;
  end;
end;


// *****************************************************************************

function FitLogExpo_Eval(State, Comp :INode): INode;
var
  X0, X1, Tau, Amp    :Double;
begin
  Result := State['Compute'];
  X0     := power(10, Double(Result.Data[0]));
  X1     := power(10, Double(Result.Data[1]));
  Tau    := Double(Comp['Tau'].Data[0]);
  Amp    := Double(Comp['Amp'].Data[0]);

  Result.Data[1] := Amp * Tau * (exp( - X0 / Tau ) - exp( - X1 / Tau ));
end;

var
  gLogExpoR         :Double;

function FitLogExpo_Fit(State, Data :INode): INode;
var
  MaxIter               :Integer;
  Components            :Integer;
  Component             :INode;
  XNode, YNode          :INode;

  N                     :Integer;
  X, Y                  :PVector;
  B, B_min, B_max       :PVector; // reg.func parameters and bounds
  V                     :PMatrix; // inverse matrix of normal eqns (output)
  i                     :Integer;
  comp                  :Integer; // start index of component in B
  //FitStats              :TRegTest;
  err                   :Integer;
begin
  Result := State;

  MaxIter        := Integer( State['MaxIter'].Data[0] );

  Components     := 0;
  Component      := State['Params'].Child;
  while not Component.IsNull do begin
    Components   := Components + 1;
    Component    := Component.Sibling;
  end;

  X     := nil;
  Y     := nil;
  B     := nil;
  B_min := nil;
  B_max := nil;
  V     := nil;

  XNode := Data['X'];
  YNode := Data['Y'];
  N     := Math.Min(XNode.DataCount, YNode.DataCount);
  DimVector(X, N);
  DimVector(Y, N);
  for i := 0 to N - 1 do begin
    X[i] := power(10, XNode.Data[i]);
    Y[i] := YNode.Data[i];
  end;
  gLogExpoR := X[N-1] / X[N-2];

  try
    DimVector(B,     Components*2);
    DimVector(B_min, Components*2);
    DimVector(B_max, Components*2);
    DimMatrix(V, Components*2, Components*2); // ?

    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 2;

      B^[comp+0] := Component['Amp'].Data[0];
      B_min^[comp+0] := 1.0e-20;
      B_max^[comp+0] := 1.0e20;

      B^[comp+1] := Component['Tau'].Data[0];
      B_min^[comp+1] := -1.0e20; // ?
      B_max^[comp+1] := 1.0e20; // ?

      Component := Component.Sibling;
    end;

    SetFirstPoint(0); // you'd think they'd at least mention this kind of weirdness in the docs
    SetOptAlgo( TOptAlgo( Integer(State['algo'].Data[0]) ) );

    Histogram_CurveFit_Components := Components;
    err := NLFit(RegLogExpo, DerivLogExpo, X, Y, N-1, 0, 2*Components-1, MaxIter, {Precision}0.01, B, B_min, B_max, V);
    State['Err'].SetData( err );

    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 2;
      Component['Amp'].Data[0]  := B[comp+0];
      Component['Tau'].Data[0]  := B[comp+1];

      Component := Component.Sibling;
    end;

    //RegTest(Y, Yout, N-1, 0, 2*Components-1, V, FitStats);
    //State['Residual'].SetData( FitStats.Vr );
  except
  end;

  if X     <> nil then DelVector(X    , N);
  if Y     <> nil then DelVector(Y    , N);
  if B     <> nil then DelVector(B    , 2*Components);
  if B_min <> nil then DelVector(B_min, 2*Components);
  if B_max <> nil then DelVector(B_max, 2*Components);
  if V     <> nil then DelMatrix(V, 2*Components, 2*Components);
end;


//******************************************************************************
function OneLogExpo(X : Float; B : PVector; comp : Integer) : Float;
// B: Amp, Tau   --  repeating per component
var
  Tau    :Double;
begin
  Result := 0.0;
  try
    comp := comp * 2; // to become the start index of parameters in B
    Tau    := B^[comp + 1];
    Result := B^[comp+0] * Tau * (exp( - X / Tau ) - exp( - X * gLogExpoR / Tau ));
  except
  end;
end;

function RegLogExpo(X : Float; B : PVector) : Float;
// B: Amp, Tau   --  repeating per component
var
  Tau           :Double;
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
begin
  Result := 0;
  for i := 0 to Histogram_CurveFit_Components - 1 do begin
    try
      comp := i * 2;
      Tau    := B^[comp + 1];
      Result := Result + B^[comp+0] * Tau * (exp( - X / Tau ) - exp( - X * gLogExpoR / Tau ));
    except
    end;
  end;
end;

procedure DerivLogExpo(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);
var
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
  expL, expR    :Double;
  Amp           :Double;
  Tau           :Double;
begin
  for i := 0 to Histogram_CurveFit_Components - 1 do begin
    try
      comp      := i * 2;
      Amp       := B^[comp + 0];
      Tau       := B^[comp + 1];

      expL      := exp(- X / Tau);
      expR      := exp(- X * gLogExpoR / Tau);

      D^[comp+0] := (expR - expL) * Tau;
      D^[comp+1] := Amp * (expL - expR) + (Amp / Tau) * (X * expL - X * gLogExpoR * expR);
    except
    end;
  end;
end;


// *****************************************************************************

function FitExpo_Eval(State, Comp :INode): INode;
var
  X, Tau, Amp    :Double;
begin
  Result := State['Compute'];
  X     := Double(Result.Data[0]);
  Tau    := Double(Comp['Tau'].Data[0]);
  Amp    := Double(Comp['Amp'].Data[0]);

  Result.Data[1] := Amp * exp( - X / Tau );
end;

function FitExpo_Fit(State, Data :INode): INode;
var
  MaxIter               :Integer;
  Components            :Integer;
  Component             :INode;
  XNode, YNode          :INode;

  N                     :Integer;
  X, Y                  :PVector;
  B, B_min, B_max       :PVector; // reg.func parameters and bounds
  V                     :PMatrix; // inverse matrix of normal eqns (output)
  i                     :Integer;
  comp                  :Integer; // start index of component in B
  //FitStats              :TRegTest;
  err                   :Integer;
begin
  Result := State;

  MaxIter        := Integer( State['MaxIter'].Data[0] );

  Components     := 0;
  Component      := State['Params'].Child;
  while not Component.IsNull do begin
    Components   := Components + 1;
    Component    := Component.Sibling;
  end;

  X     := nil;
  Y     := nil;
  B     := nil;
  B_min := nil;
  B_max := nil;
  V     := nil;

  XNode := Data['X'];
  YNode := Data['Y'];
  N     := Math.Min(XNode.DataCount, YNode.DataCount);
  DimVector(X, N);
  DimVector(Y, N);
  for i := 0 to N - 1 do begin
    X[i] := XNode.Data[i];
    Y[i] := YNode.Data[i];
  end;

  try
    DimVector(B,     Components*2);
    DimVector(B_min, Components*2);
    DimVector(B_max, Components*2);
    DimMatrix(V, Components*2, Components*2); // ?

    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 2;

      B^[comp+0] := Component['Amp'].Data[0];
      B_min^[comp+0] := 1.0e-20;
      B_max^[comp+0] := 1.0e20;

      B^[comp+1] := Component['Tau'].Data[0];
      B_min^[comp+1] := -1.0e20; // ?
      B_max^[comp+1] := 1.0e20; // ?

      Component := Component.Sibling;
    end;

    SetFirstPoint(0); // you'd think they'd at least mention this kind of weirdness in the docs
    SetOptAlgo( TOptAlgo( Integer(State['algo'].Data[0]) ) );

    Histogram_CurveFit_Components := Components;
    err := NLFit(RegExpo, DerivExpo, X, Y, N-1, 0, 2*Components-1, MaxIter, {Precision}0.01, B, B_min, B_max, V);
    State['Err'].SetData( err );

    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 2;
      Component['Amp'].Data[0]  := B[comp+0];
      Component['Tau'].Data[0]  := B[comp+1];

      Component := Component.Sibling;
    end;

    //RegTest(Y, Yout, N-1, 0, 2*Components-1, V, FitStats);
    //State['Residual'].SetData( FitStats.Vr );
  except
  end;

  if X     <> nil then DelVector(X    , N);
  if Y     <> nil then DelVector(Y    , N);
  if B     <> nil then DelVector(B    , 2*Components);
  if B_min <> nil then DelVector(B_min, 2*Components);
  if B_max <> nil then DelVector(B_max, 2*Components);
  if V     <> nil then DelMatrix(V, 2*Components, 2*Components);
end;


//******************************************************************************
function OneExpo(X : Float; B : PVector; comp : Integer) : Float;
// B: Amp, Tau   --  repeating per component
var
  Tau    :Double;
begin
  Result := 0.0;
  try
    comp := comp * 2; // to become the start index of parameters in B
    Tau    := B^[comp + 1];
    Result := B^[comp+0] * exp( - X / Tau );
  except
  end;
end;

function RegExpo(X : Float; B : PVector) : Float;
// B: Amp, Tau   --  repeating per component
var
  Tau           :Double;
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
begin
  Result := 0;
  try
    for i := 0 to Histogram_CurveFit_Components - 1 do begin
      comp := i * 2;
      Tau    := B^[comp + 1];
      Result := Result + B^[comp+0] * exp( - X / Tau );
    end;
  except
  end;
end;

procedure DerivExpo(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);
var
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
  expL          :Double;
  Amp           :Double;
  Tau           :Double;
begin
  for i := 0 to Histogram_CurveFit_Components - 1 do begin
    comp      := i * 2;
    try
      Amp       := B^[comp + 0];
      Tau       := B^[comp + 1];
      expL      := exp(- X / Tau);

      D^[comp+0] := expL;
      D^[comp+1] := X * Amp / (Tau * Tau) * expL;
    except
      D^[comp+0] := -1.0e10;
      D^[comp+1] := -1.0e10;
    end;
  end;
end;


// *****************************************************************************

function FitNegExpo_Eval(State, Comp :INode): INode;
var
  X, Tau, Amp    :Double;
begin
  Result := State['Compute'];
  X      := Double(Result.Data[0]);
  Tau    := Double(Comp['Tau'].Data[0]);
  Amp    := Double(Comp['Amp'].Data[0]);

  Result.Data[1] := Amp * (1.0 - exp( - X / Tau ));
end;

function FitNegExpo_Fit(State, Data :INode): INode;
var
  MaxIter               :Integer;
  Components            :Integer;
  Component             :INode;
  XNode, YNode          :INode;

  N                     :Integer;
  X, Y                  :PVector;
  B, B_min, B_max       :PVector; // reg.func parameters and bounds
  V                     :PMatrix; // inverse matrix of normal eqns (output)
  i                     :Integer;
  comp                  :Integer; // start index of component in B
  //FitStats              :TRegTest;
  err                   :Integer;
begin
  Result := State;

  MaxIter        := Integer( State['MaxIter'].Data[0] );

  Components     := 0;
  Component      := State['Params'].Child;
  while not Component.IsNull do begin
    Components   := Components + 1;
    Component    := Component.Sibling;
  end;

  X     := nil;
  Y     := nil;
  B     := nil;
  B_min := nil;
  B_max := nil;
  V     := nil;

  XNode := Data['X'];
  YNode := Data['Y'];
  N     := Math.Min(XNode.DataCount, YNode.DataCount);
  DimVector(X, N);
  DimVector(Y, N);
  for i := 0 to N - 1 do begin
    X[i] := XNode.Data[i];
    Y[i] := YNode.Data[i];
  end;

  try
    DimVector(B,     Components*2);
    DimVector(B_min, Components*2);
    DimVector(B_max, Components*2);
    DimMatrix(V, Components*2, Components*2); // ?

    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 2;

      B^[comp+0] := Component['Amp'].Data[0];
      B_min^[comp+0] := 1.0e-20;
      B_max^[comp+0] := 1.0e20;

      B^[comp+1] := Component['Tau'].Data[0];
      B_min^[comp+1] := -1.0e20; // ?
      B_max^[comp+1] := 1.0e20; // ?

      Component := Component.Sibling;
    end;

    SetFirstPoint(0); // you'd think they'd at least mention this kind of weirdness in the docs
    SetOptAlgo( TOptAlgo( Integer(State['algo'].Data[0]) ) );

    Histogram_CurveFit_Components := Components;
    err := NLFit(RegNegExpo, DerivNegExpo, X, Y, N-1, 0, 2*Components-1, MaxIter, {Precision}0.01, B, B_min, B_max, V);
    State['Err'].SetData( err );

    Component := State['Params'].Child;
    for i := 0 to Components - 1 do begin
      comp := i * 2;
      Component['Amp'].Data[0]  := B[comp+0];
      Component['Tau'].Data[0]  := B[comp+1];

      Component := Component.Sibling;
    end;

    //RegTest(Y, Yout, N-1, 0, 2*Components-1, V, FitStats);
    //State['Residual'].SetData( FitStats.Vr );
  except
  end;

  if X     <> nil then DelVector(X    , N);
  if Y     <> nil then DelVector(Y    , N);
  if B     <> nil then DelVector(B    , 2*Components);
  if B_min <> nil then DelVector(B_min, 2*Components);
  if B_max <> nil then DelVector(B_max, 2*Components);
  if V     <> nil then DelMatrix(V, 2*Components, 2*Components);
end;


//******************************************************************************
function OneNegExpo(X : Float; B : PVector; comp : Integer) : Float;
// B: Amp, Tau   --  repeating per component
var
  Tau    :Double;
begin
  Result   := 0.0;
  try
    comp   := comp * 2; // to become the start index of parameters in B
    Tau    := B^[comp + 1];
    Result := B^[comp+0] * (1.0 - exp( - X / Tau ));
  except
  end;
end;

function RegNegExpo(X : Float; B : PVector) : Float;
// B: Amp, Tau   --  repeating per component
var
  Tau           :Double;
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
begin
  Result     := 0;
  try
    for i := 0 to Histogram_CurveFit_Components - 1 do begin
      comp   := i * 2;
      Tau    := B^[comp + 1];
      Result := Result + B^[comp+0] * (1.0 - exp( - X / Tau ));
    end;
  except
  end;
end;

procedure DerivNegExpo(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);
var
  comp          :Integer; // start of this component's parameters in B
  i             :Integer; // component index
  expL          :Double;
  Amp           :Double;
  Tau           :Double;
begin
  for i := 0 to Histogram_CurveFit_Components - 1 do begin
    comp        := i * 2;
    try
      Amp       := B^[comp + 0];
      Tau       := B^[comp + 1];
      expL      := exp(- X / Tau);

      D^[comp+0] := 1.0 - expL;
      D^[comp+1] := - X * Amp / (Tau * Tau) * expL;
    except
      D^[comp+0] := -1.0e10;
      D^[comp+1] := -1.0e10;
    end;
  end;
end;


// *****************************************************************************

var
  gFitter              :OleVariant;
  gDerivTested         :Boolean;
  gFitterHasDeriv      :Boolean;
  // put a mutex here for good luck

procedure NLRegFit_CopyParams(Fitter :OleVariant; B :PVector);
var
  //ParamList            :OleVariant;
  ParamNodePy          :PPyObject;
  ParamNode            :INode;
  i, j, bi             :Integer;
begin
  bi                   := 0;
  with GetPythonEngine do begin
    for i := 0 to Fitter.ncomponent - 1 do begin
      // ParamList          := Fitter.getParamList(i);
      ParamNodePy        := VariantAsPyObject( Fitter.params[i].child );
      ParamNode          := PyNode_ToNode( ParamNodePy );
      Py_DECREF( ParamNodePy );
    
      // ParamNode          := Fitter.params[i].child;
      for j := 0 to Fitter.nparamToFit - 1 do begin
        // ParamList[j]     := B^[bi];
        ParamNode.Data[0]:= B^[bi];

        ParamNode        := ParamNode.Sibling;
        bi               := bi + 1;
      end;
    end;
    Fitter.updateParamLists;
  end;
end;

function NLRegFit_Func(X : Float; B : PVector) : Float;
var
  i                    :Integer;
begin
  if X = gFitter.X[0] then NLRegFit_CopyParams(gFitter, B);

  Result := 0.0;
  for i := 0 to gFitter.ncomponent - 1 do begin
    Result := Result + gFitter.func( gFitter.getParamList(i), X );
  end;
end;

procedure NLRegFit_Deriv(RegFunc : TRegFunc; X, Y : Float; B, D : PVector);
var
  OneGrad              :OleVariant;
  OneGradPy            :PPyObject;
  i, j, di             :Integer;
begin
  if gFitterHasDeriv then begin

    if X = gFitter.X[0] then NLRegFit_CopyParams(gFitter, B);

    with GetPythonEngine do begin
         di := 0;
      for i := 0 to gFitter.ncomponent - 1 do begin
        OneGrad            := gFitter.deriv( gFitter.getParamList(i), X );

        if not gDerivTested then begin
          gDerivTested      := True;
          OneGradPy         := VariantAsPyObject(OneGrad);
          gFitterHasDeriv   := PySequence_Length(OneGradPy) <> 0;
          Py_DECREF( OneGradPy );
          if not gFitterHasDeriv then Break;
        end;

        for j := 0 to gFitter.nparamToFit - 1 do begin
          D^[di] := OneGrad[j];
          di     := di + 1;
        end;
      end;
    end;

  end;
  if not gFitterHasDeriv then begin

    NumDeriv(RegFunc, X, Y, B, D);

  end;
end;

function TQUB_PyFunc_NLRegFit.Call( args, kw : PPyObject) : PPyObject;
var
  FitterPy              :PPyObject;
  Fitter                :OleVariant;
  MaxIter               :Integer;
  Precision             :Double;

  Component             :OleVariant;
  Param                 :OleVariant;
  CompCount             :Integer;
  ParamCount            :Integer;
  BCount                :Integer;
  XNode, YNode          :INode;
  NodePy                :PPyObject;
  Node                  :INode;

  N                     :Integer;
  X, Y                  :PVector;
  B, B_min, B_max       :PVector; // reg.func parameters and bounds
  V                     :PMatrix; // inverse matrix of normal eqns (output)
  i, j, bi              :Integer;
begin
  with GetPythonEngine do begin
    MaxIter        := DEF_MAXITER;
    Precision      := DEF_PRECISION;

    if PyArg_ParseTuple( args, 'O|id:QUB.NLRegFit', [@FitterPy, @MaxIter, @Precision] ) <> 0 then begin

      X            := nil;
      Y            := nil;
      B            := nil;
      B_min        := nil;
      B_max        := nil;
      V            := nil;

      Fitter       := getAtom(FitterPy);
      NodePy       := VariantAsPyObject( Fitter.data );
      Node         := PyNode_ToNode(NodePy);
      Py_DECREF(NodePy);
      XNode        := Node['X'];
      YNode        := Node['Y'];
      N            := Fitter.N;
      CompCount    := Fitter.ncomponent;
      ParamCount   := Fitter.nparamToFit;
      BCount       := CompCount * ParamCount;

      try
        DimVector(X, N);
        DimVector(Y, N);
        for i := 0 to N - 1 do begin
          X[i]     := XNode.Data[i];
          Y[i]     := YNode.Data[i];
        end;

        DimVector(B,     BCount);
        DimVector(B_min, BCount);
        DimVector(B_max, BCount);
        DimMatrix(V, BCount, BCount); // ?

        bi := 0; // index of next B
        for i := 0 to CompCount - 1 do begin
          Component   := Fitter.params[i];
          Param       := Component.child;
          j := 0;
          while (j < ParamCount) and (Integer(Param.isNull) = 0) do begin
            B^[bi]     := Param.data[0];
            try
              B_min^[bi] := Param.data[1];
            except
              B_min^[bi] := -1.0e20;
            end;
            try
              B_max^[bi] := Param.data[2];
            except
              B_max^[bi] := 1.0e20;
            end;
            j          := j + 1;
            bi         := bi + 1;
            Param      := Param.sibling;
          end;
        end;

        SetFirstPoint(0);
        SetOptAlgo( TOptAlgo(Integer(Fitter.algo)) );
        
        gFitter         := Fitter;
        gDerivTested    := False;
        gFitterHasDeriv := True;  // until proven false by an empty list returned from fitter.deriv()
        Result := PyInt_FromLong(NLFit(NLRegFit_Func, NLRegFit_Deriv, X, Y, N-1, 0, BCount-1, MaxIter, Precision, B, B_min, B_max, V));
        VarClear(gFitter);

        NLRegFit_CopyParams(Fitter, B);

      except
        on E: Exception do begin
          PyErr_SetString( PyExc_Exception^, PChar(E.Message) );
          Result        := nil;
        end;
      end;

      if X     <> nil then DelVector(X    , N);
      if Y     <> nil then DelVector(Y    , N);
      if B     <> nil then DelVector(B    , BCount);
      if B_min <> nil then DelVector(B_min, BCount);
      if B_max <> nil then DelVector(B_max, BCount);
      if V     <> nil then DelMatrix(V, BCount, BCount);

    end else begin
      Result := nil;
    end;
  end;
end;


// *****************************************************************************
var
  gNLRegFitType                 :TPythonType;

procedure InitCurveFits;
var
  PyFuncInst                    :PPyObject;
begin
  gNLRegFitType := TPythonType.Create(nil);
  gNLRegFitType.Engine := GetPythonEngine;
  gNLRegFitType.Module := PythonViewport.pmQUBTypes;
  gNLRegFitType.TypeName := 'NLRegFit Function';
  gNLRegFitType.Services.Basic := [bsCall];
  gNLRegFitType.PyObjectClass := TQUB_PyFunc_NLRegFit;
  gNLRegFitType.Initialize;

  PyFuncInst := gNLRegFitType.CreateInstance;
  PyNode_PublishPyObj( PyFuncInst, 'NLRegFit' );
  GetPythonEngine.Py_DECREF( PyFuncInst );

  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitGauss_Eval ), 'FitGauss_Eval' );
  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitGauss_Fit ), 'FitGauss_Fit' );

  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitLogExpo_Eval ), 'FitLogExpo_Eval' );
  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitLogExpo_Fit ), 'FitLogExpo_Fit' );

  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitExpo_Eval ), 'FitExpo_Eval' );
  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitExpo_Fit ), 'FitExpo_Fit' );

  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitNegExpo_Eval ), 'FitNegExpo_Eval' );
  PyNode_PublishNodeNodeFunc( TNodeNodeFunc.Create( FitNegExpo_Fit ), 'FitNegExpo_Fit' );
end;


// *****************************************************************************
end.
