{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// called by QUB_Viewport_Histogram to do the actual work


unit QUB_Task_Histogram;


interface


uses


Windows, Messages, SysUtils, Classes, Math, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, Series, ComCtrls, Menus, Buttons, TeeProcs, TeEngine, Chart, Registry,
AllocTypes, Alloc,

QUB_Python_Thread, QUB_Viewport_Python,

QUB_Console, QUB_Rich_Console,

QUB_Global,
QUB_Spectrum,
QUB_Node, QUB_Node_Python,
QUB_Thread, QUB_Task, QUB_Task_Data;


type


TQUB_Task_DataHist  = class(TQUB_Task_Data)
protected


  //Source Objects
  prSelectionList          :TObject;

  // inputs
  prBinCount               :Integer;
  prAutoScale              :Boolean;
  prMin                    :Double;
  prMax                    :Double;

  // outputs
  prBins                   :TPADouble;
  prBars                   :TPADouble;
  //Whether a bin has data or not. If not, it will be set to the minimum bin.
  prCount                  :TPAInteger;
  
  // init/get data
  prCurChunk               :Integer;
  prSetChunkCompletion     :Boolean;
  prSubtractAvg            :Boolean;
  prDataAvg                :Integer;
  prDataSel                :TQUB_FirstLast;


  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prDoOnHistogramDone;
  procedure prDoOnHistogramFailed;

  //Chunks
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prInitDataChunk;
  function prGetData(NextN :Integer; Buf :Pointer): Integer;


public


  OnHistogramDone          :TNotifyEvent;
  OnHistogramFailed        :TNotifyEvent;


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  property BinCount        :Integer         read prBinCount          write prBinCount;
  property AutoScale       :Boolean         read prAutoScale         write prAutoScale;
  property MinVal          :Double          read prMin               write prMin;
  property MaxVal          :Double          read prMax               write prMax;
  property Bins            :TPADouble       read prBins;
  property Bars            :TPADouble       read prBars;
  property Count           :TPAInteger      read prCount;


end;

TQUB_Task_AmpHist = class(TQUB_Task_DataHist)
protected


  procedure prDoTask; override;


public


  constructor Create(Suspended :Boolean); override;


end;

TQUB_Task_FFTHist  = class(TQUB_Task_DataHist)
protected


  prMaxWindowSize          :Integer;
  prWindowFunc             :Integer;
  prOverlap                :Integer;
  prXLog                   :Boolean;
  prBarsOrLine             :Boolean;


  procedure prDoTask; override;
  procedure prBinSpectrum(SpecAvg :TPADouble; WindowSize :Integer);


public


  constructor Create(Suspended :Boolean); override;


  property MaxWindowSize   :Integer         read prMaxWindowSize     write prMaxWindowSize;
  property WindowFunc      :Integer         read prWindowFunc        write prWindowFunc;
  property Overlap         :Integer         read prOverlap           write prOverlap;
  property XLog            :Boolean         read prXLog              write prXLog;
  property BarsOrLine      :Boolean         read prBarsOrLine        write prBarsOrLine;


end;


TQUB_Task_DurHist  = class(TQUB_Task)
protected


  // inputs
  prBinCount               :Integer;
  prAutoScale              :Boolean;
  prMin                    :Double;
  prMax                    :Double;
  prClass                  :Integer;
  prXLog                   :Boolean;

  prData                   :INode;

  // outputs
  prBins                   :TPADouble;
  prBars                   :TPADouble;


  procedure prDoOnHistogramDone;
  procedure prDoOnHistogramFailed;


public


  OnHistogramDone          :TNotifyEvent;
  OnHistogramFailed        :TNotifyEvent;


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Execute; override;

  // abstract bullshit
  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  property BinCount        :Integer         read prBinCount          write prBinCount;
  property AutoScale       :Boolean         read prAutoScale         write prAutoScale;
  property MinVal          :Double          read prMin               write prMin;
  property MaxVal          :Double          read prMax               write prMax;
  property Bins            :TPADouble       read prBins;
  property Bars            :TPADouble       read prBars;
  property Cls             :Integer         read prClass             write prClass;
  property XLog            :Boolean         read prXLog              write prXLog;
  property Data            :INode           read prData              write prData;     // please make this a clone if you're keeping the original


end;


const


FFTHIST_WINDOW_RECT      = 0;
FFTHIST_WINDOW_HANNING   = 1;
FFTHIST_WINDOW_HAMMING   = 2;

FFTHIST_MIN_WINDOWPOWER  = 4;


implementation


uses


QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QFS_Channels,
QUB_GUIInterface;


const


TASKHIST_MAX_BUFCOUNT = 1000000;


constructor TQUB_Task_DataHist.Create(Suspended :Boolean);
begin
  inherited;

  prBins                       := nil;
  prBars                       := nil;
  prCount                      := nil;
  prSetChunkCompletion         := False;
  prSubtractAvg                := False;

  OnChange := TaskView.OnTaskChange;

end;

destructor TQUB_Task_DataHist.Destroy;
begin
  try
    FreeMem(prBins);
    FreeMem(prBars);
    FreeMem(prCount);
  finally
    inherited;
  end;
end;

procedure TQUB_Task_DataHist.prDoTask;
begin
  prCurChunk                   := 0;
  prInitDataChunk;
end;

procedure TQUB_Task_DataHist.prDoExecuteStart;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        TQUB_QVO_QVD_SelectionList(prSelectionList).Lock;
      end;
    end;
  end;

  inherited;
end;

procedure TQUB_Task_DataHist.prDoExecuteEnd;
begin

  inherited;

  if prSelectionList <> nil then begin
    TQUB_QVO_QVD_SelectionList(prSelectionList).UnLock;
  end;

end;


function TQUB_Task_DataHist.prDoExecuteInitialize :Boolean;
begin
  Result    := False;
  try
    prBins  := DAlloc1(BinCount);
    prBars  := DAlloc1(BinCount);
    prCount := IAlloc1(BinCount);
  except
    Result  := True;
  end;
end;

procedure TQUB_Task_DataHist.prDoOnHistogramDone;
begin
  if Assigned(OnHistogramDone) then OnHistogramDone(Self);
end;

procedure TQUB_Task_DataHist.prDoOnHistogramFailed;
begin
  if Assigned(OnHistogramFailed) then OnHistogramFailed(Self);
end;

procedure TQUB_Task_DataHist.prFreeChunk(Chunk :TQUB_Thread_PChunk);
begin

  FreeMem(Chunk);

end;

function TQUB_Task_DataHist.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Chunk));

  Result.Sel := Sel;

end;

procedure TQUB_Task_DataHist.prInitDataChunk;
var
  Buf                    :TPAInteger;
  BufCount               :Integer;
  Sel                    :TQUB_FirstLast;
  Chunk                  :TQUB_Thread_PChunk;
  ChannelIndex           :Integer;
  DataSum                :Int64;
  i                      :Integer;
begin
  Chunk                := TQUB_Thread_PChunk( prChunks[ prCurChunk ] );
  Sel                  := Chunk.Sel;

  ChannelIndex           := TQUB_QVO_QVD(Data).ActiveChannel;
  Buf                    := nil;
  DataSum                := 0;

  try
    BufCount             := Min( TASKHIST_MAX_BUFCOUNT, Sel.Last - Sel.First + 1 );

    while (Buf = nil) do begin
      try
        Buf              := IAlloc1( BufCount );
      except
        if BufCount < 1000 then raise;
        BufCount         := BufCount div 2;
      end;
    end;

    Sel.Last             := Sel.First + BufCount - 1;
    while Sel.First <= Chunk.Sel.Last do begin
      TQUB_QVO_QVD(Data).LoadDataChannelProcess(Buf, ChannelIndex, Sel, TQUB_QVO_QVD(Data).FilterOn[ChannelIndex], TQUB_QVO_QVD(Data).FilterFreq[ChannelIndex], TQUB_QVO_QVD(Data).BaselineOn[ChannelIndex]);
      for i := Sel.First to Sel.Last do begin
        DataSum          := DataSum + Buf[i - Sel.First];
      end;

      Sel.First          := Sel.Last + 1;
      Sel.Last           := Min( Sel.First + BufCount - 1, Chunk.Sel.Last );
    end;

  except
  end;

  prDataAvg              := DataSum div (Chunk.Sel.Last - Chunk.Sel.First + 1);
  prDataSel              := Chunk.Sel;

  FreeMem(Buf);
end;

function TQUB_Task_DataHist.prGetData(NextN :Integer; Buf :Pointer): Integer;
var
  IBuf                   :TPAInteger;
  DBuf                   :TPADouble;
  ChannelIndex           :Integer;
  Sel                    :TQUB_FirstLast;
  Chunk                  :TQUB_Thread_PChunk;
  i                      :Integer;
  s, o                   :Double;
begin
  DBuf                   := TPADouble(Buf);
  IBuf                   := TPAInteger( @(DBuf[ NextN div (SizeOf(Double) div SizeOf(Integer)) ]) );
  ChannelIndex           := TQUB_QVO_QVD(Data).ActiveChannel;

  Chunk                  := TQUB_Thread_PChunk( prChunks[prCurChunk] );
  Sel                    := prDataSel;
  Sel.Last               := Min( Sel.Last, Sel.First + NextN - 1 );

  TQUB_QVO_QVD(Data).LoadDataChannelProcess(IBuf, ChannelIndex, Sel, TQUB_QVO_QVD(Data).FilterOn[ChannelIndex], TQUB_QVO_QVD(Data).FilterFreq[ChannelIndex], TQUB_QVO_QVD(Data).BaselineOn[ChannelIndex]);

  Result                 := Sel.Last - Sel.First + 1;
  prDataSel.First        := Sel.Last + 1;

  s                      := TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex];
  o                      := TQUB_QVO_QVD(Data).DataChannelOffset[ChannelIndex];
  if prSubtractAvg then begin
    for i := 0 to Result - 1 do begin
      DBuf[i]              := (IBuf[i] - prDataAvg) / s + o;
    end;
  end else begin
    for i := 0 to Result - 1 do begin
      DBuf[i]              := IBuf[i] / s + o;
    end;
  end;

  if prSetChunkCompletion then begin
    Chunk.Completion     := Round(100 * (prDataSel.First - Chunk.Sel.First) / (Chunk.Sel.Last - Chunk.Sel.First + 1));
    prCheckStatus;
  end;

  if prDataSel.First > prDataSel.Last then begin
    prCurChunk           := prCurChunk + 1;
    if prCurChunk < prChunks.Count then prInitDataChunk;
  end;
end;

procedure TQUB_Task_DataHist.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_DataHist.OnPutProperties(Sender :TObject);
begin
end;


// *****************************************************************************
constructor TQUB_Task_AmpHist.Create(Suspended :Boolean);
begin
  inherited;

  prName        := 'Amp Histogram';

end;

procedure TQUB_Task_AmpHist.prDoTask;
var
  // maintained by FirstBuf, NextBuf
  Buf                    :TPADouble;
  BufCount               :Integer;
  BufCountUsed           :Integer;
  MoreBufs               :Boolean;

  // maintained by main body
  Sel                    :TQUB_FirstLast;
  MinMax                 :Double;
  BinSize                :Double;
  p                      :Double;
  i, b                   :Integer;

  procedure NextBuf;
  begin
    if not MoreBufs then Exit;

    if prCurChunk < prChunks.Count then begin
      BufCountUsed         := prGetData(BufCount, Buf);
    end else begin
      MoreBufs             := False;
    end;
  end;

  procedure FirstBuf;
  begin
    MoreBufs             := True;
    prCurChunk           := 0;
    prInitDataChunk;
    NextBuf;
  end;

begin
  if Data = nil then Exit;

  inherited;

  try
    {
    prConsoleMsg  := 'Amp histogram running ... ';
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);
    }
    Buf                  := nil;

    try
      BufCount           := 0;
      for i := 0 to prChunks.Count - 1 do begin
        Sel := TQUB_Thread_PChunk( prChunks[i] ).Sel;
        b   := Sel.Last - Sel.First + 1;
        if b > BufCount then begin
          BufCount := b;
        end;
      end;
      if BufCount > TASKHIST_MAX_BUFCOUNT then BufCount := TASKHIST_MAX_BUFCOUNT;

      while (Buf = nil) do begin
        try
          Buf                := DAlloc1( BufCount );
        except
          if BufCount < 1000 then raise;
          BufCount           := BufCount div 2;
        end;
      end;

      // either use min, max as is; or loop through chunks to compute it
      if AutoScale then begin
        FirstBuf;
        prMin := Buf[0];
        prMax := prMin;
        while MoreBufs and not prStopped do begin
          for i := 0 to BufCountUsed - 1 do begin
            p := Buf[i];
            if p < prMin then begin
              prMin := p;
            end else if p > prMax then begin
              prMax := p;
            end;
          end;
          NextBuf;
        end;
      end;

      if prMax <= prMin then prMax := prMin + 1e-30;
      MinMax  := prMax - prMin;
      BinSize := MinMax / BinCount;

      for i := 0 to BinCount - 1 do begin
        Bins[i] := prMin + i*BinSize + BinSize/2;
        Bars[i] := 0;
      end;

      prSetChunkCompletion := True;

      // loop through chunks to increment bars
      FirstBuf;
      while MoreBufs and not prStopped do begin
        for i := 0 to BufCountUsed - 1 do begin
          p := Buf[i];
          if (p >= prMin) and (p <= prMax) then begin
            p := p - prMin;
            b := Min(Floor(p / BinSize), BinCount - 1);
            Bars[b] := Bars[b] + 1;
          end;
        end;
        NextBuf;
      end;

    finally
      FreeMem( Buf );
    end;


    for i := 0 to prChunks.Count - 1 do begin
      if TQUB_Thread_PChunk(prChunks[i]).Status = CHUNK_STATUS_EMPTY then begin
        TQUB_Thread_PChunk(prChunks[i]).Status := CHUNK_STATUS_DONE;
      end;
    end;

    if prStopped then begin
      Synchronize( prDoOnHistogramFailed );
    end else begin
      Synchronize( prDoOnHistogramDone );
    end;
  except
    prError    := True;
    prErrorMsg := 'Unknown error in AMP histogram';
    prWikiTopic := 'Error:Task';

    Synchronize( prDoOnHistogramFailed );
  end;

end;

// *****************************************************************************
constructor TQUB_Task_FFTHist.Create(Suspended :Boolean);
begin
  inherited;

  prName        := 'FFT Histogram';

end;

procedure TQUB_Task_FFTHist.prDoTask;
var
  SpecAvg                :TPADouble;
  ChunkSpecAvg           :TPADouble;
  Chunk                  :TQUB_Thread_PChunk;
  MinWindowSize          :Integer;
  WindowSize             :Integer;
  WindowFunction         :TFFT_WindowFunc;
  DataCount              :Integer;
  FrameCount             :Integer;
  TotalFrameCount        :Integer;
  i                      :Integer;
begin
  if Data = nil then Exit;

  inherited;

  try

    // MinWindowSize       := Min( MaxWindowSize, Max(chunksizes) );
    Chunk                  := TQUB_Thread_PChunk( prChunks[0] );
    MinWindowSize          := Chunk.Sel.Last - Chunk.Sel.First + 1;
    for i := 1 to prChunks.Count - 1 do begin
      Chunk                := TQUB_Thread_PChunk( prChunks[i] );
      MinWindowSize        := Max( MinWindowSize, Chunk.Sel.Last - Chunk.Sel.First + 1 );
    end;
    MinWindowSize          := Min( MinWindowSize, MaxWindowSize );

    // WindowSize          := smallest pow of 2 >= WindowSize;
    WindowSize             := FFTHIST_MIN_WINDOWPOWER;
    while Power2(WindowSize) < MinWindowSize do begin
      WindowSize           := WindowSize + 1;
    end;
    WindowSize             := Power2(WindowSize);

    TotalFrameCount        := 0;

    if WindowFunc = FFTHIST_WINDOW_HAMMING then begin
      WindowFunction       := FFT_HammingWindow;
    end else if WindowFunc = FFTHIST_WINDOW_HANNING then begin
      WindowFunction       := FFT_HanningWindow;
    end else begin
      WindowFunction       := FFT_RectWindow;
    end;

    prSetChunkCompletion   := True;
    prSubtractAvg          := True;

    SpecAvg                := nil;
    ChunkSpecAvg           := nil;

    try

      SpecAvg            := DAlloc1(WindowSize);
      ChunkSpecAvg       := DAlloc1(WindowSize);

      for i := 0 to WindowSize - 1 do begin
        SpecAvg[i]       := 0.0;
      end;

      while (prCurChunk < prChunks.Count) and not prStopped do begin

        DataCount        := prDataSel.Last - prDataSel.First + 1;

        if DataCount <= WindowSize then begin
          FFTSpectrum( prGetData, DataCount, WindowSize, ChunkSpecAvg, prStopped );

          FrameCount     := 1;
        end else begin
          WindowAvgSpectrum( prGetData, DataCount,
                             WindowSize, WindowFunction, Overlap,
                             ChunkSpecAvg, FrameCount,
                             prStopped );
        end;

        TotalFrameCount  := TotalFrameCount + FrameCount;

        for i := 0 to WindowSize - 1 do begin
          SpecAvg[i]     := SpecAvg[i] + FrameCount * ChunkSpecAvg[i];
        end;

      end;

      if not prStopped then begin
        for i := 0 to WindowSize - 1 do begin
          SpecAvg[i]       := SpecAvg[i] / TotalFrameCount;
        end;

        prBinSpectrum(SpecAvg, WindowSize);
      end;

    finally
      FreeMem(SpecAvg);
      FreeMem(ChunkSpecAvg);
    end;

    for i := 0 to prChunks.Count - 1 do begin
      if TQUB_Thread_PChunk(prChunks[i]).Status = CHUNK_STATUS_EMPTY then begin
        TQUB_Thread_PChunk(prChunks[i]).Status := CHUNK_STATUS_DONE;
      end;
    end;

    if prStopped then begin
      Synchronize( prDoOnHistogramFailed );
    end else begin
      Synchronize( prDoOnHistogramDone );
    end;

  except
    on E: Exception do begin
      prError    := True;
      prErrorMsg := E.Message;
      prWikiTopic := 'Error:Task';

      Synchronize( prDoOnHistogramFailed );
    end;
  end;
end;

procedure TQUB_Task_FFTHist.prBinSpectrum(SpecAvg :TPADouble; WindowSize :Integer);
var
  m                  :Double;
  ri                 :Integer;
  i, j               :Integer;
  r                  :Double;
  mn, mx             :Double;
  f, f1, f2, fs      :Integer;
  fr1, fr2, frm, frd :Double;
begin
  m := TQUB_QVO_QVD(Data).Sampling;

  //Why?! LM.
  // if the requested bin count is more than the number of freq bins in the fft,
  //   make it exactly the number of freq bins
  if BinCount > (WindowSize div 2) then begin
    BinCount := WindowSize div 2;
  end;

  mn := 0;
  mx := 0;

  if XLog then begin

    // 'log10 (Frequency [Hz])';

    if prAutoScale then begin
      fr1 := 2 / (WindowSize * m);
      fr2 := (WindowSize div 2) / (WindowSize * m);
    end else begin
      fr1 := prMin;
      fr2 := prMax;
    end;

    if fr1 = 0 then begin
      fr1 := 1e-20;
    end;
    if fr2 = 0 then begin
      fr2 := 1e-20;
    end;

    //Logarithmic boundaries
    fr1 := log10(fr1);
    fr2 := log10(fr2);
    frd := (fr2 - fr1) / BinCount;

    //First frame
    f := 2;
    for i := 0 to BinCount - 1 do begin
      r  := 0;
      ri := 0;
      frm := fr1 + i * frd + frd;
      fs := f;
      repeat
        if (log10(f / (WindowSize * m)) > frm) then begin
          Break;
        end;
        r := r + 2*Sqrt(SpecAvg[f-1]);
        // r := r + 2 * SpecAvg[f-1];
        Inc(ri);
        Inc(f);
        if (f > (WindowSize div 2)) then begin
          Break;
        end;
      until(False);
      Bins[i]     := frm - (frd / 2);
      //r := r / Sqrt(WindowSize);
      //Bars[i]     := Log10(r * WindowSize * m / (f - fs));
      if (r = 0) or (f = fs) then begin
        Bars[i]   := 123456789;
      end else begin
        Bars[i]   := Log10(r * Sqrt(WindowSize) * m / (f - fs));
        if Bars[i] < mn then mn := Bars[i];
        if Bars[i] > mx then mx := Bars[i];
      end;
      Count[i]    := ri;
      //if ri > 0 then begin
      //  r := r / ri;
      //  if r = 0 then begin
      //    r := 1e-20;
      //  end;
      //  Bars[i]   := log10(r);
      //  Count[i]  := ri;
      //end;
    end;

  end else begin

    // 'Frequency [Hz]';

    for i := 0 to BinCount - 1 do begin
      r  := 0;
      ri := 0;

      //if prFFTAutoscale then begin

        f1 := Round(i * ((WindowSize div 2) / BinCount) + 1);
        f2 := Round(i * ((WindowSize div 2) / BinCount) + ((WindowSize div 2) / BinCount));

      //end else begin

        //???

      //end;
      f  := (f1 - 1) + (f2 - f1) div 2;
      for j := f1 to f2 do begin
        //r := r + 2 * SpecAvg[j-1];
        r := r + 2*Sqrt(SpecAvg[j-1]);
        if j > 1 then begin
          Inc(ri);
        end;
      end;
      Bins[i]     := f / (WindowSize * m);
      if r = 0 then begin
        Bars[i] := 123456789;
      end else begin
        r := r / Sqrt(WindowSize); // QUB_Spectrum divides by sqrt(N) for fwd and inverse transform
        Bars[i]     := log10(r * WindowSize * m / (f2 - f1 + 1)); // sum(pA) / width-Hertz
        if Bars[i] < mn then mn := Bars[i];
        if Bars[i] > mx then mx := Bars[i];
      end;
      Count[i]    := ri;
      //if ri > 0 then begin
      //  r := r / ri;
      //  if r = 0 then begin
      //    r := 1e-20;
      //  end;
      //  Bars[i]   := log10(r);
      //  Count[i]  := ri;
      //end;
    end;

  end;

  r := mn - (mx - mn) / 7;
  if r = mn then r := mn - 1;
  for i := 0 to BinCount - 1 do begin
    if Bars[i] = 123456789 then Bars[i] := r;
  end;

end;


// *****************************************************************************
constructor TQUB_Task_DurHist.Create(Suspended :Boolean);
begin
  inherited;

  prName        := 'Dur Histogram';
  prBins                       := nil;
  prBars                       := nil;

  OnChange := TaskView.OnTaskChange;

end;

destructor TQUB_Task_DurHist.Destroy;
begin
  try
    FreeMem(prBins);
    FreeMem(prBars);
  finally
    inherited;
  end;
end;

procedure TQUB_Task_DurHist.Execute;
var
  BinCounts                :array of Integer;
  MinMax                   :Double;
  DS, Seg                  :INode;
  idwt                     :TPAInteger;
  tdwt                     :TPASingle;
  ndwt, i, j               :Integer;
  TotCount                 :Integer;
begin
  try
    try

      prBins               := DAlloc1( BinCount );
      prBars               := DAlloc1( BinCount );
      SetLength(BinCounts, BinCount);

      if prAutoScale then begin // find min and max

        prMin              := 1.0e30;
        prMax              := 0.0;

        DS                   := Data.Find('DataSet');
        while not (DS.IsNull or prStopped) do begin
          Seg                := DS.Find('Segment');
          while not (Seg.IsNull or prStopped) do begin
            ndwt           := Integer(Seg['DwellCount'].Data[0]);
            idwt           := TPAInteger(Seg['Classes'].DataPtr);
            tdwt           := TPASingle(Seg['Durations'].DataPtr);

            for i := 0 to ndwt - 1 do begin
              if idwt[i] = prClass then begin
                if tdwt[i] < prMin then prMin := tdwt[i];
                if tdwt[i] > prMax then prMax := tdwt[i];
              end;
            end;

            Seg              := Seg.SiblingSameName;
          end;

          DS                 := DS.SiblingSameName;
        end;

      end else begin;
        prMin                := prMin * 1.0e3;
        prMax                := prMax * 1.0e3;
      end;

      prMax                  := prMax + 1.0e-5; // so we can use Floor later and still catch the max dwell
      prMin                  := prMin - 1.0e-5;

      if prXLog then begin
        prMin              := log10(prMin);
        prMax              := log10(prMax);

        // convert dwells to log10
        DS                   := Data.Find('DataSet');
        while not (DS.IsNull or prStopped) do begin
          Seg                := DS.Find('Segment');
          while not (Seg.IsNull or prStopped) do begin
            ndwt           := Integer(Seg['DwellCount'].Data[0]);
            tdwt           := TPASingle(Seg['Durations'].DataPtr);

            for i := 0 to ndwt - 1 do begin
              tdwt[i]      := log10(tdwt[i]);
            end;

            Seg              := Seg.SiblingSameName;
          end;

          DS                 := DS.SiblingSameName;
        end;
      end;

      MinMax := (prMax - prMin) / BinCount;

      for j := 0 to BinCount - 1 do begin
        Bins[j]            := prMin + (j + 0.5) * MinMax;
        BinCounts[j]       := 0;
      end;

      // count events in class
      DS                   := Data.Find('DataSet');
      while not (DS.IsNull or prStopped) do begin
        Seg                := DS.Find('Segment');
        while not (Seg.IsNull or prStopped) do begin
          ndwt           := Integer(Seg['DwellCount'].Data[0]);
          idwt           := TPAInteger(Seg['Classes'].DataPtr);
          tdwt           := TPASingle(Seg['Durations'].DataPtr);

          for i := 0 to ndwt - 1 do begin
            if idwt[i] = prClass then begin
              j := Floor((tdwt[i] - prMin) / MinMax);
              if (j >= 0) and (j < BinCount) then begin
                Inc(BinCounts[j]);
              end;
            end;
          end;

          Seg              := Seg.SiblingSameName;
        end;

        DS                 := DS.SiblingSameName;
      end;

      if not prStopped then begin
        // compute bars
        TotCount             := 0;
        for j := 0 to BinCount - 1 do begin
          TotCount           := TotCount + BinCounts[j];
        end;

        if TotCount = 0 then raise Exception.Create('No idealized data.  Check the Data Source.');

        for j := 0 to BinCount - 1 do begin
          Bars[j]            := BinCounts[j] / TotCount; // Sqrt(BinCounts[j] / TotCount);
        end;

        Synchronize( prDoOnHistogramDone );
      end else begin
        Synchronize( prDoOnHistogramFailed );
      end;

    except
      on E: Exception do begin
        prError    := True;
        prErrorMsg := E.Message;
        prWikiTopic := 'Error:Task';

        Synchronize( prDoOnHistogramFailed );
      end;
    end;

  finally

    prUpdateStatus(100, prStTime);

    if not prError then begin

      if prStopped then begin
        prStatus := 'Stopping';
      end else begin
        prStatus := 'Terminating';
      end;
      Synchronize(prDoOnChange);
      
    end else if prError then begin

      prStatus := 'Error';
      Synchronize(prDoOnChange);

      Synchronize(ShowMessage);

    end;

  end;
end;

procedure TQUB_Task_DurHist.prDoOnHistogramDone;
begin
  if Assigned(OnHistogramDone) then OnHistogramDone(Self);
end;

procedure TQUB_Task_DurHist.prDoOnHistogramFailed;
begin
  if Assigned(OnHistogramFailed) then OnHistogramFailed(Self);
end;

procedure TQUB_Task_DurHist.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_DurHist.OnPutProperties(Sender :TObject);
begin
end;


// *****************************************************************************
end.
