unit QUB_Task_IdealizeMT;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

{
            Multi-threading Data Idealization
see also QUB_Thread_Idealize
}


interface


uses


Windows, SysUtils, Classes, SyncObjs, Graphics,
AllocTypes, Alloc,

QUB_Global,
QUB_Task, QUB_Task_Data, QUB_Task_DataMT,
QUB_Task_Idealize_Properties,
QUB_Thread, QUB_Thread_Idealize,
QUB_Thread_Socket, QUB_Thread_Socket_Idealize,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QVO_QVM,
QUB_QFS_Channels, QUB_QFS_QMF,
QUB_Idealize_Global,
QUB_Node, QUB_Node_Python, QUB_Node_Model,
Chart, Series;


type


TQUB_Task_IdealizeMT  = class(TQUB_Task_DataMT)
protected


  //Source Objects
  prModel                  :TQUB_QVO_QVM;

  //Parameters
  prParams                 :TQUB_Idealize_Params;
  prConfig                 :INode;
  prDatasetTemplate        :INode;

  //SKM
  prStatResult             :INode;
  prIdealResult            :INode;


  procedure prSetChunkParameters; override;

  procedure prDoExecute; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;
  procedure prDoExecuteFinalize; override;

  procedure prResultsToData; override;
  procedure prPublishResults;

  procedure prEventLoadData(Event :TQUB_Thread_Event_PRec); override;
  procedure prEventMessage(Event :TQUB_Thread_Event_PRec); override;

  //Thread
  function  prNewThread :TQUB_Thread; override;
  function  prNewSocket(const RemoteServer :String; Port :Integer) :TQUB_Thread_Socket; override;

  procedure prReadThreadOutput(Thread :TQUB_Thread); override;
  procedure prReadSocketOutput(Socket :TQUB_Thread_Socket); override;

  //Chunks
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prGetAmpsAndClss(Model :TQUB_QVO_QVM);

  procedure prSetModel(Model :TQUB_QVO_QVM);


public


  FillDataCallback           :TNodeMethPtr;


  constructor Create(Suspended :Boolean); override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  procedure Stop; override;


  //Model
  property Model             :TQUB_QVO_QVM             read prModel             write prSetModel;

  //Parameters
  property Params            :TQUB_Idealize_Params     read prParams            write prParams;


end;


implementation


uses

QUB_QVO_QVR,
QUB_Viewport_Results,
QUB_GUIInterface,
QUB_Viewport_Python,
QUB_Main;



//****************************************************************************//
//                                                                            //
//                      T Q U B _ T a s k I d e a l i ze M T                  //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_IdealizeMT.Create(Suspended :Boolean);
begin

  inherited;

  prName        := 'Idealize data';

  prParams.HalfAmp.Classes   := nil;
  prParams.HalfAmp.HalfAmps  := nil;
  prParams.HalfAmp.Amps      := nil;
  prParams.HalfAmp.IdxAmps   := nil;
  prParams.HalfAmp.AmpVals   := nil;
  prParams.HalfAmp.StdVals   := nil;

  prParams.SKM.FixA          := nil;
  prParams.SKM.FixS          := nil;
end;


procedure TQUB_Task_IdealizeMT.prSetChunkParameters;
begin
  inherited;
  prChunkSize         := 0;
end;


//******************************************************************************
procedure TQUB_Task_IdealizeMT.Stop;
begin

  prConfig['StopFlag'].Data[0] := 1;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdealizeMT.prDoExecute;
begin
  prConsoleMsg  := 'Idealize running ... ';
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
  inherited;
end;


procedure TQUB_Task_IdealizeMT.prDoExecuteEnd;
begin
  if Model <> nil then Model.UnlockAny;
  inherited;
end;

function TQUB_Task_IdealizeMT.prDoExecuteInitialize :Boolean;
var
  ProcDataNode                   :INode;
begin

  Result        := False;

  prConfig.AppendChild( GetModelAsNode );
  prConfig['StopFlag'].SetData( 0 );
  prConfig['ReestimateAmpSD'].SetData( 1 );
  prConfig['ApplyDeadTime'].SetData( Integer(prParams.ApplyDeadTime) );

  prDatasetTemplate := TNode.Create('DataSet');
  prDatasetTemplate['FileName'].SetData( Data.FileName );
  prDatasetTemplate['sampling'].SetData( TQUB_QVO_QVD(Data).Sampling * 1.0e3 );
  prDatasetTemplate['scaling'].SetData( TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex] );
  prDatasetTemplate['offset'].SetData( TQUB_QVO_QVD(Data).DataChannelOffset[ChannelIndex] );
  prDatasetTemplate['ADChannelCount'].SetData( TQUB_QVO_QVD(Data).DataObject.ADChannelCount );
  prDatasetTemplate['ActiveChannel'].SetData( TQUB_QVO_QVD(Data).ActiveChannel );
  prDatasetTemplate.AppendChild( TQUB_QVO_QVD(Data).ExpCond.ToTree );

  ProcDataNode := prConfig.Find('ProcessData');
  if not ProcDataNode.IsNull then prDatasetTemplate.AppendChild( ProcDataNode );
  if ProcDataNode.DataAsString = 'None' then begin
    ProcDataNode['Filter'].Data[0]     := 0;
    ProcDataNode['Baseline'].Data[0]   := 0;
  end else if ProcDataNode.DataAsString = 'AsDisplayed' then begin
    ProcDataNode['Filter'].Data[0]     := TQUB_QVO_QVD(Data).FilterOn[ChannelIndex];
    ProcDataNode['FilterFreq'].Data[0] := TQUB_QVO_QVD(Data).FilterFreq[ChannelIndex];
    ProcDataNode['Baseline'].Data[0]   := TQUB_QVO_QVD(Data).BaselineOn[ChannelIndex];
  end;

  if prParams.Method = METHOD_HALFAMP then begin

    prRemote := False;

    prGetAmpsAndClss(Model);

  end else begin

    prRemote := False;

  end;

  prParams.ModelObject       := Model.ModelObject;

  prParams.Scaling  := TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex];
  prParams.Sampling := TQUB_QVO_QVD(Data).Sampling;
  prParams.Offset   := TQUB_QVO_QVD(Data).DataChannelOffset[ChannelIndex];

end;

procedure TQUB_Task_IdealizeMT.prDoExecuteFinalize;
begin
  FreeMem( prParams.HalfAmp.Classes );        prParams.HalfAmp.Classes   := nil;
  FreeMem( prParams.HalfAmp.HalfAmps );       prParams.HalfAmp.HalfAmps  := nil;
  FreeMem( prParams.HalfAmp.Amps );           prParams.HalfAmp.Amps      := nil;
  FreeMem( prParams.HalfAmp.IdxAmps );        prParams.HalfAmp.IdxAmps   := nil;
  FreeMem( prParams.HalfAmp.AmpVals );        prParams.HalfAmp.AmpVals   := nil;
  FreeMem( prParams.HalfAmp.StdVals );        prParams.HalfAmp.StdVals   := nil;

  FreeMem( prParams.SKM.FixA );               prParams.SKM.FixA          := nil;
  FreeMem( prParams.SKM.FixS );               prParams.SKM.FixS          := nil;
end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdealizeMT.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  IdlChunk :TQUB_Thread_Idealize_PChunk;
begin

  IdlChunk := TQUB_Thread_Idealize_PChunk(Chunk);

  FreeMem(IdlChunk.Data);             IdlChunk.Data        := nil;

  FreeMem(IdlChunk.SegFirsts);        IdlChunk.SegFirsts   := nil;
  FreeMem(IdlChunk.SegLasts);         IdlChunk.SegLasts    := nil;
  FreeMem(IdlChunk.SegClasses);       IdlChunk.SegClasses  := nil;

  IdlChunk.ClassCount := 0;

  FreeMem(IdlChunk.Amps);             IdlChunk.Amps        := nil;
  FreeMem(IdlChunk.Stds);             IdlChunk.Stds        := nil;
  FreeMem(IdlChunk.Occupancy);        IdlChunk.Occupancy   := nil;
  FreeMem(IdlChunk.Lifetime);         IdlChunk.Lifetime    := nil;
  FreeMem(IdlChunk.NEvents);          IdlChunk.NEvents     := nil;
  Free2(TPMChar(IdlChunk.AMatrix));   IdlChunk.AMatrix     := nil;

  Free2(TPMChar(IdlChunk.Histogram)); IdlChunk.Histogram   := nil;

  IdlChunk.ReportSize := 0;

  FreeMem(IdlChunk.Report);           IdlChunk.Report      := nil;

  if IdlChunk.ResultNode <> nil then begin
    Node_DECREF( IdlChunk.ResultNode );
    IdlChunk.ResultNode  := nil;
  end;

  FreeMem(IdlChunk);

end;

function TQUB_Task_IdealizeMT.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
var
  Chunk      :TQUB_Thread_PChunk;
  IdlChunk   :TQUB_Thread_Idealize_PChunk;
  First      :Integer;
  s          :Integer;
begin

  Result := nil;

  IdlChunk := AllocMem(SizeOf(TQUB_Thread_Idealize_Chunk));

  if IdlChunk = nil then Exit;

  Chunk                 := TQUB_Thread_PChunk(IdlChunk);

  Chunk.StartTime       := 0;
  for s := 0 to TQUB_QVO_QVD(Data).DataSegmentCount - 1 do begin
    if Sel.First <= (TQUB_QVO_QVD(Data).DataBreakpoints[s] - 1) then begin
      if s = 0 then First := 0 else First := TQUB_QVO_QVD(Data).DataBreakpoints[s - 1];
      Chunk.StartTime := TQUB_QVO_QVD(Data).DataObject.Segments[s].StartTime + TQUB_QVO_QVD(Data).Sampling * (Sel.First - First);
      Break;
    end;
  end;

  IdlChunk.BaseTime    := TQUB_QVO_QVD(Data).DataObject.Segments[0].StartTime;
  IdlChunk.Sampling    := TQUB_QVO_QVD(Data).Sampling;
  IdlChunk.Scaling     := TQUB_QVO_QVD(Data).Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex];
  IdlChunk.Offset      := TQUB_QVO_QVD(Data).DataChannelOffset[ChannelIndex];

  //Clear Chunk

  IdlChunk.DataIs16Bit := 0;

  IdlChunk.SegCount    := 0;
  IdlChunk.SegFirsts   := nil;
  IdlChunk.SegLasts    := nil;
  IdlChunk.SegClasses  := nil;

  IdlChunk.ClassCount := 0;

  IdlChunk.Amps        := nil;
  IdlChunk.Stds        := nil;
  IdlChunk.Occupancy   := nil;
  IdlChunk.Lifetime    := nil;
  IdlChunk.NEvents     := nil;
  IdlChunk.AMatrix     := nil;

  IdlChunk.Histogram   := nil;

  IdlChunk.ReportSize := 0;

  IdlChunk.Report      := nil;

  IdlChunk.ResultNode  := nil;
  Result := Chunk;

end;


//****************************************************************************//
//                                                                            //
//                               T H R E A D S                                //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdealizeMT.prNewThread :TQUB_Thread;
var
  t :TQUB_Thread_Idealize;
begin

  t          := TQUB_Thread_Idealize.Create(False{Remote}, True{Suspended});

  t.Params   := Params;
  t.Config   := prConfig; // .Clone(True);  safe since they're all just reading it?
  t.DataSet  := prDatasetTemplate.Clone(True);
  t.OnEvent  := prOnThreadEvent;
  t.FillDataCallback := FillDataCallback;

  Result     := TQUB_Thread(t);

end;

function TQUB_Task_IdealizeMT.prNewSocket(const RemoteServer :String; Port :Integer) :TQUB_Thread_Socket;
var
  t :TQUB_Thread_Socket_Idealize;
begin

  t          := TQUB_Thread_Socket_Idealize.Create(RemoteServer, Port);

  t.Params   := Params;
  t.OnEvent  := prOnThreadEvent;

  Result     := TQUB_Thread_Socket(t);

end;


procedure TQUB_Task_IdealizeMT.prReadThreadOutput(Thread :TQUB_Thread);
var
  IdlChunk :TQUB_Thread_Idealize_PChunk;
  Chunk    :TQUB_Thread_PChunk;
  i        :Integer;
begin

  inherited;

  for i := 0 to Thread.ChunkCount - 1 do begin
    Chunk    := Thread.Chunk[i];
    IdlChunk := TQUB_Thread_Idealize_PChunk(Chunk);
    if IdlChunk.ReportSize <> 0 then begin
      SetLength(prConsoleMsg, IdlChunk.ReportSize - 4); // eliminate trailing CRLF CRLF
      Move(IdlChunk.Report^, PChar(prConsoleMsg)^, IdlChunk.ReportSize - 4);
      prConsoleMsg  := 'Segment ' + IntToStr(Chunk.Index + 1) + #13#10 + prConsoleMsg
                       + 'End of Segment ' + IntToStr(Chunk.Index + 1) + #13#10;
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);
    end;
  end;

end;

procedure TQUB_Task_IdealizeMT.prReadSocketOutput(Socket :TQUB_Thread_Socket);
var
  IdlChunk :TQUB_Thread_Idealize_PChunk;
  Chunk    :TQUB_Thread_PChunk;
begin

  inherited;

  Chunk    := Socket.Chunk;
  IdlChunk := TQUB_Thread_Idealize_PChunk(Chunk);
  if IdlChunk.ReportSize <> 0 then begin
    SetLength(prConsoleMsg, IdlChunk.ReportSize);
    Move(IdlChunk.Report^, PChar(prConsoleMsg)^, IdlChunk.ReportSize);
    prConsoleMsg  := 'Segment ' + IntToStr(Chunk.Index + 1) + #13#10;
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);
  end;

end;


//****************************************************************************//
//                                                                            //
//                                 E V E N T S                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdealizeMT.prEventLoadData(Event :TQUB_Thread_Event_PRec);
begin
  case DataProcessed of
    DATA_PROCESSED_NONE:
      TQUB_QVO_QVD(Data).LoadDataChannel(TQUB_Thread_Event_PRequestData(Event).Data, ChannelIndex,        TQUB_Thread_Event_PRequestData(Event).DataSel, False{Process}, False, False);
    DATA_PROCESSED_ASDISPLAYED:
      TQUB_QVO_QVD(Data).LoadDataChannelProcess(TQUB_Thread_Event_PRequestData(Event).Data, ChannelIndex, TQUB_Thread_Event_PRequestData(Event).DataSel, TQUB_QVO_QVD(Data).FilterOn[ChannelIndex], TQUB_QVO_QVD(Data).FilterFreq[ChannelIndex], TQUB_QVO_QVD(Data).BaselineOn[ChannelIndex]);
    DATA_PROCESSED_ASSUCH:
      TQUB_QVO_QVD(Data).LoadDataChannelProcess(TQUB_Thread_Event_PRequestData(Event).Data, ChannelIndex, TQUB_Thread_Event_PRequestData(Event).DataSel, ProcessFilter, ProcessFilterFreq, ProcessBaseline);
  end;
end;

procedure TQUB_Task_IdealizeMT.prEventMessage(Event :TQUB_Thread_Event_PRec);
begin
  SetLength(prConsoleMsg, TQUB_Thread_Event_PReportMessage(Event).MsgSize);
  Move(PChar(@(TQUB_Thread_Event_PReportMessage(Event).Msg))^, PChar(prConsoleMsg)^, TQUB_Thread_Event_PReportMessage(Event).MsgSize);
  prConsoleCode := TQUB_Thread_Event_PReportMessage(Event).MsgKind;
  Synchronize(prWriteNewConsole);
end;


//****************************************************************************//
//                                                                            //
//                               R E S U L T S                                //
//                                                                            //
//****************************************************************************//

procedure ComputeAvgAmps( ResultNode :INode );
var
  NClass, NPoint, SumNPoint, i        :Integer;
  Amp, Std, DS, Seg, SegAmp, SegStd   :Inode;
begin
  Amp             := ResultNode['amp'];
  Std             := ResultNode['sd'];
  NClass          := ResultNode['DataSet']['Segment']['amp'].DataCount;
  Amp.SetupNumData( NODE_DATATYPE_DOUBLE, NClass, 1 );
  Std.SetupNumData( NODE_DATATYPE_DOUBLE, NClass, 1 );
  for i := 0 to NClass - 1 do begin
    Amp.Data[i]   := 0.0;
    Std.Data[i]   := 0.0;
  end;

  SumNPoint       := 0;
  DS              := ResultNode.Find('DataSet');
  while not DS.IsNull do begin
    Seg           := DS.Find('Segment');
    while not Seg.IsNull do begin
      NPoint      := Integer(Seg.Data[1]) - Integer(Seg.Data[0]) + 1;
      SumNPoint   := SumNPoint + NPoint;

      SegAmp      := Seg['amp'];
      SegStd      := Seg['sd'];

      for i := 0 to NClass - 1 do begin
        Amp.Data[i] := Amp.Data[i] + NPoint * SegAmp.Data[i];
        Std.Data[i] := Std.Data[i] + NPoint * SegStd.Data[i] * SegStd.Data[i];
      end;

      Seg         := Seg.SiblingSameName;
    end;

    DS            := DS.SiblingSameName;
  end;

  for i := 0 to NClass - 1 do begin
    Amp.Data[i]   := Amp.Data[i] / SumNPoint;
    Std.Data[i]   := Sqrt( Std.Data[i] / SumNPoint );
  end;
end;

procedure TQUB_Task_IdealizeMT.prResultsToData;
var
  IdlChunk     :TQUB_Thread_Idealize_PChunk;
  Chunk        :TQUB_Thread_PChunk;
  Sel          :TQUB_FirstLast;
  i, a         :Integer;
  ClassCount   :Integer;
  Amps         :TPASingle;
  Stds         :TPASingle;
  Report       :String;
  DataSet      :INode;
  IdlSeg       :INode;
  ChunkResult  :INode;
  AmpNode      :INode;
  sdNode       :INode;
  Node         :INode;
  idwt         :TPAInteger;
  tdwt         :TPASingle;
  SkipTheReport :Boolean;
  LastSegIx    :Integer;
  LastChunkResult :INode;
  HistGroup    :INode;
  ActiveHist   :INode;
begin

  Amps         := nil;
  Stds         := nil;
  SkipTheReport := False;

  if prParams.Method = METHOD_SKM then begin
    prStatResult    := TNode.Create('SKM Result');
  end else begin
    prStatResult    := TNode.Create('HalfAmp Result');
  end;

  prStatResult['Date'].SetData( Format('%s %s',[DateTimeToStr( Date ),TimeToStr( Time )] ) );
  if prParams.Method = METHOD_HALFAMP then begin
    prStatResult['ResultType'].SetData( 'HalfAmp' );
  end else begin
    prStatResult['ResultType'].SetData( 'SKM' );
  end;

  prStatResult.AppendClone(prConfig, True);
  
  DataSet         := prDatasetTemplate.Clone(True);
  prStatResult.AppendChild( DataSet );

  prIdealResult  := TNode.Create('Idealization');
  prIdealResult.AppendChild( prDatasetTemplate );

  try


    //Clear Info ***************************************************************
    if Params.ClearIdl then begin
      prStatus := 'Clearing existing Idealization';
      Synchronize(prDoOnChange);
      TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].Clear;
    end;

    //Append Info
    prStatus := 'Applying Idealization';
    Synchronize(prDoOnChange);

    LastSegIx       := -1;
    LastChunkResult := TNode.CreateNull;

    for i := 0 to prChunks.Count - 1 do begin

      Chunk       := prChunks.List[i];
      IdlChunk    := TQUB_Thread_Idealize_PChunk(Chunk);

      if IdlChunk.ResultNode <> nil then begin
        ChunkResult := TNode.CreateUsing( IdlChunk.ResultNode );
      end else begin
        ChunkResult := TNode.Create('Segment');
      end;

      // split ChunkResult into IdlSeg (for data) and ChunkResult (for results)

      DataSet.AppendChild( ChunkResult );
      IdlSeg := ChunkResult.Clone(False);
      prDatasetTemplate.AppendChild( IdlSeg );

      IdlSeg.AppendChild( ChunkResult['start'].Clone(True) );
      IdlSeg.AppendChild( ChunkResult['DwellCount'].Clone(True) );
      IdlSeg.AppendChild( ChunkResult['Classes'] );
      IdlSeg.AppendChild( ChunkResult['Durations'] );
      IdlSeg.AppendChild( ChunkResult['States'] );
      IdlSeg.AppendChild( ChunkResult['StateDurations'] );

      HistGroup := ChunkResult.Find('HistogramGroup');
      HistGroup.Find('Title').SetData('Segment ' + IntToStr(i+1));
      // find relevant histogram and discard the rest:
      ActiveHist := HistGroup.Find('Histogram');
      for a := 0 to TQUB_QVO_QVD(Data).ActiveChannel - 1 do begin
        ActiveHist := ActiveHist.SiblingSameName;
      end;
      while not HistGroup.Find('Histogram').IsNull do begin
        HistGroup.RemoveChild(HistGroup.Find('Histogram'));
      end;
      if not ActiveHist.IsNull then begin
        HistGroup.AppendChild(ActiveHist);
      end;

      try

        if Chunk.Status = CHUNK_STATUS_DONE then begin

          if IdlSeg['Durations'].DataCount > 0 then begin

            //Reorder Classes
            if prParams.Method = METHOD_HALFAMP then begin
              for a := 0 to IdlChunk.SegCount - 1 do begin
                IdlChunk.SegClasses[a] := prParams.HalfAmp.IdxAmps[IdlChunk.SegClasses[a]];
              end;
            end;

            //Fix 'interburst' w.r.t prev chunk
            if LastSegIx = TQUB_QVO_QVD(Data).WhichSegmentForPoint(ChunkResult.Data[0]) then begin
              ChunkResult['interburst'].SetData(ChunkResult['start'].Data[0] - (  LastChunkResult['start'].Data[0]
                                                                                + LastChunkResult['length'].Data[0]));
            end else begin
              LastSegIx := TQUB_QVO_QVD(Data).WhichSegmentForPoint(ChunkResult.Data[0]);
            end;

            //copy idl from node
            //if prParams.Method = METHOD_SKM then begin
            if  (ChunkResult.DataCount >= 2)
            and (IdlSeg['DwellCount'].DataCount > 0)
            and (IdlSeg['Classes'].DataCount > 0)
            and (IdlSeg['Durations'].DataCount > 0) then begin
              Chunk.Sel.First       := Integer(ChunkResult.Data[0]);
              Chunk.Sel.Last        := Integer(ChunkResult.Data[1]);
              Chunk.StartTime       := Single(ChunkResult['start'].Data[0] * 1.0e-3);

              IdlChunk.SegCount     := Integer(IdlSeg['DwellCount'].Data[0]);


              idwt                  := TPAInteger(IdlSeg['Classes'].DataPtr);
              tdwt                  := TPASingle(IdlSeg['Durations'].DataPtr);
              ReallocMem( IdlChunk.SegFirsts, IdlChunk.SegCount * SizeOf(Integer) );
              ReallocMem( IdlChunk.SegLasts, IdlChunk.SegCount * SizeOf(Integer) );
              ReallocMem( IdlChunk.SegClasses, IdlChunk.SegCount * SizeOf(Integer) );

              if IdlChunk.SegCount > 0 then begin
                IdlChunk.SegFirsts[0]    := Chunk.Sel.First;
                IdlChunk.SegLasts[0]     := Chunk.Sel.First + Round(tdwt[0] * 1.0e-3 / TQUB_QVO_QVD(Data).Sampling) - 1;
                IdlChunk.SegClasses[0]   := idwt[0];
                for a := 1 to IdlChunk.SegCount - 1 do begin
                  IdlChunk.SegFirsts[a]  := IdlChunk.SegLasts[a-1] + 1;
                  IdlChunk.SegLasts[a]   := IdlChunk.SegFirsts[a] + Round(tdwt[a] * 1.0e-3 / TQUB_QVO_QVD(Data).Sampling) - 1;
                  IdlChunk.SegClasses[a] := idwt[a];
                end;
              end;
            end;


            Sel.First := 0;
            Sel.Last  := IdlChunk.SegCount - 1;

            Node                     := IdlSeg['Firsts'];
            Node.SetupNumData(NODE_DATATYPE_LONGINT, IdlChunk.SegCount, 1);
            Node.WriteRows(IdlChunk.SegFirsts, Sel);
            Node                     := IdlSeg['Lasts'];
            Node.SetupNumData(NODE_DATATYPE_LONGINT, IdlChunk.SegCount, 1);
            Node.WriteRows(IdlChunk.SegLasts, Sel);



            Report := Format('Segment %d:'#13#10'%8s%12s%12s%12s%14s%8s', [i+1, 'class', 'amp', 'std', 'occupancy', 'lifetime(ms)', 'nevent']);

            AmpNode               := ChunkResult['amp'];
            sdNode                := ChunkResult['sd'];
            ClassCount            := AmpNode.DataCount;
            ReallocMem( Amps, ClassCount * SizeOf(Single) );
            ReallocMem( Stds, ClassCount * SizeOf(Single) );
            for a := 0 to ClassCount - 1 do begin
              Amps[a]             := Single( AmpNode.Data[a] );
              Stds[a]             := Single( sdNode.Data[a] );
              if SkipTheReport then Continue;
              try
                Report              := Report + Format(#13#10'%8d%12.4f%12.4f%12.4f%14.4f%8d', [a+1, Amps[a], Stds[a], Double(ChunkResult['occupancy'].Data[a]), Double(ChunkResult['lifetime'].Data[a]), Integer(ChunkResult['nevent'].Data[a])]);
              except
                SkipTheReport     := True;
              end;
            end;


            IdlSeg.AppendClone( AmpNode, True );
            IdlSeg.AppendClone( sdNode, True );

            prConsoleMsg  := Report;
            prConsoleCode := MESSAGE_MESSAGE;
            Synchronize(prWriteNewConsole);

          end;

        end;//Done


      finally


        //Free Chunk's data
        FreeMem(IdlChunk.Data);             IdlChunk.Data        := nil;

        IdlChunk.SegCount    := 0;

        FreeMem(IdlChunk.SegFirsts);        IdlChunk.SegFirsts   := nil;
        FreeMem(IdlChunk.SegLasts);         IdlChunk.SegLasts    := nil;
        FreeMem(IdlChunk.SegClasses);       IdlChunk.SegClasses  := nil;

        IdlChunk.ClassCount := 0;

        FreeMem(IdlChunk.Amps);             IdlChunk.Amps        := nil;
        FreeMem(IdlChunk.Stds);             IdlChunk.Stds        := nil;
        FreeMem(IdlChunk.Occupancy);        IdlChunk.Occupancy   := nil;
        FreeMem(IdlChunk.Lifetime);         IdlChunk.Lifetime    := nil;
        FreeMem(IdlChunk.NEvents);          IdlChunk.NEvents     := nil;
        try
          Free2(TPMChar(IdlChunk.AMatrix));   IdlChunk.AMatrix     := nil;
        except
        end;

        Free2(TPMChar(IdlChunk.Histogram)); IdlChunk.Histogram   := nil;

        IdlChunk.ReportSize := 0;

        FreeMem(IdlChunk.Report);           IdlChunk.Report      := nil;

        if IdlChunk.ResultNode <> nil then begin
          Node_DECREF( IdlChunk.ResultNode );
          IdlChunk.ResultNode  := nil;
        end;

      end;

      LastChunkResult      := ChunkResult;

    end;

    ComputeAvgAmps( prStatResult );


    TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].SetDwells(prDatasetTemplate);
    TQUB_QVO_QVD(Data).DrawIdealized := True;


  finally


    FreeMem(Amps);
    FreeMem(Stds);

    Synchronize(prPublishResults);

  end;

end;


procedure TQUB_Task_IdealizeMT.prPublishResults;
begin
  if prParams.Method = METHOD_HALFAMP then begin
    prStatResult.Name  := 'HalfAmp ' + ChangeFileExt(ExtractFileName(TQUB_QVO_QVD(Data).FileName), '');
  end else begin
    prStatResult.Name  := 'SKM ' + ChangeFileExt(ExtractFileName(TQUB_QVO_QVD(Data).FileName), '');
  end;

  prStatResult['Hold'].SetData( 1 );
  
  ViewportResults.AddResults(prStatResult, False); // just the stats
end;


//****************************************************************************//
//                                                                            //
//                    I D E A L I Z E     S P E C I F I C S                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdealizeMT.prGetAmpsAndClss(Model :TQUB_QVO_QVM);
var
  SKMClasses :INode;
  SKMAmps    :INode;
  SKMStds    :INode;
  mc      :Integer;
  i       :Integer;
  cc      :TPAInteger;
  mx      :Integer;
  f       :Boolean;
  d       :Integer;
  x       :Double;
begin

  if Model.ModelObject.ChannelCount > 1 then begin
    SKMClasses := SKM_Get_Multiclasses(Model.ModelObject.CloneAsNode(True, TNode.CreateNull), Model.ModelObject.ChannelCount);
    SKMAmps    := SKMClasses['amp'];
    SKMStds    := SKMClasses['sd'];

    mc         := SKMAmps.DataCount - 1;

    cc := IAlloc1(mc + 1);

    prParams.HalfAmp.Classes   := IAlloc1( mc + 1 );
    prParams.HalfAmp.HalfAmps  := IAlloc1( mc + 1 );
    prParams.HalfAmp.Amps      := IAlloc1( mc + 1 );
    prParams.HalfAmp.IdxAmps   := IAlloc1( mc + 1 );
    prParams.HalfAmp.AmpVals   := DAlloc1( mc + 1 );
    prParams.HalfAmp.StdVals   := DAlloc1( mc + 1 );

    for i := 0 to mc do begin
      cc[i]                       := 1;  // let's say they're all there, but I haven't investigated
      prParams.HalfAmp.AmpVals[i] := SKMAmps.Data[i];
      prParams.HalfAmp.StdVals[i] := SKMStds.Data[i];
      prParams.HalfAmp.Classes[i] := i;
      prParams.HalfAmp.Amps[i]    := Round((SKMAmps.Data[i] - TQUB_QVO_QVD(Data).DataChannelOffset[ChannelIndex]) * TQUB_QVO_QVD(Data).DataObject.Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex]);
      prParams.HalfAmp.IdxAmps[i] := i;
    end;

    mx := mc + 1;

  end else begin

    //Find how many classes are, actually, present
    mc := Model.ModelObject.States[0].Cls;
    for i := 1 to Model.ModelObject.States.StateCount - 1 do begin
      if Model.ModelObject.States[i].Cls > mc then mc := Model.ModelObject.States[i].Cls;
    end;
    //mc is the highet class present in the model
    cc := IAlloc1(mc + 1);
    for i := 0 to mc do begin
      cc[i] := 0;
    end;

    prParams.HalfAmp.Classes   := IAlloc1( mc + 1 );
    prParams.HalfAmp.HalfAmps  := IAlloc1( mc + 1 );
    prParams.HalfAmp.Amps      := IAlloc1( mc + 1 );
    prParams.HalfAmp.IdxAmps   := IAlloc1( mc + 1 );
    prParams.HalfAmp.AmpVals   := DAlloc1( mc + 1 );
    prParams.HalfAmp.StdVals   := DAlloc1( mc + 1 );

    //Is there at least one state in class cc[i] ?
    for i := 0 to Model.ModelObject.States.StateCount - 1 do begin
      cc[Model.ModelObject.States[i].Cls] := 1;
    end;

    mx := 0;
    for i := 0 to mc do begin
      //Is class i represented ?
      if cc[i] = 1 then begin
        prParams.HalfAmp.Classes[mx] := i;
        prParams.HalfAmp.AmpVals[mx] := Model.ModelObject.Amp[prParams.HalfAmp.Classes[mx]];
        prParams.HalfAmp.StdVals[mx] := Model.ModelObject.Std[prParams.HalfAmp.Classes[mx]];
        prParams.HalfAmp.Amps[mx]    := Round((prParams.HalfAmp.AmpVals[mx] - TQUB_QVO_QVD(Data).DataChannelOffset[ChannelIndex]) * TQUB_QVO_QVD(Data).DataObject.Scaling * TQUB_QVO_QVD(Data).DataChannelScaling[ChannelIndex]);
        prParams.HalfAmp.IdxAmps[mx] := mx;
        Inc(mx);
      end;
    end;

  end;

  try

    //Sort and remember the index

    repeat
      f := False;
      for i := 0 to mx - 2 do begin
        if prParams.HalfAmp.Amps[i] > prParams.HalfAmp.Amps[i + 1] then begin
          d := prParams.HalfAmp.Amps[i]; prParams.HalfAmp.Amps[i] := prParams.HalfAmp.Amps[i + 1]; prParams.HalfAmp.Amps[i + 1] := d;
          d := prParams.HalfAmp.Classes[i]; prParams.HalfAmp.Classes[i] := prParams.HalfAmp.Classes[i + 1]; prParams.HalfAmp.Classes[i + 1] := d;
          d := prParams.HalfAmp.IdxAmps[i]; prParams.HalfAmp.IdxAmps[i] := prParams.HalfAmp.IdxAmps[i + 1]; prParams.HalfAmp.IdxAmps[i + 1] := d;
          x := prParams.HalfAmp.AmpVals[i]; prParams.HalfAmp.AmpVals[i] := prParams.HalfAmp.AmpVals[i + 1]; prParams.HalfAmp.AmpVals[i + 1] := x;
          x := prParams.HalfAmp.StdVals[i]; prParams.HalfAmp.StdVals[i] := prParams.HalfAmp.StdVals[i + 1]; prParams.HalfAmp.StdVals[i + 1] := x;
          f := True;
        end;
      end;

      prParams.HalfAmp.ClassCount := mx;

      for i := 0 to mx - 2 do begin
        prParams.HalfAmp.HalfAmps[i] := (prParams.HalfAmp.Amps[i] + prParams.HalfAmp.Amps[i + 1]) div 2;
      end;

    until (not f);

  finally

    FreeMem(cc);

  end;

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdealizeMT.OnGetProperties(Sender :TObject);
var
  c :Integer;
begin
  with Sender as TQUB_Task_Idealize_PropertiesForm do begin
    Self.ChannelIndex           := Properties['ChannelIndex'].Data[0];

    //Data Source
    Self.DataSource             := DataSource;

    //Data Processed
    Self.DataProcessed          := StringToDataProcessed(Properties['ProcessData'].DataAsString);
    Self.ProcessFilter          := 0<>Properties['ProcessData']['Filter'].Data[0];
    Self.ProcessFilterFreq      := Properties['ProcessData']['FilterFreq'].Data[0];
    Self.ProcessBaseline        := 0<>Properties['ProcessData']['Baseline'].Data[0];

    //Client/Server
    Self.ThreadCount            := Properties['ThreadCount'].Data[0];
    Self.Remote                 := 0<>Properties['Remote'].Data[0];

    //Method
    prParams.Method             := Method;

    prParams.HalfAmp.ComputeStats := HAStats;

    //Drop First/Last
    prParams.DropFirstLast      := DropFirstLast;
    prParams.DropIfClass        := DropIfClass;
    prParams.DropClass          := DropClass;

    prParams.ApplyDeadTime      := ApplyDeadTime;
    prParams.IdlDeadTime        := IdlDeadTime;

    prParams.Remote             := Remote;

    //SKM
    prParams.SKM.HistBinCount   := SKMHistBinCount;
    prParams.SKM.FixQ           := Integer(SKMFixQ);
    ReallocMem( prParams.SKM.FixA, ClassCount * SizeOf(Integer) );
    ReallocMem( prParams.SKM.FixS, ClassCount * SizeOf(Integer) );
    for c := 0 to ClassCount - 1 do begin
      prParams.SKM.FixA[c]      := Integer(Params.SKM.FixA[c+1]);
      prParams.SKM.FixS[c]      := Integer(Params.SKM.FixS[c+1]);
    end;
    prParams.SKM.LLConv         := SKMLLConv;
    prParams.SKM.MaxIter        := SKMMaxIter;
    prParams.SKM.Viterbi        := SKMViterbi;

    prParams.ClearIdl           := ClearIdl;

    prConfig                    := Properties.Clone( True )

  end;//with
end;

procedure TQUB_Task_IdealizeMT.OnPutProperties(Sender :TObject);
begin
  // this was solving no problems; just creating them by slyly setting channelindex wrong
  //with Sender as TQUB_Task_Idealize_PropertiesForm do begin
  ////  ChannelIndex                := Self.ChannelIndex;
  //end;//with
end;

procedure TQUB_Task_IdealizeMT.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  Model.LockAny;
end;

//******************************************************************************
end.

