unit QUB_Task_Idealize_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,
RegistryInterface,
Math,
QUB_QFS_Channels,
QUB_Task,
QUB_Task_Data,
QUB_Properties,
QUB_Task_Properties,
QUB_Task_Data_Properties,
QUB_Idealize_Global,
QUB_Node,
QUB_Table,
QUB_QVO_QVM, Grids;


{$include 'QUB_Task_Idealize_Properties.hct'}


const


//Data Source
DEF_DATA_SOURCE               = DATA_SOURCE_SELECTION;


type


TQUB_Task_Idealize_PropertiesForm = class(TQUB_Task_Data_PropertiesForm)

  btnReset                        :TButton;
  panMethod                       :TPanel;
  chkMethodHalfAmp                :TRadioButton;
  chkMethodSKM                    :TRadioButton;
  txtSKMHistBinCount              :TEdit;
  udSKMHistBinCount               :TUpDown;
  chkSKMFixQ                      :TCheckBox;
  panDataSource                   :TPanel;
  Label6                          :TLabel;
  chkDataSourceWholeFile          :TRadioButton;
  chkDataSourceSelectionList      :TRadioButton;
  chkDataSourceSelection          :TRadioButton;
  Label7                          :TLabel;
  Label9                          :TLabel;
  chkDropFirstLast                :TCheckBox;
  chkDropIfClass                  :TCheckBox;
  txtDropClass                    :TEdit;
  udDropClass                     :TUpDown;
  Label12                         :TLabel;
  txtSKMLLConv                    :TEdit;
  Label13                         :TLabel;
  txtSKMMaxIter                   :TEdit;
    chkSKMReestimate: TCheckBox;
    chkApplyDeadTime: TCheckBox;
  chkHAStats                      :TCheckBox;
  chkClearIdl                     :TCheckBox;
    chkIdlDeadTime: TCheckBox;
    gridFix: TStringGrid;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure chkDataSourceClick(Sender: TObject);
    procedure gridFixDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure gridFixMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);


protected


  //Data Source
  prDataSource                    :TQUB_Task_Data_Source;

  //Data Processed
  prDataProcessed                 :TQUB_Task_Data_Processed;
  prProcessFilter                 :Boolean;
  prProcessFilterFreq             :Extended;
  prProcessBaseline               :Boolean;

  //Parameters
  prParams                        :TQUB_Idealize_Params;

  prClassCount                    :Integer;

  //Data Source
  procedure prSetDataSource(s :TQUB_Task_Data_Source);

  procedure prSetClassCount(x :Integer);

  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public

  procedure PropertiesToParams;


  //Data Source
  property DataSource                     :TQUB_Task_Data_Source     read prDataSource                write prSetDataSource;

  property ClassCount                     :Integer                   read prClassCount                write prSetClassCount;

  //Parameters
  property Params                         :TQUB_Idealize_Params      read prParams;

  //Method
  property Method                         :Integer                   read prParams.Method;

  property HAStats                        :Boolean                   read prParams.HalfAmp.ComputeStats;
  property ClearIdl                       :Boolean                   read prParams.ClearIdl;

  property DropFirstLast                  :Boolean                   read prParams.DropFirstLast;
  property DropIfClass                    :Boolean                   read prParams.DropIfClass;
  property DropClass                      :Integer                   read prParams.DropClass;

  property SKMHistBinCount                :Integer                   read prParams.SKM.HistBinCount;
  property SKMFixQ                        :Integer                   read prParams.SKM.FixQ;
  property SKMLLConv                      :Double                    read prParams.SKM.LLConv;
  property SKMMaxIter                     :Integer                   read prParams.SKM.MaxIter;
  property SKMViterbi                     :Integer                   read prParams.SKM.Viterbi;

  property ApplyDeadTime                  :Boolean                   read prParams.ApplyDeadTime;
  property IdlDeadTime                    :Boolean                   read prParams.IdlDeadTime;

end;


implementation


uses

  QUB_QFS;


{$R *.DFM}


//******************************************************************************

procedure TQUB_Task_Idealize_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prParams.SKM.FixA                := nil;
  prParams.SKM.FixS                := nil;

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Method"]',
                          'Method', REGTYPE_INT, IdlMethodToString(DEF_METHOD))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMethodHalfAmp,
                                       IdlMethodToString(METHOD_HALFAMP),
                                       Integer(METHOD_HALFAMP))
                                .Enables(chkHAStats)
                             )
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMethodSKM,
                                       IdlMethodToString(METHOD_SKM),
                                       Integer(METHOD_SKM))
                                .Enables(txtSKMLLConv)
                                .Enables(txtSKMMaxIter)
                                .Enables(chkDropFirstLast)
                                .Enables(chkDropIfClass)
                                .Enables(txtDropClass)
                                .Enables(udDropClass)
                                .Enables(chkSKMReestimate)
                                .Enables(chkSKMFixQ)
                             )
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["HalfAmpStats"]',
                        'HalfAmp Stats', REGTYPE_BOOL, DEF_HASTATS,
                        chkHAStats));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ClearIdl"]',
                        'ClearIdl', REGTYPE_BOOL, DEF_CLEARIDL,
                        chkClearIdl));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DropFirstLast"]',
                        'Drop First Last', REGTYPE_BOOL, DEF_DROPFIRSTLAST,
                        chkDropFirstLast));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DropIfClass"]',
                        'Drop If Class', REGTYPE_BOOL, DEF_DROPIFCLASS,
                        chkDropIfClass));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["DropClass"]',
                      'Drop Class', REGTYPE_INT, DEF_DROPCLASS,
                      udDropClass,
                      txtDropClass,
                      AcceptIntGreaterThan(0)));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["Histograms"]["BinCount"]',
                      'SKM Hist Bin Count', REGTYPE_INT, DEF_SKMHISTBINCOUNT,
                      udSKMHistBinCount,
                      txtSKMHistBinCount,
                      AcceptIntBetween(MIN_SKMHISTBINCOUNT, MAX_SKMHISTBINCOUNT)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["FixQ"]',
                        'SKM Fix Q', REGTYPE_BOOL, DEF_SKMFIXQ,
                        chkSKMFixQ));
  AddEditProperty(TQUB_Property_Edit.Create('["ConvLL"]',
                    'SKM LL Conv', REGTYPE_FLOAT, DEF_SKMLLCONV,
                    txtSKMLLConv,
                    AcceptFloatBetween(MIN_SKMLLCONV, MAX_SKMLLCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxIterations"]',
                    'SKM MaxIter', REGTYPE_INT, DEF_SKMMAXITER,
                    txtSKMMaxIter,
                    AcceptIntBetween(MIN_SKMMAXITER, MAX_SKMMAXITER),
                   '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Reestimate"]',
                        '', REGTYPE_NONE, (DEF_SKMVITERBI + 1) mod 2,
                        chkSKMReestimate));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ApplyDeadTime"]',
                        'ApplyDeadTime', REGTYPE_BOOL, DEF_APPLYDEADTIME,
                        chkApplyDeadTime)
                      );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["IdlDeadTime"]',
                        '', REGTYPE_NONE, DEF_IDLDEADTIME,
                        chkIdlDeadTime)
                      );

  RegKeyName               := 'Software\QUB\Task Idealize data';
  ProfileCat               := 'Idl';
  prBeforeGetProperties := PropertiesToParams;

  WikiTopic                := 'Modeling:Idealize#Properties';

  gridFix.Cells[0, 0]      := 'Class';
  gridFix.Cells[1, 0]      := 'Fix Amp';
  gridFix.Cells[2, 0]      := 'Fix Std';

  ClassCount                       := QUB_QFS_QMF_DEF_NCLASS;

  {$include 'QUB_Task_Idealize_Properties.htx'}

end;

procedure TQUB_Task_Idealize_PropertiesForm.FormDestroy(Sender: TObject);
begin
  inherited;

  try
    FreeMem( prParams.SKM.FixA );
    FreeMem( prParams.SKM.FixS );
  except
  end;

end;

procedure TQUB_Task_Idealize_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  DataSource             := DataSource;

end;

procedure TQUB_Task_Idealize_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;
end;


//******************************************************************************

procedure TQUB_Task_Idealize_PropertiesForm.prSetInitialProperties(Props :INode);
var
  i, n    :Integer;
  Fix     :INode;
begin
  inherited;

  Fix     := Props['FixAmp'];
  n       := Fix.DataCount;
  if n = 0 then begin
    Fix.SetupNumData(NODE_DATATYPE_LONGINT, ClassCount, 1);
  end else if n < ClassCount then begin
    Fix.ResizeData(ClassCount);
  end;
  for i := 0 to n - 1 do begin
    if Fix.Data[i] > 1 then Fix.Data[i] := 0;
  end;
  for i := n to ClassCount - 1 do begin
    Fix.Data[i] := 0;
  end;

  Fix     := Props['FixSD'];
  n       := Fix.DataCount;
  if n = 0 then begin
    Fix.SetupNumData(NODE_DATATYPE_LONGINT, ClassCount, 1);
  end else if n < ClassCount then begin
    Fix.ResizeData(ClassCount);
  end;
  for i := 0 to n - 1 do begin
    if Fix.Data[i] > 1 then Fix.Data[i] := 0;
  end;
  for i := n to ClassCount - 1 do begin
    Fix.Data[i] := 0;
  end;

  Props['DialogOnLClick'].SetData(1);
  // a script could unset this
end;

procedure TQUB_Task_Idealize_PropertiesForm.prUpdateComponents;
var
  i         :Integer;
begin
  inherited;
  
  gridFix.RowCount := ClassCount + 1;

  for i := 0 to ClassCount - 1 do begin
    if Properties['FixAmp'].Data[i] > 1 then Properties['FixAmp'].Data[i] := 0;
    if Properties['FixSD'].Data[i] > 1 then Properties['FixSD'].Data[i] := 0;
    gridFix.Cells[0, i+1] := IntToStr(i+1);
    gridFix.Cells[1, i+1] := '';
    gridFix.Cells[2, i+1] := '';
  end;
end;

procedure TQUB_Task_Idealize_PropertiesForm.PropertiesToParams;
var
  n         :Integer;
  c         :Integer;
begin
  n         := Max(prClassCount, Properties['FixAmp'].DataCount);
  ReallocMem( prParams.SKM.FixA, n * SizeOf(Integer) );
  ReallocMem( prParams.SKM.FixS, n * SizeOf(Integer) );
  if Properties['FixAmp'].DataCount = 0 then Properties['FixAmp'].SetData(0);
  if Properties['FixSD'].DataCount = 0 then Properties['FixSD'].SetData(0);
  c         := Min(Properties['FixAmp'].DataCount, Properties['FixSD'].DataCount);
  Properties['FixAmp'].ResizeData(n);
  Properties['FixSD'].ResizeData(n);
  while c < n do begin
    Properties['FixAmp'].Data[c] := 0;
    Properties['FixSD'].Data[c]  := 0;
  end;
  // the resized area wasn't initialized, leading to nonsense in peoples' prefs.  should be zero nonsense
  for c := 0 to n - 1 do begin
    //prParams.SKM.FixA[c]        := Properties['FixAmp'].Data[c];
    //prParams.SKM.FixS[c]        := Properties['FixSD'].Data[c];
    case Integer(Properties['FixAmp'].Data[c]) of
      0:    prParams.SKM.FixA[c]   := 0;
      1:    prParams.SKM.FixA[c]   := 1;
    else begin
      Properties['FixAmp'].Data[c] := 0;
      prParams.SKM.FixA[c]         := 0;
    end end;
    case Integer(Properties['FixSD'].Data[c]) of
      0:    prParams.SKM.FixS[c]   := 0;
      1:    prParams.SKM.FixS[c]   := 1;
    else begin
      Properties['FixSD'].Data[c]  := 0;
      prParams.SKM.FixS[c]         := 0;
    end end;
  end;
  prParams.Method               := StringToIdlMethod(Properties['Method'].DataAsString);
  prParams.HalfAmp.ComputeStats := 0<>Properties['HalfAmpStats'].Data[0];
  prParams.ClearIdl             := 0<>Properties['ClearIdl'].Data[0];
  prParams.DropFirstLast        := 0<>Properties['DropFirstLast'].Data[0];
  prParams.DropIfClass          := 0<>Properties['DropIfClass'].Data[0];
  prParams.DropClass            := Properties['DropClass'].Data[0] - 1;
  prParams.SKM.HistBinCount     := Properties['Histograms']['BinCount'].Data[0];
  prParams.SKM.FixQ             := Properties['FixQ'].Data[0];
  prParams.SKM.LLConv           := Properties['ConvLL'].Data[0];
  prParams.SKM.MaxIter          := Properties['MaxIterations'].Data[0];
  prParams.SKM.Viterbi          := Integer(0=Properties['Reestimate'].Data[0]);
  prParams.ApplyDeadTime        := 0<>Properties['ApplyDeadTime'].Data[0];
  prParams.IdlDeadTime          := 0<>Properties['IdlDeadTime'].Data[0];
end;

// *****************************************************************************
procedure TQUB_Task_Idealize_PropertiesForm.prSetClassCount(x :Integer);
var
  i      :Integer;
begin
  if x <= 0 then x := 1;
  i := prClassCount;
  prClassCount           := x;

  ReallocMem( prParams.SKM.FixA, x * SizeOf(Integer) );
  ReallocMem( prParams.SKM.FixS, x * SizeOf(Integer) );
  if i = 0 then begin
    Properties['FixAmp'].SetupNumData(NODE_DATATYPE_LONGINT, x, 1);
    Properties['FixSD'].SetupNumData(NODE_DATATYPE_LONGINT, x, 1);
  end else begin
    Properties['FixAmp'].ResizeData(x);
    Properties['FixSD'].ResizeData(x);
  end;
  while i < x do begin
    Properties['FixAmp'].Data[i] := 0;
    Properties['FixSD'].Data[i]  := 0;
    Inc(i);
  end;

  udDropClass.Min        := 1;
  if udDropClass.Position > x then begin
    Properties['DropClass'].Data[0] := x;
    prUpdateComponents;
  end;
  udDropClass.Max        := x;

  prUpdateComponents;
end;


//Data Source
procedure TQUB_Task_Idealize_PropertiesForm.prSetDataSource(s :TQUB_Task_Data_Source);
begin
  prDataSource                       := s;
  chkDataSourceWholeFile.Checked     := DataSource = DATA_SOURCE_WHOLEFILE;
  chkDataSourceSelectionList.Checked := DataSource = DATA_SOURCE_SELECTIONLIST;
  chkDataSourceSelection.Checked     := DataSource = DATA_SOURCE_SELECTION;
end;

//******************************************************************************
//Data Source
procedure TQUB_Task_Idealize_PropertiesForm.chkDataSourceClick(Sender: TObject);
begin
  inherited;
  DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;

procedure TQUB_Task_Idealize_PropertiesForm.gridFixDrawCell(
  Sender: TObject; ACol, ARow: Integer; Rect: TRect;
  State: TGridDrawState);
var
  Checked      :Boolean;
  r            :TRect;
begin
  inherited;
  if ARow > 0 then begin
    if ACol = 1 then begin
      Checked  := 0 <> Properties['FixAmp'].Data[ARow-1];
    end else if ACol = 2 then begin
      Checked  := 0 <> Properties['FixSD'].Data[ARow-1];
    end else begin
      Exit;
    end;
    r        := Rect;
    InflateRect(r, 0, -((r.Bottom - r.Top + 1) - 15) div 2);
    if Checked then begin
      DrawFrameControl(gridFix.Canvas.Handle, r, DFC_BUTTON, DFCS_BUTTONCHECK or DFCS_CHECKED);
    end else begin
      DrawFrameControl(gridFix.Canvas.Handle, r, DFC_BUTTON, DFCS_BUTTONCHECK);
    end;
  end;
end;

procedure TQUB_Task_Idealize_PropertiesForm.gridFixMouseDown(
  Sender: TObject; Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var
  ACol, ARow :Integer;
begin
  gridFix.MouseToCell(X, Y, ACol, ARow);
  if ARow > 0 then begin
    if ACol = 1 then begin
      Properties['FixAmp'].Data[ARow-1] := Integer(0=Properties['FixAmp'].Data[ARow-1]);
    end else if ACol = 2 then begin
      Properties['FixSD'].Data[ARow-1] := Integer(0=Properties['FixSD'].Data[ARow-1]);
    end else begin
      Exit;
    end;
    gridFix.Cells[ACol, ARow] := '';
  end;
end;


//******************************************************************************
end.
