unit QUB_Task_IdlFit;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math, Controls, Dialogs,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QVO_QVM,

QUB_QFS_QMF,

QUB_Viewport_Data,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_IdlFit_Properties,

Chart, Series,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, QUB_Viewport_Results,
QUB_GUIInterface;


const


IDLFIT_REASONABLE_NPOINTS                              = 500000;



type


TQUB_Task_IdlFit  = class(TQUB_Task_Data)
protected

  prOptName                :String;

  //Source Objects
  prModel                  :TQUB_QVO_QVM;
  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  prConfig                 :INode;
  prDataDescription        :INode;
  prDatasets               :Array of INode;
  prResultNode             :INode;
  prReportNode             :INode;
  prRealityCheckSaysGo     :Boolean;

  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prRealityCheck; // synchronize; sets prRealityCheckSaysGo := true to proceed
  procedure prPublishResults;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetModel(Model :TQUB_QVO_QVM);

  function prReportCallback( arg :INode ): INode;
  procedure prDoReport;

  function prResultCallback( arg :INode ): INode;
  procedure prDoResult;

  procedure prShowIdealization;


public

  FillDataCallback           :TNodeMethPtr;

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Stop; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  property Model             :TQUB_QVO_QVM              read prModel             write prSetModel;

end;


function IdlFitFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; fillDataCB, rptCB, rsltCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_idlfit_tree_iface@20';


implementation


uses
  QUB_Node_Model, QUB_Python_Data, QUB_QFS_QDF, QUB_Main,
  QUB_MIL_Global,
  QUB_ActionSelector_Modeling,
  QUB_FitCurves;


//****************************************************************************//
//                                                                            //
//                       T Q U B _ T a s k _ I d l F i t                      //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_IdlFit.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'IdlFit';

end;

destructor TQUB_Task_IdlFit.Destroy;
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_IdlFit.Stop;
begin

  prConfig['StopFlag'].Data[0] := 1;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlFit.prDoTask;
var
  i               :Integer;

  DataSetPtrs     :TPAPointer;
  fillDataCB      :INodeCFunc;
  rptCB, rsltCB   :INodeCFunc;
  ResultNode      :INode;
  ResultNodePtr   :TQUB_QFS_PNodeMem;
begin
  Synchronize( prRealityCheck );
  if not prRealityCheckSaysGo then Exit;

  GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * (Length(prDatasets) + 1) );
  for i := 0 to Length(prDataSets) - 1 do begin
    DataSetPtrs[i] := Pointer(prDataSets[i].GetNodeMem);
  end;
  DataSetPtrs[ Length(prDataSets) ] := nil;

  fillDataCB  := TNodeCFunc.Wrap( TNodeMeth.Create( FillDataCallback ) );
  rptCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prReportCallback ) );
  rsltCB      := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prResultCallback ) );

  prConfig['StopFlag'].SetData( 0 );

  try
    prConsoleMsg  := 'IdlFit running ... ';
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);

    ResultNodePtr := IdlFitFromDll( prConfig.GetNodeMem, DataSetPtrs, fillDataCB, rptCB, rsltCB );
    ResultNode    := TNode.CreateUsing( ResultNodePtr );
    Node_DECREF( ResultNodePtr );

    prResultNode := ResultNode;
    prShowIdealization;
    Synchronize( prPublishResults );

    for i := 0 to prChunks.Count - 1 do begin
      if TQUB_Thread_PChunk(prChunks[i]).Status = CHUNK_STATUS_EMPTY then begin
        TQUB_Thread_PChunk(prChunks[i]).Status := CHUNK_STATUS_DONE;
      end;
    end;
  except
    prError    := True;
    prErrorMsg := 'Unknown error in IdlFit library';
    prWikiTopic := 'Error:Task';
  end;

  FreeMem( DataSetPtrs );

end;

procedure TQUB_Task_IdlFit.prRealityCheck;
var
  i, r         :Integer;
  npoint       :Integer;
  Seg          :INode;
begin
  npoint       := 0;
  for i := 0 to Length(prDataSets) - 1 do begin
    Seg        := prDataSets[i].Find('Segment');
    while not Seg.IsNull do begin
      npoint   := npoint + (Seg.Data[1] - Seg.Data[0] + 1);
      Seg      := Seg.SiblingSameName;
    end;
  end;

  if npoint <= IDLFIT_REASONABLE_NPOINTS then begin
    prRealityCheckSaysGo := True;
    Exit;
  end;

  r := MessageHelpDlg( Format('IdlFit: %d data points could take too much time and memory.  Continue anyway?', [npoint]),
                   mtConfirmation, [mbYes, mbNo], 0, 'Warning:Too_Much_Time_And_Memory');

  prRealityCheckSaysGo := (r = mrYes);
end;

procedure TQUB_Task_IdlFit.prPublishResults;
var
  ds, seg         :INode;
begin
  prResultNode.Name  := 'IdlFit ' + ChangeFileExt(ExtractFileName(Data.FileName), '');

  prResultNode['ResultType'].SetData( 'IdlFit' );
  if (prDataDescription <> nil) and not prDataDescription.IsNull then prResultNode.AppendChild( prDataDescription );

  ds              := prResultNode.Find('DataSet');
  while not ds.isNull do begin
    seg           := ds.Find('Segment');
    while not seg.isNull do begin
      seg.RemoveChild( seg['Classes'] );
      seg.RemoveChild( seg['Durations'] );
      seg.RemoveChild( seg['Firsts'] );
      seg.RemoveChild( seg['Lasts'] );
      seg.RemoveChild( seg['FitCurve'] );

      seg         := seg.SiblingSameName;
    end;
    ds            := ds.SiblingSameName;
  end;

  if prSelectionList <> nil then begin          // if the selection list is 'IdlFit Results',
    prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
    prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
  end;

  ViewportResults.AddResults( prResultNode, False );
end;

procedure TQUB_Task_IdlFit.prDoExecuteStart;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  // clear fit curves?

  inherited;

end;

procedure TQUB_Task_IdlFit.prDoExecuteEnd;
begin

  inherited;

  if prSelectionList <> nil then begin
    prSelectionList.UnLock;
  end;

  if Model <> nil then begin
    Model.UnlockAny;
  end;

end;


function TQUB_Task_IdlFit.prDoExecuteInitialize :Boolean;
var
  MdlNode       :INode;
begin
  Result        := False;

  MdlNode       := GetModelAsNode;
  MdlNode.SetData( 'Initial' );

  prConfig.AppendChild( MdlNode );
end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlFit.prChunkData :Boolean;
var
  DataSets         :INode;
  DS               :INode;
  RemDS            :INode;
  Process          :INode;
begin
  Result           := False;

  Process          := prConfig.Find('ProcessData');

  DataSets         := DataNoder.GetChosenData(TNode.CreateNull);
  prDataDescription := DataSets.Find('DataDescription');

  DS               := DataSets.Find('DataSet');
  while not DS.IsNull do begin
    if DS.Find('Segment').IsNull then begin
      RemDS          := DS;
      DS             := DS.SiblingSameName;
      DataSets.RemoveChild( RemDS );
    end else begin
      SetLength( prDataSets, Length(prDataSets) + 1 );
      prDataSets[ Length(prDataSets) - 1 ] := DS;

      if not Process.IsNull then begin
        DS.AppendChild( Process.Clone(True) );
      end;

      prConsoleMsg := DS.AsString;
      prConsoleCode := MESSAGE_ERROR;
      Synchronize(prWriteNewConsole);

      DS             := DS.SiblingSameName;
    end;
  end;
end;


procedure TQUB_Task_IdlFit.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MilChunk :TQUB_Thread_Mil_PChunk;
begin

  MilChunk := TQUB_Thread_Mil_PChunk(Chunk);

  if MilChunk.ClassDwells <> nil then begin
    FreeMem(MilChunk.ClassDwells);
    MilChunk.ClassDwells := nil;
  end;

  FreeMem(MilChunk);

end;

function TQUB_Task_IdlFit.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Mil_Chunk));

  Result.Sel := Sel;

end;


procedure TQUB_Task_IdlFit.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then prModel.LockAny;
end;



//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlFit.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_IdlFit_PropertiesForm do begin
    Self.ChannelIndex           := ChannelIndex;

    prConfig := TNode.Create('Properties');
    prConfig := Properties.Clone(True);

    //Data Source
    Self.DataProcessed          := StringToDataProcessed(Properties['ProcessData'].DataAsString);
    Self.ProcessFilter          := 0<>Properties['ProcessData']['Filter'].Data[0];
    Self.ProcessFilterFreq      := Properties['ProcessData']['FilterFreq'].Data[0];
    Self.ProcessBaseline        := 0<>Properties['ProcessData']['Baseline'].Data[0];

    //Client/Server
    Self.ThreadCount            := Properties['ThreadCount'].Data[0];
    Self.Remote                 := 0<>Properties['Remote'].Data[0];

  end;//with
end;

procedure TQUB_Task_IdlFit.OnPutProperties(Sender :TObject);
begin
end;


//****************************************************************************//
//                                                                            //
//                          C A L L B A C K S                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlFit.prReportCallback( arg :INode ): INode;
begin
  prReportNode := arg;
  Synchronize( prDoReport );
  Result := arg;
end;

procedure TQUB_Task_IdlFit.prDoReport;
begin
  // just print the node data's text rep'n for now
  prConsoleMsg := prReportNode.GetDataAsString;
  prConsoleCode := MESSAGE_ERROR;
  prWriteNewConsole;
end;

function TQUB_Task_IdlFit.prResultCallback( arg :INode ): INode;
begin
  prResultNode := arg;
  Result       := arg;
  Synchronize( prDoResult );

  Sleep(2000);
end;

procedure TQUB_Task_IdlFit.prDoResult;
var
  i, N                :Integer;
  Curve               :INode;
  CurveDbls           :TPADouble;
  s, o                :Double;
  CurveIter           :TQUB_FitCurveIter;
  CurveData           :TPAInteger;
  DataObj             :TQUB_QVO_QVD;
begin
  // LL:  %f     Iterations:  %d
  prConsoleMsg := Format('LL: %f'#9'Iter: %d  |', [Double(prResultNode['LL'].Data[0]), Integer(prResultNode['Iterations'].Data[0])]);
  Curve               := prResultNode['amp'];
  for i := 0 to Curve.DataCount - 1 do begin
    prConsoleMsg      := prConsoleMsg + #9 + FloatToStr(Curve.Data[i]);
  end;
  prConsoleCode := MESSAGE_ERROR;
  prWriteNewConsole;

  // set fit curves:  find/add,  copy/scaled
  DataObj             := TQUB_QVO_QVD( Data );
  CurveIter           := nil;
  try
    Curve := prResultNode.Find('FitCurve');
    if Curve.IsNull then Exit;

    CurveDbls          := TPADouble( Curve.DataPtr );
    N                  := Curve.DataCount;

    CurveIter          := DataObj.FitCurves.Find(prResultNode.Data[0]);
    if (0 <> CurveIter.Done){isdone} or (CurveIter.Curve.First <> prResultNode.Data[0]) or (CurveIter.Curve.Last <> prResultNode.Data[1]) then begin
      CurveIter.Free;
      CurveIter        := DataObj.FitCurves.Add(prResultNode.Data[0], prResultNode.Data[1]);
    end;
    CurveData          := CurveIter.Curve.Data;

    s := DataObj.Scaling * DataObj.DataChannelScaling[ChannelIndex];
    o := DataObj.DataChannelOffset[ChannelIndex];
    for i := 0 to N - 1 do begin
      CurveData[i] := Round((CurveDbls[i] - o) * s);
    end;

    CurveIter.MarkChanged;

  finally
    if CurveIter <> nil then CurveIter.Free;
  end;

  DataObj.FitCurves.RaiseChanged;
end;


//****************************************************************************//
//                                                                            //
//                            R E S U L T S                                   //
//                                                                            //
//****************************************************************************//

procedure TQUB_Task_IdlFit.prShowIdealization;
var
  DataSet      :INode;
begin

  //Append Info
  prStatus := 'Applying Idealization';
  Synchronize(prDoOnChange);

  DataSet := prResultNode['DataSet'];
  TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].SetDwells(DataSet);

end;


//******************************************************************************
end.

