unit QUB_Task_IdlFit_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,
RegistryInterface,

QUB_QFS_Channels,
QUB_Task,
QUB_Task_Data,
QUB_Properties,
QUB_Task_Properties,
QUB_Task_Data_Properties,
QUB_Node,
QUB_DataList;


const


//Data Processed
DEF_DATA_PROCESSED            = DATA_PROCESSED_ASDISPLAYED;

DEF_INITBASE                  = 0.0;
MIN_INITBASE                  = -1e10;
MAX_INITBASE                  = 1e10;

DEF_INITA                     = 1.0;
MIN_INITA                     = -1e10;
MAX_INITA                     = 1e10;

DEF_INITT                     = 0.01;
MIN_INITT                     = 1e-10;
MAX_INITT                     = 1e10;

DEF_INITB                     = 0.0;
MIN_INITB                     = -1e10;
MAX_INITB                     = 1e10;

DEF_CONVLL                    = 0.01;
MIN_CONVLL                    = 1e-10;
MAX_CONVLL                    = 1e5;

DEF_MAXITER                   = 10;
MIN_MAXITER                   = 1;
MAX_MAXITER                   = 10000;

type


TQUB_Task_IdlFit_PropertiesForm = class(TQUB_Task_Data_PropertiesForm)

  panMethod                       :TPanel;
  Label7                          :TLabel;
  txtInitBase                     :TEdit;
  Label11                         :TLabel;
  txtInitA                        :TEdit;
  Label2                          :TLabel;
  txtInitT                        :TEdit;
  Label3                          :TLabel;
  txtInitB                        :TEdit;
  Label4                          :TLabel;
  txtConvLL                       :TEdit;
  Label6                          :TLabel;
  txtMaxIter                      :TEdit;
  Label8                          :TLabel;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure chkDataSourceClick(Sender: TObject);


protected


  prDataList                      :TQUB_DataList;

  //IdlFit
  prInitBase                      :Double;
  prInitA                         :Double;
  prInitT                         :Double;
  prInitB                         :Double;
  prConvLL                        :Double;
  prMaxIter                       :Integer;


public

  procedure PropertiesToParams;


  //IdlFit
  property InitBase                       :Double                    read prInitBase;
  property InitA                          :Double                    read prInitA;
  property InitT                          :Double                    read prInitT;
  property InitB                          :Double                    read prInitB;
  property ConvLL                         :Double                    read prConvLL;
  property MaxIter                        :Integer                   read prMaxIter;


end;


implementation


uses

QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface, QUB_Main;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_IdlFit_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := 20;
  prDataList.Top        := 50;
  prDataList.Width      := 540;
  prDataList.Height     := 100;
  prDataList.FileListAvail := False;

  AddEditProperty(TQUB_Property_Edit.Create('["InitBaseline"]',
                    'InitBase', REGTYPE_FLOAT, DEF_INITBASE,
                    txtInitBase,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["InitA"]',
                    'InitA', REGTYPE_FLOAT, DEF_INITA,
                    txtInitA,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["InitT"]',
                    'InitT', REGTYPE_FLOAT, DEF_INITT,
                    txtInitT,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["InitB"]',
                    'InitB', REGTYPE_FLOAT, DEF_INITB,
                    txtInitB,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["ConvLL"]',
                    'ConvLL', REGTYPE_FLOAT, DEF_CONVLL,
                    txtConvLL,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxIterations"]',
                    'MaxIter', REGTYPE_INT, DEF_MAXITER,
                    txtMaxIter,
                    AcceptIntGreaterThan(0),
                    '') );

  RegKeyName            := 'Software\QUB\Task IdlFit data';
  ProfileCat            := 'IdlF';
  prBeforeGetProperties := PropertiesToParams;

end;

procedure TQUB_Task_IdlFit_PropertiesForm.FormShow(Sender: TObject);
var
  Model                  :TQUB_QVO_QVM;
begin
  inherited;

  prDataList.ShowFiles;

  prDataList.ShowVar('ChannelCount', '', 1, False);

  Model                  := g_CurrModel();
  if Model <> nil then begin
    prDataList.ShowModelVars( Model, False );
  end;

end;

procedure TQUB_Task_IdlFit_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;

  prDataList.HideAll;
end;

//Data Source
procedure TQUB_Task_IdlFit_PropertiesForm.chkDataSourceClick(Sender: TObject);
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_IdlFit_PropertiesForm.PropertiesToParams;
begin
  prInitBase          := Properties['InitBaseline'].Data[0];
  prInitA             := Properties['InitA'].Data[0];
  prInitT             := Properties['InitT'].Data[0];
  prInitB             := Properties['InitB'].Data[0];
  prConvLL            := Properties['ConvLL'].Data[0];
  prMaxIter           := Properties['MaxIterations'].Data[0];
end;


//******************************************************************************
end.
