unit QUB_Task_IdlStim_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_Task,
QUB_Properties,
QUB_Task_Properties,
QUB_Node;

//{$include 'QUB_Task_IdlStim_Properties.hct'}

const


DEF_IDLCOND_CLEARIDEALIZATION  = 1;
DEF_IDLCOND_ADDDELTAS          = 1;
DEF_IDLCOND_DELTA            :Double  = 0.5;
MIN_IDLCOND_DELTA            :Double  = 1.0e-8;
MAX_IDLCOND_DELTA            :Double  = 1.0e9;
DEF_LATENCY                  :Double  = 0.0;
DEF_MINDUR                   :Double  = 5.0;
DEF_IDLCOND_SPRINGDASHPOT    :Integer = 0;
DEF_IDLCOND_SDP_DAMPING      :Double = 0.589;
DEF_IDLCOND_SDP_FREQ         :Double = 1075;


type


TQUB_Task_IdlStim_PropertiesForm = class(TQUB_Task_PropertiesForm)
  Label3                          :TLabel;

  panDataSource                   :TPanel;
  panDataSourceFile               :TPanel;
  chkDataSourceWholeFile          :TRadioButton;
  chkDataSourceSelectionList      :TRadioButton;
  chkDataSourceSelection          :TRadioButton;
  panOptions                      :TPanel;
  Label1                          :TLabel;
  txtKnown                        :TEdit;
  chkAddDeltas                    :TCheckBox;
  lblDelta                        :TLabel;
  txtDelta                        :TEdit;
  Label2                          :TLabel;
  txtLatency                      :TEdit;
    Label4: TLabel;
    txtMinDur: TEdit;
    chkSDP: TCheckBox;
    Label5: TLabel;
    Label6: TLabel;
    txtSDPDamping: TEdit;
    txtSDPFreq: TEdit;
    chkClearIdealization: TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure chkDataSourceClick(Sender: TObject);


protected


  //Data Source
  prDataSource                    :TQUB_Task_Data_Source;
  prDataSourceInFile              :TQUB_Task_Data_Source;

  //Data Source
  procedure prSetDataSource(s :TQUB_Task_Data_Source);


public

  //Data Source
  property DataSource                     :TQUB_Task_Data_Source     read prDataSource                write prSetDataSource;

end;


implementation


uses

QUB_Python_Thread,
QUB_Node_Python,
QUB_Viewport_Python;


{$R *.DFM}


//******************************************************************************
//Data Source
procedure TQUB_Task_IdlStim_PropertiesForm.prSetDataSource(s :TQUB_Task_Data_Source);
begin
  prDataSource                       := s;
  chkDataSourceWholeFile.Checked     := DataSource =  DATA_SOURCE_WHOLEFILE;
  chkDataSourceSelectionList.Checked := DataSource =  DATA_SOURCE_SELECTIONLIST;
  chkDataSourceSelection.Checked     := DataSource =  DATA_SOURCE_SELECTION;
end;


//******************************************************************************
procedure TQUB_Task_IdlStim_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ClearIdealization"]',
                        '', REGTYPE_NONE, DEF_IDLCOND_CLEARIDEALIZATION,
                        chkClearIdealization)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["Known Amps"]',
                          '', REGTYPE_NONE, 0,
                          txtKnown,
                          AcceptFloatList(AcceptFloat),
                          'lambda node, ix: SpacedNodeData(node)') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["AddDeltas"]',
                        'AddDeltas', REGTYPE_BOOL, DEF_IDLCOND_ADDDELTAS,
                        chkAddDeltas)
                       .Enables(txtDelta)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["Delta"]',
                    'Delta', REGTYPE_FLOAT, DEF_IDLCOND_DELTA,
                    txtDelta,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Latency_MS"]',
                    '', REGTYPE_NONE, DEF_LATENCY,
                    txtLatency,
                    AcceptFloatGreaterThanOrEqualTo(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MinDur_MS"]',
                    '', REGTYPE_NONE, DEF_MINDUR,
                    txtMinDur,
                    AcceptFloatGreaterThanOrEqualTo(0.0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["SpringDashPot"]',
                        '', REGTYPE_NONE, DEF_IDLCOND_SPRINGDASHPOT,
                        chkSDP)
                       .Enables(txtSDPDamping)
                       .Enables(txtSDPFreq)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["SDP Damping"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_SDP_DAMPING,
                    txtSDPDamping,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SDP Freq"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_SDP_FREQ,
                    txtSDPFreq,
                    AcceptFloatGreaterThan(0.0),
                    '') );

  RegKeyName            := 'Software\QUB\Task IdlStim';
  ProfileCat            := 'IdlStm';

  WikiTopic             := 'Preprocessing:IdlStim';

//  {$include 'QUB_Task_IdlStim_Properties.htx'}
end;

procedure TQUB_Task_IdlStim_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  DataSource             := DataSource;
end;

//Data Source
procedure TQUB_Task_IdlStim_PropertiesForm.chkDataSourceClick(Sender: TObject);
begin
  inherited;
  DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;


//******************************************************************************
end.

