unit QUB_Task_Kalman_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry, Math,

RegistryInterface,

Alloc, AllocTypes,
QUB_Global,
QUB_QVO_QVD,
QUB_Task,
QUB_Properties,
QUB_Task_Properties,
QUB_Node;

//{$include 'QUB_Task_Kalman_Properties.hct'}

const


DEF_KALMAN_MEAS_STD           :Double  = 0.3;
DEF_KALMAN_JUMP               :Double  = 0.3;
DEF_KALMAN_BASELINE           :Integer = 0;
DEF_KALMAN_BASELINE_STD       :Double  = 0.01;
DEF_KALMAN_PERIODIC           :Integer = 1;
DEF_KALMAN_PERIODIC_FREQ      :Double  = 60.0;
DEF_KALMAN_PERIODIC_STD       :Double  = 0.01;
DEF_KALMAN_PERIODIC_FIT       :Integer = 1;

DEF_PREVIEW_SIZE               = 3200;
MAX_PREVIEW_SIZE               = 12800;

WM_NEEDREFRESH                 = 1235;


type


TQUB_Task_Kalman_PropertiesForm = class(TQUB_Task_PropertiesForm)
  Label3                          :TLabel;
  panPreview: TPaintBox;
  panDataSource                   :TPanel;
  panDataSourceFile               :TPanel;
  chkDataSourceWholeFile          :TRadioButton;
  chkDataSourceSelectionList      :TRadioButton;
  chkDataSourceSelection          :TRadioButton;
  panBaseline                     :TPanel;
  Label1                          :TLabel;
  txtBaselineStd                  :TEdit;
  panStimChannel                  :TPanel;
  Label7                          :TLabel;
  txtStimChannel                  :TEdit;
  udStimChannel                   :TUpDown;
  Panel2                          :TPanel;
  Panel3                          :TPanel;
  Label5                          :TLabel;
  txtJump                         :TEdit;
  txtMeasStd                      :TEdit;
  Label6                          :TLabel;
  chkBaseline                     :TCheckBox;
  Label8                          :TLabel;
  chkPeriodic                     :TCheckBox;
  Label9                          :TLabel;
  txtPeriodicStd                  :TEdit;
  Label14                         :TLabel;
  txtPeriodicFreq                 :TEdit;
  lblStatus                       :TLabel;
  Label4                          :TLabel;
  txtOutName                      :TEdit;
  btnPickOutName                  :TButton;
  Label10                         :TLabel;
    chkPeriodicFit: TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender :TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure chkDataSourceClick(Sender: TObject);
  procedure panPreviewPaint(Sender: TObject);
  procedure txtOutNameKeyPress(Sender: TObject; var Key: Char);
  procedure txtOutNameExit(Sender: TObject);
  procedure btnPickOutNameClick(Sender: TObject);

  procedure prOnExitWrap(Sender: TObject);
  procedure prOnCheckClickWrap(Sender: TObject);

protected

  prData                          :TQUB_QVO_QVD;
  prPreviewD                      :TPADouble;        // data (Y)
  prPreviewT                      :TPADouble;        // track (sum of jumps, baseline, periodic)
  prPreviewF                      :TPADouble;        // filtered (selected stuff subtracted)
  prN                             :Integer;
  prPreviewLo                     :Double;
  prPreviewHi                     :Double;

  prNeedRefresh                   :Boolean;
  prTask                          :TQUB_Task;

  prDataSource                    :TQUB_Task_Data_Source;
  prOutName                       :String;

  procedure prSetDataSource(s :TQUB_Task_Data_Source);
  procedure prSetOutName(s :String);

  procedure prOnNeedRefresh(var Message: TMessage); message WM_NEEDREFRESH;

  procedure prUpdateComponents; override;

  procedure prSetupPreview;
  procedure prInvalidate;
  procedure prOnTaskDone(Sender :TObject);

public

  //Data Source
  property DataSource                     :TQUB_Task_Data_Source     read prDataSource                write prSetDataSource;
  property OutName                        :String                    read prOutName                   write prSetOutName;

end;


implementation


uses

QUB_Main,
QUB_GUIInterface,
QUB_QFS_QDF,
QUB_Task_Kalman;


{$R *.DFM}


//******************************************************************************
//Data Source
procedure TQUB_Task_Kalman_PropertiesForm.prSetDataSource(s :TQUB_Task_Data_Source);
begin
  prDataSource                       := s;
  chkDataSourceWholeFile.Checked     := DataSource =  DATA_SOURCE_WHOLEFILE;
  chkDataSourceSelectionList.Checked := DataSource =  DATA_SOURCE_SELECTIONLIST;
  chkDataSourceSelection.Checked     := DataSource =  DATA_SOURCE_SELECTION;
end;

procedure TQUB_Task_Kalman_PropertiesForm.prSetOutName(s :String);
begin
  txtOutName.Text := s;
  prOutName       := s;
end;


//******************************************************************************
procedure TQUB_Task_Kalman_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddEditProperty(TQUB_Property_Edit.Create('["MeasStd"]',
                    '', REGTYPE_NONE, DEF_KALMAN_MEAS_STD,
                    txtMeasStd,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Jump"]',
                    '', REGTYPE_NONE, DEF_KALMAN_JUMP,
                    txtJump,
                    AcceptFloatGreaterThan(0.0),
                    '') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Baseline"]',
                        '', REGTYPE_NONE, DEF_KALMAN_BASELINE,
                        chkBaseline)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["BaselineStd"]',
                    '', REGTYPE_NONE, DEF_KALMAN_BASELINE_STD,
                    txtBaselineStd,
                    AcceptFloatGreaterThan(0.0),
                    '') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Periodic"]',
                        '', REGTYPE_NONE, DEF_KALMAN_PERIODIC,
                        chkPeriodic)
                       .Enables(txtPeriodicFreq)
                       .Enables(txtPeriodicStd)
                       .Enables(chkPeriodicFit)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["PeriodicFreq"]',
                    '', REGTYPE_NONE, DEF_KALMAN_PERIODIC_FREQ,
                    txtPeriodicFreq,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PeriodicStd"]',
                    '', REGTYPE_NONE, DEF_KALMAN_PERIODIC_STD,
                    txtPeriodicStd,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["PeriodicFit"]',
                      '', REGTYPE_NONE, DEF_KALMAN_PERIODIC_FIT,
                      chkPeriodicFit)
                     );

  RegKeyName            := 'Software\QUB\Task Bkg Sub';
  ProfileCat            := 'Bkg-';

  WikiTopic             := 'Preprocessing:Bkg_Sub';

  prPreviewD              := DAlloc1(MAX_PREVIEW_SIZE);
  prPreviewT              := DAlloc1(MAX_PREVIEW_SIZE);
  prPreviewF              := DAlloc1(MAX_PREVIEW_SIZE);
  prTask                  := nil;
end;

procedure TQUB_Task_Kalman_PropertiesForm.FormDestroy(Sender :TObject);
begin
  try
    FreeMem(prPreviewD);
    FreeMem(prPreviewT);
    FreeMem(prPreviewF);
  finally
    inherited;
  end;
end;

procedure TQUB_Task_Kalman_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  prData                  := ViewportData.CurrData;
  OutName                 := ChangeFileExt(prData.FileName, '_kf.qdf');
  prSetupPreview;
  prInvalidate;
end;

procedure TQUB_Task_Kalman_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;
  if prTask <> nil then prTask.Stop;
  prTask := nil;
end;

procedure TQUB_Task_Kalman_PropertiesForm.prSetupPreview;
var
  Sel              :TQUB_FirstLast;
  s, o             :Double;
  i, c             :Integer;
  y, ylo, yhi      :Double;
  BufferI          :TPAInteger; // points halfway thru prPreviewD
begin
  Sel              := prData.DataSel;
  if Sel.First >= Sel.Last then begin
    Sel.Last       := Min(Sel.First + DEF_PREVIEW_SIZE - 1, prData.DataObject.Segments.TotalDataCount-1);
  end;
  Sel.Last         := Min(Sel.First + MAX_PREVIEW_SIZE - 1, Sel.Last);
  prN              := Sel.Last - Sel.First + 1;

  c                := prData.ActiveChannel;
  s                := 1.0 / (prData.DataObject.Scaling * prData.DataChannelScaling[c]);
  o                := prData.DataChannelOffset[c];
  BufferI          := TPAInteger(@(prPreviewD[MAX_PREVIEW_SIZE div 2]));
  prData.LoadDataChannel(BufferI, c, Sel, True, True, True, True);
  ylo              := 1e30;
  yhi              := -1e30;
  for i := 0 to prN - 1 do begin
    y              := s * BufferI[i] + o;
    if y < ylo then ylo := y;
    if y > yhi then yhi := y;
    prPreviewD[i]  := y;
  end;
  prPreviewLo      := ylo;
  prPreviewHi      := yhi;
end;

procedure TQUB_Task_Kalman_PropertiesForm.chkDataSourceClick(Sender: TObject);
begin
  inherited;
  DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;

procedure TQUB_Task_Kalman_PropertiesForm.txtOutNameKeyPress(
  Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtOutNameExit(Sender);
  end;
end;

procedure TQUB_Task_Kalman_PropertiesForm.txtOutNameExit(Sender: TObject);
begin
  inherited;
  prOutName := txtOutName.Text;
end;

procedure TQUB_Task_Kalman_PropertiesForm.btnPickOutNameClick(Sender: TObject);
var
  Dlg :TSaveDialog;
begin
  inherited;
  Dlg := TSaveDialog.Create(Application);
  try
    Dlg.Filter := 'QUB data files (*.qdf)|*.qdf';
    Dlg.Options := [ofPathMustExist, ofEnableSizing];
    Dlg.FileName := OutName;
    if Dlg.Execute then begin
      OutName := Dlg.FileName;
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Task_Kalman_PropertiesForm.prOnExitWrap(Sender: TObject);
begin
  prInvalidate;
end;

procedure TQUB_Task_Kalman_PropertiesForm.prOnCheckClickWrap(Sender: TObject);
begin
  prInvalidate;
end;

procedure TQUB_Task_Kalman_PropertiesForm.prUpdateComponents;
begin
  inherited;
  prInvalidate;
end;

procedure TQUB_Task_Kalman_PropertiesForm.prInvalidate;
begin
  prNeedRefresh := True;
  PostMessage(Self.Handle, WM_NEEDREFRESH, 0, 0);
end;

procedure TQUB_Task_Kalman_PropertiesForm.prOnNeedRefresh(var Message: TMessage);
begin
  if not prNeedRefresh then Exit;
  lblStatus.Caption := 'Filtering preview...please wait...';

  if prTask <> nil then begin
    prTask.Stop;
  end else begin
    prTask := TQUB_Task_Kalman.Create(True{Suspended});
    TQUB_Task_Kalman(prTask).OnTerminate  := prOnTaskDone;
    TQUB_Task_Kalman(prTask).Properties   := Properties.Clone(True);
    TQUB_Task_Kalman(prTask).Sampling     := prData.DataObject.Sampling;
    TQUB_Task_Kalman(prTask).PreviewD     := prPreviewD;
    TQUB_Task_Kalman(prTask).PreviewT     := prPreviewT;
    TQUB_Task_Kalman(prTask).PreviewF     := prPreviewF;
    TQUB_Task_Kalman(prTask).PreviewN     := prN;
    TQUB_Task_Kalman(prTask).Resume;
    prNeedRefresh := False;
  end;
end;

procedure TQUB_Task_Kalman_PropertiesForm.prOnTaskDone(Sender :TObject);
begin
  prTask := nil;
  if not Visible then begin
    // nothing
  end else if prNeedRefresh then begin
    SyncGUIProc(prInvalidate);
  end else begin
    lblStatus.Caption := '';
    panPreview.Invalidate;
  end;
end;

procedure TQUB_Task_Kalman_PropertiesForm.panPreviewPaint(Sender: TObject);
var
  i, n              :Integer;
  w, h              :Integer;
  y0, y1, dy        :Double;

  function Inflate(var a, b :Double): Double;
  var
    d,x    :Double;
  begin
    d      := b - a;
    if d = 0 then d := 1;
    x      := (d * 15) / 100;
    a      := a - x;
    b      := b + x;
    Result := d + 2*x;
  end;
begin
  inherited;

  n                 := prN;
  w                 := panPreview.Width;
  h                 := panPreview.Height div 2;

  with panPreview.Canvas do begin
    y0 := prPreviewLo;
    y1 := prPreviewHi;
    dy := Inflate(y0, y1);

    Brush.Style := bsSolid;
    Brush.Color := clWhite;
    Pen.Color   := clWhite;
    Rectangle(0, 0, w, 2*h);

    // black data
    Pen.Color := clBlack;
    Pen.Width := 1;
    MoveTo(0, Round(((y1 - prPreviewD[0])*h) / dy));
    for i := 1 to n - 1 do begin
      LineTo((i*w) div n, Round(((y1 - prPreviewD[i])*h) / dy));
    end;

    // fuchsia track
    Pen.Color    := clFuchsia;
    MoveTo(0, Round(((y1 - prPreviewT[0])*h) / dy));
    for i := 1 to n - 1 do begin
      LineTo((i*w) div n, Round(((y1 - prPreviewT[i])*h) / dy));
    end;

    // red filtered
    y0 := prPreviewF[0];
    y1 := y0;
    for i := 1 to n - 1 do begin
      if prPreviewF[i] < y0 then y0 := prPreviewF[i];
      if prPreviewF[i] > y1 then y1 := prPreviewF[i];
    end;
    dy := Inflate(y0, y1);

    Pen.Color   := clRed;
    MoveTo(0, h+Round(((y1 - prPreviewF[0])*h) / dy));
    for i := 1 to n - 1 do begin
      LineTo((i*w) div n, h+Round(((y1 - prPreviewF[i])*h) / dy));
    end;
  end;
end;


//******************************************************************************
end.

