unit QUB_Task_MIL;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QVO_QVM, QUB_QVO_QVR,

QUB_QFS_QMF,

QUB_Viewport_Data,
QUB_Viewport_Histogram,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_Mil_Properties,

QUB_MIL_Global,
QUB_MIL,

Chart, Series,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, QUB_Viewport_Results,
QUB_Results_Histograms,
QUB_GUIInterface;


type


TQUB_MIL_Opt_Params  = record

  OptLLConv                  :Double;
  OptGradConv                :Double;
  OptMaxIter                 :Integer;
  OptRestarts                :Integer;
  OptSearchLimit             :Double;
  OptMaxStep                 :Double;
  OptMode                    :Integer;
  OptSegments                :Integer;
  OptShowIterations          :Integer;
  OptJoinSegments            :Integer;
  OptGroupViterbi            :Integer;
  OptHistBinCount            :Integer;
  OptSmoothBinning           :Integer;
  OptSqrtOrdinate            :Integer;

end;


TQUB_Task_MIL  = class(TQUB_Task_Data)
protected

  prOptName                :String;

  //Source Objects
  prModel                  :TQUB_QVO_QVM;
  //prModels                 :TList;                    //Unused

  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  //Mil
  //prLibName                :String;                   //Unused

  //prHost                   :String;                   //Unused

  //prGhCount                :Integer;                  //Unused
  //prVisGroup               :Integer;                  //Unused

  //prEventCount             :Integer;                  //Unused

  prGroupViterbi           :Boolean;

  prConfig                 :INode;
  prDataDescription        :INode;
  prDatasets               :Array of INode;
  prUseChunks              :Boolean; // or caller-supplied data
  prReportCB               :INodeFunc;
  prResultCB               :INodeFunc;
  prPercentCB              :INodeFunc;

  prResultNode             :INode;
  prReportNode             :INode;
  prHistResult             :TQUB_QVO_QVR;


  procedure prDoGroupViterbi;

  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prPublishResults;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;
  function prFindStimChannel: String;

  procedure prSetModel(Model :TQUB_QVO_QVM);

  procedure prWriteCheckLine(Result :INode);
  procedure prDoCheckReport; // using prResultNode, synchronized
  function prMakeTimeConstantsReport(FirstTC, Hist :INode): String;

  function prReportCallback( arg :INode ): INode;
  procedure prDoReport;
  function prResultCallback( arg :INode ): INode;
  procedure prDoResult;
  function prPercentCallback( arg :INode ): INode;


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure AddDataset(DS :INode); // replaces sel/list/file/list

  procedure Stop; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model             :TQUB_QVO_QVM              read prModel             write prSetModel;

  //Parameters
  property Config            :INode                     read prConfig            write prConfig; // set before doing properties and the defaults will be cur.properties.  set after and the defaults are the mysterious ones from qubopt.
  property ReportCB          :INodeFunc                 read prReportCB          write prReportCB;
  property ResultCB          :INodeFunc                 read prResultCB          write prResultCB;

  property GroupViterbi      :Boolean                   read prGroupViterbi      write prGroupViterbi;


end;

procedure SolveTcrit( TimeConsts :INode );

function MILFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; rptCB, rsltCB, pctCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_miltreeiface@20';


var


QUB_Task_MIL_LastLL           :Double;


implementation


uses


maxill_utils,
QUBFits,
QUB_Node_Model,
QUB_Python_Data,
QUB_Main;


//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _M I L                         //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_MIL.Create(Suspended :Boolean);
begin
  inherited;

  prName       := 'Mil optimizer';

  prConfig     := TNode.Create('Properties');
  prConfig['StopFlag'].SetData( 0 );

  prUseChunks  := True;

  prReportCB   := TNodeMeth.Create( Self.prReportCallback );
  prResultCB   := TNodeMeth.Create( Self.prResultCallback );
  prPercentCB  := TNodeMeth.Create( Self.prPercentCallback );

  prHistResult := nil;

end;

destructor TQUB_Task_MIL.Destroy;
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_MIL.Stop;
begin

  prConfig['StopFlag'].Data[0] := 1;

  inherited;

end;


//******************************************************************************
procedure TQUB_Task_Mil.AddDataset(DS :INode);
var
  i         :Integer;
begin
  if prUseChunks then SetLength(prDatasets, 0);
  prUseChunks  := False;

  i            := Length(prDatasets);
  SetLength(prDatasets, i+1);
  prDatasets[i] := DS;
end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Mil.prDoTask;
var
  i               :Integer;
  StimName        :String;

  DataSetPtrs     :TPAPointer;
  rptCB, rsltCB   :INodeCFunc;
  pctCB           :INodeCFunc;
  ResultNode      :INode;
  ResultNodePtr   :TQUB_QFS_PNodeMem;

  //MaxillModel     :TMaxill_Model;
  //MaxillData      :TMaxill_Interval_Data;
  //NewLL           :Double;
begin
  if Length(prDataSets) = 0 then begin
    Exit;
  end;

  GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * (Length(prDatasets) + 1) );
  for i := 0 to Length(prDataSets) - 1 do begin
    DataSetPtrs[i] := Pointer(prDataSets[i].GetNodeMem);
  end;
  DataSetPtrs[ Length(prDataSets) ] := nil;

  rptCB  := TNodeCFunc.Wrap( prReportCB );
  rsltCB := TNodeCFunc.Wrap( prResultCB );
  pctCB  := TNodeCFunc.Wrap( prPercentCB );

  StimName := prFindStimChannel;
  if StimName <> '' then begin
    prError := True;
    prErrorMsg := 'Model variable '+StimName+' is assigned to a data channel.'#13#10'Try MSL, which can handle a changing stimulus.';
    prWikiTopic := 'Modeling:MSL';
  end else if prModel.ModelObject.ChannelCount > MIL_MAX_NCHANNEL then begin
    prError := True;
    prErrorMsg := 'Channel count is too high for MIL. Try MAC instead.';
    prWikiTopic := 'Modeling:MIL';
  end else begin

    try
      prConsoleMsg  := prOptName + ' optimizer running ... ';
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);

      // set up temp results for histograms
      if prConfig['Mode'].DataAsString = MILModeStringOf(MIL_MODE_OPT) then begin
        prResultNode  := TNode.Create('MIL ' + ChangeFileExt(ExtractFileName(Data.FileName), ''));
        Synchronize( prPublishResults );
      end;

      ResultNodePtr := MILFromDll( prConfig.GetNodeMem, DataSetPtrs, rptCB, rsltCB, pctCB );
      ResultNode    := TNode.CreateUsing( ResultNodePtr );
      Node_DECREF( ResultNodePtr );

      // MILHistogramsToLogMS( ResultNode );
      prResultNode := ResultNode;
      Synchronize( prPublishResults );

      if prConfig['Mode'].DataAsString = MILModeStringOf(MIL_MODE_EVAL) then begin
        (* Example: how to call maxill.MIL:
        // Model   := QVM.GetAsNode;
        // DataSet := DataNoder.GetChosenIdealization;
        NewLL := 0;
        MaxillModel := nil;
        MaxillData  := nil;
        try
          MaxillModel  := TMaxill_Model.Create(prResultNode.Find('Properties').Find('ModelFile'));
          for i := 0 to Length(prDataSets) - 1 do begin
            MaxillData := TMaxill_Interval_Data.Create(prDataSets[i], MaxillModel);
            NewLL := NewLL + inter_ll_obj(MaxillModel, MaxillData);
            MaxillData.Free;
            MaxillData := nil;
          end;
        except
          on ex :Exception do begin
            Beep; // see ex.Message
          end;
        end;
        MaxillModel.Free;
        MaxillData.Free;
        prResultNode['NewLL'].SetData(NewLL);
        *)
        Synchronize(prDoCheckReport);
      end;

      if prGroupViterbi then begin
        Synchronize(prDoGroupViterbi);
      end;
    except
      prError    := True;
      prErrorMsg := 'Unknown error in MIL library';
      prWikiTopic := 'Error:Task';
    end;
  end;
  
  FreeMem( DataSetPtrs );

end;

procedure TQUB_Task_MIL.prPublishResults;
begin
  prResultNode.Name   := 'MIL ' + ChangeFileExt(ExtractFileName(Data.FileName), '');

  prResultNode['ResultType'].SetData( 'MIL' );
  if (prDataDescription <> nil) and not prDataDescription.IsNull then prResultNode.AppendChild( prDataDescription );

  if prSelectionList <> nil then begin          // if the selection list is 'MIL Results',
    prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
    prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
  end;

  if prHistResult <> nil then begin
    ViewportResults.CloseResults( prHistResult );
    prHistResult  := nil;
  end;
  if prResultNode.Find('DataSet').IsNull then begin // the temp. histogram result has no datasets
    prHistResult    := ViewportResults.AddResults( prResultNode, False );
  end else begin
    ViewportResults.AddResults( prResultNode, False );
  end;
end;

procedure TQUB_Task_Mil.prDoExecuteStart;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  inherited;

end;

procedure TQUB_Task_Mil.prDoExecuteEnd;
begin

  inherited;

  if prSelectionList <> nil then begin
    prSelectionList.UnLock;
  end;

  if Model <> nil then begin
    Model.UnlockAny;
  end;

end;


function TQUB_Task_Mil.prDoExecuteInitialize :Boolean;
var
  MdlNode       :INode;
begin
  Result        := False;

  MdlNode       := GetModelAsNode;
  MdlNode.SetData( 'Initial' );

  if prConfig.Find('ModelFile').IsNull then prConfig.AppendChild( MdlNode );
end;

//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Mil.prChunkData :Boolean;
var
  DataSets         :INode;
  DS               :INode;
  RemDS            :INode;
  i                :Integer;
begin

  if prUseChunks then begin

    if prScripting then begin
      DataSets         := DataNoderScript.GetChosenIdealization(TNode.CreateNull);
    end else begin
      DataSets         := DataNoder.GetChosenIdealization(TNode.CreateNull);
    end;
    prDataDescription := DataSets.Find('DataDescription');

    DS               := DataSets.Find('DataSet');
    while not DS.IsNull do begin
      if 0<>prConfig['join segments'].Data[0] then DS['Join Segments'];

      if DS.Find('Segment').IsNull then begin
        RemDS          := DS;
        DS             := DS.SiblingSameName;
        DataSets.RemoveChild( RemDS );
      end else begin
        SetLength( prDataSets, Length(prDataSets) + 1 );
        prDataSets[ Length(prDataSets) - 1 ] := DS;

        DS             := DS.SiblingSameName;
      end;
    end;

  end;

  Result               := True;
  for i := 0 to Length(prDataSets) - 1 do begin
    DS                 := prDataSets[i].Find('Segment').Find('DwellCount');
    if (DS.DataCount > 0) and (DS.Data[0] > 0) then begin
      Result           := False; // found at least one event
      Break;
    end;
  end;
  if Result then begin
    if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
      prError          := True;
      prErrorMsg       := 'There is no idealized data in the file list.'#13#10'To set up the file list, right-click MIL and put a check next to each file you''d like to include.';
      prWikiTopic      := 'Error:No_Data_FileList';
    end else begin
      prError            := True;
      prErrorMsg         := 'No idealized data.  Check the Data Source.';
      prWikiTopic        := 'Error:No_Data';
    end;
  end;
end;

procedure TQUB_Task_Mil.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MilChunk :TQUB_Thread_Mil_PChunk;
begin

  MilChunk := TQUB_Thread_Mil_PChunk(Chunk);

  if MilChunk.ClassDwells <> nil then begin
    FreeMem(MilChunk.ClassDwells);
    MilChunk.ClassDwells := nil;
  end;

  FreeMem(MilChunk);

end;

function TQUB_Task_Mil.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Mil_Chunk));

  Result.Sel := Sel;

end;

function TQUB_Task_Mil.prFindStimChannel: String;
var
  Model1Chan      :INode;
  MaxillModel     :TMaxill_Model;
  d, v            :Integer;
  ExpCond, EC     :INode;
begin
  MaxillModel     := nil;
  Result          := '';
  try
    try
      // maxill complains about multichannel models
      // but it's good at enumerating variables in play
      Model1Chan      := prConfig.Find('ModelFile').Clone(True);
      Model1Chan['ChannelCount'].Data[0] := 1;
      MaxillModel     := TMaxill_Model.Create(Model1Chan);

      // return the first variable that's assigned to an a/d channel by a dataset
      for d := 0 to Length(prDataSets) - 1 do begin
        ExpCond       := prDataSets[d].Find('ExpCond');
        for v := 1 to MaxillModel.Variables.Count - 1 do begin
          EC          := ExpCond.Find(MaxillModel.Variables[v]).Find('ChannelIndex');
          if (EC.DataCount > 0) and (EC.Data[0] >= 0) then begin
            Result    := MaxillModel.Variables[v];
            Exit;
          end;
        end;
      end;
    except
    end;
  finally
    MaxillModel.Free;
  end;
end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MIL.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_MIL_PropertiesForm do begin

    prConfig                    := Properties.Clone(True);
    prConfig['StopFlag'].SetData( 0 );

    prOptName                   := 'MIL';

    Self.GroupViterbi           := Properties['GroupViterbi'].Data[0];

  end;//with
end;

procedure TQUB_Task_MIL.OnPutProperties(Sender :TObject);
begin
end;


//******************************************************************************
procedure TQUB_Task_MIL.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then prModel.LockAny;
end;



//****************************************************************************//
//                                                                            //
//                            R E S U L T S                                   //
//                                                                            //
//****************************************************************************//
function TQUB_Task_MIL.prPercentCallback( arg :INode ): INode;
begin
  prUpdateStatus( Integer( arg.Data[0] ), prStTime );
  Result := arg;
end;

function TQUB_Task_MIL.prReportCallback( arg :INode ): INode;
begin
  prReportNode := arg;
  Synchronize( prDoReport );
  Result := arg;
end;

procedure TQUB_Task_MIL.prDoReport;
begin
  // just print the node data's text rep'n for now
  if 0<>prConfig['ShowIterations'].Data[0] then begin
    prConsoleMsg := prReportNode.GetDataAsString;
    prConsoleCode := MESSAGE_ERROR;
    prWriteNewConsole;
  end;
end;

function TQUB_Task_MIL.prResultCallback( arg :INode ): INode;
begin
  prResultNode := arg;
  Synchronize( prDoResult );
  Result := arg;
end;

procedure TQUB_Task_MIL.prWriteCheckLine(Result :INode);
var
  DwellCount         :Integer;
  DataSet, Seg       :INode;
  Grad, Ix           :INode;
  Text               :String;
  i                  :Integer;
begin
  DwellCount := 0;

  DataSet := Result.Find('DataSet');
  while not DataSet.IsNull do begin
    Seg := DataSet.Find('Segment');
    while not Seg.IsNull do begin
      DwellCount := DwellCount + Seg['DwellCount'].Data[0];
      Seg := Seg.SiblingSameName;
    end;
    DataSet := DataSet.SiblingSameName;
  end;

  Text := '';

  Ix := Result.Find('Index');
  if not Ix.isNull then begin
    Text := Format('Segment %d | ', [Integer(Ix.Data[0])+1]);
  end;

  Text := Text + Format('LL: %f'#9'Events: %d'#9'Last LL: %f'#9'Grad:', [Double(Result['LL'].Data[0]), DwellCount, QUB_Task_MIL_LastLL]);
  Grad := Result['Gradient'];
  for i := 0 to Grad.DataCount - 1 do begin
    Text := Text + Format(' %f', [Double(Grad.Data[i])]);
  end;
  QUB_Task_MIL_LastLL := Double(Result['LL'].Data[0]);

  if 0<>prConfig['ShowIterations'].Data[0] then begin
    prConsoleMsg := Text;
    prConsoleCode := MESSAGE_MESSAGE;
    prWriteNewConsole;
  end;
end;

procedure TQUB_Task_MIL.prDoCheckReport;
begin
  prWriteCheckLine( prResultNode );

  if 0<>prConfig['ShowIterations'].Data[0] then begin
    prConsoleMsg := prMakeTimeConstantsReport(prResultNode['DataSet'].Find('HistogramGroup').Find('TimeConstants'),
                                            prResultNode['DataSet'].Find('HistogramGroup').Find('Histogram'));
    prConsoleCode := MESSAGE_MESSAGE;
    prWriteNewConsole;
  end;
end;

procedure TQUB_Task_MIL.prDoResult;
var
  Final            :Boolean;
  Iter             :Integer;
  DwellCount       :Integer;
  Grad             :INode;
  DataSet, Seg     :INode;
  DataSetIx, SegIx :Integer;
  Text             :String;
  i                :Integer;
  BySegment        :Boolean;
  Rate             :INode;
  FirstTC          :INode;
  HistGroup        :TQUB_Results_HistGroup;
  HistGroupNode    :INode;
  Hist             :INode;
  LLperEvent       :Double;
begin
  SetRatesFromNode( prModel.ModelObject, prResultNode['ModelFile']['Rates'] );

  Text              := '';

  Final             := not prResultNode.Find('Final').IsNull;
  if Final then begin
    Text            := Text + 'Finished ';
  end;

  Iter              := prResultNode['Iterations'].Data[0];
  BySegment         := prConfig['use segments'].DataAsString = MILSegModeStringOf(MIL_MODE_SEG_EACH);

  if BySegment then begin
    SegIx := Integer(prResultNode['Index'].Data[0]);
    if Final or (0<>prConfig['ShowIterations'].Data[0]) then Text  := Text + Format('seg %d ', [SegIx + 1]);

    HistGroupNode   := prResultNode['HistogramGroup'];
    HistGroup       := prHistResult.Histograms.Groups[ SegIx ];
    if HistGroup <> nil then begin
      HistGroup.Node   := HistGroupNode;
    end else begin
      prHistResult.Histograms.AddGroup( HistGroupNode );
      prHistResult.Histograms.CurrIndex := prHistResult.Histograms.GroupCount - 1;
    end;

    DwellCount := prResultNode['DwellCount'].Data[0];
  end else begin
    DwellCount := 0;

    DataSetIx := 0;
    DataSet := prResultNode.Find('DataSet');
    while not DataSet.IsNull do begin
      Seg := DataSet.Find('Segment');
      while not Seg.IsNull do begin
        DwellCount := DwellCount + Seg['DwellCount'].Data[0];
        Seg := Seg.SiblingSameName;
      end;

      HistGroupNode       := DataSet['HistogramGroup'];
      HistGroup           := prHistResult.Histograms.Groups[ DataSetIx ];
      if HistGroup <> nil then begin
        HistGroup.Node    := HistGroupNode;
      end else begin
        prHistResult.Histograms.AddGroup( HistGroupNode );
        prHistResult.Histograms.CurrIndex := prHistResult.Histograms.GroupCount - 1;
      end;

      Inc(DataSetIx);
      DataSet := DataSet.SiblingSameName;
    end;
  end;

  if Final then begin
    Text := Text + Format('[%d events] after %d iterations', [DwellCount, Iter]);
  end else if 0<>prConfig['ShowIterations'].Data[0] then begin
    Grad              := prResultNode['Gradient'];

    Text := Text + Format('iter: %d'#9'LL: %.4f'#9'grad:', [Iter, Double(prResultNode['LL'].Data[0])]);
    for i := 0 to Grad.DataCount - 1 do begin
      Text := Text + Format(#9'%.6g', [Double(Grad.Data[i])]);
    end;
  end;

  if Final then begin // or (Iter = 1) or not Params.QuietOutput then begin
    Text := Text + Format(#13#10'Rates:'#13#10'  s1'#9's2'#9'%-10s %-10s %-10s %-10s', ['k0','k0.std','k1','k1.std']);

    Rate := prResultNode['ModelFile']['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      Text := Text + Format(#13#10'  %d'#9'%d'#9'%-10.6g %-10.6g %-10.6g %-10.6g',
                [ Integer(Rate['States'].Data[0]) + 1, Integer(Rate['States'].Data[1]) + 1,
                  Double(Rate['k0'].Data[0]),          Double(Rate['dk0'].Data[0]),
                  Double(Rate['k1'].Data[0]),          Double(Rate['dk1'].Data[0]) ] );
      Text := Text + Format(#13#10'  %d'#9'%d'#9'%-10.6g %-10.6g %-10.6g %-10.6g',
                [ Integer(Rate['States'].Data[1]) + 1, Integer(Rate['States'].Data[0]) + 1,
                  Double(Rate['k0'].Data[1]),          Double(Rate['dk0'].Data[1]),
                  Double(Rate['k1'].Data[1]),          Double(Rate['dk1'].Data[1]) ] );
      Rate := Rate.SiblingSameName;
    end;
  end;

  if Final then begin
    Text := Text + #13#10 + 'Time Constants (Tau, Amp):';
    if BySegment then begin
      FirstTC := prResultNode.Find('HistogramGroup').Find('TimeConstants');
      Hist    := prResultNode.Find('HistogramGroup').Find('Histogram');
      Text    := Text + prMakeTimeConstantsReport(FirstTC, Hist);
    end else begin
      DataSet := prResultNode.Find('DataSet');
      while not DataSet.IsNull do begin
        Text    := Text + #13#10' ' + ExtractFileName(DataSet['FileName'].DataAsString) + ':';
        FirstTC := DataSet.Find('HistogramGroup').Find('TimeConstants');
        Hist    := DataSet.Find('HistogramGroup').Find('Histogram');
        Text    := Text + prMakeTimeConstantsReport(FirstTC, Hist);
        DataSet := DataSet.SiblingSameName;
      end;
    end;
  end;

  if Final and not BySegment then begin
    try
      LLperEvent := Double(prResultNode['LL'].Data[0]);
      if LLperEvent <> 0.0 then LLperEvent := LLperEvent / DwellCount;
      Text  := Text + Format(#13#10'LL: %f'#9'Events: %d'#9'LL/event: %.3f'#9'Last LL: %f', [Double(prResultNode['LL'].Data[0]), DwellCount, LLperEvent, QUB_Task_MIL_LastLL]);
      QUB_Task_MIL_LastLL := Double(prResultNode['LL'].Data[0]);
    except
    end;
  end;

  if 0<>prConfig['ShowIterations'].Data[0] then begin
    if Text <> '' then begin
      prConsoleMsg := Text;
      prConsoleCode := MESSAGE_MESSAGE;
      prWriteNewConsole;
    end;
  end else if Final then begin
    Text := '';
    Rate := prResultNode['ModelFile']['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      Text := Text + Format('%-10.6g'#9'%-10.6g'#9'%-10.6g'#9'%-10.6g'#9,
                           [Double(Rate['k0'].Data[0]), Double(Rate['dk0'].Data[0]), Double(Rate['k0'].Data[1]), Double(Rate['dk0'].Data[1])]);
      Rate := Rate.SiblingSameName;
    end;
    Text := Text + Format('%-10.6g', [Double(prResultNode['LL'].Data[0])]);
    prConsoleMsg := Text;
    prConsoleCode := MESSAGE_MESSAGE;
    prWriteNewConsole;
  end;

  //ResultText
  if (Final) and (prResultText <> nil) then begin
    Text := Format('%14.10g'#9'%d'#9'%d'#9, [-Double(prResultNode['LL'].Data[0]), Iter, DwellCount]);
    Rate := prResultNode['ModelFile']['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      Text := Text + Format('%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9,
                            [
                             Double(Rate['k0'].Data[0]), Double(Rate['dk0'].Data[0]), Double(Rate['k1'].Data[0]), Double(Rate['dk1'].Data[0]),
                             Double(Rate['k0'].Data[1]), Double(Rate['dk0'].Data[1]), Double(Rate['k1'].Data[1]), Double(Rate['dk1'].Data[1])
                            ]);
      Rate := Rate.SiblingSameName;
    end;
    Grad := prResultNode['Gradient'];
    for i := 0 to Grad.DataCount - 1 do begin
      Text := Text + Format('%14.10g'#9, [Double(Grad.Data[i])]);
    end;
    prResultText^ := Text;
  end;

end;

function TQUB_Task_MIL.prMakeTimeConstantsReport(FirstTC, Hist :INode): String;
var
  TC, Tau, Amp           :INode;
  i                      :Integer;
begin
  Result := '';

  TC        := FirstTC;
  while not TC.IsNull do begin
    Result := Result + Format(#13#10'  Class %d:', [Integer(TC['Class'].Data[0])+1]);
    Tau := TC['Tau'];
    Amp := TC['Amp'];
    for i := 0 to Tau.DataCount - 1 do begin
      Result := Result + Format(' (%.4f, %.4f)', [Double(Tau.Data[i]), Double(Amp.Data[i])]);
    end;
    Result := Result + ' |  Mean: ' + TC['MeanTau'].DataAsString;

    TC := TC.SiblingSameName;
  end;

  if not FirstTC.IsNull then begin
    SolveTcrit( FirstTC );
    TC := FirstTC.Find('Tcrit');  // now TC means TCrit
    if (not TC.IsNull) and (TC.DataCount > 0) then begin
      Result := Result + #13#10'  Tcrit (closed): ';
      for i := 0 to TC.DataCount - 1 do begin
        if TC.Data[i] = 0.0 then begin
          Result := Result + '    --------    ';
        end else begin
          Result := Result + Format('%12.4g', [Double(TC.Data[i])]);
          Hist.AppendChild('VLine').SetData( log10(Double(TC.Data[i])) );
        end;
      end;
      prResultNode.AppendChild( TC.Clone(True) );
    end;
  end;
end;


const
  TCRIT_TOLER     = 1.0e-4;
  TCRIT_MAXITER   = 512;

// use newton's method:
// solve   Amp1 * exp( - Tcrit / Tau1 ) = Amp2 * (1 - exp( - Tcrit / Tau2 ))
// i.e. defining:
//         Alpha = Amp1 * exp( - Tcrit / Tau1 )
//         Beta  = Amp2 * exp( - Tcrit / Tau2 )
// solve   f(Tcrit)  = Alpha + Beta - Amp2 = 0
// with    f'(Tcrit) = - Alpha / Tau1 - Beta / Tau2
function OldSolveOneTcrit( Tau1, Amp1, Tau2, Amp2 :Double; var Iters :Integer ): Double;
var
  Tcrit, dTcrit         :Double;
  Fval, dFval           :Double;
  Alpha, Beta           :Double;
  Iter                  :Integer;
begin
  Tcrit        := (Tau1 + Tau2) / 2.0;

  Alpha        := Amp1 * Exp( - Tcrit / Tau1 );
  Beta         := Amp2 * Exp( - Tcrit / Tau2 );
  Fval         := Alpha + Beta - Amp2;

  for Iter := 1 to TCRIT_MAXITER do begin
    dFval      := - (Alpha / Tau1) - (Beta / Tau2);
    dTcrit     := - Fval / dFval;

    Tcrit      := Tcrit + dTcrit;

    Alpha      := Amp1 * Exp( - Tcrit / Tau1 );
    Beta       := Amp2 * Exp( - Tcrit / Tau2 );
    Fval       := Alpha + Beta - Amp2;

    if Abs(dTcrit) < TCRIT_TOLER then begin
      Break;
    end;
  end;

  Iters        := Iter;
  Result       := Tcrit;
end;

// using QUBFits:
// solve   Amp1 * exp( - Tcrit / Tau1 ) = Amp2 * (1 - exp( - Tcrit / Tau2 ))
// i.e. defining:
//         Alpha = Amp1 * exp( - Tcrit / Tau1 )
//         Beta  = Amp2 * exp( - Tcrit / Tau2 )
// solve   f(Tcrit)  = Alpha + Beta - Amp2 = 0
// with    f'(Tcrit) = - Alpha / Tau1 - Beta / Tau2
function SolveOneTcrit( Tau1, Amp1, Tau2, Amp2 :Double; var Iters :Integer ): Double;
var
  Vars                  :TStringList;
  X, Y, W               :Double;
  Data                  :TQF_Data;
  PData                 :TPQF_Data;
  Curve                 :TQUB_Curve;
  Fitter                :TQUB_Fitter;
begin
  QF_Data_Init(Data);
  Curve       := nil;
  Fitter      := nil;
  Vars        := nil;
  Result      := (Tau1 + Tau2) / 2.0;

  try
    X                     := 0.0;
    Y                     := 0.0;
    W                     := 1.0;
    Data.XX               := TPADouble(@X);
    Data.YY               := TPADouble(@Y);
    Data.WW               := TPADouble(@W);
    Data.N                := 1;
    Vars                  := TStringList.Create;
    Vars.AddObject('X', TObject(@X));
    Vars.AddObject('Y', TObject(@Y));
    QF_Data_SetVars(Data, Vars);

    Curve                 := CreateCurve('Custom');
    Curve.SetEqn(PChar(Format('%.8g * exp(-Tcrit / %.8g) + %.8g * exp(-Tcrit / %.8g) - %.8g',
                               [Amp1,               Tau1,   Amp2,               Tau2,   Amp2])));
    Curve.SetParamVal(0, Result, 1);

    Fitter                := CreateFitter('Levenburg-Marquardt');
    PData                 := @Data;
    Fitter.Fit(Curve, TPAQF_Data(@PData), 1);

    Result := Curve.GetParamVal(0);
    if (Result < Tau1) or (Result > Tau2) then begin
      Result := OldSolveOneTCrit(Tau1, Amp1, Tau2, Amp2, Iters);
    end;
  except
  end;

  if Curve <> nil then Curve.Free;
  if Fitter <> nil then Fitter.Free;
  if Vars <> nil then Vars.Free;
  QF_Data_FreeVars(Data);
end;

procedure SolveTcrit( TimeConsts :INode );
var
  Tau, Amp, Tcrit     :INode;
  Iter                :Integer;
  i                   :Integer;
begin
  Tau                 := TimeConsts.Find('Tau');
  Amp                 := TimeConsts.find('Amp');

  if Tau.IsNull or Amp.IsNull
  or (Tau.DataCount <> Amp.DataCount)
  or (Tau.DataCount < 2)
  then Exit;

  Tcrit               := TimeConsts['Tcrit'];
  Tcrit.SetupNumData( NODE_DATATYPE_DOUBLE, Tau.DataCount - 1, 1 );

  for i := 0 to Tcrit.DataCount - 1 do begin
    try
      Tcrit.Data[i]     := SolveOneTcrit( Double(Tau.Data[i  ]), Double(Amp.Data[i  ]),
                                          Double(Tau.Data[i+1]), Double(Amp.Data[i+1]), Iter );
      if (Iter >= TCRIT_MAXITER)
      or (Tcrit.Data[i] <= Tau.Data[i])
      or (Tcrit.Data[i] >= Tau.Data[i+1])
      then Tcrit.Data[i] := 0.0;
    except
      Tcrit.Data[i]     := 0.0;
    end;
  end;
end;


{
    This version works with idealized data
    It does NOT re-estimate
    Produces a list of selections of a given group
}
  {
  - prepare the groups
  - the group matrices G
  - the class matrices B

  - for each group allocate state vector Xt
  - for each group store the log-likelihood
  - allocate the backpath matrix(GroupCount x DataCount)


  - initialize

  - for each data point do



  - end data point

  - trace back

  - generates the selection list for the desired group, but drops first/last points if
    of a given class, if desired

  }
procedure TQUB_Task_MIL.prDoGroupViterbi;
var
  Groups     :TPAInteger;
  Classes    :TPAInteger;
  StateGroup :TPMInteger;
  StateClass :TPMInteger;
  BackPath   :TPMByte;
  Group      :TPAByte;
  LogLike    :TPADouble;
  BestLogLike:TPADouble;
  XB         :TPMDouble;
  X          :TPADouble;
  Alpha      :TPMDouble;
  GroupCount :Integer;
  StateCount :Integer;
  ClassCount :Integer;
  DataCount  :Integer;
  A          :TPMDouble;
  P0         :TPADouble;
  i, j, g, c :Integer;
  f, ch, p   :Integer;
  gg, pp, s  :Integer;
  l          :Boolean;
  SelIdx     :Integer;
  ThisClass  :Integer;
  ThisCount  :Integer;
  Sum        :Double;
  LL         :Double;
  ListName   :String;
  List       :TQUB_QVO_QVD_SelectionList;
  Sel        :TQUB_FirstLast;
  Seg        :INode;
  DwellCls   :TPAInteger;
  DwellDur   :TPASingle;
  Sampling   :Double; // in ms like DwellDur
  d          :Integer;
begin


  Groups     := nil;
  Classes    := nil;
  StateGroup := nil;
  StateClass := nil;
  LogLike    := nil;
  Alpha      := nil;
  A          := nil;
  P0         := nil;
  BestLogLike:= nil;
  XB         := nil;
  X          := nil;



  try


    try

      StateCount := Model.ModelObject.States.StateCount;

      A          := DAlloc2(StateCount, StateCount);
      P0         := DAlloc1(StateCount);

      X          := DAlloc1(StateCount);

      Classes    := IAlloc1(StateCount);
      Groups     := IAlloc1(StateCount);
      GroupCount := 0;
      ClassCount := 0;


      //A matrix
      Model.ModelObject.ADt := Double(prDataSets[0]['sampling'].Data[0]) * 1.0e-3; // TQUB_Thread_Mil_PChunk(prChunks.List[0]).Sampling;
      Model.ModelObject.CalculateMath;
      for i := 0 to StateCount - 1 do begin
        for j := 0 to StateCount - 1 do begin
          A[i, j] := Model.ModelObject.A[i, j];
        end;
      end;
      //Initial Probabilities
      Sum := 0;
      for i := 0 to StateCount - 1 do begin
        P0[i] := Model.ModelObject.States[i].Pr;
        Sum := Sum + P0[i];
      end;
      if Sum = 0 then begin
        for i := 0 to StateCount - 1 do begin
          P0[i] := 1 / StateCount;
        end;
      end else begin
        for i := 0 to StateCount - 1 do begin
          P0[i] := P0[i] / Sum;
        end;
      end;


      //Groups and Classes
      for i := 0 to StateCount - 1 do begin

        //Groups
        f := -1;
        for g := 0 to GroupCount - 1 do begin
          if Groups[g] = Model.ModelObject.States[i].Gr then begin
            f := g;
            Break;
          end;
        end;
        if f = -1 then begin
        //Add group
          Groups[GroupCount] := Model.ModelObject.States[i].Gr;
          Inc(GroupCount);
        end;

        //Classes
        f := -1;
        for c := 0 to ClassCount - 1 do begin
          if Classes[c] = Model.ModelObject.States[i].Cls then begin
            f := c;
            Break;
          end;
        end;
        if f = -1 then begin
        //Add class
          Classes[ClassCount] := Model.ModelObject.States[i].Cls;
          Inc(ClassCount);
        end;

      end;


      LogLike    := DAlloc1(GroupCount);
      Alpha      := DAlloc2(GroupCount, StateCount);

      BestLogLike:= DAlloc1(GroupCount);
      XB         := DAlloc2(GroupCount, StateCount);


      //The G and B "matrices"
      StateGroup := IAlloc2(GroupCount, StateCount);
      StateClass := IAlloc2(ClassCount, StateCount);

      //Group
      for g := 0 to GroupCount - 1 do begin
        for i := 0 to StateCount - 1 do begin
          if Groups[g] = Model.ModelObject.States[i].Gr then begin
            StateGroup[g, i] := 1;
          end else begin
            StateGroup[g, i] := 0;
          end;
        end;
      end;

      //Class
      for c := 0 to ClassCount - 1 do begin
        for i := 0 to StateCount - 1 do begin
          if Classes[c] = Model.ModelObject.States[i].Cls then begin
            StateClass[c, i] := 1;
          end else begin
            StateClass[c, i] := 0;
          end;
        end;
      end;


      //Clear Lists
      for g := 0 to GroupCount - 1 do begin

        ListName := 'Group ' + IntToStr(g) + ' selections';
        s := TQUB_QVO_QVD(Data).SelectionLists.Items.IndexOf(ListName);
        if s <> -1 then begin
          List := TQUB_QVO_QVD(Data).SelectionLists[s];
          if List <> nil then begin
            l := List.Locked;
            if l then List.Unlock;
            try
              List.Clear;
            finally
              if l then List.Lock;
            end;
          end;
        end;

      end;


      //For every Chunk
      ch := 0;
      for d := 0 to Length(prDataSets) - 1 do begin
        Sampling := prDataSets[0]['sampling'].Data[0];

        Seg := prDataSets[0].Find('Segment');
        while not Seg.IsNull do begin
          BackPath := nil;
          Group    := nil;

          try
            DataCount    := Seg['DwellCount'].Data[0];

            if DataCount > 0 then begin
              DwellCls    := TPAInteger(Seg['Classes'].DataPtr);
              DwellDur    := TPASingle( Seg['Durations'].DataPtr);

              //Do the processing here ***********************************************
              BackPath := TPMByte(CAlloc2(DataCount, GroupCount));
              Group    := TPAByte(CAlloc1(DataCount));

              SelIdx    := 0;
              ThisClass := DwellCls[SelIdx];
              ThisCount := Round(DwellDur[SelIdx] / Sampling);//Dec after each use
              Dec(ThisCount);

              //Initialize Alpha;
              for g := 0 to GroupCount - 1 do begin
                Sum := 0;
                for i := 0 to StateCount - 1 do begin
                  Alpha[g, i] := P0[i];
                  if (StateGroup[g, i] = 0) or (StateClass[ThisClass, i] = 0) then begin
                    Alpha[g, i] := 0;
                  end;
                  Sum := Sum + Alpha[g, i];
                end;
                if Sum = 0 then begin
                  Sum := 1e-30;
                end;
                for i := 0 to StateCount - 1 do begin
                  Alpha[g, i] := Alpha[g, i] / Sum;
                end;
                LogLike[g] := Ln(Sum);
              end;


              //Loop for every point
              for p := 1 to DataCount - 1 do begin

                if ThisCount = 0 then begin
                  Inc(SelIdx);
                  ThisClass := DwellCls[SelIdx];
                  ThisCount := Round(DwellDur[SelIdx] / Sampling);//Dec after each use
                end;

                for g := 0 to GroupCount - 1 do begin

                  gg  := 0;
                  //NewState = Alpha[gg] * A * G[g] * B[g]
                  Sum := 0;
                  for i := 0 to StateCount - 1 do begin
                    X[i] := 0;
                    for j := 0 to StateCount - 1 do begin
                      X[i] := X[i] + Alpha[gg, j] * A[j, i];
                    end;
                    if (StateGroup[g, i] = 0) or (StateClass[ThisClass, i] = 0) then begin
                      X[i] := 0;
                    end;
                    Sum := Sum + X[i];
                  end;
                  if Sum = 0 then begin
                    Sum := 1e-30;
                  end;
                  for i := 0 to StateCount - 1 do begin
                    X[i] := X[i] / Sum;
                  end;
                  BackPath[p, g] := gg;
                  BestLogLike[g] := Ln(Sum) + LogLike[gg];
                  for i := 0 to StateCount - 1 do begin
                    XB[g, i] := X[i];
                  end;

                  for gg := 1 to GroupCount - 1 do begin
                    //NewState = Alpha[gg] * A * G[g] * B[g]
                    Sum := 0;
                    for i := 0 to StateCount - 1 do begin
                      X[i] := 0;
                      for j := 0 to StateCount - 1 do begin
                        X[i] := X[i] + Alpha[gg, j] * A[j, i];
                      end;
                      if (StateGroup[g, i] = 0) or (StateClass[ThisClass, i] = 0) then begin
                        X[i] := 0;
                      end;
                      Sum := Sum + X[i];
                    end;
                    if Sum = 0 then begin
                      Sum := 1e-30;
                    end;
                    for i := 0 to StateCount - 1 do begin
                      X[i] := X[i] / Sum;
                    end;
                    LL := Ln(Sum) + LogLike[gg];
                    if LL > BestLogLike[g] then begin
                      BackPath[p, g] := gg;
                      BestLogLike[g] := LL;
                      for i := 0 to StateCount - 1 do begin
                        XB[g, i] := X[i];
                      end;
                    end;

                  end;//gg

                end;//g

                for g := 0 to GroupCount - 1 do begin

                  for i := 0 to StateCount - 1 do begin
                    Alpha[g, i] := XB[g, i];
                  end;

                  LogLike[g] := BestLogLike[g];

                end;//g

                Dec(ThisCount);

              end;


              //Backtrace
              g   := 0;
              gg  := g;
              Sum := LogLike[g];
              for g := 1 to GroupCount - 1 do begin
                if LogLike[g] > Sum then begin
                  gg  := g;
                  Sum := LogLike[g];
                end;
              end;
              Group[DataCount - 1] := gg;
              for p := DataCount - 2 downto 0 do begin
                Group[p] := BackPath[p + 1, Group[p + 1]];
              end;


              //Now make selection list for this chunk, one for every group
              for g := 0 to GroupCount - 1 do begin

                ListName := 'Group ' + IntToStr(g) + ' selections';
                s := TQUB_QVO_QVD(Data).SelectionLists.Items.IndexOf(ListName);
                if s = -1 then begin
                  List := TQUB_QVO_QVD(Data).AddList(ListName);
                end else begin
                  List := TQUB_QVO_QVD(Data).SelectionLists[s];
                end;

                if List = nil then Continue;

                l := List.Locked;
                if l then List.Unlock;
                try
                  p := 0;
                  repeat

                    if Group[p] = g then begin

                      Sel.First := Seg.Data[0] + p;
                      Sel.Last  := Sel.First;

                      pp := p;
                      while((pp < DataCount) and (Group[pp] = g)) do begin
                        Inc(pp);
                      end;
                      Inc(Sel.Last, pp - p - 1);

                      List.AddSelection(Sel.First, Sel.Last, '');

                      p := pp;

                    end else begin

                      Inc(p);

                    end;


                  until(p >= DataCount);
                finally
                  if l then List.Lock;
                end;

              end;

              //**********************************************************************



            end;


          finally

            Free2(TPMChar(BackPath));
            FreeMem(Group);


          end;

          Seg := Seg.SiblingSameName;
          Inc(ch);
        end;
      end;

    except

    end;


  finally


    FreeMem(Groups);
    FreeMem(Classes);
    Free2(TPMChar(StateGroup));
    Free2(TPMChar(StateClass));
    FreeMem(LogLike);
    Free2(TPMChar(Alpha));
    Free2(TPMChar(A));
    FreeMem(P0);
    FreeMem(BestLogLike);
    Free2(TPMChar(XB));
    FreeMem(X);
    

  end;


end;


//******************************************************************************
end.




