unit QUB_Task_Mil_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_Global,
QUB_QFS_Channels,
QUB_Properties,
QUB_Task,
QUB_Task_Properties,
QUB_Mil_Global,
QUB_Node,
QUB_DataList;


{$include 'QUB_Task_Mil_Properties.hct'}


type


TQUB_Task_Mil_PropertiesForm = class(TQUB_Task_PropertiesForm)
  Label4                          :TLabel;
  Label5                          :TLabel;
  Label6                          :TLabel;
  Label7                          :TLabel;
  Label8                          :TLabel;

  panParams                       :TPanel;

  txtMILHistBinCount              :TEdit;
  udMILHistBinCount               :TUpDown;
  txtMILLLConv                    :TEdit;
  txtMILGradConv                  :TEdit;
  txtMILMaxIter                   :TEdit;
  Label18                         :TLabel;
  txtMILRestarts                  :TEdit;
  Label19                         :TLabel;
  txtMILMaxStep                   :TEdit;
  chkMILShowIter                  :TCheckBox;

  Panel1                          :TPanel;
  chkMILModeOpt                   :TRadioButton;
  chkMILModeEval                  :TRadioButton;
  Label12                         :TLabel;
  txtMILSearchLimit               :TEdit;
  chkMILJoinSegments              :TCheckBox;
  Label11                         :TLabel;
  Panel3                          :TPanel;
  Label16                         :TLabel;
  chkSegModeAll                   :TRadioButton;
  chkSegModeEach                  :TRadioButton;
  chkGroupViterbi: TCheckBox;
    chkSmoothHistBins: TCheckBox;
    chkSqrtHist: TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

  procedure chkMILShowIterClick(Sender: TObject);


protected

  prDataList                      :TQUB_DataList;

  //Properties vs. Registry
  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public


end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_Mil_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := panParams.Left;
  prDataList.Top        := 50;
  prDataList.Width      := panParams.Width;
  prDataList.Height     := 100;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Histograms"]["BinCount"]',
                      'MIL Hist Bin Count', REGTYPE_INT, DEF_MILHISTBINCOUNT,
                      udMILHistBinCount,
                      txtMILHistBinCount,
                      AcceptIntBetween(MIN_MILHISTBINCOUNT, MAX_MILHISTBINCOUNT)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Histograms"]["SmoothBins"]',
                        '', REGTYPE_NONE, 1,
                        chkSmoothHistBins));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Histograms"]["Sqrt"]',
                        '', REGTYPE_NONE, 0,
                        chkSqrtHist));
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvLL"]',
                    'MIL LL Conv', REGTYPE_FLOAT, DEF_MILLLCONV,
                    txtMILLLConv,
                    AcceptFloatBetween(MIN_MILLLCONV, MAX_MILLLCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvGrad"]',
                    'MIL Grad Conv', REGTYPE_FLOAT, DEF_MILGRADCONV,
                    txtMILGradConv,
                    AcceptFloatBetween(MIN_MILGRADCONV, MAX_MILGRADCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxIterations"]',
                    'MIL Max Iter', REGTYPE_INT, DEF_MILMAXITER,
                    txtMILMaxIter,
                    AcceptIntBetween(MIN_MILMAXITER, MAX_MILMAXITER),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxRestarts"]',
                    'MIL Restarts', REGTYPE_INT, DEF_MILRESTARTS,
                    txtMILRestarts,
                    AcceptIntBetween(MIN_MILRESTARTS, MAX_MILRESTARTS),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxStep"]',
                    'MIL Max Step', REGTYPE_FLOAT, DEF_MILMAXSTEP,
                    txtMILMaxStep,
                    AcceptFloatBetween(MIN_MILMAXSTEP, MAX_MILMAXSTEP),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SearchLimit"]',
                    'MIL Search Limit', REGTYPE_FLOAT, DEF_MILSEARCHLIMIT,
                    txtMILSearchLimit,
                    AcceptFloatBetween(MIN_MILSEARCHLIMIT, MAX_MILSEARCHLIMIT),
                    '') );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Mode"]',
                          'MIL Run Mode', REGTYPE_INT, MILModeStringOf(DEF_MILMODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMILModeOpt,
                                       MILModeStringOf(MIL_MODE_OPT),
                                       Integer(MIL_MODE_OPT)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMILModeEval,
                                       MILModeStringOf(MIL_MODE_EVAL),
                                       Integer(MIL_MODE_EVAL)))
                       );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["use segments"]',
                          'MIL Seg Mode', REGTYPE_INT, MILSegModeStringOf(DEF_SEGMODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkSegModeEach,
                                       MILSegModeStringOf(MIL_MODE_SEG_EACH),
                                       Integer(MIL_MODE_SEG_EACH)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkSegModeAll,
                                       MILSegModeStringOf(MIL_MODE_SEG_ALL),
                                       Integer(MIL_MODE_SEG_ALL)))
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["join segments"]',
                        'MIL Join Segments', REGTYPE_BOOL, DEF_MILJOINSEGMENTS,
                        chkMILJoinSegments));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["GroupViterbi"]',
                        'Group Viterbi', REGTYPE_BOOL, DEF_GROUPVITERBI, 
                        chkGroupViterbi));

  // ["ShowIterations"] is done separately because it is duplicated out to
  // ["Histograms"]["PDFs Each Iteration"]
  // and because it's read from the registry as the negation of "MIL Quiet Output"

  RegKeyName            := 'Software\QUB\Task MIL';
  ProfileCat            := 'MIL';

  WikiTopic             := 'Modeling:MIL#Properties';

  {$include 'QUB_Task_Mil_Properties.htx'}
end;

procedure TQUB_Task_Mil_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  prDataList.ShowFiles;
  prDataList.ShowVar('DeadTime', 'ms', 0.05, False);

  if ViewportModel.CurrModel <> nil then begin
    prDataList.ShowModelVars(ViewportModel.CurrModel, False);
  end;

end;

procedure TQUB_Task_Mil_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;
  
  prDataList.HideAll;
end;

procedure TQUB_Task_Mil_PropertiesForm.chkMILShowIterClick(Sender: TObject);
begin
  inherited;
  Properties['ShowIterations'].Data[0] := Integer(TCheckBox(Sender).Checked);
  Properties['Histograms']['PDFs Each Iteration'].Data[0] := Integer(TCheckBox(Sender).Checked);
end;


//******************************************************************************


procedure TQUB_Task_Mil_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Reg   :TRegistry;
  Prop  :INode;
  bVal  :Boolean;
begin
  inherited;

  Reg   := nil; // will be created as needed by ReadXRegistry*
  try
    Prop  := Props['ShowIterations'];
    if Prop.DataCount = 0 then begin
      if ReadXRegistryBoolean(Reg, RegKeyName, 'MIL Quiet Output', bVal) then begin
        Prop.SetData(Integer(not bVal));
      end else begin
        Prop.SetData((DEF_MILSHOWITER + 1) mod 2);
      end;
    end;
    Prop  := Props['Histograms']['PDFs Each Iteration'];
    if Prop.DataCount = 0 then Prop.SetData(Props['ShowIterations'].Data[0]);
  finally
    if Reg <> nil then begin
      Reg.CloseKey;
      Reg.Free;
    end;
  end;
end;

procedure TQUB_Task_Mil_PropertiesForm.prUpdateComponents;
begin
  inherited;

  chkMILShowIter.Checked     := 0 <> Properties['ShowIterations'].Data[0];
end;


//******************************************************************************
end.

