unit QUB_Task_MIP;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


SysUtils, Classes, Math, SyncObjs,

Alloc, AllocTypes,
qubopt_iface,

QUB_Global,
QUB_Q_Global, QUB_Q_Model, QUB_MIP_Model, QUB_MIP_Global,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_Thread,
QUB_Task, QUB_Task_Data,
QUB_Idealization,
QUB_QFS_QIF_Segments,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix, QUB_QFS_Channels,
QUB_QVO_QVM,
//QUB_MIP_ExpCond_View,
QUB_ExpCond_Parameters,
QUB_Task_MIP_Properties,
QUB_IdlCond,
QUB_Task_IdlStim,
QUB_Task_IdlStim_Properties,
QUB_Node;
//KPL_Parser;


type


TQUB_Task_MIP  = class(TQUB_Task_Data)
protected


  //Source Objects
  prModel                  :TQUB_QVO_QVM;
  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  //Optimizer Parameters
  prOptParams              :TQUB_MIP_Opt_Params;

  //MIPModels
  prMIPModels              :TList;

  //ExpConds
  prExpCondParams          :TQUB_ExpCond_ParametersListList;
  prDelta                  :Integer;

  //Previous LL
  prPLL                    :Double;
  
  //Best Grads
  prfParsGrads             :TPADouble;

  
  //OptSegments
  prOptSegsChunks          :TList;
  prOptSegsOutput          :String;
  prOptSegsOldMdl          :TQUB_QFS_QMF;

  prQuietOutput            :Boolean;

  prGroupList              :String;
  prBatchCount             :Integer;
  prBatchIdx               :Integer;
  prBatchLabel             :String;

  prDataSetCount           :Integer;

  prQuietEntry             :Integer;

  prIdlStimSignal          :TEvent;
  prResults                :INode;


  //Execute
  procedure prDoTask; override;

  procedure prDoExecute; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;

  //Chunks
  procedure prSetChunkParameters; override;
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  function prGetIdlAndCond(QVD :TQUB_QVO_QVD; ParamTemplate :TQUB_ExpCond_Parameters; ParamClasses :TQUB_ExpCond_ParametersList): INode;
  procedure prEnsureIdlStim(QVD :TQUB_QVO_QVD; StimChannels :TList{of Integer}; SegCount :Integer; Segments :TQUB_PAFirstLast);
  procedure prStartIdlStim;
  procedure prIdlStimDone(Sender :TObject);

  procedure prSetModel(Model :TQUB_QVO_QVM);

  //Update visuals
  procedure prUpdateModel;
  procedure prResetModel;
  procedure prShowData;

  //Update results
  procedure prUpdateResults(Final :Boolean);
  procedure prPublishResults;

  function DataCallBack(MIPChunk :TQUB_MIP_PChunk) :Boolean;

  procedure prMIP;
  function prLLGrad(CompGrad :Integer) :Integer;

  procedure prPrintOptParams;
  procedure prPrintParams(Final :Boolean);
  procedure prPrintGrads;


public


  Properties                 :INode;


  constructor Create(Suspended :Boolean); override;


  function DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
  function DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model             :TQUB_QVO_QVM                 read prModel             write prSetModel;

  //Parameters
  property OptParams         :TQUB_MIP_Opt_Params          read prOptParams         write prOptParams;

  property QuietOutput       :Boolean                      read prQuietOutput       write prQuietOutput;


end;


implementation


uses


IEEE754,
QUB_Main,
QUB_GUIInterface, QUB_Viewport_Data,
QUB_Profiles,
QUB_Viewport_Results,
QUB_Node_Python, QUB_Python_Data;


//DFP Callbacks ****************************************************************
function DFPFunc(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
begin
  Result := TQUB_Task_MIP(call_obj).DFPFunc(Pars, Grad, LL, nf, ndf, CompGrad);
end;

function DFPCheck(call_obj :TObject; Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
begin
  Result := TQUB_Task_MIP(call_obj).DFPCheck(Iters, nf, ndf, LL, Pars, Grads);
end;



//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _ M I P                        //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_MIP.Create(Suspended :Boolean);
begin

  inherited;

  prName := 'MIP optimizer';

  Properties := ProfileMgr['MIP'].Active.Clone(True);
  prResults  := TNode.Create('Results');
  prResults['ResultType'].SetData('MIP');

end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MIP.prDoTask;
begin

  inherited;

  prMIP;

end;

procedure TQUB_Task_MIP.prMIP;
var
  Done             :Boolean;
  Cnt              :Integer;
  m, n             :Integer;
  BatchFirst       :Integer;
  BatchIter        :Integer;
  BatchCountCount  :Integer;
  BatchCountList   :TPAInteger;
begin


  prOptSegsChunks := TList.Create;

  BatchCountCount := 1;
  BatchCountList  := nil;


  if prOptParams.OptSegments = OPTSEGS_GROUPLIST then begin
    BatchCountCount := 100;
    BatchCountList  := IAlloc1(BatchCountCount);
    StrToArrayOfInt(prGroupList, BatchCountList, BatchCountCount);
    if BatchCountCount = 0 then begin
      prOptParams.OptSegments := OPTSEGS_GROUP;
      BatchCountCount := 1;
    end;
  end;


  try


    for BatchIter := 0 to BatchCountCount - 1 do begin


      if prStopped then Break;

      //Save them all
      for m := 0 to prChunks.Count - 1 do begin
        prOptSegsChunks.Add(prChunks[m]);
      end;

      //Initialize pointers in list
      BatchFirst := 0;
      if          prOptParams.OptSegments = OPTSEGS_TOGETHER{ALL} then begin
        prBatchCount := prChunks.Count;
      end else if prOptParams.OptSegments = OPTSEGS_GROUP then begin
        prBatchCount := prOptParams.OptSegmentsGroupCount;
      end else begin
        prBatchCount := Abs(BatchCountList[BatchIter]);
      end;


      Synchronize(prPrintOptParams);


      prBatchIdx := 0;


      while (prBatchIdx < prOptParams.OptSegmentsMaxBatch) and (BatchFirst < prOptSegsChunks.Count) and (not prStopped) do begin
      //Batches


        Inc(prBatchIdx);


        if prBatchCount < prOptSegsChunks.Count then begin
          prBatchLabel := Format('Batch %d', [prBatchIdx]);
        end;

        if (prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST) then begin
          Synchronize(prResetModel);
        end;

        if (BatchFirst + prBatchCount - 1) >= prOptSegsChunks.Count then begin
          prBatchCount := prOptSegsChunks.Count - BatchFirst;
        end;


        //Next Batch
        prChunks.Clear;


        for m := BatchFirst to BatchFirst + prBatchCount - 1 do begin
          prChunks.Add(prOptSegsChunks[m]);
        end;


        if prOptParams.OptMUXFiles = 1 then begin
          for m := BatchFirst + prBatchCount to prOptSegsChunks.Count - 1 do begin
            if TQUB_MIP_PChunk(prOptSegsChunks[BatchFirst]).QVD <> TQUB_MIP_PChunk(prOptSegsChunks[m]).QVD then begin
              for n := m + BatchFirst to m + BatchFirst + prBatchCount - 1 do begin
                if n >= prOptSegsChunks.Count then Break;
                prChunks.Add(prOptSegsChunks[n]);
              end;
              Break;
            end;
          end;
        end;


        Synchronize(prShowData);


        //**********************************************************************
        Cnt  := 0;

        repeat
        //restart if Grads > some threshold...


          if Cnt > 0 then begin
            prConsoleMsg  := 'Restarting...';
            prConsoleCode := MESSAGE_WARNING;
            if not prQuietOutput then Synchronize(prWriteNewConsole);
            prpLL := 0;
          end;

          Done := True;


          prfParsGrads := DAlloc1(TQUB_MIP_Model(prMIPModels[0]).fParCount);
          for m := 0 to TQUB_MIP_Model(prMIPModels[0]).fParCount - 1 do prfParsGrads[m] := 0;


          try


            for m := 0 to prMIPModels.Count - 1 do begin

              TQUB_MIP_Model(prMIPModels[m]).Do_fParsToModel;

            end;


            if prOptParams.OptMode = MODE_OPTIMIZE then begin
            //Optimize

              if prQuietOutput then begin
                prConsoleMsg  := Format('Batch index %d: starting ...', [prBatchIdx]);
                prConsoleCode := MESSAGE_WARNING;
                Synchronize(prWriteNewConsole);
                prQuietEntry := prConsoleEntry;
              end;

              try
                DFP_MIN(Self, TQUB_MIP_Model(prMIPModels[0]).fParCount, TQUB_MIP_Model(prMIPModels[0]).fPars, TQUB_MIP_Model(prMIPModels[0]).fParsVar,
                        QUB_Task_MIP.DFPFunc, QUB_Task_MIP.DFPCheck,
                        OptParams.OptMaxIter, OptParams.OptLLConv, OptParams.OptGradConv, OptParams.OptMaxStep,
                        TQUB_MIP_Model(prMIPModels[0]).Error);
              except
              end;

              TQUB_MIP_Model(prMIPModels[0]).Do_Varf2Varc;


            end else begin
            //Evaluate Log-Likelihood


              prLLGrad(1{CompGrad});

              TQUB_MIP_Model(prMIPModels[0]).Do_Varf2Varc;


              prPrintGrads;


            end;


            prUpdateResults(True{Final});
            prPrintParams(True{Final});


          finally


            //Check Grads:
            for m := 0 to TQUB_MIP_Model(prMIPModels[0]).fParCount - 1 do begin
              if Abs(prfParsGrads[m]) >= 1 then begin
                Done := False;
                Break;
              end;
            end;

            FreeMem(prfParsGrads);


          end;


          Inc(Cnt);


          TQUB_MIP_Model(prMIPModels[0]).Iterations := 0;

          if (OptParams.OptRestart = 0) and (Cnt > 0) then Break;

        until ((Done) or (Cnt >= 2) or (prStopped) or (prError));
        //************************************************************************


        prConsoleMsg  := prOptSegsOutput;
        prConsoleCode := MESSAGE_MESSAGE;
        if prQuietOutput then begin
          prConsoleEntry := prQuietEntry;
          Synchronize(prWriteSetConsole);
        end else begin
          Synchronize(prWriteNewConsole);
        end;


        //Next batch
        Inc(BatchFirst, prBatchCount);


      end;


      if ((prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST)) and (prBatchCount > 1) then Synchronize(prResetModel);

      prChunks.Clear;

      for m := 0 to prOptSegsChunks.Count - 1 do begin
        prChunks.Add(prOptSegsChunks[m]);
      end;

      prOptSegsChunks.Clear;


    end;//BatchIter

    Synchronize(prPublishResults);

  finally

    prOptSegsChunks.Free;
    prOptSegsOldMdl.Free;
    FreeMem(BatchCountList);

  end;

end;

procedure TQUB_Task_MIP.prPrintOptParams;
var
  t :String;
begin

  with prOptParams do begin
    t :=     #13#10'MIP parameters for:';
    t := t + #13#10#9'data  '    + Choice(prData  <> nil, ExtractFileName(prData.FileName),  '<none>');
    t := t + #13#10#9'model '    + Choice(prModel <> nil, ExtractFileName(prModel.FileName), '<none>');
    t := t + #13#10#9'Est (K), ';
    t := t +         'P0 ('      + Choice(Boolean(EstInitProbs=INITPROBS_FIXED), 'fixed', Choice(Boolean(EstInitProbs=INITPROBS_EQ), 'eq', 'condeq')) + '), ';
    t := t +         'SegCount ' + IntToStr(prBatchCount) + '.';
    t := t + #13#10;
  end;

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
end;

procedure TQUB_Task_MIP.prPrintParams(Final :Boolean);
var
  LogL :Double;
  t, tr :String;
  k, m :Integer;
begin

  t := '';
  tr := '';
  LogL := 0;
  for m := 0 to prMIPModels.Count - 1 do begin
    LogL := LogL + TQUB_MIP_Model(prMIPModels[m]).LL;
  end;

  prOptSegsOutput :=     '';

  if Final then begin
    try
      t :=            #13#10'================================================================';
      t := t + Format(#13#10'MIP final results:  LL = %-14.10g Iterations = %d, Dwells = %d, Points = %d, Multps. = %d',
                     [-LogL, TQUB_MIP_Model(prMIPModels[0]).Iterations, TQUB_MIP_Model(prMIPModels[0]).TotalDwellCount, TQUB_MIP_Model(prMIPModels[0]).TotalDataPoints, TQUB_MIP_Model(prMIPModels[0]).Multiplications]);

      tr := tr + Format('%14.10g'#9'%d'#9'%d'#9'%d'#9'%d'#9,
                     [-LogL, TQUB_MIP_Model(prMIPModels[0]).Iterations, TQUB_MIP_Model(prMIPModels[0]).TotalDwellCount, TQUB_MIP_Model(prMIPModels[0]).TotalDataPoints, TQUB_MIP_Model(prMIPModels[0]).Multiplications]);
    except
    end;
  end;

  t := t +   #13#10#9'Rates:  '#13#10#9's1  s2   k0         k0.std     k1         k1.std' +
                             #13#10#9'--------------------------------------------------';

  for k := 0 to TQUB_MIP_Model(prMIPModels[0]).RateCount - 1 do begin
    try
      t := t + Format(#13#10#9'%d   %d    %-10.6g %-10.6g %-10.6g %-10.6g',
                     [TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].x+1, TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].y+1,
                     //!!! cParsVar
                      Choice(OptParams.OptAllowNegRates,
                            TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0],
                        Exp(TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0])
                      ),
                                                                      Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 0]),
                      TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 1],
                      Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 1])
                   ]);
      prOptSegsOutput := prOptSegsOutput + Trim(Format('%10.6g', [Choice(OptParams.OptAllowNegRates,
                                                                        TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0],
                                                                    Exp(TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0])
                                                                  )
                                                                                                                        ])) + #9 +
                                           Trim(Format('%10.6g', [Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 0])])) + #9;

      tr := tr + Format('%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9,
                                              [Choice(OptParams.OptAllowNegRates,
                                                      TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0],
                                                  Exp(TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0])
                                               ),
                                               Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 0]),
                                               TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 1],
                                               Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 1])
                                              ]);
    except
    end;
  end;
  prOptSegsOutput := prOptSegsOutput + Trim(Format('%14.10g', [-LogL]));

  t := t + #13#10#9'--------------------------------------------------';

  //prBatchOutput := #13#10 + prBatchOutput;

  if Final then begin
    if OptParams.OptMode = MODE_OPTIMIZE then begin
      t := t + #13#10'Free pars grads: ';
      for k := 0 to TQUB_MIP_Model(prMIPModels[0]).fParCount - 1 do begin
        try
          t := t + Format('%-12.8g ', [-prfParsGrads[k]]);
          tr := tr + Format('%14.10g'#9, [-prfParsGrads[k]]);
        except
        end;
      end;
    end;
    t := t + #13#10 + '================================================================';
  end else begin
    t := t + #13#10;
  end;
  
  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);

  if Final then begin
    if prResultText <> nil then begin
      prResultText^ := tr;
    end;
  end;
end;

procedure TQUB_Task_MIP.prPrintGrads;
var
  LogL  :Double;
  dLogL :Double;
  t     :String;
  k, m  :Integer;
  ec    :Integer;
begin

  LogL := 0;
  for m := 0 to prMIPModels.Count - 1 do begin
    LogL := LogL + TQUB_MIP_Model(prMIPModels[m]).LL;
  end;

  prConsoleMsg  := Format('Batch %d-%d, LL: %-14.10g, Gr: ', [prBatchIdx, TQUB_MIP_Model(prMIPModels[0]).Iterations, -LogL]);

  t := Format('Iteration %d, LL = %-14.10g, Func calls = %d, Grad calls = %d',
              [TQUB_MIP_Model(prMIPModels[0]).Iterations, -LogL, TQUB_MIP_Model(prMIPModels[0]).FuncCalls, TQUB_MIP_Model(prMIPModels[0]).GradCalls]);

  t := t + #13#10#9'Free pars grads: ';
  for k := 0 to TQUB_MIP_Model(prMIPModels[0]).fParCount - 1 do begin
    dLogL := 0;
    for m := 0 to prMIPModels.Count - 1 do begin
      dLogL := dLogL + TQUB_MIP_Model(prMIPModels[m]).dLLdfPar[k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
      prConsoleMsg  := prConsoleMsg + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  //t must be assigned here
  if prQuietOutput then begin
    prConsoleCode := MESSAGE_WARNING;
    Synchronize(prWriteSetConsole);
  end;

  t := t + #13#10#9'Cstr pars grads: ';
  for k := 0 to TQUB_MIP_Model(prMIPModels[0]).cParCount - 1 do begin
    dLogL := 0;
    for m := 0 to prMIPModels.Count - 1 do begin
      dLogL := dLogL + TQUB_MIP_Model(prMIPModels[m]).dLLdcPar[k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  //This paragraph differs from Mac's ..........................................
  for ec := 0 to TQUB_MIP_Model(prMIPModels[0]).ExpCondCount - 1 do begin
    t := t + Format(#13#10#9'Rates grads (ExpCond[%d]):     ', [ec]);
    for k := 0 to TQUB_MIP_Model(prMIPModels[0]).RateCount - 1 do begin
      dLogL := 0;
      for m := 0 to prMIPModels.Count - 1 do begin
        dLogL := dLogL + TQUB_MIP_Model(prMIPModels[m]).dLLdq[ec][k];
      end;
      try
        t := t + Format('%-12.8g ', [-dLogL]);
      except
      end;
    end;
  end;
  //............................................................................

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);

end;


//******************************************************************************
procedure TQUB_Task_MIP.prUpdateResults(Final :Boolean);
var
  k, m :Integer;
  DS0   :INode;
  DS    :INode;
  Lbl   :String;
  Res   :INode;
  AAcc  :Array of Double;
  Acc   :Double;
  IAcc  :Integer;
  x     :Double;
  i     :Integer;
  N     :Integer;
  FinalMdl :INode;
  ResMdl   :INode;


  function BatchLbl(const Name :String): String;
  begin
    if prBatchLabel = ''
    then Result := Name
    else Result := prBatchLabel + ' ' + Name;
  end;
  function DResult(Parent :INode; const Lbl :String; Count :Integer): INode;
  begin
    Result := Parent[Lbl];
    if LongInt(Result.DataCount) <> Count then begin
      Result.SetupNumData(NODE_DATATYPE_DOUBLE, Count, 1);
    end;
  end;
  function IResult(Parent :INode; const Lbl :String; Count :Integer): INode;
  begin
    Result := Parent[Lbl];
    if LongInt(Result.DataCount) <> Count then begin
      Result.SetupNumData(NODE_DATATYPE_LONGINT, Count, 1);
    end;
  end;
  // array accumulator -- for summing several vectors prior to copying sum array into result node
  procedure AAccSetup(Count :Integer);
  var
    k   :Integer;
  begin
    SetLength(AAcc, Count);
    for k := 0 to Count-1 do begin
      AAcc[k] := 0.0;
    end;
  end;

  procedure UpdateRateStd(A, B :Integer; IsK1 :Boolean; Std :Double);
  var
    Rate   :INode;
    States :INode;
    DK     :INode;
    Dir    :Integer;
  begin
    Rate   := FinalMdl['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      States := Rate['States'];
      if (States.Data[0] = A) and (States.Data[1] = B) then begin
        Dir := 0;
      end else if (States.Data[0] = B) and (States.Data[1] = A) then begin
        Dir := 1;
      end else begin
        Dir := -1;
      end;

      if Dir >= 0 then begin
        if IsK1 then begin
          DK := Rate['dk1'];
        end else begin
          DK := Rate['dk0'];
        end;
        if DK.DataCount < 2 then DK.SetupNumData(NODE_DATATYPE_DOUBLE, 2, 1);
        DK.Data[(0+Dir) mod 2] := Std;

        Break;
      end;

      Rate := Rate.SiblingSameName;
    end;
  end;

begin
  if Final then begin
    FinalMdl := Model.ModelObject.CloneAsNode(True, TNode.CreateNull);
    FinalMdl.SetData(BatchLbl('Final'));
  end;

  DS0  := prResults.Find('DataSet');

  Lbl  := BatchLbl('LL');
  DS   := DS0;
  Acc  := 0.0;
  IAcc := 0;
  for m := 0 to prMIPModels.Count - 1 do begin
    x    := - TQUB_MIP_Model(prMIPModels[m]).LL;
    Acc  := Acc + x;
    i := TQUB_MIP_Model(prMIPModels[m]).TotalDwellCount;
    IAcc := IAcc + i;
    DResult(DS, Lbl, 1).Data[0]   := x;
    DResult(DS, BatchLbl('LL per event'), 1).Data[0] := x / i;
    DS   := DS.SiblingSameName;
  end;
  DResult(prResults, Lbl, 1).Data[0]   := Acc;
  DResult(prResults, BatchLbl('LL per event'), 1).Data[0] := Acc / IAcc;
  IResult(prResults, 'DwellCount', 1).Data[0] := IAcc;
  //TO DO: Change if models are additive (e.g., as in Na and K currents analyzed together). LM

  IResult(prResults, 'Iterations', 1).Data[0] := TQUB_MIP_Model(prMIPModels[0]).Iterations;

  for k := 0 to TQUB_MIP_Model(prMIPModels[0]).RateCount - 1 do begin
    try
      Lbl  := Format('%d->%d',[TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].x+1, TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].y+1]);
      DResult(prResults, BatchLbl(Lbl), 1).Data[0] := Choice(OptParams.OptAllowNegRates,
                                                       TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0],
                                                       Exp(TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 0])
                                                     );
      DResult(prResults, BatchLbl(Lbl+' std'), 1).Data[0] := Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 0]);
      DResult(prResults, BatchLbl(Lbl+' k1'), 1).Data[0]  := TQUB_MIP_Model(prMIPModels[0]).cPars[2*k + 1];
      DResult(prResults, BatchLbl(Lbl+' k1 std'), 1).Data[0] := Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 1]);
      if Final then begin
        UpdateRateStd(TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].x, TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].y, False, Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 0]));
        UpdateRateStd(TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].x, TQUB_MIP_Model(prMIPModels[0]).RateIdx[k].y, True, Sqrt(TQUB_MIP_Model(prMIPModels[0]).cParsVar[2*k + 1]));
      end;
    except
    end;
  end;

  Lbl   := BatchLbl('Free pars grads');
  N     := TQUB_MIP_Model(prMIPModels[0]).fParCount;
  AAccSetup(N);
  DS    := DS0;
  for m := 0 to prMIPModels.Count - 1 do begin
    Res := DResult(DS, Lbl, N);
    for k := 0 to N - 1 do begin
      x           := - TQUB_MIP_Model(prMIPModels[m]).dLLdfPar[k];
      Res.Data[k] := x;
      AAcc[k]     := AAcc[k] + x;
    end;
    DS  := DS.SiblingSameName;
  end;
  Res   := DResult(prResults, Lbl, N);
  for k := 0 to N - 1 do begin
    Res.Data[k]   := AAcc[k];
  end;

  Lbl   := BatchLbl('Cstr pars grads');
  N     := TQUB_MIP_Model(prMIPModels[0]).cParCount;
  AAccSetup(N);
  DS    := DS0;
  for m := 0 to prMIPModels.Count - 1 do begin
    Res := DResult(DS, Lbl, N);
    for k := 0 to N - 1 do begin
      x           := - TQUB_MIP_Model(prMIPModels[m]).dLLdcPar[k];
      Res.Data[k] := x;
      AAcc[k]     := AAcc[k] + x;
    end;
    DS  := DS.SiblingSameName;
  end;
  Res   := DResult(prResults, Lbl, N);
  for k := 0 to N - 1 do begin
    Res.Data[k]   := AAcc[k];
  end;

  Lbl   := BatchLbl('Rates grads');
  N     := TQUB_MIP_Model(prMIPModels[0]).RateCount;
  AAccSetup(N);
  DS    := DS0;
  for m := 0 to prMIPModels.Count - 1 do begin
    Res := DResult(DS, Lbl, N);
    for k := 0 to N - 1 do begin
      x           := - TQUB_MIP_Model(prMIPModels[m]).dLLdq[0,k];
      Res.Data[k] := x;
      AAcc[k]     := AAcc[k] + x;
    end;
    DS  := DS.SiblingSameName;
  end;
  Res   := DResult(prResults, Lbl, N);
  for k := 0 to N - 1 do begin
    Res.Data[k]   := AAcc[k];
  end;

  if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin
    Lbl   := BatchLbl('Cond_Rates grads');
    N     := TQUB_MIP_Model(prMIPModels[0]).RateCount;
    AAccSetup(N);
    DS    := DS0;
    for m := 0 to prMIPModels.Count - 1 do begin
      Res := DResult(DS, Lbl, N);
      for k := 0 to N - 1 do begin
        x           := - TQUB_MIP_Model(prMIPModels[m]).dLLdCond_q[k];
        Res.Data[k] := x;
        AAcc[k]     := AAcc[k] + x;
      end;
      DS  := DS.SiblingSameName;
    end;
    Res   := DResult(prResults, Lbl, N);
    for k := 0 to N - 1 do begin
      Res.Data[k]   := AAcc[k];
    end;
  end;

  if Final then begin
    ResMdl := prResults.Find('ModelFile');
    // remove any "Final" model and replace with this final model
    while not ResMdl.IsNull do begin
      if ResMdl.Name = BatchLbl('Final') then begin
        prResults.RemoveChild(ResMdl);
        Break;
      end;
      ResMdl := ResMdl.SiblingSameName;
    end;
    prResults.AppendChild(FinalMdl);
  end;
end;

procedure TQUB_Task_MIP.prPublishResults;
begin
  ViewportResults.AddResults( prResults, False );
end;


//******************************************************************************
function TQUB_Task_MIP.DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
var
  i, m :Integer;
  Bull :Boolean;
begin

  Result := 0;

  if TQUB_MIP_Model(prMIPModels[0]).Iterations > 0 then begin

    for m := 0 to prMIPModels.Count - 1 do begin

      //Copy the new pars into the models
      for i := 0 to TQUB_MIP_Model(prMIPModels[m]).fParCount - 1 do begin
        TQUB_MIP_Model(prMIPModels[m]).fPars[i] := Pars[i];
      end;

      TQUB_MIP_Model(prMIPModels[m]).Do_fParsToModel;

    end;

  end;


  try
    prLLGrad(CompGrad);
  except
  end;


  LL := 0;
  for m := 0 to prMIPModels.Count - 1 do begin
    LL := LL + TQUB_MIP_Model(prMIPModels[m]).LL;
  end;

  Bull := IsNan(LL) or IsInfinity(LL);

  if CompGrad <> 0 then begin
    for i := 0 to TQUB_MIP_Model(prMIPModels[0]).fParCount - 1 do begin
      Grad[i] := 0;
      for m := 0 to prMIPModels.Count - 1 do begin
        Grad[i] := Grad[i] + TQUB_MIP_Model(prMIPModels[m]).dLLdfPar[i];
      end;
      Bull := Bull or IsNan(Grad[i]) or IsInfinity(Grad[i]);
    end;
    Move(Grad[0], prfParsGrads[0], TQUB_MIP_Model(prMIPModels[0]).fParCount * SizeOf(Double));
  end;

  if Bull then LL := -1e10;

  Inc(TQUB_MIP_Model(prMIPModels[0]).Iterations);
  Inc(nf);
  if CompGrad <> 0 then Inc(ndf);
  TQUB_MIP_Model(prMIPModels[0]).FuncCalls := nf;
  TQUB_MIP_Model(prMIPModels[0]).GradCalls := ndf;

end;

function TQUB_Task_MIP.DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
var
  LLConverged :Boolean;
  GRConverged :Boolean;
  k, m        :Integer;
begin

  Result := 0;

  TQUB_MIP_Model(prMIPModels[0]).Iterations := Iters;

  if Iters >= OptParams.OptMaxIter then Result := 1;

  LLConverged := Abs(LL - prpLL) < OptParams.OptLLConv;
  prpLL := LL;

  GRConverged := True;
  for k := 0 to TQUB_MIP_Model(prMIPModels[0]).fParCount - 1 do begin
    if Abs(Grads[k]) > OptParams.OptGradConv then GRConverged := False;
  end;


  Synchronize(prUpdateModel);


  prUpdateResults(False{Final});

  prPrintGrads;
  prPrintParams(False{Final});


  if (LLConverged) and{or} (GRConverged) then Result := 1;
  if prStopped then Result := 1;

  //if Result = 1 then begin
    //Change the parameters to the last good ones (???)
    for m := 0 to prMIPModels.Count - 1 do begin
      //Copy the new pars into the MIP models
      for k := 0 to TQUB_MIP_Model(prMIPModels[m]).fParCount - 1 do begin
        TQUB_MIP_Model(prMIPModels[m]).fPars[k] := Pars[k];
      end;
      TQUB_MIP_Model(prMIPModels[m]).Do_fParsToModel;
    end;
    //Only the fParsGrads
    Move(Grads[0], prfParsGrads[0], TQUB_MIP_Model(prMIPModels[0]).fParCount * SizeOf(Double));
  //end;

end;



//****************************************************************************//
//                                                                            //
//                             L L   +   G R A D                              //
//                                                                            //
//****************************************************************************//
function TQUB_Task_MIP.prLLGrad(CompGrad :Integer) :Integer;
const
  VERY_SMALL     = 1e-500;
var
  MIPChunk       :TQUB_MIP_PChunk;
  MIPModel       :TQUB_MIP_Model;
  Data           :TQUB_MIP_Data;
  DataCount      :Integer;

  AlphaS         :TPMExtended;
  BetaS          :TPMExtended;
  LScale         :TPAExtended;

  cLL, cGR       :Extended;
  c, k, p, s, m  :Integer;
  ec             :Integer;
  t, dt          :Double;
  SomeConfluent  :Boolean;

  PrevExpCond    :Integer;
  ECDwell        :Integer;
  ECClass        :Integer;
  ECNextChange   :Integer;

  T0             :Integer;
  NextT0         :Integer;
  Idl            :TQUB_FirstLast;

  SmallRateWarning :Boolean;
  BigRateWarning :Boolean;

procedure ChunkData(c :Integer);
begin
  MIPChunk := prChunks[c];
  //if MIPChunk.Data = nil then begin
  if MIPChunk.AlphaS = nil then begin
    DataCallback(MIPChunk);
  end;
  Data      := MIPChunk.Data;
  DataCount := MIPChunk.DataCount;
  AlphaS    := MIPChunk.AlphaS;
  BetaS     := MIPChunk.BetaS;
  LScale    := MIPChunk.LScale;
  dt        := MIPChunk.Sampling;
end;

begin


  Result         := 0;


  SomeConfluent  := False;
  SmallRateWarning := False;
  BigRateWarning := False;

  //Reset LL and derivatives w.r.t. q's ****************************************
  for m := 0 to prMIPModels.Count - 1 do begin
    //LL
    TQUB_MIP_Model(prMIPModels[m]).LL               := 0;

    //dLLdq[ec]
    for ec := 0 to TQUB_MIP_Model(prMIPModels[m]).ExpCondCount - 1 do begin
      if TQUB_MIP_Model(prMIPModels[m]).Confluent[ec] <> 0 then SomeConfluent := True;
      for k := 0 to TQUB_MIP_Model(prMIPModels[m]).RateCount - 1 do begin
        TQUB_MIP_Model(prMIPModels[m]).dLLdq[ec, k] := 0;
      end;//k
    end;//ec
    //dLLdCond_q
    for k := 0 to TQUB_MIP_Model(prMIPModels[0]).RateCount - 1 do begin
      TQUB_MIP_Model(prMIPModels[m]).dLLdCond_q[k]  := 0;
    end;

    if TQUB_MIP_Model(prMIPModels[m]).SmallRateWarning then begin
      SmallRateWarning := True;
    end;
    if TQUB_MIP_Model(prMIPModels[m]).BigRateWarning then begin
      BigRateWarning := True;
    end;


  end;
  //****************************************************************************




  //The NEW algorithm (bogus option)
  if OptParams.OptMUXFiles = 0 then begin
  //########################################################################




  //For each chunk *************************************************************
  for c := 0 to prChunks.Count - 1 do begin


    ChunkData(c);


    try


      //Reset the experimental conditions class
      PrevExpCond := -1;

      //Choose the right MIPModel
      MIPModel := TQUB_MIP_Model(MIPChunk.MIPModel);

      //Initialize experimental condition
      ECDwell      := 0;
      ECClass      := MIPChunk.ExpCondIdl[ECDwell].Cls;

      //Reset the start time
      T0           := 0;
      NextT0       := T0 + MIPChunk.ExpCondIdl[ECDwell].N;



      //Check the logic below:


      //Idl selection
      Idl.First    := 0;
      Idl.Last     := Idl.First;
      if MIPChunk.ExpCondIdlCount = 1 then begin
        Idl.Last   := DataCount - 1;
      end else begin
        ec         := Data[Idl.First].ExpCondCls;
        for p := Idl.First + 1 to DataCount - 1 do begin
          if Data[p].ExpCondCls = ec then begin
            Idl.Last := p;
          end else begin
            Break;
          end;
        end;
      end;


      //Loop this for each stimulus segment
      for ECDwell := 0 to MIPChunk.ExpCondIdlCount - 1 do begin


        if ECDwell = 0 then begin
        //The very first stimulus segment

          //State probabilities
          MIPModel.Do_P0;
          //Derivatives of the initial state probabilities
          MIPModel.Do_dP0_dq;           //Check for Eq and CondEq !!!
          MIPModel.Do_dP0_dCond_q;      //        - ,, -

        end else begin
        //Not the first stimulus segment

          try
            {
            Assumption: the stimuli change at this very point
            Recalculate P0 and dP0/dq[ec] as (s==this seg):
              P0[s] = P0[s-1] * A[ec(s-1),dur(s-1)]
              dP0[s]/dq[ec] = (dP0[s-1]/dq[ec]) * A[ec(s-1),dur(s-1)] +
                              P0[s-1] * (dA[ec(s-1),dur(s-1)]/dq[ec])

              dA[ec(s-1),dur(s-1)]/dq[ec] <> 0 if ec(s-1)  = ec
                                           = 0 if ec(s-1) <> ec
            }

            //A matrix
            MIPModel.Do_At((NextT0-T0)*dt, ECClass{prev.seg});

            //dA[ECClass] / dq[ECClass] (the others are zero)
            MIPModel.Do_dAt_dq((NextT0-T0)*dt, ECClass);

            //Current state, later to be copied into initial state
            MIPModel.Do_Pt((NextT0-T0)*dt, ECClass);
            MIPModel.Do_dPt_dq(ECClass);
            MIPModel.Do_dPt_dCond_q(ECClass);

            //Copy Pt into P0, and the derivatives, to get:
            MIPModel.PtToP0;                        //==>P0
            MIPModel.dPtdqTodP0dq;                  //==>dP0_dq
            MIPModel.dPtdCond_qTodP0dCond_q;        //==>dP0_dCond_q


          finally

            ECClass      := MIPChunk.ExpCondIdl[ECDwell].Cls;
            ECNextChange := ECNextChange + MIPChunk.ExpCondIdl[ECDwell].N;

            T0           := NextT0;
            NextT0       := T0 + MIPChunk.ExpCondIdl[ECDwell].N;

            //Idl selection
            Idl.First    := Idl.Last + 1;
            Idl.Last     := Idl.First;
            ec         := Data[Idl.First].ExpCondCls;
            for p := Idl.First + 1 to DataCount - 1 do begin
              if Data[p].ExpCondCls = ec then begin
                Idl.Last := p;
              end else begin
                Break;
              end;
            end;

          end;


        end;//ECDwell = 0


        //Need to find the first and the last Idl segment


        //Calculate A matrix and its derivatives w.r.t. each qq, and the products and their derivatives
        MIPModel.Do_A(dt);//In fact, only the A for ECClass is needed
        MIPModel.Do_dA_dq(dt,ECClass);


        //Initialize AlphaS, LScale ................................................
        p := Idl.First;

        MulVecMatVecE(MIPModel.P0, MIPModel.B[Data[p].Cls], AlphaS[p], MIPModel.StateCount);
        LScale[p] := 0;
        for s := 0 to MIPModel.StateCount - 1 do begin
          LScale[p] := LScale[p] + AlphaS[p, s];
        end;
        if LScale[p] < VERY_SMALL then begin
          LScale[p] := VERY_SMALL;
        end;
        for s := 0 to MIPModel.StateCount - 1 do begin
          AlphaS[p, s] := AlphaS[p, s] / LScale[p];
        end;


        //For each DataPoint calculate AlphaS, LScale ..............................
        for p := Idl.First + 1 to Idl.Last do begin

          MulVecEMatVecE(AlphaS[p - 1], TPMPointer(MIPModel.AB[ECClass])[Data[p].Cls, Data[p].Pow], AlphaS[p], MIPModel.StateCount);

          LScale[p]   := 0;
          for s := 0 to MIPModel.StateCount - 1 do begin
            LScale[p] := LScale[p] + AlphaS[p, s];
          end;
          if LScale[p] < VERY_SMALL then begin
            LScale[p] := VERY_SMALL;
          end;
          for s := 0 to MIPModel.StateCount - 1 do begin
            AlphaS[p, s] := AlphaS[p, s] / LScale[p];
          end;

        end;


        //Initialize BetaS .........................................................
        p := Idl.Last;
        for s := 0 to MIPModel.StateCount - 1 do begin
          BetaS[p, s] := 1 / LScale[p];
        end;


        //For each DataPoint calculate BetaS .......................................
        for p := Idl.Last - 1 downto Idl.First do begin

          MulMatVecEVecE(TPMPointer(MIPModel.AB[ECClass])[Data[p + 1].Cls, Data[p + 1].Pow], BetaS[p + 1], BetaS[p], MIPModel.StateCount);

          for s := 0 to MIPModel.StateCount - 1 do begin
            BetaS[p, s] := BetaS[p, s] / LScale[p];
          end;

        end;


        //Log-Likelihood ...........................................................


        //For each DataPoint
        for p := Idl.First to Idl.Last do begin
          cLL := Ln(LScale[p]);
          MIPModel.LL := MIPModel.LL - cLL;//Use minus sign in order to minimize LL
        end;
        //..........................................................................



        //Rates Gradients ..........................................................

        {
          dLL/dCond_q:
            applies only to P0:

              dLL/dCond_q[k]  = dP0/dCond_q[k].B[0].BetaS[0] + P0.B0.dBetaS[0]/dCond_q[k]
                                                                    [       =0        ]

          dLL/dq[ec]:
            applies to P0 for each ec
            applies to ATBT only for ec = ECClass

              for each ec:

                dLL/dq[ec][k] = dP0/dq[ec][k].B[0].BetaS[0]  + P0.B0.dBetaS[0]/dq[ec][k]
                                [<>0 for any ec]                     [<>0 if ec=ECClass]


        }

        //dLL/dCond_q
        for k := 0 to MIPModel.RateCount - 1 do begin

          //dLL/dCond_q[k] = dP0d/Cond_q[k].B[Idl.First].BetaS[Idl.First]

          MulVecMatVecE(MIPModel.dP0dCond_q[k], MIPModel.B[Data[Idl.First].Cls], MIPModel.Vec1E, MIPModel.StateCount);

          MulVecEVecESE(MIPModel.Vec1E, BetaS[Idl.First], cGR, MIPModel.StateCount);

          MIPModel.dLLdCond_q[k] := MIPModel.dLLdCond_q[k] - cGR;//Use minus sign in order to minimize LL

        end;


        //dLL/dq[ec]

        //  dP0/dq[ec][k].B[0].BetaS[0]
        for ec := 0 to MIPModel.ExpCondCount - 1 do begin

          for k := 0 to MIPModel.RateCount - 1 do begin

            MulVecMatVecE(TPAPointer(MIPModel.dP0dq[ec])[k], MIPModel.B[Data[Idl.First].Cls], MIPModel.Vec1E, MIPModel.StateCount);

            MulVecEVecESE(MIPModel.Vec1E, BetaS[Idl.First], cGR, MIPModel.StateCount);

            MIPModel.dLLdq[ec, k] := MIPModel.dLLdq[ec, k] - cGR;//Use minus sign in order to minimize LL

          end;//k

        end;//ec

        //  P0.B0.dBetaS[0]/dq[ec][k]
        //For each DataPoint
        for p := Idl.First to Idl.Last - 1 do begin

          for k := 0 to MIPModel.RateCount - 1 do begin

            //cGR = cGR + Sum t [AlphaS[t].dAtdq[t+1].Cls[t+1].BetaS[t+1]]

            MulVecEMatVecE(AlphaS[p], TPMPointer(TPAPointer(MIPModel.dABdq[ECClass])[k])[Data[p + 1].Cls, Data[p + 1].Pow], MIPModel.Vec1E, MIPModel.StateCount);

            MulVecEVecESE(MIPModel.Vec1E, BetaS[p + 1], cGR, MIPModel.StateCount);

            MIPModel.dLLdq[ECClass][k] := MIPModel.dLLdq[ECClass][k] - cGR;//Use minus sign in order to minimize LL

          end;//of k

        end;//of DataPoint
        //..........................................................................


      end;//ECDwell


    finally

    end;


  end;//of Chunk
  //****************************************************************************




  //the OLD algorithm
  end else begin
  //########################################################################




  //For each chunk *************************************************************
  for c := 0 to prChunks.Count - 1 do begin


    ChunkData(c);


    //Choose the right MIPModel
    MIPModel := TQUB_MIP_Model(MIPChunk.MIPModel);


    //State probabilities
    MIPModel.Do_P0;
    //Derivatives of the initial state probabilities
    MIPModel.Do_dP0_dq;
    MIPModel.Do_dP0_dCond_q;


    //Calculate A matrix and its derivatives w.r.t. each qq, and the products and their derivatives
    MIPModel.Do_A(dt);
    MIPModel.Do_dA_dq(dt);


    //Initialize AlphaS, LScale ................................................
    p := 0;

    MulVecMatVecE(MIPModel.P0, MIPModel.B[Data[p].Cls], AlphaS[p], MIPModel.StateCount);
    LScale[p] := 0;
    for s := 0 to MIPModel.StateCount - 1 do begin
      LScale[p]    := LScale[p] + AlphaS[p, s];
    end;
    if LScale[p] < VERY_SMALL then begin
      LScale[p] := VERY_SMALL;
    end;
    for s := 0 to MIPModel.StateCount - 1 do begin
      AlphaS[p, s] := AlphaS[p, s] / LScale[p];
    end;


    //For each DataPoint calculate AlphaS, LScale ..............................
    for p := 1 to DataCount - 1 do begin

      ec := Data[p-1{p}].ExpCondCls;

      MulVecEMatVecE(AlphaS[p - 1], TPMPointer(MIPModel.AB[ec])[Data[p].Cls, Data[p].Pow], AlphaS[p], MIPModel.StateCount);

      LScale[p]   := 0;
      for s := 0 to MIPModel.StateCount - 1 do begin
        LScale[p] := LScale[p] + AlphaS[p, s];
      end;
      if LScale[p] < VERY_SMALL then begin
        LScale[p] := VERY_SMALL;
      end;
      for s := 0 to MIPModel.StateCount - 1 do begin
        AlphaS[p, s] := AlphaS[p, s] / LScale[p];
      end;

    end;


    //Initialize BetaS .........................................................
    p := DataCount - 1;
    for s := 0 to MIPModel.StateCount - 1 do begin
      BetaS[p, s] := 1 / LScale[p];
    end;


    //For each DataPoint calculate BetaS .......................................
    for p := DataCount - 2 downto 0 do begin

      ec := Data[p{+1}].ExpCondCls;

      MulMatVecEVecE(TPMPointer(MIPModel.AB[ec])[Data[p + 1].Cls, Data[p + 1].Pow], BetaS[p + 1], BetaS[p], MIPModel.StateCount);

      for s := 0 to MIPModel.StateCount - 1 do begin
        BetaS[p, s] := BetaS[p, s] / LScale[p];
      end;

    end;



    //Log-Likelihood ...........................................................

    //For each DataPoint
    for p := 0 to DataCount - 1 do begin
      cLL := Ln(LScale[p]);
      MIPModel.LL := MIPModel.LL - cLL;//Use minus sign in order to minimize LL
    end;
    //..........................................................................



    //Rates Gradients ..........................................................


    //dLL/dCond_q
    for k := 0 to MIPModel.RateCount - 1 do begin

      //dLL/dCond_q[k] = dP0d/Cond_q[k].B[Idl.First].BetaS[Idl.First]

      MulVecMatVecE(MIPModel.dP0dCond_q[k], MIPModel.B[Data[0].Cls], MIPModel.Vec1E, MIPModel.StateCount);

      MulVecEVecESE(MIPModel.Vec1E, BetaS[0], cGR, MIPModel.StateCount);

      MIPModel.dLLdCond_q[k] := MIPModel.dLLdCond_q[k] - cGR;//Use minus sign in order to minimize LL

    end;

    //  dP0/dq[ec][k].B[0].BetaS[0]
    for ec := 0 to MIPModel.ExpCondCount - 1 do begin

      for k := 0 to MIPModel.RateCount - 1 do begin

        MulVecMatVecE(TPAPointer(MIPModel.dP0dq[ec])[k], MIPModel.B[Data[0].Cls], MIPModel.Vec1E, MIPModel.StateCount);

        MulVecEVecESE(MIPModel.Vec1E, BetaS[0], cGR, MIPModel.StateCount);

        MIPModel.dLLdq[ec, k] := MIPModel.dLLdq[ec, k] - cGR;//Use minus sign in order to minimize LL

      end;//k

    end;//ec

    //  P0.B0.dBetaS[0]/dq[ec][k]
    //For each DataPoint
    for p := 0 to DataCount - 2 do begin

      ec := Data[p].ExpCondCls;

      for k := 0 to MIPModel.RateCount - 1 do begin

        //cGR = cGR + Sum t [AlphaS[t].dAtdq[t+1].Cls[t+1].BetaS[t+1]]

        MulVecEMatVecE(AlphaS[p], TPMPointer(TPAPointer(MIPModel.dABdq[ec])[k])[Data[p + 1].Cls, Data[p + 1].Pow], MIPModel.Vec1E, MIPModel.StateCount);

        MulVecEVecESE(MIPModel.Vec1E, BetaS[p + 1], cGR, MIPModel.StateCount);

        MIPModel.dLLdq[ec][k] := MIPModel.dLLdq[ec][k] - cGR;//Use minus sign in order to minimize LL

      end;//of k

    end;//of DataPoint
    //..........................................................................


  end;//chunk



  end;
  //########################################################################





  //Calculate free pars gradients
  for m := 0 to prMIPModels.Count - 1 do begin
    TQUB_MIP_Model(prMIPModels[m]).Do_dLLdfPar;
  end;


  if SomeConfluent then begin
    prConsoleMsg  := 'Confluent eigen values detected.';
    prConsoleCode := MESSAGE_ERROR;
    Synchronize(prWriteNewConsole);
  end;
  if SmallRateWarning then begin
    prConsoleMsg  := 'Very small rate(s) detected.';
    prConsoleCode := MESSAGE_ERROR;
    Synchronize(prWriteNewConsole);
  end;
  if BigRateWarning then begin
    prConsoleMsg  := 'Very big rate(s) detected.';
    prConsoleCode := MESSAGE_ERROR;
    Synchronize(prWriteNewConsole);
  end;


end;


//******************************************************************************
procedure TQUB_Task_MIP.prUpdateModel;
begin

  MakeRatesFromParams(Model.ModelObject, TQUB_MIP_Model(prMIPModels[0]).cPars, 2*TQUB_MIP_Model(prMIPModels[0]).RateCount, OptParams.OptAllowNegRates);

end;

procedure TQUB_Task_MIP.prResetModel;
var
  m :Integer;
begin
  for m := 0 to prOptSegsOldMdl.Rates.RateCount - 1 do begin
    prModel.ModelObject.Rates[m].k0[1] := prOptSegsOldMdl.Rates[m].k0[1];
    prModel.ModelObject.Rates[m].k0[2] := prOptSegsOldMdl.Rates[m].k0[2];
    prModel.ModelObject.Rates[m].k1[1] := prOptSegsOldMdl.Rates[m].k1[1];
    prModel.ModelObject.Rates[m].k1[2] := prOptSegsOldMdl.Rates[m].k1[2];
  end;

  for m := 0 to prMIPModels.Count - 1 do begin
    TQUB_MIP_Model(prMIPModels[m]).Do_InitcPars;
    TQUB_MIP_Model(prMIPModels[m]).Do_InitfPars;
  end;
end;

procedure TQUB_Task_MIP.prShowData;
begin
  if Data <> nil then begin
    TQUB_QVO_QVD(Data).MoveSelectionIntoWindow(QUB_FirstLast(TQUB_Thread_PChunk(prChunks[0]).Sel.First, TQUB_Thread_PChunk(prChunks[prChunks.Count - 1]).Sel.Last), TQUB_Thread_PChunk(prChunks[0]).Sel.First);
  end;
end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MIP.prDoExecute;
begin
  prConsoleMsg  := 'MIP optimizer running ... ';
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
  inherited;
end;

procedure TQUB_Task_MIP.prDoExecuteStart;
begin

  inherited;

  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  //MIPModels
  prMIPModels     := TList.Create;

  //ExpConds
  prExpCondParams := TQUB_ExpCond_ParametersListList.Create;

  if Model <> nil then begin
    Properties.AppendChild(Model.ModelObject.CloneAsNode(True, TNode.CreateNull)).SetData('Initial');
  end;

end;


procedure TQUB_Task_MIP.prDoExecuteEnd;
var
  i :Integer;
begin
  if Model <> nil then Model.UnlockAny;

  if prSelectionList <> nil then prSelectionList.UnLock;

  //Free Params & MIPModels
  if prMIPModels <> nil then begin
    for i := 0 to prMIPModels.Count - 1 do begin
      TQUB_MIP_Model(prMIPModels[i]).Free;
    end;
    prMIPModels.Free;
    prMIPModels := nil;
  end;

  prExpCondParams.Free;
  prExpCondParams := nil;

  inherited;
end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
//This is the oldest version in the file
//TO DO: LM. Chris says give it a try. Last I used this, and commented out the next
(*
function TQUB_Task_MIP.prChunkData :Boolean;
var
  DataSets         :INode;
  DS, Seg          :INode;

  Sel              :TQUB_PFirstLast;
  Dat              :TQUB_FirstLast;

  QVD              :TQUB_QVO_QVD;

  ParamClasses     :TQUB_ExpCond_ParametersList;
  Params           :TQUB_ExpCond_Parameters;
  ParamsAreVar     :TQUB_ExpCond_Parameters;  // contains defaults from file list, and isVar: whether val is different across files

  ExpCondIdlGen    :TQUB_ExpCondIdlGen;

  Chunk            :TQUB_Thread_PChunk;
  MIPChunk         :TQUB_MIP_PChunk;
  MIPModel         :TQUB_MIP_Model;
  ChunkBounds      :TList;
  FirstChunk       :Integer;

  FileIndex        :Integer;
  
  c, l, p          :Integer;
begin

  Result           := False;

  prSetChunkParameters;
                      
  ParamsAreVar     := nil;


  try


    prDataSetCount := 0;

    DataSets       := DataNoder.GetChosenIdealization( TNode.CreateNull );
    DS             := DataSets.Find('DataSet');
    FileIndex      := 0;


    while not DS.IsNull do begin

      Inc(prDataSetCount);

      FirstChunk   := prChunks.Count;
      Seg          := DS.Find('Segment');

      if not Seg.IsNull then begin

        //ExpConds
        Params := TQUB_ExpCond_Parameters.Create;
        Params.ReadTree( DS.Find('ExpCond') );
        for p := 0 to Params.Count - 1 do begin
          Params.AddPar('Cond_' + Params[p].Name, Params[p].Cond_Value,
                        Params[p].Index, Params[p].IsVar, -1);
        end;

        //Chris, the Par.IsVar function is no longer calculated ! You eliminated it with the new chunking
        //It determines whether a Par changes across files when Par.Channel = -1 (else presumably does change)
        //Task_MIP.IsPZero is no longer reliable

        QVD         := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[ DS['FileName'].DataAsString ]);
        ChunkBounds := TList.Create;


        while not Seg.IsNull do begin

          Dat                      := QUB_FirstLast(Seg.Data[0], Seg.Data[1]);
          Chunk                    := prNewChunk(Dat);

          MIPChunk                 := TQUB_MIP_PChunk(Chunk);

          prChunks.Add(MIPChunk);

          Chunk.StartTime          := Double(Seg['start'].Data[0]) * 1.0e-3;

          MIPChunk.IdlSeg          := Seg.GetNodeMem;
          Node_INCREF( MIPChunk.IdlSeg );

          MIPChunk.SkipPoints      := QUB_FirstLast(0, 0); // ??? Skp;

          MIPChunk.IdlSel          := QUB_FirstLast(0, Seg['DwellCount'].Data[0]);  // ??? Idl^; // is this used? the upper bound seems to be one too large

          MIPChunk.QVD             := QVD;

          MIPChunk.Sampling        := DS['sampling'].Data[0] * 1.0e-3;
          MIPChunk.Scaling         := QVD.DataObject.Scaling;

          MIPChunk.Data            := nil;

          New(Sel);
          Sel.First                := Seg.Data[0]; // Dat.First + Skp.First;
          Sel.Last                 := Seg.Data[1]; // Dat.Last - Skp.Last;
          ChunkBounds.Add(Sel);

          MIPChunk.MIPModel        := nil; /// MIPModel;

          MIPChunk.FileIndex       := FileIndex;

          Seg                      := Seg.SiblingSameName;

        end;//while not Seg.IsNull

        ExpCondIdlGen := TQUB_ExpCondIdlGen.Create(Params, QVD, ChunkBounds, prOptParams.OptStimDelta);
        ExpCondIdlGen.FreeOutputs := False;

        ParamClasses := TQUB_ExpCond_ParametersList.Create;
        prExpCondParams.Add(ParamClasses);

        for l := 0 to ExpCondIdlGen.ClassCount - 1 do begin
          ParamClasses.Add(ExpCondIdlGen.Params[l]);
        end;
        Params.Free;
        //Params := nil;

        MIPModel  := TQUB_MIP_Model.Create(Model.ModelObject, True{DoInitialize}, ParamClasses, @prOptParams);
        prMIPModels.Add(MIPModel);

        for l := FirstChunk to prChunks.Count - 1 do begin
          MIPChunk                   := TQUB_MIP_PChunk(prChunks[l]);
          MIPChunk.MIPModel          := MIPModel;

          MIPChunk.ExpCondIdl        := ExpCondIdlGen.Idl[l-FirstChunk];
          MIPChunk.ExpCondIdlCount   := ExpCondIdlGen.IdlCount[l-FirstChunk];

          Dispose( TQUB_PFirstLast( ChunkBounds[l - FirstChunk] ) );
        end;

        ChunkBounds.Free;
        ExpCondIdlGen.Free;

        Inc(FileIndex);

      end;

      DS           := DS.SiblingSameName;

    end;


    //**************************************************************************


    for c := 0 to prChunks.Count - 1 do begin

      Chunk               := prChunks[c];
      Chunk.Status        := CHUNK_STATUS_EMPTY;
      Chunk.Trials        := 0;
      Chunk.Index         := c;

    end;

    if prChunks.Count = 0 then begin

      Result     := True;
      prError    := True;

      if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
        prErrorMsg := 'The file list is empty'; //(View menu -> MIP Exp Cond List)';
      end else begin
        prErrorMsg := 'No Idealized Data';
      end;

      Exit;

    end;


  finally


    ParamsAreVar.Free;


  end;

end;
*)
//This is the last version, but did not work. The following call crashes:
//function MplexIdlWithCondFromDll(idl :TQUB_IdlSegments_Impl; condIdl :TQUB_QFS_PNodeMem; MaxPower :Integer): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_MplexIdlWithCond_iface@12';
//   is that still true? CN.

function TQUB_Task_MIP.prChunkData :Boolean;
var
  DataSets         :INode;
  DS, Seg          :INode;
  ResDS, ResSeg    :INode;
  PlexDS           :INode;
  PlexSeg          :INode;
  ECDS             :INode;
  ECSeg            :INode;

  DwellCount       :Integer;
  Powers           :TPAInteger;
  Classes          :TPAInteger;
  CondClasses      :TPAInteger;

  Dat              :TQUB_FirstLast;

  QVD              :TQUB_QVO_QVD;

  ParamClasses     :TQUB_ExpCond_ParametersList;
  Params           :TQUB_ExpCond_Parameters;

  Chunk            :TQUB_Thread_PChunk;
  MIPChunk         :TQUB_MIP_PChunk;
  MIPModel         :TQUB_MIP_Model;

  FileIndex        :Integer;

  i, p             :Integer;
  DataPts          :Integer;
begin

  Result           := False;

  prSetChunkParameters;


  try


    prDataSetCount := 0;

    //if prScripting then begin
    //  DataSets := DataNoderScript.GetChosenIdealization(TNode.CreateNull);
    //end else begin
      DataSets       := DataNoder.GetChosenIdealization( TNode.CreateNull );
    //end;
    DS             := DataSets.Find('DataSet');
    FileIndex      := 0;
    ResDS          := TNode.CreateNull;


    while not DS.IsNull do begin

      Inc(prDataSetCount);

      Seg          := DS.Find('Segment');

      if Seg.IsNull then begin
        DS := DS.SiblingSameName;
        Inc(FileIndex);
        Continue;
      end;

      ResDS        := prResults.InsertClone(ResDS, DS, True);
      ResSeg       := ResDS.Find('Segment');

      //Chris, the Par.IsVar function is no longer calculated ! You eliminated it with the new chunking
      //It determines whether a Par changes across files when Par.Channel = -1 (else presumably does change)
      //Task_MIP.IsPZero is no longer reliable

      QVD         := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[ DS['FileName'].DataAsString ]);

      Params      := TQUB_ExpCond_Parameters.CreateCopy(QVD.ExpCond);
      for p := 0 to Params.Count - 1 do begin
        Params.AddPar('Cond_' + Params[p].Name, Params[p].Cond_Value,
                      Params[p].Index, Params[p].IsVar, -1);
      end;

      ParamClasses := TQUB_ExpCond_ParametersList.Create;
      PlexDS       := prGetIdlAndCond(QVD, Params, ParamClasses);
      Params.Free;

      PyNode_Publish(PlexDS, 'plexds');

      PlexSeg      := PlexDS.Find('Segment');

      if PlexSeg.IsNull then begin
        for p := 0 to ParamClasses.Count - 1 do begin
          ParamClasses[p].Free;
        end;
        ParamClasses.Free;
      end;

      MIPModel  := TQUB_MIP_Model.Create(Model.ModelObject, True{DoInitialize}, ParamClasses, @prOptParams);
      prMIPModels.Add(MIPModel);

      ECDS         := PlexDS.Find('ExpCondIdl');
      ECSeg        := ECDS.Find('Segment');

      while not PlexSeg.IsNull do begin

        Dat                      := QUB_FirstLast(Seg.Data[0], Seg.Data[1]);
        Chunk                    := prNewChunk(Dat);

        MIPChunk                 := TQUB_MIP_PChunk(Chunk);
        prChunks.Add(MIPChunk);

        Chunk.StartTime          := Double(Seg['start'].Data[0]) * 1.0e-3;
        MIPChunk.SkipPoints      := QUB_FirstLast(0, 0); // ??? Skp;
        MIPChunk.IdlSel          := QUB_FirstLast(0, PlexSeg['DwellCount'].Data[0]);  // ??? Idl^; // is this used? the upper bound seems to be one too large
        MIPChunk.QVD             := QVD;
        MIPChunk.MIPModel        := MIPModel;
        MIPChunk.FileIndex       := FileIndex;
        MIPChunk.Sampling        := DS['sampling'].Data[0] * 1.0e-3;
        MIPChunk.Scaling         := QVD.DataObject.Scaling;

        // first the ExpCondIdl (not mplexed with dataidl or broken into powers)
        DwellCount               := ECSeg['DwellCount'].Data[0];
        MIPChunk.ExpCondIdlCount := DwellCount;
        GetMem(MIPChunk.ExpCondIdl, DwellCount * SizeOf(TQUB_QFS_QIF_NClass));
        Powers {actually Counts} := TPAInteger(ECSeg['Counts'].DataPtr);
        Classes                  := TPAInteger(ECSeg['Classes'].DataPtr);
        for i := 0 to DwellCount - 1 do begin
          MIPChunk.ExpCondIdl[i].N   := Powers[i];
          MIPChunk.ExpCondIdl[i].Cls := Classes[i];
        end;

        // then the full powered, multiplexed stuff
        DwellCount               := PlexSeg['DwellCount'].Data[0];
        Powers                   := TPAInteger( PlexSeg['Powers'].DataPtr );
        Classes                  := TPAInteger( PlexSeg['Classes'].DataPtr );
        CondClasses              := TPAInteger( PlexSeg['CondClasses'].DataPtr );

        MIPChunk.DataCount       := DwellCount;
        GetMem(MIPChunk.Data, DwellCount * SizeOf(TQUB_MIP_DataPoint));
        DataPts                  := 0;
        for i := 0 to DwellCount - 1 do begin
          Inc(DataPts, (1 shl Powers[i]));
          MIPChunk.Data[i].Pow   := ShortInt(Powers[i]);
          MIPChunk.Data[i].Cls   := Byte(Classes[i]);
          MIPChunk.Data[i].ExpCondCls := Byte(CondClasses[i]);
        end;
        Inc(TQUB_MIP_Model(prMIPModels[0]).TotalDataPoints, DataPts);
        Inc(TQUB_MIP_Model(prMIPModels[0]).Multiplications, DwellCount);
        Inc(TQUB_MIP_Model(prMIPModels[0]).TotalDwellCount, DwellCount);

        MIPChunk.ResultSeg       := ResSeg.GetNodeMem;
        ResSeg.RemoveChild(ResSeg['Classes']);
        ResSeg.RemoveChild(ResSeg['Durations']);
        ResSeg.RemoveChild(ResSeg['Firsts']);
        ResSeg.RemoveChild(ResSeg['Lasts']);

        // no longer used
        MIPChunk.IdlSeg          := nil;

        Seg                      := Seg.SiblingSameName;
        ResSeg                   := ResSeg.SiblingSameName;
        PlexSeg                  := PlexSeg.SiblingSameName;
        ECSeg                    := ECSeg.SiblingSameName;

      end;//while not Seg.IsNull

      Inc(FileIndex);
      DS           := DS.SiblingSameName;

    end;


    //**************************************************************************


    for i := 0 to prChunks.Count - 1 do begin

      Chunk               := prChunks[i];
      Chunk.Status        := CHUNK_STATUS_EMPTY;
      Chunk.Trials        := 0;
      Chunk.Index         := i;

    end;

    if prChunks.Count = 0 then begin

      Result     := True;
      prError    := True;

      if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
        prErrorMsg := 'The file list is empty'; //(View menu -> MIP Exp Cond List)';
      end else begin
        prErrorMsg := 'No Idealized Data.  Check the Data Source.';
      end;
      prWikiTopic := 'Error:No_Data';

      Exit;

    end;


  finally



  end;

end;


//This new version is modeled after Staircase's, which does work
(*
function TQUB_Task_MIP.prChunkData :Boolean;
var
  DataSets         :INode;
  DS, Seg          :INode;

  QVD              :TQUB_QVO_QVD;

  ParamClasses     :TQUB_ExpCond_ParametersList;
  Params           :TQUB_ExpCond_Parameters;
  ParamsAreVar     :TQUB_ExpCond_Parameters;

  Chunk            :TQUB_Thread_PChunk;
  MIPChunk         :TQUB_MIP_PChunk;
  MIPModel         :TQUB_MIP_Model;

  Dat              :TQUB_FirstLast;

  FileIndex        :Integer;

  c, p             :Integer;
begin

  Result           := False;

  prSetChunkParameters;
                      
  ParamsAreVar     := nil;//Staircase


  try

    prDataSetCount := 0;

    DataSets       := DataNoder.GetChosenIdealization( TNode.CreateNull );
    DS             := DataSets.Find('DataSet');
    FileIndex      := 0;
    while not DS.IsNull do begin
      Seg          := DS.Find('Segment');

      if not Seg.IsNull then begin
        //ExpConds
        ParamClasses := TQUB_ExpCond_ParametersList.Create;
        prExpCondParams.Add(ParamClasses);
        Params := TQUB_ExpCond_Parameters.Create;
        ParamClasses.Add(Params);
        Params.ReadTree( DS.Find('ExpCond') );
        for p := 0 to Params.Count - 1 do begin
          Params.AddPar('Cond_' + Params[p].Name, Params[p].Cond_Value,
                        Params[p].Index, Params[p].IsVar, -1);
        end;

        //MIPModel
        MIPModel  := TQUB_MIP_Model.Create(Model.ModelObject, True{DoInitialize}, ParamClasses, @prOptParams);
        prMIPModels.Add(MIPModel);

        QVD        := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DS['FileName'].DataAsString]);

        while not Seg.IsNull do begin

          Inc(prDataSetCount);

          Dat                   := QUB_FirstLast(Seg.Data[0], Seg.Data[1]);
          Chunk                 := prNewChunk(Dat);

          MIPChunk              := TQUB_MIP_PChunk(Chunk);

          prChunks.Add(MIPChunk);

          MIPChunk.IdlSeg       := Seg.GetNodeMem;
          Node_INCREF( MIPChunk.IdlSeg );

          MIPChunk.SkipPoints   := QUB_FirstLast(0, 0); // ??? Skp;

          MIPChunk.IdlSel       := QUB_FirstLast(0, Seg['DwellCount'].Data[0]); // ??? Idl^;

          MIPChunk.QVD          := QVD;

          MIPChunk.Sampling     := DS['sampling'].Data[0] * 1.0e-3;
          MIPChunk.Scaling      := QVD.DataObject.Scaling;

          MIPChunk.Data         := nil;

          MIPChunk.MIPModel     := MIPModel;

          MIPChunk.FileIndex    := FileIndex;

          Seg        := Seg.SiblingSameName;
        end;

        Inc(FileIndex);
      end;

      DS           := DS.SiblingSameName;
    end;

    for c := 0 to prChunks.Count - 1 do begin

      Chunk               := prChunks[c];
      Chunk.Status        := CHUNK_STATUS_EMPTY;
      Chunk.Trials        := 0;
      Chunk.Index         := c;

    end;

    if prChunks.Count = 0 then begin

      Result     := True;
      prError    := True;

      if DataSource = DATA_SOURCE_FILELIST then begin
        prErrorMsg := 'The file list is empty (View menu -> MIP Exp Cond List)';
      end else begin
        prErrorMsg := 'No Idealized Data';
      end;

      Exit;

    end;


  finally


    ParamsAreVar.Free;


  end;

end;
*)
procedure TQUB_Task_MIP.prSetChunkParameters;
begin
  inherited;
  prChunkSize := 0;
  prChunkSegmentation := True;
end;

procedure TQUB_Task_MIP.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MIPChunk :TQUB_MIP_PChunk;
begin

  MIPChunk := TQUB_MIP_PChunk(Chunk);

  if MIPChunk.IdlSeg <> nil then begin
    Node_DECREF( MIPChunk.IdlSeg );
    MIPChunk.IdlSeg := nil;
  end;

  if MIPChunk.Data <> nil then begin
    FreeMem(MIPChunk.Data);
    MIPChunk.Data := nil;
  end;
  if MIPChunk.ExpCondIdl <> nil then begin
    FreeMem(MIPChunk.ExpCondIdl);
    MIPChunk.ExpCondIdl := nil;
  end;
  if MIPChunk.AlphaS <> nil then begin
    Free2(TPMChar(MIPChunk.AlphaS));
    MIPChunk.AlphaS := nil;
  end;
  if MIPChunk.BetaS <> nil then begin
    Free2(TPMChar(MIPChunk.BetaS));
    MIPChunk.BetaS := nil;
  end;
  if MIPChunk.LScale <> nil then begin
    FreeMem(MIPChunk.LScale);
    MIPChunk.LScale := nil;
  end;

  FreeMem(MIPChunk);

end;

function TQUB_Task_MIP.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result           := AllocMem(SizeOf(TQUB_MIP_Chunk));

  Result.Sel       := Sel;

end;


//******************************************************************************
function TQUB_Task_MIP.prGetIdlAndCond(QVD :TQUB_QVO_QVD; ParamTemplate :TQUB_ExpCond_Parameters; ParamClasses :TQUB_ExpCond_ParametersList): INode;
var
  StimChannels     :TList;
  StimParams       :TPAInteger;
  StimChannelCount :Integer;
  Segments         :TQUB_PAFirstLast;
  SegCount         :Integer;
  i, p             :Integer;
  CondIdls         :TList;
  IdlDS            :INode;
  CondDS           :INode;
  PlexDS           :INode;
  Seg              :INode;
begin
  StimChannels     := nil;
  StimParams       := nil;
  Segments         := nil;
  CondIdls         := nil;
  Result           := QUB_NODE_NULL;

  try
    // find stimulus channels
    StimChannels     := TList.Create;
    GetMem(StimParams, QVD.DataObject.ADChannelCount * SizeOf(Integer));
    StimChannelCount := 0;
    for p := 0 to ParamTemplate.Count - 1 do begin
      if ParamTemplate[p].Channel >= 0 then begin
        StimChannels.Add(Pointer(ParamTemplate[p].Channel));;
        StimParams[StimChannelCount] := p;
        Inc(StimChannelCount);
      end;
    end;

    // get idealization and seg bounds
    IdlDS            := QVD.GetChosenIdl(True{fragments});
    SegCount         := 0;
    Seg              := IdlDS.Find('Segment');
    while not Seg.IsNull do begin
      Inc(SegCount);
      Seg            := Seg.SiblingSameName;
    end;
    GetMem(Segments, SegCount * SizeOf(TQUB_FirstLast));
    Seg              := IdlDS.Find('Segment');
    i                := 0;
    while not Seg.IsNull do begin
      Segments[i].First := Seg.Data[0];
      Segments[i].Last  := Seg.Data[1];
      Seg               := Seg.SiblingSameName;
      Inc(i);
    end;

    // make sure the condition channels are idealized
    prEnsureIdlStim(QVD, StimChannels, SegCount, Segments);

    // get multiplexed condition idealization
    CondIdls         := TList.Create;
    for i := 0 to StimChannels.Count - 1 do begin
      CondIdls.Add(QVD.DataObject.Idealization[Integer(StimChannels[i])]);
    end;
    CondDS           := MplexIdl(CondIdls, SegCount, Segments);

    // and multiplex with idealization
    PlexDS           := MplexIdlWithCond(QVD.DataObject.Idealization[QVD.ActiveChannel], CondDS, prOptParams.OptMaxPower);
    PlexDS.AppendChild(CondDS); // since the expcondidl is _also_ used !
    CondDS.Name      := 'ExpCondIdl';

    // build ParamClasses
    MplexAmpsToParamClasses(PlexDS['MplexAmps'], StimParams, ParamTemplate, ParamClasses);

    Result           := PlexDS;

  finally
    StimChannels.Free;
    FreeMem(StimParams);
    FreeMem(Segments);
    CondIdls.Free;
  end;
end;

procedure TQUB_Task_MIP.prEnsureIdlStim(QVD :TQUB_QVO_QVD; StimChannels :TList{of Integer}; SegCount :Integer; Segments :TQUB_PAFirstLast);
var
  i        :Integer;
  c        :Integer;
  s        :Integer;
  ef, el, ec :Integer;
  ea       :Double;
  Iter     :TQUB_IdlSegmentsIter;
  ReIdl    :Boolean;
begin
  i        := 0;
  while i < StimChannels.Count do begin
    c      := Integer(StimChannels[i]);
    ReIdl  := False;

    for s := 0 to SegCount - 1 do begin
      Iter := QVD.DataObject.Idealization[c].FindIter(Segments[s].First);
      while 0<>Iter.Valid do begin
        Iter.Read(@ef, @el, @ec, @ea);
        if el >= Segments[s].Last then Break;
        if ec < 0 then begin
          ReIdl := True;
          Break;
        end;
        Iter.Next;
      end;
      Iter.Free;
      if ReIdl then Break;
    end;

    if ReIdl then begin
      prIdlStimSignal   := TEvent.Create(nil, False, False, '');
      QVD.ActiveChannel := c;
      Synchronize(prStartIdlStim);
      while prIdlStimSignal.WaitFor(1000) = wrTimeout do begin end;
      prIdlStimSignal.Free;
    end else begin
      Inc(i);
    end;
  end;

  QVD.ActiveChannel := ChannelIndex;
end;

procedure TQUB_Task_MIP.prStartIdlStim;
begin
  ActionsList.ByName['Preprocessing'].ActionByName['IdlStim'].OnRClick(prIdlStimDone);
end;

procedure TQUB_Task_MIP.prIdlStimDone(Sender :TObject);
begin
  prIdlStimSignal.SetEvent;
end;


//******************************************************************************
procedure TQUB_Task_MIP.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then begin
    if (prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST) then begin
      prOptSegsOldMdl := TQUB_QFS_QMF.Create('', nil);
      prOptSegsOldMdl.CopyFrom(prModel.ModelObject);
    end;
    prModel.LockAny;
  end;
end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MIP.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_MIP_PropertiesForm do begin

    Self.ChannelIndex           := Properties['ChannelIndex'].Data[0];

    //Client/Server
    Self.ThreadCount            := Properties['ThreadCount'].Data[0];
    Self.Remote                 := 0<>Properties['Remote'].Data[0];

    //Data Source
    Self.DataSource             := DataSource;

    Self.QuietOutput            := QuietOutput;

    Self.OptParams              := Params;

    prGroupList                 := txtOptSegmentsGroupList.Text;

    Self.Properties             := prResults.AppendClone(Properties, True);

  end;//with
end;

procedure TQUB_Task_MIP.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_MIP_PropertiesForm do begin
  //  ChannelIndex                := Self.ChannelIndex;
  //end;//with
end;



//****************************************************************************//
//                                                                            //
//                      D A T A     C A L L   -   B A C K                     //
//                                                                            //
//                                    N E W                                   //
//                                                                            //
//****************************************************************************//
//This is the last version - I'm replacing it with the older below
//TO DO: LM. Chris says give it a try
function TQUB_Task_MIP.DataCallBack(MIPChunk :TQUB_MIP_PChunk) :Boolean;
begin

  Result       := False;

  try

    if MIPChunk.AlphaS = nil then begin
    //Allocate

      MIPChunk.AlphaS := EAlloc2(MIPChunk.DataCount, MIPChunk.MIPModel.StateCount);
      MIPChunk.BetaS  := EAlloc2(MIPChunk.DataCount, MIPChunk.MIPModel.StateCount);
      MIPChunk.LScale := EAlloc1(MIPChunk.DataCount);

    end else begin
    //DeAllocate

      Free2(TPMChar(MIPChunk.AlphaS));
      MIPChunk.AlphaS := nil;
      Free2(TPMChar(MIPChunk.BetaS));
      MIPChunk.BetaS := nil;
      FreeMem(MIPChunk.LScale);
      MIPChunk.LScale := nil;

    end;


  except

    Result := True;

  end;

end;
(*
function TQUB_Task_MIP.DataCallBack(MIPChunk :TQUB_MIP_PChunk) :Boolean;
var
  Seg          :INode;
  DataIdlGen   :TQUB_SegNodeIdlGen;
  ECIdlGen     :TQUB_NCIdlGen;
  Mplex        :TQUB_MPlexIdlGen;
  PrevCond     :Integer;
  Event        :TQUB_QFS_QIF_NClass;
  LastPoint    :TQUB_MIP_DataPoint;
  DataCountMax :Integer;
  Count        :Integer;
  CntPow       :Integer;
  MaxPow       :Integer;
  PowOf2       :Integer;
begin

  Result       := False;


  try

    if MIPChunk.Data = nil then begin
    //Allocate

      DataIdlGen := nil;
      ECIdlGen   := nil;
      Mplex      := nil;

      MaxPow   := MIP_PowerOf2(prOptParams.OptMaxPower);

      try

        Seg    := TNode.CreateUsing( MIPChunk.IdlSeg );

        DataIdlGen := TQUB_SegNodeIdlGen.Create(Seg);
        ECIdlGen   := TQUB_NCIdlGen.Create(MIPChunk.ExpCondIdl, MIPChunk.ExpCondIdlCount);
        Mplex      := TQUB_MPlexIdlGen.Create;
        Mplex.Add(DataIdlGen);
        Mplex.Add(ECIdlGen);

        DataCountMax       := 10000;
        MIPChunk.DataCount := 0;
        ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * DataCountMax);

        PrevCond := -1;

        while Mplex.HasMore do begin

          Mplex.Next(Event);

          Count := Event.N;

          Inc(TQUB_MIP_Model(prMIPModels[0]).TotalDataPoints, Count);

          // split off the first point if it's the first event or exp.conditions just changed.
          if (MIPChunk.DataCount = 0) or (Mplex.Classes[ Event.Cls ][1] <> PrevCond) then begin

            //Need to reallocate ?
            if (1 + MIPChunk.DataCount) > DataCountMax then begin
              Inc(DataCountMax, 10000);
              ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * DataCountMax);
            end;

            CntPow := 0;
            PowOf2 := 1;
            MIPChunk.Data[MIPChunk.DataCount].Cls := Mplex.Classes[ Event.Cls ][0]; // DC[i].Cls;
            MIPChunk.Data[MIPChunk.DataCount].ExpCondCls := Mplex.Classes[ Event.Cls ][1];
            MIPChunk.Data[MIPChunk.DataCount].Pow := CntPow;
            Inc(MIPChunk.DataCount);
            Dec(Count, PowOf2);
            Inc(TQUB_MIP_Model(prMIPModels[0]).Multiplications);

          end;

          if Count > 0 then begin

            // split off the last point (strictly necessary only for last dwell and the dwells just before a conductance change)
            LastPoint.Cls           := Mplex.Classes[ Event.Cls ][0];
            LastPoint.ExpCondCls    := Mplex.Classes[ Event.Cls ][1];
            LastPoint.Pow           := 0;
            Dec(Count, 1);

            if Count > 0 then begin
              repeat

                CntPow := MIP_MaxPower(Count, MaxPow, PowOf2);

                Inc(TQUB_MIP_Model(prMIPModels[0]).Multiplications);

                //Need to reallocate ?
                if (1 + MIPChunk.DataCount) > DataCountMax then begin
                  Inc(DataCountMax, 10000);
                  ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * DataCountMax);
                end;

                MIPChunk.Data[MIPChunk.DataCount].Cls := Mplex.Classes[ Event.Cls ][0]; // DC[i].Cls;
                MIPChunk.Data[MIPChunk.DataCount].ExpCondCls := Mplex.Classes[ Event.Cls ][1];
                MIPChunk.Data[MIPChunk.DataCount].Pow := CntPow;

                Inc(MIPChunk.DataCount);

                Dec(Count, PowOf2);

              until(Count <= 0);
            end;

            // append last point
            //Need to reallocate ?
            if (1 + MIPChunk.DataCount) > DataCountMax then begin
              Inc(DataCountMax, 10000);
              ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * DataCountMax);
            end;

            MIPChunk.Data[MIPChunk.DataCount] := LastPoint;

            Inc(MIPChunk.DataCount);
            Inc(TQUB_MIP_Model(prMIPModels[0]).Multiplications);

          end;

          PrevCond := Mplex.Classes[ Event.Cls ][1];

        end;

        Inc(TQUB_MIP_Model(prMIPModels[0]).TotalDwellCount, MIPChunk.DataCount);
        if MIPChunk.DataCount <> DataCountMax then begin
          ReAllocMem(MIPChunk.Data, SizeOf(TQUB_MIP_DataPoint) * MIPChunk.DataCount);
        end;

        MIPChunk.AlphaS := EAlloc2(MIPChunk.DataCount, MIPChunk.MIPModel.StateCount);
        MIPChunk.BetaS  := EAlloc2(MIPChunk.DataCount, MIPChunk.MIPModel.StateCount);
        MIPChunk.LScale := EAlloc1(MIPChunk.DataCount);

      finally

        Mplex.Free;
        ECIdlGen.Free;
        DataIdlGen.Free;

      end;


    end else begin
    //DeAllocate

      FreeMem(MIPChunk.Data);
      MIPChunk.Data := nil;
      MIPChunk.DataCount := 0;
      Free2(TPMChar(MIPChunk.AlphaS));
      MIPChunk.AlphaS := nil;
      Free2(TPMChar(MIPChunk.BetaS));
      MIPChunk.BetaS := nil;
      FreeMem(MIPChunk.LScale);
      MIPChunk.LScale := nil;

    end;


  except

    Result := True;

  end;

end;
*)

//******************************************************************************
end.
