unit QUB_Task_MIP_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_Task,
QUB_Properties,
QUB_Task_Properties,
QUB_Q_Model,
QUB_MIP_Global,
QUB_Node,
QUB_DataList;


{$include 'QUB_Task_MIP_Properties.hct'}


const


DEF_QUIETOUTPUT               = 0;


type


TQUB_Task_MIP_PropertiesForm      = class(TQUB_Task_PropertiesForm)


  Label2                          :TLabel;
  Label8                          :TLabel;

  btnReset                        :TButton;
  panClientServer                 :TPanel;
  chkRemote                       :TCheckBox;
  txtThreadCount                  :TEdit;
  udThreadCount                   :TUpDown;
  chkQuietOutput                  :TCheckBox;
  Panel4                          :TPanel;
  Label13                         :TLabel;
  Label14                         :TLabel;
  Label15                         :TLabel;
  Label16                         :TLabel;
  Label17                         :TLabel;
    txtOptLLConv: TEdit;
    txtOptGradConv: TEdit;
    txtOptMaxIter: TEdit;
    txtOptMaxStep: TEdit;
  chkOptRestart                   :TCheckBox;
  Panel5                          :TPanel;
  Label18                         :TLabel;
    chkEstInitProbsFixed: TRadioButton;
    chkEstInitProbsEq: TRadioButton;
    chkEstInitProbsCondEq: TRadioButton;
  Panel8                          :TPanel;
  Label20                         :TLabel;
  Label21                         :TLabel;
  chkOptSegmentsTogether          :TRadioButton;
  chkOptSegmentsGroup             :TRadioButton;
  txtOptSegmentsGroupCount        :TEdit;
  udOptSegmentsGroupCount         :TUpDown;
  chkOptIdenticalChunks           :TCheckBox;
  chkOptSegmentsGroupList         :TRadioButton;
  txtOptSegmentsGroupList         :TEdit;
  txtOptSegmentsMaxBatch          :TEdit;
  chkOptMUXFiles                  :TCheckBox;
  Panel7                          :TPanel;
  Label22                         :TLabel;
    chkOptModeOpt: TRadioButton;
    chkOptModeCheck: TRadioButton;
  Label5                          :TLabel;
  txtOptMaxPower                  :TEdit;
    txtModMinRate: TEdit;
    Label4: TLabel;
    chkOptAllowNegRates: TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);


protected


  //Data Source
  prDataList                      :TQUB_DataList;

  //QuietOutput
  prQuietOutput                   :Boolean;

  //Params
  prParams                        :TQUB_MIP_Opt_Params;


public


  procedure PropertiesToParams;


  //Quiet output
  property QuietOutput                    :Boolean                   read prQuietOutput;                   

  //Estimation
  property EstInitProbs                   :TQUB_Q_InitProbs          read prParams.EstInitProbs;           

  //Optimization
  property OptLLConv                      :Double                    read prParams.OptLLConv;
  property OptGradConv                    :Double                    read prParams.OptGradConv;
  property OptMaxStep                     :Double                    read prParams.OptMaxStep;
  
  property OptMaxIter                     :Integer                   read prParams.OptMaxIter;
  property OptMaxPower                    :Integer                   read prParams.OptMaxPower;
  property OptStimDelta                   :Integer                   read prParams.OptStimDelta;
  property OptRestart                     :Integer                   read prParams.OptRestart;
  property OptIdenticalChunks             :Integer                   read prParams.OptIdenticalChunks;
  property OptMode                        :TQUB_MIP_Opt_Mode         read prParams.OptMode;

  property OptAllowNegRates               :Boolean                   read prParams.OptAllowNegRates;
  property OptSegments                    :TQUB_OptimizeSegments     read prParams.OptSegments;
  property OptSegmentsGroupCount          :Integer                   read prParams.OptSegmentsGroupCount;
  property OptSegmentsMaxBatch            :Integer                   read prParams.OptSegmentsMaxBatch;
  property OptMUXFiles                    :Integer                   read prParams.OptMUXFiles;

  property ModQMinRate                    :Double                    read prParams.ModQMinRate;

  property Params                         :TQUB_MIP_Opt_Params       read prParams;


end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_MIP_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := 20;
  prDataList.Top        := 50;
  prDataList.Width      := 560;
  prDataList.Height     := 100;
  // prDataList.FileListAvail := False;

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["QuietOutput"]',
                        'QuietOutput', REGTYPE_BOOL, DEF_QUIETOUTPUT,
                        chkQuietOutput));
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Est"]["InitProbs"]',
                          'Estimate InitProbs', REGTYPE_INT, Q_InitProbsStringOf(DEF_EST_INITPROBS))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkEstInitProbsFixed,
                                       Q_InitProbsStringOf(INITPROBS_FIXED),
                                       Integer(INITPROBS_FIXED)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkEstInitProbsEq,
                                       Q_InitProbsStringOf(INITPROBS_EQ),
                                       Integer(INITPROBS_EQ)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkEstInitProbsCondEq,
                                       Q_InitProbsStringOf(INITPROBS_CONDEQ),
                                       Integer(INITPROBS_CONDEQ)))
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvLL"]',
                    'Optimizer LLConv', REGTYPE_FLOAT, DEF_OPT_LLCONV,
                    txtOptLLConv,
                    AcceptFloatBetween(MIN_OPT_LLCONV, MAX_OPT_LLCONV),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvGrad"]',
                    'Optimizer GradConv', REGTYPE_FLOAT, DEF_OPT_GRADCONV,
                    txtOptGradConv,
                    AcceptFloatBetween(MIN_OPT_GRADCONV, MAX_OPT_GRADCONV),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxIterations"]',
                    'Optimizer MaxIter', REGTYPE_INT, DEF_OPT_MAXITER,
                    txtOptMaxIter,
                    AcceptIntBetween(MIN_OPT_MAXITER, MAX_OPT_MAXITER),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxStep"]',
                    'Optimizer MaxStep', REGTYPE_FLOAT, DEF_OPT_MAXSTEP,
                    txtOptMaxStep,
                    AcceptFloatBetween(MIN_OPT_MAXSTEP, MAX_OPT_MAXSTEP),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxPower"]',
                    'Optimizer MaxPower', REGTYPE_INT, DEF_OPT_MAXPOWER,
                    txtOptMaxPower,
                    AcceptIntBetween(MIN_OPT_MAXPOWER, MAX_OPT_MAXPOWER),
                   '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["DFP"]["Restart"]',
                        'Optimizer Restart', REGTYPE_INT, DEF_OPT_RESTART,
                        chkOptRestart));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["AllowNegRates"]',
                        'Optimizer AllowNegRates', REGTYPE_BOOL, DEF_OPT_ALLOWNEGRATES,
                        chkOptAllowNegRates));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["IdenticalChunks"]',
                        'Optimizer IdenticalChunks', REGTYPE_INT, DEF_OPT_IDENTICALCHUNKS,
                        chkOptIdenticalChunks));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["MUXFiles"]',
                        'Optimizer MUXFiles', REGTYPE_INT, DEF_OPT_MUXFILES,
                        chkOptMUXFiles));
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Mode"]',
                          'Optimizer Mode', REGTYPE_INT, OptModeToString(DEF_OPT_MODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptModeOpt,
                                       OptModeToString(MODE_OPTIMIZE),
                                       Integer(MODE_OPTIMIZE)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptModeCheck,
                                       OptModeToString(MODE_CHECK),
                                       Integer(MODE_CHECK)))
                       );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["use segments"]',
                          'Optimizer Segments', REGTYPE_INT, SegModeToString(DEF_OPT_SEGMENTS))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsTogether,
                                       SegModeToString(OPTSEGS_TOGETHER),
                                       Integer(OPTSEGS_TOGETHER)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsGroup,
                                       SegModeToString(OPTSEGS_GROUP),
                                       Integer(OPTSEGS_GROUP))
                                .Enables(txtOptSegmentsGroupCount)
                                .Enables(udOptSegmentsGroupCount)
                            )
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptSegmentsGroupList,
                                       SegModeToString(OPTSEGS_GROUPLIST),
                                       Integer(OPTSEGS_GROUPLIST))
                                .Enables(txtOptSegmentsGroupList)
                            )
                       );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["SegGroupCount"]',
                      'Optimizer Segments GroupCount', REGTYPE_INT, DEF_OPT_SEGMENTS_GROUPCOUNT,
                      udOptSegmentsGroupCount,
                      txtOptSegmentsGroupCount,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["OptSegGroupList"]',
                    'Optimizer Segments GroupList', REGTYPE_STRING, IntToStr(DEF_OPT_SEGMENTS_GROUPCOUNT),
                    txtOptSegmentsGroupList,
                    '',
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SegMaxBatch"]',
                    'Optimizer Segments MaxBatch', REGTYPE_INT, DEF_OPT_MAXBATCH,
                    txtOptSegmentsMaxBatch,
                    AcceptIntBetween(MIN_OPT_MAXBATCH, MAX_OPT_MAXBATCH),
                    '') );

  RegKeyName            := 'Software\QUB\Task MIP optimizer';
  ProfileCat            := 'MIP';
  prBeforeGetProperties := PropertiesToParams;

  WikiTopic             := 'Modeling:MIP#Properties';

  {$include 'QUB_Task_MIP_Properties.htx'}
end;

procedure TQUB_Task_MIP_PropertiesForm.FormShow(Sender: TObject);
var
  Model                  :TQUB_QVO_QVM;
begin
  inherited;

  prDataList.ShowFiles;

  Model                  := TQUB_Viewport_Model( ViewportList.ViewportByName['Model'] ).CurrModel;
  if Model <> nil then begin
    prDataList.ShowModelVars( Model, True );
  end;

end;

procedure TQUB_Task_MIP_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  PropertiesToParams;

  prDataList.HideAll;
end;


//******************************************************************************
procedure TQUB_Task_MIP_PropertiesForm.PropertiesToParams;
begin
  prQuietOutput                  := 0<>Properties['QuietOutput'].Data[0];
  prParams.OptLLConv             := Properties['DFP']['ConvLL'].Data[0];
  prParams.OptGradConv           := Properties['DFP']['ConvGrad'].Data[0];
  prParams.OptMaxIter            := Properties['DFP']['MaxIterations'].Data[0];
  prParams.OptMaxStep            := Properties['DFP']['MaxStep'].Data[0];
  prParams.OptMaxPower           := Properties['MaxPower'].Data[0];
  prParams.OptMode               := StringToOptMode(Properties['Mode'].DataAsString);
  prParams.OptSegments           := StringToSegMode(Properties['use segments'].DataAsString);
  prParams.OptSegmentsGroupCount := Properties['SegGroupCount'].Data[0];
  txtOptSegmentsGroupList.Text   := Properties['OptSegGroupList'].DataAsString;
  prParams.OptSegmentsMaxBatch   := Properties['SegMaxBatch'].Data[0];
  prParams.OptIdenticalChunks    := Properties['IdenticalChunks'].Data[0];
  prParams.OptMUXFiles           := Properties['MUXFiles'].Data[0];
  prParams.EstInitProbs          := Q_InitProbsOfString( Properties['Est']['InitProbs'].DataAsString );
  prParams.OptRestart            := Properties['DFP']['Restart'].Data[0];
  prParams.OptStimDelta          := DEF_OPT_STIMDELTA;//Properties['OptStimDelta'].Data[0];
  prParams.OptAllowNegRates      := 0<>Properties['AllowNegRates'].Data[0];
end;


//******************************************************************************
end.
