unit QUB_Task_MMerge_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ComCtrls, ExtCtrls, Grids,

QUB_Table, 
QUB_Viewport_Model,
QUB_QVO,
QUB_QVO_QVM, QUB_Task_Properties;


type


TQUB_Task_MMerge_PropertiesForm = class(TQUB_Task_PropertiesForm)

    
  Label1                        :TLabel;
  panModels: TPanel;


  procedure FormShow(Sender: TObject);


protected

  grdModels                     :TQUB_Table;
  prModelCount                  :Integer;

  procedure grdModelsSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);

  function prGetModel(ix :Integer): TQUB_QVO_QVM;
  function prGetModelReps(ix :Integer): Integer;


public

  // you can only trust these during/immed.after showing this form
  property Models[ix :Integer]     :TQUB_QVO_QVM          read prGetModel;
  property ModelCount              :Integer               read prModelCount;
  property ModelReps[ix :Integer]  :Integer               read prGetModelReps;

end;


// todo:  make appropriate defaults in actionselector_modeling
//        come up with a shorter long-name than "Model Merge"



implementation


uses


QUB_GUIInterface,
QUB_ListBox, QUB_Main;


{$R *.DFM}


procedure TQUB_Task_MMerge_PropertiesForm.FormShow(Sender: TObject);
var
  ModelObjects         :TQUB_ListBox;
  i                    :Integer;
  AnySel               :Boolean;
begin
  inherited;

  if grdModels = nil then begin
    grdModels := TQUB_Table.Create(Self);
    grdModels.DefaultColWidth := 75;
    grdModels.DefaultRowHeight := 16;
    grdModels.Options         := [goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goRangeSelect, goRowSizing, goColSizing, goThumbTracking];
    grdModels.OnSetEditText   := grdModelsSetEditText;
    grdModels.Align := alClient;
    grdModels.Parent := panModels;
  end;

  ModelObjects         := ViewportModel.ListObjects;
  prModelCount         := ModelObjects.Items.Count;
  AnySel               := False;

  grdModels.RowCount   := prModelCount + 1;
  grdModels.ColCount   := 2;
  grdModels.Cells[0, 0] := 'Count';
  grdModels.Cells[1, 0] := 'Model';
  grdModels.FixedRows   := 1;
  grdModels.FixedCols   := 0;
  grdModels.ColWidths[0]    := 34;
  grdModels.ColWidths[1]    := 250;

  for i := 0 to ModelObjects.Items.Count - 1 do begin
    grdModels.Cells[1, i+1] := ExtractFileName(ModelObjects[i].FileName);
    if ModelObjects.Selected[i] then begin
      AnySel := True;
      grdModels.Cells[0, i+1] := '1';
    end else begin
      grdModels.Cells[0, i+1] := '0';
    end;
  end;

  btnOK.Enabled        := AnySel;

  WikiTopic            := 'Modeling:Model_Merge#Properties';
end;

function TQUB_Task_MMerge_PropertiesForm.prGetModel(ix :Integer): TQUB_QVO_QVM;
begin
  Result               := nil;
  if (ix < 0) or (ix >= ModelCount) then Exit;
  Result               := TQUB_QVO_QVM(ViewportModel.ListObjects[ix]);
// get the ix'th selected model
  //for i := 0 to lstModels.Items.Count - 1 do begin
  //  if lstModels.Selected[i] then begin
  //    if ix > 0 then begin
  //      ix             := ix - 1;
  //    end else begin
  //      Result         := TQUB_QVO_QVM(ViewportModel.ListObjects[i]);
  //      Exit;
  //    end;
  //  end;
  //end;
end;

function TQUB_Task_MMerge_PropertiesForm.prGetModelReps(ix :Integer): Integer;
begin
  try
    Result := StrToInt(grdModels.Cells[0, ix+1]);
  except
    Result := 0;
  end;
end;

procedure TQUB_Task_MMerge_PropertiesForm.grdModelsSetEditText(Sender: TObject; ACol,
  ARow: Integer; const Value: String);
var
  i         :Integer;
  AnySel    :Boolean;
begin
  if ACol <> 0 then Exit;
  if (Value = '') or (Value = '.') or (Value = '-') or (Value = '-.') then Exit;

  AnySel    := False;
  for i := 0 to ModelCount - 1 do begin
    AnySel := AnySel or (ModelReps[i] > 0);
    if AnySel then Break;
  end;
  btnOK.Enabled := AnySel;
end;

// *****************************************************************************
end.
