unit QUB_Task_MPL;


interface


uses


Windows, SysUtils, Classes, Clipbrd, Controls, Forms, Math, Dialogs,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QVO_QVM,

QUB_QFS_QMF,

QUB_Viewport_Data,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_MPL_Properties,

QUB_MIL_Global,
QUB_MIL,

Chart, Series,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, QUB_Viewport_Results,
QUB_GUIInterface;


const


MPL_REASONABLE_NPOINTS                              = 500000;



type


TQUB_MPL_Opt_Params          = record

  OptLLConv                  :Double;
  OptGradConv                :Double;
  OptMaxIter                 :Integer;
  OptRestarts                :Integer;
  OptSearchLimit             :Double;
  OptMaxStep                 :Double;
  OptMode                    :Integer;
  OptSegments                :Integer;
  OptPreprocessData          :Integer;
  OptFilterOn                :Integer;
  OptFilterFreq              :Double;
  OptBaseline                :Integer;
  OptHistBinCount            :Integer;

end;

TQUB_Task_MPL  = class(TQUB_Task_Data)
protected

  prOptName                :String;

  //Source Objects
  prModel                  :TQUB_QVO_QVM;

  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  prQuietOutput            :Boolean;

  //Mil
  prHost                   :String;

  prConfig                 :INode;
  prDataDescription        :INode;
  prDatasets               :Array of INode;
  prResultNode             :INode;
  prReportNode             :INode;
  prRealityCheckSaysGo     :Boolean;

  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prRealityCheck; // synchronize; sets prRealityCheckSaysGo := true to proceed
  procedure prPublishResults;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetModel(Model :TQUB_QVO_QVM);

  procedure prWriteCheckLine(Result :INode);
  procedure prDoCheckReport; // using prResultNode, synchronized

  function prReportCallback( arg :INode ): INode;
  procedure prDoReport;
  function prResultCallback( arg :INode ): INode;
  procedure prDoResult;
  function prPercentCallback( arg :INode ): INode;
  procedure prShowIdealization;


public

  FillDataCallback           :TNodeMethPtr;

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Stop; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model             :TQUB_QVO_QVM              read prModel             write prSetModel;

  //Parameters
  property Config            :INode                     read prConfig            write prConfig;


end;


function MPLFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; fillDataCB, rptCB, rsltCB, pctCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_mpltreeiface@24';


implementation


uses


QUB_Node_Model, QUB_Python_Data, QUB_QFS_QDF, QUB_Main;


//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _M P L                         //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_MPL.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'MPL optimizer';

  prConfig     := TNode.Create('Properties');
  prConfig['StopFlag'].SetData(0);

end;

destructor TQUB_Task_MPL.Destroy;
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_MPL.Stop;
begin

  prConfig['StopFlag'].Data[0] := 1;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MPL.prDoTask;
var
  i               :Integer;

  DataSetPtrs     :TPAPointer;
  fillDataCB      :INodeCFunc;
  rptCB, rsltCB   :INodeCFunc;
  pctCB           :INodeCFunc;
  ResultNode      :INode;
  ResultNodePtr   :TQUB_QFS_PNodeMem;
begin

  Synchronize( prRealityCheck );
  if not prRealityCheckSaysGo then Exit;

  GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * (Length(prDatasets) + 1) );
  for i := 0 to Length(prDataSets) - 1 do begin
    DataSetPtrs[i] := Pointer(prDataSets[i].GetNodeMem);
  end;
  DataSetPtrs[ Length(prDataSets) ] := nil;

  fillDataCB  := TNodeCFunc.Wrap( TNodeMeth.Create( FillDataCallback ) );
  rptCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prReportCallback ) );
  rsltCB      := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prResultCallback ) );
  pctCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prPercentCallback ) );

  prConfig['StopFlag'].SetData( 0 );

  try
    prConsoleMsg  := prOptName + ' optimizer running ... ';
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);

    ResultNodePtr := MPLFromDll( prConfig.GetNodeMem, DataSetPtrs, fillDataCB, rptCB, rsltCB, pctCB );
    if ResultNodePtr <> nil then begin
      ResultNode  := TNode.CreateUsing( ResultNodePtr );
      Node_DECREF( ResultNodePtr );
    end else begin
      ResultNode  := TNode.CreateNull;
    end;
    
    prResultNode := ResultNode;
    Synchronize( prPublishResults );

    if MILModeOfString(prConfig['Mode'].DataAsString) = MIL_MODE_EVAL then begin
      Synchronize(prDoCheckReport);
    end;

    for i := 0 to prChunks.Count - 1 do begin
      if TQUB_Thread_PChunk(prChunks[i]).Status = CHUNK_STATUS_EMPTY then begin
        TQUB_Thread_PChunk(prChunks[i]).Status := CHUNK_STATUS_DONE;
      end;
    end;
  except
    prError    := True;
    prErrorMsg := 'Unknown error in MPL library';
    prWikiTopic := 'Error:Task';
  end;

  FreeMem( DataSetPtrs );

end;

procedure TQUB_Task_MPL.prRealityCheck;
var
  i, r         :Integer;
  npoint       :Integer;
  Seg          :INode;
begin
  npoint       := 0;
  for i := 0 to Length(prDataSets) - 1 do begin
    Seg        := prDataSets[i].Find('Segment');
    while not Seg.IsNull do begin
      npoint   := npoint + (Seg.Data[1] - Seg.Data[0] + 1);
      Seg      := Seg.SiblingSameName;
    end;
  end;

  if npoint <= MPL_REASONABLE_NPOINTS then begin
    prRealityCheckSaysGo := True;
    Exit;
  end;

  r := MessageHelpDlg( Format('MPL: %d data points could take too much time and memory.  Continue anyway?', [npoint]),
                   mtConfirmation, [mbYes, mbNo], 0, 'Warning:Too_Much_Time_And_Memory');

  prRealityCheckSaysGo := (r = mrYes);
end;

procedure TQUB_Task_MPL.prPublishResults;
var
  ds, seg         :INode;
begin
  prResultNode.Name  := 'MPL ' + ChangeFileExt(ExtractFileName(Data.FileName), '');

  prResultNode['ResultType'].SetData( 'MPL' );
  if (prDataDescription <> nil) and not prDataDescription.IsNull then prResultNode.AppendChild( prDataDescription );

  ds              := prResultNode.Find('DataSet');
  while not ds.isNull do begin
    seg           := ds.Find('Segment');
    while not seg.isNull do begin
      seg.RemoveChild( seg['Classes'] );   // not helpful for Select
      seg.RemoveChild( seg['Durations'] );
      seg.RemoveChild( seg['Firsts'] );
      seg.RemoveChild( seg['Lasts'] );
      seg.RemoveChild( Seg['amp'] );       // not meaningful for Select
      seg.RemoveChild( Seg['sd'] );


      seg         := seg.SiblingSameName;
    end;
    ds            := ds.SiblingSameName;
  end;

  if prSelectionList <> nil then begin          // if the selection list is 'MPL Results',
    prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
    prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
  end;

  ViewportResults.AddResults( prResultNode, False );
end;

procedure TQUB_Task_MPL.prDoExecuteStart;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  inherited;

end;

procedure TQUB_Task_MPL.prDoExecuteEnd;
begin

  inherited;

  if prSelectionList <> nil then begin
    prSelectionList.UnLock;
  end;

  if Model <> nil then begin
    Model.UnlockAny;
  end;

end;


function TQUB_Task_MPL.prDoExecuteInitialize :Boolean;
var
  MdlNode       :INode;
begin
  Result        := False;

  MdlNode       := GetModelAsNode;
  MdlNode.SetData( 'Initial' );
  prConfig.AppendChild( MdlNode );

  prConfig['Histograms']['PDFs Each Iteration'].SetData( 0 );
  prQuietOutput := prConfig['QuietOutput'].Data[0];

end;

//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//

function TQUB_Task_MPL.prChunkData :Boolean;
var
  DataSets         :INode;
  DS               :INode;
  RemDS            :INode;
  Process          :INode;
begin
  Result           := False;

  Process          := prConfig.Find('ProcessData');

  if prScripting then begin
    DataSets         := DataNoderScript.GetChosenData(TNode.CreateNull);
  end else begin
  DataSets         := DataNoder.GetChosenData(TNode.CreateNull);
  end;
  prDataDescription := DataSets.Find('DataDescription');

  DS               := DataSets.Find('DataSet');
  while not DS.IsNull do begin
    if DS.Find('Segment').IsNull then begin
      RemDS          := DS;
      DS             := DS.SiblingSameName;
      DataSets.RemoveChild( RemDS );
    end else begin
      SetLength( prDataSets, Length(prDataSets) + 1 );
      prDataSets[ Length(prDataSets) - 1 ] := DS;

      if not Process.IsNull then begin
        DS.AppendChild( Process.Clone(True) );
      end;

      DS             := DS.SiblingSameName;
    end;
  end;
end;

procedure TQUB_Task_MPL.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MilChunk :TQUB_Thread_Mil_PChunk;
begin

  MilChunk := TQUB_Thread_Mil_PChunk(Chunk);

  if MilChunk.ClassDwells <> nil then begin
    FreeMem(MilChunk.ClassDwells);
    MilChunk.ClassDwells := nil;
  end;

  FreeMem(MilChunk);

end;

function TQUB_Task_MPL.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Mil_Chunk));

  Result.Sel := Sel;

end;



//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MPL.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_MPL_PropertiesForm do begin
    Self.ChannelIndex           := ChannelIndex;

    prConfig := Properties.Clone(True);

    //Data Processed
    Self.DataProcessed          := StringToDataProcessed(Properties['ProcessData'].DataAsString);
    Self.ProcessFilter          := 0<>Properties['ProcessData']['Filter'].Data[0];
    Self.ProcessFilterFreq      := Properties['ProcessData']['FilterFreq'].Data[0];
    Self.ProcessBaseline        := 0<>Properties['ProcessData']['Baseline'].Data[0];

    //Client/Server
    Self.ThreadCount            := Properties['ThreadCount'].Data[0];
    Self.Remote                 := 0<>Properties['Remote'].Data[0];

    prQuietOutput               := 0<>Properties['QuietOutput'].Data[0];

    prOptName                   := 'MPL';

  end;//with
end;

procedure TQUB_Task_MPL.OnPutProperties(Sender :TObject);
begin
end;


//******************************************************************************
procedure TQUB_Task_MPL.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then prModel.LockAny;
end;


//****************************************************************************//
//                                                                            //
//                          C A L L B A C K S                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Task_MPL.prReportCallback( arg :INode ): INode;
begin
  prReportNode := arg;
  Synchronize( prDoReport );
  Result := arg;
end;

procedure TQUB_Task_MPL.prDoReport;
begin
  // just print the node data's text rep'n for now
  if not prQuietOutput then begin
    prConsoleMsg := prReportNode.GetDataAsString;
    prConsoleCode := MESSAGE_ERROR;
    prWriteNewConsole;
  end;
end;

function TQUB_Task_MPL.prPercentCallback( arg :INode ): INode;
begin
  prUpdateStatus( Integer( arg.Data[0] ), prStTime );
  Result := arg;
end;

//****************************************************************************//
//                                                                            //
//                            R E S U L T S                                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MPL.prWriteCheckLine(Result :INode);
var
  Grad, Ix           :INode;
  Text               :String;
  i                  :Integer;
begin
  Text := '';

  Ix := Result.Find('Index');
  if not Ix.isNull then begin
    Text := Format('Segment %d | ', [Integer(Ix.Data[0])+1]);
  end;

  Text := Text + Format('LL: %.1f Grad:', [Double(Result['LL'].Data[0])]);
  Grad := Result['Gradient'];
  for i := 0 to Grad.DataCount - 1 do begin
    Text := Text + Format(' %f', [Double(Grad.Data[i])]);
  end;

  if not prQuietOutput then begin
    prConsoleMsg := Text;
    prConsoleCode := MESSAGE_MESSAGE;
    prWriteNewConsole;
  end;
end;

procedure TQUB_Task_MPL.prDoCheckReport;
var
  DataSet, Seg       :INode;
begin
  if MILSegModeOfString(prConfig['use segments'].DataAsString) = MIL_MODE_SEG_ALL then begin
    prWriteCheckLine(prResultNode);
  end else begin // segs separate
    DataSet := prResultNode.Find('DataSet');
    while not DataSet.isNull do begin
      Seg := DataSet.Find('Segment');
      while not Seg.isNull do begin
        prWriteCheckLine( Seg );
        Seg := Seg.SiblingSameName;
      end;
      DataSet := DataSet.SiblingSameName;
    end;
  end;
end;

function TQUB_Task_MPL.prResultCallback( arg :INode ): INode;
begin
  prResultNode := arg;
  Synchronize( prDoResult );
  Result := arg;
end;

procedure TQUB_Task_MPL.prDoResult;
var
  Final            :Boolean;
  Iter             :Integer;
  Grad             :INode;
  SegIx            :Integer;
  Text             :String;
  i                :Integer;
  BySegment        :Boolean;
  Rate             :INode;
begin
  SetRatesFromNode( prModel.ModelObject, prResultNode['ModelFile']['Rates'] );

  Text              := '';

  Final             := not prResultNode.Find('Final').IsNull;
  if Final then begin
    Text            := Text + 'Finished ';
  end;

  Iter              := prResultNode['Iterations'].Data[0];
  BySegment         := MILSegModeOfString(prConfig['use segments'].DataAsString) = MIL_MODE_SEG_EACH;

  if BySegment then begin
    SegIx := Integer(prResultNode['Index'].Data[0]);
    Text  := Text + Format('seg %d ', [SegIx + 1]);
  end;

  if not Final then begin
    Grad              := prResultNode['Gradient'];

    Text := Text + Format('iter: %d'#9'LL: %.4f'#9'grad:', [Iter, Double(prResultNode['LL'].Data[0])]);
    for i := 0 to Grad.DataCount - 1 do begin
      Text := Text + Format(#9'%.6g', [Double(Grad.Data[i])]);
    end;
  end;

  if Final or (Iter = 1) or not prQuietOutput then begin
    Text := Text + Format(#13#10'Rates:'#13#10'  s1'#9's2'#9'%-10s %-10s %-10s %-10s', ['k0','k0.std','k1','k1.std']);

    Rate := prResultNode['ModelFile']['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      Text := Text + Format(#13#10'  %d'#9'%d'#9'%-10.6g %-10.6g %-10.6g %-10.6g',
                [ Integer(Rate['States'].Data[0]) + 1, Integer(Rate['States'].Data[1]) + 1,
                  Double(Rate['k0'].Data[0]),          Double(Rate['dk0'].Data[0]),
                  Double(Rate['k1'].Data[0]),          Double(Rate['dk1'].Data[0]) ] );
      Text := Text + Format(#13#10'  %d'#9'%d'#9'%-10.6g %-10.6g %-10.6g %-10.6g',
                [ Integer(Rate['States'].Data[1]) + 1, Integer(Rate['States'].Data[0]) + 1,
                  Double(Rate['k0'].Data[1]),          Double(Rate['dk0'].Data[1]),
                  Double(Rate['k1'].Data[1]),          Double(Rate['dk1'].Data[1]) ] );
      Rate := Rate.SiblingSameName;
    end;
  end;

  //TODO: finish this, as in Mac/MIP/Staircase. LM
  if not prQuietOutput then begin
    prConsoleMsg := Text;
    prConsoleCode := MESSAGE_MESSAGE;
    prWriteNewConsole;
  end else if Final then begin
    Text := '';
    Rate := prResultNode['ModelFile']['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      Text := Text + Format('%-10.6g'#9'%-10.6g'#9'%-10.6g'#9'%-10.6g'#9,
                           [Double(Rate['k0'].Data[0]), Double(Rate['dk0'].Data[0]), Double(Rate['k0'].Data[1]), Double(Rate['dk0'].Data[1])]);
      Rate := Rate.SiblingSameName;
    end;
    Text := Text + Format('%-10.6g', [-Double(prResultNode['LL'].Data[0])]);
    prConsoleMsg := Text;
    prConsoleCode := MESSAGE_MESSAGE;
    prWriteNewConsole;
  end;

  if not prQuietOutput then begin
    prShowIdealization;
  end;

  //ResultText
  if (Final) and (prResultText <> nil) then begin
    Text := Format('%14.10g'#9'%d'#9, [-Double(prResultNode['LL'].Data[0]), Iter]);
    Rate := prResultNode['ModelFile']['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      Text := Text + Format('%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g'#9,
                            [
                             Double(Rate['k0'].Data[0]), Double(Rate['dk0'].Data[0]), Double(Rate['k1'].Data[0]), Double(Rate['dk1'].Data[0]),
                             Double(Rate['k0'].Data[1]), Double(Rate['dk0'].Data[1]), Double(Rate['k1'].Data[1]), Double(Rate['dk1'].Data[1])
                            ]);
      Rate := Rate.SiblingSameName;
    end;
    Grad := prResultNode['Gradient'];
    for i := 0 to Grad.DataCount - 1 do begin
      Text := Text + Format('%14.10g'#9, [Double(Grad.Data[i])]);
    end;
    prResultText^ := Text;
  end;

end;


procedure TQUB_Task_MPL.prShowIdealization;
var
  DataObject   :TQUB_QFS_QDF;
  DataSet      :INode;
  Seg          :INode;
begin

  prStatus := 'Applying Idealization';
  Synchronize(prDoOnChange);

  if MILSegModeOfString(prConfig['use segments'].DataAsString) <> MIL_MODE_SEG_ALL then begin
    DataObject := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[ prResultNode.Parent['FileName'].DataAsString ]).DataObject;
    DataObject.Idealization[ChannelIndex].SetDwellsSeg(prResultNode);
  end else begin
    DataSet := prResultNode.Find('DataSet');
    while not DataSet.IsNull do begin
      DataObject := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[ DataSet['FileName'].DataAsString ]).DataObject;
      Seg     := DataSet.Find('Segment');
      while not Seg.IsNull do begin
        DataObject.Idealization[ChannelIndex].SetDwellsSeg(Seg);
        Seg   := Seg.SiblingSameName;
      end;
      DataSet   := DataSet.SiblingSameName;
    end;
  end;
end;


//******************************************************************************
end.

