unit QUB_Task_MPL_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_QFS_Channels,
QUB_Task,
QUB_Task_Data,
QUB_Properties,
QUB_Task_Properties,
QUB_Task_Data_Properties,
QUB_Mil_Global,
QUB_Node,
QUB_DataList;


{$include 'QUB_Task_MPL_Properties.hct'}


const


//Mode
DEF_MILMODE                   = MIL_MODE_OPT;
DEF_SEGMODE                   = MIL_MODE_SEG_ALL;

DEF_MILQUIETOUTPUT            = 1;

DEF_MPLHISTBINCOUNT           = 50;
MIN_MPLHISTBINCOUNT           = 2;
MAX_MPLHISTBINCOUNT           = 500;

DEF_MPLMAXSTEP            :Double    = 0.1;


type


TQUB_Task_MPL_PropertiesForm = class(TQUB_Task_Data_PropertiesForm)
  Label4                          :TLabel;
  Label6                          :TLabel;
  Label7                          :TLabel;
  Label8                          :TLabel;

  panParams                       :TPanel;
  txtMILLLConv                    :TEdit;
  txtMILGradConv                  :TEdit;
  txtMILMaxIter                   :TEdit;
  Label18                         :TLabel;
  txtMILRestarts                  :TEdit;
  Label19                         :TLabel;
  txtMILMaxStep                   :TEdit;
  chkMILQuietOutput               :TCheckBox;

  Panel1                          :TPanel;
  chkMILModeOpt                   :TRadioButton;
  chkMILModeEval                  :TRadioButton;
  Label12                         :TLabel;
  txtMILSearchLimit               :TEdit;
  Label11                         :TLabel;
  Panel3                          :TPanel;
  Label16                         :TLabel;
  chkSegModeAll                   :TRadioButton;
  chkSegModeEach                  :TRadioButton;
  udMILHistBinCount               :TUpDown;
  txtMILHistBinCount              :TEdit;
  Label5                          :TLabel;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);


protected


  //Data Source
  prDataList                      :TQUB_DataList;


public



end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface, QUB_Main;


{$R *.DFM}


//******************************************************************************
//******************************************************************************
procedure TQUB_Task_MPL_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := 20;
  prDataList.Top        := 50;
  prDataList.Width      := 605;
  prDataList.Height     := 100;
  // prDataList.FileListAvail := False;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Histograms"]["BinCount"]',
                      'MIL Hist Bin Count', REGTYPE_INT, DEF_MILHISTBINCOUNT,
                      udMILHistBinCount,
                      txtMILHistBinCount,
                      AcceptIntBetween(MIN_MILHISTBINCOUNT, MAX_MILHISTBINCOUNT)));
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvLL"]',
                    'MIL LL Conv', REGTYPE_FLOAT, DEF_MILLLCONV,
                    txtMILLLConv,
                    AcceptFloatBetween(MIN_MILLLCONV, MAX_MILLLCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvGrad"]',
                    'MIL Grad Conv', REGTYPE_FLOAT, DEF_MILGRADCONV,
                    txtMILGradConv,
                    AcceptFloatBetween(MIN_MILGRADCONV, MAX_MILGRADCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxIterations"]',
                    'MIL Max Iter', REGTYPE_INT, DEF_MILMAXITER,
                    txtMILMaxIter,
                    AcceptIntBetween(MIN_MILMAXITER, MAX_MILMAXITER),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxRestarts"]',
                    'MIL Restarts', REGTYPE_INT, DEF_MILRESTARTS,
                    txtMILRestarts,
                    AcceptIntBetween(MIN_MILRESTARTS, MAX_MILRESTARTS),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxStep"]',
                    'MIL Max Step', REGTYPE_FLOAT, DEF_MILMAXSTEP,
                    txtMILMaxStep,
                    AcceptFloatBetween(MIN_MILMAXSTEP, MAX_MILMAXSTEP),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SearchLimit"]',
                    'MIL Search Limit', REGTYPE_FLOAT, DEF_MILSEARCHLIMIT,
                    txtMILSearchLimit,
                    AcceptFloatBetween(MIN_MILSEARCHLIMIT, MAX_MILSEARCHLIMIT),
                    '') );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Mode"]',
                          'MIL Run Mode', REGTYPE_INT, MILModeStringOf(DEF_MILMODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMILModeOpt,
                                       MILModeStringOf(MIL_MODE_OPT),
                                       Integer(MIL_MODE_OPT)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMILModeEval,
                                       MILModeStringOf(MIL_MODE_EVAL),
                                       Integer(MIL_MODE_EVAL)))
                       );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["use segments"]',
                          'MIL Seg Mode', REGTYPE_INT, MILSegModeStringOf(DEF_SEGMODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkSegModeEach,
                                       MILSegModeStringOf(MIL_MODE_SEG_EACH),
                                       Integer(MIL_MODE_SEG_EACH)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkSegModeAll,
                                       MILSegModeStringOf(MIL_MODE_SEG_ALL),
                                       Integer(MIL_MODE_SEG_ALL)))
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["QuietOutput"]',
                        'MIL Quiet Output', REGTYPE_BOOL, DEF_MILQUIETOUTPUT,
                        chkMILQuietOutput));

  RegKeyName            := 'Software\QUB\Task MPL';
  ProfileCat            := 'MPL';

  WikiTopic             := 'Modeling:MPL#Properties';

  {$include 'QUB_Task_MPL_Properties.htx'}
end;

procedure TQUB_Task_MPL_PropertiesForm.FormShow(Sender: TObject);
var
  Model                  :TQUB_QVO_QVM;
begin
  inherited;

  prDataList.ShowFiles;

  Model                  := g_CurrModel();
  if Model <> nil then begin
    prDataList.ShowModelVars( Model, False );
  end;

end;

procedure TQUB_Task_MPL_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  prDataList.HideAll;
end;



//******************************************************************************
end.

