unit QUB_Task_MSL_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_Global,
QUB_QFS_Channels,
QUB_Properties,
QUB_Task,
QUB_Task_Properties,
QUB_Node,
QUB_MIL_Global,
QUB_DataList;



type


TQUB_Task_MSL_PropertiesForm = class(TQUB_Task_PropertiesForm)
  Label6                          :TLabel;
  Label7                          :TLabel;
  Label8                          :TLabel;

  panParams                       :TPanel;
  txtMILLLConv                    :TEdit;
  txtMILGradConv                  :TEdit;
  txtMILMaxIter                   :TEdit;
  Label18                         :TLabel;
  txtMILRestarts                  :TEdit;
  Label19                         :TLabel;
  txtMILMaxStep                   :TEdit;
    chkUsePeq: TCheckBox;

  Panel1                          :TPanel;
  chkMILModeOpt                   :TRadioButton;
  chkMILModeEval                  :TRadioButton;
  Label12                         :TLabel;
  txtMILSearchLimit               :TEdit;
  Label11                         :TLabel;
  Panel3                          :TPanel;
  Label16                         :TLabel;
  chkSegModeAll                   :TRadioButton;
  chkSegModeEach                  :TRadioButton;
    Label1: TLabel;
    txtSimplexMaxIter: TEdit;
    Label2: TLabel;
    txtSimplexTol: TEdit;
    Label3: TLabel;
    txtSimplexStep: TEdit;
    chkShowIterations: TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);


protected

  prDataList                      :TQUB_DataList;

public


end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_MSL_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := panParams.Left;
  prDataList.Top        := 50;
  prDataList.Width      := panParams.Width;
  prDataList.Height     := 100;

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["UsePeq"]',
                        '', REGTYPE_NONE, 1,
                        chkUsePeq));
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Mode"]',
                          '', REGTYPE_NONE, MILModeStringOf(DEF_MILMODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMILModeOpt,
                                       MILModeStringOf(MIL_MODE_OPT),
                                       Integer(MIL_MODE_OPT)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMILModeEval,
                                       MILModeStringOf(MIL_MODE_EVAL),
                                       Integer(MIL_MODE_EVAL)))
                       );
  (*
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["use segments"]',
                          'MIL Seg Mode', REGTYPE_INT, MILSegModeStringOf(DEF_SEGMODE))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkSegModeEach,
                                       MILSegModeStringOf(MIL_MODE_SEG_EACH),
                                       Integer(MIL_MODE_SEG_EACH)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkSegModeAll,
                                       MILSegModeStringOf(MIL_MODE_SEG_ALL),
                                       Integer(MIL_MODE_SEG_ALL)))
                       );
  *)
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ShowIterations"]',
                        '', REGTYPE_NONE, 0,
                        chkShowIterations));
  AddEditProperty(TQUB_Property_Edit.Create('["Simplex"]["MaxIterations"]',
                    '', REGTYPE_NONE, 200,
                    txtSimplexMaxIter,
                    AcceptIntBetween(0, MAX_MILMAXITER),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Simplex"]["Tolerance"]',
                    '', REGTYPE_NONE, 1e-4,
                    txtSimplexTol,
                    AcceptFloatGreaterThan(0),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["Simplex"]["Step"]',
                    '', REGTYPE_NONE, 1.2e0,
                    txtSimplexStep,
                    AcceptFloatGreaterThan(0),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SearchLimit"]',
                    '', REGTYPE_NONE, DEF_MILSEARCHLIMIT,
                    txtMILSearchLimit,
                    AcceptFloatBetween(MIN_MILSEARCHLIMIT, MAX_MILSEARCHLIMIT),
                    '') );
  (*
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvLL"]',
                    '', REGTYPE_FLOAT, DEF_MILLLCONV,
                    txtMILLLConv,
                    AcceptFloatBetween(MIN_MILLLCONV, MAX_MILLLCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvGrad"]',
                    '', REGTYPE_FLOAT, DEF_MILGRADCONV,
                    txtMILGradConv,
                    AcceptFloatBetween(MIN_MILGRADCONV, MAX_MILGRADCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxIterations"]',
                    '', REGTYPE_INT, DEF_MILMAXITER,
                    txtMILMaxIter,
                    AcceptIntBetween(0, MAX_MILMAXITER),
                   '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxRestarts"]',
                    '', REGTYPE_NONE, DEF_MILRESTARTS,
                    txtMILRestarts,
                    AcceptIntBetween(MIN_MILRESTARTS, MAX_MILRESTARTS),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxStep"]',
                    '', REGTYPE_NONE, DEF_MILMAXSTEP,
                    txtMILMaxStep,
                    AcceptFloatBetween(MIN_MILMAXSTEP, MAX_MILMAXSTEP),
                    '') );
  *)
  RegKeyName            := 'Software\QUB\Task MSL';
  ProfileCat            := 'MSL';

  WikiTopic             := 'Modeling:MSL#Properties';
end;

procedure TQUB_Task_MSL_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  prDataList.ShowFiles;
  prDataList.ShowVar('DeadTime', 'ms', 0.05, False);

  if ViewportModel.CurrModel <> nil then begin
    prDataList.ShowModelVars(ViewportModel.CurrModel, True);
  end;

end;

procedure TQUB_Task_MSL_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  prDataList.HideAll;
end;

//******************************************************************************
end.

