unit QUB_Task_MSearch;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QVO_QVM, QUB_QVO_QVR,

QUB_QFS_QMF,

QUB_Viewport_Data,
QUB_Viewport_Histogram,

QUB_Idealization,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_MSearch_Properties,

QUB_MIL_Global,
QUB_MIL,

Chart, Series,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, QUB_Viewport_Results,
QUB_Results_Histograms,
QUB_GUIInterface;


type


TQUB_Task_MSearch  = class(TQUB_Task_Data)
protected

  prOptName                :String;

  //Source Objects
  prModel                  :TQUB_QVO_QVM;

  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  prOutput                 :Integer;

  //Mil
  prHost                   :String;

  prConfig                 :INode;
  prDataDescription        :INode;
  prStopFlag               :INode; // separate from config since msearch internally relinks it into the mil config
  prDatasets               :Array of INode;
  prResultNode             :INode;
  prReportNode             :INode;
  prHistResult             :TQUB_Results_HistGroupDisplay;
  prCurModelNumber         :Integer; // should be Int64 but the ez variant style converts it to Integer anyway
  prPrevResultNode         :INode; // different node means different trial and new histograms
  prFirstCurHistGroup      :Integer;


  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prPublishResults;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetModel(Model :TQUB_QVO_QVM);

  function prReportCallback( arg :INode ): INode;
  procedure prDoReport;
  function prResultCallback( arg :INode ): INode;
  procedure prDoResult;
  function prPercentCallback( arg :INode ): INode;
  function prFilterCallback( arg :INode ): INode;

public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Stop; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model             :TQUB_QVO_QVM              read prModel             write prSetModel;

  //Parameters
  property Output            :Integer                   read prOutput            write prOutput;


end;


procedure LoadMSearchModels; // make sure gModels is valid
function MSearchFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; models :TQUB_QFS_PNodeMem; rptCB, rsltCB, pctCB, filterCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_msearch_tree_iface@28';
procedure MSearchModelCount( model :TQUB_QFS_PNodeMem; models :TQUB_QFS_PNodeMem; allModelCount, noLoopCount :PInteger ); stdcall; external 'qubopt.dll' name '_msearch_modelCount_iface@16';


var


// the msearch model numbers (encoding the unique connection sets) (stored in msearch-models.qtr)
gModels                    :INode;


implementation


uses


QUB_Node_Model,
QUB_Python_Data,
QUB_Task_MIL,
PythonEngine,
QUB_Main;


//****************************************************************************//
//                                                                            //
//                     T Q U B _ T a s k _M S e a r c h                       //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_MSearch.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'Model Search';

end;

destructor TQUB_Task_MSearch.Destroy;
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_MSearch.Stop;
begin

  prStopFlag.Data[0] := 1;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MSearch.prDoTask;
var
  i               :Integer;

  DataSetPtrs     :TPAPointer;
  rptCB, rsltCB   :INodeCFunc;
  pctCB, fltrCB   :INodeCFunc;
  ResultNode      :INode;
  ResultNodePtr   :TQUB_QFS_PNodeMem;
begin
  if Length(prDataSets) = 0 then begin
    Exit;
  end;

  GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * (Length(prDatasets) + 1) );
  for i := 0 to Length(prDataSets) - 1 do begin
    DataSetPtrs[i] := Pointer(prDataSets[i].GetNodeMem);
  end;
  DataSetPtrs[ Length(prDataSets) ] := nil;

  pctCB  := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prPercentCallback ) );

  if prConfig.Find('ModelFilter').DataCount > 0 then begin
    fltrCB := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prFilterCallback ) );
  end else begin
    fltrCB := nil;
  end;

  if Output <> MSEARCH_OUTPUT_FINAL then begin
    rptCB  := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prReportCallback ) );
    rsltCB := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prResultCallback ) );
  end else begin
    rptCB  := nil;
    rsltCB := nil;

    if prSelectionList <> nil then begin
      prSelectionList.UnLock;
      prSelectionList := nil;
    end;

    if Model <> nil then begin
      Model.UnlockAny;
      Model := nil;
    end;
  end;

  prStopFlag := prConfig['StopFlag'];
  prStopFlag.SetData( 0 );

  if gModels.isNull then begin  // still?  there must be a problem with the file.
    prError       := True;
    prErrorMsg    := 'Missing MSearch model numbers';
    prWikiTopic   := 'Error:Missing_MSearch_model_numbers';
  end else begin

    try
      prConsoleMsg  := 'Model Search running ... ';
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);

      if Output = MSEARCH_OUTPUT_EACHMODEL then begin
        prConsoleMsg := Format('%8s%12s%8s', ['Model', 'LL   ', 'npar']);
        prConsoleCode := MESSAGE_MESSAGE;
        Synchronize(prWriteNewConsole);
      end;

      if Output <> MSEARCH_OUTPUT_FINAL then begin
        prResultNode  := TNode.Create('');
        Synchronize( prPublishResults );
        prHistResult  := ViewportResults.CurrResults.Histograms;
      end;

      //if prConfig['Method'].DataAsString = 'MIP' then begin
      //  try
      //    PythonThread.BeginPython; // calls MIP via python, this is easiest way to have GIL
      //    ResultNodePtr := MSearchFromDll( prConfig.GetNodeMem, DataSetPtrs, gModels.GetNodeMem, rptCB, rsltCB, pctCB, fltrCB );
      //  finally
      //    PythonThread.EndPython;
      //  end;
      //end else begin
        ResultNodePtr := MSearchFromDll( prConfig.GetNodeMem, DataSetPtrs, gModels.GetNodeMem, rptCB, rsltCB, pctCB, fltrCB );
      //end;
      ResultNode    := TNode.CreateUsing( ResultNodePtr );
      if ResultNodePtr <> nil then Node_DECREF( ResultNodePtr );
      ResultNode['Hold'].SetData( 1 );

      prResultNode := ResultNode;
      Synchronize( prPublishResults );

    except
      prError    := True;
      prErrorMsg := 'Unknown error in MSearch library';
      prWikiTopic := 'Error:Task';
    end;

  end;

  FreeMem( DataSetPtrs );

end;

procedure TQUB_Task_MSearch.prPublishResults;
begin
  prResultNode.Name  := 'MSearch ' + ChangeFileExt(ExtractFileName(Data.FileName), '');
  prResultNode['Hold'].SetData( 0 );
  
  prResultNode['ResultType'].SetData( 'MSearch' );
  if (prDataDescription <> nil) and not prDataDescription.IsNull then prResultNode.AppendChild( prDataDescription );

  if prSelectionList <> nil then begin          // if the selection list is 'MSearch Results',
    prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
    prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
  end;

  prResultNode['Date'].SetData( Format('%s %s',[DateTimeToStr( Date ),TimeToStr( Time )] ) );
  prResultNode.SaveAs(ChangeFileExt(ChangeFileExt(Data.FileName, '')+' MSearch', '.qrf'));
  prResultNode.Close;
  prHistResult    := ViewportResults.AddResults( prResultNode, False ).Histograms;
end;

procedure TQUB_Task_MSearch.prDoExecuteStart;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  inherited;

end;

procedure TQUB_Task_MSearch.prDoExecuteEnd;
begin

  inherited;

  if prSelectionList <> nil then begin
    prSelectionList.UnLock;
  end;

  if Model <> nil then begin
    Model.UnlockAny;
  end;

end;


function TQUB_Task_MSearch.prDoExecuteInitialize :Boolean;
var
  MdlNode       :INode;
begin
  Result        := False;

  MdlNode       := GetModelAsNode;
  MdlNode.SetData( 'Initial' );
  prConfig.AppendChild( MdlNode );

  // make sure the model lists are ready
  LoadMSearchModels;

  prCurModelNumber := 0; // the actual numbers are > 0, so this will force it to redo the rates in the first result callback
  prPrevResultNode := TNode.CreateNull;
end;

//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_MSearch.prChunkData :Boolean;
var
  DataSets         :INode;
  DS               :INode;
  RemDS            :INode;
begin
  Result           := False;

  DataSets         := DataNoder.GetChosenIdealization(TNode.CreateNull);
  prDataDescription := DataSets.Find('DataDescription');

  DS               := DataSets.Find('DataSet');
  while not DS.IsNull do begin
    if 0<>prConfig['join segments'].Data[0] then DS['Join Segments'];

    if DS.Find('Segment').IsNull then begin
      RemDS          := DS;
      DS             := DS.SiblingSameName;
      DataSets.RemoveChild( RemDS );
    end else begin
      SetLength( prDataSets, Length(prDataSets) + 1 );
      prDataSets[ Length(prDataSets) - 1 ] := DS;

      DS             := DS.SiblingSameName;
    end;
  end;
end;


procedure TQUB_Task_MSearch.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MilChunk :TQUB_Thread_Mil_PChunk;
begin

  MilChunk := TQUB_Thread_Mil_PChunk(Chunk);

  if MilChunk.ClassDwells <> nil then begin
    FreeMem(MilChunk.ClassDwells);
    MilChunk.ClassDwells := nil;
  end;

  FreeMem(MilChunk);

end;

function TQUB_Task_MSearch.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Mil_Chunk));

  Result.Sel := Sel;

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MSearch.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_MSearch_PropertiesForm do begin

    prConfig := Properties.Clone(True);

    Self.Output         := MSearchOutputOfString(prConfig['Output'].DataAsString);

  end;//with
end;

procedure TQUB_Task_MSearch.OnPutProperties(Sender :TObject);
begin
end;


//******************************************************************************
procedure TQUB_Task_MSearch.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then prModel.LockAny;
end;



//****************************************************************************//
//                                                                            //
//                            R E S U L T S                                   //
//                                                                            //
//****************************************************************************//
function TQUB_Task_MSearch.prReportCallback( arg :INode ): INode;
begin
  prReportNode := arg;
  Synchronize( prDoReport );
  Result := arg;
end;

procedure TQUB_Task_MSearch.prDoReport;
begin
  // just print the node data's text rep'n for now
  prConsoleMsg := prReportNode.GetDataAsString;
  prConsoleCode := MESSAGE_ERROR;
  prWriteNewConsole;
end;

function TQUB_Task_MSearch.prPercentCallback( arg :INode ): INode;
begin
  prUpdateStatus( Integer( arg.Data[0] ), prStTime );
  Result := arg;
end;

function TQUB_Task_MSearch.prFilterCallback( arg :INode ): INode;
var
  FilterFunc         :PPyObject;
  FilterArgs         :PPyObject;
  FilterResult       :PPyObject;
begin
  Result := QUB_Node_Null;
  FilterFunc         := nil;
  FilterArgs         := nil;
  FilterResult       := nil;

  PythonThread.BeginPython;
  try
    with GetPythonEngine do begin
      FilterFunc       := EvalString('Commands.GetByName("ModelFilters", "'+prConfig['ModelFilter'].DataAsString+'")');
      FilterArgs       := Py_BuildValue('(O)', [PyNode_FromNode( arg )]);
      FilterResult     := PyEval_CallObject(FilterFunc, FilterArgs);

      if FilterResult = nil then begin
        Result := TNode.CreateNull;
        PyErr_Print;
      end else if PyNode_IsNode( FilterResult ) then begin
        Result := PyNode_ToNode( FilterResult );
      end;
    end;
  except
  end;
  if FilterFunc <> nil then GetPythonEngine.Py_DECREF(FilterFunc);
  if FilterArgs <> nil then GetPythonEngine.Py_DECREF(FilterArgs);
  if FilterResult <> nil then GetPythonEngine.Py_DECREF(FilterResult);
  PythonThread.EndPython;
end;

function TQUB_Task_MSearch.prResultCallback( arg :INode ): INode;
begin
  prResultNode := arg;
  Synchronize( prDoResult );
  Result := arg;
end;

procedure TQUB_Task_MSearch.prDoResult;
var
  ModelNumber      :Integer;
  Final            :Boolean;
  Iter             :Integer;
  FreshTrial       :Boolean;
  DwellCount       :Integer;
  Grad             :INode;
  DataSet, Seg     :INode;
  DataSetIx        :Integer;
  Text             :String;
  i                :Integer;
  Rate             :INode;
  FirstTC          :INode;
  TC, Tau, Amp     :INode;
  HistGroup        :TQUB_Results_HistGroup;
  HistGroupNode    :INode;
begin
  FreshTrial       := prResultNode.GetNodeMem <> prPrevResultNode.GetNodeMem;
  prPrevResultNode := prResultNode;

  Final             := not prResultNode.Find('Final').IsNull;
  DwellCount        := 0;

  Text              := '';

  ModelNumber      := Integer(prResultNode['ModelNumber'].Data[0]);
  if ModelNumber = prCurModelNumber then begin
    SetRatesFromNode( prModel.ModelObject, prResultNode['ModelFile']['Rates'] );
  end else begin
    SetModelFromNode( prModel.ModelObject, prResultNode['ModelFile'] );
    prCurModelNumber := ModelNumber;
    if Output = MSEARCH_OUTPUT_EACHITER then begin
      Text             := Text + Format( 'Model %d'#13#10, [ModelNumber] );
    end;
  end;

  if FreshTrial then prFirstCurHistGroup := prHistResult.GroupCount; // ViewportHistogram.Optimize.Count;

  DataSetIx := 0;
  DataSet := prResultNode.Find('DataSet');
  while not DataSet.IsNull do begin
    Seg := DataSet.Find('Segment');
    while not Seg.IsNull do begin
      DwellCount := DwellCount + Seg['DwellCount'].Data[0];
      Seg := Seg.SiblingSameName;
    end;

    HistGroupNode  := DataSet['HistogramGroup'];
    HistGroup      := prHistResult.Groups[ prFirstCurHistGroup + DataSetIx ];
    if HistGroup <> nil then begin
      HistGroup.Node   := HistGroupNode;
    end else begin
      prHistResult.AddGroup( HistGroupNode );
    end;
    prHistResult.CurrIndex := prFirstCurHistGroup + DataSetIx;

    Inc(DataSetIx);
    DataSet := DataSet.SiblingSameName;
  end;

  if Output = MSEARCH_OUTPUT_EACHMODEL then begin
    try
      Text             := Text + Format('%8d%12f%8d', [ModelNumber, Double(prResultNode['LL'].Data[0]), Integer(prResultNode['ParameterCount'].Data[0])] );
    except
    end;
  end else begin  // MSEARCH_OUTPUT_EACHITER (there's no result callback for MSEARCH_OUTPUT_FINAL)

    if Final then begin
      Text            := Text + 'Finished ';
    end;

    Iter              := prResultNode['Iterations'].Data[0];

    if Final then begin
      Text := Text + Format('[%d events] ', [DwellCount]);
    end else begin
      Grad              := prResultNode['Gradient'];

      Text := Text + Format('iter: %d'#9'LL: %.4f'#9'grad:', [Iter, Double(prResultNode['LL'].Data[0])]);
      for i := 0 to Grad.DataCount - 1 do begin
        Text := Text + Format(#9'%.6g', [Double(Grad.Data[i])]);
      end;
    end;

    if Final or (Iter = 1) then begin
      Text := Text + Format(#13#10'Rates:'#13#10'  s1'#9's2'#9'%-10s %-10s %-10s %-10s', ['k0','k0.std','k1','k1.std']);

      Rate := prResultNode['ModelFile']['Rates'].Find('Rate');
      while not Rate.IsNull do begin
        Text := Text + Format(#13#10'  %d'#9'%d'#9'%-10.6g %-10.6g %-10.6g %-10.6g',
                  [ Integer(Rate['States'].Data[0]) + 1, Integer(Rate['States'].Data[1]) + 1,
                    Double(Rate['k0'].Data[0]),          Double(Rate['dk0'].Data[0]),
                    Double(Rate['k1'].Data[0]),          Double(Rate['dk1'].Data[0]) ] );
        Text := Text + Format(#13#10'  %d'#9'%d'#9'%-10.6g %-10.6g %-10.6g %-10.6g',
                  [ Integer(Rate['States'].Data[1]) + 1, Integer(Rate['States'].Data[0]) + 1,
                    Double(Rate['k0'].Data[1]),          Double(Rate['dk0'].Data[1]),
                    Double(Rate['k1'].Data[1]),          Double(Rate['dk1'].Data[1]) ] );
        Rate := Rate.SiblingSameName;
      end;
    end;

    if Final then begin
      FirstTC := prResultNode['DataSet'].Find('TimeConstants');

      Text := Text + #13#10'Time Constants (Tau, Amp):';

      TC        := FirstTC;
      while not TC.IsNull do begin
        Text := Text + Format(#13#10'  Class %d:', [Integer(TC['Class'].Data[0])]);
        Tau := TC['Tau'];
        Amp := TC['Amp'];
        for i := 0 to Tau.DataCount - 1 do begin
          Text := Text + Format(' (%.4f, %.4f)', [Double(Tau.Data[i]), Double(Amp.Data[i])]);
        end;

        TC := TC.SiblingSameName;
      end;

      if not FirstTC.IsNull then begin
        SolveTcrit( FirstTC );
        TC := FirstTC.Find('Tcrit');  // now TC means TCrit
        if (not TC.IsNull) and (TC.DataCount > 0) then begin
          Text := Text + #13#10'Tcrit: ';
          for i := 0 to TC.DataCount - 1 do begin
            if TC.Data[i] = 0.0 then Break;

            Text := Text + Format('%12.4g', [Double(TC.Data[i])]);
          end;
          prResultNode.AppendChild( TC.Clone(True) );
        end;
      end;
    end;

    if Final then begin
      try
        Text  := Text + Format(#13#10'LL: %f'#9'Events: %d', [Double(prResultNode['LL'].Data[0]), DwellCount]);
      except
      end;
    end;

  end;

  prConsoleMsg := Text;
  prConsoleCode := MESSAGE_MESSAGE;
  prWriteNewConsole;
end;

procedure LoadMSearchModels;
begin
  with GetPythonEngine do begin
    PythonThread.BeginPython;
    if (gModels = nil) or gModels.isNull then begin
       gModels := TNode.OpenReadOnly( ExtractFilePath(Application.ExeName) + 'msearch-models.qtr' );
       gModels.Close;
       PyNode_Publish( gModels, 'msearchModels' );
    end;

    PythonThread.EndPython;
    PythonThread.DonePython;
  end;
end;

//******************************************************************************
initialization
  gModels               := nil;

//******************************************************************************
end.

