unit QUB_Task_MSearch_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_QFS_Channels,

QUB_Profiles,
QUB_Properties,
QUB_Task,
QUB_Task_Properties,
QUB_Mil_Global,
QUB_Node,
QUB_DataList;


function MSearchOutputStringOf(x :Integer): String;
function MSearchOutputOfString(x :String): Integer;


const


DEF_MILJOINSEGMENTS           = 0;

MSEARCH_OUTPUT_EACHITER       = 0;
MSEARCH_OUTPUT_EACHMODEL      = 1;
MSEARCH_OUTPUT_FINAL          = 2;
DEF_OUTPUT                    = 2; // MSEARCH_OUTPUT_FINAL;

DEF_MS_TRIALS                 = 1;
MIN_MS_TRIALS                 = 1;
MAX_MS_TRIALS                 = 10000;

DEF_MS_NOLOOPS                = 0;
DEF_MS_BALANCELOOPS           = 1;
DEF_MS_SHOWITERATIONS         = 1;
DEF_MS_RANDOMIZE              = 1;

DEF_MS_INITIALK0       :Double       = 100.0;
MIN_MS_INITIALK0       :Double       = 1e-10;
MAX_MS_INITIALK0       :Double       = 1e10;

DEF_MS_INITIALK1       :Double       = 0.0;
MIN_MS_INITIALK1       :Double       = 0.0;
MAX_MS_INITIALK1       :Double       = 1e10;

DEF_MS_MAXRATES               = 20;
MIN_MS_MAXRATES               = 2;
MAX_MS_MAXRATES               = 10000;

DEF_MS_MAXPERSTATE            = 5;
MIN_MS_MAXPERSTATE            = 2;
MAX_MS_MAXPERSTATE            = 10000;

DEF_MS_RANDOMFROM      :Double       = 0.1;
MIN_MS_RANDOMFROM      :Double       = 1e-10;
MAX_MS_RANDOMFROM      :Double       = 1;

DEF_MS_RANDOMTO        :Double       = 10.0;
MIN_MS_RANDOMTO        :Double       = 1.0;
MAX_MS_RANDOMTO        :Double       = 1e10;

DEF_KEEP_TOP_N                = 1;
DEF_TOP_N                     = 20;

MSEARCH_OPT_MIL = 0;
MSEARCH_OPT_MIP = 1;
DEF_OPT = 0;


type


TQUB_Task_MSearch_PropertiesForm = class(TQUB_Task_PropertiesForm)
  lblModelCount                   :TLabel;
  tabsParams                      :TPageControl;
  tabMILParams                    :TTabSheet;
  panParams                       :TPanel;
  Label5                          :TLabel;
  Label6                          :TLabel;
  Label7                          :TLabel;
  Label8                          :TLabel;
  Label12                         :TLabel;
  Label18                         :TLabel;
  Label19                         :TLabel;
  txtMILHistBinCount              :TEdit;
  udMILHistBinCount               :TUpDown;
  chkMILHistPerIter               :TCheckBox;
  txtMILLLConv                    :TEdit;
  txtMILGradConv                  :TEdit;
  txtMILMaxIter                   :TEdit;
  txtMILSearchLimit               :TEdit;
  chkMILJoinSegments              :TCheckBox;
  txtMILRestarts                  :TEdit;
  txtMILMaxStep                   :TEdit;
  tabSearchParams                 :TTabSheet;
  Panel1                          :TPanel;
  Label20                         :TLabel;
  Label2                          :TLabel;
  Label11                         :TLabel;
  Label16                         :TLabel;
  Label17                         :TLabel;
  Label21                         :TLabel;
  Label22                         :TLabel;
  Label23                         :TLabel;
  chkNoLoops                      :TCheckBox;
  txtInitialK1                    :TEdit;
  txtInitialK0                    :TEdit;
  txtTrials                       :TEdit;
  chkRandom                       :TCheckBox;
  txtRandomFrom                   :TEdit;
  txtRandomTo                     :TEdit;
  txtMaxRates                     :TEdit;
  txtMaxPerState                  :TEdit;
  Label4                          :TLabel;
  cmbModelFilter                  :TComboBox;
  chkBalanceLoops                 :TCheckBox;
  grpOutput                       :TGroupBox;
  chkOutputIter                   :TRadioButton;
  chkOutputModel                  :TRadioButton;
  chkOutputFinal                  :TRadioButton;
    chkKeepTopN: TCheckBox;
    txtTopN: TEdit;
    Label1: TLabel;
    Panel2: TPanel;
    Label3: TLabel;
    chkOptMIL: TRadioButton;
    chkOptMIP: TRadioButton;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure cmbModelFilterChange(Sender: TObject);
  procedure cmbModelFilterDropDown(Sender: TObject);


protected


  prDataList                      :TQUB_DataList;

  prModelFilter                   :String;

  procedure prSetModelFilter(x :String);

  //Properties vs. Registry
  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public

  property ModelFilter                    :String                    read prModelFilter              write prSetModelFilter;

end;


implementation


uses

QUB_Task_MSearch,
QUB_Node_Model,
QUB_PyCmd_MenuButton,
QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface, QUB_Main;


{$R *.DFM}


//******************************************************************************
function MSearchOutputStringOf(x :Integer): String;
begin
  if x = MSEARCH_OUTPUT_EACHITER then begin
    Result                := 'Iteration';
  end else if x = MSEARCH_OUTPUT_EACHMODEL then begin
    Result                := 'Model';
  end else begin // if Output = MSEARCH_OUTPUT_FINAL then begin
    Result                := 'Final';
  end;
end;

function MSearchOutputOfString(x :String): Integer;
begin
  if x = 'Iteration' then begin
    Result                := MSEARCH_OUTPUT_EACHITER;
  end else if x = 'Model' then begin
    Result                := MSEARCH_OUTPUT_EACHMODEL;
  end else{if x = 'Final' then}begin
    Result                := MSEARCH_OUTPUT_FINAL;
  end;
end;

function MSearchOptStringOf(x :Integer): String;
begin
  if x = MSEARCH_OPT_MIL then begin
    Result := 'MIL';
  end else begin
    Result := 'MIP';
  end;
end;

function MSearchOptOfString(x :String): Integer;
begin
  if x = 'MIL' then begin
    Result := MSEARCH_OPT_MIL;
  end else begin
    Result := MSEARCH_OPT_MIP;
  end;
end;


//******************************************************************************
procedure TQUB_Task_MSearch_PropertiesForm.prSetModelFilter(x :String);
begin
  prModelFilter := x;
  cmbModelFilter.ItemIndex := cmbModelFilter.Items.IndexOf(x);
end;


//******************************************************************************
procedure TQUB_Task_MSearch_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prDataList            := TQUB_DataList.Create(Self);
  prDataList.Parent     := Self;
  prDataList.Left       := 20;
  prDataList.Top        := 50;
  prDataList.Width      := 560;
  prDataList.Height     := 100;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Histograms"]["BinCount"]',
                      'MIL Hist Bin Count', REGTYPE_INT, DEF_MILHISTBINCOUNT,
                      udMILHistBinCount,
                      txtMILHistBinCount,
                      AcceptIntBetween(MIN_MILHISTBINCOUNT, MAX_MILHISTBINCOUNT)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Histograms"]["PDFs Each Iteration"]',
                        'MIL Hist Per Iter', REGTYPE_BOOL, DEF_MILJOINSEGMENTS,
                        chkMILHistPerIter));
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvLL"]',
                    'MIL LL Conv', REGTYPE_FLOAT, DEF_MILLLCONV,
                    txtMILLLConv,
                    AcceptFloatBetween(MIN_MILLLCONV, MAX_MILLLCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["ConvGrad"]',
                    'MIL Grad Conv', REGTYPE_FLOAT, DEF_MILGRADCONV,
                    txtMILGradConv,
                    AcceptFloatBetween(MIN_MILGRADCONV, MAX_MILGRADCONV),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxIterations"]',
                    'MIL Max Iter', REGTYPE_INT, DEF_MILMAXITER,
                    txtMILMaxIter,
                    AcceptIntBetween(MIN_MILMAXITER, MAX_MILMAXITER),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxRestarts"]',
                    'MIL Restarts', REGTYPE_INT, DEF_MILRESTARTS,
                    txtMILRestarts,
                    AcceptIntBetween(MIN_MILRESTARTS, MAX_MILRESTARTS),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["DFP"]["MaxStep"]',
                    'MIL Max Step', REGTYPE_FLOAT, DEF_MILMAXSTEP,
                    txtMILMaxStep,
                    AcceptFloatBetween(MIN_MILMAXSTEP, MAX_MILMAXSTEP),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SearchLimit"]',
                    'MIL Search Limit', REGTYPE_FLOAT, DEF_MILSEARCHLIMIT,
                    txtMILSearchLimit,
                    AcceptFloatBetween(MIN_MILSEARCHLIMIT, MAX_MILSEARCHLIMIT),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["join segments"]',
                        'MIL Join Segments', REGTYPE_BOOL, DEF_MILJOINSEGMENTS,
                        chkMILJoinSegments));

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Output"]',
                          'Output', REGTYPE_INT, MSearchOutputStringOf(DEF_OUTPUT))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOutputIter,
                                       MSearchOutputStringOf(MSEARCH_OUTPUT_EACHITER),
                                       MSEARCH_OUTPUT_EACHITER))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOutputModel,
                                       MSearchOutputStringOf(MSEARCH_OUTPUT_EACHMODEL),
                                       MSEARCH_OUTPUT_EACHMODEL))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOutputFinal,
                                       MSearchOutputStringOf(MSEARCH_OUTPUT_FINAL),
                                       MSEARCH_OUTPUT_FINAL))
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["TrialsPerModel"]',
                    'Trials', REGTYPE_INT, DEF_MS_TRIALS,
                    txtTrials,
                    AcceptIntGreaterThan(0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["NoLoops"]',
                        'NoLoops', REGTYPE_BOOL, DEF_MS_NOLOOPS,
                        chkNoLoops));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["BalanceLoops"]',
                        'BalanceLoops', REGTYPE_BOOL, DEF_MS_BALANCELOOPS,
                        chkBalanceLoops));
  AddEditProperty(TQUB_Property_Edit.Create('["InitialK0"]',
                    'InitialK0', REGTYPE_FLOAT, DEF_MS_INITIALK0,
                    txtInitialK0,
                    AcceptFloatBetween(MIN_MS_INITIALK0, MAX_MS_INITIALK0),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["InitialK1"]',
                    'InitialK1', REGTYPE_FLOAT, DEF_MS_INITIALK1,
                    txtInitialK1,
                    AcceptFloatBetween(MIN_MS_INITIALK1, MAX_MS_INITIALK1),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxRates"]',
                    'MaxRates', REGTYPE_INT, DEF_MS_MAXRATES,
                    txtMaxRates,
                    AcceptIntGreaterThan(0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MaxExitRatesPerState"]',
                    'MaxPerState', REGTYPE_INT, DEF_MS_MAXPERSTATE,
                    txtMaxPerState,
                    AcceptIntGreaterThan(0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["RandomRates"]',
                        'Randomize', REGTYPE_BOOL, DEF_MS_RANDOMIZE,
                        chkRandom));
  AddEditProperty(TQUB_Property_Edit.Create('["RandomRates"]["From"]',
                    'RandomFrom', REGTYPE_FLOAT, DEF_MS_RANDOMFROM,
                    txtRandomFrom,
                    AcceptFloatBetween(MIN_MS_RANDOMFROM, MAX_MS_RANDOMFROM),
                    '%.6g') );
  AddEditProperty(TQUB_Property_Edit.Create('["RandomRates"]["To"]',
                    'RandomTo', REGTYPE_FLOAT, DEF_MS_RANDOMTO,
                    txtRandomTo,
                    AcceptFloatBetween(MIN_MS_RANDOMTO, MAX_MS_RANDOMTO),
                    '%.6g') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["KeepTopN"]',
                        '', REGTYPE_NONE, DEF_KEEP_TOP_N,
                        chkKeepTopN));
  AddEditProperty(TQUB_Property_Edit.Create('["TopN"]',
                    '', REGTYPE_NONE, DEF_TOP_N,
                    txtTopN,
                    AcceptIntGreaterThan(0),
                    '') );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Method"]',
                          '', REGTYPE_NONE, MSearchOptStringOf(DEF_OPT))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptMIL,
                                       MSearchOptStringOf(MSEARCH_OPT_MIL),
                                       MSEARCH_OPT_MIL))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptMIP,
                                       MSearchOptStringOf(MSEARCH_OPT_MIP),
                                       MSEARCH_OPT_MIP))
                       );

  RegKeyName            := 'Software\QUB\Task MSearch';
  ProfileCat            := 'msrch';

  WikiTopic             := 'Modeling:Model_Search#Properties';

end;

procedure TQUB_Task_MSearch_PropertiesForm.FormShow(Sender: TObject);
var
  allModelCount, noLoopCount :Integer;
  Model                  :TQUB_QVO_QVM;
begin
  inherited;

  LoadMSearchModels;
  MSearchModelCount( GetModelAsNode.GetNodeMem, gModels.getNodeMem, @allModelCount, @noLoopCount );
  lblModelCount.Caption := Format( '%d models (%d without loops)', [allModelCount, noLoopCount] );

  prDataList.ShowFiles;

  prDataList.ShowVar('DeadTime', 'ms', 0.05, False);

  Model                  := ViewportModel.CurrModel;
  if Model <> nil then begin
    prDataList.ShowModelVars( Model, False );
  end;

end;

procedure TQUB_Task_MSearch_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;
  
  prDataList.HideAll;
end;


//******************************************************************************
procedure TQUB_Task_MSearch_PropertiesForm.cmbModelFilterChange(
  Sender: TObject);
begin
  inherited;
  if cmbModelFilter.Text = '<none>' then begin
    prModelFilter := '';
  end else begin
    prModelFilter := cmbModelFilter.Text;
  end;
  Properties['ModelFilter'].SetData(prModelFilter);
end;

procedure TQUB_Task_MSearch_PropertiesForm.cmbModelFilterDropDown(
  Sender: TObject);
begin
  inherited;
  cmbModelFilter.Items.Clear;
  cmbModelFilter.Items.Add('<none>');
  PyCmd_FillMenu(cmbModelFilter, 'ModelFilters');
  cmbModelFilter.ItemIndex := cmbModelFilter.Items.IndexOf(prModelFilter);
  if cmbModelFilter.ItemIndex < 0 then cmbModelFilter.ItemIndex := 0;
end;


//******************************************************************************


procedure TQUB_Task_MSearch_PropertiesForm.prSetInitialProperties(Props :INode);
begin
  inherited;

  prModelFilter := Props['ModelFilter'].DataAsString;
end;

procedure TQUB_Task_MSearch_PropertiesForm.prUpdateComponents;
begin
  inherited;

  cmbModelFilter.Items.Clear;
  cmbModelFilter.Items.Add(Properties['ModelFilter'].DataAsString);
  cmbModelFilter.ItemIndex := 0;
end;


//******************************************************************************
end.

