unit QUB_Task_MacLiveFit;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math, SyncObjs, Controls,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,

QUB_Viewport_Data, QUB_Viewport_Model,

QUB_Thread,
QUB_Task,

QUB_Task_Mac,
QUB_Task_Mac_Properties,

QUB_QVO_QVM, QUB_QFS_QMF, QUB_QFS_Channels,

QUB_Node,
QUB_GUIInterface;


type


TQUB_Task_MacLiveFit  = class(TQUB_Task)
protected

  prNeedsUpdate         :Boolean;
  prSignal              :TEvent;
  prMacIsRunning        :Boolean; // other mac from user
  prUpdating            :Boolean; // our mac
  prMac                 :TQUB_Task_Mac;

  prViewportData        :TQUB_Viewport_Data;
  prViewportModel       :TQUB_Viewport_Model;
  prModel               :TQUB_QVO_QVM;

  procedure prClearEvents;

  procedure prOnMacDone(Sender :TObject);
  procedure prOnChangeData(Sender :TObject);
  procedure prOnCloseModel(Sender :TObject);
  procedure prOnChangeModel(Sender :TObject);
  procedure prOnModelEdit(Sender :TObject);

  procedure prRunMac;
  function prMacRunning: Boolean;
  procedure prCheckMacRunning;


public


  MacProperties         :TQUB_Task_Mac_PropertiesForm;


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  procedure Execute; override;
  procedure Stop; override;
  procedure Update;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


end;


implementation


uses


QUB_Mac_Global;


//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _M a c L i v e F i t           //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_MacLiveFit.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'Mac Live Fit';

  prNeedsUpdate         := True;
  prSignal              := TEvent.Create(nil, False, False, '');
  prUpdating            := False;
  prMac                 := nil;

  prViewportData        := TQUB_Viewport_Data(ViewportList.ViewportByName['Data']);
  prViewportModel       := TQUB_Viewport_Model(ViewportList.ViewportByName['Model']);
  prModel               := prViewportModel.CurrModel;

  prViewportData.Events.AddEvent('OnChangeFile', prOnChangeData);
  prViewportModel.Events.AddEvent('OnChangeFile', prOnChangeModel);
  prViewportModel.Events.AddEvent('OnCloseFile', prOnCloseModel);
  if prModel <> nil then prModel.Events.AddEvent('OnEdit', prOnModelEdit);

end;

destructor TQUB_Task_MacLiveFit.Destroy;
begin
  Synchronize(prClearEvents);
  prSignal.Free;

  inherited;
end;

procedure TQUB_Task_MacLiveFit.prClearEvents;
begin
  prViewportData.Events.DelEvent('OnChangeFile', prOnChangeData);
  prViewportModel.Events.DelEvent('OnChangeFile', prOnChangeModel);
  prViewportModel.Events.DelEvent('OnCloseFile', prOnCloseModel);
  if prModel <> nil then begin
    prModel.Events.DelEvent('OnEdit', prOnModelEdit);
    prModel.AllowNumEdit := False;
  end;
end;

//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MacLiveFit.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_MacLiveFit.OnPutProperties(Sender :TObject);
begin
end;


//****************************************************************************//


procedure TQUB_Task_MacLiveFit.Execute;
begin
  while not prStopped do begin
    while prNeedsUpdate and not prMacRunning do begin
      prNeedsUpdate := False;
      Synchronize(prRunMac);
      while prUpdating do begin
        prSignal.WaitFor(1000);
      end;
    end;
    prSignal.WaitFor(1000);
  end;
end;

procedure TQUB_Task_MacLiveFit.prRunMac;
var
  Data      :TQUB_QVO_QVD;
begin

  Data := TQUB_QVO_QVD(prViewportData.CurrData);

  if Data = nil then Exit;

  if Data.GetChosenData.Find('Segment').IsNull then Exit;

  MacProperties.Properties['ChannelIndex'].Data[0]   := Data.ActiveChannel;
  MacProperties.lstFitData.Items.Assign(Data.SelectionLists.Items);

  if prModel = nil then Exit;

  prUpdating := True;

  prMac := TQUB_Task_Mac.Create(True{Suspended});
  prMac.OnTerminate    := prOnMacDone;

  MacProperties.OnGetProperties := prMac.OnGetProperties;
  MacProperties.OnPutProperties := prMac.OnPutProperties;

  MacProperties.FormShow(MacProperties);
  MacProperties.ModalResult := mrOK;
  MacProperties.btnOKClick(MacProperties);

  prMac.OnChange := TaskView.OnTaskChange;
  prMac.Data     := Data;
  prMac.Model    := prModel;

  // prModel.UnlockAny;
  prModel.AllowNumEdit := True;
  // mac has locked it but we want it to stay editable during this run -- mac will unlockany when it's done but that should be safe since QVO checks the lock count and resets it to 0 if it falls below.
  // ok so it is possible that e.g. deleting a rate during mlf causes crash
  // but for the live-wheel-fit thing it is necessary to change a rate while mlf is running
  // ?? what is mac/check reading in the model ??
  // ?? must anything be locked for the duration ??
  // ?? mac/check writing anything there ??

  prMac.OptPParams.OptMode    := MODE_SIM;//MODE_CHECK; Chris, John, what do you think?
  prMac.OptPParams.OptShowFit := 1;
  prMac.SilentOutput          := True;
  prMac.PublishResults        := False;

  prMac.Resume;

end;

procedure TQUB_Task_MacLiveFit.Stop;
begin
  inherited;
  if prMac <> nil then prMac.Stop;
  prSignal.SetEvent;
end;

procedure TQUB_Task_MacLiveFit.Update;
begin
  prNeedsUpdate := True;
  prSignal.SetEvent;
end;

procedure TQUB_Task_MacLiveFit.prOnMacDone(Sender :TObject);
begin
  if prModel <> nil then prModel.AllowNumEdit := False;
  prUpdating := False;
  prMac      := nil;
  prSignal.SetEvent;
end;

procedure TQUB_Task_MacLiveFit.prOnChangeData(Sender :TObject);
begin
  Update;
end;

procedure TQUB_Task_MacLiveFit.prOnCloseModel(Sender :TObject);
begin
  prModel := nil;
end;

procedure TQUB_Task_MacLiveFit.prOnChangeModel(Sender :TObject);
begin
  if prModel <> nil then begin
    prModel.Events.DelEvent('OnEdit', prOnModelEdit);
    prModel.AllowNumEdit := False;
  end;
  prModel := prViewportModel.CurrModel;
  if prModel <> nil then begin
    prModel.Events.AddEvent('OnEdit', prOnModelEdit);
    if prMac <> nil then prModel.AllowNumEdit := True;
    Update;
  end;
end;

procedure TQUB_Task_MacLiveFit.prOnModelEdit(Sender :TObject);
begin
  Update;
end;

function TQUB_Task_MacLiveFit.prMacRunning: Boolean;
begin
  Synchronize(prCheckMacRunning);
  Result := prMacIsRunning;
end;

procedure TQUB_Task_MacLiveFit.prCheckMacRunning;
begin
  prMacIsRunning := TaskView.HasTaskOfType(TQUB_Task_Mac);
end;


//******************************************************************************
end.

