unit QUB_Task_MdlRpt;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses

Windows, SysUtils, Classes, Clipbrd, Forms, Math,

AllocTypes, Alloc,

QUB_Global,

QUB_Thread,
QUB_Task,
QUB_Task_Data,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, QUB_Viewport_Results,
QUB_GUIInterface;


type

TQUB_Task_MdlRpt  = class(TQUB_Task_Data)
protected

  prProperties             :INode;
  prData                   :INode;

  prResults                :INode;


  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prPublishResults;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Stop; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  property Properties  :INode    read prProperties  write prProperties;
  property Data        :INode    read prData        write prData;

end;


implementation

uses
  QUB_Python_Data,
  QUB_Task_MIL,
  QUB_QVO_QVR,
  QUB_Reports_Editor,
  QUB_Main;



//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _M I L                         //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_MdlRpt.Create(Suspended :Boolean);
begin
  inherited;

  prName       := 'MdlRpt';

end;

destructor TQUB_Task_MdlRpt.Destroy;
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_MdlRpt.Stop;
begin

  prProperties['StopFlag'].Data[0] := 1;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MdlRpt.prDoTask;
var
  DataSetPtrs     :TPAPointer;
  ResultNodePtr   :TQUB_QFS_PNodeMem;
  ChanCount       :INode;
begin
  prProperties['StopFlag'].SetData( 0 );
  prProperties['Mode'].SetData('model report');

  GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * (1 + 1) );
  DataSetPtrs[0] := Pointer(prData.GetNodeMem);
  DataSetPtrs[1] := nil;

  try
    prConsoleMsg  := 'MdlRpt running ... ';
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);

    ChanCount     := prProperties['ModelFiles']['ChannelCount'];
    if (ChanCount.DataCount > 0) and (ChanCount.Data[0] > 7) then begin
      ChanCount.Data[0] := 1; // prevent mil hyper-model out-of-memory
    end;

    ResultNodePtr := MILFromDll( prProperties.GetNodeMem, DataSetPtrs, nil, nil, nil );
    prResults     := TNode.CreateUsing( ResultNodePtr );
    Node_DECREF( ResultNodePtr );

    Synchronize( prPublishResults );
  except
    prError    := True;
    prErrorMsg := 'Unknown error in MdlRpt library';
    prWikiTopic := 'Error:Task';
  end;

  FreeMem( DataSetPtrs );

end;

procedure TQUB_Task_MdlRpt.prPublishResults;
var
  Results         :TQUB_QVO_QVR;
  Report          :String;
  ReportsForm     :TQUB_Reports_EditForm;
  Task            :TQUB_Reports_Task_Run;
  Options         :TStringList;
begin
  if prResults.Find('ResultType').isNull then
    prResults['ResultType'].SetData( 'MdlRpt' );

  ViewportResults.AddResults( prResults, False );
  Results         := TQUB_QVO_QVR(ViewportResults.CurrResults);

  Report          := Properties['Report'].DataAsString;
  if Report <> 'none'then begin
    if Report = 'Custom...' then begin
      ReportsForm         := TQUB_Reports_EditForm.Create(nil);
      ReportsForm.Results := Results;
      ReportsForm.ReportName := 'untitled';
      ReportsForm.ShowModal;
      ReportsForm.Release;
    end else begin
      Options         := TStringList.Create;
      Options.LoadFromFile(ExtractFilePath(Application.ExeName) + 'Reports\MdlRpt\' + Report + '.txt');
      Task            := TQUB_Reports_Task_Run.Create(Results, Options);
      PythonThread.SyncPyTask(Task, True{Async -- auto-freed}, False{not Silent}, -1{no timeout});
    end;
  end;
end;

procedure TQUB_Task_MdlRpt.prDoExecuteStart;
begin

  inherited;

end;

procedure TQUB_Task_MdlRpt.prDoExecuteEnd;
begin

  inherited;

end;


function TQUB_Task_MdlRpt.prDoExecuteInitialize :Boolean;
begin
  Result        := False;
end;

//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_MdlRpt.prChunkData :Boolean;
begin
  Result           := False;
end;

procedure TQUB_Task_MdlRpt.prFreeChunk(Chunk :TQUB_Thread_PChunk);
begin

  FreeMem(Chunk);

end;

function TQUB_Task_MdlRpt.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Chunk));

  Result.Sel := Sel;

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_MdlRpt.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_MdlRpt.OnPutProperties(Sender :TObject);
begin
end;





// *****************************************************************************
end.
 
