unit QUB_Task_MdlRpt_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, FileCtrl, Registry, Math,

RegistryInterface,

QUB_Viewport_Model,
QUB_QVO,
QUB_QVO_QVM,
QUB_Properties,
QUB_Task_Properties,
QUB_Node,
PythonEngine,
QUB_Python_Thread, ComCtrls, ExtCtrls;


const


DEF_REPORT = 'word - both';

DEF_FIRSTLAT_BINS = 64;
MIN_FIRSTLAT_BINS = 3;
MAX_FIRSTLAT_BINS = 10000;

DEF_FIRSTLAT_MAX :Double = 0.05;
MIN_FIRSTLAT_MAX :Double = 1.0e-6;
MAX_FIRSTLAT_MAX :Double = 1.0e6;


type

TQUB_Task_MdlRpt_PropertiesForm = class(TQUB_Task_PropertiesForm)
  Label1: TLabel;
  cmbReport: TComboBox;
  GroupBox1: TGroupBox;
  Label2: TLabel;
  Label3: TLabel;
    txtMax: TEdit;
  txtBins: TEdit;

  procedure FormShow(Sender: TObject);
  procedure FormCreate(Sender: TObject);

  procedure cmbReportChange(Sender: TObject);


protected

  prReport          :String;

  procedure prSetReport(x :String);

  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public


  property Report        :String        read prReport        write prSetReport;


end;


implementation

uses

QUB_GUIInterface,
QUB_Viewport_Python,
QUB_Node_Python;


{$R *.DFM}

procedure TQUB_Task_MdlRpt_PropertiesForm.prSetReport(x :String);
var
  i       :Integer;
begin
  i       := cmbReport.Items.IndexOf(x);
  if i >= 0 then begin
    cmbReport.ItemIndex := i;
    prReport := x;
    Properties['Report'].SetData(x);
  end else if cmbReport.ItemIndex < 0 then begin
    cmbReport.ItemIndex := 0; // custom
    if Visible then Properties['Report'].SetData( cmbReport.Items[0] );
  end;
end;

// *****************************************************************************

procedure TQUB_Task_MdlRpt_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddEditProperty(TQUB_Property_Edit.Create('["FirstLatMax"]',
                    'FirstLatMax', REGTYPE_FLOAT, DEF_FIRSTLAT_MAX,
                    txtMax,
                    AcceptFloatBetween(MIN_FIRSTLAT_MAX, MAX_FIRSTLAT_MAX),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["FirstLatBins"]',
                    'FirstLatBins', REGTYPE_INT, DEF_FIRSTLAT_BINS,
                    txtBins,
                    AcceptIntBetween(MIN_FIRSTLAT_BINS, MAX_FIRSTLAT_BINS),
                    '') );

  RegKeyName            := 'Software\QUB\Task MdlRpt';
  ProfileCat            := 'MdlRpt';

  WikiTopic             := 'Modeling:Mdl_Report#Properties';

  mnuProfiles.Top       := 84;
  mnuProfiles.Left      := 204;

end;

procedure TQUB_Task_MdlRpt_PropertiesForm.FormShow(Sender: TObject);
var
  Dir            :String;
  sr             :TSearchRec;
begin
  inherited;

  // clear and repopulate cmbReport from Reports\MdlRpt\*.txt
  cmbReport.Items.Clear;
  cmbReport.Items.Add('Custom...');
  cmbReport.Items.Add('none');

  Dir := ExtractFilePath(Application.ExeName) + 'Reports\MdlRpt\';
  if 0 = FindFirst(Dir + '*.txt', 0, sr) then begin
    repeat
      cmbReport.Items.Add( Copy(sr.Name, 0, Length(sr.Name) - 4) );
    until 0 <> FindNext(sr);
  end;
  Sysutils.FindClose(sr);

  Report               := Report;

end;


procedure TQUB_Task_MdlRpt_PropertiesForm.cmbReportChange(Sender: TObject);
begin
  inherited;

  Report               := cmbReport.Text;
end;


//******************************************************************************


procedure TQUB_Task_MdlRpt_PropertiesForm.prSetInitialProperties(Props :INode);
begin
  inherited;

  if Props['Report'].DataCount = 0
  then Props['Report'].SetData( DEF_REPORT );
end;                

procedure TQUB_Task_MdlRpt_PropertiesForm.prUpdateComponents;
begin
  inherited;

  Report := Properties['Report'].DataAsString;
end;


// *****************************************************************************
end.
