unit QUB_Task_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Classes, Controls, StdCtrls, Windows, Messages, SysUtils, Graphics, Forms, Dialogs,
ComCtrls, ExtCtrls, Registry,

QUB_Global,
QUB_Properties,
QUB_Node, QUB_Profiles, RegistryInterface;


const


DEF_CHANNELINDEX = 0;
DEF_THREADCOUNT = 1;
DEF_REMOTE = 0;


type


TQUB_Task_PropertiesForm          = class(TQUB_PropertiesForm)


  btnOK                           :TButton;
  btnCancel                       :TButton;
  btnHelp                         :TButton;
  panXChannelIndex                :TPanel;
  lblXChannelIndex                :TLabel;
  txtXChannelIndex                :TEdit;
  udXChannelIndex                 :TUpDown;
  panThreads                      :TPanel;
  lblClientServer                 :TLabel;
  lblThreadCount                  :TLabel;
  chkXRemote                      :TCheckBox;
  txtXThreadCount                 :TEdit;
  udXThreadCount                  :TUpDown;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure btnOKClick(Sender: TObject);
  procedure btnHelpClick(Sender: TObject);


protected


  prProfileCat                            :String;
  prProfile                               :TQUB_ProfileCat;
  mnuProfiles                             :TQUB_ProfileEditBox;

  prOnGetProperties                       :TNotifyEvent;
  prOnPutProperties                       :TNotifyEvent;
  prWikiTopic                             :String;

  prBeforeGetProperties                   :procedure of object;
  // e.g. when mac wants properties put into params structure

  prGetsFocus                             :TWinControl;


  procedure prSetProfileCat(const Cat :String);
  procedure prOnSetProperties(Sender :TQUB_ProfileCat; Updates :INode);


  // set this last, in subclass formcreate or the mechanism lies dormant
  property ProfileCat                     :String                read prProfileCat                write prSetProfileCat;


public


  function SetGetProperties(Props :INode): INode;
  // like the old PropertiesTree:
  //   sets the ones you provide,
  //   returns your tree filled with everything
  //   (unless yours is null -- then it creates one for you)
  // after updating a subclass for new-style, change references to PropertiesTree to GetSetProperties (e.g. in ActionSelector_*)


  property OnGetProperties                :TNotifyEvent          read prOnGetProperties           write prOnGetProperties;
  property OnPutProperties                :TNotifyEvent          read prOnPutProperties           write prOnPutProperties;
  property WikiTopic                      :String                read prWikiTopic                 write prWikiTopic;
  property GetsFocus                      :TWinControl           read prGetsFocus                 write prGetsFocus;


end;


implementation


{$R *.DFM}


uses


QUB_GUIInterface,
QUB_Viewport_Data,
QUB_QVO_QVD,
QUB_QFS_QDF;


//******************************************************************************
procedure TQUB_Task_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  mnuProfiles                             := TQUB_ProfileEditBox.Create(Self);
  mnuProfiles.Parent                      := Self;
  mnuProfiles.SetBounds(ClientWidth - 100 - 10,
                        ClientHeight - mnuProfiles.Height - 10,
                        100, mnuProfiles.Height);
  mnuProfiles.Anchors                     := [akRight, akBottom];
  mnuProfiles.Visible                     := False;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["ChannelIndex"]',
                      'ChannelIndex', REGTYPE_INT, DEF_CHANNELINDEX,
                      udXChannelIndex,
                      txtXChannelIndex, AcceptIntGreaterThan(-1)));
  AddUpDownProperty(TQUB_Property_UpDown.Create('["ThreadCount"]',
                      'ThreadCount', REGTYPE_INT, DEF_THREADCOUNT,
                      udXThreadCount,
                      txtXThreadCount, AcceptIntGreaterThan(0)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Remote"]',
                        'Remote', REGTYPE_BOOL, DEF_REMOTE,
                        chkXRemote));
end;

procedure TQUB_Task_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  try
    if ViewportData.CurrData <> nil then begin
      udXChannelIndex.Max  := ViewportData.CurrData.DataObject.ADChannelCount - 1;
      Properties['ChannelIndex'].SetData(ViewportData.CurrData.ActiveChannel);
    end;
  except
  end;

  if Assigned(prOnPutProperties) then begin
    prOnPutProperties(Self);
  end;

  prUpdateComponents;

  if prGetsFocus <> nil then begin
    prGetsFocus.SetFocus;
    prGetsFocus := nil;
  end;
end;

procedure TQUB_Task_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  try
    if ViewportData.CurrData <> nil then begin
      ViewportData.CurrData.ActiveChannel := Properties['ChannelIndex'].Data[0];
    end;
  except
  end;
  
  if prProfile <> nil then prProfile.SaveActive;
end;

procedure TQUB_Task_PropertiesForm.prSetProfileCat(const Cat :String);
begin
  prProfileCat           := Cat;
  prProfile              := ProfileMgr[Cat];
  prSetInitialProperties( prProfile.Active );
  prProfile.OnSet        := prOnSetProperties;
  mnuProfiles.Cat        := Cat;
  mnuProfiles.Visible    := True;
end;

procedure TQUB_Task_PropertiesForm.prOnSetProperties(Sender :TQUB_ProfileCat; Updates :INode);
begin
  SetProperties(Updates);
  prUpdateComponents;
end;

function TQUB_Task_PropertiesForm.SetGetProperties(Props :INode): INode;
  procedure Recur(Mine, Theirs :INode);
  // adds missing subtrees from Mine to Theirs
  var
    MyChild, TheirChild :INode;
  begin
    MyChild := Mine.Child;
    while not MyChild.IsNull do begin
      TheirChild := Theirs.Find(MyChild.Name);
      if TheirChild.IsNull then begin
        Theirs.AppendClone(MyChild, True);
      end else begin
        // TheirChild.EmulateData(MyChild);  // necessary? this value should have been just set.  maybe if it wasn't accepted, but considering SetProperties above, that doesn't seem possible
        Recur(MyChild, TheirChild);
      end;
      MyChild := MyChild.Sibling;
    end;
  end;
begin
  if Props.IsNull then begin
    Result   := Properties.Clone(True);
    Exit;
  end;

  if prProfile <> nil then prProfile.SetProperties(Props);
  Recur(Properties, Props);

  Result     := Props;
end;


procedure TQUB_Task_PropertiesForm.btnOKClick(Sender: TObject);
begin
  if Assigned(prBeforeGetProperties) then begin
    prBeforeGetProperties;
  end;
  if Assigned(prOnGetProperties) then begin
    prOnGetProperties(Self);
  end;
end;

procedure TQUB_Task_PropertiesForm.btnHelpClick(Sender: TObject);
begin
  if WikiTopic <> '' then begin
    ShowWiki(WikiTopic, Handle);
  end else if HelpContext <> 0 then begin
    Application.HelpCommand(HELP_CONTEXT, HelpContext);
  end;
end;


//******************************************************************************
end.
