{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Chris c.2005
// python-invokable custom modal dialog
// expects fields etc from Dialogs.py
// integrated with profiles
// supersedes QUB_Python_ModalDialog


unit QUB_Task_Python_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
StdCtrls, ExtCtrls, Math,

PythonEngine,

QUB_Node, QUB_Node_Python,

QUB_Profiles, QUB_Properties, QUB_Task_Properties, QUB_Task_Data_Properties,
QUB_DataList, ComCtrls;


type


TQUB_TPP_Layout     = class;
TQUB_TPP_Item       = class;
TQUB_TPP_Label      = class;
TQUB_TPP_TextBox    = class;
TQUB_TPP_CheckBox   = class;
TQUB_TPP_RadioGroup = class;

//----- Modal Dialogue - collection of controls
TQUB_Task_Python_PropertiesForm = class(TQUB_Task_Data_PropertiesForm)

  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);

protected

  prOldOnSet                     :TQUB_Profile_OnSet;

  prLayout                       :TQUB_TPP_Layout;
  prItems                        :TList;
  prResults                      :PPyObject;

  procedure prButtonClick(Sender: TObject);

public

  procedure Initialize(ACaption, Fields, Buttons, Cat, Options :PPyObject);

  property Results               :PPyObject        read prResults            write prResults;

end;


TQUB_TPP_Layout = class
protected
  prDialog             :TQUB_Task_Python_PropertiesForm;


public
  panDataList          :TQUB_DataList;
  Items                :Array of TQUB_TPP_Item;
  Buttons              :Array of TButton;

  HasChannel           :Boolean;
  HasDataList          :Boolean;
  HasFileList          :Boolean;
  HasProcessed         :Boolean;
  HasThreads           :Boolean;

  constructor Create(Dialog :TQUB_Task_Python_PropertiesForm; Options :PPyObject);
  destructor Destroy; override;

  procedure AddItem(PyItem :PPyObject);
  function AddButton(const Caption :String): TButton;
  procedure LayOut;

end;


//----- Base class for Python/Delphi controls
TQUB_TPP_Item = class(TPanel)
public

  Name               :String;

  constructor CreatePy(AItem :PPyObject; AParent :TWinControl); virtual;

end;

//----- Label Control
TQUB_TPP_Label = class(TQUB_TPP_Item)
public

  constructor CreatePy(AItem :PPyObject; AParent :TWinControl); override;

end;

//----- TextBox Control
TQUB_TPP_TextBox = class(TQUB_TPP_Item)
protected
  procedure prOnResize(Sender :TObject);

public
  Lbl              :TLabel;
  Txt              :TEdit;

  Acceptor         :PPyObject;
  Format           :PPyObject;

  constructor CreatePy(AItem :PPyObject; AParent :TWinControl); override;
  destructor Destroy; override;

end;

//----- Checkbox control
TQUB_TPP_CheckBox = class(TQUB_TPP_Item)
public
  CheckBox         :TCheckBox;

  constructor CreatePy(AItem :PPyObject; AParent :TWinControl); override;

end;

//----- Radio group control
TQUB_TPP_RadioGroup = class(TQUB_TPP_Item)
public
  Buttons         :array of TRadioButton;

  constructor CreatePy(AItem :PPyObject; AParent :TWinControl); override;

end;


//----- Make TaskProperties() available to Python
procedure InitPythonTaskProperties;


//------------------------------------------------------------------------------
implementation


uses


QUB_GUIInterface,
QUB_Viewport_Python, QUB_Python_Thread;


{$R *.DFM}


// *****************************************************************************

procedure TQUB_Task_Python_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prResults        := nil;
end;

procedure TQUB_Task_Python_PropertiesForm.Initialize(ACaption, Fields, Buttons, Cat, Options :PPyObject);
var
  i                :Integer;
  PyItem           :PPyObject;
  PyJunk           :PPyObject;
  CatName          :String;
begin
  PythonThread.BeginPython;

  prLayout         := TQUB_TPP_Layout.Create(Self, Options);

  with GetPythonEngine do begin
    if ACaption <> nil then begin
      Caption      := PyObject_DelphiStr( ACaption );
    end;

    if (Fields <> nil) and (0 <> PySequence_Check(Fields)) then begin
      for i := 0 to PySequence_Length(Fields) - 1 do begin
        PyItem         := PySequence_GetItem(Fields, i);
        prLayout.AddItem(PyItem);
        Py_DECREF( PyItem );
      end;
    end;

    if (Buttons <> nil) and (0 <> PySequence_Check(Buttons)) and (0 <> PySequence_Length(Buttons)) then begin
      for i := 0 to PySequence_Length(Buttons) - 1 do begin
        PyJunk         := PySequence_GetItem(Buttons, i);
        prLayout.AddButton( PyObject_DelphiStr( PyJunk ) )
          .OnClick     := prButtonClick;
        Py_DECREF( PyJunk );
      end;
    end else begin {no buttons}
      prLayout.AddButton('OK').OnClick := prButtonClick;
    end;
  end;

  CatName := PyObject_DelphiStr(Cat);

  PythonThread.EndPython;

  prOldOnSet := ProfileMgr[CatName].OnSet;
  ProfileCat := CatName;
end;

procedure TQUB_Task_Python_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  prLayout.LayOut;
end;

procedure TQUB_Task_Python_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  if prLayout.HasChannel then begin
    if ViewportData.CurrData <> nil then begin
      ViewportData.CurrData.ActiveChannel := Properties['ChannelIndex'].Data[0];
    end;
  end;

  // un initialize()
  prLayout.Free;
  prLayout := nil;
  prProfile.OnSet := prOldOnSet;
end;

procedure TQUB_Task_Python_PropertiesForm.prButtonClick(Sender: TObject);
begin
  PythonThread.BeginPython;

  with GetPythonEngine do begin
    Py_XDECREF( prResults );
    prResults   := PyString_FromString( PChar( TButton(Sender).Caption ) );
  end;

  PythonThread.EndPython;

  Close;
end;


// *****************************************************************************

constructor TQUB_TPP_Layout.Create(Dialog :TQUB_Task_Python_PropertiesForm; Options :PPyObject);
var
  i                    :Integer;
  OptItems, OptItem    :PPyObject;
  OptKey, OptValue     :PPyObject;
  OptName              :String;
  OptTrue              :Boolean;
  MethItems, Args      :PPyObject;
begin
  prDialog             := Dialog;

  // assume we have the GIL

  with GetPythonEngine do begin
    if (Options <> nil) and (0<>PyMapping_Check(Options)) then begin
      MethItems        := PyObject_GetAttrString(Options, 'items');
      Args             := PyTuple_New(0);
      OptItems         := PyObject_CallObject(MethItems, Args);
      Py_DECREF(Args);
      Py_DECREF(MethItems);
      for i := 0 to PySequence_Length(OptItems) - 1 do begin
        OptItem        := PySequence_GetItem(OptItems, i);
        OptKey         := PySequence_GetItem(OptItem, 0);
        OptValue       := PySequence_GetItem(OptItem, 1);

        OptName        := PyObject_DelphiStr(OptKey);
        OptTrue        := 0<>PyObject_IsTrue(OptValue);

        if OptName = 'DataChannel' then begin
          HasChannel   := OptTrue;
        end else if OptName = 'DataList' then begin
          HasDataList  := OptTrue;
        end else if OptName = 'FileList' then begin
          HasFileList  := OptTrue;
        end else if OptName = 'Process' then begin
          HasProcessed := OptTrue;
        end else if OptName = 'Threads' then begin
          HasThreads   := OptTrue;
        end;

        Py_DECREF(OptValue);
        Py_DECREF(OptKey);
        Py_DECREF(OptItem);
      end;
      Py_DECREF(OptItems);
    end;
  end;
end;

destructor TQUB_TPP_Layout.Destroy;
var
  i                 :Integer;
begin
  try
    for i := 0 to Length(Items) - 1 do begin
      Items[i].Free;
    end;
    SetLength(Items, 0);
    for i := 0 to Length(Buttons) - 1 do begin
      Buttons[i].Free;
    end;
    SetLength(Buttons, 0);
  finally
    inherited;
  end;
end;


procedure TQUB_TPP_Layout.AddItem(PyItem :PPyObject);
var
  Item             :TQUB_TPP_Item;
  PyJunk           :PPyObject;
  ItemType         :String;
  Path             :String;
  i, ix            :Integer;
  RGItem           :TQUB_TPP_RadioGroup;
  RGProperty       :TQUB_Property_RadioGroup;
begin
  with GetPythonEngine do begin
    ItemType       := 'bullshit';
    Item           := nil;

    PyJunk         := PyObject_GetAttrString(PyItem, PChar('type'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      ItemType     := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;

    PythonThread.EndPython;

    if ItemType = 'hidden' then begin
      Item              := TQUB_TPP_Item.CreatePy(PyItem, prDialog);
    end else if ItemType = 'label' then begin
      Item              := TQUB_TPP_Label.CreatePy(PyItem, prDialog);
    end else if ItemType = 'textbox' then begin
      Item              := TQUB_TPP_TextBox.CreatePy(PyItem, prDialog);
      Path    := Format('["%s"]', [Item.Name]);
      prDialog.AddEditProperty( TQUB_Property_Edit.CreatePy(Path, TQUB_TPP_TextBox(Item).Txt, TQUB_TPP_TextBox(Item).Acceptor, TQUB_TPP_TextBox(Item).Format) );
    end else if ItemType = 'checkbox' then begin
      Item              := TQUB_TPP_CheckBox.CreatePy(PyItem, prDialog);
      Path    := Format('["%s"]', [Item.Name]);
      prDialog.AddCheckBoxProperty( TQUB_Property_CheckBox.Create(Path, '', REGTYPE_NONE, 0, TQUB_TPP_CheckBox(Item).CheckBox) );
    end else if ItemType = 'radiogroup' then begin
      RGItem            := TQUB_TPP_RadioGroup.CreatePy(PyItem, prDialog);
      Item              := RGItem;
      Path              := Format('["%s"]', [Item.Name]);
      RGProperty        := TQUB_Property_RadioGroup.Create(Path, '', REGTYPE_NONE, '');
      for i := 0 to Length(RGItem.Buttons) - 1 do begin
        RGProperty.AddButton(
           TQUB_Property_RadioButton.Create(RGItem.Buttons[i],
                                            RGItem.Buttons[i].Caption,
                                            RGItem.Buttons[i].Caption)
        );
      end;
      prDialog.AddRadioGroupProperty(RGProperty);
    end;

    PythonThread.BeginPython;

    if Item <> nil then begin
      ix := Length(Items);
      SetLength(Items, ix+1);
      Items[ix] := Item;
    end;
  end;
end;

function TQUB_TPP_Layout.AddButton(const Caption :String): TButton;
var
  ix                :Integer;
begin
  Result            := TButton.Create(prDialog);
  Result.Parent     := prDialog;
  Result.Caption    := Caption;

  ix                := Length(Buttons);
  SetLength(Buttons, ix+1);
  Buttons[ix]       := Result;
end;


const


XPAD                      = 10;
YPAD                      = 10;
DATA_LIST_HEIGHT          = 100;


procedure TQUB_TPP_Layout.LayOut;
var
  x, y, w, i          :Integer;
  ColsY               :Integer;
  MaxY                :Integer;
  TotalHeight         :Integer;
begin
  y                   := YPAD;
  x                   := XPAD;

  PythonThread.BeginPython;

  with prDialog do begin
    if HasChannel then begin
      panXChannelIndex.Visible := True;
      panXChannelIndex.Top     := y;
      panXChannelIndex.Left    := x;
      y                       := panXChannelIndex.Top + panXChannelIndex.Height + YPAD;
    end else begin
      panXChannelIndex.Visible := False;
    end;

    if HasDataList then begin
      panDataList             := TQUB_DataList.Create(prDialog);
      panDataList.Parent      := prDialog;
      panDataList.SetBounds(x, y, ClientWidth - 2*XPAD, DATA_LIST_HEIGHT);
      panDataList.FileListAvail := HasFileList;
      panDataList.ShowFiles;
      if ViewportModel.CurrModel <> nil then begin
        panDataList.ShowModelVars(ViewportModel.CurrModel, False);
      end;
      y                       := y + DATA_LIST_HEIGHT + YPAD;
    end;

    ColsY                     := y;
    MaxY                      := ColsY;

    if HasProcessed or HasThreads then begin
      // one-column layout
      x                    := (ClientWidth - 3*XPAD) div 2 + 2*XPAD;
      y                    := ColsY;
      w                    := (ClientWidth - 3*XPAD) div 2;

      for i := 0 to Length(Items) - 1 do begin
        Items[i].SetBounds(x, y, w, Items[i].Height);
        y                  := y + Items[i].Height + YPAD;
        MaxY               := Max(y, MaxY);
      end;

    end else begin
      // two-column layout
      TotalHeight               := 0;
      for i := 0 to Length(Items) - 1 do begin
        TotalHeight  := TotalHeight + Items[i].Height + YPAD;
      end;

      y                    := ColsY;
      w                    := (ClientWidth - 3*XPAD) div 2;
      i                    := 0;
      while (i < Length(Items)) and ((y-ColsY) < (TotalHeight div 2)) do begin
        Items[i].SetBounds(x, y, w, Items[i].Height);
        y                  := y + Items[i].Height + YPAD;
        MaxY               := Max(y, MaxY);
        Inc(i);
      end;
      x                    := (ClientWidth - 3*XPAD) div 2 + 2*XPAD;
      y                    := ColsY;
      while (i < Length(Items)) do begin
        Items[i].SetBounds(x, y, w, Items[i].Height);
        y                  := y + Items[i].Height + YPAD;
        MaxY               := Max(y, MaxY);
        Inc(i);
      end;

    end;

    x  := XPAD;
    y  := ColsY;

    if HasProcessed then begin
      panXDataProcessed.Visible := True;
      if HasThreads then begin
        panXDataProcessed.SetBounds(x, y, (ClientWidth - 3*XPAD) div 2, (MaxY-ColsY-2*YPAD) div 2);
      end else begin
        panXDataProcessed.SetBounds(x, y, (ClientWidth - 3*XPAD) div 2, MaxY-ColsY);
      end;
      y                       := panXDataProcessed.Top + panXDataProcessed.Height + YPAD;
    end else begin
      panXDataProcessed.Visible := False;
    end;

    if HasThreads then begin
      panThreads.Visible     := True;
      panThreads.SetBounds(x, y, (ClientWidth - 3*XPAD) div 2, MaxY-ColsY-YPAD);
    end else begin
      panThreads.Visible     := False;
    end;

    x := XPAD;
    y := MaxY;

    for i := 0 to Length(Buttons) - 1 do begin
      Buttons[i].SetBounds(x, y, Buttons[i].Width, Buttons[i].Height);
      x            := x + Buttons[i].Width + XPAD;
    end;

    // assuming there's at least 1 button (Buttons.py ensures this)
    ClientHeight := y + Buttons[0].Height + YPAD;
  end;

  PythonThread.EndPython;
end;


// *****************************************************************************

constructor TQUB_TPP_Item.CreatePy(AItem :PPyObject; AParent :TWinControl);
var
  PyJunk               :PPyObject;
begin
  inherited Create(AParent);
  Parent := AParent;

  BevelOuter := bvNone;
  Caption    := '';

  PythonThread.BeginPython;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(AItem, PChar('name'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Self.Name        := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;
  end;

  PythonThread.EndPython;
end;


constructor TQUB_TPP_Label.CreatePy(AItem :PPyObject; AParent :TWinControl);
var
  PyJunk               :PPyObject;
  Lbl                  :TLabel;
begin
  inherited;

  Lbl                  := TLabel.Create(Self);
  Lbl.Parent           := Self;
  Lbl.Caption          := '';
  SetBounds(Lbl.Left, Lbl.Top, Lbl.Width, Lbl.Height);
  Lbl.Align            := alClient;

  PythonThread.BeginPython;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(AItem, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Lbl.Caption       := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;
  end;

  PythonThread.EndPython;
end;


constructor TQUB_TPP_TextBox.CreatePy(AItem :PPyObject; AParent :TWinControl);
var
  PyJunk               :PPyObject;
begin
  inherited;

  OnResize             := prOnResize;

  Lbl                  := TLabel.Create(Self);
  Lbl.Parent           := Self;
  Lbl.Caption          := '';

  Txt                  := TEdit.Create(Self);
  Txt.Parent           := Self;
  Txt.Text             := '';

  Height               := Max(Txt.Height, Lbl.Height);

  PythonThread.BeginPython;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(AItem, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Lbl.Caption     := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;

    PyJunk             := PyObject_GetAttrString(AItem, PChar('nodeAcceptor'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      if 0 <> PyCallable_Check(PyJunk) then begin
        Acceptor       := PyJunk;
      end else begin
        Py_DECREF( PyJunk );
      end;
    end;

    PyJunk             := PyObject_GetAttrString(AItem, PChar('nodeFormat'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      if 0 <> PyCallable_Check(PyJunk) then begin
        Format         := PyJunk;
      end else begin
        Py_DECREF( PyJunk );
      end;
    end;

    if Acceptor = nil then Acceptor := EvalString(AcceptString);
    if Format = nil then Format := EvalString('NodeFormat("%s")');
  end;

  PythonThread.EndPython;
end;

destructor TQUB_TPP_TextBox.Destroy;
begin
  try
    PythonThread.BeginPython;
    with GetPythonEngine do begin
      Py_XDECREF( Acceptor );
      Py_XDECREF( Format );
    end;
  finally
    PythonThread.EndPython;
    inherited;
  end;
end;

procedure TQUB_TPP_TextBox.prOnResize(Sender :TObject);
begin
  Lbl.Top            := 0;
  Lbl.Left           := 0;
  Lbl.Width          := ClientWidth div 2;
  Lbl.Height         := ClientHeight;
  Txt.Top            := 0;
  Txt.Left           := Lbl.Width;
  Txt.Width          := Lbl.Width;
  Txt.Height         := Lbl.Height;
end;


constructor TQUB_TPP_CheckBox.CreatePy(AItem :PPyObject; AParent :TWinControl);
var
  Chk                  :TCheckBox;
  PyJunk               :PPyObject;
begin
  inherited;

  Chk                  := TCheckBox.Create(Self);
  CheckBox             := Chk;
  Chk.Parent           := Self;
  Chk.Caption          := '';
  Chk.Checked          := False;
  Chk.Alignment        := taRightJustify;
  Height               := Chk.Height;
  Chk.Align            := alClient;

  PythonThread.BeginPython;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(AItem, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Chk.Caption      := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;
  end;

  PythonThread.EndPython;
end;


const
  RADIO_GROUP_START         = 16;
  RADIO_GROUP_END           = 9;


constructor TQUB_TPP_RadioGroup.CreatePy(AItem :PPyObject; AParent :TWinControl);
var
  Rad               :TGroupBox;
  PyJunk, PyItems   :PPyObject;
  i, y              :Integer;
begin
  inherited;

  Rad               := TGroupBox.Create(Self);
  Rad.Parent        := Self;
  Rad.Caption       := '';
  Rad.Align         := alClient;

  PythonThread.BeginPython;

  with GetPythonEngine do begin
    PyJunk             := PyObject_GetAttrString(AItem, PChar('caption'));
    if PyJunk = nil then begin
      PyErr_Clear();
    end else begin
      Rad.Caption   := PyObject_DelphiStr(PyJunk);
      Py_DECREF( PyJunk );
    end;

    PyItems         := PyObject_GetAttrString(AItem, PChar('choices'));
    if PyItems = nil then begin
      PyErr_Clear();
    end else if 0 <> PySequence_Check(PyItems) then begin
      y             := RADIO_GROUP_START;
      SetLength(Buttons, PySequence_Length(PyItems));
      for i := 0 to PySequence_Length(PyItems) - 1 do begin
        PyJunk              := PySequence_GetItem(PyItems, i);
        Buttons[i]          := TRadioButton.Create(Rad);
        Buttons[i].Parent   := Rad;
        Buttons[i].Tag      := i;
        Buttons[i].Caption  := PyObject_DelphiStr(PyJunk);
        Buttons[i].SetBounds(XPAD, y, Rad.Width-2*XPAD, Buttons[i].Height);
        y                   := y + Buttons[i].Height;
        Buttons[i].Anchors  := [akLeft, akRight, akTop];
        Py_DECREF( PyJunk );
      end;
      Py_DECREF( PyItems );
      Height            := y + RADIO_GROUP_END;
    end;
  end;

  PythonThread.EndPython;
end;


// *****************************************************************************

type

TQUB_Task_Properties_PyFunc = class(TPyObject)
protected
  prCaption         :PPyObject;
  prButtons         :PPyObject;
  prFields          :PPyObject;
  prCat             :PPyObject;
  prOpts            :PPyObject;
  prResults         :PPyObject;

  procedure prDoDialog;

public

  function  Call( args, kw : PPyObject) : PPyObject; override;

end;

function TQUB_Task_Properties_PyFunc.Call( args, kw :PPyObject ) : PPyObject;
var
  argc, argi         :Integer;
begin
  with GetPythonEngine do begin
    argc := PyTuple_Size(args);
    argi := 0;
    if argi < argc then begin
      prCaption := PyTuple_GetItem(args, argi);
      Inc(argi);
    end else if kw <> nil then begin
      prCaption := PyDict_GetItemString(kw, PChar('caption'));
    end else begin
      prCaption := nil;
    end;
    if argi < argc then begin
      prButtons := PyTuple_GetItem(args, argi);
      Inc(argi);
    end else if kw <> nil then begin
      prButtons := PyDict_GetItemString(kw, PChar('buttons'));
    end else begin
      prButtons := nil;
    end;
    if argi < argc then begin
      prFields := PyTuple_GetItem(args, argi);
      Inc(argi);
    end else if kw <> nil then begin
      prFields := PyDict_GetItemString(kw, PChar('fields'));
    end else begin
      prFields := nil;
    end;
    if argi < argc then begin
      prCat := PyTuple_GetItem(args, argi);
      Inc(argi);
    end else if kw <> nil then begin
      prCat := PyDict_GetItemString(kw, PChar('cat'));
    end else begin
      prCat := nil;
    end;
    if argi < argc then begin
      prOpts := PyTuple_GetItem(args, argi);
      // Inc(argi);
    end else if kw <> nil then begin
      prOpts := PyDict_GetItemString(kw, PChar('options'));
    end else begin
      prOpts := nil;
    end;

    PythonThread.SyncGUIProc( prDoDialog );

    Result := prResults;
  end;
end;

procedure TQUB_Task_Properties_PyFunc.prDoDialog;
var
  Form                     :TQUB_Task_Python_PropertiesForm;
begin
  Form                     := TQUB_Task_Python_PropertiesForm.Create(nil);

  Form.Initialize(prCaption, prFields, prButtons, prCat, prOpts);
  Form.ShowModal;
  prResults                := Form.Results;
  // stealing the reference
  Form.Results             := nil;

  Form.Release;
end;


var
  QUB_Task_Properties_PyFuncType      :TPythonType;


procedure InitPythonTaskProperties;
var
  PromptFunc        :PPyObject;
begin
  QUB_Task_Properties_PyFuncType := TPythonType.Create(nil);
  QUB_Task_Properties_PyFuncType.Engine := GetPythonEngine;
  QUB_Task_Properties_PyFuncType.Module := PythonViewport.pmQUBTypes;
  QUB_Task_Properties_PyFuncType.TypeName := 'TaskPropertiesFunc';
  QUB_Task_Properties_PyFuncType.Services.Basic := [bsCall];
  QUB_Task_Properties_PyFuncType.PyObjectClass := TQUB_Task_Properties_PyFunc;
  QUB_Task_Properties_PyFuncType.Initialize;

  PromptFunc := QUB_Task_Properties_PyFuncType.CreateInstance;
  PyNode_PublishPyObj( PromptFunc, 'TaskProperties' );
  GetPythonEngine.Py_DECREF( PromptFunc );
end;


// *****************************************************************************
end.
