unit QUB_Task_Sim_DoseResponse_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  QUB_Task_Properties, StdCtrls, ComCtrls, ExtCtrls, Math,

  Alloc, AllocTypes,
  QUB_Properties,
  QUB_Profiles, QUB_Node,
  QUB_Global,
  QUB_ExpCond_Parameters, QUB_ExpCond_Editor,
  QUB_QVO_QVD,
  QUB_QVO_QVM,
  QUB_QVO_QVP,
  QUB_ZoomBar;

type
TQUB_Task_Sim_DoseResponse_PropertiesForm = class(TQUB_Task_PropertiesForm)			  
   panStimulus: TPaintBox;		  
  Label3: TLabel;			  
  txtSampling: TEdit;
  Label4: TLabel;			  
  panStatus: TPanel;
    Label1: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    Label12: TLabel;
    chkPulseFilter: TCheckBox;
    Label13: TLabel;
    Label14: TLabel;
    Label15: TLabel;
    chkCurrentFilter: TCheckBox;
    Label16: TLabel;
    panMeasure: TPanel;
    Label17: TLabel;
    chkMeasurePeak: TRadioButton;
    chkMeasureTime: TRadioButton;
    Label18: TLabel;
    cmbVariable: TComboBox;
    txtRestingLevel: TEdit;
    txtPulseLevels: TEdit;
    txtPrePulseDur: TEdit;
    txtPulseDur: TEdit;
    txtSweepDur: TEdit;
    txtPulseFilterTau: TEdit;
    txtChannelCount: TEdit;
    txtCurrentFilterFreq: TEdit;
    txtMeasureTime: TEdit;
    chkNormalize: TCheckBox;
    chkSaveMeas: TCheckBox;
    txtMeasFileBase: TEdit;
    btnMeasFileChoose: TButton;
    panVar2: TPanel;
    chkVar2: TCheckBox;
    cmbVariable2: TComboBox;
    Label2: TLabel;
    txtRestingLevel2: TEdit;
    Label19: TLabel;
    txtPrePulseDur2: TEdit;
    Label20: TLabel;
    Label21: TLabel;
    txtPulseDur2: TEdit;
    Label22: TLabel;
    Label23: TLabel;
    txtPulseLevels2: TEdit;
    chkPulseFilter2: TCheckBox;
    Label24: TLabel;
    txtPulseFilterTau2: TEdit;
    Label25: TLabel;
    Label26: TLabel;
    txtRepeats: TEdit;
    udRepeats: TUpDown;
    Label27: TLabel;
    Label28: TLabel;
    txtStartRep: TEdit;
    udStartRep: TUpDown;
    txtStartRep2: TEdit;
    udStartRep2: TUpDown;
  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure panStimulusPaint(Sender: TObject);
    procedure prOnExitWrap(Sender: TObject);
    procedure prOnCheckClickWrap(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure btnMeasFileChooseClick(Sender: TObject);
    procedure cmbVariableChange(Sender: TObject);
protected
  panZoom            :TQUB_ZoomBar;
  prModel            :TQUB_QVO_QVM;
  prModelVars        :TStringList;
  prParams           :TQUB_ExpCond_Parameters;
  prStimulusInvalid  :Boolean;
  prVar2Height       :Integer;

  procedure prAdjustHeight;
  procedure prSetupVars; // list model vars and add to defexpcond on show
  procedure prInvalidateStimulus;
  procedure prSetupStimulus;
  procedure prBuildStimulus(SweepDur, Sampling :Double; PreCount, PulseCount :Integer; Filter :Boolean; FilterTau :Double; var Stimulus :TPADouble);
  procedure prSetStatus(Status :String);
  procedure prUpdateComponents; override;

  procedure prOnZoomChange(Sender :TObject);

public

  StimCount        :Integer;
  Stimulus         :TPADouble;
  Stimulus2        :TPADouble;
  VarName          :String;
  VarName2         :String;

end;


implementation

{$R *.DFM}

uses

QUB_GUIInterface,
QUB_Viewport_Protocol, QUB_QFS_QPF,
QUB_Viewport_Data, QUB_QFS_QDF,
QUBFits, // for FilterDataD
QUB_Viewport_Model, QUB_QFS_QMF, QUB_QFS_Channels;


const
  DEF_DURATION          :Double = 1000.0;
  DEF_REPEATS           :Integer = 1;
  DEF_STARTREP          :Integer = 1;
  DEF_SAMPLING          :Double = 10.0;
  DEF_RESTING           :Double = 0.0;
  DEF_PULSELEVELS       :Double = 1.0;
  DEF_PREPULSE          :Double = 10.0;
  DEF_PULSEDUR          :Double = 100.0;
  DEF_PULSEFILTER       :Integer = 0;
  DEF_PULSEFILTERTAU    :Double = 1.0;
  DEF_VARIABLE2         :Integer = 0;
  DEF_CHANNELCOUNT      :Integer = 1000;
  DEF_CURRENTFILTER     :Integer = 1;
  DEF_CURRENTFILTERFREQ :Double = 2.0;
  DEF_MEASURE           :Integer = 0;
  DEF_MEASURETIME       :Double = 100.0;
  DEF_NORMALIZE         :Integer = 1;
  DEF_SAVEMEAS          :Integer = 0;
  DEF_MEASFILEBASE      :String = 'C:\QuB-DoseResponse';


procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddEditProperty(TQUB_Property_Edit.Create('["SweepDuration"]',
                    '', REGTYPE_NONE, DEF_DURATION,
                    txtSweepDur,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["Repeats"]',
                      '', REGTYPE_NONE, DEF_REPEATS,
                      udRepeats,
                      txtRepeats,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["Sampling"]',
                    '', REGTYPE_NONE, DEF_SAMPLING,
                    txtSampling,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["StartRep"]',
                      '', REGTYPE_NONE, DEF_STARTREP,
                      udStartRep,
                      txtStartRep,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["RestingLevel"]',
                    '', REGTYPE_NONE, DEF_RESTING,
                    txtRestingLevel,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PrePulseDuration"]',
                    '', REGTYPE_NONE, DEF_PREPULSE,
                    txtPrePulseDur,
                    AcceptFloatGreaterThanOrEqualTo(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PulseDuration"]',
                    '', REGTYPE_NONE, DEF_PULSEDUR,
                    txtPulseDur,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PulseLevels"]',
                          '', REGTYPE_NONE, DEF_PULSELEVELS,
                          txtPulseLevels,
                          AcceptFloatList(AcceptFloat),
                          'lambda node, ix: SpacedNodeData(node)') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["PulseFilter"]',
                        '', REGTYPE_NONE, DEF_PULSEFILTER,
                        chkPulseFilter)
                       .Enables(txtPulseFilterTau)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["PulseFilterTau"]',
                    '', REGTYPE_NONE, DEF_PULSEFILTERTAU,
                    txtPulseFilterTau,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Variable2"]',
                        '', REGTYPE_NONE, DEF_VARIABLE2,
                        chkVar2) );
  AddUpDownProperty(TQUB_Property_UpDown.Create('["StartRep2"]',
                      '', REGTYPE_NONE, DEF_STARTREP,
                      udStartRep2,
                      txtStartRep2,
                      AcceptIntGreaterThan(0)));
  AddEditProperty(TQUB_Property_Edit.Create('["RestingLevel2"]',
                    '', REGTYPE_NONE, DEF_RESTING,
                    txtRestingLevel2,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PrePulseDuration2"]',
                    '', REGTYPE_NONE, DEF_PREPULSE,
                    txtPrePulseDur2,
                    AcceptFloatGreaterThanOrEqualTo(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PulseDuration2"]',
                    '', REGTYPE_NONE, DEF_PULSEDUR,
                    txtPulseDur2,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["PulseLevels2"]',
                          '', REGTYPE_NONE, DEF_PULSELEVELS,
                          txtPulseLevels2,
                          AcceptFloatList(AcceptFloat),
                          'lambda node, ix: SpacedNodeData(node)') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["PulseFilter2"]',
                        '', REGTYPE_NONE, DEF_PULSEFILTER,
                        chkPulseFilter2)
                       .Enables(txtPulseFilterTau2)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["PulseFilterTau2"]',
                    '', REGTYPE_NONE, DEF_PULSEFILTERTAU,
                    txtPulseFilterTau2,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["ChannelCount"]',
                    '', REGTYPE_NONE, DEF_CHANNELCOUNT,
                    txtChannelCount,
                    AcceptIntGreaterThan(0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["CurrentFilter"]',
                        '', REGTYPE_NONE, DEF_CURRENTFILTER,
                        chkCurrentFilter)
                       .Enables(txtCurrentFilterFreq)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["CurrentFilterFreq"]',
                    '', REGTYPE_NONE, DEF_CURRENTFILTERFREQ,
                    txtCurrentFilterFreq,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  // chkMeasurePeak, chkMeasureTime
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Measure"]',
                          '', REGTYPE_NONE, DEF_MEASURE)
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMeasurePeak, 0, 0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkMeasureTime, 1, 1)
                               .Enables(txtMeasureTime)
                             )
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["MeasureTime"]',
                    '', REGTYPE_NONE, DEF_MEASURETIME,
                    txtMeasureTime,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Normalize"]',
                        '', REGTYPE_NONE, DEF_NORMALIZE,
                        chkNormalize)
                     );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["SaveMeas"]',
                        '', REGTYPE_NONE, DEF_SAVEMEAS,
                        chkSaveMeas)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["MeasFileBase"]',
                    '', REGTYPE_NONE, DEF_MEASFILEBASE,
                    txtMeasFileBase,
                    AcceptString,
                    '') );

  ProfileCat             := 'DoseResponse';
  WikiTopic              := 'Simulation:DoseResponse#Properties';

  prModelVars            := TStringList.Create;
  prModelVars.Sorted     := True;
  prModelVars.Duplicates := dupIgnore;

  Stimulus               := nil;
  Stimulus2              := nil;
  StimCount              := 0;

  panZoom                := TQUB_ZoomBar.Create(Self);
  panZoom.Anchors        := [akLeft, akRight, akTop];
  panZoom.Top            := panStimulus.Top + panStimulus.Height - panZoom.Height;
  panZoom.Left           := panStimulus.Left;
  panZoom.Width          := panStimulus.Width;
  panStimulus.Height     := panStimulus.Height - panZoom.Height;
  panZoom.Parent         := Self;
  panZoom.OnZoomChange   := prOnZoomChange;
  panZoom.OnStartChange  := prOnZoomChange;

  prVar2Height           := panVar2.Height;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.FormDestroy(Sender: TObject);
begin
  try
    prModelVars.Free;
    FreeMem(Stimulus);
    FreeMem(Stimulus2);
  finally
    inherited;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.FormShow(Sender: TObject);
var
  i :Integer;
  Invalid, Invalid2 :Boolean;
begin
  inherited;

  prParams               := ViewportModel.DefExpCond;
  prModel                := ViewportModel.CurrModel;

  prSetupVars;

  Invalid  := (prModelVars.IndexOf(VarName) < 0);
  Invalid2 := (prModelVars.IndexOf(VarName2) < 0);
  if Invalid and Invalid2 then begin
    if prModelVars.Count > 0 then begin
      VarName := prModelVars[0];
    end;
    if prModelVars.Count > 1 then begin
      VarName2 := prModelVars[1];
    end;
  end else if Invalid then begin
    i := 0;
    while chkVar2.Checked and (i<prModelVars.Count) and (prModelVars[i] = VarName2) do begin
      Inc(i);
    end;
    if i < prModelVars.Count then begin
      VarName := prModelVars[i];
    end;
  end else if Invalid2 then begin
    i := 0;
    while (i<prModelVars.Count) and (prModelVars[i] = VarName) do begin
      Inc(i);
    end;
    if i < prModelVars.Count then begin
      VarName2 := prModelVars[i];
    end;
  end;

  cmbVariable.Items.Clear;
  for i := 0 to prModelVars.Count - 1 do begin
    cmbVariable.Items.Add(prModelVars[i]);
  end;
  cmbVariable2.Items.Assign(cmbVariable.Items);

  cmbVariable.ItemIndex := prModelVars.IndexOf(VarName);
  cmbVariable2.ItemIndex := prModelVars.IndexOf(VarName2);
  cmbVariableChange(cmbVariable);
  
  panZoom.Zoom := 100.0;
  prAdjustHeight;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  // harvest variable name from combo box
  if cmbVariable.ItemIndex >= 0 then begin
    VarName := cmbVariable.Items[cmbVariable.ItemIndex];
  end else begin
    VarName := '';
  end;
  if chkVar2.Checked and (cmbVariable2.ItemIndex >= 0) then begin
    VarName2 := cmbVariable2.Items[cmbVariable2.ItemIndex];
  end else begin
    VarName2 := '';
  end;

  Properties['VarName'].SetData(VarName);
  Properties['VarName2'].SetData(VarName2);
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prAdjustHeight;
begin
  if (panVar2.Height > 1) and (0=Properties['Variable2'].Data[0]) then begin
    Self.Height := Self.Height - prVar2Height + 1;
  end else if (panVar2.Height <= 1) and (0<>Properties['Variable2'].Data[0]) then begin
    Self.Height := Self.Height + prVar2Height - 1;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prSetupVars;
// list model vars and add to expcond on show
var
  i, j, k       :Integer;
begin
  prModelVars.Clear;

  // list active stimuli
  for j := 0 to prModel.ModelObject.Rates.RateCount - 1 do begin
    for k := 1 to 2 do begin
      if prModel.ModelObject.Rates[j].P[k] then prModelVars.Add(prModel.ModelObject.Rates[j].PName[k]);
      if prModel.ModelObject.Rates[j].Q[k] then prModelVars.Add(prModel.ModelObject.Rates[j].QName[k]);
    end;
  end;

  // for our purposes, there should at least be ligand, even if missing from model
  if prModelVars.Count = 0 then prModelVars.Add('Ligand');

  // add if missing
  for i := 0 to prModelVars.Count - 1 do begin
    if prParams.ParameterByName[prModelVars[i]] = nil then begin
      prParams.AddPar(prModelVars[i], 1.0, -1, False, -1);
    end;
  end;

  // knock out inactive stimuli
  for i := 0 to prParams.Count - 1 do begin
    if prModelVars.IndexOf(prParams[i].Name) < 0 then begin
      prParams.ChangeParChannel(i, -1);
    end;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prSetupStimulus;
var
  SweepDur, PreDur, PulseDur     :Double;
  Sampling                       :Double;
  Filter                         :Boolean;
  FilterTau                      :Double;
  PreCount, PulseCount           :Integer;
begin
  // GUI quantities:
  SweepDur                       := Properties['SweepDuration'].Data[0];
  Sampling                       := Properties['Sampling'].Data[0];

  PreDur                         := Properties['PrePulseDuration'].Data[0];
  PulseDur                       := Properties['PulseDuration'].Data[0];
  Filter                         := 0<>Properties['PulseFilter'].Data[0];
  FilterTau                      := Properties['PulseFilterTau'].Data[0];

  // Derived quantities
  StimCount := Round(Sampling * SweepDur);
  PreCount  := Min(Round(Sampling * PreDur), StimCount-1);
  PulseCount := Max(1, Round(Sampling * (PreDur+PulseDur)) - PreCount);
  PulseCount := Min(StimCount-PreCount, PulseCount);

  prBuildStimulus(SweepDur, Sampling, PreCount, PulseCount, Filter, FilterTau, Stimulus);

  PreDur                         := Properties['PrePulseDuration2'].Data[0];
  PulseDur                       := Properties['PulseDuration2'].Data[0];
  Filter                         := 0<>Properties['PulseFilter2'].Data[0];
  FilterTau                      := Properties['PulseFilterTau2'].Data[0];
  PreCount  := Min(Round(Sampling * PreDur), StimCount-1);
  PulseCount := Max(1, Round(Sampling * (PreDur+PulseDur)) - PreCount);
  PulseCount := Min(StimCount-PreCount, PulseCount);

  prBuildStimulus(SweepDur, Sampling, PreCount, PulseCount, Filter, FilterTau, Stimulus2);

  prStimulusInvalid := False;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prBuildStimulus(SweepDur, Sampling :Double; PreCount, PulseCount :Integer; Filter :Boolean; FilterTau :Double; var Stimulus :TPADouble);
var
  FilterInput                    :Array of Double;
  i                              :Integer;
begin
  // build normalized pulse: 0.0 to 1.0
  ReallocMem(Stimulus, StimCount*SizeOf(Double));
  for i := 0 to PreCount - 1 do begin
    Stimulus[i] := 0.0;
  end;
  for i := PreCount to PreCount+PulseCount-1 do begin
    Stimulus[i] := 1.0;
  end;
  for i := PreCount+PulseCount to StimCount - 1 do begin
    Stimulus[i] := 0.0;
  end;

  if Filter and ((2.0 * (FilterTau*2*PI)) < SweepDur) then begin
  // min freq fits 2 periods in data
    try
      SetLength(FilterInput, StimCount);
      Move(Stimulus[0], FilterInput[0], StimCount*SizeOf(Double));
      FilterDataD(@(FilterInput[0]), Stimulus, StimCount, 1e-3/Sampling, 1e3/(2*PI*FilterTau));
    except
      Move(FilterInput[0], Stimulus[0], StimCount*SizeOf(Double));
    end;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.panStimulusPaint(Sender: TObject);
var
  Resting                        :Double;
  Pulses                         :Array of Double;
  Node                           :INode;
  i, x, s                        :Integer;
  lo, hi, dy, p, off             :Double;
  w, h                           :Integer;
begin
  inherited;

  if prStimulusInvalid then prSetupStimulus;
  w  := panStimulus.Width;
  h  := panStimulus.Height;

  Resting                        := Properties['RestingLevel'].Data[0];
  lo := Resting;   hi := Resting;
  Node                           := Properties['PulseLevels'];
  SetLength(Pulses, Node.DataCount);
  for i := 0 to Node.DataCount - 1 do begin
    Pulses[i] := Node.Data[i];
    if Pulses[i] < lo then lo := Pulses[i];
    if Pulses[i] > hi then hi := Pulses[i];
  end;

  dy := hi - lo;
  if dy = 0 then dy := 1.0;
  lo := lo - dy/8;
  hi := hi + dy/8;
  dy := hi - lo;

  // drawing coords:
  // x -> (i*w) div StimCount
  // y -> h - Round((y - lo) * h/dy)

  // scaled pulse:
  // y -> (p-r)*y + r;    combined: y -> h - Round((((p-r)*y + r) - lo) * h/dy)

  with panStimulus.Canvas do begin
    Brush.Style := bsSolid;
    Brush.Color := clWhite;
    Pen.Color   := clWhite;
    Rectangle(0, 0, w, h);

    if StimCount > 0 then begin
      Pen.Color   := clBlack;
      Pen.Width   := 1;
      for s := 0 to Length(Pulses) - 1 do begin
        p   := Pulses[s] - Resting;
        off := panZoom.Start * StimCount / 100.0;
        x   := Round(off);
        MoveTo(0, h - Round(((p*Stimulus[x] + Resting) - lo) * h/dy));
        for i := 1 to w - 1 do begin
          x   := Min(StimCount-1, Round(off + (i * 100.0 * StimCount / panZoom.Zoom) / w));
          LineTo(i, h - Round(((p*Stimulus[x] + Resting) - lo) * h/dy));
        end;
      end;
    end;
  end;

  if 0=Properties['Variable2'].Data[0] then Exit;

  Resting                        := Properties['RestingLevel2'].Data[0];
  lo := Resting;   hi := Resting;
  Node                           := Properties['PulseLevels2'];
  SetLength(Pulses, Node.DataCount);
  for i := 0 to Node.DataCount - 1 do begin
    Pulses[i] := Node.Data[i];
    if Pulses[i] < lo then lo := Pulses[i];
    if Pulses[i] > hi then hi := Pulses[i];
  end;

  dy := hi - lo;
  if dy = 0 then dy := 1.0;
  lo := lo - dy/8;
  hi := hi + dy/8;
  dy := hi - lo;

  with panStimulus.Canvas do begin
    if StimCount > 0 then begin
      Pen.Color   := clRed;
      Pen.Width   := 1;
      for s := 0 to Length(Pulses) - 1 do begin
        p   := Pulses[s] - Resting;
        off := panZoom.Start * StimCount / 100.0;
        x   := Round(off);
        MoveTo(0, h - Round(((p*Stimulus2[x] + Resting) - lo) * h/dy));
        for i := 1 to w - 1 do begin
          x   := Min(StimCount-1, Round(off + (i * 100.0 * StimCount / panZoom.Zoom) / w));
          LineTo(i, h - Round(((p*Stimulus2[x] + Resting) - lo) * h/dy));
        end;
      end;
    end;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prSetStatus(Status :String);
begin
  if Status = ''
  then panStatus.Color := clBtnFace
  else panStatus.Color := TColor($ccaaff);

  panStatus.Caption := Status;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prOnExitWrap(
  Sender: TObject);
begin
  inherited;
  prInvalidateStimulus;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prOnCheckClickWrap(
  Sender: TObject);
begin
  inherited;

  prInvalidateStimulus;
  prAdjustHeight;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prInvalidateStimulus;
begin
  prStimulusInvalid := True;
  panStimulus.Invalidate;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prUpdateComponents;
var
  i        :Integer;
begin
  inherited;

  i := cmbVariable.Items.IndexOf( Properties['VarName'].DataAsString );
  if i >= 0 then cmbVariable.ItemIndex := i;

  i := cmbVariable2.Items.IndexOf( Properties['VarName2'].DataAsString );
  if i >= 0 then cmbVariable2.ItemIndex := i;

  prInvalidateStimulus;
  prAdjustHeight;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.prOnZoomChange;
begin
  panStimulus.Invalidate;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.btnMeasFileChooseClick(Sender: TObject);
var
  Dlg :TSaveDialog;
begin
  inherited;
  Dlg := TSaveDialog.Create(Application);
  try
    Dlg.Filter := 'ASCII data files (*.txt)|*.txt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist];
    Dlg.FileName := txtMeasFileBase.Text;
    if Dlg.Execute then begin
      txtMeasFileBase.Text := Dlg.FileName;
      txtMeasFileBase.OnExit(txtMeasFileBase);
    end else begin
      txtMeasFileBase.Text := Properties['MeasFileBase'].DataAsString;
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Task_Sim_DoseResponse_PropertiesForm.cmbVariableChange(
  Sender: TObject);
begin
  inherited;
  Properties['VarName'].SetData(cmbVariable.Text);
  Properties['VarName2'].SetData(cmbVariable2.Text);
end;


// *****************************************************************************
end.
