unit QUB_Task_Simulate;
{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


interface


uses


Windows, SysUtils, Classes, Clipbrd, Graphics, Forms, SyncObjs, Math,

AllocTypes, Alloc,

AMRandom,

QUB_Global,
QUB_Deviates,

QUB_Thread,
QUB_Task,
QUB_QVO_QVD_Global,
QUB_QVO_QVD,
QUB_QVO_QVM,

QUB_QFS_Segments,
QUB_QFS_QDF_MPX,
QUB_QFS_QMF,
QUB_QFS_Channels,
QUB_QFS_QIF_Segments,

QUB_QVO_QVP, QUB_QFS_QPF, QUB_Task_Protocol_To_Data,

QUB_Node,

QUB_ExpCond_Parameters;

//KPL_Parser,
//KPL_Objects, KPL_Kinetic, KPL_CollectionQUBPulse;


type


TQUB_Task_Simulate_IDLOutput   = (OUT_STATE, OUT_CLASS, OUT_GROUP);
TQUB_Task_Simulate_Stimulation = (SIM_STIM_NONE, SIM_STIM_FILE, SIM_STIM_SCRIPT);
TQUB_Task_Simulate_Response    = (SIM_RESPONSE_STOCHASTIC, SIM_RESPONSE_DETERMINISTIC);


function IDLOutputToString(x :TQUB_Task_Simulate_IDLOutput): String;
function StringToIDLOutput(const x :String): TQUB_Task_Simulate_IDLOutput;

function StimulationToString(x :TQUB_Task_Simulate_Stimulation): String;
function StringToStimulation(const x :String): TQUB_Task_Simulate_Stimulation;

function ResponseToString(x :TQUB_Task_Simulate_Response): String;
function StringToResponse(const x :String): TQUB_Task_Simulate_Response;


const


DATA_MODE_NEW                = 0;
DATA_MODE_REPLACE            = 1;
DATA_MODE_APPEND             = 2;
DATA_MODE_APPENDCURR         = 3;

DEF_STIMULATION              = SIM_STIM_NONE;
DEF_RESPONSE                 = SIM_RESPONSE_STOCHASTIC;

DEF_USECONDVOLTAGE           = 0;
MIN_CONDVOLTAGE              :Double = -10.0;
MAX_CONDVOLTAGE              :Double = +10.0;
DEF_CONDVOLTAGE              :Double = 0.0;

DEF_IFROMG                   = 0;
DEF_VFORI                    = 0.0;
DEF_ILEAK                    = 0;
MIN_GLEAK                    :Double = 0.0;
MAX_GLEAK                    :Double = 1e10;
DEF_GLEAK                    :Double = 1.0;
MIN_VLEAK                    :Double = -1e6;
MAX_VLEAK                    :Double = 1e6;
DEF_VLEAK                    :Double = -60.0;

DEF_STIMULATIONFILE          = -1;
//DEF_STIMULATIONSCRIPT        = -1;

DEF_IDEALIZEDDATA            = 0;
DEF_IDLOUTPUT                = OUT_CLASS;

DEF_DORANDOMIZE              = 1;
MIN_RANDOMSEED               = Low(Integer);
MAX_RANDOMSEED               = High(Integer);
DEF_RANDOMSEED               = 0;

DEF_APPLYDEADTIME            = 0;
DEF_DEADTIME                 :Double = 1e-2;

DEF_FILTERON                 = 0;
MIN_FILTERFREQ               :Double = 0.1;
MAX_FILTERFREQ               :Double = 100.0;
DEF_FILTERFREQ               :Double = 5.0;

GATING_V                     = 0;
GATING_MV                    = 1;
GATING_OTHER                 = 2;
DEF_GATING                   = 0;
DEF_GATING_UNITS             = GATING_V;
DEF_GATING_OTHER             :Double = 1.0;
DEF_GATING_INVERT            = 0;

MIN_SEGMENT_COUNT            = 1;
MAX_SEGMENT_COUNT            = High(Integer) - 1;
DEF_SEGMENT_COUNT            = 1;

MIN_SEGMENT_SIZE             = 10;
MAX_SEGMENT_SIZE             = High(Integer) - 1;
DEF_SEGMENT_SIZE             = 100000;

DEF_RESET_SEGMENT            = 1;

DEF_RANDOMIZE_SEGMENT        = 1;

//Staircase
DEF_STAIRCASE                = 0;
DEF_STAIRCASESTEPS           = 's = 1';
DEF_STAIRCASEUSESTEPS        = 1;
MIN_STAIRCASESTEPSTD         :Double = 0.0;
MAX_STAIRCASESTEPSTD         :Double = 1.0e6;
DEF_STAIRCASESTEPSTD         :Double = 0.0;
DEF_STAIRCASENOJUMPCLASS     = 0;


//Baseline Fluctuations
DEF_BASELINEFLUCT            = 0;

DEF_BASELINEFLUCT_LIFETIME   :Double = 1.0;
MIN_BASELINEFLUCT_LIFETIME   :Double = 1e-6;
MAX_BASELINEFLUCT_LIFETIME   :Double = 1e+6;

DEF_BASELINEFLUCT_MAXAMP     :Double = 1.0;
MIN_BASELINEFLUCT_MAXAMP     :Double = 1e-6;
MAX_BASELINEFLUCT_MAXAMP     :Double = 1e+6;

DEF_BASELINEFLUCT_N          = 1000;
MIN_BASELINEFLUCT_N          = 1;
MAX_BASELINEFLUCT_N          = 100000;

//Baseline LGM
DEF_BASELINE_LGM             = 0;

DEF_BASELINE_LGM_PROCSTD     :Double = 1e-3;
MIN_BASELINE_LGM_PROCSTD     :Double = 1e-10;
MAX_BASELINE_LGM_PROCSTD     :Double = 1e+10;

DEF_BASELINE_LGM_X0          :Double = 0.0;
MIN_BASELINE_LGM_X0          :Double = -1e10;
MAX_BASELINE_LGM_X0          :Double = +1e10;

DEF_BASELINE_LGM_X0P         :Double = 0.0;
MIN_BASELINE_LGM_X0P         :Double = -1e6;
MAX_BASELINE_LGM_X0P         :Double = +1e6;

DEF_BASELINE_LGM_2STATE      = 1;

//Baseline Drift
DEF_BASELINEDRIFT            = 0;

DEF_BASELINEDRIFT_SLOPE      :Double = 0.0;
MIN_BASELINEDRIFT_SLOPE      :Double = -1e+6;
MAX_BASELINEDRIFT_SLOPE      :Double = +1e+6;

DEF_BASELINEDRIFT_INTERCEPT  :Double = 0.0;
MIN_BASELINEDRIFT_INTERCEPT  :Double = -1e+6;
MAX_BASELINEDRIFT_INTERCEPT  :Double = +1e+6;

//Baseline Periodic
DEF_BASELINEPERIODIC         = 0;

DEF_BASELINEPERIODIC_AMP     :Double = 1.0;
MIN_BASELINEPERIODIC_AMP     :Double = 1e-6;
MAX_BASELINEPERIODIC_AMP     :Double = 1e+6;

DEF_BASELINEPERIODIC_FREQ    :Double = 60.0;
MIN_BASELINEPERIODIC_FREQ    :Double = 1e-6;
MAX_BASELINEPERIODIC_FREQ    :Double = 1e+6;

DEF_BASELINEPERIODIC_PHASE   :Double = 0.0;
MIN_BASELINEPERIODIC_PHASE   :Double = 0.0;
MAX_BASELINEPERIODIC_PHASE   :Double = 360.0;

//Baseline Jumps
DEF_BASELINEJUMPS            = 0;

DEF_BASELINEJUMPS_AMPAVG     :Double = 10.0;
MIN_BASELINEJUMPS_AMPAVG     :Double = 0;
MAX_BASELINEJUMPS_AMPAVG     :Double = +1e+6;

DEF_BASELINEJUMPS_AMPSTD     :Double = 5.0;
MIN_BASELINEJUMPS_AMPSTD     :Double = 0;
MAX_BASELINEJUMPS_AMPSTD     :Double = +1e+6;

DEF_BASELINEJUMPS_FASTW      :Double = 0.5;
MIN_BASELINEJUMPS_FASTW      :Double = 0;
MAX_BASELINEJUMPS_FASTW      :Double = 1;

DEF_BASELINEJUMPS_LIFETIME   :Double = 1.0;
MIN_BASELINEJUMPS_LIFETIME   :Double = 1e-6;
MAX_BASELINEJUMPS_LIFETIME   :Double = 1e+6;

// slow or fast of 0 leads to invalid floating point operation
DEF_BASELINEJUMPS_SLOW       :Double = 1.0;
MIN_BASELINEJUMPS_SLOW       :Double = 1e-10;
MAX_BASELINEJUMPS_SLOW       :Double = 1e+6;

DEF_BASELINEJUMPS_FAST       :Double = 1e-3;
MIN_BASELINEJUMPS_FAST       :Double = 1e-10;
MAX_BASELINEJUMPS_FAST       :Double = 1e+6;

DEF_FLUX                     = 0;
DEF_FLUX_FROM                = 1;
DEF_FLUX_TO                  = 2;

DEF_MAKEMOVIE                = 0;
DEF_MAKEMOVIEFRAMEDURATION   :Double = 1.0e0;
DEF_SAVEMOVIEFILES           = 0;


type


TQUB_Task_Simulate_PropertiesRec  = record

  //Data panel
  //ARG_Data                            = 0;
  //ARG_DataMode                        = 1;
  SegmentCount                    :Integer;
  SegmentSize                     :Integer;
  IdealizedData                   :Boolean;
  IDLOutput                       :Integer;
  Randomize                       :Boolean;
  RandomSeed                      :Integer;
  ApplyDeadTime                   :Boolean;
  DeadTime                        :Double;
  FilterOn                        :Boolean;
  FilterFreq                      :Double;

  //Model panel
  //ARG_Model                           = 10;

  //Simulation panel
  Stimulation                     :TQUB_Task_Simulate_Stimulation;
  Response                        :TQUB_Task_Simulate_Response;
  Flux                            :Boolean;
  FluxFrom                        :Integer;
  FluxTo                          :Integer;
  IFromG                          :Boolean;
  VForI                           :Double;
  ILeak                           :Boolean;
  GLeak                           :Double;
  VLeak                           :Double;
  StimulationFile                 :Integer;
  //StimulationScript               :Integer;

  UseCondVoltage                  :Boolean;
  CondVoltage                     :Double;

  //Staircase panel
  Staircase                       :Boolean;
  StaircaseUseSteps               :Boolean;
  StaircaseSteps                  :String;
  StaircaseStepStd                :Double;
  StaircaseNoJumpClass            :Boolean;

  //Baseline panel
  BaselineFluct                   :Boolean;
  BaselineFluctLifetime           :Double;
  BaselineFluctMaxAmp             :Double;
  BaselineFluctN                  :Integer;
  BaselineDrift                   :Boolean;
  BaselineDriftSlope              :Double;
  BaselineDriftIntercept          :Double;
  BaselineJumps                   :Boolean;
  BaselineJumpsAmpAvg             :Double;
  BaselineJumpsAmpStd             :Double;
  BaselineJumpsFastW              :Double;
  BaselineJumpsFast               :Double;
  BaselineJumpsSlow               :Double;
  BaselineJumpsLifetime           :Double;
  BaselinePeriodic                :Boolean;
  BaselinePeriodicAmp             :Double;
  BaselinePeriodicFreq            :Double;
  BaselinePeriodicPhase           :Double;
  BaselineLGM                     :Boolean;
  BaselineLGMProcStd              :Double;
  BaselineLGMX0                   :Double;
  BaselineLGMX0p                  :Double;
  BaselineLGM2State               :Boolean;

  TransitionStates                :Array[1..2] of Integer;
  LifetimeState                   :Integer;

  MakeMovie                       :Boolean;
  MakeMovieFrameDuration          :Double;
  MakeMovieSaveFrames             :Boolean;
end;


type


TQUB_Task_Simulate                = class(TQUB_Task)
protected


  prData                          :TQUB_QVO_QVD;
  prStimData                      :TQUB_QVO_QVD;
  prStimProto                     :TQUB_QVO_QVP;
  prModels                        :TStringList;
  prModelParameters               :TQUB_ExpCond_Parameters;

  //prKinetic                       :TKPL_Kinetic;

  //KPL Collections ************************************************************
  {
  prKPLConsoleCollection          :TKPL_Object;
  prKPLConsole                    :TKPL_Object;

  prKPLPulseCollection            :TKPL_Object;
  prKPLPulse                      :TKPL_Object;

  prKPLQUBInterfaceCollection     :TKPL_Object;
  prKPLQUBInterface               :TKPL_Object;
  }
  //****************************************************************************

  prProperties                    :TQUB_Task_Simulate_PropertiesRec;

  prCurSegmentIndex               :Integer;
  prCurSegmentCount               :Integer;

  prSignal                        :TSimpleEvent;
  prProtoToDataSignal             :TSimpleEvent;

  prState                         :TPMInteger;
  prProbs                         :TPMDouble;
  prModelCount                    :Integer;

  prTotalPoints                   :Double;
  prActualPoints                  :Double;
  prStTime                        :Comp;

  prBuffer                        :TPAInteger;

  prBufferIdl                     :TQUB_QFS_QIF_PAFirstLastClass;
  prBufferIdlCount                :Integer;
  prBufferIdlMaxCount             :Integer;
  prBufferIdlInc                  :Integer;
  prBufferStateIDL                :TPAInteger;
  prBufferIdlAmp                  :TPASingle;
  prBufferStateIDLAmp             :TPASingle;

  prTotalPointCount               :Integer;

  prNodeIdl                       :INode;

  prPars                          :Array[0..QUB_QFS_QMF_MAX_PARAMETER] of Pointer;
  prChan                          :Array[0..QUB_QFS_QMF_MAX_PARAMETER] of Integer;
  prIniV                          :Array[0..QUB_QFS_QMF_MAX_PARAMETER] of Double;

  prGatingScale                   :Double;

  prBaselineFluctStateCount       :Integer;
  prBaselineJumpsStateAmpF        :Double;
  prBaselineJumpsStateAmpL        :Double;
  prBaselineJumpsStateAmpC        :Double;
  prBaselineJumpsStateDur         :Double;
  prBaselineJumpsStateStart       :Integer;

  prBaseLineLGMState              :Array[0..1] of Double;

  prIDum                          :Integer;

  prJumpAmp                       :Double;
  prJumpCls                       :Integer;
  prJumpState                     :Integer;
  prJumpGrp                       :Integer;
  prJumpCnt                       :Array[-1..1] of Integer;
  prTotJumpCnt                    :Array[-1..1] of Integer;
  prTransitionCnt                 :Integer;
  prLifetimeTime                  :Double;
  prTotTransitionCnt              :Integer;
  prTotLifetimeTime               :Double;
  prTransitionCntAll              :TPAPointer;
  prLifetimeTimeAll               :TPAPointer;
  prTotTransitionCntAll           :TPAPointer;
  prTotLifetimeTimeAll            :TPAPointer;
  prJumpAvgSum                    :Double;
  prJumpStdSqrSum                 :Double;
  prJumpStdSqrN                   :Integer;
  prMeasStdSqrSum                 :Double;
  prStepIdx                       :Integer;
  prStepSizes                     :Array of TPADouble;
  prStepProbs                     :Array of TPADouble;
  prStepCounts                    :Array of Integer;
  prStepCount                     :Integer;


  // flux:  rate ptr, last value?
  // check up on:
  //    is file ok?  have extra channel?
  //    when does stimulus change -- update last value
  //    each macroscopic point, output value*pr[fluxFrom]
  prFluxRate                      :TQUB_QFS_Rate;
  prFluxRateDir                   :Integer; // 1 forward or 2 backward, as index to prFluxRate.Value[]
  prFluxRateValue                 :Double;
  prFluxBuffer                    :TPAInteger;
  prFluxScaling                   :Double;

  prStructureFiles                :TStringList;
  prMovieName                     :String;
  prFrameIdx                      :Integer;
  prMaxFrame                      :Integer;

  prRandSeedDwell                 :LongInt;
  prRandSeedNoise                 :LongInt;


  function prSetupFlux(ParCount, BufferCount :Integer): Boolean;

  procedure prDoSimulate_NonStimulated;
  procedure prDoSimulate_Stimulation_File;
  procedure prDoSimulate_Stimulation_File_Proto;
  procedure prProtoToDataFinished(Sender :TObject);
  procedure prDoSimulate_Stimulation_File_Data(BufferMPX :Pointer; ActualBufferCount :Integer);
  function  prDoSimulate_Stimulation_File_SetupPars: Integer;
  procedure prDoSimulate_Stimulation_File_Core(ParCount :Integer; FirstSeg :Integer; BufferMPX :Pointer; ActualBufferCount :Integer);
  //procedure prDoSimulate_Stimulation_Script;

  function prStochasticSimulationEngine(BufferMPX :Pointer; BufferCount :Integer) :Boolean;
  function prDeterministicSimulationEngine(BufferMPX :Pointer; BufferCount :Integer) :Boolean;

  procedure prSimulateBaseline(BufferCount :Integer; Sampling, Scaling, Offset :Double);

  //function prStaircaseMakeSteps :Boolean;
  //procedure prStaircaseFreeSteps;

  procedure prAppendIDL;
  procedure prPublishResultsPy;

  procedure prOnBeforeData_Stimulated(Sender :TObject; var BufferCount :Integer);
  procedure prOnAfterData_Stimulated(Sender :TObject; BufferMPX :Pointer; BufferSize :Integer; BufferCount :Integer; var Filled :Array of Boolean);

  procedure prOnEndRun_Stimulated(Sender :TObject; Failed :Boolean);

  procedure prSetData(Data :TQUB_QVO_QVD);
  procedure prSetStimData(Data :TQUB_QVO_QVD);
  procedure prSetStimProto(Proto :TQUB_QVO_QVP);

  procedure prMakeMovie;


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Execute; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  property Data                   :TQUB_QVO_QVD                     read prData                   write prSetData;
  property StimData               :TQUB_QVO_QVD                     read prStimData               write prSetStimData;
  property StimProto              :TQUB_QVO_QVP                     read prStimProto              write prSetStimProto;
  property Models                 :TStringList                      read prModels;
  property ModelParameters        :TQUB_ExpCond_Parameters          read prModelParameters        write prModelParameters;
  //property Kinetic                :TKPL_Kinetic                     read prKinetic;

  //Properties
  property Properties             :TQUB_Task_Simulate_PropertiesRec read prProperties             write prProperties;


end;


// can re-run a simulation, e.g. for LiveSim
IQUB_Task_Simulate_Stored = interface(IUnknown)
  function CanSimulate: Boolean;
  // if False, will never be able to run this sim again because a file has closed

  function ShouldSimulate: Boolean;
  // True if current model and data are involved in this sim

  function Simulate(OnTerminate :TNotifyEvent): TQUB_Task_Simulate;
end;

TQUB_Task_Simulate_Stored = class(TInterfacedObject, IQUB_Task_Simulate_Stored)
protected

  prRefCount                      :Integer;
  prModels                        :TStringList;
  prData                          :TQUB_QVO_QVD;
  prStimData                      :TQUB_QVO_QVD;
  prStimProto                     :TQUB_QVO_QVP;
  prModelParameters               :TQUB_ExpCond_Parameters;
  prProperties                    :TQUB_Task_Simulate_PropertiesRec;

public

  constructor Create(Task :TQUB_Task_Simulate);
  destructor Destroy; override;

  function CanSimulate: Boolean;
  // if False, will never be able to run this sim again because a file has closed

  function ShouldSimulate: Boolean;
  // True if current model and data are involved in this sim

  function Simulate(OnTerminate :TNotifyEvent): TQUB_Task_Simulate;

end;

TQUB_Task_SimLive  = class(TQUB_Task)
protected


  prNeedsUpdate         :Boolean;
  prSignal              :TEvent;
  prSimIsRunning        :Boolean; // other sim from user
  prUpdating            :Boolean; // our sim
  prSim                 :TQUB_Task_Simulate;
  prModel               :TQUB_QVO_QVM;


  procedure prClearEvents;

  procedure prOnSimDone(Sender :TObject);
  procedure prOnChangeModel(Sender :TObject);
  procedure prOnModelEdit(Sender :TObject);

  procedure prRunSim;
  function prSimRunning: Boolean;
  procedure prCheckSimRunning;


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Execute; override;
  procedure Stop; override;
  procedure Update;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


end;


implementation


uses


QUB_GUIInterface,
QUB_Task_Simulate_Properties,
QUB_QFS_QMF_QMatrix,
//KPL_CollectionConsole, KPL_CollectionQUBInterface,
QUB_Node_Python, QUB_Viewport_Python;


//******************************************************************************
function IDLOutputToString(x :TQUB_Task_Simulate_IDLOutput): String;
begin
  if x = OUT_STATE then begin
    Result                  := 'State';
  end else if x = OUT_CLASS then begin
    Result                  := 'Class';
  end else begin // if x = OUT_GROUP then begin
    Result                  := 'Group';
  end;
end;

function StringToIDLOutput(const x :String): TQUB_Task_Simulate_IDLOutput;
begin
  if x = 'State' then begin
    Result     := OUT_STATE;
  end else if x = 'Class' then begin
    Result     := OUT_CLASS;
  end else{if x = 'Group' then}begin
    Result     := OUT_GROUP;
  end;
end;


function StimulationToString(x :TQUB_Task_Simulate_Stimulation): String;
begin
  if x = SIM_STIM_NONE then begin
    Result     := 'None';
  end else if x = SIM_STIM_FILE then begin
    Result     := 'File';
  //end else{if x = SIM_STIM_SCRIPT then}begin
  //  Result     := 'Script';
  end;
end;

function StringToStimulation(const x :String): TQUB_Task_Simulate_Stimulation;
begin
  if x = 'None' then begin
    Result     := SIM_STIM_NONE;
  end else if x = 'File' then begin
    Result     := SIM_STIM_FILE;
  //end else{if x = 'Script' then}begin
  //  Result     := SIM_STIM_SCRIPT;
  end;
end;

function ResponseToString(x :TQUB_Task_Simulate_Response): String;
begin
  if x = SIM_RESPONSE_STOCHASTIC then begin
    Result     := 'Stochastic';
  end else{if x = SIM_RESPONSE_DETERMINISTIC then}begin
    Result     := 'Deterministic';
  end;
end;

function StringToResponse(const x :String): TQUB_Task_Simulate_Response;
begin
  if x = 'Stochastic' then begin
    Result     := SIM_RESPONSE_STOCHASTIC;
  end else{if x = 'Deterministic' then}begin
    Result     := SIM_RESPONSE_DETERMINISTIC;
  end;
end;


//******************************************************************************
constructor TQUB_Task_Simulate.Create(Suspended :Boolean);
begin
  inherited;

  Priority                := tpNormal;

  prName                  := 'Simulate data';

  prSignal                := TSimpleEvent.Create;

  prModels                := TStringList.Create;

  {
  //Collections ****************************************************************
  prKPLConsoleCollection  := TKPL_Object.Create('Collection', nil);
  prKPLConsole            := TKPL_Collection_Console.Create('console', prKPLConsoleCollection);

  prKPLPulseCollection    := TKPL_Object.Create('Collection', nil);
  prKPLPulse              := TKPL_Collection_QUB_Pulse.Create('pulse', prKPLPulseCollection);

  prKPLQUBInterfaceCollection := TKPL_Object.Create('Collection', nil);
  prKPLQUBInterface           := TKPL_Collection_QUB_Interface.Create('qubinterface', prKPLQUBInterfaceCollection);
  //****************************************************************************
  prKinetic               := TKPL_Kinetic.Create('kinetic', nil);
  prKinetic.Register(prKPLConsoleCollection);
  prKinetic.Register(prKPLPulseCollection);
  prKinetic.Register(prKPLQUBInterfaceCollection);
  }
  prStructureFiles        := TStringList.Create;
  prMaxFrame              := 5000;
end;

destructor TQUB_Task_Simulate.Destroy;
begin
  try

    prModels.Free;
    {
    prKinetic.Free;

    prKPLQUBInterfaceCollection.Free;
    prKPLPulseCollection.Free;
    prKPLConsoleCollection.Free;
    }
    prSignal.Free;

    prStructureFiles.Free;

  finally

    inherited;

  end;
end;

procedure TQUB_Task_Simulate.Execute;
var
  m, i, j :Integer;
begin

  //Lock
  for i := 0 to prModels.Count - 1 do begin
    if prModels.Objects[i] <> nil then begin
      TQUB_QVO_QVM(prModels.Objects[i]).ExpCond := ModelParameters;
      TQUB_QVO_QVM(prModels.Objects[i]).LockWrite;
    end;
  end;
  if Data <> nil then Data.LockAny;
  if StimData <> nil then StimData.LockAny;

  {
  if Properties.Staircase then begin
    if Properties.StaircaseUseSteps then begin
      if prStaircaseMakeSteps then begin
        Terminate;
      end;
    end;
  end;
  }
  if not Terminated then begin
    prTransitionCntAll    := PAlloc1(prModels.Count);
    prTotTransitionCntAll := PAlloc1(prModels.Count);
    prLifetimeTimeAll     := PAlloc1(prModels.Count);
    prTotLifetimeTimeAll  := PAlloc1(prModels.Count);
    for m := 0 to prModels.Count - 1 do begin
      prTransitionCntAll[m]    := IAlloc2(TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount, TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount);
      prTotTransitionCntAll[m] := IAlloc2(TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount, TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount);
      prLifetimeTimeAll[m]     := DAlloc1(TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount);
      prTotLifetimeTimeAll[m]  := DAlloc1(TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount);
      for i := 0 to TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount - 1 do begin
        for j := 0 to TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.States.StateCount - 1 do begin
          TPMInteger(prTotTransitionCntAll[m])[i, j] := 0;
        end;
        TPADouble(prTotLifetimeTimeAll[m])[i]        := 0;
      end;
    end;
  end else begin
    prTransitionCntAll     := nil;
    prTotTransitionCntAll  := nil;
    prLifetimeTimeAll      := nil;
    prTotLifetimeTimeAll   := nil;
  end;

  try

    if Terminated then Exit;

    prStatus  := 'Initializing';
    prPercent := 0;
    prETA     := '0:00:00';

    Synchronize(prDoOnChange);

    //TO DO: LM. Check this. I think it's about the data to be simulated
    //Chris sez: it appears to only be used to put a bogus (length -1) event in the idealization.
    //           then the bogus event is overwritten by the first actual event.
    prTotalPointCount := 0;
    //Modified by Chris & John:
    //prTotalPointCount := prData.DataObject.Segments.TotalDataCount;


    // store for later use by Mac
    if prProperties.ILeak then begin
      if prData.ExpCond.ParameterByName['GLeak'] = nil then begin
        prData.ExpCond.AddPar('GLeak', prProperties.GLeak, -1, False);
      end else begin
        prData.ExpCond.ChangePar('GLeak', prProperties.GLeak);
      end;
      if prData.ExpCond.ParameterByName['GLeak'] = nil then begin
        prData.ExpCond.AddPar('GLeak', prProperties.GLeak, -1, False);
      end else begin
        prData.ExpCond.ChangePar('GLeak', prProperties.GLeak);
      end;
    end;

    try

      //**********************************************************************//
      //                                                                      //
      //                      N O N - S T I M U L A T E D                     //
      //                                                                      //
      //**********************************************************************//
      if Properties.Stimulation = SIM_STIM_NONE then begin


        prDoSimulate_NonStimulated;


      //**********************************************************************//
      //                                                                      //
      //                         S T I M U L A T E D                          //
      //                                                                      //
      //                               F I L E                                //
      //                                                                      //
      //**********************************************************************//
      end else if Properties.Stimulation = SIM_STIM_FILE then begin


        prDoSimulate_Stimulation_File;


      //**********************************************************************//
      //                                                                      //
      //                         S T I M U L A T E D                          //
      //                                                                      //
      //                             S C R I P T                              //
      //                                                                      //
      //**********************************************************************//
      //end else if Properties.Stimulation = SIM_STIM_SCRIPT then begin


      //  prDoSimulate_Stimulation_Script;


      end;

    except
    end;

    Beep;

    //Terminate;

    prStatus  := 'Terminated';
    prPercent := 0;
    prETA     := '0:00:00';

    Synchronize(prDoOnChange);


  finally


    for m := 0 to prModels.Count - 1 do begin
      if prTransitionCntAll <> nil      then Free2(TPMChar(prTransitionCntAll[m]));
      if prTotTransitionCntAll <> nil   then Free2(TPMChar(prTotTransitionCntAll[m]));
      if prLifetimeTimeAll <> nil       then FreeMem(prLifetimeTimeAll[m]);
      if prTotLifetimeTimeAll <> nil    then FreeMem(prTotLifetimeTimeAll[m]);
    end;
    if prTransitionCntAll <> nil        then FreeMem(prTransitionCntAll);
    if prTotTransitionCntAll <> nil     then FreeMem(prTotTransitionCntAll);
    if prLifetimeTimeAll <> nil         then FreeMem(prLifetimeTimeAll);
    if prTotLifetimeTimeAll <> nil      then FreeMem(prTotLifetimeTimeAll);
    {
    if Properties.Staircase then begin
      prStaircaseFreeSteps;
    end;
    }
    //Unlock
    if Data <> nil then Data.UnLockAny;
    if StimData <> nil then StimData.UnLockAny;
    for i := 0 to prModels.Count - 1 do begin
      if prModels.Objects[i] <> nil then TQUB_QVO_QVM(prModels.Objects[i]).UnLockWrite;
    end;
    {
    if prKPLPulse <> nil then begin
      with prKPLPulse as TKPL_Collection_QUB_Pulse do begin
        OnBeforeData := nil;
        PulseToData.OnAfterData  := nil;
      end;
    end;

    if Kinetic <> nil then if Kinetic.Running then Kinetic.Stop;
    }
  end;

end;

(*
function TQUB_Task_Simulate.prStaircaseMakeSteps :Boolean;
var
  Parser :TKPL_Parser;
  Sign   :Integer;
  Step   :Double;
  Prob   :Double;
  Sum    :Double;
  i, s   :Integer;

  procedure NewStepDistribution;
  begin
    Inc(prStepCount);
    ReAllocMem(prStepSizes,  SizeOf(TPADouble)*prStepCount);
    ReAllocMem(prStepProbs,  SizeOf(TPADouble)*prStepCount);
    ReAllocMem(prStepCounts, SizeOf(Integer)  *prStepCount);
    prStepSizes[prStepCount - 1] := nil;
    prStepProbs[prStepCount - 1] := nil;
    prStepCounts[prStepCount - 1] := 0;
  end;

  procedure NewStep;
  begin
    Inc(prStepCounts[prStepCount - 1]);
    ReAllocMem(prStepSizes[prStepCount - 1], SizeOf(Double) * prStepCounts[prStepCount - 1]);
    ReAllocMem(prStepProbs[prStepCount - 1], SizeOf(Double) * prStepCounts[prStepCount - 1]);
    prStepSizes[prStepCount - 1][prStepCounts[prStepCount - 1] - 1] := Step;
    prStepProbs[prStepCount - 1][prStepCounts[prStepCount - 1] - 1] := Prob;
  end;

begin

  Result := False;

  Parser := nil;

  try

    try

      Parser               := TKPL_Parser.Create;
      Parser.ReturnComment := False;
      Parser.ReturnTab     := False;
      Parser.ReturnNewLine := False;
      Parser.Start;

      prStepCount := 0;

      NewStepDistribution;

      repeat

        Result := Parser.Next(Properties.StaircaseSteps);
        if Result then Break;

        Sign := +1;
        Step := 1;
        Prob := 1;
        Result := True;
        if Parser.Token.Token = 's' then begin
          Parser.Next(Properties.StaircaseSteps);
          if Parser.Token.Token = '=' then begin
            Parser.Next(Properties.StaircaseSteps);
            if Parser.Token.Token = '-' then begin
              Sign := -1;
              Parser.Next(Properties.StaircaseSteps);
            end;
            if Parser.Token.TokenType = TOKEN_NUMBER then begin
              //Step
              Step := Sign * Parser.Token.TokenNumber;
              Parser.Next(Properties.StaircaseSteps);
              if Parser.Token.Token = ',' then begin
                Parser.Next(Properties.StaircaseSteps);
                if Parser.Token.Token = 'p' then begin
                  Parser.Next(Properties.StaircaseSteps);
                  if Parser.Token.Token = '=' then begin
                    Parser.Next(Properties.StaircaseSteps);
                    if Parser.Token.TokenType = TOKEN_NUMBER then begin
                      Prob := Parser.Token.TokenNumber;
                      NewStep;
                      Parser.Next(Properties.StaircaseSteps);
                      if Parser.Token.TokenType = TOKEN_NULL then begin
                      //Done
                        Result := False;
                        Break;
                      end;
                      if Parser.Token.Token = ';' then begin
                        NewStepDistribution;
                        Result := False;
                      end else if Parser.Token.Token = ',' then begin
                        Result := False;
                      end else begin
                      //Error
                        Break;
                      end;
                    end;//Num
                  end;//=
                end;//p
              end;//,
            end;//Num
          end;//=
        end;//s

      until(Parser.Token.TokenType = TOKEN_NULL);

      for s := 0 to prStepCount - 1 do begin
        Sum := 0;
        for i := 0 to prStepCounts[s] - 1 do begin
          Sum := Sum + Abs(prStepProbs[s][i]);
        end;//i
        if Sum = 0 then begin
          for i := 0 to prStepCounts[s] - 1 do begin
            prStepProbs[s][i] := 1 / prStepCounts[s];
          end;//i
        end else begin
          for i := 0 to prStepCounts[s] - 1 do begin
            prStepProbs[s][i] := Abs(prStepProbs[s][i]) / Sum;
          end;//i
        end;
      end;//s

    except

      Result := True;
      
    end;

  finally

    Parser.Free;

    if Result then begin
      //Error!!!
    end;

  end;

end;

procedure TQUB_Task_Simulate.prStaircaseFreeSteps;
var
  s :Integer;
begin

  for s := 0 to prStepCount - 1 do begin
    FreeMem(prStepSizes[s]);
    prStepSizes[s] := nil;
    FreeMem(prStepProbs[s]);
    prStepProbs[s] := nil;
  end;
  FreeMem(prStepSizes);
  prStepSizes := nil;
  FreeMem(prStepProbs);
  prStepProbs := nil;
  FreeMem(prStepCounts);
  prStepCounts := nil;

  prStepCount := 0;

end;
*)
procedure TQUB_Task_Simulate.prAppendIDL;
var
  Sel             :TQUB_FirstLast;
  Amps            :TPASingle;
  Stds            :TPASingle;
  AmpCount        :Integer;
  MinCls          :Integer;
  MaxCls          :Integer;
  i, p            :Integer;

  dsNode,
  segNode,
  ampNode,
  stdNode,
  ixNode,
  durNode,
  fNode, lNode    :INode;
  ixData          :TPAInteger;
  durData         :TPASingle;
  fData, lData    :TPAInteger;
begin

  //This function is called for every segment

  if not Properties.IdealizedData then Exit;

  while (prBufferIDLCount > 0) do begin

    if prBufferIDL[prBufferIDLCount - 1].Last < prBufferIDL[prBufferIDLCount - 1].First then begin
      Dec(prBufferIDLCount);
    end else begin
      Break;
    end;

  end;

  if prBufferIDLCount <= 0 then Exit;

  Amps := nil;
  Stds := nil;

  if Properties.Staircase then begin
    //Find Min/Max Class/State/Group
    MinCls := prBufferIDL[0].Cls;
    MaxCls := prBufferIDL[0].Cls;
    for i := 0 to prBufferIDLCount - 1 do begin
      if MinCls > prBufferIDL[i].Cls then MinCls := prBufferIDL[i].Cls;
      if MaxCls < prBufferIDL[i].Cls then MaxCls := prBufferIDL[i].Cls;
    end;
    //Subtract MinCls, otherwise may get negative classes
    for i := 0 to prBufferIDLCount - 1 do begin
      prBufferIDL[i].Cls := prBufferIDL[i].Cls - MinCls;
    end;
    AmpCount := MaxCls - MinCls + 1;//prBufferIDL[prBufferIDLCount - 1].Cls + 1;
    if AmpCount = 0 then AmpCount := QUB_QFS_QMF_DEF_NCLASS;
  end else begin
    AmpCount := QUB_QFS_QMF_DEF_NCLASS;
  end;

  try

    Sel.First := prBufferIDL[0].First;
    Sel.Last  := prBufferIDL[prBufferIDLCount - 1].Last;

    Amps := SAlloc1(AmpCount);
    Stds := SAlloc1(AmpCount);

    if Properties.Staircase then begin

      for i := 0 to AmpCount - 1 do begin
        //TODO: check this
        if          Properties.IDLOutput = Integer(OUT_STATE) then begin
          Amps[i] := TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[prBufferIDL[i].Cls] * (TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[1] - TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[0]);
        end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
          Amps[i] := i * (TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[1] - TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[0]);
        end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
          Amps[i] := TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[prBufferIDL[i].Cls] * (TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[1] - TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[0]);
        end;

        //Testing:
        Amps[i] := 0;
        for p := 0 to prBufferIDLCount - 1 do begin
          if prBufferIDL[p].Cls = i then begin
            Amps[i] := prBufferIDLAmp[p];
            Break;
          end;
        end;

        Stds[i] := TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Std[0];
      end;

    end else begin

      for i := 0 to AmpCount - 1 do begin
        Amps[i] := TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Amp[i];
        Stds[i] := TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.Std[i];
      end;

    end;

    dsNode  := TNode.Create('DataSet');
    dsNode['sampling'].SetData( TQUB_QVO_QVD(Data).Sampling * 1.0e3 );

    segNode := dsNode.AppendChild('Segment');
    segNode.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
    segNode.Data[0] := Sel.First;
    segNode.Data[1] := Sel.Last;

    ampNode := segNode['amp'];
    stdNode := segNode['sd'];
    ampNode.SetupNumData( NODE_DATATYPE_DOUBLE, AmpCount, 1 );
    stdNode.SetupNumData( NODE_DATATYPE_DOUBLE, AmpCount, 1 );
    for i := 0 to AmpCount - 1 do begin
      ampNode.Data[i] := Amps[i];
      stdNode.Data[i] := Stds[i];
    end;

    //TODO John, this has to be checked
    if          Properties.IDLOutput = Integer(OUT_STATE) then begin
      ixNode  := segNode['Classes'];//segNode['States'];
      durNode := segNode['Durations'];//segNode['StateDurations'];
    end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
      ixNode  := segNode['Classes'];
      durNode := segNode['Durations'];
    end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
      ixNode  := segNode['Classes'];//segNode['Groups'];
      durNode := segNode['Durations'];//segNode['GroupDurations'];
    end;
    segNode['DwellCount'].SetData(prBufferIDLCount);
    fNode     := segNode['Firsts'];
    lNode     := segNode['Lasts'];

    segNode['start'].SetData( Data.DataObject.Segments[Data.DataObject.Segments.SegmentCount - 1].StartTime * 1.0e3 );

    ixNode.SetupNumData(NODE_DATATYPE_LONGINT, prBufferIDLCount, 1);
    durNode.SetupNumData(NODE_DATATYPE_SINGLE, prBufferIDLCount, 1);
    fNode.SetupNumData(NODE_DATATYPE_LONGINT, prBufferIDLCount, 1);
    lNode.SetupNumData(NODE_DATATYPE_LONGINT, prBufferIDLCount, 1);
    ixData  := TPAInteger(ixNode.DataPtr);
    durData := TPASingle(durNode.DataPtr);
    fData  := TPAInteger(fNode.DataPtr);
    lData  := TPAInteger(lNode.DataPtr);
    for i := 0 to prBufferIDLCount - 1 do begin
      ixData[i] := prBufferIDL[i].Cls;
      durData[i] := (prBufferIDL[i].Last - prBufferIDL[i].First + 1) * prData.Sampling * 1.0e3;
      fData[i] := prBufferIDL[i].First;
      lData[i] := prBufferIDL[i].Last;
    end;

    Data.DataObject.Idealization[0].SetDwells(dsNode);
    Data.DrawIdealized := True;

    if prNodeIdl <> nil then begin
      prNodeIdl['DataSet'].AppendChild(segNode);
    end;


  finally

    //Testing
    //if not ((Properties.MakeMovie) and (Properties.IdealizedData) and (Properties.Stimulation = SIM_STIM_SCRIPT)) then begin
    //  prBufferIDLCount := 0;
    //end;

    FreeMem(Amps);
    FreeMem(Stds);

  end;
end;

procedure TQUB_Task_Simulate.prPublishResultsPy;
begin
  PyNode_Publish( prNodeIdl, 'simResult' );
end;


function TQUB_Task_Simulate.prSetupFlux(ParCount, BufferCount :Integer): Boolean;
var
  ThisModel      :TQUB_QVO_QVM;
  c              :Integer;
begin
  Result := False;
  try
    if (Properties.Response = SIM_RESPONSE_DETERMINISTIC) and (Properties.Flux) then begin
      if (ParCount + 1 + 1) < (Data.DataObject.ADChannelCount) then begin
        prWarning    := True;
        prWarningMsg := prWarningMsg + 'Data file has more data channels than necessary.'#13#10;
        prWikiTopic  := 'Warning:Unnecessary_Channels';
      end else if (ParCount + 1 + 1) > (Data.DataObject.ADChannelCount) then begin
        prError      := True;
        prErrorMsg   := 'Data file has not enough data channels for stimulus, response and flux';
        prWikiTopic  := 'Error:Not_Enough_Channels';
        Exit;
      end;

      ThisModel      := TQUB_QVO_QVM(prModels.Objects[0]);
      if (prProperties.FluxFrom <= 0) or (prProperties.FluxFrom > ThisModel.ModelObject.States.StateCount)
      or (prProperties.FluxTo <= 0)   or (prProperties.FluxTo > ThisModel.ModelObject.States.StateCount) then begin
        prError      := True;
        prErrorMsg   := Format('Invalid state numbers (Flux %d -> %d)', [prProperties.FluxFrom, prProperties.FluxTo]);
        prWikiTopic  := 'Error:Invalid_Flux_States';
        Exit;
      end;

      with ThisModel.ModelObject do begin
        prFluxRate   := Rates.RateOfStates[ States.State[prProperties.FluxFrom-1], States.State[prProperties.FluxTo-1] ];
      end;
      if prFluxRate = nil then begin
        prError      := True;
        prErrorMsg   := Format('No rate connecting states (Flux %d -> %d)', [prProperties.FluxFrom, prProperties.FluxTo]);
        prWikiTopic  := 'Error:Invalid_Flux_States';
        Exit;
      end;

      if prFluxRate.StateIndex[1] = (prProperties.FluxFrom - 1) then begin
        prFluxRateDir := 1;
      end else begin
        prFluxRateDir := 2;
      end;

      prFluxRateValue := prFluxRate.Value[prFluxRateDir];
      prFluxBuffer  := IAlloc1(BufferCount);

      c             := Data.DataObject.ADChannelCount - 1;
      Data.DataChannelName[c]    := Format('Flux %d -> %d', [Properties.FluxFrom, Properties.FluxTo]);
      Data.DataChannelUnits[c]   := '1/s';
      // scaling: big enough for 10*current rate value?
      // or maybe we just make the user set it then re-sim if it doesn't work out
      // Data.DataChannelScaling[c] := MaxInt / (Data.DataObject.Scaling * prFluxRateValue * 10);
      prFluxScaling              := Data.DataObject.Scaling * Data.DataChannelScaling[c];
    end else begin
      prFluxRate      := nil;
      prFluxBuffer    := nil;
    end;
  finally
    Result          := prError;
  end;
end;


//****************************************************************************//
//                                                                            //
//                          N O N - S T I M U L A T E D                       //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Simulate.prDoSimulate_NonStimulated;
var
  ThisModel        :TQUB_QVO_QVM;
  BufferMPX        :Pointer;
  BufferCount      :Integer;
  s, m, c, i, j    :Integer;
  psum             :Double;
  Seg              :TQUB_QFS_Segment;
  StartTime        :Double;
  t, tr            :String;
begin

  if Properties.Randomize then begin
    Randomize;
  end else begin
    RandSeed := Properties.RandomSeed;
  end;
  prRandSeedDwell := RandSeed;
  prRandSeedNoise := RandSeed;

  prIDum := Round(Random(High(Integer)));

  //Allocate Buffer
  if Properties.SegmentSize < ONE_MEGABYTE then begin
    BufferCount       := Properties.SegmentSize;
  end else begin
    BufferCount       := ONE_MEGABYTE;
  end;
  prBuffer            := IAlloc1(BufferCount);
  BufferMPX           := AllocMem(Data.DataObject.ADChannelCount * Data.DataObject.ADDataSize * BufferCount);
  prBufferStateIDL    := IAlloc1(BufferCount);
  prBufferStateIDLAmp := SAlloc1(BufferCount);

  prBufferIDLInc := BufferCount div 100;
  if prBufferIDLInc < 1000 then prBufferIDLInc := 1000;
  prBufferIDLCount := 0;
  prBufferIDLMaxCount := 0;

  prNodeIdl        := TNode.Create('simuResult');
  prNodeIdl['DataSet']['sampling'].SetData(prData.Sampling * 1.0e3);
  prNodeIdl['DataSet']['FileName'].SetData(prData.FileName);

  //The State of each channel of each model is memorized
  prModelCount := prModels.Count;
  TPAPointer(prState) := PAlloc1(prModelCount);
  TPAPointer(prProbs) := PAlloc1(prModelCount);
  prTotalPoints  := 0;
  prActualPoints := 0;

  for m := 0 to prModelCount - 1 do begin

    ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);


    //Do it only once, here; then only re-calculate Q and P
    ThisModel.ModelObject.UpdateParameters;
    ThisModel.ModelObject.Adt := prData.Sampling;
    ThisModel.ModelObject.CalculateMath;


    psum :=        Properties.SegmentCount;
    psum := psum * Properties.SegmentSize;
    psum := psum * ThisModel.ModelObject.ChannelCount;
    prTotalPoints := prTotalPoints + psum;

    prState[m] := IAlloc1(ThisModel.ModelObject.ChannelCount);
    for c := 0 to ThisModel.ModelObject.ChannelCount - 1 do begin
      prState[m, c] := 0;
    end;//Channel

    if ThisModel.ModelObject.States.StateCount > 0 then prProbs[m] := DAlloc1(ThisModel.ModelObject.States.StateCount) else prProbs[m] := nil;
    for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
      prProbs[m, c] := 1 / ThisModel.ModelObject.States.StateCount;
    end;//State

  end;//Model

  
  prData.ExpCond.ReadTree(ModelParameters.ToTree);
  for m := 0 to prData.ExpCond.Count - 1 do begin
    prData.ExpCond[m].Channel := -1;
  end;

  prData.DataChannelName[0]   := 'Current';
  prData.DataChannelUnits[0]  := 'pA';

  // flux
  if prSetupFlux(0{ParCount}, BufferCount) then Exit;


  prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
  prStatus := 'Running';

  Synchronize(prDoOnChange);

  try


    if Properties.Staircase then begin
      prConsoleMsg  := 'Seg'#9'JumpAvg       '#9'JumpStd       '#9'Jf[s-1]'#9'Jb[s-1]';
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);
    end;

    //Segment Loop
    for s := 0 to Properties.SegmentCount - 1 do begin

      //Initialize / Reset everything
      prCurSegmentIndex := s;
      prCurSegmentCount := 0;

      //Initialize baseline state
      prBaselineFluctStateCount := 0;
      prBaselineJumpsStateAmpF  := 0;
      prBaselineJumpsStateAmpL  := 0;
      prBaselineJumpsStateAmpC  := 0;
      prBaselineJumpsStateDur   := ExpDev * Properties.BaselineJumpsLifetime;
      prBaselineJumpsStateStart := 0;
      prBaselineLGMState[0]     := Properties.BaselineLGMX0;
      prBaselineLGMState[1]     := Properties.BaselineLGMX0p;

      //Reset JumpAmp
      prJumpAmp                 := 0;
      prJumpCls                 := 0;
      prJumpState               := 0;
      prJumpGrp                 := 0;
      prJumpCnt[-1]             := 0;
      prJumpCnt[+1]             := 0;
      prTransitionCnt           := 0;
      prLifetimeTime            := 0;
      for m := 0 to prModels.Count - 1 do begin
        ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
        for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            TPMInteger(prTransitionCntAll[m])[i, j] := 0;
          end;
          TPADouble(prLifetimeTimeAll[m])[i]        := 0;
        end;
      end;
      prJumpAvgSum              := 0;
      prJumpStdSqrSum           := 0;
      prJumpStdSqrN             := 0;
      prStepIdx                 := 0;//The index of the distribution. When unequal steps alternate, StepIdx cycles through step distributions. 



      //Initialize the state or state probabilities
      for m := 0 to prModels.Count - 1 do begin

        ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);

        if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
        //Stochastic

          for c := 0 to ThisModel.ModelObject.ChannelCount - 1 do begin
            prState[m, c] := ThisModel.ModelObject.GenerateState;
          end;//Channel

          //Idealized
          if Properties.IdealizedData then begin
            if m = 0 then begin
              Inc(prBufferIDLCount);
              if prBufferIDLCount > prBufferIDLMaxCount then begin
                Inc(prBufferIDLMaxCount, prBufferIDLInc);
                try
                  ReAllocMem(prBufferIDL,    SizeOf(TQUB_QFS_QIF_FirstLastClass) * prBufferIDLMaxCount);
                  ReAllocMem(prBufferIDLAmp, SizeOf(Single)                      * prBufferIDLMaxCount);
                except
                  prProperties.IdealizedData := False;
                  FreeMem(prBufferIDL);
                  prBufferIDL := nil;
                  FreeMem(prBufferIDLAmp);
                  prBufferIDLAmp := nil;
                  prBufferIDLCount := 0;
                  prBufferIDLMaxCount := 0;
                end;
              end;
              prBufferIdl[prBufferIDLCount - 1].First := prTotalPointCount;
              prBufferIdl[prBufferIDLCount - 1].Last  := prBufferIdl[prBufferIDLCount - 1].First - 1;
              if Properties.Staircase then begin
                //TODO: Check this
                if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                  {???}prBufferIdl[prBufferIDLCount - 1].Cls := prJumpState;//prState[0, 0];                                 //TO DO
                end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls   := prJumpCls;                                                      //TO DO
                end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                  {???}prBufferIdl[prBufferIDLCount - 1].Cls := prJumpGrp;//ThisModel.ModelObject.States[prState[0, 0]].Gr;  //TO DO
                end;
                prBufferIdlAmp[prBufferIDLCount - 1] := prJumpAmp;
              end else begin
                if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := prState[0, 0];
                end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := ThisModel.ModelObject.States[prState[0, 0]].Cls;
                end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := ThisModel.ModelObject.States[prState[0, 0]].Gr;
                end;
              end;
            end;
          end;


        end else begin
        //Deterministic

          psum := 0;
          for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            psum := psum + ThisModel.ModelObject.States[c].Pr;
          end;//State
          for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            if psum <> 0 then begin
              prProbs[m, c] := ThisModel.ModelObject.States[c].Pr / psum;
            end else begin
              prProbs[m, c] := 1 / ThisModel.ModelObject.States.StateCount;
            end;
          end;//State

        end;

      end;//Model


      if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
      //Stochastic
        if Properties.Staircase then begin
          //TODO: Check this
          if          Properties.IDLOutput = Integer(OUT_STATE) then begin
            {???}prBufferStateIDL[0] := prJumpState;//prState[0, 0];
          end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
            prBufferStateIDL[0] := prJumpCls;
          end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
            {???}prBufferStateIDL[0] := prJumpGrp;//TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.States[prState[0, 0]].Gr;
          end;
          prBufferStateIDLAmp[0] := prJumpAmp;
        end else begin
          if          Properties.IDLOutput = Integer(OUT_STATE) then begin
            prBufferStateIDL[0] := prState[0, 0];
          end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
            prBufferStateIDL[0] := TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.States[prState[0, 0]].Cls;
          end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
            prBufferStateIDL[0] := TQUB_QVO_QVM(prModels.Objects[0]).ModelObject.States[prState[0, 0]].Gr;
          end;
        end;
      end;

      //Loop to generate all data in a segment - it works with 1 MP chunks
      repeat

        if (Properties.SegmentSize - prCurSegmentCount) > ONE_MEGABYTE then begin
          BufferCount := ONE_MEGABYTE;
        end else begin
          BufferCount := Properties.SegmentSize - prCurSegmentCount;
        end;

        if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
        //Stochastic

          if prStochasticSimulationEngine(BufferMPX, BufferCount) then Exit;

        end else begin
        //Deterministic

          if prDeterministicSimulationEngine(BufferMPX, BufferCount) then Exit;

        end;

        if prStopped then Exit;

        if prData.DataObject.Segments.TotalDataCount = 0 then begin
          StartTime := 0;
        end else if prCurSegmentCount = 0 then begin
          Seg := prData.DataObject.Segments[prData.DataObject.Segments.SegmentCount - 1];
          StartTime := Seg.StartTime + Seg.Channels.DataCount * prData.Sampling;
        end else begin
          StartTime := 0;
        end;

        prData.AppendData(BufferMPX, Data.DataObject.ADChannelCount * Data.DataObject.ADDataSize, BufferCount, prCurSegmentCount = 0{NewSegment}, StartTime, False{CanRefresh - Doit}, False{Acq});



        //New:
        //Make Movie ***********************************************************
        if (Properties.MakeMovie) and (Properties.IdealizedData) then begin
          Synchronize(prMakeMovie);
        end;
        //**********************************************************************


        Inc(prCurSegmentCount, BufferCount);
        Inc(prTotalPointCount, BufferCount);


      until(prCurSegmentCount >= Properties.SegmentSize);


      prAppendIdl;


      //Print true rates
      if Properties.Staircase then begin

        Inc(prTotJumpCnt[-1], prJumpCnt[-1]);
        Inc(prTotJumpCnt[+1], prJumpCnt[+1]);
        Inc(prTotTransitionCnt, prTransitionCnt);
        prTotLifetimeTime := prTotLifetimeTime + prLifetimeTime;
        for m := 0 to prModels.Count - 1 do begin
          ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
          for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
              TPMInteger(prTotTransitionCntAll[m])[i, j] := TPMInteger(prTotTransitionCntAll[m])[i, j] + TPMInteger(prTransitionCntAll[m])[i, j];
            end;
            TPADouble(prTotLifetimeTimeAll[m])[i] := TPADouble(prTotLifetimeTimeAll[m])[i] + TPADouble(prLifetimeTimeAll[m])[i];
          end;
        end;
        //prConsoleMsg  := Format('%d'#9'%d'#13#10, [prCurSegmentIndex + 1, prJumpCnt]);
        //prConsoleCode := MESSAGE_MESSAGE;
        //Synchronize(prWriteNewConsole);
        if (Properties.TransitionStates[1] <> 0) and (Properties.TransitionStates[2] <> 0) then begin
          prConsoleMsg  := Format('%d'#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.6g', [prCurSegmentIndex, prJumpAvgSum/prJumpStdSqrN, Sqrt(prJumpStdSqrSum/prJumpStdSqrN), prJumpCnt[+1]/((prCurSegmentCount - 1)*Data.Sampling), prJumpCnt[-1]/((prCurSegmentCount - 1)*Data.Sampling)]);
          prConsoleCode := MESSAGE_MESSAGE;
          Synchronize(prWriteNewConsole);
        end;

      end else begin

        Inc(prTotTransitionCnt, prTransitionCnt);
        prTotLifetimeTime := prTotLifetimeTime + prLifetimeTime;
        for m := 0 to prModels.Count - 1 do begin
          ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
          for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
              TPMInteger(prTotTransitionCntAll[m])[i, j] := TPMInteger(prTotTransitionCntAll[m])[i, j] + TPMInteger(prTransitionCntAll[m])[i, j];
            end;
            TPADouble(prTotLifetimeTimeAll[m])[i] := TPADouble(prTotLifetimeTimeAll[m])[i] + TPADouble(prLifetimeTimeAll[m])[i];
          end;
        end;
        if (Properties.TransitionStates[1] <> 0) and (Properties.TransitionStates[2] <> 0) then begin
          prConsoleMsg  := Format('%d'#9'%d'#9'%-14.6g', [prCurSegmentIndex, prTransitionCnt, prLifetimeTime]);
          t := Format('%d'#9, [prCurSegmentIndex]);
          for m := 0 to prModels.Count - 1 do begin
            ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
            for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
              for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
                if (i <> j) then begin
                  t := t + Format('%d'#9, [TPMInteger(prTransitionCntAll[m])[i, j]]);
                end;
              end;
              t := t + Format('%-14.6g'#9, [TPADouble(prLifetimeTimeAll[m])[i]]);
            end;
          end;
          prConsoleMsg  := prConsoleMsg + #13#10 + t;
          prConsoleCode := MESSAGE_MESSAGE;
          Synchronize(prWriteNewConsole);
        end;

      end;

    end;//of Segment


  finally


    if Properties.Staircase then begin
      if (Properties.TransitionStates[1] <> 0) and (Properties.TransitionStates[2] <> 0) then begin
        prConsoleMsg  := Format('Jumps (left/right): %d'#9'%d'#13#10'Transitions %d-->%d: %d, Prob(%d): %-14.6g'#13#10, [prTotJumpCnt[-1], prTotJumpCnt[+1], Properties.TransitionStates[1], Properties.TransitionStates[2], prTotTransitionCnt, Properties.TransitionStates[1], prTotLifetimeTime / (prTotalPointCount * Data.Sampling)]);
        prConsoleCode := MESSAGE_MESSAGE;
        Synchronize(prWriteNewConsole);
      end;
    end else begin
      if (Properties.TransitionStates[1] <> 0) and (Properties.TransitionStates[2] <> 0) then begin
        prConsoleMsg  := Format('Transitions %d-->%d: %d, Prob(%d): %-14.6g'#13#10, [Properties.TransitionStates[1], Properties.TransitionStates[2], prTotTransitionCnt, Properties.TransitionStates[1], prTotLifetimeTime / (prTotalPointCount * Data.Sampling)]);
        prConsoleCode := MESSAGE_MESSAGE;
        Synchronize(prWriteNewConsole);
      end;
      t := '';
      for m := 0 to prModels.Count - 1 do begin
        ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
        for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            if (i <> j) and ThisModel.ModelObject.Connected(ThisModel.ModelObject.States[i], ThisModel.ModelObject.States[j]) then begin
              t := t + Format('k(%d, %d)'#9, [i, j]);
            end;
          end;
          t := t + Format('p(%d)'#9, [i]);
        end;
      end;
      t := t + #13#10;
      tr := '';
      for m := 0 to prModels.Count - 1 do begin
        ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
        for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            if (i <> j) and ThisModel.ModelObject.Connected(ThisModel.ModelObject.States[i], ThisModel.ModelObject.States[j]) then begin
              t := t + Format('%-14.6g'#9, [TPMInteger(prTotTransitionCntAll[m])[i, j]/TPADouble(prTotLifetimeTimeAll[m])[i]]);
              tr := tr + Format('%-14.6g'#9, [TPMInteger(prTotTransitionCntAll[m])[i, j]/TPADouble(prTotLifetimeTimeAll[m])[i]]);
            end;
          end;
          t := t + Format('%-14.6g'#9, [TPADouble(prTotLifetimeTimeAll[m])[i]/(prTotalPointCount * Data.Sampling)]);
          tr := tr + Format('%-14.6g'#9, [TPADouble(prTotLifetimeTimeAll[m])[i]/(prTotalPointCount * Data.Sampling)]);
        end;
      end;
      prConsoleMsg := t;
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);
      if prResultText <> nil then begin
        prResultText^ := tr;
      end;
    end;


    if prError then begin

      Synchronize(ShowMessage);

      {???}prData.AppendData(nil, 0, 0, False, 0, True, False{Acq});

    end else begin

      prData.AppendData(nil, 0, 0, False, 0, True, False{Acq});

    end;


    //TO DO: Update experimental conditions in the file. LM

    prUpdateStatus(100, prStTime);

    PythonThread.SneakPyProc( prPublishResultsPy );

    for m := 0 to prModelCount - 1 do begin
      FreeMem(prState[m]);
    end;
    FreeMem(prState);
    prState := nil;
    for m := 0 to prModelCount - 1 do begin
      FreeMem(prProbs[m]);
    end;
    FreeMem(prProbs);
    prProbs := nil;

    prModelCount := 0;

    FreeMem(prBuffer);
    prBuffer := nil;
    FreeMem(BufferMPX);

    FreeMem(prFluxBuffer);
    prFluxBuffer := nil;

    FreeMem(prBufferStateIDL);
    prBufferStateIDL := nil;
    FreeMem(prBufferStateIDLAmp);
    prBufferStateIDLAmp := nil;

    FreeMem(prBufferIDL);
    prBufferIDL := nil;
    FreeMem(prBufferIDLAmp);
    prBufferIDLAmp := nil;
    prBufferIDLCount := 0;
    prBufferIDLMaxCount := 0;

    if prError then begin
      Synchronize(ShowMessage);
    end;

  end;

end;


//Make Movie - must be Synchronized
procedure TQUB_Task_Simulate.prMakeMovie;
var
  StructureBitmaps :TPAPointer;
  StructureBitmap  :TBitmap;
  StructureFile    :String;
  FrameName        :String;
  Sel              :TQUB_FirstLast;
  i, j             :Integer;
begin

  prData.AppendData(nil, 0, 0, False, 0, True, False{Acq});
  StructureBitmaps := nil;
  Data.Zoom.Bitmap.PixelFormat := pf24bit;
  Data.Zoom.WindowDrawMode := True;

  try

    //Create and load Structure Bitmaps
    StructureBitmaps := PAlloc1(prStructureFiles.Count);
    try
      for i := 0 to prStructureFiles.Count - 1 do begin
        StructureBitmaps[i] := Pointer(TBitmap.Create);
        StructureFile := prStructureFiles.Strings[i];
        if FileExists(StructureFile) then begin
          TBitmap(StructureBitmaps[i]).LoadFromFile(StructureFile);
        end;
      end;
    except
    end;
    try
      Sel.First := prBufferIDL[0].First;
      Sel.Last  := Sel.First;
      Data.Zoom.MustRedraw := True;
      Data.Zoom.Paint;

      for i := 0 to prBufferIDLCount - 1 do begin
        if prStopped then Break;
        StructureBitmap := nil;
        if (prBufferIDL[i].Cls >= 0) and (prBufferIDL[i].Cls < prStructureFiles.Count) then begin
          StructureBitmap := TBitmap(StructureBitmaps[prBufferIDL[i].Cls]);
        end;

        for j := prBufferIDL[i].First to prBufferIDL[i].Last do begin

          if prStopped then Break;

          FrameName := IntToStr(prFrameIdx);
          while Length(FrameName) < 4 do FrameName := '0'+ FrameName;
          FrameName := prMovieName + '_qubMovie_' + FrameName + '.bmp';

          Sel.Last := j;
          Data.Zoom.DataSel := Sel;
          Data.Zoom.MustRedraw := True;
          Synchronize(Data.Zoom.Paint);
          Data.Zoom.MustRedraw := False;
          if StructureBitmap <> nil then begin
            Data.Zoom.Bitmap.Canvas.Draw(Data.Zoom.Margin.Left + (Data.Zoom.Bitmap.Width - StructureBitmap.Width) div 2, Data.Zoom.Margin.Top, StructureBitmap);
          end;

          //Synchronize(Data.Zoom.Paint);
          //Sleep(10);

          if Properties.MakeMovieSaveFrames then begin
            Data.Zoom.Bitmap.SaveToFile(FrameName);
          end else begin
            Sleep(Round(Properties.MakeMovieFrameDuration));
          end;
          Inc(prFrameIdx);
          if prFrameIdx > prMaxFrame then Break;
          if prStopped then Break;

          Application.ProcessMessages;

        end;//j
        if prFrameIdx > prMaxFrame then Break;
        if prStopped then Break;
      end;//i
    except
    end
  finally
    for i := 0 to prStructureFiles.Count - 1 do begin
      TBitmap(StructureBitmaps[i]).Free;
    end;
    FreeMem(StructureBitmaps);
    Data.Zoom.Bitmap.PixelFormat := pfDevice;
    Data.Zoom.WindowDrawMode := False;
  end;
end;

//****************************************************************************//
//                                                                            //
//                             S T I M U L A T E D                            //
//                                                                            //
//                                   F I L E                                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Simulate.prDoSimulate_Stimulation_File;
var
  ParCount       :Integer;
  FirstSeg       :Integer;
  BufferMPX      :Pointer;
  BufferCount    :Integer;
begin

  if Properties.Randomize then begin
    Randomize;
  end else begin
    RandSeed := Properties.RandomSeed;
  end;
  prRandSeedDwell := RandSeed;
  prRandSeedNoise := RandSeed;

  prIDum := Round(Random(High(Integer)));

  prWarningMsg   := '';

  BufferMPX := nil;


  try

    // these buffers are used in two different places below; this BufferCount is their max
    BufferCount      := ONE_MEGABYTE;
    //BufferCount    := Properties.SegmentSize;
    prBuffer         := IAlloc1(BufferCount);
    
    //TO DO: LM. Check this
    //prBufferStateIDL := IAlloc1(BufferCount);
    //prBufferStateIDLAmp := SAlloc1(BufferCount);

    BufferMPX        := AllocMem(Data.DataObject.ADChannelCount * Data.DataObject.ADDataSize * BufferCount);

    prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
    prStatus := 'Running';

    Synchronize(prDoOnChange);

    if prData.DataObject.Segments.TotalDataCount = 0 then begin
      FirstSeg := 0;
    end else begin
      FirstSeg := prData.DataObject.Segments.SegmentCount;
    end;
    
    prData.DataChannelName[0]   := 'Current';
    prData.DataChannelUnits[0]  := 'pA';

    if prStimProto <> nil then begin

      prDoSimulate_Stimulation_File_Proto;

    end else begin {stimulated by prStimData :QVD}

      prDoSimulate_Stimulation_File_Data(BufferMPX, BufferCount);

    end;

    if (FirstSeg >= prData.DataObject.Segments.SegmentCount)
    or (prData.DataObject.Segments[FirstSeg].Channels.DataCount = 0) then begin
      prError    := True;
      prErrorMsg := 'Stimulus file is empty.';
      prWikiTopic := 'Error:Stimulus_Empty';
      Exit;
    end;

    ParCount := prDoSimulate_Stimulation_File_SetupPars;
    if not prError then prDoSimulate_Stimulation_File_Core(ParCount, FirstSeg, BufferMPX, BufferCount);

  finally

    prUpdateStatus(100, prStTime);

    FreeMem(prBuffer);
    prBuffer := nil;
    FreeMem(BufferMPX);

    if prError then begin
      Synchronize(ShowMessage);
    end else if prWarning then begin
      Synchronize(ShowMessage);
    end;

  end;

end;

procedure TQUB_Task_Simulate.prDoSimulate_Stimulation_File_Proto;
var
  c               :Integer;
  MaxInt          :Double;
  Low, High       :Double;
  Task            :TQUB_Task_Protocol_To_Data;
begin
  try

    // set appropriate datachannelscaling
    //   discover expressed range, then expand to be safe
    //   want t_max / (scaling * datachannelscaling) == High
    //   so datachannelscaling = t_max / (scaling * High)
    MaxInt := Power(Power(2,8), prData.DataObject.ADDataSize) / 2 - 1;
    for c := 0 to Min(prData.DataObject.ADChannelCount-1, prStimProto.Protocol.ChannelCount) - 1 do begin
      prStimProto.Protocol.GetBoundsOfChannel(c, prData.DataObject.Sampling, Low, High);
      High := Max(Abs(Low), Abs(High)) * 2;
      if High = 0.0 then High := MaxInt;
      prData.DataChannelScaling[c+1] := MaxInt / (prData.DataObject.Scaling * High);
      prData.DataChannelOffset[c+1]  := 0.0;
    end;

    prData.UnlockAny;

    // get task_p_to_d to generate the stimulus
    prProtoToDataSignal := TSimpleEvent.Create;

    Task                := TQUB_Task_Protocol_To_Data.Create(True);
    Task.OnTerminate    := prProtoToDataFinished;
    Task.Protocol       := prStimProto;
    Task.Data           := prData;
    Task.BaseChannel    := 1;

    Task.OnChange       := TaskView.OnTaskChange;
    Task.Resume;
    while wrSignaled <> prProtoToDataSignal.WaitFor(1000) do begin end;

  finally

    prProtoToDataSignal.Free;
    prData.LockAny;

  end;

end;

procedure TQUB_Task_Simulate.prProtoToDataFinished(Sender :TObject);
begin
  prProtoToDataSignal.SetEvent;
end;

procedure TQUB_Task_Simulate.prDoSimulate_Stimulation_File_Data(BufferMPX :Pointer; ActualBufferCount :Integer);
var
  c, s           :Integer;
  Seg            :TQUB_QFS_Segment;
  StimSeg        :TQUB_QFS_Segment;
  StartTime      :Double;
  BufferCount    :Integer; // the portion used
  DataFirst      :Integer;
  ADCC           :Integer;
begin
  ADCC           := Min(prData.DataObject.ADChannelCount-1, prStimData.DataChannelCount);

  // init channel scaling -- copying int values as is,
  // so must rescale channels to account for possibly different Scaling
  for c := 0 to ADCC - 1 do begin
    prData.DataChannelUnits[c+1]   := prStimData.DataChannelUnits[c];
    prData.DataChannelScaling[c+1] := prStimData.DataChannelScaling[c] * prStimData.DataObject.Scaling / prData.DataObject.Scaling;
    prData.DataChannelOffset[c+1]  := prStimData.DataChannelOffset[c];
  end;

  //Segment Loop
  DataFirst := 0;
  for s := 0 to prStimData.DataObject.Segments.SegmentCount - 1 do begin

    StimSeg           := prStimData.DataObject.Segments[s];
    prCurSegmentCount := 0;

    //Loop to generate all data in a segment - it works with 1 MP chunks
    repeat

      if (LongInt(StimSeg.Channels.DataCount) - prCurSegmentCount) > ActualBufferCount then begin
        BufferCount := ActualBufferCount;
      end else begin
        BufferCount := LongInt(StimSeg.Channels.DataCount) - prCurSegmentCount;
      end;

      //Load Buffer and MUX it
      for c := 0 to ADCC - 1 do begin

        //Load
        //prStimData.DataObject.ReadData(prBuffer, False{Raw}, c{Channel}, prCurSegmentCount{DataFirst}, prCurSegmentCount + BufferCount - 1{DataLast});
        prStimData.DataObject.ReadData(prBuffer, False{Raw}, c{Channel}, DataFirst, DataFirst + BufferCount - 1{DataLast});
        //prStimData.DataObject.ReadData(prBuffer, False{Raw}, c{Channel}, prCurSegmentCount{DataFirst}, prCurSegmentCount + BufferCount - 1{DataLast});

        //MPX
        MPXData(prBuffer, BufferMPX, prData.DataObject.ADChannelCount, c + 1{Channel}, prData.DataObject.ADDataSize, prData.DataObject.ADDataType, 0, BufferCount - 1, True{MPX});

      end;

      // append to prData
      if prData.DataObject.Segments.TotalDataCount = 0 then begin
        StartTime := 0;
      end else if prCurSegmentCount = 0 then begin
        Seg       := prData.DataObject.Segments[prData.DataObject.Segments.SegmentCount - 1];
        StartTime := Seg.StartTime + Seg.Channels.DataCount * prData.Sampling;
      end else begin
        StartTime := 0;
      end;

      prData.AppendData(BufferMPX, Data.DataObject.ADChannelCount * Data.DataObject.ADDataSize, BufferCount, prCurSegmentCount = 0{NewSegment}, StartTime, False{CanRefresh - Doit}, False{Acq});

      Inc(prCurSegmentCount, BufferCount);
      Inc(DataFirst, BufferCount);

    until(prStopped or (prCurSegmentCount >= LongInt(StimSeg.Channels.DataCount)));

  end;//s

  prData.AppendData(nil, 0, 0, False, 0, True, False{Acq});

end;

function TQUB_Task_Simulate.prDoSimulate_Stimulation_File_SetupPars: Integer;
// result is ParCount if not prError
var
  ThisModel      :TQUB_QVO_QVM;
  Par            :TQUB_ExpCond_PParameter;
  ParCount       :Integer;
  m, p, c, ch    :Integer;
begin
  Result := -1;

  for p := 0 to QUB_QFS_QMF_MAX_PARAMETER do begin
     prPars[p] := nil;
     prChan[p] := -1;
     prIniV[p] := 0;
  end;

  ParCount := 0;
  //Find parameters used by model(s)
  for p := 0 to ModelParameters.Count - 1 do begin
    Par := ModelParameters[p];

    if (Par.Name = 'Voltage') and (Properties.UseCondVoltage) then begin

      prPars[ParCount] := Par;
      prIniV[ParCount] := Par.Value;
      Inc(ParCount);

    end else begin

      for m := 0 to prModels.Count - 1 do begin
        ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);

        if ThisModel.ModelObject.UsesParameter(Par) then begin
        //ThisModel uses Par
          prPars[ParCount] := Par;
          prIniV[ParCount] := Par.Value;
          Inc(ParCount);
          Break;
        end else begin
          Par.Channel      := -1;
        end;

      end;//Model

    end;

    if ParCount > QUB_QFS_QMF_MAX_PARAMETER then Break;

  end;//Parameter

  //Now we have a list with all the parameters that are used by the model(s)

  //Any parameter at all ?!
  if ParCount = 0 then begin
    prWarning    := True;
    prWarningMsg := 'No stimuli influence upon the model(s) was detected.'#13#10;
    prWikiTopic  := 'Warning:No_Stimuli_Influence';
  end;

  if (ParCount + 1) < (Data.DataObject.ADChannelCount) then begin
    if not ((Properties.Response = SIM_RESPONSE_DETERMINISTIC) and (Properties.Flux)) then begin
      prWarning    := True;
      prWarningMsg := prWarningMsg + 'Data file has more data channels than necessary.'#13#10;
      prWikiTopic  := 'Warning:Unnecessary_Channels';
  //end else if (ParCount + 1) > (Data.DataObject.ADChannelCount) then begin
  //  prError      := True;
  //  prErrorMsg   := 'Data file has not enough data channels for stimuli and response';
  //  prWikiTopic  := 'Error:Not_Enough_Channels';
  //  Exit;
    end;
  end else if (ParCount + 1) > (Data.DataObject.ADChannelCount) then begin
    prError      := True;
    prErrorMsg   := 'Data file has not enough data channels for stimuli and response';
    Exit;
  end;

  //Check the stimuli data channels
  //First accomodate the good Pars
  for p := 0 to ParCount - 1 do begin
    Par := TQUB_ExpCond_PParameter(prPars[p]);
    if (Par.Channel > 0) and (Par.Channel < Data.DataObject.ADChannelCount) and (prChan[Par.Channel] = -1) then begin
      prChan[Par.Channel] := p;
    end;
  end;
  //Now check for the bad ones
  for p := 0 to ParCount - 1 do begin
    Par := TQUB_ExpCond_PParameter(prPars[p]);
    if (Par.Channel < 1) or (Par.Channel >= Data.DataObject.ADChannelCount) or (prChan[Par.Channel] <> p) then begin
      ch := -1;
      for c := 1 to Data.DataObject.ADChannelCount - 1 do begin
        if prChan[c] = -1 then begin
          ch := c;
          Break;
        end;
      end;
      if ch <> -1 then begin
        prWarning    := True;
        prWarningMsg := prWarningMsg + 'Parameter ' + Par.Name + ' had data channel changed from ' + IntToStr(Par.Channel) + ' to ' + IntToStr(c) + #13#10;
        prWikiTopic  := 'Warning:Parameter_Channel_Changed';
        prChan[c] := p;
        Par.Channel := c;
      end else begin
        Par.Channel  := -1;
        prWarning    := True;
        prWarningMsg := prWarningMsg + 'Parameter ' + Par.Name + ' was held constant at ' + FloatToStr(Par.Value) + '.'#13#10;
        prWikiTopic  := 'Warning:Parameter_Held_Constant';
        //prError      := True;
        //prErrorMsg   := 'Parameter ' + Par.Name + ' has an invalid data channel';
        //prWikiTopic  := 'Error:Parameter_Channel_Invalid';
        //Exit;
      end;
    end;
  end;

  for c := 1 to ParCount do begin
    if prChan[c] >= 0
    then Data.DataChannelName[c] := TQUB_ExpCond_PParameter(prPars[prChan[c]]).Name;
  end;

  prData.ExpCond.ReadTree(ModelParameters.ToTree);

  Result := ParCount;
end;

procedure TQUB_Task_Simulate.prDoSimulate_Stimulation_File_Core(ParCount :Integer; FirstSeg :Integer; BufferMPX :Pointer; ActualBufferCount :Integer);
var
  ThisModel      :TQUB_QVO_QVM;
  Seg            :TQUB_QFS_Segment;
  SegSize        :Integer;
  SegBounds      :TQUB_FirstLast;
  BufBounds      :TQUB_FirstLast;
  m, c, s        :Integer;
  psum           :Double;
  BufferCount    :Integer;

  procedure InitializeStates;
  var
    m, c, i, j   :Integer;
  begin

    //Initialize baseline state ************************************************
    prBaselineFluctStateCount := 0;
    prBaselineJumpsStateAmpF  := 0;
    prBaselineJumpsStateAmpL  := 0;
    prBaselineJumpsStateAmpC  := 0;
    prBaselineJumpsStateDur   := ExpDev * Properties.BaselineJumpsLifetime;
    prBaselineJumpsStateStart := 0;
    prBaselineLGMState[0]     := Properties.BaselineLGMX0;
    prBaselineLGMState[1]     := Properties.BaselineLGMX0p;
    //**************************************************************************

    //Reset JumpAmp
    prJumpAmp                 := 0;
    prJumpCls                 := 0;
    prJumpState               := 0;
    prJumpGrp                 := 0;
    prJumpCnt[-1]             := 0;
    prJumpCnt[+1]             := 0;
    prTransitionCnt           := 0;
    prLifetimeTime            := 0.0;
    for m := 0 to prModels.Count - 1 do begin
      ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
      for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
        for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          TPMInteger(prTransitionCntAll[m])[i, j] := 0;
        end;
        TPADouble(prLifetimeTimeAll[m])[i] := 0;
      end;
    end;                                   
    prJumpAvgSum              := 0;
    prJumpStdSqrSum           := 0;
    prJumpStdSqrN             := 0;
    prStepIdx                 := 0;


    //Initialize Markov state **************************************************
    for m := 0 to prModels.Count - 1 do begin

      ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);


      if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
      //Stochastic

        for c := 0 to ThisModel.ModelObject.ChannelCount - 1 do begin
          prState[m, c] := ThisModel.ModelObject.GenerateState;
        end;//Channel

        //Idealized
        if Properties.IdealizedData then begin
          if m = 0 then begin
            Inc(prBufferIDLCount);
            if prBufferIDLCount > prBufferIDLMaxCount then begin
              Inc(prBufferIDLMaxCount, prBufferIDLInc);
              try
                ReAllocMem(prBufferIDL,    SizeOf(TQUB_QFS_QIF_FirstLastClass) * prBufferIDLMaxCount);
                ReAllocMem(prBufferIDLAmp, SizeOf(Single)                      * prBufferIDLMaxCount);
              except
                prProperties.IdealizedData := False;
                FreeMem(prBufferIDL);
                prBufferIDL := nil;
                FreeMem(prBufferIDLAmp);
                prBufferIDLAmp := nil;
                prBufferIDLCount := 0;
                prBufferIDLMaxCount := 0;
              end;
            end;
            if Properties.IdealizedData then begin
              prBufferIdl[prBufferIDLCount - 1].First := prTotalPointCount;
              prBufferIdl[prBufferIDLCount - 1].Last  := prBufferIdl[prBufferIDLCount - 1].First - 1;
              if Properties.Staircase then begin
                //TODO: Check this
                if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                  {???}prBufferIdl[prBufferIDLCount - 1].Cls := prJumpState;//prState[0, 0];
                end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls   := prJumpCls;
                end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                  {???}prBufferIdl[prBufferIDLCount - 1].Cls := prJumpGrp;//ThisModel.ModelObject.States[prState[0, 0]].Gr;
                end;
                prBufferIdlAmp[prBufferIDLCount - 1] := prJumpAmp;
              end else begin
                if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := prState[0, 0];
                end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := ThisModel.ModelObject.States[prState[0, 0]].Cls;
                end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := ThisModel.ModelObject.States[prState[0, 0]].Gr;
                end;
              end;
            end;
          end;
        end;


      end else begin
      //Deterministic

        psum := 0;
        for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          psum := psum + ThisModel.ModelObject.States[c].Pr;
        end;//State
        if psum = 0 then begin
          for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            prProbs[m, c] := ThisModel.ModelObject.States[c].Pr / ThisModel.ModelObject.States.StateCount;
          end;//State
        end else begin
          for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            prProbs[m, c] := ThisModel.ModelObject.States[c].Pr / psum;
          end;//State
        end;

      end;

    end;//Model

  end;


begin


  try

    //Allocate Buffer
    prBufferIDLInc := ActualBufferCount div 100;
    if prBufferIDLInc < 1000 then prBufferIDLInc := 1000;
    prBufferIdlCount := 0;
    prBufferIDLMaxCount := 0;

    //TO DO: Check this. LM
    prBufferStateIdl    := IAlloc1(ActualBufferCount);
    prBufferStateIdlAmp := SAlloc1(ActualBufferCount);

    //The State of each channel of each model is memorized
    prModelCount   := prModels.Count;
    TPAPointer(prState) := PAlloc1(prModelCount);
    TPAPointer(prProbs) := PAlloc1(prModelCount);
    prTotalPoints  := 0;
    prActualPoints := 0;


    // find first seg and bounds
    SegBounds.Last      := -1;
    for s := 0 to FirstSeg - 1 do begin
      SegBounds.First   := SegBounds.Last + 1;
      SegBounds.Last    := SegBounds.Last + LongInt(prData.DataObject.Segments[s].Channels.DataCount);
    end;


    for m := 0 to prModelCount - 1 do begin

      ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);


      //Do it only once, here; then only re-calculate Q and P
      ThisModel.ModelObject.UpdateParameters;
      ThisModel.ModelObject.Adt := prData.Sampling;
      ThisModel.ModelObject.CalculateMath;

      prTotalPoints := prTotalPoints + ThisModel.ModelObject.ChannelCount * (prData.DataObject.Segments.TotalDataCount - SegBounds.Last - 1);

      prState[m] := IAlloc1(ThisModel.ModelObject.ChannelCount);
      for c := 0 to ThisModel.ModelObject.ChannelCount - 1 do begin
        prState[m, c] := 0;
      end;//Channel

      prProbs[m] := DAlloc1(ThisModel.ModelObject.States.StateCount);
      for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
        prProbs[m, c] := 1 / ThisModel.ModelObject.States.StateCount;
      end;//State

    end;//Model

    // flux
    if prSetupFlux(ParCount, ActualBufferCount) then Exit;


    //Segment Loop
    for s := FirstSeg to prData.DataObject.Segments.SegmentCount - 1 do begin


      //Initialize / Reset everything
      prCurSegmentIndex := s;
      prCurSegmentCount := 0;

      Seg             := prData.DataObject.Segments[s];
      SegSize         := prData.DataObject.Segments[s].Channels.DataCount;

      SegBounds.First   := SegBounds.Last + 1;
      SegBounds.Last    := SegBounds.Last + SegSize;

      //***************
      InitializeStates;
      //***************


      //Loop to generate all data in a segment - it works with 1 MP chunks
      repeat

        if (SegSize - prCurSegmentCount) > ActualBufferCount then begin
          BufferCount   := ActualBufferCount;
        end else begin
          BufferCount   := SegSize - prCurSegmentCount;
        end;

        // Buffer bounds within this segment
        BufBounds.First := prCurSegmentCount;
        BufBounds.Last  := prCurSegmentCount + BufferCount - 1;


        //Load MPX Buffer
        Seg.Channels.LoadDataInBuffer(@TPAByte(BufferMPX)[0], BufBounds.First, BufBounds.Last);


        //Response
        if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
        //Stochastic

          if prStochasticSimulationEngine(BufferMPX, BufferCount) then Exit;

        end else begin
        //Deterministic

          if prDeterministicSimulationEngine(BufferMPX, BufferCount) then Exit;

        end;

        if prStopped then Exit;

        // now changing to bounds within entire data file
        BufBounds.First := BufBounds.First + SegBounds.First;
        BufBounds.Last  := BufBounds.Last  + SegBounds.First;

        prData.ReplaceData(BufferMPX, BufferCount{ignored}, BufBounds, False{ignored});


        //New:
        //Make Movie ***********************************************************
        if (Properties.MakeMovie) and (Properties.IdealizedData) then begin
          Synchronize(prMakeMovie);
        end;
        //**********************************************************************



        Inc(prCurSegmentCount, BufferCount);
        Inc(prTotalPointCount, BufferCount);


      until (prCurSegmentCount >= SegSize);

      prAppendIdl;


    end;//s


  finally


    for m := 0 to prModelCount - 1 do begin
      FreeMem(prState[m]);
      FreeMem(prProbs[m]);
    end;
    FreeMem(prState);
    FreeMem(prProbs);
    prState := nil;
    prProbs := nil;

    FreeMem(prFluxBuffer);
    prFluxBuffer := nil;

    FreeMem(prBufferStateIDL);
    prBufferStateIDL := nil;
    FreeMem(prBufferStateIDLAmp);
    prBufferStateIDLAmp := nil;

    FreeMem(prBufferIDL);
    prBufferIDL := nil;
    FreeMem(prBufferIDLAmp);
    prBufferIDLAmp := nil;
    prBufferIDLCount := 0;
    prBufferIDLMaxCount := 0;

    //Restore Parameters
    for c := 0 to QUB_QFS_QMF_MAX_PARAMETER do begin
      if prPars[c] <> nil then begin
        TQUB_ExpCond_PParameter(prPars[c]).Value := prIniV[c];
      end;
    end;

    //Restore Models/Parameters
    for m := 0 to prModelCount - 1 do begin
      TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.UpdateParameters;
    end;

    prModelCount := 0;


  end;


end;




(*
//****************************************************************************//
//                                                                            //
//                             S T I M U L A T E D                            //
//                                                                            //
//                                 S C R I P T                                //
//                                                                            //
//****************************************************************************//
  {
    For Stimulated simulations, a Kinetic object is used to generate the
    stimuli waveforms, according to a protocol. The Kinetic objects acts
    as a driver for the QUB object: whenever a chunk of data (stimuli)
    is generated, QUB is signaled to do its share: generate the response
    to stimuli, until the end of segment, or end of file.
  }
procedure TQUB_Task_Simulate.prDoSimulate_Stimulation_Script;
var
  ThisModel      :TQUB_QVO_QVM;
  Par            :TQUB_ExpCond_PParameter;
  BufferCount    :Integer;
  ParCount       :Integer;
  m, c, p, ch    :Integer;
  w              :TWaitResult;
  psum           :Double;
  sum            :Double;

  procedure InitializeStates;
  var
    m, c, i, j   :Integer;
  begin

    //Initialize baseline state ************************************************
    prBaselineFluctStateCount := 0;
    prBaselineJumpsStateAmpF  := 0;
    prBaselineJumpsStateAmpL  := 0;
    prBaselineJumpsStateAmpC  := 0;
    prBaselineJumpsStateDur   := ExpDev * Properties.BaselineJumpsLifetime;
    prBaselineJumpsStateStart := 0;
    prBaselineLGMState[0]     := Properties.BaselineLGMX0;
    prBaselineLGMState[1]     := Properties.BaselineLGMX0p;
    //**************************************************************************

    //Reset JumpAmp
    prJumpAmp                 := 0;
    prJumpCls                 := 0;
    prJumpState               := 0;
    prJumpGrp                 := 0;
    prJumpCnt[-1]             := 0;
    prJumpCnt[+1]             := 0;
    prTransitionCnt           := 0;
    prLifetimeTime            := 0.0;
    for m := 0 to prModels.Count - 1 do begin
      ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);
      for i := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
        for j := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          TPMInteger(prTransitionCntAll[m])[i, j] := 0;
        end;
        TPADouble(prLifetimeTimeAll[m])[i] := 0;
      end;
    end;
    prJumpAvgSum              := 0;
    prJumpStdSqrSum           := 0;
    prJumpStdSqrN             := 0;
    prStepIdx                 := 0;


    //Initialize Markov state **************************************************
    for m := 0 to prModels.Count - 1 do begin

      ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);


      if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
      //Stochastic

        for c := 0 to ThisModel.ModelObject.ChannelCount - 1 do begin
          prState[m, c] := ThisModel.ModelObject.GenerateState;
        end;//Channel

        //Idealized
        if Properties.IdealizedData then begin
          if m = 0 then begin
            Inc(prBufferIDLCount);
            if prBufferIDLCount > prBufferIDLMaxCount then begin
              Inc(prBufferIDLMaxCount, prBufferIDLInc);
              try
                ReAllocMem(prBufferIDL,    SizeOf(TQUB_QFS_QIF_FirstLastClass) * prBufferIDLMaxCount);
                ReAllocMem(prBufferIDLAmp, SizeOf(Single)                      * prBufferIDLMaxCount);
              except
                prProperties.IdealizedData := False;
                FreeMem(prBufferIDL);
                prBufferIDL := nil;
                FreeMem(prBufferIDLAmp);
                prBufferIDLAmp := nil;
                prBufferIDLCount := 0;
                prBufferIDLMaxCount := 0;
              end;
            end;
            if Properties.IdealizedData then begin
              prBufferIdl[prBufferIDLCount - 1].First := prTotalPointCount;
              prBufferIdl[prBufferIDLCount - 1].Last  := prBufferIdl[prBufferIDLCount - 1].First - 1;
              if Properties.Staircase then begin
                //TODO: Check this
                if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                  {???}prBufferIdl[prBufferIDLCount - 1].Cls := prJumpState;//prState[0, 0];
                end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls   := prJumpCls;
                end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                  {???}prBufferIdl[prBufferIDLCount - 1].Cls := prJumpGrp;//ThisModel.ModelObject.States[prState[0, 0]].Gr;
                end;
                prBufferIdlAmp[prBufferIDLCount - 1] := prJumpAmp;
              end else begin
                if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := prState[0, 0];
                end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := ThisModel.ModelObject.States[prState[0, 0]].Cls;
                end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                  prBufferIdl[prBufferIDLCount - 1].Cls := ThisModel.ModelObject.States[prState[0, 0]].Gr;
                end;
              end;
            end;
          end;
        end;


      end else begin
      //Deterministic

        psum := 0;
        for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          psum := psum + ThisModel.ModelObject.States[c].Pr;
        end;//State
        if psum = 0 then begin
          for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            prProbs[m, c] := ThisModel.ModelObject.States[c].Pr / ThisModel.ModelObject.States.StateCount;
          end;//State
        end else begin
          for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
            prProbs[m, c] := ThisModel.ModelObject.States[c].Pr / psum;
          end;//State
        end;

      end;

    end;//Model

  end;


begin


  if Properties.Randomize then begin
    Randomize;
  end else begin
    RandSeed := Properties.RandomSeed;
  end;
  prRandSeedDwell := RandSeed;
  prRandSeedNoise := RandSeed;

  prIDum := Round(Random(High(Integer)));

  prWarningMsg   := '';

  try

    //Allocate Buffer
    BufferCount              := ONE_MEGABYTE;
    {***}prBuffer            := IAlloc1(BufferCount);
    {***}prBufferStateIDL    := IAlloc1(BufferCount);
    {***}prBufferStateIDLAmp := SAlloc1(BufferCount);


    prBufferIDLInc := BufferCount div 100;
    if prBufferIDLInc < 1000 then prBufferIDLInc := 1000;
    prBufferIdlCount := 0;
    prBufferIDLMaxCount := 0;

    //The State of each channel of each model is memorized
    prModelCount   := prModels.Count;
    {***}TPAPointer(prState) := PAlloc1(prModelCount);
    {***}TPAPointer(prProbs) := PAlloc1(prModelCount);
    prTotalPoints  := 0;
    prActualPoints := 0;

    for m := 0 to prModelCount - 1 do begin

      ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);


      //Do it only once, here; then only re-calculate Q and P
      ThisModel.ModelObject.UpdateParameters;
      ThisModel.ModelObject.Adt := prData.Sampling;
      ThisModel.ModelObject.CalculateMath;


      sum := Properties.SegmentCount;
      sum := sum * Properties.SegmentSize;
      sum := sum * ThisModel.ModelObject.ChannelCount;
      prTotalPoints := prTotalPoints + sum;

      {***}prState[m] := IAlloc1(ThisModel.ModelObject.ChannelCount);
      for c := 0 to ThisModel.ModelObject.ChannelCount - 1 do begin
        prState[m, c] := 0;
      end;//Channel

      {***}prProbs[m] := DAlloc1(ThisModel.ModelObject.States.StateCount);
      for c := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
        prProbs[m, c] := 1 / ThisModel.ModelObject.States.StateCount;
      end;//State

    end;//Model

    for p := 0 to QUB_QFS_QMF_MAX_PARAMETER do begin
       prPars[p] := nil;
       prChan[p] := -1;
       prIniV[p] := 0;
    end;

    ParCount := 0;
    //Find parameters used by model(s)
    for p := 0 to ModelParameters.Count - 1 do begin
      Par := ModelParameters[p];

      if (Par.Name = 'Voltage') and (Properties.UseCondVoltage) then begin

        prPars[ParCount] := Par;
        prIniV[ParCount] := Par.Value;
        Inc(ParCount);

      end else begin

        for m := 0 to prModels.Count - 1 do begin
          ThisModel := TQUB_QVO_QVM(prModels.Objects[m]);

          if ThisModel.ModelObject.UsesParameter(Par) then begin
          //ThisModel uses Par
            prPars[ParCount] := Par;
            prIniV[ParCount] := Par.Value;
            Inc(ParCount);
            Break;
          end else begin
            Par.Channel      := -1;
          end;

        end;//Model

      end;

      if ParCount > QUB_QFS_QMF_MAX_PARAMETER then Break;

    end;//Parameter

    //Now we have a list with all the parameters that are used by the model(s)

    //Any parameter at all ?!
    if ParCount = 0 then begin
      prWarning    := True;
      prWarningMsg := 'No stimuli influence upon the model(s) was detected.'#13#10;
      prWikiTopic  := 'Warning:No_Stimuli_Influence';
    end;

    if (ParCount + 1) < (Data.DataObject.ADChannelCount) then begin
      prWarning    := True;
      prWarningMsg := prWarningMsg + 'Data file has more data channels than necessary.'#13#10;
      prWikiTopic  := 'Warning:Unnecessary_Channels';
    end else if (ParCount + 1) > (Data.DataObject.ADChannelCount) then begin
      prError      := True;
      prErrorMsg   := 'Data file has not enough data channels for stimuli and response';
      prWikiTopic  := 'Error:Not_Enough_Channels';
      Exit;
    end;

    //Check the stimuli data channels
    //First accomodate the good Pars
    for p := 0 to ParCount - 1 do begin
      Par := TQUB_ExpCond_PParameter(prPars[p]);
      if (Par.Channel > 0) and (Par.Channel < Data.DataObject.ADChannelCount) and (prChan[Par.Channel] = -1) then begin
        prChan[Par.Channel] := p;
      end;
    end;
    //Now check for the bad ones
    for p := 0 to ParCount - 1 do begin
      Par := TQUB_ExpCond_PParameter(prPars[p]);
      if (Par.Channel < 1) or (Par.Channel >= Data.DataObject.ADChannelCount) or (prChan[Par.Channel] <> p) then begin
        ch := -1;
        for c := 1 to Data.DataObject.ADChannelCount - 1 do begin
          if prChan[c] = -1 then begin
            ch := c;
            Break;
          end;
        end;
        if ch <> -1 then begin
          prWarning    := True;
          prWarningMsg := prWarningMsg + 'Parameter ' + Par.Name + ' had data channel changed from ' + IntToStr(Par.Channel) + ' to ' + IntToStr(c) + #13#10;
          prWikiTopic  := 'Warning:Parameter_Channel_Changed';
          prChan[c] := p;
          Par.Channel := c;
        end else begin
          prError      := True;
          prErrorMsg   := 'Parameter ' + Par.Name + ' has an invalid data channel';
          prWikiTopic  := 'Error:Parameter_Channel_Invalid';
          Exit;
        end;
      end;
    end;

    prData.ExpCond.ReadTree(ModelParameters.ToTree);

    prCurSegmentIndex := 0;
    prCurSegmentCount := 0;


    prData.DataChannelName[0]   := 'Current';
    prData.DataChannelUnits[0]  := 'pA';

    // flux
    if prSetupFlux(ParCount, BufferCount) then Exit;


    //***************
    InitializeStates;
    //***************



    TKPL_Collection_QUB_Pulse(prKPLPulse).OnBeforeData       := prOnBeforeData_Stimulated;
    TKPL_Collection_QUB_Pulse(prKPLPulse).DefaultChannel     := 1;
    with prKPLPulse as TKPL_Collection_QUB_Pulse do begin
      Sampling                                               := Data.Sampling;
      PulseToData.Data                                       := Data;
      PulseToData.OnAfterData                                := prOnAfterData_Stimulated;
      PulseToData.NoFlush                                    := True;
    end;

    Kinetic.OnEndRun                                         := prOnEndRun_Stimulated;

    prStTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
    prStatus := 'Running';

    Synchronize(prDoOnChange);

    //Run Kinetic Driver
    Kinetic.Run;

    repeat

      w := prSignal.WaitFor(INFINITE);

      if prStopped then Exit;

      case w of
        wrSignaled, wrTimeout:
          begin

            if (prCurSegmentCount >= Properties.SegmentSize) or (not Kinetic.Running) then begin
            //Current Segment is done, or Kinetic has stopped

              Application.ProcessMessages;
              
              if Kinetic.Running then begin

                Kinetic.Stop;

              end else begin



                prAppendIdl;


                //New:
                //Make Movie ***********************************************************
                //Testing
                if (Properties.MakeMovie) and (Properties.IdealizedData) then begin
                  Synchronize(prMakeMovie);
                  prBufferIDLCount := 0;
                end;
                //**********************************************************************


                Inc(prCurSegmentIndex);
                prCurSegmentCount := 0;
                if prCurSegmentIndex >= Properties.SegmentCount then begin
                  Break;
                end;


                //***************
                InitializeStates;
                //***************


                Kinetic.Run;

              end;

            end;

          end;

        wrAbandoned, wrError:
          Break;

      end;

      prSignal.ResetEvent;

    until(False);


  finally


    if Kinetic.Running then Kinetic.Stop;

    {???}if not prError then begin
      prData.AppendData(nil, 0, 0, False, 0, True, False{Acq});
    end;

    prUpdateStatus(100, prStTime);

    for m := 0 to prModelCount - 1 do begin
      FreeMem(prState[m]);
      FreeMem(prProbs[m]);
    end;
    FreeMem(prState);
    FreeMem(prProbs);
    prState := nil;
    prProbs := nil;

    FreeMem(prBuffer);
    prBuffer := nil;
    FreeMem(prFluxBuffer);
    prFluxBuffer := nil;

    FreeMem(prBufferStateIDL);
    prBufferStateIDL := nil;
    FreeMem(prBufferStateIDLAmp);
    prBufferStateIDLAmp := nil;

    FreeMem(prBufferIDL);
    prBufferIDL := nil;
    FreeMem(prBufferIDLAmp);
    prBufferIDLAmp := nil;
    prBufferIDLCount := 0;
    prBufferIDLMaxCount := 0;

    //Restore Parameters
    for c := 0 to QUB_QFS_QMF_MAX_PARAMETER do begin
      if prPars[c] <> nil then begin
        TQUB_ExpCond_PParameter(prPars[c]).Value := prIniV[c];
      end;
    end;

    //Restore Models/Parameters
    for m := 0 to prModelCount - 1 do begin
      TQUB_QVO_QVM(prModels.Objects[m]).ModelObject.UpdateParameters;
    end;

    prModelCount := 0;

    if prError then begin
      Synchronize(ShowMessage);
    end else if prWarning then begin
      Synchronize(ShowMessage);
    end;


  end;


end;
*)

procedure TQUB_Task_Simulate.prOnBeforeData_Stimulated(Sender :TObject; var BufferCount :Integer);
begin
  {
  with prKPLPulse as TKPL_Collection_QUB_Pulse do begin
    PulseToData.NewSegment := prCurSegmentCount = 0;
  end;
  }
  if (BufferCount + prCurSegmentCount) >= Properties.SegmentSize then begin
    BufferCount := Properties.SegmentSize - prCurSegmentCount;
  end;

end;

procedure TQUB_Task_Simulate.prOnAfterData_Stimulated(Sender :TObject; BufferMPX :Pointer; BufferSize :Integer; BufferCount :Integer; var Filled :Array of Boolean);
var
  IsFilled :Boolean;
  p        :Integer;
begin

  //Is it filled ?
  IsFilled := True;
  if Properties.Stimulation <> SIM_STIM_NONE then begin
    for p := 1 to QUB_QFS_QMF_MAX_PARAMETER do begin
      if prChan[p] <> -1 then begin
        if not Filled[p] then begin
          IsFilled := False;
          Break;
        end;
      end;
    end;
  end;

  if not IsFilled then Exit;

  if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
  //Stochastic

    if (prStochasticSimulationEngine(BufferMPX, BufferCount)) or (prStopped) then begin
      //StopIt := True;
    end;

  end else begin
  //Deterministic

    if (prDeterministicSimulationEngine(BufferMPX, BufferCount)) or (prStopped) then begin
      //StopIt := True;
    end;

  end;

  Inc(prCurSegmentCount, BufferCount);
  Inc(prTotalPointCount, BufferCount);

  {???}prSignal.SetEvent;


end;


procedure TQUB_Task_Simulate.prOnEndRun_Stimulated(Sender :TObject; Failed :Boolean);
begin

  prSignal.SetEvent;

end;


//****************************************************************************//
//                                                                            //
//        S T O C H A S T I C     S I M U L A T I O N      E N G I N E        //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Simulate.prStochasticSimulationEngine(BufferMPX :Pointer; BufferCount :Integer) :Boolean;
var
  ThisModel             :TQUB_QVO_QVM;
  StateCount            :Integer;
  BufDuration           :Double;
  Sampling              :Double;
  Scaling               :Double;
  Offset                :Double;
  BackAmp               :Double;
  BackNoise             :Double;
  BackNoise2            :Double;
  ExcNoise              :Double;
  CurPoints             :Integer;
  CurDuration           :Double;
  EvtPoints             :Integer;
  EvtDuration           :Double;
  NxtDuration           :Double;
  NewState              :Integer;
  NewerState            :Integer;
  ThisAmp               :Double;
  OldClass              :Integer;
  OldClass_             :Integer;
  ThisClass             :Integer;
  ThisClass_            :Integer;
  ThisVoltage           :Double;
  ThisVoltageCh         :Integer;
  NewPercent            :Integer;
  IniV                  :TPAInteger;
  Vars                  :TPMInteger;
  VarScaling            :TPADouble;
  VarOffset             :TPADouble;
  ParDiff               :Boolean;
  s, m, c, ch, p, pp    :Integer;
  i, j, jj              :Integer;
  e                     :Double;
  ChPoint               :Integer;
  NxtChPoint            :Integer;
  CCls                  :Integer;
  DoChange              :Boolean;

  Found                 :Boolean;
  prevCls               :Integer;

  MinCls, MaxCls        :Integer;

  ThisJumpAmp           :Double;
  ThisStepIdx           :Integer;

  ILeak                 :Double;

  RSeedDwell            :LongInt;
  RSeedNoise            :LongInt;

  IGate, IGateSum       :Array of Double;
  GateCount, GateIx     :Integer;
  A1, A2, A3, Asum, Samp:Double;
  g                     :Integer;

  procedure DoGatingOnEvent(State1, State2 :Integer);
  // http://www.pubmedcentral.nih.gov/pagerender.fcgi?artid=54945&pageindex=2
  // They measured three time constants in squid axon.
  // http://www.jgp.org/cgi/content/abstract/98/2/265
  // Guinea pigs have similar shape and time-scale.
  var
    ri, hi, si, ii  :Integer;
    Rate            :TQUB_QFS_Rate;
    k1              :Double;
  begin
    if GateCount = 0 then Exit;

    ThisModel.ModelObject.Rates.GetRateIdxHalfIdxOfStates(State1, State2, ri, hi);
    Rate          := ThisModel.ModelObject.Rates[ri];
    if Rate.Q[hi] and (Rate.k1[hi] <> 0.0) then begin
      si := GateIx;
      k1 := Rate.k1[hi];
      for ii := 0 to GateCount - 2 do begin
        IGateSum[si] := IGateSum[si] + k1 * IGate[ii];
        Inc(si);
        if si >= GateCount then si := 0;
      end;
    end;
  end;

begin

  Result                := False;

  Sampling              := Data.Sampling;
  Scaling               := Data.Scaling * Data.DataChannelScaling[0];
  Offset                := Data.DataChannelOffset[0];
  BufDuration           := BufferCount * Sampling;

  IniV                  := nil;
  Vars                  := nil;
  VarScaling            := nil;
  VarOffset             := nil;

  // Gating current:
  // find min(nyquist, filter)
  //GateStd               := (1.0 / Data.Sampling) / 2.0; // nyquist
  //if Data.FilterOn[0] and (Data.FilterFreq[0] < GateStd) then begin
  //  GateStd             := Data.FilterFreq[0];
  //end;  // then scale from time to samples:
  //GateStd               := (1.0 / Data.Sampling) / GateStd;
  // let's say we go out 4 std devs
  //GateCount             := Integer(Ceil(4 * GateStd));
  GateIx                := 0;
  if prGatingScale = 0.0 then begin
    GateCount             := 0;
  end else begin
    GateCount             := Round(1000 * 1e-6 / Data.Sampling); // 1000 microseconds, in samples
    SetLength(IGate, GateCount);
    SetLength(IGateSum, GateCount);
    // generate half-gaussian for k1=1
    //GatePre               := 2.5e-2 * 1.6e-19 * prGatingScale; // k1*kT
    //GatePre               := 2 * GatePre / (GateStd * Sqrt(2*PI)); // scaler for half-gaussian with integral k1*kT
    //GateExp               := -1 / (2 * GateStd * GateStd);
    //for g := 0 to GateCount - 1 do begin
    //  IGate[g]            := GatePre * Exp(GateExp * g * g);
    //  IGateSum[g]         := 0.0;
    //end;
    Asum                  := 4*10e-6 + 4*100e-6 + 1*1000e-6;
    A1                    := prGatingScale * 2.5e-2 * 1.6e-19 * 4 / Asum;
    A2                    := A1;
    A3                    := prGatingScale * 2.5e-2 * 1.6e-19 / Asum;
    // area will sum to kT in coulombs
    Samp                  := Data.Sampling;
    for g := 0 to GateCount - 1 do begin
      IGate[g]            := A1 * Exp(-(g*Samp)/10e-6)
                           + A2 * Exp(-(g*Samp)/100e-6)
                           + A3 * Exp(-(g*Samp)/1000e-6);
      IGateSum[g]         := 0.0;
    end;
  end;

  ThisVoltage           := Properties.CondVoltage;

  NxtChPoint            := -1;

  try

    //Fill buffer with background noise

    //Uses class 0 of first model as the background noise
    ThisModel  := TQUB_QVO_QVM(prModels.Objects[0]);
    BackAmp    := ThisModel.ModelObject.Amp[0];//Amp[0] should not change with voltage, otherwise the whole game changes
    BackNoise  := ThisModel.ModelObject.Std[0];
    BackNoise2 := BackNoise * BackNoise;

    if Properties.Stimulation <> SIM_STIM_NONE then begin

      IniV       := IAlloc1(Data.DataObject.ADChannelCount - 1);{!!! subtract -1 from index}
      Vars       := IAlloc2(Data.DataObject.ADChannelCount - 1, BufferCount);{!!! subtract -1 from first index}
      VarScaling := DAlloc1(Data.DataObject.ADChannelCount - 1);
      VarOffset  := DAlloc1(Data.DataObject.ADChannelCount - 1);

      //De-MUX the stimuli
      for c := 1 to Data.DataObject.ADChannelCount - 1 do begin
        if prChan[c] <> - 1 then begin
          MPXData(Vars[c - 1], BufferMPX, Data.DataObject.ADChannelCount, c, Data.DataObject.ADDataSize, Data.DataObject.ADDataType, 0, BufferCount - 1, False{MPX});
        end;

        VarScaling[c - 1] := Data.Scaling * Data.DataChannelScaling[c];
        VarOffset[c - 1]  := Data.DataChannelOffset[c];
      end;

    end;

    RandSeed := prRandSeedNoise;
    if ThisModel.ModelObject.States.StateCount > 0 then begin
      for p := 0 to BufferCount - 1 do begin
        prBuffer[p] := Round((RandG(BackAmp, BackNoise) - Offset) * Scaling);
        {
        if (nar > 0) then begin
          for (j = 1; j <= nar && j <= n; j++) do begin
            noise0[n] -= (float)ar[0][j]*noise0[n - j];
          end;;
        end;
        }
      end;
    end else begin
      for p := 0 to BufferCount - 1 do begin
        prBuffer[p] := 0;
      end;
    end;
    prRandSeedNoise := RandSeed;


    //Add ILeak? ###############################################################
    ILeak := 0;
    if Properties.ILeak then begin

      ThisVoltageCh := -1;
      if Properties.Stimulation <> SIM_STIM_NONE then begin
      //Stimulation
        for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
          if prChan[ch] <> - 1 then begin
            if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
              ThisVoltageCh := ch;
              Break;
            end;
          end;
        end;
      end;

      if ThisVoltageCh = -1 then begin
      //No channel for ThisVoltage

        ILeak := Properties.GLeak*1e3 * (ThisVoltage - Properties.VLeak);
        for p := 0 to BufferCount - 1 do begin
          prBuffer[p] := Round(prBuffer[p] + (ILeak - Offset) * Scaling);
        end;

      end else begin
      //There is a channel for ThisVoltage

        for p := 0 to BufferCount - 1 do begin
          ThisVoltage := Vars[ThisVoltageCh - 1, p] / VarScaling[ThisVoltageCh - 1] + VarOffset[ThisVoltageCh - 1];//mV
          ILeak := Properties.GLeak*1e3 * (ThisVoltage - Properties.VLeak);
          prBuffer[p] := Round(prBuffer[p] + (ILeak - Offset) * Scaling);
          prData.DataChannelUnits[ThisVoltageCh] := 'mV';
        end;

      end;

    end;//ILeak?  ##############################################################


    for m := 0 to prModels.Count - 1 do begin
    //Model Loop

      ThisModel  := TQUB_QVO_QVM(prModels.Objects[m]);
      StateCount := ThisModel.ModelObject.States.StateCount;

      if StateCount = 0 then Continue;

      if Properties.Staircase then begin
        MinCls := -1;
        MaxCls := -1;
        for pp := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
          c := ThisModel.ModelObject.States[pp].Cls;
          if (c <> 0) and (    Odd(c)) and (c > MaxCls) then MaxCls := c;
          if (c <> 0) and (not Odd(c)) and (c > MinCls) then MinCls := c;
        end;

      end;

      //Calculate Q and P Matrices
      if Properties.Stimulation = SIM_STIM_NONE then begin
        //ThisModel.ModelObject.UpdateParameters;
        //ThisModel.ModelObject.ADt := Sampling;
        //ThisModel.ModelObject.CalculateMath;
        MakeQMatrix(ThisModel.ModelObject, ThisModel.ModelObject.Q);
        MakePMatrix(ThisModel.ModelObject.Q, ThisModel.ModelObject.P, ThisModel.ModelObject.MathStateCount);
      end;

      for c := 0 to ThisModel.ModelObject.ChannelCount - 1 do begin
      //Channel Loop

        CurPoints   := 0;
        CurDuration := CurPoints * Sampling;
        EvtDuration := 0;

        if Properties.Stimulation <> SIM_STIM_NONE then begin

          //Initialize parameters
          for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
            if prChan[ch] <> - 1 then begin
              TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value := Vars[ch - 1, 0] / VarScaling[ch - 1] + VarOffset[ch - 1];
              IniV[ch - 1]                                      := Vars[ch - 1, 0];
              //ThisVoltage
              if Properties.UseCondVoltage then begin
                if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
                  ThisVoltage := TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value;
                end;
              end;
            end;
          end;

          //Find next change point
          NxtChPoint := BufferCount;
          for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
            if prChan[ch] <> - 1 then begin
              for pp := CurPoints + 1 to BufferCount - 1 do begin
                if IniV[ch - 1] <> Vars[ch - 1, pp] then begin
                  if pp < NxtChPoint then NxtChPoint := pp;
                  Break;
                end;
              end;
            end;
          end;


          //ThisModel.ModelObject.UpdateParameters;
          //ThisModel.ModelObject.ADt := Sampling;
          //ThisModel.ModelObject.CalculateMath;
          MakeQMatrix(ThisModel.ModelObject, ThisModel.ModelObject.Q);
          MakePMatrix(ThisModel.ModelObject.Q, ThisModel.ModelObject.P, ThisModel.ModelObject.MathStateCount);
        end;


        //NewState := prState[m, c];

        //Apply DeadTime?
        if Properties.ApplyDeadTime then begin
          //Find the class
          OldClass  := ThisModel.ModelObject.States[prState[m, c]].Cls;
          OldClass_ := OldClass;//used for the actual class, when output is State or Group
          if Properties.IDLOutput = Integer(OUT_STATE) then begin
             OldClass := prState[m, c];
          end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
            OldClass := ThisModel.ModelObject.States[prState[m, c]].Gr;
          end;
        end;


        //Event generator loop
        repeat

          if prStopped then Exit;


          ParDiff := False;


          //Generate Event

          NewState := prState[m, c];

          DoChange := True;


          //prState[m, c] is the state of the last point
          //Next dwell will be of state NewState


          if (prCurSegmentCount = 0) and (CurPoints = 0) then begin
          //Don't change state of the first point

          //This should be an option!


            EvtDuration := 1 * Sampling;
            DoChange    := False;

            if NewState = (Properties.TransitionStates[1] - 1) then prLifetimeTime := prLifetimeTime + EvtDuration;
            TPADouble(prLifetimeTimeAll[m])[NewState] := TPADouble(prLifetimeTimeAll[m])[NewState] + EvtDuration;


            //Apply DeadTime?
            if Properties.ApplyDeadTime then begin
              //Find the class
              OldClass  := ThisModel.ModelObject.States[NewState].Cls;
              OldClass_ := OldClass;//used for the actual class, when output is State or Group
              if Properties.IDLOutput = Integer(OUT_STATE) then begin
                OldClass := prState[m, c];
              end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                OldClass := ThisModel.ModelObject.States[prState[m, c]].Gr;
              end;
            end;


          end else begin
          //Not the first point


            //Repeat until at least sampling long
            repeat


              if prStopped then Exit;


              if ThisModel.ModelObject.Q[NewState, NewState] <> 0 then begin


                //Generate new event and add its duration to whatever is current
                RandSeed := prRandSeedDwell;
                e := (ExpDev / -ThisModel.ModelObject.Q[NewState, NewState]);
                prRandSeedDwell := RandSeed;
                EvtDuration := EvtDuration + e;

                if NewState = (Properties.TransitionStates[1] - 1) then prLifetimeTime := prLifetimeTime + e;
                TPADouble(prLifetimeTimeAll[m])[NewState] := TPADouble(prLifetimeTimeAll[m])[NewState] + e;


                //Apply DeadTime?
                if Properties.ApplyDeadTime then begin
                  //Find the class
                  ThisClass  := ThisModel.ModelObject.States[NewState].Cls;
                  ThisClass_ := ThisClass;//used for the actual class, when output is State or Group
                  if Properties.IDLOutput = Integer(OUT_STATE) then begin
                    ThisClass := prState[m, c];
                  end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                    ThisClass := ThisModel.ModelObject.States[prState[m, c]].Gr;
                  end;
                  //Is it different
                  if ThisClass <> OldClass then begin
                    if e >= Properties.DeadTime*1e-3 then begin
                      OldClass  := ThisClass;
                      OldClass_ := ThisClass_;
                    end;
                  end;
                end;


                //Don't have at least one sample yet ? Not necessarily
                if EvtDuration < Sampling then begin


                  //DoChange := False;


                  //Must Change State
                  if (StateCount > 1) then begin
                    RandSeed := prRandSeedDwell;
                    NewerState := DisDev(StateCount - 1, ThisModel.ModelObject.P[NewState]);
                    prRandSeedDwell := RandSeed;
                    if (NewerState >= NewState) and (NewerState < (StateCount - 1)) then begin
                      Inc(NewerState);
                    end;


                    //Record transition ........................................
                    //TODO: make property
                    if (NewState = (Properties.TransitionStates[1]-1)) and (NewerState = (Properties.TransitionStates[2]-1)) then Inc(prTransitionCnt);
                    TPMInteger(prTransitionCntAll[m])[NewState, NewerState] := TPMInteger(prTransitionCntAll[m])[NewState, NewerState] + 1;
                    //..........................................................


                    if (Properties.Staircase) and
                       (ThisModel.ModelObject.States[NewerState].Cls{Grp} <> ThisModel.ModelObject.States[NewState].Cls{Grp}) then begin

                      if NewerState = NewState then begin
                        Beep;
                      end;

                      if Properties.StaircaseUseSteps then begin
                      //Step from Param:
                        RandSeed := prRandSeedDwell;
                        ThisStepIdx := DisDev(prStepCounts[prStepIdx], prStepProbs[prStepIdx]);
                        prRandSeedDwell := RandSeed;
                        ThisJumpAmp := RandG(prStepSizes[prStepIdx][ThisStepIdx], prStepSizes[prStepIdx][ThisStepIdx] * Properties.StaircaseStepStd);
                      end else begin
                      //Step from Model
                        ThisJumpAmp := RandG(ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewerState].Cls] -
                                             ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewState].Cls],
                                             Properties.StaircaseStepStd);
                      end;

                      prJumpAmp := prJumpAmp + ThisJumpAmp;
                      prJumpAvgSum    := prJumpAvgSum + Abs(ThisJumpAmp);
                      prJumpStdSqrSum := prJumpStdSqrSum + Sqr(ThisJumpAmp - (ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewerState].Cls] - ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewState].Cls]));
                      Inc(prJumpStdSqrN);


                      //Odd/Even give Up/Down
                      //For two steps Up --> must use only odd classes

                      if Odd(ThisModel.ModelObject.States[NewerState].Cls) then begin
                      //Up
                        if Properties.StaircaseNoJumpClass then begin
                          prJumpCls := (prJumpCls + 1) mod 2;
                        end else begin
                          Inc(prJumpCls);
                        end;
                        Inc(prJumpCnt[+1]);
                        Inc(prStepIdx);
                        if prStepIdx > (prStepCount - 1) then prStepIdx := 0;
                      end else begin
                      //Down
                        if Properties.StaircaseNoJumpClass then begin
                          prJumpCls := (prJumpCls + 1) mod 2;
                        end else begin
                          Dec(prJumpCls);
                        end;
                        Inc(prJumpCnt[-1]);
                        Dec(prStepIdx);
                        if prStepIdx < 0 then prStepIdx := (prStepCount - 1);
                      end;

                      //Testing
                      NewState := NewerState;

                      //if Grp < 0 then move to a state(-Grp), if it exists. If not, do old code
                      Found := False;
                      if ThisModel.ModelObject.States[NewerState].Gr < 0 then begin
                        for pp := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
                          if (ThisModel.ModelObject.States[pp].Gr = -(ThisModel.ModelObject.States[NewerState].Gr)) and
                             (pp <> NewerState)                                                                     then begin
                            NewState := pp;
                            Found    := True;
                            Break;
                          end;
                        end;
                      end;

                      if not Found then begin

                        if ThisModel.ModelObject.States[NewerState].Cls <> 0 then begin
                        //Not changed into Cls 0
                          if (ThisModel.ModelObject.States[NewerState].Cls = MinCls) or
                             (ThisModel.ModelObject.States[NewerState].Cls = MaxCls) then begin
                          //Terminal Cls, must Reset
                            for pp := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
                              if (ThisModel.ModelObject.States[pp].Gr = ThisModel.ModelObject.States[NewerState].Gr) and
                                 (pp <> NewerState)                                                                  and
                                 (ThisModel.ModelObject.States[pp].Cls = 0)                                          then begin
                                NewState := pp;
                              end;
                            end;
                          end;
                        end;

                      end;//Found

                      //Testing - old code
                      {
                      for pp := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
                        if (ThisModel.ModelObject.States[pp].Gr = ThisModel.ModelObject.States[NewerState].Gr) and
                           (pp <> NewerState) and
                           (ThisModel.ModelObject.States[pp].Cls = 0) then begin
                          NewState := pp;
                        end;
                      end;
                      }

                      //Testing
                      //Moved above
                      prJumpState := NewState;
                      prJumpGrp   := ThisModel.ModelObject.States[NewState].Gr;

                    end else begin

                      DoGatingOnEvent(NewState, NewerState);
                      NewState := NewerState;

                    end;

                    prState[m, c] := NewState;

                  end;//StateCount > 1

                end;//EvtDuration < Sampling

              end else begin

                EvtDuration := (BufferCount - CurPoints + 1{to make sure there is no float error}) * Sampling;

              end;

            until (EvtDuration >= Sampling);

            //State[m, c] is the entry state and NewState is the exit state


          end;//SegmentCount = 0 ?


          //Apply DeadTime?
          if Properties.ApplyDeadTime then begin
            //Use the old class
            ThisClass  := OldClass;
            ThisClass_ := OldClass_;
          end else begin
            //Find the new class ...
            ThisClass  := ThisModel.ModelObject.States[prState[m, c]].Cls;
            ThisClass_ := ThisClass;//used for the actual class, when output is State or Group
            if Properties.IDLOutput = Integer(OUT_STATE) then begin
              ThisClass := prState[m, c];
            //Testing
            //What about OUT_GROUP?! - to do
            end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
              ThisClass := ThisModel.ModelObject.States[prState[m, c]].Gr;
            end;
          end;

          //... new Amp and Noise
          if Properties.Staircase then begin

            ThisAmp   := prJumpAmp;
            ExcNoise  := 0;

          end else begin

            //Amp
            if Properties.UseCondVoltage then begin
              ThisAmp := ThisModel.ModelObject.Cond[ThisClass_] * (ThisVoltage - ThisModel.ModelObject.EffVRev[ThisClass_]) * 1e-3;
            end else begin
              ThisAmp := ThisModel.ModelObject.Amp[ThisClass_] - BackAmp;
            end;

            //ExcNoise
            ExcNoise := 0;
            if ThisClass_ <> 0 then begin
              if Properties.UseCondVoltage then begin
                with ThisModel.ModelObject do begin
                  ExcNoise := CondStd[ThisClass_] * (ThisVoltage - EffVRev[ThisClass_]) * 1e-3;
                  // ExcNoise := ExcNoise * ExcNoise;
                end;
              end else begin
                ExcNoise := ThisModel.ModelObject.Std[ThisClass_];
                ExcNoise := ExcNoise * ExcNoise;
                if ExcNoise <= BackNoise2 then ExcNoise := 0 else ExcNoise := Sqrt(ExcNoise - BackNoise2);
              end;
            end;

          end;


          //If there is a Stimulus change, may have to cut short the event
          //For now, ignore the DeadTime issue
          if Properties.Stimulation <> SIM_STIM_NONE then begin
            ParDiff := False;
            if ((CurPoints + EvtDuration/Sampling) > NxtChPoint) and (NxtChPoint < BufferCount) then begin
              EvtDuration := (NxtChPoint - CurPoints + 1) * Sampling;
              ParDiff := True;
            end;
            if ParDiff then begin
              for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
                if prChan[ch] <> - 1 then begin
                  TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value := Vars[ch - 1, NxtChPoint] / VarScaling[ch - 1] + VarOffset[ch - 1];
                  IniV[ch - 1]                                      := Vars[ch - 1, NxtChPoint];
                  //ThisVoltage
                  if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
                    ThisVoltage := Vars[ch - 1, NxtChPoint] / VarScaling[ch - 1] + VarOffset[ch - 1];
                  end;
                end;
              end;
              MakeQMatrix(ThisModel.ModelObject, ThisModel.ModelObject.Q);
              MakePMatrix(ThisModel.ModelObject.Q, ThisModel.ModelObject.P, ThisModel.ModelObject.MathStateCount);
              //Find next change point
              ChPoint    := NxtChPoint;
              NxtChPoint := BufferCount;
              for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
                if prChan[ch] <> - 1 then begin
                  for pp := ChPoint + 1 to BufferCount - 1 do begin
                    if IniV[ch - 1] <> Vars[ch - 1, pp] then begin
                      if pp < NxtChPoint then NxtChPoint := pp;
                      Break;
                    end;
                  end;
                end;
              end;
            end;

          end;
          (*
          ChPoint := BufferCount - 1;
          if Properties.Stimulation <> SIM_STIM_NONE then begin
            //Does any parameter change within the Event ?
            ParDiff := False;
            for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
              if prChan[ch] <> - 1 then begin
                for pp := CurPoints to BufferCount - 1 do begin
                  if IniV[ch - 1] <> Vars[ch - 1, pp] then begin
                    if ((pp - CurPoints + 1{}) * Sampling) <={<} EvtDuration then begin//??????
                      EvtDuration := (pp - CurPoints {+ 1}) * Sampling;//??????
                      ParDiff := True;
                      if pp < ChPoint then ChPoint := pp;
                    end;
                    Break;
                  end;
                end;
              end;
            end;
            if ParDiff then begin
              for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
                if prChan[ch] <> - 1 then begin
                  TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value := Vars[ch - 1, ChPoint] / VarScaling[ch - 1];
                  IniV[ch - 1]                                      := Vars[ch - 1, ChPoint];
                  //ThisVoltage
                  if Properties.UseCondVoltage then begin
                    if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
                      ThisVoltage := TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value;
                    end;
                  end;
                end;
              end;
              ////ThisModel.ModelObject.UpdateParameters;
              ////ThisModel.ModelObject.ADt := Sampling;
              //ThisModel.ModelObject.CalculateMath;
              MakeQMatrix(ThisModel.ModelObject, ThisModel.ModelObject.Q);
              MakePMatrix(ThisModel.ModelObject.Q, ThisModel.ModelObject.P, ThisModel.ModelObject.MathStateCount);
            end;
          end;
          *)

          //TO DO: LM. Check this
          EvtPoints     := 0;
          if (CurDuration + EvtDuration) > BufDuration then begin
          //Goes past the end, thus don't change state - maybe in the next Buffer
            EvtDuration := BufDuration - CurDuration;
            EvtPoints   := BufferCount - CurPoints;
            DoChange    := False;
          end else begin
          //Do change state
            try
              EvtPoints   := Floor(EvtDuration / Sampling);
            except
            end;
          end;
          if EvtPoints = 0 then EvtPoints := 1;


          RandSeed := prRandSeedNoise;
          //Fill raw data buffer
          if Properties.Staircase then begin
            for p := CurPoints to CurPoints + EvtPoints - 1 do begin
              prBuffer[p] := Round(prBuffer[p] + (ThisAmp - Offset) * Scaling);
            end;
          end else begin
            if ThisClass_ <> 0 then begin
              for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                prBuffer[p] := Round(prBuffer[p] + (RandG(ThisAmp, ExcNoise) - Offset) * Scaling);
              end;
            end;
            if GateCount <> 0 then begin
              for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                prBuffer[p] := Round(prBuffer[p] + Scaling * (IGateSum[GateIx] - Offset));
                IGateSum[GateIx] := 0.0;
                Inc(GateIx);
                if GateIx = GateCount then GateIx := 0;
              end;
            end;
          end;
          prRandSeedNoise := RandSeed;


          //Fill IDL data buffer
          if (Properties.IdealizedData) and (m = 0) and (c = 0) then begin
            if Properties.Staircase then begin
            //Jumps
              if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                  prBufferStateIDL[p]   := prJumpState;
                end;
              end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                  prBufferStateIDL[p]   := prJumpCls;
                end;
              end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
                for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                  prBufferStateIDL[p]   := prJumpGrp;
                end;
              end;
              for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                prBufferStateIDLAmp[p]  := prJumpAmp;
              end;
            end else begin
              {
              if          Properties.IDLOutput = Integer(OUT_STATE) then begin
                for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                  prBufferStateIDL[p] := ThisClass;//State
                end;
              end else if Properties.IDLOutput = Integer(OUT_CLASS) then begin
                for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                  prBufferStateIDL[p] := ThisClass_;//Class
                end;
              end else if Properties.IDLOutput = Integer(OUT_GROUP) then begin
              //TODO: To be done
              end;
              }
              for p := CurPoints to CurPoints + EvtPoints - 1 do begin
                prBufferStateIDL[p] := ThisClass;//Class, State or Group
              end;
            end;
          end;

          CurPoints   := CurPoints + EvtPoints;
          CurDuration := CurPoints * Sampling;


          if (((Properties.Stimulation <> SIM_STIM_NONE) and (not ParDiff)) or (Properties.Stimulation = SIM_STIM_NONE)) and (StateCount > 1) and (CurPoints < BufferCount) then begin

            if DoChange then begin
            //Must Change State ?
              RandSeed := prRandSeedDwell;
              NewerState := DisDev(StateCount - 1, ThisModel.ModelObject.P[NewState]);
              prRandSeedDwell := RandSeed;
              if (NewerState >= NewState) then begin
                if NewerState < (StateCount - 1) then begin
                  Inc(NewerState);
                end;
              end;


              //Record transition ........................................
              //TODO: make property
              if (NewState = (Properties.TransitionStates[1]-1)) and (NewerState = (Properties.TransitionStates[2]-1)) then Inc(prTransitionCnt);
              TPMInteger(prTransitionCntAll[m])[NewState, NewerState] := TPMInteger(prTransitionCntAll[m])[NewState, NewerState] + 1;
              //..........................................................


              if (Properties.Staircase) and
                 (ThisModel.ModelObject.States[NewerState].Cls{Grp} <> ThisModel.ModelObject.States[NewState].Cls{Grp}) then begin

                if NewerState = NewState then begin
                  Beep;
                end;

                if Properties.StaircaseUseSteps then begin
                //Step from Param
                  RandSeed := prRandSeedDwell;
                  ThisStepIdx := DisDev(prStepCounts[prStepIdx], prStepProbs[prStepIdx]);
                  prRandSeedDwell := RandSeed;
                  ThisJumpAmp := RandG(prStepSizes[prStepIdx][ThisStepIdx], prStepSizes[prStepIdx][ThisStepIdx] * Properties.StaircaseStepStd);
                end else begin
                //Step from Model
                  ThisJumpAmp := RandG(ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewerState].Cls] -
                                       ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewState].Cls],
                                       Properties.StaircaseStepStd);
                end;

                prJumpAmp       := prJumpAmp + ThisJumpAmp;
                prJumpAvgSum    := prJumpAvgSum + Abs(ThisJumpAmp);
                prJumpStdSqrSum := prJumpStdSqrSum + Sqr(ThisJumpAmp - (ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewerState].Cls] - ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[NewState].Cls]));
                Inc(prJumpStdSqrN);


                //Odd/Even give Up/Down
                //For two steps Up --> must use only odd classes

                if Odd(ThisModel.ModelObject.States[NewerState].Cls) then begin
                //Up
                  if Properties.StaircaseNoJumpClass then begin
                    prJumpCls := (prJumpCls + 1) mod 2;
                  end else begin
                    Inc(prJumpCls);
                  end;
                  Inc(prJumpCnt[+1]);
                  Inc(prStepIdx);
                  if prStepIdx > (prStepCount - 1) then prStepIdx := 0;
                end else begin
                //Down
                  if Properties.StaircaseNoJumpClass then begin
                    prJumpCls := (prJumpCls + 1) mod 2;
                  end else begin
                    Dec(prJumpCls);
                  end;
                  Inc(prJumpCnt[-1]);
                  Dec(prStepIdx);
                  if prStepIdx < 0 then prStepIdx := (prStepCount - 1);
                end;

                {
                Cls:             2   2   0   0   1   1
                Amp:            -1  -1   0   0   1   1
                Grp:             0   1   0   1   0   1

                Cls:             4   4   2   2   0   0   1   1   3   3
                Amp:            -2  -2  -1  -1   0   0   1   1   2   2
                Grp:             0   1   0   1   0   1   0   1   0   1

                Cls:             4   2   0   1   3
                Amp:            -2  -1   0   1   2
                Grp:             0   0   0   0   0

                If Cls changed,
                   and it changed into MaxCls or MinCls,
                   and did not change into Cls 0 then state is reset into a Cls 0 state of the same Grp
                }

                //Testing
                NewState := NewerState;


                //if Grp < 0 then move to a state(-Grp), if it exists. If not, do old code
                Found := False;
                if ThisModel.ModelObject.States[NewerState].Gr < 0 then begin
                  for pp := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
                    if (ThisModel.ModelObject.States[pp].Gr = -(ThisModel.ModelObject.States[NewerState].Gr)) and
                       (pp <> NewerState)                                                                     then begin
                      NewState := pp;
                      Found    := True;
                      Break;
                    end;
                  end;
                end;

                if not Found then begin

                  if ThisModel.ModelObject.States[NewerState].Cls <> 0 then begin
                  //Not changed into Cls 0
                    if (ThisModel.ModelObject.States[NewerState].Cls = MinCls) or
                       (ThisModel.ModelObject.States[NewerState].Cls = MaxCls) then begin
                    //Terminal Cls, must Reset
                      for pp := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
                        if (ThisModel.ModelObject.States[pp].Gr = ThisModel.ModelObject.States[NewerState].Gr) and
                           (pp <> NewerState)                                                                  and
                           (ThisModel.ModelObject.States[pp].Cls = 0)                                          then begin
                          NewState := pp;
                        end;
                      end;
                    end;
                  end;

                end;//Found

                 //Testing - old code
                {
                for pp := 0 to ThisModel.ModelObject.States.StateCount - 1 do begin
                  if (ThisModel.ModelObject.States[pp].Gr = ThisModel.ModelObject.States[NewerState].Gr) and
                     (pp <> NewerState) and
                     (ThisModel.ModelObject.States[pp].Cls = 0) then begin
                    NewState := pp;
                  end;
                end;
                }

                //Testing:
                //Moved from here to right after JumpCls is set, above
                prJumpState := NewState;
                prJumpGrp   := ThisModel.ModelObject.States[NewState].Gr;

              end else begin

                DoGatingOnEvent(NewState, NewerState);
                NewState := NewerState;

              end;

            end;//DoChange

          end;

          prState[m, c] := NewState;



          //The remaining duration of the current event (if not an exact multiple of sampling points)
          EvtDuration := EvtDuration - Sampling * EvtPoints;

          prActualPoints := prActualPoints + EvtPoints;
          NewPercent  := Round(prActualPoints / prTotalPoints * 100);
          if (NewPercent <> prPercent) and (NewPercent <= 100) then begin
            prUpdateStatus(NewPercent, prStTime);
          end;


          (*
          //Change conditions ?

          if ParDiff then begin
            for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
              if prChan[ch] <> - 1 then begin
                TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value := Vars[ch - 1, NxtChPoint] / VarScaling[ch - 1];
                IniV[ch - 1]                                      := Vars[ch - 1, NxtChPoint];
              end;
            end;
            MakeQMatrix(ThisModel.ModelObject, ThisModel.ModelObject.Q);
            MakePMatrix(ThisModel.ModelObject.Q, ThisModel.ModelObject.P, ThisModel.ModelObject.MathStateCount);
            //Find next change point
            NxtChPoint := BufferCount;
            for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
              if prChan[ch] <> - 1 then begin
                for pp := CurPoints + 1 to BufferCount - 1 do begin
                  if IniV[ch - 1] <> Vars[ch - 1, pp] then begin
                    if pp < NxtChPoint then NxtChPoint := pp;
                    Break;
                  end;
                end;
              end;
            end;
          end;
          *)

        until(CurPoints >= BufferCount);


        //Idealization *********************************************************
        if (Properties.IdealizedData) and (m = 0) and (c = 0) then begin

          prevCls  := prBufferIdl[prBufferIDLCount - 1].Cls;

          for p := 0 to BufferCount - 1 do begin

            if prBufferStateIDL[p] <> prevCls then begin
            //New event

              //Close previous event, if at least one point  -  may only happen for the first point
              if (prBufferIdl[prBufferIDLCount - 1].Last >= prBufferIdl[prBufferIDLCount - 1].First) then begin

                //New IDL segment
                Inc(prBufferIDLCount);
                if prBufferIDLCount > prBufferIDLMaxCount then begin
                  Inc(prBufferIDLMaxCount, prBufferIDLInc);
                  try
                    ReAllocMem(prBufferIDL,    SizeOf(TQUB_QFS_QIF_FirstLastClass) * prBufferIDLMaxCount);
                    ReAllocMem(prBufferIDLAmp, SizeOf(Single)                      * prBufferIDLMaxCount);
                  except
                    prProperties.IdealizedData := False;
                    FreeMem(prBufferIDL);
                    prBufferIDL := nil;
                    FreeMem(prBufferIDLAmp);
                    prBufferIDLAmp := nil;
                    prBufferIDLCount := 0;
                    prBufferIDLMaxCount := 0;
                  end;
                end;
                if Properties.IdealizedData then begin
                  prBufferIdl[prBufferIDLCount - 1].First := prBufferIdl[prBufferIDLCount - 2].Last + 1;
                  prBufferIdl[prBufferIDLCount - 1].Last  := prBufferIdl[prBufferIDLCount - 1].First;
                  prBufferIdl[prBufferIDLCount - 1].Cls   := prBufferStateIDL[p];
                  prBufferIdlAmp[prBufferIDLCount - 1]    := prBufferStateIDLAmp[p];
                end else begin
                  Break;
                end;

              end else begin
                //Change class
                prBufferIdl[prBufferIDLCount - 1].Cls := prBufferStateIDL[p];
                prBufferIdlAmp[prBufferIDLCount - 1]  := prBufferStateIDLAmp[p];
                Inc(prBufferIdl[prBufferIDLCount - 1].Last);
              end;

              prevCls := prBufferStateIDL[p];

            end else begin

              Inc(prBufferIdl[prBufferIDLCount - 1].Last);

            end;//<> cls

          end;//p

        end;//Idealized ********************************************************

      end;//Channel

    end;//Model


    prSimulateBaseline(BufferCount, Sampling, Scaling, Offset);


    MPXData(prBuffer, BufferMPX, Data.DataObject.ADChannelCount, 0{response is always Channel 0}, Data.DataObject.ADDataSize, Data.DataObject.ADDataType, 0, BufferCount - 1, True{MPX});


  finally


    FreeMem(IniV);
    Free2(TPMChar(Vars));
    FreeMem(VarScaling);
    FreeMem(VarOffset);


  end;

end;


//****************************************************************************//
//                                                                            //
//       M A C R O S C O P I C     S I M U L A T I O N      E N G I N E       //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Simulate.prDeterministicSimulationEngine(BufferMPX :Pointer; BufferCount :Integer) :Boolean;
var
  ThisModel             :TQUB_QVO_QVM;
  StateCount            :Integer;
  Sampling              :Double;
  Scaling               :Double;
  Offset                :Double;
  BackAmp               :Double;
  BackNoise             :Double;
  BackNoise2            :Double;
  ThisAmp               :Double;
  ThisNoise             :Double;
  ThisVoltage           :Double;
  ThisVoltageCh         :Integer;
  NewPercent            :Integer;
  IniV                  :TPAInteger;
  Vars                  :TPMInteger;
  VarScaling            :TPADouble;
  VarOffset             :TPADouble;
  s, m, c, ch, p, pp    :Integer;
  i, j, jj              :Integer;
  ChPoint               :Integer;
  Prob                  :TPADouble;
  nc, n                 :Integer;
  pr                    :Double;
  Amp                   :Double;
  DoChange              :Boolean;
  FluxFrom              :Integer;
  Flux                  :Double;
  FluxMin, FluxMax      :Double;

  ILeak                 :Double;
begin

  Result                := False;

  Sampling              := Data.Sampling;
  Scaling               := Data.Scaling * Data.DataChannelScaling[0];
  Offset                := Data.DataChannelOffset[0];

  IniV                  := nil;
  Vars                  := nil;
  VarScaling            := nil;
  VarOffset             := nil;
  Prob                  := nil;

  ThisVoltage           := Properties.CondVoltage;

  DoChange              := False;

  if prFluxRate <> nil then begin
    FluxFrom              := Properties.FluxFrom - 1;
    FluxMin               := - Power(2, 8*Data.DataObject.ADDataSize-1) / prFluxScaling;
    FluxMax               := (Power(2, 8*Data.DataObject.ADDataSize-1) - 1) / prFluxScaling;
  end;


  try

    //Fill buffer with background noise

    //Uses class 0 of first model as the background noise
    ThisModel  := TQUB_QVO_QVM(prModels.Objects[0]);
    BackAmp    := ThisModel.ModelObject.Amp[0];
    BackNoise  := ThisModel.ModelObject.Std[0];
    BackNoise2 := BackNoise * BackNoise;

    if Properties.Stimulation <> SIM_STIM_NONE then begin

      IniV := IAlloc1(Data.DataObject.ADChannelCount - 1);{!!! subtract -1 from index}
      Vars := IAlloc2(Data.DataObject.ADChannelCount - 1, BufferCount);{!!! subtract -1 from first index}
      VarScaling := DAlloc1(Data.DataObject.ADChannelCount - 1);
      VarOffset  := DAlloc1(Data.DataObject.ADChannelCount - 1);

      //De-MUX the stimuli
      for c := 1 to Data.DataObject.ADChannelCount - 1 do begin
        if prChan[c] <> - 1 then begin
          MPXData(Vars[c - 1], BufferMPX, Data.DataObject.ADChannelCount, c, Data.DataObject.ADDataSize, Data.DataObject.ADDataType, 0, BufferCount - 1, False{MPX});
        end;

        VarScaling[c - 1] := Data.Scaling * Data.DataChannelScaling[c];
        VarOffset[c - 1]  := Data.DataChannelOffset[c];
      end;

    end;

    RandSeed := prRandSeedNoise;
    if ThisModel.ModelObject.States.StateCount > 0 then begin
      if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
        for p := 0 to BufferCount - 1 do begin
          prBuffer[p] := Round((RandG(BackAmp, BackNoise) - Offset) * Scaling);
          {
          if (nar > 0) then begin
            for (j = 1; j <= nar && j <= n; j++) do begin
              noise0[n] -= (float)ar[0][j]*noise0[n - j];
            end;;
          end;
          }
        end;
      end else begin
        for p := 0 to BufferCount - 1 do begin
          prBuffer[p] := Round((BackAmp - Offset) * Scaling);
        end;
      end;
    end else begin
      for p := 0 to BufferCount - 1 do begin
        prBuffer[p] := 0;
      end;
    end;
    prRandSeedNoise := RandSeed;


    //Add ILeak? ###############################################################
    ILeak := 0;
    if Properties.ILeak then begin

      ThisVoltageCh := -1;
      if Properties.Stimulation <> SIM_STIM_NONE then begin
      //Stimulation
        for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
          if prChan[ch] <> - 1 then begin
            if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
              ThisVoltageCh := ch;
              Break;
            end;
          end;
        end;
      end;

      if ThisVoltageCh = -1 then begin
      //No channel for ThisVoltage

        ILeak := Properties.GLeak*1e3 * (ThisVoltage - Properties.VLeak);
        for p := 0 to BufferCount - 1 do begin
          prBuffer[p] := Round(prBuffer[p] + (ILeak - Offset) * Scaling);
        end;

      end else begin
      //There is a channel for ThisVoltage

        for p := 0 to BufferCount - 1 do begin
          ThisVoltage := Vars[ThisVoltageCh - 1, p] / VarScaling[ThisVoltageCh - 1] + VarOffset[ThisVoltageCh - 1];//mV
          ILeak := Properties.GLeak*1e3 * (ThisVoltage - Properties.VLeak);
          prBuffer[p] := Round(prBuffer[p] + (ILeak - Offset) * Scaling);
        end;

      end;

    end;//ILeak?  ##############################################################


    for m := 0 to prModels.Count - 1 do begin
    //Model Loop

      ThisModel  := TQUB_QVO_QVM(prModels.Objects[m]);
      StateCount := ThisModel.ModelObject.States.StateCount;
      if StateCount = 0 then Continue;

      ReAllocMem(Prob, SizeOf(Double) * StateCount);

      if Properties.Stimulation = SIM_STIM_NONE then begin
        {???}//ThisModel.ModelObject.UpdateParameters;
        ThisModel.ModelObject.ADt := Sampling;
        ThisModel.ModelObject.CalculateMath;
      end;

      if Properties.Stimulation <> SIM_STIM_NONE then begin
        //Initialize parameters
        for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
          if prChan[ch] <> - 1 then begin
            TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value := Vars[ch - 1, 0] / VarScaling[ch - 1] + VarOffset[ch - 1];
            IniV[ch - 1]                                      := Vars[ch - 1, 0];
            //ThisVoltage
            if Properties.UseCondVoltage then begin
              if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
                ThisVoltage := TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value;
              end;
            end;
          end;
        end;
        {???}//ThisModel.ModelObject.UpdateParameters;
        ThisModel.ModelObject.ADt := Sampling;
        ThisModel.ModelObject.CalculateMath;
      end;

      //Find next change in Stimuli
      ChPoint := BufferCount;//Next change point
      if Properties.Stimulation <> SIM_STIM_NONE then begin
        //Does any parameter change ?
        for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
          if prChan[ch] <> - 1 then begin
            for pp := 0 to ChPoint - 1 do begin
              if IniV[ch - 1] <> Vars[ch - 1, pp] then begin
                ChPoint := pp;
                Break;
              end;
            end;
          end;
        end;
      end;

      if prFluxRate <> nil then prFluxRateValue := prFluxRate.Value[prFluxRateDir];


      //Main loop
      for p := 0 to BufferCount - 1 do begin

        if prStopped then Exit;

        if Properties.Stimulation <> SIM_STIM_NONE then begin

          if p = ChPoint then begin
          //Change point
            DoChange := True;

            if Properties.UseCondVoltage then begin
              for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
                if prChan[ch] <> - 1 then begin
                  if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
                    ThisVoltage := Vars[ch - 1, ChPoint] / VarScaling[ch - 1] + VarOffset[ch - 1];
                  end;
                end;
              end;
            end;

          end else begin
            DoChange := False;
          end;

          if p = ChPoint + 1{sic} then begin
          //This is immediately past a change point
          //Do the parameter change and update QAP
            (*
            for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
              if prChan[ch] <> - 1 then begin
                TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value := Vars[ch - 1, ChPoint] / VarScaling[ch - 1];
                IniV[ch - 1]                                      := Vars[ch - 1, ChPoint];
                //ThisVoltage
                if Properties.UseCondVoltage then begin
                  if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
                    ThisVoltage := TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value;
                  end;
                end;
              end;
            end;
            {???}//ThisModel.ModelObject.UpdateParameters;
            ThisModel.ModelObject.ADt := Sampling;
            ThisModel.ModelObject.CalculateMath;
            *)

            (*
            //Find the next change point
            ChPoint := BufferCount;//Next change point
            for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
              if prChan[ch] <> - 1 then begin
                for pp := p + 1 to ChPoint - 1 do begin
                  if IniV[ch - 1] <> Vars[ch - 1, pp] then begin
                    ChPoint := pp;
                    Break;
                  end;
                end;
              end;
            end;
            *)
          end;
        end;//Stimulated

        //Advance the probability vector
        if not ((prCurSegmentCount = 0) and (p = 0)) then begin
        //if not the very first point
          for s := 0 to StateCount - 1 do begin
            Prob[s] := 0;
            for pp := 0 to StateCount - 1 do begin
              Prob[s] := Prob[s] + prProbs[m, pp] * ThisModel.ModelObject.A[pp, s];
            end;
          end;
          //Rescale probabilities
          pr := 0;
          for s := 0 to StateCount - 1 do begin
            pr := pr + Prob[s];
          end;
          if pr = 0 then begin
            for s := 0 to StateCount - 1 do begin
              prProbs[m, s] := 1 / StateCount;
            end;
          end else begin
            for s := 0 to StateCount - 1 do begin
              prProbs[m, s] := Prob[s] / pr;
            end;
          end;
        end;

        //Add amplitude
        RandSeed := prRandSeedNoise;
        Amp := 0;
        for s := 0 to StateCount - 1 do begin
          if ThisModel.ModelObject.States[s].Cls <> 0 then begin
            ThisAmp := ThisModel.ModelObject.Amp[ThisModel.ModelObject.States[s].Cls];
            if Properties.UseCondVoltage then begin
              ThisAmp := ThisModel.ModelObject.Cond[ThisModel.ModelObject.States[s].Cls] * (ThisVoltage - ThisModel.ModelObject.EffVRev[ThisModel.ModelObject.States[s].Cls]) * 1e-3;
            end;
            if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then begin
              if Properties.UseCondVoltage then begin
                ThisNoise := ThisModel.ModelObject.CondStd[ThisModel.ModelObject.States[s].Cls] * (ThisVoltage - ThisModel.ModelObject.EffVRev[ThisModel.ModelObject.States[s].Cls]) * 1e-3;
                ThisNoise := Sqrt((ThisNoise * ThisNoise) * prProbs[m, s] * ThisModel.ModelObject.ChannelCount);
              end else begin
                ThisNoise := ThisModel.ModelObject.Std[ThisModel.ModelObject.States[s].Cls];
                ThisNoise := ThisNoise * ThisNoise;
                if ThisNoise <= BackNoise2 then begin
                  ThisNoise := 0;
                end else begin
                  ThisNoise := Sqrt((ThisNoise - BackNoise2) * prProbs[m, s] * ThisModel.ModelObject.ChannelCount);
                end;
              end;
              Amp       := Amp + RandG(ThisAmp * prProbs[m, s] * ThisModel.ModelObject.ChannelCount, ThisNoise);
            end else begin
              Amp       := Amp +       ThisAmp * prProbs[m, s] * ThisModel.ModelObject.ChannelCount;
            end;
          end;
        end;
        prRandSeedNoise := RandSeed;

        prBuffer[p] := Round(prBuffer[p] + (Amp - BackAmp - Offset) * Scaling);

        if DoChange then begin
        //Do the parameter change and update QAP
          for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
            if prChan[ch] <> - 1 then begin
              TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value := Vars[ch - 1, ChPoint] / VarScaling[ch - 1] + VarOffset[ch - 1];
              IniV[ch - 1]                                      := Vars[ch - 1, ChPoint];
              //ThisVoltage
              if Properties.UseCondVoltage then begin
                if TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Name = 'Voltage' then begin
                  ThisVoltage := TQUB_ExpCond_PParameter(prPars[prChan[ch]]).Value;
                end;
              end;
            end;
          end;
          {???}//ThisModel.ModelObject.UpdateParameters;
          ThisModel.ModelObject.ADt := Sampling;
          ThisModel.ModelObject.CalculateMath(False);

          if prFluxRate <> nil then prFluxRateValue := prFluxRate.Value[prFluxRateDir];

          ChPoint := BufferCount;//Next change point
          for ch := 1 to Data.DataObject.ADChannelCount - 1 do begin
            if prChan[ch] <> - 1 then begin
              for pp := p + 1 to ChPoint - 1 do begin
                if IniV[ch - 1] <> Vars[ch - 1, pp] then begin
                  ChPoint := pp;
                  Break;
                end;
              end;
            end;
          end;
        end;

        //Calc Flux
        if prFluxBuffer <> nil then begin
          Flux            := prFluxRateValue * prProbs[0, FluxFrom];
          if (not prWarning) and (Flux < FluxMin) or (Flux > FluxMax) then begin
            prWarning     := True;
            prWarningMsg  := 'Flux went out-of-range; try a smaller calibration factor.';
            prWikiTopic   := 'Warning:Flux_out-of-range';
          end;
          prFluxBuffer[p] := Round(Flux * prFluxScaling);
        end;

        prActualPoints := prActualPoints + 1;
        NewPercent  := Round(prActualPoints / prTotalPoints * 100);
        if (NewPercent <> prPercent) and (NewPercent <= 100) then begin
          prUpdateStatus(NewPercent, prStTime);
        end;

      end;//p;

    end;//Model


    {if not (Properties.Response = SIM_RESPONSE_DETERMINISTIC) then }prSimulateBaseline(BufferCount, Sampling, Scaling, Offset);


    MPXData(prBuffer, BufferMPX, Data.DataObject.ADChannelCount, 0{response is always Channel 0}, Data.DataObject.ADDataSize, Data.DataObject.ADDataType, 0, BufferCount - 1, True{MPX});

    if prFluxBuffer <> nil then begin
      MPXData(prFluxBuffer, BufferMPX, Data.DataObject.ADChannelCount, Data.DataObject.ADChannelCount - 1{flux is last}, Data.DataObject.ADDataSize, Data.DataObject.ADDataType, 0, BufferCount - 1, True{MPX});
    end;

  finally


    FreeMem(VarScaling);
    FreeMem(VarOffset);
    Free2(TPMChar(Vars));
    FreeMem(IniV);
    FreeMem(Prob);


  end;

end;


//Baseline
procedure TQUB_Task_Simulate.prSimulateBaseline(BufferCount :Integer; Sampling, Scaling, Offset :Double);
var
  p, c                  :Integer;
  BaseCurPoints         :Integer;
  BasePoints            :Integer;
  BaseCurDuration       :Double;
  BaseDuration          :Double;
  BaseState             :Integer;
  BaseBuffer            :TPASingle;
  FlucBuffer            :TPAInteger;
  BufDuration           :Double;
  Frq, Phs              :Double;
  CloseCount            :Integer;
  e, t                  :Double;
begin

  if (not Properties.BaselineFluct) and (not Properties.BaselineJumps) and (not Properties.BaselinePeriodic) and (not Properties.BaselineDrift) and (not Properties.BaselineLGM) then Exit;

  BaseBuffer            := nil;

  BufDuration           := BufferCount * Sampling;

  try

    //Allocate the common buffer
    BaseBuffer := SAlloc1(BufferCount);//must be initialized to zero
    for p := 0 to BufferCount - 1 do begin
      BaseBuffer[p] := 0.0;
    end;


    //Baseline Fluctuations
    if Properties.BaselineFluct then begin

      FlucBuffer := nil;

      try

        FlucBuffer := IAlloc1(BufferCount);//must be initialized to zero
        for p := 0 to BufferCount - 1 do begin
          FlucBuffer[p] := 0;
        end;

        CloseCount := Abs((Properties.BaselineFluctN - Abs(prBaselineFluctStateCount)) div 2);
        if prBaselineFluctStateCount < 0 then CloseCount := -CloseCount;

        for c := 1 to Properties.BaselineFluctN do begin

          //Event generator loop
          BaseCurPoints   := 0;
          BaseCurDuration := 0;
          BaseDuration    := 0;
          if c > CloseCount then BaseState := 1 else BaseState := -1;

          repeat

            //Generate a baseline event at least Sampling long
            while (BaseDuration < Sampling) do begin
              if prStopped then Exit;
              BaseState := -BaseState;
              BaseDuration := BaseDuration + ExpDev * Properties.BaselineFluctLifetime;
            end;

            if (BaseCurDuration + BaseDuration) > BufDuration then begin
              BasePoints   := BufferCount - BaseCurPoints;
            end else begin
              BasePoints   := Floor(BaseDuration / Sampling);
            end;
            if BasePoints = 0 then BasePoints := 1;
            BaseDuration := BasePoints * Sampling;

            for p := BaseCurPoints to BaseCurPoints + BasePoints - 1 do begin
              FlucBuffer[p] := FlucBuffer[p] + BaseState;
            end;

            //The remaining duration
            BaseDuration := BaseDuration - Sampling * BasePoints;

            BaseCurPoints   := BaseCurPoints + BasePoints;
            BaseCurDuration := BaseCurPoints * Sampling;

          until(BaseCurPoints >= BufferCount);

        end;//of the ith / N particle

        t := Properties.BaselineFluctMaxAmp / Properties.BaselineFluctN;
        for p := 0 to BufferCount - 1 do begin
          try
            BaseBuffer[p] := BaseBuffer[p] + t * FlucBuffer[p];
          except
          end;
        end;

        prBaselineFluctStateCount := FlucBuffer[BufferCount - 1];

      finally

        FreeMem(FlucBuffer);

      end;

    end;//of BaselineFluct


    //Baseline Drift
    if Properties.BaselineDrift then begin

      for p := 0 to BufferCount - 1 do begin
        try
          BaseBuffer[p] := BaseBuffer[p] + Properties.BaselineDriftSlope * (p + prCurSegmentCount) * Sampling + Properties.BaselineDriftIntercept;
        except
        end;
      end;

    end;

    //Baseline Periodic
    if Properties.BaselinePeriodic then begin

      Frq   := Properties.BaselinePeriodicFreq * 2 * Pi;
      Phs   := Properties.BaselinePeriodicPhase * Pi / 180;

      for p := 0 to BufferCount - 1 do begin
        try
          BaseBuffer[p] := BaseBuffer[p] + Properties.BaselinePeriodicAmp * Sin(Frq * (p + prCurSegmentCount) * Sampling + Phs);
        except
        end;
      end;

    end;

    //Baseline Jumps
    if Properties.BaselineJumps then begin

      //Event generator loop
      BaseCurPoints   := 0;
      BaseCurDuration := 0;
      BaseDuration    := prBaselineJumpsStateDur;

      repeat

        if prStopped then Exit;

        //Generate a baseline event at least Sampling long
        while (BaseDuration < Sampling) do begin

          if prStopped then Exit;

          //Generate new jump
          prBaselineJumpsStateAmpF  := (2 * Round(Random) - 1) * RandG(prBaselineJumpsStateAmpC, Properties.BaselineJumpsAmpStd);
          prBaselineJumpsStateAmpL  := prBaselineJumpsStateAmpC;

          prBaselineJumpsStateStart := prCurSegmentCount + BaseCurPoints;
          prBaselineJumpsStateDur   := ExpDev * Properties.BaselineJumpsLifetime;
          BaseDuration := BaseDuration + prBaselineJumpsStateDur;

        end;

        if (BaseCurDuration + BaseDuration) > BufDuration then begin
          BasePoints   := BufferCount - BaseCurPoints;
        end else begin
          BasePoints   := Floor(BaseDuration / Sampling);
        end;
        if BasePoints = 0 then BasePoints := 1;
        BaseDuration := BasePoints * Sampling;

        e := prBaselineJumpsStateAmpC;
        for p := BaseCurPoints to BaseCurPoints + BasePoints - 1 do begin
          try
            e :=     Properties.BaselineJumpsFastW * exp(-(prCurSegmentCount + p - prBaselineJumpsStateStart) * Sampling / Properties.BaselineJumpsFast{sec});
            e := e + (1 - Properties.BaselineJumpsFastW) * exp(-(prCurSegmentCount + p - prBaselineJumpsStateStart) * Sampling / Properties.BaselineJumpsSlow{sec});
            e := prBaselineJumpsStateAmpL + prBaselineJumpsStateAmpF * e;
            BaseBuffer[p] := BaseBuffer[p] + e;
          except
          end;
        end;
        prBaselineJumpsStateAmpC := e;

        //The remaining duration
        BaseDuration := BaseDuration - Sampling * BasePoints;
        prBaselineJumpsStateDur := BaseDuration;

        BaseCurPoints   := BaseCurPoints + BasePoints;
        BaseCurDuration := BaseCurPoints * Sampling;

      until(BaseCurPoints >= BufferCount);


    end;//of Baseline Jumps

    //Baseline LGM
    if Properties.BaselineLGM then begin

      if Properties.BaselineLGM2State then begin
      //2 states: x and x'

        for p := 0 to BufferCount - 1 do begin

          BaseBuffer[p] := BaseBuffer[p] + prBaselineLGMState[0];
          prBaselineLGMState[0] := prBaselineLGMState[0] + prBaselineLGMState[1];
          prBaselineLGMState[1] := prBaselineLGMState[1] + RandG(0, Properties.BaselineLGMProcStd);

        end;

      end else begin
      //one state: x

        for p := 0 to BufferCount - 1 do begin

          BaseBuffer[p] := BaseBuffer[p] + prBaselineLGMState[0];
          prBaselineLGMState[0] := prBaselineLGMState[0] + RandG(0, Properties.BaselineLGMProcStd);

        end;

      end;

    end;


    //Add baseline
    for p := 0 to BufferCount - 1 do begin

      prBuffer[p] := prBuffer[p] + Round((BaseBuffer[p] - Offset) * Scaling);

    end;


  finally

    FreeMem(BaseBuffer);

  end;

end;


//******************************************************************************
procedure TQUB_Task_Simulate.OnGetProperties(Sender :TObject);
var
  i :Integer;
begin
  with Sender as TQUB_Task_Simulate_PropertiesForm do begin
    Self.Properties             := PropertiesRec;
    prStructureFiles.Assign(StructureFiles.Cols[1]);
    prStructureFiles.Delete(0);
    //Delete empty strings
    for i := prStructureFiles.Count - 1 downto 0 do begin
      if Trim(prStructureFiles.Strings[i]) = '' then begin
        prStructureFiles.Delete(i);
      end;
    end;
    prMovieName := ChangeFileExt(ExtractFileName(txtStructuresFile.Text), '');;

    if 0 = Properties['Gating'].Data[0] then begin
      prGatingScale := 0.0;
    end else if GATING_V = Properties['GatingUnits'].Data[0] then begin
      prGatingScale := 1.0;
    end else if GATING_MV = Properties['GatingUnits'].Data[0] then begin
      prGatingScale := 1000.0;
    end else begin
      prGatingScale := Properties['GatingUnitsOther'].Data[0];
    end;
    if 0<>Properties['GatingInvert'].Data[0] then prGatingScale := prGatingScale * -1.0;
  end;//with

end;

procedure TQUB_Task_Simulate.OnPutProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_Simulate_PropertiesForm do begin
  end;//with
end;

//******************************************************************************
procedure TQUB_Task_Simulate.prSetData(Data :TQUB_QVO_QVD);
begin
  prData := Data;
end;

procedure TQUB_Task_Simulate.prSetStimData(Data :TQUB_QVO_QVD);
begin
  prStimData := Data;
end;

procedure TQUB_Task_Simulate.prSetStimProto(Proto :TQUB_QVO_QVP);
begin
  prStimProto := Proto;
end;


//******************************************************************************

constructor TQUB_Task_Simulate_Stored.Create(Task :TQUB_Task_Simulate);
begin
  prModels                        := TStringList.Create;
  prModels.Assign(Task.Models);
  prData                          := Task.Data;
  prStimData                      := Task.StimData;
  prStimProto                     := Task.StimProto;
  prModelParameters               := TQUB_ExpCond_Parameters.CreateCopy(Task.ModelParameters);
  prProperties                    := Task.Properties;
end;

destructor TQUB_Task_Simulate_Stored.Destroy;
begin
  try
    prModels.Free;
    prModelParameters.Free;
  finally
    inherited;
  end;
end;

function ObjectInList(Fi :TObject; List :TStrings): Boolean;
var
  j           :Integer;
begin
  Result      := False;
  for j := 0 to List.Count - 1 do begin
    if Fi = List.Objects[j] then begin
      Result  := True;
      Exit;
    end;
  end;
end;

function TQUB_Task_Simulate_Stored.CanSimulate: Boolean;
var
  i             :Integer;
begin
  Result        := False;

  if not ObjectInList(prData, ViewportData.ListObjects.Items) then Exit;

  for i := 0 to prModels.Count - 1 do begin
    if not ObjectInList(prModels.Objects[i], ViewportModel.ListObjects.Items) then Exit;
  end;

  if prProperties.Stimulation = SIM_STIM_FILE then begin
    if prStimData <> nil then begin
      if not ObjectInList(prStimData, ViewportData.ListObjects.Items) then Exit;
    end else if prStimProto <> nil then begin
      if not ObjectInList(prStimProto, ViewportProtocol.ListObjects.Items) then Exit;
    end;
  end;

  Result        := True;
end;

function TQUB_Task_Simulate_Stored.ShouldSimulate: Boolean;
begin
  Result   :=     ObjectInList(ViewportModel.CurrModel, prModels)
              and (prData = ViewportData.CurrData);
end;

function TQUB_Task_Simulate_Stored.Simulate(OnTerminate :TNotifyEvent): TQUB_Task_Simulate;
begin
  Result                := TQUB_Task_Simulate.Create(True{Suspended});
  Result.OnTerminate    := OnTerminate;

  try

    Result.Models.Assign(prModels);
    Result.Data         := prData;
    Result.StimData     := prStimData;
    Result.StimProto    := prStimProto;
    Result.ModelParameters := prModelParameters;
    Result.Properties   := prProperties;

    Result.OnChange     := TaskView.OnTaskChange;
    Result.SilentOutput := True;

    if prData.LockedAny then begin
      // MessageDlg('Data file is in use', mtError, [mbOK], 0);
      Result.Terminate;
      Exit;
    end;

    if prData.Clear then begin
      Result.Terminate;
      Exit;
    end;

  finally

    Result.Resume;

  end;

end;

//******************************************************************************

constructor TQUB_Task_SimLive.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'Sim Live';

  prNeedsUpdate         := True;
  prSignal              := TEvent.Create(nil, False, False, '');
  prUpdating            := False;
  prSim                 := nil;

  prModel               := ViewportModel.CurrModel;

  ViewportModel.Events.AddEvent('OnChangeFile', prOnChangeModel);
  if prModel <> nil then prModel.Events.AddEvent('OnEdit', prOnModelEdit);

end;

destructor TQUB_Task_SimLive.Destroy;
begin
  Synchronize(prClearEvents);
  prSignal.Free;

  inherited;
end;

procedure TQUB_Task_SimLive.prClearEvents;
begin
  ViewportModel.Events.DelEvent('OnChangeFile', prOnChangeModel);
  if prModel <> nil then begin
    prModel.Events.DelEvent('OnEdit', prOnModelEdit);
    //prModel.AllowNumEdit := False;
  end;
end;

//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_SimLive.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_SimLive.OnPutProperties(Sender :TObject);
begin
end;


//****************************************************************************//


procedure TQUB_Task_SimLive.Execute;
begin
  while not prStopped do begin
    while prNeedsUpdate and not prSimRunning do begin
      prNeedsUpdate := False;
      Synchronize(prRunSim);
      while prUpdating do begin
        prSignal.WaitFor(1000);
      end;
    end;
    prSignal.WaitFor(1000);
  end;
end;

procedure TQUB_Task_SimLive.prRunSim;
var
  Stored    :IQUB_Task_Simulate_Stored;
begin
  if prModel = nil then Exit;
  if prModel.SimStored = nil then Exit;
  Stored    := IQUB_Task_Simulate_Stored(prModel.SimStored);
  if not Stored.CanSimulate then begin
    prModel.SimStored := nil;
    Exit;
  end;
  if not Stored.ShouldSimulate then Exit;

  prUpdating := True;
  // prModel.AllowNumEdit := True;
  prSim      := Stored.Simulate(prOnSimDone);
end;

procedure TQUB_Task_SimLive.Stop;
begin
  inherited;
  if prSim <> nil then prSim.Stop;
  prSignal.SetEvent;
end;

procedure TQUB_Task_SimLive.Update;
begin
  prNeedsUpdate := True;
  prSignal.SetEvent;
end;

procedure TQUB_Task_SimLive.prOnSimDone(Sender :TObject);
begin
  //if prModel <> nil then prModel.AllowNumEdit := False;
  prUpdating := False;
  prSim      := nil;
  prSignal.SetEvent;
end;

procedure TQUB_Task_SimLive.prOnChangeModel(Sender :TObject);
begin
  if prModel <> nil then begin
    prModel.Events.DelEvent('OnEdit', prOnModelEdit);
    prModel.AllowNumEdit := False;
  end;
  prModel := ViewportModel.CurrModel;
  if prModel <> nil then begin
    prModel.Events.AddEvent('OnEdit', prOnModelEdit);
    if prSim <> nil then prModel.AllowNumEdit := True;
    Update;
  end;
end;

procedure TQUB_Task_SimLive.prOnModelEdit(Sender :TObject);
begin
  Update;
end;

function TQUB_Task_SimLive.prSimRunning: Boolean;
begin
  Synchronize(prCheckSimRunning);
  Result := prSimIsRunning;
end;

procedure TQUB_Task_SimLive.prCheckSimRunning;
begin
  prSimIsRunning := TaskView.HasTaskOfType(TQUB_Task_Simulate);
end;


//******************************************************************************
end.

