{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// computes the stats returned by SKM without having to re-idealize


unit QUB_Task_Stat;


interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QVO_QVM, QUB_QVO_QVR,

QUB_QFS_QMF,

QUB_Viewport_Data,
QUB_Viewport_Histogram,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_Stat_Properties,

Chart, Series,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python, QUB_Viewport_Results,
QUB_Results_Histograms,
QUB_GUIInterface;


type


TQUB_Task_Stat  = class(TQUB_Task_Data)
protected

  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  //Parameters
  //prParams                 :TQUB_Mil_Params;

  prConfig                 :INode;
  prDataDescription        :INode;
  prDatasets               :Array of INode;

  prResultNode             :INode;
  prReportNode             :INode;


  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  procedure prPublishResults; virtual;

  //Chunks
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  function prReportCallback( arg :INode ): INode;
  procedure prDoReport;

public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Stop; override;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


end;


TQUB_Task_IVPlot  = class(TQUB_Task_Stat)
protected

  function prDoExecuteInitialize :Boolean; override;
  procedure prPublishResults; override;

//public

  //constructor Create(Suspended :Boolean); override;

end;


function StatDataSets(DataSetPtrs :TPAPointer; DataSetCount :Integer; Config :INode; DataCB, ReportCB :TNodeMethPtr): INode;
function StatDataSet(DataSet :INode; Config :INode; DataCB, ReportCB :TNodeMethPtr): INode;

function StatFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; fillDataCB, reportCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_stattreeiface@16';


implementation


uses


QUB_Python_Data,
QUB_Main;


procedure StatComputeGaps(StatResult :INode); forward;

function StatDataSets(DataSetPtrs :TPAPointer; DataSetCount :Integer; Config :INode; DataCB, ReportCB :TNodeMethPtr): INode;
var
  NullTermDS      :Array of Pointer;
  i               :Integer;
  rptCB, dCB      :INodeCFunc;
  ResultNodePtr   :TQUB_QFS_PNodeMem;
begin
  Result          := TNode.CreateNull;
  if DataSetCount <= 0 then Exit;

  SetLength(NullTermDS, DataSetCount+1);
  for i := 0 to DataSetCount - 1 do begin
    NullTermDS[i] := DataSetPtrs[i];
  end;
  NullTermDS[ DataSetCount ] := nil;

  rptCB     := TNodeCFunc.Wrap( TNodeMeth.Create(ReportCB) );
  dCB       := TNodeCFunc.Wrap( TNodeMeth.Create(DataCB) );

  ResultNodePtr := StatFromDll( Config.GetNodeMem, @(NullTermDS[0]), dCB, rptCB );
  Result        := TNode.CreateUsing( ResultNodePtr );
  Node_DECREF( ResultNodePtr );

  StatComputeGaps(Result);
end;

function StatDataSet(DataSet :INode; Config :INode; DataCB, ReportCB :TNodeMethPtr): INode;
var
  DataSetPtr  :Pointer;
begin
  DataSetPtr  := DataSet.GetNodeMem;
  Result      := StatDataSets(@DataSetPtr, 1, Config, DataCB, ReportCB);
end;

procedure StatComputeGaps(StatResult :INode);
var
  DS, Seg     :INode;
  Data        :TQUB_QVO_QVD;
  LastSegIx   :Integer;
  LastEndIx   :Integer;
  SegIx       :Integer;
  SegStartIx  :Integer;
begin
  DS          := StatResult.Find('DataSet');
  while not DS.IsNull do begin
    Data      := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DS['FileName'].DataAsString]);
    if Data <> nil then begin
      LastSegIx := -1;
      Seg       := DS.Find('Segment');
      while not Seg.IsNull do begin
        SegIx   := Data.WhichSegmentForPoint(Seg.Data[0]);
        if SegIx <> LastSegIx then begin
          SegStartIx  := Data.WhichSegmentFirstForSegment(SegIx);
          Seg['interburst'].SetData(Double(1e3*Data.DataObject.Sampling*(Seg.Data[0] - SegStartIx)));
          LastSegIx   := SegIx;
        end else begin
          Seg['interburst'].SetData(Double(1e3*Data.DataObject.Sampling*(Seg.Data[0] - LastEndIx)));
        end;
        Seg['interburst'].SetLineComment('ms');
        LastEndIx   := Seg.Data[1] + 1;
        Seg         := Seg.SiblingSameName;
      end;
    end;
    DS        := DS.SiblingSameName;
  end;
end;

//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _S t a t                       //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_Stat.Create(Suspended :Boolean);
begin
  inherited;

  prConfig     := TNode.Create('Properties');
  prConfig['StopFlag'].SetData( 0 );

  prName       := 'Stat';
end;

destructor TQUB_Task_Stat.Destroy;
begin
  inherited;
end;


//******************************************************************************
procedure TQUB_Task_Stat.Stop;
begin

  prConfig['StopFlag'].Data[0] := 1;

  inherited;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Stat.prDoTask;
var
  i               :Integer;

  DataSetPtrs     :Array of Pointer;
  ResultNode      :INode;
  DS, Seg         :INode;
begin
  if Length(prDataSets) = 0 then begin
    Exit;
  end;

  SetLength(DataSetPtrs, Length(prDataSets));
  for i := 0 to Length(prDataSets) - 1 do begin
    DataSetPtrs[i] := Pointer(prDataSets[i].GetNodeMem);
  end;

  try
    prConsoleMsg  := 'Stat running ... ';
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);

    ResultNode := StatDataSets(@(DataSetPtrs[0]), Length(DataSetPtrs), prConfig, DataNoder.FillData, prReportCallback);

    DS := ResultNode.Find('DataSet');
    while not DS.IsNull do begin
      Seg := DS.Find('Segment');
      while not Seg.IsNull do begin
        Seg.RemoveChild( Seg['Firsts'] );
        Seg.RemoveChild( Seg['Lasts'] );
        Seg.RemoveChild( Seg['Classes'] );
        Seg.RemoveChild( Seg['Durations'] );
        Seg.RemoveChild( Seg['amp'] );
        Seg.RemoveChild( Seg['sd'] );
        Seg['length'].SetData( (Seg.Data[1] - Seg.Data[0] + 1) * DS['sampling'].Data[0] );
        Seg := Seg.SiblingSameName;
      end;
      DS := DS.SiblingSameName;
    end;

    prResultNode := ResultNode;
    Synchronize( prPublishResults );
  except
    prError    := True;
    prErrorMsg := 'Unknown error in Stat library';
    prWikiTopic := 'Error:Task';
  end;

end;

procedure TQUB_Task_Stat.prPublishResults;
begin
  if prResultNode.Find('ResultType').isNull then
    prResultNode['ResultType'].SetData( 'Stat' );
    
  if (prDataDescription <> nil) and not prDataDescription.IsNull then prResultNode.AppendChild( prDataDescription );

  prResultNode['Hold'].SetData( 1 );

  if prSelectionList <> nil then begin          // if the selection list is 'Stat Results',
    prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
    prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
  end;

  ViewportResults.AddResults( prResultNode, False );
end;

procedure TQUB_Task_IVPlot.prPublishResults;
var
  DS, Seg, Hist         :INode;
  Charts, Chart         :INode;
  i                     :Integer;
  ChannelIndex          :Integer;
begin
  // discard histograms, recommend current vs. stimulus charts

  prResultNode['ResultType'].SetData( 'IV' );

  DS                    := prResultNode.Find('DataSet');
  while not DS.IsNull do begin
    Seg                 := DS.Find('Segment');
    while not Seg.IsNull do begin
      Hist              := Seg.Find('HistogramGroup');
      if not Hist.IsNull then Seg.RemoveChild(Hist);

      Seg               := Seg.SiblingSameName;
    end;

    DS                  := DS.SiblingSameName;
  end;

  ChannelIndex          := prConfig['ChannelIndex'].Data[0];
  Charts                := prResultNode['DefaultChartsVs'];
  for i := TQUB_QVO_QVD(Data).DataObject.ADChannelCount - 1 downto 0 do begin
    if i <> ChannelIndex then begin
      Chart               := Charts.InsertChild(QUB_Node_Null, TQUB_QVO_QVD(Data).DataChannelName[i] + ' Amp');
      Chart.SetData(TQUB_QVO_QVD(Data).DataChannelName[ChannelIndex] + ' Amp');
    end;
  end;

  inherited;
end;

procedure TQUB_Task_Stat.prDoExecuteStart;
begin
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  inherited;

end;

procedure TQUB_Task_Stat.prDoExecuteEnd;
begin

  inherited;

  if prSelectionList <> nil then begin
    prSelectionList.UnLock;
  end;

end;


function TQUB_Task_Stat.prDoExecuteInitialize :Boolean;
begin
  Result        := False;
end;

function TQUB_Task_IVPlot.prDoExecuteInitialize :Boolean;
begin
  Result        := inherited prDoExecuteInitialize;

  prConfig      := TNode.Create('Properties');
  prConfig['StopFlag'].SetData(0);
  prConfig['ChannelIndex'].SetData( TQUB_QVO_QVD(Data).ActiveChannel );
  prConfig['ProcessData'].SetData('AsDisplayed');
  prConfig['Histograms']['BinCount'].SetData( 2 );
end;

//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Stat.prChunkData :Boolean;
var
  Data             :TQUB_QVO_QVD;
  DataSets         :INode;
  DS, Seg          :INode;
  Process          :INode;
  i                :Integer;
begin
  Result           := False;

  DataSets         := DataNoder.GetChosenIdealization(TNode.CreateNull);
  prDataDescription := DataSets.Find('DataDescription');

  // nasty workaround to get fragmentary dwells:
  DataSets         := TNode.Create('DataSets');
  DataSets.AppendChild(prDataDescription);
  Process          := prConfig.Find('ProcessData');

  if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
    for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
      Data := TQUB_QVO_QVD(ViewportData.ListObjects[i]);
      if Data.Listed then begin
        DataSets.AppendChild(Data.GetChosenIdl(True{Fragments}));
      end;
    end;
  end else begin
    DataSets.AppendChild(ViewportData.CurrData.GetChosenIdl(True{Fragments}));
  end;
  
  DS               := DataSets.Find('DataSet');
  while not DS.IsNull do begin
    try
      if DS.Find('Segment').IsNull then begin
        Data           := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DS.Find('FileName').DataAsString]);
        if Data <> nil then begin
          DS           := DataSets.InsertChild(DS, Data.GetChosenData);

          Seg          := DS.Find('Segment');
          while not Seg.IsNull do begin
            Seg['amp'].SetData( 0.0 / 1.0 );
            Seg['sd'].SetData( 0.1 / 1.0 );
            Seg['DwellCount'].SetData( 1 );
            Seg['Classes'].SetData( 0 );
            Seg['Firsts'].SetData( Seg.Data[0] );
            Seg['Lasts'].SetData( Seg.Data[1] );
            Seg['Durations'].SetupNumData( NODE_DATATYPE_SINGLE, 1, 1 );
            Seg['Durations'].Data[0] := (Seg.Data[1] - Seg.Data[0] + 1) * DS['sampling'].Data[0];
            Seg        := Seg.SiblingSameName;
          end;
        end else begin
          raise Exception.Create('can''t find data file by name');
        end;
      end;

      if not Process.IsNull then begin
        DS.AppendChild( Process.Clone(True) );
      end;

      if DS['FileName'].DataAsString = g_CurrData().FileName then begin
        DS['ActiveChannel'].SetData( prConfig['ChannelIndex'].Data[0] );
      end;

      SetLength( prDataSets, Length(prDataSets) + 1 );
      prDataSets[ Length(prDataSets) - 1 ] := DS;
    except
    end;

    DS             := DS.SiblingSameName;
  end;
end;

procedure TQUB_Task_Stat.prFreeChunk(Chunk :TQUB_Thread_PChunk);
begin

  FreeMem(Chunk);

end;

function TQUB_Task_Stat.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := AllocMem(SizeOf(TQUB_Thread_Chunk));

  Result.Sel := Sel;

end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Stat.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_Stat_PropertiesForm do begin

    prConfig                    := Properties.Clone(True);
    prConfig['StopFlag'].SetData(0);

  end;//with
end;

procedure TQUB_Task_Stat.OnPutProperties(Sender :TObject);
begin
end;


//****************************************************************************//
//                                                                            //
//                            C A L L B A C K S                               //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Stat.prReportCallback( arg :INode ): INode;
begin
  prReportNode := arg;
  Synchronize( prDoReport );
  Result := arg;
end;

procedure TQUB_Task_Stat.prDoReport;
begin
  // just print the node data's text rep'n for now
  prConsoleMsg := prReportNode.GetDataAsString;
  prConsoleCode := MESSAGE_ERROR;
  prWriteNewConsole;
end;


//******************************************************************************
end.

