unit QUB_Task_StimSim_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  QUB_Task_Properties, StdCtrls, ComCtrls, ExtCtrls,

  QUB_Properties,
  QUB_Profiles, QUB_Node,
  QUB_Global,
  QUB_ExpCond_Parameters, QUB_ExpCond_Editor,
  QUB_QVO_QVD,
  QUB_QVO_QVM,
  QUB_QVO_QVP;

type
TQUB_Task_StimSim_PropertiesForm = class(TQUB_Task_PropertiesForm)
  Label1: TLabel;
  lstStimulus: TComboBox;
  Label2: TLabel;
  panExpCond: TPanel;
  chkNewData: TCheckBox;
  Label3: TLabel;
  txtSampling: TEdit;
  Label4: TLabel;
  panStatus: TPanel;
    btnOpen: TButton;
  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure lstStimulusClick(Sender: TObject);
  procedure chkNewDataClick(Sender: TObject);
    procedure btnOpenClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
protected
  grdExpCond      :TQUB_ExpCond_Editor;

  prModel         :TQUB_QVO_QVM;
  prModelVars     :TStringList;
  prParams        :TQUB_ExpCond_Parameters;
  prProtoCount    :Integer;

  procedure prSetupVars; // list model vars and add to defexpcond on show
  procedure prSetupStim; // re-list stim files on show / open
  procedure prAutoChan; // try setup obvious var channels on choose stim
  procedure prOnStimChange(Sender :TObject); // lst and grid: complain/ready
  procedure prSetStatus(Status :String);
  procedure prUpdateComponents; override;

public
  LastSimData     :TQUB_QVO_QVD;
  StimData        :TQUB_QVO_QVD;
  StimProto       :TQUB_QVO_QVP;
  StimCount       :Integer; // number of stimulus channels
  NewData         :Boolean;
  Sampling        :Double;
  Ready           :Boolean;
end;


implementation

{$R *.DFM}

uses

QUB_GUIInterface,
QUB_Viewport_Protocol, QUB_QFS_QPF,
QUB_Viewport_Data, QUB_QFS_QDF,
QUB_Viewport_Model, QUB_QFS_QMF, QUB_QFS_Channels;


const
  DEF_SAMPLING :Double = 10.0;

procedure TQUB_Task_StimSim_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  grdExpCond            := TQUB_ExpCond_Editor.Create(Self);
  grdExpCond.Align      := alClient;
  grdExpCond.Parent     := panExpCond;
  grdExpCond.ShowCond   := False;

  AddEditProperty(TQUB_Property_Edit.Create('["Sampling"]',
                    'Sampling', REGTYPE_FLOAT, DEF_SAMPLING,
                    txtSampling,
                    AcceptFloatGreaterThan(0.0),
                    '') );

  ProfileCat             := 'StimSim';
  WikiTopic              := 'Simulation:StimSim#Properties';

  prModelVars            := TStringList.Create;
  prModelVars.Sorted     := True;
  prModelVars.Duplicates := dupIgnore;
end;

procedure TQUB_Task_StimSim_PropertiesForm.FormDestroy(Sender: TObject);
begin
  try
    prModelVars.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Task_StimSim_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  chkNewData.Enabled     := LastSimData <> nil;
  chkNewData.Checked     := NewData;
  prProperties['Sampling'].Data[0] := Sampling;
  txtSampling.Text       := FloatToStr(Sampling);

  prParams               := ViewportModel.DefExpCond;
  prModel                := ViewportModel.CurrModel;

  prSetupVars;
  prSetupStim;
  prParams.Events.AddEvent('OnChange', prOnStimChange);
end;

procedure TQUB_Task_StimSim_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  prParams.Events.DelEvent('OnChange', prOnStimChange);
  Sampling               := prProperties['Sampling'].Data[0];
end;

procedure TQUB_Task_StimSim_PropertiesForm.prSetupVars;
// list model vars and add to expcond on show
var
  i, j, k       :Integer;
begin
  prModelVars.Clear;

  // list active stimuli
  for j := 0 to prModel.ModelObject.Rates.RateCount - 1 do begin
    for k := 1 to 2 do begin
      if prModel.ModelObject.Rates[j].P[k] then prModelVars.Add(prModel.ModelObject.Rates[j].PName[k]);
      if prModel.ModelObject.Rates[j].Q[k] then prModelVars.Add(prModel.ModelObject.Rates[j].QName[k]);
    end;
  end;

  // add if missing
  for i := 0 to prModelVars.Count - 1 do begin
    if prParams.ParameterByName[prModelVars[i]] = nil then begin
      prParams.AddPar(prModelVars[i], 1.0, -1, False, -1);
    end;
  end;

  // knock out inactive stimuli
  for i := 0 to prParams.Count - 1 do begin
    if prModelVars.IndexOf(prParams[i].Name) < 0 then begin
      prParams.ChangeParChannel(i, -1);
    end;
  end;

  grdExpCond.ExpCond := prParams; // update
end;

procedure TQUB_Task_StimSim_PropertiesForm.prSetupStim;
// re-list stim files on show / open
var
  i :Integer;
begin
  lstStimulus.Items.Assign(ViewportProtocol.ListObjects.Items);
  prProtoCount := ViewportProtocol.ListObjects.Items.Count;
  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    lstStimulus.Items.AddObject(ViewportData.ListObjects.Items[i], ViewportData.ListObjects.Items.Objects[i]);
  end;

  if StimData <> nil then begin
    i := lstStimulus.Items.IndexOfObject(StimData);
  end else if StimProto <> nil then begin
    i := lstStimulus.Items.IndexOfObject(StimProto);
  end else begin
    i := -1;
  end;
  if i = -1 then begin
    StimData  := nil;
    StimProto := nil;
  end;
  lstStimulus.ItemIndex  := i;

  if StimData <> nil then begin
    prProperties['Stimulus'].SetData(StimData.DataObject.FileName);
  end else if StimProto <> nil then begin
    prProperties['Stimulus'].SetData(StimProto.Protocol.FileName);
  end;

  prAutoChan;
  prOnStimChange(Self);
end;

procedure TQUB_Task_StimSim_PropertiesForm.prAutoChan;
// try setup obvious var channels on choose stim
var
  Param         :TQUB_ExpCond_PParameter;
  i             :Integer;
begin
  if StimData <> nil then begin
    StimCount := StimData.DataObject.ADChannelCount;
    if StimCount > 1 then begin
      // if channel(s) in stimdata have same name as stimulus, match 'em
      for i := 0 to prModelVars.Count - 1 do begin
        Param := StimData.ExpCond.ParameterByName[prModelVars[i]];
        if (Param <> nil) and (Param.Channel >= 0) and (Param.Channel < StimData.DataObject.ADChannelCount) then begin
          prParams.ChangeParChannel(prParams.IndexOf(prModelVars[i]), Param.Channel+1);
        end;
      end;
    end;
  end;

  if StimProto <> nil then begin
    StimCount := StimProto.Protocol.ChannelCount;
  end;

  if (StimCount = 1) and (prModelVars.Count = 1) then begin
    prParams.ChangeParChannel(prParams.IndexOf(prModelVars[0]), 1);
  end;

  if (LastSimData <> nil) then begin
    if LastSimData.DataObject.ADChannelCount = (StimCount + 1) then begin
      chkNewData.Enabled := True;
    end else begin
      chkNewData.Enabled := False;
      chkNewData.Checked := True;
    end;
  end;

  grdExpCond.ExpCond := prParams; // update
end;

procedure TQUB_Task_StimSim_PropertiesForm.prOnStimChange(Sender :TObject);
// lst and grid: complain/ready
var
  i, j       :Integer;
  Occ, Assg  :Array of Integer;
  StimCount  :Integer;
begin
  Ready := False;
  btnOK.Enabled := False;
  if (StimData = nil) and (StimProto = nil) then begin
    prSetStatus('Choose a stimulus file, or make a new one in the Acquisition Protocol window');
    Exit;
  end;

  if StimData <> nil then begin
    StimCount := StimData.DataObject.ADChannelCount;
  end else begin
    StimCount := StimProto.Protocol.ChannelCount;
  end;
  SetLength(Occ, StimCount);
  SetLength(Assg, prModelVars.Count);
  for i := 0 to StimCount - 1 do Occ[i] := -1;

  for i := 0 to prModelVars.Count - 1 do begin
    j := prParams.ParameterByName[prModelVars[i]].Channel;
    if j > StimCount then begin
      prSetStatus('The channel index for '+prModelVars[i]+' is too big');
      Exit;
    end else if j < 0 then begin
      Assg[i] := -1;
    end else if j = 0 then begin
      prSetStatus('You can''t use channel index 0.');
      Exit;
    end else begin
      Assg[i] := j-1; // source-index
      if Occ[j-1] >= 0 then begin
        prSetStatus(prModelVars[i]+' and '+prModelVars[Occ[j-1]]+' are set to the same channel');
        Exit;
      end else begin
        Occ[j-1] := i;
      end;
    end;
  end;

  Ready := True; // these guys aren't a big problem:
  btnOK.Enabled := True;

  for j := 0 to StimCount - 1 do begin
    if Occ[j] < 0 then begin
      prSetStatus(Format('Channel %d is not assigned to any variable in the model', [j+1]));
      Exit;
    end;
  end;
  for i := 0 to prModelVars.Count - 1 do begin
    if Assg[i] < 0 then begin
      prSetStatus(prModelVars[i]+' is not assigned to any channel (will be held constant)');
      Exit;
    end;
  end;

  prSetStatus(''); // quiet
end;

procedure TQUB_Task_StimSim_PropertiesForm.prSetStatus(Status :String);
begin
  if Status = ''
  then panStatus.Color := clBtnFace
  else panStatus.Color := TColor($ccaaff);

  panStatus.Caption := Status;
end;

procedure TQUB_Task_StimSim_PropertiesForm.prUpdateComponents;
var
  i        :Integer;
  FileName :String;
begin
  inherited;

  if prProperties.Find('Stimulus').DataCount > 0 then begin
    FileName := prProperties['Stimulus'].DataAsString;
    // first look in the list
    i        := lstStimulus.Items.IndexOf(ExtractFileName(FileName));
    if i >= 0 then begin
      if i <> lstStimulus.ItemIndex then begin
        lstStimulus.ItemIndex := i;
        lstStimulusClick(lstStimulus);
      end;
    end else begin
      // then try to open it
      if LowerCase(ExtractFileExt(FileName)) = '.qpf' then begin
        if not ViewportProtocol.Open(FileName, True{Silent}) then begin
          StimData  := nil;
          StimProto := ViewportProtocol.CurrProto;
          prSetupStim;
        end;
      end else begin
        if not ViewportData.Open(FileName, True{Silent}) then begin
          StimData  := ViewportData.CurrData;
          StimProto := nil;
          prSetupStim;
        end;
      end;
    end;    // or just leave it as it was
  end;
end;

procedure TQUB_Task_StimSim_PropertiesForm.btnOpenClick(Sender: TObject);
begin
  inherited;
  if not ViewportProtocol.Open('', False{Silent}) then begin
    StimData     := nil;
    StimProto    := ViewportProtocol.CurrProto;
    prSetupStim;
  end;
end;

procedure TQUB_Task_StimSim_PropertiesForm.lstStimulusClick(Sender: TObject);
begin
  inherited;
  if lstStimulus.ItemIndex >= 0 then begin
    if lstStimulus.ItemIndex < prProtoCount then begin
      StimData  := nil;
      StimProto := TQUB_QVO_QVP(lstStimulus.Items.Objects[lstStimulus.ItemIndex]);
    end else begin
      StimData  := TQUB_QVO_QVD(lstStimulus.Items.Objects[lstStimulus.ItemIndex]);
      StimProto := nil;
    end;
    prSetupStim;
  end;
end;

procedure TQUB_Task_StimSim_PropertiesForm.chkNewDataClick(Sender: TObject);
begin
  inherited;
  NewData      := chkNewData.Checked;
end;


// *****************************************************************************
end.
