{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// idealizes a channel of experimental conditions using qubopt.dll

unit QUB_Task_Stimulus;


interface


uses


Windows, SysUtils, Classes, Clipbrd, Forms, Math,

AllocTypes, Alloc,

QUB_Global,

QUB_QVO_QVD, QUB_QVO_QVD_Global,

QUB_Viewport_Data,

QUB_Thread,
QUB_Task,
QUB_Task_Data,
QUB_Task_Stimulus_Properties,

QUB_Idealization,

QUB_Node, QUB_Node_Python,
QUB_Python_Thread, QUB_Viewport_Python,
QUB_GUIInterface,

QUB_ExpCond_Parameters;


type


TQUB_Task_Stimulus  = class(TQUB_Task_Data)
protected

  prDataSet        :INode;

  prProperties     :INode;
  prReportNode     :INode;
  prResults        :INode;

  //Execute
  procedure prDoTask; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;
  function prDoExecuteInitialize :Boolean; override;

  //Chunks    -- we don't use 'em
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetProperties(x :INode);

  procedure prDoReport;
  function prReportCallback(arg :INode): INode;
  function prPercentCallback(arg :INode): INode;

  procedure prSetupData;
  procedure prPutIdl;
  procedure prPublishResults;

public

  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Stop; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  property Properties :INode    read prProperties      write prSetProperties;

end;


function StimulusFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; fillDataCB, reportCB, percentCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_IdlCond_iface@20';
function MplexIdlFromDll(idls :TPAPointer; Nseg :Integer; Firsts, Lasts: TPAInteger): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_MplexIdl_iface@16';
function MplexIdl(idls :TList{of TQUB_IdlSegments}; NSeg :Integer; Segments :TQUB_PAFirstLast): INode;
function MplexIdlWithCondFromDll(idl :TQUB_IdlSegments_Impl; condIdl :TQUB_QFS_PNodeMem; MaxPower :Integer): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_MplexIdlWithCond_iface@12';
function MplexIdlWithCond(idl :TQUB_IdlSegments; condIdl :INode; MaxPower :Integer): INode;
//function TestMplexIdl(dummy :INode): INode;

procedure RenumberIdl(Idl :TQUB_IdlSegments; DataSet :INode; DataLast :Integer);

procedure MplexAmpsToParamClasses(MplexAmps :INode; ParamIndexOfChannel :TPAInteger; ParamTemplate :TQUB_ExpCond_Parameters; ParamClasses :TQUB_ExpCond_ParametersList);

{  suggested use:
MplexIdl with your stimulus channels,
then for mip, MplexIdlWithCond(datachannel, mplexidl_result, maxpower)
}

implementation


uses


QUB_Main,
QUB_QFS_QDF,
QUB_Python_Data;


//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _I D L C O N D                 //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_Stimulus.Create(Suspended :Boolean);
begin
  inherited;

  prName := 'Stimulus';

end;

destructor TQUB_Task_Stimulus.Destroy;
begin
  inherited;
end;

procedure TQUB_Task_Stimulus.prSetProperties(x :INode);
begin
  prProperties := x;
  prProperties['StopFlag'].SetData(0);
end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Stimulus.prDoTask;
var
  DataSetPtrs         :TPAPointer;
  fillDataCB          :INodeCFunc;
  rptCB               :INodeCFunc;
  pctCB               :INodeCFunc;
  ResultNode          :INode;
  ResultNodePtr       :TQUB_QFS_PNodeMem;
begin
  Synchronize(prSetupData);

  DataSetPtrs := nil;

  try
    GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * ({Length(prDatasets)}1 + 1) );
    DataSetPtrs[0] := Pointer(prDataSet.GetNodeMem);
    DataSetPtrs[ 1 ] := nil;

    fillDataCB  := TNodeCFunc.Wrap( TNodeMeth.Create( DataNoder.FillData ) );
    rptCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prReportCallback ) );
    pctCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prPercentCallback ) );

    try

      prConsoleMsg  := 'Stimulus running ... ';
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);

      ResultNodePtr  := StimulusFromDll( Properties.GetNodeMem, DataSetPtrs, fillDataCB, rptCB, pctCB );
      ResultNode     := TNode.CreateUsing( ResultNodePtr );
      if ResultNodePtr <> nil then Node_DECREF( ResultNodePtr );
      prDataSet      := ResultNode['DataSets']['DataSet'];

      Synchronize(prPutIdl);
      Synchronize(prPublishResults);

    except


    end;


  finally
    FreeMem(DataSetPtrs);
  end;
end;

procedure TQUB_Task_Stimulus.prDoExecuteStart;
begin

  inherited;

end;

procedure TQUB_Task_Stimulus.prDoExecuteEnd;
begin

  inherited;

end;


function TQUB_Task_Stimulus.prDoExecuteInitialize :Boolean;
begin
  Result := False;
end;

procedure TQUB_Task_Stimulus.Stop;
begin

  Properties['StopFlag'].Data[0] := 1;

  inherited;

end;

procedure TQUB_Task_Stimulus.prSetupData;
begin
  ChannelIndex := Min(Properties['Stimulus Channel'].Data[0],
                      TQUB_QVO_QVD(Data).DataObject.ADChannelCount - 1);
  prDataSet    := TQUB_QVO_QVD(Data).GetChosenData;
  prDataSet['ActiveChannel'].Data[0] := ChannelIndex;

  //TO DO: We should not clear all the idealization, just what's being idealized now. LM
  //Testing ...

  if (0<>Properties['ClearIdealization'].Data[0])
  or (QUB_MainForm.DataSource = DATA_SOURCE_WHOLEFILE) then begin
    TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].Clear;
  end;

  //I found that having two separate idealizations will crash it - don't know exactly what
  //I dont know what to do. Idealization is fine, but I guess the wrong values are passed to Mac
end;

procedure RenumberIdl(Idl :TQUB_IdlSegments; DataSet :INode; DataLast :Integer);
// changes events in DataSet from class-0-based to class-N-based,
// where N = Idl.ClsCount
var
  N             :Integer;
  i, f          :Integer;
  Seg           :INode;
  Amp           :Array of Double;
  SegAmp, SegSd :INode;
  SegN          :Integer;
  Classes       :INode;

  procedure Scan(First, Last :Integer);
  var
    II          :TQUB_IdlSegmentsIter;
    ff, ll, cc  :Integer;
    aa          :Double;
  begin
    II          := Idl.FindIter(First);
    while 0<>II.Valid do begin
      II.Read(@ff, @ll, @cc, @aa);
      if ff > Last then Break;
      if cc >= 0 then Amp[cc] := aa;
      II.Next;
    end;
    II.Free;
  end;
begin
  N             := Idl.ClsCount;
  if N = 0 then Exit;

  SetLength(Amp, N);

  // read class amps still in use
  f             := 0;
  Seg           := DataSet.Find('Segment');
  while not Seg.IsNull do begin
    if f < Seg.Data[0] then Scan(f, Seg.Data[0]-1);
    f           := Seg.Data[1] + 1;
    Seg         := Seg.SiblingSameName;
  end;
  if f <= DataLast then Scan(f, DataLast);

  // renumber DataSet
  Seg           := DataSet.Find('Segment');
  while not Seg.IsNull do begin
    SegAmp      := Seg['amp'];
    SegSd       := Seg['sd'];
    SegN        := Min(SegAmp.DataCount, SegSd.DataCount);
    Classes     := Seg['Classes'];
    // expand amp and sd
    if SegN = 0 then begin
      SegAmp.SetupNumData(NODE_DATATYPE_DOUBLE, 1, 1);
      SegSd.SetupNumData (NODE_DATATYPE_DOUBLE, 1, 1);
    end;
    SegAmp.ResizeData(N+SegN);
    SegSd.ResizeData(N+SegN);
    // move new amps to end
    for i := 0 to SegN - 1 do begin
      SegAmp.Data[N+i] := SegAmp.Data[i];
      SegSd.Data[N+i]  := SegSd.Data[i];
    end;
    // fill in old amps
    for i := 0 to N - 1 do begin
      SegAmp.Data[i]   := Amp[i];
      SegSd.Data[i]    := 0.01;
    end;
    // renumber Classes
    for i := 0 to Classes.DataCount - 1 do begin
      Classes.Data [i] := Classes.Data[i] + N;
    end;

    Seg := Seg.SiblingSameName;
  end;
end;

//New from Chris 07-03-01, at my request
procedure TQUB_Task_Stimulus.prPutIdl;
var
  Idl :TQUB_IdlSegments;
begin
  Idl := TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex];
  RenumberIdl(Idl, prDataSet, TQUB_QVO_QVD(Data).DataPoint.Last);
  Idl.SetDwells(prDataSet);
  TQUB_QVO_QVD(Data).DrawIdealized := True
end;

procedure TQUB_Task_Stimulus.prPublishResults;
var
  Results :INode;
  Node, P :INode;
begin
  Results := TNode.Create('');
  Results['Held'].SetData(1);
  Results['ResultType'].SetData('Stim');
  Results['DataSet']['FileName'].SetData(Data.FileName);

  Node    := Properties.Child;
  P       := Results['DataSet'];
  while not Node.IsNull do begin
    P     := Results.InsertClone(P, Node, True);
    Node  := Node.Sibling;
  end;

  ViewportResults.AddResults(Results, False);
end;

procedure TQUB_Task_Stimulus.prDoReport;
begin
  // just print the node data's text rep'n for now
  prConsoleMsg := prReportNode.GetDataAsString;
  prConsoleCode := MESSAGE_ERROR;
  prWriteNewConsole;
end;

function TQUB_Task_Stimulus.prReportCallback( arg :INode ): INode;
begin
  Result          := arg;
  prReportNode    := arg;
  Synchronize(prDoReport);
end;

function TQUB_Task_Stimulus.prPercentCallback(arg :INode): INode;
begin
  prUpdateStatus( Integer( arg.Data[0] ), prStTime );
  Result := arg;
end;

//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Stimulus.prChunkData :Boolean;
begin
  Result := False;
end;

procedure TQUB_Task_Stimulus.prFreeChunk(Chunk :TQUB_Thread_PChunk);
begin
end;

function TQUB_Task_Stimulus.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result := nil;

end;

//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Stimulus.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_Stimulus_PropertiesForm do begin

    Self.Properties                  := Properties.Clone(True);

  end;//with
end;

procedure TQUB_Task_Stimulus.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_Stimulus_PropertiesForm do begin
  //  Scale                := Self.Scale;
  //end;//with
end;


//******************************************************************************
function MplexIdl(idls :TList{of TQUB_IdlSegments}; Nseg :Integer; Segments :TQUB_PAFirstLast): INode;
var
  idlPtrs     :TPAPointer;
  Firsts      :TPAInteger;
  Lasts       :TPAInteger;
  i           :Integer;
  Res         :TQUB_QFS_PNodeMem;
begin
  idlPtrs     := nil;
  Firsts      := nil;
  Lasts       := nil;
  Result      := QUB_Node_NULL;

  try
    GetMem( idlPtrs, SizeOf(Pointer) * (idls.Count + 1) );
    for i := 0 to idls.Count - 1 do begin
      idlPtrs[i] := Pointer(TQUB_IdlSegments(idls[i]).Impl);
    end;
    idlPtrs[idls.Count] := nil;

    GetMem(Firsts, SizeOf(Integer) * Nseg);
    GetMem(Lasts, SizeOf(Integer) * Nseg);
    for i := 0 to Nseg - 1 do begin
      Firsts[i] := Segments[i].First;
      Lasts[i]  := Segments[i].Last;
    end;

    Res       := MplexIdlFromDll(idlPtrs, Nseg, Firsts, Lasts);
    Result    := TNode.CreateUsing(Res);
    Node_DECREF(Res);
  finally
    FreeMem(idlPtrs);
    FreeMem(Firsts);
    FreeMem(Lasts)
  end;
end;

function MplexIdlWithCond(idl :TQUB_IdlSegments; condIdl :INode; MaxPower :Integer): INode;
var
  Res         :TQUB_QFS_PNodeMem;
begin
  Res := MplexIdlWithCondFromDll(idl.Impl, condIdl.GetNodeMem, MaxPower);
  Result := TNode.CreateUsing(Res);
  Node_DECREF(Res);
end;

(*
function TestMplexIdl(dummy :INode): INode;
var
  Data        :TQUB_QVO_QVD;
  n           :Integer;
  idls        :TList;
  Segments    :TQUB_PAFirstLast;
  DS          :INode;
  Seg         :INode;
  NSeg        :Integer;
  i           :Integer;
begin
  Data        := TQUB_QVO_QVD(ViewportData.CurrData);
  n           := Data.DataObject.ADChannelCount;
  idls        := TList.Create;
  for i := 0 to n - 1 do begin
    idls.Add(Data.DataObject.Idealization[i]);
  end;

  DS          := Data.GetChosenData;
  NSeg        := 0;
  Seg         := DS.Find('Segment');
  while not Seg.IsNull do begin
    Inc(NSeg);
    Seg       := Seg.SiblingSameName;
  end;
  GetMem(Segments, NSeg * SizeOf(TQUB_FirstLast));
  Seg         := DS.Find('Segment');
  for i := 0 to NSeg - 1 do begin
    Segments[i].First := Seg.Data[0];
    Segments[i].Last  := Seg.Data[1];
    Seg       := Seg.SiblingSameName;
  end;

  Result := MplexIdl(idls, NSeg, Segments, 1);

  FreeMem(Segments);
  idls.Free;
end;
*)

procedure MplexAmpsToParamClasses(MplexAmps :INode; ParamIndexOfChannel :TPAInteger; ParamTemplate :TQUB_ExpCond_Parameters; ParamClasses :TQUB_ExpCond_ParametersList);
var
  i, p           :Integer;
  Params         :TQUB_ExpCond_Parameters;
begin
  if MplexAmps.Rows > 0 then begin
    for i := 0 to MplexAmps.Rows - 1 do begin
      Params         := TQUB_ExpCond_Parameters.CreateCopy(ParamTemplate);
      for p := 0 to MplexAmps.Cols - 1 do begin
        Params[ParamIndexOfChannel[p]].Value := MplexAmps.MatrixData[i, p];
      end;
      ParamClasses.Add(Params);
    end;
  end else begin
    // one expcond class of default
    ParamClasses.Add( TQUB_ExpCond_Parameters.CreateCopy(ParamTemplate) );
  end;
end;

//******************************************************************************
end.

