unit QUB_Task_Stimulus_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry, Math,

RegistryInterface,

Alloc, AllocTypes,
QUB_Global,
QUB_QVO_QVD,
QUB_Task,
QUB_Properties,
QUB_Task_Properties,
QUB_Node;

//{$include 'QUB_Task_Stimulus_Properties.hct'}

const


DEF_IDLCOND_RESTINGTENSION    :Double  = 0.01;
DEF_IDLCOND_COLLAPSEDTENSION  :Double  = 0.0;
DEF_IDLCOND_TENSIONPRESSURE   :Double  = 1.0;
DEF_IDLCOND_TAURISING         :Double  = 10.0;
DEF_IDLCOND_TAUCOLLAPSING     :Double  = 1.0;
DEF_IDLCOND_TAURECOVERING     :Double  = 10.0;
DEF_IDLCOND_CLEARIDEALIZATION :Integer = 1;
DEF_IDLCOND_ADDDELTAS         :Integer = 1;
DEF_IDLCOND_DELTA             :Double  = 0.5;
MIN_IDLCOND_DELTA             :Double  = 1.0e-8;
MAX_IDLCOND_DELTA             :Double  = 1.0e9;
DEF_LATENCY                   :Double  = 0.0;
DEF_MINDUR                    :Double  = 0.5;
DEF_IDLCOND_SPRINGDASHPOT     :Integer = 0;
DEF_IDLCOND_SDP_DAMPING       :Double = 0.589;
DEF_IDLCOND_SDP_FREQ          :Double = 1075;
DEF_ONTHRESHOLD               :Double = 0.5;

DEF_PREVIEW_SIZE               = 32000;
MAX_PREVIEW_SIZE               = 1280000;

WM_SHOWCONTINUED               = 1234;
WM_NEEDIDL                     = 1235;


type


TQUB_Task_Stimulus_PropertiesForm = class(TQUB_Task_PropertiesForm)
  Label3                          :TLabel;
  panPreview: TPaintBox;
  panDataSource                   :TPanel;
  panDataSourceFile               :TPanel;
  chkDataSourceWholeFile          :TRadioButton;
  chkDataSourceSelectionList      :TRadioButton;
  chkDataSourceSelection          :TRadioButton;
  panOptions                      :TPanel;
  Label1                          :TLabel;
  txtKnown                        :TEdit;
  chkAddDeltas                    :TCheckBox;
  lblDelta                        :TLabel;
  txtDelta                        :TEdit;
    Label4: TLabel;
    txtMinDur: TEdit;
    panStimChannel: TPanel;
    Label7: TLabel;
    txtStimChannel: TEdit;
    udStimChannel: TUpDown;
    chkLiveMac: TCheckBox;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Label2: TLabel;
    txtLatency: TEdit;
    chkSDP: TCheckBox;
    Label5: TLabel;
    txtSDPDamping: TEdit;
    txtSDPFreq: TEdit;
    Label6: TLabel;
    chkClearIdealization: TCheckBox;
    Label8: TLabel;
    chkTension: TCheckBox;
    Label9: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    Label12: TLabel;
    Label13: TLabel;
    txtRestingTension: TEdit;
    txtCollapsedTension: TEdit;
    txtTensionPressure: TEdit;
    txtTauRising: TEdit;
    txtTauCollapsing: TEdit;
    Label14: TLabel;
    txtOnThreshold: TEdit;
    lblStatus: TLabel;
    Label15: TLabel;
    txtTauRecovering: TEdit;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure chkDataSourceClick(Sender: TObject);
    procedure panPreviewPaint(Sender: TObject);
    procedure FormHide(Sender: TObject);


protected

  prData                          :TQUB_QVO_QVD;
  prPreview                       :TQUB_QVO_QVD;
  prSel                           :TQUB_FirstLast;
  prN                             :Integer;
  prBuffer                        :TPAInteger;
  prBufferMPX                     :Pointer;

  prNeedIdl                       :Boolean;
  prNeedMac                       :Boolean;
  prLevelsMsg                     :String;
  prTaskIdl                       :TQUB_Task;
  prTaskMac                       :TQUB_Task;

  //Data Source
  prDataSourceSaved               :TQUB_Task_Data_Source;
  prDataSource                    :TQUB_Task_Data_Source;
  prDataSourceInFile              :TQUB_Task_Data_Source;

  //Data Source
  procedure prSetDataSource(s :TQUB_Task_Data_Source);

  procedure prOnShowContinued(var Message: TMessage); message WM_SHOWCONTINUED;
  procedure prOnNeedIdl(var Message: TMessage); message WM_NEEDIDL;

  procedure prOnExitWrap(Sender: TObject);
  procedure prOnCheckClickWrap(Sender: TObject);
  procedure prOnExitUpDownTextWrap(Sender :TObject);
  procedure prOnUpDownClickWrap(Sender :TObject; Button: TUDBtnType);
  procedure prUpdateComponents; override;

  procedure prSetupPreview;
  procedure prCleanupPreview;

  procedure prInvalidateData;
  procedure prInvalidateStim;
  procedure prInvalidateIdl;
  procedure prInvalidateLiveMac;
  procedure prOnIdlDone(Sender :TObject);
  procedure prOnMacDone(Sender :TObject);

public

  //Data Source
  property DataSource                     :TQUB_Task_Data_Source     read prDataSource                write prSetDataSource;

end;


implementation


uses

QUB_ActionSelector_Modeling,
QUB_FitCurves,
QUB_Idealization,
QUB_QFS_QDF_Global,
QUB_QFS_QDF_MPX,
QUB_Main,
QUB_GUIInterface,
QUB_Task_Stimulus,
QUB_Mac_Global,
QUB_Task_Mac,
QUB_Task_Mac_Properties,
QUB_Python_Thread,
QUB_Node_Python,
QUB_Viewport_Python;


{$R *.DFM}


//******************************************************************************
//Data Source
procedure TQUB_Task_Stimulus_PropertiesForm.prSetDataSource(s :TQUB_Task_Data_Source);
begin
  prDataSource                       := s;
  chkDataSourceWholeFile.Checked     := DataSource =  DATA_SOURCE_WHOLEFILE;
  chkDataSourceSelectionList.Checked := DataSource =  DATA_SOURCE_SELECTIONLIST;
  chkDataSourceSelection.Checked     := DataSource =  DATA_SOURCE_SELECTION;
end;


//******************************************************************************
procedure TQUB_Task_Stimulus_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddUpDownProperty(TQUB_Property_UpDown.Create('["Stimulus Channel"]',
                      '', REGTYPE_NONE, 1,
                      udStimChannel,
                      txtStimChannel,
                      AcceptIntGreaterThanOrEqualTo(0)));
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["LiveMac"]',
                        '', REGTYPE_NONE, 0,
                        chkLiveMac)
                     );

  AddEditProperty(TQUB_Property_Edit.Create('["Latency_MS"]',
                    '', REGTYPE_NONE, DEF_LATENCY,
                    txtLatency,
                    AcceptFloatGreaterThanOrEqualTo(0.0),
                    '') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Tension"]',
                        '', REGTYPE_NONE, 0,
                        chkTension)
                       .Enables(txtOnThreshold) 
                       .Enables(txtRestingTension)
                       .Enables(txtCollapsedTension)
                       .Enables(txtTensionPressure)
                       .Enables(txtTauRising)
                       .Enables(txtTauCollapsing)
                       .Enables(txtTauRecovering)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["OnThreshold"]',
                    '', REGTYPE_NONE, DEF_ONTHRESHOLD,
                    txtOnThreshold,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["RestingTension"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_RESTINGTENSION,
                    txtRestingTension,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["CollapsedTension"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_COLLAPSEDTENSION,
                    txtCollapsedTension,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["TensionPressure"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_TENSIONPRESSURE,
                    txtTensionPressure,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["TauRising"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_TAURISING,
                    txtTauRising,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["TauCollapsing"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_TAUCOLLAPSING,
                    txtTauCollapsing,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["TauRecovering"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_TAURECOVERING,
                    txtTauRecovering,
                    AcceptFloat,
                    '') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["SpringDashPot"]',
                        '', REGTYPE_NONE, DEF_IDLCOND_SPRINGDASHPOT,
                        chkSDP)
                       .Enables(txtSDPDamping)
                       .Enables(txtSDPFreq)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["SDP Damping"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_SDP_DAMPING,
                    txtSDPDamping,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["SDP Freq"]',
                    '', REGTYPE_NONE, DEF_IDLCOND_SDP_FREQ,
                    txtSDPFreq,
                    AcceptFloatGreaterThan(0.0),
                    '') );

  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["ClearIdealization"]',
                        '', REGTYPE_NONE, DEF_IDLCOND_CLEARIDEALIZATION,
                        chkClearIdealization)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["Known Amps"]',
                          '', REGTYPE_NONE, 0,
                          txtKnown,
                          AcceptFloatList(AcceptFloat),
                          'lambda node, ix: SpacedNodeData(node)') );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["AddDeltas"]',
                        'AddDeltas', REGTYPE_BOOL, DEF_IDLCOND_ADDDELTAS,
                        chkAddDeltas)
                       .Enables(txtDelta)
                     );
  AddEditProperty(TQUB_Property_Edit.Create('["Delta"]',
                    'Delta', REGTYPE_FLOAT, DEF_IDLCOND_DELTA,
                    txtDelta,
                    AcceptFloatGreaterThan(0.0),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["MinDur_MS"]',
                    '', REGTYPE_NONE, DEF_MINDUR,
                    txtMinDur,
                    AcceptFloatGreaterThanOrEqualTo(0.0),
                    '') );

  // wrap events which impact preview:
  txtStimChannel.OnExit         := prOnExitUpDownTextWrap;
  udStimChannel.OnClick         := prOnUpDownClickWrap;
  txtXChannelIndex.OnExit       := prOnExitUpDownTextWrap;
  udXChannelIndex.OnClick       := prOnUpDownClickWrap;
  chkLiveMac.OnClick            := prOnCheckClickWrap;
  txtLatency.OnExit             := prOnExitWrap;
  chkTension.OnClick            := prOnCheckClickWrap;
  txtOnThreshold.OnExit         := prOnExitWrap;
  txtRestingTension.OnExit      := prOnExitWrap;
  txtCollapsedTension.OnExit    := prOnExitWrap;
  txtTensionPressure.OnExit     := prOnExitWrap;
  txtTauCollapsing.OnExit       := prOnExitWrap;
  txtTauRising.OnExit           := prOnExitWrap;
  txtTauRecovering.OnExit       := prOnExitWrap;
  chkSDP.OnClick                := prOnCheckClickWrap;
  txtSDPFreq.OnExit             := prOnExitWrap;
  txtSDPDamping.OnExit          := prOnExitWrap;
  chkClearIdealization.OnClick  := prOnCheckClickWrap;
  txtKnown.OnExit               := prOnExitWrap;
  chkAddDeltas.OnClick          := prOnCheckClickWrap;
  txtDelta.OnExit               := prOnExitWrap;
  txtMinDur.OnExit              := prOnExitWrap;

  RegKeyName            := 'Software\QUB\Task Stimulus';
  ProfileCat            := 'Stim';

  WikiTopic             := 'Preprocessing:Stimulus';

  prPreview             := nil;
  prBuffer              := nil;
  prBufferMPX           := nil;
  prTaskIdl             := nil;
  prTaskMac             := nil;
  prNeedIdl             := False;
  prNeedMac             := False;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  prData                  := ViewportData.CurrData;
  if prData <> nil then begin
    Properties['Stimulus Channel'].Data[0] := Min(prData.DataObject.ADChannelCount-1, Properties['Stimulus Channel'].Data[0]);
  end;
  prDataSourceSaved       := QUB_MainForm.DataSource;
  QUB_MainForm.DataSource := DATA_SOURCE_WHOLEFILE;

  prSetupPreview;

  udStimChannel.Max      := prData.DataObject.ADChannelCount - 1;
  udXChannelIndex.Max    := prData.DataObject.ADChannelCount - 1;

  // wait for the data to be measured
  PostMessage(Self.Handle, WM_SHOWCONTINUED, 0, 0);
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnShowContinued(var Message: TMessage);
var
  i                      :Integer;
begin
  for i := 0 to prPreview.ExpCond.Count - 1 do begin
    if prPreview.ExpCond[i].Channel = udStimChannel.Position then begin
      prInvalidateStim;
      Exit;
    end;
  end;
  // stimchannel is not assigned in expcond.  Is another?
  for i := 0 to prPreview.ExpCond.Count - 1 do begin
    if prPreview.ExpCond[i].Channel >= 0 then begin
      udStimChannel.Position := prPreview.ExpCond[i].Channel;
      udStimChannel.OnClick(udStimChannel, btNext);
      Exit;
    end;
  end;
  // no?
  prInvalidateStim;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;
  if (prTaskIdl <> nil) or (prTaskMac <> nil) then begin
    if prTaskIdl <> nil then begin
      prTaskIdl.Stop;
    end;
    if prTaskMac <> nil then begin
      prTaskMac.Stop;
    end;
  end else begin
    prCleanupPreview;
  end;
  FreeMem(prBuffer);
  FreeMem(prBufferMPX);
  prBuffer    := nil;
  prBufferMPX := nil;
  QUB_MainForm.DataSource := prDataSourceSaved;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prSetupPreview;
var
  Dir   :String;
  FName :String;
  Index :Integer;
  sr    :TSearchRec;
  c         :Integer;

  function FileName: String;
  begin
    Result := FName;
    if Index > 1 then Result := Result + IntToStr(Index);
    Result := Result + '.qdf';
  end;
begin
  prPreview := nil; // cut loose any holdover

  Dir   := ExtractFilePath(Application.ExeName);
  FName := Dir+'StimulusPreview';
  // delete old StimulusPreview*
  if 0 = FindFirst(FName+'*.q*f', 0, sr) then begin
    repeat
      try
        DeleteFile(Dir+sr.Name);
      except
      end;
    until 0 <> FindNext(sr);
  end;
  Sysutils.FindClose(sr);
  // first available name
  Index := 1;
  while FileExists(FileName) do Inc(Index);

  // new 2ch qdf
  with ViewportData.NewPropertiesForm do begin
    FileFormat     := QFS_QDF_TYPE_QDF;
    ADChannelCount := prData.DataObject.ADChannelCount;
    Sampling       := 1e-3 / prData.DataObject.Sampling;
    ADDataSize     := prData.DataObject.ADDataSize;
    Scaling        := prData.DataObject.Scaling;
  end;
  if ViewportData.New(FName, True{Silent}) then begin
    MessageHelpDlg('Can''t create preview file', mtError, [mbOK], 0, 'Error:Cannot_Create_Preview');
    Exit;
  end;
  prPreview        := ViewportData.CurrData;
  ViewportData.ListObjects.SelQVO(prData);

  // adjust sel within reasonable and cache limits
  prSel            := prData.DataSel;
  if prSel.First >= prSel.Last then begin
    prSel.Last     := Min(prSel.First + DEF_PREVIEW_SIZE - 1, prData.DataObject.Segments.TotalDataCount-1);
  end;
  prSel.Last       := Min(prSel.First + MAX_PREVIEW_SIZE - 1, prSel.Last);
  prSel.Last       := Min(prSel.First + prPreview.CacheSize - 1, prSel.Last);
  prN              := prSel.Last - prSel.First + 1;

  prBuffer         := AllocMem(prN * SizeOf(Integer));
  prBufferMPX      := AllocMem(prPreview.DataObject.ADDataSize * prPreview.DataObject.ADChannelCount * prN);

  for c := 0 to prData.DataObject.ADChannelCount - 1 do begin
    prPreview.DataChannelScaling[c] := prData.DataChannelScaling[c];
    prPreview.DataChannelOffset[c]  := prData.DataChannelOffset[c];
    prData.DataObject.ReadData(prBuffer, False{Raw}, c{Channel}, prSel.First, prSel.Last);
    MPXData(prBuffer, prBufferMPX, prPreview.DataObject.ADChannelCount, c, prPreview.DataObject.ADDataSize, prPreview.DataObject.ADDataType, 0, prN-1, True{MPX});
  end;
  prPreview.ExpCond.ReadTree(prData.ExpCond.ToTree);

  prPreview.AppendData(prBufferMPX, prPreview.DataObject.ADChannelCount * prPreview.DataObject.ADDataSize, prN, True, 0.0, False{CanRefresh - Doit}, False{Acq});
  prPreview.AppendData(nil, 0, 0, False, 0, True, False{Acq});

  prInvalidateData;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prCleanupPreview;
begin
  if prPreview <> nil then begin
    ViewportData.ListObjects.SelQVO(prPreview);
    ViewportData.Close;
    prPreview := nil;
    ViewportData.ListObjects.SelQVO(prData);
  end;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnExitWrap(Sender: TObject);
begin
  prOnExit(Sender);
  prInvalidateIdl;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnCheckClickWrap(Sender: TObject);
begin
  prOnCheckClick(Sender);
  if Sender = chkLiveMac then begin
    prInvalidateLiveMac;
  end else begin
    prInvalidateIdl;
  end;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnExitUpDownTextWrap(Sender :TObject);
begin
  prOnExitUpDownText(Sender);
  if Sender = txtStimChannel then begin
    prInvalidateStim;
  end else begin
    prInvalidateData;
  end;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnUpDownClickWrap(Sender :TObject; Button: TUDBtnType);
begin
  prOnUpDownClick(Sender, Button);
  if Sender = udStimChannel then begin
    prInvalidateStim;
  end else begin
    prInvalidateData;
  end;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prUpdateComponents;
begin
  inherited;
  prInvalidateData;
  prInvalidateStim;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prInvalidateData;
begin
  panPreview.Invalidate;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prInvalidateStim;
begin
  prInvalidateIdl;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prInvalidateIdl;
begin
  if prPreview = nil then Exit;
  prNeedIdl := True;
  PostMessage(Self.Handle, WM_NEEDIDL, 0, 0);
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnNeedIdl(var Message: TMessage);
begin
  if not prNeedIdl then Exit;
  lblStatus.Caption := 'Idealizing...please wait...';

  if prTaskIdl <> nil then begin
    prTaskIdl.Stop;
  end else begin
    prPreview.DataObject.Idealization[1].Clear;
    prPreview.FitCurves.Clear;
    prTaskIdl := TQUB_Task_Stimulus.Create(True{Suspended});
    TQUB_Task_Stimulus(prTaskIdl).OnTerminate  := prOnIdlDone;
    TQUB_Task_Stimulus(prTaskIdl).Properties   := Properties.Clone(True);
    TQUB_Task_Stimulus(prTaskIdl).Properties['Stimulus Channel'].Data[0] := 1;
    TQUB_Task_Stimulus(prTaskIdl).Data         := prPreview;
    TQUB_Task_Stimulus(prTaskIdl).SilentOutput := True;
    TQUB_Task_Stimulus(prTaskIdl).Resume;
    panPreview.Invalidate;
    prNeedIdl := False;
  end;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnIdlDone(Sender :TObject);
begin
  prTaskIdl := nil;
  if (not Visible) and (prTaskMac = nil) then begin
    SyncGUIProc(prCleanupPreview);
  end else if prNeedIdl then begin
    SyncGUIProc(prInvalidateIdl);
  end else begin
    SyncGUIProc(prInvalidateLiveMac);
    prLevelsMsg := Format('Ready: %d levels found in preview',
                          [prPreview.DataObject.Idealization[Properties['Stimulus Channel'].Data[0]].ClsCount]);
    lblStatus.Caption := prLevelsMsg;
  end;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prInvalidateLiveMac;
var
  MacProperties         :TQUB_Task_Mac_PropertiesForm;
begin
  if (prPreview = nil) or (ViewportModel.CurrModel = nil) then Exit;
  prPreview.FitCurves.Clear;

  if 0<>Properties['LiveMac'].Data[0] then begin
    lblStatus.Caption := 'Simulating Mac current...please wait...';
    if prTaskMac <> nil then begin
      prNeedMac := True;
    end else begin
      prNeedMac := False;

      MacProperties := TQUB_ActionSelector_Modeling(ActionsMod).TaskMacPropertiesForm;
      MacProperties.Properties['ChannelIndex'].Data[0]   := Properties['ChannelIndex'].Data[0];
      MacProperties.lstFitData.Items.Assign(prPreview.SelectionLists.Items);

      prTaskMac := TQUB_Task_Mac.Create(True{Suspended});
      TQUB_Task_Mac(prTaskMac).OnTerminate    := prOnMacDone;

      MacProperties.OnGetProperties := TQUB_Task_Mac(prTaskMac).OnGetProperties;
      MacProperties.OnPutProperties := TQUB_Task_Mac(prTaskMac).OnPutProperties;

      MacProperties.FormShow(MacProperties);
      MacProperties.ModalResult := mrOK;
      MacProperties.btnOKClick(MacProperties);

      TQUB_Task_Mac(prTaskMac).OnChange := TaskView.OnTaskChange;
      TQUB_Task_Mac(prTaskMac).Data     := prPreview;
      TQUB_Task_Mac(prTaskMac).Model    := ViewportModel.CurrModel;

      TQUB_Task_Mac(prTaskMac).OptPParams.OptMode    := MODE_SIM;//MODE_CHECK; Chris, John, what do you think?
      TQUB_Task_Mac(prTaskMac).OptPParams.OptShowFit := 1;
      TQUB_Task_Mac(prTaskMac).SilentOutput          := True;
      TQUB_Task_Mac(prTaskMac).PublishResults        := False;

      TQUB_Task_Mac(prTaskMac).Resume;
    end;
  end;
  panPreview.Invalidate;
end;

procedure TQUB_Task_Stimulus_PropertiesForm.prOnMacDone(Sender :TObject);
begin
  prTaskMac := nil;
  if (not Visible) and (prTaskIdl = nil) then begin
    SyncGUIProc(prCleanupPreview);
  end else if prNeedMac then begin
    SyncGUIProc(prInvalidateLiveMac);
  end else begin
    SyncGUIProc(panPreview.Invalidate);
    lblStatus.Caption := prLevelsMsg;
  end;
end;

//Data Source
procedure TQUB_Task_Stimulus_PropertiesForm.chkDataSourceClick(Sender: TObject);
begin
  inherited;
  DataSource := TQUB_Task_Data_Source(TRadioButton(Sender).Tag);
end;

procedure TQUB_Task_Stimulus_PropertiesForm.panPreviewPaint(Sender: TObject);
var
  Sel               :TQUB_FirstLast;
  IdlIter           :TQUB_IdlSegmentsIter;
  F, L, C           :Integer;
  A                 :Double;

  CurveIter         :TQUB_FitCurveIter;
  CurveData         :TPAInteger;

  dc, sc            :Integer;
  i, n              :Integer;
  w, h, y0, y1, dy  :Integer;
  x, y              :Integer;
  iscal, ioff       :Double;

  function Inflate(var a, b :Integer): Integer;
  var
    d,x    :Integer;
  begin
    d      := b - a;
    if d = 0 then d := 1;
    x      := (d * 15) div 100;
    if x = 0 then x := 1;
    a      := a - x;
    b      := b + x;
    Result := d + 2*x;
  end;
begin
  inherited;
  if prPreview = nil then Exit;

  dc                := Properties['ChannelIndex'].Data[0];
  sc                := Properties['Stimulus Channel'].Data[0];
  w                 := panPreview.Width;
  h                 := panPreview.Height;
  n                 := prN;
  Sel.First         := 0;
  Sel.Last          := n - 1;

  with panPreview.Canvas do begin
    Brush.Style := bsSolid;
    Brush.Color := clWhite;
    Pen.Color   := clWhite;
    Rectangle(0, 0, w, h);

    y0 := prData.DataYMinAuto[dc];
    y1 := prData.DataYMaxAuto[dc];
    dy := Inflate(y0, y1);

    // black 'ChannelIndex'
    Pen.Color := clBlack;
    Pen.Width := 1;
    prPreview.LoadDataChannel(prBuffer, dc, Sel, False, False, False);
    MoveTo(0, ((y1 - prBuffer[0])*h) div dy);
    for i := 1 to n - 1 do begin
      LineTo((i*w) div n, ((y1 - prBuffer[i])*h) div dy);
    end;

    // fuchsia FitCurve
    Pen.Color    := clFuchsia;
    CurveIter    := nil;
    try
      CurveIter    := prPreview.FitCurves.Find(0);
      while 0 = CurveIter.Done do begin
        CurveData       := CurveIter.Curve.Data;
        MoveTo(0, ((y1 - CurveData[0])*h) div dy);
        for i := 1 to (CurveIter.Curve.Last - CurveIter.Curve.First + 1) - 1 do begin
          LineTo((i*w) div n, ((y1 - CurveData[i])*h) div dy);
        end;
        CurveIter.Next;
      end;
    finally
      CurveIter.Free;
    end;

    y0 := prData.DataYMinAuto[sc];
    y1 := prData.DataYMaxAuto[sc];
    dy := Inflate(y0, y1);

    // thick? blue 'Stimulus Channel'
    Pen.Color := clBlue;
    // Pen.Width := 2;
    prPreview.LoadDataChannel(prBuffer, sc, Sel, False, False, False);
    MoveTo(0, ((y1 - prBuffer[0])*h) div dy);
    for i := 1 to n - 1 do begin
      LineTo((i*w) div n, ((y1 - prBuffer[i])*h) div dy);
    end;

    // thick red 'Stimulus Channel'.Idealization
    Pen.Color := clRed;
    Pen.Width := 2;
    IdlIter    := prPreview.DataObject.Idealization[sc].FindIter(0);
    iscal      := prPreview.DataObject.Scaling * prPreview.DataChannelScaling[sc];
    ioff       := prPreview.DataChannelOffset[sc];
    while 0 <> IdlIter.Valid do begin
      IdlIter.Read(@F, @L, @C, @A);
      IdlIter.Next;
      if C < 0 then Continue;

      x := (F*w) div n;
      y := ((y1 - Round((A - ioff) * iscal)) * h) div dy;
      if F = 0 then begin
        MoveTo(x, y);
      end else begin
        LineTo(x, y);
      end;
      x := ((L+1)*w) div n;
      LineTo(x, y);
    end;
    IdlIter.Free;
  end;
end;


//******************************************************************************
end.

