unit QUB_Task_Survey_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Registry,

RegistryInterface,

QUB_Global,
QUB_QFS_Channels,
QUB_Properties,
QUB_Task,
QUB_Task_Properties,
QUB_Node;



const

DEF_FROM  :Double = 0.0;
DEF_TO    :Double = 1.0;
DEF_STEPS :Integer = 10;

type


TQUB_Task_Survey_PropertiesForm = class(TQUB_Task_PropertiesForm)

  panParams                       :TPanel;
  Label1                          :TLabel;
  Label2                          :TLabel;
  Label3                          :TLabel;
  Label4                          :TLabel;
  Label6                          :TLabel;
  mnuVary                         :TComboBox;
  Panel1                          :TPanel;
  chkRange                        :TRadioButton;
  chkList                         :TRadioButton;
  txtFrom                         :TEdit;
  txtTo                           :TEdit;
  txtList                         :TEdit;
  txtStep                         :TEdit;
  btnGrab                         :TButton;
  mnuAlgorithm                    :TComboBox;
  Panel2                          :TPanel;
  chkEvaluate                     :TRadioButton;
  chkOptimize                     :TRadioButton;
    chkRestore: TCheckBox;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);

  procedure btnGrabClick(Sender: TObject);
  procedure mnuVaryChange(Sender: TObject);
  procedure mnuAlgorithmChange(Sender: TObject);


protected

  prNodes       :array of INode;
  prIxs         :array of Integer;

  //Properties vs. Registry
  procedure prSetInitialProperties(Props :INode); override;
  procedure prUpdateComponents; override;


public

  Model       :INode;
  Node        :INode;
  Ix          :Integer;

end;


implementation


uses


QUB_Viewport_Model, QUB_QVO_QVM, QUB_GUIInterface, QUB_MIL_Global;


{$R *.DFM}


//******************************************************************************
procedure TQUB_Task_Survey_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["In"]',
                          '', REGTYPE_NONE, MILModeStringOf(MIL_MODE_EVAL))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkRange,
                                       'Range',
                                       0))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkList,
                                       'List',
                                       1))
                       );
  AddEditProperty(TQUB_Property_Edit.Create('["From"]',
                    '', REGTYPE_NONE, DEF_FROM,
                    txtFrom,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["To"]',
                    '', REGTYPE_NONE, DEF_TO,
                    txtTo,
                    AcceptFloat,
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["Steps"]',
                    '', REGTYPE_NONE, DEF_STEPS,
                    txtStep,
                    AcceptIntGreaterThan(1),
                    '') );
  AddEditProperty(TQUB_Property_Edit.Create('["List"]',
                          '', REGTYPE_NONE, 0,
                          txtList,
                          AcceptFloatList(AcceptFloat),
                          'lambda node, ix: SpacedNodeData(node)') );
  AddRadioGroupProperty(TQUB_Property_RadioGroup.Create('["Mode"]',
                          '', REGTYPE_NONE, MILModeStringOf(MIL_MODE_EVAL))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkOptimize,
                                       MILModeStringOf(MIL_MODE_OPT),
                                       Integer(MIL_MODE_OPT)))
                            .AddButton(TQUB_Property_RadioButton.Create(
                                       chkEvaluate,
                                       MILModeStringOf(MIL_MODE_EVAL),
                                       Integer(MIL_MODE_EVAL)))
                       );
  AddCheckBoxProperty(TQUB_Property_CheckBox.Create('["Restore"]',
                        '', REGTYPE_NONE, 1,
                        chkRestore));

  RegKeyName            := 'Software\QUB\Task Survey';
  ProfileCat            := 'Surv';

  WikiTopic             := 'Modeling:Survey#Properties';
end;

procedure TQUB_Task_Survey_PropertiesForm.FormShow(Sender: TObject);
var
  i,j,n      :Integer;
  p          :INode;
  occ        :array of Integer;

  procedure Add(Name :String; Node :INode; Ix :Integer);
  var
    ii     :Integer;
  begin
    ii     := Length(prNodes);
    SetLength(prNodes, ii+1);
    SetLength(prIxs, ii+1);
    mnuVary.Items.Add(Name);
    prNodes[ii] := Node;
    prIxs[ii]   := Ix;
  end;

begin
  inherited;

  Model := ViewportModel.CurrModel.GetAsNode;

  SetLength(prNodes, 0);
  SetLength(prIxs, 0);
  mnuVary.Items.Clear;

  Add('Channel Count', Model['ChannelCount'], 0);

  n := Model['Amps'].DataCount;
  SetLength(occ, n);
  for i := 0 to n - 1 do occ[i] := 0;
  p := Model['States'].Find('State');
  i := 0;
  while not p.IsNull do begin
    Add(Format('Pr %d', [i+1]), p['Pr'], 0);
    occ[Integer(p['Class'].Data[0])] := 1;
    p := p.SiblingSameName;
    Inc(i);
  end;
  if 0<>Model['Properties']['IeqFv'].Data[0] then begin
    Add('Vrev', Model['VRev'], 0);

    for i := 0 to n - 1 do begin
      if occ[i] = 1 then begin
        Add(Format('Cond cl.%d', [i+1]), Model['Conds'], i);
        Add(Format('Std cl.%d', [i+1]), Model['CondStds'], i);
      end;
    end;
  end else begin
    for i := 0 to n - 1 do begin
      if occ[i] = 1 then begin
        Add(Format('Amp cl.%d', [i+1]), Model['Amps'], i);
        Add(Format('Std cl.%d', [i+1]), Model['Stds'], i);
      end;
    end;
  end;

  p := Model['Rates'].Find('Rate');
  while not p.IsNull do begin
    i := p['States'].Data[0] + 1;
    j := p['States'].Data[1] + 1;
    Add(Format('k0 %d-%d', [i,j]), p['k0'], 0);
    if 0 <> p['Q'].Data[0] then Add(Format('k1 %d-%d', [i,j]), p['k1'], 0);
    Add(Format('k0 %d-%d', [j,i]), p['k0'], 1);
    if 0 <> p['Q'].Data[1] then Add(Format('k1 %d-%d', [j,i]), p['k1'], 1);

    p := p.SiblingSameName;
  end;

  i := mnuVary.Items.IndexOf(Properties['Vary'].DataAsString);
  if i < 0 then begin
    if mnuVary.ItemIndex < 0 then mnuVary.ItemIndex := 0;
    Properties['Vary'].SetData(mnuVary.Text);
  end else begin
    mnuVary.ItemIndex := i;
  end;
  Node := prNodes[mnuVary.ItemIndex];
  Ix   := prIxs[mnuVary.ItemIndex];

end;

procedure TQUB_Task_Survey_PropertiesForm.btnGrabClick(Sender: TObject);
begin
  inherited;

  txtList.Text := FloatToStr(Double(Node.Data[Ix]));
  Properties['List'].SetData(Node.Data[Ix]);
end;

procedure TQUB_Task_Survey_PropertiesForm.mnuVaryChange(Sender: TObject);
begin
  inherited;

  Properties['Vary'].SetData(mnuVary.Text);
  Node := prNodes[mnuVary.ItemIndex];
  Ix   := prIxs[mnuVary.ItemIndex];
end;

procedure TQUB_Task_Survey_PropertiesForm.mnuAlgorithmChange(Sender: TObject);
begin
  inherited;

  Properties['Algorithm'].SetData(mnuAlgorithm.Text);
end;


//******************************************************************************


procedure TQUB_Task_Survey_PropertiesForm.prSetInitialProperties(Props :INode);
var
  Prop  :INode;
begin
  inherited;

  Prop  := Props['Algorithm'];
  if Prop.DataCount = 0 then begin
    Prop.SetData('Mac');
  end;
end;

procedure TQUB_Task_Survey_PropertiesForm.prUpdateComponents;
var
  i          :Integer;
begin
  inherited;

  i := mnuVary.Items.IndexOf( Properties['Vary'].DataAsString );
  if i >= 0 then begin
    mnuVary.ItemIndex := i;
  end else if mnuVary.ItemIndex < 0 then begin
    mnuVary.ItemIndex := 0;
  end;
  if mnuVary.ItemIndex >= 0 then begin
    Node := prNodes[mnuVary.ItemIndex];
    Ix   := prIxs[mnuVary.ItemIndex];
  end else begin
    Node := QUB_Node_Null;
    Ix   := -1;
  end;

  mnuAlgorithm.ItemIndex := mnuAlgorithm.Items.IndexOf( Properties['Algorithm'].DataAsString );
end;


//******************************************************************************
end.

