{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// a sub-thread of a multi-threaded task (only Baseline and Idealize so far, probably never any more)


unit QUB_Thread;


interface


uses


SysUtils, Classes, Windows, Forms, SyncObjs, ScktComp,
AllocTypes, Alloc,

QUB_Global,
QUB_Socket;


const


CHUNK_MAX_TRIALS = 5;


type


{$MINENUMSIZE 4}
TQUB_Thread_Chunk_Status   = (CHUNK_STATUS_EMPTY, CHUNK_STATUS_RUNNING, CHUNK_STATUS_STOPPED, CHUNK_STATUS_DONE, CHUNK_STATUS_ERROR);
{$MINENUMSIZE 1}

TQUB_Thread_Chunk_PStatus  = ^TQUB_Thread_Chunk_Status;

TQUB_Thread_Chunk          = record
  Sel                      :TQUB_FirstLast;
  StartTime                :Single;
  Status                   :TQUB_Thread_Chunk_Status;
  Trials                   :Integer;
  Index                    :Integer;
  Completion               :Integer;
end;
TQUB_Thread_PChunk         = ^TQUB_Thread_Chunk;


TQUB_Thread                = class;

TQUB_Thread_Event          = (THREAD_EVENT_REQUEST_DATA, THREAD_EVENT_REPORT_STATUS, THREAD_EVENT_REPORT_TERMINATION, THREAD_EVENT_REPORT_STOPPED, THREAD_EVENT_REPORT_ERROR, THREAD_EVENT_REPORT_MESSAGE);

TQUB_Thread_Event_Rec      = record
  Thread                   :TQUB_Thread;
  Event                    :TQUB_Thread_Event;
end;
TQUB_Thread_Event_PRec     = ^TQUB_Thread_Event_Rec;

TQUB_Thread_Event_ReportStatus  = record
  Rec                      :TQUB_Thread_Event_Rec;
  Completion               :Integer;
end;
TQUB_Thread_Event_PReportStatus = ^TQUB_Thread_Event_ReportStatus;

TQUB_Thread_Event_RequestData = record
  Rec                      :TQUB_Thread_Event_Rec;
  DataSel                  :TQUB_FirstLast;
  Data                     :Pointer;
end;
TQUB_Thread_Event_PRequestData = ^TQUB_Thread_Event_RequestData;

{$MINENUMSIZE 4}
TQUB_Thread_Event_Message_Kind = (MESSAGE_MESSAGE, MESSAGE_WARNING, MESSAGE_ERROR);
{$MINENUMSIZE 1}

TQUB_Thread_Event_ReportMessage = record
  Rec                      :TQUB_Thread_Event_Rec;
  MsgKind                  :TQUB_Thread_Event_Message_Kind;
  MsgSize                  :Integer;
  Msg                      :Pointer;
end;
TQUB_Thread_Event_PReportMessage = ^TQUB_Thread_Event_ReportMessage;


TQUB_Thread_Expect         = (QUB_THREAD_EXPECT_NOTHING,
                              QUB_THREAD_EXPECT_RECV_PARAMS
                             );

TQUB_Thread_Process_Result = (QUB_THREAD_PROCESS_RESULT_NOTHING,
                              QUB_THREAD_PROCESS_RESULT_READ, QUB_THREAD_PROCESS_RESULT_WRITE
                             );


TQUB_Thread_NotifyEvent    = procedure(Sender :TObject; Event :TQUB_Thread_Event_PRec) of object;


TQUB_Thread_MessageEvent   = procedure(Sender :TObject; Msg :TQUB_Socket_PMessage) of object;


//ClientSocket
TQUB_Thread_ClientSocket   = class(TClientSocket);


//Thread
TQUB_Thread                = class(TThread)
protected


  //Chunk
  prChunkCount             :Integer;
  prChunks                 :TPAPointer;

  //Signal
  prSignal                 :TSimpleEvent;

  //The thread runs locally or remotely
  prRemote                 :Boolean;
  prHost                   :String;
  prPort                   :Integer;

  //Running flags
  prStopped                :Boolean;
  prTimedOut               :Boolean;

  //Completion
  prCompletion             :Integer;

  //Events
  prOnEvent                :TQUB_Thread_NotifyEvent;

  //This event is invoked when the thread (remote) wants to send out a message
  prOnMessage              :TQUB_Thread_MessageEvent;

  prProcessBuffer          :Pointer;
  prProcessSize            :Integer;


  procedure prReportStatus;
  procedure prReportTermination;
  procedure prReportError;

  procedure prSetChunkCount(Count :Integer);
  procedure prSetChunk(Index :Integer; Chunk :TQUB_Thread_PChunk);
  function prGetChunk(Index :Integer) :TQUB_Thread_PChunk;


public


  constructor Create(Remote :Boolean; Suspended :Boolean); virtual;
  destructor Destroy; override;


  procedure Abort; virtual;
  procedure SetEvent;

  function ProcessMessage(Msg :TQUB_Socket_PMessage) :TQUB_Thread_Process_Result; virtual; abstract;


  property ProcessBuffer           :Pointer                       read prProcessBuffer;
  property ProcessSize             :Integer                       read prProcessSize;

  property Terminated;

  //Remote and MultiThread stuff
  property Remote                  :Boolean                       read prRemote         write prRemote;
  property Host                    :String                        read prHost           write prHost;
  property Port                    :Integer                       read prPort           write prPort;

  property ChunkCount              :Integer                       read prChunkCount     write prSetChunkCount;
  property Chunk[Index :Integer]   :TQUB_Thread_PChunk            read prGetChunk       write prSetChunk;

  property Signal                  :TSimpleEvent                  read prSignal;
  property TimedOut                :Boolean                       read prTimedOut;
  property Completion              :Integer                       read prCompletion;

  property OnEvent                 :TQUB_Thread_NotifyEvent       read prOnEvent        write prOnEvent;

  property OnMessage               :TQUB_Thread_MessageEvent      read prOnMessage      write prOnMessage;


end;


TQUB_Thread_ChunkQueue = class
protected

  prQ               :TList;
  prLock            :TCriticalSection;
  prRqAvail         :TEvent; // wakes up Accept()

public

  constructor Create;
  destructor Destroy; override;

  procedure Add(Chunk :TQUB_Thread_PChunk);
  function Accept(TimeoutMS :DWord): TQUB_Thread_PChunk; // or nil if timed out

end;


implementation


//******************************************************************************
constructor TQUB_Thread.Create(Remote :Boolean; Suspended :Boolean);
begin

  inherited Create(Suspended);

  prRemote        := Remote;

  prSignal        := TSimpleEvent.Create;

  FreeOnTerminate := False;

  prStopped       := False;
  prTimedOut      := False;

  prCompletion    := 0;

  prSetChunkCount(1);

end;


destructor TQUB_Thread.Destroy;
begin

  try

    prSetChunkCount(0);
    prSignal.Free;

  finally

    inherited;

  end;

end;


procedure TQUB_Thread.Abort;
begin

  prStopped := True;

  SetEvent;

end;

procedure TQUB_Thread.SetEvent;
begin

  prSignal.Acquire;
  try
    prSignal.SetEvent;
  finally;
    prSignal.Release;
  end;

end;


//Chunks
procedure TQUB_Thread.prSetChunkCount(Count :Integer);
begin
  prChunkCount := Count;
  ReAllocMem(prChunks, prChunkCount * SizeOf(Pointer));
end;

procedure TQUB_Thread.prSetChunk(Index :Integer; Chunk :TQUB_Thread_PChunk);
begin
  if (Index >= 0) and (Index < ChunkCount) then prChunks[Index] := Chunk;
end;

function TQUB_Thread.prGetChunk(Index :Integer) :TQUB_Thread_PChunk;
begin
  if (Index >= 0) and (Index < ChunkCount) then Result := prChunks[Index] else Result := nil;
end;


//****************************************************************************//
//                                                                            //
//                                 R E P O R T                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread.prReportStatus;
var
  Rec :TQUB_Thread_Event_PReportStatus;
begin

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_ReportStatus));

    //Set the Event
    Rec.Rec.Thread := Self;
    Rec.Rec.Event  := THREAD_EVENT_REPORT_STATUS;
    Rec.Completion := prCompletion;

    prOnEvent(Self, TQUB_Thread_Event_PRec(Rec));

  end;

end;

procedure TQUB_Thread.prReportTermination;
var
  Rec :TQUB_Thread_Event_PRec;
begin

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_Rec));

    //Set the Event
    Rec.Thread := Self;
    Rec.Event  := THREAD_EVENT_REPORT_TERMINATION;

    prOnEvent(Self, Rec);

  end;

  //Must Rec be freed ?!
end;

procedure TQUB_Thread.prReportError;
var
  Rec :TQUB_Thread_Event_PRec;
begin

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_Rec));

    //Set the Event
    Rec.Thread := Self;
    Rec.Event  := THREAD_EVENT_REPORT_STOPPED; //ERROR ?

    prOnEvent(Self, Rec);

  end;

end;


//******************************************************************************

constructor TQUB_Thread_ChunkQueue.Create;
begin
  prQ          := TList.Create;
  prLock       := TCriticalSection.Create;
  prRqAvail    := TEvent.Create(nil{security}, False{auto-reset}, False{init:off}, ''{unique});
end;

destructor TQUB_Thread_ChunkQueue.Destroy;
begin
  try
    prQ.Free;
    prLock.Free;
    prRqAvail.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Thread_ChunkQueue.Add(Chunk :TQUB_Thread_PChunk);
begin
  prLock.Acquire;
  prQ.Add(Chunk);
  prLock.Release;
  prRqAvail.SetEvent;
end;

//function TQUB_Thread_ChunkQueue.Rescind(Chunk :TQUB_Thread_PChunk): Boolean; // false if it wasn't in the queue (i.e. is running or done)
//begin
//  prLock.Acquire;
//  Result := prQ.Remove(Chunk) <> -1;
//  prLock.Release;
//end;

function TQUB_Thread_ChunkQueue.Accept(TimeoutMS :DWord): TQUB_Thread_PChunk; // or nil if timed out
begin
  Result := nil;

  prLock.Acquire;
  if prQ.Count > 0 then begin
    Result := TQUB_Thread_PChunk( prQ[0] );
    prQ.Delete(0);
  end;
  prLock.Release;

  if Result = nil then begin // Q was empty
    prRqAvail.WaitFor( TimeoutMS );
    prLock.Acquire;
    if prQ.Count > 0 then begin
      Result := TQUB_Thread_PChunk( prQ[0] );
      prQ.Delete(0);
    end;
    prLock.Release;
  end;

  prRqAvail.ResetEvent;
end;


//******************************************************************************
end.

