unit QUB_Thread_Data;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


SysUtils, Classes, Windows, Forms, SyncObjs,
AllocTypes, Alloc,

QUB_Global, QUB_QVO_QVD_Global, QUB_Thread, QUB_Socket;


const


//User Defined, Idealize-specific Messages
MESSAGE_RECV_PARAMS                    = QUB_SOCKET_MESSAGE_USER_DEFINED + 1;
MESSAGE_REQUEST_DATA                   = QUB_SOCKET_MESSAGE_USER_DEFINED + 2;
MESSAGE_RECV_DATA                      = QUB_SOCKET_MESSAGE_USER_DEFINED + 3;


type


TQUB_Thread_Data_ProcessState = (PROCESSSTATE_NOTHING,
                                 PROCESSSTATE_READING_CHUNK,
                                 PROCESSSTATE_WRITING_ACK
                                );


TQUB_Thread_Data = class(TQUB_Thread)
protected


  prNeedData               :TPAInteger;
  prNeedDataSel            :TQUB_FirstLast;
  prNeedDataChunk          :TQUB_Thread_PChunk;
  prNeedData16Bit          :Boolean;
  prHasDataSel             :TQUB_FirstLast;

  prReportMsg              :String;
  prReportMsgKind          :TQUB_Thread_Event_Message_Kind;


  procedure prRequestData; virtual;
  procedure prReportMessage; virtual;

  function prReceiveParams(Msg  :TQUB_Socket_PMessage) :TQUB_Thread_Process_Result; virtual;
  function prReceiveData(Msg  :TQUB_Socket_PMessage) :TQUB_Thread_Process_Result; virtual;

  procedure prCheckNeedsLoadData(var Win :TQUB_FirstLast; Sense :Integer; DataSize :Integer); virtual;


public


  constructor Create(Remote :Boolean; Suspended :Boolean); override;
  destructor Destroy; override;

  function  ProcessMessage(Msg :TQUB_Socket_PMessage) :TQUB_Thread_Process_Result; override;


  property NeedData        :TPAInteger            read prNeedData;
  property NeedDataSel     :TQUB_FirstLast        read prNeedDataSel;
  property NeedDataChunk   :TQUB_Thread_PChunk    read prNeedDataChunk;


end;


implementation


//******************************************************************************
constructor TQUB_Thread_Data.Create(Remote :Boolean; Suspended :Boolean);
begin

  inherited Create(Remote, Suspended);

  Priority := tpHigher;

end;

destructor TQUB_Thread_Data.Destroy;
begin

  try

    FreeMem(prProcessBuffer);

  finally

    inherited;

  end;

end;


//Process Message
function TQUB_Thread_Data.ProcessMessage(Msg :TQUB_Socket_PMessage) :TQUB_Thread_Process_Result;
begin

  Result := QUB_THREAD_PROCESS_RESULT_NOTHING;

  case Msg.Message of

    MESSAGE_RECV_PARAMS:
      begin

        Result := prReceiveParams(Msg);

      end;//MESSAGE_RECV_PARAMS

    MESSAGE_RECV_DATA:
      begin

        Result := prReceiveData(Msg);

      end;//MESSAGE_RECV_DATA

  end;//case Message

end;

//Receive Params
//This function must be overriden by descendent classes
//Here does nothing
function TQUB_Thread_Data.prReceiveParams(Msg  :TQUB_Socket_PMessage) :TQUB_Thread_Process_Result;
begin
  Result := QUB_THREAD_PROCESS_RESULT_NOTHING;
end;

//Receive Data
//The base class function assumes the data is in one channel format
//If multiple data channels, this function has to be overriden
function TQUB_Thread_Data.prReceiveData(Msg  :TQUB_Socket_PMessage) :TQUB_Thread_Process_Result;
var
  p :TPAByte;
begin

  p := TPAByte(Msg);
  Move(p[SizeOf(TQUB_Socket_Message)], prNeedData^, SizeOf(Integer) * (prNeedDataSel.Last - prNeedDataSel.First + 1));

  SetEvent;

  Result := QUB_THREAD_PROCESS_RESULT_NOTHING;

end;


//Report Message
procedure TQUB_Thread_Data.prReportMessage;
var
  Rec :TQUB_Thread_Event_PReportMessage;
begin

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_ReportMessage) + Length(prReportMsg));

    //Set the Event
    Rec.Rec.Thread := Self;
    Rec.Rec.Event  := THREAD_EVENT_REPORT_MESSAGE;
    Rec.MsgKind    := prReportMsgKind;
    Rec.MsgSize    := Length(prReportMsg);
    Rec.Msg        := @(Rec.Msg);
    Move(PChar(prReportMsg)^, Rec.Msg^, Rec.MsgSize);

    prOnEvent(Self, TQUB_Thread_Event_PRec(Rec));

  end;

end;

//Request Data
procedure TQUB_Thread_Data.prRequestData;
var
  Rec :TQUB_Thread_Event_PRequestData;
begin

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_RequestData));

    //Set the Event
    Rec.Rec.Thread  := Self;
    Rec.Rec.Event   := THREAD_EVENT_REQUEST_DATA;
    Rec.DataSel     := prNeedDataSel;
    Rec.Data        := prNeedData;

    prOnEvent(Self, TQUB_Thread_Event_PRec(Rec));

  end;

end;


{ TODO -oLorin : Do this after eliminating remote stuff }
procedure TQUB_Thread_Data.prCheckNeedsLoadData(var Win :TQUB_FirstLast; Sense :Integer; DataSize :Integer);
begin

  if (prHasDataSel.Last < Win.Last) or (prHasDataSel.First > Win.First) then begin

    if Sense > 0 then begin

      prNeedDataSel.First := Win.First;
      if prNeedDataSel.First < Chunk[0].Sel.First then prNeedDataSel.First := Chunk[0].Sel.First;

      if DataSize > 0 then begin
        prNeedDataSel.Last  := prNeedDataSel.First + DataSize - 1;
      end else begin
        prNeedDataSel.Last  := Win.Last;
      end;
      if prNeedDataSel.Last  > Chunk[0].Sel.Last  then prNeedDataSel.Last  := Chunk[0].Sel.Last;

    end else begin

      prNeedDataSel.Last  := Win.Last;
      if prNeedDataSel.Last  > Chunk[0].Sel.Last  then prNeedDataSel.Last  := Chunk[0].Sel.Last;

      if DataSize > 0 then begin
        prNeedDataSel.First := prNeedDataSel.Last  - DataSize + 1;
      end else begin
        prNeedDataSel.First := Win.First;
      end;
      if prNeedDataSel.First < Chunk[0].Sel.First then prNeedDataSel.First := Chunk[0].Sel.First;

    end;

    prHasDataSel := prNeedDataSel;

    prRequestData;

    //Wait until it gets data or is aborted
    prSignal.WaitFor(INFINITE);
    prSignal.ResetEvent;

  end;

end;


//******************************************************************************
end.

