unit QUB_Thread_Idealize;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


SysUtils, Classes, Windows, Forms, SyncObjs,
AllocTypes, Alloc,

QUB_Matrices,

QUB_Global, QUB_QVO_QVD_Global,
QUB_Thread,
QUB_Thread_Data,
QUB_Idealize_Global,
QUB_QFS_QMF, QUB_QFS_Channels,
QUB_Node, QUB_Node_Python;


{$ASSERTIONS OFF}


type


TQUB_Thread_Idealize_ProcessState = (PROCESSSTATE_NOTHING,
                                     PROCESSSTATE_READING_CHUNK,
                                     PROCESSSTATE_WRITING_ACK
                                    );


TQUB_Thread_Idealize = class;


TQUB_Thread_Idealize = class(TQUB_Thread_Data)
protected


  //Parameters
  prParams                 :TQUB_Idealize_Params;
  prConfig                 :INode;
  prDataSet                :INode;

  prPath                   :String;

  prPercentDone            :Integer;

  prM                      :String;//only used for testing


  procedure prExecute_HalfAmp;
  procedure prExecute_SKM;

  function prReportCallback(arg :INode): INode;
  function prPercentCallback(arg :INode): INode;


public

  FillDataCallback           :TNodeMethPtr;

  constructor Create(Remote :Boolean; Suspended :Boolean); override;

  procedure Execute; override;

  procedure Abort; override;

  function DataCallback(Chunk :TQUB_Thread_Idealize_PChunk) :Integer; stdcall;
  function ReportCallBack(ReportKind :TQUB_Thread_Event_Message_Kind; message :PChar) :Integer; stdcall;
  function OutAllocCallBack(Chunk :TQUB_Thread_Idealize_PChunk) :Integer; stdcall;


  property Params          :TQUB_Idealize_Params        read prParams           write prParams;
  property Config          :INode                       read prConfig           write prConfig;
  property DataSet         :INode                       read prDataSet          write prDataSet;
  property PercentDone     :Integer                     read prPercentDone      write prPercentDone;

end;


function SKMFromDll( cfg :TQUB_QFS_PNodeMem; datasets :TPAPointer; fillDataCB, reportCB, percentCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_skmtreeiface@20';

function SKM_Get_Multiclasses_DLL(Model :TQUB_QFS_PNodeMem; ChannelCount :Integer): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_skm_get_multiclasses_iface@8';
function SKM_Get_Multiclasses(Model :INode; ChannelCount :Integer): INode;

procedure ApplyDeadTime( IdlSeg :INode; DeadTime :Double );


implementation


uses

QUB_Task_Stat;

{UNUSED - members of class now...?}
//Unit-level Callbacks
//function DataCallback(caller_obj :TObject; Chunk :TQUB_Thread_Idealize_PChunk) :Integer; stdcall; forward;
//function ReportCallBack(caller_obj :TObject; ReportKind :TQUB_Thread_Event_Message_Kind; message :PChar) :Integer; stdcall; forward;
//function OutAllocCallBack(caller_obj :TObject; Chunk :TQUB_Thread_Idealize_PChunk) :Integer; stdcall; forward;


//****************************************************************************//
//                                                                            //
//                 U n i t - l e v e l    C a l l b a c k s                   //
//                                                                            //
//****************************************************************************//
{
function DataCallback(caller_obj :TObject; Chunk :TQUB_Thread_Idealize_PChunk) :Integer;
begin
  Result := TQUB_Thread_Idealize(caller_obj).DataCallBack(Chunk);
end;

function ReportCallBack(caller_obj :TObject; ReportKind :TQUB_Thread_Event_Message_Kind; message :PChar) :Integer;
begin
  Result := TQUB_Thread_Idealize(caller_obj).ReportCallBack(ReportKind, message);
end;

function OutAllocCallBack(caller_obj :TObject; Chunk :TQUB_Thread_Idealize_PChunk) :Integer;
begin
  Result := TQUB_Thread_Idealize(caller_obj).OutAllocCallBack(Chunk);
end;
}


//******************************************************************************
constructor TQUB_Thread_Idealize.Create(Remote :Boolean; Suspended :Boolean);
begin

  inherited Create(Remote, Suspended);

  Priority := tpHigher;

end;


//Abort
procedure TQUB_Thread_Idealize.Abort;
begin

  inherited;

end;


//Execute **********************************************************************
procedure TQUB_Thread_Idealize.Execute;
begin

  if Params.Method = METHOD_HALFAMP then begin
    prExecute_HalfAmp;
  end else begin
    prExecute_SKM;
  end;

end;

procedure TQUB_Thread_Idealize.prExecute_SKM;
var
  Error               :Boolean;
  IdlChunk            :TQUB_Thread_Idealize_PChunk;
  i                   :Integer;

  Seg                 :INode;
  DataSetPtrs         :TPAPointer;
  fillDataCB          :INodeCFunc;
  rptCB               :INodeCFunc;
  pctCB               :INodeCFunc;
  ResultNode          :INode;
  ResultNodePtr       :TQUB_QFS_PNodeMem;
  StatResult          :INode;
  StatSeg             :INode;
begin

  prNeedData          := nil;
  prPercentDone       := 0;
  DataSetPtrs         := nil;

  //Reset error
  Error               := False;

  try

    for i := 0 to ChunkCount - 1 do begin
      IdlChunk               := TQUB_Thread_Idealize_PChunk(Chunk[i]);
      Chunk[i].Status        := CHUNK_STATUS_RUNNING;
      IdlChunk.HistBinCount  := prParams.SKM.HistBinCount;

      Seg                    := TNode.Create('Segment');
      Seg.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
      Seg.Data[0]            := Chunk[i].Sel.First;
      Seg.Data[1]            := Chunk[i].Sel.Last;
      Seg['start'].SetData( (Chunk[i].StartTime - IdlChunk.BaseTime) * 1.0e3 );
      Seg['length'].SetData( (Seg.Data[1] - Seg.Data[0] + 1) * DataSet['sampling'].Data[0] );
      Seg['Index'].SetData( i );
      DataSet.AppendChild( Seg );

    end;

    GetMem( DataSetPtrs, SizeOf(TQUB_QFS_PNodeMem) * ({Length(prDatasets)}1 + 1) );
    DataSetPtrs[0] := Pointer(DataSet.GetNodeMem);
    DataSetPtrs[ 1 ] := nil;

    fillDataCB  := TNodeCFunc.Wrap( TNodeMeth.Create( FillDataCallback ) );
    rptCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prReportCallback ) );
    pctCB       := TNodeCFunc.Wrap( TNodeMeth.Create( Self.prPercentCallback ) );

    try

      ResultNodePtr  := SKMFromDll( Config.GetNodeMem, DataSetPtrs, fillDataCB, rptCB, pctCB );
      ResultNode     := TNode.CreateUsing( ResultNodePtr );
      if ResultNodePtr <> nil then Node_DECREF( ResultNodePtr );

      Seg            := ResultNode['DataSet'].Find('Segment');
      for i := 0 to ChunkCount - 1 do begin
        if Seg.IsNull then Break;

        IdlChunk               := TQUB_Thread_Idealize_PChunk(Chunk[i]);
        IdlChunk.ResultNode    := Seg.GetNodeMem;
        Node_INCREF( IdlChunk.ResultNode );

        if prParams.IdlDeadTime then begin
          ApplyDeadTime( Seg, ResultNode['DataSet']['ExpCond']['DeadTime'].Data[0] );
        end;

        Seg          := Seg.SiblingSameName;
      end;

      // run stat
      StatResult     := StatDataSet(ResultNode['DataSet'], Config, FillDataCallback, prReportCallback);

      StatSeg          := StatResult['DataSet'].Find('Segment');
      Seg              := ResultNode['DataSet'].Find('Segment');
      while not StatSeg.IsNull do begin
        while not StatSeg.Child.IsNull do begin
          if Seg.Find( StatSeg.Child.Name ).IsNull then begin
            Seg.AppendChild( StatSeg.Child );
          end else begin
            StatSeg.RemoveChild( StatSeg.Child );
          end;
        end;

        StatSeg        := StatSeg.SiblingSameName;
        Seg            := Seg.SiblingSameName;
      end;

    except

      Error   := True;

    end;


  finally

    FreeMem(DataSetPtrs);

    prNeedData := nil;

     if (not prStopped) and (not Error) then begin

      prPercentDone := 100;
      for i := 0 to ChunkCount - 1 do begin
        Chunk[i].Status     := CHUNK_STATUS_DONE;
        Chunk[i].Completion := 100;
      end;

      prReportStatus;

    end else if (prStopped) or (Error) then begin
    //Error or Stopped

      if prStopped then begin
        for i := 0 to ChunkCount - 1 do begin
          Chunk[i].Status := CHUNK_STATUS_STOPPED;
        end;
      end else begin
        prReportMsg := 'Thread of chunks ';
        for i := 0 to ChunkCount - 1 do begin
          Chunk[i].Status := CHUNK_STATUS_ERROR;
          prReportMsg := prReportMsg + IntToStr(Chunk[i].Index);
          if i <> ChunkCount - 1 then prReportMsg := prReportMsg + ', ';
        end;
        prReportMsg := prReportMsg + ' failed to run';
        prReportMsgKind := MESSAGE_ERROR;
      end;

      prReportMessage;

      //Reset output
      for i := 0 to ChunkCount - 1 do begin

        IdlChunk := TQUB_Thread_Idealize_PChunk(Chunk[i]);

        FreeMem(IdlChunk.Data);             IdlChunk.Data        := nil;

        IdlChunk.SegCount    := 0;

        FreeMem(IdlChunk.SegFirsts);        IdlChunk.SegFirsts   := nil;
        FreeMem(IdlChunk.SegLasts);         IdlChunk.SegLasts    := nil;
        FreeMem(IdlChunk.SegClasses);       IdlChunk.SegClasses  := nil;

        IdlChunk.ClassCount := 0;

        FreeMem(IdlChunk.Amps);             IdlChunk.Amps        := nil;
        FreeMem(IdlChunk.Stds);             IdlChunk.Stds        := nil;
        FreeMem(IdlChunk.Occupancy);        IdlChunk.Occupancy   := nil;
        FreeMem(IdlChunk.Lifetime);         IdlChunk.Lifetime    := nil;
        FreeMem(IdlChunk.NEvents);          IdlChunk.NEvents     := nil;
        Free2(TPMChar(IdlChunk.AMatrix));   IdlChunk.AMatrix     := nil;

        Free2(TPMChar(IdlChunk.Histogram)); IdlChunk.Histogram   := nil;

        IdlChunk.ReportSize := 0;

        FreeMem(IdlChunk.Report);           IdlChunk.Report      := nil;

      end;

    end;

    prReportTermination;

    Terminate;

  end;

end;


//****************************************************************************//
//                                                                            //
//                          C A L L B A C K S                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Thread_Idealize.prReportCallback( arg :INode ): INode;
begin
  Result          := arg;
  prReportMsg     := arg.GetDataAsString;
  prReportMsgKind := MESSAGE_ERROR;
  prReportMessage;
end;

function TQUB_Thread_Idealize.prPercentCallback(arg :INode): INode;
var
  i           :Integer;
begin
  try
    prPercentDone     := Integer( arg.Data[0] );
  except
  end;

  if prPercentDone >= 100 then prPercentDone := 99;
  if prPercentDone < 0    then prPercentDone := 0;

  for i := 0 to ChunkCount - 1 do begin
    Chunk[i].Completion := prPercentDone;
  end;

  prReportStatus;

  Result := arg;
end;

//****************************************************************************//

{ TODO -oLorin : Calculate Amps and Stds }
procedure TQUB_Thread_Idealize.prExecute_HalfAmp;
var
  DataSize            :Integer;

  Win                 :TQUB_FirstLast;

  SegMaxCount         :Integer;

  CurrClass           :Integer;
  ThisClass           :Integer;

  SegFirst            :Integer;

  ClosedSeg           :Boolean;

  NewCompletion       :Integer;

  Error               :Boolean;
  p, c                :Integer;

  IdlChunk            :TQUB_Thread_Idealize_PChunk;
  ResultNode          :INode;
  SamplingMS          :Double;
  Clss, Durs          :INode;
  SegAmp, SegStd      :INode;

  StatResult          :INode;
  StatSeg             :INode;
begin

  IdlChunk            := TQUB_Thread_Idealize_PChunk(Chunk[0]);

  Chunk[0].Status     := CHUNK_STATUS_RUNNING;

  prNeedData          := nil;

  Error               := False;

  SegMaxCount         := 0;

  SamplingMS          := 0.01;
  //Scaling             := 320.0;

  DataSet['ChannelCount'].SetData( 1 );
  ResultNode          := DataSet['Segment'];
  IdlChunk.ResultNode := ResultNode.GetNodeMem;
  Node_INCREF( IdlChunk.ResultNode );

  try
    SamplingMS          := Double(DataSet['sampling'].Data[0]);
    //Scaling             := Double(DataSet['scaling'].Data[0]);
    ResultNode.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
    ResultNode.Data[0]            := Chunk[0].Sel.First;
    ResultNode.Data[1]            := Chunk[0].Sel.Last;
    ResultNode['start'].SetData( Chunk[0].StartTime * 1.0e3 );
    ResultNode['start'].SetLineComment('ms');
    ResultNode['length'].SetData( (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1) * SamplingMS );
    ResultNode['length'].SetLineComment('ms');
  except
  end;

  try

    //Find the Buffer Size
    if prRemote then begin
      DataSize := ONE_MEGABYTE div 16;
    end else begin
      DataSize := ONE_MEGABYTE;
    end;
    if DataSize > (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1) then DataSize := (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1);

    try

      //Allocate Buffer
      prNeedData := IAlloc1(DataSize);

    except

      Error := True;
      Exit;

    end;


    Win.First := Chunk[0].Sel.First;
    Win.Last  := Win.First + DataSize - 1;


    //It will need it !!!
    prHasDataSel.Last := Win.Last - 1;



    //**********************************************************************//
    //                                                                      //
    //                              Main loop                               //
    //                                                                      //
    //**********************************************************************//
    CurrClass := -1;
    SegFirst  := Chunk[0].Sel.First;
    ClosedSeg := False;

    repeat
    //Main loop

     //Stop thread if user-stopped
      if prStopped then begin
        Exit;
      end;

      NewCompletion := Round(100.0 * (Win.First - Chunk[0].Sel.First) / (Chunk[0].Sel.Last - Chunk[0].Sel.First + 1));
      if NewCompletion <> prCompletion then begin
        prCompletion := NewCompletion;
        Chunk[0].Completion := prCompletion;
        prReportStatus;
      end;

      prCheckNeedsLoadData(Win, +1, DataSize);

      if (prHasDataSel.Last < Win.Last) or (prHasDataSel.First > Win.First) then begin
        //Hasn't received the data
        Error := True;
      end;


      if (prStopped) or (Error) then Exit;


      //Classify points
      for p := Win.First to Win.Last do begin

        ClosedSeg := False;

        ThisClass := prParams.HalfAmp.ClassCount - 1;
        for c := 0 to prParams.HalfAmp.ClassCount - 2 do begin
          if prNeedData[p - prHasDataSel.First] < prParams.HalfAmp.HalfAmps[c] then begin
            ThisClass := c;
            Break;
          end;
        end;

        if ThisClass <> CurrClass then begin

          if CurrClass <> -1 then begin
          //Add Segment

            Inc(IdlChunk.SegCount);

            if IdlChunk.SegCount > SegMaxCount then begin

              Inc(SegMaxCount, 100000);

              try
                ReallocMem(IdlChunk.SegFirsts,    SizeOf(Integer) * SegMaxCount);
                ReallocMem(IdlChunk.SegLasts,     SizeOf(Integer) * SegMaxCount);
                ReallocMem(IdlChunk.SegClasses,   SizeOf(Integer) * SegMaxCount);
              except
                Dec(IdlChunk.SegCount);
                Error := True;
                Exit;
              end;

            end;
            IdlChunk.SegFirsts  [IdlChunk.SegCount - 1] := SegFirst;
            IdlChunk.SegLasts   [IdlChunk.SegCount - 1] := p - 1;
            IdlChunk.SegClasses [IdlChunk.SegCount - 1] := CurrClass;

            SegFirst  := p;

            ClosedSeg := True;

          end;

          CurrClass := ThisClass;

        end;

      end;


      Win.First := Win.Last + 1;
      Win.Last  := Win.First + DataSize - 1;
      if Win.Last > Chunk[0].Sel.Last then Win.Last := Chunk[0].Sel.Last;

    until (Win.First > Chunk[0].Sel.Last);
    //Main loop

    if not ClosedSeg then begin

      Inc(IdlChunk.SegCount);

      if IdlChunk.SegCount > SegMaxCount then begin

        SegMaxCount := IdlChunk.SegCount;

        try
          ReallocMem(IdlChunk.SegFirsts,    SizeOf(Integer) * SegMaxCount);
          ReallocMem(IdlChunk.SegLasts,     SizeOf(Integer) * SegMaxCount);
          ReallocMem(IdlChunk.SegClasses,   SizeOf(Integer) * SegMaxCount);
        except
          Dec(IdlChunk.SegCount);
          Error := True;
          Exit;
        end;
      end;

      IdlChunk.SegFirsts  [IdlChunk.SegCount - 1] := SegFirst;
      IdlChunk.SegLasts   [IdlChunk.SegCount - 1] := Chunk[0].Sel.Last;
      IdlChunk.SegClasses [IdlChunk.SegCount - 1] := CurrClass;

    end;

    // tree-ify the results and post-process with Stat; result node should be same as skm

    ResultNode['DwellCount'].SetData( IdlChunk.SegCount );
    Clss := ResultNode['Classes'];
    Clss.SetupNumData( NODE_DATATYPE_LONGINT, IdlChunk.SegCount, 1 );
    Durs := ResultNode['Durations'];
    Durs.SetupNumData( NODE_DATATYPE_SINGLE, IdlChunk.SegCount, 1 );

    for p := 0 to IdlChunk.SegCount - 1 do begin
      Clss.Data[p] := Params.HalfAmp.IdxAmps[ IdlChunk.SegClasses[p] ];
      Durs.Data[p] := SamplingMS * (IdlChunk.SegLasts[p] - IdlChunk.SegFirsts[p] + 1);
    end;

    SegAmp := ResultNode['amp'];
    SegStd := ResultNode['sd'];
    SegAmp.SetupNumData( NODE_DATATYPE_DOUBLE, Params.HalfAmp.ClassCount, 1 );
    SegStd.SetupNumData( NODE_DATATYPE_DOUBLE, Params.HalfAmp.ClassCount, 1 );
    for c := 0 to Params.HalfAmp.ClassCount - 1 do begin
      SegAmp.Data[c] := Params.HalfAmp.AmpVals[Params.HalfAmp.IdxAmps[c]];
      SegStd.Data[c] := Params.HalfAmp.StdVals[Params.HalfAmp.IdxAmps[c]];
    end;

    if prParams.IdlDeadTime then begin
      ApplyDeadTime( ResultNode, DataSet['ExpCond']['DeadTime'].Data[0] );
      // now the stuff in IdlChunk is outdated; we prefer the ResultNode stuff in Task_IdealizeMT
    end;

    if prParams.HalfAmp.ComputeStats then begin
      StatResult     := StatDataSet(DataSet, Config, FillDataCallback, prReportCallback);
      StatSeg          := StatResult['DataSet']['Segment'];
      while not StatSeg.Child.IsNull do begin
        if ResultNode.Find( StatSeg.Child.Name ).IsNull then begin
          ResultNode.AppendChild( StatSeg.Child );
        end else begin
          StatSeg.RemoveChild( StatSeg.Child );
        end;
      end;
    end;

  finally


    FreeMem(prNeedData);
    prNeedData := nil;

    if (not prStopped) and (not Error) then begin

      Chunk[0].Status     := CHUNK_STATUS_DONE;
      Chunk[0].Completion := 100;

      if IdlChunk.SegCount > 0 then begin

        if IdlChunk.SegCount <> SegMaxCount then begin
          ReallocMem(IdlChunk.SegFirsts,    SizeOf(Integer) * IdlChunk.SegCount);
          ReallocMem(IdlChunk.SegLasts,     SizeOf(Integer) * IdlChunk.SegCount);
          ReallocMem(IdlChunk.SegClasses,   SizeOf(Integer) * IdlChunk.SegCount);
        end;

      end;

    end else if (prStopped) or (Error) then begin
    //Error or Stopped

      if prStopped then begin
        Chunk[0].Status    := CHUNK_STATUS_STOPPED;
      end else begin
        Chunk[0].Status    := CHUNK_STATUS_ERROR;
        prReportMsg     := 'Thread ' + IntToStr(Chunk[0].Index) + ' failed to run';
        prReportMsgKind := MESSAGE_ERROR;
      end;

      prReportMessage;

      //Reset output
      FreeMem(IdlChunk.Data);             IdlChunk.Data        := nil;

      IdlChunk.SegCount    := 0;

      FreeMem(IdlChunk.SegFirsts);        IdlChunk.SegFirsts   := nil;
      FreeMem(IdlChunk.SegLasts);         IdlChunk.SegLasts    := nil;
      FreeMem(IdlChunk.SegClasses);       IdlChunk.SegClasses  := nil;

      IdlChunk.ClassCount := 0;

      FreeMem(IdlChunk.Amps);             IdlChunk.Amps        := nil;
      FreeMem(IdlChunk.Stds);             IdlChunk.Stds        := nil;
      FreeMem(IdlChunk.Occupancy);        IdlChunk.Occupancy   := nil;
      FreeMem(IdlChunk.Lifetime);         IdlChunk.Lifetime    := nil;
      FreeMem(IdlChunk.NEvents);          IdlChunk.NEvents     := nil;
      Free2(TPMChar(IdlChunk.AMatrix));   IdlChunk.AMatrix     := nil;

      Free2(TPMChar(IdlChunk.Histogram)); IdlChunk.Histogram   := nil;

      IdlChunk.ReportSize := 0;

      FreeMem(IdlChunk.Report);           IdlChunk.Report      := nil;

    end;

    prReportTermination;

    Terminate;

  end;

end;



//****************************************************************************//
//                                                                            //
//                              C A L L B A C K S                             //
//                                                                            //
//****************************************************************************//
function TQUB_Thread_Idealize.DataCallback(Chunk :TQUB_Thread_Idealize_PChunk) :Integer;
var
  BaseChunk :TQUB_Thread_PChunk;
  s, o      :Double;
  i         :Integer;
begin

  Result := 0;

  try

    BaseChunk := TQUB_Thread_PChunk(Chunk);

    if Chunk.Data = nil then begin
    //Allocate

      if (not Boolean(Chunk.DataIs16Bit)) and (SizeOf(Single) <> SizeOf(Integer)) then Halt;
      //Assumes SizeOf(Single) = SizeOf(Integer);

      Chunk.Data       := AllocMem(SizeOf(Single)   * (BaseChunk.Sel.Last - BaseChunk.Sel.First + 1));

      prNeedDataSel    := BaseChunk.Sel;
      prNeedData       := TPAInteger(Chunk.Data);
      prNeedDataChunk  := TQUB_Thread_PChunk(Chunk);
      prHasDataSel     := prNeedDataSel;

      prRequestData;

      //Wait until it gets data or is aborted
      prSignal.WaitFor(INFINITE);
      prSignal.ResetEvent;

      Result := Integer(prStopped);

      if Boolean(Chunk.DataIs16Bit) then begin
        for i := 0 to BaseChunk.Sel.Last - BaseChunk.Sel.First do begin
          try
            TPASmall(Chunk.Data)[i] := SmallInt(prNeedData[i]);
          except
          end;
        end;
      end else begin
        s := prParams.Scaling;
        o := prParams.Offset;
        for i := 0 to BaseChunk.Sel.Last - BaseChunk.Sel.First do begin
          Chunk.Data[i] := prNeedData[i] / s + o;
        end;
      end;

      //Possibly the percent is changed
      prReportStatus;


    end else begin
    //DeAllocate

      FreeMem(Chunk.Data);
      Chunk.Data := nil;

    end;

  except

    Result := 1;

  end;

end;

function TQUB_Thread_Idealize.ReportCallBack(ReportKind :TQUB_Thread_Event_Message_Kind; message :PChar) :Integer;
begin
  Result          := 0;
  prReportMsg     := message;
  prReportMsgKind := ReportKind;
  prReportMessage;
end;

function TQUB_Thread_Idealize.OutAllocCallBack(Chunk :TQUB_Thread_Idealize_PChunk) :Integer;
begin

  Result := 0;

  try

    if Chunk.SegCount > 0 then begin

      Chunk.SegClasses  := AllocMem(Chunk.SegCount    * SizeOf(Integer));
      Chunk.SegFirsts   := AllocMem(Chunk.SegCount    * SizeOf(Integer));
      Chunk.SegLasts    := AllocMem(Chunk.SegCount    * SizeOf(Integer));

      Chunk.Amps        := AllocMem(Chunk.ClassCount  * SizeOf(Single));
      Chunk.Stds        := AllocMem(Chunk.ClassCount  * SizeOf(Single));
      Chunk.Occupancy   := AllocMem(Chunk.ClassCount  * SizeOf(Single));
      Chunk.Lifetime    := AllocMem(Chunk.ClassCount  * SizeOf(Single));
      Chunk.NEvents     := AllocMem(Chunk.ClassCount  * SizeOf(Integer));

    end;

    if Chunk.StateCount > 0 then begin

      Chunk.AMatrix     := SAlloc2(Chunk.StateCount, Chunk.StateCount);

    end;

    Chunk.Histogram   := SAlloc2(Chunk.ClassCount + 3, Chunk.HistBinCount);

    Chunk.Report      := AllocMem(Chunk.ReportSize);

  except

    Result := 1;

  end;

end;


// *****************************************************************************

procedure ApplyDeadTime( IdlSeg :INode; DeadTime :Double );
var
  ndwells            :Integer;
  idwt               :TPAInteger;
  tdwt               :TPASingle;
  fdwt, ldwt         :TPAInteger;
  i, j               :Integer; // i: out_ix,  j: in_ix
  cls                :Integer;
  tm                 :Single;
begin
  ndwells            := Integer(IdlSeg['DwellCount'].Data[0]);
  idwt               := TPAInteger(IdlSeg['Classes'].DataPtr);
  tdwt               := TPASingle(IdlSeg['Durations'].DataPtr);
  fdwt               := TPAInteger(IdlSeg.Find('Firsts').DataPtr);
  ldwt               := TPAInteger(IdlSeg.Find('Lasts').DataPtr);

  // first, accumulate initial too-short events (mil drops them)
  tm                 := 0.0;
  for j := 0 to ndwells - 1 do begin
    if tdwt[j] <= DeadTime then begin
      tm             := tm + tdwt[j];
    end else begin
      tdwt[j]        := tdwt[j] + tm;
      if fdwt <> nil then fdwt[j] := fdwt[0];
      Break;
    end;
  end;
  // now j is the index of the first long-enough dwell, with its time expanded by the prev.too.shorts

  // then add the time-values of other too-shorts to the prev. dwell
  i                  := 0;
  while j < ndwells do begin
    cls              := idwt[j];
    tm               := tdwt[j];

    if (i > 0) and ((idwt[i-1] = cls) or (tm <= DeadTime)) then begin
      tdwt[i-1]      := tdwt[i-1] + tm;
      if ldwt <> nil then ldwt[i-1] := ldwt[j];
    end else begin
      idwt[i]        := cls;
      // tdwt[i]        := tm - DeadTime; how in hex would you display this over the data?
      tdwt[i]        := tm;
      if fdwt <> nil then fdwt[i] := fdwt[j];
      if ldwt <> nil then ldwt[i] := ldwt[j];
      i              := i + 1;
    end;

    j                := j + 1;
  end;

  IdlSeg['DwellCount'].Data[0] := i;
end;


//******************************************************************************

function SKM_Get_Multiclasses(Model :INode; ChannelCount :Integer): INode;
var
  ResultPtr  :TQUB_QFS_PNodeMem;
begin
  ResultPtr  := SKM_Get_Multiclasses_DLL(Model.GetNodeMem, ChannelCount);
  Result     := TNode.CreateUsing(ResultPtr);
  NODE_DECREF(ResultPtr);
end;


//******************************************************************************
end.


