unit QUB_Thread_Socket_Baseline;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


SysUtils,
AllocTypes, Alloc,

QUB_QVO_QVD_Global,
QUB_Socket,
QUB_Thread_Socket, QUB_Thread_Socket_Data,
QUB_Thread, QUB_Thread_Data, QUB_Thread_Baseline,
QUB_Baseline_Global;


type


TQUB_Thread_Socket_Baseline        = class(TQUB_Thread_Socket_Data)
protected


  //Baseline Parameters
  prParams                  :TQUB_Baseline_Params;


  procedure prSend_Params;

  procedure prReportTermination(var Msg :TQUB_Socket_Message);


public


  constructor Create(const RemoteServer :String; Port :Word); override;

  //Parameters
  property Params           :TQUB_Baseline_Params       read prParams      write prParams;


end;



implementation


//******************************************************************************
constructor TQUB_Thread_Socket_Baseline.Create(const RemoteServer :String; Port :Word);
begin

  inherited;

  prIdentification.Task   := QUB_SOCKET_ID_TASK_BASELINE;
  prIdentification.Thread := QUB_SOCKET_ID_THREAD_BASELINE;

end;


//****************************************************************************//
//                                                                            //
//                             S E N D     P A R A M S                        //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Baseline.prSend_Params;
var
  Msg :Pointer;
  s   :Integer;
  p   :TPAByte;
begin

  s := SizeOf(TQUB_Socket_Message) + SizeOf(TQUB_Thread_Baseline_Chunk) + SizeOf(prParams);


  Msg := AllocMem(s);

  try

    TQUB_Socket_PMessage(Msg).Message := MESSAGE_RECV_PARAMS;
    TQUB_Socket_PMessage(Msg).Size    := s;

    p := Msg;

    p := @p[SizeOf(TQUB_Socket_Message)];

    //Chunk
    Move(prChunk^,  p^, SizeOf(TQUB_Thread_Baseline_Chunk));

    p := @p[SizeOf(TQUB_Thread_Baseline_Chunk)];

    //Params - fixed
    Move(prParams,  p^, SizeOf(prParams));

    prWrite(Msg, s);

  finally

    FreeMem(Msg);

  end;

end;


//****************************************************************************//
//                                                                            //
//                     R E P O R T    T E R M I N A T I O N                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Baseline.prReportTermination(var Msg :TQUB_Socket_Message);
var
  Rec      :TQUB_Thread_Event_PRec;
  p        :TPAByte;
  BasChunk :TQUB_Thread_Baseline_PChunk;
begin

  BasChunk := TQUB_Thread_Baseline_PChunk(Chunk);

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_Rec));

    //Set the Event
    Rec.Thread := TQUB_Thread(Self);
    Rec.Event  := THREAD_EVENT_REPORT_TERMINATION;

    //Copy the Results found in the message
    p := TPAByte(@Msg);
    p := @p[SizeOf(TQUB_Socket_Message)];

    Chunk.Status := TQUB_Thread_Chunk_PStatus(p)^;
    p := @p[SizeOf(TQUB_Thread_Chunk_Status)];

    BasChunk.NodeCount := TPInteger(p)^;
    p := @p[SizeOf(Integer)];

    if BasChunk.NodeCount > 0 then begin

      try

        BasChunk.Nodes    := IAlloc1(BasChunk.NodeCount);
        BasChunk.NodesDBL := IAlloc1(BasChunk.NodeCount);

        Move(TPAInteger(p)[0 * BasChunk.NodeCount], BasChunk.Nodes[0],    BasChunk.NodeCount * SizeOf(Integer));
        Move(TPAInteger(p)[1 * BasChunk.NodeCount], BasChunk.NodesDBL[0], BasChunk.NodeCount * SizeOf(Integer));

      except

        BasChunk.NodeCount := 0;
        FreeMem(BasChunk.Nodes);
        FreeMem(BasChunk.NodesDBL);
        BasChunk.Nodes    := nil;
        BasChunk.NodesDBL := nil;

      end;

    end;

    prOnEvent(Self, Rec);

  end;

end;


//******************************************************************************
end.
