unit QUB_Thread_Socket_Data;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses

SysUtils,
AllocTypes, Alloc,

QUB_Global,
QUB_QVO_QVD_Global,
QUB_Socket,
QUB_Thread_Socket,
QUB_Thread, QUB_Thread_Data;


type


TQUB_Thread_Socket_ParamState = (PARAMSTATE_UNINITIALIZED,
                                 PARAMSTATE_ACK_INITIALIZED, PARAMSTATE_INITIALIZED
                                );


TQUB_Thread_Socket_Data        = class(TQUB_Thread_Socket)
protected


  prNeedData                       :TPAInteger;
  prNeedDataSel                    :TQUB_FirstLast;

  prParamState                     :TQUB_Thread_Socket_ParamState;


  procedure prSend_Params;

  procedure prInit_Service;
  procedure prStart_Service;

  procedure prRead(var Msg :TQUB_Socket_Message); override;

  procedure prRequestData(var Msg :TQUB_Socket_Message);
  procedure prReportStatus(var Msg :TQUB_Socket_Message);
  procedure prReportTermination(var Msg :TQUB_Socket_Message);

  procedure prSetEvent_RequestData; virtual;


public


  destructor Destroy; override;

  procedure SetEvent; override;


  property NeedData                :TPAInteger                    read prNeedData;
  property NeedDataSel             :TQUB_FirstLast        read prNeedDataSel;


end;



implementation


//******************************************************************************
destructor TQUB_Thread_Socket_Data.Destroy;
begin

  try

    FreeMem(prNeedData);

  finally

    inherited;

  end;

end;



//****************************************************************************//
//                                                                            //
//                                   R E A D                                  //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Data.prRead(var Msg :TQUB_Socket_Message);
begin

  inherited prRead(Msg);

  if prState = STATE_CONNECTED then begin
  //If not STATE_CONNECTED then the processing is done in inherited Read

    case prThreadState of

      THREADSTATE_AVAILABLE:
      //The Thread is available, now initialize it
        begin

          prThreadState := THREADSTATE_ACK_INITIALIZED;
          prInit_Service;

        end;

      THREADSTATE_ACK_INITIALIZED:
      //The Thread is waiting to be initialized, now pass parameters
        begin

          if Msg.Message = QUB_SOCKET_MESSAGE_ACK then begin
            if TPLongBool(@Msg.Data)^ then begin
              prThreadState := THREADSTATE_INITIALIZED;
              prParamState  := PARAMSTATE_ACK_INITIALIZED;
              prSend_Params;
            end else begin
              //Error;
            end;
          end else begin
            //Error;
          end;

        end;

      THREADSTATE_INITIALIZED:
      //The Thread is waiting for Class parameters
        begin

          case prParamState of

            PARAMSTATE_ACK_INITIALIZED:
              begin

                if Msg.Message = QUB_SOCKET_MESSAGE_ACK then begin
                  if TPLongBool(@Msg.Data)^ then begin
                    prParamState  := PARAMSTATE_INITIALIZED;
                    prThreadState := THREADSTATE_ACK_STARTED;
                    prStart_Service;
                  end else begin
                    //Error;
                  end;
                end else begin
                  //Error;
                end;

              end;

          end;//case ParamState

        end;

      THREADSTATE_ACK_STARTED:
        begin

          if Msg.Message = QUB_SOCKET_MESSAGE_ACK then begin
            if TPLongBool(@Msg.Data)^ then begin
              prThreadState := THREADSTATE_STARTED;
            end else begin
              //Error;
            end;
          end else begin
            //Error;
          end;

        end;

      THREADSTATE_STARTED:
        begin

          if Msg.Message > QUB_SOCKET_MESSAGE_USER_DEFINED then begin
          //User defined messages

            case Msg.Message of

              MESSAGE_REQUEST_DATA:
              //The client at the other end of the connection needs data
                begin

                  prRequestData(Msg);

                end;

            end;//case Msg.Message

          end else begin
          //System messages

            case Msg.Message of

              QUB_SOCKET_MESSAGE_SERVICE_STATUS:
              //Service reports Status
                begin

                  prReportStatus(Msg);

                end;

              QUB_SOCKET_MESSAGE_SERVICE_TERMINATED:
              //Service has terminated
                begin

                  prThreadState := THREADSTATE_TERMINATED;
                  prReportTermination(Msg);

                end;

            end;//case Msg.Message

          end;

        end;

    end;//case ThreadState

  end;//if prState = STATE_CONNECTED

end;


//****************************************************************************//
//                                                                            //
//                             S E N D     P A R A M S                        //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Data.prSend_Params;
begin
end;


//****************************************************************************//
//                                                                            //
//                               S E R V I C E                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Data.prInit_Service;
var
  Msg :TQUB_Socket_Message;
begin

  Msg.Message := QUB_SOCKET_MESSAGE_INIT_SERVICE;
  Msg.Size    := SizeOf(TQUB_Socket_Message);

  prWrite(@Msg, SizeOf(Msg));

end;


procedure TQUB_Thread_Socket_Data.prStart_Service;
var
  Msg :TQUB_Socket_Message;
begin

  Msg.Message := QUB_SOCKET_MESSAGE_START_SERVICE;
  Msg.Size    := SizeOf(TQUB_Socket_Message);

  prWrite(@Msg, SizeOf(Msg));

end;


//****************************************************************************//
//                                                                            //
//                           R E Q U E S T     D A T A                        //
//                                                                            //
//****************************************************************************//
//This assumes the data is in one channel format;
//If multiple data channel, it must be overriden;
procedure TQUB_Thread_Socket_Data.prRequestData(var Msg :TQUB_Socket_Message);
var
  Rec :TQUB_Thread_Event_PRequestData;
begin

  if Assigned(prOnEvent) then begin

    prLastBlockMsg := Msg;

    prNeedDataSel := TQUB_PFirstLast(@Msg.Data)^;

    FreeMem(prNeedData);
    prNeedData := IAlloc1(prNeedDataSel.Last - prNeedDataSel.First + 1);

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_RequestData));

    //Set the Event
    Rec.Rec.Thread := TQUB_Thread(Self);
    Rec.Rec.Event  := THREAD_EVENT_REQUEST_DATA;
    Rec.DataSel    := prNeedDataSel;
    Rec.Data       := prNeedData;

    prOnEvent(Self, TQUB_Thread_Event_PRec(Rec));

  end;

end;


//****************************************************************************//
//                                                                            //
//                         R E P O R T    S T A T U S                         //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Data.prReportStatus(var Msg :TQUB_Socket_Message);
var
  Rec :TQUB_Thread_Event_PReportStatus;
begin

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_ReportStatus));

    //Set the Event
    Rec.Rec.Thread := TQUB_Thread(Self);
    Rec.Rec.Event  := THREAD_EVENT_REPORT_STATUS;
    Rec.Completion := TPInteger(@Msg.Data)^;

    Chunk.Completion := Rec.Completion;

    prOnEvent(Self, TQUB_Thread_Event_PRec(Rec));

  end;

end;


//****************************************************************************//
//                                                                            //
//                     R E P O R T    T E R M I N A T I O N                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Data.prReportTermination(var Msg :TQUB_Socket_Message);
begin
end;


//****************************************************************************//
//                                                                            //
//                            R E S E T   E V E N T                           //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Data.SetEvent;
begin

  try

    case prLastBlockMsg.Message of

      MESSAGE_REQUEST_DATA:
        begin

          prSetEvent_RequestData;

        end;

    end;//case LastBlockMsg.Message

  finally

    prLastBlockMsg.Message := QUB_SOCKET_MESSAGE_NULL;

  end;

end;

procedure TQUB_Thread_Socket_Data.prSetEvent_RequestData;
var
  PMsg :TQUB_Socket_PMessage;
  SMsg :Integer;
  p    :TPAByte;
begin

  PMsg := nil;

  try

    SMsg := SizeOf(TQUB_Socket_Message) + SizeOf(Integer) * (prNeedDataSel.Last - prNeedDataSel.First + 1);
    PMsg := AllocMem(SMsg);
    PMsg.Message := MESSAGE_RECV_DATA;
    PMsg.Size    := SMsg;
    p := TPAByte(PMsg);

    Move(prNeedData^, p[SizeOf(TQUB_Socket_Message)], SizeOf(Integer) * (prNeedDataSel.Last - prNeedDataSel.First + 1));

    prWrite(PMsg, SMsg);

  finally

    FreeMem(prNeedData);
    prNeedData := nil;

    FreeMem(PMsg);

  end;

end;

//******************************************************************************
end.
