unit QUB_Thread_Socket_Idealize;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses

SysUtils,
AllocTypes, Alloc,

QUB_QVO_QVD_Global,
QUB_Socket,
QUB_Thread_Socket, QUB_Thread_Socket_Data,
QUB_Thread, QUB_Thread_Data, QUB_Thread_Idealize,
QUB_Idealize_Global;


type


TQUB_Thread_Socket_Idealize  = class(TQUB_Thread_Socket_Data)
protected


  //Parameters
  prParams                   :TQUB_Idealize_Params;


  procedure prSend_Params;

  procedure prReportTermination(var Msg :TQUB_Socket_Message);


public


  constructor Create(const RemoteServer :String; Port :Word); override;


  property Params          :TQUB_Idealize_Params        read prParams           write prParams;


end;


implementation


//******************************************************************************
constructor TQUB_Thread_Socket_Idealize.Create(const RemoteServer :String; Port :Word);
begin

  inherited;

  prIdentification.Task   := QUB_SOCKET_ID_TASK_IDEALIZE;
  prIdentification.Thread := QUB_SOCKET_ID_THREAD_IDEALIZE;

end;


//****************************************************************************//
//                                                                            //
//                             S E N D     P A R A M S                        //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Idealize.prSend_Params;
var
  Msg :Pointer;
  s   :Integer;
  p   :TPAByte;
begin

  s := SizeOf(TQUB_Socket_Message) + SizeOf(TQUB_Thread_Idealize_Chunk) + SizeOf(prParams) +
       prParams.LibNameSize + prParams.ModelSize;

  Msg := AllocMem(s);

  try

    TQUB_Socket_PMessage(Msg).Message := MESSAGE_RECV_PARAMS;
    TQUB_Socket_PMessage(Msg).Size    := s;

    p := Msg;

    p := @p[SizeOf(TQUB_Socket_Message)];

    //Chunk
    Move(prChunk^,  p^, SizeOf(TQUB_Thread_Idealize_Chunk));

    p := @p[SizeOf(TQUB_Thread_Idealize_Chunk)];

    //Params - fixed
    Move(prParams,  p^, SizeOf(prParams));

    p := @p[SizeOf(prParams)];

    //Params - variable
    Move(prParams,  p^, prParams.LibNameSize);
    p := @p[prParams.LibNameSize];
    Move(prParams,  p^, prParams.ModelSize);

    prWrite(Msg, s);

  finally

    FreeMem(Msg);

  end;

end;


{ TODO -oLorin : Include the code for Histogram and the other new junk}
//****************************************************************************//
//                                                                            //
//                     R E P O R T    T E R M I N A T I O N                   //
//                                                                            //
//****************************************************************************//
procedure TQUB_Thread_Socket_Idealize.prReportTermination(var Msg :TQUB_Socket_Message);
var
  Rec      :TQUB_Thread_Event_PRec;
  p        :TPAByte;
  IdlChunk :TQUB_Thread_Idealize_PChunk;
begin

  IdlChunk := TQUB_Thread_Idealize_PChunk(Chunk);

  if Assigned(prOnEvent) then begin

    //Allocate Event
    Rec := AllocMem(SizeOf(TQUB_Thread_Event_Rec));

    //Set the Event
    Rec.Thread := TQUB_Thread(Self);
    Rec.Event  := THREAD_EVENT_REPORT_TERMINATION;

    //Copy the Results found in the message
    p := TPAByte(@Msg);
    p := @p[SizeOf(TQUB_Socket_Message)];

    Chunk.Status := TQUB_Thread_Chunk_PStatus(p)^;
    p := @p[SizeOf(TQUB_Thread_Chunk_Status)];

    IdlChunk.SegCount := TPInteger(p)^;
    p := @p[SizeOf(Integer)];

    if IdlChunk.SegCount > 0 then begin

      try

        IdlChunk.SegFirsts  := IAlloc1(IdlChunk.SegCount);
        IdlChunk.SegLasts   := IAlloc1(IdlChunk.SegCount);
        IdlChunk.SegClasses := IAlloc1(IdlChunk.SegCount);

        Move(TPAInteger(p)[0 * IdlChunk.SegCount], IdlChunk.SegFirsts[0],  IdlChunk.SegCount * SizeOf(Integer));
        Move(TPAInteger(p)[1 * IdlChunk.SegCount], IdlChunk.SegLasts[0],   IdlChunk.SegCount * SizeOf(Integer));
        Move(TPAInteger(p)[2 * IdlChunk.SegCount], IdlChunk.SegClasses[0], IdlChunk.SegCount * SizeOf(Integer));

      except

        IdlChunk.SegCount := 0;
        FreeMem(IdlChunk.SegFirsts);
        FreeMem(IdlChunk.SegLasts);
        FreeMem(IdlChunk.SegClasses);
        IdlChunk.SegFirsts  := nil;
        IdlChunk.SegLasts   := nil;
        IdlChunk.SegClasses := nil;

      end;

    end;

    p := @p[3 * IdlChunk.SegCount * SizeOf(Integer)];

    IdlChunk.ClassCount := TPInteger(p)^;
    p := @p[SizeOf(Integer)];

    if IdlChunk.ClassCount > 0 then begin

      try

        //Change type to single
        IdlChunk.Amps := SAlloc1(IdlChunk.ClassCount);

        Move(TPAInteger(p)[0 * IdlChunk.ClassCount], IdlChunk.Amps[0],  IdlChunk.ClassCount * SizeOf(Integer));

      except

        IdlChunk.ClassCount := 0;
        FreeMem(IdlChunk.Amps);
        IdlChunk.Amps := nil;

      end;

    end;

    p := @p[1 * IdlChunk.ClassCount * SizeOf(Integer)];


    //Make sure to copy it correctly, since it does not really include the 0 terminator !!!
    IdlChunk.ReportSize := TPInteger(p)^;
    p := @p[SizeOf(Integer)];

    if IdlChunk.ReportSize > 0 then begin

      try

        IdlChunk.Report := AllocMem(IdlChunk.ReportSize);

        Move(p^, IdlChunk.Report^,  IdlChunk.ReportSize);

      except

        IdlChunk.ReportSize := 0;
        FreeMem(IdlChunk.Report);
        IdlChunk.Report := nil;

      end;

    end;

    prOnEvent(Self, Rec);

  end;

end;


//******************************************************************************
end.
