unit QUB_UserInterfaceProperties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ComCtrls,

RegistryInterface,
QUB_Interface;


type


TUserInterfaceProperties = class(TForm)

  edtText1: TEdit;
  edtName1: TEdit;
  Label1: TLabel;
  Label2: TLabel;
  Label3: TLabel;
  Label4: TLabel;
  Label5: TLabel;
  Label6: TLabel;
  Label7: TLabel;
  Label8: TLabel;
  Label9: TLabel;
  Label10: TLabel;
  edtText2: TEdit;
  edtName2: TEdit;
  edtText3: TEdit;
  edtName3: TEdit;
  edtText4: TEdit;
  edtName4: TEdit;
  edtText5: TEdit;
  edtName5: TEdit;
  edtText6: TEdit;
  edtName6: TEdit;
  edtText7: TEdit;
  edtName7: TEdit;
  edtText8: TEdit;
  edtName8: TEdit;
  btnOK: TButton;
  btnCancel: TButton;
  updnInterfaceCount: TUpDown;
  Label11: TLabel;
  edtInterfaceCount: TEdit;
  Label12: TLabel;


  procedure FormCreate(Sender: TObject);
  procedure FormKeyPress(Sender: TObject; var Key: Char);
  procedure updnInterfaceCountClick(Sender: TObject; Button: TUDBtnType);
  procedure btnOKClick(Sender: TObject);
  procedure btnCancelClick(Sender: TObject);
  procedure btnHelpClick(Sender: TObject);


private


  aText : Array[0..INTERFACE_COUNTUSER-1] of TEdit;   // Reference to edit controls as array
  aName : Array[0..INTERFACE_COUNTUSER-1] of TEdit;


  function InterfaceCount :Integer;
  procedure SetMode;


end;


implementation


uses


QUB_Global,
QUB_Main;


{$R *.DFM}


//----- Get Number of interfaces
function TUserInterfaceProperties.InterfaceCount :Integer;
begin
  Result := StrToInt(edtInterfaceCount.Text);
end;

//----- Set which control are enabled based on count of interfaces
procedure TUserInterfaceProperties.SetMode;
var
  iInterfaces, i :Integer;
begin
  iInterfaces := InterfaceCount();
  for i := 0 to INTERFACE_COUNTUSER-1 do begin
    aText[i].Visible := ( iInterfaces > i );
    aName[i].Visible := ( iInterfaces > i );
  end;
end;

//----- Up / Down control change to # of interfaces
procedure TUserInterfaceProperties.updnInterfaceCountClick(Sender: TObject; Button: TUDBtnType);
begin
  if Button = btPrev Then Begin
    if InterfaceCount > 0 Then Begin
      edtInterfaceCount.Text := IntToStr( InterfaceCount()-1 );
      SetMode;
    end;
  end else begin
    if InterfaceCount < INTERFACE_COUNTUSER Then Begin
      edtInterfaceCount.Text := IntToStr( InterfaceCount()+1 );
      SetMode;
    end;
  end;
end;

//----- Click on OK - update interface list, registry names, current interface
procedure TUserInterfaceProperties.btnOKClick(Sender: TObject);
var
  i,iInterfaces:Integer;
  oInterface : TQUB_Interface;
begin
  //-----Save Changes
  iInterfaces := InterfaceCount;
  UserInterfaceCountWriteRegistry(iInterfaces);

  for i := 0 to INTERFACE_COUNTUSER-1 do Begin
    //----- Update registry
    UserInterfaceWriteRegistry(i, aText[i].Text, aName[i].Text );
    //----- Update buttons
    oInterface := QUB_MainForm.InterfaceList.InterfaceObject[INTERFACE_USER+i];
    oInterface.InterfaceButton.Caption := aText[i].Text;
    oInterface.InterfaceButton.Hint    := aName[i].Text;
    oInterface.InterfaceButton.Visible := (i < iInterfaces);
  end;

  //----- ensure current interface is visible
  if QUB_MainForm.InterfaceIndex >= INTERFACE_USER+iInterfaces Then
    QUB_MainForm.InterfaceIndex := 0;

  ModalResult := mrOK;
end;

//----- Cancel - nothing to do
procedure TUserInterfaceProperties.btnCancelClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TUserInterfaceProperties.btnHelpClick(Sender: TObject);
begin
  ShowWiki('Interface_Theory#Windows_and_Layouts', Handle);
end;

//----- Load from qubInterfaceList
procedure TUserInterfaceProperties.FormCreate(Sender: TObject);
var
  i : Integer;
  oInterface : TQUB_Interface;
begin
  edtInterfaceCount.Text := '0';

  aText[0] := edtText1;
  aText[1] := edtText2;
  aText[2] := edtText3;
  aText[3] := edtText4;
  aText[4] := edtText5;
  aText[5] := edtText6;
  aText[6] := edtText7;
  aText[7] := edtText8;

  aName[0] := edtName1;
  aName[1] := edtName2;
  aName[2] := edtName3;
  aName[3] := edtName4;
  aName[4] := edtName5;
  aName[5] := edtName6;
  aName[6] := edtName7;
  aName[7] := edtName8;

  for i := 0 to INTERFACE_COUNTUSER-1 do Begin
    oInterface := QUB_MainForm.InterfaceList.InterfaceObject[INTERFACE_USER+i];
    aText[i].Text := oInterface.InterfaceButton.Caption;
    aName[i].Text := oInterface.InterfaceButton.Hint;
    if oInterface.InterfaceButton.Visible Then
      edtInterfaceCount.Text := IntToStr(i+1);
  end;

  SetMode;
end;

procedure TUserInterfaceProperties.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Word(Key) = VK_ESCAPE then begin
    Key := #0;
    ModalResult := mrCancel;
  end;
end;


//******************************************************************************
end.
