unit QUB_Viewport;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}



// Make sure Position is poDesigned !!!




interface


uses


Windows, SysUtils, Classes, Controls, ExtCtrls, Forms, Registry,

RegistryInterface,

QUB_Events;


type


TQUB_Viewport                                   = class(TForm)


  Conteiner                                     :TPanel;


  procedure SplitterMoved(Sender: TObject);
  procedure OnConstrainedResize(Sender: TObject; var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);


protected


  prActive                                      :Boolean;		// Remember last Activate status to determine whether to RaiseNotifyEvent for Activate / Deactivate
  prUnDocked                                    :Boolean;		// Docking status, Used by Resizing and DoropertiesDisplay
  prUnDocking                                   :Boolean;		// Unused ? {JB}

  prName                                        :String;

  prInterfaceName                               :String;

  prFloatBounds                                 :TRect;
  prDockBounds                                  :TRect;
  prUnDockBounds                                :TRect;

  prEvents                                      :TQUB_Events;


  procedure prSetName(const Name :String);
  procedure prSetInterfaceName(const Name :String);

  procedure Resize; override;
  procedure Resizing(State: TWindowState); override;

  //Event Handlers
  procedure DoOnActivate;
  procedure DoOnDeActivate;

  //Properties / Registry
  procedure DoProperties(Read :Boolean); virtual;
  //  read on create, write on destroy
  //  put customizations like colors here

  procedure DoPropertiesInternal(Read :Boolean); virtual;
  //  read/write on resize or whenever windows are rearranged
	//  strictly for window layout info expressed as percentages so resize behaves sensibly


public


  constructor Create(AOwner :TComponent); override;
  constructor CreateName(AOwner :TComponent; const Name :String); virtual;
  destructor Destroy; override;


  procedure ConstrainedResize(var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer); override;

  procedure DoPropertiesDisplay(Read :Boolean; Float :Boolean);

  procedure Activate; override;
  procedure DeActivate; override;

  function CanClose :Boolean; virtual;
  //  called from MainForm.FormCloseQuery (help: OnCloseQuery) when app wants to quit
  //  you should prompt for each modified file [save, discard, cancel]

  function CloseAll :Boolean; virtual;

  procedure ShowHide(Show :Boolean);


  property Name          :String                     read prName                write prSetName;

  property Events        :TQUB_Events                read prEvents;
  //  OnActivate -- on Viewport.Activate
  //    subscribed to by QUB_Main
  //  OnDeActivate
  //    subscribed to by QUB_Main, but handler is empty

  property InterfaceName :String                     read prInterfaceName       write prSetInterfaceName;

  property Active        :Boolean                    read prActive;

  property UnDocked      :Boolean                    read prUnDocked            write prUnDocked;
  property UnDocking     :Boolean                    read prUnDocking           write prUnDocking;

  property FloatBounds   :TRect                      read prFloatBounds         write prFloatBounds;
  property DockBounds    :TRect                      read prDockBounds          write prDockBounds;
  property UnDockBounds  :TRect                      read prUnDockBounds        write prUnDockBounds;


end;


TQUB_Viewport_List = class(TStringList)
protected


  function prGetViewport(Index :Integer) :TQUB_Viewport;
  function prGetViewportByName(Index :String) :TQUB_Viewport;


public


  destructor Destroy; override;


  procedure AddViewport(Viewport :TQUB_Viewport);


  property Viewport[Index :Integer]      :TQUB_Viewport read prGetViewport; default;
  property ViewportByName[Index :String] :TQUB_Viewport read prGetViewportByName;


end;


implementation


{$R *.DFM}


//******************************************************************************
constructor TQUB_Viewport.CreateName(AOwner :TComponent; const Name :String);
begin

  Create(AOwner);//NOT the inherited!!! LM.

  Self.Name := Name;

  DoProperties(True{Read});

end;

constructor TQUB_Viewport.Create(AOwner :TComponent);
begin

  inherited Create(AOwner);

  prActive       := False;
  prUnDocked     := False;

  prFloatBounds  := BoundsRect;
  prDockBounds   := BoundsRect;
  prUnDockBounds := BoundsRect;

  UnDockWidth    := prUnDockBounds.Right  - prUnDockBounds.Left;
  UnDockHeight   := prUnDockBounds.Bottom - prUnDockBounds.Top;

  prEvents       := TQUB_Events.Create;

end;

destructor TQUB_Viewport.Destroy;
begin
  try

    DoProperties(False{Read});

    prEvents.Free;

  finally

    inherited;

  end;
end;

//Activate/DeActivate
procedure TQUB_Viewport.Activate;
begin
  try inherited except end;
  if not prActive then begin
    prActive := True;
    //if prTaskSelector <> nil then prTaskSelector.Activate;
    DoOnActivate;
  end;
end;

procedure TQUB_Viewport.DeActivate;
begin
  try inherited except end;
  if prActive then begin
    prActive := False;
    ///if prTaskSelector <> nil then prTaskSelector.DeActivate;
    DoOnDeActivate;
  end;
end;


//CanClose
function TQUB_Viewport.CanClose :Boolean;
begin
  Result := True;
end;

function TQUB_Viewport.CloseAll :Boolean;
begin
  Result := False;
end;


//Events
procedure TQUB_Viewport.DoOnActivate;
begin
  if Events <> nil then Events.RaiseNotifyEvent('OnActivate', Self);
end;

procedure TQUB_Viewport.DoOnDeActivate;
begin
  if Events <> nil then Events.RaiseNotifyEvent('OnDeActivate', Self);
end;

//******************************************************************************
procedure TQUB_Viewport.SplitterMoved(Sender: TObject);
begin
  DoPropertiesInternal(False{Read});
end;

procedure TQUB_Viewport.ConstrainedResize(var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);
begin
  inherited;
  MinWidth  := 10;
  MinHeight := 10;
  MaxWidth  := 2 * Screen.Width;
  MaxHeight := 2 * Screen.Height;
end;

procedure TQUB_Viewport.OnConstrainedResize(Sender: TObject; var MinWidth, MinHeight, MaxWidth, MaxHeight: Integer);
begin
  MinWidth  := 10;
  MinHeight := 10;
  MaxWidth  := 2 * Screen.Width;
  MaxHeight := 2 * Screen.Height;
end;

procedure TQUB_Viewport.Resizing(State: TWindowState);
begin
  try
    if (State = wsNormal) then begin
      if (Floating) and (not UnDocked) then begin
        FloatBounds  := BoundsRect;
      end else if (Floating) and (UnDocked) then begin
      	UnDockBounds := BoundsRect;
      end;
    end;
  finally
    inherited;
  end;
  UnDockWidth  := UnDockBounds.Right  - UnDockBounds.Left;
  UnDockHeight := UnDockBounds.Bottom - UnDockBounds.Top;
end;

procedure TQUB_Viewport.Resize;
begin
  inherited;
  if (Name <> '') and (InterfaceName <> '') then begin
    if (WindowState <> wsMinimized) then begin
      DoPropertiesInternal(True{Read});
    end;
  end;
end;

procedure TQUB_Viewport.ShowHide(Show :Boolean);
begin
  Visible := Show;
  if Show then begin
    Activate;
  end else begin
    DeActivate;
  end;
end;


//Properties / Registry ********************************************************

//****************************************************************************//
//                                                                            //
//                  S P E C I F I C    P R O P E R T I E S                    //
//                                                                            //
//****************************************************************************//
//Store/Restore the specific properties, at OnCreate, and OnDestroy
procedure TQUB_Viewport.DoProperties(Read :Boolean);
begin
  //override if necessary
end;


//****************************************************************************//
//                                                                            //
//                  I N T E R N A L    P R O P E R T I E S                    //
//                                                                            //
//****************************************************************************//
//Store/Restore the internal properties (proportions)
procedure TQUB_Viewport.DoPropertiesInternal(Read :Boolean);
begin
  //override if necessary
end;


//****************************************************************************//
//                                                                            //
//                   D I S P L A Y    P R O P E R T I E S                     //
//                                                                            //
//****************************************************************************//
//Store/Restore Display properties
procedure TQUB_Viewport.DoPropertiesDisplay(Read :Boolean; Float :Boolean);
var
  Reg        :TRegistry;
  v, u       :Boolean;
  fb, db, ub :TRect;
  w, h       :Integer;
  s          :TWindowState;
begin

  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Interface ' + InterfaceName + '\Viewport ' + Name, True) then begin
        
        w := Screen.Width;
        h := Screen.Height;

        if Read then begin
        //Read

          //Visible
          if Float then begin
            v  := ReadRegistryBoolean(Reg, 'Float Visible', False);
          end else begin
            v  := ReadRegistryBoolean(Reg, 'Dock Visible', False);
          end;

          //UnDocked
          if Float then begin
            u := False;
          end else begin
            u := ReadRegistryBoolean(Reg, 'UnDocked', False);
          end;

          //WindowState
          if Float then begin
            s  := TWindowState(ReadRegistryInteger(Reg, 'Float Window State', Integer(wsNormal)));
          end else begin
            s  := TWindowState(ReadRegistryInteger(Reg, 'Dock Window State', Integer(wsNormal)));
          end;

          //Bounds
          if ValueExists('Float Bounds')         then ReadBinaryData('Float Bounds',  fb, SizeOf(fb))       else fb := Bounds(w div 4, h div 4, w div 2, h div 2);
          if ValueExists('Dock Bounds')          then ReadBinaryData('Dock Bounds',   db, SizeOf(db))       else db := fb;
          if ValueExists('UnDock Bounds')        then ReadBinaryData('UnDock Bounds', ub, SizeOf(ub))       else ub := fb;

          FloatBounds   := fb;
          DockBounds    := db;
          UnDockBounds  := ub;
          UnDockWidth   := UnDockBounds.Right  - UnDockBounds.Left;
          UnDockHeight  := UnDockBounds.Bottom - UnDockBounds.Top;

          UnDocked      := u;
          Visible       := False;
          WindowState   := wsNormal;
          if Float then begin
            BoundsRect  := FloatBounds;
            WindowState := s;
          end else if UnDocked then begin
            BoundsRect  := UnDockBounds;
            WindowState := s;
          end else begin
            BoundsRect  := DockBounds;
          end;
          Visible       := v;

        end else begin
        //Write

          //Visible
          if Float then begin
            WriteBool('Float Visible', Visible);
          end else begin
            WriteBool('Dock Visible', Visible);
          end;

          //UnDocked
          if Float then begin

          end else begin
            WriteBool('UnDocked', UnDocked);
          end;

          //WindowState
          if Float then begin
            WriteInteger('Float Window State', Integer(WindowState));
          end else if UnDocked then begin
            WriteInteger('Dock Window State', Integer(WindowState));
          end;

          //Bounds
          if (Float) then begin
            if (WindowState = wsNormal) then begin
              FloatBounds := BoundsRect;
            end;
            WriteBinaryData('Float Bounds',  prFloatBounds,  SizeOf(prFloatBounds));
          end else if UnDocked then begin
            if (WindowState = wsNormal) then begin
              UnDockBounds := BoundsRect;
              UnDockWidth   := UnDockBounds.Right  - UnDockBounds.Left;
              UnDockHeight  := UnDockBounds.Bottom - UnDockBounds.Top;
            end;
            WriteBinaryData('UnDock Bounds', prUnDockBounds, SizeOf(prUnDockBounds));
          end else begin
            DockBounds := BoundsRect;
            WriteBinaryData('Dock Bounds',   prDockBounds,   SizeOf(prDockBounds));
          end;

        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;

  DoPropertiesInternal(Read);

end;

procedure TQUB_Viewport.prSetInterfaceName(const Name :String);
begin
  prInterfaceName := Name;
end;

procedure TQUB_Viewport.prSetName(const Name :String);
begin
  prName := Name;
  Caption := Name;
end;


//Viewport List ****************************************************************
destructor TQUB_Viewport_List.Destroy;
var
  i :Integer;
begin
  try
    for i := Count - 1 downto 0 do begin
      try
        Self[i].Free;
      except
        beep;
      end;
      Delete(i);
    end;
  finally
    inherited;
  end;
end;

procedure TQUB_Viewport_List.AddViewport(Viewport :TQUB_Viewport);
begin
  AddObject(Viewport.Name, Viewport);
end;

function TQUB_Viewport_List.prGetViewport(Index :Integer) :TQUB_Viewport;
begin
  if (Index >= 0) and (Index < Count) then Result := TQUB_Viewport(Objects[Index]) else Result := nil;
end;

function TQUB_Viewport_List.prGetViewportByName(Index :String) :TQUB_Viewport;
var
  Idx :Integer;
begin
  Idx := IndexOf(Index);
  if Idx = -1 then Result := nil else Result := Self[Idx];
end;


//******************************************************************************
end.

