unit QUB_Viewport_AcquisitionInput;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, ComCtrls, Menus, Buttons,
Forms, Dialogs, StdCtrls, ExtCtrls, Registry,
Math,

RegistryInterface,

QUB_Viewport,
QUB_Viewport_Data,
QUB_QVO_QVD, QUB_QVO_QVS, QUB_QVO_QVSpectrum,
QUB_AcquisitionInput,
QUB_QVO_QVS_Properties, QUB_QVO_QVSpectrum_Properties;


type


TQUB_Viewport_AcquisitionInput = class(TQUB_Viewport)


  splScopeInput               :TSplitter;
  ConteinerInput              :TPanel;
  lblStatus                   :TLabel;
  ConteinerLeft               :TPanel;
  ConteinerScope              :TPanel;
  panGain                     :TPanel;
  lblGainUnits                :TLabel;
  lblGain                     :TLabel;
  lblChannel                  :TLabel;
  lblChannelIndex             :TLabel;
  btnFindTrace                :TSpeedButton;
  lblOffsetUnits              :TLabel;
  lblOffset                   :TLabel;
  udChannelIndex              :TUpDown;
  panGainButtons              :TPanel;
  btnGain1                    :TSpeedButton;
  btnGain50                   :TSpeedButton;
  btnGain100                  :TSpeedButton;
  btnGain2                    :TSpeedButton;
  btnGain5                    :TSpeedButton;
  btnGain200                  :TSpeedButton;
  btnGain500                  :TSpeedButton;
  btnGain10                   :TSpeedButton;
  btnGain1000                 :TSpeedButton;
  btnGain01                   :TSpeedButton;
  btnGain02                   :TSpeedButton;
  btnGain20                   :TSpeedButton;
  btnGain05                   :TSpeedButton;
  btnGain2000                 :TSpeedButton;
  panTime                     :TPanel;
  lblTime                     :TLabel;
  lblTimeUnits                :TLabel;
  btnChartMode                :TSpeedButton;
  panTimeButtons              :TPanel;
  btnTime1                    :TSpeedButton;
  btnTime2                    :TSpeedButton;
  btnTime5                    :TSpeedButton;
  btnTime10                   :TSpeedButton;
  btnTime20                   :TSpeedButton;
  btnTime50                   :TSpeedButton;
  btnTime100                  :TSpeedButton;
  btnTime200                  :TSpeedButton;
  btnTime500                  :TSpeedButton;
  btnTime1000                 :TSpeedButton;
  ConteinerSpectrum           :TPanel;
  panMode                     :TPanel;
  btnMode                     :TSpeedButton;
  panSpectrumGain             :TPanel;
  lblSpectrumDecades          :TLabel;
  lblSpectrumChannel          :TLabel;
  lblSpectrumChannelIndex     :TLabel;
  udSpectrumChannelIndex      :TUpDown;
  lblSpectrumDecadeCount      :TLabel;
  udSpectrumDecadeCount       :TUpDown;
  lblSpectrumOffset           :TLabel;
  lblSpectrumDecadeOffset     :TLabel;
  udSpectrumDecadeOffset      :TUpDown;
  Label1                      :TLabel;
  lblSpectrumTime             :TLabel;
  udSpectrumTime              :TUpDown;
  lblSpectrumBinCount         :TLabel;
  txtSpectrumBinCount         :TEdit;
  udSpectrumBinCount          :TUpDown;
  chkSpectrumAcc              :TCheckBox;
  lblSpectrumAcc              :TLabel;
  lblSpectrumAccUpdate        :TLabel;
  chkSpectrumAccUpdate        :TCheckBox;
  lblSpectrumAccPercent       :TLabel;
  txtSpectrumAccPercent       :TEdit;


  procedure FormShow(Sender: TObject);

  procedure btnModeClick(Sender: TObject);

  procedure btnGainClick(Sender: TObject);
  procedure btnTimeClick(Sender: TObject);
  procedure udChannelIndexClick(Sender: TObject; Button: TUDBtnType);
  procedure btnFindTraceClick(Sender: TObject);
  procedure btnChartModeClick(Sender: TObject);

  procedure udSpectrumChannelIndexClick(Sender: TObject; Button: TUDBtnType);
  procedure udSpectrumDecadeCountClick(Sender: TObject; Button: TUDBtnType);
  procedure udSpectrumDecadeOffsetClick(Sender: TObject; Button: TUDBtnType);
  procedure udSpectrumTimeClick(Sender: TObject; Button: TUDBtnType);
  procedure udSpectrumBinCountClick(Sender: TObject; Button: TUDBtnType);
  procedure txtSpectrumBinCountKeyPress(Sender: TObject; var Key: Char);
  procedure txtSpectrumBinCountExit(Sender: TObject);
  procedure chkSpectrumAccClick(Sender: TObject);
  procedure chkSpectrumAccUpdateClick(Sender: TObject);
  procedure txtSpectrumAccPercentKeyPress(Sender: TObject; var Key: Char);
  procedure txtSpectrumAccPercentExit(Sender: TObject);


protected


  prMode                      :Boolean;//Scope/Spectrum

  prScope                     :TQUB_QVO_QVS;
  prSpectrum                  :TQUB_QVO_QVSpectrum;

  prAcquisitionInput          :TQUB_Acquisition_Input;

  prScopePropertiesForm       :TQUB_QVO_QVS_PropertiesForm;
  prSpectrumPropertiesForm    :TQUB_QVO_QVSpectrum_PropertiesForm;

  prScopeChannelIndex         :Integer;
  prSpectrumChannelIndex      :Integer;


  procedure DoProperties(Read :Boolean); override;
  procedure DoPropertiesInternal(Read :Boolean); override;

  //Scope
  procedure prSetScopeChannelIndex(c :Integer);
  function prGetScopeChannelIndex :Integer;
  procedure prSetScopeGain(Index :Integer; g :Double);
  function prGetScopeGain(Index :Integer) :Double;
  procedure prSetScopeTime(t :Double);
  function prGetScopeTime :Double;
  procedure prSetScopeChartMode(c :Boolean);
  function prGetScopeChartMode :Boolean;

  //Spectrum
  procedure prSetSpectrumChannelIndex(c :Integer);
  function prGetSpectrumChannelIndex :Integer;
  procedure prSetSpectrumDecadeCount(Index :Integer; c :Integer);
  function prGetSpectrumDecadeCount(Index :Integer) :Integer;
  procedure prSetSpectrumDecadeOffset(Index :Integer; o :Integer);
  function prGetSpectrumDecadeOffset(Index :Integer) :Integer;
  procedure prSetSpectrumTime(t :Integer);
  function prGetSpectrumTime :Integer;
  procedure prSetSpectrumBinCount(c :Integer);
  function prGetSpectrumBinCount :Integer;
  procedure prSetSpectrumAcc(a :Boolean);
  function prGetSpectrumAcc :Boolean;
  procedure prSetSpectrumAccUpdate(u :Boolean);
  function prGetSpectrumAccUpdate :Boolean;
  procedure prSetSpectrumAccPercent(p :Double);
  function prGetSpectrumAccPercent :Double;

  //Mode
  procedure prSetMode(m :Boolean);

  //Scope
  property ScopeChannelIndex                    :Integer  read prGetScopeChannelIndex      write prSetScopeChannelIndex;
  property ScopeTime                            :Double   read prGetScopeTime              write prSetScopeTime;
  property ScopeGain[Index :Integer]            :Double   read prGetScopeGain              write prSetScopeGain;
  property ScopeChartMode                       :Boolean  read prGetScopeChartMode         write prSetScopeChartMode;

  //Spectrum
  property SpectrumChannelIndex                 :Integer  read prGetSpectrumChannelIndex   write prSetSpectrumChannelIndex;
  property SpectrumTime                         :Integer  read prGetSpectrumTime           write prSetSpectrumTime;
  property SpectrumDecadeCount[Index :Integer]  :Integer  read prGetSpectrumDecadeCount    write prSetSpectrumDecadeCount;
  property SpectrumDecadeOffset[Index :Integer] :Integer  read prGetSpectrumDecadeOffset   write prSetSpectrumDecadeOffset;
  property SpectrumBinCount                     :Integer  read prGetSpectrumBinCount       write prSetSpectrumBinCount;
  property SpectrumAcc                          :Boolean  read prGetSpectrumAcc            write prSetSpectrumAcc;
  property SpectrumAccUpdate                    :Boolean  read prGetSpectrumAccUpdate      write prSetSpectrumAccUpdate;
  property SpectrumAccPercent                   :Double   read prGetSpectrumAccPercent     write prSetSpectrumAccPercent;


public


  constructor CreateName(AOwner :TComponent; const Name :String); override;


  function CanClose :Boolean; override;

  function NewInputAcq(const FileName :String) :Boolean;
  function OpenInputAcq(const FileName :String) :Boolean;
  function AssignInputAcq(Data :TQUB_QVO_QVD) :Boolean;

  procedure DoOnCloseData(Sender :TObject);
  procedure DoOnChangeHardware(Sender :TObject);
  procedure DoOnChangeChannels(Sender :TObject);
  procedure DoOnScopeChannelChange(Sender :TObject);
  procedure DoOnSpectrumChannelChange(Sender :TObject);


  property AcquisitionInput   :TQUB_Acquisition_Input   read prAcquisitionInput;

  property Mode               :Boolean                  read prMode               write prSetMode;


end;


implementation


uses


QUB_GUIInterface;


{$R *.DFM}


//******************************************************************************
constructor TQUB_Viewport_AcquisitionInput.CreateName(AOwner :TComponent; const Name :String);
begin
  inherited CreateName(AOwner, Name);

  prScope                         := TQUB_QVO_QVS.Create(Self);
  prScope.Parent                  := ConteinerScope;
  prSpectrum                      := TQUB_QVO_QVSpectrum.Create(Self);
  prSpectrum.Parent               := ConteinerSpectrum;

  prAcquisitionInput              := TQUB_Acquisition_Input.Create(Self);
  prAcquisitionInput.Parent       := ConteinerInput;
  prAcquisitionInput.OnConstrainedResize := OnConstrainedResize;
  prAcquisitionInput.Scope        := prScope;
  prAcquisitionInput.Spectrum     := prSpectrum;
  prAcquisitionInput.Status       := lblStatus;

  prScopePropertiesForm           := TQUB_QVO_QVS_PropertiesForm.Create(Self);
  prScope.PropertiesForm          := prScopePropertiesForm;
  prScope.OnChannelChange         := DoOnScopeChannelChange;

  prSpectrumPropertiesForm        := TQUB_QVO_QVSpectrum_PropertiesForm.Create(Self);
  prSpectrum.PropertiesForm       := prSpectrumPropertiesForm;
  prSpectrum.OnChannelChange      := DoOnSpectrumChannelChange;

  prScopeChannelIndex             := prScope.ChannelIndex;
  prSpectrumChannelIndex          := prSpectrum.ChannelIndex;

  AssignInputAcq(nil);
                    
end;


//******************************************************************************

//Scope
procedure TQUB_Viewport_AcquisitionInput.prSetScopeChannelIndex(c :Integer);
begin
  prScopeChannelIndex          := c;
  //Let's go back to zero-based channels - that's how they are called by DAQ boards anyway - endless source of confusion. LM
  lblChannelIndex.Caption      := IntToStr(c);
  udChannelIndex.Position      := c;
  ScopeGain[ScopeChannelIndex] := ScopeGain[ScopeChannelIndex];
  lblOffset.Caption            := Format('%.6g', [prScope.Offset[ScopeChannelIndex] * 1000 * prScope.Scaling[ScopeChannelIndex]]);
  prScope.ChannelIndex         := c;
end;

function TQUB_Viewport_AcquisitionInput.prGetScopeChannelIndex :Integer;
begin
  Result := prScopeChannelIndex;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetScopeGain(Index :Integer; g :Double);
var
  i :Integer;
begin
  prScope.Gain[Index] := g;
  for i := 0 to panGainButtons.ControlCount - 1 do begin
    if TSpeedButton(panGainButtons.Controls[i]).Tag =  Round(prScope.Gain[Index] * 1000000) then begin
      TSpeedButton(panGainButtons.Controls[i]).Down := True;
      Break;
    end;
  end;
end;

function TQUB_Viewport_AcquisitionInput.prGetScopeGain(Index :Integer) :Double;
begin
  Result := prScope.Gain[Index];
end;

procedure TQUB_Viewport_AcquisitionInput.prSetScopeTime(t :Double);
var
  i :Integer;
begin
  prScope.XResolution := t / 1000 * 10;
  for i := 0 to panTimeButtons.ControlCount - 1 do begin
    if TSpeedButton(panTimeButtons.Controls[i]).Tag =  Round(prScope.XResolution * 1000 / 10) then begin
      TSpeedButton(panTimeButtons.Controls[i]).Down := True;
      Break;
    end;
  end;
end;

function TQUB_Viewport_AcquisitionInput.prGetScopeTime :Double;
begin
  Result := prScope.XResolution * 1000 / 10;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetScopeChartMode(c :Boolean);
begin
  prScope.ChartMode := c;
  btnChartMode.Down := c;
end;

function TQUB_Viewport_AcquisitionInput.prGetScopeChartMode :Boolean;
begin
  Result := prScope.ChartMode;
end;


//Spectrum
procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumChannelIndex(c :Integer);
begin
  prSpectrumChannelIndex                     := c;
  lblSpectrumChannelIndex.Caption            := IntToStr(c);
  udSpectrumChannelIndex.Position            := c;
  SpectrumDecadeCount[SpectrumChannelIndex]  := SpectrumDecadeCount[SpectrumChannelIndex];
  SpectrumDecadeOffset[SpectrumChannelIndex] := SpectrumDecadeOffset[SpectrumChannelIndex];
  SpectrumTime                               := SpectrumTime;
  SpectrumBinCount                           := SpectrumBinCount;
  SpectrumAcc                                := SpectrumAcc;
  prSpectrum.ChannelIndex                    := c;
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumChannelIndex :Integer;
begin
  Result := prSpectrumChannelIndex;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumDecadeCount(Index :Integer; c :Integer);
begin
  prSpectrum.DecadeCount[Index]  := c;
  udSpectrumDecadeCount.Position := c;
  lblSpectrumDecadeCount.Caption := IntToStr(c);
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumDecadeCount(Index :Integer) :Integer;
begin
  Result := prSpectrum.DecadeCount[Index];
end;

procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumDecadeOffset(Index :Integer; o :Integer);
begin
  prSpectrum.DecadeOffset[Index]  := o;
  udSpectrumDecadeOffset.Position := o;
  lblSpectrumDecadeOffset.Caption := IntToStr(o);
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumDecadeOffset(Index :Integer) :Integer;
begin
  Result := prSpectrum.DecadeOffset[Index];
end;

procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumTime(t :Integer);
begin
  prSpectrum.Time         := t;
  udSpectrumTime.Position := t;
  lblSpectrumTime.Caption := IntToStr(t);
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumTime :Integer;
begin
  Result := prSpectrum.Time;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumBinCount(c :Integer);
begin
  prSpectrum.BinCount         := c;
  udSpectrumBinCount.Position := c;
  txtSpectrumBinCount.Text    := IntToStr(c);
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumBinCount :Integer;
begin
  Result := prSpectrum.BinCount;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumAcc(a :Boolean);
begin
  prSpectrum.Acc         := a;
  chkSpectrumAcc.Checked := a;
  SpectrumAccUpdate      := SpectrumAccUpdate;
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumAcc :Boolean;
begin
  Result := prSpectrum.Acc;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumAccUpdate(u :Boolean);
begin
  prSpectrum.AccUpdate         := u;
  chkSpectrumAccUpdate.Checked := u;
  SpectrumAccPercent     := SpectrumAccPercent;
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumAccUpdate :Boolean;
begin
  Result := prSpectrum.AccUpdate;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetSpectrumAccPercent(p :Double);
begin
  prSpectrum.AccPercent      := p;
  txtSpectrumAccPercent.Text := FloatToStr(p);
end;

function TQUB_Viewport_AcquisitionInput.prGetSpectrumAccPercent :Double;
begin
  Result := prSpectrum.AccPercent;
end;

procedure TQUB_Viewport_AcquisitionInput.prSetMode(m :Boolean);
begin
  prMode := m;
  ConteinerScope.Visible    := not Mode;
  ConteinerSpectrum.Visible := Mode;
  prScope.Active            := not Mode;
  prSpectrum.Active         := Mode;
  if not Mode then begin
    btnMode.Caption         := 'Spectrum';
  end else begin
    btnMode.Caption         := 'Scope';
  end;
end;


//Events ***********************************************************************
procedure TQUB_Viewport_AcquisitionInput.FormShow(Sender: TObject);
begin
  inherited;

  panMode.Width        := btnMode.Width  + 2;
  panMode.Height       := btnMode.Height + 2;

  ScopeChannelIndex    := ScopeChannelIndex;
  ScopeTime            := ScopeTime;
  ScopeChartMode       := ScopeChartMode;

  SpectrumChannelIndex := SpectrumChannelIndex;

  Mode                 := Mode;

  //set the error

end;

procedure TQUB_Viewport_AcquisitionInput.udChannelIndexClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  if ScopeChannelIndex <> TUpDown(Sender).Position then begin
    ScopeChannelIndex := TUpDown(Sender).Position;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.btnGainClick(Sender: TObject);
begin
  inherited;
  if TSpeedButton(Sender).Tag <> Round(prScope.Gain[ScopeChannelIndex] * 1000000) then begin
    prScope.Gain[ScopeChannelIndex] := TSpeedButton(Sender).Tag / 1000000;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.btnTimeClick(Sender: TObject);
begin
  inherited;
  if TSpeedButton(Sender).Tag <> Round(prScope.XResolution * 1000 / 10) then begin
    prScope.XResolution := TSpeedButton(Sender).Tag * 10 / 1000;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.btnChartModeClick(Sender: TObject);
begin
  inherited;
  if TSpeedButton(Sender).Down <> prScope.ChartMode then begin
    prScope.ChartMode := TSpeedButton(Sender).Down;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.btnFindTraceClick(Sender: TObject);
begin
  inherited;
  prScope.FindTrace;
end;

procedure TQUB_Viewport_AcquisitionInput.btnModeClick(Sender: TObject);
begin
  inherited;
  Mode := not Mode;
end;

//Spectrum
procedure TQUB_Viewport_AcquisitionInput.udSpectrumChannelIndexClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  if SpectrumChannelIndex <> TUpDown(Sender).Position then begin
    SpectrumChannelIndex := TUpDown(Sender).Position;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.udSpectrumDecadeCountClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  if SpectrumDecadeCount[SpectrumChannelIndex] <> TUpDown(Sender).Position then begin
    SpectrumDecadeCount[SpectrumChannelIndex] := TUpDown(Sender).Position;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.udSpectrumDecadeOffsetClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  if SpectrumDecadeOffset[SpectrumChannelIndex] <> TUpDown(Sender).Position then begin
    SpectrumDecadeOffset[SpectrumChannelIndex] := TUpDown(Sender).Position;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.udSpectrumTimeClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  if SpectrumTime <> TUpDown(Sender).Position then begin
    SpectrumTime := TUpDown(Sender).Position;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.udSpectrumBinCountClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  if SpectrumBinCount <> TUpDown(Sender).Position then begin
    SpectrumBinCount := TUpDown(Sender).Position;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.txtSpectrumBinCountKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSpectrumBinCountExit(Sender);
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.txtSpectrumBinCountExit(Sender: TObject);
var
  c :Integer;
begin
  inherited;
  try
    c := StrToInt(TEdit(Sender).Text);
    if ((c < MIN_BINCOUNT) or (c > MAX_BINCOUNT)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_BINCOUNT) + ' and ' + IntToStr(MAX_BINCOUNT), mtError, [mbOK], 0);
    TEdit(Sender).Text := IntToStr(SpectrumBinCount);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if c <> SpectrumBinCount then begin
    SpectrumBinCount := c;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.chkSpectrumAccClick(Sender: TObject);
begin
  inherited;
  if SpectrumAcc <> TCheckBox(Sender).Checked then begin
    SpectrumAcc := TCheckBox(Sender).Checked;
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.chkSpectrumAccUpdateClick(Sender: TObject);
begin
  inherited;
  if SpectrumAccUpdate <> TCheckBox(Sender).Checked then begin
    SpectrumAccUpdate := TCheckBox(Sender).Checked;
  end;
end;


procedure TQUB_Viewport_AcquisitionInput.txtSpectrumAccPercentKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSpectrumAccPercentExit(Sender);
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.txtSpectrumAccPercentExit(Sender: TObject);
var
  p :Double;
begin
  inherited;
  try
    p := StrToFloat(TEdit(Sender).Text);
    if ((p < MIN_ACCPERCENT) or (p > MAX_ACCPERCENT)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(MIN_ACCPERCENT) + ' and ' + FloatToStr(MAX_ACCPERCENT), mtError, [mbOK], 0);   TEdit(Sender).Text := IntToStr(SpectrumBinCount);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if p <> SpectrumAccPercent then begin
    SpectrumAccPercent := p;
  end;
end;                       


//******************************************************************************
function TQUB_Viewport_AcquisitionInput.CanClose :Boolean;
begin
  Result := prAcquisitionInput.AcqState = ACQ_INPUT_STATE_STOPPED;
end;

function TQUB_Viewport_AcquisitionInput.NewInputAcq(const FileName :String) :Boolean;
begin

  Result := True;

  if AcquisitionInput.AcqState <> ACQ_INPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  //Instruct ViewportData to create a new data file and associate with it
  ViewportData.NewPropertiesForm.ADChannelCount := prAcquisitionInput.AIChannels.Count;
  if not ViewportData.New(FileName, False{Silent}) then begin
    Result := AssignInputAcq(ViewportData.CurrData);
  end else begin
    AssignInputAcq(AcquisitionInput.Data);
  end;

end;

function TQUB_Viewport_AcquisitionInput.OpenInputAcq(const FileName :String) :Boolean;
begin

  Result := True;

  if AcquisitionInput.AcqState <> ACQ_INPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  //Instruct ViewportData to create a new data file and associate with it
  if not ViewportData.Open(FileName, False{Silent}) then begin
    Result := AssignInputAcq(ViewportData.CurrData);
  end else begin
    AssignInputAcq(AcquisitionInput.Data);
  end;

end;

function TQUB_Viewport_AcquisitionInput.AssignInputAcq(Data :TQUB_QVO_QVD) :Boolean;
begin

  Result := True;

  if AcquisitionInput.AcqState <> ACQ_INPUT_STATE_STOPPED then Exit;
  //Acquisition already started

  if Data = nil then begin
    AcquisitionInput.Data := Data;
    Result := False;
  end else if Data.LockedAny then begin
    Exit; //Data was created but is, somehow, locked
  end else begin
    AcquisitionInput.Data := Data;
    Result := False;
  end;

end;

procedure TQUB_Viewport_AcquisitionInput.DoOnCloseData(Sender :TObject);
begin
  if Sender = AcquisitionInput.Data then begin
    if Sender <> nil then AssignInputAcq(nil);
  end;
end;

procedure TQUB_Viewport_AcquisitionInput.DoOnChangeHardware(Sender :TObject);
begin
  //...
end;

procedure TQUB_Viewport_AcquisitionInput.DoOnChangeChannels(Sender :TObject);
begin
  //...
end;

procedure TQUB_Viewport_AcquisitionInput.DoOnScopeChannelChange(Sender :TObject);
begin
  ScopeChannelIndex := TQUB_QVO_QVS(Sender).ChannelIndex;
end;

procedure TQUB_Viewport_AcquisitionInput.DoOnSpectrumChannelChange(Sender :TObject);
begin
  SpectrumChannelIndex := TQUB_QVO_QVSpectrum(Sender).ChannelIndex;
end;

procedure TQUB_Viewport_AcquisitionInput.DoProperties(Read :Boolean);
var
  Reg :TRegistry;
begin

  inherited;

  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Viewport ' + Name, True) then begin

        if Read then begin
          prMode := ReadRegistryBoolean(Reg, 'Mode', False);
        end else begin
          WriteBool('Mode', Mode);
        end;

      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;

procedure TQUB_Viewport_AcquisitionInput.DoPropertiesInternal(Read :Boolean);
var
  Reg :TRegistry;
begin

  inherited;

  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Interface ' + InterfaceName + '\Viewport ' + Name, True) then begin

        if Read then begin
          ConteinerLeft.Width := Round(ReadRegistryFloat(Reg, 'ConteinerLeft Width', 50.0) * ClientWidth / 100);
        end else begin
          WriteFloat('ConteinerLeft Width', 100 * ConteinerLeft.Width / ClientWidth);
        end;

      end;
      CloseKey;
    end;

    panMode.Left := ConteinerLeft.ClientWidth  - panMode.Width;
    panMode.Top  := ConteinerLeft.ClientHeight - panMode.Height;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
end.
