{
        A C Q U I S I T I O N     N O T E S

                (C) 11-??-04 L.M.

Creates a viewport with a number of buttons that add notes
in the data during acquisition. Can be used for, e.g., 
recording a solution change, amplifier gain, V/C-Clamp etc.

The buttons are editable with Mouse.RightClick
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

unit QUB_Viewport_AcquisitionNotes;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
StdCtrls, ExtCtrls, ComCtrls, Menus, Registry, Buttons, Math,

AllocTypes, Alloc,
QUB_Console, QUB_Rich_Console,
QUB_Viewport,
QUB_QVO_QVD;


const


BTN_WIDTH      = 100;
BTN_HEIGHT     = 20;
MAX_NOTE_COUNT = 20;


type


TQUB_Viewport_AcquisitionNotes = class(TQUB_Viewport)


  PopupMenu                    :TPopupMenu;
    AddItem                    :TMenuItem;
    InsertItem                 :TMenuItem;
    DeleteItem                 :TMenuItem;
    N1                         :TMenuItem;
    ColorItem                  :TMenuItem;
    PropertiesItem             :TMenuItem;


  procedure FormResize(Sender: TObject);
  procedure PopupMenuPopup(Sender: TObject);
  procedure AddItemClick(Sender: TObject);
  procedure InsertItemClick(Sender: TObject);
  procedure DeleteItemClick(Sender: TObject);
  procedure ColorItemClick(Sender: TObject);
  procedure PropertiesItemClick(Sender: TObject);


protected


  prData               :TQUB_QVO_QVD;

  prNoteCount          :Integer;
  prNoteButtons        :Array[0..MAX_NOTE_COUNT - 1] of TSpeedButton;
  prNotePanels         :Array[0..MAX_NOTE_COUNT - 1] of TPanel;

  prNames              :Array[0..MAX_NOTE_COUNT - 1] of String;
  prComments           :Array[0..MAX_NOTE_COUNT - 1] of String;
  prColors             :Array[0..MAX_NOTE_COUNT - 1] of TColor;

  prLastIndex          :Integer;


  procedure prResize;
  procedure prDelete(Index :Integer);
  procedure prInsert(Index :Integer);
  procedure prMove(FromIndex, ToIndex :Integer);
  procedure prColor(Index :Integer);
  procedure prNoteProperties(Index :Integer);
  procedure prProperties(Index :Integer);

  procedure prDoOnClick(Sender :TObject);

  procedure prBtnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

  procedure DoProperties(Read :Boolean); override;


public


  constructor CreateName(AOwner :TComponent; const Name :String); override;


  procedure DoOnChangeData(Sender :TObject);
  procedure DoOnClosingData(Sender :TObject);


end;


implementation


uses


QUB_GUIInterface, QUB_Node;


{$R *.DFM}


//****************************************************************************//
constructor TQUB_Viewport_AcquisitionNotes.CreateName(AOwner :TComponent; const Name :String);
var
  Btn     :TSpeedButton;
  Pan     :TPanel;
  i, w    :Integer;
begin

  prLastIndex := -1;

  inherited CreateName(AOwner, Name);

  Conteiner.PopupMenu := PopupMenu;

  //Create Buttons
  w := Round((Conteiner.ClientWidth - (prNoteCount - 1)) / prNoteCount);

  //Create all, used or not
  for i := 0 to MAX_NOTE_COUNT - 1 do begin

    Btn              := TSpeedButton.Create(Self);
    Pan              := TPanel.Create(Self);
    Pan.Visible      := False;

    Btn.OnMouseUp    := prBtnMouseUp;

    Btn.Font.Style   := [fsBold];

    Btn.ShowHint     := True;
    Pan.ShowHint     := True;

    Pan.PopupMenu    := PopupMenu;

    prNotePanels[i]  := Pan;
    prNoteButtons[i] := Btn;

    Btn.OnClick      := prDoOnClick;

    Pan.BevelOuter   := bvNone;
    Btn.Align        := alClient;

    Pan.SetBounds(0, 0, w, BTN_HEIGHT);

    Btn.Flat         := True;

    Btn.Tag          := i;
    Pan.Tag          := i;

    Btn.Parent       := Pan;
    Pan.Parent       := Conteiner;

  end;

  prResize;

end;

procedure TQUB_Viewport_AcquisitionNotes.FormResize(Sender: TObject);
begin
  inherited;
  prResize;
end;

procedure TQUB_Viewport_AcquisitionNotes.prDoOnClick(Sender :TObject);
begin
  //Insert Note in Data
  if prData <> nil then begin
    prData.AddNote('AcqNotes', prNames[TControl(Sender).Tag], prComments[TControl(Sender).Tag]);
  end;
end;

procedure TQUB_Viewport_AcquisitionNotes.DoOnChangeData(Sender :TObject);
begin
  prData    := TQUB_QVO_QVD(Sender);
end;

procedure TQUB_Viewport_AcquisitionNotes.DoOnClosingData(Sender :TObject);
begin
  if Sender = prData then DoOnChangeData(nil);
end;

//******************************************************************************
procedure TQUB_Viewport_AcquisitionNotes.PopupMenuPopup(Sender: TObject);
begin
  inherited;
  prLastIndex             := TPopupMenu(Sender).PopupComponent.Tag;
  AddItem.Enabled         := prNoteCount < MAX_NOTE_COUNT;
  InsertItem.Enabled      := prNoteCount < MAX_NOTE_COUNT;
  DeleteItem.Enabled      := prNoteCount > 0;
  ColorItem.Enabled       := prNoteCount > 0;
  PropertiesItem.Enabled  := prNoteCount > 0;
end;

procedure TQUB_Viewport_AcquisitionNotes.AddItemClick(Sender: TObject);
begin
  inherited;
  prInsert(prNoteCount);
  prResize;
end;

procedure TQUB_Viewport_AcquisitionNotes.InsertItemClick(Sender: TObject);
begin
  inherited;
  prInsert(prLastIndex);
  prResize;
end;

procedure TQUB_Viewport_AcquisitionNotes.DeleteItemClick(Sender: TObject);
begin
  inherited;
  prDelete(prLastIndex);
  prResize;
end;

procedure TQUB_Viewport_AcquisitionNotes.ColorItemClick(Sender: TObject);
begin
  inherited;
  prColor(prLastIndex);
end;

procedure TQUB_Viewport_AcquisitionNotes.PropertiesItemClick(Sender: TObject);
begin
  inherited;
  prProperties(prLastIndex);
end;

procedure TQUB_Viewport_AcquisitionNotes.prBtnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ToIndex :Integer;
begin
  inherited;
  if Button  = mbLeft then begin
    with TControl(Sender) do begin
      if (y >= 0) and (y < Height) then begin
        ToIndex := Tag + Floor(x / (Width + 1));
        ToIndex := Max(Min(ToIndex, prNoteCount - 1), 0);
        if Tag{FromIndex} <> ToIndex then begin
          prMove(Tag, ToIndex);
          prResize;
        end;
      end;//y
    end;//with
  end;//Button
end;

//******************************************************************************
procedure TQUB_Viewport_AcquisitionNotes.prResize;
var
  i, l, t, w, h :Integer;
begin
  //Resize Buttons
  l := 0;
  t := 0;
  w := Round((Conteiner.ClientWidth - (prNoteCount - 1)) / prNoteCount);
  h := Conteiner.ClientHeight;
  for i := 0 to prNoteCount - 1 do begin
    if prNoteButtons[i] <> nil then begin
      prNotePanels[i].SetBounds(l, t, w, h);
      prNotePanels[i].Visible  := True;
      prNoteProperties(i);
      Inc(l, w + 1);
    end else begin
      Exit;
    end;
  end;
  for i := prNoteCount to MAX_NOTE_COUNT - 1 do begin
    if prNoteButtons[i] <> nil then begin
      prNotePanels[i].Visible  := False;
    end else begin
      Exit;
    end;
  end;
end;

procedure TQUB_Viewport_AcquisitionNotes.prDelete(Index :Integer);
var
  i       :Integer;
begin
  if prNoteCount = 0 then Exit;
  for i := Index to prNoteCount - 2 do begin
    prNames[i]    := prNames[i + 1];
    prComments[i] := prComments[i + 1];
    prColors[i]   := prColors[i + 1];
  end;
  Dec(prNoteCount);
end;

procedure TQUB_Viewport_AcquisitionNotes.prInsert(Index :Integer);
var
  i       :Integer;
begin
  if prNoteCount = MAX_NOTE_COUNT then Exit;
  if Index = -1 then Index := 0;
  Inc(prNoteCount);
  for i := prNoteCount - 1 downto Index + 1 do begin
    prNames[i]    := prNames[i - 1];
    prComments[i] := prComments[i - 1];
    prColors[i]   := prColors[i - 1];
  end;
  prNames[Index]    := 'Note ' + IntToStr(Index);
  prComments[Index] := '';
  prColors[Index]   := clBtnFace;
end;

procedure TQUB_Viewport_AcquisitionNotes.prMove(FromIndex, ToIndex :Integer);
var
  Name    :String;
  Comment :String;
  Color   :TColor;
begin
  Name                := prNames[FromIndex];
  Comment             := prComments[FromIndex];
  Color               := prColors[FromIndex];
  prDelete(FromIndex);
  prInsert(ToIndex);
  prNames[ToIndex]    := Name;
  prComments[ToIndex] := Comment;
  prColors[ToIndex]   := Color;
end;

procedure TQUB_Viewport_AcquisitionNotes.prColor(Index :Integer);
var
  Dlg :TColorDialog;
begin
  Dlg := TColorDialog.Create(Self);
  try
    Dlg.Color := prColors[prLastIndex];
    if Dlg.Execute then begin
      prColors[prLastIndex] := Dlg.Color;
      prNoteProperties(prLastIndex);
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Viewport_AcquisitionNotes.prNoteProperties(Index :Integer);
begin
  prNoteButtons[Index].Caption := prNames[Index];
  prNoteButtons[Index].Hint    := prComments[Index];
  prNotePanels[Index].Color    := prColors[Index];
  prNoteButtons[Index].Tag     := Index;
  prNotePanels[Index].Tag      := Index;
end;

procedure TQUB_Viewport_AcquisitionNotes.prProperties(Index :Integer);
begin
  with SelectionPropertiesForm do begin
    //lblSelection.Caption := '...';
    txtSelFirst.Text     := '';
    txtSelLast.Text      := '';
    lstLabel.ItemIndex   := 0;
    lstLabel.Text        := prNames[prLastIndex];
    txtComment.Text      := prComments[prLastIndex];
    if ShowModal = mrOK then begin
      prNames[prLastIndex]    := lstLabel.Text;
      prComments[prLastIndex] := txtComment.Text;
      prNoteProperties(prLastIndex);
    end;
  end;
end;

//Store/Restore the specific properties, at OnCreate, and OnDestroy ............
procedure TQUB_Viewport_AcquisitionNotes.DoProperties(Read :Boolean);
var
  Reg :TRegistry;
  i   :Integer;
begin

  inherited;

  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Viewport ' + Name, True) then begin

        //Count
        if Read then begin
          if ValueExists('Count') then begin
            prNoteCount := ReadInteger('Count');
            if prNoteCount > MAX_NOTE_COUNT then prNoteCount := MAX_NOTE_COUNT;
          end else begin
            prNoteCount := 6;
          end;
        end else begin
          WriteInteger('Count', prNoteCount);
        end;
        //Properties
        for i := 0 to MAX_NOTE_COUNT - 1 do begin
          if Read then begin
            if ValueExists('Name ' + IntToStr(i)) then begin
              prNames[i] := ReadString('Name ' + IntToStr(i));
            end else begin
              prNames[i] := 'Note ' + IntToStr(i);
            end;
            if ValueExists('Comment ' + IntToStr(i)) then begin
              prComments[i] := ReadString('Comment ' + IntToStr(i));
            end else begin
              prComments[i] := '';
            end;
            if ValueExists('Color ' + IntToStr(i)) then begin
              prColors[i] := TColor(ReadInteger('Color ' + IntToStr(i)));
            end else begin
              prColors[i] := clBtnFace;
            end;
          end else begin
            WriteString('Name ' + IntToStr(i), prNames[i]);
            WriteString('Comment ' + IntToStr(i), prComments[i]);
            WriteInteger('Color ' + IntToStr(i), prColors[i]);
          end;
        end;//i

      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;

end;


//******************************************************************************
end.
