unit QUB_Viewport_AcquisitionOutput;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, Menus, Registry,

RegistryInterface,

QUB_Viewport,
QUB_Viewport_Data,
QUB_QVO_QVD,
QUB_AcquisitionOutput;


//{$include 'QUB_Viewport_AcquisitionOutput.hct'}


type


TQUB_Viewport_AcquisitionOutput = class(TQUB_Viewport)


  ConteinerOutput :TPanel;
  lblStatus       :TLabel;


protected


  prAcquisitionOutput         :TQUB_Acquisition_Output;


  procedure DoPropertiesInternal(Read :Boolean); override;


public


  constructor Create(AOwner :TComponent); override;


  function CanClose :Boolean; override;

  function NewOutputAcq(const FileName :String) :Boolean;
  function OpenOutputAcq(const FileName :String) :Boolean;
  function AssignOutputAcq(Data :TQUB_QVO_QVD) :Boolean;

  procedure DoOnCloseData(Sender :TObject);
  procedure DoOnChangeHardware(Sender :TObject);
  procedure DoOnChangeChannels(Sender :TObject);


  property AcquisitionOutput  :TQUB_Acquisition_Output  read prAcquisitionOutput;


end;


implementation


{$R *.DFM}


//******************************************************************************
constructor TQUB_Viewport_AcquisitionOutput.Create(AOwner :TComponent);
begin
  inherited;

  prAcquisitionOutput             := TQUB_Acquisition_Output.Create(Self);
  prAcquisitionOutput.Parent      := ConteinerOutput;
  prAcquisitionOutput.OnConstrainedResize := OnConstrainedResize;
  prAcquisitionOutput.panPanels.OnConstrainedResize := OnConstrainedResize;
  prAcquisitionOutput.Status      := lblStatus;

  AssignOutputAcq(nil);

end;


function TQUB_Viewport_AcquisitionOutput.CanClose :Boolean;
begin
  Result := prAcquisitionOutput.AcqState = ACQ_OUTPUT_STATE_STOPPED;
end;

function TQUB_Viewport_AcquisitionOutput.NewOutputAcq(const FileName :String) :Boolean;
begin

  Result := AcquisitionOutput.NewOutputAcq(FileName);

end;

function TQUB_Viewport_AcquisitionOutput.OpenOutputAcq(const FileName :String) :Boolean;
begin

  Result := AcquisitionOutput.OpenOutputAcq(FileName);

end;

function TQUB_Viewport_AcquisitionOutput.AssignOutputAcq(Data :TQUB_QVO_QVD) :Boolean;
begin

  Result := AcquisitionOutput.AssignOutputAcq(Data);

end;

procedure TQUB_Viewport_AcquisitionOutput.DoOnCloseData(Sender :TObject);
begin
  if Sender = AcquisitionOutput.Data then begin
    if Sender <> nil then AssignOutputAcq(nil);
  end;
end;

procedure TQUB_Viewport_AcquisitionOutput.DoOnChangeHardware(Sender :TObject);
begin
  //...
end;

procedure TQUB_Viewport_AcquisitionOutput.DoOnChangeChannels(Sender :TObject);
begin
  //...
end;

procedure TQUB_Viewport_AcquisitionOutput.DoPropertiesInternal(Read :Boolean);
//var
//  Reg :TRegistry;
begin

  inherited;
{ ----- middle commented out by Lorin, may as well remove the rest. JB 3/2005.
  Reg := TRegistry.Create;

  try

    with Reg do begin
      if OpenKey('Software\QUB\Interface ' + InterfaceName + '\Viewport ' + Name, True) then begin
        //if Read then begin
        //  prAcquisitionOutput.lstPanels.Width := Round(ReadRegistryFloat(Reg, 'Controls Width', 30.0) * ClientWidth / 100);
        //end else begin
        //  WriteFloat('Controls Width', 100 * prAcquisitionOutput.lstPanels.Width / ClientWidth);
        //end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
}
end;


//******************************************************************************
end.
