unit QUB_Viewport_AcquisitionProtocol;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Buttons, 
Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Menus, Registry,

QUB_Viewport,
QUB_Acquisition_Hardware,
QUB_AcquisitionProtocol_Properties,
QUB_Protocol;


type


TQUB_Viewport_AcquisitionProtocol = class(TQUB_Viewport)


  btnRun                          :TSpeedButton;
  btnStop                         :TSpeedButton;
  btnPause                        :TSpeedButton;

  btnProperties                   :TSpeedButton;

  panOverwrite                    :TPanel;


  procedure btnPropertiesClick(Sender: TObject);


protected


  prPropertiesForm             :TQUB_AcquisitionProtocol_PropertiesForm;

  prAIChannels                 :TQUB_Acquisition_Channels;
  prAOChannels                 :TQUB_Acquisition_Channels;
  prDIChannels                 :TQUB_Acquisition_Channels;
  prDOChannels                 :TQUB_Acquisition_Channels;

  prMachine                    :TQUB_Protocol_Machine;
  prOperations                 :TQUB_Protocol_Operations;
  

  procedure DoProperties(Read :Boolean); override;

  procedure prUpdateChannels;


public


  constructor CreateName(AOwner :TComponent; const Name :String); override;
  destructor Destroy; override;


  procedure DoOnChangeHardware(Sender :TObject);
  procedure DoOnChangeChannels(Sender :TObject);


end;


implementation


{$R *.DFM}


//****************************************************************************//
constructor TQUB_Viewport_AcquisitionProtocol.CreateName(AOwner :TComponent; const Name :String);
begin

  //Declare them here, so DoProperties works
  prAIChannels     := TQUB_Acquisition_Channels.Create;
  prAOChannels     := TQUB_Acquisition_Channels.Create;
  prDIChannels     := TQUB_Acquisition_Channels.Create;
  prDOChannels     := TQUB_Acquisition_Channels.Create;

  prAIChannels.OwnChannels := False;
  prAOChannels.OwnChannels := False;
  prDIChannels.OwnChannels := False;
  prDOChannels.OwnChannels := False;

  inherited;

  prPropertiesForm := TQUB_AcquisitionProtocol_PropertiesForm.Create(Self);

  prPropertiesForm.AIChannels := prAIChannels;
  prPropertiesForm.AOChannels := prAOChannels;
  prPropertiesForm.DIChannels := prDIChannels;
  prPropertiesForm.DOChannels := prDOChannels;

  prUpdateChannels;

end;

destructor TQUB_Viewport_AcquisitionProtocol.Destroy;
begin

  try

  finally

    inherited;

    prAIChannels.Free;
    prAOChannels.Free;
    prDIChannels.Free;
    prDOChannels.Free;

  end;

end;

//DoOnChangeHardware
procedure TQUB_Viewport_AcquisitionProtocol.DoOnChangeHardware(Sender :TObject);
begin
  prUpdateChannels;
end;

//DoOnChangeChannels
procedure TQUB_Viewport_AcquisitionProtocol.DoOnChangeChannels(Sender :TObject);
begin
  prUpdateChannels;
end;

//Update Channels *************************************************************
procedure TQUB_Viewport_AcquisitionProtocol.prUpdateChannels;
var
  ACh  :TQUB_Acquisition_AChannel;
  DCh  :TQUB_Acquisition_DChannel;
  c, i :Integer;
begin

  //Remove Dead Channels *******************************************************

  //AIChannels
  for c := prAIChannels.Count - 1 downto 0 do begin
    i := QUBAcqAIChannels.IndexOfObject(prAIChannels.Objects[c]);
    if i < 0 then begin
      prAIChannels.Delete(c);
    end else begin
      prAIChannels.Strings[c] := QUBAcqAIChannels.Strings[i];
    end;
  end;

  //AOChannels
  for c := prAOChannels.Count - 1 downto 0 do begin
    i := QUBAcqAOChannels.IndexOfObject(prAOChannels.Objects[c]);
    if i < 0 then begin
      prAOChannels.Delete(c);
    end else begin
      prAOChannels.Strings[c] := QUBAcqAOChannels.Strings[i];
    end;
  end;

  //DIChannels
  for c := prDIChannels.Count - 1 downto 0 do begin
    i := QUBAcqDIChannels.IndexOfObject(prDIChannels.Objects[c]);
    if i < 0 then begin
      prDIChannels.Delete(c);
    end else begin
      prDIChannels.Strings[c] := QUBAcqDIChannels.Strings[i];
    end;
  end;

  //DOChannels
  for c := prDOChannels.Count - 1 downto 0 do begin
    i := QUBAcqDOChannels.IndexOfObject(prDOChannels.Objects[c]);
    if i < 0 then begin
      prDOChannels.Delete(c);
    end else begin
      prDOChannels.Strings[c] := QUBAcqDOChannels.Strings[i];
    end;
  end;


  //Links **********************************************************************

  //AIChannels
  for c := 0 to prAIChannels.Count - 1 do begin
    ACh     := TQUB_Acquisition_AChannel(prAIChannels.Objects[c]);
    i       := QUBAcqIDevices.IndexOf(ACh.DeviceName);
    if i < 0 then begin
      ACh.Device := nil;
    end else begin
      ACh.Device := TQUB_Acquisition_IDevice(QUBAcqIDevices.Objects[i]);
    end;
  end;
  //AOChannels
  for c := 0 to prAOChannels.Count - 1 do begin
    ACh     := TQUB_Acquisition_AChannel(prAOChannels.Objects[c]);
    i       := QUBAcqODevices.IndexOf(ACh.DeviceName);
    if i < 0 then begin
      ACh.Device := nil;
    end else begin
      ACh.Device := TQUB_Acquisition_ODevice(QUBAcqODevices.Objects[i]);
    end;
  end;
  //DIChannels
  for c := 0 to prDIChannels.Count - 1 do begin
    DCh     := TQUB_Acquisition_DChannel(prDIChannels.Objects[c]);
    i       := QUBAcqIDevices.IndexOf(DCh.DeviceName);
    if i < 0 then begin
      DCh.Device := nil;
    end else begin
      DCh.Device := TQUB_Acquisition_IDevice(QUBAcqIDevices.Objects[i]);
    end;
  end;
  //DOChannels
  for c := 0 to prDOChannels.Count - 1 do begin
    DCh     := TQUB_Acquisition_DChannel(prDOChannels.Objects[c]);
    i       := QUBAcqODevices.IndexOf(DCh.DeviceName);
    if i < 0 then begin
      DCh.Device := nil;
    end else begin
      DCh.Device := TQUB_Acquisition_ODevice(QUBAcqODevices.Objects[i]);
    end;
  end;

end;


//******************************************************************************
procedure TQUB_Viewport_AcquisitionProtocol.btnPropertiesClick(Sender: TObject);
begin
  inherited;

  prPropertiesForm.ShowModal;

  prUpdateChannels;

end;

//******************************************************************************
procedure TQUB_Viewport_AcquisitionProtocol.DoProperties(Read :Boolean);
var
  Reg  :TRegistry;
  Ch   :TQUB_Acquisition_Channel;
  Idx  :Integer;
  c    :Integer;
begin

  Reg  := TRegistry.Create;

  try

    with Reg do begin

      //AIChannels *************************************************************
      if Read then begin

        c := 0;
        repeat
          if OpenKey('Software\QUB\Viewport ' + Name + '\Analog Input\AIChannel ' + IntToStr(c + 1), False) then begin
            if ValueExists('Name') then begin
              Idx := QUBAcqAIChannels.IndexOf(ReadString('Name'));
              if Idx >= 0 then begin
                Ch := TQUB_Acquisition_Channel(QUBAcqAIChannels.Objects[Idx]);
                prAIChannels.AddObject(Ch.Name, Ch);
              end;//Idx > 0
            end;//Value exists
          end else begin
            Break;
          end;
          Inc(c);
          CloseKey;
        until(False);
        CloseKey;

      end else begin

        if OpenKey('Software\QUB\Viewport ' + Name, True) then begin
          DeleteKey('Analog Input');
        end;
        CloseKey;

        for c := 0 to prAIChannels.Count - 1 do begin
          Ch := TQUB_Acquisition_Channel(prAIChannels.Objects[c]);
          if OpenKey('Software\QUB\Viewport ' + Name + '\Analog Input\AIChannel ' + IntToStr(c + 1), True) then begin
            WriteString('Name', Ch.Name);
          end;
          CloseKey;
        end;
        CloseKey;

      end;//Read/Write
      CloseKey;


      //AOChannels *************************************************************
      if Read then begin

        c := 0;
        repeat
          if OpenKey('Software\QUB\Viewport ' + Name + '\Analog Output\AOChannel ' + IntToStr(c + 1), False) then begin
            if ValueExists('Name') then begin
              Idx := QUBAcqAOChannels.IndexOf(ReadString('Name'));
              if Idx >= 0 then begin
                Ch := TQUB_Acquisition_Channel(QUBAcqAOChannels.Objects[Idx]);
                prAOChannels.AddObject(Ch.Name, Ch);
              end;//Idx > 0
            end;//Value exists
          end else begin
            Break;
          end;
          Inc(c);
          CloseKey;
        until(False);
        CloseKey;

      end else begin

        if OpenKey('Software\QUB\Viewport ' + Name, True) then begin
          DeleteKey('Analog Output');
        end;
        CloseKey;

        for c := 0 to prAOChannels.Count - 1 do begin
          Ch := TQUB_Acquisition_Channel(prAOChannels.Objects[c]);
          if OpenKey('Software\QUB\Viewport ' + Name + '\Analog Output\AOChannel ' + IntToStr(c + 1), True) then begin
            WriteString('Name', Ch.Name);
          end;
          CloseKey;
        end;
        CloseKey;

      end;//Read/Write
      CloseKey;

      //DIChannels *************************************************************
      if Read then begin

        c := 0;
        repeat
          if OpenKey('Software\QUB\Viewport ' + Name + '\Digital Input\DIChannel ' + IntToStr(c + 1), False) then begin
            if ValueExists('Name') then begin
              Idx := QUBAcqDIChannels.IndexOf(ReadString('Name'));
              if Idx >= 0 then begin
                Ch := TQUB_Acquisition_Channel(QUBAcqDIChannels.Objects[Idx]);
                prDIChannels.AddObject(Ch.Name, Ch);
              end;//Idx > 0
            end;//Value exists
          end else begin
            Break;
          end;
          Inc(c);
          CloseKey;
        until(False);
        CloseKey;

      end else begin

        if OpenKey('Software\QUB\Viewport ' + Name, True) then begin
          DeleteKey('Digital Input');
        end;
        CloseKey;

        for c := 0 to prDIChannels.Count - 1 do begin
          Ch := TQUB_Acquisition_Channel(prDIChannels.Objects[c]);
          if OpenKey('Software\QUB\Viewport ' + Name + '\Digital Input\DIChannel ' + IntToStr(c + 1), True) then begin
            WriteString('Name', Ch.Name);
          end;
          CloseKey;
        end;
        CloseKey;

      end;//Read/Write
      CloseKey;

      //DOChannels *************************************************************
      if Read then begin

        c := 0;
        repeat
          if OpenKey('Software\QUB\Viewport ' + Name + '\Digital Output\DOChannel ' + IntToStr(c + 1), False) then begin
            if ValueExists('Name') then begin
              Idx := QUBAcqDOChannels.IndexOf(ReadString('Name'));
              if Idx >= 0 then begin
                Ch := TQUB_Acquisition_Channel(QUBAcqDOChannels.Objects[Idx]);
                prDOChannels.AddObject(Ch.Name, Ch);
              end;//Idx > 0
            end;//Value exists
          end else begin
            Break;
          end;
          Inc(c);
          CloseKey;
        until(False);
        CloseKey;

      end else begin

        if OpenKey('Software\QUB\Viewport ' + Name, True) then begin
          DeleteKey('Digital Output');
        end;
        CloseKey;

        for c := 0 to prDOChannels.Count - 1 do begin
          Ch := TQUB_Acquisition_Channel(prDOChannels.Objects[c]);
          if OpenKey('Software\QUB\Viewport ' + Name + '\Digital Output\DOChannel ' + IntToStr(c + 1), True) then begin
            WriteString('Name', Ch.Name);
          end;
          CloseKey;
        end;
        CloseKey;

      end;//Read/Write
      CloseKey;

    end;//with Reg

  finally

    Reg.Free;

  end;

end;


//******************************************************************************
end.
