{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// aka Manual Control


unit QUB_Viewport_AcquisitionSShot;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Buttons, 
Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Menus, Registry,
Alloc, AllocTypes,

QUB_Viewport,
QUB_Acquisition_Global,
QUB_Acquisition_Hardware,
QUB_AcquisitionSShot_Properties,
QUB_Acquisition_AIChannel_Frame, QUB_Acquisition_AOChannel_Frame,
QUB_Acquisition_DIChannel_Frame, QUB_Acquisition_DOChannel_Frame;


type


TQUB_Viewport_AcquisitionSShot = class(TQUB_Viewport)


  panAnalogInput               :TPanel;
  Label1                       :TLabel;
  panAnalogOutput              :TPanel;
  Label2                       :TLabel;
  panDigitalInput              :TPanel;
  Label3                       :TLabel;
  panDigitalOutput             :TPanel;
  Label4                       :TLabel;
  panResistance                :TPanel;
  Panel2                       :TPanel;
  btnProperties                :TSpeedButton;
  panNameR                     :TLabel;
  chkR                         :TCheckBox;
  txtR                         :TEdit;
  panRead                      :TPanel;
  btnReadR                     :TSpeedButton;
  Label6                       :TLabel;
  Label7                       :TLabel;
  txtRmV                       :TEdit;
  txtRpA                       :TEdit;


  procedure btnPropertiesClick(Sender: TObject);
  procedure ConteinerClick(Sender: TObject);
  procedure Label4Click(Sender: TObject);
  procedure Label3Click(Sender: TObject);
  procedure Label2Click(Sender: TObject);
  procedure panAnalogInputClick(Sender: TObject);
  procedure Label1Click(Sender: TObject);
  procedure panAnalogOutputClick(Sender: TObject);
  procedure panDigitalInputClick(Sender: TObject);
  procedure panDigitalOutputClick(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure txtRmVKeyPress(Sender: TObject; var Key: Char);
  procedure txtRmVExit(Sender: TObject);
  procedure btnReadRClick(Sender: TObject);


protected


  prPropertiesForm             :TQUB_AcquisitionSShot_PropertiesForm;

  prAIChannels                 :TQUB_Acquisition_Channels;
  prAOChannels                 :TQUB_Acquisition_Channels;
  prDIChannels                 :TQUB_Acquisition_Channels;
  prDOChannels                 :TQUB_Acquisition_Channels;

  prTimer                      :TTimer;

  prRVChannel                  :Boolean;
  prRVName                     :String;
  prRVScale                    :Double; // rvscale * v-chan-reading = mV
  prRmV                        :Double;
  prRIName                     :String;


  procedure DoProperties(Read :Boolean); override;
  procedure DoPropertiesGroup(Read :Boolean; cKey :String; aChan :TQUB_Acquisition_Channels; Reg  :TRegistry; aChanLocal :TQUB_Acquisition_Channels);
  procedure DoPropertiesR(Read :Boolean; cKey :String; Reg  :TRegistry);
  
  procedure prUpdateChannels;

  procedure prOnTimer(Sender :TObject);

  procedure prOnAIRead(Sender :TObject);
  procedure prOnAOWrite(Sender :TObject);
  procedure prOnDIRead(Sender :TObject);
  procedure prOnDOWrite(Sender :TObject);

  procedure prOnAcquire(Sender: TObject; Buf :TQUB_AI_PBuffer; NotLast :Boolean);

  procedure prSetRVChannel(x :Boolean);
  procedure prSetRVName(x :String);
  procedure prSetRVScale(x :Double);
  procedure prSetRmV(x :Double);
  procedure prSetRIName(x :String);
  procedure prUpdateR(ReRead :Boolean);

public


  constructor CreateName(AOwner :TComponent; const Name :String); override;
  destructor Destroy; override;


  procedure DoOnChangeHardware(Sender :TObject);
  procedure DoOnChangeChannels(Sender :TObject);

  property R_VChannel   :Boolean      read prRVChannel    write prSetRVChannel;
  property R_VName      :String       read prRVName       write prSetRVName;
  property R_VScale     :Double       read prRVScale      write prSetRVScale;
  property R_mV         :Double       read prRmV          write prSetRmV;
  property R_IName      :String       read prRIName       write prSetRIName;

end;


implementation


uses

FMath,
QUB_GUIInterface,
QUB_AcquisitionInput,
QUB_QFS_QDF_MPX,
QUB_Node,
RegistryInterface;


{$R *.DFM}


//****************************************************************************//
constructor TQUB_Viewport_AcquisitionSShot.CreateName(AOwner :TComponent; const Name :String);
begin
  prTimer := TTimer.Create(Self);
  prTimer.Enabled  := False;
  prTimer.Interval := 50;
  prTimer.OnTimer  := prOnTimer;

  //Declare them here, so DoProperties works
  prAIChannels     := TQUB_Acquisition_Channels.Create;
  prAOChannels     := TQUB_Acquisition_Channels.Create;
  prDIChannels     := TQUB_Acquisition_Channels.Create;
  prDOChannels     := TQUB_Acquisition_Channels.Create;

  prAIChannels.OwnChannels := False;
  prAOChannels.OwnChannels := False;
  prDIChannels.OwnChannels := False;
  prDOChannels.OwnChannels := False;

  inherited;

  Caption := 'Acquisition Manual Control';

  prPropertiesForm := TQUB_AcquisitionSShot_PropertiesForm.Create(Self);

  prPropertiesForm.AIChannels := prAIChannels;
  prPropertiesForm.AOChannels := prAOChannels;
  prPropertiesForm.DIChannels := prDIChannels;
  prPropertiesForm.DOChannels := prDOChannels;

  prUpdateChannels;

end;

destructor TQUB_Viewport_AcquisitionSShot.Destroy;
begin

  try

    prTimer.Enabled := False;
    prTimer.OnTimer := nil;

  finally

    inherited;

    prAIChannels.Free;
    prAOChannels.Free;
    prDIChannels.Free;
    prDOChannels.Free;

  end;

end;

procedure TQUB_Viewport_AcquisitionSShot.FormShow(Sender: TObject);
begin
  inherited;
  ViewportAcqInput.AcquisitionInput.OnAcquire := prOnAcquire;
end;

procedure TQUB_Viewport_AcquisitionSShot.FormHide(Sender: TObject);
begin
  inherited;
  ViewportAcqInput.AcquisitionInput.OnAcquire := nil;
end;

//DoOnChangeHardware
procedure TQUB_Viewport_AcquisitionSShot.DoOnChangeHardware(Sender :TObject);
begin
  prUpdateChannels;
end;

//DoOnChangeChannels
procedure TQUB_Viewport_AcquisitionSShot.DoOnChangeChannels(Sender :TObject);
begin
  prUpdateChannels;
end;

//Timer ************************************************************************
procedure TQUB_Viewport_AcquisitionSShot.prOnTimer(Sender :TObject);
var
  AICh :TQUB_Acquisition_AIChannel;
  DICh :TQUB_Acquisition_DIChannel;
  c, e :Integer;
begin
  for c := 0 to prAIChannels.Count - 1 do begin
    AICh := TQUB_Acquisition_AIChannel(prAIChannels.Objects[c]);
    if AICh.Update then begin
      with AICh.Control as TQUB_Acquisition_AIChannelFrame do begin
        e := AICh.Read;
        if e = 0 then begin
          if panName.Color <> clBtnFace then panName.Color := clBtnFace;
          panName.Hint := '';
        end else begin
          if panName.Color <> clRed     then panName.Color := clRed;
          panName.Hint := 'Error: ' + IntToStr(e);
        end;
        Value := AICh.Value;
      end;
    end;
  end;

  for c := 0 to prDIChannels.Count - 1 do begin
    DICh := TQUB_Acquisition_DIChannel(prDIChannels.Objects[c]);
    if DICh.Update then begin
      with DICh.Control as TQUB_Acquisition_DIChannelFrame do begin
        e := DICh.Read;
        if e = 0 then begin
          if panName.Color <> clBtnFace then panName.Color := clBtnFace;
          panName.Hint := '';
        end else begin
          if panName.Color <> clRed     then panName.Color := clRed;
          panName.Hint := 'Error: ' + IntToStr(e);
        end;
        Value := DICh.Value;
      end;
    end;
  end;

end;

//Read/Write Events ************************************************************
procedure TQUB_Viewport_AcquisitionSShot.prOnAIRead(Sender :TObject);
var
  e :Integer;
begin
  with Sender as TQUB_Acquisition_AIChannelFrame do begin
    e := AIChannel.Read;
    if e = 0 then begin
      if panName.Color <> clBtnFace then panName.Color := clBtnFace;
      panName.Hint := '';
    end else begin
      if panName.Color <> clRed     then panName.Color := clRed;
      panName.Hint := 'Error: ' + IntToStr(e);
    end;
    Value := AIChannel.Value;
  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.prOnAOWrite(Sender :TObject);
var
  e :Integer;
begin
  with Sender as TQUB_Acquisition_AOChannelFrame do begin
    e := AOChannel.Write;
    if e = 0 then begin
      if panName.Color <> clBtnFace then panName.Color := clBtnFace;
      panName.Hint := '';
    end else begin
      if panName.Color <> clRed     then panName.Color := clRed;
      panName.Hint := 'Error: ' + IntToStr(e);
    end;
  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.prOnDIRead(Sender :TObject);
var
  e :Integer;
begin
  with Sender as TQUB_Acquisition_DIChannelFrame do begin
    e := DIChannel.Read;
    if e = 0 then begin
      if panName.Color <> clBtnFace then panName.Color := clBtnFace;
      panName.Hint := '';
    end else begin
      if panName.Color <> clRed     then panName.Color := clRed;
      panName.Hint := 'Error: ' + IntToStr(e);
    end;
    Value := DIChannel.Value;
  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.prOnDOWrite(Sender :TObject);
var
  e :Integer;
begin
  with Sender as TQUB_Acquisition_DOChannelFrame do begin
    e := DOChannel.Write;
    if e = 0 then begin
      if panName.Color <> clBtnFace then panName.Color := clBtnFace;
      panName.Hint := '';
    end else begin
      if panName.Color <> clRed     then panName.Color := clRed;
      panName.Hint := 'Error: ' + IntToStr(e);
    end;
  end;
end;


//Update Channels *************************************************************
procedure TQUB_Viewport_AcquisitionSShot.prUpdateChannels;
var
  AIFrame :TQUB_Acquisition_AIChannelFrame;
  AOFrame :TQUB_Acquisition_AOChannelFrame;
  DIFrame :TQUB_Acquisition_DIChannelFrame;
  DOFrame :TQUB_Acquisition_DOChannelFrame;
  ACh     :TQUB_Acquisition_AChannel;
  DCh     :TQUB_Acquisition_DChannel;
  Count   :Integer;
  c, i, t :Integer;
  Update  :Boolean;
begin

  Update  := False;

  //Remove Dead Channels *******************************************************

  //AIChannels
  for c := prAIChannels.Count - 1 downto 0 do begin
    i := QUBAcqAIChannels.IndexOfObject(prAIChannels.Objects[c]);
    if i < 0 then begin
      prAIChannels.Delete(c);
    end else begin
      prAIChannels.Strings[c] := QUBAcqAIChannels.Strings[i];
    end;
  end;

  //AOChannels
  for c := prAOChannels.Count - 1 downto 0 do begin
    i := QUBAcqAOChannels.IndexOfObject(prAOChannels.Objects[c]);
    if i < 0 then begin
      prAOChannels.Delete(c);
    end else begin
      prAOChannels.Strings[c] := QUBAcqAOChannels.Strings[i];
    end;
  end;

  //DIChannels
  for c := prDIChannels.Count - 1 downto 0 do begin
    i := QUBAcqDIChannels.IndexOfObject(prDIChannels.Objects[c]);
    if i < 0 then begin
      prDIChannels.Delete(c);
    end else begin
      prDIChannels.Strings[c] := QUBAcqDIChannels.Strings[i];
    end;
  end;

  //DOChannels
  for c := prDOChannels.Count - 1 downto 0 do begin
    i := QUBAcqDOChannels.IndexOfObject(prDOChannels.Objects[c]);
    if i < 0 then begin
      prDOChannels.Delete(c);
    end else begin
      prDOChannels.Strings[c] := QUBAcqDOChannels.Strings[i];
    end;
  end;


  //Frames *********************************************************************

  //AIChannels
  Count               := panAnalogInput.ControlCount;
  t                   := 0 + Count * 20;
  while prAIChannels.Count > Count do begin
    AIFrame           := TQUB_Acquisition_AIChannelFrame.Create(Self);
    AIFrame.OnRead    := prOnAIRead;
    AIFrame.Top       := t;
    AIFrame.Parent    := panAnalogInput;
    AIFrame.Name      := 'AIChannelFrame' + IntToStr(Count);
    Inc(Count);
    Inc(t, AIFrame.Height);
  end;
  while prAIChannels.Count < Count do begin
    AIFrame           := TQUB_Acquisition_AIChannelFrame(panAnalogInput.Controls[panAnalogInput.ControlCount - 1]);
    Dec(Count);
    Dec(t, AIFrame.Height);
    AIFrame.OnRead    := nil;
    AIFrame.Free;
  end;

  if t = 0 then t := 20;
  if panAnalogInput.Height <> t + 2 then panAnalogInput.Height := t + 2;

  //AOChannels
  Count               := panAnalogOutput.ControlCount;
  t                   := 0 + Count * 20;
  while prAOChannels.Count > Count do begin
    AOFrame           := TQUB_Acquisition_AOChannelFrame.Create(Self);
    AOFrame.OnWrite   := prOnAOWrite;
    AOFrame.Top       := t;
    AOFrame.Parent    := panAnalogOutput;
    AOFrame.Name      := 'AOChannelFrame' + IntToStr(Count);
    Inc(Count);
    Inc(t, AOFrame.Height);
  end;
  while prAOChannels.Count < Count do begin
    AOFrame           := TQUB_Acquisition_AOChannelFrame(panAnalogOutput.Controls[panAnalogOutput.ControlCount - 1]);
    Dec(Count);
    Dec(t, AOFrame.Height);
    AOFrame.OnWrite   := nil;
    AOFrame.Free;
  end;

  if t = 0 then t := 20;
  if panAnalogOutput.Height <> t + 2 then panAnalogOutput.Height := t + 2;

  //DIChannels
  Count               := panDigitalInput.ControlCount;
  t                   := 0 + Count * 20;
  while prDIChannels.Count > Count do begin
    DIFrame           := TQUB_Acquisition_DIChannelFrame.Create(Self);
    DIFrame.OnRead    := prOnDIRead;
    DIFrame.Top       := t;
    DIFrame.Parent    := panDigitalInput;
    DIFrame.Name      := 'DIChannelFrame' + IntToStr(Count);
    Inc(Count);
    Inc(t, DIFrame.Height);
  end;
  while prDIChannels.Count < Count do begin
    DIFrame           := TQUB_Acquisition_DIChannelFrame(panDigitalInput.Controls[panDigitalInput.ControlCount - 1]);
    Dec(Count);
    Dec(t, DIFrame.Height);
    DIFrame.OnRead    := nil;
    DIFrame.Free;
  end;

  if t = 0 then t := 20;
  if panDigitalInput.Height <> t + 2 then panDigitalInput.Height := t + 2;

  //DOChannels
  Count               := panDigitalOutput.ControlCount;
  t                   := 0 + Count * 20;
  while prDOChannels.Count > Count do begin
    DOFrame           := TQUB_Acquisition_DOChannelFrame.Create(Self);
    DOFrame.OnWrite   := prOnDOWrite;
    DOFrame.Top       := t;
    DOFrame.Parent    := panDigitalOutput;
    DOFrame.Name      := 'DOChannelFrame' + IntToStr(Count);
    Inc(Count);
    Inc(t, DOFrame.Height);
  end;
  while prDOChannels.Count < Count do begin
    DOFrame           := TQUB_Acquisition_DOChannelFrame(panDigitalOutput.Controls[panDigitalOutput.ControlCount - 1]);
    Dec(Count);
    Dec(t, DOFrame.Height);
    DOFrame.OnWrite   := nil;
    DOFrame.Free;
  end;

  if t = 0 then t := 20;
  if panDigitalOutput.Height <> t + 2 then panDigitalOutput.Height := t + 2;


  //Links **********************************************************************

  //AIChannels
  for c := 0 to prAIChannels.Count - 1 do begin
    AIFrame := TQUB_Acquisition_AIChannelFrame(panAnalogInput.Controls[c]);
    ACh     := TQUB_Acquisition_AChannel(prAIChannels.Objects[c]);
    ACh.Control := AIFrame;
    if not Update then Update := ACh.Update;
    AIFrame.AIChannel := TQUB_Acquisition_AIChannel(ACh);
    AIFrame.panName.Hint := '';
    i       := QUBAcqIDevices.IndexOf(ACh.DeviceName);
    if i < 0 then begin
      ACh.Device := nil;
      AIFrame.panName.Color := clRed;
      AIFrame.panName.Hint  := 'Device unassigned';
    end else begin
      ACh.Device := TQUB_Acquisition_IDevice(QUBAcqIDevices.Objects[i]);
      AIFrame.panName.Color := clBtnFace;
    end;
    AIFrame.panName.Caption  := ACh.Name;
    AIFrame.Value            := ACh.Value;
    AIFrame.panUnits.Caption := '[' + ACh.Units + ']';
  end;
  //AOChannels
  for c := 0 to prAOChannels.Count - 1 do begin
    AOFrame := TQUB_Acquisition_AOChannelFrame(panAnalogOutput.Controls[c]);
    ACh     := TQUB_Acquisition_AChannel(prAOChannels.Objects[c]);
    ACh.Control := AOFrame;
    AOFrame.AOChannel := TQUB_Acquisition_AOChannel(ACh);
    AOFrame.panName.Hint := '';
    i       := QUBAcqODevices.IndexOf(ACh.DeviceName);
    if i < 0 then begin
      ACh.Device := nil;
      AOFrame.panName.Color := clRed;
      AOFrame.panName.Hint  := 'Device unassigned';
    end else begin
      ACh.Device := TQUB_Acquisition_ODevice(QUBAcqODevices.Objects[i]);
      AOFrame.panName.Color := clBtnFace;
    end;
    AOFrame.panName.Caption  := ACh.Name;
    AOFrame.Value            := ACh.Value;
    AOFrame.panUnits.Caption := '[' + ACh.Units + ']';
  end;
  //DIChannels
  for c := 0 to prDIChannels.Count - 1 do begin
    DIFrame := TQUB_Acquisition_DIChannelFrame(panDigitalInput.Controls[c]);
    DCh     := TQUB_Acquisition_DChannel(prDIChannels.Objects[c]);
    DCh.Control := DIFrame;
    if not Update then Update := DCh.Update;
    DIFrame.DIChannel := TQUB_Acquisition_DIChannel(DCh);
    DIFrame.panName.Hint := '';
    i       := QUBAcqIDevices.IndexOf(DCh.DeviceName);
    if i < 0 then begin
      DCh.Device := nil;
      DIFrame.panName.Color := clRed;
      DIFrame.panName.Hint  := 'Device unassigned';
    end else begin
      DCh.Device := TQUB_Acquisition_IDevice(QUBAcqIDevices.Objects[i]);
      DIFrame.panName.Color := clBtnFace;
    end;
    DIFrame.panName.Caption  := DCh.Name;
    DIFrame.Value            := DCh.Value;
  end;
  //DOChannels
  for c := 0 to prDOChannels.Count - 1 do begin
    DOFrame := TQUB_Acquisition_DOChannelFrame(panDigitalOutput.Controls[c]);
    DCh     := TQUB_Acquisition_DChannel(prDOChannels.Objects[c]);
    DCh.Control := DOFrame;
    DOFrame.DOChannel := TQUB_Acquisition_DOChannel(DCh);
    DOFrame.panName.Hint := '';
    i       := QUBAcqODevices.IndexOf(DCh.DeviceName);
    if i < 0 then begin
      DCh.Device := nil;
      DOFrame.panName.Color := clRed;
      DOFrame.panName.Hint  := 'Device unassigned';
    end else begin
      DCh.Device := TQUB_Acquisition_ODevice(QUBAcqODevices.Objects[i]);
      DOFrame.panName.Color := clBtnFace;
    end;
    DOFrame.panName.Caption  := DCh.Name;
    DOFrame.Value            := DCh.Value;
  end;

  prTimer.Enabled := Update;

end;


//******************************************************************************
procedure TQUB_Viewport_AcquisitionSShot.prOnAcquire(Sender: TObject; Buf :TQUB_AI_PBuffer; NotLast :Boolean);
var
  BufDataSize  :Integer;
  BufDataType  :TQUB_QFS_Node_DataType;
  FBuf         :TPASingle;
  IorV         :Boolean;
  Channel      :TQUB_Acquisition_AIChannel;
  i, c         :Integer;

  procedure Average(c :Integer; Channel :TQUB_Acquisition_AIChannel);
  var
    Scaling      :Double;
    p, j         :Integer;
    Val          :Double;
  begin
    with ViewportAcqInput.AcquisitionInput do begin
      p       := Round(IntPower(2, Hardware.InputDeviceParams[Device.Index].m_Bits));
      Scaling := Channel.Gain * Channel.Scaling * p / Hardware.InputDeviceParams[Device.Index].m_Range;
      MPXDataX(FBuf, Buf.Buffer, AIChannels.Count, c, SizeOf(Single), BufDataSize, NODE_DATATYPE_SINGLE, BufDataType, 0, Buf.BufferCount-1, False, Scaling);
    end;
    Val     := 0;
    for j := 0 to Buf.BufferCount - 1 do begin
      Val   := Val + FBuf[j];
    end;
    if Buf.BufferCount > 0 then Val := Val / Buf.BufferCount;
    Channel.Value := Val;
  end;
begin
  if (prAIChannels.Count = 0) and not chkR.Checked then Exit; // and not prUpdateResistance then Exit;
  with Sender as TQUB_Acquisition_Input do begin
    BufDataSize := ((Hardware.InputDeviceParams[Device.Index].m_Bits + 7) div 8);
    if BufDataSize = 1 then begin
      BufDataType := NODE_DATATYPE_SHORTINT;
    end else if BufDataSize = 2 then begin
      BufDataType := NODE_DATATYPE_SMALLINT;
    end else if BufDataSize = 4 then begin
      BufDataType := NODE_DATATYPE_LONGINT;
    end else if BufDataSize = 8 then begin
      BufDataType := NODE_DATATYPE_INT64;
    end else begin
      raise Exception.Create(Format('SShot: Unknown Acq DataSize: %d', [BufDataSize]));
    end;
    if NotLast then Exit; // already stale
    FBuf       := SAlloc1(Buf.BufferCount);
    IorV       := False;
    try
      for i := 0 to prAIChannels.Count - 1 do begin
        Channel := TQUB_Acquisition_AIChannel(prAIChannels.Objects[i]);
        c       := TQUB_Acquisition_Input(Sender).AIChannels.IndexOfObject(Channel);
        if c >= 0 then begin
          Average(c, Channel);
          TQUB_Acquisition_AIChannelFrame(Channel.Control).Value := Channel.Value;
          IorV := (IorV or
                   (prRVChannel and (Channel.Name = prRVName)) or
                   (Channel.Name = prRIName));
        end;
      end;
      if chkR.Checked then begin
        if -1 = prAIChannels.IndexOf(prRIName) then begin
          c      := TQUB_Acquisition_Input(Sender).AIChannels.IndexOf(prRIName);
          if c >= 0 then begin
            Average(c, TQUB_Acquisition_AIChannel(TQUB_Acquisition_Input(Sender).AIChannels.Objects[c]));
            IorV := True;
          end;
        end;
        if prRVChannel and (-1 = prAIChannels.IndexOf(prRVName)) then begin
          c      := TQUB_Acquisition_Input(Sender).AIChannels.IndexOf(prRVName);
          if c >= 0 then begin
            Average(c, TQUB_Acquisition_AIChannel(TQUB_Acquisition_Input(Sender).AIChannels.Objects[c]));
            IorV := True;
          end;
        end;
        if IorV then prUpdateR(False);
      end;
    except
    end;
    FreeMem(FBuf);
  end;
end;



//******************************************************************************
procedure TQUB_Viewport_AcquisitionSShot.btnPropertiesClick(Sender: TObject);
begin
  inherited;
  SetFocus;
  prPropertiesForm.ShowModal;

  prUpdateChannels;

end;


//******************************************************************************
procedure TQUB_Viewport_AcquisitionSShot.DoPropertiesGroup(Read :Boolean; cKey :String; aChan :TQUB_Acquisition_Channels; Reg  :TRegistry; aChanLocal :TQUB_Acquisition_Channels);
var
  Idx  :Integer;
  i    :Integer;
  Ch   :TQUB_Acquisition_Channel;
begin
  with Reg do begin

    if Read then begin
    //----- Create local subset of global channel array from registry list
      i := 0;
      while OpenKey('Software\QUB\Viewport ' +  Name + cKey + IntToStr(i + 1), False) and ValueExists('Name') do begin
        Idx := aChan.IndexOf(ReadString('Name'));
        if Idx >= 0 then begin
          Ch := TQUB_Acquisition_Channel(aChan.Objects[Idx]);
          aChanLocal.AddObject(Ch.Name, Ch);
        end;
        CloseKey; // Unnecc.
        Inc(i);
      end;

    end else begin
    //----- Write local subset list to registry
      for i := 0 to aChanLocal.Count - 1 do begin
        Ch := TQUB_Acquisition_Channel(aChanLocal.Objects[i]);
        if OpenKey('Software\QUB\Viewport ' + Name + cKey + IntToStr(i + 1), True) then begin
          WriteString('Name', Ch.Name);
        end;
        CloseKey;
      end;
      //----- DeleteKey with subkeys does not work on all win32's, so specifically
      // delete the next higher (which may quietly fail) then read will not move
      // on to any others.
      DeleteKey('Software\QUB\Viewport ' + Name + cKey + IntToStr(aChanLocal.Count+1));

    end;
    CloseKey;

  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.DoPropertiesR(Read :Boolean; cKey :String; Reg  :TRegistry);
begin
  with Reg do begin

    if OpenKey('Software\QUB\Acquisition Input', True) then begin
      if Read then begin

        R_VChannel    := ReadRegistryBoolean(Reg, 'R_VChannel', False);
        R_VName       := ReadRegistryString(Reg, 'R_VName', '');
        R_VScale      := ReadRegistryFloat(Reg, 'R_VScale', 1.0); //mV
        R_IName       := ReadRegistryString(Reg, 'R_IName', '');
        R_mV          := ReadRegistryFloat(Reg, 'R_mV', 10.0);

      end else begin

        WriteBool('R_VChannel', prRVChannel);
        WriteString('R_VName', prRVName);
        WriteFloat('R_VScale', prRVScale);
        WriteString('R_IName', prRIName);
        WriteFloat('R_mV', prRmV);

      end;
      CloseKey;
    end;

  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.DoProperties(Read :Boolean);
var
  Reg  :TRegistry;
begin
  Reg  := TRegistry.Create;
  try
    DoPropertiesGroup(Read, '\Analog Input\AIChannel ',  QUBAcqAIChannels,  Reg, prAIChannels);
    DoPropertiesGroup(Read, '\Analog Output\AOChannel ',  QUBAcqAOChannels, Reg, prAOChannels);
    DoPropertiesGroup(Read, '\Digital Input\DIChannel ',  QUBAcqDIChannels, Reg, prDIChannels);
    DoPropertiesGroup(Read, '\Digital Output\DOChannel ', QUBAcqDOChannels, Reg, prDOChannels);
    DoPropertiesR(Read, '\Resistance', Reg);
  finally
    Reg.Free;
  end;
end;


procedure TQUB_Viewport_AcquisitionSShot.ConteinerClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.Label4Click(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.Label3Click(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.Label2Click(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.panAnalogInputClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.Label1Click(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.panAnalogOutputClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.panDigitalInputClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_AcquisitionSShot.panDigitalOutputClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;


//******************************************************************************
procedure TQUB_Viewport_AcquisitionSShot.txtRmVKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtRmVExit(Sender);
  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.txtRmVExit(Sender: TObject);
begin
  if txtRmV.ReadOnly then Exit;
  try
    R_mV   := StrToFloat(txtRmV.Text);
  except
    txtRmV.Text := Format('%.2g', [prRmV]);
  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.btnReadRClick(Sender: TObject);
begin
  prUpdateR(True);
end;

procedure TQUB_Viewport_AcquisitionSShot.prSetRVChannel(x :Boolean);
begin
  prRVChannel := x;
  if x then begin
    txtR.Text := '';
    txtRmV.Text := '';
  end;
  txtRmV.ReadOnly := x;
end;

procedure TQUB_Viewport_AcquisitionSShot.prSetRVName(x :String);
begin
  prRVName := x;
  if prRVChannel then begin
    txtR.Text := '';
    txtRmV.Text := '';
  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.prSetRVScale(x :Double);
var
  ReScale :Double;
begin
  ReScale := x / prRVScale;
  prRVScale := x;
  if prRVChannel then begin
    try
      if txtR.Text <> ''   then txtR.Text   := Format('%.2g', [ReScale * StrToFloat(txtR.Text)]);
      if txtRmV.Text <> '' then txtRmV.Text := Format('%.2g', [ReScale * StrToFloat(txtRmV.Text)]);
    except
      txtR.Text := '';
      txtRmV.Text := '';
    end;
  end;
end;

procedure TQUB_Viewport_AcquisitionSShot.prSetRmV(x :Double);
begin
  prRmV := x;
  txtRmV.Text := Format('%.2g', [x]);
  prUpdateR(False);
end;

procedure TQUB_Viewport_AcquisitionSShot.prSetRIName(x :String);
begin
  prRIName := x;
  txtR.Text := '';
  txtRpA.Text := '';
end;

procedure TQUB_Viewport_AcquisitionSShot.prUpdateR(ReRead :Boolean);
var
  RecChannels :TQUB_Acquisition_Channels;
  c           :Integer;
  pA          :Double;
begin
  if ViewportAcqInput = nil then Exit;
  RecChannels := ViewportAcqInput.AcquisitionInput.AIChannels;
  if prRVChannel then begin
    c := RecChannels.IndexOf(prRVName);
    if c >= 0 then begin
      if ReRead then begin
        if 0 = TQUB_Acquisition_AIChannel(RecChannels.Objects[c]).Read then begin
          prRmV :=  prRVScale * TQUB_Acquisition_AIChannel(RecChannels.Objects[c]).Value;
          txtRmV.Text := Format('%.2g', [prRmV]);
          panNameR.Hint := '';
        end else begin
          txtRmV.Text := 'err';
          panNameR.Hint := 'Failed to read V';
        end;
      end else begin
        prRmV := prRVScale * TQUB_Acquisition_AIChannel(RecChannels.Objects[c]).Value;
        txtRmV.Text := Format('%.2g', [prRmV]);
      end;
    end else begin
      panNameR.Hint := Format('Invalid V channel: %s', [prRVName]);
    end;
  end else begin
    panNameR.Hint := '';
  end;
  c := RecChannels.IndexOf(prRIName);
  if c >= 0 then begin
    if ReRead then begin
      if 0 = TQUB_Acquisition_AIChannel(RecChannels.Objects[c]).Read then begin
        pA := TQUB_Acquisition_AIChannel(RecChannels.Objects[c]).Value;
        txtRpA.Text := Format('%.2g', [pA]);
      end else begin
        txtRpA.Text := 'err';
        txtR.Text   := 'err';
        panNameR.Hint := 'Failed to read pA';
        Exit;
      end;
    end else begin
      pA := TQUB_Acquisition_AIChannel(RecChannels.Objects[c]).Value;
      txtRpA.Text := Format('%.2g', [pA]);
    end;
    if pA = 0.0 then begin
      txtR.Text := 'inf';
    end else begin
      txtR.Text := Format('%.2g', [prRmV * 1e3 / pA]);
    end;
  end else begin
    txtRpA.Text := 'err';
    txtR.Text := 'err';
    panNameR.Hint := Format('Invalid pA channel: %s', [prRIName]);
  end;
end;

// show RmV at start
// show RmV on update



// ****************************************************************************
end.
