unit QUB_Viewport_Data;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Math, Graphics, Controls, Forms, Dialogs, ComCtrls,
ExtCtrls, StdCtrls, Registry, Buttons, TeeProcs, TeEngine, Chart, Series, Menus, ImgList, Clipbrd,

AllocTypes, Alloc,
RegistryInterface,

QUB_ListBox,

QUB_Node,

QUB_Global,
QUB_Viewport_Files,
QUB_Viewport_Data_New_Properties,
QUB_QVO, QUB_QVO_QVD, QUB_QVO_QVD_Global, QUB_QVO_QVT,
QUB_QVO_QVD_XProperties, QUB_QVO_QVD_Print_Properties, QUB_QVO_QVD_SelectionProperties,
QUB_QVO_QVD_Chop_Properties, QUB_QVO_QVD_EventProcessor_Properties, QUB_QVO_QVD_GoToPoint_Properties,
QUB_QVO_QVM,
QUB_Task_EvP_Properties,
QUB_Python_Thread,
PythonEngine;


//{$include 'QUB_Viewport_Data.hct'}


type


TQUB_Viewport_Data = class(TQUB_Viewport_Files)


  panScroll                                     :TPanel;
  DisplayScrollBar                              :TScrollBar;
  splFilesResults                               :TSplitter;
  ConteinerLists                                :TPanel;
  splListsSelections                            :TSplitter;
  ConteinerListsList                            :TPanel;
  lblLists                                      :TLabel;
  ConteinerListsSelections                      :TPanel;
  lblSelections                                 :TLabel;
  panDisplay                                    :TPanel;
  splDisplayTop                                 :TSplitter;
  ConteinerZoom                                 :TPanel;
  ConteinerDisplay                              :TPanel;
  SelScrollBar                                  :TScrollBar;
  Panel2                                        :TPanel;
  ConteinerResults                              :TPanel;
  splResultsLists                               :TSplitter;
  lblResults                                    :TLabel;
  DataImageList                                 :TImageList;
  N1                                            :TMenuItem;
  DrawListOverlappedItem                        :TMenuItem;


  procedure DisplayScrollBarScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
  procedure SelScrollBarScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
  procedure OnClickGeneric(Sender: TObject);
  procedure DrawListOverlappedItemClick(Sender: TObject);
  procedure PopupMenuPopup(Sender: TObject); override;


protected


  prShortcuts                                   :TFrame;

  prNewPropertiesForm                           :TQUB_Viewport_Data_New_PropertiesForm;

  prPropertiesForm                              :TQUB_QVO_QVD_XPropertiesForm;
  prChopPropertiesForm                          :TQUB_QVO_QVD_Chop_PropertiesForm;
  prPrintPropertiesForm                         :TQUB_QVO_QVD_Print_PropertiesForm;
  prEvtListProcessorPropertiesForm              :TQUB_Task_EvP_Properties; // TQUB_QVO_QVD_EventProcessor_PropertiesForm;
  prGoToPointPropertiesForm                     :TQUB_QVO_QVD_GoToPoint_PropertiesForm;
  prSelectionPropertiesForm                     :TQUB_QVO_QVD_SelectionPropertiesForm;

  prCurrModel                                   :TQUB_QVO;

  //prEvtpPyFunc                                  :TPythonType;

  //Defaults
  prADChannelCount                              :Variant;
  prADDataSize                                  :Variant;
  prSampling                                    :Variant;
  prScaling                                     :Variant;
  prTimeColumn                                  :Variant;
  prUnevenSampling                              :Variant;
  prSinglePrec                                  :Variant;
  prSegmented                                   :Variant;

  prOpenDlg                                     :TOpenDialog;
  prOpenIdlDlg                                  :TOpenDialog;
  prOpenBaseDlg                                 :TOpenDialog;
  prOpenListDlg                                 :TOpenDialog;
  prSaveDlg                                     :TSaveDialog;
  prSaveIdlDlg                                  :TSaveDialog;
  prSaveBaseDlg                                 :TSaveDialog;
  prSaveListDlg                                 :TSaveDialog;

  
  procedure ListObjectsDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);

  function prCurrData :TQUB_QVO_QVD;

  function prGetHasData :Boolean;
  function prGetFileName :String;

  procedure DoPropertiesInternal(Read :Boolean); override;

  procedure prOnChangeDataLock(Sender :TObject);


public


  prevData                                      :TQUB_QVO_QVD;// Saves properties from close of last files for next open


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure Activate; override;
  procedure DeActivate; override;

  //File
  function New(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean; override;
  function Save(const FileName :String; Silent :Boolean) :Boolean; override;
  function SaveAs(const FileName :String; Silent :Boolean) :Boolean; override;
  function Close :Boolean; override;
  function CloseAll :Boolean; override;

  function OpenFromString(const MemFile :String) :Boolean; override;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

  function LoadIdealizedData(const FileName :String) :Boolean;
  function SaveIdealizedData(const FileName :String; EventStats :Boolean) :Boolean;
  function SaveIdealizedDataASCII(const FileName :String) :Boolean;
  procedure SaveIdealizedToClipBoard;

  function LoadBaselineNodes(const FileName :String) :Boolean;
  function SaveBaselineNodes(const FileName :String) :Boolean;

  function CanClose :Boolean; override;
  function CanCloseActive :Boolean; override;
  function CanSave :Boolean;

  function CloseData(Data :TQUB_QVO_QVD) :Boolean;
  function SaveData(Data :TQUB_QVO_QVD; const FileName :String; Silent :Boolean) :Boolean;
  function SaveAsData(Data :TQUB_QVO_QVD; const FileName :String; Silent :Boolean) :Boolean;

  function DataListSaveAs :Boolean;
  function DataListOpen :Boolean;

  function Properties :Boolean; override;

  function DataCount :Integer;

  //Tasks
  function AddSegmentList :Boolean;
  function AddIdlSegmentList :Boolean;
  function ChopSegments(Properties :Boolean) :Boolean;
  function AddList :Boolean; overload;
  function AddList(const Name  :String) :Boolean; overload;
  function DelList :Boolean;
  function AddSelection :Boolean;
  function DelSelection :Boolean;
  function ShowSelectionProperties :Boolean;
  function AddNote :Boolean;
  function EvtListProcess(Properties :Boolean; OnActionDone :TNotifyEvent) :Boolean;
  function GoToPoint(Properties :Boolean) :Boolean;
  function PlaySelection :Boolean;
  function StopSelection :Boolean;

  procedure InitPython;


  property CurrData                          :TQUB_QVO_QVD                                 read prCurrData;
  property CurrModel                         :TQUB_QVO                                     read prCurrModel                       write prCurrModel;

  property Shortcuts                         :TFrame                                       read prShortcuts;

  property HasData                           :Boolean                                      read prGetHasData;
  property FileName                          :String                                       read prGetFileName;

  property NewPropertiesForm                 :TQUB_Viewport_Data_New_PropertiesForm        read prNewPropertiesForm;
  property ChopPropertiesForm                :TQUB_QVO_QVD_Chop_PropertiesForm             read prChopPropertiesForm;
  property EvtListProcessorPropertiesForm    :TQUB_Task_EvP_Properties                     read prEvtListProcessorPropertiesForm;
  property GoToPointPropertiesForm           :TQUB_QVO_QVD_GoToPoint_PropertiesForm        read prGoToPointPropertiesForm;

  property ADChannelCount                    :Variant                                      read prADChannelCount                  write prADChannelCount;
  property ADDataSize                        :Variant                                      read prADDataSize                      write prADDataSize;
  property Sampling                          :Variant                                      read prSampling                        write prSampling;
  property Scaling                           :Variant                                      read prScaling                         write prScaling;
  property TimeColumn                        :Variant                                      read prTimeColumn                      write prTimeColumn;
  property UnevenSampling                    :Variant                                      read prUnevenSampling                  write prUnevenSampling;
  property SinglePrec                        :Variant                                      read prSinglePrec                      write prSinglePrec;
  property Segmented                         :Variant                                      read prSegmented                       write prSegmented;


end;


implementation


uses


QUB_Viewport_Data_Shortcuts,
QUB_GUIInterface,
QUB_Interface,
QUB_Node_Python,
QUB_Viewport_Python,
QUB_Python_Data, QUB_Main;



{$R *.DFM}
{$R *.RES}


//******************************************************************************
constructor TQUB_Viewport_Data.Create(AOwner :TComponent);
begin

  inherited;

  splFilesResults.Align   := alNone;

  ListObjects.Width       := 97;
  ListObjects.Height      := 70;
  ListObjects.Align       := alTop;
  ListObjects.Style       := lbOwnerDrawFixed;
  ListObjects.OnDrawItem  := ListObjectsDrawItem;

  splFilesResults.Top     := ListObjects.Top + ListObjects.Height;
  splFilesResults.Align   := alTop;

  //Image resources
  DataImageList.ResourceLoad(rtBitmap, 'DATARECORD',   clWhite);
  DataImageList.ResourceLoad(rtBitmap, 'DATAPLAYBACK', clWhite);
  DataImageList.ResourceLoad(rtBitmap, 'RESULTSHOLD',  clWhite);


  //Shortcuts
  prShortcuts                                              := TQUB_Viewport_Data_Shortcuts.Create(Self);
  Events.AddEvent('OnChangeFile', TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnChangeData);
  Events.AddEvent('OnOpenFile',   TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnOpenData);
  Events.AddEvent('OnCloseFile',  TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnCloseData);

  prPropertiesForm                                         := TQUB_QVO_QVD_XPropertiesForm.Create(Self);

  prNewPropertiesForm                                      := TQUB_Viewport_Data_New_PropertiesForm.Create(Self);

  prChopPropertiesForm                                     := TQUB_QVO_QVD_Chop_PropertiesForm.Create(Self);

  prPrintPropertiesForm                                    := TQUB_QVO_QVD_Print_PropertiesForm.Create(Self);

  prEvtListProcessorPropertiesForm                         := TQUB_Task_EvP_Properties.Create(Self);{TQUB_QVO_QVD_EventProcessor_PropertiesForm}

  prGoToPointPropertiesForm                                := TQUB_QVO_QVD_GoToPoint_PropertiesForm.Create(Self);

  prSelectionPropertiesForm                                := TQUB_QVO_QVD_SelectionPropertiesForm.Create(Self);
  QUB_GUIInterface.SelectionPropertiesForm                 := prSelectionPropertiesForm;

  PythonThread.SyncPyProc( InitPython, True, False, -1 );

  TVarData(prADChannelCount).VType := varEmpty;
  TVarData(prADDataSize).VType     := varEmpty;
  TVarData(prSampling).VType       := varEmpty;
  TVarData(prScaling).VType        := varEmpty;
  TVarData(prTimeColumn).VType     := varEmpty;
  TVarData(prUnevenSampling).VType := varEmpty;
  TVarData(prSinglePrec).VType     := varEmpty;
  TVarData(prSegmented).VType      := varEmpty;

end;

destructor TQUB_Viewport_Data.Destroy;
begin
  try

    Events.DelEvent('OnChangeFile', TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnChangeData);
    Events.DelEvent('OnOpenFile',   TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnOpenData);
    Events.DelEvent('OnCloseFile',  TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnCloseData);

  finally
    inherited;
  end;
end;

procedure TQUB_Viewport_Data.DrawListOverlappedItemClick(Sender: TObject);
begin
  inherited;
  if CurrData <> nil then begin
    CurrData.DrawListOverlapped    := not CurrData.DrawListOverlapped;
    CurrData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA, 0);
    DrawListOverlappedItem.Checked := CurrData.DrawListOverlapped;
  end;
end;

procedure TQUB_Viewport_Data.PopupMenuPopup(Sender: TObject);
begin
  inherited;
  DrawListOverlappedItem.Enabled := CurrData <> nil;
  if CurrData <> nil then begin
    DrawListOverlappedItem.Checked := CurrData.DrawListOverlapped;
  end else begin
    DrawListOverlappedItem.Checked := False;
  end;
end;

//Activate/DeActivate **********************************************************
procedure TQUB_Viewport_Data.Activate;
begin
  inherited;
  if CurrData <> nil then if not CurrData.Active then CurrData.Activate;
end;

procedure TQUB_Viewport_Data.DeActivate;
begin
  inherited;
  if CurrData <> nil then if CurrData.Active then CurrData.DeActivate;
end;

//****************************************************************************//
//                                                                            //
//                         F I L E     M E T H O D S                          //
//                                                                            //
//****************************************************************************//
function TQUB_Viewport_Data.New(const FileName :String; Silent :Boolean) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  Zoom     :TQUB_QVO_QVT;
  f        :String;
begin

  //Process FileName;
  f                                    := FileName;
  if f = '' then f                     := GetBaseStampedFileName(ExpandFileName(prNewPropertiesForm.BaseName), prNewPropertiesForm.FileTypeToExt(prNewPropertiesForm.FileFormat));
  //if f = '' then f                     := GetNewUniqueFileName(ExpandFileName('noname'), prNewPropertiesForm.FileTypeToExt(prNewPropertiesForm.FileFormat), prNoNameCount);

  //Create new Data
  Data                                 := TQUB_QVO_QVD.Create(Self);
  Zoom                                 := TQUB_QVO_QVT.Create(Data);
  Data.Parent                          := ConteinerDisplay;

  Data.ScrollBar                       := DisplayScrollBar;
  Data.Zoom                            := Zoom;
  Data.ConteinerList                   := ConteinerListsList;
  Data.ConteinerSel                    := ConteinerListsSelections;
  Data.ConteinerResults                := ConteinerResults;
  Zoom.Parent                          := ConteinerZoom;
  Zoom.PopUpMenu                       := Data.PopUpMenu;

  Data.PropertiesForm                  := prPropertiesForm;
  Data.ChopPropertiesForm              := prChopPropertiesForm;
  Data.PrintPropertiesForm             := prPrintPropertiesForm;
  Data.EvtListProcessorPropertiesForm  := prEvtListProcessorPropertiesForm;
  Data.GoToPointPropertiesForm         := prGoToPointPropertiesForm;
  Data.SelectionPropertiesForm         := prSelectionPropertiesForm;

  Data.ResultsList.DataImageList       := DataImageList;

  Data.TipsOn                          := prTipsOn;
  Data.Events.AddEvent('OnActiveChannelChange', TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnActiveChannelChange);

  prNewPropertiesForm.FileName         := f;

  prNewPropertiesForm.OnGetProperties  := Data.DataObject.OnGetPropertiesNew;
  prNewPropertiesForm.OnPutProperties  := Data.DataObject.OnPutPropertiesNew;

  if Silent then begin
    prNewPropertiesForm.FormShow(prNewPropertiesForm);
    prNewPropertiesForm.ModalResult := mrOK;
    prNewPropertiesForm.btnOKClick(prNewPropertiesForm.btnOK);
  end else begin
    prNewPropertiesForm.ShowModal;
  end;

  Result := prNewPropertiesForm.ModalResult = mrCancel;

  if not Result then begin
    f := Data.DataObject.FileName;
    Data.DataObject.OpenSilently := Silent;
    Result := Data.New(f);
  end;

  if not Result then begin
    ListObjects.AddQVO(Data, ExtractFileName(f));
    DoOnOpenFile(Data);
    DoOnChangeFile(Data);
    prAddFileEvents(Data);
    Data.OnChangeLockRecord   := prOnChangeDataLock;
    Data.OnChangeLockPlayback := prOnChangeDataLock;
    Data.ExpCond.ReadTree( prNewPropertiesForm.ExpCond );
  end else begin
    Data.OnChangeLockRecord   := nil;
    Data.OnChangeLockPlayback := nil;
    prDelFileEvents(Data);
    Data.Free;
  end;

  DisplayScrollBar.Enabled := ListObjects.Items.Count <> 0;
  DisplayScrollBar.Max     := DisplayScrollBar.Min;
  SelScrollBar.Enabled     := ListObjects.Items.Count <> 0;

end;

function TQUB_Viewport_Data.Open(const FileName :String; Silent :Boolean) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  Zoom     :TQUB_QVO_QVT;
  //Dlg      :TOpenDialog;
  Count    :Integer;
  f, fi    :String;
  d        :Integer;
begin

  Result := False;

  //Process FileName;
  try

    f := FileName;
    if f = '' then begin
      if prOpenDlg = nil then prOpenDlg := TOpenDialog.Create(Application);
      prOpenDlg.Filter := 'All QUB data files (*.qdf;*.dat;*.ldt;*.txt;*.abf;*.wav;*.pul;*.acquire;*.dwt;*.evl;*.prt;*.tacidl;*.qsf;*.ibw)|' +
                                              '*.qdf;*.dat;*.ldt;*.txt;*.abf;*.wav;*.pul;*.acquire;*.dwt;*.evl;*.prt;*.tacidl;*.qsf;*.ibw|'  +
                          'QUB data files (*.qdf)|*.qdf|'            +
                          'QUB old binary files (*.dat)|*.dat|'      +
                          'QUB old ldt files (*.ldt)|*.ldt|'         +
                          'ASCII data files (*.txt)|*.txt|'          +
                          'ABF files (*.abf)|*.abf|'                 +
                          'WAVE sound files (*.wav)|*.wav|'          +
                          'HEKA Pulse files (*.pul)|*.pul|'          +
                          'Bruxton Acquire files (*.acquire)|*.acquire|' +
                          'QUB idealized data files (*.dwt)|*.dwt|'  +
                          'Axon idealized data files (*.evl)|*.evl|' +
                          'EKDIST printout files (*.prt)|*.prt|'     +
                          'TacFit transition table (*.tacidl)|*.tacidl|' +
                          'QUB session files (*.qsf)|*.qsf|'         +
                          'IGOR binary wave(*.ibw)|*.ibw|'           +
                          'All files (*.*)|*.*';
      prOpenDlg.Options := [ofPathMustExist, ofFileMustExist, ofAllowMultiSelect, ofEnableSizing];
      if not prOpenDlg.Execute then begin
        Result := True;
        Exit;
      end;
    end;

    if f <> '' then begin
      Count := 1;
    end else begin
      Count := prOpenDlg.Files.Count;
    end;

    for d := 0 to Count - 1 do begin

      if f <> '' then begin
        fi := f;
      end else begin
        if d < (Count - 1) then begin
          fi := prOpenDlg.Files[d+1];
        end else begin
          fi := prOpenDlg.Files[0];
        end;
      end;

      //Create new Data and open the file

      Data                                 := TQUB_QVO_QVD.Create(Self);
      Zoom                                 := TQUB_QVO_QVT.Create(Data);
      Data.Parent                          := ConteinerDisplay;

      Data.ScrollBar                       := DisplayScrollBar;
      Data.Zoom                            := Zoom;
      Data.ConteinerList                   := ConteinerListsList;
      Data.ConteinerSel                    := ConteinerListsSelections;
      Data.ConteinerResults                := ConteinerResults;
      Zoom.Parent                          := ConteinerZoom;
      Zoom.PopUpMenu                       := Data.PopUpMenu;

      Data.PropertiesForm                  := prPropertiesForm;
      Data.ChopPropertiesForm              := prChopPropertiesForm;
      Data.PrintPropertiesForm             := prPrintPropertiesForm;
      Data.EvtListProcessorPropertiesForm  := prEvtListProcessorPropertiesForm;
      Data.GoToPointPropertiesForm         := prGoToPointPropertiesForm;
      Data.SelectionPropertiesForm         := prSelectionPropertiesForm;
 
      Data.ResultsList.DataImageList       := DataImageList;

      Data.TipsOn                          := prTipsOn;
      Data.Events.AddEvent('OnActiveChannelChange', TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnActiveChannelChange);

      Data.DataObject.OpenSilently         := Silent;
      if TVarData(prADChannelCount).VType  <> varEmpty then Data.DataObject.ADChannelCount     := prADChannelCount;
      if TVarData(prADDataSize).VType      <> varEmpty then Data.DataObject.ADDataSize         := prADDataSize;
      if TVarData(prSampling).VType        <> varEmpty then Data.DataObject.Sampling           := prSampling;
      if TVarData(prScaling).VType         <> varEmpty then Data.DataObject.Scaling            := prScaling;
      if TVarData(prTimeColumn).VType      <> varEmpty then Data.DataObject.OpenTimeColumn     := prTimeColumn;
      if TVarData(prUnevenSampling).VType  <> varEmpty then Data.DataObject.OpenUnevenSampling := prUnevenSampling;
      if TVarData(prSinglePrec).VType      <> varEmpty then Data.DataObject.OpenSinglePrec     := prSinglePrec;
      if TVarData(prSegmented).VType       <> varEmpty then Data.DataObject.OpenSegmented      := prSegmented;

      Result                               := Data.Open(fi);

      if not Result then begin
        ListObjects.AddQVO(Data, ExtractFileName(fi));
        DoOnOpenFile(Data);
        DoOnChangeFile(Data);
        prAddFileEvents(Data);
        Data.OnChangeLockRecord   := prOnChangeDataLock;
        Data.OnChangeLockPlayback := prOnChangeDataLock;

        {
        //Testing - to make it remember settings, e.g., when opening multiple files
        
        //TO DO:
        // Lorin, does it behave like you like with my changes in qdf?
        // I can't tell what this is about here.  -- Chris

        //!!!! revert to old code and find a better way. This overrides the settings for files that have qsf !!!
        prADChannelCount          := Data.DataObject.ADChannelCount;
        prADDataSize              := Data.DataObject.ADDataSize;
        prSampling                := Data.DataObject.Sampling;
        prScaling                 := Data.DataObject.Scaling;
        prTimeColumn              := Data.DataObject.OpenTimeColumn;
        prUnevenSampling          := Data.DataObject.OpenUnevenSampling;
        prSinglePrec              := Data.DataObject.OpenSinglePrec;
        prSegmented               := Data.DataObject.OpenSegmented;
        }

      end else begin

        //Testing
        {
        TVarData(prADChannelCount).VType := varEmpty;
        TVarData(prADDataSize).VType     := varEmpty;
        TVarData(prSampling).VType       := varEmpty;
        TVarData(prScaling).VType        := varEmpty;
        TVarData(prTimeColumn).VType     := varEmpty;
        TVarData(prUnevenSampling).VType := varEmpty;
        TVarData(prSinglePrec).VType     := varEmpty;
        TVarData(prSegmented).VType      := varEmpty;
        }

        Data.OnChangeLockRecord   := nil;
        Data.OnChangeLockPlayback := nil;
        prDelFileEvents(Data);
        Data.Free;
        if not Silent then MessageHelpDlg('Could not open file ' + fi, mtError, [mbOK], 0, 'Error:Open_Data');
      end;

      DisplayScrollBar.Enabled := ListObjects.Items.Count <> 0;
      DisplayScrollBar.Max     := DisplayScrollBar.Min;
      SelScrollBar.Enabled     := ListObjects.Items.Count <> 0;

    end;


  finally

    //Testing - to remember settings for opening multiple files
    //{
    TVarData(prADChannelCount).VType := varEmpty;
    TVarData(prADDataSize).VType     := varEmpty;
    TVarData(prSampling).VType       := varEmpty;
    TVarData(prScaling).VType        := varEmpty;
    TVarData(prTimeColumn).VType     := varEmpty;
    TVarData(prUnevenSampling).VType := varEmpty;
    TVarData(prSinglePrec).VType     := varEmpty;
    TVarData(prSegmented).VType      := varEmpty;
    //}

  end;

end;

function TQUB_Viewport_Data.Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  Zoom     :TQUB_QVO_QVT;
begin

  try

    //Create new Data and open the file
    Data                   := TQUB_QVO_QVD.Create(Self);
    Zoom                   := TQUB_QVO_QVT.Create(Data);
    Data.Parent            := ConteinerDisplay;

    Data.ScrollBar         := DisplayScrollBar;
    Data.Zoom              := Zoom;
    Data.ConteinerList     := ConteinerListsList;
    Data.ConteinerSel      := ConteinerListsSelections;
    Data.ConteinerResults  := ConteinerResults;
    Zoom.Parent            := ConteinerZoom;
    Zoom.PopUpMenu         := Data.PopUpMenu;

    Data.PropertiesForm                  := prPropertiesForm;
    Data.ChopPropertiesForm              := prChopPropertiesForm;
    Data.PrintPropertiesForm             := prPrintPropertiesForm;
    Data.EvtListProcessorPropertiesForm  := prEvtListProcessorPropertiesForm;
    Data.GoToPointPropertiesForm         := prGoToPointPropertiesForm;
    Data.SelectionPropertiesForm         := prSelectionPropertiesForm;

    Data.ResultsList.DataImageList       := DataImageList;

    Data.TipsOn                          := prTipsOn;
    Data.Events.AddEvent('OnActiveChannelChange', TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnActiveChannelChange);

    Data.DataObject.OpenSilently         := Silent;
    if TVarData(prADChannelCount).VType  <> varEmpty then Data.DataObject.ADChannelCount      := prADChannelCount;
    if TVarData(prADDataSize).VType      <> varEmpty then Data.DataObject.ADDataSize          := prADDataSize;
    if TVarData(prSampling).VType        <> varEmpty then Data.DataObject.Sampling            := prSampling;
    if TVarData(prScaling).VType         <> varEmpty then Data.DataObject.Scaling             := prScaling;
    if TVarData(prTimeColumn).VType      <> varEmpty then Data.DataObject.OpenTimeColumn      := prTimeColumn;
    if TVarData(prUnevenSampling).VType  <> varEmpty then Data.DataObject.OpenUnevenSampling  := prUnevenSampling;
    if TVarData(prSinglePrec).VType      <> varEmpty then Data.DataObject.OpenSinglePrec      := prSinglePrec;
    if TVarData(prSegmented).VType       <> varEmpty then Data.DataObject.OpenSegmented       := prSegmented;

    Result                               := Data.Open(FileName, OpenStream);

    if not Result then begin
      ListObjects.AddQVO(Data, ExtractFileName(FileName));
      DoOnOpenFile(Data);
      DoOnChangeFile(Data);
      prAddFileEvents(Data);
      Data.OnChangeLockRecord   := prOnChangeDataLock;
      Data.OnChangeLockPlayback := prOnChangeDataLock;
    end else begin
      Data.OnChangeLockRecord   := nil;
      Data.OnChangeLockPlayback := nil;
      prDelFileEvents(Data);
      Data.Free;
      if not Silent then MessageHelpDlg('Could not open file ' + FileName, mtError, [mbOK], 0, 'Error:Open_Data');
    end;

    DisplayScrollBar.Enabled := ListObjects.Items.Count <> 0;
    DisplayScrollBar.Max     := DisplayScrollBar.Min;
    SelScrollBar.Enabled     := ListObjects.Items.Count <> 0;

  finally

    TVarData(prADChannelCount).VType := varEmpty;
    TVarData(prADDataSize).VType     := varEmpty;
    TVarData(prSampling).VType       := varEmpty;
    TVarData(prScaling).VType        := varEmpty;
    TVarData(prTimeColumn).VType     := varEmpty;
    TVarData(prUnevenSampling).VType := varEmpty;
    TVarData(prSinglePrec).VType     := varEmpty;
    TVarData(prSegmented).VType      := varEmpty;

  end;

end;



function TQUB_Viewport_Data.Save(const FileName :String; Silent :Boolean) :Boolean;
var
  Data  :TQUB_QVO_QVD;
begin

  Data   := CurrData;
  Result := Data = nil;
  if Result then Exit;

  Data.DataObject.SaveSilently := Silent;
  Data.DataObject.OpenSilently := Silent;
  Result := SaveData(Data, FileName, Silent);

end;

function TQUB_Viewport_Data.SaveAs(const FileName :String; Silent :Boolean) :Boolean;
var
  Data     :TQUB_QVO_QVD;
begin

  Data    := CurrData;
  Result  := Data = nil;
  if Result then Exit;

  Data.DataObject.SaveSilently := Silent;
  Data.DataObject.OpenSilently := Silent;
  Result := SaveAsData(Data, FileName, Silent);

end;

function TQUB_Viewport_Data.Close :Boolean;
var
  Data     :TQUB_QVO_QVD;
begin
  Result := False;

  Data   := CurrData;
  if Data = nil then Exit;

  Result := CloseData(Data);

end;


function TQUB_Viewport_Data.SaveData(Data :TQUB_QVO_QVD; const FileName :String; Silent :Boolean) :Boolean;
var
  FName :String;
begin

  Result := (Data.LockedAny) or (Data.LockedRecord) or (Data.LockedPlayback);
  if Result then Exit;

  //Process FileName
  FName := FileName;
  if FName = '' then FName := Data.FileName;
  if ((Data.Saved) or (Silent)) and (Data.CanSaveAs(FName, True)) then begin
    Data.DataObject.SaveSilently := Silent;
    Data.DataObject.OpenSilently := Silent;
    Result := Data.Save(FName);
    if Data.DataObject.Opened then begin
      ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Data)] := ExtractFileName(Data.FileName);
      DoOnChangeFile(Data);
    end else begin
      Result := True;
      CloseData(Data);
    end;
  end else begin
    Result := SaveAsData(Data, FName, Silent);
  end;
end;

function TQUB_Viewport_Data.SaveAsData(Data :TQUB_QVO_QVD; const FileName :String; Silent :Boolean) :Boolean;
var
  f        :String;
  r        :Integer;
begin

  Result := (Data.LockedAny) or (Data.LockedRecord) or (Data.LockedPlayback);
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := ChangeFileExt(Data.FileName, '.qdf');

  if prSaveDlg = nil then prSaveDlg := TSaveDialog.Create(Application);
  prSaveDlg.Filter := 'All QUB data files (*.qdf;*.ldt;*.dat;*.wav;*.txt;*.dwt)|*.qdf;*.ldt;*.dat;*.wav;*.txt;*.dwt|' +
                'QUB data files (*.qdf)|*.qdf|' +
                'QUB old ldt files (*.ldt)|*.ldt|' +
                'QUB old binary files (*.dat)|*.dat|' +
                'WAVE sound files (*.wav)|*.wav|' +
                'ASCII data files (*.txt)|*.txt|' +
                'DWT files (*.dwt)|*.dwt|' +
                'All files (*.*)|*.*';

  if Data.FileExt = '.ldt' then begin
    prSaveDlg.FilterIndex := 3;
  end else if Data.FileExt = '.dat' then begin
    prSaveDlg.FilterIndex := 4;
  end else if Data.FileExt = '.wav' then begin
    prSaveDlg.FilterIndex := 5;
  end else if Data.FileExt = '.txt' then begin
    prSaveDlg.FilterIndex := 6;
  end else if Data.FileExt = '.dwt' then begin
    prSaveDlg.FilterIndex := 7;
  end else begin
    prSaveDlg.FilterIndex := 2; // default save as qdf
  end;

  prSaveDlg.Options := [ofPathMustExist{, ofOverwritePrompt}, ofEnableSizing];
  prSaveDlg.Title := 'Save data file as';

  try

    repeat

      prSaveDlg.FileName := ChangeFileExt(f, '');
      if prSaveDlg.Execute then begin
        f := prSaveDlg.FileName;
        case prSaveDlg.FilterIndex of
          2: f := ChangeFileExt(f, '.qdf');
          3: f := ChangeFileExt(f, '.ldt');
          4: f := ChangeFileExt(f, '.dat');
          5: f := ChangeFileExt(f, '.wav');
          6: f := ChangeFileExt(f, '.txt');
          7: f := ChangeFileExt(f, '.dwt');
        else
          f := ChangeFileExt(f, Data.FileExt);
        end;//case FilterIndex
      end else begin
        Result := True;
        Exit;
      end;

      if (FileExists(f)) and (f <> ExpandFileName(Data.FileName)) then begin
        case MessageDlg('File ''' + f + ''' exists. Do you want to overwrite it ?', mtWarning, [mbYes, mbNo, mbCancel], 0) of
          mrCancel:
            begin
              Result := True;
              Exit;
            end;
          mrNo:
            Continue;
        end;//case
      end;

      Data.DataObject.SaveSilently := Silent;
      Data.DataObject.OpenSilently := Silent;
      if Data.CanSaveAs(f, True){Ask confirmation for losing features} then begin
        Result := Data.Save(f);
        Exit;
      end else begin
        r := MessageHelpDlg('Data file cannot be saved in the selected format.'#13#10'Please choose one of the writable formats.', mtError, [mbOK, mbCancel], 0, 'Error:Data_File_Format');
        Result := r = mrCancel;
        if Result then Exit;
      end;

    until (Result);

  finally

    if Data.DataObject.Opened then begin
      ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Data)] := ExtractFileName(Data.FileName);
      DoOnChangeFile(Data);
    end else begin
      Result := True;
      CloseData(Data);
    end;

  end;

end;


function TQUB_Viewport_Data.CloseData(Data :TQUB_QVO_QVD) :Boolean;
var
  Zoom     :TQUB_QVO_QVT;
begin
  Zoom     := Data.Zoom;

  // if closing last data, save properties for next open ( Tony request ... ) 
{  if ListObjects.Items.Count = 1 Then begin //Objects[ListObjects.ItemIndex]);
    if prevData = nil Then
      prevData := TQUB_QVO_QVD.Create(self);
    prevData.TracesPerPage := Data.TracesPerPage;
  end;
}

  DoOnClosingFile(Data);  // TQUB_Viewport_Files.

  Result := Data.Close;
  if Result then Exit;
  Data.Closing := True;

  DoOnCloseFile(Data);    //TQUB_Viewport_Files.

  prDelFileEvents(Data);

  if Data = CurrData then DoOnChangeFile(nil);

  Data.Zoom := nil;
  Zoom.Free;
  ListObjects.DelQVO(Data);

  Data.Free;

  DisplayScrollBar.Enabled := ListObjects.Items.Count <> 0;
  DisplayScrollBar.Max     := DisplayScrollBar.Min;
  SelScrollBar.Enabled     := ListObjects.Items.Count <> 0;

end;


function TQUB_Viewport_Data.CloseAll :Boolean;
begin
  Result := False;
  while CurrData <> nil do begin
    Result := Close;
    if Result then Break;
  end;
end;

function TQUB_Viewport_Data.CanCloseActive :Boolean;
var
  Data :TQUB_QVO_QVD;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  Data   := CurrData;
  if Data = nil then Exit;

  if          Data.LockedRecord then begin
    MessageHelpDlg('File ''' + Data.FileName + ''' is being recorded to and cannot be closed (stop acquisition)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end else if Data.LockedPlayback then begin
    MessageHelpDlg('File ''' + Data.FileName + ''' is being played back and cannot be closed (stop acquisition)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end else if Data.LockedClose then begin
    MessageHelpDlg('File ''' + Data.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end else if Data.LockedAny then begin
    MessageHelpDlg('File ''' + Data.FileName + ''' is executing a task and cannot be closed (check TaskView)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end;

  if not Result then Exit;

  if Data.Modified then begin
    case MessageDlg('File ''' + Data.FileName + ''' is modified.'#13#10'Do you want to save it ?', mtConfirmation, [mbYes, mbNo, mbCancel], 0) of
      mrCancel:
      //Cancel
        begin
          Result := False;
          Exit;
        end;

      mrYes:
      //Yes
        begin
          if Data.Saved then begin
            Result := not Data.Save(Data.FileName);
          end else begin
            Result := not SaveAs('', False{Silent});
          end;
        end;


      mrNo:
      //No
        begin
          Data.DataObject.DiscardChanges := True;
        end;

    end;//case r
  end;

end;

function TQUB_Viewport_Data.CanClose :Boolean;
var
  ada :TQUB_QVO_QVD;
  qvd :TQUB_QVO_QVD;
  All :Boolean;
  None :Boolean;
  i   :Integer;
  r   :Integer;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  All := False;
  None := False;

  ada := CurrData;

  try

    for i := ListObjects.Items.Count - 1 downto 0 do begin

      qvd := TQUB_QVO_QVD(ListObjects[i]);

      if          qvd.LockedRecord then begin
        MessageHelpDlg('File ''' + qvd.FileName + ''' is being recorded to and cannot be closed (stop acquisition)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end else if qvd.LockedPlayback then begin
        MessageHelpDlg('File ''' + qvd.FileName + ''' is being played back and cannot be closed (stop acquisition)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end else if qvd.LockedClose then begin
        MessageHelpDlg('File ''' + qvd.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end else if qvd.LockedAny then begin
        MessageHelpDlg('File ''' + qvd.FileName + ''' is executing a task and cannot be closed (check TaskView)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end;

      if not Result then Exit;

      if qvd.Modified then begin

        ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(qvd);
        qvd.Activate;

        if All then begin

          if qvd.Saved then begin
            Result := not qvd.Save(qvd.FileName);
          end else begin
            Result := not SaveAs('', False{Silent});
          end;
          if not Result then Exit;

        end else if None then begin

          qvd.DataObject.DiscardChanges := True;

        end else begin

          r := MessageDlg('File ''' + qvd.FileName + ''' is modified.'#13#10'Do you want to save it ?', mtConfirmation, [mbYes, mbNo, mbYesToAll, mbNoToAll, mbCancel], 0);

          case r of

            mrCancel:
            //Cancel
              begin
                Result := False;
                Exit;
              end;

            mrYes:
            //Yes
              begin
                if qvd.Saved then begin
                  Result := not qvd.Save(qvd.FileName);
                end else begin
                  Result := not SaveAs('', False{Silent});
                end;
                if not Result then Exit;
              end;

            mrNo:
            //No
              begin
                qvd.DataObject.DiscardChanges := True;
              end;

            mrNotoAll://9
            //NotoAll
              begin
                None := True;
                qvd.DataObject.DiscardChanges := True;
              end;

            mrYestoAll://10:
            //YestoAll
              begin
                All := True;
                if qvd.Saved then begin
                  Result := not qvd.Save(qvd.FileName);
                end else begin
                  Result := not SaveAs('', False{Silent});
                end;
                if not Result then Exit;
              end;

          end;//case r

        end;//All

      end;

    end;

  finally

    if ada <> nil then begin
      ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(ada);
      if ListObjects.ItemIndex >= 0 then ada.Activate;
    end;

  end;

end;

function TQUB_Viewport_Data.CanSave :Boolean;
begin
  if CurrData <> nil then Result := CurrData.Modified else Result := False;
end;

function TQUB_Viewport_Data.OpenFromString(const MemFile :String) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  Zoom     :TQUB_QVO_QVT;
  f        :String;
begin
  Result := False;

  //Process FileName;
  f := MemFile;
  if f = '' then Exit;

  //Create new Data and open the file

  Data                   := TQUB_QVO_QVD.Create(Self);
  Zoom                   := TQUB_QVO_QVT.Create(Data);
  Data.Parent            := ConteinerDisplay;

  Data.ScrollBar         := DisplayScrollBar;
  Data.Zoom              := Zoom;
  Data.ConteinerList     := ConteinerListsList;
  Data.ConteinerSel      := ConteinerListsSelections;
  Data.ConteinerResults  := ConteinerResults;
  Zoom.Parent            := ConteinerZoom;
  Zoom.PopUpMenu         := Data.PopUpMenu;

  Data.ResultsList.DataImageList       := DataImageList;

  Data.TipsOn            := prTipsOn;
  Data.Events.AddEvent('OnActiveChannelChange', TQUB_Viewport_Data_Shortcuts(prShortcuts).DoOnActiveChannelChange);

  Data.PropertiesForm    := prPropertiesForm;

  Result                 := Data.OpenFromString(f);

  if not Result then begin
    ListObjects.AddQVO(Data, ExtractFileName(f));
    DoOnOpenFile(Data);
    DoOnChangeFile(Data);
    prAddFileEvents(Data);
    Data.OnChangeLockRecord   := prOnChangeDataLock;
    Data.OnChangeLockPlayback := prOnChangeDataLock;
  end else begin
    Data.OnChangeLockRecord   := nil;
    Data.OnChangeLockPlayback := nil;
    prDelFileEvents(Data);
    Data.Free;
    MessageHelpDlg('Could not open file ' + f, mtError, [mbOK], 0, 'Error:Open_Data');
  end;

  DisplayScrollBar.Enabled := ListObjects.Items.Count <> 0;
  DisplayScrollBar.Max     := DisplayScrollBar.Min;
  SelScrollBar.Enabled     := ListObjects.Items.Count <> 0;

end;

function TQUB_Viewport_Data.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.SaveToString(MemFile, FileType);
end;


//Idealized Data
function TQUB_Viewport_Data.LoadIdealizedData(const FileName :String) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  f        :String;
begin

  Data    := CurrData;
  Result  := Data = nil;
  if Result then Exit;
  Result := Data.LockedAny;
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Data.FileName;

  f := ChangeFileExt(f, '.dwt');

  if prOpenIdlDlg = nil then prOpenIdlDlg := TOpenDialog.Create(Application);
  prOpenIdlDlg.DefaultExt := 'dwt';
  prOpenIdlDlg.Filter := 'QUB dwell time files (*.dwt)|*.dwt|All files (*.*)|*.*';
  prOpenIdlDlg.Options := [ofPathMustExist, ofFileMustExist, ofEnableSizing];
  prOpenIdlDlg.Title := 'Load idealized data file';

  prOpenIdlDlg.FileName := ChangeFileExt(f, '');
  if prOpenIdlDlg.Execute then begin
    f := ChangeFileExt(prOpenIdlDlg.FileName, '.dwt');
    Result := Data.LoadInfo(Data.ActiveChannel, f);
  end else begin
    Result := True;
  end;

end;

function TQUB_Viewport_Data.SaveIdealizedData(const FileName :String; EventStats :Boolean) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  f        :String;
begin

  Data    := CurrData;
  Result  := Data = nil;
  if Result then Exit;
  Result := Data.LockedAny;
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Data.FileName;

  f := ChangeFileExt(f, '.dwt');

  if prSaveIdlDlg = nil then prSaveIdlDlg := TSaveDialog.Create(Application);
  prSaveIdlDlg.DefaultExt := 'dwt';
  if EventStats then begin
    prSaveIdlDlg.Filter := 'Text files (*.txt)|*.txt|All files (*.*)|*.*';
  end else begin
    prSaveIdlDlg.Filter := 'QUB dwell time files (*.dwt)|*.dwt|All files (*.*)|*.*';
  end;
  prSaveIdlDlg.Options := [ofPathMustExist, ofOverwritePrompt];
  prSaveIdlDlg.Title := 'Save idealized data file as';

  prSaveIdlDlg.FileName := ChangeFileExt(f, '');
  if prSaveIdlDlg.Execute then begin
    if EventStats then begin
      f := ChangeFileExt(prSaveIdlDlg.FileName, '.txt');
      Result := Data.SaveInfoStat(Data.ActiveChannel, f);
    end else begin
      f := ChangeFileExt(prSaveIdlDlg.FileName, '.dwt');
      Result := Data.SaveInfo(Data.ActiveChannel, f);
    end;
  end else begin
    Result := True;
  end;

end;

//------------------------------------------------------------------------------
procedure TQUB_Viewport_Data.SaveIdealizedToClipBoard;
var
  Model        :TQUB_QVO_QVM;
  Data         :TQUB_QVO_QVD;
  seg, idl     :INode;
  cText        :String;
  iRow, iRows  :Integer;
  aClasses     :TPAInteger;
  aDurations   :TPASingle;
begin

  Model        := ViewportModel.CurrModel;
  Data         := ViewportData.CurrData;

  if (Model <> nil) and (Data <> nil) then begin
    idl := Data.GetChosenIdl(False);
    seg := idl.Find('Segment');
    cText := '';
    while not seg.IsNull do begin           // For each segment of data
      iRows      := seg['DwellCount'].Data[0];
      aClasses   := TPAInteger( seg['Classes'].DataPtr );
      aDurations := TPASingle( seg['Durations'].DataPtr );

      for iRow := 0 to iRows - 1 do begin
        cText := cText + Format('%10.5f'#9'%8.3f'#13#10, [ aDurations[iRow]/1000.0, Model.ModelObject.Amp[aClasses[iRow]] ]);
      end;
      seg := seg.SiblingSameName;
    end;
    Clipboard.SetTextBuf(PChar(cText));
  end;

end;


//------------------------------------------------------------------------------
// Accessed from main menu.  Get a filename, save idealized data to ascii file.
function TQUB_Viewport_Data.SaveIdealizedDataASCII(const FileName :String) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  f        :String;
begin

  //----- Check Data
  Data    := CurrData;
  Result  := Data = nil;
  if Result then Exit;
  Result := Data.LockedAny;
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Data.FileName;
  f := ChangeFileExt(f, '.txt');

  //----- Save as Filename dialogue
  if prSaveIdlDlg = nil then prSaveIdlDlg := TSaveDialog.Create(Application);
  prSaveIdlDlg.DefaultExt := 'txt';
  prSaveIdlDlg.Filter := 'QUB ASCII files (*.txt)|*.txt|All files (*.*)|*.*';
  prSaveIdlDlg.Options := [ofPathMustExist, ofOverwritePrompt, ofEnableSizing];
  prSaveIdlDlg.Title := 'Save idealized data file as';

  //----- Execute dialogue, save if True.
  // Note - changes file extension to .txt, even if user supplied something else.

  prSaveIdlDlg.FileName := ChangeFileExt(f, '');
  if prSaveIdlDlg.Execute then begin
    f := ChangeFileExt(prSaveIdlDlg.FileName, '.txt');
    Result := Data.SaveInfoASCII(Data.ActiveChannel, f);
  end else begin
    Result := True;
  end;

end;


//Baseline Nodes
function TQUB_Viewport_Data.LoadBaselineNodes(const FileName :String) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  f        :String;
begin

  Data    := CurrData;
  Result  := Data = nil;
  if Result then Exit;
  Result := Data.LockedAny;
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Data.FileName;

  f := ChangeFileExt(f, '.qbn');

  if prOpenBaseDlg = nil then prOpenBaseDlg := TOpenDialog.Create(Application);
  prOpenBaseDlg.DefaultExt := 'qbn';
  prOpenBaseDlg.Filter := 'All QUB baseline nodes files (*.qbn;*.qbf)|*.qbn;*.qbf|' +
                'QUB old baseline nodes files (*.qbn)|*.qbn|' +
                'QUB baseline nodes files (*.qbf)|*.qbf|' +
                'All files (*.*)|*.*';
  prOpenBaseDlg.Options := [ofPathMustExist, ofFileMustExist, ofEnableSizing];
  prOpenBaseDlg.Title := 'Load baseline nodes file';

  prOpenBaseDlg.FileName := ChangeFileExt(f, '');
  if prOpenBaseDlg.Execute then begin
    f := prOpenBaseDlg.FileName;
    Result := Data.LoadBaselineNodes(Data.ActiveChannel, f);
  end else begin
    Result := True;
  end;

end;

function TQUB_Viewport_Data.SaveBaselineNodes(const FileName :String) :Boolean;
var
  Data     :TQUB_QVO_QVD;
  f        :String;
begin

  Data    := CurrData;
  Result  := Data = nil;
  if Result then Exit;
  Result := Data.LockedAny;
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Data.FileName;

  f := ChangeFileExt(f, '.qbf');

  if prSaveBaseDlg = nil then prSaveBaseDlg := TSaveDialog.Create(Application);
  prSaveBaseDlg.DefaultExt := 'qbf';//'qbn';LM.
  prSaveBaseDlg.Filter := 'QUB baseline nodes files (*.qbf)|*.qbf|All files (*.*)|*.*';
  prSaveBaseDlg.Options := [ofPathMustExist, ofOverwritePrompt, ofEnableSizing];
  prSaveBaseDlg.Title := 'Save baseline nodes file as';

  prSaveBaseDlg.FileName := ChangeFileExt(f, '');
  if prSaveBaseDlg.Execute then begin
    f := ChangeFileExt(prSaveBaseDlg.FileName, '.qbf');
    Result := Data.SaveBaselineNodes(Data.ActiveChannel, f);
  end else begin
    Result := True;
  end;

end;


//Data Lists
function TQUB_Viewport_Data.DataListSaveAs :Boolean;
var
  FilesNode :INode;
  FileNode  :INode;
  f         :String;
  c         :Integer;
begin

  Result := False;

  if prSaveListDlg = nil then prSaveListDlg := TSaveDialog.Create(Application);
  prSaveListDlg.Title := 'Save data list';
  prSaveListDlg.Filter := 'QUB data list files (*.qdl)|*.qdl|'               +
                'List files (*.lst)|*.lst|'        +
                'All files (*.*)|*.*';
  prSaveListDlg.FilterIndex := 1;
  prSaveListDlg.Options := [ofPathMustExist, ofOverwritePrompt, ofEnableSizing];
  prSaveListDlg.FileName := 'datalist';
  if not prSaveListDlg.Execute then begin
    Result := True;
    Exit;
  end;
  if prSaveListDlg.FilterIndex = 1 then begin
    f := ChangeFileExt(prSaveListDlg.FileName, '.qdl');
  end else if prSaveListDlg.FilterIndex = 2 then begin
    f := ChangeFileExt(prSaveListDlg.FileName, '.lst');
  end else begin
    f := ChangeFileExt(prSaveListDlg.FileName, '.qdl');
  end;

  FilesNode := TNode.Create('DataFiles');

  for c := 0 to ListObjects.Items.Count - 1 do begin

    FileNode := TNode.CreateParented('DataFile', FilesNode);
    FileNode['FileName'].SetData(ListObjects[c].FileName);

  end;//c

  //Save the FilesNode
  FilesNode.SaveAs(f);

end;

function TQUB_Viewport_Data.DataListOpen :Boolean;
var
  FilesNode :INode;
  FileNode  :INode;
  PathName  :String;
  Found     :Boolean;
  c, r      :Integer;
  f         :String;
begin

  Result    := False;

  //Process FileName;
  if prOpenListDlg = nil then prOpenListDlg := TOpenDialog.Create(Application);
  prOpenListDlg.Title := 'Open data list';
  prOpenListDlg.Filter := 'QUB data list files (*.qdl)|*.qdl|'               +
              'List files (*.lst)|*.lst|'        +
              'All files (*.*)|*.*';
  prOpenListDlg.FilterIndex := 1;
  prOpenListDlg.Options := [ofPathMustExist, ofFileMustExist, ofEnableSizing];
  prOpenListDlg.FileName := 'datalist.qdl';
  if not prOpenListDlg.Execute then begin
    Result := True;
    Exit;
  end;

  PathName := ExtractFilePath(prOpenListDlg.FileName);
  FilesNode := TNode.Open(prOpenListDlg.FileName);
  if not FilesNode.IsNull then begin
    FileNode := FilesNode['DataFile'];
    while not FileNode.IsNull do begin
      f := FileNode['FileName'].DataAsString;
      Found := False;
      for c := 0 to ListObjects.Items.Count - 1 do begin
        if UpperCase(ListObjects[c].FileName) = UpperCase(f) then begin
          Found := True;
          Break;
        end;
      end;
      if not Found then begin
          if not FileExists(f) then begin
            r := MessageDlg('File ''' + f + ''' could not be found.'#13#10'Would you like to try in the folder '''#13#10+PathName+'''?', mtConfirmation, [mbYes, mbNo, mbCancel], 0);
            case r of
              mrCancel:
                Exit;
              mrNo:
                Found := True;
              mrYes:
                f := PathName + ExtractFileName(f);
            end;
          end;
        end;
        if not Found then begin
        if Open(f, True{Silent}) then begin
          Result := True;
          Exit;
        end;
      end;
      FileNode := FileNode.SiblingSameName;
    end;
  end;

end;



//Properties
function TQUB_Viewport_Data.Properties :Boolean;
begin
  Result := False;
  if CurrData <> nil then CurrData.Properties else Result := True;
end;



//****************************************************************************//
//                                                                            //
//                                  T A S K S                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Viewport_Data.AddSegmentList :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.AddSegmentList;
end;

function TQUB_Viewport_Data.AddIdlSegmentList :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.AddIdlSegmentList;
end;

function TQUB_Viewport_Data.ChopSegments(Properties :Boolean) :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.ChopSegments(Properties);
end;

function TQUB_Viewport_Data.AddList :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.AddList;
end;

function TQUB_Viewport_Data.AddList(const Name  :String) :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.AddList(Name) <> nil;
end;

function TQUB_Viewport_Data.DelList :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.DelList;
end;

function TQUB_Viewport_Data.AddSelection :Boolean;
begin
  Result := False;
  if CurrData <> nil then begin
    if CurrData.DrawListOverlapped then Result := CurrData.AddSelections(CurrData.KeyShift{FromZoom}) else Result := CurrData.AddSelection;
  end;
end;

function TQUB_Viewport_Data.DelSelection :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.DelSelection;
end;

function TQUB_Viewport_Data.ShowSelectionProperties :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.ShowSelectionProperties;
end;

function TQUB_Viewport_Data.AddNote :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.AddNote('Notes');
end;

function TQUB_Viewport_Data.EvtListProcess(Properties :Boolean; OnActionDone :TNotifyEvent) :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.EvtListProcess(Properties, OnActionDone);
end;

function TQUB_Viewport_Data.GoToPoint(Properties :Boolean) :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.GoToPoint(Properties);
end;

function TQUB_Viewport_Data.PlaySelection :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.PlaySelection(-1{Chanel});
end;

function TQUB_Viewport_Data.StopSelection :Boolean;
begin
  Result := False;
  if CurrData <> nil then Result := CurrData.StopSelection;
end;


//****************************************************************************//
//                                                                            //
//                                  O T H E R                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Viewport_Data.prCurrData :TQUB_QVO_QVD;
begin
  Result := nil;
  if ListObjects = nil then Exit;
  if ListObjects.Items.Count = 0  then Exit;
  if ListObjects.ItemIndex   = -1 then Exit;
  Result := TQUB_QVO_QVD(ListObjects.Items.Objects[ListObjects.ItemIndex]);
end;

procedure TQUB_Viewport_Data.ListObjectsDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  Offset :Integer;
begin

  inherited;

  with (Control as TListBox).Canvas, (Control as TListBox) do begin

    FillRect(Rect);

    Offset := 2;

    if Items.Objects[Index] <> nil then begin
      if          TQUB_QVO_QVD(Items.Objects[Index]).LockedRecord   then begin
        //-- places red circle at start of item in list (record icon)
        //DataImageList.GetBitmap(0, Bitmap);
        DataImageList.Draw((Control as TListBox).Canvas, Rect.Left + Offset, Rect.Top, 0, True);
        Inc(Offset, 14);
      end else if TQUB_QVO_QVD(Items.Objects[Index]).LockedPlayback then begin
        //-- Places blue triangle at start of item in list (Play icon)
        //DataImageList.GetBitmap(1, Bitmap);
        DataImageList.Draw((Control as TListBox).Canvas, Rect.Left + Offset, Rect.Top, 1, True);
        Inc(Offset, 14);
      end;
      TextOut(Rect.Left + Offset, Rect.Top, (Control as TListBox).Items[Index]);
    end;

  end;

end;

procedure TQUB_Viewport_Data.SelScrollBarScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
var
  Data :TQUB_QVO_QVD;
  Sel  :TQUB_FirstLast;
begin
  inherited;

  Data := prCurrData;
  if Data = nil then Exit;

  Sel := Data.DataSel;
  if Sel.First = Sel.Last then Exit;

  case ScrollCode of

    scLineUp:	//User clicked the top or left scroll arrow or pressed the Up or Left arrow key.
      begin
        if Sel.First <= Data.DataPoint.First then Exit;
        if (GetAsyncKeyState(VK_CONTROL) < 0) then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_SEL_UPSEG){BrowseCmd});
        end else begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_SEL_UP){BrowseCmd});
        end;
        if not Data.Active then Data.Activate;
      end;

    scLineDown:	//User clicked the top or left scroll arrow or pressed the Up or Left arrow key.
      begin
        if Sel.Last >= Data.DataPoint.Last then Exit;
        if (GetAsyncKeyState(VK_CONTROL) < 0) then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_SEL_DNSEG){BrowseCmd});
        end else begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_SEL_DN){BrowseCmd});
        end;
        if not Data.Active then Data.Activate;
      end;

  end;

end;



procedure TQUB_Viewport_Data.DisplayScrollBarScroll(Sender: TObject; ScrollCode: TScrollCode; var ScrollPos: Integer);
var
  Data :TQUB_QVO_QVD;
  //Max  :Integer;
begin
  inherited;

  //Max := TScrollBar(Sender).Max;
  //if Max = TScrollBar(Sender).Min then Exit;

  Data := prCurrData;
  if Data = nil then Exit;

  case ScrollCode of

    scLineUp:	//User clicked the top or left scroll arrow or pressed the Up or Left arrow key.

      begin
        if Data.DataDraw.First > Data.DataPoint.First then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_TRACE_UP){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_TRACE_UP, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end;
      end;

    scLineDown:	//User clicked the bottom or right scroll arrow or pressed the Down or Right arrow key.

      begin
        if Data.DataDraw.Last < Data.DataPoint.Last then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_TRACE_DN){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_TRACE_DN, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end;
      end;

    scPageUp:	//User clicked the area to the left of the thumb tab or pressed the PgUp key.

      begin
        if Data.DataDraw.First > Data.DataPoint.First then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_PAGE_UP){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_PAGE_UP, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end;
      end;

    scPageDown:	//User clicked the area to the right of the thumb tab or pressed the PgDn key.

      begin
        if Data.DataDraw.Last < Data.DataPoint.Last then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_PAGE_DN){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_PAGE_DN, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end;
      end;

    scPosition:	//User positioned the thumb tab and released it.

      begin
        if ScrollPos = TScrollBar(Sender).Min then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_HOME){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_HOME, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end else if ScrollPos = TScrollBar(Sender).Max then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_END){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_END, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end else begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_TO_T, ScrollPos);
          //Data.Browse(QUB_QVO_QVD_CMD_GOTO_TRACE, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end;
      end;

    scTrack:	//User is moving the thumb tab.

      begin
      end;

    scTop:	//User moved the thumb tab to the top or far left on the scroll bar.

      begin
        if Data.DataDraw.First > Data.DataPoint.First then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_HOME){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_HOME, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end;
      end;

    scBottom:	//User moved the thumb tab to the bottom or far right on the scroll bar.

      begin
        if Data.DataDraw.Last < Data.DataPoint.Last then begin
          Data.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA or DATA_MSG_SET_SCROLL or DATA_MSG_BROWSE_CMD, Longint(QUB_QVO_QVD_CMD_END){BrowseCmd});
          //Data.Browse(QUB_QVO_QVD_CMD_END, ScrollPos, Max, True{Display}, False{SetScroll}, True{Process}, False);
        end;
      end;

    scEndScroll: //User finished moving the thumb tab on the scroll bar.

      begin
      end;

  end;//case

  //TScrollBar(Sender).Max := Max;
  if not Data.Active then Data.Activate;//if Data.CanFocus then try Data.SetFocus except end;            ???

end;




//Registry
procedure TQUB_Viewport_Data.DoPropertiesInternal(Read :Boolean);
var
  Reg :TRegistry;
begin

  inherited;

  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Interface ' + InterfaceName + '\Viewport ' + Name, True) then begin

        if Read then begin
          ConteinerDisplay.Height   := Round(ReadRegistryFloat(Reg,'ConteinerDisplay Height',80) * ConteinerLeft.ClientHeight / 100);
          ListObjects.Height        := Round(ReadRegistryFloat(Reg,'ListObjects Height',25) * ConteinerRight.ClientHeight / 100);
          ConteinerResults.Height   := Round(ReadRegistryFloat(Reg,'ConteinerResults Height',30) * ConteinerLists.ClientHeight / 100);
          ConteinerListsList.Height := Round(ReadRegistryFloat(Reg, 'ConteinerListsList Height', 30) * ConteinerLists.ClientHeight / 100);
          if ConteinerResults.Height < 16 then ConteinerResults.Height := 16;
          if ConteinerListsList.Height < 16 then ConteinerListsList.Height := 16;
        end else begin
          WriteFloat('ConteinerDisplay Height',   100 * ConteinerDisplay.Height   / ConteinerLeft.ClientHeight);
          WriteFloat('ListObjects Height',        100 * ListObjects.Height        / ConteinerRight.ClientHeight);
          WriteFloat('ConteinerResults Height',   100 * ConteinerResults.Height   / ConteinerLists.ClientHeight);
          WriteFloat('ConteinerListsList Height', 100 * ConteinerListsList.Height / ConteinerLists.ClientHeight);
        end;

      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;

function TQUB_Viewport_Data.DataCount :Integer;
begin
  Result := ListObjects.Items.Count;
end;

function TQUB_Viewport_Data.prGetHasData :Boolean;
begin
  Result := CurrData <> nil;
end;

function TQUB_Viewport_Data.prGetFileName :String;
begin
  if CurrData <> nil then Result := CurrData.FileName else Result := '';
end;

//******************************************************************************
procedure TQUB_Viewport_Data.prOnChangeDataLock(Sender :TObject);
begin
  ListObjects.Refresh;
end;

//******************************************************************************
procedure TQUB_Viewport_Data.InitPython;
begin
  QUB_Python_DataView_Init(Self);
end;

//----- Generic click handling to set focus when panels are clicked on.
//----- shared by several controls on this form.
procedure TQUB_Viewport_Data.OnClickGeneric(Sender: TObject);
begin
  inherited;
  SetFocus;
end;


//******************************************************************************
end.
