unit QUB_Viewport_DataOverlap;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, 
StdCtrls, ExtCtrls, ComCtrls, Menus, Registry, Math,

RegistryInterface,

QUB_Viewport,

QUB_Global,
QUB_QVO_QVD,
QUB_QVO_QVD_Global,
QUB_QVO_QVD_Overlap;


type


TQUB_Viewport_DataOverlap = class(TQUB_Viewport)


  PopupMenu          :TPopupMenu;
  PropertiesItem     :TMenuItem;
  CopyItem           :TMenuItem;
  N1                 :TMenuItem;
  ClearItem          :TMenuItem;
  ConteinerLeft      :TPanel;
  splLeftRight       :TSplitter;
  ConteinerRight     :TPanel;
  ConteinerTop       :TPanel;
  splTopBottom       :TSplitter;
  ConteinerBottom    :TPanel;
  ConteinerFiles     :TPanel;
  splFilesLists      :TSplitter;
  ConteinerList      :TPanel;
  ConteinerLists     :TPanel;
  Label1             :TLabel;
  cmbLists           :TComboBox;
  Label2             :TLabel;
  panStatus          :TPanel;
  lblStatus          :TLabel;
  lblStatusActive    :TLabel;


  procedure FormCreate(Sender: TObject);
  procedure panStatusResize(Sender: TObject);
  procedure cmbListsChange(Sender: TObject);
    procedure FormDestroy(Sender: TObject);


protected


  prData               :TQUB_QVO_QVD;

  prListIndex          :Integer;
  prList               :TQUB_QVO_QVD_SelectionList;

  prFileView           :TQUB_QVO_QVD_DataListMirror;
  prListView           :TQUB_QVO_QVD_ActiveSelList;
  prOverlapAll         :TQUB_QVO_QVD_Overlap;
  prOverlapActive      :TQUB_QVO_QVD_OverlapActive;
  prZoomStack          :TQUB_QVO_QVD_ListZoomStack;


  procedure prSetListIndex(x :Integer);

  procedure prOnListAdded(Sender :TObject; Index :Integer);
  procedure prOnListDeleting(Sender :TObject; Index :Integer);
  procedure prOnListRenamed(Sender :TObject; Index :Integer);

  procedure DoPropertiesInternal(Read :Boolean); override;


public


  constructor Create(AOwner :TComponent); override;


  procedure DoOnChangeData(Sender :TObject);
  procedure DoOnClosingData(Sender :TObject);


  property ListIndex         :Integer                     read prListIndex       write prSetListIndex;
  property List              :TQUB_QVO_QVD_SelectionList  read prList;

  property FileView          :TQUB_QVO_QVD_DataListMirror read prFileView;
  property ListView          :TQUB_QVO_QVD_ActiveSelList  read prListView;
  property OverlapAll        :TQUB_QVO_QVD_Overlap        read prOverlapAll;
  property OverlapActive     :TQUB_QVO_QVD_OverlapActive  read prOverlapActive;
  property ZoomStack         :TQUB_QVO_QVD_ListZoomStack  read prZoomStack;


end;


implementation



{$R *.DFM}


//****************************************************************************//
constructor TQUB_Viewport_DataOverlap.Create(AOwner :TComponent);
begin
  inherited;

  {
  prPropertiesForm := TQUB_Viewport_DataOverlap_PropertiesForm.Create(Self);
  prPropertiesForm.OnGetProperties := prOnGetProperties;
  prPropertiesForm.OnPutProperties := prOnPutProperties;
  }
end;

procedure TQUB_Viewport_DataOverlap.FormCreate(Sender: TObject);
begin
  inherited;

  prFileView            := TQUB_QVO_QVD_DataListMirror.Create(Self);
  prFileView.Align      := alClient;
  prFileView.Parent     := ConteinerFiles;

  prListView            := TQUB_QVO_QVD_ActiveSelList.Create(Self);
  prListView.Align      := alClient;
  prListView.Parent     := ConteinerList;

  prZoomStack           := TQUB_QVO_QVD_ListZoomStack.Create;

  prOverlapAll          := TQUB_QVO_QVD_Overlap.Create(Self);
  prOverlapAll.Align    := alClient;
  prOverlapAll.LblStatus := lblStatus;
  prOverlapAll.LblStatusActive := lblStatusActive;
  prOverlapAll.Parent   := ConteinerTop;
  prOverlapAll.ZoomStack := prZoomStack;

  prOverlapActive          := TQUB_QVO_QVD_OverlapActive.Create(Self);
  prOverlapActive.Align    := alClient;
  prOverlapActive.LblStatus := lblStatus;
  prOverlapActive.LblStatusActive := lblStatusActive;
  prOverlapActive.Parent   := ConteinerBottom;
  prOverlapActive.ZoomStack := prZoomStack;

  prListIndex   := -1;
end;

procedure TQUB_Viewport_DataOverlap.FormDestroy(Sender: TObject);
begin
  inherited;
  prZoomStack.Free;
end;

procedure TQUB_Viewport_DataOverlap.DoOnChangeData(Sender :TObject);
var
  i             :Integer;
begin
  if prData = Sender then Exit;

  if prData <> nil then begin
    ListIndex   := -1;
    prData.SelectionLists.OnListAdded    := nil;
    prData.SelectionLists.OnListDeleting := nil;
    prData.SelectionLists.OnListRenamed  := nil;
  end;

  prData    := TQUB_QVO_QVD(Sender);

  cmbLists.Items.Clear;
  cmbLists.Items.Add('<Pick a list to start...>');
  cmbLists.ItemIndex := 0;
  cmbLists.Enabled := prData <> nil;

  if prData <> nil then begin
    with prData.SelectionLists do begin
      for i := 0 to Items.Count - 1 do begin
        cmbLists.Items.Add( Items[i] );
      end;
    end;

    prData.SelectionLists.OnListAdded    := prOnListAdded;
    prData.SelectionLists.OnListDeleting := prOnListDeleting;
    prData.SelectionLists.OnListRenamed  := prOnListRenamed;
  end;

  prListView.Data       := prData;
  prOverlapAll.Data     := prData;
  prOverlapActive.Data  := prData;
end;

procedure TQUB_Viewport_DataOverlap.DoOnClosingData(Sender :TObject);
begin
  if Sender = prData then DoOnChangeData(nil);
end;

procedure TQUB_Viewport_DataOverlap.prSetListIndex(x :Integer);
begin
  if prData = nil then Exit;
  if x = prListIndex then Exit;

  prListIndex := x;
  cmbLists.ItemIndex := x + 1;
  
  if (x >= 0) and (x < prData.SelectionLists.Items.Count) then begin
    prList    := prData.SelectionLists[x];
    if prList <> prData.CurrList then prData.ChooseList( prListIndex );
  end else begin
    prList    := nil;
  end;

  prZoomStack.List      := prList;
  prListView.List       := prList;
  prOverlapAll.List     := prList;
  prOverlapActive.List  := prList;
end;

procedure TQUB_Viewport_DataOverlap.prOnListAdded(Sender :TObject; Index :Integer);
begin
  cmbLists.Items.Insert(Index + 1, prData.SelectionLists.Items[Index]);
  if ListIndex >= Index then Inc(prListIndex);
end;

procedure TQUB_Viewport_DataOverlap.prOnListDeleting(Sender :TObject; Index :Integer);
begin
  cmbLists.Items.Delete(Index + 1);
  if Index = ListIndex then begin
    ListIndex := -1;
  end else if ListIndex > Index then begin
    prListIndex := prListIndex - 1;
  end;
end;

procedure TQUB_Viewport_DataOverlap.prOnListRenamed(Sender :TObject; Index :Integer);
begin
  cmbLists.Items[Index + 1] := prData.SelectionLists.Items[Index];
end;

procedure TQUB_Viewport_DataOverlap.DoPropertiesInternal(Read :Boolean);
var
  Reg :TRegistry;
begin

  inherited;

  Reg := TRegistry.Create;

  try
    with Reg do begin
      if OpenKey('Software\QUB\Viewport ' + Name, True) then begin
        if Read then begin
          ConteinerRight.Width := Round(ReadRegistryFloat(Reg , 'List Width', 33.3 ) * ClientWidth / 100);
          ConteinerTop.Height  := Round(ReadRegistryFloat(Reg , 'Top Height', 33.3 ) * ClientHeight / 100);
        end else begin
          WriteFloat('List Width', Min(97.0, Max(3.0, 100.0 * ConteinerRight.Width / ClientWidth)));
          WriteFloat('Top Height', Min(97.0, Max(3.0, 100.0 * ConteinerTop.Height / ClientHeight )));
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
procedure TQUB_Viewport_DataOverlap.cmbListsChange(Sender: TObject);
begin
  inherited;
  ListIndex := cmbLists.ItemIndex - 1;
end;

procedure TQUB_Viewport_DataOverlap.panStatusResize(Sender: TObject);
begin
  inherited;
  lblStatus.Width := panStatus.Width div 2;
end;


//******************************************************************************
end.
