{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// dialog on new file to determine sampling etc.


unit QUB_Viewport_Data_New_Properties;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls, StdCtrls, ExtCtrls, Registry,

RegistryInterface,

QUB_Node,
QUB_Classic_Properties,
QUB_Profiles,
QUB_Global,
QUB_QVO_QVD_Global,
QUB_QFS_QDF_ScalingPanel,
QUB_QFS_File, QUB_QFS_QDF, QUB_QFS_QDF_Global;


{$include 'QUB_Viewport_Data_New_Properties.hct'}


const


DEF_BASENAME = 'noname';


type


TQUB_Viewport_Data_New_PropertiesForm = class(TQUB_Classic_PropertiesForm)


  Label1                 :TLabel;
  Label2                 :TLabel;
  Label3                 :TLabel;
  Label4                 :TLabel;
  Label6                 :TLabel;
  Label7                 :TLabel;
  Label8                 :TLabel;
  Label9                 :TLabel;

  txtFileName            :TEdit;
  btnBrowseFileName      :TButton;

  Panel1                 :TPanel;
  optFileFormat0         :TRadioButton;
  optFileFormat1         :TRadioButton;
  optFileFormat2         :TRadioButton;
  optFileFormat3         :TRadioButton;
  optFileFormat5         :TRadioButton;

  txtChannelCount        :TEdit;
  udChannelCount         :TUpDown;

  txtSamplingFrequency   :TEdit;
  txtSamplingTime        :TEdit;
  txtScaling             :TQUB_QFS_QDF_ScalingPanel;
  
  Panel2                 :TPanel;
  chkADDataSize2         :TRadioButton;
  chkADDataSize4         :TRadioButton;
  Label11                :TLabel;
  txtBaseName            :TEdit;
  Label10                :TLabel;


  procedure FormCreate(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure FormHide(Sender: TObject);
  procedure btnBrowseFileNameClick(Sender: TObject);
  procedure optFileFormatClick(Sender: TObject);
  procedure btnOKClick(Sender: TObject);
  procedure txtFileNameExit(Sender: TObject);
  procedure txtFileNameKeyPress(Sender: TObject; var Key: Char);
  procedure udChannelCountClick(Sender: TObject; Button: TUDBtnType);
  procedure txtSamplingFrequencyKeyPress(Sender: TObject; var Key: Char);
  procedure txtSamplingFrequencyExit(Sender: TObject);
  procedure txtSamplingTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtSamplingTimeExit(Sender: TObject);
  procedure chkADDataSizeClick(Sender: TObject);
  procedure txtBaseNameKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
  procedure txtScalingEditDblClick(Sender: TObject);


protected


  mnuProfiles                             :TQUB_ProfileEditBox;
  mnuExpCond                              :TQUB_ProfileReadBox;
  prProfile                               :INode;

  prBaseName                              :String;
  prFileName                              :String;
  prFileFormat                            :TQUB_QFS_QDF_FileType;
  prADChannelCount                        :Integer;
  prADDataSize                            :Integer;
  prSampling                              :Double;
  prScaling                               :Double;
  prHideScaling                           :Boolean;


  procedure prInitProfile;
  procedure prOnSetProfile(Cat :TQUB_ProfileCat; Updates :INode);
  function  prGetExpCond: INode;

  procedure prOnScalingChange(Sender :TObject);
  procedure prOnScalingEditing(Sender :TObject);
  procedure prOnScalingMinimize(Sender :TObject);
  procedure prOnScalingClose(Sender :TObject);

  procedure prSetBaseName(f :String);
  procedure prSetFileName(f :String);
  procedure prSetFileFormat(f :TQUB_QFS_QDF_FileType);
  procedure prSetADChannelCount(c :Integer);
  procedure prSetADDataSize(s :Integer);
  procedure prSetSampling(s :Double);
  procedure prSetScaling(s :Double);


public


  function FileTypeToExt(FType :TQUB_QFS_QDF_FileType) :String;


  property BaseName                       :String                read prBaseName                  write prSetBaseName;
  property FileName                       :String                read prFileName                  write prSetFileName;
  property FileFormat                     :TQUB_QFS_QDF_FileType read prFileFormat                write prSetFileFormat;
  property ADChannelCount                 :Integer               read prADChannelCount            write prSetADChannelCount;
  property ADDataSize                     :Integer               read prADDataSize                write prSetADDataSize;
  property Sampling                       :Double                read prSampling                  write prSetSampling;//Frequency, in kHz
  property Scaling                        :Double                read prScaling                   write prSetScaling;
  property ExpCond                        :INode                 read prGetExpCond;


end;


implementation


uses


QUB_GUIInterface;


{$R *.DFM}


//******************************************************************************
function TQUB_Viewport_Data_New_PropertiesForm.FileTypeToExt(FType :TQUB_QFS_QDF_FileType) :String;
begin
  Result := '';
  case FType of
    QFS_QDF_TYPE_QDF:
      Result := '.qdf';
    QFS_QDF_TYPE_LDT:
      Result := '.ldt';
    QFS_QDF_TYPE_DAT:
      Result := '.dat';
    QFS_QDF_TYPE_ASC:
      Result := '.txt';
    QFS_QDF_TYPE_WAV:
      Result := '.wav';
  end;//case
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prSetBaseName(f :String);
var
  baselen     :Integer;
  fdir, fname :String;
begin
  baselen     := Length(prBaseName);
  fname       := ExtractFileName(prFileName);
  if Copy(fname, 1, baselen) = prBaseName then begin
    fdir      := ExtractFileDir(prFileName);
    // big nasty surprise: Copy is 1-based
    fname     := Copy(fname, baselen+1, 10000);
    FileName  := fdir+'\'+f+fname;
  end;
  prBaseName       := f;
  txtBaseName.Text := f;
  prProfile['BaseName'].SetData(f);
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prSetFileName(f :String);
var
  FExt       :String;
begin
  prFileName := f;

  FExt       := LowerCase(ExtractFileExt(f));
  if (FExt = '.qmf') or (FExt = '.mdl') then begin
    prFileName := ChangeFileExt(prFileName, '');
  end;

  txtFileName.Text := prFileName;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prSetFileFormat(f :TQUB_QFS_QDF_FileType);
begin
  prFileFormat := f;
  case f of
    QFS_QDF_TYPE_QDF: optFileFormat0.Checked := True;
    QFS_QDF_TYPE_LDT: optFileFormat1.Checked := True;
    QFS_QDF_TYPE_DAT: optFileFormat2.Checked := True;
    QFS_QDF_TYPE_ASC: optFileFormat3.Checked := True;
    QFS_QDF_TYPE_WAV: optFileFormat5.Checked := True;
    else
      begin
        prFileFormat := QFS_QDF_TYPE_QDF;
        optFileFormat0.Checked := True;
      end;
  end;//case

  udChannelCount.Max      := QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[prFileFormat];
  txtChannelCount.Enabled := QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[prFileFormat] > 1;
  if ADChannelCount > QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[prFileFormat] then begin
    ADChannelCount := QUB_QFS_QDF_MAX_ADCHANNELCOUNTS[prFileFormat];
  end;
  prProfile['FileFormat'].Data[0] := Integer(prFileFormat);
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prSetADChannelCount(c :Integer);
begin
  prADChannelCount := c;
  if udChannelCount.Position <> c then udChannelCount.Position := c;
  prProfile['ADChannelCount'].Data[0] := c;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prSetADDataSize(s :Integer);
begin
  prADDataSize := s;
  chkADDataSize2.Checked := s = 2;
  chkADDataSize4.Checked := s = 4;
  prProfile['ADDataSize'].Data[0] := s;
  if s = 2 then begin
    txtScaling.MaxInt := High(SmallInt);
  end else begin
    txtScaling.MaxInt := High(LongInt);
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prSetSampling(s :Double);
begin
  prSampling := s;
  txtSamplingFrequency.Text := FloatToStr(Sampling);
  txtSamplingTime.Text      := FloatToStr(1 / Sampling);
  prProfile['Sampling'].Data[0] := s;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prSetScaling(s :Double);
begin
  prScaling := s;
  prProfile['Scaling'].Data[0] := s;
  txtScaling.OnScalingChange := nil;
  txtScaling.Scaling := s;
  txtScaling.OnScalingChange := prOnScalingChange;
end;

//Events
procedure TQUB_Viewport_Data_New_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prProfile       := ProfileMgr['Data'].Active;
  prInitProfile;

  mnuProfiles                             := TQUB_ProfileEditBox.Create(Self);
  mnuProfiles.Parent                      := Self;
  mnuProfiles.SetBounds(ClientWidth - 100 - 10,
                        ClientHeight - mnuProfiles.Height - 10,
                        100, mnuProfiles.Height);
  mnuProfiles.Anchors                     := [akRight, akBottom];
  mnuProfiles.Cat                         := 'Data';

  mnuExpCond                              := TQUB_ProfileReadBox.Create(Self);
  mnuExpCond.Parent                       := Self;
  mnuExpCond.SetBounds(263, 189, 114, mnuExpCond.Height);
  mnuExpCond.Cat                          := 'ExpCond';

  txtScaling.OnScalingChange              := prOnScalingChange;
  txtScaling.OnEditing                    := prOnScalingEditing;
  txtScaling.OnMinimize                   := prOnScalingMinimize;
  txtScaling.OnClose                      := prOnScalingClose;

  {$include 'QUB_Viewport_Data_New_Properties.htx'}
  WikiTopic                               := 'New_Data';
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  BaseName        := BaseName;

  prOnSetProfile(ProfileMgr['Data'], prProfile);
  ProfileMgr['Data'].OnSet := prOnSetProfile;

  mnuExpCond.Refresh;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.FormHide(Sender: TObject);
begin
  inherited;

  ProfileMgr['Data'].OnSet := nil;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prInitProfile;
var
  Node            :INode;
  function HasData(const Name :String): Boolean;
  begin
    Node          := prProfile[Name];
    Result        := Node.DataCount > 0;
  end;
begin
  if not HasData('BaseName')       then Node.SetData(DEF_BASENAME);
  if not HasData('FileFormat')     then Node.SetData(Integer(QFS_QDF_TYPE_QDF));
  if not HasData('ADChannelCount') then Node.SetData(1);
  if not HasData('ADDataSize')     then Node.SetData(SizeOf(SmallInt));
  if not HasData('Sampling')       then Node.SetData(QUB_QVO_QVD_DEF_SAMPLING);
  if not HasData('Scaling')        then Node.SetData(QUB_QVO_QVD_DEF_SCALING);

  prOnSetProfile(ProfileMgr['Data'], prProfile);
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prOnSetProfile(Cat :TQUB_ProfileCat; Updates :INode);
var
  Node            :INode;
  function HasData(const Name :String): Boolean;
  begin
    Node          := Updates.Find(Name);
    Result        := Node.DataCount > 0;
  end;
begin
  if HasData('BaseName')       then BaseName       := Node.DataAsString;
  if HasData('FileFormat')     then FileFormat     := TQUB_QFS_QDF_FileType(Node.Data[0]);
  if HasData('ADChannelCount') then ADChannelCount := Node.Data[0];
  if HasData('ADDataSize')     then ADDataSize     := Node.Data[0];
  if HasData('Sampling')       then Sampling       := Double(Node.Data[0]);
  if HasData('Scaling')        then Scaling        := Node.Data[0];
end;

function TQUB_Viewport_Data_New_PropertiesForm.prGetExpCond: INode;
begin
  Result := mnuExpCond.Properties;
end;

//******************************************************************************

procedure TQUB_Viewport_Data_New_PropertiesForm.prOnScalingChange(Sender :TObject);
begin
  Scaling := txtScaling.Scaling;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prOnScalingEditing(Sender :TObject);
begin
  if prHideScaling then Exit;
  with TQUB_QFS_QDF_ScalingDetails(txtScaling.Details) do begin
    if Parent <> Self then begin
      Parent := Self;
      SetBounds(300, 120, Width, Height);
    end;
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prOnScalingMinimize(Sender :TObject);
begin
  prHideScaling := True;
  prOnScalingClose(Sender);
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.prOnScalingClose(Sender :TObject);
begin
  txtScaling.Details.Parent := nil;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.txtScalingEditDblClick(Sender: TObject);
begin
  inherited;
  prHideScaling := False;
  prOnScalingEditing(txtScaling);
end;

//FileName
procedure TQUB_Viewport_Data_New_PropertiesForm.txtBaseNameKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  f :String;
begin
  inherited;
  f := TEdit(Sender).Text;
  if f <> BaseName then begin
    BaseName := f;
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.btnBrowseFileNameClick(Sender: TObject);
var
  Dlg :TSaveDialog;
begin
  inherited;
  Dlg := TSaveDialog.Create(Application);
  try
    Dlg.Filter := 'All QUB data files (*.qdf;*.dat;*.ldt;*.txt)|*.qdf;*.dat;*.ldt;*.txt|QUB data files (*.qdf)|*.qdf|QUB old binary files (*.dat)|*.dat|QUB old ldt files (*.ldt)|*.ldt|ASCII data files (*.txt)|*.txt|WAVE sound files (*.wav)|*.wav|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofEnableSizing];
    Dlg.FileName := FileName;
    if Dlg.Execute then begin
      FileName := Dlg.FileName;
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.txtFileNameKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtFileNameExit(Sender);
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.txtFileNameExit(Sender: TObject);
var
  f :String;
begin
  inherited;
  f := TEdit(Sender).Text;
  if f <> FileName then begin
    FileName := f;
  end;
end;

//Properties
procedure TQUB_Viewport_Data_New_PropertiesForm.optFileFormatClick(Sender: TObject);
begin
  inherited;
  FileFormat := TQUB_QFS_QDF_FileType(TRadioButton(Sender).Tag);
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.udChannelCountClick(Sender: TObject; Button: TUDBtnType);
begin
  inherited;
  ADChannelCount := TUpDown(Sender).Position;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.chkADDataSizeClick(Sender: TObject);
begin
  inherited;
  ADDataSize := TRadioButton(Sender).Tag;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.txtSamplingFrequencyKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSamplingFrequencyExit(Sender);
  end;
end;

//Sampling
procedure TQUB_Viewport_Data_New_PropertiesForm.txtSamplingFrequencyExit(Sender: TObject);
var
  s :Extended;
begin
  inherited;
  try
    s := StrToFloat(TEdit(Sender).Text);
    if ((s < QUB_QVO_QVD_MIN_SAMPLING) or (s > QUB_QVO_QVD_MAX_SAMPLING)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(QUB_QVO_QVD_MIN_SAMPLING) + ' and ' + FloatToStr(QUB_QVO_QVD_MAX_SAMPLING), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(Sampling);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if s <> Sampling then begin
    Sampling := s;
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.txtSamplingTimeKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtSamplingTimeExit(Sender);
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.txtSamplingTimeExit(Sender: TObject);
var
  s :Extended;
begin
  inherited;
  try
    s := 1 / StrToFloat(TEdit(Sender).Text);
    if ((s < QUB_QVO_QVD_MIN_SAMPLING) or (s > QUB_QVO_QVD_MAX_SAMPLING)) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + FloatToStr(QUB_QVO_QVD_MIN_SAMPLING) + ' and ' + FloatToStr(QUB_QVO_QVD_MAX_SAMPLING), mtError, [mbOK], 0);
    TEdit(Sender).Text := FloatToStr(1 / Sampling);
    TEdit(Sender).SetFocus;
    Exit;
  end;
  if s <> Sampling then begin
    Sampling := s;
  end;
end;

procedure TQUB_Viewport_Data_New_PropertiesForm.btnOKClick(Sender: TObject);
var
  Dlg :TSaveDialog;
  r   :Integer;
  f   :String;
begin

  f := ChangeFileExt(FileName, FileTypeToExt(FileFormat));
  case FileFormat of
    QFS_QDF_TYPE_WAV:
      begin
        if ADChannelCount > 2 then ADChannelCount := 2;
      end;
  end;//case

  Dlg := nil;

  try

    while FileExists(f) do begin

      Dlg := TSaveDialog.Create(Application);
      Dlg.Filter := 'All QUB data files (*.qdf;*.dat;*.ldt;*.txt)|*.qdf;*.dat;*.ldt;*.txt|QUB data files (*.qdf)|*.qdf|QUB old binary files (*.dat)|*.dat|QUB old ldt files (*.ldt)|*.ldt|ASCII data files (*.txt)|*.txt|WAVE sound files (*.wav)|*.wav|All files (*.*)|*.*';
      Dlg.Options := [ofPathMustExist, ofEnableSizing];
      Dlg.FileName := ChangeFileExt(f, '');
      r := MessageHelpDlg('File ''' + f + ''' already exists. Do you want to overwrite it ?', mtConfirmation, [mbYes, mbNo, mbCancel], 0, 'Warning:File_Already_Exists');
      case r of
        idCancel:
          begin
            ModalResult := mrNone;
            Exit;
          end;
        idYes:
          begin
            FileName   := f;
            prFileName := f;
            inherited;
            Exit;
          end;
        idNo:
          begin
            if Dlg.Execute then begin
              f := ChangeFileExt(Dlg.FileName, FileTypeToExt(FileFormat));
            end else begin
              ModalResult := mrNone;
              Exit;
            end;
          end;
      end;//case
    end;//while

    FileName   := f;
    prFileName := f;
    inherited;

  finally
    Dlg.Free;
  end;

end;


//******************************************************************************
end.
