{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the bar across the top of QUB that reflects settings in the QUB Visual Data


unit QUB_Viewport_Data_Shortcuts;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls, StdCtrls, Buttons,

QUB_TrackBar,
QUB_Global,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_QFS_File, QUB_QFS_QDF_Global,
QUB_Viewport_Data,
QUB_PyCmd_MenuButton;


//{$include 'QUB_Viewport_Data_Shortcuts.hct'}


type


TQUB_Viewport_Data_Shortcuts = class(TFrame)


  lblActiveChannel          :TLabel;
  txtActiveChannel          :TEdit;
  udActiveChannel           :TUpDown;

  chkAutoScale              :TCheckBox;

  btnDataZoomIn             :TSpeedButton;
  btnDataZoomOut            :TSpeedButton;
  btnDataUp                 :TSpeedButton;
  btnDataDown               :TSpeedButton;
  btnFindTrace              :TSpeedButton;

  chkFilterOn               :TCheckBox;
  txtFilterFreq             :TEdit;
  chkBaselineOn             :TCheckBox;
  chkDrawList               :TCheckBox;
  chkDrawIdealized          :TCheckBox;
  chkDrawIdealizedAbove     :TCheckBox;
  lblDeadTime               :TLabel;
  txtDeadTime               :TEdit;
  chkAutoScroll             :TCheckBox;


  procedure udActiveChannelClick(Sender: TObject; Button: TUDBtnType);
  procedure chkAutoScaleClick(Sender: TObject);
  procedure btnDataZoomInClick(Sender: TObject);
  procedure btnDataZoomOutClick(Sender: TObject);
  procedure btnDataUpClick(Sender: TObject);
  procedure btnDataDownClick(Sender: TObject);
  procedure btnFindTraceClick(Sender: TObject);
  procedure chkFilterOnClick(Sender: TObject);
  procedure txtFilterFreqKeyPress(Sender: TObject; var Key: Char);
  procedure txtFilterFreqExit(Sender: TObject);
  procedure chkBaselineOnClick(Sender: TObject);
  procedure chkDrawListClick(Sender: TObject);
  //procedure chkDrawParametersClick(Sender: TObject);
  procedure chkDrawIdealizedAboveClick(Sender: TObject);
  procedure chkDrawIdealizedClick(Sender: TObject);
  procedure lblDeadTimeClick(Sender: TObject);
  procedure txtDeadTimeKeyPress(Sender: TObject; var Key: Char);
  procedure txtDeadTimeExit(Sender: TObject);
  procedure chkAutoScrollClick(Sender: TObject);


protected


  trkFilter                 :TQUB_TrackBar;

  prData                    :TQUB_QVO_QVD;
  prActiveChannel           :Integer;

  prChangingFilter          :Boolean;

  prDeadTimeInSamples       :Boolean;

  mnuPython                 :TQUB_PyCmd_MenuButton;


  procedure trkFilterChange(Sender: TObject);

  procedure prSetData(d :TQUB_QVO_QVD);
  procedure prSetActiveChannel(c :Integer);

  function prFilterTable(f :Integer) :Double;
  function prFilterTrackTable(f :Double) :Integer;

  procedure prOnExpCondChange(Sender :TObject);


public

  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure DoOnOpenData(Sender :TObject);
  procedure DoOnCloseData(Sender :TObject);
  procedure DoOnChangeData(Sender :TObject);
  procedure DoOnActiveChannelChange(Sender :TObject);

  // toggle 5 zoom controls '.Enabled' based on AutoScroll setting  2004-02 {JB}
  procedure EnableZoom( l : boolean );

  property Data             :TQUB_QVO_QVD                              write prSetData;
  property ActiveChannel    :Integer          read prActiveChannel     write prSetActiveChannel;


end;


implementation


uses

QUB_ExpCond_Parameters;


{$R *.DFM}


//******************************************************************************
constructor TQUB_Viewport_Data_Shortcuts.Create(AOwner :TComponent);
begin
  inherited;

  TrkFilter                 := TQUB_TrackBar.Create(Self);
  TrkFilter.Parent          := Self;
  TrkFilter.Left            := 388;
  TrkFilter.Top             := 6;
  TrkFilter.Width           := 51;
  TrkFilter.Height          := 17;
  TrkFilter.Enabled         := False;
  TrkFilter.Min             := 1;
  TrkFilter.Orientation     := trHorizontal;
  TrkFilter.Frequency       := 1;
  TrkFilter.Position        := 1;
  TrkFilter.SelEnd          := 0;
  TrkFilter.SelStart        := 0;
  TrkFilter.TabOrder        := 6;
  TrkFilter.ThumbLength     := 10;
  TrkFilter.TickMarks       := tmBoth;
  TrkFilter.TickStyle       := tsNone;
  TrkFilter.OnChange        := trkFilterChange;

  mnuPython                 := TQUB_PyCmd_MenuButton.Create('Python');
  mnuPython.Category        := 'General';
  mnuPython.Parent          := Self;
  mnuPython.Button.Top      := 2;
  mnuPython.Button.Left     := chkDrawIdealizedAbove.Left + chkDrawIdealizedAbove.Width + 10;
  mnuPython.Button.Width    := 60;
  mnuPython.Button.Visible  := False;
  //mnuPython.Button.Anchors  := [akTop, akRight];

  lblDeadTimeClick( lblDeadTime );
  
end;

destructor TQUB_Viewport_Data_Shortcuts.Destroy;
begin
  try
    mnuPython.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.DoOnOpenData(Sender :TObject);
begin
  Data := TQUB_QVO_QVD(Sender);
end;

procedure TQUB_Viewport_Data_Shortcuts.DoOnChangeData(Sender :TObject);
begin
  Data := TQUB_QVO_QVD(Sender);
end;

procedure TQUB_Viewport_Data_Shortcuts.DoOnCloseData(Sender :TObject);
begin
  if TQUB_QVO_QVD(Sender) <> nil then TQUB_QVO_QVD(Sender).ExpCond.Events.DelEvent('OnChange', prOnExpCondChange);
end;

procedure TQUB_Viewport_Data_Shortcuts.prSetData(d :TQUB_QVO_QVD);
var
  IsDWT :Boolean;
begin

  if prData <> nil then prData.ExpCond.Events.DelEvent('OnChange', prOnExpCondChange);

  if d <> nil then IsDWT := (d.DataObject.FileType = QFS_QDF_TYPE_DWT) or (d.DataObject.FileType = QFS_QDF_TYPE_EVL) or (d.DataObject.FileType = QFS_QDF_TYPE_PRT) or (d.DataObject.FileType = QFS_QDF_TYPE_QSF) or (d.DataObject.FileType = QFS_QDF_TYPE_TAC) else IsDWT := False;

  lblActiveChannel.Enabled       := d <> nil;
  txtActiveChannel.Enabled       := d <> nil;
  udActiveChannel.Enabled        := d <> nil;
  udActiveChannel.Refresh;

  chkAutoScale.Enabled           := d <> nil;

  chkAutoScroll.Enabled          := d <> nil;

  EnableZoom(d <> nil );
  if not IsDWT then btnFindTrace.Enabled := False;
  //btnFindTrace.Enabled           := (d <> nil) and (not IsDWT);

  chkFilterOn.Enabled            := (d <> nil) and (not IsDWT);
  txtFilterFreq.Enabled          := (d <> nil) and (not IsDWT);
  trkFilter.Enabled              := (d <> nil) and (not IsDWT);

  chkBaselineOn.Enabled          := (d <> nil) and (not IsDWT);
  chkDrawList.Enabled            := (d <> nil);

  chkDrawIdealized.Enabled       := (d <> nil) and (not IsDWT);
  chkDrawIdealizedAbove.Enabled  := (d <> nil) and (not IsDWT);

  lblDeadTime.Enabled            := d <> nil;
  txtDeadTime.Enabled            := d <> nil;

  prData                         := d;
  if prData = nil then begin
    udActiveChannel.Position     := 0;
    udActiveChannel.Max          := 0;
    prActiveChannel              := -1;
    txtDeadTime.Text             := '';
  end else begin
    udActiveChannel.Position     := 0;
    udActiveChannel.Max          := prData.DataChannelCount - 1;
    ActiveChannel                := prData.ActiveChannel;
    prData.ExpCond.Events.AddEvent('OnChange', prOnExpCondChange);
    prOnExpCondChange( prData.ExpCond );
  end;

end;

procedure TQUB_Viewport_Data_Shortcuts.prSetActiveChannel(c :Integer);
begin
  prData.ActiveChannel := c;
end;

procedure TQUB_Viewport_Data_Shortcuts.DoOnActiveChannelChange(Sender :TObject);
begin
  if Sender = prData then begin
    prActiveChannel               := prData.ActiveChannel;
    udActiveChannel.Position      := prActiveChannel;
    chkAutoscale.Checked          := prData.Autoscale;
    EnableZoom( not prData.Autoscale );
    chkAutoScroll.Checked         := prData.AutoScroll;
    prChangingFilter              := True;
    try
      chkFilterOn.Checked         := prData.FilterOn[ActiveChannel];
      txtFilterFreq.Text          := FloatToStr(prData.FilterFreq[ActiveChannel] / 1e3);
      trkFilter.Position          := prFilterTrackTable(prData.FilterFreq[ActiveChannel] / 1000);
    finally
      prChangingFilter            := False;
    end;
    chkBaselineOn.Checked         := prData.BaselineOn[ActiveChannel];
    chkDrawList.Checked           := prData.DrawSelList;
    chkDrawIdealized.Checked      := prData.DrawIdealized;
    chkDrawIdealizedAbove.Checked := prData.DrawIdealizedAbove;
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.udActiveChannelClick(Sender: TObject; Button: TUDBtnType);
begin
  ActiveChannel := TUpDown(Sender).Position;
end;

// toggle 5 zoom controls '.Enabled' based on AutoScroll setting  2004-02 {JB}
procedure TQUB_Viewport_Data_Shortcuts.EnableZoom( l : boolean );
begin
    btnDataZoomIn.Enabled := l;
    btnDataZoomOut.Enabled := l;
    btnDataUp.Enabled := l;
    btnDataDown.Enabled := l;
    btnFindTrace.Enabled := l;
end;

procedure TQUB_Viewport_Data_Shortcuts.chkAutoScaleClick(Sender: TObject);
var l : boolean;
begin
  if prData.AutoScale <> TCheckBox(Sender).Checked then begin
    l := TCheckBox(Sender).Checked;
    prData.AutoScale := l;
    EnableZoom(not l);

    //DrawAndZoom;
    prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA, 0);
  end
end;

procedure TQUB_Viewport_Data_Shortcuts.chkAutoScrollClick(Sender: TObject);
begin
  if prData.AutoScroll <> TCheckBox(Sender).Checked then begin
    prData.AutoScroll := TCheckBox(Sender).Checked;
  end
end;

procedure TQUB_Viewport_Data_Shortcuts.btnDataZoomInClick(Sender: TObject);
begin
  if prData.KeyShift then begin
    prData.ZoomIn(2);
  end else if prData.KeyControl then begin
    prData.ZoomIn(100);
  end else begin
    prData.ZoomIn(10);
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.btnDataZoomOutClick(Sender: TObject);
begin
  if prData.KeyShift then begin
    prData.ZoomOut(2);
  end else if prData.KeyControl then begin
    prData.ZoomOut(100);
  end else begin
    prData.ZoomOut(10);
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.btnDataUpClick(Sender: TObject);
begin
  if prData.KeyShift then begin
    prData.OffsetUp(2);
  end else if prData.KeyControl then begin
    prData.OffsetUp(200);
  end else begin
    prData.OffsetUp(20);
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.btnDataDownClick(Sender: TObject);
begin
  if prData.KeyShift then begin
    prData.OffsetDown(2);
  end else if prData.KeyControl then begin
    prData.OffsetDown(200);
  end else begin
    prData.OffsetDown(20);
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.btnFindTraceClick(Sender: TObject);
begin
  prData.FindTrace;
end;

procedure TQUB_Viewport_Data_Shortcuts.chkFilterOnClick(Sender: TObject);
begin
  if prData.FilterOn[ActiveChannel] <> TCheckBox(Sender).Checked then begin
    prData.FilterOn[ActiveChannel] := TCheckBox(Sender).Checked;
    //DrawAndProcessAndZoom;
    prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA, 0); //LM
  end
end;

procedure TQUB_Viewport_Data_Shortcuts.txtFilterFreqKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtFilterFreqExit(Sender);
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.txtFilterFreqExit(Sender: TObject);
var
  Frq :Double;
begin
  prChangingFilter := True;
  try
    try
      Frq := StrToFloat(TEdit(Sender).Text) * 1000;
      if ((Frq < QUB_QVO_QVD_MIN_FILTERFREQ) or (Frq > QUB_QVO_QVD_MAX_FILTERFREQ)) then raise Exception.Create('');
    except
      MessageDlg('Invalid value; must be between ' + FloatToStr(QUB_QVO_QVD_MIN_FILTERFREQ / 1000) + ' and ' + FloatToStr(QUB_QVO_QVD_MAX_FILTERFREQ / 1000), mtError, [mbOK], 0);
      TEdit(Sender).Text := FloatToStr(prData.FilterFreq[ActiveChannel] / 1000);
      TEdit(Sender).SetFocus;
      Exit;
    end;
    if Frq <> prData.FilterFreq[ActiveChannel] then begin
      prData.FilterFreq[ActiveChannel] := Frq;
      trkFilter.Position := prFilterTrackTable(prData.FilterFreq[ActiveChannel] / 1000);
      //DrawAndProcessAndZoom;
      prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA, 0); //LM
    end;
  finally
    prChangingFilter := False;
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.trkFilterChange(Sender: TObject);
var
  Frq :Double;
begin
  if prChangingFilter then Exit;
  prChangingFilter := True;
  try
    Frq := prFilterTable(trkFilter.Position) * 1000;
    if Frq <> (prData.FilterFreq[ActiveChannel] / 1000) then begin
      prData.FilterFreq[ActiveChannel] := Frq;
      txtFilterFreq.Text := FloatToStr(prData.FilterFreq[ActiveChannel] / 1000);
      //DrawAndProcessAndZoom;
      prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA, 0); //LM
    end;
  finally
    prChangingFilter := False;
  end;
end;

function TQUB_Viewport_Data_Shortcuts.prFilterTable(f :Integer) :Double;
begin
  // unusual selection of filters - 12,15 are not smooth.  Perhaps commonly used values ?
  //Yes, 12 doesn't make much sense. Maybe 10, 15, 20, 30, 50? LM.
  // I would think {... .1,.2,.5, 1,2,5, 10,20,50, 100,200,500, ...} {JB}
  //todo
  case f of
    1:   Result := 0.2;
    2:   Result := 0.5;
    3:   Result := 1;
    4:   Result := 2;
    5:   Result := 5;
    6:   Result := 10;
    7:   Result := 12;
    8:   Result := 15;
    9:   Result := 20;
    else Result := 50;
  end;
end;

function TQUB_Viewport_Data_Shortcuts.prFilterTrackTable(f :Double) :Integer;
begin
  // These should be maintained as semi midpoints of ranges in prFilterTable.
  if f <= 0.3 then begin
    Result := 1;
  end else if f<=0.7  then begin
    Result := 2;
  end else if f<=1.5  then begin
    Result := 3;
  end else if f<=3.0  then begin
    Result := 4;
  end else if f<=7.0  then begin
    Result := 5;
  end else if f<=11.0 then begin
    Result := 6;
  end else if f<=13.0 then begin
    Result := 7;
  end else if f<=17.0 then begin
    Result := 8;
  end else if f<=30.0 then begin
    Result := 9;
  end else begin
    Result := 10;
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.chkBaselineOnClick(Sender: TObject);
begin
  if prData.BaselineOn[ActiveChannel] <> TCheckBox(Sender).Checked then begin
    prData.BaselineOn[ActiveChannel] := TCheckBox(Sender).Checked;
    //DrawAndProcessAndZoom;
    prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA or DATA_MSG_PROC_DATA, 0); //LM
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.chkDrawListClick(Sender: TObject);
begin
  if prData.DrawSelList <> TCheckBox(Sender).Checked then begin
    prData.DrawSelList := TCheckBox(Sender).Checked;
    //DrawAndZoom;
    prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA, 0);
  end;
end;

{
procedure TQUB_Viewport_Data_Shortcuts.chkDrawParametersClick(Sender: TObject);
begin
  if prData.DrawParameters <> TCheckBox(Sender).Checked then begin
    prData.DrawParameters := TCheckBox(Sender).Checked;
    //DrawAndZoom;
    prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA, 0);
  end;
end;
}

procedure TQUB_Viewport_Data_Shortcuts.chkDrawIdealizedClick(Sender: TObject);
begin
  if prData.DrawIdealized <> TCheckBox(Sender).Checked then begin
    prData.DrawIdealized := TCheckBox(Sender).Checked;
    //prData.DrawAndZoom;
    prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA, 0);
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.chkDrawIdealizedAboveClick(Sender: TObject);
begin
  if prData.DrawIdealizedAbove <> TCheckBox(Sender).Checked then begin
    prData.DrawIdealizedAbove := TCheckBox(Sender).Checked;
    //DrawAndZoom;
    prData.DataMessage(DATA_MSG_BROWSE or DATA_MSG_DRAW_DATA, 0);
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.lblDeadTimeClick(Sender: TObject);
begin
  txtDeadTimeExit( txtDeadTime );

  prDeadTimeInSamples := not prDeadTimeInSamples;
  if prDeadTimeInSamples then begin
    lblDeadTime.Caption := 'td [samples]:';
  end else begin
    lblDeadTime.Caption := 'td [ms]:';
  end;
  if prData <> nil then prOnExpCondChange( prData.ExpCond );
end;

procedure TQUB_Viewport_Data_Shortcuts.txtDeadTimeKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtDeadTimeExit(Sender);
  end else begin
    TEdit(Sender).Font.Color := clRed;
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.txtDeadTimeExit(Sender: TObject);
var
  TDead :Double;
begin
  TEdit(Sender).Font.Color := clWindowText;

  if TEdit(Sender).Text = '' then begin
    if prData <> nil then prOnExpCondChange( prData.ExpCond );
    Exit;
  end;

  try
    TDead := StrToFloat(TEdit(Sender).Text);
    if prDeadTimeInSamples then TDead := TDead * prData.Sampling * 1.0e3;
    if TDead < 0.0 then raise Exception.Create('');

    prData.ExpCond.ChangePar('DeadTime', TDead);
  except
    MessageDlg('Invalid value; must be positive or zero', mtError, [mbOK], 0);
    prOnExpCondChange( prData.ExpCond );
    TEdit(Sender).SetFocus;
    Exit;
  end;
end;

procedure TQUB_Viewport_Data_Shortcuts.prOnExpCondChange(Sender :TObject);
var
  Param          :TQUB_ExpCond_PParameter;
  TDead          :Double;
begin
  Param          := prData.ExpCond.ParameterByName['DeadTime'];
  if Param = nil then begin
    TDead        := prData.Sampling * 1.0e3;
    prData.ExpCond.AddPar('DeadTime', TDead, -1, False);
  end else begin
    TDead        := Param.Value;
  end;

  if prDeadTimeInSamples then begin
    txtDeadTime.Text := FloatToStr(TDead * 1.0e-3 / prData.Sampling);
  end else begin
    txtDeadTime.Text := FloatToStr(TDead);
  end;
end;


//******************************************************************************
end.
