{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
//ancestor class of Viewports such as Data, Model that deal with files


unit QUB_Viewport_Files;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, Menus, Registry,

RegistryInterface, 

QUB_ListBox,

QUB_Viewport,
QUB_QVO;


const


DEF_TIPSON = True;


type


TQUB_Viewport_Files     = class(TQUB_Viewport)


  ConteinerLeft         :TPanel;
  splLeftRight          :TSplitter;
  ConteinerRight        :TPanel;
  lblFiles              :TLabel;
  PopupMenu             :TPopupMenu;
    CloseItem           :TMenuItem;
    PropertiesItem      :TMenuItem;
    TipsItem            :TMenuItem;


  procedure PropertiesItemClick(Sender: TObject);
  procedure CloseItemClick(Sender: TObject);
  procedure TipsItemClick(Sender: TObject);
  procedure PopupMenuPopup(Sender: TObject); virtual;


protected


  prNoNameCount         :Integer;

  //Tips
  prTipsOn              :Boolean;

  //prOpenDlg             :TOpenDialog;


  procedure ListObjectsDblClick(Sender: TObject);

  procedure DoPropertiesInternal(Read :Boolean); override;

  //Events
  procedure DoOnOpenFile(Sender :TObject);
  procedure DoOnClosingFile(Sender :TObject);
  procedure DoOnCloseFile(Sender :TObject);
  procedure DoOnChangeFile(Sender :TObject);
  procedure DoOnActivateFile(Sender :TObject);
  procedure DoOnDeActivateFile(Sender :TObject);
  procedure DoOnEditFile(Sender :TObject);
  procedure DoOnEnterFile(Sender :TObject);
  procedure DoOnExitFile(Sender :TObject);

  procedure prAddFileEvents(QVO :TQUB_QVO);
  procedure prDelFileEvents(QVO :TQUB_QVO);


public


  ListObjects           :TQUB_ListBox;


  constructor Create(AOwner :TComponent); override;

  procedure Activate; override;
  procedure DeActivate; override;

  function New(const FileName :String; Silent :Boolean) :Boolean; virtual; abstract;
  function Open(const FileName :String; Silent :Boolean) :Boolean; overload; virtual; abstract;
  function Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean; overload; virtual; abstract;
  function Save(const FileName :String; Silent :Boolean) :Boolean; virtual; abstract;
  function SaveAs(const FileName :String; Silent :Boolean) :Boolean; virtual; abstract;
  function Close :Boolean; virtual; abstract;
  function OpenFromString(const MemFile :String) :Boolean; virtual; abstract;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; virtual; abstract;

  function CanCloseActive :Boolean; virtual;

  function Properties :Boolean; virtual; abstract;


end;


implementation


{$R *.DFM}


//******************************************************************************
constructor TQUB_Viewport_Files.Create(AOwner :TComponent);
begin
  inherited;

  ListObjects              := TQUB_ListBox.Create(Self);
  ListObjects.Parent       := ConteinerRight;
  ListObjects.Left         := 0;
  ListObjects.Top          := 13;
  ListObjects.Width        := 183;
  ListObjects.Height       := 239;
  ListObjects.Align        := alClient;
  ListObjects.ItemHeight   := 13;
  ListObjects.MultiSelect  := True;
  ListObjects.ParentColor  := True;
  ListObjects.ParentShowHint := False;
  ListObjects.PopupMenu    := PopupMenu;
  ListObjects.ShowHint     := True;
  ListObjects.TabOrder     := 0;
  ListObjects.OnDblClick   := ListObjectsDblClick;

  //prOpenDlg                := TOpenDialog.Create(Self);
end;

procedure TQUB_Viewport_Files.Activate;
begin
  inherited;
end;

procedure TQUB_Viewport_Files.DeActivate;
begin
  inherited;
end;

//Open
procedure TQUB_Viewport_Files.DoOnOpenFile(Sender :TObject);
begin
  Events.RaiseNotifyEvent('OnOpenFile', Sender);
end;

//Close
procedure TQUB_Viewport_Files.DoOnClosingFile(Sender :TObject);
begin
  Events.RaiseNotifyEvent('OnClosingFile', Sender);
end;

procedure TQUB_Viewport_Files.DoOnCloseFile(Sender :TObject);
begin
  Events.RaiseNotifyEvent('OnCloseFile', Sender);
end;

//Change
procedure TQUB_Viewport_Files.DoOnChangeFile(Sender :TObject);
begin
  Events.RaiseNotifyEvent('OnChangeFile', Sender);
end;


//Activate
procedure TQUB_Viewport_Files.DoOnActivateFile(Sender :TObject);
begin
  DoOnChangeFile(Sender);
end;

//DeActivate
procedure TQUB_Viewport_Files.DoOnDeActivateFile(Sender :TObject);
begin
end;

//Edit
procedure TQUB_Viewport_Files.DoOnEditFile(Sender :TObject);
begin
  DoOnChangeFile(Sender);
end;

//Enter
procedure TQUB_Viewport_Files.DoOnEnterFile(Sender :TObject);
begin
  if not Active then Activate;
end;

//Exit
procedure TQUB_Viewport_Files.DoOnExitFile(Sender :TObject);
begin
end;


//Properties Registry
procedure TQUB_Viewport_Files.DoPropertiesInternal(Read :Boolean);
var
  Reg :TRegistry;
begin
  Reg := TRegistry.Create;

  try

    with Reg do begin
      if OpenKey('Software\QUB\Interface ' + InterfaceName + '\Viewport ' + Name, True) then begin
        if Read then begin
          ConteinerLeft.Width := Round(ReadRegistryFloat(Reg, 'ConteinerLeft Width', 80.0) * ClientWidth / 100);
          prTipsOn := ReadRegistryBoolean(Reg, 'Tips On', DEF_TIPSON);
          TipsItem.Checked := prTipsOn;
        end else begin
          WriteFloat('ConteinerLeft Width', 100 * ConteinerLeft.Width / ClientWidth);
          WriteBool('Tips On', prTipsOn);
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;

end;

function TQUB_Viewport_Files.CanCloseActive :Boolean;
begin
  Result := True;
end;


//File Events;
procedure TQUB_Viewport_Files.prAddFileEvents(QVO :TQUB_QVO);
begin
  QVO.Events.AddEvent('OnActivate',   DoOnActivateFile);
  QVO.Events.AddEvent('OnDeActivate', DoOnDeActivateFile);
  QVO.Events.AddEvent('OnEdit',       DoOnEditFile);
  QVO.Events.AddEvent('OnEnter',      DoOnEnterFile);
  QVO.Events.AddEvent('OnExit',       DoOnExitFile);
end;

procedure TQUB_Viewport_Files.prDelFileEvents(QVO :TQUB_QVO);
begin
  QVO.Events.DelEvent('OnActivate',   DoOnActivateFile);
  QVO.Events.DelEvent('OnDeActivate', DoOnDeActivateFile);
  QVO.Events.DelEvent('OnEdit',       DoOnEditFile);
  QVO.Events.DelEvent('OnEnter',      DoOnEnterFile);
  QVO.Events.DelEvent('OnExit',       DoOnExitFile);
end;


//******************************************************************************
procedure TQUB_Viewport_Files.ListObjectsDblClick(Sender: TObject);
begin
  inherited;
  Properties;
end;

procedure TQUB_Viewport_Files.PropertiesItemClick(Sender: TObject);
begin
  inherited;
  Properties;
end;

procedure TQUB_Viewport_Files.CloseItemClick(Sender: TObject);
var
  i, nSel, iSel :Integer;
  Sels          :array of Integer;
begin
  inherited;
  nSel   := 0;
  for i := 0 to ListObjects.Items.Count - 1 do begin
    if ListObjects.Selected[i] then Inc(nSel);
  end;
  if (0 = nSel) or ((1 = nSel) and (ListObjects.ItemIndex >= 0) and
                    ListObjects.Selected[ListObjects.ItemIndex]) then begin
  // if none selected, or current is only selected
    if CanCloseActive then Close;
  end else begin
  // else choose/close all selected
    SetLength(Sels, nSel);
    iSel := 0;
    for i := 0 to ListObjects.Items.Count - 1 do begin
      if ListObjects.Selected[i] then begin
        Sels[iSel] := i;
        Inc(iSel);
      end;
    end;

    iSel := ListObjects.ItemIndex;
    for i := nSel - 1 downto 0 do begin
      ListObjects.SelQVOAt(Sels[i]);
      if CanCloseActive then Close else Break;
      if Sels[i] <= iSel then Dec(iSel);
    end;
    ListObjects.ItemIndex := iSel;
  end;
end;

procedure TQUB_Viewport_Files.TipsItemClick(Sender: TObject);
var
  i :Integer;
begin
  inherited;
  prTipsOn := not prTipsOn;
  TipsItem.Checked := prTipsOn;
  for i := 0 to ListObjects.Items.Count - 1 do begin
    TQUB_QVO(ListObjects.Items.Objects[i]).TipsOn := prTipsOn;
  end;
end;

procedure TQUB_Viewport_Files.PopupMenuPopup(Sender: TObject);
begin
  inherited;
  CloseItem.Enabled      := (ListObjects.ItemIndex <> -1) and (ListObjects.Items.Count <> 0);
  PropertiesItem.Enabled := (ListObjects.ItemIndex <> -1) and (ListObjects.Items.Count <> 0);
  TipsItem.Checked       := prTipsOn;
end;


//******************************************************************************
end.
