unit QUB_Viewport_Histogram;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Math, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls,
Menus, Buttons, Registry, Clipbrd,
TeeProcs, TeEngine, Chart, Series,

AllocTypes, Alloc,
RegistryInterface,

QUB_QVO,

QUB_Global,
QUB_Viewport,
QUB_Histogram_Amp_Properties, QUB_Histogram_FFT_Properties, QUB_Histogram_Dur_Properties,
QUB_Histogram_Print_Properties,
QUB_Spectrum,
QUB_Node,
QUB_Console, QUB_Rich_Console, 
QUBFits_Dialog, QUBFits,
QUB_Task_Histogram;


{$include 'QUB_Viewport_Histogram.hct'}


const


ORIENT_H = True;
ORIENT_V = not ORIENT_H;


type


TQUB_Viewport_Histogram    = class(TQUB_Viewport)


  ConteinerDur             :TPanel;
  HistogramDur             :TChart;
  btnHistogramDur          :TSpeedButton;
  splFFTDur                :TSplitter;
  ConteinerFFT             :TPanel;
  HistogramFFT             :TChart;
  btnHistogramFFT          :TSpeedButton;
  ConteinerAmp             :TPanel;
  HistogramAmp             :TChart;
  btnHistogramAmp          :TSpeedButton;
  splAmpFFT                :TSplitter;
  mnuHistogram             :TPopupMenu;
  RunHistogramItem         :TMenuItem;
  N1                       :TMenuItem;
  TileItem                 :TMenuItem;
  N2                       :TMenuItem;
  PropertiesItem           :TMenuItem;
  CopyHistogramPictureItem :TMenuItem;
  CurveFittingItem         :TMenuItem;
  HideCurveItem            :TMenuItem;
  CopyHistogramDataItem    :TMenuItem;
  ExtractHistogramItem     :TMenuItem;
  ExtracttoExcelItem       :TMenuItem;
  N3                       :TMenuItem;
  PrintPicture1            :TMenuItem;
  SendToItem               :TMenuItem;


  procedure btnHistogramAmpClick(Sender: TObject);
  procedure btnHistogramFFTClick(Sender: TObject);
  procedure btnHistogramDurClick(Sender: TObject);

  procedure RunHistogramItemClick(Sender: TObject);
  procedure ExtractHistogramItemClick(Sender: TObject);
  procedure TileItemClick(Sender: TObject);
  procedure PropertiesItemClick(Sender: TObject);
  procedure mnuHistogramPopup(Sender: TObject);

  procedure CopyHistogramPictureItemClick(Sender: TObject);
  procedure CurveFittingItemClick(Sender: TObject);
  //procedure tabOptResize(Sender: TObject);
  //procedure tabIdlResize(Sender: TObject);
  procedure ExtracttoExcelItemClick(Sender: TObject);
  procedure HideCurveItemClick(Sender: TObject);
  procedure HistogramFFTMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  procedure CopyHistogramDataItemClick(Sender: TObject);
  procedure PrintPicture1Click(Sender: TObject);
  procedure ExtracttoQtiPlotItemClick(Sender: TObject);


protected


  prData                         :TQUB_QVO;
  prDataIdealize                 :TQUB_QVO;

  prHistOrientation              :Boolean;

  prHistogramAmpPropertiesForm   :TQUB_Histogram_Amp_PropertiesForm;
  prHistogramFFTPropertiesForm   :TQUB_Histogram_FFT_PropertiesForm;
  prHistogramDurPropertiesForm   :TQUB_Histogram_Dur_PropertiesForm;

  prPrintPropertiesForm          :TQUB_Histogram_Print_PropertiesForm;

  prAmpCount                     :Integer;
  prFFTCount                     :Integer;
  prDurCount                     :Integer;
  prFFTLog                       :Boolean;
  prFFTBarsOrLine                :Boolean;
  prDurLog                       :Boolean;
  prDurClass                     :Integer;
  //prDurTimes                     :TArrayOfDouble;
  prAmpAutoscale                 :Boolean;
  prFFTAutoscale                 :Boolean;
  prDurAutoscale                 :Boolean;
  prAmpMin                       :Double;
  prAmpMax                       :Double;
  prFFTMin                       :Double;
  prFFTMax                       :Double;
  prDurMin                       :Double;
  prDurMax                       :Double;
  prFFTMaxWindow                 :Integer;
  prFFTWindowFunc                :Integer;
  prFFTOverlap                   :Integer;

  prHistToExtractToExcel         :INode;

  prFitHist                      :TChart;
  prFitData                      :INode;
  prAmpData                      :INode;
  prFFTData                      :INode;
  prDurData                      :INode;
  prVisDataSet                   :TQUB_FitDataSet;
  prAmpVisDataSet                :TQUB_FitDataSet;
  prFFTVisDataSet                :TQUB_FitDataSet;
  prDurVisDataSet                :TQUB_FitDataSet;


  procedure prShowAmpFit;
  procedure prShowFFTFit;
  procedure prShowDurFit;
  procedure prHideAmpFit;
  procedure prHideFFTFit;
  procedure prHideDurFit;
  procedure prOnCurvesChanged(Sender :TObject);


  procedure prRunHistogramAmp;
  procedure prHistogramAmpDone(Sender :TObject);
  procedure prRunHistogramFFT;
  procedure prHistogramFFTDone(Sender :TObject);
  procedure prRunHistogramDur;
  procedure prHistogramDurDone(Sender :TObject);

  procedure prExtractHistogramAmp;
  procedure prExtractHistogramFFT;
  procedure prExtractHistogramDur;
  procedure prExtractQtiPlot(Script :String; Chart :TChart; PlotTitle :String);

  function prGetHistogramAmpAsNode: INode;
  function prGetHistogramFFTAsNode: INode;
  function prGetHistogramDurAsNode: INode;
  function prGetHistogramAsNode(Histogram :TChart): INode;
  procedure prExtractToExcel( HistNode :INode );
  procedure prExtractExcelPy;

  function prCopyHistogram(Histogram :TChart) :Boolean;

  function prSaveHistogram(Histogram :TChart; const FileName :String) :Boolean;

  procedure prPropertiesHistogramAmp;
  procedure prPropertiesHistogramFFT;
  procedure prPropertiesHistogramDur;

  procedure DoPropertiesInternal(Read :Boolean); override;

  procedure prSetHistOrientation(o :Boolean);


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure DoOnChangeData(Sender :TObject);
  procedure DoOnCloseData(Sender :TObject);


  property HistOrientation                     :Boolean                      read prHistOrientation        write prSetHistOrientation;

  property Data                                :TQUB_QVO                     read prData;

  property PrintPropertiesForm                 :TQUB_Histogram_Print_PropertiesForm  read prPrintPropertiesForm;

end;


implementation


uses


IEEE754,
ShellAPI,
QUB_QtiPlot,
PythonEngine,
QUB_Viewport_Python,
QUB_Python_Thread,
QUB_Node_Python,
QUB_Python_Data,
QUB_PyCmd_MenuButton,
QUB_QFS_Channels,
QUB_GUIInterface,
QUB_Main,
QUB_Profiles;


{$R *.DFM}


//****************************************************************************//
constructor TQUB_Viewport_Histogram.Create(AOwner :TComponent);
begin
  inherited;

  prHistogramAmpPropertiesForm            := TQUB_Histogram_Amp_PropertiesForm.Create(Self);
  prHistogramFFTPropertiesForm            := TQUB_Histogram_FFT_PropertiesForm.Create(Self);
  prHistogramDurPropertiesForm            := TQUB_Histogram_Dur_PropertiesForm.Create(Self);

  prHistogramAmpPropertiesForm.PropertiesToParams;
  prHistogramFFTPropertiesForm.PropertiesToParams;
  prHistogramDurPropertiesForm.PropertiesToParams;

  prPrintPropertiesForm                   := TQUB_Histogram_Print_PropertiesForm.Create(Self);

  prAmpCount                              := prHistogramAmpPropertiesForm.BinCount;
  prFFTCount                              := prHistogramFFTPropertiesForm.BinCount;
  prFFTLog                                := prHistogramFFTPropertiesForm.FFTLog;
  prFFTBarsOrLine                         := prHistogramFFTPropertiesForm.FFTBarsOrLine;
  prDurCount                              := prHistogramDurPropertiesForm.BinCount;
  prDurClass                              := prHistogramDurPropertiesForm.DurClass;
  prDurLog                                := prHistogramDurPropertiesForm.DurLog;
  prAmpAutoscale                          := prHistogramAmpPropertiesForm.Autoscale;
  prFFTAutoscale                          := prHistogramFFTPropertiesForm.Autoscale;
  prDurAutoscale                          := prHistogramDurPropertiesForm.Autoscale;
  prAmpMin                                := prHistogramAmpPropertiesForm.Min;
  prAmpMax                                := prHistogramAmpPropertiesForm.Max;
  prFFTMin                                := prHistogramFFTPropertiesForm.Min;
  prFFTMax                                := prHistogramFFTPropertiesForm.Max;
  prDurMin                                := prHistogramDurPropertiesForm.Min;
  prDurMax                                := prHistogramDurPropertiesForm.Max;
  prFFTMaxWindow                          := Power2(prHistogramFFTPropertiesForm.MaxWindow);
  prFFTWindowFunc                         := prHistogramFFTPropertiesForm.WindowFunc;
  prFFTOverlap                            := Power2(prHistogramFFTPropertiesForm.Overlap);

  HistogramAmp.LeftAxis.Title.Caption     := prHistogramAmpPropertiesForm.YTitle;
  HistogramAmp.BottomAxis.Title.Caption   := prHistogramAmpPropertiesForm.XTitle;

  if prFFTLog then begin
    HistogramFFT.BottomAxis.Title.Caption := 'log10 (Frequency [Hz])';
  end else begin
    HistogramFFT.BottomAxis.Title.Caption := 'Frequency [Hz]';
  end;
  if prDurLog then begin
    HistogramDur.BottomAxis.Title.Caption := 'log10 (Duration [ms])';
  end else begin
    HistogramDur.BottomAxis.Title.Caption := 'Duration [ms]';
  end;
end;


destructor TQUB_Viewport_Histogram.Destroy;
begin
  try

  finally
    inherited;
  end;
end;


//****************************************************************************//
//                                                                            //
//                                   R U N                                    //
//                                                                            //
//****************************************************************************//
procedure TQUB_Viewport_Histogram.prRunHistogramAmp;
var
  Task                         :TQUB_Task_AmpHist;
  i                            :Integer;
begin
  if Data = nil then Exit;

  prHideAmpFit;

  for i := HistogramAmp.SeriesCount - 1 downto 0 do begin
    HistogramAmp.Series[i].Free;
  end;

  Task                         := TQUB_Task_AmpHist.Create(True);
  Task.Data                    := Data;
  Task.DataSource              := QUB_MainForm.DataSource;
  Task.BinCount                := prAmpCount;
  Task.AutoScale               := prAmpAutoscale;
  Task.MinVal                  := prAmpMin;
  Task.MaxVal                  := prAmpMax;
  Task.OnHistogramDone         := prHistogramAmpDone;

  Task.Resume;
end;

procedure TQUB_Viewport_Histogram.prHistogramAmpDone(Sender :TObject);
var
  Task                         :TQUB_Task_AmpHist;
  Bins                         :TPADouble;
  Bars                         :TPADouble;
  BarSeries                    :TBarSeries;
  i                            :Integer;
begin
  Task                         := TQUB_Task_AmpHist(Sender);
  Bins                         := Task.Bins;
  Bars                         := Task.Bars;

  BarSeries                    := TBarSeries.Create(nil);
  BarSeries.BarStyle           := bsCilinder;
  BarSeries.MultiBar           := mbSide;
  BarSeries.UseYOrigin         := False;
  BarSeries.Marks.Visible      := False;
  BarSeries.BarWidthPercent    := 100;
  BarSeries.ParentChart        := HistogramAmp;
  BarSeries.Title              := '';

  for i := 0 to prAmpCount - 1 do begin
    BarSeries.AddXY(Bins[i], Bars[i], '', prHistogramAmpPropertiesForm.BinColor{clTeeColor});
  end;
  BarSeries.Active             := True;
end;

procedure TQUB_Viewport_Histogram.prRunHistogramFFT;
var
  Task                         :TQUB_Task_FFTHist;
  i                            :Integer;
begin
  if Data = nil then Exit;

  prHideFFTFit;

  for i := HistogramFFT.SeriesCount - 1 downto 0 do begin
    HistogramFFT.Series[i].Free;
  end;

  Task                         := TQUB_Task_FFTHist.Create(True);
  Task.Data                    := Data;
  Task.DataSource              := QUB_MainForm.DataSource;
  Task.BinCount                := prFFTCount;
  Task.AutoScale               := prFFTAutoscale;
  Task.MinVal                  := prFFTMin;
  Task.MaxVal                  := prFFTMax;

  Task.MaxWindowSize           := prFFTMaxWindow;
  Task.WindowFunc              := prFFTWindowFunc;
  Task.Overlap                 := prFFTOverlap;
  Task.XLog                    := prFFTLog;
  Task.BarsOrLine              := prFFTBarsOrLine;

  Task.OnHistogramDone         := prHistogramFFTDone;

  Task.Resume;
end;

procedure TQUB_Viewport_Histogram.prHistogramFFTDone(Sender :TObject);
var
  Task                         :TQUB_Task_FFTHist;
  Bins                         :TPADouble;
  Bars                         :TPADouble;
  Count                        :TPAInteger;
  BarSeries                    :TBarSeries;
  LineSeries                   :TAreaSeries;
  i                            :Integer;
begin
  Task                         := TQUB_Task_FFTHist(Sender);

  Bins                         := Task.Bins;
  Bars                         := Task.Bars;
  Count                        := Task.Count;

  if prFFTBarsOrLine then begin
    BarSeries                  := TBarSeries.Create(nil);
    BarSeries.BarStyle         := bsCilinder;
    BarSeries.BarPen.Style     := psClear;
    BarSeries.MultiBar         := mbSide;
    BarSeries.UseYOrigin       := False;
    BarSeries.Marks.Visible    := False;
    BarSeries.BarWidthPercent  := 100;
    BarSeries.ParentChart      := HistogramFFT;
    BarSeries.Title            := '';
    for i := 0 to prFFTCount - 1 do begin
      if Count[i] = 0 then begin
        // exclude
      end else if not (IsNAN(Bars[i]) or IsInfinity(Bars[i])) then begin
        BarSeries.AddXY(Bins[i], Bars[i], '', prHistogramFFTPropertiesForm.BinColor);
      end else begin
        BarSeries.AddXY(Bins[i], 0.0, '', prHistogramFFTPropertiesForm.BinColor);
      end;
    end;
    BarSeries.Active           := True;
  end else begin
    LineSeries                 := TAreaSeries.Create(nil);
    LineSeries.ParentChart     := HistogramFFT;
    LineSeries.Title           := '';
    LineSeries.AreaLinesPen.Style := psSolid;//Clear;
    LineSeries.AreaColor       := prHistogramFFTPropertiesForm.BinColor;
    LineSeries.LinePen.Style   := psSolid;
    LineSeries.LinePen.Color   := clBlack;
    LineSeries.LinePen.Visible := True;
    for i := 0 to prFFTCount - 1 do begin
      if Count[i] = 0 then begin
        // exclude
      end else if not (IsNAN(Bars[i]) or IsInfinity(Bars[i])) then begin
        LineSeries.AddXY(Bins[i], Bars[i], '', prHistogramFFTPropertiesForm.BinColor);
      end else begin
        LineSeries.AddXY(Bins[i], 0.0, '', prHistogramFFTPropertiesForm.BinColor);
      end;
    end;
    LineSeries.Active          := True;
  end;
end;

procedure TQUB_Viewport_Histogram.prRunHistogramDur;
var
  Task                         :TQUB_Task_DurHist;
  i                            :Integer;
begin
  if Data = nil then Exit;
     
  prHideDurFit;

  for i := HistogramDur.SeriesCount - 1 downto 0 do begin
    HistogramDur.Series[i].Free;
  end;

  Task                         := TQUB_Task_DurHist.Create(True);
  Task.Data                    := DataNoder.GetChosenIdealization(TNode.CreateNull);
  Task.BinCount                := prDurCount;
  Task.AutoScale               := prDurAutoscale;
  Task.MinVal                  := prDurMin;
  Task.MaxVal                  := prDurMax;
  Task.Cls                     := prDurClass;
  Task.XLog                    := prDurLog;

  Task.OnHistogramDone         := prHistogramDurDone;

  Task.Resume;
end;

procedure TQUB_Viewport_Histogram.prHistogramDurDone(Sender :TObject);
var
  Task                         :TQUB_Task_DurHist;
  Bins                         :TPADouble;
  Bars                         :TPADouble;
  BarSeries                    :TBarSeries;
  i                            :Integer;
begin
  Task                         := TQUB_Task_DurHist(Sender);
  Bins                         := Task.Bins;
  Bars                         := Task.Bars;

  BarSeries                    := TBarSeries.Create(nil);
  BarSeries.BarStyle           := bsCilinder;
  BarSeries.MultiBar           := mbSide;
  BarSeries.UseYOrigin         := False;
  BarSeries.Marks.Visible      := False;
  BarSeries.BarWidthPercent    := 100;
  BarSeries.ParentChart        := HistogramDur;
  BarSeries.Title              := '';

  for i := 0 to prDurCount - 1 do begin
    BarSeries.AddXY(Bins[i], Bars[i], '', QUB_QFS_QMF_CLASS_COLORS[prDurClass]);
  end;
  BarSeries.Active             := True;
end;

//****************************************************************************//
//                                                                            //
//                               E X T R A C T                                //
//                                                                            //
//****************************************************************************//
procedure TQUB_Viewport_Histogram.prExtractHistogramAmp;
var
  Dlg :TSaveDialog;
  FName :String;
begin
  if HistogramAmp.SeriesCount < 1 then begin
    MessageHelpDlg('No data available for extraction', mtError, [mbOK], 0, 'Error:No_Histogram_Data_Available');
    Exit;
  end;
  Dlg := TSaveDialog.Create(Self);
  try
    if Data <> nil then Dlg.FileName := ChangeFileExt(Data.FileName, 'amphst.txt') else Dlg.FileName := 'amphst.txt';
    Dlg.Filter := 'Text files (*.txt)|*.txt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofOverwritePrompt];
    Dlg.Title := 'Save AMP histogram for Notepad';
    if Dlg.Execute then begin
      FName := ChangeFileExt(Dlg.FileName, '.txt');
      prSaveHistogram(HistogramAmp, ChangeFileExt(Dlg.FileName, '.txt'));
      ShellExecute(Application.Handle, 'open', 'notepad', PChar(FName), PChar(ExtractFilePath(FName)), SW_SHOWNORMAL);
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Viewport_Histogram.prExtractHistogramFFT;
var
  Dlg :TSaveDialog;
  FName :String;
begin
  if HistogramFFT.SeriesCount < 1 then begin
    MessageHelpDlg('No data available for extraction', mtError, [mbOK], 0, 'Error:No_Histogram_Data_Available');
    Exit;
  end;
  Dlg := TSaveDialog.Create(Self);
  try
    if Data <> nil then Dlg.FileName := ChangeFileExt(Data.FileName, 'ffthst.txt') else Dlg.FileName := 'ffthst.txt';
    Dlg.Filter := 'Text files (*.txt)|*.txt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofOverwritePrompt];
    Dlg.Title := 'Save FFT histogram for Notepad';
    if Dlg.Execute then begin
      FName := ChangeFileExt(Dlg.FileName, '.txt');
      prSaveHistogram(HistogramFFT, FName);
      ShellExecute(Application.Handle, 'open', 'notepad', PChar(FName), PChar(ExtractFilePath(FName)), SW_SHOWNORMAL);
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Viewport_Histogram.prExtractHistogramDur;
var
  Dlg :TSaveDialog;
  FName :String;
begin
  if HistogramDur.SeriesCount < 1 then begin
    MessageHelpDlg('No data available for extraction', mtError, [mbOK], 0, 'Error:No_Histogram_Data_Available');
    Exit;
  end;
  Dlg := TSaveDialog.Create(Self);
  try
    if Data <> nil then Dlg.FileName := ChangeFileExt(Data.FileName, 'durhst.txt') else Dlg.FileName := 'durhst.txt';
    Dlg.Filter := 'Text files (*.txt)|*.txt|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofOverwritePrompt];
    Dlg.Title := 'Save DUR histogram for Notepad';
    if Dlg.Execute then begin
      FName := ChangeFileExt(Dlg.FileName, '.txt');
      prSaveHistogram(HistogramDur, ChangeFileExt(Dlg.FileName, '.txt'));
      ShellExecute(Application.Handle, 'open', 'notepad', PChar(FName), PChar(ExtractFilePath(FName)), SW_SHOWNORMAL);
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Viewport_Histogram.prExtractQtiPlot(Script :String; Chart :TChart; PlotTitle :String);
var
  QtiPlot           :TQUB_QtiPlot_Connection;
  Buf               :TPMSingle;
  i, j              :Integer;
begin
  with Chart do begin
    Buf               := nil;
    QtiPlot           := TQUB_QtiPlot_Connection.Create;
    try
      Buf             := SAlloc2(Series[0].Count, SeriesCount+1);
      for i := 0 to Series[0].Count - 1 do begin
        for j := 0 to SeriesCount do begin
          if j = 0 then begin
            Buf[i][j] := Series[0].XValue[i];
          end else begin
            Buf[i][j] := Series[j-1].YValue[i];
          end;
        end;
      end;

      QtiPlot.BringToFront;
      QtiPlot.ExecScript('Histogram', Script);
      QtiPlot.SetArray(Buf[0], Series[0].Count, SeriesCount+1, 'float32', 'a');
      QtiPlot.Exec(Format('add_hist("%s", "%s", "%s", a[:,0], a[:,1], None, *[a[:,i] for i in xrange(2, a.shape[1])])',
                          [PlotTitle,
                           BottomAxis.Title.Caption,
                           LeftAxis.Title.Caption]));

    finally
      QtiPlot.Free;
      Free2(TPMChar(Buf));
    end;
  end;
end;


function  TQUB_Viewport_Histogram.prCopyHistogram(Histogram :TChart) :Boolean;
var
  Series   :TChartSeries;
  Text     :String;
  i, j     :Integer;
begin

  Result   := False;

  try

    Text   := '';

    Series := Histogram.Series[0];

    for i := 0 to Series.Count - 2 do begin
      Text := Text + FloatToStr(Series.XValues[i]);
      for j := 0 to Histogram.SeriesCount - 1 do begin
        Text := Text + #9 + FloatToStr(Histogram.Series[j].YValues[i]);
      end;
      Text := Text + #13#10;
    end;
    i := Series.Count - 1;
    Text := Text + FloatToStr(Series.XValues[i]);
    for j := 0 to Histogram.SeriesCount - 1 do begin
      Text := Text + #9 + FloatToStr(Histogram.Series[j].YValues[i]);
    end;

    Clipboard.SetTextBuf(PChar(Text));
    
  except
  end;

end;

function TQUB_Viewport_Histogram.prSaveHistogram(Histogram :TChart; const FileName :String) :Boolean;
var
  Series      :TChartSeries;
  Text        :String;
  f           :File;
  i, j        :Integer;
  Entry       :Integer;
begin

  Result := False;

  Series := Histogram.Series[0];

  AssignFile(f, FileName);
  try
    Rewrite(f, 1);
  except
    Result := True;
    MessageDlg('Cannot open file ' + FileName + ' for writing', mtError, [mbOK], 0);
    Exit;
  end;

  Entry := Console.NewEntry('Extracting histogram file ' + FileName + ' ...', COLOR_TEXT);

  try

    Text := '';

    for i := 0 to Series.Count - 2 do begin
      Text := Text + FloatToStr(Series.XValues[i]);
      for j := 0 to Histogram.SeriesCount - 1 do begin
        Text := Text + #9 + FloatToStr(Histogram.Series[j].YValues[i]);
      end;
      Text := Text + #13#10;
    end;
    i := Series.Count - 1;
    Text := Text + FloatToStr(Series.XValues[i]);
    for j := 0 to Histogram.SeriesCount - 1 do begin
      Text := Text + #9 + FloatToStr(Histogram.Series[j].YValues[i]);
    end;

    try
      BlockWrite(f, PChar(Text)^, Length(Text));
    except
      Result := True;
      MessageDlg('Could not write data to file ' + FileName, mtError, [mbOK], 0);
    end;

  finally

    try
      CloseFile(f);
    except
    end;

    Console.SetEntry(Entry, 'Extracting histogram file ' + FileName + ' ... Done.', COLOR_GRAY);
    //Beep;

  end;

end;


function TQUB_Viewport_Histogram.prGetHistogramAmpAsNode: INode;
begin
  Result := prGetHistogramAsNode(HistogramAmp);
  // Result['Title'].SetData('Amplitude');
end;

function TQUB_Viewport_Histogram.prGetHistogramFFTAsNode: INode;
begin
  Result := prGetHistogramAsNode(HistogramFFT);
  Result['Title'].SetData('Power Spectrum');
end;

function TQUB_Viewport_Histogram.prGetHistogramDurAsNode: INode;
begin
  Result := prGetHistogramAsNode(HistogramDur);
  // Result['Title'].SetData('');
end;

function TQUB_Viewport_Histogram.prGetHistogramAsNode(Histogram :TChart): INode;
var
  Series      :TChartSeries;
  i, j        :Integer;
  Bins, Bars  :INode;
  Line        :INode;
begin
  Series      := Histogram.Series[0];

  Result      := TNode.Create('Histogram');
  Result['XLabel'].SetData(Histogram.BottomAxis.Title.Caption);
  Result['YLabel'].SetData(Histogram.LeftAxis.Title.Caption);
  Result['BinCount'].SetData( Series.Count );

  Bins        := Result['Bins'];
  Bins.SetupNumData( NODE_DATATYPE_DOUBLE, Series.Count, 1 );
  Bars        := Result['Bars'];
  Bars.SetupNumData( NODE_DATATYPE_DOUBLE, Series.Count, 1 );
  for i := 0 to Series.Count - 1 do begin
    Bins.Data[i] := Series.XValue[i];
    Bars.Data[i] := Series.YValue[i];
  end;

  for j := 1 to Histogram.SeriesCount - 1 do begin
    Series    := Histogram.Series[j];

    Line      := TNode.Create('Line');
    Line.SetupNumData( NODE_DATATYPE_DOUBLE, Series.Count, 1 );
    Result.AppendChild(Line);

    if j < 2 then begin
      Line['Color'].SetData( 0 );
    end else begin
      Line['Color'].SetData( j - 2 );
    end;

    for i := 0 to Series.Count - 1 do begin
      Line.Data[i] := Series.YValue[i];
    end;
  end;
end;

procedure TQUB_Viewport_Histogram.prExtractToExcel( HistNode :INode );
begin
  prHistToExtractToExcel := HistNode;

  if PythonThread.SyncPyProc( prExtractExcelPy, False, True, 2000 ) then begin
    MessageHelpDlg('Python is busy.  Try again later.', mtError, [mbOK], 0, 'Error:Python_Is_Busy');
  end;
end;

procedure TQUB_Viewport_Histogram.prExtractExcelPy;
var
  InvokeCmd  :PPyObject;
  histarg    :PPyObject;
  arglist    :PPyObject;
  cmdresult  :PPyObject;
begin
  with GetPythonEngine do begin
    InvokeCmd   := nil;
    histarg     := nil;
    arglist     := nil;
    cmdresult   := nil;
    try
      InvokeCmd := EvalString('Commands.CallByName');
      histarg   := PyNode_FromNode( prHistToExtractToExcel );
      arglist   := Py_BuildValue('(ssO)', [PChar('Histogram'), PChar('Extract to Excel'), histarg]);
      cmdresult := PyEval_CallObject( InvokeCmd, arglist );
      if cmdresult = nil then PyErr_Print;
    finally
      Py_XDECREF( cmdresult );
      Py_XDECREF( arglist );
      Py_XDECREF( histarg );
      Py_XDECREF( InvokeCmd );
    end;
  end;
end;

procedure TQUB_Viewport_Histogram.prShowAmpFit;
var
  X, Y, F                  :INode;
  i, N                     :Integer;
  Series                   :TChartSeries;
  DataSet                  :TQUB_FitDataSet;
begin
  if prAmpData = nil then begin
    Series                   := HistogramAmp.Series[0];
    N                        := Series.Count;

    prAmpData                := TNode.Create('DataSet');
    X                        := prAmpData['X'];
    X.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
    Y                        := prAmpData['Y'];
    Y.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
    F                        := prAmpData['Line'];
    F.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );

    for i := 0 to N - 1 do begin
      X.Data[i]              := Series.XValue[i];
      Y.Data[i]              := Series.YValue[i];
    end;
  end;

  prFitData            := prAmpData;
  prFitHist            := HistogramAmp;
  FitDlog.Data.AbsX    := True;
  FitDlog.Data.DisplayStyle := QUBFITS_DISPLAY_HIST;
  FitDlog.SetState(ProfileMgr['Fitter-HistogramA'].Active); // init from last Select fit
  DataSet              := TQUB_FitDataSet.Create(prAmpData['X'].DataCount);
  DataSet.AddVar('X', TPADouble(prAmpData['X'].DataPtr));
  DataSet.AddVar('Y', TPADouble(prAmpData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prAmpVisDataSet = nil then begin
    prAmpVisDataSet               := TQUB_FitDataSet.Create;
    prAmpVisDataSet.Assign(DataSet);
    // prAmpVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  prVisDataSet := prAmpVisDataSet;
  DataSet.Free;
  FitDlog.Fitter.Events.AddEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.ShowModal;
  FitDlog.Fitter.Events.DelEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  ProfileMgr['Fitter-HistogramA'].Active := FitDlog.GetState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
  prFitData            := nil;
  prFitHist            := nil;
  prVisDataSet         := nil;
end;

procedure TQUB_Viewport_Histogram.prHideAmpFit;
begin
  if prAmpData <> nil then begin
    while HistogramAmp.SeriesCount > 1 do begin
      HistogramAmp.Series[ HistogramAmp.SeriesCount - 1 ].Free;
    end;
    prAmpData   := nil;
    prAmpVisDataSet.Free;
    prAmpVisDataSet := nil;
  end;
end;

procedure TQUB_Viewport_Histogram.prShowFFTFit;
var
  X, Y, F                  :INode;
  i, N                     :Integer;
  Series                   :TChartSeries;
  DataSet                  :TQUB_FitDataSet;
begin
  if prFFTData = nil then begin
    Series                   := HistogramFFT.Series[0];
    N                        := Series.Count;

    prFFTData                := TNode.Create('DataSet');
    X                        := prFFTData['X'];
    X.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
    Y                        := prFFTData['Y'];
    Y.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
    F                        := prFFTData['Line'];
    F.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );

    for i := 0 to N - 1 do begin
      X.Data[i]              := Series.XValue[i];
      Y.Data[i]              := Series.YValue[i];
    end;
  end;

  prFitData            := prFFTData;
  prFitHist            := HistogramFFT;
  FitDlog.Data.AbsX    := True;
  FitDlog.Data.DisplayStyle := QUBFITS_DISPLAY_HIST;
  FitDlog.SetState(ProfileMgr['Fitter-HistogramF'].Active); // init from last Select fit
  DataSet              := TQUB_FitDataSet.Create(prFFTData['X'].DataCount);
  DataSet.AddVar('X', TPADouble(prFFTData['X'].DataPtr));
  DataSet.AddVar('Y', TPADouble(prFFTData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prFFTVisDataSet = nil then begin
    prFFTVisDataSet               := TQUB_FitDataSet.Create;
    prFFTVisDataSet.Assign(DataSet);
    // prFFTVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  prVisDataSet := prFFTVisDataSet;
  DataSet.Free;
  FitDlog.Fitter.Events.AddEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.ShowModal;
  FitDlog.Fitter.Events.DelEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  ProfileMgr['Fitter-HistogramF'].Active := FitDlog.GetState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
  prFitData            := nil;
  prFitHist            := nil;
  prVisDataSet         := nil;
end;

procedure TQUB_Viewport_Histogram.prHideFFTFit;
begin
  if prFFTData <> nil then begin
    while HistogramFFT.SeriesCount > 1 do begin
      HistogramFFT.Series[ HistogramFFT.SeriesCount - 1 ].Free;
    end;
    prFFTData   := nil;
    prFFTVisDataSet.Free;
    prFFTVisDataSet := nil;
  end;
end;

procedure TQUB_Viewport_Histogram.prShowDurFit;
var
  X, Y, F                  :INode;
  i, N                     :Integer;
  Series                   :TChartSeries;
  DataSet                  :TQUB_FitDataSet;
begin
  if prDurData = nil then begin
    Series                   := HistogramDur.Series[0];
    N                        := Series.Count;

    prDurData                := TNode.Create('DataSet');
    X                        := prDurData['X'];
    X.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
    Y                        := prDurData['Y'];
    Y.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );
    F                        := prDurData['Line'];
    F.SetupNumData( NODE_DATATYPE_DOUBLE, N, 1 );

    for i := 0 to N - 1 do begin
      X.Data[i]              := Series.XValue[i];
      Y.Data[i]              := Series.YValue[i];
    end;
  end;

  prFitData            := prDurData;
  prFitHist            := HistogramDur;
  FitDlog.Data.AbsX    := True;
  FitDlog.Data.DisplayStyle := QUBFITS_DISPLAY_HIST;
  FitDlog.SetState(ProfileMgr['Fitter-HistogramD'].Active); // init from last Select fit
  DataSet              := TQUB_FitDataSet.Create(prDurData['X'].DataCount);
  DataSet.AddVar('X', TPADouble(prDurData['X'].DataPtr));
  DataSet.AddVar('Y', TPADouble(prDurData['Y'].DataPtr));
  FitDlog.Data.UseData(DataSet);
  if prDurVisDataSet = nil then begin
    prDurVisDataSet               := TQUB_FitDataSet.Create;
    prDurVisDataSet.Assign(DataSet);
    // prDurVisDataSet.MakeOwn; // only if you'll modify it e.g. x-shift
  end;
  prVisDataSet := prDurVisDataSet;
  DataSet.Free;
  FitDlog.Fitter.Events.AddEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  FitDlog.ShowModal;
  FitDlog.Fitter.Events.DelEvent('OnCurveParamVal', prOnCurvesChanged);
  prOnCurvesChanged(FitDlog.Fitter);
  ProfileMgr['Fitter-HistogramD'].Active := FitDlog.GetState;   // post as last Select fit settings
  FitDlog.Data.UseData(nil);
  prFitData            := nil;
  prFitHist            := nil;
  prVisDataSet         := nil;
end;

procedure TQUB_Viewport_Histogram.prHideDurFit;
begin
  if prDurData <> nil then begin
    while HistogramDur.SeriesCount > 1 do begin
      HistogramDur.Series[ HistogramDur.SeriesCount - 1 ].Free;
    end;
    prDurData   := nil;
    prDurVisDataSet.Free;
    prDurVisDataSet := nil;
  end;
end;

procedure TQUB_Viewport_Histogram.prOnCurvesChanged(Sender :TObject);
var
  Fitter                       :TQUB_Fits;
  i, N, c                      :Integer;
  Curve                        :INode;
  Line                         :TLineSeries;
  Color                        :TColor;
  NComp                        :Integer;
begin
  Fitter := FitDlog.Fitter;
  NComp  := FitDlog.Fitter.Curve.GetComponentCount();

  while prFitHist.SeriesCount > (NComp+2) do begin
    prFitHist.Series[ prFitHist.SeriesCount - 1 ].Free;
  end;

  N := prFitHist.Series[0].Count;

  while prFitHist.SeriesCount < (NComp+2) do begin
    Line                     := TLineSeries.Create(prFitHist);
    Line.ParentChart         := prFitHist;
    Line.Marks.Visible       := False;
    if prFitHist.SeriesCount = 2 then begin
      Line.LinePen.Width     := 3;
      Color                  := clRed;
    end else begin
      Line.LinePen.Width     := 2;
      Color                  := clBlack;
    end;
    for i := 0 to N - 1 do begin
      Line.AddXY(prFitHist.Series[0].XValue[i], 0.0, '', Color);
    end;
  end;

  Curve := prFitData['Line'];
  Fitter.Curve.EvalArrayAll(prVisDataSet.Data,
                            TPADouble(Curve.DataPtr));
  Line                := TLineSeries( prFitHist.Series[1] );
  for i := 0 to N - 1 do begin
    Line.YValue[i]    := Double(Curve.Data[i]);
  end;

  Curve := prFitData.Find('Component');
  for c := 0 to NComp - 1 do begin
    if Curve.IsNull then Curve := prFitData.AppendChild('Component');
    if LongInt(Curve.DataCount) <> N then Curve.SetupNumData(NODE_DATATYPE_DOUBLE, N, 1);
    Fitter.Curve.EvalArrayOne(prVisDataSet.Data,
                              TPADouble(Curve.DataPtr),
                              c);
    Line                := TLineSeries( prFitHist.Series[1+c+1] );
    for i := 0 to N - 1 do begin
      Line.YValue[i]    := Double(Curve.Data[i]);
    end;

    Curve := Curve.SiblingSameName;
  end;
end;

//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Viewport_Histogram.prPropertiesHistogramAmp;
begin
  prHistogramAmpPropertiesForm.ShowModal;
  if prHistogramAmpPropertiesForm.ModalResult = mrCancel then Exit;
  //Assign properties;
  prAmpCount                              := prHistogramAmpPropertiesForm.BinCount;
  prAmpAutoscale                          := prHistogramAmpPropertiesForm.Autoscale;
  prAmpMin                                := prHistogramAmpPropertiesForm.Min;
  prAmpMax                                := prHistogramAmpPropertiesForm.Max;
  HistogramAmp.LeftAxis.Title.Caption     := prHistogramAmpPropertiesForm.YTitle;
  HistogramAmp.BottomAxis.Title.Caption   := prHistogramAmpPropertiesForm.XTitle;

end;

procedure TQUB_Viewport_Histogram.prPropertiesHistogramFFT;
begin
  prHistogramFFTPropertiesForm.ShowModal;
  if prHistogramFFTPropertiesForm.ModalResult = mrCancel then Exit;
  //Assign properties;
  prFFTCount      := prHistogramFFTPropertiesForm.BinCount;
  prFFTLog        := prHistogramFFTPropertiesForm.FFTLog;
  prFFTBarsOrLine := prHistogramFFTPropertiesForm.FFTBarsOrLine;
  if prFFTLog then begin
    HistogramFFT.BottomAxis.Title.Caption := 'log10 (Frequency [Hz])';
  end else begin
    HistogramFFT.BottomAxis.Title.Caption := 'Frequency [Hz]';
  end;
  prFFTAutoscale  := prHistogramFFTPropertiesForm.Autoscale;
  prFFTMin        := prHistogramFFTPropertiesForm.Min;
  prFFTMax        := prHistogramFFTPropertiesForm.Max;

  prFFTMaxWindow  := Power2(prHistogramFFTPropertiesForm.MaxWindow);
  prFFTWindowFunc := prHistogramFFTPropertiesForm.WindowFunc;
  prFFTOverlap    := Power2(prHistogramFFTPropertiesForm.Overlap);
end;

procedure TQUB_Viewport_Histogram.prPropertiesHistogramDur;
begin
  prHistogramDurPropertiesForm.ShowModal;
  if prHistogramDurPropertiesForm.ModalResult = mrCancel then Exit;
  //Assign properties;
  prDurCount := prHistogramDurPropertiesForm.BinCount;
  prDurClass := prHistogramDurPropertiesForm.DurClass;
  prDurLog   := prHistogramDurPropertiesForm.DurLog;
  if prDurLog then begin
    HistogramDur.BottomAxis.Title.Caption := 'log10 (Duration [ms])';
  end else begin
    HistogramDur.BottomAxis.Title.Caption := 'Duration [ms]';
  end;
  prDurAutoscale := prHistogramDurPropertiesForm.Autoscale;
  prDurMin       := prHistogramDurPropertiesForm.Min;
  prDurMax       := prHistogramDurPropertiesForm.Max;
end;


procedure TQUB_Viewport_Histogram.DoOnChangeData(Sender :TObject);
begin

  prData := TQUB_QVO(Sender);

end;

procedure TQUB_Viewport_Histogram.DoOnCloseData(Sender :TObject);
begin
end;

//******************************************************************************
procedure TQUB_Viewport_Histogram.prSetHistOrientation(o :Boolean);
begin
  if o = HistOrientation then Exit;

  prHistOrientation := o;

  ConteinerDur.Align := alNone;
  splFFTDur.Align    := alNone;
  ConteinerFFT.Align := alNone;
  splAmpFFT.Align    := alNone;
  ConteinerAmp.Align := alNone;

  if HistOrientation = ORIENT_H then begin

    ConteinerAmp.Align := alLeft;
    splAmpFFT.Align    := alLeft;
    ConteinerFFT.Align := alLeft;
    splFFTDur.Align    := alLeft;

    ConteinerAmp.SetBounds(0,                                      0, (Conteiner.ClientWidth - 6) div 3,             Conteiner.ClientHeight);
    splAmpFFT   .SetBounds(ConteinerAmp.Left + ConteinerAmp.Width, 0, 3,                                             Conteiner.ClientHeight);
    ConteinerFFT.SetBounds(splAmpFFT.Left    + splAmpFFT.Width,    0, (Conteiner.ClientWidth - 6) div 3,             Conteiner.ClientHeight);
    splFFTDur   .SetBounds(ConteinerFFT.Left + ConteinerFFT.Width, 0, 3,                                             Conteiner.ClientHeight);
    ConteinerDur.SetBounds(splFFTDur.Left    + splFFTDur.Width,    0, Conteiner.ClientWidth - (splFFTDur.Left + splFFTDur.Width), Conteiner.ClientHeight);

  end else begin

    ConteinerAmp.Align := alTop;
    splAmpFFT.Align    := alTop;
    ConteinerFFT.Align := alTop;
    splFFTDur.Align    := alTop;

    ConteinerAmp.SetBounds(0, 0,                                      Conteiner.ClientWidth, (Conteiner.ClientHeight - 6) div 3);
    splAmpFFT   .SetBounds(0, ConteinerAmp.Top + ConteinerAmp.Height, Conteiner.ClientWidth, 3);
    ConteinerFFT.SetBounds(0, splAmpFFT.Top    + splAmpFFT.Height,    Conteiner.ClientWidth, (Conteiner.ClientHeight - 6) div 3);
    splFFTDur   .SetBounds(0, ConteinerFFT.Top + ConteinerFFT.Height, Conteiner.ClientWidth, 3);
    ConteinerDur.SetBounds(0, splFFTDur.Top    + splFFTDur.Height,    Conteiner.ClientWidth, Conteiner.ClientHeight - (splFFTDur.Top + splFFTDur.Height));

  end;

  ConteinerDur.Align := alClient;
end;


//Registry
procedure TQUB_Viewport_Histogram.DoPropertiesInternal(Read :Boolean);
var
  Reg :TRegistry;
begin

  inherited;

  Reg := TRegistry.Create;

  try

    with Reg do begin
      if OpenKey('Software\QUB\Interface ' + InterfaceName + '\Viewport ' + Name, True) then begin
        if Read then begin

          HistOrientation := ReadRegistryBoolean( Reg, 'Orientation', ORIENT_H);
          if ConteinerAmp.Align = alTop then begin
            ConteinerAmp.Height := Round(ReadRegistryFloat(Reg, 'ConteinerAmp Height', 33.0) * Conteiner.ClientHeight / 100)
          end else if ConteinerAmp.Align = alLeft then begin
            ConteinerAmp.Width := Round(ReadRegistryFloat(Reg, 'ConteinerAmp Width', 33.0) * Conteiner.ClientWidth / 100);
          end;
          if ConteinerFFT.Align = alTop then begin
            ConteinerFFT.Height := Round(ReadRegistryFloat(Reg, 'ConteinerFFT Height', 33.0) * Conteiner.ClientHeight / 100)
          end else if ConteinerFFT.Align = alLeft then begin
            ConteinerFFT.Width := Round(ReadRegistryFloat(Reg, 'ConteinerFFT Width', 33.0) * Conteiner.ClientWidth / 100);
          end;
              
        end else begin
        //Write

          WriteBool('Orientation', HistOrientation);
          WriteFloat('ConteinerAmp Height',   100 * ConteinerAmp.Height / Conteiner.ClientHeight);
          WriteFloat('ConteinerAmp Width',    100 * ConteinerAmp.Width  / Conteiner.ClientWidth);
          WriteFloat('ConteinerFFT Height',   100 * ConteinerFFT.Height / Conteiner.ClientHeight);
          WriteFloat('ConteinerFFT Width',    100 * ConteinerFFT.Width  / Conteiner.ClientWidth);

        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
procedure TQUB_Viewport_Histogram.btnHistogramAmpClick(Sender: TObject);
begin
  inherited;
  prRunHistogramAmp;
end;

procedure TQUB_Viewport_Histogram.btnHistogramFFTClick(Sender: TObject);
begin
  inherited;
  prRunHistogramFFT;
end;

procedure TQUB_Viewport_Histogram.btnHistogramDurClick(Sender: TObject);
begin
  inherited;
  prRunHistogramDur;
end;

procedure TQUB_Viewport_Histogram.RunHistogramItemClick(Sender: TObject);
begin
  inherited;
  if          (mnuHistogram.PopUpComponent = HistogramAmp) or (mnuHistogram.PopUpComponent = btnHistogramAmp) then begin
    prRunHistogramAmp;
  end else if (mnuHistogram.PopUpComponent = HistogramFFT) or (mnuHistogram.PopUpComponent = btnHistogramFFT) then begin
    prRunHistogramFFT;
  end else if (mnuHistogram.PopUpComponent = HistogramDur) or (mnuHistogram.PopUpComponent = btnHistogramDur) then begin
    prRunHistogramDur;
  end;
end;

procedure TQUB_Viewport_Histogram.CopyHistogramPictureItemClick(Sender: TObject);
var
  Chart     :TChart;
  Tree      :INode;
begin
  inherited;

  if not (mnuHistogram.PopUpComponent is TChart) then Exit;

  Chart := mnuHistogram.PopUpComponent as TChart;
  Tree  := prGetHistogramAsNode(Chart);

  if Chart = HistogramAmp then begin
    prPrintPropertiesForm.UseProfile('Amp');
  end else if Chart = HistogramDur then begin
    prPrintPropertiesForm.UseProfile('Dur');
  end else if Chart = HistogramFFT then begin
    prPrintPropertiesForm.UseProfile('FFT');
  end;
  prPrintPropertiesForm.Print(Chart, Tree, True{Copy}, True{Properties});
end;

procedure TQUB_Viewport_Histogram.PrintPicture1Click(Sender: TObject);
var
  Chart     :TChart;
  Tree      :INode;
begin
  inherited;

  if not (mnuHistogram.PopUpComponent is TChart) then Exit;

  Chart := mnuHistogram.PopUpComponent as TChart;
  Tree  := prGetHistogramAsNode(Chart);

  if Chart = HistogramAmp then begin
    prPrintPropertiesForm.UseProfile('Amp');
  end else if Chart = HistogramDur then begin
    prPrintPropertiesForm.UseProfile('Dur');
  end else if Chart = HistogramFFT then begin
    prPrintPropertiesForm.UseProfile('FFT');
  end;
  prPrintPropertiesForm.Print(Chart, Tree, False{Copy}, True{Properties});
end;

procedure TQUB_Viewport_Histogram.CopyHistogramDataItemClick(Sender: TObject);
var
  Chart     :TChart;
begin
  inherited;

  if not (mnuHistogram.PopUpComponent is TChart) then Exit;

  Chart := mnuHistogram.PopUpComponent as TChart;

  prCopyHistogram(Chart);

end;

procedure TQUB_Viewport_Histogram.ExtractHistogramItemClick(Sender: TObject);
begin
  inherited;
  if          (mnuHistogram.PopUpComponent = HistogramAmp) or (mnuHistogram.PopUpComponent = btnHistogramAmp) then begin
    prExtractHistogramAmp;
  end else if (mnuHistogram.PopUpComponent = HistogramFFT) or (mnuHistogram.PopUpComponent = btnHistogramFFT) then begin
    prExtractHistogramFFT;
  end else if (mnuHistogram.PopUpComponent = HistogramDur) or (mnuHistogram.PopUpComponent = btnHistogramDur) then begin
    prExtractHistogramDur;
  end;
end;

procedure TQUB_Viewport_Histogram.ExtracttoQtiPlotItemClick(Sender: TObject);
var
  Script   :String;
begin
  inherited;
  Script   := ChangeFileExt(StringReplace(TMenuItem(Sender).Caption, '&', '', [rfReplaceAll]), '.py');
  if          (mnuHistogram.PopUpComponent = HistogramAmp) or (mnuHistogram.PopUpComponent = btnHistogramAmp) then begin
    prExtractQtiPlot( Script, HistogramAmp, 'Amplitude' );
  end else if (mnuHistogram.PopUpComponent = HistogramFFT) or (mnuHistogram.PopUpComponent = btnHistogramFFT) then begin
    prExtractQtiPlot( Script, HistogramFFT, 'Spectrum' );
  end else if (mnuHistogram.PopUpComponent = HistogramDur) or (mnuHistogram.PopUpComponent = btnHistogramDur) then begin
    prExtractQtiPlot( Script, HistogramDur, 'Duration' );
  end;
end;

procedure TQUB_Viewport_Histogram.ExtracttoExcelItemClick(Sender: TObject);
begin
  inherited;
  if          (mnuHistogram.PopUpComponent = HistogramAmp) or (mnuHistogram.PopUpComponent = btnHistogramAmp) then begin
    prExtractToExcel( prGetHistogramAmpAsNode );
  end else if (mnuHistogram.PopUpComponent = HistogramFFT) or (mnuHistogram.PopUpComponent = btnHistogramFFT) then begin
    prExtractToExcel( prGetHistogramFFTAsNode );
  end else if (mnuHistogram.PopUpComponent = HistogramDur) or (mnuHistogram.PopUpComponent = btnHistogramDur) then begin
    prExtractToExcel( prGetHistogramDurAsNode );
  end;
end;

procedure TQUB_Viewport_Histogram.CurveFittingItemClick(Sender: TObject);
begin
  inherited;

  if          (mnuHistogram.PopUpComponent = HistogramAmp) or (mnuHistogram.PopUpComponent = btnHistogramAmp) then begin
    if HistogramAmp.SeriesCount > 0 then begin
      prShowAmpFit;
    end;
  end else if (mnuHistogram.PopUpComponent = HistogramFFT) or (mnuHistogram.PopUpComponent = btnHistogramFFT) then begin
    if HistogramFFT.SeriesCount > 0 then begin
      prShowFFTFit;
    end;
  end else if (mnuHistogram.PopUpComponent = HistogramDur) or (mnuHistogram.PopUpComponent = btnHistogramDur) then begin
    if HistogramDur.SeriesCount > 0 then begin
      prShowDurFit;
    end;
  end;
end;

procedure TQUB_Viewport_Histogram.HideCurveItemClick(Sender: TObject);
begin
  inherited;

  if          (mnuHistogram.PopUpComponent = HistogramAmp) or (mnuHistogram.PopUpComponent = btnHistogramAmp) then begin
    prHideAmpFit;
  end else if (mnuHistogram.PopUpComponent = HistogramFFT) or (mnuHistogram.PopUpComponent = btnHistogramFFT) then begin
    prHideFFTFit;
  end else if (mnuHistogram.PopUpComponent = HistogramDur) or (mnuHistogram.PopUpComponent = btnHistogramDur) then begin
    prHideDurFit;
  end;
end;

procedure TQUB_Viewport_Histogram.TileItemClick(Sender: TObject);
begin
  inherited;

  prHistOrientation := not prHistOrientation;
  try
    //Store the current settings
    DoPropertiesInternal(False{Read});
  finally
    prHistOrientation := not prHistOrientation;
  end;

  HistOrientation := not HistOrientation;

  DoPropertiesInternal(False{Read});

  DoPropertiesInternal(True{Read});

end;

procedure TQUB_Viewport_Histogram.PropertiesItemClick(Sender: TObject);
begin
  inherited;
  if          (mnuHistogram.PopUpComponent = HistogramAmp) or (mnuHistogram.PopUpComponent = btnHistogramAmp) then begin
    prPropertiesHistogramAmp;
  end else if (mnuHistogram.PopUpComponent = HistogramFFT) or (mnuHistogram.PopUpComponent = btnHistogramFFT) then begin
    prPropertiesHistogramFFT;
  end else if (mnuHistogram.PopUpComponent = HistogramDur) or (mnuHistogram.PopUpComponent = btnHistogramDur) then begin
    prPropertiesHistogramDur;
  end;
end;

procedure TQUB_Viewport_Histogram.mnuHistogramPopUp(Sender :TObject);
var
  IsAmp, IsFFT, IsDur      :Boolean;
  i                        :Integer;
  Items                    :TStringList;
  Item                     :TMenuItem;
begin
  if HistOrientation = ORIENT_H then TileItem.Caption := 'Tile Vertically' else TileItem.Caption := 'Tile Horizontally';

  IsAmp                    := (mnuHistogram.PopUpComponent = btnHistogramAmp) or
                              (mnuHistogram.PopUpComponent = HistogramAmp);
  IsFFT                    := (mnuHistogram.PopUpComponent = btnHistogramFFT) or
                              (mnuHistogram.PopUpComponent = HistogramFFT);
  IsDur                    := (mnuHistogram.PopUpComponent = btnHistogramDur) or
                              (mnuHistogram.PopUpComponent = HistogramDur);

  RunHistogramItem.Visible := IsAmp or IsFFt or IsDur;

  TileItem.Visible         := RunHistogramItem.Visible;
  PropertiesItem.Visible   := RunHistogramItem.Visible;
  CurveFittingItem.Visible := RunHistogramItem.Visible;
  HideCurveItem.Visible    := (IsAmp and (prAmpData <> nil)) or
                              (IsFFT and (prFFTData <> nil)) or
                              (IsDur and (prDurData <> nil));

  for i := SendToItem.Count - 1 downto 2 do begin
    SendToItem.Items[i].Free;
  end;
  Items                    := TStringList.Create;
  Items.Sorted             := True;
  TQUB_QtiPlot_Connection.ListScripts('Histogram', Items);
  for i := 0 to Items.Count - 1 do begin
    Item           := TMenuItem.Create(Self);
    Item.Caption   := ChangeFileExt(Items[i], '');
    Item.OnClick   := ExtracttoQtiPlotItemClick;
    SendToItem.Add(Item);
  end;
  Items.Free;
end;

procedure TQUB_Viewport_Histogram.HistogramFFTMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  v :Double;
  t :String;
begin
  inherited;
  with Sender as TChart do begin
    if SeriesCount > 0 then begin
      t := 'x = ';
      v := Series[0].XScreenToValue(X);
      if prFFTLog then begin
        t := t + FloatToStr(Math.Power(10.0, v));
      end else begin
        t := t + FloatToStr(v);
      end;
      t := t + #13#10;
      v := Series[0].YScreenToValue(Y);
      t := t + 'y = ' + FloatToStr(Math.Power(10, v));
      Hint := t;
    end;
  end;
end;


//******************************************************************************
end.
