unit QUB_Viewport_Model;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls, ExtCtrls, StdCtrls, Menus, Registry,

RegistryInterface,

QUB_Global,

QUB_ListBox,

QUB_Node, QUB_Node_Python,
QUB_Viewport_Files,
QUB_QVO, QUB_QVO_QVM,
QUB_QFS_Channels,
QUB_QVO_QVM_XProperties, QUB_QVO_QVM_State_Properties, QUB_QVO_QVM_Print_Properties,
QUB_Viewport_Model_New_Properties,
QUB_ExpCond_Parameters, QUB_ExpCond_Editor, QUB_ExpCond_Form;


//{$include 'QUB_Viewport_Model.hct'}


type


TQUB_Viewport_Model                             = class(TQUB_Viewport_Files)


  itemUnMerge                                   :TMenuItem;
  DefExpCondItem                                :TMenuItem;
  itemRevert: TMenuItem;


  procedure itemUnMergeClick(Sender: TObject);
  procedure PopupMenuPopup(Sender: TObject); override;
  procedure ConteinerLeftClick(Sender: TObject);
  procedure lblFilesClick(Sender: TObject);
  procedure DefExpCondItemClick(Sender: TObject);
    procedure itemRevertClick(Sender: TObject);


protected


  prNewPropertiesForm                           :TQUB_Viewport_Model_New_PropertiesForm;

  prShortcuts                                   :TFrame;

  prPropertiesForm                              :TQUB_QVO_QVM_XPropertiesForm;
  prStatePropertiesForm                         :TQUB_QVO_QVM_State_PropertiesForm;
  prPrintPropertiesForm                         :TQUB_QVO_QVM_Print_PropertiesForm;
  prDefExpCondForm                              :TQUB_ExpCond_Form;

  prCurrData                                    :TQUB_QVO;

  prDefExpCond                                  :TQUB_ExpCond_Parameters;
  prExpCond                                     :TQUB_ExpCond_Parameters;

  prOpenDlg                                     :TOpenDialog;
  prSaveDlg                                     :TSaveDialog;


  procedure ListObjectsClick(Sender: TObject);

  procedure prInitPython;

  function prCurrModel :TQUB_QVO_QVM;

  procedure prSetExpCond(EC :TQUB_ExpCond_Parameters);
  function prGetDefExpCond: TQUB_ExpCond_Parameters;

  function prGetHasModel :Boolean;
  function prGetFileName :String;

  procedure ModelPropertiesRegistry(Read :Boolean);

  function prDoMerge(Factors :INode): INode;
  function prDoUnmerge(Product, Factors :INode): INode;
  function prMergeReportCallback(arg :INode): INode;

  function prGetModelByFullFileName(FileName :String): TQUB_QVO_QVM; // creates if none


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure Activate; override;
  procedure DeActivate; override;

  //File
  function New(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean; override;
  function Save(const FileName :String; Silent :Boolean) :Boolean; override;
  function SaveAs(const FileName :String; Silent :Boolean) :Boolean; override;
  function Close :Boolean; override;
  function CloseAll :Boolean; override;
  function Revert :Boolean; // put in base vp_files?

  function OpenFromString(const MemFile :String) :Boolean; override;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

  function CanClose :Boolean; override;
  function CanCloseActive :Boolean; override;
  function CanSave :Boolean;

  function CloseModel(Model :TQUB_QVO_QVM) :Boolean;
  function SaveModel(Model :TQUB_QVO_QVM; const FileName :String; Silent :Boolean) :Boolean;
  function SaveAsModel(Model :TQUB_QVO_QVM; const FileName :String; Silent :Boolean) :Boolean;

  function Merge(QVMs :TList): TQUB_QVO_QVM; // result is added as the active model
  function UnMerge(Merged :TQUB_QVO_QVM): Boolean;

  function Properties :Boolean; override;
  function EditDefExpCond: Boolean;

  function ModelCount :Integer;

  procedure DoOnChangeData(Sender :TObject);
  procedure DoOnClosingData(Sender :TObject);


  property CurrModel           :TQUB_QVO_QVM            read prCurrModel;

  property Shortcuts           :TFrame                  read prShortcuts;

  property HasModel            :Boolean                 read prGetHasModel;
  property FileName            :String                  read prGetFileName;

  property DefExpCond          :TQUB_ExpCond_Parameters read prGetDefExpCond;
  property ExpCond             :TQUB_ExpCond_Parameters read prExpCond          write prSetExpCond;


end;


function DllMergeModels(Factors :TQUB_QFS_PNodeMem; reportCB :INodeCFunc ): TQUB_QFS_PNodeMem; stdcall; external 'qubopt.dll' name '_DoMergeModels@8';
function DllUnmergeModels(Product, Factors :TQUB_QFS_PNodeMem; reportCB :INodeCFunc ): Integer; stdcall; external 'qubopt.dll' name '_DoUnmergeModels@12';


implementation


uses


QUB_Python_Model,
QUB_Python_Thread,
QUB_Console, QUB_Rich_Console,
QUB_Viewport_Python,
QUB_QVO_QVD;


{$R *.DFM}


//******************************************************************************
constructor TQUB_Viewport_Model.Create(AOwner :TComponent);
begin

  inherited;

  ListObjects.OnClick                := ListObjectsClick;

  //The DefExpCond are owned by ViewportModel (should it be the Viewport or some other object, like Main?)
  prDefExpCond                       := TQUB_ExpCond_Parameters.Create;

  //ExpCond are a dynamic reference; they point to the default if no other ExpCond is available
  prExpCond                          := prDefExpCond;

  prPropertiesForm                   := TQUB_QVO_QVM_XPropertiesForm.Create(Self);
  prStatePropertiesForm              := TQUB_QVO_QVM_State_PropertiesForm.Create(Self);

  prPrintPropertiesForm              := TQUB_QVO_QVM_Print_PropertiesForm.Create(Self);

  prNewPropertiesForm                := TQUB_Viewport_Model_New_PropertiesForm.Create(Self);

  prDefExpCondForm                   := TQUB_ExpCond_Form.Create(Self);
  prDefExpCondForm.grdExpCond.ShowValueAlways := True;
  prDefExpCondForm.grdExpCond.ShowChannel     := False;

  ModelPropertiesRegistry(True{Read});

  PythonThread.SyncPyProc(prInitPython, True, False, -1);

end;

destructor TQUB_Viewport_Model.Destroy;
begin
  try

    ModelPropertiesRegistry(False{Write});

    prDefExpCond.Free;

  finally

    inherited;

  end;
end;

//******************************************************************************
procedure TQUB_Viewport_Model.prInitPython;
begin
  QUB_Python_ModelView_Init(Self);
  PyNode_PublishNodeFunc( TNodeMeth.Create(prDoMerge), 'MergeModels' );
  PyNode_PublishNodeNodeFunc( TNodeNodeMeth.Create(prDoUnmerge), 'UnmergeModels' );
end;

//Activate/DeActivate **********************************************************
procedure TQUB_Viewport_Model.Activate;
begin
  inherited;
  if CurrModel <> nil then CurrModel.Activate;
end;

procedure TQUB_Viewport_Model.DeActivate;
begin
  inherited;
  if CurrModel <> nil then CurrModel.DeActivate;
end;


//****************************************************************************//
//                                                                            //
//                         F I L E     M E T H O D S                          //
//                                                                            //
//****************************************************************************//
function TQUB_Viewport_Model.New(const FileName :String; Silent :Boolean) :Boolean;
var
  Model    :TQUB_QVO_QVM;
  f        :String;
begin

  //Process FileName;
  f := FileName;
  if f = '' then f := GetBaseStampedFileName(ExpandFileName(prNewPropertiesForm.BaseName), '.qmf');
  //if f = '' then f := GetBaseStampedFileName(ExpandFileName(prNewPropertiesForm.BaseName), prNewPropertiesForm.FileTypeToExt(prNewPropertiesForm.FileFormat));
  // if f = '' then f := GetNewUniqueFileName(ExpandFileName('noname'), prNewPropertiesForm.FileTypeToExt(prNewPropertiesForm.FileFormat), prNoNameCount);
  
  //Create new Model
  Model                        := TQUB_QVO_QVM.Create(Self);

  //Model.ModelObject.Parameters := ExpCond;
  //Testing ..............................
  if Model.UseGlobalCond then begin
    Model.ExpCond              := DefExpCond;
  end else begin
  Model.ExpCond                := ExpCond;
  end;
  //......................................

  Model.Parent                 := ConteinerLeft;

  Model.PropertiesForm         := prPropertiesForm;
  Model.StatePropertiesForm    := prStatePropertiesForm;
  Model.PrintPropertiesForm    := prPrintPropertiesForm;

  Model.TipsOn                 := prTipsOn;

  prNewPropertiesForm.FileName := f;

  prNewPropertiesForm.OnGetProperties := Model.ModelObject.OnGetPropertiesNew;
  prNewPropertiesForm.OnPutProperties := Model.ModelObject.OnPutPropertiesNew;

  if Silent then begin
    prNewPropertiesForm.FormShow(prNewPropertiesForm);
    prNewPropertiesForm.ModalResult := mrOK;
    prNewPropertiesForm.btnOKClick(prNewPropertiesForm.btnOK);
  end else begin
    prNewPropertiesForm.ShowModal;
  end;

  Result := prNewPropertiesForm.ModalResult = mrCancel;

  if not Result then begin
    Model.ModelObject.OpenSilently := Silent;
    f := Model.ModelObject.FileName;
    Result := Model.New(f);
  end;

  if not Result then begin
    ListObjects.AddQVO(Model, ExtractFileName(f));
    DoOnOpenFile(Model);
    DoOnChangeFile(Model);
    prAddFileEvents(Model);
  end else begin
    prDelFileEvents(Model);
    Model.Free;
  end;

end;

function TQUB_Viewport_Model.Open(const FileName :String; Silent :Boolean) :Boolean;
var
  Model    :TQUB_QVO_QVM;
  Count    :Integer;
  f, fi    :String;
  m        :Integer;
begin

  Result := False;

  //Process FileName;
  f := FileName;
  if f = '' then begin
    if prOpenDlg = nil then prOpenDlg := TOpenDialog.Create(Application);
    prOpenDlg.Filter := 'All QUB model files (*.qmf; *.mdl)|*.qmf; *.mdl|' +
                  'QUB model files (*.qmf)|*.qmf|' +
                  'QUB old model files (*.mdl)|*.mdl|' +
                  'Synaptosoft model files (*.mod)|*.mod|' +
                  'All files (*.*)|*.*';
    prOpenDlg.Options := [ofPathMustExist, ofFileMustExist, ofAllowMultiSelect, ofEnableSizing];
    if not prOpenDlg.Execute then Exit;
  end;

  if f <> '' then begin
    Count := 1;
  end else begin
    Count := prOpenDlg.Files.Count;
  end;

  for m := Count - 1 downto 0 do begin

    if f = '' then fi := prOpenDlg.Files[m] else fi := f;

    //Create new Model and open the file

    Model                        := TQUB_QVO_QVM.Create(Self);

    //Model.ModelObject.Parameters := ExpCond;
    //Testing ..............................
    if Model.UseGlobalCond then begin
      Model.ExpCond              := DefExpCond;
    end else begin
    Model.ExpCond                := ExpCond;
    end;
    //......................................

    Model.Parent                 := ConteinerLeft;

    Model.PropertiesForm         := prPropertiesForm;
    Model.StatePropertiesForm    := prStatePropertiesForm;
    Model.PrintPropertiesForm    := prPrintPropertiesForm;

    Model.TipsOn                 := prTipsOn;

    Result                       := Model.Open(fi);

    if not Result then begin
      ListObjects.AddQVO(Model, ExtractFileName(fi));
      DoOnOpenFile(Model);
      DoOnChangeFile(Model);
      prAddFileEvents(Model);
    end else begin
      prDelFileEvents(Model);
      Model.Free;
      MessageDlg('Could not open file ' + fi, mtError, [mbOK], 0);
    end;

  end;

end;

function TQUB_Viewport_Model.Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean;
var
  Model    :TQUB_QVO_QVM;
begin

  //Create new Model and open the file

  Model                        := TQUB_QVO_QVM.Create(Self);

  //Model.ModelObject.Parameters := ExpCond;
  //Testing ..............................
  if Model.UseGlobalCond then begin
    Model.ExpCond              := DefExpCond;
  end else begin
  Model.ExpCond                := ExpCond;
  end;
  //......................................

  Model.Parent                 := ConteinerLeft;

  Model.PropertiesForm         := prPropertiesForm;
  Model.StatePropertiesForm    := prStatePropertiesForm;
  Model.PrintPropertiesForm    := prPrintPropertiesForm;

  Model.TipsOn                 := prTipsOn;

  Result                       := Model.Open(FileName);

  if not Result then begin
    ListObjects.AddQVO(Model, ExtractFileName(FileName));
    DoOnOpenFile(Model);
    DoOnChangeFile(Model);
    prAddFileEvents(Model);
  end else begin
    prDelFileEvents(Model);
    Model.Free;
    MessageDlg('Could not open file ' + FileName, mtError, [mbOK], 0);
  end;

end;

function TQUB_Viewport_Model.Save(const FileName :String; Silent :Boolean) :Boolean;
var
  Model :TQUB_QVO_QVM;
begin

  Model  := CurrModel;
  Result := Model = nil;
  if Result then Exit;

  Result := SaveModel(Model, FileName, Silent);

end;

function TQUB_Viewport_Model.SaveAs(const FileName :String; Silent :Boolean) :Boolean;
var
  Model    :TQUB_QVO_QVM;
begin

  Model    := CurrModel;
  Result   := Model = nil;
  if Result then Exit;

  Result := SaveAsModel(Model, FileName, Silent);

end;

function TQUB_Viewport_Model.Close :Boolean;
var
  Model    :TQUB_QVO_QVM;
begin
  Result := False;

  Model   := CurrModel;
  if Model = nil then Exit;

  Result := CloseModel(Model);

end;

function TQUB_Viewport_Model.SaveModel(Model :TQUB_QVO_QVM; const FileName :String; Silent :Boolean) :Boolean;
var
  FName :String;
begin

  Result := Model.LockedAny;
  if Result then Exit;

  //Process FileName
  FName := FileName;
  if FName = '' then FName := Model.FileName;
  if (Model.Saved) and (Model.CanSaveAs(FName, True)) then begin
    Result := Model.Save(FName);
    if Model.ModelObject.Opened then begin
      ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Model)] := ExtractFileName(Model.FileName);
      DoOnChangeFile(Model);
    end else begin
      Result := True;
      CloseModel(Model);
    end;
  end else begin
    Result := SaveAsModel(Model, FName, Silent);
  end;

end;

function TQUB_Viewport_Model.SaveAsModel(Model :TQUB_QVO_QVM; const FileName :String; Silent :Boolean) :Boolean;
var
  f        :String;
  r        :Integer;
begin

  Result := Model.LockedAny;
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Model.FileName;

  if prSaveDlg = nil then prSaveDlg := TSaveDialog.Create(Application);
  prSaveDlg.Filter := 'All QUB model files (*.qmf;*.mdl)|*.qmf;*.mdl|' +
                'QUB model files (*.qmf)|*.qmf|' +
                'QUB old model files (*.mdl)|*.mdl|' +
                'All files (*.*)|*.*';

  if Model.FileExt = '.mdl' then begin
    prSaveDlg.FilterIndex := 3;
  end else begin
    prSaveDlg.FilterIndex := 2;
  end;

  prSaveDlg.Options := [ofPathMustExist{, ofOverwritePrompt}, ofEnableSizing];
  prSaveDlg.Title := 'Save model file as';

  try

    repeat

      prSaveDlg.FileName := ChangeFileExt(f, '');
      if prSaveDlg.Execute then begin
        f := prSaveDlg.FileName;
        case prSaveDlg.FilterIndex of
          2: f := ChangeFileExt(f, '.qmf');
          3: f := ChangeFileExt(f, '.mdl');
        else
            f := ChangeFileExt(f, Model.FileExt);
        end;//case FilterIndex
      end else begin
        Result := True;
        Exit;
      end;

      if (FileExists(f)) and (f <> ExpandFileName(Model.FileName)) then begin
        case MessageDlg('File ''' + f + ''' exists. Do you want to overwrite it ?', mtWarning, [mbYes, mbNo, mbCancel], 0) of
          mrCancel:
            begin
              Result := True;
              Exit;
            end;
          mrNo:
            Continue;
        end;//case
      end;

      if Model.CanSaveAs(f, True){Ask confirmation for losing features} then begin
        Result := Model.Save(f);
        Exit;
      end else begin
        r := MessageHelpDlg('Model file cannot be saved in the selected format.'#13#10'Please choose one of the writable formats.', mtError, [mbOK, mbCancel], 0, 'Error:Model_File_Format');
        Result := r = mrCancel;
        if Result then Exit;
      end;

    until (Result);

  finally

    if Model.ModelObject.Opened then begin
      ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Model)] := ExtractFileName(Model.FileName);
      DoOnChangeFile(Model);
    end else begin
      Result := True;
      CloseModel(Model);
    end;

  end;

end;

function TQUB_Viewport_Model.CloseModel(Model :TQUB_QVO_QVM) :Boolean;
begin

  DoOnClosingFile(Model);

  Result := Model.Close;
  if Result then Exit;

  DoOnCloseFile(Model);

  prDelFileEvents(Model);

  ListObjects.DelQVO(Model);
  Model.Free;

  if ListObjects.Items.Count = 0 then DoOnChangeFile(nil);

end;


function TQUB_Viewport_Model.CloseAll :Boolean;
begin
  Result := False;
  while CurrModel <> nil do begin
    Result := Close;
    if Result then Break;
  end;
end;


function TQUB_Viewport_Model.Revert :Boolean;
var
  FName          :String;
begin
  Result := not ((CurrModel <> nil) and
                 (CurrModel.Saved) and
                 (CurrModel.Modified));
  if not Result then begin
    FName        := CurrModel.FileName;
    CurrModel.ModelObject.UnModify;
    Result       := CloseModel(CurrModel);
    if not Result then begin
      Result     := Open(FName, True{Silent});
    end;
  end;
end;


function TQUB_Viewport_Model.CanCloseActive :Boolean;
var
  Model :TQUB_QVO_QVM;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  Model   := CurrModel;
  if Model = nil then Exit;

  if Model.LockedAny then begin
    MessageHelpDlg('File ''' + Model.FileName + ''' is executing a task and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end else if Model.LockedClose then begin
    MessageHelpDlg('File ''' + Model.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end;

  if not Result then Exit;

  if Model.Modified then begin
    case MessageDlg('File ''' + Model.FileName + ''' is modified.'#13#10'Do you want to save it ?', mtConfirmation, [mbYes, mbNo, mbCancel], 0) of
      mrCancel:
      //Cancel
        begin
          Result := False;
          Exit;
        end;

      mrYes:
      //Yes
        begin
          if Model.Saved then begin
            Result := not Model.Save(Model.FileName);
          end else begin
            Result := not SaveAs('', False{Silent});
          end;
        end;

    end;//case r
  end;

end;

function TQUB_Viewport_Model.CanClose :Boolean;
var
  amo :TQUB_QVO_QVM;
  qvm :TQUB_QVO_QVM;
  All :Boolean;
  i   :Integer;
  r   :Integer;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  All := False;

  amo := CurrModel;

  try

    for i := ListObjects.Items.Count - 1 downto 0 do begin

      qvm := TQUB_QVO_QVM(ListObjects[i]);

      if qvm.LockedAny then begin
        MessageHelpDlg('File ''' + qvm.FileName + ''' is executing a task and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end else if qvm.Lockedclose then begin
        MessageHelpDlg('File ''' + qvm.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end;

      if not Result then Exit;

      if qvm.Modified then begin

        ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(qvm);
        qvm.Activate;

        if All then begin

          if qvm.Saved then begin
            Result := not qvm.Save(qvm.FileName);
          end else begin
            Result := not SaveAs('', False{Silent});
          end;
          if not Result then Exit;

        end else begin

          r := MessageDlg('File ''' + qvm.FileName + ''' is modified.'#13#10'Do you want to save it ?', mtConfirmation, [mbYes, mbNo, mbYesToAll, mbNoToAll, mbCancel], 0);

          case r of

            mrCancel:
            //Cancel
              begin
                Result := False;
                Exit;
              end;

            mrYes:
            //Yes
              begin
                if qvm.Saved then begin
                  Result := not qvm.Save(qvm.FileName);
                end else begin
                  Result := not SaveAs('', False{Silent});
                end;
                if not Result then Exit;
              end;

            //mrNo:
            //No

            mrNotoAll://9
            //NotoAll
              begin
                Exit;
              end;

            mrYestoAll://10:
            //YestoAll
              begin
                All := True;
                if qvm.Saved then begin
                  Result := not qvm.Save(qvm.FileName);
                end else begin
                  Result := not SaveAs('', False{Silent});
                end;
                if not Result then Exit;
              end;

          end;//case r

        end;//All

      end;

    end;

  finally

    if amo <> nil then begin
      ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(amo);
      if ListObjects.ItemIndex >= 0 then amo.Activate;
    end;

  end;

end;

function TQUB_Viewport_Model.CanSave :Boolean;
begin
  if CurrModel <> nil then Result := CurrModel.Modified else Result := False;
end;

function TQUB_Viewport_Model.OpenFromString(const MemFile :String) :Boolean;
var
  Model    :TQUB_QVO_QVM;
  f        :String;
begin
  Result := False;

  //Process FileName;
  f := MemFile;
  if f = '' then Exit;

  //Create new Data and open the file
  Model                        := TQUB_QVO_QVM.Create(Self);

  //Model.ModelObject.Parameters := ExpCond;
  //Testing ..............................
  if Model.UseGlobalCond then begin
    Model.ExpCond              := DefExpCond;
  end else begin
  Model.ExpCond                := ExpCond;
  end;
  //......................................

  Model.Parent                 := ConteinerLeft;

  Model.TipsOn                 := prTipsOn;

  Result                       := Model.OpenFromString(f);

  if not Result then begin
    ListObjects.AddQVO(Model, ExtractFileName(f));
    DoOnOpenFile(Model);
    DoOnChangeFile(Model);
    prAddFileEvents(Model);
  end else begin
    prDelFileEvents(Model);
    Model.Free;
    MessageHelpDlg('Could not open file ' + f, mtError, [mbOK], 0, 'Error:Open_File');
  end;

end;

function TQUB_Viewport_Model.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  Result := False;

  if CurrModel <> nil then Result := CurrModel.SaveToString(MemFile, FileType);

end;

function TQUB_Viewport_Model.Properties :Boolean;
begin
  Result := False;
  if CurrModel <> nil then CurrModel.Properties else Result := True;
end;

function TQUB_Viewport_Model.EditDefExpCond :Boolean;
begin
  Result := False;
  prDefExpCondForm.grdExpCond.ExpCond := DefExpCond;
  prDefExpCondForm.ShowModal;
  prDefExpCondForm.grdExpCond.ExpCond := nil;
end;

procedure TQUB_Viewport_Model.DoOnChangeData(Sender :TObject);
begin
  if Sender = nil then begin
    ExpCond := DefExpCond;
  end else begin
    ExpCond := TQUB_QVO_QVD(Sender).ExpCond;
  end;
end;

procedure TQUB_Viewport_Model.DoOnClosingData(Sender :TObject);
begin
  if (Sender <> nil) and (ExpCond <> nil) then begin
    if ExpCond = TQUB_QVO_QVD(Sender).ExpCond then begin
      ExpCond := DefExpCond;
    end;
  end;
end;

//****************************************************************************//
//                                                                            //
//                                  O T H E R                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Viewport_Model.prSetExpCond(EC :TQUB_ExpCond_Parameters);
var
  Model :TQUB_QVO_QVM;
  i     :Integer;
begin
  if EC = nil then EC := prDefExpCond;
  if EC <> prExpCond then begin
    prExpCond := EC;

    for i := 0 to ListObjects.Items.Count - 1 do begin
      Model := TQUB_QVO_QVM(ListObjects[i]);
      if not Model.UseGlobalCond then begin
        Model.ExpCond              := EC;
      end;
    end;
  end;
end;

function TQUB_Viewport_Model.prGetDefExpCond: TQUB_ExpCond_Parameters;
(*var
  EC           :TQUB_ExpCond_Parameters;
  i            :Integer;
  Par          :TQUB_ExpCond_PParameter; *)
begin
  Result       := prDefExpCond;
  (*
  EC           := ExpCond;
  if EC <> nil then begin
    for i := 0 to EC.Count - 1 do begin
      Par      := EC[i];
      if prDefExpCond.ParameterByName[Par.Name] = nil then begin
        prDefExpCond.AddPar(Par.Name, Par.Value, Par.Index, Par.IsVar, Par.Channel);
      end;
    end;
  end;
  *)
end;

function TQUB_Viewport_Model.prCurrModel :TQUB_QVO_QVM;
begin
  Result := nil;
  if ListObjects = nil then Exit;
  if ListObjects.Items.Count = 0  then Exit;
  if ListObjects.ItemIndex   = -1 then Exit;
  Result := TQUB_QVO_QVM(ListObjects.Items.Objects[ListObjects.ItemIndex]);
end;

function TQUB_Viewport_Model.ModelCount :Integer;
begin
  Result := ListObjects.Items.Count;
end;

function TQUB_Viewport_Model.prGetHasModel :Boolean;
begin
  Result := CurrModel <> nil;
end;

function TQUB_Viewport_Model.prGetFileName :String;
begin
  if CurrModel <> nil then Result := CurrModel.FileName else Result := '';
end;


//Registry *********************************************************************
//TODO - is the parameters section of model still utilized ?
//     - the 'Default Parameters' section exists for read but not write ?
procedure TQUB_Viewport_Model.ModelPropertiesRegistry(Read :Boolean);
var
  Reg     :TRegistry;
  i       :Integer;
  n       :String;
  v, cv   :Double;
  c, x    :Integer;
begin
  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Viewport Model\Parameters', True) then begin

        //Parameters
        for i := 0 to QUB_QFS_QMF_MAX_PARAMETER - 1 do begin

          if Read then begin
          //Read

            if ValueExists('Name ' + IntToStr(i)) then begin

              n  := ReadString('Name ' + IntToStr(i));
              v  := ReadRegistryFloat(Reg, 'Value '      + IntToStr(i), 0);
              cv := ReadRegistryFloat(Reg, 'Cond_Value ' + IntToStr(i), v);
              //x  := ReadRegistryInteger(Reg, 'Index '    + IntToStr(i), 1);
              //c  := ReadRegistryInteger(Reg, 'Channel '  + IntToStr(i), 1);
              x  := ReadRegistryInteger(Reg, 'Index '    + IntToStr(i), -11);
              c  := ReadRegistryInteger(Reg, 'Channel '  + IntToStr(i), -11);

              // no reason for it to be here;
              // and it overwrites user value each time you sim
              if n <> 'DeadTime' then begin
              DefExpCond.AddPar(n, v, x, False, c);
              DefExpCond[DefExpCond.Count - 1].Cond_Value := cv;
              end;

            end else begin

              Break;

            end;

          end else begin
          //Write

            if i >= DefExpCond.Count then begin

              DeleteValue('Name '         + IntToStr(i));
              DeleteValue('Value '        + IntToStr(i));
              DeleteValue('Cond_Value '   + IntToStr(i));
              DeleteValue('Index '        + IntToStr(i));
              DeleteValue('Channel '      + IntToStr(i));

            end else begin

              WriteString ('Name '        + IntToStr(i), DefExpCond[i].Name);
              WriteFloat  ('Value '       + IntToStr(i), DefExpCond[i].Value);
              WriteFloat  ('Cond_Value '  + IntToStr(i), DefExpCond[i].Cond_Value);
              WriteInteger('Index '       + IntToStr(i), DefExpCond[i].Index);
              WriteInteger('Channel '     + IntToStr(i), DefExpCond[i].Channel);

            end;

          end;

        end;//for i

        if Read then begin
        //Default parameters
        {
          Read them only if they don't exist.
          The point is to already have a Voltage and a Ligand, but once they are set,
          they become ordinary parameters.
          LM.
        }
          if DefExpCond.ParameterByName['Ligand'] = nil then begin
            //DefExpCond.AddPar('Ligand', 1, 1, False, 1);
            DefExpCond.AddPar('Ligand', 1, -1, False, -1);
          end;
          if DefExpCond.ParameterByName['Voltage'] = nil then begin
            //DefExpCond.AddPar('Voltage', 0, 2, False, 2);
            DefExpCond.AddPar('Voltage', 0, -1, False, -1);
          end;
        end;

      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;

function TQUB_Viewport_Model.prDoMerge(Factors :INode): INode;
var
  Prod      :TQUB_QFS_PNodeMem;
begin
  try
    Prod      := DllMergeModels( Factors.GetNodeMem, TNodeCFunc.Wrap(TNodeMeth.Create(prMergeReportCallback)) );
    Result    := TNode.CreateUsing( Prod );
    Node_DECREF( Prod );
  except
    Result    := TNode.CreateNull;
  end;
end;


function TQUB_Viewport_Model.prDoUnmerge(Product, Factors :INode): INode;
begin
  Result    := TNode.Create('');
  try
    Result.SetData( DllUnmergeModels( Product.GetNodeMem, Factors.GetNodeMem, TNodeCFunc.Wrap(TNodeMeth.Create(prMergeReportCallback)) ) );
  except
    Result.SetData( -666 );
  end;
end;


function TQUB_Viewport_Model.prMergeReportCallback(arg :INode): INode;
begin
  Result := arg;
  Console.NewEntry(arg.DataAsString, COLOR_TEXT);
end;

function TQUB_Viewport_Model.Merge(QVMs :TList): TQUB_QVO_QVM;
// result is added as the active model
var
  Factors           :INode;
  MergedNode        :INode;
  i                 :Integer;
begin
  Factors           := TNode.Create('Factors');
  for i := 0 to QVMs.Count - 1 do begin
    Factors.AppendChild( TQUB_QVO_QVM(QVMs[i]).GetAsNode );
  end;

  MergedNode        := prDoMerge(Factors);
  if MergedNode.IsNull then begin
    MessageHelpDlg('Could not merge models.', mtError, [mbOK], 0, 'Error:Merge');
    Result := nil;
    Exit;
  end;

  if not New('', True) then begin
    Result := CurrModel;
    Result.SetAsNode( MergedNode );
    Result.Factors    := Factors;
  end else begin
    Result := nil;
  end;
end;

function TQUB_Viewport_Model.UnMerge(Merged :TQUB_QVO_QVM): Boolean;
var
  Factor           :INode;
  FactorName       :String;
  FactorObject     :TQUB_QVO_QVM;
begin
  Result           := False;
  if not Merged.Factors.IsNull then begin
    if prDoUnmerge(Merged.GetAsNode, Merged.Factors).Data[0] < 0 then begin
      MessageHelpDlg('Could not unmerge model.', mtError, [mbOK], 0, 'Error:Merge');
      Result       := True;
      Exit;
    end;

    Factor := Merged.Factors.Find('ModelFile');
    while not Factor.IsNull do begin
      FactorName := Factor['FileName'].DataAsString;
      FactorObject := prGetModelByFullFileName( FactorName );
      if (FactorObject <> nil) then begin
        Factor['FileName'].SetData( FactorObject.FileName );
        if FactorObject.LockedAny then begin
          Console.NewEntry( 'Skipping ' + ExtractFileName(FactorObject.FileName) + ' because it is in use.', COLOR_TEXT );
        end else begin
          FactorObject.SetAsNode( Factor );
        end;
      end;

      Factor := Factor.SiblingSameName;
    end;
  end;
end;

function TQUB_Viewport_Model.prGetModelByFullFileName(FileName :String): TQUB_QVO_QVM;
// creates if none
var
  i           :Integer;
begin
  for i := 0 to ListObjects.Items.Count - 1 do begin
    Result    := TQUB_QVO_QVM( ListObjects[i] );
    if Result.FileName = FileName then Exit;
  end;

  Result := nil;
  if not New(FileName, True) then begin
    Result := CurrModel;
  end;
end;

procedure TQUB_Viewport_Model.PopupMenuPopup(Sender: TObject);
begin
  inherited;
  itemUnMerge.Visible := (CurrModel <> nil) and not CurrModel.Factors.IsNull;
  itemRevert.Visible  := (CurrModel <> nil) and (CurrModel.Saved) and (CurrModel.Modified);
end;

procedure TQUB_Viewport_Model.itemUnMergeClick(Sender: TObject);
begin
  inherited;
  if CurrModel <> nil then UnMerge(CurrModel);
end;

procedure TQUB_Viewport_Model.itemRevertClick(Sender: TObject);
begin
  inherited;
  Revert;
end;

procedure TQUB_Viewport_Model.DefExpCondItemClick(Sender: TObject);
begin
  inherited;
  EditDefExpCond;
end;

procedure TQUB_Viewport_Model.ConteinerLeftClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_Model.lblFilesClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_Model.ListObjectsClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;


//******************************************************************************
end.
