unit QUB_Viewport_Model_New_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls, StdCtrls, ExtCtrls, Registry,

RegistryInterface,

QUB_Classic_Properties,

QUB_QFS_QMF,

QUB_QVO_QVM;

{$include 'QUB_Viewport_Model_New_Properties.hct'}


const

DEF_BASENAME = 'noname';


type


TQUB_Viewport_Model_New_PropertiesForm = class(TQUB_Classic_PropertiesForm)


  Label1                 :TLabel;
  Label2                 :TLabel;

  txtFileName            :TEdit;
  btnBrowseFileName      :TButton;

  Panel1                 :TPanel;
  optFileFormat0         :TRadioButton;
  optFileFormat1         :TRadioButton;
  Label11: TLabel;
  txtBaseName: TEdit;


  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
  procedure FormShow(Sender: TObject);
  procedure btnBrowseFileNameClick(Sender: TObject);
  procedure optFileFormatClick(Sender: TObject);
  procedure btnOKClick(Sender: TObject);
  procedure txtFileNameExit(Sender: TObject);
  procedure txtFileNameKeyPress(Sender: TObject; var Key: Char);
  procedure txtBaseNameKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);


protected


  prBaseName                              :String;
  prFileName                              :String;
  prFileFormat                            :TQUB_QFS_QMF_FileType;


  procedure prSetBaseName(f :String);
  procedure prSetFileName(f :String);
  procedure prSetFileFormat(f :TQUB_QFS_QMF_FileType);

  procedure prPropertiesRegistry(Read :Boolean);


public


  function FileTypeToExt(FType :TQUB_QFS_QMF_FileType) :String;


  property BaseName                       :String                read prBaseName                  write prSetBaseName;
  property FileName                       :String                read prFileName                  write prSetFileName;
  property FileFormat                     :TQUB_QFS_QMF_FileType read prFileFormat                write prSetFileFormat;


end;


implementation


{$R *.DFM}


//******************************************************************************
function TQUB_Viewport_Model_New_PropertiesForm.FileTypeToExt(FType :TQUB_QFS_QMF_FileType) :String;
begin
  Result := '';
  case FType of
    QFS_QMF_TYPE_QMF:
      begin
        Result := '.qmf';
      end;
    QFS_QMF_TYPE_MDL:
      begin
        Result := '.mdl';
      end;
  end;//case
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.prSetBaseName(f :String);
var
  baselen     :Integer;
  fdir, fname :String;
begin
  baselen     := Length(prBaseName);
  fname       := ExtractFileName(prFileName);
  if Copy(fname, 1, baselen) = prBaseName then begin
    fdir      := ExtractFileDir(prFileName);
    // big nasty surprise: Copy is 1-based
    fname     := Copy(fname, baselen+1, 10000);
    FileName  := fdir+'\'+f+fname;
  end;
  prBaseName       := f;
  txtBaseName.Text := f;
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.prSetFileName(f :String);
begin
  if ExtractFileExt(f) = '.mdl' then begin
    FileFormat := QFS_QMF_TYPE_MDL;
    prFileName := ChangeFileExt(f, '');
  end else if ExtractFileExt(f) = '.qmf' then begin
    FileFormat := QFS_QMF_TYPE_QMF;
    prFileName := ChangeFileExt(f, '');
  end else begin
    prFileName := f;
  end;
  txtFileName.Text := prFileName;
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.prSetFileFormat(f :TQUB_QFS_QMF_FileType);
begin
  prFileFormat := f;
  optFileFormat0.Checked := prFileFormat = QFS_QMF_TYPE_QMF;
  optFileFormat1.Checked := prFileFormat = QFS_QMF_TYPE_MDL;
end;

//Events
procedure TQUB_Viewport_Model_New_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(True{Read});
  {$include 'QUB_Viewport_Model_New_Properties.htx'}

  WikiTopic := 'New_Model';
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.FormDestroy(Sender: TObject);
begin
  inherited;

  prPropertiesRegistry(False{Write});

end;

procedure TQUB_Viewport_Model_New_PropertiesForm.FormShow(Sender: TObject);
begin
  inherited;

  BaseName     := BaseName;
  FileFormat   := FileFormat;
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.txtBaseNameKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  f :String;
begin
  inherited;
  f := TEdit(Sender).Text;
  if f <> BaseName then begin
    BaseName := f;
  end;
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.btnBrowseFileNameClick(Sender: TObject);
var
  Dlg :TSaveDialog;
begin
  inherited;
  Dlg := TSaveDialog.Create(Application);
  try
    Dlg.Filter := 'All QUB model files (*.qmf;*.mdl)|*.qmf;*.mdl|QUB model files (*.qmf)|*.qmf|QUB old model files (*.mdl)|*.mdl|All files (*.*)|*.*';
    Dlg.Options := [ofPathMustExist, ofEnableSizing];
    Dlg.FileName := FileName;
    if Dlg.Execute then begin
      FileName := Dlg.FileName;
    end;
  finally
    Dlg.Free;
  end;
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.txtFileNameKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtFileNameExit(Sender);
  end;
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.txtFileNameExit(Sender: TObject);
var
  f :String;
begin
  inherited;
  f := TEdit(Sender).Text;
  if f <> FileName then begin
    FileName := f;
  end;
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.optFileFormatClick(Sender: TObject);
begin
  inherited;
  FileFormat := TQUB_QFS_QMF_FileType(TRadioButton(Sender).Tag);
end;

procedure TQUB_Viewport_Model_New_PropertiesForm.btnOKClick(Sender: TObject);
var
  Dlg :TSaveDialog;
  r   :Integer;
  f   :String;
  xt  :String;
begin
  xt  := LowerCase(ExtractFileExt(FileName));
  if (xt = '.qmf') or (xt = '.mdl') then begin
    f := ChangeFileExt(FileName, FileTypeToExt(FileFormat));
  end else begin
    f := FileName + FileTypeToExt(FileFormat);
  end;
  
  Dlg := nil;

  try

    while FileExists(f) do begin

      Dlg := TSaveDialog.Create(Application);
      Dlg.Filter := 'All QUB model files (*.qmf;*.mdl)|*.qmf;*.mdl|QUB model files (*.qmf)|*.qmf|QUB old model files (*.mdl)|*.mdl|All files (*.*)|*.*';
      Dlg.Options := [ofPathMustExist, ofEnableSizing];
      Dlg.FileName := ChangeFileExt(f, '');
      r := MessageDlg('File ''' + f + ''' already exists. Do you want to overwrite it ?', mtConfirmation, [mbYes, mbNo, mbCancel], 0);
      case r of
        idCancel:
          begin
            ModalResult := mrNone;
            Exit;
          end;
        idYes:
          begin
            FileName   := f;
            prFileName := f;
            inherited;
            Exit;
          end;
        idNo:
          begin
            if Dlg.Execute then begin
              f := ChangeFileExt(Dlg.FileName, FileTypeToExt(FileFormat));
            end else begin
              ModalResult := mrNone;
              Exit;
            end;
          end;
      end;//case
    end;//while

    FileName   := f;
    prFileName := f;
    inherited;

  finally
    Dlg.Free;
  end;

end;

procedure TQUB_Viewport_Model_New_PropertiesForm.prPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
begin
  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\New Model File', True) then begin
        if Read then begin
          prBaseName   := ReadRegistryString(Reg, 'Base Name', DEF_BASENAME);
          prFileFormat := TQUB_QFS_QMF_FileType(ReadRegistryInteger(Reg, 'File Format', Integer(QFS_QMF_TYPE_QMF)));
        end else begin
          WriteString('Base Name', BaseName);
          WriteInteger('File Format', Integer(FileFormat));
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;


//******************************************************************************
end.
