{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
//
// Acquisition Protocol - Building blocks (pulse, sine, ramp, curve, ... )
// to create data files for Analog / Digital output.
//


unit QUB_Viewport_Protocol;


interface


uses


Windows, Messages, SysUtils, Classes, Math, Graphics, Controls, Forms, Dialogs, ComCtrls, ExtCtrls, StdCtrls, Registry,
Buttons, TeeProcs, TeEngine, Chart, Series, Menus, ImgList, ToolWin,

AllocTypes, Alloc,
RegistryInterface,

QUB_ListBox, 
QUB_Global,
QUB_Viewport_Files,
QUB_Viewport_Data_New_Properties,
QUB_QFS_QPF,
QUB_QVO, QUB_QVO_QVP,
PythonEngine,
QUB_Python_Thread;


type


TQUB_ProtocolShapeButton = class(TBitBtn)
protected

  prShape            :TQUB_QFS_QPF_ShapeFactory;
  prClickY           :Integer;

  procedure prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  procedure prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

public

  OnClickNoDrag      :TNotifyEvent;

  constructor Create(AOwner :TComponent; AShape :TQUB_QFS_QPF_ShapeFactory); reintroduce; virtual;

  property Shape     :TQUB_QFS_QPF_ShapeFactory           read prShape;
end;

  //-----
TQUB_ProtocolShapePanel = class(TPanel)
protected
  prShapeLib         :TQUB_QFS_QPF_ShapeLib;

  procedure prSetShapeLib(x :TQUB_QFS_QPF_ShapeLib);
  procedure prOnShapeAdded(Sender :TObject);
  procedure prAddShape(i :Integer);

public
  OnShapeAdded       :procedure(Sender :TObject; ShapeButton :TQUB_ProtocolShapeButton) of object;

  property ShapeLib  :TQUB_QFS_QPF_ShapeLib   read prShapeLib     write prSetShapeLib;
end;

//-----
TQUB_Python_ProtocolView = class;

//-----
TQUB_Viewport_Protocol = class(TQUB_Viewport_Files)
  splFilesLists                                 :TSplitter;
  ConteinerLists                                :TPanel;
    ConteinerStages: TPanel;
    lblStages: TLabel;
    panToolbar: TPanel;
    panToolbarFile: TPanel;
    btnNewFile: TSpeedButton;
    btnOpenFile: TSpeedButton;
    btnSaveFile: TSpeedButton;
    panToolbarProperties: TPanel;
    lblChannels: TLabel;
    txtChannels: TEdit;
    udChannels: TUpDown;
    btnProperties: TButton;
    panToolbarPlay: TPanel;
    Panel1: TPanel;
    btnCompile: TSpeedButton;
    panToolbarShapes: TPanel;
    Panel3: TPanel;
    Panel2: TPanel;
    Saveas1: TMenuItem;
    procedure btnNewFileClick(Sender: TObject);
    procedure btnOpenFileClick(Sender: TObject);
    procedure btnSaveFileClick(Sender: TObject);
    procedure btnPropertiesClick(Sender: TObject);
    procedure udChannelsChangingEx(Sender: TObject; var AllowChange: Boolean; NewValue: Smallint; Direction: TUpDownDirection);
    procedure btnCompileClick(Sender: TObject);
    procedure Saveas1Click(Sender: TObject);

protected

  prPyHandle             :TQUB_Python_ProtocolView;

  prShapeLib             :TQUB_QFS_QPF_ShapeLib;
  panShapes              :TQUB_ProtocolShapePanel;

  prChangingChannelCount :Boolean;

  procedure prInitPyHandle;
  procedure prFreePyHandle;

  function prCurrProto   :TQUB_QVO_QVP;
  function prGetFileName :String;

  procedure prOnChangeFile(Sender :TObject);
  procedure prOnOpenFile(Sender :TObject);
  procedure prOnCloseFile(Sender :TObject);
  procedure prOnChannelCountChanged(Sender :TObject);
  procedure prOnShapeAdded(Sender :TObject; ShapeButton :TQUB_ProtocolShapeButton);
  procedure prOnShapeEndDrag(Sender, Target: TObject; X, Y: Integer);
  procedure prOnShapeClick(Sender :TObject);
  procedure prOnCompileDone(Sender :TObject);

  procedure DoPropertiesInternal(Read :Boolean); override;


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;

  procedure Activate; override;
  procedure DeActivate; override;

  //File
  function New(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean; override;
  function Save(const FileName :String; Silent :Boolean) :Boolean; override;
  function SaveAs(const FileName :String; Silent :Boolean) :Boolean; override;
  function Close :Boolean; override;
  function CloseAll :Boolean; override;

  function OpenFromString(const MemFile :String) :Boolean; override;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

  function CanClose :Boolean; override;
  function CanCloseActive :Boolean; override;
  function CanSave :Boolean;

  function CloseProto(Proto :TQUB_QVO_QVP) :Boolean;
  function SaveProto(Proto :TQUB_QVO_QVP; const FileName :String; Silent :Boolean) :Boolean;
  function SaveAsProto(Proto :TQUB_QVO_QVP; const FileName :String; Silent :Boolean) :Boolean;

  function Properties :Boolean; override;

  property PyHandle                          :TQUB_Python_ProtocolView                     read prPyHandle;
  property CurrProto                         :TQUB_QVO_QVP                                 read prCurrProto;
  property FileName                          :String                                       read prGetFileName;

end;

  //-----
TQUB_Python_ProtocolView = class(TPyObject)
public
  ViewportProtocol      :TQUB_Viewport_Protocol;

  function  GetAttr(key : PChar) : PPyObject; override;

  // Sequence services
  function  SqLength : Integer; override;
  function  SqItem( idx : Integer ) : PPyObject; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;
end;


implementation


uses


QUB_GUIInterface,
QUB_Viewport_AcquisitionOutput, QUB_AcquisitionOutput,
QUB_QVO_QVD,
QUB_Viewport_Python,
QUB_Node,
QUB_Node_Python,
QUB_Task_Protocol_To_Data,
QUB_Viewport_Data,
QUB_QFS_QDF_Global, QUB_Main;


{$R *.DFM}


var


ProtocolViewPyType                    :TPythonType;


//******************************************************************************
constructor TQUB_Viewport_Protocol.Create(AOwner :TComponent);
begin

  inherited;

  PythonThread.SneakPyProc(prInitPyHandle);

  ListObjects.Width      := 128;
  ListObjects.Height     := 80;
  ListObjects.Align      := alTop;
  ListObjects.Style      := lbOwnerDrawFixed;

  prShapeLib             := TQUB_QFS_QPF_ShapeLib.Create;
  panShapes              := TQUB_ProtocolShapePanel.Create(Self);
  panShapes.Align        := alClient;
  panShapes.Caption      := '';
  panShapes.BevelInner   := bvNone;
  panShapes.BevelOuter   := bvNone;
  panShapes.Parent       := panToolbarShapes;
  panShapes.OnShapeAdded := prOnShapeAdded;
  panShapes.ShapeLib     := prShapeLib;

  prChangingChannelCount := False;

  Events.AddEvent('OnChangeFile', prOnChangeFile);
  Events.AddEvent('OnOpenFile',   prOnOpenFile);
  Events.AddEvent('OnCloseFile',  prOnCloseFile);

end;

destructor TQUB_Viewport_Protocol.Destroy;
begin
  try
    PythonThread.SneakPyProc( prFreePyHandle );
    prShapeLib.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Viewport_Protocol.prInitPyHandle;
var
  PyInst                  :PPyObject;
begin
  ProtocolViewPyType                    := TPythonType.Create(nil);
  ProtocolViewPyType.Engine             := GetPythonEngine;
  ProtocolViewPyType.Module             := PythonViewport.pmQUBTypes;
  ProtocolViewPyType.TypeName           := 'ProtocolViewType';
  ProtocolViewPyType.Services.Basic     := [bsGetAttr];
  ProtocolViewPyType.Services.Sequence  := [ssItem, ssLength];
  ProtocolViewPyType.PyObjectClass      := TQUB_Python_ProtocolView;
  ProtocolViewPyType.Initialize;

  PyInst := ProtocolViewPyType.CreateInstance;
  prPyHandle := TQUB_Python_ProtocolView( PythonToDelphi(PyInst) );
  prPyHandle.ViewportProtocol := Self;
  PyNode_PublishPyObj( PyInst, 'ProtocolView' );
end;

procedure TQUB_Viewport_Protocol.prFreePyHandle;
begin
  GetPythonEngine.Py_DECREF( prPyHandle.GetSelf );
end;


//Activate/DeActivate **********************************************************
procedure TQUB_Viewport_Protocol.Activate;
begin
  inherited;
  if CurrProto <> nil then if not CurrProto.Active then CurrProto.Activate;
end;

procedure TQUB_Viewport_Protocol.DeActivate;
begin
  inherited;
  if CurrProto <> nil then if CurrProto.Active then CurrProto.DeActivate;
end;

//****************************************************************************//
//                                                                            //
//                         F I L E     M E T H O D S                          //
//                                                                            //
//****************************************************************************//
function TQUB_Viewport_Protocol.New(const FileName :String; Silent :Boolean) :Boolean;
var
  Proto    :TQUB_QVO_QVP;
  f        :String;
begin

  //Process FileName;
  f                                    := FileName;
  if f = '' then f                     := GetNewUniqueFileName(ExpandFileName('noname'), '.qpf', prNoNameCount);

  //Create new Data
  Proto                                := TQUB_QVO_QVP.Create(Self, prShapeLib);
  Proto.Parent                         := ConteinerLeft;
  Proto.StagesPanel                    := ConteinerStages;
  Proto.Initialize;

  Proto.Protocol.FileName              := f;
  Proto.TipsOn                         := prTipsOn;

  Result                               := False;

  if not Result then begin
    ListObjects.AddQVO(Proto, ExtractFileName(Proto.FileName));
    DoOnOpenFile(Proto);
    DoOnChangeFile(Proto);
    prAddFileEvents(Proto);
  end else begin
    prDelFileEvents(Proto);
    Proto.Free;
  end;

end;

function TQUB_Viewport_Protocol.Open(const FileName :String; Silent :Boolean) :Boolean;
var
  Proto    :TQUB_QVO_QVP;
  Dlg      :TOpenDialog;
  Count    :Integer;
  f, fi    :String;
  d        :Integer;
begin

  Result := False;

  Dlg := nil;

  //Process FileName;
  try

    f := FileName;
    if f = '' then begin
      Dlg := TOpenDialog.Create(Application);
      Dlg.Filter := 'QUB protocol files (*.qpf)|*.qpf|' +
                    'All files (*.*)|*.*';
      Dlg.Options := [ofPathMustExist, ofFileMustExist, ofAllowMultiSelect, ofEnableSizing];
      if not Dlg.Execute then begin
        Result := True;
        Exit;
      end;
    end;

    if f <> '' then begin
      Count := 1;
    end else begin
      Count := Dlg.Files.Count;
    end;

    for d := Count - 1 downto 0 do begin

      if f = '' then fi := Dlg.Files[d] else fi := f;

      //Create new Proto and open the file

      Proto                                := TQUB_QVO_QVP.Create(Self, prShapeLib);
      Proto.Parent                         := ConteinerLeft;
      Proto.StagesPanel                    := ConteinerStages;
      Proto.Initialize;

      Proto.TipsOn                         := prTipsOn;

      Result                               := Proto.Open(fi);

      if not Result then begin
        ListObjects.AddQVO(Proto, ExtractFileName(fi));
        DoOnOpenFile(Proto);
        DoOnChangeFile(Proto);
        prAddFileEvents(Proto);
      end else begin
        prDelFileEvents(Proto);
        Proto.Free;
        if not Silent then MessageHelpDlg('Could not open file ' + fi, mtError, [mbOK], 0, 'Error:Open_File');
      end;

    end;


  finally

    Dlg.Free;

  end;

end;

function TQUB_Viewport_Protocol.Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean;
var
  Proto     :TQUB_QVO_QVP;
begin

  try

    //Create new Data and open the file
    Proto                  := TQUB_QVO_QVP.Create(Self, prShapeLib);
    Proto.Parent           := ConteinerLeft;
    Proto.StagesPanel      := ConteinerStages;
    Proto.Initialize;

    Proto.TipsOn           := prTipsOn;

    Result                               := Proto.Open(FileName, OpenStream);

    if not Result then begin
      ListObjects.AddQVO(Proto, ExtractFileName(FileName));
      DoOnOpenFile(Proto);
      DoOnChangeFile(Proto);
      prAddFileEvents(Proto);
    end else begin
      prDelFileEvents(Proto);
      Proto.Free;
      if not Silent then MessageHelpDlg('Could not open file ' + FileName, mtError, [mbOK], 0, 'Error:Open_File');
    end;

  finally

  end;

end;



function TQUB_Viewport_Protocol.Save(const FileName :String; Silent :Boolean) :Boolean;
var
  Proto  :TQUB_QVO_QVP;
begin

  Proto   := CurrProto;
  Result  := Proto = nil;
  if Result then Exit;

  Result := SaveProto(Proto, FileName, Silent);

end;

function TQUB_Viewport_Protocol.SaveAs(const FileName :String; Silent :Boolean) :Boolean;
var
  Proto     :TQUB_QVO_QVP;
begin

  Proto    := CurrProto;
  Result   := Proto = nil;
  if Result then Exit;

  Result := SaveAsProto(Proto, FileName, Silent);

end;

function TQUB_Viewport_Protocol.Close :Boolean;
var
  Proto     :TQUB_QVO_QVP;
begin
  Result := False;

  Proto   := CurrProto;
  if Proto = nil then Exit;

  Result := CloseProto(Proto);

end;


function TQUB_Viewport_Protocol.SaveProto(Proto :TQUB_QVO_QVP; const FileName :String; Silent :Boolean) :Boolean;
var
  FName :String;
begin

  Result := (Proto.LockedAny);
  if Result then Exit;

  //Process FileName
  FName := FileName;
  if FName = '' then FName := Proto.FileName;
  if (Proto.Saved) and (Proto.CanSaveAs(FName, True)) then begin
    Result := Proto.Save(FName);
    ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Proto)] := ExtractFileName(Proto.FileName);
    DoOnChangeFile(Proto);
  end else begin
    Result := SaveAsProto(Proto, FName, Silent);
  end;
end;

function TQUB_Viewport_Protocol.SaveAsProto(Proto :TQUB_QVO_QVP; const FileName :String; Silent :Boolean) :Boolean;
var
  Dlg      :TSaveDialog;
  f        :String;
  r        :Integer;
begin

  Result := (Proto.LockedAny);
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Proto.FileName;

  Dlg := TSaveDialog.Create(Application);
  Dlg.Filter := 'QUB protocol files (*.qpf)|*.qpf|' +
                'All files (*.*)|*.*';

  Dlg.Options := [ofPathMustExist{, ofOverwritePrompt}, ofEnableSizing];
  Dlg.Title := 'Save protocol file as';

  try

    repeat

      Dlg.FileName := ChangeFileExt(f, '');
      if Dlg.Execute then begin
        f := ChangeFileExt(Dlg.FileName, '.qpf');
      end else begin
        Result := True;
        Exit;
      end;

      if (FileExists(f)) and (f <> ExpandFileName(Proto.FileName)) then begin
        case MessageDlg('File ''' + f + ''' exists. Do you want to overwrite it ?', mtWarning, [mbYes, mbNo, mbCancel], 0) of
          mrCancel:
            begin
              Result := True;
              Exit;
            end;
          mrNo:
            Continue;
        end;//case
      end;

      if Proto.CanSaveAs(f, True){Ask confirmation for losing features} then begin
        Result := Proto.Save(f);
        Exit;
      end else begin
        r := MessageDlg('Protocol file cannot be saved in the selected format.'#13#10'Please choose one of the writable formats.', mtError, [mbOK, mbCancel], 0);
        Result := r = mrCancel;
        if Result then Exit;
      end;

    until (Result);

  finally

    Dlg.Free;

    ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Proto)] := ExtractFileName(Proto.FileName);
    DoOnChangeFile(Proto);

  end;

end;


function TQUB_Viewport_Protocol.CloseProto(Proto :TQUB_QVO_QVP) :Boolean;
begin
  Result := False;
  
  DoOnClosingFile(Proto);
  DoOnCloseFile(Proto);
  prDelFileEvents(Proto);
  ListObjects.DelQVO(Proto);

  if ListObjects.Items.Count = 0 then DoOnChangeFile(nil);

  Proto.Free;
end;


function TQUB_Viewport_Protocol.CloseAll :Boolean;
begin
  Result := False;
  while CurrProto <> nil do begin
    Result := Close;
    if Result then Break;
  end;
end;

function TQUB_Viewport_Protocol.CanCloseActive :Boolean;
var
  Proto :TQUB_QVO_QVP;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  Proto   := CurrProto;
  if Proto = nil then Exit;

  if Proto.LockedClose then begin
    MessageHelpDlg('File ''' + Proto.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end else if Proto.LockedAny then begin
    MessageHelpDlg('File ''' + Proto.FileName + ''' is executing a task and cannot be closed (check TaskView)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end;

  if not Result then Exit;

  if Proto.Modified then begin
    case MessageDlg('File ''' + Proto.FileName + ''' is modified.'#13#10'Do you want to save it ?', mtConfirmation, [mbYes, mbNo, mbCancel], 0) of
      mrCancel:
        Result := False;

      mrYes:
        if Proto.Saved then begin
          Result := not Proto.Save(Proto.FileName);
        end else begin
          Result := not SaveAs('', False{Silent});
        end;

      //mrNo:

    end;
  end;

end;

function TQUB_Viewport_Protocol.CanClose :Boolean;
var
  ada :TQUB_QVO_QVP;
  qvp :TQUB_QVO_QVP;
  All :Boolean;
  i   :Integer;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  All := False;
  ada := CurrProto;

  //John, check similar function in ViewportData. I think there is some bug in the logic that I fixed. LM
  
  try

    for i := ListObjects.Items.Count - 1 downto 0 do begin

      qvp := TQUB_QVO_QVP(ListObjects[i]);

      if qvp.LockedClose then begin
        MessageHelpDlg('File ''' + qvp.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end else if qvp.LockedAny then begin
        MessageHelpDlg('File ''' + qvp.FileName + ''' is executing a task and cannot be closed (check TaskView)', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end;

      if not Result then Exit;

      if qvp.Modified then begin

        ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(qvp);
        qvp.Activate;

        if All then begin

          if qvp.Saved then begin
            Result := not qvp.Save(qvp.FileName);
          end else begin
            Result := not SaveAs('', False{Silent});
          end;
          if not Result then Exit;

        end else begin

          case MessageDlg('File ''' + qvp.FileName + ''' is modified.'#13#10'Do you want to save it ?', mtConfirmation, [mbYes, mbNo, mbYesToAll, mbNoToAll, mbCancel], 0) of
            mrCancel:
            //Cancel
              begin
                Result := False;
                Exit;
              end;

            mrYes:
            //Yes
              begin
                if qvp.Saved then begin
                  Result := not qvp.Save(qvp.FileName);
                end else begin
                  Result := not SaveAs('', False{Silent});
                end;
                if not Result then Exit;
              end;

            mrYestoAll://10:
            //YestoAll
              begin
                All := True;
                if qvp.Saved then begin
                  Result := not qvp.Save(qvp.FileName);
                end else begin
                  Result := not SaveAs('', False{Silent});
                end;
                if not Result then Exit;
              end;

            //mrNotoAll://9
            //NotoAll

          end;//case 

        end;//All

      end;

    end;

  finally

    if ada <> nil then begin
      ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(ada);
      ada.Activate;
    end;

  end;

end;

function TQUB_Viewport_Protocol.CanSave :Boolean;
begin
  if CurrProto <> nil then Result := CurrProto.Modified else Result := False;
end;

function TQUB_Viewport_Protocol.OpenFromString(const MemFile :String) :Boolean;
var
  Proto     :TQUB_QVO_QVP;
begin
  Result := False;

  //Process FileName;
  if MemFile = '' then Exit;

  //Create new Data and open the file

  Proto                   := TQUB_QVO_QVP.Create(Self, prShapeLib);
  Proto.Parent            := ConteinerLeft;
  Proto.StagesPanel       := ConteinerStages;
  Proto.Initialize;
  Proto.TipsOn            := prTipsOn;
  Result                  := Proto.OpenFromString(MemFile);

  if not Result then begin
    ListObjects.AddQVO(Proto, 'noname');
    DoOnOpenFile(Proto);
    DoOnChangeFile(Proto);
    prAddFileEvents(Proto);
  end else begin
    prDelFileEvents(Proto);
    Proto.Free;
    MessageDlg('Could not open file noname', mtError, [mbOK], 0);
  end;

end;

function TQUB_Viewport_Protocol.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  Result := False;
  if CurrProto <> nil then Result := CurrProto.SaveToString(MemFile, FileType);
end;


//Properties
function TQUB_Viewport_Protocol.Properties :Boolean;
begin
  Result := False;
  if CurrProto <> nil then CurrProto.Properties else Result := True;
end;



//****************************************************************************//
//                                                                            //
//                                  O T H E R                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Viewport_Protocol.prCurrProto :TQUB_QVO_QVP;
begin
  Result := nil;
  if ListObjects = nil then Exit;
  if ListObjects.Items.Count = 0  then Exit;
  if ListObjects.ItemIndex   = -1 then Exit;
  Result := TQUB_QVO_QVP(ListObjects.Items.Objects[ListObjects.ItemIndex]);
end;


function TQUB_Viewport_Protocol.prGetFileName :String;
begin
  if CurrProto <> nil then Result := CurrProto.FileName else Result := '';
end;

procedure TQUB_Viewport_Protocol.prOnChangeFile(Sender :TObject);
begin
  if Sender <> nil then begin
    btnSaveFile.Enabled     := True;
    lblChannels.Enabled     := True;
    txtChannels.Enabled     := True;
    udChannels.Enabled      := True;
    btnProperties.Enabled   := True;
    udChannels.Position     := TQUB_QVO_QVP(Sender).Protocol.ChannelCount;
    btnCompile.Enabled      := True;
  end else begin
    btnSaveFile.Enabled     := False;
    lblChannels.Enabled     := False;
    txtChannels.Enabled     := False;
    udChannels.Enabled      := False;
    btnProperties.Enabled   := False;
    btnCompile.Enabled      := False;
  end;
end;

procedure TQUB_Viewport_Protocol.prOnOpenFile(Sender :TObject);
begin
  TQUB_QVO_QVP(Sender).OnChannelCountChanged := prOnChannelCountChanged;
end;

procedure TQUB_Viewport_Protocol.prOnCloseFile(Sender :TObject);
begin

end;

procedure TQUB_Viewport_Protocol.prOnChannelCountChanged(Sender :TObject);
begin
  if (Sender = CurrProto) and not prChangingChannelCount then begin
    udChannels.Position := CurrProto.Protocol.ChannelCount;
  end;
end;

procedure TQUB_Viewport_Protocol.prOnShapeAdded(Sender :TObject; ShapeButton :TQUB_ProtocolShapeButton);
begin
  ShapeButton.OnEndDrag       := prOnShapeEndDrag;
  ShapeButton.OnClickNoDrag   := prOnShapeClick;
end;

procedure TQUB_Viewport_Protocol.prOnShapeEndDrag(Sender, Target: TObject; X, Y: Integer);
begin
  if CurrProto <> nil then CurrProto.DoneDragging; // clean up any residual highlights
end;

procedure TQUB_Viewport_Protocol.prOnShapeClick(Sender :TObject);
begin
  if CurrProto <> nil then CurrProto.AddShape( TQUB_ProtocolShapeButton(Sender).Shape );
end;

procedure TQUB_Viewport_Protocol.btnNewFileClick(Sender: TObject);
begin
  inherited;
  New('', False);
end;

procedure TQUB_Viewport_Protocol.Saveas1Click(Sender: TObject);
begin
  inherited;
  SaveAs('', False);
end;

procedure TQUB_Viewport_Protocol.btnCompileClick(Sender: TObject);
var
  Task                         :TQUB_Task_Protocol_To_Data;
  Data                         :TQUB_QVO_QVD;
begin
  if CurrProto = nil then Exit;

  Task                         := TQUB_Task_Protocol_To_Data.Create(True);
  Task.Protocol                := CurrProto;
  Data                         := ViewportAcqOutput.AcquisitionOutput.Data;
  if Data = nil then begin
    with ViewportData.NewPropertiesForm do begin
      FileFormat               := QFS_QDF_TYPE_QDF;
      ADChannelCount           := CurrProto.Protocol.ChannelCount;
      Sampling                 := ViewportAcqOutput.AcquisitionOutput.Sampling * 1.0e-3;
    end;
    if not ViewportData.New(ChangeFileExt(CurrProto.FileName, '.qdf'), False) then begin
          Data                     := TQUB_QVO_QVD(ViewportData.ListObjects[ViewportData.ListObjects.Items.Count - 1]);
      ViewportAcqOutput.AcquisitionOutput.AssignOutputAcq( Data );
    end;
  end;
  if Data <> nil then begin
    Data.Activate;
    if Data.Clear then Task.Terminate;
  end;
  Task.Data                    := Data;
  Task.OnTerminate             := prOnCompileDone;
  Task.OnChange                := TaskView.OnTaskChange;

  Task.Resume;
  panToolbar.Enabled           := False;
end;

procedure TQUB_Viewport_Protocol.prOnCompileDone(Sender :TObject);
begin
  panToolbar.Enabled           := True;
end;

procedure TQUB_Viewport_Protocol.btnOpenFileClick(Sender: TObject);
begin
  inherited;
  Open('', False);
end;

procedure TQUB_Viewport_Protocol.btnSaveFileClick(Sender: TObject);
begin
  inherited;
  Save('', False);
end;

procedure TQUB_Viewport_Protocol.btnPropertiesClick(Sender: TObject);
begin
  inherited;
  CurrProto.Properties;
end;

procedure TQUB_Viewport_Protocol.udChannelsChangingEx(Sender: TObject;
  var AllowChange: Boolean; NewValue: Smallint;
  Direction: TUpDownDirection);
begin
  inherited;
  if (NewValue >= udChannels.Min) and (NewValue <= udChannels.Max) then begin
    prChangingChannelCount := True;
    CurrProto.Protocol.ChannelCount := NewValue;
    prChangingChannelCount := False;
  end;
end;



//Registry
procedure TQUB_Viewport_Protocol.DoPropertiesInternal(Read :Boolean);
var
  Reg :TRegistry;
begin

  inherited;

  Reg := TRegistry.Create;

  try

    with Reg do begin

      if OpenKey('Software\QUB\Interface ' + InterfaceName + '\Viewport ' + Name, True) then begin
        if Read then begin
          ListObjects.Height := Round(ReadRegistryFloat(Reg, 'ListObjects Height', 25.0) * ConteinerRight.ClientHeight / 100);
        end else begin
          WriteFloat('ListObjects Height', 100 * ListObjects.Height / ConteinerRight.ClientHeight);
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;

//******************************************************************************

constructor TQUB_ProtocolShapeButton.Create(AOwner :TComponent; AShape :TQUB_QFS_QPF_ShapeFactory);
begin
  inherited Create(AOwner);

  prShape     := AShape;
  Align       := alLeft;
  Hint        := prShape.Name + ': drag-and-drop into the protocol area';
  ShowHint    := True;
  OnMouseDown := prOnMouseDown;
  OnMouseUp   := prOnMouseUp;
  OnMouseMove := prOnMouseMove;
  prClickY    := -1;

  if prShape.Icon = nil then begin
    Caption  := prShape.Name;
    Width    := 0;
    // make wider
  end else begin
    Glyph    := prShape.Icon;
    Width    := prShape.Icon.Width + 4;
  end;
end;

// here ensues some bullshit:
//   apparently BeginDrag triggers an immediate OnClick,
//   but we want to present _either_ a drag or a click
//   (dragging begins when drag.y exceeds buttonheight)

procedure TQUB_ProtocolShapeButton.prOnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  //if Button = mbLeft then BeginDrag(False, 3);
  if Button = mbLeft then begin
    prClickY := Y;
  end;
end;

procedure TQUB_ProtocolShapeButton.prOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if (prClickY >= 0) and (Y >= Height) then begin
    prClickY := -2;
    BeginDrag(True, -1);
  end;
end;

procedure TQUB_ProtocolShapeButton.prOnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then begin
    if prClickY <> -2 then begin
      if (X >= 0) and (X < Width) and (Y >= 0) and (Y < Height) then begin
        if Assigned(OnClickNoDrag) then OnClickNoDrag(Self);
      end;
    end;
  end;
  prClickY := -1;
end;


procedure TQUB_ProtocolShapePanel.prSetShapeLib(x :TQUB_QFS_QPF_ShapeLib);
var
  i                :Integer;
begin
  prShapeLib       := x;
  for i := 1 to ShapeLib.ShapeCount - 1 do begin  // skip the <missing> template
    prAddShape(i);
  end;
end;

procedure TQUB_ProtocolShapePanel.prOnShapeAdded(Sender :TObject);
begin
  prAddShape( ShapeLib.ShapeCount - 1 );
end;

procedure TQUB_ProtocolShapePanel.prAddShape(i :Integer);
var
  Btn              :TQUB_ProtocolShapeButton;
begin
  Btn             := TQUB_ProtocolShapeButton.Create(Self, prShapeLib[i]);
  Btn.Parent      := Self;
  Btn.Height      := ClientHeight;
  if Btn.Width = 0 then Btn.Width := Canvas.TextWidth(Btn.Caption) + 10;

  if Assigned(OnShapeAdded) then OnShapeAdded(Self, Btn);
end;



//******************************************************************************

function  TQUB_Python_ProtocolView.GetAttr(key : PChar) : PPyObject;
begin
  Result := inherited GetAttr(key);
end;

function  TQUB_Python_ProtocolView.SqLength : Integer;
begin
  Result := 0; // ViewportProtocol.ListObjects.Items.Count;
end;

function  TQUB_Python_ProtocolView.SqItem( idx : Integer ) : PPyObject;
begin
{
  with GetPythonEngine do begin
    if (idx >= 0) and (idx < ViewportProtocol.ListObjects.Items.Count) then begin
      Result := TQUB_QVO_QVR(ViewportProtocol.ListObjects[idx]).PyHandle.GetSelf;
      Py_INCREF( Result );
    end else begin
      PyErr_SetString( PyExc_IndexError^, PChar('index out of range') );
      Result := nil;
    end;
  end;
}
  GetPythonEngine.PyErr_SetString( GetPythonEngine.PyExc_IndexError^, PChar('index out of range') );
  Result := nil;
end;

class procedure TQUB_Python_ProtocolView.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  {with PythonType do begin
    AddMethod( 'addShape',      @TQUB_Python_ProtocolView.AddShape,   'ProtocolView.addShape(AcquisitionProtocol.Shape)'  );
  end;}
end;

{
function TQUB_Python_ProtocolView.AddShape( args :PPyObject ): PPyObject;
var
  Shape        :PPyObject;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'O:ProtocolView.addShape(AcquisitionProtocol.Shape)', [@Shape] ) <> 0 then begin
      PythonThread.EndPython;
      ViewportProtocol.AddShape(Shape);
      PythonThread.BeginPython;
      Result := ReturnNone;
    end else begin
      Result := nil;
    end;
  end;
end;
}


//******************************************************************************
end.
