{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// holds the PythonReadEvalPrint, and handles a lot of the initialization sequence


unit QUB_Viewport_Python;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
StdCtrls, ExtCtrls, ComCtrls, Menus, Registry, SyncObjs,

QUB_Node_Python, QUB_Python_Prompt, QUB_Python_ModalDialog,
QUB_Viewport,
QUB_Console, QUB_Rich_Console,
PythonEngine, PythonReadEvalPrint, QUB_Python_Thread, PythonAtom,
AtomPythonEngine;


//{$include 'QUB_Viewport_Python.hct'}


const


PYTHON_GUI_WAIT       = 200;



type


TQUB_Viewport_Python = class(TQUB_Viewport)


  PythonEngine          :TAtomPythonEngine;
  PythonIO              :TPythonInputOutput;
  ptReport              :TPythonType;
  pmQUB                 :TPythonModule;
  butBreak              :TButton;
  pmQUBTypes            :TPythonModule;


  procedure PythonIOSendData(Sender: TObject; const Data: String);
  procedure FormCreate(Sender: TObject);
  procedure ptReportInitialization(Sender: TObject);
  procedure butBreakClick(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
    procedure PythonEngineBeforeLoad(Sender: TObject);

protected


  PREP                  :TPythonReadEvalPrint;

  prPyThread            :TQUB_Python_Thread;

  procedure prFinishInitialization;
  procedure prCheckPythonAvailable;


public


  constructor Create(AOwner :TComponent); override;

  procedure FinishInitialization;
  function PythonAvailable: Boolean;

  procedure DoInterrupt;
  procedure FinalizePython;


end;
(*
TPyPrompt = class(TPyObject)
protected

  prMsg             :String;
  prButtons         :PPyObject;
  prFields          :PPyObject;
  prResults         :PPyObject;

  procedure prDoPrompt;

public

  function  Repr : PPyObject; override;
  function  Call( args, kw : PPyObject) : PPyObject; override;

end;
*)
TPyReport = class(TPyObject)
protected

  prMessage   :String;
  prColor     :TQUB_Rich_Console_ColorCode;

  procedure prDoWrite;

public

  // Constructors & Destructors
  constructor Create( APythonType : TPythonType ); override;
  constructor CreateWith( PythonType : TPythonType; args : PPyObject ); override;

  // Type services

  // Basic services
  function  GetAttr(key : PChar) : PPyObject; override;
  function  SetAttr(key : PChar; value : PPyObject) : Integer; override;
  function  Repr : PPyObject; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;

  // Interface methods
  function DoWrite( args : PPyObject ) : PPyObject; cdecl;
  //function DoRaiseError( args : PPyObject ) : PPyObject; cdecl;
end;

var

PythonThread               :TQUB_Python_Thread;
PythonViewport             :TQUB_Viewport_Python;

implementation


{$R *.DFM}


uses

QUB_Main,
QUB_GUIInterface;


//****************************************************************************//
constructor TQUB_Viewport_Python.Create(AOwner :TComponent);
begin
  inherited;

end;

//******************************************************************************
procedure TQUB_Viewport_Python.PythonIOSendData(Sender: TObject; const Data: String);
begin
  inherited;
  PREP.PythonPrint(Sender,Data);
end;

procedure TQUB_Viewport_Python.FormCreate(Sender: TObject);
begin
  inherited;

  PythonViewport := Self;

  PREP                 := TPythonReadEvalPrint.Create(Self);
  PREP.Parent          := Conteiner;
  PREP.Left            := 0;
  PREP.Top             := 0;
  PREP.Width           := 808;
  PREP.Height          := 253;
  PREP.Align           := alClient;
  PREP.Font.Charset    := DEFAULT_CHARSET;
  PREP.Font.Color      := clWindowText;
  PREP.Font.Height     := -13;
  PREP.Font.Name       := 'Courier New';
  PREP.Font.Style      := [];
  PREP.Lines.Add('>>> ');
  PREP.ParentFont      := False;
  PREP.ScrollBars      := ssBoth;
  PREP.TabOrder        := 0;
  PREP.WantTabs        := True;
  PREP.WordWrap        := False;
  PREP.IO              := PythonIO;

  prPyThread := TQUB_Python_Thread.Create(False{wait to start});
  prPyThread.FreeOnTerminate := True;
  PythonThread := prPyThread;

  PREP.OnSendData   := prPyThread.InterpretFromEvent;
  prPyThread.Engine := PythonEngine;

  // pump output through pythread to synchronize with gui thread
  PythonIO.OnSendData := prPyThread.PythonPrint;
  prPyThread.OnSendData := PREP.PythonPrint;

  myPyThreadState       := GetPythonEngine.PyThreadState_Get;
  guiPyThreadState      := myPyThreadState;
  GetPythonEngine.PyThreadState_Swap(nil);
  GetPythonEngine.PyEval_ReleaseLock; // the python thread will take charge of it

  prPyThread.Resume;

  PREP.Text := 'You can use Python to add features to QUB.  See http://www.qub.buffalo.edu/qubdoc/ for details.'#13#10 + PREP.Text;

//   {$include 'QUB_Viewport_Python.htx'}
end;

procedure TQUB_Viewport_Python.PythonEngineBeforeLoad(Sender: TObject);
var
  PythonHome :String;
begin
  inherited;
  //PythonHome := ExtractFilePath(Application.ExeName)+'PythonHome';
  //PythonHome := PythonHome +';'
  //                         +'\Lib;'
  //                         +'\Lib\site-packages;'
  //                         +'\Lib\site-packages\win32;'
  //                         +'\Lib\site-packages\win32\lib';
  //SetEnvironmentVariable(PChar('PYTHONPATH'), PChar(PythonHome));
end;

procedure TQUB_Viewport_Python.FinishInitialization;
begin
  PythonThread.SyncPyProc(prFinishInitialization, False, False, -1);
end;

procedure TQUB_Viewport_Python.prFinishInitialization;
var
  libpath                       :String;
begin
  libpath := ExtractFileDir(Application.ExeName) + '\PythonScripts';
  GetPythonEngine.Run_CommandAsString('QUBB.LoadCommands(''' + libpath + ''')', single_input);  // read in qubfolder/PythonScripts/*.py
end;

function TQUB_Viewport_Python.PythonAvailable: Boolean;
begin
  Result := not PythonThread.SyncPyProc(prCheckPythonAvailable, False, True, 100);
end;

procedure TQUB_Viewport_Python.prCheckPythonAvailable;
begin
end;

procedure TQUB_Viewport_Python.FormDestroy(Sender: TObject);
begin
  inherited;
  FinalizePython;
end;

procedure TQUB_Viewport_Python.FinalizePython;
begin
  if PythonEngine = nil then Exit;
  
  while prPyThread.RequestTerminate(1000) do Application.ProcessMessages;

  GetPythonEngine.PyEval_AcquireLock;
  GetPythonEngine.PyThreadState_Swap( myPyThreadState );

  PythonEngine.Free;
  PythonEngine := nil;
end;

procedure TQUB_Viewport_Python.DoInterrupt;
begin
  prPyThread.DoInterrupt;
end;

procedure TQUB_Viewport_Python.ptReportInitialization(Sender: TObject);
begin
  inherited;
  ptReport.PyObjectClass := TPyReport;
end;

procedure TQUB_Viewport_Python.butBreakClick(Sender: TObject);
begin
  inherited;
  DoInterrupt;
end;

//******************************************************************************

constructor TPyReport.Create( APythonType : TPythonType );
begin
  inherited;
end;

// Don't call the Create constructor of TPyPoint, because
// we call the inherited constructor CreateWith that calls
// the Create constructor first, and because the constructors
// are virtual, TPyPoint.Create will be automatically be called.

constructor TPyReport.CreateWith( PythonType : TPythonType; args : PPyObject );
begin
  inherited;
  exit;
  {
  with GetPythonEngine do
    begin
      if PyArg_ParseTuple( args, 'ii:CreatePoint', [@x, @y] ) = 0 then
        exit;
    end;
  }
end;

// Then we override the needed services

function  TPyReport.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do
    begin
      if key = 'TextColor' then
        Result := PyInt_FromLong( Integer(COLOR_TEXT) )
      else if key = 'GrayColor' then
        Result := PyInt_FromLong( Integer(COLOR_GRAY) )
      else if key = 'ErrorColor' then
        Result := PyInt_FromLong( Integer(COLOR_ERROR) )
      else if key = 'WarningColor' then
        Result := PyInt_FromLong( Integer(COLOR_WARNING) )
      {else if key = 'BlackAndWhite' then
        Result := PyBooleanWhatever( whatever )}
      else
        Result := inherited GetAttr(key);
    end;
end;

function  TPyReport.SetAttr(key : PChar; value : PPyObject) : Integer;
begin
  //Result := 0;
  with GetPythonEngine do
    begin
      {if key = 'BlackAndWhite' then
        begin
          whatever
          if PyArg_Parse( value, 'i:Point.SetAttr', [@x] ) = 0 then
            Result := -1;
        end
      else    }
        Result := inherited SetAttr(key, value);
    end;
end;

function  TPyReport.Repr : PPyObject;
begin
  with GetPythonEngine do
    Result := PyString_FromString( PChar('<QUB Report Viewport>') );
end;

// Class methods
// We register the methods of our type

class procedure TPyReport.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do
    begin
      AddMethod( 'Write', @TPyReport.DoWrite, 'ReportViewport.Write( msg, color )' );
      //AddMethod( 'RaiseError', @TPyPoint.DoRaiseError, 'Point.RaiseError()' );
    end;
end;

// Interface methods
// They will be called directly by Python, so we extract the
// python arguments and we call the method that will really do
// the action.

function TPyReport.DoWrite( args : PPyObject ) : PPyObject;
var
  s      :PChar;
  color  :Integer;
begin
  with GetPythonEngine do begin
    // We adjust the transmitted self argument
    Adjust(@Self);

    // set defaults for optional arguments
    color := Integer(COLOR_TEXT);

    // extract the arguments
    if PyArg_ParseTuple( args, 's|i:ReportViewport.Write', [@s, @color] ) <> 0 then begin
      // if it's ok, then we call the method that does the job
      // with the correct arguments
      prMessage := s;
      prColor   := TQUB_Rich_Console_ColorCode(color);

      PythonThread.SyncGUIProc( prDoWrite );

      // Finally, we return nothing
      Result := ReturnNone;
    end else begin// the arguments were not right
      Result := nil;
    end;
  end;
end;

procedure TPyReport.prDoWrite;
begin
  Console.NewEntry(prMessage, prColor);
end;


//******************************************************************************
end.
