{
/* Copyright 1998-2014 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the text-output window


unit QUB_Viewport_Report;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, ComCtrls, Menus, Registry,

RegistryInterface,

QUB_Rich_Console,
QUB_Viewport,
QUB_Viewport_Report_Properties,

//KPL_Expression,

Buttons;


//{$include 'QUB_Viewport_Report.hct'}


const


MAX_EVALUATOR_HISTORY_COUNT = 50;


type


TQUB_Viewport_Report = class(TQUB_Viewport)


  panEvaluator       :TPanel;
  lblEvaluator       :TLabel;
  lstEvaluator       :TComboBox;

  PopupMenu          :TPopupMenu;
    PropertiesItem   :TMenuItem;
    CopyItem         :TMenuItem;
    N1               :TMenuItem;
    ClearItem        :TMenuItem;


  procedure FormCreate(Sender: TObject);
  procedure FormResize(Sender: TObject);

  procedure PropertiesItemClick(Sender: TObject);
  procedure CopyItemClick(Sender: TObject);
  procedure ClearItemClick(Sender: TObject);

  procedure txtEvaluatorKeyPress(Sender: TObject; var Key: Char);
  procedure txtEvaluatorEnter(Sender: TObject);


protected


  prPropertiesForm   :TQUB_Viewport_Report_PropertiesForm;


  procedure DoPropertiesInternal(Read :Boolean); override;
  procedure DoProperties(Read :Boolean); override;


  //Properties
  procedure prOnGetProperties(Sender :TObject);
  procedure prOnPutProperties(Sender :TObject);


public


  Console            :TQUB_Rich_Console;


  constructor Create(AOwner :TComponent); override;


end;


implementation

uses

QUB_Viewport_Python, QUB_Python_Thread, PythonEngine;


{$R *.DFM}


//****************************************************************************//
constructor TQUB_Viewport_Report.Create(AOwner :TComponent);
begin
  inherited;

  prPropertiesForm := TQUB_Viewport_Report_PropertiesForm.Create(Self);
  prPropertiesForm.OnGetProperties := prOnGetProperties;
  prPropertiesForm.OnPutProperties := prOnPutProperties;

end;

procedure TQUB_Viewport_Report.FormCreate(Sender: TObject);
begin
  inherited;

  Console                := TQUB_Rich_Console.Create(Self);
  Console.Parent         := Conteiner;
  Console.Initialize;
  Console.Left           := 0;
  Console.Top            := 24;
  Console.Width          := 374;
  Console.Height         := 228;
  Console.Align          := alClient;
  Console.Color          := clNavy;
  Console.Font.Charset   := ANSI_CHARSET;
  Console.Font.Color     := clCaptionText;
  Console.Font.Height    := -11;
  Console.Font.Name      := 'Courier New';
  Console.Font.Style     := [];
  Console.ParentFont     := False;
  Console.PopupMenu      := PopupMenu;
  Console.ScrollBars     := ssBoth;
  Console.TabOrder       := 0;
  Console.WordWrap       := False;

  Console.Lines.Append('/* Copyright 1998-2014 Research Foundation State University of New York */');
  Console.Lines.Append('');
  Console.Lines.Append('/* QuB is free software; you can redistribute it and/or modify          */');
  Console.Lines.Append('/* it under the terms of the GNU General Public License as published by */');
  Console.Lines.Append('/* the Free Software Foundation, either version 3 of the License, or    */');
  Console.Lines.Append('/* (at your option) any later version.                                  */');
  Console.Lines.Append('');
  Console.Lines.Append('/* QuB is distributed in the hope that it will be useful,               */');
  Console.Lines.Append('/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */');
  Console.Lines.Append('/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */');
  Console.Lines.Append('/* GNU General Public License for more details.                         */');
  Console.Lines.Append('');
  Console.Lines.Append('/* You should have received a copy of the GNU General Public License,   */');
  Console.Lines.Append('/* named LICENSE.txt, in the QuB program directory.  If not, see        */');
  Console.Lines.Append('/* <http://www.gnu.org/licenses/>.                                      */');

end;

procedure TQUB_Viewport_Report.prOnGetProperties(Sender :TObject);
var
  lRedraw :Boolean;
begin
  with Sender as TQUB_Viewport_Report_PropertiesForm do begin
    lRedraw := (Console.BlackAndWhite <> BlackAndWhite) or (Console.Colors[COLOR_BACK] <> ColorBack) ;
    if Console.Colors[COLOR_BACK]    <> ColorBack     then Console.Colors[COLOR_BACK]    := ColorBack;
    if Console.Colors[COLOR_TEXT]    <> ColorText     then Console.Colors[COLOR_TEXT]    := ColorText;
    if Console.Colors[COLOR_GRAY]    <> ColorGray     then Console.Colors[COLOR_GRAY]    := ColorGray;
    if Console.Colors[COLOR_ERROR]   <> ColorError    then Console.Colors[COLOR_ERROR]   := ColorError;
    if Console.Colors[COLOR_WARNING] <> ColorWarning  then Console.Colors[COLOR_WARNING] := ColorWarning;
    if Console.BlackAndWhite         <> BlackAndWhite then Console.BlackAndWhite         := BlackAndWhite;
    if Console.MaxLines              <> MaxLines      then Console.MaxLines              := MaxLines;

    if lRedraw then Console.Redraw;

    if BlackAndWhite then begin
      lstEvaluator.Color      := clWhite;
      lstEvaluator.Font.Color := clBlack;
    end else begin
      lstEvaluator.Color      := Console.Colors[COLOR_BACK];
      lstEvaluator.Font.Color := Console.Colors[COLOR_TEXT];
    end;

  end;//with

  {
  // retrieve control values
  TQUB_Viewport_Report_PropertiesForm(Sender).GetControls( @clBack, @clText,
            @clGray, @clError, @clWarning, @lBW, @iMaxLines );
  // save to console member vars
  with Sender as TQUB_Viewport_Report_PropertiesForm do begin
    lRedraw := (Console.BlackAndWhite<>lBW) or (Console.Colors[COLOR_BACK] <> clBack) ;
    Console.Colors[COLOR_BACK]    := clBack;
    Console.Colors[COLOR_TEXT]    := clText;
    Console.Colors[COLOR_GRAY]    := clGray;
    Console.Colors[COLOR_ERROR]   := clError;
    Console.Colors[COLOR_WARNING] := clWarning;
    Console.BlackAndWhite         := lBW;
    Console.MaxLines              := iMaxLines;

    if lRedraw Then Console.Redraw();

    txtEvaluator.Color            := Console.Color;
    txtEvaluator.Font             := Console.Font;

  end;
  }

end;

procedure TQUB_Viewport_Report.prOnPutProperties(Sender :TObject);
begin
  {
  TQUB_Viewport_Report_PropertiesForm(Sender).SetControls(
              Console.Colors[COLOR_BACK], Console.Colors[COLOR_TEXT],
              Console.Colors[COLOR_GRAY], Console.Colors[COLOR_ERROR],
              Console.Colors[COLOR_WARNING], Console.BlackAndWhite, Console.MaxLines);
  }
  with Sender as TQUB_Viewport_Report_PropertiesForm do begin
    ColorBack     := Console.Colors[COLOR_BACK];
    ColorText     := Console.Colors[COLOR_TEXT];
    ColorGray     := Console.Colors[COLOR_GRAY];
    ColorError    := Console.Colors[COLOR_ERROR];
    ColorWarning  := Console.Colors[COLOR_WARNING];
    BlackAndWhite := Console.BlackAndWhite;
    MaxLines      := Console.MaxLines;
  end;//with
end;

procedure TQUB_Viewport_Report.DoPropertiesInternal(Read :Boolean);
var
  Reg :TRegistry;
begin
  Reg := TRegistry.Create;

  try
    if Reg.OpenKey('Software\QUB\Interface '+InterfaceName+'\Viewport '+Name, not Read) then begin
      if Read then
        Console.Width := Round(ReadRegistryFloat(Reg,'Console Width',100.0) * ClientWidth / 100)
      else
        Reg.WriteFloat('Console Width', 100 * Console.Width / ClientWidth);
    end;
    Reg.CloseKey;

  finally

    Reg.Free;

  end;

end;

//Store/Restore the specific properties, at OnCreate, and OnDestroy
procedure TQUB_Viewport_Report.DoProperties(Read :Boolean);
var
  Reg :TRegistry;
begin

  inherited;

  Reg := TRegistry.Create;

  try

    with Reg do begin
      if Read then begin
        if OpenKey('Software\QUB\Viewport ' + Name, False) then begin
          Console.BlackAndWhite       := ReadRegistryBoolean(Reg,'Console Black and White',    Console.BlackAndWhite);
          Console.Colors[COLOR_BACK]  := TColor(ReadRegistryInteger(Reg,'Console Color Back',  clNavy));
          Console.Colors[COLOR_TEXT]  := TColor(ReadRegistryInteger(Reg,'Console Color Text',  clWhite));
          Console.Colors[COLOR_GRAY]  := TColor(ReadRegistryInteger(Reg,'Console Color Gray',  clGray));
          Console.Colors[COLOR_ERROR] := TColor(ReadRegistryInteger(Reg,'Console Color Error', clRed));
          Console.MaxLines            := ReadRegistryInteger(Reg,'Console MaxLines', 2000);
          Console.Redraw;
          if Console.BlackAndWhite then begin
            lstEvaluator.Color        := clWhite;
            lstEvaluator.Font.Color   := clBlack
          end else begin
            lstEvaluator.Color        := Console.Colors[COLOR_BACK];
            lstEvaluator.Font.Color   := Console.Colors[COLOR_TEXT];
          end;
          lstEvaluator.Items.Text     := ReadRegistryString(Reg, 'Evaluator Expression', '');
        end;
      end else begin
        if OpenKey('Software\QUB\Viewport ' + Name, True) then begin
          WriteBool('Console Black and White', Console.BlackAndWhite);
          WriteInteger('Console Color Back', Integer(Console.Colors[COLOR_BACK]));
          WriteInteger('Console Color Text', Integer(Console.Colors[COLOR_TEXT]));
          WriteInteger('Console Color Gray', Integer(Console.Colors[COLOR_GRAY]));
          WriteInteger('Console Color Error', Integer(Console.Colors[COLOR_ERROR]));
          WriteInteger('Console Color Warning', Integer(Console.Colors[COLOR_WARNING]));
          WriteInteger('Console MaxLines', Console.MaxLines);
          WriteString('Evaluator Expression', lstEvaluator.Items.Text);
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;

end;


//******************************************************************************
procedure TQUB_Viewport_Report.PropertiesItemClick(Sender: TObject);
begin
  inherited;
  prPropertiesForm.ShowModal;
end;


//******************************************************************************
procedure TQUB_Viewport_Report.CopyItemClick(Sender: TObject);
begin
  inherited;
  Console.CopyToClipboard;
end;

procedure TQUB_Viewport_Report.ClearItemClick(Sender: TObject);
begin
  inherited;
  Console.Clear;
end;

procedure TQUB_Viewport_Report.FormResize(Sender: TObject);
begin
  inherited;
  lstEvaluator.Width := panEvaluator.ClientWidth - lstEvaluator.Left;
end;

procedure TQUB_Viewport_Report.txtEvaluatorKeyPress(Sender: TObject; var Key: Char);
var
  //Expression :TKPL_Expression;
  Text       :String;
  Index      :Integer;
  CompStop   :Boolean;
  PyResult   :PPyObject;
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    Text := Trim(lstEvaluator.Text);
    if Text = '' then Exit;
    Index := lstEvaluator.Items.IndexOf(Text);
    if Index < 0 then begin
      lstEvaluator.Items.Insert(0, Text);
      if lstEvaluator.Items.Count > MAX_EVALUATOR_HISTORY_COUNT then begin
        lstEvaluator.Items.Delete(lstEvaluator.Items.Count - 1);
      end;
    end else begin
      lstEvaluator.Items.Move(Index, 0);
    end;
    PythonThread.BeginPython;
    try
      PyResult := GetPythonEngine.EvalString('str('+lstEvaluator.Text+')');
      if PyResult = nil then begin
        GetPythonEngine.PyErr_Print;
        raise Exception.Create('');
      end;
      Console.NewEntry(GetPythonEngine.PyString_AsString(PyResult), COLOR_TEXT);
      GetPythonEngine.Py_DECREF(PyResult);
    except 
      on E :Exception do begin
        if E.Message <> '' then begin
          Console.NewEntry('Error: '+E.Message, COLOR_ERROR);
        end else begin
          Console.NewEntry('Error: View -> Python Scripts for details', COLOR_ERROR);
        end;
      end;
    end;
    PythonThread.EndPython;
    lstEvaluator.SelStart      := Length(lstEvaluator.Text);
    lstEvaluator.SelLength     := Length(lstEvaluator.Text);

    {
    Expression := TKPL_Expression.Create('Evaluator', nil);
    try
      Expression.Expression := lstEvaluator.Text;
      if not Expression.Compile(CompStop) then begin
        Console.NewEntry(Expression.Value, COLOR_TEXT);
        lstEvaluator.SelStart  := Length(lstEvaluator.Text);
        lstEvaluator.SelLength := Length(lstEvaluator.Text);
      end else begin
        Console.NewEntry('Error in evaluating expression: ' + Expression.Error, COLOR_ERROR);
        lstEvaluator.SelStart  := Expression.Bookmark.TokenPos - 1;
        lstEvaluator.SelLength := Expression.Bookmark.TokenLen;
      end;
      lstEvaluator.SetFocus;
    finally
      Expression.Free;
    end;
    }
  end;
end;

procedure TQUB_Viewport_Report.txtEvaluatorEnter(Sender: TObject);
begin
  inherited;
  lstEvaluator.SelStart  := 0;
  lstEvaluator.SelLength := Length(lstEvaluator.Text);
end;


//******************************************************************************
end.
