unit QUB_Viewport_Report_Properties;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, StdCtrls,
QUB_Global,
QUB_QVO_Properties;

{$include 'QUB_Viewport_Report_Properties.hct'}

const


DEF_MAXLINES = 512;
MIN_MAXLINES = 10;
MAX_MAXLINES = 1000000;


type


TQUB_Viewport_Report_PropertiesForm = class(TQUB_QVO_PropertiesForm)


  Label1                            :TLabel;
  Label2                            :TLabel;
  Label3                            :TLabel;
  Label4                            :TLabel;
  Label5                            :TLabel;
  Label7                            :TLabel;

  panColorBack                      :TPanel;
  panColorText                      :TPanel;
  panColorGray                      :TPanel;
  panColorError                     :TPanel;
  panColorWarning                   :TPanel;

  chkBlackAndWhite                  :TCheckBox;
  txtMaxLines                       :TEdit;
    btnHelp: TButton;


  procedure panColorBackClick(Sender: TObject);
  procedure panColorTextClick(Sender: TObject);
  procedure panColorGrayClick(Sender: TObject);
  procedure panColorErrorClick(Sender: TObject);
  procedure panColorWarningClick(Sender: TObject);
  procedure chkBlackAndWhiteClick(Sender: TObject);
  procedure txtMaxLinesKeyPress(Sender: TObject; var Key: Char);
  procedure txtMaxLinesExit(Sender: TObject);

  procedure FormCreate(Sender: TObject);
  procedure FormDestroy(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);


protected


  prDlgColor                              :TColorDialog;

  prColorBack                             :TColor;
  prColorText                             :TColor;
  prColorGray                             :TColor;
  prColorError                            :TColor;
  prColorWarning                          :TColor;

  prBlackAndWhite                         :Boolean;

  prMaxLines                              :Integer;

  procedure prSetColorBack(c :TColor);
  procedure prSetColorText(c :TColor);
  procedure prSetColorGray(c :TColor);
  procedure prSetColorError(c :TColor);
  procedure prSetColorWarning(c :TColor);
  procedure prSetBlackAndWhite(bw :Boolean);
  procedure prSetMaxLines(l :Integer);


public


  property ColorBack                      :TColor                read prColorBack             write prSetColorBack;
  property ColorText                      :TColor                read prColorText             write prSetColorText;
  property ColorGray                      :TColor                read prColorGray             write prSetColorGray;
  property ColorError                     :TColor                read prColorError            write prSetColorError;
  property ColorWarning                   :TColor                read prColorWarning          write prSetColorWarning;

  property BlackAndWhite                  :Boolean               read prBlackAndWhite         write prSetBlackAndWhite;

  property MaxLines                       :Integer               read prMaxLines              write prSetMaxLines;


end;


implementation


{$R *.DFM}


//******************************************************************************
procedure TQUB_Viewport_Report_PropertiesForm.prSetColorBack(c :TColor);
begin
  prColorBack := c;
  panColorBack.Color := prColorBack;
  prPropertyChanged := True;
end;

procedure TQUB_Viewport_Report_PropertiesForm.prSetColorText(c :TColor);
begin
  prColorText := c;
  panColorText.Color := prColorText;
  prPropertyChanged := True;
end;

procedure TQUB_Viewport_Report_PropertiesForm.prSetColorGray(c :TColor);
begin
  prColorGray := c;
  panColorGray.Color := prColorGray;
  prPropertyChanged := True;
end;

procedure TQUB_Viewport_Report_PropertiesForm.prSetColorError(c :TColor);
begin
  prColorError := c;
  panColorError.Color := prColorError;
  prPropertyChanged := True;
end;

procedure TQUB_Viewport_Report_PropertiesForm.prSetColorWarning(c :TColor);
begin
  prColorWarning := c;
  panColorWarning.Color := prColorWarning;
  prPropertyChanged := True;
end;

procedure TQUB_Viewport_Report_PropertiesForm.prSetBlackAndWhite(bw :Boolean);
begin
  prBlackAndWhite := bw;
  if bw <> chkBlackAndWhite.Checked then begin
    chkBlackAndWhite.Checked := bw;
  end;
  prPropertyChanged := True;
end;

procedure TQUB_Viewport_Report_PropertiesForm.prSetMaxLines(l :Integer);
begin
  prMaxLines := l;
  txtMaxLines.Text := IntToStr(l);
  prPropertyChanged := True;
end;


//******************************************************************************
procedure TQUB_Viewport_Report_PropertiesForm.panColorBackClick(Sender: TObject);
begin
  inherited;
  prDlgColor.Color := ColorBack;
  if prDlgColor.Execute then begin
    ColorBack := prDlgColor.Color;
  end;
end;

procedure TQUB_Viewport_Report_PropertiesForm.panColorTextClick(Sender: TObject);
begin
  inherited;
  prDlgColor.Color := ColorText;
  if prDlgColor.Execute then begin
    ColorText := prDlgColor.Color;
  end;
end;

procedure TQUB_Viewport_Report_PropertiesForm.panColorGrayClick(Sender: TObject);
begin
  inherited;
  prDlgColor.Color := ColorGray;
  if prDlgColor.Execute then begin
    ColorGray := prDlgColor.Color;
  end;
end;

procedure TQUB_Viewport_Report_PropertiesForm.panColorErrorClick(Sender: TObject);
begin
  inherited;
  prDlgColor.Color := ColorError;
  if prDlgColor.Execute then begin
    ColorError := prDlgColor.Color;
  end;
end;

procedure TQUB_Viewport_Report_PropertiesForm.panColorWarningClick(Sender: TObject);
begin
  inherited;
  prDlgColor.Color := ColorWarning;
  if prDlgColor.Execute then begin
    ColorWarning := prDlgColor.Color;
  end;
end;

procedure TQUB_Viewport_Report_PropertiesForm.chkBlackAndWhiteClick(Sender: TObject);
begin
  inherited;
  BlackAndWhite := TCheckBox(Sender).Checked;
end;

procedure TQUB_Viewport_Report_PropertiesForm.txtMaxLinesKeyPress(Sender: TObject; var Key: Char);
begin
  inherited;
  if Key = #13 then begin
    Key := #0;
    txtMaxLinesExit(Sender);
  end;
end;

procedure TQUB_Viewport_Report_PropertiesForm.txtMaxLinesExit(Sender: TObject);
var
  l :Integer;
begin
  inherited;
  try
    l := StrToInt(TEdit(Sender).Text);
    if (l < MIN_MAXLINES) or (l > MAX_MAXLINES) then raise Exception.Create('');
  except
    MessageDlg('Invalid value; must be between ' + IntToStr(MIN_MAXLINES) + ' and ' + IntToStr(MAX_MAXLINES), mtError, [mbOK], 0);
    MaxLines := MaxLines;
    txtMaxLines.SetFocus;
    Exit;
  end;
  if l <> MaxLines then begin
    MaxLines := l;
  end;
end;


//******************************************************************************
procedure TQUB_Viewport_Report_PropertiesForm.FormCreate(Sender: TObject);
begin
  inherited;
  prDlgColor := TColorDialog.Create(Self);
  {$include 'QUB_Viewport_Report_Properties.htx'}
end;

procedure TQUB_Viewport_Report_PropertiesForm.FormDestroy(Sender: TObject);
begin
  try
    prDlgColor.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_Viewport_Report_PropertiesForm.btnHelpClick(
  Sender: TObject);
begin
  //inherited;
  ShowWiki('Report', Handle);
end;


//******************************************************************************
end.
