{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// was formerly based on the idea that a Result was a document that was listed in ListObjects and could be saved/reopened (before there were session files)


unit QUB_Viewport_Results;


interface


uses


Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ComCtrls, ExtCtrls, StdCtrls, Menus, Registry, 

RegistryInterface,

QUB_Global,

QUB_ListBox,

QUB_Viewport_Files,
QUB_QVO, QUB_QVO_QVR, QUB_Results,
QUB_Node, QUB_QFS,
QUB_PyCmd_MenuButton,
PythonEngine,
QUB_Python_Thread,
QUB_Task;


//{$include 'QUB_Viewport_Results.hct'}


const

WM_QUB_RESULTS_ADD = WM_USER + 105;
DEF_SHOW_HIST_COMPONENTS = True;


type


TQUB_Viewport_Results             = class(TQUB_Viewport_Files)


  procedure PopupMenuPopup(Sender: TObject); override;
  procedure lblFilesClick(Sender: TObject);


protected


  itemRename                      :TMenuItem;
  itemOpen                        :TMenuItem;
  itemSave                        :TMenuItem;

  prMultiResultItemCount          :Integer;

  prShowHistComponents            :Boolean;

  prCurrResults                   :TQUB_QVO_QVR;
  prPopupResults                  :TQUB_QVO_QVR;
  prBlankResults                  :TPanel;


  procedure ListObjectsClick(Sender: TObject);

  procedure prSetCurrResults(Res :TQUB_QVO_QVR);
  function prGetCurrResults: TQUB_QVO_QVR;

  procedure prOnResultsCmdClick(Sender :TObject);
  procedure prOnMultiResultsCmdClick(Sender :TObject);

  procedure prAddBuiltResults(Sender :TObject);

  function prGetHasResults :Boolean;
  function prGetFileName   :String;

  procedure prOnRenameClick(Sender :TObject);
  procedure prOnOpenClick(Sender :TObject);
  procedure prOnSaveClick(Sender :TObject);

  procedure prInitPython;


public


  constructor Create(AOwner :TComponent); override;
  destructor Destroy; override;


  procedure Activate; override;
  procedure DeActivate; override;

  function AddResults(ResultsNode :INode; ASync :Boolean): TQUB_QVO_QVR;
  procedure AddToDisplay(Results :TQUB_QVO_QVR); // ignores file list

  //File
  function New(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean) :Boolean; override;
  function Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean; override;
  function Save(const FileName :String; Silent :Boolean) :Boolean; override;
  function SaveAs(const FileName :String; Silent :Boolean) :Boolean; override;
  function Close :Boolean; override;
  function CloseAll :Boolean; override;

  function OpenFromString(const MemFile :String) :Boolean; override;
  function SaveToString(var MemFile :String; FileType :Integer) :Boolean; override;

  function CanClose :Boolean; override;
  function CanCloseActive :Boolean; override;
  function CanSave :Boolean;

  function CloseResults(Results :TQUB_QVO_QVR) :Boolean;
  function SaveResults(Results :TQUB_QVO_QVR; const FileName :String) :Boolean;
  function SaveAsResults(Results :TQUB_QVO_QVR; const FileName :String) :Boolean;

  function Properties :Boolean; override;

  procedure ResultsPropertiesRegistry(Read :Boolean);


  property CurrResults         :TQUB_QVO_QVR        read prGetCurrResults        write prSetCurrResults;
  property PopupResults        :TQUB_QVO_QVR        read prPopupResults          write prPopupResults;

  property HasResults          :Boolean             read prGetHasResults;
  property FileName            :String              read prGetFileName;

  property ShowHistComponents  :Boolean             read prShowHistComponents    write prShowHistComponents;


end;


TQUB_Python_ResultsView = class(TPyObject)
protected


  prViewportResults     :TQUB_Viewport_Results;

  prResultsToAdd        :INode;
  prAddedResults        :TQUB_QVO_QVR;

  prResult              :PPyObject;
  prIndex               :Integer;


  procedure prGetCurrResults;
  procedure prGetSelectedResults;
  procedure prSqLength;
  procedure prSqItem;

  procedure prAddResultsSync;


public


  function  GetAttr(key : PChar) : PPyObject; override;
  //function  SetAttr(key : PChar; value : PPyObject) : Integer; override;

  // Sequence services
  function  SqLength : Integer; override;
  //function  SqConcat( obj : PPyObject) : PPyObject; override;
  //function  SqRepeat( val : Integer ) : PPyObject; override;
  function  SqItem( idx : Integer ) : PPyObject; override;
  //function  SqSlice( idx1, idx2 : Integer ) : PPyObject; override;
  //function  SqAssItem( idx : integer; obj : PPyObject) : Integer; override;
  //function  SqAssSlice( idx1, idx2 : Integer; obj : PPyObject): integer; override;
  //function  SqContains( obj: PPyObject): integer; override;
  //function  SqInplaceConcat( obj : PPyObject): PPyObject; override;
  //function  SqInplaceRepeat( i: integer): PPyObject; override;

  // Class methods
  class procedure RegisterMethods( PythonType : TPythonType ); override;

  function AddResults( args :PPyObject ): PPyObject; cdecl;

  property ViewportResults :TQUB_Viewport_Results  read prViewportResults   write prViewportResults;
end;


TQUB_Task_BuildResults = class(TQUB_Task)
protected

  procedure prDoOnDone;

public

  ResultsNode                    :INode;
  FileName                       :String;
  Results                        :TQUB_Results;
  QVR                            :TQUB_QVO_QVR;
  OnDone                         :TNotifyEvent;
  Done                           :Boolean;

  constructor Create(Suspended :Boolean); override;
  procedure Execute; override;

  procedure OnGetProperties(Sender :TObject); override; // useless;
  procedure OnPutProperties(Sender :TObject); override; // useless;

end;

TQUB_Task_BuildResultsFromFile = class(TQUB_Task_BuildResults)
public

  procedure Execute; override;

end;



var
  ResultsViewPyType                    :TPythonType;


implementation


uses

QUB_Main,
QUB_Viewport_Python,
QUB_Node_Python,
QUB_GUIInterface,
QUB_Viewport_Data, QUB_QVO_QVD,
QUB_QVO_QVM_Print_Properties;


{$R *.DFM}

//******************************************************************************
constructor TQUB_Viewport_Results.Create(AOwner :TComponent);
begin

  inherited;

  ListObjects.OnClick := ListObjectsClick;

  ResultsPropertiesRegistry(True{Read});

  itemRename := TMenuItem.Create(PopupMenu);
  itemRename.Caption := 'Rename';
  itemRename.OnClick := prOnRenameClick;
  PopupMenu.Items.Add( itemRename );

  itemOpen := TMenuItem.Create(PopupMenu);
  itemOpen.Caption := 'Open...';
  itemOpen.OnClick := prOnOpenClick;
  PopupMenu.Items.Add( itemOpen );

  itemSave := TMenuItem.Create(PopupMenu);
  itemSave.Caption := 'Save As...';
  itemSave.OnClick := prOnSaveClick;
  PopupMenu.Items.Add( itemSave );

  PythonThread.SyncPyProc( prInitPython, True, False, -1 );

  prMultiResultItemCount := 0;

  prCurrResults := nil;

  prBlankResults          := TPanel.Create(Self);
  prBlankResults.Left     := 0;
  prBlankResults.Top      := 0;
  prBlankResults.Width    := ConteinerLeft.Width;
  prBlankResults.Height   := ConteinerLeft.Height;
  prBlankResults.Parent   := ConteinerLeft;
  prBlankResults.Align    := alClient;
  prBlankResults.Anchors  := [akLeft, akTop, akRight, akBottom];

end;

destructor TQUB_Viewport_Results.Destroy;
begin
  try

    ResultsPropertiesRegistry(False{Write});

  finally

    inherited;

  end;
end;


procedure TQUB_Viewport_Results.PopupMenuPopup(Sender :TObject);
var
  i          :Integer;
begin
  for i := PopupMenu.Items.Count - 1 downto PopupMenu.Items.Count - prMultiResultItemCount do begin
    PopupMenu.Items[i].Free;
  end;
  prMultiResultItemCount := 0;

  if ListObjects.SelCount > 0 then begin
    prMultiResultItemCount := prMultiResultItemCount + PyCmd_FillMenu( PopupMenu.Items, 'Results', prOnResultsCmdClick );
  end;

  if ListObjects.SelCount > 1 then begin
    prMultiResultItemCount := prMultiResultItemCount + PyCmd_FillMenu( PopupMenu.Items, 'MultiResults', prOnMultiResultsCmdClick );
  end;
end;

procedure TQUB_Viewport_Results.prOnResultsCmdClick(Sender :TObject);
begin
  PythonThread.InterpretNonSilently(
       Format('Commands.CallByName( "%s", "%s" )', ['Results', TMenuItem(Sender).Caption]) );
end;

procedure TQUB_Viewport_Results.prOnMultiResultsCmdClick(Sender :TObject);
begin
  PythonThread.InterpretNonSilently(
       Format('Commands.CallByName( "%s", "%s" )', ['MultiResults', TMenuItem(Sender).Caption]) );
end;

procedure TQUB_Viewport_Results.prOnRenameClick(Sender :TObject);
var
  Name    :String;
begin
  if ListObjects.ItemIndex >= 0 then begin
    Name := CurrResults.FileName;
    if InputQuery('Rename Results', 'Name: ', Name) then begin
      CurrResults.Rename( Name );
      ListObjects.Items[ ListObjects.ItemIndex ] := Name;
    end;
  end;
end;

procedure TQUB_Viewport_Results.prOnOpenClick(Sender :TObject);
begin
  Open('', False{Silent});
end;

procedure TQUB_Viewport_Results.prOnSaveClick(Sender :TObject);
begin
  SaveAs('', False{Silent});
end;

procedure TQUB_Viewport_Results.prSetCurrResults(Res :TQUB_QVO_QVR);
begin
  if (Res = nil) or (Res.Parent = ConteinerLeft) then prCurrResults := Res;
  if prCurrResults <> nil then begin
    Caption := Format('Results: %s %s', [Res.Results.Node['ResultType'].DataAsString, Res.Results.Node['Date'].DataAsString]);
  end else begin
    prBlankResults.BringToFront;
    Caption := 'Results';
  end;
end;

function TQUB_Viewport_Results.prGetCurrResults: TQUB_QVO_QVR;
begin
  if prPopupResults <> nil then Result := prPopupResults
  else Result := prCurrResults;
end;

//Activate/DeActivate **********************************************************
procedure TQUB_Viewport_Results.Activate;
begin
  inherited;
  if prCurrResults <> nil then prCurrResults.Activate;
end;

procedure TQUB_Viewport_Results.DeActivate;
begin
  inherited;
  if prCurrResults <> nil then prCurrResults.DeActivate;
end;


function TQUB_Viewport_Results.AddResults(ResultsNode :INode; ASync :Boolean): TQUB_QVO_QVR;
var
  Name            :String;
  DSHolder        :INode;
  DS, Seg         :INode;
  i               :Integer;
  f               :String;
  Task            :TQUB_Task_BuildResults;
  Results         :TQUB_QVO_QVR;
  Data            :TQUB_QVO_QVD;
  DataNotOpenCount :Integer;
begin
  Result          := nil;
  if ResultsNode.Find('Date').IsNull then
    ResultsNode['Date'].SetData( Format('%s %s',[DateTimeToStr( Date ),TimeToStr( Time )] ) );

  DSHolder        := ResultsNode;
  if DSHolder.Find('DataSet').IsNull then DSHolder := ResultsNode.Find('ModelTrial');
  
  // set segment indices
  i := 0;
  DS := DSHolder.Find('DataSet');
  while not DS.IsNull do begin
    Seg := DS.Find('Segment');
    while not Seg.IsNull do begin
      Seg['Index'].SetData(i);
      Seg := Seg.SiblingSameName;
      i   := i + 1;
    end;
    DS  := DS.SiblingSameName;
  end;

  // add to open data files
  DataNotOpenCount := 0;
  DS := DSHolder.Find('DataSet');  
  while not DS.IsNull do begin
    Name := ExtractFileName(DS['FileName'].DataAsString);
    Data := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByName[Name]);
    if Data <> nil then begin
      Data.ResultsList.AddResults( ResultsNode );
      Result := Data.ResultsList.CurrResults;
    end else begin
      Inc(DataNotOpenCount);
    end;
    DS := DS.SiblingSameName;
  end;

  if (DataNotOpenCount > 0) or DSHolder.Find('DataSet').IsNull then begin
    Name            := Format('%s %s', [ResultsNode['ResultType'].DataAsString, ResultsNode['Date'].DataAsString]);

    Results := TQUB_QVO_QVR( ListObjects.QVOByName[Name] );
    if Results <> nil then begin
      CloseResults(Results);
    end;

    //Process FileName;
    f := ResultsNode['ResultType'].DataAsString;
    if f = '' then f := GetNewUniqueFileName(ExpandFileName('noname'), '.qrf', prNoNameCount); // should never happen: results come as "...\Results\Latest MIL.qrf" or off disk

    Task                     := TQUB_Task_BuildResults.Create(True);
    Task.FileName            := f;
    Task.ResultsNode         := ResultsNode;
    Task.OnDone              := prAddBuiltResults;
    Task.FreeOnTerminate     := ASync;
    Task.Resume;

    if ASync then Exit;

    while not Task.Done do Application.ProcessMessages;
    Task.WaitFor;
    Result                   := Task.QVR;
    Task.Free;
  end;
end;

procedure TQUB_Viewport_Results.prAddBuiltResults(Sender :TObject);
var
  Task                     :TQUB_Task_BuildResults;
  Results                  :TQUB_QVO_QVR;
  i                        :Integer;
begin
  Task                     := TQUB_Task_BuildResults(Sender);
  if Task.Results = nil then begin
    MessageHelpDlg('Could not open file ' + Task.FileName, mtError, [mbOK], 0, 'Error:Open_File');
    Exit;
  end;

  //Create new Results
  Results                  := TQUB_QVO_QVR.Create(Self, Task.Results);
  Results.Results.FileName := Task.FileName;

  AddToDisplay(Results);
  Results.BuildInterface;

  ListObjects.AddQVO(Results, ExtractFileName(Task.FileName));
  DoOnOpenFile(Results);
  DoOnChangeFile(Results);
  prAddFileEvents(Results);

  for i := 0 to ListObjects.Items.Count - 1 do begin
    ListObjects.Selected[i] := Results = ListObjects.QVO[i];
  end;
  Results.Activate;

  Task.QVR                := Results;
end;

procedure TQUB_Viewport_Results.AddToDisplay(Results :TQUB_QVO_QVR);
begin
  Results.Left             := 0;
  Results.Top              := 0;
  Results.Width            := ConteinerLeft.Width;
  Results.Height           := ConteinerLeft.Height;
  Results.Parent           := ConteinerLeft;
  Results.Align            := alClient;
  Results.Anchors          := [akLeft, akTop, akRight, akBottom];
  Results.TipsOn           := prTipsOn;
  Results.SendToBack;
end;


//****************************************************************************//
//                                                                            //
//                         F I L E     M E T H O D S                          //
//                                                                            //
//****************************************************************************//

// this shit just does not happen like that
function TQUB_Viewport_Results.New(const FileName :String; Silent :Boolean) :Boolean;
begin

  MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  Result := True;
  {
  //Process FileName;
  f := FileName;
  if f = '' then f := GetNewUniqueFileName(ExpandFileName('noname'), '.qrf'); // should never happen: results come as "...\Results\Latest MIL.qrf" or off disk

  //Create new Results
  Results                  := TQUB_QVO_QVR.Create(Self);

  Results.Left             := 0;
  Results.Top              := 0;
  Results.Width            := ConteinerLeft.Width;
  Results.Height           := ConteinerLeft.Height;
  Results.Parent           := ConteinerLeft;
  Results.Align            := alClient;
  Results.Anchors          := [akLeft, akTop, akRight, akBottom];
  Results.BuildInterface;

  Result := Results.New(f);
  ListObjects.AddQVO(Results, ExtractFileName(f));
  DoOnOpenFile(Results);
  DoOnChangeFile(Results);
  prAddFileEvents(Results);
  }
end;

function TQUB_Viewport_Results.Open(const FileName :String; Silent :Boolean) :Boolean;
var
  Task            :TQUB_Task_BuildResultsFromFile;
  Dlg             :TOpenDialog;
  Count           :Integer;
  f, fi           :String;
  m               :Integer;
begin

  Result  := False;
  Dlg := nil;

  //Process FileName;
  try

    f := FileName;
    if f = '' then begin
      Dlg := TOpenDialog.Create(Application);
      Dlg.Filter := 'QUB result files (*.qrf; *.qrt)|*.qrf; *.qrt|' +
                    'All files (*.*)|*.*';
      Dlg.Options := [ofPathMustExist, ofFileMustExist, ofAllowMultiSelect, ofEnableSizing];
      if not Dlg.Execute then Exit;
    end;

    if f <> '' then begin
      Count := 1;
    end else begin
      Count := Dlg.Files.Count;
    end;

    for m := Count - 1 downto 0 do begin

      if f = '' then fi := Dlg.Files[m] else fi := f;

      Task                     := TQUB_Task_BuildResultsFromFile.Create(True);
      Task.FileName            := fi;
      Task.OnDone              := prAddBuiltResults;
      Task.Resume;

    end;

     {
      //Create new Results and open the file
      if ExtractFileExt(fi) = '.qrt' then begin
        ResultsObj          := TQUB_Results.Create(TNode.CreateFromText(fi));
        ResultsObj.FileName := FileName;
        Results             := TQUB_QVO_QVR.Create(Self, ResultsObj);
        Result              := False;
      end else begin
        ResultsObj          := TQUB_Results.Create(fi, False);
        if not ResultsObj.FileIsOpen then begin
          ResultsObj        := TQUB_Results.Create(fi, True);
        end;
        if ResultsObj.FileIsOpen then begin
          Results           := TQUB_QVO_QVR.Create(Self, ResultsObj);
          Result := False;
        end else begin
          Result := True;
        end;
      end;

      if not Result then begin
        Results.Parent           := ConteinerLeft;
        Results.TipsOn           := prTipsOn;
        Results.BuildInterface;
        ListObjects.AddQVO(Results, ExtractFileName(fi));
        DoOnOpenFile(Results);
        DoOnChangeFile(Results);
        prAddFileEvents(Results);
      end else begin
        MessageHelpDlg('Could not open file ' + fi, mtError, [mbOK], 0, 'Error:Open_File');
      end;

    end;

   }
  finally

    Dlg.Free;

  end;

end;

function TQUB_Viewport_Results.Open(const FileName :String; Silent :Boolean; OpenStream :TStream) :Boolean;
begin
  //Chris, do whatever you please here. I did not follow your Results code. Lorin, 07/19/03
  Result  := True;
end;

function TQUB_Viewport_Results.Save(const FileName :String; Silent :Boolean) :Boolean;
var
  Results :TQUB_QVO_QVR;
begin

  Results  := CurrResults;
  Result   := Results = nil;
  if Result then Exit;

  Result := SaveResults(Results, FileName);

end;

function TQUB_Viewport_Results.SaveAs(const FileName :String; Silent :Boolean) :Boolean;
var
  Results    :TQUB_QVO_QVR;
begin

  Results    := CurrResults;
  Result     := Results = nil;
  if Result then Exit;

  Result := SaveAsResults(Results, FileName);

end;

function TQUB_Viewport_Results.Close :Boolean;
var
  Results    :TQUB_QVO_QVR;
begin
  Result   := False;

  if ListObjects.ItemIndex < 0 then Exit;
  Results   := TQUB_QVO_QVR( ListObjects[ ListObjects.ItemIndex ] );

  Result := CloseResults(Results);

end;


function TQUB_Viewport_Results.SaveResults(Results :TQUB_QVO_QVR; const FileName :String) :Boolean;
//var
//  FName :String;
begin
  MessageDlg('This code should never be run!', mtError, [mbOK], 0);
  {
  Result := Results.LockedAny;
  if Result then Exit;

  //Process FileName
  FName := FileName;
  if FName = '' then FName := Results.FileName;
  if (Results.Saved) and (Results.CanSaveAs(FName, True)) then begin
    Result := Results.Save(FName);
   // if Results.ModelObject.Opened then begin     // i.e. this is not part of a close n' save operation?
      ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Results)] := ExtractFileName(Results.FileName);
      DoOnChangeFile(Results);
   // end else begin
   //   Result := True;
   //   CloseResults(Results);
   // end;
  end else begin
    Result := SaveAsResults(Results, FName);
  end;
  }
  Result := False;
end;

function TQUB_Viewport_Results.SaveAsResults(Results :TQUB_QVO_QVR; const FileName :String) :Boolean;
var
  Dlg      :TSaveDialog;
  f        :String;
begin

  Result := Results.LockedAny;
  if Result then Exit;

  //Process FileName
  f := FileName;
  if f = '' then f := Results.FileName;

  Dlg := TSaveDialog.Create(Application);
  Dlg.Filter := 'All QUB result files (*.qrf;*.qrt)|*.qrf;*.qrt|' +
                'QUB result files (*.qrf)|*.qrf|' +
                'QUB results as text (*.qrt)|*.qrt|' +
                'All files (*.*)|*.*';

  if Results.FileExt = '.qrt' then begin
    Dlg.FilterIndex := 3;
  end else begin
    Dlg.FilterIndex := 2;
  end;

  Dlg.Options := [ofPathMustExist{, ofOverwritePrompt}, ofEnableSizing];
  Dlg.Title := 'Save results file as';

  try

    repeat

      Dlg.FileName := ChangeFileExt(f, '');
      if Dlg.Execute then begin
        f := Dlg.FileName;
        case Dlg.FilterIndex of
          2: f := ChangeFileExt(f, '.qrf');
          3: f := ChangeFileExt(f, '.qrt');
        else
            f := ChangeFileExt(f, Results.FileExt);
        end;//case FilterIndex
      end else begin
        Result := True;
        Exit;
      end;

      if (FileExists(f)) and (f <> ExpandFileName(Results.FileName)) then begin
        case MessageDlg('File ''' + f + ''' exists. Do you want to overwrite it ?', mtWarning, [mbYes, mbNo, mbCancel], 0) of
          mrCancel:
            begin
              Result := True;
              Exit;
            end;
          mrNo:
            Continue;
        end;//case
      end;

      //if Results.CanSaveAs(f, True) then begin
        Result := Results.Save(f);
        //Exit;
      //end else begin
      //  r := MessageDlg('Results file cannot be saved in the selected format.'#13#10'Please choose one of the writable formats.', mtError, [mbOK, mbCancel], 0);
      //  Result := r = mrCancel;
      //  if Result then Exit;
      //end;

    until (Result);

  finally

    Dlg.Free;

    //if Results{.ModelObject}.Opened then begin
      //ListObjects.Items.Strings[ListObjects.Items.IndexOfObject(Results)] := ExtractFileName(Results.FileName);
      //DoOnChangeFile(Results);
    //end else begin
    //  Result := True;
    //  CloseResults(Results);
    //end;

  end;

end;

function TQUB_Viewport_Results.CloseResults(Results :TQUB_QVO_QVR) :Boolean;
begin
{
  Result := Results.Close;
  if Result then Exit;
}
  DoOnClosingFile(Results);
  DoOnCloseFile(Results);

  prDelFileEvents(Results);

  ListObjects.DelQVO(Results);
  Results.Free;

  if ListObjects.Items.Count = 0 then DoOnChangeFile(nil);

  Result := False;

end;


function TQUB_Viewport_Results.CloseAll :Boolean;
begin
  Result := False;
  while ListObjects.Items.Count > 0 do begin
    Result := Close;
    if Result then Break;
  end;
end;

function TQUB_Viewport_Results.CanCloseActive :Boolean;
var
  Results :TQUB_QVO_QVR;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  if ListObjects.ItemIndex < 0 then Exit;
  Results   := TQUB_QVO_QVR( ListObjects[ ListObjects.ItemIndex ] );

  if Results.LockedAny then begin
    MessageHelpDlg('File ''' + Results.FileName + ''' is executing a task and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end else if Results.LockedClose then begin
    MessageHelpDlg('File ''' + Results.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
    Result := False;
  end;

  //if not Result then Exit;
end;

function TQUB_Viewport_Results.CanClose :Boolean;
var
  Curr    :TQUB_QVO_QVR;
  Results :TQUB_QVO_QVR;
  // All     :Boolean;
  i       :Integer;
  //r       :Integer;
begin
  Result := inherited CanClose;
  if not Result then Exit;

  // All := False;

  Curr := CurrResults;

  try

    for i := ListObjects.Items.Count - 1 downto 0 do begin

      Results := TQUB_QVO_QVR(ListObjects[i]);

      if Results.LockedAny then begin
        MessageHelpDlg('File ''' + Results.FileName + ''' is executing a task and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end else if Results.Lockedclose then begin
        MessageHelpDlg('File ''' + Results.FileName + ''' is in use and cannot be closed', mtError, [mbOK], 0, 'Error:File_Closing_Busy');
        Result := False;
      end;

      if not Result then Exit;
      {
      if Results.Modified then begin

        ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(Results);
        Results.Activate;

       // if All then begin

          if Results.Saved then begin
            Result := not Results.Save(Results.FileName);
          end else begin
            Result := not SaveAs('');
          end;
          if not Result then Exit;

        end else begin

          r := MessageDlg('File ''' + qvm.FileName + ''' is modified.'#13#10'Do you want to save it ?', mtConfirmation, [mbYes, mbNo, mbYesToAll, mbNoToAll, mbCancel], 0);

          case r of

            mrCancel:
            //Cancel
              begin
                Result := False;
                Exit;
              end;

            mrYes:
            //Yes
              begin
                if qvm.Saved then begin
                  Result := not qvm.Save(qvm.FileName);
                end else begin
                  Result := not SaveAs('');
                end;
                if not Result then Exit;
              end;

            //mrNo:
            //No

            mrNotoAll://9
            //NotoAll
              begin
                Exit;
              end;

            mrYestoAll://10:
            //YestoAll
              begin
                All := True;
                if qvm.Saved then begin
                  Result := not qvm.Save(qvm.FileName);
                end else begin
                  Result := not SaveAs('');
                end;
                if not Result then Exit;
              end;

          end;//case r

        end;//All

      end;
      }
    end;

  finally

    if Curr <> nil then begin
      Curr.Activate;
      ListObjects.ItemIndex := ListObjects.Items.IndexOfObject(Curr);
    end;

  end;

end;

function TQUB_Viewport_Results.CanSave :Boolean;
begin
  if CurrResults <> nil then Result := CurrResults.Modified else Result := False;
end;

function TQUB_Viewport_Results.OpenFromString(const MemFile :String) :Boolean;
begin
  MessageDlg('This code should never be run!', mtError, [mbOK], 0);

  Result := False;
{
  //Process FileName;
  f := MemFile;
  if f = '' then Exit;

  //Create new Data and open the file
  Results                  := TQUB_QVO_QVR.Create(Self);
  Results.Parent           := ConteinerLeft;

  Result                   := Results.OpenFromString(f);

  if not Result then begin
    ListObjects.AddQVO(Results, ExtractFileName(f));        // how can this be a good file name?
    DoOnOpenFile(Results);
    DoOnChangeFile(Results);
    prAddFileEvents(Results);
  end else begin
    prDelFileEvents(Results);
    Results.Free;
    MessageHelpDlg('Could not open file ' + f, mtError, [mbOK], 0, 'Error:Open_File');
  end;
}
end;

function TQUB_Viewport_Results.SaveToString(var MemFile :String; FileType :Integer) :Boolean;
begin
  Result := False;

  if CurrResults <> nil then Result := CurrResults.SaveToString(MemFile, FileType);

end;


function TQUB_Viewport_Results.Properties :Boolean;
begin
  Result := False;
  if CurrResults <> nil then CurrResults.Properties else Result := True;
end;


//****************************************************************************//
//                                                                            //
//                                  O T H E R                                 //
//                                                                            //
//****************************************************************************//
function TQUB_Viewport_Results.prGetHasResults :Boolean;
begin
  Result := CurrResults <> nil;
end;

function TQUB_Viewport_Results.prGetFileName :String;
begin
  if CurrResults <> nil then Result := CurrResults.FileName else Result := '';
end;

procedure TQUB_Viewport_Results.prInitPython;
var
  PyInst                  :PPyObject;
  ResultsViewObj          :TQUB_Python_ResultsView;
begin
  ResultsViewPyType                    := TPythonType.Create(nil);
  ResultsViewPyType.Engine             := GetPythonEngine;
  ResultsViewPyType.Module             := PythonViewport.pmQUBTypes;
  ResultsViewPyType.TypeName           := 'ResultsViewType';
  ResultsViewPyType.Services.Basic     := [bsGetAttr];
  ResultsViewPyType.Services.Sequence  := [ssItem, ssLength];
  ResultsViewPyType.PyObjectClass      := TQUB_Python_ResultsView;
  ResultsViewPyType.Initialize;

  PyInst := ResultsViewPyType.CreateInstance;
  ResultsViewObj := TQUB_Python_ResultsView( PythonToDelphi(PyInst) );
  ResultsViewObj.ViewportResults := Self;
  PyNode_PublishPyObj( PyInst, 'ResultsView' );
  GetPythonEngine.Py_DECREF( PyInst );

  ResultsPyType                    := TPythonType.Create(nil);
  ResultsPyType.Engine             := GetPythonEngine;
  ResultsPyType.Module             := PythonViewport.pmQUBTypes;
  ResultsPyType.TypeName           := 'ResultsType';
  ResultsPyType.Services.Basic     := [bsGetAttr];
  ResultsPyType.Services.Mapping   := [msLength, msSubscript];
  ResultsPyType.PyObjectClass      := TQUB_Python_Results;
  ResultsPyType.Initialize;

  ResultsVarPyType                   := TPythonType.Create(nil);
  ResultsVarPyType.Engine             := GetPythonEngine;
  ResultsVarPyType.Module             := PythonViewport.pmQUBTypes;
  ResultsVarPyType.TypeName           := 'ResultsVarType';
  ResultsVarPyType.Services.Basic     := [bsGetAttr];
  ResultsVarPyType.Services.Sequence  := [ssItem, ssLength];
  ResultsVarPyType.PyObjectClass      := TQUB_Results_Var;
  ResultsVarPyType.Initialize;
end;

{//Parameters Change
procedure TQUB_Viewport_Model.prModelParametersChange(Sender :TObject);
var
  m :Integer;
begin

  for m := 0 to ListObjects.Items.Count - 1 do begin

    TQUB_QVO_QVM(ListObjects.Items.Objects[m]).ModelObject.UpdateParameters;

  end;

end;
}

//Registry *********************************************************************
procedure TQUB_Viewport_Results.ResultsPropertiesRegistry(Read :Boolean);
var
  Reg :TRegistry;
begin
  Reg := TRegistry.Create;
  try
    with Reg do begin
      if OpenKey('Software\QUB\Viewport Results', True) then begin
        if Read then begin
          ShowHistComponents := ReadRegistryBoolean(Reg, 'ShowHistComponents', DEF_SHOW_HIST_COMPONENTS);
        end else begin
          WriteBool('ShowHistComponents', ShowHistComponents);
        end;
      end;
      CloseKey;
    end;

  finally

    Reg.Free;

  end;
end;



//******************************************************************************

function  TQUB_Python_ResultsView.GetAttr(key : PChar) : PPyObject;
begin
  with GetPythonEngine do begin
    if key = 'currResults' then begin
      PythonThread.SyncGUIProc( prGetCurrResults );
      if prResult <> nil then begin
        Result := prResult;
      end else begin
        PyErr_SetString( PyExc_IndexError^, PChar('no results') );
        Result := nil;
      end;
    end else if key = 'selectedResults' then begin
      PythonThread.SyncGUIProc( prGetSelectedResults );
      Result := prResult;
    end else begin
      Result := inherited GetAttr(key);
    end;
  end;
end;

procedure TQUB_Python_ResultsView.prGetCurrResults;
begin
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    if ViewportResults.CurrResults <> nil then begin
      prResult := ViewportResults.CurrResults.PyHandle.GetSelf;
      Py_INCREF( prResult );
    end else begin
      prResult := nil;
    end;
  end;
  PythonThread.EndPython;
end;

procedure TQUB_Python_ResultsView.prGetSelectedResults;
var
  PyResults             :PPyObject;
  i                     :Integer;
begin
  PythonThread.BeginPython;
  try
  with GetPythonEngine do begin
    prResult := PyList_New( 0 );
      for i := 0 to ViewportResults.ListObjects.Items.Count - 1 do begin
        if ViewportResults.ListObjects.Selected[i] then begin
          PyResults := TQUB_QVO_QVR(ViewportResults.ListObjects.QVO[i]).PyHandle.GetSelf;
        PyList_Append( prResult, PyResults );
        end;
      end;
    end;
  finally
    PythonThread.EndPython;
  end;
end;

function  TQUB_Python_ResultsView.SqLength : Integer;
begin
  PythonThread.SyncGUIProc( prSqLength );
  Result := prIndex;
end;

procedure TQUB_Python_ResultsView.prSqLength;
begin
  prIndex := ViewportResults.ListObjects.Items.Count;
end;

function  TQUB_Python_ResultsView.SqItem( idx : Integer ) : PPyObject;
begin
  with GetPythonEngine do begin
    prIndex := idx;
    PythonThread.SyncGUIProc(prSqItem);
    if prResult <> nil then begin
      Result := prResult;
    end else begin
      PyErr_SetString( PyExc_IndexError^, PChar('index out of range') );
      Result := nil;
    end;
  end;
end;

procedure TQUB_Python_ResultsView.prSqItem;
begin
  PythonThread.BeginPython;
  with GetPythonEngine do begin
    if (prIndex >= 0) and (prIndex < ViewportResults.ListObjects.Items.Count) then begin
      prResult := TQUB_QVO_QVR(ViewportResults.ListObjects[prIndex]).PyHandle.GetSelf;
      Py_INCREF( prResult );
    end else begin
      prResult := nil;
    end;
  end;
  PythonThread.EndPython;
end;

class procedure TQUB_Python_ResultsView.RegisterMethods( PythonType : TPythonType );
begin
  inherited;
  with PythonType do begin
    AddMethod( 'addResults',      @TQUB_Python_ResultsView.AddResults,   'ResultsView.addResults(node, name)'  );
  end;
end;

function TQUB_Python_ResultsView.AddResults( args :PPyObject ): PPyObject;
var
  Results               :PPyObject;
  QVR                   :TQUB_QVO_QVR;
begin
  Adjust(@Self);

  with GetPythonEngine do begin
    if PyArg_ParseTuple( args, 'O:ResultView.addResults( node )', [@Results] ) <> 0 then begin
      if not PyNode_IsNode( Results ) then begin
        PyErr_SetString( PyExc_TypeError^, PChar('results must be a qtr.Node') );
        Result := nil;
      end else begin
        prResultsToAdd    := PyNode_ToNode(Results);
        PythonThread.SyncGUIProc( prAddResultsSync );
        QVR               := prAddedResults;
        prResultsToAdd    := nil;

        if QVR = nil then begin
          Result          := ReturnNone;
        end else begin
          Result          := QVR.PyHandle.GetSelf;
          Py_INCREF( Result );
        end;
      end;
    end else begin
      Result := nil;
    end;
  end;
end;

procedure TQUB_Python_ResultsView.prAddResultsSync;
begin
  prAddedResults := ViewportResults.AddResults( prResultsToAdd, False );
end;

//******************************************************************************
constructor TQUB_Task_BuildResults.Create(Suspended :Boolean);
begin
  inherited;
  
  Done        := False;
  Results     := nil;
  QVR         := nil;
end;

procedure TQUB_Task_BuildResults.Execute;
begin
  Results     := TQUB_Results.Create( ResultsNode );
  Synchronize(prDoOnDone);
end;

procedure TQUB_Task_BuildResults.prDoOnDone;
begin
  Done        := True;
  if Assigned(OnDone) then OnDone(Self);
end;

procedure TQUB_Task_BuildResults.OnGetProperties(Sender :TObject);
begin
end;

procedure TQUB_Task_BuildResults.OnPutProperties(Sender :TObject);
begin
end;

//******************************************************************************
procedure TQUB_Task_BuildResultsFromFile.Execute;
begin
  if ExtractFileExt(FileName) = '.qrt' then begin
    Results          := TQUB_Results.Create(TNode.CreateFromText(FileName));
    Results.FileName := FileName;
  end else begin
    Results          := TQUB_Results.Create(FileName, False);
    if not Results.FileIsOpen then begin
      Results        := TQUB_Results.Create(FileName, True);
    end;
    if not Results.FileIsOpen then begin
      Results.Free;
      Results        := nil;
    end;
  end;

  Synchronize(prDoOnDone);
end;

procedure TQUB_Viewport_Results.ListObjectsClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;

procedure TQUB_Viewport_Results.lblFilesClick(Sender: TObject);
begin
  inherited;
  SetFocus;
end;


//******************************************************************************
end.
