unit QUB_WizardScaling;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls,
  RegistryInterface,
  AllocTypes,
  QUB_Viewport_Data,
  QUB_QVO_QVD,
  QUB_Global,
  QUB_Acquisition_Hardware;

type
  TfrmWizardScaling = class(TForm)
    lblWarning: TLabel;
    lblDirections: TLabel;
    btnClose: TButton;
    btnHelp: TButton;
    grpCalibrate: TGroupBox;
    Label13: TLabel;
    Label14: TLabel;
    edtKnownValue: TEdit;
    edtDataAverage: TEdit;
    Label17: TLabel;
    Label18: TLabel;
    btnUpdate: TButton;
    lblChannelName: TLabel;
    lblWarning1: TLabel;
    edtChannelUnits: TEdit;
    Label1: TLabel;
    btnContinue: TButton;
    cmbAIChannels: TComboBox;
    lblCalibrationFactor: TLabel;
    procedure btnCloseClick(Sender: TObject);
    procedure btnContinueClick(Sender: TObject);
    procedure btnHelpClick(Sender: TObject);
    procedure btnUpdateClick(Sender: TObject);
    procedure cmbAIChannelsChange(Sender: TObject);
    procedure SetFactor(Sender: TObject);
  private
    { Private declarations }
  public
    Procedure SetStatus();
    Function CalcFactor() : Double;
  end;

var
  frmWizardScaling: TfrmWizardScaling;

implementation

uses QUB_Main;

//-- Check if hdw, chnl, data set up etc., change warning if necc.
Procedure TfrmWizardScaling.SetStatus();
var
  lEnable : Boolean;
  qvd :TQUB_QVO_QVD;
  avg, std : Double;
  iCh : Integer;
  cLbl1 : String; // string table ?
  cName : String;
  //IBuf : Array of Integer;

begin
  lblWarning.Width := 475;
  lblWarning1.Width := 475;

  qvd := g_CurrData();
  cLbl1 := 'Correct this problem, then click [Grab Data] to try again.';
  if QUBAcqIDevices.Count=0 Then begin
    lblWarning.Caption := 'No  acquisition hardware : You must set up your acquisition hardware prior to setting scaling factors.  Select [Options][Acquisition Hardware] from the main menu to do this.';
    lblWarning1.Caption := cLbl1;
    lEnable := False;
  end else if QUBAcqAIChannels.Count=0 Then begin
    lblWarning.Caption := 'No acquisition channels : You must create an initial configuration for the desired analog input prior to setting a scaling factor.   Select [Options][Acquisition Channels] from the main menu to do this.';
    lblWarning1.Caption := 'Note : You can use a scaling factor of 1 in the initial configuration, then use this utility to correct it. '+cLbl1;
    lEnable := False;
  end else if qvd = nil Then Begin
    lblWarning.Caption := 'No Analog Input data : You must collect some data on the analog input you want to configure prior to setting the scaling factor. ';
    lblWarning1.Caption := cLbl1;
    lEnable := False;
  end else if (qvd.DataSel.Last<=qvd.DataSel.First ) or (not qvd.GetAverage(0,avg,std)) Then begin
    lblWarning.Caption := 'No data selected : To use this utility, you need to select some data of a known magnitude on the desired channel.  Click and drag the mouse in the data window to select the data.';
    lblWarning1.Caption := cLbl1;
    lEnable := False;
  end else begin
    cmbAIChannels.Clear();
    for iCh := 0 to QUBAcqAIChannels.Count-1 do begin
      cName := TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[iCh]).Name;
      cmbAIChannels.Items.Add(cName);
      if (qvd.ActiveChannel>=0) and (qvd.ActiveChannel<qvd.DataChannelCount)
          and (cName = qvd.DataChannelName[qvd.ActiveChannel]) Then
        cmbAIChannels.ItemIndex:=cmbAIChannels.Items.Count-1;
    end;
    if cmbAIChannels.ItemIndex < 0 Then cmbAIChannels.ItemIndex := 0;

    //SetLength(IBuf, (qvd.DataSel.Last-qvd.DataSel.First+1));
    //qvd.LoadDataChannel(@(IBuf[0]), qvd.ActiveChannel, qvd.DataSel, False, False, False);
    //avg := IBuf[0]/qvd.Scaling;
    qvd.GetAverage(0,avg,std);
    edtDataAverage.Text := FloatToStr(avg);

    cmbAIChannelsChange(self);

    lblWarning.Caption := '';
    lblWarning1.Caption := '';
    lEnable := True;
  end;
  lblWarning.Width := 475;
  lblWarning1.Width := 475;
  btnUpdate.Enabled := lEnable;
  edtKnownValue.Enabled := lEnable;
  edtDataAverage.Enabled := lEnable;
end;

{$R *.DFM}

procedure TfrmWizardScaling.btnCloseClick(Sender: TObject);
begin
  Hide();
end;

//----- Recheck the dialogue status
procedure TfrmWizardScaling.btnContinueClick(Sender: TObject);
begin
  SetStatus();
end;

procedure TfrmWizardScaling.btnHelpClick(Sender: TObject);
begin
  ShowWiki('Calibration_Assistant', Handle);
end;

//----- Save changes to scaling factor in channel description
procedure TfrmWizardScaling.btnUpdateClick(Sender: TObject);
var
  oAIChannel : TQUB_Acquisition_AIChannel;
begin
  //-- ensure channel selected
  if (cmbAIChannels.ItemIndex >= QUBAcqAIChannels.Count) or (cmbAIChannels.ItemIndex=-1) then
    exit;
  oAIChannel := TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[cmbAIChannels.ItemIndex]);

  //-- Avoid false edits
  if (oAIChannel.Units = edtChannelUnits.Text) and (oAIChannel.Scaling = CalcFactor()) Then
    exit;

  //-- set new values
  oAIChannel.Units := edtChannelUnits.Text;
  oAIChannel.Scaling := CalcFactor();
  //if oAIChannel.Gain <> 1.0 Then

  //-- Close form, save changes, invoke Channel Change event
  Hide();
  ChannelsWriteRegistry();
  QUB_MainForm.OnChangeChannels(self);
end;

//----- if channel to configure changes (in combo box) look up the units and scaling
procedure TfrmWizardScaling.cmbAIChannelsChange(Sender: TObject);
var
  oAIChannel : TQUB_Acquisition_AIChannel;
begin
  if (cmbAIChannels.ItemIndex>=0) and (cmbAIChannels.ItemIndex < QUBAcqAIChannels.Count) then Begin
    oAIChannel := TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[cmbAIChannels.ItemIndex]);
    edtKnownValue.Text := FloatToStr(ValidStrToFloat(edtDataAverage.Text,1.0)/oAIChannel.Scaling);
    edtChannelUnits.Text := oAIChannel.Units;
  end;
end;

//----- When average or known value changes, set text
procedure TfrmWizardScaling.SetFactor(Sender: TObject);
begin
  lblCalibrationFactor.Caption := 'Calibration factor :' +FloatToStr(CalcFactor());
end;

//----- Calculate factor from average and known value
Function TFrmWizardScaling.CalcFactor() : Double;
begin
  Result := ValidStrToFloat(edtDataAverage.Text,1.0) / ValidStrToFloat(edtKnownValue.Text,1.0);
end;

initialization

end.
