{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
{
  04-01-2004  JB
  Form contains pages for :
    WIZ   Discover available hardware ( QUB configured or not )
    NI    Help user setup National Instrument Card
    CHN   I/O Channel configuration
                                                                           R
  The DLL load routines create devices for all discovered cards, whether
  configured or not.  This form is used to setup configuration parameters for
  the cards.  Config params are saved to registry, then card may be used for
  acquisition.

  Notes/Bugs :
    If two drivers report the same card name, they will conflict.
    NI will only alow 8 AI channels (in library) though cards have 16.
    Scaling needs more help, and needs to be independent of gain.
    Need to add a hint during NI channel edit for possible gains.
    Hardware configuration.  First time QUB is run should open with page 0.
    Channels selection - link in ?
  Note : Not set by NISave, unused ?
    hdw.InputDevices[iDevice].m_ChannelCount
    hdw.InputDevices[iDevice].m_Channels        : Array[ 0..ACQ_MAX_CHANNELS - 1] of Integer;
    hdw.InputDevices[iDevice].m_Gain            : Array[ 0..ACQ_MAX_CHANNELS - 1] of Double;

  05-2005 - Added support for NI - MX library
}

unit QUB_WizardSetup;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, StdCtrls, Grids, ExtCtrls,
  Math,

  RegistryInterface,
  QUB_Global,
  QUB_Events,
  QUB_Acquisition_Global,
  QUB_Acquisition_Hardware, Buttons
  ;

Const
  //----- Captions for Tabs - runtime creation to limit default tab width per sheet
  HDW_CAPTION : String = 'Hardware Configure';
  NI_CAPTION : String = 'National Instruments Card Setup';
  CHN_CAPTION : String = 'Channels Configure';

  HDW_PAGE : Integer = 0;
  NI_PAGE : Integer = 1;
  CHN_PAGE : Integer = 2;

type
  enumChnType = ( CHN_AI, CHN_AO, CHN_DI, CHN_DO ) ;
  enumHdwViewMode = ( HVM_NORMAL, HVM_NISETUP, HVM_DXSETUP, HVM_CHANNELSELECT, HVM_WIZEDIT );
  enumHdwSetupMode = (
                  HSM_NORMAL,                 //All pages available
                  HSM_CHN_EDIT, HSM_CHN_ADD,  //Configuring channels
                  HSM_HDW_EDIT
                  );

  TCardDefaults = class(TObject)
    { Contains default settings for NI card }
    public
      m_lKnownCard      : Boolean;                  // Whether card is a known type or defaults
      m_iCardType       : Integer;                  // N.I. card type returned by NI DLL
      m_iAIChannels     : Integer;                  // Input channels count
      m_iAOChannels     : Integer;                  // Output channels count
      m_iMaxAIRate      : Integer;                  // Max samples per second
      m_iMaxAORate      : Integer;                  // max output per second
      m_iAIUnipolarRange: Integer;                  // Range at gain of 1
      m_iAIBipolarRange : Integer;                  // Range at gain of 1
      m_iAORange        : Integer;                  // Output range default
      m_aiUnipolarGain  : Array[0..9] of Integer;   // array of possible gains
      m_iUnipolarGains  : Integer;                  // # of gains in array, -1=unknown
      m_aiBipolarGain   : Array[0..9] of Integer;   // array of possible gains
      m_iBipolarGains   : Integer;                  // # of gains in array, -1=unknown
      m_iAIBits         : Integer;                  // Analog In bits (usually 12 or 16)
      m_iAOBits         : Integer;                  // Analog Out bits (usually 12 or 16)
      Procedure SetCardParameters(lKnown:Boolean; iAICh,iAOCh,iAIUnipolarRange,iAIBipolarRange,iAORange : Integer; iGainCode,iAIBits,iAOBits,iAIRate,iAORate : Integer);
    Public
      Procedure SetNICardType(iCard:Integer);  // Initialize everything
    end;


  TfrmHardwareConfigure = class(TForm)
    pgsConfigure: TPageControl;

    shtWizard: TTabSheet;
    grpWizNewHdw: TGroupBox;
    lblWiz0: TLabel;
    lblWiz1: TLabel;
    lblWiz2: TLabel;
    lblWiz3: TLabel;
    btnWizSetup: TButton;
    lstWizHardware: TListBox;
    btnWizSearch: TButton;
    btnWizDelete: TButton;
    btnWizEdit: TButton;
    grdWizHardware: TStringGrid;
    grpWizEdit: TGroupBox;
    btnWizSave: TButton;
    btnWizCancel: TButton;
    grpWizInput: TGroupBox;
    lblWizInputBipolar: TLabel;
    lblWizInputUnipolar: TLabel;
    optWizInputPolarity: TRadioButton;
    optWizInputPolarity1: TRadioButton;
    edtWizInputRange: TEdit;
    grpWizOutput: TGroupBox;
    lblWizOutputBipolar: TLabel;
    lblWizOutputUnipolar: TLabel;
    optWizOutputPolarity: TRadioButton;
    optWizOutputPolarity1: TRadioButton;
    edtWizOutputRange: TEdit;
    lblWizInputBits: TLabel;
    edtWizInputBits: TEdit;
    lblWizOutputBits: TLabel;
    edtWizOutputBits: TEdit;
    shtNISetup: TTabSheet;
    lblNI1: TLabel;
    lblNI2: TLabel;
    lblNIWarning: TLabel;
    btnNISave: TButton;
    btnNICancel: TButton;
    grpNIRange: TGroupBox;
    lblNI7: TLabel;
    lblNI6: TLabel;
    grpNIInputRange: TGroupBox;
    optNIInputRange: TRadioButton;
    optNIInputRange2: TRadioButton;
    edtNIInputRange: TEdit;
    grpNIOutputRange: TGroupBox;
    optNIOutputRange: TRadioButton;
    optNIOutputRange2: TRadioButton;
    edtNIOutputRange: TEdit;
    grpNIIOInit: TGroupBox;
    lblNI4: TLabel;
    lblNIInputAChannels: TLabel;
    lblNIInputDChannels: TLabel;
    lblNIBasename: TLabel;
    lblNIOutputAChannels: TLabel;
    lblNIOutputDChannels: TLabel;
    lblNI3: TLabel;
    edtNIInputAChannels: TEdit;
    edtNIInputDChannels: TEdit;
    edtNIBaseName: TEdit;
    edtNIOutputDChannels: TEdit;
    edtNIOutputAChannels: TEdit;
    grpNIBits: TGroupBox;
    lblNIInputBits: TLabel;
    lblNI8: TLabel;
    lblNIOutputBits: TLabel;
    edtNIInputBits: TEdit;
    edtNIOutputBits: TEdit;
    shtChannels: TTabSheet;
    lblChnName: TLabel;
    lblChnUnits: TLabel;
    lblChnIOChannel: TLabel;
    lblChnScaling: TLabel;
    lblChnHdwGain: TLabel;
    lblChnDevice: TLabel;
    lblChnIncrement: TLabel;
    lblChnValue: TLabel;
    Label6: TLabel;
    edtChnName: TEdit;
    edtChnUnits: TEdit;
    edtChnIOChannel: TEdit;
    chkChnUpdate: TCheckBox;
    edtChnScaling: TEdit;
    edtChnHdwGain: TEdit;
    selChnDevice: TComboBox;
    treeChn: TTreeView;
    btnChnEdit: TButton;
    btnChnSave: TButton;
    btnChnCancel: TButton;
    btnChnDelete: TButton;
    btnChnNew: TButton;
    edtChnValue: TEdit;
    edtChnIncrement: TEdit;
    chkChnValue: TCheckBox;
    updownChnIOChannel: TUpDown;
    selChnIOType: TComboBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    btnClose: TButton;
    btnHelp: TButton;
    Label5: TLabel;
    Label7: TLabel;

    procedure KeyNumericOnly(Sender: TObject; var Key: Char);
    procedure pgsConfigureChange(Sender: TObject);
    procedure pgsConfigureChanging(Sender: TObject; var AllowChange: Boolean);
    procedure FormCreate(Sender: TObject);

    procedure btnWizSearchClick(Sender: TObject);
    procedure btnWizSetupClick(Sender: TObject);
    procedure lstWizHardwareClick(Sender: TObject);

    procedure btnNICancelClick(Sender: TObject);
    procedure btnNISaveClick(Sender: TObject);

    procedure btnChnEditClick(Sender: TObject);
    procedure btnChnDeleteClick(Sender: TObject);
    procedure OnBtnCancel(Sender: TObject);
    procedure btnChnSaveClick(Sender: TObject);
    procedure treeChnChange(Sender: TObject; Node: TTreeNode);
    procedure updownChnIOChannelClick(Sender: TObject; Button: TUDBtnType);
    procedure selChnIOTypeChange(Sender: TObject);
    procedure btnChnNewClick(Sender: TObject);
    procedure NISetWarning(Sender: TObject);
    procedure OnBtnClose(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure btnWizDeleteClick(Sender: TObject);
    procedure btnWizEditClick(Sender: TObject);
    procedure btnWizSaveClick(Sender: TObject);
    procedure btnWizCancelClick(Sender: TObject);
    procedure grdWizHardwareSelectCell(Sender: TObject; ACol,
      ARow: Integer; var CanSelect: Boolean);
    procedure btnHelpClick(Sender: TObject);
  private
    { Private declarations }
    m_aiHardwareFound : Array[1..5] of Integer;   // Number of devices found by type in hardware discover.
    m_iPreviousTab : Integer;                     // During tab change, remember which tab it came from.
    m_lChannelSetup : Boolean;                    // Whether Channel controls have been loaded yet.
    m_lHardwareSetup : Boolean;                   // Whether hardware controls have yet been setup.
    m_iMode : enumHdwSetupMode ;                  // Current Mode
    m_iType : enumChnType;                        // During edit / insert : Type of I/O Channel under construction
    m_aiNICodes : TStringList;                    // Array of card codes for NI cards found.  (Delphi needs an IntArray class.)
    m_oCardDefaults : TCardDefaults ;             // Defaults for NI card currently being configured
    m_cNewHdwDLL, m_cNewHdwName : String;         // During setup, name of hardware being added.
    m_lModifiedHardware : Boolean;                // Whether any hardware changes have occurred and need to be written
    m_lModifiedChannels : Boolean;                // Whether any channels changes have occurred and need to be written
    m_iSelected : Integer;                        // List position of item being setup.
    m_lstWizDLLs : TStringList;
    m_lstHWType : Array of Integer;
    m_lstCardIndex : Array of Integer;

    Procedure SetVisiblePage( lView : Boolean; sht : TTabSheet; cCaption : String );
    procedure SetupSelectedHardware();
    Procedure SetupHardwareSheet();
    Procedure SetupWizEditControls(iRow : Integer);
    Procedure SetupChannelsSheet();
    procedure SaveChanges();

  public
    { Public declarations }
    procedure ConfigureDefault(cName, cPrefix : String);
    procedure ConfigureUnknown();
    Procedure SetMode( iMode : enumHdwSetupMode );
    Procedure SetVisible( iView : enumHdwViewMode; iPage : Integer );
    Procedure EnableChnControls( iType : enumChnType );
    Procedure SetChnControls( iType : enumChnType; cName,cUnits : String; iChannel : Integer; lUpdate : Boolean; nScale,nGain,nValue,nIncrement : Double; lValue:Boolean; cDevice : String );
    Procedure GetChnControls( iType : enumChnType; pcName, pcUnits : PString;
                      piChannel : PInteger;  plUpdate : PBool;
                      pnScale, pnGain, pnValue, pnIncrement : PDouble;
                      plValue:PBool; pcDevice : PString; piDevice : PInteger );
  end;

//------------------------------------------------------------------------------
implementation

uses
  QUB_Main;

{$R *.DFM}

Procedure TCardDefaults.SetNICardType(iCard:Integer);
begin
  m_iCardType := iCard;
  case iCard of
    // Note : 6013, 6014 same as 6036 except check AO rate.  What are card #'s for these ?
    //                           (Knwn,ICH,OCH, I+,I+-,O+-,Gain,IBit,OBit,   IRate,   ORate)
    210 :       SetCardParameters(True, 16,  2, 10,  5,	10,		3, 	16,	 12,	 20000,	  20000); //6011E
    76 :        SetCardParameters(True, 16,  2, 10,  5,	10,		1,	12,	 12,  100000,	     20);	//6020E USB
    90,267,268 :SetCardParameters(True, 16,  0, 10,  5,	 0,   4,	12,	  0,  200000,	      0);	//6023E
    91,269,270 :SetCardParameters(True, 16,  2, 10,  5,	10,   4,	12,	 12,	200000,	  10000);	//6024E
    271, 272 :	SetCardParameters(True, 16,  2, 20,  5,	10,   4,	12,	 12,	200000,	  10000);	//6025E
    209 :		    SetCardParameters(True, 16,  2, 20, 10,	10,   2,	16,	 16,	100000,	 100000);	//6030E
    220,259 :	  SetCardParameters(True, 64,  2, 20, 10,	10,   2,	16,	 16,	100000,	 100000);	//6031E
    221 :		    SetCardParameters(True, 16,  0, 20, 10,	 0,   2,	16,	  0,  100000,	      0);	//6032E
    222 :		    SetCardParameters(True, 64,  0, 10, 10,	 0,   2,	16,	  0,  100000,	      0);	//6033E
    314,315 :	  SetCardParameters(True, 16,  0, 10,  5,	 0,   4,	16,	  0,  200000,	      0);	//6034E
    316,317 :	  SetCardParameters(True, 16,  2, 10,  5,	10,   4,	16,	 12,  200000,	  10000);	//6035E
    348 :	      SetCardParameters(True, 16,  2, 10,  5,	10,   4,	16,	 16,  200000,	  10000);	//6036E
    208 :		    SetCardParameters(True, 16,  2, 10,  5,	10,		1,	12,	 12,	500000,	1000000);	//6040E
    273,274,276,
  	332,361 :	  SetCardParameters(True, 16,  2, 10,  5,	10,		1,	16,	 16,	333000,	 333000);	//6052E
    88 :		    SetCardParameters(True, 16,  2, 10,  5,	10,		1,	12,	 12,	500000,	 850000);	//6062E
    207,275 :	  SetCardParameters(True, 16,	 2, 10,  5,	10,		1,	12,	 12, 1250000, 1000000);	//6070E
    223,258 :	  SetCardParameters(True, 64,  2, 20, 10,	10,		1,	12,	 12, 1250000, 1000000);	//6071E
    6229 :      SetCardParameters(True, 32,  4, 20, 10,	10,		5,	16,	 16,  250000,  625000);	//6229 MX
    else
      SetCardParameters(False, 16,	2, 20, 10, 10, -1, 16, 16, 100000, 100000);	//Defaults
    end;
end;


Procedure TCardDefaults.SetCardParameters(lKnown:Boolean; iAICh,iAOCh,iAIUnipolarRange,iAIBipolarRange,iAORange : Integer;
                                          iGainCode,iAIBits,iAOBits,iAIRate,iAORate:Integer);
begin
  m_lKnownCard      := lKnown ;
  m_iAIChannels     := iAICh;
  m_iAOChannels     := iAOCh;
  m_iMaxAIRate      := iAIRate;
  m_iMaxAORate      := iAORate;
  m_iAIUnipolarRange:= iAIUnipolarRange;
  m_iAIBipolarRange := iAIBipolarRange;
  m_iAORange        := iAORange;
  m_iAIBits         := iAIBits;
  m_iAOBits         := iAOBits;

  if iGainCode = 1 Then begin
    m_iUnipolarGains:=7;
    m_aiUnipolarGain[0]:= 1;
    m_aiUnipolarGain[1]:= 2;
    m_aiUnipolarGain[2]:= 5;
    m_aiUnipolarGain[3]:= 10;
    m_aiUnipolarGain[4]:= 20;
    m_aiUnipolarGain[5]:= 50;
    m_aiUnipolarGain[6]:= 100;

    m_iBipolarGains:=8;
    m_aiBipolarGain[0]:= -1;
    m_aiBipolarGain[1]:= 1;
    m_aiBipolarGain[2]:= 2;
    m_aiBipolarGain[3]:= 5;
    m_aiBipolarGain[4]:= 10;
    m_aiBipolarGain[5]:= 20;
    m_aiBipolarGain[6]:= 50;
    m_aiBipolarGain[7]:= 100;
  end else if iGainCode = 2 Then begin
    m_iUnipolarGains:=7;
    m_aiUnipolarGain[0]:= 1;
    m_aiUnipolarGain[1]:= 2;
    m_aiUnipolarGain[2]:= 5;
    m_aiUnipolarGain[3]:= 10;
    m_aiUnipolarGain[4]:= 20;
    m_aiUnipolarGain[5]:= 50;
    m_aiUnipolarGain[6]:= 100;

    m_iBipolarGains:=7;
    m_aiBipolarGain[0]:= 1;
    m_aiBipolarGain[1]:= 2;
    m_aiBipolarGain[2]:= 5;
    m_aiBipolarGain[3]:= 10;
    m_aiBipolarGain[4]:= 20;
    m_aiBipolarGain[5]:= 50;
    m_aiBipolarGain[6]:= 100;
  end else if iGainCode = 3 Then begin
    m_iBipolarGains:=4;
    m_aiBipolarGain[0]:= 1;
    m_aiBipolarGain[1]:= 2;
    m_aiBipolarGain[2]:= 10;
    m_aiBipolarGain[3]:= 100;

    m_iUnipolarGains:=4;
    m_aiUnipolarGain[0]:= 1;
    m_aiUnipolarGain[1]:= 2;
    m_aiUnipolarGain[2]:= 10;
    m_aiUnipolarGain[3]:= 100;
  end else if iGainCode = 4 Then begin
    m_iUnipolarGains:=0;

    m_iBipolarGains:=4;
    m_aiBipolarGain[0]:=-1;
    m_aiBipolarGain[1]:= 1;
    m_aiBipolarGain[2]:= 10;
    m_aiBipolarGain[3]:= 100;
  end else if iGainCode = 5 Then begin
    m_iUnipolarGains:=0;

    m_iBipolarGains:=4;
    m_aiBipolarGain[0]:= 1;
    m_aiBipolarGain[1]:= 2;
    m_aiBipolarGain[2]:= 10;
    m_aiBipolarGain[3]:= 50;
  end else begin
    m_iBipolarGains:=-1;
    m_iUnipolarGains:=-1;
  end;
end;

//----- Hide or show the pages depending on view mode.
Procedure TfrmHardwareConfigure.SetVisible( iView : enumHdwViewMode; iPage : Integer);
var
  lConfigure : Boolean; // True if new visible mode is a hardware configuration page
begin
  lConfigure := (iView=HVM_DXSETUP) or (iView=HVM_NISETUP);
  btnClose.Visible := (iView<>HVM_NISETUP);

  //----- Determine which pages are visible and set captions
  SetVisiblePage((not lConfigure) and (iView=HVM_NORMAL), shtWizard, HDW_CAPTION);
  SetVisiblePage( iView = HVM_NISETUP, shtNISetup, NI_CAPTION);
  SetVisiblePage((not lConfigure) and (iView = HVM_NORMAL) , shtChannels, CHN_CAPTION);

  //----- Switch to specified page
  if iPage >= 0 Then begin
    pgsConfigure.ActivePageIndex := iPage;
    pgsConfigure.ActivePage.Visible := True;
    pgsConfigureChange(self);
  end;

end;

Procedure TfrmHardwareConfigure.SetVisiblePage( lView : Boolean; sht : TTabSheet; cCaption : String );
  begin
  sht.Caption := Choice(lView, cCaption, '');
  sht.Enabled := lView;
end;

//----- Enable / disable various controls depending on configuration mode.
Procedure TfrmHardwareConfigure.SetMode( iMode : enumHdwSetupMode );
var
  l : Boolean;
begin
  m_iMode := iMode;
  btnClose.Enabled := (iMode=HSM_NORMAL);

  l := ((iMode=HSM_CHN_EDIT) or (iMode=HSM_CHN_ADD));
  treeChn.Enabled :=  not l;
  btnChnEdit.Visible :=  not l;
  btnChnDelete.Visible :=  not l;
  btnChnNew.Visible :=  not l;
  selChnIOType.Enabled := (iMode=HSM_CHN_ADD);
  edtChnName.Enabled := l;
  edtChnUnits.Enabled := l;
  edtChnIOChannel.Enabled := l;
  updownChnIOChannel.Enabled := l;
  chkChnUpdate.Enabled := l;
  edtChnValue.Enabled := l;
  edtChnScaling.Enabled := l;
  edtChnHdwGain.Enabled := l;
  edtChnIncrement.Enabled := l;
  chkChnValue.Enabled := l;
  selChnDevice.Enabled := l;
  btnChnSave.Visible := l;
  btnChnCancel.Visible := l;

  l := ( iMode = HSM_HDW_EDIT);
  grpWizEdit.Enabled := l;
  optWizInputPolarity.Enabled := l;
  optWizInputPolarity1.Enabled := l;
  edtWizInputRange.Enabled := l;
  edtWizInputBits.Enabled := l;
  optWizOutputPolarity.Enabled := l;
  optWizOutputPolarity1.Enabled := l;
  edtWizOutputRange.Enabled := l;
  edtWizOutputBits.Enabled := l;
  btnWizSave.Visible := l;
  btnWizCancel.Visible := l;
  grdWizHardware.Enabled := not l;
  grpWizNewHdw.Visible := not l;

end;

//----- On 'Wizard' tab, Search button : This will fill list box with all items
// from each of the supported DLL's.
procedure TfrmHardwareConfigure.btnWizSearchClick(Sender: TObject);
var
  iHWType, iHdw : Integer;                        // index to hardware types
  iDev, iDevCount : Integer;                      // Number of devices found in current DLL
  funcIDevCode : TQUB_Acquisition_InputDeviceCode;// ptr to DLL func for device code
  hdw : TQUB_Acquisition_Hardware;                // AcqHdw object being enumerated
  cQUBDLLName : String;                           // DLL being enumerated

  pchFile : Array[0..MAX_PATH+1] of char;         // Junk buffer to hold file search result
  iLibFind : Integer;                             // Flag whether base (non-QUB) DLL exists
  pchJunk : PChar;                                // Used by SearchPath call
  cFindPath : String;
  iFound : Integer;
  SearchRec : TSearchRec;
begin

  hdw := nil;

  //----- Clean out discovered hardware list from prior run
  m_aiHardwareFound[1] := 0;
  m_aiHardwareFound[2] := 0;
  m_aiHardwareFound[3] := 0;
  m_aiHardwareFound[4] := 0;
  m_aiHardwareFound[5] := 0;
  lstWizHardware.Items.Clear();
  m_aiNICodes.Clear();
  m_lstWizDLLs.Clear();

  //----- For each DLL (hardware type) - check if already loaded.  If so, check
  //----- for new devices.  If not, attempt to load and list devices.
  cFindPath := ExtractFilePath(Application.ExeName) + 'qubacq*.dll';
  iFound := FindFirst(cFindPath, faAnyFile, SearchRec);
  while iFound = 0 do begin
    cQUBDLLName := LowerCase(SearchRec.Name);
    if cQUBDLLName = 'qubacqdx.dll' then begin
      iHWType := 1;
    end else if cQUBDLLName = 'qubacqni.dll' then begin
      iHWType := 2;
    end else if cQUBDLLName = 'qubacqsc.dll' then begin
      iHWType := 3;
    end else if cQUBDLLName = 'qubacqnm.dll' then begin
      iHWType := 4;
    end else begin
      iHWType := 5;
    end;
    //----- Open Hardware DLL.  At completion, iHdw=-1 if no load to be done.

    //----- find hardware in list
    iHdw := QUBHardwareList.IndexOf(cQUBDLLName);

    //----- if DLL not in list and it exists, load it.
    if iHdw = -1 Then begin
      //----- if DLL not already in list, ensure it exists
      if iHWType = 1 Then
        iLibFind := SearchPath( NIL, 'OLMEM32.DLL', NIL, MAX_PATH, pchFile, pchJunk )
      else if iHWType = 2 Then
        //iLibFind := SearchPath( NIL, 'NICAIU.DLL', NIL, MAX_PATH, pchFile, pchJunk )
        iLibFind := SearchPath( NIL, 'NIDAQ32.DLL', NIL, MAX_PATH, pchFile, pchJunk )
      else if iHWType = 4 Then
        iLibFind := SearchPath( NIL, 'NICAIU.DLL', NIL, MAX_PATH, pchFile, pchJunk )
      else
        iLibFind := 1;

      //----- if DLL exists, load it
      if iLibFind > 0 Then Begin
        try
          hdw := TQUB_Acquisition_Hardware.Create(cQUBDLLName);
          if hdw.Loaded Then begin
            iHdw := QUBHardwareList.Count;
            QUBHardwareList.AddObject(cQUBDLLName,hdw);
          end else begin
            hdw.Free;
          end;
        except
          iHdw := -1;
        end;
      end;
    end else begin
      hdw := TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iHdw]);
    end;

    //----- Locate any new devices
    if iHdw >= 0 Then begin
      iDevCount := hdw.{GetLib}IDeviceCount;
      for iDev := 0 to iDevCount-1 do begin
        if not TDeviceParams(hdw.InputDeviceParams[iDev]).m_Loaded Then begin
          lstWizHardware.Items.Add( hdw.IDeviceName[iDev] );
          if Length(m_lstHWType) <= m_lstWizDLLs.Count then begin
            SetLength(m_lstHWType, m_lstWizDLLs.Count+1);
            SetLength(m_lstCardIndex, m_lstWizDLLs.Count+1);
          end;
          m_lstHWType[m_lstWizDlls.Count] := iHWType;
          m_lstWizDLLs.Add( cQUBDLLName );
          Inc(m_aiHardwareFound[iHWType]);
          if (iHWType=2) or (iHWType=4) then begin
            funcIDevCode := GetProcAddress(hdw.LibraryHandle, 'InputDeviceCode');
            if @funcIDevCode=nil then
              funcIDevCode := GetProcAddress(hdw.LibraryHandle, '_InputDeviceCode@4');
            if @funcIDevCode=nil then begin
              m_aiNICodes.Add(IntToStr(iDev));
              m_lstCardIndex[m_lstWizDLLs.Count-1] := iDev;
            end else begin
              m_aiNICodes.Add(IntToStr(funcIDevCode(iDev)));  // For NI cards, store card code
              m_lstCardIndex[m_lstWizDLLs.Count-1] := iDev;
            end;
          end;
        end;
      end;

      iDevCount := hdw.{GetLib}ODeviceCount;
      for iDev := 0 to iDevCount-1 do begin
        if (not TDeviceParams(hdw.OutputDeviceParams[iDev]).m_Loaded)
              and (lstWizHardware.Items.IndexOf(hdw.ODeviceName[iDev])=-1) Then begin
          lstWizHardware.Items.Add( hdw.ODeviceName[iDev] );
          if Length(m_lstHWType) <= m_lstWizDLLs.Count then begin
            SetLength(m_lstHWType, m_lstWizDLLs.Count+1);
            SetLength(m_lstCardIndex, m_lstWizDLLs.Count+1);
          end;
          m_lstHWType[m_lstWizDlls.Count] := iHWType;
          m_lstWizDLLs.Add( cQUBDLLName );
          Inc(m_aiHardwareFound[iHWType]);
          if (iHWType=2) or (iHWType=4) then begin
            funcIDevCode := GetProcAddress(hdw.LibraryHandle, 'InputDeviceCode');
            if @funcIDevCode=nil then
              funcIDevCode := GetProcAddress(hdw.LibraryHandle, '_InputDeviceCode@4');
            if @funcIDevCode=nil then begin
              m_aiNICodes.Add(IntToStr(iDev));
              m_lstCardIndex[m_lstWizDLLs.Count-1] := iDev;
            end else begin
              m_aiNICodes.Add(IntToStr(funcIDevCode(iDev)));  // For NI cards, store card code
              m_lstCardIndex[m_lstWizDLLs.Count-1] := iDev;
            end;
          end;
        end;
      end;
    end;

    iFound := FindNext(SearchRec);
  end;

  //----- If any errors occurred, show combined message
  if (m_aiHardwareFound[1]=0) and (m_aiHardwareFound[2]=0) and (m_aiHardwareFound[3]=0) and (m_aiHardwareFound[4]=0) and (m_aiHardwareFound[5]=0) Then
    MessageHelpDlg('No new hardware discovered.', mtWarning, [mbOk], 0, 'Error:No_New_Hardware');

end;

//----- On 'Wizard' tab, Setup button : Switch to correct config page for
// selected device or message if no currently selected device.
procedure TfrmHardwareConfigure.SetupSelectedHardware();
var
  iCardIndex : Integer;   // Index of selected card within array of all NI cards
begin
  //----- Determine hardware type selected based on position in list (iSel)
  //----- and known count of discovered devices of each type(m_aiHardwareFound).
  if m_lstHWType[m_iSelected] = 1 Then begin

    //----- Data Translation card Setup
    if Application.MessageBox('The Data Translation card will be configured for 16 bits, 10 volts, 2 inputs and 2 outputs.'#13#10'You can change these settings when configuration is complete. Continue with setup ? ','Data Translation Card Setup',MB_ICONQUESTION or MB_OKCANCEL) = IDOK Then Begin
      m_cNewHdwDLL := 'QUBACQDX.DLL';
      m_cNewHdwName := lstWizHardware.Items[m_iSelected];
      ConfigureDefault('Data Translation','DX');
      lstWizHardware.Items.Clear();
    end;

  end else if (m_lstHWType[m_iSelected] = 2)
         or   (m_lstHWType[m_iSelected] = 4) Then Begin
    //----- National Instruments Card Setup
    SetVisible(HVM_NISETUP, NI_PAGE);
    lblNIWarning.Caption:='';
    if m_lstHWType[m_iSelected] = 2 Then Begin
      m_cNewHdwDLL := 'QUBACQNI.DLL';
    end else begin
      m_cNewHdwDLL := 'QUBACQNM.DLL';
    end;
    iCardIndex    := m_lstCardIndex[m_iSelected];
    m_cNewHdwName := lstWizHardware.Items[m_iSelected];

    //----- Default NI Setup values
    m_oCardDefaults.SetNICardType(StrToInt(m_aiNICodes[iCardIndex]));

    optNIInputRange.Checked := True;
    edtNIInputRange.Text := IntToStr(m_oCardDefaults.m_iAIBipolarRange);
    edtNIInputBits.Text := IntToStr(m_oCardDefaults.m_iAIBits);
    edtNIInputAChannels.Text := IntToStr(Min(4,m_oCardDefaults.m_iAIChannels));
    edtNIInputDChannels.Text := '1';

    optNIOutputRange.Checked := True;
    edtNIOutputRange.Text := IntToStr(m_oCardDefaults.m_iAORange);
    edtNIOutputBits.Text := IntToStr(m_oCardDefaults.m_iAOBits);
    edtNIOutputAChannels.Text := IntToStr(m_oCardDefaults.m_iAOChannels);
    edtNIOutputDChannels.Text := '1';

  end else if m_lstHWType[m_iSelected] = 3 Then Begin

    //----- Sound Card Setup
    if Application.MessageBox('The sound card will be configured for 16 bits, 10 volts, 2 inputs and 2 outputs.'#13#10'You can change these settings when configuration is complete. Continue with setup ? ','Sound Card Setup',MB_ICONQUESTION or MB_OKCANCEL) = IDOK Then Begin
      m_cNewHdwDLL := 'QUBACQSC.DLL';
      m_cNewHdwName := lstWizHardware.Items[m_iSelected];
      ConfigureDefault('Sound Card','SC');
      lstWizHardware.Items.Clear();
    end;

  end else if m_lstHWType[m_iSelected] = 5 Then Begin

    //----- Unknown Setup
    m_cNewHdwDLL := m_lstWizDLLs[m_iSelected];
    m_cNewHdwName := lstWizHardware.Items[m_iSelected];
    ConfigureUnknown;
    lstWizHardware.Items.Clear();

  end;
end;

//----- First time onto Hardware page, load grid
Procedure TfrmHardwareConfigure.SetupHardwareSheet();
  var
    i, iRow, ii, iRows : Integer;
    IDev : TQUB_Acquisition_IDevice;
    ODev : TQUB_Acquisition_ODevice;
    DeviceParams : TDeviceParams;
begin
  m_lHardwareSetup := True;
  //----- Setup grid
  iRows := 0;
  for i:=0 to QUBAcqIDevices.Count-1 do begin
    IDev := TQUB_Acquisition_IDevice(QUBAcqIDevices.Objects[i]);
    if IDev.Hardware.InputDeviceParams[IDev.Index].m_Loaded Then
      Inc(iRows);
  end;
  grdWizHardware.RowCount := 1 + iRows;
  grdWizHardware.FixedRows := Min(1,iRows);

  grdWizHardware.ColCount := 8;
  grdWizHardware.ColWidths[0] := 120;
  grdWizHardware.ColWidths[1] := 120;
  grdWizHardware.ColWidths[2] := 50;
  grdWizHardware.ColWidths[3] := 70;
  grdWizHardware.ColWidths[4] := 70;
  grdWizHardware.ColWidths[5] := 50;
  grdWizHardware.ColWidths[6] := 70;
  grdWizHardware.ColWidths[7] := 70;

  grdWizHardware.Cells[0,0] := 'Type';
  grdWizHardware.Cells[1,0] := 'Device Name';
  grdWizHardware.Cells[2,0] := 'In Bits';
  grdWizHardware.Cells[3,0] := 'In Polarity';
  grdWizHardware.Cells[4,0] := 'In Range';
  grdWizHardware.Cells[5,0] := 'Out Bits';
  grdWizHardware.Cells[6,0] := 'Out Polarity';
  grdWizHardware.Cells[7,0] := 'Out Range';

  //---- Fill in Input data
  iRow := 1;
  for i:=0 to QUBAcqIDevices.Count-1 do begin
    IDev := TQUB_Acquisition_IDevice(QUBAcqIDevices.Objects[i]);
    DeviceParams := IDev.Hardware.InputDeviceParams[IDev.Index];
    if DeviceParams.m_Loaded Then begin
      grdWizHardware.Cells[0,iRow] := IDev.Hardware.LibraryName;
      grdWizHardware.Cells[1,iRow] := IDev.Name;
      grdWizHardware.Cells[2,iRow] := IntToStr(DeviceParams.m_Bits);
      if DeviceParams.m_Differential Then Begin
        grdWizHardware.Cells[3,iRow] := 'Bipolar';
        grdWizHardware.Cells[4,iRow] := '+/- '+FloatToStr(DeviceParams.m_Range/2.0);
      end else begin
        grdWizHardware.Cells[3,iRow] := 'Unipolar';
        grdWizHardware.Cells[4,iRow] := '0 to '+FloatToStr(DeviceParams.m_Range);
      end;
      Inc(iRow);
    end;
  end;

  //---- Fill in Output data
  for i:=0 to QUBAcqODevices.Count-1 do begin
    ODev := TQUB_Acquisition_ODevice(QUBAcqODevices.Objects[i]);
    DeviceParams := ODev.Hardware.OutputDeviceParams[ODev.Index];
    if DeviceParams.m_Loaded Then begin
      // Find matching input row or create a new row
      iRow := -1;
      for ii := 1 to grdWizHardware.RowCount-1 do
        if grdWizHardware.Cells[1,ii]=ODev.Name Then
          iRow := ii;

      if iRow = -1 then begin
        iRow := grdWizHardware.RowCount;
        grdWizHardware.RowCount := grdWizHardware.RowCount + 1;
        grdWizHardware.Cells[0,iRow] := ODev.Hardware.LibraryName;
        grdWizHardware.Cells[1,iRow] := ODev.Name;
      end;

      grdWizHardware.Cells[5,iRow] := IntToStr(DeviceParams.m_Bits);
      if DeviceParams.m_Differential Then Begin
        grdWizHardware.Cells[6,iRow] := 'Bipolar';
        grdWizHardware.Cells[7,iRow] := '+/- '+FloatToStr(DeviceParams.m_Range/2.0);
      end else begin
        grdWizHardware.Cells[6,iRow] := 'Unipolar';
        grdWizHardware.Cells[7,iRow] := '0 to '+FloatToStr(DeviceParams.m_Range);
      end;
    end;
  end;

end;

//----- Fill in wizard edit controls with selected row from grid
Procedure TfrmHardwareConfigure.SetupWizEditControls(iRow : Integer);
begin
   edtWizInputBits.Text := grdWizHardware.Cells[2,iRow];
  if grdWizHardware.Cells[3,iRow] = 'Bipolar' Then Begin
    optWizInputPolarity.Checked := True;
    edtWizInputRange.Text := Copy(grdWizHardware.Cells[4,iRow],4,9);
  end else begin
    optWizInputPolarity1.Checked := True;
    edtWizInputRange.Text := Copy(grdWizHardware.Cells[4,iRow],5,9);
  end;

  edtWizOutputBits.Text := grdWizHardware.Cells[5,iRow];
  if grdWizHardware.Cells[6,iRow] = 'Bipolar' Then Begin
    optWizOutputPolarity.Checked := True;
    edtWizOutputRange.Text := Copy(grdWizHardware.Cells[7,iRow],4,9);
  end else begin
    optWizOutputPolarity1.Checked := True;
    edtWizOutputRange.Text := Copy(grdWizHardware.Cells[7,iRow],5,9);
  end;
end;

//----- First time onto Channels page, load tree control
Procedure TfrmHardwareConfigure.SetupChannelsSheet();
var
  nodes : TTreeNodes ;
  nodeAI, nodeAO, nodeDI, nodeDO : TTreeNode ;
  chnAI : TQUB_Acquisition_AIChannel;
  chnAO : TQUB_Acquisition_AOChannel;
  chnDI : TQUB_Acquisition_DIChannel;
  chnDO : TQUB_Acquisition_DOChannel;
  i : Integer;
  c : String;

begin
  m_lChannelSetup := True;

  //----- Root nodes
  nodes := treeChn.Items;
  nodeAI := nodes[0];
  nodeAI.DeleteChildren();

  nodeAO := nodes[1];
  nodeAO.DeleteChildren();

  nodeDI := nodes[2];
  nodeDI.DeleteChildren();

  nodeDO := nodes[3];
  nodeDO.DeleteChildren();

  //----- setup AI Channels
  for i:=0 to QUBAcqAIChannels.Count-1 do begin
    chnAI := TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[i]);
    c := Copy(chnAI.Name+'                ',0,16);
    c := c + ' I/O='+IntToStr(chnAI.Channel)+' | Units='+chnAI.Units+' | Gain='+FloatToStr(chnAI.Gain)+' | Device='+chnAI.DeviceName;
    nodes.AddChild(nodeAI, c );
  end;

  //----- setup AO Channels
  for i:=0 to QUBAcqAOChannels.Count-1 do begin
    chnAO := TQUB_Acquisition_AOChannel(QUBAcqAOChannels.Objects[i]);
    c := Copy(chnAO.Name+'                ',0,16);
    c := c + ' I/O='+IntToStr(chnAO.Channel)+' | Units='+chnAO.Units+' | Device='+chnAO.DeviceName;
    nodes.AddChild(nodeAO, c );
  end;

  //----- setup DI Channels
  for i:=0 to QUBAcqDIChannels.Count-1 do begin
    chnDI := TQUB_Acquisition_DIChannel(QUBAcqDIChannels.Objects[i]);
    c := Copy(chnDI.Name+'                ',0,16);
    c := c + ' I/O='+IntToStr(chnDI.Channel)+' | Device='+chnDI.DeviceName;
    nodes.AddChild(nodeDI, c );
  end;

  //----- setup DO Channels
  for i:=0 to QUBAcqDOChannels.Count-1 do begin
    chnDO := TQUB_Acquisition_DOChannel(QUBAcqDOChannels.Objects[i]);
    c := Copy(chnDO.Name+'                ',0,16);
    c := c + ' I/O='+IntToStr(chnDO.Channel)+' | Device='+chnDO.DeviceName;
    nodes.AddChild(nodeDO, c );
  end;

  //----- Initially expand all nodes
  nodeAO.Expand(False);
  nodeAI.Expand(False);
  nodeDI.Expand(False);
  nodeDO.Expand(False);

end;


//--------------------------------------------------------------Main Form Events
procedure TfrmHardwareConfigure.FormCreate(Sender: TObject);
begin
  //Width := 725;//pgsConfigure.Width + 2 * pgsConfigure.Left ;
  //Height := 620;
  m_lChannelSetup := False;
  m_lHardwareSetup := False;
  SetupHardwareSheet();         // Page 0 by default

  SetMode(HSM_NORMAL);

  m_aiNICodes := TStringList.Create();
  m_oCardDefaults := TCardDefaults.Create() ;
  m_lstWizDLLs := TStringList.Create();

  //----- fill in channel type combo box
  selChnIOType.Items.Clear();
  selChnIOType.Items.Add('Analog Input');
  selChnIOType.Items.Add('Analog Output');
  selChnIOType.Items.Add('Digital Input');
  selChnIOType.Items.Add('Digital Output');

end;

//----- Before page change :
procedure TfrmHardwareConfigure.pgsConfigureChanging(Sender: TObject;  var AllowChange: Boolean);
begin
  //----- Remember current page prior to switch
  m_iPreviousTab := pgsConfigure.ActivePageIndex;

  //----- Do not allow exit from configure pages - Cancel/Save to exit.
  if {( pgsConfigure.ActivePage.Caption = DX_CAPTION ) or}
   ( pgsConfigure.ActivePage.Caption = NI_CAPTION ) Then
      AllowChange := False;
end;

//----- After page change : If switching to unavailable page, switch back.
procedure TfrmHardwareConfigure.pgsConfigureChange(Sender: TObject);
begin
  if pgsConfigure.ActivePage.Caption = '' Then
    pgsConfigure.ActivePageIndex := m_iPreviousTab ;
  if (pgsConfigure.ActivePage.PageIndex=CHN_PAGE) and (not m_lChannelSetup) Then
    SetupChannelsSheet();
  if (pgsConfigure.ActivePage.PageIndex=HDW_PAGE) and (not m_lHardwareSetup) Then
    SetupHardwareSheet();
end;


//------------------------------------------------------------Wizard page Events
//----- [Setup] button or double click list item : Go to hardware page
Procedure TfrmHardwareConfigure.btnWizSetupClick(Sender: TObject);
begin
  m_iSelected := lstWizHardware.ItemIndex;
  if m_iSelected < 0 Then
    Application.MessageBox('You must select an item in the hardware list to Configure.','Hardware Wizard',MB_ICONSTOP or MB_OK)
  else
    SetupSelectedHardware();
end;

procedure TfrmHardwareConfigure.lstWizHardwareClick(Sender: TObject);
begin
  m_iSelected := lstWizHardware.ItemIndex;
  if m_iSelected >= 0 Then
    SetupSelectedHardware();
end;

//----------------------------------------------National Instruments page Events
//----- NI Setup Complete
procedure TfrmHardwareConfigure.btnNISaveClick(Sender: TObject);
var
  iHdw, iDevice, iDevices, iChannel, iChannels, i : Integer;
  chnAI : TQUB_Acquisition_AIChannel;
  chnAO : TQUB_Acquisition_AOChannel;
  chnDI : TQUB_Acquisition_DIChannel;
  chnDO : TQUB_Acquisition_DOChannel;
  IDev : TQUB_Acquisition_IDevice;
  ODev : TQUB_Acquisition_ODevice;
  hdw : TQUB_Acquisition_Hardware;
begin
  m_lChannelSetup := False;
  m_lHardwareSetup := False;
  m_lModifiedHardware:=True;
  m_lModifiedChannels:=True;

  //----- Find hardware
  iHdw := QUBHardwareList.IndexOf(m_cNewHdwDLL);
  if iHdw =-1 Then begin
    Application.MessageBox('Error accessing hardware dll','National Instruments Setup',MB_ICONEXCLAMATION or MB_OK);
    exit;
  end;
  hdw := TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iHdw]);

  //----- find device being setup in Hardware list and global lists
  iDevice := -1;
  iDevices := hdw.{GetLib}IDeviceCount;
  for i := 0 to iDevices-1 do
    if hdw.IDeviceName[i] = m_cNewHdwName Then
      iDevice := i;
  if iDevice=-1 Then begin
    Application.MessageBox('Error accessing hardware device','National Instruments Setup',MB_ICONEXCLAMATION or MB_OK);
    exit;
  end;
  IDev := TQUB_Acquisition_IDevice(QUBAcqIDevices.FindDevice(m_cNewHdwName));
  ODev := TQUB_Acquisition_ODevice(QUBAcqODevices.FindDevice(m_cNewHdwName));
  if (IDev = nil) and{or} (ODev = nil) Then begin
    Application.MessageBox('Error creating hardware device','National Instruments Setup',MB_ICONEXCLAMATION or MB_OK);
    Exit;
  end;

  //----- Create hardware device configuration
  hdw.InputDeviceParams[iDevice].m_Bits := ValidStrToInt( edtNIInputBits.Text, 12 );
  hdw.InputDeviceParams[iDevice].m_Range := ValidStrToInt( edtNIInputRange.Text, 10 );
  hdw.InputDeviceParams[iDevice].m_Differential := optNIInputRange.Checked ;
  hdw.InputDeviceParams[iDevice].m_Loaded := True;
  if hdw.InputDeviceParams[iDevice].m_Differential Then
    hdw.InputDeviceParams[iDevice].m_Range := hdw.InputDeviceParams[iDevice].m_Range * 2;
    //?

  hdw.OutputDeviceParams[iDevice].m_Bits := ValidStrToInt( edtNIOutputBits.Text, 12 );
  hdw.OutputDeviceParams[iDevice].m_Range := ValidStrToInt( edtNIOutputRange.Text, 10 );
  hdw.OutputDeviceParams[iDevice].m_Differential := optNIOutputRange.Checked ;
  hdw.OutputDeviceParams[iDevice].m_Loaded := True;
  if hdw.OutputDeviceParams[iDevice].m_Differential Then
    hdw.OutputDeviceParams[iDevice].m_Range := hdw.OutputDeviceParams[iDevice].m_Range * 2;
    //?

  //----- Create Channels configuration
  iChannels := ValidStrToInt(edtNIInputAChannels.Text,0);       //AI
  for iChannel := 0 to iChannels-1 do begin
    chnAI := TQUB_Acquisition_AIChannel.Create(Trim(edtNIBaseName.Text)+'AI'+IntToStr(iChannel));
    chnAI.Units       := 'V';
    chnAI.Channel     := iChannel;
    chnAI.Update      := False;
    chnAI.Scaling     := 1;
    chnAI.Gain        := 1;
    chnAI.DeviceName  := m_cNewHdwName;
    chnAI.Device      := IDev;
  end;

  iChannels := ValidStrToInt(edtNIOutputAChannels.Text,0);      //AO
  for iChannel := 0 to iChannels-1 do begin
    chnAO := TQUB_Acquisition_AOChannel.Create(Trim(edtNIBaseName.Text)+'AO'+IntToStr(iChannel));
    chnAO.Units       := 'V';
    chnAO.Channel     := iChannel;
    chnAO.Scaling     := 1;
    chnAO.Value       := 0;
    chnAO.Increment   := 0.1;
    chnAO.DeviceName  := m_cNewHdwName;
    chnAO.Device      := ODev;
  end;

  iChannels := ValidStrToInt(edtNIInputDChannels.Text,0);       //DI
  for iChannel := 0 to iChannels-1 do begin
    chnDI := TQUB_Acquisition_DIChannel.Create(Trim(edtNIBaseName.Text)+'DI'+IntToStr(iChannel));
    chnDI.Channel     := iChannel;
    chnDI.Update      := False;
    chnDI.DeviceName  := m_cNewHdwName;
    chnDI.Device      := IDev;
  end;

  iChannels := ValidStrToInt(edtNIOutputDChannels.Text,0);      //DO
  for iChannel := 0 to iChannels-1 do begin
    chnDO := TQUB_Acquisition_DOChannel.Create(Trim(edtNIBaseName.Text)+'DO'+IntToStr(iChannel));
    chnDO.Channel     := iChannel;
    chnDO.Value       := False;
    chnDO.DeviceName  := m_cNewHdwName;
    chnDO.Device      := ODev;
  end;

  //----- Go to hardware page
  lstWizHardware.Items.Clear();
  SetVisible(HVM_NORMAL,CHN_PAGE);
end;

//----- Cancel setup
procedure TfrmHardwareConfigure.btnNICancelClick(Sender: TObject);
begin
  lstWizHardware.Items.Clear();
  SetVisible(HVM_NORMAL,HDW_PAGE);
end;

//----------------------------------------------Sound card configuration save
//----- NI Setup Complete
procedure TfrmHardwareConfigure.ConfigureDefault(cName, cPrefix : String);
var
  iHdw, iDevice, iDevices, iChannel, i : Integer;
  chnAI : TQUB_Acquisition_AIChannel;
  chnAO : TQUB_Acquisition_AOChannel;
  IDev : TQUB_Acquisition_IDevice;
  ODev : TQUB_Acquisition_ODevice;
  hdw : TQUB_Acquisition_Hardware;
  iPrefixLen, iSuffix : Integer;
begin
  //----- Find hardware
  iHdw := QUBHardwareList.IndexOf(m_cNewHdwDLL);
  if iHdw =-1 Then begin
    Application.MessageBox('Error accessing hardware DLL',PChar(cName),MB_ICONEXCLAMATION or MB_OK);
    exit;
  end;
  hdw := TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iHdw]);

  //----- find device being setup in Hardware list and global lists
  iDevice := -1;
  iDevices := hdw.{GetLib}IDeviceCount;
  for i := 0 to iDevices-1 do
    if hdw.IDeviceName[i] = m_cNewHdwName Then
      iDevice := i;
  if iDevice=-1 Then begin
    Application.MessageBox('Error accessing hardware device',PChar(cName),MB_ICONEXCLAMATION or MB_OK);
    Exit;
  end;
  IDev := TQUB_Acquisition_IDevice(QUBAcqIDevices.FindDevice(m_cNewHdwName));
  ODev := TQUB_Acquisition_ODevice(QUBAcqODevices.FindDevice(m_cNewHdwName));
  //if (IDev = nil) or (ODev = nil) Then begin
  if (IDev = nil) and (ODev = nil) Then begin
    Application.MessageBox('Error creating hardware device',PChar(cName),MB_ICONEXCLAMATION or MB_OK);
    Exit;
  end;

  //----- Create hardware device configuration
  hdw.InputDeviceParams[iDevice].m_Bits := 16;
  hdw.InputDeviceParams[iDevice].m_Range := 20;
  hdw.InputDeviceParams[iDevice].m_Differential := True;
  hdw.InputDeviceParams[iDevice].m_Loaded := True;

  hdw.OutputDeviceParams[iDevice].m_Bits := 16;
  hdw.OutputDeviceParams[iDevice].m_Range := 20;
  hdw.OutputDeviceParams[iDevice].m_Differential := True;
  hdw.OutputDeviceParams[iDevice].m_Loaded := True;

  //----- Create Channels configuration
  //-- check for existing channels of new name, change name if found.
  iPrefixLen := Length(cPrefix);
  iSuffix := 0;
  while (QUBAcqAIChannels.IndexOf(cPrefix+' AI0')>=0 )
      or (QUBAcqAIChannels.IndexOf(cPrefix+' AI1')>=0 )
      or (QUBAcqAOChannels.IndexOf(cPrefix+' AO0')>=0 )
      or (QUBAcqAOChannels.IndexOf(cPrefix+' AO1')>=0 ) do begin
      Inc(iSuffix);
      cPrefix := Copy(cPrefix,1,iPrefixLen) + IntToStr(iSuffix);
  end;

  for iChannel := 0 to 1 do begin
    chnAI := TQUB_Acquisition_AIChannel.Create(cPrefix+' AI'+IntToStr(iChannel));
    chnAI.Units       := 'V';
    chnAI.Channel     := iChannel;
    chnAI.Update      := False;
    chnAI.Scaling     := 1;
    chnAI.Gain        := 1;
    chnAI.DeviceName  := m_cNewHdwName;
    chnAI.Device      := IDev;
  end;

  for iChannel := 0 to 1 do begin
    chnAO := TQUB_Acquisition_AOChannel.Create(cPrefix+' AO'+IntToStr(iChannel));
    chnAO.Units       := 'V';
    chnAO.Channel     := iChannel;
    chnAO.Scaling     := 1;
    chnAO.Value       := 0;
    chnAO.Increment   := 0.1;
    chnAO.DeviceName  := m_cNewHdwName;
    chnAO.Device      := ODev;
  end;

  //----- Go to hardware page
  m_lChannelSetup := False;
  m_lHardwareSetup := False;
  m_lModifiedHardware:=True;
  m_lModifiedChannels:=True;
  SetVisible(HVM_NORMAL,CHN_PAGE);

end;

procedure TfrmHardwareConfigure.ConfigureUnknown();
var
  cName, cPrefix : String;
  iHdw, iDevice, iDevices, iChannel, i : Integer;
  chnAI : TQUB_Acquisition_AIChannel;
  chnAO : TQUB_Acquisition_AOChannel;
  chnDI : TQUB_Acquisition_DIChannel;
  chnDO : TQUB_Acquisition_DOChannel;
  IDev : TQUB_Acquisition_IDevice;
  ODev : TQUB_Acquisition_ODevice;
  hdw : TQUB_Acquisition_Hardware;
  iPrefixLen, iSuffix : Integer;
begin
  cName   := 'DAQ Card';
  cPrefix := 'DAQ';

  //----- Find hardware
  iHdw := QUBHardwareList.IndexOf(m_cNewHdwDLL);
  if iHdw =-1 Then begin
    Application.MessageBox('Error accessing hardware DLL',PChar(cName),MB_ICONEXCLAMATION or MB_OK);
    exit;
  end;
  hdw := TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iHdw]);

  //----- find device being setup in Hardware list and global lists
  iDevice := -1;
  iDevices := hdw.{GetLib}IDeviceCount;
  for i := 0 to iDevices-1 do
    if hdw.IDeviceName[i] = m_cNewHdwName Then
      iDevice := i;
  if iDevice=-1 Then begin
    Application.MessageBox('Error accessing hardware device',PChar(cName),MB_ICONEXCLAMATION or MB_OK);
    Exit;
  end;
  IDev := TQUB_Acquisition_IDevice(QUBAcqIDevices.FindDevice(m_cNewHdwName));
  ODev := TQUB_Acquisition_ODevice(QUBAcqODevices.FindDevice(m_cNewHdwName));
  //if (IDev = nil) or (ODev = nil) Then begin
  if (IDev = nil) and (ODev = nil) Then begin
    Application.MessageBox('Error creating hardware device',PChar(cName),MB_ICONEXCLAMATION or MB_OK);
    Exit;
  end;

  //----- Create hardware device configuration
  hdw.InputDeviceParams[iDevice].m_Bits := hdw.GetLibIBits(iDevice);
  hdw.InputDeviceParams[iDevice].m_Range := hdw.GetLibIRange(iDevice);
  hdw.InputDeviceParams[iDevice].m_Differential := 0<>hdw.GetLibIDifferential(iDevice);
  hdw.InputDeviceParams[iDevice].m_Loaded := True;

  hdw.OutputDeviceParams[iDevice].m_Bits := hdw.GetLibOBits(iDevice);
  hdw.OutputDeviceParams[iDevice].m_Range := hdw.GetLibIRange(iDevice);
  hdw.OutputDeviceParams[iDevice].m_Differential := 0<>hdw.GetLibODifferential(iDevice);
  hdw.OutputDeviceParams[iDevice].m_Loaded := True;

  //----- Create Channels configuration
  //-- check for existing channels of new name, change name if found.
  iPrefixLen := Length(cPrefix);
  iSuffix := 0;
  while (QUBAcqAIChannels.IndexOf(cPrefix+' AI0')>=0 )
      or (QUBAcqAIChannels.IndexOf(cPrefix+' AI1')>=0 )
      or (QUBAcqAOChannels.IndexOf(cPrefix+' AO0')>=0 )
      or (QUBAcqAOChannels.IndexOf(cPrefix+' AO1')>=0 ) do begin
      Inc(iSuffix);
      cPrefix := Copy(cPrefix,1,iPrefixLen) + IntToStr(iSuffix);
  end;

  for iChannel := 0 to IDev.Hardware.IDataChannelCount - 1 do begin
    chnAI := TQUB_Acquisition_AIChannel.Create(cPrefix+' AI'+IntToStr(iChannel));
    chnAI.Units       := 'V';
    chnAI.Channel     := iChannel;
    chnAI.Update      := False;
    chnAI.Scaling     := 1;
    chnAI.Gain        := 1;
    chnAI.DeviceName  := m_cNewHdwName;
    chnAI.Device      := IDev;
  end;

  for iChannel := 0 to ODev.Hardware.IDataChannelCount - 1 do begin
    chnAO := TQUB_Acquisition_AOChannel.Create(cPrefix+' AO'+IntToStr(iChannel));
    chnAO.Units       := 'V';
    chnAO.Channel     := iChannel;
    chnAO.Scaling     := 1;
    chnAO.Value       := 0;
    chnAO.Increment   := 0.1;
    chnAO.DeviceName  := m_cNewHdwName;
    chnAO.Device      := ODev;
  end;

  for iChannel := 0 to IDev.Hardware.ODigitalDataChannelCount - 1 do begin
    chnDI := TQUB_Acquisition_DIChannel.Create(cPrefix+' DI'+IntToStr(iChannel));
    chnDI.Channel     := iChannel;
    chnDI.Update      := False;
    chnDI.DeviceName  := m_cNewHdwName;
    chnDI.Device      := IDev;
  end;

  for iChannel := 0 to ODev.Hardware.ODigitalDataChannelCount - 1 do begin
    chnDO := TQUB_Acquisition_DOChannel.Create(cPrefix+'DO'+IntToStr(iChannel));
    chnDO.Channel     := iChannel;
    chnDO.Value       := False;
    chnDO.DeviceName  := m_cNewHdwName;
    chnDO.Device      := ODev;
  end;

  //----- Go to hardware page
  m_lChannelSetup := False;
  m_lHardwareSetup := False;
  m_lModifiedHardware:=True;
  m_lModifiedChannels:=True;
  SetVisible(HVM_NORMAL,CHN_PAGE);

end;

//-----------------------------------------------------------------------Helpers
//----- Limit acceptable characters in numeric edit fields
procedure TfrmHardwareConfigure.KeyNumericOnly(Sender: TObject;  var Key: Char);
begin
  // Not exhaustive.  ex: Cut and paste text
  if (Key<>'.')
    and (Key <> '-')
    and (((Key>=' ') and (Key<'0')) or ((Key > '9') {and (Key<#128)} )) then
    Key := #0;
end;

//------------------------------------------------------Channel Tree view events
//----- Edit selected channel
procedure TfrmHardwareConfigure.btnChnEditClick(Sender: TObject);
var
  nodeCurrent : TTreeNode ;
begin
  nodeCurrent := treeChn.Selected;
  if (nodeCurrent<> nil) and (nodeCurrent.Level=1) Then
    SetMode(HSM_CHN_EDIT);
end;

procedure TfrmHardwareConfigure.btnChnNewClick(Sender: TObject);
begin
  SetMode(HSM_CHN_ADD);
end;

//----- delete I/O Channel
procedure TfrmHardwareConfigure.btnChnDeleteClick(Sender: TObject);
var
  nodeCurrent : TTreeNode ;
begin
  nodeCurrent := treeChn.Selected;
  if (nodeCurrent.Level=1)
    and ( MessageHelpDlg('Are you sure you want to delete this channel ? ', mtConfirmation, [mbYes,mbNo], 0, 'Warning:Delete_This_Channel')= mrYes ) Then Begin
      // determine type of IOChannel to delete
      if nodeCurrent.Parent.Index = 0 Then
        TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[nodeCurrent.Index]).Free()
      else if nodeCurrent.Parent.Index = 1 Then
        TQUB_Acquisition_AOChannel(QUBAcqAOChannels.Objects[nodeCurrent.Index]).Free()
      else if nodeCurrent.Parent.Index = 2 Then
        TQUB_Acquisition_DIChannel(QUBAcqDIChannels.Objects[nodeCurrent.Index]).Free()
      else if nodeCurrent.Parent.Index = 3 Then
        TQUB_Acquisition_DOChannel(QUBAcqDOChannels.Objects[nodeCurrent.Index]).Free();

      m_lModifiedChannels := True;
      SetMode(HSM_NORMAL);  // Unnecc.
      SetupChannelsSheet();
  end;
end;

//----- Update / insert modified channel
procedure TfrmHardwareConfigure.btnChnSaveClick(Sender: TObject);
var
  chnAI : TQUB_Acquisition_AIChannel;
  chnAO : TQUB_Acquisition_AOChannel;
  chnDI : TQUB_Acquisition_DIChannel;
  chnDO : TQUB_Acquisition_DOChannel;
  iCurrent : Integer ;
  nBaseScaling : Double;
begin
  iCurrent := treeChn.Selected.Index;
  //----- For each type, set pointer to edit channel or new channel; Load values.
  // Note : .Create methods automatically add new channels to lists.
  if m_iType = CHN_AI Then Begin
    if m_iMode = HSM_CHN_ADD then
      chnAI := TQUB_Acquisition_AIChannel.Create(edtChnName.Text)
    else
      chnAI := TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[iCurrent]);
    GetChnControls( CHN_AI, @chnAI.Name, @chnAI.Units, @chnAI.Channel,
                    @chnAI.Update, @nBaseScaling, @chnAI.Gain, nil,
                    nil, nil, @chnAI.DeviceName, @chnAI.Device );
    chnAI.BaseScaling := nBaseScaling;
  end;
  if m_iType = CHN_AO Then Begin
    if m_iMode = HSM_CHN_ADD then
      chnAO := TQUB_Acquisition_AOChannel.Create(edtChnName.Text)
    else
      chnAO := TQUB_Acquisition_AOChannel(QUBAcqAOChannels.Objects[iCurrent]);
    GetChnControls( CHN_AO, @chnAO.Name, @chnAO.Units, @chnAO.Channel,
                    nil, @chnAO.Scaling, nil, @chnAO.Value,
                    @chnAO.Increment, nil, @chnAO.DeviceName, @chnAO.Device);
  end;
  if m_iType = CHN_DI Then Begin
    if m_iMode = HSM_CHN_ADD then
      chnDI := TQUB_Acquisition_DIChannel.Create(edtChnName.Text)
    else
      chnDI := TQUB_Acquisition_DIChannel(QUBAcqDIChannels.Objects[iCurrent]);
    GetChnControls( CHN_DI, @chnDI.Name, nil, @chnDI.Channel,
                    @chnDI.Update, nil, nil, nil,
                    nil, nil, @chnDI.DeviceName, @chnDI.Device);
  end;
  if m_iType = CHN_DO Then Begin
    if m_iMode = HSM_CHN_ADD then
      chnDO := TQUB_Acquisition_DOChannel.Create(edtChnName.Text)
    else
      chnDO := TQUB_Acquisition_DOChannel(QUBAcqDOChannels.Objects[iCurrent]);
    GetChnControls( CHN_DO, @chnDO.Name, nil, @chnDO.Channel,
                    nil, nil, nil, nil,
                    nil, @chnDO.Value, @chnDO.DeviceName, @chnDO.Device);
  end;
  //----- Reload tree control
  SetMode(HSM_NORMAL);
  SetupChannelsSheet();
  m_lModifiedChannels := True;
end;

//----- Cancel channel insert / edit
procedure TfrmHardwareConfigure.OnBtnCancel(Sender: TObject);
begin
  SetMode(HSM_NORMAL);
end;

//----- save channel controls values to vars
Procedure TfrmHardwareConfigure.GetChnControls( iType : enumChnType; pcName,pcUnits : PString;
        piChannel : PInteger;  plUpdate : PBool; pnScale, pnGain, pnValue, pnIncrement : PDouble;
        plValue : PBool; pcDevice : PString; piDevice : PInteger );
begin
  //-----Name
  pcName^ := edtChnName.Text;
  //-----Units
  if (iType=CHN_AI) or (iType=CHN_AO) Then
    pcUnits^ := edtChnUnits.Text;
  //-----Channel#
  piChannel^ := StrToInt(edtChnIOChannel.Text);
  //-----Update
  if (iType=CHN_AI) or (iType=CHN_DI) Then plUpdate^ := chkChnUpdate.Checked;
  //-----Value (Double)
  if iType=CHN_AO Then pnValue^ := StrToFloat(edtChnValue.Text);
  //-----Scaling
  if (iType=CHN_AI) or (iType=CHN_AO) Then pnScale^ := StrToFloat(edtChnScaling.Text);
  //----- Gain
  if iType=CHN_AI Then pnGain^ := StrToFloat(edtChnHdwGain.Text);
  //-----Increment
  if iType=CHN_AO Then pnIncrement^ := StrToFloat(edtChnIncrement.Text);
  //-----Value (Boolean)
  if iType=CHN_DO Then plValue^ := chkChnValue.Checked ;
  //----- Device
  piDevice^ := selChnDevice.ItemIndex;
  pcDevice^ := Choice(piDevice^ >= 0 , selChnDevice.Items[piDevice^], '');
end;

//----- Enable controls appropriate to channel type
Procedure TfrmHardwareConfigure.EnableChnControls( iType : enumChnType );
begin
  m_iType := iType;
  //-----Units
  lblChnUnits.Visible := (iType=CHN_AI) or (iType=CHN_AO);
  edtChnUnits.Visible := (iType=CHN_AI) or (iType=CHN_AO);
  //-----Update
  chkChnUpdate.Visible := (iType=CHN_AI) or (iType=CHN_DI);
  //-----Value (Double)
  lblChnValue.Visible := (iType=CHN_AO);
  edtChnValue.Visible := (iType=CHN_AO);
  //-----Scaling
  lblChnScaling.Visible := (iType=CHN_AI) or (iType=CHN_AO);
  edtChnScaling.Visible := (iType=CHN_AI) or (iType=CHN_AO);
  //----- Gain
  lblChnHdwGain.Visible := (iType=CHN_AI);
  edtChnHdwGain.Visible := (iType=CHN_AI);
  //-----Increment
  lblChnIncrement.Visible := (iType=CHN_AO);
  edtChnIncrement.Visible := (iType=CHN_AO);
  //-----Value (Boolean)
  chkChnValue.Visible := (iType=CHN_DO);
end;

//----- Load channel controls
Procedure TfrmHardwareConfigure.SetChnControls( iType : enumChnType; cName,cUnits : String; iChannel : Integer; lUpdate : Boolean; nScale,nGain,nValue,nIncrement : Double; lValue:Boolean; cDevice : String );
var
  i, iSel : Integer;
begin
  m_iType := iType;
  EnableChnControls(m_iType);
  //----- Channel Type
  selChnIOType.ItemIndex := Ord(m_iType);
  //-----Name
  edtChnName.Text := cName;
  //-----Units
  if (iType=CHN_AI) or (iType=CHN_AO) Then edtChnUnits.Text := cUnits;
  //-----Channel#
  edtChnIOChannel.Text := IntToStr(iChannel);
  updownChnIOChannel.Position := iChannel;
  //-----Update
  if (iType=CHN_AI) or (iType=CHN_DI) Then chkChnUpdate.Checked := lUpdate;
  //-----Value (Double)
  if iType=CHN_AO Then edtChnValue.Text := FloatToStr(nValue);
  //-----Scaling
  if (iType=CHN_AI) or (iType=CHN_AO) Then edtChnScaling.Text := FloatToStr(nScale);
  //----- Gain
  if iType=CHN_AI Then edtChnHdwGain.Text := FloatToStr(nGain);
  //-----Increment
  if iType=CHN_AO Then edtChnIncrement.Text := FloatToStr(nIncrement);
  //-----Value (Boolean)
  if iType=CHN_DO Then chkChnValue.Checked := lValue;

  //----- Device
  selChnDevice.Clear();
  iSel := -1;
  if (iType=CHN_AI) or (iType=CHN_DI) Then begin
    for i:=0 to QUBAcqIDevices.Count-1 do begin
      selChnDevice.Items.Add( QUBAcqIDevices.Strings[i] );
      if QUBAcqIDevices.Strings[i] = cDevice Then iSel := i;
    end;
  end else begin
    for i:=0 to QUBAcqIDevices.Count-1 do begin
      selChnDevice.Items.Add( QUBAcqIDevices.Strings[i] );
      if QUBAcqIDevices.Strings[i] = cDevice Then iSel := i;
    end;
  end;
  if iSel>=0 then
    selChnDevice.ItemIndex := iSel;
end;

//---------------------------------------------------------------Experimentation
procedure TfrmHardwareConfigure.treeChnChange(Sender: TObject;  Node: TTreeNode);
var
  nodeParent : TTreeNode;
  chnAI : TQUB_Acquisition_AIChannel;
  chnAO : TQUB_Acquisition_AOChannel;
  chnDI : TQUB_Acquisition_DIChannel;
  chnDO : TQUB_Acquisition_DOChannel;
begin
  if Node.Level = 1 Then begin
    nodeParent := Node.Parent;
    if nodeParent.Index = 0 Then begin
      chnAI := TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[Node.Index]);
      SetChnControls( CHN_AI, chnAI.Name, chnAI.Units, chnAI.Channel,
                      chnAI.Update,chnAI.BaseScaling,chnAI.Gain,0.0{A Value},
                      0.0{Increment},False{D Value},chnAI.DeviceName);
    end;

    if nodeParent.Index = 1 Then begin
      chnAO := TQUB_Acquisition_AOChannel(QUBAcqAOChannels.Objects[Node.Index]);
      SetChnControls( CHN_AO, chnAO.Name, chnAO.Units, chnAO.Channel,
                      False{Update},chnAO.Scaling,0.0{Gain},chnAO.Value,
                      chnAO.Increment,False{D Value},chnAO.DeviceName);
    end;

    if nodeParent.Index = 2 Then begin
      chnDI := TQUB_Acquisition_DIChannel(QUBAcqDIChannels.Objects[Node.Index]);
      SetChnControls( CHN_DI, chnDI.Name, '' {Units}, chnDI.Channel,
                      chnDI.Update,0.0{Scaling},0.0{Gain},0.0{A Value},
                      0.0{Increment},False{D Value},chnDI.DeviceName);
    end;

    if nodeParent.Index = 3 Then begin
      chnDO := TQUB_Acquisition_DOChannel(QUBAcqDOChannels.Objects[Node.Index]);
      SetChnControls( CHN_DO, chnDO.Name, '' {Units}, chnDO.Channel,
                      False{Update},0.0{Scaling},0.0{Gain},0.0{A Value},
                      0.0{Increment},chnDO.Value,chnDO.DeviceName);
    end;

  end;
end;

//----- Change channel via Up/Down arrows
procedure TfrmHardwareConfigure.updownChnIOChannelClick(Sender: TObject; Button: TUDBtnType);
begin
  edtChnIOChannel.Text := IntToStr(StrToInt(edtChnIOChannel.Text) + Choice(Button=btNext,1,-1) );
end;

//----- Close form -> save to registry any updates.
procedure TfrmHardwareConfigure.selChnIOTypeChange(Sender: TObject);
begin
  m_iType := enumChnType(selChnIOType.ItemIndex);
  EnableChnControls(m_iType);
end;

//----- Check NI card setup parameters.  Fill in warning label with any errors.
procedure TfrmHardwareConfigure.NISetWarning(Sender: TObject);
var
  cWarning : String;
begin
  btnNISave.Enabled := True;

  //----- check setting for known card types
  if m_oCardDefaults.m_lKnownCard Then Begin
    //----- Check input range
    if not IsValidInteger(edtNIInputRange.Text)  Then begin
      cWarning := cWarning + 'Input range is required.; ';
      btnNISave.Enabled := False;
    end else
      if optNIInputRange.Checked Then Begin // Bipolar
        if m_oCardDefaults.m_iAIBipolarRange <> StrToInt(edtNIInputRange.Text) Then cWarning := cWarning + 'Card range is usually +/- '+IntToStr(m_oCardDefaults.m_iAIBipolarRange)+'; ';
      end else begin                        // Unipolar
        if m_oCardDefaults.m_iAIUnipolarRange <> StrToInt(edtNIInputRange.Text) Then cWarning := cWarning + 'Card range is usually  0 - '+IntToStr(m_oCardDefaults.m_iAIUnipolarRange)+'; ';
      end;

    //----- Check output range
    if not IsValidInteger(edtNIOutputRange.Text) Then begin
      cWarning := cWarning + 'Output range must be specified or 0.; ';
      btnNISave.Enabled := False;
    end else
      if (optNIOutputRange2.Checked) or (m_oCardDefaults.m_iAORange <> StrToInt(edtNIOutputRange.Text)) Then
        cWarning := cWarning + 'Card range is usually +/- '+IntToStr(m_oCardDefaults.m_iAORange)+'; ';

    //----- Check Analog channel counts
    if not IsValidInteger(edtNIInputAChannels.Text) Then begin
      cWarning := cWarning + 'Input channels must be specified or 0.; ';
      btnNISave.Enabled := False;
    end else if m_oCardDefaults.m_iAIChannels < StrToInt(edtNIInputAChannels.Text) Then
      cWarning := cWarning + 'Maximum of '+IntToStr(m_oCardDefaults.m_iAIChannels)+' AI Channels; ';

    if not IsValidInteger(edtNIOutputAChannels.Text) Then begin
      cWarning := cWarning + 'Output channels must be specified or 0.; ';
      btnNISave.Enabled := False;
    end else if m_oCardDefaults.m_iAOChannels < StrToInt(edtNIOutputAChannels.Text) Then begin
      if m_oCardDefaults.m_iAOChannels=0 Then
        cWarning := cWarning + 'This card may not have analog outputs; '
      else
        cWarning := cWarning + 'Maximum of '+IntToStr(m_oCardDefaults.m_iAOChannels)+' AO Channels; ';
    end;

    //----- Check Input bits
    if not IsValidInteger(edtNIInputBits.Text) Then begin
      cWarning := cWarning + 'Input bits is required.; ';
      btnNISave.Enabled := False;
    end else if m_oCardDefaults.m_iAIBits <> StrToInt(edtNIInputBits.Text) Then
      cWarning := cWarning + 'This card should be set to '+IntToStr(m_oCardDefaults.m_iAIBits)+' input bits; ';

    //----- Check Output bits
    if not IsValidInteger(edtNIOutputBits.Text) Then begin
      cWarning := cWarning + 'Output bits is required.; ';
      btnNISave.Enabled := False;
    end else if m_oCardDefaults.m_iAOBits <> StrToInt(edtNIOutputBits.Text) Then
      cWarning := cWarning + 'This card should be set to '+IntToStr(m_oCardDefaults.m_iAOBits)+' output bits; ';

  end else begin
    cWarning := cWarning + 'Unknown card.  Check settings carefully.; ';
  end;

  if Length(edtNIBaseName.Text) > 8 Then
    cWarning := cWarning + 'Short base names are preferable; ';

  if Length(cWarning) > 0 Then
    cWarning := 'Warning : '+cWarning;
  lblNIWarning.Caption := cWarning;
  lblNIWarning.Width := 550;
end;

procedure TfrmHardwareConfigure.OnBtnClose(Sender: TObject);
begin
  Hide;
end;

procedure TfrmHardwareConfigure.SaveChanges();
begin
  //----- Save hardware changes to registry
  if m_lModifiedHardware Then begin
    HardwareWriteRegistry();
    QUB_MainForm.OnChangeHardware(self);
    m_lModifiedHardware := False;
  end;
  //----- Save channels changes to registry
  if m_lModifiedChannels Then begin
    ChannelsWriteRegistry();
    QUB_MainForm.OnChangeChannels(self);
    m_lModifiedChannels := False;
  end;
end;

procedure TfrmHardwareConfigure.FormHide(Sender: TObject);
begin
  SaveChanges();
  ModalResult := mrOK;
end;

procedure TfrmHardwareConfigure.btnWizDeleteClick(Sender: TObject);
var
  iCh, iHdw, iDev : Integer;
  iRow : Integer;                   // Row being removed
  cName : String;                   // Device being removed.
  hdw : TQUB_Acquisition_Hardware;  // Hardware device
begin
  iRow := grdWizHardware.Row;
  if iRow < 1 Then
    exit;
  if Application.MessageBox('Delete selected card configuration ? ','QUB Hardware Configuration',MB_ICONQUESTION or MB_OKCANCEL) = IDCANCEL Then
    exit;

  cName := grdWizHardware.Cells[1,iRow];
  //---- Delete input channels
  for iCh := QUBAcqAIChannels.Count-1 downTo 0 do
    if TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[iCh]).DeviceName = cName Then
      TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[iCh]).Free();

  for iCh := QUBAcqDIChannels.Count-1 downTo 0 do
    if TQUB_Acquisition_DIChannel(QUBAcqDIChannels.Objects[iCh]).DeviceName = cName Then
      TQUB_Acquisition_DIChannel(QUBAcqDIChannels.Objects[iCh]).Free();

  //---- Delete input device
  for iHdw := 0 to QUBHardwareList.Count-1 do Begin
    hdw := TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iHdw]);
    for iDev := 0 to hdw.{GetLib}IDeviceCount do
      if hdw.IDeviceName[iDev]=cName Then
        TDeviceParams(hdw.InputDeviceParams[iDev]).m_Loaded := False;
  end;

  //---- Delete output channels
  for iCh := QUBAcqAOChannels.Count-1 downTo 0 do
    if TQUB_Acquisition_AOChannel(QUBAcqAOChannels.Objects[iCh]).DeviceName = cName Then
      TQUB_Acquisition_AOChannel(QUBAcqAOChannels.Objects[iCh]).Free();

  for iCh := QUBAcqDOChannels.Count-1 downTo 0 do
    if TQUB_Acquisition_DOChannel(QUBAcqDOChannels.Objects[iCh]).DeviceName = cName Then
      TQUB_Acquisition_DOChannel(QUBAcqDOChannels.Objects[iCh]).Free();

  //---- Delete Output device
  for iHdw := 0 to QUBHardwareList.Count-1 do Begin
    hdw := TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iHdw]);
    for iDev := 0 to hdw.GetLibODeviceCount do
      if hdw.ODeviceName[iDev]=cName Then
        TDeviceParams(hdw.OutputDeviceParams[iDev]).m_Loaded := False;
  end;

  //----- Rebuild grid
  m_lModifiedHardware := True;
  m_lModifiedChannels := True;

  SetupHardwareSheet();
  SetupChannelsSheet(); //m_lChannelSetup := False;
  if iRow < grdWizHardware.RowCount Then
    grdWizHardware.Row := iRow;
end;

//----- Hardware wizard page - edit existing hardware
procedure TfrmHardwareConfigure.btnWizEditClick(Sender: TObject);
begin
  if grdWizHardware.Row >= 1 Then begin
    SetMode( HSM_HDW_EDIT );
    SetupWizEditControls(grdWizHardware.Row);
  end;
end;

//----- Save changes to Hardware, set update flag.
procedure TfrmHardwareConfigure.btnWizSaveClick(Sender: TObject);
var
  iRow : Integer;
  iDevice, iDevices : Integer;
  cName : String;
  i : Integer;
  IDev : TQUB_Acquisition_IDevice;
  ODev : TQUB_Acquisition_ODevice;
  hdw : TQUB_Acquisition_Hardware;  // Hardware device
begin
  //----- Find hardware
  iRow := grdWizHardware.Row;
  hdw := TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iRow-1]);
  cName := grdWizHardware.Cells[1,iRow];

  //----- find device being setup in Hardware list and global lists
  iDevice := -1;
  iDevices := hdw.{GetLib}IDeviceCount;
  for i := 0 to iDevices-1 do
    if hdw.IDeviceName[i] = cName Then
      iDevice := i;
  if iDevice=-1 Then begin
    Application.MessageBox('Error accessing hardware device','Hardware Setup',MB_ICONEXCLAMATION or MB_OK);
    exit;
  end;
  IDev := TQUB_Acquisition_IDevice(QUBAcqIDevices.FindDevice(cName));
  ODev := TQUB_Acquisition_ODevice(QUBAcqODevices.FindDevice(cName));
  if (IDev=nil) or (ODev=nil) Then begin
    Application.MessageBox('Error finding hardware device','Hardware Setup',MB_ICONEXCLAMATION or MB_OK);
    exit;
  end;

  //----- Create hardware device configuration
  hdw.InputDeviceParams[iDevice].m_Bits := ValidStrToInt( edtWizInputBits.Text, 12 );
  hdw.InputDeviceParams[iDevice].m_Range := ValidStrToInt( edtWizInputRange.Text, 10 );
  hdw.InputDeviceParams[iDevice].m_Differential := optWizInputPolarity.Checked ;
  hdw.InputDeviceParams[iDevice].m_Loaded := True;
  if hdw.InputDeviceParams[iDevice].m_Differential Then
    hdw.InputDeviceParams[iDevice].m_Range := hdw.InputDeviceParams[iDevice].m_Range * 2;
    //?

  hdw.OutputDeviceParams[iDevice].m_Bits := ValidStrToInt( edtWizOutputBits.Text, 12 );
  hdw.OutputDeviceParams[iDevice].m_Range := ValidStrToInt( edtWizOutputRange.Text, 10 );
  hdw.OutputDeviceParams[iDevice].m_Differential := optWizOutputPolarity.Checked ;
  hdw.OutputDeviceParams[iDevice].m_Loaded := True;
  if hdw.OutputDeviceParams[iDevice].m_Differential Then
    hdw.OutputDeviceParams[iDevice].m_Range := hdw.OutputDeviceParams[iDevice].m_Range * 2;
    //?

  m_lModifiedHardware := True;
  m_lHardwareSetup := False;
  SetVisible(HVM_NORMAL,HDW_PAGE);
  SetMode( HSM_NORMAL );
end;

//----- Cancel HDW edit - reload controls
procedure TfrmHardwareConfigure.btnWizCancelClick(Sender: TObject);
begin
  SetMode( HSM_NORMAL );
  if grdWizHardware.Row >= 1 Then
    SetupWizEditControls(grdWizHardware.Row);
end;

//----- When grid row changes, reload edit controls
procedure TfrmHardwareConfigure.grdWizHardwareSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
begin
  if ARow >= 1 then
    SetupWizEditControls(ARow);
end;

procedure TfrmHardwareConfigure.btnHelpClick(Sender: TObject);
begin
  //Application.HelpCommand(HELP_CONTEXT,15552);
  ShowWiki('Hardware_Configuration', Handle);
end;

end.

