{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the zoom/scroll stuff in QUB Visual Protocol


unit QUB_ZoomBar;

interface

uses 
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Math,
  StdCtrls, Buttons;

type

TQUB_ZoomBar = class(TFrame)
  btnZoomIn: TBitBtn;
  btnZoomOut: TBitBtn;
  txtPercent: TEdit;
  Label1: TLabel;
  scrStart: TScrollBar;

  procedure btnZoomInClick(Sender: TObject);
  procedure btnZoomOutClick(Sender: TObject);
  procedure txtPercentKeyPress(Sender: TObject; var Key: Char);
  procedure txtPercentExit(Sender: TObject);
  procedure scrStartScroll(Sender: TObject; ScrollCode: TScrollCode;
      var ScrollPos: Integer);


protected
  prZoom             :Double;
  prZoomMin          :Double;
  prZoomMax          :Double;

  prStart            :Double;

  procedure prSetZoom(x :Double);
  procedure prSetZoomMin(x :Double);
  procedure prSetZoomMax(x :Double);
  procedure prSetStart(x :Double);

public
  OnZoomChange       :TNotifyEvent;
  OnStartChange      :TNotifyEvent;

  constructor Create(AOwner :TComponent); override;

  property Zoom      :Double          read prZoom       write prSetZoom;
  property ZoomMin   :Double          read prZoomMin    write prSetZoomMin;
  property ZoomMax   :Double          read prZoomMax    write prSetZoomMax;

  property Start     :Double          read prStart      write prSetStart;
end;

implementation

{$R *.DFM}

constructor TQUB_ZoomBar.Create(AOwner :TComponent);
begin
  inherited;
  prZoomMin          := 100;
  prZoomMax          := 102400;
  Zoom               := 100;
  Start              := 0;
end;

procedure TQUB_ZoomBar.prSetZoom(x :Double);
begin
  x := Max(x, ZoomMin);
  x := Min(x, ZoomMax);
  prZoom := x;
  txtPercent.Text := FloatToStr(x);
  scrStart.Max := Round(1000 * (1 - 100 / x));
  scrStart.Enabled := x > 100;
  Start := Start;
end;

procedure TQUB_ZoomBar.prSetZoomMin(x :Double);
begin
  prZoomMin := x;
  Zoom := Zoom;
end;

procedure TQUB_ZoomBar.prSetZoomMax(x :Double);
begin
  prZoomMax := x;
  Zoom := Zoom;
end;

procedure TQUB_ZoomBar.prSetStart(x :Double);
var
  MaxStart    :Double;
begin
  MaxStart    := 100 * (1 - 100 / Zoom);
  if x > MaxStart then x := MaxStart;
  if x < 0 then x := 0;
  prStart := x;
  scrStart.Position := Round(10*x);
end;



procedure TQUB_ZoomBar.btnZoomInClick(Sender: TObject);
begin
  Zoom := Zoom * 2;
  if Assigned(OnZoomChange) then OnZoomChange(Self);
end;

procedure TQUB_ZoomBar.btnZoomOutClick(Sender: TObject);
begin
  Zoom := Zoom / 2;
  if Assigned(OnZoomChange) then OnZoomChange(Self);
end;

procedure TQUB_ZoomBar.txtPercentKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then begin
    Key := #0;
    txtPercentExit(Sender);
  end;
end;

procedure TQUB_ZoomBar.txtPercentExit(Sender: TObject);
var
  x    :Double;
begin
  inherited;
  try
    x := StrToFloat(TEdit(Sender).Text);
    if (x < ZoomMin) or (x > ZoomMax) then raise Exception.Create('');
    Zoom := x;
    if Assigned(OnZoomChange) then OnZoomChange(Self);
  except
    MessageDlg('Invalid value; must be between '+FloatToStr(ZoomMin)+' and '+FloatToStr(ZoomMax)+'.', mtError, [mbOK], 0);
    txtPercent.Text := FloatToStr(Zoom);
    TEdit(Sender).SetFocus;
    Exit;
  end;
end;

procedure TQUB_ZoomBar.scrStartScroll(Sender: TObject;
  ScrollCode: TScrollCode; var ScrollPos: Integer);
begin
  Start := ScrollPos / 10;
  if Assigned(OnStartChange) then OnStartChange(Self);
end;


// *****************************************************************************

end.
