{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// Custom control to allow only numeric values on inputs and check the range
// for the value before exiting.

unit RangedEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

type
  TRangedEdit = class(TEdit)
  private
    F_MinValue:Double;
    F_MaxValue:Double;
    F_ErrorMessage:String;
    F_OriginalValue:Double;

  protected
    procedure KeyPress(var Key: Char); override;
    procedure DoExit(); override;
    procedure DoEnter(); override;

  public
    procedure Setup( parent : TWinControl; iLft, iTop, iWid, iHgt : Integer;
                    iTabOrder : Integer; cText : String; nMin, nMax : Double );

  published
    { Published declarations }
    property MinValue : Double read F_MinValue write F_MinValue ;
    property MaxValue : Double read F_MaxValue write F_MaxValue ;
    property ErrorMessage : String read F_ErrorMessage write F_ErrorMessage;
  end;

procedure Register;

implementation

// on exit - warning if not a valid numeric or out of acceptable range.
procedure TRangedEdit.DoEnter();
begin
  try
    F_OriginalValue := StrToFloat(Text);
    if F_OriginalValue < F_MinValue Then
      F_OriginalValue := F_MinValue;
    if F_OriginalValue > F_MaxValue Then
      F_OriginalValue := F_MaxValue;
  except
    F_OriginalValue := F_MinValue;
  end;


end;

procedure TRangedEdit.DoExit();
var
  n : Double;
  c : String;
  lReset : Boolean;
begin
  lReset:=False;
  try
    n := StrToFloat(Text);
    if (n<F_MinValue) or (n>F_MaxValue) Then begin
      c := format('Valid Range is [%f] - [%f].'+ #13 +F_ErrorMessage, [F_MinValue,F_MaxValue]);
      MessageDlg( c, mtError, [mbOK], 0 );
      lReset:=True;
    end;
  except
    lReset:=True;
    MessageDlg('Not a valid numeric input'+ #13 +F_ErrorMessage, mtError, [mbOK], 0 );
  end;

  if lReset Then Begin
    try
      Text := FloatToStr(F_OriginalValue);
      SetFocus();
    except
    end;
  end;

end;

procedure TRangedEdit.KeyPress(var Key: Char);
begin
  //----- allow "+-.0123456789" and special chars
  if not( (Key<' ') or (Key='.') or (Key='-') or (Key='+') or (Key='E') or (Key='e')
              or ((Key>='0') and (Key<='9')) or (Key>'') ) then
    Key:=char(0);
end;

procedure Register;
begin
  RegisterComponents('QUB', [TRangedEdit]);
end;

procedure TRangedEdit.Setup( parent : TWinControl; iLft, iTop, iWid, iHgt : Integer;
            iTabOrder : Integer; cText : String; nMin, nMax : Double );
begin
  SetParent(parent);
  Left     := iLft;
  Top      := iTop;
  Width    := iWid;
  Height   := iHgt;
  TabOrder := iTabOrder;
  Text     := cText;
  F_MinValue := nMin;
  F_MaxValue := nMax;
end;


end.

