{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
{
  This unit encapsulates the hardware registry interface.
  This could be class wrapped once completed.
  Keep changes compatible with writing to a registry or other persist.
  Note : Previously Channel writes executed every time QUB exits.
    Inform users - now it occurs at form close ?
  Note : DeviceListReadRegistry Ignores Gain 0..Gain 7 from older QUB

}

unit RegistryInterface;


interface


uses


Windows, Messages, SysUtils, Registry, Classes, Forms,

QUB_Globals,
QUB_Acquisition_Hardware;


//----- General pupose
Function ReadRegistryInteger( reg : TRegistry; cName : String; iDefault : Integer ) : Integer ;
Function ReadRegistryIntBool( reg : TRegistry; cName : String; lDefault : Boolean ) : Integer ;
Function ReadRegistryBoolean( reg : TRegistry; cName : String; lDefault : Boolean ) : Boolean ;
Function ReadRegistryString( reg : TRegistry; cName : String; cDefault : String ) : String ;
Function ReadRegistryStringFloat( reg : TRegistry; cName : String; nDefault : double ) : double ;
Function ReadRegistryFloat( reg : TRegistry; cName : String; nDefault : double ) : double ;

// open-as-needed for the phaseout
Function ReadXRegistryInteger( var reg : TRegistry; const kName, cName : String; var iResult : Integer ) : Boolean ;
Function ReadXRegistryIntBool( var reg : TRegistry; const kName, cName : String; var iResult : Integer ) : Boolean ;
Function ReadXRegistryBoolean( var reg : TRegistry; const kName, cName : String; var bResult : Boolean ) : Boolean ;
Function ReadXRegistryString( var reg : TRegistry; const kName, cName : String; var cResult : String ) : Boolean ;
Function ReadXRegistryStringFloat( var reg : TRegistry; const kName, cName : String; var nResult : double ) : Boolean ;
Function ReadXRegistryFloat( var reg : TRegistry; const kName, cName : String; var nResult : double ) : Boolean ;

Function WriteXRegistryString( var reg : TRegistry; const kName, cName : String; cVal : String ) : Boolean ;

  //----- Generic window settings (for future use)
  //Procedure WindowPositionReadRegistry( cKey : String; var plVisible : Boolean; var piLeft : Integer; var piTop : Integer );
  //Procedure WindowPositionWriteRegistry( cKey : String; lVisible : Boolean; iLeft : Integer; iTop : Integer );

  //----- Viwports default settings - write
  Function ViewportWriteRegistry( reg : TRegistry; visible : Boolean; rct : TRect ) : Boolean;

  //----- Hardware settings
  Procedure HardwareReadRegistry( ) ;
  Procedure HardwareWriteRegistry( );

  procedure DeviceListReadRegistry(oHdw : TQUB_Acquisition_Hardware);
  Function DeviceListWriteRegistry(oHdw : TQUB_Acquisition_Hardware):Boolean;

  //----- Channel Settings
  Procedure ChannelsReadRegistry();
    Procedure AChannelReadRegistry( reg : TRegistry; channel : TQUB_Acquisition_AChannel );
    Procedure DChannelReadRegistry( reg : TRegistry; channel : TQUB_Acquisition_DChannel );
      Procedure ChannelReadRegistry( reg : TRegistry; channel : TQUB_Acquisition_Channel );
  Procedure ChannelsWriteRegistry();
    Procedure AChannelWriteRegistry( reg : TRegistry; channel : TQUB_Acquisition_AChannel );
    Procedure DChannelWriteRegistry( reg : TRegistry; channel : TQUB_Acquisition_DChannel );
      Procedure ChannelWriteRegistry( reg : TRegistry; channel : TQUB_Acquisition_Channel );

  //----- Read user interface count / text / name from registry
  Function UserInterfaceCountReadRegistry():Integer;
  Procedure UserInterfaceCountWriteRegistry(iInterfaces:Integer);
  Procedure UserInterfaceReadRegistry( iInterface : Integer; var cText : String; var cName : String );
  Procedure UserInterfaceWriteRegistry( iInterface : Integer; cText, cName : String );

  //----- Model Properties
  Procedure ModelRateReadRegistry( var FwdRate, BckRate : TRate);
  Procedure ModelRateWriteRegistry( FwdRate, BckRate : TRate);


implementation


uses


QUB_QFS_Channels;


//---------------------------------------------------------------------------------------------
//----- Read Integer / Boolean / String / Float value from Registry
Function ReadRegistryInteger( reg : TRegistry; cName : String; iDefault : Integer ) : Integer ; 
  begin
    if reg.ValueExists(cName) then
      result := reg.ReadInteger(cName)
    else
      result := iDefault;
  end;

Function ReadRegistryBoolean( reg : TRegistry; cName : String; lDefault : Boolean ) : Boolean ;
  begin
    if reg.ValueExists(cName) then
      result := reg.ReadBool(cName)
    else
      result := lDefault;
  end;

Function ReadRegistryIntBool( reg : TRegistry; cName : String; lDefault : Boolean ) : Integer ;
  begin
    if reg.ValueExists(cName) then
      result := Integer(reg.ReadBool(cName))
    else
      result := Integer(lDefault);
  end;

Function ReadRegistryString( reg : TRegistry; cName : String; cDefault : String ) : String ;
  begin
    if reg.ValueExists(cName) then
      result := reg.ReadString(cName)
    else
      result := cDefault;
  end;

Function ReadRegistryFloat( reg : TRegistry; cName : String; nDefault : double ) : double ;
  begin
    if reg.ValueExists(cName) then
      result := reg.ReadFloat(cName)
    else
      result := nDefault;
  end;

Function ReadRegistryStringFloat( reg : TRegistry; cName : String; nDefault : double ) : double ;
  begin
    if reg.ValueExists(cName) then
      result := StrToFloat(reg.ReadString(cName))
    else
      result := nDefault;
  end;


//---------------------------------------------------------------------------------------------
function EnsureRegistry(var reg :TRegistry; const kName :String): Boolean;
begin
  Result := True;
  if reg = nil then begin
    reg := TRegistry.Create;
  end;
  if reg.CurrentPath <> kName then begin
    reg.CloseKey;
    Result := reg.OpenKey(kName, True);
  end;
  if not Result then begin
    reg.Free;
    reg := nil;
  end;
end;

//----- Read Integer / Boolean / String / Float value from Registry
Function ReadXRegistryInteger( var reg : TRegistry; const kName, cName : String; var iResult : Integer ) : Boolean ;
  begin
    if EnsureRegistry(reg, kName) and reg.ValueExists(cName) then begin
      iResult := reg.ReadInteger(cName);
      Result  := True;
    end else begin
      Result  := False;
    end;
  end;

Function ReadXRegistryBoolean( var reg : TRegistry; const kName, cName : String; var bResult : Boolean ) : Boolean ;
  begin
    if EnsureRegistry(reg, kName) and reg.ValueExists(cName) then begin
      bResult := reg.ReadBool(cName);
      Result  := True;
    end else begin
      Result  := False;
    end;
  end;

Function ReadXRegistryIntBool( var reg : TRegistry; const kName, cName : String; var iResult : Integer ) : Boolean ;
  begin
    if EnsureRegistry(reg, kName) and reg.ValueExists(cName) then begin
      iResult := Integer(reg.ReadBool(cName));
      Result  := True;
    end else begin
      Result  := False;
    end;
  end;

Function ReadXRegistryString( var reg : TRegistry; const kName, cName : String; var cResult : String ) : Boolean ;
  begin
    if EnsureRegistry(reg, kName) and reg.ValueExists(cName) then begin
      cResult := reg.ReadString(cName);
      Result  := True;
    end else begin
      Result  := False;
    end;
  end;

Function ReadXRegistryFloat( var reg : TRegistry; const kName, cName : String; var nResult : double ) : Boolean ;
  begin
    if EnsureRegistry(reg, kName) and reg.ValueExists(cName) then begin
      nResult := reg.ReadFloat(cName);
      Result  := True;
    end else begin
      Result  := False;
    end;
  end;

Function ReadXRegistryStringFloat( var reg : TRegistry; const kName, cName : String; var nResult : double ) : Boolean ;
  begin
    if EnsureRegistry(reg, kName) and reg.ValueExists(cName) then begin
      nResult := StrToFloat(reg.ReadString(cName));
      Result  := True;
    end else begin
      Result  := False;
    end;
  end;



Function WriteXRegistryString( var reg : TRegistry; const kName, cName : String; cVal : String ) : Boolean ;
  begin
    if EnsureRegistry(reg, kName) and reg.ValueExists(cName) then begin
      reg.WriteString(cName, cVal);
      Result  := True;
    end else begin
      Result  := False;
    end;
  end;



//---------------------------------------------------------------------------------------------
//----- Read list of hardware settings from Registry.
Procedure HardwareReadRegistry( );
  var
    Reg : TRegistry;
    iDev : Integer;
    cDev : String;
    oHdw : TQUB_Acquisition_Hardware;
  begin
    iDev := 1;
    Reg := nil;
    try
      Reg := TRegistry.Create;
      if Reg.OpenKey('Software\QUB\Acquisition Hardware View\Hardware', False) then begin
        while Reg.ValueExists('Hardware ' + IntToStr(iDev)) do begin
          cDev := Reg.ReadString('Hardware ' + IntToStr(iDev));
          oHdw := TQUB_Acquisition_Hardware.Create(cDev);
          QUBHardwareList.AddObject(cDev, oHdw );
          if oHdw.Loaded then DeviceListReadRegistry(oHdw);
          Inc(iDev);
        end;
      end;
    finally
      Reg.Free;
    end;
  end;

//----- Write list of hardware settings to registry
Procedure HardwareWriteRegistry( );
  var
    Reg : TRegistry;
    iHdw : Integer;
    iDev : Integer;
  begin
    iDev := 0;
    Reg := nil;
    try
      Reg := TRegistry.Create;
      Reg.DeleteKey('Software\QUB\Acquisition Hardware View\Hardware');
      if Reg.OpenKey('Software\QUB\Acquisition Hardware View\Hardware', True) then
        for iHdw := 0 to QUBHardwareList.Count-1 do begin
          if DeviceListWriteRegistry(TQUB_Acquisition_Hardware(QUBHardwareList.Objects[iHdw])) Then Begin
            Inc(iDev);
            Reg.WriteString('Hardware ' + IntToStr(iDev), QUBHardwareList[iHdw] );
          end;
        end;
      Reg.CloseKey();
    finally
      Reg.Free;
    end;
  end;

//------------------------------------------------- Device Lists
procedure DeviceListReadRegistry(oHdw : TQUB_Acquisition_Hardware);
var
  Reg  : TRegistry;
  iDev : Integer;
  cKey : String;
begin
  Reg := TRegistry.Create();
  try
    //----- Input device settings
    cKey := 'Software\QUB\Hardware\Library\' + oHdw.LibraryName + '\Input\';
    for iDev := 0 to oHdw.IDeviceCount - 1 do
      if Reg.OpenKey(cKey + oHdw.IDeviceName[iDev], False) then Begin
        //----- read registry
        oHdw.InputDeviceParams[iDev].m_Bits := ReadRegistryInteger(Reg,'Bits',oHdw.GetLibIBits(iDev));
        oHdw.InputDeviceParams[iDev].m_Differential := ReadRegistryBoolean( Reg, 'Differential', (oHdw.GetLibIDifferential(iDev)<>0) );
        oHdw.InputDeviceParams[iDev].m_Range := ReadRegistryFloat(Reg,'Range',oHdw.GetLibIRange(iDev));
        oHdw.InputDeviceParams[iDev].m_Loaded := True;
        Reg.CloseKey();
      end;

    //----- Output device settings
    cKey := 'Software\QUB\Hardware\Library\' + oHdw.LibraryName + '\Output\' ;
    for iDev := 0 to oHdw.ODeviceCount - 1 do
      if Reg.OpenKey(cKey + oHdw.ODeviceName[iDev], False) then Begin
        //----- read registry
        oHdw.OutputDeviceParams[iDev].m_Bits := ReadRegistryInteger(Reg,'Bits',oHdw.GetLibOBits(iDev));
        oHdw.OutputDeviceParams[iDev].m_Differential := ReadRegistryBoolean( Reg, 'Differential', (oHdw.GetLibODifferential(iDev)<>0) );
        oHdw.OutputDeviceParams[iDev].m_Range := ReadRegistryFloat(Reg,'Range',oHdw.GetLibORange(iDev));
        oHdw.OutputDeviceParams[iDev].m_Loaded := True;
        Reg.CloseKey();
      end;

  finally
    Reg.CloseKey();
    Reg.Free();

  end;
end;

//----- Write list of HDW, return true if any active devices
Function DeviceListWriteRegistry(oHdw : TQUB_Acquisition_Hardware):Boolean;
var
  Reg  : TRegistry;
  iDev : Integer;
  cKey : String;
begin
  Result := False;
  Reg := TRegistry.Create();
  try
    Reg.DeleteKey('Software\QUB\Hardware\Library\' + oHdw.LibraryName );

    //----- Input device settings
    cKey := 'Software\QUB\Hardware\Library\' + oHdw.LibraryName + '\Input\';
    for iDev := 0 to oHdw.IDeviceCount - 1 do
      if oHdw.InputDeviceParams[iDev].m_Loaded Then begin
        Result := True;
        if Reg.OpenKey(cKey + oHdw.IDeviceName[iDev], True) then begin
          Reg.WriteInteger('Bits', oHdw.InputDeviceParams[iDev].m_Bits);
          Reg.WriteBool('Differential', oHdw.InputDeviceParams[iDev].m_Differential);
          Reg.WriteFloat('Range', oHdw.InputDeviceParams[iDev].m_Range );
          Reg.CloseKey();
        end;
      end;

    //----- Output device settings
    cKey := 'Software\QUB\Hardware\Library\' + oHdw.LibraryName + '\Output\' ;
    for iDev := 0 to oHdw.ODeviceCount - 1 do
      if oHdw.OutputDeviceParams[iDev].m_Loaded Then Begin
        Result := True;
        if Reg.OpenKey(cKey + oHdw.ODeviceName[iDev], True) then Begin
          Reg.WriteInteger('Bits', oHdw.OutputDeviceParams[iDev].m_Bits);
          Reg.WriteBool('Differential', oHdw.OutputDeviceParams[iDev].m_Differential);
          Reg.WriteFloat('Range', oHdw.OutputDeviceParams[iDev].m_Range );
          Reg.CloseKey();
        end;
      end;

  finally
    Reg.CloseKey();
    Reg.Free();

  end;
end;

{
//---------------------------------------------------------------------------------------------
//----- Read window position from Registry
Procedure WindowPositionReadRegistry( cKey : String; var plVisible : Boolean; var piLeft : Integer; var piTop : Integer );
  var
    Reg :TRegistry;
  begin
    Reg := TRegistry.Create;
    try
      if Reg.OpenKey(cKey, True) then begin
        plVisible := ReadRegistryBoolean( Reg, 'Visible', False);
        piLeft := ReadRegistryInteger( Reg, 'Left', 100 );
        piTop := ReadRegistryInteger( Reg, 'Top', 100 );
      end;
      Reg.CloseKey();
    finally
      Reg.Free;
    end;
  end;

//----- Write window position to Registry
Procedure WindowPositionWriteRegistry( cKey : String; lVisible : Boolean; iLeft : Integer; iTop : Integer );
  var
    Reg :TRegistry;
  begin
    Reg := TRegistry.Create;
    try
      if Reg.OpenKey(cKey, True) then begin
          Reg.WriteBool('Visible', lVisible);
          Reg.WriteInteger('Left', iLeft);
          Reg.WriteInteger('Top', iTop);
      end;
      Reg.CloseKey();
    finally
      Reg.Free;
    end;
  end;
}

//----- Write viewport parameters to registry
Function ViewportWriteRegistry( reg : TRegistry; visible : Boolean; rct : TRect ) : Boolean;
begin
  //----- Write view parameters to registry
  try
    reg.WriteBool('Float Visible', visible);
    reg.WriteBool('Dock Visible', visible);

    reg.WriteBool('UnDocked', False);

    reg.WriteInteger('Float Window State', Integer(wsNormal));
    reg.WriteInteger('Dock Window State', Integer(wsNormal));

    reg.WriteBinaryData('Float Bounds',  rct, SizeOf(rct));
    reg.WriteBinaryData('Dock Bounds',   rct, SizeOf(rct));
    reg.WriteBinaryData('UnDock Bounds', rct, SizeOf(rct));
  Finally
  end;
  Result := True;
end;

//---------------------------------------------------------------------------------------------
//-----Read/Write Channel config from registry (CHANNEL,DEVICE NAME, UPDATE)
Procedure ChannelReadRegistry( reg : TRegistry; channel : TQUB_Acquisition_Channel );
  begin
    channel.Channel := ReadRegistryInteger( reg, 'Channel', 0);
    channel.DeviceName := ReadRegistryString( reg, 'Device Name', '' );
    channel.Update := ReadRegistryBoolean( reg, 'Update', False );
  end;

Procedure ChannelWriteRegistry( reg : TRegistry; channel : TQUB_Acquisition_Channel );
  begin
    reg.WriteString('Name', channel.Name);
    reg.WriteInteger('Channel', channel.Channel);
    reg.WriteString('Device Name', channel.DeviceName);
    reg.WriteBool('Update', channel.Update);
  end;

//----- Read Analog channel properties (VALUE, UNITS, SCALING, GAIN, INCREMENT)
Procedure AChannelReadRegistry( reg : TRegistry; channel : TQUB_Acquisition_AChannel );
  begin
    ChannelReadRegistry(reg, TQUB_Acquisition_Channel(channel) );
    channel.Value := ReadRegistryFloat( reg, 'Value', 0.0 );
    channel.Units := ReadRegistryString( reg, 'Units', '' );
    channel.Scaling := ReadRegistryFloat( reg, 'Scaling', 0.0 );
    channel.Gain := ReadRegistryFloat( reg, 'Gain', 0.0 );
  end;

//----- Write Analog channel properties
Procedure AChannelWriteRegistry( reg : TRegistry; channel : TQUB_Acquisition_AChannel );
  begin
    ChannelWriteRegistry(reg,channel);
    reg.WriteString('Units', channel.Units);
    reg.WriteFloat('Scaling', channel.Scaling);
    reg.WriteFloat('Gain', channel.Gain);
    reg.WriteFloat('Value', channel.Value);
  end;

//----- Read Digital channel properties (VALUE)
Procedure DChannelReadRegistry( reg : TRegistry; channel : TQUB_Acquisition_DChannel );
  begin
    ChannelReadRegistry(reg, TQUB_Acquisition_Channel(channel) );
    channel.Value := ReadRegistryBoolean( reg, 'Value', False );
  end;

Procedure DChannelWriteRegistry( reg : TRegistry; channel : TQUB_Acquisition_DChannel );
  begin
    ChannelWriteRegistry(reg,channel);
    reg.WriteBool('Value', channel.Value);
  end;

//----- Read the lists of AI,AO,DI,DO channels From the registry.
Procedure ChannelsReadRegistry();
  var
  	Reg  : TRegistry;
  	AICh : TQUB_Acquisition_AIChannel;
    AOCh : TQUB_Acquisition_AOChannel;
    DICh : TQUB_Acquisition_DIChannel;
    DOCh : TQUB_Acquisition_DOChannel;
    iCh  : Integer;
    cKey : String;
  begin
    cKey := 'Software\QUB\Acquisition Channels View\';

    Reg := TRegistry.Create;
    try
      //----------- READ channels from Registry
    	//-----AIChannels Read
      iCh := 1;
      while reg.OpenKey(cKey+'Analog Input\AIChannel ' + IntToStr(iCh), False) do begin
        if reg.ValueExists('Name') then begin
          AICh := TQUB_Acquisition_AIChannel.Create(reg.ReadString('Name')); // side effect - adds to array of channels
          AChannelReadRegistry( reg , AICh );
          AICh.Device := QUBAcqIDevices.FindDevice(AICh.DeviceName);
        end;
        reg.CloseKey();
        Inc(iCh);
      end;

      //-----AOChannels Read
			iCh := 1;
			while reg.OpenKey(cKey+'Analog Output\AOChannel ' + IntToStr(iCh), False) do begin
				if reg.ValueExists('Name') then begin
					AOCh := TQUB_Acquisition_AOChannel.Create(reg.ReadString('Name'));
          AChannelReadRegistry( reg, AOCh );
          AOCh.Device := QUBAcqODevices.FindDevice(AOCh.DeviceName);
          AOCh.Increment := ReadRegistryFloat( reg, 'Increment', 0.0 );
				end;
				reg.CloseKey();
				Inc(iCh);
			end;

      //----- DI Channels Read
      iCh := 1;
      while reg.OpenKey(cKey+'Digital Input\DIChannel ' + IntToStr(iCh), False) do begin
			  if reg.ValueExists('Name') then begin
          DICh := TQUB_Acquisition_DIChannel.Create(reg.ReadString('Name'));
          DChannelReadRegistry( reg, DICH );
          DICh.Device := QUBAcqIDevices.FindDevice(DICh.DeviceName);
        end;
				reg.CloseKey();
        Inc(iCh);
      end;

      //----- DO Channels Read
      iCh := 1;
      while reg.OpenKey(cKey+'Digital Output\DOChannel ' + IntToStr(iCh), False) do begin
			  if reg.ValueExists('Name') then begin
          DOCh := TQUB_Acquisition_DOChannel.Create(reg.ReadString('Name'));
          DChannelReadRegistry( reg, DOCH );
          DOCh.Device := QUBAcqODevices.FindDevice(DOCh.DeviceName);
        end;
				reg.CloseKey();
        Inc(iCh);
      end;
		finally
      reg.CloseKey();
			Reg.Free();
		end;
  end;

//----- Write the lists of AI,AO,DI,DO channels to the registry.
Procedure ChannelsWriteRegistry();
  var
  	Reg  : TRegistry;
    AOCh : TQUB_Acquisition_AOChannel;
    iCh  : Integer;
    cKey : String;
  begin
    cKey := 'Software\QUB\Acquisition Channels View\';
    Reg := TRegistry.Create;

    try
      //----------Write Channels to Registry
    	//-----AIChannels Write
      reg.DeleteKey(cKey+'Analog Input');
      for iCh := 0 to QUBAcqAIChannels.Count - 1 do begin
        if reg.OpenKey(cKey+'Analog Input\AIChannel ' + IntToStr(iCh + 1), True) then
          AChannelWriteRegistry( reg, TQUB_Acquisition_AIChannel(QUBAcqAIChannels.Objects[iCh]) );
				reg.CloseKey();
      end;

      //-----AOChannels Write - Note : Slightly different than others - inclusion of 1 extra value
      reg.DeleteKey(cKey+'Analog Output');
			for iCh := 0 to QUBAcqAOChannels.Count - 1 do begin
				if reg.OpenKey(cKey+'Analog Output\AOChannel ' + IntToStr(iCh + 1), True) then begin
          AOCh := TQUB_Acquisition_AOChannel(QUBAcqAOChannels.Objects[iCh]);
          AChannelWriteRegistry( reg, AOCh);
          reg.WriteFloat('Increment', AOCh.Increment);
        end;
				reg.CloseKey();
			end;

      //----- DI Channels Write
      reg.DeleteKey(cKey+'Digital Input');
      for iCh := 0 to QUBAcqDIChannels.Count - 1 do begin
        if reg.OpenKey(cKey+'Digital Input\DIChannel ' + IntToStr(iCh + 1), True) then
          DChannelWriteRegistry( reg, TQUB_Acquisition_DIChannel(QUBAcqDIChannels.Objects[iCh]));
				reg.CloseKey();
			end;

      //----- DO Channels Write
      reg.DeleteKey(cKey+'Digital Output');
      for iCh := 0 to QUBAcqDOChannels.Count - 1 do begin
        if reg.OpenKey(cKey+'Digital Output\DOChannel ' + IntToStr(iCh + 1), True) then
          DChannelWriteRegistry(reg, TQUB_Acquisition_DOChannel(QUBAcqDOChannels.Objects[iCh]));
				reg.CloseKey();
			end;

		finally
      reg.CloseKey();
			Reg.Free();
		end;
  end;

//----- Read user interface count
Function UserInterfaceCountReadRegistry():Integer;
var
  	Reg  : TRegistry;
begin
    Reg := TRegistry.Create;
    try
      if reg.OpenKey('Software\QUB\Main Form', False) Then
        Result := ReadRegistryInteger( Reg, 'UserInterfaceCount', 0 )
      else
        Result := 0;
      Reg.CloseKey();
    Finally
      Reg.Free();
    end;
end;

//----- Write user interface count
Procedure UserInterfaceCountWriteRegistry(iInterfaces:Integer);
var
  	Reg  : TRegistry;
begin
    Reg := TRegistry.Create;
    try
      if reg.OpenKey('Software\QUB\Main Form', False) Then
        Reg.WriteInteger('UserInterfaceCount', iInterfaces );
      Reg.CloseKey();
    Finally
      Reg.Free();
    end;
end;

//----- Read user interface button text and name
Procedure UserInterfaceReadRegistry( iInterface : Integer; var cText : String; var cName : String );
var
  	Reg  : TRegistry;
begin
    Reg := TRegistry.Create;
    try
      if reg.OpenKey('Software\QUB\Interface User' + IntToStr(iInterface), False) Then Begin
        cText := ReadRegistryString( Reg, 'ButtonText', IntToStr(iInterface) );
        cName := ReadRegistryString( Reg, 'ButtonName', IntToStr(iInterface) );
      end else Begin
        cText := IntToStr(iInterface);
        cName := '';
      end;
      Reg.CloseKey();
    Finally
      Reg.Free();
    end;
end;

//----- Write user interface button text and name
Procedure UserInterfaceWriteRegistry( iInterface : Integer; cText, cName : String );
var
  	Reg  : TRegistry;
begin
    Reg := TRegistry.Create;
    try
      if reg.OpenKey('Software\QUB\Interface User' + IntToStr(iInterface), True) Then Begin
        Reg.WriteString('ButtonText', cText );
        Reg.WriteString('ButtonName', cName );
      end;
      Reg.CloseKey();
    Finally
      Reg.Free();
    end;
end;

Procedure ModelRateReadRegistry( var FwdRate, BckRate : TRate );
var
  Reg : TRegistry;
begin
    Reg := TRegistry.Create;
    try
        if Reg.OpenKey('Software\QUB\Visual Model\Rate', False) then begin
            FwdRate.m_K0 := ReadRegistryFloat( Reg, 'Rate K0', QUB_QFS_QMF_DEF_K0 );
            FwdRate.m_K1 := ReadRegistryFloat( Reg, 'Rate K1', QUB_QFS_QMF_DEF_K1 );
            FwdRate.m_P := ReadRegistryBoolean( Reg, 'Rate P', QUB_QFS_QMF_DEF_P );
            FwdRate.m_Q := ReadRegistryBoolean( Reg, 'Rate Q', QUB_QFS_QMF_DEF_Q );
            FwdRate.m_PName := ReadRegistryString( Reg, 'Rate PName', QUB_QFS_QMF_DEF_PNAME );
            FwdRate.m_QName := ReadRegistryString( Reg, 'Rate QName', QUB_QFS_QMF_DEF_QNAME );
            FwdRate.m_RateFormat := ReadRegistryString( Reg, 'Rate Format', QUB_QFS_QMF_DEF_RATEFORMAT );

            BckRate.m_K0 := ReadRegistryFloat( Reg, 'Rate K0 b',        FwdRate.m_K0);
            BckRate.m_K1 := ReadRegistryFloat( Reg, 'Rate K1 b',        FwdRate.m_K1);
            BckRate.m_P := ReadRegistryBoolean( Reg, 'Rate P b',        FwdRate.m_P);
            BckRate.m_Q := ReadRegistryBoolean( Reg, 'Rate Q b',        FwdRate.m_Q);
            BckRate.m_PName := ReadRegistryString( Reg, 'Rate PName b' , FwdRate.m_PName);
            BckRate.m_QName := ReadRegistryString( Reg, 'Rate QName b', FwdRate.m_QName);
            BckRate.m_RateFormat := ReadRegistryString( Reg, 'Rate Format b', FwdRate.m_RateFormat);
            Reg.CloseKey;
        end;
    finally
        Reg.Free;
    end;
end;

Procedure ModelRateWriteRegistry( FwdRate, BckRate : TRate );
var
  Reg : TRegistry;
begin
    Reg := TRegistry.Create;
    try
        if Reg.OpenKey('Software\QUB\Visual Model\Rate', True) then begin
          Reg.WriteFloat('Rate K0', FwdRate.m_K0);
          Reg.WriteFloat('Rate K1', FwdRate.m_K1);
          Reg.WriteBool('Rate P', FwdRate.m_P);
          Reg.WriteBool('Rate Q', FwdRate.m_Q);
          Reg.WriteString('Rate PName', FwdRate.m_PName);
          Reg.WriteString('Rate QName', FwdRate.m_QName);

          Reg.WriteFloat('Rate K0 b', BckRate.m_K0);
          Reg.WriteFloat('Rate K1 b', BckRate.m_K1);
          Reg.WriteBool('Rate P b', BckRate.m_P);
          Reg.WriteBool('Rate Q b', BckRate.m_Q);
          Reg.WriteString('Rate PName b', BckRate.m_PName);
          Reg.WriteString('Rate QName b', BckRate.m_QName);
        end;
    finally
        Reg.Free;
    end;
end;

//-----------------
end.

