unit nidaqmx_types;


interface


uses


AllocTypes;


type


TtaskHandle  = LongWord;
TPtaskHandle = ^TtaskHandle;


const


DAQmx_Val_Volts                        = 10348;

DAQmx_Val_GroupByChannel               = 0;// Group by Channel
DAQmx_Val_GroupByScanNumber            = 1;// Group by Scan Number

//*** Value for the Terminal Config parameter of DAQmxCreateAIVoltageChan, DAQmxCreateAICurrentChan and DAQmxCreateAIVoltageChanWithExcit ***
DAQmx_Val_Cfg_Default                  = -1;// Default

//*** Value set for the ActiveEdge parameter of DAQmxCfgSampClkTiming ***
DAQmx_Val_Rising                       = 10280; // Rising
DAQmx_Val_Falling                      = 10171; // Falling

//*** Values for DAQmx_SampQuant_SampMode ***
//*** Value set AcquisitionType ***
DAQmx_Val_FiniteSamps                  = 10178; // Finite Samples
DAQmx_Val_ContSamps                    = 10123; // Continuous Samples
DAQmx_Val_HWTimedSinglePoint           = 12522; // Hardware Timed Single Point

DAQmx_Val_SampClk                      = 10388; // Use a sample clock to determine how often to acquire or generate samples.
DAQmx_Val_BurstHandshake               = 12548; // Determine sample timing using burst handshaking between the device and a peripheral device.
DAQmx_Val_Handshake                    = 10389; // Determine sample timing by using digital handshaking between the device and a peripheral device.
DAQmx_Val_Implicit                     = 10451; // Configure only the duration of the task.
DAQmx_Val_OnDemand                     = 10390; // Acquire or generate a sample on each read or write operation.
DAQmx_Val_ChangeDetection              = 12504; // Acquire samples when a change occurs in the state of one or more digital input lines. The lines must be contained within a digital input channel.
DAQmx_Val_PipelinedSampClk             = 14668; // Device acquires or generates samples on each sample clock edge, but does not respond to certain triggers
                                                //   until a few sample clock edges later. Pipelining allows higher data transfer rates at the cost of increased
                                                //   trigger response latency. Refer to the device documentation for information about which triggers pipelining
                                                //   affects. This timing type allows handshaking with some devices using the Pause trigger, the Ready for Transfer event,
                                                //   or the Data Active event. Refer to the device documentation for more information.

DAQmx_Val_WaitForInterrupt             = 12523; // Check for Sample Clock pulses when the system receives an interrupt service request. This mode is the most CPU efficient, but results in lower possible sampling rates.
DAQmx_Val_Poll                         = 12524; // Repeatedly check for Sample Clock pulses as fast as possible. This mode allows for the highest sampling rates at the expense of CPU efficiency.
DAQmx_Val_Yield                        = 12525; // Repeatedly check for Sample Clock pulses, but yield control to other threads after each check. This mode offers a balance between sampling rate and CPU efficiency.

DAQmx_Val_FirstSample                  = 10424; // Start reading samples relative to the first sample acquired.
DAQmx_Val_CurrReadPos                  = 10425; // Start reading samples relative to the last sample returned by the previous read. For the first read operation, this position is the first sample acquired or the first pretrigger sample if you configured a reference trigger for the task.
DAQmx_Val_RefTrig                      = 10426; // Start reading samples relative to the first sample after the reference trigger occurred.
DAQmx_Val_FirstPretrigSamp             = 10427; // Start reading samples relative to the first pretrigger sample. You specify the number of pretrigger samples to acquire when you configure a reference trigger.
DAQmx_Val_MostRecentSamp               = 10428; // Start reading samples relative to the next sample acquired. For example, use this value and set Offset to -1 to read the last sample acquired.


DAQmx_Val_Task_Start                   = 0;   // Start
DAQmx_Val_Task_Stop                    = 1;   // Stop
DAQmx_Val_Task_Verify                  = 2;   // Verify
DAQmx_Val_Task_Commit                  = 3;   // Commit
DAQmx_Val_Task_Reserve                 = 4;   // Reserve
DAQmx_Val_Task_Unreserve               = 5;   // Unreserve
DAQmx_Val_Task_Abort                   = 6;   // Abort


type


//Task
TDAQmxCreateTask                       = function (const taskName :PChar; var taskHandle :TtaskHandle) :LongInt; stdcall;
TDAQmxClearTask                        = function (taskHandle :TtaskHandle) :LongInt; stdcall;
TDAQmxStartTask                        = function (taskHandle :TtaskHandle) :LongInt; stdcall;
TDAQmxStopTask                         = function (taskHandle :TtaskHandle) :LongInt; stdcall;
TDAQmxTaskControl                      = function (taskHandle :TtaskHandle; action :LongInt) :LongInt; stdcall;

//Channels
TDAQmxCreateAIVoltageChan              = function (taskHandle :TtaskHandle; const physicalChannel, nameToAssignToChannel :PChar; terminalConfig :LongInt; minVal, maxVal :Double; units :LongInt; const customScaleName :PChar) :LongInt; stdcall;
TDAQmxCreateAOVoltageChan              = function (taskHandle :TtaskHandle; const physicalChannel, nameToAssignToChannel :PChar; minVal, maxVal :Double; units :LongInt; const customScaleName :PChar) :LongInt; stdcall;

TDAQmxGetAIResolutionUnits             = function (taskHandle :TtaskHandle; const channel :PChar; var data :LongInt) :LongInt; stdcall;
TDAQmxGetAIResolution                  = function (taskHandle :TtaskHandle; const channel :PChar; var data :Double) :LongInt; stdcall;
TDAQmxGetAOResolutionUnits             = function (taskHandle :TtaskHandle; const channel :PChar; var data :LongInt) :LongInt; stdcall;
TDAQmxGetAOResolution                  = function (taskHandle :TtaskHandle; const channel :PChar; var data :Double) :LongInt; stdcall;
TDAQmxGetAIRngHigh                     = function (taskHandle :TtaskHandle; const channel :PChar; var data :Double) :LongInt; stdcall;
TDAQmxGetAIRngLow                      = function (taskHandle :TtaskHandle; const channel :PChar; var data :Double) :LongInt; stdcall;
TDAQmxGetAODACRngHigh                  = function (taskHandle :TtaskHandle; const channel :PChar; var data :Double) :LongInt; stdcall;
TDAQmxGetAODACRngLow                   = function (taskHandle :TtaskHandle; const channel :PChar; var data :Double) :LongInt; stdcall;

//Error
TDAQmxGetErrorString                   = function (errorCode :LongInt; errorString :PChar; bufferSize :LongWord) :LongInt; stdcall;
TDAQmxGetExtendedErrorInfo             = function (errorString :PChar; bufferSize :LongWord) :LongInt; stdcall;

//Write
TDAQmxWriteBinaryI16                   = function (taskHandle :TtaskHandle; numSampsPerChan :LongInt; autoStart :LongBool; timeout :Double; dataLayout :LongInt; writeArray :TPASmallInt; var sampsPerChanWritten :LongInt; Reserved :Pointer) :LongInt; stdcall;
TDAQmxWriteRaw                         = function (taskHandle :TtaskHandle; numSamps :LongInt; autoStart :LongBool; timeout :Double; writeArray :Pointer; var sampsPerChanWritten :LongInt; Reserved :Pointer) :LongInt; stdcall;

//Read
TDAQmxReadAnalogF64                    = function (taskHandle :TtaskHandle; numSampsPerChan :LongInt; timeout :Double; fillMode :LongInt; readArray :TPADouble; arraySizeInSamps :LongWord; var sampsPerChanRead :LongInt; Reserved :Pointer) :LongInt; stdcall;
TDAQmxReadBinaryI16                    = function (taskHandle :TtaskHandle; numSampsPerChan :LongInt; timeout :Double; fillMode :LongInt; readArray :TPASmallInt; arraySizeInSamps :LongWord; var sampsPerChanRead :LongInt; Reserved :Pointer) :LongInt; stdcall;
TDAQmxReadRaw                          = function (taskHandle :TtaskHandle; numSampsPerChan :LongInt; timeout :Double; readArray :Pointer; arraySizeInBytes :LongWord; var sampsRead :LongInt; var numBytesPerSamp :LongInt; Reserved :Pointer) :LongInt; stdcall;

//Timing
TDAQmxCfgPipelinedSampClkTiming        = function (taskHandle :TtaskHandle; const source :PChar; rate :Double; activeEdge :LongInt; sampleMode :LongInt; sampsPerChan :{u}Int64) :LongInt; stdcall;
TDAQmxCfgSampClkTiming                 = function (taskHandle :TtaskHandle; const source :PChar; rate :Double; activeEdge :LongInt; sampleMode :LongInt; sampsPerChan :{u}Int64) :LongInt; stdcall;

TDAQmxGetSampQuantSampMode             = function (taskHandle :TtaskHandle; var data :LongInt) :LongInt; stdcall;
TDAQmxSetSampQuantSampMode             = function (taskHandle :TtaskHandle; data :LongInt) :LongInt; stdcall;

TDAQmxGetSampTimingType                = function (taskHandle :TtaskHandle; var data :LongInt) :LongInt; stdcall;
TDAQmxSetSampTimingType                = function (taskHandle :TtaskHandle; data :LongInt) :LongInt; stdcall;

TDAQmxGetRealTimeWaitForNextSampClkWaitMode = function (taskHandle :TtaskHandle; var data :LongInt) :LongInt; stdcall;
TDAQmxSetRealTimeWaitForNextSampClkWaitMode = function (taskHandle :TtaskHandle; data :LongInt) :LongInt; stdcall;

TDAQmxGetSampClkRate                   = function (taskHandle :TtaskHandle; var data :Double) :LongInt; stdcall;
TDAQmxSetSampClkRate                   = function (taskHandle :TtaskHandle; data :Double) :LongInt; stdcall;

TDAQmxGetReadCurrReadPos               = function (taskHandle :TtaskHandle; var data :{u}Int64) :LongInt; stdcall;
TDAQmxGetWriteCurrWritePos             = function (taskHandle :TtaskHandle; var data :{u}Int64) :LongInt; stdcall;

TDAQmxGetBufInputBufSize               = function (taskHandle :TtaskHandle; var data :{u}LongWord) :LongInt; stdcall;
TDAQmxSetBufInputBufSize               = function (taskHandle :TtaskHandle; data :{u}LongWord) :LongInt; stdcall;
TDAQmxCfgInputBuffer                   = function (taskHandle :TtaskHandle; numSamplesPerChan :LongWord) :LongInt; stdcall;

TDAQmxGetReadRelativeTo                = function (taskHandle :TtaskHandle; var data :LongInt) :LongInt; stdcall;
TDAQmxSetReadRelativeTo                = function (taskHandle :TtaskHandle; data :LongInt) :LongInt; stdcall;

TDAQmxGetReadOffset                    = function (taskHandle :TtaskHandle; var data :LongInt) :LongInt; stdcall;
TDAQmxSetReadOffset                    = function (taskHandle :TtaskHandle; data :LongInt) :LongInt; stdcall;

TDAQmxGetSyncPulseSrc                  = function (taskHandle :TtaskHandle; const source :PChar; buffersize :LongWord) :LongInt; stdcall;

TDAQmxCfgDigEdgeStartTrig              = function (taskHandle :TtaskHandle; const source :PChar; triggeredge :LongInt) :LongInt; stdcall;

//int32 DllExport __CFUNC     DAQmxClearTask            (TaskHandle taskHandle);
//int32 DllExport __CFUNC     DAQmxCreateTask           (const char taskName[], TaskHandle *taskHandle);

//int32 DllExport __CFUNC     DAQmxStartTask            (TaskHandle taskHandle);
//int32 DllExport __CFUNC     DAQmxStopTask             (TaskHandle taskHandle);

//int32 DllExport __CFUNC     DAQmxGetErrorString       (int32 errorCode, char errorString[], uInt32 bufferSize);
//int32 DllExport __CFUNC     DAQmxGetExtendedErrorInfo (char errorString[], uInt32 bufferSize);

//int32 DllExport __CFUNC     DAQmxCreateAIVoltageChan          (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], int32 terminalConfig, float64 minVal, float64 maxVal, int32 units, const char customScaleName[]);
//int32 DllExport __CFUNC     DAQmxCreateAOVoltageChan  (TaskHandle taskHandle, const char physicalChannel[], const char nameToAssignToChannel[], float64 minVal, float64 maxVal, int32 units, const char customScaleName[]);

//int32 DllExport __CFUNC DAQmxGetAIResolutionUnits(TaskHandle taskHandle, const char channel[], int32 *data);
//int32 DllExport __CFUNC DAQmxGetAIResolution(TaskHandle taskHandle, const char channel[], float64 *data);
//int32 DllExport __CFUNC DAQmxGetAOResolutionUnits(TaskHandle taskHandle, const char channel[], int32 *data);
//int32 DllExport __CFUNC DAQmxGetAOResolution(TaskHandle taskHandle, const char channel[], float64 *data);
//int32 DllExport __CFUNC DAQmxGetAIRngHigh(TaskHandle taskHandle, const char channel[], float64 *data);
//int32 DllExport __CFUNC DAQmxGetAIRngLow(TaskHandle taskHandle, const char channel[], float64 *data);
//int32 DllExport __CFUNC DAQmxGetAODACRngHigh(TaskHandle taskHandle, const char channel[], float64 *data);
//int32 DllExport __CFUNC DAQmxGetAODACRngLow(TaskHandle taskHandle, const char channel[], float64 *data);

//int32 DllExport __CFUNC     DAQmxReadAnalogF64         (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, float64 readArray[], uInt32 arraySizeInSamps, int32 *sampsPerChanRead, bool32 *reserved);
//int32 DllExport __CFUNC     DAQmxReadBinaryI16        (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, bool32 fillMode, int16 readArray[], uInt32 arraySizeInSamps, int32 *sampsPerChanRead, bool32 *reserved);
//int32 DllExport __CFUNC     DAQmxReadRaw              (TaskHandle taskHandle, int32 numSampsPerChan, float64 timeout, void *readArray, uInt32 arraySizeInBytes, int32 *sampsRead, int32 *numBytesPerSamp, bool32 *reserved);

//int32 DllExport __CFUNC     DAQmxWriteBinaryI16       (TaskHandle taskHandle, int32 numSampsPerChan, bool32 autoStart, float64 timeout, bool32 dataLayout, int16 writeArray[], int32 *sampsPerChanWritten, bool32 *reserved);
//int32 DllExport __CFUNC     DAQmxWriteRaw             (TaskHandle taskHandle, int32 numSamps, bool32 autoStart, float64 timeout, void *writeArray, int32 *sampsPerChanWritten, bool32 *reserved);

//int32 DllExport __CFUNC     DAQmxCfgSampClkTiming          (TaskHandle taskHandle, const char source[], float64 rate, int32 activeEdge, int32 sampleMode, uInt64 sampsPerChan);


implementation


end.
