unit QUB_Acquisition_Global;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


Classes, SyncObjs, contnrs,
AllocTypes;

type


TQUB_Acquisition_Operation   = record
  Points                     :LongInt;
  Started                    :Boolean;
  Recording                  :Boolean;
end;
TQUB_Acquisition_POperation  = ^TQUB_Acquisition_Operation;


TQUB_Acquisition_Operations  = class(TQueue)
protected


  prCritSection              :TCriticalSection;
  prStarted                  :Boolean;
  prStopped                  :Boolean;
  prStartTime                :LongInt;


  procedure PushItem(AItem: Pointer); override;
  function PopItem: Pointer; override;
  function PeekItem: Pointer; override;


public


  constructor Create;
  destructor Destroy; override;


  procedure Reset;


  property Started           :Boolean    read prStarted      write prStarted;
  property Stopped           :Boolean    read prStopped      write prStopped;
  property StartTime         :LongInt    read prStartTime    write prStartTime;


end;

// used by QUB
TQUB_Acquisition_Mode                             = (ACQ_MODE_CONT, ACQ_MODE_TIME, ACQ_MODE_POINTS, ACQ_MODE_ONCE, ACQ_MODE_PROT);


TQUB_Acquisition_LibraryInitialize                = function :Integer; stdcall;
TQUB_Acquisition_LibraryTerminate                 = function :Integer; stdcall;

TQUB_Acquisition_LibraryNameLength                = function :Integer; stdcall;
TQUB_Acquisition_LibraryName                      = procedure(Name :PChar); stdcall;


//Input
TQUB_Acquisition_InputCallBack                    = procedure(Caller :TObject; Buffer :Pointer; BufferSize, BufferCount :Integer; Error :Integer); stdcall;

TQUB_Acquisition_InputStart_Params                = packed record
  Size                                            :Integer;
  ChannelCount                                    :Integer;
  Channels                                        :TPAInteger;
  Gain                                            :TPADouble;
  Bits                                            :Integer;
  Sampling                                        :Double;
  CallBack                                        :TQUB_Acquisition_InputCallBack;
  Caller                                          :TObject;
  BufferLength                                    :Integer;
  Buffer                                          :Pointer;//Is filled by the driver with the address of the internal buffer
  SingleBuffer                                    :Boolean;
  BufferCount                                     :Integer;
  BufferElemSize                                  :Integer;//Filled by the driver

  DeviceData                                      :Pointer;//Pointer to a hardware-specific structure
                                                           //Allocated/freed by the QUBDAQ driver Start/Stop functions
end;
TQUB_Acquisition_InputStart_PParams               = ^TQUB_Acquisition_InputStart_Params;

TQUB_Acquisition_InputStart                       = function(Device :Integer; var InputParams :TQUB_Acquisition_InputStart_Params) :Integer; stdcall;
TQUB_Acquisition_InputStop                        = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_InputDeviceCount                 = function :Integer; stdcall;
TQUB_Acquisition_InputDeviceName                  = procedure(Device :Integer; Name :PChar); stdcall;
TQUB_Acquisition_InputDeviceCode                  = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_InputChannelCount                = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_InputGain                        = function(Device, Channel :Integer) :Double; stdcall;
TQUB_Acquisition_InputRange                       = function(Device :Integer) :Double; stdcall;
TQUB_Acquisition_InputDifferential                = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_InputBits                        = function(Device :Integer) :Integer; stdcall;

TQUB_Acquisition_InputMonitor                     = function(Device, Channel, NumPoints :Integer; Buffer :Pointer; var LastPointIndex :LongWord) :Integer; stdcall;

TQUB_Acquisition_ReadDBLChannel                   = function(Device :Integer; Channel :Integer; Gain :Double; var Value :Double) :Integer; stdcall;

TQUB_Acquisition_AIRead                           = function(Device :Integer; Channel :Integer; Gain :Double; var Value :Integer) :Integer; stdcall;
TQUB_Acquisition_DIRead                           = function(Device :Integer; Channel :Integer; var Value :Integer) :Integer; stdcall;


//Output
TQUB_Acquisition_OutputCallBack                   = procedure(Caller :TObject; Buffer :Pointer; BufferSize :Integer; var BufferCount :Integer; var Error :Integer); stdcall;
TQUB_Acquisition_DigitalOutputCallBack            = procedure(Caller :TObject; var DigitalValue, DigitalCount :Integer; var Error :Integer); stdcall;
TQUB_Acquisition_OutputStart_Params               = packed record
  Size                                            :Integer;
  ChannelCount                                    :Integer;
  DigitalChannelCount                             :Integer;
  Channels                                        :TPAInteger;
  Gain                                            :TPADouble;
  Bits                                            :Integer;
  Sampling                                        :Double;
  DigitalSampling                                 :Double;
  CallBack                                        :TQUB_Acquisition_OutputCallBack;
  DigitalCallBack                                 :TQUB_Acquisition_DigitalOutputCallBack;
  Caller                                          :TObject;
  BufferLength                                    :Integer;
  Buffer                                          :Pointer;//Is filled by the driver with the address of the internal buffer
  SingleBuffer                                    :Boolean;
  BufferCount                                     :Integer;
  BufferElemSize                                  :Integer;//Filled by the driver

  DeviceData                                      :Pointer;//Pointer to a hardware-specific structure
                                                           //Allocated/freed by the QUBDAQ driver Start/Stop functions
end;
TQUB_Acquisition_OutputStart_PParams              = ^TQUB_Acquisition_OutputStart_Params;

TQUB_Acquisition_OutputStart                      = function(Device :Integer; var OutputParams :TQUB_Acquisition_OutputStart_Params) :Integer; stdcall;
TQUB_Acquisition_OutputStop                       = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_OutputDeviceCount                = function :Integer; stdcall;
TQUB_Acquisition_OutputDeviceName                 = procedure(Device :Integer; Name :PChar); stdcall;
TQUB_Acquisition_OutputChannelCount               = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_OutputDigitalChannelCount        = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_OutputGain                       = function(Device, Channel :Integer) :Double; stdcall;
TQUB_Acquisition_OutputRange                      = function(Device :Integer) :Double; stdcall;
TQUB_Acquisition_OutputDifferential               = function(Device :Integer) :Integer; stdcall;
TQUB_Acquisition_OutputBits                       = function(Device :Integer) :Integer; stdcall;

TQUB_Acquisition_WriteDBLChannel                  = function(Device :Integer; Channel :Integer; const Value :Double) :Integer; stdcall;

TQUB_Acquisition_AOWrite                          = function(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall;
TQUB_Acquisition_DOWrite                          = function(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall;


//Synchronized start / stop - Input and output almost simulataneous.
TQUB_Acquisition_SynchStart = function( InputDevice, OutputDevice :Integer; var InputParams :TQUB_Acquisition_InputStart_Params; var OutputParams :TQUB_Acquisition_OutputStart_Params; var r_i, r_o :Integer ) :Integer; stdcall;
TQUB_Acquisition_SynchStop = function( ) : Integer; stdcall;

const

ACQ_MAX_DEVICES                                   = 16;
ACQ_MAX_CHANNELS                                  = 16;

ACQ_MIN_DIGITALSAMPLING                           = 1e-6;
ACQ_MAX_DIGITALSAMPLING                           = 100000;


type

TQUB_AI_Buffer = record
  Buffer      :Pointer;
  BufferSize  :Integer;
  BufferCount :Integer;
  StartTime   :Double;
  NewSegment  :Boolean;
  Refresh     :Boolean;
  Recording   :Boolean;
end;

TQUB_AI_PBuffer = ^TQUB_AI_Buffer;

TQUB_BufferSupply = class
protected

  prSupply          :TList;
  prLock            :TCriticalSection;
  prMax             :Integer;
  prBytes           :Integer;

  function prAlloc: TQUB_AI_PBuffer;
  procedure prFree(Buf :TQUB_AI_PBuffer);

public

  constructor Create(Count, Max, Bytes :Integer);
  // Count: start with this many buffers
  // Max: store up to this many for re-use
  // Bytes: size of Buffer.Buffer
  destructor Destroy; override;

  function Get: TQUB_AI_PBuffer;
  procedure Put(Buf :TQUB_AI_PBuffer);

end;

TQUB_BufferQueue = class
protected

  prQ               :TList;
  prLock            :TCriticalSection;
  prRqAvail         :TEvent; // wakes up Accept()

public

  constructor Create;
  destructor Destroy; override;

  procedure Put(Buf :TQUB_AI_PBuffer);
  function Get(TimeoutMS :Integer): TQUB_AI_PBuffer; // or nil if timed out
  function Count: Integer;
  procedure Clear(Supply :TQUB_BufferSupply);

end;


const
{ Supply params:
An acquired buffer is copied into a buffer for queueing.
This capacity of buffers is readied ahead of time:             }
DEF_SPARE_BYTES                                 =  4*1024*1024;
// An unlimited additional amount will be GetMem'd as needed,
// and when the GUI message finishes with them they are returned to the "Supply",
// where up to this capacity will be retained for re-use:
MAX_SPARE_BYTES                                 = 16*1024*1024;

{ _AVAIL handler params:
I don't think queue reads should wait for something to arrive: }
BUFQ_TIMEOUT                                    = 0;
// An _AVAIL message can process more than one buffer if available,
// but not too many for responsiveness' sake.  The proper number is unknown/untested.
MAX_BUF_AT_ONCE                                 = 4;



implementation


//******************************************************************************
constructor TQUB_Acquisition_Operations.Create;
begin
  inherited;
  prCritSection := TCriticalSection.Create;
end;

destructor TQUB_Acquisition_Operations.Destroy;
begin
  try
    Reset;
    prCritSection.Free;
    prCritSection := nil;
  finally
    inherited;
  end;
end;

procedure TQUB_Acquisition_Operations.Reset;
var
  Op :TQUB_Acquisition_POperation;
begin

  repeat

    if Count <> 0 then Op := Pop else Op := nil;
    if Op <> nil then begin
      Dispose(Op);
    end;

  until(Op = nil);

  prStarted    := False;
  prStopped    := False;

end;

procedure TQUB_Acquisition_Operations.PushItem(AItem: Pointer);
begin
  try
    prCritSection.Acquire;
    inherited;
  finally
    prCritSection.Release;
  end;
end;

function TQUB_Acquisition_Operations.PopItem: Pointer;
begin
  try
    prCritSection.Acquire;
    if Count = 0 then Result := nil else Result := inherited PopItem;
  finally
    prCritSection.Release;
  end;
end;

function TQUB_Acquisition_Operations.PeekItem: Pointer;
begin
  try
    prCritSection.Acquire;
    if Count = 0 then Result := nil else Result := inherited PeekItem;
  finally
    prCritSection.Release;
  end;
end;


//******************************************************************************
constructor TQUB_BufferSupply.Create(Count, Max, Bytes :Integer);
var
  i                 :Integer;
begin
  prSupply          := TList.Create;
  prLock            := TCriticalSection.Create;
  prMax             := Max;
  prBytes           := Bytes;

  for i := 0 to Count - 1 do begin
    Put(prAlloc());
  end;
end;

destructor TQUB_BufferSupply.Destroy;
var
  i                 :Integer;
begin
  try
    prLock.Free;
    for i := 0 to prSupply.Count - 1 do begin
      prFree(prSupply[i]);
    end;
    prSupply.Free;
  finally
    inherited;
  end;
end;

function TQUB_BufferSupply.Get: TQUB_AI_PBuffer;
begin
  prLock.Acquire;
  try
    if prSupply.Count = 0 then begin
      Result := prAlloc();
    end else begin
      Result := TQUB_AI_PBuffer(prSupply[prSupply.Count - 1]);
      prSupply.Delete(prSupply.Count - 1);
    end;
  finally
    prLock.Release;
  end;
end;

procedure TQUB_BufferSupply.Put(Buf :TQUB_AI_PBuffer);
begin
  prLock.Acquire;
  try
    if prSupply.Count >= prMax then begin
      prFree(Buf);
    end else begin
      prSupply.Add(Buf);
    end;
  finally
    prLock.Release;
  end;
end;

function TQUB_BufferSupply.prAlloc: TQUB_AI_PBuffer;
begin
  GetMem(Result, SizeOf(TQUB_AI_Buffer));
  GetMem(Result.Buffer, prBytes);
end;

procedure TQUB_BufferSupply.prFree(Buf :TQUB_AI_PBuffer);
begin
  FreeMem(Buf.Buffer);
  FreeMem(Buf);
end;


constructor TQUB_BufferQueue.Create;
begin
  prQ          := TList.Create;
  prLock       := TCriticalSection.Create;
  prRqAvail    := TEvent.Create(nil{security}, False{auto-reset}, False{init:off}, ''{unique});
end;

destructor TQUB_BufferQueue.Destroy;
begin
  try
    prQ.Free;
    prLock.Free;
    prRqAvail.Free;
  finally
    inherited;
  end;
end;

procedure TQUB_BufferQueue.Put(Buf :TQUB_AI_PBuffer);
begin
  prLock.Acquire;
  prQ.Add(Buf);
  prLock.Release;
  prRqAvail.SetEvent;
end;

function TQUB_BufferQueue.Get(TimeoutMS :Integer): TQUB_AI_PBuffer; // or nil if timed out
begin
  Result := nil;

  prLock.Acquire;
  if prQ.Count > 0 then begin
    Result := TQUB_AI_PBuffer(prQ[0]);
    prQ.Delete(0);
  end;
  prLock.Release;

  if Result = nil then begin // Q was empty
    prRqAvail.WaitFor( TimeoutMS );
    prLock.Acquire;
    if prQ.Count > 0 then begin
      Result := TQUB_AI_PBuffer(prQ[0]);
      prQ.Delete(0);
    end;
    prLock.Release;
  end;

  prRqAvail.ResetEvent;
end;

function TQUB_BufferQueue.Count: Integer;
begin
  prLock.Acquire;
  Result := prQ.Count;
  prLock.Release;
end;

procedure TQUB_BufferQueue.Clear(Supply :TQUB_BufferSupply);
var
  i        :Integer;
begin
  prLock.Acquire;
  for i := 0 to prQ.Count - 1 do begin
    Supply.Put(TQUB_AI_PBuffer(prQ[i]));
  end;
  prQ.Clear;
  prLock.Release;
end;


//******************************************************************************
end.
