library qubacqdx;


{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
//05-27-2001


{
                         QUB data acquisition library interface

                             for Data Translation Cards
}


//The IMPORTANT assumption is that only ONE device /board is in use at a time. This can be changed later


uses


SysUtils, Windows,

AllocTypes,

OLErrors,
OLTypes,
OLdaapi,
OLdaDefs,
OLMem,

QUB_Global,
QUB_Acquisition_Global;


{$R *.RES}


const


LIBRARY_NAME                       = 'QUB DATA TRANSLATION 1.4';


var


//Input
//Board
InputBoardCount                    :Integer                                     = 0;
InputBoardName                     :Array[0..ACQ_MAX_DEVICES - 1] of String;
InputBoardHandle                   :HDEV                                        = 0;
//AD subsystem
InputADHandle                      :HDASS                                       = 0;
//ChannelCount
InputChCount                       :Integer                                     = 0;
//Buffer
InputElemSize                      :Integer                                     = 0;
InputBufferHandle                  :Array[1..2] of HBUF                         = (0, 0);
//Flags
InputCalling                       :Boolean                                     = False;
InputStopped                       :Boolean                                     = True;
InputError                         :Integer                                     = 0;
InputUserCallBack                  :TQUB_Acquisition_InputCallBack              = nil;
InputUserCaller                    :TObject                                     = nil;



//Output
OutputBoardCount                   :Integer                                     = 0;
OutputBoardName                    :Array[0..ACQ_MAX_DEVICES - 1] of String;
OutputBoardHandle                  :HDEV                                        = 0;
//DA subsystem
OutputDAHandle                     :HDASS                                       = 0;
//D subsystem
OutputDDHandle                     :HDASS                                       = 0;
//Channels
OutputChCount                      :Integer                                     = 0;
//Buffer
OutputElemSize                     :Integer                                     = 0;
OutputBufferSize                   :Integer                                     = 0;
OutputBufferHandle                 :Array[1..2] of HBUF                         = (0, 0);
//Flags
OutputCalling                      :Boolean                                     = False;
OutputStopped                      :Boolean                                     = True;
OutputError                        :Integer                                     = 0;
OutputUserCallBack                 :TQUB_Acquisition_OutputCallBack             = nil;
OutputUserCaller                   :TObject                                     = nil;
//Digital
OutputDigitalChCount               :Integer;
//OutputDigitalValue                 :Integer;
//OutputDigitalCount                 :Integer;
//OutputDigitalError                 :Integer;
OutputDigitalUserCallBack          :TQUB_Acquisition_DigitalOutputCallBack;
OutputDigitalCalling               :Boolean                                     = False;


//Export functions
function LibraryInitialize :Integer; stdcall; forward;
function LibraryTerminate :Integer; stdcall; forward;


function LibraryNameLength :Integer; stdcall; forward;
procedure LibraryName(Name :PChar); stdcall; forward;


//Input
function InputStart(Device :Integer; var InputParams :TQUB_Acquisition_InputStart_Params) :Integer; stdcall; forward;
function InputStop(Device :Integer) :Integer; stdcall; forward;

function InputDeviceCount :Integer; stdcall; forward;
procedure InputDeviceName(Device :Integer; Name :PChar); stdcall; forward;
function InputChannelCount(Device :Integer) :Integer; stdcall; forward;
function InputGain(Device, Channel :Integer) :Double; stdcall; forward;
function InputRange(Device :Integer) :Double; stdcall; forward;
function InputDifferential(Device :Integer) :Integer; stdcall; forward;
function InputBits(Device :Integer) :Integer; stdcall; forward;

function ReadDBLChannel(Device :Integer; Channel :Integer; Gain :Double; var Value :Double) :Integer; stdcall; forward;

function AIRead(Device :Integer; Channel :Integer; Gain :Double; var Value :Integer) :Integer; stdcall; forward;
function DIRead(Device :Integer; Channel :Integer; var Value :Integer) :Integer; stdcall; forward;

procedure InputDeviceCallBack(uiMsg :UINT; wParam :WPARAM; lParam :LPARAM); stdcall; forward;


//Output
function OutputStart(Device :Integer; var OutputParams :TQUB_Acquisition_OutputStart_Params) :Integer; stdcall; forward;
function OutputStop(Device :Integer) :Integer; stdcall; forward;

function OutputDeviceCount :Integer; stdcall; forward;
procedure OutputDeviceName(Device :Integer; Name :PChar); stdcall; forward;
function OutputChannelCount(Device :Integer) :Integer; stdcall; forward;
function OutputDigitalChannelCount(Device :Integer) :Integer; stdcall; forward;
function OutputGain(Device, Channel :Integer) :Double; stdcall; forward;
function OutputRange(Device :Integer) :Double; stdcall; forward;
function OutputDifferential(Device :Integer) :Integer; stdcall; forward;
function OutputBits(Device :Integer) :Integer; stdcall; forward;

function WriteDBLChannel(Device :Integer; Channel :Integer; const Value :Double) :Integer; stdcall; forward;

function AOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall; forward;
function DOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall; forward;

procedure OutputDeviceCallBack(uiMsg :UINT; wParam :WPARAM; lParam :LPARAM); stdcall; forward;
procedure OutputDigitalDeviceCallBack(uiMsg :UINT; wParam :WPARAM; lParam :LPARAM); stdcall; forward;


//Synch
function SynchStart(InputDevice, OutputDevice :Integer; var InputParams :TQUB_Acquisition_InputStart_Params; var OutputParams :TQUB_Acquisition_OutputStart_Params; var r_i, r_o :Integer) :Integer; stdcall; forward;




//****************************************************************************//
//                                                                            //
//                 I N I T I A L I Z E   /   T E R M I N A T E                //
//                                                                            //
//****************************************************************************//
function LibraryInitialize :Integer;
begin
  Result := 0;

end;

function LibraryTerminate :Integer;
begin
  Result := 0;

end;



//****************************************************************************//
//                                                                            //
//                          A C Q U I S I T I O N                             //
//                                                                            //
//****************************************************************************//

//Input ************************************************************************
function InputStart(Device :Integer; var InputParams :TQUB_Acquisition_InputStart_Params) :Integer;
var
  ADssCount    :UINT;
  goodADFound  :Boolean;
  maxFreq      :DBL;
  dma          :UINT;
  gainsup      :UINT;
  BufferSize   :Integer;
  i, c         :Integer;
begin

  Result := 1;

  if not InputStopped  then Exit;


  InputCalling      := False;
  InputUserCallBack := InputParams.CallBack;
  InputUserCaller   := InputParams.Caller;


  InputBoardHandle  := 0;
  InputADHandle     := 0;

  InputChCount      := InputParams.ChannelCount;


  try

    try


      //Get a handle to the board
      Result := olDaInitialize(PChar(InputBoardName[Device]), InputBoardHandle);
      if Result <> OLNOERROR then Exit;

      Result := Integer(InputBoardHandle = 0);
      if Result <> OLNOERROR then Exit;

      //How many DA subsystems available
      Result := olDaGetDevCaps(InputBoardHandle, OLDC_ADELEMENTS, ADssCount);
      if Result <> OLNOERROR then Exit;

      Result := Integer(ADssCount = 0);
      if Result <> OLNOERROR then Exit;

      GoodADFound := False;
      for i := 0 to ADssCount - 1 do begin
        if olDaGetDASS(InputBoardHandle, OLSS_AD, i, InputADHandle) = OLNOERROR then begin
        //Found a good AD ss;
          GoodADFound := True;
          Break;
        end;
      end;
      Result := Integer(GoodADFound = False);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCapsEx(InputADHandle, OLSSCE_MAXTHROUGHPUT, MaxFreq);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCaps(InputADHandle, OLSSC_NUMDMACHANS, dma);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCaps(InputADHandle, OLSSC_SUP_PROGRAMGAIN, gainsup);
      if Result <> OLNOERROR then Exit;

      if dma > 1 then dma := 1;

      Result := Integer(MaxFreq < InputParams.Sampling);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetDataFlow(InputADHandle, OL_DF_CONTINUOUS);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetWrapMode(InputADHandle, OL_WRP_MULTIPLE);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetClockFrequency(InputADHandle, DBL(InputParams.Sampling));
      if Result <> OLNOERROR then Exit;

      Result := olDaSetDmaUsage(InputADHandle, dma);
      if Result <> OLNOERROR then Exit;

      //Set Channels and Gains
      for c := 0 to InputParams.ChannelCount - 1 do begin

        Result := olDaSetChannelListEntry(InputADHandle, c, InputParams.Channels[c]);
        if Result <> OLNOERROR then Exit;

        //Only set the gain if the board supports it !!!
        if Boolean(gainsup) then begin
          olDaSetGainListEntry(InputADHandle, c, InputParams.Gain[c]);
          if Result <> OLNOERROR then Exit;
        end;

      end;

      //Set ChannelCount
      Result := olDaSetChannelListSize(InputADHandle, InputParams.ChannelCount);
      if Result <> OLNOERROR then Exit;

      //Set the Notification procedure
      Result := olDaSetNotificationProcedure(InputADHandle, InputDeviceCallBack, 0);
      if Result <> OLNOERROR then Exit;

      //Allocate Buffers
      InputElemSize := InputParams.ChannelCount * (((InputParams.Bits - 1) div 8) + 1);
      BufferSize    := ((Round((InputParams.BufferLength / 1000) * InputParams.Sampling) + (512 - 1)) div 512) * 512;
      if BufferSize > ONE_MEGABYTE then BufferSize := ONE_MEGABYTE;             //But not larger than one MegaSamples

      for i := 1 to 2 do begin

        Result := olDmCallocBuffer(GPTR, 0, BufferSize, InputElemSize, InputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

        Result := olDaPutBuffer(InputADHandle, InputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

      end;

      //Config - updates the configuration
      Result := olDaConfig(InputADHandle);
      if Result <> OLNOERROR then Exit;


      //*********************
      InputStopped  := False;
      //*********************


      Result := olDaStart(InputADHandle);
      if Result <> OLNOERROR then Exit;


    except

      Result := 1;

    end;


  finally


    if Result <> 0 then begin

      InputStop(Device);
      Result := 1;

    end;

  end;

end;

function InputStop(Device :Integer) :Integer;
var
  i :Integer;
begin

  Result := 0;

  InputStopped := True;

  try

    while InputCalling do begin
    end;

    olDaSetNotificationProcedure(InputADHandle, nil, 0);

    olDaFlushBuffers(InputADHandle);
    for i := 1 to 2 do begin
      olDaGetBuffer(InputADHandle, InputBufferHandle[i]);
      olDmFreeBuffer(InputBufferHandle[i]);
    end;

    olDaReleaseDASS(InputADHandle);
    olDaTerminate(InputBoardHandle);

  except
  end;

end;



//****************************************************************************//
//                                                                            //
//                               C A L L B A C K                              //
//                                                                            //
//****************************************************************************//
procedure InputDeviceCallBack(uiMsg :UINT; wParam :WPARAM; lParam :LPARAM);
var
  Buffer  :Pointer;
  hBuffer :HBUF;
  Samples :UINT;
  r       :Integer;
begin

  InputCalling := True;

  try


    //************************
    if InputStopped then Exit;
    //************************


    case uiMsg of

      OLDA_WM_BUFFER_DONE:

        begin

          //Get buffer off done list
          r := olDaGetBuffer(wParam, hBuffer);
          if r = OLNOERROR then r := Integer(hBuffer = 0);
          if r = OLNOERROR then begin
          //If no error and there is a buffer

            //get max samples in input buffer
            r := olDmGetValidSamples(hBuffer, Samples);
            if r = OLNOERROR then begin

              //get pointer to the buffer
              r := olDmGetBufferPtr(hBuffer, Buffer);
              if r = OLNOERROR then begin

                try
                  InputUserCallBack(InputUserCaller, Buffer, InputElemSize, (Samples div UINT(InputChCount)), Error);
                except
                end;

              end;

            end;

          end;

          //put buffer back to list
          if hBuffer <> 0 then olDaPutBuffer(wParam, hBuffer);

          InputError := r;

        end;//BUFFER_DONE

    end;//case uiMsg


  finally

    InputCalling := False;

  end;

end;

function ReadDBLChannel(Device :Integer; Channel :Integer; Gain :Double; var Value :Double) :Integer;
begin
  //Function not yet implemented
  Result := 1;
end;

function AIRead(Device :Integer; Channel :Integer; Gain :Double; var Value :Integer) :Integer;
var
  v        :LongInt;
begin
  //Not fully implemented - update in the new driver style
  if          InputBoardHandle = 0 then begin
    Result := olDaInitialize(PChar(InputBoardName[Device]), InputBoardHandle);
    if Result <> 0 then Exit;
    Result := olDaGetDASS(InputBoardHandle, OLSS_AD, 0, InputADHandle);
    if Result <> 0 then Exit;
    Result := olDaSetDataFlow(InputADHandle, OL_DF_SINGLEVALUE);
    if Result <> 0 then Exit;
    Result := olDaConfig(InputADHandle);
    if Result <> 0 then Exit;
  end else if InputADHandle    = 0 then begin
    Result := olDaGetDASS(InputBoardHandle, OLSS_AD, 0, InputADHandle);
    if Result <> 0 then Exit;
    Result := olDaSetDataFlow(InputADHandle, OL_DF_SINGLEVALUE);
    if Result <> 0 then Exit;
    Result := olDaConfig(InputADHandle);
    if Result <> 0 then Exit;
  end;
  Result   := olDaGetSingleValue(InputADHandle, @v, Channel, Gain);
  Value    := v;
end;

function DIRead(Device :Integer; Channel :Integer; var Value :Integer) :Integer;
begin
  Result := 1;//Not yet implemented
end;


//Output ***********************************************************************
function OutputStart(Device :Integer; var OutputParams :TQUB_Acquisition_OutputStart_Params) :Integer;
var
  DAssCount    :UINT;
  goodDAFound  :Boolean;
  maxFreq      :DBL;
  dma          :UINT;
  i, c         :Integer;
  //DigVal     :Integer;
  Count        :Integer;
  Buffer       :Pointer;
begin

  Result := 1;

  if not OutputStopped  then Exit;


  OutputCalling             := False;
  OutputDigitalCalling      := False;
  OutputUserCallBack        := OutputParams.CallBack;
  OutputDigitalUserCallBack := OutputParams.DigitalCallBack;
  OutputUserCaller          := OutputParams.Caller;


  OutputBoardHandle         := 0;
  OutputDAHandle            := 0;
  OutputDDHandle            := 0;

  OutputChCount             := OutputParams.ChannelCount;
  OutputDigitalChCount      := OutputParams.DigitalChannelCount;


  try

    try


      //Get a handle to the board
      Result := olDaInitialize(PChar(OutputBoardName[Device]), OutputBoardHandle);
      if Result <> OLNOERROR then Exit;

      Result := Integer(OutputBoardHandle = 0);
      if Result <> OLNOERROR then Exit;

      //How many DD subsystems available
      Result := olDaGetDevCaps(OutputBoardHandle, OLDC_DAELEMENTS, DAssCount);
      if Result <> OLNOERROR then Exit;

      Result := Integer(DAssCount = 0);
      if Result <> OLNOERROR then Exit;

      GoodDAFound := False;
      for i := 0 to DAssCount - 1 do begin
        if olDaGetDASS(OutputBoardHandle, OLSS_DA, i, OutputDAHandle) = OLNOERROR then begin
        //Found a good DA ss;
          GoodDAFound := True;
          Break;
        end;
      end;
      Result := Integer(GoodDAFound = False);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCapsEx(OutputDAHandle, OLSSCE_MAXTHROUGHPUT, MaxFreq);
      if Result <> OLNOERROR then Exit;

      Result := Integer(MaxFreq < OutputParams.Sampling);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCaps(OutputDAHandle, OLSSC_NUMDMACHANS, dma);
      if Result <> OLNOERROR then Exit;
      if dma > 1 then dma := 1;

      Result := olDaSetDataFlow(OutputDAHandle, OL_DF_CONTINUOUS);
      if Result <> OLNOERROR then Exit;
      //for 9802 returns 36

      Result := olDaSetWrapMode(OutputDAHandle, OL_WRP_MULTIPLE);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetClockFrequency(OutputDAHandle, DBL(OutputParams.Sampling));
      if Result <> OLNOERROR then Exit;

      Result := olDaSetDmaUsage(OutputDAHandle, dma);
      if Result <> OLNOERROR then Exit;

      //Set ChannelCount
      //Result := olDaSetChannelListSize(OutputDAHandle, OutputParams.ChannelCount);
      //if Result <> OLNOERROR then Exit;

      //Set Channels and Gains
      for c := 0 to OutputParams.ChannelCount - 1 do begin

        Result := olDaSetChannelListEntry(OutputDAHandle, c, OutputParams.Channels[c]);
        if Result <> OLNOERROR then Exit;

      end;

      //Before ListSize or After ?!
      //Set ChannelCount
      Result := olDaSetChannelListSize(OutputDAHandle, OutputParams.ChannelCount);
      if Result <> OLNOERROR then Exit;

      //Set the Notification procedure
      Result := olDaSetNotificationProcedure(OutputDAHandle, OutputDeviceCallBack, 0);
      if Result <> OLNOERROR then Exit;

      //Allocate Buffers
      OutputElemSize   := OutputParams.ChannelCount * (((OutputParams.Bits - 1) div 8) + 1);
      OutputBufferSize := ((Round((OutputParams.BufferLength / 1000) * OutputParams.Sampling) + (512 - 1)) div 512) * 512;
      if OutputBufferSize > ONE_MEGABYTE then OutputBufferSize := ONE_MEGABYTE;        //But not larger than one MegaSamples

      for i := 1 to 2 do begin

        Result := olDmCallocBuffer(GPTR, 0, OutputBufferSize, OutputElemSize, OutputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

        Result := olDaPutBuffer(OutputDAHandle, OutputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

        Count := OutputBufferSize;
        try
          olDmGetBufferPtr(OutputBufferHandle[i], Buffer);
          OutputUserCallBack(OutputUserCaller, Buffer, OutputElemSize, Count, OutputError);
        except
        end;

      end;


      //**********************************************************************//
      //                                                                      //
      //                             D I G I T A L                            //
      //                                                                      //
      //**********************************************************************//

      //For now let's not do digital

      OutputParams.DigitalChannelCount := 0;

      if OutputParams.DigitalChannelCount > 0 then begin

        (*
        OutputDigitalCount := Round(Sampling / DigitalSampling);
        if (OutputDigitalCount < 1)   then OutputDigitalCount := 1;
        case OutputDigitalCount of
          1:        OutputDigitalCount := 1;
          2:        OutputDigitalCount := 2;
          3..4:     OutputDigitalCount := 4;
          5..8:     OutputDigitalCount := 8;
          9..16:    OutputDigitalCount := 16;
          17..32:   OutputDigitalCount := 32;
          33..64:   OutputDigitalCount := 64;
          65..128:  OutputDigitalCount := 128;
          129..256: OutputDigitalCount := 256;
          257..512: OutputDigitalCount := 512;
          else  OutputDigitalCount := ((OutputDigitalCount + (512 - 1)) div 512) * 512;
        end;
        OutputDigitalUserCallBack(OutputUserCaller, OutputDigitalValue, OutputDigitalCount, OutputDigitalError);
        DigVal := OutputDigitalValue;
        for c := 0 to DigitalChannelCount - 1 do begin
          Result := DIG_Prt_Config(Devices[Device], c, 0, 1);
          Result := DIG_Out_Port(Devices[Device], c, DigVal and $ff);
          DigVal := DigVal shr 8;
        end;
        //Set the Event trapping
        Result := Config_DAQ_Event_Message(Devices[Device], 1{Add}, PChar(OutputDigitalChStr), 1, OutputDigitalCount, 0, 0, 0, 0, 0, 0, Cardinal(@OutputDigitalDeviceCallback));

        Result := olDaStart(OutputDDHandle);
        if Result <> OLNOERROR then Exit;
        *)
      end;


      Result := olDaConfig(OutputDAHandle);
      if Result <> OLNOERROR then Exit;

      //********************
      OutputStopped := False;
      //********************


      Result := olDaStart(OutputDAHandle);
      if Result <> OLNOERROR then Exit;


    except

      Result := 1;

    end;


  finally


    if Result <> 0 then begin

      Result := 1;
      OutputStop(Device);

    end;

  end;

end;

function OutputStop(Device :Integer) :Integer;
var
  i :Integer;
begin

  Result := 0;

  OutputStopped := True;

  try

    //Analog
    while OutputCalling do begin
    end;

    olDaSetNotificationProcedure(OutputDAHandle, nil, 0);

    olDaFlushBuffers(OutputDAHandle);
    for i := 1 to 2 do begin
      olDaGetBuffer(OutputDAHandle, OutputBufferHandle[i]);
      olDmFreeBuffer(OutputBufferHandle[i]);
    end;

    olDaReleaseDASS(OutputDAHandle);
    olDaTerminate(OutputBoardHandle);

    //Digital
    while OutputDigitalCalling do begin
    end;

  except
  end;

end;


//****************************************************************************//
//                                                                            //
//                               C A L L B A C K                              //
//                                                                            //
//****************************************************************************//
procedure OutputDeviceCallBack(uiMsg :UINT; wParam :WPARAM; lParam :LPARAM);
var
  Buffer  :Pointer;
  hBuffer :HBUF;
  Count   :Integer;
  r       :Integer;
begin

  OutputCalling := True;

  try


    //*************************
    if OutputStopped then Exit;
    //*************************


    case uiMsg of

      OLDA_WM_BUFFER_DONE:

        begin

          //Get buffer off done list
          r := olDaGetBuffer(wParam, hBuffer);
          if r = OLNOERROR then r := Integer(hBuffer = 0);
          if r = OLNOERROR then begin
          //If no error and there is a buffer

            //get pointer to the buffer
            r := olDmGetBufferPtr(hBuffer, Buffer);
            if r = OLNOERROR then begin

              Count := OutputBufferSize;
              try
                OutputUserCallBack(OutputUserCaller, Buffer, OutputElemSize, Count, OutputError);
              except
              end;

            end;

          end;

          //put buffer back to list
          if hBuffer <> 0 then olDaPutBuffer(wParam, hBuffer);

          OutputError := r;

        end;//BUFFER_DONE

    end;//case uiMsg


  finally

    OutputCalling := False;

  end;

end;


//****************************************************************************//
//                                                                            //
//                     D I G I T A L     C A L L B A C K                      //
//                                                                            //
//****************************************************************************//
procedure OutputDigitalDeviceCallBack(uiMsg :UINT; wParam :WPARAM; lParam :LPARAM);
{
var
  DigVal     :Integer;
  c          :Integer;
}
begin

  OutputDigitalCalling := True;

  (*

  try

    if OutputStopped then Exit;

    try
      //Get new values
      OutputDigitalUserCallBack(OutputUserCaller, OutputDigitalValue, OutputDigitalCount, OutputDigitalError);
      //Write new value
      DigVal := OutputDigitalValue;
      for c := 0 to OutputDeviceDigitalChannelCount - 1 do begin
        OutputDigitalError := DIG_Out_Port(OutputDeviceHandle, c, DigVal and $ff);
        DigVal := DigVal shr 8;
      end;
    except
    end;

  finally

    OutputDigitalCalling := False;

  end;
  *)

end;


function WriteDBLChannel(Device :Integer; Channel :Integer; const Value :Double) :Integer;
begin
  Result := 1;//Function not yet implemented
end;

function AOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer;
var
  v        :LongInt;
begin
  if          OutputBoardHandle = 0 then begin
    Result := olDaInitialize(PChar(OutputBoardName[Device]), OutputBoardHandle);
    if Result <> 0 then Exit;
    Result := olDaGetDASS(OutputBoardHandle, OLSS_DA, 0, OutputDAHandle);
    if Result <> 0 then Exit;
    Result := olDaSetDataFlow(OutputDAHandle, OL_DF_SINGLEVALUE);
    if Result <> 0 then Exit;
    Result := olDaConfig(OutputDAHandle);
    if Result <> 0 then Exit;
  end else if OutputDAHandle    = 0 then begin
   Result := olDaGetDASS(OutputBoardHandle, OLSS_DA, 0, OutputDAHandle);
    if Result <> 0 then Exit;
    Result := olDaSetDataFlow(OutputDAHandle, OL_DF_SINGLEVALUE);
    if Result <> 0 then Exit;
    Result := olDaConfig(OutputDAHandle);
    if Result <> 0 then Exit;
  end;
  v        := Value;
  Result   := olDaPutSingleValue(OutputDAHandle, v, Channel, 1{Gain});
end;

function DOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer;
begin
  Result := 1;//Function not yet implemented
end;


//Synch ************************************************************************
function SynchStart(InputDevice, OutputDevice :Integer; var InputParams :TQUB_Acquisition_InputStart_Params; var OutputParams :TQUB_Acquisition_OutputStart_Params; var r_i, r_o :Integer) :Integer;
var
  ADssCount    :UINT;
  goodADFound  :Boolean;
  gainsup      :UINT;
  BufferSize   :Integer;

  DAssCount    :UINT;
  goodDAFound  :Boolean;
  Count        :Integer;
  Buffer       :Pointer;

  sstart       :UINT;
  maxFreq      :DBL;
  dma          :UINT;
  i, c         :Integer;

  SSList       :HSSLIST;

begin

  r_i    := 1;
  r_o    := 1;
  Result := 1;

  if not InputStopped   then Exit;
  if not OutputStopped  then Exit;


  InputCalling              := False;
  InputUserCallBack         := InputParams.CallBack;
  InputUserCaller           := InputParams.Caller;
  InputBoardHandle          := 0;
  InputADHandle             := 0;
  InputChCount              := InputParams.ChannelCount;

  OutputCalling             := False;
  OutputDigitalCalling      := False;
  OutputUserCallBack        := OutputParams.CallBack;
  OutputDigitalUserCallBack := OutputParams.DigitalCallBack;
  OutputUserCaller          := OutputParams.Caller;
  OutputBoardHandle         := 0;
  OutputDAHandle            := 0;
  OutputDDHandle            := 0;
  OutputChCount             := OutputParams.ChannelCount;
  OutputDigitalChCount      := OutputParams.DigitalChannelCount;


  r_o    := 0;//preemptive


  try

    try


      //Input ******************************************************************
      r_i := 1;

      //Get a handle to the board
      Result := olDaInitialize(PChar(InputBoardName[InputDevice]), InputBoardHandle);
      if Result <> OLNOERROR then Exit;

      Result := Integer(InputBoardHandle = 0);
      if Result <> OLNOERROR then Exit;

      //How many DA subsystems available
      Result := olDaGetDevCaps(InputBoardHandle, OLDC_ADELEMENTS, ADssCount);
      if Result <> OLNOERROR then Exit;

      Result := Integer(ADssCount = 0);
      if Result <> OLNOERROR then Exit;

      GoodADFound := False;
      for i := 0 to ADssCount - 1 do begin
        if olDaGetDASS(InputBoardHandle, OLSS_AD, i, InputADHandle) = OLNOERROR then begin
        //Found a good AD ss;
          GoodADFound := True;
          Break;
        end;
      end;
      Result := Integer(GoodADFound = False);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCapsEx(InputADHandle, OLSSCE_MAXTHROUGHPUT, MaxFreq);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCaps(InputADHandle, OLSSC_NUMDMACHANS, dma);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCaps(InputADHandle, OLSSC_SUP_PROGRAMGAIN, gainsup);
      if Result <> OLNOERROR then Exit;

      if dma > 1 then dma := 1;

      Result := Integer(MaxFreq < InputParams.Sampling);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetDataFlow(InputADHandle, OL_DF_CONTINUOUS);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetWrapMode(InputADHandle, OL_WRP_MULTIPLE);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetClockFrequency(InputADHandle, DBL(InputParams.Sampling));
      if Result <> OLNOERROR then Exit;

      Result := olDaSetDmaUsage(InputADHandle, dma);
      if Result <> OLNOERROR then Exit;

      //Set Channels and Gains
      for c := 0 to InputParams.ChannelCount - 1 do begin

        Result := olDaSetChannelListEntry(InputADHandle, c, InputParams.Channels[c]);
        if Result <> OLNOERROR then Exit;

        //Only set the gain if the board supports it !!!
        if Boolean(gainsup) then begin
          olDaSetGainListEntry(InputADHandle, c, InputParams.Gain[c]);
          if Result <> OLNOERROR then Exit;
        end;

      end;

      //Set ChannelCount
      Result := olDaSetChannelListSize(InputADHandle, InputParams.ChannelCount);
      if Result <> OLNOERROR then Exit;

      //Set the Notification procedure
      Result := olDaSetNotificationProcedure(InputADHandle, InputDeviceCallBack, 0);
      if Result <> OLNOERROR then Exit;

      //Allocate Buffers
      InputElemSize := InputParams.ChannelCount * (((InputParams.Bits - 1) div 8) + 1);
      BufferSize    := ((Round((InputParams.BufferLength / 1000) * InputParams.Sampling) + (512 - 1)) div 512) * 512;
      if BufferSize > ONE_MEGABYTE then BufferSize := ONE_MEGABYTE;             //But not larger than one MegaSamples

      for i := 1 to 2 do begin

        Result := olDmCallocBuffer(GPTR, 0, BufferSize, InputElemSize, InputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

        Result := olDaPutBuffer(InputADHandle, InputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

      end;

      //Config
      Result := olDaConfig(InputADHandle);
      if Result <> OLNOERROR then Exit;


      r_i := 0;
      //************************************************************************


      //Output *****************************************************************
      r_o := 1;

      //Get a handle to the board
      Result := olDaInitialize(PChar(OutputBoardName[OutputDevice]), OutputBoardHandle);
      if Result <> OLNOERROR then Exit;

      Result := Integer(OutputBoardHandle = 0);
      if Result <> OLNOERROR then Exit;

      //How many DD subsystems available
      Result := olDaGetDevCaps(OutputBoardHandle, OLDC_DAELEMENTS, DAssCount);
      if Result <> OLNOERROR then Exit;

      Result := Integer(DAssCount = 0);
      if Result <> OLNOERROR then Exit;

      GoodDAFound := False;
      for i := 0 to DAssCount - 1 do begin
        if olDaGetDASS(OutputBoardHandle, OLSS_DA, i, OutputDAHandle) = OLNOERROR then begin
        //Found a good DA ss;
          GoodDAFound := True;
          Break;
        end;
      end;
      Result := Integer(GoodDAFound = False);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCapsEx(OutputDAHandle, OLSSCE_MAXTHROUGHPUT, MaxFreq);
      if Result <> OLNOERROR then Exit;

      Result := Integer(MaxFreq < OutputParams.Sampling);
      if Result <> OLNOERROR then Exit;

      Result := olDaGetSSCaps(OutputDAHandle, OLSSC_NUMDMACHANS, dma);
      if Result <> OLNOERROR then Exit;

      if dma > 1 then dma := 1;

      Result := olDaSetDataFlow(OutputDAHandle, OL_DF_CONTINUOUS);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetWrapMode(OutputDAHandle, OL_WRP_MULTIPLE);
      if Result <> OLNOERROR then Exit;

      Result := olDaSetClockFrequency(OutputDAHandle, DBL(OutputParams.Sampling));
      if Result <> OLNOERROR then Exit;

      Result := olDaSetDmaUsage(OutputDAHandle, dma);
      if Result <> OLNOERROR then Exit;

      //Set Channels and Gains
      for c := 0 to OutputParams.ChannelCount - 1 do begin

        Result := olDaSetChannelListEntry(OutputDAHandle, c, OutputParams.Channels[c]);
        if Result <> OLNOERROR then Exit;

      end;

      //Set ChannelCount
      Result := olDaSetChannelListSize(OutputDAHandle, OutputParams.ChannelCount);
      if Result <> OLNOERROR then Exit;

      //Set the Notification procedure
      Result := olDaSetNotificationProcedure(OutputDAHandle, OutputDeviceCallBack, 0);
      if Result <> OLNOERROR then Exit;

      //Allocate Buffers
      OutputElemSize   := OutputParams.ChannelCount * (((OutputParams.Bits - 1) div 8) + 1);
      OutputBufferSize := ((Round((OutputParams.BufferLength / 1000) * OutputParams.Sampling) + (512 - 1)) div 512) * 512;
      if OutputBufferSize > ONE_MEGABYTE then OutputBufferSize := ONE_MEGABYTE;        //But not larger than one MegaSamples

      for i := 1 to 2 do begin

        Result := olDmCallocBuffer(GPTR, 0, OutputBufferSize, OutputElemSize, OutputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

        Result := olDaPutBuffer(OutputDAHandle, OutputBufferHandle[i]);
        if Result <> OLNOERROR then Exit;

        Count := OutputBufferSize;
        try
          olDmGetBufferPtr(OutputBufferHandle[i], Buffer);
          OutputUserCallBack(OutputUserCaller, Buffer, OutputElemSize, Count, OutputError);
        except
        end;

      end;

      //Config
      Result := olDaConfig(OutputDAHandle);
      if Result <> OLNOERROR then Exit;


      //**********************************************************************//
      //                                                                      //
      //                             D I G I T A L                            //
      //                                                                      //
      //**********************************************************************//

      //For now let's not do digital

      OutputParams.DigitalChannelCount := 0;

      if OutputParams.DigitalChannelCount > 0 then begin
        //do it when in OutputStart ...
      end;

      r_o := 0;
      //************************************************************************


      //*********************
      InputStopped  := False;
      OutputStopped := False;
      //*********************


      sstart := 0;
      if olDaGetSSCaps(InputADHandle, OLSSC_SUP_SIMULTANEOUS_START, sstart) = OLNOERROR then begin
        if sstart <> 0 then begin
          sstart := 0;
          if olDaGetSSCaps(OutputDAHandle, OLSSC_SUP_SIMULTANEOUS_START, sstart) = OLNOERROR then begin
            if sstart <> 0 then begin
              sstart := 0;
              //Can do simultaneous start
              if olDaGetSSList(SSList) <> OLNOERROR then begin
                try
                  if olDaPutDassToSSList(SSList, InputADHandle) <> OLNOERROR then begin
                  end else if olDaPutDassToSSList(SSList, OutputDAHandle) <> OLNOERROR then begin
                  end else if olDaSimultaneousPrestart(SSList) <> OLNOERROR then begin
                  end else if olDaSimultaneousStart(SSList) = OLNOERROR then begin
                    sstart := 1;
                  end;
                finally
                  olDaReleaseSSList(SSList);
                end;
              end;
            end;
          end;
        end;
      end;

      if sstart = 0 then begin
      //Could not do simultaneous start
        //Input
        r_i := 1;
        Result := olDaStart(InputADHandle);
        if Result <> OLNOERROR then Exit;
        r_i := 0;

        //Output
        r_o := 1;
        Result := olDaStart(OutputDAHandle);
        if Result <> OLNOERROR then Exit;
        r_o := 0;
      end;


    except

      Result := 1;

    end;


  finally


    if Result <> 0 then begin

      InputStop(InputDevice);
      OutputStop(OutputDevice);
      Result := 1;

    end;

  end;


end;



//Interfacing section:



//****************************************************************************//
//                                                                            //
//                           L I B R A R Y    N A M E                         //
//                                                                            //
//****************************************************************************//
function LibraryNameLength :Integer;
begin
  Result := Length(LIBRARY_NAME);
end;

procedure LibraryName(Name :PChar);
begin
  Move(PChar(LIBRARY_NAME)^, Name^, Length(LIBRARY_NAME));
end;



//****************************************************************************//
//                                                                            //
//                           D E V I C E    C O U N T                         //
//                                                                            //
//****************************************************************************//

//Device Count callback
function InputBoardEnum(lpszBrdName, lpszDriverName :PChar; lParam :LPARAM) :BOOL; stdcall;
begin
  InputBoardName[InputBoardCount] := lpszBrdName;
  Inc(InputBoardCount);
  Result := InputBoardCount < ACQ_MAX_DEVICES;//stop enumeration, if necessary
end;

function OutputBoardEnum(lpszBrdName, lpszDriverName :PChar; lParam :LPARAM) :BOOL; stdcall;
begin
  OutputBoardName[OutputBoardCount] := lpszBrdName;
  Inc(OutputBoardCount);
  Result := OutputBoardCount < ACQ_MAX_DEVICES;//stop enumeration, if necessary
end;

//Input
function InputDeviceCount :Integer;
begin

  //Initialize the input board count to 0
  InputBoardCount := 0;

  if ECODE(olDaEnumBoards(InputBoardEnum, 0)) <> OLNOERROR then begin
    Result := 0;
  end else begin
    Result := InputBoardCount;
  end;

end;

procedure InputDeviceName(Device :Integer; Name :PChar);
var
  n :String;
begin
  n := InputBoardName[Device];
  Move(PChar(n)^, Name^, Length(n));
  Name[Length(n)] := #0;
end;

//Output
function OutputDeviceCount :Integer;
begin

  //Initialize the output board count to 0
  OutputBoardCount := 0;

  if ECODE(olDaEnumBoards(OutputBoardEnum, 0)) <> OLNOERROR then begin
    Result := 0;
  end else begin
    Result := OutputBoardCount;
  end;

end;

procedure OutputDeviceName(Device :Integer; Name :PChar);
var
  n :String;
begin
  n := OutputBoardName[Device];
  Move(PChar(n)^, Name^, Length(n));
  Name[Length(n)] := #0;
end;



//****************************************************************************//
//                                                                            //
//                          C H A N N E L    C O U N T                        //
//                                                                            //
//****************************************************************************//

function MaxChannels(Device :Integer; oldc :OLDC; olss :OLSS) :Integer;
var
  h  :HDEV;
  hs :HDASS;
  m  :UINT;
  c  :UINT;
  i  :Integer;
begin
  Result := 0;
  if olDaInitialize(PChar(InputBoardName[Device]), h) = OLNOERROR then begin
    try
      if olDaGetDevCaps(h, oldc, m) = OLNOERROR then begin
        for i := 0 to m - 1 do begin
          if olDaGetDASS(h, olss, i, hs) = OLNOERROR then begin
            try
              c := 0;
              if olDaGetSSCaps(hs, OLSSC_NUMCHANNELS, c) = OLNOERROR then begin
                Result := c;
              end;
              Exit;
            finally
              olDaReleaseDASS(hs);
            end;
          end;
        end;
      end;
    finally
      olDaTerminate(h);
    end;
  end;
end;

//Input
function InputChannelCount(Device :Integer) :Integer;
begin
  Result := MaxChannels(Device, OLDC_ADELEMENTS, OLSS_AD);
end;

//Output
function OutputChannelCount(Device :Integer) :Integer;
begin
  Result := MaxChannels(Device, OLDC_DAELEMENTS, OLSS_DA);
end;

function OutputDigitalChannelCount(Device :Integer) :Integer;
begin
  Result := MaxChannels(Device, OLDC_DOUTELEMENTS, OLSS_DOUT);
end;


//****************************************************************************//
//                                                                            //
//                               D E F A U L T S                              //
//                                                                            //
//****************************************************************************//


//Do these !!!

//Input
function InputGain(Device, Channel :Integer) :Double;
begin
  Result := 1;
end;

function InputRange(Device :Integer) :Double;
begin
  Result := 20;
end;

function InputDifferential(Device :Integer) :Integer;
begin
  Result := Integer(True);
end;

function InputBits(Device :Integer) :Integer;
begin
  Result := 12;
end;

//Output
function OutputGain(Device, Channel :Integer) :Double;
begin
  Result := 1;
end;

function OutputRange(Device :Integer) :Double;
begin
  Result := 20;
end;

function OutputDifferential(Device :Integer) :Integer;
begin
  Result := Integer(True);
end;

function OutputBits(Device :Integer) :Integer;
begin
  Result := 12;
end;


//****************************************************************************//
//                                                                            //
//                               E X P O R T S                                //
//                                                                            //
//****************************************************************************//
exports


  LibraryInitialize,
  LibraryTerminate,

  LibraryNameLength,
  LibraryName,

  //Input
  InputStart,
  InputStop,

  InputDeviceCount,
  InputDeviceName,
  InputChannelCount,
  InputGain,
  InputRange,
  InputDifferential,
  InputBits,

  ReadDBLChannel,

  AIRead,
  DIRead,

  //Output
  OutputStart,
  OutputStop,

  OutputDeviceCount,
  OutputDeviceName,
  OutputChannelCount,
  OutputDigitalChannelCount,
  OutputGain,
  OutputRange,
  OutputDifferential,
  OutputBits,

  WriteDBLChannel,

  AOWrite,
  DOWrite,

  //Synch
  SynchStart;


end.
//******************************************************************************

