library qubacqni;
{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
//11-10-2000, 08-12-03, 10-12-03 (LM)

{
                         QUB data acquisition library interface
                              for National Instruments  Cards
                          - Each Device can be used separately -

Output file is QUBACQNI.DLL
  This library matches the generic QUB hardware interface to the National
Instruments NIDAQ32.DLL.  Nearly all included functions are exported.


--------------------------------------------------------
Summary of Usage of NI DLL functions by QUBACQNI library
--------------------------------------------------------
QUBACQNI export function    NI DLL functions used
--------------------------------------------------------
InputStart                  DAQ_Rate
                            DAQ_Clear
                            DAQ_DB_Config
                            Config_DAQ_Event_Message
                            DAQ_Start
                            SCAN_Setup
                            SCAN_Start
InputStop		    DAQ_Clear
InputDeviceCallBack	    Timeout_Config
ReadDBLChannel		    DAQ_DB_Config
AIRead								      Config_DAQ_Event_Message
DIRead								      DAQ_Start
						      		    	SCAN_Setup
OutputStart		    					SCAN_Start
		          							Config_DAQ_Event_Message
			      				    		DAQ_Clear
						    		      	DAQ_DB_Transfer
		          							AI_VRead
			  		        				AI_Read
		    						      	DIG_Line_Config
OutputStop					    		DIG_In_Line
      			    						Config_DAQ_Event_Message
		    			      				Timeout_Config
OutputDeviceCallBack		    WFM_DB_Config
        						      	DIG_Prt_Config
OutputDigitalDeviceCallBack	DIG_Out_Port
WriteDBLChannel			      	Config_DAQ_Event_Message
AOWrite         						WFM_Op
DOWrite					          	Config_DAQ_Event_Message
				              			WFM_Group_Control
SynchStart	        				AO_VWrite
														WFM_DB_HalfReady
														WFM_DB_Transfer
			 											DIG_Out_Port
														AO_VWrite
														AO_Write
		 												DIG_Line_Config
	 													DIG_Out_Line
														DAQ_Clear
														DAQ_Rate
													  Timeout_Config
														DAQ_DB_Config
 														Config_DAQ_Event_Message
 														Timeout_Config
LibraryInitialize		      	WFM_DB_Config
}


uses


SysUtils, Windows,

AllocTypes,

nidaq,                      // NI dll interface (from NI)

QUB_Global,                 // Included for definition of ONE_MEGABYTE
QUB_Acquisition_Global    { Included only for definitions of
                            TQUB_Acquisition_InputCallBack
                            TQUB_Acquisition_OutputCallBack
                            TQUB_Acquisition_DigitalOutputCallBack
                            TQUB_Acquisition_InputStart_Params = record
                            TQUB_Acquisition_OutputStart_Params = record
                            const ACQ_MAX_DEVICES = 16;
                            const ACQ_MAX_CHANNELS = 16;
                            }
;

{$R *.RES}


{$INCLUDE nidaqcns.pas}


const


LIBRARY_NAME                       = 'QUB NIDAQ 2.0';

DBmodeON                           :SmallInt = 1;
DBmodeOFF                          :SmallInt = 0;


var


//Local variables, set by Initialize/Terminate
DeviceCount                        :Integer = 0;
Devices                            :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
DeviceCodes                        :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
DeviceNames                        :Array[0..ACQ_MAX_DEVICES - 1] of String;

//Input
IChannelCount                      :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
IElemSize                          :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
IBufferSize                        :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
IBufferSize2                       :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
IBuffer                            :Array[0..ACQ_MAX_DEVICES - 1] of Pointer;
IBuffer2                           :Array[0..ACQ_MAX_DEVICES - 1] of Pointer;
IRunning                           :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;
IUserCallBack                      :Array[0..ACQ_MAX_DEVICES - 1] of TQUB_Acquisition_InputCallBack;
IUserCaller                        :Array[0..ACQ_MAX_DEVICES - 1] of TObject;
IError                             :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
IDAQStopped                        :Array[0..ACQ_MAX_DEVICES - 1] of i16;
IPointsTfr                         :Array[0..ACQ_MAX_DEVICES - 1] of u32;
IChStr                             :Array[0..ACQ_MAX_DEVICES - 1] of String;
ICalling                           :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;


//Output
OChannelCount                      :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OFirstChannel                      :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
ODChannelCount                     :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OElemSize                          :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OBufferSize                        :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OBufferSize2                       :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OBuffer                            :Array[0..ACQ_MAX_DEVICES - 1] of Pointer;
OBuffer2                           :Array[0..ACQ_MAX_DEVICES - 1] of Pointer;
ORunning                           :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;
OUserCallBack                      :Array[0..ACQ_MAX_DEVICES - 1] of TQUB_Acquisition_OutputCallBack;
OUserCaller                        :Array[0..ACQ_MAX_DEVICES - 1] of TObject;
OError                             :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OChStr                             :Array[0..ACQ_MAX_DEVICES - 1] of String;
ODChStr                            :Array[0..ACQ_MAX_DEVICES - 1] of String;
OChanVector                        :Array[0..ACQ_MAX_DEVICES - 1, 0..ACQ_MAX_CHANNELS - 1] of i16;
OCalling                           :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;
ODValue                            :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
ODCount                            :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
ODError                            :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
ODUserCallBack                     :Array[0..ACQ_MAX_DEVICES - 1] of TQUB_Acquisition_DigitalOutputCallBack;
ODCalling                          :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;
{$ifdef DEBUG}
fLog : TextFile;
{$endif}

//Export functions
function LibraryInitialize :Integer; stdcall; forward;
function LibraryTerminate :Integer; stdcall; forward;

function LibraryNameLength :Integer; stdcall; forward;
procedure LibraryName(Name :PChar); stdcall; forward;

//Input
function InputStart(Device :Integer; var InputParams :TQUB_Acquisition_InputStart_Params) :Integer; stdcall; forward;
function InputStop(Device :Integer) :Integer; stdcall; forward;

function InputDeviceCount :Integer; stdcall; forward;
procedure InputDeviceName(Device :Integer; Name :PChar); stdcall; forward;
function InputDeviceCode(Device :Integer):Integer; stdcall; forward;
function InputChannelCount(Device :Integer) :Integer; stdcall; forward;
function InputGain(Device, Channel :Integer) :Double; stdcall; forward;
function InputRange(Device :Integer) :Double; stdcall; forward;
function InputDifferential(Device :Integer) :Integer; stdcall; forward;
function InputBits(Device :Integer) :Integer; stdcall; forward;

function ReadDBLChannel(Device :Integer; Channel :Integer; Gain :Double; var Value :Double) :Integer; stdcall; forward;

function AIRead(Device :Integer; Channel :Integer; Gain :Double; var Value :Integer) :Integer; stdcall; forward;
function DIRead(Device :Integer; Channel :Integer; var Value :Integer) :Integer; stdcall; forward;

procedure InputDeviceCallBack(handle :HWND; Msg :UINT; wParam :WPARAM; lParam :LPARAM); cdecl; forward;

//Output
function OutputStart(Device :Integer; var OutputParams :TQUB_Acquisition_OutputStart_Params) :Integer; stdcall; forward;
function OutputStop(Device :Integer) :Integer; stdcall; forward;

function OutputDeviceCount :Integer; stdcall; forward;
procedure OutputDeviceName(Device :Integer; Name :PChar); stdcall; forward;
function OutputChannelCount(Device :Integer) :Integer; stdcall; forward;
function OutputDigitalChannelCount(Device :Integer) :Integer; stdcall; forward;
function OutputGain(Device, Channel :Integer) :Double; stdcall; forward;
function OutputRange(Device :Integer) :Double; stdcall; forward;
function OutputDifferential(Device :Integer) :Integer; stdcall; forward;
function OutputBits(Device :Integer) :Integer; stdcall; forward;

function WriteDBLChannel(Device :Integer; Channel :Integer; const Value :Double) :Integer; stdcall; forward;

function AOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall; forward;
function DOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall; forward;

procedure OutputDeviceCallBack(handle :HWND; Msg :UINT; wParam :WPARAM; lParam :LPARAM); cdecl; forward;
procedure OutputDigitalDeviceCallBack(handle :HWND; Msg :UINT; wParam :WPARAM; lParam :LPARAM); cdecl; forward;


//Synch
function SynchStart(InputDevice, OutputDevice :Integer; var InputParams :TQUB_Acquisition_InputStart_Params; var OutputParams :TQUB_Acquisition_OutputStart_Params; var r_i, r_o :Integer) :Integer; stdcall; forward;


//----- The following methods are local - not exported in DLL interface.
function Code2Name(Code :Integer) :String; forward;
function NI_To_Device(NIDevice :Integer) :Integer; forward;
function NI_TimeBaseToResolution(TimeBase :i16) :Double; forward;
procedure IFreeBuff(Device :Integer); forward;
procedure OFreeBuff(Device :Integer); forward;

//****************************************************************************//
//                                                                            //
//                          A C Q U I S I T I O N                             //
//                                                                            //
//****************************************************************************//

//Input ************************************************************************
function InputStart(Device :Integer; var InputParams :TQUB_Acquisition_InputStart_Params) :Integer;
var
  ChanVector :Array[0..ACQ_MAX_CHANNELS - 1] of i16;
  GainVector :Array[0..ACQ_MAX_CHANNELS - 1] of i16;
  SampTB     :SmallInt;
  SampInt    :SmallInt;
  c          :Integer;
  {$ifdef DEBUG}
  i1,i2,i3 : Integer;
  {$endif}
begin


  Result := 1;


  if IRunning[Device]  then Exit;


  ICalling[Device]      := False;
  IUserCallBack[Device] := InputParams.CallBack;
  IUserCaller[Device]   := InputParams.Caller;

  IChannelCount[Device] := InputParams.ChannelCount;

  IChStr[Device] := '';
  for c := 0 to InputParams.ChannelCount - 1 do begin
    ChanVector[c] := InputParams.Channels[c];
    try
      if (InputParams.Gain[c] > 0.49) and (InputParams.Gain[c] < 0.51) Then
        GainVector[c] := -1
      else
        GainVector[c] := Round(InputParams.Gain[c]);
    except
      GainVector[c] := 1;
    end;
    IChStr[Device] := IChStr[Device] + 'AI' + IntToStr(ChanVector[c]);
    if c < (InputParams.ChannelCount - 1) then begin
      IChStr[Device] := IChStr[Device] + ',';
    end;
  end;


  //Free Buffers (just in case)
  IFreeBuff(Device);


  try

    try


      //Reset DAQ **************************************************************
      DAQ_Clear(Devices[Device]);
      {$ifdef DEBUG}
        WriteLn( fLog, Format('DAQ_Clear(%d)',[Devices[Device]]));
      {$endif}


      //Set the Sampling
      Result := DAQ_Rate(InputParams.ChannelCount * InputParams.Sampling, 0, @SampTB, @SampInt);
      {$ifdef DEBUG}
        i1:=InputParams.ChannelCount;
        i2:=Round(InputParams.Sampling);
        i3:=Round(InputParams.ChannelCount*InputParams.Sampling);
        WriteLn( fLog, Format('DAQ_Rate(%d*%d=%d,0,->%d,->%d)->%d',[i1,i2,i3,SampTB, SampInt, Result]));
      {$endif}
      if Result <> 0 then Exit;


      //Allocate Buffers
      try
        IElemSize[Device] := InputParams.ChannelCount * (((InputParams.Bits - 1) div 8) + 1);
        if InputParams.BufferCount = 0 then begin
          IBufferSize[Device] := ((Round((InputParams.BufferLength / 1000) * InputParams.Sampling) + (512 - 1)) div 512) * 512;
          if IBufferSize[Device] > ONE_MEGABYTE then IBufferSize[Device] := ONE_MEGABYTE; //But not larger than one MegaSamples
        end else begin
          IBufferSize[Device] := InputParams.BufferCount;//in samples
        end;
        if InputParams.SingleBuffer then begin
          IBufferSize2[Device] := IBufferSize[Device];
        end else begin
          IBufferSize2[Device] := 2 * IBufferSize[Device];//in samples
        end;
        // There might be an issue with buffersize overruns.  Adding some (1024)
        // padding seems to reduce (eliminate?) the problem.  Possibly related to
        // buffer overruns too, as caused by debugging and stopping while Acquiring
        // data.  03-18-2004 {JB}
        IBuffer[Device]            := AllocMem(1024+IBufferSize[Device]  * IElemSize[Device]);
        IBuffer2[Device]           := AllocMem(1024+IBufferSize2[Device] * IElemSize[Device]);
        InputParams.BufferCount    := IBufferSize2[Device];//in samples
        InputParams.Buffer         := IBuffer2[Device];
        InputParams.BufferElemSize := IElemSize[Device] div InputParams.ChannelCount;
      except
        Result := 1;
        Exit;
      end;


      //Set TimeOut - for synchronous operations
      Timeout_Config(Devices[Device], 1091{60 secs});
      {$ifdef DEBUG}
        WriteLn( fLog, Format('Timeout_Config(%d,1091=60 secs)',[Devices[Device]]));
      {$endif}


      //Set the Double Buffer mode ON
      if InputParams.SingleBuffer then begin
        Result := DAQ_DB_Config(Devices[Device], DBmodeOFF);
        {$ifdef DEBUG}
          WriteLn( fLog, Format('DAQ_DB_Config(%d,DBmodeOFF)->%d',[Devices[Device],Result]));
        {$endif}
      end else begin
        Result := DAQ_DB_Config(Devices[Device], DBmodeON);
        {$ifdef DEBUG}
          WriteLn( fLog, Format('DAQ_DB_Config(%d,DBmodeON)->%d',[Devices[Device],Result]));
        {$endif}
      end;
      if Result <> 0 then Exit;


      //Set the Event trapping
      if @IUserCallBack[Device] <> nil then begin
        Result := Config_DAQ_Event_Message(Devices[Device], 1{Add}, PChar(IChStr[Device]), 1, IBufferSize[Device], 0, 0, 0, 0, 0, 0, Cardinal(@InputDeviceCallback));
        {$ifdef DEBUG}
          WriteLn( fLog, Format('Config_DAQ_Event_Message(%d, 1{Add}, "%s", 1, %d, 0, 0, 0, 0, 0, 0, [ADDRESS])',[Devices[Device],IChStr[Device],IBufferSize[Device]]));
        {$endif}
        if Result <> 0 then Exit;
      end;

      //*********************
      IRunning[Device] := True;
      //*********************


      //Start Acquisition
      if InputParams.ChannelCount = 1 then begin
      //Single Channel

        Result := DAQ_Start(Devices[Device], ChanVector[0], GainVector[0], IBuffer2[Device], IBufferSize2[Device], SampTB, SampInt);
        {$ifdef DEBUG}
          WriteLn( fLog, Format('DAQ_Start(%d, %d, %d, [BUFFER], %d, %d, %d)->%d',[Devices[Device],ChanVector[0],GainVector[0],IBufferSize2[Device],SampTB,SampInt,Result]));
        {$endif}
        if Result <> 0 then Exit;

      end else begin
      //Multiple Channels

        Result := SCAN_Setup(Devices[Device], InputParams.ChannelCount, @ChanVector[0], @GainVector[0]);
        {$ifdef DEBUG}
          WriteLn( fLog, Format('SCAN_Setup(%d,%d, [%d,%d,%d,...],[%d,%d,%d,...])->%d',[Devices[Device],InputParams.ChannelCount,ChanVector[0],ChanVector[1],ChanVector[2],GainVector[0],GainVector[1],GainVector[2],Result]));
        {$endif}
        if Result <> 0 then Exit;

        Result := SCAN_Start(Devices[Device], IBuffer2[Device], IBufferSize2[Device] * InputParams.ChannelCount, sampTB, sampInt, SampTB, 0);
        {$ifdef DEBUG}
          i1 := IBufferSize2[Device] * InputParams.ChannelCount;
          WriteLn( fLog, Format('SCAN_Start(%d, [BUFFER], %d,%d,%d,%d,0)->%d', [Devices[Device],i1, sampTB, sampInt, SampTB,Result]));
        {$endif}
        if Result <> 0 then Exit;

      end;


    except

      Result := 1;

    end;


  finally


    if Result <> 0 then begin

      InputStop(Devices[Device]);
      Result := 1;

    end;


  end;

end;

function InputStop(Device :Integer) :Integer;
begin

  Result := 0;
  IRunning[Device] := False;

  try

    while ICalling[Device] do begin
    end;
    if @IUserCallBack[Device] <> nil then begin
      Config_DAQ_Event_Message(Devices[Device], 2, PChar(IChStr[Device]), 1, IBufferSize[Device], 0, 0, 0, 0, 0, 0, Cardinal(@InputDeviceCallback));
    end;
    DAQ_Clear(Devices[Device]);
    {$ifdef DEBUG}
    WriteLn( fLog, Format('DAQ_Clear(%d)',[Devices[Device]]));
    {$endif}

    IFreeBuff(Device);

  except
  end;

end;


//****************************************************************************//
//                                                                            //
//                               C A L L B A C K                              //
//                                                                            //
//****************************************************************************//
procedure InputDeviceCallBack(handle :HWND; Msg :UINT; wParam :WPARAM; lParam :LPARAM);
var
  Device :Integer;
begin

  Device := NI_To_Device(wParam and $00FF);

  ICalling[Device] := True;

  try


    //*********************
    if not IRunning[Device] then Exit;
    //*********************

    IError[Device] := DAQ_DB_Transfer(Devices[Device], IBuffer[Device], @IPointsTfr[Device], @IDAQStopped[Device]);

    try
      if @IUserCallBack[Device] <> nil then begin
        IUserCallBack[Device](IUserCaller[Device], IBuffer[Device], IElemSize[Device], (IPointsTfr[Device] div u32(IChannelCount[Device])), Error);
      end;
    except
    end;

  finally

    ICalling[Device] := False;

  end;

end;

function ReadDBLChannel(Device :Integer; Channel :Integer; Gain :Double; var Value :Double) :Integer;
begin
  Result := AI_VRead(Devices[Device], Channel, Round(Gain), @Value);
end;

function AIRead(Device :Integer; Channel :Integer; Gain :Double; var Value :Integer) :Integer;
var
  i :I16;
begin
  //AI_Configure(Devices[Device], Channel, 0{inputMode}, 10{inputRange}, 0{polarity}, 0{driveAIS});
  Result := AI_Read(Devices[Device], Channel, Round(Gain), @i);
  Value  := i;
end;

function DIRead(Device :Integer; Channel :Integer; var Value :Integer) :Integer;
var
  i :I16;
begin
  DIG_Line_Config(Devices[Device], 0{Port}, Channel{Line}, 0{dir});
  Result := DIG_In_Line(Devices[Device], 0{Port}, Channel{Line}, @i);
  Value  := i;
end;

//Output ***********************************************************************
function OutputStart(Device :Integer; var OutputParams :TQUB_Acquisition_OutputStart_Params) :Integer;
var
  DigVal     :Integer;
  Count      :Integer;
  TBase      :i16;
  UpdateInt  :u32;
  c          :Integer;
begin

  Result := 1;

  if ORunning[Device] then Exit;


  OCalling[Device]                   := False;
  ODCalling[Device]                  := False;
  OUserCallBack[Device]              := OutputParams.CallBack;
  ODUserCallBack[Device]             := OutputParams.DigitalCallBack;
  OUserCaller[Device]                := OutputParams.Caller;


  OChannelCount[Device]              := OutputParams.ChannelCount;
  OFirstChannel[Device]              := OutputParams.Channels[0];
  ODChannelCount[Device]             := OutputParams.DigitalChannelCount;

  ODChStr[Device]                    := 'AO' + IntToStr(OFirstChannel[Device]);

  OChStr[Device] := '';
  for c := 0 to OutputParams.ChannelCount - 1 do begin
    OChanVector[Device][c] := OutputParams.Channels[c];
    OChStr[Device] := OChStr[Device] + 'AO' + IntToStr(OChanVector[Device][c]);
    if c < (OutputParams.ChannelCount - 1) then begin
      OChStr[Device] := OChStr[Device] + ',';
    end;
  end;


  //Free Buffers (just in case)
  OFreeBuff(Device);


  try

    try


      Config_DAQ_Event_Message(Devices[Device], 0, nil, 0, 0, 0, 0, 0, 0, 0, 0, 0);


      //Set TimeOut - for synchronous operations
      Timeout_Config(Devices[Device], 1091{60 secs});


      //Set the Double Buffer mode ON/OFF, according to the user Params
      if OutputParams.SingleBuffer then begin
      //Single Buffer

        Result := WFM_DB_Config(Devices[Device], OutputParams.ChannelCount, @OChanVector[Device][0], DBModeOFF, 0{iOldDataStop}, 0{iPartialTransferStop});

      end else begin
      //Double Buffer

        //Result := WFM_DB_Config(Devices[Device], OutputParams.ChannelCount, @OChanVector[Device][0], DBModeON,  1{iOldDataStop}, 0{iPartialTransferStop});

        //Testing
        Result := WFM_DB_Config(Devices[Device], OutputParams.ChannelCount, @OChanVector[Device][0], DBModeON,  0{iOldDataStop}, 0{iPartialTransferStop});
        {
        iOldDataStop = 0 works with AcqOutput, but also works with DynClamp code, with DoubleBuffer (it doesn't, otherwise)
        It's not clear yet what the effects are with AcqOutput
        }
        //.......

      end;
      if Result <> 0 then Exit;


      //Allocate Buffers
      try
        OElemSize[Device]    := OutputParams.ChannelCount * (((OutputParams.Bits - 1) div 8) + 1);
        if OutputParams.BufferCount = 0 then begin
          OBufferSize[Device]  := ((Round((OutputParams.BufferLength / 1000) * OutputParams.Sampling) + (512 - 1)) div 512) * 512;
          if OBufferSize[Device] > ONE_MEGABYTE then OBufferSize[Device] := ONE_MEGABYTE;                 //But not larger than one MegaSamples
        end else begin
          OBufferSize[Device]  := OutputParams.BufferCount;//in samples
        end;
        if OutputParams.SingleBuffer then begin
          OBufferSize2[Device] := OBufferSize[Device];
        end else begin
          OBufferSize2[Device] := 2 * OBufferSize[Device];//in samples
        end;
        OBuffer[Device]             := AllocMem(OBufferSize[Device]  * OElemSize[Device]);
        OBuffer2[Device]            := AllocMem(OBufferSize2[Device] * OElemSize[Device]);
        OutputParams.Buffer         := OBuffer2[Device];
        OutputParams.BufferCount    := OBufferSize2[Device];//in samples
        OutputParams.BufferElemSize := OElemSize[Device] div OutputParams.ChannelCount;
      except
        OFreeBuff(Device);
        Result := 1;
        Exit;
      end;


      //Load the first two buffers with data
      Count := OBufferSize2[Device];//samples
      try
        if @OUserCallBack[Device] <> nil then begin
          OUserCallBack[Device](OUserCaller[Device], OBuffer2[Device], OElemSize[Device], Count, OError[Device]);
        end else begin
          FillChar(OBuffer2[Device]^, Count * OElemSize[Device], 0);
        end;
      except
      end;



      //**********************************************************************//
      //                                                                      //
      //                             D I G I T A L                            //
      //                                                                      //
      //**********************************************************************//
      if OutputParams.DigitalChannelCount > 0 then begin

        if @ODUserCallBack[Device] <> nil then begin

          ODCount[Device] := Round(OutputParams.Sampling / OutputParams.DigitalSampling);
          if (ODCount[Device] < 1)   then ODCount[Device] := 1;
          case ODCount[Device] of
            1:        ODCount[Device] := 1;
            2:        ODCount[Device] := 2;
            3..4:     ODCount[Device] := 4;
            5..8:     ODCount[Device] := 8;
            9..16:    ODCount[Device] := 16;
            17..32:   ODCount[Device] := 32;
            33..64:   ODCount[Device] := 64;
            65..128:  ODCount[Device] := 128;
            129..256: ODCount[Device] := 256;
            257..512: ODCount[Device] := 512;
            else  ODCount[Device] := ((ODCount[Device] + (512 - 1)) div 512) * 512;
          end;
          ODUserCallBack[Device](OUserCaller[Device], ODValue[Device], ODCount[Device], ODError[Device]);
          DigVal := ODValue[Device];
          for c := 0 to OutputParams.DigitalChannelCount - 1 do begin
            Result := DIG_Prt_Config(Devices[Device], c, 0, 1);
            Result := DIG_Out_Port(Devices[Device], c, DigVal and $ff);
            DigVal := DigVal shr 8;
          end;
          //Set the Event trapping
          Result := Config_DAQ_Event_Message(Devices[Device], 1{Add}, PChar(ODChStr[Device]), 1, ODCount[Device], 0, 0, 0, 0, 0, 0, Cardinal(@OutputDigitalDeviceCallback));

        end;

      end;



      //Set the Event trapping
      if @OUserCallBack[Device] <> nil then begin

        Result := Config_DAQ_Event_Message(Devices[Device], 1{Add}, PChar(OChStr[Device]), 1, OBufferSize[Device], 0, 0, 0, 0, 0{Handle}, 0{Message}, Cardinal(@OutputDeviceCallback));
        if Result <> 0 then Exit;

      end;



      //***************
      ORunning[Device] := True;
      //***************


      //Find out what the user-provided Sampling actually translates into
      Result := WFM_Rate(OutputParams.Sampling, 0{units; pts/sec}, @TBase, @UpdateInt);
      OutputParams.Sampling := 1/(NI_TimebaseToResolution(TBase) * UpdateInt);


      //Start Acquisition
      Result := WFM_Op(Devices[Device], OutputParams.ChannelCount, @OChanVector[Device][0], OBuffer2[Device], OBufferSize2[Device] * OutputParams.ChannelCount, 0, OutputParams.Sampling);
      if Result <> 0 then Exit;


    except

      Result := 1;

    end;


  finally


    if Result <> 0 then begin

      OutputStop(Devices[Device]);
      //Result := 1;

    end;


  end;

end;

function OutputStop(Device :Integer) :Integer;
var
  i :Integer;
begin

  Result := 0;
  ORunning[Device] := False;

  try

    //Analog
    while OCalling[Device] do begin
    end;
    if @OUserCallBack[Device] <> nil then begin
      Result := Config_DAQ_Event_Message(Devices[Device], 2, PChar(OChStr[Device]), 1, OBufferSize[Device], 0, 0, 0, 0, 0, 0, Cardinal(@OutputDeviceCallback));
    end;

    //Digital
    while ODCalling[Device] do begin
    end;
    if @ODUserCallBack[Device] <> nil then begin
      Result := Config_DAQ_Event_Message(Devices[Device], 2, PChar(ODChStr[Device]), 1, ODCount[Device], 0, 0, 0, 0, 0, 0, Cardinal(@OutputDigitalDeviceCallback));
    end;

    Result := WFM_Group_Control(Devices[Device], 1{Group}, 0{Clear});
    for i := 0 to OChannelCount[Device] - 1 do begin
      Result := AO_VWrite(Devices[Device], OChanVector[Device][i], 0.0);
    end;

    OFreeBuff(Device);

  except
  end;

end;



//****************************************************************************//
//                                                                            //
//                               C A L L B A C K                              //
//                                                                            //
//****************************************************************************//
procedure OutputDeviceCallBack(handle :HWND; Msg :UINT; wParam :WPARAM; lParam :LPARAM);
var
  Device :Integer;
  Count  :Integer;
  hReady :i16;
  Result :Integer;
begin


  Device := NI_To_Device(wParam and $00FF);

  OCalling[Device] := True;

  try


    //*************************
    if not ORunning[Device] then Exit;
    //*************************


    Result := WFM_DB_HalfReady(Devices[Device], OChannelCount[Device], @OChanVector[Device][0], @hReady);

    if (hReady <> 0) and (Result = 0) then begin

      Count := OBufferSize[Device];
      try
        if @OUserCallBack[Device] <> nil then begin
          OUserCallBack[Device](OUserCaller[Device], OBuffer[Device], OElemSize[Device], Count, OError[Device]);
        end;
      except
      end;

      OError[Device] := WFM_DB_Transfer(Devices[Device], OChannelCount[Device], @OChanVector[Device][0], OBuffer[Device], Count * OChannelCount[Device]);

    end;


  finally

    OCalling[Device] := False;

  end;

end;


//****************************************************************************//
//                                                                            //
//                     D I G I T A L     C A L L B A C K                      //
//                                                                            //
//****************************************************************************//
procedure OutputDigitalDeviceCallBack(handle :HWND; Msg :UINT; wParam :WPARAM; lParam :LPARAM);
var
  Device :Integer;
  DigVal :Integer;
  c      :Integer;
begin

  Device := NI_To_Device(wParam and $00FF);

  ODCalling[Device] := True;

  try


    //*************************
    if not ORunning[Device] then Exit;
    //*************************


    try
      //Get new values
      if @ODUserCallBack[Device] <> nil then begin
        ODUserCallBack[Device](OUserCaller[Device], ODValue[Device], ODCount[Device], ODError[Device]);
        //Write new value
        DigVal := ODValue[Device];
        for c := 0 to ODChannelCount[Device] - 1 do begin
          ODError[Device] := DIG_Out_Port(Devices[Device], c, DigVal and $ff);
          DigVal := DigVal shr 8;
        end;
      end;
    except
    end;

  finally

    ODCalling[Device] := False;

  end;

end;


function WriteDBLChannel(Device :Integer; Channel :Integer; const Value :Double) :Integer;
begin
  Result := AO_VWrite(Devices[Device], Channel, Value);
end;

function AOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer;
begin
  //AO_Configure(Devices[Device], Channel, 0{outputPolarity}, 0{intOrExtRef}, 10{refVoltage}, 0{updateMode});
  Result := AO_Write(Devices[Device], Channel, Value);
end;

function DOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer;
begin
  DIG_Line_Config(Devices[Device], 0{Port}, Channel{Line}, 1{dir});
  Result := DIG_Out_Line(Devices[Device], 0{Port}, Channel{Line}, Value);
end;


//Synch
function SynchStart(InputDevice, OutputDevice :Integer; var InputParams :TQUB_Acquisition_InputStart_Params; var OutputParams :TQUB_Acquisition_OutputStart_Params; var r_i, r_o :Integer) :Integer;
var
  IChanVector :Array[0..ACQ_MAX_CHANNELS - 1] of i16;
  IGainVector :Array[0..ACQ_MAX_CHANNELS - 1] of i16;
  ISampTB     :SmallInt;
  ISampInt    :SmallInt;
  DigVal      :Integer;
  Count       :Integer;
  c           :Integer;

begin

  Result := 1;
  r_i    := 1;
  r_o    := 1;


  if IRunning[InputDevice]  then Exit;
  if ORunning[OutputDevice] then Exit;


  r_i    := 0;
  r_o    := 0;


  //Input
  ICalling[InputDevice]      := False;
  IUserCallBack[InputDevice] := InputParams.CallBack;
  IUserCaller[InputDevice]   := InputParams.Caller;
  IChannelCount[InputDevice] := InputParams.ChannelCount;


  IChStr[InputDevice] := '';
  for c := 0 to InputParams.ChannelCount - 1 do begin
    IChanVector[c] := InputParams.Channels[c];
    try
      if (InputParams.Gain[c] >= 0.49) and (InputParams.Gain[c] >= 0.51) Then
        IGainVector[c] := -1
      else
        IGainVector[c] := Round(InputParams.Gain[c]);

    except
      IGainVector[c] := 1;
    end;

    IChStr[InputDevice] := IChStr[InputDevice] + 'AI' + IntToStr(IChanVector[c]);
    if c < (InputParams.ChannelCount - 1) then begin
      IChStr[InputDevice] := IChStr[InputDevice] + ',';
    end;
  end;


  //Output
  OCalling[OutputDevice]             := False;
  ODCalling[OutputDevice]            := False;
  OUserCallBack[OutputDevice]        := OutputParams.CallBack;
  ODUserCallBack[OutputDevice]       := OutputParams.DigitalCallBack;
  OUserCaller[OutputDevice]          := OutputParams.Caller;


  OChannelCount[OutputDevice]        := OutputParams.ChannelCount;
  OFirstChannel[OutputDevice]        := OutputParams.Channels[0];
  ODChannelCount[OutputDevice]       := OutputParams.DigitalChannelCount;

  ODChStr[OutputDevice]              := 'AO' + IntToStr(OFirstChannel[OutputDevice]);

  OChStr[OutputDevice] := '';
  for c := 0 to OutputParams.ChannelCount - 1 do begin
    OChanVector[OutputDevice][c] := OutputParams.Channels[c];
    OChStr[OutputDevice] := OChStr[OutputDevice] + 'AO' + IntToStr(OChanVector[OutputDevice][c]);
    if c < (OutputParams.ChannelCount - 1) then begin
      OChStr[OutputDevice] := OChStr[OutputDevice] + ',';
    end;
  end;


  //Free Buffers (just in case)
  IFreeBuff(InputDevice);
  OFreeBuff(OutputDevice);


  try

    try


      //Input ******************************************************************
      r_i := 1;

      //Reset DAQ **************************************************************
      DAQ_Clear(Devices[InputDevice]);


      //Set the Sampling
      Result := DAQ_Rate(InputParams.ChannelCount * InputParams.Sampling, 0, @ISampTB, @ISampInt);
      if Result <> 0 then Exit;


      //Allocate Buffers
      try
        IElemSize[InputDevice]   := InputParams.ChannelCount * (((InputParams.Bits - 1) div 8) + 1);
        if InputParams.BufferCount = 0 then begin
          IBufferSize[InputDevice] := ((Round((InputParams.BufferLength / 1000) * InputParams.Sampling) + (512 - 1)) div 512) * 512;
          if IBufferSize[InputDevice] > ONE_MEGABYTE then IBufferSize[InputDevice] := ONE_MEGABYTE;           //But not larger than one MegaSamples
        end else begin
          IBufferSize[InputDevice] := InputParams.BufferCount;//in samples
        end;
        if InputParams.SingleBuffer then begin
          IBufferSize2[InputDevice] := IBufferSize[InputDevice];
        end else begin
          IBufferSize2[InputDevice] := 2 * IBufferSize[InputDevice];//in samples
        end;
        IBuffer[InputDevice]       := AllocMem(IBufferSize[InputDevice]  * IElemSize[InputDevice]);
        IBuffer2[InputDevice]      := AllocMem(IBufferSize2[InputDevice] * IElemSize[InputDevice]);
        InputParams.BufferCount    := IBufferSize2[InputDevice];//in samples
        InputParams.Buffer         := IBuffer2[InputDevice];
        InputParams.BufferElemSize := IElemSize[InputDevice] div InputParams.ChannelCount;
      except
        Result := 1;
        Exit;
      end;


      //Set TimeOut - for synchronous operations
      Timeout_Config(Devices[InputDevice], 1091{60 secs});


      //Set the Double Buffer mode ON
      if InputParams.SingleBuffer then begin
        Result := DAQ_DB_Config(Devices[InputDevice], DBmodeOFF);
      end else begin
        Result := DAQ_DB_Config(Devices[InputDevice], DBmodeON);
      end;
      if Result <> 0 then Exit;


      //Set the Event trapping
      if @IUserCallBack[InputDevice] <> nil then begin
        Result := Config_DAQ_Event_Message(Devices[InputDevice], 1{Add}, PChar(IChStr[InputDevice]), 1, IBufferSize[InputDevice], 0, 0, 0, 0, 0, 0, Cardinal(@InputDeviceCallback));
        if Result <> 0 then Exit;
      end;

      r_i := 0;

      //************************************************************************



      //Output *****************************************************************
      r_o := 1;

      Config_DAQ_Event_Message(Devices[OutputDevice], 0, nil, 0, 0, 0, 0, 0, 0, 0, 0, 0);


      //Set TimeOut - for synchronous operations
      Timeout_Config(Devices[OutputDevice], 1091{60 secs});


      //Set the Double Buffer mode ON
      if OutputParams.SingleBuffer then begin
        Result := WFM_DB_Config(Devices[OutputDevice], OutputParams.ChannelCount, @OChanVector[OutputDevice][0], DBModeOFF, 0{iOldDataStop}, 0{iPartialTransferStop});
      end else begin
        Result := WFM_DB_Config(Devices[OutputDevice], OutputParams.ChannelCount, @OChanVector[OutputDevice][0], DBModeON,  1{iOldDataStop}, 0{iPartialTransferStop});
        //****************//
        //   See OStart   //
        //****************//
      end;
      if Result <> 0 then Exit;


      //Allocate Buffers
      try
        OElemSize[OutputDevice]    := OutputParams.ChannelCount * (((OutputParams.Bits - 1) div 8) + 1);
        if OutputParams.BufferCount = 0 then begin
          OBufferSize[OutputDevice]  := ((Round((OutputParams.BufferLength / 1000) * OutputParams.Sampling) + (512 - 1)) div 512) * 512;
          if OBufferSize[OutputDevice] > ONE_MEGABYTE then OBufferSize[OutputDevice] := ONE_MEGABYTE;                 //But not larger than one MegaSamples
        end else begin
          OBufferSize[OutputDevice]  := OutputParams.BufferCount;//in samples
        end;
        if OutputParams.SingleBuffer then begin
          OBufferSize2[OutputDevice] := OBufferSize[OutputDevice];
        end else begin
          OBufferSize2[OutputDevice] := 2 * OBufferSize[OutputDevice];//in samples
        end;
        OBuffer[OutputDevice]       := AllocMem(OBufferSize[OutputDevice]  * OElemSize[OutputDevice]);
        OBuffer2[OutputDevice]      := AllocMem(OBufferSize2[OutputDevice] * OElemSize[OutputDevice]);
        OutputParams.Buffer         := OBuffer2[OutputDevice];
        OutputParams.BufferCount    := OBufferSize2[OutputDevice];//in samples
        OutputParams.BufferElemSize := OElemSize[OutputDevice] div OutputParams.ChannelCount;
      except
        OFreeBuff(OutputDevice);
        Result := 1;
        Exit;
      end;

      //Load the first two buffers with data
      Count := OBufferSize2[OutputDevice];//samples
      try
        if @OUserCallBack[OutputDevice] <> nil then begin
          OUserCallBack[OutputDevice](OUserCaller[OutputDevice], OBuffer2[OutputDevice], OElemSize[OutputDevice], Count, OError[OutputDevice]);
        end else begin
          FillChar(OBuffer2[OutputDevice]^, Count * OutputParams.ChannelCount * OElemSize[OutputDevice], 0);
        end;
      except
      end;



      //**********************************************************************//
      //                                                                      //
      //                             D I G I T A L                            //
      //                                                                      //
      //**********************************************************************//
      if OutputParams.DigitalChannelCount > 0 then begin

        if @ODUserCallBack[OutputDevice] <> nil then begin
          ODCount[OutputDevice] := Round(OutputParams.Sampling / OutputParams.DigitalSampling);
          if (ODCount[OutputDevice] < 1)   then ODCount[OutputDevice] := 1;
          case ODCount[OutputDevice] of
            1:        ODCount[OutputDevice] := 1;
            2:        ODCount[OutputDevice] := 2;
            3..4:     ODCount[OutputDevice] := 4;
            5..8:     ODCount[OutputDevice] := 8;
            9..16:    ODCount[OutputDevice] := 16;
            17..32:   ODCount[OutputDevice] := 32;
            33..64:   ODCount[OutputDevice] := 64;
            65..128:  ODCount[OutputDevice] := 128;
            129..256: ODCount[OutputDevice] := 256;
            257..512: ODCount[OutputDevice] := 512;
            else  ODCount[OutputDevice] := ((ODCount[OutputDevice] + (512 - 1)) div 512) * 512;
          end;
          ODUserCallBack[OutputDevice](OUserCaller[OutputDevice], ODValue[OutputDevice], ODCount[OutputDevice], ODError[OutputDevice]);
          DigVal := ODValue[OutputDevice];
          for c := 0 to OutputParams.DigitalChannelCount - 1 do begin
            Result := DIG_Prt_Config(Devices[OutputDevice], c, 0, 1);
            Result := DIG_Out_Port(Devices[OutputDevice], c, DigVal and $ff);
            DigVal := DigVal shr 8;
          end;
          //Set the Event trapping
          Result := Config_DAQ_Event_Message(Devices[OutputDevice], 1{Add}, PChar(ODChStr[OutputDevice]), 1, ODCount[OutputDevice], 0, 0, 0, 0, 0, 0, Cardinal(@OutputDigitalDeviceCallback));
        end;
      end;


      //Set the Event trapping
      if @OUserCallBack[OutputDevice] <> nil then begin
        Result := Config_DAQ_Event_Message(Devices[OutputDevice], 1{Add}, PChar(OChStr[OutputDevice]), 1, OBufferSize[OutputDevice], 0, 0, 0, 0, 0{Handle}, 0{Message}, Cardinal(@OutputDeviceCallback));
        if Result <> 0 then Exit;
      end;

      r_o := 0;

      //************************************************************************



      IRunning[InputDevice]  := True;
      ORunning[OutputDevice] := True;


      //Start Acquisition ******************************************************
      //Input
      r_i := 1;
      if InputParams.ChannelCount = 1 then begin
      //Single Channel
        Result := DAQ_Start(Devices[InputDevice], IChanVector[0], IGainVector[0], IBuffer2[InputDevice], IBufferSize2[InputDevice], ISampTB, ISampInt);
        if Result <> 0 then Exit;
      end else begin
      //Multiple Channel
        Result := SCAN_Setup(Devices[InputDevice], InputParams.ChannelCount, @IChanVector[0], @IGainVector[0]);
        if Result <> 0 then Exit;
        Result := SCAN_Start(Devices[InputDevice], IBuffer2[InputDevice], IBufferSize2[InputDevice] * InputParams.ChannelCount, IsampTB, IsampInt, ISampTB, 0);
        if Result <> 0 then Exit;
      end;
      r_i := 0;
      //Output
      r_o := 1;
      Result := WFM_Op(Devices[OutputDevice], OutputParams.ChannelCount, @OChanVector[OutputDevice][0], OBuffer2[OutputDevice], OBufferSize2[OutputDevice] * OutputParams.ChannelCount, 0, OutputParams.Sampling);
      if Result <> 0 then Exit;
      r_o := 0;


    except

      Result := 1;

    end;


  finally


    if Result <> 0 then begin

      InputStop(Devices[InputDevice]);
      OutputStop(Devices[OutputDevice]);
      Result := 1;

    end;


  end;

end;



//****************************************************************************//
//                                                                            //
//                 I N I T I A L I Z E   /   T E R M I N A T E                //
//                                                                            //
//****************************************************************************//
function LibraryInitialize :Integer;
var
  i :I16;
  iDevCode :U32;
  iResult : Integer;
begin
  {$ifdef DEBUG}
  AssignFile(fLog,'C:\QUB.LOG');
  Rewrite(fLog);
  {$endif}
  Result := 0;

  //Count devices, and get their names
  DeviceCount := 0;
  for i := 0 to ACQ_MAX_DEVICES - 1 do begin
    iResult := Get_DAQ_Device_Info(i + 1, ND_DEVICE_TYPE_CODE, @iDevCode);
    {$ifdef DEBUG}
    WriteLn( fLog, Format('Get_DAQ_Device_Info(%d,ND_DEVICE_TYPE_CODE,->%d)->%d',[i+1,iDevCode,iResult]));
    {$endif}
    if iResult = 0 then begin
      Devices    [DeviceCount] := i + 1;
      DeviceNames[DeviceCount] := Code2Name(iDevCode) + ' (' + IntToStr(Devices[DeviceCount]) + ')';
      DeviceCodes[DeviceCount] := iDevCode;
      Inc(DeviceCount);
    end;
    {
    if IBuffer[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
    if IBuffer2[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
    if OBuffer[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
    if OBuffer2[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
    }
  end;

end;

function LibraryTerminate :Integer;
//var
//  i :Integer;
begin
  {$ifdef DEBUG}
  CloseFile(fLog);
  {$endif}
  Result := 0;
  {
  for i := 0 to ACQ_MAX_DEVICES - 1 do begin
    if IBuffer[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
    if IBuffer2[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
    if OBuffer[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
    if OBuffer2[i] <> nil then begin
      MessageDlg('Non-nil buffer!', mtInformation, [mbOK], 0);
    end;
  end;
  }
end;



//****************************************************************************//
//                                                                            //
//                           L I B R A R Y    N A M E                         //
//                                                                            //
//****************************************************************************//
function LibraryNameLength :Integer;
begin
  Result := Length(LIBRARY_NAME);
end;

procedure LibraryName(Name :PChar);
begin
  Move(PChar(LIBRARY_NAME)^, Name^, Length(LIBRARY_NAME));
end;


//****************************************************************************//
//                                                                            //
//                           D E V I C E    C O U N T                         //
//                                                                            //
//****************************************************************************//

//Input
function InputDeviceCount :Integer;
begin

  Result := DeviceCount;

end;

procedure InputDeviceName(Device :Integer; Name :PChar);
begin

  Move(PChar(DeviceNames[Device])^, Name^, Length(DeviceNames[Device]));
  Name[Length(DeviceNames[Device])] := #0;

end;

function InputDeviceCode(Device :Integer):Integer;
begin
	result := DeviceCodes[Device];
end;

//Output
function OutputDeviceCount :Integer;
begin

  Result := DeviceCount;

end;

procedure OutputDeviceName(Device :Integer; Name :PChar);
begin

  Move(PChar(DeviceNames[Device])^, Name^, Length(DeviceNames[Device]));
  Name[Length(DeviceNames[Device])] := #0;

end;


//Some local functions
//-- Clear input buffer
procedure IFreeBuff(Device :Integer);
begin
  FreeMem(IBuffer[Device]);
  IBuffer[Device]  := nil;
  FreeMem(IBuffer2[Device]);
  IBuffer2[Device] := nil;
end;

//-- Clear output buffer
procedure OFreeBuff(Device :Integer);
begin
  FreeMem(OBuffer[Device]);
  OBuffer[Device]  := nil;
  FreeMem(OBuffer2[Device]);
  OBuffer2[Device] := nil;
end;

//Used to calculate the actual Sampling rate used by the card
function NI_TimeBaseToResolution(TimeBase :i16) :Double;
begin
  case TimeBase of
    -3: Result := 50e-9;    //20 MHz clock used as a timebase (50 ms) (E Series, 671X, and 673X only).
    -1: Result := 200e-9;   //5 MHz clock used as timebase (200 ns resolution).
     1: Result := 1e-6;     //1 MHz clock used as timebase (1 s resolution).
     2: Result := 10e-6;    //100 kHz clock used as timebase (10 s resolution).
     3: Result := 100e-6;   //10 kHz clock used as timebase (100 s resolution).
     4: Result := 1e-3;     //1 kHz clock used as timebase (1 ms resolution).
     5: Result := 10e-3;    //100 Hz clock used as timebase (10 ms resolution).
  else  Result := 0;
  end;
end;

//-- Return 0 or device array index for specified device number
function NI_To_Device(NIDevice :Integer) :Integer;
var
  i :Integer;
begin
  Result := 0;
  for i := 0 to DeviceCount - 1 do begin
    if Devices[i] = NIDevice then begin
      Result := i;
      Exit;
    end;
  end;
end;

//-- Text versions of NI card names to fill DeviceNames array.
function Code2Name(Code :Integer) :String;
begin
  case Code of
    1:     Result := 'Unknown - ' + IntToStr(Code);
    7:     Result := 'NI PC-DIO-24';
    8:     Result := 'NI AT-DIO-32F';
    12:    Result := 'NI PC-DIO-96';
    13:    Result := 'NI PC-LPM-16';
    14:    Result := 'NI PC-TIO-10';
    15:    Result := 'NI AT-AO-6';
    25:    Result := 'NI AT-MIO-16E-2';
    26:    Result := 'NI AT-AO-10';
    28:    Result := 'NI Lab-PC+';
    30:    Result := 'NI SCXI-1200';
    31:    Result := 'NI DAQCard-700';
    32:    Result := 'NI NEC-MIO-16E-4';
    33:    Result := 'NI DAQPad-1200';
    35:    Result := 'NI DAQCard DIO-24';
    36:    Result := 'NI AT-MIO-16E-10';
    37:    Result := 'NI AT-MIO-16DE-10';
    38:    Result := 'NI AT-MIO-64E-3';
    39:    Result := 'NI AT-MIO-16XE-50';
    40:    Result := 'NI NEC-AI-16E-4';
    41:    Result := 'NI NEC-MIO-16XE-50';
    42:    Result := 'NI NEC-AI-16XE-50';
    43:    Result := 'NI DAQPad-MIO-16XE-50';
    44:    Result := 'NI AT-MIO-16E-1';
    45:    Result := 'NI PC-OPDIO-16';
    46:    Result := 'NI PC-AO-2DC';
    47:    Result := 'NI DAQCard-AO-2DC';
    48:    Result := 'NI DAQCard-1200';
    49:    Result := 'NI DAQCard-500';
    50:    Result := 'NI AT-MIO-16XE-10';
    51:    Result := 'NI AT-AI-16XE-10';
    52:    Result := 'NI DAQCard-AI-16XE-50';
    53:    Result := 'NI DAQCard-AI-16E-4';
    54:    Result := 'NI DAQCard-516';
    55:    Result := 'NI PC-516';
    56:    Result := 'NI PC-LPM-16PnP';
    57:    Result := 'NI Lab-PC-1200';
    58:    Result := 'NI Lab-PC-1200AI';
    59:    Result := 'NI VXI-MIO-64E-1';
    60:    Result := 'NI VXI-MIO-64XE-10';
    61:    Result := 'NI VXI-AO-48XDC';
    62:    Result := 'NI VXI-DIO-128';
    65:    Result := 'NI PC-DIO-24/PnP';
    66:    Result := 'NI PC-DIO-96/PnP';
    67:    Result := 'NI AT-DIO-32HS';
    68:    Result := 'NI DAQCard-6533';
    75:    Result := 'NI DAQPad-6507/8';
    //75:    Result := 'NI DAQPad-6508';
    76:    Result := 'NI DAQPad-6020E for USB';
    88:    Result := 'NI DAQCard-6062E';
    89:    Result := 'NI DAQCard-6715';
    90:    Result := 'NI DAQCard-6023E';
    91:    Result := 'NI DAQCard-6024E';
    200:   Result := 'NI PCI-DIO-96';
    201:   Result := 'NI PCI-1200';
    202:   Result := 'NI PCI-MIO-16XE-50';
    203:   Result := 'NI PCI-5102';
    204:   Result := 'NI PCI-MIO-16XE-10';
    205:   Result := 'NI PCI-MIO-16E-1';
    206:   Result := 'NI PCI-MIO-16E-4';
    207:   Result := 'NI PXI-6070E';
    208:   Result := 'NI PXI-6040E';
    209:   Result := 'NI PXI-6030E';
    210:   Result := 'NI PXI-6011E';
    211:   Result := 'NI PCI-DIO-32HS';
    216:   Result := 'NI PCI-6534';
    218:   Result := 'NI PXI-6534';
    220:   Result := 'NI PCI-6031E (MIO-64XE-10)';
    221:   Result := 'NI PCI-6032E (AI-16XE-10)';
    222:   Result := 'NI PCI-6033E (AI-64XE-10)';
    223:   Result := 'NI PCI-6071E (MIO-64E-1)';
    232:   Result := 'NI PCI-6602';
    233:   Result := 'NI 4451 for PCI';
    234:   Result := 'NI 4452 for PCI';
    235:   Result := 'NI 4551 for PCI';
    236:   Result := 'NI 4552 for PCI';
    237:   Result := 'NI PXI-6602';
    240:   Result := 'NI PXI-6508';
    241:   Result := 'NI PCI-6110';
    244:   Result := 'NI PCI-6111';
    256:   Result := 'NI PCI-6503';
    257:   Result := 'NI PXI-6503';
    258:   Result := 'NI PXI-6071E';
    259:   Result := 'NI PXI-6031E';
    261:   Result := 'NI PCI-6711';
    262:   Result := 'NI PCI-6711';
    263:   Result := 'NI PCI-6713';
    264:   Result := 'NI PXI-6713';
    265:   Result := 'NI PCI-6704';
    266:   Result := 'NI PXI-6704';
    267:   Result := 'NI PCI-6023E';
    268:   Result := 'NI PXI-6023E';
    269:   Result := 'NI PCI-6024E';
    270:   Result := 'NI PXI-6024E';
    271:   Result := 'NI PCI-6025E';
    272:   Result := 'NI PXI-6025E';
    273:   Result := 'NI PCI-6052E';
    274:   Result := 'NI PXI-6052E';
    275:   Result := 'NI DAQPad-6070E for 1394';
    276:   Result := 'NI DAQPad-6052E for 1394';
    285:   Result := 'NI PCI-6527';
    286:   Result := 'NI PXI-6527';
    308:   Result := 'NI PCI-6601';
    311:   Result := 'NI PCI-6703';
    314:   Result := 'NI PCI-6034E';
    315:   Result := 'NI PXI-6034E';
    316:   Result := 'NI PCI-6035E';
    317:   Result := 'NI PXI-6035E';
    318:   Result := 'NI PXI-6703';
    319:   Result := 'NI PXI-6608';
    321:   Result := 'NI 4454 for PCI';
    327:   Result := 'NI PCI-6608';
    329:   Result := 'NI 6222 for PCI';
    330:   Result := 'NI 6222 for PXI';
    331:   Result := 'NI 6224 for Ethernet';
    332:   Result := 'NI DAQPad-6052E for USB';
    335:   Result := 'NI 4472 for PXI/CompactPCI';
    338:   Result := 'NI PCI-6115';
    339:   Result := 'NI PXI-6115';
    340:   Result := 'NI PCI-6120';
    341:   Result := 'NI PXI-6120';
    342:   Result := 'NI 4472 for PCI';
    347:   Result := 'NI 4472 for 1394';
    361:   Result := 'NI DAQPad-6052E for 1394';
    // Codes added 4-6-2004 {JB}
    215:   Result := 'NI PXI-6533';
    348:   Result := 'NI DAQCard-6036E';
    349:   Result := 'NI 6731 for PCI';
    350:   Result := 'NI 6733 for PCI';
    351:   Result := 'NI 6731 for PXI/Compact PCI';
    352:   Result := 'NI 6733 for PXI/Compact PCI';
    353:   Result := 'NI PCI-4474';
    354:   Result := 'NI PXI-4474';
    366:   Result := 'NI PCI-6013';
    367:   Result := 'NI PCI-6014';

    else   Result := 'NI Unknown - ' + IntToStr(Code);

  end;//case Code

end;


//****************************************************************************//
//                                                                            //
//                          C H A N N E L    C O U N T                        //
//                                                                            //
//****************************************************************************//

//Input
function InputChannelCount(Device :Integer) :Integer;
begin
  Result := 8;
end;

//Output
function OutputChannelCount(Device :Integer) :Integer;
begin
  Result := 2;
end;

function OutputDigitalChannelCount(Device :Integer) :Integer;
begin
  Result := 2;
end;


//****************************************************************************//
//                                                                            //
//                               D E F A U L T S                              //
//                                                                            //
//****************************************************************************//

//Input
function InputGain(Device, Channel :Integer) :Double;
begin
  Result := 1;
end;

function InputRange(Device :Integer) :Double;
begin
  Result := 20;
end;

function InputDifferential(Device :Integer) :Integer;
begin
  Result := Integer(True);
end;

function InputBits(Device :Integer) :Integer;
begin
  Result := 12;
end;

//Output
function OutputGain(Device, Channel :Integer) :Double;
begin
  Result := 1;
end;

function OutputRange(Device :Integer) :Double;
begin
  Result := 20;
end;

function OutputDifferential(Device :Integer) :Integer;
begin
  Result := Integer(True);
end;

function OutputBits(Device :Integer) :Integer;
begin
  Result := 12;
end;


//****************************************************************************//
//                                                                            //
//                               E X P O R T S                                //
//                                                                            //
//****************************************************************************//
exports


  LibraryInitialize,
  LibraryTerminate,

  LibraryNameLength,
  LibraryName,

  //Input
  InputStart,
  InputStop,

  InputDeviceCount,
  InputDeviceName,
  InputDeviceCode,
  InputChannelCount,
  InputGain,
  InputRange,
  InputDifferential,
  InputBits,

  ReadDBLChannel,

  AIRead,
  DIRead,

  //Output
  OutputStart,
  OutputStop,

  OutputDeviceCount,
  OutputDeviceName,
  OutputChannelCount,
  OutputDigitalChannelCount,
  OutputGain,
  OutputRange,
  OutputDifferential,
  OutputBits,

  WriteDBLChannel,

  AOWrite,
  DOWrite,

  //Synch
  SynchStart;


end.
