library qubacqsc;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
//11-10-2000


{
                         QUB data acquisition library interface

                                  for Sound Cards
}


uses


SysUtils, Windows, MMSystem, 

AllocTypes,

QUB_Global,
QUB_Acquisition_Global;


{$R *.RES}


const


LIBRARY_NAME                    = 'QUB SoundCard 2.0';


var


//Input
IDeviceCount                    :Integer;
IDeviceHandle                   :Array[0..ACQ_MAX_DEVICES - 1] of HWAVEIN;
IBufferSize                     :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
IBufferIndex                    :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
IBuffer                         :Array[0..ACQ_MAX_DEVICES - 1, 1..2] of Pointer;
IHeader                         :Array[0..ACQ_MAX_DEVICES - 1, 1..2] of TWAVEHDR;
IRunning                        :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;
IUserCallBack                   :Array[0..ACQ_MAX_DEVICES - 1] of TQUB_Acquisition_InputCallBack;
IUserCaller                     :Array[0..ACQ_MAX_DEVICES - 1] of TObject;
IWaveFormat                     :Array[0..ACQ_MAX_DEVICES - 1] of TWAVEFORMATEX;
ICalling                        :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;

//Output
ODeviceCount                    :Integer;
ODeviceHandle                   :Array[0..ACQ_MAX_DEVICES - 1] of HWAVEIN;
OBufferSize                     :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OBufferIndex                    :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OBuffer                         :Array[0..ACQ_MAX_DEVICES - 1, 1..2] of Pointer;
OHeader                         :Array[0..ACQ_MAX_DEVICES - 1, 1..2] of TWAVEHDR;
ORunning                        :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;
OUserCallBack                   :Array[0..ACQ_MAX_DEVICES - 1] of TQUB_Acquisition_OutputCallBack;
OUserCaller                     :Array[0..ACQ_MAX_DEVICES - 1] of TObject;
OWaveFormat                     :Array[0..ACQ_MAX_DEVICES - 1] of TWAVEFORMATEX;
OError                          :Array[0..ACQ_MAX_DEVICES - 1] of Integer;
OCalling                        :Array[0..ACQ_MAX_DEVICES - 1] of Boolean;


//Export functions
function LibraryInitialize :Integer; stdcall; forward;
function LibraryTerminate :Integer; stdcall; forward;


function LibraryNameLength :Integer; stdcall; forward;
procedure LibraryName(Name :PChar); stdcall; forward;


//Input
function InputStart(Device :Integer; var InputParams :TQUB_Acquisition_InputStart_Params) :Integer; stdcall; forward;
function InputStop(Device :Integer) :Integer; stdcall; forward;

function InputDeviceCount :Integer; stdcall; forward;
procedure InputDeviceName(Device :Integer; Name :PChar); stdcall; forward;
function InputChannelCount(Device :Integer) :Integer; stdcall; forward;
function InputGain(Device, Channel :Integer) :Double; stdcall; forward;
function InputRange(Device :Integer) :Double; stdcall; forward;
function InputDifferential(Device :Integer) :Integer; stdcall; forward;
function InputBits(Device :Integer) :Integer; stdcall; forward;

function ReadDBLChannel(Device :Integer; Channel :Integer; Gain :Double; var Value :Double) :Integer; stdcall; forward;

function AIRead(Device :Integer; Channel :Integer; Gain :Double; var Value :Integer) :Integer; stdcall; forward;
function DIRead(Device :Integer; Channel :Integer; var Value :Integer) :Integer; stdcall; forward;

procedure InputDeviceCallBack(hwi :HWAVEIN; uMsg :UINT; dwInstance :DWORD; dwParam1, dwParam2 :DWORD); stdcall; forward;


//Output
function OutputStart(Device :Integer; var OutputParams :TQUB_Acquisition_OutputStart_Params) :Integer; stdcall; forward;
function OutputStop(Device :Integer) :Integer; stdcall; forward;

function OutputDeviceCount :Integer; stdcall; forward;
procedure OutputDeviceName(Device :Integer; Name :PChar); stdcall; forward;
function OutputChannelCount(Device :Integer) :Integer; stdcall; forward;
function OutputDigitalChannelCount(Device :Integer) :Integer; stdcall; forward;
function OutputGain(Device, Channel :Integer) :Double; stdcall; forward;
function OutputRange(Device :Integer) :Double; stdcall; forward;
function OutputDifferential(Device :Integer) :Integer; stdcall; forward;
function OutputBits(Device :Integer) :Integer; stdcall; forward;

function WriteDBLChannel(Device :Integer; Channel :Integer; const Value :Double) :Integer; stdcall; forward;

function AOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall; forward;
function DOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer; stdcall; forward;

procedure OutputDeviceCallBack(hwo :HWAVEOUT; uMsg :UINT; dwInstance :DWORD; dwParam1, dwParam2 :DWORD); stdcall; forward;


//Synch
function SynchStart(InputDevice, OutputDevice :Integer; var InputParams :TQUB_Acquisition_InputStart_Params; var OutputParams :TQUB_Acquisition_OutputStart_Params; var r_i, r_o :Integer) :Integer; stdcall; forward;


function SC_To_IDevice(hwi :HWAVEIN) :Integer; forward;
function SC_To_ODevice(hwo :HWAVEOUT) :Integer; forward;

procedure IFreeBuff(Device :Integer); forward;
procedure OFreeBuff(Device :Integer); forward;



//****************************************************************************//
//                                                                            //
//                          A C Q U I S I T I O N                             //
//                                                                            //
//****************************************************************************//

//Input
function InputStart(Device :Integer; var InputParams :TQUB_Acquisition_InputStart_Params) :Integer;
var
  i, j :Integer;

  procedure ErrorFreeAndClose;
  begin
    IFreeBuff(Device);
    waveInClose(IDeviceHandle[Device]);
  end;

  procedure ErrorUnPrepareAndFreeAndClose;
  var
    i :Integer;
  begin
    for i := 2 downto 1 do begin
      waveInUnprepareHeader(IDeviceHandle[Device], @IHeader[Device, i], SizeOf(IHeader[Device, i]));
    end;
    ErrorFreeAndClose;
  end;

begin

  Result := 1;


  if IRunning[Device] then Exit;


  ICalling[Device]      := False;
  IUserCallBack[Device] := InputParams.CallBack;
  IUserCaller[Device]   := InputParams.Caller;


  //Free Buffers (just in case)
  IFreeBuff(Device);


  //Prepare Format
  try
    with IWaveFormat[Device] do begin
      wFormatTag            := WAVE_FORMAT_PCM;
      nChannels             := InputParams.ChannelCount;
      nSamplesPerSec        := Round(InputParams.Sampling);
      nAvgBytesPerSec       := nSamplesPerSec * DWORD(((InputParams.Bits - 1) div 8) + 1);
      nBlockAlign           := InputParams.ChannelCount * (((InputParams.Bits - 1) div 8) + 1);
      wBitsPerSample        := (((InputParams.Bits - 1) div 8) + 1) * 8;
      cbSize                := 0;
    end;
  except
    Result := 1;
    Exit;
  end;


  //Open Device
  Result := waveInOpen(@IDeviceHandle[Device], Device, @IWaveFormat[Device], DWORD(@InputDeviceCallback), DWORD(IUserCaller[Device]), CALLBACK_FUNCTION);
  if Result <> MMSYSERR_NOERROR then begin
    Result := 1;
    Exit;
  end;


  //Allocate Buffers
  try
    if InputParams.BufferCount = 0 then begin
      IBufferSize[Device]  := Round((InputParams.BufferLength / 1000) * InputParams.Sampling);
      if IBufferSize[Device] > ONE_MEGABYTE then IBufferSize[Device] := ONE_MEGABYTE;               //But not larger than one MegaSamples
    end else begin
      IBufferSize[Device]  := InputParams.BufferCount;
    end;
    IBufferSize[Device]    := IBufferSize[Device] * IWaveFormat[Device].nBlockAlign;//in bytes

    //Code for single buffer

    //Allocate contiguous buffers
    IBuffer[Device, 1]         := AllocMem(2*IBufferSize[Device]);
    IBuffer[Device, 2]         := @TPAByte(IBuffer[Device, 1])[IBufferSize[Device]];
    InputParams.Buffer         := IBuffer[Device, 1];
    InputParams.BufferCount    := 2 * IBufferSize[Device] div IWaveFormat[Device].nBlockAlign;//in samples
    InputParams.BufferElemSize := IWaveFormat[Device].nBlockAlign div InputParams.ChannelCount;
  except;
    Result := 1;
    ErrorFreeAndClose;
    Exit;
  end;


  //Prepare Headers:
  for i := 1 to 2 do begin
    with IHeader[Device, i] do begin
      lpData             := IBuffer[Device, i];
      dwBufferLength     := IBufferSize[Device];
      dwBytesRecorded    := 0;
      dwUser             := 0;
      dwFlags            := 0;
      dwLoops            := 0;
    end;
    Result := waveInPrepareHeader(IDeviceHandle[Device], @IHeader[Device, i], SizeOf(IHeader[Device, i]));
    if Result <> MMSYSERR_NOERROR then begin
      for j := i downto 1 do begin
        waveInUnprepareHeader(IDeviceHandle[Device], @IHeader[Device, j], SizeOf(IHeader[Device, j]));
      end;
      ErrorFreeAndClose;
      Result := 1;
      Exit;
    end;
  end;


  //Add Headers:
  for i := 1 to 2 do begin
    Result := waveInAddBuffer(IDeviceHandle[Device], @IHeader[Device, i], SizeOf(IHeader[Device, i]));
    if Result <> MMSYSERR_NOERROR then begin
      ErrorUnprepareAndFreeAndClose;
      Result := 1;
      Exit;
    end;
  end;


  //Start Device:
  IRunning[Device]     := True;
  IBufferIndex[Device] := 1;

  Result := waveInStart(IDeviceHandle[Device]);
  if Result <> MMSYSERR_NOERROR then begin
    ErrorUnprepareAndFreeAndClose;
    Result := 1;
    IRunning[Device] := False;
    Exit;
  end;


end;

function InputStop(Device :Integer) :Integer;
var
  i :Integer;
begin

  Result := 0;


  IRunning[Device] := False;

  while ICalling[Device] do begin
  end;
  waveInReset(IDeviceHandle[Device]);
  for i := 1 to 2 do begin
    waveInUnPrepareHeader(IDeviceHandle[Device], @IHeader[Device, i], SizeOf(IHeader[Device, i]));
  end;
  waveInClose(IDeviceHandle[Device]);

end;



//****************************************************************************//
//                                                                            //
//                               C A L L B A C K                              //
//                                                                            //
//****************************************************************************//
procedure InputDeviceCallBack(hwi :HWAVEIN; uMsg :UINT; dwInstance :DWORD; dwParam1, dwParam2 :DWORD);
var
  Device :Integer;
begin

  Device := SC_To_IDevice(hwi);

  ICalling[Device] := True;

  try

    case uMsg of

      WIM_OPEN:
        begin
          IRunning[Device]     := True;
          IBufferIndex[Device] := 1;
        end;

      WIM_DATA:
        begin

          try
            if @IUserCallBack[Device] <> nil then begin
              IUserCallBack[Device](IUserCaller[Device], IHeader[Device, IBufferIndex[Device]].lpData, IWaveFormat[Device].nBlockAlign, IHeader[Device, IBufferIndex[Device]].dwBytesRecorded div IWaveFormat[Device].nBlockAlign, 0);
            end;
          except
          end;

          if IRunning[Device] then begin

            try

              waveInUnPrepareHeader(IDeviceHandle[Device], @IHeader[Device, IBufferIndex[Device]], SizeOf(IHeader[Device, IBufferIndex[Device]]));

              IHeader[Device, IBufferIndex[Device]].dwFlags := 0;
              waveInPrepareHeader(IDeviceHandle[Device], @IHeader[Device, IBufferIndex[Device]], SizeOf(IHeader[Device, IBufferIndex[Device]]));
              waveInAddBuffer(IDeviceHandle[Device], @IHeader[Device, IBufferIndex[Device]], SizeOf(IHeader[Device, IBufferIndex[Device]]));

            except
            end;

          end;

          if IBufferIndex[Device] = 1 then IBufferIndex[Device] := 2 else IBufferIndex[Device] := 1;

        end;

      WIM_CLOSE:
        begin
          IRunning[Device] := False;
          IFreeBuff(Device);
        end;

    end;//case uMsg

  finally

    ICalling[Device] := False;

  end;

end;

function ReadDBLChannel(Device :Integer; Channel :Integer; Gain :Double; var Value :Double) :Integer;
begin
  Result := 1;//SoundCard doesn't Read/Write single-shot
end;

function AIRead(Device :Integer; Channel :Integer; Gain :Double; var Value :Integer) :Integer;
begin
  Result := 1;//SoundCard doesn't Read/Write single-shot
end;

function DIRead(Device :Integer; Channel :Integer; var Value :Integer) :Integer;
begin
  Result := 1;//SoundCard doesn't Read/Write Digital
end;


{

 SoundCard makes no use of the digital output

}

//Output
function OutputStart(Device :Integer; var OutputParams :TQUB_Acquisition_OutputStart_Params) :Integer;
var
  Count :Integer;
  i, j  :Integer;

  procedure ErrorFreeAndClose;
  begin
    OFreeBuff(Device);
    waveOutClose(ODeviceHandle[Device]);
  end;

  procedure ErrorUnPrepareAndFreeAndClose;
  var
    i :Integer;
  begin
    for i := 2 downto 1 do begin
      waveOutUnprepareHeader(ODeviceHandle[Device], @OHeader[Device, i], SizeOf(OHeader[Device, i]));
    end;
    ErrorFreeAndClose;
  end;

begin

  Result := 1;

  if ORunning[Device] then Exit;


  OCalling[Device]      := False;
  OUserCallBack[Device] := OutputParams.CallBack;
  OUserCaller[Device]   := OutputParams.Caller;


  //Free Buffers (just in case)
  OFreeBuff(Device);


  //Prepare Format
  try
    with OWaveFormat[Device] do begin
      wFormatTag            := WAVE_FORMAT_PCM;
      nChannels             := OutputParams.ChannelCount;
      nSamplesPerSec        := Round(OutputParams.Sampling);
      nAvgBytesPerSec       := nSamplesPerSec * DWORD(((OutputParams.Bits - 1) div 8) + 1);
      nBlockAlign           := OutputParams.ChannelCount * (((OutputParams.Bits - 1) div 8) + 1);
      wBitsPerSample        := (((OutputParams.Bits - 1) div 8) + 1) * 8;
      cbSize                := 0;
    end;
  except
    Result := 1;
    Exit;
  end;


  //Open Device
  Result := waveOutOpen(@ODeviceHandle[Device], Device, @OWaveFormat[Device], DWORD(@OutputDeviceCallback), DWORD(OUserCaller[Device]), CALLBACK_FUNCTION);
  if Result <> MMSYSERR_NOERROR then begin
    Result := 1;
    Exit;
  end;


  //Allocate Buffers
  try
    if OutputParams.BufferCount = 0 then begin
      OBufferSize[Device]  := Round((OutputParams.BufferLength / 1000) * OutputParams.Sampling);
      if OBufferSize[Device] > ONE_MEGABYTE then OBufferSize[Device] := ONE_MEGABYTE;               //But not larger than one MegaSamples
    end else begin
      OBufferSize[Device]  := OutputParams.BufferCount;
    end;
    OBufferSize[Device]    := OBufferSize[Device] * OWaveFormat[Device].nBlockAlign;//in bytes

    //Code for single buffer

    //Allocate contiguous buffers
    OBuffer[Device, 1]          := AllocMem(2*OBufferSize[Device]);
    OBuffer[Device, 2]          := @TPAByte(OBuffer[Device, 1])[OBufferSize[Device]];
    OutputParams.Buffer         := OBuffer[Device, 1];
    OutputParams.BufferCount    := 2 * OBufferSize[Device] div OWaveFormat[Device].nBlockAlign;//in samples
    OutputParams.BufferElemSize := OWaveFormat[Device].nBlockAlign div OutputParams.ChannelCount;
  except;
    Result := 1;
    ErrorFreeAndClose;
    Exit;
  end;


  //Prepare Headers:
  for i := 1 to 2 do begin
    with OHeader[Device, i] do begin
      Count := OBufferSize[Device] div OWaveFormat[Device].nBlockAlign;
      if @OUserCallBack[Device] <> nil then begin
        OUserCallBack[Device](OUserCaller[Device], OBuffer[Device, i], OWaveFormat[Device].nBlockAlign, Count, OError[Device]);
      end else begin
        FillChar(TPAByte(OBuffer[Device, i])[0], OBufferSize[Device], 0);
      end;
      dwBufferLength     := OBufferSize[Device];
      lpData             := OBuffer[Device, i];
      dwUser             := 0;
      dwFlags            := 0;
      dwLoops            := 1;
    end;
    Result := waveOutPrepareHeader(ODeviceHandle[Device], @OHeader[Device, i], SizeOf(OHeader[Device, i]));
    if Result <> MMSYSERR_NOERROR then begin
      for j := i downto 1 do begin
        waveOutUnprepareHeader(ODeviceHandle[Device], @OHeader[Device, j], SizeOf(OHeader[Device, j]));
      end;
      ErrorFreeAndClose;
      Result := 1;
      Exit;
    end;
  end;


  //Add Headers:
  for i := 1 to 2 do begin
    OHeader[Device, i].dwFlags := OHeader[Device, i].dwFlags or WHDR_BEGINLOOP;
    Result := waveOutWrite(ODeviceHandle[Device], @OHeader[Device, i], SizeOf(OHeader[Device, i]));
    if Result <> MMSYSERR_NOERROR then begin
      ErrorUnprepareAndFreeAndClose;
      Result := 1;
      Exit;
    end;
  end;


  //Start Device:
  ORunning[Device]     := True;
  OBufferIndex[Device] := 1;
  Result := waveOutReStart(ODeviceHandle[Device]);
  if Result <> MMSYSERR_NOERROR then begin
    ErrorUnprepareAndFreeAndClose;
    Result := 1;
    ORunning[Device] := False;
    Exit;
  end;

end;

function OutputStop(Device :Integer) :Integer;
var
  i :Integer;
begin

  Result := 0;
  ORunning[Device] := False;
  while OCalling[Device] do begin
  end;
  waveOutReset(ODeviceHandle[Device]);
  for i := 1 to 2 do begin
    waveOutUnPrepareHeader(ODeviceHandle[Device], @OHeader[Device, i], SizeOf(OHeader[Device, i]));
  end;
  waveOutClose(ODeviceHandle[Device]);

end;



//****************************************************************************//
//                                                                            //
//                               C A L L B A C K                              //
//                                                                            //
//****************************************************************************//
procedure OutputDeviceCallBack(hwo :HWAVEOUT; uMsg :UINT; dwInstance :DWORD; dwParam1, dwParam2 :DWORD);
var
  Device :Integer;
  Count  :Integer;
begin

  Device := SC_To_ODevice(hwo);

  OCalling[Device] := True;

  try

    case uMsg of

      WOM_OPEN:
        begin
          ORunning[Device]     := True;
          OBufferIndex[Device] := 1;
        end;

      WOM_DONE:
        begin

          if ORunning[Device] then begin

            waveOutUnPrepareHeader(ODeviceHandle[Device], @OHeader[Device, OBufferIndex[Device]], SizeOf(OHeader[Device, OBufferIndex[Device]]));

            Count := OBufferSize[Device] div OWaveFormat[Device].nBlockAlign;
            try
              if @OUserCallBack[Device] <> nil then begin
                OUserCallBack[Device](OUserCaller[Device], OHeader[Device, OBufferIndex[Device]].lpData, OWaveFormat[Device].nBlockAlign, Count, OError[Device]);
              end;
            except
            end;
            OHeader[Device, OBufferIndex[Device]].dwFlags        := 0;
            OHeader[Device, OBufferIndex[Device]].dwBufferLength := OBufferSize[Device];
            waveOutPrepareHeader(ODeviceHandle[Device], @OHeader[Device, OBufferIndex[Device]], SizeOf(OHeader[Device, OBufferIndex[Device]]));
            waveOutWrite(ODeviceHandle[Device], @OHeader[Device, OBufferIndex[Device]], SizeOf(OHeader[Device, OBufferIndex[Device]]));

          end;

          if OBufferIndex[Device] = 1 then OBufferIndex[Device] := 2 else OBufferIndex[Device] := 1;

        end;

      WOM_CLOSE:
        begin
          ORunning[Device] := False;
          OFreeBuff(Device);
        end;

    end;//case uMsg

  finally

    OCalling[Device] := False;

  end;

end;


function WriteDBLChannel(Device :Integer; Channel :Integer; const Value :Double) :Integer;
begin
  Result := 1;//SoundCard doesn't Read/Write single-shot
end;

function AOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer;
begin
  Result := 1;//SoundCard doesn't Read/Write single-shot
end;

function DOWrite(Device :Integer; Channel :Integer; const Value :Integer) :Integer;
begin
  Result := 1;//SoundCard doesn't Read/Write Digital
end;


//Synch
function SynchStart(InputDevice, OutputDevice :Integer; var InputParams :TQUB_Acquisition_InputStart_Params; var OutputParams :TQUB_Acquisition_OutputStart_Params; var r_i, r_o :Integer) :Integer;
var
  i, j  :Integer;
  Count :Integer;

  procedure ErrorFreeAndClose;
  begin
    IFreeBuff(InputDevice);
    OFreeBuff(OutputDevice);
    waveInClose(IDeviceHandle[InputDevice]);
    waveOutClose(ODeviceHandle[OutputDevice]);
  end;

  procedure ErrorUnPrepareAndFreeAndClose;
  var
    i :Integer;
  begin
    for i := 2 downto 1 do begin
      waveInUnprepareHeader(IDeviceHandle[InputDevice],   @IHeader[InputDevice, i],  SizeOf(IHeader[InputDevice, i]));
      waveOutUnprepareHeader(ODeviceHandle[OutputDevice], @OHeader[OutputDevice, i], SizeOf(OHeader[OutputDevice, i]));
    end;
    ErrorFreeAndClose;
  end;

begin

  Result := 1;
  r_i    := 1;
  r_o    := 1;


  if IRunning[InputDevice]  then Exit;
  if ORunning[OutputDevice] then Exit;


  r_i    := 0;
  r_o    := 0;


  ICalling[InputDevice]        := False;
  IUserCallBack[InputDevice]   := InputParams.CallBack;
  IUserCaller[InputDevice]     := InputParams.Caller;
  OCalling[OutputDevice]       := False;
  OUserCallBack[OutputDevice]  := OutputParams.CallBack;
  OUserCaller[OutputDevice]    := OutputParams.Caller;


  //Free Buffers (just in case)
  IFreeBuff(InputDevice);
  OFreeBuff(OutputDevice);


  //Prepare Formats
  try
    with IWaveFormat[InputDevice] do begin
      wFormatTag            := WAVE_FORMAT_PCM;
      nChannels             := InputParams.ChannelCount;
      nSamplesPerSec        := Round(InputParams.Sampling);
      nAvgBytesPerSec       := nSamplesPerSec * DWORD(((InputParams.Bits - 1) div 8) + 1);
      nBlockAlign           := InputParams.ChannelCount * (((InputParams.Bits - 1) div 8) + 1);
      wBitsPerSample        := (((InputParams.Bits - 1) div 8) + 1) * 8;
      cbSize                := 0;
    end;
  except
    r_i    := 1;
    Result := 1;
    Exit;
  end;
  try
    with OWaveFormat[OutputDevice] do begin
      wFormatTag            := WAVE_FORMAT_PCM;
      nChannels             := OutputParams.ChannelCount;
      nSamplesPerSec        := Round(OutputParams.Sampling);
      nAvgBytesPerSec       := nSamplesPerSec * DWORD(((OutputParams.Bits - 1) div 8) + 1);
      nBlockAlign           := OutputParams.ChannelCount * (((OutputParams.Bits - 1) div 8) + 1);
      wBitsPerSample        := (((OutputParams.Bits - 1) div 8) + 1) * 8;
      cbSize                := 0;
    end;
  except
    r_o    := 1;
    Result := 1;
    Exit;
  end;


  //Open Device
  //Input
  Result := waveInOpen(@IDeviceHandle[InputDevice], InputDevice, @IWaveFormat[InputDevice], DWORD(@InputDeviceCallback), DWORD(IUserCaller[InputDevice]), CALLBACK_FUNCTION);
  if Result <> MMSYSERR_NOERROR then begin
    r_i    := 1;
    Result := 1;
    Exit;
  end;
  //Output
  Result := waveOutOpen(@ODeviceHandle[OutputDevice], OutputDevice, @OWaveFormat[OutputDevice], DWORD(@OutputDeviceCallback), DWORD(OUserCaller[OutputDevice]), CALLBACK_FUNCTION);
  if Result <> MMSYSERR_NOERROR then begin
    waveInClose(IDeviceHandle[InputDevice]);
    r_o    := 1;
    Result := 1;
    Exit;
  end;


  //Allocate Buffers
  try
    //Input
    if InputParams.BufferCount = 0 then begin
      IBufferSize[InputDevice]  := Round((InputParams.BufferLength / 1000) * InputParams.Sampling);
      if IBufferSize[InputDevice] > ONE_MEGABYTE then IBufferSize[InputDevice] := ONE_MEGABYTE;//But not larger than one MegaSamples
    end else begin
      IBufferSize[InputDevice]  := InputParams.BufferCount;
    end;
    IBufferSize[InputDevice]    := IBufferSize[InputDevice] * IWaveFormat[InputDevice].nBlockAlign;//in bytes

    //Single buffer code

    //Allocate contiguous buffers
    IBuffer[InputDevice, 1]    := AllocMem(2*IBufferSize[InputDevice]);
    IBuffer[InputDevice, 2]    := @TPAByte(IBuffer[InputDevice, 1])[IBufferSize[InputDevice]];
    InputParams.Buffer         := IBuffer[InputDevice, 1];
    InputParams.BufferCount    := 2 * IBufferSize[InputDevice] div IWaveFormat[InputDevice].nBlockAlign;//in samples
    InputParams.BufferElemSize := IWaveFormat[InputDevice].nBlockAlign div InputParams.ChannelCount;
  except;
    r_i    := 1;
    Result := 1;
    ErrorFreeAndClose;
    Exit;
  end;

  try
    //Output
    if OutputParams.BufferCount = 0 then begin
      OBufferSize[OutputDevice]  := Round((OutputParams.BufferLength / 1000) * OutputParams.Sampling);
      if OBufferSize[OutputDevice] > ONE_MEGABYTE then OBufferSize[OutputDevice] := ONE_MEGABYTE;//But not larger than one MegaSamples
    end else begin
      OBufferSize[OutputDevice]  := OutputParams.BufferCount;
    end;
    OBufferSize[OutputDevice]    := OBufferSize[OutputDevice] * OWaveFormat[OutputDevice].nBlockAlign;//in bytes

    //Single buffer code

    //Allocate contiguous buffers
    OBuffer[OutputDevice, 1]    := AllocMem(2*OBufferSize[OutputDevice]);
    OBuffer[OutputDevice, 2]    := @TPAByte(OBuffer[OutputDevice, 1])[OBufferSize[OutputDevice]];
    OutputParams.Buffer         := OBuffer[OutputDevice, 1];
    OutputParams.BufferCount    := 2 * OBufferSize[OutputDevice] div OWaveFormat[OutputDevice].nBlockAlign;//in samples
    OutputParams.BufferElemSize := OWaveFormat[OutputDevice].nBlockAlign div OutputParams.ChannelCount;
  except;
    r_o    := 1;
    Result := 1;
    ErrorFreeAndClose;
    Exit;
  end;


  //Prepare Headers:
  //Input
  for i := 1 to 2 do begin
    with IHeader[InputDevice, i] do begin
      lpData             := IBuffer[InputDevice, i];
      dwBufferLength     := IBufferSize[InputDevice];
      dwBytesRecorded    := 0;
      dwUser             := 0;
      dwFlags            := 0;
      dwLoops            := 0;
    end;
    Result := waveInPrepareHeader(IDeviceHandle[InputDevice], @IHeader[InputDevice, i], SizeOf(IHeader[InputDevice, i]));
    if Result <> MMSYSERR_NOERROR then begin
      for j := i downto 1 do begin
        waveInUnprepareHeader(IDeviceHandle[InputDevice], @IHeader[InputDevice, j], SizeOf(IHeader[InputDevice, j]));
      end;
      ErrorFreeAndClose;
      r_i    := 1;
      Result := 1;
      Exit;
    end;
  end;
  //Output
  for i := 1 to 2 do begin
    //Load the first two output buffers
    with OHeader[OutputDevice, i] do begin
      Count := OBufferSize[OutputDevice] div OWaveFormat[OutputDevice].nBlockAlign;
      if @OUserCallBack[OutputDevice] <> nil then begin
        OUserCallBack[OutputDevice](OUserCaller[OutputDevice], OBuffer[OutputDevice, i], OWaveFormat[OutputDevice].nBlockAlign, Count, OError[OutputDevice]);
      end else begin
        FillChar(TPAByte(OBuffer[OutputDevice, i])[0], OBufferSize[OutputDevice], 0);
      end;
      dwBufferLength     := Count * OWaveFormat[OutputDevice].nBlockAlign;
      lpData             := OBuffer[OutputDevice, i];
      dwUser             := 0;
      dwFlags            := 0;
      dwLoops            := 1;
    end;
    Result := waveOutPrepareHeader(ODeviceHandle[OutputDevice], @OHeader[OutputDevice, i], SizeOf(OHeader[OutputDevice, i]));
    if Result <> MMSYSERR_NOERROR then begin
      for j := i downto 1 do begin
        waveOutUnprepareHeader(ODeviceHandle[OutputDevice], @OHeader[OutputDevice, j], SizeOf(OHeader[OutputDevice, j]));
      end;
      for j := 2 downto 1 do begin
        waveInUnprepareHeader(IDeviceHandle[InputDevice], @IHeader[InputDevice, j], SizeOf(IHeader[InputDevice, j]));
      end;
      ErrorFreeAndClose;
      r_o    := 1;
      Result := 1;
      Exit;
    end;
  end;


  //Add Headers:
  //Input
  for i := 1 to 2 do begin
    Result := waveInAddBuffer(IDeviceHandle[InputDevice], @IHeader[InputDevice, i], SizeOf(IHeader[InputDevice, i]));
    if Result <> MMSYSERR_NOERROR then begin
      r_i    := 1;
      Result := 1;
      ErrorUnPrepareAndFreeAndClose;
      Exit;
    end;
  end;
  //Output
  for i := 1 to 2 do begin
    OHeader[OutputDevice, i].dwFlags := OHeader[OutputDevice, i].dwFlags or WHDR_BEGINLOOP;
    Result := waveOutWrite(ODeviceHandle[OutputDevice], @OHeader[OutputDevice, i], SizeOf(OHeader[OutputDevice, i]));
    if Result <> MMSYSERR_NOERROR then begin
      r_o    := 1;
      Result := 1;
      ErrorUnPrepareAndFreeAndClose;
      Exit;
    end;
  end;


  //Start Device:
  //Input
  IRunning[InputDevice]      := True;
  IBufferIndex[InputDevice]  := 1;
  Result := waveInStart(IDeviceHandle[InputDevice]);
  if Result <> MMSYSERR_NOERROR then begin
    r_i    := 1;
    Result := 1;
    ErrorUnPrepareAndFreeAndClose;
    IRunning[InputDevice] := False;
    Exit;
  end;
  //Output
  ORunning[OutputDevice]     := True;
  OBufferIndex[OutputDevice] := 1;
  Result := waveOutReStart(ODeviceHandle[OutputDevice]);
  if Result <> MMSYSERR_NOERROR then begin
    r_o    := 1;
    Result := 1;
    ErrorUnPrepareAndFreeAndClose;
    IRunning[InputDevice]  := False;
    ORunning[OutputDevice] := False;
    Exit;
  end;

end;


//Some Local functions
function SC_To_IDevice(hwi :HWAVEIN) :Integer;
var
  i :Integer;
begin
  Result := 0;
  for i := 0 to IDeviceCount - 1 do begin
    if IDeviceHandle[i] = hwi then begin
      Result := i;
      Exit;
    end;
  end;
end;

function SC_To_ODevice(hwo :HWAVEOUT) :Integer;
var
  i :Integer;
begin
  Result := 0;
  for i := 0 to ODeviceCount - 1 do begin
    if ODeviceHandle[i] = hwo then begin
      Result := i;
      Exit;
    end;
  end;
end;

procedure IFreeBuff(Device :Integer);
begin
  //Buffers are consecutive, thus free only the first
  FreeMem(IBuffer[Device, 1]);
  IBuffer[Device, 1] := nil;
  IBuffer[Device, 2] := nil;
end;

procedure OFreeBuff(Device :Integer);
begin
  //Buffers are consecutive, thus free only the first
  FreeMem(OBuffer[Device, 1]);
  OBuffer[Device, 1] := nil;
  OBuffer[Device, 2] := nil;
end;



//****************************************************************************//
//                                                                            //
//                 I N I T I A L I Z E   /   T E R M I N A T E                //
//                                                                            //
//****************************************************************************//
function LibraryInitialize :Integer;
begin
  Result := 0;

  //Count devices
  IDeviceCount := InputDeviceCount;
  if IDeviceCount > ACQ_MAX_DEVICES then IDeviceCount := ACQ_MAX_DEVICES;
  ODeviceCount := OutputDeviceCount;
  if ODeviceCount > ACQ_MAX_DEVICES then ODeviceCount := ACQ_MAX_DEVICES;

end;

function LibraryTerminate :Integer;
begin
  Result := 0;

end;


//****************************************************************************//
//                                                                            //
//                           L I B R A R Y    N A M E                         //
//                                                                            //
//****************************************************************************//
function LibraryNameLength :Integer;
begin
  Result := Length(LIBRARY_NAME);
end;

procedure LibraryName(Name :PChar);
begin
  Move(PChar(LIBRARY_NAME)^, Name^, Length(LIBRARY_NAME));
end;



//****************************************************************************//
//                                                                            //
//                           D E V I C E    C O U N T                         //
//                                                                            //
//****************************************************************************//

//Input
function InputDeviceCount :Integer;
begin
  Result := waveInGetNumDevs;
end;

procedure InputDeviceName(Device :Integer; Name :PChar);
var
  wicaps :WAVEINCAPS;
begin
  if waveInGetDevCaps(Device, @wicaps, SizeOf(wicaps)) = MMSYSERR_NOERROR then begin
    Move(wicaps.szPName, Name^, Length(String(wicaps.szPName)) + 1);
  end;
end;

//Output
function OutputDeviceCount :Integer;
begin
  Result := waveOutGetNumDevs;
end;

procedure OutputDeviceName(Device :Integer; Name :PChar);
var
  wocaps :WAVEOUTCAPS;
begin
  if waveOutGetDevCaps(Device, @wocaps, SizeOf(wocaps)) = MMSYSERR_NOERROR then begin
    Move(wocaps.szPName, Name^, Length(String(wocaps.szPName)) + 1);
  end;
end;


//****************************************************************************//
//                                                                            //
//                          C H A N N E L    C O U N T                        //
//                                                                            //
//****************************************************************************//

//Input
function InputChannelCount(Device :Integer) :Integer;
var
  wicaps :WAVEINCAPS;
begin
  if waveInGetDevCaps(Device, @wicaps, SizeOf(wicaps)) = MMSYSERR_NOERROR then begin
    Result := wicaps.wChannels;
  end else begin
    Result := 0;
  end;
end;

//Output
function OutputChannelCount(Device :Integer) :Integer;
var
  wocaps :WAVEOUTCAPS;
begin
  if waveOutGetDevCaps(Device, @wocaps, SizeOf(wocaps)) = MMSYSERR_NOERROR then begin
    Result := wocaps.wChannels;
  end else begin
    Result := 0;
  end;
end;

function OutputDigitalChannelCount(Device :Integer) :Integer;
begin
  Result := 0;
end;


//****************************************************************************//
//                                                                            //
//                               D E F A U L T S                              //
//                                                                            //
//****************************************************************************//

//Input
function InputGain(Device, Channel :Integer) :Double;
begin
  Result := 1;
end;

function InputRange(Device :Integer) :Double;
begin
  Result := 1;
end;

function InputDifferential(Device :Integer) :Integer;
begin
  Result := Integer(True);
end;

function InputBits(Device :Integer) :Integer;
begin
  Result := 16;
end;

//Output
function OutputGain(Device, Channel :Integer) :Double;
begin
  Result := 1;
end;

function OutputRange(Device :Integer) :Double;
begin
  Result := 1;
end;

function OutputDifferential(Device :Integer) :Integer;
begin
  Result := Integer(True);
end;

function OutputBits(Device :Integer) :Integer;
begin
  Result := 16;
end;



//****************************************************************************//
//                                                                            //
//                               E X P O R T S                                //
//                                                                            //
//****************************************************************************//
exports


  LibraryInitialize,
  LibraryTerminate,

  LibraryNameLength,
  LibraryName,


  //Input
  InputStart,
  InputStop,

  InputDeviceCount,
  InputDeviceName,
  InputChannelCount,
  InputGain,
  InputRange,
  InputDifferential,
  InputBits,

  ReadDBLChannel,

  AIRead,
  DIRead,

  //Output
  OutputStart,
  OutputStop,

  OutputDeviceCount,
  OutputDeviceName,
  OutputChannelCount,
  OutputDigitalChannelCount,
  OutputGain,
  OutputRange,
  OutputDifferential,
  OutputBits,

  WriteDBLChannel,

  AOWrite,
  DOWrite,

  //Synch
  SynchStart;
  

end.
//******************************************************************************

