"""Date/time conversion between module datetime, Delphi, and "HH:MM:SS"."""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import datetime

epoch = datetime.datetime(1899, 12, 30)

DayPerSec = 1.0 / (3600*24)
DayPerUsec = DayPerSec / 1e6

def DelphiToDatetime(delphi):
    """Delphi stores datetime as double-precision seconds since 12/30/1899."""
    return epoch + datetime.timedelta(delphi)
def DatetimeToDelphi(dt):
    """Delphi stores datetime as double-precision seconds since 12/30/1899."""
    delta = dt - epoch
    return (delta.days
            + delta.seconds * DayPerSec
            + delta.microseconds * DayPerUsec)

# *****************************************************************************

def SecondsToHMS(x):
    """Returns x seconds formatted as "[-] [[HH:]MM:]SS." """
    if x < 0:
        x = -x
        prefix = '-'
    else:
        prefix = ''
    s = int(x) % 60
    m = int(x) / 60
    h, m = m/60, m%60
    if h:
        return "%s%d:%.2d:%.2d" % (prefix,h,m,s)
    elif m:
        return "%s%d:%.2d" % (prefix,m,s)
    else:
        return "%s%d" % (prefix,s)

def HMSToSeconds(hms):
    """Returns the number of seconds in "[-] [[HH:]MM:]SS." """
    result = 0
    part = hms
    while True:
        i = part.find(':')
        if i < 0:
            return 60*result + int(part)
        result = 60*result + int(part[:i])
        part = part[i+1:]
