"""GUI support for user scripting"""

import qub.pyenv
from qub.GTK import *
import os

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

class PythonOut(gtk.Frame):
    """A widget showing sys.stdout."""
    def __init__(self):
        gtk.Frame.__init__(self, 'Output')
        scr = gtk.ScrolledWindow()
        self.txtOutput = gtk.TextView()
        self.txtOutput.set_editable(False)
        SetFixedWidth(self.txtOutput)
        scr.add(self.txtOutput)
        self.txtOutput.show()
        self.output = TextViewAppender(self.txtOutput)
        qub.pyenv.env.OnOutput += self.output
        self.add(scr)
        scr.show()
        self.write = qub.pyenv.env.write # backwards-compatibility
    def detach(self):
        """Stops receiving stdout.  Call this before discarding a short-lived PythonOut."""
        qub.pyenv.env.OnOutput -= self.output

class PythonPrompt(gtk.HBox):
    """A widget which lets the user execute a line of python."""
    def __init__(self):
        gtk.HBox.__init__(self)
        lbl = gtk.Label('>>> ')
        self.pack_start(lbl, False, True)
        lbl.show()
        self.entry = gtk.Entry()
        self.entry.connect('key_press_event', self._onCmdKey)
        self.pack_start(self.entry, True, True)
        self.entry.show()
        self.btnRun = gtk.Button('Run')
        self.btnRun.connect('clicked', self._onPressRun)
        self.pack_start(self.btnRun, False, True)
        self.btnRun.show()
    def _onCmdKey(self, txt, event):
        if event.keyval == keysyms.Return:
            self.run()
            self.entry.select_region(0, -1)
    def _onPressRun(self, btn):
        self.run()
    def run(self, cmd=None):
        """Prints cmd to stdout and runs it."""
        cmd = cmd or self.entry.get_text()
        if cmd:
            qub.pyenv.env.write('>>> %s\n' % cmd)
            try:
                e_val = eval(cmd, qub.pyenv.env.globals)
                if e_val: print repr(e_val)
            except:
                try:
                    exec cmd in qub.pyenv.env.globals
                except:
                    traceback.print_exc()


class PythonEnvironmentDialog(gtk.Dialog):
    """PythonEnvironmentDialog(appName="", parent=None):
    Dialog for editing sys.path extensions and the startup script
    (see qub.pyenv.PythonEnvironment).

    appName: top-level application name, for the title bar
    parent:  parent gtk.Window, for centering the dialog
"""
    def __init__(self, appName="", parent=None):
        appTitle = appName and (appName+" - ")
        gtk.Dialog.__init__(self, appTitle+'Python Environment', parent, gtk.DIALOG_MODAL)
        self._build_ui()
        self.ref = Reffer()
        self.txtPath.OnChange += self.ref(self._onChangePath)
        self.btnReload.connect('clicked', self._onPressReload)
    def _build_ui(self):
        vsplit = gtk.VPaned()
        v = gtk.VBox()
        h = gtk.HBox()
        lbl = gtk.Label('sys.path.append:')
        h.pack_start(lbl, False, True)
        lbl.show()
        self.txtPath = NumEntry("")
        h.pack_start(self.txtPath, True, True)
        self.txtPath.show()
        v.pack_start(h, False, True)
        h.show()
        
        msg = gtk.TextView()
        buf = msg.get_buffer()
        buf.set_text("""The following script is run during startup, and when you press Reload below.""")
        buf.select_range(buf.get_start_iter(), buf.get_end_iter())
        v.pack_start(msg, False, True)
        msg.show()
        sep = gtk.HSeparator()
        v.pack_start(sep, False, True)
        sep.show()
        
        scroll = gtk.ScrolledWindow()
        scroll.set_size_request(600, 300)
        self.txtScript = gtk.TextView()
        SetFixedWidth(self.txtScript)
        self.bufScript = self.txtScript.get_buffer()
        scroll.add(self.txtScript)
        self.txtScript.show()
        v.pack_start(scroll, True, True)
        scroll.show()
        
        vsplit.pack1(v, True, True)
        v.show()
        v = gtk.VBox()
        
        self.txtOut = PythonOut()
        self.txtOut.set_size_request(620, 200)
        v.pack_start(self.txtOut, True, True)
        self.txtOut.show()
        self.prompt = PythonPrompt()
        v.pack_start(self.prompt, False, True)
        self.prompt.show()
        
        vsplit.pack2(v, False, True)
        v.show()
        self.vbox.pack_start(vsplit, True, True)
        vsplit.show()

        self.btnReload = gtk.Button('Reload')
        self.action_area.pack_end(self.btnReload, False, True)
        self.btnReload.show()
        self.add_button('Close', gtk.RESPONSE_ACCEPT)
    def run(self):
        self.txtPath.value = ';'.join(qub.pyenv.env.path)
        buf = self.bufScript
        buf.set_text(qub.pyenv.env.script)
        response = gtk.Dialog.run(self)
        qub.pyenv.env.script = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
        qub.pyenv.env.save()
        return response
    def destroy(self):
        qub.pyenv.env.save()
        self.txtOut.detach()
        gtk.Dialog.destroy(self)
    def _onChangePath(self, txt, val):
        try:
            qub.pyenv.env.path = re.split(" *; *", val)
        except:
            traceback.print_exc()
    def _onPressReload(self, btn):
        buf = self.bufScript
        qub.pyenv.env.script = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
        qub.pyenv.env.save()
        qub.pyenv.env.reload()



class PythonScriptView(gtk.VBox):
    """Widget for editing and running user scripts.

Properties:
    clean:   True if the script has not been modified
    script:  Contents of the script pane (string)
"""
    def __init__(self, vertical=True):
        gtk.VBox.__init__(self)
        self._build_ui(vertical)

        self.btnPlay.connect('clicked', self._onClickPlay)
        self.btnSaveAs.connect('clicked', self._onClickSaveAs)
        self.btnSave.connect('clicked', self._onClickSave)
        self.btnOpen.connect('clicked', self._onClickOpen)
        self.btnNew.connect('clicked', self._onClickNew)
        self.bufScript.connect('changed', self._onScriptChanged)

        self.path, self.fname = '', ''
        self.clean = True

    def _build_ui(self, vertical):
        v = self
        hsplit = gtk.HPaned()
        self.toolbar = gtk.Toolbar()
        self.toolbar.set_style(gtk.TOOLBAR_ICONS)
        self.toolbar.set_tooltips(True)
        self.btnNew = gtk.ToolButton(gtk.STOCK_NEW)
        self.btnNew.show()
        self.toolbar.insert(self.btnNew, -1)
        self.btnOpen = gtk.ToolButton(gtk.STOCK_OPEN)
        self.btnOpen.show()
        self.toolbar.insert(self.btnOpen, -1)
        self.btnSave = gtk.ToolButton(gtk.STOCK_SAVE)
        self.btnSave.show()
        self.toolbar.insert(self.btnSave, -1)
        self.btnSaveAs = gtk.ToolButton(gtk.STOCK_SAVE_AS)
        self.btnSaveAs.show()
        self.toolbar.insert(self.btnSaveAs, -1)
        sep = gtk.SeparatorToolItem()
        sep.show()
        self.toolbar.insert(sep, -1)
        self.btnPlay = gtk.ToolButton(gtk.STOCK_MEDIA_PLAY)
        self.btnPlay.show()
        self.toolbar.insert(self.btnPlay, -1)
        hsplit.pack1(self.toolbar, False, True)
        self.toolbar.show()
        vh = gtk.HBox()
        self.txtName = gtk.Entry()
        self.txtName.set_editable(False)
        vh.pack_start(self.txtName, True, True)
        self.txtName.show()
        lbl = gtk.Label('in')
        vh.pack_start(lbl, False, True)
        lbl.show()
        self.txtPath = gtk.Entry()
        self.txtPath.set_editable(False)
        vh.pack_start(self.txtPath, True, True)
        self.txtPath.show()
        self.btnEnv = gtk.Button('Environment...')
        self.btnEnv.connect('clicked', self._onClickEnv)
        vh.pack_start(self.btnEnv, False, True)
        self.btnEnv.show()
        hsplit.pack2(vh, True, True)
        vh.show()
        v.pack_start(hsplit, False, True)
        hsplit.set_position(200)
        hsplit.show()

        vsplit = (vertical and gtk.VPaned or gtk.HPaned)()
        vsv = gtk.VBox()
        scroll = gtk.ScrolledWindow()
        scroll.set_size_request(200, 200)
        self.txtScript = gtk.TextView()
        SetFixedWidth(self.txtScript)
        self.bufScript = self.txtScript.get_buffer()
        scroll.add(self.txtScript)
        self.txtScript.show()
        vsv.pack_start(scroll, True, True)
        scroll.show()
        
        vsplit.pack1(vsv, True, True)
        vsv.show()
        vsv = gtk.VBox()

        self.txtOut = PythonOut()
        self.txtOut.set_size_request(200, 200)
        vsv.pack_start(self.txtOut, True, True)
        self.txtOut.show()
        self.prompt = PythonPrompt()
        vsv.pack_start(self.prompt, False, True)
        self.prompt.show()
        
        vsplit.pack2(vsv, False, True)
        vsv.show()
        v.pack_start(vsplit, True, True)
        vsplit.show()

    def set_clean(self, clean):
        self._clean = clean
        self.btnSave.set_sensitive(not clean)
    clean = property(lambda s: s._clean, set_clean)
    def get_script(self):
        buf = self.bufScript
        return buf.get_text(buf.get_start_iter(), buf.get_end_iter())
    script = property(get_script)
    def set_file(self, path, fname):
        """Sets the folder and file name.  Doesn't open or save."""
        self.path = path
        self.fname = fname
        self.txtPath.set_text(path)
        self.txtName.set_text(fname)
    def _onScriptChanged(self, buf):
        self.clean = False
    def promptSave(self):
        """If there are changes, ask to save them.
Returns True unless there are changes and the user says Cancel."""
        if not self.clean:
            dlg = gtk.MessageDialog(None, buttons=gtk.BUTTONS_NONE,
                                    message_format='Save changes to python script %s?' % self.fname)
            dlg.add_buttons('Yes',gtk.RESPONSE_YES,
                            'No',gtk.RESPONSE_NO,
                            'Cancel',gtk.RESPONSE_CANCEL)
            response = dlg.run()
            dlg.destroy()
            if response == gtk.RESPONSE_YES:
                return self.saveAs(self.path, self.fname)
            elif response == gtk.RESPONSE_NO:
                return True
            elif response == gtk.RESPONSE_CANCEL:
                return False
        else:
            return True
    def saveAs(self, _path=None, _fname=None):
        """Saves script to _fname in folder _path.  If None specified, prompts the user for a location."""
        if _path and _fname:
            path, fname = _path, _fname
        else:
            dlg = gtk.FileChooserDialog('Save script as...', None, gtk.FILE_CHOOSER_ACTION_SAVE,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
            dlg.set_default_response(gtk.RESPONSE_OK)
            dlg.set_filename(os.path.join(_path or self.path, _fname or self.fname))
            filter = gtk.FileFilter()
            filter.set_name("Python scripts")
            filter.add_pattern("*.py")
            dlg.add_filter(filter)
            dlg.set_current_folder(_path or self.path)
            response = dlg.run()
            if response == gtk.RESPONSE_OK:
                path, fname = os.path.split( dlg.get_filename() )
            dlg.destroy()
            if response != gtk.RESPONSE_OK:
                return False
        try:
            open(os.path.join(path, fname), 'w').write(self.script)
            self.set_file(path, fname)
            self.clean = True
            return True
        except:
            traceback.print_exc()
            return False
    def _onClickNew(self, item):
        self.bufScript.set_text("")
        self.set_file('', '')
        gobject.idle_add(self.set_clean, True)
    def _onClickOpen(self, item):
        try:
            dlg = gtk.FileChooserDialog('Open script...', None, gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
            dlg.set_default_response(gtk.RESPONSE_OK)
            filter = gtk.FileFilter()
            filter.set_name("Python scripts")
            filter.add_pattern("*.py")
            dlg.add_filter(filter)
            if self.path:
                dlg.set_current_folder(self.path)
            response = dlg.run()
            if response == gtk.RESPONSE_OK:
                path, fname = os.path.split(dlg.get_filename())
                self.bufScript.set_text(open(os.path.join(path, fname), 'r').read())
                self.set_file(path, fname)
                self.clean = True
            dlg.destroy()
        except:
            traceback.print_exc()
    def _onClickSave(self, item):
        self.saveAs(self.path, self.fname)
    def _onClickSaveAs(self, item):
        self.saveAs()
    def _onClickPlay(self, item):
        try:
            exec(self.script, qub.pyenv.env.globals)
        except:
            traceback.print_exc()
        qub.pyenv.env.write('Done.\n')
    def _onClickEnv(self, btn):
        d = PythonEnvironmentDialog()
        d.run()
        d.destroy()



class PythonScriptWindow(gtk.Window):
    """Window for editing and running user scripts.

Properties:
    clean:   True if the script has not been modified
    script:  Contents of the script pane (string)
"""
    def __init__(self, appName=""):
        gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)
        appTitle = appName and (appName+" - ")
        self.set_title(appTitle+'Python Scripts')
        self.connect('delete_event', self._onDelete)
        
        self.view = PythonScriptView()
        self.add(self.view)
        self.view.show()
    clean = property(lambda s: s._clean, lambda s,x: s.view.set_clean(x))
    script = property(lambda s: s.view.get_script())
    def set_file(self, path, fname):
        """Sets the folder and file name.  Doesn't open or save."""
        self.path = path
        self.fname = fname
        self.txtPath.set_text(path)
        self.txtName.set_text(fname)
    def _onDelete(self, w, evt):
        try:
            if self.view.promptSave():
                self.hide()
        except:
            traceback.print_exc()
        return True # we have handled it
    def _onScriptChanged(self, buf):
        self.clean = False
    def promptSave(self):
        """If there are changes, ask to save them.
Returns True unless there are changes and the user says Cancel."""
        return self.view.promptSave()



class HelpTestDialog(gtk.Dialog):
    """HelpTestDialog(accept, caption, label, help_msg, bind, write_test, parent=None):
    Dialog to help write and debug one-liners.

    accept: a func(string) -> val; such as qub.accept.acceptF
    caption: top line of text explaining what's wanted
    label: e.g. "f(x) = "
    help_msg: a multiline explanation of what's wanted, with examples
    bind: a func(val) that puts val into the global dict
          e.g. qub.pyenv.env.globals['f'] = val
    write_test: a func(val) -> expr; where expr is a one-liner that tests the global binding
          e.g. 'f(1.0)'
"""
    def __init__(self, accept, caption, label, help_msg, bind, write_test, parent=None):
        gtk.Dialog.__init__(self, 'Python Expression Help', parent, gtk.DIALOG_MODAL)
        self.expr = ""
        self.accept = accept
        self.caption = caption
        self.label = label
        self.help_msg = help_msg
        self.bind = bind
        self.write_test = write_test
        self._args = []
        self._build_ui()
        
        self.ref = Reffer()
        self.txtF.OnChange += self.ref(self._onChangeF)
        self.txtF.OnReject += self.ref(self._onRejectF)
        self.txtF.OnEdit   += self.ref(self._onEditF)
        self.txtF.OnReset  += self.ref(self._onResetF)

        buf = self.txtAbout.get_buffer()
        buf.insert(buf.get_end_iter(), help_msg)

    def _build_ui(self):
        h = gtk.HBox(False)
        hv = gtk.VBox(False)
        self.heading = gtk.Label(self.caption)
        hv.pack_start(self.heading, False, True)
        self.heading.show()
        hvh = gtk.HBox(False)
        lbl = gtk.Label(self.label)
        hvh.pack_start(lbl, False, True)
        lbl.show()
        self.txtF = NumEntry(None, self.accept)
        hvh.pack_start(self.txtF, True, True)
        self.txtF.show()
        hv.pack_start(hvh, False, True)
        hvh.show()
        self.txtOut = PythonOut()
        self.txtOut.set_size_request(620, 200)
        hv.pack_start(self.txtOut, True, True)
        self.txtOut.show()
        self.prompt = PythonPrompt()
        hv.pack_start(self.prompt, False, True)
        self.prompt.show()
        h.pack_start(hv, True, True)
        hv.show()
        scr = gtk.ScrolledWindow()
        scr.set_size_request(300, 300)
        self.txtAbout = gtk.TextView()
        self.txtAbout.set_editable(False)
        self.txtAbout.set_wrap_mode(gtk.WRAP_WORD)
        scr.add(self.txtAbout)
        self.txtAbout.show()
        h.pack_start(scr, False, True)
        scr.show()
        self.vbox.pack_start(h, True, True)
        h.show()
        
        self.btnCancel = self.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT)
        self.btnOK = self.add_button(gtk.STOCK_OK, gtk.RESPONSE_ACCEPT)
            
    def run(self, expr):
        """Presents expr for testing and editing.
If the return value is gtk.RESPONSE_ACCEPT, self.expr has the edited text."""
        self.expr = expr
        try:
            self.txtF.setValue(self.accept(expr), expr)
            self._onChangeF(self.txtF, self.txtF.value)
        except Exception, e:
            self._onRejectF(self.txtF, expr, e)
            self.txtF.set_text(expr)
        try:
            self.show()
            response = gtk.Dialog.run(self)
        except:
            traceback.print_exc()
            response = gtk.RESPONSE_REJECT
        return response
    def destroy(self):
        self.txtOut.detach()
        gtk.Dialog.destroy(self)
    def _onEditF(self, txt, text):
        self.btnOK.set_sensitive(False)
    def _onRejectF(self, txt, text, exc):
        self.btnOK.set_sensitive(False)
        traceback.print_exc()
    def _onChangeF(self, txt, val):
        self.btnOK.set_sensitive(True)
        self.bind(val)
        try:
            if self._args != val.args:
                self.prompt.entry.set_text(self.write_test(val))
        except:
            self.prompt.entry.set_text(self.write_test(val))
        self.expr = txt.text
        self.prompt.run()
    def _onResetF(self, txt, val):
        self.btnOK.set_sensitive(True)

class HelpTestButton(gtk.Button):
    """A gtk.Button linked to a qub.GTK.NumEntry, labeled '?',
which brings up an appropriate HelpTestDialog.

Properties:
    entry:    the linked NumEntry
    caption: top line of text explaining what's wanted
    label: e.g. "f(x) = "
    help_msg: a multiline explanation of what's wanted, with examples
    bind: a func(val) that puts val into the global dict
          e.g. qub.pyenv.env.globals['f'] = val
    write_test: a func(val) -> expr; where expr is a one-liner that tests the global binding
          e.g. 'f(1.0)'
"""
    def __init__(self, entry):
        gtk.Button.__init__(self, '?')
        self.connect('clicked', self._onPress)
        self.entry = entry
        self.caption = ""
        self.label = 'f = '
        self.help_msg = ""
        self.bind = lambda x: qub.pyenv.env.globals.__setitem__('f', x.f)
        self.write_test = lambda x: ""
    def _onPress(self, btn):
        dlg = HelpTestDialog(self.entry.accept, self.caption, self.label, self.help_msg, self.bind, self.write_test)
        if gtk.RESPONSE_ACCEPT == dlg.run(self.entry.get_text()):
            self.entry.set_text(dlg.expr) # setValue(self.entry.accept(dlg.expr), dlg.expr, True)
            self.entry.onParse()          # self.entry.OnChange(self.entry, self.entry.value)
        dlg.destroy()
