"""pygtk GUI for qub.settings."""

#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import pygtk
pygtk.require('2.0')
import gtk
from gtk import gdk
from gtk import keysyms
import gobject

import qub.settings
from qub.settings import *
from qub.GTK import *
import qub.treeGTK
import qub.tree

class PresetsMenu(gtk.EventBox):
    """Combo box for managing a category of presets/settings.

constructor:
    PresetsMenu(name, appName="", parent=None, includeF=lambda name:True, mask='.qprs')
        name        the category name
        appName     top-level application name, for dialog titles
        parent      parent gtk.Window, for centering dialogs
        includeF    filter function; we show only presets for which includeF(name) is True
        mask        preferred file extension for import/export

notes:
    If you have a presets menu for a category, you must
        * keep category.active up-to-date, in case the user saves a preset
        * provide category.OnSet, in case the user chooses a preset
"""
    def __init__(self, name, appName="", parent=None, includeF=lambda name:True, mask='.qprs'):
	MapSettings()
        gtk.EventBox.__init__(self)
        self._name = name
        self._appName = appName
        self._parent = parent
        self._include = includeF
        self._mask = mask
        self._perma = 3
        self._count = 0
        self._box = gtk.combo_box_new_text()
        self.add(self._box)
        self._box.show()
        self.connect('enter-notify-event', self.enter_notify_event)
        self._box.connect('changed', self.combo_changed)
        self._box.append_text('Presets...')
        self._box.append_text('Add to menu...')
        self._box.append_text('Manage...')
        self._box.set_active(0)
    def enter_notify_event(self, widget, event):
        self._box.set_active(0)
        for i in xrange(self._count):
            self._box.remove_text(self._perma)
        self._count = 0
        for name in SettingsMgr.listNames(self._name):
            if self._include(name):
                self._box.append_text(name)
                self._count += 1
    def combo_changed(self, combo):
        active_i = combo.get_active()
        if active_i == 1:
            dlog = InputDialog('Add preset to menu...', None, 'Preset name:', '')
            if dlog.run() == gtk.RESPONSE_ACCEPT:
                name = dlog.text
                if name: # should check for special characters
                    SettingsMgr[self._name].save(name)
            dlog.destroy()
        elif active_i == 2:
            dlog = PresetsDialog(self._name, self._parent, self._appName, self._mask)
            dlog.run()
            dlog.destroy()
        elif active_i >= self._perma:
            SettingsMgr[self._name].load(combo.get_active_text())
        combo.set_active(0)
        
class PresetsDialog(gtk.Dialog):
    """Manage Presets dialog"""
    def __init__(self, cat, parent, appName="", mask=".qprs"):
        appPrefix = appName and (appName+" - ") or ""
        gtk.Dialog.__init__(self, appPrefix+"Manage Presets", parent, gtk.DIALOG_MODAL)
        self.cat = cat
        self.mask = mask
        self.ref = Reffer()

        btns = gtk.HBox(False, 10)
        self.btnImport = gtk.Button('Import...')
        self.btnImport.connect('clicked', self._onClickImport)
        btns.pack_start(self.btnImport, False, True)
        self.btnImport.show()
        self.btnExport = gtk.Button('Export...')
        self.btnExport.connect('clicked', self._onClickExport)
        btns.pack_start(self.btnExport, False, True)
        self.btnExport.show()
        self.btnRename = gtk.Button('Rename...')
        self.btnRename.connect('clicked', self._onItemRename)
        btns.pack_start(self.btnRename, False, True)
        self.btnRename.show()
        self.btnDelete = gtk.Button('Delete')
        self.btnDelete.connect('clicked', self._onItemDelete)
        btns.pack_start(self.btnDelete, False, True)
        self.btnDelete.show()
        self.vbox.pack_start(btns, False, True)
        btns.show()
        
        lr = gtk.HPaned()
        scr = gtk.ScrolledWindow()
        scr.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scr.set_size_request(200, 200)
        self.presets = SimpleList(sortable=False)
        for name in SettingsMgr.listNames(cat):
            self.presets.append(name)
        self.presets.OnSelect += self.ref(self._onSelect)
        scr.add(self.presets)
        self.presets.show()
        lr.pack1(scr, False, True)
        scr.show()
        
        scr = gtk.ScrolledWindow()
        scr.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.treeView = qub.treeGTK.TreeView()
        scr.add(self.treeView)
        self.treeView.show()
        lr.pack2(scr, True, True)
        scr.show()
        self.vbox.pack_start(lr, True, True)
        lr.show()
        
        self.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_REJECT)
        
        popup = gtk.Menu()
        item = gtk.MenuItem('Rename')
        item.connect('activate', self._onItemRename)
        item.show()
        popup.append(item)
        item = gtk.MenuItem('Delete')
        item.connect('activate', self._onItemDelete)
        item.show()
        popup.append(item)
        self.presets.popup = popup
    def _onSelect(self, list, index):
        if index == self._lastIndex:
            return
        if self.promptSave():
            self._lastIndex = index
            self._lastName = self.presets.active
            self.treeView.root = SettingsMgr[self.cat].read(self.presets.active)
            self.btnExport.set_sensitive(index >= 0)
            self.btnRename.set_sensitive(index >= 0)
            self.btnDelete.set_sensitive(index >= 0)
        else:
            gobject.idle_add(self._fixSel)
    def _fixSel(self):
        self.presets.index = self._lastIndex
    def run(self):
        self._lastIndex = -1
        if len(SettingsMgr.listNames(self.cat)):
            if self.presets.index < 0:
                self.presets.index = 0
            self._onSelect(self.presets, self.presets.index)
        response = gtk.Dialog.run(self)
        self.promptSave(False)
        return response
    def promptSave(self, canCancel=True):
        if self._lastIndex < 0 or not self.treeView.changed:
            return True
        dlg = gtk.MessageDialog(self, buttons=gtk.BUTTONS_NONE,
                                message_format='Save changes to %s?' % self._lastName)
        dlg.add_buttons('Yes',gtk.RESPONSE_YES,
                        'No',gtk.RESPONSE_NO)
        if canCancel:
            dlg.add_buttons('Cancel',gtk.RESPONSE_CANCEL)
        response = dlg.run()
        dlg.destroy()
        if response == gtk.RESPONSE_YES:
            SettingsMgr.save(self.treeView.root, self.cat, self._lastName)
            return True
        elif response == gtk.RESPONSE_CANCEL:
            return False
        elif response == gtk.RESPONSE_NO:
            return True
    def _onClickImport(self, btn):
        dlg = gtk.FileChooserDialog('Import presets...', self, gtk.FILE_CHOOSER_ACTION_OPEN,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("Presets files")
        filter.add_pattern("*"+self.mask)
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dlg.add_filter(filter)
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            try:
                qprs = qub.tree.Open(dlg.get_filename(), True)
                qprs.close()
                if str(qprs['Target'].data) != self.cat:
                    mdlg = gtk.MessageDialog(self, buttons=gtk.BUTTONS_NONE,
                     message_format='This looks like a preset for a different category, "%s."  Import it anyway?'%str(qprs['Target'].data))
                    mdlg.add_buttons('Yes', gtk.RESPONSE_YES,
                                     'No', gtk.RESPONSE_NO)
                    if gtk.RESPONSE_NO == mdlg.run():
                        qprs = None
                    mdlg.destroy()
                if qprs:
                    name = str(qprs['Label'].data) or "untitled"
                    try:
                        prev_ix = SettingsMgr.listNames(self.cat).index(name)
                    except:
                        prev_ix = -1
                    SettingsMgr.save(qprs['Presets'], self.cat, name)
                    if prev_ix >= 0:
                        self.presets.index = prev_ix
                    else:
                        self.presets.append(name)
                        self.presets.index = len(self.presets.model) - 1
            except Exception, e:
                traceback.print_exc()
        dlg.destroy()
    def _onClickExport(self, btn):
        dlg = gtk.FileChooserDialog('Export presets as...', self, gtk.FILE_CHOOSER_ACTION_SAVE,
                                    buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        dlg.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("Presets files")
        filter.add_pattern('*'+self.mask)
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dlg.add_filter(filter)
        response = dlg.run()
        if response == gtk.RESPONSE_OK:
            try:
                qprs = qub.tree.Node('Presets')
                qprs['Target'].data = self.cat
                qprs['Label'].data = self.presets.active
                qprs.appendClone(self.treeView.root).name = 'Presets'
                filename = dlg.get_filename()
                if not os.path.splitext(filename)[1]:
                    filename = filename + self.mask
                qprs.saveAs(filename)
            except Exception, e:
                traceback.print_exc()
        dlg.destroy()
    def _onItemRename(self, item):
        if self.presets.index < 0:
            return
        dlog = InputDialog('Rename preset...', None, 'Preset name:', self.presets.active)
        try:
            if dlog.run() == gtk.RESPONSE_ACCEPT:
                name = dlog.text
                if name: # should check for special characters
                    SettingsMgr.rename(self.cat, self.presets.active, name)
                    self.presets.edit(self.presets.index, name)
        except Exception, e:
            d = gtk.MessageDialog(self, buttons=gtk.BUTTONS_NONE, message_format=str(e))
            d.add_buttons('OK', gtk.RESPONSE_OK)
            d.run()
            d.destroy()
        dlog.destroy()
    def _onItemDelete(self, item):
        if self.presets.index < 0:
            return
        dlg = gtk.MessageDialog(self, buttons=gtk.BUTTONS_NONE,
                                message_format="Permanently delete %s?" % self.presets.active)
        dlg.add_buttons('Yes',gtk.RESPONSE_YES,
                        'No', gtk.RESPONSE_NO)
        response = dlg.run()
        dlg.destroy()
        if response == gtk.RESPONSE_NO:
            return
        try:
            self.treeView.changed = False
            self._lastIndex = -1
            if SettingsMgr.delete(self.cat, self.presets.active):
                self.presets.remove(self.presets.index)
        except:
            traceback.print_exc()

def MapSettings():
    """Sets up global SettingsMgr."""
    global SettingsMgr
    qub.settings.InitSettings()
    SettingsMgr = qub.settings.SettingsMgr


