import QUB
from Buttons import *
from Dialogs import *
from qubtree import Node
from math import *
from itertools import *
import os
import tempfile
import time

def children(tree, name=None):
    if name == None:
        child = tree.child
    else:
        child = tree.find(name)
    while not child.isNull:
        yield child
        if name == None:
            child = child.sibling
        else:
            child = child.nextSameName()


def TBurst(props):
    reIdealize = props['reidl']
    T1c, T2c, I1, I2 = props['T1c'], props['T2c'], props['I1'], props['I2']
    T1 = T1c + 273.15
    T2 = T2c + 273.15
    if I1 == I2:
        Tmean = (T1 + T2) / 2.0
        CurrentToK = lambda I: Tmean
    else:
        Tslope = (T2 - T1) / (I2 - I1)
        Tintercept = T1 - Tslope * I1
        CurrentToK = lambda I: Tslope * I + Tintercept

    seg_Ti = []
    ab_seg_Af = {}
    ab_seg_Ab = {}
    ab_seg_vH = {}
    
    modelObj = QUB.ModelView.currModel
    if not modelObj:
        QUB.Report('no model')
        return
    model = modelObj.getAsNode()
    QUB.ModelView.New(os.path.join(tempfile.gettempdir(), 'QuB - TBurst temp %f.qmf' % time.time()))
    modelObj = QUB.ModelView.currModel
    model_amps = model['Amps'].data

    state_x = []
    for state in children(model['States'], 'State'):
        state_x.append(state['x'].data[0])    
    for rate in children(model['Rates'], 'Rate'):
        a, b = rate['States'].data[:2]
        if state_x[a] > state_x[b]:
            a, b = b, a
        ab_seg_Af[(a,b)] = []
        ab_seg_Ab[(a,b)] = []
        ab_seg_vH[(a,b)] = []
        
    datasource = QUB.GetDataSource()

    if not reIdealize:
        try:
            datasets = QUB.getChosenIdealization()
            for ds in children(datasets, 'DataSet'):
                for seg in children(ds, 'Segment'):
                    amps = seg['amp'].data
                    xamp = amps[1] - amps[0]
                    T = CurrentToK(xamp)
                    if T != 0.0:
                        Ti = 1.0e3 / T
                    else:
                        Ti = 0.0
                    seg['xamp'].data = xamp
                    seg['T'].data = T
                    seg['Tinv'].data = Ti
                    seg_Ti.append(Ti)
        except:
            reIdealize = True
            
    if reIdealize:
        datasets = QUB.getChosenData()    

        #  each segment: idealize; save amp (open - closed); save T(amp), Ti = 1e3/T; update model amps
        QUB.SetDataSource('sel')
        for ds in children(datasets, 'DataSet'):
            # show this file
            if QUB.DataView.currData.fileName != str(ds['FileName'].data):
                for i in xrange(len(QUB.DataView)):
                    if QUB.DataView[i].fileName == str(ds['FileName'].data):
                        QUB.DataView.currIndex = i
                        break
            for seg in children(ds, 'Segment'):
                QUB.DataView.currData.select(seg.data[0], seg.data[1])
                Run('Modeling', 'Idealize')
                results = QUB.ResultsView.currResults.node
                results['Hold'].data = 0
                amps = results['DataSet']['Segment']['amp'].data
                xamp = amps[1] - amps[0]
                T = CurrentToK(xamp)
                if T != 0.0:
                    Ti = 1.0e3 / T
                else:
                    Ti = 0.0
                seg['amp'].data = amps[:2]
                seg['xamp'].data = xamp
                seg['T'].data = T
                seg['Tinv'].data = Ti
                seg_Ti.append(Ti)
                model_amps[:2] = amps[:2]
                modelObj.setAsNode(model)

    #  MIL it all, "opt", "separate"
    #  store each segment's f_ab, b_ab, Af=log(f_ab), Ab=log(b_ab), vH=log(f_ab/b_ab) where state a.x <= b.x
    milOptions = Node()
    milOptions['Mode'].data = 'optimize'
    milOptions['use segments'].data = 'separately'
    QUB.SetDataSource(datasource)
    Run('Modeling', 'MIL', milOptions)
    results = QUB.ResultsView.currResults.node
    for ds_out, ds_mil in izip(children(datasets, 'DataSet'), children(results, 'DataSet')):
        for seg_out, seg_mil in izip(children(ds_out, 'Segment'), children(ds_mil, 'Segment')):
            for rate in children(seg_mil['ModelFile']['Rates'], 'Rate'):
                a, b = rate['States'].data[:2]
                k0f, k0b = rate['k0'].data[:2]
                if state_x[a] > state_x[b]:
                    a, b, k0f, k0b = b, a, k0b, k0f
                seg_out['f_%d%d'%(a+1,b+1)].data = k0f
                seg_out['b_%d%d'%(a+1,b+1)].data = k0b
                seg_out['ln_f_%d%d'%(a+1,b+1)].data = log(k0f)
                ab_seg_Af[(a,b)].append(log(k0f))
                seg_out['ln_b_%d%d'%(a+1,b+1)].data = log(k0b)
                ab_seg_Ab[(a,b)].append(log(k0b))
                seg_out["van'tHoff_%d%d"%(a+1,b+1)].data = log(k0f / k0b)
                ab_seg_vH[(a,b)].append(log(k0f / k0b))
                
    datasets['ResultType'].data = 'TBurst'
    datasets['Hold'].data = 1
    
    for a,b in ab_seg_Af.keys():
        for yseries in ['ln_f_%d%d'%(a+1,b+1),
                        'ln_b_%d%d'%(a+1,b+1),
                        "van'tHoff_%d%d"%(a+1,b+1)]:
            entry = datasets['DefaultChartsVs'].append('Tinv')
            entry.data = yseries
            entry['AutoFit']['Curve'].data = 'Linear'
            entry['AutoFit']['Params'].data = [1.0, 1.0]
            entry['AutoFit']['Const'].data = [0, 0]
    
    QUB.ResultsView.addResults( datasets )
        


AddTaskButton( 'TBurst', # button text
               TBurst,   # function
               'TBurst',       # short text for small buttons
               0,            # show settings dialog on left-click?
               1,            # show settings dialog on right-click?
               "Create Arrhenius and van't Hoff plots from bursts at varying temperatures.",   # tool-tip
               "Create Arrhenius and van't Hoff plots from bursts at varying temperatures.",   # description in "Edit Buttons"
               "TBurst properties:", # caption for the settings dialog
               [#DialogLabel('Linear relationship between temperature and current:'),
                DialogCheckBox("reidl", "Re-idealize", False),
                DialogText("T1c", "Temp 1 (C):", 20.0, acceptFloat),
                DialogText("I1", "Excess Current 1 (pA):", 1.0, acceptFloat),
                DialogText("T2c", "Temp 2 (C):", 30.0, acceptFloat),
                DialogText("I2", "Excess Current 2 (pA):", 0.8, acceptFloat)],
               'Modeling', DataChannel=True, DataList=True)
