#~ Cube2.py
#~ Charles Bowman
#~ February 2005

#~ Plot Channel Records from Qub
	# To use: make a .txt file from Qub using Extract
	#~ The ordering of the columns reflects the order in which the A/D channels are polled (I guess).
	#~ So, A/D Ch 0 is column 2 below, A/D Ch 1 is column 3, etc
	#~ Qub hardware configuration settings are listed at the end of this file...
	#~ My input txt file from have the following columns:
		#~ column 	1	time (sec)
		#~ column 	2	membrane potential (mV)
		#~ column	3	current (pA)
		#~ column	4	pressure (mmHg)
		#~ column	5	patch clamp gain (see below for table)
	#~ If you do not have patch clamp gain, comment out the relevent section below...
	#~ If your ordering is different, rearrange the relevent sections of the program below
	
	#~ Output: an eps file showing the data. View using ghostview.

#~ Requirements:
	#~ Python: http://www.pyton.org
	#~ PyX: http://pyx.sourceforge.net/
	#~ Ghostscript: http://www.cs.wisc.edu/~ghost/doc/AFPL/get814.htm
	#~ Install Ghostcript. Then install GSview:
	#~ GSview: http://www.cs.wisc.edu/~ghost/gsview/index.htm

from pyx import *
from math import exp
import sys,re

text.set(mode='latex')
#~ text.preamble(r'\usepackage{fancybox}')

#~ Change this stuff.....
#~ =========================================	
pathToFile = 'D:\\Charley\\AcquisitionInput\\'
TextFileName = "Feb82005cx"  # name of txt file holding data (w/o .txt extension)
MinimumTime = 0  # min time in seconds
MaximumTime = 18  # max time in seconds
DecimateEvery = 10  # plot every xth point
SubtractFirstTime =       390.015	  # subtract first time from all other times......
#========================================
Max_mV = 2  # max membrane potential
Min_mV = -2 # min membrane potential
Max_pA = 300  # max membrane current;   assumes patch clamp gain = 100 mV/pA; see below
Min_pA = 0 # min membrane current;   assumes patch clamp gain = 100 mV/pA; see below
Max_mmHg = 50 # max pressure
Min_mmHg  = -50  # min pressure
Min_Gain = .1  # min patch clamp gain
Max_Gain = 1000 # max patch clamp gain
#=======================================

PlotWidth = 15 # (cm)

#~ No changes below this point.....
#~ ===========================
c = canvas.canvas()

epsOutputFileName = TextFileName
FinalTextFileName = pathToFile + TextFileName + '.txt'
print FinalTextFileName
SubtractFirstTime = str(SubtractFirstTime)
#================
# Read in the first and last times...from the data file
beginTimeC = re.compile(r'\d+.\d+')
f = open(FinalTextFileName,'r')
ml = f.readlines()
findBeginTime = beginTimeC.search(ml[0])
MinStartTime = findBeginTime.group(0)
z = len(ml)-3
findEndTime = beginTimeC.search(ml[z])
MaxEndTime = findEndTime.group(0)
MaxDeltaTime = float(MaxEndTime) - float(MinStartTime)
PlotDeltaTime = MaximumTime - MinimumTime
#===================================

print 'Min Start Time: ',MinStartTime
print '  Max End Time: ',MaxEndTime
print 'Max Delta Time: ',MaxDeltaTime
print ''
print '  Plot Delta Time: ',PlotDeltaTime,' sec'
print 'Plot Minimum Time: ',MinimumTime
print 'Plot Maximum Time: ',MaximumTime
print ''
QubStartTime = str(MinimumTime+float(MinStartTime))
QubEndTime = str(MaximumTime + float(MinStartTime))
print 'Qub Start Time: ',QubStartTime,' sec'
print '  Qub End Time: ',QubEndTime
f.close()

#~ =======================
#~ Label the Graph
File = TextFileName + '.txt'
c.text(10,20,File)
stuff = 'Qub Start Time: '+QubStartTime+r' sec'
c.text(10,19.0,stuff)
morestuff = ' Qub End Time: '+QubEndTime
c.text(10,18.4,morestuff)
#~ ======================

SubtractFirstTime = MinStartTime


#         Voltage,   column 2
voltage = graph.graphxy(xpos=1,ypos=13,width=PlotWidth,height=2,ratio=None,x2=None,y2=None,
	x=graph.axis.lin(title="",min=MinimumTime,max=MaximumTime,painter=None),
	y=graph.axis.linear(title = 'potential (mV)',min=Min_mV,max=Max_mV))
voltage.plot(graph.data.file(FinalTextFileName,every=DecimateEvery,x="$1 - "+SubtractFirstTime,y=2),
	[graph.style.line(lineattrs=[style.linewidth.Thin])])
c.insert(voltage)

#~ print voltage.vpos(0,0)

#           Current,   column 3
current = graph.graphxy(xpos= voltage.vpos(0,0)[0],ypos=10,width=PlotWidth,height=2,x2=None,y2=None,
	x=graph.axis.lin(title="",min=MinimumTime,max=MaximumTime,painter=None),
	y=graph.axis.linear(title = 'current (pA)',min=Min_pA,max=Max_pA))
current.plot(graph.data.file(FinalTextFileName,every=DecimateEvery,x="$1 - "+SubtractFirstTime,
	y="$3*100"),
	[graph.style.line(lineattrs=[style.linewidth.Thin])])
	#~ [graph.style.symbol(graph.style.symbol.circle,size=0.05,
	#~ symbolattrs=[deco.stroked])])
				#~ Current Scaling:
				#~ if patch clamp gain  is 100 mV/pA:  leave alone
				#~ if patch clamp gain is 50 mV/pA: y = y*2
				#~ if patch clamp gain is 10 mV/pA: y = y*10
				#~ if patch clamp gain is 1 mV/pA: y = y*100

c.insert(current)
#~ print current.vpos(0,0)

#             Pressure,  column 4
pressure = graph.graphxy(xpos = voltage.vpos(0,0)[0],ypos=6,width=PlotWidth,height=2,x2=None,y2=None,
	x=graph.axis.lin(title="time (sec)",min=MinimumTime,max=MaximumTime),
	y=graph.axis.linear(title = 'pressure (mmHg)',min=Min_mmHg,max=Max_mmHg))
pressure.plot(graph.data.file(FinalTextFileName, every=DecimateEvery,x="$1 -"+SubtractFirstTime,y=4),
	[graph.style.line(lineattrs=[style.linewidth.Thin])])
#~ print dir(pressure)
#~ print pressure.vpos(pressure.pos(50,30)[0],pressure.pos(50,30)[1])
c.insert(pressure)

#~ print pressure.vpos(0,0)

#             Patch Clamp Gain,  column 5
#~ PatchClampGain = graph.graphxy(xpos=voltage.vpos(0,0)[0],ypos=2,width=PlotWidth,height=2,x2=None,y2=None,
	#~ x=graph.axis.lin(title="time (sec)",min=MinimumTime,max=MaximumTime),
	#~ y=graph.axis.log(title = 'clamp gain (mV/pA)',min=Min_Gain,max=Max_Gain))
#~ PatchClampGain.plot(graph.data.file(FinalTextFileName, every=DecimateEvery,x="$1 -"+SubtractFirstTime,y="0.0221*exp(1.5334*$5)", context={"exp(1.5334*$5)":exp(1.5334*5)}),
	#~ [graph.style.line(lineattrs=[style.linewidth.Thin])])
#~ c.insert(PatchClampGain)

#~ y=graph.axis.lin(title = 'clamp gain (mV/pA)',min=Min_Gain,max=Max_Gain))
#~ PatchClampGain.plot(graph.data.file(FinalTextFileName, every=DecimateEvery,x="$1 -"+SubtractFirstTime,y=5),
	#~ [graph.style.line(lineattrs=[style.linewidth.Thin])])

#~ print canvas.canvas.bbox(c)

#~ [line(lineattrs=[linewidth.Thick])])
#~ g.plot(graph.data.file("noname8.txt", x=6, y=3),
	#~ [graph.style.symbol(graph.style.symbol.circle,size=0.1,
	#~ symbolattrs=[deco.stroked])])


#~ a = graph.axis.axis.tick(38000)
#~ print a

# graph styles can be modified by a second parameter to the plot method:
	# g.plot(graph.data.file("minimal.dat", x=1, y=2), graph.line())

#~ h.writeEPSfile('textStuff.eps',rotated=1)
#~ c.writeEPSfile('textStuff.eps',rotated=1)
#~ final.insert(epsfile.epsfile(0,1,'voltage.eps'))   #  x,y in units of cm; adjust to line up graphs' y-axes.
#~ final.insert(epsfile.epsfile(5,1.177,'current.eps'))
#~ final.insert(epsfile.epsfile(9,1,'pressure.eps'))
#~ final.insert(epsfile.epsfile(20,1,'textStuff.eps'))
#~ final.insert(epsfile.epsfile(13,1,'patchClampGain.eps'))

c.writeEPSfile(epsOutputFileName,rotated=1)  # rotated =1 makes Landscape mode
print ''
print 'View ',epsOutputFileName+'.eps with ghostView'
print ''

			#~ Patch Clamp Gain   Axopatch Telegraph Output
			#~ 0.5 mV/pA            	2.0 volts
			#~ 1                       	2.5
			#~ 2                       	3.0
			#~ 5                       	3.5
			#~ 10                      	4.0
			#~ 20                      	4.5
			#~ 50                      	5.0
			#~ 100                    	5.5
			#~ 200                    	6.0
			#~ 500                    	6.5

		#~ Best Fit:
		#~ Patch Clamp Gain = 0.0221 * exp(1.5334 * volts)

#~ 				Qub Hardware  Configuration Setup:
					#~ Inputs
	#~ I/O Channel	Input		Unit		Gain		Calibration Factor
	#~ 0			Voltage	mV		0.5		0.01 (volts/unit)
	#~ 1			Current	pA		0.5		0.1
	#~ 2			Pressure	mmHg		1.0		0.018
	#~ 3		Patch Gain		Volts		0.5		1.0
	
						#~ Outputs
	#~ I/O Channel	Output	Increment	Calibration Factor
	#~ 0			Vout		0.1		0.025 volts/unit
	#~ 1			Pout		0.1		0.009


#~ mytexter = graph.axis.texter.decimal(equalprecision=0,thousandsep=',')
#~ myXparter = graph.axis.parter.linear(tickdist=[60,6])

#~ myYparter = graph.axis.parter.linear(tickdist=[10,1])

#~ g.plot(graph.data.file("noname8xx.txt", x=1, y=3),
	#~ [graph.style.symbol(graph.style.symbol.circle,size=0.05,	
	#~ symbolattrs=[deco.filled])])
