import math
import zlib
import struct
    
### -------------------------------------------------------------------------
### create a SWF movie from idealized data.
### Insert a frame (nClasses+1) on segment changes 
#
# IntToBin( i, iBits ) :
# SWFFB( nVal, iBits ) :
# BinToSWFBit( c ) :
# SWFRectangle(iBits, ix0, ix1, iy0, iy1 ) :
# SWFMatrix( iXScale, iYScale, iXSkew, iYSkew, iXTranslate, iYTranslate ) :
# SWFShowFrame(iCount) :
# SWFEndFile() :
# RGB(r,g,b) :
# RGBA(r,g,b) :
# SWFDefineBits2(id, iWid, iHgt, cPalette, cImage) :
# 
# SWFBitmapShapeWithStyle( iCharID, iX0, iX1, iY0, iY1 ) :
#   SWFFillStyleBitmap( iCharID ) :
#   SWFLineStyle( iPixelWidth, r,g,b ) :
#
# SWFPlaceObject2( iDepth, iCharID ) :
# SWFRemoveObject2( iDepth ) :
# SWFSetBGR( r,g,b ) :
# BMPLoad(cFile,iID) -> (iWid, iHgt, cImage)
# SWFFile( cRct, iFrameRate, iFrameCount , cData ) :

#-- convert int to binary string.  Ex : (25,6) -> '011001'
def IntToBin( i, iBits ) :
    i=int(i)
    iBits=int(iBits)
    c=''
    for iBit in range(iBits):
        if i & (2**(iBits-iBit-1)) :
            c=c+'1'
        else:
            c=c+'0'
    return c

#-- convert float to binary string
# Float is stored in 16+iBits bits - first iBits are integer, next 16 are fraction of 2^16
def SWFFB( nVal, iBits ) :
    return IntToBin(int(nVal),iBits) + IntToBin(int((nVal-math.floor(nVal))*(2**16)),16)
    
#-- Convert binary string to ascii.  Ex : '01000001' -> 'A'
def BinToSWFBit( c ) :
    cc = '' 
    c = c + '00000000'
    while len(c)>8 :
        cc = cc + chr(int(c[0])*128 +int(c[1])*64 +int(c[2])*32 +int(c[3])*16 +int(c[4])*8 +int(c[5])*4 +int(c[6])*2 +int(c[7]))
        c = c[8:]        
    return cc

#-- Create SWF rectagle string : 1st 5 bits are number of bits for each subsequent
#-- value.  Followed by x0,x1,y0,y1 and padded with 0's to byte length.
def SWFRectangle(iBits, ix0, ix1, iy0, iy1 ) :
    if iBits == 0 :
        iBits=math.ceil(math.log(max(ix0,ix1,iy0,iy1),2))+1
    return BinToSWFBit(IntToBin(iBits,5)+IntToBin(ix0,iBits)+IntToBin(ix1,iBits)+IntToBin(iy0,iBits)+IntToBin(iy1,iBits))

#-- 
def SWFMatrix( nXScale, nYScale, nXSkew, nYSkew, iXTranslate, iYTranslate ) :
    c = ''

    if nXScale==-1 :
        c='0'
    else :
        iBits=int(math.ceil(math.log(max(nXScale*20.0,nYScale*20.0),2)))+1
        c = '1' + IntToBin(iBits+16,5) + SWFFB(nXScale*20.0,iBits) +SWFFB(nYScale*20.0,iBits)

    if nXSkew==-1 :
        c=c+'0'
    else :
        iBits=int(math.ceil(math.log(max(iXSkew,iYSkew),2)))+1
        c = c + '1' +IntToBin(iBits+16,5) +SWFFB(nXSkew,iBits) +SWFFB(nYSkew,iBits)

    if iXTranslate==-1 :
        c=c+'00000'
    else :
        iBits=int(math.ceil(math.log(max(iXTranslate*20,iYTranslate*20),2)))+1
        c=c+IntToBin(iBits,5)+IntToBin(int(iXTranslate*20),iBits)+IntToBin(int(iYTranslate*20),iBits)
    return BinToSWFBit(c)        

#-- Create string of <iCount> SWF file Show Frame commands
def SWFShowFrame(iCount) :
    return (chr(64)+chr(0))*iCount

#-- Required at end of SWF file
def SWFEndFile() :
    return chr(0)+chr(0)

#-- Convert (R,G,B) to 4 byte string (alpha=255)
def RGB(r,g,b) :
    return struct.pack('l',r*0x10000+g*0x100+b)[0:3]
def RGBA(r,g,b) :
    return struct.pack('l',0xFF000000+r*0x10000+g*0x100+b)

#-- Create SWF lossless Bitmap type 3 - 256 color bitmap index
def SWFDefineBits2(id, iWid, iHgt, cPalette, cImage) :
    cComp = zlib.compress(cPalette+cImage,9)
    return chr(63)+chr(9) +struct.pack('l',8+len(cComp)) +struct.pack('h',id) +chr(3) +struct.pack('h',iWid) +struct.pack('h',iHgt) + chr(len(cPalette)/4-1) + cComp
#-- Create SWF lossless Bitmap type 5 - RGBA array (Alpha   .Blue.Green.Red)
def SWFDefineBits2a(id, iWid, iHgt, cImage) :
    cComp = zlib.compress(cImage,9)
    return chr(63)+chr(9) +struct.pack('l',7+len(cComp)) +struct.pack('h',id) +chr(5) +struct.pack('h',iWid) +struct.pack('h',iHgt) + cComp

#-- Bitmap character placement
def SWFFillStyleBitmap( iCharID, iDX, iDY ) :
    return chr(0x41) + struct.pack('h',iCharID) + SWFMatrix(1.0,1.0,-1,-1,iDX,iDY)
def SWFLineStyle( iPixelWidth, r,g,b ) :
    return struct.pack('h',iPixelWidth*20) + RGB(r,g,b)
def SWFBitmapShapeWithStyle( iCharID, iX0, iX1, iY0, iY1 ) :
    dY=iY1-iY0
    dX=iX1-iX0

    #----- SHAPEWITHSTYLE tag contents
    # Fill Styles
    c = chr(2) + SWFFillStyleBitmap( 0x7FFF, iX0, iY0 ) + SWFFillStyleBitmap( iCharID, iX0, iY0 )
    # line Styles
    c = c + chr(0)
    # Fill, line index bits = 2,0
    c = c + chr(0x20)
    # Style change - NEWSTYLE + EDGE + EDGE + EDGE + EDGE + END
    # type=0 newstyle=0 linestyle=0 fill1=1 fill0=0 move=1 + movesize[5] + dx[] + dy[] + Fill1=2
    iBits=math.ceil(math.log(20*max(iX1,iY0),2))+1
    cc = '000101' + IntToBin(iBits,5) +IntToBin(iX1*20,iBits) + IntToBin(iY0*20,iBits) +'10'
    # y=y+dy : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 1(=Ver) DY
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'01'+IntToBin(20*dY,iBits)
    # x=x-dx : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 0(=Hor) DX
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'00'+IntToBin(-20*dX,iBits)
    # y=y-dy 
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'01'+IntToBin(-20*dY,iBits)
    # x=x+dx : type=1(edge) straight=1 bits[4](add+2) general=0(V or H) 0(=Hor) DX
    iBits=math.ceil(math.log(20*dY,2))+1
    cc = cc + '11'+IntToBin(iBits-2,4)+'00'+IntToBin(20*dX,iBits)
    # End Shape Record
    cc = cc + '000000'
    return c + BinToSWFBit( cc )

def SWFBitmapPlace( iShapeID, iCharID, iX0, iX1, iY0, iY1 ) :
    cc = SWFBitmapShapeWithStyle( iCharID, iX0, iX1, iY0, iY1 )
    cRct = SWFRectangle(0, int(20*iX0), int(20*iX1), int(20*iY0), int(20*iY1) )
    ilen = len(cc)+len(cRct)+2
    return chr(0xBF)+chr(0) +struct.pack('l',ilen) +struct.pack('h',iShapeID) + cRct + cc

#-- Place a previously defined shape object 
def SWFPlaceObject2( iDepth, iCharID ) :
    # change flags 6->2 and len 6->5 and remove matrix at end.
    return chr(0x86)+chr(6) +chr(6) +struct.pack('h',iDepth) +struct.pack('h',iCharID)+chr(0)

#-- Remove object at specified depth 
def SWFRemoveObject2( iDepth ) :
    return chr(2)+chr(7)+struct.pack('h',iDepth)

#-- Set background tag (id=9)
def SWFSetBGR( r,g,b ) : 
    return chr(0x43)+chr(2)+chr(r)+chr(g)+chr(b)

#-- Construct file header
def SWFFile( cRct, iFrameRate, iFrameCount , cData ) :
    return 'FWS'+chr(5)+struct.pack('l',12+len(cRct)+len(cData)) +cRct +chr(0)+chr(iFrameRate) + struct.pack('h',iFrameCount) + cData

#---------------------------------------------------------------------------------
def BMPLoad(cFile, iID) :
    f = file(cFile,'rb')
    f.seek(10)
    iPos=struct.unpack('l',f.read(4))[0]
    f.seek(18)
    iWid=struct.unpack('l',f.read(4))[0]
    iHgt=struct.unpack('l',f.read(4))[0]
    f.seek(28)
    iBpp=struct.unpack('h',f.read(2))[0]
    iHgt=min(800,iHgt)
    cImage=''

    f.seek(iPos)
    for iRow in range(iHgt) :
        cc=''
        for iCol in range(iWid) :
            if iBpp==32 :
                cRGB=f.read(4)
            else :
                cRGB=f.read(3)
            if iCol < 800 :
                cc=cc+chr(255)+cRGB[2]+cRGB[1]+cRGB[0]
        cImage=cc+cImage
    iWid = min(800,iWid)
    f.close()
    print 'Loaded',cFile,iHgt,iWid
    return (iHgt,iWid,SWFDefineBits2a(iID, iWid, iHgt, cImage ))

def SWFTest() :
    cImage1 = BMPLoad( 'C:\\temp\\swf\\ACHRSERIES1.bmp', 1)
    cImage2 = BMPLoad( 'C:\\temp\\swf\\ACHRSERIES2.bmp', 2)
    cImage3 = BMPLoad( 'C:\\temp\\swf\\ACHRSERIES3.bmp', 3)
    cImage4 = BMPLoad( 'C:\\temp\\swf\\ACHRSERIES4.bmp', 4)
    cImage5 = BMPLoad( 'C:\\temp\\swf\\ACHRSERIES5.bmp', 5)
    cImage6 = BMPLoad( 'C:\\temp\\swf\\ACHRSERIES6.bmp', 6)
    cImage7 = BMPLoad( 'C:\\temp\\swf\\ACHRSERIES7.bmp', 7)

    iMaxHgt = max(cImage1[0],cImage2[0],cImage3[0],cImage4[0],cImage5[0],cImage6[0],cImage7[0])
    iMaxWid = max(cImage1[1],cImage2[1],cImage3[1],cImage4[1],cImage5[1],cImage6[1],cImage7[1])

    cRct = SWFRectangle(0,0,20*iMaxWid,0,20*iMaxHgt)      # bits,x0,x1,y0,y1

    c = SWFSetBGR(64,64,64)
    c = c + cImage1[2] + cImage2[2] + cImage3[2] + cImage4[2]
    c = c + cImage5[2] + cImage6[2] + cImage7[2]
    c = c + SWFBitmapPlace(11, 1, 1.0, cImage1[1], 1.0, cImage1[0] )
    c = c + SWFBitmapPlace(12, 2, 1.0, cImage2[1], 1.0, cImage2[0] )
    c = c + SWFBitmapPlace(13, 3, 1.0, cImage3[1], 1.0, cImage3[0] )
    c = c + SWFBitmapPlace(14, 4, 1.0, cImage4[1], 1.0, cImage4[0] )
    c = c + SWFBitmapPlace(15, 5, 1.0, cImage5[1], 1.0, cImage5[0] )
    c = c + SWFBitmapPlace(16, 6, 1.0, cImage6[1], 1.0, cImage6[0] )
    c = c + SWFBitmapPlace(17, 7, 1.0, cImage7[1], 1.0, cImage7[0] )

    c = c + SWFPlaceObject2(1,11) + SWFShowFrame(10) + SWFRemoveObject2(1)
    c = c + SWFPlaceObject2(2,12) + SWFShowFrame(10) + SWFRemoveObject2(2)
    c = c + SWFPlaceObject2(1,13) + SWFShowFrame(10) + SWFRemoveObject2(1)
    c = c + SWFPlaceObject2(1,14) + SWFShowFrame(10) + SWFRemoveObject2(1)
    c = c + SWFPlaceObject2(1,15) + SWFShowFrame(10) + SWFRemoveObject2(1)
    c = c + SWFPlaceObject2(3,16) + SWFShowFrame(10) + SWFRemoveObject2(3)
    c = c + SWFPlaceObject2(1,17) + SWFShowFrame(10) + SWFRemoveObject2(1)
    c = c + SWFShowFrame(10)
    iFrames=7

    c = c + SWFEndFile()

    f = open('C:\\Temp\\SWF\\Test.swf', 'wb')
    f.write(SWFFile(cRct,12,iFrames,c))
    f.close()
