"""Implements menuitem Send to -> QtiPlot.  This script is executed inside QtiPlot.
Do not edit this file -- it may be replaced by QuB Upgrade.  Instead, make a copy,
and it will be a separate option in the Send To menu.
"""

import numpy
from math import *

def add_hist(title, xlabel, ylabel, bins, bars, colors, *lines):
    """Adds a Table and Graph (histogram) with optional trend line(s).

    @param title: name for the table and graph
    @param xlabel: label for the x axis
    @param ylabel: label for the y axis
    @param bins: array of x coordinates
    @param bars: array of y coordinates
    @param colors: array of color indices (each trend line)
    @param lines: optional: one or more arrays of y coordinates
"""
    if not colors:
        colors = [0] + range(len(lines))
    if len(colors) <= len(lines):
        colors = list(colors)
        colors[len(colors):] = [0]*(len(lines) - len(colors) + 1)
    t = newTable(title, len(bins), 2+len(lines))
    t.setColData(1, bins, 0)
    t.setColData(2, bars, 0)
    g = newGraph()
    l = g.activeLayer()
    l.addCurves(t, (2,), Layer.VerticalBars)
    l.curve(0).setBrush(QtGui.QBrush(Qt.NoBrush))
    l.setCurveLineColor(0, colors[0]);
    for i, line in enumerate(lines):
        t.setColData(i+3, line, 0)
        l.addCurves(t, (i+3,), Layer.Line, 2.0, 0)
        l.setCurveLineColor(i+1, colors[i+1])
    l.setCurveLineWidth(1, 3.0)
    l.setAntialiasing()
    l.setTitle(title)
    l.enableAxis(Layer.Bottom, True)
    l.setAxisTitle(Layer.Bottom, xlabel)
    l.enableAxis(Layer.Left, True)
    l.setAxisTitle(Layer.Left, ylabel)
