"""Implements menuitem Send to -> QtiPlot.  This script is executed inside QtiPlot.
Do not edit this file -- it may be replaced by QuB Upgrade.  Instead, make a copy,
and it will be a separate option in the Send To menu.
"""

import numpy
from numpy import *

def add_scatter(title, xlabel, ylabel, xser, yser, colors, line_colors, line_x, *lines):
    """Adds a Table and Graph (scatter plot) with multi-color data points and optional trend line(s).

    @param title: name for the table and graph
    @param xlabel: label for the x axis
    @param ylabel: label for the y axis
    @param xser: array of x coordinates (each point)
    @param yser: array of y coordinates
    @param colors: array of color indices (each point)
    @param line_colors: array of color indices (each trend line)
    @param line_x: array of x-coords for lines
    @param lines: optional: one or more arrays of y coordinates
    @returns: (table, graph, layer)
"""
    c_lo, c_hi = int(min(colors)), int(max(colors))
    Nc = c_hi - c_lo + 1
    t = newTable(title, len(yser), 4+Nc+len(lines))
    for i in xrange(len(xser)):
        t.setCell(1, i+1, xser[i])
        t.setCell(2, i+1, yser[i])
        t.setCell(3, i+1, colors[i])
        t.setCell(4+int(colors[i])-c_lo, i+1, yser[i])
    g = newGraph()
    l = g.activeLayer()
    for c in xrange(Nc):
        l.addCurves(t, (c+4,), Layer.Scatter, 0.0, 8.0)
        l.setCurveLineColor(c, c-c_lo)
    if not line_colors:
        line_colors = [0] + range(len(lines)-1)
    if len(line_colors) <= len(lines):
        line_colors = list(line_colors)
        line_colors[len(line_colors):] = [0]*(len(lines) - len(line_colors) + 1)
    if line_x != None:
        t.setColData(4+Nc, line_x, 0)
    for i, line in enumerate(lines):
        t.setColData(5+Nc+i, line, 0)
        c = l.insertCurve(t, str(4+Nc), str(5+Nc+i), Layer.Line)
        c.symbol().setStyle(PlotSymbol.NoSymbol)
        l.setCurveLineWidth(4+Nc+i, 1.0)
        l.setCurveLineColor(4+Nc+i, int(line_colors[i]))
    l.setAntialiasing()
    l.setTitle(title)
    l.enableAxis(Layer.Bottom, True)
    l.setAxisTitle(Layer.Bottom, xlabel)
    l.enableAxis(Layer.Left, True)
    l.setAxisTitle(Layer.Left, ylabel)
    return (t,g,l)

def add_hist_to_scatter(table, graph, layer, bins, bars):
    iCol = table.numCols() + 1
    table.addColumn()
    table.addColumn()
    if len(bins) > table.numRows():
        table.setNumRows(len(bins))
    for i in xrange(len(bins)):
        table.setCell(iCol, i+1, bins[i])
        table.setCell(iCol+1, i+1, bars[i])
    curve = layer.insertCurve(table, str(iCol), str(iCol+1), Layer.VerticalBars)
    curve.setBrush(QtGui.QBrush(Qt.NoBrush))

def add_lbl_to_scatter(table, graph, layer, lbl):
    legend = layer.newLegend(lbl)
