/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifndef SKM_EVAL_TREE_H
#define SKM_EVAL_TREE_H

using namespace std;

class skm_model{
	public:
		skm_model( QUB_Tree modelNode, int *fFixAmp=0, int *fFixSD=0 );
		
		QUB_Tree        node;
		ModelTree       tree;
		mdlinf          mi;

		int             nclass, nchannel;
		fq::vector<int>   path, fixAmp, fixSD;
		fq::vector<float> x, q, amp, xms;
	};

class skm_metamodel{
	public:
		skm_metamodel( skm_model& mdl, int nchan, double dtMS );
		~skm_metamodel();

		skm_model& model;
		int nchannel;

		int             nmutistate, nmuticlass, nmutipath;
		int             *mutistate, *mutipath; // allocated in skm_mutimakv() [skm-utils.cpp]
		fq::vector<int>   cmutistate, mutiFixAmp, mutiFixSD;
		fq::vector<float> mutiq, mutia, mutipr, mutiamp, mutixms;
	};

QUB_Tree MakeIdealNode( int *byPoint, int ndata, float dtMS );

class skm_workspace {	
	// can idealize segments for one file (dataset with same conditions)
	public:
		skm_workspace( QUB_Tree dataset, skm_model& mdl, double *aMatrixOverride,
						QTR_NodeFptr fillDatCB, std::ostream& errstream, int *StopFlag );
		~skm_workspace(); 

		int segCount();
		int classCount();

		int runSKM( bool reestimate, int maxIter, double llConv, bool fixQ, int seg=-1 ); // default: all segs

		QUB_Tree         dataSet; // filled originals
		QUB_Tree         dataOut; // results and idealization
		vector<QUB_Tree> segments;
		vector<QUB_Tree> segOuts;

	protected:
		void setSegIdealization( int seg_ix, int ndata ); // changes stateSeq to classSeq
		void setSegStat( int seg_ix );
		//void setSegHist( int seg_ix, double *data, int ndata );
		void setErrIdlAndStat( int seg_ix );

		void setupData();
		void setupArrays();

		int              maxndata;
		double           dtMS;

		skm_metamodel    model;

		QTR_NodeFptr     fillDataCB;
		std::ostream&    milerr;
		int*             Stopped;

		double *         initialAMatrix; // ignore if null, otherwise use it instead of model rates

		fq::vector<float>  mutia, mutipr, mutiamp, mutixms; // being reestimated
		fq::vector<int>    stateSeq;
	};

typedef CountedPtr<skm_workspace> skm_worksptr;

class skm_eval_tree{
	public:
		skm_eval_tree( vector<QUB_Tree>& dataSets, QUB_Tree config,
						QTR_NodeFptr fillDatCB, QTR_NodeFptr rptCB, QTR_NodeFptr percentCB );

		QUB_Tree execute();

	protected:
		void initResults();
		void finishResults();

		bool canRun();

		skm_model model;
		vector<skm_worksptr> workspaces;
		QUB_Tree cfg;
		QTR_NodeFptr fillDataCB, reportCB, pctCB;

		mil_reportstream milerr;

		bool reestimate, fixQ;
		int maxIter;
		double llConv;

		int *Stopped;

		QUB_Tree result;
	};

QUB_Tree skm_unlimited( vector<QUB_Tree>& dataSets, QUB_Tree config,
					    QTR_NodeFptr fillDataCB, QTR_NodeFptr reportCB, QTR_NodeFptr pctCB, bool doStat = false );

extern "C" MIL_API QTR_Impl * __stdcall
skmtreeiface( QTR_Impl *cfg, QTR_Impl **data,
			  QTR_INodeCFunc *fillDataCB, QTR_INodeCFunc *rptCB, QTR_INodeCFunc *pctCB );

extern "C" PyObject * skmtreePyiface( PyObject *self, PyObject *args );
// it will work with a bogus fillDataCB as long as the data is already there

extern "C" MIL_API QTR_Impl * __stdcall skm_get_multiclasses_iface(QTR_Impl *model, int channelCount);


#endif
