{ **********************************************************************
  *                         Unit PLOTVAR.PAS                           *
  *                            Version 1.0                             *
  *                    (c) J. Debord, August 2001                      *
  **********************************************************************
                 Graphic constants, types and variables
                    (used by units PLOT and TEXPLOT)
  ********************************************************************** }

unit plotvar;

interface

uses
  fmath;

const
  MAXSYMBOL = 9;        { Max. number of graphic symbols }
  EPS       = 1.0E-10;  { Lower limit for an axis label }

type
  TScale  = (LIN_SCALE,        { Scale }
             LOG_SCALE);

  TGrid   = (NO_GRID,          { Grid }
             HORIZ_GRID,
             VERTIC_GRID,
             BOTH_GRID);

  TAxis   = record             { Coordinate axis }
              Scale : TScale;
              Min   : Float;
              Max   : Float;
              Step  : Float;
            end;

  TTitle  = record             { Title for main graph or axis }
              Text       : String[70];
              Font       : Integer;
              CharWidth  : Integer;
              CharHeight : Integer;
            end;

  TLegend = record             { Legends of plotted curves }
              Text       : array[1..MAXSYMBOL] of String[40];
              Font       : Integer;
              CharWidth  : Integer;
              CharHeight : Integer;
              SymbolSize : Integer;
            end;

var
  Xwin1, Ywin1, Xwin2, Ywin2 : Integer; { Window limits in % }

  GraphBorder : Boolean;  { Plot graph border }
  XAxis       : TAxis;    { Horizontal axis }
  YAxis       : TAxis;    { Vertical axis }
  Grid        : TGrid;    { Grid }
  GraphTitle  : TTitle;   { Title of graph }
  XTitle      : TTitle;   { Title of X axis }
  YTitle      : TTitle;   { Title of Y axis }
  Legend      : TLegend;  { Legends of curves }

implementation

var
  I : Integer;

begin
  Xwin1 := 15;
  Ywin1 := 15;
  Xwin2 := 85;
  Ywin2 := 85;

  GraphBorder := True;

  with XAxis do
    begin
      Scale := LIN_SCALE;
      Min   := 0.0;
      Max   := 1.0;
      Step  := 0.2;
    end;

  with YAxis do
    begin
      Scale := LIN_SCALE;
      Min   := 0.0;
      Max   := 1.0;
      Step  := 0.2;
    end;

  Grid := NO_GRID;

  with GraphTitle do
    begin
      Text       := '';
      Font       := 2;
      CharWidth  := 300;
      CharHeight := 350;
    end;

  with XTitle do
    begin
      Text       := 'X';
      Font       := 2;
      CharWidth  := 200;
      CharHeight := 250;
    end;

  with YTitle do
    begin
      Text       := 'Y';
      Font       := 2;
      CharWidth  := 200;
      CharHeight := 250;
    end;

  with Legend do
    begin
      Font       := 2;
      CharWidth  := 50;
      CharHeight := 50;
      SymbolSize := 3;

      for I := 1 to MAXSYMBOL do
        Text[I] := Chr(32 + I);
    end;

end.
