{

                                 Qub Data File
                                     unit



                        Interface unit for QUB Data files


(C) 2000 Lorin Milescu
}
{
/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}


//01-31-2000   Creation date

unit QUB_QFS_QDF;


//******************************************************************************
interface


uses


Windows, SysUtils, Classes, Dialogs, Forms, Graphics, SyncObjs, Clipbrd, Controls,

AllocTypes, Alloc,

QUB_Global,

QUB_Console, QUB_Rich_Console,

QUB_BusyDialog,
QUB_QFS, QUB_QFS_File, QUB_Node,
QUB_QFS_QDF_Global,
QUB_QFS_Edit,
QUB_QFS_Segments, QUB_QFS_DWT_Segments,
QUB_QFS_QDF_MPX, QUB_QVO_QVD_Global,
QUB_Node_HEKA,
QUB_Idealization,
QUB_QFS_QDF_Properties,
QUB_Parser;


type


//Forward class declarations ***************************************************
TQUB_QFS_QDF                                   = class;


//Class definitions

//QDF
TQUB_QFS_QDF                                   = class(TQUB_QFS_File)
protected


  prQVDParent                                  :TObject;

  m_nMultiplier                                :Double;
  // For floating point loads, calculate maximum power
  // of 10 to multiply data and still maintain < longint
  // precision.  Then suggest inverse as scaling value.

  m_lMacData                                   :Boolean;
  // Set to true for data which requires byte-swap

  prFileType                                   :TQUB_QFS_QDF_FileType;


  prSession                                    :INode;


  //Streamed properties ********************************************************
  prSegments                                   :TQUB_QFS_Segments;

  prRecSampling                                :TQUB_QFS_Record;
  prRecScaling                                 :TQUB_QFS_Record;

  prRecADChannelCount                          :TQUB_QFS_Record;
  prRecADDataSize                              :TQUB_QFS_Record;
  prRecADDataType                              :TQUB_QFS_Record;
  //****************************************************************************


  prIdealization                               :Array[0..QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1] of TQUB_IdlSegments;
  //****************************************************************************


  prLDTLeader                                  :Integer;
  prLDTSampling                                :Word;
  prLDTScaling                                 :Word;

  prWAVBitsPerSample                           :Integer;

  //CriticalSections
  prSegInfoCritSection                         :TCriticalSection;
  prReadWriteCritSection                       :TCriticalSection;

  //InfoEvents
  prOnInfoEdit                                 :TQUB_EditEvent;

  prOpenTimeColumn                             :Boolean;
  prOpenUnevenSampling                         :Boolean;
  prOpenSinglePrec                             :Boolean;
  prOpenSegmented                              :Boolean;
  prOpenMultiplexed                            :Boolean;
  prOpenMPXAA                                  :Boolean;

  //Add Series/Sweep events - used with HEKA files. LM
  prOnAddSeries                                :TQUB_QFS_QDF_OnAddSeries;
  prOnAddSweep                                 :TQUB_QFS_QDF_OnAddSweep;


  // used during Save:
  prSaveStream :TFileStream;
  prSaveType   :TQUB_QFS_QDF_FileType;
  prSaveName   :String;
  prSaveEntry  :Integer;
  prSaveUpdate :TQUB_BusyUpdate;
  prSaveResult :Boolean;
  

  // update and save the session unless the user has explicitly said "No" don't save
  prDiscardChanges                             :Boolean;


  //Streamed properties ********************************************************
  function prGetSampling :Double;
  procedure prSetSampling(s :Double);
  procedure prSetScaling(s :Double);
  function prGetScaling :Double;
  procedure prSetADChannelCount(c :Integer);
  function prGetADChannelCount :Integer;
  procedure prSetADDataSize(s :Integer);
  function prGetADDataSize :Integer;
  procedure prSetADDataType(t :TQUB_QFS_Node_DataType);
  function prGetADDataType :TQUB_QFS_Node_DataType;
  //****************************************************************************


  function prGetIdealization(Index :Integer) :TQUB_IdlSegments;


  //Low level Read / Write *****************************************************
  function prReadQDF :Boolean;
  function prReadDAT :Boolean;
  function prReadLDT :Boolean;
  function prReadASC :Boolean;
  function prReadABF :Boolean;
  function prReadWAV :Boolean;
  function prReadPUL :Boolean;
  function prReadPUL9:Boolean;
  procedure prAddPulse9Sweep(Sweep :THEKA_Tree_Pulse9_Sweep; Scale :TPADouble);
  function prIsPUL9Dat(const FName :String): Boolean;
  function prReadACQ :Boolean;
  function prReadDWT :Boolean;
  function prReadEVL :Boolean;
  function prReadPRT :Boolean;
  function prReadTACidl :Boolean;
  function prReadQSF :Boolean;
  function prReadIBW :Boolean;
  function prReadEvents(DWT :INode) :Boolean;
  
  //-- prReadSegmentOf... Various methods for reading arrays of Int2, Int4,
  // Single, Double data from other (possibly Mac byte ordered ) data files to a
  // Qub 2 or 4 byte scaled integer data segment.
  // Each of these return a TQUB_QFS_Segment object, or nil on failure.
  // SampleCount : Samples per channel
  // iReadSize : 4 for single, 8 for double
  // aFactor :  array of scaling factors, or nil to use single Scaling member variable for all data
  // bSwapBytes : Reverse the 2, 4, or 8 bytes of data prior to conversion.  ( converts Macintosh data )
  //
  // The '2Chan' versions are For Heka Files pulse files, which have channels
  // stored one after another, with a possible intervening leak trace (ignored).
  // if not SecondTrace: fill channel2 with zeros
  //
  // Some of these also set the Scaling to Scaling/2.  Not sure why, and it is
  // course not commented, but my guess is the data coming in can use unsigned
  // integers and qub uses signed integers so use half the range....
  //
  function prReadSegmentAddResult( iDataType : TQUB_QFS_Node_DataType; iChannels, iDataSize, iSamples, iEditPos : Integer ): TQUB_QFS_Segment;
  function prReadSegmentOfReals( SampleCount :Integer; iReadSize:Integer; aFactor:TPASingle ): TQUB_QFS_Segment;
  function prReadSegmentOfInts(SampleCount, iReadSize :Integer; bSwapBytes:Boolean): TQUB_QFS_Segment; // or nil on failure
  function prReadSegmentOfReals2Chan(SampleCount, iReadSize :Integer; LeakTrace, SecondTrace :Boolean): TQUB_QFS_Segment;
  function prReadSegmentOfInts2Chan(SampleCount, iReadSize :Integer; bSwapBytes, LeakTrace, SecondTrace :Boolean): TQUB_QFS_Segment; // or nil on failure

  //LM. I added the leak trace, and generalized to 2 or 3 channels. n = [2, 3], but Second trace has priority
  function prReadSegment2or3Chan(SampleCount :Integer; DataType :TQUB_QFS_Node_DataType; LittleEndian :Boolean; LeakTrace, SecondTrace :Boolean): TQUB_QFS_Segment; // or nil on failure

  function prConvertASCII(Stream :TStream; ChannelCount :Integer; TimeColumn, UnevenSampling, Segmented :Boolean; var DataCount :Integer; Segment :Integer; Entry :Integer) :Boolean;
  function prConvertASCIIMultiplexed(Stream :TStream; ChannelCount :Integer; TimeColumn :Boolean; var DataCount :Integer; Segment :Integer; MPXAA :Boolean; Entry :Integer) :Boolean;
  function prConvertSingle(Stream :TStream; ChannelCount :Integer; TimeColumn, UnevenSampling :Boolean; var DataCount :Integer; Segment :Integer; Entry :Integer) :Boolean;
  function prConvertInteger(Stream :TStream; ChannelCount :Integer; TimeColumn, UnevenSampling :Boolean; var DataCount :Integer; Segment :Integer; Entry :Integer) :Boolean;

  function prWriteDATHeader(Stream :TStream) :Boolean;
  function prWriteLDTHeader(Stream :TStream) :Boolean;
  function prWriteASCHeader(Stream :TStream) :Boolean;
  function prWriteABFHeader(Stream :TStream) :Boolean;
  function prWriteWAVHeader(Stream :TStream) :Boolean;

  function prWriteDATSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteLDTSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteASCSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteABFSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteWAVSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;

  function prWriteDATSegment(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteLDTSegment(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteASCSegment(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteABFSegment(Segment :Integer; Stream :TStream) :Boolean;
  function prWriteWAVSegment(Segment :Integer; Stream :TStream) :Boolean;


  //Low level File *************************************************************
  function prNew :Boolean; override;
  function prOpen(const FileName :String) :Boolean; override;
  function prSave(const FileName :String) :Boolean; override;
  procedure prDoBusySave(OnUpdate :TQUB_BusyUpdate);

  procedure prReadSession(const FileName :String);
  procedure prCheckSession;
  procedure prOverlayQDFSession;
  procedure prUpdateSession;

  procedure prNewIdealization;

  function prGetFileTypeFromFileName(const FileName :String) :TQUB_QFS_QDF_FileType;
  function prGetFileExt :String;

  function prGetUniqueFileName(const FileName :String) :String;

  function prGetLDTSampling :Word;
  function prGetLDTScaling :Word;
  function prGetLDTLeader :Integer;

  function prGetWAVBitsPerSample :Integer;
  function prGetWAVBlockAlign :Integer;
  function prGetWAVAvgBytesPerSec :Integer;


  //Can's **********************************************************************
  function prGetCanCut :Boolean;
  function prGetCanCopy :Boolean;
  function prGetCanPaste :Boolean;
  function prGetCanDelete :Boolean;
  function prGetCanErase :Boolean;
  function prGetCanFill :Boolean;
  function prGetCanBreak :Boolean;
  function prGetCanUndo :Boolean;
  function prGetCanAppend :Boolean;
  //****************************************************************************

  //Edit methods
  function prCut(First, Last :Integer; DoOnePoint :Boolean; CopyASCII :Boolean) :Boolean;
  function prCopy(First, Last :Integer; DoOnePoint :Boolean; CopyASCII :Boolean) :Boolean;
  function prCopyBinary(First, Last :Integer; DoOnePoint :Boolean) :Boolean;
  function prCopyASCII(First, Last :Integer; DoOnePoint :Boolean) :Boolean;
  function prCopyList(List :TQUB_QVO_QVD_SelectionList; ListName :String; CopyASCII :Boolean) :Boolean;
  function prCopyListBinary(List :TQUB_QVO_QVD_SelectionList; ListName :String) :Boolean;
  function prCopyListASCII(List :TQUB_QVO_QVD_SelectionList; ListName :String) :Boolean;
  function prPaste(First, Last :Integer; PasteAsSeg :Boolean; ReplaceSel :Boolean) :Boolean;
  function prDelete(First, Last :Integer; DoOnePoint :Boolean; DelOperation :TQUB_QFS_EditOperation; AllowEmptySeg :Boolean) :Boolean;
  function prErase(First, Last :Integer; BufFirst, BufLast :Integer; DoOnePoint :Boolean) :Boolean;
  function prFill(First, Last :Integer; Channel :Integer; Value :Variant) :Boolean;
  function prBreak(First, Last :Integer) :Boolean;
  function prEditPoint(Channel :Integer; Point :Integer; Change :Integer) :Boolean;
  function prClear :Boolean;
  function prUndo :Boolean;
  function prAppendData(Buffer :Pointer; BufferSize, BufferCount :Integer; BufferIsWritten :Boolean; BufferAtPos :Integer; NewSegment :Boolean; StartTime :Double) :Boolean;
  function prAppendDataASCII(const Text :String; NewSegment, TimeColumn, UnevenSampling, Segmented :Boolean) :Boolean;
  function prReplaceData(Buffer :Pointer; BufferSize :Integer; Sel :TQUB_FirstLast) :Boolean;
  function prNewSegment(StartTime :Double; ForceNewIfEmpty :Boolean) :Boolean;


  procedure DoOnInfoOpen(Sender :TObject);
  procedure DoOnInfoClose(Sender :TObject);
  procedure DoOnInfoEdit(Sender :TObject; EditOper :TQUB_EditOperation; var Param; var Param2);

  procedure prOnEditIdealization(Sender :TObject);


  //procedure DoOnEdit(EditOper :TQUB_EditOperation; var Param; var Param2); virtual;
  procedure prInfoEdit(EditOper :TQUB_EditOperation; var Param; var Param2); virtual;


public



  constructor Create(const Name :String; Parent :TQUB_QFS_Record); override;
  destructor Destroy; override;


  //File
  function Open(const FileName :String) :Boolean; override;
  function Close :Boolean; override;

  function SaveSession: Boolean;

  function CanLoadInfo(ChannelIndex :Integer) :Boolean;
  function CanSaveInfo(ChannelIndex :Integer) :Boolean;

  function CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;

  //Edit
  function CutSelection(First, Last :Integer; CopyASCII :Boolean) :Boolean;
  function CopySelection(First, Last :Integer; CopyASCII :Boolean) :Boolean;
  function CopyList(List :TQUB_QVO_QVD_SelectionList; ListName :String; CopyASCII :Boolean) :Boolean;
  function Append :Boolean;
  function PasteSelection(First, Last :Integer) :Boolean;
  function PasteSegment(First, Last :Integer) :Boolean;
  function DeleteSelection(First, Last :Integer) :Boolean;
  function EraseSelection(First, Last :Integer; BufFirst, BufLast :Integer) :Boolean;
  function FillSelection(First, Last :Integer; Channel :Integer; Value :Variant) :Boolean;
  function BreakSegment(First, Last :Integer) :Boolean;
  function EditPoint(Channel :Integer; Point :Integer; Change :Integer) :Boolean;
  function Undo :Boolean;
  function RecordData(Buffer :Pointer; BufferSize, BufferCount :Integer; NewSegment :Boolean; StartTime :Double; Refresh :Boolean) :Boolean;
  function AppendData(Buffer :Pointer; BufferSize, BufferCount :Integer; NewSegment :Boolean; StartTime :Double; Refresh :Boolean) :Boolean;
  function AppendDataASCII(const Text :String; NewSegment, TimeColumn, UnevenSampling, Segmented :Boolean; Refresh :Boolean) :Boolean;
  function ReplaceData(Buffer :Pointer; BufferSize :Integer; Sel :TQUB_FirstLast; Refresh :Boolean) :Boolean;
  function NewSegment(StartTime :Double; Refresh :Boolean; ForceNewIfEmpty :Boolean) :Boolean;
  function Clear :Boolean;

  function Read(ReadStream :TStream) :Boolean; override;

  function NewChild(ReadStream :TStream) :TQUB_QFS_Record; override;

  //Read / Write Data
  function ReadData(Buffer :Pointer; Raw :Boolean; Channel :Integer; DataFirst, DataLast :Integer) :Boolean;
  procedure ListSegmentBreaks(iOffset :Integer; aBreaks :TStringList);

  procedure WhichPointForRealTime(Time :Double; var Point :Integer);

  procedure OnGetPropertiesNew(Sender :TObject); override;
  procedure OnPutPropertiesNew(Sender :TObject); override;

  procedure DoMsgOut(Msg :PChar);

  //Streamed properties ********************************************************
  property Segments                            :TQUB_QFS_Segments        read prSegments;
  property Scaling                             :Double                   read prGetScaling          write prSetScaling;
  property Sampling                            :Double                   read prGetSampling         write prSetSampling;
  property ADChannelCount                      :Integer                  read prGetADChannelCount   write prSetADChannelCount;
  property ADDataSize                          :Integer                  read prGetADDataSize       write prSetADDataSize;
  property ADDataType                          :TQUB_QFS_Node_DataType   read prGetADDataType       write prSetADDataType;
  //****************************************************************************


  property Idealization[Index :Integer]        :TQUB_IdlSegments         read prGetIdealization;


  //LDT compatibility
  property LDTSampling                         :Word                     read prGetLDTSampling;
  property LDTScaling                          :Word                     read prGetLDTScaling;
  property LDTLeader                           :Integer                  read prGetLDTLeader;

  //WAV compatibility
  property WAVBitsPerSample                    :Integer                  read prGetWAVBitsPerSample;
  property WAVBlockAlign                       :Integer                  read prGetWAVBlockAlign;
  property WAVAvgBytesPerSec                   :Integer                  read prGetWAVAvgBytesPerSec;

  property FileType                            :TQUB_QFS_QDF_FileType    read prFileType;
  property FileExt                             :String                   read prGetFileExt;


  //Can's ......................................................................
  property CanCut                              :Boolean                  read prGetCanCut;
  property CanCopy                             :Boolean                  read prGetCanCopy;
  property CanPaste                            :Boolean                  read prGetCanPaste;
  property CanDelete                           :Boolean                  read prGetCanDelete;
  property CanErase                            :Boolean                  read prGetCanErase;
  property CanFill                             :Boolean                  read prGetCanFill;
  property CanBreak                            :Boolean                  read prGetCanBreak;
  property CanUndo                             :Boolean                  read prGetCanUndo;
  property CanAppend                           :Boolean                  read prGetCanAppend;

  property OnInfoEdit                          :TQUB_EditEvent           read prOnInfoEdit          write prOnInfoEdit;

  property Session                             :INode                    read prSession;
  property DiscardChanges                      :Boolean                  read prDiscardChanges      write prDiscardChanges;

  property OpenTimeColumn                      :Boolean                  read prOpenTimeColumn      write prOpenTimeColumn;
  property OpenUnevenSampling                  :Boolean                  read prOpenUnevenSampling  write prOpenUnevenSampling;
  property OpenSinglePrec                      :Boolean                  read prOpenSinglePrec      write prOpenSinglePrec;
  property OpenSegmented                       :Boolean                  read prOpenSegmented       write prOpenSegmented;
  property OpenMultiplexed                     :Boolean                  read prOpenMultiplexed     write prOpenMultiplexed;
  property OpenMPXAA                           :Boolean                  read prOpenMPXAA           write prOpenMPXAA;

  property OnAddSeries                         :TQUB_QFS_QDF_OnAddSeries read prOnAddSeries         write prOnAddSeries;
  property OnAddSweep                          :TQUB_QFS_QDF_OnAddSweep  read prOnAddSweep          write prOnAddSweep;

  property QVDParent                           :TObject                  read prQVDParent           write prQVDParent;
  

end;
//..............................................................................


type
TAxoMsgOutProc = function(Msg :PChar; Data :Pointer): Integer; stdcall;
function Convert_ABF_To_QDF(InPath, OutPath :PChar; MsgOut :TAxoMsgOutProc; MsgOutData :Pointer): Integer; stdcall; external 'qubaxo.dll' name '_Convert_ABF_To_QDF@16';
function Convert_Acquire_To_QDF(InPath, OutPath :PChar; MsgOut :TAxoMsgOutProc; MsgOutData :Pointer): Integer; stdcall; external 'qubaxo.dll' name '_Convert_Acquire_To_QDF@16';
function QAX_IsABFFile(Path :PChar): Integer; stdcall; external 'qubaxo.dll' name '_QAX_IsABFFile@4';
function DoAxoMsgOut(Msg :PChar; Data :Pointer): Integer; stdcall;


//******************************************************************************
implementation


uses


Math, FileCtrl,
QUB_QFS_QDF_WAV, QUB_Viewport_Data_New_Properties,
QUB_QVO_QVR,
QUB_QVO_QVD,
QUB_Viewport_Results,
QUB_GUIInterface,
QUB_Profiles,
QUB_Main;


var


ClipboardFormatName :String = 'QUB_QFS_QDF_CLIP_FORMAT';
ClipboardFormat     :Word   = 0;


//****************************************************************************//
//                                                                            //
//                              TQUB_QFS_QDF                                  //
//                                                                            //
//****************************************************************************//


//****************************************************************************//
//                                                                            //
//              c o n s t r u c t o r   /   d e s t r u c t o r               //
//                                                                            //
//****************************************************************************//
constructor TQUB_QFS_QDF.Create(const Name :String; Parent :TQUB_QFS_Record);
begin

  inherited;

  prSegInfoCritSection    := TCriticalSection.Create;
  prReadWriteCritSection  := TCriticalSection.Create;



  //Create permanent streamed properties ***************************************
  //Segments
  prSegments              := TQUB_QFS_Segments.Create(QUB_QFS_QDF_SEGMENTS_NAME, Self);
  prSegments.Permanent    := True;
  //Sampling/Scaling
  prRecSampling           := TQUB_QFS_Record.CreateAsValue(QUB_QFS_QDF_SAMPLING_NAME,        Self, True{Perm}, 1 / (QUB_QVO_QVD_DEF_SAMPLING * 1e3), NODE_DATATYPE_DOUBLE,  1{Count});
  prRecScaling            := TQUB_QFS_Record.CreateAsValue(QUB_QFS_QDF_SCALING_NAME,         Self, True{Perm}, QUB_QVO_QVD_DEF_SCALING,              NODE_DATATYPE_DOUBLE,  1{Count});
  //ADChannels
  prRecADChannelCount     := TQUB_QFS_Record.CreateAsValue(QUB_QFS_QDF_ADCHANNELCOUNT_NAME,  Self, True{Perm}, 1,                                    NODE_DATATYPE_LONGINT, 1{Count});
  prRecADDataSize         := TQUB_QFS_Record.CreateAsValue(QUB_QFS_QDF_ADDATASIZE_NAME,      Self, True{Perm}, SizeOf(SmallInt),                     NODE_DATATYPE_LONGINT, 1{Count});
  prRecADDataType         := TQUB_QFS_Record.CreateAsValue(QUB_QFS_QDF_ADDATATYPE_NAME,      Self, True{Perm}, Integer(NODE_DATATYPE_SMALLINT),      NODE_DATATYPE_LONGINT, 1{Count});
  //****************************************************************************



  //Idealization - do it later


  //Other defaults here
  prLDTLeader             := 1000;

  prSession               := TNode.Create('SessionFile');
  prDiscardChanges        := False;

  prOpenSilently          := False;
  prSaveSilently          := False;
  prOpenTimeColumn        := False;
  prOpenUnevenSampling    := False;
  prOpenSinglePrec        := False;
  prOpenSegmented         := True;
  prOpenMultiplexed       := False;
  prOpenMPXAA             := False;

  m_lMacData              := False;
  m_nMultiplier           := 1.0;


end;

destructor TQUB_QFS_QDF.Destroy;
var
  i         :Integer;
begin
  try

    Close;

    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
      if prIdealization[i] <> nil then begin
        prIdealization[i].Free;
        prIdealization[i] := nil;
      end;
    end;

    prSegInfoCritSection.Free;
    prReadWriteCritSection.Free;

  finally

    inherited;


  end;
end;


//Info Events
procedure TQUB_QFS_QDF.DoOnInfoOpen(Sender :TObject);
var
  Sel :TQUB_FirstLast;
begin
  if Assigned(OnInfoEdit) then OnInfoEdit(Self, EDIT_OPER_UPDATE, Sel, Sel);
end;

procedure TQUB_QFS_QDF.DoOnInfoClose(Sender :TObject);
var
  Sel :TQUB_FirstLast;
begin
  if Assigned(OnInfoEdit) then OnInfoEdit(Self, EDIT_OPER_UPDATE, Sel, Sel);
end;

procedure TQUB_QFS_QDF.DoOnInfoEdit(Sender :TObject; EditOper :TQUB_EditOperation; var Param; var Param2);
begin
  if Assigned(OnInfoEdit) then OnInfoEdit(Self, EditOper, Param, Param2);
end;

procedure TQUB_QFS_QDF.prOnEditIdealization(Sender :TObject);
var
  Sel :TQUB_FirstLast;
begin
  if Assigned(OnInfoEdit) then OnInfoEdit(Self, EDIT_OPER_UPDATE, Sel, Sel);
end;

procedure TQUB_QFS_QDF.prInfoEdit(EditOper :TQUB_EditOperation; var Param; var Param2);
var
  UndoInfo, UndoChan     :INode;
  c                      :Integer;
begin

  case EditOper of

    EDIT_OPER_INSERT_SEG:
      begin
        UndoInfo  := TNode.CreateUsing( TQUB_QFS_PNodeMem(Param2) );
        UndoChan  := UndoInfo.Find('DataSet');
        for c := 0 to ADChannelCount - 1 do begin
          Idealization[c].AddSeg( TQUB_FirstLast(Param).First, TQUB_FirstLast(Param).Last );
          if not UndoChan.IsNull then Idealization[c].SetDwells( UndoChan );
          UndoChan := UndoChan.SiblingSameName;
        end;
      end;

    EDIT_OPER_INSERT_INTOSEG:
      begin
        UndoInfo  := TNode.CreateUsing( TQUB_QFS_PNodeMem(Param2) );
        UndoChan  := UndoInfo.Find('DataSet');
        for c := 0 to ADChannelCount - 1 do begin
          Idealization[c].InsertPoints( TQUB_FirstLast(Param).First, TQUB_FirstLast(Param).Last);
          if not UndoChan.IsNull then Idealization[c].SetDwells( UndoChan );
          UndoChan := UndoChan.SiblingSameName;
        end;
      end;

    EDIT_OPER_DEL_SEG:
      begin
        UndoInfo  := TNode.CreateUsing( TQUB_QFS_PNodeMem(Param2) );
        for c := 0 to ADChannelCount - 1 do begin
          if not UndoInfo.IsNull then UndoInfo.AppendChild(Idealization[c].GetDwells(TQUB_FirstLast(Param).First, TQUB_FirstLast(Param).Last, 1, 0) );
          Idealization[c].RemovePoints( TQUB_FirstLast(Param).First, TQUB_FirstLast(Param).Last );
        end;
      end;

    EDIT_OPER_DEL_SEL:
      begin
        UndoInfo  := TNode.CreateUsing( TQUB_QFS_PNodeMem(Param2) );
        for c := 0 to ADChannelCount - 1 do begin
          if not UndoInfo.IsNull then UndoInfo.AppendChild( Idealization[c].GetDwells(TQUB_FirstLast(Param).First, TQUB_FirstLast(Param).Last, 1, 0) );
          Idealization[c].RemovePoints( TQUB_FirstLast(Param).First, TQUB_FirstLast(Param).Last );
        end;
      end;

    EDIT_OPER_IBREAK:
      begin
        for c := 0 to ADChannelCount - 1 do begin
          Idealization[c].BreakSeg( TQUB_FirstLast(Param).First );
        end;
      end;

    EDIT_OPER_JOIN:
      begin
        for c := 0 to ADChannelCount - 1 do begin
          Idealization[c].JoinSegs( TQUB_FirstLast(Param).First );
        end;
     end;

  end;//EditOper

  if Assigned(OnEdit) then OnEdit(Self, EditOper, Param, Param2);

end;


//****************************************************************************//
//                                                                            //
//                            F I L E    M E T H O D S                        //
//                                                                            //
//****************************************************************************//
procedure TQUB_QFS_QDF.prNewIdealization;
var
  i, s, p          :Integer;
  DataBreakPoints  :Array of Integer;
begin
  for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
    try
      prIdealization[i].Free;
      prIdealization[i] := nil;
    except
    end;
  end;

  SetLength(DataBreakPoints, Segments.SegmentCount);
  p := 0;
  for s := 0 to Segments.SegmentCount - 1 do begin
    Inc(p, Segments[s].Channels.DataCount);
    DataBreakPoints[s] := p;
  end;

  for i := 0 to ADChannelCount - 1 do begin
    try
      prIdealization[i]             := TQUB_IdlSegments.Create(Sampling);
      prIdealization[i].OnEdit      := prOnEditIdealization;
      prIdealization[i].UseTempFile(1);
      if Segments.SegmentCount > 0 then begin
        prIdealization[i].AddSeg(0, DataBreakPoints[0] - 1);
      end;
      for s := 1 to Segments.SegmentCount - 1 do begin
        prIdealization[i].AddSeg(DataBreakPoints[s-1], DataBreakPoints[s] - 1);
      end;
    except
    end;
  end;
end;

function TQUB_QFS_QDF.prGetUniqueFileName(const FileName :String) :String;
var
  Ext :String;
  i   :Integer;
begin

  Ext := ExtractFileExt(FileName);
  Result := FileName;
  i := 0;
  while ((FileExists(Result)) or (i > 1000)) do begin
    Result := ChangeFileExt(FileName, '_' + IntToStr(i) + Ext);
    Inc(i);
  end;

end;


function TQUB_QFS_QDF.Open(const FileName :String) :Boolean;
begin
  Result := inherited Open(FileName);

  //Session open
  if Result then begin
    prSession := TNode.CreateNull;
  end else begin
    prCheckSession;
    prOverlayQDFSession;
  end;
end;

function TQUB_QFS_QDF.Close :Boolean;
var
  i :Integer;
begin
  //Session save
  if (FileName <> '') and not DiscardChanges then begin
    prUpdateSession;

    Session.SaveAs(ChangeFileExt(FileName, '.qsf'));
    Session.Close;
  end;
  
  prSession := TNode.CreateNull;

  Result := inherited Close;
  if not Result then begin
    for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
      if prIdealization[i] <> nil then prIdealization[i].Clear;
    end;
  end;
end;

function TQUB_QFS_QDF.SaveSession: Boolean;
begin
  if FileName <> '' then begin
    prUpdateSession;
    Result := not Session.SaveAs( ChangeFileExt( FileName, '.qsf' ) );
    Session.Close;
  end else begin
    Result := True;
  end;
end;

procedure TQUB_QFS_QDF.prReadSession(const FileName :String);
var
  DataProperties :INode;

  function ReadSessPropOrDef(const Name :String; Default :Variant): Variant;
  var
    Node         :INode;
  begin
    Node         := prSession.Find(Name);
    if Node.DataCount > 0 then begin
      Result     := Node.Data[0];
      Exit;
    end;
    Node         := DataProperties.Find(Name);
    if Node.DataCount > 0 then begin
      Result     := Node.Data[0];
      Exit;
    end;
    Result       := Default;
  end;
begin
  if not prOpenedFromStream then begin
  //FromFile

    prSession := TNode.OpenReadOnly(ChangeFileExt(FileName, '.qsf'));
    prSession.Close;
    if prSession.IsNull then prSession := TNode.Create('SessionFile');

  //end else begin
  //FromStream - no session is/should be available

  end;

  DataProperties        := ProfileMgr['Data'].Active;

  if prSession.Find('Sampling').DataCount > 0 then begin
    Sampling            := prSession['Sampling'].Data[0];
  end else if DataProperties.Find('Sampling').DataCount > 0 then begin
    Sampling            := 1.0e-3 / DataProperties['Sampling'].Data[0];
  end;

  Scaling               := ReadSessPropOrDef('Scaling',         Scaling);
  ADChannelCount        := ReadSessPropOrDef('ADChannelCount',  ADChannelCount);
  ADDataSize            := ReadSessPropOrDef('ADDataSize',      ADDataSize);
  prOpenTimeColumn      := ReadSessPropOrDef('TimeColumn',      prOpenTimeColumn);
  prOpenUnevenSampling  := ReadSessPropOrDef('UnevenSampling',  prOpenUnevenSampling);
  prOpenSinglePrec      := ReadSessPropOrDef('SinglePrecision', prOpenSinglePrec);
  prOpenSegmented       := ReadSessPropOrDef('Segmented',       prOpenSegmented);
  prOpenMultiplexed     := ReadSessPropOrDef('Multiplexed',     prOpenMultiplexed);
  prOpenMPXAA           := ReadSessPropOrDef('MPXAA',           prOpenMPXAA);
end;


//TO DO: John, this function should be carefully checked, because there are many instances
//when it fails without a good reason, and all session info is lost, possibly for good.
//It should rather adapt to changes in the file structure. Very important. LM.
procedure TQUB_QFS_QDF.prCheckSession;
begin
end;
(*
var
  SegSizesNode  :INode;
  SegSizes      :TPAInteger;
  i             :Integer;
begin
  SegSizesNode  := Session.Find('SegmentSizes');
  if SegSizesNode.IsNull then Exit; // it's ok 'cause it's new

  if LongInt(SegSizesNode.DataCount) <> Segments.SegmentCount then begin
    prSession   := TNode.Create('SessionFile');
    Exit;
  end;

  SegSizes      := TPAInteger( SegSizesNode.DataPtr );
  for i := 0 to Segments.SegmentCount - 1 do begin
    if (SegSizes = nil) or (Segments[i].Channels.DataCount <> SegSizes[i]) then begin
      prSession := TNode.Create('SessionFile');
      Exit;
    end;
  end;
end;
*)
procedure TQUB_QFS_QDF.prOverlayQDFSession;
var
  SessionQFS            :TQUB_QFS_Record;
  SessionQDF            :INode;
  Child                 :INode;
begin
  // if have child 'SessionFile'
  // foreach toplevel child of 'SessionFile'
  //   add to prSession unless it's already there
  SessionQFS            := RecordsByName['SessionFile'];
  if SessionQFS = nil then Exit;
  SessionQDF            := SessionQFS.CloneAsNode(True, TNode.CreateNull);
  Child                 := SessionQDF.Child;
  while not Child.IsNull do begin
    if prSession.Find(Child.Name).IsNull then prSession.AppendClone(Child, True);
    Child               := Child.Sibling;
  end;
end;

procedure TQUB_QFS_QDF.prUpdateSession;
var
  SegSizesNode  :INode;
  SegSizes      :TPAInteger;
  i             :Integer;
begin
  if Session.IsNull then Exit;

  Session['Sampling'].SetData(Sampling);
  Session['Scaling'].SetData(Scaling);
  Session['ADChannelCount'].SetData(ADChannelCount);
  Session['ADDataSize'].SetData(ADDataSize);
  Session['TimeColumn'].SetData(Integer(prOpenTimeColumn));
  Session['UnevenSampling'].SetData(Integer(prOpenUnevenSampling));
  Session['SinglePrecision'].SetData(Integer(prOpenSinglePrec));
  Session['Segmented'].SetData(Integer(prOpenSegmented));
  Session['Multiplexed'].SetData(Integer(prOpenMultiplexed));
  Session['MPXAA'].SetData(Integer(prOpenMPXAA));

  SegSizesNode  := Session['SegmentSizes'];
  if FileExt = '.dat' then begin // no segmentation
    SegSizesNode.SetData(Segments.TotalDataCount);
  end else begin
    SegSizesNode.SetupNumData(NODE_DATATYPE_LONGINT, Segments.SegmentCount, 1);
    SegSizes      := TPAInteger(SegSizesNode.DataPtr);
    for i := 0 to Segments.SegmentCount - 1 do begin
      SegSizes[i] := Segments[i].Channels.DataCount;
    end;
  end;
end;


//****************************************************************************//
//                                                                            //
//                                    N e w                                   //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prNew :Boolean;
var
  Segment :TQUB_QFS_Segment;
begin

  Result := False;

  try

    prEditCritSection.Acquire;


    //Data
    //Sampling, Scaling, ADChannelCount, ADDataSize, ADDataType are set by the new file dialog


    //Segments
    prSegments.Stream                := Stream;
    prSegments.StreamPos             := Stream.Position;

    prWAVBitsPerSample               := ADDataSize * 8;

    //Segment
    Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
    Segment.Stream                   := Stream;
    Segment.StreamPos                := Stream.Position;

    //Channels
    Segment.Channels.CanEdit         := CanEdit;
    Segment.Channels.EditStream      := EditStream;
    Segment.Channels.Stream          := Stream;
    Segment.Channels.StreamPos       := Stream.Position;
    Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
    Segment.Channels.DataType        := ADDataType;
    Segment.Channels.DataCount       := 0;
    Segment.Channels.DataStreamPos   := Stream.Position;

    //Channels is empty and thus has no selection

    prSegments.AddRecord(Segment, prSegments.SegmentCount);

    prNewIdealization;

    prSession                        := TNode.Create('SessionFile');

  finally

    prEditCritSection.Release;

  end;

end;

//****************************************************************************//
//                                                                            //
//                                   O p e n                                  //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prOpen(const FileName :String) :Boolean;
var
  Ext        :String;
  FName      :PChar;
begin

  Result     := False;

  prReadSession(FileName);


  Ext        := LowerCase(ExtractFileExt(FileName));

  if          Ext = '.qdf' then begin
  //QDF
    prFileType := QFS_QDF_TYPE_QDF;

  end else if Ext = '.dat' then begin
  //DAT
    FName := PChar(FileName);        
    try
      Result := Boolean(QAX_IsABFFile(FName));
    except
      Result := False;
    end;
    if Result then begin
      prFileType := QFS_QDF_TYPE_ABF;
      Result := False;
    end else if prIsPUL9Dat(FName) then begin
      prFileType := QFS_QDF_TYPE_PUL;
    end else begin
      prFileType := QFS_QDF_TYPE_DAT;
    end;

  end else if Ext = '.ldt' then begin
  //LDT
    prFileType := QFS_QDF_TYPE_LDT;

  end else if Ext = '.pul' then begin
  //PUL
    prFileType := QFS_QDF_TYPE_PUL;

  end else if Ext = '.acquire' then begin
  //ACQ
    prFileType := QFS_QDF_TYPE_ACQ;

  end else if Ext = '.txt' then begin
  //ASC
    prFileType := QFS_QDF_TYPE_ASC;

  end else if Ext = '.abf' then begin
  //ABF
    prFileType := QFS_QDF_TYPE_ABF;

  end else if Ext = '.wav' then begin
  //WAV
    prFileType := QFS_QDF_TYPE_WAV;

  end else if Ext = '.dwt' then begin
  //DWT
    prFileType := QFS_QDF_TYPE_DWT;
    prCanEdit  := False;

  end else if Ext = '.evl' then begin
  //EVL
    prFileType := QFS_QDF_TYPE_EVL;
    prCanEdit  := False;

  end else if Ext = '.prt' then begin
  //PRT
    prFileType := QFS_QDF_TYPE_PRT;
    prCanEdit  := False;

  end else if Ext = '.tacidl' then begin
  //TAC
    prFileType := QFS_QDF_TYPE_TAC;
    prCanEdit  := False;

  end else if Ext = '.qsf' then begin
  //QSF
    prFileType := QFS_QDF_TYPE_QSF;
    prCanEdit  := False;

  end else if Ext = '.ibw' then begin
  //IBW = Igor Binary Wave
    prFileType := QFS_QDF_TYPE_IBW;
    prCanEdit  := False;

  end else begin
  //UNK
    prFileType := QFS_QDF_TYPE_DAT;
    Console.NewEntry('File type unknown: ' + FileName + ' - opened as raw binary', COLOR_WARNING);
  end;

end;


//****************************************************************************//
//                                                                            //
//                                   S a v e                                  //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prSave(const FileName :String) :Boolean;
var
  FStream :TFileStream;
  FType   :TQUB_QFS_QDF_FileType;
  OldType :TQUB_QFS_QDF_FileType;
  FName   :String;
  SName   :String;
  OldName :String;
  Entry   :Integer;
begin

  SName := FileName;
  FType := prGetFileTypeFromFileName(SName);
  FName := prGetFileNameForSave(ChangeFileExt(SName, ''), '.~s', False{IncFile}, 1000{MaxInc});

  Result := FName = '';
  if Result then Exit;

  FStream := nil;

  Entry := Console.NewEntry('Saving data file ' + SName + ' ...', COLOR_TEXT);

  try

    try

      prEditCritSection.Acquire;

      try
        FStream := TFileStream.Create(FName, fmCreate or fmOpenWrite or fmShareExclusive);
      except
        Result := True;
        Console.NewEntry('Could not create file ' + SName, COLOR_ERROR);
        Exit;
      end;

      prSaveStream := FStream;
      prSaveName   := SName;
      prSaveType   := FType;
      prSaveEntry  := Entry;

      QUB_BusyDlog.Caption := 'Saving '+SName+'...';
      QUB_BusyDlog.Action  := prDoBusySave;
      Result := prSaveResult or (QUB_BusyDlog.ShowModal <> mrOK);

    finally

      prSaveStream := nil;
      prEditCritSection.Release;

    end;


  finally


    Beep;

    if Result then begin
      Console.SetEntry(Entry, 'Saving data file ' + SName + ' ... Error!', COLOR_ERROR);
    end else begin
      Console.SetEntry(Entry, 'Saving data file ' + SName + ' ... Done.', COLOR_GRAY);
    end;

    try FStream.Free; except end;

    if Result then begin

      DeleteFile(FName);

    end else begin

      prSaved := True;

      // temporarily change filename so session file is saved with the new one
      OldName    := Self.FileName;
      OldType    := prFileType;
      prFileName := SName;
      prFileType := FType;
      Close;
      prFileName := OldName;
      prFileType := OldType;

      try

        if Self.FileName = SName then begin
          DeleteFile(Self.FileName);
        end;

        if FileExists(SName) then begin
          Result := not DeleteFile(SName);
          if Result then begin
            SName := prGetFileNameForSave(ChangeFileExt(SName, ''), ExtractFileExt(SName), True{IncFile}, 1000{MaxInc});
            Result := SName = '';
            if Result then begin
              Console.NewEntry('Could not save as requested. Destination file ' + FileName  + ' could not be overwritten', COLOR_ERROR);
            end else begin
              Console.NewEntry('Could not save as requested. Instead saved as ' + SName, COLOR_WARNING);
            end;
          end;
        end;

        if not Result then begin
          RenameFile(FName, SName);
        end;

      except

        Result := True;
        New(SName);

      end;

      if not Result then Result := Open(SName);

    end;

  end;

end;

procedure TQUB_QFS_QDF.prDoBusySave(OnUpdate :TQUB_BusyUpdate);
var
  s          :Integer;
begin
  prSaveResult := False;
  prSaveUpdate := OnUpdate;
  try
    if prSaveType = QFS_QDF_TYPE_QDF then begin
    //Native type

      prSaveResult := Write(prSaveStream);

    end else begin

      //File Header
      case prSaveType of
        QFS_QDF_TYPE_LDT:
          prSaveResult := prWriteLDTHeader(prSaveStream);
        QFS_QDF_TYPE_DAT:
          prSaveResult := prWriteDATHeader(prSaveStream);
        QFS_QDF_TYPE_ASC:
          prSaveResult := prWriteASCHeader(prSaveStream);
        QFS_QDF_TYPE_ABF:
          prSaveResult := prWriteABFHeader(prSaveStream);
        QFS_QDF_TYPE_WAV:
          prSaveResult := prWriteWAVHeader(prSaveStream);
      end;//case FType

      if prSaveResult then Exit;

      //Segments
      for s := 0 to prSegments.SegmentCount - 1 do begin

        Console.SetEntry(prSaveEntry, 'Saving data file ' + prSaveName + ' ... Segment ' + IntToStr(s + 1) + '/' + IntToStr(prSegments.SegmentCount), COLOR_TEXT);

        //Segment Header
        case prSaveType of
          QFS_QDF_TYPE_LDT:
            prSaveResult := prWriteLDTSegmentHeader(s, prSaveStream);
          QFS_QDF_TYPE_DAT:
            prSaveResult := prWriteDATSegmentHeader(s, prSaveStream);
          QFS_QDF_TYPE_ASC:
            prSaveResult := prWriteASCSegmentHeader(s, prSaveStream);
          QFS_QDF_TYPE_ABF:
            prSaveResult := prWriteABFSegmentHeader(s, prSaveStream);
          QFS_QDF_TYPE_WAV:
            prSaveResult := prWriteWAVSegmentHeader(s, prSaveStream);
        end;//case FType

        if prSaveResult then Exit;

        //Segment Data
        case prSaveType of
          QFS_QDF_TYPE_LDT:
            prSaveResult := prWriteLDTSegment(s, prSaveStream);
          QFS_QDF_TYPE_DAT:
            prSaveResult := prWriteDATSegment(s, prSaveStream);
          QFS_QDF_TYPE_ASC:
            prSaveResult := prWriteASCSegment(s, prSaveStream);
          QFS_QDF_TYPE_ABF:
            prSaveResult := prWriteABFSegment(s, prSaveStream);
          QFS_QDF_TYPE_WAV:
            prSaveResult := prWriteWAVSegment(s, prSaveStream);
        end;//case FType

        if prSaveResult then Exit;

        if not prSaveUpdate(s*100/prSegments.SegmentCount) then Exit;

      end;//of s

    end;//FType
  except
    prSaveResult := True;
  end;
  prSaveUpdate := nil;
end;


function TQUB_QFS_QDF.prGetFileExt :String;
begin
  Result := QUB_QFS_QDF_FILE_EXTS[prFileType];
end;

function TQUB_QFS_QDF.prGetFileTypeFromFileName(const FileName :String) :TQUB_QFS_QDF_FileType;
var
  Ext :String;
begin
  Ext := LowerCase(ExtractFileExt(FileName));
  if Ext = '' then begin
  //QDF
    Result := prFileType;
  end else if Ext = '.qdf' then begin
  //QDF
    Result := QFS_QDF_TYPE_QDF;
  end else if Ext = '.dat' then begin
  //DAT - must eliminate confusion between QUB dat and ABF dat
    if FileName = Self.FileName then begin
      if Self.FileType = QFS_QDF_TYPE_ABF then begin
        Result := QFS_QDF_TYPE_ABF;
      end else begin
        Result := QFS_QDF_TYPE_DAT;
      end;
    end else begin
      Result := QFS_QDF_TYPE_DAT;
    end;
  end else if Ext = '.ldt' then begin
  //LDT
    Result := QFS_QDF_TYPE_LDT;
  end else if Ext = '.pul' then begin
  //PUL
    Result := QFS_QDF_TYPE_PUL;
  end else if Ext = '.acquire' then begin
  //ACQ
    Result := QFS_QDF_TYPE_ACQ;
  end else if Ext = '.txt' then begin
  //ASC
    Result := QFS_QDF_TYPE_ASC;
  end else if Ext = '.abf' then begin
  //ABF
    Result := QFS_QDF_TYPE_ABF;
  end else if Ext = '.wav' then begin
  //WAV
    Result := QFS_QDF_TYPE_WAV;
  end else if Ext = '.dwt' then begin
  //DWT
    Result := QFS_QDF_TYPE_DWT;
  end else if Ext = '.evl' then begin
  //EVL
    Result := QFS_QDF_TYPE_EVL;
  end else if Ext = '.prt' then begin
  //PRT
    Result := QFS_QDF_TYPE_PRT;
  end else if Ext = '.tac' then begin
  //TAC
    Result := QFS_QDF_TYPE_TAC;
  end else if Ext = '.qsf' then begin
  //QSF
    Result := QFS_QDF_TYPE_QSF;
  end else if Ext = '.ibw' then begin
  //IBW
    Result := QFS_QDF_TYPE_IBW;
  end else begin
  //UNK
    Result := QFS_QDF_TYPE_UNK;
  end;
end;

function TQUB_QFS_QDF.CanSaveAs(var FileName :String; Confirmation :Boolean) :Boolean;
var
  FType :TQUB_QFS_QDF_FileType;
begin

  FType := prGetFileTypeFromFileName(FileName);

  Result := False;
  case FType of

    QFS_QDF_TYPE_QDF:
      Result := (True)  and (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_QSF) and (FileType <> QFS_QDF_TYPE_TAC);
    QFS_QDF_TYPE_DAT:
      Result := (True)  and (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_QSF) and (FileType <> QFS_QDF_TYPE_TAC);
    QFS_QDF_TYPE_LDT:
      Result := (True)  and (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_QSF) and (FileType <> QFS_QDF_TYPE_TAC);
    QFS_QDF_TYPE_ASC:
      Result := (True)  and (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_QSF) and (FileType <> QFS_QDF_TYPE_TAC);
    QFS_QDF_TYPE_ABF:
      Result := (False) and (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_QSF) and (FileType <> QFS_QDF_TYPE_TAC);
    QFS_QDF_TYPE_WAV:
      Result := (True)  and (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_QSF) and (FileType <> QFS_QDF_TYPE_TAC) and (ADChannelCount <= 2);
    QFS_QDF_TYPE_PUL:
      Result := False;
    QFS_QDF_TYPE_ACQ:
      Result := False;
    QFS_QDF_TYPE_DWT:
      Result := False;
    QFS_QDF_TYPE_EVL:
      Result := False;
    QFS_QDF_TYPE_PRT:
      Result := False;
    QFS_QDF_TYPE_TAC:
      Result := False;
    QFS_QDF_TYPE_QSF:
      Result := False;
    QFS_QDF_TYPE_IBW:
      Result := False;
    QFS_QDF_TYPE_UNK:
      Result := False;

  end;//case

  if (Result) and (Confirmation) then begin
  //Can be saved but the user wants Confirmation regarding lost features
    if (prSegments.SegmentCount > 1) and
      ((FType = QFS_QDF_TYPE_DAT) or (FType = QFS_QDF_TYPE_ASC) or (FType = QFS_QDF_TYPE_WAV)) then begin
      if not prSaveSilently then begin
        Result := not (MessageHelpDlg('Some features may be lost after saving (segmentation etc)'#13#10'Are you sure you want to save ?', mtConfirmation, [mbYes, mbNo], 0, 'Warning:Some_Data_Features_Lost') = idNo);
      end else begin
        Result := True;
      end;
    end;
  end;

end;

function TQUB_QFS_QDF.CanLoadInfo(ChannelIndex :Integer) :Boolean;
begin
  //TO DO: What calls this function? LM
  //File->load idealized data, which shouldn't load into a pure idealized file
  Result := (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_TAC);
end;

function TQUB_QFS_QDF.CanSaveInfo(ChannelIndex :Integer) :Boolean;
begin
  Result := prIdealization[ChannelIndex] <> nil;
end;


procedure TQUB_QFS_QDF.WhichPointForRealTime(Time :Double; var Point :Integer);
var
  p, s, t, i :Integer;
begin
  p := Point;
  for s := Segments.SegmentCount - 1 downto 0 do begin
    if ((Time + 0.4999999 * Sampling) >= Segments[s].StartTime) and
       ((Time - 0.4999999 * Sampling) <= (Segments[s].StartTime + Segments[s].Channels.DataCount * Sampling))
    then begin
      t := 0;
      for i := 0 to s - 1 do begin
        Inc(t, Segments[i].Channels.DataCount);
      end;
      p := t + Round((Time - Segments[s].StartTime) / Sampling);
      Break;
    end;
  end;
  Point := p;
end;

//procedure TQUB_QFS_QDF.WhichPointForRelTime(Time :Double; var Point :Integer);
//begin
//  WhichPointForRealTime(Time + Segments[0].StartTime, Point);
//end;


function TQUB_QFS_QDF.NewChild(ReadStream :TStream) :TQUB_QFS_Record;
var
  ReadNodeStream :TQUB_QFS_NodeStream;
  NodeName       :String;
  OldPos         :LongInt;
begin
  OldPos := ReadStream.Position;
  try
    ReadStream.ReadBuffer(ReadNodeStream, SizeOf(ReadNodeStream));
    SetLength(NodeName, ReadNodeStream.NameLen);
    ReadStream.ReadBuffer(PChar(NodeName)^, ReadNodeStream.NameLen);
    if NodeName          = QUB_QFS_QDF_SEGMENTS_NAME        then begin
      Result := prSegments;
    end else if NodeName = QUB_QFS_QDF_SAMPLING_NAME        then begin
      Result := prRecSampling;
    end else if NodeName = QUB_QFS_QDF_SCALING_NAME         then begin
      Result := prRecScaling;
    end else if NodeName = QUB_QFS_QDF_ADCHANNELCOUNT_NAME  then begin
      Result := prRecADChannelCount;
    end else if NodeName = QUB_QFS_QDF_ADDATASIZE_NAME      then begin
      Result := prRecADDataSize;
    end else if NodeName = QUB_QFS_QDF_ADDATATYPE_NAME      then begin
      Result := prRecADDataType;
    end else begin
      Result := inherited NewChild(ReadStream);
    end;
  finally
    ReadStream.Position := OldPos;
  end;
end;

//****************************************************************************//
//                                                                            //
//                       R e a d   -   l o w   l e v e l                      //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.Read(ReadStream :TStream) :Boolean;
var
  Entry :Integer;
begin
  Result := False;

  Entry := Console.NewEntry('Reading data file ' + FileName + ' ...', COLOR_TEXT);

  try

    prEditCritSection.Acquire;

    //Read according to the file type
    case prFileType of

      //Native QDF file format
      QFS_QDF_TYPE_QDF: Result := prReadQDF;

      //Old raw binary format
      QFS_QDF_TYPE_DAT: Result := prReadDAT;

      //Old LDT binary format
      QFS_QDF_TYPE_LDT: Result := prReadLDT;

      //HEKA Pulse
      QFS_QDF_TYPE_PUL: Result := prReadPUL;

      QFS_QDF_TYPE_ACQ: Result := prReadACQ;
      //ASCII format
      QFS_QDF_TYPE_ASC: Result := prReadASC;

      //Axon Binary File
      QFS_QDF_TYPE_ABF:
        begin
          Result := prReadABF;
          if Result then begin
            if MessageDlg('Failed to open file ' + FileName + ' as ABF. Would you like to try it as DAT (raw binary)?', mtError, [mbYes, mbNo], 0) = idYes then begin
              prFileType := QFS_QDF_TYPE_DAT;
              Result := prReadDAT;
            end;
          end;
        end;

      //WAVE Sound File
      QFS_QDF_TYPE_WAV: Result := prReadWAV;

      //DWT Idealized data
      QFS_QDF_TYPE_DWT: Result := prReadDWT;

      //EVL Idealized data
      QFS_QDF_TYPE_EVL: Result := prReadEVL;

      //PRT -- EKDIST printout
      QFS_QDF_TYPE_PRT: Result := prReadPRT;

      //TAC -- TAC table of idealized transitions
      QFS_QDF_TYPE_TAC: Result := prReadTACidl;

      //QSF's Idealized data
      QFS_QDF_TYPE_QSF: Result := prReadQSF;

      //IBW - Igor Binary Wave
      QFS_QDF_TYPE_IBW: Result := prReadIBW;

    end;//case


    if Result then Exit;


    if (FileType <> QFS_QDF_TYPE_DWT) and (FileType <> QFS_QDF_TYPE_EVL) and (FileType <> QFS_QDF_TYPE_PRT) and (FileType <> QFS_QDF_TYPE_QSF) and (FileType <> QFS_QDF_TYPE_TAC) then begin
      prNewIdealization;
    end;


  finally


    prEditCritSection.Release;

    if Result then begin
      Console.SetEntry(Entry, 'Reading data file ' + FileName + ' ... Error!', COLOR_ERROR);
    end else begin
      Console.SetEntry(Entry, 'Reading data file ' + FileName + ' ... Done.', COLOR_GRAY);
    end;

  end;

end;

//------------------------------------------------------------------------------
// Read IGOR binary Wave file.
// 126 byte header followed by array of data (floats).
//****************************************************************************//
function TQUB_QFS_QDF.prReadIBW :Boolean;
var
  iSig, iType, iFullScaleValid            :Smallint;
  i, iHd1, iHd2, iSize, iPoints           :Integer;
  nSampling, nTopFullScale, nBotFullScale :Double ;
  Selection                               :TQUB_QFS_PEditSelection;
  PropForm                                :TQUB_QFS_QDF_PropertiesForm;
  Segment                                 :TQUB_QFS_Segment;
  nMax, n                                 :Single;

begin

  Reset;

  Result := false;

  try

    //----- Read File Header
    StreamPos                      := Stream.Position;

    // iSig : Should be 1,2,3, or 5.
    Stream.Seek(0, soFromBeginning);
    Stream.ReadBuffer(iSig, 2);
    m_lMacData := ((iSig=256) or (iSig=2*256) or (iSig=3*256) or (iSig=5*256));
    if m_lMacData then Swap2Bytes(@iSig);

    if iSig=1 Then Begin
      iHd1:=8;
      iHd2:=110;
    end else if iSig=2 Then Begin
      iHd1:=16;
      iHd2:=110;
    end else if iSig=3 Then Begin
      iHd1:=20;
      iHd2:=110;
    end else if iSig=5 Then Begin
      iHd1:=64;
      iHd2:=320;
    end else Begin
      Result := True;
      exit;
    end;

    // iSize : Data size + headers.
    Stream.Seek( Choice(iSig=5,4,2), soFromBeginning );
    Stream.ReadBuffer( iSize, 4);

    // iType : bitmask - 2=float 4=double 8=byte 16=short 32=long ( +64 for unsigned ) (+1 for complex)
    // Only accept 2 for now
    Stream.Seek( iHd1 + Choice(iSig=5,16,0), soFromBeginning);
    Stream.ReadBuffer( iType, 2);

    // iPoints : 3 traces of 100 pts = 300 pts.
    Stream.Seek( iHd1 + Choice(iSig=5,12,42), soFromBeginning);
    Stream.ReadBuffer( iPoints, 4);

    // nSampling : i.e.  .0002 for 2 microseconds.
    Stream.Seek( iHd1 + Choice(iSig=5,84,48), soFromBeginning);
    Stream.ReadBuffer( nSampling, 8);

    // iFullScaleValid : 0=no 1=yes use "FullScale" scaling values
    Stream.Seek( iHd1 + Choice(iSig=5,168,68), soFromBeginning);
    Stream.ReadBuffer( iFullScaleValid, 2);
    Stream.Seek( iHd1 + Choice(iSig=5,172,70), soFromBeginning);
    Stream.ReadBuffer( nTopFullScale, 8);
    Stream.ReadBuffer( nBotFullScale, 8);

    if m_lMacData Then Begin
      Swap4Bytes(@iSize);
      Swap2Bytes(@iType);
      Swap4Bytes(@iPoints);
      Swap8Bytes(@nSampling);
      Swap2Bytes(@iFullScaleValid);
      Swap8Bytes(@nTopFullScale);
      Swap8Bytes(@nBotFullScale);
    end;

    // err if any header issues arose ...
    // Only accept 4 byte floats for now ...
    if (iSize<=126) or (iType<>2) Then Begin
      Result := True;
      exit;
    end;

    //----- Pre-read the data to calculate a multiplier
    Stream.Seek( iHd1 + iHd2, soFromBeginning);
    nMax := 0.0;
    for i := 0 to iPoints-1 do Begin
      Stream.ReadBuffer( n, 4);
      if m_lMacData Then Swap4Bytes( @n );
      if abs(n)>nMax Then nMax := abs(n);
    end;
    if nMax > 0.0 Then Begin
      m_nMultiplier := Power(10,Round(log10(10000/nMax)));
      Scaling := m_nMultiplier;
    end else begin
      m_nMultiplier := 1.0;
      Scaling := 1.0;
    end;

    //----- Fill in QUB Values
    Sampling := nSampling;  // * 1e-6
    Scaling := Scaling * 1e-12;
    //Scaling := 1;   // 1E12

    ADDataSize := 4;
    ADDataType := NODE_DATATYPE_SINGLE;
    ADChannelCount := Round((iSize-iHd1-iHd2)/(sizeof(Single)*iPoints));

    //Interactive properties - for confirmation of Scaling only
    if not prOpenSilently then begin

      PropForm                          := TQUB_QFS_QDF_PropertiesForm.Create(nil);
      PropForm.chkTimeColumn.Enabled    := False;
      PropForm.chkSegmented.Enabled     := False;
      PropForm.chkMultiplexed.Enabled   := False;
      PropForm.chkMPXAA.Enabled         := False;
      PropForm.txtChannelCount.Enabled  := False;
      PropForm.chkADDataSize2.Enabled   := False;
      PropForm.chkADDataSize4.Enabled   := False;
      PropForm.txtSamplingMS.Enabled    := False;
      PropForm.txtSamplingKHz.Enabled   := False;

      PropForm.ADChannelCount           := ADChannelCount;
      PropForm.ADDataSize               := ADDataSize;
      PropForm.Scaling                  := Scaling;
      PropForm.Sampling{milisecs}       := Sampling{secs}*1.0e3;

      PropForm.lblFileName.Caption      := FileName;

      try

        PropForm.ShowModal;

        case PropForm.ModalResult of

          idOK:
            begin
              Scaling    := PropForm.Scaling;
            end;

          idCancel:
            begin
              Result := True;
            end;

        end;//case ModalResult

      finally

        PropForm.Free;

      end;

    end;//Silently

    if Result then Exit;

    //----------------------------------------------------------
    Stream.Seek(126, soFromBeginning);
    StreamPos             := Stream.Position;

    ADDataType := NODE_DATATYPE_SINGLE;

    //----- Read file DATA
    //Segments
    prSegments.Stream                := Stream;
    prSegments.StreamPos             := Stream.Position;

    //Segment
    Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
    Segment.Stream                   := Stream;
    Segment.StreamPos                := Stream.Position;

    //Channels
    Segment.Channels.CanEdit         := CanEdit;
    Segment.Channels.EditStream      := EditStream;
    Segment.Channels.Stream          := Stream;
    Segment.Channels.StreamPos       := Stream.Position;
    Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
    Segment.Channels.DataType        := ADDataType;
    Segment.Channels.DataCount       := 0;
    Segment.Channels.DataStreamPos   := Stream.Position;

    DataCount := iPoints;
    if DataCount <> 0 then begin
      Selection                      := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount            := DataCount;
      Selection.StreamPos            := Segment.Channels.DataStreamPos;
      Selection.FromEditStream       := False;
      Segment.Channels.InsertSelection(Selection, 0);
    end;

    prSegments.AddRecord(Segment, prSegments.SegmentCount);

  except
    Result := true;

  end;

end;

//****************************************************************************//
//                                                                            //
//                               R e a d    Q D F                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadQDF :Boolean;
var
  i      :Integer;
begin

  Result := inherited Read(Stream);

  if not Result then begin
    for i := 0 to prSegments.SegmentCount - 1 do begin
      prSegments[i].Channels.EditStream      := EditStream;
    end;
  end;
end;

//****************************************************************************//
//                                                                            //
//                               R e a d    D A T                             //
//                                                                            //
//****************************************************************************//
//12-13-05 LM
//I modified this procedure to allow timecolumn and Single and UnevenSampling
function TQUB_QFS_QDF.prReadDAT :Boolean;
var
  Segment    :TQUB_QFS_Segment;
  Selection  :TQUB_QFS_PEditSelection;
  PropForm   :TQUB_QFS_QDF_PropertiesForm;
  DataCount  :Integer;
  SegPos     :Integer;
  Entry      :Integer;
begin

  Result := False;

  Reset;


  //Interactive properties
  if not prOpenSilently then begin

    PropForm                            := TQUB_QFS_QDF_PropertiesForm.Create(nil);
    PropForm.chkTimeColumn.Enabled      := True;
    PropForm.chkUnevenSampling.Enabled  := True;
    PropForm.chkSinglePrec.Enabled      := True;
    PropForm.chkSegmented.Enabled       := False;//Applies only to ASCII files
    PropForm.chkMultiplexed.Enabled     := False;//- ,, -
    PropForm.chkMPXAA.Enabled           := False;//- ,, -

    PropForm.TimeColumn                 := prOpenTimeColumn;
    PropForm.UnevenSampling             := prOpenUnevenSampling;
    PropForm.SinglePrec                 := prOpenSinglePrec;

    PropForm.ADChannelCount             := ADChannelCount;
    PropForm.ADDataSize                 := ADDataSize;
    PropForm.Scaling                    := Scaling;
    PropForm.Sampling{milisecs}         := Sampling{secs}*1.0e3;

    PropForm.lblFileName.Caption        := FileName;

    try

      PropForm.ShowModal;

      case PropForm.ModalResult of

        idOK:
          begin
            prOpenTimeColumn            := PropForm.TimeColumn;
            prOpenUnevenSampling        := PropForm.UnevenSampling;
            prOpenSinglePrec            := PropForm.SinglePrec;
            ADChannelCount              := PropForm.ADChannelCount;
            ADDataSize                  := PropForm.ADDataSize;
            Sampling{secs}              := PropForm.Sampling{milisecs}*1.0e-3;
            Scaling                     := PropForm.Scaling;
          end;

        idCancel:
          begin
            Result := True;
          end;

      end;//case ModalResult

    finally

      PropForm.Free;

    end;

  end;

  if Result then Exit;


  try

    //Data
    if ADDataSize = 2 then ADDataType := NODE_DATATYPE_SMALLINT else ADDataType := NODE_DATATYPE_LONGINT;


    if (not prOpenTimeColumn) and (not prOpenSinglePrec) then begin
    //Just Data, EvenSampling

      //Stream is already assigned, supposedly
      StreamPos                        := Stream.Position;

      //Segments
      prSegments.Stream                := Stream;
      prSegments.StreamPos             := Stream.Position;

      //Segment
      Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      Segment.Stream                   := Stream;
      Segment.StreamPos                := Stream.Position;

      //Channels
      Segment.Channels.CanEdit         := CanEdit;
      Segment.Channels.EditStream      := EditStream;
      Segment.Channels.Stream          := Stream;
      Segment.Channels.StreamPos       := Stream.Position;
      Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
      Segment.Channels.DataType        := ADDataType;
      Segment.Channels.DataCount       := 0;
      Segment.Channels.DataStreamPos   := Stream.Position;
      DataCount := Stream.Size div (ADDataSize * ADChannelCount);


    end else begin
    //It may be Single and may have TimeColumn, and possibly UnevenSampling

      Entry                            := Console.NewEntry('', COLOR_TEXT);

      StreamPos                        := 0;

      //Segments
      prSegments.Stream                := EditStream;
      prSegments.StreamPos             := 0;

      SegPos                           := prEditStream.Position;

      if prOpenSinglePrec then begin
        Result := prConvertSingle(Stream{Source}, ADChannelCount, prOpenTimeColumn, prOpenUnevenSampling, DataCount, Segments.SegmentCount, Entry);
      end else begin
        Result := prConvertInteger(Stream{Source}, ADChannelCount, prOpenTimeColumn, prOpenUnevenSampling, DataCount, Segments.SegmentCount, Entry);
      end;

      //Segment
      Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      Segment.Stream                   := EditStream;
      Segment.StreamPos                := SegPos;

      Segment.StartTime                := prSegments.TotalDataCount * Sampling;
      
      //Channels
      Segment.Channels.CanEdit         := CanEdit;
      Segment.Channels.EditStream      := EditStream;
      Segment.Channels.Stream          := EditStream;
      Segment.Channels.StreamPos       := SegPos;
      Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
      Segment.Channels.DataType        := ADDataType;
      Segment.Channels.DataCount       := 0;
      Segment.Channels.DataStreamPos   := SegPos;

    end;//TimeColumn

    if DataCount > 0 then begin
      Selection                      := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount            := DataCount;
      Selection.StreamPos            := Segment.Channels.DataStreamPos;
      Selection.FromEditStream       := False;
      Segment.Channels.InsertSelection(Selection, 0);
    end;

    prSegments.AddRecord(Segment, prSegments.SegmentCount);

  except

    Result := True;

  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    L D T                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadLDT :Boolean;
var
  Segment    :TQUB_QFS_Segment;
  Selection  :TQUB_QFS_PEditSelection;
  StartTime  :Integer;
  DataPoints :Integer;
  DataCount  :Integer;
  SessScaling :INode;
begin

  Result := False;

  Reset;

  try

    StreamPos                      := Stream.Position;

    //Read file  H E A D E R ***************************************************

    //Leader
    Stream.ReadBuffer(prLDTLeader, SizeOf(prLDTLeader));


    //Data
    //Sampling
    Stream.ReadBuffer(prLDTSampling, SizeOf(prLDTSampling));
    Sampling       := prLDTSampling * 1.0e-6;
    //Scaling
    Stream.ReadBuffer(prLDTScaling, SizeOf(prLDTScaling));
    SessScaling    := prSession.Find('Scaling');
    if SessScaling.DataCount > 0 then begin
      Scaling      := Double(SessScaling.Data[0]);
    end else begin
      Scaling      := prLDTScaling / 100;
    end;

    //ADChannels
    ADChannelCount := 1;
    ADDataSize     := SizeOf(SmallInt);
    ADDataType     := NODE_DATATYPE_SMALLINT;


    prLDTLeader := prLDTLeader div prLDTSampling;



    //Read file  D A T A *******************************************************


    //Segments
    prSegments.Stream                := Stream;
    prSegments.StreamPos             := Stream.Position;

    //Segment
    while (Stream.Position < Stream.Size) do begin

      //Segment
      Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      Segment.Stream                   := Stream;
      Segment.StreamPos                := Stream.Position;

      //StartTime
      Stream.ReadBuffer(StartTime, SizeOf(StartTime));
      Segment.StartTime := StartTime * 1e-6;

      //DataPoints
      Stream.ReadBuffer(DataPoints, SizeOf(DataPoints));
      DataPoints := DataPoints - 2 * prLDTLeader;

      Stream.Seek(prLDTLeader * ADChannelCount * ADDataSize, soFromCurrent);

      //Channels
      Segment.Channels.CanEdit         := CanEdit;
      Segment.Channels.EditStream      := EditStream;
      Segment.Channels.Stream          := Stream;
      Segment.Channels.StreamPos       := Stream.Position;
      Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
      Segment.Channels.DataType        := ADDataType;
      Segment.Channels.DataCount       := 0;
      Segment.Channels.DataStreamPos   := Stream.Position;

      DataCount := DataPoints;
      if DataCount <> 0 then begin
        Selection                      := AllocMem(SizeOf(TQUB_QFS_EditSelection));
        Selection.DataCount            := DataCount;
        Selection.StreamPos            := Segment.Channels.DataStreamPos;
        Selection.FromEditStream       := False;
        Segment.Channels.InsertSelection(Selection, 0);
      end;

      prSegments.AddRecord(Segment, prSegments.SegmentCount);

      Stream.Seek((DataPoints + prLDTLeader) * ADChannelCount * ADDataSize, soFromCurrent);

    end;

  except

    Result := True;

  end;

end;



//****************************************************************************//
//                                                                            //
//                               R e a d    P U L                             //
//                                                                            //
//****************************************************************************//

// must do:
//    find out where the data actually is in the .dat
//    verify sampling and scaling
//    find out how LeakTrace, SecondTrace, ZeroCurrent, DataAbscissa play into things
//    straighten out the start time between root, series, and sweep


function TQUB_QFS_QDF.prReadPUL :Boolean;
var
  PropForm        :TQUB_QFS_QDF_PropertiesForm;
  Segment         :TQUB_QFS_Segment;
  Selection       :TQUB_QFS_PEditSelection;
  PulseTree       :THEKA_Tree_Pulse;
  StimTree        :THEKA_Tree_StimRoot;
  StimStream      :TFileStream;
  Sweep           :TPSweepRecord;
  SweepDataType   :TQUB_QFS_Node_DataType;
  SweepName       :String;
  SeriesName      :String;
  SweepSel        :TQUB_FirstLast;
  i, j, k         :Integer;
  //PulseStat       :TStringStream;
begin

  Result          := False;

  PulseTree       := nil;
  StimTree        := nil;
  StimStream      := nil;

  Reset;


  try


    try

      //PulseStat                      := TStringStream.Create('');

      StreamPos                      := Stream.Position;

      // read pulse tree (description)

      PulseTree                      := LoadPulseTree(Stream);
      if PulseTree = nil then begin
        Stream.Seek(0, soFromBeginning);
        Result := prReadPUL9;
        Exit;
      end;

      //PulseTree.Print('', PulseStat);

      // switch over to the raw data stream
      Stream.Free;
      Stream := nil;
      //Stream := TFileStream.Create( ExtractFilePath(FileName) + PulseTree.Rec.FileName + '.dat', fmOpenRead or fmShareDenyWrite );
      {
        I used commas in the name and it comes all messed up in PulseTree.Rec.FileName.
        LM.
      }
      Stream := TFileStream.Create(ChangeFileExt(FileName, '.dat'), fmOpenRead or fmShareDenyWrite);

      // get basic info from the first sweep
      Sweep          := PulseTree[0][0][0].Rec;
      ADDataSize     := SizeOf(SmallInt);
      ADDataType     := NODE_DATATYPE_SMALLINT;

      if Session.Find('Scaling').IsNull then begin
        Scaling        := 1.0e-12 / Sweep.DataFactor1; //  * 32768.0);
      end;

      if Session.Find('Sampling').IsNull then begin
        Sampling       := 1.0 / 20000;                    // default in case pgf can't be read
        try
          StimStream   := TFileStream.Create( ChangeFileExt(FileName, '.pgf'), fmOpenRead or fmShareDenyWrite );
          StimTree     := LoadStimTree( StimStream );
          if StimTree <> nil then begin
            Sampling   := StimTree[ Sweep.StimCount-1 ].Rec.SampleInterval;
              //TODO : should above be Sweep.StimCount-1 ?  (Heka files error -1 added 7/2004 )
            if (Sampling < 1.0e-8) or (Sampling > 1.0e1) then begin
              Sampling := 1.0e-5;
            end;
            //StimTree.Print('', PulseStat);
          end;
        except
        end;
        StimStream.Free;
      end;

      //Console.NewEntry(PulseStat.DataString, COLOR_TEXT);
      //PulseStat.Free;

      ADChannelCount := 1;
      for i := 0 to PulseTree.Count - 1 do begin             // foreach group
        for j := 0 to PulseTree[i].Count - 1 do begin        // foreach series
          for k := 0 to PulseTree[i][j].Count - 1 do begin   // foreach sweep
            Sweep                := PulseTree[i][j][k].Rec;
            if (Sweep.SecondTrace) and (ADChannelCount < 2) then begin
              ADChannelCount := 2;
            end;
            if (Sweep.Leak) then begin
              ADChannelCount := 3;   //TO DO: LM. Make this an option to see Leak trace?
            end;
          end;
        end;
      end;


      //Interactive properties
      if not prOpenSilently then begin

        PropForm                             := TQUB_QFS_QDF_PropertiesForm.Create(nil);
        PropForm.chkTimeColumn.Enabled       := False;
        PropForm.chkUnevenSampling.Enabled   := False;
        PropForm.chkSinglePrec.Enabled       := False;
        PropForm.chkSegmented.Enabled        := False;
        PropForm.chkMultiplexed.Enabled      := False;
        PropForm.chkMPXAA.Enabled            := False;
        PropForm.txtChannelCount.Enabled     := False;
        PropForm.udChannelCount.Enabled      := False;
        PropForm.chkADDataSize2.Enabled      := False;
        PropForm.chkADDataSize4.Enabled      := False;

        PropForm.ADChannelCount              := ADChannelCount;
        PropForm.ADDataSize                  := ADDataSize;
        PropForm.Scaling                     := Scaling;             //only Scaling and Sampling
        PropForm.Sampling{milisecs}          := Sampling{secs}*1.0e3;//must be confirmed by the user

        PropForm.lblFileName.Caption         := FileName;

        try

          PropForm.ShowModal;

          case PropForm.ModalResult of

            idOK:
              begin
                Sampling{secs} := PropForm.Sampling{milisecs}*1.0e-3;
                Scaling        := PropForm.Scaling;
              end;

            idCancel:
              begin
                Result := True;
              end;

          end;//case ModalResult

        finally

          PropForm.Free;

        end;

        if Result then Exit;

      end;//Silently


      prSegments.Stream          := Stream;
      prSegments.StreamPos       := Stream.Position;

      for i := 0 to PulseTree.Count - 1 do begin             // foreach group
      //Group

        for j := 0 to PulseTree[i].Count - 1 do begin        // foreach series
        //Series

          //Add list - use an event mechanism to communicate with Parent QVD, if any
          SeriesName := 'Series ' + IntToStr(j);
          if Assigned(OnAddSeries) then begin
            OnAddSeries(Self, SeriesName, True{Private});
          end;

          for k := 0 to PulseTree[i][j].Count - 1 do begin   // foreach sweep
          //Sweep

            Sweep                := PulseTree[i][j][k].Rec;

            Stream.Seek(Sweep.Data, soFromBeginning);

            SweepDataType      := NODE_DATATYPE_SMALLINT;
            if 0 <> (Sweep.SweepKind and (ExtendedBitMask or DataFormatBitMask)) then begin
              case Sweep.DataFormat of
                int16:  SweepDataType := NODE_DATATYPE_SMALLINT;
                int32:  SweepDataType := NODE_DATATYPE_LONGINT;   //I fixed this mistake (was it?). LM.
                real32: SweepDataType := NODE_DATATYPE_SINGLE;
                real64: SweepDataType := NODE_DATATYPE_DOUBLE;
              end;
            end;

            Segment              := nil;
            if ADChannelCount > 1 then begin // i.e. if 2 or 3 channels

              Segment := prReadSegment2or3Chan(Sweep.TotalPoints, SweepDataType, 0 = (Sweep.SweepKind and LittleEndianBitMask), Sweep.Leak, Sweep.SecondTrace);
              {if 0 = (Sweep.SweepKind and LittleEndianBitMask) then begin
                if SweepDataType = NODE_DATATYPE_LONGINT then begin
                  Segment := prReadSegmentOfInts2Chan(Sweep.TotalPoints, SizeOf(Integer), True, Sweep.Leak, Sweep.SecondTrace);
                end else if SweepDataType = NODE_DATATYPE_SMALLINT then begin
                  Segment := prReadSegmentOfInts2Chan(Sweep.TotalPoints, SizeOf(SmallInt), True, Sweep.Leak, Sweep.SecondTrace);
                end;
              end else begin
                if SweepDataType = NODE_DATATYPE_DOUBLE then begin
                  Segment := prReadSegmentOfReals2Chan(Sweep.TotalPoints, SizeOf(Double), Sweep.Leak, Sweep.SecondTrace);
                end else if SweepDataType = NODE_DATATYPE_SINGLE then begin
                  Segment := prReadSegmentOfReals2Chan(Sweep.TotalPoints, SizeOf(Single), Sweep.Leak, Sweep.SecondTrace);
                end else if SweepDataType = NODE_DATATYPE_LONGINT then begin
                  Segment := prReadSegmentOfInts2Chan(Sweep.TotalPoints, SizeOf(Integer), False, Sweep.Leak, Sweep.SecondTrace);
                end else if SweepDataType = NODE_DATATYPE_SMALLINT then begin
                  Segment := prReadSegmentOfInts2Chan(Sweep.TotalPoints, SizeOf(SmallInt), False, Sweep.Leak, Sweep.SecondTrace);
                end;
              end;
              }

            end else begin

              if 0 = (Sweep.SweepKind and LittleEndianBitMask) then begin
                if SweepDataType = NODE_DATATYPE_LONGINT then begin
                  Segment := prReadSegmentOfInts(Sweep.TotalPoints, SizeOf(Integer), True);
                end else if SweepDataType = NODE_DATATYPE_SMALLINT then begin
                  Segment := prReadSegmentOfInts(Sweep.TotalPoints, SizeOf(SmallInt), True);
                end;
              end else begin
                if SweepDataType = NODE_DATATYPE_DOUBLE then begin
                  Segment := prReadSegmentOfReals( Sweep.TotalPoints, SizeOf(double), Nil );
                end else if SweepDataType = NODE_DATATYPE_SINGLE then begin
                  Segment := prReadSegmentOfReals( Sweep.TotalPoints, SizeOf(single), Nil );
                end else if SweepDataType = NODE_DATATYPE_LONGINT then begin
                  Segment := prReadSegmentOfInts(Sweep.TotalPoints, SizeOf(Integer), False);
                end else if SweepDataType = NODE_DATATYPE_SMALLINT then begin

                  Segment                     := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
                  Segment.Stream              := Stream;
                  Segment.StreamPos           := Sweep.Data;

                  Segment.Channels.CanEdit    := CanEdit;
                  Segment.Channels.EditStream := EditStream;
                  Segment.Channels.Stream     := Stream;
                  Segment.StreamPos           := Sweep.Data;
                  Segment.Channels.DataSize   := ADChannelCount * ADDataSize;
                  Segment.Channels.DataType   := ADDataType;
                  Segment.Channels.DataCount  := 0;
                  Segment.Channels.DataStreamPos := Sweep.Data;

                  Selection                   := AllocMem(SizeOf(TQUB_QFS_EditSelection));
                  Selection.DataCount         := Sweep.TotalPoints;
                  Selection.StreamPos         := Segment.Channels.DataStreamPos;
                  Selection.FromEditStream    := False;
                  Segment.Channels.InsertSelection(Selection, 0);

                end;
              end;

            end;//ADChannelCount

            if Segment <> nil then begin
              // Segment.StartTime    := Sweep.Time;

              SweepSel.First := Segments.TotalDataCount;
              SweepSel.Last  := SweepSel.First + Sweep.TotalPoints - 1;
              prSegments.AddRecord(Segment, prSegments.SegmentCount);

              SweepName            := 'Sweep ' + IntToStr(k);
              //Add Selection - use an event mechanism to communicate with Parent QVD, if any
              if Assigned(OnAddSweep) then begin
                OnAddSweep(Self, SeriesName, SweepName, SweepSel, True{ForceAdd});
              end;
            end;

          end;//k - Sweep == Segment

        end;//j - Series

      end;//i - Group

    except

      Result := True;

    end;

  finally

    PulseTree.Free;
    StimTree.Free;

  end;

end;


function TQUB_QFS_QDF.prReadPUL9: Boolean;
var
  SweepSel        :TQUB_FirstLast;
  PropForm        :TQUB_QFS_QDF_PropertiesForm;
  PulseTree       :THEKA_Tree_Pulse9;
  Sweep           :THEKA_Tree_Pulse9_Sweep;
  Trace           :TPTraceRecord9;
  SweepName       :String;
  SeriesName      :String;
  i, j, k, t      :Integer;
  Ch              :INode;
  Scale           :Array of Double;
  Units           :String;
begin

  Result          := False;

  PulseTree       := nil;

  Reset;


  try


    try

      //PulseStat                      := TStringStream.Create('');

      StreamPos                      := Stream.Position;

      // read pulse tree (description)

      PulseTree                      := LoadPulse9Tree(Stream);
      if PulseTree = nil then begin
        Result := True;
        Exit;
      end;

      //PulseTree.Print('', PulseStat);

      if (PulseTree.Count = 0)             // no groups
      or (PulseTree[0].Count = 0)          // no series
      or (PulseTree[0][0].Count = 0)       // no sweeps
      or (PulseTree[0][0][0].Count = 0)    // no traces
      then begin
        Result := True;
        Exit;
      end;

      // switch over to the raw data stream
      Stream.Free;
      Stream := nil;
      //Stream := TFileStream.Create( ExtractFilePath(FileName) + PulseTree.Rec.FileName + '.dat', fmOpenRead or fmShareDenyWrite );
      {
        I used commas in the name and it comes all messed up in PulseTree.Rec.FileName.
        LM.
      }
      Stream := TFileStream.Create(ChangeFileExt(FileName, '.dat'), fmOpenRead or fmShareDenyWrite);

      ADDataSize     := SizeOf(SmallInt);
      ADDataType     := NODE_DATATYPE_SMALLINT;

      // get basic info from the first trace
      Sweep          := PulseTree[0][0][0];
      Trace          := Sweep[0].Rec;

      if Session.Find('Scaling').IsNull then begin
        Scaling        := 1.0; // 1.0e-12 / Trace.TrDataScaler;
      end else begin
        Scaling        := Session['Scaling'].Data[0];
      end;

      if Session.Find('Sampling').IsNull then begin
        Sampling       := Trace.XInterval;
      end else begin
        Sampling       := Session['Sampling'].Data[0];
      end;

      ADChannelCount := Sweep.Count;
      SetLength(Scale, ADChannelCount);
      Ch             := Session['DataChannels'].Find('Channel');
      for t := 0 to ADChannelCount - 1 do begin
        if Ch.IsNull then begin
          Ch         := Session['DataChannels'].AppendChild('Channel');
          Ch['Name'].SetData(String(Sweep[t].Rec.Lbl));
          Units      := Sweep[t].Rec.YUnit;
          Scale[t]   := Sweep[t].Rec.DataScaler;
          if Units = 'V' then begin
            Units    := 'mV';
            Scale[t] := Scale[t] * 1e3;
          end else if Units = 'A' then begin
            Units    := 'pA';
            Scale[t] := Scale[t] * 1e12;
          end;
          Ch['Units'].SetData(Units);
          Ch['Scaling'].SetData(1.0 / Scale[t]);
        end;
        Ch           := Ch.SiblingSameName;
      end;



      prSegments.Stream          := Stream;
      prSegments.StreamPos       := Stream.Position;

      for i := 0 to PulseTree.Count - 1 do begin             // foreach group
      //Group

        for j := 0 to PulseTree[i].Count - 1 do begin        // foreach series
        //Series

          //Add list - use an event mechanism to communicate with Parent QVD, if any
          SeriesName := PulseTree[i][j].Rec.Lbl;
          SeriesName := 'Series ' + IntToStr(j);
          if PulseTree[i][j].Rec.Lbl <> '' then begin
            SeriesName := SeriesName + ': ' + PulseTree[i][j].Rec.Lbl;
          end;
          if Assigned(OnAddSeries) then begin
            OnAddSeries(Self, SeriesName, True{Private});
          end;

          for k := 0 to PulseTree[i][j].Count - 1 do begin   // foreach sweep
          //Sweep

            Sweep                := PulseTree[i][j][k];
            SweepSel.First       := Segments.TotalDataCount;
            SweepSel.Last        := SweepSel.First + Sweep[0].Rec.DataPoints - 1;

            prAddPulse9Sweep(Sweep, @(Scale[0]));

            //Add Selection - use an event mechanism to communicate with Parent QVD, if any
            if Assigned(OnAddSweep) then begin
              OnAddSweep(Self, SeriesName, Sweep.Rec.Lbl, SweepSel, True{ForceAdd});
            end;

          end;//k - Sweep == Segment

        end;//j - Series

      end;//i - Group

    except

      Result := True;

    end;

  finally

    PulseTree.Free;

  end;

end;

procedure TQUB_QFS_QDF.prAddPulse9Sweep(Sweep :THEKA_Tree_Pulse9_Sweep; Scale :TPADouble);
var
  t               :Integer;
  Leaf            :Integer;
  Segment         :TQUB_QFS_Segment;
  Selection       :TQUB_QFS_PEditSelection;
  BigEndian       :Boolean;
  TraceDataType   :TQUB_QFS_Node_DataType;
  TraceDataSize   :Integer;
  Trace           :TPTraceRecord9;
  BuffRaw         :Array of Char;
  BuffMPX         :Array of SmallInt;
  LeafSize        :Integer;
  LeafStart       :Integer;

  procedure CopySmallTrace;
  var
    Buff          :TPASmallInt;
    i             :Integer;
  begin
    Buff          := TPASmallInt(@(BuffRaw[0]));
    for i := 0 to LeafSize - 1 do begin
      if BigEndian then begin
        Swap2Bytes(@(Buff[i]));
      end;
      BuffMPX[i*ADChannelCount+t] := Buff[i];
    end;
  end;

  // TODO: get sample data and get the scaling right for LongInt, Single, Double
  procedure CopyLongTrace;
  var
    Buff          :TPALongInt;
    i             :Integer;
  begin
    Buff          := TPALongInt(@(BuffRaw[0]));
    for i := 0 to LeafSize - 1 do begin
      if BigEndian then begin
        Swap4Bytes(@(Buff[i]));
      end;
      BuffMPX[i*ADChannelCount+t] := SmallInt(Buff[i]); // bad scaling?
    end;
  end;

  procedure CopySingleTrace;
  var
    Buff          :TPASingle;
    i             :Integer;
  begin
    Buff          := TPASingle(@(BuffRaw[0]));
    for i := 0 to LeafSize - 1 do begin
      if BigEndian then begin
        Swap4Bytes(@(Buff[i]));
      end;
      BuffMPX[i*ADChannelCount+t] := SmallInt(Round(Buff[i])); // bad scaling?
    end;
  end;

  procedure CopyDoubleTrace;
  var
    Buff          :TPADouble;
    i             :Integer;
  begin
    Buff          := TPADouble(@(BuffRaw[0]));
    for i := 0 to LeafSize - 1 do begin
      if BigEndian then begin
        Swap4Bytes(@(Buff[i]));
      end;
      BuffMPX[i*ADChannelCount+t] := SmallInt(Round(Buff[i])); // bad scaling?
    end;
  end;
begin
  Trace                       := Sweep[0].Rec;
  Segment                     := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
  Segment.Channels.CanEdit    := CanEdit;   //?
  Segment.Channels.EditStream := EditStream;
  Segment.Channels.DataSize   := ADChannelCount * ADDataSize;
  Segment.Channels.DataType   := ADDataType;
  Segment.Channels.DataCount  := 0;

  BigEndian     := 0 = (Trace.DataKind and (1 shl LittleEndianBit9));
  case Trace.DataFormat of
      int16:  begin  TraceDataType := NODE_DATATYPE_SMALLINT;
                     TraceDataSize := 2;  end;
      int32:  begin  TraceDataType := NODE_DATATYPE_LONGINT;   //I fixed this mistake (was it?). LM.
                     TraceDataSize := 4;  end;
      real32: begin  TraceDataType := NODE_DATATYPE_SINGLE;
                     TraceDataSize := 4;  end;
      real64: begin  TraceDataType := NODE_DATATYPE_DOUBLE;
                     TraceDataSize := 8;  end;
  end;

  if (Sweep.Count = 1) and (   (TraceDataType = NODE_DATATYPE_SMALLINT)
                            or (TraceDataType = NODE_DATATYPE_LONGINT)  )
                       and (not BigEndian) then begin

    if Trace.DataFormat = int32 then begin
      ADDataSize     := SizeOf(LongInt);
      ADDataType     := NODE_DATATYPE_LONGINT;
    end;

    Segment.Stream              := Stream;
    Segment.Channels.Stream     := Stream;
    Segment.StreamPos           := Trace.Data;
    Segment.Channels.DataStreamPos := Trace.Data;

    Selection                   := AllocMem(SizeOf(TQUB_QFS_EditSelection));
    Selection.DataCount         := Trace.DataPoints;
    Selection.StreamPos         := Segment.Channels.DataStreamPos;
    Selection.FromEditStream    := False;

    Segment.Channels.InsertSelection(Selection, 0);

  end else begin // cast / mpx

    Segment.Stream              := EditStream;
    Segment.Channels.Stream     := EditStream;
    Segment.StreamPos           := EditStream.Position;
    Segment.Channels.DataStreamPos := EditStream.Position;

    Leaf             := 0;
    LeafStart        := 0;
    LeafSize         := Trace.InterleaveSize;
    if LeafSize = 0 then begin
      LeafSize       := Trace.DataPoints;
    end;
    SetLength(BuffRaw, TraceDataSize * LeafSize);
    SetLength(BuffMPX, Sizeof(SmallInt) * ADChannelCount * LeafSize);

    while LeafStart < Sweep[0].Rec.DataPoints do begin
      LeafSize := Min(LeafSize, Sweep[0].Rec.DataPoints - LeafStart);
    // each interleaf:
      for t := 0 to Sweep.Count - 1 do begin
        Trace := Sweep[t].Rec;
        Stream.Seek(Trace.Data, soFromBeginning);
        Stream.ReadBuffer(PChar(BuffRaw)^, TraceDataSize * LeafSize);
        case TraceDataType of
          NODE_DATATYPE_SMALLINT:  CopySmallTrace;
          NODE_DATATYPE_LONGINT:   CopyLongTrace;
          NODE_DATATYPE_SINGLE:    CopySingleTrace;
          NODE_DATATYPE_DOUBLE:    CopyDoubleTrace;
        end;
      end;
      Selection                := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount      := LeafSize;
      Selection.StreamPos      := EditStream.Position;
      Selection.FromEditStream := False;

      EditStream.WriteBuffer(PChar(BuffMpx)^, Sizeof(SmallInt)*ADChannelCount*LeafSize);
      Segment.Channels.InsertSelection(Selection, Leaf);

      LeafStart := LeafStart + LeafSize;
      Inc(Leaf);
    end;

  end;

  prSegments.AddRecord(Segment, prSegments.SegmentCount);
end;


function TQUB_QFS_QDF.prIsPUL9Dat(const FName :String): Boolean;
var
  S      :TFileStream;
  Magic  :PChar;
begin
  Magic  := nil;
  S      := nil;
  Result := False;
  try
    GetMem(Magic, 5);
    FillChar(Magic[0], 5, 0);
    S := TFileStream.Create(FName, fmOpenRead);
    S.Read(Magic[0], 4);
    Result := (String(Magic) = 'DAT2');
  finally
    FreeMem(Magic);
    S.Free;
  end;
end;


//****************************************************************************//
//                                                                            //
//                               R e a d    A S C                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadASC :Boolean;
var
  Segment    :TQUB_QFS_Segment;
  Selection  :TQUB_QFS_PEditSelection;
  PropForm   :TQUB_QFS_QDF_PropertiesForm;
  DataCount  :Integer;
  SegPos     :Integer;
  Entry      :Integer;
begin

  Result := False;

  Reset;

  //Interactive properties
  if not prOpenSilently then begin

    PropForm                         := TQUB_QFS_QDF_PropertiesForm.Create(nil);

    PropForm.Segmented               := prOpenSegmented;
    PropForm.Multiplexed             := prOpenMultiplexed;
    PropForm.MPXAA                   := prOpenMPXAA;
    PropForm.TimeColumn              := prOpenTimeColumn;
    PropForm.UnevenSampling          := prOpenUnevenSampling;
    PropForm.SinglePrec              := prOpenSinglePrec;

    PropForm.ADChannelCount          := ADChannelCount;
    PropForm.ADDataSize              := ADDataSize;
    PropForm.Sampling{milisecs}      := Sampling{secs}*1.0e3;
    PropForm.Scaling                 := Scaling;
    PropForm.SetupAscPreScan(Stream);

    PropForm.lblFileName.Caption     := FileName;

    try

      PropForm.ShowModal;

      case PropForm.ModalResult of

        idOK:
          begin
            prOpenSegmented          := PropForm.Segmented;
            prOpenMultiplexed        := PropForm.Multiplexed;
            prOpenMPXAA              := PropForm.MPXAA;
            prOpenTimeColumn         := PropForm.TimeColumn;
            prOpenUnevenSampling     := PropForm.UnevenSampling;
            prOpenSinglePrec         := PropForm.SinglePrec;
            ADChannelCount           := PropForm.ADChannelCount;
            ADDataSize               := PropForm.ADDataSize;
            Scaling                  := PropForm.Scaling;
            Sampling{secs}           := PropForm.Sampling{milisecs}*1.0e-3;
          end;

        idCancel:
          begin
            Result := True;
          end;

      end;//case ModalResult

    finally

      PropForm.Free;

    end;

  end;//Silently


  if Result then Exit;


  //Data
  if ADDataSize = 2 then ADDataType := NODE_DATATYPE_SMALLINT else ADDataType := NODE_DATATYPE_LONGINT;

  { TODO -oLorin :
It would be interesting to use Int32 instead of Int16,
or even Double/Single. If Float, a convertor must be intercalated
in Channels.LoadData. The convertor concept should be explored at large.
08-03-03 }


  //Use 32-bit integer instead of 16, for more precision
  //(presumably required for ASCII data, usually computer generated)
  //Do the same for ABF files.


  //Segments
  prSegments.Stream                := EditStream;
  prSegments.StreamPos             := 0;

  StreamPos                        := 0;

  Entry                            := Console.NewEntry('', COLOR_TEXT);

  repeat

    try

      SegPos := prEditStream.Position;

      if prOpenMultiplexed then begin
        Result := prConvertASCIIMultiplexed(Stream{Source}, ADChannelCount, prOpenTimeColumn, DataCount, Segments.SegmentCount, prOpenMPXAA, Entry);
      end else begin
        Result := prConvertASCII(Stream{Source}, ADChannelCount, prOpenTimeColumn, prOpenUnevenSampling, prOpenSegmented, DataCount, Segments.SegmentCount, Entry);
      end;

      if Result then Break;

      if DataCount > 0 then begin

        //Segment
        Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
        Segment.Stream                   := EditStream;
        Segment.StreamPos                := SegPos;

        Segment.StartTime                := prSegments.TotalDataCount * Sampling;// - prSegments.SegmentCount * Sampling;

        //Channels
        Segment.Channels.CanEdit         := CanEdit;
        Segment.Channels.EditStream      := EditStream;
        Segment.Channels.Stream          := EditStream;
        Segment.Channels.StreamPos       := SegPos;
        Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
        Segment.Channels.DataType        := ADDataType;
        Segment.Channels.DataCount       := 0;
        Segment.Channels.DataStreamPos   := SegPos;

        Selection                        := AllocMem(SizeOf(TQUB_QFS_EditSelection));
        Selection.DataCount              := DataCount;
        Selection.StreamPos              := Segment.Channels.DataStreamPos;
        Selection.FromEditStream         := False;
        Segment.Channels.InsertSelection(Selection, 0);

        prSegments.AddRecord(Segment, prSegments.SegmentCount);

      end;

    except

      Result := True;

    end;

  until((DataCount < 0) or (Result));

end;

{
Reads ASCII data organized as:
t0, A1_0, A2_0, ..., B1_0, B2_0,...
t1, A1_1, A2_1, ..., B1_1, B2_1,...
...

The time column is optional. These files are producd by, e.g., Origin, from pClamp.

For the time being, writing to binary is non-buffered (presumably these files are small)
}
function TQUB_QFS_QDF.prConvertASCIIMultiplexed(Stream :TStream; ChannelCount :Integer; TimeColumn :Boolean; var DataCount :Integer; Segment :Integer; MPXAA :Boolean; Entry :Integer) :Boolean;
var
  Parser       :TQUB_Parser;
  Text         :String;
  Sign         :Integer;
  OldPos       :LongInt;
  TimeCnt      :Integer;
  ColCnt       :Integer;
  ColIdx       :Integer;
  DatIdx       :Integer;
  SegCnt       :Integer;
  Buffer       :Pointer;

  function GetTextChunk(var Text :String) :Boolean;
  var
    o, p, r :Integer;
  begin

    Result := False;
    SetLength(Text, ONE_MEGABYTE);

    o := Stream.Position;
    r := Stream.Read(PChar(Text)^, ONE_MEGABYTE);
    if r = 0 then begin
      Result := True;
      Text := '';
      Exit;
    end;

    if Length(Text) <> r then SetLength(Text, r);

    if (o + r) >= Stream.Size then begin
      p := r;
    end else begin
      p := LastDelimiter(' ,'#9#13#10, Text);
      if p < 1 then p := r;
      if p <> r then SetLength(Text, p);
    end;

    Stream.Position := o + p;

  end;

  function GetColCount :Integer;
  begin
    Result := 0;
    repeat

      GetTextChunk(Text);

      Parser.Start;
      repeat

        Parser.Next(Text);

        if Parser.Token.TokenType = TOKEN_NEWLINE then begin
          if Result > 0 then Break;//in case there are comment lines
        end else if Parser.Token.TokenType = TOKEN_NULL then begin
          Break;
        end else if Parser.Token.TokenType = TOKEN_NUMBER then begin
          Inc(Result);
        end;

      until False;

    until((Text = '') or (Parser.Token.TokenType = TOKEN_NEWLINE) or (Parser.Token.TokenType = TOKEN_NULL));
  end;

  procedure WriteValueToBuffer;
  var
    Value :Extended;
    i, s  :Integer;
  begin

    //if Time, just ignore it
    if (ColIdx = 0) and (TimeCnt = 1) then Exit;

    //if too many channels, or in the wrong segment, ignore it
    if MPXAA then begin
      s := (ColIdx - TimeCnt) mod SegCnt;//ChannelCount;
    end else begin
      s := (ColIdx - TimeCnt) div ChannelCount;
    end;
    if s <> Segment then Exit;
    if MPXAA then begin
      i := (ColIdx - TimeCnt) div SegCnt;
    end else begin
      i := (ColIdx - TimeCnt) - (s*ChannelCount);
    end;
    if (i < 0) or (i >= ChannelCount) then Exit;

    DatIdx := i + 1;
    Value := Sign * Parser.Token.TokenNumber * Scaling;

    case ADDataType of

      NODE_DATATYPE_BYTE:
        if Value > High(Byte)        then TPAByte(Buffer)[i]     := High(Byte)        else if Value < Low(Byte)       then TPAByte(Buffer)[i]      := Low(Byte)        else TPAByte(Buffer)[i]      := Round(Value);
      NODE_DATATYPE_SHORTINT:
        if Value > High(ShortInt)    then TPAShort(Buffer)[i]    := High(ShortInt)    else if Value < Low(ShortInt)   then TPAShort(Buffer)[i]     := Low(ShortInt)    else TPAShort(Buffer)[i]     := Round(Value);
      NODE_DATATYPE_WORD:
        if Value > High(Word)        then TPAWord(Buffer)[i]     := High(Word)        else if Value < Low(Word)       then TPAWord(Buffer)[i]      := Low(Word)        else TPAWord(Buffer)[i]      := Round(Value);
      NODE_DATATYPE_SMALLINT:
        if Value > High(SmallInt)    then TPASmall(Buffer)[i]    := High(SmallInt)    else if Value < Low(SmallInt)   then TPASmall(Buffer)[i]     := Low(SmallInt)    else TPASmall(Buffer)[i]     := Round(Value);
      NODE_DATATYPE_LONGWORD:
        if Value > High(LongWord)    then TPALongWord(Buffer)[i] := High(LongWord)    else if Value < Low(LongWord)   then TPALongWord(Buffer)[i]  := Low(LongWord)    else TPALongWord(Buffer)[i]  := Round(Value);
      NODE_DATATYPE_LONGINT:
        if Value > High(LongInt)     then TPALongInt(Buffer)[i]  := High(LongInt)     else if Value < Low(LongInt)    then TPALongInt(Buffer)[i]   := Low(LongInt)     else TPALongInt(Buffer)[i]   := Round(Value);
      NODE_DATATYPE_UINT64:
        if Value > High(UInt64)      then TPAUInt64(Buffer)[i]   := High(UInt64)      else if Value < Low(UInt64)     then TPAUInt64(Buffer)[i]    := Low(UInt64)      else TPAUInt64(Buffer)[i]    := Round(Value);
      NODE_DATATYPE_INT64:
        if Value > High(Int64)       then TPAInt64(Buffer)[i]    := High(Int64)       else if Value < Low(Int64)      then TPAInt64(Buffer)[i]     := Low(Int64)       else TPAInt64(Buffer)[i]     := Round(Value);
      NODE_DATATYPE_SINGLE:
        TPASingle(Buffer)[i]         := Value;
      NODE_DATATYPE_DOUBLE:
        TPADouble(Buffer)[i]         := Value;
      NODE_DATATYPE_EXTENDED:
        TPAExtended(Buffer)[i]       := Value;
    end;

  end;

  procedure WriteBufferToStream;
  begin
    //Write buffer
    EditStream.WriteBuffer(Buffer^, ADDataSize * ChannelCount);
    //Clean buffer
    FillChar(Buffer^, ADDataSize * ChannelCount, 0);
    //update DataCount
    if DataCount = -1 then DataCount := 0;
    Inc(DataCount);
  end;

begin

  Result := False;

  DataCount := -1;

  Parser := nil;

  Buffer := nil;

  //Save position in Stream, to restore it later
  OldPos := Stream.Position;

  try

    Parser               := TQUB_Parser.Create;
    Parser.ReturnTab     := False;
    Parser.ReturnNewLine := True;

    //First, check how many columns
    ColCnt := GetColCount;

    //Restore position
    Stream.Position := OldPos;

    //Calculate Segment count
    if TimeColumn then TimeCnt := 1 else TimeCnt := 0;
    SegCnt := (ColCnt - TimeCnt) div ChannelCount;

    //Done?
    if Segment >= SegCnt then Exit;

    Buffer := AllocMem(ChannelCount * ADDataSize);

    //Now read actual segment

    ColIdx := 0;
    DatIdx := 0;
    Sign   := 1;

    Text   := '';

    Parser.Start;
    repeat

      Parser.Next(Text);

      if Parser.Token.TokenType = TOKEN_NEWLINE then begin

        //Write the data, if DatIdx > 0;
        if DatIdx > 0 then begin
          WriteBufferToStream;
        end;

        //Re-initialize ColIdx
        ColIdx := 0;
        DatIdx := 0;
        Sign   := 1;

      end else if Parser.Token.TokenType = TOKEN_NULL then begin

        GetTextChunk(Text);
        Parser.Start;

      end else if Parser.Token.TokenType = TOKEN_SEPARATOR then begin

        if          Parser.Token.Token = '-' then begin
          Sign := -Sign;
        end else if Parser.Token.Token = '+' then begin
          Sign := Sign;
        end;

      end else if Parser.Token.TokenType = TOKEN_NUMBER then begin

        //Read number
        WriteValueToBuffer;

        Sign := 1;
        Inc(ColIdx);

      end;

    until(Text = '');

    //Write the data, if DatIdx > 0;
    if DatIdx > 0 then begin
      WriteBufferToStream;
    end;


  finally

    FreeMem(Buffer);
    Parser.Free;
    Stream.Position := OldPos;

  end;

end;

//12-12-05 LM
//I modified this procedure to accept unevenly sampled data files.
//It resamples the data at even dt, without filtering.
//Not thoroughly checked, but it seems it works.

//May 2006 CN
//Various problems reading files so I studied this code:
// Fixed bug at 1MB of text by moving Channel initialization out of loop.
// Noticed that newlines are ignored except for segmentation, so a missing number
// will shift all the channels off by one.
// Noticed that macintosh-style newlines (bare CR#13)
// and Unix-style (bare LF#10) are parsed as TOKEN_BAD_CHAR and ignored.

function TQUB_QFS_QDF.prConvertASCII(Stream :TStream; ChannelCount :Integer; TimeColumn, UnevenSampling, Segmented :Boolean; var DataCount :Integer; Segment :Integer; Entry :Integer) :Boolean;
var
  Parser       :TQUB_Parser;
  Text         :String;
  Channel      :Integer;
  Buffer       :Pointer;//TPASmall;
  BufferCount  :Integer;
  BufferRest   :Integer;
  Number       :Integer;
  Sign         :Integer;
  p, r, s, i   :Integer;
  TotalCount   :Cardinal;
  OldPos       :LongInt;
  SegDone      :Boolean;
  NewLineCount :Integer;
  ElemSize     :Integer;
  Values       :TPMDouble;
  ValIdx       :Integer;
  dCount       :Integer;
  PointTime    :Double;
  Percent      :Integer;
  NewPercent   :Integer;

  procedure WriteBuffer;
  begin
    if ElemSize = SizeOf(SmallInt) then begin
      if Number > High(SmallInt) then begin
        TPASmallInt(Buffer)[BufferCount] := High(SmallInt);
      end else if Number < Low(SmallInt) then begin
        TPASmallInt(Buffer)[BufferCount] := Low(SmallInt);
      end else begin
        TPASmallInt(Buffer)[BufferCount] := Number;
      end;
    end else if ElemSize = SizeOf(Integer) then begin
      TPAInteger(Buffer)[BufferCount] := Number;
    end;
  end;

begin

  Result       := False;

  SegDone      := False;
  NewLineCount := 0;
  DataCount    := 0;
  dCount       := DataCount;

  Parser               := TQUB_Parser.Create;
  Parser.ReturnTab     := False;
  Parser.ReturnNewLine := True;

  Buffer       := nil;
  Values       := nil;
  ValIdx       := 0;

  TotalCount   := Stream.Size;
  if TotalCount = 0 then TotalCount := 1;

  Percent := Cardinal(Round(100.0 * (Stream.Position + 1) / TotalCount));
  Console.SetEntry(Entry, Format('Converting ASCII to binary for segment %d ... %d%%', [Segment + 1, Percent]), COLOR_TEXT);

  try

    ElemSize := ADDataSize;
    if ElemSize = SizeOf(SmallInt) then begin
      Buffer := SmAlloc1(ONE_MEGABYTE);
    end else if ADDataSize = SizeOf(Integer) then begin
      Buffer := IAlloc1(ONE_MEGABYTE);
    end;

    //Where to store the First/Last values for interpolation, when UnevenSampling
    Values := DAlloc2(ADChannelCount + 1, 2);

    s := Stream.Position;
    if s >= Stream.Size then begin
      DataCount := -1;
      Exit;
    end;

    BufferRest := 0;

    Percent    := -1;

    //TO DO: Chris moved it here - Check! LM.
    if TimeColumn then Channel := 0 else Channel := 1;

    repeat


      OldPos := Stream.Position;

      SetLength(Text, ONE_MEGABYTE);

      r := Stream.Read(PChar(Text)^, ONE_MEGABYTE);
      if r = 0 then Break;

      if Length(Text) <> r then SetLength(Text, r);

      if (OldPos + r) >= Stream.Size then begin

        p := r;

      end else begin

        p := LastDelimiter(' ,'#9#13#10, Text);
        if p < 1 then p := r;

        if p <> r then SetLength(Text, p);

      end;

      Inc(s, p);

      Stream.Position := s;

      Parser.Start;

      BufferCount := 0;

      //TO DO: See above. LM
      //if TimeColumn then Channel := 0 else Channel := 1;

      Sign := 1;

      repeat

        Parser.Next(Text);

        if Parser.Token.TokenType <> TOKEN_NEWLINE then NewLineCount := 0;
        //Reset NewLineCount

        if Parser.Token.TokenType = TOKEN_NEWLINE then begin

          if BufferCount > 0 then begin
            Inc(NewLineCount);
            if (NewLineCount > 1) and (Segmented) then begin
            //End of segment
              Stream.Position := OldPos + Parser.Token.TokenPos + Parser.Token.TokenLen - 1{!!!};
              SegDone := True;
              Break;
            end;
          end;

        end else if Parser.Token.TokenType = TOKEN_SEPARATOR then begin

          if Parser.Token.Token = '-' then begin
            Sign := -1;
          end else begin
            Sign := 1;
          end;

        end else if Parser.Token.TokenType = TOKEN_NUMBER then begin

          if Channel = 0 then begin
          //Time

            if UnevenSampling then begin
              try
                Values[Channel, ValIdx] := {Sign * }Parser.Token.TokenNumber;//Can the time be negative?!
              except
                Values[Channel, ValIdx] := 0;
              end;
            end;
            Sign    := 1;
            Channel := 1;

          end else if Channel > 0 then begin
          //Data

            try
              Number                  := Sign * Parser.Token.TokenNumber * Scaling;
              Values[Channel, ValIdx] := Sign * Parser.Token.TokenNumber;
            except
              Number                  := 0;
              Values[Channel, ValIdx] := 0;
            end;

            if not UnevenSampling then begin
            //Even Sampling, nothing special
              WriteBuffer;
              Inc(BufferCount);
            end;

            if Channel = ChannelCount then begin
            //All channels complete

              if UnevenSampling then begin
              //UnevenSampling

                if ValIdx = 0 then begin

                  ValIdx := 1;

                end else begin
                //Both sides acquired, now do interpolation

                  //Interpolation
                  while PointTime <= Values[0, 1] do begin
                    for i := 1 to ADChannelCount do begin
                      //Interpolated Value
                      Number := Round((Values[i, 0]{v1} + (Values[i, 1]{v2} - Values[i, 0]{v1})*(PointTime - Values[0, 0{t1}])/(Values[0, 1]{t2} - Values[0, 0]{t1})) * Scaling);
                      WriteBuffer;
                      Inc(BufferCount);
                    end;
                    Inc(dCount);
                    PointTime := dCount * Sampling;
                  end;

                  //Cycle Values
                  for i := 0 to ADChannelCount do begin
                    Values[i, 0] := Values[i, 1];
                  end;

                  ValIdx := 0;
                end;//ValIdx

              end;//UnevenSampling

              if TimeColumn then begin
                Channel := 0;
              end else begin
                Channel := 1;
              end;

            end else begin

              Inc(Channel);

            end;

            Sign := 1;

          end else begin

            Sign := 1;

          end;

        end;

      until((SegDone) or (Parser.Token.TokenType = TOKEN_NULL));

      NewPercent := Cardinal(Round(100.0 * (Stream.Position + 1) / TotalCount));
      if NewPercent <> Percent then begin
        Console.SetEntry(Entry, Format('Converting ASCII to binary for segment %d ... %d%%', [Segment + 1, NewPercent]), COLOR_TEXT);
      end;
      Percent := NewPercent;

      Application.ProcessMessages;

      if BufferCount > 0 then begin
        try
          EditStream.WriteBuffer(Buffer^, ElemSize{SizeOf(SmallInt)} * BufferCount);
          Sleep(5); // ?
          Inc(DataCount, (BufferCount + BufferRest) div ChannelCount);
          dCount := DataCount;
          BufferRest := (BufferCount + BufferRest) mod ChannelCount;
        except
          Result := True;
          Exit;
        end;
      end;

    until(SegDone or (r < ONE_MEGABYTE));


  finally


    Console.SetEntry(Entry, 'Converting ASCII to binary ... Done.', COLOR_GRAY);

    FreeMem(Buffer);
    Free2(TPMChar(Values));
    Parser.Free;

  end;

end;

//12-13-05 LM
//It converts from Single into native format
//It accepts uneven sampling, but not finished
function TQUB_QFS_QDF.prConvertSingle(Stream :TStream; ChannelCount :Integer; TimeColumn, UnevenSampling :Boolean; var DataCount :Integer; Segment :Integer; Entry :Integer) :Boolean;
const
  BUFFER_SIZE  = ONE_MEGABYTE div 10;
var
  IBuffer      :Pointer;
  OBuffer      :Pointer;
  IBufferCount :Integer;
  //OBufferCount :Integer;
  IChCount     :Integer;
  IIndex       :Integer;
  OIndex       :Integer;
  Number       :Integer;
  c, p, r      :Integer;
  TotalCount   :Cardinal;
  ElemSize     :Integer;
  Values       :TPMDouble;
  //ValIdx       :Integer;
  //PointTime    :Double;
  Percent      :Integer;
  NewPercent   :Integer;
  Count        :Integer;
  
  procedure WriteBuffer;
  begin
    if ElemSize = SizeOf(SmallInt) then begin
      if Number > High(SmallInt) then begin
        TPASmallInt(OBuffer)[OIndex] := High(SmallInt);
      end else if Number < Low(SmallInt) then begin
        TPASmallInt(OBuffer)[OIndex] := Low(SmallInt);
      end else begin
        TPASmallInt(OBuffer)[OIndex] := Number;
      end;
    end else if ElemSize = SizeOf(Integer) then begin
      TPAInteger(OBuffer)[OIndex] := Number;
    end;
  end;

begin

  Result       := False;

  DataCount    := 0;

  IBuffer      := nil;
  OBuffer      := nil;
  Values       := nil;
  //never used ValIdx       := 0;

  TotalCount   := Stream.Size;
  if TotalCount = 0 then TotalCount := 1;

  Percent := Cardinal(Round(100.0 * (Stream.Position + 1) / TotalCount));
  Console.SetEntry(Entry, Format('Converting binary data from Single to Integer for segment %d ... %d%%', [Segment + 1, Percent]), COLOR_TEXT);

  try

    ElemSize := ADDataSize;
    if ElemSize = SizeOf(SmallInt) then begin
      OBuffer := SmAlloc1(ADChannelCount * BUFFER_SIZE);
    end else if ADDataSize = SizeOf(Integer) then begin
      OBuffer := IAlloc1(ADChannelCount * BUFFER_SIZE);
    end;
    if TimeColumn then begin
      IChCount := ADChannelCount + 1;
    end else begin
      IChCount := ADChannelCount;
    end;
    IBufferCount := IChCount * BUFFER_SIZE;
    IBuffer := SAlloc1(IBufferCount);

    //Where to store the First/Last values for interpolation, when UnevenSampling
    Values := DAlloc2(ADChannelCount + 1, 2);

    if Stream.Position >= Stream.Size then begin
      DataCount := -1;
      Exit;
    end;

    Percent    := -1;

    repeat
    //Buffer

      r := Stream.Read(IBuffer^, IBufferCount * SizeOf(Single)) div SizeOf(Single);
      if r = 0 then Break;

      IIndex := 0;
      OIndex := 0;
      try
        Count  := Floor(r / IChCount);
      except
        try
          Count := r div IChCount;
        except
        end;
      end;
      for p := 0 to Count - 1 do begin
        if TimeColumn then begin
          Inc(IIndex);
          //Not used yet - only for UnevenSampling

          //For UnevenSampling, watch for OIndex to be < OBufferCount

        end;
        for c := 0 to ADChannelCount - 1 do begin
          try
            Number := Round(TPASingle(IBuffer)[IIndex] * Scaling);
          except
          end;
          WriteBuffer;
          Inc(IIndex);
          Inc(OIndex);
        end;//c
      end;
    
      NewPercent := Cardinal(Round(100.0 * (Stream.Position + 1) / TotalCount));
      if NewPercent <> Percent then begin
        Console.SetEntry(Entry, Format('Converting binary data from Single to Integer for segment %d ... %d%%', [Segment + 1, NewPercent]), COLOR_TEXT);
      end;
      Percent := NewPercent;

      Application.ProcessMessages;

      if OIndex > 0 then begin
        try
          EditStream.WriteBuffer(OBuffer^, ElemSize * OIndex);
          Sleep(5); // ?
          Inc(DataCount, OIndex div ChannelCount);
        except
          Result := True;
          Exit;
        end;
      end;

    until(r < IBufferCount);


  finally


    FreeMem(IBuffer);
    FreeMem(OBuffer);
    Free2(TPMChar(Values));

  end;

end;

//Converts Integer to native - use only when data have TimeColumn
function TQUB_QFS_QDF.prConvertInteger(Stream :TStream; ChannelCount :Integer; TimeColumn, UnevenSampling :Boolean; var DataCount :Integer; Segment :Integer; Entry :Integer) :Boolean;
const
  BUFFER_SIZE  = ONE_MEGABYTE div 10;
var
  IBuffer      :Pointer;
  OBuffer      :Pointer;
  IBufferCount :Integer;
  //OBufferCount :Integer;
  IChCount     :Integer;
  IIndex       :Integer;
  OIndex       :Integer;
  Number       :Integer;
  c, p, r      :Integer;
  TotalCount   :Cardinal;
  ElemSize     :Integer;
  Values       :TPMDouble;
  //ValIdx       :Integer;
  //PointTime    :Double;
  Percent      :Integer;
  NewPercent   :Integer;
  Count        :Integer;

  procedure WriteBuffer;
  begin
    if ElemSize = SizeOf(SmallInt) then begin
      if Number > High(SmallInt) then begin
        TPASmallInt(OBuffer)[OIndex] := High(SmallInt);
      end else if Number < Low(SmallInt) then begin
        TPASmallInt(OBuffer)[OIndex] := Low(SmallInt);
      end else begin
        TPASmallInt(OBuffer)[OIndex] := Number;
      end;
    end else if ElemSize = SizeOf(Integer) then begin
      TPAInteger(OBuffer)[OIndex] := Number;
    end;
  end;

begin

  Result       := False;

  DataCount    := 0;

  IBuffer      := nil;
  OBuffer      := nil;
  Values       := nil;
  //never used ValIdx       := 0;

  TotalCount   := Stream.Size;
  if TotalCount = 0 then TotalCount := 1;

  Percent := Cardinal(Round(100.0 * (Stream.Position + 1) / TotalCount));
  Console.SetEntry(Entry, Format('Converting binary data from Integer to Integer for segment %d ... %d%%', [Segment + 1, Percent]), COLOR_TEXT);

  try

    ElemSize := ADDataSize;
    if ElemSize = SizeOf(SmallInt) then begin
      OBuffer := SmAlloc1(ADChannelCount * BUFFER_SIZE);
    end else if ADDataSize = SizeOf(Integer) then begin
      OBuffer := IAlloc1(ADChannelCount * BUFFER_SIZE);
    end;
    if TimeColumn then begin
      IChCount := ADChannelCount + 1;
    end else begin
      IChCount := ADChannelCount;
    end;
    IBufferCount := IChCount * BUFFER_SIZE;
    IBuffer := SAlloc1(IBufferCount);

    //Where to store the First/Last values for interpolation, when UnevenSampling
    Values := DAlloc2(ADChannelCount + 1, 2);

    if Stream.Position >= Stream.Size then begin
      DataCount := -1;
      Exit;
    end;

    Percent    := -1;

    repeat
    //Buffer

      r := Stream.Read(IBuffer^, IBufferCount * ElemSize) div ElemSize;
      if r = 0 then Break;

      IIndex := 0;
      OIndex := 0;
      try
        Count  := Floor(r / IChCount);
      except
        try
          Count := r div IChCount;
        except
        end;
      end;
      for p := 0 to Count - 1 do begin
        if TimeColumn then begin
          Inc(IIndex);
          //Not used yet - only for UnevenSampling

          //For UnevenSampling, watch for OIndex to be < OBufferCount

        end;
        for c := 0 to ADChannelCount - 1 do begin
          try
            if ElemSize = SizeOf(SmallInt) then begin
              Number := TPASmallInt(IBuffer)[IIndex];
            end else begin
              Number := TPAInteger(IBuffer)[IIndex];
            end;
          except
          end;
          WriteBuffer;
          Inc(IIndex);
          Inc(OIndex);
        end;//c
      end;

      NewPercent := Cardinal(Round(100.0 * (Stream.Position + 1) / TotalCount));
      if NewPercent <> Percent then begin
        Console.SetEntry(Entry, Format('Converting binary data from Integer to Integer for segment %d ... %d%%', [Segment + 1, NewPercent]), COLOR_TEXT);
      end;
      Percent := NewPercent;

      Application.ProcessMessages;

      if OIndex > 0 then begin
        try
          EditStream.WriteBuffer(OBuffer^, ElemSize * OIndex);
          Sleep(5);
          Inc(DataCount, OIndex div ChannelCount);
        except
          Result := True;
          Exit;
        end;
      end;

    until(r < IBufferCount);


  finally

    FreeMem(IBuffer);
    FreeMem(OBuffer);
    Free2(TPMChar(Values));

  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    A B F                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadABF :Boolean;
var
  d, f        :String;
  QdfStream   :TFileStream;
  SessionQFS            :TQUB_QFS_Record;
  SessionQDF            :INode;
  OffNode               :INode;
begin

  try
    d := ExtractFilePath(Application.ExeName) + 'Temp\';
    if not DirectoryExists(d) then begin
      if not CreateDirectory(PChar(d), nil) then begin
        d := ExtractFilePath(prFileName);
      end;
    end;
  except
    d := ExtractFilePath(prFileName);
  end;

  f := d + ChangeFileExt(ExtractFileName(prFileName), '.qdf');
  Result := 0 <> Convert_ABF_To_QDF(PChar(FileName), PChar(f), @DoAxoMsgOut, Pointer(Self));
  if Result then Exit;

  try
    QdfStream := TFileStream.Create(f, fmOpenRead or fmShareDenyWrite);
    Stream.Free;
    Stream := QdfStream;
    Result := prReadQDF;

    // fix dc. scaling/offset corruption from early abf2 reader
    SessionQFS            := RecordsByName['SessionFile'];
    if SessionQFS = nil then Exit;
    SessionQDF            := SessionQFS.CloneAsNode(True, TNode.CreateNull);
    OffNode := Session.Find('DataChannels').Find('Channel').Find('Offset');
    if (OffNode.DataCount > 0) and (OffNode.Data[0] <> 0.0) then begin
      Session.RemoveChild(Session['DataChannels']);
      Session.AppendClone(SessionQDF['DataChannels'], True);
    end;

  except
    Result := True;
  end;

end;

//****************************************************************************//
//                                                                            //
//                               R e a d    A C Q                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadACQ :Boolean;
var
  d, f        :String;
  QdfStream   :TFileStream;
begin

  try
    d := ExtractFilePath(Application.ExeName) + 'Temp\';
    if not DirectoryExists(d) then begin
      if not CreateDirectory(PChar(d), nil) then begin
        d := ExtractFilePath(prFileName);
      end;
    end;
  except
    d := ExtractFilePath(prFileName);
  end;

  f := d + ChangeFileExt(ExtractFileName(prFileName), '.qdf');
  Result := 0 <> Convert_Acquire_To_QDF(PChar(FileName), PChar(f), @DoAxoMsgOut, Pointer(Self));
  if Result then Exit;

  try
    QdfStream := TFileStream.Create(f, fmOpenRead or fmShareDenyWrite);
    Stream.Free;
    Stream := QdfStream;
    Result := prReadQDF;
  except
    Result := True;
  end;

end;

//****************************************************************************//
//                                                                            //
//                               R e a d    W A V                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadWAV :Boolean;
var
  WaveChunk  :TQUB_QFS_QDF_WAV_Chunk;
  Segment    :TQUB_QFS_Segment;
  Selection  :TQUB_QFS_PEditSelection;
  p          :Integer;
  DataCount  :Integer;
begin

  Result     := False;

  Reset;

  try

    StreamPos := Stream.Position;

    //Read file  H E A D E R ***************************************************
    Stream.ReadBuffer(WaveChunk.Riff, SizeOf(WaveChunk.Riff));
    if (WaveChunk.riff.riff <> QUB_QFS_QDF_STRING_RIFF) or
       (WaveChunk.riff.wave <> QUB_QFS_QDF_STRING_WAVE)
    then begin
      Result := True;
      Console.NewEntry('Wrong file format', COLOR_ERROR);
      Exit;
    end;

    p := Stream.Position;
    Stream.ReadBuffer(WaveChunk.Format, SizeOf(WaveChunk.Format));
    if ( WaveChunk.format.fmt <> QUB_QFS_QDF_STRING_FMT)                   or
       ((WaveChunk.format.bytespersample / WaveChunk.format.channels) > 4) or
       ( WaveChunk.format.formattag <> 1)
    then begin
      Result := True;
      Console.NewEntry('Data format currently not supported', COLOR_ERROR);
      Exit;
    end;

    p := p + 4{Tag} + 4{Length} + WaveChunk.Format.Length;
    if Odd(WaveChunk.Format.Length) then Inc(p);

    repeat

      //prStream.Position := p;
      Stream.Seek(p, soFromBeginning);
      try
        Stream.ReadBuffer(WaveChunk.Data, SizeOf(WaveChunk.Data));
        if WaveChunk.Data.Data = QUB_QFS_QDF_STRING_DATA then begin
          Break;
        end else begin
          p := p + 4 + 4 + WaveChunk.Data.Length;
          if Odd(WaveChunk.Data.Length) then Inc(p);
        end;
      except
        Result := True;
      end;

    until((Result));

    if (Result) or (WaveChunk.Data.Data <> QUB_QFS_QDF_STRING_DATA) then begin
      Result := True;
      Exit;
    end;

    Stream.Seek(p + 4 + 4, soFromBeginning);

    //Data
    Sampling       := 1 / WaveChunk.format.samplerate;
    //Scaling is defaulted...
    //ADChannels
    ADChannelCount := WaveChunk.format.channels;
    ADDataSize     := WaveChunk.format.bytespersample div WaveChunk.format.channels;
    if WaveChunk.format.bitspersample <= 8 then begin
      ADDataType   := NODE_DATATYPE_BYTE;
    end else if WaveChunk.format.bitspersample <= 16 then begin
      ADDataType   := NODE_DATATYPE_SMALLINT;
    end else begin
      ADDataType   := NODE_DATATYPE_LONGINT;
    end;


    //Read file  D A T A *******************************************************


    //Segments
    prSegments.Stream                := Stream;
    prSegments.StreamPos             := Stream.Position;


    //LDT Legacy
    prLDTLeader   := 0;
    prLDTSampling := 0;
    prLDTScaling  := 0;

    //WAV settings
    prWAVBitsPerSample               := WaveChunk.format.BitsPerSample;

    //Segment
    Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
    Segment.Stream                   := Stream;
    Segment.StreamPos                := Stream.Position;

    //Channels
    Segment.Channels.CanEdit         := CanEdit;
    Segment.Channels.EditStream      := EditStream;
    Segment.Channels.Stream          := Stream;
    Segment.Channels.StreamPos       := Stream.Position;
    Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
    Segment.Channels.DataType        := ADDataType;
    Segment.Channels.DataCount       := 0;
    Segment.Channels.DataStreamPos   := Stream.Position;

    DataCount := WaveChunk.data.length div (WaveChunk.format.bytespersample);
    if DataCount <> 0 then begin
      Selection                      := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount            := DataCount;
      Selection.StreamPos            := Segment.Channels.DataStreamPos;
      Selection.FromEditStream       := False;
      Segment.Channels.InsertSelection(Selection, 0);
    end;

    prSegments.AddRecord(Segment, prSegments.SegmentCount);

  except

    Result := True;

  end;
end;


//****************************************************************************//
//                                                                            //
//                               R e a d    E v e n t s                       //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadEvents(DWT :INode) :Boolean;
var
  SegDWT          :TQUB_QFS_DWT_Segment;
  i               :Integer;
  Seg             :INode;
  SampMS          :Double;
  At              :Integer;
  Firsts, Lasts   :TPAInteger;
  Durs            :TPASingle;
  DwellCount      :Integer;

begin

  Reset;

  try

    Session.RemoveChild( Session.Find('Idealization') );

    StreamPos                        := Stream.Position;

    //Segments
    prSegments.Stream                := Stream;
    prSegments.StreamPos             := Stream.Position;

    prNewIdealization;

    Result     := DWT.Find('Segment').IsNull;
    if Result then Exit;

    //Sampling/Scaling
    SampMS         := Double(DWT['sampling'].Data[0]);
    Sampling       := SampMS * 1.0e-3;
    //Scaling        :=
    //ADChannels
    ADChannelCount := 1;
    ADDataSize     := SizeOf(SmallInt);
    ADDataType     := NODE_DATATYPE_SMALLINT;


    At             := 0;
    Seg            := DWT.Find('Segment');
    while not Seg.IsNull do begin
      DwellCount   := Seg['DwellCount'].Data[0];
      Durs         := TPASingle( Seg['Durations'].DataPtr );
      Seg['Firsts'].SetupNumData( NODE_DATATYPE_LONGINT, DwellCount, 1 );
      Firsts       := TPAInteger( Seg['Firsts'].DataPtr );
      Seg['Lasts'].SetupNumData( NODE_DATATYPE_LONGINT, DwellCount, 1 );
      Lasts        := TPAInteger( Seg['Lasts'].DataPtr );

      for i := 0 to DwellCount - 1 do begin
        Firsts[i]  := At;
        if Durs[i] <= 0.0 then Durs[i] := SampMS; // protect against negative time
        Inc(At, Round(Durs[i] / SampMS));
        Lasts[i]   := At-1;
      end;

      Seg.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
      Seg.Data[0]  := Firsts[0];
      Seg.Data[1]  := Lasts[ DwellCount-1 ];

      Seg          := Seg.SiblingSameName;
    end;

    Seg   := DWT.Find('Segment');
    while not Seg.IsNull do begin
      //Segment
      SegDWT                          := TQUB_QFS_DWT_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      SegDWT.Stream                   := Stream;
      SegDWT.StreamPos                := Stream.Position;
      SegDWT.StartTime                := Double(Seg['start'].Data[0]) * 1.0e-3;

      //Channels
      SegDWT.Channels.CanEdit         := False;
      SegDWT.Channels.EditStream      := nil;
      SegDWT.Channels.Stream          := Stream;
      SegDWT.Channels.StreamPos       := Stream.Position;
      SegDWT.Channels.DataSize        := ADChannelCount * ADDataSize;
      SegDWT.Channels.DataType        := ADDataType;
      SegDWT.Channels.DataCount       := Seg.Data[1] - Seg.Data[0] + 1;
      SegDWT.Channels.DataStreamPos   := Stream.Position;

      TQUB_QFS_DWT_Channels(SegDWT.Channels).Scaling   := Scaling;
        // assign (self) instead ? - scaling is A/D only, not updated when changed in file ?
        
      TQUB_QFS_DWT_Channels(SegDWT.Channels).Idl       := Idealization[0];
      TQUB_QFS_DWT_Channels(SegDWT.Channels).Offset    := Seg.Data[0];

      prSegments.AddRecord(SegDWT, prSegments.SegmentCount);
      Idealization[0].AddSeg(Seg.Data[0], Seg.Data[1]);

      Seg      := Seg.SiblingSameName;

    end;

    Idealization[0].SetDwells(DWT);

  except

    Result := True;

  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    D W T                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadDWT :Boolean;
var
  TBLName         :String;
  TBL             :INode;
  DataSet         :INode;
  TBLSeg          :INode;
  DWT, Seg        :INode;
  SampMS          :Double;
begin

  try

    DWT        := QUB_Idl_ReadDWT(FileName);
    Result     := prReadEvents(DWT);
    if Result then Exit;

    TBL                               := ReadTBLAsNode( ChangeFileExt(FileName, '.tbl') );
    TBL['Date'].SetData( Format('%s %s',[DateTimeToStr( Date ),TimeToStr( Sysutils.Time )] ) );
    TBL['ResultType'].SetData( 'SKM' );
    DataSet                           := TBL['DataSet'];
    DataSet['FileName'].SetData( FileName );

    //Sampling/Scaling
    SampMS         := Double(DWT['sampling'].Data[0]);
    Sampling       := SampMS * 1.0e-3;
    DataSet['sampling'].SetData( Sampling * 1.0e3 );

    TBLSeg        := DataSet.Find('Segment');
    Seg           := DWT.Find('Segment');
    while not Seg.IsNull do begin

      if not TBLSeg.IsNull then begin
        TBLSeg.SetupNumData( NODE_DATATYPE_LONGINT, 2, 1 );
        TBLSeg.Data[0] := Seg.Data[0];
        TBLSeg.Data[1] := Seg.Data[1];

        TBLSeg := TBLSeg.SiblingSameName;
      end;

      Seg      := Seg.SiblingSameName;

    end;

    TBLName                           := 'SKM ' + ChangeFileExt(ExtractFileName(FileName), '');
    TBL.Name                          := TBLName;
    ViewportResults.AddResults(TBL, True);

  except

    Result := True;

  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    E V L                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadEVL :Boolean;
var
  DWT        :INode;
begin

  try

    DWT        := QUB_Idl_ReadEVL(FileName);
    Result     := prReadEvents(DWT);

  except

    Result := True;

  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    P R T                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadPRT :Boolean;
var
  DWT        :INode;
  SampMS     :Double;
  Acrit      :String;
  PropForm   :TQUB_QFS_QDF_PropertiesForm;
begin

  Result     := False;

  //Interactive properties
  if not prOpenSilently then begin

    PropForm                         := TQUB_QFS_QDF_PropertiesForm.Create(nil);
    PropForm.panPRT.Visible          := True;
    PropForm.panPRT.BringToFront;
    PropForm.lblFileName.Caption     := FileName;

    try
      PropForm.ShowModal;

      case PropForm.ModalResult of
        idOK:
          begin
            SampMS                   := PropForm.PRTSampling;
            Acrit                    := '';
            if PropForm.PRTUseAcrit then begin
              Acrit                  := PropForm.PRTAcrit;
            end;
          end;
        idCancel:
          begin
            Result := True;
          end;
      end;//case ModalResult
    finally
      PropForm.Free;
    end;
  end;//Silently


  if Result then Exit;

  try

    DWT        := QUB_Idl_ReadPRT(FileName, Acrit);
    DWT['sampling'].SetData(SampMS);
    Result     := prReadEvents(DWT);

  except

    Result := True;

  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    P R T                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadTACidl :Boolean;
var
  DWT        :INode;
  SampMS     :Double;
  PropForm   :TQUB_QFS_QDF_PropertiesForm;
begin

  Result     := False;

  //Interactive properties
  if not prOpenSilently then begin

    PropForm                         := TQUB_QFS_QDF_PropertiesForm.Create(nil);
    PropForm.panPRT.Visible          := True;
    PropForm.panPRT.BringToFront;
    PropForm.panAcrit.Visible        := False;
    PropForm.lblFileName.Caption     := FileName;

    try
      PropForm.ShowModal;

      case PropForm.ModalResult of
        idOK:
          begin
            SampMS                   := PropForm.PRTSampling;
          end;
        idCancel:
          begin
            Result := True;
          end;
      end;//case ModalResult
    finally
      PropForm.Free;
    end;
  end;//Silently


  if Result then Exit;

  try

    DWT        := QUB_Idl_ReadTAC(FileName);
    DWT['sampling'].SetData(SampMS);
    Result     := prReadEvents(DWT);

  except

    Result := True;

  end;

end;


//****************************************************************************//
//                                                                            //
//                               R e a d    Q S F                             //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prReadQSF :Boolean;
var
  SegDWT          :TQUB_QFS_DWT_Segment;
  DWT, Seg        :INode;

begin

  Reset;

  try
    StreamPos                        := Stream.Position;

    //Segments
    prSegments.Stream                := Stream;
    prSegments.StreamPos             := Stream.Position;

    DWT               := Session.Find('Idealization').Find('Channel');
    Session.RemoveChild( DWT );
    prNewIdealization;

    //Sampling/Scaling
    Sampling       := Session['Sampling'].Data[0];
    Scaling        := Session['Scaling'].Data[0];
    ADChannelCount := Session['ADChannelCount'].Data[0];
    ADDataSize     := SizeOf(SmallInt);
    ADDataType     := NODE_DATATYPE_SMALLINT;

    // we are piggybacking on dwt code which assumes 1 channel
    Result         := ADChannelCount > 1;
    if Result then begin
      Console.NewEntry('Directly opening QSF (session) files with more than one A/D channel is not currently supported.', COLOR_ERROR);
      Exit;
    end;

    Seg            := DWT.Find('Segment');
    while not Seg.IsNull do begin
      //Segment
      SegDWT                          := TQUB_QFS_DWT_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      SegDWT.Stream                   := Stream;
      SegDWT.StreamPos                := Stream.Position;
      SegDWT.StartTime                := Sampling * Seg.Data[0];

      //Channels
      SegDWT.Channels.CanEdit         := False;
      SegDWT.Channels.EditStream      := nil;
      SegDWT.Channels.Stream          := Stream;
      SegDWT.Channels.StreamPos       := Stream.Position;
      SegDWT.Channels.DataSize        := ADChannelCount * ADDataSize;
      SegDWT.Channels.DataType        := ADDataType;
      SegDWT.Channels.DataCount       := Seg.Data[1] - Seg.Data[0] + 1;
      SegDWT.Channels.DataStreamPos   := Stream.Position;

      TQUB_QFS_DWT_Channels(SegDWT.Channels).Scaling   := Scaling;
      TQUB_QFS_DWT_Channels(SegDWT.Channels).Idl       := Idealization[0];
      TQUB_QFS_DWT_Channels(SegDWT.Channels).Offset    := Seg.Data[0];

      prSegments.AddRecord(SegDWT, prSegments.SegmentCount);
      Idealization[0].AddSeg(Seg.Data[0], Seg.Data[1]);

      Seg      := Seg.SiblingSameName;

    end;

    Idealization[0].SetDwells(DWT);

  except

    Result := True;

  end;

end;


const
  RDSEG_MAX_WINDOW : Integer = 262144;


//-- Helper function used by various prReadSegments
function TQUB_QFS_QDF.prReadSegmentAddResult( iDataType : TQUB_QFS_Node_DataType;
          iChannels, iDataSize, iSamples, iEditPos : Integer ) : TQUB_QFS_Segment;
var
  Selection : TQUB_QFS_PEditSelection;
begin
  Result := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
  if Result<>nil Then begin
    with Result do begin
      Result.Stream := EditStream;
      StreamPos := iEditPos;

      Channels.CanEdit := CanEdit;
      Channels.EditStream := EditStream;
      Channels.Stream := EditStream;
      Channels.StreamPos := iEditPos;
      Channels.DataSize := iChannels*iDataSize;
      Channels.DataType := iDataType;
      Channels.DataCount := 0;
      Channels.DataStreamPos := iEditPos;

      Selection := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount := iSamples;
      Selection.StreamPos := iEditPos;
      Selection.FromEditStream := False;
      Channels.InsertSelection(Selection, 0);
    end;
  end;
end;

function TQUB_QFS_QDF.prReadSegmentOfReals( SampleCount:Integer; iReadSize:Integer;
              aFactor:TPASingle ): TQUB_QFS_Segment; // or nil on failure
var
  EditPos                      :Integer;
  Buf                          :TPAByte;
  WindowSize                   :Integer;
  i                            :Integer;
  Number                       :Integer;
  ElemSize                     :Integer;
  dScale : double;             // Scaling factor to apply to current sample - from provided array or default single value
  iSamplesLeft : Integer;      // Buffer read iterator : Decremented from SampleCount to 0 by windowSize
  iChannels : Integer;         // Channel count from ADChannelCount

begin

  Result                       := nil;
  Buf                          := nil;
  ElemSize                     := ADDataSize;

  try

    try
      //-- Validate inputs
      if ElemSize > iReadSize then
        raise Exception.Create('Cannot do in place conversion from DOUBLE to integer');
      if (iReadSize<>4) and (iReadSize<>8) then
        raise Exception.Create('Unrecognized real number format');
      if (ElemSize<>SizeOf(SmallInt)) and (ElemSize<>Sizeof(Integer)) then
        raise Exception.Create('Unrecognized integer number format');

      //-- save the position of this data in the file for later use.
      EditPos := EditStream.Position;

      //-- Set windowsize to a multiple of ADChannelCount, <= 2^18 samples ( 1MB or 2MB total size )
      // This is required for aFactor[ mod() ] to work on blocks larger than a single window of data.
      iChannels := ADChannelCount;
      iSamplesLeft := SampleCount;
      WindowSize := min( RDSEG_MAX_WINDOW Div iChannels, SampleCount ) * iChannels ;
      Buf := Alloc1( WindowSize, iReadSize );

      while iSamplesLeft>0 do begin
        WindowSize := Min( WindowSize, iSamplesLeft * iChannels );

        Stream.ReadBuffer(Buf^, WindowSize * iReadSize);

        for i := 0 to WindowSize - 1 do begin
          //-- pick Scaling factor to use
          if aFactor=Nil Then
            dScale := Scaling
          else
            dScale := 1.0 / aFactor[ i mod iChannels ];

          //-- scale and convert to integer
          if iReadSize=4 Then
            Number := Round( TPASingle(Buf)[i]*dScale )
          else
            Number := Round( TPADouble(Buf)[i]*dScale );

          //-- Store back to array.  Reuses original array since element size <= original.
          if ElemSize = SizeOf(SmallInt) then
            TPASmall(Buf)[i] := max( min( Number, High(SmallInt) ), Low( SmallInt ) ) // limit to +/- 2^15 for 2 byte int's
          else
            TPAInteger(Buf)[i] := Number;
        end;

        EditStream.WriteBuffer(Buf^, WindowSize * ElemSize);

        Dec( iSamplesLeft, WindowSize );
      end;

      Result := prReadSegmentAddResult( ADDataType, ADChannelCount, ADDataSize, SampleCount, EditPos );

    except
      Result.Free;
      Result := nil;
    end;

  finally
    FreeMem(Buf);
  end;

end;

function TQUB_QFS_QDF.prReadSegmentOfInts(SampleCount, iReadSize :Integer; bSwapBytes:Boolean): TQUB_QFS_Segment; // or nil on failure
var
  EditPos                      :Integer;
  Buf                          :TPAByte;
  N, WindowSize                :Integer;
  i                            :Integer;

begin
  Result := nil;
  Buf    := nil;

  try
    EditPos                    := EditStream.Position;

    N                          := ADChannelCount * SampleCount;
    WindowSize                 := Min(RDSEG_MAX_WINDOW, N);
    Buf                        := Alloc1(WindowSize, iReadSize);

    if iReadSize=4 Then
      Scaling                    := Scaling / 2.0;

    while N > 0 do begin
      if WindowSize > N then WindowSize := N;

      Stream.ReadBuffer(Buf^, WindowSize * iReadSize );

      if bSwapBytes Then
        SwapBytesA( Buf, iReadSize, WindowSize );

      if iReadSize=4 Then
        for i := 0 to WindowSize - 1 do
          TPASmallInt(Buf)[i] := Buf[i] div 2;

      EditStream.WriteBuffer(Buf^, WindowSize * SizeOf(SmallInt));

      Dec( N , WindowSize );
    end;

    Result := prReadSegmentAddResult( ADDataType, ADChannelCount, ADDataSize, SampleCount, EditPos );

  except
    Result.Free;
    Result := nil;
  end;

  FreeMem(Buf);
end;

function TQUB_QFS_QDF.prReadSegmentOfReals2Chan( SampleCount, iReadSize :Integer; LeakTrace, SecondTrace :Boolean): TQUB_QFS_Segment; // or nil on failure
var
  StreamPos, StreamPos2        :Integer;
  EditPos                      :Integer;
  Buf, Buf2					   :TPAByte;
  N, WindowSize                :Integer;
  i                            :Integer;
  Number                       :Integer;

begin

  Result := nil;
  Buf    := nil;
  Buf2   := nil;

  try
    EditPos                    := EditStream.Position;
    StreamPos                  := Stream.Position;
    StreamPos2                 := StreamPos; // so it shuts up already

    N                          := SampleCount;
    WindowSize                 := Min(RDSEG_MAX_WINDOW, N);
    Buf                        := Alloc1(WindowSize,iReadSize);

    if SecondTrace then begin
      Buf2                     := Alloc1(WindowSize, iReadSize);
      StreamPos2               := StreamPos + SampleCount * iReadSize;
      if LeakTrace then StreamPos2 := StreamPos2 + SampleCount * iReadSize;
    end;

	try
      while N > 0 do begin
        if WindowSize > N
          then WindowSize := N;

        Stream.Seek(StreamPos, soFromBeginning);
        Stream.ReadBuffer(Buf^, WindowSize * iReadSize);
        StreamPos := StreamPos + WindowSize * iReadSize;

        if SecondTrace then begin
          Stream.Seek(StreamPos2, soFromBeginning);
          Stream.ReadBuffer(Buf2^, WindowSize * iReadSize);
          StreamPos2 := StreamPos2 + WindowSize * iReadSize;
        end;

        for i := 0 to WindowSize - 1 do begin
          //-- scale and convert to integer
          if iReadSize=4 Then
            Number := Round( TPASingle(Buf)[i]*Scaling )
          else
            Number := Round( TPADouble(Buf)[i]*Scaling );
          TPASmall(Buf)[2*i] := max( min( Number, High(SmallInt) ), Low( SmallInt ) ); // limit to +/- 2^15 for 2 byte int's

          if SecondTrace then begin
  		      if iReadSize=4 Then
	  		      Number := Round( TPASingle(Buf2)[i]*Scaling )
		        else
              Number := Round( TPADouble(Buf2)[i]*Scaling );
	          TPASmall(Buf)[2*i+1] := max( min( Number, High(SmallInt) ), Low( SmallInt ) ); // limit to +/- 2^15 for 2 byte int's
          end;
        end;

        EditStream.WriteBuffer( Buf^, WindowSize * 2 * SizeOf(SmallInt));

        Dec( N, WindowSize);
      end;

      Result := prReadSegmentAddResult( ADDataType, ADChannelCount, ADDataSize, SampleCount, EditPos );

    except
      Result.Free;
      Result := nil;
    end;

  finally
    FreeMem(Buf);
    FreeMem(Buf2);
  end;

end;

function TQUB_QFS_QDF.prReadSegmentOfInts2Chan(SampleCount, iReadSize :Integer; bSwapBytes, LeakTrace, SecondTrace :Boolean): TQUB_QFS_Segment; // or nil on failure
var
  StreamPos1, StreamPos2       :Integer;
  EditPos                      :Integer;
  Buf1, Buf2                   :TPAByte;
  BufOut                       :TPASmall;
  N, WindowSize                :Integer;
  i                            :Integer;
  iDiv : Integer;              // For 4byte ints, value is Div'd by 2

begin
  Result := nil;
  Buf1 := nil;
  Buf2 := nil;
  BufOut := nil;

  try
    EditPos                    := EditStream.Position;
    StreamPos1                 := Stream.Position;
    StreamPos2                 := StreamPos1; // so it shuts up already

    N                          := SampleCount;
    WindowSize                 := Min(RDSEG_MAX_WINDOW, N);
    Buf1 := Alloc1(WindowSize, iReadSize);
    BufOut := SmAlloc1( 2*WindowSize );

    if iReadSize=4 Then Begin
      iDiv := 2;
      Scaling := Scaling / 2.0;
    end else
      iDiv := 1;

    if SecondTrace then begin   // 2nd trace is read after 1st trace, with possibly a 'leak' trace in between them.
      Buf2 := Alloc1(WindowSize,iReadSize);
      StreamPos2 := StreamPos1 + SampleCount * iReadSize * Choice(LeakTrace,2,1);
    end;

    while N > 0 do begin
      if WindowSize > N then
        WindowSize := N;

      // Read channel 1 to buf1
      Stream.Seek(StreamPos1, soFromBeginning);
      Stream.ReadBuffer(Buf1^, WindowSize * iReadSize);
      StreamPos1 := StreamPos1 + WindowSize * iReadSize;
      if bSwapBytes Then
        SwapBytesA( Buf1, iReadSize, WindowSize );

      // Read channel 2 (if any) to Buf2
      if SecondTrace then begin
        Stream.Seek(StreamPos2, soFromBeginning);
        Stream.ReadBuffer(Buf2^, WindowSize * iReadSize);
        StreamPos2 := StreamPos2 + WindowSize * iReadSize;
        if bSwapBytes Then
          SwapBytesA( Buf2, iReadSize, WindowSize );
      end;

      // shuffle buf1 and buf2 to bufout with integer->smallint conversion if rqd.
      for i := 0 to WindowSize - 1 do begin
        BufOut[2*i] := Buf1[i] div iDiv;
        if SecondTrace then
          BufOut[2*i+1] := Buf2[i] div iDiv
        else
          BufOut[2*i+1] := 0;
      end;

      EditStream.WriteBuffer(BufOut^, WindowSize * 2 * SizeOf(SmallInt));

      Dec( N, WindowSize );
    end;

    Result := prReadSegmentAddResult( ADDataType, ADChannelCount, ADDataSize, SampleCount, EditPos );

  except
    Result.Free;
    Result := nil;
  end;

  FreeMem(Buf1);
  FreeMem(Buf2);
  FreeMem(BufOut);
end;


//06-11-23, while waiting for the turkey. LM
//To be used when the Leak trace is desired - will fill with zeroes if absent, same as Second trace.
//I assume all three traces have the same DataType (and DataSize)
//QuB data will be MPX-ed as: FirstTrace/SecondTrace/Leak
function TQUB_QFS_QDF.prReadSegment2or3Chan(SampleCount :Integer; DataType :TQUB_QFS_NODE_DataType; LittleEndian :Boolean; LeakTrace, SecondTrace :Boolean): TQUB_QFS_Segment; // or nil on failure
var
  StreamPos                    :Array[0..2] of Integer;
  EditPos                      :Integer;
  Buf                          :Array[0..2] of Pointer;
  HasCh                        :Array[0..2] of Boolean;
  WantsCh                      :Array[0..2] of Boolean;
  BufMPX                       :Pointer;
  DataSize                     :Integer;
  N2or3                        :Integer;
  N, WindowSize                :Integer;
  Number                       :Array[0..2] of Double;
  i, j                         :Integer;
  Selection                    :TQUB_QFS_PEditSelection;
begin

  Result                       := nil;

  for j := 0 to 2 do Buf[j]    := nil;
  BufMPX                       := nil;//For all, multiplexed

  DataSize := SizeOf(SmallInt);
  case DataType of
    NODE_DATATYPE_SMALLINT     :DataSize := SizeOf(SmallInt);
    NODE_DATATYPE_LONGINT      :DataSize := SizeOf(LongInt);
    NODE_DATATYPE_SINGLE       :DataSize := SizeOf(Single);
    NODE_DATATYPE_DOUBLE       :DataSize := SizeOf(Double);
  end;

  N2or3                        := ADChannelCount;

  HasCh[0]                     := True;
  HasCh[1]                     := SecondTrace;
  HasCh[2]                     := LeakTrace;
  WantsCh[0]                   := True;
  WantsCh[1]                   := N2or3 > 1;
  WantsCh[2]                   := N2or3 > 2;

  try

    try
      EditPos                  := EditStream.Position;
      StreamPos[0]             := Stream.Position;
      if HasCh[2] then StreamPos[2] := StreamPos[0] + SampleCount * DataSize else StreamPos[2] := StreamPos[0];
      if HasCh[1] then StreamPos[1] := StreamPos[2] + SampleCount * DataSize;

      N                        := SampleCount;
      WindowSize               := Min(RDSEG_MAX_WINDOW, N);

      //Multiplexed
      BufMPX                   := AllocMem(ADDataSize * 3 * WindowSize);

      for j := 0 to 2 do begin
        if HasCh[j] then Buf[j]:= AllocMem(DataSize * WindowSize);
      end;

      while N > 0 do begin

        if WindowSize > N then WindowSize := N;

        //Read data into buffer
        for j := 0 to 2 do begin
          if (HasCh[j]) and (WantsCh[j]) then begin
            Stream.Seek(StreamPos[j], soFromBeginning);
            Stream.ReadBuffer(Buf[j]^, WindowSize * DataSize);
            StreamPos[j] := StreamPos[j] + WindowSize * DataSize;
          end;
        end;

        for i := 0 to WindowSize - 1 do begin

          //Read Number from Buf or set it to 0;
          for j := 0 to 2 do begin
            Number[j] := 0;
            case DataType of
              NODE_DATATYPE_SMALLINT:
                begin
                  if (WantsCh[j]) and (HasCh[j]) then begin
                    if LittleEndian then begin
                      Swap2Bytes(@(TPASmallInt(Buf[j])[i]));
                    end;
                    Number[j] := TPASmallInt(Buf[j])[i];
                  end;
                end;
              NODE_DATATYPE_LONGINT:
                begin
                  if (WantsCh[j]) and (HasCh[j]) then begin
                    if LittleEndian then begin
                      Swap4Bytes(@(TPALongInt(Buf[j])[i]));
                    end;
                    Number[j] := TPALongInt(Buf[j])[i];
                  end;
                end;
              NODE_DATATYPE_SINGLE:   if (WantsCh[j]) and (HasCh[j]) then Number[j] := Scaling * TPASingle(Buf[j])[i];
              NODE_DATATYPE_DOUBLE:   if (WantsCh[j]) and (HasCh[j]) then Number[j] := Scaling * TPADouble(Buf[j])[i];
            end;
          end;

          //Set BufMPX
          for j := 0 to 2 do begin

            if WantsCh[j] then begin

              case ADDataType of

                NODE_DATATYPE_BYTE:
                  if Number[j] > High(Byte)        then TPAByte(BufMPX)[i*N2or3+j] := High(Byte)          else if Number[j] < Low(Byte)       then TPAByte(BufMPX)[i*N2or3+j]      := Low(Byte)        else TPAByte(BufMPX)[i*N2or3+j]     := Round(Number[j]);
                NODE_DATATYPE_SHORTINT:
                  if Number[j] > High(ShortInt)    then TPAShort(BufMPX)[i*N2or3+j] := High(ShortInt)     else if Number[j] < Low(ShortInt)   then TPAShort(BufMPX)[i*N2or3+j]     := Low(ShortInt)    else TPAShort(BufMPX)[i*N2or3+j]    := Round(Number[j]);
                NODE_DATATYPE_WORD:
                  if Number[j] > High(Word)        then TPAWord(BufMPX)[i*N2or3+j] := High(Word)          else if Number[j] < Low(Word)       then TPAWord(BufMPX)[i*N2or3+j]      := Low(Word)        else TPAWord(BufMPX)[i*N2or3+j]     := Round(Number[j]);
                NODE_DATATYPE_SMALLINT:
                  if Number[j] > High(SmallInt)    then TPASmall(BufMPX)[i*N2or3+j] := High(SmallInt)     else if Number[j] < Low(SmallInt)   then TPASmall(BufMPX)[i*N2or3+j]     := Low(SmallInt)    else TPASmall(BufMPX)[i*N2or3+j]    := Round(Number[j]);
                NODE_DATATYPE_LONGWORD:
                  if Number[j] > High(LongWord)    then TPALongWord(BufMPX)[i*N2or3+j] := High(LongWord)  else if Number[j] < Low(LongWord)   then TPALongWord(BufMPX)[i*N2or3+j]  := Low(LongWord)    else TPALongWord(BufMPX)[i*N2or3+j] := Round(Number[j]);
                NODE_DATATYPE_LONGINT:
                  if Number[j] > High(LongInt)     then TPALongInt(BufMPX)[i*N2or3+j] := High(LongInt)    else if Number[j] < Low(LongInt)    then TPALongInt(BufMPX)[i*N2or3+j]   := Low(LongInt)     else TPALongInt(BufMPX)[i*N2or3+j]  := Round(Number[j]);
                NODE_DATATYPE_UINT64:
                  if Number[j] > High(UInt64)      then TPAUInt64(BufMPX)[i*N2or3+j] := High(UInt64)      else if Number[j] < Low(UInt64)     then TPAUInt64(BufMPX)[i*N2or3+j]    := Low(UInt64)      else TPAUInt64(BufMPX)[i*N2or3+j]   := Round(Number[j]);
                NODE_DATATYPE_INT64:
                  if Number[j] > High(Int64)       then TPAInt64(BufMPX)[i*N2or3+j] := High(Int64)        else if Number[j] < Low(Int64)      then TPAInt64(BufMPX)[i*N2or3+j]     := Low(Int64)       else TPAInt64(BufMPX)[i*N2or3+j]    := Round(Number[j]);
                NODE_DATATYPE_SINGLE:
                  TPASingle(BufMPX)[i*N2or3+j]     := Number[j];
                NODE_DATATYPE_DOUBLE:
                  TPADouble(BufMPX)[i*N2or3+j]     := Number[j];
                NODE_DATATYPE_EXTENDED:
                  TPAExtended(BufMPX)[i*N2or3+j]   := Number[j];
              end;//case

            end;//WantsCh

          end;//j

        end;//i

        EditStream.WriteBuffer(BufMPX^, WindowSize * ADChannelCount * ADDataSize);

        N                        := N - WindowSize;

      end;


      Result                     := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      with Result do begin
        Result.Stream            := EditStream;
        StreamPos                := EditPos;

        Channels.CanEdit         := CanEdit;
        Channels.EditStream      := EditStream;
        Channels.Stream          := EditStream;
        Channels.StreamPos       := EditPos;
        Channels.DataSize        := ADChannelCount * ADDataSize;
        Channels.DataType        := ADDataType;
        Channels.DataCount       := 0;
        Channels.DataStreamPos   := EditPos;

        Selection                := AllocMem(SizeOf(TQUB_QFS_EditSelection));
        Selection.DataCount      := SampleCount;
        Selection.StreamPos      := EditPos;
        Selection.FromEditStream := False;
        Channels.InsertSelection(Selection, 0);
      end;

    except
      Result.Free;
      Result := nil;
    end;

  finally

    for j := 0 to 2 do FreeMem(Buf[j]);
    FreeMem(BufMPX);

  end;
end;



//****************************************************************************//
//                                                                            //
//                      W r i t e   -   l o w   l e v e l                     //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prWriteDATHeader(Stream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_QDF.prWriteLDTHeader(Stream :TStream) :Boolean;
var
  LDTLeader   :Integer;
  LDTSampling :Word;
  LDTScaling  :Word;
begin
  Result := False;

  try

    //it strips the leader
    LDTLeader   := 0;//Self.LDTLeader;
    LDTSampling := Self.LDTSampling;
    LDTScaling  := Self.LDTScaling;

    //Leader
    Stream.WriteBuffer(LDTLeader, SizeOf(LDTLeader));

    //Sampling
    Stream.WriteBuffer(LDTSampling, SizeOf(LDTSampling));

    //Scaling
    Stream.WriteBuffer(LDTScaling, SizeOf(LDTScaling));

  except

    Result := True;

  end;

end;

function TQUB_QFS_QDF.prWriteASCHeader(Stream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_QDF.prWriteABFHeader(Stream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_QDF.prWriteWAVHeader(Stream :TStream) :Boolean;
var
  WaveChunk :TQUB_QFS_QDF_WAV_Chunk;
begin
  Result := False;

  try

    WaveChunk.riff.riff              := QUB_QFS_QDF_STRING_RIFF;
    WaveChunk.riff.length            := SizeOf(WaveChunk) - SizeOf(WaveChunk.riff.riff) - SizeOf(WaveChunk.riff.length) + prSegments.TotalDataCount * ADChannelCount * ADDataSize;
    WaveChunk.riff.wave              := QUB_QFS_QDF_STRING_WAVE;

    WaveChunk.format.fmt             := QUB_QFS_QDF_STRING_FMT;
    WaveChunk.format.length          := $10;
    WaveChunk.format.formattag       := $0001;
    WaveChunk.format.channels        := ADChannelCount;
    WaveChunk.format.samplerate      := Round(1 / Sampling);
    WaveChunk.format.bytespersample  := ADDataSize * ADChannelCount;
    WaveChunk.format.bytespersec     := WaveChunk.format.bytespersample * WaveChunk.format.samplerate;
    WaveChunk.format.bitspersample   := WAVBitsPerSample;

    WaveChunk.data.data              := QUB_QFS_QDF_STRING_DATA;
    WaveChunk.data.length            := prSegments.TotalDataCount * WaveChunk.format.bytespersample;

    Stream.WriteBuffer(WaveChunk, SizeOf(WaveChunk));

  except

    Result := True;

  end;

end;

//Segment Header
function TQUB_QFS_QDF.prWriteDATSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
begin
  Result := False
end;

function TQUB_QFS_QDF.prWriteLDTSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
var
  StartTime :Integer;
  DataCount :Integer;
begin
  Result := False;

  try

    //StartTime
    StartTime := Round(prSegments[Segment].StartTime / 1e-6);
    Stream.WriteBuffer(StartTime, SizeOf(StartTime));

    //DataPoints
    DataCount := prSegments[Segment].Channels.DataCount;
    Stream.WriteBuffer(DataCount, SizeOf(DataCount));

  except

    Result := True;

  end;

end;

function TQUB_QFS_QDF.prWriteASCSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_QDF.prWriteABFSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_QDF.prWriteWAVSegmentHeader(Segment :Integer; Stream :TStream) :Boolean;
begin
  Result := False;
end;

//Segment
function TQUB_QFS_QDF.prWriteDATSegment(Segment :Integer; Stream :TStream) :Boolean;
var
  Buffer :Pointer;
  Sel    :TQUB_FirstLast;
begin
  Result := False;

  try

    Buffer := AllocMem(ONE_MEGABYTE * prSegments[Segment].Channels.DataSize);

  except

    Result := True;
    Exit;

  end;

  try

    try

      Sel.First := 0;
      Sel.Last  := 0;
      repeat

        Sel.Last := Sel.First + ONE_MEGABYTE - 1;
        if Sel.Last > (prSegments[Segment].Channels.DataCount - 1) then Sel.Last := (prSegments[Segment].Channels.DataCount - 1);

        prSegments[Segment].Channels.LoadDataInBuffer(Buffer, Sel.First, Sel.Last);

        Stream.WriteBuffer(Buffer^, (Sel.Last - Sel.First + 1) * prSegments[Segment].Channels.DataSize);
        Sleep(5); // ?

        Sel.First := Sel.Last + 1;

        if Assigned(prSaveUpdate) then prSaveUpdate(-1.0{no pct info});

      until(Sel.First >= prSegments[Segment].Channels.DataCount);

    except

      Result := True;
      Exit;

    end;

  finally

    FreeMem(Buffer);

  end;

end;

function TQUB_QFS_QDF.prWriteLDTSegment(Segment :Integer; Stream :TStream) :Boolean;
begin
  Result := prWriteDATSegment(Segment, Stream);
end;

function TQUB_QFS_QDF.prWriteASCSegment(Segment :Integer; Stream :TStream) :Boolean;
var
  BufferMPX   :Pointer;
  Buffer      :TPMInteger;
  BufferCount :Integer;
  ASCII       :String;
  Seg         :TQUB_QFS_Segment;
  Chunk       :TQUB_FirstLast;
  p, c, cc    :Integer;
begin

  Result      := False;

  BufferMPX   := nil;
  Buffer      := nil;

  try

    Seg := Segments[Segment];
    if Seg.Channels.DataCount > ONE_MEGABYTE then begin
      BufferCount := ONE_MEGABYTE;
    end else begin
      BufferCount := Seg.Channels.DataCount;
    end;
    cc := ADChannelCount;
    Buffer     := IAlloc2(cc, BufferCount);
    BufferMPX  := AllocMem(cc * ADDataSize * BufferCount);

    Chunk.First := 0;

    repeat

      Chunk.Last := Chunk.First + BufferCount - 1;
      if Chunk.Last >= Seg.Channels.DataCount then Chunk.Last := Seg.Channels.DataCount - 1;

      if (Chunk.Last - Chunk.First + 1) < BufferCount then begin
        BufferCount := (Chunk.Last - Chunk.First + 1);
      end;

      if Seg.Channels.LoadDataInBuffer(BufferMPX, Chunk.First, Chunk.Last) then begin
        Result := True;
        Exit;
      end;

      if MPXData(Buffer, BufferMPX, cc, -1{All Channels}, ADDataSize, ADDataType, 0, BufferCount - 1, False{DE-MPX}) then begin
        Result := True;
        Exit;
      end;

      ASCII := '';

      try

        for p := 0 to BufferCount - 1 do begin
          for c := 0 to cc - 1 do begin
            ASCII := ASCII + Format('%.6g', [Buffer[c, p] / Scaling]);
            if c < (cc - 1) then begin
              ASCII := ASCII + #9;
            end;
          end;
          ASCII := ASCII + #13#10;
        end;

        Stream.WriteBuffer(PChar(ASCII)^, Length(ASCII));

      except

        Result := True;
        Exit;

      end;

      Chunk.First := Chunk.Last + 1;

      if Assigned(prSaveUpdate) then prSaveUpdate(-1.0{no pct info});

    until Chunk.First >= Seg.Channels.DataCount;

  finally

    FreeMem(BufferMPX);
    Free2(TPMChar(Buffer));

  end;

end;

function TQUB_QFS_QDF.prWriteABFSegment(Segment :Integer; Stream :TStream) :Boolean;
begin
  Result := False;
end;

function TQUB_QFS_QDF.prWriteWAVSegment(Segment :Integer; Stream :TStream) :Boolean;
begin
  Result := prWriteDATSegment(Segment, Stream);
end;



//****************************************************************************//
//                                                                            //
//                            R E A D    M E T H O D S                        //
//                                                                            //
//****************************************************************************//


{
     These functions may be called by QVD (Browse etc) concurrently with AppendData
       (called by Simu or Acquisition).
}



//****************************************************************************//
//                                                                            //
//                              R e a d   D a t a                             //
//                                                                            //
//****************************************************************************//

//This function is safe as long as no other thread DELETES data before or within the First/Last

function TQUB_QFS_QDF.ReadData(Buffer :Pointer; Raw :Boolean; Channel :Integer; DataFirst, DataLast :Integer) :Boolean;
var
  RawBuffer   :Pointer;
  pData2      :TPSmallInt;
  pData4      :TPSingle;
  pData8      :TPDouble;
  BufferSize  :Integer;
  NewSize     :Integer;
  SegCount    :Integer;
  ChDataCount :Integer;
  f, l        :Integer;
  cf, cl      :Integer;
  i, s        :Integer;
begin

  //Read data information
  try
    prSegInfoCritSection.Acquire;
    Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
    if Result then Exit;
    SegCount    := prSegments.SegmentCount;
    ChDataCount := prSegments[SegCount - 1].Channels.DataCount;
  finally
    prSegInfoCritSection.Release;
  end;

  try

    //Testing: let's try to concentrate this around the actual Read/Write. LM

    //Acquire ReadWrite Critical Section
    //prReadWriteCritSection.Acquire;

    if Raw then RawBuffer := Buffer else RawBuffer := nil;
    BufferSize := 0;

    try

      f := 0;
      s := 0;

      repeat

        //calculate the last point
        if s = (SegCount - 1) then begin
        //If it is the last segment use the value it had at the time of entry
        //(which may change due to appending)
          l := f + ChDataCount - 1;
        end else begin
          l := f + prSegments[s].Channels.DataCount - 1;
        end;

        //Segment s'th has (some) data within First to Last range
        if not ((DataLast < f) or (DataFirst > l)) then begin

          cf := f;
          if DataFirst > cf then cf := DataFirst;
          cl := l;
          if DataLast < cl  then cl := DataLast;

          if not Raw then begin

            //Read the data within the Segment's range
            NewSize := (cl - cf + 1) * prSegments[s].Channels.DataSize;
            if BufferSize < NewSize then begin
              BufferSize := NewSize;
              try
                ReAllocMem(RawBuffer, BufferSize);
              except
                Result := True;
              end;
            end;
            if Result then Exit;

          end;


          //....................................................................
          //Testing: let's try to concentrate this around the actual Read/Write. LM
          try

            //Acquire ReadWrite Critical Section
            prReadWriteCritSection.Acquire;

            if Raw then begin
              Result := prSegments[s].Channels.LoadDataInBuffer(@TPAByte(RawBuffer)[prSegments[s].Channels.DataSize * (cf - DataFirst)], cf - f, cl - f);
            end else begin
              Result := prSegments[s].Channels.LoadDataInBuffer(RawBuffer, cf - f, cl - f);
              if m_lMacData then begin
                pData2 := TPSmallInt(RawBuffer);  // one of these three statements matter
                pData4 := TPSingle(RawBuffer);    // single or 4 byte int
                pData8 := TPDouble(RawBuffer);
                for i := 0 to cl-cf do Begin
                  if prSegments[s].Channels.DataSize = 2 then begin
                    Swap2Bytes( pData2 );
                    inc(pData2,1);
                  end else if prSegments[s].Channels.DataSize = 4 then begin
                    Swap4Bytes( pData4 );
                    inc(pData4,1);
                  end else if prSegments[s].Channels.DataSize = 8 then begin
                    Swap8Bytes( pData8 );
                    inc(pData8,1);
                  end;
                end;
              end;
            end;
 
          finally

            //Release ReadWrite Critical Section
            prReadWriteCritSection.Release;

          end;
          //....................................................................


          if Result then Exit;

          if not Raw then begin
            try
              Result := MPXData(Buffer, RawBuffer, ADChannelCount, Channel{-1 = ALL}, ADDataSize, ADDataType, cf - DataFirst, cl - DataFirst, False {De-MPX}, m_nMultiplier);
            except
              Result := True;
            end;
          end;
          if Result then Exit;

        end;

        Inc(s);

        f := l + 1;

      until ((s = SegCount) or (f > DataLast));

    finally

      if not Raw then FreeMem(RawBuffer);

    end;

  finally

    //Testing: let's try to concentrate this around the actual Read/Write. LM

    //Release ReadWrite Critical Section
    //prReadWriteCritSection.Release;

  end;

end;


//----- Companion to ReadData - generate list of segment breaks in data buffer
procedure TQUB_QFS_QDF.ListSegmentBreaks( iOffset : Integer; aBreaks : TStringList );
var
  SegCount : Integer;
  iSeg, iPoints : Integer;
begin
  //----- Get Segment Count
  SegCount := 0;
  try
    prSegInfoCritSection.Acquire;
    if prOpened and (prSegments<>nil) Then
      SegCount := prSegments.SegmentCount;
  finally
    prSegInfoCritSection.Release;
  end;

  //----- Add end of each segment to break list
  iPoints := 0;
  iSeg := 0;
  while (iSeg<SegCount) {and (iPoints<iLast)} do Begin
    //-- running total of point count
    inc( iPoints, prSegments[iSeg].Channels.DataCount );
    //-- add to list of segment breaks
    aBreaks.Add( IntToStr(iPoints+iOffset) );
    //-- continue with next segment
    Inc(iSeg);
  end;
end;


//****************************************************************************//
//                                                                            //
//                             W r i t e   D a t a                            //
//                                                                            //
//****************************************************************************//




//****************************************************************************//
//                                                                            //
//                          E D I T     M E T H O D S                         //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.CutSelection(First, Last :Integer; CopyASCII :Boolean) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanCut then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prCut(First, Last, True{DoOnePoint}, CopyASCII);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_CUT, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_CUT, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.CopySelection(First, Last :Integer; CopyASCII :Boolean) :Boolean;
begin
  Result := True;
  if CanCopy then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prCopy(First, Last, True{DoOnePoint}, CopyASCII);
      except
        Result := True;
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.CopyList(List :TQUB_QVO_QVD_SelectionList; ListName :String; CopyASCII :Boolean) :Boolean;
begin
  Result := True;
  if CanCopy then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prCopyList(List, ListName, CopyASCII);
      except
        Result := True;
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.PasteSelection(First, Last :Integer) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanPaste then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prPaste(First, Last, False{PasteAsSeg}, True{ReplaceSel});
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_PASTE, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_PASTE, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.PasteSegment(First, Last :Integer) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanPaste then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prPaste(First, Last, True{PasteAsSeg}, False{ReplaceSel});
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_PASTE, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_PASTE, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.Append :Boolean;
var
  Sel   :TQUB_FirstLast;
  First :Integer;
  Last  :Integer;
begin
  Result := True;
  if CanPaste then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      First     := Segments.TotalDataCount;
      Last      := First;
      try
        Result := prPaste(First, Last, True{PasteAsSeg}, False{ReplaceSel});
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_PASTE, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_PASTE, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.DeleteSelection(First, Last :Integer) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanDelete then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prDelete(First, Last, True{DoOnePoint}, nil{DelOperation}, False{AllowEmptySeg});
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_DELETE, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_DELETE, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.EraseSelection(First, Last :Integer; BufFirst, BufLast :Integer) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanErase then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prErase(First, Last, BufFirst, BufLast, True{DoOnePoint});
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_ERASE, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_ERASE, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.FillSelection(First, Last :Integer; Channel :Integer; Value :Variant) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanFill then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prFill(First, Last, Channel, Value);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_FILL, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_FILL, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.BreakSegment(First, Last :Integer) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanBreak then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prBreak(First, Last);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := First;
        Sel.Last   := Last;
        //DoOnEdit(EDIT_OPER_BREAK, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_BREAK, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.EditPoint(Channel :Integer; Point :Integer; Change :Integer) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prEditPoint(Channel, Point, Change);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        Sel.First  := Point;
        Sel.Last   := Point;
        //DoOnEdit(EDIT_OPER_EDITPOINT, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_EDITPOINT, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.Clear :Boolean;
var
  Sel :TQUB_FirstLast;
  i   :Integer;
begin
  Result := True;
  if CanDelete then begin

    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      Sel.First  := 0;
      Sel.Last   := prSegments.TotalDataCount - 1;
      try
        Result := prDelete(0, prSegments.TotalDataCount - 1, True{DoOnePoint}, nil{DelOperation}, False{AllowEmptySeg});
      except
        Result := True;
      end;
      if not Result then begin
        prClear;
        prModified := True;
        //DoOnEdit(EDIT_OPER_CLEAR, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_CLEAR, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;

  end else if prSegments.TotalDataCount = 0 then begin

    Result := False;

  end else begin

    Result := True;

  end;

  if not Result then begin
    for i := 0 to ADChannelCount - 1 do begin
      try
        prIdealization[i].Clear;
      except
      end;
    end;
  end;
end;

function TQUB_QFS_QDF.Undo :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanUndo then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prUndo;
      except
        Result := True;
      end;
      if not Result then begin
        prModified := prEditOperations.Count <> 0;
        Sel.First  := -1;
        Sel.Last   := -1;
        //DoOnEdit(EDIT_OPER_UNDO, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UNDO, Sel, Sel);
      end;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;


//****************************************************************************//
//                                                                            //
//                            R E C O R D    D A T A                          //
//                                                                            //
//****************************************************************************//
{
This procedure is supposed to have higher priority than Append,
and to be used only by Acquisition. Needs more work. 
12-2004 LM.
}
function TQUB_QFS_QDF.RecordData(Buffer :Pointer; BufferSize, BufferCount :Integer; NewSegment :Boolean; StartTime :Double; Refresh :Boolean) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanAppend then begin
    //if prEditing then Exit;
    try
      //prEditCritSection.Acquire;
      //prEditing := True;
      Sel.First  := prSegments.TotalDataCount;
      Sel.Last   := prSegments.TotalDataCount + BufferCount - 1;
      try
        Result := prAppendData(Buffer, BufferSize, BufferCount, False{BufferIsWritten}, 0{BufferAtPos}, NewSegment, StartTime);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        //if Refresh then DoOnEdit(EDIT_OPER_APPEND, Sel, Sel);
        if Refresh then if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_APPEND, Sel, Sel);
      end;
      //prEditing := False;
    finally
      //prEditing := False;
      //prEditCritSection.Release;
    end;
  end;
end;


//****************************************************************************//
//                                                                            //
//                A P P E N D   /   R E P L A C E    D A T A                  //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.AppendData(Buffer :Pointer; BufferSize, BufferCount :Integer; NewSegment :Boolean; StartTime :Double; Refresh :Boolean) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanAppend then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      Sel.First  := prSegments.TotalDataCount;
      Sel.Last   := prSegments.TotalDataCount + BufferCount - 1;
      try
        Result := prAppendData(Buffer, BufferSize, BufferCount, False{BufferIsWritten}, 0{BufferAtPos}, NewSegment, StartTime);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        //if Refresh then DoOnEdit(EDIT_OPER_APPEND, Sel, Sel);
        if Refresh then if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_APPEND, Sel, Sel);
      end;
      prEditing := False;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.AppendDataASCII(const Text :String; NewSegment, TimeColumn, UnevenSampling, Segmented :Boolean; Refresh :Boolean) :Boolean;
var
  Sel    :TQUB_FirstLast;
begin
  Result := True;
  if CanAppend then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      Sel.First := prSegments.TotalDataCount;
      try
        Result  := prAppendDataASCII(Text, NewSegment, TimeColumn, UnevenSampling, Segmented);
      except
        Result  := True;
      end;
      Sel.Last  := prSegments.TotalDataCount - 1;
      if not Result then begin
        prModified := True;
        //DoOnEdit(EDIT_OPER_APPEND, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_APPEND, Sel, Sel);
      end;
      prEditing := False;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;

function TQUB_QFS_QDF.NewSegment(StartTime :Double; Refresh :Boolean; ForceNewIfEmpty :Boolean) :Boolean;
var
  Sel :TQUB_FirstLast;
begin
  Result := True;
  if CanAppend then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      Sel.First  := prSegments.TotalDataCount;
      Sel.Last   := prSegments.TotalDataCount + 0 - 1;
      try
        Result := prNewSegment(StartTime, ForceNewIfEmpty);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        //if Refresh then DoOnEdit(EDIT_OPER_APPEND, Sel, Sel);
        if Refresh then if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_APPEND, Sel, Sel);
      end;
      prEditing := False;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;


function TQUB_QFS_QDF.prAppendDataASCII(const Text :String; NewSegment, TimeColumn, UnevenSampling, Segmented :Boolean) :Boolean;
var
  TextStream :TMemoryStream;
  DataCount  :Integer;
  SegPos     :Integer;
  FirstSeg   :Boolean;
  Entry      :Integer;
begin

  Entry      := Console.NewEntry('', COLOR_TEXT);

  TextStream := TMemoryStream.Create;

  FirstSeg   := True;

  try

    TextStream.WriteBuffer(Text[1], Length(Text));
    TextStream.Position := 0;

    repeat

      try

        SegPos := prEditStream.Size;

        Result := prConvertASCII(TextStream, ADChannelCount, TimeColumn, UnevenSampling, Segmented, DataCount, Segments.SegmentCount, Entry);
        if Result then Break;

        if DataCount > 0 then begin

          Result := prAppendData(nil, ADChannelCount * ADDataSize, DataCount, True{BufferIsWritten}, SegPos{BufferAtPos}, ((FirstSeg) and (NewSegment)) or (not FirstSeg), prSegments.TotalDataCount * Sampling);
          if Result then Break;

          FirstSeg := False;

        end;

      except

        Result := True;

      end;

    until((DataCount < 0) or (Result));

  finally

    TextStream.Free;

  end;

end;

function TQUB_QFS_QDF.ReplaceData(Buffer :Pointer; BufferSize :Integer; Sel :TQUB_FirstLast; Refresh :Boolean) :Boolean;
begin
  Result := True;
  if CanEdit then begin
    if prEditing then Exit;
    try
      prEditCritSection.Acquire;
      prEditing := True;
      try
        Result := prReplaceData(Buffer, BufferSize, Sel);
      except
        Result := True;
      end;
      if not Result then begin
        prModified := True;
        //DoOnEdit(EDIT_OPER_UPDATE, Sel, Sel);
        if Assigned(OnEdit) then OnEdit(Self, EDIT_OPER_UPDATE, Sel, Sel);
      end;
      prEditing := False;
    finally
      prEditing := False;
      prEditCritSection.Release;
    end;
  end;
end;


//****************************************************************************//
//                                                                            //
//                     Protected    E D I T    methods                        //
//                                                                            //
//****************************************************************************//
function TQUB_QFS_QDF.prCut(First, Last :Integer; DoOnePoint :Boolean; CopyASCII :Boolean) :Boolean;
begin
  Result := prCopy(First, Last, DoOnePoint, CopyASCII);
  if not Result then begin
    Result := prDelete(First, Last, DoOnePoint, nil{DelOperation}, False{AllowEmptySeg});
  end;
end;

function TQUB_QFS_QDF.prCopy(First, Last :Integer; DoOnePoint :Boolean; CopyASCII :Boolean) :Boolean;
begin
  if CopyASCII then begin
    Result := prCopyASCII(First, Last, DoOnePoint);
  end else begin
    Result := prCopyBinary(First, Last, DoOnePoint);
  end;
end;

function TQUB_QFS_QDF.prCopyASCII(First, Last :Integer; DoOnePoint :Boolean) :Boolean;
const
  BUF_INCR        = ONE_MEGABYTE div 8;
var
  ASCII           :String;
  AsciiBufStr     :String;
  AsciiBufLen     :Integer;
  AsciiBufOffset  :Integer;

  BufferMPX   :Pointer;
  Buffer      :TPMInteger;
  BufferCount :Integer;
  Chunk       :TQUB_FirstLast;
  p, c, cc, l     :Integer;
begin

  Result      := False;

  BufferMPX   := nil;
  Buffer      := nil;

  try

    if (Last - First + 1) > ONE_MEGABYTE then begin
      BufferCount := ONE_MEGABYTE;
    end else begin
      BufferCount := Last - First + 1;
    end;
    cc := ADChannelCount;
    Buffer     := IAlloc2(cc, BufferCount);
    BufferMPX  := AllocMem(cc * ADDataSize * BufferCount);

    Chunk.First := First;

    AsciiBufOffset  := 1;//0;Strings are 1-based
    AsciiBufStr     := '';
    AsciiBufLen     := BufferCount * cc * 16;
    SetLength(AsciiBufStr, AsciiBufLen);


    repeat

      Chunk.Last := Chunk.First + BufferCount - 1;
      if Chunk.Last > Last then Chunk.Last := Last;

      if (Chunk.Last - Chunk.First + 1) < BufferCount then begin
        BufferCount := (Chunk.Last - Chunk.First + 1);
      end;

      if ReadData(Buffer, False{Raw}, -1, Chunk.First, Chunk.Last) then begin
        Result := True;
        Exit;
      end;

      try

        for p := 0 to BufferCount - 1 do begin
          ASCII := '';
          for c := 0 to cc - 1 do begin
            ASCII := Format('%.6g', [Buffer[c, p] / Scaling]);
            if c < (cc - 1) then begin
              ASCII := ASCII + #9;
            end;
          end;
          ASCII := ASCII + #13#10;

          l := Length(ASCII);
          if (AsciiBufOffset + l) > AsciiBufLen then begin
            AsciiBufLen  := AsciiBufLen + BUF_INCR;
            SetLength(AsciiBufStr, AsciiBufLen);
          end;
          Move(PChar(ASCII)^, (@AsciiBufStr[AsciiBufOffset])^, l);
          AsciiBufOffset := AsciiBufOffset + l;
        end;

      except

        Result := True;
        Exit;

      end;

      Chunk.First := Chunk.Last + 1;

    until Chunk.First > Last;

    ASCII := #0;
    l := Length(ASCII);
    if (AsciiBufOffset + l) > AsciiBufLen then begin
      AsciiBufLen  := AsciiBufLen + BUF_INCR;
      SetLength(AsciiBufStr, AsciiBufLen);
    end;
    Move(PChar(ASCII)^, (@AsciiBufStr[AsciiBufOffset])^, l);
    AsciiBufOffset := AsciiBufOffset + l;

    SetLength(AsciiBufStr, AsciiBufOffset);

    Clipboard.SetTextBuf(PChar(AsciiBufStr));


  finally

    FreeMem(BufferMPX);
    Free2(TPMChar(Buffer));

  end;

end;

function TQUB_QFS_QDF.prCopyBinary(First, Last :Integer; DoOnePoint :Boolean) :Boolean;
var
  CInfo  :TQUB_QFS_QDF_PCopyInfo;
  Mem    :Integer;
  Size   :Integer;
  cCount :Integer;
  dCount :Integer;
  sCount :Integer;
  SegCnt :Integer;
  r      :Integer;

  function ReadSegInfo(First, Last :Integer; Segs :TPAInteger) :Integer;
  var
    f, l, ff, ll, s, c :Integer;
  begin
    f := 0;
    c := 0;
    for s := 0 to prSegments.SegmentCount - 1 do begin
      l := f + prSegments[s].Channels.DataCount - 1;
      //if ((f >= First) and (f <= Last)) or ((l >= First) and (l <= Last)) then begin
      if not ((First > l) or (Last < f)) then begin
        if Segs <> nil then begin
          //ff := f;
          //ll := l;
          //if ff < First then ff := First;
          //if ll > Last  then ll := Last;
          ff := First;
          ll := Last;
          if ff < f then ff := f;
          if ll > l then ll := l;
          Segs[c] := ll - ff + 1;
        end;
        Inc(c);
      end;
      f := l + 1;
      if f > Last then Break;
    end;
    Result := c;
  end;

begin

  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  Result := ((First = Last) and (not DoOnePoint));
  if Result then Exit;

  //Is the Clipboard Format Registered ?
  Result := ClipboardFormat = 0;
  if Result then Exit;

  //Open the Clipboard
  Result := not OpenClipboard(Application.Handle);
  if Result then Exit;

  try

    //New: copy the segmentation info. LM

    cCount := SizeOf(TQUB_QFS_QDF_CopyInfo);
    dCount := ADDataSize * ADChannelCount * (Last - First + 1);
    SegCnt := ReadSegInfo(First, Last, nil);
    sCount := SizeOf(Integer)*SegCnt;

    //Allocate memory
    Size := cCount + dCount + sCount;
    Mem  := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Size);
    if Mem = 0{NULL} then begin
    //Memory allocation failure
      Result := True;
      Exit;
    end;

    CInfo := GlobalLock(Mem);

    try

      CInfo.ChannelCount  := ADChannelCount;
      CInfo.DataType      := ADDataType;
      CInfo.DataSize      := ADDataSize;
      CInfo.DataCount     := Last - First + 1;
      CInfo.SegCount      := SegCnt;
      FillChar(CInfo.ListName, Length(CInfo.ListName), 0);
      CInfo.ListName      := '';

      //Get Data
      Result := ReadData(@TPAByte(CInfo)[cCount], True{Raw}, -1{Channel}, First, Last);
      if not Result then begin
        Result := ReadSegInfo(First, Last, @TPAByte(CInfo)[cCount + dCount]) = 0;
      end;

    finally

      GlobalUnlock(Mem);

    end;

    if Result then begin
      GlobalFree(Mem);
      Exit;
    end;

    r := SetClipboardData(ClipboardFormat, Mem);

    if r = 0{NULL} then begin
      Result := True;
      GlobalFree(Mem);
      Exit;
    end;

  finally

    CloseClipboard;

  end;
end;

//Copy List
function TQUB_QFS_QDF.prCopyList(List :TQUB_QVO_QVD_SelectionList; ListName :String; CopyASCII :Boolean) :Boolean;
begin
  if CopyASCII then begin
    Result := prCopyListASCII(List, ListName);
  end else begin
    Result := prCopyListBinary(List, ListName);
  end;
end;

//Copy the list as it is shown
function TQUB_QFS_QDF.prCopyListASCII(List :TQUB_QVO_QVD_SelectionList; ListName :String) :Boolean;
const
  BUF_INCR        = ONE_MEGABYTE div 8;
var
  ASCII       :String;
  AsciiBufStr     :String;
  AsciiBufLen     :Integer;
  AsciiBufOffset  :Integer;

  Sel         :TQUB_FirstLast;
  MaxCount    :Integer;
  TotCount        :Integer;
  Count       :Integer;
  CCount          :Integer;
  Value       :Double;
  c, li, p, l     :Integer;
begin

  Result      := False;
  if prQVDParent = nil then Exit;

  try

    List.Lock;

    try

      //Find the longest selection in the list
      MaxCount := 0;
      TotCount := 0;
      for li := 0 to List.Items.Count - 1 do begin
        Sel := List.Selection[li].Sel;
        Count := Sel.Last - Sel.First + 1;
        Inc(TotCount, Count);
        if Count > MaxCount then MaxCount := Count;
      end;

      CCount := 0;
      for c := 0 to ADChannelCount - 1 do begin
        if TQUB_QVO_QVD(prQVDParent).ChannelVisible[c] then Inc(CCount);
      end;

      AsciiBufOffset  := 1;//0;Strings are 1-based
      AsciiBufStr     := '';
      AsciiBufLen     := TotCount * CCount * 16;
      SetLength(AsciiBufStr, AsciiBufLen);

      CCount := ADChannelCount;

      //For each sample
      for p := 0 to MaxCount - 1 do begin
        ASCII := Format('%.6g', [p * TQUB_QVO_QVD(prQVDParent).Sampling]) + #9;
        //For each channel
        for c := 0 to CCount - 1 do begin
          if TQUB_QVO_QVD(prQVDParent).ChannelVisible[c] then begin
            //For each selection in the list
            for li := 0 to List.Items.Count - 1 do begin
              Sel := List.Selection[li].Sel;
              Count := Sel.Last - Sel.First + 1;
              //if (Count - 1) <= MaxCount then begin
              if p < Count then begin
              //Sel is not short
                Value := TQUB_QVO_QVD(prQVDParent).DataChannelDisplay[c, p + Sel.First - TQUB_QVO_QVD(prQVDParent).DataDraw.First];
                ASCII := ASCII + Format('%.6g', [Value/TQUB_QVO_QVD(prQVDParent).Scaling/TQUB_QVO_QVD(prQVDParent).DataChannelScaling[c]]) + #9;
              end else begin
              //Sel is short, done
                ASCII := ASCII + ' ' + #9;
              end;
            end;
          end;
        end;
        ASCII := ASCII + #13#10;

        l := Length(ASCII);
        if (AsciiBufOffset + l) > AsciiBufLen then begin
          AsciiBufLen  := AsciiBufLen + BUF_INCR;
          SetLength(AsciiBufStr, AsciiBufLen);
        end;
        Move(PChar(ASCII)^, (@AsciiBufStr[AsciiBufOffset])^, l);
        AsciiBufOffset := AsciiBufOffset + l;

      end;


      ASCII := #0;
      l := Length(ASCII);
      if (AsciiBufOffset + l) > AsciiBufLen then begin
        AsciiBufLen  := AsciiBufLen + BUF_INCR;
        SetLength(AsciiBufStr, AsciiBufLen);
      end;
      Move(PChar(ASCII)^, (@AsciiBufStr[AsciiBufOffset])^, l);
      AsciiBufOffset := AsciiBufOffset + l;

      SetLength(AsciiBufStr, AsciiBufOffset);

      Clipboard.SetTextBuf(PChar(AsciiBufStr));


    except

      Result := True;

    end;

  finally

    List.Unlock;

  end;

end;

function TQUB_QFS_QDF.prCopyListBinary(List :TQUB_QVO_QVD_SelectionList; ListName :String) :Boolean;
var
  CInfo  :TQUB_QFS_QDF_PCopyInfo;
  Mem    :Integer;
  Size   :Integer;
  cCount :Integer;
  dCount :Integer;
  sCount :Integer;
  SegCnt :Integer;
  Sel    :TQUB_FirstLast;
  DatCnt :Integer;
  dCnt   :Integer;
  l, r   :Integer;

begin

  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  Result := List = nil;
  if Result then Exit;

  //Is the Clipboard Format Registered ?
  Result := ClipboardFormat = 0;
  if Result then Exit;

  //Open the Clipboard
  Result := not OpenClipboard(Application.Handle);
  if Result then Exit;

  try

    List.Lock;

    cCount := SizeOf(TQUB_QFS_QDF_CopyInfo);
    DatCnt := 0;
    for l := 0 to List.Items.Count - 1 do begin
      Sel := List.Selection[l].Sel;
      Inc(DatCnt, Sel.Last - Sel.First + 1);
    end;
    dCount := ADDataSize * ADChannelCount * DatCnt;
    SegCnt := List.Items.Count;
    sCount := SizeOf(Integer)*SegCnt;

    //Allocate memory
    Size := cCount + dCount + sCount;
    Mem  := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, Size);
    if Mem = 0{NULL} then begin
    //Memory allocation failure
      Result := True;
      Exit;
    end;

    CInfo := GlobalLock(Mem);

    try

      CInfo.ChannelCount  := ADChannelCount;
      CInfo.DataType      := ADDataType;
      CInfo.DataSize      := ADDataSize;
      CInfo.DataCount     := DatCnt;
      ListName            := ListName + '_Pasted';
      FillChar(CInfo.ListName[0], SizeOf(CInfo.ListName), 0);
      Move(PChar(ListName)[0], CInfo.ListName[0], Min(99, Length(ListName)));
      CInfo.SegCount      := SegCnt;

      dCnt := 0;
      for l := 0 to List.Items.Count - 1 do begin
        Sel := List.Selection[l].Sel;
        //Get Data
        Result := ReadData(@TPAByte(CInfo)[cCount + dCnt], True{Raw}, -1{Channel}, Sel.First, Sel.Last);
        Inc(dCnt, ADChannelCount * ADDataSize * (Sel.Last - Sel.First + 1));
        TPAInteger(@TPAByte(CInfo)[cCount + dCount])[l] := (Sel.Last - Sel.First + 1);
      end;

    finally

      GlobalUnlock(Mem);

    end;

    if Result then begin
      GlobalFree(Mem);
      Exit;
    end;

    r := SetClipboardData(ClipboardFormat, Mem);

    if r = 0{NULL} then begin
      Result := True;
      GlobalFree(Mem);
      Exit;
    end;

  finally

    CloseClipboard;
    List.Unlock;

  end;
end;

//Paste from clipboard
function TQUB_QFS_QDF.prPaste(First, Last :Integer; PasteAsSeg :Boolean; ReplaceSel :Boolean) :Boolean;
var
  OperCount   :Integer;
  Operation   :TQUB_QFS_EditOperation;

  OperAddSeg  :TQUB_QFS_PEditOperationAddSegment;
  Segment     :TQUB_QFS_Segment;

  OperAddSel  :TQUB_QFS_PEditOperationAddSelection;
  Selection   :TQUB_QFS_PEditSelection;

  OperInsSeg  :TQUB_QFS_PEditOperationInsertSeg;
  OperInsInto :TQUB_QFS_PEditOperationInsertIntoSeg;

  CInfo       :TQUB_QFS_QDF_PCopyInfo;
  Mem         :Integer;
  SelPos      :Integer;
  SelIdx      :Integer;
  SegIdx      :Integer;
  s, f, l     :Integer;
  Sel         :TQUB_FirstLast;

  cCount      :Integer;
  dCount      :Integer;

  ListName    :String;
  SelName     :String;

begin

  //Replace this with paste as segment (?!)
  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  //Is the Clipboard Format Registered ?
  Result := ClipboardFormat = 0;
  if Result then Exit;

  //Is the Clipboard Format Available ?
  Result := not IsClipboardFormatAvailable(ClipboardFormat);
  if Result then Exit;

  //OpenClipboard
  Result := not OpenClipboard(Application.Handle);
  if Result then Exit;

  //Retrieve Data Handle
  Mem := GetClipboardData(ClipboardFormat);
  if Mem = 0 then begin
    Result := True;
    CloseClipboard;
    Exit;
  end;

  try

    CInfo := GlobalLock(Mem);

    //Check data format
    if (CInfo.ChannelCount <> ADChannelCount) or (CInfo.DataType <> ADDataType) then begin
      Result := True;
      Console.NewEntry('Cannot paste data from clipboard - different number of channels or different data type', COLOR_ERROR);
      Exit;
    end;

    //This should go into Write Data &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

    //Write Data into EditStream
    try
      prReadWriteCritSection.Acquire;
      SelPos := prEditStream.Size;
      prEditStream.Position := SelPos;
      try
        prEditStream.WriteBuffer(TPAByte(CInfo)[SizeOf(TQUB_QFS_QDF_CopyInfo)], CInfo.ChannelCount * CInfo.DataCount * CInfo.DataSize);
        Sleep(5); // ?
      except
        Result := True;
        Exit;
      end;
    finally
      prReadWriteCritSection.Release;
    end;

    OperCount := prEditOperations.Count;

    //Before Paste, Delete Selection, if any !!!
    if not PasteAsSeg then begin
      Result := prDelete(First, Last, False{DoOnePoint}, nil{DelOperation}, False{AllowEmptySeg});
      if Result then Exit;
    end;

    if prEditOperations.Count > OperCount then begin
      Operation := prEditOperations[prEditOperations.Count -1];
    end else begin
      Operation := TQUB_QFS_EditOperation.Create;
      prEditOperations.Add(Operation);
    end;

    //Find where to insert the new selection
    SegIdx := -1;
    SelIdx := -1;
    f := 0;
    for s := 0 to prSegments.SegmentCount - 1 do begin
      l := f + prSegments[s].Channels.DataCount - 1;
      if (First >= f) and (First <= l) then begin
      //Found the Segment
        SegIdx := s;
        if not PasteAsSeg then SelIdx := prSegments[s].Channels.WhichSelectionForPoint(First - f, True{IncIfLast});
        //Found the Selection
        Break;
      end;
      f := l + 1;
    end;

    if SegIdx = -1 then begin
      if PasteAsSeg then begin
        SegIdx := prSegments.SegmentCount;
      end else begin
        SegIdx := prSegments.SegmentCount - 1;
        SelIdx := prSegments[SegIdx].Channels.WhichSelectionForPoint(-1, True{IncIfLast});
      end;
    end;

    if PasteAsSeg then begin
    //Segment(s)

      cCount := SizeOf(TQUB_QFS_QDF_CopyInfo);
      dCount := CInfo.DataSize * CInfo.ChannelCount * CInfo.DataCount;

      if (First = prSegments.TotalDataCount - 1) then begin
        SegIdx := prSegments.SegmentCount;
        Inc(First); //This is new --- 01-27-08 LM
      end;

      ListName := CInfo.ListName;
      if ListName <> '' then begin
        if Assigned(OnAddSeries) then begin
          OnAddSeries(Self, ListName, False{Private});
        end;
      end;

      for s := 0 to CInfo.SegCount - 1 do begin

        //Create new Selection
        Selection                        := AllocMem(SizeOf(TQUB_QFS_EditSelection));
        Selection.DataCount              := TPAInteger(@TPAByte(CInfo)[cCount + dCount])[s];
        Selection.StreamPos              := SelPos;
        Selection.FromEditStream         := True;

        Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
        Segment.Stream                   := Stream;
        Segment.StreamPos                := Stream.Position;
        Segment.Channels.CanEdit         := CanEdit;
        Segment.Channels.EditStream      := EditStream;
        Segment.Channels.Stream          := Stream;
        Segment.Channels.StreamPos       := Stream.Size + 1;
        Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
        Segment.Channels.DataType        := ADDataType;
        Segment.Channels.DataCount       := 0;//it is empty, since the data is brought by the Selection
        Segment.Channels.DataStreamPos   := Stream.Size + 1;

        //Insert Selection in Segment
        Segment.Channels.InsertSelection(Selection, 0);//Now DataCount becomes <> 0 (Selection.DataCount

        //Insert Segment in Segments
        prSegments.AddRecord(Segment, SegIdx);

        //Create operation AddSeg
        OperAddSeg                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSegment));
        OperAddSeg.OperType       := OPER_TYPE_ADD_SEGMENT;
        OperAddSeg.Segment        := Segment;
        OperAddSeg.SegmentIndex   := SegIdx;
        Operation.Add(OperAddSeg);

        //Create operation for Info
        OperInsSeg                := AllocMem(SizeOf(TQUB_QFS_EditOperationInsertSeg));
        OperInsSeg.OperType       := OPER_TYPE_INS_ISEG;
        OperInsSeg.First          := First;
        OperInsSeg.Last           := First + Selection.DataCount - 1;
        Operation.Add(OperInsSeg);
        Sel                       := QUB_FirstLast(OperInsSeg.First, OperInsSeg.Last);
        prInfoEdit(EDIT_OPER_INSERT_SEG, Sel, QUB_QFS_PNodeMem_Null);

        if ListName <> '' then begin
          SelName                   := IntToStr(s);
          //Add Selection - use an event mechanism to communicate with Parent QVD, if any
          if Assigned(OnAddSweep) then begin
            OnAddSweep(Self, ListName, SelName, Sel, True{ForceAdd});
          end;
        end;

        SelPos := SelPos + Segment.Channels.DataCount*ADChannelCount*ADDataSize;
        Inc(SegIdx);
        Inc(First, Selection.DataCount);

      end;//SegCount

    end else begin
    //Selection

      //Create new Selection
      Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount       := CInfo.DataCount;
      Selection.StreamPos       := SelPos;
      Selection.FromEditStream  := True;

      //Insert Selection in Segment
      prSegments[SegIdx].Channels.InsertSelection(Selection, SelIdx);//DataCount is augmented by Selection.DataCount

      //Create operation AddSel
      OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
      OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
      OperAddSel.Selection      := Selection;
      OperAddSel.SelectionIndex := SelIdx;
      OperAddSel.EditRecord     := prSegments[SegIdx].Channels;
      Operation.Add(OperAddSel);

      //Create operation for Info
      OperInsInto               := AllocMem(SizeOf(TQUB_QFS_EditOperationInsertIntoSeg));
      OperInsInto.OperType      := OPER_TYPE_INS_INTO_ISEG;
      OperInsInto.First         := First;
      OperInsInto.Last          := First + Selection.DataCount - 1;
      Operation.Add(OperInsInto);
      Sel                       := QUB_FirstLast(OperInsInto.First, OperInsInto.Last);
      prInfoEdit(EDIT_OPER_INSERT_INTOSEG, Sel, QUB_QFS_PNodeMem_Null);

    end;

  finally

    GlobalUnlock(Mem);
    CloseClipboard;

  end;
end;

{
I made a small change: I added the optional DelOperation, so prDelete can be used in conjunction with other edit operations
If AllowEmptySeg = True, then empty segments are not deleted after data were deleted - also to be used with other edit operations - EXCLUSIVELY!!!
}
function TQUB_QFS_QDF.prDelete(First, Last :Integer; DoOnePoint :Boolean; DelOperation :TQUB_QFS_EditOperation; AllowEmptySeg :Boolean) :Boolean;
var
  Deleted       :Boolean;
  Operation     :TQUB_QFS_EditOperation;
  OperDelSeg    :TQUB_QFS_PEditOperationDelSegment;
  OperDeleteSeg :TQUB_QFS_PEditOperationDeleteSeg;
  OperDeleteSel :TQUB_QFS_PEditOperationDeleteSel;
  OperBreak     :TQUB_QFS_PEditOperationBreakSeg;
  DynamicOffset :Integer; // for the last two operations, once you've deleted stuff before [f,l] you have to offset [f,l] by that much
  Sel           :TQUB_FirstLast;
  f, l          :Integer;
  cf, cl        :Integer;
  s             :Integer;
begin

  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  Deleted  := False;

  Operation := nil;

  DynamicOffset := 0;

  try

    if DelOperation <> nil then begin
      Operation := DelOperation;
    end else begin
      Operation := TQUB_QFS_EditOperation.Create;
    end;

    f := 0;
    s := 0;
    repeat

      //calculate the last point
      l := f + prSegments[s].Channels.DataCount - 1;

      //Segment s'th has (some) data within First to Last range
      if not ((Last < f) or (First > l)) then begin

        cf := f;
        if First > cf then cf := First;
        cl := l;
        if Last < cl  then cl := Last;

        if (cf = f) and (cl = l) then begin
        //Delete the whole segment

          if (prSegments.SegmentCount = 1) or (AllowEmptySeg) then begin
          //Never delete the only segment - just clear it
          //Also, don't delete any segment if AllowEmptySeg = True

            Result := prSegments[s].Channels.DeleteSelection(cf - f, cl - f, Operation);
            if not Result then Deleted := True;

            OperDeleteSel           := AllocMem(SizeOf(TQUB_QFS_EditOperationDeleteSel));
            OperDeleteSel.OperType  := OPER_TYPE_DEL_ISEL;
            OperDeleteSel.First     := cf - DynamicOffset;
            OperDeleteSel.Last      := cl - DynamicOffset;
            OperDeleteSel.Data      := NODE_Create( PChar('') );
            Operation.Add(OperDeleteSel);
            Sel                     := QUB_FirstLast(OperDeleteSel.First, OperDeleteSel.Last);
            prInfoEdit(EDIT_OPER_DEL_SEL, Sel, OperDeleteSel.Data);

            DynamicOffset           := DynamicOffset + (cl - cf + 1);

            Inc(s);

          end else begin

            OperDelSeg              := AllocMem(SizeOf(TQUB_QFS_EditOperationDelSegment));
            OperDelSeg.OperType     := OPER_TYPE_DEL_SEGMENT;
            OperDelSeg.Segment      := prSegments[s];
            OperDelSeg.SegmentIndex := s;

            Operation.Add(OperDelSeg);

            //Info operation
            OperDeleteSeg           := AllocMem(SizeOf(TQUB_QFS_EditOperationDeleteSeg));
            OperDeleteSeg.OperType  := OPER_TYPE_DEL_ISEG;
            OperDeleteSeg.First     := cf - DynamicOffset;
            OperDeleteSeg.Last      := cl - DynamicOffset;
            OperDeleteSeg.Data      := NODE_Create( PChar('') );
            Operation.Add(OperDeleteSeg);
            Sel                     := QUB_FirstLast(OperDeleteSeg.First, OperDeleteSeg.Last);
            prInfoEdit(EDIT_OPER_DEL_SEG, Sel, OperDeleteSeg.Data);

            DynamicOffset           := DynamicOffset + (cl - cf + 1);

            prSegments.DelRecord(prSegments[s]);

            Deleted := True;

          end;

        end else begin
        //Delete just a selection of the segment

          if (First = Last) and (not DoOnePoint) then begin
          //Break Selection
            Result := prSegments[s].Channels.BreakSelection(cf - f, Operation);
          end else begin
          //Delete Selection
            Result := prSegments[s].Channels.DeleteSelection(cf - f, cl - f, Operation);

            if cf <> f then begin

              OperBreak               := AllocMem(SizeOf(TQUB_QFS_EditOperationBreakSeg));
              OperBreak.OperType      := OPER_TYPE_BREAK;
              OperBreak.Point         := cf - DynamicOffset;
              Operation.Add(OperBreak);
              Sel                     := QUB_FirstLast(OperBreak.Point, OperBreak.Point);
              prInfoEdit(EDIT_OPER_IBREAK, Sel, Sel);

            end;

            if cl <> l then begin

              OperBreak               := AllocMem(SizeOf(TQUB_QFS_EditOperationBreakSeg));
              OperBreak.OperType      := OPER_TYPE_BREAK;
              OperBreak.Point         := cl - DynamicOffset + 1;
              Operation.Add(OperBreak);
              Sel                     := QUB_FirstLast(OperBreak.Point, OperBreak.Point);
              prInfoEdit(EDIT_OPER_IBREAK, Sel, Sel);

            end;

            OperDeleteSeg           := AllocMem(SizeOf(TQUB_QFS_EditOperationDeleteSeg));
            OperDeleteSeg.OperType  := OPER_TYPE_DEL_ISEG;
            OperDeleteSeg.First     := cf - DynamicOffset;
            OperDeleteSeg.Last      := cl - DynamicOffset;
            OperDeleteSeg.Data      := NODE_Create( PChar('') );
            Operation.Add(OperDeleteSeg);
            Sel                     := QUB_FirstLast(OperDeleteSeg.First, OperDeleteSeg.Last);
            prInfoEdit(EDIT_OPER_DEL_SEG, Sel, OperDeleteSeg.Data);

            DynamicOffset           := DynamicOffset + (cl - cf + 1);

            // now,  if cf <> f and cl <> l then infoedit: joinat cf?
            // since it appears to break off the piece to be deleted as its own seg,
            // leaving a phony seg-break.
            // unfortunately this will also require writing the undo for join etc etc and i'd have to relearn the whole system

          end;

          if not Result then Deleted := True;

          Inc(s);

        end;

        if Result then Exit;

      end else begin

        Inc(s);

      end;

      f := l + 1;

    until ((s >= prSegments.SegmentCount) or (f > Last));

  finally

    if DelOperation = nil then begin
    //else is managed by the caller
      if Deleted then begin
        prEditOperations.Add(Operation);
      end else begin
        Operation.Free;
      end;
    end;

  end;
end;

function TQUB_QFS_QDF.prErase(First, Last :Integer; BufFirst, BufLast :Integer; DoOnePoint :Boolean) :Boolean;
var
  Buffer     :Pointer;
  Size       :Integer;
  SelPos     :Integer;
  Count      :Integer;
  OperCount  :Integer;

  Operation  :TQUB_QFS_EditOperation;
  OperAddSel :TQUB_QFS_PEditOperationAddSelection;
  Selection  :TQUB_QFS_PEditSelection;

  OperInsInto :TQUB_QFS_PEditOperationInsertIntoSeg;

  SelIdx     :Integer;
  SegIdx     :Integer;
  s, f, l    :Integer;

  Sel        :TQUB_FirstLast;

begin

  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  Buffer := nil;

  try

    //Alloc Memory
    Size := BufLast - BufFirst + 1;
    if Size > ONE_MEGABYTE then Size := ONE_MEGABYTE;
    if Size > (Last - First + 1) then Size := (Last - First + 1);
    BufLast := BufFirst + Size - 1;
    Buffer := AllocMem(ADDataSize * ADChannelCount * Size);

    Result := ReadData(Buffer, True{Raw}, -1{Channel},  BufFirst, BufLast);
    if Result then Exit;

    //This should go into Write Data &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

    //Write Erase Buffer into EditStream
    try
      prReadWriteCritSection.Acquire;
      SelPos := prEditStream.Size;
      prEditStream.Position := SelPos;
      Count := 0;
      while (Count < (Last - First + 1)) do begin
        try
          prEditStream.WriteBuffer(Buffer^, ADDataSize * ADChannelCount * Size);
          Sleep(5); // ?
        except
          prEditStream.Size := SelPos;
          Result := True;
          Exit;
        end;
        Inc(Count, Size);
        if Size > ((Last - First + 1) - Count) then Size := (Last - First + 1) - Count;
      end;
    finally
      prReadWriteCritSection.Release;
    end;

    //Now Delete Selection

    OperCount := prEditOperations.Count;

    //Before Paste, Delete Selection
    Result := prDelete(First, Last, DoOnePoint, nil{DelOperation}, False{AllowEmptySeg});
    if Result then Exit;

    if prEditOperations.Count > OperCount then begin
      Operation := prEditOperations[prEditOperations.Count -1];
    end else begin
      Operation := TQUB_QFS_EditOperation.Create;
      prEditOperations.Add(Operation);
    end;

    //Find where to insert the new selection
    SegIdx := -1;
    SelIdx := -1;
    f := 0;
    for s := 0 to prSegments.SegmentCount - 1 do begin
      l := f + prSegments[s].Channels.DataCount - 1;
      if (First >= f) and (First <= l) then begin
      //Found the Segment
        SegIdx := s;
        SelIdx := prSegments[s].Channels.WhichSelectionForPoint(First - f, True{IncIfLast});
        //Found the Selection
        Break;
      end;
      f := l + 1;
    end;

    if SegIdx = -1 then begin
      SegIdx := prSegments.SegmentCount - 1;
      SelIdx := prSegments[SegIdx].Channels.WhichSelectionForPoint(-1, True{IncIfLast});
    end;

    //Create new Selection
    Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
    Selection.DataCount       := Last - First + 1;
    Selection.StreamPos       := SelPos;
    Selection.FromEditStream  := True;

    //Insert Selection in Segment
    prSegments[SegIdx].Channels.InsertSelection(Selection, SelIdx);//DataCount is augmented by Selection.DataCount

    //Create operation AddSel
    OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
    OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
    OperAddSel.Selection      := Selection;
    OperAddSel.SelectionIndex := SelIdx;
    OperAddSel.EditRecord     := prSegments[SegIdx].Channels;
    Operation.Add(OperAddSel);

    //Info operation
    OperInsInto               := AllocMem(SizeOf(TQUB_QFS_EditOperationInsertIntoSeg));
    OperInsInto.OperType      := OPER_TYPE_INS_INTO_ISEG;
    OperInsInto.First         := First;
    OperInsInto.Last          := Last;
    Operation.Add(OperInsInto);
    Sel                       := QUB_FirstLast(OperInsInto.First, OperInsInto.Last);
    prInfoEdit(EDIT_OPER_INSERT_INTOSEG, Sel, QUB_QFS_PNodeMem_Null);

  finally

    FreeMem(Buffer);

  end;

end;

function TQUB_QFS_QDF.prFill(First, Last :Integer; Channel :Integer; Value :Variant) :Boolean;
var
  Buffer      :Pointer;
  BufferI     :TPMInteger;
  BufferSize  :Integer;

  Operation   :TQUB_QFS_EditOperation;

  OperAddSel  :TQUB_QFS_PEditOperationAddSelection;
  Selection   :TQUB_QFS_PEditSelection;
  OperInsInto :TQUB_QFS_PEditOperationInsertIntoSeg;

  SegCount    :Integer;

  OldPos      :Integer;
  SelPos      :Integer;
  SelIdx      :Integer;
  s, f, l, p  :Integer;
  Sel         :TQUB_FirstLast;

begin

  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  //Read Seg Info
  SegCount    := prSegments.SegmentCount;

  Buffer      := nil;
  BufferI     := nil;

  try

    BufferSize := Last - First + 1;
    if BufferSize > ONE_MEGABYTE then BufferSize := ONE_MEGABYTE;

    Buffer  := AllocMem(ADChannelCount * ADDataSize * BufferSize);
    BufferI := IAlloc2(ADChannelCount, BufferSize);

    //Change and Write data to edit stream
    try
      prReadWriteCritSection.Acquire;
      OldPos := prEditStream.Position;
      try
        try
          SelPos := prEditStream.Size;
          prEditStream.Position := SelPos;

          f := First;
          l := f + BufferSize - 1;
          if l > Last then l := Last;

          repeat

            Result := ReadData(Buffer, True{Raw}, -1{Channel}, f, l);
            if Result then Exit;

            if MPXData(BufferI, Buffer, ADChannelCount, -1, ADDataSize, ADDataType, f-f, l-f, False) then Exit;
            for s := f - f to l - f do begin
              BufferI[Channel, s] := Value;
            end;
            if MPXData(BufferI, Buffer, ADChannelCount, -1, ADDataSize, ADDataType, f-f, l-f, True)  then Exit;

            prEditStream.WriteBuffer(Buffer^, ADChannelCount * ADDataSize * (l - f + 1));
            Sleep(5);

            f := l + 1;

          until(f > Last);

        except
          Result := True;
          Exit;
        end;
      finally
        prEditStream.Position := OldPos;
      end;
    finally
      prReadWriteCritSection.Release;
    end;


    //Create Edit Operation
    Operation := TQUB_QFS_EditOperation.Create;

    //Delete old data and insert new one - all the while keeping the segmentation
    f := 0;
    for s := 0 to SegCount - 1 do begin

      //calculate the last point
      l := f + prSegments[s].Channels.DataCount - 1;
      p := f;

      if ((First >= f) and (First <= l)) or ((Last >= f) and (Last <= l)) then begin
      //Found a Segment that contains (some of) the data to be filled

        if f < First then f := First;
        if l > Last  then l := Last;

        SelIdx := prSegments[s].Channels.WhichSelectionForPoint(f - p, False{IncIfLast});
        if f > p then Inc(SelIdx);
        Result := prDelete(f, l, True{DoOnePoint}, Operation, True{AllowEmptySeg});

        //Create new Selection
        Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
        Selection.DataCount       := l - f + 1;
        Selection.StreamPos       := SelPos + (f - First) * ADChannelCount * ADDataSize;
        Selection.FromEditStream  := True;

        //Insert Selection in Segment
        prSegments[s].Channels.InsertSelection(Selection, SelIdx);

        //Create operation AddSel
        OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
        OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
        OperAddSel.Selection      := Selection;
        OperAddSel.SelectionIndex := SelIdx;
        OperAddSel.EditRecord     := prSegments[s].Channels;
        Operation.Add(OperAddSel);

        //Info operation
        OperInsInto               := AllocMem(SizeOf(TQUB_QFS_EditOperationInsertIntoSeg));
        OperInsInto.OperType      := OPER_TYPE_INS_INTO_ISEG;
        OperInsInto.First         := f;
        OperInsInto.Last          := l;
        Operation.Add(OperInsInto);
        Sel                       := QUB_FirstLast(OperInsInto.First, OperInsInto.Last);
        prInfoEdit(EDIT_OPER_INSERT_INTOSEG, Sel, QUB_QFS_PNodeMem_Null);

      end else if f > Last then begin

        Break;

      end;

      f := l + 1;

    end;

    Result := False;

  finally

    FreeMem(Buffer);
    Free2(TPMChar(BufferI));

    prEditOperations.Add(Operation);

  end;

end;

function TQUB_QFS_QDF.prBreak(First, Last :Integer) :Boolean;
var
  Segment     :TQUB_QFS_Segment;
  Selection   :TQUB_QFS_PEditSelection;
  SegIdx      :Integer;
  SelIdx      :Integer;
  Operation   :TQUB_QFS_EditOperation;
  OperAddSeg  :TQUB_QFS_PEditOperationAddSegment;
  OperAddSel  :TQUB_QFS_PEditOperationAddSelection;
  OperBreak   :TQUB_QFS_PEditOperationBreakSeg;
  SegCount    :Integer;
  ChDataCount :Integer;
  f, l, s     :Integer;
  Sel         :TQUB_FirstLast;
begin

  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  //Read Seg Info
  SegCount    := prSegments.SegmentCount;
  ChDataCount := prSegments[SegCount - 1].Channels.DataCount;

  //Find the Segment
  f := 0;
  SegIdx := 0;
  repeat

    //calculate the last point
    if SegIdx = (SegCount - 1) then begin
      l := f + ChDataCount - 1;
    end else begin
      l := f + prSegments[SegIdx].Channels.DataCount - 1;
    end;

    Result := ((First = f) or (First = l));
    if Result then Exit;

    //Segment s'th has First within
    if (First >= f) and (First <= l) then Break;

    Inc(SegIdx);

    f := l + 1;

  until ((SegIdx >= SegCount) or (f > First));

  Result := SegIdx > SegCount;
  if Result then Exit;

  Operation := TQUB_QFS_EditOperation.Create;

  try

    //Break the Selection
    Result := prSegments[SegIdx].Channels.BreakSelection(First - f, Operation);
    if Result then Exit;

    //Create a new Segment
    try
      prReadWriteCritSection.Acquire;
      Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      Segment.Stream                   := prSegments[SegIdx].Stream;
      Segment.StreamPos                := Stream.Position;
      Segment.Channels.CanEdit         := CanEdit;
      Segment.Channels.EditStream      := EditStream;
      Segment.Channels.Stream          := prSegments[SegIdx].Channels.Stream;
      Segment.Channels.StreamPos       := Stream.Size + 1;
      Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
      Segment.Channels.DataType        := ADDataType;
      Segment.Channels.DataCount       := 0;
      Segment.Channels.DataStreamPos   := Stream.Size + 1;
    finally
      prReadWriteCritSection.Release;
    end;

    //Create operation AddSeg
    OperAddSeg                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSegment));
    OperAddSeg.OperType       := OPER_TYPE_ADD_SEGMENT;
    OperAddSeg.Segment        := Segment;
    OperAddSeg.SegmentIndex   := SegIdx + 1;
    Operation.Add(OperAddSeg);

    //Find first Selection to be moved
    SelIdx := prSegments[SegIdx].Channels.WhichSelectionForPoint(First - f, True{IncIfLast});

    //Create new Selections and insert them in Segment
    f := 0;
    for s := SelIdx to prSegments[SegIdx].Channels.Selections.Count - 1 do begin

      //Create Selection
      Selection  := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection^ := prSegments[SegIdx].Channels.Selections[s]^;
      Segment.Channels.InsertSelection(Selection, s - SelIdx);//DataCount is augmented by Selection.DataCount

      //Create operation AddSel
      OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
      OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
      OperAddSel.Selection      := Selection;
      OperAddSel.SelectionIndex := s - SelIdx;
      OperAddSel.EditRecord     := TQUB_QFS_Edit(Segment.Channels);
      Operation.Add(OperAddSel);

      Inc(f, Selection.DataCount);
    end;

    //Delete Selections from split Segment
    f := prSegments[SegIdx].Channels.DataCount - f;
    l := prSegments[SegIdx].Channels.DataCount - 1;
    Result := prSegments[SegIdx].Channels.DeleteSelection(f, l, Operation);

    //Insert Segment in Segments
    prSegments.AddRecord(Segment, SegIdx + 1);

    OperBreak                 := AllocMem(SizeOf(TQUB_QFS_EditOperationBreakSeg));
    OperBreak.OperType        := OPER_TYPE_BREAK;
    OperBreak.Point           := First;
    Operation.Add(OperBreak);
    Sel                       := QUB_FirstLast(First, First);
    prInfoEdit(EDIT_OPER_IBREAK, Sel, Sel);

  finally

    if Operation.Count <> 0 then begin
      prEditOperations.Add(Operation);
    end else begin
      Result := True;
      Operation.Free;
    end;

  end;

end;

function TQUB_QFS_QDF.prEditPoint(Channel :Integer; Point :Integer; Change :Integer) :Boolean;
var
  Buffer      :Pointer;
  BufferI     :TPMInteger;

  Operation   :TQUB_QFS_EditOperation;

  OperAddSel  :TQUB_QFS_PEditOperationAddSelection;
  Selection   :TQUB_QFS_PEditSelection;
  OperInsInto :TQUB_QFS_PEditOperationInsertIntoSeg;

  SegCount    :Integer;
  ChDataCount :Integer;

  Pnt         :Integer;
  OldPos      :Integer;
  SelPos      :Integer;
  SelIdx      :Integer;
  SegIdx      :Integer;
  s, f, l     :Integer;
  Sel         :TQUB_FirstLast;
begin

  Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  //Read Seg Info
  SegCount    := prSegments.SegmentCount;
  ChDataCount := prSegments[SegCount - 1].Channels.DataCount;

  Buffer      := nil;
  BufferI     := nil;

  Pnt         := 0;

  try

    //Read Data from stream
    Buffer  := AllocMem(ADChannelCount * ADDataSize * 1);
    BufferI := IAlloc2(ADChannelCount, 1);

    //Find where to insert the new selection
    SegIdx := -1;
    f := 0;
    for s := 0 to SegCount - 1 do begin

      //calculate the last point
      if s = (SegCount - 1) then begin
        l := f + ChDataCount - 1;
      end else begin
        l := f + prSegments[s].Channels.DataCount - 1;
      end;

      if (Point >= f) and (Point <= l) then begin
      //Found the Segment
        SegIdx := s;
        Pnt    := Point - f;
        Break;
      end;
      f := l + 1;

    end;

    if SegIdx = -1 then Exit;

    Result := ReadData(Buffer, True{Raw}, -1{Channel}, Point, Point);
    if Result then Exit;

    if MPXData(BufferI, Buffer, ADChannelCount, -1, ADDataSize, ADDataType, 0, 0, False) then Exit;
    BufferI[Channel, 0] := BufferI[Channel, 0] + Change;
    if MPXData(BufferI, Buffer, ADChannelCount, -1, ADDataSize, ADDataType, 0, 0, True) then Exit;

    //Write changed data to edit stream
    try
      prReadWriteCritSection.Acquire;
      OldPos := prEditStream.Position;
      try
        try
          SelPos := prEditStream.Size;
          prEditStream.Position := SelPos;
          prEditStream.WriteBuffer(Buffer^, ADChannelCount * 1 * ADDataSize);
          Sleep(5); // ?
        except
          Result := True;
          Exit;
        end;
      finally
        prEditStream.Position := OldPos;
      end;
    finally
      prReadWriteCritSection.Release;
    end;

    //Delete Old Point
    Result := prDelete(Point, Point, True{DoOnePoint}, nil{DelOperation}, False{AllowEmptySeg}); //Change it?!
    if Result then Exit;

    if prEditOperations.Count = 0 then Exit;

    SelIdx := prSegments[SegIdx].Channels.WhichSelectionForPoint(Pnt, False{IncIfLast});

    Operation := prEditOperations[prEditOperations.Count - 1];

    //Create new Selection
    Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
    Selection.DataCount       := 1;
    Selection.StreamPos       := SelPos;
    Selection.FromEditStream  := True;

    //Insert Selection in Segment
    prSegments[SegIdx].Channels.InsertSelection(Selection, SelIdx);

    //Create operation AddSel
    OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
    OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
    OperAddSel.Selection      := Selection;
    OperAddSel.SelectionIndex := SelIdx;
    OperAddSel.EditRecord     := prSegments[SegIdx].Channels;
    Operation.Add(OperAddSel);

    //Info operation
    OperInsInto               := AllocMem(SizeOf(TQUB_QFS_EditOperationInsertIntoSeg));
    OperInsInto.OperType      := OPER_TYPE_INS_INTO_ISEG;
    OperInsInto.First         := Point;
    OperInsInto.Last          := Point;
    Operation.Add(OperInsInto);
    Sel                       := QUB_FirstLast(OperInsInto.First, OperInsInto.Last);
    prInfoEdit(EDIT_OPER_INSERT_INTOSEG, Sel, QUB_QFS_PNodeMem_Null);

    Result := False;

  finally

    FreeMem(Buffer);
    Free2(TPMChar(BufferI));

  end;

end;

function TQUB_QFS_QDF.prUndo :Boolean;
var
  Operation     :TQUB_QFS_EditOperation;
  OperBase      :TQUB_QFS_PEditOperationBase;
  OperDelSeg    :TQUB_QFS_PEditOperationDelSegment;
  OperAddSeg    :TQUB_QFS_PEditOperationAddSegment;
  OperDelSel    :TQUB_QFS_PEditOperationDelSelection;
  OperAddSel    :TQUB_QFS_PEditOperationAddSelection;
  OperInsSeg    :TQUB_QFS_PEditOperationInsertSeg;
  OperInsInto   :TQUB_QFS_PEditOperationInsertIntoSeg;
  OperDeleteSeg :TQUB_QFS_PEditOperationDeleteSeg;
  OperDeleteSel :TQUB_QFS_PEditOperationDeleteSel;
  OperBreak     :TQUB_QFS_PEditOperationBreakSeg;
  EditRecord    :TQUB_QFS_Edit;
  o             :Integer;
  Sel           :TQUB_FirstLast;
begin

  Result := ((prEditOperations.Count = 0) or (not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
  if Result then Exit;

  Operation := prEditOperations[prEditOperations.Count - 1];

  try

    for o := Operation.Count - 1 downto 0 do begin

      OperBase := Operation[o];

      case OperBase.OperType of

        //Data
        OPER_TYPE_DEL_SEGMENT:
           begin

             OperDelSeg := TQUB_QFS_PEditOperationDelSegment(OperBase);
             prSegments.AddRecord(TQUB_QFS_Segment(OperDelSeg.Segment), OperDelSeg.SegmentIndex);
             FreeMem(OperDelSeg);

           end;

        //Data
        OPER_TYPE_ADD_SEGMENT:
           begin

             OperAddSeg := TQUB_QFS_PEditOperationAddSegment(OperBase);
             prSegments.DelRecord(OperAddSeg.Segment);
             FreeMem(OperAddSeg);

           end;

        //Data
        OPER_TYPE_DEL_SELECTION:
           begin

             //Selection was removed, now has to be moved back
             OperDelSel := TQUB_QFS_PEditOperationDelSelection(OperBase);
             EditRecord := OperDelSel.EditRecord;
             EditRecord.InsertSelection(OperDelSel.Selection, OperDelSel.SelectionIndex);
             FreeMem(OperDelSel);

           end;

        //Data
        OPER_TYPE_ADD_SELECTION:
           begin

             //Selection was added, now has to be removed
             OperAddSel := TQUB_QFS_PEditOperationAddSelection(OperBase);
             EditRecord := OperAddSel.EditRecord;
             EditRecord.RemoveSelection(OperAddSel.SelectionIndex);
             FreeMem(OperAddSel.Selection);
             FreeMem(OperAddSel);

           end;

        //Info
        OPER_TYPE_INS_ISEG:
           begin
             OperInsSeg := TQUB_QFS_PEditOperationInsertSeg(OperBase);
             Sel        := QUB_FirstLast(OperInsSeg.First, OperInsSeg.Last);
             prInfoEdit(EDIT_OPER_DEL_SEG, Sel, QUB_QFS_PNodeMem_Null);
             FreeMem(OperInsSeg);

           end;

        //Info
        OPER_TYPE_INS_INTO_ISEG:
           begin
             OperInsInto := TQUB_QFS_PEditOperationInsertIntoSeg(OperBase);
             Sel         := QUB_FirstLast(OperInsInto.First, OperInsInto.Last);
             prInfoEdit(EDIT_OPER_DEL_SEL, Sel, QUB_QFS_PNodeMem_Null);
             FreeMem(OperInsInto);

           end;

        //Info
        OPER_TYPE_DEL_ISEG:
           begin
             OperDeleteSeg  := TQUB_QFS_PEditOperationDeleteSeg(OperBase);
             Sel            := QUB_FirstLast(OperDeleteSeg.First, OperDeleteSeg.Last);
             prInfoEdit(EDIT_OPER_INSERT_SEG, Sel, OperDeleteSeg.Data);
             FreeMem(OperDeleteSeg);

           end;

        //Info
        OPER_TYPE_DEL_ISEL:
           begin
             OperDeleteSel  := TQUB_QFS_PEditOperationDeleteSel(OperBase);
             Sel            := QUB_FirstLast(OperDeleteSel.First, OperDeleteSel.Last);
             prInfoEdit(EDIT_OPER_INSERT_INTOSEG, Sel, OperDeleteSel.Data);
             FreeMem(OperDeleteSel);

           end;

        //Data + Info (?!)
        OPER_TYPE_BREAK:
           begin
             OperBreak      := TQUB_QFS_PEditOperationBreakSeg(OperBase);
             Sel            := QUB_FirstLast(OperBreak.Point, OperBreak.Point);
             prInfoEdit(EDIT_OPER_JOIN, Sel, Sel);
             FreeMem(OperBreak);

           end;

      end;//case OperType

      Operation.Delete(o);

    end;//for Operation.Count

  finally

    //Don't forget to resize the stream !!!

    prEditOperations.Delete(prEditOperations.Count - 1);
    Operation.Free;

  end;

end;

function TQUB_QFS_QDF.prAppendData(Buffer :Pointer; BufferSize, BufferCount :Integer; BufferIsWritten :Boolean; BufferAtPos :Integer; NewSegment :Boolean; StartTime :Double) :Boolean;
var
  Operation   :TQUB_QFS_EditOperation;

  OperAddSeg  :TQUB_QFS_PEditOperationAddSegment;
  Segment     :TQUB_QFS_Segment;

  OperAddSel  :TQUB_QFS_PEditOperationAddSelection;
  Selection   :TQUB_QFS_PEditSelection;

  OperInsSeg  :TQUB_QFS_PEditOperationInsertSeg;
  OperInsInto :TQUB_QFS_PEditOperationInsertIntoSeg;

  SelPos      :Integer;
  SelIdx      :Integer;
  SegIdx      :Integer;

  OldPos      :Integer;
  First       :Integer;
  Sel         :TQUB_FirstLast;
begin

  //Check data format
  Result := (BufferSize <> (ADDataSize * ADChannelCount));
  if Result then Exit;

  try

    prReadWriteCritSection.Acquire;

    //Write Data into EditStream, unless already written
    if not BufferIsWritten then begin

      OldPos     := prEditStream.Position;
      try
        try
          SelPos := prEditStream.Size;
          prEditStream.Position := SelPos;
          prEditStream.WriteBuffer(Buffer^, BufferSize * BufferCount);
          Sleep(5); // ?
        except
          Result := True;
          Exit;
        end;
      finally
        prEditStream.Position := OldPos;
      end;

    end else begin

      SelPos     := BufferAtPos;

    end;

    First     := prSegments.TotalDataCount;

    Operation := TQUB_QFS_EditOperation.Create;
    prEditOperations.Add(Operation);

    //Create new Selection
    if (NewSegment) and (prSegments[prSegments.SegmentCount - 1].Channels.DataCount > 0) then begin
    //Segment

      SegIdx                           := prSegments.SegmentCount;

      Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
      Segment.Stream                   := Stream;
      Segment.StreamPos                := Stream.Position;
      Segment.Channels.CanEdit         := CanEdit;
      Segment.Channels.EditStream      := EditStream;
      Segment.Channels.Stream          := Stream;
      Segment.Channels.StreamPos       := SelPos;
      Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
      Segment.Channels.DataType        := ADDataType;
      Segment.Channels.DataCount       := 0;//Data is contained in selection
      Segment.Channels.DataStreamPos   := SelPos;

      Segment.StartTime := StartTime;

      //Insert Selection in Segment
      Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount       := BufferCount;
      Selection.StreamPos       := SelPos;
      Selection.FromEditStream  := True;
      Segment.Channels.InsertSelection(Selection, Segment.Channels.Selections.Count);

      //Insert Segment in Segments
      prSegments.AddRecord(Segment, SegIdx);

      //Create operation AddSeg
      OperAddSeg                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSegment));
      OperAddSeg.OperType       := OPER_TYPE_ADD_SEGMENT;
      OperAddSeg.Segment        := Segment;
      OperAddSeg.SegmentIndex   := SegIdx;
      Operation.Add(OperAddSeg);

      //Info operation
      OperInsSeg                := AllocMem(SizeOf(TQUB_QFS_EditOperationInsertSeg));
      OperInsSeg.OperType       := OPER_TYPE_INS_ISEG;
      OperInsSeg.First          := First;
      OperInsSeg.Last           := First + BufferCount - 1;
      Operation.Add(OperInsSeg);
      Sel                       := QUB_FirstLast(OperInsSeg.First, OperInsSeg.Last);
      prInfoEdit(EDIT_OPER_INSERT_SEG, Sel, QUB_QFS_PNodeMem_Null);

    end else begin
    //Selection

      SegIdx                    := prSegments.SegmentCount - 1;
      SelIdx                    := prSegments[SegIdx].Channels.Selections.Count;

      //Insert Selection in Segment
      Selection                 := AllocMem(SizeOf(TQUB_QFS_EditSelection));
      Selection.DataCount       := BufferCount;
      Selection.StreamPos       := SelPos;
      Selection.FromEditStream  := True;
      prSegments[SegIdx].Channels.InsertSelection(Selection, SelIdx);

      if NewSegment then begin
        //prSegments[SegIdx].StartTime := 0;
        prSegments[SegIdx].StartTime := StartTime;
      end;

      //Create operation AddSel
      OperAddSel                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSelection));
      OperAddSel.OperType       := OPER_TYPE_ADD_SELECTION;
      OperAddSel.Selection      := Selection;
      OperAddSel.SelectionIndex := SelIdx;
      OperAddSel.EditRecord     := prSegments[SegIdx].Channels;
      Operation.Add(OperAddSel);

      //Info operation
      OperInsInto               := AllocMem(SizeOf(TQUB_QFS_EditOperationInsertIntoSeg));
      OperInsInto.OperType      := OPER_TYPE_INS_INTO_ISEG;
      OperInsInto.First         := First;
      OperInsInto.Last          := First + BufferCount - 1;
      Operation.Add(OperInsInto);
      Sel                       := QUB_FirstLast(OperInsInto.First, OperInsInto.Last);
      prInfoEdit(EDIT_OPER_INSERT_INTOSEG, Sel, QUB_QFS_PNodeMem_Null);

    end;

  finally

    prReadWriteCritSection.Release;

  end;

end;

function TQUB_QFS_QDF.prNewSegment(StartTime :Double; ForceNewIfEmpty :Boolean) :Boolean;
var
  Operation   :TQUB_QFS_EditOperation;

  OperAddSeg  :TQUB_QFS_PEditOperationAddSegment;
  Segment     :TQUB_QFS_Segment;

  SelPos      :Integer;
  SegIdx      :Integer;

begin

  Result      := False;

  try

    prReadWriteCritSection.Acquire;

    //Last segment is empty, no need to insert a new one
    if (not ForceNewIfEmpty) and (prSegments.SegmentCount > 0) and (prSegments.Segment[prSegments.SegmentCount - 1].Channels.DataCount = 0) then Exit;

    SelPos    := prEditStream.Size;

    Operation := TQUB_QFS_EditOperation.Create;
    prEditOperations.Add(Operation);

    SegIdx                           := prSegments.SegmentCount;

    Segment                          := TQUB_QFS_Segment.Create(QUB_QFS_QDF_SEGMENT_NAME, nil);
    Segment.Stream                   := Stream;
    Segment.StreamPos                := Stream.Position;
    Segment.Channels.CanEdit         := CanEdit;
    Segment.Channels.EditStream      := EditStream;
    Segment.Channels.Stream          := Stream;
    Segment.Channels.StreamPos       := SelPos;
    Segment.Channels.DataSize        := ADChannelCount * ADDataSize;
    Segment.Channels.DataType        := ADDataType;
    Segment.Channels.DataCount       := 0;//Data is contained in selection
    Segment.Channels.DataStreamPos   := SelPos;

    Segment.StartTime := StartTime;

    //Insert Segment in Segments
    prSegments.AddRecord(Segment, SegIdx);

    //Create operation AddSeg
    OperAddSeg                := AllocMem(SizeOf(TQUB_QFS_EditOperationAddSegment));
    OperAddSeg.OperType       := OPER_TYPE_ADD_SEGMENT;
    OperAddSeg.Segment        := Segment;
    OperAddSeg.SegmentIndex   := SegIdx;
    Operation.Add(OperAddSeg);

  finally

    prReadWriteCritSection.Release;

  end;

end;



//This function cannot be undone. It is meant for situations when data
//comes separately for different channels. The first Channel will Append
// - which can be undone, and the rest will replace.
//it is the complementary - sort of - of LoadData
function TQUB_QFS_QDF.prReplaceData(Buffer :Pointer; BufferSize :Integer; Sel :TQUB_FirstLast) :Boolean;
var
  SegCount    :Integer;
  ChDataCount :Integer;
  f, l        :Integer;
  cf, cl      :Integer;
  s           :Integer;
begin

  //Read data information
  try
    prSegInfoCritSection.Acquire;
    Result := ((not prOpened) or (prSegments = nil) or (prSegments.SegmentCount = 0));
    if Result then Exit;
    SegCount    := prSegments.SegmentCount;
    ChDataCount := prSegments[SegCount - 1].Channels.DataCount;
  finally
    prSegInfoCritSection.Release;
  end;

  try

    //Acquire ReadWrite Critical Section
    prReadWriteCritSection.Acquire;

    f := 0;
    s := 0;

    repeat

      //calculate the last point
      if s = (SegCount - 1) then begin
      //If it is the last segment use the value it had at the time of entry
      //(which may change due to appending)
        l := f + ChDataCount - 1;
      end else begin
        l := f + prSegments[s].Channels.DataCount - 1;
      end;

      //Segment s'th has (some) data within First to Last range
      if not ((Sel.Last < f) or (Sel.First > l)) then begin

        cf := f;
        if Sel.First > cf then cf := Sel.First;
        cl := l;
        if Sel.Last < cl  then cl := Sel.Last;

        Result := prSegments[s].Channels.ReplaceData(@TPAByte(Buffer)[prSegments[s].Channels.DataSize * (cf - Sel.First)], cf - f, cl - f);
        if Result then Exit;

      end;

      Inc(s);

      f := l + 1;

    until ((s = SegCount) or (f > Sel.Last));

  finally

    //Release ReadWrite Critical Section
    prReadWriteCritSection.Release;

  end;

end;

function TQUB_QFS_QDF.prClear :Boolean;
begin

  Result := False;

  try

    prReadWriteCritSection.Acquire;

    try
      prEditOperations.Clear;
      prEditStream.Size := 0;
    except
      Result := True;
    end;

  finally

    prReadWriteCritSection.Release;

  end;

end;


//LDT compatibility functions **************************************************
function TQUB_QFS_QDF.prGetLDTSampling :Word;
begin
  Result := Round(Sampling * 1e6);
end;

function TQUB_QFS_QDF.prGetLDTScaling :Word;
var
  IntScaling      :LongWord;
begin
  IntScaling      := Round(Scaling * 100);
  while IntScaling > 32767 do begin
    IntScaling    := IntScaling div 2;
  end;
  Result          := Word(IntScaling);
end;

function TQUB_QFS_QDF.prGetLDTLeader :Integer;
begin
  if prFileType = QFS_QDF_TYPE_LDT then begin
    Result := prLDTLeader;
  end else begin
    Result := 0;
  end;
end;

//WAV compatibility functions
function TQUB_QFS_QDF.prGetWAVBitsPerSample :Integer;
begin
  if prFileType = QFS_QDF_TYPE_WAV then begin
    Result := prWAVBitsPerSample;
  end else begin
    Result := 8 * ADDataSize;
  end;
end;

function TQUB_QFS_QDF.prGetWAVBlockAlign :Integer;
begin
  Result := ADChannelCount * ADDataSize;
end;

function TQUB_QFS_QDF.prGetWAVAvgBytesPerSec :Integer;
begin
  Result := Round(1 / Sampling) * WAVBlockAlign;
end;


//Can's ************************************************************************
function TQUB_QFS_QDF.prGetCanCut :Boolean;
begin
  Result := (CanCopy) and (CanDelete);
end;

function TQUB_QFS_QDF.prGetCanCopy :Boolean;
begin
  Result := (ClipboardFormat <> 0) and (prSegments.TotalDataCount > 0) and (not LockedReadWrite);
end;

function TQUB_QFS_QDF.prGetCanPaste :Boolean;
begin
  Result := (CanEdit) and (not LockedAny) and (ClipboardFormat <> 0) and (IsClipboardFormatAvailable(ClipboardFormat));
  if Result then begin
    Result := OpenClipboard(Application.Handle);
    if Result then begin
      Result := GetClipboardData(ClipboardFormat) <> 0;
      CloseClipboard;
    end;
  end;
end;

function TQUB_QFS_QDF.prGetCanAppend :Boolean;
begin
  Result := (CanEdit);
end;

function TQUB_QFS_QDF.prGetCanDelete :Boolean;
begin
  Result := (CanEdit) and (prSegments.TotalDataCount > 0) and (not LockedAny);
end;

function TQUB_QFS_QDF.prGetCanErase :Boolean;
begin
  Result := (CanEdit) and (prSegments.TotalDataCount > 0) and (not LockedAny);
end;

function TQUB_QFS_QDF.prGetCanFill :Boolean;
begin
  Result := (CanEdit) and (prSegments.TotalDataCount > 0) and (not LockedAny);
end;

function TQUB_QFS_QDF.prGetCanBreak :Boolean;
begin
  Result := (CanEdit) and (prSegments.TotalDataCount > 0) and (not LockedAny);
end;

function TQUB_QFS_QDF.prGetCanUndo :Boolean;
begin
  Result := (CanEdit) and (prEditOperations.Count > 0) and (not LockedAny);
end;

//Sampling / Scaling
procedure TQUB_QFS_QDF.prSetSampling(s :Double);
var
  i :Integer;
begin
  prRecSampling.DataValue := s;
  for i := 0 to QUB_QFS_QDF_MAX_ADCHANNELCOUNT - 1 do begin
    if prIdealization[i] <> nil then prIdealization[i].Sampling := Sampling;
  end;
end;

function TQUB_QFS_QDF.prGetSampling :Double;
begin
  Result := prRecSampling.DataValue;
end;

procedure TQUB_QFS_QDF.prSetScaling(s :Double);
begin
  prRecScaling.DataValue := s;
end;

function TQUB_QFS_QDF.prGetScaling :Double;
begin
  Result := prRecScaling.DataValue;
end;

//ADChannels
procedure TQUB_QFS_QDF.prSetADChannelCount(c :Integer);
begin
  prRecADChannelCount.DataValue := c;
end;

function TQUB_QFS_QDF.prGetADChannelCount :Integer;
begin
  Result := prRecADChannelCount.DataValue;
end;

procedure TQUB_QFS_QDF.prSetADDataSize(s :Integer);
begin
  prRecADDataSize.DataValue := s;
end;

function TQUB_QFS_QDF.prGetADDataSize :Integer;
begin
  Result := prRecADDataSize.DataValue;
end;

procedure TQUB_QFS_QDF.prSetADDataType(t :TQUB_QFS_Node_DataType);
begin
  prRecADDataType.DataValue := Integer(t);
end;
                                                                              
function TQUB_QFS_QDF.prGetADDataType :TQUB_QFS_Node_DataType;
begin
  Result := TQUB_QFS_Node_DataType(prRecADDataType.DataValue);
end;


//Data Info
function TQUB_QFS_QDF.prGetIdealization(Index :Integer) :TQUB_IdlSegments;
begin
  if (Index >= 0) and (Index < QUB_QFS_QDF_MAX_ADCHANNELCOUNT) then Result := prIdealization[Index] else Result := nil;
end;


//New File Properties
procedure TQUB_QFS_QDF.OnGetPropertiesNew(Sender :TObject);
begin
  prFileName              := TQUB_Viewport_Data_New_PropertiesForm(Sender).FileName;
  prFileType              := TQUB_Viewport_Data_New_PropertiesForm(Sender).FileFormat;
  ADChannelCount          := TQUB_Viewport_Data_New_PropertiesForm(Sender).ADChannelCount;
  ADDataSize              := TQUB_Viewport_Data_New_PropertiesForm(Sender).ADDataSize;
  //TO DO: Elaborate on this. LM
  if ADDataSize = 2 then ADDataType := NODE_DATATYPE_SMALLINT else ADDataType := NODE_DATATYPE_LONGINT;
  try Sampling            := 1 / (TQUB_Viewport_Data_New_PropertiesForm(Sender).Sampling * 1000); except end;
  Scaling                 := TQUB_Viewport_Data_New_PropertiesForm(Sender).Scaling;
  //TO DO: This was duplicated for a valid but now forgotten reason. Check if still necessary. LM
  ADChannelCount          := TQUB_Viewport_Data_New_PropertiesForm(Sender).ADChannelCount;
end;

procedure TQUB_QFS_QDF.OnPutPropertiesNew(Sender :TObject);
begin
end;


function DoAxoMsgOut(Msg :PChar; Data :Pointer): Integer; stdcall;
begin
  TQUB_QFS_QDF(Data).DoMsgOut(Msg);
  Result := 0;
end;

procedure TQUB_QFS_QDF.DoMsgOut(Msg :PChar);
begin
  Console.NewEntry(Msg, COLOR_ERROR);
end;


//******************************************************************************
initialization
begin
  //Register Clipboard Format
  ClipboardFormat := RegisterClipboardFormat(PChar(ClipboardFormatName));
end;

finalization
begin
  //Clear Clipboard
  if ClipboardFormat <> 0 then begin
    if IsClipboardFormatAvailable(ClipboardFormat) then begin
      if OpenClipboard(Application.Handle) then begin
        SetClipboardData(ClipboardFormat, 0);
        CloseClipboard;
      end;
    end;
  end;
end;


//******************************************************************************
end.

