{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}
// the Extract button's task

//05-03-08 Added P/n functionality. LM

//11-26-08 Added functionality to include kinetic parameters in the data, and some other minor changes. LM

// 03-21-09 cleaned up accumulated messes with scalings, function_fit, sourcesample, ...
//          added file list ability as independent flag to e.g. sum congruent lists
//          no more direct prData access due to file list

unit QUB_Task_ExtractSelection;


// change join to a processing step so all kinpars etc are included
// warn about different scaling


interface


uses


Windows, SysUtils, Classes, Dialogs, Forms, Math,
AllocTypes, Alloc,

QUB_Global,
QUB_Globals,
QUB_QVO_QVD, QUB_QVO_QVD_Global,

QUB_Thread,
QUB_Task,
QUB_Task_Data,

QUB_QFS_File, QUB_QFS_QDF, QUB_QFS_QDF_Global, QUB_QFS_QDF_MPX, QUB_QFS_QDF_WAV,
QUB_Node,

QUB_Task_ExtractSelection_Properties,
QUB_Task_Exp_Properties,

QUB_QFS_QIF_Segments,

QUB_Extract_Global;


type


TQUB_Task_ExtractSelection  = class(TQUB_Task)
protected


  prData                                 :TQUB_QVO_QVD;
  prSelectionList                        :TQUB_QVO_QVD_SelectionList;
  prSubSelectionList                     :TQUB_QVO_QVD_SelectionList;

  prFileName                             :String;

  prOpenFileAfter                        :Boolean;

  //Data Source
  prDataSource                           :TQUB_Task_Data_Source;
  prExtractFileList                      :Boolean;

  //Data Processed
  prDataProcessed                        :TQUB_Task_Data_Processed;
  prProcessFilter                        :Boolean;
  prProcessFilterFreq                    :Extended;
  prProcessBaseLine                      :Boolean;
  prProcessResample                      :Boolean;
  prProcessResampleFactor                :Extended;
  prProcOnlyActiveChannel                :Boolean;

  //File Format
  prFileFormat                           :TQUB_QFS_QDF_FileType;
  prASCIITime                            :Boolean;
  prFloat4Byte                           :Boolean;
  prDropFirstLast                        :Boolean;
  prDropIfClass                          :Boolean;
  prDropClass                            :Integer;

  //Other Options
  prChannels                             :TQUB_Task_ExtractSelection_Channels;
  prChannelList                          :TList;
  prJoinSegments                         :Boolean;
  prASCIIFit                             :Boolean;
  prASCIISegCol                          :Boolean;
  prKinPars                              :Boolean;
  prKinParsAll                           :Boolean;
  prKinParsList                          :TStringList;
  prKinParCount                          :Integer;

  //Operation
  prOperation                            :TQUB_Task_ExtractSelection_Operation;
  prOperationList                        :Boolean;
  prOperationListName                    :String;
  prOperationFiles                       :Boolean;
  prOperationScale                       :Boolean;
  prOperationScaleFactor                 :Double;

  //Function
  prFunctionDrift                        :Boolean;
  prFunctionDriftA                       :Double;
  prFunctionDriftB                       :Double;
  prFunctionSine                         :Boolean;
  prFunctionSineFrq                      :Double;//Hz
  prFunctionSineAmp                      :Double;
  prFunctionSinePh                       :Double;//Deg
  prFunctionDigitize                     :Boolean;
  prFunctionDigitizeThreshold            :Double;
  prFunctionDigitizeToMin                :Double;
  prFunctionDigitizeToMax                :Double;
  prFunctionDerivate                     :Boolean;
  prFunctionLKC                          :Boolean;
  prFunctionItoG                         :Boolean;
  prFunctionItoGVRev                     :Double;
  prFunctionItoGVCh                      :Integer;
  prFunctionRsCorr                       :Boolean;
  prFunctionRsValue                      :Double;
  prFunctionRsPer                        :Double;
  prFunctionRsAddLeak                    :Boolean;
  prFunctionRsLeakCh                     :Integer;
  prFunctionFit                          :Boolean;
  prFunctionPnAdd                        :Boolean;
  prFunctionPnSub                        :Boolean;
  prFunctionPnOffset                     :Double;
  prFunctionPnDelay                      :Double;
  prFunctionPnScale                      :Double;
  prFunctionPnCount                      :Integer;
  prFunctionPnVm                         :Double;
  prFunctionPnInterval                   :Double;

  prLKCVChannel                          :Integer; // copied from Exp properties

  //Chunks
  prChunks                               :TList;


  procedure prSetData(Data :TQUB_QVO_QVD);

  procedure prDoExtractSelection;
  procedure prDoExtractIdealized;

  function prChunkData(QVD :TQUB_QVO_QVD) :Boolean;
  function prChunkSubData(QVD :TQUB_QVO_QVD) :Boolean;
  function prChunkLKCData(QVD :TQUB_QVO_QVD) :Boolean;
  procedure prFreeChunks;

  function prJoinIdlSegments(DS :INode): INode;

  //procedure prResample(Buffer, PrevBufferLastPoint, PrevBufferLastValue :TPMInteger; BufferR :TPAInteger; ChannelCount :Integer; var CountR :Integer; var ResampledPoint :Integer; Sel :TQUB_FirstLast; Sampling, Factor :Extended);


public


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;

  procedure Execute; override;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;

  property Data                          :TQUB_QVO_QVD                          read prData                  write prSetData;
  property SelectionList                 :TQUB_QVO_QVD_SelectionList            read prSelectionList;
  property SubSelectionList              :TQUB_QVO_QVD_SelectionList            read prSubSelectionList;

  property FileName                      :String                                read prFileName              write prFileName;

  property OpenFileAfter                 :Boolean                               read prOpenFileAfter         write prOpenFileAfter;

  //Data Source
  property DataSource                    :TQUB_Task_Data_Source                 read prDataSource            write prDataSource;
  property ExtractFileList               :Boolean                               read prExtractFileList       write prExtractFileList;

  //Data Processed
  property DataProcessed                 :TQUB_Task_Data_Processed              read prDataProcessed         write prDataProcessed;
  property ProcessFilter                 :Boolean                               read prProcessFilter         write prProcessFilter;
  property ProcessFilterFreq             :Extended                              read prProcessFilterFreq     write prProcessFilterFreq;
  property ProcessBaseLine               :Boolean                               read prProcessBaseLine       write prProcessBaseLine;
  property ProcessResample               :Boolean                               read prProcessResample       write prProcessResample;
  property ProcessResampleFactor         :Extended                              read prProcessResampleFactor write prProcessResampleFactor;
  property ProcOnlyActiveChannel         :Boolean                               read prProcOnlyActiveChannel write prProcOnlyActiveChannel;

  //File Format
  property FileFormat                    :TQUB_QFS_QDF_FileType                 read prFileFormat            write prFileFormat;
  property ASCIITime                     :Boolean                               read prASCIITime             write prASCIITime;
  property Float4Byte                    :Boolean                               read prFloat4Byte            write prFloat4Byte;
  property DropFirstLast                 :Boolean                               read prDropFirstLast         write prDropFirstLast;
  property DropIfClass                   :Boolean                               read prDropIfClass           write prDropIfClass;
  property DropClass                     :Integer                               read prDropClass             write prDropClass;

  //Channels, Segments
  property Channels                      :TQUB_Task_ExtractSelection_Channels   read prChannels              write prChannels;
  property ChannelList                   :TList                                 read prChannelList;
  property ASCIIFit                      :Boolean                               read prASCIIFit              write prASCIIFit;
  property ASCIISegCol                   :Boolean                               read prASCIISegCol           write prASCIISegCol;
  property JoinSegments                  :Boolean                               read prJoinSegments          write prJoinSegments;
  property KinPars                       :Boolean                               read prKinPars               write prKinPars;
  property KinParsAll                    :Boolean                               read prKinParsAll            write prKinParsAll;
  property KinParsList                   :TStringList                           read prKinParsList           write prKinParsList;

  //Operation
  property Operation                     :TQUB_Task_ExtractSelection_Operation  read prOperation             write prOperation;
  property OperationList                 :Boolean                               read prOperationList         write prOperationList;
  property OperationListName             :String                                read prOperationListName     write prOperationListName;
  property OperationFiles                :Boolean                               read prOperationFiles        write prOperationFiles;
  property OperationScale                :Boolean                               read prOperationScale        write prOperationScale;
  property OperationScaleFactor          :Double                                read prOperationScaleFactor  write prOperationScaleFactor;

  //Function
  property FunctionDrift                 :Boolean                               read prFunctionDrift         write prFunctionDrift;
  property FunctionDriftA                :Double                                read prFunctionDriftA        write prFunctionDriftA;
  property FunctionDriftB                :Double                                read prFunctionDriftB        write prFunctionDriftB;
  property FunctionSine                  :Boolean                               read prFunctionSine          write prFunctionSine;
  property FunctionSineFrq               :Double                                read prFunctionSineFrq       write prFunctionSineFrq;
  property FunctionSineAmp               :Double                                read prFunctionSineAmp       write prFunctionSineAmp;
  property FunctionSinePh                :Double                                read prFunctionSinePh        write prFunctionSinePh;
  property FunctionDigitize              :Boolean                               read prFunctionDigitize      write prFunctionDigitize;
  property FunctionDigitizeThreshold     :Double                                read prFunctionDigitizeThreshold write prFunctionDigitizeThreshold;
  property FunctionDigitizeToMin         :Double                                read prFunctionDigitizeToMin write prFunctionDigitizeToMin;
  property FunctionDigitizeToMax         :Double                                read prFunctionDigitizeToMax write prFunctionDigitizeToMax;
  property FunctionDerivate              :Boolean                               read prFunctionDerivate      write prFunctionDerivate;
  property FunctionLKC                   :Boolean                               read prFunctionLKC           write prFunctionLKC;
  property FunctionItoG                  :Boolean                               read prFunctionItoG          write prFunctionItoG;
  property FunctionItoGVRev              :Double                                read prFunctionItoGVRev      write prFunctionItoGVRev;
  property FunctionItoGVCh               :Integer                               read prFunctionItoGVCh       write prFunctionItoGVCh;
  property FunctionRsCorr                :Boolean                               read prFunctionRsCorr        write prFunctionRsCorr;
  property FunctionRsValue               :Double                                read prFunctionRsValue       write prFunctionRsValue;
  property FunctionRsPer                 :Double                                read prFunctionRsPer         write prFunctionRsPer;
  property FunctionRsAddLeak             :Boolean                               read prFunctionRsAddLeak     write prFunctionRsAddLeak;
  property FunctionRsLeakCh              :Integer                               read prFunctionRsLeakCh      write prFunctionRsLeakCh;
  property FunctionFit                   :Boolean                               read prFunctionFit           write prFunctionFit;
  property FunctionPnAdd                 :Boolean                               read prFunctionPnAdd         write prFunctionPnAdd;
  property FunctionPnSub                 :Boolean                               read prFunctionPnSub         write prFunctionPnSub;
  property FunctionPnOffset              :Double                                read prFunctionPnOffset      write prFunctionPnOffset;
  property FunctionPnDelay               :Double                                read prFunctionPnDelay       write prFunctionPnDelay;
  property FunctionPnScale               :Double                                read prFunctionPnScale       write prFunctionPnScale;
  property FunctionPnCount               :Integer                               read prFunctionPnCount       write prFunctionPnCount;
  property FunctionPnVm                  :Double                                read prFunctionPnVm          write prFunctionPnVm;
  property FunctionPnInterval            :Double                                read prFunctionPnInterval    write prFunctionPnInterval;
  property LKCVChannel                   :Integer                               read prLKCVChannel           write prLKCVChannel;

end;


//******************************************************************************

{
The following are for extraction of data (not idealization).
The main procedure in prDoExtractSelection is to read from a _DataSet
and write to a _File in buffer-size chunks.

The DataSet describes the data as it will be in the extracted file -- post-processed.
Several steps of processing are accomplished by having one DataSet subclass
responsible for each step -- loading data from a "Source" DataSet and modifying it,
while presenting appropriately modified Session, SegmentCount, etc.

The base DataSets are _Chunks or _Chunks_Joined.  They interpret the Task's Chunks
and request data from QVO_QVD.  In terms of options, they handle processing
(except resampling), joining segments, and selecting channels.

The rest are constructed with a Source :_DataSet.  In this way we build a
processing chain, and the file is written with data from the final link in the chain.

On buffers:  only one DataSet in the chain can safely obliterate Buf.Alt.  This
role is played by Add, Sub, Avg, Std, of which only one can be chosen.

}

ArrayOfTQUB_FirstLast = Array of TQUB_FirstLast;

TQUB_SrcSegCount          = record
  Src                   :Integer;
  Seg                   :Integer;
  Count                 :Integer;
end;
TQUB_PSrcSegCount         = ^TQUB_SrcSegCount;
TQUB_ASrcSegCount         = Array[0..(MaxInt div (SizeOf(TQUB_SrcSegCount)*2))] of TQUB_SrcSegCount;
TQUB_PASrcSegCount        = ^TQUB_ASrcSegCount;
ArrayOfTQUB_SrcSegCount = Array of TQUB_SrcSegCount;


TQUB_Task_Extract_DataSet                = class
protected
  prData                  :TQUB_QVO_QVD;

  function prGetChannelCount: Integer; virtual; abstract;
  function prGetSegmentCount: Integer; virtual; abstract;
  function prGetSegmentStart(Ix :Integer): Double; virtual; abstract;
  function prGetSegmentLength(Ix :Integer): Integer; virtual; abstract;

public
  Session                             :INode;
  ActiveChannel                       :Integer;
  Sampling                            :Double;

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; virtual; abstract;

  property ChannelCount               :Integer     read prGetChannelCount;
  property SegmentCount               :Integer     read prGetSegmentCount;
  property SegmentStart[Ix :Integer]  :Double      read prGetSegmentStart;
  property SegmentLength[Ix :Integer] :Integer     read prGetSegmentLength;
end;

TQUB_Task_Extract_DataSet_Chunks        = class(TQUB_Task_Extract_DataSet)
// handles:  process (except resample), channel selection, fit curve
protected
  //prData                  :TQUB_QVO_QVD;
  prChunks                :TList;
  prChannels              :ArrayOfInteger;
  prFitCurve              :Integer;
  prTotalCount            :Integer;
  prJoinSels              :ArrayOfTQUB_FirstLast;
  prKinParsCount          :Integer;
  prKinPars               :TStringList;
  prTaskChannels          :TQUB_Task_ExtractSelection_Channels;
  prDataProcessed         :TQUB_Task_Data_Processed;
  prProcOnlyActive        :Boolean;
  prProcessFilter         :Boolean;
  prProcessBaseline       :Boolean;
  prActiveChannel         :Integer;

  function prSetupIdealization: INode;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

  function prGetChunk(Ix :Integer): TQUB_Thread_PChunk;
  function prGetChunkCount: Integer;

  property Chunks[Ix :Integer] :TQUB_Thread_PChunk  read prGetChunk;
  property ChunkCount          :Integer             read prGetChunkCount;

public
  constructor Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Chunks :TList);
  destructor Destroy; override;

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

TQUB_Task_Extract_DataSet_Chunks_Joined = class(TQUB_Task_Extract_DataSet)
// handles:  process (except resample), join, channel selection
protected
  //prData                  :TQUB_QVO_QVD;
  prChunks                :TList;
  prChannels              :ArrayOfInteger;
  prTotalCount            :Integer;
  prJoinSels              :ArrayOfTQUB_FirstLast;
  prTaskChannels          :TQUB_Task_ExtractSelection_Channels;
  prDataProcessed         :TQUB_Task_Data_Processed;
  prProcOnlyActive        :Boolean;
  prProcessFilter         :Boolean;
  prProcessBaseline       :Boolean;
  prActiveChannel         :Integer;

  function prSetupIdealization: INode;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

  function prGetChunk(Ix :Integer): TQUB_Thread_PChunk;
  function prGetChunkCount: Integer;

  procedure prGetJoinSources(First, Last :Integer; var Sources :ArrayOfTQUB_FirstLast);

  property Chunks[Ix :Integer] :TQUB_Thread_PChunk  read prGetChunk;
  property ChunkCount          :Integer             read prGetChunkCount;

public
  constructor Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Chunks :TList);
  destructor Destroy; override;

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

TQUB_Task_Extract_DataSet_FileList = class(TQUB_Task_Extract_DataSet)
// handles: file list
// it seems join could work the same way...
// join should still happen first, so you can treat each file as a segment e.g. for adding signals.
protected
  prSources               :Array of TQUB_Task_Extract_DataSet;
  prDatas                 :Array of TQUB_QVO_QVD;
  prCounts                :Array of Integer;
  prDataFront             :Integer;
  prTotalCount            :Integer;
  prChannelCount          :Integer;
  prSrcSels               :ArrayOfTQUB_SrcSegCount;
  prSegStarts             :Array of Double;
  prLastTime              :Double;

  procedure prSetupSession;
  procedure prSetupIdealization;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

  procedure prGetSourceSel(Seg :Integer; var Source :TQUB_SrcSegCount);

public
  constructor Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD);

  procedure AddSource(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; Last :Boolean);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

TQUB_Task_Extract_DataSet_Add       = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prSampleCount           :Integer;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
  // uses Buf.Alt // Source can't
end;

TQUB_Task_Extract_DataSet_Sub       = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prSampleCount           :Integer;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
  // uses Buf.Alt // Source can't
end;

TQUB_Task_Extract_DataSet_AddSubList   = class(TQUB_Task_Extract_DataSet)
protected


  prSource                :TQUB_Task_Extract_DataSet;
  prAddSubSource          :TQUB_Task_Extract_DataSet;
  prSampleCount           :Integer;
  prChannelIndex          :Integer;
  prAddSub                :Integer;


  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  constructor Create(Source :TQUB_Task_Extract_DataSet; AddSubSource :TQUB_Task_Extract_DataSet; ChannelIndex :Integer; AddSub :Integer);


  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
  // uses Buf.Alt // Source can't

end;

TQUB_Task_Extract_DataSet_AvgLists   = class(TQUB_Task_Extract_DataSet)//I don't know yet if this is the best ancestor
protected


  prSource                :TQUB_Task_Extract_DataSet;
  prSources               :TList;// of QUB_Task_Extract_DataSet;
  prChunksList            :TList;
  prSourceCount           :Integer;
  prSampleCount           :Integer;
  prChannelIndex          :Integer;


  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  constructor Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Source :TQUB_Task_Extract_DataSet; SelectionLists :TQUB_QVO_QVD_SelectionLists; ChannelIndex :Integer; Joined :Boolean);
  destructor Destroy; override;

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
  // uses Buf.Alt // Source can't
end;

TQUB_Task_Extract_DataSet_AvgListsFiles = class(TQUB_Task_Extract_DataSet_AvgLists)
protected
public


  constructor Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer; Joined :Boolean);


end;

TQUB_Task_Extract_DataSet_Avg       = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prSampleCount           :Integer;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
  // uses Buf.Alt // Source can't
end;

TQUB_Task_Extract_DataSet_Std       = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prSampleCount           :Integer;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
  // uses Buf.Alt // Source can't
end;

TQUB_Task_Extract_DataSet_Scale     = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prScale                 :Double;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; Scale :Double; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

TQUB_Task_Extract_DataSet_Drift     = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prA, prB                :Double;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; A, B :Double; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

TQUB_Task_Extract_DataSet_Sine      = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prHz, prAmp, prDeg      :Double;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; Hz, Amp, Deg :Double; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

TQUB_Task_Extract_DataSet_Digitize = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prThr, prMin, prMax     :Double;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; Thr, Min, Max :Double; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

//Derivate
TQUB_Task_Extract_DataSet_Derivate = class(TQUB_Task_Extract_DataSet)
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prChannelIndex          :Integer;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;
end;

//Leak-Cap
TQUB_Task_Extract_DataSet_LKC     = class(TQUB_Task_Extract_DataSet)
protected


  prSource                :TQUB_Task_Extract_DataSet;
  prVSource               :TQUB_Task_Extract_DataSet;
  prVBuf                  :TQUB_DBuf2_Double;
  prChannelIndex          :Integer;

  prData                  :TQUB_QVO_QVD;

  prLKC_I                 :TPADouble;
  prLKC_V0                :Double;


  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  constructor Create(Source, SubSource :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; ChannelIndex :Integer);
  destructor Destroy; override;


  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;


end;

//I to G
TQUB_Task_Extract_DataSet_ItoG     = class(TQUB_Task_Extract_DataSet)
protected


  prSource                :TQUB_Task_Extract_DataSet;
  prVRev                  :Double;
  prVCh                   :Integer;
  prChannelIndex          :Integer;

  prData                  :TQUB_QVO_QVD;


  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  constructor Create(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; VRev :Double; VCh :Integer; ChannelIndex :Integer);


  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;


end;

//Rs Correction
TQUB_Task_Extract_DataSet_RsCorr    = class(TQUB_Task_Extract_DataSet)
protected


  prSource                :TQUB_Task_Extract_DataSet;
  prRsValue               :Double;
  prRsPer                 :Double;
  prVCh                   :Integer;
  prRsAddLeak             :Boolean;
  prRsLeakCh              :Integer;
  prChannelIndex          :Integer;

  prData                  :TQUB_QVO_QVD;


  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  constructor Create(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; RsValue, RsPer :Double; VCh :Integer; AddLeak :Boolean; LeakCh :Integer; ChannelIndex :Integer);


  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;


end;

//P/n protocol
//Add
TQUB_Task_Extract_DataSet_PnAdd    = class(TQUB_Task_Extract_DataSet)
protected


  prSource                         :TQUB_Task_Extract_DataSet;
  prOffset                         :Double;
  prDelay                          :Double;
  prScale                          :Double;
  prCount                          :Integer;
  prVm                             :Double;
  prInterval                       :Double;

  prChannelIndex                   :Integer;

  prData                           :TQUB_QVO_QVD;

  prMin                            :Double;
  prFirstTime                      :Boolean;
  prCurrSeg                        :Integer;
  prBuffer                         :TQUB_DBuf2_Double;
  prBufferCount                    :Integer;


  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  constructor Create(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; Offset, Delay, Scale :Double; Count :Integer; Vm, Interval :Double; ChannelIndex :Integer);
  destructor Destroy; override;


  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;


end;

//Sub
TQUB_Task_Extract_DataSet_PnSub    = class(TQUB_Task_Extract_DataSet_PnAdd)
protected


  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;


end;


//Fit
TQUB_Task_Extract_DataSet_Fit     = class(TQUB_Task_Extract_DataSet)
protected


  prSource                :TQUB_Task_Extract_DataSet;
  prFitSource             :TQUB_Task_Extract_DataSet;
  prFitBuf                :TQUB_DBuf2_Double;
  prChannelIndex          :Integer;

  prData                  :TQUB_QVO_QVD;


  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;


public


  constructor Create(Source, FitSource :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; ChannelIndex :Integer);
  destructor Destroy; override;


  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;


end;


TQUB_Task_Extract_DataSet_Resample = class(TQUB_Task_Extract_DataSet)
// each segment separately, by linear interpolation [user is advised to filter accordingly]
protected
  prSource                :TQUB_Task_Extract_DataSet;
  prFactor                :Double;

  procedure prResampleIdealization;

  function prGetChannelCount: Integer; override;
  function prGetSegmentCount: Integer; override;
  function prGetSegmentStart(Ix :Integer): Double; override;
  function prGetSegmentLength(Ix :Integer): Integer; override;

public
  constructor Create(Source :TQUB_Task_Extract_DataSet; Factor :Double);

  function Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean; override;

end;



//******************************************************************************

TQUB_Task_Extract_File                   = class
protected
  prChannelCount                         :Integer;
  prSampling                             :Double;
  Fi                                     :File of Byte;
  prFileOpen                             :Boolean;
  prData                                 :TQUB_QVO_QVD;//I added this because Channels node does not
                                                       //work properly when the same channel is multiply extracted
  prChannels                             :TList;
public
  constructor Create(ChannelCount :Integer; Sampling :Double; Data :TQUB_QVO_QVD; ChannelList :TList); virtual;
  destructor Destroy; override;

  function Open(FileName :String): Boolean; virtual;
  function WriteFileHeader: Boolean; virtual; abstract;
  function WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean; virtual; abstract;
  function WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean; virtual; abstract;
  function Close: Boolean; virtual;
end;

TQUB_Task_Extract_File_Int               = class(TQUB_Task_Extract_File)
protected
  prDataSize                             :Integer;
  prDataType                             :TQUB_QFS_Node_DataType;
  prScaling                              :Double;
  prChannelScaling                       :array of Double; // prScaling * DataCHannelScaling[i]
  prChannelOffset                        :array of Double;

public
  constructor Create(ChannelCount :Integer; Sampling :Double; DataSize :Integer; DataType :TQUB_QFS_Node_DataType; Scaling :Double; Channels :INode; Data :TQUB_QVO_QVD; ChannelList :TList); reintroduce;

  property Scaling                       :Double         read prScaling;
  property DataSize                      :Integer        read prDataSize;
end;

TQUB_Task_Extract_File_QDF               = class(TQUB_Task_Extract_File_Int)
protected
  QDF                                    :INode;
  QDFSegs                                :INode;
  QDFSeg                                 :INode;
  QDFChan                                :INode;
  SamplesWritten                         :Integer;

public
  function Open(FileName :String): Boolean; override;
  function WriteFileHeader: Boolean; override;
  function WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean; override;
  function WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean; override;
  function Close: Boolean; override;
end;

TQUB_Task_Extract_File_LDT               = class(TQUB_Task_Extract_File_Int)
protected
  prRescaling                            :Integer;
  prIntScal                              :Integer;

public
  Scaling                                :Double;

  constructor Create(ChannelCount :Integer; Sampling :Double; Scaling :Double; Channels :INode; Data :TQUB_QVO_QVD; ChannelList :TList);

  function WriteFileHeader: Boolean; override;
  function WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean; override;
  function WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean; override;
end;

TQUB_Task_Extract_File_DAT               = class(TQUB_Task_Extract_File_Int)
public
  Float4Byte                             :Boolean;

  function WriteFileHeader: Boolean; override;
  function WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean; override;
  function WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean; override;
end;

TQUB_Task_Extract_File_ASC               = class(TQUB_Task_Extract_File)
protected
  prTimeColumn                           :Boolean;
  prFirstSeg                             :Boolean;
  prStartTime                            :Double;
  prSamplesWritten                       :Integer;
  prTimeFormat                           :String;

public
  constructor Create(ChannelCount :Integer; Sampling :Double; TimeColumn :Boolean; Data :TQUB_QVO_QVD; ChannelList :TList); reintroduce; virtual;

  function WriteFileHeader: Boolean; override;
  function WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean; override;
  function WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean; override;
  function Close: Boolean; override;
end;
TQUB_Task_Extract_File_ASC_SegCol        = class(TQUB_Task_Extract_File_ASC)
protected
  prTmpName                              :String;
  prTmpData                              :INode;
  prTmpSeg                               :INode;
public

  function WriteFileHeader: Boolean; override;
  function WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean; override;
  function WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean; override;
  function Close: Boolean; override;
end;


TQUB_Task_Extract_File_WAV               = class(TQUB_Task_Extract_File_Int)
public
  function WriteFileHeader: Boolean; override;
  function WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean; override;
  function WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean; override;
end;




implementation


uses


QUB_Main,
QUB_GUIInterface,
QUB_FitCurves,
QUB_Python_Data,
QUB_Idealization,
QUB_Parser;


//******************************************************************************
constructor TQUB_Task_Extract_DataSet_Chunks.Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Chunks :TList);
var
  i, p, q          :Integer;
  Node             :INode;
  Chan             :INode;
  Chans            :Array of INode;
  List, Sels       :INode;
  SelNode          :INode;
  CSel             :TQUB_FirstLast;
begin
  prData           := Data;
  prChunks         := Chunks;
  Sampling         := prData.Sampling;
  Session          := prData.CloneSession;

  if Task.Channels = CHANNELS_ACTIVE then begin
    SetLength(prChannels, 1);
    prChannels[0] := prData.ActiveChannel;
    ActiveChannel := 0;
  end else if Task.Channels = CHANNELS_LIST then begin
    SetLength(prChannels, Task.ChannelList.Count);
    ActiveChannel := Length(prChannels) + 1; // i.e., not in the list
    for i := 0 to Length(prChannels) - 1 do begin
      prChannels[i]   := Integer(Task.ChannelList[i]);
      if prChannels[i] = prData.ActiveChannel then begin
        ActiveChannel := i;
        Break;
      end;
    end;
  end else begin
    SetLength(prChannels, prData.DataObject.ADChannelCount);
    for i := 0 to prData.DataObject.ADChannelCount - 1 do begin
      prChannels[i] := i;
    end;
    ActiveChannel := prData.ActiveChannel;
  end;

  //ASCII Fit - to do: make it work with other formats
  if (Task.ASCIIFit) {and (Task.FileFormat = QFS_QDF_TYPE_ASC)} then begin
    SetLength(prChannels, Length(prChannels)+1);
    prChannels[Length(prChannels)-1] := -1;
    prFitCurve := 1;
  end else begin
    prFitCurve := 0;
  end;

  prTotalCount := 0;
  SetLength(prJoinSels, prChunks.Count);
  for i := 0 to prChunks.Count - 1 do begin
    CSel := Self.Chunks[i].Sel;
    prJoinSels[i].First := prTotalCount;
    prTotalCount := prTotalCount + (CSel.Last - CSel.First + 1);
    prJoinSels[i].Last := prTotalCount - 1;
  end;


  //Kinetic Parameters
  if (Task.KinPars) and (Task.KinParsList.Count > 0) then begin
    prKinParsCount := Task.KinParsList.Count;
    SetLength(prChannels, Length(prChannels) + prKinParsCount);
    for i := Length(prChannels) - 1 downto Length(prChannels) - prKinParsCount do begin
      prChannels[i] := -1;
    end;
  end else begin
    prKinParsCount := 0;
  end;


  //****************************************************************************
  Session.RemoveChild(Session['Results']);
  Session.RemoveChild(Session['Baseline']);
  Session.RemoveChild(Session['Markers']);
  Session.RemoveChild(Session['Lists']);

  Session.RemoveChild(Session['Idealization']);
  Session.AppendChild( prSetupIdealization );

  Session['Scaling'].SetData(prData.DataObject.Scaling);
  Session['Sampling'].SetData(Sampling );
  Session['ADChannelCount'].SetData(ChannelCount);
  SetLength(Chans, prData.DataObject.ADChannelCount);

  Node       := Session.Find('Display').Find('Channels');
  Chan       := Node.Find('Channel');
  i          := 0;
  while not Chan.IsNull do begin
    Chans[i] := Chan;
    Chan     := Chan.SiblingSameName;
    Inc(i);
  end;
  Session.Find('Display').RemoveChild(Node);
  Node       := Session['Display']['Channels'];
  for i := 0 to ChannelCount - prFitCurve - prKinParsCount - 1 do begin
    if Chans[prChannels[i]] <> nil{It was crashing here - this whole bussiness is loose. LM} then Node.AppendClone(Chans[prChannels[i]], True);
  end;
  if prFitCurve > 0 then begin
    Node.AppendClone(Chans[ActiveChannel], True);
  end;
  if prKinParsCount > 0 then begin
  end;

  //Copy the existing DataChannels in Chans
  Node       := Session.Find('DataChannels');
  Chan       := Node.Find('Channel');
  i          := 0;
  while not Chan.IsNull do begin
    Chans[i] := Chan;
    Chan     := Chan.SiblingSameName;
    Inc(i);
  end;
  //Remove the existing DataChannels
  Session.RemoveChild(Node);
  //Create new DataChannels
  Node       := Session['DataChannels'];
  //Append previous channels
  for i := 0 to ChannelCount - prFitCurve - prKinParsCount - 1 do begin
    Node.AppendClone(Chans[prChannels[i]], True);
  end;
  //Append DataChannel for Fit curve
  if prFitCurve > 0 then begin
    Chan := Chans[ActiveChannel].Clone(True);
    Chan['Name'].SetData('Fit(' + Chan['Name'].DataAsString + ')');
    Node.AppendChild(Chan);
  end;
  //Append DataChannels for Kinetic Parameters
  if prKinParsCount > 0 then begin
    for i := 0 to prKinParsCount - 1 do begin
      Chan := Chans[ActiveChannel].Clone(True);
      Chan['Name'].SetData(Task.KinParsList.Strings[i]);
      Chan['Units'].SetData('');
      Chan['Scaling'].SetData(VarAsType(1.0, varDouble));
      Node.AppendChild(Chan);
    end;
  end;

  Node       := Session['SegmentSizes'];
  Node.SetupNumData( NODE_DATATYPE_LONGINT, SegmentCount, 1 );
  for i := 0 to SegmentCount - 1 do begin
    Node.Data[i] := SegmentLength[i];
  end;

  if (Task.DataSource = DATA_SOURCE_SELECTIONLIST)
  and (Task.SelectionList <> nil)
  then begin
    List      := Task.SelectionList.AsTree;
    List.SetData( prData.CurrListName );
    Session['Lists'].AppendChild( List );

    Sels      := List['Sels'];
    SelNode   := Sels.Sibling;
    p         := 0;
    for i := 0 to Sels.Rows - 1 do begin
      q := Sels.MatrixData[i,1] - Sels.MatrixData[i,0] + 1;
      Sels.MatrixData[i, 0] := p;
      if not SelNode.IsNull then SelNode.Data[0] := p;
      p      := p + q;
      Sels.MatrixData[i, 1] := p-1;
      if not SelNode.IsNull then SelNode.Data[1] := p-1;
      SelNode    := SelNode.Sibling;
    end;
  end;

  prKinPars               := TStringList.Create;
  prKinPars.Assign(Task.KinParsList);

  prTaskChannels          := Task.Channels;
  prDataProcessed         := Task.DataProcessed;
  prProcOnlyActive        := Task.ProcOnlyActiveChannel;
  prProcessFilter         := Task.ProcessFilter;
  prProcessBaseline       := Task.ProcessBaseline;
end;

destructor TQUB_Task_Extract_DataSet_Chunks.Destroy;
var
  i        :Integer;
begin
  if prChunks <> nil then begin
    for i := 0 to prChunks.Count - 1 do begin
      FreeMem(prChunks.List[i]);
    end;
    prChunks.Free;
  end;
  prKinPars.Free;

  inherited;
end;

function TQUB_Task_Extract_DataSet_Chunks.prSetupIdealization: INode;
var
  Chan, Seg          :INode;
  i, c, s, h, p, d   :Integer;
  Firsts, Lasts      :TPAInteger;
  off                :Integer;
begin
  // each chan: fill a Session:Idealization:Channel with Segment of idl per chunk
  //            adjust indices Segment, Firsts, Lasts for skipped data
  Result         := TNode.Create('Idealization');
  for i := 0 to ChannelCount - 1 do begin
    Chan         := Result.AppendChild('Channel');
    c    := prChannels[i];
    if c < 0 then Continue;

    for s := 0 to ChunkCount - 1 do begin
      prData.DataObject.Idealization[c].GetDwellsInto(Chunks[s].Sel.First, Chunks[s].Sel.Last, 1{Fragments}, 1{Durations}, Chan, 0{not insert first});
    end;

    h       := -1; // chunk index
    p       := 0; // expected next point, looking for gaps
    off     := 0; // accumulating the offset as we slide to fill gaps
    Seg     := Chan.Find('Segment');
    while not Seg.IsNull do begin
      while (h < 0) or ((h < (ChunkCount-1)) and (Seg.Data[0] > Chunks[h].Sel.Last)) do begin
        Inc(h);
        off := off + Chunks[h].Sel.First - p;
        p   := Chunks[h].Sel.Last + 1;
      end;
      Seg.Data[0] := Seg.Data[0] - off;
      Seg.Data[1] := Seg.Data[1] - off;
      Firsts      := TPAInteger(Seg['Firsts'].DataPtr);
      Lasts       := TPAInteger(Seg['Lasts'].DataPtr);
      for d := 0 to Integer(Seg['DwellCount'].Data[0]) - 1 do begin
        Firsts[d] := Firsts[d] - off;
        Lasts[d]  := Lasts[d] - off;
      end;
      Seg   := Seg.SiblingSameName;
    end;
  end;

  // if joined: replace with concatenated Segment, mean amp,std
  // if scale:  scale amp,std
  // if any ensemble proc: remove Idl
  // if resample: carefully adjust indices Segment, Firsts, Lasts, Durations
  //              same rounding as data indexing, no 0-sample events
  //              check list indices
end;

function TQUB_Task_Extract_DataSet_Chunks.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  c, i, lim    :Integer;
  IBuf         :TPMInteger;
  Sel          :TQUB_FirstLast;
  Scal         :Double;
  Process      :Boolean;
  CurveIter    :TQUB_FitCurveIter;
  CurveFirst   :Integer;
  CurveData    :TPAInteger;
  ParName      :String;
  ParValue     :Double;
  NextValue    :Double;
  ChangePoint  :TQUB_FirstLast;
  ParamsToDraw :INode;

  //Assumes the ChangePoints are in order
  function FindNextChangePoint(const ParName :String; var ParValue :Double; PrevChangePoint :Integer) :Integer;
  var
    Params        :INode;
    Par           :INode;
    ChangePoint   :INode;
    CurrPoint     :INode;
    ThisPoint     :Integer;
    p             :Integer;
  begin
    Result := -1;//No more change points;
    for p := 0 to prData.DataKinParamsMarkersCount - 1 do begin
      if prData.DataKinParamsMarkers[p] > PrevChangePoint then begin
        Result := prData.DataKinParamsMarkers[p];
        ChangePoint := ParamsToDraw.Find('ChangePoint');
        while not ChangePoint.IsNull do begin
          CurrPoint := ChangePoint.Find('CurrPoint');
          ThisPoint := CurrPoint.Data[0];
          if ThisPoint = Result then begin
            Params    := ChangePoint.Find('Parameters');
            if not Params.IsNull then begin
              Par := Params.Find(ParName);
              if not Par.IsNull then begin
                try
                  ParValue := Par.Data[0];
                except
                end;
              end;
            end;
            Break;
          end;
          ChangePoint := ChangePoint.SiblingSameName;
        end;
        Break;
      end;
    end;
  end;

begin
  IBuf         := nil;
  Result       := False;

  try
    IBuf       := IAlloc2(Buf.RowCount, Buf.ColCount);

    Sel.First  := Chunks[Seg].Sel.First + First;
    Sel.Last   := Chunks[Seg].Sel.First + Last;

    // if all channels & same process
    if  (prTaskChannels = CHANNELS_ALL)
    and ((prDataProcessed = DATA_PROCESSED_NONE) or (not prProcOnlyActive)) then begin
      Result := prData.LoadData(IBuf, Sel, prDataProcessed <> DATA_PROCESSED_NONE, prProcessFilter, prProcessBaseLine);
      if not Result then begin
        for c := 0 to ChannelCount - prFitCurve - prKinParsCount - 1 do begin
          Scal := 1.0 / (prData.Scaling * prData.DataChannelScaling[prChannels[c]]);
          for i := 0 to (Sel.Last - Sel.First + 1) - 1 do begin
            Buf.Buf[c][i] := Scal * IBuf[c][i];
          end;
        end;
      end;
    end else begin
      for c := 0 to ChannelCount - prFitCurve - prKinParsCount - 1 do begin
        Process := (prDataProcessed <> DATA_PROCESSED_NONE)
                   and ((prProcOnlyActive and (prChannels[c] = prData.ActiveChannel))
                        or not prProcOnlyActive);
        Result  := prData.LoadDataChannel(IBuf[c], prChannels[c], Sel, Process, prProcessFilter, prProcessBaseLine);
        if not Result then begin
          Scal    := 1.0 / (prData.Scaling * prData.DataChannelScaling[prChannels[c]]);
          for i := 0 to (Sel.Last - Sel.First + 1) - 1 do begin
            Buf.Buf[c][i] := Scal * IBuf[c][i];
          end;
        end;
      end;
    end;

    //Add the Fit Curve ********************************************************
    if prFitCurve > 0 then begin

      c := ChannelCount - prKinParsCount - 1;
      Scal := 1.0 / (prData.Scaling * prData.DataChannelScaling[ prData.ActiveChannel ]);

      CurveIter := nil;
      try

        i := Sel.First;
        try
          {TO DO: LM. John, this function (Find) never returns. What's the problem?}
          //Please do it high priority, as fit curves cannot be extracted
          //The same function works fine in Data.Draw, for example.
          //CurveIter := prData.FitCurves.Find(Sel.First);//doesn't ever return

          //Testing
          prData.FitCurves.FirstFindSynch := Sel.First;
          SyncGUIProc(prData.FitCurves.FindSynch);
          CurveIter := prData.FitCurves.FitFoundSynch;
          //

          while (0 = CurveIter.Done) and (CurveIter.Curve.First <= Sel.Last) do begin
            CurveFirst := CurveIter.Curve.First;
            while i < CurveFirst do begin
              Buf.Buf[c][i - Sel.First] := 0;
              Inc(i);
            end;
            lim       := Min(CurveIter.Curve.Last, Sel.Last);
            CurveData := CurveIter.Curve.Data;
            while i <= lim do begin
              Buf.Buf[c][i - Sel.First] := Scal * CurveData[i - CurveFirst];
              Inc(i);
            end;
            CurveIter.Next;
          end;
        except
        end;

      finally
        if CurveIter <> nil then CurveIter.Free;
      end;

      while i <= Sel.Last do begin
        Buf.Buf[c][i - Sel.First] := 0;
        Inc(i);
      end;

    end;

    //Add the Kinetic Parameters
    if prKinParsCount > 0 then begin

      ParamsToDraw := prData.DataObject.Session.Find('KineticParameters');

      for c := ChannelCount - prKinParsCount to ChannelCount - 1 do begin

        ParName := prKinPars[c - (ChannelCount - prKinParsCount)];

        //Find the first change point:
        ParValue := 0;
        ChangePoint.First := FindNextChangePoint(ParName, ParValue,  -1);
        NextValue := ParValue;
        ChangePoint.Last  := FindNextChangePoint(ParName, NextValue, ChangePoint.First);

        for i := Sel.First to Sel.Last do begin

          while not ((i >= ChangePoint.First) and (i <= ChangePoint.Last)) do begin
            ChangePoint.First := ChangePoint.Last + 1;
            ParValue          := NextValue;
            ChangePoint.Last  := FindNextChangePoint(ParName, NextValue, ChangePoint.First);
            if ChangePoint.Last = -1 then begin
              ChangePoint.Last := Sel.Last;
              Break;
            end;
          end;

          Buf.Buf[c][i - Sel.First] := ParValue;

        end;//i

      end;//c

    end;//KinParsCount

  except
    Result := True;
  end;

  Free2(TPMChar(IBuf));
end;

function TQUB_Task_Extract_DataSet_Chunks.prGetChannelCount: Integer;
begin
  Result := Length(prChannels);
end;

function TQUB_Task_Extract_DataSet_Chunks.prGetSegmentCount: Integer;
begin
  Result := prChunks.Count;
end;

function TQUB_Task_Extract_DataSet_Chunks.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := Chunks[Ix].StartTime;
end;

function TQUB_Task_Extract_DataSet_Chunks.prGetSegmentLength(Ix :Integer): Integer;
begin
  with Chunks[Ix].Sel do begin
    Result := Last - First + 1;
  end;
end;

function TQUB_Task_Extract_DataSet_Chunks.prGetChunk(Ix :Integer): TQUB_Thread_PChunk;
begin
  Result := TQUB_Thread_PChunk(prChunks[Ix]);
end;

function TQUB_Task_Extract_DataSet_Chunks.prGetChunkCount: Integer;
begin
  Result := prChunks.Count;
end;

constructor TQUB_Task_Extract_DataSet_Chunks_Joined.Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Chunks :TList);
var
  i, p, q          :Integer;
  Sel              :TQUB_FirstLast;
  Node             :INode;
  Chan             :INode;
  Chans            :Array of INode;
  List, Sels, SelNode :INode;
begin
  prData           := Data;
  prChunks         := Chunks;
  Sampling         := prData.Sampling;
  Session          := prData.CloneSession;

  // figure out channels
  if Task.Channels = CHANNELS_ACTIVE then begin
    SetLength(prChannels, 1);
    prChannels[0] := prData.ActiveChannel;
    ActiveChannel := 0;
  end else if Task.Channels = CHANNELS_LIST then begin
    SetLength(prChannels, Task.ChannelList.Count);
    ActiveChannel := Length(prChannels) + 1; // i.e. not in the list
    for i := 0 to Length(prChannels) - 1 do begin
      prChannels[i]   := Integer(Task.ChannelList[i]);
      if prChannels[i] = prData.ActiveChannel then begin
        ActiveChannel := i;
        Break;
      end;
    end;
  end else begin
    SetLength(prChannels, prData.DataObject.ADChannelCount);
    for i := 0 to prData.DataObject.ADChannelCount - 1 do begin
      prChannels[i] := i;
    end;
    ActiveChannel := prData.ActiveChannel;
  end;


  //What about ASCIIFit and KinPars?!
  // to do


  prTotalCount := 0;
  SetLength(prJoinSels, prChunks.Count);
  for i := 0 to prChunks.Count - 1 do begin
    Sel := Self.Chunks[i].Sel;
    prJoinSels[i].First := prTotalCount;
    prTotalCount := prTotalCount + (Sel.Last - Sel.First + 1);
    prJoinSels[i].Last := prTotalCount - 1;
  end;

  Session.RemoveChild( Session['Results'] );
  Session.RemoveChild( Session['Baseline'] );
  Session.RemoveChild( Session['Markers'] );
  Session.RemoveChild( Session['Lists'] );

  Session.RemoveChild( Session['Idealization'] );
  Session.AppendChild( prSetupIdealization );

  Session['Scaling'].SetData(prData.DataObject.Scaling);
  Session['Sampling'].SetData( Sampling );
  Session['ADChannelCount'].SetData( ChannelCount );
  SetLength( Chans, prData.DataObject.ADChannelCount );

  Node       := Session.Find('Display').Find('Channels');
  Chan       := Node.Find('Channel');
  i          := 0;
  while not Chan.IsNull do begin
    Chans[i] := Chan;
    Chan     := Chan.SiblingSameName;
    Inc(i);
  end;
  while i < prData.DataObject.ADChannelCount do begin
    Chans[i] := TNode.Create('Channel');
    Inc(i);
  end;
  Session.Find('Display').RemoveChild(Node);
  Node       := Session['Display']['Channels'];
  for i := 0 to ChannelCount - 1 do begin
    Node.AppendClone( Chans[ prChannels[i] ], True );
  end;

  Node       := Session.Find('DataChannels');
  Chan       := Node.Find('Channel');
  i          := 0;
  while not Chan.IsNull do begin
    Chans[i] := Chan;
    Chan     := Chan.SiblingSameName;
    Inc(i);
  end;
  while i < prData.DataObject.ADChannelCount do begin
    Chans[i] := TNode.Create('Channel');
    Inc(i);
  end;
  Session.RemoveChild(Node);
  Node       := Session['DataChannels'];
  for i := 0 to ChannelCount - 1 do begin
    Node.AppendClone( Chans[ prChannels[i] ], True );
  end;

  Node       := Session['SegmentSizes'];
  Node.SetupNumData( NODE_DATATYPE_LONGINT, SegmentCount, 1 );
  for i := 0 to SegmentCount - 1 do begin
    Node.Data[i] := SegmentLength[i];
  end;

  if (Task.DataSource = DATA_SOURCE_SELECTIONLIST)
  and (Task.SelectionList <> nil)
  then begin
    List      := Task.SelectionList.AsTree;
    List.SetData( prData.CurrListName );
    Session['Lists'].AppendChild( List );

    Sels      := List['Sels'];
    SelNode   := Sels.Sibling;
    p         := 0;
    for i := 0 to Sels.Rows - 1 do begin
      q := Sels.MatrixData[i,1] - Sels.MatrixData[i,0] + 1;
      Sels.MatrixData[i, 0] := p;
      if not SelNode.IsNull then SelNode.Data[0] := p;
      p      := p + q;
      Sels.MatrixData[i, 1] := p-1;
      if not SelNode.IsNull then SelNode.Data[1] := p-1;
      SelNode := SelNode.Sibling;
    end;
  end;

  prTaskChannels          := Task.Channels;
  prDataProcessed         := Task.DataProcessed;
  prProcOnlyActive        := Task.ProcOnlyActiveChannel;
  prProcessFilter         := Task.ProcessFilter;
  prProcessBaseline       := Task.ProcessBaseline;
end;

destructor TQUB_Task_Extract_DataSet_Chunks_Joined.Destroy;
var
  i        :Integer;
begin
  if prChunks <> nil then begin
    for i := 0 to prChunks.Count - 1 do begin
      FreeMem(prChunks.List[i]);
    end;
    prChunks.Free;
  end;

  inherited;
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.prSetupIdealization: INode;
var
  Unjoined       :INode;
  Chan, Seg      :INode;
  i, c, s, p,d   :Integer;
  Firsts, Lasts  :TPAInteger;
  off            :Integer;

  ClassCount     :Integer;
  Amp, AmpCount  :Array of Double;
  Std, StdCount  :Array of Double;
  JChan, JSeg    :INode;
  DwellCount     :Integer;
  JF, JL, JC, Classes :TPAInteger;
  JD, Durations  :TPASingle;
begin
  // each chan: fill a Session:Idealization:Channel with Segment of idl per chunk
  //            adjust indices Segment, Firsts, Lasts for skipped data
  Result         := TNode.Create('Idealization');
  for i := 0 to ChannelCount - 1 do begin
    Chan         := Result.AppendChild('Channel');
    c    := prChannels[i];
    for s := 0 to ChunkCount - 1 do begin
      prData.DataObject.Idealization[c].GetDwellsInto(Chunks[s].Sel.First, Chunks[s].Sel.Last, 1{Fragments}, 1{Durations}, Chan, 0{not insert first});
    end;
    p       := 0; // expected next point, looking for gaps
    off     := 0; // accumulating the offset as we slide to fill gaps
    Seg     := Chan.Find('Segment');
    while not Seg.IsNull do begin
      off   := off + Seg.Data[0] - p;
      p     := Seg.Data[1] + 1;
      Seg.Data[0] := Seg.Data[0] - off;
      Seg.Data[1] := Seg.Data[1] - off;
      Firsts      := TPAInteger(Seg['Firsts'].DataPtr);
      Lasts       := TPAInteger(Seg['Lasts'].DataPtr);
      for d := 0 to Integer(Seg['DwellCount'].Data[0]) - 1 do begin
        Firsts[d] := Firsts[d] - off;
        Lasts[d]  := Lasts[d] - off;
      end;
      Seg   := Seg.SiblingSameName;
    end;
  end;

  // if joined: replace with concatenated Segment, mean amp,std
  Unjoined := Result;
  Result   := TNode.Create('Idealization');
  Chan     := Unjoined.Find('Channel');
  for i := 0 to ChannelCount - 1 do begin
    JChan  := Result.AppendChild('Channel');
    JSeg   := JChan['Segment'];
    JSeg.SetupNumData(NODE_DATATYPE_LONGINT, 2, 1);
    JSeg.Data[0] := 0;
    JSeg.Data[1] := prTotalCount - 1;
    ClassCount := 0;
    DwellCount := 0;
    Seg    := Chan.Find('Segment');
    while not Seg.IsNull do begin
      p := Seg['amp'].DataCount;
      if p > ClassCount then begin
        SetLength(Amp, p);
        SetLength(AmpCount, p);
        SetLength(Std, p);
        SetLength(StdCount, p);
        for d := ClassCount to p - 1 do begin
          Amp[d] := 0;
          AmpCount[d] := 0;
          Std[d] := 0;
          StdCount[d] := 0;
        end;
        ClassCount := p;
      end;
      for d := 0 to p - 1 do begin
        Amp[d] := Amp[d] + Seg['amp'].Data[d];
        AmpCount[d] := AmpCount[d] + 1;
        Std[d] := Std[d] + Seg['sd'].Data[d];
        StdCount[d] := StdCount[d] + 1;
      end;
      DwellCount := DwellCount + Seg['DwellCount'].Data[0];

      Seg  := Seg.SiblingSameName;
    end;
    for d := 0 to ClassCount - 1 do begin
      if AmpCount[d] > 0 then begin
        Amp[d] := Amp[d] / AmpCount[d];
      end;
      if StdCount[d] > 0 then begin
        Std[d] := Std[d] / StdCount[d];
      end;
    end;
    Seg['amp'].SetData( Amp );
    Seg['sd'].SetData( Std );

    JSeg['DwellCount'].SetData(DwellCount);
    JSeg['Firsts'].SetupNumData(NODE_DATATYPE_LONGINT, DwellCount, 1);
    JSeg['Lasts'].SetupNumData(NODE_DATATYPE_LONGINT, DwellCount, 1);
    JSeg['Classes'].SetupNumData(NODE_DATATYPE_LONGINT, DwellCount, 1);
    JSeg['Durations'].SetupNumData(NODE_DATATYPE_SINGLE, DwellCount, 1);
    JF     := TPAInteger(JSeg['Firsts'].DataPtr);
    JL     := TPAInteger(JSeg['Lasts'].DataPtr);
    JC     := TPAInteger(JSeg['Classes'].DataPtr);
    JD     := TPASingle(JSeg['Durations'].DataPtr);
    p      := 0;
    Seg    := Chan.Find('Segment');
    while not Seg.IsNull do begin
      Firsts := TPAInteger(Seg['Firsts'].DataPtr);
      Lasts := TPAInteger(Seg['Lasts'].DataPtr);
      Classes := TPAInteger(Seg['Classes'].DataPtr);
      Durations := TPASingle(Seg['Durations'].DataPtr);
      for d := 0 to Integer(Seg['DwellCount'].Data[0]) - 1 do begin
        JF[p] := Firsts[d];
        JL[p] := Lasts[d];
        JC[p] := Classes[d];
        JD[p] := Durations[d];
        Inc(p);
      end;

      Seg  := Seg.SiblingSameName;
    end;

    Chan   := Chan.SiblingSameName;
  end;

  // TODO ELSEWHERE:
  // if any ensemble proc: remove Idl
  // if resample: carefully adjust indices Segment, Firsts, Lasts, Durations
  //              same rounding as data indexing, no 0-sample events
  //              check list indices
end;

procedure TQUB_Task_Extract_DataSet_Chunks_Joined.prGetJoinSources(First, Last :Integer; var Sources :ArrayOfTQUB_FirstLast);
var
  s, n      :Integer;
begin
  n         := 0;
  s         := 0;
  while (s < prChunks.Count) and (First <= Last) do begin
    if First < prJoinSels[s].Last then begin
      Inc(n);
      SetLength(Sources, n);
      Sources[n-1].First := Chunks[s].Sel.First + First - prJoinSels[s].First;
      Sources[n-1].Last  := Min(Chunks[s].Sel.Last, Chunks[s].Sel.First + Last - prJoinSels[s].First);
      First      := First + (Sources[n-1].Last - Sources[n-1].First + 1);
    end;
    Inc(s);
  end;
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  s, c, i, p   :Integer;
  IBuf         :TPMInteger;
  Sels         :ArrayOfTQUB_FirstLast;
  Scal         :Double;
  Process      :Boolean;
begin
  IBuf         := nil;
  Result       := False;

  try
    IBuf       := IAlloc2(Buf.RowCount, Buf.ColCount);

    prGetJoinSources(First, Last, Sels);

    // if all channels/same process
    if  (prTaskChannels = CHANNELS_ALL)
    and ((prDataProcessed = DATA_PROCESSED_NONE) or (not prProcOnlyActive)) then begin
      p     := 0;
      for s := 0 to Length(Sels) - 1 do begin
        Result := prData.LoadData(IBuf, Sels[s], prDataProcessed <> DATA_PROCESSED_NONE, prProcessFilter, prProcessBaseLine);
        if Result then Break;

        for c := 0 to ChannelCount - 1 do begin
          Scal    := 1.0 / (prData.Scaling * prData.DataChannelScaling[ prChannels[c] ]);
          for i := 0 to (Sels[s].Last - Sels[s].First + 1) - 1 do begin
            Buf.Buf[c][p+i] := Scal * IBuf[c][i];
          end;
        end;
        p   := p + (Sels[s].Last - Sels[s].First + 1);
      end;
    end else begin
      p     := 0;
      for s := 0 to Length(Sels) - 1 do begin
        for c := 0 to ChannelCount - 1 do begin
          Process := (prDataProcessed <> DATA_PROCESSED_NONE)
                     and ((prProcOnlyActive and (prChannels[c] = prData.ActiveChannel))
                          or not prProcOnlyActive);
          Result  := prData.LoadDataChannel(IBuf[c], prChannels[c], Sels[s], Process, prProcessFilter, prProcessBaseLine);
          if Result then Break;
          Scal    := 1.0 / (prData.Scaling * prData.DataChannelScaling[ prChannels[c] ]);
          for i := 0 to (Sels[s].Last - Sels[s].First + 1) - 1 do begin
            Buf.Buf[c][p+i] := Scal * IBuf[c][i];
          end;
        end;
        p   := p + (Sels[s].Last - Sels[s].First + 1);
      end;
    end;
  except
    Result := True;
  end;

  Free2(TPMChar(IBuf));
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.prGetChannelCount: Integer;
begin
  Result := Length(prChannels);
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.prGetSegmentCount: Integer;
begin
  Result := 1;
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := Chunks[0].StartTime;
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prTotalCount;
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.prGetChunk(Ix :Integer): TQUB_Thread_PChunk;
begin
  Result := TQUB_Thread_PChunk(prChunks[Ix]);
end;

function TQUB_Task_Extract_DataSet_Chunks_Joined.prGetChunkCount: Integer;
begin
  Result := prChunks.Count;
end;

constructor TQUB_Task_Extract_DataSet_FileList.Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD);
begin
  prData           := Data;
  ActiveChannel    := Data.ActiveChannel;
  Sampling         := Data.DataObject.Sampling;
  prDataFront      := -1;
  prTotalCount     := 0;
  prLastTime       := UNSET_VALUE;
  Session          := TNode.CreateNull;
end;

procedure TQUB_Task_Extract_DataSet_FileList.AddSource(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; Last :Boolean);
var
  Ix            :Integer;
  i, p, n       :Integer;
begin
  if Source.SegmentCount = 0 then Exit;

  Ix            := Length(prSources);
  SetLength(prSources, Ix+1);
  SetLength(prDatas, Ix+1);
  SetLength(prCounts, Ix+1);
  prSources[Ix] := Source;
  prDatas[Ix]   := Data;

  if prLastTime = UNSET_VALUE then begin
    prLastTime := Source.SegmentStart[0];
  end;
  if prData = Data then prDataFront := Ix;
  if Sampling <> Source.Sampling then begin
    // warn
  end;

    // build array of segment start, beginning with first file's,
    //   then (prev-segn-start + prev-segn-len) + (segi-start - seg0-start)
  p             := Length(prSrcSels);
  SetLength(prSrcSels, p + Source.SegmentCount);
  SetLength(prSegStarts, p + Source.SegmentCount);
  n := 0;
  for i := p to p + Source.SegmentCount - 1 do begin
    n                  := n + Source.SegmentLength[i-p];
    prSrcSels[i].Src   := Ix;
    prSrcSels[i].Seg   := i-p;
    prSrcSels[i].Count := Source.SegmentLength[i-p];
    if Ix = 0 then begin
      prSegStarts[i]   := Source.SegmentStart[i-p];
    end else begin
      prSegStarts[i]   := prLastTime + Source.SegmentStart[i-p] - Source.SegmentStart[0];
    end;
    prLastTime         := prSegStarts[i] + Source.Sampling * Source.SegmentLength[i-p]; // time units?
  end;
  prTotalCount         := prTotalCount + n;
  prCounts[Ix]         := n;

  if Last then begin
    // find source with most a/d channels;
    n                  := prSources[0].ChannelCount;
    p                  := 0;
    for i := 1 to Length(prSources) - 1 do begin
      if n < prSources[i].ChannelCount then begin
        n  := prSources[i].ChannelCount;
        p  := i;
      end;
    end;
    prChannelCount := n;
    // if the front file prData is not in the list, or it has fewer channels, pick the mostest
    if (prDataFront = -1) or (prSources[prDataFront].ChannelCount < n) then begin
      prDataFront    := p;
      prData         := prDatas[p];
    end;
    prSetupSession;
    // build session out of prData's session, with first file's lists, idl, comment, segsizes
    // augment session with 2..n files' lists, idl, comment, segsizes
    // augment session with non-prData files' expcond
    // adjust session datachannel sampling to least across sources
    prSetupIdealization;
  end;
end;

procedure TQUB_Task_Extract_DataSet_FileList.prSetupSession;
var
  i, s, n, p   :Integer;
  a            :Array of Double;
  Chan, Node   :INode;
  Last, New    :INode;
  Sels         :INode;
begin
  // take prData.session.clone, get lists, idl, comment, segsizes of first source
  Session      := prSources[prDataFront].Session.Clone(True);
  Session.RemoveChild( Session['Results'] );
  Session.RemoveChild( Session['Baseline'] );
  Session.RemoveChild( Session['Markers'] );

  if prDataFront <> 0 then begin
    Session.RemoveChild( Session['Lists'] );
    Session.RemoveChild( Session['Idealization'] );
    Session.RemoveChild( Session['Comment'] );
    Session.RemoveChild( Session['SegmentSizes'] );
    Session.AppendClone( prSources[0].Session['Lists'], True );
    Session.AppendClone( prSources[0].Session['Idealization'], True );
    Session.AppendClone( prSources[0].Session['Comment'], True );
    Session.AppendClone( prSources[0].Session['SegmentSizes'], True );
  end;

  // adjust session datachannel scaling to least across sources
  SetLength(a, ChannelCount);
  Chan  := prSources[0].Session['DataChannels']['Channel'];
  for i := 0 to ChannelCount - 1 do begin
    if Chan.IsNull then Break;
    Node := Chan['Scaling'];
    a[i] := Node.Data[0];
    Chan := Chan.SiblingSameName;
  end;
  for s := 0 to Length(prSources) - 1 do begin
    Chan  := prSources[s].Session['DataChannels']['Channel'];
    for i := 0 to ChannelCount - 1 do begin
      if Chan.IsNull then Break;
      if Abs(Chan['Scaling'].Data[0]) < a[i] then begin
        a[i] := Chan['Scaling'].Data[0];
      end;
      Chan := Chan.SiblingSameName;
    end;
  end;
  Chan := Session['DataChannels'].Find('Channel');
  for i := 0 to ChannelCount - 1 do begin
    if Chan.IsNull then Chan := Session['DataChannels'].AppendChild('Channel');
    Chan['Scaling'].SetData(a[i]);
    Chan := Chan.SiblingSameName;
  end;

  // add 2..n files' lists, with f,l += prev-source-points
  Last       := Session['Lists'].Child;
  n          := prCounts[0];
  for s := 1 to Length(prSources) - 1 do begin
    Node     := prSources[s].Session['Lists'].Find('List');
    while not Node.IsNull do begin
      New := Session['Lists'].Find('List');
      while (not New.IsNull) and (New.DataAsString <> Node.DataAsString) do begin
        New := New.SiblingSameName;
      end;
      if New.IsNull then begin
        New := Session['Lists'].InsertClone(Last, Node, True);
        Last := New;
        p := 0;
      end else if New['Sels'].DataCount = 0 then begin
        Session['Lists'].RemoveChild(New);
        New := Session['Lists'].InsertClone(Last, Node, True);
        Last := New;
        p := 0;
      end else begin
        Sels := New['Sels'];
        p    := Sels.Rows;
        Sels.ResizeData(p + LongInt(Node['Sels'].Rows));
        Move(Node['Sels'].DataPtr^, PChar(Sels.DataPtr)[2*p*SizeOf(Integer)], 2*Node['Sels'].Rows*SizeOf(Integer));
      end;

      for i := 0 to Node['Sels'].Rows - 1 do begin
        Sels.MatrixData[p+i, 0] := Sels.MatrixData[p+i, 0] + n;
        Sels.MatrixData[p+i, 1] := Sels.MatrixData[p+i, 1] + n;
      end;

      Node := Node.SiblingSameName;
    end;
    n := n + prCounts[s];
  end;

  // add 2..n files' segsizes, comments
  New := Session['SegmentSizes'];
  for s := 1 to Length(prSources) - 1 do begin
    Node := prSources[s].Session['SegmentSizes'];
    n    := New.Rows;
    New.ResizeData(n + LongInt(Node.Rows));
    Move(Node.DataPtr^, PChar(New.DataPtr)[n*SizeOf(Integer)], Node.Rows*SizeOf(Integer));
  end;
  New := Session['Comment'];
  for s := 1 to Length(prSources) - 1 do begin
    Node := prSources[s].Session['Comment'];
    if Node.DataAsString <> '' then begin
      New.SetData(New.DataAsString + Node.DataAsString);
    end;
  end;

  // add any missing expcond from non-prData files
  New := Session['ExpCond'];
  for s := 0 to Length(prSources) - 1 do begin
    if s = prDataFront then Continue;
    Node := prSources[s].Session['ExpCond'].Child;
    while not Node.IsNull do begin
      if New.Find(Node.Name).IsNull then begin
        New.AppendClone(Node, True);
      // end else if something doesn't match like channel, warn?
      end;
      Node := Node.Sibling;
    end;
  end;
end;

procedure TQUB_Task_Extract_DataSet_FileList.prSetupIdealization;
  // add 2..n files' ideal segments with bounds, firsts, lasts += prev-source-points
var
  Idl            :INode;
  s, n, d        :Integer;
  Chan, XChan    :INode;
  Seg, XSeg      :INode;
  Firsts, Lasts  :TPAInteger;
  DwellCount     :Integer;
begin
  Idl            := Session['Idealization'];
  n              := prCounts[0];
  for s := 1 to Length(prSources) - 1 do begin
    Chan         := Idl.Find('Channel');
    XChan        := prSources[s].Session['Idealization'].Find('Channel');
    while not (Chan.IsNull or XChan.IsNull) do begin
      Seg        := Chan.Find('Segment');
      while not Seg.Sibling.IsNull do Seg := Seg.Sibling;
      XSeg       := XChan.Find('Segment');
      while not XSeg.isNull do begin
        Seg      := Chan.InsertClone(Seg, XSeg, True);
        Seg.Data[0] := Seg.Data[0] + n;
        Seg.Data[1] := Seg.Data[1] + n;
        Firsts      := TPAInteger(Seg['Firsts'].DataPtr);
        Lasts       := TPAInteger(Seg['Lasts'].DataPtr);
        DwellCount  := Seg['DwellCount'].Data[0];
        for d := 0 to DwellCount - 1 do begin
          Firsts[d] := Firsts[d] + n;
          Lasts[d]  := Lasts[d] + n;
        end;

        XSeg     := XSeg.SiblingSameName;
      end;
      Chan       := Chan.SiblingSameName;
      XChan      := XChan.SiblingSameName;
    end;
    n            := n + prCounts[s];
  end;
end;

procedure TQUB_Task_Extract_DataSet_FileList.prGetSourceSel(Seg :Integer; var Source :TQUB_SrcSegCount);
begin
  Source    := prSrcSels[Seg];
end;

function TQUB_Task_Extract_DataSet_FileList.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  Sel          :TQUB_SrcSegCount;
begin
  prGetSourceSel(Seg, Sel);
  Result       := prSources[Sel.Src].Load(Buf, Sel.Seg, First, Last);
end;

function TQUB_Task_Extract_DataSet_FileList.prGetChannelCount: Integer;
begin
  Result := prChannelCount;
end;

function TQUB_Task_Extract_DataSet_FileList.prGetSegmentCount: Integer;
begin
  Result := Length(prSrcSels);
end;

function TQUB_Task_Extract_DataSet_FileList.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSegStarts[Ix];
end;

function TQUB_Task_Extract_DataSet_FileList.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSrcSels[Ix].Count;
end;

constructor TQUB_Task_Extract_DataSet_Add.Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);
var
  i         :Integer;
  k         :Integer;
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prSampleCount  := 0;
  prChannelIndex := ChannelIndex;

  for i := 0 to SegmentCount - 1 do begin
    k       := prSource.SegmentLength[i];
    if (prSampleCount = 0) or (k < prSampleCount) then begin
      prSampleCount := k;
    end;
  end;

  Session['SegmentSizes'].SetData( prSampleCount );
  Session.RemoveChild( Session.Find('Lists') );
  Session.RemoveChild(Session['Idealization']);
end;

function TQUB_Task_Extract_DataSet_Add.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  s, c, i      :Integer;
begin
  Result       := True;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := 0.0;
      end;
    end;
  end else begin
    c := prChannelIndex;
    for i := 0 to (Last - First + 1) - 1 do begin
      Buf.Buf[c][i] := 0.0;
    end;
  end;

  for s := 0 to prSource.SegmentCount - 1 do begin
    Buf.Swap;
    if prSource.Load(Buf, s, First, Last) then Exit;
    Buf.Swap;

    if prChannelIndex < 0 then begin
      for c := 0 to ChannelCount - 1 do begin
        for i := 0 to (Last - First + 1) - 1 do begin
          Buf.Buf[c][i] := Buf.Buf[c][i] + Buf.Alt[c][i];
        end;
      end;
    end else begin
      c := prChannelIndex;
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Buf[c][i] + Buf.Alt[c][i];
      end;
    end;
  end;

  Result := False;
end;

function TQUB_Task_Extract_DataSet_Add.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Add.prGetSegmentCount: Integer;
begin
  Result := 1;
end;

function TQUB_Task_Extract_DataSet_Add.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[0];
end;

function TQUB_Task_Extract_DataSet_Add.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSampleCount;
end;

//Operation SUB - subtracts one selection from the source
constructor TQUB_Task_Extract_DataSet_Sub.Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);
var
  i         :Integer;
  k         :Integer;
  List, Sels :INode;
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prSampleCount  := 0;
  prChannelIndex := ChannelIndex;

  for i := 0 to SegmentCount - 1 do begin
    k       := prSource.SegmentLength[i];
    if (prSampleCount = 0) or (k < prSampleCount) then begin
      prSampleCount := k;
    end;
  end;

  Session['SegmentSizes'].SetData( prSampleCount );
  Session.RemoveChild(Session['Idealization']);

  List           := Session.Find('Lists').Find('List');
  if not List.IsNull then begin
    Sels         := List.Find('Sels');
    if LongInt(Sels.Rows) = (SegmentCount + 1) then begin
      Move(TPAInteger(Sels.DataPtr)[2], TPAInteger(Sels.DataPtr)[0], SegmentCount * 2 * SizeOf(Integer));
      Sels.ResizeData( SegmentCount );
      List.RemoveChild( Sels.Sibling ); // first named sel
    end;
  end;
end;

function TQUB_Task_Extract_DataSet_Sub.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  c, i         :Integer;
begin
  Result       := True;

  //This should subtract the last (not first!!!) trace from all other traces. LM.
  //Think about usage: usually the first trace is the one to be subtracted, and qub keeps lists in order by Sel.First -- Chris
  //TO DO: LM. Check this
  if prSource.Load(Buf, SegmentCount, First, Last) then Exit;
  Buf.Swap;
  if prSource.Load(Buf, Seg, First, Last) then Exit;
  Buf.Swap;
  //if prSource.Load(Buf, 0, First, Last) then Exit;
  //Buf.Swap;
  //if prSource.Load(Buf, Seg+1, First, Last) then Exit;
  //Buf.Swap;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Alt[c][i] - Buf.Buf[c][i];
      end;
    end;
  end else begin
    c := prChannelIndex;
    for i := 0 to (Last - First + 1) - 1 do begin
      Buf.Buf[c][i] := Buf.Alt[c][i] - Buf.Buf[c][i];
    end;
  end;

  Result := False;
end;

function TQUB_Task_Extract_DataSet_Sub.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Sub.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount - 1;
end;

function TQUB_Task_Extract_DataSet_Sub.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[0];
end;

function TQUB_Task_Extract_DataSet_Sub.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSampleCount;
end;

//Operation ADD/SUB LIST - subtracts/adds one list of selections from/to the source
constructor TQUB_Task_Extract_DataSet_AddSubList.Create(Source :TQUB_Task_Extract_DataSet; AddSubSource :TQUB_Task_Extract_DataSet; ChannelIndex :Integer; AddSub :Integer);
var
  i              :Integer;
  k              :Integer;
  List, Sels     :INode;
begin
  prSource       := Source;
  prAddSubSource := AddSubSource;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prSampleCount  := 0;
  prChannelIndex := ChannelIndex;
  prAddSub       := AddSub;

  for i := 0 to SegmentCount - 1 do begin
    k       := prSource.SegmentLength[i];
    //if (prSampleCount = 0) or (k < prSampleCount) then begin
    if k > prSampleCount then begin
      prSampleCount := k;
    end;
  end;

  Session['SegmentSizes'].SetData( prSampleCount );
  Session.RemoveChild(Session['Idealization']);

  List           := Session.Find('Lists').Find('List');
  if not List.IsNull then begin
    Sels         := List.Find('Sels');
    if LongInt(Sels.Rows) = (SegmentCount + 1) then begin
      Move(TPAInteger(Sels.DataPtr)[2], TPAInteger(Sels.DataPtr)[0], SegmentCount * 2 * SizeOf(Integer));
      Sels.ResizeData( SegmentCount );
      List.RemoveChild( Sels.Sibling ); // first named sel
    end;
  end;
end;

function TQUB_Task_Extract_DataSet_AddSubList.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  SFirst, SLast :Integer;
  SSeg          :Integer;
  i             :Integer;
begin
  Result        := True;

  //Load Source
  if prSource.Load(Buf, Seg, First, Last) then Exit;
  Buf.Swap;

  SSeg          := -1;

  //Check SubSource
  if prAddSubSource.SegmentCount > 0 then begin
    //Recycle the segments in SubSource;
    //this way, the user can have the same sub list subtracted from several source lists
    SSeg := Seg mod prAddSubSource.SegmentCount;
    //Source and SubSource might be different lengths
    if First <= (prAddSubSource.SegmentLength[SSeg] - 1) then begin
      SFirst := First;
      SLast  := Min(prAddSubSource.SegmentLength[SSeg] - 1, Last);
    end else begin
      SFirst := -1;
      SLast  := -1;
    end;
  end else begin
    SFirst := -1;
    SLast  := -1;
  end;

  //Load SubSource
  if SFirst >= 0 then begin
    if prAddSubSource.Load(Buf, SSeg, SFirst, SLast) then Exit;
  end;
  Buf.Swap;

  //Subtract
  if SFirst >= 0 then begin
    //Do it only for the ActiveChannel
    for i := 0 to (SLast - SFirst + 1) - 1 do begin
      Buf.Buf[ActiveChannel][i] := Buf.Buf[ActiveChannel][i] + prAddSub*Buf.Alt[ActiveChannel][i];
    end;
    {
    if prChannelIndex < 0 then begin
      for c := 0 to ChannelCount - 1 do begin
        for i := 0 to (SLast - SFirst + 1) - 1 do begin
          Buf.Buf[c][i] := Buf.Alt[c][i] - Buf.Buf[c][i];
        end;
      end;
    end else begin
      c := prChannelIndex;
      for i := 0 to (SLast - SFirst + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Alt[c][i] - Buf.Buf[c][i];
      end;
    end;
    }
  end;

  Result := False;
end;

function TQUB_Task_Extract_DataSet_AddSubList.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_AddSubList.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_AddSubList.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[0];
end;

function TQUB_Task_Extract_DataSet_AddSubList.prGetSegmentLength(Ix :Integer): Integer;
begin
  //Result := prSampleCount;
  Result := prSource.SegmentLength[Ix];
end;

//Operation AVGLISTS - averages several lists, the result being another list. All lists must have identical structure
//If OperationFiles is checked, the current lists in each selected file are averaged
constructor TQUB_Task_Extract_DataSet_AvgLists.Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Source :TQUB_Task_Extract_DataSet; SelectionLists :TQUB_QVO_QVD_SelectionLists; ChannelIndex :Integer; Joined :Boolean);
var
  SourceA        :TQUB_Task_Extract_DataSet;
  SelList        :TQUB_QVO_QVD_SelectionList;
  List, Sels     :INode;
  Chunks         :TList;
  Chunk          :TQUB_Thread_PChunk;
  i, k, l        :Integer;
begin

  prData         := Data;
  prSource       := Source;
  prSources      := TList.Create;
  prChunksList   := TList.Create;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prSampleCount  := 0;
  prChannelIndex := ChannelIndex;

  prSourceCount  := 0;
  for i := 0 to SelectionLists.Items.Count - 1 do begin
    //Selected?
    if SelectionLists.Selected[i] then begin

      SelList := SelectionLists[i];

      Chunks  := TList.Create;
      prChunksList.Add(Chunks);

        //Chunk this SelList:
      for l := 0 to SelList.Items.Count - 1 do begin
        Chunk                 := AllocMem(SizeOf(TQUB_Thread_Chunk));
        Chunks.Add(Chunk);
        Chunk.Sel             := SelList.Selection[l].Sel;
        Chunk.Status          := CHUNK_STATUS_EMPTY;
        Chunk.Trials          := 0;
        Chunk.Index           := l;
      end;

      if Joined then begin
        SourceA := TQUB_Task_Extract_DataSet_Chunks_Joined.Create(Task, Data, Chunks);
      end else begin
        SourceA := TQUB_Task_Extract_DataSet_Chunks.Create(Task, Data, Chunks);
      end;
      prSources.Add(SourceA);

      Inc(prSourceCount);

    end;//Selected
  end;//i

  for i := 0 to SegmentCount - 1 do begin
    k       := prSource.SegmentLength[i];
    //if (prSampleCount = 0) or (k < prSampleCount) then begin
    if k > prSampleCount then begin
      prSampleCount := k;
    end;
  end;

  Session['SegmentSizes'].SetData( prSampleCount );
  Session.RemoveChild(Session['Idealization']);

  List           := Session.Find('Lists').Find('List');
  if not List.IsNull then begin
    Sels         := List.Find('Sels');
    if LongInt(Sels.Rows) = (SegmentCount + 1) then begin
      Move(TPAInteger(Sels.DataPtr)[2], TPAInteger(Sels.DataPtr)[0], SegmentCount * 2 * SizeOf(Integer));
      Sels.ResizeData( SegmentCount );
      List.RemoveChild( Sels.Sibling ); // first named sel
    end;
  end;
end;


destructor TQUB_Task_Extract_DataSet_AvgLists.Destroy;
var
  Chunks :TList;
  i, l   :Integer;
begin
  try
    for l := 0 to prSources.Count - 1 do begin
      TQUB_Task_Extract_DataSet(prSources.List[l]).Free;
    end;
    prSources.Free;
    for l := 0 to prChunksList.Count - 1 do begin
      Chunks := prChunksList.List[l];
      for i := 0 to Chunks.Count - 1 do begin
        FreeMem(Chunks.List[i]);
      end;
      Chunks.Clear;
    end;
    prChunksList.Free;
  finally
    inherited;
  end;
end;


function TQUB_Task_Extract_DataSet_AvgLists.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  Count  :Integer;
  i, l   :Integer;
begin
  Result := True;

  Count  := prSources.Count;

  //Load First Source
  if TQUB_Task_Extract_DataSet(prSources.List[0]).Load(Buf, Seg, First, Last) then Exit;
  Buf.Swap;

  for l := 1 to Count - 1 do begin

    //Load Next Source
    if TQUB_Task_Extract_DataSet(prSources.List[l]).Load(Buf, Seg, First, Last) then Exit;
    Buf.Swap;

    //Add:
    //Do it only for the ActiveChannel
    for i := 0 to (Last - First + 1) - 1 do begin
      Buf.Buf[ActiveChannel][i] := Buf.Buf[ActiveChannel][i] + Buf.Alt[ActiveChannel][i];
    end;

    Buf.Swap;

  end;

  //Divide to get the Avg:
  Buf.Swap;
  for i := 0 to (Last - First + 1) - 1 do begin
    Buf.Buf[ActiveChannel][i] := Buf.Buf[ActiveChannel][i] / Count;
  end;

  Result := False;
end;

function TQUB_Task_Extract_DataSet_AvgLists.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_AvgLists.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_AvgLists.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[0];
end;

function TQUB_Task_Extract_DataSet_AvgLists.prGetSegmentLength(Ix :Integer): Integer;
begin
  //Result := prSampleCount;
  Result := prSource.SegmentLength[Ix];
end;

//If OperationFiles is checked, the current lists in each selected file are averaged
constructor TQUB_Task_Extract_DataSet_AvgListsFiles.Create(Task :TQUB_Task_ExtractSelection; Data :TQUB_QVO_QVD; Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer; Joined :Boolean);
var
  SourceA        :TQUB_Task_Extract_DataSet;
  SelList        :TQUB_QVO_QVD_SelectionList;
  List, Sels     :INode;
  Chunks         :TList;
  Chunk          :TQUB_Thread_PChunk;
  ThisData       :TQUB_QVO_QVD;
  i, k, l        :Integer;
begin

  prData         := Data;
  prSource       := Source;
  prSources      := TList.Create;
  prChunksList   := TList.Create;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prSampleCount  := 0;
  prChannelIndex := ChannelIndex;

  prSourceCount  := 0;
  for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
    //Selected?
    if ViewportData.ListObjects.Selected[i] then begin
      ThisData := TQUB_QVO_QVD(ViewportData.ListObjects[i]);
      if ThisData.SelectionLists.ItemIndex <> - 1 then begin

        SelList := ThisData.SelectionLists[ThisData.SelectionLists.ItemIndex];
        Chunks  := TList.Create;
        prChunksList.Add(Chunks);

        //Chunk this SelList:
        for l := 0 to SelList.Items.Count - 1 do begin
          Chunk                 := AllocMem(SizeOf(TQUB_Thread_Chunk));
          Chunks.Add(Chunk);
          Chunk.Sel             := SelList.Selection[l].Sel;
          Chunk.Status          := CHUNK_STATUS_EMPTY;
          Chunk.Trials          := 0;
          Chunk.Index           := l;
        end;

        if Joined then begin
          SourceA := TQUB_Task_Extract_DataSet_Chunks_Joined.Create(Task, ThisData, Chunks);
        end else begin
          SourceA := TQUB_Task_Extract_DataSet_Chunks.Create(Task, ThisData, Chunks);
        end;
        prSources.Add(SourceA);

        Inc(prSourceCount);

      end;//ItemIndex
    end;//Selected
  end;//i

  for i := 0 to SegmentCount - 1 do begin
    k       := prSource.SegmentLength[i];
    if (prSampleCount = 0) or (k < prSampleCount) then begin
      prSampleCount := k;
    end;
  end;

  Session['SegmentSizes'].SetData(prSampleCount);
  Session.RemoveChild(Session['Idealization']);

  List           := Session.Find('Lists').Find('List');
  if not List.IsNull then begin
    Sels         := List.Find('Sels');
    if LongInt(Sels.Rows) = (SegmentCount + 1) then begin
      Move(TPAInteger(Sels.DataPtr)[2], TPAInteger(Sels.DataPtr)[0], SegmentCount * 2 * SizeOf(Integer));
      Sels.ResizeData( SegmentCount );
      List.RemoveChild( Sels.Sibling ); // first named sel
    end;
  end;
end;


//Operation AVG
constructor TQUB_Task_Extract_DataSet_Avg.Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);
var
  i         :Integer;
  k         :Integer;
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prSampleCount  := 0;
  prChannelIndex := ChannelIndex;

  for i := 0 to SegmentCount - 1 do begin
    k       := prSource.SegmentLength[i];
    if (prSampleCount = 0) or (k < prSampleCount) then begin
      prSampleCount := k;
    end;
  end;

  Session['SegmentSizes'].SetData( prSampleCount );
  Session.RemoveChild( Session.Find('Lists') );
end;

function TQUB_Task_Extract_DataSet_Avg.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  s, c, i      :Integer;
  N            :Double;
begin
  Result       := True;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := 0.0;
      end;
    end;
    Buf.Swap;

    for s := 0 to prSource.SegmentCount - 1 do begin
      if prSource.Load(Buf, s, First, Last) then Exit;

      for c := 0 to ChannelCount - 1 do begin
        for i := 0 to (Last - First + 1) - 1 do begin
          Buf.Alt[c][i] := Buf.Alt[c][i] + Buf.Buf[c][i];
        end;
      end;
    end;
    Buf.Swap;

    N := prSource.SegmentCount;
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Buf[c][i] / N;
      end;
    end;
  end else begin
    // load seg 0 on all channels
    c := prChannelIndex;
    if prSource.Load(Buf, 0, First, Last) then Exit;
    Buf.Swap;

    for s := 1 to prSource.SegmentCount - 1 do begin
      if prSource.Load(Buf, s, First, Last) then Exit;

      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Alt[c][i] := Buf.Alt[c][i] + Buf.Buf[c][i];
      end;
    end;
    Buf.Swap;

    N := prSource.SegmentCount;
    for i := 0 to (Last - First + 1) - 1 do begin
      Buf.Buf[c][i] := Buf.Buf[c][i] / N;
    end;
  end;

  Result := False;
end;

function TQUB_Task_Extract_DataSet_Avg.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Avg.prGetSegmentCount: Integer;
begin
  Result := 1;
end;

function TQUB_Task_Extract_DataSet_Avg.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[0];
end;

function TQUB_Task_Extract_DataSet_Avg.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSampleCount;
end;

//Std
constructor TQUB_Task_Extract_DataSet_Std.Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);
var
  i         :Integer;
  k         :Integer;
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prSampleCount  := 0;
  prChannelIndex := ChannelIndex;

  for i := 0 to SegmentCount - 1 do begin
    k       := prSource.SegmentLength[i];
    if (prSampleCount = 0) or (k < prSampleCount) then begin
      prSampleCount := k;
    end;
  end;

  Session['SegmentSizes'].SetData( prSampleCount );
  Session.RemoveChild(Session['Idealization']);
  Session.RemoveChild( Session.Find('Lists') );
end;

function TQUB_Task_Extract_DataSet_Std.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  s, c, i      :Integer;
  N, x         :Double;
  SumVar       :TPMDouble;
  SumVar1      :TPADouble;
begin
  Result       := True;
  SumVar       := nil;
  SumVar1      := nil;

  try

    if prChannelIndex < 0 then begin
    //All channels

      // mean as above
      for c := 0 to ChannelCount - 1 do begin
        for i := 0 to (Last - First + 1) - 1 do begin
          Buf.Buf[c][i] := 0.0;
        end;
      end;
      Buf.Swap;

      for s := 0 to prSource.SegmentCount - 1 do begin
        if prSource.Load(Buf, s, First, Last) then Exit;

        for c := 0 to ChannelCount - 1 do begin
          for i := 0 to (Last - First + 1) - 1 do begin
            Buf.Alt[c][i] := Buf.Alt[c][i] + Buf.Buf[c][i];
          end;
        end;
      end;
      Buf.Swap;

      N := prSource.SegmentCount;
      for c := 0 to ChannelCount - 1 do begin
        for i := 0 to (Last - First + 1) - 1 do begin
          Buf.Buf[c][i] := Buf.Buf[c][i] / N;
        end;
      end;
      Buf.Swap;    // leaving avg in Buf.Alt

      // std dev
      SumVar := DAlloc2(ChannelCount, Last - First + 1);
      for c := 0 to ChannelCount - 1 do begin
        for i := 0 to (Last - First + 1) - 1 do begin
          SumVar[c][i] := 0.0;
        end;
      end;

      for s := 0 to prSource.SegmentCount - 1 do begin
        if prSource.Load(Buf, s, First, Last) then Exit;
        for c := 0 to ChannelCount - 1 do begin
          for i := 0 to (Last - First + 1) - 1 do begin
            x            := Buf.Alt[c][i] - Buf.Buf[c][i];
            SumVar[c][i] := SumVar[c][i] + x * x;
          end;
        end;
      end;

      for c := 0 to ChannelCount - 1 do begin
        for i := 0 to (Last - First + 1) - 1 do begin
          Buf.Buf[c][i] := Sqrt( SumVar[c][i] / N );
        end;
      end;

    end else begin
    //Just one channel


      c := prChannelIndex;
      if prSource.Load(Buf, 0, First, Last) then Exit;
      Buf.Swap;

      for s := 1 to prSource.SegmentCount - 1 do begin
        if prSource.Load(Buf, s, First, Last) then Exit;

        for i := 0 to (Last - First + 1) - 1 do begin
          Buf.Alt[c][i] := Buf.Alt[c][i] + Buf.Buf[c][i];
        end;
      end;
      Buf.Swap;

      N := prSource.SegmentCount;
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Buf[c][i] / N;
      end;
      Buf.Swap;    // leaving avg in Buf.Alt

      // std dev
      SumVar1 := DAlloc1(Last - First + 1);
      for i := 0 to (Last - First + 1) - 1 do begin
        SumVar1[i] := 0.0;
      end;

      for s := 0 to prSource.SegmentCount - 1 do begin
        if prSource.Load(Buf, s, First, Last) then Exit;
        for i := 0 to (Last - First + 1) - 1 do begin
          x            := Buf.Alt[c][i] - Buf.Buf[c][i];
          SumVar1[i]   := SumVar1[i] + x * x;  //Ouch!
        end;
      end;

      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Sqrt(SumVar1[i] / N);
      end;


      (*
      c := prChannelIndex;
      N := prSource.SegmentCount;
      SumVar1 := DAlloc1(Last - First + 1);
      for i := 0 to (Last - First + 1) - 1 do begin
        SumVar1[i] := 0.0;
      end;
      if prSource.Load(Buf, 0, First, Last) then Exit;
      Buf.Swap;

      for s := 1 to prSource.SegmentCount - 1 do begin
        if prSource.Load(Buf, s, First, Last) then Exit;
        for i := 0 to (Last - First + 1) - 1 do begin
          SumVar1[i] := SumVar1[i] + sqr(Buf.Alt[c][i] - Buf.Buf[c][i]) / 2;
          Buf.Buf[c][i] := Buf.Alt[c][i];
        end;
      end;
      Buf.Swap;

      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := sqrt(SumVar1[i] / N);
      end;
      *)


    end;

    Result := False;

  finally
    Free2(TPMChar(SumVar));
    FreeMem(SumVar1);
  end;
end;

function TQUB_Task_Extract_DataSet_Std.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Std.prGetSegmentCount: Integer;
begin
  Result := 1;
end;

function TQUB_Task_Extract_DataSet_Std.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[0];
end;

function TQUB_Task_Extract_DataSet_Std.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSampleCount;
end;


constructor TQUB_Task_Extract_DataSet_Scale.Create(Source :TQUB_Task_Extract_DataSet; Scale :Double; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  // TODO: scale idealization amps, stds
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prScale        := Scale;
  prChannelIndex := ChannelIndex;
end;

function TQUB_Task_Extract_DataSet_Scale.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  c, i      :Integer;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Buf[c][i] * prScale;
      end;
    end;
  end else begin
    c := prChannelIndex;
    for i := 0 to (Last - First + 1) - 1 do begin
      Buf.Buf[c][i] := Buf.Buf[c][i] * prScale;
    end;
  end;
end;

function TQUB_Task_Extract_DataSet_Scale.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Scale.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_Scale.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[0];
end;

function TQUB_Task_Extract_DataSet_Scale.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;

//Drift
constructor TQUB_Task_Extract_DataSet_Drift.Create(Source :TQUB_Task_Extract_DataSet; A, B :Double; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  Session.RemoveChild(Session['Idealization']);
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prA            := A;
  prB            := B;
  prChannelIndex := ChannelIndex;
end;

function TQUB_Task_Extract_DataSet_Drift.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  c, i      :Integer;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Buf[c][i] + prA * (First+i) * Sampling + prB;
      end;
    end;
  end else begin
    c := prChannelIndex;
    for i := 0 to (Last - First + 1) - 1 do begin
      Buf.Buf[c][i] := Buf.Buf[c][i] + prA * (First+i) * Sampling + prB;
    end;
  end;
end;

function TQUB_Task_Extract_DataSet_Drift.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Drift.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_Drift.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_Drift.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;

//Derivate
constructor TQUB_Task_Extract_DataSet_Derivate.Create(Source :TQUB_Task_Extract_DataSet; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  Session.RemoveChild(Session['Idealization']);
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prChannelIndex := ChannelIndex;
end;

function TQUB_Task_Extract_DataSet_Derivate.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  v, d      :Double;
  c, i      :Integer;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      v := Buf.Buf[c][First - First];
      for i := First - First to Last - First do begin
        //Assume it loads the entire segment
        if i = First then begin
          d := (Buf.Buf[c][i + 1] - Buf.Buf[c][i]) / Sampling;
        end else if i = Last then begin
          d := (Buf.Buf[c][i] - v) / Sampling;
        end else begin
          d := 0.5*(Buf.Buf[c][i + 1] - Buf.Buf[c][i]) / Sampling + 0.5*(Buf.Buf[c][i] - v) / Sampling;
        end;
        v := Buf.Buf[c][i];
        Buf.Buf[c][i] := d;
      end;
    end;
  end else begin
    c := prChannelIndex;
    v := Buf.Buf[c][First - First];
    for i := First - First to Last - First do begin
      if i = First then begin
        d := (Buf.Buf[c][i + 1] - Buf.Buf[c][i]) / Sampling;
      end else if i = Last then begin
        d := (Buf.Buf[c][i] - v) / Sampling;
      end else begin
        d := 0.5*(Buf.Buf[c][i + 1] - Buf.Buf[c][i]) / Sampling + 0.5*(Buf.Buf[c][i] - v) / Sampling;
      end;
      v := Buf.Buf[c][i];
      Buf.Buf[c][i] := d;
    end;
  end;
end;

function TQUB_Task_Extract_DataSet_Derivate.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Derivate.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_Derivate.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_Derivate.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;

//LKC
constructor TQUB_Task_Extract_DataSet_LKC.Create(Source, SubSource :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; ChannelIndex :Integer);
begin
  prSource       := Source;
  prVSource      := SubSource;
  Session        := Source.Session;
  Session.RemoveChild(Session['Idealization']);
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prChannelIndex := ChannelIndex;

  prData         := Data;

  if TaskExpPropertiesForm <> nil then begin
    with TaskExpPropertiesForm do begin
      prLKC_I    := DAlloc1(LKCN);
    end;//with
  end;//if
end;

destructor TQUB_Task_Extract_DataSet_LKC.Destroy;
begin
  try
    FreeMem(prLKC_I);
  finally
    inherited;
  end;
end;

{
LM. This is the old style LKC correction, which only works for one segment following a jump in voltage.
I'm replacing it here with a version that follows the changes in voltage, whatever they are.
Usage: first, calculate the VL, GL and the capacitive Ai,and taui, using ExpFit(LKC), then extract
}
(*
function TQUB_Task_Extract_DataSet_LKC.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  Sel       :TQUB_FirstLast;
  V, V0     :Double;
  SumExp    :Double;
  c, i      :Integer;
  t         :Double;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  Result := TaskExpPropertiesForm = nil;
  if Result then Exit;

  with TaskExpPropertiesForm do begin

    V  := 0;
    V0 := 0;

    if First = 0 then begin
      Sel.First  := TQUB_Thread_PChunk(prChunks[Seg]).Sel.First + First;
      Sel.Last   := TQUB_Thread_PChunk(prChunks[Seg]).Sel.First + Last;
      //V
      V    := prData.Average(LKCVChannel{StimCh}, Sel.First + LKCSkipPoints, Sel.Last - LKCSkipPoints);
      //V0
      V0   := prData.Average(LKCVChannel{StimCh}, Sel.First - LKCV0Points + LKCSkipPoints, Sel.First - LKCSkipPoints);
    end;


    for i := First to Last do begin
      SumExp := 0;
      t := i * Sampling;
      for c := 0 to LKCN - 1 do begin
        SumExp := SumExp + (V-V0)*LKCAmp[c]*exp(-t/LKCTau[c]);
      end;
      //Only do it for ActiveChannel
      Buf.Buf[ActiveChannel][i-First] := Buf.Buf[ActiveChannel][i-First] - LKCGL*(V-LKCVL) - SumExp;
      {
      if prChannelIndex < 0 then begin
        for c := 0 to ChannelCount - 1 do begin
          Buf.Buf[c][i-First] := Buf.Buf[c][i-First] - LKCGL*(V-LKCVL) - SumExp;
        end;//c
      end else begin
        c := prChannelIndex;
        Buf.Buf[c][i-First] := Buf.Buf[c][i-First] - LKCGL*(V-LKCVL) - SumExp;
      end;//ChIdx
      }
    end;//i

  end;//with

end;
*)

function TQUB_Task_Extract_DataSet_LKC.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  V         :Double;
  IR        :Double;
  SumIc     :Double;
  c, i      :Integer;
  Nv        :Integer;
  //t         :Double;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  Result := TaskExpPropertiesForm = nil;
  if Result then Exit;

  with TaskExpPropertiesForm do begin

    if First = 0 then begin
      //Initialize the Ic components
      for i := 0 to LKCN - 1 do begin
        prLKC_I[i] := 0;
      end;

      // load prior V
      Nv        := LKCV0Points - (2 * LKCSkipPoints) + 1;
      if Nv > 0 then begin
        if (prVBuf = nil) or (prVBuf.ColCount < Nv) then begin
          prVBuf.Free;
          prVBuf := TQUB_DBuf2_Double.Create(prVSource.ChannelCount, Nv);
        end;
        Result    := prVSource.Load(prVBuf, Seg, 0, Nv);
        if Result then Exit;

        // Average the V
        V := 0;
        for i := 0 to Nv - 1 do begin
          V := V + prVBuf.Buf[0][i];
        end;
        prLKC_V0   := V / Nv;
      end else begin
        prLKC_V0   := Buf.Buf[LKCVChannel][0];
      end;
    end;

    for i := First - First to Last - First do begin

      //Calculate the V
      V := Buf.Buf[LKCVChannel][i];

      //Calculate the resistive current
      IR := LKCGL*(V - LKCVL);

      //Update the capacitive currents, and calculate their sum
      SumIc := 0;
      for c := 0 to LKCN - 1 do begin
        prLKC_I[c] := (prLKC_I[c] + (V - prLKC_V0)*LKCAmp[c])*exp(-Sampling/LKCTau[c]);
        SumIc      := SumIc + prLKC_I[c];
      end;

      //Make the correction, but only for ActiveChannel
      Buf.Buf[ActiveChannel][i] := Buf.Buf[ActiveChannel][i] - IR - SumIc;

      //Update V0
      prLKC_V0 := V;

    end;//i

  end;//with TaskExpPropertiesForm

end;

function TQUB_Task_Extract_DataSet_LKC.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_LKC.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_LKC.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_LKC.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;


//I to G
constructor TQUB_Task_Extract_DataSet_ItoG.Create(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; VRev :Double; VCh :Integer; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prVRev         := VRev;
  prVCh          := VCh;
  prChannelIndex := ChannelIndex;

  prData         := Data;

end;

function TQUB_Task_Extract_DataSet_ItoG.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  V :Double;
  i :Integer;
begin
  Result := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  for i := First - First to Last - First do begin

    //Calculate the V
    V := Buf.Buf[prVCh][i];

    //Make the conversion, but only for ActiveChannel
    if Abs(V - prVRev) > 1e-3 then begin
      Buf.Buf[ActiveChannel][i] := Buf.Buf[ActiveChannel][i] / (V - prVRev);
    end;

  end;//i

end;

function TQUB_Task_Extract_DataSet_ItoG.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_ItoG.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_ItoG.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_ItoG.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;

//Rs Correction
constructor TQUB_Task_Extract_DataSet_RsCorr.Create(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; RsValue, RsPer :Double; VCh :Integer; AddLeak :Boolean; LeakCh :Integer; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prRsValue      := RsValue;
  prRsPer        := RsPer;
  prVCh          := VCh;
  prRsAddLeak    := AddLeak;
  prRsLeakCh     := LeakCh;
  prChannelIndex := ChannelIndex;

  prData         := Data;

end;

function TQUB_Task_Extract_DataSet_RsCorr.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  Curr :Double;
  Rs   :Double;
  i    :Integer;
begin
  Result := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  Rs := prRsValue*(1-prRsPer/100);
  for i := First - First to Last - First do begin

    Curr := Buf.Buf[ActiveChannel][i]{Current};
    if prRsAddLeak then Curr := Curr + Buf.Buf[prRsLeakCh][i]{LeakCurrent};
    //Make the correction for the Voltage channel
    Buf.Buf[prVCh][i] := Buf.Buf[prVCh][i] - Curr * Rs;

  end;//i

end;

function TQUB_Task_Extract_DataSet_RsCorr.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_RsCorr.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_RsCorr.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_RsCorr.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;

//P/n protocol
constructor TQUB_Task_Extract_DataSet_PnAdd.Create(Source :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; Offset, Delay, Scale :Double; Count :Integer; Vm, Interval :Double; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  Session.RemoveChild(Session['Idealization']);
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;

  prOffset       := Offset;
  prDelay        := Delay*1e-3;
  prScale        := Scale;
  prCount        := Count;
  prVm           := Vm;
  prInterval     := Interval*1e-3;

  prChannelIndex := ChannelIndex;

  prData         := Data;

  prCurrSeg      := -1;
  prFirstTime    := True;
  prBuffer       := nil;
  prBufferCount  := 0;
end;

destructor TQUB_Task_Extract_DataSet_PnAdd.Destroy;
begin
  try
    prBuffer.Free;
  finally
    inherited;
  end;
end;

function TQUB_Task_Extract_DataSet_PnAdd.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  SegLen     :Integer;
  SourceData :TQUB_FirstLast;
  PnData     :TQUB_FirstLast;
  PnDataLen  :Integer;
  PnIntLen   :Integer;
  SecStart   :Integer;
  i, c       :Integer;
begin

  Result := False;
  SegLen := prSource.SegmentLength[Seg];

  if Seg <> prCurrSeg then begin
  //Loading from a new segment

    prCurrSeg := Seg;

    //Re-allocate Buffer is necessary
    if (prBuffer = nil) or (prBufferCount < SegLen) then begin
      prBuffer.Free;
      prBuffer := nil;
      prBufferCount := SegLen;
      prBuffer := TQUB_DBuf2_Double.Create(prSource.ChannelCount, prBufferCount);
    end;

    //Load data from Source
    Result := prSource.Load(prBuffer, Seg, 0, SegLen - 1);
    if Result then Exit;

    //Find the minimum
    prMin := prBuffer.Buf[0{Channel???}][0];
    for i := 1 to SegLen - 1 do begin
      if prMin > prBuffer.Buf[0][i] then prMin := prBuffer.Buf[0][i];
    end;

  end;

  PnIntLen  := Round(Abs(prDelay) / Sampling);
  PnDataLen := SegLen + PnIntLen;

  //Calculate the pointers
  if prDelay >= 0 then begin
    SourceData.First := 0;
    SourceData.Last  := SegLen - 1;
    PnData.First     := SourceData.Last + 1;
    PnData.Last      := SegmentLength[Seg] - 1;
  end else begin
    SourceData.First := SegmentLength[Seg] - SegLen;
    SourceData.Last  := SegmentLength[Seg] - 1;
    PnData.First     := 0;
    PnData.Last      := SourceData.First - 1;
  end;


  for i := First to Last do begin
  //Source

    //Where in the segment:

    if (i >= SourceData.First) and (i <= SourceData.Last) then begin
    //Source Data

      for c := 0 to ChannelCount - 1 do begin
        Buf.Buf[c][i-First] := prBuffer.Buf[c][i-SourceData.First];
      end;

    end else if (i >= PnData.First) and (i <= PnData.Last) then begin
    //Pn Data

      //Which section?
      SecStart := PnDataLen * ((i - PnData.First) div PnDataLen) + PnData.First;

      if prDelay < 0 then begin
        if ((i - SecStart) >= SegLen) then begin
          for c := 0 to ChannelCount - 1 do begin
            if c = prChannelIndex then begin
              Buf.Buf[c][i-First] := prVm;
            end else begin
              Buf.Buf[c][i-First] := 0;
            end;
          end;
        end else begin
          //Subtract Min, Scale and Offset
          for c := 0 to ChannelCount - 1 do begin
            if c = prChannelIndex then begin
              Buf.Buf[c][i-First] := (prBuffer.Buf[c][i-SecStart] - prMin) * prScale + prOffset;
            end else begin
              Buf.Buf[c][i-First] :=  prBuffer.Buf[c][i-SecStart];
            end;
          end;
        end;
      end else if prDelay >= 0 then begin
        if ((i - SecStart) < PnIntLen) then begin
          for c := 0 to ChannelCount - 1 do begin
            if c = prChannelIndex then begin
              Buf.Buf[c][i-First] := prVm;
            end else begin
              Buf.Buf[c][i-First] := 0;
            end;
          end;
        end else begin
          for c := 0 to ChannelCount - 1 do begin
            if c = prChannelIndex then begin
              Buf.Buf[c][i-First] := (prBuffer.Buf[c][i-SecStart - PnIntLen] - prMin) * prScale + prOffset;
            end else begin
              Buf.Buf[c][i-First] :=  prBuffer.Buf[c][i-SecStart - PnIntLen];
            end;
          end;
        end;
      end;

    end;

  end;   

end;

function TQUB_Task_Extract_DataSet_PnAdd.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_PnAdd.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_PnAdd.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_PnAdd.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
  Result := Result + prCount * (Result + Round(Abs(prDelay) / Sampling));
end;

//Sub
function TQUB_Task_Extract_DataSet_PnSub.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  SegLen     :Integer;
  PnDataLen  :Integer;
  PnIntLen   :Integer;
  SecStart   :Integer;
  i, c       :Integer;
begin

  SegLen := SegmentLength[Seg];

  PnIntLen  := Round(Abs(prDelay) / Sampling);
  PnDataLen := SegLen + PnIntLen;

  if Seg <> prCurrSeg then begin
  //Loading from a new segment

    prCurrSeg := Seg;

    //Re-allocate Buffer is necessary
    if (prBuffer = nil) or (prBufferCount < SegLen) then begin
      prBuffer.Free;
      prBuffer := nil;
      prBufferCount := SegLen;
      prBuffer := TQUB_DBuf2_Double.Create(prSource.ChannelCount, prBufferCount);
    end;

    if prDelay >= 0 then begin
      SecStart := PnDataLen;
    end else begin
      SecStart := 0;
    end;

    prBuffer.Swap;
    for c := 0 to SegLen - 1 do begin
      prBuffer.Alt[ActiveChannel][c] := 0;
    end;
    for i := 0 to prCount - 1 do begin
      //Load data from Source
      Result := prSource.Load(prBuffer, Seg, SecStart, SecStart + SegLen - 1);
      if Result then Exit;
      for c := 0 to SegLen - 1 do begin
        prBuffer.Alt[ActiveChannel][c] := prBuffer.Alt[ActiveChannel][c] + prBuffer.Buf[ActiveChannel][c];
      end;
      SecStart := SecStart + PnDataLen;
    end;
    prBuffer.Swap;
    for c := 0 to SegLen - 1 do begin
      prBuffer.Buf[ActiveChannel][c] := prBuffer.Buf[ActiveChannel][c] / prCount;
    end;

  end;

  Result := prSource.Load(Buf, Seg, First, Last);

  for i := First to Last do begin
  //Source

    Buf.Buf[ActiveChannel][i-First] := Buf.Buf[ActiveChannel][i-First] - prBuffer.Buf[ActiveChannel][i] / prScale;

  end;

end;

function TQUB_Task_Extract_DataSet_PnSub.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_PnSub.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
  Result := (Result - prCount * Round(Abs(prDelay) / Sampling)) div (prCount + 1);
end;


//Fit
constructor TQUB_Task_Extract_DataSet_Fit.Create(Source, FitSource :TQUB_Task_Extract_DataSet; Data :TQUB_QVO_QVD; ChannelIndex :Integer);
begin
  prSource       := Source;
  prFitSource    := FitSource;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prChannelIndex := ChannelIndex;
  prData         := Data;
  prFitBuf       := nil;
end;

destructor TQUB_Task_Extract_DataSet_Fit.Destroy;
begin
  prFitBuf.Free;
  inherited;
end;

function TQUB_Task_Extract_DataSet_Fit.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  c, i           :Integer;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  if (prFitBuf = nil) or (prFitBuf.ColCount < (Last - First + 1)) then begin
    prFitBuf.Free;
    prFitBuf := TQUB_DBuf2_Double.Create(prFitSource.ChannelCount, 2*(Last-First+1));
  end;
  Result    := prFitSource.Load(prFitBuf, Seg, First, Last);
  if Result then Exit;

  c := ActiveChannel;
  for i := 0 to Last - First do begin
    Buf.Buf[c][i] := Buf.Buf[c][i] - prFitBuf.Buf[1][i];
  end;
end;

function TQUB_Task_Extract_DataSet_Fit.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Fit.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_Fit.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_Fit.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;


//Sine
constructor TQUB_Task_Extract_DataSet_Sine.Create(Source :TQUB_Task_Extract_DataSet; Hz, Amp, Deg :Double; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prHz           := Hz;
  prAmp          := Amp;
  prDeg          := Deg;
  prChannelIndex := ChannelIndex;
end;

function TQUB_Task_Extract_DataSet_Sine.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  c, i         :Integer;
  Amp, Frq, Ph :Double;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  Amp       := prAmp;
  Frq       := prHz * 2 * Pi * Sampling;
  Ph        := prDeg * Pi / 180.0;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        Buf.Buf[c][i] := Buf.Buf[c][i] + Amp * Sin(Frq * (First+i) + Ph);
      end;
    end;
  end else begin
    c := prChannelIndex;
    for i := 0 to (Last - First + 1) - 1 do begin
      Buf.Buf[c][i] := Buf.Buf[c][i] + Amp * Sin(Frq * (First+i) + Ph);
    end;
  end;
end;

function TQUB_Task_Extract_DataSet_Sine.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Sine.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_Sine.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_Sine.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;

//Digitize
constructor TQUB_Task_Extract_DataSet_Digitize.Create(Source :TQUB_Task_Extract_DataSet; Thr, Min, Max :Double; ChannelIndex :Integer);
begin
  prSource       := Source;
  Session        := Source.Session;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling;
  prThr          := Thr;
  prMin          := Min;
  prMax          := Max;
  prChannelIndex := ChannelIndex;
end;

function TQUB_Task_Extract_DataSet_Digitize.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  c, i            :Integer;
  Thr, Min, Max   :Double;
begin
  Result    := prSource.Load(Buf, Seg, First, Last);
  if Result then Exit;

  Thr             := prThr;
  Min             := prMin;
  Max             := prMax;

  if prChannelIndex < 0 then begin
    for c := 0 to ChannelCount - 1 do begin
      for i := 0 to (Last - First + 1) - 1 do begin
        if Buf.Buf[c][i] >= Thr then begin
          Buf.Buf[c][i] := Max;
        end else begin
          Buf.Buf[c][i] := Min;
        end;
      end;
    end;
  end else begin
    c := prChannelIndex;
    for i := 0 to (Last - First + 1) - 1 do begin
      if Buf.Buf[c][i] >= Thr then begin
        Buf.Buf[c][i] := Max;
      end else begin
        Buf.Buf[c][i] := Min;
      end;
    end;
  end;
end;

function TQUB_Task_Extract_DataSet_Digitize.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Digitize.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_Digitize.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_Digitize.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := prSource.SegmentLength[Ix];
end;

//Resample
constructor TQUB_Task_Extract_DataSet_Resample.Create(Source :TQUB_Task_Extract_DataSet; Factor :Double);
var
  SegSizes, PPT  :INode;
  i, p           :Integer;
  Lists, List, Sels, Sel :INode;
begin
  prSource       := Source;
  ActiveChannel  := Source.ActiveChannel;
  Sampling       := Source.Sampling * Factor;
  Session        := Source.Session.Clone(True);
  prFactor       := Factor;

  Session['Sampling'].SetData( Sampling );
  SegSizes       := Session['SegmentSizes'];
  for i := 0 to SegmentCount - 1 do begin
    SegSizes.Data[i]  := SegmentLength[i];
  end;

  PPT := Session['Display']['PointsPerTrace'];
  PPT.Data[0] := Integer(Max(10, Round(PPT.Data[0] / Factor)));

  Lists          := Session.Find('Lists');
  List           := Lists.Find('List');
  if not List.IsNull then begin
    Sels       := Lists.Find('Sels');
    Sel        := Sels.Sibling;
    if Integer(Sels.Rows) = SegmentCount then begin
      p        := 0;
      for i := 0 to SegmentCount - 1 do begin
        Sels.MatrixData[i, 0] := p;
        if not Sel.IsNull then Sel.Data[0] := p;
        p      := p + SegmentLength[i];
        Sels.MatrixData[i, 1] := p-1;
        if not Sel.IsNull then Sel.Data[1] := p-1;
        Sel    := Sel.Sibling;
      end;
    end else begin
      Lists.RemoveChild(List);
    end;
  end;

  prResampleIdealization;
end;

procedure TQUB_Task_Extract_DataSet_Resample.prResampleIdealization;
var
  Chan, Seg              :INode;
  Firsts, Lasts, Classes :TPAInteger;
  Durations              :TPASingle;
  DwellCount             :Integer;
  f, l, ff, i, j         :Integer;
  iFactor, SamplingMS    :Double;
begin
  iFactor    := 1.0 / prFactor;
  SamplingMS := 1e3 * Session['Sampling'].Data[0];

  // carefully resize idl segments and dwells...
  Chan := Session['Idealization'].Find('Channel');
  while not Chan.IsNull do begin
    Seg := Chan.Find('Segment');
    while not Seg.IsNull do begin
      f    := Round(Seg.Data[0] * iFactor);
      l    := Round((Seg.Data[1] + 1) * iFactor) - 1;
      Seg.Data[0] := f;
      Seg.Data[1] := l;

      Firsts      := TPAInteger(Seg['Firsts'].DataPtr);
      Lasts       := TPAInteger(Seg['Lasts'].DataPtr);
      Classes     := TPAInteger(Seg['Classes'].DataPtr);
      Durations   := TPASingle(Seg['Durations'].DataPtr);
      DwellCount  := Seg['DwellCount'].Data[0];

      // rescaling indices: round(ix/factor) for ix in Firsts, Lasts;
      // drop events whose first == next first or > p; consolidating
      j     := -1; // index of last written event
      for i := 0 to DwellCount - 1 do begin
        ff  := Round(Firsts[i] * iFactor);
        if (j >= 0) and (Classes[j] = Classes[i]) then begin
          Lasts[j]   := Round((Lasts[i]+1) * iFactor) - 1;
          Durations[j] := SamplingMS * (Lasts[j] - Firsts[j] + 1);
        end else begin
          if (j < 0) or (Firsts[j] <> ff) then Inc(j);
          // else (if the last event also starts at ff) overwrite the last event
          Firsts[j]    := ff;
          Lasts[j]     := Round((Lasts[i] + 1) * iFactor) - 1;
          Classes[j]   := Classes[i];
          Durations[j] := SamplingMS * (Lasts[j] - ff + 1);
        end;
      end;
      DwellCount  := j + 1;

      // make sure first event starts on f, last event ends on l
      Firsts[0] := f;
      Lasts[DwellCount-1] := l;

      // resize
      Seg['DwellCount'].Data[0] := DwellCount;
      Seg['Firsts'].ResizeData(DwellCount);
      Seg['Lasts'].ResizeData(DwellCount);
      Seg['Classes'].ResizeData(DwellCount);
      Seg['Durations'].ResizeData(DwellCount);

      Seg := Seg.SiblingSameName;
    end;
    Chan := Chan.SiblingSameName;
  end;
end;

function TQUB_Task_Extract_DataSet_Resample.Load(Buf :TQUB_DBuf2_Double; Seg, First, Last :Integer): Boolean;
var
  SrcBuf     :TQUB_DBuf2_Double;
  AtATime    :Integer;
  At         :Integer;
  SrcFirst   :Integer;
  SrcLast    :Integer;
  c, i       :Integer;
  iSrc       :Double;
  il, ir     :Integer; // left- and right-correspondents of i
  pl, pr     :Double;  // pct contribution of il, ir for linear interpolation
begin
  SrcBuf     := nil;
  Result     := False;

  try
    SrcBuf   := TQUB_DBuf2_Double.Create(Buf.RowCount, Buf.ColCount+2);

    AtATime  := Last - First + 1;
    At       := First;
    while At <= Last do begin
      //TO DO: LM. Check if First should be there
      // no I think it was a bug that only struck with huge data files -- Chris
      SrcFirst  := Floor(prFactor * ({First + }At));
      SrcLast   := Min( Ceil(prFactor * Min(Last, {First + }At + AtATime - 1)),
                        prSource.SegmentLength[Seg] - 1 );
      while (SrcLast - SrcFirst + 1) >= SrcBuf.ColCount do begin
        AtATime := AtATime div 2;
        SrcLast := Min( Ceil(prFactor * Min(Last, {First + }At + AtATime - 1)),
                        prSource.SegmentLength[Seg] - 1 );
      end;
      AtATime   := Min(AtATime, Last - At + 1);

      Result    := prSource.Load(SrcBuf, Seg, SrcFirst, SrcLast);
      if Result then Break;

      for i := 0 to AtATime - 1 do begin
        iSrc    := prFactor * ({First + }At + i);
        il      := Floor(iSrc) - SrcFirst;
        ir      := Min( Ceil(iSrc), prSource.SegmentLength[Seg] - 1 ) - SrcFirst;
        pr      := Frac(iSrc);
        pl      := 1.0 - pr;
        for c := 0 to ChannelCount - 1 do begin
          Buf.Buf[c][At-First{no}+i] := pl * SrcBuf.Buf[c][il] + pr * SrcBuf.Buf[c][ir];
        end;
      end;

      At        := At + AtATime;
    end;
  except
    Result   := True;
  end;

  if SrcBuf <> nil then SrcBuf.Free;
end;

function TQUB_Task_Extract_DataSet_Resample.prGetChannelCount: Integer;
begin
  Result := prSource.ChannelCount;
end;

function TQUB_Task_Extract_DataSet_Resample.prGetSegmentCount: Integer;
begin
  Result := prSource.SegmentCount;
end;

function TQUB_Task_Extract_DataSet_Resample.prGetSegmentStart(Ix :Integer): Double;
begin
  Result := prSource.SegmentStart[Ix];
end;

function TQUB_Task_Extract_DataSet_Resample.prGetSegmentLength(Ix :Integer): Integer;
begin
  Result := Floor((prSource.SegmentLength[Ix] - 1) / prFactor) + 1;
end;


//******************************************************************************
constructor TQUB_Task_Extract_File.Create(ChannelCount :Integer; Sampling :Double; Data :TQUB_QVO_QVD; ChannelList :TList);
begin
  prChannelCount := ChannelCount;
  prSampling     := Sampling;
  prFileOpen     := False;
  prData         := Data;
  prChannels     := ChannelList;
end;

destructor TQUB_Task_Extract_File.Destroy;
begin
  try
    if prFileOpen then Close;
  finally
    inherited;
  end;
end;

function TQUB_Task_Extract_File.Open(FileName :String): Boolean;
begin
  try
    AssignFile(Fi, FileName);
    prFileOpen        := True;
    ReWrite(Fi);
    Result            := False;
  except
    Result            := True;
  end;
end;

function TQUB_Task_Extract_File.Close: Boolean;
begin
  try
    if prFileOpen then CloseFile(Fi);
    prFileOpen        := False;
    Result            := False;
  except
    Result            := True;
  end;
end;


constructor TQUB_Task_Extract_File_Int.Create(ChannelCount :Integer; Sampling :Double; DataSize :Integer; DataType :TQUB_QFS_Node_DataType; Scaling :Double; Channels :INode; Data :TQUB_QVO_QVD; ChannelList :TList);
var
  Chan           :INode;
  i              :Integer;
begin
  inherited Create(ChannelCount, Sampling, Data, ChannelList);

  prDataSize     := DataSize;
  prDataType     := DataType;
  prScaling      := Scaling;

  SetLength(prChannelScaling, ChannelCount);
  i              := 0;
  Chan           := Channels.Find('Channel');

  while (i < ChannelCount) and not Chan.IsNull do begin
    prChannelScaling[i] := Chan['Scaling'].Data[0];
    Inc(i);
    Chan         := Chan.SiblingSameName;
  end;
end;


function TQUB_Task_Extract_File_QDF.Open(FileName :String): Boolean;
begin
  QDF                 := TNode.Create('DataFile');
  Result              := not QDF.SaveAs(FileName);
  QDFSeg              := QUB_Node_NULL;
  QDFChan             := QUB_Node_NULL;
end;

function TQUB_Task_Extract_File_QDF.WriteFileHeader: Boolean;
begin
  QDFSegs             := QDF['Segments'];
  QDF['Sampling'].SetData( prSampling );
  QDF['Scaling'].SetData( prScaling );
  QDF['ADChannelCount'].SetData( prChannelCount );
  QDF['ADDataSize'].SetData( prDataSize );
  QDF['ADDataType'].SetData( Integer(prDataType) );
  Result              := False;
end;

function TQUB_Task_Extract_File_QDF.WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean;
begin
  QDFSeg              := QDFSegs.InsertChild(QDFSeg, 'Segment');
  QDFSeg['StartTime'].SetData( StartTime );
  QDFChan             := QDFSeg['Channels'];
  QDFChan.Preload     := False;
  QDFChan.SetupNumData(prDataType, SampleCount, prChannelCount);
  SamplesWritten      := 0;
  Result              := False;
end;

function TQUB_Task_Extract_File_QDF.WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean;
var
  BufferInt           :TPMInteger;
  c, i                :Integer;
  Scal                :Double;
  Sel                 :TQUB_FirstLast;
begin
  Sel.First           := SamplesWritten;
  Sel.Last            := SamplesWritten + SampleCount - 1;
  SamplesWritten      := SamplesWritten + SampleCount;
  Result              := not QDFChan.LoadRows(Sel, False);
  if Result then Exit;

  BufferInt          := nil;
  try
    BufferInt        := IAlloc2(prChannelCount, SampleCount);
    for c := 0 to prChannelCount - 1 do begin
      Scal := Scaling * prChannelScaling[c];
      for i := 0 to SampleCount - 1 do begin
        BufferInt[c][i] := Round(Scal * Buffer[c][i]);
      end;
    end;

    MPXData(BufferInt, QDFChan.DataPtr, prChannelCount, -1{all ch}, prDataSize, prDataType, 0, SampleCount-1, True{MPX});
    Result           := False;
  except
    Result           := True;
  end;

  Free2(TPMChar(BufferInt));
end;

function TQUB_Task_Extract_File_QDF.Close: Boolean;
begin
  Result              := False;
  if QDF = nil then Exit;

  Result              := not QDF.Save;
  QDFChan             := nil;
  QDFSeg              := nil;
  QDFSegs             := nil;
  QDF                 := nil;
end;


constructor TQUB_Task_Extract_File_LDT.Create(ChannelCount :Integer; Sampling :Double; Scaling :Double; Channels :INode; Data :TQUB_QVO_QVD; ChannelList :TList);
begin
  inherited Create(ChannelCount, Sampling, 2, NODE_DATATYPE_SMALLINT, Scaling, Channels, Data, ChannelList);

  prIntScal          := Round(prScaling * 100);
  prRescaling        := 1;
  while prIntScal > 32767 do begin
    prIntScal        := prIntScal div 2;
    prRescaling      := prRescaling * 2;
  end;
  Self.Scaling       := Scaling / prRescaling;
end;

function TQUB_Task_Extract_File_LDT.WriteFileHeader: Boolean;
var
  Leader             :Integer;
  Samp, Scal         :Word;
begin
  Leader             := 0;
  Samp               := Round(prSampling * 1.0e6);
  Scal               := Word(prIntScal);

  try
    BlockWrite(Fi, Leader, SizeOf(Leader));
    BlockWrite(Fi, Samp, SizeOf(Samp));
    BlockWrite(Fi, Scal, SizeOf(Scal));
    Result           := False;
  except
    Result           := True;
  end;
end;

function TQUB_Task_Extract_File_LDT.WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean;
var
  StartInt           :Integer;
begin
  StartInt           := Round(StartTime * 1.0e6);
  try
    BlockWrite(Fi, StartInt, SizeOf(StartInt));
    BlockWrite(Fi, SampleCount, SizeOf(SampleCount));
    Result           := False;
  except
    Result           := True;
  end;
end;

function TQUB_Task_Extract_File_LDT.WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean;
var
  IBuf               :TPASmall;
  i                  :Integer;
  Scal               :Double;
begin
  IBuf               := nil;
  Scal               := Scaling * prChannelScaling[0];
  try
    IBuf             := SmAlloc1(SampleCount);
    for i := 0 to SampleCount - 1 do begin
      IBuf[i]        := Round(Scal * Buffer[0][i]);
    end;
    BlockWrite(Fi, IBuf^, prDataSize * SampleCount);
    Result           := False;
  except
    Result           := True;
  end;
  FreeMem(IBuf);
end;


function TQUB_Task_Extract_File_DAT.WriteFileHeader: Boolean;
begin
  Result := False;
end;

function TQUB_Task_Extract_File_DAT.WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean;
begin
  Result := False;
end;

function TQUB_Task_Extract_File_DAT.WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean;
var
  BufferInt          :TPMInteger;
  BufferFloat        :TPMSingle;
  BufferMPX          :Pointer;
  c, i               :Integer;
  Scal               :Double;
begin
  BufferInt          := nil;
  BufferFloat        := nil;
  BufferMPX          := nil;

  if Float4Byte then begin
  //Float

    try
      BufferFloat    := SAlloc2(prChannelCount, SampleCount);
      for c := 0 to prChannelCount - 1 do begin
        for i := 0 to SampleCount - 1 do begin
          BufferFloat[c][i] := Buffer[c][i];
        end;
      end;

      BufferMPX := AllocMem(4{prDataSize} * prChannelCount * SampleCount);
      MPXDataFloat(BufferFloat, BufferMPX, prChannelCount, -1{all ch}, 4{prDataSize}, NODE_DATATYPE_SINGLE{prDataType}, 0, SampleCount-1, True{MPX});
      BlockWrite(Fi, BufferMPX^, 4{prDataSize} * prChannelCount * SampleCount);
      Result           := False;
    except
      Result           := True;
    end;

  end else begin
  //Integer

  try
    BufferInt        := IAlloc2(prChannelCount, SampleCount);
    for c := 0 to prChannelCount - 1 do begin
      Scal           := Scaling * prChannelScaling[c];
      for i := 0 to SampleCount - 1 do begin
        BufferInt[c][i] := Round(Scal * Buffer[c][i]);
      end;
    end;

    BufferMPX := AllocMem(prDataSize * prChannelCount * SampleCount);
    MPXData(BufferInt, BufferMPX, prChannelCount, -1{all ch}, prDataSize, prDataType, 0, SampleCount-1, True{MPX});
    BlockWrite(Fi, BufferMPX^, prDataSize * prChannelCount * SampleCount);
    Result           := False;
  except
    Result           := True;
  end;

  end;//Float or Int?

  Free2(TPMChar(BufferFloat));
  Free2(TPMChar(BufferInt));
  FreeMem(BufferMPX);
end;

constructor TQUB_Task_Extract_File_ASC.Create(ChannelCount :Integer; Sampling :Double; TimeColumn :Boolean; Data :TQUB_QVO_QVD; ChannelList :TList);
var
  Digits             :Integer;
  Place              :Double;
begin
  inherited Create(ChannelCount, Sampling, Data, ChannelList);
  prTimeColumn       := TimeColumn;
  prFirstSeg         := True;
  Digits             := 1;
  Place              := 0.1;
  while (Digits < 9) and (Abs(Place * Round(Sampling/Place) - Sampling) > 1.0e-6) do begin
    Digits           := Digits + 1;
    Place            := Place * 0.1;
  end;
  prTimeFormat       := Format('%%.%df', [Digits]);
end;

function TQUB_Task_Extract_File_ASC.WriteFileHeader: Boolean;
begin
  Result := False;
end;

function TQUB_Task_Extract_File_ASC.WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean;
var
  s                  :String;
begin
  prStartTime        := StartTime;
  prSamplesWritten   := 0;

  Result             := False;
  if prFirstSeg then begin
    prFirstSeg       := False;
  end else begin
    s                := #13#10;
    try
      BlockWrite(Fi, PChar(s)^, Length(s));
    except
      Result         := True;
    end;
  end;
end;

function TQUB_Task_Extract_File_ASC.WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean;
const
  BUF_INCR           = ONE_MEGABYTE;
var
  AsciiBuf           :TPAByte;
  AsciiBufLen        :Integer;
  AsciiBufOffset     :Integer;
  s                  :String;
  c, i, len          :Integer;
begin
  AsciiBuf           := nil;
  AsciiBufLen        := 0;
  AsciiBufOffset     := 0;
  Result             := False;

  try
    for i := 0 to SampleCount - 1 do begin
      if prTimeColumn then begin
        //s              := Format(prTimeFormat+#9, [prStartTime + prSampling * (prSamplesWritten + i)]);
        //TO DO: Let's try the g format - it will use as many digits as necessary. LM
        s              := Format('%g'#9, [prStartTime + prSampling * (prSamplesWritten + i)]);
        {
          Any reason for as many digits? File size shouldn't grow unnecessarily
          I'm thinking to let the user specify precision
          LM.

          Sample output from %.8g formatting:
39360.141	-0.896875	0	0.908688
39360.141	0.09375	-0.0180636	0.975177
39360.141	-0.35	0.0361272	0.99734
39360.141	-0.021875	0	1.2633
39360.141	-0.26875	0.0361272	1.2633
39360.142	-0.1375	0	1.57358
39360.142	0.434375	-0.0180636	1.4406
39360.142	0.21875	-0.0541908	1.17465
39360.142	-0.20625	0.0361272	1.19681
39360.142	0.765625	0.0541908	1.13032

          Also, I think the user will like to specify precision as much
          as they already love Scaling.
        }
      end else begin
        s              := '';
      end;

      for c := 0 to prChannelCount - 1 do begin
        if c > 0 then s := s + #9;
        s              := s + Format('%.9g', [Buffer[c][i]]);
        {Idem Idem}
        //s              := s + Format('%.6g', [Buffer[c][i]]);
      end;

      s                := s + #13#10;
      len              := Length(s);

      if (AsciiBufOffset + len) > AsciiBufLen then begin
        AsciiBufLen    := AsciiBufLen + BUF_INCR;
        ReallocMem(AsciiBuf, AsciiBufLen);
      end;
      Move(PChar(s)^, AsciiBuf[AsciiBufOffset], len);
      AsciiBufOffset   := AsciiBufOffset + len;
    end;
    BlockWrite(Fi, AsciiBuf^, AsciiBufOffset);
  except
    Result            := True;
  end;
  FreeMem(AsciiBuf);
end;

function TQUB_Task_Extract_File_ASC.Close: Boolean;
var
  Terminator      :String;
begin
  if prFileOpen then begin
    Terminator      := #13#10#13#10;
    try
      BlockWrite(Fi, PChar(Terminator)^, Length(Terminator));
    except
    end;
  end;

  Result := inherited Close;
end;


function TQUB_Task_Extract_File_ASC_SegCol.WriteFileHeader: Boolean;
begin
  prTmpName := GetBaseStampedFileName(ExtractFilePath(Application.ExeName)+'Temp\extract', '.qtr');
  prTmpData := TNode.Create('');
  prTmpData.SaveAs(prTmpName);
  prTmpSeg  := TNode.CreateNull;
  Result := False;
end;

function TQUB_Task_Extract_File_ASC_SegCol.WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean;
begin
  prTmpSeg  := prTmpData.InsertChild(prTmpSeg, 'Segment');
  prTmpSeg.SetupNumData(NODE_DATATYPE_DOUBLE, SampleCount, prChannelCount);
  prSamplesWritten   := 0;
  Result    := False;
end;

function TQUB_Task_Extract_File_ASC_SegCol.WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean;
var
  c, i              :Integer;
begin
  for c := 0 to prChannelCount - 1 do begin
    for i := 0 to SampleCount - 1 do begin
      prTmpSeg.MatrixData[prSamplesWritten+i,c] := Buffer[c][i];
    end;
  end;
  prSamplesWritten  := prSamplesWritten + SampleCount;
  if LongWord(prSamplesWritten) = prTmpSeg.Rows then prTmpSeg.UnloadRows(True);
  Result            := False;
end;

function TQUB_Task_Extract_File_ASC_SegCol.Close: Boolean;
var
  Sel               :TQUB_FirstLast;
  Seg               :INode;
  Segs              :array of INode;
  Sizes             :array of Integer;
  i, j, c           :Integer;
  RowsPerBlock      :Integer;
  Row0, RowMax      :Integer;
  AsciiBuf          :TPAByte;
  AsciiBufOffset    :Integer;
  OptionalTab       :String;
const
  POINTS_PER_BLOCK  :Integer    = ONE_MEGABYTE;
  ASCIIBUF_LEN      :Integer    = ONE_MEGABYTE;
  ASCIIBUF_GUARD    :Integer    = 8192;

  procedure Flush;
  begin
    if AsciiBufOffset > 0 then begin
      BlockWrite(Fi, AsciiBuf^, AsciiBufOffset);
      AsciiBufOffset  := 0;
    end;
  end;
  procedure Write(s :String);
  var
    len     :Integer;
  begin
    len     := Length(s);
    Move(PChar(s)^, AsciiBuf[AsciiBufOffset], len);
    AsciiBufOffset     := AsciiBufOffset + len;
    if AsciiBufOffset > (ASCIIBUF_LEN - ASCIIBUF_GUARD) then Flush;
  end;
begin
  AsciiBuf           := AllocMem(ASCIIBUF_LEN);
  AsciiBufOffset     := 0;
  try
    Seg               := prTmpData.Find('Segment');
    i                 := 0;
    RowMax            := 0;
    while not Seg.IsNull do begin
      SetLength(Segs, i+1);
      Segs[i]         := Seg;
      SetLength(Sizes, i+1);
      Sizes[i]        := Seg.Rows;
      RowMax          := Max(RowMax, Seg.Rows);
      Seg             := Seg.SiblingSameName;
      i               := i + 1;
    end;
    RowsPerBlock      := POINTS_PER_BLOCK div (Length(Segs) * prChannelCount);
    Row0              := 0;
    while Row0 < RowMax do begin
      for i := 0 to Length(Segs) - 1 do begin
        Sel           := QUB_FirstLast(Row0, Row0+RowsPerBlock-1);
        Segs[i].LoadRows(Sel, True);
      end;
      for j := Row0 to Min(Row0+RowsPerBlock, RowMax) - 1 do begin
        OptionalTab   := '';
        if prTimeColumn then Write(Format(prTimeFormat+#9, [j*prSampling]));
        for i := 0 to Length(Segs) - 1 do begin
          for c := 0 to prChannelCount - 1 do begin
            if j < Sizes[i] then begin
              Write(Format('%s%.9g', [OptionalTab, Double(Segs[i].MatrixData[j,c])]));
            end else begin
              Write(OptionalTab);
            end;
            OptionalTab := #9;
          end;
        end;
        Write(#13#10);
      end;
      Row0            := Row0 + RowsPerBlock;
    end;

  except
  end;

  Flush;
  FreeMem(AsciiBuf);

  try
    // release all refs into tmp tree
    SetLength(Segs, 0);
    Seg       := TNode.CreateNull;
    prTmpSeg  := Seg;
    prTmpData := Seg;
    // and now that it's closed and gone, delete it
    DeleteFile(prTmpName);
  except
  end;

  Result := inherited Close;
end;


function TQUB_Task_Extract_File_WAV.WriteFileHeader: Boolean;
begin
  Result                := False;
end;

function TQUB_Task_Extract_File_WAV.WriteSegmentHeader(StartTime :Double; SampleCount :Integer): Boolean;
var
  WaveChunk             :TQUB_QFS_QDF_WAV_Chunk;
begin
  Result                := False;
  try
    WaveChunk.riff.riff              := QUB_QFS_QDF_STRING_RIFF;//(LongInt('R')) or (LongInt('I') shl 8) or (LongInt('F') shl 16) or (LongInt('F') shl 24);
    // length: why *2?
    WaveChunk.riff.length            := SizeOf(WaveChunk) - SizeOf(WaveChunk.riff.riff) - SizeOf(WaveChunk.riff.length) + SampleCount * 2;
    WaveChunk.riff.wave              := QUB_QFS_QDF_STRING_WAVE;//(LongInt('W')) or (LongInt('A') shl 8) or (LongInt('V') shl 16) or (LongInt('E') shl 24);
    WaveChunk.format.fmt             := QUB_QFS_QDF_STRING_FMT;//(LongInt('f')) or (LongInt('m') shl 8) or (LongInt('t') shl 16) or (LongInt(' ') shl 24);
    WaveChunk.format.length          := $10;
    WaveChunk.format.formattag       := $0001;
    WaveChunk.format.channels        := prChannelCount;
    WaveChunk.format.samplerate      := Round(1 / prSampling);
    WaveChunk.format.bytespersample  := prDataSize * WaveChunk.format.channels;
    WaveChunk.format.bytespersec     := WaveChunk.format.bytespersample * WaveChunk.format.samplerate;
    WaveChunk.format.bitspersample   := 8 * prDataSize;
    WaveChunk.data.data              := QUB_QFS_QDF_STRING_DATA;//(LongInt('d')) or (LongInt('a') shl 8) or (LongInt('t') shl 16) or (LongInt('a') shl 24);
    WaveChunk.data.length            := SampleCount            * WaveChunk.format.bytespersample;

    BlockWrite(Fi, WaveChunk, SizeOf(WaveChunk));
  except
    Result              := True;
  end;
end;

function TQUB_Task_Extract_File_WAV.WriteBuffer(Buffer :TPMDouble; SampleCount :Integer): Boolean;
var
  BufferInt          :TPMInteger;
  BufferMPX          :Pointer;
  c, i               :Integer;
  Scal               :Double;
begin
  BufferInt          := nil;
  BufferMPX          := nil;
  try
    BufferInt        := IAlloc2(prChannelCount, SampleCount);
    for c := 0 to prChannelCount - 1 do begin
      Scal           := Scaling * prChannelScaling[c];
      for i := 0 to SampleCount - 1 do begin
        BufferInt[c][i] := Round(Scal * Buffer[c][i]);
      end;
    end;

    BufferMPX := AllocMem(prDataSize * prChannelCount * SampleCount);
    MPXData(BufferInt, BufferMPX, prChannelCount, -1{all ch}, prDataSize, prDataType, 0, SampleCount-1, True{MPX});
    BlockWrite(Fi, BufferMPX^, prDataSize * prChannelCount * SampleCount);
    Result           := False;
  except
    Result           := True;
  end;

  Free2(TPMChar(BufferInt));
  FreeMem(BufferMPX);
end;


//******************************************************************************
constructor TQUB_Task_ExtractSelection.Create(Suspended :Boolean);
begin
  inherited;

  Priority      := tpNormal;
  prName        := 'Extract selection';
  prFileFormat  := DEF_FILEFORMAT;//QFS_QDF_TYPE_LDT;

  prChunks      := nil;
  prChannelList := TList.Create;
  prKinParsList := TStringList.Create;

end;

destructor TQUB_Task_ExtractSelection.Destroy;
begin

  try

    if prChunks <> nil then prFreeChunks;
    prChannelList.Free;
    prKinParsList.Free;

  finally

    inherited;

  end;

end;

procedure TQUB_Task_ExtractSelection.Execute;
begin

  try

    if Terminated then Exit;


    prStatus := 'Initializing';
    prPercent := 0;
    prETA := '0:00:00';

    Synchronize(prDoOnChange);

    try


      if FileFormat = QFS_QDF_TYPE_DWT then begin
        prDoExtractIdealized;
      end else begin
        prDoExtractSelection;
      end;


    except
    end;

    Beep;

    //Terminate;
        
    prStatus := 'Terminated';
    prPercent := 0;
    prETA := '0:00:00';

    Synchronize(prDoOnChange);

  finally

    if Data <> nil then Data.UnLockWrite;
    if prSelectionList    <> nil then prSelectionList.UnLock;
    if prSubSelectionList <> nil then prSubSelectionList.UnLock;

  end;

end;


type TChunker = function(QVD:TQUB_QVO_QVD): Boolean of object;

procedure TQUB_Task_ExtractSelection.prDoExtractSelection;
var
  Sources             :TList; // of TQUB_Task_Extract_DataSet
  Source              :TQUB_Task_Extract_DataSet;
  AddSource           :TQUB_Task_Extract_DataSet;
  SubSource           :TQUB_Task_Extract_DataSet;
  Fi                  :TQUB_Task_Extract_File;
  fn                  :String;
  Ext                 :String;
  Buf                 :TQUB_DBuf2_Double;
  BufCount            :Integer;
  ActualBufCount      :Integer;
  i, s, At, N         :Integer;
  OperChannel         :Integer;
  StTime              :Comp;
  NewPercent          :Integer;
  MaxSeg              :Integer;
  TotalCount          :Integer;
  ActualCount         :Integer;
  Sess                :INode;
  Chan                :INode;
  ActualDataSize      :Integer;
  ActualScaling       :Double;

  // build variations e.g. fit-source by modifying Self.Channels etc.
  function BuildInitialSource(ChunkData :TChunker): TQUB_Task_Extract_DataSet;
  var
    i, last           :Integer;
    QVD               :TQUB_QVO_QVD;
  begin
    Result            := nil;

    if ExtractFileList then begin

      last  := 0;
      for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
        if TQUB_QVO_QVD(ViewportData.ListObjects[i]).Listed then last := i;
      end;

      Result := TQUB_Task_Extract_DataSet_FileList.Create(Self, Data);
      for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
        QVD := TQUB_QVO_QVD(ViewportData.ListObjects[i]);
        if QVD.Listed then begin
          ChunkData(QVD);

          if JoinSegments then begin
            TQUB_Task_Extract_DataSet_FileList(Result).AddSource(
                         TQUB_Task_Extract_DataSet_Chunks_Joined.Create(Self, QVD, prChunks),
                         QVD, i = last );
          end else begin
            TQUB_Task_Extract_DataSet_FileList(Result).AddSource(
                         TQUB_Task_Extract_DataSet_Chunks.Create(Self, QVD, prChunks),
                         QVD, i = last );
          end;

          prChunks := nil;
        end;
      end;

    end else begin

      if prChunkData(Data) then Exit;
      if prChunkSubData(Data) then Exit;

      if JoinSegments then begin
        Result := TQUB_Task_Extract_DataSet_Chunks_Joined.Create(Self, Data, prChunks);
      end else begin
        Result := TQUB_Task_Extract_DataSet_Chunks.Create(Self, Data, prChunks);
      end;

      prChunks := nil;

    end;
  end;

begin

  Sources             := TList.Create;
  Fi                  := nil;
  Buf                 := nil;

  try

    //Start here
    StTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
    prStatus := 'Running ';

    Synchronize(prDoOnChange);

    // setup processing chain

    if (FileFormat = QFS_QDF_TYPE_WAV) or (FileFormat = QFS_QDF_TYPE_DAT) then begin
      if (Operation = OPER_NONE) {or (OperationScale)} then begin
        JoinSegments  := True;
      end;
    end;

    if ((Operation = OPER_AVG) or (Operation = OPER_STD)) and (OperationList) then begin
    //It would be kind of pointless, otherwise
      JoinSegments := False;
    end;

    //if {(FileFormat <> QFS_QDF_TYPE_ASC) or} prData.FitCurves.IsEmpty then begin
    //  ASCIIFit := False; // TODO: this could be annoying with file list
    //end;

    //This if the origin of the Source Chain. Includes ASCIIFit and KinPars
    Source        := BuildInitialSource(prChunkData);
    Sources.Add(Source);

    if (Source.ChannelCount = 0) or (Source.SegmentCount = 0) then begin
      prError := True;
      prErrorMsg  := 'No data.  Check the Data Source.';
      prWikiTopic := 'Error:No_Data';
      Exit;
    end;

    if ProcOnlyActiveChannel then begin
      OperChannel := Source.ActiveChannel;
    end else begin
      OperChannel := -1;
    end;

    if OperationScale then begin
      Source := TQUB_Task_Extract_DataSet_Scale.Create(Source, OperationScaleFactor, OperChannel);
      Sources.Add(Source);
    end;
    //This can be optimized
    if Operation = OPER_ADD then begin
      if OperationList then begin
      //Add a list of selections
        //TO DO: LM. See Chris' comments below
        // optional: modify self.channels etc here
        AddSource   := BuildInitialSource(prChunkData); // prChunkAddData?
        Sources.Add(AddSource); // garbage collection list
        Source := TQUB_Task_Extract_DataSet_AddSubList.Create(Source, AddSource, OperChannel, 1);
      end else begin
        Source := TQUB_Task_Extract_DataSet_Add.Create(Source, OperChannel);
      end;
      Sources.Add(Source);
    end;
    if Operation = OPER_SUB then begin
      if OperationList then begin
      //Subtract a list of selections (e.g., subtract the TTX-blocked data from raw data)
        //TO DO: LM. Check this
        SubSource   := BuildInitialSource(prChunkSubData);
        Sources.Add(SubSource);
        Source := TQUB_Task_Extract_DataSet_AddSubList.Create(Source, SubSource, OperChannel, -1);
      end else begin
        Source := TQUB_Task_Extract_DataSet_Sub.Create(Source, OperChannel);
      end;
      Sources.Add(Source);
    end;
    if Operation = OPER_AVG then begin
      if OperationList then begin
        if OperationFiles then begin
          Source := TQUB_Task_Extract_DataSet_AvgListsFiles.Create(Self{Task}, prData, Source, OperChannel, JoinSegments);
        end else begin
          Source := TQUB_Task_Extract_DataSet_AvgLists.Create(Self{Task}, prData, Source, prData.SelectionLists, OperChannel, JoinSegments);
        end;
      end else begin
        Source := TQUB_Task_Extract_DataSet_Avg.Create(Source, OperChannel);
      end;
      Sources.Add(Source);
    end;
    if Operation = OPER_STD then begin
      if OperationList then begin
        if OperationFiles then begin
          //Source := TQUB_Task_Extract_DataSet_StdListsFiles.Create(Self{Task}, prData, Source, OperChannel, JoinSegments);
        end else begin
          //Source := TQUB_Task_Extract_DataSet_StdLists.Create(Self{Task}, prData, Source, prData.SelectionLists, OperChannel, JoinSegments);
        end;
      end else begin
        Source := TQUB_Task_Extract_DataSet_Std.Create(Source, OperChannel);
      end;
      Sources.Add(Source);
    end;
    if FunctionDrift then begin
      Source := TQUB_Task_Extract_DataSet_Drift.Create(Source, FunctionDriftA, FunctionDriftB, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionSine then begin
      Source := TQUB_Task_Extract_DataSet_Sine.Create(Source, FunctionSineFrq, FunctionSineAmp, FunctionSinePh, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionDigitize then begin
      Source := TQUB_Task_Extract_DataSet_Digitize.Create(Source, FunctionDigitizeThreshold, FunctionDigitizeToMin, FunctionDigitizeToMax, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionDerivate then begin
      Source := TQUB_Task_Extract_DataSet_Derivate.Create(Source, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionLKC then begin
      Self.Channels := CHANNELS_List;
      Self.ChannelList.Clear;
      Self.ChannelList.Add(Pointer(Integer(LKCVChannel)));
      Self.ASCIIFit := False;
      // if we keep building sub-sources some of these above might have to be restored...
      SubSource := BuildInitialSource(prChunkLKCData);
      Sources.Add(SubSource);
      Source := TQUB_Task_Extract_DataSet_LKC.Create(Source, SubSource, prData, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionItoG then begin
      Source := TQUB_Task_Extract_DataSet_ItoG.Create(Source, prData, FunctionItoGVRev, FunctionItoGVCh, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionRsCorr then begin
      Source := TQUB_Task_Extract_DataSet_RsCorr.Create(Source, prData, FunctionRsValue, FunctionRsPer, FunctionItoGVCh, FunctionRsAddLeak, FunctionRsLeakCh, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionPnAdd then begin
      Source := TQUB_Task_Extract_DataSet_PnAdd.Create(Source, prData,  FunctionPnOffset, FunctionPnDelay, FunctionPnScale, FunctionPnCount, FunctionPnVm, FunctionPnInterval, OperChannel);
      Sources.Add(Source);
    end else if FunctionPnSub then begin
      Source := TQUB_Task_Extract_DataSet_PnSub.Create(Source, prData, FunctionPnOffset, FunctionPnDelay, FunctionPnScale, FunctionPnCount, FunctionPnVm, FunctionPnInterval, OperChannel);
      Sources.Add(Source);
    end;
    if FunctionFit then begin
      Self.Channels := CHANNELS_ACTIVE;
      Self.ASCIIFit := True;
      SubSource := BuildInitialSource(prChunkData);
      Sources.Add(SubSource);
      Source := TQUB_Task_Extract_DataSet_Fit.Create(Source, SubSource, prData, OperChannel{ignored/uses ActiveChannel});
      Sources.Add(Source);
    end;

    //The last is always Resample - Shouldn't Filter be done right before this?
    if (DataProcessed <> DATA_PROCESSED_NONE) and ProcessResample then begin
      Source := TQUB_Task_Extract_DataSet_Resample.Create(Source, ProcessResampleFactor);
      Sources.Add(Source);
    end;

    // setup buffer
    try

      MaxSeg      := 0;
      ActualCount := 0;
      TotalCount  := 0;
      for s := 0 to Source.SegmentCount - 1 do begin
        MaxSeg    := Max(MaxSeg, Source.SegmentLength[s]);
        TotalCount := TotalCount + Source.SegmentLength[s];
      end;
      if (DataProcessed <> DATA_PROCESSED_NONE) and ProcessResample then begin
        MaxSeg    := Round(MaxSeg * ProcessResampleFactor);
      end;
      BufCount    := Min( ({4*}ONE_MEGABYTE) div Source.ChannelCount, MaxSeg );
      Buf         := TQUB_DBuf2_Double.Create(Source.ChannelCount, BufCount);

    except

      prError := True;
      prErrorMsg := 'Cannot allocate memory.'#13#10'Make sure you don''t have too many files opened';
      prWikiTopic := 'Error:Cannot_Allocate_Memory';
      Exit;

    end;

    // setup file

    case FileFormat of

      QFS_QDF_TYPE_QDF:
        begin
          Fi  := TQUB_Task_Extract_File_QDF.Create(Source.ChannelCount, Source.Sampling, prData.DataObject.ADDataSize, prData.DataObject.ADDataType, Source.Session['Scaling'].Data[0], Source.Session['DataChannels'], prData, Self.ChannelList);
          Ext := '.qdf';
          ActualDataSize := TQUB_Task_Extract_File_QDF(Fi).DataSize;
          ActualScaling  := TQUB_Task_Extract_File_QDF(Fi).Scaling;
        end;

      QFS_QDF_TYPE_LDT:
        begin
          Fi  := TQUB_Task_Extract_File_LDT.Create(Source.ChannelCount, Source.Sampling, Source.Session['Scaling'].Data[0], Source.Session['DataChannels'], prData, Self.ChannelList);
          Ext := '.ldt';
          ActualDataSize := 2;
          ActualScaling  := TQUB_Task_Extract_File_LDT(Fi).Scaling;
        end;

      QFS_QDF_TYPE_DAT:
        begin
          Fi  := TQUB_Task_Extract_File_DAT.Create(Source.ChannelCount, Source.Sampling, prData.DataObject.ADDataSize, prData.DataObject.ADDataType, Source.Session['Scaling'].Data[0], Source.Session['DataChannels'], prData, Self.ChannelList);
          TQUB_Task_Extract_File_DAT(Fi).Float4Byte := Float4Byte;
          Ext := '.dat';
          ActualDataSize := TQUB_Task_Extract_File_DAT(Fi).DataSize;
          ActualScaling  := TQUB_Task_Extract_File_DAT(Fi).Scaling;
        end;

      QFS_QDF_TYPE_ASC:
        begin
          if ASCIISegCol then begin
            Fi  := TQUB_Task_Extract_File_ASC_SegCol.Create(Source.ChannelCount, Source.Sampling, ASCIITime, prData, Self.ChannelList);
          end else begin
            Fi  := TQUB_Task_Extract_File_ASC.Create(Source.ChannelCount, Source.Sampling, ASCIITime, prData, Self.ChannelList);
          end;
          Ext := '.txt';
          ActualDataSize := prData.DataObject.ADDataSize;
          ActualScaling  := prData.Scaling;
          Source.Session['TimeColumn'].SetData(Integer(ASCIITime));
        end;

      QFS_QDF_TYPE_WAV:
        begin
          if Source.ChannelCount > 2 then begin
            prError := True;
            prErrorMsg := 'Cannot save more than two data channels in WAV format';
            prWikiTopic := 'Error:Data_File_Format';
            Exit;
          end else begin
            Fi  := TQUB_Task_Extract_File_WAV.Create(Source.ChannelCount, Source.Sampling, prData.DataObject.ADDataSize, prData.DataObject.ADDataType, Source.Session['Scaling'].Data[0], Source.Session['DataChannels'], prData, Self.ChannelList);
            Ext := '.wav';
            ActualDataSize := TQUB_Task_Extract_File_WAV(Fi).DataSize;
            ActualScaling  := TQUB_Task_Extract_File_WAV(Fi).Scaling;
          end;
        end;

    else//case
      begin
        prError := True;
        prErrorMsg := 'File format not supported yet';
        prWikiTopic := 'Error:Data_File_Format';
        Exit;
      end;
    end;//case FileFormat

    // write file

    fn := ChangeFileExt(FileName, Ext);
    prFileName := fn;
    if Fi.Open(fn) then begin
      prError := True;
      prErrorMsg := 'Cannot create file.';
      prWikiTopic := 'Error:Cannot_Create_New_File';
      Exit;
    end;

    if Fi.WriteFileHeader then begin
      prError := True;
      prErrorMsg := 'Cannot write file.';
      prWikiTopic := 'Error:Cannot_Create_New_File';
      Exit;
    end;

    for s := 0 to Source.SegmentCount - 1 do begin
      N  := Source.SegmentLength[s];
      if Fi.WriteSegmentHeader(Source.SegmentStart[s], N) then begin
        prError := True;
        prErrorMsg := 'Cannot write file.';
        prWikiTopic := 'Error:Cannot_Create_New_File';
        Exit;
      end;

      At := 0;
      while At < N do begin
        NewPercent  := Round(ActualCount / TotalCount * 100);
        if NewPercent <> prPercent then begin
          prUpdateStatus(NewPercent, StTime);
        end;
        if prStopped then Break;

        ActualBufCount := Min(N - At, BufCount);
        if Source.Load(Buf, s, At, At + ActualBufCount - 1) then begin
          prError := True;
          prErrorMsg := 'Cannot read data for extraction.';
          prWikiTopic := 'Error:Cannot_read_data_for_extraction';
          Exit;
        end;
        if Fi.WriteBuffer(Buf.Buf, ActualBufCount) then begin
          prError := True;
          prErrorMsg := 'Cannot write data.';
          prWikiTopic := 'Error:Cannot_Create_New_File';
          Exit;
        end;

        At := At + ActualBufCount;
        ActualCount := ActualCount + ActualBufCount;
      end;
    end;

    Fi.Close;

    // write session file
    Sess := Source.Session;

    Sess['FileExt'].SetData( Ext );
    Sess['ADDataSize'].SetData( ActualDataSize );
    Sess['Scaling'].SetData( ActualScaling );

    if FileFormat = QFS_QDF_TYPE_ASC then begin
      Chan := Sess['DataChannels'].Find('Channel');
      while not Chan.IsNull do begin
        Chan['Scaling'].SetData( 1.0/1.0 ); // file is written in Units not Volts
        Chan := Chan.SiblingSameName;
      end;
    end;

    Sess.SaveAs( ChangeFileExt(FileName, '.qsf') );

  finally

    for i := Sources.Count - 1 downto 0 do begin
      TQUB_Task_Extract_DataSet(Sources[i]).Free;
    end;
    Sources.Free;

    if Fi <> nil then Fi.Free;
    if Buf <> nil then Buf.Free;

    if prStopped or prError then begin
    //Delete file
      try
        DeleteFile(PChar(fn));
      except
      end;
    end;

    if prError then begin

      Synchronize(ShowMessage);

    end;

  end;

end;


procedure TQUB_Task_ExtractSelection.prDoExtractIdealized;
var
  FileOpened            :Boolean;

  NewPercent            :Integer;
  TotalCount            :Integer;
  ActualCount           :Integer;
  StTime                :Comp;

  f                     :TextFile;
  fn                    :String;

  i, s                  :Integer;

  DSs, DS, DSj, Seg     :INode;
  DwellCount            :Integer;
  Classes               :TPAInteger;
  Durations             :TPASingle;
  Amp, Std              :INode;
  Last                  :INode;

  DecSeparator          :Char;

begin
  DecSeparator          := DecimalSeparator; // to restore global setting at end
  DecimalSeparator      := '.'; // no decimal comma allowed

  DSs                   := TNode.Create('DataSets');
  if ExtractFileList then begin
    Last                  := TNode.CreateNull;
    for i := 0 to ViewportData.ListObjects.Items.Count - 1 do begin
      if TQUB_QVO_QVD(ViewportData.ListObjects[i]).Listed then begin
        DS := TQUB_QVO_QVD(ViewportData.ListObjects[i]).GetChosenIdl(DropFirstLast, DropIfClass, DropClass);
        if DS.Find('Segment').Find('Classes').DataCount > 0 then begin
          Last := DSS.InsertChild(Last, DS);
        end;
      end;
    end;
  end else begin
    DSs.AppendChild( Data.GetChosenIdl(DropFirstLast, DropIfClass, DropClass) );
  end;

  if DSs.Find('DataSet').Find('Segment').IsNull then begin
    prError := True;
    prErrorMsg := 'No idealized data.  Check the Data Source.';
    prWikiTopic := 'Error:No_Data';
    Exit;
  end;

  FileOpened            := False;

  try

    try
      prFileName := ChangeFileExt(FileName, '.dwt');
      fn := prFileName;
      AssignFile(f, fn);
      ReWrite(f);
      FileOpened := True;
    except
      prError := True;
      prErrorMsg := 'Cannot open output file.'#13#10'Medium may be read-only or a file with the same name is in use';
      prWikiTopic := 'Error:Cannot_Create_New_File';
      Exit;
    end;

    StTime := TimeStampToMSecs(DateTimeToTimeStamp(Now));
    prStatus := 'Running ';
    Synchronize(prDoOnChange);

    TotalCount  := 0;
    ActualCount := 0;

    DS := DSs.Find('DataSet');
    while not DS.IsNull do begin
      if JoinSegments then begin
        DSj := prJoinIdlSegments(DS);
        DSs.InsertChild(DS, DSj);
        DSs.RemoveChild(DS);
        DS := DSj;
      end;
      
      Seg         := DS.Find('Segment');
      while not Seg.IsNull do begin
        DwellCount := Seg['DwellCount'].Data[0];
        Inc(TotalCount, DwellCount);

        Seg        := Seg.SiblingSameName;
      end;
      DS           := DS.SiblingSameName;
    end;

    s          := 1;
    DS         := DSs.Find('DataSet');
    while not DS.IsNull do begin
      Seg        := DS.Find('Segment');
      while not Seg.IsNull do begin
        DwellCount   := Seg['DwellCount'].Data[0];
        Classes      := TPAInteger( Seg['Classes'].DataPtr );
        Durations    := TPASingle( Seg['Durations'].DataPtr );
        Amp          := Seg['amp'];
        Std          := Seg['sd'];

        // write header
        Write(f, Format('Segment: %d Dwells: %d Sampling(ms): %.8g Start(ms): %.8g ClassCount: %d',
                        [s, DwellCount, Data.Sampling * 1.0e3,
                         Double(Seg['start'].Data[0]), Amp.DataCount]) );

        for i := 0 to Amp.DataCount - 1 do begin
          Write(f, Format(' %.6g %.6g', [Double(Amp.Data[i]), Double(Std.Data[i])]));
        end;
        Write(f, #13#10);

        for i := 0 to DwellCount - 1 do begin
          Write(f, Format('%d'#9'%.6g'#13#10, [Classes[i], Durations[i]]));
        end;

        Inc(ActualCount, DwellCount);
        NewPercent  := Round(100 * ActualCount / TotalCount);
        if NewPercent <> prPercent then begin
          prUpdateStatus(NewPercent, StTime);
        end;

        Inc(s);
        Seg      := Seg.SiblingSameName;
      end;
      DS         := DS.SiblingSameName;
    end;

  finally

    if FileOpened then begin
      try
        CloseFile(f);
      except
      end;
    end;

    if prStopped then begin
    //Delete file
      try
        DeleteFile(PChar(fn));
      except
      end;
    end;

    if prError then begin

      Synchronize(ShowMessage);

    end;

    DecimalSeparator := DecSeparator;

  end;

end;

function TQUB_Task_ExtractSelection.prJoinIdlSegments(DS :INode): INode;
var
  Seg             :INode;
  JoinIdl         :TQUB_IdlSegments;
  First, Last     :Integer;
begin
  Seg             := DS.Find('Segment');
  if Seg.IsNull then begin
    Result        := DS.Clone(True);
    Exit;
  end;
  JoinIdl         := TQUB_IdlSegments.Create(DS['sampling'].Data[0]*1.0e-3);
  First           := Seg.Data[0];
  Last            := First; // shut up compiler
  while not Seg.IsNull do begin
    Last          := Seg.Data[1];
    Seg           := Seg.SiblingSameName;
  end;
  JoinIdl.AddSeg(0, Last);
  JoinIdl.SetDwell(First, Last, 0);
  JoinIdl.SetDwells(DS);
  Result          := JoinIdl.GetDwells(First, Last, 1, 1);
  JoinIdl.Free;

  Seg := Result.Find('Segment');
  while not Seg.IsNull do begin
    Seg['start'].SetData( Data.WhichTimeForPoint( Seg.Data[0] ) * 1.0e3 );
    Seg := Seg.SiblingSameName;
  end;
end;

function TQUB_Task_ExtractSelection.prChunkData(QVD :TQUB_QVO_QVD) :Boolean;
var
  ISels            :TList;
  ODats            :TList;
  OIdlx            :TList;
  OIdls            :TList;
  Sel              :TQUB_PFirstLast;
  ListIter         :TQUB_QVO_QVD_SelectionListIter;

  First            :Integer;

  l, s             :Integer;

  Chunk            :TQUB_Thread_PChunk;

begin

  if FileFormat = QFS_QDF_TYPE_DWT then begin
    Result := False;
    Exit;
  end;

  ISels := nil;
  ODats := nil;
  OIdlx := nil;
  OIdls := nil;

  if prChunks <> nil then prFreeChunks;
  prChunks := TList.Create;

  try

    ISels := TList.Create;
    ODats := TList.Create;
    OIdlx := TList.Create;
    OIdls := TList.Create;

    if (DataSource = DATA_SOURCE_WHOLEFILE) or ((DataSource = DATA_SOURCE_SELECTIONLIST) and (prSelectionList = nil)) then begin

      New(Sel);
      ISels.Add(Sel);
      Sel^ := QVD.DataPoint;

    end else if DataSource = DATA_SOURCE_SELECTIONLIST then begin

      ListIter := prSelectionList.Find(0, QVD.DataPoint.Last, True{ActiveOnly});
      if not ListIter.Done then begin
        repeat
          New(Sel);
          ISels.Add(Sel);
          Sel^ := ListIter.Sel;
        until ListIter.Next;
      end;
      ListIter.Free;

    end else if DataSource = DATA_SOURCE_SELECTION then begin

      if QVD.DataSel.First < QVD.DataSel.Last then begin
        New(Sel);
        ISels.Add(Sel);
        Sel^ := QVD.DataSel;
      end;

    end;

    if ISels.Count < 1 then begin
      Result     := True;
      prError    := True;
      prErrorMsg := 'No selection active or empty selection list or empty selection';
      prWikiTopic := 'Error:No_Selection';
      Exit;
    end;

    //if FileFormat = QFS_QDF_TYPE_DWT then begin
    //  Result := QVD.SegmentIdlData(ISels, OIdlx, OIdls, nil{Dats}, True{Segmentation}, False{PartIDLSeg});
    //  if not Result then Result := OIdls.Count = 0;
    //end else begin
      Result := QVD.SegmentData(ISels, ODats, True{Segmentation}, 0{MaxSize});
      if not Result then Result := ODats.Count = 0;
    //end;

    if Result then begin
      prError    := True;
      prErrorMsg := 'Data could not be segmented';
      prWikiTopic := 'Error:Data_could_not_be_segmented';
      Exit;
    end;

    for l := 0 to ODats.Count - 1 do begin

      Sel                   := TQUB_PFirstLast(ODats.List[l]);
      Chunk                 := AllocMem(SizeOf(TQUB_Thread_Chunk));

      if Chunk <> nil then begin

        prChunks.Add(Chunk);
        Chunk.Sel             := Sel^;
        Chunk.Status          := CHUNK_STATUS_EMPTY;
        Chunk.Trials          := 0;
        Chunk.Index           := l;

        Chunk.StartTime       := 0;
        for s := 0 to QVD.DataSegmentCount - 1 do begin
          if Sel.First <= (QVD.DataBreakpoints[s] - 1) then begin
            if s = 0 then First := 0 else First := QVD.DataBreakpoints[s - 1];
            Chunk.StartTime := QVD.DataObject.Segments[s].StartTime + QVD.Sampling * (Sel.First - First);
            Break;
          end;
        end;

      end;

    end;

    if prChunks.Count = 0 then begin
      Result     := True;
      prError    := True;
      prErrorMsg := 'Data could not be segmented';
      prWikiTopic := 'Error:Data_could_not_be_segmented';
      Exit;
    end;

  finally

    if ISels <> nil then begin
      for s := 0 to ISels.Count - 1 do begin
        Dispose(TQUB_PFirstLast(ISels.List[s]));
      end;
      ISels.Free;
    end;
    if ODats <> nil then begin
      for s := 0 to ODats.Count - 1 do begin
        Dispose(TQUB_PFirstLast(ODats.List[s]));
      end;
      ODats.Free;
    end;
    if OIdlx <> nil then begin
      for s := 0 to OIdlx.Count - 1 do begin
        Dispose(TQUB_PFirstLast(OIdlx.List[s]));
      end;
      OIdlx.Free;
    end;
    if OIdls <> nil then begin
      for s := 0 to OIdls.Count - 1 do begin
        Dispose(TQUB_PFirstLast(OIdls.List[s]));
      end;
      OIdls.Free;
    end;

  end;

end;

function TQUB_Task_ExtractSelection.prChunkSubData(QVD: TQUB_QVO_QVD) :Boolean;
var
  ISels            :TList;
  ODats            :TList;
  OIdlx            :TList;
  OIdls            :TList;
  Sel              :TQUB_PFirstLast;
  ListIter         :TQUB_QVO_QVD_SelectionListIter;

  First            :Integer;

  l, s             :Integer;

  Chunk            :TQUB_Thread_PChunk;

begin

  if FileFormat = QFS_QDF_TYPE_DWT then begin
    Result := False;
    Exit;
  end;

  if prSubSelectionList = nil then begin
    Result := False;
    Exit;
  end;

  ISels := nil;
  ODats := nil;
  OIdlx := nil;
  OIdls := nil;

  if prChunks <> nil then prFreeChunks;
  prChunks := TList.Create;

  try

    ISels := TList.Create;
    ODats := TList.Create;
    OIdlx := TList.Create;
    OIdls := TList.Create;

    {
    if (DataSource = DATA_SOURCE_WHOLEFILE) or ((DataSource = DATA_SOURCE_SELECTIONLIST) and (prSelectionList = nil)) then begin

      New(Sel);
      ISels.Add(Sel);
      Sel^ := QVD.DataPoint;

    end else if DataSource = DATA_SOURCE_SELECTIONLIST then begin
    }

      ListIter := prSubSelectionList.Find(0, QVD.DataPoint.Last, True{ActiveOnly});
      if not ListIter.Done then begin
        repeat
          New(Sel);
          ISels.Add(Sel);
          Sel^ := ListIter.Sel;
        until ListIter.Next;
      end;
      ListIter.Free;

    {
    end else if DataSource = DATA_SOURCE_SELECTION then begin

      if QVD.DataSel.First < QVD.DataSel.Last then begin
        New(Sel);
        ISels.Add(Sel);
        Sel^ := QVD.DataSel;
      end;

    end;
    }

    if ISels.Count < 1 then begin
      Result     := True;
      prError    := True;
      prErrorMsg := 'No selection active or empty selection list or empty selection';
      prWikiTopic := 'Error:Subtract_Data';
      Exit;
    end;

    //if FileFormat = QFS_QDF_TYPE_DWT then begin
    //  Result := QVD.SegmentIdlData(ISels, OIdlx, OIdls, nil{Dats}, True{Segmentation}, False{PartIDLSeg});
    //  if not Result then Result := OIdls.Count = 0;
    //end else begin
      Result := QVD.SegmentData(ISels, ODats, True{Segmentation}, 0{MaxSize});
      if not Result then Result := ODats.Count = 0;
    //end;

    if Result then begin
      prError    := True;
      prErrorMsg := 'Subtract data could not be segmented';
      prWikiTopic := 'Error:Subtract_Data';
      Exit;
    end;

    for l := 0 to ODats.Count - 1 do begin

      Sel                   := TQUB_PFirstLast(ODats.List[l]);
      Chunk                 := AllocMem(SizeOf(TQUB_Thread_Chunk));

      if Chunk <> nil then begin

        prChunks.Add(Chunk);
        Chunk.Sel             := Sel^;
        Chunk.Status          := CHUNK_STATUS_EMPTY;
        Chunk.Trials          := 0;
        Chunk.Index           := l;

        Chunk.StartTime       := 0;
        for s := 0 to QVD.DataSegmentCount - 1 do begin
          if Sel.First <= (QVD.DataBreakpoints[s] - 1) then begin
            if s = 0 then First := 0 else First := QVD.DataBreakpoints[s - 1];
            Chunk.StartTime := QVD.DataObject.Segments[s].StartTime + QVD.Sampling * (Sel.First - First);
            Break;
          end;
        end;

      end;

    end;

    if prChunks.Count = 0 then begin
      Result     := True;
      prError    := True;
      prErrorMsg := 'Subtract data could not be segmented';
      prWikiTopic := 'Error:Subtract_Data';
      Exit;
    end;

  finally

    if ISels <> nil then begin
      for s := 0 to ISels.Count - 1 do begin
        Dispose(TQUB_PFirstLast(ISels.List[s]));
      end;
      ISels.Free;
    end;
    if ODats <> nil then begin
      for s := 0 to ODats.Count - 1 do begin
        Dispose(TQUB_PFirstLast(ODats.List[s]));
      end;
      ODats.Free;
    end;
    if OIdlx <> nil then begin
      for s := 0 to OIdlx.Count - 1 do begin
        Dispose(TQUB_PFirstLast(OIdlx.List[s]));
      end;
      OIdlx.Free;
    end;
    if OIdls <> nil then begin
      for s := 0 to OIdls.Count - 1 do begin
        Dispose(TQUB_PFirstLast(OIdls.List[s]));
      end;
      OIdls.Free;
    end;

  end;

end;

function TQUB_Task_ExtractSelection.prChunkLKCData(QVD :TQUB_QVO_QVD): Boolean;
var
  i          :Integer;
  N          :Integer;
  Off        :Integer;
  Chunk      :TQUB_Thread_PChunk;
begin
  Result     := False;
  prChunkData(QVD);
  if TaskExpPropertiesForm = nil then Exit;

  with TaskExpPropertiesForm do begin
  //        prLKC_V0   := prData.Average(LKCVChannel{StimCh}, Sel.First - LKCV0Points + LKCSkipPoints, Sel.First - LKCSkipPoints);
    Off        := LKCSkipPoints - LKCV0Points;
    N          := Max(1, - LKCSkipPoints - Off + 1);
  end;

  for i := 0 to prChunks.Count - 1 do begin
    Chunk    := TQUB_Thread_PChunk(prChunks[i]);
    Chunk.Sel.First := Max(0, Chunk.Sel.First + Off);
    Chunk.Sel.Last  := Chunk.Sel.First + N - 1;
  end;
end;

procedure TQUB_Task_ExtractSelection.prFreeChunks;
var
  i :Integer;
begin
  for i := 0 to prChunks.Count - 1 do begin
    FreeMem(prChunks.List[i]);
  end;
  prChunks.Free;
end;


//******************************************************************************
procedure TQUB_Task_ExtractSelection.OnGetProperties(Sender :TObject);
var
  i      :Integer;
  Form   :TQUB_Task_ExtractSelection_PropertiesForm;
  Props  :INode;
  Prop   :INode;

  procedure MakeKinParsList(Text :String);
  var
    Parser :TQUB_Parser;
  begin
    Parser               := TQUB_Parser.Create;
    Parser.ReturnTab     := False;
    Parser.ReturnNewLine := True;
    try
      Parser.Start;
      repeat
        Parser.Next(Text);
        if Parser.Token.TokenType = TOKEN_NAME then begin
          prKinParsList.Add(Parser.Token.Token);
        end;
      until (Parser.Token.TokenType = TOKEN_NULL);
    finally
      Parser.Free;
    end;
  end;

begin
  Form   := TQUB_Task_ExtractSelection_PropertiesForm(Sender);
  Props  := Form.Properties;

  Self.FileName                  := Form.FileName;

  Self.OpenFileAfter             := 0<>Props['OpenFileAfter'].Data[0];

  //Data Source
  Self.DataSource                := Form.DataSource;
  Self.ExtractFileList           := 0<>Props['ExtractFileList'].Data[0];

  //Data Processed
  Self.DataProcessed             := StringToDataProcessed(Props['ProcessData'].DataAsString);
  Self.ProcessFilter             := 0<>Props['ProcessData']['Filter'].Data[0];
  Self.ProcessFilterFreq         := Props['ProcessData']['FilterFreq'].Data[0] * 1000;
  Self.ProcessBaseLine           := 0<>Props['ProcessData']['Baseline'].Data[0];
  Self.ProcessResample           := 0<>Props['ProcessData']['Resample'].Data[0];
  Self.ProcessResampleFactor     := Props['ProcessData']['ResampleFactor'].Data[0];
  Self.ProcOnlyActiveChannel     := 0<>Props['ProcOnlyActiveChannel'].Data[0];

  //File Format
  Self.FileFormat                := Form.FileFormat;
  Self.Float4Byte                := 0<>Props['Float4Byte'].Data[0];
  Self.ASCIITime                 := 0<>Props['ASCIIHasTime'].Data[0];
  Self.DropFirstLast             := 0<>Props['DropFirstLast'].Data[0];
  Self.DropIfClass               := 0<>Props['DropIfClass'].Data[0];
  Self.DropClass                 := Props['DropClass'].Data[0];

  //Channels, Segments
  Self.Channels                  := Props['Channels'].Data[0];
  Self.ChannelList.Clear;
  Prop                           := Props['ChannelList'];
  if Self.Channels = CHANNELS_LIST then begin
    for i := 0 to Prop.DataCount - 1 do begin
      Self.ChannelList.Add(Pointer(Integer(Prop.Data[i])));
    end;
  end;
  Self.JoinSegments              := 0<>Props['JoinSegments'].Data[0];
  Self.ASCIIFit                  := 0<>Props['ASCIIHasFit'].Data[0];
  Self.ASCIISegCol               := 0<>Props['ASCIISegCol'].Data[0];
  Self.KinPars                   := 0<>Props['KinPars'].Data[0];
  Self.KinParsAll                := False;//0<>Props['KinParsAll'].Data[0];
  Self.KinParsList.Clear;
  MakeKinParsList(Props['KinParsList'].DataAsString);

  //Operation
  Self.Operation                 := StringToOperation(Props['Operation'].DataAsString);
  Self.OperationList             := 0<>Props['Operation']['List'].Data[0];
  Self.OperationListName         := Form.OperationListName;
  Self.OperationFiles            := 0<>Props['Operation']['Files'].Data[0];
  Self.OperationScale            := 0<>Props['Operation']['Scale'].Data[0];
  Self.OperationScaleFactor      := Props['Operation']['ScaleFactor'].Data[0];
  if 0<>Props['Operation']['ScaleOneOver'].Data[0] then Self.OperationScaleFactor := 1 / Self.OperationScaleFactor;

  //Function
  Self.FunctionDrift             := 0<>Props['Function']['Drift'].Data[0];
  Self.FunctionDriftA            := Props['Function']['DriftA'].Data[0];
  Self.FunctionDriftB            := Props['Function']['DriftB'].Data[0];

  Self.FunctionSine              := 0<>Props['Function']['Sine'].Data[0];
  Self.FunctionSineFrq           := Props['Function']['SineFreq'].Data[0];
  Self.FunctionSineAmp           := Props['Function']['SineAmp'].Data[0];
  Self.FunctionSinePh            := Props['Function']['SinePhase'].Data[0];

  Self.FunctionDigitize          := 0<>Props['Function']['Digitize'].Data[0];
  Self.FunctionDigitizeThreshold := Props['Function']['DigitizeThreshold'].Data[0];
  Self.FunctionDigitizeToMin     := Props['Function']['DigitizeToMin'].Data[0];
  Self.FunctionDigitizeToMax     := Props['Function']['DigitizeToMax'].Data[0];

  Self.FunctionDerivate          := 0<>Props['Function']['Derivate'].Data[0];

  Self.FunctionLKC               := 0<>Props['Function']['LKC'].Data[0];

  Self.FunctionItoG              := 0<>Props['Function']['ItoG'].Data[0];
  Self.FunctionItoGVRev          := Props['Function']['ItoGVRev'].Data[0];
  Self.FunctionItoGVCh           := Props['Function']['ItoGVCh'].Data[0];

  Self.FunctionRsCorr            := 0<>Props['Function']['RsCorr'].Data[0];
  Self.FunctionRsValue           := Props['Function']['RsValue'].Data[0];
  Self.FunctionRsPer             := Props['Function']['RsPer'].Data[0];
  Self.FunctionRsAddLeak         := 0<>Props['Function']['RsAddLeak'].Data[0];
  Self.FunctionRsLeakCh          := Props['Function']['RsLeakCh'].Data[0];

  Self.FunctionPnAdd             := 0<>Props['Function']['PnAdd'].Data[0];
  Self.FunctionPnSub             := 0<>Props['Function']['PnSub'].Data[0];
  Self.FunctionPnOffset          := Props['Function']['PnOffset'].Data[0];
  Self.FunctionPnDelay           := Props['Function']['PnDelay'].Data[0];
  Self.FunctionPnScale           := Props['Function']['PnScale'].Data[0];
  Self.FunctionPnCount           := Props['Function']['PnCount'].Data[0];
  Self.FunctionPnVm              := Props['Function']['PnVm'].Data[0];
  Self.FunctionPnInterval        := Props['Function']['PnInterval'].Data[0];

  Self.FunctionFit               := 0<>Props['Function']['Fit'].Data[0];

  if TaskExpPropertiesForm <> nil then begin
    Self.LKCVChannel             := TaskExpPropertiesForm.LKCVChannel;
  end;
end;

procedure TQUB_Task_ExtractSelection.OnPutProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_ExtractSelection_PropertiesForm do begin
    FileName                        := Self.FileName;
    //FileFormat                      := Self.FileFormat;
    //JoinSegments                    := Self.JoinSegments;
  end;//with
end;

//******************************************************************************
procedure TQUB_Task_ExtractSelection.prSetData(Data :TQUB_QVO_QVD);
var
  Idx :Integer;
begin
  prData := Data;
  Data.LockWrite;
  prSelectionList    := nil;
  prSubSelectionList := nil;
  if Data.SelectionLists.Items.Count > 0 then begin
    if Data.SelectionLists.ItemIndex <> -1 then begin
      prSelectionList := Data.SelectionLists[Data.SelectionLists.ItemIndex];
      prSelectionList.Lock;
    end;
  end;
  if OperationList then begin
    Idx  := Data.SelectionLists.Items.IndexOf(OperationListName);
    if Idx >= 0 then begin
      prSubSelectionList := Data.SelectionLists[Idx];
      prSubSelectionList.Lock;
    end;
  end;
end;


//******************************************************************************
end.
