unit QUB_Task_IdlBase;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


QUB_Matrices,

Windows, SysUtils, Classes, Math, IEEE754, Dialogs, Controls, Graphics, SyncObjs, Forms,

Alloc, AllocTypes,
qubopt_iface,

QUB_Matrix,

QUB_Global,
QUB_QVO_QVD_Global,
QUB_Thread,
QUB_Task, QUB_Task_Data,
QUB_IdlBase_Global,
QUB_QFS_QMF, QUB_QFS_QMF_QMatrix,
QUB_QFS_Channels,
QUB_QVO_QVM,
QUB_Kalman,
QUB_Task_IdlBase_Properties,
QUB_Node,
QUB_Node_Python,
QUB_Movies,
QUB_ExpCond_Parameters,
QUB_Q_Model, QUB_A2Q_Model, QUB_A2Q_Global;


const


USE_P                                               = False;
IDLBASE_REASONABLE_NPOINTS                          = 500000;

BASE_CONV                                           = 1e-10;//@@@
BASE_EST_MAX_COUNT                                  = 20;
MIN_REPR_LEVEL_GAMM                                 = 0.1;


type


TQUB_Task_IdlBase = class(TQUB_Task_Data)
protected


  //Source Objects
  prModel                  :TQUB_QVO_QVM;
  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  //Parameters
  prParams                 :TQUB_IdlBase_Params;

  //Model
  prStateCount             :Integer;
  prClsCount               :Integer;
  prQ                      :TPMDouble;
  prA                      :TPMDouble;
  prP0                     :TPADouble;
  prClsIdx                 :TPAInteger;
  prClsIdxR                :TPAInteger;
  prClsOfState             :TPAInteger;
  prGrpOfState             :TPAInteger;
  prAmp                    :TPADouble;
  prStd                    :TPADouble;

  prJumpAmp                :Double;
  prJumpAmps               :TPADouble;
  prJumpAmpsCount          :Integer;
  prMeasStd                :Double;
  prBaseAmp                :Double;

  //TotStats
  prTotA                   :TPMDouble;
  prTotJumpAmps            :TPADouble;
  prTotJumpAmp             :Double;
  prTotJumpStd             :Double;
  prTotMeasStd             :Double;
  prTotBaseAmp             :Double;
  prTotKf                  :Double;
  prTotKb                  :Double;
  prTotLL                  :Double;
  prTotCount               :Integer;

  prGlobalFillData         :TNodeMethPtr; // Interface_Modeling.FillDataCallback used by FillStatData
  prResultNode             :INode;

  prRealityCheckSaysGo     :Boolean;

  prMovie                  :TQUB_MovieRec;
  prMovieState             :TQUB_MovieState;
  prMoviePointIndex        :Integer;
  prMovieFrameIndex        :Integer;
  prMovieIteration         :Integer;
  prMovieFBKIter           :Integer;


  prA2QModel               :TQUB_A2Q_Model;
  prA2QParams              :TQUB_A2Q_Opt_Params;
  prfParsGrads             :TPADouble;
  prpLL                    :Double;
  prSampling               :Double;
  prReestA                 :TPMDouble;


  // used by prFillDataBaseline during Stat during prReest*
  prDataSet                :INode;
  prStatChunk              :TQUB_IdlBase_PChunk;
  prStatBase               :TPADouble;
  prCurrChunkIndex         :Integer;


  //Execute
  procedure prDoTask; override;
  procedure prRealityCheck; // synchronize; sets prRealityCheckSaysGo := true to proceed

  procedure prDoExecute; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;

  //Chunks
  procedure prSetChunkParameters; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prSetModel(Model :TQUB_QVO_QVM);

  function DataCallBack(IdlBaseChunk :TQUB_IdlBase_PChunk) :Integer;

  procedure prFinishResults;

  procedure prIdlBase;
  procedure prIdl_Staircase;

  function prReestSKM(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
  function prReestBW(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
  function prReestBWKII(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;

  function prForwardViterbi(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;

  function prReest_Staircase_SKM(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
  function prReest_Staircase_BW(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;

  function prBase2Nodes(IdlBaseChunk :TQUB_IdlBase_PChunk; BaseX :TPADouble) :Boolean;
  procedure prMakeDwells(IdlBaseChunk :TQUB_IdlBase_PChunk; State :TPAInteger; DataCount :Integer);

  function prMakeStats(IdlBaseChunk :TQUB_IdlBase_PChunk; BaseX :TPADouble) :Boolean;
  function prBuildIdlSegNode(IdlBaseChunk :TQUB_IdlBase_PChunk): INode;
  function prRunStat(SegNode :INode): INode;
  function prFillDataBaseline(Seg :INode): INode;

  procedure prApplyDeadTimeSamples(IdlBaseChunk :TQUB_IdlBase_PChunk; DeadCount :Integer);

  //Used to get an A that closest matches the re-estimated AMatrix and satisfies the constraints on Q
  function prA2Q(AMatrix, AMatrixN :TPMDouble; IsQ :Boolean) :Boolean;
  function prA2Q_SolvOpt(AMatrix :TPMDouble) :Boolean;
  function prLLGrad(CompGrad :Integer) :Integer;
  function prLLGradQ(CompGrad :Integer) :Integer;
  procedure prPrintParams;
  procedure prUpdateModel;

  //Constrained Amplitude/Variance re-estimation
  function CstrAmpEst(Ct, Sum_Gamma, Sum_Gamma_y, Amp :TPADouble) :Boolean;
  function CstrVarEst(Ct, Sum_Gamma, Sum_Gamma_y, Std :TPADouble) :Boolean;


  procedure  prResultsToData; override;

  procedure prPublishResultsPy;


  //Movies
  procedure prMakeMovie_ViterbiKalman;
  procedure prMakeMovie_ViterbiKalman_BeforeDrawChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_ViterbiKalman_AfterDrawChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_ViterbiKalman_BeforeDrawIdlChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_ViterbiKalman_AfterDrawIdlChannel(Sender :TObject; Channel :Integer);

  procedure prMakeMovie_ForwardBackwardKalman;
  procedure prMakeMovie_ForwardBackwardKalman_BeforeDrawChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_ForwardBackwardKalman_AfterDrawChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_ForwardBackwardKalman_BeforeDrawIdlChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_ForwardBackwardKalman_AfterDrawIdlChannel(Sender :TObject; Channel :Integer);

  procedure prMakeMovie_Staircase;
  procedure prMakeMovie_Staircase_BeforeDrawChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_Staircase_AfterDrawChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_Staircase_BeforeDrawIdlChannel(Sender :TObject; Channel :Integer);
  procedure prMakeMovie_Staircase_AfterDrawIdlChannel(Sender :TObject; Channel :Integer);

  function prMovie_BitmapFileName(const Name :String; Digits :Integer; Index :Integer) :String;
  procedure prMovie_DrawAndSave(const MovieName :String; var FrameIndex :Integer);


public


  constructor Create(Suspended :Boolean); override;


  function SOLVOPTFunc(Pars :TPADouble) :Double; stdcall;
  procedure SOLVOPTGrad(Pars :TPADouble; Grad :TPADouble); stdcall;
  function DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
  function DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
  function DFPFuncQ(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;


  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model              :TQUB_QVO_QVM                 read prModel              write prSetModel;

  //Parameters
  property Params             :TQUB_IdlBase_Params          read prParams             write prParams;

  //Movies
  property Movie              :TQUB_MovieRec                read prMovie              write prMovie;

  //Interface_Modeling's filldatacallback used by stat
  property GlobalFillData     :TNodeMethPtr                 read prGlobalFillData     write prGlobalFillData;


end;


var


//Used to make movies
MovieTree                 :INode;


implementation


uses


QUB_QFS,
QUB_Viewport_Results,
//QUB_Viewport_Histogram,
QUB_Viewport_Python,
QUB_Python_Data,
QUB_GUIInterface,
QUB_Viewport_Data,
QUB_QVO_QVD,
QUB_Task_Stat,
QUB_Main;


//DFP Callbacks ****************************************************************
function DFPFunc(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
begin
  Result := TQUB_Task_IDLBase(call_obj).DFPFunc(Pars, Grad, LL, nf, ndf, CompGrad);
end;

function DFPFuncQ(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
begin
  Result := TQUB_Task_IDLBase(call_obj).DFPFuncQ(Pars, Grad, LL, nf, ndf, CompGrad);
end;

function DFPCheck(call_obj :TObject; Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
begin
  Result := TQUB_Task_IDLBase(call_obj).DFPCheck(Iters, nf, ndf, LL, Pars, Grads);
end;

//SOLVOPT Callbacks ************************************************************
function SOLVOPTFunc(call_obj :TObject; Pars :TPADouble) :Double; stdcall;
begin
  Result := TQUB_Task_IDLBase(call_obj).SOLVOPTFunc(Pars);
end;

procedure SOLVOPTGrad(call_obj :TObject; Pars :TPADouble; Grad :TPADouble); stdcall;
begin
  TQUB_Task_IDLBase(call_obj).SOLVOPTGrad(Pars, Grad);
end;


//SOLVOPT optimization
function TQUB_Task_IdlBase.SOLVOPTFunc(Pars :TPADouble) :Double; stdcall;
var
  i :Integer;
begin

  //Copy the new pars into the model
  for i := 0 to prA2QModel.fParCount - 1 do begin
    prA2QModel.fPars[i] := Pars[i];
  end;
  prA2QModel.Do_fParsToModel;

  try
    prLLGrad(0{CompGrad});
  except
  end;

  Result := prA2QModel.LL;

  Inc(prA2QModel.Iterations);
  Inc(prA2QModel.FuncCalls);

end;

procedure TQUB_Task_IdlBase.SOLVOPTGrad(Pars :TPADouble; Grad :TPADouble); stdcall;
var
  i :Integer;
begin

  //Copy the new pars into the model
  for i := 0 to prA2QModel.fParCount - 1 do begin
    prA2QModel.fPars[i] := Pars[i];
  end;
  prA2QModel.Do_fParsToModel;

  try
    prLLGrad(1{CompGrad});
  except
  end;

  for i := 0 to prA2QModel.fParCount - 1 do begin
    Grad[i] := prA2QModel.dLLdfPar[i];
  end;
  Move(Grad[0], prfParsGrads[0], prA2QModel.fParCount * SizeOf(Double));

  Inc(prA2QModel.Iterations);
  Inc(prA2QModel.GradCalls);

end;

//DFP optimization
function TQUB_Task_IdlBase.DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
var
  i :Integer;
begin

  Result := 0;

  if prA2QModel.Iterations > 0 then begin

    //Copy the new pars into the model
    for i := 0 to prA2QModel.fParCount - 1 do begin
      prA2QModel.fPars[i] := Pars[i];
    end;

    prA2QModel.Do_fParsToModel;

  end;


  try
    prLLGrad(CompGrad);
  except
  end;


  LL := prA2QModel.LL;


  if CompGrad <> 0 then begin
    for i := 0 to prA2QModel.fParCount - 1 do begin
      Grad[i] := prA2QModel.dLLdfPar[i];
    end;
    Move(Grad[0], prfParsGrads[0], prA2QModel.fParCount * SizeOf(Double));
  end;


  Inc(prA2QModel.Iterations);
  Inc(nf);
  if CompGrad <> 0 then Inc(ndf);
  prA2QModel.FuncCalls := nf;
  prA2QModel.GradCalls := ndf;

end;

function TQUB_Task_IdlBase.DFPFuncQ(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
var
  i :Integer;
begin

  Result := 0;

  if prA2QModel.Iterations > 0 then begin

    //Copy the new pars into the model
    for i := 0 to prA2QModel.fParCount - 1 do begin
      prA2QModel.fPars[i] := Pars[i];
    end;

    prA2QModel.Do_fParsToModel;

  end;


  try
    prLLGradQ(CompGrad);
  except
  end;


  LL := prA2QModel.LL;


  if CompGrad <> 0 then begin
    for i := 0 to prA2QModel.fParCount - 1 do begin
      Grad[i] := prA2QModel.dLLdfPar[i];
    end;
    Move(Grad[0], prfParsGrads[0], prA2QModel.fParCount * SizeOf(Double));
  end;


  Inc(prA2QModel.Iterations);
  Inc(nf);
  if CompGrad <> 0 then Inc(ndf);
  prA2QModel.FuncCalls := nf;
  prA2QModel.GradCalls := ndf;

end;

function TQUB_Task_IDLBase.DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
var
  LLConverged :Boolean;
  GRConverged :Boolean;
  k           :Integer;
begin

  Result := 0;

  prA2QModel.Iterations := Iters;

  if Iters >= prA2QParams.OptMaxIter then Result := 1;

  LLConverged := ((Abs(LL - prpLL)) / (LL + 1e-30)) < prA2QParams.OptLLConv;
  //LLConverged := Abs(LL / Sqr(prStateCount)) < prA2QParams.OptLLConv;
  prpLL := LL;

  GRConverged := True;
  for k := 0 to prA2QModel.fParCount - 1 do begin
    if Abs(Grads[k]) > prA2QParams.OptGradConv then GRConverged := False;
  end;


  if (LLConverged) and {or} (GRConverged) then Result := 1;
  if prStopped then Result := 1;

  if Result = 1 then begin
    //Change the parameters to the last good ones (???)
    //Copy the new pars into the A2Q models
    for k := 0 to prA2QModel.fParCount - 1 do begin
      prA2QModel.fPars[k] := Pars[k];
    end;
    prA2QModel.Do_fParsToModel;
    //Only the fParsGrads
    Move(Grads[0], prfParsGrads[0], prA2QModel.fParCount * SizeOf(Double));
  end;

end;

//****************************************************************************//
//                                                                            //
//                             L L   +   G R A D                              //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IDLBase.prLLGrad(CompGrad :Integer) :Integer;
var
  a, ar, dadq    :Double;
  k, i, j        :Integer;
  SomeConfluent  :Boolean;
begin


  Result    := 0;


  SomeConfluent := False;

  if prA2QModel.Confluent[0] <> 0 then SomeConfluent := True;

  //Reset LL and derivatives w.r.t. q's ****************************************
  for k := 0 to prA2QModel.RateCount - 1 do begin
    prA2QModel.dLLdq[0][k] := 0;
  end;

  // reset deriv for cond
  for k := 0 to prA2QModel.RateCount - 1 do begin
    prA2QModel.dLLdCond_q[k] := 0;
  end;

  prA2QModel.LL         := 0;
  //****************************************************************************


  //Calculate A matrix and its derivatives w.r.t. each qq, and the products and their derivatives
  prA2QModel.Do_A(prSampling);
  prA2QModel.Do_dA_dq(prSampling, 0{ECClass});


  //Cost function ..............................................................
  for i := 0 to prStateCount - 1 do begin
    for j := 0 to prStateCount - 1 do begin
      //if i = j then begin
        a  := Abs(TPMDouble(prA2QModel.A[0])[i, j]);
        ar := prReestA[i, j];
        //r  := a / ar;
        //prA2QModel.LL := prA2QModel.LL + Sqr(ln(r));
        //prA2QModel.LL := prA2QModel.LL + Sqr( (ln(a) - ln(ar)) / (ln(ar)) );
        //prA2QModel.LL := prA2QModel.LL + Sqr((a - ar)/ar);
        prA2QModel.LL := prA2QModel.LL + Sqr((a - ar)/a);
        //prA2QModel.LL := prA2QModel.LL + Sqr(a - ar);
      //end;
    end;
  end;
  //............................................................................


  //Rates Gradients ............................................................
  for k := 0 to prA2QModel.RateCount - 1 do begin

    for i := 0 to prStateCount - 1 do begin
      for j := 0 to prStateCount - 1 do begin
        //if i = j then begin
          a    := Abs(TPMDouble(prA2QModel.A[0])[i, j]);
          ar   := prReestA[i, j];
          //r    := a / ar;
          dadq := TPMDouble(TPAPointer(prA2QModel.dAdq[0])[k])[i, j];
          //prA2QModel.dLLdq[0][k] := prA2QModel.dLLdq[0][k] + 2 * ln(r) * (1/a) * dadq;
          //prA2QModel.dLLdq[0][k] := prA2QModel.dLLdq[0][k] + 2 * (ln(a) - ln(ar)) * (1 / Sqr(ar)) * (1 / a) * dadq;
          //prA2QModel.dLLdq[0][k] := prA2QModel.dLLdq[0][k] + 2 * (a - ar) * (1 / Sqr(ar)) * dadq;
          prA2QModel.dLLdq[0][k] := prA2QModel.dLLdq[0][k] + 2 * dadq * (a - ar) * ( (1 / (a*a)) - (a - ar) * (1 / (a*a*a)) );
          //prA2QModel.dLLdq[0][k] := prA2QModel.dLLdq[0][k] + 2 * (a - ar) * dadq;
        //end;
      end;
    end;

  end;//of k


  //Calculate free pars gradients
  prA2QModel.Do_dLLdfPar;


  if SomeConfluent then begin
    prConsoleMsg  := 'Confluent eigen values detected.';
    prConsoleCode := MESSAGE_ERROR;
    Synchronize(prWriteNewConsole);
  end;
  if prA2QModel.SmallRateWarning then begin
    prConsoleMsg  := 'Very small rate(s) detected.';
    prConsoleCode := MESSAGE_WARNING;
    Synchronize(prWriteNewConsole);
  end;
  if prA2QModel.BigRateWarning then begin
    prConsoleMsg  := 'Very big rate(s) detected.';
    prConsoleCode := MESSAGE_WARNING;
    Synchronize(prWriteNewConsole);
  end;

end;

function TQUB_Task_IDLBase.prLLGradQ(CompGrad :Integer) :Integer;
var
  r, a, ar, dadq :Double;
  i, j, k        :Integer;
  SomeConfluent  :Boolean;
begin


  Result    := 0;


  SomeConfluent := False;

  if prA2QModel.Confluent[0] <> 0 then SomeConfluent := True;

  //Reset LL and derivatives w.r.t. q's ****************************************
  for k := 0 to prA2QModel.RateCount - 1 do begin
    prA2QModel.dLLdq[0][k] := 0;
  end;

  // reset deriv for cond
  for k := 0 to prA2QModel.RateCount - 1 do begin
    prA2QModel.dLLdCond_q[k] := 0;
  end;

  prA2QModel.LL         := 0;
  //****************************************************************************


  //Calculate A matrix and its derivatives w.r.t. each qq, and the products and their derivatives
  prA2QModel.Do_A(prSampling);
  prA2QModel.Do_dA_dq(prSampling, 0{ECClass});


  //Cost function ..............................................................
  for i := 0 to prStateCount - 1 do begin
    for j := 0 to prStateCount - 1 do begin
      a  := Abs(TPMDouble(prA2QModel.Q[0])[i, j]);
      ar := prReestA{Q}[i, j];
      if ar <> 0 then begin
        r  := a / ar;
        prA2QModel.LL := prA2QModel.LL + Sqr(ln(r));
      end;
    end;
  end;
  //............................................................................


  //Rates Gradients ............................................................
  for k := 0 to prA2QModel.RateCount - 1 do begin

    for i := 0 to prStateCount - 1 do begin
      for j := 0 to prStateCount - 1 do begin
        a    := Abs(TPMDouble(prA2QModel.Q[0])[i, j]);
        ar   := prReestA{Q}[i, j];
        if ar <> 0 then begin
          r    := a / ar;
          dadq := TPMDouble(prA2QModel.dQdq[k])[i, j];
          prA2QModel.dLLdq[0][k] := prA2QModel.dLLdq[0][k] + 2 * ln(r) * (1/a) * dadq;
        end;
      end;
    end;

  end;//of k

  //Calculate free pars gradients
  prA2QModel.Do_dLLdfPar;


  if SomeConfluent then begin
    prConsoleMsg  := 'Confluent eigen values detected.';
    prConsoleCode := MESSAGE_ERROR;
    Synchronize(prWriteNewConsole);
  end;
  if prA2QModel.SmallRateWarning then begin
    prConsoleMsg  := 'Very small rate(s) detected.';
    prConsoleCode := MESSAGE_WARNING;
    Synchronize(prWriteNewConsole);
  end;
  if prA2QModel.BigRateWarning then begin
    prConsoleMsg  := 'Very big rate(s) detected.';
    prConsoleCode := MESSAGE_WARNING;
    Synchronize(prWriteNewConsole);
  end;


end;

//******************************************************************************
procedure TQUB_Task_IDLBase.prUpdateModel;
begin

  MakeRatesFromParams(Model.ModelObject, prA2QModel.cPars, 2*prA2QModel.RateCount, False{AllowNegRates});

end;

procedure TQUB_Task_IDLBase.prPrintParams;
var
  LogL :Double;
  t    :String;
  k    :Integer;
begin

  t :=     #13#10'================================================================';

  LogL := prA2QModel.LL;

  try
    t := t + Format(#13#10'A2Q final results:  SS = %-14.10g Iterations = %d,'#13#10 +
                   'Rates:  '#13#10#9's1  s2   k0         k0.std     k1         k1.std' +
                             #13#10#9'--------------------------------------------------',
                   [LogL, prA2QModel.Iterations]
                   );
  except
  end;

  for k := 0 to prA2QModel.RateCount - 1 do begin
    try
      //t := t + Format(#13#10#9'%d   %d    %-10.6g %-10.6g %-10.6g %-10.6g',
      t := t + Format(#13#10#9'%d   %d    %-10.6g'#9'%-10.6g'#9'%-10.6g'#9'%-10.6g',
                     [prA2QModel.RateIdx[k].x+1, prA2QModel.RateIdx[k].y+1,
                     //!!! cParsVar
                      Exp(prA2QModel.cPars[2*k + 0]), Sqrt(prA2QModel.cParsVar[2*k + 0]),
                          prA2QModel.cPars[2*k + 1],  Sqrt(prA2QModel.cParsVar[2*k + 1])
                     ]);
    except
    end;
  end;
  t := t + #13#10#9'--------------------------------------------------';

  t := t + #13#10'Free pars grads: ';
  for k := 0 to prA2QModel.fParCount - 1 do begin
    try
      t := t + Format('%-12.8g ', [prfParsGrads[k]]);
    except
    end;
  end;

  t := t + #13#10'================================================================';


  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
end;


//Constrained Amplitude / Variance re-estimation
function TQUB_Task_IdlBase.CstrAmpEst(Ct, Sum_Gamma, Sum_Gamma_y, Amp :TPADouble) :Boolean;
var
  Cstr        :TQUB_QFS_ConstraintAmpVar;
  Amp_CstrCnt :Integer;
  Amp_Size    :Integer;
  Amp_A       :TPMDouble;
  Amp_AI      :TPMDouble;
  Amp_B       :TPADouble;
  Amp_X       :TPADouble;
  i, j, k, c  :Integer;
begin


  Result      := False;


  Amp_A       := nil;
  Amp_AI      := nil;
  Amp_B       := nil;
  Amp_X       := nil;


  try


    {
      Must solve the equation Ax = B
        x is the vector of unknowns (amplitudes), including the lambda multipliers, if any;
        A and B are obtained by setting the derivatives of the cost function to zero
        First row is for Amp0, the rest for Ampi>0x, and for lambda
    }

    //Allocate A and B:
    Amp_CstrCnt := 0;
    for i := 0 to prModel.ModelObject.ConstraintsAmpVar.ConstraintCount - 1 do begin
      case prModel.ModelObject.ConstraintsAmpVar[i].cType of
        CSTR_FIX_AMP, CSTR_FIX_EXC_AMP, CSTR_FIX_DIFF_EXC_AMP, CSTR_SCALE_EXC_AMP, CSTR_SCALE_DIFF_EXC_AMP:
          Inc(Amp_CstrCnt);
      end;
    end;
    if Amp_CstrCnt = 0 then Exit;

    Amp_Size    := prClsCount + Amp_CstrCnt;
    Amp_A       := DAlloc2(Amp_Size, Amp_Size);
    Amp_AI      := DAlloc2(Amp_Size, Amp_Size);
    Amp_B       := DAlloc1(Amp_Size);
    Amp_X       := DAlloc1(Amp_Size);


    //Zero the matrices
    for i := 0 to Amp_Size - 1 do begin
      Amp_B[i] := 0;
      Amp_X[i] := 0;
      for j := 0 to Amp_Size - 1 do begin
        Amp_A[i, j] := 0;
      end;
    end;


    //Amp_B
    i := 0;
    for j := 0 to prClsCount - 1 do begin
      Amp_B[i] := Amp_B[i] + 2 * Sum_Gamma_y[j];
    end;
    for i := 1 to prClsCount - 1 do begin
      Amp_B[i] :=            2 * Sum_Gamma_y[i];
    end;

    //Amp_A
    i := 0;
    j := 0;
    for k := 0 to prClsCount - 1 do begin
      Amp_A[i, j] := Amp_A[i, j] + 2 * Sum_Gamma[k];
    end;
    for j := 1 to prClsCount - 1 do begin
      Amp_A[i, j] :=               2 * Sum_Gamma[j];
    end;
    for i := 1 to prClsCount - 1 do begin
      Amp_A[i, 0] :=               2 * Sum_Gamma[i];
      Amp_A[i, i] :=               2 * Sum_Gamma[i];
    end;


    //Lagrange multipliers
    c := 0;
    for i := 0 to prModel.ModelObject.ConstraintsAmpVar.ConstraintCount - 1 do begin

      Cstr := prModel.ModelObject.ConstraintsAmpVar[i];

      case Cstr.cType of

        CSTR_FIX_AMP:
          begin
            Amp_B[prClsCount + c]                  := Amp_B[prClsCount + c]                  + Ct[c];
            if (Cstr.Index[0] = 0) then begin
              Amp_A[prClsCount + c, 0]             := Amp_A[prClsCount + c, 0]               + 1;
              Amp_A[0, prClsCount + c]             := Amp_A[0, prClsCount + c]               + 1;
            end else begin
              Amp_A[prClsCount + c, 0]             := Amp_A[prClsCount + c, 0]               + 1;
              Amp_A[prClsCount + c, Cstr.Index[0]] := Amp_A[prClsCount + c, Cstr.Index[0]]   + 1;
              Amp_A[0,             prClsCount + c] := Amp_A[0,             prClsCount + c]   + 1;
              Amp_A[Cstr.Index[0], prClsCount + c] := Amp_A[Cstr.Index[0], prClsCount + c]   + 1;
            end;
            Inc(c);
          end;
        CSTR_FIX_EXC_AMP:
          begin
            Amp_B[prClsCount + c]                  := Amp_B[prClsCount + c]                  + Ct[c];
            Amp_A[prClsCount + c, Cstr.Index[0]]   := Amp_A[prClsCount + c, Cstr.Index[0]]   + 1;
            Amp_A[Cstr.Index[0], prClsCount + c]   := Amp_A[Cstr.Index[0], prClsCount + c]   + 1;
            Inc(c);
          end;
        CSTR_FIX_DIFF_EXC_AMP:
          begin
            if (Cstr.Index[0] <> 0) and (Cstr.Index[1] <> 0) then begin
              Amp_B[prClsCount + c]                  := Amp_B[prClsCount + c]                + Ct[c];
              Amp_A[prClsCount + c, Cstr.Index[1]]   := Amp_A[prClsCount + c, Cstr.Index[1]] + 1;
              Amp_A[prClsCount + c, Cstr.Index[0]]   := Amp_A[prClsCount + c, Cstr.Index[0]] - 1;
              Amp_A[Cstr.Index[1], prClsCount + c]   := Amp_A[Cstr.Index[1], prClsCount + c] + 1;
              Amp_A[Cstr.Index[0], prClsCount + c]   := Amp_A[Cstr.Index[0], prClsCount + c] - 1;
            end else if (Cstr.Index[0] = 0)  and (Cstr.Index[1] <> 0) then begin
              Amp_B[prClsCount + c]                  := Amp_B[prClsCount + c]                + Ct[c];
              Amp_A[prClsCount + c, Cstr.Index[1]]   := Amp_A[prClsCount + c, Cstr.Index[1]] + 1;
              Amp_A[Cstr.Index[1], prClsCount + c]   := Amp_A[Cstr.Index[1], prClsCount + c] + 1;
            end else if (Cstr.Index[0] <> 0) and (Cstr.Index[1] =  0) then begin
              Amp_B[prClsCount + c]                  := Amp_B[prClsCount + c]                - Ct[c];
              Amp_A[prClsCount + c, Cstr.Index[0]]   := Amp_A[prClsCount + c, Cstr.Index[0]] + 1;
              Amp_A[Cstr.Index[0], prClsCount + c]   := Amp_A[Cstr.Index[0], prClsCount + c] + 1;
            end;
            Inc(c);
          end;
        CSTR_SCALE_EXC_AMP:
          begin
            Amp_A[prClsCount + c, Cstr.Index[1]]   := Amp_A[prClsCount + c, Cstr.Index[1]]   + 1;
            Amp_A[prClsCount + c, Cstr.Index[0]]   := Amp_A[prClsCount + c, Cstr.Index[0]]   - Ct[c];
            Amp_A[Cstr.Index[1], prClsCount + c]   := Amp_A[Cstr.Index[1], prClsCount + c]   + 1;
            Amp_A[Cstr.Index[0], prClsCount + c]   := Amp_A[Cstr.Index[0], prClsCount + c]   - Ct[c];
            Inc(c);
          end;
        CSTR_SCALE_DIFF_EXC_AMP:
          begin
            Amp_A[prClsCount + c, Cstr.Index[1]]   := Amp_A[prClsCount + c, Cstr.Index[1]]   + 1;
            Amp_A[prClsCount + c, Cstr.Index[0]]   := Amp_A[prClsCount + c, Cstr.Index[0]]   - 1;
            Amp_A[prClsCount + c, Cstr.Index[3]]   := Amp_A[prClsCount + c, Cstr.Index[3]]   - Ct[c];
            Amp_A[prClsCount + c, Cstr.Index[2]]   := Amp_A[prClsCount + c, Cstr.Index[2]]   + Ct[c];
            Amp_A[Cstr.Index[1], prClsCount + c]   := Amp_A[Cstr.Index[1], prClsCount + c]   + 1;
            Amp_A[Cstr.Index[0], prClsCount + c]   := Amp_A[Cstr.Index[0], prClsCount + c]   - 1;
            Amp_A[Cstr.Index[3], prClsCount + c]   := Amp_A[Cstr.Index[3], prClsCount + c]   - Ct[c];
            Amp_A[Cstr.Index[2], prClsCount + c]   := Amp_A[Cstr.Index[2], prClsCount + c]   + Ct[c];
            Inc(c);
          end;

      end;//case

    end;//for i


    //X = Inv(A) * B
    InvMat(PMatrix(Amp_A), 0, Amp_Size - 1, PMatrix(Amp_AI));

    for k := 0 to Amp_Size - 1 do begin

      for i := 0 to Amp_Size - 1 do begin

        Amp_X[k] := Amp_X[k] + Amp_AI[k, i] * Amp_B[i];

      end;


    end;

    Amp[0] := Amp_X[0];
    for i := 1 to prClsCount - 1 do begin

      Amp[i] := Amp_X[i] + Amp_X[0];

    end;

  finally

    Free2(TPMChar(Amp_A));
    Free2(TPMChar(Amp_AI));
    FreeMem(Amp_B);
    FreeMem(Amp_X);

  end;

end;

function TQUB_Task_IdlBase.CstrVarEst(Ct, Sum_Gamma, Sum_Gamma_y, Std :TPADouble) :Boolean;
var
  Cstr        :TQUB_QFS_ConstraintAmpVar;
  Var_CstrCnt :Integer;
  Var_Size    :Integer;
  Var_A       :TPMDouble;
  Var_AI      :TPMDouble;
  Var_B       :TPADouble;
  Var_X       :TPADouble;
  i, j, k, c  :Integer;
begin


  Result      := False;


  Var_A       := nil;
  Var_AI      := nil;
  Var_B       := nil;
  Var_X       := nil;


  try


    {
      Must solve the equation Ax = B
        x is the vector of unknowns (amplitudes), including the lambda multipliers, if any;
        A and B are obtained by setting the derivatives of the cost function to zero
        First row is for Var0, the rest for Vari>0x, and for lambda
    }

    //Allocate A and B:
    Var_CstrCnt := 0;
    for i := 0 to prModel.ModelObject.ConstraintsAmpVar.ConstraintCount - 1 do begin
      case prModel.ModelObject.ConstraintsAmpVar[i].cType of
        CSTR_FIX_VAR, CSTR_FIX_EXC_VAR, CSTR_FIX_DIFF_EXC_VAR, CSTR_SCALE_EXC_VAR, CSTR_SCALE_DIFF_EXC_VAR:
          Inc(Var_CstrCnt);
      end;
    end;
    if Var_CstrCnt = 0 then Exit;

    Var_Size    := prClsCount + Var_CstrCnt;
    Var_A       := DAlloc2(Var_Size, Var_Size);
    Var_AI      := DAlloc2(Var_Size, Var_Size);
    Var_B       := DAlloc1(Var_Size);
    Var_X       := DAlloc1(Var_Size);


    //Zero the matrices
    for i := 0 to Var_Size - 1 do begin
      Var_B[i] := 0;
      Var_X[i] := 0;
      for j := 0 to Var_Size - 1 do begin
        Var_A[i, j] := 0;
      end;
    end;


    //Var_B
    i := 0;
    for j := 0 to prClsCount - 1 do begin
      Var_B[i] := Var_B[i] + 2 * Sum_Gamma_y[j];
    end;
    for i := 1 to prClsCount - 1 do begin
      Var_B[i] :=            2 * Sum_Gamma_y[i];
    end;

    //Var_A
    i := 0;
    j := 0;
    for k := 0 to prClsCount - 1 do begin
      Var_A[i, j] := Var_A[i, j] + 2 * Sum_Gamma[k];
    end;
    for j := 1 to prClsCount - 1 do begin
      Var_A[i, j] :=               2 * Sum_Gamma[j];
    end;
    for i := 1 to prClsCount - 1 do begin
      Var_A[i, 0] :=               2 * Sum_Gamma[i];
      Var_A[i, i] :=               2 * Sum_Gamma[i];
    end;


    //Lagrange multipliers
    c := 0;
    for i := 0 to prModel.ModelObject.ConstraintsAmpVar.ConstraintCount - 1 do begin

      Cstr := prModel.ModelObject.ConstraintsAmpVar[i];

      case Cstr.cType of

        CSTR_FIX_VAR:
          begin
            Var_B[prClsCount + c]                  := Var_B[prClsCount + c]                  + Ct[c];
            if (Cstr.Index[0] = 0) then begin
              Var_A[prClsCount + c, 0]             := Var_A[prClsCount + c, 0]               + 1;
              Var_A[0, prClsCount + c]             := Var_A[0, prClsCount + c]               + 1;
            end else begin
              Var_A[prClsCount + c, 0]             := Var_A[prClsCount + c, 0]               + 1;
              Var_A[prClsCount + c, Cstr.Index[0]] := Var_A[prClsCount + c, Cstr.Index[0]]   + 1;
              Var_A[0,             prClsCount + c] := Var_A[0,             prClsCount + c]   + 1;
              Var_A[Cstr.Index[0], prClsCount + c] := Var_A[Cstr.Index[0], prClsCount + c]   + 1;
            end;
            Inc(c);
          end;
        CSTR_FIX_EXC_VAR:
          begin
            Var_B[prClsCount + c]                  := Var_B[prClsCount + c]                  + Ct[c];
            Var_A[prClsCount + c, Cstr.Index[0]]   := Var_A[prClsCount + c, Cstr.Index[0]]   + 1;
            Var_A[Cstr.Index[0], prClsCount + c]   := Var_A[Cstr.Index[0], prClsCount + c]   + 1;
            Inc(c);
          end;
        CSTR_FIX_DIFF_EXC_VAR:
          begin
            if (Cstr.Index[0] <> 0) and (Cstr.Index[1] <> 0) then begin
              Var_B[prClsCount + c]                  := Var_B[prClsCount + c]                + Ct[c];
              Var_A[prClsCount + c, Cstr.Index[1]]   := Var_A[prClsCount + c, Cstr.Index[1]] + 1;
              Var_A[prClsCount + c, Cstr.Index[0]]   := Var_A[prClsCount + c, Cstr.Index[0]] - 1;
              Var_A[Cstr.Index[1], prClsCount + c]   := Var_A[Cstr.Index[1], prClsCount + c] + 1;
              Var_A[Cstr.Index[0], prClsCount + c]   := Var_A[Cstr.Index[0], prClsCount + c] - 1;
            end else if (Cstr.Index[0] = 0)  and (Cstr.Index[1] <> 0) then begin
              Var_B[prClsCount + c]                  := Var_B[prClsCount + c]                + Ct[c];
              Var_A[prClsCount + c, Cstr.Index[1]]   := Var_A[prClsCount + c, Cstr.Index[1]] + 1;
              Var_A[Cstr.Index[1], prClsCount + c]   := Var_A[Cstr.Index[1], prClsCount + c] + 1;
            end else if (Cstr.Index[0] <> 0) and (Cstr.Index[1] =  0) then begin
              Var_B[prClsCount + c]                  := Var_B[prClsCount + c]                - Ct[c];
              Var_A[prClsCount + c, Cstr.Index[0]]   := Var_A[prClsCount + c, Cstr.Index[0]] + 1;
              Var_A[Cstr.Index[0], prClsCount + c]   := Var_A[Cstr.Index[0], prClsCount + c] + 1;
            end;
            Inc(c);
          end;
        CSTR_SCALE_EXC_VAR:
          begin
            Var_A[prClsCount + c, Cstr.Index[1]]   := Var_A[prClsCount + c, Cstr.Index[1]]   + 1;
            Var_A[prClsCount + c, Cstr.Index[0]]   := Var_A[prClsCount + c, Cstr.Index[0]]   - Ct[c];
            Var_A[Cstr.Index[1], prClsCount + c]   := Var_A[Cstr.Index[1], prClsCount + c]   + 1;
            Var_A[Cstr.Index[0], prClsCount + c]   := Var_A[Cstr.Index[0], prClsCount + c]   - Ct[c];
            Inc(c);
          end;
        CSTR_SCALE_DIFF_EXC_VAR:
          begin
            Var_A[prClsCount + c, Cstr.Index[1]]   := Var_A[prClsCount + c, Cstr.Index[1]]   + 1;
            Var_A[prClsCount + c, Cstr.Index[0]]   := Var_A[prClsCount + c, Cstr.Index[0]]   - 1;
            Var_A[prClsCount + c, Cstr.Index[3]]   := Var_A[prClsCount + c, Cstr.Index[3]]   - Ct[c];
            Var_A[prClsCount + c, Cstr.Index[2]]   := Var_A[prClsCount + c, Cstr.Index[2]]   + Ct[c];
            Var_A[Cstr.Index[1], prClsCount + c]   := Var_A[Cstr.Index[1], prClsCount + c]   + 1;
            Var_A[Cstr.Index[0], prClsCount + c]   := Var_A[Cstr.Index[0], prClsCount + c]   - 1;
            Var_A[Cstr.Index[3], prClsCount + c]   := Var_A[Cstr.Index[3], prClsCount + c]   - Ct[c];
            Var_A[Cstr.Index[2], prClsCount + c]   := Var_A[Cstr.Index[2], prClsCount + c]   + Ct[c];
            Inc(c);
          end;

      end;//case

    end;//for i


    //X = Inv(A) * B
    InvMat(PMatrix(Var_A), 0, Var_Size - 1, PMatrix(Var_AI));

    for k := 0 to Var_Size - 1 do begin

      for i := 0 to Var_Size - 1 do begin

        Var_X[k] := Var_X[k] + Var_AI[k, i] * Var_B[i];

      end;


    end;

    Std[0] := Sqrt(Abs(Var_X[0]));
    for i := 1 to prClsCount - 1 do begin

      Std[i] := Sqrt(Abs(Var_X[i] + Var_X[0]));

    end;

  finally

    Free2(TPMChar(Var_A));
    Free2(TPMChar(Var_AI));
    FreeMem(Var_B);
    FreeMem(Var_X);

  end;

end;


function TQUB_Task_IdlBase.prA2Q(AMatrix, AMatrixN :TPMDouble; IsQ :Boolean) :Boolean;
var
  ParamClasses  :TQUB_ExpCond_ParametersList;
  Params        :TQUB_ExpCond_Parameters;
  i             :Integer;
  t             :String;
begin


  Result        := False;

  ParamClasses  := nil;
  //Params        := nil;
  prA2QModel    := nil;

  prpLL         := 1e100;
  prReestA      := AMatrix;


  //Chris: make them user settings !!!
  with prA2QParams do begin

    OptLLConv    := 1e-4;
    OptGradConv  := 1e-4;
    OptMaxIter   := 50;
    OptMaxStep   := 1;

  end;


  try


    //ExpConds
    ParamClasses := TQUB_ExpCond_ParametersList.Create;
    Params       := TQUB_ExpCond_Parameters.CreateCopy(prModel.ModelObject.Parameters);
    ParamClasses.Add(Params);

    prA2QModel  := TQUB_A2Q_Model.Create(prModel.ModelObject, True{DoInitialize}, ParamClasses, @prA2QParams);


    prfParsGrads := DAlloc1(prA2QModel.fParCount);
    for i := 0 to prA2QModel.fParCount - 1 do prfParsGrads[i] := 0;

    prA2QModel.Do_fParsToModel;


    if not prParams.A2QEval then begin
    //Optimize


      try
        if not IsQ then begin
          DFP_MIN(Self, prA2QModel.fParCount, prA2QModel.fPars, prA2QModel.fParsVar,
                  QUB_Task_IDLBase.DFPFunc, QUB_Task_IDLBase.DFPCheck,
                  prA2QParams.OptMaxIter, prA2QParams.OptLLConv, prA2QParams.OptGradConv, prA2QParams.OptMaxStep,
                  prA2QModel.Error);
        end else begin
          DFP_MIN(Self, prA2QModel.fParCount, prA2QModel.fPars, prA2QModel.fParsVar,
                  QUB_Task_IDLBase.DFPFuncQ, QUB_Task_IDLBase.DFPCheck,
                  prA2QParams.OptMaxIter, prA2QParams.OptLLConv, prA2QParams.OptGradConv, prA2QParams.OptMaxStep,
                  prA2QModel.Error);
        end;
      except
        Result := True;
      end;


      if prA2QModel.Error <> 0 then Result := True;

      prA2QModel.Do_fParsToModel;

      prA2QModel.Do_Varf2Varc;


    end else begin
    //Evaluate Log-Likelihood


      prA2QModel.Do_fParsToModel;

      if not IsQ then begin
        prLLGrad(1{CompGrad});
      end else begin
        prLLGradQ(1{CompGrad});
      end;

      prA2QModel.Do_Varf2Varc;

      //prPrintGrads;
      t := t + #13#10#9'Free pars grads: ';
      for i := 0 to prA2QModel.fParCount - 1 do begin
        t := t + Format('%-12.8g ', [prA2QModel.dLLdfPar[i]]);
      end;
      t := t + #13#10#9'Rates grads: ';
      for i := 0 to prA2QModel.RateCount - 1 do begin
        t := t + Format('%-12.8g ', [prA2QModel.dLLdq[0][i]]);
      end;
      prConsoleMsg  := t;
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);


    end;


    prPrintParams;


    if AMatrixN <> nil then begin
      MakeAMatrixPEq(TPMDouble(prA2QModel.Q[0]), AMatrixN, nil{PEq}, nil{Ak}, nil{Lambda}, nil{Tau}, nil{Area}, prSampling, prStateCount);
    end else begin
      MakeAMatrixPEq(TPMDouble(prA2QModel.Q[0]), AMatrix,  nil{PEq}, nil{Ak}, nil{Lambda}, nil{Tau}, nil{Area}, prSampling, prStateCount);
    end;


    Synchronize(prUpdateModel);


  finally


    FreeMem(prfParsGrads);
    ParamClasses.Free;
    prA2QModel.Free;


  end;


end;

function TQUB_Task_IdlBase.prA2Q_SolvOpt(AMatrix :TPMDouble) :Boolean;
var
  ParamClasses  :TQUB_ExpCond_ParametersList;
  sopt          :TPADouble;
  i             :Integer;
  t, w          :String;
begin


  Result        := False;

  ParamClasses  := nil;
  prA2QModel    := nil;
  sopt          := nil;

  prpLL         := 1e100;

  prReestA      := AMatrix;


  //Chris: make them user settings !!!
  with prA2QParams do begin

    OptLLConv    := 1e-3;
    OptGradConv  := 1e-3;
    OptMaxIter   := 100;
    OptMaxStep   := 1;

  end;


  try


    //ExpConds
    ParamClasses := TQUB_ExpCond_ParametersList.Create;
    ParamClasses.Add(prModel.ModelObject.Parameters);


    prA2QModel  := TQUB_A2Q_Model.Create(Model.ModelObject, True{DoInitialize}, ParamClasses, @prA2QParams);


    prfParsGrads := DAlloc1(prA2QModel.fParCount);
    for i := 0 to prA2QModel.fParCount - 1 do prfParsGrads[i] := 0;


    prA2QModel.Do_fParsToModel;


    if not prParams.A2QEval then begin
    //Optimize

      try
        sopt     := DAlloc1(14);
        sopt[0]  := -1.e0;
        sopt[1]  := 1e-6;//varprec
        sopt[2]  := 1e-6;//funprec
        sopt[3]  := 100;//maxiter
        sopt[4]  := 1.e0;
        sopt[5]  := 1.e-8;
        sopt[6]  := 2.5;//Dilation;
        sopt[7]  := 1.e-11;
        sopt[8]  := 0.e0;
        sopt[9]  := 0.e0;
        sopt[10] := 0.e0;
        sopt[11] := 0.e0;
        sopt[12] := 0.e0;

        try
          solvopt(Self, prStopped, prA2QModel.fParCount, prA2QModel.fPars, @QUB_Task_IdlBase.SOLVOPTFunc, @QUB_Task_IdlBase.SOLVOPTGrad, Pointer(sopt), nil, nil, nil, nil);
        except
          Result := True;
          t := 'Error in optimization routine (dll)';
        end;
        if sopt[8] = -1 then begin
          t := 'Allocation error';
          Result := True;
        end else if sopt[8] = -2 then begin
          t := 'Improper space dimension';
          Result := True;
        end else if sopt[8] = -3 then begin
          t := '<fun> returns an improper value';
          Result := True;
        end else if sopt[8] = -4 then begin
          t := '<grad> returns a zero or improper vector at the starting point';
          Result := True;
        end else if sopt[8] = -5 then begin
          t := '<func> returns an improper value';
          Result := True;
        end else if sopt[8] = -6 then begin
          t := '<gradc> returns an improper vector';
          Result := True;
        end else if sopt[8] = -7 then begin
          t := 'Function is unbounded';
          Result := True;
        end else if sopt[8] = -8 then begin
          t := 'Gradient is zero, but stopping criteria are not fulfilled';
          Result := True;
        end else if sopt[8] = -9 then begin
          t := 'Iterations limit exceeded';
          Result := True;
        end else if sopt[8] = -11 then begin
          w := 'Premature stop is possible';
        end else if sopt[8] = -12 then begin
          w := 'Result may not provide the true optimum';
        end else if sopt[8] = -13 then begin
          w := 'Result may be inaccurate in view of a point';
        end else if sopt[8] = -14 then begin
          w := 'Result may be inaccurate in view of a function value';
        end else if sopt[8] > 0 then begin
          prA2QModel.Iterations := Round(sopt[8]);
        end;

      except
        Result := True;
      end;

      if w <> '' then prConsoleMsg  := w;
      if t <> '' then prConsoleMsg  := t;
      if (w <> '') or (t <> '') then begin
        prConsoleCode := MESSAGE_MESSAGE;
        Synchronize(prWriteNewConsole);
      end;

      if prA2QModel.Error <> 0 then Result := True;

      prA2QModel.Do_fParsToModel;

      //prA2QSecDerivative;

      prA2QModel.Do_Varf2Varc;


    end else begin
    //Evaluate Log-Likelihood


      prLLGrad(1{CompGrad});

      prA2QModel.Do_Varf2Varc;

      //prPrintGrads;
      t := t + #13#10#9'Free pars grads: ';
      for i := 0 to prA2QModel.fParCount - 1 do begin
        t := t + Format('%-12.8g ', [prA2QModel.dLLdfPar[i]]);
      end;
      t := t + #13#10#9'Rates grads: ';
      for i := 0 to prA2QModel.RateCount - 1 do begin
        t := t + Format('%-12.8g ', [prA2QModel.dLLdq[0][i]]);
      end;
      prConsoleMsg  := t;
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);


    end;


    prPrintParams;

    MakeAMatrixPEq(TPMDouble(prA2QModel.Q[0]), AMatrix, nil{PEq}, nil{Ak}, nil{Lambda}, nil{Tau}, nil{Area}, prSampling, prStateCount);

    Synchronize(prUpdateModel);


  finally


    FreeMem(prfParsGrads);
    ParamClasses.Free;
    prA2QModel.Free;
    FreeMem(sopt);


  end;


end;


//****************************************************************************//
//                                                                            //
//                       T Q U B _ T a s k _ I d l B a s e                    //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_IdlBase.Create(Suspended :Boolean);
begin

  inherited;

  prName := 'Idealize with Baseline';

  prCanPause := True;
  prCanSkip  := True;

end;


//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlBase.prDoTask;
begin

  inherited;

  Synchronize(prRealityCheck);
  if not prRealityCheckSaysGo then begin
    prStopped := True;
    Exit;
  end;

  if prParams.Staircase then begin
    prIdl_Staircase;
  end else begin
    prIdlBase;
  end;

  prFinishResults;

  // if not prParams.Staircase then prRunStat;

end;


procedure TQUB_Task_IdlBase.prRealityCheck;
var
  i, r         :Integer;
  npoint       :Integer;
  maxnpoint    :Integer;
  Chunk        :TQUB_Thread_PChunk;
begin
  maxnpoint    := 0;
  for i := 0 to prChunks.Count - 1 do begin
    Chunk      := TQUB_Thread_PChunk( prChunks[i] );
    npoint     := Chunk.Sel.Last - Chunk.Sel.First + 1;
    if npoint > maxnpoint then maxnpoint := npoint;
  end;

  if maxnpoint <= IDLBASE_REASONABLE_NPOINTS then begin
    prRealityCheckSaysGo := True;
    Exit;
  end;

  r := MessageHelpDlg( Format('IB: %d data points could take too much time and memory.  Consider chopping the data first.  Continue anyway?', [maxnpoint]),
                   mtConfirmation, [mbYes, mbNo], 0, 'Warning:Too_Much_Time_And_Memory');

  prRealityCheckSaysGo := (r = mrYes);
end;



//****************************************************************************//
//                                                                            //
//                              S T A I R C A S E                             //
//                                                                            //
//****************************************************************************//
{

  One jump or two unequal, sequential jumps

  The repeating unit is one state

  The input model must have two or three states of different class. It uses:
    - the jump rates (forward and backward);
    - the difference in amplitude is assigned to the jump amplitude;
    - the variance of the first state is assigned to the measurement variance.
}
procedure TQUB_Task_IdlBase.prIdl_Staircase;
var
  IdlBaseChunk   :TQUB_IdlBase_PChunk;
  RateForward    :Double;
  RateBackward   :Double;
  Iteration      :Integer;
  Entry          :Integer;
  i, j, c        :Integer;
const
  SMALL_STD      = 1e-20;
begin


  if Model.ModelObject.States.StateCount < 2 then Exit;
  if Model.ModelObject.States[0].Cls = Model.ModelObject.States[1].Cls then Exit;


  prQ            := nil;
  prA            := nil;
  prP0           := nil;


  try


    //The number of repeating units is equal to MaxJump + 1
    prStateCount := Max(6, ((prParams.StaircaseMaxJump div 4) * 2 + 1) * 2);
    //Q Matrix
    prQ          := DAlloc2(prStateCount, prStateCount);
    if prParams.StaircaseKineticsUseInit then begin
    //Get rates from Panel
      RateForward  := prParams.StaircaseKinetics[0];
      RateBackward := prParams.StaircaseKinetics[1];
    end else begin
    //Get Rates from Model
      MakeQMatrix(Model.ModelObject, prQ);
      RateForward  := prQ[0, 1];
      RateBackward := prQ[1, 0];
    end;
    for i := 0 to prStateCount - 1 do begin
      for j := 0 to prStateCount - 1 do begin
        prQ[i, j] := 0;
      end;
    end;
    for i := 0 to prStateCount - 1 do begin
      if i > 0 then begin
      //Can do backward
        prQ[i, i - 1] := RateBackward;
      end;
      if i < (prStateCount - 1) then begin
      //Can do forward
        prQ[i, i + 1] := RateForward;
      end;
      for j := 0 to prStateCount - 1 do begin
        if j <> 0 then prQ[i, i] := prQ[i, i] - prQ[i, j];
      end;
    end;
    //A Matrix
    prA          := DAlloc2(prStateCount, prStateCount);
    MakeAMatrixPEq(prQ, prA, nil, nil, nil, nil, nil, TQUB_QVO_QVD(Data).DataObject.Sampling, prStateCount);
    if prParams.StaircaseRightOnly then begin
      for i := 0 to prStateCount - 1 do begin
        for j := 0 to i - 1 do begin
          prA[i, j] := 0;
        end;
        //for j := i to prStateCount - 1 do begin
        //  prA[i, j] := Abs(prA[i, j]);
        //end;
      end;
    end;
    //P0
    prP0         := DAlloc1(prStateCount);

    //JumpAmpsCount
    if prParams.StaircaseJumpAmpUseInit then begin
    //Get JumpAmpsCount from Panel
      if prParams.StaircaseJumpAmpStepModel = 0 then begin
        prJumpAmpsCount := 1;
      end else begin
        prJumpAmpsCount := 2;
      end;
    end else begin
    //Get JumpAmpsCount from Model
      if (Model.ModelObject.States.StateCount > 2) and
         (Model.ModelObject.States[1].Cls <> Model.ModelObject.States[2].Cls) then begin
      //Two sequential, unequal jumps
        prJumpAmpsCount := 2;
      end else begin
      //One jump
        prJumpAmpsCount := 1;
      end;
    end;

    prJumpAmps   := DAlloc1(prJumpAmpsCount);
    prJumpAmp    := 0;

    //TotStats
    prTotA        := DAlloc2(prStateCount, prStateCount);
    for i := 0 to prStateCount - 1 do begin
      for j := 0 to prStateCount - 1 do begin
        prTotA[i, j] := 0;
      end;
    end;
    prTotJumpAmps := DAlloc1(prJumpAmpsCount);
    for i := 0 to prJumpAmpsCount - 1 do begin
      prTotJumpAmps[i] := 0;
    end;
    prTotJumpAmp  := 0;
    prTotJumpStd  := 0;
    prTotMeasStd  := 0;
    prTotBaseAmp  := 0;
    prTotKf       := 0;
    prTotKb       := 0;
    prTotLL       := 0;
    prTotCount    := 0;

    //JumpAmps
    if prParams.StaircaseJumpAmpUseInit then begin
    //Get JumpAmps from Panel
      for i := 0 to prJumpAmpsCount - 1 do begin
        prJumpAmps[i] := prParams.StaircaseJumpAmps[i];
        prJumpAmp     := prJumpAmp + prJumpAmps[i];
      end;
    end else begin
    //Get JumpAmps from Model
      for i := 0 to prJumpAmpsCount - 1 do begin
        prJumpAmps[i] := Model.ModelObject.Amp[Model.ModelObject.States[i + 1].Cls];// - Model.ModelObject.Amp[Model.ModelObject.States[i].Cls];
        prJumpAmp     := prJumpAmp + prJumpAmps[i];
      end;
    end;

    //MeasStd
    if prParams.StaircaseMeasStdUseInit then begin
    //Get MeasStd from Panel
      prMeasStd    := prParams.StaircaseMeasStd;
    end else begin
    //Get MeasStd from Model
      prMeasStd    := Model.ModelObject.Std[Model.ModelObject.States[0].Cls];
    end;

    //BaseAmp
    if prParams.StaircaseBaseAmpUseInit then begin
    //Get BaseAmp from Panel
      prBaseAmp    := prParams.StaircaseBaseAmp;
    end else begin
    //Get BaseAmp from Model
      prBaseAmp    := Model.ModelObject.Amp[Model.ModelObject.States[0].Cls];
    end;

    Entry        := 0;

    Iteration    := 0;


    prStatus  := 'Running';
    prPercent := 0;
    prETA     := '0:00:00';
    Synchronize(prDoOnChange);


    prConsoleMsg := 'Idealizing data file: ' + TQUB_QVO_QVD(prData).FileName + #13#10;
    prConsoleMsg := prConsoleMsg + 'Parameters:'#13#10;
    with prParams do begin
      prConsoleMsg := prConsoleMsg + Format('Reestimation: %s, MLSS: %s, Output: %s'#13#10 +
                                            'LLConv: %6.4g, MaxIter: %d'#13#10 +
                                            'A2Q: %s, EveryIter: %s, Eval: %s'#13#10 +
                                            'FixK: %s, FixA: %s, FixS: %s'#13#10 +
                                            'Baseline Track: %s, Std: %6.4g, Correct: %s, MinY: %d, BackKalman: %s, ReestPar: %s, FBKCount: %d, FKCount: %d'#13#10 +
                                            'Staircase MaxJump: %d, ROnly: %s, BW: %s, LLAdd: %s, JAmp: %6.4g(%s, w %s), JStd: %6.4g(%s, w %s), BVar: %6.4g, Merge: %6.4g(%s), Split: %6.4g(%s)',
                                            [Choice(Reestimation = IDLBASE_REEST_NONE, 'None', Choice(Reestimation = IDLBASE_REEST_SKM, 'SKM', 'BW')), Choice(MLSS = IDLBASE_MLSS_VITERBI, 'Viterbi', 'Gamma'),
                                             Choice(Output = IDLBASE_OUTPUT_CLASS, 'class', 'group'),
                                             LLConv, MaxIter,
                                             Choice(A2Q, 'Y', 'N'), Choice(A2QEveryIter, 'Y', 'N'), Choice(A2QEval, 'Y', 'N'),
                                             Choice(StaircaseKineticsReest, 'N', 'Y'), Choice(StaircaseJumpAmpReestimate, 'N', 'Y'), Choice(StaircaseMeasStdReest, 'N', 'Y'),
                                             Choice(BaselineTrack, 'Y', 'N'), BaselineStd, Choice(BaselineCorrect, 'Y', 'N'), BaselineMinY, Choice(BaselineBackKalman, 'Y', 'N'), Choice(BaselineReestPar, 'Y', 'N'), BaselineFBKCount, BaselineFKCount,
                                             StaircaseMaxJump, Choice(StaircaseRightOnly, 'Y', 'N'), Choice(StaircaseReestBW, 'Y', 'N'), Choice(StaircaseLLAddJump, 'Y', 'N'),
                                                prJumpAmp, Choice(StaircaseJumpAmpReestMethod = JUMPAMP_CONSECUTIVE, 'Cons', Choice(StaircaseJumpAmpReestMethod = JUMPAMP_COMBINATIONS, 'Comb', 'Constr')), Choice(StaircaseJumpAmpWeighed, 'Y', 'N'), StaircaseJumpStd, Choice(StaircaseJumpStdReest, 'Y', 'N'), Choice(StaircaseJumpStdWeighed, 'Y', 'N'), StaircaseBaseJumpVar,
                                                StaircaseMergeDwellsPercent, Choice(StaircaseMergeDwells, 'Y', 'N'), StaircaseSplitDwellsPercent, Choice(StaircaseSplitDwells, 'Y', 'N')]);
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    Synchronize(prWriteNewConsole);


    {

    Two scenarios:
    1. Each chunk is estimated separately
    2. In each iteration, each Chunk is estimated separately, but parameters are reestimated together
    }


    repeat


      if prSkipped then begin
        prSkipped := False;
        Continue;
      end;

      if not prParams.QuietOutput then begin
        prConsoleMsg  := Format('Idealizing data, iteration %d', [Iteration + 1]);
        prConsoleCode := MESSAGE_MESSAGE;
        Synchronize(prWriteSetConsole);
        Entry         := prConsoleEntry;
      end;

      //For each chunk
      for c := 0 to prChunks.Count - 1 do begin

        if prSkipped then begin
          prSkipped := False;
          Continue;
        end;

        prCurrChunkIndex := c;

        prStatus  := 'Running chunk ' + IntToStr(c);
        try
          prPercent := Round(((Iteration * prChunks.Count) + c) / (prChunks.Count * prParams.MaxIter));
        except
        end;
        Synchronize(prDoOnChange);


        IdlBaseChunk := prChunks[c];

        try


          //Load Chunk Data ......................................................
          DataCallBack(IdlBaseChunk);
          if IdlBaseChunk.Data = nil then Continue;

          FreeMem(IdlBaseChunk.SegFirsts);
          IdlBaseChunk.SegFirsts := nil;
          FreeMem(IdlBaseChunk.SegLasts);
          IdlBaseChunk.SegLasts := nil;
          FreeMem(IdlBaseChunk.SegClasses);
          IdlBaseChunk.SegClasses := nil;

          FreeMem(IdlBaseChunk.Amps);
          IdlBaseChunk.Amps := nil;
          FreeMem(IdlBaseChunk.Stds);
          IdlBaseChunk.Stds := nil;
          Free2(TPMChar(IdlBaseChunk.AMatrix));
          IdlBaseChunk.AMatrix := nil;

          FreeMem(IdlBaseChunk.Nodes);
          IdlBaseChunk.Nodes := nil;
          FreeMem(IdlBaseChunk.NodesDBL);
          IdlBaseChunk.NodesDBL := nil;

          //Re-estimation ........................................................
          if          (prParams.Reestimation = IDLBASE_REEST_BW)  or (prParams.MLSS = IDLBASE_MLSS_GAMMA)   then begin
            if not prParams.QuietOutput then begin
              prConsoleMsg   := Format('Idealizing chunk %d/%d with Staircase Baum-Welch - Kalman algorithm',        [c + 1, prChunks.Count]);
              prConsoleCode  := MESSAGE_MESSAGE;
              prConsoleEntry := Entry;
              Synchronize(prWriteSetConsole);
            end;
            prReest_Staircase_BW(IdlBaseChunk);
          end else if (prParams.Reestimation = IDLBASE_REEST_SKM) or (prParams.MLSS = IDLBASE_MLSS_VITERBI) then begin
            if not prParams.QuietOutput then begin
              prConsoleMsg   := Format('Idealizing chunk %d/%d with Staircase Segmental k-means - Kalman algorithm', [c + 1, prChunks.Count]);
              prConsoleCode  := MESSAGE_MESSAGE;
              prConsoleEntry := Entry;
              Synchronize(prWriteSetConsole);
            end;
            prReest_Staircase_SKM(IdlBaseChunk);
          end;



        finally


          //Free Chunk Data ......................................................
          if IdlBaseChunk.Data <> nil then DataCallBack(IdlBaseChunk);


        end;


        if (prStopped) and not (prParams.StaircaseGlobalReest) then Break;


      end;


      //Print TotStats
      if prTotCount > 0 then begin

        prConsoleMsg := '';
        if not prParams.QuietOutput then begin
          prConsoleMsg := #13#10'Overall jump parameters:'#13#10;
          prConsoleMsg := prConsoleMsg + 'iter';
          for i := 0 to prJumpAmpsCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'JumpAmp[%d]    ', [i]);
          end;
          prConsoleMsg := prConsoleMsg + #9'JumpStd       '#9'MeasStd       '#9'BaseAmp       '#9'LL            '#13#10;
        end;
        prConsoleMsg := prConsoleMsg + Format('%d', [Iteration + 1]);
        for i := 0 to prJumpAmpsCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [prTotJumpAmps[i]/prTotCount]);
        end;
        prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.10g', [prTotJumpStd/prTotCount, prTotMeasStd/prTotCount, prTotBaseAmp/prTotCount, prTotKf/prTotCount, prTotKb/prTotCount, prTotLL]);
        prConsoleCode := MESSAGE_MESSAGE;
        Synchronize(prWriteNewConsole);
      end;

      if prParams.StaircaseGlobalReest then begin

        for i := 0 to prStateCount - 1 do begin
          for j := 0 to prStateCount - 1 do begin
            prA[i, j] := prTotA[i, j]/prTotCount;
          end;
        end;
        prJumpAmp       := 0;
        for i := 0 to prJumpAmpsCount - 1 do begin
          prJumpAmps[i] := prTotJumpAmps[i]/prTotCount;
          prJumpAmp     := prJumpAmp + prJumpAmps[i];
        end;
        if prParams.StaircaseJumpStdReest then begin
          prParams.StaircaseJumpStd := prTotJumpStd/prTotCount;
        end;
        prMeasStd       := prTotMeasStd/prTotCount;
        prBaseAmp       := prTotBaseAmp/prTotCount;


        //Print Q matrix as the logarithm of the A matrix ......................
        //prTotA[0, 0] := prA[0, 0];
        //prTotA[0, 1] := 1 - prA[0, 0];
        //prTotA[1, 1] := prA[prStateCount - 1, prStateCount - 1];
        //prTotA[1, 0] := 1 - prA[prStateCount - 1, prStateCount - 1];

        //MatrixLogarithm(prTotA, prTotA, IDLBaseChunk.Sampling, 2, 100, 1e-4);
        //for s := 0 to {prStateCount -} 1 do begin
        //  for ss := 0 to {prStateCount -} 1 do begin
        //    if prTotA[s, ss] < 0 then begin
        //      prTotA[s, ss] := 0;
        //      prTotA[ss, s] := 0;
        //    end;
        //  end;
        //end;
        //prConsoleMsg := #13#10'Small (LeftTop-RightBottom) global Q Matrix:'#13#10;
        //for s := 0 to {prStateCount -} 1 do begin
        //  for ss := 0 to {prStateCount -} 1 do begin
        //    if (s = ss) or (prQ[s, ss] = 0)then begin
        //      prConsoleMsg := prConsoleMsg + #9'...           ';
        //    end else begin
        //      prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [prTotA[s, ss]]);
        //    end;            end;
        //  prConsoleMsg := prConsoleMsg + #13#10;
        //end;
        //prConsoleCode := MESSAGE_MESSAGE;
        //Synchronize(prWriteNewConsole);


        //Clear Totals
        for i := 0 to prStateCount - 1 do begin
          for j := 0 to prStateCount - 1 do begin
            prTotA[i, j] := 0;
          end;
        end;
        for i := 0 to prJumpAmpsCount - 1 do begin
          prTotJumpAmps[i] := 0;
        end;
        prTotJumpAmp  := 0;
        prTotJumpStd  := 0;
        prTotMeasStd  := 0;
        prTotBaseAmp  := 0;
        prTotKf       := 0;
        prTotKb       := 0;
        prTotLL       := 0;
        prTotCount    := 0;


      end else begin

        Break;

      end;

      Inc(Iteration);

    until(prStopped) or (Iteration >= prParams.MaxIter) or (not prParams.StaircaseGlobalReest);


    //Print all dAmps to do a histogram


  finally


    Free2(TPMChar(prQ));
    prQ := nil;
    Free2(TPMChar(prA));
    prA := nil;
    FreeMem(prP0);
    prP0 := nil;
    FreeMem(prJumpAmps);
    FreeMem(prTotJumpAmps);
    Free2(TPMChar(prTotA));


  end;

end;




{
  For now, the segments are estimated individually.
  Later, add Together option
}
procedure TQUB_Task_IdlBase.prIdlBase;
var
  IdlBaseChunk   :TQUB_IdlBase_PChunk;
  Found          :Boolean;
  c, s           :Integer;
const
  SMALL_STD      = 1e-20;
begin


  prQ            := nil;
  prA            := nil;
  prP0           := nil;
  prClsIdx       := nil;
  prClsIdxR      := nil;
  prClsOfState   := nil;
  prGrpOfState   := nil;
  prAmp          := nil;
  prStd          := nil;


  try


    prStateCount := Model.ModelObject.States.StateCount;
    prQ          := DAlloc2(prStateCount, prStateCount);
    MakeQMatrix(Model.ModelObject, prQ);
    prA          := DAlloc2(prStateCount, prStateCount);
    MakeAMatrixPEq(prQ, prA, nil, nil, nil, nil, nil, TQUB_QVO_QVD(Data).DataObject.Sampling, prStateCount);
    prP0         := DAlloc1(prStateCount);
    MakeInitProbs(Model.ModelObject, prP0, prStateCount);

    prClsIdx     := IAlloc1(prStateCount);
    prClsIdxR    := IAlloc1(prStateCount);
    prClsOfState := IAlloc1(prStateCount);
    prGrpOfState := IAlloc1(prStateCount);
    prClsCount   := 0;
    for s := 0 to prStateCount - 1 do begin
      prClsOfState[s] := Model.ModelObject.States[s].Cls;
      prGrpOfState[s] := Model.ModelObject.States[s].Gr;
    end;
    for s := 0 to prStateCount - 1 do begin
      Found := False;
      for c := 0 to prClsCount - 1 do begin
        if prClsIdx[c] = prClsOfState[s] then begin
          prClsIdxR[s] := c;
          Found := True;
          Break;
        end;
      end;
      if not Found then begin
        prClsIdx[prClsCount] := prClsOfState[s];
        prClsIdxR[s]         := prClsCount;
        Inc(prClsCount);
      end;
    end;
    prAmp        := DAlloc1(prClsCount);
    prStd        := DAlloc1(prClsCount);
    for c := 0 to prClsCount - 1 do begin
      prAmp[c] := Model.ModelObject.Amp[prClsIdx[c]];
      prStd[c] := Model.ModelObject.Std[prClsIdx[c]];
      if prStd[c] < SMALL_STD then prStd[c] := SMALL_STD;
    end;



    //For each chunk
    for c := 0 to prChunks.Count - 1 do begin

      prCurrChunkIndex := c;

      IdlBaseChunk := prChunks[c];
      prSampling   := IDLBaseChunk.Sampling;


      try


        //Load Chunk Data ......................................................
        DataCallBack(IdlBaseChunk);
        if IdlBaseChunk.Data = nil then Continue;


        prConsoleMsg  := Format(#13#10'Idealizing chunk %d/%d'#13#10, [c + 1, prChunks.Count]);
        prConsoleCode := MESSAGE_MESSAGE;
        Synchronize(prWriteNewConsole);

        (*
        //Re-estimation ........................................................
        if          (prParams.Reestimation = IDLBASE_REEST_SKM) or (prParams.MLSS = IDLBASE_MLSS_VITERBI) then begin
        //Segmental k-means / Viterbi
          prReestSKM(IdlBaseChunk);
        end else if (prParams.Reestimation = IDLBASE_REEST_BW)  or (prParams.MLSS = IDLBASE_MLSS_GAMMA)   then begin
        //Baum-Welch / Gamma
          //if prParams.BaselineBackKalman then begin
          //  prReestBWKII(IdlBaseChunk);
          //end else begin
            prReestBW(IdlBaseChunk);
          //end;
        end else if prParams.MLSS = IDLBASE_MLSS_FORWARDVITERBI then begin
        //Forward-Viterbi (no re-estimation yet)
          prForwardViterbi(IdlBaseChunk);
        end;
        *)

        if          (prParams.Reestimation = IDLBASE_REEST_SKM)then begin
        //Segmental k-means / Viterbi
          prReestSKM(IdlBaseChunk);
        end else if (prParams.Reestimation = IDLBASE_REEST_BW) then begin
        //Baum-Welch / Gamma
          prReestBW(IdlBaseChunk);
          //Viterbi
          if (prParams.MLSS = IDLBASE_MLSS_VITERBI) then begin
            //prViterbi;
          end;
        end else if prParams.MLSS = IDLBASE_MLSS_FORWARDVITERBI then begin
        //Forward-Viterbi (no re-estimation yet)
          prForwardViterbi(IdlBaseChunk);
        end;


      finally


        //Free Chunk Data ......................................................
        if IdlBaseChunk.Data <> nil then DataCallBack(IdlBaseChunk);


      end;

      if prStopped then Break;

    end;//Chunk


  finally


    Free2(TPMChar(prQ));
    prQ := nil;
    Free2(TPMChar(prA));
    prA := nil;
    FreeMem(prP0);
    prP0 := nil;
    FreeMem(prClsIdx);
    prClsIdx := nil;
    FreeMem(prClsIdxR);
    prClsIdxR := nil;
    FreeMem(prClsOfState);
    prClsOfState := nil;
    FreeMem(prGrpOfState);
    prGrpOfState := nil;
    FreeMem(prAmp);
    prAmp := nil;
    FreeMem(prStd);
    prStd := nil;


  end;

end;

procedure TQUB_Task_IdlBase.prApplyDeadTimeSamples(IdlBaseChunk :TQUB_IdlBase_PChunk; DeadCount :Integer);
var
  First :LongInt;
  Last  :LongInt;
  Count :LongInt;
  i, j  :Integer;
begin

  if DeadCount = 0 then Exit;//Find the first long enough dwell:
  First := -1;
  for i := 0 to IdlBaseChunk.SegCount - 1 do begin
    Count := IdlBaseChunk.SegLasts[i] - IdlBaseChunk.SegFirsts[i] + 1;
    if Count >= DeadCount then begin
      IdlBaseChunk.SegFirsts[i] := IdlBaseChunk.SegFirsts[0];
      First := i;
      Break;
    end else begin
      IdlBaseChunk.SegLasts[i] := IdlBaseChunk.SegFirsts[i] - 1;
    end;
  end;
  if First = -1 then begin
    IdlBaseChunk.SegCount := 0;
    Exit;
  end;

  //Find all the too-short segments and mark them; also expand the long enough ones
  Last := First;
  for i := First + 1 to IdlBaseChunk.SegCount - 1 do begin
    Count := IdlBaseChunk.SegLasts[i] - IdlBaseChunk.SegFirsts[i] + 1;
    if Count >= DeadCount then begin
      Last := i;
    end else begin
      IdlBaseChunk.SegLasts[Last] := IdlBaseChunk.SegLasts[i];
      IdlBaseChunk.SegLasts[i] := IdlBaseChunk.SegFirsts[i] - 1;
    end;
  end;

  //Collapse the list and set the new count;
  j := 0;
  for i := First to Last do begin
    if IdlBaseChunk.SegLasts[i] > IdlBaseChunk.SegFirsts[i] then begin
      IdlBaseChunk.SegFirsts[j]  := IdlBaseChunk.SegFirsts[i];
      IdlBaseChunk.SegLasts[j]   := IdlBaseChunk.SegLasts[i];
      IdlBaseChunk.SegClasses[j] := IdlBaseChunk.SegClasses[i];
      Inc(j);
    end;
  end;
  IdlBaseChunk.SegCount := Last - First + 1;
end;

procedure  TQUB_Task_IdlBase.prResultsToData;
var
  IdlBaseChunk      :TQUB_IdlBase_PChunk;
  Chunk             :TQUB_Thread_PChunk;
  Sel               :TQUB_FirstLast;
  c, p              :Integer;

  //ViewportHistogram :TQUB_Viewport_Histogram;
  //Seg               :INode;
  //HistGroupNode     :INode;
  //Group             :TQUB_Histogram_SimulGroup;
  //j                 :Integer;

  //Seg               :INode;

begin


  //Clear Info .................................................................
  if prParams.ReplaceIDL then begin
    prStatus := 'Clearing existing Idealization';
    Synchronize(prDoOnChange);
    TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].Clear;
  end;

  //Append Info ................................................................
  prStatus := 'Applying Idealization and Baseline';
  Synchronize(prDoOnChange);


  //First pass &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
  for c := prChunks.Count - 1 downto 0 do begin

    prCurrChunkIndex := c;

    IdlBaseChunk := prChunks[c];
    Chunk        := prChunks[c];

    if Chunk.Status = CHUNK_STATUS_DONE then begin


      //Idealization .........................................................

      //Apply DeadTime?
      if (IdlBaseChunk.SegCount > 0) and (prParams.ApplyDeadTime) then begin
        prApplyDeadTimeSamples(IdlBaseChunk, Round(prParams.DeadTime*1e-3 / TQUB_QVO_QVD(prData).Sampling));
      end;

      //Drop First?
      if IdlBaseChunk.SegCount > 0 then begin
        if (prParams.DropFirstLastIfClosed and (IdlBaseChunk.SegClasses[0] =  0)) or
           (prParams.DropFirstLastIfOpen   and (IdlBaseChunk.SegClasses[0] <> 0)) then begin
          for p := 0 to IdlBaseChunk.SegCount - 2 do begin
            IdlBaseChunk.SegFirsts[p]  := IdlBaseChunk.SegFirsts[p + 1];
            IdlBaseChunk.SegLasts[p]   := IdlBaseChunk.SegLasts[p + 1];
            IdlBaseChunk.SegClasses[p] := IdlBaseChunk.SegClasses[p + 1];
          end;
          IdlBaseChunk.SegCount := IdlBaseChunk.SegCount - 1;
        end;
      end;
      //Drop Last?
      if IdlBaseChunk.SegCount > 0 then begin
        if (prParams.DropFirstLastIfClosed and (IdlBaseChunk.SegClasses[IdlBaseChunk.SegCount - 1] =  0)) or
           (prParams.DropFirstLastIfOpen   and (IdlBaseChunk.SegClasses[IdlBaseChunk.SegCount - 1] <> 0)) then begin
          IdlBaseChunk.SegCount := IdlBaseChunk.SegCount - 1;
        end;
      end;

      if IdlBaseChunk.SegCount > 0 then begin
        Sel.First := IdlBaseChunk.SegFirsts[0];
        Sel.Last  := IdlBaseChunk.SegLasts[IdlBaseChunk.SegCount - 1];
        Chunk.StartTime := Chunk.StartTime + IdlBaseChunk.Sampling * (Sel.First - Chunk.Sel.First);
        Chunk.Sel := Sel;
      end;//SegCount > 0

      //Baseline Correction ..................................................
      if IdlBaseChunk.NodeCount > 0 then begin

        //Account here for drop first/last
        for p := 0 to IdlBaseChunk.NodeCount - 1 do begin
          Inc(IdlBaseChunk.Nodes[p], Chunk.Sel.First);
        end;

      end;//NodeCount > 0

    end;//Done

  end;//c



  //Second pass &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
  for c := 0 to prChunks.Count - 1 do begin

    prCurrChunkIndex := c;

    IdlBaseChunk := prChunks[c];
    Chunk        := prChunks[c];


    try


      if Chunk.Status = CHUNK_STATUS_DONE then begin


        //Idealization .........................................................
        if IdlBaseChunk.SegCount > 0 then begin

          Sel.First := IdlBaseChunk.SegFirsts[0];
          Sel.Last  := IdlBaseChunk.SegLasts[IdlBaseChunk.SegCount - 1];

          TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].SetDwellsSeg(prBuildIdlSegNode(IdlBaseChunk));

        end;//SegCount > 0


        //Baseline Correction ..................................................
        if IdlBaseChunk.NodeCount > 0 then begin

          //Append
          TQUB_QVO_QVD(Data).AddBaseLineNodes(ChannelIndex, IdlBaseChunk.NodeCount, IdlBaseChunk.Nodes, IdlBaseChunk.NodesDBL);

        end;//NodeCount > 0


      end;//Done


    finally


      IdlBaseChunk.SegCount    := 0;
      FreeMem(IdlBaseChunk.SegFirsts);        IdlBaseChunk.SegFirsts   := nil;
      FreeMem(IdlBaseChunk.SegLasts);         IdlBaseChunk.SegLasts    := nil;
      FreeMem(IdlBaseChunk.SegClasses);       IdlBaseChunk.SegClasses  := nil;

      IdlBaseChunk.NodeCount   := 0;
      IdlBaseChunk.Nodes       := nil;
      IdlBaseChunk.NodesDBL    := nil;

      IdlBaseChunk.ClassCount  := 0;
      FreeMem(IdlBaseChunk.Amps);             IdlBaseChunk.Amps        := nil;
      FreeMem(IdlBaseChunk.Stds);             IdlBaseChunk.Stds        := nil;
      Free2(TPMChar(IdlBaseChunk.AMatrix));   IdlBaseChunk.AMatrix     := nil;


    end;


  end;//c


  //This below does not remove FirstLast, as far as I could see
  //TQUB_QVO_QVD(Data).DataObject.Idealization[ChannelIndex].SetDwells(prDataSet);


  try

    if prSelectionList <> nil then begin          // if the selection list is 'IdlBase Results',
      prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
      prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
    end;

    prResultNode['LL'].SetData(prTotLL);

    prResultNode.Name   := 'IdlBase ' + ChangeFileExt(ExtractFileName(TQUB_QVO_QVD(Data).FileName), '');
    ViewportResults.AddResults(prResultNode, False);

  except
  end;


  inherited;


end;

procedure TQUB_Task_IdlBase.prPublishResultsPy;
begin
  PyNode_Publish( prResultNode, 'ibResult' );
end;

procedure TQUB_Task_IdlBase.prMakeDwells(IdlBaseChunk :TQUB_IdlBase_PChunk; State :TPAInteger; DataCount :Integer);
var
  DwellMaxCount       :Integer;
  DwellFirst          :Integer;
  CurrCls, ThisCls    :Integer;
  p                   :Integer;
begin

  IdlBaseChunk.SegCount := 0;
  DwellMaxCount       := 0;
  p                   := 0;
  if prParams.Staircase then begin
    CurrCls           := State[p];
  end else begin
    if prParams.Output = IDLBASE_OUTPUT_GROUP then begin
      CurrCls         := prGrpOfState[State[p]];
    end else begin
      CurrCls         := prClsIdxR[State[p]];
    end;
  end;
  DwellFirst          := p;
  p                   := 1;
  repeat

    if prParams.Staircase then begin
      ThisCls         := State[p];
    end else begin
      if prParams.Output = IDLBASE_OUTPUT_GROUP then begin
        ThisCls         := prGrpOfState[State[p]];
      end else begin
        ThisCls         := prClsIdxR[State[p]];
      end;
    end;

    if (ThisCls <> CurrCls) or (p = (DataCount - 1)) then begin

      Inc(IdlBaseChunk.SegCount);
      if IdlBaseChunk.SegCount > DwellMaxCount then begin
        Inc(DwellMaxCount, 10000);
        try
          ReallocMem(IdlBaseChunk.SegFirsts,    SizeOf(Integer) * DwellMaxCount);
          ReallocMem(IdlBaseChunk.SegLasts,     SizeOf(Integer) * DwellMaxCount);
          ReallocMem(IdlBaseChunk.SegClasses,   SizeOf(Integer) * DwellMaxCount);
        except
          Dec(IdlBaseChunk.SegCount);
          Break;
        end;
      end;
      IdlBaseChunk.SegFirsts  [IdlBaseChunk.SegCount - 1] := DwellFirst;
      DwellFirst := p;
      if ThisCls = CurrCls then begin
        IdlBaseChunk.SegLasts [IdlBaseChunk.SegCount - 1] := p;
      end else begin
        IdlBaseChunk.SegLasts [IdlBaseChunk.SegCount - 1] := p - 1;
        if p = (DataCount - 1) then begin
          DwellFirst := DataCount;
        end;
      end;
      IdlBaseChunk.SegClasses [IdlBaseChunk.SegCount - 1] := CurrCls;

    end;

    CurrCls    := ThisCls;

    Inc(p);

  until (p >= DataCount);

  if DwellFirst = DataCount then begin
  //Last Point, if unclosed
    Inc(IdlBaseChunk.SegCount);
    if IdlBaseChunk.SegCount > DwellMaxCount then begin
      Inc(DwellMaxCount, 10000);
      try
        ReallocMem(IdlBaseChunk.SegFirsts,    SizeOf(Integer) * DwellMaxCount);
        ReallocMem(IdlBaseChunk.SegLasts,     SizeOf(Integer) * DwellMaxCount);
        ReallocMem(IdlBaseChunk.SegClasses,   SizeOf(Integer) * DwellMaxCount);
      except
        Dec(IdlBaseChunk.SegCount);
      end;
    end;
    IdlBaseChunk.SegFirsts [IdlBaseChunk.SegCount - 1] := DwellFirst - 1;
    IdlBaseChunk.SegLasts  [IdlBaseChunk.SegCount - 1] := DwellFirst - 1;
    IdlBaseChunk.SegClasses[IdlBaseChunk.SegCount - 1] := CurrCls;

  end;

  if IdlBaseChunk.SegCount > 0 then begin

    if IdlBaseChunk.SegCount <> DwellMaxCount then begin
      ReallocMem(IdlBaseChunk.SegFirsts,    SizeOf(Integer) * IdlBaseChunk.SegCount);
      ReallocMem(IdlBaseChunk.SegLasts,     SizeOf(Integer) * IdlBaseChunk.SegCount);
      ReallocMem(IdlBaseChunk.SegClasses,   SizeOf(Integer) * IdlBaseChunk.SegCount);
    end;

  end;

end;


function TQUB_Task_IdlBase.prBase2Nodes(IdlBaseChunk :TQUB_IdlBase_PChunk; BaseX :TPADouble) :Boolean;
var
  Chunk               :TQUB_Thread_PChunk;

  DataCount           :Integer;
  Scale               :Double;
  Off                 :Double;

  NodeMaxCount        :Integer;

  PrevP               :Integer;
  PrevY               :Integer;
  ThisY               :Integer;

  p                   :Integer;
begin

  Result              := False;

  try

    Chunk               := TQUB_Thread_PChunk(IdlBaseChunk);

    DataCount           := Chunk.Sel.Last - Chunk.Sel.First + 1;
    Scale               := IdlBaseChunk.Scaling;
    Off                 := IdlBaseChunk.Offset;

    p                   := 0;

    Inc(IdlBaseChunk.NodeCount);
    NodeMaxCount        := 10000;
    ReallocMem(IdlBaseChunk.Nodes,    SizeOf(Integer) * NodeMaxCount);
    ReallocMem(IdlBaseChunk.NodesDBl, SizeOf(Integer) * NodeMaxCount);

    PrevY               := Round((BaseX[p]-Off) * Scale);
    PrevP               := p;
    IdlBaseChunk.Nodes   [IdlBaseChunk.NodeCount - 1] := p;
    IdlBaseChunk.NodesDBl[IdlBaseChunk.NodeCount - 1] := -PrevY;

    for p := 1 to DataCount - 1 do begin

      ThisY             := Round((BaseX[p]-Off) * Scale);

      if Abs(ThisY - PrevY) >= prParams.BaselineMinY then begin
      //Add Node

        if p > PrevP then begin
          Inc(IdlBaseChunk.NodeCount);
          if IdlBaseChunk.NodeCount > NodeMaxCount then begin
            Inc(NodeMaxCount, 10000);
            try
              ReallocMem(IdlBaseChunk.Nodes,    SizeOf(Integer) * NodeMaxCount);
              ReallocMem(IdlBaseChunk.NodesDBl, SizeOf(Integer) * NodeMaxCount);
            except
              Dec(IdlBaseChunk.NodeCount);
              Break;
            end;
          end;
          IdlBaseChunk.Nodes   [IdlBaseChunk.NodeCount - 1] := p - 1;
          IdlBaseChunk.NodesDBl[IdlBaseChunk.NodeCount - 1] := -PrevY;
        end;

        Inc(IdlBaseChunk.NodeCount);
        if IdlBaseChunk.NodeCount > NodeMaxCount then begin
          Inc(NodeMaxCount, 10000);
          try
            ReallocMem(IdlBaseChunk.Nodes,    SizeOf(Integer) * NodeMaxCount);
            ReallocMem(IdlBaseChunk.NodesDBl, SizeOf(Integer) * NodeMaxCount);
          except
            Dec(IdlBaseChunk.NodeCount);
            Break;
          end;
        end;
        IdlBaseChunk.Nodes   [IdlBaseChunk.NodeCount - 1] := p;
        IdlBaseChunk.NodesDBl[IdlBaseChunk.NodeCount - 1] := -ThisY;

        PrevP := p;
        PrevY := ThisY;

      end;//D >MinY

    end;//p

    if IdlBaseChunk.NodeCount > 0 then begin
      p := DataCount - 1;
      if IdlBaseChunk.Nodes[IdlBaseChunk.NodeCount - 1] < p then begin
        Inc(IdlBaseChunk.NodeCount);
        if IdlBaseChunk.NodeCount > NodeMaxCount then begin
          Inc(NodeMaxCount, 10000);
          try
            ReallocMem(IdlBaseChunk.Nodes,    SizeOf(Integer) * NodeMaxCount);
            ReallocMem(IdlBaseChunk.NodesDBl, SizeOf(Integer) * NodeMaxCount);
          except
            Dec(IdlBaseChunk.NodeCount);
          end;
        end;
        IdlBaseChunk.Nodes   [IdlBaseChunk.NodeCount - 1] := p;
        IdlBaseChunk.NodesDBl[IdlBaseChunk.NodeCount - 1] := -PrevY;
      end;
    end;

  except

    Result := True;

  end;

end;



// *****************************************************************************

function TQUB_Task_IdlBase.prMakeStats(IdlBaseChunk :TQUB_IdlBase_PChunk; BaseX :TPADouble) :Boolean;
var
  IdlSeg      :INode;
begin

  prStatChunk := IdlBaseChunk;
  prStatBase  := BaseX;

  // (we know the data is loaded, and Base is size dataCount or nil)

  // TODO: it appears dead time is applied after stats...unlike SKM...bad?

  IdlSeg      := prBuildIdlSegNode(IdlBaseChunk);
  IdlBaseChunk.StatResult := prRunStat(IdlSeg); // this and below prevent idealizing all chunks
  prDataSet.AppendChild(IdlSeg);
  IdlBaseChunk.StatResult.Find('HistogramGroup').Find('Title').SetData( Format('Segment %d', [prCurrChunkIndex+1]) );

  Result := False;

end;

function TQUB_Task_IdlBase.prBuildIdlSegNode(IdlBaseChunk :TQUB_IdlBase_PChunk): INode;
var
  p, cls                     :Integer;
  Chunk                      :TQUB_Thread_PChunk;
  Sel                        :TQUB_FirstLast;
  idwt, tdwt, firsts, lasts  :INode;
begin

  Result                   := TNode.CreateNull;
  Chunk                    := TQUB_Thread_PChunk(IdlBaseChunk);

  if IdlBaseChunk.SegCount > 0 then begin

    // set chunk bounds and start time
    Sel.First := IdlBaseChunk.SegFirsts[0];
    Sel.Last  := IdlBaseChunk.SegLasts[IdlBaseChunk.SegCount - 1];
    Chunk.StartTime := Chunk.StartTime + IdlBaseChunk.Sampling * Sel.First;

    // build idl seg node for stat
    Result              := TNode.Create('Segment');
    Result.SetupNumData(NODE_DATATYPE_LONGINT, 2, 1);
    Result.Data[0]      := Chunk.Sel.First;
    Result.Data[1]      := Chunk.Sel.Last;

    Result['start'].SetData(Chunk.StartTime * 1.0e3);
    Result['start'].SetLineComment('ms');
    Result['length'].SetData((Sel.Last - Sel.First + 1) * IdlBaseChunk.Sampling * 1.0e3);
    Result['length'].SetLineComment('ms');

    Result['amp'].SetupNumData(NODE_DATATYPE_DOUBLE, IdlBaseChunk.ClassCount, 1);
    Result['sd'].SetupNumData(NODE_DATATYPE_DOUBLE, IdlBaseChunk.ClassCount, 1);
    for cls := 0 to IdlBaseChunk.ClassCount - 1 do begin
      Result['amp'].Data[cls] := IdlBaseChunk.Amps[cls];
      Result['sd'].Data[cls]  := IdlBaseChunk.Stds[cls];
    end;

    Result['LL'].SetData(IdlBaseChunk.LL);

    Result['DwellCount'].SetData(IdlBaseChunk.SegCount);
    idwt                   := Result['Classes'];
    idwt.SetupNumData(NODE_DATATYPE_LONGINT, IdlBaseChunk.SegCount, 1);
    tdwt                   := Result['Durations'];
    tdwt.SetupNumData(NODE_DATATYPE_SINGLE,  IdlBaseChunk.SegCount, 1);
    firsts                 := Result['Firsts'];
    firsts.SetupNumData(NODE_DATATYPE_LONGINT, IdlBaseChunk.SegCount, 1);
    lasts                  := Result['Lasts'];
    lasts.SetupNumData(NODE_DATATYPE_LONGINT, IdlBaseChunk.SegCount, 1);
    for p := 0 to IdlBaseChunk.SegCount - 1 do begin
      idwt.Data[p]         := IdlBaseChunk.SegClasses[p];
      tdwt.Data[p]         := (IdlBaseChunk.SegLasts[p] - IdlBaseChunk.SegFirsts[p] + 1) * IdlBaseChunk.Sampling * 1.0e3;
      firsts.Data[p]       := IdlBaseChunk.SegFirsts[p];
      lasts.Data[p]        := IdlBaseChunk.SegLasts[p];
    end;

  end;//SegCount > 0
end;

function TQUB_Task_IdlBase.prRunStat(SegNode :INode): INode;   // returns 'Segment'
var
  StatResult               :INode;
  DataSetNode              :INode;
  Config                   :INode;
begin

  Result := SegNode;
  if SegNode.IsNull then Exit;

  if prParams.Staircase then begin
    Exit;
  end;

  DataSetNode            := TQUB_QVO_QVD(Data).CreateDataSet;
  DataSetNode.AppendChild( SegNode );

  Config                 := TNode.Create('Properties');
  Config['ReestimateAmpSD'].SetData( 0 );
  Config['Histograms']['BinCount'].SetData( 64 ); // get this from the properties

  StatResult             := StatDataSet(DataSetNode, Config, prFillDataBaseline, DataNoder.NullCallback );
  Result               := StatResult['DataSet']['Segment'];

end;

function TQUB_Task_IdlBase.prFillDataBaseline(Seg :INode): INode;
var
  ChanNode            :INode;
  ChanData            :TPADouble;
  SegOff              :Integer;
  Scale               :Double;
  Off                 :Double;
  p                   :Integer;
begin
  if Seg.Parent['ActiveChannel'].Data[0] = ChannelIndex then begin
    ChanNode            := Seg['Channel'];
    ChanNode.SetupNumData( NODE_DATATYPE_DOUBLE, Seg.Data[1] - Seg.Data[0] + 1, 1 );
    ChanData            := TPADouble( ChanNode.DataPtr );

    SegOff              := Seg.Data[0] - prStatChunk.Chunk.Sel.First;
    Scale               := 1 / prStatChunk.Scaling;
    Off                 := prStatChunk.Offset;

    for p := 0 to ChanNode.DataCount - 1 do begin
      ChanData[p]       := prStatChunk.Data[p + SegOff] * Scale + Off
    end;

    if prStatBase <> nil then begin
      for p := 0 to ChanNode.DataCount - 1 do begin
        ChanData[p]     := ChanData[p] - prStatBase[p + SegOff];
      end;
    end;

    Result              := Seg;
  end else begin
    if prScripting then begin
      Result            := DataNoderScript.FillData(Seg);
    end else begin
      Result            := DataNoder.FillData(Seg);
    end;
  end;
end;


//****************************************************************************//
//                                                                            //
//                                  D R I F T                                 //
//                                                                            //
//                                                                            //
//                        V I T E R B I  -  K A L M A N                       //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlBase.prReest_Staircase_SKM(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
var
  Chunk               :TQUB_Thread_PChunk;

  Data                :TPAInteger;
  DataCount           :Integer;
  Scale               :Double;
  Off                 :Double;

  HalfStateCount      :Integer;

  Iteration           :Integer;

  //*********************************
  VProb               :TPMDouble;
  VBack               :TPMInteger;

  Alpha               :TPMDouble;
  Betta               :TPMDouble;
  Gamma               :TPMDouble;
  Gauss               :TPMDouble;
  Rho                 :TPADouble;
  Epsilon             :TPAPointer;

  State               :TPAInteger;
  StateOffset         :TPAInteger;

  AMatrix             :TPMDouble;
  LogAMatrix          :TPMDouble;

  Prob0               :TPADouble;

  Amps                :TPADouble;
  Gamm                :TPADouble;

  BaseX               :TPMDouble;
  BaseP               :TPMDouble;

  SBaseX              :TPADouble;
  SBaseP              :TPADouble;
  SBaseQ              :TPADouble;

  CBaseX              :TPADouble;
  CBaseP              :TPADouble;

  //*********************************

  Sum                 :Double;

  pm, pc              :Double;

  y                   :Double;

  c, p, pp, s, ss, im :Integer;
  j                   :Integer;

  OldLL, NewLL        :Double;

  JumpAmp             :Double;
  JumpAmps            :TPADouble;
  JumpStd             :Double;
  JumpVar             :Double;

  OldJumpAmp          :Double;
  OldJumpStd          :Double;

  MeasStd             :Double;
  MeasVar             :Double;

  BaseAmp             :Double;
  BaseVar             :Double;
  BaseStd             :Double;

  RefAmp              :Double;
  RefState            :Integer;

  ThisAmp             :Double;
  ThisStd             :Double;

  MinState            :Integer;
  MaxState            :Integer;

  Min1, Max1          :Integer;

  lState              :Integer;
  rState              :Integer;

  Middle              :Integer;

  UsedCount           :Integer;

  CanQuit             :Boolean;
  BaseEstCount        :Integer;

  Offset              :Integer;

  oJ, cJ              :Double;

  yl, hl, kl          :Double;

  k, kxp              :Double;

  LevelExists         :TPAInteger;
  LevelCount          :Integer;

  dl                  :Integer;
  ThisJump            :Double;

  
  procedure MakeAStates(var lState, rState :Integer; l, r :Integer; p :Integer);
  var
    dState :Integer;
  begin
    dState := (r + StateOffset[p]) - (l + StateOffset[p - 1]);
    if dState >= 0 then begin
      lState := 0;
      rState := lState + dState;
    end else begin
      rState := 0;
      lState := rState - dState;
    end;
  end;

  procedure ReestJumpAmpStdCombinations;
  var
    l, s, ss :Integer;
  begin

    //First calculate the mean...
    //JumpAmp

    if prParams.StaircaseJumpAmpReestimate then begin

      OldJumpAmp := JumpAmp;
      OldJumpStd := JumpStd;
      JumpAmp    := 0;
      Sum        := 0;
      for s := MinState to MaxState do begin
        if (Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
          for ss := s + 1 to MaxState do begin
            if (Gamm[ss - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
              ThisJump  := (Amps[ss - MinState] - Amps[s - MinState]);
              dl        := ss - s;
              im        := dl;
              pm        := 0;
              for l := dl - prParams.StaircaseJumpAmpPlsMin to dl + prParams.StaircaseJumpAmpPlsMin do begin
                pc    := Gaussian(l * OldJumpAmp, OldJumpStd, ThisJump);
                if pc > pm then begin
                  im    := l;
                  pm    := pc;
                end;
              end;
              dl        := im;
              if dl <> 0 then begin
                JumpAmp := JumpAmp +
                           (ThisJump / dl)  *
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
                           //(Gamm[ss - MinState] * Gamm[s - MinState]);
                Sum     := Sum     +
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
                           //(Gamm[ss - MinState] * Gamm[s - MinState]);
              end;

            end;
          end;
        end;
      end;
      if Sum <> 0 then JumpAmp := JumpAmp / Sum;
      JumpAmps[0] := JumpAmp;

    end;


    //...and then the std
    //JumpStd
    if prParams.StaircaseJumpStdReest then begin

      JumpStd    := 0;
      Sum        := 0;
      for s := MinState to MaxState do begin
        if (Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
          for ss := s + 1 to MaxState do begin
            if (Gamm[ss - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
              ThisJump  := (Amps[ss - MinState] - Amps[s - MinState]);
              dl        := ss - s;
              im        := dl;
              pm        := 0;
              for l := dl - prParams.StaircaseJumpAmpPlsMin to dl + prParams.StaircaseJumpAmpPlsMin do begin
                pc    := Gaussian(l * OldJumpAmp, OldJumpStd, ThisJump);
                if pc > pm then begin
                  im    := l;
                  pm    := pc;
                end;
              end;
              dl        := im;
              if dl = 1 {<> 0} then begin
                JumpStd := JumpStd +
                           Sqr((ThisJump / dl) - JumpAmp) *
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
                Sum     := Sum +
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
              end;

            end;
          end;
        end;
      end;
      if Sum <> 0 then JumpStd := Sqrt(JumpStd / Sum);

    end;//Reest JumpStd ?

  end;

  procedure ReestJumpAmpStd;
  var
    j, s :Integer;
  begin

    if prParams.StaircaseJumpAmpReestimate then begin

      //JumpAmp
      if prJumpAmpsCount = 1 then begin
      //1

        JumpAmp := 0;
        Sum         := 0;
        for s := MinState to MaxState - 1 do begin
          if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
            JumpAmp := JumpAmp +
                       (Amps[s + 1 - MinState] - Amps[s - MinState]) *
                       (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));

            Sum     := Sum     +
                       (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
          end;
        end;
        if Sum <> 0 then JumpAmp := JumpAmp / Sum;
        JumpAmps[0] := JumpAmp;

      end else begin
      //2

        JumpAmp := 0;
        for j := 0 to prJumpAmpsCount - 1 do begin

          JumpAmps[j] := 0;
          Sum         := 0;
          for s := MinState to MaxState - 1 do begin
            if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
              //What is RefState here ?!
              if ((s - RefState) mod prJumpAmpsCount) = j then begin
                JumpAmps[j] := JumpAmps[j] +
                               (Amps[s + 1 - MinState] - Amps[s - MinState]) *
                               (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
                Sum         := Sum         +
                               (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
              end;
            end;
          end;
          if Sum <> 0 then JumpAmps[j] := JumpAmps[j] / Sum;
          JumpAmp := JumpAmp + JumpAmps[j];
        end;

        //Switch JumpAmps, if necessary - may need to get rid of this !!!
        p   := 0;
        s   := 0;
        im  := s;
        Sum := Gamma[s, p];
        for s := 1 to HalfStateCount - 1 do begin
          if Gamma[s, p] > Sum then begin
            im := s;
            Sum := Gamma[s, p];
          end;
        end;
        if im = Middle + 1 then begin
          Sum         := JumpAmps[1];
          JumpAmps[1] := JumpAmps[0];
          JumpAmps[0] := Sum;
        end;

      end;//JumpAmpsCount

    end;//Reest JumpAmp ?

    //JumpStd
    if prParams.StaircaseJumpStdReest then begin
      JumpStd := 0;
      Sum     := 0;
      for s := MinState to MaxState - 1 do begin
        if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
          JumpStd := JumpStd +
                     Sqr((Amps[s + 1 - MinState] - Amps[s - MinState]) - JumpAmp) *
                     (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
          Sum     := Sum     +
                     (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
        end;
      end;
      if Sum <> 0 then JumpStd := Sqrt(JumpStd / Sum);
    end;

  end;

  function GetThisAmp(State, Point, RState :Integer) :Double;
  begin

    try
      if prJumpAmpsCount = 1 then begin
          Result := JumpAmp     * (State + StateOffset[Point] - RState);
      end else begin
        if Odd(RState) then begin
          Result := JumpAmp     * ((State + StateOffset[Point] - RState) div prJumpAmpsCount) +
                    JumpAmps[1] * ((State + StateOffset[Point] - RState) mod prJumpAmpsCount);
        end else begin
          Result := JumpAmp     * ((State + StateOffset[Point] - RState) div prJumpAmpsCount) +
                    JumpAmps[0] * ((State + StateOffset[Point] - RState) mod prJumpAmpsCount);
        end;
      end;
    except
      Result := 0;
    end;

  end;

  procedure DoGamma;
  var
    p, s :Integer;
  begin

    for p := 0 to DataCount - 1 do begin
      for s := 0 to HalfStateCount - 1 do begin
        Gamma[s, p] := 0;
      end;//of s
      Gamma[State[p], p] := 1;
    end;//of p

  end;

  procedure DoEpsilon;
  var
    l, p, s, ss :Integer;
  begin

    for p := 0 to DataCount - 2 do begin

      //Clear
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          TPMDouble(Epsilon[p])[s, ss] := 0;
        end;//of ss
      end;//of s

      //Calculate and add
      for s := 0 to HalfStateCount - 1 do begin

        for ss := 0 to HalfStateCount - 1 do begin

          MakeAStates(lState, rState, s, ss, p + 1);

          for l := 0 to prStateCount - 1 do begin

            if ((lState + l) < prStateCount) and ((rState + l) < prStateCount) then begin
              TPMDouble(Epsilon[p])[lState + l, rState + l] := TPMDouble(Epsilon[p])[lState + l, rState + l] +
                                                               Gamma[s,  p] *
                                                               Gamma[ss, p + 1];
            end else begin
              Break;//from l
            end;

          end;//of l

        end;//of ss

      end;//of s

    end;//of p

  end;

  procedure ReestAMatrix;
  var
    p, s, ss :Integer;
  begin

    //Clear
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss] := 0;
      end;//of ss
    end;//of s
    //Add Epsilon
    for p := 0 to DataCount - 2 do begin
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          AMatrix[s, ss] := AMatrix[s, ss] + TPMDouble(Epsilon[p])[s, ss];
        end;//of ss
      end;//of s
    end;
    //Rescale
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        Sum := Sum + AMatrix[s, ss];
      end;//of ss
      for ss := 0 to prStateCount - 1 do begin
        if Sum <> 0 then AMatrix[s, ss] := AMatrix[s, ss] / Sum;
        LogAMatrix[s, ss] := SafeLog(AMatrix[s, ss]);
      end;//of ss
    end;//of s

  end;//ReestAMatrix

  procedure DoBaselineBack;
  var
    p :Integer;
  begin

    p := DataCount - 1;

    //Recalculate the last point gain k:
    //k := (BaseP[State[p-1], p-1] + BaseVar) / (BaseP[State[p-1], p-1] + BaseVar + MeasVar);

    //SBaseQ[p] = (I - k . c) . a . BaseP[p-1]
    //SBaseQ[p] := (1 - k) * BaseP[State[p-1], p-1];

    //SBase[p] = Base[p]
    SBaseX[p] := BaseX[State[p], p];
    SBaseP[p] := BaseP[State[p], p];

    //oJ := 0;

    for p := DataCount - 1 downto 1 do begin

      //pp = a . P[p-1] . a'  + BaseVar
      //cJ = P[p-1] . a' . (pp)-1
      cJ          := BaseP[State[p-1], p-1] / (BaseP[State[p-1], p-1] + BaseVar);

      //SBaseX[p-1] = BaseX[p-1] + cJ . (SBaseX[p] - a . BaseX[p-1])
      SBaseX[p-1] := BaseX[State[p-1], p-1] + cJ * (SBaseX[p] - BaseX[State[p-1], p-1]);

      //SBaseP[p-1] = BaseP[p-1] + cJ . (SBaseP[p] - pp) . cJ'
      SBaseP[p-1] := BaseP[State[p-1], p-1] + cJ * (SBaseP[p] - (BaseP[State[p-1], p-1] + BaseVar)) * cJ;

      //if p < (DataCount - 1) then begin
        //SBaseQ[p] = BaseP[p] . cJ'  +  oJ . (SBaseQ[p+1] - a . BaseP[p]) . cJ'
        //SBaseQ[p] := BaseP[State[p], p] * cJ  +  oJ * (SBaseQ[p+1] - BaseP[State[p], p]) * cJ;
      //end;

      //oJ := cJ;

    end;

  end;//DoBaselineBack

  procedure ReestBaselineVar;
  var
    p :Integer;
  begin

    cJ := 0;
    oJ := 0;

    for p := 1 to DataCount - 1 do begin

      //cJ = cJ + SBaseX[p]*SBaseX[p]' + SBaseP[p];
      cJ := cJ + SBaseX[p] * SBaseX[p] + SBaseP[p];

      //oJ = oJ + SBaseX[p]*SBaseX[p-1]' + SBaseQ[p];
      oJ := oJ + SBaseX[p] * SBaseX[p-1] + SBaseQ[p];
    end;

    //BaseVar = (cJ - A . oJ') / (DataCount - 1)
    BaseVar := (cJ - oJ) / (DataCount - 1);

    if BaseVar < KALMAN_SMALL then BaseVar := KALMAN_SMALL;
    BaseStd := Sqrt(BaseVar);

  end;//ReestBaselineVar;

  procedure FindMinMaxState;
  var
    p :Integer;
  begin

    MinState  := 0                  + StateOffset[0];
    MaxState  := HalfStateCount - 1 + StateOffset[0];
    for p := 1 to DataCount - 1 do begin
      if StateOffset[p] < MinState then begin
        MinState := StateOffset[p];
      end;
      if (StateOffset[p] + HalfStateCount - 1) > MaxState then begin
        MaxState := HalfStateCount - 1 + StateOffset[p];
      end;
    end;


  end;

  procedure DoAmpsGamm;
  var
    p, s :Integer;
  begin

    //Resize Amps and Gamm
    ReAllocMem(Amps, SizeOf(Double) * (MaxState - MinState + 1));
    ReAllocMem(Gamm, SizeOf(Double) * (MaxState - MinState + 1));

    //Clear Amps, Gamm
    for s := MinState to MaxState do begin
      Amps[s - MinState] := 0;
      Gamm[s - MinState] := 0;
    end;

    //Calculate Amps and Gamm
    for p := 0 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      for s := MinState to MaxState do begin
        if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
          Amps[s - MinState] := Amps[s - MinState] + y * Gamma[s - StateOffset[p], p];
          Gamm[s - MinState] := Gamm[s - MinState] +     Gamma[s - StateOffset[p], p];
        end;
      end;
    end;
    for s := MinState to MaxState do begin
      if Gamm[s - MinState] <> 0 then Amps[s - MinState] := Amps[s - MinState] / Gamm[s - MinState];
    end;

  end;

  procedure ReestMeasStd;
  var
    p, s :Integer;
  begin

    if prParams.StaircaseJumpStdReest then begin

      Sum     := 0;
      MeasStd := 0;
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        for s := MinState to MaxState do begin
          if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
            MeasStd := MeasStd + Gamma[s - StateOffset[p], p] * Sqr(y - Amps[s - MinState]);
            Sum     := Sum     + Gamma[s - StateOffset[p], p];
          end;
        end;
      end;
      if Sum <> 0 then MeasStd := Sqrt(MeasStd / Sum);

    end;//Reest JumpStd ?

  end;//ReestMeasStd


  //Calculate the LL after re-estimation
  function GetNewLL :Double;
  var
    p, s, ss :Integer;
  begin

    //Mark the weakly represented levels - although they are not used
    ReAllocMem(LevelExists, SizeOf(Integer) * (MaxState - MinState + 1));
    for s := MinState to MaxState do begin
      if Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM then begin
        LevelExists[s - MinState] := 1;
      end else begin
        LevelExists[s - MinState] := 0;
      end;
    end;


    //Calculate ll
    p        := 0;
    s        := State[p];
    y        := Data[p] * Scale + Off;

    Result   := SafeLog(Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y)) +
                SafeLog(Prob0[s]);

    for p := 1 to DataCount - 1 do begin

      s        := State[p];
      ss       := State[p-1];
      y        := Data[p] * Scale + Off;

      MakeAStates(lState, rState, ss, s, p);

      Result   := Result +
                  SafeLog(Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y)) +
                  LogAMatrix[lState, rState];

    end;

  end;//GetNewLL


begin


  Result              := False;


  //Nullify pointers ...........................................................
  VProb               := nil;
  VBack               := nil;

  Alpha               := nil;
  Betta               := nil;
  Gamma               := nil;
  Gauss               := nil;
  Rho                 := nil;
  Epsilon             := nil;

  State               := nil;
  StateOffset         := nil;

  AMatrix             := nil;
  LogAMatrix          := nil;

  Prob0               := nil;

  Amps                := nil;
  Gamm                := nil;

  BaseX               := nil;
  BaseP               := nil;

  SBaseX              := nil;
  SBaseP              := nil;
  SBaseQ              := nil;

  CBaseX              := nil;
  CBaseP              := nil;

  LevelExists         := nil;

  JumpAmps            := nil;

  JumpAmp             := 0;
  MeasStd             := 0;



  //Signal Running status ......................................................
  Chunk               := TQUB_Thread_PChunk(IdlBaseChunk);
  Chunk.Status        := CHUNK_STATUS_RUNNING;


  try//finally



    //Get data parameters ......................................................
    //   - Data, DataCount
    Data              := IdlBaseChunk.Data;
    DataCount         := Chunk.Sel.Last - Chunk.Sel.First + 1;
    //   - Scale
    Scale             := 1 / IdlBaseChunk.Scaling;
    Off               := IdlBaseChunk.Offset;


    HalfStateCount    := prStateCount div 2;
    Middle            := (HalfStateCount div 2) + 1 - 1;



    //Allocate memory ..........................................................


    //   - VProb - the probabilities used by Viterbi;
    VProb             := DAlloc2(HalfStateCount, DataCount);
    //   - VBack - backtracing used by Viterbi;
    VBack             := IAlloc2(HalfStateCount, DataCount);

    //   - Vectors
    Alpha             := DAlloc2(HalfStateCount, DataCount);
    Betta             := DAlloc2(HalfStateCount, DataCount);
    Gamma             := DAlloc2(HalfStateCount, DataCount);
    Gauss             := DAlloc2(HalfStateCount, DataCount);
    Rho               := DAlloc1(DataCount);
    Epsilon           := PAlloc1(DataCount);
    for p := 0 to DataCount - 1 do begin
      Epsilon[p]      := DAlloc2(prStateCount, prStateCount);
    end;

    //   - State - the idealized state, at each t;
    State             := IAlloc1(DataCount);
    //   - StateOffset - the offset of the state vector, at each t;
    StateOffset       := IAlloc1(DataCount);


    //   - A matrix
    AMatrix           := DAlloc2(prStateCount, prStateCount);
    LogAMatrix        := DAlloc2(prStateCount, prStateCount);


    //   - Prob0 - initial probabilities;
    Prob0             := DAlloc1(HalfStateCount);


    if prParams.BaselineTrack then begin
      //   - BaseX|P - baseline state|variance, for each data point, for each state;
      BaseX           := DAlloc2(HalfStateCount, DataCount);
      BaseP           := DAlloc2(HalfStateCount, DataCount);
      //   - SBaseX|P|Q - smoothed baseline state, variance and BaseVar, for each data point
      SBaseX          := DAlloc1(DataCount);
      SBaseP          := DAlloc1(DataCount);
      SBaseQ          := DAlloc1(DataCount);
      //   - CBaseX|P - current baseline, for each state
      CBaseX          := DAlloc1(HalfStateCount);
      CBaseP          := DAlloc1(HalfStateCount);
    end;


    JumpAmps          := DAlloc1(prJumpAmpsCount);



    //Initialize variables .....................................................


    //   - A matrix
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss]    := Abs(prA[s, ss]);//to prevent some occasional negative entries
        Sum               := Sum + AMatrix[s, ss];
      end;
      for ss := 0 to prStateCount - 1 do begin //re-normalize
        if Sum <> 0 then AMatrix[s, ss]    := AMatrix[s, ss] / Sum;
        LogAMatrix[s, ss] := SafeLog(AMatrix[s, ss]);
      end;
    end;



    //   - Prob0
    //Start with equal probability and let the Gaussian make the difference
    for s := 0 to HalfStateCount - 1 do begin
      Prob0[s]     := 1 / HalfStateCount;
    end;



    //   - JumpAmp, MeasStd, BaseAmp
    JumpAmp        := prJumpAmp;
    for s := 0 to prJumpAmpsCount - 1 do begin
      JumpAmps[s]  := prJumpAmps[s];
    end;

    JumpStd        := prParams.StaircaseJumpStd;

    MeasStd        := prMeasStd;
    //MeasVar        := Sqr(MeasStd);

    BaseAmp        := prBaseAmp;

    BaseStd        := prParams.BaselineStd;
    BaseVar        := Sqr(BaseStd);


    //Initialize BaseX[0] and BaseP[0]
    if prParams.BaselineTrack then begin
      p                 := 0;
      y                 := Data[p] * Scale + Off;
      for s := 0 to HalfStateCount - 1 do begin
        BaseX[s, p]     := BaseAmp;
        if Abs(BaseAmp - y) > 3*MeasStd then begin
          BaseX[s, p]   := y;
        end;
        BaseP[s, p]     := Sqr(MeasStd);
      end;
    end;


    //StateOffset
    StateOffset[0] := - Middle;
    RefState       := 0;

    MinState       := 0;
    MaxState       := 0;


    //Print starting values ....................................................
    prConsoleMsg := 'Starting values:'#13#10;
    prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
      end;
      prConsoleMsg := prConsoleMsg + #13#10;
    end;
    prConsoleMsg := prConsoleMsg + Format('Jump amp = %-14.6g, Meas std = %-14.6g, Base amp = %-14.6g'#13#10, [JumpAmp, MeasStd, BaseAmp]);
    for c := 0 to prJumpAmpsCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [c, JumpAmps[c]]);
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    if not prParams.QuietOutput then Synchronize(prWriteNewConsole);



    try


      OldLL := 0;


      //Iterate
      for Iteration := 1 to prParams.MaxIter do begin


        NewLL      := 0;

        RefState   := 0;

        if prParams.BaselineTrack then begin
          RefAmp   := BaseX[0, 0];
        end else begin
          RefAmp   := BaseAmp;//This is always the case because BaseX is reestimated as BaseAmp
        end;


        //Variances
        MeasVar    := Sqr(MeasStd);
        JumpVar    := Sqr(JumpStd);


        //Forward **************************************************************


        //Makes the assumption that the first point IS A (not B)
        //which is wrong. It may get past it. Or use a combination of amps ?


        //Initialize the first point
        p          := 0;
        y          := Data[p] * Scale + Off;


        for s := 0 to HalfStateCount - 1 do begin

          //ThisAmp is what the data should be, in absence of baseline or baseamp
          ThisAmp := GetThisAmp(s, p, RefState);

          kl          := 0;//LL of Kalman process
          if prParams.BaselineTrack then begin
            //Save the previous values
            kxp       := BaseX[s, p];
            KalmanForward(BaseX[s, p], BaseP[s, p], BaseX[s, p], BaseP[s, p], MeasVar, BaseVar, (y - ThisAmp){BaseData});
            kl        := SafeLog(Gaussian(kxp, BaseStd, BaseX[s, p]));
            RefAmp    := BaseX[s, p];
          end;

          //p(y|s0, x0) * p(s0) * p(x0)
          VProb[s, 0] := SafeLog(Gaussian(ThisAmp{expected mean}, MeasStd, y - RefAmp{base.corr. measurement})) +
                         SafeLog(Prob0[s]) +
                         kl;

        end;


        //Loop for every point
        for p := 1 to DataCount - 1 do begin


          StateOffset[p] := StateOffset[p - 1];

          y              := Data[p] * Scale + Off;


          for Offset := 1 to 2 do begin


            //for each state
            for s := 0 to HalfStateCount - 1 do begin

              //ThisAmp is what the data should be, in absence of baseline or baseamp
              ThisAmp     := GetThisAmp(s, p, RefState);

              for ss := 0 to HalfStateCount - 1 do begin

                MakeAStates(lState, rState, ss, s, p);

                kl          := 0;//LL of Kalman process
                if prParams.BaselineTrack then begin

                  //Increase the baseline variance if there is a jump ?
                  //ThisStd   := Abs(rState - lState) *
                  //             prParams.StaircaseBaseJumpVar *
                  //             JumpVar;
                  ThisStd   := {ThisStd + }BaseVar;

                  KalmanForward(CBaseX[ss], CBaseP[ss], BaseX[ss, p - 1], BaseP[ss, p - 1], MeasVar, ThisStd{BaseVar}, (y - ThisAmp){BaseData});
                  kl        := SafeLog(Gaussian(BaseX[ss, p - 1], Sqrt(ThisStd{=BaseVar here}){BaseStd},  CBaseX[ss]));
                  RefAmp    := CBaseX[ss];
                end;

                //p(y|s0, x0) * p(s0) * p(x0)
                pc          := SafeLog(Gaussian(ThisAmp{expected mean}, MeasStd, y - RefAmp{base.corr. measurement})) +
                               LogAMatrix[lState, rState] +
                               kl +
                               VProb[ss, p - 1];

                //Max of ss
                if (ss = 0) then begin
                  im := ss;
                  pm := pc;
                end else if (pc > pm) then begin
                  im := ss;
                  pm := pc;
                end;

              end;//of ss

              VProb[s, p] := pm;
              VBack[s, p] := im;

              if prParams.BaselineTrack then begin
                BaseX[s, p] := CBaseX[im];
                BaseP[s, p] := CBaseP[im];
              end;

            end;//of s



            //Find the Max of s and do Offset, if necessary .........................
            s   := 0;
            im  := s;
            Sum := VProb[s, p];
            for s := 1 to HalfStateCount - 1 do begin
              if VProb[s, p] > Sum then begin
                im := s;
                Sum := VProb[s, p];
              end;
            end;

            if Offset = 1 then begin
              StateOffset[p] := StateOffset[p - 1] + (im - Middle);
            end;


          end;//Offset


        end;//of p



        //BackTrace
        s        := 0;
        p        := DataCount - 1;
        im       := 0;
        pm       := VProb[s, p];
        for s := 1 to HalfStateCount - 1 do begin
          if VProb[s, p] > pm then begin
            im := s;
            pm := VProb[s, p];
          end;
        end;
        State[p] := im;
        for p := DataCount - 2 downto 0 do begin
          State[p] := VBack[State[p + 1], p + 1];
        end;



        //Backward step on Baseline
        if prParams.BaselineTrack then begin

          DoBaselineBack;

        end;//Backward on baseline



        //Backward *************************************************************


        //Gamma
        DoGamma;

        if prParams.StaircaseKineticsReest then begin

          //Epsilon
          DoEpsilon;

        end;//Reest Kinetics ?



        //Re-estimation ********************************************************


        //Find Min/MaxState;
        FindMinMaxState;


        //Amps and Gamm
        DoAmpsGamm;


        //Calculate MeasStd
        ReestMeasStd;


        //Update the BaseAmp
        s       := 0;
        p       := 0;
        im      := s;
        Sum     := Gamma[s, p];
        for s := 1 to HalfStateCount - 1 do begin
          if Gamma[s, p] > Sum then begin
            im  := s;
            Sum := Gamma[s, p];
          end;
        end;
        BaseAmp := Amps[im + StateOffset[p] - MinState];


        //Update Baseline start
        if prParams.BaselineTrack then begin
          for s := 0 to HalfStateCount - 1 do begin
            BaseX[s, 0] := BaseAmp;
            BaseP[s, 0] := SBaseP[0];
          end;
        end;


       //Update the JumpAmps and JumpStd
        if (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_COMBINATIONS) and (prJumpAmpsCount = 1) then begin

          ReestJumpAmpStdCombinations;

        end else begin //(prParams.StaircaseJumpAmpReest = JUMPAMP_CONSECUTIVE) 

          ReestJumpAmpStd;

        end;


        if prParams.StaircaseKineticsReest then begin

          //A Matrix
          ReestAMatrix;


          //Print Q matrix as the logarithm of the A matrix ......................
          MatrixLogarithm(AMatrix, Epsilon[0], IDLBaseChunk.Sampling, prStateCount, 100, 1e-4);
          for s := 0 to prStateCount - 1 do begin
            for ss := 0 to prStateCount - 1 do begin
              if TPMDouble(Epsilon[0])[s, ss] < 0 then begin
                TPMDouble(Epsilon[0])[s, ss] := 0;
                TPMDouble(Epsilon[0])[ss, s] := 0;
              end;
            end;
          end;
          prConsoleMsg := #13#10'Big Q Matrix:'#13#10;
          for s := 0 to prStateCount - 1 do begin
            for ss := 0 to prStateCount - 1 do begin
              if (s = ss) or (prQ[s, ss] = 0)then begin
                prConsoleMsg := prConsoleMsg + #9'...           ';
              end else begin
                prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [TPMDouble(Epsilon[0])[s, ss]]);
              end;
            end;
            prConsoleMsg := prConsoleMsg + #13#10;
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);

          //Print Q matrix as the logarithm of the A matrix ......................
          TPMDouble(Epsilon[0])[0, 0] := AMatrix[0, 0];
          TPMDouble(Epsilon[0])[0, 1] := 1 - AMatrix[0, 0];
          TPMDouble(Epsilon[0])[1, 1] := AMatrix[prStateCount - 1, prStateCount - 1];
          TPMDouble(Epsilon[0])[1, 0] := 1 - AMatrix[prStateCount - 1, prStateCount - 1];

          MatrixLogarithm(Epsilon[0], Epsilon[0], IDLBaseChunk.Sampling, 2, 100, 1e-4);
          for s := 0 to {prStateCount -} 1 do begin
            for ss := 0 to {prStateCount -} 1 do begin
              if TPMDouble(Epsilon[0])[s, ss] < 0 then begin
                TPMDouble(Epsilon[0])[s, ss] := 0;
                TPMDouble(Epsilon[0])[ss, s] := 0;
              end;
            end;
          end;
          prConsoleMsg := #13#10'Small (LeftTop-RightBottom) Q Matrix:'#13#10;
          for s := 0 to {prStateCount -} 1 do begin
            for ss := 0 to {prStateCount -} 1 do begin
              if (s = ss) or (prQ[s, ss] = 0)then begin
                prConsoleMsg := prConsoleMsg + #9'...           ';
              end else begin
                prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [TPMDouble(Epsilon[0])[s, ss]]);
              end;            end;
            prConsoleMsg := prConsoleMsg + #13#10;
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


        end;//Reest Kinetics ?


        //Calculate LL with the reestimated parameters
        NewLL := GetNewLL;


        //Print current iteration values .......................................
        prConsoleMsg := Format(#13#10'Iteration %d'#13#10, [Iteration]);
        prConsoleMsg := prConsoleMsg + Format('LL: %14.10g'#13#10, [NewLL]);
        prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
        for s := 0 to prStateCount - 1  do begin
          for ss := 0 to prStateCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleMsg := prConsoleMsg + Format('Jump amp = %-14.6g Jump std = %-14.6g'#13#10'Meas std = %-14.6g'#13#10'Base amp = %-14.6g'#13#10, [JumpAmp, JumpStd, MeasStd, BaseAmp]);
        prConsoleMsg := prConsoleMsg + 'Jump amps:'#13#10;
        for c := 0 to prJumpAmpsCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [c, JumpAmps[c]]);
        end;
        if prParams.BaselineTrack then begin
          prConsoleMsg := prConsoleMsg + Format('BaseVar = %-14.6g BaseStd = %-14.6g'#13#10, [BaseVar, BaseStd]);
          prConsoleMsg := prConsoleMsg + Format('BaseX[0] = %-14.6g BaseP[0] = %-14.6g'#13#10, [SBaseX[0], SBaseP[0]]);
        end;
        prConsoleMsg := prConsoleMsg + 'Level, Amp, dAmp, Gamm:'#13#10;
        for c := MinState to MaxState do begin
          if c > MinState then begin
            prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#13#10,        [c, Amps[c - MinState], Amps[c - MinState] - Amps[c - 1 - MinState], Gamm[c - MinState]]);
          end else begin
            prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'              '#9'%-14.6g'#13#10, [c, Amps[c - MinState],                                              Gamm[c - MinState]]);
          end;
        end;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


        if (prStopped) and (not prParams.StaircaseGlobalReest) then Break;


        //Check convergence
        if Abs(NewLL - OldLL) <= prParams.LLConv then Break;


        if prParams.Reestimation = IDLBASE_REEST_NONE then Break;


        OldLL := NewLL;


      end;//of Iteration



      //**********************************************************************//
      //                                                                      //
      //          B a u m - W e l c h     R e - e s t i m a t i o n           //
      //                                                                      //
      //**********************************************************************//
      if prParams.StaircaseReestBW then begin


        prConsoleMsg := #13#10'Second step: Baum-Welch reestimation'#13#10;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


        ReAllocMem(LevelExists, SizeOf(Integer) * (MaxState - MinState + 1));

        OldLL := 0;


        //Iterate
        for Iteration := 1 to prParams.MaxIter do begin


          //Mark the weakly represented levels
          for s := MinState to MaxState do begin
            if Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM then begin
              LevelExists[s - MinState] := 1;
            end else begin
              LevelExists[s - MinState] := 0;
            end;
          end;


          NewLL      := 0;


          //Forward ************************************************************


          //Initialize the first point
          p      := 0;
          y      := Data[p] * Scale + Off;

          Rho[p] := 0;
          for s := 0 to HalfStateCount - 1 do begin

            Gauss[s, p] := Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y);
            //p(y|s0) * p(s0)
            Alpha[s, p] := {@@@}LevelExists[s + StateOffset[p] - MinState] *
                           Prob0[s] *
                           Gauss[s, p];
            Rho[p]      := Rho[p] + Alpha[s, p];

          end;

          NewLL := NewLL + SafeLog(Rho[p]);

          //Normalize Alpha
          for s := 0 to HalfStateCount - 1 do begin
            Alpha[s, p] := Alpha[s, p] / Rho[p];
          end;


          //Loop for every point
          for p := 1 to DataCount - 1 do begin

            y := Data[p] * Scale + Off;

            Rho[p] := 0;
            for s := 0 to HalfStateCount - 1 do begin

              Gauss[s, p] := Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y);
              Alpha[s, p] := 0;

              for ss := 0 to HalfStateCount - 1 do begin

                MakeAStates(lState, rState, ss, s, p);
                Alpha[s, p] := Alpha[s, p] +
                               AMatrix[lState, rState] *
                               Alpha[ss, p - 1];

              end;//of ss

              Alpha[s, p] := {@@@}LevelExists[s + StateOffset[p] - MinState] *
                             Alpha[s, p] *
                             Gauss[s, p];
              Rho[p]      := Rho[p] + Alpha[s, p];

            end;//of s

            //Normalize Alpha
            for s := 0 to HalfStateCount - 1 do begin
              Alpha[s, p] := Alpha[s, p] / Rho[p];
            end;


            NewLL := NewLL + SafeLog(Rho[p]);


          end;//of p


          //Backward *************************************************************


          //Betta
          p := DataCount - 1;

          for s := 0 to HalfStateCount - 1 do begin
            Betta[s, p] := 1 / Rho[p];
          end;//of s

          for p := DataCount - 2 downto 0 do begin

            for s := 0 to HalfStateCount - 1 do begin

              Betta[s, p] := 0;

              for ss := 0 to HalfStateCount - 1 do begin

                MakeAStates(lState, rState, s, ss, p + 1);
                Betta[s, p] := Betta[s, p] +
                               {@@@}LevelExists[ss + StateOffset[p + 1] - MinState] *
                               AMatrix[lState, rState] *
                               Gauss[ss, p + 1] *
                               Betta[ss, p + 1];

              end;//of ss

              Betta[s, p] := Betta[s, p] / Rho[p];

            end;//of s

          end;//of p


          //Gamma
          for p := 0 to DataCount - 1 do begin

            Sum := 0;
            for s := 0 to HalfStateCount - 1 do begin
              Gamma[s, p] := Alpha[s, p] * Betta[s, p];
              Sum := Sum + Gamma[s, p];
            end;//of s

            //Rescale
            for s := 0 to HalfStateCount - 1 do begin
              Gamma[s, p] := Gamma[s, p] / Sum;
            end;//of s

          end;//of p


          //Epsilon
          for p := 0 to DataCount - 2 do begin

            Sum := 0;

            for s := 0 to HalfStateCount - 1 do begin

              for ss := 0 to HalfStateCount - 1 do begin

                MakeAStates(lState, rState, s, ss, p + 1);

                TPMDouble(Epsilon[p])[s, ss] := Alpha[s, p] *
                                                AMatrix[lState, rState] *
                                                Gauss[ss, p + 1] *
                                                Betta[ss, p + 1];
                Sum := Sum + TPMDouble(Epsilon[p])[s, ss];

              end;//of ss

            end;//of s

            //Rescale
            for s := 0 to HalfStateCount - 1 do begin

              for ss := 0 to HalfStateCount - 1 do begin
                TPMDouble(Epsilon[p])[s, ss] := TPMDouble(Epsilon[p])[s, ss] / Sum;
              end;//of ss

            end;//of s

          end;//of p


          //Make the state vector
          for p := 0 to DataCount - 1 do begin
            State[p] := 0;
            Sum      := Gamma[0, p];
            for s := 1 to HalfStateCount - 1 do begin
              if Gamma[s, p] > Sum then begin
                State[p] := s;
                Sum := Gamma[s, p];
              end;
            end;
          end;


          //Re-estimation ********************************************************


          //Clear Amps, Gamm
          for s := MinState + Middle to MaxState - Middle - 1 do begin
            Amps[s - MinState] := 0;
            Gamm[s - MinState] := 0;
          end;

          //Calculate Amps and Gamm
          for p := 0 to DataCount - 1 do begin
            y := Data[p] * Scale + Off;
            for s := MinState + Middle to MaxState - Middle - 1 do begin
              if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
                Amps[s - MinState] := Amps[s - MinState] + y * Gamma[s - StateOffset[p], p];
                Gamm[s - MinState] := Gamm[s - MinState] +     Gamma[s - StateOffset[p], p];
              end;
            end;
          end;
          for s := MinState + Middle to MaxState - Middle - 1 do begin
            if Gamm[s - MinState] <> 0 then Amps[s - MinState] := Amps[s - MinState] / Gamm[s - MinState];
          end;

          //Calculate MeasStd
          Sum     := 0;
          MeasStd := 0;
          for p := 0 to DataCount - 1 do begin
            y := Data[p] * Scale + Off;
            for s := MinState + Middle to MaxState - Middle - 1 do begin
              if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
                MeasStd := MeasStd + Gamma[s - StateOffset[p], p] * Sqr(y - Amps[s - MinState]);
                Sum     := Sum     + Gamma[s - StateOffset[p], p];
              end;
            end;
          end;
          if Sum <> 0 then MeasStd := Sqrt(MeasStd / Sum);



          //Update the BaseAmp
          BaseAmp := Amps[Middle];



          //Update the JumpAmps and JumpStd
          if (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_COMBINATIONS) and (prJumpAmpsCount = 1) then begin


            //JumpAmp
            JumpAmp := 0;
            Sum     := 0;
            for s := MinState to MaxState do begin
              if (Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
                for ss := s + 1 to MaxState do begin
                  if (Gamm[ss - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
                    JumpAmp := JumpAmp +
                               (Gamm[s - MinState] * Gamm[ss - MinState]) *
                               Abs(Amps[s - MinState] - Amps[ss - MinState]) * (1 / Abs(s - ss));
                    Sum     := Sum     +
                               (Gamm[s - MinState] * Gamm[ss - MinState]);
                  end;
                end;
              end;
            end;
            if Sum <> 0 then JumpAmp := JumpAmp / Sum;
            JumpAmps[0] := JumpAmp;


            //JumpStd
            JumpStd := 0;
            Sum     := 0;
            for s := MinState to MaxState do begin
              if (Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
                for ss := s + 1 to MaxState do begin
                  if (Gamm[ss - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
                    JumpStd := JumpStd +
                               (Gamm[s - MinState] * Gamm[ss - MinState]) *
                               Sqr((Abs(Amps[s - MinState] - Amps[ss - MinState]) * (1 / Abs(s - ss))) - JumpAmp);
                    Sum     := Sum     +
                               (Gamm[s - MinState] * Gamm[ss - MinState]);
                  end;
                end;
              end;
            end;
            if Sum <> 0 then JumpStd := Sqrt(JumpStd / Sum);


          end else begin //if (prParams.StaircaseJumpAmpReest = JUMPAMP_CONSECUTIVE) 


            //JumpAmp
            if prJumpAmpsCount = 1 then begin
            //1

              JumpAmp := 0;
              Sum         := 0;
              for s := MinState to MaxState - 1 do begin
                if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
                  JumpAmp := JumpAmp + (Gamm[s - MinState] * Gamm[s + 1 - MinState]) * (Amps[s + 1 - MinState] - Amps[s - MinState]);
                  Sum     := Sum     + (Gamm[s - MinState] * Gamm[s + 1 - MinState]);
                end;
              end;
              if Sum <> 0 then JumpAmp := JumpAmp / Sum;
              JumpAmps[0] := JumpAmp;

            end else begin
            //2

              JumpAmp := 0;
              for j := 0 to prJumpAmpsCount - 1 do begin

                JumpAmps[j] := 0;
                Sum         := 0;
                for s := MinState to MaxState - 1 do begin
                  if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
                    if ((s - RefState) mod prJumpAmpsCount) = j then begin
                      JumpAmps[j] := JumpAmps[j] + (Gamm[s - MinState] * Gamm[s + 1 - MinState]) * (Amps[s + 1 - MinState] - Amps[s - MinState]);
                      Sum         := Sum         + (Gamm[s - MinState] * Gamm[s + 1 - MinState]);
                    end;
                  end;
                end;
                if Sum <> 0 then JumpAmps[j] := JumpAmps[j] / Sum;
                JumpAmp := JumpAmp + JumpAmps[j];
              end;

              //Switch JumpAmps, if necessary - may need to get rid of this !!!
              p   := 0;
              if State[p] = Middle + 1 then begin
                Sum         := JumpAmps[1];
                JumpAmps[1] := JumpAmps[0];
                JumpAmps[0] := Sum;
              end;

            end;//JumpAmpsCount


            //JumpStd
            JumpStd := 0;
            Sum     := 0;
            for s := MinState to MaxState - 1 do begin
              if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
                JumpStd := JumpStd + (Gamm[s - MinState] * Gamm[s + 1 - MinState]) * Sqr((Amps[s + 1 - MinState] - Amps[s - MinState]) - JumpAmp);
                Sum     := Sum     + (Gamm[s - MinState] * Gamm[s + 1 - MinState]);
              end;
            end;
            if Sum <> 0 then JumpStd := Sqrt(JumpStd / Sum);


          end;//LevelViterbi



          //A Matrix
          // . . .



          //Print current iteration values .......................................
          prConsoleMsg := Format(#13#10'Second step Iteration %d'#13#10, [Iteration]);
          prConsoleMsg := prConsoleMsg + Format('LL: %14.10g'#13#10, [NewLL]);
          prConsoleMsg := prConsoleMsg + Format('Jump amp = %-14.6g Jump std = %-14.6g'#13#10'Meas std = %-14.6g'#13#10'Base amp = %-14.6g'#13#10, [JumpAmp, JumpStd, MeasStd, BaseAmp]);
          prConsoleMsg := prConsoleMsg + 'Level, Exists, Amp, Gamm:'#13#10;
          for c := MinState + Middle to MaxState - Middle - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [c, LevelExists[c - MinState], Amps[c - MinState], Gamm[c - MinState]]);
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


          if prStopped then Break;


          //Check convergence
          if Abs(NewLL - OldLL) <= prParams.LLConv then Break;


          OldLL := NewLL;


        end;//of Iteration


      end;//Second step: Baum-Welch reestimation





      //Results ................................................................
      if not ((prStopped) and (not prParams.StaircaseGlobalReest)) then begin


        //Which states are actually used
        Min1 := State[0] + StateOffset[0];
        Max1 := State[0] + StateOffset[0];
        for p := 1 to DataCount - 1 do begin
          if (State[p] + StateOffset[p]) < Min1 then begin
            Min1 := State[p] + StateOffset[p];
          end;
          if (State[p] + StateOffset[p]) > Max1 then begin
            Max1 := State[p] + StateOffset[p];
          end;
        end;

        UsedCount := Max1 - Min1 + 1;

        IdlBaseChunk.ClassCount := UsedCount;
        IdlBaseChunk.StateCount := prStateCount;
        IdlBaseChunk.Amps       := SAlloc1(UsedCount);
        IdlBaseChunk.Stds       := SAlloc1(UsedCount);
        IdlBaseChunk.AMatrix    := DAlloc2(prStateCount, prStateCount);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            IdlBaseChunk.AMatrix[s, ss] := AMatrix[s, ss];
          end;
        end;
        for s := Min1 to Max1 do begin
          IdlBaseChunk.Amps[s - Min1]  := Amps[s - MinState];
          IdlBaseChunk.Stds[s - Min1]  := MeasStd;
        end;


        //Make Dwells
        for p := 0 to DataCount - 1 do begin
          Inc(State[p], StateOffset[p] - Min1);
        end;

        prMakeDwells(IdlBaseChunk, State, DataCount);


        //Make Baseline Nodes
        if (prParams.BaselineTrack) and (prParams.BaselineCorrect) then begin

          prBase2Nodes(IdlBaseChunk, SBaseX);

        end;


        //Make Stats and Histograms
        prMakeStats(IdlBaseChunk, SBaseX); // SBaseX is nil if not prParams.BaselineTrack


        IdlBaseChunk.LL         := NewLL;
        IdlBaseChunk.Iterations := Iteration - 1;


        if not prParams.StaircaseGlobalReest then begin
          prConsoleMsg := Format('%d', [TQUB_Thread_PChunk(IdlBaseChunk).Index, IdlBaseChunk.LL]);
          for p := 0 to prJumpAmpsCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [JumpAmps[p]]);
          end;
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.10g', [JumpStd, MeasStd, BaseAmp, IdlBaseChunk.LL]);
          prConsoleCode := MESSAGE_MESSAGE;
          Synchronize(prWriteNewConsole);
        end;
                                                 

      end;//Results


    except


      Result            := True;

      Chunk.Status      := CHUNK_STATUS_ERROR;
      Chunk.Completion  := 0;


    end;


  finally



    //Free Memory ..............................................................
    Free2(TPMChar(VProb));
    Free2(TPMChar(VBack));

    Free2(TPMChar(Alpha));
    Free2(TPMChar(Betta));
    Free2(TPMChar(Gamma));
    Free2(TPMChar(Gauss));
    FreeMem(Rho);
    if Epsilon <> nil then begin
      for p := 0 to DataCount - 1 do begin
        Free2(TPMChar(Epsilon[p]));
      end;
    end;
    FreeMem(Epsilon);

    FreeMem(State);
    FreeMem(StateOffset);

    FreeMem(Prob0);

    FreeMem(Amps);
    FreeMem(Gamm);

    Free2(TPMChar(BaseX));
    Free2(TPMChar(BaseP));
    FreeMem(SBaseX);
    FreeMem(SBaseP);
    FreeMem(SBaseQ);
    FreeMem(CBaseX);
    FreeMem(CBaseP);

    FreeMem(LevelExists);


    //Signal new status
    if Chunk.Status <> CHUNK_STATUS_ERROR then begin

      Chunk.Status := CHUNK_STATUS_DONE;
      Chunk.Completion := 100;

      if (not prStopped) and (not prSkipped) then begin

        Inc(prTotCount);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            prTotA[s, ss] := prTotA[s, ss] + AMatrix[s, ss];
          end;
        end;
        for p := 0 to prJumpAmpsCount - 1 do begin
          prTotJumpAmps[p] := prTotJumpAmps[p] + JumpAmps[p];
        end;
        prTotJumpAmp     := prTotJumpAmp     + JumpAmp;
        prTotJumpStd     := prTotJumpStd     + JumpStd;
        prTotBaseAmp     := prTotBaseAmp     + BaseAmp;
        prTotKf          := prTotKf          + 0{Jf};
        prTotKb          := prTotKb          + 0{Jb};
        prTotMeasStd     := prTotMeasStd     + MeasStd;
        prTotLL          := prTotLL          + IdlBaseChunk.LL;

      end;
    end;


    FreeMem(JumpAmps);

    Free2(TPMChar(AMatrix));
    Free2(TPMChar(LogAMatrix));


  end;

end;



//****************************************************************************//
//                                                                            //
//                              S T A I R C A S E                             //
//                                                                            //
//                                                                            //
//                    B A U M  -  W E L C H  -  K A L M A N                   //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlBase.prReest_Staircase_BW(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
var
  Chunk               :TQUB_Thread_PChunk;

  Data                :TPAInteger;
  DataCount           :Integer;
  Scale               :Double;
  Off                 :Double;

  HalfStateCount      :Integer;

  Iteration           :Integer;

  //*********************************
  Alpha               :TPMDouble;
  Betta               :TPMDouble;
  Gamma               :TPMDouble;
  Gauss               :TPMDouble;
  Rho                 :TPADouble;
  Epsilon             :TPAPointer;

  State               :TPAInteger;
  StateOffset         :TPAInteger;

  AMatrix             :TPMDouble;

  Prob0               :TPADouble;

  Amps                :TPADouble;
  Gamm                :TPADouble;

  BaseX               :TPADouble;
  BaseP               :TPADouble;

  SBaseX              :TPADouble;
  SBaseP              :TPADouble;
  SBaseQ              :TPADouble;

  //*********************************

  Sum                 :Double;

  pm, pc              :Double;

  y                   :Double;

  c, p, pp, s, ss, im :Integer;
  j                   :Integer;

  OldLL, NewLL        :Double;

  JumpAmp             :Double;
  JumpAmps            :TPADouble;
  JumpStd             :Double;//Used for calculations
  JumpStdReest        :Double;//Used for output
  JumpVar             :Double;

  MeasStd             :Double;
  MeasVar             :Double;

  BaseAmp             :Double;
  BaseVar             :Double;
  BaseStd             :Double;

  RefAmp              :Double;
  RefState            :Integer;

  ThisAmp             :Double;
  ThisStd             :Double;

  MinState            :Integer;
  MaxState            :Integer;
  //MinWState           :Integer;
  //MaxWState           :Integer;

  Min1, Max1          :Integer;

  lState              :Integer;
  rState              :Integer;

  Middle              :Integer;

  UsedCount           :Integer;

  CanQuit             :Boolean;
  BaseEstCount        :Integer;

  Offset              :Integer;

  oJ, cJ              :Double;

  ya, yv              :Double;

  yl, hl, kl          :Double;

  kxp, kpp            :Double;

  k                   :Double;

  LevelExists         :TPAInteger;
  LevelCount          :Integer;

  dl                  :Integer;
  ThisJump            :Double;
  OldJumpAmp          :Double;
  OldJumpStd          :Double;

  JumpAmpFixed        :Double;
  JumpAmpLS           :Double;
  BaseAmpFixed        :Double;

  //Movies
  MovieData           :INode;
  MovieOffset         :INode;

  Jf                  :Double;
  Jb                  :Double;

  OldStateOffset      :Integer;

    
  //**************************************************************************//
  //                                                                          //
  //                                   H M M                                  //
  //                                                                          //
  //**************************************************************************//
  procedure MakeAStates(var lState, rState :Integer; l, r :Integer; p :Integer);
  var
    dState :Integer;
  begin
    dState := (r + StateOffset[p]) - (l + StateOffset[p - 1]);
    if dState >= 0 then begin
      lState := 0;
      rState := lState + dState;
    end else begin
      rState := 0;
      lState := rState - dState;
    end;
  end;

  procedure DoBetta;
  var
    p, s, ss :Integer;
  begin

    p := DataCount - 1;

    for s := 0 to HalfStateCount - 1 do begin
      Betta[s, p] := 1 / Rho[p];
    end;//of s


    for p := DataCount - 2 downto 0 do begin

      for s := 0 to HalfStateCount - 1 do begin

        Betta[s, p] := 0;
        for ss := 0 to HalfStateCount - 1 do begin

          MakeAStates(lState, rState, s, ss, p + 1);

          Betta[s, p] := Betta[s, p] +
                         AMatrix[lState, rState] *
                         Gauss[ss, p + 1] *
                         Betta[ss, p + 1];

        end;//of ss

        Betta[s, p] := Betta[s, p] / Rho[p];

      end;//of s

      if prStoppedCount >= 3 then Exit;

    end;//of p

  end;

  procedure DoBettaII;
  var
    p, s, ss :Integer;
  begin

    p := DataCount - 1;

    for s := 0 to HalfStateCount - 1 do begin
      Betta[s, p] := 1 / Rho[p];
    end;//of s

    for p := DataCount - 2 downto 0 do begin

      for s := 0 to HalfStateCount - 1 do begin

        Betta[s, p] := 0;

        for ss := 0 to HalfStateCount - 1 do begin

          MakeAStates(lState, rState, s, ss, p + 1);
          Betta[s, p] := Betta[s, p] +
                         {@@@}LevelExists[ss + StateOffset[p + 1] - MinState] *
                         AMatrix[lState, rState] *
                         Gauss[ss, p + 1] *
                         Betta[ss, p + 1];

        end;//of ss

        Betta[s, p] := Betta[s, p] / Rho[p];

      end;//of s

      if prStoppedCount >= 3 then Exit;

    end;//of p

  end;

  procedure DoGamma;
  var
    p, s :Integer;
  begin

    for p := 0 to DataCount - 1 do begin

      Sum := 0;
      for s := 0 to HalfStateCount - 1 do begin

        Gamma[s, p] := Alpha[s, p] * Betta[s, p];
        Sum := Sum + Gamma[s, p];

      end;//of s

      //Rescale
      for s := 0 to HalfStateCount - 1 do begin

        Gamma[s, p] := Gamma[s, p] / Sum;

      end;//of s

      if prStoppedCount >= 3 then Exit;

    end;//of p

  end;

  procedure DoEpsilon;
  var
    l, p, s, ss :Integer;
  begin

    for p := 0 to DataCount - 2 do begin

      //Clear
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          TPMDouble(Epsilon[p])[s, ss] := 0;
        end;//of ss
      end;//of s

      //Calculate and add
      for s := 0 to HalfStateCount - 1 do begin

        for ss := 0 to HalfStateCount - 1 do begin

          MakeAStates(lState, rState, s, ss, p + 1);

          for l := 0 to prStateCount - 1 do begin

            if ((lState + l) < prStateCount) and ((rState + l) < prStateCount) then begin
              TPMDouble(Epsilon[p])[lState + l, rState + l] := TPMDouble(Epsilon[p])[lState + l, rState + l] +
                                                               Alpha[s, p] *
                                                               AMatrix[lState, rState] *
                                                               Gauss[ss, p + 1] *
                                                               Betta[ss, p + 1];
            end else begin
              Break;//from l
            end;

          end;//of l

        end;//of ss

      end;//of s

      if prStoppedCount >= 3 then Exit;

    end;//of p

  end;

  procedure ReestAMatrix;
  var
    p, s, ss :Integer;
  begin

    //Clear
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss] := 0;
      end;//of ss
    end;//of s
    //Add Epsilon
    for p := 0 to DataCount - 2 do begin
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          AMatrix[s, ss] := AMatrix[s, ss] + TPMDouble(Epsilon[p])[s, ss];
        end;//of ss
      end;//of s
      if prStoppedCount >= 3 then Exit;
    end;
    //Rescale
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        Sum := Sum + AMatrix[s, ss];
      end;//of ss
      for ss := 0 to prStateCount - 1 do begin
        if Sum <> 0 then AMatrix[s, ss] := AMatrix[s, ss] / Sum;
      end;//of ss
    end;//of s

  end;

  //**************************************************************************//
  //                                                                          //
  //                              B a s e l i n e                             //
  //                                                                          //
  //**************************************************************************//
  procedure DoBaselineBack;
  var
    p :Integer;
  begin

    p  := DataCount - 1;

    //Recalculate the last point gain k:
    //k := (BaseP[p-1] + BaseVar) / (BaseP[p-1] + BaseVar + MeasVar);

    //SBaseQ[p] = (I - k . c) . a . BaseP[p-1]
    //SBaseQ[p] := (1 - k) * BaseP[p-1];

    //SBase[p] = Base[p]
    SBaseX[p] := BaseX[p];
    SBaseP[p] := BaseP[p];

    //oJ := 0;

    for p := DataCount - 1 downto 1 do begin

      //pp = a . P[p-1] . a'  + BaseVar
      //cJ = P[p-1] . a' . (pp)-1
      cJ          := BaseP[p-1] / (BaseP[p-1] + BaseVar);

      //SBaseX[p-1] = BaseX[p-1] + cJ . (SBaseX[p] - a . BaseX[p-1])
      SBaseX[p-1] := BaseX[p-1] + cJ * (SBaseX[p] - BaseX[p-1]);

      //SBaseP[p-1] = BaseP[p-1] + cJ . (SBaseP[p] - pp) . cJ'
      SBaseP[p-1] := BaseP[p-1] + cJ * (SBaseP[p] - (BaseP[p-1] + BaseVar)) * cJ;

      //if p < (DataCount - 1) then begin
        //SBaseQ[p] = BaseP[p] . cJ'  +  oJ . (SBaseQ[p+1] - a . BaseP[p]) . cJ'
        //SBaseQ[p] := BaseP[p] * cJ  +  oJ * (SBaseQ[p+1] - BaseP[p]) * cJ;
      //end;

      //oJ := cJ;

      if prStoppedCount >= 3 then Exit;
    end;

  end;

  procedure ReestBaselineVar;
  var
    p :Integer;
  begin

    cJ := 0;
    oJ := 0;

    for p := 1 to DataCount - 1 do begin

      //cJ = cJ + SBaseX[p]*SBaseX[p]' + SBaseP[p];
      cJ := cJ + SBaseX[p] * SBaseX[p] + SBaseP[p];

      //oJ = oJ + SBaseX[p]*SBaseX[p-1]' + SBaseQ[p];
      oJ := oJ + SBaseX[p] * SBaseX[p-1] + SBaseQ[p];
      if prStoppedCount >= 3 then Exit;
    end;

    //BaseVar = (cJ - A . oJ') / (DataCount - 1)
    BaseVar := (cJ - oJ) / (DataCount - 1);

    if BaseVar < KALMAN_SMALL then BaseVar := KALMAN_SMALL;
    BaseStd := Sqrt(BaseVar);

  end;

  procedure FindMinMaxState;
  var
    p{, s, im} :Integer;
  begin

    MinState  := 0                  + StateOffset[0];
    MaxState  := HalfStateCount - 1 + StateOffset[0];
    for p := 1 to DataCount - 1 do begin
      if StateOffset[p] < MinState then begin
        MinState := StateOffset[p];
      end;
      if (StateOffset[p] + HalfStateCount - 1) > MaxState then begin
        MaxState := HalfStateCount - 1 + StateOffset[p];
      end;
      if prStoppedCount >= 3 then Exit;
    end;

    //Testing
    //Now find the Min/Max OCCUPIED state
    {
    for p := 0 to DataCount - 1 do begin
      s       := 0;
      im      := s;
      Sum     := Gamma[s, p];
      for s := 1 to HalfStateCount - 1 do begin
        if Gamma[s, p] > Sum then begin
          im  := s;
          Sum := Gamma[s, p];
        end;
      end;
      if p = 0 then begin
        MinWState := im + StateOffset[p];
        MaxWState := MinWState;
      end else begin
        if MinWState > (im + StateOffset[p]) then MinWState := im + StateOffset[p];
        if MaxWState < (im + StateOffset[p]) then MaxWState := im + StateOffset[p];
      end;
    end;
    }

  end;

  //**************************************************************************//
  //                                                                          //
  //                                J u m p s                                 //
  //                                                                          //
  //**************************************************************************//
  procedure DoAmpsGamm;
  var
    p, s :Integer;
  begin

    //Resize Amps and Gamm
    ReAllocMem(Amps, SizeOf(Double) * (MaxState - MinState + 1));
    ReAllocMem(Gamm, SizeOf(Double) * (MaxState - MinState + 1));

    //Clear Amps, Gamm
    for s := MinState to MaxState do begin
      Amps[s - MinState] := 0;
      Gamm[s - MinState] := 0;
    end;

    //Calculate Amps and Gamm
    for p := 0 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      for s := MinState to MaxState do begin
        if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
          Amps[s - MinState] := Amps[s - MinState] + y * Gamma[s - StateOffset[p], p];
          Gamm[s - MinState] := Gamm[s - MinState] +     Gamma[s - StateOffset[p], p];
        end;
      end;
      if prStoppedCount >= 3 then Exit;
    end;
    for s := MinState to MaxState do begin
      if Gamm[s - MinState] <> 0 then Amps[s - MinState] := Amps[s - MinState] / Gamm[s - MinState];
    end;

  end;

  procedure DoAmpsGammBW;
  var
    p, s :Integer;
  begin

    //Clear Amps, Gamm
    for s := MinState + Middle to MaxState - Middle do begin
      Amps[s - MinState] := 0;
      Gamm[s - MinState] := 0;
    end;

    //Calculate Amps and Gamm
    for p := 0 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      for s := MinState + Middle to MaxState - Middle do begin
        if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
          Amps[s - MinState] := Amps[s - MinState] + y * Gamma[s - StateOffset[p], p];
          Gamm[s - MinState] := Gamm[s - MinState] +     Gamma[s - StateOffset[p], p];
        end;
      end;
      if prStoppedCount >= 3 then Exit;
    end;
    for s := MinState + Middle to MaxState - Middle do begin
      if Gamm[s - MinState] <> 0 then Amps[s - MinState] := Amps[s - MinState] / Gamm[s - MinState];
    end;

  end;

  //Calculates JumpAmp assuming there is no Jump variance
  procedure DoJumpAmpNoVar;
  var
    Suml_Sumt_Gamma        :Double;
    Suml_Sumt_Gamma_y      :Double;
    Suml_L_Sumt_Gamma      :Double;
    Suml_L_Sumt_Gamma_y    :Double;
    Suml_LL_Sumt_Gamma     :Double;
    Sumt_Gamma             :Double;
    Sumt_Gamma_y           :Double;
    a, b                   :Double;
    l, p                   :Integer;
  begin

    Suml_Sumt_Gamma        := 0;
    Suml_Sumt_Gamma_y      := 0;
    Suml_L_Sumt_Gamma      := 0;
    Suml_L_Sumt_Gamma_y    := 0;
    Suml_LL_Sumt_Gamma     := 0;

    for l := MinState to MaxState do begin

      Sumt_Gamma           := 0;
      Sumt_Gamma_y         := 0;

      for p := 0 to DataCount - 1 do begin

        if ((l - StateOffset[p]) >= 0) and ((l - StateOffset[p]) < HalfStateCount) then begin

          y                := Data[p] * Scale + Off;

          Sumt_Gamma       := Sumt_Gamma   +     Gamma[l - StateOffset[p], p];
          Sumt_Gamma_y     := Sumt_Gamma_y + y * Gamma[l - StateOffset[p], p];

        end;

      if prStoppedCount >= 3 then Exit;
      end;//p

      Suml_Sumt_Gamma        := Suml_Sumt_Gamma        +         Sumt_Gamma;
      Suml_Sumt_Gamma_y      := Suml_Sumt_Gamma_y      +         Sumt_Gamma_y;

      Suml_L_Sumt_Gamma      := Suml_L_Sumt_Gamma      + l *     Sumt_Gamma;
      Suml_L_Sumt_Gamma_y    := Suml_L_Sumt_Gamma_y    + l *     Sumt_Gamma_y;
      Suml_LL_Sumt_Gamma     := Suml_LL_Sumt_Gamma     + l * l * Sumt_Gamma;

    end;//l


    a            := Suml_Sumt_Gamma_y - ((Suml_L_Sumt_Gamma_y * Suml_Sumt_Gamma) / (Suml_L_Sumt_Gamma));
    b            := Suml_L_Sumt_Gamma - ((Suml_LL_Sumt_Gamma  * Suml_Sumt_Gamma) / (Suml_L_Sumt_Gamma));

    JumpAmpFixed := a / b;
    BaseAmpFixed := (Suml_L_Sumt_Gamma_y - JumpAmpFixed * Suml_LL_Sumt_Gamma) / Suml_L_Sumt_Gamma;

  end;

  procedure ReestJumpStd;
  var
    s               :Integer;
  begin

    JumpStdReest    := 0;
    Sum             := 0;
    if prJumpAmpsCount = 1 then begin
      for s := MinState to MaxState - 1 do begin
        if (Gamm[s - MinState] > prParams.StaircaseJumpStdMinGamma) and (Gamm[s + 1 - MinState] > prParams.StaircaseJumpStdMinGamma) then begin
          if prParams.StaircaseJumpStdWeighed then begin
            JumpStdReest := JumpStdReest +
                       Sqr((Amps[s + 1 - MinState] - Amps[s - MinState]) - JumpAmp) *
                       ((Gamm[s + 1 - MinState] * Gamm[s - MinState]) / ((Gamm[s + 1 - MinState]) + (Gamm[s - MinState])));
                       //(Gamm[s + 1 - MinState] + Gamm[s - MinState]);
            Sum     := Sum     +
                       ((Gamm[s + 1 - MinState] * Gamm[s - MinState]) / ((Gamm[s + 1 - MinState]) + (Gamm[s - MinState])));
                       //(Gamm[s + 1 - MinState] + Gamm[s - MinState]);
          end else begin
            JumpStdReest := JumpStdReest +
                       Sqr((Amps[s + 1 - MinState] - Amps[s - MinState]) - JumpAmp);
            Sum     := Sum     +
                       1;
          end;
        end;
      end;
      if Sum <> 0 then JumpStdReest := Sqrt(JumpStdReest / Sum);
    end else begin
      s := MinState;
      repeat
        if (Gamm[s - MinState] > prParams.StaircaseJumpStdMinGamma) and (Gamm[s + 2 - MinState] > prParams.StaircaseJumpStdMinGamma) then begin
          if prParams.StaircaseJumpStdWeighed then begin
            JumpStdReest := JumpStdReest +
                       Sqr((Amps[s + 2 - MinState] - Amps[s - MinState]) - JumpAmp) *
                       ((Gamm[s + 2 - MinState] * Gamm[s - MinState]) / ((Gamm[s + 2 - MinState]) + (Gamm[s - MinState])));
                       //(Gamm[s + 2 - MinState] + Gamm[s - MinState]);
            Sum     := Sum     +
                       ((Gamm[s + 2 - MinState] * Gamm[s - MinState]) / ((Gamm[s + 2 - MinState]) + (Gamm[s - MinState])));
                       //(Gamm[s + 2 - MinState] + Gamm[s - MinState]);
          end else begin
            JumpStdReest := JumpStdReest +
                       Sqr((Amps[s + 2 - MinState] - Amps[s - MinState]) - JumpAmp);
            Sum     := Sum     +
                       1;
          end;
        end;
        Inc(s);
      until s > (MaxState - 2);
      if Sum <> 0 then JumpStdReest := Sqrt(JumpStdReest / (2*Sum));
    end;
  end;

  //Calculates JumpAmp and Amps simultaneously, given JumpStd and MeasStd
  procedure DoJumpAmpLS(UpdateAmps, UpdateJumpAmp :Boolean);
  var
    Amp_Size    :Integer;
    JA1         :Integer;
    Amp_A       :TPMDouble;
    Amp_AI      :TPMDouble;
    Amp_B       :TPADouble;
    Amp_X       :TPADouble;
    Gammat      :TPADouble;
    Gammat_yt   :TPADouble;
    Sm2         :Double;
    Sj2         :Double;
    SumW        :Double;

    i, j, p     :Integer;

    //t           :String;

    function W(l :Integer) :Double;
    begin
      if prParams.StaircaseJumpAmpWeighed then begin
        //Result := (Gammat[l] * Gammat[l + 1]) / (Gammat[l] + Gammat[l + 1]);

        //Testing!!!
        {
        0.1 may be too small for lots of missed events and high jump variance and high noise.
        }
        Result := Max(0.5, (Gammat[l] * Gammat[l + 1]) / (Gammat[l] + Gammat[l + 1]));
        //Result := 0.5 + (Gammat[l] * Gammat[l + 1]) / (Gammat[l] + Gammat[l + 1]);
        //Result := Gammat[l] + Gammat[l + 1];
      end else begin
        Result := 1;
      end;
    end;

  begin


    Amp_A       := nil;
    Amp_AI      := nil;
    Amp_B       := nil;
    Amp_X       := nil;
    Gammat      := nil;
    Gammat_yt   := nil;


    { value assigned to p never used
    p := 0;
    for i := MinState to MaxState - 1 do begin
      if (Gamm[i - MinState] > 0.5) and (Gamm[i + 1 - MinState] > 0.5) then begin
        Inc(p);
      end;
    end;
    }



    //Testing
    //Sm2         := (MeasStd*MeasStd)/DataCount;
    Sm2         := (MeasStd*MeasStd)*DataCount;
    if not prParams.StaircaseJumpAmpWeighed then begin
      //Sj2         := (JumpStd*JumpStd)/(2 * (MaxState - MinState + 1) - 1);
      //Sj2         := (JumpStd*JumpStd)/(2 * (prA[0, 0] + prA[0, 1]) * (MaxState - MinState + 1) - 1);
      //Sj2         := (JumpStd*JumpStd)/(2 * Sqr(p/(MaxState - MinState))*(MaxState - MinState));
      //Sj2         := (JumpStd*JumpStd) / (2 * prA[0, 0] * prA[0, 1] * (MaxState - MinState - (HalfStateCount div 2)));
      //Sj2         := (JumpStd*JumpStd) / (2 * prA[0, 0] * (MaxState - MinState - (HalfStateCount div 2)));
      //Sj2         := (JumpStd*JumpStd) / (2.5 * (MaxState - MinState - (HalfStateCount div 2)));
      Sj2         := prParams.StaircaseJumpAmpWeight*(1/40)*DataCount*(JumpStd*JumpStd)*(MaxState - MinState - (HalfStateCount div 2));
      //Sj2         := prParams.StaircaseJumpAmpWeight*(1/40)*DataCount*(JumpStd*JumpStd)*p;
    end else begin
      //Sj2         := (JumpStd*JumpStd) / (2 * prA[0, 0] * prA[0, 1] * (MaxState - MinState - (HalfStateCount div 2)));
      //Sj2         := (JumpStd*JumpStd) / ((MaxState - MinState));
      Sj2         := (JumpStd*JumpStd)*(MaxState - MinState - (HalfStateCount div 2));
    end;

    try


      {
        Must solve the equation Ax = B
          x is the vector of unknowns (amplitudes), including the JumpAmp;
          A and B are obtained by setting the derivatives of the cost function to zero
          First Amp_Size rows are for Amps, the last for JumpAmp
      }


      //Allocate A and B:

      //Testing
      Amp_Size    := MaxState - MinState + 1;
      JA1         := Amp_Size;
      Amp_A       := DAlloc2(Amp_Size + 1, Amp_Size + 1);
      Amp_AI      := DAlloc2(Amp_Size + 1, Amp_Size + 1);
      Amp_B       := DAlloc1(Amp_Size + 1);
      Amp_X       := DAlloc1(Amp_Size + 1);
      Gammat      := DAlloc1(Amp_Size);
      Gammat_yt   := DAlloc1(Amp_Size);


      //Clear Gammat, Gammat_yt
      for i := 0 to Amp_Size - 1 do begin
        Gammat   [i] := 1e-6;//To prevent numerical screw-up (with 0);
        Gammat_yt[i] := 1e-6;//Idem
      end;


      //Calculate Gammat and Gammat_yt
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        for i := MinState to MaxState do begin
          if ((i - StateOffset[p]) >= 0) and ((i - StateOffset[p]) < HalfStateCount) then begin
            Gammat   [i - MinState] := Gammat   [i - MinState] +     Gamma[i - StateOffset[p], p];
            Gammat_yt[i - MinState] := Gammat_yt[i - MinState] + y * Gamma[i - StateOffset[p], p];
          end;
        end;
        if prStoppedCount >= 3 then Exit;
      end;


      //Calculate SumW
      if prParams.StaircaseJumpAmpWeighed then begin
        SumW := 0;
        for i := 0 to Amp_Size - 2 do begin
          SumW := SumW + W(i);
        end;
      end else begin
        SumW := Amp_Size - 1;
      end;

      //Testing
      {
      t := 'Cls     '#9'Gammat  '#9'Gammat_yt'#13#10;
      for i := 0 to Amp_Size - 1 do begin
        t := t + Format('%d'#9'%-8.6g'#9'%-8.6g'#13#10, [i, Gammat[i], Gammat_yt[i]]);
      end;
      prConsoleMsg := t;
      prConsoleCode := MESSAGE_WARNING;
      if not prParams.QuietOutput then Synchronize(prWriteNewConsole);
      }


      //Zero the matrices
      for i := 0 to Amp_Size do begin
        Amp_B[i] := 0;
        Amp_X[i] := 0;
      end;
      for i := 0 to Amp_Size do begin
        for j := 0 to Amp_Size do begin
          Amp_A[i, j] := 0;
        end;
      end;


      //Amp_B
      for i := 0 to Amp_Size - 1 do begin
        Amp_B[i] := Gammat_yt[i] / Sm2;
      end;


      //Three different methods - which is correct?!

      //Amp_A - I -

      for i := 0 to Amp_Size do begin
        for j := 0 to Amp_Size do begin
          Amp_A[i, j] := 0;
        end;
      end;
      for i := 0 to Amp_Size - 1 do begin
        Amp_A[i, i] := Gammat[i] / Sm2;
        if i > 0              then begin
          Amp_A[i, i]         := Amp_A[i, i]        + W(i - 1) / Sj2;
          Amp_A[i, i - 1]     :=                    - W(i - 1) / Sj2;
          Amp_A[i, JA1]       := Amp_A[i, JA1]      - W(i - 1) / Sj2;
          Amp_A[JA1, i]       := Amp_A[JA1, i]      - W(i - 1) / Sj2;
        end;
        if i < (Amp_Size - 1) then begin
          Amp_A[i, i]         := Amp_A[i, i]        + W(i)     / Sj2;
          Amp_A[i, i + 1]     :=                    - W(i)     / Sj2;
          Amp_A[i, JA1]       := Amp_A[i, JA1]      + W(i)     / Sj2;
          Amp_A[JA1, i]       := Amp_A[JA1, i]      + W(i)     / Sj2;
        end;
      end;
      //for i := 0 to Amp_Size - 2 do begin
      //  Amp_A[JA1, JA1] := Amp_A[JA1, JA1] + W(i) / Sj2;
      //end;
      Amp_A[JA1, JA1] := SumW / Sj2;


      //Testing
      {
      t := 'Amp_A, Amp_B - I - :'#13#10;
      for i := 0 to Amp_Size do begin
        for j := 0 to Amp_Size do begin
          t := t + Format('%-8.6g'#9, [Amp_A[i, j]]);
        end;
        t := t + Format('%-8.6g'#13#10, [Amp_B[i]]);
      end;
      prConsoleMsg := t;
      prConsoleCode := MESSAGE_WARNING;
      Synchronize(prWriteNewConsole);
      /}


      //Amp_A - II -
      {
      for i := 0 to Amp_Size do begin
        for j := 0 to Amp_Size do begin
          Amp_A[i, j] := 0;
        end;
      end;
      for i := 0 to Amp_Size - 1 do begin
        if          (i = 0)                              then begin
        //First line

          Amp_A[i, i]         := Gammat[i]/Sm2 + 1/Sj2;
          Amp_A[i, i + 1]     :=               - 1/Sj2;
          Amp_A[i, Amp_Size]  :=               + 1/Sj2;

        end else if (i > 0) and (i < (Amp_Size - 1))     then begin
        //Middle

          Amp_A[i, i]         := Gammat[i]/Sm2 + 2/Sj2;
          Amp_A[i, i - 1]     :=               - 1/Sj2;
          Amp_A[i, i + 1]     :=               - 1/Sj2;

        end else if             (i = (Amp_Size - 1))     then begin
        //Last line

          Amp_A[i, i]         := Gammat[i]/Sm2 + 1/Sj2;
          Amp_A[i, i - 1]     :=               - 1/Sj2;
          Amp_A[i, Amp_Size]  :=               - 1/Sj2;

        end;
      end;
      Amp_A[Amp_Size, 0]            :=   1/Sj2;
      Amp_A[Amp_Size, Amp_Size - 1] := - 1/Sj2;
      Amp_A[Amp_Size, Amp_Size]     :=   (Amp_Size - 1)/Sj2;
      }


      //Testing
      {
      t := 'Amp_A, Amp_B - II - :'#13#10;
      for i := 0 to Amp_Size do begin
        for j := 0 to Amp_Size do begin
          t := t + Format('%-8.6g'#9, [Amp_A[i, j]]);
        end;
        t := t + Format('%-8.6g'#13#10, [Amp_B[i]]);
      end;
      prConsoleMsg := t;
      prConsoleCode := MESSAGE_WARNING;
      Synchronize(prWriteNewConsole);
      }


      //Amp_A - III -
      {
      for i := 0 to Amp_Size do begin
        for j := 0 to Amp_Size do begin
          Amp_A[i, j] := 0;
        end;
      end;
      for i := 0 to Amp_Size - 1 do begin
        if          (i = 0)                              then begin
        //First line

          Amp_A[i, i]           := Gammat[i]/Sm2 + W(i)/(Sj2*SumW);
          Amp_A[i, i + 1]       :=               - W(i)/(Sj2*SumW);
          Amp_A[i, Amp_Size]    :=               + W(i)/(Sj2*SumW);

        end else if (i > 0) and (i < (Amp_Size - 1))     then begin
        //Middle

          Amp_A[i, i]           := Gammat[i]/Sm2 + (W(i) + W(i - 1))/(Sj2*SumW);
          Amp_A[i, i + 1]       :=               -  W(i)            /(Sj2*SumW);
          Amp_A[i, Amp_Size]    :=               + (W(i) - W(i - 1))/(Sj2*SumW);

        end else if             (i = (Amp_Size - 1))     then begin
        //Last line

          Amp_A[i, i]           := Gammat[i]/Sm2 + W(i - 1)/(Sj2*SumW);
          Amp_A[i, Amp_Size]    :=               - W(i - 1)/(Sj2*SumW);

        end;
      end;
      Amp_A[Amp_Size, Amp_Size] := 1/Sj2;
      //The matrix is symmetric:
      for i := 0 to Amp_Size do begin
        for j := i to Amp_Size do begin
          Amp_A[j, i] := Amp_A[i, j];
        end;
      end;
      }


      //Testing
      {
      t := 'Amp_A, Amp_B - III - :'#13#10;
      for i := 0 to Amp_Size do begin
        for j := 0 to Amp_Size do begin
          t := t + Format('%-8.6g'#9, [Amp_A[i, j]]);
        end;
        t := t + Format('%-8.6g'#13#10, [Amp_B[i]]);
      end;
      prConsoleMsg := t;
      prConsoleCode := MESSAGE_WARNING;
      Synchronize(prWriteNewConsole);
      }



      //X = Inv(A) * B
      {
      InvMat(PMatrix(Amp_A), 0, Amp_Size, PMatrix(Amp_AI));
      for k := 0 to Amp_Size do begin
        for i := 0 to Amp_Size do begin
          Amp_X[k] := Amp_X[k] + Amp_AI[k, i] * Amp_B[i];
        end;
      end;
      }

      //Testing:

      if SolveLinEqs(Amp_A, Amp_B, Amp_Size + 1, Amp_X, True) then begin
        //Beep;
      end;




      JumpAmpLS := Amp_X[Amp_Size];

      //Testing
      {
      t := 'Cls     '#9'Amp     '#9'AmpLS   '#9'Gamm    '#13#10;
      for i := 0 to Amp_Size - 1 do begin
        t := t + Format('%d'#9'%-8.6g'#9'%-8.6g'#9'%-8.6g'#13#10, [i, Amps[i], Amp_X[i], Gamm[i]]);
      end;
      prConsoleMsg := t;
      prConsoleCode := MESSAGE_WARNING;
      if not prParams.QuietOutput then Synchronize(prWriteNewConsole);
      }


      if UpdateJumpAmp then begin
        JumpAmp     := Amp_X[JA1];
        JumpAmps[0] := JumpAmp;
      end;
      if UpdateAmps then begin
        for i := 0 to Amp_Size - 1 do begin
          //testing
          //if Gamm[i] <> 0 then begin
            Amps[i]   := Amp_X[i];
          //end;
        end;
      end;


    finally

      Free2(TPMChar(Amp_A));
      Free2(TPMChar(Amp_AI));
      FreeMem(Amp_B);
      FreeMem(Amp_X);
      FreeMem(Gammat);
      FreeMem(Gammat_yt);

    end;

  end;

  //Calculates JumpAmps (two unequal, alternative steps) and Amps simultaneously, given JumpStd and MeasStd
  procedure DoJumpAmpLS_UnequalSteps(UpdateAmps, UpdateJumpAmp :Boolean);
  var
    Amp_Size    :Integer;
    JA1         :Integer;
    JA2         :Integer;
    Amp_A       :TPMDouble;
    Amp_AI      :TPMDouble;
    Amp_B       :TPADouble;
    Amp_X       :TPADouble;
    Gammat      :TPADouble;
    Gammat_yt   :TPADouble;
    Sm2         :Double;
    Sj2         :Double;
    SumW1       :Double;
    SumW2       :Double;

    i, j, k, p  :Integer;

    t           :String;

    function W(l :Integer) :Double;
    begin
      Result := 1;
    end;

    function Alpha(l, j :Integer) :Double;
    var
      p :Array[0..1] of Double;
    begin

      p[0] := Gaussian(JumpAmps[0], JumpStd, Amps[l + 1] - Amps[l]);
      p[1] := Gaussian(JumpAmps[1], JumpStd, Amps[l + 1] - Amps[l]);

      //12-07-05
      //Levels with only one point, belong equally - it doesn't work if many levels are short
      {
      if (Gammat[l + 1] <= 1.5) or (Gammat[l] <= 1.5) then begin
        p[0] := 0.5;
        p[1] := 0.5;
      end;
      }

      //testing - all or none
      {
      if p[0] > p[1] then begin
        p[0] := 1;
        p[1] := 0;
      end else begin
        p[0] := 0;
        p[1] := 1;
      end;
      }
      Result := p[j] / (p[0] + p[1]);



    end;

  begin


    Amp_A       := nil;
    Amp_AI      := nil;
    Amp_B       := nil;
    Amp_X       := nil;
    Gammat      := nil;
    Gammat_yt   := nil;

    { value assigned to p never used
    p := 0;
    for i := MinState to MaxState - 1 do begin
      if (Gamm[i - MinState] > 0.5) and (Gamm[i + 1 - MinState] > 0.5) then begin
        Inc(p);
      end;
    end;
    }

    Sm2         := (MeasStd*MeasStd)*DataCount;
    Sj2         := prParams.StaircaseJumpAmpWeight*(1/40)*DataCount*(JumpStd*JumpStd)*(MaxState - MinState - (HalfStateCount div 2));

    try


      {
        Must solve the equation Ax = B
          x is the vector of unknowns (amplitudes), including the JumpAmp;
          A and B are obtained by setting the derivatives of the cost function to zero
          First Amp_Size rows are for Amps, the last for JumpAmp
      }


      //Allocate A and B:

      //Testing
      Amp_Size    := MaxState - MinState + 1;
      JA1         := Amp_Size;
      JA2         := Amp_Size + 1;
      Amp_A       := DAlloc2(Amp_Size + 2, Amp_Size + 2);
      Amp_AI      := DAlloc2(Amp_Size + 2, Amp_Size + 2);
      Amp_B       := DAlloc1(Amp_Size + 2);
      Amp_X       := DAlloc1(Amp_Size + 2);
      Gammat      := DAlloc1(Amp_Size);
      Gammat_yt   := DAlloc1(Amp_Size);


      //Clear Gammat, Gammat_yt
      for i := 0 to Amp_Size - 1 do begin
        Gammat   [i] := 1e-6;//To prevent numerical screw-up (with 0);
        Gammat_yt[i] := 1e-6;//Idem
      end;


      //Calculate Gammat and Gammat_yt
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        for i := MinState to MaxState do begin
          if ((i - StateOffset[p]) >= 0) and ((i - StateOffset[p]) < HalfStateCount) then begin
            Gammat   [i - MinState] := Gammat   [i - MinState] +     Gamma[i - StateOffset[p], p];
            Gammat_yt[i - MinState] := Gammat_yt[i - MinState] + y * Gamma[i - StateOffset[p], p];
          end;
        end;
        if prStoppedCount >= 3 then Exit;
      end;


      //12-07-05
      //Try to eliminate levels that have only one point - it doesn't work if many levels are short
      {
      for i := MinState to MaxState do begin
        if Gammat   [i - MinState] <= 1.5 then begin
          Gammat   [i - MinState] := 1e-6;
          Gammat_yt[i - MinState] := 1e-6;
        end;
        if prStoppedCount >= 3 then Exit;
      end;
      }

      //Calculate SumW
      SumW1 := 0;
      SumW2 := 0;
      for i := 0 to Amp_Size - 2 do begin
        SumW1 := SumW1 + Alpha(i, 0);
        SumW2 := SumW2 + Alpha(i, 1);
      end;

      //Zero the matrices
      for i := 0 to Amp_Size + 1 do begin
        Amp_B[i] := 0;
        Amp_X[i] := 0;
      end;
      for i := 0 to Amp_Size + 1 do begin
        for j := 0 to Amp_Size + 1 do begin
          Amp_A[i, j] := 0;
        end;
      end;

      //Amp_B
      for i := 0 to Amp_Size - 1 do begin
        Amp_B[i] := Gammat_yt[i] / Sm2;
      end;


      //Amp_A - I -
      for i := 0 to Amp_Size + 1 do begin
        for j := 0 to Amp_Size + 1 do begin
          Amp_A[i, j] := 0;
        end;
      end;
      for i := 0 to Amp_Size - 1 do begin
        Amp_A[i, i] := Gammat[i] / Sm2;
        if i > 0              then begin
          Amp_A[i, i]         := Amp_A[i, i]        +  W(i - 1) / Sj2;
          Amp_A[i, i - 1]     :=                    -  W(i - 1) / Sj2;
          Amp_A[i, JA1]       := Amp_A[i, JA1]      - (W(i - 1) / Sj2)*Alpha(i-1, 0);
          Amp_A[JA1, i]       := Amp_A[JA1, i]      - (W(i - 1) / Sj2)*Alpha(i-1, 0);
          Amp_A[i, JA2]       := Amp_A[i, JA2]      - (W(i - 1) / Sj2)*Alpha(i-1, 1);
          Amp_A[JA2, i]       := Amp_A[JA2, i]      - (W(i - 1) / Sj2)*Alpha(i-1, 1);
        end;
        if i < (Amp_Size - 1) then begin
          Amp_A[i, i]         := Amp_A[i, i]        +  W(i)     / Sj2;
          Amp_A[i, i + 1]     :=                    -  W(i)     / Sj2;
          Amp_A[i, JA1]       := Amp_A[i, JA1]      + (W(i)     / Sj2)*Alpha(i, 0);
          Amp_A[JA1, i]       := Amp_A[JA1, i]      + (W(i)     / Sj2)*Alpha(i, 0);
          Amp_A[i, JA2]       := Amp_A[i, JA2]      + (W(i)     / Sj2)*Alpha(i, 1);
          Amp_A[JA2, i]       := Amp_A[JA2, i]      + (W(i)     / Sj2)*Alpha(i, 1);
        end;
      end;
      Amp_A[JA1, JA1] := SumW1 / Sj2;
      Amp_A[JA2, JA2] := SumW2 / Sj2;


      if SolveLinEqs(Amp_A, Amp_B, Amp_Size + 2, Amp_X, True) then begin
        //Beep;
      end;

      JumpAmpLS := Amp_X[JA1] + Amp_X[JA2];

      if UpdateJumpAmp then begin
        JumpAmp     := Amp_X[JA1] + Amp_X[JA2];
        JumpAmps[0] := Amp_X[JA1];
        JumpAmps[1] := Amp_X[JA2];
      end;
      if UpdateAmps then begin
        for i := 0 to Amp_Size - 1 do begin
          Amps[i]   := Amp_X[i];
        end;
      end;


    finally

      Free2(TPMChar(Amp_A));
      Free2(TPMChar(Amp_AI));
      FreeMem(Amp_B);
      FreeMem(Amp_X);
      FreeMem(Gammat);
      FreeMem(Gammat_yt);

    end;

  end;

  procedure ReestJumpAmpStdConstrained;
  begin

    //JumpAmp
    if prParams.StaircaseJumpAmpReestimate then begin
      if prJumpAmpsCount = 1 then begin
        DoJumpAmpLS(True{UpdateAmps}, True{UpdateJumpAmp});
      end else begin
        DoJumpAmpLS_UnequalSteps(True{UpdateAmps}, True{UpdateJumpAmp});
      end;
    end;//Reest JumpAmp ?

    //JumpStd
    ReestJumpStd;
    if prParams.StaircaseJumpStdReest then begin
      JumpStd := Max(JumpStdReest, 1e-3);
    end;//Reest JumpStd  ?

  end;

  procedure ReestMeasStd;
  var
    p, s :Integer;
  begin

    if prParams.StaircaseMeasStdReest then begin

      Sum     := 0;
      MeasStd := 0;
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        for s := MinState to MaxState do begin
          if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
            MeasStd := MeasStd + Gamma[s - StateOffset[p], p] * Sqr(y - Amps[s - MinState]);
            Sum     := Sum     + Gamma[s - StateOffset[p], p];
          end;
        end;
        if prStoppedCount >= 3 then Exit;
      end;
      if Sum <> 0 then MeasStd := Sqrt(MeasStd / Sum);

    end;//Reest MeasStd ?

  end;

  procedure ReestMeasStdBW;
  var
    p, s :Integer;
  begin

    if prParams.StaircaseMeasStdReest then begin

      Sum     := 0;
      MeasStd := 0;
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        for s := MinState + Middle to MaxState - Middle do begin
          if ((s - StateOffset[p]) >= 0) and ((s - StateOffset[p]) < HalfStateCount) then begin
            MeasStd := MeasStd + Gamma[s - StateOffset[p], p] * Sqr(y - Amps[s - MinState]);
            Sum     := Sum     + Gamma[s - StateOffset[p], p];
          end;
        end;
        if prStoppedCount >= 3 then Exit;
      end;
      if Sum <> 0 then MeasStd := Sqrt(MeasStd / Sum);

    end;//Reest MeasStd ?

  end;

  procedure ReestJumpAmpStdCombinations;
  var
    l, s, ss :Integer;
  begin

    //First calculate the mean...
    //JumpAmp

    if prParams.StaircaseJumpAmpReestimate then begin

      OldJumpAmp := JumpAmp;
      OldJumpStd := JumpStd;
      JumpAmp    := 0;
      Sum        := 0;
      for s := MinState to MaxState do begin
        if (Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
          for ss := s + 1 to MaxState do begin
            if (Gamm[ss - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
              ThisJump  := (Amps[ss - MinState] - Amps[s - MinState]);
              dl        := ss - s;
              im        := dl;
              pm        := 0;
              for l := dl - prParams.StaircaseJumpAmpPlsMin to dl + prParams.StaircaseJumpAmpPlsMin do begin
                pc    := Gaussian(l * OldJumpAmp, OldJumpStd, ThisJump);
                if pc > pm then begin
                  im    := l;
                  pm    := pc;
                end;
              end;
              dl        := im;
              if dl <> 0 then begin
                JumpAmp := JumpAmp +
                           (ThisJump / dl)  *
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
                           //(Gamm[ss - MinState] * Gamm[s - MinState]);
                Sum     := Sum     +
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
                           //(Gamm[ss - MinState] * Gamm[s - MinState]);
              end;

            end;
          end;
        end;
      end;
      if Sum <> 0 then JumpAmp := JumpAmp / Sum;
      JumpAmps[0] := JumpAmp;

    end;


    //...and then the std
    //JumpStd
    if prParams.StaircaseJumpStdReest then begin

      JumpStd    := 0;
      Sum        := 0;
      for s := MinState to MaxState do begin
        if (Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
          for ss := s + 1 to MaxState do begin
            if (Gamm[ss - MinState] >= MIN_REPR_LEVEL_GAMM) then begin
              ThisJump  := (Amps[ss - MinState] - Amps[s - MinState]);
              dl        := ss - s;
              im        := dl;
              pm        := 0;
              for l := dl - prParams.StaircaseJumpAmpPlsMin to dl + prParams.StaircaseJumpAmpPlsMin do begin
                pc    := Gaussian(l * OldJumpAmp, OldJumpStd, ThisJump);
                if pc > pm then begin
                  im    := l;
                  pm    := pc;
                end;
              end;
              dl        := im;
              if dl = 1 {<> 0} then begin
                JumpStd := JumpStd +
                           Sqr((ThisJump / dl) - JumpAmp) *
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
                Sum     := Sum +
                           (Sqr(Gamm[ss - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[ss - MinState]) + Sqr(Gamm[s - MinState])));
              end;

            end;
          end;
        end;
      end;
      if Sum <> 0 then JumpStd := Sqrt(JumpStd / Sum);
      JumpStdReest := JumpStd;

    end;//Reest JumpStd ?

  end;

  procedure ReestJumpAmpStdConsecutive;
  var
    j, s :Integer;
  begin

    if prParams.StaircaseJumpAmpReestimate then begin

      //JumpAmp
      if prJumpAmpsCount = 1 then begin
      //1

        JumpAmp := 0;
        Sum         := 0;
        for s := MinState to MaxState - 1 do begin
          if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
            JumpAmp := JumpAmp +
                       (Amps[s + 1 - MinState] - Amps[s - MinState]) *
                       (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));

            Sum     := Sum     +
                       (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
          end;
        end;
        if Sum <> 0 then JumpAmp := JumpAmp / Sum;
        JumpAmps[0] := JumpAmp;

      end else begin
      //2

        JumpAmp := 0;
        for j := 0 to prJumpAmpsCount - 1 do begin

          JumpAmps[j] := 0;
          Sum         := 0;
          for s := MinState to MaxState - 1 do begin
            if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
              //What is RefState here ?!
              if ((s - RefState) mod prJumpAmpsCount) = j then begin
                JumpAmps[j] := JumpAmps[j] +
                               (Amps[s + 1 - MinState] - Amps[s - MinState]) *
                               (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
                Sum         := Sum         +
                               (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
              end;
            end;
          end;
          if Sum <> 0 then JumpAmps[j] := JumpAmps[j] / Sum;
          JumpAmp := JumpAmp + JumpAmps[j];
        end;

        //Switch JumpAmps, if necessary - may need to get rid of this !!!
        p   := 0;
        s   := 0;
        im  := s;
        Sum := Gamma[s, p];
        for s := 1 to HalfStateCount - 1 do begin
          if Gamma[s, p] > Sum then begin
            im := s;
            Sum := Gamma[s, p];
          end;
        end;
        if im = Middle + 1 then begin
          Sum         := JumpAmps[1];
          JumpAmps[1] := JumpAmps[0];
          JumpAmps[0] := Sum;
        end;

      end;//JumpAmpsCount

    end;//Reest JumpAmp ?

    //JumpStd
    if prParams.StaircaseJumpStdReest then begin
      JumpStd := 0;
      Sum     := 0;
      for s := MinState to MaxState - 1 do begin
        if (Gamm[s - MinState] <> 0) and (Gamm[s + 1 - MinState] <> 0) then begin
          JumpStd := JumpStd +
                     Sqr((Amps[s + 1 - MinState] - Amps[s - MinState]) - JumpAmp) *
                     (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
          Sum     := Sum     +
                     (Sqr(Gamm[s + 1 - MinState] * Gamm[s - MinState]) / (Sqr(Gamm[s + 1 - MinState]) + Sqr(Gamm[s - MinState])));
        end;
      end;
      if Sum <> 0 then JumpStd := Sqrt(JumpStd / Sum);
      JumpStdReest := JumpStd;
    end;

  end;

  function GetThisAmp(State, Point, RState :Integer) :Double;
  begin

    try
      if prJumpAmpsCount = 1 then begin
          Result := JumpAmp     * (State + StateOffset[Point] - RState);
      end else begin
        if Odd(RState) then begin
          Result := JumpAmp     * ((State + StateOffset[Point] - RState) div prJumpAmpsCount) +
                    JumpAmps[1] * ((State + StateOffset[Point] - RState) mod prJumpAmpsCount);
        end else begin
          Result := JumpAmp     * ((State + StateOffset[Point] - RState) div prJumpAmpsCount) +
                    JumpAmps[0] * ((State + StateOffset[Point] - RState) mod prJumpAmpsCount);
        end;
      end;
    except
      Result := 0;
    end;

  end;

  //Calculate the LL after re-estimation
  function GetNewLL :Double;
  var
    p, s, ss :Integer;
  begin

    Result := 0;

    //Mark the weakly represented levels
    ReAllocMem(LevelExists, SizeOf(Integer) * (MaxState - MinState + 1));
    for s := MinState to MaxState do begin
      //Testing
      //if Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM then begin
        LevelExists[s - MinState] := 1;
      //end else begin
      //  LevelExists[s - MinState] := 0;
      //end;
    end;

    //Initialize the first point
    p      := 0;
    y      := Data[p] * Scale + Off;

    Rho[p] := 0;
    for s := 0 to HalfStateCount - 1 do begin

      Gauss[s, p] := Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y);
      //p(y|s0) * p(s0)
      Alpha[s, p] := LevelExists[s + StateOffset[p] - MinState] *
                     Prob0[s] *
                     Gauss[s, p];
      Rho[p]      := Rho[p] + Alpha[s, p];

    end;

    Result := Result + SafeLog(Rho[p]);

    //Normalize Alpha
    for s := 0 to HalfStateCount - 1 do begin
      Alpha[s, p] := Alpha[s, p] / Rho[p];
    end;

    //Loop for every point
    for p := 1 to DataCount - 1 do begin

      y      := Data[p] * Scale + Off;
      Rho[p] := 0;

      for s := 0 to HalfStateCount - 1 do begin

        Gauss[s, p] := Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y);
        Alpha[s, p] := 0;

        for ss := 0 to HalfStateCount - 1 do begin

          MakeAStates(lState, rState, ss, s, p);
          Alpha[s, p] := Alpha[s, p] +
                         AMatrix[lState, rState] *
                         Alpha[ss, p - 1];

        end;//of ss

        Alpha[s, p] := LevelExists[s + StateOffset[p] - MinState] *
                       Alpha[s, p] *
                       Gauss[s, p];
        Rho[p]      := Rho[p] + Alpha[s, p];

      end;//of s

      //Normalize Alpha
      for s := 0 to HalfStateCount - 1 do begin
        Alpha[s, p] := Alpha[s, p] / Rho[p];
      end;

      Result := Result + SafeLog(Rho[p]);

      if prStoppedCount >= 3 then Exit;

    end;//of p

    //Add a term for Jumps:
    //For equal steps only
    //Testing
    if prParams.StaircaseLLAddJump then begin
      if prJumpAmpsCount = 1 then begin
        for s := MinState to MaxState - 1 do begin
          Result := Result + SafeLog(Gaussian(JumpAmp, JumpStd, Amps[s + 1 - MinState] - Amps[s - MinState]));
        end;
      end;
    end;


  end;

  procedure DoMovies;
  var
    s :Integer;
  begin

    MovieData       := MovieTree['Alpha'];
    MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, HalfStateCount, DataCount );
    for s := 0 to HalfStateCount - 1 do begin
      Move(Alpha[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
    end;
    MovieData       := MovieTree['Gamma'];
    MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, HalfStateCount, DataCount );
    for s := 0 to HalfStateCount - 1 do begin
      Move(Gamma[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
    end;
    MovieData       := MovieTree['MovieOffset'];
    MovieData.SetupNumData( NODE_DATATYPE_LONGINT, DataCount, 1 );
    Move(StateOffset^, MovieData.DataPtr^, SizeOf(Integer) * DataCount);

    if prParams.BaselineTrack then begin
      MovieData       := MovieTree['BaseX'];
      MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1{#col} );
      Move(BaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
      MovieData       := MovieTree['BaseP'];
      MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
      Move(BaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
      MovieData       := MovieTree['SBaseX'];
      MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
      Move(SBaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
      MovieData       := MovieTree['SBaseP'];
      MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
      Move(SBaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
    end;

    prMovieIteration  := Iteration;
    Synchronize(prMakeMovie_Staircase);

  end;//DoMovies

  procedure PrintQMatrix;
  var
    s, ss :Integer;
  begin
    //Print Q matrix as the logarithm of the A matrix ......................
    MatrixLogarithm(AMatrix, Epsilon[0], IDLBaseChunk.Sampling, prStateCount, 100, 1e-4);
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        if TPMDouble(Epsilon[0])[s, ss] < 0 then begin
          TPMDouble(Epsilon[0])[s, ss] := 0;
          TPMDouble(Epsilon[0])[ss, s] := 0;
        end;
      end;
    end;
    prConsoleMsg := #13#10'Big Q Matrix:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        if (s = ss) or (prQ[s, ss] = 0)then begin
          prConsoleMsg := prConsoleMsg + #9'...           ';
        end else begin
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [TPMDouble(Epsilon[0])[s, ss]]);
        end;
      end;
      prConsoleMsg := prConsoleMsg + #13#10;
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    if not prParams.QuietOutput then Synchronize(prWriteNewConsole);

    //Print Q matrix as the logarithm of the A matrix ......................
    TPMDouble(Epsilon[0])[0, 0] := AMatrix[0, 0];
    TPMDouble(Epsilon[0])[0, 1] := 1 - AMatrix[0, 0];
    TPMDouble(Epsilon[0])[1, 1] := AMatrix[prStateCount - 1, prStateCount - 1];
    TPMDouble(Epsilon[0])[1, 0] := 1 - AMatrix[prStateCount - 1, prStateCount - 1];

    MatrixLogarithm(Epsilon[0], Epsilon[0], IDLBaseChunk.Sampling, 2, 100, 1e-4);
    for s := 0 to {prStateCount -} 1 do begin
      for ss := 0 to {prStateCount -} 1 do begin
        if TPMDouble(Epsilon[0])[s, ss] < 0 then begin
          TPMDouble(Epsilon[0])[s, ss] := 0;
          TPMDouble(Epsilon[0])[ss, s] := 0;
        end;
      end;
    end;
    prConsoleMsg := #13#10'Small (LeftTop-RightBottom) Q Matrix:'#13#10;
    for s := 0 to {prStateCount -} 1 do begin
      for ss := 0 to {prStateCount -} 1 do begin
        if (s = ss) or (prQ[s, ss] = 0)then begin
          prConsoleMsg := prConsoleMsg + #9'...           ';
        end else begin
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [TPMDouble(Epsilon[0])[s, ss]]);
        end;            end;
      prConsoleMsg := prConsoleMsg + #13#10;
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    if not prParams.QuietOutput then Synchronize(prWriteNewConsole);

  end;//PrintQMatrix

  procedure ReestJumpAmpStdMeasStdBaseAmpI;
  var
    s :Integer;
  begin

    //Update JumpAmps, JumpStd and MeasStd
    if          (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_CONSECUTIVE) {or (prJumpAmpsCount > 1)} then begin

      ReestJumpAmpStdConsecutive;
      ReestMeasStd;

    end else if (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_COMBINATIONS)                         then begin

      ReestJumpAmpStdCombinations;
      ReestMeasStd;

    end else if (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_CONSTRAINED)                          then begin

      ReestJumpAmpStdConstrained;
      ReestMeasStd;

    end;

    //Update BaseAmp
    s       := 0;
    p       := 0;
    im      := s;
    Sum     := Gamma[s, p];
    for s := 1 to HalfStateCount - 1 do begin
      if Gamma[s, p] > Sum then begin
        im  := s;
        Sum := Gamma[s, p];
      end;
    end;
    BaseAmp := Amps[im + StateOffset[p] - MinState];

    //Update Baseline start
    if prParams.BaselineTrack then begin
      BaseX[0] := BaseAmp;
      BaseP[0] := SBaseP[0];
    end;

  end;//ReestJumpAmpStdMeasStdBaseAmpI

  procedure ReestJumpAmpStdMeasStdBaseAmpII;
//  var
//    s :Integer;
  begin

    //Update JumpAmps, JumpStd and MeasStd
    if          (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_CONSECUTIVE) {or (prJumpAmpsCount > 1)} then begin

      ReestJumpAmpStdConsecutive;
      ReestMeasStdBW;

    end else if (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_COMBINATIONS)                         then begin

      ReestJumpAmpStdCombinations;
      ReestMeasStdBW;

    end else if (prParams.StaircaseJumpAmpReestMethod = JUMPAMP_CONSTRAINED)                          then begin

      ReestJumpAmpStdConstrained;
      ReestMeasStdBW;

    end;

    //Update BaseAmp
    BaseAmp := Amps[Middle];


    //Update BaseAmp
    {
    s       := 0;
    p       := 0;
    im      := s;
    Sum     := Gamma[s, p];
    for s := 1 to HalfStateCount - 1 do begin
      if Gamma[s, p] > Sum then begin
        im  := s;
        Sum := Gamma[s, p];
      end;
    end;
    BaseAmp := Amps[im + StateOffset[p] - MinState];
    }


  end;//ReestJumpAmpStdMeasStdBaseAmpII

  //As it is, it only works for irreversible staircase data
  procedure DoMergeDwells;
  var
    Merge    :Boolean;
    Sel      :TQUB_FirstLast;
    ix       :Integer;
    p, s, ss :Integer;
  begin

    //if prJumpAmpsCount > 1 then Exit;

    s := MinState;// + HalfStateCount div 2;//Testing
    repeat

      if prStoppedCount >= 3 then Exit;

      Merge := False;

      if prJumpAmpsCount = 1 then begin
        Merge := Abs(Amps[s + 1 - MinState] - Amps[s - MinState]) < prParams.StaircaseMergeDwellsPercent*JumpAmp;
      end else begin
        if s < (MaxState - 1) then begin
          Merge := Abs(Amps[s + 2 - MinState] - Amps[s - MinState]) < prParams.StaircaseMergeDwellsPercent*(JumpAmps[0] + JumpAmps[1]);
          //if not Merge then begin
          //  Merge := Abs(Amps[s + 1 - MinState] - Amps[s - MinState]) < prParams.StaircaseMergeDwellsPercent*Min(JumpAmps[0], JumpAmps[1]);
          //end;
        end;
      end;

      if Merge then begin
      //Merge!

        //Collapse StateOffset
        //Find First/Last of ss
        Sel.First := -1;
        Sel.Last  := -1;
        for p := 0 to DataCount - 1 do begin
          ix := 0;
          for ss := 1 to HalfStateCount - 1 do begin
            if Gamma[ss, p] > Gamma[ix, p] then begin
              ix := ss;
            end;
          end;
          if (ix + StateOffset[p]) = (s + 1) then begin
            Sel.First := p;
            Break;
          end;
          if prStoppedCount >= 3 then Exit;
        end;
        if Sel.First >= 0 then begin
          Sel.Last := Sel.First;
          for p := Sel.Last + 1 to DataCount - 1 do begin
            ix := 0;
            for ss := 1 to HalfStateCount - 1 do begin
              if Gamma[ss, p] > Gamma[ix, p] then begin
                ix := ss;
              end;
            end;
            if (ix - StateOffset[p]) <> (s + 1) then begin
              Break;
            end else begin
              Sel.Last := p;
            end;
            if prStoppedCount >= 3 then Exit;
          end;
        end;

        //We have First/Last
        if Sel.First >= 0 then begin
          for p := Sel.First to DataCount - 1 do begin
            Dec(StateOffset[p]);
          end;
        end;

        //Collapse Amps and Gamm
        Amps[s - MinState] := (Gamm[s - MinState]*Amps[s - MinState] + Gamm[s + 1 - MinState]*Amps[s + 1 - MinState]) / (Gamm[s - MinState] + Gamm[s + 1 - MinState]);
        Gamm[s - MinState] := Gamm[s - MinState] + Gamm[s + 1 - MinState];
        for ss := s + 1 to MaxState - 1 do begin
          Amps[ss - MinState] := Amps[ss + 1 - MinState];
          Gamm[ss - MinState] := Gamm[ss + 1 - MinState];
        end;

        //Inc(s);
        Dec(MaxState);
      end;
      Inc(s);
    until s > (MaxState - 1);
    //Testing. It shouldn't be necessary once the constrained re-estimation for unequal steps is implemented
    //until s > (MaxState - 1 - HalfStateCount div 2);

  end;

  //As it is, it only works for irreversible staircase data
  procedure DoSplitDwells;
  var
    Split    :Boolean;
    OldSize  :Integer;
    Sel      :TQUB_FirstLast;
    ix       :Integer;
    p, s, ss :Integer;
  begin

    //if prJumpAmpsCount > 1 then Exit;

    OldSize := MaxState - MinState + 1;

    s := MinState;// + HalfStateCount div 2;//Testing
    repeat

      if prStoppedCount >= 3 then Exit;

      Split := False;

      if prJumpAmpsCount = 1 then begin
        Split := Abs(Amps[s + 1 - MinState] - Amps[s - MinState]) > (1 + prParams.StaircaseSplitDwellsPercent)*JumpAmp;
      end else begin
        if s < (MaxState - 1) then begin
          Split := Abs(Amps[s + 2 - MinState] - Amps[s - MinState]) > (1 + prParams.StaircaseSplitDwellsPercent)*(JumpAmps[0] + JumpAmps[1]);
        end;
      end;

      if Split then begin
      //Split!

        //Resize Amps and Gamms, if necessary
        if OldSize < (MaxState - MinState + 1 + 1) then begin
          Inc(OldSize, 10);
          ReAllocMem(Amps, SizeOf(Double) * OldSize);
          ReAllocMem(Gamm, SizeOf(Double) * OldSize);
        end;

        //Insert StateOffset
        //Find First/Last of ss
        Sel.First := -1;
        Sel.Last  := -1;
        for p := 0 to DataCount - 1 do begin
          ix := 0;
          for ss := 1 to HalfStateCount - 1 do begin
            if Gamma[ss, p] > Gamma[ix, p] then begin
              ix := ss;
            end;
          end;
          if (ix + StateOffset[p]) = s then begin
            Sel.First := p;
            Break;
          end;
          if prStoppedCount >= 3 then Exit;
        end;
        if Sel.First >= 0 then begin
          Sel.Last := Sel.First;
          for p := Sel.Last + 1 to DataCount - 1 do begin
            ix := 0;
            for ss := 1 to HalfStateCount - 1 do begin
              if Gamma[ss, p] > Gamma[ix, p] then begin
                ix := ss;
              end;
            end;
            if (ix - StateOffset[p]) <> s then begin
              Break;
            end else begin
              Sel.Last := p;
            end;
            if prStoppedCount >= 3 then Exit;
          end;
        end;

        //We have First/Last
        if Sel.First >= 0 then begin
          for p := Sel.Last + 1 to DataCount - 1 do begin
            Inc(StateOffset[p]);
          end;
        end;

        //Move up Amps and Gamm
        for ss := MaxState + 1 downto s + 2 do begin
          Amps[ss - MinState] := Amps[ss - 1 - MinState];
          Gamm[ss - MinState] := Gamm[ss - 1 - MinState];
        end;
        Amps[s + 1 - MinState] := (Amps[s - MinState] + Amps[s + 2 - MinState]) / 2;
        Gamm[s + 1 - MinState] := 1;

        //Inc(s);
        Inc(MaxState);
      end;
      Inc(s);
    until s > (MaxState - 1);
    //Testing. It shouldn't be necessary once the constrained re-estimation for unequal steps is implemented
    //until s > (MaxState - 1 - HalfStateCount div 2);

  end;

  procedure MakeState(State :TPAInteger);
  var
    Sum   :Double;
    p, s  :Integer;
  begin
    for p := 0 to DataCount - 1 do begin
      State[p] := 0;
      Sum      := Gamma[0, p];
      for s := 1 to HalfStateCount - 1 do begin
        if Gamma[s, p] > Sum then begin
          State[p] := s;
          Sum      := Gamma[s, p];
        end;
      end;
      if prStoppedCount >= 3 then Exit;
    end;
  end;

  //Calculate forward and backward Jump rates (assuming one state per position
  procedure CalculateJfJb;
  var
    State   :TPAInteger;
    //Minl    :Integer;
    fc, bc  :Integer;
    d, p    :Integer;
  begin
    State := IAlloc1(DataCount);
    try
      MakeState(State);
      Min1 := State[0] + StateOffset[0];
      for p := 1 to DataCount - 1 do begin
        if (State[p] + StateOffset[p]) < Min1 then begin
          Min1 := State[p] + StateOffset[p];
        end;
        if prStoppedCount >= 3 then Exit;
      end;
      for p := 0 to DataCount - 1 do begin
        Inc(State[p], StateOffset[p] - Min1);
      end;
      //Count
      fc := 0;
      bc := 0;
      for p := 0 to DataCount - 2 do begin
        d := State[p + 1] - State[p];
        if          d > 0 then begin
          Inc(fc, d);
        end else if d < 0 then begin
          Inc(bc, d);
        end;
        if prStoppedCount >= 3 then Exit;
      end;
      //It's not accurate, since missing events may be mixture of f and b
      Jf := fc / ((DataCount - 1) * IDLBaseChunk.Sampling);
      Jb := bc / ((DataCount - 1) * IDLBaseChunk.Sampling);
    finally
      FreeMem(State);
    end;
  end;

  //Make the state vector
begin



  //Epsilon might be just 2-dimensional: add Epsilon for each point



  Result              := False;


  //Nullify pointers ...........................................................
  Alpha               := nil;
  Betta               := nil;
  Gamma               := nil;
  Gauss               := nil;
  Rho                 := nil;
  Epsilon             := nil;

  State               := nil;
  StateOffset         := nil;

  AMatrix             := nil;

  Prob0               := nil;

  Amps                := nil;
  Gamm                := nil;

  BaseX               := nil;
  BaseP               := nil;

  SBaseX              := nil;
  SBaseP              := nil;
  SBaseQ              := nil;

  LevelExists         := nil;

  JumpAmps            := nil;

  JumpAmp             := 0;
  MeasStd             := 0;



  //Signal Running status ......................................................
  Chunk               := TQUB_Thread_PChunk(IdlBaseChunk);
  Chunk.Status        := CHUNK_STATUS_RUNNING;


  //Movies
  if prMovie.Make then begin

    MovieTree := TNode.Create('IBMovie');
    PythonThread.BeginPython;
    PyNode_Publish(MovieTree, 'IBMovie'); // make tree available in Python as QUB.IBMovie
    PythonThread.EndPython;
    PythonThread.DonePython;

  end;



  try//finally



    //Get data parameters ......................................................
    //   - Data, DataCount
    Data              := IdlBaseChunk.Data;
    DataCount         := Chunk.Sel.Last - Chunk.Sel.First + 1;
    //   - Scale
    Scale             := 1 / IdlBaseChunk.Scaling;
    Off               := IdlBaseChunk.Offset;


    HalfStateCount    := prStateCount div 2;
    Middle            := (HalfStateCount div 2) + 1 - 1;



    //Movies
    if prMovie.Make then begin

      MovieTree['Scaling'].SetData(IdlBaseChunk.Scaling);
      MovieTree['Offset'].SetData(IdlBaseChunk.Offset);
      MovieTree['DataCount'].SetData(DataCount);
      MovieTree['HalfStateCount'].SetData(HalfStateCount);
      MovieTree['JumpAmp'].SetData(prJumpAmp);
      MovieTree['BaseAmp'].SetData(prBaseAmp);

    end;



    //Allocate memory ..........................................................


    //   - Vectors
    Alpha             := DAlloc2(HalfStateCount, DataCount);
    Betta             := DAlloc2(HalfStateCount, DataCount);
    Gamma             := DAlloc2(HalfStateCount, DataCount);
    Gauss             := DAlloc2(HalfStateCount, DataCount);
    Rho               := DAlloc1(DataCount);
    Epsilon           := PAlloc1(DataCount);
    for p := 0 to DataCount - 1 do begin
      Epsilon[p]      := DAlloc2(prStateCount, prStateCount);
    end;

    //   - State - the idealized state, at each t;
    State             := IAlloc1(DataCount);
    //   - StateOffset - the offset of the state vector, at each t;
    StateOffset       := IAlloc1(DataCount);


    //   - A matrix
    AMatrix           := DAlloc2(prStateCount, prStateCount);


    //   - Prob0 - initial probabilities;
    Prob0             := DAlloc1(HalfStateCount);


    if prParams.BaselineTrack then begin
      //   - BaseX|P - baseline state|variance, for each data point
      BaseX           := DAlloc1(DataCount);
      BaseP           := DAlloc1(DataCount);
      //   - SBaseX|P|Q - smoothed baseline state, variance and BaseVar, for each data point
      SBaseX          := DAlloc1(DataCount);
      SBaseP          := DAlloc1(DataCount);
      SBaseQ          := DAlloc1(DataCount);
    end;


    JumpAmps          := DAlloc1(prJumpAmpsCount);



    //Initialize variables .....................................................


    //   - A matrix
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss]    := Abs(prA[s, ss]);//to prevent some occasional negative entries
        Sum               := Sum + AMatrix[s, ss];
      end;
      for ss := 0 to prStateCount - 1 do begin //re-normalize
        if Sum <> 0 then AMatrix[s, ss]    := AMatrix[s, ss] / Sum;
      end;
    end;



    //   - Prob0
    //Start with equal probability and let the Gaussian make the difference
    for s := 0 to HalfStateCount - 1 do begin
      Prob0[s]     := 1 / HalfStateCount;
    end;



    //   - JumpAmp, JumpStd, MeasStd, BaseAmp, BaseStd
    JumpAmp        := prJumpAmp;
    for s := 0 to prJumpAmpsCount - 1 do begin
      JumpAmps[s]  := prJumpAmps[s];
    end;

    JumpStd        := prParams.StaircaseJumpStd;
    JumpStdReest   := JumpStd;

    MeasStd        := prMeasStd;
    MeasVar        := Sqr(MeasStd);

    BaseAmp        := prBaseAmp;

    BaseStd        := prParams.BaselineStd;
    BaseVar        := Sqr(BaseStd);


    //Initialize BaseX[0] and BaseP[0]
    if prParams.BaselineTrack then begin
      p            := 0;
      y            := Data[p] * Scale + Off;
      BaseX[p]     := BaseAmp;
      if Abs(BaseAmp - y) > 3*MeasStd then begin
        BaseX[p]   := y;
      end;
      BaseP[p]     := Sqr(MeasStd);
    end;


    //StateOffset
    StateOffset[0] := - Middle;
    RefState       := 0;

    MinState       := 0;
    MaxState       := 0;


    //Print starting values ....................................................
    prConsoleMsg := 'Starting values:'#13#10;
    prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
      end;
      prConsoleMsg := prConsoleMsg + #13#10;
    end;
    prConsoleMsg := prConsoleMsg + Format('Jump amp = %-14.6g, Meas std = %-14.6g, Base amp = %-14.6g'#13#10, [JumpAmp, MeasStd, BaseAmp]);
    for c := 0 to prJumpAmpsCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [c, JumpAmps[c]]);
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


    try


      OldLL := 0;


      //**********************************************************************//
      //                                                                      //
      //                              S t e p   I                             //
      //                                                                      //
      //                    D w e l l    D e t e c t i o n                    //
      //                                                                      //
      //**********************************************************************//
      for Iteration := 1 to prParams.MaxIter do begin


        if prStoppedCount >= 3 then Exit;

        NewLL      := 0;

        RefState   := 0;

        if prParams.BaselineTrack then begin
          RefAmp   := BaseX[0];
        end else begin
          RefAmp   := BaseAmp;//This is always the case because BaseX is reestimated as BaseAmp
        end;


        //Variances
        MeasVar    := Sqr(MeasStd);
        JumpVar    := Sqr(JumpStd);



        //********************************************************************//
        //                                                                    //
        //                            F o r w a r d                           //
        //                                                                    //
        //********************************************************************//

        //Makes the assumption that the first point IS A (not B)
        //which is wrong. It may get past it. Or use a combination of amps ?

        //Initialize the first point
        p         := 0;
        y         := Data[p] * Scale + Off;
        Rho[p]    := 0;


        for s := 0 to HalfStateCount - 1 do begin

          //ThisAmp is what the data should be, in absence of baseline or baseamp
          ThisAmp := GetThisAmp(s, p, RefState);

          Gauss[s, p] := Gaussian(ThisAmp, MeasStd, y - RefAmp);

          //p(y|s0) * p(s0)
          Alpha[s, p] := Prob0[s] * Gauss[s, p];
          Rho[p]      := Rho[p] + Alpha[s, p];

        end;

        NewLL := NewLL + SafeLog(Rho[p]);


        //Normalize Alpha
        for s := 0 to HalfStateCount - 1 do begin
          Alpha[s, p] := Alpha[s, p] / Rho[p];
        end;


        //Loop for every point
        for p := 1 to DataCount - 1 do begin

          if prStoppedCount >= 3 then Exit;

          if prParams.BaselineTrack then begin
            BaseX[p] := BaseX[p-1];
            BaseP[p] := BaseP[p-1];
          end;

          StateOffset[p] := StateOffset[p - 1];

          y              := Data[p] * Scale + Off;

          CanQuit        := True;
          BaseEstCount   := 0;


          //Baseline EM loop
          repeat


            if prStoppedCount >= 3 then Exit;

            for Offset := 1 to 2 do begin


              Rho[p] := 0;


              //for each state
              for s := 0 to HalfStateCount - 1 do begin

                //ThisAmp is what the data should be, in absence of baseline or baseamp
                ThisAmp     := GetThisAmp(s, p, RefState);

                Gauss[s, p] := Gaussian(ThisAmp, MeasStd, y - RefAmp);

                Alpha[s, p] := 0;

                for ss := 0 to HalfStateCount - 1 do begin

                  MakeAStates(lState, rState, ss, s, p);

                  Alpha[s, p] := Alpha[s, p] +
                                 AMatrix[lState, rState] *
                                 Alpha[ss, p - 1]
                                 ;

                end;//of ss

                Alpha[s, p] := Gauss[s, p] * Alpha[s, p];

                Rho[p]      := Rho[p]      + Alpha[s, p];

              end;//of s


              //Normalize Alpha
              for s := 0 to HalfStateCount - 1 do begin
                Alpha[s, p] := Alpha[s, p] / Rho[p];
              end;


              //Find the Max
              s   := 0;
              im  := s;
              Sum := Alpha[s, p];
              for s := 1 to HalfStateCount - 1 do begin
                if Alpha[s, p] > Sum then begin
                  im := s;
                  Sum := Alpha[s, p];
                end;
              end;
              //Do StateOffset
              if Offset = 1 then begin
                StateOffset[p] := StateOffset[p - 1] + (im - Middle);
              end;


            end;//Offset



            //Update baseline
            if prParams.BaselineTrack then begin

              //ThisAmp is a linear combination of states
              ThisAmp := 0;
              for s := 0 to HalfStateCount - 1 do begin

                //Alpha is already normalized here
                ThisAmp := ThisAmp + Alpha[s, p] * GetThisAmp(s, p, RefState);

              end;

              //When a transition is likely, increase BaseVar, so the tracking occurs faster
              ThisStd := 0;
              for s := 0 to HalfStateCount - 1 do begin
                for ss := 0 to HalfStateCount - 1 do begin
                  dl := (s + StateOffset[p]) - (ss + StateOffset[p-1]);
                  ThisStd := ThisStd + Abs(dl) * Alpha[s,  p] *  Alpha[ss, p-1];
                end;
              end;

              //ThisStd := BaseVar + prParams.StaircaseBaseJumpVar * JumpVar * ThisStd;
              //New:
              ThisStd := BaseVar + prParams.StaircaseBaseJumpVar * JumpVar * ThisStd * 0.5;

              KalmanForward(BaseX[p], BaseP[p], BaseX[p-1], BaseP[p-1], MeasVar, ThisStd{BaseVar}, (y - ThisAmp){BaseData});

              Inc(BaseEstCount);

              CanQuit := (Abs((RefAmp - BaseX[p] + 1e-10) / (RefAmp + 1e-10)) < BASE_CONV) or (BaseEstCount > prParams.BaselineFKCount);

              RefAmp  := BaseX[p];

            end;


          until(CanQuit);


          NewLL := NewLL + SafeLog(Rho[p]);


          //Update RefState
          if (prParams.StaircaseUpdateRefState) and (prParams.BaselineTrack) then begin
            if Abs(im + StateOffset[p] - RefState) > prParams.StaircaseUpdateRefStateEvery then begin
              ThisAmp := 0;
              for s := 0 to HalfStateCount - 1 do begin
                ThisAmp := ThisAmp + Alpha[s, p] * GetThisAmp(s, p, RefState);
              end;
              RefAmp   := RefAmp + ThisAmp;
              RefState := im + StateOffset[p];
              BaseX[p] := RefAmp;
            end;
          end;



        end;//of p



        //Backward step on Baseline
        if prParams.BaselineTrack then begin

          DoBaselineBack;

        end;//Backward on baseline



        //********************************************************************//
        //                                                                    //
        //                           B a c k w a r d                          //
        //                                                                    //
        //********************************************************************//

        //Betta
        DoBetta;

        //Gamma
        DoGamma;

        if prParams.StaircaseKineticsReest then begin

          //Epsilon
          DoEpsilon;

        end;//Reest Kinetics ?



        //********************************************************************//
        //                                                                    //
        //                              M O V I E S                           //
        //                                                                    //
        //********************************************************************//
        if prMovie.Make then begin

          DoMovies;

        end;



        //********************************************************************//
        //                                                                    //
        //                       R e - E s t i m a t i o n                    //
        //                                                                    //
        //********************************************************************//

        //Find Min/MaxState;
        FindMinMaxState;


        //Amps and Gamm
        DoAmpsGamm;
        DoJumpAmpNoVar;
        //DoJumpAmpLS(False, False);


        //Reestimate stepping parameters: JumpAmp, JumpStd, MeasStd and BaseAmp
        ReestJumpAmpStdMeasStdBaseAmpI;



        if prParams.StaircaseKineticsReest then begin

          //A Matrix
          ReestAMatrix;
          PrintQMatrix;

        end;//Reest Kinetics ?


        //Calculate LL with the reestimated parameters
        NewLL := GetNewLL;


        //Print current iteration values .......................................
        prConsoleMsg := Format(#13#10'Iteration %d'#13#10, [Iteration]);
        prConsoleMsg := prConsoleMsg + Format('LL: %14.10g'#13#10, [NewLL]);
        prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
        for s := 0 to prStateCount - 1  do begin
          for ss := 0 to prStateCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleMsg := prConsoleMsg + Format('Jump amp = %-14.6g Jump std = %-14.6g'#13#10'Meas std = %-14.6g'#13#10'Base amp = %-14.6g'#13#10, [JumpAmp, JumpStdReest, MeasStd, BaseAmp]);
        prConsoleMsg := prConsoleMsg + Format('Jump amp fixed = %-14.6g Base amp fixed = %-14.6g'#13#10, [JumpAmpFixed, BaseAmpFixed]);
        prConsoleMsg := prConsoleMsg + Format('Jump amp LS = %-14.6g'#13#10, [JumpAmpLS]);
        prConsoleMsg := prConsoleMsg + 'Jump amps:'#13#10;
        for c := 0 to prJumpAmpsCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [c, JumpAmps[c]]);
        end;
        if prParams.BaselineTrack then begin
          prConsoleMsg := prConsoleMsg + Format('BaseVar = %-14.6g BaseStd = %-14.6g'#13#10, [BaseVar, BaseStd]);
          prConsoleMsg := prConsoleMsg + Format('BaseX[0] = %-14.6g BaseP[0] = %-14.6g'#13#10, [SBaseX[0], SBaseP[0]]);
        end;
        prConsoleMsg := prConsoleMsg + 'Level, Amp, dAmp, Gamm:'#13#10;
        for c := MinState to MaxState do begin
          if c > MinState then begin
            prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g%s' + #9 + '%-14.6g'#13#10, [c, Amps[c - MinState], Amps[c - MinState] - Amps[c - 1 - MinState], Choice((Amps[c - MinState] - Amps[c - 1 - MinState]) < prParams.StaircaseMergeDwellsPercent*JumpAmp, 'L', Choice((Amps[c - MinState] - Amps[c - 1 - MinState]) > (1 + prParams.StaircaseSplitDwellsPercent)*JumpAmp, 'H', ' ')), Gamm[c - MinState]]);
          end else begin
            prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'               '#9'%-14.6g'#13#10, [c, Amps[c - MinState],                                              Gamm[c - MinState]]);
          end;
        end;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


        if prParams.QuietOutput and prParams.QuietOutputEveryIter then begin
          prConsoleMsg := Format('%d'#9'FBK', [Iteration]);
          for p := 0 to prJumpAmpsCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [JumpAmps[p]]);
          end;
          CalculateJfJb;
          {
          with TQUB_Thread_PChunk(IdlBaseChunk)^ do begin
            Jf := (MaxState - MinState - 2*Middle) / ((Sel.Last - Sel.First) * IDLBaseChunk.Sampling);
          end;
          }
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.10g'#9'%-14.10g'#9'%-14.6g', [JumpStdReest, MeasStd, BaseAmp, Jf, Jb, NewLL]);
          prConsoleCode := MESSAGE_MESSAGE;
          Synchronize(prWriteNewConsole);
        end;


        //Movies
        if prMovie.Make then begin

          MovieTree['JumpAmp'].SetData(JumpAmp);
          MovieTree['BaseAmp'].SetData(BaseAmp);

        end;



        if (prStopped) and (not prParams.StaircaseGlobalReest) then Break;


        //Check convergence
        if Abs(NewLL - OldLL) <= prParams.LLConv then Break;


        if prParams.Reestimation = IDLBASE_REEST_NONE then Break;


        OldLL := NewLL;


      end;//of Iteration




      //**********************************************************************//
      //                                                                      //
      //                            S t e p   I I                             //
      //                                                                      //
      //          B a u m - W e l c h     R e - e s t i m a t i o n           //
      //                                                                      //
      //**********************************************************************//
      if prParams.StaircaseReestBW then begin


        prConsoleMsg := #13#10'Second step: Baum-Welch reestimation'#13#10;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


        ReAllocMem(LevelExists, SizeOf(Integer) * (MaxState - MinState + 1));

        OldLL := 0;


        //Iterate
        for Iteration := 1 to prParams.MaxIter do begin


          if prStoppedCount >= 3 then Break;

          //Mark the weakly represented levels
          if (prParams.StaircaseMergeDwells) or (prParams.StaircaseSplitDwells) then begin
            ReAllocMem(LevelExists, SizeOf(Integer) * (MaxState - MinState + 1));
          end;
          for s := MinState to MaxState do begin
            //if Gamm[s - MinState] >= MIN_REPR_LEVEL_GAMM then begin
              LevelExists[s - MinState] := 1;
            //end else begin
            //  LevelExists[s - MinState] := 0;
            //end;
          end;


          NewLL := 0;



          //******************************************************************//
          //                                                                  //
          //                           F o r w a r d                          //
          //                                                                  //
          //******************************************************************//


          //Initialize the first point
          p      := 0;
          y      := Data[p] * Scale + Off;

          Rho[p] := 0;
          for s := 0 to HalfStateCount - 1 do begin

            Gauss[s, p] := Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y);
            //p(y|s0) * p(s0)
            Alpha[s, p] := {@@@}LevelExists[s + StateOffset[p] - MinState] *
                           Prob0[s] *
                           Gauss[s, p];
            Rho[p]      := Rho[p] + Alpha[s, p];

          end;

          NewLL := NewLL + SafeLog(Rho[p]);

          //Normalize Alpha
          for s := 0 to HalfStateCount - 1 do begin
            Alpha[s, p] := Alpha[s, p] / Rho[p];
          end;


          //Loop for every point
          for p := 1 to DataCount - 1 do begin

            if prStoppedCount >= 3 then Exit;

            y := Data[p] * Scale + Off;

            Rho[p] := 0;
            for s := 0 to HalfStateCount - 1 do begin

              Gauss[s, p] := Gaussian(Amps[s + StateOffset[p] - MinState], MeasStd, y);
              Alpha[s, p] := 0;

              for ss := 0 to HalfStateCount - 1 do begin

                MakeAStates(lState, rState, ss, s, p);
                Alpha[s, p] := Alpha[s, p] +
                               AMatrix[lState, rState] *
                               Alpha[ss, p - 1];

              end;//of ss

              Alpha[s, p] := {@@@}LevelExists[s + StateOffset[p] - MinState] *
                             Alpha[s, p] *
                             Gauss[s, p];
              Rho[p]      := Rho[p] + Alpha[s, p];

            end;//of s

            //Normalize Alpha
            for s := 0 to HalfStateCount - 1 do begin
              Alpha[s, p] := Alpha[s, p] / Rho[p];
            end;


            NewLL := NewLL + SafeLog(Rho[p]);


          end;//of p



          //******************************************************************//
          //                                                                  //
          //                          B a c k w a r d                         //
          //                                                                  //
          //******************************************************************//


          //Betta
          DoBettaII;


          //Gamma
          DoGamma;


          if prParams.StaircaseKineticsReest then begin

            //Epsilon
            DoEpsilon;

          end;//Reest Kinetics ?



          //******************************************************************//
          //                                                                  //
          //                     R e - E s t i m a t i o n                    //
          //                                                                  //
          //******************************************************************//


          //Amps and Gamm
          DoAmpsGammBW;


          ReestJumpAmpStdMeasStdBaseAmpII;


          if prParams.StaircaseKineticsReest then begin

            //A Matrix
            ReestAMatrix;

          end;//Reest Kinetics ?
          //********************************************************************



          //Calculate LL with the reestimated parameters
          //Testing
          NewLL := GetNewLL;



          //Print current iteration values .......................................
          prConsoleMsg := Format(#13#10'Second step Iteration %d'#13#10, [Iteration]);
          prConsoleMsg := prConsoleMsg + Format('LL: %14.10g'#13#10, [NewLL]);
          prConsoleMsg := prConsoleMsg + Format('Jump amp = %-14.6g Jump std = %-14.6g'#13#10'Meas std = %-14.6g'#13#10'Base amp = %-14.6g'#13#10, [JumpAmp, JumpStdReest, MeasStd, BaseAmp]);
          prConsoleMsg := prConsoleMsg + 'Level, Exists, Amp, Gamm:'#13#10;
          for c := MinState + Middle to MaxState - Middle do begin
            //prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [c, LevelExists[c - MinState], Amps[c - MinState], Gamm[c - MinState]]);
            if c > MinState then begin
              prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%d'#9'%-14.6g'#9'%-14.6g%s' + #9 + '%-14.6g'#13#10, [c, LevelExists[c - MinState], Amps[c - MinState], Amps[c - MinState] - Amps[c - 1 - MinState], Choice((Amps[c - MinState] - Amps[c - 1 - MinState]) < prParams.StaircaseMergeDwellsPercent*JumpAmp, 'L', Choice((Amps[c - MinState] - Amps[c - 1 - MinState]) > (1 + prParams.StaircaseSplitDwellsPercent)*JumpAmp, 'H', ' ')), Gamm[c - MinState]]);
            end else begin
              prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%d'#9'%-14.6g'#9'               '#9'%-14.6g'#13#10, [c, LevelExists[c - MinState], Amps[c - MinState],                                              Gamm[c - MinState]]);
            end;
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


          if prParams.QuietOutput and prParams.QuietOutputEveryIter then begin
            prConsoleMsg := Format('%d'#9'BW', [Iteration]);
            for p := 0 to prJumpAmpsCount - 1 do begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [JumpAmps[p]]);
            end;
            CalculateJfJb;
            {
            with TQUB_Thread_PChunk(IdlBaseChunk)^ do begin
              Jf := (MaxState - MinState - 2*Middle) / ((Sel.Last - Sel.First) * IDLBaseChunk.Sampling);
            end;
            }
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.10g'#9'%-14.10g'#9'%-14.6g', [JumpStdReest, MeasStd, BaseAmp, Jf, Jb, NewLL]);
            prConsoleCode := MESSAGE_MESSAGE;
            Synchronize(prWriteNewConsole);
          end;




          if (prStopped) and (not prParams.StaircaseGlobalReest) then Break;


          //Check convergence
          if Abs(NewLL - OldLL) <= prParams.LLConv then Break;


          OldLL := NewLL;



          //Testing
          if Iteration < prParams.MaxIter then begin
            if prParams.StaircaseMergeDwells then begin
              DoMergeDwells;
            end;
            if prParams.StaircaseSplitDwells then begin
              DoSplitDwells;
            end;
          end;


        end;//of Iteration


      end;//Second step: Baum-Welch reestimation




      //**********************************************************************//
      //                                                                      //
      //                             R e s u l t s                            //
      //                                                                      //
      //**********************************************************************//
      if not ((prStopped) and (not prParams.StaircaseGlobalReest)) then begin


        //Make the state vector
        State := IAlloc1(DataCount);
        for p := 0 to DataCount - 1 do begin
          State[p] := 0;
          Sum      := Gamma[0, p];
          for s := 1 to HalfStateCount - 1 do begin
            if Gamma[s, p] > Sum then begin
              State[p] := s;
              Sum := Gamma[s, p];
            end;
          end;
          if prStoppedCount >= 3 then Exit;
        end;


        //Which states are actually used
        Min1 := State[0] + StateOffset[0];
        Max1 := State[0] + StateOffset[0];
        for p := 1 to DataCount - 1 do begin
          if (State[p] + StateOffset[p]) < Min1 then begin
            Min1 := State[p] + StateOffset[p];
          end;
          if (State[p] + StateOffset[p]) > Max1 then begin
            Max1 := State[p] + StateOffset[p];
          end;
          if prStoppedCount >= 3 then Exit;
        end;

        UsedCount := Max1 - Min1 + 1;

        IdlBaseChunk.ClassCount := UsedCount;
        IdlBaseChunk.StateCount := prStateCount;
        IdlBaseChunk.Amps       := SAlloc1(UsedCount);
        IdlBaseChunk.Stds       := SAlloc1(UsedCount);
        IdlBaseChunk.AMatrix    := DAlloc2(prStateCount, prStateCount);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            IdlBaseChunk.AMatrix[s, ss] := AMatrix[s, ss];
          end;
        end;
        if (prParams.BaselineTrack) or (prParams.Reestimation <> IDLBASE_REEST_NONE) then begin
          for s := Min1 to Max1 do begin
            IdlBaseChunk.Amps[s - Min1]  := Amps[s - MinState];
            IdlBaseChunk.Stds[s - Min1]  := MeasStd;
          end;
        end else begin
          s := Min1;
          IdlBaseChunk.Amps[s - Min1]  := Amps[s - MinState];
          IdlBaseChunk.Stds[s - Min1]  := MeasStd;
          for s := Min1 + 1 to Max1 do begin
            IdlBaseChunk.Amps[s - Min1]  := Amps[Min1 - MinState] + JumpAmps[0] * (s - Min1);
            IdlBaseChunk.Stds[s - Min1]  := MeasStd;
          end;
        end;

        //Make Dwells
        for p := 0 to DataCount - 1 do begin
          Inc(State[p], StateOffset[p] - Min1);
        end;

        prMakeDwells(IdlBaseChunk, State, DataCount);


        //Make Baseline Nodes
        if (prParams.BaselineTrack) and (prParams.BaselineCorrect) then begin

          prBase2Nodes(IdlBaseChunk, SBaseX);

        end;


        //Make Stats and Histograms
        prMakeStats(IdlBaseChunk, SBaseX); // SBaseX is nil if not prParams.BaselineTrack


        IdlBaseChunk.LL         := NewLL;
        IdlBaseChunk.Iterations := Iteration - 1;


        if not prParams.StaircaseGlobalReest then begin
          prConsoleMsg := Format('%d', [TQUB_Thread_PChunk(IdlBaseChunk).Index]);
          for p := 0 to prJumpAmpsCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [JumpAmps[p]]);
          end;
          CalculateJfJb;
          {
          with TQUB_Thread_PChunk(IdlBaseChunk)^ do begin
            Jf := (MaxState - MinState - 2*Middle) / ((Sel.Last - Sel.First) * IDLBaseChunk.Sampling);
          end;
          }
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#9'%-14.6g'#9'%-14.10g'#9'%-14.10g'#9'%-14.6g'#9'%d', [JumpStdReest, MeasStd, BaseAmp, Jf, Jb, IdlBaseChunk.LL, IdlBaseChunk.Iterations]);
          prConsoleCode := MESSAGE_MESSAGE;
          Synchronize(prWriteNewConsole);
        end;


      end;//Results


    except


      Result            := True;

      Chunk.Status      := CHUNK_STATUS_ERROR;
      Chunk.Completion  := 0;


    end;


  finally



    //Free Memory ..............................................................
    Free2(TPMChar(Alpha));
    Free2(TPMChar(Betta));
    Free2(TPMChar(Gamma));
    Free2(TPMChar(Gauss));
    FreeMem(Rho);
    if Epsilon <> nil then begin
      for p := 0 to DataCount - 1 do begin
        Free2(TPMChar(Epsilon[p]));
      end;
    end;
    FreeMem(Epsilon);

    FreeMem(State);
    FreeMem(StateOffset);

    FreeMem(Prob0);

    FreeMem(Amps);
    FreeMem(Gamm);

    FreeMem(BaseX);
    FreeMem(BaseP);
    FreeMem(SBaseX);
    FreeMem(SBaseP);
    FreeMem(SBaseQ);

    FreeMem(LevelExists);


    //Signal new status
    if Chunk.Status <> CHUNK_STATUS_ERROR then begin

      Chunk.Status := CHUNK_STATUS_DONE;
      Chunk.Completion := 100;

      if (not prStopped) and (not prSkipped) then begin

        Inc(prTotCount);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            prTotA[s, ss] := prTotA[s, ss] + AMatrix[s, ss];
          end;
        end;
        for p := 0 to prJumpAmpsCount - 1 do begin
          prTotJumpAmps[p] := prTotJumpAmps[p] + JumpAmps[p];
        end;
        prTotJumpAmp     := prTotJumpAmp     + JumpAmp;
        prTotJumpStd     := prTotJumpStd     + JumpStdReest;
        prTotBaseAmp     := prTotBaseAmp     + BaseAmp;
        prTotKf          := prTotKf          + Jf;
        prTotKb          := prTotKb          + Jb;
        prTotMeasStd     := prTotMeasStd     + MeasStd;
        prTotLL          := prTotLL          + IdlBaseChunk.LL;

      end;

    end;


    FreeMem(JumpAmps);

    Free2(TPMChar(AMatrix));


  end;

end;



//****************************************************************************//
//                                                                            //
//                       F O R W A R D  -  V I T E R B I                      //
//                                                                            //
//                           (no re-estimation yet)                           //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlBase.prForwardViterbi(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
var
  Chunk               :TQUB_Thread_PChunk;

  Data                :TPAInteger;
  DataCount           :Integer;
  Scale               :Double;
  Off                 :Double;

  AMatrix             :TPMDouble;

  Prob0               :TPADouble;

  State               :TPAInteger;

  BackPath            :TPMByte;
  LogLike             :TPADouble;
  BestLogLike         :TPADouble;
  X                   :TPADouble;
  XB                  :TPMDouble;
  Alpha               :TPMDouble;

  y                   :Double;
  Sum                 :Double;

  c, cc, p, s, ss     :Integer;

begin


  Result              := False;


  //Nullify pointers ...........................................................
  AMatrix             := nil;
  Prob0               := nil;

  State               := nil;
  LogLike             := nil;
  BestLogLike         := nil;
  Alpha               := nil;
  X                   := nil;
  XB                  := nil;


  //Signal Running status ......................................................
  Chunk               := TQUB_Thread_PChunk(IdlBaseChunk);
  Chunk.Status        := CHUNK_STATUS_RUNNING;


  try//finally



    //Get data parameters ......................................................
    //   - Data, DataCount
    Data              := IdlBaseChunk.Data;
    DataCount         := Chunk.Sel.Last - Chunk.Sel.First + 1;
    //   - Scale
    Scale             := 1 / IdlBaseChunk.Scaling;
    Off               := IdlBaseChunk.Offset;



    //Allocate memory ..........................................................


    //   - A matrix
    AMatrix           := DAlloc2(prStateCount, prStateCount);


    //   - Prob0 - initial probabilities;
    Prob0             := DAlloc1(prStateCount);


    //   - Vectors
    LogLike           := DAlloc1(prClsCount);
    BestLogLike       := DAlloc1(prClsCount);
    Alpha             := DAlloc2(prClsCount, prStateCount);
    X                 := DAlloc1(prStateCount);
    XB                := DAlloc2(prClsCount, prStateCount);
    BackPath          := TPMByte(CAlloc2(DataCount, prClsCount));

    State             := IAlloc1(DataCount);



    //Initialize variables .....................................................


    //   - A matrix
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss]    := Abs(prA[s, ss]);//to prevent some occasional negative entries
        Sum               := Sum + AMatrix[s, ss];
      end;
      for ss := 0 to prStateCount - 1 do begin //re-normalize
        if Sum <> 0 then AMatrix[s, ss] := AMatrix[s, ss] / Sum;
      end;
    end;



    //   - Prob0
    for s := 0 to prStateCount - 1 do begin
      Prob0[s] := prP0[s];
    end;



    try



      //Initialize the first point
      p       := 0;
      y       := Data[p] * Scale + Off;


      for c := 0 to prClsCount - 1 do begin

        Sum := 0;
        for s := 0 to prStateCount - 1 do begin

          if prClsIdxR[s] = c then begin
            Alpha[c, s] := Prob0[s] * Gaussian(prAmp[prClsIdxR[s]], prStd[prClsIdxR[s]], y);
          end else begin
            Alpha[c, s] := 0;
          end;
          Sum := Sum + Alpha[c, s];

        end;//s

        if Sum = 0 then begin
          Sum := 1e-30;
        end;

        //Normalize Alpha
        for s := 0 to prStateCount - 1 do begin
          Alpha[c, s] := Alpha[c, s] / Sum;
        end;

        LogLike[c] := Ln(Sum);

      end;


      //Loop for every point
      for p := 1 to DataCount - 1 do begin

        y := Data[p] * Scale + Off;

        for c := 0 to prClsCount - 1 do begin

          BestLogLike[c] := SafeLog(0);

          for cc := 0 to prClsCount - 1 do begin

            //NewState = Alpha[cc] * A * Gaussian[c] * B[c]
            Sum := 0;
            for s := 0 to prStateCount - 1 do begin
              X[s] := 0;
              if prClsIdxR[s] = c then begin
                for ss := 0 to prStateCount - 1 do begin
                  X[s] := X[s] + Alpha[cc, ss] * AMatrix[ss, s];
                end;//ss
                X[s] := X[s] * Gaussian(prAmp[prClsIdxR[s]], prStd[prClsIdxR[s]], y);
              end;
              Sum := Sum + X[s];
            end;//s

            if Sum = 0 then begin
              Sum := 1e-30;
            end;

            //Normalize State probabilities
            for s := 0 to prStateCount - 1 do begin
              X[s] := X[s] / Sum;
            end;

            //if cc is best so far (or the first in the loop):
            if (cc > 0) and (LogLike[cc] + Ln(Sum) >= BestLogLike[c]) then begin
              BestLogLike[c] := LogLike[cc] + Ln(Sum);
              BackPath[p, c] := cc;
              for s := 0 to prStateCount - 1 do begin
                XB[c, s] := X[s];
              end;
            end else if cc = 0 then begin
              BestLogLike[c] := LogLike[cc] + Ln(Sum);
              BackPath[p, c] := cc;
              for s := 0 to prStateCount - 1 do begin
                XB[c, s] := X[s];
              end;
            end;

          end;//cc

        end;//c

        for c := 0 to prClsCount - 1 do begin

          LogLike[c] := BestLogLike[c];
          for s := 0 to prStateCount - 1 do begin
            Alpha[c, s] := XB[c, s];
          end;

        end;//c

      end;//p


      //Backtrace
      c   := 0;
      Sum := LogLike[c];
      cc  := c;
      for c := 1 to prClsCount - 1 do begin
        if LogLike[c] > Sum then begin
          cc  := c;
          Sum := LogLike[c];
        end;
      end;
      p := DataCount - 1;
      for s := 0 to prStateCount - 1 do begin
        if prClsIdxR[s] = cc then begin
          State[p] := s;
          Break;
        end;
      end;
      for p := DataCount - 2 downto 0 do begin
        c  := BackPath[p + 1, cc];
        cc := c;
        for s := 0 to prStateCount - 1 do begin
          if prClsIdxR[s] = cc then begin
            State[p] := s;
            Break;
          end;
        end;
      end;



      IdlBaseChunk.ClassCount := prClsCount;
      IdlBaseChunk.StateCount := prStateCount;
      IdlBaseChunk.Amps       := SAlloc1(prClsCount);
      IdlBaseChunk.Stds       := SAlloc1(prClsCount);
      IdlBaseChunk.AMatrix    := DAlloc2(prStateCount, prStateCount);
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          IdlBaseChunk.AMatrix[s, ss] := AMatrix[s, ss];
        end;
      end;
      for c := 0 to prClsCount - 1 do begin
        IdlBaseChunk.Amps[c]  := prAmp[c];
        IdlBaseChunk.Stds[c]  := prStd[c];
      end;

      IdlBaseChunk.LL         := 0;
      IdlBaseChunk.Iterations := 0;


      //Make Dwells
      prMakeDwells(IdlBaseChunk, State, DataCount);


      prMakeStats(IdlBaseChunk, nil);

   
    except


      Result            := True;

      Chunk.Status      := CHUNK_STATUS_ERROR;
      Chunk.Completion  := 0;


    end;


  finally


    FreeMem(State);
    Free2(TPMChar(AMatrix));
    FreeMem(Prob0);
    FreeMem(LogLike);
    FreeMem(BestLogLike);
    Free2(TPMChar(Alpha));
    FreeMem(X);
    Free2(TPMChar(XB));


    //Signal new status
    if Chunk.Status <> CHUNK_STATUS_ERROR then begin
      Chunk.Status := CHUNK_STATUS_DONE;
      Chunk.Completion := 100;
    end;


  end;


end;

//****************************************************************************//
//                                                                            //
//                        V I T E R B I  -  K A L M A N                       //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlBase.prReestSKM(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
var
  Chunk               :TQUB_Thread_PChunk;

  Data                :TPAInteger;
  DataCount           :Integer;
  Scale               :Double;
  Off                 :Double;

  Iteration           :Integer;

  //*********************************
  VProb               :TPMDouble;
  VBack               :TPMInteger;

  State               :TPAInteger;

  BaseX               :TPMDouble;
  BaseP               :TPMDouble;

  SBaseX              :TPADouble;
  SBaseP              :TPADouble;
  SBaseQ              :TPADouble;

  CBaseX              :TPADouble;
  CBaseP              :TPADouble;

  AMatrix             :TPMDouble;
  AMatrixN            :TPMDouble;
  LogAMatrix          :TPMDouble;

  Prob0               :TPADouble;

  Cnt                 :TPAInteger;
  Amp, AmpN           :TPADouble;
  Std                 :TPADouble;
  //*********************************

  Sum                 :Double;

  pm, pc              :Double;

  y                   :Double;

  c, p, s, ss, im     :Integer;

  OldLL, NewLL        :Double;

  BaseVar             :Double;
  BaseStd             :Double;

  nm, np              :Double;

  oJ, cJ              :Double;

  SeqDiff             :Boolean;

  yl, hl, kl          :Double;

  kxp, kpp            :Double;

  k                   :Double;

  RefAmp              :Double;

  //Movies
  MovieData           :INode;


  procedure DoBaselineBack;
  var
    p :Integer;
  begin

    p  := DataCount - 1;
    //nm := Std[prClsIdxR[State[p]]] * Std[prClsIdxR[State[p]]];

    //Recalculate the last point gain kt:
    //k  := (BaseP[State[p-1], p - 1] + BaseVar) / (BaseP[State[p-1], p-1] + BaseVar + nm);

    //SBaseQ[p] = (I - k . c) . a . BaseP[p-1]
    //SBaseQ[p] := (1 - k) * BaseP[State[p-1], p-1];

    //SBase[p] = Base[p]
    SBaseX[p] := BaseX[State[p], p];
    SBaseP[p] := BaseP[State[p], p];


    //oJ := 0;

    for p := DataCount - 1 downto 1 do begin

      //pp = a . P[p-1] . a'  + BaseVar
      //cJ = P[p-1] . a' . (pp)-1
      cJ          := BaseP[State[p-1], p-1] / (BaseP[State[p-1], p-1] + BaseVar);

      //SBaseX[p-1] = BaseX[p-1] + cJ . (SBaseX[p] - a . BaseX[p-1])
      SBaseX[p-1] := BaseX[State[p-1], p-1] + cJ *      (SBaseX[p] - BaseX[State[p-1], p-1]);

      //SBaseP[p-1] = BaseP[p-1] + cJ . (SBaseP[p] - pp) . cJ'
      SBaseP[p-1] := BaseP[State[p-1], p-1] + cJ * (SBaseP[p] - (BaseP[State[p-1], p-1] + BaseVar)) * cJ;

      //if p < (DataCount - 1) then begin
        //SBaseQ[p] = BaseP[p] . cJ'  +  oJ . (SBaseQ[p+1] - a . BaseP[p]) . cJ'
        //SBaseQ[p] := BaseP[State[p], p] * cJ  +  oJ * (SBaseQ[p+1] - BaseP[State[p], p]) * cJ;
      //end;

      //oJ := cJ;

    end;

  end;//DoBaselineBack

  procedure ReestBaselineVar;
  var
    p :Integer;
  begin

    cJ := 0;
    oJ := 0;

    for p := 1 to DataCount - 1 do begin

      //cJ = cJ + SBaseX[p]*SBaseX[p]' + SBaseP[p];
      cJ := cJ + SBaseX[p]   * SBaseX[p] + SBaseP[p];

      //oJ = oJ + SBaseX[p]*SBaseX[p-1]' + SBaseQ[p];
      oJ := oJ + SBaseX[p] * SBaseX[p-1] + SBaseQ[p];
    end;

    //BaseVar = (cJ - A . oJ') / (DataCount - 1)
    BaseVar := (cJ - oJ) / (DataCount - 1);
    if BaseVar < KALMAN_SMALL then BaseVar := KALMAN_SMALL;
    BaseStd := Sqrt(BaseVar);

  end;//ReestBaselineVar

  procedure ReestAmpsStds;
  var
    c, p :Integer;
  begin

    for c := 0 to prClsCount - 1 do begin
      AmpN[c] := 0;
      Cnt[c]  := 0;
    end;

    //Amplitudes
    if prParams.BaselineTrack then begin
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        y := y - SBaseX[p];
        AmpN[prClsIdxR[State[p]]] := AmpN[prClsIdxR[State[p]]] + y;
        Inc(Cnt[prClsIdxR[State[p]]]);
      end;
    end else begin
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        AmpN[prClsIdxR[State[p]]] := AmpN[prClsIdxR[State[p]]] + y;
        Inc(Cnt[prClsIdxR[State[p]]]);
      end;
    end;
    for c := 0 to prClsCount - 1 do begin
      //Only represented classes
      if Cnt[c] <> 0 then begin
        Amp[c]  := AmpN[c] / Cnt[c];
        Std[c]  := 0;
        AmpN[c] := 0;
      end;
    end;

    //Noise
    if prParams.BaselineTrack then begin
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        y := y - SBaseX[p];
        Std[prClsIdxR[State[p]]] := Std[prClsIdxR[State[p]]] + Sqr(y - Amp[prClsIdxR[State[p]]]);
      end;
    end else begin
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        Std[prClsIdxR[State[p]]] := Std[prClsIdxR[State[p]]] + Sqr(y - Amp[prClsIdxR[State[p]]]);
      end;
    end;
    for c := 0 to prClsCount - 1 do begin
      //Only represented classes
      if Cnt[c] <> 0 then begin
        Std[c]  := Sqrt(Std[c] / Cnt[c]);
      end;
    end;

  end;//ReestAmpsStds

  procedure ReestP0;
  var
    s :Integer;
  begin
    for s := 0 to prStateCount - 1 do begin
      Prob0[s] := 0;
      if s = State[0] then Prob0[s] := 1;
    end;
  end;//ReestP0

  procedure ReestAMatrix;
  var
    p, s, ss :Integer;
  begin

    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        AMatrixN[s, ss] := 0;
      end;
    end;

    for p := 0 to DataCount - 2 do begin
      AMatrixN[State[p], State[p + 1]] := AMatrixN[State[p], State[p + 1]] + 1;
    end;

    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        Sum := Sum + AMatrixN[s, ss];
      end;
      if Sum = 0 then begin
        for ss := 0 to prStateCount - 1 do begin
          AMatrixN[s, ss] := AMatrix[s, ss];
          if AMatrixN[s, ss] = 0 then AMatrixN[s, ss] := 1e-10;
        end;
      end else begin
        for ss := 0 to prStateCount - 1 do begin
          AMatrixN[s, ss] := AMatrixN[s, ss] / Sum;
          if AMatrixN[s, ss] = 0 then AMatrixN[s, ss] := 1e-10;
        end;
      end;
      //Do it again
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        Sum := Sum + AMatrixN[s, ss];
      end;
      for ss := 0 to prStateCount - 1 do begin
        AMatrixN[s, ss] := AMatrixN[s, ss] / Sum;
      end;
    end;
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss]    := AMatrixN[s, ss];
        LogAMatrix[s, ss] := SafeLog(AMatrix[s, ss]);
      end;
    end;

  end;//ReestAMatrix

  function GetNewLL :Double;
  var
    p    :Integer;
  begin

    p    := 0;
    y    := Data[p] * Scale + Off;

    if prParams.BaselineTrack then begin
      yl   := SafeLog(Gaussian(Amp[prClsIdxR[State[p]]], Std[prClsIdxR[State[p]]], y - SBaseX[p]));
      hl   := SafeLog(Prob0[State[p]]);
      kl   := SafeLog(Gaussian(y - Amp[prClsIdxR[State[p]]], BaseStd, SBaseX[p]));
    end else begin
      yl   := SafeLog(Gaussian(Amp[prClsIdxR[State[p]]], Std[prClsIdxR[State[p]]], y));
      hl   := SafeLog(Prob0[State[p]]);
      kl   := 0;
    end;

    for p := 1 to DataCount - 1 do begin

      y  := Data[p] * Scale + Off;

      if prParams.BaselineTrack then begin
        yl := yl + SafeLog(Gaussian(Amp[prClsIdxR[State[p]]], Std[prClsIdxR[State[p]]], y - SBaseX[p]));
        hl := hl + LogAMatrix[State[p - 1], State[p]];
        kl := kl + SafeLog(Gaussian(SBaseX[p - 1], BaseStd, SBaseX[p]));
      end else begin
        yl := yl + SafeLog(Gaussian(Amp[prClsIdxR[State[p]]], Std[prClsIdxR[State[p]]], y));
        hl := hl + LogAMatrix[State[p - 1], State[p]];
      end;

    end;

    Result := yl + hl;// + kl;

  end;//GetNewLL


begin


  Result              := False;


  //Nullify pointers ...........................................................
  VProb               := nil;
  VBack               := nil;

  State               := nil;

  AMatrix             := nil;
  AMatrixN            := nil;
  LogAMatrix          := nil;

  Prob0               := nil;

  Cnt                 := nil;
  Amp                 := nil;
  AmpN                := nil;
  Std                 := nil;

  BaseX               := nil;
  BaseP               := nil;

  SBaseX              := nil;
  SBaseP              := nil;
  SBaseQ              := nil;

  CBaseX              := nil;
  CBaseP              := nil;


  //Signal Running status ......................................................
  Chunk               := TQUB_Thread_PChunk(IdlBaseChunk);
  Chunk.Status        := CHUNK_STATUS_RUNNING;


  //Movies
  if prMovie.Make then begin

    MovieTree := TNode.Create('IBMovie');
     PythonThread.BeginPython;
     PyNode_Publish(MovieTree, 'IBMovie'); // make tree available in Python as QUB.IBMovie
     PythonThread.EndPython;
     PythonThread.DonePython;

  end;



  try//finally



    //Get data parameters ......................................................
    //   - Data, DataCount
    Data              := IdlBaseChunk.Data;
    DataCount         := Chunk.Sel.Last - Chunk.Sel.First + 1;
    //   - Scale
    Scale             := 1 / IdlBaseChunk.Scaling;
    Off               := IdlBaseChunk.Offset;


    //Movies
    if prMovie.Make then begin

      MovieTree['Scaling'].SetData( IdlBaseChunk.Scaling );
      MovieTree['Offset'].SetData( IdlBaseChunk.Offset );
      MovieTree['DataCount'].SetData( DataCount );

      MovieData       := MovieTree['Amps'];
      MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, 1 );
      for s := 0 to prStateCount - 1 do begin
        MovieData.Data[s] := prAmp[prClsIdxR[s]];
      end;

    end;



    //Allocate memory ..........................................................


    //   - VProb - the probabilities used by Viterbi;
    VProb             := DAlloc2(prStateCount, DataCount);
    //   - VBack - backtracing used by Viterbi;
    VBack             := IAlloc2(prStateCount, DataCount);

    //   - State - the idealized state, at each t;
    State             := IAlloc1(DataCount);
    for p := 0 to DataCount - 1 do State[p] := -1;


    //   - A matrix
    AMatrix           := DAlloc2(prStateCount, prStateCount);
    AMatrixN          := DAlloc2(prStateCount, prStateCount);
    LogAMatrix        := DAlloc2(prStateCount, prStateCount);



    //   - Prob0 - initial probabilities;
    Prob0             := DAlloc1(prStateCount);


    if prParams.BaselineTrack then begin
      //   - BaseX|P - baseline state|variance, for each data point, for each state;
      BaseX           := DAlloc2(prStateCount, DataCount);
      BaseP           := DAlloc2(prStateCount, DataCount);
      //   - SBaseX|P|Q - smoothed baseline state, variance and BaseVar, for each data point
      SBaseX          := DAlloc1(DataCount);
      SBaseP          := DAlloc1(DataCount);
      SBaseQ          := DAlloc1(DataCount);
      //   - CBaseX|P - current baseline, for each state
      CBaseX          := DAlloc1(prStateCount);
      CBaseP          := DAlloc1(prStateCount);
    end;


                                                                                                                   
    //   - Cnt, Amp, Std
    Cnt               := IAlloc1(prStateCount);
    Amp               := DAlloc1(prStateCount);
    AmpN              := DAlloc1(prStateCount);
    Std               := DAlloc1(prStateCount);



    //Initialize variables .....................................................


    //   - A matrix
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss]    := Abs(prA[s, ss]);//to prevent some occasional negative entries
        Sum               := Sum + AMatrix[s, ss];
      end;
      for ss := 0 to prStateCount - 1 do begin //re-normalize
        if Sum <> 0 then AMatrix[s, ss]    := AMatrix[s, ss] / Sum;
        LogAMatrix[s, ss] := SafeLog(AMatrix[s, ss]);
      end;
    end;



    //   - Prob0
    for s := 0 to prStateCount - 1 do begin
      Prob0[s] := prP0[s];
    end;



    //   - Amp, Std
    for c := 0 to prClsCount - 1 do begin
      Amp[c] := prAmp[c];
      Std[c] := prStd[c];
    end;


    BaseStd := 0;
    BaseVar := 0;


    //Initialize BaseX[0] and BaseP[0]
    if prParams.BaselineTrack then begin
      p             := 0;
      y             := Data[p] * Scale + Off;
      for s := 0 to prStateCount - 1 do begin
        BaseX[s, p] := y - Amp[prClsIdxR[s]];
        BaseP[s, p] := Sqr(Std[prClsIdxR[s]]);
      end;
      BaseStd       := prParams.BaselineStd;
      BaseVar       := Sqr(BaseStd);
    end;



    //Print starting values ....................................................
    prConsoleMsg := 'Starting values:'#13#10;
    prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
      end;
      prConsoleMsg := prConsoleMsg + #13#10;
    end;
    prConsoleMsg := prConsoleMsg + 'State, Prob0:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [s, Prob0[s]]);
    end;
    prConsoleMsg := prConsoleMsg + 'Cls, Amp, Std:'#13#10;
    for c := 0 to prClsCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [prClsIdx[c], Amp[c], Std[c]]);
    end;
    if prParams.BaselineTrack then begin
      prConsoleMsg := prConsoleMsg + 'Baseline Std, Var:'#13#10;
      prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#13#10, [BaseStd, BaseVar]);
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    if not prParams.QuietOutput then Synchronize(prWriteNewConsole);



    try


      OldLL := 0;


      //Iterate
      for Iteration := 1 to prParams.MaxIter do begin


        SeqDiff := False;


        RefAmp  := 0;


        //Forward **************************************************************


        //Initialize the first point
        p       := 0;
        y       := Data[p] * Scale + Off;


        for s := 0 to prStateCount - 1 do begin


          //Measurement noise
          nm    := Std[prClsIdxR[s]]*Std[prClsIdxR[s]];


          //Baseline
          //The prior x is y - Amp[ClassOfState[s]]
          //The prior p is Std[ClassOfState[s]]
          kl       := 0;//LL of Kalman process
          if prParams.BaselineTrack then begin
            //Save the previous values
            kxp    := BaseX[s, p];
            KalmanForward(BaseX[s, p], BaseP[s, p], BaseX[s, p], BaseP[s, p], nm, BaseVar, y - Amp[prClsIdxR[s]]{Data});
            kl     := SafeLog(Gaussian(kxp, BaseStd,  BaseX[s, p]));
            RefAmp := BaseX[s, p];
          end;

          //p(y|s0, x0) * p(s0) * p(x0)
          VProb[s, 0] := SafeLog(Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y - RefAmp)) +
                         SafeLog(Prob0[s]) +
                         kl;

        end;

        im := 0;
        pm := 0;

        //Loop for every point
        for p := 1 to DataCount - 1 do begin

          y := Data[p] * Scale + Off;

          //for each state
          for s := 0 to prStateCount - 1 do begin

            nm         := Std[prClsIdxR[s]]*Std[prClsIdxR[s]];

            for ss := 0 to prStateCount - 1 do begin

              kl       := 0;
              if prParams.BaselineTrack then begin
                KalmanForward(CBaseX[ss], CBaseP[ss], BaseX[ss, p - 1], BaseP[ss, p - 1], nm, BaseVar, y - Amp[prClsIdxR[s]]);
                kl     := SafeLog(Gaussian(BaseX[ss, p - 1], BaseStd,  CBaseX[ss]));
                RefAmp := CBaseX[ss];
              end;

              //p(y|s0, x0) * p(s0) * p(x0)
              pc       := SafeLog(Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y - RefAmp)) +
                          LogAMatrix[ss, s] +
                          kl +
                          VProb[ss, p - 1];

              //Max of ss
              if (ss = 0) then begin
                im := ss;
                pm := pc;
              end else if (pc > pm) then begin
                im := ss;
                pm := pc;
              end;

            end;//of ss

            VProb[s, p] := pm;
            VBack[s, p] := im;

            if prParams.BaselineTrack then begin
              BaseX[s, p] := CBaseX[im];
              BaseP[s, p] := CBaseP[im];
            end;

          end;//of s


        end;//of p


        //BackTrace
        s       := 0;
        p       := DataCount - 1;
        im      := 0;
        pm      := VProb[s, p];
        for s := 1 to prStateCount - 1 do begin
          if VProb[s, p] > pm then begin
            im  := s;
            pm  := VProb[s, p];
          end;
        end;
        if im <> State[p] then SeqDiff := True;//???
        State[p] := im;
        for p := DataCount - 2 downto 0 do begin
          if State[p] <> VBack[State[p + 1], p + 1] then SeqDiff := True;//???
          State[p] := VBack[State[p + 1], p + 1];
        end;



        //Backward step on Baseline
        if prParams.BaselineTrack then begin

          DoBaselineBack;

        end;//Backward on Baseline



        if prParams.Reestimation = IDLBASE_REEST_NONE then begin
          //Recalculate LL and quit
          NewLL := GetNewLL;
          Break;
        end;



        //Re-estimation ********************************************************



        //Amplitudes and Noise
        ReestAmpsStds;


        //Eliminate offset in Amp and Baseline
        if prParams.BaselineTrack then begin
          for s := 0 to prStateCount - 1 do begin                    
            BaseX[s, 0] := SBaseX[0] + Amp[prClsIdx[0]];
          end;
          for s := 0 to prStateCount - 1 do begin
            BaseP[s, 0] := SBaseP[0];
          end;
          Sum := Amp[prClsIdx[0]];
          for c := 0 to prClsCount - 1 do begin
            Amp[c] := Amp[c] - Sum;
          end;
        end;

        if SeqDiff then begin

          //Initial probabilities
          ReestP0;

          //A Matrix
          if not prParams.FixKinetics then begin
            ReestAMatrix;
          end;

          if (prParams.A2Q) and (prParams.A2QEveryIter) then begin
            prA2Q(AMatrix, nil, False{Q});
            if not prParams.FixKinetics then begin
              for s := 0 to prStateCount - 1 do begin
                for ss := 0 to prStateCount - 1 do begin
                  LogAMatrix[s, ss] := SafeLog(AMatrix[s, ss]);
                end;
              end;
            end;
          end;

        end;


        //Recalculate LL with reestimated parameters
        NewLL := GetNewLL;


        //Print Q matrix as the logarithm of the A matrix ......................
        MatrixLogarithm(AMatrix, AMatrixN, IDLBaseChunk.Sampling, prStateCount, 100, 1e-4);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            if AMatrixN[s, ss] < 0 then begin
              AMatrixN[s, ss] := 0;
              AMatrixN[ss, s] := 0;
            end;
          end;
        end;
        prConsoleMsg := #13#10'Q Matrix:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            if (s = ss) or (prQ[s, ss] = 0)then begin
              prConsoleMsg := prConsoleMsg + #9'...           ';
            end else begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrixN[s, ss]]);
            end;
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);



        //Print current iteration values .......................................
        prConsoleMsg := Format(#13#10'Iteration %d'#13#10, [Iteration]);
        if prParams.BaselineTrack then begin
          prConsoleMsg := prConsoleMsg + 'LL_HMM, LL_LGM, LL_TOT:'#13#10;
          prConsoleMsg := prConsoleMsg + Format(#9'%14.10g'#13#10#9'%14.10g'#13#10#9'%14.10g'#13#10, [NewLL, kl, NewLL + kl]);
        end else begin
          prConsoleMsg := prConsoleMsg + Format('LL: %14.10g'#13#10, [NewLL]);
        end;
        prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleMsg := prConsoleMsg + 'State, Prob0:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [s, Prob0[s]]);
        end;
        prConsoleMsg := prConsoleMsg + 'Cls, Amp, Std:'#13#10;
        for c := 0 to prClsCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [prClsIdx[c], Amp[c], Std[c]]);
        end;
        if prParams.BaselineTrack then begin
          prConsoleMsg := prConsoleMsg + 'Baseline Var, Std, X0, P0:'#13#10;
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#13#10#9'%-14.6g'#13#10#9'%-14.6g'#13#10#9'%-14.6g'#13#10, [BaseVar, Sqrt(Abs(BaseVar)), SBaseX[0], SBaseP[0]]);
        end;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);



        //********************************************************************//
        //                                                                    //
        //                              M O V I E S                           //
        //                                                                    //
        //********************************************************************//
        if prMovie.Make then begin


          MovieData       := MovieTree['VProb'];
          MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
          for s := 0 to prStateCount - 1 do begin
            Move(VProb[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
          end;
          MovieData       := MovieTree['VBack'];
          MovieData.SetupNumData( NODE_DATATYPE_LONGINT, prStateCount, DataCount );
          for s := 0 to prStateCount - 1 do begin
            Move(VBack[s][0], TPAInteger(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Integer));
          end;


          if prParams.BaselineTrack then begin
            MovieData       := MovieTree['BaseX'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(BaseX[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;
            MovieData       := MovieTree['BaseP'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(BaseP[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;

            MovieData       := MovieTree['SBaseX'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(SBaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['SBaseP'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(SBaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
          end;

          MovieData       := MovieTree['State'];
          MovieData.SetupNumData( NODE_DATATYPE_LONGINT, DataCount, 1 );
          Move(State^, MovieData.DataPtr^, SizeOf(Integer) * DataCount);


          prMovieIteration  := Iteration;
          Synchronize(prMakeMovie_ViterbiKalman);


          MovieData       := MovieTree['Amps'];
          for s := 0 to prStateCount - 1 do begin
            MovieData.Data[s] := Amp[prClsIdxR[s]];
          end;


        end;



        if prStopped then Break;


        //Check convergence
        if Abs(NewLL - OldLL) <= prParams.LLConv then Break;


        OldLL := NewLL;


      end;//of Iteration



      //Results ................................................................
      //if not prStopped then begin


        //Make Dwells
        prMakeDwells(IdlBaseChunk, State, DataCount);

        IdlBaseChunk.ClassCount := prClsCount;
        IdlBaseChunk.StateCount := prStateCount;
        IdlBaseChunk.Amps       := SAlloc1(prClsCount);
        IdlBaseChunk.Stds       := SAlloc1(prClsCount);
        IdlBaseChunk.AMatrix    := DAlloc2(prStateCount, prStateCount);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            IdlBaseChunk.AMatrix[s, ss] := AMatrix[s, ss];
          end;
        end;
        for c := 0 to prClsCount - 1 do begin
          IdlBaseChunk.Amps[c]  := Amp[c];
          IdlBaseChunk.Stds[c]  := Std[c];
        end;

        IdlBaseChunk.LL         := NewLL;//HMM_LL
        IdlBaseChunk.Iterations := Iteration - 1;

        //Make Baseline Nodes
        if (prParams.BaselineTrack) and (prParams.BaselineCorrect) then begin

          prBase2Nodes(IdlBaseChunk, SBaseX);

        end;


        if (prParams.A2Q) and (not prParams.A2QEveryIter) then begin
          prA2Q(AMatrix, nil, False{Q});
        end;


        //Make Stats and Histograms
        prMakeStats(IdlBaseChunk, SBaseX); // SBaseX is nil if not prParams.BaselineTrack


      //end;//Results


    except


      Result            := True;

      Chunk.Status      := CHUNK_STATUS_ERROR;
      Chunk.Completion  := 0;


    end;


  finally


    //Free Memory ..............................................................
    Free2(TPMChar(VProb));
    Free2(TPMChar(VBack));

    FreeMem(State);

    Free2(TPMChar(BaseX));
    Free2(TPMChar(BaseP));
    FreeMem(SBaseX);
    FreeMem(SBaseP);
    FreeMem(SBaseQ);
    FreeMem(CBaseX);
    FreeMem(CBaseP);

    Free2(TPMChar(LogAMatrix));
    Free2(TPMChar(AMatrix));
    Free2(TPMChar(AMatrixN));

    FreeMem(Prob0);

    FreeMem(Cnt);
    FreeMem(Amp);
    FreeMem(AmpN);
    FreeMem(Std);


    //Signal new status
    if Chunk.Status <> CHUNK_STATUS_ERROR then begin
      Chunk.Status := CHUNK_STATUS_DONE;
      Chunk.Completion := 100;
    end;


  end;

end;



//****************************************************************************//
//                                                                            //
//                    B A U M  -  W E L C H  -  K A L M A N                   //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlBase.prReestBW(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
var
  Chunk               :TQUB_Thread_PChunk;

  Data                :TPAInteger;
  DataCount           :Integer;
  Scale               :Double;
  Off                 :Double;

  Iteration           :Integer;

  //*********************************
  Alpha               :TPMDouble;
  Betta               :TPMDouble;
  Gamma               :TPMDouble;
  Gauss               :TPMDouble;
  Rho                 :TPADouble;
  Epsilon             :TPAPointer;

  State               :TPAInteger;

  BaseX               :TPADouble;
  BaseP               :TPADouble;

  SBaseX              :TPADouble;
  SBaseP              :TPADouble;
  SBaseQ              :TPADouble;

  AMatrix             :TPMDouble;
  AMatrixN            :TPMDouble;
  AM                  :TPADouble;

  Prob0               :TPADouble;

  Amp, AmpN           :TPADouble;
  Std                 :TPADouble;

  Sum_Gamma_y         :TPADouble;
  Sum_Gamma           :TPADouble;

  Cta                 :TPADouble;
  Ctv                 :TPADouble;
  //*********************************

  Cstr                :TQUB_QFS_ConstraintAmpVar;

  Sum                 :Double;

  y                   :Double;

  c, i, p, s, ss      :Integer;

  OldLL, NewLL        :Double;

  CanQuit             :Boolean;
  BaseEstCount        :Integer;

  RefAmp              :Double;
  ThisAmp             :Double;

  MeasVar             :Double;

  BaseVar             :Double;
  BaseStd             :Double;

  nm, np              :Double;

  oJ, cJ              :Double;

  yl, hl, kl          :Double;

  kxp, kpp            :Double;

  k                   :Double;
  ca, cv              :Integer;

  //JumpAmp,   JumpStd  :Double;
  //JumpAmpLS, JumpStdLS:Double;

  //Movies
  MovieData           :INode;


  procedure DoBaselineForward;
  var
    p, s :Integer;
  begin

    //First point
    p := 0;
    y  := Data[p] * Scale + Off;
    nm := 0;
    for s := 0 to prStateCount - 1 do begin
      y  := y  - Gamma[s, p] * Amp[prClsIdxR[s]];
      nm := nm + Gamma[s, p] * Sqr(Std[prClsIdxR[s]]);
    end;
    KalmanForward(BaseX[p], BaseP[p], SBaseX[p], SBaseP[p], nm, BaseVar, y);

    //Every point
    for p := 1 to DataCount - 1 do begin
      if (p mod prParams.BaselineDecFactor) = 0 then begin

        y  := Data[p] * Scale + Off;
        nm := 0;
        for s := 0 to prStateCount - 1 do begin
          y  := y  - Gamma[s, p] * Amp[prClsIdxR[s]];
          nm := nm + Gamma[s, p] * Sqr(Std[prClsIdxR[s]]);
        end;
        KalmanForward(BaseX[p], BaseP[p], BaseX[p-1], BaseP[p-1], nm, BaseVar, y);

      end else begin

        BaseX[p] := BaseX[p - 1];
        BaseP[p] := BaseP[p - 1];

      end;
    end;//p

  end;//DoBaselineForward

  procedure DoBaselineBack;
  var
    p :Integer;
  begin

    p  := DataCount - 1;
    //nm := 0;
    //for s := 0 to prStateCount - 1 do begin
    //  nm := nm + Alpha[s, p] * Sqr(Std[prClsIdxR[s]]);
    //end;


    //Recalculate the last point gain k:
    //k := (BaseP[p-1] + BaseVar) / (BaseP[p-1] + BaseVar + nm);

    //SBaseQ[p] = (I - k . c) . a . BaseP[p-1]
    //SBaseQ[p] := (1 - k) * BaseP[p-1];

    //SBase[p] = Base[p]
    SBaseX[p] := BaseX[p];
    SBaseP[p] := BaseP[p];

    //oJ := 0;

    for p := DataCount - 1 downto 1 do begin
      if (p mod prParams.BaselineDecFactor) = 0 then begin

        //pp = a . P[p-1] . a'  + BaseVar
        //cJ = P[p-1] . a' . (pp)-1
        cJ          := BaseP[p-1] / (BaseP[p-1] + BaseVar);

        //SBaseX[p-1] = BaseX[p-1] + cJ . (SBaseX[p] - a . BaseX[p-1])
        SBaseX[p-1] := BaseX[p-1] + cJ * (SBaseX[p] - BaseX[p-1]);

        //SBaseP[p-1] = BaseP[p-1] + cJ . (SBaseP[p] - pp) . cJ'
        SBaseP[p-1] := BaseP[p-1] + cJ * (SBaseP[p] - (BaseP[p-1] + BaseVar)) * cJ;

        //if p < (DataCount - 1) then begin
          //SBaseQ[p] = BaseP[p] . cJ'  +  oJ . (SBaseQ[p+1] - a . BaseP[p]) . cJ'
          //SBaseQ[p] := BaseP[p] * cJ  +  oJ * (SBaseQ[p+1] - BaseP[p]) * cJ;
        //end;

        //oJ := cJ;

      end else begin

        SBaseX[p-1] := SBaseX[p];
        SBaseP[p-1] := SBaseP[p];

      end;
    end;//p

  end;//DoBaselineBack;

  procedure DoBaselineBBack;
  var
    LastX, LastP :Double;
    p            :Integer;
  begin

    p     := 0;

    LastX := BaseX[p];
    LastP := BaseP[p];

    if LastP < SBaseP[p] then begin
      SBaseX[p] := LastX;
      SBaseP[p] := LastP;
    end;

    for p := 0 to DataCount - 2 do begin
      if ((p+1) mod prParams.BaselineDecFactor) = 0 then begin

        cJ    := BaseP[p+1] / (BaseP[p+1] + BaseVar);
        LastX := BaseX[p+1] + cJ * (LastX -  BaseX[p+1]);
        LastP := BaseP[p+1] + cJ * (LastP - (BaseP[p+1] + BaseVar)) * cJ;
        if LastP < SBaseP[p+1] then begin
          SBaseX[p+1] := LastX;
          SBaseP[p+1] := LastP;
        end;

      end else begin

        SBaseX[p+1] := SBaseX[p];
        SBaseP[p+1] := SBaseP[p];

      end;
    end;//p

  end;//DoBaselineBBack;

  procedure ReestBaselineVar;
  var
    p :Integer;
    Count :LongInt;
  begin

    cJ := 0;
    oJ := 0;

    Count := 0;

    for p := 1 to DataCount - 1 do begin
      if (p mod prParams.BaselineDecFactor) = 0 then begin
        //cJ = cJ + SBaseX[p]*SBaseX[p]' + SBaseP[p];
        cJ := cJ + SBaseX[p]   * SBaseX[p] + SBaseP[p];
        //oJ = oJ + SBaseX[p]*SBaseX[p-1]' + SBaseQ[p];
        oJ := oJ + SBaseX[p] * SBaseX[p-1] + SBaseQ[p];
        Inc(Count);
      end;
    end;

    //BaseVar = (cJ - A . oJ') / Count
    BaseVar := (cJ - oJ) / Count;

    if BaseVar < KALMAN_SMALL then BaseVar := KALMAN_SMALL;
    BaseStd := Sqrt(BaseVar);

  end;//ReestBaselineVar

  procedure DoBetta;
  var
    p, s, ss :Integer;
  begin

    //Initialize last point
    p := DataCount - 1;
    for s := 0 to prStateCount - 1 do begin
      Betta[s, p] := 1 / Rho[p];
    end;//of s
    //Calculate Betta
    for p := DataCount - 2 downto 0 do begin
      for s := 0 to prStateCount - 1 do begin
        Betta[s, p] := 0;
        for ss := 0 to prStateCount - 1 do begin
          Betta[s, p] := Betta[s, p] + AMatrix[s, ss] * Gauss[ss, p + 1] * Betta[ss, p + 1];
        end;//of ss
        Betta[s, p] := Betta[s, p] / Rho[p];
      end;//of s
    end;//of p

  end;//DoBetta

  procedure DoGamma;
  var
    p, s :Integer;
  begin

    for p := 0 to DataCount - 1 do begin
      Sum := 0;
      for s := 0 to prStateCount - 1 do begin
        Gamma[s, p] := Alpha[s, p] * Betta[s, p];
        Sum := Sum + Gamma[s, p];
      end;//of s
      //Rescale
      for s := 0 to prStateCount - 1 do begin
        Gamma[s, p] := Gamma[s, p] / Sum;
      end;//of s
    end;//of p

  end;//DoGamma

  procedure DoEpsilon;
  var
    p, s, ss :Integer;
  begin

    for p := 0 to DataCount - 2 do begin
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          TPMDouble(Epsilon[p])[s, ss] := Alpha[s, p] * AMatrix[s, ss] * Gauss[ss, p + 1] * Betta[ss, p + 1];
        end;//of ss
      end;//of s
    end;//of p

  end;//DoEpsilon

  function DoForward :Double;
  var
    p, s, ss :Integer;
  begin
    Result := 0;

    //Initialize the first point
    p := 0;
    y := Data[p] * Scale + Off;
    if prParams.BaselineTrack then begin
      y := y - SBaseX[p];
    end;
    Rho[p] := 0;
    for s := 0 to prStateCount - 1 do begin
      Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
      Alpha[s, p] := Prob0[s] * Gauss[s, p];
      Rho[p]      := Rho[p] + Alpha[s, p];
    end;
    for s := 0 to prStateCount - 1 do begin
      Alpha[s, p] := Alpha[s, p] / Rho[p];         //Check for Rho to be > 0;
    end;
    Result := Result + SafeLog(Rho[p]);
    //Loop for every point
    for p := 1 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      if prParams.BaselineTrack then begin
        y := y - SBaseX[p];
      end;
      Rho[p] := 0;
      for s := 0 to prStateCount - 1 do begin
        Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
        Alpha[s, p] := 0;
        for ss := 0 to prStateCount - 1 do begin
          Alpha[s, p] := Alpha[s, p] + AMatrix[ss, s] * Alpha[ss, p - 1];
        end;
        Alpha[s, p] := Alpha[s, p] * Gauss[s, p];
        Rho[p] := Rho[p] + Alpha[s, p];
      end;//of s
      for s := 0 to prStateCount - 1 do begin
        Alpha[s, p] := Alpha[s, p] / Rho[p];
      end;
      Result := Result + SafeLog(Rho[p]);
    end;

  end;//DoForward

  (*
  procedure DoAmpStdLS;
  var
    Amp_Size    :Integer;
    Amp_A       :TPMDouble;
    Amp_AI      :TPMDouble;
    Amp_B       :TPADouble;
    Amp_X       :TPADouble;
    Gammat      :TPADouble;
    Gammat_yt   :TPADouble;

    i, j, k, p  :Integer;

    // t           :String;

    function W(l :Integer) :Double;
    begin
      Result := (Gammat[l] * Gammat[l + 1]) / (Gammat[l] + Gammat[l + 1]);
      if Result < 1 then Result := 1;
    end;

  begin


    Amp_A       := nil;
    Amp_AI      := nil;
    Amp_B       := nil;
    Amp_X       := nil;
    Gammat      := nil;
    Gammat_yt   := nil;


    try


      {
        Must solve the equation Ax = B
          x is the vector of unknowns (amplitudes), including the JumpAmp;
          A and B are obtained by setting the derivatives of the cost function to zero
          First Amp_Size rows are for Amps, the last for JumpAmp
      }

      //Allocate A and B:
      Amp_Size    := prClsCount;

      Amp_A       := DAlloc2(Amp_Size + 1, Amp_Size + 1);
      Amp_AI      := DAlloc2(Amp_Size + 1, Amp_Size + 1);
      Amp_B       := DAlloc1(Amp_Size + 1);
      Amp_X       := DAlloc1(Amp_Size + 1);
      Gammat      := DAlloc1(Amp_Size);
      Gammat_yt   := DAlloc1(Amp_Size);


      //Clear Gammat, Gammat_yt
      for i := 0 to Amp_Size - 1 do begin
        Gammat   [i] := 0;
        Gammat_yt[i] := 0;
      end;

      //Calculate Gammat and Gammat_yt
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        if prParams.BaselineTrack then begin
          y := y - SBaseX[p];
        end;
        for i := 0 to prStateCount - 1 do begin
          Gammat_yt[prClsIdxR[i]] := Gammat_yt[prClsIdxR[i]] + Gamma[i, p] * y;
          Gammat   [prClsIdxR[i]] := Gammat   [prClsIdxR[i]] + Gamma[i, p];
        end;
      end;


      //Zero the matrices
      for i := 0 to Amp_Size do begin
        Amp_B[i] := 0;
        Amp_X[i] := 0;
        for j := 0 to Amp_Size do begin
          Amp_A[i, j] := 0;
        end;
      end;


      //Amp_B
      for i := 0 to Amp_Size - 1 do begin
        Amp_B[i] := Gammat_yt[i];
      end;

      //Amp_A
      for i := 0 to Amp_Size - 1 do begin
        Amp_A[i, i] := Gammat[i];
        if i > 0              then begin
          Amp_A[i, i]         := Amp_A[i, i]        + W(i - 1);
          Amp_A[i, i - 1]     :=                    - W(i - 1);
          Amp_A[i, Amp_Size]  := Amp_A[i, Amp_Size] - W(i - 1);
          Amp_A[Amp_Size, i]  := Amp_A[Amp_Size, i] - W(i - 1);
        end;
        if i < (Amp_Size - 1) then begin
          Amp_A[i, i]         := Amp_A[i, i]        + W(i);
          Amp_A[i, i + 1]     :=                    - W(i);
          Amp_A[i, Amp_Size]  := Amp_A[i, Amp_Size] + W(i);
          Amp_A[Amp_Size, i]  := Amp_A[Amp_Size, i] + W(i);
        end;
      end;
      for i := 0 to Amp_Size - 2 do begin
        Amp_A[Amp_Size, Amp_Size] := Amp_A[Amp_Size, Amp_Size] + W(i);
      end;


      {
      t := 'Amp_A'#13#10'<';
      for i := 0 to Amp_Size do begin
        t := t + '<';
        for j := 0 to Amp_Size do begin
          t := t + FloatToStr(Amp_A[i, j]);
          if j < Amp_Size then t := t + '|';
        end;
        t := t + '>';
        if i < Amp_Size then t := t + ',';
      end;
      t := t + '>';
      prConsoleMsg := #13#10 + t + #13#10;
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);
      }


      //X = Inv(A) * B
      InvMat(PMatrix(Amp_A), 0, Amp_Size, PMatrix(Amp_AI));


      {
      t := 'Amp_A'#13#10'<';
      for i := 0 to Amp_Size do begin
        t := t + '<';
        for j := 0 to Amp_Size do begin
          t := t + FloatToStr(Amp_AI[i, j]);
          if j < Amp_Size then t := t + '|';
        end;
        t := t + '>';
        if i < Amp_Size then t := t + ',';
      end;
      t := t + '>';
      prConsoleMsg := #13#10 + t + #13#10;
      prConsoleCode := MESSAGE_MESSAGE;
      Synchronize(prWriteNewConsole);
      }


      for k := 0 to Amp_Size do begin
        for i := 0 to Amp_Size do begin
          Amp_X[k] := Amp_X[k] + Amp_AI[k, i] * Amp_B[i];
        end;
      end;


      JumpAmp := 0;
      JumpStd := 0;
      for k := 0 to Amp_Size - 2 do begin
        JumpAmp := JumpAmp + W(k) * (Amp[k + 1] - Amp[k])
      end;
      JumpAmp := JumpAmp / Amp_A[Amp_Size, Amp_Size];
      for k := 0 to Amp_Size - 2 do begin
        JumpStd := JumpStd + W(k) * Sqr(Amp[k + 1] - Amp[k] - JumpAmp)
      end;
      JumpStd := Sqrt(JumpStd / Amp_A[Amp_Size, Amp_Size]);



      //Copy the Amps and re-calculate Stds

      for k := 0 to Amp_Size - 1 do begin
        Amp[k] := Amp_X[k];
      end;

      for k := 0 to prClsCount - 1 do begin
        Std[k] := 0;
      end;
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        if prParams.BaselineTrack then begin
          y := y - SBaseX[p];
        end;
        for i := 0 to prStateCount - 1 do begin
          Std[prClsIdxR[i]] := Std[prClsIdxR[i]] + Gamma[i, p] * Sqr(y - Amp[prClsIdxR[i]]);
        end;
      end;
      for k := 0 to prClsCount - 1 do begin
        if Gammat[k] <> 0 then begin
          Std[k] := Sqrt(Std[k] / Gammat[k]);
        end;
      end;

      JumpAmpLS := Amp_X[Amp_Size];//This result is identical to doing the average
      //JumpAmpLS := 0;
      JumpStdLS := 0;
      {
      for k := 0 to Amp_Size - 2 do begin
        JumpAmpLS := JumpAmpLS + W(k) * (Amp[k + 1] - Amp[k])
      end;
      JumpAmpLS := JumpAmpLS / Amp_A[Amp_Size, Amp_Size];
      }
      for k := 0 to Amp_Size - 2 do begin
        JumpStdLS := JumpStdLS + W(k) * Sqr(Amp[k + 1] - Amp[k] - JumpAmpLS)
      end;
      JumpStdLS := Sqrt(JumpStdLS / Amp_A[Amp_Size, Amp_Size]);


    finally

      Free2(TPMChar(Amp_A));
      Free2(TPMChar(Amp_AI));
      FreeMem(Amp_B);
      FreeMem(Amp_X);
      FreeMem(Gammat);
      FreeMem(Gammat_yt);

    end;

  end;
  *)

  procedure ReestAmpsStds;
  var
    c, p, s :Integer;
  begin

    //Amplitudes
    for c := 0 to prClsCount - 1 do begin
      Amp[c] := 0;//receives Gamma * y
      AM [c] := 0;//receives Gamma
    end;
    for p := 0 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      if prParams.BaselineTrack then begin
        y := y - SBaseX[p];
      end;
      for s := 0 to prStateCount - 1 do begin
        Amp[prClsIdxR[s]] := Amp[prClsIdxR[s]] + Gamma[s, p] * y;
        AM [prClsIdxR[s]] := AM [prClsIdxR[s]] + Gamma[s, p];
      end;
    end;
    for c := 0 to prClsCount - 1 do begin
      if AM[c] <> 0 then begin
        Amp[c] := Amp[c] / AM[c];
      end;
    end;

    //Noise
    for c := 0 to prClsCount - 1 do begin
      Std[c] := 0;
    end;
    for p := 0 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      if prParams.BaselineTrack then begin
        y := y - SBaseX[p];
      end;
      for s := 0 to prStateCount - 1 do begin
        Std[prClsIdxR[s]] := Std[prClsIdxR[s]] + Gamma[s, p] * Sqr(y - Amp[prClsIdxR[s]]);
      end;
    end;
    for c := 0 to prClsCount - 1 do begin
      if AM[c] <> 0 then begin
        Std[c] := Sqrt(Std[c] / AM[c]);
      end;
    end;



    //************************************************************************//
    //                                                                        //
    //                    C O N S T R A I N E D     A M P S                   //
    //                                                                        //
    //************************************************************************//
    if prModel.ModelObject.ConstraintsAmpVar.ConstraintCount > 0 then begin
    //Constrained re-estimation

      //Amplitudes
      for c := 0 to prClsCount - 1 do begin
        Sum_Gamma_y[c] := 0;
        Sum_Gamma  [c] := 0;
      end;
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        if prParams.BaselineTrack then begin
          y := y - SBaseX[p];
        end;
        for s := 0 to prStateCount - 1 do begin
          Sum_Gamma_y[prClsIdxR[s]] := Sum_Gamma_y[prClsIdxR[s]] + Gamma[s, p] * y;
          Sum_Gamma  [prClsIdxR[s]] := Sum_Gamma  [prClsIdxR[s]] + Gamma[s, p];
        end;
      end;

      CstrAmpEst(Cta, Sum_Gamma, Sum_Gamma_y, Amp);

      //Variances
      for c := 0 to prClsCount - 1 do begin
        Sum_Gamma_y[c] := 0;
        Sum_Gamma  [c] := 0;
      end;
      for p := 0 to DataCount - 1 do begin
        y := Data[p] * Scale + Off;
        if prParams.BaselineTrack then begin
          y := y - SBaseX[p];
        end;
        for s := 0 to prStateCount - 1 do begin
          Sum_Gamma_y[prClsIdxR[s]] := Sum_Gamma_y[prClsIdxR[s]] + Gamma[s, p] * Sqr(y - Amp[prClsIdxR[s]]);
          Sum_Gamma  [prClsIdxR[s]] := Sum_Gamma  [prClsIdxR[s]] + Gamma[s, p];
        end;
      end;
      CstrVarEst(Ctv, Sum_Gamma, Sum_Gamma_y, Std);


    end;

    //Bogus switch !!! Remove after testing
    //if prParams.BaselineReestPar then begin
    //  DoAmpStdLS;//!!!
    //end;

  end;//ReestAmpsStds

  procedure ReestAMatrix;
  var
    p, s, ss :Integer;
  begin

    //Clear
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss] := 0;
      end;//of ss
    end;//of s
    //Add Epsilon
    for p := 0 to DataCount - 2 do begin
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          AMatrix[s, ss] := AMatrix[s, ss] + TPMDouble(Epsilon[p])[s, ss];
        end;//of ss
      end;//of s
    end;
    //Rescale
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        Sum := Sum + AMatrix[s, ss];
      end;//of ss
      for ss := 0 to prStateCount - 1 do begin
        if Sum <> 0 then AMatrix[s, ss] := AMatrix[s, ss] / Sum;
      end;//of ss
    end;//of s

  end;//ReestAMatrix


begin


  //Make epsilon 2-d and add entries for each point

  Result              := False;


  //Nullify pointers ...........................................................
  Alpha               := nil;
  Betta               := nil;
  Gamma               := nil;
  Gauss               := nil;
  Rho                 := nil;
  Epsilon             := nil;

  State               := nil;

  AMatrix             := nil;
  AMatrixN            := nil;
  AM                  := nil;

  Prob0               := nil;

  BaseX               := nil;
  BaseP               := nil;

  SBaseX              := nil;
  SBaseP              := nil;
  SBaseQ              := nil;

  Amp                 := nil;
  AmpN                := nil;
  Std                 := nil;

  Sum_Gamma_y         := nil;
  Sum_Gamma           := nil;

  Cta                 := nil;
  Ctv                 := nil;



  //Signal Running status ......................................................
  Chunk               := TQUB_Thread_PChunk(IdlBaseChunk);
  Chunk.Status        := CHUNK_STATUS_RUNNING;


  //Movies
  if prMovie.Make then begin

    MovieTree := TNode.Create('IBMovie');
     PythonThread.BeginPython;
     PyNode_Publish(MovieTree, 'IBMovie'); // make tree available in Python as QUB.IBMovie
     PythonThread.EndPython;
     PythonThread.DonePython;

  end;



  try//finally



    //Get data parameters ......................................................
    //   - Data, DataCount
    Data              := IdlBaseChunk.Data;
    DataCount         := Chunk.Sel.Last - Chunk.Sel.First + 1;
    //   - Scale
    Scale             := 1 / IdlBaseChunk.Scaling;
    Off               := IdlBaseChunk.Offset;


    //Movies
    if prMovie.Make then begin

      MovieTree['Scaling'].SetData( IdlBaseChunk.Scaling );
      MovieTree['Offset'].SetData( IdlBaseChunk.Offset );
      MovieTree['DataCount'].SetData( DataCount );

      MovieData       := MovieTree['Amps'];
      MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, 1 );
      for s := 0 to prStateCount - 1 do begin
        MovieData.Data[s] := prAmp[prClsIdxR[s]];
      end;

    end;



    //Allocate memory ..........................................................


    //   - Vectors
    Alpha             := DAlloc2(prStateCount, DataCount);
    Betta             := DAlloc2(prStateCount, DataCount);
    Gamma             := DAlloc2(prStateCount, DataCount);
    Gauss             := DAlloc2(prStateCount, DataCount);
    Rho               := DAlloc1(DataCount);
    Epsilon           := PAlloc1(DataCount);
    for p := 0 to DataCount - 1 do begin
      Epsilon[p]      := DAlloc2(prStateCount, prStateCount);
    end;


    //   - A matrix
    AMatrix           := DAlloc2(prStateCount, prStateCount);
    AMatrixN          := DAlloc2(prStateCount, prStateCount);

    AM                := DAlloc1(prStateCount);


    //   - Prob0 - initial probabilities;
    Prob0             := DAlloc1(prStateCount);


    if prParams.BaselineTrack then begin
      //   - BaseX|P - baseline state|variance, for each data point
      BaseX           := DAlloc1(DataCount);
      BaseP           := DAlloc1(DataCount);
      //   - SBaseX|P|Q - smoothed baseline state, variance and BaseVar, for each data point
      SBaseX          := DAlloc1(DataCount);
      SBaseP          := DAlloc1(DataCount);
      SBaseQ          := DAlloc1(DataCount);
    end;


    //   - Amp, Std
    Amp               := DAlloc1(prStateCount);
    AmpN              := DAlloc1(prStateCount);
    Std               := DAlloc1(prStateCount);

    Sum_Gamma_y       := DAlloc1(prClsCount);
    Sum_Gamma         := DAlloc1(prClsCount);
    Cta               := DAlloc1(prModel.ModelObject.ConstraintsAmpVar.ConstraintCount + 1);
    Ctv               := DAlloc1(prModel.ModelObject.ConstraintsAmpVar.ConstraintCount + 1);


    //Initialize variables .....................................................


    //Amplitude Constraints ****************************************************
    if prModel.ModelObject.ConstraintsAmpVar.ConstraintCount > 0 then begin

      ca := 0;
      cv := 0;
      for i := 0 to prModel.ModelObject.ConstraintsAmpVar.ConstraintCount - 1 do begin

        Cstr := prModel.ModelObject.ConstraintsAmpVar[i];

        case Cstr.cType of

          //Amplitude
          CSTR_FIX_AMP:
            begin
              Cta[ca] := prAmp[Cstr.Index[0]];
              Inc(ca);
            end;
          CSTR_FIX_EXC_AMP:
            begin
              if Cstr.Index[0] = 0 then begin
                Cta[ca] := prAmp[0];
              end else begin
                Cta[ca] := prAmp[Cstr.Index[0]] - prAmp[0];
              end;
              Inc(ca);
            end;
          CSTR_FIX_DIFF_EXC_AMP:
            begin
              Cta[ca] := prAmp[Cstr.Index[1]] - prAmp[Cstr.Index[0]];
              Inc(ca);
            end;
          CSTR_SCALE_EXC_AMP:
            begin
              Cta[ca] := prAmp[Cstr.Index[1]] / prAmp[Cstr.Index[0]];
              Inc(ca);
            end;
          CSTR_SCALE_DIFF_EXC_AMP:
            begin
              Cta[ca] := (prAmp[Cstr.Index[1]] - prAmp[Cstr.Index[0]]) / (prAmp[Cstr.Index[3]] - prAmp[Cstr.Index[2]]);
              Inc(ca);
            end;
         //Variance
         CSTR_FIX_VAR:
            begin
              Ctv[cv] := Sqr(prStd[Cstr.Index[0]]);
              Inc(cv);
            end;
          CSTR_FIX_EXC_VAR:
            begin
              if Cstr.Index[0] = 0 then begin
                Ctv[cv] := Sqr(prStd[0]);
              end else begin
                Ctv[cv] := Sqr(prStd[Cstr.Index[0]]) - Sqr(prStd[0]);
              end;
              Inc(cv);
            end;
          CSTR_FIX_DIFF_EXC_VAR:
            begin
              Ctv[cv] := Sqr(prStd[Cstr.Index[1]]) - Sqr(prStd[Cstr.Index[0]]);
              Inc(cv);
            end;
          CSTR_SCALE_EXC_VAR:
            begin
              Ctv[cv] := Sqr(prStd[Cstr.Index[1]]) / Sqr(prStd[Cstr.Index[0]]);
              Inc(cv);
            end;
          CSTR_SCALE_DIFF_EXC_VAR:
            begin
              Ctv[cv] := (Sqr(prStd[Cstr.Index[1]]) - Sqr(prStd[Cstr.Index[0]])) / (Sqr(prStd[Cstr.Index[3]]) - Sqr(prStd[Cstr.Index[2]]));
              Inc(cv);
            end;

        end;//case

      end;

    end;


    //   - A matrix
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss]    := Abs(prA[s, ss]);//to prevent some occasional negative entries
        Sum               := Sum + AMatrix[s, ss];
      end;
      for ss := 0 to prStateCount - 1 do begin //re-normalize
        if Sum <> 0 then AMatrix[s, ss]    := AMatrix[s, ss] / Sum;
        AMatrixN[s, ss]   := AMatrix[s, ss];
      end;
    end;


    //   - Prob0
    for s := 0 to prStateCount - 1 do begin
      Prob0[s] := prP0[s];
    end;


    //   - Amp, Std
    for c := 0 to prClsCount - 1 do begin
      Amp[c] := prAmp[c];
      Std[c] := prStd[c];
    end;
    if prParams.BaselineTrack then begin
      for c := 1 to prClsCount - 1 do begin
        Amp[c] := prAmp[c] - prAmp[0];
        Std[c] := prStd[c];
      end;
      Amp[0]   := 0;
      Std[0]   := prStd[0];
    end;


    BaseStd := 0;
    BaseVar := 0;


    //   Baseline X0 and P0
    if prParams.BaselineTrack then begin
      BaseStd    := prParams.BaselineStd;
      BaseVar    := Sqr(BaseStd);

      //Experimental: initialize baseline as the first point
      p          := 0;
      y          := Data[p] * Scale + Off;
      BaseX[0]   := y;
      BaseP[0]   := 0;
      for s := 0 to prStateCount - 1 do begin
        BaseX[0] := BaseX[0] - Prob0[s] *     Amp[prClsIdxR[s]];
        BaseP[0] := BaseP[0] + Prob0[s] * Sqr(Std[prClsIdxR[s]]);
      end;

    end;


    RefAmp  := 0;


    //Print starting values ....................................................
    prConsoleMsg := 'Starting values:'#13#10;
    prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
      end;
      prConsoleMsg := prConsoleMsg + #13#10;
    end;
    prConsoleMsg := prConsoleMsg + 'State, Prob0:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [s, Prob0[s]]);
    end;
    prConsoleMsg := prConsoleMsg + 'Cls, Amp, Std:'#13#10;
    for c := 0 to prClsCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [prClsIdx[c], Amp[c], Std[c]]);
    end;
    if prParams.BaselineTrack then begin
      prConsoleMsg := prConsoleMsg + 'Baseline Std, Var:'#13#10;
      prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#13#10, [BaseStd, BaseVar]);
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    if not prParams.QuietOutput then Synchronize(prWriteNewConsole);



    try


      OldLL := 0;


      //Iterate
      for Iteration := 1 to prParams.MaxIter do begin


        NewLL := 0;


        //********************************************************************//
        //                                                                    //
        //                   F o r w a r d   -   K a l m a n                  //
        //                                                                    //
        //********************************************************************//

        //Experimental:

        //Initialize the first point
        p      := 0;
        y      := Data[p] * Scale + Off;
        if prParams.BaselineTrack then begin
          y    := y - BaseX[p];
        end;
        Rho[p] := 0;
        for s := 0 to prStateCount - 1 do begin
          Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
           //p(y|s0) * p(s0)
          Alpha[s, p] := Prob0[s] * Gauss[s, p];
          Rho[p]      := Rho[p] + Alpha[s, p];
        end;
        for s := 0 to prStateCount - 1 do begin
          Alpha[s, p] := Alpha[s, p] / Rho[p];
        end;


        NewLL := NewLL + SafeLog(Rho[p]);


        //Loop for every point .................................................
        for p := 1 to DataCount - 1 do begin


          if prParams.BaselineTrack then begin
            BaseX[p] := BaseX[p-1];
            BaseP[p] := BaseP[p-1];
          end;


          y              := Data[p] * Scale + Off;

          CanQuit        := True;
          BaseEstCount   := 0;


          repeat


            if prParams.BaselineTrack then begin
              RefAmp := BaseX[p];
            end;


            Rho[p] := 0;


            //for each state
            for s := 0 to prStateCount - 1 do begin

              Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y - RefAmp);

              Alpha[s, p] := 0;

              for ss := 0 to prStateCount - 1 do begin

                Alpha[s, p] := Alpha[s, p] +
                               AMatrix[ss, s] *
                               Alpha[ss, p - 1];

              end;//of ss

              Alpha[s, p] := Gauss[s, p] * Alpha[s, p];

              Rho[p]      := Rho[p]      + Alpha[s, p];

            end;//of s


            //Normalize Alpha
            for s := 0 to prStateCount - 1 do begin
              Alpha[s, p] := Alpha[s, p] / Rho[p];
            end;


            //Update baseline
            if prParams.BaselineTrack then begin
              if (p mod prParams.BaselineDecFactor) = 0 then begin

                //ThisAmp and ThisStd are linear combinations of states
                ThisAmp := 0;
                MeasVar := 0;
                for s := 0 to prStateCount - 1 do begin
                  ThisAmp := ThisAmp + Alpha[s, p] *     Amp[prClsIdxR[s]];
                  MeasVar := MeasVar + Alpha[s, p] * Sqr(Std[prClsIdxR[s]]);
                end;

                KalmanForward(BaseX[p], BaseP[p], BaseX[p-1], BaseP[p-1], MeasVar, BaseVar, (y - ThisAmp){BaseData});

                Inc(BaseEstCount);

                CanQuit := (Abs((RefAmp - BaseX[p] + 1e-10) / (RefAmp + 1e-10)) < BASE_CONV) or (BaseEstCount > prParams.BaselineFKCount);

              end else begin

                CanQuit := True;

              end;
            end;


          until(CanQuit);


          NewLL := NewLL + SafeLog(Rho[p]);


        end;//of p


        //==>Alpha


        if not prParams.BaselineTrack then begin


          DoBetta;//==>Betta

          DoGamma;//==>Gamma


          //********************************************************************//
          //                                                                    //
          //                              M O V I E S                           //
          //                                                                    //
          //********************************************************************//
          if prMovie.Make then begin

            MovieData       := MovieTree['Alpha'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(Alpha[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;
            MovieData       := MovieTree['Gamma'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(Gamma[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;

            prMovieIteration  := Iteration;
            prMovieFBKIter    := 0;
            Synchronize(prMakeMovie_ForwardBackwardKalman);

          end;



        end else begin
        //BaselineTrack


          //Now go backwards on the baseline to get the smooth baseline
          DoBaselineBack;//==>SBaseX, SBaseP


          if prParams.BaselineBackKalman then begin


            //********************************************************************//
            //                                                                    //
            //                  B a c k w a r d   -   K a l m a n                 //
            //                                                                    //
            //********************************************************************//


            //Initialize last point
            p         := DataCount - 1;
            y         := Data[p] * Scale - BaseX[p];

            //Betta of last point
            for s := 0 to prStateCount - 1 do begin
              Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
              Betta[s, p] := Gauss[s, p] * (1 / prStateCount);
            end;


            //Loop for every point .................................................
            for p := DataCount - 2 downto 0 do begin


              BaseX[p]      := BaseX[p+1];
              BaseP[p]      := BaseP[p+1];

              y             := Data[p] * Scale + Off;

              BaseEstCount  := 0;


              repeat


                RefAmp := BaseX[p];


                //Betta t

                Rho[p]  := 0;

                //for each state
                for s := 0 to prStateCount - 1 do begin

                  Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y - RefAmp);

                  Betta[s, p] := 0;

                  for ss := 0 to prStateCount - 1 do begin

                    Betta[s, p] := Betta[s, p] +
                                   AMatrix[s, ss] *
                                   Betta[ss, p + 1];

                  end;//of ss

                  Betta[s, p]  := Gauss[s, p] * Betta[s, p];

                  Rho[p]       := Rho[p]      + Betta[s, p];

                end;//of s

                //Normalize Betta
                for s := 0 to prStateCount - 1 do begin
                  Betta[s, p] := Betta[s, p] / Rho[p];
                end;


                //Update baseline

                if ((p + 1) mod prParams.BaselineDecFactor) = 0 then begin

                  //ThisAmp and ThisStd are linear combinations of states
                  ThisAmp := 0;
                  MeasVar := 0;
                  Sum     := 0;
                  for s := 0 to prStateCount - 1 do begin
                    ThisAmp := ThisAmp + Betta[s, p] *     Amp[prClsIdxR[s]];
                    MeasVar := MeasVar + Betta[s, p] * Sqr(Std[prClsIdxR[s]]);
                    Sum     := Sum     + Betta[s, p];
                  end;
                  ThisAmp := ThisAmp / Sum;
                  MeasVar := MeasVar / Sum;

                  KalmanForward(BaseX[p], BaseP[p], BaseX[p+1], BaseP[p+1], MeasVar, BaseVar, (y - ThisAmp){BaseData});

                  Inc(BaseEstCount);

                  CanQuit := (Abs((RefAmp - BaseX[p] + 1e-10) / (RefAmp + 1e-10)) < BASE_CONV) or (BaseEstCount > prParams.BaselineFKCount);

                end else begin

                  CanQuit := True;

                end;

              until(CanQuit);


            end;//of p


            //Smooth back baseline and choose best
            DoBaselineBBack;//==> SBaseX, SBaseP

            NewLL := DoForward;//==> Alpha


          end;


          DoBetta;//==>Betta

          DoGamma;//==>Gamma



          //********************************************************************//
          //                                                                    //
          //                              M O V I E S                           //
          //                                                                    //
          //********************************************************************//
          if prMovie.Make then begin

            MovieData       := MovieTree['Alpha'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(Alpha[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;
            MovieData       := MovieTree['Gamma'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(Gamma[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;

            MovieData       := MovieTree['BaseX'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(BaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['BaseP'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(BaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['SBaseX'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(SBaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['SBaseP'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(SBaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);

            prMovieIteration  := Iteration;
            prMovieFBKIter    := 0;
            Synchronize(prMakeMovie_ForwardBackwardKalman);

          end;



          //Now alternate between HMM forward-backward and LGM forward-backward
          for BaseEstCount := 1 to prParams.BaselineFBKCount do begin


            //****************************************************************//
            //                                                                //
            //                           K a l m a n                          //
            //                                                                //
            //                         (given  Markov)                        //
            //                                                                //
            //****************************************************************//

            //Forward **********************************************************
            DoBaselineForward;//==>BaseX, BaseP

            //Backward *********************************************************
            DoBaselineBack;//==> SBaseX, SBaseP


            //****************************************************************//
            //                                                                //
            //                            M O V I E S                         //
            //                                                                //
            //****************************************************************//
            if prMovie.Make then begin

              MovieData       := MovieTree['BaseX'];
              MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
              Move(BaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
              MovieData       := MovieTree['BaseP'];
              MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
              Move(BaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
              MovieData       := MovieTree['SBaseX'];
              MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
              Move(SBaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
              MovieData       := MovieTree['SBaseP'];
              MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
              Move(SBaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);

              prMovieIteration  := Iteration;
              prMovieFBKIter    := BaseEstCount;
              Synchronize(prMakeMovie_ForwardBackwardKalman);

            end;


            //****************************************************************//
            //                                                                //
            //               F o r w a r d   -   B a c k w a r d              //
            //                                                                //
            //                        (given Baseline)                        //
            //                                                                //
            //****************************************************************//

            //Forward **********************************************************

            NewLL := DoForward;//==>Alpha

            //Backward *********************************************************

            DoBetta;//==>Betta

            DoGamma;//==>Gamma

            //Initial Probabilities
            for s := 0 to prStateCount - 1 do begin
              Prob0[s] := Gamma[s, 0];
            end;

            //==> Prob0

            if prParams.BaselineReestPar then begin

              //Amplitudes and Noise
              ReestAmpsStds;

              //Eliminate offset in Amp and Baseline
              if prParams.BaselineTrack then begin
                SBaseX[0] := SBaseX[0] + Amp[prClsIdx[0]];
                BaseX[0]  := SBaseX[0];
                BaseP[0]  := SBaseP[0];
                Sum       := Amp[prClsIdx[0]];
                for c := 0 to prClsCount - 1 do begin
                  Amp[c] := Amp[c] - Sum;
                end;
              end;
              //==> Amps, Stds

              //Movies
              if prMovie.Make then begin

                MovieData       := MovieTree['Amps'];
                for s := 0 to prStateCount - 1 do begin
                  MovieData.Data[s] := Amp[prClsIdxR[s]];
                end;

              end;

              //A Matrix
              if not prParams.FixKinetics then begin
                DoEpsilon;    //==>Epsilon
                ReestAMatrix; //==>A
              end;
              if (prParams.A2Q) and (prParams.A2QEveryIter) then begin
                prA2Q(AMatrix, nil, False{Q});
              end;

            end;


            //****************************************************************//
            //                                                                //
            //                            M O V I E S                         //
            //                                                                //
            //****************************************************************//
            if prMovie.Make then begin

              MovieData       := MovieTree['Alpha'];
              MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
              for s := 0 to prStateCount - 1 do begin
                Move(Alpha[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
              end;
              MovieData       := MovieTree['Gamma'];
              MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
              for s := 0 to prStateCount - 1 do begin
                Move(Gamma[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
              end;

              prMovieIteration  := Iteration;
              prMovieFBKIter    := -BaseEstCount;
              Synchronize(prMakeMovie_ForwardBackwardKalman);

            end;


          end;//of BaseEstCount


        end;//of BaselineTrack


        if prParams.Reestimation = IDLBASE_REEST_NONE then begin
          Break;
        end;



        //Re-estimation ********************************************************


        //Amplitudes and Noise
        ReestAmpsStds;

        //Eliminate offset in Amp and Baseline
        if prParams.BaselineTrack then begin
          SBaseX[0] := SBaseX[0] + Amp[prClsIdx[0]];
          BaseX[0]  := SBaseX[0];
          BaseP[0]  := SBaseP[0];
          Sum       := Amp[prClsIdx[0]];
          for c := 0 to prClsCount - 1 do begin
            Amp[c] := Amp[c] - Sum;
          end;
        end;
        //==> Amps, Stds

        //Movies
        if prMovie.Make then begin
          MovieData       := MovieTree['Amps'];
          for s := 0 to prStateCount - 1 do begin
            MovieData.Data[s] := Amp[prClsIdxR[s]];
          end;
        end;

        //Initial Probabilities
        for s := 0 to prStateCount - 1 do begin
          Prob0[s] := Gamma[s, 0];
        end;

        //==> Prob0

        //A Matrix
        if not prParams.FixKinetics then begin
          DoEpsilon;    //==>Epsilon
          ReestAMatrix; //==>A
        end;

        //Print Q matrix as the logarithm of the A matrix ......................
        MatrixLogarithm(AMatrix, AMatrixN{Q}, IDLBaseChunk.Sampling, prStateCount, 100, 1e-4);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            if AMatrixN[s, ss] < 0 then begin
              AMatrixN[s, ss] := 0;
              AMatrixN[ss, s] := 0;
            end;
          end;
        end;
        prConsoleMsg := #13#10'Q Matrix:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            if (s = ss) or (prQ[s, ss] = 0)then begin
              prConsoleMsg := prConsoleMsg + #9'...           ';
            end else begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrixN[s, ss]]);
            end;
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);

        if (prParams.A2Q) and (prParams.A2QEveryIter) then begin
          (*
          prA2Q(AMatrixN{Q}, nil, True{Q});
          prConsoleMsg := #13#10'A(QQ) Matrix:'#13#10;
          for s := 0 to prStateCount - 1 do begin
            for ss := 0 to prStateCount - 1 do begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrixN[s, ss]]);
            end;
            prConsoleMsg := prConsoleMsg + #13#10;
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);
          *)
          {
          AMatrix[0,0] := 0.98259481166;
          AMatrix[0,1] := 0.017405182501;
          AMatrix[0,2] := 5.8406285565e-9;
          AMatrix[1,0] := 0.00042073968802;
          AMatrix[1,1] := 0.94931094501;
          AMatrix[1,2] := 0.050268315299;
          AMatrix[2,0] := 4.9966791294e-11;
          AMatrix[2,1] := 0.017759241948;
          AMatrix[2,2] := 0.982240758;
          }

          prA2Q(AMatrix, AMatrixN, False{Q});
          //prA2Q_SolvOpt(AMatrix);
          {
          prConsoleMsg := #13#10'A* Matrix:'#13#10;
          for s := 0 to prStateCount - 1 do begin
            for ss := 0 to prStateCount - 1 do begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
            end;
            prConsoleMsg := prConsoleMsg + #13#10;
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);
          prConsoleMsg := #13#10'A Matrix:'#13#10;
          for s := 0 to prStateCount - 1 do begin
            for ss := 0 to prStateCount - 1 do begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrixN[s, ss]]);
            end;
            prConsoleMsg := prConsoleMsg + #13#10;
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);
          prConsoleMsg := #13#10'Sqr((A-A*)/A) Matrix:'#13#10;
          for s := 0 to prStateCount - 1 do begin
            for ss := 0 to prStateCount - 1 do begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [Sqr((AMatrixN[s, ss]-AMatrix[s, ss])/AMatrixN[s, ss])]);
            end;
            prConsoleMsg := prConsoleMsg + #13#10;
          end;
          prConsoleCode := MESSAGE_MESSAGE;
          if not prParams.QuietOutput then Synchronize(prWriteNewConsole);
          }
          for s := 0 to prStateCount - 1 do begin
            for ss := 0 to prStateCount - 1 do begin
              AMatrix[s, ss] := AMatrixN[s, ss];
            end;
          end;
        end;

        //==> A

        //Recalculate LL with reestimated parameters
        NewLL := DoForward;





        //Print current iteration values .......................................
        prConsoleMsg := Format(#13#10'Iteration %d'#13#10, [Iteration]);
        prConsoleMsg := prConsoleMsg + Format('LL: %14.10g'#13#10, [NewLL]);
        prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleMsg := prConsoleMsg + 'State, Prob0:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [s, Prob0[s]]);
        end;
        prConsoleMsg := prConsoleMsg + 'Cls, Amp, Std:'#13#10;
        for c := 0 to prClsCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [prClsIdx[c], Amp[c], Std[c]]);
        end;
        if prParams.BaselineTrack then begin
          prConsoleMsg := prConsoleMsg + 'Baseline Var, Std, X0, P0:'#13#10;
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#13#10#9'%-14.6g'#13#10#9'%-14.6g'#13#10#9'%-14.6g'#13#10, [BaseVar, Sqrt(Abs(BaseVar)), SBaseX[0], SBaseP[0]]);
        end;
        //{testing}prConsoleMsg := prConsoleMsg + Format(#9'JumpAmp   = %-14.6g, JumpStd   = %-14.6g'#13#10#9'JumpAmpLS = %-14.6g, JumpStdLS = %-14.6g', [JumpAmp, JumpStd, JumpAmpLS, JumpStdLS]);
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


        if prStopped then Break;


        //Check convergence
        if Abs(NewLL - OldLL) <= prParams.LLConv then Break;


        OldLL := NewLL;


      end;//of Iteration


      //Results ................................................................
      //if not prStopped then begin

        //Make the state vector
        State := IAlloc1(DataCount);
        for p := 0 to DataCount - 1 do begin
          //{
          State[p] := 0;
          Sum := Gamma[0, p];//it's not Sum but Max
          for s := 1 to prStateCount - 1 do begin
            if Gamma[s, p] > Sum then begin
              State[p] := s;
              Sum := Gamma[s, p];
            end;
          end;
          //}
          {
          for c := 0 to prClsCount - 1 do begin
            AmpN[c] := 0;
          end;
          for s := 0 to prStateCount - 1 do begin
            AmpN[prClsIdxR[s]] := AmpN[prClsIdxR[s]] + Gamma[s, p];
          end;
          State[p] := 0;
          Sum      := AmpN[0];
          for c := 1 to prClsCount - 1 do begin
            if AmpN[c] > Sum then begin
              State[p] := c;
              Sum      := AmpN[c];
            end;
          end;
          for s := 0 to prStateCount - 1 do begin
            if prClsOfState[s] = State[p] then begin
              State[p] := s;
              Break;
            end;
          end;
          }
        end;


        //Make Dwells
        prMakeDwells(IdlBaseChunk, State, DataCount);

        IdlBaseChunk.ClassCount := prClsCount;
        IdlBaseChunk.StateCount := prStateCount;
        IdlBaseChunk.Amps       := SAlloc1(prClsCount);
        IdlBaseChunk.Stds       := SAlloc1(prClsCount);
        IdlBaseChunk.AMatrix    := DAlloc2(prStateCount, prStateCount);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            IdlBaseChunk.AMatrix[s, ss] := AMatrix[s, ss];
          end;
        end;
        for c := 0 to prClsCount - 1 do begin
          IdlBaseChunk.Amps[c]  := Amp[c];
          IdlBaseChunk.Stds[c]  := Std[c];
        end;

        IdlBaseChunk.LL         := NewLL;
        IdlBaseChunk.Iterations := Iteration - 1;


        //Make Baseline Nodes
        if (prParams.BaselineTrack) and (prParams.BaselineCorrect) then begin

          prBase2Nodes(IdlBaseChunk, SBaseX);

        end;


        if (prParams.A2Q) and (not prParams.A2QEveryIter) then begin
          prA2Q{_SolvOpt}(AMatrix, nil, False{Q});
        end;


        //Make Stats and Histograms
        prMakeStats(IdlBaseChunk, SBaseX); // SBaseX is nil if not prParams.BaselineTrack


      //end;//Results



    except


      Result            := True;

      Chunk.Status      := CHUNK_STATUS_ERROR;
      Chunk.Completion  := 0;


    end;


  finally


    //Free Memory ..............................................................
    Free2(TPMChar(Alpha));
    Free2(TPMChar(Betta));
    Free2(TPMChar(Gamma));
    Free2(TPMChar(Gauss));
    FreeMem(Rho);
    if Epsilon <> nil then begin
      for p := 0 to DataCount - 1 do begin
        Free2(TPMChar(Epsilon[p]));
      end;
    end;
    FreeMem(Epsilon);

    FreeMem(State);

    FreeMem(BaseX);
    FreeMem(BaseP);
    FreeMem(SBaseX);
    FreeMem(SBaseP);
    FreeMem(SBaseQ);

    Free2(TPMChar(AMatrix));
    Free2(TPMChar(AMatrixN));

    FreeMem(AM);

    FreeMem(Prob0);

    FreeMem(Amp);
    FreeMem(AmpN);
    FreeMem(Std);

    FreeMem(Sum_Gamma_y);
    FreeMem(Sum_Gamma);

    FreeMem(Cta);
    FreeMem(Ctv);


    //Signal new status
    if Chunk.Status <> CHUNK_STATUS_ERROR then begin
      Chunk.Status := CHUNK_STATUS_DONE;
      Chunk.Completion := 100;
    end;


  end;

end;



//****************************************************************************//
//                                                                            //
//                  B A U M  -  W E L C H  -  K A L M A N   II                //
//                                                                            //
//****************************************************************************//
function TQUB_Task_IdlBase.prReestBWKII(IdlBaseChunk :TQUB_IdlBase_PChunk) :Boolean;
var
  Chunk               :TQUB_Thread_PChunk;

  Data                :TPAInteger;
  DataCount           :Integer;
  Scale               :Double;
  Off                 :Double;

  Iteration           :Integer;

  //*********************************
  Alpha               :TPMDouble;
  Betta               :TPMDouble;
  Gamma               :TPMDouble;
  Omega               :TPMDouble;
  Gauss               :TPMDouble;
  BGauss              :TPMDouble;
  Rho                 :TPADouble;
  Epsilon             :TPAPointer;

  State               :TPAInteger;

  BaseX               :TPADouble;
  BaseP               :TPADouble;
  BBaseX              :TPADouble;
  BBaseP              :TPADouble;
  SBaseX              :TPADouble;
  SBaseP              :TPADouble;
  BSBaseX             :TPADouble;
  BSBaseP             :TPADouble;

  AMatrix             :TPMDouble;
  AMatrixN            :TPMDouble;
  AM                  :TPADouble;

  Prob0               :TPADouble;

  Amp, AmpN           :TPADouble;
  Std                 :TPADouble;

  //*********************************

  Sum                 :Double;

  y                   :Double;

  c, p, s, ss         :Integer;

  OldLL, NewLL        :Double;

  CanQuit             :Boolean;
  BaseEstCount        :Integer;

  RefAmp              :Double;
  ThisAmp             :Double;

  MeasVar             :Double;

  BaseVar             :Double;
  BaseStd             :Double;

  nm, np              :Double;

  oJ, cJ              :Double;

  yl, hl, kl          :Double;

  kxp, kpp            :Double;

  k                   :Double;

  //Movies
  MovieData           :INode;


  procedure DoBaselineBack;
  var
    p, s :Integer;
  begin

    p  := DataCount - 1;
    nm := 0;
    for s := 0 to prStateCount - 1 do begin
      nm := nm + Alpha[s, p] * Sqr(Std[prClsIdxR[s]]);
    end;

    //SBase[p] = Base[p]
    SBaseX[p] := BaseX[p];
    SBaseP[p] := BaseP[p];

    oJ := 0;

    for p := DataCount - 1 downto 1 do begin

      //pp = a . P[p-1] . a'  + BaseVar
      //cJ = P[p-1] . a' . (pp)-1
      cJ          := BaseP[p-1] / (BaseP[p-1] + BaseVar);

      //SBaseX[p-1] = BaseX[p-1] + cJ . (SBaseX[p] - a . BaseX[p-1])
      SBaseX[p-1] := BaseX[p-1] + cJ *      (SBaseX[p] - BaseX[p-1]);

      //SBaseP[p-1] = BaseP[p-1] + cJ . (SBaseP[p] - pp) . cJ'
      SBaseP[p-1] := BaseP[p-1] + cJ * (SBaseP[p] - (BaseP[p-1] + BaseVar)) * cJ;

      oJ := cJ;

    end;

  end;//DoBaselineBack;

  procedure DoBaselineBBack;
  var
    p, s :Integer;
  begin

    p  := 0;
    nm := 0;
    for s := 0 to prStateCount - 1 do begin
      nm := nm + Alpha[s, p] * Sqr(Std[prClsIdxR[s]]);
    end;

    BSBaseX[p] := BBaseX[p];
    BSBaseP[p] := BBaseP[p];

    oJ := 0;

    for p := 1 to DataCount - 2 do begin

      cJ          := BBaseP[p+1] / (BBaseP[p+1] + BaseVar);

      BSBaseX[p+1] := BBaseX[p+1] + cJ * (BSBaseX[p] -  BBaseX[p+1]);

      BSBaseP[p+1] := BBaseP[p+1] + cJ * (BSBaseP[p] - (BBaseP[p+1] + BaseVar)) * cJ;

      oJ := cJ;

    end;

  end;//DoBaselineBBack;

  procedure DoBetta;
  var
    p, s, ss :Integer;
  begin

    //Initialize last point
    p := DataCount - 1;
    for s := 0 to prStateCount - 1 do begin
      Betta[s, p] := 1 / Rho[p];
    end;//of s
    //Calculate Betta
    for p := DataCount - 2 downto 0 do begin
      for s := 0 to prStateCount - 1 do begin
        Betta[s, p] := 0;
        for ss := 0 to prStateCount - 1 do begin
          Betta[s, p] := Betta[s, p] + AMatrix[s, ss] * Gauss[ss, p + 1] * Betta[ss, p + 1];
        end;//of ss
        Betta[s, p] := Betta[s, p] / Rho[p];
      end;//of s
    end;//of p

  end;//DoBetta

  procedure DoGamma;
  var
    p, s :Integer;
  begin

    for p := 0 to DataCount - 1 do begin
      Sum := 0;
      for s := 0 to prStateCount - 1 do begin
        Gamma[s, p] := Alpha[s, p] * Betta[s, p];
        Sum := Sum + Gamma[s, p];
      end;//of s
      //Rescale
      for s := 0 to prStateCount - 1 do begin
        Gamma[s, p] := Gamma[s, p] / Sum;
      end;//of s
    end;//of p

  end;//DoGamma

  procedure DoEpsilon;
  var
    p, s, ss :Integer;
  begin

    for p := 0 to DataCount - 2 do begin
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          TPMDouble(Epsilon[p])[s, ss] := Alpha[s, p] * AMatrix[s, ss] * Gauss[ss, p + 1] * Betta[ss, p + 1];
        end;//of ss
      end;//of s
    end;//of p

  end;//DoEpsilon

  function DoForward :Double;
  var
    p, s, ss :Integer;
  begin
    Result := 0;

    //Initialize the first point
    p := 0;
    y := Data[p] * Scale + Off;
    if prParams.BaselineTrack then begin
      y := y - SBaseX[p];
    end;
    Rho[p] := 0;
    for s := 0 to prStateCount - 1 do begin
      Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
      Alpha[s, p] := Prob0[s] * Gauss[s, p];
      Rho[p]      := Rho[p] + Alpha[s, p];
    end;
    for s := 0 to prStateCount - 1 do begin
      Alpha[s, p] := Alpha[s, p] / Rho[p];
    end;
    Result := Result + SafeLog(Rho[p]);
    //Loop for every point
    for p := 1 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      if prParams.BaselineTrack then begin
        y := y - SBaseX[p];
      end;
      Rho[p] := 0;
      for s := 0 to prStateCount - 1 do begin
        Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
        Alpha[s, p] := 0;
        for ss := 0 to prStateCount - 1 do begin
          Alpha[s, p] := Alpha[s, p] + AMatrix[ss, s] * Alpha[ss, p - 1];
        end;
        Alpha[s, p] := Alpha[s, p] * Gauss[s, p];
        Rho[p] := Rho[p] + Alpha[s, p];
      end;//of s
      for s := 0 to prStateCount - 1 do begin
        Alpha[s, p] := Alpha[s, p] / Rho[p];
      end;
      Result := Result + SafeLog(Rho[p]);
    end;

  end;//DoForward

  procedure ReestAmpsStds;
  var
    c, p, s :Integer;
  begin


    //Amplitudes
    for c := 0 to prClsCount - 1 do begin
      Amp[c] := 0;//receives Gamma * y
      AM [c] := 0;//receives Gamma
    end;
    for p := 0 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      if prParams.BaselineTrack then begin
        y := y - SBaseX[p];
      end;
      for s := 0 to prStateCount - 1 do begin
        Amp[prClsIdxR[s]] := Amp[prClsIdxR[s]] + Gamma[s, p] * y;
        AM [prClsIdxR[s]] := AM [prClsIdxR[s]] + Gamma[s, p];
      end;
    end;
    for c := 0 to prClsCount - 1 do begin
      if AM[c] <> 0 then begin
        Amp[c] := Amp[c] / AM[c];
      end;
    end;

    //Noise
    for c := 0 to prClsCount - 1 do begin
      Std[c] := 0;
    end;
    for p := 0 to DataCount - 1 do begin
      y := Data[p] * Scale + Off;
      if prParams.BaselineTrack then begin
        y := y - SBaseX[p];
      end;
      for s := 0 to prStateCount - 1 do begin
        Std[prClsIdxR[s]] := Std[prClsIdxR[s]] + Gamma[s, p] * Sqr(y - Amp[prClsIdxR[s]]);
      end;
    end;
    for c := 0 to prClsCount - 1 do begin
      if AM[c] <> 0 then begin
        Std[c] := Sqrt(Std[c] / AM[c]);
      end;
    end;


  end;//ReestAmpsStds

  procedure ReestAMatrix;
  var
    p, s, ss :Integer;
  begin

    //Clear
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss] := 0;
      end;//of ss
    end;//of s
    //Add Epsilon
    for p := 0 to DataCount - 2 do begin
      for s := 0 to prStateCount - 1 do begin
        for ss := 0 to prStateCount - 1 do begin
          AMatrix[s, ss] := AMatrix[s, ss] + TPMDouble(Epsilon[p])[s, ss];
        end;//of ss
      end;//of s
    end;
    //Rescale
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        Sum := Sum + AMatrix[s, ss];
      end;//of ss
      for ss := 0 to prStateCount - 1 do begin
        if Sum <> 0 then AMatrix[s, ss] := AMatrix[s, ss] / Sum;
      end;//of ss
    end;//of s

  end;//ReestAMatrix


begin


  //Make epsilon 2-d and add entries for each point

  Result              := False;


  //Nullify pointers ...........................................................
  Alpha               := nil;
  Betta               := nil;
  Gamma               := nil;
  Omega               := nil;
  Gauss               := nil;
  BGauss              := nil;
  Rho                 := nil;
  Epsilon             := nil;

  State               := nil;

  AMatrix             := nil;
  AMatrixN            := nil;
  AM                  := nil;

  Prob0               := nil;

  BaseX               := nil;
  BaseP               := nil;

  BBaseX              := nil;
  BBaseP              := nil;

  SBaseX              := nil;
  SBaseP              := nil;

  BSBaseX             := nil;
  BSBaseP             := nil;

  Amp                 := nil;
  AmpN                := nil;
  Std                 := nil;




  //Signal Running status ......................................................
  Chunk               := TQUB_Thread_PChunk(IdlBaseChunk);
  Chunk.Status        := CHUNK_STATUS_RUNNING;


  //Movies
  if prMovie.Make then begin

    MovieTree := TNode.Create('IBMovie');
     PythonThread.BeginPython;
     PyNode_Publish(MovieTree, 'IBMovie'); // make tree available in Python as QUB.IBMovie
     PythonThread.EndPython;
     PythonThread.DonePython;

  end;



  try//finally



    //Get data parameters ......................................................
    //   - Data, DataCount
    Data              := IdlBaseChunk.Data;
    DataCount         := Chunk.Sel.Last - Chunk.Sel.First + 1;
    //   - Scale
    Scale             := 1 / IdlBaseChunk.Scaling;
    Off               := IdlBaseChunk.Offset;


    //Movies
    if prMovie.Make then begin

      MovieTree['Scaling'].SetData( IdlBaseChunk.Scaling );
      MovieTree['Offset'].SetData( IdlBaseChunk.Offset );
      MovieTree['DataCount'].SetData( DataCount );

      MovieData       := MovieTree['Amps'];
      MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, 1 );
      for s := 0 to prStateCount - 1 do begin
        MovieData.Data[s] := prAmp[prClsIdxR[s]];
      end;

    end;



    //Allocate memory ..........................................................


    //   - Vectors
    Alpha             := DAlloc2(prStateCount, DataCount);
    Betta             := DAlloc2(prStateCount, DataCount);
    Gamma             := DAlloc2(prStateCount, DataCount);
    Omega             := DAlloc2(prStateCount, DataCount);
    Gauss             := DAlloc2(prStateCount, DataCount);
    BGauss            := DAlloc2(prStateCount, DataCount);
    Rho               := DAlloc1(DataCount);
    Epsilon           := PAlloc1(DataCount);
    for p := 0 to DataCount - 1 do begin
      Epsilon[p]      := DAlloc2(prStateCount, prStateCount);
    end;


    //   - A matrix
    AMatrix           := DAlloc2(prStateCount, prStateCount);
    AMatrixN          := DAlloc2(prStateCount, prStateCount);

    AM                := DAlloc1(prStateCount);


    //   - Prob0 - initial probabilities;
    Prob0             := DAlloc1(prStateCount);


    //   - BaseX|P - baseline state|variance, for each data point
    BaseX           := DAlloc1(DataCount);
    BaseP           := DAlloc1(DataCount);
    //   - BBaseX|P - baseline state|variance, for each data point
    BBaseX          := DAlloc1(DataCount);
    BBaseP          := DAlloc1(DataCount);
    //   - SBaseX|P - smoothed baseline state|variance, for each data point
    SBaseX          := DAlloc1(DataCount);
    SBaseP          := DAlloc1(DataCount);
    //   - BSBaseX|P - smoothed baseline state|variance, for each data point
    BSBaseX         := DAlloc1(DataCount);
    BSBaseP         := DAlloc1(DataCount);


    //   - Amp, Std
    Amp               := DAlloc1(prStateCount);
    AmpN              := DAlloc1(prStateCount);
    Std               := DAlloc1(prStateCount);



    //Initialize variables .....................................................


    //   - A matrix
    for s := 0 to prStateCount - 1 do begin
      Sum := 0;
      for ss := 0 to prStateCount - 1 do begin
        AMatrix[s, ss]    := Abs(prA[s, ss]);//to prevent some occasional negative entries
        Sum               := Sum + AMatrix[s, ss];
      end;
      for ss := 0 to prStateCount - 1 do begin //re-normalize
        if Sum <> 0 then AMatrix[s, ss]    := AMatrix[s, ss] / Sum;
        AMatrixN[s, ss]   := AMatrix[s, ss];
      end;
    end;


    //   - Prob0
    for s := 0 to prStateCount - 1 do begin
      Prob0[s] := prP0[s];
    end;


    //   - Amp, Std
    for c := 0 to prClsCount - 1 do begin
      Amp[c] := prAmp[c];
      Std[c] := prStd[c];
    end;
    for c := 1 to prClsCount - 1 do begin
      Amp[c] := prAmp[c] - prAmp[0];
      Std[c] := prStd[c];
    end;
    Amp[0]   := 0;
    Std[0]   := prStd[0];


    BaseStd := 0;
    BaseVar := 0;


    //   Baseline X0 and P0
    BaseStd    := prParams.BaselineStd;
    BaseVar    := Sqr(BaseStd);

    //Experimental: initialize baseline as the first point
    p          := 0;
    y          := Data[p] * Scale + Off;
    BaseX[0]   := y;
    BaseP[0]   := 0;
    for s := 0 to prStateCount - 1 do begin
      BaseX[0] := BaseX[0] - Prob0[s] *     Amp[prClsIdxR[s]];
      BaseP[0] := BaseP[0] + Prob0[s] * Sqr(Std[prClsIdxR[s]]);
    end;


    //RefAmp  := 0;


    //Print starting values ....................................................
    prConsoleMsg := 'Starting values:'#13#10;
    prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      for ss := 0 to prStateCount - 1 do begin
        prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
      end;
      prConsoleMsg := prConsoleMsg + #13#10;
    end;
    prConsoleMsg := prConsoleMsg + 'State, Prob0:'#13#10;
    for s := 0 to prStateCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [s, Prob0[s]]);
    end;
    prConsoleMsg := prConsoleMsg + 'Cls, Amp, Std:'#13#10;
    for c := 0 to prClsCount - 1 do begin
      prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [prClsIdx[c], Amp[c], Std[c]]);
    end;
    if prParams.BaselineTrack then begin
      prConsoleMsg := prConsoleMsg + 'Baseline Std, Var:'#13#10;
      prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#9'%-14.6g'#13#10, [BaseStd, BaseVar]);
    end;
    prConsoleCode := MESSAGE_MESSAGE;
    if not prParams.QuietOutput then Synchronize(prWriteNewConsole);



    try


      OldLL := 0;


      //Iterate
      for Iteration := 1 to prParams.MaxIter do begin


        NewLL := 0;


        //********************************************************************//
        //                                                                    //
        //                   F o r w a r d   -   K a l m a n                  //
        //                                                                    //
        //********************************************************************//


        //Initialize the first point
        p      := 0;
        y      := Data[p] * Scale + Off - BaseX[p];
        Rho[p] := 0;
        for s := 0 to prStateCount - 1 do begin
          Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
           //p(y|s0) * p(s0)
          Alpha[s, p] := Prob0[s] * Gauss[s, p];
          Rho[p]      := Rho[p] + Alpha[s, p];
        end;
        for s := 0 to prStateCount - 1 do begin
          Alpha[s, p] := Alpha[s, p] / Rho[p];
        end;


        NewLL := NewLL + SafeLog(Rho[p]);


        //Loop for every point .................................................
        for p := 1 to DataCount - 1 do begin


          BaseX[p]       := BaseX[p-1];
          BaseP[p]       := BaseP[p-1];

          y              := Data[p] * Scale + Off;

          BaseEstCount   := 0;


          repeat


            RefAmp := BaseX[p];


            Rho[p] := 0;


            //for each state
            for s := 0 to prStateCount - 1 do begin

              Gauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y - RefAmp);

              Alpha[s, p] := 0;

              for ss := 0 to prStateCount - 1 do begin

                Alpha[s, p] := Alpha[s, p] +
                               AMatrix[ss, s] *
                               Alpha[ss, p - 1];

              end;//of ss

              Alpha[s, p] := Gauss[s, p] * Alpha[s, p];

              Rho[p]      := Rho[p]      + Alpha[s, p];

            end;//of s


            //Normalize Alpha
            for s := 0 to prStateCount - 1 do begin
              Alpha[s, p] := Alpha[s, p] / Rho[p];
            end;


            //Update baseline

            //ThisAmp and ThisStd are linear combinations of states
            ThisAmp := 0;
            MeasVar := 0;
            for s := 0 to prStateCount - 1 do begin
              ThisAmp := ThisAmp + Alpha[s, p] *     Amp[prClsIdxR[s]];
              MeasVar := MeasVar + Alpha[s, p] * Sqr(Std[prClsIdxR[s]]);
            end;

            KalmanForward(BaseX[p], BaseP[p], BaseX[p-1], BaseP[p-1], MeasVar, BaseVar, (y - ThisAmp){BaseData});

            Inc(BaseEstCount);

            CanQuit := (Abs((RefAmp - BaseX[p] + 1e-10) / (RefAmp + 1e-10)) < BASE_CONV) or (BaseEstCount > prParams.BaselineFKCount);


          until(CanQuit);


          NewLL := NewLL + SafeLog(Rho[p]);


        end;//of p



        //********************************************************************//
        //                                                                    //
        //                  B a c k w a r d   -   K a l m a n                 //
        //                                                                    //
        //********************************************************************//


        //Initialize last point
        p         := DataCount - 1;

        //Last BBase is equal to last Base
        BBaseX[p] := BaseX[p];
        BBaseP[p] := BaseP[p];

        y         := Data[p] * Scale + Off - BBaseX[p];

        //Betta of last point
        for s := 0 to prStateCount - 1 do begin
          Betta[s, p] := 1 / Rho[p];
        end;//of s

        for s := 0 to prStateCount - 1 do begin
          BGauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y);
          Omega[s, p]  := BGauss[s, p] * Betta[s, p];
        end;


        //Loop for every point .................................................
        for p := DataCount - 2 downto 0 do begin


          BBaseX[p]      := BBaseX[p+1];
          BBaseP[p]      := BBaseP[p+1];

          y              := Data[p] * Scale + Off;

          BaseEstCount   := 0;


          repeat


            RefAmp := BBaseX[p];


            //Betta t

            //for each state
            for s := 0 to prStateCount - 1 do begin

              Betta[s, p] := 0;

              for ss := 0 to prStateCount - 1 do begin

                Betta[s, p] := Betta[s, p] + AMatrix[s, ss] * Omega[ss, p + 1];

              end;//of ss

              //Scale Betta t
              Betta[s, p] := Betta[s, p] / Rho[p];


            end;//of s


            //Omega t

            //for each state
            for s := 0 to prStateCount - 1 do begin

              BGauss[s, p] := Gaussian(Amp[prClsIdxR[s]], Std[prClsIdxR[s]], y - RefAmp);

              Omega[s, p]  := BGauss[s, p] * Betta[s, p];

            end;//of s



            //Update baseline

            //ThisAmp and ThisStd are linear combinations of states
            ThisAmp := 0;
            MeasVar := 0;
            Sum     := 0;
            for s := 0 to prStateCount - 1 do begin
              ThisAmp := ThisAmp + Omega[s, p] *     Amp[prClsIdxR[s]];
              MeasVar := MeasVar + Omega[s, p] * Sqr(Std[prClsIdxR[s]]);
              Sum     := Sum     + Omega[s, p];
            end;
            ThisAmp := ThisAmp / Sum;
            MeasVar := MeasVar / Sum;

            KalmanForward(BBaseX[p], BBaseP[p], BBaseX[p+1], BBaseP[p+1], MeasVar, BaseVar, (y - ThisAmp){BaseData});

            Inc(BaseEstCount);

            CanQuit := (Abs((RefAmp - BBaseX[p] + 1e-10) / (RefAmp + 1e-10)) < BASE_CONV) or (BaseEstCount > prParams.BaselineFKCount);


          until(CanQuit);



        end;//of p



        //Smooth both baselines
        DoBaselineBack;
        DoBaselineBBack;


        //Smooth baseline
        for p := 0 to DataCount - 1 do begin

          if SBaseP[p] < BSBaseP[p] then begin
            SBaseX[p] := SBaseX[p];
            SBaseP[p] := SBaseP[p];
          end else begin
            SBaseX[p] := BSBaseX[p];
            SBaseP[p] := BSBaseP[p];
          end;

        end;



        //Now alternate between Markov forward-backward and Kalman forward-backward
        for BaseEstCount := 1 to prParams.BaselineFBKCount do begin


          //****************************************************************//
          //                                                                //
          //               F o r w a r d   -   B a c k w a r d              //
          //                                                                //
          //                        (given Baseline)                        //
          //                                                                //
          //****************************************************************//


          //Forward **********************************************************


          NewLL := DoForward;


          //Backward *********************************************************


          //Betta
          DoBetta;

          //Gamma
          DoGamma;

          //Initial Probabilities
          for s := 0 to prStateCount - 1 do begin
            Prob0[s] := Gamma[s, 0];
          end;



          //****************************************************************//
          //                                                                //
          //                           K a l m a n                          //
          //                                                                //
          //                          (given Markov)                        //
          //                                                                //
          //****************************************************************//


          //Forward **********************************************************

          //First point
          p := 0;
          y  := Data[p] * Scale + Off;
          nm := 0;
          for s := 0 to prStateCount - 1 do begin
            y  := y  - Gamma[s, p] * Amp[prClsIdxR[s]];
            nm := nm + Gamma[s, p] * Sqr(Std[prClsIdxR[s]]);
          end;
          KalmanForward(BaseX[p], BaseP[p], SBaseX[p], SBaseP[p], nm, BaseVar, y);

          //Every point
          for p := 1 to DataCount - 1 do begin
            y  := Data[p] * Scale + Off;
            nm := 0;
            for s := 0 to prStateCount - 1 do begin
              y  := y  - Gamma[s, p] * Amp[prClsIdxR[s]];
              nm := nm + Gamma[s, p] * Sqr(Std[prClsIdxR[s]]);
            end;
            KalmanForward(BaseX[p], BaseP[p], BaseX[p-1], BaseP[p-1], nm, BaseVar, y);
          end;

          //Backward *********************************************************
          p := DataCount - 1;
          SBaseX[p] := BaseX[p];
          SBaseP[p] := BaseP[p];
          for p := DataCount - 1 downto 1 do begin
            cJ          := BaseP[p-1] / (BaseP[p-1] + BaseVar);
            SBaseX[p-1] := BaseX[p-1] + cJ *      (SBaseX[p] - BaseX[p-1]);
            SBaseP[p-1] := BaseP[p-1] + cJ * (SBaseP[p] - (BaseP[p-1] + BaseVar)) * cJ;
          end;



          //Add noise to the baseline %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
          //for p := 0 to DataCount - 1 do begin
          //  SBaseX[p] := RandG(SBaseX[p], Sqrt(SBaseP[p] / BaseEstCount));
          //end;
          //********************************************************************



          //Try re-estimation inside FBK ...
          if prParams.Reestimation <> IDLBASE_REEST_NONE then begin
            //Epsilon
            if not prParams.FixKinetics then begin
              DoEpsilon;
            end;//FixA ?
            //Amplitudes and Noise
            ReestAmpsStds;
            //Eliminate offset in Amp and Baseline
            SBaseX[0] := SBaseX[0] + Amp[prClsIdx[0]];
            BaseX[0]  := SBaseX[0];
            BaseP[0]  := SBaseP[0];
            Sum       := Amp[prClsIdx[0]];
            for c := 0 to prClsCount - 1 do begin
              Amp[c] := Amp[c] - Sum;
            end;
          end;


          //********************************************************************//
          //                                                                    //
          //                              M O V I E S                           //
          //                                                                    //
          //********************************************************************//
          if prMovie.Make then begin


            MovieData       := MovieTree['Alpha'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(Alpha[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;
            MovieData       := MovieTree['Betta'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(Betta[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;
            MovieData       := MovieTree['Gamma'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, prStateCount, DataCount );
            for s := 0 to prStateCount - 1 do begin
              Move(Gamma[s][0], TPADouble(MovieData.DataPtr)[s * DataCount], DataCount * SizeOf(Double));
            end;

            MovieData       := MovieTree['BaseX'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(BaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['BaseP'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(BaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['BBaseX'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(BBaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['BBaseP'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(BBaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['SBaseX'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(SBaseX^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['SBaseP'];
            MovieData.SetupNumData( NODE_DATATYPE_DOUBLE, DataCount, 1 );
            Move(SBaseP^, MovieData.DataPtr^, SizeOf(Double) * DataCount);
            MovieData       := MovieTree['Amps'];
            for s := 0 to prStateCount - 1 do begin
              MovieData.Data[s] := Amp[prClsIdxR[s]];
            end;

            prMovieIteration  := Iteration;
            prMovieFBKIter    := BaseEstCount;
            Synchronize(prMakeMovie_ForwardBackwardKalman);


          end;//MovieMake


        end;//of BaseEstCount



        //Epsilon
        if not prParams.FixKinetics then begin

          DoEpsilon;

        end;//FixA ?


        if prParams.Reestimation = IDLBASE_REEST_NONE then begin
          Break;
        end;



        //Re-estimation ********************************************************



        //Amplitudes and Noise
        ReestAmpsStds;



        //Movies
        if prMovie.Make then begin
          MovieData       := MovieTree['Amps'];
          for s := 0 to prStateCount - 1 do begin
            MovieData.Data[s] := Amp[prClsIdxR[s]];
          end;
        end;


        //Eliminate offset in Amp and Baseline
        SBaseX[0] := SBaseX[0] + Amp[prClsIdx[0]];
        BaseX[0]  := SBaseX[0];
        BaseP[0]  := SBaseP[0];
        Sum       := Amp[prClsIdx[0]];
        for c := 0 to prClsCount - 1 do begin
          Amp[c] := Amp[c] - Sum;
        end;


        //Initial Probabilities
        for s := 0 to prStateCount - 1 do begin
          Prob0[s] := Gamma[s, 0];
        end;


        //A Matrix
        if not prParams.FixKinetics then begin
          ReestAMatrix;
        end;


        //Recalculate LL with reestimated parameters
        NewLL := DoForward;


        //Print Q matrix as the logarithm of the A matrix ......................
        MatrixLogarithm(AMatrix, AMatrixN, IDLBaseChunk.Sampling, prStateCount, 100, 1e-4);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            if AMatrixN[s, ss] < 0 then begin
              AMatrixN[s, ss] := 0;
              AMatrixN[ss, s] := 0;
            end;
          end;
        end;
        prConsoleMsg := #13#10'Q Matrix:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            if (s = ss) or (prQ[s, ss] = 0)then begin
              prConsoleMsg := prConsoleMsg + #9'...           ';
            end else begin
              prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrixN[s, ss]]);
            end;
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);



        //Print current iteration values .......................................
        prConsoleMsg := Format(#13#10'Iteration %d'#13#10, [Iteration]);
        prConsoleMsg := prConsoleMsg + Format('LL: %14.10g'#13#10, [NewLL]);
        prConsoleMsg := prConsoleMsg + 'A Matrix:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g', [AMatrix[s, ss]]);
          end;
          prConsoleMsg := prConsoleMsg + #13#10;
        end;
        prConsoleMsg := prConsoleMsg + 'State, Prob0:'#13#10;
        for s := 0 to prStateCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#13#10, [s, Prob0[s]]);
        end;
        prConsoleMsg := prConsoleMsg + 'Cls, Amp, Std:'#13#10;
        for c := 0 to prClsCount - 1 do begin
          prConsoleMsg := prConsoleMsg + Format(#9'%d'#9'%-14.6g'#9'%-14.6g'#13#10, [prClsIdx[c], Amp[c], Std[c]]);
        end;
        if prParams.BaselineTrack then begin
          prConsoleMsg := prConsoleMsg + 'Baseline Var, Std, X0, P0:'#13#10;
          prConsoleMsg := prConsoleMsg + Format(#9'%-14.6g'#13#10#9'%-14.6g'#13#10#9'%-14.6g'#13#10#9'%-14.6g'#13#10, [BaseVar, Sqrt(Abs(BaseVar)), SBaseX[0], SBaseP[0]]);
        end;
        prConsoleCode := MESSAGE_MESSAGE;
        if not prParams.QuietOutput then Synchronize(prWriteNewConsole);


        if prStopped then Break;


        //Check convergence
        if Abs(NewLL - OldLL) <= prParams.LLConv then Break;


        OldLL := NewLL;


      end;//of Iteration


      //Results ................................................................
      //if not prStopped then begin

        //Make the state vector
        State := IAlloc1(DataCount);
        for p := 0 to DataCount - 1 do begin
          //{
          State[p] := 0;
          Sum := Gamma[0, p];//it's not Sum but Max
          for s := 1 to prStateCount - 1 do begin
            if Gamma[s, p] > Sum then begin
              State[p] := s;
              Sum := Gamma[s, p];
            end;
          end;
          //}
          {
          for c := 0 to prClsCount - 1 do begin
            AmpN[c] := 0;
          end;
          for s := 0 to prStateCount - 1 do begin
            AmpN[prClsIdxR[s]] := AmpN[prClsIdxR[s]] + Gamma[s, p];
          end;
          State[p] := 0;
          Sum      := AmpN[0];
          for c := 1 to prClsCount - 1 do begin
            if AmpN[c] > Sum then begin
              State[p] := c;
              Sum      := AmpN[c];
            end;
          end;
          for s := 0 to prStateCount - 1 do begin
            if prClsOfState[s] = State[p] then begin
              State[p] := s;
              Break;
            end;
          end;
          }
        end;


        //Make Dwells
        prMakeDwells(IdlBaseChunk, State, DataCount);

        IdlBaseChunk.ClassCount := prClsCount;
        IdlBaseChunk.StateCount := prStateCount;
        IdlBaseChunk.Amps       := SAlloc1(prClsCount);
        IdlBaseChunk.Stds       := SAlloc1(prClsCount);
        IdlBaseChunk.AMatrix    := DAlloc2(prStateCount, prStateCount);
        for s := 0 to prStateCount - 1 do begin
          for ss := 0 to prStateCount - 1 do begin
            IdlBaseChunk.AMatrix[s, ss] := AMatrix[s, ss];
          end;
        end;
        for c := 0 to prClsCount - 1 do begin
          IdlBaseChunk.Amps[c]  := Amp[c];
          IdlBaseChunk.Stds[c]  := Std[c];
        end;

        IdlBaseChunk.LL         := NewLL;
        IdlBaseChunk.Iterations := Iteration - 1;


        //Make Baseline Nodes
        if (prParams.BaselineTrack) and (prParams.BaselineCorrect) then begin

          prBase2Nodes(IdlBaseChunk, SBaseX);

        end;


        //Make Stats and Histograms
        prMakeStats(IdlBaseChunk, SBaseX); // SBaseX is nil if not prParams.BaselineTrack


      //end;//Results



    except


      Result            := True;

      Chunk.Status      := CHUNK_STATUS_ERROR;
      Chunk.Completion  := 0;


    end;


  finally


    //Free Memory ..............................................................
    Free2(TPMChar(Alpha));
    Free2(TPMChar(Betta));
    Free2(TPMChar(Gamma));
    Free2(TPMChar(Omega));
    Free2(TPMChar(Gauss));
    Free2(TPMChar(BGauss));
    FreeMem(Rho);
    if Epsilon <> nil then begin
      for p := 0 to DataCount - 1 do begin
        Free2(TPMChar(Epsilon[p]));
      end;
    end;
    FreeMem(Epsilon);

    FreeMem(State);

    FreeMem(BaseX);
    FreeMem(BaseP);
    FreeMem(BBaseX);
    FreeMem(BBaseP);
    FreeMem(SBaseX);
    FreeMem(SBaseP);
    FreeMem(BSBaseX);
    FreeMem(BSBaseP);

    Free2(TPMChar(AMatrix));
    Free2(TPMChar(AMatrixN));

    FreeMem(AM);

    FreeMem(Prob0);

    FreeMem(Amp);
    FreeMem(AmpN);
    FreeMem(Std);


    //Signal new status
    if Chunk.Status <> CHUNK_STATUS_ERROR then begin
      Chunk.Status := CHUNK_STATUS_DONE;
      Chunk.Completion := 100;
    end;


  end;

end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlBase.prDoExecute;
begin
  prConsoleMsg  := 'IdlBase running ... ';
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
  inherited;
end;

procedure TQUB_Task_IdlBase.prDoExecuteStart;
begin

  inherited;

  prDataSet := TNode.Create('DataSet');

  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;

    prDataSet['sampling'].SetData(TQUB_QVO_QVD(Data).Sampling * 1.0e3);
  end;

  prResultNode         := TNode.Create('IdlBase Result');
  prResultNode['ResultType'].SetData( 'IdlBase' );
  if Data <> nil then prResultNode['DataSet']['FileName'].SetData(TQUB_QVO_QVD(Data).FileName);

end;


procedure TQUB_Task_IdlBase.prDoExecuteEnd;
begin
  if Model <> nil then Model.UnlockAny;

  if prSelectionList <> nil then prSelectionList.UnLock;

  inherited;
end;


procedure TQUB_Task_IdlBase.prSetChunkParameters;
begin
  inherited;
  prChunkSize := 0;
  prChunkSegmentation := True;
end;

procedure TQUB_Task_IdlBase.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  IdlBaseChunk :TQUB_IdlBase_PChunk;
begin

  IdlBaseChunk := TQUB_IdlBase_PChunk(Chunk);

  if IdlBaseChunk.Data <> nil then begin
    FreeMem(IdlBaseChunk.Data);
    IdlBaseChunk.Data := nil;
  end;

  FreeMem(IdlBaseChunk.Amps);
  IdlBaseChunk.Amps := nil;
  FreeMem(IdlBaseChunk.Stds);
  IdlBaseChunk.Stds := nil;
  Free2(TPMChar(IdlBaseChunk.AMatrix));
  IdlBaseChunk.AMatrix := nil;

  IdlBaseChunk.StatResult := nil;

  FreeMem(IdlBaseChunk);
                   
end;

function TQUB_Task_IdlBase.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
begin

  Result           := AllocMem(SizeOf(TQUB_IdlBase_Chunk));

  Result.Sel       := Sel;

  with TQUB_IdlBase_PChunk(Result)^ do begin
  //Set Chunk properties
  
    Sampling       := TQUB_QVO_QVD(Self.Data).Sampling;
    Scaling        := TQUB_QVO_QVD(Self.Data).Scaling * TQUB_QVO_QVD(Self.Data).DataChannelScaling[ChannelIndex];
    Offset         := TQUB_QVO_QVD(Self.Data).DataChannelOffset[ChannelIndex];

    SegCount       := 0;
    SegFirsts      := nil;
    SegLasts       := nil;
    SegClasses     := nil;

    NodeCount      := 0;
    Nodes          := nil;
    NodesDBL       := nil;

    LL             := 0;
    Iterations     := 0;

    StatResult     := nil;

  end;

end;

//******************************************************************************
procedure TQUB_Task_IdlBase.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then prModel.LockAny;
end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlBase.OnGetProperties(Sender :TObject);
var
  QVD                           :TQUB_QVO_QVD;
begin
  with Sender as TQUB_Task_IdlBase_PropertiesForm do begin

    Self.ChannelIndex           := Properties['ChannelIndex'].Data[0];

    //Data Source
    Self.DataSource             := DataSource;

    //Data Processed
    Self.DataProcessed          := StringToDataProcessed(Properties['ProcessData'].DataAsString);
    Self.ProcessFilter          := 0<>Properties['ProcessData']['Filter'].Data[0];
    Self.ProcessFilterFreq      := Properties['ProcessData']['FilterFreq'].Data[0];

    //Client/Server
    Self.ThreadCount            := Properties['ThreadCount'].Data[0];
    Self.Remote                 := 0<>Properties['Remote'].Data[0];

    Self.Movie                  := Movie;
    {
    Self.MovieSaveBitmaps       := MovieSaveBitmaps;
    Self.MovieShowTitles        := MovieShowTitles;
    Self.MovieMaxFrameCount     := MovieMaxFrameCount;
    }
    
    //Params
    Self.Params                 := Params;

  end;//with

   // prevents corrected corrected data and accidentally making it twice as off
  if DataProcessed = DATA_PROCESSED_ASDISPLAYED then begin
    QVD := ViewportData.CurrData;
    if QVD <> nil then QVD.BaselineOn[ QVD.ActiveChannel ] := False;
  end;
end;

procedure TQUB_Task_IdlBase.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_IdlBase_PropertiesForm do begin
  //  ChannelIndex                := Self.ChannelIndex;
  //end;//with
end;


function TQUB_Task_IdlBase.DataCallBack(IdlBaseChunk :TQUB_IdlBase_PChunk) :Integer;
var
  BaseChunk :TQUB_Thread_PChunk;
begin
  Result := 0;

  try

    BaseChunk := TQUB_Thread_PChunk(IdlBaseChunk);

    if IdlBaseChunk.Data = nil then begin
    //Allocate

      IdlBaseChunk.Data := AllocMem(SizeOf(Integer) * (BaseChunk.Sel.Last - BaseChunk.Sel.First + 1));

      case DataProcessed of
        DATA_PROCESSED_NONE:
          TQUB_QVO_QVD(Data).LoadDataChannel       (IdlBaseChunk.Data, ChannelIndex, BaseChunk.Sel, False{Process}, False, False);
        DATA_PROCESSED_ASDISPLAYED:
          TQUB_QVO_QVD(Data).LoadDataChannelProcess(IdlBaseChunk.Data, ChannelIndex, BaseChunk.Sel, TQUB_QVO_QVD(Data).FilterOn[ChannelIndex], TQUB_QVO_QVD(Data).FilterFreq[ChannelIndex], TQUB_QVO_QVD(Data).BaselineOn[ChannelIndex]);
        DATA_PROCESSED_ASSUCH:
          TQUB_QVO_QVD(Data).LoadDataChannelProcess(IdlBaseChunk.Data, ChannelIndex, BaseChunk.Sel, ProcessFilter,               ProcessFilterFreq,             ProcessBaseline);
      end;

    end else begin
    //DeAllocate

      FreeMem(IdlBaseChunk.Data);
      IdlBaseChunk.Data := nil;

    end;

  except

    Result := 1;

  end;

end;


procedure TQUB_Task_IdlBase.prFinishResults;
var
  c, p                     :Integer;
  IdlBaseChunk             :TQUB_IdlBase_PChunk;
  Chunk                    :TQUB_Thread_PChunk;
  DataSet                  :INode;
begin

  //To do: Chris, please check this procedure.

  DataSet                  := prResultNode['DataSet'];
  DataSet['FileName'].SetData(TQUB_QVO_QVD(Data).FileName);
  DataSet['sampling'].SetData(TQUB_QVO_QVD(Data).Sampling * 1.0e3);

  prTotLL := 0.0;

  for c := 0 to prChunks.Count - 1 do begin

    IdlBaseChunk := prChunks[c];
    Chunk        := prChunks[c];

    if IdlBaseChunk.StatResult <> nil then begin
      DataSet.AppendClone(IdlBaseChunk.StatResult, True);
    end;

    if Chunk.Status = CHUNK_STATUS_DONE then begin

      if IdlBaseChunk.SegCount > 0 then begin

        // make dwell bounds global in data file
        for p := 0 to IdlBaseChunk.SegCount - 1 do begin
          Inc(IdlBaseChunk.SegFirsts[p], Chunk.Sel.First);
          Inc(IdlBaseChunk.SegLasts [p], Chunk.Sel.First);
        end;

        prTotLL := prTotLL + IdlBaseChunk.LL;

      end;//SegCount > 0

    end;//Done

  end;//c

  prResultNode['LL'].SetData(prTotLL);

end;


//****************************************************************************//
//                                                                            //
//                                  M O V I E S                               //
//                                                                            //
//                         V I T E R B I  -  K A L M A N                      //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlBase.prMakeMovie_ViterbiKalman;
const
  MovieName = 'Movie_VK_';
var
  MovieMargin :TRect;
  WaitEvent   :TEvent;
  dPoint      :Integer;
  DCount      :Integer;
  p           :Integer;

begin

  WaitEvent := nil;


  try


    WaitEvent := TEvent.Create(nil, True, False, '');

    TQUB_QVO_QVD(Data).Zoom.Bitmap.PixelFormat := pf24Bit;
    //TQUB_QVO_QVD(Data).Zoom.Bitmap.Width       := 350;
    //TQUB_QVO_QVD(Data).Zoom.Bitmap.Height      := 300;

    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawChannel    := prMakeMovie_ViterbiKalman_BeforeDrawChannel;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawChannel     := prMakeMovie_ViterbiKalman_AfterDrawChannel;
    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawIdlChannel := prMakeMovie_ViterbiKalman_BeforeDrawIdlChannel;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawIdlChannel  := prMakeMovie_ViterbiKalman_AfterDrawIdlChannel;


    MovieMargin            := TQUB_QVO_QVD(Data).Zoom.Margin;
    TQUB_QVO_QVD(Data).Zoom.MarginLeft   := 5;
    if prMovie.ShowTitles then begin
      TQUB_QVO_QVD(Data).Zoom.MarginTop  := 60;
    end else begin
      TQUB_QVO_QVD(Data).Zoom.MarginTop  := 5;
    end;
    TQUB_QVO_QVD(Data).Zoom.MarginRight  := 5;
    TQUB_QVO_QVD(Data).Zoom.MarginBottom := 5;


    DCount    := MovieTree['DataCount'].Data[0];
    if Movie.MaxFrameCount = 0 then begin
      dPoint := 1;
    end else begin
      dPoint := DCount div Movie.MaxFrameCount;
      if dPoint < 1 then dPoint := 1;
    end;


    //First frame --> Just data
    if prMovieIteration = 1 then begin
      prMovieState := MOVIE_JUSTDATA;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
    //WaitEvent.WaitFor(1000);
    end;

    //Forward
    prMovieState := MOVIE_FORWARD;
    p := 0;
    if Movie.MaxFrameCount = 0 then p := DCount - 1;
    repeat
      Application.ProcessMessages;
      if prStopped then Exit;
      prMoviePointIndex := p;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
      //WaitEvent.Wait         For(5);
      if (p = DCount - 1) then Break;
      Inc(p, dPoint);
      if p >= DCount then p := DCount - 1;
    until(False);

    //WaitEvent.WaitFor(1000);


    if prParams.BaselineTrack then begin

      //Backward
      prMovieState := MOVIE_BACKWARD;
      p := DCount - 1;
      if prMovie.MaxFrameCount = 0 then p := 0;
      repeat
        Application.ProcessMessages;
        if prStopped then Exit;
        prMoviePointIndex := p;
        prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
        //WaitEvent.WaitFor(5);
        if p = 0 then Break;
        Dec(p, dPoint);
        if p < 0 then p := 0;
      until(False);

    end;

    //WaitEvent.WaitFor(1000);


  finally


    WaitEvent.Free;

    TQUB_QVO_QVD(Data).Zoom.Bitmap.PixelFormat     := pfDevice;

    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawChannel    := nil;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawChannel     := nil;
    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawIdlChannel := nil;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawIdlChannel  := nil;

    TQUB_QVO_QVD(Data).Zoom.MarginLeft             := MovieMargin.Left;
    TQUB_QVO_QVD(Data).Zoom.MarginTop              := MovieMargin.Top;
    TQUB_QVO_QVD(Data).Zoom.MarginRight            := MovieMargin.Right;
    TQUB_QVO_QVD(Data).Zoom.MarginBottom           := MovieMargin.Bottom;


  end;

end;


procedure TQUB_Task_IdlBase.prMakeMovie_ViterbiKalman_BeforeDrawChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  MovieVar       :INode;
  MovieVProb     :INode;
  MovieVBack     :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  pMax           :Double;
  d, p, s, v, w  :Integer;
  x, y           :Integer;
begin


  Exit;

  if not prParams.BaselineTrack then Exit;


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        Brush.Style     := bsClear;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];

              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(0, 1, x, w);


              //Forward Baseline
              Pen.Color := prMovie.ForBasePColor;

              MovieData       := MovieTree['BaseX'];
              MovieVar        := MovieTree['BaseP'];
              MovieVProb      := MovieTree['VProb'];
              MovieVBack      := MovieTree['VBack'];


              p    := prMoviePointIndex;
              v    := 0;
              pMax := MovieVProb.Data[p + v*DCount];
              for s := 1 to prStateCount - 1 do begin
                if MovieVProb.Data[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := MovieVProb.Data[p + s*DCount];
                end;
              end;
              d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p + v*DCount])) + 1;
              MoveTo(x, y);
              for p := prMoviePointIndex - 1 downto 0 do begin
                v := MovieVBack.Data[p+1 + v*DCount];
                d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p + v*DCount])) + 1;
                LineTo(x, y);
              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];

              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(0, 1, x, w);


              //Forward Baseline
              Pen.Color := prMovie.ForBasePColor;

              MovieData       := MovieTree['BaseX'];
              MovieVar        := MovieTree['BaseP'];
              MovieVProb      := MovieTree['VProb'];
              MovieVBack      := MovieTree['VBack'];


              p    := DCount - 1;
              v    := 0;
              pMax := MovieVProb.Data[p + v*DCount];
              for s := 1 to prStateCount - 1 do begin
                if MovieVProb.Data[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := MovieVProb.Data[p + s*DCount];
                end;
              end;
              d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p + v*DCount])) + 1;
              MoveTo(x, y);
              for p := DCount - 1 - 1 downto 0 do begin
                v := MovieVBack.Data[p+1 + v*DCount];
                d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p + v*DCount])) + 1;
                LineTo(x, y);
              end;


              //Backward Baseline
              Pen.Color := prMovie.BacBasePColor;


              MovieData       := MovieTree['SBaseX'];
              MovieVar        := MovieTree['SBaseP'];


              p := prMoviePointIndex;

              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
              MoveTo(x, y);
              for p := prMoviePointIndex + 1 to DCount - 1 do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                LineTo(x, y);
              end;


            end;


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);


        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_ViterbiKalman_AfterDrawChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  MovieVProb     :INode;
  MovieVBack     :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  TitleText      :String;
  ForText        :String;
  BackText       :String;
  pMax           :Double;
  l, t           :Integer;
  d, p, s, v     :Integer;
  x, y           :Integer;
begin


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        if prMovie.ShowTitles then begin

          SelectClipRgn(Handle, DrawRegion);

          Brush.Style     := bsClear;


          if (prParams.BaselineTrack) and (prStateCount = 1) then begin
            TitleText       := 'Kalman filter-smoother';
          end else if (prParams.BaselineTrack) and (prStateCount > 1) then begin
            TitleText       := 'Viterbi-Kalman';
          end else begin
            TitleText       := 'Viterbi';
          end;

          ForText         := 'forward -->';
          BackText        := '<-- backward';


          //Title
          SetTextAlign(Handle, TA_TOP + TA_LEFT);
          Font.Height     := 30;
          Font.Name       := 'Arial';
          Font.Style      := [fsBold];
          l               := DrawRect.Left + 10;
          t               := DrawRect.Top  + 5;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, TitleText);
          Font.Color := clWhite;
          TextOut(l, t, TitleText);

          //ForTitle
          SetTextAlign(Handle, TA_TOP + TA_RIGHT);
          Font.Height     := 24;
          Font.Style      := [];
          l               := DrawRect.Right - 10;
          t               := DrawRect.Top   + 5 + 6;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, ForText);
          Font.Color := prMovie.ForIDLColor;
          TextOut(l, t, ForText);

          //BackTitle
          SetTextAlign(Handle, TA_TOP + TA_RIGHT);
          l               := DrawRect.Right - 10;
          t               := DrawRect.Top   + 5 + 6 + 24;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, BackText);
          Font.Color := prMovie.BacIDLColor;
          TextOut(l, t, BackText);

          if (prMovieState <> MOVIE_JUSTDATA) or (prMovieIteration > 1) then begin
            SetTextAlign(Handle, TA_TOP + TA_LEFT);
            Font.Height     := 20;
            Font.Name       := 'Arial';
            Font.Style      := [];
            l               := DrawRect.Left + 10;
            t               := DrawRect.Top  + 5 + 4 + 30;
            Font.Color := clBlack;
            TextOut(l + 3, t + 3, 'iteration ' + IntToStr(prMovieIteration));
            Font.Color := clWhite;
            TextOut(l, t, 'iteration ' + IntToStr(prMovieIteration));
          end;

        end;

        
        if not prParams.BaselineTrack then Exit;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              SelectClipRgn(Handle, DataRegion);


              //Forward Baseline
              Pen.Color := prMovie.ForBaseXColor;
              Pen.Width := prMovie.BaseXPenWidth;


              MovieData       := MovieTree['BaseX'];
              MovieVProb      := MovieTree['VProb'];
              MovieVBack      := MovieTree['VBack'];


              p    := prMoviePointIndex;
              v    := 0;
              pMax := MovieVProb.Data[p + v*DCount];
              for s := 1 to prStateCount - 1 do begin
                if MovieVProb.Data[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := MovieVProb.Data[p + s*DCount];
                end;
              end;
              d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              MoveTo(x, y);
              for p := prMoviePointIndex - 1 downto 0 do begin
                v := MovieVBack.Data[p+1 + v*DCount];
                d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                LineTo(x, y);
              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              SelectClipRgn(Handle, DataRegion);


              //Forward Baseline
              Pen.Color := prMovie.ForBaseXColor;
              Pen.Width := prMovie.BaseXPenWidth;


              MovieData       := MovieTree['BaseX'];
              MovieVProb      := MovieTree['VProb'];
              MovieVBack      := MovieTree['VBack'];


              p    := DCount - 1;
              v    := 0;
              pMax := MovieVProb.Data[p + v*DCount];
              for s := 1 to prStateCount - 1 do begin
                if MovieVProb.Data[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := MovieVProb.Data[p + s*DCount];
                end;
              end;
              d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              MoveTo(x, y);
              for p := DCount - 1 - 1 downto 0 do begin
                v := MovieVBack.Data[p+1 + v*DCount];
                d := Round((MovieData.Data[p + v*DCount] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                LineTo(x, y);
              end;


              //Backward Baseline
              Pen.Color := prMovie.BacBaseXColor;
              Pen.Width := prMovie.BaseXPenWidth;


              MovieData := MovieTree['SBaseX'];
              p := prMoviePointIndex;
              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              MoveTo(x, y);
              for p := prMoviePointIndex + 1 to DCount - 1 do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                LineTo(x, y);
              end;


            end;


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);

        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_ViterbiKalman_BeforeDrawIdlChannel(Sender :TObject; Channel :Integer);
var
  MovieVProb     :INode;
  MovieVBack     :INode;
  MovieAmps      :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  pMax           :Double;
  d, p, s, v, vn :Integer;
  x, y           :Integer;
begin


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        Brush.Style     := bsClear;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              if prStateCount = 1 then Exit;


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              //Forward Idealization

              Pen.Color := prMovie.ForIDLColor;
              Pen.Width := prMovie.IDLPenWidth;


              MovieAmps       := MovieTree['Amps'];
              MovieVProb      := MovieTree['VProb'];
              MovieVBack      := MovieTree['VBack'];


              p               := prMoviePointIndex;


              //Find Max(VProb)
              v    := 0;
              pMax := MovieVProb.Data[p + v*DCount];
              for s := 1 to prStateCount - 1 do begin
                if MovieVProb.Data[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := MovieVProb.Data[p + s*DCount];
                end;
              end;


              d := Round((MovieAmps.Data[v] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
              MoveTo(x, y);


              for p := prMoviePointIndex - 1 downto 0 do begin
                vn := MovieVBack.Data[p+1 + v*DCount];
                if vn <> v then begin
                  d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                  d := Round((MovieAmps.Data[vn] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;
                v := vn;
                d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                LineTo(x, y);
              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              if prStateCount = 1 then Exit;


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              //Forward Idealization

              Pen.Color := prMovie.ForIDLColor;
              Pen.Width := prMovie.IDLPenWidth;

              MovieAmps       := MovieTree['Amps'];
              MovieVProb      := MovieTree['VProb'];
              MovieVBack      := MovieTree['VBack'];


              p               := DCount - 1;


              //Find Max(VProb)
              v    := 0;
              pMax := MovieVProb.Data[p + v*DCount];
              for s := 1 to prStateCount - 1 do begin
                if MovieVProb.Data[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := MovieVProb.Data[p + s*DCount];
                end;
              end;


              d := Round((MovieAmps.Data[v] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
              MoveTo(x, y);


              for p := DCount - 2 downto 0 do begin
                vn := MovieVBack.Data[p+1 + v*DCount];
                if vn <> v then begin
                  d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                  d := Round((MovieAmps.Data[vn] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;
                v := vn;
                d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                LineTo(x, y);
              end;

            end;

        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);


        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_ViterbiKalman_AfterDrawIdlChannel(Sender :TObject; Channel :Integer);
begin
end;


//****************************************************************************//
//                                                                            //
//                                  M O V I E S                               //
//                                                                            //
//               F O R W A R D  -  B A C K A R D  -  K A L M A N              //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlBase.prMakeMovie_ForwardBackwardKalman;
const
  MovieName   = 'Movie_FBK_';
var
  MovieMargin :TRect;
  WaitEvent   :TEvent;
  dPoint      :Integer;
  DCount      :Integer;
  p           :Integer;

begin

  WaitEvent := nil;


  try


    WaitEvent := TEvent.Create(nil, True, False, '');

    TQUB_QVO_QVD(Data).Zoom.Bitmap.PixelFormat := pf24Bit;
    //TQUB_QVO_QVD(Data).Zoom.Bitmap.Width       := 624;
    //TQUB_QVO_QVD(Data).Zoom.Bitmap.Height      := 336;

    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawChannel    := prMakeMovie_ForwardBackwardKalman_BeforeDrawChannel;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawChannel     := prMakeMovie_ForwardBackwardKalman_AfterDrawChannel;
    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawIdlChannel := prMakeMovie_ForwardBackwardKalman_BeforeDrawIdlChannel;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawIdlChannel  := prMakeMovie_ForwardBackwardKalman_AfterDrawIdlChannel;


    MovieMargin            := TQUB_QVO_QVD(Data).Zoom.Margin;
    TQUB_QVO_QVD(Data).Zoom.MarginLeft   := 5;
    if prMovie.ShowTitles then begin
      TQUB_QVO_QVD(Data).Zoom.MarginTop  := 60;
    end else begin
      TQUB_QVO_QVD(Data).Zoom.MarginTop  := 5;
    end;
    TQUB_QVO_QVD(Data).Zoom.MarginRight  := 5;
    TQUB_QVO_QVD(Data).Zoom.MarginBottom := 5;


    DCount    := MovieTree['DataCount'].Data[0];
    if Movie.MaxFrameCount = 0 then begin
      dPoint := 1;
    end else begin
      dPoint := DCount div Movie.MaxFrameCount;
      if dPoint < 1 then dPoint := 1;
    end;


    //First frame --> Just data
    if (prMovieIteration = 1) and (prMovieFBKIter = 0) then begin
      prMovieState := MOVIE_JUSTDATA;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
    end;


    //Forward
    prMovieState := MOVIE_FORWARD;
    p := 0;
    if Movie.MaxFrameCount = 0 then p := DCount - 1;
    repeat
      Application.ProcessMessages;
      if prStopped then Exit;
      prMoviePointIndex := p;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
      if (p = DCount - 1) then Break;
      Inc(p, dPoint);
      if p >= DCount then p := DCount - 1;
    until(False);


    //Backward
    prMovieState := MOVIE_BACKWARD;
    p := DCount - 1;
    if Movie.MaxFrameCount = 0 then p := 0;
    repeat
      Application.ProcessMessages;
      if prStopped then Exit;
      prMoviePointIndex := p;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
      if p = 0 then Break;
      Dec(p, dPoint);
      if p < 0 then p := 0;
    until(False);


  finally


    WaitEvent.Free;

    TQUB_QVO_QVD(Data).Zoom.Bitmap.PixelFormat     := pfDevice;

    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawChannel    := nil;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawChannel     := nil;
    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawIdlChannel := nil;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawIdlChannel  := nil;

    TQUB_QVO_QVD(Data).Zoom.MarginLeft             := MovieMargin.Left;
    TQUB_QVO_QVD(Data).Zoom.MarginTop              := MovieMargin.Top;
    TQUB_QVO_QVD(Data).Zoom.MarginRight            := MovieMargin.Right;
    TQUB_QVO_QVD(Data).Zoom.MarginBottom           := MovieMargin.Bottom;


  end;


end;

procedure TQUB_Task_IdlBase.prMakeMovie_ForwardBackwardKalman_BeforeDrawChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  MovieVar       :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  d, p, w        :Integer;
  x, y           :Integer;
begin


  Exit;

  if not prParams.BaselineTrack then Exit;


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        Brush.Style     := bsClear;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];

              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(0, 1, x, w);



              //Forward Baseline ...............................................

              if prMovieFBKIter >= 0 then begin
              //Initialization or conditional on HMM


                Pen.Color := prMovie.ForBasePColor;


                MovieData       := MovieTree['BaseX'];
                MovieVar        := MovieTree['BaseP'];


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                MoveTo(x, y);
                for p := 1 to prMoviePointIndex do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                  LineTo(x, y);
                end;


              end else begin
              //Conditional on Baseline - draw it all


                Pen.Color := prMovie.BacBasePColor;


                MovieData       := MovieTree['SBaseX'];
                MovieVar        := MovieTree['SBaseP'];


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                  LineTo(x, y);
                end;


              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];

              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(0, 1, x, w);


              //Forward Baseline ...............................................

              if prMovieFBKIter >=0 then begin
              //Initialization or conditional on HMM


                Pen.Color := prMovie.ForBasePColor;


                MovieData       := MovieTree['BaseX'];
                MovieVar        := MovieTree['BaseP'];


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                  LineTo(x, y);
                end;


              end;


              //Backward Baseline ...............................................

              Pen.Color := prMovie.BacBasePColor;


              MovieData       := MovieTree['SBaseX'];
              MovieVar        := MovieTree['SBaseP'];


              if prMovieFBKIter >=0 then begin
              //Initialization or conditional on HMM


                p := prMoviePointIndex;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                MoveTo(x, y);
                for p := prMoviePointIndex + 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                  LineTo(x, y);
                end;


              end else begin
              //Conditional on Baseline


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                  LineTo(x, y);
                end;


              end;


            end;


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);


        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_ForwardBackwardKalman_AfterDrawChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  TitleText      :String;
  ForText        :String;
  BackText       :String;
  Text           :String;
  l, t           :Integer;
  d, p           :Integer;
  x, y           :Integer;
begin


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        if prMovie.ShowTitles then begin

          SelectClipRgn(Handle, DrawRegion);

          Brush.Style     := bsClear;


          if (prParams.BaselineTrack) and (prStateCount = 1) then begin
            TitleText       := 'Kalman filter-smoother';
          end else if (prParams.BaselineTrack) and (prStateCount > 1) then begin
            TitleText       := 'Forward-Backward-Kalman';
          end else begin
            TitleText       := 'Forward-Backward';
          end;

          ForText         := 'forward -->';
          BackText        := '<-- backward';


          //Title
          SetTextAlign(Handle, TA_TOP + TA_LEFT);
          Font.Height     := 30;
          Font.Name       := 'Arial';
          Font.Style      := [fsBold];
          l               := DrawRect.Left + 10;
          t               := DrawRect.Top  + 5;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, TitleText);
          Font.Color := clWhite;
          TextOut(l, t, TitleText);

          //ForTitle
          SetTextAlign(Handle, TA_TOP + TA_RIGHT);
          Font.Height     := 24;
          Font.Style      := [];
          l               := DrawRect.Right - 10;
          t               := DrawRect.Top   + 5 + 6;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, ForText);
          Font.Color := prMovie.ForIDLColor;
          TextOut(l, t, ForText);

          //BackTitle
          SetTextAlign(Handle, TA_TOP + TA_RIGHT);
          l               := DrawRect.Right - 10;
          t               := DrawRect.Top   + 5 + 6 + 24;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, BackText);
          Font.Color := prMovie.BacIDLColor;
          TextOut(l, t, BackText);


          if (prMovieState <> MOVIE_JUSTDATA) or (prMovieIteration > 1) then begin
            SetTextAlign(Handle, TA_TOP + TA_LEFT);
            Font.Height     := 20;
            Font.Name       := 'Arial';
            Font.Style      := [];
            l               := DrawRect.Left + 10;
            t               := DrawRect.Top  + 5 + 4 + 30;
            Font.Color := clBlack;
            Text := 'iteration ' + IntToStr(prMovieIteration);
            if prMovieFBKIter = 0 then begin
              if prParams.BaselineTrack then begin
                Text := Text + ' (initialize)';
              end;
            end else if prMovieFBKIter < 0 then begin
            //Conditional on baseline
              Text := Text + ' (HMM|LGM, iter ' + IntToStr(Abs(prMovieFBKIter)) + ')';
            end else begin
            //Conditional on HMM
              Text := Text + ' (LGM|HMM, iter ' + IntToStr(Abs(prMovieFBKIter)) + ')';
            end;

            TextOut(l + 3, t + 3, Text);
            Font.Color := clWhite;
            TextOut(l, t, Text);
          end;

        end;
        

        if not prParams.BaselineTrack then Exit;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              SelectClipRgn(Handle, DataRegion);


              //Forward Baseline ...............................................

              if prMovieFBKIter >= 0 then begin
              //Initialization or conditional on HMM


                Pen.Color := prMovie.ForBaseXColor;
                Pen.Width := prMovie.BaseXPenWidth;


                MovieData       := MovieTree['BaseX'];


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                MoveTo(x, y);
                for p := 1 to prMoviePointIndex do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  LineTo(x, y);
                end;


              end else begin
              //Conditional on Baseline - draw it all


                Pen.Color := prMovie.BacBaseXColor;
                Pen.Width := prMovie.BaseXPenWidth;


                MovieData       := MovieTree['SBaseX'];


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  LineTo(x, y);
                end;


              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              SelectClipRgn(Handle, DataRegion);


              //Forward Baseline ...............................................

              if prMovieFBKIter >=0 then begin
              //Initialization or conditional on HMM


                Pen.Color := prMovie.ForBaseXColor;
                Pen.Width := prMovie.BaseXPenWidth;


                MovieData       := MovieTree['BaseX'];


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  LineTo(x, y);
                end;


              end;


              //Backward Baseline ...............................................

              Pen.Color := prMovie.BacBaseXColor;
              Pen.Width := prMovie.BaseXPenWidth;


              MovieData       := MovieTree['SBaseX'];


              if prMovieFBKIter >=0 then begin
              //Initialization or conditional on HMM


                p := prMoviePointIndex;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                MoveTo(x, y);
                for p := prMoviePointIndex + 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  LineTo(x, y);
                end;


              end else begin
              //Conditional on Baseline


                p := 0;
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  d := Round((MovieData.Data[p] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                  LineTo(x, y);
                end;


              end;


            end;


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);

        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_ForwardBackwardKalman_BeforeDrawIdlChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  MovieAmps      :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  pMax           :Double;
  d, p, s, v, vn :Integer;
  x, y           :Integer;
begin


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        Brush.Style     := bsClear;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              //Forward Idealization

              if prMovieFBKIter <= 0 then begin
              //Initialization or conditional on Baseline


                Pen.Color := prMovie.ForIDLColor;
                Pen.Width := prMovie.IDLPenWidth;


                MovieAmps       := MovieTree['Amps'];
                MovieData       := MovieTree['Alpha'];


                p    := 0;
                v    := 0;
                pMax := MovieData.Data[p + v*DCount];
                for s := 1 to prStateCount - 1 do begin
                  if MovieData.Data[p + s*DCount] > pMax then begin
                    v    := s;
                    pMax := MovieData.Data[p + s*DCount];
                  end;
                end;
                d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                MoveTo(x, y);
                for p := 1 to prMoviePointIndex do begin
                  vn    := 0;
                  pMax := MovieData.Data[p + vn*DCount];
                  for s := 1 to prStateCount - 1 do begin
                    if MovieData.Data[p + s*DCount] > pMax then begin
                      vn    := s;
                      pMax := MovieData.Data[p + s*DCount];
                    end;
                  end;
                  if vn <> v then begin
                    d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                    d := Round((MovieAmps.Data[vn] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                  end;
                  v := vn;
                  d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;


              end else begin
              //Conditional on HMM


                Pen.Color := prMovie.BacIDLColor;
                Pen.Width := prMovie.IDLPenWidth;

                MovieAmps       := MovieTree['Amps'];
                MovieData       := {MovieTree['Betta'];//}MovieTree['Gamma'];


                p    := 0;
                v    := 0;
                pMax := MovieData.Data[p + v*DCount];
                for s := 1 to prStateCount - 1 do begin
                  if MovieData.Data[p + s*DCount] > pMax then begin
                    v    := s;
                    pMax := MovieData.Data[p + s*DCount];
                  end;
                end;
                d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  vn    := 0;
                  pMax := MovieData.Data[p + vn*DCount];
                  for s := 1 to prStateCount - 1 do begin
                    if MovieData.Data[p + s*DCount] > pMax then begin
                      vn    := s;
                      pMax := MovieData.Data[p + s*DCount];
                    end;
                  end;
                  if vn <> v then begin
                    d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                    d := Round((MovieAmps.Data[vn] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                  end;
                  v := vn;
                  d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;


              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              //Forward Idealization

              if prMovieFBKIter <= 0 then begin
              //Initialization or conditional on Baseline


                Pen.Color := prMovie.ForIDLColor;
                Pen.Width := prMovie.IDLPenWidth;

                MovieAmps       := MovieTree['Amps'];
                MovieData       := MovieTree['Alpha'];


                p    := 0;
                v    := 0;
                pMax := MovieData.Data[p + v*DCount];
                for s := 1 to prStateCount - 1 do begin
                  if MovieData.Data[p + s*DCount] > pMax then begin
                    v    := s;
                    pMax := MovieData.Data[p + s*DCount];
                  end;
                end;
                d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  vn    := 0;
                  pMax := MovieData.Data[p + vn*DCount];
                  for s := 1 to prStateCount - 1 do begin
                    if MovieData.Data[p + s*DCount] > pMax then begin
                      vn    := s;
                      pMax := MovieData.Data[p + s*DCount];
                    end;
                  end;
                  if vn <> v then begin
                    d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                    d := Round((MovieAmps.Data[vn] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                  end;
                  v := vn;
                  d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;


              end;


              //Backward Idealization


              Pen.Color := prMovie.BacIDLColor;
              Pen.Width := prMovie.IDLPenWidth;


              MovieAmps       := MovieTree['Amps'];
              MovieData       := {MovieTree['Betta'];//}MovieTree['Gamma'];


              //if (prMovieFBKIter < 0) or ((prMovieFBKIter = 0) and (not prParams.BaselineTrack)) then begin
              //Conditional on Baseline
              if prMovieFBKIter <= 0 then begin
              //Initialization or conditional on Baseline

                p    := prMoviePointIndex;
                v    := 0;
                pMax := MovieData.Data[p + v*DCount];
                for s := 1 to prStateCount - 1 do begin
                  if MovieData.Data[p + s*DCount] > pMax then begin
                    v    := s;
                    pMax := MovieData.Data[p + s*DCount];
                  end;
                end;
                d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                MoveTo(x, y);
                for p := prMoviePointIndex + 1 to DCount - 1 do begin
                  vn    := 0;
                  pMax := MovieData.Data[p + vn*DCount];
                  for s := 1 to prStateCount - 1 do begin
                    if MovieData.Data[p + s*DCount] > pMax then begin
                      vn    := s;
                      pMax := MovieData.Data[p + s*DCount];
                    end;
                  end;
                  if vn <> v then begin
                    d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                    d := Round((MovieAmps.Data[vn] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                  end;
                  v := vn;
                  d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;


              end else if prMovieFBKIter > 0 then begin
              //Conditional on HMM


                p    := 0;
                v    := 0;
                pMax := MovieData.Data[p + v*DCount];
                for s := 1 to prStateCount - 1 do begin
                  if MovieData.Data[p + s*DCount] > pMax then begin
                    v    := s;
                    pMax := MovieData.Data[p + s*DCount];
                  end;
                end;
                d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                MoveTo(x, y);
                for p := 1 to DCount - 1 do begin
                  vn    := 0;
                  pMax := MovieData.Data[p + vn*DCount];
                  for s := 1 to prStateCount - 1 do begin
                    if MovieData.Data[p + s*DCount] > pMax then begin
                      vn    := s;
                      pMax := MovieData.Data[p + s*DCount];
                    end;
                  end;
                  if vn <> v then begin
                    d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                    d := Round((MovieAmps.Data[vn] - Offset) * Scaling);
                    TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                    LineTo(x, y);
                  end;
                  v := vn;
                  d := Round((MovieAmps.Data[v] - Offset) * Scaling);
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;


              end;//FBKIter


            end;//Backward


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);


        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_ForwardBackwardKalman_AfterDrawIdlChannel(Sender :TObject; Channel :Integer);
begin
end;



//****************************************************************************//
//                                                                            //
//                                  M O V I E S                               //
//                                                                            //
//                               S T A I R C A S E                            //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_IdlBase.prMakeMovie_Staircase;
const
  MovieName   = 'Movie_Staircase_';
var
  MovieMargin :TRect;
  WaitEvent   :TEvent;
  dPoint      :Integer;
  DCount      :Integer;
  p           :Integer;

begin

  WaitEvent := nil;


  try


    WaitEvent := TEvent.Create(nil, True, False, '');

    TQUB_QVO_QVD(Data).Zoom.Bitmap.PixelFormat := pf24Bit;
    //TQUB_QVO_QVD(Data).Zoom.Bitmap.Width       := 500;
    //TQUB_QVO_QVD(Data).Zoom.Bitmap.Height      := 500;

    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawChannel    := prMakeMovie_Staircase_BeforeDrawChannel;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawChannel     := prMakeMovie_Staircase_AfterDrawChannel;
    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawIdlChannel := prMakeMovie_Staircase_BeforeDrawIdlChannel;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawIdlChannel  := prMakeMovie_Staircase_AfterDrawIdlChannel;


    MovieMargin            := TQUB_QVO_QVD(Data).Zoom.Margin;
    TQUB_QVO_QVD(Data).Zoom.MarginLeft   := 5;
    if prMovie.ShowTitles then begin
      TQUB_QVO_QVD(Data).Zoom.MarginTop  := 60;
    end else begin
      TQUB_QVO_QVD(Data).Zoom.MarginTop  := 5;
    end;
    TQUB_QVO_QVD(Data).Zoom.MarginRight  := 5;
    TQUB_QVO_QVD(Data).Zoom.MarginBottom := 5;


    DCount    := MovieTree['DataCount'].Data[0];
    if Movie.MaxFrameCount = 0 then begin
      dPoint := 1;
    end else begin
      dPoint := DCount div Movie.MaxFrameCount;
      if dPoint < 1 then dPoint := 1;
    end;


    //First frame --> Just data
    if prMovieIteration = 1 then begin
      prMovieState := MOVIE_JUSTDATA;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
    end;


    //Forward
    prMovieState := MOVIE_FORWARD;
    p := 0;
    if Movie.MaxFrameCount = 0 then p := DCount - 1;
    repeat
      Application.ProcessMessages;
      if prStopped then Exit;
      prMoviePointIndex := p;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
      if (p = DCount - 1) then Break;
      Inc(p, dPoint);
      if p >= DCount then p := DCount - 1;
    until(False);


    //Backward
    prMovieState := MOVIE_BACKWARD;
    p := DCount - 1;
    if Movie.MaxFrameCount = 0 then p := 0;
    repeat
      Application.ProcessMessages;
      if prStopped then Exit;
      prMoviePointIndex := p;
      prMovie_DrawAndSave(MovieName, prMovieFrameIndex);
      if p = 0 then Break;
      Dec(p, dPoint);
      if p < 0 then p := 0;
    until(False);


  finally


    WaitEvent.Free;

    TQUB_QVO_QVD(Data).Zoom.Bitmap.PixelFormat     := pfDevice;

    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawChannel    := nil;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawChannel     := nil;
    TQUB_QVO_QVD(Data).Zoom.OnBeforeDrawIdlChannel := nil;
    TQUB_QVO_QVD(Data).Zoom.OnAfterDrawIdlChannel  := nil;

    TQUB_QVO_QVD(Data).Zoom.MarginLeft             := MovieMargin.Left;
    TQUB_QVO_QVD(Data).Zoom.MarginTop              := MovieMargin.Top;
    TQUB_QVO_QVD(Data).Zoom.MarginRight            := MovieMargin.Right;
    TQUB_QVO_QVD(Data).Zoom.MarginBottom           := MovieMargin.Bottom;


  end;


end;

procedure TQUB_Task_IdlBase.prMakeMovie_Staircase_BeforeDrawChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  MovieVar       :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  d, p, w        :Integer;
  x, y           :Integer;
begin


  Exit;

  if not prParams.BaselineTrack then Exit;


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        Brush.Style     := bsClear;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];

              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(0, 1, x, w);



              //Forward Baseline ...............................................
              Pen.Color := prMovie.ForBasePColor;


              MovieData       := MovieTree['BaseX'];
              MovieVar        := MovieTree['BaseP'];


              p := 0;
              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
              MoveTo(x, y);
              for p := 1 to prMoviePointIndex do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                LineTo(x, y);
              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];

              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(0, 1, x, w);


              //Forward Baseline ...............................................
              Pen.Color := prMovie.ForBasePColor;


              MovieData       := MovieTree['BaseX'];
              MovieVar        := MovieTree['BaseP'];


              p := 0;
              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
              MoveTo(x, y);
              for p := 1 to DCount - 1 do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                LineTo(x, y);
              end;


              //Backward Baseline ...............................................
              Pen.Color := prMovie.BacBasePColor;


              MovieData       := MovieTree['SBaseX'];
              MovieVar        := MovieTree['SBaseP'];


              p := prMoviePointIndex;
              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
              MoveTo(x, y);
              for p := prMoviePointIndex + 1 to DCount - 1 do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                Pen.Width := 2 * Round(w * Sqrt(MovieVar.Data[p])) + 1;
                LineTo(x, y);
              end;


            end;


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);


        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_Staircase_AfterDrawChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  TitleText      :String;
  ForText        :String;
  BackText       :String;
  Text           :String;
  l, t           :Integer;
  d, p           :Integer;
  x, y           :Integer;
begin


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        if Movie.ShowTitles then begin


          SelectClipRgn(Handle, DrawRegion);

          Brush.Style     := bsClear;


          TitleText       := 'Staircase algorithm';
          ForText         := 'forward -->';
          BackText        := '<-- backward';


          //Title
          SetTextAlign(Handle, TA_TOP + TA_LEFT);
          Font.Height     := 30;
          Font.Name       := 'Arial';
          Font.Style      := [fsBold];
          l               := DrawRect.Left + 10;
          t               := DrawRect.Top  + 5;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, TitleText);
          Font.Color := clWhite;
          TextOut(l, t, TitleText);

          //ForTitle
          SetTextAlign(Handle, TA_TOP + TA_RIGHT);
          Font.Height     := 24;
          Font.Style      := [];
          l               := DrawRect.Right - 10;
          t               := DrawRect.Top   + 5 + 6;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, ForText);
          Font.Color := prMovie.ForIDLColor;
          TextOut(l, t, ForText);

          //BackTitle
          SetTextAlign(Handle, TA_TOP + TA_RIGHT);
          l               := DrawRect.Right - 10;
          t               := DrawRect.Top   + 5 + 6 + 24;
          Font.Color := clBlack;
          TextOut(l + 3, t + 3, BackText);
          Font.Color := prMovie.BacIDLColor;
          TextOut(l, t, BackText);


          if (prMovieState <> MOVIE_JUSTDATA) or (prMovieIteration > 1) then begin
            SetTextAlign(Handle, TA_TOP + TA_LEFT);
            Font.Height     := 20;
            Font.Name       := 'Arial';
            Font.Style      := [];
            l               := DrawRect.Left + 10;
            t               := DrawRect.Top  + 5 + 4 + 30;
            Font.Color := clBlack;
            Text := 'iteration ' + IntToStr(prMovieIteration);
            TextOut(l + 3, t + 3, Text);
            Font.Color := clWhite;
            TextOut(l, t, Text);
          end;

        end;//ShowTitles


        //*****
        Exit;
        //*****


        if not prParams.BaselineTrack then Exit;


        DCount := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              SelectClipRgn(Handle, DataRegion);


              //Forward Baseline ...............................................
              Pen.Color := prMovie.ForBaseXColor;
              Pen.Width := prMovie.BaseXPenWidth;


              MovieData       := MovieTree['BaseX'];


              p := 0;
              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              MoveTo(x, y);
              for p := 1 to prMoviePointIndex do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                LineTo(x, y);
              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              SelectClipRgn(Handle, DataRegion);


              //Forward Baseline ...............................................
              Pen.Color := prMovie.ForBaseXColor;
              Pen.Width := prMovie.BaseXPenWidth;


              MovieData       := MovieTree['BaseX'];


              p := 0;
              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              MoveTo(x, y);
              for p := 1 to DCount - 1 do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                LineTo(x, y);
              end;


              //Backward Baseline ...............................................

              Pen.Color := prMovie.BacBaseXColor;
              Pen.Width := prMovie.BaseXPenWidth;


              MovieData       := MovieTree['SBaseX'];


              p := prMoviePointIndex;
              d := Round((MovieData.Data[p] - Offset) * Scaling);
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
              MoveTo(x, y);
              for p := prMoviePointIndex + 1 to DCount - 1 do begin
                d := Round((MovieData.Data[p] - Offset) * Scaling);
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p, d, x, y);
                LineTo(x, y);
              end;


            end;


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);

        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_Staircase_BeforeDrawIdlChannel(Sender :TObject; Channel :Integer);
var
  MovieData      :INode;
  MovieBase      :INode;
  MovieOffset    :INode;
  JumpAmp        :Double;
  BaseAmp        :Double;
  HalfStateCount :Integer;
  DCount         :Integer;
  Scaling        :Double;
  Offset         :Double;
  OldBrushStyle  :TBrushStyle;
  OldPenColor    :TColor;
  OldPenWidth    :Integer;
  pMax           :Double;
  d, p, s, v, vn :Integer;
  x, y           :Integer;
begin


  with TQUB_QVO_QVD(Data).Zoom.DrawMap do begin
    with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas do begin


      //Save
      OldBrushStyle         := Brush.Style;
      OldPenColor           := Pen.Color;
      OldPenWidth           := Pen.Width;


      try


        Brush.Style     := bsClear;


        DCount    := MovieTree['DataCount'].Data[0];


        case prMovieState of

          MOVIE_JUSTDATA:
            begin
            end;

          //Forward
          MOVIE_FORWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              //Forward Idealization
              Pen.Color := prMovie.ForIDLColor;
              Pen.Width := prMovie.IDLPenWidth;


              JumpAmp         := MovieTree['JumpAmp'].Data[0];
              BaseAmp         := MovieTree['BaseAmp'].Data[0];
              HalfStateCount  := MovieTree['HalfStateCount'].Data[0];
              MovieData       := MovieTree['Alpha'];
              MovieBase       := MovieTree['BaseX'];
              MovieOffset     := MovieTree['MovieOffset'];


              p    := 0;
              v    := 0;
              pMax := TPADouble(MovieData.DataPtr)[p + v*DCount];
              for s := 1 to HalfStateCount - 1 do begin
                if TPADouble(MovieData.DataPtr)[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := TPADouble(MovieData.DataPtr)[p + v*DCount];
                end;
              end;
              if prParams.BaselineTrack then begin
                d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
              end else begin
                d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
              end;
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
              MoveTo(x, y);
              for p := 1 to prMoviePointIndex do begin
                vn    := 0;
                pMax := TPADouble(MovieData.DataPtr)[p + vn*DCount];
                for s := 1 to HalfStateCount - 1 do begin
                  if TPADouble(MovieData.DataPtr)[p + s*DCount] > pMax then begin
                    vn    := s;
                    pMax := TPADouble(MovieData.DataPtr)[p + vn*DCount];
                  end;
                end;
                if (vn + TPAInteger(MovieOffset.DataPtr)[p]) <> (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) then begin
                  if prParams.BaselineTrack then begin
                    d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) + TPADouble(MovieBase.DataPtr)[p - 1]) * Scaling - Offset*Scaling);
                  end else begin
                    d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) + BaseAmp) * Scaling - Offset*Scaling);
                  end;
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                  if prParams.BaselineTrack then begin
                    d := Round((JumpAmp * (vn + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
                  end else begin
                    d := Round((JumpAmp * (vn + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
                  end;
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;
                v := vn;
                if prParams.BaselineTrack then begin
                  d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
                end else begin
                  d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
                end;
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                LineTo(x, y);
              end;


            end;


          //Backward
          MOVIE_BACKWARD:
            begin


              SelectClipRgn(Handle, DataRegion);


              Scaling := MovieTree['Scaling'].Data[0];
              Offset  := MovieTree['Offset'].Data[0];


              //Forward Idealization
              Pen.Color := prMovie.ForIDLColor;
              Pen.Width := prMovie.IDLPenWidth;


              JumpAmp         := MovieTree['JumpAmp'].Data[0];
              BaseAmp         := MovieTree['BaseAmp'].Data[0];
              HalfStateCount  := MovieTree['HalfStateCount'].Data[0];
              MovieData       := MovieTree['Alpha'];
              MovieBase       := MovieTree['BaseX'];
              MovieOffset     := MovieTree['MovieOffset'];


              p    := 0;
              v    := 0;
              pMax := TPADouble(MovieData.DataPtr)[p + v*DCount];
              for s := 1 to HalfStateCount - 1 do begin
                if TPADouble(MovieData.DataPtr)[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := TPADouble(MovieData.DataPtr)[p + v*DCount];
                end;
              end;
              if prParams.BaselineTrack then begin
                d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
              end else begin
                d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
              end;
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
              MoveTo(x, y);
              for p := 1 to DCount - 1 do begin
                vn    := 0;
                pMax := TPADouble(MovieData.DataPtr)[p + vn*DCount];
                for s := 1 to HalfStateCount - 1 do begin
                  if TPADouble(MovieData.DataPtr)[p + s*DCount] > pMax then begin
                    vn    := s;
                    pMax := TPADouble(MovieData.DataPtr)[p + vn*DCount];
                  end;
                end;
                if (vn + TPAInteger(MovieOffset.DataPtr)[p]) <> (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) then begin
                  if prParams.BaselineTrack then begin
                    d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) + TPADouble(MovieBase.DataPtr)[p - 1]) * Scaling - Offset*Scaling);
                  end else begin
                    d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) + BaseAmp) * Scaling - Offset*Scaling);
                  end;
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                  if prParams.BaselineTrack then begin
                    d := Round((JumpAmp * (vn + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
                  end else begin
                    d := Round((JumpAmp * (vn + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
                  end;
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;
                v := vn;
                if prParams.BaselineTrack then begin
                  d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
                end else begin
                  d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
                end;
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                LineTo(x, y);
              end;



              //Backward Idealization
              Pen.Color := prMovie.BacIDLColor;
              Pen.Width := prMovie.IDLPenWidth;


              JumpAmp         := MovieTree['JumpAmp'].Data[0];
              BaseAmp         := MovieTree['BaseAmp'].Data[0];
              HalfStateCount  := MovieTree['HalfStateCount'].Data[0];
              MovieData       := MovieTree['Gamma'];
              MovieBase       := MovieTree['BaseX'];//SBaseX
              MovieOffset     := MovieTree['MovieOffset'];


              p    := prMoviePointIndex;
              v    := 0;
              pMax := TPADouble(MovieData.DataPtr)[p + v*DCount];
              for s := 1 to HalfStateCount - 1 do begin
                if TPADouble(MovieData.DataPtr)[p + s*DCount] > pMax then begin
                  v    := s;
                  pMax := TPADouble(MovieData.DataPtr)[p + v*DCount];
                end;
              end;
              if prParams.BaselineTrack then begin
                d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
              end else begin
                d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
              end;
              TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
              MoveTo(x, y);
              for p := prMoviePointIndex + 1 to DCount - 1 do begin
                vn    := 0;
                pMax := TPADouble(MovieData.DataPtr)[p + vn*DCount];
                for s := 1 to HalfStateCount - 1 do begin
                  if TPADouble(MovieData.DataPtr)[p + s*DCount] > pMax then begin
                    vn    := s;
                    pMax := TPADouble(MovieData.DataPtr)[p + vn*DCount];
                  end;
                end;
                if (vn + TPAInteger(MovieOffset.DataPtr)[p]) <> (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) then begin
                  if prParams.BaselineTrack then begin
                    d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) + TPADouble(MovieBase.DataPtr)[p - 1]) * Scaling - Offset*Scaling);
                  end else begin
                    d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p - 1]) + BaseAmp) * Scaling - Offset*Scaling);
                  end;
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                  if prParams.BaselineTrack then begin
                    d := Round((JumpAmp * (vn + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
                  end else begin
                    d := Round((JumpAmp * (vn + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
                  end;
                  TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                  LineTo(x, y);
                end;
                v := vn;
                if prParams.BaselineTrack then begin
                  d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + TPADouble(MovieBase.DataPtr)[p]) * Scaling - Offset*Scaling);
                end else begin
                  d := Round((JumpAmp * (v + TPAInteger(MovieOffset.DataPtr)[p]) + BaseAmp) * Scaling - Offset*Scaling);
                end;
                TQUB_QVO_QVD(Data).Zoom.MapPhysToDrawCoord(p - 0.5, d, x, y);
                LineTo(x, y);
              end;


            end;//Backward


        end;//case prMovieState


      finally


        //Restore
        SelectClipRgn(Handle, DataRegion);


        Pen.Color   := OldPenColor;
        Pen.Width   := OldPenWidth;
        Brush.Style := OldBrushStyle;


      end;


    end;//with TQUB_QVO_QVD(Data).Zoom.BitmapCanvas
  end;//with TQUB_QVO_QVD(Data).Zoom.DrawMap


end;

procedure TQUB_Task_IdlBase.prMakeMovie_Staircase_AfterDrawIdlChannel(Sender :TObject; Channel :Integer);
begin
end;


function TQUB_Task_IdlBase.prMovie_BitmapFileName(const Name :String; Digits :Integer; Index :Integer) :String;
begin

  Result := IntToStr(Index);
  while Length(Result) < Digits do Result := '0'+ Result;
  Result := Name + Result + '.bmp';

end;

procedure TQUB_Task_IdlBase.prMovie_DrawAndSave(const MovieName :String; var FrameIndex :Integer);
begin
  TQUB_QVO_QVD(Data).Zoom.Draw;
  if prMovie.SaveBitmaps then TQUB_QVO_QVD(Data).Zoom.Bitmap.SaveToFile(prMovie_BitmapFileName(MovieName, 4, FrameIndex));
  Inc(FrameIndex);
end;




//******************************************************************************
initialization


MovieTree := nil;


finalization


MovieTree := nil;


//******************************************************************************
end.
