unit QUB_Task_Mac;

{
/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */
}

interface


uses


SysUtils, Classes, Windows, SyncObjs,

Alloc, AllocTypes, Math,
qubopt_iface,

QUB_Node,
QUB_Global,
QUB_Q_Global, QUB_Q_Model, QUB_Mac_Model, QUB_Mac_Global,
QUB_QVO_QVD, QUB_QVO_QVD_Global,
QUB_Thread,
QUB_Task, QUB_Task_Data,
QUB_QFS_Channels, QUB_QFS_QMF, QUB_QFS_QMF_QMatrix,
QUB_QVO_QVM,
QUB_QFS_QDF,
QUB_IdlCond,
QUB_Task_IdlStim,
QUB_Task_IdlStim_Properties,
QUB_ExpCond_Parameters,
QUB_Task_Mac_Properties,

QUB_Cuda;

//KPL_Parser;


type


//forward class declarations
TQUB_Task_Mac_ChunkThread  = class;
TQUB_Task_Mac              = class;


//class declarations
TQUB_Task_Mac_ChunkThread  = class(TThread)
protected


  prReady                  :Boolean;


public


  MacChunkIdx              :Integer;
  MacTask                  :TQUB_Task_Mac;
  MacResult                :Boolean;
  Affinity                 :Integer;


  constructor Create(Suspended :Boolean);


  procedure Execute; override;

  procedure LLGradOfChunk;


  property Ready :Boolean read prReady;


end;


TQUB_Task_Mac              = class(TQUB_Task_Data)
protected


  //Source Objects
  prModel                  :TQUB_QVO_QVM;
  prSelectionList          :TQUB_QVO_QVD_SelectionList;

  //Optimizer Parameters
  prOptParams              :TQUB_Mac_Opt_Params;

  //MacModels
  prMacModels              :TList;

  //ExpConds
  prExpCondParams          :TQUB_ExpCond_ParametersListList;

  //Previous LL
  prPLL                    :Double;

  //prScriptPenaltyLL        :Double;

  //MaxChannel in Data
  prMaxChannel             :Integer;

  //Best Grads
  prfParsGrads             :TPADouble;
  prfParsOld               :TPADouble;

  //Used for numeric grads of Penalty
  prPenaltyLL              :Double;
  prPenaltyGrads           :TPADouble;

  //OptSegments
  prOptSegsChunks          :TList;
  prOptSegsOutput          :String;
  prOptSegsOldMdl          :TQUB_QFS_QMF;

  prQuietOutput            :Boolean;
  prDoPublishResults       :Boolean;

  prFitPreCleared          :Boolean;

  prFitDataListName        :String;
  prExcDataListName        :String;
  prGroupList              :String;
  prBatchCount             :Integer;
  prBatchIdx               :Integer;
  prBatchLabel             :String;

  prDataSetCount           :Integer;

  prQuietEntry             :Integer;

  prIdlStimSignal          :TEvent;
  prIdlStimQVD             :TQUB_QVO_QVD;
  prIdlStimCh              :Integer;
  prIdlStimDS              :INode;

  prSkipSlowChange         :Boolean;
  prSkipSlowStd            :Double;

  prResults                :INode;

  prChunkThreadCount       :Integer;
  prChunkThreads           :TPAPointer;
  prSignal                 :TSimpleEvent;

  prStartWatch             :Int64;

  prCuda_Lib               :TQUB_Cuda_Lib;

  //Execute
  procedure prDoTask; override;

  procedure prDoExecute; override;
  procedure prDoExecuteStart; override;
  procedure prDoExecuteEnd; override;

  //Chunks
  procedure prSetChunkParameters; override;
  function prChunkData :Boolean; override;
  function prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk; override;
  procedure prFreeChunk(Chunk :TQUB_Thread_PChunk); override;

  procedure prGetIdlStim(QVD :TQUB_QVO_QVD; ParamTemplate :TQUB_ExpCond_Parameters; Segments :TQUB_PAFirstLast; SegCount :Integer; FirstChunk :Integer; ParamClasses :TQUB_ExpCond_ParametersList);
  procedure prEnsureIdlStim(QVD :TQUB_QVO_QVD; StimChannels :TList{of Integer}; SegCount :Integer; Segments :TQUB_PAFirstLast);
  procedure prSimpleIdlStim(QVD :TQUB_QVO_QVD; Ch :Integer);
  procedure prIdlStimCommit;
  procedure prStartIdlStim;
  procedure prIdlStimDone(Sender :TObject);

  procedure prSetModel(Model :TQUB_QVO_QVM);

  //Update visuals
  procedure prUpdateModel;
  procedure prResetModel;
  procedure prUpdateData;
  procedure prShowData;
  procedure prRefreshZoom;
  //procedure prUpdateLLPlot;

  procedure prDoOverlapData;

  function DataCallBack(MacChunk :TQUB_Mac_PChunk) :Integer;
  procedure prSetupResampling(MacChunk :TQUB_Mac_PChunk);

  procedure prFitPointsOfChunk(MacChunk :TQUB_Mac_PChunk);

  procedure prMac;
  function prLLGrad(CompGrad :Integer) :Integer;
  {
  procedure prCalculateScriptPenalty;
  procedure prRunScriptPenalty;
  procedure prRunScriptIteration;
  procedure prRunScriptSetPars;
  }
  procedure prPrintOptParams;
  procedure prPrintParams(Final :Boolean);
  procedure prPrintGrads;
  procedure prUpdateResults(Final :Boolean);
  procedure prPublishResults;

  function prGetOptPParams: TQUB_Mac_Opt_PParams;


public


  Properties                            :INode;


  constructor Create(Suspended :Boolean); override;
  destructor Destroy; override;


  function DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
  function DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;

  function LLGradOfChunk(c :Integer; CompGrad, NumGrad :Boolean; var TotalLL :Double) :Boolean;
  //function SimOfChunk(c :Integer) :Boolean;
  function LLGradOfAllChunks(CompGrad, NumGrad :Boolean; var TotalLL :Double) :Boolean;

  procedure OnGetProperties(Sender :TObject); override;
  procedure OnPutProperties(Sender :TObject); override;


  //Model
  property Model                        :TQUB_QVO_QVM                 read prModel             write prSetModel;

  //Parameters
  property OptParams                    :TQUB_Mac_Opt_Params          read prOptParams         write prOptParams;
  property OptPParams                   :TQUB_Mac_Opt_PParams         read prGetOptPParams;

  property QuietOutput                  :Boolean                      read prQuietOutput       write prQuietOutput;
  property PublishResults               :Boolean                      read prDoPublishResults  write prDoPublishResults;

  property FitDataListName              :String                       read prFitDataListName   write prFitDataListName;
  property ExcDataListName              :String                       read prExcDataListName   write prExcDataListName;

  property Signal                       :TSimpleEvent                 read prSignal;


end;


implementation


uses


QUB_Main, QUB_GUIInterface, QUB_Viewport_Data, QUB_Python_Data, QUB_FitCurves,
QUB_QFS_QIF_Segments,
QUB_Task_AMP,
QUB_Idealization,
QUB_ActionSelector_Preprocessing,
QUB_Viewport_Results,
QUB_AdaptiveResample;
//QUB_QVO_QVGraph;


//DFP Callbacks ****************************************************************
function DFPFunc(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
begin
  Result := TQUB_Task_Mac(call_obj).DFPFunc(Pars, Grad, LL, nf, ndf, CompGrad);
end;

function DFPCheck(call_obj :TObject; Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
begin
  Result := TQUB_Task_Mac(call_obj).DFPCheck(Iters, nf, ndf, LL, Pars, Grads);
end;

(*
//SolvOpt Callbacks ************************************************************
function SolvOptFunc(call_obj :TObject; Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
begin
  Result := TQUB_Task_Mac(call_obj).DFPFunc(Pars, Grad, LL, nf, ndf, CompGrad);
end;

function SolvOptCheck(call_obj :TObject; Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
begin
  Result := TQUB_Task_Mac(call_obj).DFPCheck(Iters, nf, ndf, LL, Pars, Grads);
end;
*)


//****************************************************************************//
//                                                                            //
//                           T Q U B _ T a s k _ M a c                        //
//                                                                            //
//****************************************************************************//
constructor TQUB_Task_Mac.Create(Suspended :Boolean);
var
  i :Integer;
begin

  inherited;

  prName     := 'Mac optimizer';

  prCanPause := True;
  prCanSkip  := True;

  prSignal   := TSimpleEvent.Create;

  //TO DO: LM. I did not finish modifying the code for two processor machines (local multi-threading).
  prChunkThreadCount := 0;//For two-processor machines. Make it an option
  if prChunkThreadCount > 0 then begin
    prChunkThreads := PAlloc1(prChunkThreadCount);
    for i := 0 to prChunkThreadCount - 1 do begin
      prChunkThreads[i] := TQUB_Task_Mac_ChunkThread.Create(True);
      TQUB_Task_Mac_ChunkThread(prChunkThreads[i]).MacTask  := Self;
      TQUB_Task_Mac_ChunkThread(prChunkThreads[i]).Affinity := i + 1;
    end;
  end;

  prResults  := TNode.Create('Mac Results');
  prResults['ResultType'].SetData('Mac');

  prDoPublishResults := True;
end;

destructor TQUB_Task_Mac.Destroy;
var
  i :Integer;
begin
  try
    if prChunkThreadCount > 0 then begin
      for i := 0 to prChunkThreadCount - 1 do begin
        TQUB_Task_Mac_ChunkThread(prChunkThreads[i]).Terminate;;
      end;
      prSignal.Free;
    end;
  finally
    inherited;
  end;
end;



//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Mac.prDoTask;
begin

  inherited;

  prMac;

end;

procedure TQUB_Task_Mac.prMac;
var
  Done             :Boolean;
  Cnt              :Integer;
  m, n             :Integer;
  BatchFirst       :Integer;
  BatchIter        :Integer;
  BatchCountCount  :Integer;
  BatchCountList   :TPAInteger;
begin


  prOptSegsChunks := TList.Create;

  BatchCountCount := 1;
  BatchCountList  := nil;


  if prOptParams.OptSegments = OPTSEGS_GROUPLIST then begin
    BatchCountCount := 100;
    BatchCountList  := IAlloc1(BatchCountCount);
    StrToArrayOfInt(prGroupList, BatchCountList, BatchCountCount);
    if BatchCountCount = 0 then begin
      prOptParams.OptSegments := OPTSEGS_GROUP;
      BatchCountCount := 1;
    end;
  end;


  try


    for BatchIter := 0 to BatchCountCount - 1 do begin


      prSkipped := False;

      if prStopped then Break;

      //Save them all
      for m := 0 to prChunks.Count - 1 do begin
        prOptSegsChunks.Add(prChunks[m]);
      end;

      //Initialize pointers in list
      BatchFirst := 0;
      if          prOptParams.OptSegments = OPTSEGS_TOGETHER{ALL} then begin
        prBatchCount := prChunks.Count;
      end else if prOptParams.OptSegments = OPTSEGS_GROUP then begin
        prBatchCount := prOptParams.OptSegmentsGroupCount;
      end else begin
        prBatchCount := Abs(BatchCountList[BatchIter]);
      end;


      Synchronize(prPrintOptParams);


      prBatchIdx := 0;


      while (prBatchIdx < prOptParams.OptSegmentsMaxBatch) and (BatchFirst < prOptSegsChunks.Count) and (not prStopped) do begin
      //Batches


        prSkipped := False;

        Inc(prBatchIdx);

        if prBatchCount < prOptSegsChunks.Count then begin
          prBatchLabel := Format('Batch %d', [prBatchIdx]);
        end;

        if (prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST) then begin
          Synchronize(prResetModel);
        end;

        if (BatchFirst + prBatchCount - 1) >= prOptSegsChunks.Count then begin
          prBatchCount := prOptSegsChunks.Count - BatchFirst;
        end;


        //Next Batch
        prChunks.Clear;


        for m := BatchFirst to BatchFirst + prBatchCount - 1 do begin
          prChunks.Add(prOptSegsChunks[m]);
        end;


        //Multiplex files (e.g., to test Nc from multiple (2) patches)
        if prOptParams.OptMUXFiles = 1 then begin
          for m := BatchFirst + prBatchCount to prOptSegsChunks.Count - 1 do begin
            if TQUB_Mac_PChunk(prOptSegsChunks[BatchFirst]).QVD <> TQUB_Mac_PChunk(prOptSegsChunks[m]).QVD then begin
              for n := m + BatchFirst to m + BatchFirst + prBatchCount - 1 do begin
                if n >= prOptSegsChunks.Count then Break;
                prChunks.Add(prOptSegsChunks[n]);
              end;
              Break;
            end;
          end;
        end;


        Synchronize(prShowData);


        //**********************************************************************
        Cnt  := 0;

        repeat
        //restart if Grads > some threshold...


          if Cnt > 0 then begin
            prConsoleMsg  := 'Restarting...';
            prConsoleCode := MESSAGE_WARNING;
            if not prQuietOutput then Synchronize(prWriteNewConsole);
            prpLL := 0;
          end;

          Done := True;


          prMaxChannel := -1;


          prfParsGrads := DAlloc1(TQUB_Mac_Model(prMacModels[0]).fParCount);
          prfParsOld   := DAlloc1(TQUB_Mac_Model(prMacModels[0]).fParCount);
          for m := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do prfParsGrads[m] := 0;


          //Used for Penalty numerical gradients
          prPenaltyGrads := DAlloc1(TQUB_Mac_Model(prMacModels[0]).fParCount);


          try


            for m := 0 to prMacModels.Count - 1 do begin

              TQUB_Mac_Model(prMacModels[m]).Do_fParsToModel;

            end;


            prStartWatch := Round(TimeStampToMSecs(DateTimeToTimeStamp(Now)));

            if prOptParams.OptMode = MODE_OPTIMIZE then begin
            //Optimize

              if prQuietOutput then begin
                prConsoleMsg  := Format('Batch index %d: starting ...', [prBatchIdx]);
                prConsoleCode := MESSAGE_WARNING;
                Synchronize(prWriteNewConsole);
                prQuietEntry := prConsoleEntry;
              end;

              try
                DFP_MIN(Self, TQUB_Mac_Model(prMacModels[0]).fParCount, TQUB_Mac_Model(prMacModels[0]).fPars, TQUB_Mac_Model(prMacModels[0]).fParsVar,
                        QUB_Task_Mac.DFPFunc, QUB_Task_Mac.DFPCheck,
                        OptParams.OptMaxIter, OptParams.OptLLConv, OptParams.OptGradConv, OptParams.OptMaxStep,
                        TQUB_Mac_Model(prMacModels[0]).Error);
              except
              end;

              TQUB_Mac_Model(prMacModels[0]).Do_Varf2Varc;


            end else if prOptParams.OptMode = MODE_CHECK then begin
            //Evaluate Log-Likelihood


              try
                prLLGrad(1{CompGrad});
              except
                SysUtils.Beep;
              end;

              TQUB_Mac_Model(prMacModels[0]).Do_Varf2Varc;

              prPrintGrads;

              Synchronize(prDoOverlapData);

            end else if prOptParams.OptMode = MODE_SIM then begin
            //Simulate the response to the same protocol as the data

              try
                prLLGrad(0{new}{CompGrad});
              except
                SysUtils.Beep;
              end;

              Synchronize(prDoOverlapData);

            end;


            //TO DO: idem. LM
            if prChunkThreadCount <= 0 then prUpdateResults(True{Final});
            prPrintParams(True{Final});


          finally


            //Check Grads:
            for m := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
              if Abs(prfParsGrads[m]) >= 1 then begin
                Done := False;
                Break;
              end;
            end;

            FreeMem(prfParsOld);
            FreeMem(prfParsGrads);

            //Penalty Grads
            FreeMem(prPenaltyGrads);


          end;


          Inc(Cnt);


          TQUB_Mac_Model(prMacModels[0]).Iterations := 0;

          if (OptParams.OptRestart = 0) and (Cnt > 0) then Break;

        until ((Done) or (Cnt >= 2) or (prStopped) or (prError));
        //**********************************************************************


        prConsoleMsg  := prOptSegsOutput;
        if prErrorMsg <> '' then begin
          prConsoleMsg := prConsoleMsg + #9 + 'Error: ' + prErrorMsg;
          prConsoleCode := MESSAGE_ERROR;
        end else begin
          prConsoleCode := MESSAGE_MESSAGE;
        end;
        if prQuietOutput then begin
          prConsoleEntry := prQuietEntry;
          Synchronize(prWriteSetConsole);
        end else begin
          Synchronize(prWriteNewConsole);
        end;


        //Next batch
        Inc(BatchFirst, prBatchCount);


      end;


      if ((prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST)) and (prBatchCount > 1) then Synchronize(prResetModel);

      prChunks.Clear;

      for m := 0 to prOptSegsChunks.Count - 1 do begin
        prChunks.Add(prOptSegsChunks[m]);
      end;

      prOptSegsChunks.Clear;


    end;//BatchIter

    if prDoPublishResults then Synchronize(prPublishResults);


  finally

    prOptSegsChunks.Free;
    prOptSegsOldMdl.Free;
    FreeMem(BatchCountList);

  end;

end;


procedure TQUB_Task_Mac.prPrintOptParams;
var
  t, h, s :String;
  k, m    :Integer;
begin

  with prOptParams do begin

    t :=     #13#10'Mac parameters for:';
    t := t + #13#10#9'data  '    + Choice(prData  <> nil, ExtractFileName(prData.FileName),  '<none>');
    t := t + #13#10#9'model '    + Choice(prModel <> nil, ExtractFileName(prModel.FileName), '<none>');
    t := t + #13#10#9'Est (K'    + Choice(Boolean(EstChannelCount), ', CC' + Choice(Boolean(EstChannelCountPerFile), '(PF)', ''), '') +
                                   Choice(Boolean(EstAmpOpen), ', AO', '') +
                                   Choice(Boolean(EstVarOpen), ', SO', '') +
                                   '), ';
    t := t +         'Method '   + Choice(Boolean(EstUseVariance), 'LL', 'SS') + ', ';
    t := t +         'P0 ('      + Choice(Boolean(EstInitProbs = INITPROBS_FIXED), 'fixed', Choice(Boolean(EstInitProbs = INITPROBS_EQ), 'eq', 'condeq')) + '), ';
    t := t +         'SegCount ' + IntToStr(prBatchCount) + '.';
    t := t + #13#10;


    //TO DO: rework for multiple models - do it as a different task. LM

    //Print column headers for quiet output:
    if QuietOutput then begin
      //TO DO: finish it. LM
      h := '';
      with TQUB_Mac_Model(prMacModels[0]) do begin

        //Rates
        for k := 0 to RateCount - 1 do begin
          h := h + Format('%.10s', [Format('k%d-%d', [RateIdx[k].x + 1, RateIdx[k].y + 1]) + '          ']) + #9 +
                   Format('%.10s', ['std' + '          ']) + #9;
        end;

        h := h + Format('%.14s', ['LL'    + '              ']) + #9 +
                 Format('%.14s', ['SS'    + '              ']) + #9 +
                 Format('%.14s', ['AIC'   + '              ']) + #9 +
                 Format('%.14s', ['BIC'   + '              ']) + #9 +
                 Format('%.14s', ['MaxNc' + '              ']) + #9;

        //Extra kinetic parameters
        for k := 0 to ExtraKinParCount - 1 do begin
          h := h + Format('%.10s', [prModel.ModelObject.ExtraKineticPars[k].Name + '          ']) + #9 +
                   Format('%.10s', ['std' + '          ']) + #9;
        end;

        //ChannelCount
        if OptParams.EstChannelCount <> 0 then begin
          if prOptParams.EstChannelCountPerFile = 0 then begin
            h := Format('%.10s', ['Nc'    + '          ']) + #9 +
                 Format('%.10s', ['std'   + '          ']) + #9 +
                 h;
          end else begin
            s := '';
            for m := 0 to ChannelCountCount - 1 do begin
              s := s + Format('%.10s', ['Nc'    + '          ']) + #9 +
                       Format('%.10s', ['std'   + '          ']) + #9;
            end;
            h := s + h;
          end;
        end;

        //AmpOpen
        if OptParams.EstAmpOpen <> 0 then begin
          h := Format('%.10s', ['AmpO'  + '          ']) + #9 +
               Format('%.10s', ['std'   + '          ']) + #9 +
               h;
        end;

        //VarOpen
        if OptParams.EstVarOpen <> 0 then begin
          h := Format('%.10s', ['StdO'  + '          ']) + #9 +
               Format('%.10s', ['std'   + '          ']) + #9 +
               h;
        end;

      end;//with

      h := #13#10 + h;

    end;

  end;

  prConsoleMsg  := t + h;
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
end;

procedure TQUB_Task_Mac.prPrintParams(Final :Boolean);
var
  LogL :Double;
  SS   :Double;
  AIC  :Double;
  BIC  :Double;
  p    :Double;
  ekp  :Double;
  DC   :Integer;
  t, s :String;
  k, m :Integer;
  Diff :Int64;
  RateCount        :Integer;
  ExtraKinParCount :Integer;
begin

  LogL := 0;
  SS   := 0;
  DC   := 0;
  for m := 0 to prMacModels.Count - 1 do begin
    LogL := LogL + TQUB_Mac_Model(prMacModels[m]).LL;
    SS   := SS   + TQUB_Mac_Model(prMacModels[m]).SS;
    DC   := DC   + TQUB_Mac_Model(prMacModels[m]).PointCount;
  end;

  //TO DO: Change if models are additive (e.g., as in Na and K currents analyzed together). LM
  AIC := 2*(TQUB_Mac_Model(prMacModels[0]).fParCount            + LogL{it's already negative});
  BIC := 2*(TQUB_Mac_Model(prMacModels[0]).fParCount*ln(DC)*0.5 + LogL);

  RateCount := TQUB_Mac_Model(prMacModels[0]).RateCount;
  ExtraKinParCount := TQUB_Mac_Model(prMacModels[0]).ExtraKinParCount;

  prOptSegsOutput :=     '';

  if Final then begin
    try
      t :=            #13#10'================================================================';
      t := t + Format(#13#10'MAC final results:  LL = %-14.10g SS = %-14.10g AIC = %-14.10g BIC = %-14.10g Iterations = %d',
                     [-LogL, -SS, AIC, BIC, TQUB_Mac_Model(prMacModels[0]).Iterations]);
    except
    end;
  end;

  //Rates
  t := t +   #13#10#9'Rates:  '#13#10#9's1  s2   k0         k0.std     k1         k1.std' +
                             #13#10#9'--------------------------------------------------';

  for k := 0 to RateCount - 1 do begin
    try
      t := t + Format(#13#10#9'%d   %d    %-10.6g %-10.6g %-10.6g %-10.6g',
                     [TQUB_Mac_Model(prMacModels[0]).RateIdx[k].x+1, TQUB_Mac_Model(prMacModels[0]).RateIdx[k].y+1,
                     //!!! cParsVar
                      Choice(OptParams.OptAllowNegRates,
                            TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0],
                        Exp(TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0])
                      ),
                                                                      Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 0]),
                      TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 1],  Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 1])
                   ]);
      prOptSegsOutput := prOptSegsOutput + Trim(Format('%10.6g', [Choice(OptParams.OptAllowNegRates,
                                                                        TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0],
                                                                    Exp(TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0])
                                                                  )
                                                                                                                        ])) + #9 +
                                           Trim(Format('%10.6g', [Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 0])])) + #9;
    except
    end;
  end;
  prOptSegsOutput := prOptSegsOutput + Trim(Format('%14.10g'#9'%14.10g'#9'%14.10g'#9'%14.10g', [-LogL, -SS, AIC, BIC])) + #9 + IntToStr(prMaxChannel);

  t := t + #13#10#9'--------------------------------------------------';


  //Exra Kinetic Parameters
  if ExtraKinParCount > 0 then begin
    t := t + #13#10#13#10#9'Extra kinetic parameters:  '#13#10#9'name      mean      std' +
             #13#10#9'--------------------------------------------------';
    for k := 0 to ExtraKinParCount - 1 do begin
      try
        case TQUB_Mac_Model(prMacModels[0]).Model.ExtraKineticPars[k].EKPType of
          EKP_TYPE_LIN:
            ekp :=     TQUB_Mac_Model(prMacModels[0]).cPars[2*RateCount + k];
          EKP_TYPE_LOG:
            ekp := Exp(TQUB_Mac_Model(prMacModels[0]).cPars[2*RateCount + k]);
          EKP_TYPE_EXP:
            ekp := Ln (TQUB_Mac_Model(prMacModels[0]).cPars[2*RateCount + k]);
          else
            ekp := Exp(TQUB_Mac_Model(prMacModels[0]).cPars[2*RateCount + k]);
        end;
        t := t + Format(#13#10#9'%.10s %-10.6g %-10.6g',
                       [TQUB_Mac_Model(prMacModels[0]).Model.ExtraKineticPars[k].Name + '          ',
                        ekp,
                        Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + k])
                       ]);
        prOptSegsOutput := prOptSegsOutput + Trim(Format('%10.6g', [ekp])) + #9 +
                                             Trim(Format('%10.6g', [Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + k])])) + #9;
      except
      end;
    end;
    t := t + #13#10#9'--------------------------------------------------';

  end;

  //ChannelCount
  if OptParams.EstChannelCount <> 0 then begin
    try
      if prOptParams.EstChannelCountPerFile = 0 then begin
        t := t + Format(#13#10#13#10#9'ChannelCount:  '#13#10#9'%d   %-10.6g',
                        [Round(Exp (TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount])),
                               Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount])]);
        prOptSegsOutput := Trim(Format('%d',     [Round(Exp(TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount]))])) + #9 +
                           Trim(Format('%10.6g', [Sqrt(     TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount]) ])) + #9 +
                           prOptSegsOutput;
      end else begin
        t := t + #13#10#13#10'ChannelCount:  '#13#10#9;
        s := '';
        for m := 0 to TQUB_Mac_Model(prMacModels[0]).ChannelCountCount - 1 do begin
          t := t + Format('%d   %-10.6g  ',
                          [Round(Exp(TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + m])),
                                Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + m])]);
          s := s + Trim(Format('%d',     [Round(Exp (TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + m]))])) + #9 +
                   Trim(Format('%10.6g', [      Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + m]) ])) + #9;
        end;
        prOptSegsOutput := s + prOptSegsOutput;

      end;
    except
    end;
  end;

  //AmpOpen
  if OptParams.EstAmpOpen <> 0 then begin
    try
      t := t + Format(#13#10#13#10#9'AmpOpen:  '#13#10#9'%-10.6g   %-10.6g',
                      [     TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount],
                       Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount])]);
      prOptSegsOutput := Trim(Format('%10.6g', [     TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount]])) + #9 +
                         Trim(Format('%10.6g', [Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount]) ])) + #9 +
                         prOptSegsOutput;
    except
    end;
  end;

  //VarOpen
  if OptParams.EstVarOpen <> 0 then begin
    try
      t := t + Format(#13#10#13#10#9'VarOpen:  '#13#10#9'%-10.6g   %-10.6g',
                      [ Exp(TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount + TQUB_Mac_Model(prMacModels[0]).AmpOpenCount]),
                       Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount + TQUB_Mac_Model(prMacModels[0]).AmpOpenCount])]);
      prOptSegsOutput := Trim(Format('%10.6g', [ Exp(TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount + TQUB_Mac_Model(prMacModels[0]).AmpOpenCount])])) + #9 +
                         Trim(Format('%10.6g', [Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount + TQUB_Mac_Model(prMacModels[0]).AmpOpenCount])])) + #9 +
                         prOptSegsOutput;
    except
    end;
  end;

  //prBatchOutput := #13#10 + prBatchOutput;

  if Final then begin
    if OptParams.OptMode = MODE_OPTIMIZE then begin
      //Free Pars Grads
      t := t + #13#10'Free pars grads: ';
      for k := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
        try
          t := t + Format('%-12.8g ', [-prfParsGrads[k]]);
        except
        end;
      end;
      //Free Pars mean and covariance
      t := t + #13#10'Free pars mean and covariance: '#13#10;
      for k := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
        p := TQUB_Mac_Model(prMacModels[0]).fPars[k];
        try
          t := t + Format(#9'%-12.8g '#9, [p]);
        except
        end;
        for m := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
          p := TQUB_Mac_Model(prMacModels[0]).fParsVar[k, m];
          try
            t := t + Format('%-12.8g ', [p]);
          except
          end;
        end;
        t := t + #13#10;
      end;
    end;
    t := t + #13#10 + '================================================================';
  end else begin
    t := t + #13#10;
  end;

  if Final then begin
  //Running time
    Diff := Round(TimeStampToMSecs(DateTimeToTimeStamp(Now))) - prStartWatch;
    t := t + #13#10'Running time: ' + FloatToStr(Diff / 1000) + ' secs, ' + FloatToStr(Diff/1000/Max(1, TQUB_Mac_Model(prMacModels[0]).FuncCalls)) + ' secs/function call';
  end;


  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);
  // Always show convergence info at Fred's request.  11-2004 {JB}
  {
    Not possible, Quiet output is meant for batch processing mainly,
    for which tabular output is generated. Tell Fred to turn off Quiet.
    He can talk to me about this. I will re-work the output anyway,
    to make it more customizable
  }
end;

procedure TQUB_Task_Mac.prPrintGrads;
var
  LogL  :Double;
  dLogL :Double;
  t     :String;
  k, m  :Integer;
begin

  LogL := 0;
  for m := 0 to prMacModels.Count - 1 do begin
    LogL := LogL + TQUB_Mac_Model(prMacModels[m]).LL;
  end;

  prConsoleMsg  := Format('Batch %d-%d, %s LL: %-14.10g, Gr: ', [prBatchIdx, TQUB_Mac_Model(prMacModels[0]).Iterations, prErrorMsg, -LogL]);

  t := Format('Iteration %d, LL = %-14.10g, Func calls = %d, Grad calls = %d',
              [TQUB_Mac_Model(prMacModels[0]).Iterations, -LogL, TQUB_Mac_Model(prMacModels[0]).FuncCalls, TQUB_Mac_Model(prMacModels[0]).GradCalls]);

  t := t + #13#10#9'Free pars grads: ';
  for k := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
    dLogL := 0;
    for m := 0 to prMacModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Mac_Model(prMacModels[m]).dLLdfPar[k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
      prConsoleMsg  := prConsoleMsg + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  //t must be assigned here
  if prQuietOutput then begin
    if prErrorMsg <> '' then begin
      prConsoleCode := MESSAGE_ERROR;
    end else begin
      prConsoleCode := MESSAGE_WARNING;
    end;
    Synchronize(prWriteSetConsole);
  end;

  t := t + #13#10#9'Cstr pars grads: ';
  for k := 0 to TQUB_Mac_Model(prMacModels[0]).cParCount - 1 do begin
    dLogL := 0;
    for m := 0 to prMacModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Mac_Model(prMacModels[m]).dLLdcPar[k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  t := t + #13#10#9'Rates grads:     ';
  for k := 0 to TQUB_Mac_Model(prMacModels[0]).RateCount - 1 do begin
    dLogL := 0;
    for m := 0 to prMacModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Mac_Model(prMacModels[m]).dLLdq[0,k];
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin
    t := t + #13#10#9'Cond_Rates grads:     ';
    for k := 0 to TQUB_Mac_Model(prMacModels[0]).RateCount - 1 do begin
      dLogL := 0;
      for m := 0 to prMacModels.Count - 1 do begin
        dLogL := dLogL + TQUB_Mac_Model(prMacModels[m]).dLLdCond_q[k];
      end;
      try
        t := t + Format('%-12.8g ', [-dLogL]);
      except
      end;
    end;
  end;

  //Extra kinetic parameters - to do
  // . . .

  //ChannelCount
  if OptParams.EstChannelCount <> 0 then begin
    if prOptParams.EstChannelCountPerFile = 0 then begin
      t := t + #13#10#9'ChCount grads:   ';
      dLogL := 0;
      for m := 0 to prMacModels.Count - 1 do begin
        dLogL := dLogL + TQUB_Mac_Model(prMacModels[m]).dLLdN;
      end;
      try
        t := t + Format('%-12.8g ', [-dLogL]);
      except
      end;
    end else begin
      t := t + #13#10#9'ChCount grads:   ';
      for m := 0 to prMacModels.Count - 1 do begin
        dLogL := TQUB_Mac_Model(prMacModels[m]).dLLdN;
        try
          t := t + Format('%-12.8g ', [-dLogL]);
        except
        end;
      end;
    end;
  end;

  //AmpOpen
  if OptParams.EstAmpOpen <> 0 then begin
    t := t + #13#10#9'AmpO grads:   ';
    dLogL := 0;
    for m := 0 to prMacModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Mac_Model(prMacModels[m]).dLLdA;
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  //VarOpen
  if OptParams.EstVarOpen <> 0 then begin
    t := t + #13#10#9'VarO grads:   ';
    dLogL := 0;
    for m := 0 to prMacModels.Count - 1 do begin
      dLogL := dLogL + TQUB_Mac_Model(prMacModels[m]).dLLdV;
    end;
    try
      t := t + Format('%-12.8g ', [-dLogL]);
    except
    end;
  end;

  prConsoleMsg  := t;
  prConsoleCode := MESSAGE_MESSAGE;
  if not prQuietOutput then Synchronize(prWriteNewConsole);

end;

//TO DO: LM. Check the costs of this
procedure TQUB_Task_Mac.prUpdateResults(Final :Boolean);
var
  k, m :Integer;
  PointCount :Integer;
  DS0   :INode;
  DS    :INode;
  Lbl   :String;
  Res   :INode;
  AAcc  :Array of Double;
  Acc   :Double;
  IAcc  :Integer;
  x     :Double;
  i     :Integer;
  N     :Integer;
  FinalMdl :INode;
  RateCount :Integer;
  ExtraKinParCount :Integer;

  function BatchLbl(const Name :String): String;
  begin
    if prBatchLabel = ''
    then Result := Name
    else Result := prBatchLabel + ' ' + Name;
  end;
  function DResult(Parent :INode; const Lbl :String; Count :Integer): INode;
  begin
    Result := Parent[Lbl];
    if LongInt(Result.DataCount) <> Count then begin
      Result.SetupNumData(NODE_DATATYPE_DOUBLE, Count, 1);
    end;
  end;
  function IResult(Parent :INode; const Lbl :String; Count :Integer): INode;
  begin
    Result := Parent[Lbl];
    if LongInt(Result.DataCount) <> Count then begin
      Result.SetupNumData(NODE_DATATYPE_LONGINT, Count, 1);
    end;
  end;
  // array accumulator -- for summing several vectors prior to copying sum array into result node
  procedure AAccSetup(Count :Integer);
  var
    k   :Integer;
  begin
    SetLength(AAcc, Count);
    for k := 0 to Count-1 do begin
      AAcc[k] := 0.0;
    end;
  end;

  procedure UpdateRateStd(A, B :Integer; IsK1 :Boolean; Std :Double);
  var
    Rate   :INode;
    States :INode;
    DK     :INode;
    Dir    :Integer;
  begin
    Rate   := FinalMdl['Rates'].Find('Rate');
    while not Rate.IsNull do begin
      States := Rate['States'];
      if (States.Data[0] = A) and (States.Data[1] = B) then begin
        Dir := 0;
      end else if (States.Data[0] = B) and (States.Data[1] = A) then begin
        Dir := 1;
      end else begin
        Dir := -1;
      end;

      if Dir >= 0 then begin
        if IsK1 then begin
          DK := Rate['dk1'];
        end else begin
          DK := Rate['dk0'];
        end;
        if DK.DataCount < 2 then DK.SetupNumData(NODE_DATATYPE_DOUBLE, 2, 1);
        DK.Data[(0+Dir) mod 2] := Std;

        Break;
      end;

      Rate := Rate.SiblingSameName;
    end;
  end;

begin
  if Final then begin
    FinalMdl := Model.ModelObject.CloneAsNode(True, TNode.CreateNull);
    FinalMdl.SetData(BatchLbl('Final'));
  end;

  RateCount := TQUB_Mac_Model(prMacModels[0]).RateCount;
  ExtraKinParCount := TQUB_Mac_Model(prMacModels[0]).ExtraKinParCount;

  DS0  := prResults.Find('DataSet');

  Lbl  := BatchLbl('PointCount');
  DS   := DS0;
  PointCount  := 0;
  for m := 0 to prMacModels.Count - 1 do begin
    IResult(DS, Lbl, 1).Data[0] := TQUB_Mac_Model(prMacModels[m]).PointCount;
    PointCount    := PointCount + TQUB_Mac_Model(prMacModels[m]).PointCount;
    DS   := DS.SiblingSameName;
  end;
  IResult(prResults, Lbl, 1).Data[0] := PointCount;

  Lbl  := BatchLbl('LL');
  DS   := DS0;
  Acc  := 0.0;
  for m := 0 to prMacModels.Count - 1 do begin
    x    := - TQUB_Mac_Model(prMacModels[m]).LL;
    Acc  := Acc + x;
    DResult(DS, Lbl, 1).Data[0] := x;
    DResult(DS, 'AIC', 1).Data[0] := 2*(TQUB_Mac_Model(prMacModels[0]).fParCount            - x);
    DResult(DS, 'BIC', 1).Data[0] := 2*(TQUB_Mac_Model(prMacModels[0]).fParCount*ln(PointCount)*0.5 - x);
    DS   := DS.SiblingSameName;
  end;
  DResult(prResults, Lbl, 1).Data[0]   := Acc;
  //TO DO: Change if models are additive (e.g., as in Na and K currents analyzed together). LM
  DResult(prResults, 'AIC', 1).Data[0] := 2*(TQUB_Mac_Model(prMacModels[0]).fParCount            - Acc);
  DResult(prResults, 'BIC', 1).Data[0] := 2*(TQUB_Mac_Model(prMacModels[0]).fParCount*ln(PointCount)*0.5 - Acc);

  Lbl  := BatchLbl('SS');
  DS   := DS0;
  Acc  := 0.0;
  for m := 0 to prMacModels.Count - 1 do begin
    x  := - TQUB_Mac_Model(prMacModels[m]).SS;
    DResult(DS, Lbl, 1).Data[0] := x;
    Acc         := Acc + x;
    DS   := DS.SiblingSameName;
  end;
  DResult(prResults, Lbl, 1).Data[0] := Acc;

  IResult(prResults, 'Iterations', 1).Data[0] := TQUB_Mac_Model(prMacModels[0]).Iterations;

  //Rates
  for k := 0 to RateCount - 1 do begin
    try
      Lbl  := Format('%d->%d',[TQUB_Mac_Model(prMacModels[0]).RateIdx[k].x+1, TQUB_Mac_Model(prMacModels[0]).RateIdx[k].y+1]);
      DResult(prResults, BatchLbl(Lbl), 1).Data[0] := Choice(OptParams.OptAllowNegRates,
                                                       TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0],
                                                       Exp(TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0])
                                                     );
      DResult(prResults, BatchLbl(Lbl+' std'), 1).Data[0] := Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 0]);
      DResult(prResults, BatchLbl(Lbl+' k1'), 1).Data[0]  := TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 1];
      DResult(prResults, BatchLbl(Lbl+' k1 std'), 1).Data[0] := Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 1]);
      if Final then begin
        UpdateRateStd(TQUB_Mac_Model(prMacModels[0]).RateIdx[k].x, TQUB_Mac_Model(prMacModels[0]).RateIdx[k].y, False, Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 0]));
        UpdateRateStd(TQUB_Mac_Model(prMacModels[0]).RateIdx[k].x, TQUB_Mac_Model(prMacModels[0]).RateIdx[k].y, True, Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 1]));
      end;
    except
    end;
  end;

  //Extra kinetic parameters - to do
  //. . .

  //ChannelCount
  if OptParams.EstChannelCount <> 0 then begin

    //TO DO: There might be a problem with CCFixed - ignore it for now. LM

    IAcc := 0;
    Acc  := 0.0;
    try
      if prOptParams.EstChannelCountPerFile <> 0 then begin
        DS   := DS0;
        for m := 0 to TQUB_Mac_Model(prMacModels[0]).ChannelCountCount - 1 do begin
          i    := Round(Exp(TQUB_Mac_Model(prMacModels[0]).cPars[2*RateCount + ExtraKinParCount + m]));
          IAcc := IAcc + i;
          x    := Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + m]);
          Acc  := Acc + x;
          IResult(DS, BatchLbl('ChannelCount'), 1).Data[0] := i;
          DResult(DS, BatchLbl('ChannelCount std'), 1).Data[0] := x;

          DS := DS.SiblingSameName;
        end;
      end else begin
        IAcc := Round(Exp(TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount]));
        Acc  := Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount]);
      end;
    except
    end;
    IResult(prResults, BatchLbl('ChannelCount'), 1).Data[0] := IAcc;
    DResult(prResults, BatchLbl('ChannelCount std'), 1).Data[0] := Acc;
  end;


  //AmpOpen
  if OptParams.EstAmpOpen <> 0 then begin
    try
      DResult(prResults, BatchLbl('AmpOpen'), 1).Data[0] := TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount];
      DResult(prResults, BatchLbl('AmpOpen std'), 1).Data[0] := Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount]);
    except
    end;
  end;

  //VarOpen
  if OptParams.EstVarOpen <> 0 then begin
    try
      DResult(prResults, BatchLbl('VarOpen'), 1).Data[0] := Exp(TQUB_Mac_Model(prMacModels[0]).cPars   [2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount + TQUB_Mac_Model(prMacModels[0]).AmpOpenCount]);
      DResult(prResults, BatchLbl('VarOpen std'), 1).Data[0] := Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*RateCount + ExtraKinParCount + TQUB_Mac_Model(prMacModels[0]).ChannelCountCount + TQUB_Mac_Model(prMacModels[0]).AmpOpenCount]);
    except
    end;
  end;

  Lbl   := BatchLbl('Free pars grads');
  N     := TQUB_Mac_Model(prMacModels[0]).fParCount;
  AAccSetup(N);
  DS    := DS0;
  for m := 0 to prMacModels.Count - 1 do begin
    Res := DResult(DS, Lbl, N);
    for k := 0 to N - 1 do begin
      x           := - TQUB_Mac_Model(prMacModels[m]).dLLdfPar[k];
      Res.Data[k] := x;
      AAcc[k]     := AAcc[k] + x;
    end;
    DS  := DS.SiblingSameName;
  end;
  Res   := DResult(prResults, Lbl, N);
  for k := 0 to N - 1 do begin
    Res.Data[k]   := AAcc[k];
  end;

  Lbl   := BatchLbl('Cstr pars grads');
  N     := TQUB_Mac_Model(prMacModels[0]).cParCount;
  AAccSetup(N);
  DS    := DS0;
  for m := 0 to prMacModels.Count - 1 do begin
    Res := DResult(DS, Lbl, N);
    for k := 0 to N - 1 do begin
      x           := - TQUB_Mac_Model(prMacModels[m]).dLLdcPar[k];
      Res.Data[k] := x;
      AAcc[k]     := AAcc[k] + x;
    end;
    DS  := DS.SiblingSameName;
  end;
  Res   := DResult(prResults, Lbl, N);
  for k := 0 to N - 1 do begin
    Res.Data[k]   := AAcc[k];
  end;

  Lbl   := BatchLbl('Rates grads');
  N     := TQUB_Mac_Model(prMacModels[0]).RateCount;
  AAccSetup(N);
  DS    := DS0;
  for m := 0 to prMacModels.Count - 1 do begin
    Res := DResult(DS, Lbl, N);
    for k := 0 to N - 1 do begin
      x           := - TQUB_Mac_Model(prMacModels[m]).dLLdq[0,k];
      Res.Data[k] := x;
      AAcc[k]     := AAcc[k] + x;
    end;
    DS  := DS.SiblingSameName;
  end;
  Res   := DResult(prResults, Lbl, N);
  for k := 0 to N - 1 do begin
    Res.Data[k]   := AAcc[k];
  end;

  if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin
    Lbl   := BatchLbl('Cond_Rates grads');
    N     := TQUB_Mac_Model(prMacModels[0]).RateCount;
    AAccSetup(N);
    DS    := DS0;
    for m := 0 to prMacModels.Count - 1 do begin
      Res := DResult(DS, Lbl, N);
      for k := 0 to N - 1 do begin
        x           := - TQUB_Mac_Model(prMacModels[m]).dLLdCond_q[k];
        Res.Data[k] := x;
        AAcc[k]     := AAcc[k] + x;
      end;
      DS  := DS.SiblingSameName;
    end;
    Res   := DResult(prResults, Lbl, N);
    for k := 0 to N - 1 do begin
      Res.Data[k]   := AAcc[k];
    end;
  end;

  //Extra kinetic parameters - to do
  //. . .

  //ChannelCount
  if OptParams.EstChannelCount <> 0 then begin
    //CCFixed?
    Lbl   := BatchLbl('ChannelCount grads');
    Acc   := 0.0;
    DS    := DS0;
    for m := 0 to prMacModels.Count - 1 do begin
      x   := - TQUB_Mac_Model(prMacModels[m]).dLLdN;
      DResult(DS, Lbl, N).Data[0] := x;
      Acc := Acc + x;
      DS  := DS.SiblingSameName;
    end;
    DResult(prResults, Lbl, 1).Data[0] := Acc;
  end;

  if OptParams.EstAmpOpen <> 0 then begin
    Lbl   := BatchLbl('Amp0 grads');
    Acc   := 0.0;
    DS    := DS0;
    for m := 0 to prMacModels.Count - 1 do begin
      x   := - TQUB_Mac_Model(prMacModels[m]).dLLdA;
      DResult(DS, Lbl, N).Data[0] := x;
      Acc := Acc + x;
      DS  := DS.SiblingSameName;
    end;
    DResult(prResults, Lbl, 1).Data[0] := Acc;
  end;

  if OptParams.EstVarOpen <> 0 then begin
    Lbl   := BatchLbl('Var0 grads');
    Acc   := 0.0;
    DS    := DS0;
    for m := 0 to prMacModels.Count - 1 do begin
      x   := - TQUB_Mac_Model(prMacModels[m]).dLLdV;
      DResult(DS, Lbl, N).Data[0] := x;
      Acc := Acc + x;
      DS  := DS.SiblingSameName;
    end;
    DResult(prResults, Lbl, 1).Data[0] := Acc;
  end;

  if Final then begin
    prResults.AppendChild(FinalMdl);
  end;
end;

procedure TQUB_Task_Mac.prPublishResults;
begin
  if prSelectionList <> nil then begin          // if the selection list is 'Mac Results',
    prSelectionList.UnLock;                     // publishing the results will cause it to be replaced,
    prSelectionList := nil;                     // which can't happen while it's locked (make it nil so it's not doubly-unlocked)
  end;

  ViewportResults.AddResults( prResults, False );
end;



//******************************************************************************
function TQUB_Task_Mac.DFPFunc(Pars :TPADouble; Grad :TPADouble; var LL :Double; var nf, ndf :Integer; CompGrad{yes/no} :Integer) :Integer; stdcall;
var
  NANGrads :Boolean;
  i, m     :Integer;
begin

  Result := 0;

  if TQUB_Mac_Model(prMacModels[0]).Iterations > 0 then begin

    for m := 0 to prMacModels.Count - 1 do begin

      //Copy the new pars into the Mac models
      for i := 0 to TQUB_Mac_Model(prMacModels[m]).fParCount - 1 do begin
        TQUB_Mac_Model(prMacModels[m]).fPars[i] := Pars[i];
      end;

      TQUB_Mac_Model(prMacModels[m]).Do_fParsToModel;

    end;

  end;


  try
    prLLGrad(CompGrad);
  except
    SysUtils.Beep;
  end;


  try

    LL := 0;
    for m := 0 to prMacModels.Count - 1 do begin
      LL := LL + TQUB_Mac_Model(prMacModels[m]).LL;
    end;
    if Abs(LL) > 1e100 then begin
      LL := 1e100;
      //testing
      for i := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
        Grad[i] := 0;
      end;
    end;


    if CompGrad <> 0 then begin
      for i := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
        Grad[i] := 0;
        for m := 0 to prMacModels.Count - 1 do begin
          Grad[i] := Grad[i] + TQUB_Mac_Model(prMacModels[m]).dLLdfPar[i];
        end;
      end;

      //Testing
      //Check if the gradients are +/-NAN, in which case they should be set to reasonable values
      //  and LL made very small.
      NANGrads := False;
      for i := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
        if Abs(Grad[i]) > 1e100 then begin
          //Maybe it's better to set the grads to zero?

          //Testing
          //if Grad[i] < 0 then Grad[i] := -1e100 else Grad[i] := 1e100;
          Grad[i] := 0;

          NANGrads := True;
        end;
      end;
      if NANGrads then begin
        LL := 1e100;
      end;

      Move(Grad[0], prfParsGrads[0], TQUB_Mac_Model(prMacModels[0]).fParCount * SizeOf(Double));
    end;

  except
    SysUtils.Beep;
  end;

  //Inc(TQUB_Mac_Model(prMacModels[0]).Iterations);
  Inc(nf);
  if CompGrad <> 0 then Inc(ndf);
  TQUB_Mac_Model(prMacModels[0]).FuncCalls := nf;
  TQUB_Mac_Model(prMacModels[0]).GradCalls := ndf;

end;

function TQUB_Task_Mac.DFPCheck(Iters, nf, ndf :Integer; LL :Double; Pars :TPADouble; Grads :TPADouble) :Integer; stdcall;
var
  LLConverged :Boolean;
  GRConverged :Boolean;
  k, m        :Integer;
begin

  Result := 0;

  TQUB_Mac_Model(prMacModels[0]).Iterations := Iters;

  if Iters >= OptParams.OptMaxIter then Result := 1;

  LLConverged := Abs(LL - prpLL) < OptParams.OptLLConv;
  prpLL := LL;

  GRConverged := True;
  for k := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin
    if Abs(Grads[k]) > OptParams.OptGradConv then GRConverged := False;
  end;


  Synchronize(prDoOverlapData);
  Synchronize(prUpdateData);
  Synchronize(prUpdateModel);
  //if OptParams.OptPlotLL <> 0 then Synchronize(prUpdateLLPlot);


  //TO DO: Chris, there is an error in UpdateResults when using threads. Please check. LM.
  if prChunkThreadCount <= 0 then prUpdateResults(False{Final});
  //Testing

  prPrintGrads;
  prPrintParams(False{Final});


  if (LLConverged) and{or} (GRConverged) then Result := 1;
  if prStopped then Result := 1;
  if prSkipped then Result := 1;

  //if Result = 1 then begin
    //Change the parameters to the last good ones (???)
    for m := 0 to prMacModels.Count - 1 do begin
      //Copy the new pars into the Mac models
      for k := 0 to TQUB_Mac_Model(prMacModels[m]).fParCount - 1 do begin
        TQUB_Mac_Model(prMacModels[m]).fPars[k] := Pars[k];
      end;
      TQUB_Mac_Model(prMacModels[m]).Do_fParsToModel;
    end;
    //Only the fParsGrads
    Move(Grads[0], prfParsGrads[0], TQUB_Mac_Model(prMacModels[0]).fParCount * SizeOf(Double));
  //end;

end;



//****************************************************************************//
//                                                                            //
//                             L L   +   G R A D                              //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Mac.prLLGrad(CompGrad :Integer) :Integer;
var
  c, k, m, t, f                :Integer;
  SomeConfluent                :Boolean;
  ErrorA                       :Boolean;
  i, j                         :Integer;
  ThreadsDone                  :Boolean;
  SmallRateWarning             :Boolean;
  BigRateWarning               :Boolean;

  NumGrad                      :Boolean;//On, if calculating LL for the purpose of
                                        // computing the numerical Gradients
  DoGrad                       :Boolean;
  
  TotalLL                      :Double;
  ThisLL, ThisGrad             :Double;
  LLLeft, LLRight              :Double;

  procedure CalculateLL;
  var
    c :Integer;
  begin

    //**************************************************************************//
    //                                                                          //
    //                                                                          //
    //                   C H U N K S         I D E N T I C A L                  //
    //                                                                          //
    //                                                                          //
    //**************************************************************************//
    if OptParams.OptIdenticalChunks = 1 then begin
    //Identical chunks: faster, but it loads all data
    //Obviously, use only with identical chunks, whether in the same file or not


      LLGradOfAllChunks(DoGrad, NumGrad, TotalLL);


    end else begin
    //Separate treatment of each chunk - It still loads all data
    //**************************************************************************//
    //                                                                          //
    //                                                                          //
    //                   C H U N K S        D I F F E R E N T                   //
    //                                                                          //
    //                                                                          //
    //**************************************************************************//


      if prChunkThreadCount > 0 then begin

        //This code does not work well
        (*
        c := 0;
        repeat

          if prStopped then begin
            if prStoppedCount >= 3 then Break;
          end;

          for i := 0 to prChunkThreadCount - 1 do begin
            if prChunkThreads[i] <> nil then begin
              if TQUB_Task_Mac_ChunkThread(prChunkThreads[i]).Ready then begin
                TQUB_Task_Mac_ChunkThread(prChunkThreads[i]).MacChunkIdx := c;
                TQUB_Task_Mac_ChunkThread(prChunkThreads[i]).Resume;
                Inc(c);
              end;
            end;
          end;

          prSignal.WaitFor(100);
          prSignal.Acquire;
          try
            prSignal.ResetEvent;
          finally;
            prSignal.Release;
          end;

        until(c >= prChunks.Count);

        //Now all chunks are done or in progress
        //Let's wait until all are done
        repeat

          if prStopped then begin
            if prStoppedCount >= 3 then Break;
          end;

          ThreadsDone := True;
          for i := 0 to prChunkThreadCount - 1 do begin
            if prChunkThreads[i] <> nil then begin
              if not TQUB_Task_Mac_ChunkThread(prChunkThreads[i]).Ready then begin
                ThreadsDone := False;
                Break;
              end;
            end;
          end;

          prSignal.WaitFor(100);
          prSignal.Acquire;
          try
            prSignal.ResetEvent;
          finally;
            prSignal.Release;
          end;

        until(ThreadsDone);
        *)

      end else begin

        //For each chunk *********************************************************
        for c := 0 to prChunks.Count - 1 do begin

          if prStopped then begin
            if prStoppedCount >= 3 then Break;
          end;

          //if prOptParams.OptMode = MODE_SIM then begin
          //  SimOfChunk(c);
          //end else begin
            LLGradOfChunk(c, DoGrad, NumGrad, TotalLL);
          //end;

        end;

      end;

    end;//identical chunks?

  end;//CalculateLL

  procedure CalculateNumGrads;
  var
    OldPar :Double;
    f, m   :Integer;
  begin

    OldPar := 0;

    try

      if prOptParams.OptNumGradsFast = 1 then begin
        //Left derivative
        TotalLL := 0;
        CalculateLL;
        LLLeft := TotalLL;
      end;

      //Calculate LLLeft and LLRight for each parameter
      for f := 0 to TQUB_Mac_Model(prMacModels[0]).fParCount - 1 do begin

        if prStopped then begin
          if prStoppedCount >= 3 then Break;
        end;

        prStatus := Format('Opt. num. grad. %d/%d', [f + 1, TQUB_Mac_Model(prMacModels[0]).fParCount]);
        Synchronize(prDoOnChange);


        try

          OldPar := TQUB_Mac_Model(prMacModels[0]).fPars[f];

          if prOptParams.OptNumGradsFast = 0 then begin
            //Left derivative
            for m := 0 to prMacModels.Count - 1 do begin
              TQUB_Mac_Model(prMacModels[m]).fPars[f] := OldPar - prOptParams.OptNumGradStep;
              TQUB_Mac_Model(prMacModels[m]).Do_fParsToModel;
            end;
            TotalLL := 0;
            CalculateLL;
            LLLeft := TotalLL;
          end;

          //Right derivative
          for m := 0 to prMacModels.Count - 1 do begin
            TQUB_Mac_Model(prMacModels[m]).fPars[f] := OldPar + prOptParams.OptNumGradStep;
            TQUB_Mac_Model(prMacModels[m]).Do_fParsToModel;
          end;
          TotalLL := 0;
          CalculateLL;
          LLRight := TotalLL;

          //The Weights are included in the LL for the numerical gradients
          if prOptParams.OptNumGradsFast = 0 then begin
            ThisGrad := (LLRight - LLLeft)/(2*prOptParams.OptNumGradStep);
          end else begin
            ThisGrad := (LLRight - LLLeft)/prOptParams.OptNumGradStep;
          end;
          TQUB_Mac_Model(prMacModels[0]).dLLdfPar[f] := {TQUB_Mac_Model(prMacModels[0]).dLLdfPar[f]} + ThisGrad;
          for m := 1 to prMacModels.Count - 1 do begin
            TQUB_Mac_Model(prMacModels[m]).dLLdfPar[f] := 0;
          end;

        finally

          //Restore this parameter
          for m := 0 to prMacModels.Count - 1 do begin
            TQUB_Mac_Model(prMacModels[m]).fPars[f] := OldPar;
          end;

        end;

      end;//f

    finally

      //Restore the old fPars
      for m := 0 to prMacModels.Count - 1 do begin
        TQUB_Mac_Model(prMacModels[m]).Do_fParsToModel;
      end;

    end;

  end;//CalculateNumGrads

begin

  Result               := 0;

  prErrorMsg           := '';
  //****************************************************************************

  SomeConfluent := False;
  SmallRateWarning := False;
  BigRateWarning := False;
  //Reset LL and derivatives w.r.t. q's ****************************************
  for m := 0 to prMacModels.Count - 1 do begin
    if TQUB_Mac_Model(prMacModels[m]).ResetLLandGR then begin
      SomeConfluent := True;
    end;
    if TQUB_Mac_Model(prMacModels[m]).SmallRateWarning then begin
      SmallRateWarning := True;
    end;
    if TQUB_Mac_Model(prMacModels[m]).BigRateWarning then begin
      BigRateWarning := True;
    end;
  end;
  //****************************************************************************


  //Calculate LL
  TotalLL := 0;
  DoGrad  := (CompGrad <> 0) and (prOptParams.OptNumGrads = 0);
  NumGrad := False;
  CalculateLL;
  ThisLL := TotalLL;


  //So far we have the gradients with respect to k's, and Nc, Amp, or Var
  //Now we need to calculate the free pars gradients
  if (prOptParams.OptMode <> MODE_SIM) and (CompGrad <> 0) and (prOptParams.OptNumGrads = 0) then begin
    for m := 0 to prMacModels.Count - 1 do begin
      TQUB_Mac_Model(prMacModels[m]).Do_dLLdfPar;
    end;
  end;


  //Calculate the numerical gradients
  if (CompGrad <> 0) and (prOptParams.OptNumGrads <> 0) and (prStoppedCount < 3) then begin
    DoGrad  := False;
    NumGrad := True;
    CalculateNumGrads;
    NumGrad := False;
  end;//NumGrad

  {
  //Add a scripted penalty?
  if prOptParams.OptScriptPenalty <> 0 then begin
    prCalculateScriptPenalty;
  end;
  }

  ErrorA               := False;
  ////Check if the A matrix was calculated correctly, i.e., there are no negative entries
  //for m := 0 to prMacModels.Count - 1 do begin
  //  if TQUB_Mac_Model(prMacModels[m]).ErrorA then begin
  //    //TO DO: LM. Turn this message off, optionally
  //    ErrorA := True;
  //  end;
  //end;

  if ErrorA then begin
    prErrorMsg := prErrorMsg + 'Error in computing the A matrix. ';
    prWikiTopic := 'Error:Eigenvalues';
  end;
  if SomeConfluent then begin
    prErrorMsg := prErrorMsg + 'Confluent eigen values detected. ';
    prWikiTopic := 'Error:Eigenvalues';
  end;
  if SmallRateWarning then begin
    prErrorMsg := prErrorMsg + 'Very small rate(s) detected. ';
  end;
  if BigRateWarning then begin
    prErrorMsg := prErrorMsg + 'Very big rate(s) detected. ';
  end;


  if prErrorMsg <> '' then begin
    (*
    cLL := 0;
    for m := 0 to prMacModels.Count - 1 do begin
      cLL := cLL + TQUB_Mac_Model(prMacModels[m]).LL;
    end;
    prErrorMsg := prErrorMsg + #13#10 + ' LL = ' + FloatToStr(cLL);
    prErrorMsg := prErrorMsg + #13#10#9'Rates:  '#13#10#9's1  s2   k0         k0.std     k1         k1.std' +
                               #13#10#9'--------------------------------------------------';
    for k := 0 to TQUB_Mac_Model(prMacModels[0]).RateCount - 1 do begin
      try
        prErrorMsg := prErrorMsg + Format(#13#10#9'%d   %d    %-10.6g %-10.6g %-10.6g %-10.6g',
                                         [TQUB_Mac_Model(prMacModels[0]).RateIdx[k].x+1, TQUB_Mac_Model(prMacModels[0]).RateIdx[k].y+1,
                                          //!!! cParsVar
                                          Choice(OptParams.OptAllowNegRates,
                                                     TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0],
                                                 Exp(TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 0])
                                          ),
                                          Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 0]),
                                          TQUB_Mac_Model(prMacModels[0]).cPars[2*k + 1],  Sqrt(TQUB_Mac_Model(prMacModels[0]).cParsVar[2*k + 1])
                                         ]
                                   );
      except
      end;
    end;
    prErrorMsg := prErrorMsg + #13#10#9'--------------------------------------------------';
    *)
    if not prQuietOutput then begin
      prConsoleMsg  := ErrorMsg;
      prConsoleCode := MESSAGE_ERROR;
      Synchronize(prWriteNewConsole);
    end;
  end;

end;

function TQUB_Task_Mac.LLGradOfAllChunks(CompGrad, NumGrad :Boolean; var TotalLL :Double) :Boolean;
var

  MacChunk                     :TQUB_Mac_PChunk;
  FitChunk                     :TQUB_Mac_PChunk;

  MacModel                     :TQUB_Mac_Model;
  Chunk                        :TQUB_Thread_PChunk;
  DataCount                    :Integer;
  IData                        :TPAInteger;
  FData                        :Double;
  cLL, cGR, cSS                :Double;
  c, k, m, t                   :Integer;
  l, s, ec, q                  :Integer;
  ln2PI_05                     :Double;
  Scale                        :Double;
  Off                          :Double;
  dt                           :Double;

  mt, Vt, VtInv                :Double;
  dmt, dVt                     :Double;

  FitThisPoint                 :Boolean;
  ECDwell                      :Integer;
  ECClass                      :Integer;
  ECNextChange                 :Integer;
  PrevT0                       :Double;
  mc                           :Integer;

  ThisVoltage                  :Double;
  VoltageCh                    :Integer;

  i, j                         :Integer;
  pq, pk0, pk1, pC, pV         :Double;
  resample_i                   :Integer;

  VRev                         :Double;
  VLeak                        :Double;
  GLeak                        :Double;

  FitList                      :TPAPoint;
  FitListCount                 :Integer;
  FitListIdx                   :Integer;
  ExcList                      :TPAPoint;
  ExcListCount                 :Integer;
  ExcListIdx                   :Integer;

  //-----
  function CalcGradient :Double;
  begin
    //Rates gradients
    if OptParams.EstUseVariance = 0 then begin
      Result := 2*(FData - mt)*dmt;
    end else begin
      Result := -0.5*VtInv*dVt + (FData - mt)*VtInv*MacModel.AverageCount*dmt + 0.5*Sqr((FData - mt)*VtInv)*dVt*MacModel.AverageCount;
    end;
  end;

  //-----
  function SumGradient :Double;
  var
    c :Integer;
  begin
    Result := 0;

    for c := 0 to prChunks.Count - 1 do begin

      if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin

        FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;

        if OptParams.EstUseVariance = 0 then begin
          Result := Result + 2*(FData - mt)*dmt
        end else begin
          Result := Result - 0.5*VtInv*dVt + (FData - mt)*VtInv*dmt + 0.5*Sqr((FData - mt)*VtInv)*dVt;
        end;

      end;

    end;//c

  end;

  //TO DO: Optimize it to discard memory once used, if possible. LM
  procedure ChunkData(c :Integer);
  begin
    MacChunk  := prChunks[c];
    Chunk     := prChunks[c];
    if MacChunk.Data = nil then begin
      DataCallback(MacChunk);
    end;
    IData     := MacChunk.Data;
    DataCount := TQUB_Thread_PChunk(MacChunk).Sel.Last - TQUB_Thread_PChunk(MacChunk).Sel.First + 1;
    Scale     := 1 / MacChunk.Scaling;
    Off       := MacChunk.Offset;
    dt        := MacChunk.Sampling;
  end;

  //Penalty ....................................................................
  //A hard threshold is not handled well by the optimizer (unequal left and right derivatives)
  function RatePenaltyMin(q, k0, k1, C, V, minq :Double) :Double;
  const
    A = 0.01;
  var
    q0 :Double;
  begin
    q0 := q / (C*exp(k1*V));
    //1st penalty: q > qmin
    Result :=          (ln(minq) - ln(q))  / (1 + exp((ln(q)  - ln(minq))/A));
    //Result :=          ((minq - q)/minq)  / (1 + exp(((q  - minq)/minq)/A));
    //2nd penalty; k0 > qmin
    Result := Result + (ln(minq) - ln(q0)) / (1 + exp((ln(q0) - ln(minq))/A));
  end;

  function DRatePenaltyMin(q, k0, k1, C, V, minq :Double) :Double;
  const
    A = 0.01;
  begin
    //Use a numerical derivative, until figure out the analytical solution
    Result := (RatePenaltyMin(q + 1e-4, k0, k1, C, V, minq) - RatePenaltyMin(q - 1e-4, k0, k1, C, V, minq)) / 2e-4;
  end;

  function RatePenaltyMax(q, k0, k1, C, V, maxq :Double) :Double;
  const
    A = 0.01;
  var
    q0 :Double;
  begin
    q0 := q / (C*exp(k1*V));
    //1st penalty: q < qmax
    Result :=          (ln(q)  - ln(maxq)) / (1 + exp((ln(maxq) -  ln(q))/A));
    //Result :=          ((q  - maxq)/maxq) / (1 + exp(((maxq -  q)/maxq)/A));
    //2nd penalty; k0 < qmin
    Result := Result + (ln(q0) - ln(maxq)) / (1 + exp((ln(maxq) - ln(q0))/A));
  end;

  function DRatePenaltyMax(q, k0, k1, C, V, maxq :Double) :Double;
  const
    A = 0.01;
  begin
    //Use a numerical derivative, until figure out the analytical solution
    Result := (RatePenaltyMax(q + 1e-4, k0, k1, C, V, maxq) - RatePenaltyMax(q - 1e-4, k0, k1, C, V, maxq)) / 2e-4;
  end;

  procedure CalculateRateMinMaxPenalty;
  var
    k, ec :Integer;
  begin

    for k := 0 to MacModel.RateCount - 1 do begin

      i := MacModel.c2qS[2*k + 0].x;
      j := MacModel.c2qS[2*k + 0].y;

      //Cond Equilibrium
      if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

        //Need:

        //q[ec][i, j]
        pq  := MacModel.Cond_Q[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.Cond_ExpCondP[2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.Cond_ExpCondQ[2*k + 1];

        cSS := 0;
        if OptParams.EstUseMinRate then begin
          cSS := cSS + RatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cSS := cSS + RatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        cLL := - cSS;
        MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
        MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;

      end;//Cond eq ?

      for ec := 0 to MacModel.ExpCondCount - 1 do begin

        //Need:

        //q[ec][i, j]
        pq  := TPMDouble(MacModel.Q[ec])[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.ExpCondP[ec][2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.ExpCondQ[ec][2*k + 1];

        cSS := 0;
        if OptParams.EstUseMinRate then begin
          cSS := css + RatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cSS := css + RatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        cLL := - cSS;
        MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
        MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;

      end;//ec

    end;//k

  end;//CalculateRateMinMaxPenalty

  procedure CalculateRateMinMaxPenaltyGradients;
  var
    k, ec :Integer;
  begin

    for k := 0 to MacModel.RateCount - 1 do begin

      i := MacModel.c2qS[2*k + 0].x;
      j := MacModel.c2qS[2*k + 0].y;

      //Cond Equilibrium
      if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

        //Need:

        //q[ec][i, j]
        pq  := MacModel.Cond_Q[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.Cond_ExpCondP[2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.Cond_ExpCondQ[2*k + 1];

        cGR := 0;
        if OptParams.EstUseMinRate then begin
          cGR := cGR - DRatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cGR := cGR - DRatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        MacModel.dLLdCond_q[k] := MacModel.dLLdCond_q[k] - MacModel.FitWeight*cGR;

      end;//Cond eq ?

      for ec := 0 to MacModel.ExpCondCount - 1 do begin

        //Need:

        //q[ec][i, j]
        pq  := TPMDouble(MacModel.Q[ec])[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.ExpCondP[ec][2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.ExpCondQ[ec][2*k + 1];

        cGR := 0;
        if OptParams.EstUseMinRate then begin
          cGR := cGR - DRatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cGR := cGR - DRatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        MacModel.dLLdq[ec, k] := MacModel.dLLdq[ec, k] - MacModel.FitWeight*cGR;

      end;//ec

    end;//k

  end;//CalculateRateMinMaxPenaltyGradients

begin

  Result               := False;

  VoltageCh            := -1;

  ln2PI_05             := 0.5*ln(2*PI);
  VtInv                := 1;

  ThisVoltage          := prOptParams.OptVRev;


  try


    //For each MacModel
    //TO DO: LM. Are there ever more than one?!
    for m := 0 to prMacModels.Count - 1 do begin


      if prStopped then begin
        if prStoppedCount >= 3 then Break;
      end;


      MacModel     := TQUB_Mac_Model(prMacModels[m]);


      VRev         := MacModel.VRev;
      VLeak        := MacModel.VLeak;
      GLeak        := MacModel.GLeak;


      FitList      := nil;
      FitListCount := -1;
      ExcList      := nil;
      ExcListCount := -1;


      //Load data
      for c := prChunks.Count - 1 downto 0 do begin
        if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin
          ChunkData(c);
          if FitListCount = -1 then begin
            FitListCount := MacChunk.FitListCount;
            FitList      := MacChunk.FitList;
            FitListIdx   := 0;
          end;
          if ExcListCount = -1 then begin
            ExcListCount := MacChunk.ExcListCount;
            ExcList      := MacChunk.ExcList;
            ExcListIdx   := 0;
          end;
        end;
      end;


      //Reset the previous start time
      PrevT0      := 0;


      //Initial state probabilities
      MacModel.Do_P0;
      if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
        //Derivatives of the initial state probabilities
        MacModel.Do_dP0_dq;           //Check for Eq and CondEq !!!
        MacModel.Do_dP0_dCond_q;      //        - ,, -
      end;

      //Initialize experimental condition
      ECDwell      := 0;
      ECClass      := MacChunk.ExpCondIdl[ECDwell].Cls;
      ECNextChange := MacChunk.ExpCondIdl[ECDwell].N;

      if prOptParams.OptUseVRev <> 0 then begin
        VoltageCh := -1;
        for ec := 0 to MacChunk.MacModel.ExpCondParams[ECClass].Count - 1 do begin
          if MacChunk.MacModel.ExpCondParams[ECClass][ec].Name = 'Voltage' then begin
            VoltageCh   := ec;
            Break;
          end;
        end;
        if VoltageCh > -1 then begin
          ThisVoltage := MacChunk.MacModel.ExpCondParams[ECClass][VoltageCh].Value;
        end;
      end;


      //Chris, the segmentation does not appear to be right - last segment is one point short


      //For each DataPoint
      for t := 0 to DataCount - 1 do begin

        if prStopped then begin
          if prStoppedCount >= 3 then Break;
        end;

        //Change ECClass?
        if (t >= ECNextChange) and ((ECDwell + 1) < MacChunk.ExpCondIdlCount){Apparently, the last dwell is one point short} then begin
        //something fishy in this condition !!!

          try

            {
            Assumption: the stimuli change at this very point
            Recalculate P0 and dP0/dQ as
              P0[this_seg] = P0[prev_seg] * A[prev_seg,dur]
              dP0[this_seg] / dQ[prev_seg] = (dP0[prev_seg] / dQ[prev_seg]) * A[prev_seg_dur] +
                                             P0[prev_seg] * (dA[prev_seg,dur] / dQ[prev_seg])
            }

            //A matrix
            MacModel.Do_At((t-PrevT0)*dt, ECClass);
            if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
              //dA[ECClass] / dq[ec]
              MacModel.Do_dAt_dq((t-PrevT0)*dt, ECClass);
              //dA[ECClass] / dCond_q[ec]
              //MacModel.Do_dAt_dCond_q((t-PrevT0)*dt, ECClass);// - not necessary
            end;

            //Current state, later to be copied into initial state
            MacModel.Do_Pt((t-PrevT0)*dt, ECClass);
            if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
              MacModel.Do_dPt_dq(ECClass);
              MacModel.Do_dPt_dCond_q(ECClass);
            end;

            //Copy Pt into P0, and the derivatives
            MacModel.PtToP0;
            if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
              MacModel.dPtdqTodP0dq;
              MacModel.dPtdCond_qTodP0dCond_q;
            end;


          finally

            Inc(ECDwell);
            ECClass      := MacChunk.ExpCondIdl[ECDwell].Cls;
            ECNextChange := ECNextChange + MacChunk.ExpCondIdl[ECDwell].N;
            PrevT0       := t;

            if prOptParams.OptUseVRev <> 0 then begin
              if VoltageCh > -1 then begin
                ThisVoltage := MacChunk.MacModel.ExpCondParams[ECClass][VoltageCh].Value;
              end;
            end;

          end;

        end;


        //Fit this point ?
        FitThisPoint := True;
        if (prOptParams.OptSkip = 1) then begin
          if (t*dt) <= (prOptParams.OptSkipTime*1e-3) then begin
            FitThisPoint := False;
          end;
        end;
        if (FitThisPoint) and (FitListCount > 0) then begin
          FitThisPoint := False;
          for l := FitListIdx to FitListCount - 1 do begin // This may have been out of synch ? JB.
            if (t >= FitList[l].x) and (t <= FitList[l].y) then begin
              FitThisPoint := True;
              FitListIdx := l;
              Break;
            end;
          end;
        end;
        if (FitThisPoint) and (ExcListCount > 0) then begin
          for l := ExcListIdx to ExcListCount - 1 do begin // This may have been out of synch ? JB.
            if (t >= ExcList[l].x) and (t <= ExcList[l].y) then begin
              FitThisPoint := False;
              ExcListIdx := l;
              Break;
            end;
          end;
        end;


        //**********************************************************************//
        //                                                                      //
        //                             Fit this point !                         //
        //                                                                      //
        //**********************************************************************//
        if FitThisPoint then begin


          if not NumGrad then Inc(MacModel.PointCount);//, MacModel.AverageCount);

          //A matrix
          MacModel.Do_At((t-PrevT0)*dt, ECClass);
          if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
            MacModel.Do_dAt_dq((t-PrevT0)*dt, ECClass);
          end;

          //Current state
          MacModel.Do_Pt((t-PrevT0)*dt, ECClass);
          if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
            MacModel.Do_dPt_dq(ECClass);
            MacModel.Do_dPt_dCond_q(ECClass);
          end;

          //I Mean

          //Correct for I = g*(V-Vrev)
          if prOptParams.OptUseVRev <> 0 then begin
            MacModel.Do_Mean_t(mt, 1e-3*(ThisVoltage - Vrev), GLeak*1e-3*(ThisVoltage - VLeak));
          end else begin
            MacModel.Do_Mean_t(mt);
          end;

          if OptParams.EstUseVariance = 1 then begin

            //State Variance
            MacModel.Do_VarPt;

            //State Variance derivatives
            if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
              MacModel.Do_dVarPt_dq;
              if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin
                MacModel.Do_dVarPt_dCond_q;
              end;
            end;

            //I Variance
            if prOptParams.OptUseVRev <> 0 then begin
              MacModel.Do_Var_t(Vt, 1e-3*(ThisVoltage - Vrev));
            end else begin
              MacModel.Do_Var_t(Vt);
            end;
            if Vt <= 0 then Vt := 1e-30;
            VtInv := 1 / Vt;

          end;


          //Min ChannelCount
          //TO DO: LM. This won't work if I = f(V).
          if not NumGrad then begin
            if TQUB_Mac_Model(TQUB_Mac_PChunk(prChunks[0]).MacModel).Iterations = 0 then begin
              for c := 0 to prChunks.Count - 1 do begin
                if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin
                  FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;
                  mc := Round(Abs(FData - MacModel.bAmp) / Abs(MacModel.MinAmp));
                  if mc > prMaxChannel then begin
                    prMaxChannel := mc;
                  end;
                end;
              end;
            end;
          end;


          //Log-Likelihood .....................................................
          for c := 0 to prChunks.Count - 1 do begin
            if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin

              //For each Chunk of this model
              //Inc(MacModel.PointCount);//, MacModel.AverageCount);

              FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;

              cSS := Sqr(FData - mt);
              if OptParams.EstUseVariance = 0 then begin
                cLL := - cSS;
              end else begin
                cLL := - ln2PI_05 - 0.5*ln(Vt) - 0.5*cSS*VtInv;
              end;

              if not NumGrad then begin
                MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
                MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;
              end;

              TotalLL := TotalLL - cLL;

            end;
          end;


          if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
            //Rates Gradients ....................................................
            for k := 0 to MacModel.RateCount - 1 do begin

              for ec := 0 to MacModel.ExpCondCount - 1 do begin

                MacModel.Do_dMean_t_dq(dmt, k, ec);
                //LM. New: testing
                if prOptParams.OptUseVRev <> 0 then begin
                  dmt := dmt * (ThisVoltage - VRev) * 1e-3; // bAmp not part of deriv.
                end;
                if OptParams.EstUseVariance = 1 then begin
                  MacModel.Do_dVar_t_dq(dVt, k, ec);
                  if prOptParams.OptUseVRev <> 0 then begin
                    dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
                  end;
                end;

                for c := 0 to prChunks.Count - 1 do begin
                  if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin

                    FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;

                    MacModel.dLLdq[ec, k] := MacModel.dLLdq[ec, k] - MacModel.FitWeight*CalcGradient;

                  end;
                end;//c

              end;//ec

              //Cond Equilibrium
              if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

                MacModel.Do_dMean_t_dCond_q(dmt, k);
                //LM. New: testing
                if prOptParams.OptUseVRev <> 0 then begin
                  dmt := dmt * (ThisVoltage - VRev) * 1e-3;
                end;
                if OptParams.EstUseVariance = 1 then begin
                  MacModel.Do_dVar_t_dCond_q(dVt, k);
                  if prOptParams.OptUseVRev <> 0 then begin
                    dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
                  end;
                end;

                for c := 0 to prChunks.Count - 1 do begin
                  if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin

                    FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;

                    MacModel.dLLdCond_q[k] := MacModel.dLLdCond_q[k] - MacModel.FitWeight*CalcGradient;

                  end;
                end;//c

              end;//Cond eq ?

            end;//k


            //ChannelCount Gradient ..............................................
            if OptParams.EstChannelCount <> 0 then begin
              if not MacModel.CCFixed then begin

                MacModel.Do_dMean_t_dN(dmt);
                //LM. New: testing
                if prOptParams.OptUseVRev <> 0 then begin
                  dmt := dmt * (ThisVoltage - VRev) * 1e-3;
                end;
                if OptParams.EstUseVariance = 1 then begin
                  MacModel.Do_dVar_t_dN(dVt);
                  if prOptParams.OptUseVRev <> 0 then begin
                    dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
                  end;
                end;

                for c := 0 to prChunks.Count - 1 do begin
                  if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin

                    FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;

                    MacModel.dLLdN := MacModel.dLLdN - MacModel.FitWeight*CalcGradient;

                  end;
                end;

              end;
            end;

            //AmpOpen Gradient ...................................................
            if OptParams.EstAmpOpen <> 0 then begin

              MacModel.Do_dMean_t_dA(dmt);
              //LM. New: testing
              if prOptParams.OptUseVRev <> 0 then begin
                dmt := dmt * (ThisVoltage - VRev) * 1e-3;
              end;
              if OptParams.EstUseVariance = 1 then begin
                MacModel.Do_dVar_t_dA(dVt);
                if prOptParams.OptUseVRev <> 0 then begin
                  dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
                end;
              end;

              for c := 0 to prChunks.Count - 1 do begin
                if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin

                  FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;

                  MacModel.dLLdA := MacModel.dLLdA - MacModel.FitWeight*CalcGradient;

                end;
              end;

            end;

            //VarOpen Gradient .................................................
            if OptParams.EstVarOpen <> 0 then begin

              MacModel.Do_dMean_t_dV(dmt);
              //LM. New: testing
              if prOptParams.OptUseVRev <> 0 then begin
                dmt := dmt * (ThisVoltage - VRev) * 1e-3;
              end;
              if OptParams.EstUseVariance = 1 then begin
                MacModel.Do_dVar_t_dV(dVt);
              end;

              for c := 0 to prChunks.Count - 1 do begin
                if TQUB_Mac_PChunk(prChunks[c]).MacModel = MacModel then begin

                  FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale + Off;

                  MacModel.dLLdV := MacModel.dLLdV - MacModel.FitWeight*CalcGradient;

                end;
              end;

            end;

          end;//Grads

        end;//Fit this point!


        if not NumGrad then begin
          //Show fitting curve ?
          if (OptParams.OptShowFit <> 0) then begin

            for c := 0 to prChunks.Count - 1 do begin

              FitChunk := TQUB_Mac_PChunk(prChunks[c]);
              if (FitChunk.MacModel = MacModel) then begin

                //************************************************************
                if FitThisPoint then begin
                  //LM. New: testing
                  //if prOptParams.OptUseVRev <> 0 then begin
                  //  mt := mt*ThisVoltage/prOptParams.OptVRev;
                  //end;
                  if          OptParams.OptShowFitWhat = SHOWFIT_AVG then begin
                    FitChunk.FitCurve[t] := Round((mt - FitChunk.Offset) * FitChunk.Scaling);
                  end else if OptParams.OptShowFitWhat = SHOWFIT_STD then begin
                    FitChunk.FitCurve[t] := Round((Sqrt(Vt) - FitChunk.Offset) * FitChunk.Scaling);
                  end else if OptParams.OptShowFitWhat = SHOWFIT_RES then begin
                    FitChunk.FitCurve[t] := Round((mt - FitChunk.Offset) * FitChunk.Scaling - TQUB_Mac_PChunk(prChunks[c]).Data[t]);
                  end;
                end else begin
                  FitChunk.FitCurve[t] := TQUB_Mac_PChunk(prChunks[c]).Data[t];
                end;
                //****************************************************************

              end;//MacModel

            end;//c

          end;//ShowFit
        end;


      end;//of t



      //********************************************************************//
      //                                                                    //
      //                   Penalty for Rates' Upper/LowerBound              //
      //                                                                    //
      //********************************************************************//
      if (OptParams.EstUseMinRate) or (OptParams.EstUseMaxRate) then begin

        //Log-likelihood .....................................................
        CalculateRateMinMaxPenalty;

        if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
          //Rates Gradients ....................................................
          CalculateRateMinMaxPenaltyGradients
        end;

      end;//Penalty ..........................................................



    end;//MacModel


  finally

  end;


end;//LLGradOfAllChunks

function TQUB_Task_Mac.LLGradOfChunk(c :Integer; CompGrad, NumGrad :Boolean; var TotalLL :Double) :Boolean;
var
  MacChunk                     :TQUB_Mac_PChunk;
  FitChunk                     :TQUB_Mac_PChunk;
  MacModel                     :TQUB_Mac_Model;
  Chunk                        :TQUB_Thread_PChunk;
  DataCount                    :Integer;
  IData                        :TPAInteger;
  TData                        :TPASingle;
  FData                        :Double;
  cLL, cGR, cSS                :Double;
  k, m, t                      :Integer;
  l, ec                        :Integer;
  ln2PI_05                     :Double;
  Scale                        :Double;
  Off                          :Double;
  dt                           :Double;
  mt, Vt, VtInv                :Double;
  dmt, dVt                     :Double;

  ECDwell                      :Integer;
  ECClass                      :Integer;
  ECNextChange                 :Integer;

  PrevT0                       :Integer;
  mc                           :Integer;

  ThisVoltage                  :Double;
  VoltageCh                    :Integer;
  i, j                         :Integer;
  pq, pk0, pk1, pC, pV         :Double;
  resample_i                   :Integer;

  VRev                         :Double;
  VLeak                        :Double;
  GLeak                        :Double;

  ChunkPercent                 :Integer;
  PrevPercent                  :Integer;
  Percent                      :Integer;

  PrevTime0                    :Single;
  Time                         :Single;
  DTime                        :Single;

  function CalcGradient :Double;
  begin
    //Rates gradients
    if (OptParams.EstUseVariance = 0) or (MacModel.HasFitP) then begin
      Result := 2*(FData - mt)*dmt;
    end else begin
      Result := -0.5*VtInv*dVt + (FData - mt)*VtInv*MacModel.AverageCount*dmt + 0.5*Sqr((FData - mt)*VtInv)*dVt*MacModel.AverageCount;
    end;
  end;

  //Penalty ....................................................................
  //A hard threshold is not handled well by the optimizer (unequal left and right derivatives)
  function RatePenaltyMin(q, k0, k1, C, V, minq :Double) :Double;
  const
    A = 0.01;
  var
    q0 :Double;
  begin
    q0 := q / (C*exp(k1*V));
    //1st penalty: q > qmin
    Result :=          (ln(minq) - ln(q))  / (1 + exp((ln(q)  - ln(minq))/A));
    //Result :=          ((minq - q)/minq)  / (1 + exp(((q  - minq)/minq)/A));
    //2nd penalty; k0 > qmin
    Result := Result + (ln(minq) - ln(q0)) / (1 + exp((ln(q0) - ln(minq))/A));
  end;

  function DRatePenaltyMin(q, k0, k1, C, V, minq :Double) :Double;
  const
    A = 0.01;
  begin
    //Use a numerical derivative, until figure out the analytical solution
    Result := (RatePenaltyMin(q + 1e-4, k0, k1, C, V, minq) - RatePenaltyMin(q - 1e-4, k0, k1, C, V, minq)) / 2e-4;
  end;

  function RatePenaltyMax(q, k0, k1, C, V, maxq :Double) :Double;
  const
    A = 0.01;
  var
    q0 :Double;
  begin
    q0 := q / (C*exp(k1*V));
    //1st penalty: q < qmax
    Result :=          (ln(q)  - ln(maxq)) / (1 + exp((ln(maxq) -  ln(q))/A));
    //Result :=          ((q  - maxq)/maxq) / (1 + exp(((maxq -  q)/maxq)/A));
    //2nd penalty; k0 < qmin
    Result := Result + (ln(q0) - ln(maxq)) / (1 + exp((ln(maxq) - ln(q0))/A));
  end;

  function DRatePenaltyMax(q, k0, k1, C, V, maxq :Double) :Double;
  const
    A = 0.01;
  begin
    //Use a numerical derivative, until figure out the analytical solution
    Result := (RatePenaltyMax(q + 1e-4, k0, k1, C, V, maxq) - RatePenaltyMax(q - 1e-4, k0, k1, C, V, maxq)) / 2e-4;
  end;

  procedure CalculateRateMinMaxPenalty;
  var
    k, ec :Integer;
  begin

    for k := 0 to MacModel.RateCount - 1 do begin

      i := MacModel.c2qS[2*k + 0].x;
      j := MacModel.c2qS[2*k + 0].y;

      //Cond Equilibrium
      if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

        //Need:

        //q[ec][i, j]
        pq  := MacModel.Cond_Q[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.Cond_ExpCondP[2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.Cond_ExpCondQ[2*k + 1];

        cSS := 0;
        if OptParams.EstUseMinRate then begin
          cSS := cSS + RatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cSS := cSS + RatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        cLL := - cSS;
        MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
        MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;
        MacChunk.LL := MacChunk.LL - MacModel.FitWeight*cLL;

      end;//Cond eq ?

      for ec := 0 to MacModel.ExpCondCount - 1 do begin

        //Need:

        //q[ec][i, j]
        pq  := TPMDouble(MacModel.Q[ec])[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.ExpCondP[ec][2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.ExpCondQ[ec][2*k + 1];

        cSS := 0;
        if OptParams.EstUseMinRate then begin
          cSS := css + RatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cSS := css + RatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        cLL := - cSS;
        MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
        MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;
        MacChunk.LL := MacChunk.LL - MacModel.FitWeight*cLL;

      end;//ec

    end;//k

  end;//CalculateRateMinMaxPenalty

  procedure CalculateRateMinMaxPenaltyGradients;
  var
    k, ec :Integer;
  begin

    for k := 0 to MacModel.RateCount - 1 do begin

      i := MacModel.c2qS[2*k + 0].x;
      j := MacModel.c2qS[2*k + 0].y;

      //Cond Equilibrium
      if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

        //Need:

        //q[ec][i, j]
        pq  := MacModel.Cond_Q[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.Cond_ExpCondP[2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.Cond_ExpCondQ[2*k + 1];

        cGR := 0;
        if OptParams.EstUseMinRate then begin
          cGR := cGR - DRatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cGR := cGR - DRatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        MacModel.dLLdCond_q[k] := MacModel.dLLdCond_q[k] - MacModel.FitWeight*cGR;

      end;//Cond eq ?

      for ec := 0 to MacModel.ExpCondCount - 1 do begin

        //Need:

        //q[ec][i, j]
        pq  := TPMDouble(MacModel.Q[ec])[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.ExpCondP[ec][2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.ExpCondQ[ec][2*k + 1];

        cGR := 0;
        if OptParams.EstUseMinRate then begin
          cGR := cGR - DRatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cGR := cGR - DRatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        MacModel.dLLdq[ec, k] := MacModel.dLLdq[ec, k] - MacModel.FitWeight*cGR;

      end;//ec

    end;//k

  end;//CalculateRateMinMaxPenaltyGradients


begin

  Result                       := False;

  TData                        := nil;

  //This was not calculated here ***********************************************
  VoltageCh                    := -1;
  ln2PI_05                     := 0.5*ln(2*PI);
  VtInv                        := 1;
  //****************************************************************************

  //ChunkData
  MacChunk                     := prChunks[c];
  Chunk                        := prChunks[c];
  if MacChunk.Data = nil then begin
    DataCallback(MacChunk);
  end;
  IData                        := MacChunk.Data;
  if MacChunk.HasTime then begin
    TData                      := MacChunk.TimeData;
  end;
  DataCount                    := TQUB_Thread_PChunk(MacChunk).Sel.Last - TQUB_Thread_PChunk(MacChunk).Sel.First + 1;
  Scale                        := 1 / MacChunk.Scaling;
  Off                          := MacChunk.Offset;
  dt                           := MacChunk.Sampling;

  if not NumGrad then begin
    Percent := Round((TQUB_Mac_Model(prMacModels[0]).Iterations + c/prChunks.Count) / prOptParams.OptMaxIter * 100);
    ChunkPercent := 0;
    PrevPercent  := 0;
    prStatus := Format('Opt. chunk %d/%d: %d%%', [c + 1, prChunks.Count, ChunkPercent]);
    prUpdateStatus(Percent, prStTime);
  end;
  
  try


    //Reset the previous start time
    PrevT0      := 0;
    PrevTime0   := 0;
    Time        := 0;
    if MacChunk.HasTime then begin
      PrevTime0 := TData[PrevT0];
      Time      := PrevTime0;
    end;


    //Choose the right MacModel
    MacModel    := TQUB_Mac_Model(MacChunk.MacModel);
    if not NumGrad then MacChunk.LL := 0;

    VRev        := MacModel.VRev;
    VLeak       := MacModel.VLeak;
    GLeak       := MacModel.GLeak;

    VoltageCh   := -1;

    //State probabilities
    MacModel.Do_P0;
    if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
      //Derivatives of the initial state probabilities
      MacModel.Do_dP0_dq;           //Check for Eq and CondEq !!!
      MacModel.Do_dP0_dCond_q;      //        - ,, -
    end;

    //Initialize experimental condition
    ECDwell      := 0;
    ECClass      := MacChunk.ExpCondIdl[ECDwell].Cls;
    ECNextChange := MacChunk.ExpCondIdl[ECDwell].N;


    ThisVoltage  := 0;
    if prOptParams.OptUseVRev <> 0 then begin
      VoltageCh := -1;
      for ec := 0 to MacChunk.MacModel.ExpCondParams[ECClass].Count - 1 do begin
        if MacChunk.MacModel.ExpCondParams[ECClass][ec].Name = 'Voltage' then begin
          VoltageCh   := ec;
          Break;
        end;
      end;
      if VoltageCh > -1 then begin
        ThisVoltage := MacChunk.MacModel.ExpCondParams[ECClass][VoltageCh].Value;
      end;
    end;


    //Chris, the segmentation does not appear to be right - last segment is one point short

    resample_i   := 0;

    //For each DataPoint
    for t := 0 to DataCount - 1 do begin

      if prStopped then begin
        if prStoppedCount >= 3 then Break;
      end;

      if prSkipSlowChange then begin
        if t <> MacChunk.ResampleSamples[resample_i] then begin
          Continue;
        end;
      end;

      //New - for data with TimeCh
      if MacChunk.HasTime then begin
        Time := TData[t];
      end;

      //Change ECClass?
      if (t >= ECNextChange) and ((ECDwell + 1) < MacChunk.ExpCondIdlCount){Apparently, the last dwell is one point short} then begin
      //something fishy in this condition !!!

        try

          {
          Assumption: the stimuli change at this very point
          Recalculate P0 and dP0/dQ as
            P0[this_seg] = P0[prev_seg] * A[prev_seg,dur]
            dP0[this_seg] / dQ[prev_seg] = (dP0[prev_seg] / dQ[prev_seg]) * A[prev_seg_dur] +
                                           P0[prev_seg] * (dA[prev_seg,dur] / dQ[prev_seg])
          }

          //New - for data with time channel - uneven samples etc;
          if MacChunk.HasTime then begin
            DTime := Time - PrevTime0;
          end else begin
            DTime := (t-PrevT0)*dt;
          end;

          //A matrix
          MacModel.Do_At(DTime, ECClass);//All matrices or just ECClass ?!
          if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
            //dA[ECClass] / dq[ec]
            MacModel.Do_dAt_dq(DTime, ECClass);
            //MacModel.Do_dAt_dCond_q(DTime, ECClass);
          end;

          //Current state, later to be copied into initial state
          MacModel.Do_Pt(DTime, ECClass);
          if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
            MacModel.Do_dPt_dq(ECClass);
            MacModel.Do_dPt_dCond_q(ECClass);
          end;

          //Copy Pt into P0, and the derivatives
          MacModel.PtToP0;
          if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
            MacModel.dPtdqTodP0dq;
            MacModel.dPtdCond_qTodP0dCond_q;
            //MacModel.AMatrixAndCurrent(DTime,ECClass);
            //MacModel.CopyPtToP0();
          end;

        finally

          Inc(ECDwell);
          ECClass      := MacChunk.ExpCondIdl[ECDwell].Cls;
          ECNextChange := ECNextChange + MacChunk.ExpCondIdl[ECDwell].N;
          PrevT0       := t;
          PrevTime0    := Time;

          if prOptParams.OptUseVRev <> 0 then begin
            if VoltageCh > -1 then begin
              ThisVoltage := MacChunk.MacModel.ExpCondParams[ECClass][VoltageCh].Value;
            end;
          end;

        end;

      end;


      //**********************************************************************//
      //                                                                      //
      //                             Fit this point !                         //
      //                                                                      //
      //**********************************************************************//
      if MacChunk.FitPoint[t] then begin

        if not NumGrad then begin
          ChunkPercent := Round(t / DataCount * 100);
          if (ChunkPercent - PrevPercent) >= 10 then begin
            prStatus := Format('Opt. chunk %d/%d: %d%%', [c+1, prChunks.Count, ChunkPercent]);
            Synchronize(prDoOnChange);
            PrevPercent := ChunkPercent;
          end;
        end;

        //Testing. LM
        if not NumGrad then Inc(MacModel.PointCount);//, MacModel.AverageCount);

        //New - for data with time channel - uneven samples etc;
        if MacChunk.HasTime then begin
          DTime := Time - PrevTime0;
        end else begin
          DTime := (t-PrevT0)*dt;
        end;

        //A matrix
        MacModel.Do_At(DTime, ECClass);
        if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
          MacModel.Do_dAt_dq(DTime, ECClass);
        end;

        //Current state
        MacModel.Do_Pt(DTime, ECClass);
        if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
          MacModel.Do_dPt_dq(ECClass);
          MacModel.Do_dPt_dCond_q(ECClass);
        end;

        //I Mean

        //Correct for I = g*(V-Vrev)
        if (prOptParams.OptUseVRev <> 0) and (not MacModel.HasFitP) then begin
          MacModel.Do_Mean_t(mt, 1e-3*(ThisVoltage - VRev), 1e-3*GLeak*(ThisVoltage - VLeak));
        end else begin
          MacModel.Do_Mean_t(mt);
        end;

        if (OptParams.EstUseVariance <> 0) and (not MacModel.HasFitP) then begin
          //State Variance
          MacModel.Do_VarPt;

          if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
            //State Variance derivatives
            MacModel.Do_dVarPt_dq;
            if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin
              MacModel.Do_dVarPt_dCond_q;
            end;
          end;

          //I Variance
          if prOptParams.OptUseVRev <> 0 then begin
            MacModel.Do_Var_t(Vt, 1e-3*(ThisVoltage - Vrev));
          end else begin
            MacModel.Do_Var_t(Vt);
          end;
          if Vt <= 0 then Vt := 1e-30;
          VtInv := 1 / Vt;
        end;


        if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
          //Min ChannelCount
          //TO DO: LM. This won't work if I = f(V).
          if TQUB_Mac_Model(prMacModels[0]).Iterations = 0 then begin
            if MacModel.HasFitP then begin
              prMaxChannel := 1;
            end else begin
              //FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale;
              FData := MacChunk.Data[t]*Scale + Off;
              mc := Round(Abs(FData - MacModel.bAmp) / Abs(MacModel.MinAmp));
              if mc > prMaxChannel then begin
                prMaxChannel := mc;
              end;
            end;
          end;
        end;

        //Log-Likelihood .....................................................
        //FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale;
        FData := MacChunk.Data[t]*Scale + Off;

        cSS := Sqr(FData - mt);
        if OptParams.EstUseVariance = 0 then begin
          cLL := - cSS;
        end else begin
          cLL := - ln2PI_05 - 0.5*ln(Vt) - 0.5*cSS*VtInv;
        end;

        if not NumGrad then begin
          MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
          MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;
          MacChunk.LL := MacChunk.LL - MacModel.FitWeight*cLL;
        end;

        TotalLL := TotalLL - MacModel.FitWeight*cLL;

        if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin

          //Rates Gradients ....................................................
          for k := 0 to MacModel.RateCount - 1 do begin

            for ec := 0 to MacModel.ExpCondCount - 1 do begin

              MacModel.Do_dMean_t_dq(dmt, k, ec);
              //LM. New: testing
              if (prOptParams.OptUseVRev <> 0) and (not MacModel.HasFitP) then begin
                dmt := dmt * (ThisVoltage - VRev) * 1e-3;
              end;
              if (OptParams.EstUseVariance <> 0) and (not MacModel.HasFitP) then begin
                MacModel.Do_dVar_t_dq(dVt, k, ec);
                if prOptParams.OptUseVRev <> 0 then begin
                  dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
                end;
              end;

              //FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale;
              FData := MacChunk.Data[t]*Scale + Off;

              MacModel.dLLdq[ec, k] := MacModel.dLLdq[ec, k] - MacModel.FitWeight*CalcGradient;

            end;//ec

            //Cond Equilibrium
            if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

              MacModel.Do_dMean_t_dCond_q(dmt, k);
              //LM. New: testing
              if (prOptParams.OptUseVRev <> 0) and (not MacModel.HasFitP) then begin
                dmt := dmt * (ThisVoltage - VRev) * 1e-3;
              end;
              if (OptParams.EstUseVariance <> 0) and (not MacModel.HasFitP) then begin
                MacModel.Do_dVar_t_dCond_q(dVt, k);
                if prOptParams.OptUseVRev <> 0 then begin
                  dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
                end;
              end;

              //FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale;
              FData := MacChunk.Data[t]*Scale + Off;

              MacModel.dLLdCond_q[k] := MacModel.dLLdCond_q[k] - MacModel.FitWeight*CalcGradient;

            end;//Cond eq ?

          end;//k

          //ChannelCount Gradient ............................................
          if (OptParams.EstChannelCount <> 0) and (not MacModel.HasFitP) then begin
            if not MacModel.CCFixed then begin

              MacModel.Do_dMean_t_dN(dmt);
              //LM. New: testing
              if prOptParams.OptUseVRev <> 0 then begin
                dmt := dmt * (ThisVoltage - VRev) * 1e-3;
              end;

              if OptParams.EstUseVariance = 1 then begin
                MacModel.Do_dVar_t_dN(dVt);
                if prOptParams.OptUseVRev <> 0 then begin
                  dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
                end;
              end;

              //FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale;
              FData := MacChunk.Data[t]*Scale + Off;

              MacModel.dLLdN := MacModel.dLLdN - MacModel.FitWeight*CalcGradient;

            end;
          end;

          //AmpOpen Gradient .................................................
          if (OptParams.EstAmpOpen <> 0) and (not MacModel.HasFitP) then begin

            MacModel.Do_dMean_t_dA(dmt);
            //LM. New: testing
            if OptParams.EstUseVariance = 1 then begin
              MacModel.Do_dVar_t_dA(dVt);
              if prOptParams.OptUseVRev <> 0 then begin
                dVt := dVt * (ThisVoltage - VRev) * 1e-3; // bVar not part of deriv.
              end;
            end;

            //FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale;
            FData := MacChunk.Data[t]*Scale + Off;

            MacModel.dLLdA := MacModel.dLLdA - MacModel.FitWeight*CalcGradient;

          end;

          //VarOpen Gradient .................................................
          if (OptParams.EstVarOpen <> 0) and (not MacModel.HasFitP) then begin

            MacModel.Do_dMean_t_dV(dmt);
            //LM. New: testing
            if prOptParams.OptUseVRev <> 0 then begin
              dmt := dmt * (ThisVoltage - VRev) * 1e-3;
            end;
            if OptParams.EstUseVariance = 1 then begin
              MacModel.Do_dVar_t_dV(dVt);
            end;

            //FData := TQUB_Mac_PChunk(prChunks[c]).Data[t]*Scale;
            FData := MacChunk.Data[t]*Scale + Off;

            MacModel.dLLdV := MacModel.dLLdV - MacModel.FitWeight*CalcGradient;

          end;

        end;//not SIM

      end;//Fit this point!


      if not NumGrad then begin
        //Show fitting curve ?
        if (OptParams.OptShowFit <> 0) then begin
          //FitChunk := TQUB_Mac_PChunk(prChunks[c]);
          FitChunk := MacChunk;
          //************************************************************
          //if FitThisPoint then begin
          if MacChunk.FitPoint[t] then begin
            if          OptParams.OptShowFitWhat = SHOWFIT_AVG then begin
              FitChunk.FitCurve[t] := Round((mt - FitChunk.Offset) * FitChunk.Scaling);
            end else if OptParams.OptShowFitWhat = SHOWFIT_STD then begin
              FitChunk.FitCurve[t] := Round((Sqrt(Vt) - FitChunk.Offset) * FitChunk.Scaling);
            end else if OptParams.OptShowFitWhat = SHOWFIT_RES then begin
              //FitChunk.FitCurve[t] := Round(mt * FitChunk.Scaling - TQUB_Mac_PChunk(prChunks[c]).Data[t]);
              FitChunk.FitCurve[t] := Round((mt - FitChunk.Offset) * FitChunk.Scaling - MacChunk.Data[t]);
            end;
          end else begin
            //FitChunk.FitCurve[t] := TQUB_Mac_PChunk(prChunks[c]).Data[t];
            FitChunk.FitCurve[t] := MacChunk.Data[t];
          end;
          //****************************************************************

        end;//ShowFit
      end;//not NumGrad

      Inc(resample_i);

    end;//of t


    //********************************************************************//
    //                                                                    //
    //                   Penalty for Rates' Upper/LowerBound              //
    //                                                                    //
    //********************************************************************//
    if (OptParams.EstUseMinRate) or (OptParams.EstUseMaxRate) then begin

      //Log-likelihood .....................................................
      CalculateRateMinMaxPenalty;

      if (prOptParams.OptMode <> MODE_SIM) and (CompGrad = True) and (NumGrad = False) then begin
        //Rates Gradients ....................................................
        CalculateRateMinMaxPenaltyGradients;
      end;

    end;//Penalty ..........................................................
 

  finally

  end;


end;

(*
function TQUB_Task_Mac.SimOfChunk(c :Integer) :Boolean;
var
  MacChunk                     :TQUB_Mac_PChunk;
  FitChunk                     :TQUB_Mac_PChunk;
  MacModel                     :TQUB_Mac_Model;
  Chunk                        :TQUB_Thread_PChunk;
  DataCount                    :Integer;
  IData                        :TPAInteger;
  TData                        :TPASingle;
  FData                        :Double;
  cLL, cGR, cSS                :Double;
  k, m, t                      :Integer;
  l, ec                        :Integer;
  ln2PI_05                     :Double;
  Scale                        :Double;
  dt                           :Double;
  mt, Vt, VtInv                :Double;
  dmt, dVt                     :Double;

  ECDwell                      :Integer;
  ECClass                      :Integer;
  ECNextChange                 :Integer;

  PrevT0                       :Integer;
  mc                           :Integer;

  ThisVoltage                  :Double;
  VoltageCh                    :Integer;
  i, j                         :Integer;
  pq, pk0, pk1, pC, pV         :Double;
  resample_i                   :Integer;

  VRev                         :Double;
  VLeak                        :Double;
  GLeak                        :Double;

  ChunkPercent                 :Integer;
  PrevPercent                  :Integer;
  Percent                      :Integer;

  PrevTime0                    :Single;
  Time                         :Single;
  DTime                        :Single;

  //Penalty ....................................................................
  //A hard threshold is not handled well by the optimizer (unequal left and right derivatives)
  function RatePenaltyMin(q, k0, k1, C, V, minq :Double) :Double;
  const
    A = 0.01;
  var
    q0 :Double;
  begin
    q0 := q / (C*exp(k1*V));
    //1st penalty: q > qmin
    Result :=          (ln(minq) - ln(q))  / (1 + exp((ln(q)  - ln(minq))/A));
    //Result :=          ((minq - q)/minq)  / (1 + exp(((q  - minq)/minq)/A));
    //2nd penalty; k0 > qmin
    Result := Result + (ln(minq) - ln(q0)) / (1 + exp((ln(q0) - ln(minq))/A));
  end;

  function DRatePenaltyMin(q, k0, k1, C, V, minq :Double) :Double;
  const
    A = 0.01;
  begin
    //Use a numerical derivative, until figure out the analytical solution
    Result := (RatePenaltyMin(q + 1e-4, k0, k1, C, V, minq) - RatePenaltyMin(q - 1e-4, k0, k1, C, V, minq)) / 2e-4;
  end;

  function RatePenaltyMax(q, k0, k1, C, V, maxq :Double) :Double;
  const
    A = 0.01;
  var
    q0 :Double;
  begin
    q0 := q / (C*exp(k1*V));
    //1st penalty: q < qmax
    Result :=          (ln(q)  - ln(maxq)) / (1 + exp((ln(maxq) -  ln(q))/A));
    //Result :=          ((q  - maxq)/maxq) / (1 + exp(((maxq -  q)/maxq)/A));
    //2nd penalty; k0 < qmin
    Result := Result + (ln(q0) - ln(maxq)) / (1 + exp((ln(maxq) - ln(q0))/A));
  end;

  function DRatePenaltyMax(q, k0, k1, C, V, maxq :Double) :Double;
  const
    A = 0.01;
  begin
    //Use a numerical derivative, until figure out the analytical solution
    Result := (RatePenaltyMax(q + 1e-4, k0, k1, C, V, maxq) - RatePenaltyMax(q - 1e-4, k0, k1, C, V, maxq)) / 2e-4;
  end;

  procedure CalculateRateMinMaxPenalty;
  var
    k, ec :Integer;
  begin

    for k := 0 to MacModel.RateCount - 1 do begin

      i := MacModel.c2qS[2*k + 0].x;
      j := MacModel.c2qS[2*k + 0].y;

      //Cond Equilibrium
      if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

        //Need:

        //q[ec][i, j]
        pq  := MacModel.Cond_Q[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.Cond_ExpCondP[2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.Cond_ExpCondQ[2*k + 1];

        cSS := 0;
        if OptParams.EstUseMinRate then begin
          cSS := cSS + RatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cSS := cSS + RatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        cLL := - cSS;
        MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
        MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;
        MacChunk.LL := MacChunk.LL - MacModel.FitWeight*cLL;

      end;//Cond eq ?

      for ec := 0 to MacModel.ExpCondCount - 1 do begin

        //Need:

        //q[ec][i, j]
        pq  := TPMDouble(MacModel.Q[ec])[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.ExpCondP[ec][2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.ExpCondQ[ec][2*k + 1];

        cSS := 0;
        if OptParams.EstUseMinRate then begin
          cSS := css + RatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cSS := css + RatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        cLL := - cSS;
        MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
        MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;
        MacChunk.LL := MacChunk.LL - MacModel.FitWeight*cLL;

      end;//ec

    end;//k

  end;//CalculateRateMinMaxPenalty

  procedure CalculateRateMinMaxPenaltyGradients;
  var
    k, ec :Integer;
  begin

    for k := 0 to MacModel.RateCount - 1 do begin

      i := MacModel.c2qS[2*k + 0].x;
      j := MacModel.c2qS[2*k + 0].y;

      //Cond Equilibrium
      if OptParams.EstInitProbs = INITPROBS_CONDEQ then begin

        //Need:

        //q[ec][i, j]
        pq  := MacModel.Cond_Q[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.Cond_ExpCondP[2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.Cond_ExpCondQ[2*k + 1];

        cGR := 0;
        if OptParams.EstUseMinRate then begin
          cGR := cGR - DRatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cGR := cGR - DRatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        MacModel.dLLdCond_q[k] := MacModel.dLLdCond_q[k] - MacModel.FitWeight*cGR;

      end;//Cond eq ?

      for ec := 0 to MacModel.ExpCondCount - 1 do begin

        //Need:

        //q[ec][i, j]
        pq  := TPMDouble(MacModel.Q[ec])[i, j];
        //k0[i, j]
        pk0 := Exp(MacModel.cPars[2*k + 0]);
        //k1[i, j]
        pk1 := MacModel.cPars[2*k + 1];
        //P[ec][i, j]
        pC  := MacModel.ExpCondP[ec][2*k + 0];
        //Q[ec][i, j]
        pV  := MacModel.ExpCondQ[ec][2*k + 1];

        cGR := 0;
        if OptParams.EstUseMinRate then begin
          cGR := cGR - DRatePenaltyMin(pq, pk0, pk1, pC, pV, OptParams.EstMinRate);
        end;
        if OptParams.EstUseMaxRate then begin
          cGR := cGR - DRatePenaltyMax(pq, pk0, pk1, pC, pV, OptParams.EstMaxRate);
        end;
        MacModel.dLLdq[ec, k] := MacModel.dLLdq[ec, k] - MacModel.FitWeight*cGR;

      end;//ec

    end;//k

  end;//CalculateRateMinMaxPenaltyGradients

begin

  Result                       := False;

  TData                        := nil;

  //This was not calculated here ***********************************************
  VoltageCh                    := -1;
  ln2PI_05                     := 0.5*ln(2*PI);
  VtInv                        := 1;
  //****************************************************************************

  //ChunkData
  MacChunk                     := prChunks[c];
  Chunk                        := prChunks[c];
  if MacChunk.Data = nil then begin
    DataCallback(MacChunk);
  end;
  IData                        := MacChunk.Data;
  if MacChunk.HasTime then begin
    TData                      := MacChunk.TimeData;
  end;
  DataCount                    := TQUB_Thread_PChunk(MacChunk).Sel.Last - TQUB_Thread_PChunk(MacChunk).Sel.First + 1;
  Scale                        := 1 / MacChunk.Scaling;
  dt                           := MacChunk.Sampling;

  Percent := Round((TQUB_Mac_Model(prMacModels[0]).Iterations + c/prChunks.Count) / prOptParams.OptMaxIter * 100);
  ChunkPercent := 0;
  PrevPercent  := 0;
  prStatus := Format('Opt. chunk %d/%d: %d%%', [c + 1, prChunks.Count, ChunkPercent]);
  prUpdateStatus(Percent, prStTime);

  try


    //Reset the previous start time
    PrevT0      := 0;
    PrevTime0   := 0;
    Time        := 0;
    if MacChunk.HasTime then begin
      PrevTime0 := TData[PrevT0];
      Time      := PrevTime0;
    end;

    //Choose the right MacModel
    MacModel    := TQUB_Mac_Model(MacChunk.MacModel);
    MacChunk.LL := 0;

    VRev        := MacModel.VRev;
    VLeak       := MacModel.VLeak;
    GLeak       := MacModel.GLeak;

    VoltageCh   := -1;

    //State probabilities
    MacModel.Do_P0;


    //Initialize experimental condition
    ECDwell      := 0;
    ECClass      := MacChunk.ExpCondIdl[ECDwell].Cls;
    ECNextChange := MacChunk.ExpCondIdl[ECDwell].N;


    ThisVoltage  := 0;
    if prOptParams.OptUseVRev <> 0 then begin
      VoltageCh := -1;
      for ec := 0 to MacChunk.MacModel.ExpCondParams[ECClass].Count - 1 do begin
        if MacChunk.MacModel.ExpCondParams[ECClass][ec].Name = 'Voltage' then begin
          VoltageCh   := ec;
          Break;
        end;
      end;
      if VoltageCh > -1 then begin
        ThisVoltage := MacChunk.MacModel.ExpCondParams[ECClass][VoltageCh].Value;
      end;
    end;


    //Chris, the segmentation does not appear to be right - last segment is one point short

    resample_i   := 0;

    //For each DataPoint
    for t := 0 to DataCount - 1 do begin

      if prStopped then begin
        if prStoppedCount >= 3 then Break;
      end;

      if prSkipSlowChange then begin
        if t <> MacChunk.ResampleSamples[resample_i] then begin
          Continue;
        end;
      end;

      //New - for data with TimeCh
      if MacChunk.HasTime then begin
        Time := TData[t];
      end;

      //Change ECClass?
      if (t >= ECNextChange) and ((ECDwell + 1) < MacChunk.ExpCondIdlCount){Apparently, the last dwell is one point short} then begin
      //something fishy in this condition !!!

        try

          {
          Assumption: the stimuli change at this very point
          Recalculate P0 and dP0/dQ as
            P0[this_seg] = P0[prev_seg] * A[prev_seg,dur]
            dP0[this_seg] / dQ[prev_seg] = (dP0[prev_seg] / dQ[prev_seg]) * A[prev_seg_dur] +
                                           P0[prev_seg] * (dA[prev_seg,dur] / dQ[prev_seg])
          }

          //New - for data with time channel - uneven samples etc;
          if MacChunk.HasTime then begin
            DTime := Time - PrevTime0;
          end else begin
            DTime := (t-PrevT0)*dt;
          end;

          //A matrix
          MacModel.Do_At(DTime, ECClass);//All matrices or just ECClass ?!

          //Current state, later to be copied into initial state
          MacModel.Do_Pt(DTime, ECClass);

          //Copy Pt into P0, and the derivatives
          MacModel.PtToP0;

        finally

          Inc(ECDwell);
          ECClass      := MacChunk.ExpCondIdl[ECDwell].Cls;
          ECNextChange := ECNextChange + MacChunk.ExpCondIdl[ECDwell].N;
          PrevT0       := t;
          PrevTime0    := Time;

          if prOptParams.OptUseVRev <> 0 then begin
            if VoltageCh > -1 then begin
              ThisVoltage := MacChunk.MacModel.ExpCondParams[ECClass][VoltageCh].Value;
            end;
          end;

        end;

      end;


      //**********************************************************************//
      //                                                                      //
      //                             Fit this point !                         //
      //                                                                      //
      //**********************************************************************//
      if MacChunk.FitPoint[t] then begin


        ChunkPercent := Round(t / DataCount * 100);
        if (ChunkPercent - PrevPercent) >= 10 then begin
          prStatus := Format('Opt. chunk %d/%d: %d%%', [c+1, prChunks.Count, ChunkPercent]);
          Synchronize(prDoOnChange);
          PrevPercent := ChunkPercent;
        end;

        //Testing. LM
        Inc(MacModel.PointCount);//, MacModel.AverageCount);

        //New - for data with time channel - uneven samples etc;
        if MacChunk.HasTime then begin
          DTime := Time - PrevTime0;
        end else begin
          DTime := (t-PrevT0)*dt;
        end;

        //A matrix
        MacModel.Do_At(DTime, ECClass);

        //Current state
        MacModel.Do_Pt(DTime, ECClass);

        //I Mean
        MacModel.Do_Mean_t(mt);

        //Correct for I = g*(V-Vrev)
        if (prOptParams.OptUseVRev <> 0) and (not MacModel.HasFitP) then begin
          mt := mt * (ThisVoltage - VRev) * 1e-3;
          mt := mt + GLeak * (ThisVoltage - VLeak) * 1e-3;
        end;

        if (OptParams.EstUseVariance <> 0) and (not MacModel.HasFitP) then begin
          //State Variance
          MacModel.Do_VarPt;

          //I Variance
          MacModel.Do_Var_t(Vt);
          if Vt <= 0 then Vt := 1e-30;
          VtInv := 1 / Vt;

        end;


        //Log-Likelihood .....................................................
        FData := MacChunk.Data[t]*Scale;

        cSS := Sqr(FData - mt);
        if OptParams.EstUseVariance = 0 then begin
          cLL := - cSS;
        end else begin
          cLL := - ln2PI_05 - 0.5*ln(Vt) - 0.5*cSS*VtInv;
        end;

        MacModel.LL := MacModel.LL - MacModel.FitWeight*cLL;//Use minus sign in order to minimize LL
        MacModel.SS := MacModel.SS + MacModel.FitWeight*cSS;
        MacChunk.LL := MacChunk.LL - MacModel.FitWeight*cLL;


      end;//Fit this point!


      //Show fitting curve ?
      if (OptParams.OptShowFit <> 0) then begin
        //FitChunk := TQUB_Mac_PChunk(prChunks[c]);
        FitChunk := MacChunk;
        //************************************************************
        //if FitThisPoint then begin
        if MacChunk.FitPoint[t] then begin
          if          OptParams.OptShowFitWhat = SHOWFIT_AVG then begin
            FitChunk.FitCurve[t] := Round(mt * FitChunk.Scaling);
          end else if OptParams.OptShowFitWhat = SHOWFIT_STD then begin
            FitChunk.FitCurve[t] := Round(Sqrt(Vt) * FitChunk.Scaling);
          end else if OptParams.OptShowFitWhat = SHOWFIT_RES then begin
            //FitChunk.FitCurve[t] := Round(mt * FitChunk.Scaling - TQUB_Mac_PChunk(prChunks[c]).Data[t]);
            FitChunk.FitCurve[t] := Round(mt * FitChunk.Scaling - MacChunk.Data[t]);
          end;
        end else begin
          //FitChunk.FitCurve[t] := TQUB_Mac_PChunk(prChunks[c]).Data[t];
          FitChunk.FitCurve[t] := MacChunk.Data[t];
        end;
        //****************************************************************

      end;//ShowFit

      Inc(resample_i);

    end;//of t


    //********************************************************************//
    //                                                                    //
    //                   Penalty for Rates' Upper/LowerBound              //
    //                                                                    //
    //********************************************************************//
    if (OptParams.EstUseMinRate) or (OptParams.EstUseMaxRate) then begin

      //Log-likelihood .....................................................
      CalculateRateMinMaxPenalty;

    end;//Penalty ..........................................................



  finally

  end;


end;
*)





(*
//Calculate a penalty and its derivatives with respect to the free parameters
//by calling the PenaltyScript
procedure TQUB_Task_Mac.prCalculateScriptPenalty;
var
  MacModel :TQUB_Mac_Model;
  LLLeft   :Double;
  LLRight  :Double;
  OldPar   :Double;
  f        :Integer;
begin

  MacModel := prMacModels[0];

  OldPar := 0;
  
  try


    prStatus := 'Opt. script penalty';
    Synchronize(prDoOnChange);

    //First, calculate the LL for the current set of parameters
    MacModel.Do_fParsTocPars;
    Synchronize(prUpdateModel);
    Synchronize(prRunScriptPenalty);
    prPenaltyLL := prScriptPenaltyLL;

    if prOptParams.OptNumGradsFast = 1 then begin
      LLLeft :=  prScriptPenaltyLL;
    end;

    //Use + sign because the Penalty LL is actually a SS, which has to be minimized
    MacModel.LL := MacModel.LL + prPenaltyLL;
    MacModel.SS := MacModel.SS - prPenaltyLL;


    if prOptParams.OptMode <> MODE_SIM then begin

      try

        for f := 0 to MacModel.fParCount - 1 do begin

          if prStopped then begin
            if prStoppedCount >= 3 then Break;
          end;

          prStatus := Format('Opt. script penalty num. grad. %d/%d', [f + 1, MacModel.fParCount]);
          Synchronize(prDoOnChange);

          OldPar := MacModel.fPars[f];

          if prOptParams.OptNumGradsFast = 0 then begin
            //Left derivative
            MacModel.fPars[f] := OldPar - prOptParams.OptNumGradStep;
            MacModel.Do_fParsTocPars;
            Synchronize(prUpdateModel);
            Synchronize(prRunScriptPenalty);
            LLLeft := prScriptPenaltyLL;
          end;

          //Right derivative
          MacModel.fPars[f] := OldPar + prOptParams.OptNumGradStep;
          MacModel.Do_fParsTocPars;
          Synchronize(prUpdateModel);
          Synchronize(prRunScriptPenalty);
          LLRight := prScriptPenaltyLL;

          if prOptParams.OptNumGradsFast = 0 then begin
            prPenaltyGrads[f] := (LLRight - LLLeft)/(2*prOptParams.OptNumGradStep);
          end else begin
            prPenaltyGrads[f] := (LLRight - LLLeft)/prOptParams.OptNumGradStep;
          end;

          MacModel.dLLdfPar[f] := MacModel.dLLdfPar[f] + prPenaltyGrads[f];
          //MacModel.dLLdfPar[f] := MacModel.dLLdfPar[f] - prPenaltyGrads[f];

          //Reset this parameter
          MacModel.fPars[f] := OldPar;

        end;

      finally

        MacModel.Do_fParsTocPars;
        Synchronize(prUpdateModel);

      end;

    end;

  except
  end;

end;

procedure TQUB_Task_Mac.prRunScriptPenalty;
var
  Result :Variant;
begin
  ViewportScriptlets.RunScriptlet(ViewportScriptlets.ScriptletIndexOfName[prOptParams. OptScriptPenaltyName, ''], False{ASynchronous}, False{DontSynchThread}, False{WaitUntilAvailable}, -1{Affinity}, 0{Priority}, Result, nil, 0);
  prScriptPenaltyLL := Result;
end;

procedure TQUB_Task_Mac.prRunScriptIteration;
var
  Result :Variant;
begin
  ViewportScriptlets.RunScriptlet(ViewportScriptlets.ScriptletIndexOfName[prOptParams. OptScriptIterationName, ''], False{ASynchronous}, False{DontSynchThread}, False{WaitUntilAvailable}, -1{Affinity}, 0{Priority}, Result, nil, 0);
end;

procedure TQUB_Task_Mac.prRunScriptSetPars;
var
  Result :Variant;
begin
  ViewportScriptlets.RunScriptlet(ViewportScriptlets.ScriptletIndexOfName[prOptParams. OptScriptSetParsName, ''], False{ASynchronous}, False{DontSynchThread}, False{WaitUntilAvailable}, -1{Affinity}, 0{Priority}, Result, nil, 0);
end;
*)
procedure TQUB_Task_Mac.prFitPointsOfChunk(MacChunk :TQUB_Mac_PChunk);
var
  DataCount                    :Integer;
  FitList                      :TPAPoint;
  FitListCount                 :Integer;
  FitListIdx                   :Integer;
  ExcList                      :TPAPoint;
  ExcListCount                 :Integer;
  ExcListIdx                   :Integer;
  dt                           :Double;
  l, t                         :Integer;
begin

  DataCount                    := TQUB_Thread_PChunk(MacChunk).Sel.Last - TQUB_Thread_PChunk(MacChunk).Sel.First + 1;
  dt                           := MacChunk.Sampling;

  FitListCount                 := MacChunk.FitListCount;
  FitList                      := MacChunk.FitList;
  FitListIdx                   := 0;
  ExcListCount                 := MacChunk.ExcListCount;
  ExcList                      := MacChunk.ExcList;
  ExcListIdx                   := 0;

  //For each DataPoint
  for t := 0 to DataCount - 1 do begin

    //Fit this point ?
    MacChunk.FitPoint[t] := True;
    if (prOptParams.OptSkip = 1) and (t*dt <= prOptParams.OptSkipTime*1e-3) then begin
      MacChunk.FitPoint[t] := False;        //^^^ do it for TimeCh?
    end;
    if (MacChunk.FitPoint[t]) and (FitListCount > 0) then begin
      MacChunk.FitPoint[t] := False;
      for l := FitListIdx to FitListCount - 1 do begin // This may have been out of synch ? JB.
        if (t >= FitList[l].x) and (t <= FitList[l].y) then begin
          MacChunk.FitPoint[t] := True;
          FitListIdx := l;
          Break;
        end;
      end;
    end;
    if (MacChunk.FitPoint[t]) and (ExcListCount > 0) then begin
      for l := ExcListIdx to ExcListCount - 1 do begin // This may have been out of synch ? JB.
        if (t >= ExcList[l].x) and (t <= ExcList[l].y) then begin
          MacChunk.FitPoint[t] := False;
          ExcListIdx := l;
          Break;
        end;
      end;
    end;

  end;//of t

end;


//------------------------------------
procedure TQUB_Task_Mac.prDoOverlapData;
var
  FitIter         :TQUB_FitCurveIter;
  Chunk           :TQUB_Mac_PChunk;
  QVD             :TQUB_QVO_QVD;
  i, di, ri, x    :Integer;
  FitData         :TPAInteger;
begin
  if OptParams.OptShowFit = 0 then Exit;

  if not prFitPreCleared then begin
    if prChunks.Count > 0 then begin
      QVD := TQUB_Mac_PChunk(prChunks[0]).QVD;
      QVD.FitCurves.Clear;
      for i := 1 to prChunks.Count - 1 do begin
        if TQUB_Mac_PChunk(prChunks[i]).QVD <> QVD then begin
          QVD := TQUB_Mac_PChunk(prChunks[i]).QVD;
          QVD.FitCurves.Clear;
        end;
      end;
    end;
  end;

  for i := 0 to prChunks.Count - 1 do begin
    Chunk         := TQUB_Mac_PChunk(prChunks[i]);
    FitIter       := nil;
    try
      FitIter       := Chunk.QVD.FitCurves.Find(Chunk.Chunk.Sel.First);
      if (0 <> FitIter.Done) or (FitIter.Curve.First <> Chunk.Chunk.Sel.First) or (FitIter.Curve.Last <> Chunk.Chunk.Sel.Last) then begin
        FitIter.Free;
        FitIter     := Chunk.QVD.FitCurves.Add(Chunk.Chunk.Sel.First, Chunk.Chunk.Sel.Last);
      end;
      if prSkipSlowChange and (Chunk.ResampleSamples <> nil) then begin
        FitData     := Chunk.FitCurve;
        for ri := 0 to Chunk.ResampleCount - 1 do begin
          x         := FitData[Chunk.ResampleSamples[ri]];
          for di := Chunk.ResampleIntervals[ri].First to Chunk.ResampleIntervals[ri].Last do begin
            FitData[di] := x;
          end;
        end;
      end;
      Move(Chunk.FitCurve^, FitIter.Curve.Data^, (Chunk.Chunk.Sel.Last - Chunk.Chunk.Sel.First + 1) * SizeOf(Integer));
      FitIter.MarkChanged;
    finally
      FitIter.Free;
    end;
  end;

  if prChunks.Count > 0 then begin
    QVD := TQUB_Mac_PChunk(prChunks[0]).QVD;
    QVD.FitCurves.RaiseChanged;
    for i := 1 to prChunks.Count - 1 do begin
      if TQUB_Mac_PChunk(prChunks[i]).QVD <> QVD then begin
        QVD := TQUB_Mac_PChunk(prChunks[i]).QVD;
        QVD.FitCurves.RaiseChanged;
      end;
    end;
  end;

end;

//******************************************************************************
procedure TQUB_Task_Mac.prUpdateModel;
var
  c :Integer;
begin

  //Rates
  MakeRatesFromParams(Model.ModelObject, TQUB_Mac_Model(prMacModels[0]).cPars, 2*TQUB_Mac_Model(prMacModels[0]).RateCount, OptParams.OptAllowNegRates);

  //Extra kinetic parameters
  MakeExtraKinPars(Model.ModelObject, TQUB_Mac_Model(prMacModels[0]).ExtraKinPars, TQUB_Mac_Model(prMacModels[0]).ExtraKinParCount);

  //ChannelCount
  if OptParams.EstChannelCount <> 0 then begin

    //CCFixed?

    if OptParams.EstChannelCountPerFile = 0 then begin
      c := Round(Min(TQUB_Mac_Model(prMacModels[0]).ChannelCount, High(Integer)));
      //When MacModel.ChannelCount was > High(Integer) it caused a roll-over of c into a negative and resulted in c = 1. LM.
      if c < 1 then c := 1;
      Model.ModelObject.ChannelCount := c;
    end else begin
      //NewChannelCount ???
      c := Round(Min(TQUB_Mac_Model(prMacModels[0]).ChannelCount, High(Integer)));
      if c < 1 then c := 1;
      Model.ModelObject.ChannelCount := c;
    end;
  end;

  //AmpOpen
  if OptParams.EstAmpOpen <> 0 then begin
    //???
  end;

  //VarOpen
  if OptParams.EstVarOpen <> 0 then begin
    //???
  end;

end;

procedure TQUB_Task_Mac.prResetModel;
var
  m :Integer;
begin

  if OptParams.OptPropagateParams = 0 then begin

    //Rates
    for m := 0 to prOptSegsOldMdl.Rates.RateCount - 1 do begin
      prModel.ModelObject.Rates[m].k0[1] := prOptSegsOldMdl.Rates[m].k0[1];
      prModel.ModelObject.Rates[m].k0[2] := prOptSegsOldMdl.Rates[m].k0[2];
      prModel.ModelObject.Rates[m].k1[1] := prOptSegsOldMdl.Rates[m].k1[1];
      prModel.ModelObject.Rates[m].k1[2] := prOptSegsOldMdl.Rates[m].k1[2];
    end;

    //Extra kinetic parameters
    for m := 0 to prOptSegsOldMdl.ExtraKineticPars.ExtraKineticParCount - 1 do begin
      prModel.ModelObject.ExtraKineticPars[m].Value := prOptSegsOldMdl.ExtraKineticPars[m].Value;
    end;

    //ChannelCount
    //CCFixed?
    if OptParams.EstChannelCountPerFile = 0 then begin
      prModel.ModelObject.ChannelCount := prOptSegsOldMdl.ChannelCount;
    end else begin
      //NewChannelCount ???
      prModel.ModelObject.ChannelCount := prOptSegsOldMdl.ChannelCount;
    end;

    //AmpOpen
    //???

    //VarOpen
    //???

  end;

  for m := 0 to prMacModels.Count - 1 do begin
    TQUB_Mac_Model(prMacModels[m]).Do_InitcPars;
    TQUB_Mac_Model(prMacModels[m]).Do_InitfPars;
  end;
end;

procedure TQUB_Task_Mac.prUpdateData;
var
  QVD  :TQUB_QVO_QVD;
  i    :Integer;
begin

  if OptParams.OptShowFit <> 0 then begin
    if prChunks.Count > 0 then begin
      QVD := TQUB_Mac_PChunk(prChunks[0]).QVD;
      QVD.Zoom.Draw;
      for i := 1 to prChunks.Count - 1 do begin
        if TQUB_Mac_PChunk(prChunks[i]).QVD <> QVD then begin
          QVD := TQUB_Mac_PChunk(prChunks[i]).QVD;
          QVD.Zoom.Draw;
        end;
      end;
    end;
  end;

end;

procedure TQUB_Task_Mac.prShowData;
begin
  //TO DO :Do it for FileList. LM
  if Data <> nil then begin
    //TO DO: Make it an option. LM
    if TQUB_QVO_QVD(Data).DrawListOverlapped then begin
      //TQUB_QVO_QVD(Data).MoveSelectionIntoWindow(QUB_FirstLast(0, TQUB_Thread_PChunk(prChunks[0]).Sel.Last - TQUB_Thread_PChunk(prChunks[0]).Sel.First), TQUB_Thread_PChunk(prChunks[0]).Sel.First);
    end else begin
      TQUB_QVO_QVD(Data).MoveSelectionIntoWindow(QUB_FirstLast(TQUB_Thread_PChunk(prChunks[0]).Sel.First, TQUB_Thread_PChunk(prChunks[prChunks.Count - 1]).Sel.Last), TQUB_Thread_PChunk(prChunks[0]).Sel.First);
    end;
  end;
end;

procedure TQUB_Task_Mac.prRefreshZoom;
begin
  //TO DO:Do it for FileList. LM
  if Data <> nil then begin
    TQUB_QVO_QVD(Data).Zoom.Draw;
  end;
end;
(*
procedure TQUB_Task_Mac.prUpdateLLPlot;
var
  Model :TQUB_Mac_Model;
  i, m  :Integer;
begin

  Model := TQUB_Mac_Model(prMacModels[0]);
  if Model.Iterations = 1 then begin

    //LL
    ViewportGraph.Graph[prOptParams.OptPlotLLPage].AllocData(prOptParams.OptPlotLLTrace, prOptParams.OptMaxIter);
    ViewportGraph.Graph[prOptParams.OptPlotLLPage].ChannelNames[prOptParams.OptPlotLLTrace] := 'Mac fit LL';

    {
    //Rates
    m := 1;
    for k := 0 to (Model.RateCount div 2) - 1 do  begin

      i := Model.c2qS[4*k].x;
      j := Model.c2qS[4*k].y;

      ViewportGraph.Graph.AllocData(m, prOptParams.OptMaxIter);
      ViewportGraph.Graph.ChannelNames[m] := Format('k0(%d,%d)', [i, j]);
      Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
      ViewportGraph.Graph.AllocData(m, prOptParams.OptMaxIter);
      ViewportGraph.Graph.ChannelNames[m] := Format('k1(%d,%d)', [i, j]);
      Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
      ViewportGraph.Graph.AllocData(m, prOptParams.OptMaxIter);
      ViewportGraph.Graph.ChannelNames[m] := Format('k0(%d,%d)', [j, i]);
      Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
      ViewportGraph.Graph.AllocData(m, prOptParams.OptMaxIter);
      ViewportGraph.Graph.ChannelNames[m] := Format('k1(%d,%d)', [j, i]);
      Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;

    end;

    //Extra kinetic parameters - to do
    //. . .

    //Channel Count
    if m <= QUB_QVO_QVGRAPH_MAXCHANNEL then begin
      for k := 0 to Model.ChannelCountCount - 1 do begin
        ViewportGraph.Graph.AllocData(m, prOptParams.OptMaxIter);
        ViewportGraph.Graph.ChannelNames[m] := Format('Nc(%d)', [k + 1]);
        Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
      end;
    end;
    }

    for i := 0 to ViewportGraph.Graph[prOptParams.OptPlotLLPage].DataCount[prOptParams.OptPlotLLTrace] - 1 do begin
      ViewportGraph.Graph[prOptParams.OptPlotLLPage].Data[prOptParams.OptPlotLLTrace, i].x := i;
      ViewportGraph.Graph[prOptParams.OptPlotLLPage].Data[prOptParams.OptPlotLLTrace, i].y := 0;
    end;

  end;

  ViewportGraph.Graph[prOptParams.OptPlotLLPage].DataXMin := 0;
  ViewportGraph.Graph[prOptParams.OptPlotLLPage].DataXMax := OptParams.OptMaxIter{Model.Iterations} - 1;

  //LL
  ViewportGraph.Graph[prOptParams.OptPlotLLPage].Data[prOptParams.OptPlotLLTrace, Model.Iterations - 1].y := prPLL;

  {
  //Rates
  m := 1;
  for k := 0 to (Model.RateCount div 2) - 1 do  begin

    ViewportGraph.Graph.Data[m, Model.Iterations - 1].y := Exp(Model.cPars[4*k + 0]);
    Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
    ViewportGraph.Graph.Data[m, Model.Iterations - 1].y :=     Model.cPars[4*k + 1];
    Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
    ViewportGraph.Graph.Data[m, Model.Iterations - 1].y := Exp(Model.cPars[4*k + 2]);
    Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
    ViewportGraph.Graph.Data[m, Model.Iterations - 1].y :=     Model.cPars[4*k + 3];
    Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;

  end;

  //Extra kinetic parameters - to do
  //. . .

  //Channel Count
  if m <= QUB_QVO_QVGRAPH_MAXCHANNEL then begin
    for k := 0 to Model.ChannelCountCount - 1 do begin
      ViewportGraph.Graph.Data[m, Model.Iterations - 1].y := Exp(Model.cPars[Model.RateCount + k]);
      Inc(m); if m > QUB_QVO_QVGRAPH_MAXCHANNEL then Break;
    end;
  end;
  }

  ViewportGraph.Graph[prOptParams.OptPlotLLPage].Draw;

end;
*)

//****************************************************************************//
//                                                                            //
//                              E X E C U T E                                 //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Mac.prDoExecute;
begin
  prConsoleMsg  := 'Mac optimizer running ... ';
  prConsoleCode := MESSAGE_MESSAGE;
  Synchronize(prWriteNewConsole);
  inherited;
end;

procedure TQUB_Task_Mac.prDoExecuteStart;
begin

  inherited;

  //TO DO: Do it for FileList. LM
  prSelectionList := nil;
  if Data <> nil then begin
    if TQUB_QVO_QVD(Data).SelectionLists.Items.Count > 0 then begin
      if TQUB_QVO_QVD(Data).SelectionLists.ItemIndex <> -1 then begin
        prSelectionList := TQUB_QVO_QVD(Data).SelectionLists[TQUB_QVO_QVD(Data).SelectionLists.ItemIndex];
        prSelectionList.Lock;
      end;
    end;
  end;

  //MacModels
  prMacModels     := TList.Create;

  //ExpConds
  prExpCondParams := TQUB_ExpCond_ParametersListList.Create;

  if Model <> nil then begin
    Properties.AppendChild(Model.ModelObject.CloneAsNode(True, TNode.CreateNull)).SetData('Initial');
  end;

  prCuda_Lib  := TQUB_Cuda_Lib.Create;

end;


procedure TQUB_Task_Mac.prDoExecuteEnd;
var
  i :Integer;
begin
  if Model <> nil then Model.UnlockAny;

  if prSelectionList <> nil then prSelectionList.UnLock;

  //Free Params & MacModels
  if prMacModels <> nil then begin
    for i := 0 to prMacModels.Count - 1 do begin
      TQUB_Mac_Model(prMacModels[i]).Free;
    end;
    prMacModels.Free;
    prMacModels := nil;
  end;

  prExpCondParams.Free;
  prExpCondParams := nil;

  Synchronize(prRefreshZoom);

  prCuda_Lib.Free;

  inherited;
end;


//****************************************************************************//
//                                                                            //
//                               C H U N K S                                  //
//                                                                            //
//****************************************************************************//
function TQUB_Task_Mac.prChunkData :Boolean;
var
  DataSets         :INode;
  DS, Seg          :INode;
  ResDS, ResSeg    :INode;

  QVD              :TQUB_QVO_QVD;
  QDF              :TQUB_QFS_QDF;

  ParamClasses     :TQUB_ExpCond_ParametersList;
  Params           :TQUB_ExpCond_Parameters;
  ParamsAreVar     :TQUB_ExpCond_Parameters;

  //ExpCondIdlGen    :TQUB_ExpCondIdlGen;

  MacModel         :TQUB_Mac_Model;

  FileIndex        :Integer;

  Chunk            :TQUB_Thread_PChunk;
  MacChunk         :TQUB_Mac_PChunk;
  //ChunkBounds      :TList;
  Segments         :array of TQUB_FirstLast;
  FirstChunk       :Integer;

  //Sel              :TQUB_PFirstLast;
  Dat              :TQUB_FirstLast;
  PrevDat          :TQUB_FirstLast;

  DList            :TQUB_QVO_QVD_SelectionList;

  c, l, p, m, n, s :Integer;

  CCFixed          :Boolean;
  TimeCh           :Integer;

  FitStatesList    :TQUB_QVO_QVD_SelectionList;
  FitStates        :TPAInteger;
  FitStatesCount   :Integer;
  
begin

  Result           := False;

  prSetChunkParameters;

  ParamsAreVar     := nil;
  Params           := nil;

  FitStatesCount   := 100;
  FitStates        := IAlloc1(FitStatesCount);


  try


    //Need to know how many models (one per data set), in case ChannelCount is optimized per file (or fixed)
    //It's also possible that ChannelCount is per file but not optimized
    prOptParams.ChannelCountCount := 0;
    prOptParams.ChannelCountIdx   := 0;
    if          (prOptParams.EstChannelCount = 1) and (prOptParams.EstChannelCountPerFile = 0) then begin
      prOptParams.ChannelCountCount := 1;
    end else if (prOptParams.EstChannelCount = 1) and (prOptParams.EstChannelCountPerFile = 1) then begin
      DataSets := DataNoder.GetChosenData( TNode.CreateNull );
      DS       := DataSets.Find('DataSet');
      while not DS.IsNull do begin
        CCFixed := False;
        try
          Params := TQUB_ExpCond_Parameters.Create;
          Params.ReadTree(DS.Find('ExpCond'));
          for p := 0 to Params.Count - 1 do begin
            if Params[p].Name = 'CCFixed' then begin
              CCFixed := Boolean(Round(Params[p].Value));
              Break;
            end;
          end;
        finally
          Params.Free;
          Params := nil;
        end;
        if not CCFixed then begin
          Inc(prOptParams.ChannelCountCount);
        end;
        DS := DS.SiblingSameName;
      end;
    end;


    prDataSetCount := 0;

    DataSets       := DataNoder.GetChosenData( TNode.CreateNull );
    prResults.AppendClone(DataSets['DataDescription'], True);
    DS             := DataSets.Find('DataSet');
    FileIndex      := 0;
    ResDS          := TNode.CreateNull;


    while not DS.IsNull do begin

      //Find the QVD object for this DS
      QVD         := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DS['FileName'].DataAsString]);
      //Find if there is a FitStates list in this QVD
      FitStatesCount := 0;
      if QVD <> nil then begin
        FitStatesList := QVD.SelectionLists.ListByName['FitStates'];
        if FitStatesList <> nil then begin
          FitStatesCount := 100;
          //Read the list of Fit States
          StrToArrayOfInt(FitStatesList.Selection[0].Comment, FitStates, FitStatesCount);
        end;
      end;

      CCFixed := False;
      TimeCh  := -1;
      try
        Params := TQUB_ExpCond_Parameters.Create;
        Params.ReadTree(DS.Find('ExpCond'));
        for p := 0 to Params.Count - 1 do begin
          if Params[p].Name = 'CCFixed' then begin
            CCFixed := Boolean(Round(Params[p].Value));
          end else if Params[p].Name = 'TimeCh' then begin
            TimeCh  := Round(Params[p].Value);
          end;
        end;
      finally
        Params.Free;
        Params := nil;
      end;

      Inc(prDataSetCount);

      ResDS        := prResults.InsertClone(ResDS, DS, True);

      FirstChunk   := prChunks.Count;
      Seg          := DS.Find('Segment');
      ResSeg       := ResDS.Find('Segment');

      if not Seg.IsNull then begin

        //ExpConds
        Params := TQUB_ExpCond_Parameters.Create;
        Params.ReadTree( DS.Find('ExpCond') );
        for p := 0 to Params.Count - 1 do begin
          Params.AddPar('Cond_' + Params[p].Name, Params[p].Cond_Value, Params[p].Index, Params[p].IsVar, -1);
        end;

        //Chris, the Par.IsVar function is no longer calculated ! You eliminated it with the new chunking
        //It determines whether a Par changes across files when Par.Channel = -1 (else presumably does change)
        //Task_Mac.IsPZero is no longer reliable

        QVD         := TQUB_QVO_QVD(ViewportData.ListObjects.QVOByPath[DS['FileName'].DataAsString]);
        QDF         := QVD.DataObject;


        //**********************
        //1-point extra seg fix:
        PrevDat.First := -1;
        PrevDat.Last  := -1;
        //**********************


        while not Seg.IsNull do begin

          Dat                      := QUB_FirstLast(Seg.Data[0] , Seg.Data[1]);


          //************************
          //TO DO: LM
          {
          There is a little problem: when idealizing twice in the same data file (without erasing),
            IdlStim puts the last point of the last selection into a separate Seg, one point long.
            I'm adding some ad-hoc code to eliminate this extra segment,
            but the proper fix should be done on IdlStim (by Chris?... Please?...)
            Note that until the proper fix, this may remove genuine segments
          }
          if PrevDat.Last <> -1 then begin
            if (Dat.Last = Dat.First{OnePoint}) and (Dat.First = PrevDat.Last + 1) then begin
              Break;
            end;
          end;
          PrevDat := Dat;
          //************************

          Chunk                    := prNewChunk(Dat);

          MacChunk                 := TQUB_Mac_PChunk(Chunk);

          prChunks.Add(MacChunk);

          Chunk.StartTime          := Double(Seg['start'].Data[0]) * 1.0e-3;

          MacChunk.QVD             := QVD;
          MacChunk.QDF             := QDF;

          MacChunk.Sampling        := QDF.Sampling;
          MacChunk.Scaling         := QDF.Scaling * QVD.DataChannelScaling[QVD.ActiveChannel];
          MacChunk.Offset          := QVD.DataChannelOffset[QVD.ActiveChannel];

          MacChunk.Data            := nil;
          MacChunk.TimeData        := nil;
          MacChunk.FitPoint        := nil;
          MacChunk.HasTime         := TimeCh <> -1;
          if MacChunk.HasTime then begin
            MacChunk.TimeChIndex   := TimeCh;
            MacChunk.TimeScaling   := QDF.Scaling * QVD.DataChannelScaling[MacChunk.TimeChIndex];
            MacChunk.TimeOffset    := QVD.DataChannelOffset[MacChunk.TimeChIndex];
          end;
            
          //TO DO: this may unnecessarily fragment memory. Replace with allocation in chunks - there are examples in the code. LM.
          c                        := Length(Segments);
          SetLength(Segments, c+1);
          Segments[c].First        := Seg.Data[0];
          Segments[c].Last         := Seg.Data[1];


          MacChunk.MacModel        := nil;

          MacChunk.FileIndex       := FileIndex;

          MacChunk.ResultSeg       := ResSeg.GetNodeMem;

          //Fit/Exc lists
          MacChunk.FitListCount    := 0;
          MacChunk.FitList         := nil;
          MacChunk.FitListIdx      := 0;
          if prOptParams.OptFitDataMode = FITDATA_LIST then begin
            //If identical chunks, do it only  for the first chunk of this model, else do it for all
            if (prOptParams.OptIdenticalChunks = 0) or (FirstChunk = prChunks.Count) then begin
              DList := MacChunk.QVD.SelectionLists.ListByName[FitDataListName];
              if DList <> nil then begin
                try
                  DList.Lock;
                  //TO DO: LM.
                  //For the time being, do it brute force.
                  //In the (far) future, only allocate as many fit/exclusion selections as concern this Chunk
                  MacChunk.FitListCount := DList.Items.Count;
                  MacChunk.FitList      := PointAlloc1(MacChunk.FitListCount);
                  for s := 0 to DList.Items.Count - 1 do begin
                    MacChunk.FitList[s].x := DList.Selection[s].Sel.First - Chunk.Sel.First;
                    MacChunk.FitList[s].y := DList.Selection[s].Sel.Last  - Chunk.Sel.First;
                  end;
                finally
                  DList.Unlock;
                end;
              end;
            end;
          end;//FitData = List

          MacChunk.ExcListCount    := 0;
          MacChunk.ExcList         := nil;
          MacChunk.ExcListIdx      := 0;
          if prOptParams.OptExcData <> 0 then begin
            if (prOptParams.OptIdenticalChunks = 0) or (FirstChunk = prChunks.Count) then begin
              DList := MacChunk.QVD.SelectionLists.ListByName[ExcDataListName];
              if DList <> nil then begin
                try
                  DList.Lock;
                  MacChunk.ExcListCount := DList.Items.Count;
                  MacChunk.ExcList      := PointAlloc1(MacChunk.ExcListCount);
                  for s := 0 to DList.Items.Count - 1 do begin
                    MacChunk.ExcList[s].x := DList.Selection[s].Sel.First - Chunk.Sel.First;
                    MacChunk.ExcList[s].y := DList.Selection[s].Sel.Last  - Chunk.Sel.First;
                  end;
                finally
                  DList.Unlock;
                end;
              end;
            end;
          end;//ExcData = True

          Seg                      := Seg.SiblingSameName;
          ResSeg                   := ResSeg.SiblingSameName;

        end;//while not Seg.IsNull

        ParamClasses := TQUB_ExpCond_ParametersList.Create;
        prExpCondParams.Add(ParamClasses);
        prGetIdlStim(QVD, Params, @(Segments[0]), Length(Segments), FirstChunk, ParamClasses);

        Params.Free;
        //Params := nil;


        //Testing ************************************************************
        if prChunkThreadCount > 0 then begin
        //Multi-thread(processor) computation, must have one model per chunk, otherwise concurrent access will screw it up

          if (prOptParams.EstChannelCount = 1) and (prOptParams.EstChannelCountPerFile = 1) then begin
            if not CCFixed then begin
              Inc(prOptParams.ChannelCountIdx);
            end;
          end;

          for l := FirstChunk to prChunks.Count - 1 do begin

            //TO DO: LM. I'm not sure what happens to ChannelCount estimation. The grads must be summed up across all models of the same file
            //Must check !!!
            MacModel  := TQUB_Mac_Model.Create(Model.ModelObject, False{DoInitialize}, ParamClasses, @prOptParams, prCuda_Lib);
            prMacModels.Add(MacModel);

            if FitStatesCount > 0 then begin
              MacModel.HasFitP := True;
              for p := 0 to FitStatesCount - 1 do begin
                MacModel.FitP[FitStates[p]] := 1;
              end;
            end;

            MacChunk                   := TQUB_Mac_PChunk(prChunks[l]);
            MacChunk.MacModel          := MacModel;

          end;

        end else begin
        //Single thread

          MacModel  := TQUB_Mac_Model.Create(Model.ModelObject, False{DoInitialize}, ParamClasses, @prOptParams, prCuda_Lib);
          prMacModels.Add(MacModel);

          if FitStatesCount > 0 then begin
            MacModel.HasFitP := True;
            for p := 0 to FitStatesCount - 1 do begin
              MacModel.FitP[FitStates[p] - 1] := 1;
            end;
          end;

          if (prOptParams.EstChannelCount = 1) and (prOptParams.EstChannelCountPerFile = 1) then begin
            if not CCFixed then begin
              Inc(prOptParams.ChannelCountIdx);
            end;
          end;

          for l := FirstChunk to prChunks.Count - 1 do begin
            MacChunk                   := TQUB_Mac_PChunk(prChunks[l]);
            MacChunk.MacModel          := MacModel;
          end;

        end;
        //Testing*************************************************************


        SetLength(Segments, 0);

        Inc(FileIndex);

      end;

      DS           := DS.SiblingSameName;

    end;


    //Delayed initialization of models, due to EstChannelCountPerFile
    if (prOptParams.EstChannelCount = 1) and (prOptParams.EstChannelCountPerFile = 1) then begin
      //Messy but important
      for m := 0 to prMacModels.Count - 1 do begin
        if TQUB_Mac_Model(prMacModels[m]).ChannelCountIdx >= 0 then begin
          TQUB_Mac_Model(prMacModels[m]).cPars[2*TQUB_Mac_Model(prMacModels[m]).RateCount + TQUB_Mac_Model(prMacModels[m]).ExtraKinParCount + TQUB_Mac_Model(prMacModels[m]).ChannelCountIdx] := Ln(TQUB_Mac_Model(prMacModels[m]).CCount);
        end;
      end;
      for m := 0 to prMacModels.Count - 1 do begin
        for n := 0 to prMacModels.Count - 1 do begin
          if n <> m then begin
            if (TQUB_Mac_Model(prMacModels[m]).ChannelCountIdx >= 0) and (TQUB_Mac_Model(prMacModels[n]).ChannelCountIdx >= 0) then begin
              TQUB_Mac_Model(prMacModels[n]).cPars[2*TQUB_Mac_Model(prMacModels[n]).RateCount + TQUB_Mac_Model(prMacModels[m]).ExtraKinParCount + TQUB_Mac_Model(prMacModels[m]).ChannelCountIdx] := TQUB_Mac_Model(prMacModels[m]).cPars   [2*TQUB_Mac_Model(prMacModels[m]).RateCount + TQUB_Mac_Model(prMacModels[m]).ExtraKinParCount + TQUB_Mac_Model(prMacModels[m]).ChannelCountIdx];
            end;
          end;
        end;
      end;
    end;
    for m := 0 to prMacModels.Count - 1 do begin
      TQUB_Mac_Model(prMacModels[m]).Initialize;
    end;


    //**************************************************************************


    for c := 0 to prChunks.Count - 1 do begin

      Chunk               := prChunks[c];
      Chunk.Status        := CHUNK_STATUS_EMPTY;
      Chunk.Trials        := 0;
      Chunk.Index         := c;

    end;

    if prChunks.Count = 0 then begin

      Result     := True;
      prError    := True;

      if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin
        prErrorMsg := 'The file list is empty'; // (View menu -> Mac Exp Cond List)';
      end else begin
        prErrorMsg := 'No Data';
      end;
      prWikiTopic := 'Error:No_Data';

      Exit;

    end;


  finally


    FreeMem(FitStates);
    ParamsAreVar.Free;


  end;

end;

procedure TQUB_Task_Mac.prSetChunkParameters;
begin
  inherited;
  prChunkSize := 0;
  prChunkSegmentation := True;
end;

procedure TQUB_Task_Mac.prFreeChunk(Chunk :TQUB_Thread_PChunk);
var
  MacChunk :TQUB_Mac_PChunk;
begin

  MacChunk := TQUB_Mac_PChunk(Chunk);

  if MacChunk.Data <> nil then begin
    FreeMem(MacChunk.Data);
    MacChunk.Data := nil;
    FreeMem(MacChunk.TimeData);
    MacChunk.TimeData := nil;
    FreeMem(MacChunk.FitPoint);
    MacChunk.FitPoint := nil;
    FreeMem(MacChunk.ResampleIntervals);
    MacChunk.ResampleIntervals := nil;
    FreeMem(MacChunk.ResampleSamples);
    MacChunk.ResampleSamples := nil;
  end;

  FreeMem(MacChunk.FitCurve);
  MacChunk.FitCurve := nil;

  FreeMem(MacChunk.FitList);
  MacChunk.FitList := nil;
  MacChunk.FitListCount := 0;
  MacChunk.FitListIdx := 0;

  FreeMem(MacChunk.ExcList);
  MacChunk.ExcList := nil;
  MacChunk.ExcListCount := 0;
  MacChunk.ExcListIdx := 0;

  MacChunk.MacModels.Free;

  FreeMem(MacChunk);

end;

function TQUB_Task_Mac.prNewChunk(var Sel :TQUB_FirstLast) :TQUB_Thread_PChunk;
var
  i                :Integer;
begin

  Result           := AllocMem(SizeOf(TQUB_Mac_Chunk));

  Result.Sel       := Sel;

  TQUB_Mac_PChunk(Result).FitCurve := IAlloc1(Sel.Last - Sel.First + 1);
  for i := 0 to (Sel.Last - Sel.First) do begin
    TQUB_Mac_PChunk(Result).FitCurve[i] := 0;
  end;

  TQUB_Mac_PChunk(Result).MacModels := TList.Create;

end;

procedure TQUB_Task_Mac.prGetIdlStim(QVD :TQUB_QVO_QVD; ParamTemplate :TQUB_ExpCond_Parameters; Segments :TQUB_PAFirstLast; SegCount :Integer; FirstChunk :Integer; ParamClasses :TQUB_ExpCond_ParametersList);
var
  StimChannels     :TList;
  StimParams       :TPAInteger;
  StimChannelCount :Integer;
  i, p             :Integer;
  CondIdls         :TList;
  CondDS           :INode;
  Seg              :INode;

  procedure IdlStimToChunk(Seg :INode; Chunk :TQUB_Mac_PChunk);
  var
    Cls, Count     :TPAInteger;
    i              :Integer;
  begin
    Chunk.ExpCondIdlCount := Seg['DwellCount'].Data[0];
    Chunk.ExpCondIdl      := AllocMem(Chunk.ExpCondIdlCount * SizeOf(TQUB_QFS_QIF_NClass));
    Cls                   := TPAInteger(Seg['Classes'].DataPtr);
    Count                 := TPAInteger(Seg['Counts'].DataPtr);
    for i := 0 to Chunk.ExpCondIdlCount - 1 do begin
      Chunk.ExpCondIdl[i].Cls := Cls[i];
      Chunk.ExpCondIdl[i].N   := Count[i];
    end;
  end;
begin
  StimChannels     := nil;
  StimParams       := nil;
  CondIdls         := nil;

  try
    // find stimulus channels
    StimChannels     := TList.Create;
    GetMem(StimParams, QVD.DataObject.ADChannelCount * SizeOf(Integer));
    StimChannelCount := 0;
    for p := 0 to ParamTemplate.Count - 1 do begin
      if ParamTemplate[p].Channel >= 0 then begin
        if ParamTemplate[p].Channel >= QVD.DataObject.ADChannelCount then begin
          raise Exception.Create(Format('Data channel for ''%s'' is out of range (%d > %d)', [ParamTemplate[p].Name, ParamTemplate[p].Channel, QVD.DataObject.ADChannelCount-1]));
        end;
        StimChannels.Add(Pointer(ParamTemplate[p].Channel));;
        StimParams[StimChannelCount] := p;
        Inc(StimChannelCount);
      end;
    end;

    // make sure the condition channels are idealized
    prEnsureIdlStim(QVD, StimChannels, SegCount, Segments);

    // get multiplexed condition idealization
    CondIdls         := TList.Create;
    for i := 0 to StimChannels.Count - 1 do begin
      CondIdls.Add(QVD.DataObject.Idealization[Integer(StimChannels[i])]);
    end;
    CondDS           := MplexIdl(CondIdls, SegCount, Segments);

    // build ParamClasses
    MplexAmpsToParamClasses(CondDS['MplexAmps'], StimParams, ParamTemplate, ParamClasses);

    // copy idlcond to chunks
    i   := FirstChunk;
    Seg := CondDS.Find('Segment');
    while not Seg.IsNull do begin
      IdlStimToChunk(Seg, TQUB_Mac_PChunk(prChunks[i]));

      Inc(i);
      Seg := Seg.SiblingSameName;
    end;

  finally
    StimChannels.Free;
    FreeMem(StimParams);
    CondIdls.Free;
  end;
end;

procedure TQUB_Task_Mac.prEnsureIdlStim(QVD :TQUB_QVO_QVD; StimChannels :TList{of Integer}; SegCount :Integer; Segments :TQUB_PAFirstLast);
var
  i        :Integer;
  c        :Integer;
  s        :Integer;
  ef, el, ec :Integer;
  ea       :Double;
  Iter     :TQUB_IdlSegmentsIter;
  ReIdl    :Boolean;
  Multi    :Integer; // -1: none; >=0: only index seen; -2: multiple classes
begin
  i        := 0;
  while i < StimChannels.Count do begin
    c      := Integer(StimChannels[i]);
    ReIdl  := False;
    Multi  := -1;

    for s := 0 to SegCount - 1 do begin
      Iter := QVD.DataObject.Idealization[c].FindIter(Segments[s].First);
      while 0<>Iter.Valid do begin
        Iter.Read(@ef, @el, @ec, @ea);
        if ef > Segments[s].Last then Break;
        if ec < 0 then begin
          ReIdl := True;
          Break;
        end;
        if Multi = -1 then Multi := ec;
        if Multi <> ec then Multi := -2;
        Iter.Next;
      end;
      Iter.Free;
      if ReIdl then Break;
    end;

    if ReIdl then prSimpleIdlStim(QVD, c);
    if Multi >= 0 then begin
      // raise Exception.Create('Please run IdlStim on data channel '+IntToStr(Integer(StimChannels[i])));
      // but maybe they've just got a flat selection or extract, and they mean it?
      prConsoleMsg  := 'The stimulus on data channel '+IntToStr(Integer(StimChannels[i]))+' appears to be flat.  If this is not the case, please'#13#10'run IdlStim on data channel '+IntToStr(Integer(StimChannels[i]))+'.';
      prConsoleCode := MESSAGE_WARNING;
      Synchronize(prWriteNewConsole);
    end;


    Inc(i);
  end;

  QVD.ActiveChannel := ChannelIndex;
end;



procedure CleanupAmpForIdlStim(DataSet :INode; AbsMin :Double);
// replace 0 with +/- 1e-12
var
  Seg        :INode;
  Amp        :INode;
  i          :Integer;
  Sign       :Double;
begin
  Seg        := DataSet.Find('Segment');
  while not Seg.IsNull do begin
    Amp      := Seg['amp'];
    for i := 0 to Amp.DataCount - 1 do begin
      if Abs(Amp.Data[i]) < AbsMin then begin
        Sign := Amp.Data[(i+1) mod Amp.DataCount];
        Sign := Sign / Max(1e-12, Abs(Sign));
        Amp.Data[i] := Sign * AbsMin;
        Break;
      end;
    end;
    Seg      := Seg.SiblingSameName;
  end;
end;

procedure TQUB_Task_Mac.prSimpleIdlStim(QVD :TQUB_QVO_QVD; Ch :Integer);
var
  i        :Integer;
  Sel      :TQUB_FirstLast;
  Mdl      :INode;
  St1, St2 :INode;
  Result   :INode;
begin
  for i := 0 to prChunks.Count - 1 do begin
    Sel := TQUB_Mac_PChunk(prChunks[i]).Chunk.Sel;
    if(Sel.Last - Sel.First + 1) > 1000000 then begin
      raise Exception.Create('Before running Mac, please run IdlStim on A/D channel '+IntToStr(Ch));
    end;
  end;
  
  // build simple model
  Mdl      := TNode.Create('ModelFile');
  St1      := Mdl['States']['State'];
  St1['Class'].SetData(Integer(0));
  St1['Pr'].SetData(0.0/1.0);
  St2      := St1.Parent.AppendClone(St1, True);
  St2['Class'].SetData(Integer(1));
  Mdl['Amps'].SetupNumData(NODE_DATATYPE_DOUBLE, 2, 1);
  Mdl['Amps'].Data[0] := 0.0;
  Mdl['Amps'].Data[1] := 1.0; // they just have to be different
  Mdl['Stds'].SetupNumData(NODE_DATATYPE_DOUBLE, 2, 1);
  Mdl['NAr'].SetupNumData(NODE_DATATYPE_DOUBLE, 2, 1);
  St1      := Mdl['Rates'].AppendChild(Model.MakeNewRateNode);
  St1['States'].Data[0] := 0;
  St1['States'].Data[1] := 1;
  St1['k0'].Data[0]     := 1.0;
  St1['k0'].Data[1]     := 1.0;

  // run baum-welch + viterbi
  QVD.ActiveChannel := Ch;
  Result   := SimpleAmp(TNode.CreateNull, Mdl, True, True);

  // clean up and put idlzation in data
  CleanupAmpForIdlStim(Result['DataSet'], Properties['MinAbsStimulus'].Data[0]);
  prIdlStimQVD := QVD;
  prIdlStimCh  := Ch;
  prIdlStimDS  := Result['DataSet'];
  Synchronize(prIdlStimCommit);
  QVD.DrawIdealized := True;

      { rejected as obnoxious
      prIdlStimSignal    := TEvent.Create(nil, False, False, '');
      QVD.ActiveChannel := c;
      ViewportData.ListObjects.SelQVO(QVD);
      Synchronize(prStartIdlStim);
      while prIdlStimSignal.WaitFor(1000) = wrTimeout do begin end;
      prIdlStimSignal.Free;
      }
end;

procedure TQUB_Task_Mac.prIdlStimCommit;
begin
  prIdlStimQVD.DataObject.Idealization[prIdlStimCh].SetDwells(prIdlStimDS);
end;

procedure TQUB_Task_Mac.prStartIdlStim;
begin
  ActionsList.ByName['Preprocessing'].ActionByName['IdlStim'].OnRClick(prIdlStimDone);
end;

procedure TQUB_Task_Mac.prIdlStimDone(Sender :TObject);
begin
  prIdlStimSignal.SetEvent;
end;

//******************************************************************************
procedure TQUB_Task_Mac.prSetModel(Model :TQUB_QVO_QVM);
begin
  prModel := Model;
  if Model <> nil then begin
    if (prOptParams.OptSegments = OPTSEGS_GROUP) or (prOptParams.OptSegments = OPTSEGS_GROUPLIST) then begin
      prOptSegsOldMdl := TQUB_QFS_QMF.Create('', nil);
      prOptSegsOldMdl.CopyFrom(prModel.ModelObject);
    end;
    prModel.LockAny;
  end;
end;


//****************************************************************************//
//                                                                            //
//                            P R O P E R T I E S                             //
//                                                                            //
//****************************************************************************//
procedure TQUB_Task_Mac.OnGetProperties(Sender :TObject);
begin
  with Sender as TQUB_Task_Mac_PropertiesForm do begin

    Self.ChannelIndex           := Properties['ChannelIndex'].Data[0];

    //Data Processed
    Self.DataProcessed          := StringToDataProcessed(Properties['ProcessData'].DataAsString);
    Self.ProcessFilter          := 0<>Properties['ProcessData']['Filter'].Data[0];
    Self.ProcessFilterFreq      := Properties['ProcessData']['FilterFreq'].Data[0];

    //Client/Server
    Self.ThreadCount            := Properties['ThreadCount'].Data[0];
    Self.Remote                 := 0<>Properties['Remote'].Data[0];

    Self.QuietOutput            := QuietOutput;

    Self.OptParams              := Params;

    Self.FitDataListName        := lstFitData.Items.Strings[FitDataListIdx];
    Self.ExcDataListName        := lstExcData.Items.Strings[ExcDataListIdx];

    prGroupList                 := txtOptSegmentsGroupList.Text;

    Self.Properties             := prResults.AppendClone(Properties, True);

    prSkipSlowChange            := 0<>Properties['SkipSlowChange'].Data[0];
    prSkipSlowStd               := Properties['SkipSlowStd'].Data[0];

  end;//with
  if OptParams.OptIdenticalChunks <> 0 then ThreadCount := 1;
end;

procedure TQUB_Task_Mac.OnPutProperties(Sender :TObject);
begin
  //with Sender as TQUB_Task_Mac_PropertiesForm do begin
  //  ChannelIndex                := Self.ChannelIndex;
  //end;//with
end;


function TQUB_Task_Mac.DataCallBack(MacChunk :TQUB_Mac_PChunk) :Integer;
var
  BaseChunk :TQUB_Thread_PChunk;
  TScale    :Double;
  TOff      :Double;
  i         :Integer;
begin

  Result := 0;

  TScale := 1;
  TOff   := 0;

  try

    BaseChunk := TQUB_Thread_PChunk(MacChunk);

    if MacChunk.Data = nil then begin
    //Allocate

      MacChunk.Data := AllocMem(SizeOf(Integer) * (BaseChunk.Sel.Last - BaseChunk.Sel.First + 1));
      if MacChunk.HasTime then begin
        MacChunk.TimeData := AllocMem(SizeOf(Single) * (BaseChunk.Sel.Last - BaseChunk.Sel.First + 1));
        TScale := 1 / MacChunk.TimeScaling;
        TOff   := MacChunk.TimeOffset;
      end;
      MacChunk.FitPoint := AllocMem(SizeOf(Boolean) * (BaseChunk.Sel.Last - BaseChunk.Sel.First + 1));

      //TO DO: LM
      //Chris, this is bad!!! The DataSource must be local and frozen during the task execution.
      //Apparently, no answer from Chris.
      if QUB_MainForm.DataSource = DATA_SOURCE_FILELIST then begin

        //case DataProcessed of
          //DATA_PROCESSED_NONE:
            begin
              MacChunk.QDF.ReadData(MacChunk.Data, False{Raw}, ChannelIndex, BaseChunk.Sel.First, BaseChunk.Sel.Last);
              if MacChunk.HasTime then begin
                MacChunk.QDF.ReadData(TPAInteger(MacChunk.TimeData), False{Raw}, MacChunk.TimeChIndex, BaseChunk.Sel.First, BaseChunk.Sel.Last);
              end;
            end;
        //end;

      end else begin

        case DataProcessed of
          DATA_PROCESSED_NONE:
            begin
              with MacChunk.QVD do LoadDataChannel(MacChunk.Data, ChannelIndex, BaseChunk.Sel, False{Process}, False, False);
              if MacChunk.HasTime then begin
                with MacChunk.QVD do LoadDataChannel(TPAInteger(MacChunk.TimeData), MacChunk.TimeChIndex, BaseChunk.Sel, False{Process}, False, False);
              end;
            end;
          DATA_PROCESSED_ASDISPLAYED:
            //use QVD's properties
            begin
              with MacChunk.QVD do LoadDataChannelProcess(MacChunk.Data, ChannelIndex, BaseChunk.Sel, FilterOn[ChannelIndex], FilterFreq[ChannelIndex], BaselineOn[ChannelIndex]);
              if MacChunk.HasTime then begin
                with MacChunk.QVD do LoadDataChannelProcess(TPAInteger(MacChunk.TimeData), MacChunk.TimeChIndex, BaseChunk.Sel, FilterOn[ChannelIndex], FilterFreq[ChannelIndex], BaselineOn[ChannelIndex]);
              end;
            end;
          DATA_PROCESSED_ASSUCH:
            begin
              with MacChunk.QVD do LoadDataChannelProcess(MacChunk.Data, ChannelIndex, BaseChunk.Sel, ProcessFilter, ProcessFilterFreq, ProcessBaseline);
              if MacChunk.HasTime then begin
                with MacChunk.QVD do LoadDataChannelProcess(TPAInteger(MacChunk.TimeData), MacChunk.TimeChIndex, BaseChunk.Sel, ProcessFilter, ProcessFilterFreq, ProcessBaseline);
              end;
            end;
        end;

      end;

      if MacChunk.HasTime then begin
        for i := BaseChunk.Sel.First - BaseChunk.Sel.First to BaseChunk.Sel.Last - BaseChunk.Sel.First do begin
          MacChunk.TimeData[i] := TPAInteger(MacChunk.TimeData)[i] * TScale + TOff;
        end;
      end;

      
      //Do the FitPoint list
      prFitPointsOfChunk(MacChunk);

      if prSkipSlowChange and (MacChunk.ResampleSamples = nil)
      then prSetupResampling(MacChunk);


    end else begin
    //DeAllocate

      FreeMem(MacChunk.Data);
      MacChunk.Data := nil;
      FreeMem(MacChunk.TimeData);
      MacChunk.TimeData := nil;
      FreeMem(MacChunk.FitPoint);
      MacChunk.FitPoint := nil;
      FreeMem(MacChunk.ResampleIntervals);
      MacChunk.ResampleIntervals := nil;
      FreeMem(MacChunk.ResampleSamples);
      MacChunk.ResampleSamples := nil;

    end;

  except

    Result := 1;

  end;

end;

(*
function ExpCondEndPoints(ECIdl :TQUB_QFS_QIF_PANClass; Count :Integer; var EndCount :Integer): TPAInteger;
var
  i, j, At :Integer;
begin
  Result   := AllocMem(2 * Count * SizeOf(Integer));
  At       := 0;
  j        := 0;
  for i := 0 to Count - 1 do begin
    Result[j] := At;
    Inc(j);
    if ECIdl[i].N > 1 then begin
      Result[j] := At + ECIdl[i].N - 1;
      Inc(j);
    end;
    At            := At + EcIdl[i].N;
  end;
  EndCount := j;
end;
*)

function InterleaveIndices(A1 :TPAInteger; A1Count :Integer;
                           A2 :TPAInteger; A2Count :Integer;
                           var Count :Integer)
             :TPAInteger;
var
  i1, i2     :Integer;
begin
  Result     := AllocMem((A1Count + A2Count) * SizeOf(Integer));
  Count      := 0;

  i1         := 0;
  i2         := 0;

  while (i1 < A1Count) or (i2 < A2Count) do begin
    if (i1 < A1Count) then begin
      if (i2 < A2Count) then begin
        if A1[i1] = A2[i2] then begin
          Result[Count] := A1[i1];
          Inc(i1);
          Inc(i2);
        end else if A1[i1] < A2[i2] then begin
          Result[Count] := A1[i1];
          Inc(i1);
        end else begin
          Result[Count] := A2[i2];
          Inc(i2);
        end;
      end else begin
        Result[Count] := A1[i1];
        Inc(i1);
      end;
    end else begin
      Result[Count] := A2[i2];
      Inc(i2);
    end;
    Inc(Count);
  end;

  ReallocMem(Result, Count*SizeOf(Integer));
end;

procedure TQUB_Task_Mac.prSetupResampling(MacChunk :TQUB_Mac_PChunk);
var
  MaxStd               :Double;
  dN                   :Integer;
begin
  MaxStd               := prSkipSlowStd * MacChunk.QVD.Scaling * MacChunk.QVD.DataChannelScaling[ChannelIndex];
  dN                   := MacChunk.Chunk.Sel.Last - MacChunk.Chunk.Sel.First + 1;
  AdaptiveResample2(MacChunk.Data, dN, MacChunk.ExpCondIdl, MacChunk.ExpCondIdlCount, MaxStd,
                    MacChunk.ResampleCount, MacChunk.ResampleIntervals, MacChunk.ResampleSamples);
end;

function TQUB_Task_Mac.prGetOptPParams: TQUB_Mac_Opt_PParams;
begin
  Result := @prOptParams;
end;


//******************************************************************************
constructor TQUB_Task_Mac_ChunkThread.Create(Suspended :Boolean);
begin
  FreeOnTerminate := True;
  inherited;
  prReady   := True;
  MacResult := False;
end;

procedure TQUB_Task_Mac_ChunkThread.Execute;
var
  c :THandle;
begin
  c := GetCurrentThread;
  SetThreadAffinityMask(c, Affinity);
  repeat
    if MacChunkIdx <> -1 then begin
      prReady   := False;
      Synchronize(LLGradOfChunk);//MacResult := MacTask.LLGradOfChunk(MacChunkIdx);
      MacChunkIdx := -1;
    end;
    prReady := True;
    MacTask.Signal.SetEvent;
    Suspend;
  until(Terminated);
end;

procedure TQUB_Task_Mac_ChunkThread.LLGradOfChunk;
var
  LL :Double;
begin
  MacResult := MacTask.LLGradOfChunk(MacChunkIdx, True{CompGrad}, False{NumGrad}, LL);
end;


//******************************************************************************
end.
